
#ifndef _elfapi_h_
#define _elfapi_h_

#include <stdio.h>
#include <setjmp.h>

#ifndef DLL_EXPORT
#if AXMS
#define DLL_EXPORT __declspec( dllexport )
#else
#define DLL_EXPORT
#endif
#endif

#ifndef DLL_IMPORT
#if AX_DLL
#define DLL_IMPORT __declspec( dllimport )
#else
#define DLL_IMPORT
#endif
#endif


#if !ELF_DATA_DONE
typedef struct data_any *elfData;
#define ELF_DATA_DONE 1
#endif

#undef AXP_

#ifdef AX_USE_PROTOS
#define AXP_(body) body
#else
#define AXP_(body) ()
#endif

#define AX_CALL_MODE_SS_DSP_FLAG  0x01  /* use this flag in the call mode 
                                        *  if function needs to be displayed
                                        *  in the SS functions box */       

#define AX_FUNC_VOLATILE_ON_DEMAND	0x02	/* used to define individual 
											calculation for external 
											functions. on demand NOT USED */

#define AX_FUNC_VOLATILE			0x04	/* used to define individual 
											calculation for external 
											functions. always NOT USED */

/* call modes for the addin functions */
#define AX_CALL_MODE_NORMAL     0x00    /*  ERROR and NA values are sent in
                                            as NULLs */

#define AX_CALL_MODE_SS         0x10    /*  ERROR and NA values are sent in 
                                            as binary objects */

typedef struct AxCallInfo_t {   /* C-Library dynamic linkage info */
    char *category; /* func type ... "financial", "math" ... */
    elfData (*func) AXP_((elfData data));   /* the C routine to call */
    char *alias;    /* name to use in Applixware (usually same as name) */
    char *argStr;   /* describes function args (per declaration) */
    int callMode;   /* AX_CALL_MODE_XX | AX_CALL_MODE_SS_DSP_FLAG */
} AxCallInfo_t;


#ifndef wchar
#if BUILD_WIDE
/* build the product using wide characters */
#define wchar unsigned short
#define mbchar char
#else
#define wchar unsigned char
#define mbchar char
#endif
#endif

#define int32 int

#define ELFC_INIT 0     /* Automatically called once during server init */
#define ELFC_EXIT -1    /* Reserved convention for shutdown sequence */
#define ELFC_AXNET -2   /* Allows C servers to look like ELF<->ELF servers */
#define ELFC_PORT -3    /* Return port number from (e.g. from /etc/services) */
#define ELFC_HS1 -4     /* Private handshake support for axmain */
#define ELFC_HS2 -5     /* Private handshake support for axmain */
#define ELFC_HS3 -6     /* Private handshake support for axmain */
#define ELFC_GET_PID -7     /* For axmain to request pid of child process */
#define ELFC_OGET -8    /* Reserved for Builder->C interface */
#define ELFC_OSET -9    /* Reserved for Builder->C interface */
#define ELFC_PACK_DATA -10	/* pack ELF socket data */

#define ELFSVR_OK        0
#define ELFSVR_ERROR    -1
#define ELFSVR_NOCHANNEL 1    /* obsolete */
#define ELFSVR_NOCLIENT  2

#ifdef __cplusplus
#define NO_MANGLE_NAMES 1
#endif

#if NO_MANGLE_NAMES
extern "C" {       /* make sure names to not get "mangled" */
#endif

int     AxInitializeBuilderObject AXP_((char *bldrObjectName, char *host, char **errStr));
elfData AxCallBuilderObjectMethod AXP_((int id, char *methodName, elfData args, int *error, char **errStr));
int	    AxTerminateBuilderObject  AXP_((int id));

DLL_EXPORT AxCallInfo_t *AxGetCallInfo AXP_((void));
AxCallInfo_t *AxGetCallInfo_ud_1 AXP_((void));
AxCallInfo_t	*AxGetCallInfo_ud_2 AXP_((void));
AxCallInfo_t *AxGetCallInfo_ud_3 AXP_((void));
AxCallInfo_t	*AxGetCallInfo_ud_4 AXP_((void));
AxCallInfo_t *AxGetCallInfo_ud_5 AXP_((void));

int AxIsArray    AXP_((elfData data));
int AxArraySize  AXP_((elfData data));
int AxIsBinary   AXP_((elfData data));
int AxBinarySize AXP_((elfData data));
int AxIsInt      AXP_((elfData data));
int AxIsNumber   AXP_((elfData data));
int AxIsFloat    AXP_((elfData data));
int AxIsString   AXP_((elfData data));
int AxIsObject   AXP_((elfData data));
int AxCanBeString AXP_((elfData data));	/* return TRUE if convertable to string */


#if !eldataHinclude
elfData  AxAddArrayToArray AXP_((elfData ap,int dex,elfData p));
elfData  AxAddBoolToArray AXP_((elfData ap,int dex,int n));
elfData  AxAddDataToArray AXP_((elfData ap,int dex,elfData p));
elfData  AxAddFloatToArray AXP_((elfData ap,int dex, double dbl));
elfData AxAddBinaryToArray AXP_((elfData ap,int dex,int dataSize,void *binaryData));
elfData  AxAddIntToArray AXP_((elfData ap,int dex,int n));
elfData  AxAddStrToArray AXP_((elfData ap,int dex, mbchar *str));
elfData  AxAddWideStrToArray AXP_((elfData ap, int dex, wchar *str));
elfData AxAppendArray AXP_((elfData ap1, elfData ap2 ));
elfData AxMakeDblData AXP_((double num));
elfData AxMakeNumData AXP_((int num));
elfData AxMakeObjectData AXP_((int objdex,int objuid));
elfData AxMakeWideStrData AXP_((int len, wchar *str));
elfData AxMemRead AXP_((char *bufP));
elfData AxTaskCopyElfData AXP_((int task, elfData aP));
void AxTaskFreeElfData AXP_((int task, elfData aP));

elfData AxWriteArray AXP_((elfData ap,int dex,elfData new_val));
void AxOverwriteArray AXP_((elfData ap,int dex,elfData new_val));
elfData ElfGetMsgData AXP_((void));
int AxListWalkerInit AXP_((elfData aP));
elfData AxListWalkerNext AXP_(( void ));

elfData  AxArrayFromArray AXP_((elfData ap,int dex));
int      AxBinaryFromArray AXP_((elfData data,int dex,char **binP));
int      AxBinaryFromDataPtr AXP_((elfData data, char **binP));
int      AxBoolFromArray AXP_((elfData data,int dex));
mbchar    *AxStrFromArray AXP_((elfData ap,int dex));
mbchar    *AxStrFromDataPtr AXP_((elfData ap));
mbchar    *AxStrPtrFromArray AXP_((elfData ap,int dex));
double AxDblFromArray AXP_((elfData ap,int dex));
int      AxIntFromArray AXP_((elfData aP,int dex));
double   AxFloatFromArray AXP_((elfData aP,int dex));
wchar *AxWideStrPtrFromArray AXP_((elfData ap, int dex));

elfData *AxArrayDataAddr AXP_((elfData data));
elfData  AxArrayElement AXP_((elfData ap,int dex));
char    *AxBinaryBytesPtr AXP_((elfData data));
int      AxBoolFromDataPtr AXP_((elfData aP));
wchar *AxWideStrPtrFromDataPtr AXP_(( elfData ap));
elfData  AxBuildArray();
elfData  AxCopyData AXP_((elfData data));
void     AxDecomposeArray();
void     AxError AXP_((int code, char *str, char *obj));
int      AxIntFromDataPtr AXP_((elfData aP));
double   AxFloatFromDataPtr AXP_((elfData ap));
void     AxFree AXP_(( char *p ));
void     AxFreeData AXP_((elfData p));
elfData  AxTruncArray AXP_((elfData p, int newsize));
void     AxFreeArrayButNotContents AXP_((elfData p));
elfData  AxMakeArray AXP_((int elements));
elfData  AxMakeBinaryData AXP_((int elements, char *dataBytes));
elfData  AxMakeFloatData AXP_((double val));
elfData  AxMakeIntData AXP_((int num));
elfData  AxMakeBoolData AXP_((int num));
elfData  AxMakeStrData AXP_((int len, mbchar *str));
char    *AxMalloc AXP_((int size));
char *AxRealloc AXP_((char *oldP, int new_size));
void     AxMallocAborter AXP_((jmp_buf abortBuf));
mbchar    *AxStrPtrFromDataPtr AXP_((elfData aP));
char    *TrueCalloc AXP_(( int n, int size ));
char    *TrueMalloc AXP_(( int size ));
char    *TrueRealloc AXP_(( char *p, int size ));
void    *AxInitCube AXP_(( elfData , int *, int *, int *));
void    AxFreeCube AXP_(( void *));
elfData  AxCubeElement AXP_((void *, int, int, int));
elfData  AxSSMakeNaData AXP_((void));
elfData  AxSSMakeErrorData AXP_((void));
int      AxSSIsNa AXP_((elfData));
int      AxSSIsError AXP_((elfData));

elfData AxDispatch AXP_((int funcId, elfData funcData, int32 *quitNow));



/*
The following function is not used in standard ELF->C RPC programs
and shared libraries but may be required in certain C->ELF programs and
non-standard ELF->C RPC programs.
*/
extern int      AxMemWrite AXP_((elfData data,int makeRoom,char **buf));

/* The following function is not for customer use */
extern elfData  AxMemRead AXP_((char*buf));

#endif

/*
The following functions are not used in standard ELF->C RPC programs
and shared libraries but may be required in certain C->ELF programs and
non-standard ELF->C RPC programs.
*/
extern int      AxOpenServer AXP_((char *sockname,int forkIt));
extern int      AxSockTop AXP_((int fd));
extern elfData  AxRPCRead AXP_((char *bufP));

extern void     AxClientDisConnect AXP_((int channel));
extern int      AxNetConnect AXP_((char *hostNameIn,
                                  int *code, char **str, char **obj));
extern elfData  AxNetElfChannelCall AXP_((int channel,char *funcname,
                                  elfData args,int *code,char **str,char **obj));
extern int      AxNetElfServiceConnect AXP_((char *host,char *service,
                                  int *code,char **str,char **obj));
extern void     AxPrintfDatum AXP_((elfData data));
extern void		AxDumpDatum AXP_((elfData datum, FILE *fp));
extern void		AxFileDatum AXP_((char *filename, elfData datum));
extern char *	AxTempFileDatum AXP_((elfData datum));
extern void     AxSendMessageToServer AXP_((int channel, int command, elfData data,
                                  int *code,char **str,char **obj));
extern elfData  AxGetResponseFromServer AXP_((int channel,
                                   int *codeP, char **strP, char **objP));

extern int      AxNetProcSlave AXP_((char *hostName,elfData argvP,
                                   int *code,char **str,char **obj));

typedef struct {
	unsigned io_packed:1;
} axio_channel_t_flags;

typedef struct axio_channel_t {
	int freed;
	void *calledInfo;	/* Used by callback functions, can be anything */
	int sock_fd;		/* accept on this... */
	int chan_fd;		/* read/write on this */
	int reading;	/* If true, post a read */
	int accepting;	/* If true, accept connections rather than read */
	int writing;	/* If true, write Buffer has something */
	int threadOnAccept;	/* If true, a new thread is started after the accept */
	int threadUid;		/* non-repeating thread id */
	int threadDex;		/* repeating, zero-based thread id */
	char *readBuf;		/* malloc'd buffer */
	int readBufSize;	/* current buffer (malloc'd) size */
	int readPos;		/* pos in readBuf data will get written to */
	int readOffset;		/* ptr past already processed data in readBuf */
	int readNeed;		/* accumulate these many bytes before callbacking again */
	int (*readFunc)();	/* func to call on read selects wakeups */
	void (*acceptFunc)();/* func to call on accept selects wakeups */
	struct axio_channel_t *socket;	/* parent of accepted channel */
	char *writeBuf;		/* malloc'd buffer */
	int writeBufSize;	/* current buffer (malloc'd) size */
	int writePos;		/* pos in writeBuf data will get sent from */
	int writeCnt;		/* how many bytes left to write (from writePos) */
	int closeOnWrite;	/* If TRUE, will shutown when writeCnt==0 */
	void (*writeNtfy)();	/* called after each write and on connection closed */
	int channelClosed;	/* may be set true after read/write */
	void (*closeNtfy)();/* called on connection close for read and write */
	char *clientIP;		/* client IP address */
	int validated;		/* TRUE for AA protocol clients only */
	void (*aaAccepted) AXP_((struct axio_channel_t *)); /* for AA protocol clients */
	void (*aaAcceptNotify) AXP_((struct axio_channel_t *)); /* for AA protocol clients */
	void (*aaDispatch) AXP_((struct axio_channel_t *,elfData)); /* for AA protocol clients */
	void (*aaClosed) AXP_((struct axio_channel_t *)); /* for AA protocol clients */
	axio_channel_t_flags axio_flags;	/* bit flags for channel attributes */
	struct axio_channel_t *next;	/* threader or NULL */
} axio_channel_t;

extern axio_channel_t *AxIORegisterService AXP_((int channel,
		void (*onAccept)(axio_channel_t *, struct sockaddr_in *)));
#if 1
axio_channel_t *AxIOAcceptClient (
    axio_channel_t *chanP,
    int (*readFunc)(axio_channel_t *,char *,int),
    void (*writFunc)(axio_channel_t *,char *, int),
    void (*closFunc)(axio_channel_t *));
#else
axio_channel_t *AxIOAcceptClient AXP_((axio_channel_t *chanP,
					int (*readFunc)(axio_channel_t *, char *, int),
					void (*writFunc)(axio_channel_t *, char *, int),
					void (*closFunc)(axio_channel_t *)));
#endif
extern axio_channel_t *AxIOAddDuplexChannel AXP_((int channel,
					int (*readFunc)(axio_channel_t *, char *, int),
					void (*writFunc)(axio_channel_t *, char *, int),
					void (*closFunc)(axio_channel_t *)));
extern void AxIOSetClientInfo AXP_((axio_channel_t *chanP, void *anyData));
extern void *AxIOGetClientInfo AXP_((axio_channel_t *chanP));
extern int AxIOIsClosing AXP_((axio_channel_t *chanP));
extern void AxIOCloseChannel AXP_((axio_channel_t *chanP));
extern void AxIOHackShutdownForEntsvr AXP_((axio_channel_t *chanP)); /* TEMP */
extern void AxIORejectClient AXP_((axio_channel_t *chanP, char *msgBuf, int msgLen));
extern int AxIOSelect AXP_((struct timeval *timeP));
extern int AxIOSelectOne AXP_((struct timeval *timeP));
extern void AxIOWriteSome AXP_((void));
extern axio_channel_t *AxIOAddConnectionInfo AXP_((void));
extern void AxIOAddToWriteBuffer AXP_((axio_channel_t *chanP, char *buf, int cnt, int closeOnWrite));
extern int AxIOMakeChannelNonBlocking AXP_((int channel));
extern int AxIOMakeChannelBlocking AXP_((int channel));
extern void AxIOWrapUp AXP_((void));
extern int AxIOElfReadFromClient AXP_((axio_channel_t *chanP, char *buf, int cnt));
extern void AxIOElfSvrAccept AXP_((axio_channel_t *chanP, struct sockaddr_in *inP));
extern void AxIOSetAcceptFunc AXP_((axio_channel_t *chanP,
				void (*acceptFunc)(axio_channel_t *,struct sockaddr_in*)));
extern int AxIOIsElfMsgComplete (char *buf, int cnt);
extern void AxIODispatchElfMsg AXP_((axio_channel_t *chanP, elfData aP));
extern elfData AxDispatchMsg AXP_((axio_channel_t *chanP, int funcId, elfData funcData));
extern elfData AxReadElfDatumFromBuf AXP_((char *buf));
extern elfData AxReadElfFile AXP_((char *filename));
extern int AxChoosePort AXP_((void));
extern int AxMemWritePacked AXP_((elfData , int, char**, int));
extern void EncryptPassword AXP_((char *in_password, char *encrypt_pwd));
extern void GetClearPassword AXP_((char *encrypt_pwd, char *in_password));

/* Not for customer use */
extern elfData  AxNetCall AXP_((char *host, char *service, elfData args,
                         char *passwd, int *code, char **str, char **obj));
extern elfData  AxNetServices AXP_((char *hostName, char *qualifier,
                                    int *code, char **str, char **obj));
extern int AxClientConnect AXP_((int socknum,char *sockname,char *passed_hostname,int *code,char **str,char **obj));
extern int AxnetNoResponse;

#ifndef ELF_LOADED
#define data_ptr elfData	/* mapping for internal/external use */
#endif

#if NO_MANGLE_NAMES
} /* extern "C" to prevent names from being mangled */
#endif

#endif /* !_elfapi_h_ */
