/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: acopy1.c
   Applix ASC version: 3
   Date: Sun Jan 24 21:13:16 1999
*/

#include <am_aos.h>
#include <sys/types.h>
#include <sys/stat.h>

#define UT_CP_SAME 0
#define UT_CP_MODE 1
#define UT_CP_ALLPRIV 2

#define MDBLK 8192

/*************************************************************************
function: ACopyOneFile - Create a Copy of a File
Returns TRUE/FALSE on success/failure.
*/
int ACopyOneFile (code,src,dst,mode)
register int code;              /* operation type */
register char *src, *dst;       /* source & destination filenames */
register int mode;              /* optional new file mode */
{
    register int fd, fd2, num_read, num_writ;
    char buf[MDBLK];
    struct stat file_info;

/* set the mode based on passed code */

    if ( code == UT_CP_SAME ) {
        if ( 0 != Astat(src,&file_info) )
            return(FALSE);
        mode = file_info.st_mode;
    }
    else if ( code == UT_CP_ALLPRIV )
        mode = 0777;
    else if ( code != UT_CP_MODE )
        abort();

/* Perform the copy */

    fd = open(src,0);
    if ( fd == -1 )
        return(FALSE);
    fd2 = creat(dst,0777);
    if ( fd2 == -1 ) {
        close(fd);
        return(FALSE);
    }
    while ( num_read = read(fd,buf,MDBLK) ) {
        if ( (num_read == -1) || (num_read != write(fd2,buf,num_read)) ) {
            close(fd);
            close(fd2);
            Adelete(dst);
            return(FALSE);
        }
    }
    close(fd);
    close(fd2);
    return(TRUE);
}

