/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: am_bld.h
   Applix ASC version: 77
   Date: Thu May 20 21:13:16 1999
*/

#include <stdio.h>
#include <am_aos.h>

#define ErrStream stdout

#define ID unsigned short

#define MALLOC_TEST 0
#if MALLOC_TEST
char *MyMalloc();
char *MyRealloc();
#define free(x) MyFree(x)
#define malloc(x) MyMalloc( (unsigned)(x) )
#define realloc(p,x) MyRealloc( p,(unsigned)(x) )
#define calloc(x,s) MyCalloc( (unsigned)(x),(unsigned)(s) )
#endif

#include <malloc.h>

#ifdef strcmp   /* do not use Istrcmp */
#undef strcmp
#endif

#ifdef assert
#undef assert
#endif

#if SYS&VMSSYS
#define assert(ex) {if (!(ex)) {printf("Bye!\n");exit(3);}}
#else
#if AXMS
#define assert(ex) {if (!(ex)){fprintf(stderr,"Assertion failed: file %s, line %d\n", __FILE__, __LINE__);exit(1);}}
#else
#define assert(ex) {if (!(ex)){fprintf(stderr,"Assertion failed: file %s, line %d\n", __FILE__, __LINE__);abort();}}
#endif
#endif

extern int InputLine;           /* input line number */
char *AMakePath();

#define TYPE_NODE 1         /* module node (struct) */
#define TYPE_CMLVAR 2       /* stored command line variable (str) */
#define TYPE_CTRLFILE 3     /* name of a control file (str) */
#define TYPE_DIR 4          /* a dircetory object (struct) */
#define TYPE_COMP 5         /* a component object (struct) */
#define TYPE_STR 6          /* a string object (str) */
#define TYPE_ID 7           /* an ID (ID) */
#define TYPE_TAPE 8         /* tape generation info (struct) */
#define TYPE_UNUSED 9   /* always highest */

#define OUTPUT_NODE(p) (p->ruleid && (p->ruleid != DUMMY_RULE))

#define MAX_LOGICAL_DIRS 500    /* .defdirs */
#define MAX_COMPONENTS 1000
#define MAX_BUILD_FILES 1000
#define MAX_APPVARS 2000
#define MAX_LOCDIRS 4000
#define MAX_OUTDIRS 100
#define MAX_INCLIST 20
#define MAX_VARNAME 32      /* length of variable names */
#define MAX_DIR_USERS 20
#define MAX_VARS 400
#define MAX_LINELEN 32000
#define COMMENT_CHAR ';'
#define CONTINUE_CHAR '\\'
#define TAB 9
#define VAR_FLAG '$'
#define AM_HASH_PRIME 1011

/* prototype creation levels */
#define PROTO_NONE 0        /* no prototypes are generated              */
#define PROTO_FUNCDEF 1     /* int main();                              */
#define PROTO_PROTOTYPE 2   /* int main(int, char*[]);                  */
#define PROTO_PROTOFULL 3   /* int main(int argc, char *argv[]);        */
#define PROTO_MACRO 4       /* int main P_((int argc, char *argv[]));   */

#if (SYS&B4_2) || (SYS&VMSSYS)
#define FORK vfork
#else
#define FORK fork
#endif

#if AXMS
#define OEXT ".obj"
#define EEXT ".exe"
#define LEXT ".lib"
#define XEXT ".obj"
#define AEXT ".obj"
#define PCHEXT ".pch"  /* pre-compiled header suffix */
#else
#define OEXT ".o"   /* ?? should be .o when all under amake */
#define EEXT NULL
#define LEXT NULL   /* ?? should be .a when all under amake */
#define XEXT NULL   /* ?? should be .o when all under amake */
#define AEXT NULL   /* ?? should be .o when all under amake */
#define PCHEXT ".pch"
#endif


#if AXMS
#define TEMPNAME "\\tmp"
#else
#define TEMPNAME "/tmp"
#endif  /* !MS */

/* locdir/checksir AttsDirs encoding */

#define DIRAT_LOC1 1      /* non-recursive locdir */
#define DIRAT_CHK  2      /* is a checkdir versus locdir */

#define streq(a,b) (!strcmp(a,b))

#ifdef DEBUG
#undef DEBUG
#endif
#define CASE break; case
#define DEFAULT break; default

#define DEBUG(body) {if ( Debug ) DebugLog body; }
#define LOGGER(body) {if ( Logging ) LogLog body; }
#define TRACE(body) {if ( Trace ) TraceLog  body; }
#define TELL(body) {if ( !Quiet ) TellLog body; }
#define CMPTELL(body){if (!Quiet) TellComponent(nodeP),TellLog body; }
#define VERBOSE(body) {if ( Verbose ) VerboseLog body; }
#define COMMANDLOG(body) {if ( CommandLog ) VerboseLog body; }
#define bld_assert(x,body) { if (!(x)) aborter body; }
#define bld_verify(x,body) { if (!(x)) ErrorMsg body; }

/* special object numbers */

#define HASH_ARR_ID 1       /* Node Hash Table */
#define RULE_ARR_ID 2       /* Ruledir array */

typedef struct {
    char *name;         /* name of this rule */
    int (*func)();      /* the function to call */
	char tool_based;	/* true if first input actually a tool name */
    char multi_inputs;  /* accepts multiple inputs */
    char multi_outputs; /* accepst multiple outputs */
    char allow;         /* TRUE if should allow rule to execute */
    char ok_with;       /* accepts 'with' */
    char matched_io;    /* input/output one-for-one */
    char dep_srch;      /* true if can dependency search */
    char **dep_cmd;     /* additional command based dependencies */
    char *extension;    /* system overriding extension or NULL */
} dispatch_rule;

/*
    Source nodes (no rule) have a varying extent name and no data after the
    name.
*/

#define MAX_NODENAME 128     /* name of an output node only */

typedef struct node {
    unsigned long dtmod;        /* timestamp */
    ID next;                    /* next in hash */
    ID dirid;                 /* containing directory */
	ID baseDirId;	/* if file .checkdir'd, this is base locdir of file */
    ID comp_id;        /* component object id */
    unsigned up_to_date : 1;    /* TRUE if node built */
    unsigned exists : 1;        /* TRUE if file exists */
    unsigned resolving: 1; /* TRUE while building a node (for circs) */
    unsigned busted : 1;        /* TRUE if build failed for this */
    unsigned dep_searching : 1;/* TRUE while dependency searching (for circs) */
    unsigned deps_done : 1;/* TRUE if have completed dependency checking for node */
    unsigned call_deps_done : 1;/* TRUE if have completed dependency checking for node */
    unsigned no_deps : 1;   /* TRUE if should suppress dependency checking */
    unsigned am_used : 1;       /* TRUE if this is input/dep to a rule */
    unsigned relocated : 1;     /* TRUE if node was .relocated */
    unsigned located : 1;      /* TRUE if node was .locdir/.checkdir */
    unsigned moved : 1;         /* TRUE if src moved since last build */
    unsigned conflict : 1;      /* TRUE if multiple of this exist */
    unsigned refers_to_moved : 1;  /* TRUE if uses a moved module */
    unsigned is_global : 1; /* relocated or .globaled */
    unsigned in_opensrc : 1;   /* TRUE if used in opensrc directory */
    unsigned user_deps : 1; /* TRUE if user supplied deps in .amk file */
    unsigned has_siblings : 1;  /* TRUE if first of multi-out rule */
    unsigned cproto_excluded : 1; /* TRUE if should NOT prototype */
    unsigned strict_excluded : 1; /* TRUE if should NOT use CC to compile */
    unsigned cpyright_excluded : 1; /* TRUE if binary or not ours */
    unsigned opensrc_excluded : 1; /* TRUE absolutely prevents copying this file */
	unsigned cpyright_nowarn : 1;
	unsigned opensrc_builtfile : 1;/* Use derived rather than base input */
    unsigned haveProtoed : 1; /* TRUE if have gen'd prototypes this pass */
    unsigned char ruleid;       /* rule to use to build (0 == input file) */
    char name[MAX_NODENAME+1]; /* varying extent if input node */
    ID withstr;        /* string object */
    short incount;     /* how many inputs there are */
    short depcount;    /* how many depends there are */
    short callcount;   /* how many external function calls there are */
    ID inputs[1];      /* the inputs/depends ids, varying extent */
} node;

typedef struct {
    long dtmod;
    char name[STRPATH+1];
} ctrlfile_info;

typedef struct {        /* component info */
    ID dflt;            /* default output directory */
    ID person;          /* person's name - for future use */
	int num_opensrc_lines;
	int used_opensrc_lines;
	char **opensrc_lines;
    char cname[1];      /* varying extent */
} comp_info;

typedef struct {
    short priority;     /* priority number for -I ordering */
    ID owner;           /* component that owns it (or 0 for global) */
/*    ID users[MAX_DIR_USERS];    * components that can see into it */
    ID defdirname;   /* first defdir associated with this dir */
    unsigned recursive : 1; /* for amcache, true if not loc1dir or check1dir */
    unsigned is_locdir : 1;
    unsigned is_checkdir : 1;
    unsigned loc_visited :1;
    unsigned is_outdir : 1;
    unsigned is_altdir : 1;
    unsigned is_defdir : 1;
	unsigned is_shlib_dir : 1;
	unsigned is_opensrc: 1;
/*    unsigned is_unique : 1; */
    long dtmod;
    char dname[1];      /* varying extent */
} dir_info;

typedef struct {
    char *name;     /* the text of the name (lower case) */
    char *protoString;  /* full cproto string */
    unsigned int time;  /* time proto was created */
	char *fromFile;		/* filename of reference */
    ID id;              /* id reference for node arrays */
    int inCurrentFile;  /* prevents multi-refs */
} callInfo;
extern callInfo **CallsById;    /* callInfo.id links to calls info */

extern dispatch_rule RuleArray[];

/* Aos vars (macro'd for compatability to older rules files) */

#define AosRts FindDefine("AOS_RTS")
#define AosDir FindDefine("AOSDIR")
#define AosUnix FindDefine("AOS_UNIX")
#define AosMachine FindDefine("AOS_MACHINE")
#define AosCompiler FindDefine("AOS_COMPILER")
#define AosCompilerMT FindDefine("AOS_COMPILER_MT")
#define AosAnsiCompiler FindDefine("AOS_ANSI_COMPILER")
#define AosCppCompiler FindDefine("AOS_COMPILER_CPP")
#define AosCppCompilerMT FindDefine("AOS_COMPILER_CPP_MT")
#define AosCCflags FindDefine("AOS_CCFLAGS")
#define AosCPPflags FindDefine("AOS_CPPFLAGS")
#define AosOptimize FindDefine("AOS_OPTIMIZE")
#define AosLinker FindDefine("AOS_LINKER")
#define AosLinkerCpp FindDefine("AOS_LINKER_C_PLUS")
#define AosLdFlags FindDefine("AOS_LDFLAGS")
#define AosLibs FindDefine("AOS_LIBS")
#define AosLibraryBuilder FindDefine("AOS_LIB_BUILDER")
#define AosFinal FindDefine("AOS_FINAL")
#define AosFinalMT FindDefine("AOS_FINAL_MT")
#define AosArchiverCPlus FindDefine("AOS_AR_C_PLUS")
#define AosAnsiFinal FindDefine("AOS_ANSI_FINAL")
#define AosFinalCPlus FindDefine("AOS_FINAL_C_PLUS")
#define AosFinalCPlusMT FindDefine("AOS_FINAL_C_PLUS_MT")
#define AosFinalCPlusSHMT FindDefine("AOS_FINAL_C_PLUS_SHMT")
#define AosFnlflags FindDefine("AOS_FNLFLAGS")
#define AosGuiflags FindDefine("AOS_GUIFLAGS")
#define AosDebug FindDefine("AOS_DEBUG")
#define AosDebugFlag FindDefine("AOS_DEBUG_FLAG")
#define AosShlibCCFlags FindDefine("AOS_SHLIB_CC_FLAGS")
#define AosShlibLDFlags FindDefine("AOS_SHLIB_LD_FLAGS")
#define AosShlibCCFinalFlags FindDefine("AOS_SHLIB_CC_FINAL_FLAGS")
#define AosShlibLdCppFlags FindDefine("AOS_SHLIB_LD_CPP_FLAGS")
#define AosShlibLibs FindDefine("AOS_SHLIB_LIBS")
#define AosStrictCompiler AosCppCompiler
#define AosStrictCompilerMT AosCppCompilerMT

/* Globals */

extern ID *HashArray;   /* Node hash array */
extern int Quiet;       /* Error messages only */
extern int Verbose;     /* log commands while doing them */
extern int CommandLog;	/* Log command lines only */
extern int Trace;           /* tell about state */
extern int Logging;     /* more telling */
extern int Debug;           /* most telling */
extern char *ModsToBuild[]; /* module names to build */
extern char ModsToForce[];  /* modules to force build */
extern int BuildModsCnt;    /* how many to build */
extern int BuildPhase;      /* TRUE while building */
extern char *ErrorBuf;      /* Buffer containing error info */
extern int CreateDependencies;  /* TRUE if should depend scan on compilation */
extern int CompList[];      /* array of component ids */
extern int NumComps;        /* number of defined components */
extern char CacheLocateFile[256];   /* created locator file */
extern char CacheBuildFile[256];    /* created build file */
extern int NeedNewBuildFile;    /* TRUE if should regenerate cached buildfile */
extern int CreatedDependencies; /* TRUE if new dependencies found */
extern int NumNodes;                /* total number of nodes */
extern char *TempDir;  /* directory to put tempfiles */
extern char *LastIncs[];        /* list of .last_include's */
extern int NumLastIncs;
extern int NumAppsToBuild;      /* How many -app apps */
extern char *AppNamesToBuild[100];  /* Their names */
extern int AppIdsToBuild[100];      /* Their comp number */
extern int DoingReport;    /* send errors to a report file */
extern FILE *ReportFp;     /* File to send errors to */
extern int TotalErrors;
extern char StdOutFile[STRPATH+1];
extern char StdErrFile[STRPATH+1];  /* for reports & outfile commands */
extern char ReportFile[STRPATH+1];
extern int BuildPhase;     /* TRUE while building */
extern ID NodeIdLooked;    /* set by LookupNode */

/* Common routines */

#if !ALPHA
#define sprintf MySprintf
char *MySprintf();
#endif
char *MyMalloc();
node *CheckForNode ();   /* find node given name */
node *LookupNode();     /* find/create node given name */
node *LookupOwnedNode();  /* find node given name in owner's dir only */
char *GetComponent();   /* get component name from node ptr */
char *GetCompStr();     /* get component name from comp_id */
long GetSimpleDtmod();  /* simplified getter of date/time modified for file */
char *AllcStr();        /* Malloc a string */
char *GenErrnoMsg();    /* Return an 'errno' message */
char *ExtractOutputs();
char *NodeDirName();    /* Return dirname node associated with */
char *DefaultDirName(); /* default directory of a node */
node *NodeInVal();  /* Return a node's input value from index */
ID NodeInId();  /* Return a node's input value from index */
node *NodeDepVal();  /* Return a node's dependency value from index */
char *NodeWithStr(); /* Return a node's 'with' string */
ID *AddrOfInVals(); /* Inputs array address in node */
ID *AddrOfDepVals();    /* Depends array address in node */
ID *AddrOfCallVals();   /* Calls array in node */
char *FindEntryName();  /* return ptr to filename in pathname */
ID LookupDir();         /* given a dir name, return object id */
ID LookupIdForMe();     /* get a visible node */
node *LookupForMe();     /* get a visible node */
ID FillInNode();        /* Create/Fill node */
char *FindDefine();     /* Find var, abort if not set */
char *Pstreq();     /* partial strcmp, rtn's ptr to next token (or NULL) */
char *DosName();    /* given node, return Dos entryname if there is one */

/* object manager calls */

ID ObjStr();            /* Allocate a string object */
ID AllocObj();          /* allocate an object */
char *StrPtr();         /* get ptr of string */
node *NodePtr();        /* Return node ptr from ID */
ID *IdPtr();            /* Return ID ptr */
char *ComVarPtr();      /* get ptr to command line var string */
dir_info *DirPtr();     /* directory info ptr from ID */
comp_info *CompPtr();   /* component info ptr from ID */
ctrlfile_info *CtrlFilePtr(); /* control file struct from ID */
char *CmlVarPtr();      /* command line var ptr from ID */

/******************* RULES INFO ***************************************/

#define DUMMY_RULE 1
#define TRUE_NOP_RULE 2
#define NOP_RULE 3
#define NUM_RULES 67

/* Need to declare each rule's routine here because otherwise the
    Rules table cannot compile */

int dbg_compile_mt(), mkshared_cpp_mt(), mkshared_cc_mt();
int dbg_compile(), nodbg_compile(), cc_exec(), same_source(), same_file();
int default_link(), build_browser();
int ss_strgen(), ss_datagen(), archiver(), archiver_cplus(), ld_plink();
int namelist(), cat_sources(), cmacs(), yacc(), run_tool(), run_tool_loc();
int aroff(), hcmprs(), rtest(), dolint(), lintlib(), lex(), strfl_incl();
int strfl_file(), bldint(), deltemps(), cpp(), assemble();
int elf_compile(), archive_extract(), copy_file(), bldr_compile();
int link_or_copy(), elf_portable(),untar(),sqlgate();
int true_nop(), tape_copy(), nop_rule(), do_translate(), ld_mkshared();
int ld_mksharedcpp(), cpp_compile(), cpp_compile_mt(), cpp_precompile(), cplus_exec(), mkshared_cc();
int mkshared_libcpp_mt();
int make_import_lib(), untarcompressed();
int ansi_compile(), ansi_exec(), noopt_compile(), tarfile();
int mkshared_cpp(), java_compile(), java_archive(), sub_amake();
int ld_mksharedcpp_mt(), genshlibintf();
int cc_exec_mt(), cplus_exec_mt();

#ifdef AM_DCL_RULES

/* If a rule has implicit dependencies, they need to be called out here.
    The general form is a null terminated array of strings */
 
char *AD02[] = {"axstrfl","remove_TR",NULL};
char *AD03[] = {"ss_strgen",NULL};
char *AD04[] = {"axstrfl","ss_datagen",NULL};
char *AD05[] = {"namelist",NULL};
char *AD06[] = {"cmacs",NULL};
char *AD07[] = {"aroff",NULL};
char *AD08[] = {"hcompress",NULL};
char *AD09[] = {"sgenincl",NULL};
char *AD10[] = {"sgenfile",NULL};
char *AD11[] = {"bldint",NULL};
char *AD12[] = {"axstrfl","elfc",NULL};
char *AD13[] = {"remove_TR",NULL};
char *AD14[] = {"*JDK_NAME",NULL};
char *AD15[] = {"head_off",NULL};
char *AD16[] = {"elfmain","databld.am",NULL};

dispatch_rule RuleArray[NUM_RULES+1] = {
/*
                                   +-- 1st arg is tool name
                                   |  +-- multiple inputs ok
                                   |  |  + multiple outputs ok
                                   |  |  |  +-- allowed
                                   |  |  |  |  +-- with options ok
                                   |  |  |  |  |  +-- matched i/o
                                   |  |  |  |  |  |  +-- dep srch
  name             routine         |  |  |  |  |  |  |    + auto deps
     |                 |           |  |  |  |  |  |  |    |    +-- rule extension
     |                 |           |  |  |  |  |  |  |    |    |  */
{"NOT_HERE",       true_nop,       0, 0, 0, 1, 0, 0, 0, NULL, NULL  },
{"DUMMY",          true_nop,       0, 0, 0, 0, 0, 0, 1, NULL, NULL  },
{"TRUE_NOP",       true_nop,       0, 0, 0, 0, 0, 0, 0, NULL, NULL  },
{"nop",            nop_rule,       0, 1, 0, 0, 0, 0, 0, NULL, NULL  },
{"compile",        dbg_compile,    0, 1, 1, 1, 1, 1, 1, NULL, OEXT  },
{"compile_mt",     dbg_compile_mt, 0, 1, 1, 1, 1, 1, 1, NULL, OEXT  },
{"nodebug_compile",nodbg_compile,  0, 1, 1, 1, 1, 1, 1, NULL, OEXT  },
{"nooptimized_compile",noopt_compile,
                                   0, 1, 1, 1, 1, 1, 1, NULL, OEXT  },
{"shlib_compile",  mkshared_cc,    0, 1, 1, 1, 1, 1, 1, NULL, OEXT  },
{"shlib_compile_mt",  mkshared_cc_mt,
                                   0, 1, 1, 1, 1, 1, 1, NULL, OEXT  },
{"shlib_cpp_compile", mkshared_cpp,0, 1, 1, 1, 1, 1, 1, NULL, OEXT  },
{"shlib_cpp_compile_mt", mkshared_cpp_mt,
                                   0, 1, 1, 1, 1, 1, 1, NULL, OEXT  },
{"assemble",       assemble,       0, 1, 1, 1, 1, 1, 1, NULL, AEXT  },
{"final_link",     cc_exec,        0, 1, 0, 1, 1, 0, 0, NULL, EEXT  },
{"final_link_mt",  cc_exec_mt,     0, 1, 0, 1, 1, 0, 0, NULL, EEXT  },
{"default_link",   default_link,   0, 1, 0, 1, 1, 0, 0, NULL, EEXT  },
{"build_browser",  build_browser,  0, 1, 0, 1, 1, 0, 0, NULL, NULL  },
#if AXMS
{"partial_link",   ld_plink,       0, 1, 0, 1, 1, 0, 0, NULL, LEXT  },
{"mkshared_lib",   ld_mkshared,    0, 1, 1, 1, 1, 0, 0, NULL, NULL  },
{"mkshared_libcpp",   ld_mksharedcpp,
                                   0, 1, 1, 1, 1, 0, 0, NULL, NULL  },
{"mkshared_libcpp_mt",ld_mksharedcpp_mt,
                                   0, 1, 1, 1, 1, 0, 0, NULL, NULL  },
#else
{"partial_link",   ld_plink,       0, 1, 0, 1, 1, 0, 0, NULL, XEXT  },
{"mkshared_lib",   ld_mkshared,    0, 1, 0, 1, 1, 0, 0, NULL, LEXT  },
{"mkshared_libcpp",   ld_mksharedcpp,
                                   0, 1, 0, 1, 1, 0, 0, NULL, LEXT  },
{"mkshared_libcpp_mt",ld_mksharedcpp_mt,
                                   0, 1, 0, 1, 1, 0, 0, NULL, LEXT  },
#endif
{"archive",        archiver,       0, 1, 0, 1, 0, 0, 0, NULL, LEXT  },
{"archiver_cplus", archiver_cplus, 0, 1, 0, 1, 0, 0, 0, NULL, LEXT  },
{"same_source",    same_file,      0, 1, 1, 1, 0, 1, 0, NULL, NULL  },
{"same_file",      same_file,      0, 1, 1, 1, 0, 1, 0, NULL, NULL  },
{"ss_strgen",      ss_strgen,      0, 0, 1, 1, 1, 0, 0, AD03, NULL  },
{"ss_datagen",     ss_datagen,     0, 1, 0, 1, 0, 0, 0, AD04, NULL  },
{"namelist",       namelist,       0, 1, 0, 0, 0, 0, 0, AD05, NULL  },
{"cat_sources",    cat_sources,    0, 1, 0, 1, 0, 0, 0, NULL, NULL  },
{"cmacs",          cmacs,          0, 1, 1, 1, 1, 1, 1, AD06, NULL  },
{"yacc",           yacc,           0, 0, 1, 1, 1, 0, 0, NULL, NULL  },
{"aroff",          aroff,          0, 0, 0, 1, 1, 0, 0, AD07, NULL  },
{"hcompress",      hcmprs,         0, 1, 0, 1, 1, 0, 0, AD08, NULL  },
{"lint",           dolint,         0, 1, 0, 0, 1, 0, 1, NULL, NULL  },
{"lintlib",        lintlib,        0, 1, 1, 0, 1, 0, 1, NULL, NULL  },
{"test",           rtest,          0, 1, 1, 1, 1, 0, 0, NULL, NULL  },
{"lex",            lex,            0, 0, 0, 1, 0, 0, 1, NULL, NULL  },
{"bldint",         bldint,         0, 0, 0, 1, 0, 0, 0, AD11, NULL  },
{"strfl_genincl",  strfl_incl,     0, 1, 1, 1, 1, 1, 0, AD09, NULL  },
{"strfl_genfile",  strfl_file,     0, 0, 0, 1, 1, 0, 0, AD10, NULL  },
{"cpp",            cpp,            0, 1, 1, 1, 1, 1, 1, NULL, NULL  },
{"elf_compile",    elf_compile,    0, 1, 1, 1, 0, 1, 1, AD12, NULL  },
{"elf_portable",   elf_portable,   0, 1, 1, 1, 0, 1, 1, AD12, NULL  },
{"archive_extract",archive_extract,0, 0, 1, 1, 1, 0, 0, NULL, NULL  },
{"copy_file",      copy_file,      0, 1, 1, 1, 0, 1, 0, NULL, NULL  },
{"run_tool",       run_tool,       1, 1, 1, 1, 1, 0, 0, NULL, NULL  },
{"run_tool_loc",   run_tool_loc,   1, 1, 1, 1, 1, 0, 0, NULL, NULL  },
{"run_mytool",     run_tool,       1, 1, 1, 0, 1, 0, 0, NULL, NULL  },
{"link_or_copy",   link_or_copy,   0, 1, 1, 1, 0, 1, 0, NULL, NULL  },
{"translate",      do_translate,   0, 1, 1, 1, 1, 1, 1, AD13, NULL},
{"untar",          untar,          0, 0, 1, 0, 1, 0, 0, NULL, NULL  },
{"sqlgate",        sqlgate,        0, 1, 0, 0, 1, 0, 1, NULL, NULL  },
{"cpp_compile",    cpp_compile,    0, 1, 1, 1, 1, 1, 1, NULL, OEXT  },    
{"cpp_compile_mt", cpp_compile_mt, 0, 1, 1, 1, 1, 1, 1, NULL, OEXT  },    
{"cpp_precompile",cpp_precompile,  0, 1, 1, 1, 1, 1, 1, NULL, NULL  },    
{"make_import_lib",make_import_lib,0, 1, 0, 1, 1, 0, 0, NULL, NULL  },
{"untarcompressed",untarcompressed,0, 0, 1, 0, 1, 0, 0, AD15, NULL  },
{"final_link_c++",  cplus_exec,    0, 1, 1, 1, 1, 0, 0, NULL, EEXT  },
{"final_link_c++_mt",cplus_exec_mt,0, 1, 1, 1, 1, 0, 0, NULL, EEXT  },
{"ansi_compile",    ansi_compile,  0, 1, 1, 1, 1, 1, 1, NULL, OEXT  },
{"ansi_final_link", ansi_exec,     0, 1, 0, 1, 1, 0, 0, NULL, EEXT  },
{"java_compile",   java_compile,   0, 1, 1, 1, 1, 0, 0, AD14, NULL  },
{"java_archive",   java_archive,   0, 1, 0, 1, 1, 0, 0, AD14, NULL  },
{"sub_amake",sub_amake,            0, 1, 0, 1, 1, 0, 0, NULL, NULL  },
{"genshlibintf",   genshlibintf,   0, 1, 0, 1, 1, 0, 0, NULL, NULL  },
{"tarfile",        tarfile,        0, 1, 0, 1, 1, 0, 0, NULL, NULL  },
{"bldr_compile",   bldr_compile,   0, 1, 1, 1, 1, 0, 1, AD16, NULL  },
{"",               0,              0, 0, 0, 1, 0, 0, 0, NULL, NULL  }
};

/*
The DUMMY_RULE does not have the allow flag primarily for use with
lintlib. That it is not auto-allowed is harmless since the allow
flag is overridden if the node is really needed.
*/
#ifdef AM_RULE_STUBS

/* Rule Stubs for table to resolve linking */

cc_compile(){} cc_exec(){} same_source(){} same_file(){}
ss_strgen(){} ss_datagen(){} archiver(){} archiver_cplus(){} ld_plink(){}
namelist(){} cat_sources(){} cmacs(){} yacc(){} run_tool(){}
aroff(){} hcmprs(){} rtest(){} dolint(){} lintlib(){} lex(){} strfl_incl(){}
strfl_file(){} bldint(){} deltemps(){} cpp(){} atm_tbl(){} assemble(){}
elf_compile(){} xpndMacs(){} archive_extract(){} copy_file(){}
kcompile(){} klink(){} link_or_copy(){} mkshared_cc(){}
true_nop(){} ld_mkshared(){} ld_mksharedcpp(){} untar(){} sqlgate(){} cpp_compile(){}
make_import_lib(){} cplus_exec(){} noopt_compile(){} mkshared_cpp(){} cpp_compile_mt{}()
ld_mksharedcpp_mt(){}
#endif

#endif  /* of AM_DCL_RULES */
