/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: am_rules.c
   Applix ASC version: 139.45
   Date: Wed May 19 21:14:58 1999
*/

#include <string.h>    /* for stricmp */

#define AM_DCL_RULES
#include <am_bld.h>

#define BUMP(p) (p+=strlen(p))

#define ND_INM0 (NodeInVal(nodeP,0)->name)
#define ND_IVAL0 NodeInVal(nodeP,0)
#define ND_IVAL(x) NodeInVal(nodeP,x)
#define ND_INCOUNT NodeInCount(nodeP)

char *getenv();

#define GIVE_UP(body) { ErrorMsg body; return(3); }

char ThisCommand[64000];

char *strstr();
/*char *strrchr();
char *strchr();
*/

extern int DoingProfileBuild;
extern int SkipGeneratingProtos;
extern int LimitedDebug;
extern char *DirSlash, *MyBuildTree;
extern char *TempDir;
extern int CprotoForcing;       /* TRUE if ModsToForce set by -fcp */
extern char *CProtoMacro;   /* macro name for cproto */

char *readFileIntoBuf();
char *GetVar();
char *NeedVar();
char *NeedFullVar();
int Arename();
char *NodeSysSuffix();
node *CrtOrFindNode();
char *NodeName();
char *GetCProtoDirName();
int DoingLink;
extern int PretendBuild;
extern char *SrcRelativeToDst();	/* defined in am_utils.c */

static char *ProtoNames[] = {
    "devsrc",   "devsrc",
    "pf",       "pfproto",
    "efs",      "efsproto",
    "ax",       "axfuncs",
    "ge",       "geproto",
    "rts",      "rtsproto",
    "axatmrt",  "atmproto",
    "gefilter", "geflfunc",
    "ss",       "ssproto",
    "axelf",    "elfproto",
    "axhelp",   "helpfunc",
    "tools",    "toolfunc",
    "axnlm",    "nlmproto",
    "utils",    "utilfunc",
    "axtutor",  "tutrfunc",
    "wp",       "wpfuncs",
    "biff",     "bifffunc",
    NULL
};

/*************************************************************************
function: rtest - invoke rules test program
*/
rtest (nodeP)
node *nodeP;
{
    SetNodeDir(nodeP);
    return(PipeProc("pipetest",nodeP->name));
}

/*************************************************************************
function: dolint - run lint
*/
dolint (nodeP)
node *nodeP;
{
    char outfile[300];
    char *cmdP;
    char *dirs[100];    /* list of -I dirs */
    int ix;
    extern int AllowDeletes;
    int code;
    register char *name;
    char *dirname;
    node *childP;
    char elfrt_str[30]; 
    char *elfrtP;

    elfrtP = NeedVar("ELFRT_ONLY");
    if ( (*elfrtP == '\0') || (strcmp(elfrtP, "0") == 0 ) )
       /* normal build, do not include switch */
       elfrt_str[0] = '\0';
    else
       strcpy( elfrt_str, "-DELFRT_ONLY=1" );

    NodeFilename(nodeP,outfile,TRUE);   /* absolute result file */
    SetNodeDir(ND_IVAL0);   /* optimal command line length */

    cmdP = ThisCommand;
    sprintf(cmdP,"lint -D%s -DDEBUG=%s -DSYS=%s %s %s %s %s ",
              AosMachine,AosDebug,AosUnix,NeedVar("BLD_FLAGS"),
              NeedVar("AOS_FLAGS"),
              elfrt_str, NodeWithStr(nodeP));
    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, FALSE);
    strcat(cmdP," ");
    BUMP(cmdP);

   /* Gen list of input files" */
    dirname = NodeDirName(ND_IVAL0);
    for ( ix = 0; ix < ND_INCOUNT; ix++ ) {
        *cmdP++ = ' ';
        childP = NodeInVal(nodeP,ix);
        name = childP->name;
        if ( NodeDirName(childP) == dirname )
            strcpy(cmdP,name);
        else
            NodeFilename(childP,cmdP,TRUE);
        BUMP(cmdP);
    }
    AllowDeletes = FALSE;
    return(PipeProc(ThisCommand,outfile));
}

/*************************************************************************
function: lintlib - run lint to get a lint library

Note: This actually runs a shell script called "lint_pass1" because lint
      is too stupid to be able to put its output in an arbitrary place.
      The arguments to the shell script are:
      lint_pass1 <full_path_to_source> <input_filename_prefix> \
                            <output_file_prefix> <-D's and includes>

      for example:
      lint_pass1 /ax/pgms/axelf/elintr.c elintr eeintr -DSUN4S -I/ax/pgms/includes ...

  A sample shell script which moves all the output to "/ax/axobj/lint",
  ,for system-V lint where a "-c" produces a ".ln" lint library is:
:
FULLPATH=$1
shift
INPUTFILE=$1
shift
OUTPUTFILE=$1
shift
lint -c $* $FULLPATH 1>$INPUTFILE.pass1  2>&1
#
rm -f /ax/axobj/lint/$OUTPUTFILE.pass1
cp $INPUTFILE.pass1 /ax/axobj/lint/$OUTPUTFILE.pass1
rm $INPUTFILE.pass1
#
rm -f /ax/axobj/$OUTPUTFILE.ln
cp $INPUTFILE.ln /ax/axobj/lint/$OUTPUTFILE.ln
rm $INPUTFILE.ln


*/
int lintlib (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *dirs[100];    /* list of -I dirs */
    int ix;
    extern int AllowDeletes;
    int code;
    char *p;
    char *withstr;
    char elfrt_str[30]; 
    char *elfrtP;
    char srcfile[200], dstfile[200];
    char base_input_name[50], base_output_name[50];
 
    elfrtP = NeedVar("ELFRT_ONLY");
    if ( (*elfrtP == '\0') || (strcmp(elfrtP, "0") == 0 ) )
       /* normal build, do not include switch */
       elfrt_str[0] = '\0';
    else
       strcpy( elfrt_str, "-DELFRT_ONLY=1" );

    src = ND_IVAL0;
    NodeFilename(src,srcfile,TRUE);
    ReSuffix(base_input_name, FindEntryName(srcfile), "" );
    
    NodeFilename(nodeP,dstfile,TRUE);
    SetNodeDir(nodeP);

    withstr = NodeWithStr(nodeP);

    /** nodeP->name is the output file, with no path */
    ReSuffix( base_output_name, nodeP->name, "");

    CMPTELL(("lint %s \n", srcfile ));
    cmdP = ThisCommand;
    sprintf(cmdP,"lint_pass1 %s %s %s -D%s -DDEBUG=%s %s -DSYS=%s %s %s %s %s",
             srcfile, base_input_name, base_output_name,
             AosMachine, AosDebug, AosCCflags, AosUnix,
             NeedVar("AOS_FLAGS"),
             NeedVar("BLD_FLAGS"),
             elfrt_str, withstr);

    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, FALSE);

    code = proc(ThisCommand, TRUE);
    return( 0 );
}

/*************************************************************************
function: cpp - Run the C preprocessor (acpp)
*/
cpp (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char srcfile[300];
    char *p;
    char *acpp_file[300];

    NodeFilename(LookupNode("acpp"),acpp_file,TRUE);
    src = ND_IVAL0;
    SetNodeDir(nodeP);
    cmdP = ThisCommand;
    sprintf(cmdP,"%s %s %s -DDEBUG=%s -D%s -DSYS=%s %s ",
            acpp_file, NeedVar("BLD_FLAGS"),
            NeedVar("AOS_FLAGS"),
            AosDebug,AosMachine,AosUnix,NodeWithStr(nodeP));
    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, TRUE);
    strcat(cmdP," ");
    BUMP(cmdP);
    NodeFilename(src,cmdP,TRUE);
    BUMP(cmdP);
    strcat(cmdP," ");
    strcat(cmdP,NodeName(nodeP));
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function: assemble - assmble assembly code
*/
assemble (nodeP)
node *nodeP;
{
    node *src;
    char srcfile[300];
    char *asname;

    src = ND_IVAL0;
    SetNodeDir(nodeP);
    NodeFilename(src,srcfile,TRUE);
    asname = FindDefine("AOS_ASSEMBLER");
#if AXMS
//    ReSuffix(tstname, NodeName(nodeP), OEXT);
    sprintf(ThisCommand,"%s %s %s %s NUL.LST NUL.CRF",asname,NodeWithStr(nodeP),srcfile,NodeName(nodeP));
#else
    sprintf(ThisCommand,"%s %s %s -o %s",asname,NodeWithStr(nodeP),srcfile,NodeName(nodeP));
#endif
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function:  cpp_precompile - precompile a ".cxx" into a .pch pre-compiled
           header file.
*/
 
cpp_precompile (nodeP)
node* nodeP;
{
    char *cmdP;
    node *src;
    char srcfile[STRPATH+1];
    char dstfile[STRPATH+1];
    char tstname[STRLONG+1];
    int code;
    char *p;
    char *ccflags;
    char asrcfile[STRPATH+1];
    char *sysname, *str1;
    char *GetComponent();
    char debug_flag[250];
    char optimize_flag[4];
    char *compiler;
    static char DebugFlag[250];
    char elfrt_str[30];
    char *elfrtP;
 
    compiler = AosCppCompiler;

    if ( DebugFlag[0] == '\0' )
        {
        /* initialize the DebugFlag name */
        if ( IsDebugBuild() )
           {
           strcpy( DebugFlag, AosDebugFlag);
           }
        else
           strcpy( DebugFlag, " " );
        }
 
    src = ND_IVAL0;
    NodeFilename(src,srcfile,TRUE);
    NodeFilename(nodeP,dstfile,TRUE);
    SetNodeDir(nodeP);

    if ( LimitedDebug ) {
       if ( file_in_debug_list( srcfile) ) {
           /* turn on debug mode for this file */
           strcpy( debug_flag, DebugFlag );
           optimize_flag[0] = '\0';      /* turn off optimization */
       }
       else  {
           debug_flag[0] = '\0';  /* turn off debug info */
           strcpy( optimize_flag, AosOptimize );
       }
    }
    else {
       /* do the normal build thing */
       strcpy( debug_flag, DebugFlag );
       strcpy( optimize_flag, AosOptimize );
    }
 
    sysname = AosUnix;

    cmdP = ThisCommand;

    sprintf(cmdP,
        "%s -c %s %s %s %s -DDEBUG=%s -D%s -DSYS=%s %s -Yc -Fp%s ",
            compiler, debug_flag,NeedVar("BLD_FLAGS"),
            NeedVar("AOS_FLAGS"),
            optimize_flag,AosDebug,AosMachine,sysname,
            NodeWithStr(nodeP), dstfile);
 
    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, FALSE);
    strcat(cmdP," ");
    BUMP(cmdP);
    GenProtosForMe(nodeP);
#if AXMS
    strcat( cmdP, "-Tp" );    /* force compiler to treat file as C++ file */
    BUMP(cmdP);
#endif
    strcpy(cmdP,srcfile);
 
    code = proc(ThisCommand, TRUE);
    return(code);
}

cpp_compile(nodeP) 
node* nodeP;
{
	return(do_cpp_compile(nodeP, FALSE));
}

cpp_compile_mt(nodeP) 
node* nodeP;
{
	return(do_cpp_compile(nodeP, TRUE));
}


/*************************************************************************
function:  cpp_compile - Compile a ".cxx" into a ".o"
This rule is used to compile C++ programs.
*/
 
do_cpp_compile (nodeP,multy_thread)
node* nodeP;
int multy_thread;
{
    char *cmdP;
    node *src;
    char srcfile[STRPATH+1];
    char dstfile[STRPATH+1];
    char tstname[STRLONG+1];
    int code;
    char *p;
    char *ccflags;
    char asrcfile[STRPATH+1];
    char *sysname, *str1;
    char *GetComponent();
    char debug_flag[250];
    char optimize_flag[4];
    char *compiler;
    static char DebugFlag[250];
    char elfrt_str[30];
    char *elfrtP;
 
    elfrtP = NeedVar("ELFRT_ONLY");
    if ( (*elfrtP == '\0') || (strcmp(elfrtP, "0") == 0 ) )
       /* normal build, do not include switch */
       elfrt_str[0] = '\0';
    else
       strcpy( elfrt_str, "-DELFRT_ONLY=1" );
 
 
	if ( multy_thread == TRUE ) {
	    compiler = AosCppCompilerMT;
	}
	else {
	    compiler = AosCppCompiler;
	}
 
    if ( DebugFlag[0] == '\0' )
        {
        /* initialize the DebugFlag name */
        if ( IsDebugBuild() )
           {
              strcpy( DebugFlag, AosDebugFlag);
           }
        else
           strcpy( DebugFlag, " " );
        }

 
    src = ND_IVAL0;
    NodeFilename(src,srcfile,TRUE);
    NodeFilename(nodeP,dstfile,TRUE);
    SetNodeDir(nodeP);
 
    if ( LimitedDebug ) {
       if ( file_in_debug_list( srcfile) ) {
           /* turn on debug mode for this file */
           strcpy( debug_flag, DebugFlag );
           optimize_flag[0] = '\0';      /* turn off optimization */
       }
       else  {
           debug_flag[0] = '\0';  /* turn off debug info */
           strcpy( optimize_flag, AosOptimize );
       }
    }
    else {
       /* do the normal build thing */
       strcpy( debug_flag, DebugFlag );
       strcpy( optimize_flag, AosOptimize );
    }
 
    cmdP = ThisCommand;
/*    ccflags = AosCCflags; OBSOLETED!!  */    /* Let's obsolete THIS !!! */
 
    sysname = AosUnix;
    sprintf(cmdP,
        "%s -c %s %s %s %s -DDEBUG=%s -D%s -DSYS=%s %s %s ",
            compiler, debug_flag,NeedVar("BLD_FLAGS"),
            NeedVar("AOS_FLAGS"),
            optimize_flag,AosDebug,AosMachine,sysname,
            elfrt_str, NodeWithStr(nodeP));
    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, FALSE);
    strcat(cmdP," ");
    BUMP(cmdP);
    GenProtosForMe(nodeP);
#if AXMS
    strcat( cmdP, "-Tp" );    /* force compiler to treat file as C++ file */
    BUMP(cmdP);
#endif
    strcpy(cmdP,srcfile);
    ReSuffix(tstname,FindEntryName(srcfile),OEXT);
 
#if AXMS
    if ( streq(tstname,FindEntryName(dstfile)) )
#else
    if ( Cstreq(tstname,FindEntryName(dstfile)) )
#endif
        code = proc(ThisCommand, TRUE);
    else {
        SetDir(TempDir);
        code = proc(ThisCommand, TRUE);
        ReSuffix(tstname,FindEntryName(srcfile),OEXT);
        NodeFilename(nodeP,dstfile,TRUE);
        if ( code == 0  ) {
            VERBOSE(("renaming %s to %s\n",tstname,dstfile));
            Arename(tstname,dstfile);
        }
        else
            unlink(tstname);
    }
    return(code);
}

/*************************************************************************
function: IsLoose - return whether to use the regular, rather than C++ compiler
The nodebug compilations used for ELF modules are always strict compiles
for historical reasons
This includes a bogus check for strictness based on whether any nodes
have been excluded from strictness!
*/
int IsLoose(node *nodeP) {
	int strict;
	static int haveCheckedForStrictness = FALSE;
	static int StrictCompiles = FALSE;
	int id;

	if ( !haveCheckedForStrictness ) {
		haveCheckedForStrictness = TRUE;
		for ( id = NextType(0,TYPE_NODE); id ;id = NextType(id,TYPE_NODE) ) {
			if ( NodePtr(id)->strict_excluded ) {
				StrictCompiles = TRUE;
				break;
			}
		}
	}
	if ( RuleArray[nodeP->ruleid].func == nodbg_compile )
		strict = TRUE;
	else {
		strict = StrictCompiles;
/*
		if ( strict && !nodeP->strict_excluded )
			printf("%s warning: strict compiling forced on.\n",nodeP->name);
*/
	}
	return(!strict || (nodeP->strict_excluded && !CprotoForcing));
}

/*************************************************************************
function:  cc_compile - Compile a ".c" into a ".o"
NOTE, If you make a generic change here, also make it in ansi_compile.
*/
cc_compile (nodeP, enable_debug, use_optimizer, multi_thread)
node* nodeP;
int enable_debug;
int use_optimizer;
int multi_thread;
{
    char *cmdP;
    node *src;
    char srcfile[STRPATH+1];
    char dstfile[STRPATH+1];
    char tstname[STRLONG+1];
    int code;
    char *p;
    char *ccflags;
    char asrcfile[STRPATH+1];
    char *sysname, *str1;
    char *GetComponent();
    char debug_flag[250];
    char optimize_flag[4];
    char *compiler;
    static char DebugFlag[250];
    char elfrt_str[30]; 
    char *elfrtP;
 
    elfrtP = NeedVar("ELFRT_ONLY");
    if ( (*elfrtP == '\0') || (strcmp(elfrtP, "0") == 0 ) )
       /* normal build, do not include switch */
       elfrt_str[0] = '\0';
    else
       strcpy( elfrt_str, "-DELFRT_ONLY=1" );

	if ( multi_thread )
    	compiler = IsLoose(nodeP) ? AosCompilerMT : AosStrictCompilerMT;
	else    
    	compiler = IsLoose(nodeP) ? AosCompiler : AosStrictCompiler;

    if ( DebugFlag[0] == '\0' )
       {
        /* initialize the DebugFlag name */
          if ( IsDebugBuild() )
          {
              strcpy( DebugFlag, AosDebugFlag );
          }
          else
             strcpy( DebugFlag, " " );
        }
 

    src = ND_IVAL0;
    NodeFilename(src,srcfile,TRUE);
    NodeFilename(nodeP,dstfile,TRUE);
    SetNodeDir(nodeP);

    if ( LimitedDebug ) {
       if( enable_debug && file_in_debug_list( srcfile) ) {
           /* turn on debug mode for this file */
           strcpy( debug_flag, DebugFlag );
           optimize_flag[0] = '\0';      /* turn off optimization */
       }
       else {
           debug_flag[0] = '\0';  /* turn off debug info */
           strcpy( optimize_flag, AosOptimize );
       }
    } 
    else if(enable_debug) {       /* do the normal build thing */
       strcpy( debug_flag, DebugFlag );
       strcpy( optimize_flag, AosOptimize );
    }
	else if (use_optimizer) {
		 strcpy( optimize_flag, "" );
	}
    else {
       debug_flag[0]='\0';
       strcpy( optimize_flag, AosOptimize );
    }


    cmdP = ThisCommand;
    ccflags = AosCCflags;     /* Let's obsolete THIS !!! */

#if SUN4S	/* shame on me for doing this!!!! */
	if ( 0 == strcmp(compiler,AosStrictCompiler) ) {
		p = strstr(ccflags,"-Xa");
		if ( p ) {
			*p++ = ' ';
			*p++ = ' ';
			*p++ = ' ';
		}
	}
#endif

    sysname = AosUnix;
    sprintf(cmdP,
        "%s -c %s %s %s %s %s -DDEBUG=%s -D%s -DSYS=%s %s %s ",
            compiler, debug_flag, ccflags,NeedVar("BLD_FLAGS"),
            NeedVar("AOS_FLAGS"),
            optimize_flag,AosDebug,AosMachine,sysname,
            elfrt_str, NodeWithStr(nodeP));
    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, FALSE);
    strcat(cmdP," ");
    BUMP(cmdP);
    strcpy(cmdP,srcfile);
    ReSuffix(tstname,FindEntryName(srcfile),OEXT);
    GenProtosForMe(nodeP);

#if AXMS
    if ( streq(tstname,FindEntryName(dstfile)) )
#else
    if ( Cstreq(tstname,FindEntryName(dstfile)) )
#endif
        code = proc(ThisCommand, TRUE);
    else {
        SetDir(TempDir);
        code = proc(ThisCommand, TRUE);
        ReSuffix(tstname,FindEntryName(srcfile),OEXT);
        NodeFilename(nodeP,dstfile,TRUE);
        if ( code == 0  ) {
            VERBOSE(("renaming %s to %s\n",tstname,dstfile));
            Arename(tstname,dstfile);
        }
        else
            unlink(tstname);
    }
    return(code);
}

noopt_compile(nodeP)
node* nodeP;
{
    return(cc_compile(nodeP,FALSE,TRUE,FALSE));
}

dbg_compile(nodeP)
node* nodeP;
{
    return(cc_compile(nodeP,TRUE,FALSE,FALSE));
}

dbg_compile_mt(nodeP)
node* nodeP;
{
    return(cc_compile(nodeP,TRUE,FALSE,TRUE));
}

nodbg_compile(nodeP)
node* nodeP;
{
    return(cc_compile(nodeP,FALSE,FALSE,FALSE));
}

/*************************************************************************
function:  ansi_compile - Compile a ".c" into a ".o" using the ANSI compiler
NOTE, this is identical to cc_compile, except that it uses AosAnsiCompiler
*/
ansi_compile (nodeP)
node* nodeP;
{
    char *cmdP;
    node *src;
    char srcfile[STRPATH+1];
    char dstfile[STRPATH+1];
    char tstname[STRLONG+1];
    int code;
    char *p;
    char *ccflags;
    char asrcfile[STRPATH+1];
    char *sysname, *str1;
    char *GetComponent();
    char debug_flag[250];
    char optimize_flag[4];
    char *compiler;
    char *DebugFlag;
    char elfrt_str[30]; 
    char *elfrtP;
 
    elfrtP = NeedVar("ELFRT_ONLY");
    if ( (*elfrtP == '\0') || (strcmp(elfrtP, "0") == 0 ) )
       /* normal build, do not include switch */
       elfrt_str[0] = '\0';
    else
       strcpy( elfrt_str, "-DELFRT_ONLY=1" );

    
    compiler = IsLoose(nodeP) ? AosAnsiCompiler : AosStrictCompiler;

    DebugFlag = (IsDebugBuild() ? AosDebugFlag : " ");

    src = ND_IVAL0;
    NodeFilename(src,srcfile,TRUE);
    NodeFilename(nodeP,dstfile,TRUE);
    SetNodeDir(nodeP);

    if ( LimitedDebug ) {
      if( file_in_debug_list( srcfile) ) {
       /* turn on debug mode for this file */
       strcpy( debug_flag, DebugFlag );
       optimize_flag[0] = '\0';      /* turn off optimization */
      }
      else  {
       debug_flag[0] = '\0';  /* turn off debug info */
       strcpy( optimize_flag, AosOptimize );
      }
    }
    else {
       /* do the normal build thing */
       strcpy( debug_flag, DebugFlag );
       strcpy( optimize_flag, AosOptimize );
    }

    cmdP = ThisCommand;
    ccflags = AosCCflags;     /* Let's obsolete THIS !!! */

    sysname = AosUnix;
    sprintf(cmdP,
        "%s -c %s %s %s %s %s -DDEBUG=%s -D%s -DSYS=%s %s %s ",
            compiler, debug_flag, ccflags,NeedVar("BLD_FLAGS"),
            NeedVar("AOS_FLAGS"),
            optimize_flag,AosDebug,AosMachine,sysname,
            elfrt_str, NodeWithStr(nodeP));
    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, FALSE);
    strcat(cmdP," ");
    BUMP(cmdP);
    strcpy(cmdP,srcfile);
    ReSuffix(tstname,FindEntryName(srcfile),OEXT);
    GenProtosForMe(nodeP);

#if AXMS
    if ( streq(tstname,FindEntryName(dstfile)) )
#else
    if ( Cstreq(tstname,FindEntryName(dstfile)) )
#endif
        code = proc(ThisCommand, TRUE);
    else {
        SetDir(TempDir);
        code = proc(ThisCommand, TRUE);
        ReSuffix(tstname,FindEntryName(srcfile),OEXT);
        NodeFilename(nodeP,dstfile,TRUE);
        if ( code == 0  ) {
            VERBOSE(("renaming %s to %s\n",tstname,dstfile));
            Arename(tstname,dstfile);
        }
        else
            unlink(tstname);
    }
    return(code);
}


/*************************************************************************
function: aroff - Invoke aroff
*/
aroff (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *p;
    char srcname[300], dstname[300];

    src = ND_IVAL0;
    SetNodeDir(nodeP);
    NodeFilename(src,srcname,TRUE);
    NodeFilename(nodeP,dstname,TRUE);
    sprintf(ThisCommand,"aroff -f %s %s %s",dstname,NodeWithStr(nodeP),srcname);
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function: hcmprs - Help database compression program

Note that when compressing level 1 help, the input file is listed as the
".list" files.  In fact, the compressor searches the current dircetory
for all numeric files.

*/
hcmprs (nodeP)
node *nodeP;
{
    node *src;
    char file1[256];
    char file2[256];
    char *with;
    int ix;
    char *cmdP = ThisCommand;

    with = NodeWithStr(nodeP);
    if ( 0 == strcmp(with,"-c1") ) {
        SetNodeDir(ND_IVAL0);
        NodeFilename(nodeP,file1,TRUE);
        sprintf(cmdP,"hcompress -c1 %s ",file1);
        BUMP(cmdP);
        ListInFiles(nodeP,cmdP,NodeDirName(ND_IVAL0));
    }
    else if ( 0 == strcmp(with,"-c") ) {
        src = ND_IVAL0;
        SetNodeDir(src);
        NodeFilename(src,file1,TRUE);
        NodeFilename(nodeP,file2,TRUE);
        sprintf(ThisCommand,"hcompress -c %s %s",file1,file2);
    }
    else if ( 0 == strcmp(with,"-cb") ) {
        DefaultOutputDir(nodeP);
        src = ND_IVAL0;
        NodeFilename(ND_IVAL0,file1,TRUE);
        NodeFilename(nodeP,file2,TRUE);
        sprintf(ThisCommand,"hcompress -cb %s %s",file1,file2);
    }
    else {
        assert(FALSE);
    }
    return(proc(ThisCommand, FALSE));
}




/*************************************************************************
function:  build_browser - build the browse info
*/
#if !AXMS                         /*  {  */
int build_browser(nodeP)
{
    ErrorMsg("build_browser - No a know function on UNIX");
    return(3);
}

#else                              /*  }  {  */

build_browser(nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *p;
    char *name;
    char *fsize;
    char *segs = NULL;
    int is_dbx = FALSE;
    static char linker[100];
    static char DebugFlag[100];
    char *debugP;
    int ret;


    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;


    strcpy( cmdP, "BSCMAKE -nologo  -o" );

    BUMP(cmdP);
    NodeFilename(nodeP,cmdP,FALSE);   /* add output filename */
    BUMP(cmdP);

    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    strcat(cmdP," ");
    strcat(cmdP,NodeWithStr(nodeP));

    /* The input files will be ".obj" filename.  Convert these all to */
    /* ".sbr" names                                                   */
    p = cmdP;
    while ( *p != '\0' )
       {
       if ( *p == '.' )
           {
           if ( strnicmp( p, ".obj", 4 ) == 0 )
              {
              strncpy( p, ".sbr", 4 );
              p = p + 3;
              }
           }
       p++;
       }

    ret = proc(ThisCommand, FALSE);


   return( ret );

}
#endif                          /*  AXMS }  */





/*************************************************************************
function:  default_link - Link files into an executable
           using the system default libraries
*/
#if !AXMS                         /*  {  */
int default_link (nodeP)
{
    return( do_cc_exec( nodeP,FALSE ) );
}

#else                              /*  }  {  */

default_link (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *p;
    char *name;
    char *fsize;
    char *segs = NULL;
    int is_dbx = FALSE;
    static char linker[100];
    static char DebugFlag[100];
    char *debugP;
    int ret;


    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;

    if ( linker[0] == '\0' )
        {
        /* initialize the linker name */
           strcpy( linker, AosFinal );
        }

    if ( DebugFlag[0] == '\0' )
        {
        /* initialize the DebugFlag name */
        if ( IsDebugBuild() )
          {
            /* We do not use the "debug flag" in the ctl file because that is
             * a compiler flag only.            -donato
             */  
             strcpy( DebugFlag, "-debug" );
          }
        else
          strcpy( DebugFlag, " " );
        }

    if ( file_in_console_list( nodeP->name ) )
        /* link with the "console" libraries */
        sprintf(cmdP,"%s %s %s -out:", AosFinal, 
                             FindDefine("AOS_DEF_FNLFLAGS"), DebugFlag);
    else
        /* link with the GUI libraries */
        sprintf(cmdP,"%s %s %s -out:", AosFinal, 
                             FindDefine("AOS_DEF_GUIFLAGS"), DebugFlag);

    BUMP(cmdP);
    NodeFilename(nodeP,cmdP,FALSE);
    BUMP(cmdP);

    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    strcat(cmdP," ");
    strcat(cmdP,NodeWithStr(nodeP));

    ret = proc(ThisCommand, FALSE);

    /* the link process produces unwanted ".exp" and ".lib" files.  The */
    /* only ones we want are for axmain and elfrt (things which may load */
    /* shared libraries.  The ".lib" file is needed to build the shaared */
    /* library).  Delete the unwanted ones here                          */
    /* delete the files */
    cmdP = ThisCommand;
    *cmdP = '\0';
    NodeFilename(nodeP,cmdP,FALSE);
    p = cmdP + strlen(cmdP);
    while ( *p != '.' )
      { p--; }
    *p = '\0';
    strcat( cmdP, ".EXP" );
    VERBOSE(("Deleting unused file: < %s >\n", cmdP));
    unlink( cmdP );

    if ( (stricmp( nodeP->name, "axmain" ) != 0) &&
         (stricmp( nodeP->name, "elfrt" ) != 0) )
       {
       p = cmdP + strlen(cmdP);
       while ( *p != '.' )
         { p--; }
       *p = '\0';
       strcat( cmdP, ".LIB" );
       VERBOSE(("Deleting unused file: < %s >\n", cmdP));
       unlink( cmdP );
       }

   return( ret );

}
#endif                          /*  AXMS }  */

/*************************************************************************
function:  cc_exec - Link files into an executable
NOTE, If you make a generic change here, also make it in ansi_exec.
*/
do_cc_exec (nodeP, multi_thread)
node *nodeP;
int multi_thread;
{
    char *cmdP;
    node *src;
    char *p;
    char *name;
    char *fsize;
    char *segs = NULL;
    int is_dbx = FALSE;
    static char linker[100];
    static char DebugFlag[100];
    char *debugP;
    int ret;
#if AXMS
    char with_str[500];
#endif

    if ( getenv("DO_PURIFY") != (char *)NULL )
        do_purify( nodeP );

    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;

    if ( linker[0] == '\0' )
    {
		/* initialize the linker name */
		if ( multi_thread )
	    	strcpy(linker, AosFinalMT);
		else
	    	strcpy(linker, AosFinal);
    }

    if ( DebugFlag[0] == '\0' )
        {
        /* initialize the DebugFlag name */
        if ( IsDebugBuild() )
          {
#if AXMS
            /* We do not use the "debug flag" in the ctl file because that is
             * a compiler flag only.            -donato
             */  
             strcpy( DebugFlag, "-debug:full -debugtype:BOTH" );
#else
             strcpy( DebugFlag, AosDebugFlag );
#endif
          }
        else
          strcpy( DebugFlag, " " );
        }

#if AXMS                        /* {  Windows NT linker */
    if ( file_in_console_list( nodeP->name ) )
        /* link with the "console" libraries */
        sprintf(cmdP,"%s %s %s -out:",AosFinal, AosFnlflags, DebugFlag);
    else
        /* link with the GUI libraries */
        sprintf(cmdP,"%s %s %s -out:",AosFinal, AosGuiflags, DebugFlag);

#else    /*  }  {  */

    /* UNIX */
    sprintf(cmdP,"%s %s %s -o ", linker,AosFnlflags,DebugFlag);
#endif   /*  }  */

    BUMP(cmdP);
    NodeFilename(nodeP,cmdP,FALSE);
    BUMP(cmdP);
#if RIOS
	if ( !multi_thread )
	{
    	strcat(cmdP," /lib/crt0.o ");
    	BUMP(cmdP);
	}
#endif

    DoingLink = TRUE;      /* turn on ".DEF" handling for Windows */
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    DoingLink = FALSE;

    strcat(cmdP," ");

#if AXMS
    /* check for -no_default_libs. if it exists then use the subsequent list
     * of libs rather than the default list of libs (AosLibs).
     */
    strcpy(with_str, NodeWithStr(nodeP));
    p = strstr(with_str, "-no_default_libs");
    if ( p ) {
        memset(p, ' ', 16);   /* blank special switch out */
        strcat(cmdP, with_str);
    }
    else {
        strcat(cmdP,with_str);
        strcat(cmdP," ");
        strcat(cmdP,AosLibs);
    }
#else
    strcat(cmdP,NodeWithStr(nodeP));
    strcat(cmdP," ");
    strcat(cmdP,AosLibs);
#endif

#if LINUX
    BUMP(cmdP);
	sprintf(cmdP," -L%s",NeedFullVar("AOS_SHLIB"));
    BUMP(cmdP);
	sprintf(cmdP," -L%s/rts",NeedFullVar("AOS_OBJECTS"));
#endif
    p = getenv("AX_MORE_LIBS");
    if ( p != (char *)NULL )
       {
       strcat(cmdP," ");
       strcat(cmdP, p );
       }

    ret = proc(ThisCommand, FALSE);

#if AXMS    /*  {  */
    /* the link process produces unwanted ".exp" and ".lib" files.  The */
    /* only ones we want are for axmain and elfrt (things which may load */
    /* shared libraries.  The ".lib" file is needed to build the shaared */
    /* library).  Delete the unwanted ones here                          */
    /* delete the files */
    cmdP = ThisCommand;
    *cmdP = '\0';
    NodeFilename(nodeP,cmdP,FALSE);
    p = cmdP + strlen(cmdP);
    while ( *p != '.' )
      { p--; }
    *p = '\0';
    strcat( cmdP, ".EXP" );
    VERBOSE(("Deleting unused file: < %s >\n", cmdP));
    unlink( cmdP );

    if ( (stricmp( nodeP->name, "axmain" ) != 0) &&
         (stricmp( nodeP->name, "entsvr" ) != 0) &&
         (stricmp( nodeP->name, "elfrt" ) != 0) )
       {
       p = cmdP + strlen(cmdP);
       while ( *p != '.' )
         { p--; }
       *p = '\0';
       strcat( cmdP, ".LIB" );
       VERBOSE(("Deleting unused file: < %s >\n", cmdP));
       unlink( cmdP );
       }
#endif    /*  }  AXMS  */

   return( ret );

}

int cc_exec(nodeP)
node *nodeP;
{
    return(do_cc_exec(nodeP,FALSE));
}

int cc_exec_mt(nodeP)
node *nodeP;
{
    return(do_cc_exec(nodeP,TRUE));
}

/*************************************************************************
function:  ansi_exec - Link files into an executable
Identical to cc_exec except that it uses AOS_ANSI_LINKER or AosAnsiFinal.
*/
ansi_exec (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *p;
    char *name;
    char *fsize;
    char *segs = NULL;
    int is_dbx = FALSE;
    static char linker[100];
    static char DebugFlag[100];
    char *debugP;
    int ret;

 
#if AXMS   /*  {  */
    /** on NT the basic linker is ansi */
    ret = do_cc_exec( nodeP,FALSE );
    return( ret );
#else     /*  }  {  */

    if ( getenv("DO_PURIFY") != (char *)NULL )
        do_purify( nodeP );

 
    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;

    if ( linker[0] == '\0' )
        {
        /* initialize the linker name */
           p = GetVar("AOS_ANSI_FINAL");
           if ( p != (char *)NULL )
               strcpy( linker, p );
           else
               /* use the "normal" linker */
               strcpy( linker, AosFinal );
        }
    if ( DebugFlag[0] == '\0' )
        {
        /* initialize the DebugFlag name */
        if ( IsDebugBuild() )
          {
             strcpy( DebugFlag, AosDebugFlag );
          }
        else
          strcpy( DebugFlag, " " );
        }


    sprintf(cmdP,"%s %s %s -o ", linker,AosFnlflags,DebugFlag);

    BUMP(cmdP);
    NodeFilename(nodeP,cmdP,FALSE);
    BUMP(cmdP);
#if RIOS
    strcat(cmdP," /lib/crt0.o ");
    BUMP(cmdP);
#endif
    DoingLink = TRUE;      /* turn on ".DEF" handling for Windows */
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    DoingLink = FALSE;

    strcat(cmdP," ");
    strcat(cmdP,NodeWithStr(nodeP));
    strcat(cmdP," ");
    strcat(cmdP,AosLibs);

#if LINUX
    BUMP(cmdP);
	sprintf(cmdP," -L%s",NeedFullVar("AOS_SHLIB"));
    BUMP(cmdP);
	sprintf(cmdP," -L%s/rts",NeedFullVar("AOS_OBJECTS"));
#endif
    p = getenv("AX_MORE_LIBS");
    if ( p != (char *)NULL )
       {
       strcat(cmdP," ");
       strcat(cmdP, p );
       }

    ret = proc(ThisCommand, FALSE);

#endif        /*   } not AXMS */

   return( ret );

}

/*************************************************************************
function:  lint_final   - lint a bunch of stuff which makes up an executable
*
*
* This call a shell script to actually run lint.  The arguments to the
* script are:
*
*  lint_final <full_path_to_output> <simple_output_name> <list-of-.ln's>
*
*  For example:
*    lint_final /ax/axexec/axdata/axmain axmain ssinit.ln uimain.ln ...
*
* Note: The ".ln" files are simple names.  The shell script must know 
*      where these things were put by an earlier call to "lint_pass1" and
*      lint_partial.  In the sample below all lint files are assumed to be
*      in "/ax/axobj/lint"
*      See the functions "lintlib" and "lint_partial" here in amake.
*
* An example shell script might be  (for system-V line)

:
FULLPATH=$1
shift
FILE=$1
shift
cd /ax/axobj/lint
lint  $* 1>$FILE.lint_final  2>&1

*/

lint_final (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *p;
    char destfile[STRPATH+1], base_name[60], this_file[STRPATH+1];
    int ix, incount;
    char simple_name[60];
    node *childP;
    int code;

    cmdP = ThisCommand;

    strcpy( cmdP, "lint_final " );

    BUMP(cmdP);

    NodeFilename(nodeP,cmdP,FALSE);
    BUMP(cmdP);

    /** nodeP->name is the output file, with no path */
    ReSuffix( base_name, nodeP->name, "");
  
    sprintf( cmdP, " %s", base_name );

    CMPTELL(("lint_final into %s\n",  nodeP->name ));

    /* now enumerate all of the input files as simple files.  It is up to */
    /* the shell script to know where these things are!        */
    incount = NodeInCount(nodeP);
    for ( ix=0; ix<incount; ix++ )
        {
        childP = NodeInVal(nodeP,ix);
        NodeFilename(childP,this_file,TRUE);
        if ( strncmp( this_file, "/usr", 4 ) == 0 )
           /* ignore system files */
           continue;
         if ( *this_file == '-' )
           /* ignore library files */
           continue;

        strcpy( simple_name, FindEntryName(this_file) );
        ReSuffix( base_name, simple_name, ".ln" );
        strcat( cmdP, " " );
        strcat( cmdP, base_name );
        BUMP(cmdP);
        }
        
    code = proc(ThisCommand, TRUE);
    return( code );
}




/*************************************************************************
function:  lint_partial   - lint a bunch of libraries into a larger library
*
*
* This call a shell script "lint_partial" to actually run lint.  
*The arguments to the script are:
*
*  lint_partial <full_path_to_.o_output> <simple_output_name> <list-of-.ln's>
*
*  For example:
*    lint_partial /ax/axobj/axdata/rts/ss_atm.o ss_atm ssinit.ln uimain.ln ...
*
* Note: The ".ln" files are simple names.  The shell script must know 
*      where these things were put by an earlier call to "lint_pass1".
*      In the sample below, all lint output is assumed to be in "/ax/axobj/lint"
*      See the function "lintlib" here in amake.
*
* An example shell script might be  (for system-V line)
:
FULLPATH=$1
shift
FILE=$1
shift
cd /ax/axobj/lint
lint -o$FILE $* 1>$FILE.pass2  2>&1
#
# The output is "llib-l$FILE.ln", we re-name it to a simple name so that
#   it can be found easily, since the build process sometimes produces ".o"
#   files as the result of a prtial link, and sometimes ".a" files so it
#   is hard in amake to know for sure what is being linked on the full link
rm -f /ax/axobj/$FILE.ln
mv llib-l$FILE.ln $FILE.ln

*/

lint_partial (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *p;
    char destfile[STRPATH+1], base_name[60], this_file[STRPATH+1];
    int ix, incount;
    char simple_name[60];
    node *childP;
    int code;

    cmdP = ThisCommand;

    strcpy( cmdP, "lint_partial " );

    BUMP(cmdP);

    NodeFilename(nodeP,cmdP,FALSE);
    BUMP(cmdP);

    /** nodeP->name is the output file, with no path */
    ReSuffix( base_name, nodeP->name, "");
  
    CMPTELL(("lint_partial into %s\n",  nodeP->name ));
    sprintf( cmdP, " %s", base_name );

    /* now enumerate all of the input files as simple files.  It is up to */
    /* the shell script to know where these things are!        */
    incount = NodeInCount(nodeP);
    for ( ix=0; ix<incount; ix++ )
        {
        childP = NodeInVal(nodeP,ix);
        NodeFilename(childP,this_file,TRUE);
        strcpy( simple_name, FindEntryName(this_file) );
        ReSuffix( base_name, simple_name, ".ln" );
        strcat( cmdP, " " );
        strcat( cmdP, base_name );
        BUMP(cmdP);
        }
        
    code = proc(ThisCommand, TRUE);
    return( code );
}



/*************************************************************************
function:  ld_plink - Link files into a ".out"
*/

ld_plink (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *p;
    char destfile[STRPATH+1];

    if ( 0 == strcmp(AosLinker,"ar") ) {
        archiver(nodeP);
        return;
    }
    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;
#if AXMS
    sprintf(cmdP,"%s %s /out:", AosLibraryBuilder, AosLdFlags);
#else
    sprintf(cmdP,"%s -r %s -o ",AosLinker,AosLdFlags);
#endif
    BUMP(cmdP);
    NodeFilename(nodeP,destfile,FALSE);
    strcat( cmdP, destfile );
    BUMP(cmdP);
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    strcat(cmdP," ");
    strcat(cmdP,NodeWithStr(nodeP));
    return(proc(ThisCommand, FALSE));
}






/**===============================================================**/
#if AXMS   /*  {  */

/** make_import_lib - create the import library, and the exports definition **/
/**                   file. Both are created at the same time, though the   **/
/**                   rule is repeated in the .amk file since amake does    **/
/**                   not handle multiple output from a single rule well    **/
make_import_lib( nodeP )
node *nodeP;
{
    char *cmdP;
    node *src;
    char *p;
    char destfile[STRPATH+1];
    int ret;
 
    DefaultOutputDir( nodeP );
    cmdP = ThisCommand;

#if NT_ALPHA
    sprintf( cmdP, "%s -machine:ALPHA -out:", AosLibraryBuilder );
#endif
#if NT_X86
    sprintf( cmdP, "%s -machine:i386 -out:", AosLibraryBuilder );
#endif
#if NT_MIPS
    sprintf( cmdP, "%s -machine:MIPS -out:", AosLibraryBuilder );
#endif
#if NT_PPC
    sprintf( cmdP, "%s -machine:PPC -out:", AosLibraryBuilder );
#endif

    BUMP(cmdP);
    NodeFilename(nodeP,destfile,FALSE);
    strcat( cmdP, destfile );
    p = cmdP + strlen(cmdP) - 3;
    if ( stricmp( p, "exp" ) == 0 )
       {
       /* the command really takes a ".lib", but we have a double output */
       /* situation so the rule is run twice in the ".amk" files         */
       strncpy( p, "lib", 3 );
       }
    BUMP(cmdP);

    strcat( cmdP, " -def" );

    /* a bit of a hack here.  The name of the "def" file must immediatly */
    /* follow the colon of "-def:", but ListInFiles() will put on a leading */
    /* blank.  So anticipate this, and convert the blank to the required ":" */
    p = cmdP + strlen(cmdP);    /* end of string will be a blank */

    BUMP(cmdP);
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    *p = ':';        /* create -def:<output-file>  */

    ret = proc(ThisCommand, FALSE);
    return( ret );
}




ld_mkshared (nodeP)
node *nodeP;
{
    Do_ld_mkshared (nodeP, FALSE);
}




/*** This is the NT/95 version of make-shared-lib! ***/
Do_ld_mkshared (nodeP, use_mfc)
node *nodeP;
int use_mfc;
{
    char *cmdP;
    node *src;
    char *p;
    char destfile[STRPATH+1];
    int ret;
    static char DebugFlag[100];
    char with_str[200];
    int keep_exp, keep_lib;

    if ( DebugFlag[0] == '\0' )
        {
        /* initialize the DebugFlag name */
        if ( IsDebugBuild() )
            {
            /* These flags should be moved to  ctl file */
            /* We do not use the "debug flag" in the ctl file because that is
             * a compiler flag only.            -donato
             */
            p = getenv( "AOS_DEBUG_FLAG" );
            if ( p != (char *)NULL )
                strcpy( DebugFlag, p );
            else
                strcpy( DebugFlag, "-debug:full -debugtype:cv" );
            }
        else
            strcpy( DebugFlag, " " );
        }

    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;
    sprintf(cmdP, "%s %s %s -out:", AosFinal, AosShlibLDFlags, DebugFlag);
 
    BUMP(cmdP);
    NodeFilename(nodeP,destfile,FALSE);
    strcat( cmdP, destfile );
    BUMP(cmdP);

    DoingLink = TRUE;      /* turn on ".DEF" handling for Windows */
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    DoingLink = FALSE;

    strcat(cmdP," ");

    /** see if the with str contains the string "-keep_exp" or "-keep_lib" */
    /** to prevent eletion of te libraries below                          */
    strcpy( with_str, NodeWithStr(nodeP) );
    keep_exp = keep_lib = FALSE;
    p = strstr( with_str, "-keep_exp" );
    if ( p != (char *)NULL )
       {
       keep_exp = TRUE;
       memset( p, ' ', 9 );   /* blank special switch out */
       }
    p = strstr( with_str, "-keep_lib" );
    if ( p != (char *)NULL )
       {
       keep_lib = TRUE;
       memset( p, ' ', 9 );   /* blank special switch out */
       }
    
    strcat(cmdP, with_str);
    strcat(cmdP," ");
    if ( use_mfc )
       {
       /* add on the mfc libraries */
       strcat(cmdP, FindDefine("AOS_MFC_LIBS") );
       strcat( cmdP, " " );
       }
    strcat(cmdP,AosShlibLibs);
    ret = proc(ThisCommand, FALSE);

    /* delete any ".EXP" and ".LIB" files produced by the linker.  These */
    /* are not needed since we do not link one shared lib to another     */
    cmdP = ThisCommand;
    *cmdP = '\0';
    NodeFilename(nodeP,cmdP,FALSE);

    p = cmdP + strlen(cmdP);
    while ( *p != '.' )
      { p--; }
    *p = '\0';
    strcat( cmdP, ".EXP" );
    if ( !keep_exp )
      {
      VERBOSE(("Deleting unused file: < %s >\n", cmdP));
      unlink( cmdP );
      }

    p = cmdP + strlen(cmdP);
    while ( *p != '.' )
      { p--; }
    *p = '\0';
    strcat( cmdP, ".LIB" );
    if ( !keep_lib )
      {
      VERBOSE(("Deleting unused file: < %s >\n", cmdP));
      unlink( cmdP );
      }

    return( ret );
}

ld_mksharedcpp (nodeP)
node *nodeP;
{
    ld_mkshared(nodeP);
}

/********************************************************************/
/* This does a multi-threaded build, which assumes that we are built */
/* using the MFC stuff                                               */

ld_mksharedcpp_mt (nodeP)
node *nodeP;
{
    Do_ld_mkshared(nodeP,TRUE);
}

#else     /* }  AXMS {  */

make_import_lib( nodeP )
node *nodeP;
{
    ErrorMsg("make_import_lib not valid on UNIX");
	return(1);
}

#if SGI
static char *
basename(str)
char *str;
{
      char *p = strrchr(str, '/');
      if (!p)
              return(str);
      return(p+1);
}
#endif

#if SUN4S
#define START_SKIP (strncmp(buf,"Undefined",9)==0)
#define END_SKIP   (strncmp(buf,"ld: warning: Symbol refrencing errors",37)==0)
#elif ALPHA
#define START_SKIP (strncmp(buf,"Warning: Unresolved:",20)==0)
#define END_SKIP 0
#else
#define START_SKIP 0
#define END_SKIP 0
#endif	   

/*************************************************************************
  function:  ld_stripUndef - strip Undefine symbol warnings from ld output
             and print to StdOutFile
*/
void ld_stripUndef(errFile)
char* errFile;
{
	FILE* fp;
	char buf[1024];
	int inSkipMode=FALSE;
	int endOfSkipFound=TRUE;
	fp = fopen(errFile,"r");
	if (Verbose) {
		VERBOSE(("copying %s file contents to stdout\n",errFile));
	}
	while(fgets(buf,1024,fp) != NULL) {
		if (START_SKIP) {
			if (Verbose) {
				VERBOSE(("***Following lines are skipped in non-verbose mode\n"));
				printf("%s",buf);
				endOfSkipFound = FALSE;
			}
			else
				inSkipMode = TRUE;
		}
		else if (END_SKIP) {
			if (Verbose) {
				printf("%s",buf);
				VERBOSE(("***End of lines skipped in non-verbose mode\n"));
				endOfSkipFound = TRUE;
			}
			inSkipMode = FALSE;
		}
		else if (!inSkipMode)
			printf("%s",buf);
	}
	if (Verbose) {
		if (!endOfSkipFound)
			VERBOSE(("***End of lines skipped in non-verbose mode\n"));
		VERBOSE(("removing file %s\n",errFile));
	}
	fclose(fp);
	unlink(errFile);
}

/*************************************************************************
function:  ld_mkshared - Link files into a ".so" 
*/

ld_mkshared (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *shldflags;
    char *p;
	int pe;
    char destfile[STRPATH+1];
    char errFile[STRPATH+1];

    if ( 0 == strcmp(AosLinker,"ar") ) {
        archiver(nodeP);
        return;
    }

    shldflags = AosShlibLDFlags; 

    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;
#if AXMS
    sprintf(cmdP,"%s %s /out:", AosLibraryBuilder, AosLdFlags);
#elif RIOS
    sprintf(cmdP,"%s %s -o ",AosCompiler,shldflags);
#elif LINUX || LINUX_LIBC5
    sprintf(cmdP,"%s %s -o ",AosFinal,shldflags);
#else
    sprintf(cmdP,"%s %s -o ",AosLinker,shldflags);
#endif

#if !SGI
    BUMP(cmdP);
    NodeFilename(nodeP,destfile,FALSE);
    strcat( cmdP, destfile );
#if LINUX
    BUMP(cmdP);
	sprintf(cmdP," -L%s",NeedFullVar("AOS_SHLIB"));
    BUMP(cmdP);
	sprintf(cmdP," -L%s/rts",NeedFullVar("AOS_OBJECTS"));
#endif
    BUMP(cmdP);
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    strcat(cmdP," ");
    strcat(cmdP,NodeWithStr(nodeP));
	GetTemp("Errors",errFile);
	pe = PipeProc(ThisCommand, errFile);
	ld_stripUndef(errFile);
	return(pe);
#else
    {
        char mv_cmd[(STRPATH * 2) + 30];
        int pe, se;
 
        BUMP(cmdP);
        NodeFilename(nodeP,destfile,FALSE);
        strcat( cmdP, basename(destfile) );
        BUMP(cmdP);
        ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
        strcat(cmdP," ");
        strcat(cmdP,NodeWithStr(nodeP));
		GetTemp("Errors",errFile);
        pe = PipeProc(ThisCommand, errFile);
		ld_stripUndef(errFile);
        sprintf(mv_cmd, "mv %s %s", basename(destfile), destfile);
        se = system(mv_cmd);
        return(pe);
    }
#endif

}

/*************************************************************************
function:  ld_mksharedcpp - Link C++files into a ".so" 
*/

ld_mksharedcpp (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *shldflags;
    char *p;
	int pe;
    char destfile[STRPATH+1];
    char errFile[STRPATH+1];

    if ( 0 == strcmp(AosLinker,"ar") ) {
        archiver(nodeP);
        return;
    }

    shldflags = AosShlibLDFlags; 

    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;
#if AXMS
    sprintf(cmdP,"%s %s /out:", AosLibraryBuilder, AosLdFlags);
#elif RIOS
    sprintf(cmdP,"%s %s -o ",AosCppCompiler,shldflags);
#elif SGI
    sprintf(cmdP,"%s %s -o ",AosLinkerCpp,shldflags);
#elif SUN4S
    sprintf(cmdP,"%s %s -o ",AosLinkerCpp,shldflags);
#elif SUNIS
    sprintf(cmdP,"%s %s -o ",AosFinalCPlus,shldflags);
#elif LINUX || LINUX_LIBC5
	sprintf(cmdP,"%s %s -o ",AosFinalCPlus,shldflags);
#else
    sprintf(cmdP,"%s %s -o ",AosLinker,shldflags);
#endif

#if !SGI
    BUMP(cmdP);
    NodeFilename(nodeP,destfile,FALSE);
    strcat( cmdP, destfile );
#if LINUX
    BUMP(cmdP);
	sprintf(cmdP," -L%s",NeedFullVar("AOS_SHLIB"));
    BUMP(cmdP);
	sprintf(cmdP," -L%s/rts",NeedFullVar("AOS_OBJECTS"));
#endif
    BUMP(cmdP);
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    strcat(cmdP," ");
    strcat(cmdP,NodeWithStr(nodeP));
	GetTemp("Errors",errFile);
	pe = PipeProc(ThisCommand, errFile);
	ld_stripUndef(errFile);
	return(pe);
#else
    {
        char mv_cmd[(STRPATH * 2) + 30];
        int pe, se;
 
        BUMP(cmdP);
        NodeFilename(nodeP,destfile,FALSE);
        strcat( cmdP, basename(destfile) );
        BUMP(cmdP);
        ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
        strcat(cmdP," ");
        strcat(cmdP,NodeWithStr(nodeP));
		GetTemp("Errors",errFile);
		pe = PipeProc(ThisCommand, errFile);
		ld_stripUndef(errFile);
        sprintf(mv_cmd, "mv %s %s", basename(destfile), destfile);
        se = system(mv_cmd);
        return(pe);
    }
#endif

}

/*************************************************************************
function:  ld_mksharedcpp_mt - Link C++files into a ".so" 
*/

ld_mksharedcpp_mt (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *shldflags;
    char *p;
	int pe;
    char destfile[STRPATH+1];
    char errFile[STRPATH+1];

    if ( 0 == strcmp(AosLinker,"ar") ) {
        archiver(nodeP);
        return;
    }

    shldflags = AosShlibLDFlags; 

    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;
#if AXMS
    sprintf(cmdP,"%s %s /out:", AosLibraryBuilder, AosLdFlags);
#elif RIOS
    sprintf(cmdP,"%s -o ",AosFinalCPlusSHMT);
#elif SGI
    sprintf(cmdP,"%s %s -o ",AosLinkerCpp,shldflags);
#elif SUN4S
    sprintf(cmdP,"%s %s -o ",AosLinkerCpp,shldflags);
#elif SUNIS
    sprintf(cmdP,"%s %s -o ",AosFinalCPlus,shldflags);
#elif LINUX || LINUX_LIBC5
	sprintf(cmdP,"%s %s -o ",AosFinalCPlus,shldflags);
#else
    sprintf(cmdP,"%s %s -o ",AosLinker,shldflags);
#endif

#if !SGI
    BUMP(cmdP);
    NodeFilename(nodeP,destfile,FALSE);
    strcat( cmdP, destfile );
#if LINUX
    BUMP(cmdP);
	sprintf(cmdP," -L%s",NeedFullVar("AOS_SHLIB"));
    BUMP(cmdP);
	sprintf(cmdP," -L%s/rts",NeedFullVar("AOS_OBJECTS"));
#endif
    BUMP(cmdP);
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    strcat(cmdP," ");
    strcat(cmdP,NodeWithStr(nodeP));
	GetTemp("Errors",errFile);
	pe = PipeProc(ThisCommand, errFile);
	ld_stripUndef(errFile);
	return(pe);
#else
    {
        char mv_cmd[(STRPATH * 2) + 30];
        int pe, se;
 
        BUMP(cmdP);
        NodeFilename(nodeP,destfile,FALSE);
        strcat( cmdP, basename(destfile) );
        BUMP(cmdP);
        ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
        strcat(cmdP," ");
        strcat(cmdP,NodeWithStr(nodeP));
		GetTemp("Errors",errFile);
		pe = PipeProc(ThisCommand, errFile);
		ld_stripUndef(errFile);
        sprintf(mv_cmd, "mv %s %s", basename(destfile), destfile);
        se = system(mv_cmd);
        return(pe);
    }
#endif

}

#endif      /* }  !AXMS   */
/*************************************************************************
function:  archiver - Link .o's into a .a
*/
archiver (nodeP)
node *nodeP;
{
    char *cmdP;
    node dst[256];
    char *p;
    int code;
    char *ranlibCommand;

    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;
    NodeFilename(nodeP,dst,TRUE);
#if AXMS
    sprintf(cmdP,"%s /out:%s ",AosLibraryBuilder, dst);
#else
    sprintf(cmdP,"ar rc %s ",dst);
#endif
    BUMP(cmdP);
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    code = proc(ThisCommand, FALSE);
#if AXMS
    return(code);
#else
    if ( code )
        return(code);
    ranlibCommand = NeedVar("AOS_RANLIB");
    if ( strcmp(ranlibCommand, "") == 0 )
        return(0);
    else
        {
        sprintf(ThisCommand,"%s %s", ranlibCommand, dst);
        return(proc(ThisCommand, FALSE));
        }
#endif
}

/*************************************************************************
function:  archiver_cplus - Link .o's into a .a
*/
archiver_cplus (nodeP)
node *nodeP;
{
    char *cmdP;
    node dst[256];
    char *p;
    int code;
    char *ranlibCommand;

    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;
    NodeFilename(nodeP,dst,TRUE);
#if AXMS
    sprintf(cmdP,"%s /out:%s ",AosLibraryBuilder, dst);
#else
    sprintf(cmdP,"%s rc %s ",AosArchiverCPlus, dst);
#endif
    BUMP(cmdP);
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    code = proc(ThisCommand, FALSE);
#if AXMS
    return(code);
#else
    if ( code )
        return(code);
    ranlibCommand = NeedVar("AOS_RANLIB");
    if ( strcmp(ranlibCommand, "") == 0 )
        return(0);
    else
        {
        sprintf(ThisCommand,"%s %s", ranlibCommand, dst);
        return(proc(ThisCommand, FALSE));
        }
#endif
}

/*************************************************************************
function: archive_extract - Remove a file from a library
*/
archive_extract (nodeP)
node *nodeP;
{
    char src[256];
    int code;

    SetNodeDir(nodeP);
    NodeFilename(ND_IVAL0,src,TRUE);
    sprintf(ThisCommand,"ar -x %s %s",src,NodeName(nodeP));
    if ( code = proc(ThisCommand, FALSE) )
        return(code);
}

/*************************************************************************
 * function:  ss_strgen - Create ssStrXX.h for spreadsheet private build
*/
ss_strgen (nodeP)
node *nodeP;
{
    char *cmdP;
    char src[256];

    DefaultOutputDir(nodeP);
    NodeFilename(ND_IVAL0,src,TRUE);
    cmdP = ThisCommand;
    sprintf(cmdP,"ss_strgen %s %s %s",src,NodeName(nodeP),NodeWithStr(nodeP));
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function:  ss_datagen - Create SSdata.sw for spreadsheet
*/
ss_datagen (nodeP)
node *nodeP;
{
    char *cmdP;
    char dst[256];
    char *NodeDirName();

    SetDir(NodeDirName(ND_IVAL0));
    NodeFilename(nodeP,dst,TRUE);
    cmdP = ThisCommand;
    sprintf(cmdP,"ss_datagen %s",dst);
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function:  namelist - make a global varibales name list
*/
namelist (nodeP)
node *nodeP;
{
    char *cmdP;
    char dst[256];

    DefaultOutputDir(nodeP);
    NodeFilename(nodeP,dst,TRUE);
    cmdP = ThisCommand;
    sprintf(cmdP,"namelist %s",dst);
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function: copy_file - Make a duplicate of a file under a new name
*/
copy_file (nodeP)
node *nodeP;
{
    char src[256];
    char *dst;

    SetNodeDir(nodeP);
    NodeFilename(ND_IVAL0,src,TRUE);
    dst = NodeName(nodeP);
    unlink(dst);
    return(AmCopyFile(src,dst));
}

/*************************************************************************
function: link_or_copy - Make a duplicate, use a link if possible
*/
link_or_copy (nodeP)
node *nodeP;
{
    char src[256];
    char *dst, *srcRel;
    char dstfile[STRPATH+1];
	int status;

    SetNodeDir(nodeP);
    NodeFilename(nodeP,dstfile,TRUE);
    NodeFilename(ND_IVAL0,src,TRUE);
    dst = FindEntryName(dstfile);
    unlink(dst);
#if !AXMS
	/* find source relative to destination */
	srcRel = SrcRelativeToDst(dstfile,src);
	if (srcRel) {
    	if ( 0 == symlink(srcRel,dst) ) {
			free(srcRel);
       	 	return(0);
		}
	}
#endif
    status = AmCopyFile(src,dst);
	return(status);
}

/*************************************************************************
function:  same _file- Make a link to or copy of a file
*/
same_file (nodeP)
node *nodeP;
{
    char *cmdP;
    char src[STRPATH+1];
    char dst[STRPATH+1];

    ThisCommand[0] = 0;
    SetNodeDir(nodeP);
    NodeFilename(ND_IVAL0,src,TRUE);
    NodeFilename(nodeP,dst,TRUE);
    if ( FilesIdentical(src,dst) ) {
        VERBOSE(("%s already same as %s\n",dst,src));
        return(0);
    }
    unlink(dst);
    CMPTELL(("Copying %s into %s\n",src,dst));
    return(AmCopyFile(src,dst));
}

/*************************************************************************
function:  elf_compile - Compile a ".els" or into a ".elo"
*/
elf_compile (nodeP)
node* nodeP;
{
    char *cmdP;
    node *src;
    char srcfile[STRPATH+1], objfile[STRPATH+1], strfl[STRPATH+1];

    src = ND_IVAL0;
    SetNodeDir(nodeP);
    cmdP = ThisCommand;
    NodeFilename(src,srcfile,TRUE);
    NodeFilename(nodeP,objfile,TRUE);
    NodeFilename(LookupNode("axstrfl"),strfl,TRUE);
    sprintf(cmdP,"elfc %s -x -o %s -s %s",srcfile, objfile, strfl);
    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, FALSE);
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function:  bldr_compile - Compile a ".ab" into a ".abo"
*/
bldr_compile (nodeP)
node* nodeP;
{
    char *cmdP;
    node *src;
    char objfile[STRPATH+1], amfile[STRPATH+1];
	char *progname = "elfmain";

    src = ND_IVAL0;
    SetNodeDir(nodeP);
    cmdP = ThisCommand;
    NodeFilename(LookupNode("databld.am"),amfile,TRUE);
    NodeFilename(nodeP,objfile,TRUE);
    sprintf(cmdP,"%s -test -quiet %s",progname,NodeWithStr(nodeP));
    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, FALSE);
	if ( Verbose )
		strcat(cmdP," -amake_verbose");
    BUMP(cmdP);
	if ( IsDebugBuild() )
		strcat(cmdP," -BLD_ELF_DEBUG");
	else
		strcat(cmdP," -BLD_NO_ELF_DEBUG");
    BUMP(cmdP);
    sprintf(cmdP," -outfile %s %s",objfile,amfile);
    BUMP(cmdP);
	ListInFiles(nodeP,cmdP,"");
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function:  elf_portable - Compile a ".els" or into a ".elo"
*/
elf_portable (nodeP)
node* nodeP;
{
    char *cmdP;
    node *src;
    char srcfile[STRPATH+1], objfile[STRPATH+1], strfl[STRPATH+1];

    src = ND_IVAL0;
    SetNodeDir(nodeP);
    cmdP = ThisCommand;
    NodeFilename(src,srcfile,TRUE);
    NodeFilename(nodeP,objfile,TRUE);
    NodeFilename(LookupNode("axstrfl"),strfl,TRUE);
    sprintf(cmdP,"elfc %s -x -o %s -s %s -portable",srcfile, objfile, strfl);
    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, FALSE);
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function:  cat_sources - cat some source files into a single file
*/
cat_sources (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char *p;

    SetNodeDir(nodeP);
    cmdP = ThisCommand;
#if AXMS
    /* to cat on Windows 95 we need to do:
     * copy src1 + src2 + src3 + ... dst
     */
    strcpy(cmdP, "copy ");
    ListInFilesForWindowsCat(nodeP,0,BUMP(cmdP),NodeDirName(nodeP));
    strcat(cmdP, " ");
#else
    strcpy(cmdP,"cat ");
    ListInFiles(nodeP,BUMP(cmdP),NodeDirName(nodeP));
    strcat(cmdP," > ");
#endif

    NodeFilename(nodeP,BUMP(cmdP),FALSE);

    VERBOSE(("%s\n",ThisCommand));
    system(ThisCommand);
    return(0);
}

/*************************************************************************
function:  cmacs - Perform cmacs preprocessing of a document source
*/
cmacs (nodeP)
node *nodeP;
{
    char src[256];
    char *cmdP;

    SetNodeDir(nodeP);
    cmdP = ThisCommand;
    NodeFilename(ND_IVAL0,src,TRUE);
    sprintf(cmdP,"cmacs -f%s %s %s ",NodeName(nodeP), NodeWithStr(nodeP),src);
    GetIncludes(nodeP,BUMP(cmdP), FALSE);
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function: bldint - Build ApplixInt
*/
bldint (nodeP)
node *nodeP;
{
    char src[256];

    SetNodeDir(nodeP);
    NodeFilename(ND_IVAL0,src,TRUE);
    sprintf(ThisCommand,"bldint %s %s",src,NodeName(nodeP));
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function: lex - Execute lex program
*/
lex (nodeP)
node *nodeP;
{
    char src[256];
    int code;

    SetNodeDir(nodeP);
    NodeFilename(ND_IVAL0,src,TRUE);
    sprintf(ThisCommand,"lex %s",src);
    code = proc(ThisCommand, FALSE);
    if ( !code )
        Arename("lex.yy.c",NodeName(nodeP));
    else
        unlink("lex.yy.c");
    return(code);
}

/*************************************************************************
function: strfl_incl - Generate stringfile include files
*/
strfl_incl (nodeP)
node *nodeP;
{
    char src[STRPATH+1], dst[STRPATH+1];

    SetNodeDir(nodeP);
    NodeFilename(ND_IVAL0,src,TRUE);
    NodeFilename(nodeP,dst,TRUE);
    sprintf(ThisCommand,"sgenincl %s %s",src,dst);
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function: strfl_file - Generate Applixstrfl
*/
strfl_file (nodeP)
node *nodeP;
{
    char src[256];

    SetNodeDir(nodeP);
    NodeFilename(ND_IVAL0,src,TRUE);
    sprintf(ThisCommand,"sgenfile %s %s",src, NodeWithStr(nodeP) );
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function: AppendExpandedWithStr - Expand to paths if -expand_paths <names> present
*/
void AppendExpandedWithStr (node *nodeP,char *cmdP) {
	char *withstr, *p;
	char copy_withstr[10000];	/* plenty */
	char *names[1000];
	int num_names = 0, ix,toklen;
	char filename[STRPATH+1];
	char tok[STRPATH+1];

	withstr = NodeWithStr(nodeP);
	if ( !withstr || !withstr[0] )
		return;

	BUMP(cmdP);
	p = strstr(withstr,"-expand_paths");
	if ( !p ) {
		sprintf(cmdP," %s",withstr);
		return;
	}
/* capture node names */

	bld_assert(strlen(p)<10000,("withstr too long"));
	strcpy(copy_withstr,p+strlen("-expand_paths "));
	p = copy_withstr;
	while ( *p ) {
		while ( *p && *p == ' ' )
			p++;
		bld_assert(num_names<1000,("withstr too long"));
		names[num_names++] = p;
		while ( *p && *p != ' ' )
			p++;
		if ( *p )
			*p++ = 0;
	}
	
/* convert simple names in withstr to full pathnames */

	while ( *withstr ) {
		while ( *withstr && *withstr == ' ' )
			withstr++;
		for ( toklen = 0; *withstr && *withstr != ' '; withstr++, toklen++ )
			tok[toklen] = *withstr;
		tok[toklen] = 0;
		if ( toklen != 0 ) {
			if ( streq(tok,"-expand_paths") )
				return;
			for ( ix = 0; ix < num_names; ix++ ) {
				if ( streq(names[ix],tok) )
					break;
			}
			if ( ix != num_names ) {
			    NodeFilename(LookupNode(tok),filename,TRUE);
				sprintf(cmdP," %s",filename);
			}
			else
				sprintf(cmdP," %s",tok);
			BUMP(cmdP);
		}
	}

}


/*************************************************************************
function:  run_tool - Execute a tool
*/
run_tool (nodeP)
node *nodeP;
{
    char dst[STRPATH+1];
	char cmd[STRPATH+1];
    char *dir, *p;

    DefaultOutputDir(nodeP);
    dir = DefaultDirName(nodeP);
    assert(ND_INCOUNT >= 1);
    NodeFilename(nodeP,dst,TRUE);
    NodeFilename(ND_IVAL0, cmd, TRUE); 
	strcpy(ThisCommand,cmd);
	AppendExpandedWithStr(nodeP,ThisCommand);
	if ( !Verbose && (p = strstr(ThisCommand,"-amake_verbose")) )
		strcpy(p,p+strlen("-noamk_verbose"));
    sprintf(ThisCommand+strlen(ThisCommand)," %s",dst);
    ListSomeInFiles(nodeP,1,ThisCommand + strlen(ThisCommand),dir);
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function:  run_tool_loc - Execute a tool in the output directory
*/
run_tool_loc (nodeP)
node *nodeP;
{
    char dst[STRPATH+1];
    char *dir, *p;

    assert(ND_INCOUNT >= 1);
    NodeFilename(nodeP,dst,TRUE);
    SetNodeDir(nodeP);
       dir = DefaultDirName(nodeP);
	strcpy(ThisCommand,NodeName(ND_IVAL0));
	AppendExpandedWithStr(nodeP,ThisCommand);
	if ( !Verbose && (p = strstr(ThisCommand,"-amake_verbose")) )
		strcpy(p,p+strlen("-noamk_verbose"));
    sprintf(ThisCommand+strlen(ThisCommand)," %s",dst);
    ListSomeInFiles(nodeP,1,ThisCommand + strlen(ThisCommand),dir);
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function: sub_amake - Run a subordinate (from built source tree) amake
*/
sub_amake (node *nodeP)
{
	char *amake_dir;
	int code;
    char dst[STRPATH+1];

    DefaultOutputDir(nodeP);
    NodeFilename(nodeP,dst,TRUE);
	amake_dir = NodeDirName(ND_IVAL0);
	sprintf(ThisCommand,"%s/amake %s",amake_dir,NodeWithStr(nodeP));
	TELL(("Starting %s...\n",ThisCommand));	/* otherwise log gets VERY confusing! */
	code = proc(ThisCommand, FALSE);
	TELL(("Completed %s...\n",ThisCommand));	/* otherwise log gets VERY confusing! */
	if ( code == 0 ) {
		PretendBuild = FALSE;
		TouchFile(dst);
	}
	return(code);
}

/*************************************************************************
function: java_archive - Produce a jar file
*/
java_archive (node *nodeP)
{
    char dst[STRPATH+1];
    char compilerPath[STRPATH+1], tmpFile[STRPATH+1];
    char *dir;
	char *cmdP;
	int code;

    DefaultOutputDir(nodeP);
    dir = DefaultDirName(nodeP);
    NodeFilename(nodeP,dst,TRUE);
    cmdP = ThisCommand;
	NodeFilename(LookupNode(FindDefine("JDK_NAME")),compilerPath,TRUE);
	if ( Verbose )
		sprintf(ThisCommand,"%s/bin/jar -cvf %s",compilerPath,dst);
	else
		sprintf(ThisCommand,"%s/bin/jar -cf %s",compilerPath,dst);
	AxSysDosUnixPath(ThisCommand);
	BUMP(cmdP);
	ListInFiles(nodeP,cmdP,dir);
	BUMP(cmdP);
	if ( GetSimpleDtmod("images") )
		strcpy(cmdP," images");

	if ( Verbose )
	    return(proc(ThisCommand, FALSE));
	else {
		GetTemp("jar_errs",tmpFile);
		code = PipeProc(ThisCommand, tmpFile);
		if ( code != 0 )
			StderrFile(tmpFile);
		unlink(tmpFile);
		return(code);
	}
}

int StderrFile(char *name) {
    int err, got;
	char buf[8096];

	err = open(name,0,0);
    while ( (got = read(err,buf,8096)) > 0 ) fwrite(buf,got,1,stderr);
    fflush(stderr);
    close(err);
}

/*************************************************************************
function: genshlibintf - Run PERL script to enerate shared library files
*/
genshlibintf (node *nodeP) {
	char protofilename[STRPATH+1];
	char perlfilename[STRPATH+1];
	char filename[STRPATH+1];
	char tok[STRPATH+1];
	int toklen, ix, numInputs;
	char *cmdP, *withstr, *p;


    SetNodeDir(nodeP);
    NodeFilename(NodeInVal(nodeP,0),perlfilename,TRUE);
    sprintf(protofilename,"%s/protos.dct",GetCProtoDirName());
	sprintf(ThisCommand,"%s -proto_file %s",perlfilename,protofilename);
    cmdP = ThisCommand;
	BUMP(cmdP);
	withstr = NodeWithStr(nodeP);
    assert(strlen(withstr));
	numInputs = NodeInCount(nodeP);
/* convert simple names in withstr to full pathnames */
	while ( *withstr ) {
		while ( *withstr && *withstr == ' ' )
			withstr++;
		for ( toklen = 0; *withstr && *withstr != ' '; withstr++, toklen++ )
			tok[toklen] = *withstr;
		tok[toklen] = 0;
		if ( toklen != 0 ) {
			for ( ix = 0; ix < numInputs; ix++ ) {
				if ( streq(NodeInVal(nodeP,ix)->name,tok) )
					break;
			}
			if ( ix != numInputs ) {
			    NodeFilename(NodeInVal(nodeP,ix),filename,TRUE);
				sprintf(cmdP," %s",filename);
			}
			else
				sprintf(cmdP," %s",tok);
			BUMP(cmdP);
		}
	}
    NodeFilename(nodeP,filename,TRUE);
	sprintf(cmdP," %s",filename);
    return(proc(ThisCommand, FALSE));
}

/*************************************************************************
function: java_compile - Run current java compiler
The java compier is a funny animal, it peeks into class files to verify
consistancy.  This means that if two sources change in a dependent manner,
they need to be presented to the compiler at the saem time or neither will
compile.  This fights amake's one-rule per object model, so this rules
is a macro rule that should only run ONCE per build.

Compiles ALL updated '.java' files into .class files.
Updates ALL final .class files.
Sets datestamp on ALL .class files to prevent amake from triggering
recompiles.
*/
java_compile (nodeP)
node *nodeP;
{
	int buildAll = FALSE;
    char src[STRPATH+1], dst[STRPATH+1],dep[STRPATH+1],depc[STRPATH+1];
    int code;
	int numInputs;
	node *outputs[10000], *inP, *srcP, *otherP;
	int numOutputs, ix;
	node *src1;
	long newest;
	char *currdirname, *cmdP, *outdir1, *p, *depstring;
	ID id;
	int numDepends, px, num_paths;
	node *depP;
	char *new_paths[3000];
	char *depdir, old_classpath[30000],new_classpath[30000];
	char old_javahome[STRPATH+1];
	char compilerPath[STRPATH+1];

	src1 = NodeInVal(nodeP,0);
    SetNodeDir(src1);
	currdirname = NodeDirName(src1);
	outdir1 = strdup(NodeDirName(nodeP));

/* Produce a list of all of the output nodes */

	numOutputs = 0;
	outputs[numOutputs++] = nodeP;
	for ( id = NextType(0,TYPE_NODE); id != 0; id = NextType(id,TYPE_NODE) ) {
		otherP = NodePtr(id);
		if ( otherP->ruleid == TRUE_NOP_RULE && NodeInVal(otherP,0) == nodeP ) {
			bld_assert(numOutputs < 10000,("Too many outputs to java_compile rule"));
			outputs[numOutputs++] = otherP;
		}
	}

/* Verify inputs/outputs line up */
/* This should do main name test for balancing!!! */

	numInputs = NodeInCount(nodeP);
	if ( numInputs > numOutputs ) {
		ErrorMsg("java_compile requires at least as many outputs as inputs");
		return(1);
	}
	code = 0;
	for ( ix = 0; ix < numInputs; ix++ ) {
		p = NodeInVal(nodeP,ix)->name;
		if ( !streq(p+strlen(p)-5,".java") ) {
			ErrorMsg("sources %s does not end with .java",p);
			code++;
		}
		if ( 0 != strncmp(p,outputs[ix]->name,strlen(p)-4) ) {
			ErrorMsg("%s -> %s must namestem match, add excess classes at end of outputs list",p,nodeP->name);
			code++;
			break;
		}
	}
	if ( code != 0 )
		return(code);

/* Find newest output node */

	newest = 0;
	for ( ix = 0; ix < numOutputs; ix++ ) {
		StatNode(outputs[ix]);
		newest = MAX(newest,outputs[ix]->dtmod);
	}

	numDepends = NodeDepCount(nodeP);
	for ( ix = 0; ix < numDepends; ix++ ) {
		depP = NodeDepVal(nodeP,ix);
		StatNode(depP);
		if ( depP->dtmod > newest )
			buildAll = TRUE;
	}

/* Add to CLASSPATH if there are any .class dependencies */
/* 
	File dependecies:
		.class -> add directory holding the classfile
		.jar -> add the file
		directories -> add the directory, but if directory/classes exists,
						the the directory/classes instead - gross hack needed
						to support packages from third parties!
		others -> do not affect CLASSPATH
*/

	p = getenv("CLASSPATH");
	if ( !p )
		p = "";
	sprintf(old_classpath,"CLASSPATH=%s",p);

	p = getenv("JAVA_HOME");
	if ( !p )
		p = "";
	sprintf(old_javahome,"JAVA_HOME=%s",p);

	numDepends = NodeDepCount(nodeP);
	num_paths = 0;
	new_paths[num_paths++] = strdup(outdir1);
	for ( ix = 0; ix < numDepends; ix++ ) {
		depP = NodeDepVal(nodeP,ix);
		depdir = NodeDirName(depP);
		NodeFilename(depP,dep,TRUE);
		depstring = NULL;
		if ( (strlen(depP->name) > 6) && streq(".class",depP->name+strlen(depP->name)-6) )
			depstring = depdir;
		else if ( (strlen(depP->name) > 4) && streq(".jar",depP->name+strlen(depP->name)-4) )
			depstring = dep;
		else if ( DirExists(dep) ) {
			sprintf(depc,"%s/classes",dep);
			AxSysDosUnixPath(depc);
			if ( DirExists(depc) )
				depstring = depc;
			else
				depstring = dep;
		}
		if ( depstring ) {
			for ( px = 0; px < num_paths; px++ )
				if ( streq(new_paths[px],depstring) )
					break;
			if ( px == num_paths )
				new_paths[num_paths++] = strdup(depstring);
		}
	}
	strcpy(new_classpath,"CLASSPATH=");
	for ( ix = 0; ix < num_paths; ix++ ) {
		strcat(new_classpath,new_paths[ix]);
#if AXMS
		strcat(new_classpath,";");
#else
		strcat(new_classpath,":");
#endif
		free(new_paths[ix]);
	}
	new_classpath[strlen(new_classpath)-1] = 0;	/* toss trailing ':' */

/* build command, including inputs newer than outputs */

    cmdP = ThisCommand;
	NodeFilename(LookupNode(FindDefine("JDK_NAME")),compilerPath,TRUE);
	sprintf(ThisCommand,"%s/bin/javac -d %s -J-mx32m ",compilerPath,outdir1);
	AxSysDosUnixPath(ThisCommand);
	BUMP(cmdP);

	/*ok to free now*/
	free(outdir1);

	for ( ix = 0; ix < numInputs; ix++ ) {
		inP = NodeInVal(nodeP,ix);
		StatNode(inP);
		if ( buildAll || (inP->dtmod > newest) || (inP->dtmod > outputs[ix]->dtmod) ) {
			if ( NodeDirName(inP) == currdirname )
				strcpy(src,inP->name);
			else
				NodeFilename(inP,src,TRUE);
			CMPTELL(("...compiling %s\n",inP->name));
			strcpy(cmdP,src);
			BUMP(cmdP);
			*cmdP++ = ' ';
		}
	}
	*cmdP = 0;
	putenv(new_classpath);
	putenv("JAVA_HOME=");
	VERBOSE(("Using %s\n",new_classpath));
	code = proc(ThisCommand, FALSE);
	putenv(old_classpath);
	putenv(old_javahome);
	if ( code != 0 )
		return(code);

/* touch all of the output files */

	for ( ix = 0; ix < numOutputs; ix++ ) {
		NodeFilename(outputs[ix],dst,TRUE);
		TouchFile(dst);
		ReStatNode(outputs[ix]);
	}
	return(0);
}

/*************************************************************************
function:  yacc - Execute yacc program
*/
yacc (nodeP)
node *nodeP;
{
    char src[256];
    char c_out[256], h_out[256];
    int code;

    DefaultOutputDir(nodeP);
    assert(strlen(NodeWithStr(nodeP)));
    NodeFilename(ND_IVAL0,src,TRUE);
    sprintf(ThisCommand,"yacc -dv %s",src);
    if ( 0 != (code = proc(ThisCommand, FALSE)) )
        return(code);
    NodeFilename(nodeP,c_out,TRUE);
    Arename("y.tab.c",c_out);
    VERBOSE(("created %s from y.tab.c\n",c_out));
    NodeFilename(LookupForMe(NodeWithStr(nodeP),NodeCompId(nodeP)),h_out,TRUE);
    Arename("y.tab.h",h_out);
    VERBOSE(("created %s from y.tab.h\n",h_out));
    return(0);
}

/*************************************************************************
function: nop_rule - Do nothing except create a (zero length) tag file
*/
nop_rule (nodeP)
node *nodeP;
{
    char dst[STRPATH+1];
    FILE *fp;

    SetNodeDir(nodeP);
    NodeFilename(nodeP,dst,TRUE);
    fp = fopen(dst,"w");
    if ( fp ) {
        fclose(fp);
        return(0);
    }
    else
        return(1);
}


/*************************************************************************
function: translate - remove the $TR strings from source files
*/
do_translate (nodeP)
node *nodeP;
{
    char *cmdP;
    node *src;
    char srcfile[300];
    char *p;
    int code;

#ifdef DBXTOOL
    code = 99;
    while ( code == 99 )
       sleep(4);
#endif
    src = ND_IVAL0;
    DefaultOutputDir(nodeP);
    NodeFilename(src,srcfile,TRUE);
    cmdP = ThisCommand;
    sprintf(cmdP,"remove_TR %s", srcfile);  /* output is file.NTR */
    code = proc(ThisCommand, FALSE);
    return( code );

}






#define NUM_NAMES 250
#define MAX_NAME_LEN 50

file_in_debug_list( in_srcfile )
char *in_srcfile;
{
    static inited = 0;
    static char names[NUM_NAMES][MAX_NAME_LEN];
    static index;
    int i;
    FILE *ff;
    char *p;
    char debug_list_file[120];
    char srcfile[1025];

    if ( inited == 2 )
       {
       /* no debug file -==> do not build anything debug */
       return( FALSE );
       }
    if ( !inited )
       {
       p = getenv("DEBUG_LIST_FILE");
       if ( p != (char *)NULL )
          strcpy( debug_list_file, p );
       else
          AMakePath(debug_list_file,MyBuildTree,"debug_list");
       TELL(("Using debug list file %s\n", debug_list_file));
       ff = fopen( debug_list_file, "r" );
#if AXMS
       if ( ff == (FILE *)0 )
          {
          /* for backwards compatibility, look for the old name */
          TELL(("Trying debug list file \\ax\\deb_list \n"));
          ff = fopen( "\\ax\\deb_list", "r" );
          }
#endif
       if ( ff == (FILE *)0 )
          {
          /* no debug file no debug stuff */
          inited = 2;
          return( FALSE );
          }
       index = 0;
       while ( fgets( names[index], MAX_NAME_LEN, ff ) != (char *)0 )
          {
          p = names[index] + strlen(names[index]) - 1;
          *p = '\0';    /* step on carriage return */
          if ( strlen( names[index] ) == 0 )
             /* blank line */
             continue;
#if AXMS
          p = names[index];
          while ( *p != '\0' )
             {
             if ( islower( *p ) )
                *p = toupper( *p );
             p++;
             }
#endif
          index++;
          }
       inited = 1;
       fclose ( ff );
       }

    strcpy( srcfile, in_srcfile );
#if AXMS
    /* mash source to upper case for "strstr" below */
    p = srcfile;
    while ( *p != '\0' )
       {
       if ( islower( *p ) )
          *p = toupper( *p );
       p++;
       }
#endif

    for ( i=0; i<index; i++ )
       {
       p = strstr( srcfile, names[i] );
       if ( p != (char *)NULL )
          return( TRUE );
       }
    return( FALSE );
}


file_in_purify_list( srcfile )
char *srcfile;
{
    static inited = 0;
    static char names[NUM_NAMES][MAX_NAME_LEN];
    static index;
    int i;
    FILE *ff;
    char *p;

    if ( inited == 2 )
       {
       /* no purify list file -==> do not build anything purify */
       return( FALSE );
       }
    if ( !inited )
       {
       ff = fopen( "/ax/purify_list", "r" );
       if ( ff == (FILE *)0 )
          {
          /* no purify file no purify stuff */
          inited = 2;
          return( FALSE );
          }
       index = 0;
       while ( fgets( names[index], MAX_NAME_LEN, ff ) != (char *)0 )
          {
          p = names[index] + strlen(names[index]) - 1;
          *p = '\0';    /* step on carriage return */
          if ( strlen( names[index] ) == 0 )
             /* blank line */
             continue;
          index++;
          }
       inited = 1;
       fclose ( ff );
       }

    for ( i=0; i<index; i++ )
       {
       if ( strcmp( srcfile, names[i] ) == 0 )
          return( TRUE );
       }
    return( FALSE );
}


#if AXMS        /*  {  */

file_in_console_list( srcfile )
char *srcfile;
{
    static inited = 0;
    static char names[NUM_NAMES][MAX_NAME_LEN];
    static index;
    char line[300];
    char file_name[300];
    node *nodeP;
    int i;
    FILE *ff;
    char *p;
    ID id;

    if ( inited == 2 )
       {
       /* no list file -==> build anything console */
       return( TRUE );
       }
    if ( !inited )
       {
       nodeP = CrtOrFindNode( "win_list.mak", FALSE, 0, 0 );
       if ( nodeP != (node *)NULL )
          {
          /* found it */
          NodeFilename(nodeP,file_name,TRUE);
          }
       else
          {
          /* not found, try hard default */
          strcpy( file_name, "/ax/pgms/axbuild/win_list.mak" );
          }
       ff = fopen( file_name, "r" );
       if ( ff == (FILE *)0 )
          {
          /* no file, build everything console */
          inited = 2;
          return( TRUE );
          }
       index = 0;
       while ( fgets( line, sizeof(line), ff ) != (char *)0 )
          {
          if ( line[0] == '#' )
             /* comment line */
             continue;
  
          /* a name to remember */
          p = line + strlen(line) - 1;
          while ( *p <= ' ' )
             { 
             *p = '\0';    /* step on any CR or NL or trailing blank */
             p--;
             }

          if ( strlen( line ) == 0 )
             /* blank line */
             continue;
          strcpy( names[index], line );
          index++;
          }
       inited = 1;
       fclose ( ff );
       }

    for ( i=0; i<index; i++ )
       {
       if ( _stricmp( srcfile, names[i] ) == 0 )
          return( TRUE );
       }
    return( FALSE );
}


#endif   /*  } AXMS  */


do_purify (nodeP)
node *nodeP;
{
#if SUN4 | SUN4S
    char *cmdP;
    node *src;
    char srcfile[STRPATH+1];
    char dstfile[STRPATH+1];
    char linker[100];
    char DebugFlag[100];
    char *p;

    /* initialize the linker name */
    strcpy( linker, AosFinalCPlus);

    /* initialize the DebugFlag name */
    strcpy( DebugFlag, AosDebugFlag );


    NodeFilename(nodeP,dstfile,TRUE);
    p = strrchr( dstfile, '/' );
    if ( p == (char *)NULL )
       p = dstfile;
    else
       p++;    /* skip the slash */

    if ( !file_in_purify_list( p ) )
       return( 0 );

    SetNodeDir(nodeP);

    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;

    strcpy( cmdP, "purify -first_only " );
    BUMP(cmdP);

    sprintf(cmdP,"%s %s %s -o %s.purifed", linker,AosFnlflags,DebugFlag,dstfile);

    BUMP(cmdP);
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    strcat(cmdP," ");
    strcat(cmdP,NodeWithStr(nodeP));
    strcat(cmdP," ");
    strcat(cmdP,AosLibs);

    /* if we are linking with "tmzmallo.o" we need to replace this with */
    /* "tmzmpure.o", the version of the malloc code built for purify     */
    p = strstr( ThisCommand, "tmzmallo.o" );
    if ( p != (char *)NULL )
       {
       /* replace tmzmallo.o  */
       strncpy( p, "tmzmpure", 8 );
       }

    return(proc(ThisCommand, FALSE));
#else
    return( 0 );
#endif
}


/*************************************************************************
function: tarfile - Create tar file
The input node is only used to determine the directory to run in
Processes the withstr for input names to allow for paths...
*/
tarfile (node *nodeP) {
	char dst[STRPATH+1];
	char *cmdP;

    SetDir(NodeDirName(NodeInVal(nodeP,0)));
	strcpy(ThisCommand,"tar cf");
	if ( Verbose )
		strcat(ThisCommand,"v");
	cmdP = ThisCommand;
    BUMP(cmdP);
    NodeFilename(nodeP,dst,TRUE);
	sprintf(cmdP," %s %s",dst,NodeWithStr(nodeP));
    BUMP(cmdP);
    return(proc(ThisCommand, FALSE));
}


/*************************************************************************
function: untarcompressed - remove_header, uncompress and extract all from 
a .tar file.
Note: this rule builds a directory containing the contents of the tar.
Individual files do not participate in the amake dependency process.
*/
untarcompressed (nodeP)
node *nodeP;
{
    int ret;
    ret = untar_common (nodeP, TRUE);

    return( ret );
}


/*************************************************************************
function: untar - hex to bin, then extract all from a .tar file.
Note: this rule builds a directory containing the contents of the tar.
Individual files do not participate in the amake dependency process.
*/
untar (nodeP )
node *nodeP;
{
    int ret;
    ret = untar_common (nodeP, FALSE);

    return( ret );
}



untar_common (nodeP, compressed_source)
node *nodeP;
int compressed_source;
{
    char tmp[STRPATH+1];
    char dst[STRPATH+1];
    char filename[STRPATH+1];
    char dirname[STRPATH+1];
    char pathname[STRPATH+1];
    char targetDir[STRPATH+1];
    char *dir, *finddir;
    char *ranlibCommand;
    int status,ret, gotanA,len;

    SetNodeDir(nodeP);
    dir = NodeDirName(nodeP);
    assert(ND_INCOUNT >= 1);
    NodeFilename(nodeP,dst,TRUE);

    /* hex to bin into a temp file */
    GetTemp("x",tmp);
    if ( compressed_source )
        {
        strcat( tmp, ".Z" );   /* uncompress insists on a ".Z" */
        sprintf(ThisCommand, "head_off %s ",tmp);
        }
    else
        /* older uncompresed style used hex2bin */
        sprintf(ThisCommand, "hex2bin %s ",tmp);

    ListSomeInFiles(nodeP,0,ThisCommand + strlen(ThisCommand),dir);
    status = proc(ThisCommand, FALSE);
    if (status) goto cleanup;

    /* node is actually a directory that contains the stuff extracted
    from the tar file */
    AMakePath(targetDir,dir,NodeSysSuffix(nodeP));
    SetDir(targetDir);

    /* tar into destination directory */
#if AX_X11
    if ( compressed_source )
        {
        sprintf(ThisCommand, "uncompress -c %s | tar -xf -",tmp);
        status = system( ThisCommand );
        }
    else
#endif
        {
        sprintf(ThisCommand, "tar -xf %s",tmp);
        status = proc(ThisCommand, FALSE);
        }
	TouchFile(targetDir);
    if (status) goto cleanup;

    /* ranlib all .a files found in destination directory 
    if there is a "with ranlib" clause */
    if (strcmp(NodeWithStr(nodeP), "ranlib") == 0) {
        ranlibCommand = NeedVar("AOS_RANLIB");
        if ( strcmp(ranlibCommand, "") == 0 )
            goto cleanup;
        strcpy(ThisCommand,ranlibCommand);
        finddir =  targetDir;
        gotanA = FALSE; /* set when archive file is found */
        while (TRUE) {
            ret = AFindFiles(finddir,8,dirname,filename);
            finddir = NULL;
            if (ret == 3) continue; /* sub-directory */
            if (ret == 1) break; /* end */
            if (ret != 0) continue; /* stat failed */
            /* got a file. if is ".a" file, add to ranlib command */
            len = strlen(filename);
            if (len <= 2 || strcmp(filename+len-2,".a") != 0) continue;
            AMakePath(pathname,dirname,filename);
            gotanA = TRUE;
            sprintf(ThisCommand + strlen(ThisCommand), " %s",pathname);
        }

        /* do ranlib */
        if (gotanA)
            status = proc(ThisCommand, FALSE);
    }

cleanup:
    unlink(tmp); /* delete temp file */
    return(status);
}





/*************************************************************************
function: sqlgate - build SQL database gateway using database toolkit and
DBMS vendor tools.
sqlgate <scriptname> <database libary pathname> into <gateway> with <make args>
*/
sqlgate (nodeP)
node *nodeP;
{
    node *inNodeP, *dblibNodeP, *elfapiNodeP;
    int status;
    char outfile[STRPATH+1];
    char tmpOutfile[STRPATH+1];
    char scriptfile[STRPATH+1];
    char dblib[STRPATH+1];
    char *NodeDirName(),*FindEntryName(),*NodeWithStr(),*NodeSysSuffix();

    /* pathname of gateway to build */
    NodeFilename(nodeP,outfile,FALSE);

    /* target for makefile */
    strcpy(tmpOutfile,FindEntryName(outfile));

    /* modify target for Oracle builds with different libraries */
    if (0 == strncmp(tmpOutfile,"elfora",6)) {
        strcpy(tmpOutfile, "elforacl"); /* makefile target for oracle */
    }


    /* pathname of script to run */
    inNodeP = NodeInVal(nodeP,0);
    NodeFilename(inNodeP,scriptfile,FALSE);

    /* pathname of database libraries */
    dblibNodeP = NodeInVal(nodeP,1);
    AMakePath(dblib,NodeDirName(dblibNodeP),NodeSysSuffix(dblibNodeP));

    /* change directory to location of script */
    SetDir(NodeDirName(NodeInVal(nodeP,0)));

    /* find location of elfapi.a */
    elfapiNodeP = CrtOrFindNode("elfapi.a",FALSE,0,0);
    if (!elfapiNodeP) return (1); /* has to exist */

    sprintf(ThisCommand,"%s -d%s -e%s %s %s",
        FindEntryName(scriptfile),  /* shell script to run make */
        dblib,                      /* DBMS_HOME */
        NodeDirName(elfapiNodeP),   /* ELF_LOC */
        NodeWithStr(nodeP),         /* other switches */
        tmpOutfile);                 /* target */

    status = proc(ThisCommand, FALSE);

    /* move built file to output directory */
    if(status == 0) {
        unlink(outfile);
        status = AmCopyFile(tmpOutfile, outfile);
        unlink(tmpOutfile);
    }

    /* delete temp files */
    sprintf(ThisCommand,"%s -d%s -e%s %s %s",
        FindEntryName(scriptfile),  /* shell script to run make */
        dblib,                      /* DBMS_HOME */
        NodeDirName(elfapiNodeP),   /* ELF_LOC */
        NodeWithStr(nodeP),         /* other switches */
        "clean");                /* target */
    proc(ThisCommand, FALSE); 

    return(status);
}

/*************************************************************************
function:  cplus_exec - Link files into an executable
*/

do_cplus_exec (nodeP, multi_thread)
node *nodeP;
int multi_thread;
{
    char *cmdP;
    node *src;
    char *p;
    char *name;
    char *fsize;
    char *segs = NULL;
    int is_dbx = FALSE;
    static char linker[100];
    static char DebugFlag[100];
    char *debugP;
    int ret;

#if AXMS   /*  {  */
    /** c++ is the same as the regular 'C' on NT */
    ret = do_cc_exec(nodeP, multi_thread);
    return( ret );
#else     /*  }  {  */

    if ( getenv("DO_PURIFY") != (char *)NULL )
        do_purify( nodeP );


    DefaultOutputDir(nodeP);
    cmdP = ThisCommand;

    if ( linker[0] == '\0' )
		{
        /* initialize the linker name */
			if ( multi_thread )
            	strcpy( linker, AosFinalCPlusMT );
			else
            	strcpy( linker, AosFinalCPlus );
        }
    if ( DebugFlag[0] == '\0' )
        {
        /* initialize the DebugFlag name */
        if ( IsDebugBuild() )
           {
              strcpy( DebugFlag, AosDebugFlag );
           }
        else
           strcpy( DebugFlag, " " );
        }
 
    sprintf(cmdP,"%s %s %s -o ", linker,AosFnlflags,DebugFlag);

    BUMP(cmdP);
    NodeFilename(nodeP,cmdP,FALSE);
    BUMP(cmdP);

    DoingLink = TRUE;      /* turn on ".DEF" handling for Windows */
    ListInFiles(nodeP,cmdP,DefaultDirName(nodeP));
    DoingLink = FALSE;

    strcat(cmdP," ");
    strcat(cmdP,NodeWithStr(nodeP));
    strcat(cmdP," ");
    strcat(cmdP,AosLibs);

#if LINUX
    BUMP(cmdP);
	sprintf(cmdP," -L%s",NeedFullVar("AOS_SHLIB"));
    BUMP(cmdP);
	sprintf(cmdP," -L%s/rts",NeedFullVar("AOS_OBJECTS"));
#endif
    p = getenv("AX_MORE_LIBS");
    if ( p != (char *)NULL )
       {
       strcat(cmdP," ");
       strcat(cmdP, p );
       }

    return(proc(ThisCommand, FALSE));
#endif    /* } not AXMS */
}

int cplus_exec(nodeP)
node *nodeP;
{
	return do_cplus_exec(nodeP, FALSE);
}

int cplus_exec_mt(nodeP)
node *nodeP;
{
	return do_cplus_exec(nodeP, TRUE);
}



int IsDebugBuild()
{
    char *p;
    
    if ( LimitedDebug )
      return( TRUE );

    p = GetVar( "TURBO" );
    if ( p && *p == '1' )
       return( FALSE );
    else
       return( TRUE );
}

/*************************************************************************
function: GetCProtoDirName - Return name of prototypes directory
*/
char *GetCProtoDirName ()
{
    static char *CprotoDir;         /* dir to put Cproto files */
    if ( !CprotoDir ) {
        CprotoDir = NodeDirName(LookupNode("cproto.tag"));
        bld_assert(CprotoDir,("Could not establish proto dir from cproto.tag"));
    }
    return(CprotoDir);
}


/*************************************************************************
function: GenCProtoFiles - generate C prototyping files into /of/obj/cproto
*/
GenCProtoFiles (id)
ID id;
{
    node *nodeP = NodePtr(id);
    char *cmdP, *dirP, *protoP, *dstP, *dstPathP;
    char srcfile[STRPATH+1], tmpFile[STRPATH+1];
    int ix, nspaces;
    extern int AllowDeletes;
    int code;
    register char *name;
    char *dirname;
    node *childP;
    char elfrt_str[30]; 
    char *elfrtP;
    node *src;
    static int Inited = FALSE;
    char outfile[256];
    char sfinalentry[256];
    char sfinalfile[256];
    char reffile[256];
    node *finalP, *sfinalP;
    extern int CProtosCreated;
    int created_new = FALSE;
    char macrostr[256];  /* macro name & options */
    char fullDir[256];
    char includer[256];
    char incdir[256];
    char cprotoexec[256];
    char specialName[STRPATH+1];
    char *fileBuf, *dataBuf, *newlineAt;
	char withstr[800], *withP;
    FILE *fp;

    if ( nodeP->haveProtoed )
        return(FALSE);
    nodeP->haveProtoed = TRUE;
    elfrtP = NeedVar("ELFRT_ONLY");
    if ( (*elfrtP == '\0') || (strcmp(elfrtP, "0") == 0 ) )
       /* normal build, do not include switch */
       elfrt_str[0] = '\0';
    else
       strcpy( elfrt_str, "-DELFRT_ONLY=1" );

    src = ND_IVAL0;
    NodeFilename(src,srcfile,TRUE);

	if ( SkipGeneratingProtos ) {
		TELL(("skipping prototype scan for %s\n", srcfile));
		return( FALSE ); 
	}

/* Create proto dir, paths for temp outfile xxx.z and permanent xxx.s */

    sprintf(fullDir,"%s%s%s",GetCProtoDirName(),DirSlash,GetComponent(nodeP));
    dirP = fullDir;
    CreateDir(dirP);
    AMakePath(outfile,dirP,nodeP->name);
    outfile[strlen(outfile)-1] = 'z';
    strcpy(sfinalfile,outfile);
    sfinalfile[strlen(sfinalfile)-1] = 's';
    strcpy(sfinalentry,FindEntryName(sfinalfile));
    sfinalP = LookupForMe(sfinalentry,0);
    sfinalP->dirid = LookupDir(dirP);
    sfinalP->located = TRUE;

/* create zero length .p so prototyper will not complain */

    strcpy(specialName,src->name);
    specialName[strlen(specialName)-1] = 'p';
    sprintf(includer,"%s/%s",DefaultDirName(nodeP),specialName);
    close(creat(includer,0777));

/* build 'references' filename for acproto to side-effect produce */

	strcpy(reffile,sfinalfile);
    reffile[strlen(reffile)-1] = 'r';

/* Fabricate includes if needed */

    sprintf(incdir,"%s%s%s",GetCProtoDirName(),DirSlash,"proto_incs");
    CreateDir(incdir);
	VfyProtoableIncludes(id,incdir);

/* generate prototypes for globals & statics */

    if ( CProtoMacro && CProtoMacro[0] )
        sprintf(macrostr, "-m -M %s -d", CProtoMacro);
	else
	    macrostr[0] = 0;
    NodeFilename(LookupNode("acproto"),cprotoexec,TRUE);
    cmdP = ThisCommand;
	withstr[0] = '\0';
	withP = NodeWithStr(nodeP);
	for (ix=0; *withP; ) { /* ignore with options except -D,-U,-I */
		if (*withP == '-' ) {
			int c = *(withP+1);
			if (c != 'D' && c != 'U' && c != 'I') {
				while(*withP && (*withP != ' ' && *withP != '\t')) withP++;
#if ALPHA
				if (c == 'x') {
				  nspaces = 0;
				  while(*(withP+nspaces) == ' ')
					nspaces++;
				  if (strncmp(withP+nspaces,"cxx",3) == 0) {
					withP += nspaces;
					while(*withP && (*withP != ' ' && *withP != '\t')) withP++;
				  }
				}
#endif
				continue;
			}
		}
		else if (*withP == '+') {
			while(*withP && (*withP != ' ' && *withP != '\t')) withP++;
			continue;
		}
		withstr[ix++] = *withP++;
	}
	withstr[ix] = '\0';
	
#if AXMS
    sprintf(cmdP,"%s -D_MSC_VER=1199 -q %s -s -o %s -R %s -DCPROTO -D%s -DDEBUG=%s -DSYS=%s %s %s %s %s ",
#else
    sprintf(cmdP,"%s -q %s -s -o %s -R %s -DCPROTO -D%s -DDEBUG=%s -DSYS=%s %s %s %s %s ",
#endif
            cprotoexec,
              macrostr,
              outfile,
              reffile,
              AosMachine,AosDebug,AosUnix,NeedVar("BLD_FLAGS"),
              NeedVar("AOS_FLAGS"),
              elfrt_str, withstr);
    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, FALSE);
    BUMP(cmdP);
	sprintf(cmdP," -I%s",incdir);
    BUMP(cmdP);
    sprintf(cmdP," %s",srcfile);

#if SUN4S
	if(DoingProfileBuild)	{
		/* doing a profile build, remove -xpg flag, can't c proto can't handle*/
		char *lP, *cP = strstr(cmdP, "-xpg");	
		if(cP != NULL)	{
			/* remove the -xpg from the command line */
			lP = cP + strlen("-xpg");
            memcpy( cP, lP, strlen(lP)+1);    /* copy the null char also */
		}
	}
#endif

	if ( Verbose )
	    code = proc(ThisCommand,FALSE);
	else {
		GetTemp("acp_errs",tmpFile);
		code = PipeProc(ThisCommand, tmpFile);
		if ( code != 0 )
			StderrFile(tmpFile);
		unlink(tmpFile);
	}

	FixBugOnVarargProtos(outfile);
    if ( !GetSimpleDtmod(sfinalfile) || !FilesIdentical(outfile,sfinalfile) ) {
        created_new = TRUE;
        Arename(outfile,sfinalfile);
        UpdateProtoTables(sfinalfile);
    }
    else {
        /* no change, just use the new files so that the date of creation */
        /* can be seen accuratly. - jon                                   */
        Arename(outfile,sfinalfile);
    }

/* Force later call to FindMyDeps so the ".r" will get processed */

	nodeP->deps_done = FALSE;

    return(created_new);
}

static char *varbugs[] = {"int va_alist",
                       "long va_alist",
                       "va_list va_alist",
                       "int __builtin_va_alist",
                       NULL};

/*************************************************************************
function: FixBugOnVarargProtos - hide proto error
*/
FixBugOnVarargProtos (filename)
char *filename;
{
	char *sgibug = "register ";
	int did_arg_fix = FALSE;
	char *p, *fileBuf;
	int ix;

    fileBuf = readFileIntoBuf(filename);
	if ( !fileBuf )
	{
		ErrorMsg("FixBugOnVarargProtos: could not read %s", filename);
		return 0;
	}
	/* convert "int __builtin_va_alist" to "..." */
	for ( ix = 0; varbugs[ix]; ix++ ) {
		p = fileBuf;
		while ( NULL != (p = strstr(p,varbugs[ix]) ) ) {
			char *beginP;
			char *endP = strchr(p,')');
			for ( beginP = p; beginP > fileBuf && *beginP != '('; beginP-- );
			if ( (*beginP=='(') && endP )
				memset(beginP+1,' ',endP-beginP-1);
			else {
				memset(p,' ',strlen(varbugs[ix]));
				p[0] = '.';
				p[1] = '.';
				p[2] = '.';
			}
			did_arg_fix = TRUE;
		}
	}
	p = fileBuf;
	while ( NULL != (p = strstr(p,sgibug) ) ) {
		strcpy(p,p+strlen(sgibug));
		did_arg_fix = TRUE;
	}
	if ( did_arg_fix ) {
		FILE *fp;
		fp = fopen(filename,"w");
		fwrite(fileBuf,strlen(fileBuf),1,fp);
		fclose(fp);
	}
	free(fileBuf);
}

/*************************************************************************
function:  mkshared_cc - Compile a ".c" into a ".o"
NOTE, If you make a generic change here, also make it in ansi_compile.
*/
static int
mkshared_compile (nodeP, cpp, multi_thread)
node* nodeP;
int cpp;
int multi_thread;
{
    char *cmdP;
    node *src;
    char srcfile[STRPATH+1];
    char dstfile[STRPATH+1];
    char tstname[STRLONG+1];
    int code;
    char *p;
    char *shccflags;
    char *ccflags;
    char asrcfile[STRPATH+1];
    char *sysname, *str1;
    char *GetComponent();
    char debug_flag[250];
    char optimize_flag[4];
    char *compiler;
    static char DebugFlag[250];
    char elfrt_str[30]; 
    char *elfrtP;
 
    elfrtP = NeedVar("ELFRT_ONLY");
    if ( (*elfrtP == '\0') || (strcmp(elfrtP, "0") == 0 ) )
       /* normal build, do not include switch */
       elfrt_str[0] = '\0';
    else
       strcpy( elfrt_str, "-DELFRT_ONLY=1" );

    
    if ( cpp )
	{
		if ( multi_thread )
			compiler = AosCppCompilerMT;
		else
        	compiler = AosCppCompiler;
	}
    else
	{
		if ( multi_thread )
        	compiler = IsLoose(nodeP) ? AosCompilerMT : AosStrictCompilerMT;
		else
        	compiler = IsLoose(nodeP) ? AosCompiler : AosStrictCompiler;
	}

    if ( DebugFlag[0] == '\0' )
        {
        /* initialize the DebugFlag name */
        if ( IsDebugBuild() )
           {
              strcpy( DebugFlag, AosDebugFlag );
           }
        else
           strcpy( DebugFlag, " " );
        }


    src = ND_IVAL0;
    NodeFilename(src,srcfile,TRUE);
    NodeFilename(nodeP,dstfile,TRUE);
    SetNodeDir(nodeP);

    if ( LimitedDebug ) {
        if ( file_in_debug_list( srcfile) ) {
            /* turn on debug mode for this file */
            strcpy( debug_flag, DebugFlag );
            optimize_flag[0] = '\0';      /* turn off optimization */
        }
        else  {
            debug_flag[0] = '\0';  /* turn off debug info */
            strcpy( optimize_flag, AosOptimize );
        }
    }
    else {
       /* do the normal build thing */
       strcpy( debug_flag, DebugFlag );
       strcpy( optimize_flag, AosOptimize );
    }

    cmdP = ThisCommand;
    shccflags = AosShlibCCFlags; 
    ccflags = cpp ? AosCPPflags : AosCCflags; 

    sysname = AosUnix;
    sprintf(cmdP,
        "%s -c %s %s %s %s %s %s -DDEBUG=%s -D%s -DSYS=%s %s %s ",
            compiler, debug_flag, ccflags, shccflags ,NeedVar("BLD_FLAGS"),
            NeedVar("AOS_FLAGS"),
            optimize_flag,AosDebug,AosMachine,sysname,
            elfrt_str, NodeWithStr(nodeP));
    BUMP(cmdP);
    GetIncludes(nodeP,cmdP, FALSE);
    strcat(cmdP," ");
    BUMP(cmdP);
#if AXMS
    if ( cpp ) {
        strcat(cmdP, "-Tp");  /* force compiler to treat file as C++ file */
        BUMP(cmdP);
    }
#endif 
    strcpy(cmdP,srcfile);
    ReSuffix(tstname,FindEntryName(srcfile),OEXT);
    if ( !cpp )
        GenProtosForMe(nodeP);

#if AXMS
    if ( streq(tstname,FindEntryName(dstfile)) )
#else
    if ( Cstreq(tstname,FindEntryName(dstfile)) )
#endif
        code = proc(ThisCommand, TRUE);
    else {
        SetDir(TempDir);
        code = proc(ThisCommand, TRUE);
        ReSuffix(tstname,FindEntryName(srcfile),OEXT);
        NodeFilename(nodeP,dstfile,TRUE);
        if ( code == 0  ) {
            VERBOSE(("renaming %s to %s\n",tstname,dstfile));
            Arename(tstname,dstfile);
        }
        else
            unlink(tstname);
    }
    return(code);
}

int
mkshared_cc(nodeP)
node *nodeP;
{
    return mkshared_compile(nodeP, FALSE, FALSE);
}

int
mkshared_cc_mt(nodeP)
node *nodeP;
{
    return mkshared_compile(nodeP, FALSE, TRUE);
}

int
mkshared_cpp(nodeP)
node *nodeP;
{
    return mkshared_compile(nodeP, TRUE, FALSE);
}

int
mkshared_cpp_mt(nodeP)
node *nodeP;
{
    return mkshared_compile(nodeP, TRUE, TRUE);
}
