/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: am_unlnk.c
   Applix ASC version: 2
   Date: Sun Jan 24 21:13:17 1999
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <errno.h>

#ifdef unlink
#undef unlink
#endif


extern char *sys_errlist[];

static char *cleanup_file = "/tmp/am_cleanup";

static int addcleanup();


/***********************************************************************
  PORT HERE!!
  Set `UNLINK_SAYS_ETXTBSY' to 1 if unlink(2) for this port can fail
  if the text file is busy;
  Default value is 0.
***********************************************************************/
#if (HP800V3)
#define UNLINK_SAYS_ETXTBSY 1
#else
#define UNLINK_SAYS_ETXTBSY 0
#endif

#if !UNLINK_SAYS_ETXTBSY
/* Simply call unlink(2). */
int
am_unlink(file)
char *file;
{
	return(unlink(file));
}

#else
/* Remove the directory entry named by FILE by calling unlink(2).
   unlink(2) returns -1, and sets errno to ETXTBSY if the text file is busy.
   If that is so, am_unlink will rename(2) the file to FILE ++ '#' and make
   a note of that in "/tmp/am_cleanup".  The user will need to clean
   up files recorded in "/tmp/am_cleanup" from time to time. */
int
am_unlink(file)
char *file;
{
	char dstfile[(1 + MAXNAMLEN) + 1];
	int  err;

	if (!file)
		return(-1);

	err = unlink(file);
	if (!err)
		return(0);

	if (errno == ETXTBSY) {
		sprintf(dstfile, "%s#", file);
		if (rename(file, dstfile) == -1) {
			fprintf(stderr, "am_unlink: rename %s %s: %s\n", file, dstfile,
					sys_errlist[errno]);
			return(err);
		}
		if (addcleanup(dstfile) == -1)
			fprintf(stderr, "*** %s moved to %s\n", file, dstfile);
		return(err);
	}
	return(err); /* some other error */
}

static int
addcleanup(file)
char *file;
{
	FILE *clfp;

	if (!file)
		return(-1);

	clfp = fopen(cleanup_file, "a+");
	if (!clfp) {
		fprintf(stderr, "open: %s: %s\n", cleanup_file, sys_errlist[errno]);
		return(-1);
	}
	fprintf(clfp, "%s\n", file);
	fclose(clfp);
	return(0);
}

#endif /* HP800V3 */

#if TEST_AM_UNLINK
/*
cc -DHP800V3 -DTEST_AM_UNLINK -o arm am_unlink.c
*/
main(argc, argv)
int  argc;
char *argv[];
{
	int ix;

	if (argc == 1)
		exit(0);
	for (ix = 1; ix < argc; ix++) {
		printf("%s...\n", argv[ix]);
		am_unlink(argv[ix]);
	}
	printf("done\n");
	exit(0);
}
#endif


/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
