/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: am_utils.c
   Applix ASC version: 43.2
   Date: Wed May 19 21:14:59 1999
*/


#include <am_bld.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#if !NO_VARARGS
#include <varargs.h>
#endif

#if AXMS
#include <io.h>
#endif

int CalledSetDir = FALSE;   /* Sanity check on rules */
int NodeCreated;            /* TRUE if most recent LookupNode created */
int DirObjCreated;          /* TRUE if most recent LookupDir created */
int CompObjCreated;         /* TRUE if most recent LookupComp created */
ID NodeIdLooked;            /* last node found/created by lookupnode */

extern int RunningNT3_1;
void AxSysDosUnixPath();
char *NodeName ();
extern int DoingLink;
extern char *strdup();
static int vErrorMsg();
extern char InputFilename[STRPATH+1];	/* file matching LineNumber */

/*************************************************************************
function: NodeSysSuffix - Gien a node, return its VMS/Unix file entryname
*/
char *NodeSysSuffix (nodeP)
node *nodeP;
{
    static char name[STRLONG+1], *sfx;

    if ( sfx = RuleArray[nodeP->ruleid].extension ) {
        ReSuffix(name,NodeName(nodeP),sfx);
        return(name);
    }
    else
        return(NodeName(nodeP));
}

/*************************************************************************
function:  DefaultOutputDir - Set dir to node's default dir
*/
DefaultOutputDir (nodeP)
register node *nodeP;
{
    if ( !SetDir(DefaultDirName(nodeP)) ) {
        nodeP->busted = TRUE;
        nodeP->up_to_date = TRUE;
    }
}

/*************************************************************************
function:  SetNodeDir - Set dir to node's dir
*/
SetNodeDir (nodeP)
register node *nodeP;
{
    if ( nodeP->dirid ) {
        if ( !SetDir(DirPtr(nodeP->dirid)->dname) ) {
            nodeP->busted = TRUE;
            nodeP->up_to_date = TRUE;
        }
    }
    else
        DefaultOutputDir(nodeP);
}

/*************************************************************************
function: CreateNodeDir - Create node's directory if it doesn't already exist
*/
CreateNodeDir (nodeP)
node *nodeP;
{
    if ( nodeP->dirid )
        CreateDir(DirPtr(nodeP->dirid)->dname);
    else
        CreateDir(DefaultDirName(nodeP));
}

/*************************************************************************
function: NodeDirName - Get name of node's directory or NULL if /tmp
*/
char *NodeDirName (nodeP)
register node *nodeP;
{
    register ID id;
    ID NodeDirId();

    id = NodeDirId(nodeP);
    return( id ? DirPtr(id)->dname : NULL );
}

/*************************************************************************
function: NodeIsShlib - return true if shlib is true on node
*/
int NodeIsShlib (nodeP)
register node *nodeP;
{
    register ID id;
    ID NodeDirId();

    id = NodeDirId(nodeP);
    return( id ? DirPtr(id)->is_shlib_dir : FALSE );
}

/*************************************************************************
function: NodeDirId - Get the ID of the dir a node is in
*/
ID NodeDirId (nodeP)
register node *nodeP;
{
    if ( nodeP->dirid )
        return(nodeP->dirid);
    else if ( nodeP->ruleid )
        return(CompPtr(nodeP->comp_id)->dflt);
    else
        return(0);
}

/*************************************************************************
function: DefaultDirName - Get name of node's default output directory

The NULL case is needed for unlocated nodes for the deltemps rule!!

*/
char *DefaultDirName (nodeP)
register node *nodeP;
{
    register ID compid;

    if ( !nodeP->dirid )
        return(NULL);
    if ( 0 == (compid = NodeCompId(nodeP)) )
        return(NULL);
    return(DirPtr(CompPtr(compid)->dflt)->dname);
}

/*************************************************************************
function: GetCIdFromName - Get component object given name
*/
ID GetCIdFromName (name)
char *name;
{
    ID id;

    for ( id = NextType(0,TYPE_COMP); id; id = NextType(id,TYPE_COMP) ) {
        if ( streq(CompPtr(id)->cname,name) )
            return(id);
    }
    return(0);
}

/*************************************************************************
function: LookupDir - Find/create directory object from name
Note that an object table search occurs here, so this shouldn't happen
too often - would otherwise need to chain these guys.
*/
ID LookupDir (name)
char *name;
{
    ID id;
    dir_info *p;

    DirObjCreated = FALSE;
    for ( id = NextType(0,TYPE_DIR); id; id = NextType(id,TYPE_DIR) ) {
        if ( streq(DirPtr(id)->dname,name) )
            return(id);
    }
    DirObjCreated = TRUE;
    id = AllocObj(sizeof(dir_info)+strlen(name),TYPE_DIR);
    p = DirPtr(id);
    memset((char*)p,0,sizeof(dir_info));
    AxSysDosUnixPath( name );
    strcpy(p->dname,name);
    return(id);
}

/*************************************************************************
function: LookupComp - Find/Create a component object from name
Note that an object table search occurs here, so this shouldn't happen
too often - would otherwise need to chain these guys.
*/
LookupComp (name)
char *name;
{
    ID id;
    comp_info *p;

    CompObjCreated = FALSE;
    for ( id = NextType(0,TYPE_COMP); id; id = NextType(id,TYPE_COMP) ) {
        if ( streq(CompPtr(id)->cname,name) )
            return(id);
    }
    CompObjCreated = TRUE;
    id = AllocObj(sizeof(comp_info)+strlen(name),TYPE_COMP);
    p = CompPtr(id);
    memset((char*)p,0,sizeof(comp_info));
    strcpy(p->cname,name);
    return(id);
}

/*************************************************************************
function:  SetDir - Change directory if needed
*/
SetDir (newdir)
register char *newdir;
{
    static char current_dir[STRPATH+1];
    int ix;
    extern int StatingWorld;
    extern int CalledSetDir;

    CalledSetDir = TRUE;


    if ( !streq(current_dir,newdir) ) {
#ifdef APOLLO
        CreateDir(newdir);
#else
        if ( 0 != (ix = CreateDir(newdir)) ) {
            ErrorMsg("could not create directory %s, %s",newdir, GenErrnoMsg(ix));
            return(FALSE);
        }
#endif
        strcpy(current_dir,newdir);
        if ( !StatingWorld ) {
            COMMANDLOG(("cd %s\n",newdir));
        }
        chdir(newdir);
        return(TRUE);
    }
    return(TRUE);
}



#if AXMS
/** if the filename is a ".DEF" file, change the name to include a "-DEF:" */
/** prefix, because the linker will not handle a ".DEF" file just by the   */
/*  filename suffix  (how stupid!)                                         */
void CheckForDefFile( char *buf )
{
    char *p;
    char filename[1026];

    p = buf + strlen(buf) - 4;
    if ( strcmp( p, ".def" ) == 0 )
       {
       /* add file with a special switch */
       strcpy( filename, buf );
       strcpy( buf, "-DEF:" );
       strcat( buf, filename );
       }
    return;
}
#endif


/*************************************************************************
function: ListInFiles - Given a node arraym create file list string
*/
ListInFiles (nodeP,buf,dirname)
node *nodeP;
char *buf;
char *dirname;     /* default (current) dir */
{
    ListSomeInFiles(nodeP,0,buf,dirname);
}

/*************************************************************************
function:  ListSomeInFiles - Given node & index, create file list string
*/
ListSomeInFiles (nodeP,ix,buf,dirname)
register node *nodeP;
register int ix;
register char *buf;
register char *dirname;     /* default (current) dir */
{
    char this_file[256];
    node *childP;
    int incount;

    incount = NodeInCount(nodeP);
    for ( ; ix < incount; ix++ ) {
        *buf++ = ' ';
        childP = NodeInVal(nodeP,ix);
        NodeFilename(childP,this_file,TRUE);
        if ( NodeIsShlib(childP) || NodeDirName(childP) == dirname )
            strcpy(buf,FindEntryName(this_file));
        else
            strcpy(buf,this_file);
#if AXMS
        if ( DoingLink )
           /* we need to handle ".DEF" files special */
           CheckForDefFile( buf );
#endif
        buf += strlen(buf);
    }
}

#if AXMS  /* { */
/* this function is simply to get cat_sources to work on Windows.
 * on NT we got away with: type src1 src2 ... > dst
 * but on Windows 95 we need to do:  copy src1 + src2 + ... dst
 * which also works on NT.      -donato
 */
void
ListInFilesForWindowsCat(register node *nodeP,
                        register int ix,
                        register char *buf,
                        register char *dirname)
{
    char this_file[256];
    node *childP;
    int incount;

    incount = NodeInCount(nodeP);
    for ( ; ix < incount; ix++ )
        {
        childP = NodeInVal(nodeP,ix);
        NodeFilename(childP,this_file,TRUE);
        if ( NodeDirName(childP) == dirname )
            strcpy(buf,FindEntryName(this_file));
        else
            strcpy(buf,this_file);
        buf += strlen(buf);
        if ( ix < (incount - 1) )
            {
            *buf++ = ' ';
            *buf++ = '+';
            *buf++ = ' ';
            }
        }
}
#endif  /* } AXMS */

/*************************************************************************
function:  ReStatNode - Re-get node status
*/
ReStatNode (nodeP)
node *nodeP;
{
    nodeP->dtmod = 0;
    StatNode(nodeP);
}

/*************************************************************************
function:  StatNode - Get file status of node if dtmod==0

Returns false on fatal stat failure

*/
StatNode(nodeP)
node *nodeP;
{
    int code;
    char name[256];
    struct stat status;

    DEBUG(("StatNode: %s",nodeP->name));

    if ( nodeP->dtmod != 0 ) {
        DEBUG(("StatNode: %s already %d",nodeP->name,nodeP->dtmod));
        return(TRUE);
    }

    NodeFilename(nodeP,name,TRUE);
    code = stat(name,&status);
    UpgradeToPretendTime(name,&status);

    if ( -1 == code ) {
        if ( errno == ENOENT ) {
            LOGGER(("StatNode: %s does not exist",name));
            nodeP->dtmod = 1;
            nodeP->exists = FALSE;
            return(TRUE);
        }
        else {
            LOGGER(("StatNode: %s errno %d",name,errno));
            nodeP->busted = TRUE;
            nodeP->up_to_date = TRUE;
            ErrorMsg("File access errno %d on file %s",errno,name);
            return(FALSE);
        }
    }
    nodeP->exists = TRUE;
    nodeP->dtmod = status.st_mtime;
    LOGGER(("StatNode: %s dtmod %ld",name,nodeP->dtmod));
    return(TRUE);
}

/*************************************************************************
function:  NodeFilename - Create filename from node info
If use_default is true, there will always be a full path.
If use_default is false, there will only be a path if the node
    does not lies in the component's defalt dir
*/
NodeFilename (nodeP,buf,use_default)
node *nodeP;
char *buf;
int use_default;
{
    register char *df, *sysname;

    sysname = NodeSysSuffix(nodeP);
    if ( !nodeP->dirid ) {
        AMakePath(buf,TempDir,sysname);
        return(FALSE);
    }
    else if ( use_default )
        AMakePath(buf,NodeDirName(nodeP),sysname);
    else if ( streq(df=NodeDirName(nodeP),DefaultDirName(nodeP)) )
        strcpy(buf,sysname);
    else
        AMakePath(buf,df,sysname);
    return(TRUE);
}

/*************************************************************************
function:  DeleteNodeFile - Delete a file
*/
DeleteNodeFile (nodeP)
node *nodeP;
{
    char name[256];

    NodeFilename(nodeP,name,TRUE);
    if ( GetSimpleDtmod(name) ) {

        /* for the untar rule, the output is a directory. delete
        entire directory. */
        if (strcmp(RuleArray[nodeP->ruleid].name, "untar") == 0) {
            char dirname[STRPATH+1],filename[STRPATH+1], pathname[STRPATH+1];
            char *finddir;
            int ret,dirDeleted;

            TRACE(("Deleting directory %s\n",name));
            /* delete files */
            finddir = name;
            while (TRUE) {
                ret = AFindFiles(finddir,8,dirname,filename);
                finddir = NULL;
                if (ret == 1) break; /* end */
                AMakePath(pathname,dirname,filename);
                if (ret == 0) {
                    unlink(pathname);
                }
            }

            /* delete directories. repeat until none are left. */
            do {
                dirDeleted = FALSE;
                finddir = name;
                while (TRUE) {
                    ret = AFindFiles(finddir,8,dirname,filename);
                    finddir = NULL;
                    if (ret == 1) break; /* end */
                    AMakePath(pathname,dirname,filename);
                    if (ret == 3) {
                        if (0 == rmdir(dirname)) dirDeleted = TRUE;
                    }
                }
            } while (dirDeleted);
            rmdir(name);
        }

        /* delete a file */
        else {
            TRACE(("Deleting file %s\n",name));
            unlink(name);
        }
        return(TRUE);
    }
    else
        return(FALSE);
}

/*************************************************************************
function:  AmCopyFile - Make a copy of a file
*/
AmCopyFile(srcname,dstname)
char *srcname, *dstname;
{
    int fd, fd2, num_read, num_writ;
#define MDBLK 8192
    char buf[MDBLK];

#if AXMS
    fd = open(srcname, O_RDONLY|O_BINARY);
#else
    fd = open(srcname,0);
#endif
    if ( fd == -1 )
        return(errno);
#if AXMS
    fd2 = open(dstname, (O_WRONLY | O_CREAT | O_TRUNC | O_BINARY),
                            (_S_IREAD | _S_IWRITE));
#else
    fd2 = creat(dstname,0777);
#endif
    if ( fd2 == -1 ) {
        close(fd);
        return(errno);
    }
    while ( num_read = read(fd,buf,MDBLK) ) {
        if ( num_read != write(fd2,buf,num_read) ) {
            close(fd);
            close(fd2);
            return(errno);
        }
    }
    close(fd);
    close(fd2);
    return(0);
}

/*************************************************************************
function: ExtractOutputs - get list of output nodes from a 'with' string
Returns the balance of the with string
Used only by the rules that have multiple unmatched outputs
*/
char *ExtractOutputs(outputs,str,max)
node *outputs[];
char *str;
int max;
{
    char *p;
    char name[100];
    char *dst;
    int count = 0;

    outputs[0] = NULL;
    if ( !str )
        return("");
    for ( p = str; *p; p++ ) {
        if ( *p == ' ' ) continue;
        else if ( *p == '/' ) {     /* hack! */
            if ( count == max ) {
                ErrorMsg("too many outputs, %d max",max);
                break;
            }
            p++;
            dst = name;
            while ( (*p != ' ') && *p )
                *dst++ = *p++;
            *dst = 0;
            outputs[count++] = LookupNode(name);
            VERBOSE(("multi-output sibling: %s\n",name));
            outputs[count] = NULL;
            if ( *p == 0 )
                break;
        }
        else break;
    }
    return(p);
}
/*************************************************************************
function: UpDateExcept - restat nodes in list except for one
Used only by the rules that have multiple unmatched outputs
*/
UpDateExcept(outputs,not_me)
node *outputs[];
node *not_me;
{
    register node **p;

    for ( p = outputs; *p; p++ ) {
        if ( *p != not_me ) {
            ReStatNode(*p);
            (*p)->up_to_date = TRUE;
        }
    }
}

/*************************************************************************
function: NodeCompId - get component id from a node
*/
NodeCompId (nodeP)
register node *nodeP;
{
    if ( nodeP->comp_id )
        return(nodeP->comp_id);
    else if ( nodeP->dirid )
        return(DirPtr(nodeP->dirid)->owner);
    else
        return(0);
}

/*************************************************************************
function:  GetComponent - Extract component name from node
*/
char *GetComponent (nodeP)
node *nodeP;
{
    register char *compid;

    return(GetCompStr(NodeCompId(nodeP)));
}

/*************************************************************************
function:  GetCompStr - Extract component name from compid
*/
char *GetCompStr (compid)
int compid;
{
    if ( compid )
        return(CompPtr(compid)->cname);
    else
        return(NULL);
}

/*************************************************************************
function:  GetSimpleDtmod - Get dtmod, all errors are treated as dtmod 0
*/
long GetSimpleDtmod (name)
char *name;
{
    struct stat status;
    long result;

    if ( -1 == stat(name,&status) )
        result = 0;
    else {
        UpgradeToPretendTime(name,&status);
        result = status.st_mtime;
    }
#if 0
    printf ("GetSimpleDtmod returns %d for %s\n", result, name);
#endif
    return (result);
}

/*************************************************************************
function: DirExists - Return TRUE if name exists as a directory
*/
int DirExists (char *name_in) {
    struct stat     lstat;
	char name[STRPATH+1];

    if (!name_in || !*name_in)
        return(FALSE);

	strcpy(name,name_in);
    AxSysDosUnixPath(name);		/* Can't alter 'name_in'! */
	if ( (0 == stat(name,&lstat)) && ((lstat.st_mode & S_IFDIR) == S_IFDIR) )
		return(TRUE);
	else
		return(FALSE);
}

/*************************************************************************
function:  CreateDir - Create directory if doesn't already exist
*/
CreateDir (name)
char *name;
{
 return(CreateOneDir(name,TRUE));
}

/*************************************************************************
function: CreateOneDir - Create directory, optionally checking sanity
*/
CreateOneDir (passed_name,check)
char *passed_name;
int check;
{
    char name[1000];
    char temp[256];
    int mode, ix;
    char *FindEntryName();
    char *p;

    strcpy(name,passed_name);
    if ( '.' == name[0] ) {
        TELL(("Bad directory name - %s, aborting\n",name));
        exit(EXIT_FAILURE);
    }
    if ( GetSimpleDtmod(name) )
        goto finish_up;         /* dir exists */

    mode = 0777;

    umask(0);
    VERBOSE(("Creating directory %s\n",name));
    strcpy(temp,name);
    while ( TRUE ) {
#if (SYS&SYS3) || (SYS&SYS5)
        {
        char buf[300];
        sprintf(buf,"mkdir %s >/tmp/zzx 2>/tmp/zzx2",name);
        system(buf);
        if ( GetSimpleDtmod(name) )
            goto finish_up;
        }
        errno = ENOENT;
#else
        errno = 0;
        if ( 0 == mkdir(name,0777) )
            goto finish_up;
        VERBOSE(("mkdir returned %d\n", errno));
#endif
        if ( errno == EEXIST )
            goto finish_up;
        if ( (errno == ENOENT) || (errno == ENOTDIR) ) {
            int stubbed = FALSE;
            for ( ix = strlen(temp)-1; ix > 0; ix-- ) {
#if AXMS
                if ( temp[ix] == '\\' || temp[ix] == '/' )
#else
                if ( temp[ix] == '/' ) 
#endif
                {
                    temp[ix] = 0;
                    stubbed = TRUE;
                    CreateOneDir(temp,FALSE);  /* recursive create */
                    break;
                }
            }
            if ( !stubbed ) {
                ErrorMsg("Creation of %s failed, %s",name,GenErrnoMsg(errno));
                abort();
                return(ENOENT);
            }
        }
        else {
            ErrorMsg("Creation of %s failed, %s",name,GenErrnoMsg(errno));
            return(errno);
        }
    }

finish_up:
    /* change the permissions onthe directory if we were instructed to */
    /* do so            */
    strcpy(temp,name);
    for ( ix = strlen(temp)-1; ix > 0; ix-- ) 
        {
#if AXMS
        if ( temp[ix] == '\\' || temp[ix] == '/' )
#else
        if ( temp[ix] == '/' )
#endif
            {
            p = &temp[ix+1];
            if ( IsDirOnChmodList(p, &mode ) )
               {
               chmod( name, mode );
               }
            break;
            }
        }
    return( 0 );
}

/*************************************************************************
function:  AllcStr - Malloc a string
*/
char *AllcStr (str)
register char *str;
{
    register char *p;
    p = malloc(strlen(str)+1);
    strcpy(p,str);
    return(p);
}

/*************************************************************************
function:  GetTemp - Create a tempfile name
*/
GetTemp (type,buf)
char *type;
char *buf;
{
#if AXMS
    sprintf(buf,"%s\\zzXXXXXX",TempDir);
#else
    sprintf(buf,"%s/zz%sXXXXXX",TempDir,type);
#endif
    mktemp(buf);
}

/*************************************************************************
function: FilesIdentical - Compare whether two files are identical
*/
FilesIdentical (f1,f2)
char *f1, *f2;
{
#define FI_BUF 8192
    register char *p1, *p2;
    register int ix;
    int did1, did2;
    int fd1 = -1, fd2 = -1;
    char buf1[FI_BUF],buf2[FI_BUF];

    fd1 = open(f1,0);
    if ( fd1 == -1 )
        goto nogood;
    fd2 = open(f2,0);
    if ( fd2 == -1 )
        goto nogood;
    while ( TRUE ) {
        did1 = read(fd1,buf1,FI_BUF);
        did2 = read(fd2,buf2,FI_BUF);
        if ( did1 != did2 )
            goto nogood;
        if ( did1 == 0 ){       /* EOF */
            close(fd1);
            close(fd2);
            return(TRUE);
        }
        p1 = buf1;
        p2 = buf2;
        for ( ix = 0; ix < did1; ix++ ) {
            if ( *p1++ != *p2++ )
                goto nogood;
        }
    }
nogood:
    if ( fd1 != -1 )
        close(fd1);
    if ( fd2 != -1 )
        close(fd2);
    return(FALSE);
}

/* Logging */

int OkToIndent = TRUE;  /* FALSE if trace/tell didn't have an ending newline */
char LogBuffer[64000];

#if NO_VARARGS
/*VARARGS1*/
DebugLog (a,b,c,d,e,f,g,h,i,j,k)
char *a;
int b,c,d,e,f,g,h,i,j,k;
{
    if ( !OkToIndent )
        printf("\n");
    printf("\t");
    printf(a,b,c,d,e,f,g,h,i,j,k);
    printf("\n");
    fflush(stdout);
    OkToIndent = TRUE;
}
#else
int vDebugLog (fmt, args)
  char *fmt;
  va_list args;
{
  if ( !OkToIndent )
    printf("\n");
  printf("\t");

  vprintf(fmt, args);

  printf("\n");
  fflush(stdout);
  OkToIndent = TRUE;
}

int DebugLog(fmt, va_alist)
  char *fmt;
  va_dcl
{
  va_list args;

  va_start(args);
  vDebugLog(fmt, args);
  va_end(args);
}
#endif

#if NO_VARARGS
/*VARARGS1*/
LogLog (a,b,c,d,e,f,g,h,i,j,k)
char *a;
int b,c,d,e,f,g,h,i,j,k;
{
    DebugLog(a,b,c,d,e,f,g,h,i,j,k);
}
#else
LogLog (fmt, va_alist)
char *fmt;
va_dcl
{
        va_list args;

        va_start(args);
        vDebugLog(fmt, args);
        va_end(args);
}
#endif

#if NO_VARARGS
/*VARARGS1*/
TraceLog (a,b,c,d,e,f,g,h,i,j,k)
char *a;
int b,c,d,e,f,g,h,i,j,k;
{
    LogBuffer[0] = 0;
    if ( OkToIndent )
        sprintf(LogBuffer,"\t");
    sprintf(LogBuffer+strlen(LogBuffer),a,b,c,d,e,f,g,h,i,j,k);
    OkToIndent = (LogBuffer[strlen(LogBuffer)-1] == '\n');
    printf(LogBuffer);
    fflush(stdout);
}
#else
TraceLog (fmt, va_alist)
char *fmt;
va_dcl
{
        va_list args;

        LogBuffer[0] = 0;
        if ( OkToIndent )
                sprintf(LogBuffer,"\t");

        va_start(args);
        vsprintf(LogBuffer+strlen(LogBuffer), fmt, args);
        va_end(args);

        OkToIndent = (LogBuffer[strlen(LogBuffer)-1] == '\n');
        printf(LogBuffer);
        fflush(stdout);
}
#endif

TellComponent(nodeP)
node *nodeP;
{
	if ( CommandLog && !Verbose )	/* do nothing for command-only logging */
		return;
    if ( !Trace && OkToIndent )
        printf("\t");
    printf( "%s\t", GetComponent(nodeP) );
    fflush(stdout);
    OkToIndent = FALSE;
}

#if NO_VARARGS
/*VARARGS1*/
TellLog (a,b,c,d,e,f,g,h,i,j,k)
char *a;
int b,c,d,e,f,g,h,i,j,k;
{
	if ( CommandLog && !Verbose )	/* do nothing for command-only logging */
		return;
    LogBuffer[0] = 0;
    if ( !Trace && OkToIndent )
        sprintf(LogBuffer,"\t");
    sprintf(LogBuffer+strlen(LogBuffer),a,b,c,d,e,f,g,h,i,j,k);
    OkToIndent = (LogBuffer[strlen(LogBuffer)-1] == '\n');
    printf(LogBuffer);
    fflush(stdout);
}
#else
TellLog (fmt, va_alist)
char *fmt;
va_dcl
{
        va_list args;

	if ( CommandLog && !Verbose )	/* do nothing for command-only logging */
		return;
    LogBuffer[0] = 0;
    if ( !Trace && OkToIndent )
            sprintf(LogBuffer,"\t");

    va_start(args);
    vsprintf(LogBuffer+strlen(LogBuffer), fmt, args);
    va_end(args);

    OkToIndent = (LogBuffer[strlen(LogBuffer)-1] == '\n');
    printf(LogBuffer);
    fflush(stdout);
}
#endif

#if NO_VARARGS
/*VARARGS1*/
VerboseLog (a,b,c,d,e,f,g,h,i,j,k)
char *a;
int b,c,d,e,f,g,h,i,j,k;
{
    LogBuffer[0] = 0;
    if ( !Trace && OkToIndent )
        sprintf(LogBuffer,"\t");
    sprintf(LogBuffer+strlen(LogBuffer),a,b,c,d,e,f,g,h,i,j,k);
    OkToIndent = (LogBuffer[strlen(LogBuffer)-1] == '\n');
    printf(LogBuffer);
    fflush(stdout);
}
#else
VerboseLog (fmt, va_alist)
char *fmt;
va_dcl
{
        va_list args;

        LogBuffer[0] = 0;
        if ( !Trace && OkToIndent )
                sprintf(LogBuffer,"\t");

        va_start(args);
        vsprintf(LogBuffer+strlen(LogBuffer), fmt, args);
        va_end(args);

        OkToIndent = (LogBuffer[strlen(LogBuffer)-1] == '\n');
        printf(LogBuffer);
        fflush(stdout);
}
#endif

#if (SYS&SYS3)
/*
 * function:  memcpy -
*/
char *memcpy(dst,src,cnt)
register char *dst;
register char *src;
register int cnt;
{
    while ( --cnt >= 0 )
        *dst++ = *src++;
}

char *memset(dst,val,cnt)
register char *dst;
register char val;
register int cnt;
{
    while ( --cnt >= 0 )
        *dst++ = val;
}
#endif

#if SYS&VMSSYS
/*
 * function:  GenErrnoMsg - Return an 'errno' message
*/
char *GenErrnoMsg (err)
int err;
{
    perror("");
}
#else
/*
 * function:  GenErrnoMsg - Return an 'errno' message
*/
char *GenErrnoMsg (err)
int err;
{
    static char buf[100];
#if !(LINUX || LINUX_LIBC5)
    extern char *sys_errlist[];
#endif
    extern int sys_nerr;

    if ( err < sys_nerr )
        return((char*)sys_errlist[err]);
    sprintf(buf,"err %d",err);
    return(buf);
}
#endif

#if !ALPHA
/* MySprintf is needed for the AT and other SYS3/SYS5 machines that do
    not return buffer.  Why did Berkley change this without a comment.
    The SOBS!!!
*/

#undef sprintf

#if NO_VARARGS
/*VARARGS2*/
char *MySprintf(buf,str,a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,x,y,z)
char *str, *buf;
int a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,x,y,z;
{
    sprintf(buf,str,a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,x,y,z);
    return(buf);
}
#else
char *MySprintf(buf, str, va_alist)
char *buf, *str;
va_dcl
{
        va_list args;

        va_start(args);
        vsprintf(buf, str, args);
        va_end(args);
        return(buf);
}
#endif
#endif

/*************************************************************************
function:  Pstrcmp - Do strcmp until second arg's NULL
*/
Pstrcmp (p1,p2)
register char *p1, *p2;
{
    while ( *p2 )
        if ( *p1++ != *p2++ )
            return(1);
    return(0);
}

/*************************************************************************
function: Pstreq - Perform partial strcmp, return ptr to after test
*/
char *Pstreq (bigstr,test)
register char *test, *bigstr;
{
    while ( *test ) {
        if ( *test++ != *bigstr++ )
            return(NULL);
    }
    return(bigstr);
}

/*************************************************************************
function:  TokenizeString - Convert a string to a list of tokens
returns number of tokens
*/
TokenizeString (str,tokens)
register char *str;     /* input string */
register char *tokens;  /* output string segemnts, null separated */
{
    register int count = 0;

    *tokens = 0;
    while ( TRUE ) {
        while ( *str == ' ' ) str++;
        if ( !(*str) )
            break;
        count++;
        while ( *str && (*str != ' ') )
            *tokens++ = *str++;
        *tokens++ = 0;
    }
    return(count);
}

/*************************************************************************
function: FindEntryName - Find entry name in pathname
*/
char *FindEntryName (name)
char *name;
{
    char *p;

    for ( p = name + strlen(name); p >= name; p-- )
#if AXMS
        if ( *p == '\\' || *p == '/' )
#else
        if ( *p == '/' )
#endif
            return(p+1);
    return(p+1);
}

#if NO_VARARGS
/*************************************************************************
function:  aborter - print a message and abort processing
*/
/*VARARGS1*/
aborter (a,b,c,d,e,f,g,h,i,j,k)
char *a;
int b,c,d,e,f,g,h,i,j,k;
{
    ErrorMsg(a,b,c,d,e,f,g,h,i,j,k);
    exit(EXIT_FAILURE);
}
#else
int aborter (fmt, va_alist)
  char *fmt;
  va_dcl
{
  va_list args;

  va_start(args);
  vErrorMsg(fmt, args);
  va_end(args);
  exit(EXIT_FAILURE);
}
#endif

#if NO_VARARGS
/*************************************************************************
function:  line_error - Error message for a bad input line
*/
/*VARARGS1*/
line_error (a,b,c,d,e,f,g,h,i,j,k)
char *a;
int b,c,d,e,f,g,h,i,j,k;
{
    ErrorMsg(a,b,c,d,e,f,g,h,i,j,k);
}
#else
line_error (fmt, va_alist)
  char *fmt;
  va_dcl
{
  va_list args;

  va_start(args);
  vErrorMsg(fmt, args);
  va_end(args);
}
#endif

#if NO_VARARGS
/*************************************************************************
function:  ErrorMsg - Send error msg to error stream
*/
/*VARARGS1*/
ErrorMsg (a,b,c,d,e,f,g,h,i,j,k)
char *a;
int b,c,d,e,f,g,h,i,j,k;
{
    extern int NoReportErrMsg; /* Used to briefly keep messages out of report */

    TotalErrors++;
    fprintf(ErrStream,"*** ");
    fprintf(ErrStream,a,b,c,d,e,f,g,h,i,j,k);
    if ( !BuildPhase ) {
        if ( InputLine )
            fprintf(ErrStream,": file %s line %d: %s",InputFilename,InputLine,ErrorBuf);
    }
    fprintf(ErrStream,"\n");
    fflush(ErrStream);
    if ( DoingReport && !NoReportErrMsg )
        DoReport((node*)NULL,(char*)NULL,0,a,b,c,d,e,f,g,h,i,j,k);
}
#else
static int vErrorMsg (fmt, args)
  char *fmt;
  va_list args;
{
  extern int NoReportErrMsg; /* Used to briefly keep messages out of report */
  va_list argsl;
  
#if (LINUX && __powerpc__)
	memcpy(argsl,args,sizeof(va_list));
#else
	argsl = args;
#endif

  TotalErrors++;
  fprintf(ErrStream,"*** ");

  vfprintf(ErrStream, fmt, argsl);

  if (!BuildPhase) {
    if ( InputLine )
      fprintf(ErrStream,": file %s line %d: %s",InputFilename,InputLine,ErrorBuf);
  }

  fprintf(ErrStream,"\n");
  fflush(ErrStream);
  if (DoingReport && !NoReportErrMsg)
    vDoReport((node*)NULL, (char*)NULL, 0, fmt, args);
}

int ErrorMsg(fmt, va_alist)
  char *fmt;
  va_dcl
{
  va_list args;

  va_start(args);
  vErrorMsg(fmt, args);
  va_end(args);
}
#endif

#define NDLOOK_NONE 0
#define NDLOOK_OWNER 1
#define NDLOOK_USER 2

/*************************************************************************
function: CrtOrFindNode - Find/create a node
*/
node *CrtOrFindNode (name,create,mode,user)
register char *name;
int create;
int mode;
ID user;
{
    ID id, dirid;
    register node *nodeP;
    int hashval;
    ID may_use, global_id = 0;
    ID *users;
    int usage_cnt = 0;
    dir_info *dirP;
    int ix;
    int size, nmsize;
    node NodeModel;
    char dosdir[300];

    DEBUG(("LookupNode: '%s'",name));
    hashval = am_hash(name);
    DEBUG(("LookupNode: hash %d",hashval));
    NodeCreated = FALSE;
    for ( id = HashArray[hashval]; id != 0; id = nodeP->next ) {
        nodeP = NodePtr(id);
        if ( strcmp(nodeP->name,name) == 0 ) 
            {
            NodeIdLooked = id;
            return(nodeP);
            }
    }
    if ( !create )
        return(NULL);
    DEBUG(("LookupNode: creating"));
    id = AllocObj(sizeof(node),TYPE_NODE);
    NodeIdLooked = id;
    nodeP = NodePtr(id);
    memset(nodeP,0,sizeof(node));
    nodeP->next = HashArray[hashval];
    HashArray[hashval] = id;
	if ( strlen(name) > MAX_NODENAME ) {
		aborter("node name is too long: %s", name);
	}
    strcpy(nodeP->name,name);
    if ( mode == NDLOOK_OWNER )
        nodeP->comp_id = user;
    LOGGER(("LookupNode: created node '%s'",name));
    NodeCreated = TRUE;
    return(nodeP);
}

/*************************************************************************
function:  LookupNode - Find/create a node
*/
node *LookupNode (name)
char *name;
{
    return(CrtOrFindNode(name,TRUE,NDLOOK_NONE,(ID)0));
}

/*************************************************************************
function: CheckForNode - Find node or return NULL
*/
node *CheckForNode (name)
char *name;
{
    return(CrtOrFindNode(name,FALSE,NDLOOK_NONE,(ID)0));
}

/*************************************************************************
function: LookupOwnedNode - Find/Create node based on owner only
Note that in LookupNode, relocated nodes are always FOUND by this, never
created.  Since relocation ocurrs BEFORE location, all relocations are global.
*/
node *LookupOwnedNode (name,owner)
char *name;
ID owner;
{
    return(CrtOrFindNode(name,TRUE,NDLOOK_OWNER,owner));
}

/*************************************************************************
function: LookupForMe - Find/Create a node I can use
*/
node *LookupForMe (name,owner)
char *name;
ID owner;
{
    return(CrtOrFindNode(name,TRUE,NDLOOK_USER,owner));
}
/*************************************************************************
function: LookupIdForMe - Find/Create a node I can use
*/
ID LookupIdForMe (name,owner)
char *name;
ID owner;
{
    CrtOrFindNode(name,TRUE,NDLOOK_USER,owner);
    return(NodeIdLooked);
}

static int sigCatch();

#if !AXMS && !ALPHA
/*************************************************************************
function: am_set_sigs - Set the Unix signal catchers
*/
am_set_sigs ()
{
#if 1
        /* catch appropriate signals */
        /* killers */
#if SYS&SYS5
    if ( isatty(0) )
#endif

#if !AXMS
    signal(SIGQUIT, (void (*)())sigCatch);
    signal(SIGTRAP, (void (*)())sigCatch);
    signal(SIGIOT,  (void (*)())sigCatch);
#if !(LINUX || LINUX_LIBC5)
    signal(SIGEMT,  (void (*)())sigCatch);
#endif
    signal(SIGBUS,  (void (*)())sigCatch);
#if !(LINUX || LINUX_LIBC5)
    signal(SIGSYS,  (void (*)())sigCatch);
#endif
    signal(SIGHUP,   SIG_IGN);
    signal(SIGPIPE, (void (*)())sigCatch);
    signal(SIGALRM, (void (*)())sigCatch);
#if SYS&B4_2
        /* killers */
    signal(SIGXCPU, (void (*)())sigCatch);
    signal(SIGXFSZ, (void (*)())sigCatch);
#endif
#endif  /* !AXMS */

    signal(SIGILL,  (void (*)())sigCatch);
    signal(SIGFPE,  (void (*)())sigCatch);
    signal(SIGSEGV, (void (*)())sigCatch);
        /* close down */
    signal(SIGTERM, (void (*)())sigCatch);
        /* Don't die */
#if SYS&SYS5
    if ( isatty(0) )
#endif
    signal(SIGINT,  (void (*)())sigCatch);

    return;
#endif
}
#endif /* if !AXMS && !ALPHA */

/*************************************************************************
function: SelectProg - Use path info to select an executable
*/
SelectProg (dst,src)
char *src;  /* prog name */
char *dst;  /* quailified prog name */
{
    if ( (0 == strcmp("acpp",src)) || (0 == strcmp("acpp.exe",src)) ) {
        strcpy(dst,src);
        return;
    }
    if ( NodeFilename(LookupNode(src),dst,TRUE) )
        return;
    else
        strcpy(dst,src);
}

#if NO_VARARGS
/*************************************************************************
function: DoReport - Report an error to the errors file
*/
DoReport (nodeP,cmd,code,a,b,c,d,e,f,g,h,i,j,k)
node *nodeP;
char *cmd;
int code;
int a,b,c,d,e,f,g,h,i,j,k;   /* only if nodeP == NULL */
{
    static FILE *ReportFp;
    char *GetComponent();
    static char pagebreak = 12;
    char name[256];
    char rpt_file;
    int err, out, got;
    static char *last_comp = "dummy_str_not_null";
    char *comp;
    char buf[8096];

    if ( !ReportFp ) {
        GetTemp("Errors",ReportFile);
        unlink (ReportFile);
        ReportFp = fopen(ReportFile,"w");
        TELL(("Created Error Report File %s\n",ReportFile));
    }
    if ( !nodeP ) {         /* internally generated error messages */
        if ( last_comp ) {
            fwrite(&pagebreak,1,1,ReportFp);
            fprintf(ReportFp,"Amake structural problem, see the log!\n\n");
            last_comp = NULL;
        }
        fprintf(ReportFp,a,b,c,d,e,f,g,h,i,j,k);
        fprintf(ReportFp,"\n");
        fflush(ReportFp);
        return;
    }
    comp = GetComponent(nodeP);
    if ( last_comp != comp ) {
        fwrite(&pagebreak,1,1,ReportFp);
        fprintf(ReportFp,"\nComponent:     %s\n\n",comp);
        last_comp = comp;
    }
    else
        fprintf(ReportFp,"\n\n\n");

    NodeFilename(nodeP,name,TRUE);
    fprintf(ReportFp,"File To Build: %s\n",name);
    fprintf(ReportFp,"Command Line:  %s\n",cmd);
    fprintf(ReportFp,"Return Error:  %d\n\n",code);
    fflush(ReportFp);
    err = open(StdErrFile,0);
    if ( err != -1 ) {
        while ( got = read(err,buf,8096) )
            write(fileno(ReportFp),buf,got);
        close(err);
    }
    out = open(StdOutFile,0);
    if ( out != -1 ) {
        while ( got = read(out,buf,8096) )
            write(fileno(ReportFp),buf,got);
        close(out);
    }
    fflush(ReportFp);
}
#else
int vDoReport (nodeP, cmd, code, a, args)
  node *nodeP;
  char *cmd;
  int code;
  char *a;
  va_list args;
{
  static FILE *ReportFp;
  char *GetComponent();
  static char pagebreak = 12;
  char name[256];
  char rpt_file;
  int err, out, got;
  static char *last_comp = "dummy_str_not_null";
  char *comp;
  char buf[8096];

  if ( !ReportFp ) {
    GetTemp("Errors",ReportFile);
    unlink (ReportFile);
    ReportFp = fopen(ReportFile,"w");
    TELL(("Created Error Report File %s\n",ReportFile));
  }
  if ( !nodeP ) {         /* internally generated error messages */
    if ( last_comp ) {
      fwrite(&pagebreak,1,1,ReportFp);
      fprintf(ReportFp,"Amake structural problem, see the log!\n\n");
      last_comp = NULL;
    }
    vfprintf(ReportFp, a, args);
    fprintf(ReportFp,"\n");
    fflush(ReportFp);
    return;
  }
  comp = GetComponent(nodeP);
  if ( last_comp != comp ) {
    fwrite(&pagebreak,1,1,ReportFp);
    fprintf(ReportFp,"\nComponent:     %s\n\n",comp);
    last_comp = comp;
  } else fprintf(ReportFp,"\n\n\n");

  NodeFilename(nodeP,name,TRUE);
  fprintf(ReportFp,"File To Build: %s\n",name);
  fprintf(ReportFp,"Command Line:  %s\n",cmd);
  fprintf(ReportFp,"Return Error:  %d\n\n",code);
  fflush(ReportFp);
  err = open(StdErrFile,0);
  if ( err != -1 ) {
    while ( got = read(err,buf,8096) )
      write(fileno(ReportFp),buf,got);
    close(err);
  }
  out = open(StdOutFile,0);
  if ( out != -1 ) {
    while ( got = read(out,buf,8096) )
      write(fileno(ReportFp),buf,got);
    close(out);
  }
  fflush(ReportFp);
}

int DoReport (nodeP, cmd, code, a, va_alist)
  node *nodeP;
  char *cmd;
  int code;
  char *a;
  va_dcl
{
  va_list args;

  va_start(args);
  vDoReport(nodeP, cmd, code, a, args);
  va_end(args);
}
#endif

/*************************************************************************
function: FilterToTokArray - Create Token ptr array, return token count
Assumes passed string has exactly one space as token separator and
do not begin/end with spaces.
On most machines (Unix), the filter has no effect and the passed string
is itself munged into nulls.  If filtering occurs, the null munged string
is a separately stored character array (FilteredTokens).
*/
FilterToTokArray (p,toks)
register char *p;
register char **toks;
{
    register int count = 1;

    if ( *p == 0 ) {
        toks[0] = NULL;
        return(0);
    }
    toks[0] = p;
    while ( *p ) {
        if ( *p == ' ' ) {
            *p++ = 0;
            toks[count++] = p;
        }
        p++;
    }
    toks[count] = NULL;
    return(count);
}

/*************************************************************************
function:  LocateRule - Find the rule to use given rule name
*/
dispatch_rule *LocateRule (name)
register char *name;
{
    register dispatch_rule *ruleP;
    extern dispatch_rule RuleArray[];

    DEBUG(("LocateRule: '%s'",name));
    for ( ruleP = RuleArray; ruleP->name[0]; ruleP++ ) {
        DEBUG(("LocateRule: '%s'",ruleP->name));
        if ( 0 == strcmp(name,ruleP->name) )
            return(ruleP);
    }
    return(NULL);
}

/*************************************************************************
function: ReFindDeps - Force a finding of dependencies
*/
ReFindDeps (nodeP)
node *nodeP;
{
    nodeP->deps_done = FALSE;
    FindMyDeps(nodeP);
}

/*************************************************************************
function: ReSuffix - Generate a filename with changed/added suffix
*/
ReSuffix (new,old,sfx)
register char *old, *new, *sfx;
{
    register char *p;

    strcpy(new,old);
    for ( p = new + strlen(new) - 1; p != new; p-- ) {
        if ( *p == '.' ) {
            strcpy(p,sfx);
            return;
        }
    }
    strcat(new,sfx);
}

/*************************************************************************
function:  GetIncludes - Create an includes file includer string for cc
*/
GetIncludes (nodeP,buf)
register node *nodeP;
register char *buf;
{
    register int ix, best, bestix;
    register int copied = 0;
    register int num_indirs;
    dir_info *dirs[200];
    dir_info *unsorted[200];
    ID dir_ids[200];

    num_indirs = GetIncDirs(AddrOfDepVals(nodeP),NodeDepCount(nodeP),dir_ids,0);
    for ( ix = 0; ix < num_indirs; ix++ )
        unsorted[ix] = DirPtr(dir_ids[ix]);
    while ( copied < num_indirs ) {
        best = 0;
        for ( ix = 0; ix < num_indirs; ix++ ) {
            if ( unsorted[ix] && (unsorted[ix]->priority >= best) ) {
                best = unsorted[ix]->priority;
                bestix = ix;
            }
        }
        dirs[copied++] = unsorted[bestix];
        unsorted[bestix] = NULL;
    }
    *buf = 0;
    for ( ix = 0; ix < num_indirs; ix++ ) {
        sprintf(buf," -I%s",dirs[ix]->dname);
        buf += strlen(buf);
    }
	sprintf(buf," -I%s",DefaultDirName(nodeP));
}

/*************************************************************************
function:  GetIncDirs - Get unique directories from array of node ids
*/
GetIncDirs(arrP,total,dirs,count)
register ID *arrP;
int total;
register ID dirs[];
register int count;
{
    register int ix, jx;
    register ID id;
    register node *nodeP;

    for ( ix = 0; ix < total; ix++ ) {
        nodeP = NodePtr(*arrP++);
        if ( nodeP->ruleid == DUMMY_RULE )
            count = GetIncDirs(AddrOfDepVals(nodeP),
                                 NodeDepCount(nodeP),dirs,count);
        if ( 0 == (id = NodeDirId(nodeP)) )
            continue;
        for ( jx = 0; (jx < count) && (dirs[jx] != id); jx++ );
        if ( jx == count )
            dirs[count++] = id;
    }
    return(count);
}

/*************************************************************************
function: DeleteAll - Delete all output files (modules with a rule)
*/
DeleteAll()
{
    register node *nodeP;
    register ID id;
    char name[STRPATH+1];

    if ( TotalErrors != 0 ) {
        TELL(("Not risking a -clean due to prior errors\n"));
        return;
    }
    for ( id = NextType(0,TYPE_NODE); id ;id = NextType(id,TYPE_NODE) ) {
        nodeP = NodePtr(id);
        if ( OUTPUT_NODE(nodeP) ) {
            NodeFilename(nodeP,name,TRUE);
            if ( unlink(name) )
                VERBOSE(("Deleted %s\n",name));
        }
    }
}

#if !AXMS && !ALPHA
/*************************************************************************
function: sigCatch - The amake cleanup on abort code
*/
static int sigCatch (signum)
int signum;
{
    int i;
    int ix;
    extern char *FilesToDelete[];
    extern int NumToDelete;     /* delete these guys on succesfull build */
    extern char StartDir[];
#if !(LINUX || LINUX_LIBC5)
    extern char *sys_siglist[];
#else
	extern __const char * __const sys_siglist[];
#endif
    extern node *CurrentOutputNode;

#if (SYS&SYS3) || (SYS&SYS5) || PORT_SUNX
#else
    if ( signum < NSIG )
        ErrorMsg("%s, aborting",sys_siglist[signum]);
    else
#endif
        ErrorMsg("caught signal %d, aborting",signum);
    for ( i = 1; i < NSIG; i++ )
        signal(i,SIG_DFL);
    SetDir(StartDir);
    if ( CurrentOutputNode )
        if ( DeleteNodeFile(CurrentOutputNode) )
            ErrorMsg("%s deleted",CurrentOutputNode->name);
#if !AXMS
    if ( (signum != SIGINT) && (signum != SIGTERM) )
        kill(getpid(),SIGIOT);
#endif
    exit(EXIT_FAILURE);
}
#endif /* if !AXMS && !ALPHA */

/*************************************************************************
function: NodeName - Return string of node's name
*/
char *NodeName (nodeP)
node *nodeP;
{
    return(nodeP->name);
}

/*************************************************************************
function: AddDepend - Add dependency ID to a node
*/
AddDepend (id,dep)
ID id;          /* node to add to */
ID dep;      /* value to Add */
{
    node *nodeP;
    ID *depsP;
    int ix;
    int depcount;

    if ( !dep ) return;
    depsP = AddrOfDepVals(nodeP = NodePtr(id));
    depcount = NodeDepCount(nodeP);
    for ( ix = 0; ix < depcount; ix++ ) {
        if ( *depsP++ == dep )
            return;
    }
    ResizeObj(id,SizeObj(id)+sizeof(ID));
    depsP = AddrOfDepVals(nodeP = NodePtr(id));
	assert(nodeP->callcount==0);
    nodeP->inputs[nodeP->incount+nodeP->depcount] = dep;
    nodeP->depcount++;
/*  VERBOSE(("AddDep %s\n",NodePtr(dep)->name)); */
}


/*************************************************************************
function: NodeCvtToRuled - Convert an input to an output node
*/
node *NodeCvtToRuled (id,rule,compid)
ID id;
int rule;
ID compid;
{
    register node *nodeP;
    int was_global;

    nodeP = NodePtr(id);
    was_global = NodeVisGlobal(nodeP);
    if ( nodeP->ruleid ) return(nodeP);
    bld_assert(strlen(nodeP->name) < MAX_NODENAME,
            ("name too long - %s",nodeP->name));
    ResizeObj(id,sizeof(node)-sizeof(ID));
    nodeP = NodePtr(id);
    nodeP->ruleid = rule;
    nodeP = NodePtr(id);
    nodeP->withstr = 0;
    nodeP->incount = 0;
    nodeP->depcount = 0;
    nodeP->callcount = 0;
    if ( !was_global )
        nodeP->comp_id = compid;
    return(nodeP);
}

/*************************************************************************
function: NodeDepCount - Return how many depends in an outputs node
*/
NodeDepCount (nodeP)
register node *nodeP;
{
    if ( !nodeP->ruleid ) return(0);
    return(nodeP->depcount);
}

/*************************************************************************
function: NodeInCount - Return how many inputs to an outputs node
*/
NodeInCount (nodeP)
register node *nodeP;
{
    if ( !nodeP->ruleid ) return(0);
    return(nodeP->incount);
}

/*************************************************************************
function: NodeInVal - Return a node's input value from index
*/
node *NodeInVal (nodeP,ix)
register node *nodeP;
register int ix;
{
    assert(nodeP->ruleid);
    return(NodePtr(nodeP->inputs[ix]));
}

/*************************************************************************
function: NodeInId - Return a node's input value from index
*/
ID NodeInId (nodeP,ix)
register node *nodeP;
register int ix;
{
    assert(nodeP->ruleid);
    return(nodeP->inputs[ix]);
}

/*************************************************************************
function: NodeDepVal - Return a node's dependecy value from index
*/
node *NodeDepVal (nodeP,ix)
register node *nodeP;
register int ix;
{
    assert(nodeP->ruleid);
    return(NodePtr(nodeP->inputs[nodeP->incount+ix]));
}

/*************************************************************************
function: NodeDepId - Return a node's dependecy value from index
*/
ID NodeDepId (nodeP,ix)
register node *nodeP;
register int ix;
{
    assert(nodeP->ruleid);
    return(nodeP->inputs[nodeP->incount+ix]);
}

/*************************************************************************
function: NodeCallVal - Return a node's call value from index
*/
ID NodeCallVal (nodeP,ix)
register node *nodeP;
register int ix;
{
    assert(nodeP->ruleid);
    return(nodeP->inputs[nodeP->incount+nodeP->depcount+ix]);
}

/*************************************************************************
function: AddrOfDepVals - Return addr of deps array
*/
ID *AddrOfDepVals (nodeP)
node *nodeP;
{
    assert(nodeP->ruleid);
    return(nodeP->inputs+nodeP->incount);
}

/*************************************************************************
function: AddrOfCallVals - Return addr of calls array
*/
ID *AddrOfCallVals (nodeP)
node *nodeP;
{
    assert(nodeP->ruleid);
    return(nodeP->inputs+nodeP->incount+nodeP->depcount);
}

/*************************************************************************
function: AddrOfInVals - Return addr of inputs/deps array
*/
ID *AddrOfInVals (nodeP)
node *nodeP;
{
    assert(nodeP->ruleid);
    return(nodeP->inputs);
}


/* The following will go through the "with" str, looking for particular */
/* tokens, and will possibly alter the with str.  This is most useful   */
/* when one needs to expans a simple name to a real path                */
char *MangleWithStr( char *with_str )
{
#if !AXMS                   /*  {  */
    return( with_str );
#else                       /*  }  {  */
    static char *with_buff;
    static int size_with_buff;
    char tmp_buf[2000];                  /* real big */
    char *p, *p2, *p3, *p4;
    int len;
    char pathname[1026];
    char *filename;
    node *nodeP;

    if ( size_with_buff == 0 )
        {
        size_with_buff = 300;    /* start out size */
        with_buff = malloc( size_with_buff );
        }

    p = with_str;
    p2 = tmp_buf;
    while ( *p != '\0' )
        {
        if ( strncmp( p, "-Fp", 3 ) == 0 )
            {
            /* this is a pre-compiled header file name, get the name */
            /* and subs=titute a full path                            */
            p = p + 3;                         /* skip "-Fp" */
            filename = p;
            while ( *p > ' ' )
               { p++; }     /* find the end of the name */
            if ( p != '\0' )
               {
               *p = '\0';
               }
            p--; /* so we can increment below and hit next char */

            nodeP = CheckForNode( filename );
            if ( nodeP == (node *)NULL )
               printf("Could not find \"-Fp\" argument <%s>\n", filename );
            else
               {
               NodeFilename( nodeP, pathname, TRUE );
               sprintf( p2, "-Fp%s ", pathname );
               p2 = p2 + strlen(p2);
               }
            }
        else
            {
            *p2 = *p;
            p2++;
            }
        
        p++;
        }
    *p2 = '\0';
    len = strlen( tmp_buf );
    if ( len >= size_with_buff )
        {
        size_with_buff = len;
        with_buff = realloc( with_buff, size_with_buff );
        size_with_buff = 300;    /* start out size */
        }
    strcpy( with_buff, tmp_buf );
    return( with_buff );   

#endif                     /* } AXMS */
}


/*************************************************************************
function: NodeWithStr - Return a node's 'with' string
*/
char *NodeWithStr (nodeP)
register node *nodeP;
{
    register ID id;
    char *with_str;

    assert(nodeP->ruleid);
    if ( id = nodeP->withstr )
        {
        with_str = StrPtr(id);
        with_str = MangleWithStr( with_str );
        return(with_str);
        }
    else
        return("");
}

/*************************************************************************
function: NodeVisGlobal - return T/F whether node is globally visible
*/
NodeVisGlobal (nodeP)
register node *nodeP;
{
    if ( (NodeCompId(nodeP)==0) || nodeP->is_global )
        return(TRUE);
    if ( nodeP->dirid )
        return(DirPtr(nodeP->dirid)->owner == 0);
}

/*************************************************************************
function: NodeFrcGlobal - return T/F whether node is forced global
*/
NodeFrcGlobal (nodeP)
node *nodeP;
{
    return(nodeP->is_global);
}

/*************************************************************************
function: FindDefine - Find a var, abort if not set
*/
char *FindDefine (name)
char *name;
{
    char *p, *GetVar();

#if 0
    char buff[80];
    
    if ( !RunningNT3_1 )
       {
       /* first try for the newer rev stuff */
       strcpy( buff, name );
       strcat( buff, "3_5" );
       p = GetVar(buff);
       if ( p != (char *)NULL )
          /* found it */
          return( p );
       }
#endif

    p = GetVar(name);
    bld_assert(p != NULL,("%s not set",name));
    return(p);
}

/*************************************************************************
function: GrepFiles - Grep thru files, print result at stdout
*/
GrepFiles (key,count,paths)
char *key;  /* what to look for */
int count;  /* how many paths there are */
char *paths[];  /* pathname fragments */
{
    register char *p, *p2, *lp;
    register node *nodeP;
    register int ix, id;
    char pathname[STRPATH+1];
    char line[1024];
    FILE *fp;
    char *fgets(), *strchr();

    if ( count < 1 ) return;
    for ( id = NextType(0,TYPE_NODE); id != 0; id = NextType(id,TYPE_NODE) ) {
        nodeP = NodePtr(id);
        if ( nodeP->ruleid && (nodeP->ruleid != DUMMY_RULE) )
            continue;
        if ( (NULL == (p = strchr(nodeP->name,'.'))) ||
             streq(".o",p) || streq(".out",p) || streq(".a",p) )
            continue;
        NodeFilename(nodeP,pathname,TRUE);
        if ( streq(paths[0],"*") ) goto doit;
        for ( ix = 0; ix < count; ix++ ) {
            for ( lp = pathname+strlen(pathname)-strlen(paths[ix]);
                     (p=lp) >= pathname; lp-- ){
                for ( p2 = paths[ix]; (*p2 == *p) && *p2; p2++, p++ );
                if ( *p2 == '\0' )
                    goto doit;
            }
        }
        continue;
doit:
        if ( NULL == (fp = fopen(pathname,"r")) )
            continue;
        while ( fgets(line,1022,fp) ) {
            for ( lp = line+strlen(line)-1-strlen(key);(p = lp) >= line; lp-- ){
                for ( p2 = key; *p2 == *p; p2++, p++ );
                if ( *p2 == '\0' ) {
                    printf("%s: %s",nodeP->name,line);
                    goto nexter;
                }
            }
nexter:;
        }
        fclose(fp);
nextfile:;
    }
}


/*************************************************************************
function: FindFiles - Print at stdout the dirs of files
*/
FindFiles (count,names)
int count;  /* how many paths there are */
char *names[];
{
    int ix;
    char filename[STRPATH+1];

    for ( ix = 0; ix < count; ix++ ) {
        NodeFilename(LookupNode(names[ix]),filename,TRUE);
        printf("%s\n",filename);
    }
}

/*************************************************************************
function: StripTypes - remove "non-standard" types and static funcs
                       function proto lines
*/
StripTypes(inFileName, outFileName)
char *inFileName, *outFileName;
{
    FILE *inFP, *outFP;
    char buff1[512], buff2[128], *p;
    char *NextWord();
    int IsOneOfTheITypes(), IsOneOfTheUTypes();

    if ((inFP = fopen(inFileName,"r")) == NULL)
        return;
    if ((outFP = fopen(outFileName,"a")) == NULL)
        return;

    /* while get next line isn't end_of_file */
    while (fgets(buff1,512,inFP) != NULL)
    {
        /* if the first thing is a comment
            print and continue
        */
        if ((buff1[0] == '/') && (buff1[1] == '*'))
        {
                fputs(buff1,outFP);
                continue;
        }

        /* point to the first word */
        p = buff1;

        /* if the word is "static" skip it */
        if (!strncmp(buff1,"static", 6))
            continue;
            

        /* if the word is "unsigned" or "signed"
            point to the next word
        */
        if (!strncmp(buff1,"unsigned", 8))
            p = NextWord(p);
        else if (!strncmp(buff1,"signed", 6))
            p = NextWord(p);

        /* terminate the pointed_to_word - this allows us
           to distinguish between, for example, type
           int and type intercept.
        */
        TerminateWord(p,buff2);

        /* if the pointed_to_word is one of the basic types, keep it */
        switch (p[0])
        {
            case 'i':
                if (IsOneOfTheITypes(buff2))
                    fputs(buff1,outFP);
                continue;
            case 'u':
                if (IsOneOfTheUTypes(buff2))
                    fputs(buff1,outFP);
                continue;
            case 'v':
                TestAndKeep(outFP,buff1,buff2,"void");
                continue;
            case 'c':
                TestAndKeep(outFP,buff1,buff2,"char");
                continue;
            case 's':
                TestAndKeep(outFP,buff1,buff2,"short");
                continue;
            case 'l':
                TestAndKeep(outFP,buff1,buff2,"long");
                continue;
            case 'f':
                TestAndKeep(outFP,buff1,buff2,"float");
                continue;
            case 'd':
                TestAndKeep(outFP,buff1,buff2,"double");
                continue;
            case 'w':
                TestAndKeep(outFP,buff1,buff2,"wchar");
                continue;
            case 'm':
                TestAndKeep(outFP,buff1,buff2,"mbchar");
                continue;
            case 'g':
                TestAndKeep(outFP,buff1,buff2,"generic_ptr");
                continue;
            default:
                continue;
        }
    }
    fclose(inFP);
    fclose(outFP);
}


char *NextWord(cur)
char *cur;
{
    char *p;

    p = cur;
    while (*p != ' ')
        p++;
    while (*p == ' ')
        p++;
    return(p);
}

IsOneOfTheITypes(p)
char *p;
{

    if (!strcmp(p,"int32"))
        return(TRUE);
    else if (!strcmp(p,"int16"))
        return(TRUE);
    else if (!strcmp(p,"int8"))
        return(TRUE);
    else if (!strcmp(p,"int"))
        return(TRUE);
    else
        return(FALSE);
}

IsOneOfTheUTypes(p)
char *p;
{

    if (!strcmp(p,"uint32"))
        return(TRUE);
    else if (!strcmp(p,"uint16"))
        return(TRUE);
    else if (!strcmp(p,"uint8"))
        return(TRUE);
    else
        return(FALSE);
}

TerminateWord(inputBuff, outputBuff)
char *inputBuff, *outputBuff;
{
    char *p;
    int len;

    p = inputBuff;
    while (*p != ' ')
        p++;

    len = p-inputBuff;
    strncpy(outputBuff,inputBuff,len);
    outputBuff[len] = 0;
}

TestAndKeep(outFP,lineBuf,testBuf,testString)
FILE *outFP;
char *lineBuf,*testBuf,*testString;
{
    if (!strcmp(testBuf,testString))
        fputs(lineBuf,outFP);
}



typedef struct chmod_rec_ {
    char *file;
    int mode;
    }  chmod_rec;

static chmod_rec *ChmodList = (chmod_rec *)NULL;
static int ChmodListSize = 0;

void AddDirToChmodList( filename, mode_str )
char *filename;
char *mode_str;
{
    char *p;
    int mode;

    /* first compute the mode.  It is passed in as an octal string, ie */
    /*   755, 444, etc.                                                */
    p = mode_str;
    while ( *p ==' ' )
       { p++; }
    
    /* now the three digits */
    mode = *p - '0';
    mode = mode << 3;
    p++;
    mode = mode + (*p - '0');
    mode = mode << 3;
    p++;    
    mode = mode + (*p - '0');

    
    if ( ChmodList == (chmod_rec *)NULL )
        ChmodList = (chmod_rec *)malloc( sizeof(chmod_rec) );
    else
        ChmodList = (chmod_rec *)realloc( ChmodList,
                                   (ChmodListSize+1) * sizeof(chmod_rec) );

    ChmodList[ChmodListSize].file = AllcStr( filename );
    ChmodList[ChmodListSize].mode = mode;
    
    ChmodListSize++;
    return;
}


int IsDirOnChmodList( file, modeP )
char *file;
int *modeP;
{
    chmod_rec *chmodP;
    int i;

    chmodP = ChmodList;
    for ( i=0; i<ChmodListSize; i++ )
       {
       if ( strcmp(chmodP->file, file ) == 0 )
          {
          *modeP = chmodP->mode;
          return( TRUE );
          }
       chmodP++;
       }
    return( FALSE );
}





#define DOS_SLASH '\\'
#define UNIX_SLASH '/'
#if AXMS
#define SLASH_CHAR DOS_SLASH
#else
#define SLASH_CHAR UNIX_SLASH
#endif

char *strchr(), *strrchr();

/*************************************************************************
function: AxSysDosUnixPath - Convert / to \ to keep varois DOS commands happy
*/
void AxSysDosUnixPath (name)
char *name;
{
#if !AXMS
    return;
#else
    char *p1, *p2;
    p1 = p2 = name;
    while ( *p2 ) 
        {
        if ( *p2 == UNIX_SLASH )
            {
            *p1++ = DOS_SLASH;
            p2++;
            }
        else
            /* a regular character */
            *p1++ = *p2++;
        }
     *p1 = '\0';
#endif
}

/* SrcRelativeToDst() - returns the relative path from the dst to the src
for eg. if dst = /a/b/c/foo and src = /a/b/c/d/bar then ret = ./d/bar
		   dst = /a/b/c/d/foo and src = /a/b/c/bar then ret = ./../bar
THE RETURN VALUE SHOULD BE FREED BY THE CALLER. 
*/
char *
SrcRelativeToDst(dst,src)
char *dst, *src;
{
	char *s, *d, *r;
	char *s_slash, *d_slash;
	char ret[1024];

	if (!dst || !src)
		return((char *)NULL);

	d = dst;
	s = src;
	d_slash = d;
	s_slash = s;

	while (*s == *d) {
		if (*s == SLASH_CHAR) {
			s_slash = s+1;
			d_slash = d+1;
		}
		*s++;
		*d++;
	}
	s = s_slash ;
	d = d_slash ;

	/* add ./ at the begining of the path */
	ret[0] = '.';
	ret[1] = SLASH_CHAR;
	r = &ret[2];
	for (;*d!='\0';d++) {
		if (*d == SLASH_CHAR) {
			/* add ../ for every extra directory in dst */
			*r = '.' ; r++;
			*r = '.' ; r++;
			*r = SLASH_CHAR; r++;	
		}
	}
	*r = '\0';
	r = ret;	

	/* add the remaining portion of the source */
	strcat(r,s);

	/* to be freed by the caller */
	r = malloc(strlen(ret)+1);
	strcpy(r,ret);
	return(r);
}

/*************************************************************************
function: AmListFiles - Return allocated array of entrynames in directory
*/
char **AmListFiles (char *dirname) {
	int numEntries = 0;
	char *entries[10000];
	int code;
	char dir[STRPATH+1];
	char entryname[STRPATH+1];
	char **retArray;

	while ( TRUE ) {
		code = AFindFiles(dirname,0,dir,entryname);
		dirname = NULL;
		if ( code == 1 )
			break;
		else if ( code == 2 ) {
			TELL(("could not stat %s\n",entryname));
			continue;
		}
		bld_assert(numEntries<10000,("Too many names in AmListFiles"));
		entries[numEntries++] = AllcStr(entryname);
	}
	retArray = (char**)malloc((numEntries+1)*sizeof(char*));
	memcpy(retArray,entries,numEntries*sizeof(char*));
	retArray[numEntries] = NULL;
	return(retArray);
}
/*************************************************************************
function: AmListFilesFreer - Free malloc'd result f AmListFiles
*/
void AmListFilesFreer (char **retArray) {
	char **pp;

	for ( pp = retArray; *pp; pp++ )
		free(*pp);
	free(retArray);
}

#if MALLOC_TEST
#undef free
#undef malloc
#undef calloc
#undef realloc

char *malloc(), *realloc();

typedef struct malc_hdr {
	int magic1;
	struct malc_hdr *next;
	struct malc_hdr *prev;
	unsigned int size;
	int magic2;
	int magic3;
} malc_hdr;

malc_hdr *MalcList;
int AmMagic = 0xFEDCBA9;

char *MyMalloc(unsigned passedIn)
{
	unsigned n = passedIn;
	malc_hdr *p;

    n += sizeof(malc_hdr) + sizeof(AmMagic); /* make room for security checks */
    p = (malc_hdr*)malloc(n);
	p->magic1 = AmMagic;
	p->magic2 = AmMagic;
	p->magic3 = AmMagic;
	if ( MalcList )
		MalcList->prev = p;
	p->prev = NULL;
	p->next = MalcList;
	p->size = passedIn;
	MalcList = p;
	memcpy( ((char*)(p+1))+passedIn, &AmMagic, sizeof(int));
    return((char*)(p+1));
}

char *MyCalloc(unsigned n,unsigned s) {
	char *p;
	n *= s;
	p = MyMalloc(n);
	memset(p,0,n);
	return(p);
}

MyFree(char *ps)
{
	int ix;
	malc_hdr *p = (malc_hdr*)ps;

	p--;
	assert(p->magic1 == AmMagic);
	assert(p->magic2 == AmMagic);
	assert(p->magic3 == AmMagic);
	memcpy( &ix,((char*)(p+1))+p->size, sizeof(int));
	assert(ix == AmMagic);
	if ( MalcList == p )
		MalcList = p->next;
	if ( p->prev )
		p->prev->next = p->next;
	if ( p->next )
		p->next->prev = p->prev;
	p->magic1 = -1;
	p->magic2 = -1;
	p->magic3 = -1;
	memcpy( ((char*)(p+1))+p->size, &p->magic1, sizeof(int));

    free((char*)p);
}

char *MyRealloc(char *ps,unsigned passedIn)
{
	unsigned n = passedIn;
	malc_hdr *p = (malc_hdr*)ps, *oldP, *oldNext, *oldPrev;
	int ix;

	p--;
	assert(p->magic1 == AmMagic);
	assert(p->magic2 == AmMagic);
	assert(p->magic3 == AmMagic);
	memcpy( &ix,((char*)(p+1))+p->size, sizeof(int));
	assert(ix == AmMagic);

    n += sizeof(malc_hdr) + sizeof(AmMagic); /* make room for security checks */
	p->size = passedIn;
	oldP = p;
	oldPrev = p->prev;
	oldNext = p->next;
	p = (malc_hdr*)realloc(p,n);
	memcpy( ((char*)(p+1))+passedIn, &AmMagic, sizeof(int));
	if ( p == oldP )
		return((char*)(p+1));
	if ( oldPrev )
		oldPrev->next = p;
	if ( oldNext )
		oldNext->prev = p;
	if ( MalcList == oldP )
		MalcList = p;
	return((char*)(p+1));
}

CheckMemory()
{
	malc_hdr *p;
    int ix;

	for ( p = MalcList; p; p = p->next ) {
		assert(p->magic1 == AmMagic);
		assert(p->magic2 == AmMagic);
		assert(p->magic3 == AmMagic);
		memcpy( &ix,((char*)(p+1))+p->size, sizeof(int));
		assert(ix == AmMagic);
	}
}
#endif
