/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: cpr_yyerror.h
   Applix ASC version: 5
   Date: Mon May 17 21:13:26 1999
*/
/* $Id: yyerror.c,v 4.2 1995/01/01 19:34:59 cthuang Exp $
 *
 * This file is included into cpr_grammar.y to provide the 'yyerror()' function. 
 * If the yacc/bison parser is one that we know how to backtrack, we'll augment
 * the "syntax error" message with information that shows what type of token we
 * expected.
 */

/* 'yyerror()' has to be a macro, since it must be expanded inline to subvert
 * the internal state of 'yyparse()'.
 */

#if YACC_HAS_YYTOKS_2
#undef  YACC_HAS_YYTOKS
#define YACC_HAS_YYTOKS 1
#define YaccState yy_state
#endif

#if YACC_HAS_YYNAME	/* Linux's yacc */
#if YYDEBUG
#define	yyerror(text) {\
    register int n, x, c1, count = 0;\
    yaccError(text);\
    if (((n = yysindex[yystate]) != 0) && (n < YYTABLESIZE)) {\
	for (x = ((n > 0) ? n : 0); x < YYTABLESIZE; ++x) {\
	    c1 = x - n;\
	    if ((yycheck[x] == c1) && (c1 != YYERRCODE)) {\
		yaccExpected(yyname[c1] ? yyname[c1] : "illegal symbol",\
		    count++);\
	    }\
	}\
    }\
    yaccExpected("", -1);\
} 
#endif
#endif	/* YACC_HAS_YYNAME */


/*
 * Any way we define it, 'yyerror()' is a real function (that we provide,
 * rather than use the one from a library).
 */
static void yaccError    ARGS((char *));

#ifdef yyerror
static int  compar       ARGS((CONST void *a, CONST void *b));
static void yaccExpected ARGS((CONST char *, int count));

static
int compar(p1, p2)
CONST void *p1;
CONST void *p2;
{
    return (strcmp(*(char **)p1, *(char **)p2));
}
extern int ExpectedOK;

static
void yaccExpected (s, count)
CONST char *s;
int count;
{
    static struct {
	char *actual, *name;
    } tbl[] = {
	{"...",	"T_ELLIPSIS"},
	{"[]",	"T_BRACKETS"},
	{"{",	"T_LBRACE"},
	{"}",	"T_MATCHRBRACE"},
    };
    register int j, k, x;
    char *t = (char *)s;
    char *tag;
    char tmp[80];
    static int last;
    static char	*vec[10];	/* patch: don't know how big */

#if MSDOS
    if ( !ExpectedOK )
        /* error output has been turned off */
        return;
#endif

    if (count < 0) {
	if (last++ >= 0) {
	    qsort((char *)vec, (size_t)last, sizeof(vec[0]), compar);
	    /* limit length of error message */
if ( last > 10 || !vec[last-1] )
  vec[last-1] = "???????";
	    k = 80 - (strlen(vec[last-1]) + 2);
	    for (j = 0; j < last; j++) {
		tag = j ? " " : "Expected: ";
		s = vec[j];
		if (j != (last - 1)) {
		    x = strlen(s) + strlen(tag);
		    if (k <= 0)
			continue;
		    else if ((k - x) <= 0)
			s = "...";
		    k -= x;
		}
		fprintf(stderr, "%s%s", tag, s);
	    }
	    fprintf(stderr, "\n");
	}
    } else {
	if (!strncmp(t, "T_", 2)) {
	    int	found = FALSE;
	    for (j = 0; j < sizeof(tbl)/sizeof(tbl[0]); j++) {
		if (!strcmp(t, tbl[j].name)) {
		    t = tbl[j].actual;
		    found = TRUE;
		    break;
		}
	    }
	    if (!found) {
		t = strncpy(tmp, t + 2, sizeof(tmp)-1);
		for (k = 0; t[k] != '\0'; k++) {
		    if (t[k] == '_')
			t[k] = '-';
		    else if (isalpha(t[k]) && isupper(t[k]))
			t[k] = tolower(t[k]);
		}
	    }
	}
	if (count >= (sizeof(vec)/sizeof(vec[0]))-1)
	    count = (sizeof(vec)/sizeof(vec[0]))-1;
	vec[count] = xstrdup(t);
    }
    last = count;
}
#else
#define yyerror(s) yaccError(s)
#endif	/* yyerror */
