/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: cproto.c
   Applix ASC version: 9.1
   Date: Sun Jan 24 21:13:19 1999
*/
/* $Id: cproto.c,v 4.4 1995/08/24 01:03:48 cthuang Exp $
 *
 * C function prototype generator and function definition converter
 */
static char rcsid[] = "$Id: cproto.c,v 4.4 1995/08/24 01:03:48 cthuang Exp $";

#include <stdio.h>
#include <ctype.h>
#include "cproto.h"

#define PATCHLEVEL 4    /* originally ina file called "patchlev.h" */

/* getopt declarations */
#if HAVE_GETOPT_H
#include <getopt.h>
#else
#if MSDOS
/* we will be defining our own getopt at the bottom of this file.
 * taken from /ax/pgms/utils/utgtopt.c              -donato
 */
int getopt ARGS((int argc, char * *argv,  char *shortopts));
int     opterr = 1,      /* useless, never set or used */
        optind = 1,      /* index into parent argv vector */
        optopt=0;        /* character checked for validity */
char    *optarg=NULL;    /* argument associated with option */
#else
extern int getopt ARGS((int argc, char * *argv,  char *shortopts));
extern char *optarg;
extern int optind;
#endif
#endif

/* Name of the program (from argv[0]) */
char *progname;

/* Program options */

/* If nonzero, output variables declared "extern" in include-files */
int extern_in = 0;

/* When TRUE, suppress return-statements in function-bodies */
int exitlike_func = FALSE;

/* If TRUE, output "extern" before global declarations */
boolean extern_out = FALSE;

/* By default, generate global declarations only */
Scope scope_out = SCOPE_EXTERN;

/* If TRUE, export typedef declarations (needed for lint-libs) */
#if OPT_LINTLIBRARY
boolean types_out = FALSE;
#endif

/* If TRUE, undef functions to avoid shadowing problems */
#if OPT_LINTLIBRARY
boolean lint_shadowed = FALSE;
#endif

/* If TRUE, generate variable declarations */
boolean variables_out = FALSE;

/* If TRUE, enable formal parameter promotion */
boolean promote_param = TRUE;

/* Style of function prototype to generate */
PrototypeStyle proto_style = PROTO_ANSI;

/* Function definition style converted to */
FuncDefStyle func_style = FUNC_NONE;

/* If TRUE, put guard macro around prototype parameters */
boolean proto_macro = FALSE;

/* Name of macro to guard prototypes */
char *macro_name = "P_";

/* If TRUE, output prototype macro definition */
boolean define_macro = TRUE;

/* If TRUE, output comments in prototypes */
boolean proto_comments = FALSE;

/* If TRUE, output comments naming source files */
boolean file_comments = TRUE;

/* Conditional compilation directive output in front of function definitions */
char *func_directive = "#ifdef ANSI_FUNC";

/* Output formats for function declarators */
FuncFormat fmt[] = {
    /* miscellaneous function declarator */
    { "", " ", "", "", " ", "" },
    /* prototype */
    { "", " ", "", "", " ", "" },
    /* function definition */
    { "", "\n", " ", "", " ", "" },
    /* function definition with parameter comments */
    { "", "\n", " ", "\n    ", "\n    ", "\n" },
};

/* If TRUE, don't output message if unable to read an include file */
boolean quiet = FALSE;

/* Include file directories */
#ifdef MSDOS
# ifdef __TURBOC__
int num_inc_dir = 2;
char *inc_dir[MAX_INC_DIR] = { "" , "/tc/include" };
# else
int num_inc_dir = 1;
char *inc_dir[MAX_INC_DIR] = { "" };
# endif
#else
# ifdef vms
int num_inc_dir = 2;
char *inc_dir[MAX_INC_DIR] = { "[]", "sys$library:" };
# else
int num_inc_dir = 2;
char *inc_dir[MAX_INC_DIR] = { "", "/usr/include" };
# endif
#endif

/* Run the C preprocessor */
#ifdef CPP
# if !HAVE_POPEN_PROTOTYPE
extern FILE *popen ARGS(( char *c,  char *m));
# endif
#if MSDOS
#define popen _popen
#define pclose _pclose
#else
extern int pclose ARGS((FILE *p));
#endif
static char *cpp = CPP, *cpp_opt, *cpp_cmd;
#endif

char *AxRefFileName;	/* produced '.r' file... see cpr_lex.h */

static  char *  escape_string   ARGS((char *src));
static  void    usage           ARGS((void));
static  void    process_options ARGS((int *pargc, char ***pargv));
static  void    parse_options   ARGS((char *src, int maxargc, int *pargc, char **argv));

/* Try to allocate some memory.
 * If unsuccessful, output an error message and exit.
 */
#ifdef NO_LEAKS
char *xMalloc(n,f,l) unsigned n; char *f; int l;
#else
char *xmalloc (n) unsigned n;
#endif
{
    char *p;
#if HAVE_LIBDBMALLOC
    p = debug_malloc(f, l, n);
#else
    p = malloc(n);
#endif

    if (p == NULL) {
    fprintf(stderr, "%s: out of memory (cannot allocate %d bytes)\n",
        progname, n);
    exit(EXIT_FAILURE);
    }
    *p = '\0';
    return p;
}

/* Copy the string into allocated memory.
 * If unsuccessful, output an error message and exit.
 */
#ifdef NO_LEAKS
char *xStrdup(src, f, l) char *src; char *f; int l;
#else
char *xstrdup (src) char *src;
#endif
{
#if defined(NO_LEAKS)
    return strcpy(xMalloc(strlen(src)+1, f, l), src);
#else
    return strcpy(xmalloc(strlen(src)+1), src);
#endif
}

/* Output the current source file name and line number.
 */
#if MSDOS       /*  {  */
    /* on NT we use this as a flag to avoid generating error lines for */
    /* the system includes, which are all below the "msdev" directory  */

    /* Extended this bad hack to include "devstudio" directories       */
int put_error ()
{
    char *filename, *p;
    int len, i;
    filename = cur_file_name();
    len = strlen(filename);
    for ( i=0; i<len; i++ )
    {
        if ( islower(filename[i]) )
            filename[i] = toupper(filename[i]);
    }
    p = strstr( filename, "MSDEV" );
    if ( p != (char *)NULL )
       return( FALSE );
    p = strstr( filename, "DEVSTUDIO" ) ;
    if ( p != (char *)NULL )
       return( FALSE );
    fprintf(stderr, "\"%s\", line %u: ", filename, cur_line_num());
    return( TRUE );

}
#else           /* }  {  */
void
put_error ()
{
    fprintf(stderr, "\"%s\", line %u: ", cur_file_name(), cur_line_num());
}
#endif    /* }  */






/* Scan for options from a string.
 */
static void
parse_options (src, maxargc, pargc, argv)
char *src;
int maxargc, *pargc;
char **argv;
{
    char *g, *p, c;
    int argc;

    argc = 0;
    g = xstrdup(src);
    c = *g;
    while (c != '\0' && argc < maxargc) {
    while (c == ' ' || c == '\t')
        c = *++g;
    if (c == '\0')
        break;
    argv[argc++] = g;

    p = g;
    while (1) {
        if (c == ' ' || c == '\t' || c == '\0') {
        *p = '\0';
        break;
        } else if (c == '"') {
        while (1) {
            c = *++g;
            if (c == '"') {
            c = *++g;
            break;
            } else if (c == '\0') {
            break;
            } else {
            *p++ = c;
            }
        }
        } else {
        *p++ = c;
        c = *++g;
        }
    }
    if (c != '\0')
        c = *++g;
    }

    *pargc = argc;
}

/* Replace any character escape sequences in a string with the actual
 * characters.  Return a pointer to malloc'ed memory containing the result.
 * This function knows only a few escape sequences.
 */
static char *
escape_string (src)
char *src;
{
    char *result, *get, *put;

    result = xstrdup(src);
    put = result;
    get = src;
    while (*get != '\0') {
    if (*get == '\\') {
        switch (*(++get)) {
        case 'n':
        *put++ = '\n';
        ++get;
        break;
        case 's':
        *put++ = ' ';
        ++get;
        break;
        case 't':
        *put++ = '\t';
        ++get;
        break;
        default:
        if (*get != '\0')
            *put++ = *get++;
        }
    } else {
        *put++ = *get++;
    }
    }
    *put = *get;
    return result;
}

/*
 * Returns true iff the character is a pathleaf separator
 */
int
is_path_sep (ch)
int ch;
{
#if defined(MSDOS) || defined(OS2)
    return ch == '/' || ch == '\\';
#else
    return ch == '/';
#endif
}

/* Trim any path name separator from the end of the string.
 * Return a pointer to the string.
 */
char *
trim_path_sep (s)
char *s;
{
    int n;

    n = strlen(s);
    if (n > 0) {
    if (is_path_sep(s[n-1]))
        s[n-1] = '\0';
    }
    return s;
}

/* Output usage message and exit.
 */
static void
usage ()
{
    fprintf(stderr, "usage: %s [ option ... ] [ file ... ]\n", progname);
    fputs("Options:\n", stderr);
    fputs("  -a, -t           Convert function definitions to ANSI or traditional style\n", stderr);
    fputs("  -b               Rewrite function definitions in both styles\n", stderr);
    fputs("  -c               Enable comments in prototype parameters\n", stderr);
    fputs("  -e               Output \"extern\" keyword before global declarations\n", stderr);
    fputs("  -f n             Set function prototype style (0 to 3)\n", stderr);
#if OPT_LINTLIBRARY
    fputs("  -l               Generate output in lint-library style\n", stderr);
#endif
    fputs("  -o file          Redirect output to file\n", stderr);
    fputs("  -O file          Redirect errors to file\n", stderr);
    fputs("  -p               Disable formal parameter promotion\n", stderr);
    fputs("  -q               Disable include file read failure messages\n", stderr);
    fputs("  -s               Output static declarations also\n", stderr);
    fputs("  -S               Output static declarations only\n", stderr);
#if OPT_LINTLIBRARY
    fputs("  -T               Output type definitions\n", stderr);
#endif
    fputs("  -v               Output variable declarations\n", stderr);
    fputs("  -x               Output variables and functions declared \"extern\"\n", stderr);
    fputs("  -m               Put macro around prototype parameters\n", stderr);
    fputs("  -M name          Set name of prototype macro\n", stderr);
    fputs("  -d               Omit prototype macro definition\n", stderr);
    fputs("  -P template      Set prototype format template \" int f (a, b)\"\n", stderr);
    fputs("  -F template      Set function definition format template \" int f (a, b)\"\n", stderr);
    fputs("  -C template      Set format for function definition with parameter comments\n", stderr);
    fputs("  -D name[=value]  Define C preprocessor symbol\n", stderr);
    fputs("  -U name          Undefine C preprocessor symbol\n", stderr);
    fputs("  -I directory     Add #include search directory\n", stderr);
    fputs("  -E command       Run specified C preprocessor command\n", stderr);
    fputs("  -E 0             Do not run any C preprocessor\n", stderr);
    fputs("  -V               Print version information\n", stderr);
    exit(EXIT_FAILURE);
}

#ifdef  vms
static  char    *cpp_defines;
static  char    *cpp_include;
static  char    *cpp_undefns;

static  void    add2list ARGS((char *dst, char *src));
static  void    add_option ARGS((char *keyword, char *src));

static void
add2list(dst, src)
    char    *dst;
    char    *src;
{
    if (*dst)
        strcat(dst, ",");
    strcat(dst, src);
}

static void
add_option(keyword, src)
    char    *keyword;
    char    *src;
{
    if (*src)
        (void)sprintf(cpp_opt + strlen(cpp_opt),
            " /%s=(%s)", keyword, src);
}
#endif  /* vms */

#ifdef QUOTE_POPEN_ARGS

/* Calculate length of string including shell quoting characters that
 * must be inserted to preserve the string when it is passed to /bin/sh.
 * On UNIX systems, popen() runs /bin/sh.
 */
#define QUOTECHARS "\"\'\t\n "

static int
quote_length (s)
char *s;
{
    int len = strlen(s);

    if (strpbrk(s, QUOTECHARS))  {
    len += 2;
    while (*s)
        if (*s++ == '\'')
        len += 4; /* replace ' with '"'"' (ick!) */
    }
    return len;
}

/* Insert quoting characters to preserve the string when it is passed to
 * /bin/sh.
 */
static char *
quote_string (s) 
char *s;
{
    if (strpbrk(s, QUOTECHARS))  {
    char *src = s;
    char *dup, *dup_orig;
    
    dup = dup_orig = xstrdup(s);

    while (isspace(*src))
        *src++ = *dup++;

    *src++ = '\'';
    while (*dup) {
        if (*dup == '\'') {
        *src++ = '\'';
        *src++ = '"';
        *src++ = '\'';
        *src++ = '"';
        *src++ = '\'';
        dup++;
        } else {
            *src++ = *dup++;
        }
    }

    *src++ = '\'';
    *src = '\0';
    free(dup_orig);
    }

    return s;
}
#else
#define quote_length(s) strlen(s)
#define quote_string(s) (s)
#endif /*QUOTE_POPEN_ARGS*/

#define MAX_OPTIONS 40

/* Process the command line options.
 */
static void
process_options (pargc, pargv)
int *pargc;
char ***pargv;
{
    int argc, eargc, nargc;
    char **argv, *eargv[MAX_OPTIONS], **nargv;
    int i, c;
    char *s;
#if defined(CPP)
    unsigned n;
#endif
#if defined(CPP) && !defined(vms)
    char tmp[MAX_TEXT_SIZE];
#endif

    argc = *pargc;
    argv = *pargv;
#ifndef vms /* this conflicts with use of foreign commands... */
    if ((s = getenv("CPROTO")) != NULL) {
    parse_options(s, MAX_OPTIONS, &eargc, eargv);
    nargv = (char **)xmalloc((eargc+argc+1)*sizeof(char *));
    nargv[0] = argv[0];
    nargc = 1;
    for (i = 0; i < eargc; ++i)
        nargv[nargc++] = eargv[i];
    for (i = 1; i < argc; ++i)
        nargv[nargc++] = argv[i];
    nargv[nargc] = NULL;
    argc = nargc;
    argv = nargv;
    }
#endif

#ifdef CPP
    /* Allocate buffer for C preprocessor command line. */
    n = strlen(cpp) + 1;
    for (i = 0; i < argc; ++i) {
    n += quote_length(argv[i]) + 1;  /* add more for possible quoting */
    }
#ifdef  vms
    cpp_include = xmalloc(n+argc);
    cpp_defines = xmalloc(n+argc);
    cpp_undefns = xmalloc(n+argc);
    n += 30;    /* for keywords */
#endif
    cpp_opt = xmalloc(n);
    cpp_cmd = xmalloc(n);
#endif

    while ((c = getopt(argc, argv, "aB:bC:cD:dE:eF:f:I:mM:P:pqR:SstU:Vvo:O:Tlx")) != EOF) {
    switch (c) {
    case 'I':
#ifdef  vms
        add2list(cpp_include, optarg);
        break;
#else   /* unix */
        if (num_inc_dir < MAX_INC_DIR) {
        inc_dir[num_inc_dir++] = trim_path_sep(xstrdup(optarg));
        } else {
        fprintf(stderr, "%s: too many include directories\n",
            progname);
        }
#endif
        /*FALLTHRU*/
    case 'D':
#ifdef  vms
        add2list(cpp_defines, optarg);
        break;
#endif
        /*FALLTHRU*/
    case 'U':
#ifdef  vms
        add2list(cpp_undefns, optarg);
        break;
#else   /* UNIX, etc. */
#ifdef CPP
        sprintf(tmp, " -%c%s", c, optarg);
        strcat(cpp_opt, quote_string(tmp));
#endif
#endif
        break;

    case 'a':
        func_style = FUNC_ANSI;
        break;
    case 'B':
        func_directive = optarg;
        break;
    case 'b':
        func_style = FUNC_BOTH;
        break;
    case 'c':
        proto_comments = TRUE;
        break;
    case 'd':
        define_macro = FALSE;
        break;
    case 'E':
#ifdef CPP
        if (strcmp(optarg, "0") == 0) {
        cpp = NULL;
        } else {
        cpp = optarg;
        }
#endif
        break;
    case 'e':
        extern_out = TRUE;
        break;
    case 'C':
    case 'F':
    case 'P':
        s = escape_string(optarg);
        i = (c == 'C') ? FMT_FUNC_COMMENT :
        ((c == 'F') ? FMT_FUNC : FMT_PROTO);

        fmt[i].decl_spec_prefix = s;
        while (*s != '\0' && isascii(*s) && !isalnum(*s)) ++s;
        if (*s == '\0') usage();
        *s++ = '\0';
        while (*s != '\0' && isascii(*s) && isalnum(*s)) ++s;
        if (*s == '\0') usage();

        fmt[i].declarator_prefix = s;
        while (*s != '\0' && isascii(*s) && !isalnum(*s)) ++s;
        if (*s == '\0') usage();
        *s++ = '\0';
        while (*s != '\0' && isascii(*s) && isalnum(*s)) ++s;
        if (*s == '\0') usage();

        fmt[i].declarator_suffix = s;
        while (*s != '\0' && *s != '(') ++s;
        if (*s == '\0') usage();
        *s++ = '\0';

        fmt[i].first_param_prefix = s;
        while (*s != '\0' && isascii(*s) && !isalnum(*s)) ++s;
        if (*s == '\0') usage();
        *s++ = '\0';
        while (*s != '\0' && *s != ',') ++s;
        if (*s == '\0') usage();

        fmt[i].middle_param_prefix = ++s;
        while (*s != '\0' && isascii(*s) && !isalnum(*s)) ++s;
        if (*s == '\0') usage();
        *s++ = '\0';
        while (*s != '\0' && isascii(*s) && isalnum(*s)) ++s;
        if (*s == '\0') usage();

        fmt[i].last_param_suffix = s;
        while (*s != '\0' && *s != ')') ++s;
        *s = '\0';

        break;
    case 'f':
        proto_style = atoi(optarg);
        if (proto_style < 0 || proto_style > PROTO_ANSI)
        usage();
        break;
    case 'm':
        proto_macro = TRUE;
        break;
    case 'M':
        macro_name = optarg;
        break;
    case 'p':
        promote_param = FALSE;
        break;
    case 'q':
        quiet = TRUE;
        break;
	case 'R':
		AxRefFileName = optarg;
		break;
    case 'S':
        scope_out = SCOPE_STATIC;
        break;
    case 's':
        scope_out = SCOPE_ALL;
        break;
    case 't':
        func_style = FUNC_TRADITIONAL;
        break;
    case 'V':
        fprintf(stderr, "%s patchlevel %d\n", rcsid, PATCHLEVEL);
        exit(EXIT_FAILURE);
        break;
    case 'v':
        variables_out = TRUE;
        break;
    case 'o':
        if (freopen(optarg, "w", stdout) == 0) {
        perror(optarg);
        exit(EXIT_FAILURE);
        }
        break;
    case 'O':
        if (freopen(optarg, "w", stderr) == 0) {
        perror(optarg);
        exit(EXIT_FAILURE);
        }
        break;
#if OPT_LINTLIBRARY
    case 'T':   /* emit typedefs */
        types_out = TRUE;
        break;
    case 'l':
        func_style    = FUNC_NONE;
        proto_style   = PROTO_LINTLIBRARY;
        extern_out    = FALSE;
        types_out     = TRUE;
        variables_out = TRUE;
#ifdef unix
        (void)strcat(cpp_opt, " -C");   /* pass-through comments */
#endif
        break;
#endif
    case 'x':
        extern_in = MAX_INC_DEPTH;
        break;
    default:
        usage();
    }
    }

#ifdef  vms
    add_option("includes", cpp_include);
    add_option("define",   cpp_defines);
    add_option("undefine", cpp_undefns);
#endif

    *pargc = argc;
    *pargv = argv;
}

int
main (argc, argv)
int argc;
char **argv;
{
    int i;
    FILE *inf;

	if (getenv("CATCH_CPROTO")) {
		static int spin = 1;
		while (spin);
	}
    /* Get the program name from the 0th argument, stripping the pathname
     * for readability.
     */
    progname = xstrdup(argv[0]);
#ifdef vms
    for (i = strlen(progname)-1; i >= 0; i--) {
        if (progname[i] == SQUARE_R
     || progname[i] == ':') {
        progname += (i + 1);
        break;
    } else if (progname[i] == '.') {
        progname[i] = '\0';
    }
    }
#else
    for (i = strlen(progname)-1; i >= 0; i--) {
    if (is_path_sep(progname[i])) {
        progname += (i + 1);
        break;
# if defined(MSDOS) || defined(OS2)
    } else if (progname[i] == '.') {
        progname[i] = '\0';
# endif
    }
    }
#endif
    argv[0] = progname; /* do this so getopt is consistent with us */

    process_options(&argc, &argv);

#if OPT_LINTLIBRARY
    if (lintLibrary())
        put_string(stdout, "/* LINTLIBRARY */\n");
#endif

    if (proto_macro && define_macro) {
    printf("#if __STDC__ || defined(__cplusplus)\n");
    printf("#define %s(s) s\n", macro_name);
    printf("#else\n");
    printf("#define %s(s) ()\n", macro_name);
    printf("#endif\n\n");
    }

    init_parser();
    if (optind == argc) {
    if (func_style != FUNC_NONE) {
        proto_style = PROTO_NONE;
        variables_out = FALSE;
        file_comments = FALSE;
    }
    process_file(stdin, "stdin");
    pop_file();
    } else {
    if (!optind)
        optind++;
    for (i = optind; i < argc; ++i) {
#ifdef CPP
# if CPP_DOES_ONLY_C_FILES
        /*
         * GCC (and possibly other compilers) don't pass-through the ".l"
         * and ".y" files to the C preprocessor stage.  Fix this by
         * temporarily linking the input file to a temporary-file with a
         * ".c" suffix.
         *
         * Of course, this solution assumes that the input directory is
         * writeable.
         */
        char temp[BUFSIZ];
        char *s = strcpy(temp, argv[i]);
        int len = strlen(temp);
        s += len - 1;
        if ((len > 2)
         && (s[-1] == '.')
         && (*s == 'l' || *s == 'y')) {
        while (s != temp && s[-1] != '/')
            s--;
        (void)strcpy(s, "XXXXXX.c");
            mktemp(temp);
            if (link(argv[i], temp) < 0)
            (void)strcpy(temp, argv[i]);
        }
#  define FileName temp
# else
#  define FileName argv[i]
#  ifdef vms
        char temp[BUFSIZ];
        (void)strcpy(temp, FileName);
#  endif
# endif
        if (func_style == FUNC_NONE && cpp != NULL) {
#ifdef vms
        /*
         * Assume the 'cpp' command contains a "%s" for the name of
         * the file that we're writing to.
         */
        sprintf(cpp_cmd, cpp,
            mktemp(strcpy(temp, "sys$scratch:XXXXXX.i")));
        sprintf(cpp_cmd + strlen(cpp_cmd), "%s %s", cpp_opt, FileName);
        system(cpp_cmd);
        inf = fopen(temp, "r");
#else
        sprintf(cpp_cmd, "%s%s %s", cpp, cpp_opt, FileName);
#if 0
	{
        /* output the command line to a file so we can run it for debug */
        FILE *fp = fopen("d:\\temp\\acproto.bat", "w+");
        if ( fp != (FILE *)NULL )
            {
            fprintf(fp, "%s\n", cpp_cmd );
            fclose( fp );
             }
	}
#endif

        inf = popen(cpp_cmd, "r");
#endif
        if (inf == NULL) {
            fprintf(stderr, "%s: error running %s\n", progname,
             cpp_cmd);
            continue;
        }
        } else {
        if ((inf = fopen(argv[i], "r")) == NULL) {
            fprintf(stderr, "%s: cannot read file %s\n", progname,
             argv[i]);
            continue;
        }
        }
#else
        if ((inf = fopen(argv[i], "r")) == NULL) {
        fprintf(stderr, "%s: cannot read file %s\n", progname, argv[i]);
        continue;
        }
#endif
        process_file(inf, argv[i]);
#ifdef CPP
        if (func_style == FUNC_NONE && cpp != NULL) {
#ifdef vms
        fclose(inf);
#else
        pclose(inf);
#endif
#if CPP_DOES_ONLY_C_FILES || defined(vms)
        if (strcmp(argv[i], temp)) {
            (void)unlink(temp);
        }
        pop_file();
#endif
        } else {
        pop_file();
        }
#else   /* no CPP defined */
        pop_file();
#endif
    }
    }

    if (proto_macro && define_macro) {
    printf("\n#undef %s\n", macro_name);
    }

#ifdef NO_LEAKS
# ifdef CPP
    if (cpp_opt != 0) free(cpp_opt);
    if (cpp_cmd != 0) free(cpp_cmd);
#  ifdef vms
    if (cpp_include != 0) free(cpp_include);
    if (cpp_defines != 0) free(cpp_defines);
    if (cpp_undefns != 0) free(cpp_undefns);
#  endif
# endif
    while (num_inc_dir-- > 2) {
    free(inc_dir[num_inc_dir]);
    }
    free_parser();
# if OPT_LINTLIBRARY
    free_lintlibs();
# endif

# ifdef DOALLOC
    show_alloc();
# endif
#endif

#if HAVE_LIBDBMALLOC
    malloc_dump(fileno(stderr));
#endif
    exit(EXIT_SUCCESS);
    return EXIT_SUCCESS;
}

#if PMAX  /* { */
char *strdup(str)
char *str;
{
  return(strcpy(malloc(strlen(str)+1),str));
}
#endif  /* } PMAX */


#if MSDOS           /*  {  */

/*** our own version of getopt. taken from /ax/pgms/utils/utgtopt.c ***/
#define BADCH   (int)'?'
#define EMSG    ""
#ifdef VERBOSE
#define tell(s) fputs(*nargv,stderr);fputs(s,stderr); \
        fputc(optopt,stderr);fputc('\n',stderr);return(BADCH);
#else
#define tell(s) return(BADCH);
#endif
int
getopt(nargc, nargv, ostr)
int     nargc;
char *  nargv[];
char  *ostr;
{
    static char *place = EMSG;  /* option letter processing */
    register char   *oli;       /* option letter list index */

    if(!*place) {
        /* update scanning pointer */
        if(   optind >= nargc
           || *(place = nargv[optind]) != '-'
           || !*++place)
        {
            return(EOF);
        }
        if (*place == '-') {
            /* found "--" */
            ++optind;
            return(EOF);
        }
    }
    /* option letter okay? */
    if (   (optopt = (int)*place++) == (int)':'
        || !(oli = strchr(ostr, optopt)))
    {
        if(!*place) ++optind;
        tell(": illegal option -- ");
    }
    if (*++oli != ':') {
        /* don't need argument */
        optarg = NULL;
        if (!*place) ++optind;
    }
    else {
        /* need an argument */
        if (*place) {
            /* no white space */
            optarg = place;
        }
        else if (nargc <= ++optind) {
            /* no arg */
            place = EMSG;
            tell(": option requires an argument -- ");
        }
        else {
            /* white space */
            optarg = nargv[optind];
        }
        place = EMSG;
        ++optind;
    }
    return(optopt);                     /* dump back option letter */
}






/* Copyright (C) 1994 Free Software Foundation, Inc.
This file is part of the GNU C Library.

The GNU C Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU C Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU C Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

/*
 * Tree search generalized from Knuth (6.2.2) Algorithm T just like
 * the AT&T man page says.
 *
 * The node_t structure is for internal use only, lint doesn't grok it.
 *
 * Written by reading the System V Interface Definition, not the code.
 *
 * Totally public domain.
 */
/*LINTLIBRARY*/

/* This routine is not very bad. It makes many assumptions about
 * the compiler. It assumpts that the first field in node must be
 * the "key" field, which points to the datum. It is a very trick
 * stuff. H.J.
 */

typedef struct node_t
{
    void	*key;
    struct node_t *left, *right;
} node;

/* find or insert datum into search tree.
char 	*key;			 key to be located
register node	**rootp;	 address of tree root
int	(*compar)();		 ordering function
*/
void *tsearch( void *key, void **vrootp, int (*compar)() )
{
    register node *q;
    register node **rootp = (node **) vrootp;

    if (rootp == (struct node_t **)0)
	return ((struct node_t *)0);
    while (*rootp != (struct node_t *)0)	/* Knuth's T1: */
    {
	int r;

	if ((r = (*compar)(key, (*rootp)->key)) == 0)	/* T2: */
	    return (*rootp);		/* we found it! */
	rootp = (r < 0) ?
	    &(*rootp)->left :		/* T3: follow left branch */
	    &(*rootp)->right;		/* T4: follow right branch */
    }
    q = (node *) malloc(sizeof(node));	/* T5: key not found */
    if (q != (struct node_t *)0)	/* make new node */
    {
	*rootp = q;			/* link new node to old */
	q->key = key;			/* initialize new node */
	q->left = q->right = (struct node_t *)0;
    }
    return (q);
}

typedef enum { preorder, postorder, endorder, leaf } VISIT;

/* Walk the nodes of a tree
register node	*root;		Root of the tree to be walked
register void	(*action)();	Function to be called at each node
register int	level;
*/
static void trecurse( struct node_t *vroot, void (*action)(), int level )
{
    register node *root = (node *) vroot;

    if (root->left == (struct node_t *)0 && root->right == (struct node_t *)0)
	(*action)(root, leaf, level);
    else
    {
	(*action)(root, preorder, level);
	if (root->left != (struct node_t *)0)
	    trecurse(root->left, action, level + 1);
	(*action)(root, postorder, level);
	if (root->right != (struct node_t *)0)
	    trecurse(root->right, action, level + 1);
	(*action)(root, endorder, level);
    }
}

/* void twalk(root, action)		Walk the nodes of a tree 
node	*root;			Root of the tree to be walked
void	(*action)();		Function to be called at each node
PTR
*/
void twalk( struct node_t *vroot, void (*action)() )
{
    register  node *root = (node *) vroot;

    if (root != (node *)0 && action != NULL )
	trecurse(root, action, 0);
}


#endif            /*  }  */
