/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: junkscan.c
   Applix ASC version: 13.1
   Date: Thu May 13 21:13:47 1999
*/

/*
 *
 *      AFindFiles is used to recursive search thru a directory.
 *      It is a slightly frendlier interface to AgetNfname.
 *      This is called repetitively to get all of the files in a directory.
 *      The FIRST call has topdir being the directory to search.  Subsequent
 *      calls have topdir NULL.
 *
 *  AFindFiles(topdir,maxlevels,dir,fname)
 *  char *topdir;           /* IN Name of directory to recursively search
 *  int maxlevels;          /* maximumn levels of recursion, 0 for non-recurse
 *  char *dir;              /* OUT Name of file's (sub)directory
 *  char *fname;            /* OUT Name of entry file
 *
 *  Return Values:
 *
 *      0 - File found
 *      1 - No file found, end of directory search
 *      2 - File found but stat failed
 *      3 - Subdirectory encounter, subsequent calls report files
 *
 *  #include <stdio.h> 
 *  #include "am_aos.h"
 *
 *  main()
 *  {
 *      char *first;
 *      char dir[STRPATH + 1];
 *      char fname[STRPATH + 1];
 *
 *      first = dir;
 *      strcpy(first,"/user/hae");
 *      while(AFndFile(first,dir,fname) != 1) {
 *          first = NULL;
 *          printf("directory = %s, file = %s\n",dir,fname);
 *      }
 *  }
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <am_aos.h>

#if !AXMS
#include <sys/dir.h>
#define DIRSLASH '/'
#endif

#if _MSDOS        /*  {  */
#include <windows.h>
#define DIRSLASH '\\'
#define ALCSTR(p) strcpy(malloc(strlen(p)+1),p)
BOOL IsDir(LPWIN32_FIND_DATA);

struct direct {
    char d_name[256];
} direct;
#endif           /*  }  */

int AgetNfilename();
char *AMakePath();


int AFindFiles(topdir,maxlevels,dir,fname)
register char *topdir;       /* IN Name of directory to recursively search */
register int maxlevels;      /* maximumn levels of recursion */
register char *dir;          /* OUT Name of file's (sub)directory */
register char *fname;        /* OUT Name of entry file */
{
    static struct getinfo *place = NULL;
    static char directory[STRPATH + 1];
    static int level;
    struct stat file_info;
    char *strrchr();
    static int init;

    if(topdir) {
        init = TRUE;
        level = 0;
        strcpy(directory,topdir);
    }

try_again:
    if(AgetNfilename(init,directory,fname,&place)) {
        init = FALSE;
        AMakePath(dir,directory,fname);
        
        if(stat(dir,&file_info)) {
            strcpy(dir,directory);
            return(2);
        }
        if(file_info.st_mode&S_IFDIR        &&
                (strcmp(fname, ".")) != 0   &&        
                (strcmp(fname, "..")) != 0) {
            if ( level == maxlevels )
                goto try_again;
            level++;
            init = TRUE;
            strcpy(directory,dir);
            return(3);
        }
        strcpy(dir,directory);
        return(0);
    }
    
    if(level) {
        level--;
        if(topdir = strrchr(directory, DIRSLASH))
            *topdir = '\0';
        init = FALSE;
        goto try_again;
    }
    return(1);
}


static int checkSkip(name)
struct direct *name;
{
	/*
	 *	if(!strcmp((*name).d_name,".") ||
	 *		!strcmp((*name).d_name,"..")) return (0);
	 */
	if((*name).d_name[0] == '.' && 
	   (((*name).d_name[1] == '.' && (*name).d_name[2] == '\0') ||
		(*name).d_name[1] == '\0'))
		return(0);
	return(1);

}


#define MAXLEVEL 20

#if SYS&B4_2 || _MSDOS  /*  {  */
struct getinfo {
	int current;
	int count;
	struct direct **dirlist;
	struct direct  *p;
	struct direct **startlist;
	struct direct **dirstack[MAXLEVEL];
	struct direct **startstack[MAXLEVEL];
	int level;
	int countstack[MAXLEVEL];
	int currstack[MAXLEVEL];
};
#else  /* }  {  */

#include <stdio.h>
struct getinfo {
	struct direct direct;       
	int level;
	long posstack[MAXLEVEL];
	char *filestack[MAXLEVEL];
	char  pathname[256];
	FILE *fp;
};
#endif  /*  }  */




int AgetNfilename(init, dirpath, buffer, place)
int init;
char *dirpath;
char *buffer;
struct getinfo **place;
{

#if SYS&B4_2 || _MSDOS    /*  {  */
	int scandir ();

	if ((*place) == NULL) {
		(*place) = (struct getinfo *) calloc(1,sizeof(struct getinfo));
		(*place)->level = 0;
	}
	
	if (init) {
		/* Sanity check. */
		if ((*place)->level >= MAXLEVEL)
			return(FALSE);
		if ((*place)->level > 0) {
			(*place)->dirstack[(*place)->level] = (*place)->dirlist;
			(*place)->startstack[(*place)->level] = (*place)->startlist;
			(*place)->countstack[(*place)->level] = (*place)->count;
			(*place)->currstack[(*place)->level] = (*place)->current;
		}
		(*place)->level++;
		(*place)->startlist = NULL;
		(*place)->count = scandir (dirpath, &(*place)->startlist, checkSkip, NULL);
		(*place)->dirlist = (*place)->startlist;
		(*place)->current = 0;
	}
	if ((*place)->current++ >= (*place)->count) {
		if((*place)->startlist != NULL)
			free ((*place)->startlist);
		if(--((*place)->level) > 0) {
			(*place)->dirlist = (*place)->dirstack[(*place)->level];
			(*place)->startlist = (*place)->startstack[(*place)->level];
			(*place)->current = (*place)->currstack[(*place)->level];
			(*place)->count = (*place)->countstack[(*place)->level];
		}
		else {
			free((*place));
			(*place) = NULL;
		}
		return (FALSE);
	}
	(*place)->p = *(*place)->dirlist++;
	strcpy (buffer, (*(*place)->p).d_name);
	return (TRUE);

#else  /* } SYS&B4_2 || _MSDOS {  */

	char *p;
	char *malloc ();
	int count;

	if ((*place) == NULL) {
		(*place) = (struct getinfo *) calloc(1,sizeof(struct getinfo));
		(*place)->level = 0;
	}
	
	if (init) {
		/* Sanity check. */
		if ((*place)->level >= MAXLEVEL)
			return(FALSE);
		if ((*place)->level > 0) {
			p = malloc (strlen ((*place)->pathname) + 1);
			strcpy (p, (*place)->pathname);
			(*place)->filestack[(*place)->level] = p;
			(*place)->posstack[(*place)->level] = ftell ((*place)->fp);
			fclose ((*place)->fp);
		}
		(*place)->level ++;
		(*place)->fp = fopen (dirpath, "r");
		strcpy ((*place)->pathname, dirpath);
		if ((*place)->fp == NULL) {
			if ((*place)->level > 0) {
				(*place)->level --;
				p = (*place)->filestack[(*place)->level];
				strcpy ((*place)->pathname, p);
				free (p);
				(*place)->fp = fopen ((*place)->pathname, "r");
				if ( (*place)->fp )
					fseek ((*place)->fp, (*place)->posstack[(*place)->level], 0);
			}
			else {
				free((*place));
				(*place) = NULL;
			}
			return (FALSE);
		}
	}
	while (TRUE) {
		count = fread ((char *) &(*place)->direct, sizeof (struct direct),
					   1, (*place)->fp);
		if (count < 1) {
			fclose ((*place)->fp);
			(*place)->level--;
			if ((*place)->level > 0) {
				p = (*place)->filestack[(*place)->level];
				strcpy ((*place)->pathname, p);
				free (p);
				(*place)->fp = fopen ((*place)->pathname, "r");
				if ( (*place)->fp )
					fseek ((*place)->fp, (*place)->posstack[(*place)->level], 0);
			}
			else {
				free((*place));
				(*place) = NULL;
			}
			return (FALSE);
		}
		if (((*place)->direct.d_ino != 0) && (checkSkip (&(*place)->direct) == 1))
			break;
	}
	/* copy name and append a null */
{
	char *bp, *dp, *edp;
	edp = (*place)->direct.d_name + sizeof((*place)->direct.d_name);
	dp = (*place)->direct.d_name;
	bp = buffer;
	while (dp < edp && *dp) 
		*bp++ = *dp++;
	*bp = 0;
}
	return (TRUE);
#endif   /*  }  SYS&B4_2 || _MSDOS   */
}






#if _MSDOS      /*  {  */

int scandir( dirname, ret_list, select, comp_func )
char *dirname;
struct direct ***ret_list;
int (*select)();
void (*comp_func)();
{
   int num_files = 0;
   char name[STRPATH + 1];
   WIN32_FIND_DATA data;
   HANDLE          h;
   int return_count;
   struct direct *entryP;
   struct direct **filesP;


   return_count = 0;
   filesP = (struct direct **)malloc( sizeof(struct direct *) );

   sprintf(name,"%s\\*", dirname);

   h = FindFirstFile(name,&data);
   if (h == INVALID_HANDLE_VALUE)
      {
      /*printf("\nNo files in %s. Check directory name.\n", name);*/
      return( 0 );
      }

   if ( (strcmp(data.cFileName, ".") != 0) &&
        (strcmp(data.cFileName, "..") != 0 ) )
      {
      filesP = (struct direct **)realloc( filesP, 
                              ((return_count+1) * sizeof(struct direct *)) );
      entryP = (struct direct *)malloc( sizeof(direct) );
      *(filesP + return_count) = entryP;
      
      return_count++;
      strcpy( entryP->d_name, data.cFileName );
      }

   while (FindNextFile(h,&data))
      {
      if ( (strcmp(data.cFileName, ".") == 0) || 
           (strcmp(data.cFileName, "..") == 0 ) )
         continue;
      filesP = (struct direct **)realloc( filesP, 
                                  ((return_count+1)* sizeof(struct direct *)) );
      entryP = (struct direct *)malloc( sizeof(direct) );
      *(filesP + return_count) = entryP;
      return_count++;
      strcpy( entryP->d_name, data.cFileName );
      }

   FindClose(h);
   *ret_list = filesP;
   return( return_count );
}


#endif        /*  }  */
