/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: scansys5.c
   Applix ASC version: 10
   Date: Sun Jan 24 21:13:17 1999
*/

/*
.BEGIN MANPAGE
.NAME AgetNfilename - Get the names of files in a directory
.DESCRIPTION
	AgetNfilename() sequentially returns the entries in a directory. 

	Complete directory trees can be walked by calling AgetNfilename() with
	init equal to TRUE each time it returns a directory name.  The type
	of a directory entry (file, subdirectory, symbolic link, etc.) can be
	obtained from Astat() on all systems.  Up to 20 levels of AgetNfilename()
	can be nested.  Levels are automatically popped when all the entries have
	been returned.
.ARGUMENT init
	Set to TRUE when beginning new directory, FALSE otherwise.
.ARGUMENT dirpath
	When init is TRUE, dirpath is the address of a string which identifies
	the directory to be processed, otherwise it is ignored.
.ARGUMENT buffer
	The address of a buffer to which a directory entry should be returned.
	The buffer should be at least 33 bytes long.
.ARGUMENT stack
	The address of a pointer used to maintain currency.
	The pointer should be set to NULL on the first AgetNfilename() call.
.SEE ALSO
	Astat(3A)
.RETURN VALUE
	AgetNfilename() returns TRUE when it finds a valid entry and FALSE when
	it runs out of entries.
.END MANPAGE
*/

#include <am_aos.h>
#include <stdio.h>  	/* For pre-System 5.3 systems */
#include <sys/stat.h>
#define TRUE 1 
#define FALSE 0 

#define MAXLEVEL 20		/* Max directory levels */


	/****************************************************************/
	/*		AEGIS and VMS includes, defines, and declarations		*/
    /*      have been removed. See Rev 6.5 or earlier of this file  */
    /*      if you need this code.                                  */
	/****************************************************************/


    /********************************************************/
    /*      UNIXy includes, defines, and declarations       */
	/*			beware all who enter this code...			*/
    /********************************************************/


#include <sys/types.h>

#if HP800
#include <ndir.h>
#else
#if X386_NEW || M_88K || AVIION
#define MAXNAMLEN   512     /* maximum filename length */
struct dirent               /* data from readdir() */
    {
    long        d_ino;      /* inode number of entry */
    off_t       d_off;      /* offset of disk direntory entry */
    unsigned short  d_reclen;   /* length of this record */
    char        d_name[255];    /* JJM was 1 - name of file */
    };
#define DIRBUF      1048        /* buffer size for fs-indep. dirs */
 
typedef struct
    {
    int dd_fd;          /* file descriptor */
    int dd_loc;         /* offset in block */
    int dd_size;        /* amount of valid data */
    char    *dd_buf;        /* directory block */
    }   DIR;            /* stream data from opendir() */
 
extern DIR      *opendir();
extern struct dirent    *readdir();
extern long     telldir();
extern void     seekdir();
extern int      closedir();
#define direct dirent       /* JJM */
 
#else
#if CLIX 
#include <dirent.h>
#define direct dirent
#else
#if RIOS

typedef struct _dirdesc {
	int	dd_fd;		/* file descriptor of directory */
	int	dd_blksize;	/* this filesystem's block size */
	char	*dd_buf;	/* malloc'd buffer depending of fs bsize */
	long	dd_size;	/* size of buffer */
	long	dd_flag;	/* private flags for readdir, unused */
	long	dd_loc;		/* logical(dirent) offset in  directory */
	long	dd_curoff;	/* real offset in directory corresponding
				 * to dd_loc */
} DIR;

struct  dirent {
	ulong_t     d_offset;   /* real off after this entry */
	ino_t       d_ino;      /* inode number of entry */
	/* make ino_t when it's ulong */
	ushort_t    d_reclen;   /* length of this record */
	ushort_t    d_namlen;   /* length of string in d_name */
	char            d_name[255+1];
};
	
#define direct dirent

#else
#if SYSV4 || SUN4S || NEC_EWS
 
#define direct dirent
#include <sys/types.h>  /* for ino_t and off_t */
#include <sys/param.h>  /* for MAXNAMELEN */

/* just in case */
#ifndef MAXNAMELEN
#define MAXNAMELEN 1024
#endif


/*
 * NOTE:
 * We need our own "struct dirent" because we do a structure assignment in
 * scandir().  On SYSV ports, the d_name[] element of "struct dirent" is a
 * char array of size 2, so a structure assignment copies only the first two
 * chars of the filename.
 */
/* from <dirent.h> */
typedef struct {
    int  dd_fd; /* file descriptor */
    int  dd_loc;    /* offset in block */
    int  dd_size;   /* amount of valid data */
    char *dd_buf;   /* directory block */
} DIR;  /* stream data from opendir() */
 
/* from <sys/dirent.h> */
struct dirent {
    ino_t d_ino;    /* "inode number of entry */
    off_t d_off;    /* offset of disk directory entry */
    unsigned short d_reclen;    /* length of this record */
    char  d_name[MAXNAMELEN + 1];/* <--- name of file -- for struct assign.*/
};
 
extern struct dirent *readdir(/* DIR * */);
extern int closedir(/* DIR * */);
extern DIR *opendir(/* const char * */);
/* end of SYSV4 */
 
#else

/* DEFAULT */
#include <sys/dir.h>

#endif  /* if SYSV4 */
#endif  /* if CLIX */
#endif  /* RIOS */
#endif  /* if X386_NEW */
#endif  /* if HP800 */

int AFindFiles(topdir,maxlevels,dir,fname)
register char *topdir;       /* IN Name of directory to recursively search */
register int maxlevels;      /* maximumn levels of recursion */
register char *dir;          /* OUT Name of file's (sub)directory */
register char *fname;        /* OUT Name of entry file */
{
    static struct getinfo *place = NULL;
    static char directory[STRPATH + 1];
    static int level;
    struct stat file_info;
    char *strrchr();
    static int init;
 
    if(topdir) {
        init = TRUE;
        level = 0;
        strcpy(directory,topdir);
    }
 
try_again:
    if(AgetNfilename(init,directory,fname,&place)) {
        init = FALSE;
        AMakePath(dir,directory,fname);
      
        if(stat(dir,&file_info)) {
            strcpy(dir,directory);
            return(2);
        }
        if(file_info.st_mode&S_IFDIR        &&
                (strcmp(fname, ".")) != 0   &&
                (strcmp(fname, "..")) != 0) {
            if ( level == maxlevels )
                goto try_again;
            level++;
            init = TRUE;
            strcpy(directory,dir);
            return(3);
        }
        strcpy(dir,directory);
        return(0);
    }
    
    if(level) {
        level--;
        if(topdir = strrchr(directory, '/') )
            *topdir = '\0';
        init = FALSE;
        goto try_again;
    }
    return(1);
}

static int checkSkip(name)
struct direct *name;
{
	/*
	 *	if(!strcmp((*name).d_name,".") ||
	 *		!strcmp((*name).d_name,"..")) return (0);
	 */
	if(name->d_name[0] == '.' && 
	   ((name->d_name[1] == '.' && name->d_name[2] == '\0') ||
		name->d_name[1] == '\0'))
		return(0);
	return(1);
}

struct getinfo {
	int current;
	int count;
	struct direct **dirlist;
	struct direct  *p;
	struct direct **startlist;
	struct direct **dirstack[MAXLEVEL];
	struct direct **startstack[MAXLEVEL];
	int level;
	int countstack[MAXLEVEL];
	int currstack[MAXLEVEL];
};







/* BEGIN SYNOPSIS */
AgetNfilename(init, dirpath, buffer, stack)
int init;
char *dirpath;
char *buffer;
char **stack;
/* END SYNOPSIS */
{
	/*****************************************************************/
	/*		VMS and AEGIS code sections have                         */
    /*      been removed. See rev 6.5 or earlier of this file        */
    /*      if you need to see that code.                            */
	/*****************************************************************/


	/********************************************************/
	/*		UNIXy code section								*/
	/********************************************************/
	struct getinfo **place = (struct getinfo **)stack;

	if (*place == NULL) {
		*place = (struct getinfo *)malloc(sizeof(struct getinfo));
		(*place)->level = 0;
	}
	
	if (init) {
		/* Sanity check. */
		if ((*place)->level >= MAXLEVEL)
			return(FALSE);
		if ((*place)->level > 0) {
			(*place)->dirstack[(*place)->level] = (*place)->dirlist;
			(*place)->startstack[(*place)->level] = (*place)->startlist;
			(*place)->countstack[(*place)->level] = (*place)->count;
			(*place)->currstack[(*place)->level] = (*place)->current;
		}
		(*place)->level++;
		(*place)->startlist = NULL;
		(*place)->count = scandir(dirpath,&(*place)->startlist,checkSkip,
			NULL);
		(*place)->dirlist = (*place)->startlist;
		(*place)->current = 0;
	}
	if ((*place)->current++ >= (*place)->count) {
		if((*place)->startlist != NULL)
			free ((*place)->startlist);
		if(--((*place)->level) > 0) {
			(*place)->dirlist = (*place)->dirstack[(*place)->level];
			(*place)->startlist = (*place)->startstack[(*place)->level];
			(*place)->current = (*place)->currstack[(*place)->level];
			(*place)->count = (*place)->countstack[(*place)->level];
		}
		else {
			free(*place);
			*place = NULL;
		}
		return (FALSE);
	}
	(*place)->p = *(*place)->dirlist++;


	strcpy(buffer,(*(*place)->p).d_name);

	free((char *)(*place)->p);
	return(TRUE);
}






#if SYS&SYS5 || SYSV4
/*
 *	Name: scandir() - scan a directory.
 *
 *	Description: Scandir reads the directory dirname and builds an array of
 *		pointers to directory entries using malloc(3).  The second parameter is
 *		a pointer to an array of structure pointers. The third parameter is a
 *		pointer to a routine which is called with a pointer to a directory
 *		entry and should return a none zero value if the directory should be
 *		included in the array.  If this pointer is null, than all the directory
 *		entries will be included.  The last argument is a pointer to a routine
 *		which is passed to qsort(3) to sort the completed array.  If this
 *		pointer is null, the array is not sorted. 
 *
 *		Scandir returns the number of entries in the array and a pointer to the
 *		array through the parameter namelist.
 */

scandir(dirname,namelist,select,compar)
char *dirname;
struct direct ***namelist;
int (*select)();
int (*compar)();
{
#if (SYSV3 | HP300 | X386_NEW | HP800 | SGI | AUX | RIOS | AVIION | M_88K | SUN4S | CLIX | SYSV4 | NEC_EWS) && (!UNISYS)
	DIR *fp;
#else
	FILE *fp;
	struct direct ldir;
#endif
#if PCS
	struct dirent *mips_dir;
	struct direct ldir;
#endif
	struct direct *dir,*dirptr,**nameptr;
	int array_cnt = 0;
	int dir_cnt = 0;

#if (SYSV3 | HP300 | X386_NEW | HP800 | SGI | AUX | RIOS | AVIION | M_88K | SUN4S | CLIX | SYSV4 | NEC_EWS) && (!UNISYS)
	if(!(fp = opendir(dirname)))
#else
	dir = &ldir;
	if(!(fp = fopen(dirname,"r")))
#endif
		return(-1);

#if PCS
	dir = &ldir;
	while(mips_dir = readdir(fp)) {
		dir->d_ino = mips_dir->d_ino;
		strcpy(dir->d_name,mips_dir->d_name);
#else
#if (SYSV3 | HP300 | X386_NEW | HP800 | SGI | AUX | RIOS | AVIION | M_88K | SUN4S | CLIX | SYSV4 | NEC_EWS) && (!UNISYS) && (!PCS)
	while(dir = readdir(fp)) {
#else
	while(fread(dir,sizeof(struct direct),1,fp)) {
#endif
#endif
		if(!dir->d_ino || (select && !(*select)(dir)))
			continue;
		if(dir_cnt == array_cnt) {
			if(array_cnt) {
				array_cnt *= 2;
				if(!(nameptr = (struct direct **)
					malloc((unsigned int)
						(sizeof(struct direct *) * array_cnt))))
					goto free_mem;
                memmove( (char *)nameptr, (char *)*namelist,
                         (int)(sizeof(struct direct *) * dir_cnt) );
				free((char *)*namelist);
				*namelist = nameptr;
			}
			else {
				array_cnt = 64;
				if(!(*namelist = nameptr = (struct direct **)
					malloc((unsigned int)(sizeof(struct direct *) * array_cnt))))
					return(-1);
			}
		}
		if(!(dirptr = (struct direct *)
			malloc((unsigned int)sizeof(struct direct))))
			goto free_mem;
		*dirptr = *dir;
		nameptr[dir_cnt++] = dirptr;
	}
#if (SYSV3 | HP300 | X386_NEW | HP800 | SGI | AUX | RIOS | AVIION | M_88K | SUN4S | CLIX | SYSV4 | NEC_EWS) && (!UNISYS)
	closedir(fp);
#else 
	fclose(fp); 
#endif

	if(compar && dir_cnt > 1)
		qsort(nameptr,dir_cnt,(int)sizeof(struct direct *),*compar);

	return(dir_cnt);

free_mem:
#if (SYSV3 | HP300 | X386_NEW | HP800 | SGI | AUX | RIOS | AVIION | M_88K | SUN4S | CLIX | SYSV4 | NEC_EWS) && (!UNISYS)
	closedir(fp);
#else
	fclose(fp);
#endif
	nameptr = *namelist;
	while(dir_cnt--)
		free((char *)nameptr[dir_cnt]);
	free((char *)nameptr);
	return(-1);
}
#endif	/* SYS&SYS5 */
