/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axlist.cc
   Applix ASC version: 1
   Date: Mon Nov 16 14:27:53 1998
*/
#include "axlist.hh"

// Constructor
list::list(int (*cmpFunc)(void*, void*) )
{
	head 	= 0;
	tail 	= 0;
	cmp 	= cmpFunc;
	iter 	= 0;
	n_elem 	= 0;
}

// Destructor
list::~list()
{
	node *tmp, *rem;
	
	tmp = head;
	rem = head;
	for (int ix=0;ix<n_elem;ix++) {
		tmp = rem->next;
		delete rem;
		rem = tmp;
	}	

}

long
list::number_of_elements(void)
{
	return(n_elem);
}

void *
list::element(long n)
{
	int ix;
	node *tmp;

	if (n < 0 || n > n_elem)
		return(0);
	for(ix=0,tmp=head;ix!=n && tmp;ix++,tmp=tmp->next)
		;
	if (tmp)
		return(tmp->elem);
	else
		return(0);
}
long 
list::unlink(void *info, long n)
{
	node *tmp, *nextNode, *prevNode;	
	int ix;
	if (n >= 0 && n <= n_elem)  {
		/* unlink the nth element */
		for(ix=0,tmp=head;ix!=n && tmp;ix++,tmp=tmp->next)
			;
		if (tmp) {
			this->_unlink(tmp);
			n_elem = n_elem-1;
		}
	}
	else if (info) {
		for (ix=0,tmp=head;ix<n_elem && tmp;ix++,tmp=tmp->next) {
			if (0 == this->_cmp(info, tmp)) {
				this->_unlink(tmp);
				n_elem = n_elem-1;
				break;
			}
		}
	}
	return(n_elem);
}

int
list::comp_function(int (*cmpFunc)(void*, void*) )
{
	cmp = cmpFunc;
	return(OK);
}

long 
list::add(void *info)
{
	if (!head) {
		head = tail = this->new_node(info);
	}
	else {
		tail->next = this->new_node(info);
		tail->next->prev = tail;
		tail = tail->next;
	}
	n_elem = n_elem+1;
	return(n_elem);
}

void *
list::find(void *info)
{
	node *tmp = 0;
	int ix = 0;
	for (ix=0,tmp=head;ix<n_elem && tmp;ix++,tmp=tmp->next) 
		if (0 == this->_cmp(info, tmp)) 
			return(tmp->elem);					// Found

	return(0);
}

void 
list::init_iter(void *elem)
{
	if (elem) {
		node *tmp;
		int ix = 0;
		for (ix=0,tmp=head;ix<n_elem && tmp;ix++,tmp=tmp->next) {
			if (0 == this->_cmp(elem, tmp)) {
				iter = tmp;
				return;
			}
		}
	}
	iter = head;
}

void *
list::next(void)
{
	void *ret_elem = 0;
	if (!iter)
		return 0;

	ret_elem = iter->elem;
	iter = iter->next;
	return(ret_elem);
}
	
// Private Methods 
int
list::_cmp(void *info, node *tmp)
{
    return(cmp(info, tmp->elem));
}


node *
list::new_node(void *info)
{
	node *new_node;

	new_node = new node;

	new_node->next	= 0;
	new_node->prev 	= 0;
	new_node->elem 	= info;
	return(new_node);
}

void
list::_unlink(node *tmp) 
{
	node *nextNode, *prevNode;

	if (tmp == head) {
		head = tmp->next;
		if (head)
			head->prev = 0;
	}
	else if (tmp == tail) {
		tail = tmp->prev;
		tail->next = 0;
	}
	else {
		prevNode = tmp->prev;
		nextNode = tmp->next;
		prevNode->next = nextNode;
		nextNode->prev = prevNode;
	}
	tmp->elem = 0;
	delete tmp;
}

