/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: bldrbltn.c
   Applix ASC version: 14
   Date: Mon Nov 16 13:56:45 1998
*/
#include <stdio.h>
#include <math.h>
 
#include <aos.h>
#include <global.h>
#include <elf.h>
#include <elf_ss.h>
#include <afile.h>
#include <aiconids.h>
#include <ax_inset.h>
#include <atm.h>
#include <axtext.h>
#include <atxint.h>
#include <app_type.h>
#include <bldrbltn.h>


/********************************************/
/* ProtoTypes (External), Data & Data Types */
/********************************************/

extern int grRenderAssociatedGraphic AXP_((long w, data_ptr objP, data_ptr infP));
extern grMeasureInsetObject AXP_((data_ptr objP, data_ptr *unitP, data_ptr *areaP));
extern void *AxGetSSTaskInfo AXP_((int pid));
extern GPHOBJ *GetObjFromUID AXP_((ss_value *ssP, int uid, int appid));

extern data_ptr ElfRetData; 

/* note that if any fields added to this strcture, and if those fields
   contain allocated data not pointed to by anything else, then that
   data must be de-alloc'd in the "bldrFreeEntry" procedure */

typedef struct {
  int uid;                 /* inset handle/instance - unique identifier */
  int pid;                 /* process id of container application */
  int container_taskid;    /* task id of container application */
  int taskid;              /* task id of builder inset, if running */
  data_ptr gfx_objP;       /* graphics inset object used for inset display */
  data_ptr gfx_inst_uid;   /* uid of graphics inset object */
  data_ptr appobj;         /* application object of this builder inset */
  int task_just_started;   /* flag indicating whether bldr inset is starting */
  int container_app;       /* type of application which contains this inset */
  void *container_app_data;/* data regarding exact position of inset
			      in container application -- see
			      bldrbltn.h for possible datatypes */
  char *root_pathname;     /* filename of the file that supplied the
			      current contents of the inset */
  data_ptr builder_docinfo; /* contains a copy of the */
			    /* "builder_docinfo@" structure created */
			    /* during the reading of the builder file */
			    /* -- used for subsequent writes of the */
			    /* file */
} handle_table_entry;

static handle_table_entry *htable = (handle_table_entry *)0;
static int n_htentries = 0;
static int ptr_uid = 1;

/**********************************************/
/*   Prototypes for Procedures in This File   */
/* which are Not in the Corresponding .h File */
/**********************************************/

typeErr ElfbCreateBuilder AXP_((data_ptr aP, data_ptr jnk1P, data_ptr jnk2P, data_ptr jnk3P, data_ptr jnk4P, data_ptr jnk5P, data_ptr jnk6P));
typeErr CreateBldrInstance AXP_((int *bldr_inst_ptr, int container_taskid, int pid, data_ptr gfx_objP, data_ptr gfx_inst_uid, int cont_apptype));
void bldrFreeEntry AXP_((int slot_ix));
data_ptr AxfBuilderDestroyInset AXP_((data_ptr argDatum));
int bldrRetrieveSlotIndex AXP_((int uid));
data_ptr AxfBuilderGetInsetAppobj AXP_((data_ptr argDatum));
data_ptr AxfBuilderGetGraphicsInset AXP_((data_ptr argDatum));
void bldrExtractUnits AXP_((data_ptr unitsP, ax_units *units));
data_ptr AxfBuilderRegisterAppobjWithInset AXP_((data_ptr argDatum));
data_ptr bldrMakeObjData AXP_((ax_units *units_ptr, data_ptr uidP));
data_ptr AxfBuilderGetGraphicsInsetUID AXP_((data_ptr argDatum));
data_ptr AxfBuilderFindInstanceGivenFilenameAndParentTaskid AXP_((data_ptr argDatum));
int bldrIntInList AXP_((int *intList, int listSize, int val));
int *bldrAddToIntList AXP_((int *intList, int listSize, int newint));
data_ptr AxfBuilderGetContainerInfo AXP_((data_ptr argDatum));
data_ptr AxfBuilderGetContainerApptype AXP_((data_ptr argDatum));
data_ptr AxfBuilderSetSourceInfo AXP_((data_ptr argDatum));
data_ptr AxfBuilderGetSourceInfo AXP_((data_ptr argDatum));
data_ptr *AxFindBuilderInsetObjectAndPID AXP_((data_ptr argDatum, int *pid));
data_ptr *AxFindBldrInsetObjFromApptypePIDandUID AXP_((int cont_app, int cont_app_pid, int bldr_inst));
void *bldrFindContextInfoFromApptypePIDAndInstance AXP_((int cont_app, int cont_app_pid, int bldr_inst));
data_ptr AxfBuilderSetInsetDocInfo AXP_((data_ptr argDatum));
data_ptr AxfBuilderGetInsetDocInfo AXP_((data_ptr argDatum));

/*******************************************************/
/* Procedures for Creating Instances of Builder Insets */
/*******************************************************/

/************************************************************************
 * CreateBldrInstance - this procedure adds a builder instance
 *     to the instance table, returns in the "bldr_inst_ptr" out
 *     parameter the instance handle (an integer), and returns as the
 *     value of the procedure an error code indicating if the
 *     operation was successful or not; note that the "appobj" field
 *     of the table entry is not filled in, as we haven't yet read a
 *     particular builder file into the inset, and hence have no
 *     application object to instantiate
 */
typeErr
CreateBldrInstance(bldr_inst_ptr, container_taskid, pid, gfx_objP,
		   gfx_inst_uid, cont_apptype) 
     int *bldr_inst_ptr;    /* OUT */
     int container_taskid;  /* IN */
     int pid;               /* IN */
     data_ptr gfx_objP;     /* IN */
     data_ptr gfx_inst_uid; /* IN */
     int cont_apptype;      /* IN */
{
  int uid, size;
  handle_table_entry *hinst;
  void *c_app_data;

  /* figure out the size of the memory required for the new table */
  size = (n_htentries+1) * sizeof(handle_table_entry);
 
  /* allocate the new table, copying over the data from the old table */
  htable = (handle_table_entry *)TrueRealloc(htable, size);
  if (! htable) {
    return((! AOK));
  }

  /* get a pointer to the current table entry */
  hinst = &(htable[n_htentries++]);

  /* zero out its associated memory */
  memset((char *)hinst, 0, sizeof(handle_table_entry));

  /* set the instance handle (uid) of the new entry to the next
     available instance number, and increment the uid counter */
  hinst->uid = ptr_uid++;

  /* note that we cannot get the application context data yet, because
     (at least in the spreadsheet), the data structure which holds the
     information regarding the placement of the inset does not have
     the relevant information in at at the time the inset is created */

  /* define the rest of the entry with the given input parameters */
  hinst->container_taskid = container_taskid;
  hinst->pid = pid;
  hinst->gfx_objP = AxTaskCopyData(0,gfx_objP);
  hinst->gfx_inst_uid = AxTaskCopyData(0,gfx_inst_uid);
  hinst->task_just_started = FALSE;
  hinst->container_app = cont_apptype;
  hinst->container_app_data = NULL;
  hinst->root_pathname = NULL;

  /* set up the value of the return (OUT) parameter */
  *bldr_inst_ptr = hinst->uid;

  /* return an error code indicating all's well */
  return(AOK);
} /* CreateBldrInstance */

/************************************************************************
 * ElfbCreateBuilder - creates a new builder instance in the builder
 *       instance handle table; returns the builder instance handle
 *       (an integer); for further info, see the header comment
 *        section of the "CreateBldrInstance" procedure
 */
typeErr
ElfbCreateBuilder(aP, jnk1P, jnk2P, jnk3P, jnk4P, jnk5P, jnk6P)
     data_ptr aP, jnk1P, jnk2P, jnk3P, jnk4P, jnk5P, jnk6P; /* IN */
{
  int container_taskid, pid, err, slot_ix, bldr_inst, cont_apptype;
  data_ptr gfx_objP, container_taskidP, gfx_inst_uid;

  /* get the pid:  first by checking if we can get it from the input
     taskid argument -- otherwise by using the THIMpid */
  if ( !aP || !AxIsArray(aP) || AxArraySize(aP) < 1 ) {
    pid = ThimTopPidFromPid(THIMpid);
  } else {
    /* extract the pid (check container_taskidP, taskMaster, then task)  */
    container_taskidP = AxArrayElement(aP, 0);
    if (! container_taskidP) {
      pid = ThimTopPidFromPid(THIMpid);
    } else {
      container_taskid = AxIntFromDataPtr(container_taskidP);
      pid = ThimPidFromUID(container_taskid);
    }
  }

  /* get the graphic inset object & the graphic inset instance pointer
     for the builder icon graphic */ 
  gfx_objP = AxArrayElement(aP,1);
  gfx_inst_uid = AxArrayElement(aP,2);

  /* get the type of the application which contains this inset */
  cont_apptype = AxIntFromArray(aP,3);

  /* create and init a builder instance */
  err = CreateBldrInstance(&bldr_inst, container_taskid, pid,
			   gfx_objP, gfx_inst_uid, cont_apptype);
  if (err != AOK) {
    return((! AOK));
  }

  /* make sure the return value is the integer representing the
     builder instance handle */
  ElfRetData = AxMakeIntData(bldr_inst);

  /* done */
  return(AOK);
} /* ElfbCreateBuilder */

/*********************************************************/
/* Procedures for Destroying Instances of Builder Insets */
/*********************************************************/

/****************************************************************************
 * AxFreeBldrByPID - when a container process, such as the spreadsheet
 *   terminates, this procedure is called, so that the insets in that
 *   spreadsheet can be terminated also
 */

void
AxFreeBldrByPID (pid)
     int pid; /* IN */
{
  handle_table_entry *hinst;
  data_ptr bldr_instP, retData = NULL;
  typeErr err;
  int ix, destroyed_inset;

 do_again:
  destroyed_inset = FALSE;
  for (ix=n_htentries-1; (ix >= 0); ix--) {
    hinst = &(htable[ix]);
    if (hinst && (hinst->pid == pid)) {
      bldr_instP = AxMakeIntData(hinst->uid);
      err = ElfCallByName(NXLT("destroy_builder@"), &retData, 1,
			  bldr_instP);
      AxFreeData(bldr_instP);
      AxFreeData(retData);
      /* note that this macro call has changed the contents of htable,
         so we need to start over */
      destroyed_inset = TRUE;
      break;
    }
  }
  if (destroyed_inset) {
    goto do_again;
  }
} /* AxFreeBldrByPID */

/****************************************************************************
 * bldrFreeEntry - removes an entry from the builder inset instance
 *    table; note this procedure assumes that the application object
 *    pointed to by the "appobj" field in the table entry has already
 *    been terminated and destroyed, and that the graphics inset
 *    object pointed to by the "gfx_objP" field in the table entry has
 *    already been destroyed -- both these actions take place in the
 *    "destroy_builder@" macro in the Elf layer of the code
 */
void
bldrFreeEntry(slot_ix)
     int slot_ix;  /* IN */
{
  int i, size;
  handle_table_entry *hinst1, *hinst2;

  /* find the slot we wish to delete */
  hinst1 = &(htable[slot_ix]);

  /* free up the data_ptr associated with the copied application
     object and the graphics inset which represents it; free up any
     other memory allocated for this instance */ 
  AxFreeData(hinst1->appobj);
  AxFreeData(hinst1->gfx_objP);
  AxFreeData(hinst1->gfx_inst_uid);
  AxFreeData(hinst1->builder_docinfo);
  TrueFree(hinst1->container_app_data);
  TrueFree(hinst1->root_pathname);
  

  /* copy the rest of the table above the current slot into the
     current slot */
  size = sizeof(handle_table_entry);
  for (i=slot_ix+1; (i < n_htentries); i++) {
    hinst2 = &(htable[i]);
    memcpy((void *)hinst1, (void *)hinst2, size);
    hinst1 = hinst2;
  }

  size = (--n_htentries) * sizeof(handle_table_entry);
  if (size > 0) {
    htable = (handle_table_entry *)TrueRealloc(htable, size);
  } else {
    TrueFree(htable);
    htable = (handle_table_entry *)0;
  }
} /* bldrFreeEntry */

/************************************************************************
 * AxfBuilderDestroyInset - destroys the table entry data associated
 *     with the given builder inset instance
 */
data_ptr
AxfBuilderDestroyInset(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to builder_register_appobj_with_inset$ */
{
  data_ptr appobj;
  int bldr_inst, slot_ix;

  /* get the actual arguments out of the argument list */
  bldr_inst = AxIntFromArray(argDatum,0); /* IN */

  /* find the correct entry in the inset instance table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return(AxMakeIntData((! AOK)));
  }
  bldrFreeEntry(slot_ix);

  /* return "AOK" flag */
  return(AxMakeIntData(AOK));
} /* AxfBuilderDestroyInset */

/******************************************************/
/* Procedures for Finding Instances of Builder Insets */
/******************************************************/

/************************************************************************
 * bldrRetrieveSlotIndex - retrieves the index into the table of
 *      handle at which the data for the handle corresponding to the
 *      given uid resides.  returns -1 if there is no such uid present
 *      in the table
 */
int
bldrRetrieveSlotIndex (uid)
     int uid; /* IN */
{
  int slot_ix = -1;
  int i;

  for (i=0; (i < n_htentries); i++) {
    if (htable[i].uid == uid) {
      slot_ix = i;
      break;
    }
  }
  return(slot_ix);
} /* bldrRetrieveSlotIndex */

/************************************************************************
 * AxfBuilderFindInstanceGivenFilenameAndParentTaskid - given a
 *    filename and the taskid of the parent (container) application,
 *    this procedure tries to find the corresponding builder inset
 *    instance handle.  Note that if it cannot find it, it returns -1.
 */
data_ptr
AxfBuilderFindInstanceGivenFilenameAndParentTaskid(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to builder_register_appobj_with_inset$ */
{
  char *rootpname;
  int taskid, ix, val;
  int bldr_inst = -1;
  int slot_ix = -1;
  int *active_list = NULL;
  int *pname_list = NULL;
  int *task_list = NULL;
  int active_size = 0;
  int pname_size = 0;
  int task_size = 0;

  /* get the actual arguments out of the argument list */
  rootpname = AxStrFromArray(argDatum, 0); /* IN */
  taskid = AxIntFromArray(argDatum, 1);    /* IN */

  /* find out which slots in the builder inset table contain insets
     that match the given container taskid, that match the given root
     pathname, or that are marked as just started */
  for (ix=0; (ix < n_htentries); ix++) {
    if (taskid == htable[ix].container_taskid) {
      task_list = bldrAddToIntList(task_list, task_size++, ix);
    }
    if (strcmp(rootpname, htable[ix].root_pathname) == 0) {
      pname_list = bldrAddToIntList(pname_list, pname_size++, ix);
    }
    if (htable[ix].task_just_started) {
      active_list = bldrAddToIntList(active_list, active_size++, ix);
    }
  }

  /* try to find a builder inset instance on all three lists; in most
     cases, the "active_list" should be the shortest -- in fact, it
     should in most cases have size == 1 */ 
  for (ix=0; (ix < active_size); ix++) {
    val = active_list[ix];
    if (bldrIntInList(pname_list, pname_size, val) &&
	bldrIntInList(task_list, task_size, val)) {
      slot_ix = val;
      goto do_return;
    }
  }
  /* next, try to find something with the same filename and that's
     just been marked as activated */
  for (ix=0; (ix < pname_size); ix++) {
    val = pname_list[ix];
    if (bldrIntInList(active_list, active_size, val)) {
      slot_ix = val;
      goto do_return;
    }
  }
  /* next, try finding a single "activated" inset */ 
  if (1 == active_size) {
    slot_ix = *active_list;
    goto do_return;
  }
  /* next, try to find something with the same filename and the same
     parent task */
  for (ix=0; (ix < pname_size); ix++) {
    val = pname_list[ix];
    if (bldrIntInList(task_list, task_size, val)) {
      slot_ix = val;
      goto do_return;
    }
  }
  /* next, try just something with the same filename */
  if (pname_size > 0) {
    slot_ix = *pname_list;
    goto do_return;
  }
  /* if we can't get it by filename, try something in common with the 
     other two, remaining lists */
  for (ix=0; (ix < active_size); ix++) {
    val = active_list[ix];
    if (bldrIntInList(task_list, task_size, val)) {
      slot_ix = val;
      goto do_return;
    }
  }
  /* if there's nothing activated, or more than one, try to see if
     there's only one inset associated with the given task */
  if (1 == task_size) {
    slot_ix = *task_list;
    goto do_return;
  }

  /* at this point, we know the following:
     1.  Either no task has been marked as just started, or more than
         one has.
     2.  We have no inset that matches the filename as given.
     3.  Either the given task has no insets associated in the table,
         or it has more than one.
     4.  We may have some marked-active insets and some insets of the
         current task, but these two lists do not intersect.
     At this point, I think we can safely throw up our hands and say,
        "Sorry, I don't understand."
  */

 do_return:
  if (-1 != slot_ix) {
    /* reset the "just started" flag to FALSE */
    htable[slot_ix].task_just_started = FALSE;

    /* return the unique identifier (builder instance) */
    bldr_inst = htable[slot_ix].uid;
  }
  return(AxMakeIntData(bldr_inst));
} /* AxfBuilderFindInstanceGivenFilenameAndParentTaskid */


/************************************************************************
 * bldrIntInList - auxillary procedure for
 *       AxfBuilderFindInstanceGivenFilenameAndParentTaskid 
 */
int
bldrIntInList(intList, listSize, val)
     int *intList;
     int listSize;
     int val;
{
  int ix;
  int *tmp = intList;

  for (ix=0; (ix < listSize); ix++) {
    if (*tmp == val) {
      return(TRUE);
    }
  }
  return(FALSE);
} /* bldrIntInList */

/************************************************************************
 * bldrAddToIntList - auxillary procedure for
 *       AxfBuilderFindInstanceGivenFilenameAndParentTaskid 
 */
int *
bldrAddToIntList(intList, listSize, newint)
     int *intList;
     int listSize;
     int newint;
{
  int *tmp;
  int size;

  if (0 == listSize) {
    tmp = (int *)TrueMalloc(sizeof(int));
    *tmp = newint;
  } else {
    size = (listSize + 1) * sizeof(int);
    tmp = (int *)TrueRealloc(intList, size);
    tmp[listSize] = newint;
  }
  return(tmp);
} /* bldrAddToIntList */

/****************************************************************/
/* Procedures for Retrieving Pieces of Info from Builder Insets */
/****************************************************************/

/************************************************************************
 * AxfBuilderGetInsetDocInfo - returns a copy of the "builder_docinfo@"
 *      structure associated with the given builder inset instance
 */
data_ptr
AxfBuilderGetInsetDocInfo(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to this macro */
{
  int bldr_inst, slot_ix;
  data_ptr docinfo;

  /* get the actual arguments out of the argument list */
  bldr_inst = AxIntFromArray(argDatum,0); /* IN */

  /* find the correct entry in the handle table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return((data_ptr)0);
  }

  /* get the application object pointer and return it */
  docinfo = htable[slot_ix].builder_docinfo;
  return(AxCopyData(docinfo));
} /* AxfBuilderGetInsetDocInfo */

/************************************************************************
 * AxfBuilderGetInsetAppobj - returns the application object
 *      associated with the given builder inset instance
 */
data_ptr
AxfBuilderGetInsetAppobj(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to this macro */
{
  int bldr_inst, slot_ix;
  data_ptr appobj;

  /* get the actual arguments out of the argument list */
  bldr_inst = AxIntFromArray(argDatum,0); /* IN */

  /* find the correct entry in the handle table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return((data_ptr)0);
  }

  /* get the application object pointer and return it */
  appobj = htable[slot_ix].appobj;
  return(AxCopyData(appobj));
} /* AxfBuilderGetInsetAppobj */

/************************************************************************
 * AxfBuilderGetContainerInfo - returns the information about the
 *     context of the builder inset in the containing application 
 */
data_ptr
AxfBuilderGetContainerInfo(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to builder_register_appobj_with_inset$ */
{
  int bldr_inst, slot_ix;
  data_ptr container_infoP;
  handle_table_entry *hinst;
  bldr_container_info_for_ss *cont_info;

  /* get the actual arguments out of the argument list */
  bldr_inst = AxIntFromArray(argDatum,0); /* IN */

  /* find the correct entry in the handle table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return((data_ptr)0);
  }
  hinst = &(htable[slot_ix]);

  switch (hinst->container_app) {
  case AX_APP_TYPE_SPREADSHEET:
    cont_info = (bldr_container_info_for_ss *)
      hinst->container_app_data;
    if (NULL == cont_info) {
      cont_info = (bldr_container_info_for_ss *)
	bldrFindContextInfoFromApptypePIDAndInstance( 
	     AX_APP_TYPE_SPREADSHEET, hinst->pid, bldr_inst);
    }
    if (NULL != cont_info) {
      container_infoP = AxMakeArray(6);
      container_infoP = AxAddIntToArray(container_infoP, 0,
					cont_info->tl_row); 
      container_infoP = AxAddIntToArray(container_infoP, 1,
					cont_info->br_row); 
      container_infoP = AxAddIntToArray(container_infoP, 2,
					cont_info->tl_col); 
      container_infoP = AxAddIntToArray(container_infoP, 3,
					cont_info->br_col); 
      container_infoP = AxAddIntToArray(container_infoP, 4,
					cont_info->tl_plate); 
      container_infoP = AxAddIntToArray(container_infoP, 5,
					cont_info->br_plate); 
    } else {
      container_infoP = (data_ptr)0;
    }
    break;
  default:
    container_infoP = (data_ptr)0;
    break;
  }
  return(container_infoP);
} /* AxfBuilderGetContainerInfo */

/************************************************************************
 * bldrFindContextInfoFromApptypePIDAndInstance - auxillary procedure
 *   for AxfBuilderGetContainerInfo; given the containing
 *   application's type and process ID, the the builder inset
 *   instance's unique id, this procedure finds, constructs and
 *   returns a pointer to the "context information" -- a data
 *   structure which tells *where* in the containing application this
 *   inset is; the structure of this information varies with the type
 *   of containing application 
 */
void *
bldrFindContextInfoFromApptypePIDAndInstance(cont_app, cont_app_pid,
					     bldr_inst) 
     int cont_app;     /* IN */
     int cont_app_pid; /* IN */
     int bldr_inst;    /* IN */
{
  data_ptr *inset_obj;
  GPHOBJ *obj;
  ss_value *ssP;
  bldr_container_info_for_ss *c_app_data = NULL;
  int size;

  /* each type of application which can contain a builder inset
     instance stores the data regarding that inset object in a
     different place */
  switch(cont_app) {
  case AX_APP_TYPE_SPREADSHEET:
    ssP = AxGetSSTaskInfo(cont_app_pid);
    obj = GetObjFromUID(ssP, bldr_inst, AX_APP_TYPE_BUILDER);
    if (obj) {
      size = sizeof(bldr_container_info_for_ss);
      c_app_data = (bldr_container_info_for_ss *)TrueMalloc(size);
      memset((char *)c_app_data, 0, (size_t)size);
      c_app_data->tl_row = (obj->rng).tl_row;
      c_app_data->br_row = (obj->rng).br_row;
      c_app_data->tl_col = (obj->rng).tl_col;
      c_app_data->br_col = (obj->rng).br_col;
      c_app_data->tl_plate = (obj->rng).tl_plate;
      c_app_data->br_plate = (obj->rng).br_plate;
    }
    break;
  default:
    break;
  }
  return((void *)c_app_data);
} /* bldrFindContextInfoFromApptypePIDAndInstance */

/************************************************************************
 * AxfBuilderGetContainerApptype - returns the application type of the
 *      application which is the containing application for the
 *      specified builder inset
 */
data_ptr
AxfBuilderGetContainerApptype(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to builder_register_appobj_with_inset$ */
{
  int bldr_inst, slot_ix;
  data_ptr container_info;
  handle_table_entry *hinst;
  bldr_container_info_for_ss *cont_info;
  int container_apptype;

  /* get the actual arguments out of the argument list */
  bldr_inst = AxIntFromArray(argDatum,0); /* IN */

  /* find the correct entry in the handle table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return((data_ptr)0);
  }
  hinst = &(htable[slot_ix]);

  container_apptype = hinst->container_app;
  return(AxMakeIntData(container_apptype));
} /* AxfBuilderGetContainerApptype */

/************************************************************************
 * AxfBuilderGetGraphicsInset - returns the graphics inset object
 *       associated with the given builder inset instance
 */
data_ptr
AxfBuilderGetGraphicsInset(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to builder_register_appobj_with_inset$ */
{
  data_ptr gfx_objP;
  int bldr_inst, slot_ix;

  /* get the actual arguments out of the argument list */
  bldr_inst = AxIntFromArray(argDatum,0); /* IN */

  /* find the correct entry in the inset instance table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return((data_ptr)0);
  }

  /* get the graphics inset data pointer and return it */
  gfx_objP = htable[slot_ix].gfx_objP;
  return(AxCopyData(gfx_objP));
} /* AxfBuilderGetGraphicsInset */

/************************************************************************
 * AxfBuilderGetGraphicsInsetUID - get the UID of the graphics inset
 *      object associated with the given builder inset instance
 */
data_ptr
AxfBuilderGetGraphicsInsetUID(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to builder_register_appobj_with_inset$ */
{
  data_ptr gfx_inst_uid;
  int bldr_inst, slot_ix;

  /* get the actual arguments out of the argument list */
  bldr_inst = AxIntFromArray(argDatum,0); /* IN */

  /* find the correct entry in the inset instance table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return((data_ptr)0);
  }

  /* get the graphics inset UID and return it */
  gfx_inst_uid = htable[slot_ix].gfx_inst_uid;
  return(AxCopyData(gfx_inst_uid));
} /* AxfBuilderGetGraphicsInsetUID */

/************************************************************************
 * AxfBuilderSetRootPathname - gets the root pathname of the inset's
 *    corresponding file into a character string, and stores that away
 *    in the builder inset instance corresponding to "bldr_inst"
 */
data_ptr
AxfBuilderSetRootPathname(argDatum)
     data_ptr argDatum;
{
  int bldr_inst, slot_ix, size;
  char *temp, *rootpname;
  handle_table_entry *hinst;

  /* get the actual arguments out of the argument list */
  bldr_inst = AxIntFromArray(argDatum,0); /* IN */
  temp = AxStrFromArray(argDatum,1); /* IN */

  /* find the correct entry in the inset instance table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return(AxMakeIntData((! AOK)));
  }
  hinst = &(htable[slot_ix]);

  /* if we've already got a pathname here, get rid of it */
  if (NULL != hinst->root_pathname) {
    TrueFree(hinst->root_pathname);
    hinst->root_pathname = NULL;
  }

  /* if we've said just to NULL out the field, then we're done */
  if (NULL == temp) {
    return(AxMakeIntData(AOK));
  }

  /* temp is in volatile memory, so copy it into something we control */
  size = strlen(temp) + 1;
  rootpname = (char *)TrueMalloc(size);
  if (NULL == rootpname) {
    /* memory allocation failure */
    return(AxMakeIntData((! AOK)));
  }
  strcpy(rootpname, temp);

  /* assign it to the correct place in the data structure and return */
  hinst->root_pathname = rootpname;
  return(AxMakeIntData(AOK));
} /* AxfBuilderSetRootPathname */

/********************************************/
/* Procedures for Mucking Around with Units */
/********************************************/

/************************************************************************
 * bldrExtractUnits - takes the data provided in the "unitsP" input
 *    parameter, and if it is valid/readable, puts it in the "units"
 *    output paramter; otherwise, it supplies default values to the
 *    output parameter.  This is an auxillary procedure for
 *    bldrPackageInsetObject.
 */
void
bldrExtractUnits(unitsP, units)
     data_ptr unitsP;  /* IN */
     ax_units *units;  /* OUT */
{
  int n;
  data_ptr valP;

  /* something of a hack here, to figure out how much of the input
     "unitsP" array we can read */
  n = ((! unitsP) || (! AxIsArray(unitsP))) ? 0 : AxArraySize(unitsP);

  /* do first argument, the dpi value */
  if (n < 1) {
    units->dpi = 1000;
  } else {
    valP = AxArrayElement(unitsP, 0);
    units->dpi = valP ? AxIntFromDataPtr(valP) : 1000;
  }

  /* do second argument, the dpu value */
  if (n < 2) {
    units->dpu = 1;
  } else {
    valP = AxArrayElement(unitsP, 1);
    units->dpu = valP ? AxIntFromDataPtr(valP) : 1;
  }

  /* do third argument, the precision value */
  if (n < 3) {
    units->precision = 0;
  } else {
    valP = AxArrayElement(unitsP, 2);
    units->precision = valP ? AxIntFromDataPtr(valP) : 0;
  }

  /* do fourth argument, the unit name value */
  if (n < 4) {
    units->name = NXLT("mils");
  } else {
    valP = AxArrayElement(unitsP, 3);
    units->name = valP ? AxStrPtrFromDataPtr(valP) : (mbchar *)(NXLT("mils"));
  }
} /* bldrExtractUnits */

/**************************************************/
/* Procedures for Setting the Values of Pieces of */
/*       Particular Builder Inset Instances       */
/**************************************************/

/************************************************************************
 * AxfBuilderSetInsetDocInfo - copies the given "builder_docinfo@"
 *     data structure into the given builder inset instance data
 *     structure 
 */
data_ptr
AxfBuilderSetInsetDocInfo(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to this macro call */
{
  data_ptr bldr_docinfo;
  int bldr_inst, slot_ix;

  /* get the actual arguments out of the argument list */
  bldr_inst = AxIntFromArray(argDatum,0);    /* IN */
  bldr_docinfo = AxArrayElement(argDatum,1); /* IN */

  /* find the correct entry in the inset instance table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return(AxMakeIntData((! AOK)));
  }

  /* do the assignment into the proper field */
  htable[slot_ix].builder_docinfo = AxTaskCopyData(0, bldr_docinfo);

  /* return done */
  return(AxMakeIntData(AOK));
} /* AxfBuilderSetInsetDocInfo */

/************************************************************************
 * AxfBuilderRegisterAppobjWithInset - the argument list here is
 *    composed of three values -- the first of which is the builder
 *    instance handle number, the second of which is an application
 *    object, and the third of which is the taskid of the application
 *    object if it is running, and -1 if it is not.  This procedure
 *    puts the application object and the taskid in the inset instance
 *    table at the entry associated with the given instance/uid.
 */
data_ptr
AxfBuilderRegisterAppobjWithInset(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to builder_register_appobj_with_inset$ */
{
  data_ptr appobj;
  int bldr_inst, slot_ix, bldr_taskid;

  /* get the actual arguments out of the argument list */
  bldr_inst = AxIntFromArray(argDatum,0);    /* IN */
  appobj = AxArrayElement(argDatum,1);       /* IN */
  bldr_taskid = AxIntFromArray(argDatum,2);  /* IN */

  /* find the correct entry in the inset instance table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return(AxMakeIntData((! AOK)));
  }

  /* do the assignments into the proper fields */
  htable[slot_ix].appobj = AxTaskCopyData(0,appobj);
  htable[slot_ix].taskid = bldr_taskid;

  /* return done */
  return(AxMakeIntData(AOK));
} /* AxfBuilderRegisterAppobjWithInset */

/************************************************************************
 * bldrSetContainerInfo - takes an application type from <app_types.h>
 *    and data (whose format depends upon the type of application (see
 *    bldrbltn.h for possible container data structures)), and stores
 *    this away in the inset
 *
 * IMPORTANT NOTE:  This procedure assumes that what comes into it in
 * the "c_app_data" parameter has been copied by the *caller* using
 * TrueMalloc (or its equivalent, Task-0 memory allocation), and
 * hence, that the builder insetting code is at liberty to deallocate
 * it upon inset destruction using "TrueFree".
 */
typeErr
bldrSetContainerInfo(bldr_inst, c_app, c_app_data)
     int bldr_inst;     /* IN */
     int c_app;         /* IN */
     void *c_app_data;  /* IN */
{
  int slot_ix;
  handle_table_entry *hinst;
  data_ptr bldr_instP, retData;
  typeErr err;

  /* find the correct entry in the inset instance table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return((! AOK));
  }
  hinst =  &(htable[slot_ix]);

  /* set the appropriate fields of the builder instance */
  hinst->container_app = c_app;
  /* check first to make sure that if we're overwriting valid data, we 
     clean up the memory allocation first */
  if (NULL != hinst->container_app_data) {
    TrueFree(hinst->container_app_data);
  }
  hinst->container_app_data = c_app_data;

  /* tell the application object that it needs to update its context
     info */
  bldr_instP = AxMakeIntData(hinst->uid);
  err = ElfCallByName(NXLT("builder_inset_update_context_info$"),
		      &retData, 1, bldr_instP);
  AxFreeData(bldr_instP);
  AxFreeData(retData);

  return(AOK);
} /* bldrSetContainerInfo */

/************************************************************************
 * bldrSetTaskJustStartedFlag - sets a flag in the builder inset
 *   instance that indicates that the inset has just been started up
 */
typeErr
bldrSetTaskJustStartedFlag(bldr_inst, flag_val)
     int bldr_inst; /* IN */
     int flag_val;  /* IN */
{
  int slot_ix;
  handle_table_entry *hinst;

  /* find the correct entry in the inset instance table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return((! AOK));
  }
  hinst = &(htable[slot_ix]);

  /* set the appropriate fields of the builder instance */
  hinst->task_just_started = flag_val;

  return(AOK);
} /* bldrSetTaskJustStartedFlag */

/***************************************************************/
/* Procedures for Mucking Around with the ax_inset_object-type */
/*           Objects that Enclose the Actual Insets            */
/***************************************************************/

/************************************************************************
 * AxfBuilderSetSourceInfo - changes the "source" information
 *     associated with the given builder inset instance; first
 *     argument is the builder inset instance UID, the second is an
 *     ElfData array representing the source info (in whatever format
 *     the user desires) to be saved away with the inset
 */
data_ptr
AxfBuilderSetSourceInfo(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to builder_register_appobj_with_inset$ */
{
  int pid;
  data_ptr src_info;
  data_ptr *inset_objP;

  /* find the location in the current data structures for whatever
     application we're in where the definition of the builder
     application inset object resides; pid is used here so that we'll
     knkow what task to allocated memory in when we allocate and store
     the source information */
  inset_objP = AxFindBuilderInsetObjectAndPID(argDatum, &pid);
  if ((data_ptr *)0 == inset_objP) {
    return(AxMakeIntData((! AOK)));
  }

  /* get the other argument out of the argument list */
  src_info = AxArrayElement(argDatum,1);     /* IN */

  /* put the source information provided by the user in the
     correct place in the ax_inset_object@ format object, and assign
     the object back to where it was */ 
  *inset_objP = AxAddArrayToArray(*inset_objP, 2, AxTaskCopyData(pid,src_info));

  /* return done */
  return(AxMakeIntData(AOK));
} /* AxfBuilderSetSourceInfo */

/************************************************************************
 * AxfBuilderGetSourceInfo - retrieves the "source" information
 *     associated with the given builder inset; first argument is the
 *     builder inset instance UID
 */
data_ptr
AxfBuilderGetSourceInfo(argDatum)
     data_ptr argDatum;   /* an array that constitutes the argument */
			  /* list to builder_register_appobj_with_inset$ */
{
  int pid;
  data_ptr src_info;
  data_ptr *inset_objP;

  /* find the location in the current data structures for whatever
     application we're in where the definition of the builder
     application inset object resides. (Note that the pid is unused
     here */ 
  inset_objP = AxFindBuilderInsetObjectAndPID(argDatum, &pid);
  if ((data_ptr *)0 == inset_objP) {
    return((data_ptr) 0);
  }

  /* get the source information out of the ax_inset_object@ format */
  src_info = AxArrayElement(*inset_objP, 2);

  /* return the source information */
  return(AxCopyData(src_info));
} /* AxfBuilderGetSourceInfo */

/************************************************************************
 * AxFindBuilderInsetObjectAndPID - finds the location in the current
 *    data structures for whatever application we're in where the
 *    def'n of the builder application inset object resides; auxillary 
 *    procedure for AxfBuilder{Set,Get}SourceInfo
 */
data_ptr *
AxFindBuilderInsetObjectAndPID(argDatum, pid)
     data_ptr argDatum;  /* IN */
     int *pid;           /* OUT */
{
  data_ptr *inset_objP;
  int bldr_inst, slot_ix, cont_app, cont_app_pid;
  handle_table_entry *hinst;

  /* get the actual arguments out of the argument list */
  bldr_inst = AxIntFromArray(argDatum,0);    /* IN */

  /* find the correct entry in the inset instance table */
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return((data_ptr *)0);
  }
  hinst = &(htable[slot_ix]);
  cont_app = hinst->container_app;
  cont_app_pid = hinst->pid;

  inset_objP = AxFindBldrInsetObjFromApptypePIDandUID(cont_app,
			      cont_app_pid, bldr_inst);
  *pid = cont_app_pid;
  return(inset_objP);
} /* AxFindBuilderInsetObjectAndPID */

/************************************************************************
 * AxFindBldrInsetObjFromApptypePIDandUID - auxillary procedure for
 *       AxFindBuilderInsetObjectAndPID; given the type of application in
 *       which the builder inset instance is contained and its process
 *       id, and the unique id which represents that builder inset
 *       instance, this function returns a pointer to the location
 *       (wherever each individual application may choose to store it)
 *       of the ax_inset_object@ thing the application is storing as
 *       representing that inset object.
 */
data_ptr *
AxFindBldrInsetObjFromApptypePIDandUID(cont_app, cont_app_pid, bldr_inst)
     int cont_app;     /* IN */
     int cont_app_pid; /* IN */
     int bldr_inst;    /* IN */
{
  data_ptr *inset_obj;
  GPHOBJ *obj;
  ss_value *ssP;

  /* each type of application which can contain a builder inset
     instance stores the data regarding that inset object in a
     different place */
  switch(cont_app) {
  case AX_APP_TYPE_SPREADSHEET:
    ssP = AxGetSSTaskInfo(cont_app_pid);
    obj = GetObjFromUID(ssP, bldr_inst, AX_APP_TYPE_BUILDER);
    inset_obj = &(obj->axInsetObjP);
    break;
  default:
    inset_obj = (data_ptr *)0;
    break;
  }
  return(inset_obj);
} /* AxFindBldrInsetObjFromApptypePIDandUID */

/************************************************************************
 * bldrMakeObjData - packages up the input data into an inset object
 *     whose ElfData form matches that of the "ax_inset_object@"
 *     structure.  Auxillary procedure for bldrPackageInsetObject.
 */
data_ptr
bldrMakeObjData(units_ptr, uidP, srcP)
     ax_units *units_ptr;  /* IN */
     data_ptr uidP;        /* IN */
     data_ptr srcP;        /* IN */
{
  data_ptr objP, valP, unitsP;

  /* do unit data first */
  unitsP = AxMakeArray(4);
  AxAddIntToArray(unitsP, 0, units_ptr->dpi);
  AxAddIntToArray(unitsP, 1, units_ptr->dpu);
  AxAddIntToArray(unitsP, 2, units_ptr->precision);
  AxAddStrToArray(unitsP, 3, units_ptr->name);

  /* since we have no x, y, wid or hyt data for a builder inset, these
     are just place-holders that are left empty/blank */
/*
  valP = AxMakeArray(4);
*/

  /* the second argument here is just an empty/blank place-holder as
     well */ 
/*
  srcP = AxMakeArray(2);
  AxAddArrayToArray(srcP, 0, valP);
*/

  /* create the object package */
  objP = AxMakeArray(5);
  AxAddIntToArray  (objP, 0, IC_AB);	              /* builder object type */
  AxAddIntToArray  (objP, 1, AxIntFromDataPtr(uidP)); /* builder instance */
  AxAddArrayToArray(objP, 2, AxCopyData(srcP));       /* source info */
  AxAddArrayToArray(objP, 3, unitsP);                 /* units of measure */
  AxAddArrayToArray(objP, 4, (data_ptr)0);            /* parent array */

  /* return the inset object value */
  return(objP);
} /* bldrMakeObjData */

/************************************************************************
 * bldrPackageInsetObject - checks out the input "unitsP" data, to
 *     calculate the correct units for the inset.  (Actually this is
 *     unnecessary, as the associated graphics inset will be the thing
 *     actually displayed, but other parts of the code depend upon
 *     having reasonable values returned here.)  Then this procedure
 *     packages up the input data into an insetobject whose ElfData
 *     form matches that of the "ax_inset_object@" structure 
 */
data_ptr
bldrPackageInsetObject (uidP, srcP, unitsP)
     data_ptr uidP, srcP, unitsP; /* IN */
{
  data_ptr objP;
  ax_units units;

  /* read the units */
  bldrExtractUnits(unitsP, &units);

  /* create an inset object from the application object and the unit
     information */
  objP = bldrMakeObjData(&units, uidP, srcP);

  /* return the inset object */
  return(objP);
} /* bldrPackageInsetObject */

/*****************************************/
/*       Procedures for Measuring        */
/* the Graphical Size of a Builder Inset */
/*****************************************/

/************************************************************************
 * bldrMeasureInsetObject - this procedure merely pulls out the
 *     builder inset's associated graphic inset, and passes that along
 *     to the appropriate graphics measuring routine -- coding
 *     sleight-of-hand 
 */
int
bldrMeasureInsetObject (objP, unitP, areaP)
     data_ptr objP;        /* IN */
     data_ptr *unitP;      /* IN/OUT */
     data_ptr *areaP;      /* IN/OUT */
{
  data_ptr gfx_objP;
  int bldr_inst, slot_ix;

  /* find the correct entry in the inset instance table */
  bldr_inst = AxIntFromArray(objP,1);
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return(FALSE);
  }

  /* get the *graphic* inset object which is the builder icon */
  gfx_objP = htable[slot_ix].gfx_objP;

  /* measure *that* object */
  grMeasureInsetObject(gfx_objP, unitP, areaP);

  return(TRUE);
} /* bldrMeasureInsetObject */

/*******************************************************************/
/* Procedure for Painting (Graphics) a Builder Inset on the Screen */
/*******************************************************************/

/************************************************************************
 * bldrRenderAssociatedGraphic - this procedure merely pulls out the
 *    builder inset's associated graphic inset, and passes that along
 *    to the appropriate graphics rendering routine -- coding
 *    sleight-of-hand 
 */
void
bldrRenderAssociatedGraphic (w, objP, infP, imP)
     AXWidget w;
     data_ptr objP;
     data_ptr infP;
     data_ptr imP;
{
  data_ptr gfx_objP;
  int bldr_inst, slot_ix;

  /* find the correct entry in the inset instance table */
  bldr_inst = AxIntFromArray(objP,1);
  slot_ix = bldrRetrieveSlotIndex(bldr_inst);
  if (-1 == slot_ix) {
    return;
  }
  gfx_objP = htable[slot_ix].gfx_objP;

  /* render the builder icon */
  grRenderAssociatedGraphic(w, gfx_objP, infP, (data_ptr)0);

} /* bldrRenderAssociatedGraphic */
