/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elprint.c
   Applix ASC version: 23
   Date: Tue May 11 21:11:24 1999
*/
#include <stdio.h>
#include <basicwin.h>
#include <ax_xm.h>
#include <aos.h>
#include <wchars.h>
#include <getstr5.h>
#include <thimble.h>
#include <ctype.h>
#include <axtext.h>
#include <elf_env.h>
#include <eldraw.h>
#include <xmpixel.h>
#include <colors.h>

/*-----------------------------------------------------------------------------
	define
-----------------------------------------------------------------------------*/
#define MBALCSTR(p)			/* Alloc & copy string */	\
		((mbchar *)strcpy((char *)malloc((strlen((char *)p)+1)*sizeof(mbchar)),(char *)p))

/*-----------------------------------------------------------------------------
	global data
-----------------------------------------------------------------------------*/
extern	int		TmAmSE ;
extern	char	*SEhackname ;
extern	char	*SEhackfilename ;
extern	Colormap ApplixCmap;
extern	Display *Dpy;

/*-----------------------------------------------------------------------------
	global procedures
-----------------------------------------------------------------------------*/
extern	typeErr		SEinit();
extern	typeErr		SEterm();
extern	char		*GetProfileByCode() ;
extern	char		*AxTempFile() ;
extern	void		axDrawText() ;
extern	void		*AxGFXFromData();
extern	data_ptr	AxMakeGFXData();
extern	data_ptr	ElfGetSysVar();
extern	pixcache_t	*Pixmapper();

/*-----------------------------------------------------------------------------
	local data
-----------------------------------------------------------------------------*/
static	char		seFileName[STRPATH+1];
static	data_ptr	pen ;
static	textAttrs	tattrs ;		/* text attributes */
static	unsigned int tflags ;
static	attr_type	pattrs ;		/* polyline attributes */
static	int			fill ;
static	poly_type	pski ;			/* in honor of Scott */

static struct {
	int		start_page ;
	int		end_page ;
	int		color ;
	char	*printer ;
	char	*print_file ;
	int		copies ;
	char	*language ;
	int		banner ;
	data_ptr baggage ;
	} doc_data ;


static struct {
	int		height ;
	int		width ;
	char	*orientation ;
	int		lm ;
	int		rm ;
	int		tm ;
	int		bm ;
	} page_data ;


/*-----------------------------------------------------------------------------
	local procedures
-----------------------------------------------------------------------------*/
static	int		init() ;
static	void	newpage() ;
static	void	text_type() ;
static	void	text_pen() ;
static	void	color_pen() ;
static	void	text_draw() ;
static	void	polygon() ;
static	void	poly_pen() ;
static	void	poly_line() ;
static	void	poly_rectangle() ;
static	void	poly_arc() ;
static	void	poly_poly() ;
static  void 	chart() ;
static	void	icon() ;
static	pixcache_t *find_icon() ;
static	void	SetGEunits() ;

/*-----------------------------------------------------------------------------
	ElfbPrint

	Print from an array of material

		stuff[0] is the doc parameters

		stuff[1] ... stuff[n] are a mixture of types:
			PAGE	new page
			TEXT	text
			POLY	polygon
			ICON	icon
			CHART	chart
			
-----------------------------------------------------------------------------*/
void ElfbPrint(stuff)
	data_ptr stuff;
{
	int i, n ;
	data_ptr p, s ;
	data_ptr doc ;
	int err ;
	char err_obj[STRPATH + 1];
	char err_str[STRPATH + 1];
	int type ;
	int page ;

	if (!AxIsArray(stuff))
		return ;

	n = AxArraySize(stuff) ;

	/* the first element contains document parameters */
	doc = AxArrayFromArray(stuff, 0) ;
	doc_data.start_page = AxIntFromArray(doc, 1) ;
	doc_data.end_page = AxIntFromArray(doc, 2) ;
	doc_data.color = AxIntFromArray(doc, 3) ;
	doc_data.printer = AxStrFromArray(doc, 4) ;
	doc_data.print_file = AxStrFromArray(doc, 5) ;
	doc_data.copies = AxIntFromArray(doc, 6) ;
	doc_data.language = AxStrFromArray(doc, 7) ;
	doc_data.banner = AxIntFromArray(doc, 8) ;
	doc_data.baggage = AxArrayFromArray(doc, 9) ;

	init() ;
	page = 0 ;

	for (i = 1 ; i < n ; i++)
	{
		p = AxArrayFromArray(stuff, i) ;
		if (!AxIsArray(p))
			return ;
		type = AxIntFromArray(p, 0) ;
		if (type == PRINT_PAGE_TYPE)
		{
			page++ ;
			if (doc_data.end_page > 0 && page > doc_data.end_page)
				break ;
			else if (page < doc_data.start_page)
				continue ;
		}

		switch (type)
		{
		case PRINT_PAGE_TYPE:
			newpage(p) ;
 			break ;

		case PRINT_TEXT_TYPE:
			text_type(p) ;
			break ;

		case PRINT_POLY_TYPE:
			polygon(p) ;
			break ;

		case PRINT_ICON_TYPE:
			icon(p) ;
			break ;

		case PRINT_CHART_TYPE:
			chart(p) ;
			break ;
		}
	}

	err = SEterm() ;

	if (err == AOK)
		err = AxPrintFinish2(doc_data.printer,
							doc_data.copies,
							doc_data.print_file,
							doc_data.banner,
							"",
							doc_data.language,
							"",
							FALSE,
							doc_data.baggage,
							err_obj,
							err_str);
	/* Reset TmAmSE */
	TmAmSE = FALSE;

	/* if error return, throw Elf error */
	if (err != AOK)
		ElfStrAbort(err, err_str, err_obj) ;
}

/*-----------------------------------------------------------------------------
	init

	initialize
-----------------------------------------------------------------------------*/
static int init()
{
	int err ;
	char *printTempDir;
	char buf[STRPATH+1];

	/* determine directory for serializer file */
	printTempDir = GetProfileByCode(PROF_PRT_TEMP);
	if ( !printTempDir || !printTempDir[0] )
		printTempDir = ThimTempDir;
	AxMakePath(buf, printTempDir, NXLT("bd%%%%%%.se"));

	if (!AxTempFile(buf, seFileName))
	{
#if 0
		strcpy(WP.errorObj, printTempDir);
		StrGet(ErrnoErr(), buf);
		sprintf((char *)WP.errorStr, NXLT("%s (%s)"),
						XLT("Could not create print serializer temp file"), buf);
#endif
		return(ErrnoErr());
	}

	/* initialize serializer */

	err = SEinit(seFileName);
	if (AOK != err)
	{
#if 0
		StrGet(ErrnoErr(), buf);
		sprintf((char *)WP.errorStr, NXLT("%s (%s)"),
						XLT("Could not initialize print serializer"), buf);
#endif
		return(err);
	}

	SEhackname = NXLT("Printer") ;
	SEhackfilename = seFileName;

	/* Set TmAmSE */
	TmAmSE = TRUE;
#if 0
	TmDefDev = WP.printerDevice;
#endif

	PFSetColordev(doc_data. color);

/* #if AX_X11 */
#if 0
#if 1
	TmDchar(&hDesk, &wDesk, &hDefault, &wDefault, &DcharData, TmDefDev);
	WP.tmMilHt = DcharData.grMHt;
	WP.tmPixHt = DcharData.grPHt;
#endif  /* DAMNT */
#endif

}

/*-----------------------------------------------------------------------------
	newpage

	new page
-----------------------------------------------------------------------------*/
static void newpage(p)
	data_ptr p ;

{
	data_ptr s ;
	data_ptr hf, ptext, ftext, text ;
	data_ptr point, tsize ;
	char *string ;
	int landscape ;
	int i, k, y ;
	int n, ix ;

	s = AxArrayFromArray(p, 2) ;

	page_data.height = AxIntFromArray(s, 2) ;
	page_data.width = AxIntFromArray(s, 3) ;
	page_data.orientation = AxStrPtrFromArray(s, 4) ;
	landscape = strcmp(page_data.orientation, "Landscape") == 0 ;
	page_data.lm = AxIntFromArray(s, 5) ;
	page_data.rm = AxIntFromArray(s, 6) ;
	page_data.tm = AxIntFromArray(s, 7) ;
	page_data.bm = AxIntFromArray(s, 8) ;

	se_newpage(landscape,
				page_data.width,
				page_data.height,
				page_data.lm,
				FALSE,
				FALSE,
				FALSE,
				NULL, NULL,
				0, 0, FALSE) ;

	/* header/footer  text */
	for (k = 9 ; k <= 10 ; k++)
	{
		hf = AxArrayFromArray(s, k) ;
		for (i = 0 ; i <= 2 ; i++)
		{
			ptext = AxArrayFromArray(hf, i) ;
			ftext = AxArrayFromArray(ptext, 2) ;
			if (ftext)
			{
				n = AxArraySize(ftext) ;
				pen = AxArrayFromArray(ptext, 0) ;
				text_pen() ;
				for (ix = 0 ; ix < n ; ix++)
				{
					text = AxArrayFromArray(ftext, ix) ;
					point = AxArrayFromArray(text, 0) ;
					string = AxStrFromArray(text, 1) ;
					tsize = AxArrayFromArray(text, 2) ;
					text_draw(point, tsize, string) ;
				}
			}
		}
	}
}

/*-----------------------------------------------------------------------------
	text_type


-----------------------------------------------------------------------------*/
static void text_type(p)
	data_ptr p ;

{
	data_ptr text, point, tsize ;
	char *str ;

	pen = AxArrayFromArray(p, 1) ;
	text_pen() ;
	text = AxArrayFromArray(p, 2) ;
	point = AxArrayFromArray(text, 0) ;
	str = AxStrFromArray(text, 1) ;
	tsize = AxArrayFromArray(text, 2) ;
	text_draw(point, tsize, str) ;
}

/*-----------------------------------------------------------------------------
	text_pen
	
	get the text attributes off pen to tattrs
-----------------------------------------------------------------------------*/
static void text_pen()
{
	data_ptr font ;
	data_ptr fattr ;
	data_ptr point ;
	int c, m, y, k ;
	int r, g, b ;

	memset(&tattrs, 0, sizeof(textAttrs)) ;

	font = AxArrayFromArray(pen, 15) ;
	tflags = 0 ;

	fattr = AxArrayFromArray(font, 0) ;
	tattrs.fontFamily = AxStrFromArray(fattr, 1) ;
	tflags |= TXfontFamily ;

	fattr = AxArrayFromArray(font, 1) ;
	tattrs.pointsize = 10000 * AxIntFromArray(fattr, 1) ;
	tflags |= TXpointsize ;

	fattr = AxArrayFromArray(font, 2) ;
	if (AxIntFromArray(fattr, 1))
		tattrs.weight = 1 ;
	tflags |= TXweight ;

	fattr = AxArrayFromArray(font, 3) ;
	if (AxIntFromArray(fattr, 1))
		tattrs.slant = 1 ;
	tflags |= TXslant ;

#if 0
	fattr = AxArrayFromArray(font, 4) ;
	tattrs.dropShadow = AxIntFromArray(fattr, 1) ;
	tflags |= TXdropShadow ;
#endif

	color_pen(2, &c, &m, &y, &k) ;

	tattrs.fillcolor = (((((c << 8) | m) << 8) | y) << 8) | k ;
	tflags |= TXfillcolor ;

}

/*-----------------------------------------------------------------------------
	color_pen

	determine cmyk from pen color 
-----------------------------------------------------------------------------*/
static void color_pen(n, c, m, y, k)
	int n ;
	int *c, *m, *y, *k ;
{
	data_ptr color ;
	int ctype ;
	char *color_name ;
	int r, g, b ;
	XColor cdef;

	color = AxArrayFromArray(pen, n) ;

	ctype = AxIntFromArray(color, 0) ;
	switch (ctype)
	{
	case 0:
		/* rgb */
		r = AxIntFromArray(color, 1) ;
		g = AxIntFromArray(color, 2) ;
		b = AxIntFromArray(color, 3) ;
		rgb2cmyk(r, g, b, c, m, y, k, 0) ;
		break ;

	case 2:
		/* cmyk */
		*c = AxIntFromArray(color, 1) ;
		*m = AxIntFromArray(color, 2) ;
		*y = AxIntFromArray(color, 3) ;
		*k = AxIntFromArray(color, 4) ;

	case 5:
		/* name */
		color_name = AxStrFromArray(color, 1) ;
		if (XParseColor(Dpy, ApplixCmap, color_name, &cdef))
		{
			r = cdef.red >> 8 ;
			g = cdef.green >> 8 ;
			b = cdef.blue >> 8 ;
			rgb2cmyk(r, g, b, c, m, y, k, 0) ;
		}
		else
		{
			*c = 0 ;
			*m = 0 ;
			*y = 0 ;
			*k = 255 ;
		}
		break ;

	default:
		/* any color you want, as long as it's black */
		*c = 0 ;
		*m = 0 ;
		*y = 0 ;
		*k = 255 ;
		break ;

	}
}


/*-----------------------------------------------------------------------------
	text_draw

	Draw text at point size rsize
-----------------------------------------------------------------------------*/
static void text_draw(point, tsize, text)
	data_ptr point, tsize ;
	char *text ;
{
	tx_char *textstr ;
	int len, px, py ;
	int i ;

	px = AxIntFromArray(point, 0) + page_data.lm ;
	py = AxIntFromArray(point, 1) + page_data.tm ;
	py += AxIntFromArray(tsize, 2) ;
	len = strlen(text) ;

#if BUILD_WIDE
   textstr = (tx_char *)malloc((len+1) * sizeof(tx_char)) ;
   len = MbToWStr(text, textstr);
#else
	textstr = (tx_char *)malloc(len * sizeof(tx_char)) ;

	for (i = 0 ; i < len ; i++)
		textstr[i] = text[i] ;
#endif

	axDrawText(0L, textstr, len, px, py, &tattrs, tflags) ;
	free(textstr) ;
}



/*-----------------------------------------------------------------------------
	polygon

-----------------------------------------------------------------------------*/
static void polygon(p)
	data_ptr p ;

{
	int polytype, i, n ;
	data_ptr pt, pp ;

	pen = AxArrayFromArray(p, 1) ;
	poly_pen() ;
	polytype = AxIntFromArray(p, 2) ;
	fill = AxIntFromArray(p, 3) ;
	if (fill)
	{
		pattrs.f_type = GRAY_FILL_TYPE ;
		pattrs.f_pat = 1000 ;
		pattrs.ffg = pattrs.lfg ;
		pattrs.fbg = pattrs.lbg ;
		pattrs.fill_flag = TRUE ;
	}
	pski.nCtrls = 0 ;
	pski.controls = NULL ;


	switch (polytype)
	{
	case PRINT_POLY_LINE:
		poly_line(p) ;
		break ;

	case PRINT_POLY_RECTANGLE:
		poly_rectangle(p) ;
		break ;

	case PRINT_POLY_ARC:
		poly_arc(p) ;
		break ;

	case PRINT_POLY_POLY:
		poly_poly(p) ;
		break ;
	}

}


/*-----------------------------------------------------------------------------
	poly_pen

	get the polyline attributes off pen to to pattrs
-----------------------------------------------------------------------------*/
static void poly_pen()
{
	int c, m, y, k ;

	memset(&pattrs, 0, sizeof(attr_type)) ;
	color_pen(2, &c, &m, &y, &k) ;
	pattrs.lfg.c = c ;
	pattrs.lfg.m = m ;
	pattrs.lfg.y = y ;
	pattrs.lfg.k = k ;
	color_pen(3, &c, &m, &y, &k) ;
	pattrs.lbg.c = c ;
	pattrs.lbg.m = m ;
	pattrs.lbg.y = y ;
	pattrs.lbg.k = k ;

	pattrs.l_wid = AxIntFromArray(pen, 4) ;
	if (pattrs.l_wid == 0)
		pattrs.l_wid = 1 ;
	pattrs.l_style = AxIntFromArray(pen, 5) ;
	pattrs.l_cap = AxIntFromArray(pen, 6) ;

#if 0
	join style = AxIntFromArray(pen, 7) ;
	fill style = AxIntFromArray(pen, 8) ;
	fill rule = AxIntFromArray(pen, 9) ;
	arc mode = AxIntFromArray(pen, 10) ;
	tile = AxStrFromArray(pen, 11) ;
	stipple = AxStrFromArray(pen, 12) ;
#endif
}


/*-----------------------------------------------------------------------------
	poly_line

	Simple line
-----------------------------------------------------------------------------*/
static void poly_line(p)
	data_ptr p ;
{
	int i ;
	data_ptr pt, pp ;

	pski.nPoints = 2 ;
	pski.points = (TmPoint32 *)malloc(pski.nPoints * sizeof(TmPoint32)) ;
	pt = AxArrayFromArray(p, 5) ;

	for (i = 0 ; i < 2 ; i++)
	{
		pp = AxArrayFromArray(pt, i) ;
		pski.points[i].x = AxFloatFromArray(pp, 0) + page_data.lm ;
		pski.points[i].y = AxFloatFromArray(pp, 1) + page_data.tm ;
	}
	pattrs.l_pat = 4 ;
	ax_se_pline(0L, 1, &pattrs, &pski, NULL, NULL) ;
	free(pski.points) ;
}

/*-----------------------------------------------------------------------------
	poly_rectangle

-----------------------------------------------------------------------------*/
static void poly_rectangle(p)
	data_ptr p ;
{
	data_ptr pt, pp ;
	TmPoint32 lrc ;
	int i ;

	pski.nPoints = 5 ;
	pski.points = (TmPoint32 *)malloc(pski.nPoints * sizeof(TmPoint32)) ;
	pt = AxArrayFromArray(p, 5) ;
	pp = AxArrayFromArray(pt, 0) ;

	pski.points[0].x = AxFloatFromArray(pp, 0) + page_data.lm ;
	pski.points[0].y = AxFloatFromArray(pp, 1) + page_data.tm ;
	for (i = 1 ; i <= 4 ; i++)
		pski.points[i] = pski.points[0] ;

	pp = AxArrayFromArray(pt, 1) ;
	lrc.x = AxFloatFromArray(pp, 0) ;
	lrc.y = AxFloatFromArray(pp, 1) ;

	pski.points[1].x += lrc.x ;
	pski.points[2].x += lrc.x ;
	pski.points[2].y += lrc.y ;
	pski.points[3].y += lrc.y ;

	pattrs.l_pat = 4 ;
	ax_se_pline(0L, 1, &pattrs, &pski, NULL, NULL) ;
	free(pski.points) ;
}

/*-----------------------------------------------------------------------------
	poly_arc

-----------------------------------------------------------------------------*/
static void poly_arc(p)
	data_ptr p ;
{
}

/*-----------------------------------------------------------------------------
	poly_poly
-----------------------------------------------------------------------------*/
static void poly_poly(p)
	data_ptr p ;
{
}


/*-----------------------------------------------------------------------------
	chart

-----------------------------------------------------------------------------*/
static void chart(p)
		data_ptr p ;
{
	data_ptr pointP ;
	data_ptr geP, valP;
	data_ptr insetP, retData, ret;
	int err, gfx;
	data_ptr dpu, prc;

	geP = AxArrayElement(p, 1);
	insetP = AxArrayElement(p, 2);

	pointP = AxMakeArray(2);
	AxAddIntToArray(pointP, 0, page_data.lm);
	AxAddIntToArray(pointP, 1, page_data.tm);
	err = ElfCallByName(NXLT("correction_for_margins$"), &ret,2, pointP, p);
	insetP = AxArrayElement(ret, 0);

	err = ElfCallByName(NXLT("gfx_render_graphic@"),&retData, 2, geP, insetP);
}
 
 
/*-----------------------------------------------------------------------------
	icon

-----------------------------------------------------------------------------*/
static void icon(p)
	data_ptr p ;
{
	data_ptr point1 ;
	char *name ;
		char pathname[STRPATH+1] ;
		data_ptr geP, fileP, insetP, wP, srcP, retData;
	data_ptr sP, dP, clP ;
		char *gePtr;
		int err;
		pixcache_t *pix;
	int x, y ;
	int srcX, srcY, srcWidth, srcHeight ;
		AXWidget w;

	pen = AxArrayFromArray(p, 1) ;
	point1 = AxArrayFromArray(p, 2) ;
	name = AxStrFromArray(p, 3) ;

	pix = find_icon(name, pathname) ;
	if (!pix)
		return ;

	x = AxIntFromArray(point1, 0) + page_data.lm ;
	y = AxIntFromArray(point1, 1) + page_data.tm ;

	srcX = 0 ;
	srcY = 0 ;
	srcWidth = (1000 * pix->width) / 75 ;
	srcHeight = (1000 * pix->height) / 75 ;


	/* create the graphic */
	srcP = AxMakeArray(1);
	err = ElfCallByName(NXLT("create_graphic$"),&retData, 1, srcP);
	AxFreeData(srcP);
	if ( err != AOK )
	{
		AxFreeData(retData);
		return ;
	}

	gePtr = (void *)AxGFXFromData(retData);
	geP = (data_ptr)AxMakeGFXData(gePtr);
	AxFreeData(retData);

	/* read the file */
	fileP = AxMakeStrData(-1, pathname);
	wP = AxMakeIntData(TRUE);
	srcP = AxMakeArray(3);
	AxAddDataToArray(srcP, 0, geP);
	AxAddDataToArray(srcP, 1, fileP);
	AxAddDataToArray(srcP, 2, wP);
	err = ElfCallByName(NXLT("read_graphic_file$"),&retData, 1, srcP);
	AxFreeData(srcP);
	if ( err != AOK )
		AxFreeData(retData);

	/* render the graphic unto Caesar */

	/* Build the inset struct */
#if 0
	/* Window info */
	w = WidgetOfWindow(WP.wid);
	wP = (AXWidget)AxMakeBinaryData(sizeof(AXWidget), &w);
#endif
	w = (AXWidget) calloc(1, sizeof(axw_any));
	w->type = W_CANVAS; /* bogus widget type */
	w->parent = (AXWidget_main)w;
	w->width = srcWidth;
	w->height = srcHeight;
/*	AxSetDrawableOfWidget(w, pmap);
*/
	wP = AxMakeBinaryData(sizeof(AXWidget), &w);


	/* Source info */
	sP = AxMakeArray(4);
	AxAddDblToArray(sP, 0, (double)srcX);
	AxAddDblToArray(sP, 1, (double)srcY);
	AxAddDblToArray(sP, 2, (double)srcWidth);
	AxAddDblToArray(sP, 3, (double)srcHeight);

#if 0
	/* Destination info */
	destr.xpos = WPXPIX2SCREENMIL(x);
	destr.ypos = WPYPIX2SCREENMIL(y);
	destr.width = width * WP.dsZoomFactor / 100;
	destr.height = height * WP.dsZoomFactor / 100;
#endif

	dP = AxMakeArray(4);
	AxAddDblToArray(dP, 0, (double)x);
	AxAddDblToArray(dP, 1, (double)y);
	AxAddDblToArray(dP, 2, (double)srcWidth);
	AxAddDblToArray(dP, 3, (double)srcHeight);

	clP = AxMakeArray(4);
	AxAddIntToArray(clP, 0, x);
	AxAddIntToArray(clP, 1, y);
	AxAddIntToArray(clP, 2, srcWidth);
	AxAddIntToArray(clP, 3, srcHeight);

	insetP = AxMakeArray(12);
	AxAddIntToArray(insetP, 0, 2);		/* ??? clearing 'window' */
#if 0
	AxAddIntToArray(insetP, 1, WpToGrClipMode(clipMode));
	AxAddIntToArray(insetP, 2, proportional);
#endif
	AxAddIntToArray(insetP, 1, 0);
	AxAddIntToArray(insetP, 2, 0);
	AxAddDblToArray(insetP, 3, 100.0);
	AxAddDblToArray(insetP, 4, 100.0);
	AxAddDataToArray(insetP, 5, sP);
	AxAddDataToArray(insetP, 6, dP);
	AxAddDataToArray(insetP, 7, clP);
	AxAddDataToArray(insetP, 11, wP);

	/* set GE to use mils */
	SetGEunits(geP) ;

	err = ElfCallByName(NXLT("gfx_render_graphic@"),&retData, 2, geP, insetP);
	AxFreeData(insetP);
	free(w) ;

	srcP = AxMakeArray(1);
	AxAddDataToArray(srcP, 0, geP);
	err = ElfCallByName(NXLT("destroy_graphic@"),&retData, 1, srcP);
	AxFreeData(srcP);
}



/*-----------------------------------------------------------------------------
	find_icon

-----------------------------------------------------------------------------*/
static pixcache_t *find_icon(name, pathname)
	char *name ;
	char *pathname ;
{
	char *strP;
	char glom_name[STRPATH+1];
	int ix, cacheIx, size;
	data_ptr searchList;
	pixcache_t *p = NULL;



	/* Search standard places for .im's and .ag's */
	searchList = ElfGetSysVar(NXLT("elf_search_list@"));
	if (!searchList)
	{
		ElfSetStandardSearchList();
		searchList = ElfGetSysVar(NXLT("elf_search_list@"));
	}

	sprintf(pathname,"%s", name);
	AxSysDosUnixPath(pathname);
	if ( Aaccess(pathname,0) == 0)
		goto ok ;

	size = AxArraySize(searchList);
	for (ix = 0; ix < size; ix++)
	{
		strP = (char *)AxStrFromArray(searchList, ix);
		if (!strP)
			return NULL ;
		sprintf(pathname,"%s/%s", strP, name);
		AxSysDosUnixPath(pathname);
		if ( Aaccess(pathname,0) == 0)
			goto ok ;
		sprintf(pathname,"%s/%s.im", strP, name);
		AxSysDosUnixPath(pathname);
		if ( Aaccess(pathname,0) == 0)
			goto ok ;
		sprintf(pathname,"%s/%s.ag", strP, name);
		AxSysDosUnixPath(pathname);
		if ( Aaccess(pathname,0) == 0)
			 goto ok;
	}
	*pathname = 0 ;
ok:
	if (*pathname)
		p = Pixmapper(name);
	return p ;
}

/*-----------------------------------------------------------------------------
	SetGEunits

-----------------------------------------------------------------------------*/
static void SetGEunits (geP)
	data_ptr geP;
{
	typeErr err;
	data_ptr retData;
	data_ptr geUnitsNameP;
	data_ptr geUnitsPrecisP;
	data_ptr wpUnitsDpuP;
	mbchar *geUnitsName;
	int geUnitsDpu;
	int geUnitsPrecis;

	/* get GE units */
	err = ElfCallByName(NXLT("WP_GFX_GET_UNIT_NAME$"), &retData, 1, geP);
	geUnitsName = MBALCSTR((mbchar *)AxStrPtrFromDataPtr(retData));
	AxFreeData(retData);

	err = ElfCallByName(NXLT("WP_GFX_GET_UNIT_SIZE$"), &retData, 1, geP);
	geUnitsDpu = AxIntFromDataPtr(retData);
	AxFreeData(retData);

	err = ElfCallByName(NXLT("WP_GFX_GET_UNIT_PRECISION$"), &retData, 1, geP);
	geUnitsPrecis = AxIntFromDataPtr(retData);
	AxFreeData(retData);

	/* reset GE precision to 1 'ge dot' (mil) per WP unit (mil)  */
	geUnitsNameP = AxMakeStrData(-1, geUnitsName);
	wpUnitsDpuP = AxMakeNumData(1);
	geUnitsPrecisP = AxMakeNumData(geUnitsPrecis);

	err = ElfCallByName(NXLT("WP_GFX_SET_UNITS$"), &retData, 4, geP,
									geUnitsNameP, wpUnitsDpuP, geUnitsPrecisP);

	AxFreeData(retData);
	AxFreeData(geUnitsNameP);
	AxFreeData(wpUnitsDpuP);
	AxFreeData(geUnitsPrecisP);

	return ;
}


