/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elfnet.c
   Applix ASC version: 6
   Date: Mon Nov 16 13:53:47 1998
*/
#include <stdio.h> /* just for printfs */ 
#if AXMS
#include <basicwin.h>
#endif
#include <aos.h>
#include <elfapi.h>
#include <axnet.h>
#include <signal.h>
#include <elfnet.h>
/***********************************************************************
function: AxNetSignalProcess - Have axnet send a signal to a process.
If 'hostName' is NULL or zero-length string, use own host.
Return 0 on success; -1 on failure
-- UnPublished!!
*/
int
AxNetSignalProcess(hostName, pidAndSig, code, str, obj)
char *hostName;
elfData pidAndSig; /* elf array of pid num and signal num */
int *code;
char **str, **obj;
{
    int channel = -1;
    elfData request, response;
#if AXMS
    int ret;
    DWORD size_username;
    char username[21];    /* NT max username = 20 */
#endif

    channel = AxNetConnect(hostName, code, str, obj);
    if (*code != 0)
        goto badret;
    request = AxCopyData(pidAndSig);

#if AXMS
    /* there is no uid on NT, so pass -1 as a flag. */
    request = AxAddIntToArray(request, 2, (int)-1);
    /* pass the username so that it can be verified on the other end */
    /* since the 'uid' is meaningless on NT                          */
    size_username = sizeof(username);
    ret = MS32_GetUserName(username, &size_username);
    if ( !ret )
        /* do not know why this would fail, but ... */
        username[0] = 0;
    request = AxAddStrToArray(request, 3, username);
#else
    request = AxAddIntToArray(pidAndSig, 2, (int) getuid());
#endif

    AxSendMessageToServer(channel,AXNET_SIGNAL_PROCESS,request,code,str,obj);
    if (*code != 0)
        goto badret;
    response = AxGetResponseFromServer(channel,code,str,obj);
    if (*code != 0)
        goto badret;
    AxClientDisConnect(channel);
    return(0);
badret:
    if (channel != -1)
        AxClientDisConnect(channel);
    return(-1);
}

/* convert platform independant signal into port specific.
   Returns AXNOSIG if not found.
   All the # ifdefs below are to avoid build errors due to
   platform specific variations
   */
int portSpecificSignal( genericSignal )
int genericSignal;
{
    switch( genericSignal ){
        
#ifdef SIGHUP
        case AXSIGHUP: return(SIGHUP);
#endif
#ifdef SIGINT
        case AXSIGINT: return(SIGINT);
#endif
#ifdef SIGQUIT
        case AXSIGQUIT: return(SIGQUIT);
#endif
#ifdef SIGILL
        case AXSIGILL: return(SIGILL);
#endif
#ifdef SIGTRAP
        case AXSIGTRAP: return(SIGTRAP);
#endif
#ifdef SIGABRT
        case AXSIGABRT: return(SIGABRT);
#endif
#ifdef SIGEMT
        case AXSIGEMT: return(SIGEMT);
#endif
#ifdef SIGFPE
        case AXSIGFPE: return(SIGFPE);
#endif
#ifdef SIGKILL
        case AXSIGKILL: return(SIGKILL);
#endif
#ifdef SIGBUS
        case AXSIGBUS: return(SIGBUS);
#endif
#ifdef SIGSEGV
        case AXSIGSEGV: return(SIGSEGV);
#endif
#ifdef SIGSYS
        case AXSIGSYS: return(SIGSYS);
#endif
#ifdef SIGPIPE
        case AXSIGPIPE: return(SIGPIPE);
#endif
#ifdef SIGALRM
        case AXSIGALRM: return(SIGALRM);
#endif
#ifdef SIGTERM
        case AXSIGTERM: return(SIGTERM);
#endif
#ifdef SIGURG
        case AXSIGURG: return(SIGURG);
#endif
#ifdef SIGSTOP
        case AXSIGSTOP: return(SIGSTOP);
#endif
#ifdef SIGTSTP
        case AXSIGTSTP: return(SIGTSTP);
#endif
#ifdef SIGCONT
        case AXSIGCONT: return(SIGCONT);
#endif
#ifdef SIGCHLD
        case AXSIGCHLD: return(SIGCHLD);
#endif
#ifdef SIGTTIN
        case AXSIGTTIN: return(SIGTTIN);
#endif
#ifdef SIGTTOU
        case AXSIGTTOU: return(SIGTTOU);
#endif
#ifdef SIGIO
        case AXSIGIO: return(SIGIO);
#endif
#ifdef SIGXCPU
        case AXSIGXCPU: return(SIGXCPU);
#endif
#ifdef SIGXFSZ
        case AXSIGXFSZ: return(SIGXFSZ);
#endif
#ifdef SIGVTALRM
        case AXSIGVTALRM: return(SIGVTALRM);
#endif
#ifdef SIGPROF
        case AXSIGPROF: return(SIGPROF);
#endif
#ifdef SIGWINCH
        case AXSIGWINCH: return(SIGWINCH);
#endif
#ifdef SIGLOST
        case AXSIGLOST: return(SIGLOST);
#endif
#ifdef SIGUSR1
        case AXSIGUSR1: return(SIGUSR1);
#endif
#ifdef SIGUSR2
        case AXSIGUSR2: return(SIGUSR2);
#endif
        default: return(AXNOSIG);
    }   
}

