/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winservice.c
   Applix ASC version: 15
   Date: Mon Nov 16 13:53:27 1998
*/
#include <stdio.h>
#include <windows.h>
#include <winmap.h>

#define STRIEQ( a, b )  (stricmp(a,b) == 0)

/* move to winmap.h! */
#define MSNT_ControlService             ControlService
#define MSNT_StartService               StartService
#define MSNT_SetServiceStatus           SetServiceStatus
#define MSNT_RegisterServiceCtrlHandler RegisterServiceCtrlHandler
#define MSNT_RegisterEventSource        RegisterEventSource
#define MSNT_OpenSCManager              OpenSCManager
/********************/


/**********************************************************************/
/* The following block of stuff is the code to implement the interface */
/* to the NT "Services" manager. To get this to work right, when things */
/* are to exit, the code MUST return back up through the function specified */
/* in the dispatch table passed to MS_StartServiceCtrlDispatcher().  Trying */
/* to exit otherwise seems to always result in some error being reported */
/* by the service manager, or worse, the whole service manager gets hung. */

/* Since the service manager interface to do the "stop" is implemented as */
/*   a callback, the most the callback can do is set some flag.  Because */
/*   of this one MUST implement a service task as a multi-threaded entity.  */
/* To keep the main-line axnet processing as generic as possible, the code  */
/*   below spwaws a small thread simple to interface with the service   */
/*   manager.  This thread will call "exit()" on the stop command to shut */
/*   down all of axnet.  - jon                                            */
/**********************************************************************/

SERVICE_STATUS_HANDLE   sshStatusHandle;

char ServiceName[MAX_PATH+1];

// this event is signaled when the worker thread ends
HANDLE                  hServDoneEvent = NULL;

/* struct to pass standard argc/argv to threads. */
typedef struct thread_args_t {
    int argc;
    char **argv;
} thread_args_t;

extern DWORD WINAPI AxStartServiceThread(void*);

#if !_CRXRPT
#include <winservice.p>
#else
  /* local prototypes, so that we can use this as part of a VisualStudio */
  /* project which has no idea where the ".p" files live                 */
void ax_service_main(DWORD argc, LPSTR *argv);
VOID WINAPI ax_service_ctrl(DWORD dwCtrlCode);
VOID AddToMessageLog(LPTSTR lpszMsg);
BOOL AxReportStatusToSCMgr(DWORD dwCurrentState,
                           DWORD dwWin32ExitCode,
                           DWORD dwServiceSpecificExitCode,
                           DWORD dwCheckPoint,
                           DWORD dwWaitHint);
int RemoveProgramAsNtService(char *program_name, int force_quiet);
void AddNamedNtService(char *program_name, char *display_name, 
                  char *account_name, char *clear_password);
#endif






#if 0
DWORD WINAPI 
AxStartServiceThread(void)
{
    char buf[MAX_PATH];
    SERVICE_TABLE_ENTRY dispatchTable[] =
       {
          { "", (LPSERVICE_MAIN_FUNCTION)ax_service_main },
          { NULL, NULL }
       };

    /* dispatch table to the service manager libraries, which will */
    /* call back into our normal main loop "elfmain".  This never  */
    /* comes back until the service is stopped. - jon              */
    if ( !MSNT_StartServiceCtrlDispatcher(dispatchTable) )
       {
       sprintf(buf, "%s: StartServiceCtrlDispatcher failed.", ServiceName);
       AddToMessageLog(buf);
       return 0;
       }

    //WSACleanup();
    return 0;
}
#endif

int
AxInitService(char *service_name)
{
    char buf[MAX_PATH];
    SERVICE_TABLE_ENTRY dispatchTable[] =
       {
          { "", (LPSERVICE_MAIN_FUNCTION)ax_service_main },
          { NULL, NULL }
       };

    strcpy(ServiceName, service_name);

    /* dispatch table to the service manager libraries, which will */
    /* call back into our normal main loop "elfmain".  This never  */
    /* comes back until the service is stopped. - jon              */
    if ( !MSNT_StartServiceCtrlDispatcher(dispatchTable) )
       {
       sprintf(buf, "%s: StartServiceCtrlDispatcher failed.", ServiceName);
       AddToMessageLog(buf);
       }

    return 0;
}


BOOL
AxStartService(int argc, char **argv)
{
    BOOL retval = FALSE;

#if !_CRXRPT    /*  {  */
    HANDLE hThread;
    DWORD thread_id;
    thread_args_t *thread_args;

    thread_args = (thread_args_t*)malloc(sizeof(thread_args_t));
    thread_args->argc = argc;
    thread_args->argv = argv;

    /* start the thread that does the real work. the thread should be placed
     * in its appropriate module (e.g javanet.c for aagtwy) and call the
     * appropriate function, in this case elfmain().
     */
    hThread = MS_CreateThread((LPSECURITY_ATTRIBUTES)NULL, 0,
                              (LPTHREAD_START_ROUTINE)AxStartServiceThread,
                              (LPVOID)thread_args, 0, &thread_id);
    if ( hThread )
        {
        MS_CloseHandle(hThread);
        retval = TRUE;
        }
#endif      /*  !_CRXRPT  }   */
    return retval;
}


#if 0
void
CreateServiceThread(char *service_name)
{
    HANDLE threadH;
    DWORD thread_id;

    strcpy(ServiceName, service_name);

    threadH = MS_CreateThread( (LPSECURITY_ATTRIBUTES)NULL,
                               0,            /* initial stack size default */
                               (LPTHREAD_START_ROUTINE)AxStartServiceThread,
                               (LPVOID)NULL, /* optional argument */
                               0,            /* run thread now */
                               &thread_id );
    MS_CloseHandle(threadH);
    return;
}
#endif

void
AxStopService(DWORD error_code)
{
    if ( hServDoneEvent)
        MS_CloseHandle(hServDoneEvent);
    if( sshStatusHandle )
        AxReportStatusToSCMgr(SERVICE_STOPPED, error_code, 0, 0, 0);
}

void
AxSignalServiceStop()
{
/* Stops the service by allowing ax_service_main to complete */
    MS_SetEvent(hServDoneEvent);
}

BOOL AxReportStatusToSCMgr(DWORD dwCurrentState,
                           DWORD dwWin32ExitCode,
                           DWORD dwServiceSpecificExitCode,
                           DWORD dwCheckPoint,
                           DWORD dwWaitHint)
{
    SERVICE_STATUS status;       // current status of the service
    BOOL fResult = TRUE;


    status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    status.dwCurrentState = dwCurrentState;
    if ( dwCurrentState == SERVICE_START_PENDING )
        status.dwControlsAccepted = 0;
    else
        status.dwControlsAccepted = SERVICE_ACCEPT_STOP;
    if ( dwServiceSpecificExitCode == 0 )
        status.dwWin32ExitCode = dwWin32ExitCode;
    else
        status.dwWin32ExitCode = ERROR_SERVICE_SPECIFIC_ERROR;
    status.dwServiceSpecificExitCode = dwServiceSpecificExitCode;

    status.dwCheckPoint = dwCheckPoint;
    status.dwWaitHint = dwWaitHint;

    // Report the status of the service to the service control manager.
    //
    fResult = MSNT_SetServiceStatus(sshStatusHandle, &status);
    return fResult;
}




VOID WINAPI 
ax_service_ctrl(DWORD dwCtrlCode)
{
    DWORD currentState = 0;

    // Handle the requested control code.
    //
    switch (dwCtrlCode)
        {
        // Stop the service.
        //
        case SERVICE_CONTROL_STOP:
            currentState = SERVICE_STOP_PENDING;
            AxReportStatusToSCMgr(currentState, NO_ERROR, 0, 1, 5000);

            /* signal the event the service theread is waiting on */
            AxSignalServiceStop();
            return;

        // Update the service status.
        //
        case SERVICE_CONTROL_INTERROGATE:
            break;

        default:
            break;
        }

    AxReportStatusToSCMgr(currentState, NO_ERROR, 0, 0, 0);
}



void ax_service_main(DWORD argc, LPSTR *argv)
{
    /* first register our service control handler for callback from the */
    /* service manager                                                  */
    
    sshStatusHandle = 
                MSNT_RegisterServiceCtrlHandler(ServiceName, ax_service_ctrl);
    if ( !sshStatusHandle )
        {
        AxStopService(MS_GetLastError());
        return;
        }

    /* report the status to the service control manager */
    if ( !AxReportStatusToSCMgr(SERVICE_START_PENDING,
                                NO_ERROR,
                                0, 1, 5000) )
        {
        AxStopService(MS_GetLastError());
        return;
        }

    // Create the event object. The control handler function signals
    // this event when it receives the "stop" control code.
    //
    hServDoneEvent = MS_CreateEvent((LPSECURITY_ATTRIBUTES)NULL,
                                    TRUE,    // manual reset event
                                    FALSE,   // not-signalled
                                    (LPCTSTR)NULL);   // no name
    if ( hServDoneEvent == (HANDLE)NULL )
        {
        AxStopService(MS_GetLastError());
        return;
        }

    if ( !AxReportStatusToSCMgr(SERVICE_START_PENDING,
                                NO_ERROR,
                                0, 2, 5000) )
        {
        AxStopService(MS_GetLastError());
        return;
        }

    if ( !AxStartService(argc, argv) )
        {
        AxStopService(MS_GetLastError());
        return;
        }

    if ( !AxReportStatusToSCMgr(SERVICE_RUNNING,       // service state
                              NO_ERROR,              // exit code
                              0, 0, 0) )
        {
        AxStopService(MS_GetLastError());
        return;
        }

    /* wait for the termination event to get thrown */
    MS_WaitForSingleObject(hServDoneEvent, INFINITE);

    AxStopService(0);
}




VOID AddToMessageLog(LPTSTR lpszMsg)
{
    HANDLE  hEventSource;
    LPTSTR  lpszStrings[2];
    int ret;
    PVOID err_str;
    
    hEventSource = MSNT_RegisterEventSource((LPCTSTR)NULL, TEXT(ServiceName));

    if ( hEventSource != NULL ) 
        {
        ret = MS_GetLastError();
        MS_FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
                              FORMAT_MESSAGE_FROM_SYSTEM, NULL, ret,
                            MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
                            (LPTSTR) &err_str, 0, NULL);
    
        lpszStrings[0] = lpszMsg;
        lpszStrings[1] = err_str;
    
        MSNT_ReportEvent(hEventSource, // handle of event source
                EVENTLOG_ERROR_TYPE,  // event type
                0,                    // event category
                0,                    // event ID
                NULL,                 // current user's SID
                1,                    // strings in lpszStrings
                0,                    // no bytes of raw data
                lpszStrings,          // array of error strings
                NULL);                // no raw data

        MS_LocalFree(err_str);
        (VOID)MSNT_DeregisterEventSource(hEventSource);
        }

    return;
}

    

static void mess_out( char *mess, char *service_name )
{
    char msg_title[80];
    sprintf(msg_title, "\"%s\" service", service_name);
    MS_MessageBox( (HWND)NULL, mess, msg_title, MB_OK );
    return;
}

static void format_post_error( int ret, char *service_name )
{
    PVOID err_str;

    ret = MS_GetLastError();
    MS_FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
                          FORMAT_MESSAGE_FROM_SYSTEM, NULL, ret,
                        MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
                        (LPTSTR) &err_str, 0, NULL);
    mess_out( err_str, service_name );
    MS_LocalFree(err_str);
    return;
}

static SC_HANDLE open_manager_channel(char *service_name)
{
    SC_HANDLE schSCManager;
    int ret;

    /* open a handle to the service manager */
    schSCManager = MSNT_OpenSCManager(NULL,        // machine (NULL == local)
                                      NULL,        // database (NULL == default)
                                      SC_MANAGER_ALL_ACCESS); // access required
    if ( NULL == schSCManager )
        { 
        ret = MS_GetLastError();
        format_post_error( ret, service_name );
        exit( 1 );
        }
    return schSCManager; 
}

static SC_HANDLE axmsCreateNtServiceWithAccount(char *program_name, 
                                         char *display_str,
                                         char *account_name, 
                                         char *clear_password)
{
    HMODULE hxll;
    char tmp_loc[MAX_PATH+1];
    char ExecLocation[MAX_PATH+1];
    char display_name[180];
    OSVERSIONINFO osversion;
    int ret;
    SC_HANDLE schService;
    SC_HANDLE schSCManager;
    char buff[100];
    char *p;

    /* get the location of this executable */
    hxll = MS_GetModuleHandle( NULL );
    if ( hxll == (HMODULE)NULL )
        return NULL;

    ret = MS_GetModuleFileName(hxll, tmp_loc, sizeof(tmp_loc));
    if ( ret <= 0 )
        /* failure fallback */
        return NULL;
    strcpy(ExecLocation, "\"");
    strcat(ExecLocation, tmp_loc);
    strcat(ExecLocation, "\"");

    /* we only do this on Windows NT */
    osversion.dwOSVersionInfoSize = sizeof( OSVERSIONINFO );
    ret = MS_GetVersionEx( &osversion );
    if ( !ret )
        return NULL;

    if ( osversion.dwPlatformId != VER_PLATFORM_WIN32_NT )
        {
        /* not windows NT */
        char buf[256];
        sprintf(buf, "%s can only be installed as a service on Windows NT",
                                                             program_name);
        mess_out(buf, program_name);
        return NULL;
        }

    schSCManager = open_manager_channel(program_name);

    if ( (display_str == NULL) || (*display_str == '\0') )
        {
        /* use a default */
        strcpy(display_name, "Applix ");
        strcat(display_name, program_name);
        }
    else
        strcpy( display_name, display_str );

    /* verify that the account name is domain qualified */
    if ( account_name != NULL )
        {
        /* a user account name was passed */
        p = strchr( account_name, '\\' );
        if ( p == (char *)NULL )
            {
            /* not domain qualified, use the default of the local machine */
            sprintf( buff, ".\\%s", account_name );
            account_name = buff;
            }
        }


    /* now create the service */
    schService = MSNT_CreateService(
                        schSCManager,               // SCManager database
                        program_name,               // name of service
                        display_name,               // name to display
                        SERVICE_ALL_ACCESS,         // desired access
                        SERVICE_WIN32_OWN_PROCESS,  // service type
                        SERVICE_AUTO_START,         // start type
                        SERVICE_ERROR_NORMAL,       // error control type
                        ExecLocation,               // service's binary
                        NULL,                       // no load ordering group
                        NULL,                       // no tag identifier
                        NULL,                       // no dependencies
                        account_name,               // Local System Account
                        clear_password);            // NULL password
 
    MSNT_CloseServiceHandle(schSCManager);
    return schService;
}





void
AddNtService(char *program_name, char *account_name, char *clear_password)
{
    AddNamedNtService(program_name, NULL, account_name, clear_password);
}



void
AddNamedNtService(char *service_name, char *display_name, 
                  char *account_name, char *clear_password)
{
    char buf[500];
    DWORD last_error;
    SC_HANDLE serviceH;
    char *nameP;

    serviceH = axmsCreateNtServiceWithAccount( service_name, display_name,
                                             account_name, clear_password );
    if ( NULL == serviceH )
        {
        last_error = MS_GetLastError();
        format_post_error(last_error, service_name);
        }
    else
        {
        if ( display_name && (*display_name != '\0') )
            nameP = display_name;
        else
            nameP = service_name;
        sprintf(buf,"The \"%s\" service has been created.\n", nameP);
        strcat(buf,"Make sure there is an entry for it in the services file.\n");
        strcat(buf,"You can then start it by going to Control Panel/Services,\n");
        strcat(buf,"selecting the service, and clicking the \"Start\" button.\n");
        mess_out(buf, nameP);
        }
    MSNT_CloseServiceHandle(serviceH);
}




void 
axmsInstallAndStartNamedNtServiceAccount(char *service_name, 
                                         char *display_name,
                                         int force_quiet,
                                         char *account_name,
                                         char *clear_password )
{
    char buf[80];
    DWORD last_error;
    SC_HANDLE serviceH;
    int already_tried = FALSE;

try_again:
    serviceH = axmsCreateNtServiceWithAccount(service_name, display_name,
                                         account_name, clear_password );
    if ( !serviceH )
        {
        /* creation of service failed. if the reason is because it already
         * exists, and we've been passed the force flag, remove the already
         * installed service, and create a new one.
         */
        last_error = MS_GetLastError();
        if ( !already_tried &&
             (force_quiet && last_error == ERROR_SERVICE_EXISTS) )
            {
            already_tried = TRUE;
            if ( RemoveProgramAsNtService(service_name, force_quiet) )
                goto try_again;
            }
        else if ( !force_quiet )
            {
            format_post_error(last_error, service_name);
            return;
            }
        else
            return;
        }

    if ( !MSNT_StartService(serviceH, 0, (LPCTSTR*)NULL) )
        {
        last_error = MS_GetLastError();
        if ( !force_quiet )
            format_post_error(last_error, service_name);
        } 
    else if ( !force_quiet )
        {
        sprintf(buf,"The \"%s\" service has been created and started.\n", 
                                                           service_name);
        mess_out(buf, service_name);
        }
    MSNT_CloseServiceHandle(serviceH); 
}





void 
axmsInstallAndStartNtServiceAccount(char *service_name, 
                                    int force_quiet,
                                    char *account_name,
                                    char *clear_password )
{
    axmsInstallAndStartNamedNtServiceAccount(service_name, NULL,
                                 force_quiet, account_name, clear_password );
}

int RemoveProgramAsNtService(char *program_name, int force_quiet)
{ 

    #define SZ_ENUM_BUF 4096
    ENUM_SERVICE_STATUS        essServiceStatus[SZ_ENUM_BUF];
    DWORD   dwBufSize = sizeof(essServiceStatus);
    DWORD   dwBytesNeeded      = 0;
    DWORD   dwServicesReturned = 0;
    DWORD   dwResumeHandle     = 0;
    DWORD   dwI                = 0;
    BOOLEAN bFound = FALSE;
    SC_HANDLE schService, schSCManager;
    int ret;
    char buff[300];
    SERVICE_STATUS status_info;

    schSCManager = open_manager_channel(program_name);

    if ( !MSNT_EnumServicesStatus(schSCManager,
                            SERVICE_WIN32,
                            SERVICE_ACTIVE,
                            (LPENUM_SERVICE_STATUS)&essServiceStatus,
                            dwBufSize,
                            &dwBytesNeeded,
                            &dwServicesReturned,
                            &dwResumeHandle) )
        { 
        MSNT_CloseServiceHandle(schSCManager);
        ret = MS_GetLastError();
        if ( !force_quiet )
            format_post_error( ret, program_name );
        }

    for ( dwI=0; dwI<dwServicesReturned; dwI++ )
        { 
        if ( STRIEQ(essServiceStatus[dwI].lpServiceName,program_name) )
            { 
            bFound = TRUE;
            break;
            }
        }

    if ( bFound && !force_quiet )
        { 
        sprintf(buff,
                 "The %s service cannot be removed until it has been stopped.",
                 program_name);
        mess_out( buff, program_name );
        MSNT_CloseServiceHandle(schSCManager);
        return( FALSE );
        }

    schService = MSNT_OpenService(schSCManager,program_name,SERVICE_ALL_ACCESS);
    if ( NULL == schService )
        {
        ret = MS_GetLastError();
        if ( !force_quiet )
            format_post_error ( ret, program_name );
        MSNT_CloseServiceHandle(schSCManager);
        return( FALSE );
        }

    if ( bFound )
        {
        /* the service is currently running, so stop it. */
        MSNT_ControlService(schService, SERVICE_CONTROL_STOP, &status_info);
        }

    ret = TRUE;       /* assume good return */
    if ( MSNT_DeleteService(schService) )
        { 
        if ( !force_quiet )
            {
            sprintf(buff,"\"%s\" Service Deleted", program_name );
            mess_out( buff, program_name );
            ret = FALSE;
            }
        }
    else
        {
        ret = MS_GetLastError();
        if ( !force_quiet )
            format_post_error ( ret, program_name );
        ret = FALSE;
        }

    MSNT_CloseServiceHandle(schService);
    MSNT_CloseServiceHandle(schSCManager);
    return(ret);
}

