/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elbapps.c
   Applix ASC version: 36
   Date: Tue Feb  2 15:21:19 1999
*/
#define BOOL_DEFINED 1

#include <stdlib.h>

#if USL54
#include <afile.h>
#endif
#include <elbltn.h>

#include <aos.h>
#include <atm.h>
#include <afileimg.h>
#include <aiconids.h>

extern char *DirSlash;

/*************************************************************************
function: ElfbEditIcon - Start bitmap/graphics editor on an icon image
*/
ElfbEditIcon(nameP, widP, hytP, depP, titleP, readP, jnkP)
data_ptr nameP; /* pathname */
data_ptr widP;   /* width */
data_ptr hytP;   /* height */
data_ptr depP;   /* depth */
data_ptr titleP;  /* title */
data_ptr readP;  /* read */
data_ptr jnkP;  /* junk */
{
	char cmd[STRPATH], buf[512], pathname[STRPATH], *title, *str;
	int len, i, wid, hyt, dep, uid, doctype, on_disk, err, read_only;

	/* get the filename to edit */
	if ( nameP && AxIsString(nameP) )
		strcpy(pathname , AxStrFromDataPtr(nameP));
	else
		pathname[0] = (char)0;

	if ( strlen(pathname) <= 0 )
		ElfStrAbort(FEBadName, NULL, NULL);

	/* valid path ? */
	err = IsValidPathFilename(pathname);
	if (err != AOK )
		ElfStrAbort(err, NULL, NULL);

	/* get the image info(optional) */
	wid = widP ? AxIntFromDataPtr(widP) : 10;
	hyt = hytP ? AxIntFromDataPtr(hytP) : 10;
	dep = depP ? AxIntFromDataPtr(depP) : 1;

	if ( readP )
		read_only = AxIntFromDataPtr(readP) ? TRUE : FALSE;
	else
		read_only = FALSE;

	if ( wid == 0 || hyt == 0 || dep == 0 )
		ElfStrAbort(ERR_ILLVAL, XLT("Bad image data"), NULL);

	/* get the title to display */
	if ( titleP && AxIsString(titleP) )
		str = (char *)AxStrFromDataPtr(titleP);
	else
		str = (char *)0;
	if ( str )
	{
		len = strlen(str)+1;
		title = (char *)TrueMalloc(len);
		for ( i=0; i< len; i++ )
			if ( str[i] == ' ' )
				str[i] = '_';
		strcpy(title, str);
	}
	else
		title = (char *)0;

	/* existing doc  ?? */
    on_disk = Aaccess(pathname, F_OK) == 0 ? TRUE : FALSE;

	/* determine the file type passed in (.im or .ag only!) */
	doctype = DocTypeFromName(pathname);

	/* then open the file */
	switch ( doctype )
	{
		case MDrawing: /*  graphics editor */
			if ( on_disk ) /* existing ag file */
			{
				if ( title )
					sprintf(cmd, NXLT("\"-v%s\" -N%s"), pathname, title);
				else
					sprintf(cmd, NXLT("\"-v%s\""), pathname);
				if ( read_only )
				{
					sprintf(buf, NXLT(" -q"));
					strcat(cmd, buf);
				}
			}
			else /* need to create */
			{
				if ( title )
					sprintf(cmd, NXLT("\"-c%s\" -N%s"), pathname, title);
				else
					sprintf(cmd, NXLT("\"-c%s\""), pathname);
			}
		break;
		case MBitmap: /* bitmap editor */
			if ( on_disk ) /* existing im file */
			{
				if ( title )
					sprintf(cmd, NXLT("\"-v%s\" -N%s -R"), pathname,title);
				else
					sprintf(cmd, NXLT("\"-v%s\" -R"), pathname);
				if ( read_only )
				{
					sprintf(buf, NXLT(" -q"));
					strcat(cmd, buf);
				}
			}
			else /* need to create */
			{
				if ( title )
					sprintf(cmd, NXLT("\"-c%s\"  -N%s  -I -Ih %d -Iw %d -Id %d"),
				             pathname,title, hyt,wid,dep);
				else
					sprintf(cmd, NXLT("\"-c%s\"  -I -Ih %d -Iw %d -Id %d"),
				             pathname,hyt,wid,dep);
			}
		break;
		default:
			ElfStrAbort(FEnotDoc, NULL, NULL);
		break;
	}
	uid = ACpproc(APGE, cmd);

	/* done */
	if ( title )
		TrueFree(title);
	return(uid);
}

/*************************************************************************
function: ElfbGEQuietOpen - Open a GE windowless & return pid
*/
#ifdef AX_USE_PROTOS
ElfbGEQuietOpen( )
#else
ElfbGEQuietOpen(name, parent_wid)
char *name;
int parent_wid;
#endif
{
    char opts[STRPATH + 20];
    int pid;

    sprintf(opts, NXLT("-pA -n -v%s"), name);
    opts[2] += parent_wid;
    pid = ElfbProc(APGE, opts);
    return (pid);
}

/*************************************************************************
function: ElfbGEQuietCreate - Open a GE windowless & return pid
*/
#ifdef AX_USE_PROTOS
ElfbGEQuietCreate( )
#else
ElfbGEQuietCreate(name, parent_wid)
char *name;
int parent_wid;
#endif
{
    char opts[STRPATH + 20];
    int pid;

    sprintf(opts, NXLT("-pA -n -c%s"), name);
    opts[2] += parent_wid;
    pid = ElfbProc(APGE, opts);
    return (pid);
}

/*************************************************************************
macro: ElfbUpdateBitmap - re-read the named pixmap
*/
#ifdef AX_USE_PROTOS
ElfbUpdateBitmap( char *name  )
#else
ElfbUpdateBitmap(name)
    char *name;
#endif
{
#if !FOR_TESTER
    AxUpdatePixCache(name);
#endif
}

/*************************************************************************
function: ElfbImageFromGE - Produce an Icon image from a GE file
This gets the first image found in the GE file.
*/
#ifdef AX_USE_PROTOS
ElfbImageFromGE( char *imgfile, char *gefile )
#else
ElfbImageFromGE(imgfile, gefile)
    char *imgfile, *gefile;
#endif
{
#if !FOR_TESTER
    char buf[STRPATH];

    sprintf(buf,NXLT("%s%sfilters%sgeinout %s %s -t %s -s %s -o im"),ThimApplixDir,DirSlash,DirSlash, gefile,imgfile,ThimTempDir,ThimApplixDir);
    if ( 0 != AxRunProgram(buf,NULL,NULL,NULL,FALSE) )
        ElfAbort(ERR_NOIMAGE);
#endif
}

/*************************************************************************
function: ElfbImportSS - Kick off spreadsheet
*/
#ifdef AX_USE_PROTOS
ElfbImportSS( char *opts  )
#else
ElfbImportSS(opts)
    char *opts;
#endif
{
    int code;

    code = ACpproc(APSS, opts);
    VERIFY(code == 0, ERR_ILLVAL);
}


/*************************************************************************
function: ElfbLocalizeSS - Localize spreadsheet links
*/
#ifdef AX_USE_PROTOS
ElfbLocalizeSS( char *Source, char *Dest )
#else
ElfbLocalizeSS(Source, Dest)
    char *Source, *Dest;
#endif
{
    char buf[1000];
    int code;

    sprintf(buf, NXLT("cp %s %s"), Source, Dest);
    system(buf);

    sprintf(buf, NXLT("-v%s  -L -m -q"), Dest);
    code = ACpproc(APSS, buf);
    VERIFY(code == 0, ERR_ILLVAL);
}


/*************************************************************************
function: ElfbImportAscii - Load ascii file to DC mash format
*/
#ifdef AX_USE_PROTOS
ElfbImportAscii( char *opts  )
#else
ElfbImportAscii(opts)
    char *opts;
#endif
{
    int code;

    code = ACpproc(APIDC, opts);
    if (code && (code != 4))
        SuperMessage(XLT("!Import failed."));
}


/*************************************************************************
function: writeImageSegment - create a graphics segment containing an image
*/
#ifdef AX_USE_PROTOS
writeImageSegment(
    char *docname,
    unsigned char *buf,
    unsigned int height, 
    unsigned int width,
    unsigned int depth,
    int inBytewidth,
    unsigned char *map)
#else
writeImageSegment(docname, buf, height, width, depth, inBytewidth, map)
    char *docname;
    unsigned char *buf;
    unsigned int height, width, depth;
    int inBytewidth;
    unsigned char *map;
#endif
{
}
