/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elbatm.c
   Applix ASC version: 37
   Date: Sat Feb 20 21:13:14 1999
*/
#include <stdlib.h>
#include <elbltn.h>
#include <axtext.h>
#include <java_syms.h>
#include <elbatm.p>


/*************************************************************************
function: ElfbXwinOfWin - Return X-wins of shell and work area
*/
#ifdef AX_USE_PROTOS
ElfbXwinOfWin( int win         /* ATM window id */ )
#else
ElfbXwinOfWin(win)
    int win;        /* ATM window id */
#endif
{
#if !FOR_TESTER && !AX_STUB /* !!cvt AX_STUB from 3.0 */
    ElfRetData = MkArray(2);

    VERIFY((win < TM_MAXWIN) && (win > 0) && (window[win].busy_flag), ERR_NOWINDOW);
    ElfAddDblToArray(ElfRetData, 0, (double) window[win].xwin);
    ElfAddDblToArray(ElfRetData, 0, (double) window[win].topwindow);
#endif              /* !!cvt */
}

/*************************************************************************
function: ElfbSelByXId - Select window from X id
*/
#ifdef AX_USE_PROTOS
ElfbSelByXId(
    long xid,
    long grab  )
#else
ElfbSelByXId(xid, grab)
    long xid;
    long grab;
#endif
{
#if !FOR_TESTER && !AX_STUB /* !!cvt AX_STUB from 3.0 */
    int ix;

    for (ix = 1; ix < TM_MAXWIN; ix++)
        {
        if (window[ix].busy_flag &&
          ((window[ix].xwin == xid) || (window[ix].topwindow == xid)))
            {
            ElfbMkWinActive(ix, grab);
            break;
            }
        }
#endif              /* !!cvt */
}

/*************************************************************************
function: ElfbPrivateMenuBar - Assign a menu bar to this window
*/
ElfbPrivateMenuBar(type)
int type;
{
#if !FOR_TESTER && !AX_STUB /* !!cvt AX_STUB from 3.0 */

	if ( TaskIsJavaBased(THIMpid))
	{
		data_ptr infoP, menuP, aP;
		char name[100];
		int pid;

		sprintf(name,"menubar$vals$%d",type);
		menuP = (data_ptr)ElfGetSysVar(name);
		if ( menuP != (data_ptr)NULL)
		{
			aP = AxMakeArray(1);
			AxAddDataToArray(aP, 0, menuP);
			AxfJavaMenubarNotify(aP);
			infoP = AxMakeArray(1);
			AxAddIntToArray(infoP, 0, type);
			pid = ThimTopPidFromPid(THIMpid);
			AxJavaPidNotifier(pid,JV_SET_MENU_BAR,infoP);
			if ( AxIsGraphicsTask(pid))
				GEUpdateMenuBarInfo(type);
			else if ( AxIsWPTask(pid))
				WPUpdateMenuBarInfo(type);
			else if ( AxIsSSTask(pid))
				SSUpdateMenuBarInfo(type);
		}
	}
	else
	{
		VERIFY((cur_win != TM_NULWIN) && (window[cur_win].busy_flag), ERR_NOWINDOW);
		window[cur_win].privateMenuBar = type;
		CreateOrChangeMenuBar(WidgetOfWindow(cur_win));

		/* Poke all dialogs */
		{
			char buf[20];
			sprintf(buf, "%d", type);
			TmPokeDialogs(0, POKE_MENU_BAR, buf);
        }
	}
#endif
}

/*************************************************************************
function: ElfFindSlider - Look up an Elf specified selection line
In the future, this may be able to distinguish between a window having
its own menu bar (from doc file).
*/
#ifdef AX_USE_PROTOS
menu_item_def *ElfFindSlider( struct wind_struct *wp  )
#else
menu_item_def *ElfFindSlider(wp)
    register struct wind_struct *wp;
#endif
{
    menu_item_def *mp = NULL;

    if (wp->privateMenuBar)
        mp = ElfLookupSlider(wp->privateMenuBar);
    if (!mp)
        mp = ElfLookupSlider(wp->wind_attr.menubar_id);
    return (mp);
}

/*************************************************************************
function: ElfFindMenuBar - ElfFindSlider wrapper, passed atm window id
*/
#ifdef AX_USE_PROTOS
menu_item_def *ElfFindMenuBar( int window_id  )
#else
menu_item_def *ElfFindMenuBar(window_id)
register int window_id;
#endif
{
    int good_id;

    good_id = (window_id < TM_MAXWIN) && (window_id > 0) &&
        (window[window_id].busy_flag);

    VERIFY(good_id, ERR_NOWINDOW);

    return( ElfFindSlider(&window[window_id]) );
}

ElfbFuncKey(){}     /* OBSOLETE */


/*************************************************************************
function: ElfbWindowList - Return array of windows' state
This used to have an "all" flag which was not documented... removed by zane
*/
void ElfbWindowList()
{
    register data_ptr aP, saP;
    register int wid, ix = 0;
    data_ptr Elf1WindowInfo(), axmDialogsList();

    aP = MkArray(0);
    for (wid = 0; wid < TM_MAXWIN; wid++) {
        saP = Elf1WindowInfo(wid);
        if (saP != NULL)
            aP = WriteArray(aP, ix++, saP);
    }
    ElfRetData = aP;
}

/*************************************************************************
function: ElfbListOfWindows - Return complete list of windows... UID based
*/
void ElfbListOfWindows ()
{
    register data_ptr aP, saP, dlgs;
    register int wid, ix = 0, numDlgs;
    data_ptr AxDialogsList();

    aP = MkArray(0);

/* C-based windows first */

    for (wid = 1; wid < TM_MAXWIN; wid++) {
        if ( !window[wid].busy_flag ) continue;
        saP = MkArray(5);
        ElfAddNumToArray(saP, 0, window[wid].widget_uid);
        ElfAddStrToArray(saP, 1, window[wid].wind_attr.title_line);
        ElfAddNumToArray(saP, 2, window[wid].wind_attr.icon_id);
        ElfAddNumToArray(saP, 3, TaskInfo[window[wid].p_id].uid);
        ElfAddNumToArray(saP, 4, window[wid].wind_attr.menubar_id);
        aP = WriteArray(aP, ix++, saP);
    }

/* Now Elf based */

    dlgs = AxDialogsList();
    numDlgs = AxArraySize(dlgs);
    for (wid = 0; wid < numDlgs; wid++, ix++)
        aP = WriteArray(aP, ix, ElfDataCopy(AxArrayElement(dlgs, wid)));
    AxFreeData(dlgs);
    ElfRetData = aP;
}

/*************************************************************************
function: ElfbMkWinActive - Make a given window id the active window
This code copied from tmwin.c -> atm_mk_curwin
*/
ElfbMkWinActive(cw, grab)
register int cw;
{
    int ix;

	if (TaskIsJavaBased(THIMpid) && cw <= 0) {
		return; /* currently, we don't use cur_win, so select_window@ can
				   get -1 as the current window from current_window_num@ */
	}

	VERIFY(cw > 0, ERR_NOWINDOW);
    for ( ix = 0; ix < TM_MAXWIN; ix++ ) {
        if ( window[ix].busy_flag && (window[ix].widget_uid == cw) )
            break;
    }
    if ( ix == TM_MAXWIN ) {
        if ( (cw < TM_MAXWIN) && window[cw].busy_flag )
            make_curwin(cw);
        else
            goto try_to_select_elf_app;
    }
    else
        make_curwin(ix);
    if ( grab )       /* Grab now optional are requested by Raytheon */
        GrabCurWin();
    if ((cur_win > 0) && window[cur_win].busy_flag)
        AxActiveTask = TaskInfo[window[cur_win].p_id].uid;
    AxSetTaskMaster(THIMpid,AxActiveTask);
    return;

try_to_select_elf_app:          /* !!0 TODO */
    if ( !AxmSelectWinByWinUid(cw,grab) )
        ElfAbort(ERR_NOWINDOW); 
}


/*************************************************************************
function: ElfbWinTitle - Title of current window or "" if nulwin
*/
ElfbWinTitle()
{
#if !FOR_TESTER && !AX_STUB /* !!cvt from 3.0 */
    register char *p;

    if (cur_win == TM_NULWIN)
        p = "";
    else
        p = window[cur_win].wind_attr.title_line;
    strcpy(ElfStrPtrRet = ElfMalloc(strlen(p) + 1), p);
#endif              /* !!cvt from 3.0 */
}

/*************************************************************************
function: ElfbWindowInfo - Return array of info about a given window
*/
ElfbWindowInfo( int wid_in  )
{
	int wid=wid_in;
    data_ptr Elf1WindowInfo();

	/* Convert large window id to small, if necessary */
	if (wid > TM_MAXWIN) {
		wid = AxFindSmallWindowId(wid);
	}
    VERIFY(wid != TM_NULWIN && winExists(wid), ERR_NOWINDOW);
    ElfRetData = Elf1WindowInfo(wid);   /* Alis func generic return value */
}

/*************************************************************************
function: Elf1WindowInfo - Return array describing a window
*/
data_ptr Elf1WindowInfo(wid)
register int wid;
{
    register struct wind_struct *p;
    register data_ptr saP;
    data_ptr eb1AtmWinInfo();

    p = window + wid;
    if (!p->busy_flag || !(TaskInfo[p->p_id].in_use)) {
        return (NullDataPtr);
    }
    saP = MkArray(5);
    ElfAddNumToArray(saP, 0, wid);
    ElfAddStrToArray(saP, 1, p->wind_attr.title_line);
    ElfAddNumToArray(saP, 2, p->wind_attr.icon_id);
    ElfAddNumToArray(saP, 3, TaskInfo[p->p_id].uid);
    ElfAddNumToArray(saP, 4, p->wind_attr.menubar_id);
    return (saP);
}


/****************************************************************************/
/* Return list of available printer types.  The return value is through the */
/*    "ElfRetData" global                                                   */
/****************************************************************************/
void ElfbGetPrinterTypes()
{

    data_ptr        aP = NULL;

    aP = ELF_CREATE_ARRAY(0);
    aP = ElfAddStrToArray(aP, 0, NXLT("PostScript") );
    aP = ElfAddStrToArray(aP, 0, NXLT("PCL5") );
#if BUILD_ORIENT
    aP = ElfAddStrToArray(aP, 0, NXLT("LIPS3") );
#endif

    ElfRetData = aP;
    return;
}




/*************************************************************************
function: ElfbCurWindow - Return current window id
*/
ElfbCurWindow()
{
    extern short cur_win;

    return (cur_win);
}

/***************************************************************************
function: ElfbGetXLFD - get nearest X font match for font 
*/
ElfbGetXLFD(familyP,boldP,italicP,sizeP)
data_ptr familyP;
data_ptr boldP;
data_ptr italicP;
data_ptr sizeP;
{
    mbchar *family, *xlfd, *screenFamily;
    int bold, italic, size, exact, freeXLFD;
	fontTabEnt *screenRec, *targetRec;
	char *err;

    family = AxStrPtrFromDataPtr(familyP);
    bold = (AxIntFromDataPtr(boldP) != 0);
    italic = (AxIntFromDataPtr(italicP) != 0);
    size = AxIntFromDataPtr(sizeP);
	
	if (err = axGetFontRecordForScreen(family, bold, italic, size,
									   &targetRec, &screenRec)) {
		freeErrString(err);
		screenFamily = family;
	}
	else {
		screenFamily = screenRec->familyName;
	}
	
#if AX_X11
 	xlfd = (mbchar *) axGetXLFD(screenFamily, bold, italic, size,
								&exact, &freeXLFD);
#else
    /* Windows-NT */
    xlfd = axMakeCompoundWinName(screenFamily, size, bold, italic);
	freeXLFD = TRUE;
#endif
	
    ElfRetData = AxMakeStrData(-1, xlfd);
	if (freeXLFD) {
		axtFreeXLFD(xlfd);
	}
}

/*************************************************************************
function: ElfbPromoteWindow - Make the window associated with uid toplevel
*/
ElfbPromoteWindow(uid)
register int uid;
{
	int ix, pid;
    register struct wind_struct *wp;

	pid = ThimTopPidFromPid(ThimPidFromUID(uid));
	if ( TaskIsJavaBased(pid))
	{
		data_ptr infoP;

		infoP = AxMakeArray(1);
		AxAddIntToArray(infoP, 0, AxJavaGetCurwinId(uid));
		AxJavaPidNotifier(pid,JV_PROMOTE_WINDOW,infoP);
	}
	else
	{
		for ( ix = 0; ix < TM_MAXWIN; ix++ )
		{
			if ( window[ix].busy_flag && (window[ix].p_id == pid) )
			{
				wp = &window[ix];
				TmRaiseWindow(wp);
				break;
			}
		}
	}
}

/*************************************************************************
function: ElfbFindWindow - Given a pid, find the matching window.
Returns the small window id, or -1 if no window belongs to the passed pid
*/
int
ElfbFindWindow(pid)
register int pid;
{
	int ix;

	/* C (application) windows only */
	for ( ix = 0; ix < TM_MAXWIN; ix++ ) {
		if ( window[ix].busy_flag && (window[ix].p_id == pid) )
			return( ix);
	}
	return( -1);
}

/*************************************************************************
function: ElfbFindDialog - Given a pid, find the matching window.
Returns the large dlg window id, or -1 if no dialog belongs to the passed pid
*/
int
ElfbFindDialog(pid)
register int pid;
{
    data_ptr aP, dlgs;
    int uid, wid, ix, numDlgs;
    data_ptr AxDialogsList();

	/* ELF windows (dboxes) only */
    dlgs = AxDialogsList();
    numDlgs = AxArraySize(dlgs);
    for (ix = 0; ix < numDlgs; ix++) {
        aP = AxArrayElement(dlgs, ix);
		uid = AxIntFromArray(aP, 3);
		if (uid && (uid == TaskInfo[pid].uid)) {
			wid = AxIntFromArray(aP, 0);
    		AxFreeData(dlgs);
			return(wid);
		}
	}
    AxFreeData(dlgs);
	return( -1);
}

/*
 * Window id's come in 2 flavors: small and large.
 *
 * The small window id's run from 0 to TM_MAXWIN (currently 21).
 *	The numbers are reused as slots open up (when old windows are 
 *	shut down). This id is used as the index into the 'window' array.
 *
 * The large window id's run upwards from 100001; they are not reused.
 *	They are stored in the widget_uid member of the widget structure.
 *
 * The relationship between the id's is:
 *			window[smallId].widget_uid = largeId
 */

/*************************************************************************
function: AxfFindLargeWindowId - Given a small window id, return the
window's large id.
Returns -1 if no window has the passed small id
*/

data_ptr AxfFindLargeWindowId(data_ptr argDatum){
	register int smallId = AxIntFromArray(argDatum,0);
	return(MK_INTEGER(AxFindLargeWindowId(smallId)));
}
int AxFindLargeWindowId(int smallId) {
	VERIFY( (smallId >= 0) && (smallId < TM_MAXWIN), ERR_NOWINDOW);
	if ( window[smallId].busy_flag)
		return(window[smallId].widget_uid);
	return(-1);
}


/*************************************************************************
function: AxfFindSmallWindowId - Given a large window id, return the
window's small id.
Returns -1 if no window has the passed large id
*/
data_ptr AxfFindSmallWindowId(data_ptr argDatum) {
	register int largeId = AxIntFromArray(argDatum,0);
	return(MK_INTEGER(AxFindSmallWindowId(largeId)));
}
int AxFindSmallWindowId(int largeId) {
	register int ix;
	for ( ix = 0; ix < TM_MAXWIN; ix++ ) {
		if ( window[ix].busy_flag && (window[ix].widget_uid == largeId) )
			return(ix);
	}
	return(-1);
}
