/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elccc.c
   Applix ASC version: 30
   Date: Tue Feb  2 15:21:06 1999
*/
#include <elcomp.h>
#include <eltokens.h>
#include <elnodes.h>
#include <elcodes.h>
#include <sqlops.h>
#include <getstr4.h>
#include <fsrts.h>
#include <elf_env.h>
#include <eldata.h>
#if HP800V3
#include <time.h> /* for gettimeofday() */
#endif

#define TESTING 0	/* set this to emit to stdout... with COMMNENTS */

data_ptr ElfcPackDatum();

int InText = 0;	/* tracks which segment we are putting to */
int NeedEnd = 0;
int UID;	/* counter for labels */


#define USE_COMMENTS TRUE

#if TESTING
#	define AsmFP stdout
#else
	extern FILE *AsmFP;
#endif

#if USE_COMMENTS
#	define EMIT_COMMENT(str) fprintf(AsmFP,"/* %s */\n",str);
#else
#	define EMIT_COMMENT(str)
#endif

#if HP800V3
static int compute_uid();
#endif

#if HP800V3
/***********************************************************************
function: compute_uid - 
*/
static int
compute_uid()
{
	int uid = 0;
	struct timeval t;
	struct timezone tz;
	unsigned long sec;
	unsigned long usec;

	t.tv_sec = 0;
	t.tv_usec = 0;
	gettimeofday(&t, &tz);
	sec = (unsigned long) t.tv_sec;
	usec = (unsigned long) t.tv_usec;
	sec  = (unsigned long) (sec & 0x0000ffff);
	usec = (unsigned long) (usec << 16);
	usec = (unsigned long) (usec & 0xffff0000);
	uid  = (int) (sec | usec);
	uid  = uid < 0 ? -uid : uid;
	return(uid);
}
#endif

/*************************************************************************
function: EmitBytes - Emits Any number of bytes
Attempts to load 16 bytes per assembly source line.
*/
EmitBytes(buf,cnt)
char *buf;
int cnt;
{
	int ix, dex;

	dex = 0;
	while ( dex < cnt ) {
		fprintf(AsmFP,"\t");
		for ( ix = 0; (ix < 16) && (dex < cnt); ix++,dex++ ) {
			if ( dex == cnt-1 )
				fprintf(AsmFP,"%d",buf[dex]);
			else
				fprintf(AsmFP,"%d,",buf[dex]);
		}
		fprintf(AsmFP,"\n");
	}
}
/*************************************************************************
function: EmitShrts - Emits Any number of bytes
Attempts to load 16 bytes per assembly source line.
*/
EmitShrts(buf,cnt)
shrt *buf;
int cnt;
{
	int ix, dex;

	dex = 0;
	while ( dex < cnt ) {
		for ( ix = 0; (ix < 16) && (dex < cnt); ix++,dex++ ) {
			if ( dex == cnt-1 )
				fprintf(AsmFP,"%d",buf[dex]);
			else
				fprintf(AsmFP,"%d,",buf[dex]);
		}
		fprintf(AsmFP,"\n");
	}
}

/*************************************************************************
function: EmitInts - Emits Any number of bytes
Attempts to load 16 bytes per assembly source line.
*/
EmitInts(buf,cnt,cont_flag)
int *buf;
int cnt;
int cont_flag;
{
	int ix, dex;

	dex = 0;
	while ( dex < cnt ) {
		fprintf(AsmFP,"\t");
		for ( ix = 0; (ix < 8) && (dex < cnt); ix++,dex++ ) {
			if ( dex == cnt-1 ) 
				fprintf(AsmFP,"%d",buf[dex]);
			else
				fprintf(AsmFP,"%d,",buf[dex]);
		}
		if(dex == cnt && cont_flag)
			fprintf(AsmFP,",\n");
		else
			fprintf(AsmFP,"\n");
	}
}

static int CodeStartsInText;

/*************************************************************************
function: ElfcAsmGen - Generate assembly code
*/
ElfcAsmGen (allP)
data_ptr allP;
{
	data_ptr headerP, gvarsP, ovarsP, svarsP, gfuncsP, litsP, extsP, codeP, varnamesP, formatsP;
	int codeStart, codeSize;
	int gvStart, ovStart, svStart, gfStart, extStart, litsStart, formatsStart, varsStart;
	char *name;
	char nameBuf[STRLONG+1];
        int ix;
	int nameid, pathid;
	int trueInt = TRUE, falseInt = FALSE, zeroInt = 0;
	int numLiterals;
	char *path;
	char *ComputeName();

	allP = ElfcPackDatum(allP);
	headerP = ReadArray(allP,0);
	gvarsP = ReadArray(allP,1);
	gfuncsP = ReadArray(allP,2);
	extsP = ReadArray(allP,3);
	litsP = ReadArray(allP,4);
	codeP = ReadArray(allP,5);
	varnamesP = ReadArray(allP,6);
	formatsP = ReadArray(allP,7);
	ovarsP = ReadArray(allP,8);
	svarsP = ReadArray(allP,9);

	name = ComputeName(headerP);
#if AXMS    /* name read in uppercase, change to lower */
        for (ix=0; *name; name++,ix++)
            *name = tolower(*name);
        name = name - ix;
#endif
	sprintf(nameBuf,"ELF_BIN_%s",name);

	fprintf(AsmFP,"#include <aos.h>\n");
#if (M_88K && !AVIION) 
	fprintf(AsmFP,"#define CONST\n");
#else
	fprintf(AsmFP,"#define CONST const\n");
#endif
	fprintf(AsmFP,"#include <elf.h>\n");

    /* on some machines NULL is defined as either "(char *)0" or   */
    /*  as "(void *)0" which will generate a lot of warnings  */
    /* when used as an initializer by the code we output here when using an */
    /*  ansi compiler, so set NULL back to a plain old zero          */
    fprintf(AsmFP,"#ifdef NULL\n");
    fprintf(AsmFP,"#undef NULL\n");
    fprintf(AsmFP,"#endif\n");
    fprintf(AsmFP,"#define NULL 0\n");


/*
	fprintf(AsmFP,"#define TRUE 1\n");
	fprintf(AsmFP,"#define FALSE 0\n");
	fprintf(AsmFP,"#define NULL 0\n");
*/
#if HP800V3
    /* Since on HP we can not have "data_number" types be 'static',   */
    /*  (see comments at "%g%" below), we must generate unique labels */
    /*  across all of the elf code.  To do this we use the time in seconds */
	/*    UID = time();  -- produces symbol redefinitions. */
	UID = compute_uid();
#endif

	PutCode(codeP,&codeStart,&codeSize); 
	CodeStartsInText = codeStart;
	PutGvars(gvarsP,&gvStart); 
	PutGfuncs(gfuncsP,&gfStart); 
	PutExts(extsP,&extStart);
	PutLiterals(litsP,&litsStart);
	numLiterals = AX(litsP)->size;
	PutFormats(formatsP,&formatsStart);
	PutVars(varnamesP,&varsStart);
	PutOvars(ovarsP,&ovStart); 
	PutSvars(svarsP,&svStart); 

/* Produce Package Structure */

	EMIT_COMMENT("Module Name");
	EmitNamedStr(nameid = ++UID,name);

	EMIT_COMMENT("Source File Name");
	path = StrFromArray(headerP,1);
	EmitNamedStr(pathid = ++UID,path);

	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("PACKAGE STRUCT");

    fprintf(AsmFP, "#ifdef __cplusplus\n");
	fprintf(AsmFP,"extern \"C\" {\n");
    fprintf(AsmFP, "#endif\n");
	fprintf(AsmFP,"extern CONST package_info %s;\n",nameBuf);
	fprintf(AsmFP,"CONST package_info %s = {\n",nameBuf);
	fprintf(AsmFP,"\t(shrt*)L%d,\t\t/* shrt *code_starts */ \n",codeStart);
	fprintf(AsmFP,"\t(shrt*)(L%d+%d),\t\t/* shrt *code_ends */\n",
                                             codeStart, (codeSize-1) );
	fprintf(AsmFP,"\t(gvar_info**)L%d,\t\t/* gvar_info **gvar_list */\n",gvStart);
	fprintf(AsmFP,"\t(gfunc_info**)L%d,\t\t/* gfunc_info **gfun_list */\n",gfStart);
	fprintf(AsmFP,"\t(ext_link**)L%d,\t\t/* ext_link_for_elccc **ext_list */\n",extStart);		
	fprintf(AsmFP,"\t(data_ptr*)L%d,\t\t/* data_ptr *literals_table */\n",litsStart);		
	fprintf(AsmFP,"\t(gvar_info**)L%d,\t\t/* gvar_info **ovar_list */\n",ovStart);
	fprintf(AsmFP,"\t(gvar_info**)L%d,\t\t/* gvar_info **svar_list */\n",svStart);
	fprintf(AsmFP,"\t(char *)L%d,\t\t/* char *varNames */\n",varsStart);		
	fprintf(AsmFP,"\t(data_ptr)&L%d,\t\t/* data_ptr formats */\n",formatsStart);	
	fprintf(AsmFP,"\t(char*)L%d,\t\t/* char *moduleName */\n",nameid);			
	fprintf(AsmFP,"\t(char*)L%d,\t\t/* char *source_name */\n",pathid);			
#if BLD_NO_ELF_DEBUG	/* set for turbo builds */
	fprintf(AsmFP,"\tFALSE,\t\t/* int src_map */\n");
#else
	fprintf(AsmFP,"\tTRUE,\t\t/* int src_map */\n");
#endif
	fprintf(AsmFP,"\tFALSE,\t\t/* int dynamic_load */\n");
	fprintf(AsmFP,"\t0,\t\t/* int task_uid */\n");
	fprintf(AsmFP,"\t%d }; /* int num_literals */\n",numLiterals);
    fprintf(AsmFP, "#ifdef __cplusplus\n");
	fprintf(AsmFP,"}\n");
    fprintf(AsmFP, "#endif\n");
	EMIT_COMMENT("END PROGRAM");
	EMIT_COMMENT("");
	EMIT_COMMENT("");

#if !TESTING
	if ( 0 != fclose(stdout) ) {
		printf("Could not write ouput files!!!\n");
		abort();
	}
#endif
}

/*************************************************************************
function: EmitNamedStr - Put out an ascii string
*/
EmitNamedStr (id,str)
int id;
char *str;
{
	int len = strlen(str) + 1;
	int ix;

#if USE_COMMENTS
	fprintf(AsmFP,"/* %s */\t\t",str);
#endif
	fprintf(AsmFP,"CONST static char L%d[%d] = {",id,len);
	for ( ix = 0; ix < len-1; ix++ )
		fprintf(AsmFP,"%d,",str[ix] & 0xFF);
	fprintf(AsmFP,"0};\n");
#if USE_COMMENTS
	fprintf(AsmFP,"\n",str);
#endif
}

/*************************************************************************
function: EmitStringTable - Put out a string table from <name_ind> index of each
element of an array
*/
EmitStringTable(id, num_strs, tblP, name_ind)
int id;
int num_strs;
data_ptr tblP;
int name_ind; /* index of the name in the str array */
{

	int ix, jx, len;
	data_ptr strP;
	char* name;
	int array_defined=0; /* to handle empty tables */
	
	for ( ix = 0; ix < num_strs; ix++ ) {
		strP = ReadArray(tblP,ix);
		name = StrFromArray(strP,name_ind);
		if(name[0] == '\0')
			continue;
		if(!array_defined) {
	      fprintf(AsmFP,"CONST static char L%d[] = {\n",id);
		  array_defined = TRUE;
	    }
#if USE_COMMENTS
		fprintf(AsmFP,"/* %s */\t\t",name);
#endif
		len = strlen(name);
		for ( jx = 0; jx < len; jx++)
			fprintf(AsmFP,"%d,", name[jx]&0xFF);
		fprintf(AsmFP,"0,\n");
#if USE_COMMENTS
		fprintf(AsmFP,"\n");
#endif
	}
	if(array_defined)
	   fprintf(AsmFP,"};\n");
}

/*************************************************************************
function: ComputeName - Compute module name from filename
module name from file name... clip off trailing dot & leading path
*/
char *ComputeName (headerP)
data_ptr headerP;
{
	static char filename[STRPATH+1];
	char *name;

	strcpy(filename,StrFromArray(headerP,1));
	name = strrchr(filename,'.');
	if ( name )
		*name = 0;
	name = strrchr(filename,'/');
	if ( !name )
		name = strrchr(filename,'\\');
	if ( name )
		name = name + 1;
	else
		name = filename;
	return(name);
}

/*************************************************************************
function: PutCode - Emit code frame
*/
PutCode (codeP,startP,sizeP)
data_ptr codeP;
int *startP, *sizeP;
{
	char *bytes = BX(codeP)->bytes;
	int size = BX(codeP)->size;
	int zero = 0;

	
	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("EXECUTBALE OPCODES");

	size = (size + sizeof(int) - 1) / sizeof(int);	/* int size, rnded up */
	fprintf(AsmFP,"static CONST int L%d[%d] = {\n", *startP = ++UID,size);
	EmitInts((int*)bytes,size,FALSE);
	fprintf(AsmFP,"};\n");
    *sizeP = size; 

	EMIT_COMMENT("END OPCODES");
	EMIT_COMMENT("");
	EMIT_COMMENT("");
}

/*************************************************************************
function: PutGvars - Put Globals list out
The Array of pointers is loaded into the TextFrame.
The actual structures are loaded into the DataFrame.
The names are loaded into the TextFrame.
*/
PutGvars (gvarsP,startP)
data_ptr gvarsP;
int *startP;
{
	int num_gvars;
	char *name, *format;
	int var_ids[10000];
	int names[10000], formats[10000];
	int zero = 0;
	int hashval, ix;
	data_ptr avarP;
	data_ptr *nullptr = NULL;
	int gvars_id, gvar_info_id, format_id;
	int varname_ind, format_ind;

/* Emit the gvar objects in the data segment */

	num_gvars = ElfbArraySize(gvarsP);
	assert(num_gvars<10000);

	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("GLOBAL VARIABLES");
	EMIT_COMMENT("");
	EMIT_COMMENT("");


	gvars_id = ++UID;
	EmitStringTable(gvars_id, num_gvars, gvarsP, 0); /* names */
	format_id = ++UID;
	EmitStringTable(format_id, num_gvars, gvarsP, 1); /* formats */

	gvar_info_id = ++UID;
	varname_ind = 0;
	format_ind = 0;
	if(num_gvars)
		fprintf(AsmFP,"static gvar_info L%d[%d] = {\n",gvar_info_id,num_gvars);
	for ( ix = 0; ix < num_gvars; ix++ ) {
		avarP = ReadArray(gvarsP,ix);
		name = StrFromArray(avarP,0);
		format = StrFromArray(avarP,1);
		hashval = hash(name);
		fprintf(AsmFP,"\t{\n\t0,\n");
		fprintf(AsmFP,"\t%d,\n",hashval);
		fprintf(AsmFP,"\t(L%d+%d),\n",gvars_id,varname_ind);
		varname_ind += strlen(name)+1;
		if(format[0] != '\0') {
			fprintf(AsmFP,"\t(L%d+%d)\n\t},\n",format_id,format_ind);
			format_ind += strlen(format)+1;
		}
		else
			fprintf(AsmFP,"\tNULL\n\t},\n");
	}
	if(num_gvars)
		fprintf(AsmFP,"};\n");

/* Emit the array of ptrs in the text segment */

	EMIT_COMMENT("GLOBALS TABLE");

	*startP = ++UID;
	fprintf(AsmFP,"static CONST gvar_info *L%d[%d] = {\n",*startP,num_gvars+1);

	for ( ix = 0; ix < num_gvars; ix++ )
		fprintf(AsmFP,"\t(L%d+%d),\n",gvar_info_id,ix);
	fprintf(AsmFP,"\tNULL\n};\n");

	EMIT_COMMENT("END GLOBALS INFO");
	EMIT_COMMENT("");
	EMIT_COMMENT("");
}

/*************************************************************************
function: PutOvars - Put Object Vars list out
The stuff is only needed by the debugger
For "laziness" reasons, this uses the gvar_info structure.
*/
PutOvars (ovarsP,startP)
data_ptr ovarsP;
int *startP;
{
	int num_ovars;
	char *name, *format;
	int zero = 0;
	int hashval, ix;
	data_ptr avarP;
	data_ptr *nullptr = NULL;
	int ovars_id, gvar_info_id, format_id;
	int varname_ind, format_ind;

/* Emit the ovar objects in the data segment */

	num_ovars = ElfbArraySize(ovarsP);
	assert(num_ovars<10000);

	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("OBJECTVAR VARIABLES");
	EMIT_COMMENT("");
	EMIT_COMMENT("");

	ovars_id = ++UID;
	EmitStringTable(ovars_id, num_ovars, ovarsP, 0); /* names */
	format_id = ++UID;
	EmitStringTable(format_id, num_ovars, ovarsP, 1); /* formats */

	gvar_info_id = ++UID;
	varname_ind = 0;
	format_ind = 0;
	if(num_ovars)
		fprintf(AsmFP,"static gvar_info L%d[%d] = {\n",gvar_info_id,num_ovars);
	for ( ix = 0; ix < num_ovars; ix++ ) {
		avarP = ReadArray(ovarsP,ix);
		name = StrFromArray(avarP,0);
		format = StrFromArray(avarP,1);
		hashval = hash(name);
		fprintf(AsmFP,"\t{\n\t0,\n");
		fprintf(AsmFP,"\t%d,\n",hashval);
		fprintf(AsmFP,"\t(L%d+%d),\n",ovars_id,varname_ind);
		varname_ind += strlen(name)+1;
		if(format[0] != '\0') {
			fprintf(AsmFP,"\t(L%d+%d)\n\t},\n",format_id,format_ind);
			format_ind += strlen(format)+1;
		}
		else
			fprintf(AsmFP,"\tNULL\n\t},\n");
	}
	if(num_ovars)
		fprintf(AsmFP,"};\n");

/* Emit the array of ptrs in the text segment */

	EMIT_COMMENT("OBJECTVARS TABLE");

	*startP = ++UID;
	fprintf(AsmFP,"static CONST gvar_info *L%d[%d] = {\n",*startP,num_ovars+1);

	for ( ix = 0; ix < num_ovars; ix++ )
		fprintf(AsmFP,"\t(L%d+%d),\n",gvar_info_id,ix);
	fprintf(AsmFP,"\tNULL\n};\n");

	EMIT_COMMENT("END OBJECTVARS INFO");
	EMIT_COMMENT("");
	EMIT_COMMENT("");
}

/*************************************************************************
function: PutSvars - Put System Vars list out
The stuff is only needed by the debugger
For "laziness" reasons, this uses the gvar_info structure.
*/
PutSvars (svarsP,startP)
data_ptr svarsP;
int *startP;
{
	int num_svars;
	char *name, *format;
	int zero = 0;
	int hashval, ix;
	data_ptr avarP;
	data_ptr *nullptr = NULL;
	int svars_id, gvar_info_id, format_id;
	int varname_ind, format_ind;

/* Emit the svar objects in the data segment */

	num_svars = ElfbArraySize(svarsP);
	assert(num_svars<10000);

	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("SYSTEMVAR VARIABLES");
	EMIT_COMMENT("");
	EMIT_COMMENT("");

	svars_id = ++UID;
	EmitStringTable(svars_id, num_svars, svarsP, 0); /* names */
	format_id = ++UID;
	EmitStringTable(format_id, num_svars, svarsP, 1); /* formats */

	gvar_info_id = ++UID;
	varname_ind = 0;
	format_ind = 0;
	if(num_svars)
		fprintf(AsmFP,"static gvar_info L%d[%d] = {\n",gvar_info_id,num_svars);
	for ( ix = 0; ix < num_svars; ix++ ) {
		avarP = ReadArray(svarsP,ix);
		name = StrFromArray(avarP,0);
		format = StrFromArray(avarP,1);
		hashval = hash(name);
		fprintf(AsmFP,"\t{\n\t0,\n");
		fprintf(AsmFP,"\t%d,\n",hashval);
		fprintf(AsmFP,"\t(L%d+%d),\n",svars_id,varname_ind);
		varname_ind += strlen(name)+1;
		if(format[0] != '\0') {
			fprintf(AsmFP,"\t(L%d+%d)\n\t},\n",format_id,format_ind);
			format_ind += strlen(format)+1;
		}
		else
			fprintf(AsmFP,"\tNULL\n\t},\n");
	}
	if(num_svars)
		fprintf(AsmFP,"};\n");

/* Emit the array of ptrs in the text segment */

	EMIT_COMMENT("SYSTEMVARS TABLE");

	*startP = ++UID;
	fprintf(AsmFP,"static CONST gvar_info *L%d[%d] = {\n",*startP,num_svars+1);

	for ( ix = 0; ix < num_svars; ix++ )
		fprintf(AsmFP,"\t(L%d+%d),\n",gvar_info_id,ix);
	fprintf(AsmFP,"\tNULL\n};\n");

	EMIT_COMMENT("END SYSTEMVARS INFO");
	EMIT_COMMENT("");
	EMIT_COMMENT("");
}

/*************************************************************************
function: PutGfuncs - Put Global Functions list out
The Array of pointers is loaded into the TextFrame.
The actual structures are loaded into the DataFrame.
The names are loaded into the TextFrame.
*/
PutGfuncs (gfuncsP,startP)
data_ptr gfuncsP;
int *startP;
{
	int num_gfuncs;
	char *name;
	int varname_ind;
	int zero = 0;
	int hashval, ix, jx, type, srcLine;
	int gfuncs_id, gfunc_info_id;
	data_ptr afuncP;
	data_ptr *nullptr = NULL;

/* Emit the gvar objects in the data segment */

	num_gfuncs = ElfbArraySize(gfuncsP);

	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("GLOBAL FUNCTIONS");
	EMIT_COMMENT("");
	EMIT_COMMENT("");

	gfuncs_id = ++UID;
	EmitStringTable(gfuncs_id, num_gfuncs, gfuncsP, 0);

	gfunc_info_id = ++UID;
	varname_ind = 0;
	
	if(num_gfuncs)
		fprintf(AsmFP,"static CONST gfunc_info L%d[%d] = {\n",
				gfunc_info_id, num_gfuncs);
	for ( ix = 0; ix < num_gfuncs; ix++ ) {
		afuncP = ReadArray(gfuncsP,ix);
		name = StrFromArray(afuncP,0);
		srcLine = IntFromArray(afuncP,2);
		type = IntFromArray(afuncP,1);
		if ( (type == TYPE_SETOBJ) || (type == TYPE_GETOBJ) )
			hashval = hash(name+1);
		else
			hashval = hash(name);
		
		fprintf(AsmFP,"\t{\n");
		fprintf(AsmFP,"\t\t(shrt*)(L%d+%d),\n",CodeStartsInText,IntFromArray(afuncP,3)/2);
		fprintf(AsmFP,"\t\t%d,\n",srcLine);
		fprintf(AsmFP,"\t\t%d,\n",hashval);
		fprintf(AsmFP,"\t\t(L%d+%d)\n\t},\n",gfuncs_id,varname_ind);
		varname_ind += strlen(name)+1;
	}
	if(num_gfuncs)
		fprintf(AsmFP,"};\n");

/* Emit the array of ptrs in the text segment */

	*startP = ++UID;
	
	fprintf(AsmFP,"static CONST gfunc_info *L%d[%d] = {\n",*startP,num_gfuncs+1);

	for ( ix = 0; ix < num_gfuncs; ix++ )
		fprintf(AsmFP,"\t(L%d+%d),\n",gfunc_info_id,ix);
	fprintf(AsmFP,"\tNULL\n};\n");

	EMIT_COMMENT("END FUNCTIONS INFO");
	EMIT_COMMENT("");
	EMIT_COMMENT("");
}

/*************************************************************************
function: PutExts - Put Externla linkage list out
The Array of pointers is loaded into the TextFrame.
The actual structures are loaded into the DataFrame.
The names are loaded into the TextFrame.
*/
PutExts (extsP,startP)
data_ptr extsP;
int *startP;
{
	int num_exts;
	char *name, *moduleName;
	int zero = 0;
	int hashval=0, ix;
	data_ptr afuncP;
	data_ptr *nullptr = NULL;
	int extTableId, extLinkId, moduleId;
    int name_ind, moduleName_ind;
	
/* Emit the gvar objects in the data segment */

	num_exts = ElfbArraySize(extsP);

	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("EXTERNAL LINKAGE");
	EMIT_COMMENT("");
	EMIT_COMMENT("");
	
	extTableId = ++UID;
	EmitStringTable(extTableId, num_exts, extsP, 0);
	moduleId = ++UID;
	EmitStringTable(moduleId, num_exts, extsP, 2);
	
	extLinkId = ++UID;
	name_ind = 0;
	moduleName_ind = 0;
	
	if(num_exts)
		fprintf(AsmFP,"static ext_link_for_elccc L%d[%d] = {\n", extLinkId, num_exts);
	for ( ix = 0; ix < num_exts; ix++ ) {
		afuncP = ReadArray(extsP,ix);
		name = StrFromArray(afuncP,0);
		
		fprintf(AsmFP,"\t{\n");
		fprintf(AsmFP,"\t\tNULL,\n");	/* addr */
		fprintf(AsmFP,"\t\tNULL,\n");	/* packP */
		fprintf(AsmFP,"\t\t%d,\n",hashval);
		fprintf(AsmFP,"\t\t(L%d+%d),\n",extTableId,name_ind);
		name_ind += strlen(name)+1;
		if ( AX(afuncP)->size > 2 ) {
			moduleName = StrFromArray(afuncP,2);
			fprintf(AsmFP,"\t\t(L%d+%d)\n\t},\n",moduleId,moduleName_ind);
			moduleName_ind += strlen(moduleName)+1;
	    }
		else
			fprintf(AsmFP,"\t\tNULL\n\t},\n");
	}
	if(num_exts)
		fprintf(AsmFP,"};\n");

/* Emit the array of ptrs in the text segment */

	EMIT_COMMENT("EXTERNS TABLE");

	*startP = ++UID;
	fprintf(AsmFP,"static CONST ext_link_for_elccc *L%d[%d] = {\n",*startP,num_exts+1);

	for ( ix = 0; ix < num_exts; ix++ )
		fprintf(AsmFP,"\t(L%d+%d),\n",extLinkId,ix);
	fprintf(AsmFP,"\tNULL\n};\n");

	EMIT_COMMENT("END EXTERNS INFO");
	EMIT_COMMENT("");
	EMIT_COMMENT("");
}

/*************************************************************************
function: PutVars - Names of local vars
*/
PutVars (varsP,startP)
data_ptr varsP;
int *startP;
{
	char *bytes = BX(varsP)->bytes;
	int zero = 0;
	int size;
	
	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("VARIABLE NAMES");

	size = BX(varsP)->size;
	fprintf(AsmFP,"static char L%d[%d] = {\n",*startP = ++UID,size);
	EmitBytes(bytes,size);
	fprintf(AsmFP,"\n};\n");

	EMIT_COMMENT("END VARIABLES");
	EMIT_COMMENT("");
	EMIT_COMMENT("");
}

/*************************************************************************
function: PutFormats - Dump Formats
*/
PutFormats (formatsP,startP)
data_ptr formatsP;
int *startP;
{
	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("FORMATS");
	EMIT_COMMENT("");
	EMIT_COMMENT("");

	PutArray(formatsP,startP,FALSE,FALSE);

	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("END FORMATS");
	EMIT_COMMENT("");
	EMIT_COMMENT("");
}

/*************************************************************************
function: PutLiterals - Dump Literals
*/
PutLiterals (litsP,startP)
data_ptr litsP;
int *startP;
{
	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("LITERALS");
	EMIT_COMMENT("");
	EMIT_COMMENT("");

	PutArray(litsP,startP,TRUE,TRUE);

	EMIT_COMMENT("");
	EMIT_COMMENT("");
	EMIT_COMMENT("END LITERALS");
	EMIT_COMMENT("");
	EMIT_COMMENT("");
}

/*************************************************************************
function: PutDatum - Write an Elf datum
*/
int
PutDatum (inP,startP,set_fault_bit,skip_array_header)
data_ptr inP;
int *startP;
int set_fault_bit;
int skip_array_header;
{
	int type;
	int size;

	if ( !inP )
		return(0);

	type = inP->type;

	switch ( type ) {

	case TYPE_NUMBER:
#if HP800V3
		fprintf(AsmFP, NXLT("{ %d,%.15gL},\n"),inP->type,NX(inP)->value);
#else
		fprintf(AsmFP, NXLT("{ %d,%.15g},\n"),inP->type,NX(inP)->value);
#endif
		return(1);
		break;

	case TYPE_STRING:
#if USE_COMMENTS
		fprintf(AsmFP,"/* TYPE_STRING - '%s' */\n",SX(inP)->text);
#endif
		size = sizeof(data_string) + strlen(SX(inP)->text);
		size = (size + sizeof(int) - 1) / sizeof(int);	/* int size, rnded up */
		EmitInts(inP,size,TRUE);
		return(size);
		break;

	case TYPE_BINARY:
		EMIT_COMMENT("TYPE_BINARY");
		size = sizeof(data_binary) + BX(inP)->size - 1;
		size = (size + sizeof(int) - 1) / sizeof(int);	/* int size, rnded up */
		EmitInts(inP,size,TRUE);
		return(size);
		break;

	case TYPE_ARRAY:
		PutArray (inP,startP,set_fault_bit,skip_array_header);
		return(1);
    }
}

/*************************************************************************
function: PutArray - Write an Elf Array
*/
int
PutArray (inP,startP,set_fault_bit,skip_array_header)
data_ptr inP;
int *startP;
int set_fault_bit;
int skip_array_header;
{
			int next_ind, ix, cnt, var_defined;
			int numberId, subId;
			int subStart[10000];
			data_ptr dP;
			data_array* aP=AX(inP);
			int has_int = FALSE;
			
			cnt = AX(inP)->size;
			assert(cnt < 10000);
			/* Emit sub arrays */
			for ( ix = 0; ix < cnt; ix++ ) {
				dP = AX(inP)->vals[ix];
				if ( dP && !IS_INTEGER(dP) && dP->type == TYPE_ARRAY ) {
					PutArray(dP,&subStart[ix],FALSE,FALSE);
				}
			}
			next_ind = 0;
			var_defined = FALSE;
			/* Emit all TYPE_NUMBER's */
			for ( ix = 0; ix < cnt; ix++ ) {
				dP = AX(inP)->vals[ix];
				if ( dP && !IS_INTEGER(dP) && dP->type == TYPE_NUMBER ) {
					subStart[ix] = next_ind;
					if(!var_defined) {
						EMIT_COMMENT("TYPE_NUMBER's");
#if HP800V3
/* Need leave "data_number" elements in normal data space on HP (UX9.x) 
  due to a compiler bug with the +ESlit switch.   This switch must be used to
  get the ELF data into read-only text space, but in the process the 
  compiler/linker packs static stuff tight, so some "data_number" structures do 
  not end up 8-byte aligned, as required by a double.  By getting ridd of the
  "static" word, things seem to work ok.
*/
						fprintf(AsmFP,NXLT("CONST data_number L%d[] = {"),
								numberId=++UID);
#else
						fprintf(AsmFP,NXLT("static CONST data_number L%d[] = {"),
								numberId=++UID);
#endif
						var_defined = TRUE;
					}
					next_ind += PutDatum(dP,&numberId,FALSE,FALSE);
				}
			}
			if(var_defined)
				fprintf(AsmFP,"};\n");
			next_ind = 0;
			var_defined = FALSE;
			/* Emit rest of the types, except integers */
			for ( ix = 0; ix < cnt; ix++ ) {
				dP = AX(inP)->vals[ix];
				if(dP && (IS_INTEGER(dP) || (dP->type != TYPE_ARRAY && 
			                            dP->type != TYPE_NUMBER))) {
					subStart[ix] = next_ind;
					if(!var_defined) {
						fprintf(AsmFP,"static CONST int L%d[] = {\n", subId= ++UID);
						var_defined = TRUE;
					}
					if (IS_INTEGER(dP)) {
						EMIT_COMMENT("TYPE_INTEGER");
						fprintf(AsmFP,"\t%ld,%ld,\n",TYPE_INTEGER,dP);
						has_int = TRUE;
						next_ind+=2;
					}
					else
						next_ind += PutDatum(dP,&subId,FALSE,FALSE);
				}
			}
			if(var_defined)
				fprintf(AsmFP,"};\n");

			EMIT_COMMENT("TYPE_ARRAY");
			if ( skip_array_header ) {
				fprintf(AsmFP,"static %s char *L%d[%d] = {\n",has_int ? "":"CONST",
						*startP = ++UID,cnt+1);
			}
			else {
				fprintf(AsmFP,"static CONST struct {\n");
				fprintf(AsmFP,"\tshrt type;\n");
				fprintf(AsmFP,"\tint size;\n");
				/* can't have 0 size array, so make it size 1 if cnt is 0 */
				fprintf(AsmFP,"\tchar *vals[%d];\n",(cnt == 0) ? 1 : cnt);
				fprintf(AsmFP,"} L%d = {\n", *startP = ++UID);
				fprintf(AsmFP,"\t(shrt)%d,%d,\n",aP->type,aP->size);
			}

			for ( ix = 0; ix < cnt; ix++ ) {
				dP = AX(inP)->vals[ix];
				if (dP) {
					if(!IS_INTEGER(dP) && dP->type == TYPE_ARRAY) {
						fprintf(AsmFP,"\t((char *)(&L%d))%s",subStart[ix],
								(set_fault_bit) ? "+1":"");
					}
					else {
						fprintf(AsmFP,"\t((char *)(L%d+%d))%s",
							(!IS_INTEGER(dP) && dP->type==TYPE_NUMBER) ? numberId : subId,
							subStart[ix],(set_fault_bit) ? "+1":"");
					}
				}
				else {
					fprintf(AsmFP,"\t(char *)(0)",dP);
				}
				fprintf(AsmFP,",\n");
			}
			if (skip_array_header)
				fprintf(AsmFP,"\t(char *)(0)\n");
			fprintf(AsmFP,"\n};\n");
}






