/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elcode.c
   Applix ASC version: 118
   Date: Thu Feb 25 21:11:07 1999
*/
#define FOR_ASM_GEN DEBUG       /* stuff needed for assembly code generator */
#define COMPILER_OPCODES 1      /* to get opcodes in compiler form */


#include <elcomp.h>
#include <eltokens.h>
#include <elnodes.h>
#include <elcodes.h>
#include <sqlops.h>
#include <getstr4.h>
#include <fsrts.h>
#include <elf_env.h>
#include <eldata.h>
/* #include <revstamp.h> */
#define APX_MASTER_REVSTAMP 301 /* ??1 create by build! */

#define MAX_SYMS 16000         /* Symbol Table size */

#define MAX_MALLOC 130000
#define MAX_CODE_WORDS (MAX_MALLOC/sizeof(shrt))    /* 1st release limit */


#if SGI|| ALPHA || HP800 || M_88K || CLIX || NT_MIPS || NT_PPC || (LINUX && (__alpha__ || __sparc__))
#define MAX_SYMS_WORDS (2 * 130000)
#else
#define MAX_SYMS_WORDS 130000
#endif

#define MAX_EA_REFS    (MAX_CODE_WORDS*sizeof(data_id*)) /* size elcodeStr.EaRefs array */
#define MAX_LITWORDS 130000 /* 260000 bytes of literals (used by WriteCodeFile) */

#if 0
#define CODE_CHECK(x,y) if ( !(x) ) {printf("check failed at %d\n",__LINE__); abort();}
#else
#define CODE_CHECK(x,y) if ( !(x) ) {CodeError y; goto cleanup;}
#endif

#define CFH_REVISION 0
#define CFH_SIZE 1
#define CFH_GLOBALS_START 2
#define CFH_GLOBALS_WORDS 3
#define CFH_CODE_STARTS 4
#define CFH_CODE_ENDS 5
#define CFH_SRCNAME 6

#if FOR_TESTER
FILE *Xrf;  /* cross reference file */
#define XREF(body) {if ( Xrf ) { fprintf body ; }}
static char *XRname();
#else
#define XREF(body)
#endif

axstatic struct {
    shrt *CodeFrame;     /* Text & Symbols during compilation */
    int MaxProgDex;      /* highest allowed index to Program.code */
    shrt SzCode;        /* size in shrts of code frame */
    shrt *SymsFrame;    /* temporary summation of symbols table */
    int SzSyms;        /* size in shrts of symbols frame */
    data_id **EaRefs;    /* parallel to code frame, contains EA ptrs (symP ptrs) */
    int InFunction;     /* TRUE while parsing function decl */
    int IsMethod;     /* TRUE while parsing function decl */
    int HaveFunctions;  /* becomes TRUE after first func/cmd dcl */
    int NumNodes;           /* how many nodes exist */
    int *SymTab;           /* malloced symbol table */
    int *SymHash;           /* parallel array of hash values */
    shrt SymOfNumberOne;    /* symtab entry for 1 */
    shrt SymOfTrue;         /* symtab entry for -1 */
    shrt SymOfFalse;        /* symtab entry for 0 */
    long GsymList;          /* head of global symbol thread */
    long ExtSymList;        /* head of extern symbols thread */
    shrt LineLastNoted;     /* line number of last stmt in src map */
    int LocalVarsOkNow;     /* TRUE during beginning of a function */
    int HaveDoneAssignment; /* TRUE after first assignment in a function */
    int LocalVarsCount;     /* Counter for local dcls within a function */
    int LocalVarsEntryWord; /* Where to HACK in local vars count */
    int FreeNdex;    /* Free node number index */
    int NewSwitch;   /* TRUE when a select first made */
    char LockedRegisters[MAX_REGS];/* regs used for for-loops, selects... */
    char RegistersAvailable[MAX_REGS];/* fo intra-inst register re-use */
    int NextRegister;    /* decrementing counter for temp vars */
    int LowestRegister;  /* lowest register number needed for function */
    char *ElfcVarnames; /* malloced symbol table for debugger */
    int ElfcSizeVarnames;   /* size of malloced buffer */
    int ElfcVarnamePos; /* size of ElfcVarnames */
	int HavePublicObjvars;	/* If TRUE, "public objvar" has been done */
}elcodeStr;

axstatic format_info *Formats;    /* chain of format info */
axglobal int ElfSysMode; /* TRUE if ELF compiled in system mode */
axglobal int AllowSourceInfo;    /* 'debug' mode compilation */
axglobal shrt NumSyms;       /* number of symbols in symbol table */
axglobal int CGcurLine;     /* current line code generator working on */
axglobal node_ptr *NodeArray;     /* malloced array to put nodes */
axglobal int ElfIntMode;    /* ?? TRUE if generating integer mode arthmetic */
axglobal int ElfcLanguage = 1;  /* Gothic default */
axglobal int ElfcAllowSQL = FALSE;  /* If true, accpet SQL keywords */
axglobal int ElfcAllowOracle = FALSE;   /* If true, accpet SQL keywords */
axglobal int ElfcObjectMode = FALSE; /* If true, compiling methods */
axglobal int ElfcTimeSliceMode = FALSE; /* whether @@@ TIME_SLICE enabled */
axglobal int ElfcPortableBytes = TRUE;  /* By default, machine independent! */
axglobal int ElfcPatchMode = FALSE; /* Used only by debugger */

extern char **ElfProtected;    /* Elf names accesable only is sys mode */
extern int ElfcNumErrs;        /* error counter */
extern comp_loop_info CG;   /* code generator loop info */
extern int ElfcLastFuncKeyAtLine;   /* hack for object code buster-upper */
extern typeErr ElfValidNameBasedOnSysMode();
extern data_ptr AxTaskCreateElfBinary();

/*************** DECODE ARRAYS **************************/

#define GET_SYMP(id) ((data_id*)(elcodeStr.SymsFrame + elcodeStr.SymTab[id-SYM_FUDGE]))

#define IS_ID 1
#define IS_FUNC 2
#define IS_LOCAL 4
#define IS_VAR 8

axstatic char SymInfo[] = {
IS_ID,                /* TYPE_UNKNOWN 0 */
0,                    /* TYPE_NUMBER 1 */
0,                    /* TYPE_STRING 2 */
0,                    /* TYPE_A_INTEGER 3 */
0,                    /* TYPE_A_STRING 4 */
0,                    /* TYPE_C_FUNC 5 */
0,                    /* TYPE_C_MACRO 6 */
0,                    /* TYPE_A_KEY 7 */
0,                    /* TYPE_ARRAY 8 */
IS_ID | IS_LOCAL | IS_VAR, /* TYPE_LOCVAR 9 */
IS_ID | IS_LOCAL | IS_VAR, /* TYPE_ARGVAR 10 */
IS_ID | IS_VAR,       /* TYPE_GLOVAR 11 */
IS_ID | IS_VAR,       /* TYPE_EXTVAR 12 */
IS_ID | IS_FUNC,      /* TYPE_QFUNC 13 */
IS_ID | IS_FUNC,      /* TYPE_UIFUNC 14 */
IS_ID | IS_FUNC,      /* TYPE_MACRO 15 */
IS_ID | IS_FUNC,      /* TYPE_EXTFUNC 16 */
IS_ID | IS_FUNC,      /* TYPE_NRFUNC 17 */
IS_ID,                /* TYPE_LABEL 18 */
0,                    /* TYPE_INTEGER 19 */
0,                    /* TYPE_ENTRY 20 */
IS_ID | IS_VAR,       /* TYPE_OBJECT 21 */
0,                    /* TYPE_BINARY 22 */
IS_ID | IS_FUNC,      /* TYPE_SETOBJ 23 */
IS_ID | IS_FUNC,      /* TYPE_GETOBJ 24 */
IS_ID | IS_VAR,       /* TYPE_OBJVAR 25 */
IS_ID | IS_VAR,       /* TYPE_SYSVAR 26 */
};
#define SYM_IS_ID(symP) (SymInfo[(symP)->type]&IS_ID)
#define FUNC_TYPE(t) (SymInfo[t]&IS_FUNC)
#define LOCAL_TYPE(t) (SymInfo[t]&IS_LOCAL)
#define VAR_TYPE(t) (SymInfo[t]&IS_VAR)

#include <elcode.p>

/*************************************************************************
function: InitCompiler - Do whatever inits are needed
*/
InitCompiler ()
{
    extern int LineCount;   /* in lexer */
    extern int DoingComment;
    extern int TokPos, DoingInclude;
    extern FILE *FpInclude;
    extern char *ElfcLineBuf;   /* current line being lexed */
    extern char *getenv();

    elcodeStr.LocalVarsOkNow = FALSE;
    elcodeStr.HaveDoneAssignment = FALSE;
    LineCount = 0;
    ElfIntMode = FALSE;
    ElfcNumErrs = 0;
    Formats = NULL;
    elcodeStr.CodeFrame = (shrt*)ElfcAlloc(MAX_MALLOC);
    MEMSET(&CG,0,sizeof(comp_loop_info));
    elcodeStr.SzCode = 0;
    elcodeStr.SymsFrame = (shrt*)ElfcAlloc(MAX_SYMS_WORDS*sizeof(shrt));
    elcodeStr.ElfcVarnamePos = 0;
    elcodeStr.ElfcSizeVarnames = 32000;
    elcodeStr.ElfcVarnames = ElfcAlloc(elcodeStr.ElfcSizeVarnames);
    elcodeStr.MaxProgDex = (MAX_MALLOC - sizeof(package_info))/sizeof(shrt);
    elcodeStr.SzSyms = 4;     /* Avoids a elcodeStr.Symtab[x] == 0 for elcodeStr.SymOfNumberOne */
    elcodeStr.SymsFrame[0] = elcodeStr.SymsFrame[1] = elcodeStr.SymsFrame[2] = elcodeStr.SymsFrame[3] = 0;
    elcodeStr.NumNodes = 0;
    NodeArray = (node_ptr*)ElfcAlloc(sizeof(node_ptr)*MAX_NODES);
    elcodeStr.FreeNdex = 0;
    NumSyms = 0;
    ElfcLanguage = 1;   /* Gothic default */
    CGcurLine = 0;
    elcodeStr.SymTab = (int*)ElfcAlloc(sizeof(int)*MAX_SYMS);
    elcodeStr.SymHash = (int*)ElfcAlloc(sizeof(int)*MAX_SYMS);
    elcodeStr.SymOfNumberOne = AddIntConst(1);
    elcodeStr.SymOfTrue = AddIntConst(-1);
    elcodeStr.SymOfFalse = AddIntConst(0);
    Formats = NULL;
    elcodeStr.GsymList = -1;
    elcodeStr.ExtSymList = -1;
    elcodeStr.HaveFunctions = FALSE;
	elcodeStr.HavePublicObjvars = FALSE;
    elcodeStr.LineLastNoted = -1;
#if !BLD_NO_ELF_DEBUG   /* set for elfc on turbo builds */
    AllowSourceInfo = TRUE;
#endif
    LineCount = 0;
    DoingComment = FALSE;
    elcodeStr.EaRefs = (data_id**)ElfcAlloc(MAX_EA_REFS);
    MEMSET(elcodeStr.EaRefs,0,MAX_EA_REFS);
    ElfTokArray = (tok_ptr*)ElfcAlloc(sizeof(tok_ptr)*(MAX_TOKENS+1));
    MEMSET(ElfTokArray,0,sizeof(tok_ptr)*(MAX_TOKENS+1));
    Defines = (def_info**)ElfcAlloc(sizeof(def_info*)*(ELF_HASH_PRIME+1));
    MEMSET(Defines,0,sizeof(def_info*)*(ELF_HASH_PRIME+1));
    ElfcLineBuf = ElfcAlloc(MAX_LINELEN+1);
    ElfSetupQsrch();
    TokPos = 0xFFFF;
    DoingInclude = FALSE;
    FpInclude = NULL;
    elcodeStr.InFunction = FALSE;
    elcodeStr.IsMethod = FALSE;
    ElfcAllowSQL = FALSE;
    ElfcAllowOracle = FALSE;
#if FOR_TESTER
    ElfcObjectMode = TRUE;
#else
    ElfcObjectMode = FALSE;
#endif
    ElfcTimeSliceMode = FALSE;
    AddWord(TYPE_OBJVAR);    /* so first function does not point to zero pos */
		/* Note that gfuncs supplying objvars locate to 0... to TYPE_OBJVAR! */

    /* If for debugger, flesh out symbols tab from runtime environment */
    if ( ElfcPatchMode )
        LoadSymbolsFromRuntime();
}

/*************************************************************************
function: LoadSymbolsFromRuntime - Load globals & this routine's stack vars
*/
static LoadSymbolsFromRuntime ()
{
    gvar_info **gvar_list;  /* null ptr termed array */
    ext_link **ext_list;    /* null ptr termed array of external vars/funcs */
    register data_id *symP = NULL; /* 18-Jan-94  steve  Fix this! */
    ndata symId;
    int ix;
    package_info *packP;
    extern package_info *ElfGetBreakPackP();

    packP = ElfGetBreakPackP(0);
    if ( !packP )
        return;
    gvar_list = packP->gvar_list;
    for ( ix = 0; gvar_list[ix]; ix++ ) {
        symId = AddIdentifier(gvar_list[ix]->name);
        symP->type = TYPE_GLOVAR;       /* format TODO!!0 */
    }
    ext_list = packP->ext_list;
    for ( ix = 0; ext_list[ix]; ix++ ) {
        symId = AddIdentifier(ext_list[ix]->nameP);
        symP->type = TYPE_UNKNOWN;
    }
}

/*************************************************************************
function: AlcNode - Create node storage
*/
any_node *AlcNode ()
{
    return((any_node*)ElfcAlloc(sizeof(any_node)));
}

/*************************************************************************
function: CrtNode - Create a node in parse tree
??c Sometimes the line number is incorrect because it is based on the linecount of the lexer.  Thus, a 'next qq' error where there is no qq has the line number of the next line because the parser had to look-ahead for a comma.
*/
LINTCOMMENT(VARARGS2)
ndata CrtNode (num_args,linenum,type,arg1,arg2,arg3,arg4,arg5,arg6)
register int num_args;
#ifdef AT
int linenum;
#else
unsigned long linenum;       /* line number in high 16 bits */
#endif
int type;
unsigned arg1,arg2,arg3,arg4,arg5,arg6;
{
    register node_ptr p;
    register int ix;


    ELFC_VERIFY(elcodeStr.NumNodes < MAX_NODES,XLT("Macro too big, too many parse nodes"));

    p.any = AlcNode();
    switch ( num_args ) {
    case 6: p.any->data[5] = arg6 & 0xFFFF;
    case 5: p.any->data[4] = arg5 & 0xFFFF;
    case 4: p.any->data[3] = arg4 & 0xFFFF;
    case 3: p.any->data[2] = arg3 & 0xFFFF;
    case 2: p.any->data[1] = arg2 & 0xFFFF;
    case 1: p.any->data[0] = arg1 & 0xFFFF;
    case 0:
            break;
    default:
        abort();
    }
    p.any->type = type;
    p.any->linenumber = linenum >> 16;
    if ( 0 == (ix = elcodeStr.FreeNdex) )
        ix = elcodeStr.NumNodes++;
    else
        elcodeStr.FreeNdex = NodeArray[ix].num;
    NodeArray[ix] = p;
    LOGGER(("CrtNode: %d (0x%lx) type %d, numargs %d: %d %d %d %d %d %d",
        ix+NODE_FUDGE,p,type,num_args,arg1,arg2,arg3,arg4,arg5,arg6));
    return(ix+NODE_FUDGE);
}

/*************************************************************************
function: FreeNode - release position in node table
*/
FreeNode (n)
register shrt n;
{
    n -= NODE_FUDGE;
    NodeArray[n].num = elcodeStr.FreeNdex;
    elcodeStr.FreeNdex = n;
}

/*************************************************************************
function: NeedLocal - Create/verify local var
*/
data_id *NeedLocal (id)
register shrt id;
{
    register data_id *symP;

    symP = GetSymP(id);
    if ( symP->type != TYPE_UNKNOWN ) {
        CODE_CHECK(!LOCAL_TYPE(symP->type),(XLT("%s multiply declared"),symP->name));
        id = AddOrSrchForId(symP->name,TRUE);   /* forced add */
        symP = GetSymP(id);
    }
    symP->type = TYPE_LOCVAR;
cleanup:
    return(symP);
}

/*************************************************************************
function: EmitStatement - General dispatcher for statement emitters
Because of recursion due to statment lists, this should not perform any
work that involves local vars.  The only local vars should be the passed
stmt var and the register node_ptr.
*/
EmitStatement (stmt)
register ndata stmt;
{
    register node_ptr np;

    np = NODEPTR(stmt);
    LOGGER(("EmitStmt:(%d) %d",np.any->linenumber,stmt));
    CGcurLine = np.any->linenumber;

    elcodeStr.NextRegister = MAX_REGS-1;  /* regs available for temp expressions */

    if ( elcodeStr.NewSwitch && (np.any->type != ND_CASE) )
        CodeError(XLT("CASE must be first statement after SELECT"));
    elcodeStr.NewSwitch = FALSE;  /* only one msg */

    if ( np.any->type == ND_FUNC ) {
        elcodeStr.LocalVarsCount = 0;
        elcodeStr.LocalVarsOkNow = TRUE;
        elcodeStr.HaveDoneAssignment = FALSE;
    }
    else if ( np.any->type == ND_VARS ) {
        CODE_CHECK(elcodeStr.LocalVarsOkNow || !elcodeStr.InFunction,(XLT("declaration in improper place")));
    }
    else if ( elcodeStr.LocalVarsOkNow ) {  /* set num locs/registers values */
        elcodeStr.CodeFrame[elcodeStr.LocalVarsEntryWord] = elcodeStr.LocalVarsCount+MAX_REGS;
        elcodeStr.LocalVarsOkNow = FALSE;
    }
#if 0
printf("emit type %d\n",NodeType(stmt));
#endif
    switch ( NodeType(stmt) ) {
    case ND_STMTLIST:
                        if ( np.stmtlist->list )
                            EmitStatement(np.stmtlist->list);
                        EmitStatement(np.stmtlist->stmt);
    CASE ND_SQLPREPARE: EmitSQLPrepare(np.any->data[0],np.any->data[1]);
    CASE ND_SQLUNPREPARE:   EmitSQLUnPrepare(np.any->data[0]);
    CASE ND_DCLCURSOR:  EmitSQLDeclare(np.any->data[0],np.any->data[1]);
    CASE ND_OPENCURSOR: EmitSQLOpen(np.any->data[0],np.any->data[1]);
    CASE ND_CLOSECURSOR:    EmitSQLClose(np.any->data[0]);
    CASE ND_EXECUTE:    EmitSQLExecute(np.any->data[0],np.any->data[1]);
    CASE ND_UPDATE: EmitSQLUpdate(np.any->data[0],np.any->data[1],
                                    np.any->data[2]);
    CASE ND_COMPILE:ElfcDirective(np.any->data[0]); goto ok_extern;
    CASE ND_ASSIGN: EmitAsgnStmt(np.assign->id,np.assign->ref,np.assign->expr);
    CASE ND_FUNCEXPR:   EmitFuncCall(TRUE,np.funcexpr->id,
                                np.funcexpr->args,np.funcexpr->object_def);
    CASE ND_ENTRYFUNC:   EmitEntryCall(TRUE,np.funcexpr->id,np.funcexpr->args);
    CASE ND_WHILE:  EmitWhileStmt(np.any->data[0]);
    CASE ND_WEND:   EmitWendStmt();
    CASE ND_FOR:    EmitForStmt(np.fr->id,np.fr->expr,np.fr->stop,np.fr->step);
    CASE ND_NEXT:   EmitNextStmt(np.any->data[0]);
    CASE ND_BREAKFOR: EmitBreakForStmt(np.any->data[0]);
    CASE ND_BREAKWHILE: EmitBreakWhileStmt();
    CASE ND_CONTINFOR: EmitContinForStmt(np.any->data[0]);
    CASE ND_CONTINWHILE: EmitContinWhileStmt();
    CASE ND_GOTO:   EmitGotoStmt(np.any->data[0]);
    CASE ND_RETURN: EmitReturnStmt(np.any->data[0]);
    CASE ND_IF:     EmitIfStmt(np.ife->expr,np.ife->st_then,np.ife->st_else);
    CASE ND_ONGOTO: EmitCompGotoStmt(np.cgoto->expr,np.cgoto->labels);
    CASE ND_ONCALL: EmitCompCallStmt(np.ccall->expr,np.ccall->calls);
    CASE ND_LABELED:EmitLabeledStmt(np.label->label,np.label->stmt);
    CASE ND_NULLST: AddGenericInst(OP_NOP);
    CASE ND_VARS:   EmitVarsStmt(np.any->data[0],np.any->data[1],
                    np.any->data[2],np.any->data[3],np.any->data[4]);
                    goto ok_extern;
    CASE ND_FORMAT: EmitFormatStmt(np.any->data[0],np.any->data[1],
                    np.any->data[2]);
                    goto ok_extern;
    CASE ND_FUNC:   EmitFunc(np.func->ftype,np.func->id,
                                np.func->args,np.func->functype);
                    goto ok_extern;
    CASE ND_EXTERN: EmitExterns(np.any->data[0],np.any->data[1]);goto ok_extern;
    CASE ND_END:    EmitEnder();goto ok_extern;
    CASE ND_ONERROR:    EmitOnError(np.any->data[0]);
    CASE ND_ERRERR:     EmitErrorOnError();
    CASE ND_SWITCH:     EmitSwitch(np.any->data[0]);
    CASE ND_CASE:       EmitCase(np.any->data[0],FALSE);
    CASE ND_ENDSWITCH:  EmitEndSwitch();
    CASE ND_DEFAULT:    EmitDefault();
    CASE ND_ADDTO:      EmitAddTo(TRUE,np.any->data[0],np.any->data[1]);
    CASE ND_SUBTO:      EmitAddTo(FALSE,np.any->data[0],np.any->data[1]);
    DEFAULT:
        CODE_CHECK(FALSE,(XLT("compiler failure (%d)"),NodeType(stmt)));
    }
#if 0
printf("emit done\n");
#endif
#if 1   /* No longer allow statements outside of functions */
    if ( !ElfcPatchMode ) {
        CODE_CHECK(elcodeStr.InFunction || (NodeType(stmt)==ND_LABELED),
                     (XLT("statement not in function")));
    }
#else
    CODE_CHECK(!elcodeStr.HaveFunctions || elcodeStr.InFunction || (NodeType(stmt)==ND_LABELED),
                     (XLT("statement not in function")));
#endif
ok_extern:
    FreeNode(stmt);
cleanup:;
 }

/*************************************************************************
function: EmitExterns - Tag identifiers as extern vars
*/
EmitExterns (vars,moduleName)
register ndata vars, moduleName;
{
    register int ix, num_vars;
    register node_ptr np;
    register data_id *symP;

    num_vars = SizeList(vars);
    for ( ix = 0; ix < num_vars; ix++ ) {
        np = NODEPTR(GetListMember(vars,ix));
        symP = GetSymP(np.fid->id);
        CODE_CHECK( (symP->type == 0) || (symP->type == TYPE_EXTVAR),
             (XLT("inconsistent usage of %s"),symP->name));
        CODE_CHECK(ElfSysMode || !ElfIsProtected(symP->name),
             (XLT("%s available only to system programmers"),symP->name));
        symP->type = TYPE_EXTVAR;
        if ( np.fid->fmt ) {
            symP->format = FindFormat(GetSymP(np.fid->fmt)->name);
            CODE_CHECK(symP->format,(XLT("no such format as %s"),GetSymP(np.fid->fmt)->name));
        }
#if 0   /* screws object model */
        if ( ElfcAllowSQL ) {
            CODE_CHECK(moduleName,(XLT("module name not specified for %s"),symP->name));
        }
#endif
        symP->moduleName = moduleName;
    }
cleanup:;
}

/*************************************************************************
function: EmitFunc - Start a function
User defined local vars begin at MAX_REGS.  Expression temp registers
are allocated from MAX_REGS down to 1.
*/
static EmitFunc (type,id,args,functype)
register ndata type;
register ndata id;
register ndata args, functype;
{
    register data_id *symP;
    register shrt num_args, ix;
    register node_ptr np;
	ndata obj;
    data_id *NeedLocal();
    char name[ELFIDSTR+1];

    typeErr badname, ElfNameValid();
    int isMethod = 0;   /* values 0/1, sidesteps ?/! on methods */

    CODE_CHECK(!elcodeStr.InFunction,(XLT("function in function")));
    if ( !elcodeStr.HaveFunctions )
        AddGenericInst(OP_RETURN);

    symP = GetSymP(id);
    if ( (type == TK_SET) || (type == TK_GET) ) {
        isMethod = 1;
        CODE_CHECK(strlen(symP->name)<ELFIDSTR,(XLT("name too long: %s"),symP->name));
        strcpy(name+1,symP->name);
        name[0] = (type == TK_SET) ? '!' : '?';
        id = AddIdentifier(name);
        symP = GetSymP(id);
    }
    CODE_CHECK( (symP->type == 0) || (symP->type == TYPE_EXTFUNC),
            (XLT("%s already declared"),symP->name+isMethod));
    if ( badname = ElfValidNameBasedOnSysMode(symP->name+isMethod) ) {
        char temp[100];
        StrGet(badname,temp);
        CodeError(XLT("%s illegal name (%s)"),symP->name+isMethod,temp);
    }
    Align4Word();   /* all functions on 64bit boundaries */
    symP->locate = elcodeStr.SzCode;
    symP->srcLine = ElfcLastFuncKeyAtLine;  /* from lexer */
    switch ( type ) {
    case TK_UIPROC:     symP->type = TYPE_UIFUNC;       break;
    case TK_NRPROC:     symP->type = TYPE_NRFUNC;       break;
    case TK_PROC:       symP->type = TYPE_MACRO;        break;
    case TK_FUNC:       symP->type = TYPE_QFUNC;        break;
    case TK_SET:        symP->type = TYPE_SETOBJ;       break;
    case TK_GET:        symP->type = TYPE_GETOBJ;       break;
    default:            abort();
    }
	symP->isPrivate = (functype == TK_PRIVATE);
    XREF((Xrf,"(%d) %s\n",symP->type,XRname(symP->name)));
    elcodeStr.InFunction = TRUE;
	elcodeStr.IsMethod = isMethod;
    elcodeStr.HaveFunctions = TRUE;
    num_args = SizeList(args);
    AddWord(symP->type);
    AddWord(OP_ENTER);      /* cannot do as inst cause of SrcMap */
    AddWord(num_args+isMethod);
    elcodeStr.LocalVarsEntryWord = elcodeStr.SzCode; /* where to put real locals count */
    AddWord(0);     /* to be mashed over by locals count */
    AddWord(0);     /* to be mashed over by registers count */
    AddWord(elcodeStr.ElfcVarnamePos);
    if ( isMethod ) {
        obj = AddOrSrchForId(NXLT("THIS"),FALSE);   /* Re-fetch based on name */
        symP = NeedLocal(obj);
        symP->type = TYPE_ARGVAR;
        symP->locate = 0;
        symP->isObject = TRUE;
        ElfcAddVarName(symP);
    }
    for ( ix = 0; ix < num_args; ix++ ) {
        np = NODEPTR(GetListMember(args,(int)ix));
        symP = NeedLocal(np.fid->id);
        if ( np.fid->fmt ) {
            symP->format = FindFormat(GetSymP(np.fid->fmt)->name);
            CODE_CHECK(symP->format,(XLT("no such format as %s"),GetSymP(np.fid->fmt)->name));
        }
        else if ( np.fid->isObject )
            symP->isObject = TRUE;
        symP->formatDim = np.fid->fmtDim;
        symP->type = TYPE_ARGVAR;
        symP->locate = ix+isMethod;  /* runtime table offs for arg vars */
        ElfcAddVarName(symP);
    }
    memset(elcodeStr.LockedRegisters,0,MAX_REGS);
    memset(elcodeStr.RegistersAvailable,0,MAX_REGS);
    elcodeStr.LowestRegister = MAX_REGS;
cleanup:;
}

/*************************************************************************
function: ElfcAddVarName - Add a variable name & its format to the table
*/
static ElfcAddVarName (symP)
register data_id *symP;
{
    char *name;
    char buf[200];

/* make sure buffer big enough */

    if ( elcodeStr.ElfcVarnamePos + 100 > elcodeStr.ElfcSizeVarnames ) {
        char *p;
        p = ElfcAlloc(elcodeStr.ElfcSizeVarnames*2);
        memcpy(p,elcodeStr.ElfcVarnames,elcodeStr.ElfcSizeVarnames);
        elcodeStr.ElfcSizeVarnames = elcodeStr.ElfcSizeVarnames * 2;
        elcodeStr.ElfcVarnames = p; /* old one freed when compiler completes */
    }

/* add var name */

    name = symP->name;
    strcpy(elcodeStr.ElfcVarnames+elcodeStr.ElfcVarnamePos,name);
    elcodeStr.ElfcVarnamePos += strlen(name)+1;

/* add format name */

    if ( symP->format ) {
        name = symP->format->myname;
        symP->format->used = TRUE;
    }
    else
        name = "";
    strcpy(buf,name);

/* if the var is actually array of format, add braces */

    if ( symP->formatDim )
        encodeFormatDim(buf,symP->formatDim,name);

    strcpy(elcodeStr.ElfcVarnames+elcodeStr.ElfcVarnamePos,buf);
    elcodeStr.ElfcVarnamePos += strlen(buf)+1;
}

/*************************************************************************
function: EmitEnder - End a function
*/
EmitEnder ()
{
    register int ix;
    register data_id *symP;
    int check_labels = TRUE;

    CODE_CHECK(elcodeStr.InFunction,(XLT("misplaced ENDMACRO statement")));
    AddGenericInst(OP_RETURN);
    elcodeStr.CodeFrame[elcodeStr.LocalVarsEntryWord+1] = MAX_REGS - elcodeStr.LowestRegister;

/* verify termination of for, while loops */

    if ( CG.num_for_loops ) {
        CodeError(XLT("FOR without NEXT, variable %s"),
                     GetSymP(CG.for_loops[CG.num_for_loops-1].id)->name);
        check_labels = FALSE;
        CG.num_for_loops = 0;
    }
    if ( CG.num_while_loops ) {
        CodeError(XLT("WHILE without WEND"));
        check_labels = FALSE;
        CG.num_while_loops = 0;
    }
    if ( CG.num_selects ) {
        CodeError(XLT("CASE OF without ENDCASE"));
        check_labels = FALSE;
        CG.num_selects = 0;
    }

/* get rid of local symbols */

    for ( ix = 0; ix < NumSyms; ix++ ) {
        symP = GetSymP((shrt)(ix+SYM_FUDGE));
        if ( check_labels && (symP->type == TYPE_LABEL )) {
            if ( !symP->locate ) {
                if ( symP->name[0] == '^' )
                    CodeError(XLT("branch point not resolvable"));
                else
                    CodeError(XLT("label %s not specified"),symP->name);
            }
            elcodeStr.SymTab[ix] = 0;
        }
        else if ( LOCAL_TYPE(symP->type) )
            elcodeStr.SymTab[ix] = 0;
    }

cleanup:
    elcodeStr.InFunction = FALSE;
}

/*************************************************************************
function: FinishSrcFile - Verify state of world at end of file
*/
FinishSrcFile ()
{
    if ( ElfcPatchMode ) {
        AddGenericInst(OP_PATCH_RET);
    }
    else {
        CODE_CHECK(!elcodeStr.InFunction,(XLT("missing ENDMACRO statement")));
    }
cleanup:;
}

/*************************************************************************
function: EmitVarsStmt - Create local/global/object vars
*/
EmitVarsStmt (tokType,format,vars,fmtDim,isObject)
register ndata vars, format, fmtDim, isObject, tokType;
{
    register data_id *symP;
    register int ix, num_vars;
    typeErr badname, ElfNameValid();
    data_id *NeedLocal();

    num_vars = SizeList(vars);
    LOGGER(("EmitVarsStmt: vars %d count %d",vars,num_vars));
    for ( ix = 0; ix < num_vars; ix++ ) {
        if ( tokType == TK_OBJVAR || tokType == TK_PUBLIC ) {
            symP = GetSymP(GetListMember(vars,ix));
            CODE_CHECK( !symP->type || (symP->type == TYPE_OBJVAR),
                    (XLT("%s already declared"),symP->name));
            if ( badname = ElfValidNameBasedOnSysMode(symP->name) ) {
                char temp[100];
                StrGet(badname,temp);
                CodeError(XLT("%s illegal name (%s)"),symP->name,temp);
            }
            symP->type = TYPE_OBJVAR;
            if ( format ) {
                symP->format = FindFormat(GetSymP(format)->name);
                CODE_CHECK(symP->format,(XLT("no such format as %s"),GetSymP(format)->name));
            }
            symP->formatDim = fmtDim;
            symP->isObject = isObject;
			if ( tokType == TK_PUBLIC ) {
				elcodeStr.HavePublicObjvars = TRUE;
				symP->isPublic = TRUE;
		        CODE_CHECK(strlen(symP->name)<ELFIDSTR,(XLT("Name too long for a public objvar: %s"),symP->name)); /* Stricter limit because of matching SET/GET funcs */
			}
        }
        else if ( tokType == TK_SYSVAR ) {
            symP = GetSymP(GetListMember(vars,ix));
            CODE_CHECK( !symP->type || (symP->type == TYPE_SYSVAR),
                    (XLT("%s already declared"),symP->name));
            if ( badname = ElfValidNameBasedOnSysMode(symP->name) ) {
                char temp[100];
                StrGet(badname,temp);
                CodeError(XLT("%s illegal name (%s)"),symP->name,temp);
            }
            symP->type = TYPE_SYSVAR;
            if ( format ) {
                symP->format = FindFormat(GetSymP(format)->name);
                CODE_CHECK(symP->format,(XLT("no such format as %s"),GetSymP(format)->name));
            }
            symP->formatDim = fmtDim;
            symP->isObject = isObject;
        }
        else if ( elcodeStr.InFunction ) {
            symP = NeedLocal(GetListMember(vars,ix));
            symP->locate = MAX_REGS + elcodeStr.LocalVarsCount++;
            symP->formatDim = fmtDim;
            symP->isObject = isObject;
            CODE_CHECK(elcodeStr.LocalVarsCount < MAX_REGS,(XLT("Macro too big, too many local variables")));
            if ( format ) {
                symP->format = FindFormat(GetSymP(format)->name);
                CODE_CHECK(symP->format,(XLT("no such format as %s"),GetSymP(format)->name));
            }
            ElfcAddVarName(symP);
        }
        else {
            symP = GetSymP(GetListMember(vars,ix));
#if 0
            if ( ElfSysMode ) {
                printf("Warning: %s is a non-local variable\n",symP->name);
/*              CODE_CHECK(FALSE,(XLT("Non-local vars not supported for standard libraries"))); */
            }
#endif
            CODE_CHECK( !symP->type || (symP->type == TYPE_EXTVAR),
                    (XLT("%s already declared"),symP->name));
            if ( badname = ElfValidNameBasedOnSysMode(symP->name) ) {
                char temp[100];
                StrGet(badname,temp);
                CodeError(XLT("%s illegal name (%s)"),symP->name,temp);
            }
            symP->type = TYPE_GLOVAR;
            if ( format ) {
                symP->format = FindFormat(GetSymP(format)->name);
                CODE_CHECK(symP->format,(XLT("no such format as %s"),GetSymP(format)->name));
            }
            symP->formatDim = fmtDim;
            symP->isObject = isObject;
        }
    }
cleanup:;
}

/*************************************************************************
function: EmitFormatStmt - Create format info
*/
EmitFormatStmt (id,elements,fmtDim)
register ndata id, elements,fmtDim;
{
    register format_info *p;
    register node_ptr np;
    register int ix, count;
    register data_id *symP;
    register ndata this;
    extern char ElfcLastInclude[STRPATH+1];
    char *ElfcStrAlloc(), *ElfcAlloc();

    CODE_CHECK(!elcodeStr.InFunction,(XLT("format specifications must be outside functions")));
    symP = GetSymP(id);
    CODE_CHECK(!FindFormat(symP->name),(XLT("format %s already declared"),symP->name));
    count = SizeList(elements);
    CODE_CHECK(count < MAX_FORMATS,(XLT("too many elements in format")));
    p = (format_info*)ElfcAlloc(sizeof(format_info));
    memset(p,0,sizeof(format_info));
    p->myname = ElfcStrAlloc(symP->name);
    p->fileName = ElfcStrAlloc(ElfcLastInclude);
    for ( ix = 0; ix < count; ix++ ) {
        this = GetListMember(elements,ix);
        np = NODEPTR(this);
        p->names[ix] = ElfcStrAlloc(GetSymP(np.fid->id)->name);
        if ( np.fid->fmt ) {
            p->formats[ix] = ElfcStrAlloc(GetSymP(np.fid->fmt)->name);
            p->formatDim[ix] = np.fid->fmtDim;
        }
    }
    p->next = Formats;
    Formats = p;
cleanup:;
}

/*************************************************************************
function: EmitReturnStmt - return from function
*/
EmitReturnStmt (expr)
register ndata expr;
{
    if ( expr ) {
        EmitExpression(expr);
        if ( ElfcPatchMode )
            AddGenericInst(OP_PATCH_RET);
        else
            AddGenericInst(OP_RETVAL);
    }
    else {
        if ( ElfcPatchMode ) {
            EmitExpression(TK_NULL);
            AddGenericInst(OP_PATCH_RET);
        }
        else
            AddGenericInst(OP_RETURN);
    }
}

/*************************************************************************
function: ElfcDirective - Process a compiler directive
*/
ElfcDirective (id)
register ndata id;       /* destination variable */
{
    register data_id *symP;

    symP = GetSymP(id);
    if ( streq(symP->name,NXLT("GOTHIC")) )
        ElfcLanguage = 0;
    if ( streq(symP->name,NXLT("ENGLISH")) )
        ElfcLanguage = 1;

    /* Embedded SQL mode turn on/off directives */
    if ( streq(symP->name,NXLT("SQL")) )
        ElfcAllowSQL++;
    if ( streq(symP->name,NXLT("NO_SQL")) ) {
        ElfcAllowSQL--;
        if (ElfcAllowSQL < 0)
            ElfcAllowSQL = 0;
    }

    /* Oracle mode turn on/off directives */
    if ( streq(symP->name,NXLT("ORACLE")) ) {
        ElfcAllowSQL++;
        ElfcAllowOracle = TRUE;
    }
    if ( streq(symP->name,NXLT("NO_ORACLE")) ) {
        ElfcAllowSQL--;
        if (ElfcAllowSQL < 0)
            ElfcAllowSQL = 0;
        ElfcAllowOracle = FALSE;
    }

    /* object mode turn on/off directives */
    if ( streq(symP->name,NXLT("OBJECTS")) )
        ElfcObjectMode = TRUE;
    if ( streq(symP->name,NXLT("NO_OBJECTS")) )
        ElfcObjectMode = FALSE;
    
    /* alarm based tiome slicing */
    if ( streq(symP->name,NXLT("TIME_SLICE")) )
        ElfcTimeSliceMode = TRUE;
    if ( streq(symP->name,NXLT("NO_TIME_SLICE")) )
        ElfcTimeSliceMode = FALSE;

#if 0
    if ( streq(symP->name,NXLT("REV3")) )
        ElfcAllowSQL = TRUE;
#endif
}

/*************************************************************************
function: EmitAsgnStmt - emit assignment statement
*/

EmitAsgnStmt (id,array_ref,expr)
register ndata id;       /* destination variable */
register ndata array_ref;    /* array subscript */
register ndata expr;     /* expression to set it to */
{
    register data_id *symP;
    register int num_datums = 1;
    register int is_data;

    LOGGER(("EmitAsgnStmt: id %d array_ref %d expr %d",id,array_ref,expr));
    is_data = IS_NODE(expr) && (NodeType(expr) == ND_NUMLIST) &&
                                   (1 < (num_datums = SizeList(expr)) );
    EmitExpression(expr);

/* Special hack to solve the following case:

    var foo         ' global foo
    macro a
        var foo     ' local decl
        foo = 1     ' THE ASSIGNMENT GOES TO THE GLOBAL FOO, NOT THE LOCAL!
    endmacro

Without the HaveDoneAssignment test, the parser had pre-fetched the
assignment's foo in order to close the declaration.  Thus, the
local foo hadn't been created yet and the assignment goes to the
global foo.  Any statement before the assignment would solve (hide)
the bug.  The real fix (below) is to re-fetch the identifier if the assignment
is the first executable line after the var decls.
*/

    if ( !elcodeStr.HaveDoneAssignment ) {      /* This is the hack */
        symP = GetSymP(id);     /* Get name of Identifier */
        id = AddOrSrchForId(symP->name,FALSE);  /* Re-fetch based on name */
        elcodeStr.HaveDoneAssignment = TRUE;    /* Save performance for later assignments */
    }
    symP = NeedVariable(id);
    if ( symP->isObject && array_ref )
        EmitSetObj(id,array_ref,-1,is_data,num_datums);
    else
        EmitLeftHandSideAsgn(id,array_ref,is_data,num_datums);
}

/*************************************************************************
function: EmitLeftHandSideAsgn - Emit LHS of assignment statement
*/
static EmitLeftHandSideAsgn (id,array_ref,is_data,num_datums)
register ndata id;       /* destination variable */
register ndata array_ref;    /* array subscript */
register int is_data, num_datums;
{
    int refs; /* number of array_refs */
    ndata ref_list[MAX_DIMS];
    register data_id *symP;
    register int ix;
    ndata methodId;
    register node_ptr np;

    symP = NeedVariable(id);
    if ( array_ref ) {
        RefListToArray(symP,array_ref,ref_list,&refs);
        CODE_CHECK(refs<10,(XLT("too many array subscripts")));
        for ( ix = refs-1; ix >= 0; ix-- ) {
            EmitExpression(ref_list[ix]);
        }
        if ( is_data ) {
            AddEaWithImmediateInst(OP_POPDARR,symP,refs);
            AddWord(num_datums);
        }
        else {
            if ( refs == 1 )
                AddEaInstruction(OP_POPTO1AR,symP);
            else
                AddEaWithImmediateInst(OP_POPTOARR,symP,refs);
        }
        FreeNode(array_ref);
    }
    else if ( is_data )
        AddEaWithImmediateInst(OP_POPDATA,symP,num_datums);
    else
        AddEaInstruction(OP_POPTO,symP);
cleanup:;
}


/*************************************************************************
function: EmitSetObj - Emit a set method call
    OP_SETOBJ+Ea
    0
    methodName literal offset, if 0 next item not there, pop name from stk
    methodHash literal offset
    arg_count of args that where pushed before the OP_SETOBJ instruction
*/
EmitSetObj (id,methodlist,numLevels,is_data,num_datums)
register ndata id;       /* destination variable */
register ndata methodlist;    /* array subscript */
register int numLevels; /* how many levels to process in list */
register int is_data, num_datums;
{
    int refs; /* number of methodlists */
    ndata ref_list[MAX_DIMS];
    register data_id *symP;
    int thisLevel;
    ndata methodId;
    register data_id *methodP;
    register node_ptr np;
    int dynamic_method_name;        /* foo.!bar evaluates bar to get method name */
    int is_broadcast;
    int objRegister;

    assert(IS_NODE(methodlist));
    symP = NeedVariable(id);

/* check for "objs[ix] = foo" ... do as normal array assignment */

    methodId = GetListMember(methodlist, 0);
    if ( IS_NODE(methodId) && (NodeType(methodId) == ND_NUMLIST) ) {
        EmitLeftHandSideAsgn(id,methodlist,is_data,num_datums);
        return;
    }

    if ( numLevels == -1 )  /* only -1 at top level */
        numLevels = SizeList(methodlist);
    thisLevel = numLevels - 1;
    if ( thisLevel != 0 ) {
        objRegister = EmitGetObj(id,methodlist,thisLevel,0);
        AddGenericInst(OP_TOSSTOS);
    }
    np = NODEPTR(methodlist);
    methodId = GetListMember(methodlist, thisLevel);
    if ( IS_NODE(methodId) ) {
        np = NODEPTR(methodId);
        switch ( np.any->data[2] ) {
        case TK_ENTRY:
            dynamic_method_name = TRUE;
            is_broadcast = FALSE;
            EmitExpression(np.any->data[0]);
            break;
        case TK_MULT:
            is_broadcast = TRUE;
            dynamic_method_name = FALSE;
            methodId = np.any->data[0];
            break;
        case TK_EXP:
            is_broadcast = 2;		/* false + call_inherit */
            dynamic_method_name = FALSE;
            methodId = np.any->data[0];
            break;
        case TK_ENTRY+TK_MULT:
            dynamic_method_name = TRUE;
            is_broadcast = TRUE;
            EmitExpression(np.any->data[0]);
            break;
        case TK_ENTRY+TK_EXP:
            dynamic_method_name = TRUE;
            is_broadcast = 2;		/* false + call_inherit */
            EmitExpression(np.any->data[0]);
            break;
        case TK_MULT+TK_EXP:
            is_broadcast = 3;		/* true + call_inherit */
            dynamic_method_name = FALSE;
            methodId = np.any->data[0];
            break;
        case TK_ENTRY+TK_MULT+TK_EXP:
            dynamic_method_name = TRUE;
            is_broadcast = 3;		/* true + call_inherit */
            EmitExpression(np.any->data[0]);
            break;
        default:
            CODE_CHECK(FALSE,(XLT("Invalid object method")));
        }
    }
    else {
/*printf("final set method %s\n",GetSymP(methodId)->name);*/
        dynamic_method_name = FALSE;
        is_broadcast = FALSE;
    }
    if ( thisLevel == 0 )
        AddEaInstruction(OP_SETOBJ,symP);
    else {
        AddGenericInst(OP_SETOBJ);
        AddWord(objRegister);
    }
    AddWord(0);     /* is a 'set' method, no return value */
    AddWord(is_broadcast);

    if ( dynamic_method_name )
        AddWord(0); /* tell intr to pop value off stack */
    else {
        methodP = GetSymP(methodId);
        AddLiteralEaWord(GetSymP(AddStrConst(methodP->name)));
        AddWord(hash(methodP->name));
    }
    if ( is_data )
        AddWord(-num_datums);
    else
        AddWord(num_datums);
cleanup:;
}

/*************************************************************************
function: EmitExpression - push expression onto stack
*/
EmitExpression (expr)
register ndata expr;
{
    ExprEmitter(expr);
}

/*************************************************************************
function: ExprEmitter - expr emitter, may pass vars themselves
*/
ExprEmitter (expr)
register ndata expr;
{
    register node_ptr np;

    LOGGER(("EmitExpr: expr %d",expr));

    expr = ReduceExpression(expr);

    if ( expr == TK_TRUE )
        PushNumber((data_comp_integer *)GetSymP(elcodeStr.SymOfTrue));
    else if ( expr == TK_FALSE )
        PushNumber((data_comp_integer *)GetSymP(elcodeStr.SymOfFalse));
    else if ( expr == TK_NULL )
        AddGenericInst(OP_PUSH_NULL);
    else if ( IS_SYM(expr) ) {   /* identifier or constant in symbol table */
        register data_id *symP;
        symP = GetSymP(expr);
        if ( SYM_IS_ID(symP) ) {
            symP = NeedVariable(expr);
            AddEaInstruction(OP_PUSH,symP);
        }
		/* try for small instruction first */
        else if ( !PushNumber((data_comp_integer *)symP) )   
            AddEaInstruction(OP_PUSH_LIT,symP);
    }
    else {
        np = NODEPTR(expr);
        switch ( NodeType(expr) ) {
        case ND_NUMLIST:    /* only used in assignments & SQL open cursor */
            CODE_CHECK(SizeList(expr) <= 1000,(/* prevent elintr stack overflow */
                XLT("Array assignment exceeds 1000 limit. Use multiple assignment statements.")));
            if ( np.numlist->list )
                ExprEmitter(np.numlist->list);
            ExprEmitter(np.numlist->expr);
 
        CASE ND_FUNCEXPR:
            EmitFuncCall(FALSE,np.funcexpr->id,np.funcexpr->args,
                            np.funcexpr->object_def);

        CASE ND_ENTRY: {
            register data_id *symP;

            symP = NeedFunc(np.any->data[0]);           
            AddEaInstruction(OP_PUSH_ENT,GetSymP(AddStrConst(symP->name)));
            symP->need_link = TRUE;
            AddWord(GetRegister());
        }

        CASE ND_SELECT:
            EmitSelect(expr);

        CASE ND_FETCHCURSOR:
            EmitSQLFetch(expr);

        CASE ND_ENTRYFUNC:
            EmitEntryCall(FALSE,np.funcexpr->id,np.funcexpr->args);

        CASE ND_AREF:
        {
            register int ix;
            register data_id *symP;
            int refs;
            ndata ref_list[MAX_DIMS];
 
            symP = NeedVariable(np.aref->id);
            if ( symP->isObject ) {
                ndata methodId; /* check for foo = objs[j] special */
                methodId = GetListMember(np.aref->ref, 0);
                if ( IS_NODE(methodId) && (NodeType(methodId) == ND_NUMLIST) )
                    goto do_arr;
                EmitFuncCall(FALSE,np.aref->id,np.aref->objArgs,np.aref->ref);
            }
            else {
do_arr:         RefListToArray(symP,np.aref->ref,ref_list,&refs);
                for ( ix = refs-1; ix >= 0; ix-- )
                    EmitExpression(ref_list[ix]);
                if ( refs == 1 )
                    AddEaInstruction(OP_PUSH1ARR,symP);
                else
                    AddEaWithImmediateInst(OP_PUSHARR,symP,refs);
            }
        }
        CASE ND_OP:
        {
            register ndata op;
            register ndata lhs, rhs;
 
            op = np.op->op;
            lhs = np.op->lhs;
            rhs = np.op->rhs;
 
            if ( (op == OP_UMINUS) || (op == OP_IUMINUS) || (op == OP_NOT) ) {
                EmitExpression(rhs);
                AddDoubleInst(op,GetRegister());
            }
            else {
                EmitExpression(rhs);
                EmitExpression(lhs);
                AddDoubleInst(op,GetRegister());
            }
        }
        CASE ND_ARRLIT:
        {
            register int num_datums = 1;
            register int reg;
            register ndata list;
            list = np.any->data[0];
            if ( list && IS_NODE(list) && (NodeType(list) == ND_NUMLIST) )
                num_datums = SizeList(list);
            reg = GetRegister();
            if ( list == 0 )
                num_datums = 0;
            else
                ExprEmitter(list);
            AddDoubleInst(OP_POPDATA,reg);
            AddWord(num_datums);
            AddDoubleInst(OP_PUSH,reg);
/*          RlsRegister(reg); */
        }
        }       /* end of switch */
        FreeNode(expr);
    }
    return;

cleanup:
    return;
}

/*************************************************************************
function: PushNumber - If possible, create a quick immediate for number push
This is primarily usefull for array indexes and bulk data assignments
where the benifit of generating a single word instruction is
worth the extra work.
*/
PushNumber (symP)
register data_comp_integer *symP;
{
    if ( (symP->type != TYPE_INTEGER) || 
         (symP->value < -1) || (symP->value > MAX_QPUSH) )
        return(FALSE);
    AddDoubleInst(OP_PUSH_NUM,(symP->value + 1));
    return(TRUE);
}

/*************************************************************************
function: EmitErrorOnError - Turn off frame's error catcher
*/
EmitErrorOnError ()
{
    AddGenericInst(OP_ERRERR);
}

/*************************************************************************
function: EmitOnError - Set error catcher
*/
EmitOnError(stmt)
register ndata stmt;
{
    data_id *NeedLabel();
    shrt bypass;

    bypass = CrtLabel(CG.cur_label++);
    AddEaInstruction(OP_ONERROR,GetSymP(bypass));
    EmitStatement(stmt);
    SetLabelHere(bypass);
}

/*************************************************************************
function: EmitFuncCall - Emit an object method "get" or a normal function call
*/
EmitFuncCall (is_stmt,id,args,methodlist)
register int is_stmt;   /* TRUE if return value is to be ignored */
register ndata id;      /* id of function symbol */
register ndata args;    /* arg list or 0 */
register ndata methodlist;  /* 0 or object method list */
{
    int num_datums, ix;
    register data_id *symP;

    if ( methodlist ) {
        symP = NeedVariable(id);
        CODE_CHECK(symP->isObject,(XLT("%s[...] is not a valid macro call"),symP->name));
        num_datums = SizeList(args);
        for ( ix = num_datums-1; ix >= 0; ix-- )
            ExprEmitter(GetListMember(args,ix));
        if ( is_stmt )
            EmitSetObj(id,methodlist,-1,FALSE,num_datums);
        else
            EmitGetObj(id,methodlist,-1,num_datums);
    }
    else
        EmitNormalFuncCall(is_stmt,id,args);
cleanup:;
}

/*************************************************************************
function: EmitNormalFuncCall - Emit code for a function call
*/
static EmitNormalFuncCall (is_stmt,id,args)
register int is_stmt;   /* TRUE if return value is to be ignored */
register ndata id;      /* id of function symbol */
register ndata args;    /* arg list or 0 */
{
    register data_id *symP;
    char *name;
    register int arg_count, ix;
    int numLevels;
    ndata object_var, methodId;
    register data_id *methodP;
    register node_ptr np;
    int dynamic_method_name;
    int is_broadcast;

    LOGGER(("EmitFuncCall: stmt? %d id %d args %d",is_stmt,id,args));
    symP = GetSymP(id);
    name = symP->name;

    if ( !streq(NXLT("NXLT"),name) && !streq(NXLT("XLT"),name) && !streq(NXLT("XLT2"),name) ) {
        symP = NeedFunc(id);
        XREF((Xrf,"\t%s\n",XRname(symP->name)));
        CheckBltinSym(symP);    /* set is_bltin, is_funckey bits, bltinpos */
        CODE_CHECK(ElfSysMode || !ElfIsProtected(symP->name), (XLT("%s available only to system programmers"),symP->name));
    }
    if ( args ) {
        arg_count = SizeList(args);
        for ( ix = arg_count-1; ix >= 0; ix-- )
            ExprEmitter(GetListMember(args,ix));

        if ( streq(NXLT("NXLT"),name) )
            ;       /* totally ignore NXLT */
        else if ( streq(NXLT("XLT"),name)  ) {
            ELFC_VERIFY(arg_count == 1,XLT("use XLT for one arg, XLT2 for two"));
            AddDoubleInst(OP_TRCALL,GetRegister());
        }
        else if ( streq(NXLT("XLT2"),name)  ) {
            ELFC_VERIFY(arg_count == 2,XLT("use XLT for one arg, XLT2 for two"));
            AddDoubleInst(OP_XTRCALL,GetRegister());
        }
        else if ( is_stmt ) {
            if ( symP->is_builtin ) {
                AddGenericInst(OP_BCALL);
                AddWord((shrt)symP->bltinpos);
                AddWord((shrt)arg_count);
            }
            else if ( symP->is_funckey ) {
                AddGenericInst(OP_KCALL);
                AddWord((shrt)symP->bltinpos);
                AddWord((shrt)arg_count);
            }
            else
                AddEaWithImmediateInst(OP_CALL,symP,(shrt)arg_count);
        }
        else {
            if ( symP->is_builtin ) {
                AddGenericInst(OP_RBCALL);
                AddWord((shrt)symP->bltinpos);
                AddWord((shrt)arg_count);
            }
            else if ( symP->is_funckey ) {
                AddGenericInst(OP_RKCALL);
                AddWord((shrt)symP->bltinpos);
                AddWord((shrt)arg_count);
            }
            else
                AddEaWithImmediateInst(OP_RCALL,symP,(shrt)arg_count);
            AddWord(GetRegister());
        }
        FreeNode(args);
    }
    else if ( is_stmt ) {
        if ( symP->is_builtin ) {
            AddGenericInst(OP_BCALL0);
            AddWord((shrt)symP->bltinpos);
        }
        else if ( symP->is_funckey ) {
            AddGenericInst(OP_KCALL0);
            AddWord((shrt)symP->bltinpos);
        }
        else
            AddEaInstruction(OP_CALL0,symP);
    }
    else {
        if ( symP->is_builtin ) {
            AddGenericInst(OP_RBCALL0);
            AddWord((shrt)symP->bltinpos);
        }
        else if ( symP->is_funckey ) {
            AddGenericInst(OP_RKCALL0);
            AddWord((shrt)symP->bltinpos);
        }
        else
            AddEaInstruction(OP_RCALL0,symP);
        AddWord(GetRegister());
    }
cleanup:;
}

/*************************************************************************
function: EmitGetObj - Emit code for an object method "get"
*/
static EmitGetObj (id,methodlist,numLevels,num_datums)
register ndata id;      /* id of function symbol */
register ndata methodlist;  /* 0 or object method list */
register int numLevels;     /* how many of methods to process */
register int num_datums;
{
    register data_id *symP;
    char *name;
    int thisLevel;
    ndata methodId;
    register data_id *methodP;
    register node_ptr np;
    int dynamic_method_name;
    int is_broadcast;
    int retval = 0;
    int objRegister;

    symP = NeedVariable(id);
    name = symP->name;
    if ( !symP->isObject ) {
        CODE_CHECK(FALSE,(
            XLT("Array or Formatted Array being used as a statement.")));
    }

    if ( numLevels == -1 )
        numLevels = SizeList(methodlist);
    thisLevel = numLevels - 1;
    if ( thisLevel != 0 ) {
        objRegister = EmitGetObj(id,methodlist,thisLevel,0);
        AddGenericInst(OP_TOSSTOS);
    }
    methodId = GetListMember(methodlist, thisLevel);
    if ( IS_NODE(methodId) ) {
        np = NODEPTR(methodId);
        switch ( np.any->data[2] ) {
        case TK_ENTRY:
            dynamic_method_name = TRUE;
            is_broadcast = FALSE;
            EmitExpression(np.any->data[0]);
            break;
        case TK_MULT:
            is_broadcast = TRUE;
            dynamic_method_name = FALSE;
            methodId = np.any->data[0];
            break;
        case TK_EXP:
            is_broadcast = 2;		/* false + call_inherit */
            dynamic_method_name = FALSE;
            methodId = np.any->data[0];
            break;
        case TK_MULT+TK_EXP:
            is_broadcast = 3;		/* true + call_inherit */
            dynamic_method_name = FALSE;
            methodId = np.any->data[0];
            break;
        case TK_ENTRY+TK_MULT:
            dynamic_method_name = TRUE;
            is_broadcast = TRUE;
            EmitExpression(np.any->data[0]);
            break;
        case TK_ENTRY+TK_EXP:
            dynamic_method_name = TRUE;
            is_broadcast = 2;		/* false + call_inherit */
            EmitExpression(np.any->data[0]);
            break;
        case TK_ENTRY+TK_MULT+TK_EXP:
            dynamic_method_name = TRUE;
            is_broadcast = 3;		/* true + call_inherit */
            EmitExpression(np.any->data[0]);
            break;
        default:
            CODE_CHECK(FALSE,(XLT("Invalid object method")));
        }
    }
    else {
#if 0
if ( thisLevel == 0 )
printf("initial ");
printf("get method %s\n",GetSymP(methodId)->name);
#endif
        dynamic_method_name = FALSE;
        is_broadcast = FALSE;
    }

    if ( thisLevel == 0 )
        AddEaInstruction(OP_SETOBJ,symP);
    else {
        AddGenericInst(OP_SETOBJ);
        AddWord(objRegister);
    }
    AddWord(retval = GetRegister());
    AddWord(is_broadcast);

    if ( dynamic_method_name ) {
        AddWord(0); /* tell intr to pop value off stack */
    }
    else {
        methodP = GetSymP(methodId);
/*      printf("  method name %s\n",methodP->name); */
        AddLiteralEaWord(GetSymP(AddStrConst(methodP->name)));
        AddWord(hash(methodP->name));
    }
    AddWord(num_datums);
cleanup:
    return(retval);
}

/*************************************************************************
function: EmitEntryCall - Emit code for a function call thru entry var
*/
EmitEntryCall (is_stmt,id,args)
register int is_stmt;   /* TRUE if return value is to be ignored */
register ndata id;      /* id of function symbol */
register ndata args;    /* arg list or 0 */
{
    register int arg_count, ix;

    LOGGER(("EmitEntryCall: stmt? %d id %d args %d",is_stmt,id,args));
    arg_count = (args ? SizeList(args) : 0);
    LOGGER(("  num_args %d",arg_count));
    for ( ix = arg_count-1; ix >= 0; ix-- ) {
        LOGGER(("   arg %d",ix));
        ExprEmitter(GetListMember(args,ix));
    }
    XREF((Xrf,"\t%s\n","***CALLBYNAME***"));
    EmitExpression(id);
    if ( is_stmt ) {
        if ( arg_count ) {
            AddGenericInst(OP_ENTRYCALL);
            AddWord(arg_count);
        }
        else
            AddGenericInst(OP_0ENTRYCALL);
    }
    else {
        if ( arg_count ) {
            AddGenericInst(OP_RENTRYCALL);
            AddWord(arg_count);
            AddWord(GetRegister());
        }
        else {
            AddGenericInst(OP_0RENTRYCALL);
            AddWord(GetRegister());
        }
    }
    if ( args )
        FreeNode(args);
cleanup:;
}

/*************************************************************************
function: NeedFunc - Get/Verify symbol is a function
*/
data_id *NeedFunc (id)
register shrt id;
{
    register data_id *symP;

    symP = GetSymP(id);
    if ( symP->type == TYPE_UNKNOWN )
        symP->type = TYPE_EXTFUNC;  /* assumption */
    CODE_CHECK(FUNC_TYPE(symP->type),(XLT("%s not a function"),symP->name));
cleanup:
    return(symP);
}


/*************************************************************************
function: EmitWhileStmt - start a while loop
*/
EmitWhileStmt (expr)
ndata expr;
{
    register while_loop_info *wp;

    LOGGER(("EmitWhileStmt: expr %d",expr));
    wp = CG.while_loops + CG.num_while_loops++;
    CODE_CHECK(CG.num_while_loops < MAX_WHILES,(XLT("nesting too deep")));
    wp->label = CrtLabelHere(CG.cur_label++);
    EmitExpression(expr);
    wp->end_label = CrtLabel(CG.cur_label++);
    AddEaInstruction(OP_JZ,GetSymP(wp->end_label));
cleanup:
    return;
}

/*************************************************************************
function: EmitWendStmt - end a while loop
*/
EmitWendStmt ()
{
    register while_loop_info *wp;

    LOGGER(("EmitWendStmt"));
    CODE_CHECK(CG.num_while_loops,(XLT("WEND without WHILE")));
    wp = CG.while_loops + (--CG.num_while_loops);
    AddEaInstruction(OP_JMP,GetSymP(wp->label));
    SetLabelHere(wp->end_label);
cleanup:
    return;
}

/*************************************************************************
function: EmitSwitch - Top of switch statement
*/
EmitSwitch (expr)
register ndata expr;
{
    register select_info *sp;

    EmitExpression(expr);
    sp = CG.selects + CG.num_selects++;
    CODE_CHECK(CG.num_selects < MAX_WHILES,(XLT("nesting too deep")));
    memset(sp,0,sizeof(select_info));
    sp->end_label = CrtLabel(CG.cur_label++);
    sp->data_label = CrtLabel(CG.cur_label++);
    sp->dflt_label = CrtLabel(CG.cur_label++);
    AddSrcMap();/* so label below is ON switch, not srcmap */
    sp->opcode_label = CrtLabelHere(CG.cur_label++);
    AddEaInstruction(OP_SWITCH,GetSymP(sp->data_label));
    AddEa(GetSymP(sp->dflt_label));
    sp->last_case_ended_pc = elcodeStr.SzCode;
cleanup:;
}

/*************************************************************************
function: EmitCase - Switch Case
!!0 DROPS THRU IF EMPTY CASE STATEMENT!!!!!
*/
EmitCase (expr,allowGroup)
register ndata expr;
int allowGroup;
{
    register select_info *sp;
    data_id *symP;
    select_item *sip;
    int is_data, ix, num_exprs;

    CODE_CHECK(CG.num_selects,(XLT("CASE without CASE OF")));

    expr = ReduceExpression(expr);

    is_data = IS_NODE(expr) && (NodeType(expr) == ND_NUMLIST);
    if ( is_data ) {
        num_exprs = SizeList(expr);
        for ( ix = 0; ix < num_exprs; ix++ )
            EmitCase(GetListMember(expr,ix),TRUE);
        return;
    }

    sp = CG.selects + CG.num_selects - 1;
    if ( expr == TK_TRUE ) {
        symP = GetSymP(elcodeStr.SymOfTrue);
        sp->is_numbers = TRUE;
    }
    else if ( expr == TK_FALSE ) {
        symP = GetSymP(elcodeStr.SymOfFalse);
        sp->is_numbers = TRUE;
    }
    else if ( IS_SYM(expr) ) {
        symP = GetSymP(expr);
        if ( symP->type == TYPE_INTEGER )
            sp->is_numbers = TRUE;
        else if ( symP->type == TYPE_STRING )
            sp->is_strings = TRUE;
        else {
            CODE_CHECK(FALSE,(XLT("only constant number or string allowed in case expression")));
        }
    }
    else {
        CODE_CHECK(FALSE,(XLT("only constant number or string allowed in case expression")));
    }
    if ( sp->is_numbers && sp->is_strings ) {
        sp->is_numbers = FALSE; /* reset so won't keep getting error messages */
        sp->is_strings = FALSE; /* reset so won't keep getting error messages */
        CODE_CHECK(FALSE,(XLT("case expressions must all be strings or all be numbers")));
    }
    for ( sip = sp->items; sip; sip = sip->next ) {
        CODE_CHECK(symP != sip->item,(XLT("duplicate case in switch")));
    }
    if ( !allowGroup || (elcodeStr.SzCode != sp->last_case_ended_pc) ) {
        AddEaInstruction(OP_JMP,GetSymP(sp->end_label));
        sp->last_case_ended_pc = elcodeStr.SzCode;
    }
    sip = (select_item*)ElfcAlloc(sizeof(select_item));
    sip->code_label = CrtLabelHere(CG.cur_label++);
    sip->item = symP;
    sip->next = sp->items;
    sp->items = sip;
cleanup:;
}

/*************************************************************************
function: EmitDefault - For swicth statement
*/
EmitDefault ()
{
    register select_info *sp;

    CODE_CHECK(CG.num_selects,(XLT("CASE without CASE OF")));
    sp = CG.selects + CG.num_selects - 1;
    CODE_CHECK(!sp->did_dflt,(XLT("more than one DEFAULT in switch")));
    AddEaInstruction(OP_JMP,GetSymP(sp->end_label));
    sp->last_case_ended_pc = elcodeStr.SzCode;
    sp->did_dflt = TRUE;
    SetLabelHere(sp->dflt_label);
cleanup:;
}

/*************************************************************************
function: EmitEndSwitch - end of switch
Puts out target label and array of values.
*/
EmitEndSwitch ()
{
    register select_info *sp;
    select_item *sip;
    int count;
    data_id *symP;

    CODE_CHECK(CG.num_selects,(XLT("ENDCASE without CASE OF")));
    sp = CG.selects + (--CG.num_selects);
    AddEaInstruction(OP_JMP,GetSymP(sp->end_label));

/* Emit the data list */

    SetLabelHere(sp->data_label);
    for ( sip = sp->items, count = 0; sip; sip = sip->next ) count++;
    AddWord(count);

    for ( sip = sp->items; sip; sip = sip->next ) {
        AddEa(sip->item);
        AddEa(GetSymP(sip->code_label));
    }
    symP = GetSymP(sp->opcode_label);
    assert(elcodeStr.CodeFrame[symP->locate] == OP_SWITCH);
    if ( sp->is_strings )
        elcodeStr.CodeFrame[symP->locate] = OP_SSWITCH;
    if ( !sp->did_dflt )
        SetLabelHere(sp->dflt_label);
    SetLabelHere(sp->end_label);

cleanup:
    return;
}

/*************************************************************************
function: EmitForStmt - start a for loop
*/
EmitForStmt (id,expr,stop,step)
register ndata id,expr,stop,step;
{
    register for_loop_info *fp;
    register shrt loop_num = CG.num_for_loops;

    LOGGER(("EmitForStmt: id %d expr %d stop %d step %d", id,expr,stop,step));
    CODE_CHECK(loop_num < MAX_FORS,(XLT("nesting too deep")));
    EmitExpression(stop);
    if ( step )
        EmitExpression(step);
    else
        PushNumber((data_comp_integer *)GetSymP(elcodeStr.SymOfNumberOne));
    fp = CG.for_loops + CG.num_for_loops++;
    EmitExpression(expr);
    AddEaInstruction(OP_FORLOOP,MustBeLocal(id));
    elcodeStr.LockedRegisters[fp->step_reg = GetRegister()] = TRUE;
    AddWord(fp->step_reg);
    elcodeStr.LockedRegisters[fp->stop_reg = GetRegister()] = TRUE;
    AddWord(fp->stop_reg);
    fp->id = id;
    fp->label = CrtLabelHere(CG.cur_label++);
    fp->next_label = CrtLabel(CG.cur_label++);
    fp->end_label = CrtLabel(CG.cur_label++);
    AddEaInstruction(OP_PUSH,MustBeLocal(id));
    AddEaInstruction(ElfIntMode?OP_IFORTEST:OP_FORTEST,GetSymP(fp->end_label));
    AddWord(fp->step_reg);
    AddWord(fp->stop_reg);
cleanup:;
}

/*************************************************************************
function: EmitBreakForStmt - Break out of for-loop
*/
EmitBreakForStmt (id)
register ndata id;
{
    register for_loop_info *fp;
    int ix;

    CODE_CHECK(CG.num_for_loops,(XLT("BREAK without FOR")));
    for ( ix = CG.num_for_loops - 1; ix >= 0; ix-- ) {
        fp = &CG.for_loops[ix];
        if ( fp->id == id ) {
            AddEaInstruction(OP_JMP,GetSymP(fp->end_label));
            return;
        }
    }
    CODE_CHECK(FALSE,(XLT("BREAK variable does not match any FOR variables")));
cleanup:
    return;
}

/*************************************************************************
function: EmitBreakWhileStmt - Break out of while-loop
*/
EmitBreakWhileStmt ()
{
    register while_loop_info *wp;

    CODE_CHECK(CG.num_while_loops,(XLT("NEXT WHILE without WHILE")));
    wp = CG.while_loops + CG.num_while_loops - 1;
    AddEaInstruction(OP_JMP,GetSymP(wp->end_label));
cleanup:
    return;
}


/*************************************************************************
function: EmitContinForStmt - Continue thru for-loop
*/
EmitContinForStmt (id)
register ndata id;
{
    register for_loop_info *fp;
    int ix;

    CODE_CHECK(CG.num_for_loops,(XLT("NEXT STEP without FOR")));
    for ( ix = CG.num_for_loops - 1; ix >= 0; ix-- ) {
        fp = &CG.for_loops[ix];
        if ( fp->id == id ) {
            AddEaInstruction(OP_JMP,GetSymP(fp->next_label));
            return;
        }
    }
    CODE_CHECK(FALSE,(XLT("NEXT STEP variable does not match any FOR variables")));
cleanup:
    return;
}

/*************************************************************************
function: EmitContinWhileStmt - Continue thru while-loop
*/
EmitContinWhileStmt ()
{
    register while_loop_info *wp;

    CODE_CHECK(CG.num_while_loops,(XLT("BREAK without WHILE")));
    wp = CG.while_loops + CG.num_while_loops - 1;
    AddEaInstruction(OP_JMP,GetSymP(wp->label));
cleanup:
    return;
}

/*************************************************************************
function: EmitNextStmt - end some for loops
*/
EmitNextStmt (idlist)
register ndata idlist;
{
    register for_loop_info *fp;
    register int num_ids, ix;

    LOGGER(("EmitNextStmt: ids %d",idlist));
    num_ids = idlist ? SizeList(idlist) : 1;
    for ( ix = 0; ix < num_ids; ix++ ) {
        CODE_CHECK(CG.num_for_loops,(XLT("NEXT without FOR")));
        CG.num_for_loops--;
        fp = CG.for_loops + CG.num_for_loops;
        if ( idlist ){
            CODE_CHECK(fp->id == GetListMember(idlist,ix),(XLT("NEXT variable does not match FOR variable"))); }
        SetLabelHere(fp->next_label);
        AddDoubleInst(OP_PUSH,fp->step_reg);
        AddEaInstruction(OP_PUSH,GetSymP(fp->id));
        AddDoubleInst(OP_ADD, GetRegister());
        AddEaInstruction(OP_POPTO,GetSymP(fp->id));
        AddEaInstruction(OP_JMP,GetSymP(fp->label));
        SetLabelHere(fp->end_label);
        elcodeStr.LockedRegisters[fp->step_reg] = FALSE;
        elcodeStr.LockedRegisters[fp->stop_reg] = FALSE;
    }
cleanup:
    return;
}

/*************************************************************************
function: EmitIfStmt - If expr then statement else statement
*/
EmitIfStmt (expr,thn,els)
register ndata expr,thn,els;
{
    register shrt else_starts, else_ends;

    LOGGER(("EmitIfStmt: expr %d then %d else %d",expr,thn,els));
    EmitExpression(expr);
    else_starts = CrtLabel(CG.cur_label++);
    AddEaInstruction(OP_JZ,GetSymP(else_starts));
    if ( !thn )
        AddGenericInst(OP_NOP);
    else
        EmitStatement(thn);
    if ( els ) {
        else_ends = CrtLabel(CG.cur_label++);
        AddEaInstruction(OP_JMP,GetSymP(else_ends));
        SetLabelHere(else_starts);
        EmitStatement(els);
        SetLabelHere(else_ends);
    }
    else
        SetLabelHere(else_starts);
}

/*************************************************************************
function: EmitAddTo - Emit a quick-add/sub if possible
*/
EmitAddTo (adder,id,expr)
register int adder;
ndata id,expr;
{
#if 0
    register data_id *symP, *xp;
    register double dval;
    register int val;

#if 0       /* eventually */
    if ( ((symP = GetSymP(id))->type == TYPE_LOCVAR) && IS_SYM(expr) ) {
        if ( (xp = GetSymP(expr))->type == TYPE_NUMBER ) {
            if ( ((dval = ((data_comp_number *)xp)->value) <= 32767) ||
                  (dval >= -32768) ) goto punt;
            val = dval;
        }
        else if ( (xp = GetSymP(expr))->type == TYPE_INTEGER ) {
            if ( ((val = ((data_comp_integer *)xp)->value) <= 32767) ||
                  (val >= -32768) ) goto punt;
        }
        else
            goto punt;
        if ( val == 1 )
            AddEaInstruction(adder ? OP_INC : OP_DEC,symP);
        else
            AddEaWithImmediateInst(adder ? OP_POPADD : OP_POPSUB,symP,val);
        return;
    }
punt:
#endif
    EmitExpression(id);
    EmitExpression(expr);
    AddDoubleInst((adder ? TK_ADD : TK_SUB) , GetRegister());
    AddEaInstruction(OP_POPTO,symP);
#endif
}

/*************************************************************************
function: EmitLabeledStmt - Create a label, emit statement
??x The following code crashes compiler rather than producing label redef err
        on depth goto a1,a2,a3,a4,a5
        passed_data = data
a1:     passed_data[dexes[0]] = data
a2:     passed_data[dexes[0],dexes[1]] = data
a3:     passed_data[dexes[0],dexes[1],dexes[2]] = data
a4:     passed_data[dexes[0],dexes[1],dexes[2],dexes[3]] = data
a5:     passed_data[dexes[0],dexes[1],dexes[2],dexes[3],dexes[4]] = data
up_one:
        depth = depth - 1
        on depth goto a1,a2,a3,a4,a5
        data = passed_data
a1:     data = passed_data[dexes[0]]
a2:     data = passed_data[dexes[0],dexes[1]]
a3:     data = passed_data[dexes[0],dexes[1],dexes[2]]
a4:     data = passed_data[dexes[0],dexes[1],dexes[2],dexes[3]]
a5:     data = passed_data[dexes[0],dexes[1],dexes[2],dexes[3],dexes[4]]
        goto start_over
*/
EmitLabeledStmt (label,stmt)
ndata label;
ndata stmt;
{
    register data_id *symP;

    symP = GetSymP(label);
    if ( symP->type == TYPE_UNKNOWN )
        symP->type = TYPE_LABEL;
    CODE_CHECK(symP->type == TYPE_LABEL,(XLT("%s not a label"),symP->name));
    CODE_CHECK(!symP->locate,(XLT("%s already in use"),symP->name));
    SetLabelHere(label);
    EmitStatement(stmt);
cleanup:
    return;
}

/*************************************************************************
function: EmitGotoStmt - Goto label
*/
EmitGotoStmt (label)
register ndata label;
{
    data_id *NeedLabel();

    AddEaInstruction(OP_JMP,NeedLabel(label));
}

/*************************************************************************
function: EmitCompGotoStmt - Computed Goto statement
*/
EmitCompGotoStmt (expr,labels)
register ndata expr,labels;
{
    register int num_labs, ix;
    data_id *NeedLabel();

    EmitExpression(expr);
    num_labs = SizeList(labels);
    AddGenericInst(OP_COMPGOTO);
    AddWord((shrt)num_labs);
    for ( ix = 0; ix < num_labs; ix++ )
        AddEa(NeedLabel(GetListMember(labels,ix)));
}

/*************************************************************************
function: EmitCompCallStmt - Computed Call statement
This is implemented in the brute force manner of creating a group
of function calls, each followed by a goto to the statement following this
computed call.
*/
EmitCompCallStmt (expr,funcs)
register ndata expr,funcs;
{
    register int num_funcs, ix, count;
    register ndata bypass;  /* addr after this instruction */
    data_id *NeedFunc();

    EmitExpression(expr);
    num_funcs = SizeList(funcs);
    AddGenericInst(OP_COMPCALL);
    AddWord((shrt)num_funcs);
    bypass = CrtLabel(CG.cur_label++);
    for ( ix = 0; ix < num_funcs; ix++ ) {
        count = elcodeStr.SzCode;
        EmitFuncCall(TRUE,GetListMember(funcs,ix),0,0);
        if ( elcodeStr.SzCode - count != 2 ){   /* pad single word calls with NOP */
            assert((elcodeStr.SzCode - count) == 1);
            AddWord(OP_NOP);
        }
        AddEaInstruction(OP_JMP,GetSymP(bypass));
        assert((elcodeStr.SzCode - count) == 4);
    }
    SetLabelHere(bypass);
cleanup:;
}

/*************************************************************************
function: NeedLabel - Verify/Set passed sym id is a label
*/
data_id *NeedLabel (label)
register ndata label;
{
    register data_id *symP;

    symP = GetSymP(label);
    if ( symP->type == TYPE_UNKNOWN )
        symP->type = TYPE_LABEL;
    CODE_CHECK(symP->type==TYPE_LABEL,(XLT("%s not a label"),symP->name));
cleanup:
    return(symP);
}

/*************************************************************************
function: SizeList - return number of elements in a stmt_list/num_list/arg_list
*/

SizeList (list)
register ndata list;
{
    register node_ptr np;
    register int count = 1;
#if SGI
    register ndata x;
#endif  

    if ( list == 0 )
        return(0);
#if SGI     /* silly optimizer bug in SGI cannot build standalone libs */
            /* The SGI OS 3 Compiler has trouble with "np.any->data[0] ==0" */
    np = NODEPTR(list);
    while ( TRUE ) {
        x = np.any->data[0];
        if ( !x )
            break;
        np = NODEPTR(x);
        count++;
    }

#else   /* working compiler optimizer */

    for ( np = NODEPTR(list); np.any->data[0]; np = NODEPTR(np.any->data[0]) )
        count++;

#endif  /* SGI optimizer bug */

    return(count);
}

/*************************************************************************
function: GetListMember - Get indexed member from list
The index is from the physical end of the list, which maps to the first statement encountered in the source.
*/
ndata GetListMember (list,dex)
ndata list;
register int dex;
{
    register node_ptr np;
    register int ix;

    dex = SizeList(list) - dex - 1;
    for ( np = NODEPTR(list), ix = 0; ix < dex; ix++ )
        np = NODEPTR(np.any->data[0]);
    return(np.any->data[1]);
}

/*************************************************************************
function: GetListMemberAndAttr - Get indexed member from list
The index is from the physical end of the list, which maps to the first statement encountered in the source.
Also get data[2] into passed attrP... Needed only by emit_order_names.
The right fix would have been to syntax emit_order_names correctly.
*/
ndata GetListMemberAndAttr (list,dex,attrP)
ndata list;
register int dex;
ndata *attrP;
{
    register node_ptr np;
    register int ix;

    dex = SizeList(list) - dex - 1;
    for ( np = NODEPTR(list), ix = 0; ix < dex; ix++ )
        np = NODEPTR(np.any->data[0]);
    *attrP = np.any->data[2];
    return(np.any->data[1]);
}

/*************************************************************************
function: RefListToArray - Convert an array-reference-list to an array of refs
*/
RefListToArray (symP,list,resultP,countP)
data_id *symP;      /* var to use for decoding format names */
ndata list;
ndata *resultP;
int *countP;
{
    register int count, ix, dex = 0, jx;
    register ndata info;
    register char *p;
    register struct format_info *fP, *newP;

    count = SizeList(list);
    fP = symP->format;
    for ( ix = 0; ix < count; ix++ ){
        CODE_CHECK(dex<MAX_DIMS,(XLT("nesting too deep")));  /* ??c poor msg */
        info = GetListMember(list,ix);
        if ( IS_SYM(info) ) {   /* format name */
            CODE_CHECK(fP,(XLT("format not specified for %s"),symP->name));
            p = GetSymP(info)->name;
            for ( jx = 0; fP->names[jx]; jx++ ) {
                if ( streq(p,fP->names[jx]) ) {
                    resultP[dex++] = AddIntConst(jx);
                    if ( fP->formats[jx] ) {
                        newP = FindFormat(fP->formats[jx]);
                        CODE_CHECK(newP,(XLT("no such format as %s"),fP->formats[jx]));
                        fP = newP;
                    }
                    else
                        fP = NULL;
                    goto got_it;
                }
            }
            CodeError(XLT("name %s not in format %s"),p,fP->myname);
got_it:;
        }
        else if(info) {
            node_ptr np;
            np = NODEPTR(info);
            if(np.any->type == ND_ENTRYFUNC && !symP->isObject) {
                CodeError(XLT("<%s> is not an object, it cannot reference methods"),symP->name);
                *countP = 0;
                return;
            }
            for ( jx = 0; jx < SizeList(info); jx++ ) {
                CODE_CHECK(dex<MAX_DIMS,(XLT("nesting too deep")));  /* ??c poor msg */
                resultP[dex++] = GetListMember(info,jx);
            }
        }
    }
    *countP = dex;
    return;
cleanup:
    *countP = 0;
}

/*************************************************************************
function: NeedVariable - GetSymP a variable
*/
data_id *NeedVariable (id)
register ndata id;
{
    register data_id *p;

    p = GetSymP(id);
    if ( !VAR_TYPE(p->type) ) {
        CodeError(XLT("%s not declared"),p->name);
        p->type = TYPE_LOCVAR;
    }
    return(p);
}

/*************************************************************************
function: MustBeLocal - Verify symbol is a local var, return the symP
*/
data_id *MustBeLocal (id)
register ndata id;
{
    register data_id *p;

    p = NeedVariable(id);
    if ( p->type != TYPE_LOCVAR ) {
        CodeError(XLT("variable must be a local variable"),p->name);
        p->type = TYPE_LOCVAR;
    }
    return(p);
}


/*************************************************************************
function: AddSymWords - Add some words to symbols frame
*/
AddSymWords (p,c)
register shrt *p;
register int c;
{
#if ALPHA || (LINUX && __alpha__)
    register int pad = 8-(c & 7);
#else
#if SGI || HP800 || M_88K || CLIX || NT_MIPS || NT_PPC || (LINUX && __sparc__)
    register int pad = 4-(c & 3);
#else
    register int pad = c & 1;
#endif
#endif

    while ( --c >= 0 ) {
        if ( elcodeStr.SzSyms >= MAX_SYMS_WORDS )
    {
       printf("Crash\n");
           ErrOut(XLT("Macro too big, symbols frame overflow"));
    }
        elcodeStr.SymsFrame[elcodeStr.SzSyms++] = *p++;
    }
    while ( --pad >= 0) {
        if ( elcodeStr.SzSyms >= MAX_SYMS_WORDS )
           ErrOut(XLT("Macro too big, symbols frame overflow"));
        elcodeStr.SymsFrame[elcodeStr.SzSyms++] = 0;
    }
}

/*************************************************************************
function: AddSrcMap - If source info enabled, add source info
*/
AddSrcMap ()
{
    if ( elcodeStr.LineLastNoted == CGcurLine )
        return;
    elcodeStr.LineLastNoted = CGcurLine;
    if ( !AllowSourceInfo && !ElfcTimeSliceMode )
        return;
    if ( AllowSourceInfo && ElfcTimeSliceMode ) {
        AddWord(OP_SRCTIMECHECK);
        AddWord(elcodeStr.LineLastNoted);
    }
    else if ( ElfcTimeSliceMode )
        AddWord(OP_TIMECHECK);
    else {
        AddWord(OP_SRCMAP);
        AddWord(elcodeStr.LineLastNoted);
    }
}

/*************************************************************************
function: Align4Word - Insert NOPS to get PC to 64bit boundary
This is used to allow addressing capabilyt of 256Kbytes...(+-32768 * 8bytes)
*/
Align4Word ()
{
    while ( elcodeStr.SzCode & 0x03 )
        AddWord(OP_NOP);
}

/*************************************************************************
function: AddWord - Add a word to the code frame
Complain if overflow
*/
AddWord (x)
register shrt x;
{
    if ( elcodeStr.SzCode == MAX_CODE_WORDS )
        ErrOut(XLT("Macro too big, code area overflow"));
    elcodeStr.CodeFrame[elcodeStr.SzCode++] = x;
}

/*************************************************************************
function: AddWords - Add some words to code frame
*/
AddWords (p,c)
register shrt* p;
register int c;
{
    while ( --c >= 0 )
        AddWord(*p++);
}

/*************************************************************************
function: AddImmediateInst - Add a two word instruction
*/
AddImmediateInst (x,i)
register shrt x, i;
{
    AddSrcMap();
    AddWord(x);
    AddWord(i);
}

/*************************************************************************
function: AddGenericInst - Add a single word instruction
*/
AddGenericInst (x)
register shrt x;
{
    AddSrcMap();
    AddWord(x);
}

/*************************************************************************
function: AddDoubleInst - Add a double word instruction
*/
AddDoubleInst (x,reg)
register shrt x, reg;
{
    AddSrcMap();
    AddWord(x);
    AddWord(reg);
}

/*************************************************************************
function: AddEaInstruction - Add instruction with eff addr
*/
AddEaInstruction (x,symP)
register shrt x;
register data_id *symP;
{
    AddSrcMap();
    AddWord(x);
    AddEaCanMerge(symP);
}

/*************************************************************************
function: AddEa - Add eff addr ref to code frame
This can only be used for PC addrs.  Variable addrs must be routed
thru the AddEaCanMerge which modifies the opcode.
*/
AddEa (symP)
register data_id *symP;
{
    elcodeStr.EaRefs[elcodeStr.SzCode] = symP;
    AddWord(0);
}

/*************************************************************************
function: AddLiteralEaWord - Add literal reference to code
*/
AddLiteralEaWord (symP)
register data_id *symP;
{
    elcodeStr.EaRefs[elcodeStr.SzCode] = symP;
    AddWord(0);
}

/*************************************************************************
function: AddEaCanMerge - Add eff addr ref to code frame
If the sym type is a GLOVAR/EXTVAR/OBJVAR/SYSVAR then
if the opcode/ea is a single word type (OP1) this converts it to a
two word instruction (OP2).
*/
AddEaCanMerge (symP)
register data_id *symP;
{
    register int last = elcodeStr.SzCode-1;

    if ( (symP->type == TYPE_GLOVAR) || (symP->type == TYPE_EXTVAR) ) {
        elcodeStr.CodeFrame[last] = elcodeStr.CodeFrame[last] + OP_GLO_OFFSET;
        elcodeStr.EaRefs[elcodeStr.SzCode] = symP;
        AddWord(0);
    }
    else if ( symP->type == TYPE_OBJVAR ) {
	    CODE_CHECK(elcodeStr.IsMethod,
				   (XLT("%s objvar reference not allowed in macro/function"),symP->name));
        elcodeStr.CodeFrame[last] = elcodeStr.CodeFrame[last] + OP_OBJ_OFFSET;
        elcodeStr.EaRefs[elcodeStr.SzCode] = symP;
        AddLiteralEaWord(GetSymP(AddStrConst(symP->name))); /* var name */
        AddWord(hash(symP->name));                          /* hash value */
    }
    else if ( symP->type == TYPE_SYSVAR ) {
        elcodeStr.CodeFrame[last] = elcodeStr.CodeFrame[last] + OP_SYS_OFFSET;
        elcodeStr.EaRefs[elcodeStr.SzCode] = symP;
        AddLiteralEaWord(GetSymP(AddStrConst(symP->name))); /* var name */
        AddWord(hash(symP->name));                          /* hash value */
    }
    else {
        elcodeStr.EaRefs[elcodeStr.SzCode] = symP;
        AddWord(0);
    }
cleanup:;
}

/*************************************************************************
function: AddEaWithImmediateInst - Add instruction with eff addr & immed word
*/
AddEaWithImmediateInst (x,symP,immed)
register shrt x;
register data_id *symP;
register shrt immed;
{
    AddSrcMap();
    AddWord(x);
    AddEaCanMerge(symP);
    AddWord(immed);
}

/*************************************************************************
function: SymToEa - Convert symbol ptr to effective addr
*/
shrt SymToEa (p)
data_id *p;
{
    return(((shrt*)p)-elcodeStr.SymsFrame);
}

/*************************************************************************
function: CrtLabel - Crt an internal label
*/
shrt CrtLabel (num)
shrt num;
{
    register shrt id;
    register data_id *symP;
    char name[10];

    sprintf(name,"^%d",num);
    id = AddIdentifier(name);
    symP = GetSymP(id);
    symP->type = TYPE_LABEL;
    LOGGER(("CrtLabel: %s",symP->name));
    return(id);
}

/*************************************************************************
function: CrtLabelHere - Crt an internal label at current code location
*/
shrt CrtLabelHere (num)
shrt num;
{
    return(SetLabelHere(CrtLabel(num)));
}

/*************************************************************************
function: SetLabelHere - Set location of created label to current code loc
*/
shrt SetLabelHere (id)
register shrt id;
{
    register data_id *symP;

    symP = GetSymP(id);
    symP->locate = elcodeStr.SzCode;
    LOGGER(("SetLabelHere: %s at %d",symP->name,symP->locate));
    return(id);
}

/*************************************************************************
function: GetRegister - Allocate a register for temp usage
*/
GetRegister ()
{
    register int retval;

    if ( retval = elcodeStr.RegistersAvailable[0] ) {
        memcpy(elcodeStr.RegistersAvailable,elcodeStr.RegistersAvailable+1,MAX_REGS-1);
        return(retval);
    }
    while ( elcodeStr.LockedRegisters[elcodeStr.NextRegister] ) elcodeStr.NextRegister--;
    elcodeStr.LowestRegister = MIN(elcodeStr.LowestRegister,elcodeStr.NextRegister);
    CODE_CHECK(elcodeStr.NextRegister > 0,(XLT("expression too complex, break into smaller expressions")));
    elcodeStr.NextRegister--;
    return(elcodeStr.NextRegister+1);
cleanup:
    return(elcodeStr.NextRegister = MAX_REGS-1);
}

/*************************************************************************
function: RlsRegister - Release a register so rest of statemtn can use it
This is ONLY called by the Multi-dimension array assignent which goes
thru registers like mad on a per-statement basis.
*/
RlsRegister (reg)
int reg;
{
    register char *p;

    for ( p = elcodeStr.RegistersAvailable; *p != 0; p++ );
    *p = reg;
}

/*************************************************************************
function: AddIdentifier - Add an identifier if not already there, rtn id
*/
AddIdentifier (p)
register char *p;
{
    return(AddOrSrchForId(p,FALSE));
}

/*************************************************************************
function: AddOrSrchForId - Add or srch for existing identifer
*/
AddOrSrchForId (p,force)
register char *p;
register int force; /* TRUE if must add */
{
    register int ix;
    register data_id *symP;
    register int hashval;
    union {
        data_id datum;
        shrt buf[1000];
    } store;
    register int sym_to_re_use = -1;

    hashval = hash(p);
    if ( !force ) {
        for ( ix = NumSyms-1; ix >= 0; ix-- ) {
            if ( (elcodeStr.SymHash[ix] == hashval) && elcodeStr.SymTab[ix]
                    && SYM_IS_ID(symP = (data_id*)(elcodeStr.SymTab[ix]+elcodeStr.SymsFrame))
                    && streq(symP->name,p) )
                return(ix+SYM_FUDGE);
        }
        for ( ix = NumSyms-1; ix >= 0; ix-- ) {
            if ( !elcodeStr.SymTab[ix] ) {
                sym_to_re_use = ix;
                break;
            }
        }
    }
    ix = (sym_to_re_use == -1) ? NumSyms : sym_to_re_use;
    ELFC_VERIFY(NumSyms < MAX_SYMS-1,XLT("Macro too big, too many symbols"));
    LOGGER(("AddIdentifier: adding %s, val %d",p,ix));
    elcodeStr.SymHash[ix] = hashval;
    elcodeStr.SymTab[ix] = elcodeStr.SzSyms;
    MEMSET(&store.datum,0,sizeof(store.datum));
    strcpy(store.datum.name,p);
    AddSymWords(store.buf,(sizeof(store.datum)+strlen(p)+1)/2);
    return( ((sym_to_re_use == -1) ? NumSyms++ : sym_to_re_use) + SYM_FUDGE );
}

/*************************************************************************
function: CheckBltinSym - Scan builtins tables for symbol
*/
static CheckBltinSym (symP)
register data_id *symP;
{
    int code;

    if ( symP->blt_checked )
        return;
    symP->blt_checked = TRUE;
    if ( IsBltinSymCode(code = ElfcLookupSymbol(symP->name,ElfcLanguage)) ) {
        symP->is_builtin = TRUE;
        symP->bltinpos = code;
    }
}

/*************************************************************************
function: AddStrConst - Add string constant to symbol table
*/
AddStrConst (p)
register char *p;
{
    register int ix;
    register int hashval;
    register data_comp_string *symP;
    union {
        data_comp_string datum;
        shrt buf[1000];
        char plenty[sizeof(data_comp_string)+5001];
    } store;
    register int sym_to_re_use = -1;

    ELFC_VERIFY(strlen(p)<5000,XLT("String constant too big"));
    hashval = hash(p);
    for ( ix = 0; ix < NumSyms; ix++ ) {
        if ( (elcodeStr.SymHash[ix] == hashval) && elcodeStr.SymTab[ix] ) {
            symP = (data_comp_string*)(elcodeStr.SymTab[ix]+elcodeStr.SymsFrame);
            if ((symP->type == TYPE_STRING) && streq(symP->text,p))
                return(ix+SYM_FUDGE);
        }
    }
    for ( ix = NumSyms-1; ix >= 0; ix-- ) {
        if ( !elcodeStr.SymTab[ix] ) {
            sym_to_re_use = ix;
            break;
        }
    }
    ix = (sym_to_re_use == -1) ? NumSyms : sym_to_re_use;
    ELFC_VERIFY(NumSyms < MAX_SYMS-1,XLT("Macro too big, too many symbols"));
    LOGGER(("AddStrConst: adding %s, val %d",p,(sym_to_re_use == -1) ? NumSyms : sym_to_re_use));
    elcodeStr.SymHash[ix] = hashval;
    elcodeStr.SymTab[ix] = elcodeStr.SzSyms;
    MEMSET(&store.datum,0,sizeof(store.datum));
    strcpy(store.datum.text,p);
    store.datum.type = TYPE_STRING;
    AddSymWords((shrt*)(&store.datum),(sizeof(store.datum)+strlen(p)+1)/2);
    return( ((sym_to_re_use == -1) ? NumSyms++ : sym_to_re_use) + SYM_FUDGE );
}

/*************************************************************************
function: AddNumConst - Add number constant to symbol table
*/
AddNumConst (num)
double num;
{
    register int *tabP, *endP = elcodeStr.SymTab+NumSyms;
    register data_comp_number *symP;
    union {
        data_comp_number datum;
        shrt buf[1000];
    } store;
    register int sym_to_re_use = -1;

    for ( tabP = elcodeStr.SymTab; tabP < endP; tabP++ ) {
        if ( *tabP == 0 )
            sym_to_re_use = tabP-elcodeStr.SymTab;
        else {
            symP = (data_comp_number*)(*tabP+elcodeStr.SymsFrame);
            if ( (symP->type == TYPE_NUMBER) && (symP->value == num) )
                return((tabP-elcodeStr.SymTab)+SYM_FUDGE);
        }
    }
    ELFC_VERIFY(NumSyms < MAX_SYMS-1,XLT("Macro too big, too many symbols"));
    LOGGER(("AddNumConst: adding %g, val %d",num,(sym_to_re_use == -1) ? NumSyms : sym_to_re_use));
    elcodeStr.SymTab[(sym_to_re_use == -1) ? NumSyms : sym_to_re_use] = elcodeStr.SzSyms;
    MEMSET(&store.datum,0,sizeof(store.datum));
    store.datum.value = num;
    store.datum.type = TYPE_NUMBER;
    AddSymWords(store.buf,sizeof(store.datum)/2);
    return( ((sym_to_re_use == -1) ? NumSyms++ : sym_to_re_use) + SYM_FUDGE );
}
/*************************************************************************
function: AddIntConst - Add an integer constant to symbol table
*/
AddIntConst (num)
int num;
{
    register int *tabP, *endP = elcodeStr.SymTab+NumSyms;
    register data_comp_integer *symP;
    union {
        data_comp_integer datum;
        shrt buf[1000];
    } store;
    register int sym_to_re_use = -1;

    for ( tabP = elcodeStr.SymTab; tabP < endP; tabP++ ) {
        if ( *tabP == 0 )
            sym_to_re_use = tabP-elcodeStr.SymTab;
        else {
            symP = (data_comp_integer*)(*tabP+elcodeStr.SymsFrame);
            if ( (symP->type == TYPE_INTEGER) && (symP->value == num) )
                return((tabP-elcodeStr.SymTab)+SYM_FUDGE);
        }
    }
    ELFC_VERIFY(NumSyms < MAX_SYMS-1,XLT("Macro too big, too many symbols"));
    LOGGER(("AddIntConst: adding %d, val %d",num,(sym_to_re_use == -1) ? NumSyms : sym_to_re_use));
    elcodeStr.SymTab[(sym_to_re_use == -1) ? NumSyms : sym_to_re_use] = elcodeStr.SzSyms;
    MEMSET(&store.datum,0,sizeof(store.datum));
    store.datum.value = num;
    store.datum.type = TYPE_INTEGER;
    AddSymWords(store.buf,sizeof(store.datum)/2);
    return( ((sym_to_re_use == -1) ? NumSyms++ : sym_to_re_use) + SYM_FUDGE );
}


/*************************************************************************
function: GetSymP - Get ptr to symbol table entry data
*/
data_id *GetSymP (id)
register shrt id;
{
    assert(IS_SYM(id));
    return( (data_id*)(elcodeStr.SymsFrame + elcodeStr.SymTab[id-SYM_FUDGE]) );
}

/*************************************************************************
function: NodeType - Extract type from node, verify is a node
*/
NodeType (n)
register shrt n;
{
    assert(IS_NODE(n));
    return((NodeArray[n-NODE_FUDGE]).any->type);
}

/*************************************************************************
function: FindFormat - Find a format by name
Returns NULL if no such format.
*/
format_info *FindFormat (name)
register char *name;
{
    register format_info *p;

    for ( p = Formats; p; p = p->next ) {
        if ( streq(p->myname,name) )
            return(p);
    }
    return(NULL);
}

/*************************************************************************
function: UseTagSubFormats - Tag subformats as being used if this is used
*/
void UseTagSubFormats (p)
register format_info *p;
{
    register format_info *subP;
    register int ix;

    if ( !p->used )
        return;
    for ( ix = 0; (ix < MAX_FORMATS) && p->names[ix]; ix++ ) {
        if ( !p->formats[ix] )
            continue;
        subP = FindFormat(p->formats[ix]);
        CODE_CHECK(subP,(XLT("no such format as %s"),p->formats[ix]));
        if ( !subP->used ) {
            subP->used = TRUE;
            UseTagSubFormats(subP);
        }
    }
cleanup:;       /* for CODE_CHECK above */
}

/*************************************************************************
function: FormatsData - Put format info for debugger & recorder
Loads the EWlcVarnames array, updating ElfcVarnamePos.
*/
data_ptr FormatsData()
{
#define MAX_FMTINFO 20000

    register format_info *p;
    register int ix;
    data_ptr allP;
    data_ptr formP;
    data_ptr elementP;
    char buf[2000]; /* plenty! */

    allP = MkArray(0);
    
    for ( p = Formats; p; p = p->next )
        UseTagSubFormats(p);

    for ( p = Formats; p; p = p->next ) {
        if ( !p->used )
            continue;

        formP = MkArray(1);
        formP = ElfAddStrToArray(formP,0,p->myname);

/* load element names and formats */

        for ( ix = 0; (ix < MAX_FORMATS) && p->names[ix]; ix++ ) {
            if ( p->formats[ix] ) {
                elementP = MkArray(2);
                elementP = ElfAddStrToArray(elementP,0,p->names[ix]);
                if ( p->formatDim[ix] )
                    encodeFormatDim(buf,p->formatDim[ix],p->formats[ix]);
                else
                    strcpy(buf,p->formats[ix]);
                elementP = ElfAddStrToArray(elementP,1,buf);
                formP = WriteArray(formP,ElfbArraySize(formP),elementP);
            }
            else
                formP = ElfAddStrToArray(formP,ElfbArraySize(formP),p->names[ix]);
        }
        allP = WriteArray(allP,ElfbArraySize(allP),formP);
    }
    return(allP);
}

#if FOR_TESTER
static char *PackBuf;
static int PackSize = 0;

/*************************************************************************
function: ElfcPackDatum - Re-package Elf array to fit in minimum memory
*/
data_ptr ElfcPackDatum (inP)
data_ptr inP;
{
    int ix, cnt, type;
    data_ptr dP;
    int size;

    if ( !PackBuf )
        PackBuf = (char*)malloc(1000000);

    if ( !inP || IS_INTEGER(inP) )
        return(inP);
    type = inP->type;
    switch ( type ) {

    case TYPE_NUMBER:
        while ( PackSize & 0x07 )       /* pad out to nearest 8-byte */
            PackSize++;
        size = sizeof(data_number);
        break;

    case TYPE_STRING:
        while ( PackSize & 0x03 )       /* pad out to nearest 4-byte */
            PackSize++;
        size = sizeof(data_string) + strlen(SX(inP)->text);
        break;

    case TYPE_BINARY:
        while ( PackSize & 0x03 )       /* pad out to nearest 4-byte */
            PackSize++;
        size = sizeof(data_binary) + BX(inP)->size - 1;
        break;

    case TYPE_ARRAY:
        cnt = AX(inP)->size;
        for ( ix = 0; ix < cnt; ix++ ) {
            dP = AX(inP)->vals[ix];
            dP = ElfcPackDatum(dP);
            AX(inP)->vals[ix] = dP;
        }
        while ( PackSize & 0x07 )       /* pad out to nearest 8-byte */
            PackSize++;
        size = sizeof(data_array) + (sizeof(data_ptr) * (cnt - 1));
    }

    memcpy(PackBuf+PackSize,(char*)inP,size);
    dP = (data_ptr)(PackBuf + PackSize);
    PackSize += size;
    return(dP);
}
#endif
	
/*************************************************************************
function: WriteCodeArray - Return Elf array of result
*/
data_ptr WriteCodeArray (src)
char *src;
{
    extern FILE *ElfCompSrcFp;  /* non-null if should created real file */
    FILE *OpenFileObjFmode();
    register int cdex, ix, span;
    register data_id *symP;
    data_ptr allP, headerP, gvarsP, gfuncsP, afuncP, avarP, litsP, extsP, extP;
    data_ptr codeP, varnamesP, formatsP, ovarsP, svarsP;
    char *name;
    int nameid;

    allP = MkArray(10);
    headerP = MkArray(1);
    gvarsP = MkArray(0);
    ovarsP = MkArray(0);
    gfuncsP = MkArray(0);
    litsP = MkArray(1);     /* bottom slot cannot be used cause of OP_SETOBJ */
    extsP = MkArray(0);
    svarsP = MkArray(0);

    headerP = ElfAddNumToArray(headerP,0,APX_MASTER_REVSTAMP);
    headerP = ElfAddStrToArray(headerP,1,src);
    headerP = ElfAddNumToArray(headerP,2,AllowSourceInfo);

/* Store Global Vars */

    for ( ix = 0; ix < NumSyms; ix++ ) {
        if ( (symP = GetSymP((shrt)(ix+SYM_FUDGE)))->type != TYPE_GLOVAR )
            continue;
        symP->locate = ElfbArraySize(gvarsP);
        avarP = MkArray(1);
        avarP = ElfAddStrToArray(avarP,0,symP->name);
        if ( symP->format ) {
            if ( symP->formatDim ) {
                char buf[200];
                encodeFormatDim(buf,symP->formatDim,symP->format->myname);
                avarP = ElfAddStrToArray(avarP,1,buf);
            }
            else
                avarP = ElfAddStrToArray(avarP,1,symP->format->myname);
            symP->format->used = TRUE;
        }
        gvarsP = WriteArray(gvarsP,symP->locate,avarP);
    }

/* Store Object Vars info for Debugger */

    for ( ix = 0; ix < NumSyms; ix++ ) {
        if ( (symP = GetSymP((shrt)(ix+SYM_FUDGE)))->type != TYPE_OBJVAR )
            continue;
        avarP = MkArray(1);
        avarP = ElfAddStrToArray(avarP,0,symP->name);
        if ( symP->format ) {
            if ( symP->formatDim ) {
                char buf[200];
                encodeFormatDim(buf,symP->formatDim,symP->format->myname);
                avarP = ElfAddStrToArray(avarP,1,buf);
            }
            else
                avarP = ElfAddStrToArray(avarP,1,symP->format->myname);
            symP->format->used = TRUE;
        }
        ovarsP = WriteArray(ovarsP,ElfbArraySize(ovarsP),avarP);
    }

/* Store System Vars info for Debugger */

    for ( ix = 0; ix < NumSyms; ix++ ) {
        if ( (symP = GetSymP((shrt)(ix+SYM_FUDGE)))->type != TYPE_SYSVAR )
            continue;
        avarP = MkArray(1);
        avarP = ElfAddStrToArray(avarP,0,symP->name);
        if ( symP->format ) {
            if ( symP->formatDim ) {
                char buf[200];
                encodeFormatDim(buf,symP->formatDim,symP->format->myname);
                avarP = ElfAddStrToArray(avarP,1,buf);
            }
            else
                avarP = ElfAddStrToArray(avarP,1,symP->format->myname);
            symP->format->used = TRUE;
        }
        svarsP = WriteArray(svarsP,ElfbArraySize(svarsP),avarP);
    }

/* Store Global Funcs info for others to dynlink to (quad aligned) */

    for ( ix = 0; ix < NumSyms; ix++ ) {
        register int type;
        symP = GetSymP((shrt)(ix+SYM_FUDGE));
        type = symP->type;
        if ( (type != TYPE_QFUNC) && 
             (type != TYPE_UIFUNC) &&
             (type != TYPE_NRFUNC) &&
             (type != TYPE_GETOBJ) &&
             (type != TYPE_SETOBJ) &&
             (type != TYPE_MACRO) )
            continue;

		if ( symP->isPrivate &&
			((type != TYPE_GETOBJ) && (type != TYPE_SETOBJ)) )
			continue;		/* don't export private funcs */

        name = symP->name;
        afuncP = MkArray(4);
        afuncP = ElfAddStrToArray(afuncP,0,name);
        afuncP = ElfAddNumToArray(afuncP,1,symP->type);
		if ( symP->isPrivate )
			afuncP = ElfAddNumToArray(afuncP,2,-1);
		else
			afuncP = ElfAddNumToArray(afuncP,2,symP->srcLine);
        afuncP = ElfAddNumToArray(afuncP,3,symP->locate);
        gfuncsP = WriteArray(gfuncsP,ElfbArraySize(gfuncsP),afuncP);
    }

/* Create Global Funcs for Public Objvars */

	if ( elcodeStr.HavePublicObjvars ) {
		for ( ix = 0; ix < NumSyms; ix++ ) {
			register int type;
		    char vname[ELFIDSTR+1];
			symP = GetSymP((shrt)(ix+SYM_FUDGE));
			type = symP->type;
			if ( type != TYPE_OBJVAR || !symP->isPublic )
				continue;
			strcpy(vname+1,symP->name);
			/* Create GET function if not already existant */
			vname[0] = '?';
			symP = GetSymP(AddIdentifier(vname));
			if ( symP->type == 0 ) {
		        afuncP = MkArray(4);
		        afuncP = ElfAddStrToArray(afuncP,0,vname);
		        afuncP = ElfAddNumToArray(afuncP,1,TYPE_GETOBJ);
		        afuncP = ElfAddNumToArray(afuncP,2,0);
		        afuncP = ElfAddNumToArray(afuncP,3,0);
		        gfuncsP = WriteArray(gfuncsP,ElfbArraySize(gfuncsP),afuncP);
			}
			/* Create SET function if not already existant */
			vname[0] = '!';
			symP = GetSymP(AddIdentifier(vname));
			if ( symP->type == 0 ) {
		        afuncP = MkArray(4);
		        afuncP = ElfAddStrToArray(afuncP,0,vname);
		        afuncP = ElfAddNumToArray(afuncP,1,TYPE_SETOBJ);
		        afuncP = ElfAddNumToArray(afuncP,2,0);
		        afuncP = ElfAddNumToArray(afuncP,3,0);
		        gfuncsP = WriteArray(gfuncsP,ElfbArraySize(gfuncsP),afuncP);
			}
		}
    }
	
	/* sort Global Funcs */
	ElfSortGfuncs(gfuncsP);
	
/* Store Dynamic linkage info for externs */

    for ( ix = 0; ix < NumSyms; ix++ ) {
        symP = GetSymP((shrt)(ix+SYM_FUDGE));
        if ( (symP->type != TYPE_EXTFUNC) && (symP->type != TYPE_EXTVAR) )
            continue;
        name = symP->name;
        symP->locate = ElfbArraySize(extsP);
        extP = MkArray(2);
        extP = ElfAddStrToArray(extP,0,name);
        extP = ElfAddNumToArray(extP,1,symP->type);
        if ( symP->moduleName ) {
            symP = GetSymP(symP->moduleName);
            extP = ElfAddStrToArray(extP,2,((data_comp_string*)symP)->text);
        }
        extsP = WriteArray(extsP,ElfbArraySize(extsP),extP);
    }

/* Store literals */

    for ( cdex = 0; cdex < elcodeStr.SzCode; cdex++ ) {
        if ( !elcodeStr.EaRefs[cdex] || (symP = elcodeStr.EaRefs[cdex])->loaded )
            continue;
        symP->loaded = TRUE;
        switch ( symP->type ) {
        case TYPE_NUMBER:
            symP->locate = ix = ElfbArraySize(litsP);
            litsP = ElfAddDblToArray(litsP,ix,((data_comp_number*)symP)->value);
            break;
        case TYPE_STRING:
            symP->locate = ix = ElfbArraySize(litsP);
            litsP = ElfAddStrToArray(litsP,ix,((data_comp_string*)symP)->text);
            break;
        case TYPE_INTEGER:
            symP->locate = ix = ElfbArraySize(litsP);
            litsP = ElfAddNumToArray(litsP,ix,((data_comp_integer*)symP)->value);
            break;
        }
    }

/*  Resolve all effective addrs */

    for ( cdex = 0; cdex < elcodeStr.SzCode; cdex++ ) {
        if ( !elcodeStr.EaRefs[cdex] ) continue;
        symP = elcodeStr.EaRefs[cdex];
        switch ( symP->type ) {

        case TYPE_NUMBER:
        case TYPE_INTEGER:
        case TYPE_STRING:
            elcodeStr.CodeFrame[cdex] = symP->locate;

        CASE TYPE_LABEL:        /* make PC based */
            span = symP->locate - cdex - 1;
            ELFC_VERIFY((span < 32767) && (span > -32768),XLT("Branch distance too far"));
            elcodeStr.CodeFrame[cdex] = span;

        CASE TYPE_LOCVAR:
            elcodeStr.CodeFrame[cdex] = symP->locate;
        CASE TYPE_ARGVAR:
            elcodeStr.CodeFrame[cdex-1]++;  /* set opcode mode for argvar */
            elcodeStr.CodeFrame[cdex] = symP->locate;
        CASE TYPE_GLOVAR:
            elcodeStr.CodeFrame[cdex] = symP->locate;
        CASE TYPE_EXTVAR:
            elcodeStr.CodeFrame[cdex-1]++;  /* set opcode mode for extref */
            elcodeStr.CodeFrame[cdex] = symP->locate;
        CASE TYPE_EXTFUNC:
            elcodeStr.CodeFrame[cdex-1]++;  /* set opcode mode & fall thru */
            elcodeStr.CodeFrame[cdex] = symP->locate;

        CASE TYPE_QFUNC:        /* make PC based, quad word offset */
        case TYPE_UIFUNC:
        case TYPE_NRFUNC:
        case TYPE_MACRO:
        case TYPE_GETOBJ:
        case TYPE_SETOBJ:
            assert((symP->locate & 0x03) == 0);
            elcodeStr.CodeFrame[cdex] = symP->locate >> 2;  /* quad word alignment */
            break;
        }
    }
    codeP = MkBinary(elcodeStr.SzCode*sizeof(shrt));
    if ( ElfcPortableBytes ) {  /* By default, machine independent! */
        for ( ix = 0; ix < elcodeStr.SzCode; ix++ ) {
            BX(codeP)->bytes[ix*2] = (elcodeStr.CodeFrame[ix] >> 8) & 0xFF;
            BX(codeP)->bytes[ix*2+1] = elcodeStr.CodeFrame[ix] & 0xFF;
        }
    }
    else    /* non-portable only for files to be assembled/compiled */
        memcpy(BX(codeP)->bytes,(char*)elcodeStr.CodeFrame,elcodeStr.SzCode*2);

    /* variable names */

    elcodeStr.ElfcVarnames[elcodeStr.ElfcVarnamePos++] = 0; /* empty last str */
    varnamesP = MkBinary(elcodeStr.ElfcVarnamePos);
    memcpy(BX(varnamesP)->bytes,elcodeStr.ElfcVarnames,elcodeStr.ElfcVarnamePos);

    /* formats */

    formatsP = FormatsData();
    
/* assemble pieces */

    allP = WriteArray(allP,0,headerP);
    allP = WriteArray(allP,1,gvarsP);
    allP = WriteArray(allP,2,gfuncsP);
    allP = WriteArray(allP,3,extsP);
    allP = WriteArray(allP,4,litsP);
    allP = WriteArray(allP,5,codeP);
    allP = WriteArray(allP,6,varnamesP);
    allP = WriteArray(allP,7,formatsP);
    allP = WriteArray(allP,8,ovarsP);
    allP = WriteArray(allP,9,svarsP);
#if FOR_TESTER
    if ( !ElfcPortableBytes )   /* By default, machine independent! */
        ElfcAsmGen(allP);
#endif
    return(allP);
}

#if FOR_TESTER
/*************************************************************************
function: XRname - Produce gothic/english names
*/
static char *XRname (instr)
char *instr;
{
    static char result[200];
    int gothic;
    char *ElfSymbolNameFromId();

    gothic = ElfLookupSymbol(instr,0);
    if ( gothic != 0 ) {
        strcpy(result,ElfSymbolNameFromId(gothic,1));
        strcat(result," | ");
        strcat(result,instr);
    }
    else
        strcpy(result,instr);
    return(result);
}
#endif

ElfcLookupSymbol(name,bestLanguage) /* TEMP till ALL english! */
char *name;
int bestLanguage;
{
    int sym;

    sym = ElfLookupSymbol(name,bestLanguage);
	if ( bestLanguage==0 && sym ) {/* old APPLIX files only! */
		if ( sym != ElfLookupSymbol(name,1) )
			CodeError(NXLT("GOTHIC '%s' must be ENGLISH '%s'"),name,ElfSymbolNameFromId(sym,1));
	}
    if ( !sym ) {
        sym = ElfLookupSymbol(name,0);
        if ( sym )
            CodeError("Obsolete function spelling '%s'",name);
    }
    return(sym);
}

static encodeFormatDim(buf,dim,name)
char *buf,*name;
int dim;
{
    *buf = 0;
    while ( --dim >= 0 )
        *buf++ = '[';
    strcpy(buf,name);
}

ReduceExpression(expr)
register ndata expr;
{
    register data_comp_number *numP;
    register data_comp_integer *intP;
    register data_id *symP;
    register node_ptr np;
    register ndata op, lhs, rhs;
    register ndata id;
    register int val;

    if ( !IS_NODE(expr) )
        return(expr);
    
    np = NODEPTR(expr);

/* Check for NXLT... reduce in compiler */

    switch ( NodeType(expr) ) {
    case ND_FUNCEXPR:
        symP = GetSymP(np.funcexpr->id);
        if ( !streq(symP->name,NXLT("NXLT")) )
            return(expr);
        CODE_CHECK(np.funcexpr->args,(XLT("NXLT requires args")));
        CODE_CHECK(SizeList(np.funcexpr->args)==1,(XLT("NXLT requires exactly 1 arg")));
        return(GetListMember(np.funcexpr->args,0));
    case ND_OP:
        break;
    default:
        return(expr);
    }

/* Reduce operators */

    op = np.op->op;
    lhs = np.op->lhs;
    rhs = np.op->rhs;

    switch ( op ) {     /* !!2 should do multiply & divide... */

    case OP_ADD:
    case OP_IADD:
        rhs = ReduceExpression(rhs);
        lhs = ReduceExpression(lhs);
        if ( !IS_SYM(rhs) || !IS_SYM(lhs)
            || (GetSymP(rhs)->type != TYPE_INTEGER)
            || (GetSymP(lhs)->type != TYPE_INTEGER) )
            return(expr);
        val = ((data_comp_integer *)GetSymP(lhs))->value + 
                ((data_comp_integer *)GetSymP(rhs))->value;
#if 0
printf("aval %d\n",val);
#endif
        id = AddIntConst(val);
        return(id);

    case OP_SUB:
    case OP_ISUB:
        rhs = ReduceExpression(rhs);
        lhs = ReduceExpression(lhs);
        if ( !IS_SYM(rhs) || !IS_SYM(lhs)
            || (GetSymP(rhs)->type != TYPE_INTEGER)
            || (GetSymP(lhs)->type != TYPE_INTEGER) )
            return(expr);
        val = ((data_comp_integer *)GetSymP(lhs))->value - 
                ((data_comp_integer *)GetSymP(rhs))->value;
#if 0
printf("sval %d\n",val);
#endif
        id = AddIntConst(val);
        return(id);

    case OP_UMINUS:
    case OP_IUMINUS:
        rhs = ReduceExpression(rhs);
        if ( !IS_SYM(rhs) )
            return(expr);
        symP = GetSymP(rhs);
        if ( symP->type == TYPE_INTEGER ) {
            intP = (data_comp_integer *)symP;
            id = AddIntConst(-(intP->value));
            return(id);
        }
        else if ( symP->type == TYPE_NUMBER ) {
            numP = (data_comp_number *)symP;
            id = AddNumConst(-(numP->value));
#if 0
printf("uval %d\n",-(numP->value));
#endif
            return(id);
        }
        else
            return(expr);
        break;

    case OP_NOT:
        rhs = ReduceExpression(rhs);
        if ( !IS_SYM(rhs) )
            return(expr);
        symP = GetSymP(rhs);
        if ( symP->type == TYPE_INTEGER ) {
            intP = (data_comp_integer *)symP;
            id = AddIntConst(~(intP->value));
            return(id);
        }
        else
            return(expr);
        break;

    default:
        return(expr);
    }

cleanup:
    return(expr);
}
