/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: eldraw.c
   Applix ASC version: 41
   Date: Tue Feb  2 15:21:34 1999
*/
#include <stdio.h>
#include <math.h>
#include <basicwin.h>
#include <ax_xm.h>
#include <aos.h>
#include <dwt2xm.h>
#include <xmlib.h>
#include <xmpixel.h>
#include <wchars.h>
#include <thimble.h>
#include <ctype.h>
#include <elf_env.h>
#include <getstr4.h>
#include <xmcanvas.h>
#include <eldraw.h>
#include <ax_inset.h>

#define DRAW_NO_FILL	0
#define DRAW_FILL		1
#define DRAW_CURVE		2

extern GC ElfGC, ForeGC, DisabledGC;
extern Display *Dpy;
extern int AxCharAbove, AxCharHeight, ScreenRes;
extern unsigned long ForePixel, BackPixel;
extern XFontStruct *AxDlgMainFont;

typedef struct { /* modelled on GC */
    int function;       /* logical operation */
    unsigned long plane_mask;/* plane mask */
    unsigned long foreground;/* foreground pixel */
    unsigned long background;/* background pixel */
    int line_width;     /* line width */
    int line_style;     /* LineSolid, LineOnOffDash, LineDoubleDash */
    int cap_style;      /* CapNotLast, CapButt, CapRound, CapProjecting */
    int join_style;     /* JoinMiter, JoinRound, JoinBevel */
    int fill_style;     /* FillSolid, FillTiled, FillStippled,
						   FillOpaeueStippled */
    int fill_rule;      /* EvenOddRule, WindingRule */
    int arc_mode;       /* ArcChord, ArcPieSlice */
    Pixmap tile;        /* tile pixmap for tiling operations */
    Pixmap stipple;     /* stipple 1 plane pixmap for stipping */
    int ts_x_origin;    /* offset for tile or stipple operations */
    int ts_y_origin;
	Font font;
    XFontStruct *font_struct; /* default text font for text operations */
    int subwindow_mode;     /* ClipByChildren, IncludeInferiors */
    int graphics_exposures;/* boolean, should exposures be generated */
    int clip_x_origin;  /* origin for clipping */
    int clip_y_origin;
    Pixmap clip_mask;   /* bitmap clipping; other calls for rects */
    int dash_offset;    /* patterned/dashed line information */
	int n_dashes;
    char *dash_list;
} PenValues;

#include <eldraw.p>

static PenValues elf_pen;

/*************************************************************************
function: ElfbDraw - Dispatcher for drawing to canvas widgets
*/
void
ElfbDraw (ctrlP, cmdCodeP, pen, stuff, scrollposP)
data_ptr ctrlP, cmdCodeP, pen, stuff, scrollposP;
{
    AXWidget_canvas w;
    int cmdCode;
    canvas_info *cP;
	int val, wid, hyt, xoff, yoff;

	data_ptr objP, infP;

    cmdCode = IntFromDataPtr(cmdCodeP);
    elfPen(pen);

	/* command codes which dont require canvas widget ptr */
	val = 0;
	switch (cmdCode) {
	case ELFDRAW_GET_TEXT_WIDTH_:
		val = elfMeasureText(stuff);
		ElfRetData = MkNumData((double)val);
		return;

	case ELFDRAW_GET_FONT_HEIGHT_:
		val = (elf_pen.font_struct) ? (elf_pen.font_struct->ascent+
						   elf_pen.font_struct->descent) : AxCharHeight;
		ElfRetData = MkNumData((double)val);
		return;

	case ELFDRAW_GET_FONT_BASELINE_:
		val = (elf_pen.font_struct) ? elf_pen.font_struct->ascent : AxCharAbove;
		ElfRetData = MkNumData((double)val);
		return;

	case ELFDRAW_GET_ICON_SIZE_:
		elfMeasureIcon(stuff,&wid,&hyt);
		ElfRetData = MkArray(2);
		ElfAddNumToArray(ElfRetData,0,wid);
		ElfAddNumToArray(ElfRetData,1,hyt);
		return;

	case ELFDRAW_PRINT_:
		ElfbPrint(stuff) ;
		return ;
	}

	/* get the canvas widget */
    w = (AXWidget_canvas)WidgetPointerFromHandleData(ctrlP);
    if ( !w )			 /* spurious ref to non-widgets ok */
        return;

    if (w->type != W_CANVAS)
        return;
	
	if (!w->window) {
		PaintCanvas (w, NULL, TRUE, TRUE);  /* this will make the window */
	}

    cP = (canvas_info *)w->info.canvas;
	getScrollOffset(cP, scrollposP, &xoff, &yoff);

    switch ( cmdCode ) {

    case ELFDRAW_LINE_:
        elfDrawLine(w->window, stuff, xoff, yoff);
        break;

    case ELFDRAW_RECT_:
		elfDrawRectangle(w->window, stuff, DRAW_NO_FILL, xoff, yoff);
        break;

	case ELFDRAW_FILL_RECT_:
		elfDrawRectangle(w->window, stuff, DRAW_FILL, xoff, yoff);
		break;
		
	case ELFDRAW_POLY_:
		elfDrawPolygon(w->window, stuff, DRAW_NO_FILL, xoff, yoff);
		break;

	case ELFDRAW_FILL_POLY_:
		elfDrawPolygon(w->window, stuff, DRAW_FILL, xoff, yoff);
		break;
		
/*
	case ELFDRAW_CURVE_:
		elfDrawPolygon(w->window, stuff, DRAW_CURVE|DRAW_NO_FILL, xoff, yoff);
		break;

	case ELFDRAW_FILL_CURVE_:
		elfDrawPolygon(w->window, stuff, DRAW_CURVE|DRAW_FILL, xoff, yoff);
		break;
*/
		
	case ELFDRAW_ARC_:
		elfDrawArc(w->window, stuff, DRAW_NO_FILL, xoff, yoff);
		break;
		
	case ELFDRAW_FILL_ARC_:
		elfDrawArc(w->window, stuff, DRAW_FILL, xoff, yoff);
		break;
		
    case ELFDRAW_CLEAR_:
        elfClearArea(w->window,stuff, xoff, yoff);
        break;

    case ELFDRAW_TEXT_:
        elfDrawText(w->window,stuff, w->disabled, xoff, yoff);
        break;

    case ELFDRAW_IMAGE_:
        elfDrawImage(w->window,stuff, xoff, yoff);
        break;

    case ELFDRAW_PAINT_WIDTH_:
        cP->paintWidth = IntFromDataPtr(stuff);
        CanvasUpdateScrollbars(w,FALSE);
        break;

    case ELFDRAW_PAINT_HEIGHT_:
        cP->paintHeight = IntFromDataPtr(stuff);
        CanvasUpdateScrollbars(w,FALSE);
        break;

    case ELFDRAW_SCROLL_POS_:
        if (ElfbIsArray(stuff)) {
            cP->leftPoint = IntFromArray(stuff,0);
            cP->topPoint = IntFromArray(stuff,1);
            CanvasUpdateScrollbars(w,FALSE);
        }
        break;

	case ELFDRAW_SCROLL_INCR_:
		if (ElfbIsArray(stuff)) {
			cP->incX = IntFromArray(stuff,0);
			cP->incY = IntFromArray(stuff,1);
		}
		break;

	case ELFDRAW_SCROLL_PAGE_INCR_:
		if (ElfbIsArray(stuff)) {
			cP->page_incX = IntFromArray(stuff,0);
			cP->page_incY = IntFromArray(stuff,1);
		}
		break;
	case ELFDRAW_SCROLL_GRID_:
		if (ElfbIsArray(stuff)) {
			cP->use_grid = TRUE;
			cP->hgrid_size = IntFromArray(stuff,0);
			cP->vgrid_size = IntFromArray(stuff,1);

			if ( cP->hgrid_size == 0 )
				cP->hgrid_size = cP->incX;
			if ( cP->vgrid_size == 0 )
				cP->vgrid_size = cP->incY;
		}
		break;
	case ELFDRAW_SUPPRESS_COPY_:
		cP->supress_copy = IntFromArray(stuff,0);
		break;

	case ELFDRAW_INSET_:
		objP = AxArrayFromArray(stuff, 0);
		infP = AxArrayFromArray(stuff, 1);
		infP = applyOffsetToInset(infP, xoff, yoff);
		ElfbPaintInsetObject(ctrlP, objP, infP, NULL, NULL, NULL, NULL);
		break;
		
    default:
        return;     /* just ignore nonsense request */

	}
}

static void
convDotsToPix(units, dotx, doty, pixx, pixy)
ax_units *units;     /* in */
int dotx, doty;
int *pixx, *pixy;    /* out */
{
    *pixx = (int) floor(((((double) dotx * ScreenRes) / units->dpi)) + 0.5);
    *pixy = (int) floor(((((double) doty * ScreenRes) / units->dpi)) + 0.5);
}

static void
convPixToDots(units, pixx, pixy, dotx, doty)
ax_units *units;
int pixx, pixy;
int *dotx, *doty;    /* out */
{
    int round_up = ScreenRes - 1;
    if (pixx < 0)
        round_up = -round_up;
    *dotx = (int)(((long) (pixx * units->dpi) + round_up) / ScreenRes);
    round_up = ScreenRes - 1;
    if (pixy < 0)
        round_up = -round_up;
    *doty = (int)(((long) (pixy * units->dpi) + round_up) / ScreenRes);
	return;
}

static data_ptr
applyOffsetToInset(insetP, xoff, yoff)
data_ptr insetP;    /* ASSUMED to be in format ax_inset_info@ */
int xoff, yoff;     /* in canvas device pixels (screenpix) */
{
	int ixpos, iypos;
	ax_units units;
	
    data_ptr unitsP = AxArrayElement(insetP, 0);
	data_ptr destP = AxArrayElement(insetP, 1);
	if (destP) {
		memset((void *) &units, 0, sizeof(ax_units));
		if (unitsP) {
			AxExtractUnits(0, unitsP, &units);
		}
		else {
			units.dpi = 1000;   /* use mils as default dot & unit size */
			units.dpu = 1;
			units.precision = 0;
		}
		ixpos = AxIntFromArray(destP, 0);
		iypos = AxIntFromArray(destP, 1);
		ixpos = AxUnitsToDots(&units, (double) ixpos);
		iypos = AxUnitsToDots(&units, (double) iypos);

		/* convert dots to screenpix, add offset, put back
		 */
		convDotsToPix(&units, ixpos, iypos, &ixpos, &iypos);
		ixpos += xoff;
		iypos += yoff;
		convPixToDots(&units, ixpos, iypos, &ixpos, &iypos);
		ixpos = (int) AxDotsToUnits(&units, ixpos);
		iypos = (int) AxDotsToUnits(&units, iypos);
		if (units.name) {
			TaskFree(0, units.name);
		}
		destP = AxAddIntToArray(destP, 0, ixpos);
		destP = AxAddIntToArray(destP, 1, iypos);
		return (AxAddArrayToArray(insetP, 1, AxCopyData(destP)));
	}
	return (insetP);
}

/***************************************************************************
function: initElfPen - set default values on Elf Pen
*/
static void
initElfPen (valuesP)
XGCValues *valuesP;
{
	memset(valuesP, 0, sizeof(XGCValues));
	
	valuesP->function = XVAL_GXcopy;
	valuesP->foreground = BlackPixel(Dpy, DefaultScreen(Dpy));
	valuesP->font = AxDlgMainFont->fid;
	valuesP->line_width = 1;
	valuesP->plane_mask = AllPlanes;
	valuesP->tile = None;
	valuesP->stipple = None;
	valuesP->clip_mask = None;
}


/***************************************************************************
function: changeElfPen - change values of ElfGC from previuos state
*/
static void
changeElfPen (valuesP)
XGCValues *valuesP;
{
	PenValues *penP;
	XFontStruct *fs;
	unsigned long valuemask = 0;
	
	/* look for fields that changed from previous state
	 * by comparing with static "penP", and change those
	 * values on GC, and update static GC values struct
	 */
	penP = &elf_pen;
	if (penP->function != valuesP->function) {
		valuemask |= GCFunction;
		penP->function = valuesP->function;
	}
	if (penP->plane_mask != valuesP->plane_mask) {
		valuemask |= GCPlaneMask;
		penP->plane_mask = valuesP->plane_mask;
	}
	if (penP->foreground != valuesP->foreground) {
		valuemask |= GCForeground;
		penP->foreground = valuesP->foreground;
	}
	if (penP->background != valuesP->background) {
		valuemask |= GCBackground;
		penP->background = valuesP->background;
	}
	if (penP->line_width != valuesP->line_width) {
		valuemask |= GCLineWidth;
		penP->line_width = valuesP->line_width;
	}
	if (penP->line_style != valuesP->line_style) {
		valuemask |= GCLineStyle;
		penP->line_style = valuesP->line_style;
	}
	if (penP->cap_style != valuesP->cap_style) {
		valuemask |= GCCapStyle;
		penP->cap_style = valuesP->cap_style;
	}
	if (penP->join_style != valuesP->join_style) {
		valuemask |= GCJoinStyle;
		penP->join_style = valuesP->join_style;
	}
	if (penP->fill_style != valuesP->fill_style) {
		valuemask |= GCFillStyle;
		penP->fill_style = valuesP->fill_style;
	}
	if (penP->fill_rule != valuesP->fill_rule) {
		valuemask |= GCFillRule;
		penP->fill_rule = valuesP->fill_rule;
	}
	if (penP->tile != valuesP->tile) {
		valuemask |= GCTile;
		penP->tile = valuesP->tile;
	}
	if (penP->stipple != valuesP->stipple) {
		valuemask |= GCStipple;
		penP->stipple = valuesP->stipple;
	}
	if (penP->ts_x_origin != valuesP->ts_x_origin) {
		valuemask |= GCTileStipXOrigin;
		penP->ts_x_origin = valuesP->ts_x_origin;
	}
	if (penP->ts_y_origin != valuesP->ts_y_origin) {
		valuemask |= GCTileStipYOrigin;
		penP->ts_y_origin = valuesP->ts_y_origin;
	}
	if (penP->font != valuesP->font) {
		valuemask |= GCFont;
		penP->font = valuesP->font;
	}
	if (penP->subwindow_mode != valuesP->subwindow_mode) {
		valuemask |= GCSubwindowMode;
		penP->subwindow_mode = valuesP->subwindow_mode;
	}
	if (penP->graphics_exposures != valuesP->graphics_exposures) {
		valuemask |= GCGraphicsExposures;
		penP->graphics_exposures = valuesP->graphics_exposures;
	}
	if (penP->clip_x_origin != valuesP->clip_x_origin) {
		valuemask |= GCClipXOrigin;
		penP->clip_x_origin = valuesP->clip_x_origin;
	}
	if (penP->clip_y_origin != valuesP->clip_y_origin) {
		valuemask |= GCClipYOrigin;
		penP->clip_y_origin = valuesP->clip_y_origin;
	}
	if (penP->clip_mask != valuesP->clip_mask) {
		valuemask |= GCClipMask;
		penP->clip_mask = valuesP->clip_mask;
	}
	if (penP->arc_mode != valuesP->arc_mode) {
		valuemask |= GCArcMode;
		penP->arc_mode = valuesP->arc_mode;
	}
	if (elf_pen.font_struct) {
		fs = elf_pen.font_struct;
	}
	else {
		fs = AxDlgMainFont;
	}
 	axChangeGC(Dpy, ElfGC, fs, valuemask, valuesP);
	if (penP->dash_list && penP->n_dashes) {
		XSetDashes(Dpy, ElfGC, valuesP->dash_offset,
				   penP->dash_list, penP->n_dashes);
	}
}

/*************************************************************************
function: elfPen - Update GC from pen format
*/
static
elfPen (pen)
data_ptr pen;
{
    XGCValues values;
	int color_type;
    unsigned long pixel;
    data_ptr xP;
    pixcache_t *p;
    char *name;
	int ix;

    initElfPen(&values);

    if (!pen)
		goto change_gc;
    
    /* logical drawing function */
    if (xP = AnyFromArray(pen, XGC_function_))
        values.function = IntFromDataPtr(xP);

    /* plane mask */
    if (xP = AnyFromArray(pen, XGC_plane_mask_))
        values.plane_mask = IntFromDataPtr(xP);

    /* foreground color */
    if (xP = AnyFromArray(pen,XGC_foreground_)) {
        if (ElfbIsArray(xP)) {
			color_type = IntFromArray(xP,0);
			if (color_type == 5) {
				AxPixelFromColorname(StrFromArray(xP,1), &pixel);
			}
			else {
				AxGetRealPixel(color_type, IntFromArray(xP,1), IntFromArray(xP,2),
							   IntFromArray(xP,3), IntFromArray(xP,4), 255, &pixel);
			}
            values.foreground = pixel;
        }
    }

    /* background color */
    if (xP = AnyFromArray(pen,XGC_background_)) {
        if (ElfbIsArray(xP)) {
            color_type = IntFromArray(xP,0);
			if (color_type == 5) {
				AxPixelFromColorname(StrFromArray(xP,1), &pixel);
			}
			else {
				AxGetRealPixel(IntFromArray(xP,0), IntFromArray(xP,1), IntFromArray(xP,2),
							   IntFromArray(xP,3), IntFromArray(xP,4), 255, &pixel);
			}
            values.background = pixel;
        }
    }

    /* line width in pixels */
    if (xP = AnyFromArray(pen,XGC_line_width_))
        values.line_width = IntFromDataPtr(xP);

    /* line style */
    if (xP = AnyFromArray(pen,XGC_line_style_))
        values.line_style = IntFromDataPtr(xP);

    /* cap style */
    if (xP = AnyFromArray(pen,XGC_cap_style_))
        values.cap_style = IntFromDataPtr(xP);

    /* line join style */
    if (xP = AnyFromArray(pen,XGC_join_style_))
        values.join_style = IntFromDataPtr(xP);

    /* fill style */
    if (xP = AnyFromArray(pen,XGC_fill_style_))
        values.fill_style = IntFromDataPtr(xP);

    /* fill rule */
    if (xP = AnyFromArray(pen,XGC_fill_rule_))
        values.fill_rule = IntFromDataPtr(xP);
        
    if (xP = AnyFromArray(pen,XGC_arc_mode_))
        values.arc_mode = IntFromDataPtr(xP);

    /* tile pixmap */
    name = StrFromArray(pen,XGC_tile_);
    if (name && *name) {
        if (p = Pixmapper(name))
            values.tile = p->pix;
    }

    /* stipple pixmap */
    name = StrFromArray(pen,XGC_stipple_);
    if (name && *name) {
        if (p = Pixmapper(name))
            values.stipple = p->pix;
    }

    if (xP = AnyFromArray(pen,XGC_ts_x_origin_))
        values.ts_x_origin = IntFromDataPtr(xP);

    if (xP = AnyFromArray(pen,XGC_ts_y_origin_))
        values.ts_y_origin = IntFromDataPtr(xP);

    /* font for text operations */
    if (xP = AnyFromArray(pen,XGC_font_)) {
        if (elf_pen.font_struct = lookupNamedFont(xP))
            values.font = elf_pen.font_struct->fid;
    }
	else {
		elf_pen.font_struct = (XFontStruct *)NULL;
	}

    if (xP = AnyFromArray(pen,XGC_subwindow_mode_))
        values.subwindow_mode = IntFromDataPtr(xP);

    if (xP = AnyFromArray(pen,XGC_graphics_exposures_))
        values.graphics_exposures = IntFromDataPtr(xP);

    if (xP = AnyFromArray(pen,XGC_clip_x_origin_))
        values.clip_x_origin = IntFromDataPtr(xP);

    if (xP = AnyFromArray(pen,XGC_clip_y_origin_))
        values.clip_y_origin = IntFromDataPtr(xP);

    name = StrFromArray(pen,XGC_clip_mask_);
    if (name && *name) {
        if (p = Pixmapper(name))
            values.clip_mask = p->pix;
    }

    if (xP = AnyFromArray(pen,XGC_dash_offset_))
        values.dash_offset = IntFromDataPtr(xP);

	elf_pen.n_dashes = 0;
	elf_pen.dash_list = (char *)TrueRealloc(elf_pen.dash_list, 2);
	elf_pen.dash_list[0] = 0;

    if (xP = AnyFromArray(pen,XGC_dashes_)) {
		if (IS_INTEGER(xP))
			elf_pen.dash_list[0] = IntFromDataPtr(xP);
		else if (ElfbIsArray(xP)) {
			elf_pen.n_dashes = ElfbArraySize(xP);
			elf_pen.dash_list = (char *)TrueRealloc(elf_pen.dash_list, elf_pen.n_dashes+1);
			for (ix = 0; ix < elf_pen.n_dashes; ++ix)
				elf_pen.dash_list[ix] = IntFromArray(xP,ix);
			elf_pen.dash_list[ix] = 0;
		}
	}

 change_gc:
	changeElfPen(&values);
}

static pixcache_t *
hackForWhitePixmap(name)
char *name;
{
    return(Pixmapper(name));
}


/*************************************************************************
function: elfMeasureIcon - Return icon size
*/
static
elfMeasureIcon (stuff, widthP, heightP)
data_ptr stuff;
int *widthP, *heightP;
{
    char *name;
    pixcache_t *p;

    name = StrFromDataPtr(stuff);
    p = Pixmapper(name);
    if ( !p ) {
        *widthP = 0;
        *heightP = 0;
    }
    else {
        *widthP = p->width;
        *heightP = p->height;
    }
}

/*************************************************************************
function: elfMeasureText - Return text width
*/
static int
elfMeasureText (stuff)
data_ptr stuff;
{
    char *str;
    int result;
    
    str = StrFromDataPtr(stuff);
    result = ((elf_pen.font_struct) ? (AXTextWidth(elf_pen.font_struct, str, strlen(str)))
                                        : BasicStrlen(str));
    return(result);
}

/* : addInScrollOffset - return an offset taken as the difference between
 * the canvas widget's current scrollpos and the calling-elftask's notion
 * of current scrollpos (which could be a bit old)
 */
static void
getScrollOffset(canv, scrollposP, xp, yp)
canvas_info *canv;
data_ptr scrollposP;
int *xp, *yp;              /* out: return the offset vector */
{
	int scroll_x = canv->leftPoint;
	int scroll_y = canv->topPoint;
	if (scrollposP && AxIsArray(scrollposP)) {
		scroll_x = AxIntFromArray(scrollposP, 0);
		scroll_y = AxIntFromArray(scrollposP, 1);
	}
	*xp = scroll_x - canv->leftPoint;
	*yp = scroll_y - canv->topPoint;
	return;
}

/*************************************************************************
function: elfDrawLine - Draw a line
The draw is EXCLUSIVE of bottom/right point.
*/
static
elfDrawLine (win, coords, xoff, yoff)
Window win;
data_ptr coords;
int xoff, yoff;
{
    data_ptr pt1, pt2;
    int x1, y1, x2, y2;
	
    pt1 = AnyFromArray(coords,0);
    pt2 = AnyFromArray(coords,1);
    x1 = IntFromArray(pt1,0) + xoff;
    y1 = IntFromArray(pt1,1) + yoff;
    x2 = IntFromArray(pt2,0) + xoff;
    y2 = IntFromArray(pt2,1) + yoff;

    XDrawLine(Dpy,win,ElfGC,x1,y1,x2,y2);
}

/*************************************************************************
function: elfClearArea - Draw a line
The draw is EXCLUSIVE of bottom/right point.
*/
static
elfClearArea (win, coords, xoff, yoff)
Window win;
data_ptr coords;
int xoff, yoff;
{
    data_ptr pt1, pt2;
    int x1, y1, wd, ht;

    x1 = IntFromArray(coords,0) + xoff;
    y1 = IntFromArray(coords,1) + yoff;
    wd = IntFromArray(coords,2);
    ht = IntFromArray(coords,3);

    XClearArea(Dpy,win,x1,y1,wd,ht,FALSE);
}

/***************************************************************************
function: elfDrawRectangle - draw rectangle
*/
static
elfDrawRectangle (win, coords, fill, xoff, yoff)
Window win;
data_ptr coords;
int fill, xoff, yoff;
{
    int x1, y1, wd, ht;
	
    x1 = IntFromArray(coords,0) + xoff;
    y1 = IntFromArray(coords,1) + yoff;
    wd = IntFromArray(coords,2);
    ht = IntFromArray(coords,3);

	if (fill & DRAW_FILL)
		XFillRectangle(Dpy, win, ElfGC, x1,y1, wd, ht);
	else 
		XDrawRectangle(Dpy, win, ElfGC, x1,y1, wd, ht);
}


/***************************************************************************
function: elfDrawPolygon - draw polyline or curve
*/
static int
elfDrawPolygon (win, stuff, flags, xoff, yoff)
Window win;
data_ptr stuff;
int flags;
int xoff, yoff;
{
	int ix, npoints, mode;
	data_ptr pt, pts;
	XPoint *points;

	pts = AnyFromArray(stuff,0);
	npoints = IntFromArray(stuff, 1);
	mode = IntFromArray(stuff, 2);

	points = (XPoint *)malloc(npoints * sizeof(XPoint));

	for (ix = 0; ix < npoints; ix++) {
		pt = AnyFromArray(pts, ix);
		points[ix].x = IntFromArray(pt,0) + xoff;
		points[ix].y = IntFromArray(pt,1) + yoff;
	}

	if (flags & DRAW_FILL)
		XFillPolygon(Dpy, win, ElfGC, points, npoints, Complex, CoordModeOrigin);
	else
		XDrawLines(Dpy, win, ElfGC, points, npoints, CoordModeOrigin);

	free(points);
}


/***************************************************************************
function: elfDrawArc - draw arc
*/
static int
elfDrawArc (win, coords, fill, xoff, yoff)
Window win;
data_ptr coords;
int fill;
int xoff, yoff;
{
    int x1, y1, wd, ht, ang1, ang2;

    x1 = IntFromArray(coords,0) + xoff;
    y1 = IntFromArray(coords,1) + yoff;
    wd = IntFromArray(coords,2);
    ht = IntFromArray(coords,3);
	ang1 = IntFromArray(coords,4);
	ang2 = IntFromArray(coords,5);

	if (fill & DRAW_FILL)
		XFillArc(Dpy, win, ElfGC, x1,y1, wd, ht, ang1*16, ang2*16);
	else
		XDrawArc(Dpy, win, ElfGC, x1,y1, wd, ht, ang1*16, ang2*16);
}

/*************************************************************************
function: elfDrawText - Put text down
Recieves multi-byte text that is assumed to represent 16bit charset
*/
static
elfDrawText (win, stuff, disabled, xoff, yoff)
Window win;
data_ptr stuff;
int disabled;
int xoff, yoff;
{
    int x,y;
    char *str;
	XFontStruct *fs;
    data_ptr pt1;
		
    pt1 = AnyFromArray(stuff,0);
    x = IntFromArray(pt1,0) + xoff;
    y = IntFromArray(pt1,1) +
		((elf_pen.font_struct) ? elf_pen.font_struct->ascent : AxCharAbove) +
			yoff;
    str = StrFromArray(stuff,1);

	fs = elf_pen.font_struct ? elf_pen.font_struct : AxDlgMainFont;

	if ( disabled == FALSE )
		axDrawString(Dpy, win, ElfGC, fs, x, y, str, strlen(str));
	else
	{
		/* set the font info of the gc */
		axSetFont(Dpy, DisabledGC, fs);
		axDrawString(Dpy, win, DisabledGC, (XFontStruct *) 0,
						  x, y, str, strlen(str));

		/* restore the gc */
		axSetFont(Dpy, DisabledGC, AxDlgMainFont);
	}
}

/*************************************************************************
function: elfDrawImage - Draw a ".im" file
*/
static void
elfDrawImage (win, stuff, xoff, yoff)
Window win;
data_ptr stuff;
int xoff, yoff;
{
    int x,y;
    char *name;
    data_ptr pt1;
    pixcache_t *p;
    XImage *ximg;
    unsigned long fgMapColor, bgMapColor, bgPixInImage;
    extern unsigned long ForePixel, BackPixel;

    pt1 = AnyFromArray(stuff,0);
    x = IntFromArray(pt1,0) + xoff;
    y = IntFromArray(pt1,1) + yoff;
    name = StrFromArray(stuff,1);
    if ( !name || !*name )
        return;
    p = Pixmapper(name);
    if ( !p )
        return;
	if (p->clip_mask) {
        XSetClipOrigin(Dpy, ElfGC, x, y);
        XSetClipMask(Dpy, ElfGC, p->clip_mask);
	}

    XCopyArea(Dpy,p->pix,win,ElfGC,0,0,p->width,p->height,x,y);
	
	if (p->clip_mask)
		XSetClipMask(Dpy, ElfGC, None);
}
