/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elipc.c
   Applix ASC version: 31
   Date: Tue Feb  2 15:21:07 1999
*/
#include <math.h>
#include <elf_env.h>
#include <paripc.h>
#include <elf_strs.h>
#include <elipc.h>
#include <thimble.h>
#include <scac.h>
#include <elhelper.h>
#include <eldata.h>

void ElfbSendMesWaiter AXP_((int pid));
extern elf_object *ElfObjPtr AXP_((data_ptr));

/*************************************************************************
function: ElfbSendMesToApp - Send an IPC message to a process
If the command code is positive, the current field read is first terminated
via the LK_ELFMSG function key.  If the code is negative, the code is
re-inverted and the function key is not sent and the message is sent in the
form of an ELFM msg.
10-26-89: Response format passed is ignored, this now uses the actual
			response message.
*/
void ElfbSendMesToApp(pidP,cmdP,formatP,aP)
data_ptr pidP;
data_ptr cmdP;
data_ptr formatP;                  /* control string */
register data_array *aP;    /* array to decompose */
{
	int pid, cmd;
	char *msg;
	int size;

	pid = ThimPidFromUID(IntFromDataPtr(pidP));
	cmd = IntFromDataPtr(cmdP);
    VERIFY(pid > 0,ERR_BADCHANNEL);
	if ( cmd < 0 )		/* historical */
		cmd = -cmd;

	if ( ElfbIsArray(formatP) ) {
		size = AxMemWrite(formatP,sizeof(int),&msg);		/* pack into a single buffer */
		memcpy(msg,&cmd,sizeof(int));
		AxSendMes(pid,msg,size,APELFM);
		THIMfree(msg);
	}
	else {
		ElfbArraySize(aP);	/* just to verify is an array (or null datum) */
		size = AxMemWrite(aP,sizeof(int),&msg);		/* pack into a single buffer */
		memcpy(msg,&cmd,sizeof(int));
		AxSendMes(pid,msg,size,APELFM);
		THIMfree(msg);
	}
	ElfbSendMesWaiter(pid);
}

/*************************************************************************
function: ElfbSendMesWaiter - Wait for a response
*/
void ElfbSendMesWaiter (pid)
int pid;
{
	int err_code;
	char err_str[1001], err_obj[1001];
	int responseCode;
	char *response_buf;
	int response_size;
	register data_ptr rP;	/* response array */

/* Read response into an elf array */

	if ( TaskInfo[AxInTask].appid == APELFM )
		ACImOk();	/* prevents communications deadlocks */
	response_size = IPCresponseSize(pid);
	if ( response_size == 0 )
		exit(0);	/* task must have gone away */
	response_buf = ElfMalloc(response_size);
	if ( IPC_OK != AxGetResponse(pid,response_buf,response_size) )
		exit(0);	/* task must have gone away */

	memcpy(&responseCode,response_buf,sizeof(int));
	rP = AxMemRead(response_buf + sizeof(responseCode));
	ElfFree(response_buf);

	if ( responseCode == 0 ) {	/* reject */
		err_code = AxIntFromArray(rP,0);
		strncpy(err_str,AxStrPtrFromArray(rP,1),1000);
		strncpy(err_obj,AxStrPtrFromArray(rP,2),1000);
		err_str[1000] = err_obj[1000] = 0;
		AxFreeData(rP);
		ElfStrAbort(err_code,err_str,err_obj);
	}
    ElfRetData = rP;    /* generic return value */
}


/*************************************************************************
function: SendElfStyleMessage - Similar to ElfbSendMes but for C<->C messages
If the command code is positive, the current field read is first terminated
via the LK_ELFMSG function key.  If the code is negative, the code is
re-inverted and the function key is not sent and the message is sent in the
form of an ELFM msg.
10-26-89: Response format passed is ignored, this now uses the actual
			response message.
*/
void SendElfStyleMessage (va_alist)
va_dcl
{
	va_list ptr;
    int size, dex, pid, cmd, window_id;
	register char *format;                  /* control string */
	char *msg;
    register char *p;
	data_ptr aP;

    va_start(ptr);

	window_id = va_arg(ptr,int);		/* not used */
    pid = va_arg(ptr,int);
    cmd = va_arg(ptr,int);
	if ( cmd < 0 )		/* historical */
		cmd = -cmd;
	format = va_arg(ptr,char*);

	aP = AxMakeArray(strlen(format));
    for ( p = format,dex = 0; *p; p++,dex++ ) {
        switch ( *p ) {
        case 'I':
			AxAddIntToArray(aP,dex,*(va_arg(ptr,int*)));
            break;
        case 'N':
            AxAddDblToArray(aP,dex,*(va_arg(ptr,double*)));
            break;
        case 'S':
			AxAddStrToArray(aP,dex,*(va_arg(ptr,char**)));
			break;
        case 'C':
			AxAddStrToArray(aP,dex,va_arg(ptr,char*));
			break;
        default:
			tempAbort("bad ipc format");
		}
    }
	size = AxMemWrite(aP,sizeof(int),&msg);		/* pack into a single buffer */
	memcpy(msg,&cmd,sizeof(int));
	AxSendMes(pid,msg,size,APELFM);
	AxFreeData(aP);
	THIMfree(msg);
}

#if AXMS   /* {  */
/* check to see if an elf message being sent is a duplicate of one in */
/* the queue.  If it is a duplicate, the new one will replace the old  */
/* At this point, we are only interested in dialog-resize message posted */
/* from RebuildGDIWindow, where "duplicate" is determined by looking at the */
/* firt 50 bytes                                                         */

extern int UniqueOnly;

static int check_elf_message( char *old_mess, char *new_mess )
{
     int i;
     for ( i=0; i<50; i++ )
        {
        if ( *old_mess != *new_mess )
           return( FALSE );
        old_mess++;
        new_mess++;
        }
     return( TRUE );
}
#endif   /*  }  AXMS  */

/*************************************************************************
function: SendElfDataToTask - Can be called by C code
See the file eldata.c for how to make an elf arary from c code.
*/
void SendElfData (pid,cmd,aP)
int pid;
int cmd;
data_ptr aP;		/* Must be an Elf Array */
{
	char *msg;
	int size;

	size = AxMemWrite(aP,sizeof(int),&msg);		/* pack into a single buffer */
	memcpy(msg,&cmd,sizeof(int));
#if AXMS
    if ( UniqueOnly )
        AxSendUniqueMes( pid,msg,size,APELFASYNC,check_elf_message );
    else
#endif
	AxSendMes(pid,msg,size,APELFASYNC);
	THIMfree(msg);
}

/*************************************************************************
function: ElfbSendElfData - Send a message to someone
*/
void ElfbSendElfData(pidP,cmdP,aP)
data_ptr pidP;			/* uid of task to send message to */
data_ptr cmdP;				/* command code */
register data_ptr aP;    /* array to decompose */
{
	int pid;
	elf_object *oP;
	register data_ptr dataP;    /* array to decompose */

	dataP = aP;
	if (ElfbIsObject(pidP)) {
		pid = (oP = ElfObjPtr(pidP)) ? oP->task_uid : 0;

		/* repackage data for object */
		dataP = ELF_CREATE_ARRAY(3);
		dataP = WriteArray(dataP,0,ELF_DATA_COPY(aP));
		dataP = WriteArray(dataP,1,ELF_CREATE_STRING(NXLT("_async_")));
	}
	else
		pid = IntFromDataPtr(pidP);
	pid = ThimPidFromUID(pid);
    VERIFY(pid > 0,ERR_BADCHANNEL);

	SendElfData(pid,IntFromDataPtr(cmdP),dataP);

	if (dataP != aP)
		ELF_DATA_FREE(dataP);
}

/*************************************************************************
function: ElfbWaitForMessage - Wait for event
Returns array of 2 elements.  The first element is the command code,
the second is the data.
Returns NULL datum on timeout.
*/
ElfbWaitForMessage (timeout)
int timeout;	/* secs or -1 for infinite */
{
	int code;
	char *msgP;
	long size, senderType, senderTask, ljunk;
	int cmd;
	data_ptr aP;

	ACImOk();
	if ( (AxInTask != THIMpid) || (TaskInfo[AxInTask].appid != APELFM) )
		ElfStrAbort(ERR_ILLVAL,
			XLT("Macro called in this manner cannot pend for input."),NULL);
	if ( TaskInfo[AxInTask].showing_hourglass )
		AXmRestoreNormalCursorPointer(TaskInfo[AxInTask].uid);
	code = AxGetMessage(timeout,&msgP,&size,&senderType,&senderTask,&ljunk);
	if ( IPC_TO == code ) {
		ElfRetData = NULL;
		return;
	}
	memcpy(&cmd,msgP,sizeof(int));

	ElfRetData = aP = MkArray(2);
	AxAddDblToArray(aP,0,(double)cmd);
	AxAddArrayToArray(aP,1,AxMemRead(msgP+sizeof(int)));
	AxMessageAbsorbed();
}

/*************************************************************************
function: ElfbSendAndAckElfData - Send message to another Elf task, return response
*/
ElfbSendAndAckElfData (taskIdP,cmdCodeP,aP)
data_ptr taskIdP, cmdCodeP;
data_ptr aP;
{
	int taskId, cmdCode;
	char *msg;
	int size;
	elf_object *oP;
	register data_ptr dataP;

/* get args */

	dataP = aP;
	if (ElfbIsObject(taskIdP)) {
		taskId = (oP = ElfObjPtr(taskIdP)) ? oP->task_uid : 0;

		/* repackage data for object */
		dataP = ELF_CREATE_ARRAY(3);
		dataP = WriteArray(dataP,0,ELF_DATA_COPY(aP));
		dataP = WriteArray(dataP,1,ELF_CREATE_STRING(NXLT("_sync_")));
		dataP = WriteArray(dataP,2,MK_INTEGER(TaskInfo[AxInTask].uid));
	}
	else
		taskId = IntFromDataPtr(taskIdP);
	taskId = ThimPidFromUID(taskId);
	VERIFY(taskId > 0,ERR_BADCHANNEL);

	cmdCode = IntFromDataPtr(cmdCodeP);

/* send message */

	size = AxMemWrite(dataP,sizeof(int),&msg);	/* pack into a single buffer */
	memcpy(msg,&cmdCode,sizeof(int));
	AxSendMes(taskId,msg,size,APELFASYNC);		/* !!0 new cmd code needed? */
	THIMfree(msg);

	if (dataP != aP)
		ELF_DATA_FREE(dataP);

/* get response */

	ElfbSendMesWaiter(taskId);
}

/*************************************************************************
function: ElfbRespondElf - Respond to a ElfSendAndAckElfData
*/
ElfbRespondElf (taskIdP,data)
data_ptr taskIdP;
data_ptr data;
{
	int size;
	char *msg;
	int code1 = 1;
	int pid;
	elf_object *oP;

	if (ElfbIsObject(taskIdP))
		pid = (oP = ElfObjPtr(taskIdP)) ? oP->task_uid : 0;
	else
		pid = IntFromDataPtr(taskIdP);
	pid = ThimPidFromUID(pid);
	VERIFY(pid > 0,ERR_BADCHANNEL);

	size = AxMemWrite(data,sizeof(int),&msg);	/* pack into a single buffer */
	memcpy(msg,&code1,sizeof(int));
	IPCsendMes(IPC_RES,pid,msg,size);
	THIMfree(msg);
}

/*************************************************************************
function: ElfbRespondElfErr - Throw an error back to ElfSendAndAckElfData
*/
ElfbRespondElfErr (taskIdP,code,errstr,obj)
data_ptr taskIdP;
data_ptr code, errstr, obj;
{
	int size;
	char *msg;
	int code0 = 0;
	int pid;
	data_ptr data;
	elf_object *oP;

	data = MkArray(3);
	WriteArray(data,0,ElfDataCopy(code));
	WriteArray(data,1,ElfDataCopy(errstr));
	WriteArray(data,2,ElfDataCopy(obj));

	if (ElfbIsObject(taskIdP))
		pid = (oP = ElfObjPtr(taskIdP)) ? oP->task_uid : 0;
	else
		pid = IntFromDataPtr(taskIdP);
	pid = ThimPidFromUID(pid);
	VERIFY(pid > 0,ERR_BADCHANNEL);

	size = AxMemWrite(data,sizeof(int),&msg);	/* pack into a single buffer */
	memcpy(msg,&code0,sizeof(int));
	IPCsendMes(IPC_RES,pid,msg,size);
	ElfDataFree(data);
	THIMfree(msg);
}

/*************************************************************************
function: AxRejectElfMessage - Send a reject to calling Elf code.
Like ElfbRespondElfErr (taskIdP,code,errstr,obj).
*/
AxRejectElfMessage (pid)
int pid;
{
	int size;
	char *msg;
	int code0 = 0;
	data_ptr data;

	if ( pid <= 0 )
		return;
	data = AxMakeArray(3);
	AxAddIntToArray(data,0,ERR_NOTHELPER);
	size = AxMemWrite(data,sizeof(int),&msg);		/* pack into a single buffer */
	memcpy(msg,&code0,sizeof(int));
	IPCsendMes(IPC_RES,pid,msg,size);
	AxFreeData(data);
	TaskFree(THIMpid,msg);
}

