/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elmdd.c
   Applix ASC version: 11
   Date: Tue Feb  2 15:22:22 1999
*/
#include <string.h>
#include <elf_env.h>
#include <eldata.h>
#include <datetime.h>
#include <arghdr.h>
#include <elf_strs.h>

static data_ptr astore(),bsearcher();
static struct astore_info {
	int ndim;
	data_ptr subs;
	data_ptr val;
	double numval;
	int mode; /* 0 to store val, 1 to add numval */
} astoreInfo;

#define TASK0 0

/*************************************************************************
function: AxfMDDFuncs - Do calculations for multi-dimensional databases
*/
data_ptr AxfMDDFuncs(argDatum)
data_ptr argDatum;
{
	register data_ptr values;
	register int ix;

	switch (AxIntFromArray(argDatum,0)) {
	/* val = fetch_element(values,subs) of array */
	case 1: {
		int endx;
		data_ptr subs;
		values = AxArrayFromArray(AxArrayFromArray(argDatum,1),0);
		subs = AxArrayFromArray(argDatum,2);
		endx = AxArraySize(subs); /* number of dimensions */
		for (ix=0; ix<endx; ix++) {
			values = AxArrayElement(values, AxIntFromArray(subs,ix));
		}
		return(AxCopyData(values));
	} 
	
	/* values = store_element(values,subs,value) into array.
	The values array has as its first element the actual
	values array. The indirection is so that we can modify
	the values array in place, without copying. Also,
	the values array is assumed to be allocated in task 0.
	Objvars are allocated in task 0.
	*/
	case 2: {
		astoreInfo.mode = 0; /* store val */
		astoreInfo.subs = AxArrayFromArray(argDatum,2); /* subscripts */
		astoreInfo.ndim = AxArraySize(astoreInfo.subs)-1;
													/* max dimension index*/
		astoreInfo.val = AxArrayElement(argDatum,3); /* value to store */
		values = AxArrayElement(argDatum,1);
		AX(values)->vals[0] = astore(AxArrayElement(values,0),0);
		return NULL;
	}
	case 3: { /* search_sorted_array(array,value)
				array elements are strings or numbers.
				If the datatype of value is the same as the elements
				of the array, performance will be enhanced.
				returns a two element array:
				arr[0] is true iff value is found
				arr[1] is location of/for value in sorted array 
				*/
		return
			bsearcher(AxArrayFromArray(argDatum,1),AxArrayElement(argDatum,2));
	}
	case 4: { /* add_element(values,subs,value) 
				where summary function is sum. */

		/* add the value into the values array at the point indicated
		by the subscripts array */
		values = AxArrayElement(argDatum,1);
		astoreInfo.subs = AxArrayElement(argDatum,2);
		astoreInfo.numval = AxFloatFromArray(argDatum,3);
		astoreInfo.mode = 1; /* add val */
		astoreInfo.ndim = AxArraySize(astoreInfo.subs)-1;
		AX(values)->vals[0] = astore(AxArrayElement(values,0),0);
		return NULL;
	}
	case 5: { /* rollup_element(values,coordsArray,value) 
				where summary function is sum. */
		data_ptr coordsArray;

		/* add the value into the values array at all the points indicated
		by the array of subscript arrays */
		values = AxArrayElement(argDatum,1);
		coordsArray = AxArrayElement(argDatum,2);
		astoreInfo.numval = AxFloatFromArray(argDatum,3);
		astoreInfo.mode = 1; /* add val */
		astoreInfo.ndim = AxArraySize(AxArrayFromArray(coordsArray,0))-1;
		for (ix = 0; ix<AxArraySize(coordsArray);ix ++) {
			astoreInfo.subs = AxArrayFromArray(coordsArray,ix);
			AX(values)->vals[0] = astore(AxArrayElement(values,0),0);
		}
		return NULL;
	}
	case 6: { /* allocate_values(values,maxcoord)
				returns an array of nulls. number of dimensions
				is array_size@(maxcoord).
				size of dimension n is maxcoord(n)+1.
				values array is fully populated with nulls. */
		values = AxArrayElement(argDatum,1);
		astoreInfo.mode = 2; /* allocate storage */
		astoreInfo.subs = AxArrayFromArray(argDatum,2); /* max coords */
		astoreInfo.ndim = AxArraySize(astoreInfo.subs)-1;
													/* max dimension index*/
		astoreInfo.val = NULL;
		AX(values)->vals[0] = astore(AxArrayElement(values,0),0);
		return NULL;
	}
	}
}
/* store astore_val into arr, where astore_subs is an array of subscripts. */
/* arr is assumed to be an objvar. Therefore, memory is allocated in task 0.
Care is taken to modify arrays in place. */
static data_ptr astore(arr,ix)
register data_ptr arr;
register int ix;
{
	int ixs;
	data_ptr AxTaskCreateElfArray(), TaskWriteArray(), AxTaskCreateElfNumber();

	ixs = AxIntFromArray(astoreInfo.subs,ix); /* ix'th subscript */

	/* make or extend array */
	if (!AxIsArray(arr))
		arr = AxTaskCreateElfArray(TASK0,ixs+1);
	/* resize now */
	else if (AxArraySize(arr) < ixs+1)
		arr = TaskWriteArray(TASK0, arr, ixs, (data_ptr)0);

	/* store sub-array or value */
	if (ix == astoreInfo.ndim) {
		if (astoreInfo.mode == 1) /* add value */
			AX(arr)->vals[ixs] =
					AxTaskCreateElfNumber(TASK0,
							AxFloatFromArray(arr,ixs) + astoreInfo.numval);
		else /* store value */
			AX(arr)->vals[ixs] = AxTaskCopyElfData(TASK0,astoreInfo.val);
	}
	else {
		/* store array element ixs */
		AX(arr)->vals[ixs] = astore(AxArrayElement(arr,ixs),ix+1);
	}

	if (astoreInfo.mode == 2) { /* populate full arrays */
		int ix;
		data_ptr val;

		val = AX(arr)->vals[ixs];
		for (ix=0; ix<ixs; ix++) {
			AX(arr)->vals[ix] = AxTaskCopyElfData(TASK0,val);
		}
	}

	return(arr);
}

static data_ptr bsearcher(arr,val)
data_ptr arr;
data_ptr val;
{
	register int compval,ix;
	int ihi, ilow, found;
	int location;
	int is_string;
	mbchar *strval;
	double numval,numaval;
	data_ptr aval;
	data_ptr retval;

	found = FALSE;
	ihi = AxArraySize(arr)-1;

	/* handle null val (it comes first, always) */
	if (val == 0) {
		ix = 0;
		if (ihi >= 0)
			found = (AxArrayElement(arr,0) == 0);
	} else {
		if (is_string = AxIsString(val)) {
			strval = AxStrPtrFromDataPtr(val);
		}
		else {
			numval = AxFloatFromDataPtr(val);
		}

		ix = ilow = 0;
		while (ilow <= ihi) {
			ix = ilow + ((ihi - ilow) / 2);	/* mid point */
			aval = AxArrayElement(arr,ix);
			if (aval == 0) {
				compval = 1; /* anything is bigger than null */
			}
			else {
				if (is_string) {
					compval = strcmp(strval,AxStrPtrFromDataPtr(aval));
				}
				else {
					numaval = AxFloatFromDataPtr(aval);
					if (numval == numaval)
						compval = 0;
					else if (numval > numaval)
						compval = 1;
					else
						compval = -1;
				}

			}
			if (compval == 0) {
				found = TRUE;
				break;
			}
			else if (compval < 0) {
				ihi = ix -1; /* search below */
			}
			else {
				ilow = ix + 1; /* search above */
				ix = ilow;
			}
		}
	}

	retval = AxMakeArray(2);
	retval = AxAddBoolToArray(retval,0,found);
	retval = AxAddIntToArray(retval,1,ix);
	return retval;
}

/*************************************************************************
function: AxfChangeDateFormat - convert a datetime string to a given format.
Precise to the second, unlike the Applix datetime values.
Returns a datetime string or an AxTimeStruct array of values. AxTimeStruct,
defined in datetime.h, corresponds to format datetime_ defined in datetim_.am.
*/
data_ptr AxfChangeDateFormat(argDatum)
data_ptr argDatum;
{
	register data_ptr dP;
    char           *datestr;
	char           *newfmt=NULL;
    short           precision=0;
    AxTimeStruct 	axt;
    char            string[240];
	int				formatNumber;
 
	/* return datetime now */
	if (AxArraySize(argDatum)==0) {
		if (AOK !=
		  ParseDateToAxTimeStruct(&axt,PDCURRBOTH,NULL,NULL,&precision,string))
 		ElfStrAbort(ERR_ILLVAL,XLT("datetime error"),datestr);

	}
	else {
		/* first arg is either a datetime string to be parsed,
		or an AxTimeStruct */
		dP = AxArrayElement(argDatum,0);
		if (AxIsArray(dP)) {
			/* first arg is AxTimeStruct */
			axt.year = AxIntFromArray(dP,0);
			axt.month = AxIntFromArray(dP,1);
			axt.day = AxIntFromArray(dP,2);
			axt.hour = AxIntFromArray(dP,3);
			axt.minute = AxIntFromArray(dP,4);
			axt.second = AxIntFromArray(dP,5);
			axt.fraction = AxDblFromArray(dP,6);
		}
		else {
			/* parse first arg as datetime string, into an AxTimeStruct */
			datestr  = AxStrPtrFromDataPtr(dP);
			if (AOK !=
			  ParseDateToAxTimeStruct(&axt,
									PDNORMAL,datestr,NULL,&precision,string))
			ElfStrAbort(ERR_ILLVAL,XLT("Invalid datetime"),datestr);
		}
	}

	/* less than two args, return now */
	if (AxArraySize(argDatum) < 2) {
		dP = AxMakeArray(7);
		dP = AxAddIntToArray(dP,0,axt.year);
		dP = AxAddIntToArray(dP,1,axt.month);
		dP = AxAddIntToArray(dP,2,axt.day);
		dP = AxAddIntToArray(dP,3,axt.hour);
		dP = AxAddIntToArray(dP,4,axt.minute);
		dP = AxAddIntToArray(dP,5,axt.second);
		dP = AxAddFloatToArray(dP,6,axt.fraction);
		return(dP);
	}

	/* if there is a second arg, it is a format, given either as a format
	string, like "yyyy-mm-dd", or a format number defined by the
	datetime.sp file. If there is no second arg, simply return the parsed
	result as an AxTimeStruct. */
	dP = AxArrayElement(argDatum,1);
	if (AxIsString(dP)) { /* string format given */
		newfmt = AxStrPtrFromDataPtr(dP);
		if (AOK != Formatdt_str_AxTime(axt, newfmt, string))
			ElfStrAbort(ERR_ILLVAL,XLT("Invalid datetime format"),newfmt);
	}
	else {
		formatNumber = AxIntFromDataPtr(dP); /* numeric format given */
		if (AOK != Formatdt_AxTime(axt, formatNumber, string))
			ElfStrAbort(ERR_ILLVAL,XLT("Invalid datetime format"),newfmt);
	}

	return(ELF_CREATE_STRING(string));
}

/*************************************************************************
function: AxfTokenize - implements STRING_TOKENS@(string,separators,escapeChar)
given a string, one or more separators, and optionally,
an escape character. get all the substrings from the string, Returning an
array of strings. the escape character is used to embed a separator, or
the escape character itself, in a token.
For example STRING_TOKENS@("this i\ s a test   with a \\", " ", "\")
returns
	{"this", "i s", "a", "test", "with", "a", "\"}
*/
data_ptr AxfTokenize(argDatum)
data_ptr argDatum;
{
	data_ptr retData;
	mbchar *string = (mbchar *)StrFromArgArray(argDatum,0);
	mbchar *separators = (mbchar *)StrFromArgArray(argDatum,1);
	mbchar *escapeChar = (mbchar *)StrFromArgArray(argDatum,2);
	mbchar *token, *cP, *tP;
	mbchar escaper;
	int sepLen, tokcnt;

	retData = ELF_CREATE_ARRAY(0);
	tokcnt = 0;
	if (string == 0 || strlen(string) == 0
			|| separators == 0 || strlen(separators) == 0)
		return(retData);

	token = AxMalloc(strlen(string)+1);
	if (escapeChar)
		escaper = *escapeChar;
	else
		escaper = 0;
	cP = string;
	tP = token;
	while (*cP) {

		if (escaper && *cP == *escapeChar) {
			cP++;
		}
		else {
			sepLen = strspn(cP,separators);
			if (sepLen > 0) {
				*tP = 0;
				retData = ElfAddStrToArray(retData,tokcnt++,token);
				tP = token;
				cP = cP + sepLen;
				continue;
			}
			else {
			}
		}
		*tP++ = *cP++;
	}
	if (tP > token) { /* final token */
		*tP = 0;
		retData = ElfAddStrToArray(retData,tokcnt++,token);
	}
	AxFree(token);
	return(retData);
}
