/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elnodes.h
   Applix ASC version: 19
   Date: Tue Feb  2 15:21:26 1999
*/
#define MAX_NODES 32000

#define NODEPTR(n) (NodeArray[(n)-NODE_FUDGE])

/* decode node values */

#define IS_TOK(v) ( ((v) < SYM_FUDGE )		/* TK_... */
#define IS_NODE(v) ( (v) >= NODE_FUDGE )	/* Created by CrtNode */
#define IS_SYM(v) ( ((v) < NODE_FUDGE) && ((v) >= SYM_FUDGE) )	/* by lexer */

#define ND_STMTLIST     1   /* statement pair */
#define ND_IDLIST       2   /* identifier pair */
#define ND_NUMLIST      3   /* expression pair */
#define ND_ASSIGN       4   /* assignment statement */
#define ND_FUNC         5   /* function/command starter */
#define ND_WHILE        6   /* while statement */
#define ND_FOR          7   /* for statement */
#define ND_GOTO         8   /* goto statement */
#define ND_RETURN       9   /* return statement */
#define ND_IF           10  /* if statement */
#define ND_ONGOTO       11  /* computed goto statement */
#define ND_ONCALL       12  /* computed call statement */
#define ND_LABELED      13  /* labelled statement */
#define ND_NULLST       14  /* null statement */
#define ND_VARS         15  /* variable decl statement */
#define ND_FUNCEXPR     16  /* function call expression */
#define ND_AREF         17  /* array reference expression */
#define ND_OP           18  /* arith/relate/bool operator expression */
#define ND_NEXT         19  /* the nexter for a for loop */
#define ND_WEND         20  /* the nexter for a while loop */
#define ND_DIMLIST      21  /* variable dimensioning */
#define ND_ERASE        22  /* erase statement */
#define ND_EXTERN       23  /* dcl for extern vars */
#define ND_END          24  /* funtion/command ender */
#define ND_RESVD99      25  /* unused */
#define ND_RESVD98      26  /* unused */
#define ND_ONERROR      27  /* catch an error */
#define ND_ERRERR       28  /* on error return error error_number */
#define ND_SELECT       29  /* select statement */
#define ND_CASE         30  /* case within select */
#define ND_DEFAULT      31  /* default case within select */
#define ND_ENDSELECT    32  /* end of select stmt */
#define ND_ADDTO        33  /* add a value to a var */
#define ND_SUBTO        34  /* subtract a value from a var */
#define ND_ENTRYFUNC	35	/* call a function via entry var */
#define ND_ENTRY		36	/* push an entry var */
#define ND_FORMAT		37	/* define a format */
#define ND_REFLIST		38	/* list of multi-format array references */
#define ND_ARRLIT		39	/* literal array ref */
#define ND_FID			40  /* formatted identifier */
#define ND_COMPILE		41  /* compiler directive */
#define ND_SQW_IN		42	/* SQL Select's where clause */
#define ND_COLLIST		43	/* SQL column list */
#define ND_BINDER		44	/* SQL bindable name */
#define ND_USERTAB		45	/* SQL username.tablename */
#define ND_COLSTAR		46	/* SQL column.* */
#define ND_TABLIST		47	/* SQL tables list */
#define ND_TABEXPR		48	/* SQL from.where.using expression */
#define ND_SQLOR 49
#define ND_SQLAND 50
#define ND_SQLNOT 51
#define ND_SQLPAREN 52
#define ND_DCLCURSOR 53
#define ND_OPENCURSOR 54
#define ND_FETCHCURSOR 55
#define ND_CLOSECURSOR 56
#define ND_FROMTAB 57
#define ND_WHERE 58
#define ND_GROUPBY 59
#define ND_ORDERBY 60
#define ND_UNION 61
#define ND_INTERSECT 62
#define ND_MINUS 63
#define ND_OREF 64
#define ND_OREFLIST 65
#define ND_SQLPREPARE 66
#define ND_UPDATE 67
#define ND_COLSET 68
#define ND_COLSETLIST 69
#define ND_EXECUTE 70
#define ND_SWITCH 71
#define ND_ENDSWITCH 72
#define ND_CONNECT 73
#define ND_SQLUNPREPARE 74
#define ND_SQLFUNC 75
#define ND_BREAKFOR 76
#define ND_BREAKWHILE 77
#define ND_CONTINFOR 78
#define ND_CONTINWHILE 79

#define ndata shrt          /* node data element */


typedef struct {        /* std template */
    ndata type;
    ndata linenumber;
    ndata data[6];
} any_node;

typedef struct {        /* format definition */
    ndata type;
    ndata linenumber;
	ndata fmt;
    ndata id;
	ndata fmtDim;
	ndata isObject;
} NT_FID;

typedef struct {        /* function/command starter */
    ndata type;
    ndata linenumber;
    ndata ftype;
    ndata id;
    ndata args;
	ndata functype;
} NT_FUNC;

typedef struct {        /* statement pair */
    ndata type;
    ndata linenumber;
    ndata list;
    ndata stmt;
} NT_STMTLIST;

typedef struct {        /* identifier pair */
    ndata type;
    ndata linenumber;
    ndata list;
    ndata id;
} NT_IDLIST;

typedef struct {        /* expression pair */
    ndata type;
    ndata linenumber;
    ndata list;
    ndata expr;
} NT_NUMLIST;

typedef struct {        /* assignment statement */
    ndata type;
    ndata linenumber;
    ndata id;
    ndata ref;
    ndata expr;
} NT_ASSIGN;

typedef struct {        /* while statement */
    ndata type;
    ndata linenumber;
    ndata expr;
} NT_WHILE;

typedef struct {        /* for statement */
    ndata type;
    ndata linenumber;
    ndata id;
    ndata expr;
    ndata stop;
    ndata step;
} NT_FOR;

typedef struct {        /* the nexter for a for */
    ndata type;
    ndata linenumber;
    ndata ids;
} NT_NEXT;

typedef struct {        /* goto statement */
    ndata type;
    ndata linenumber;
    ndata id;
} NT_GOTO;

typedef struct {        /* return statement */
    ndata type;
    ndata linenumber;
    ndata expr;
} NT_RETURN;

typedef struct {        /* if statement */
    ndata type;
    ndata linenumber;
    ndata expr;
    ndata st_then;
    ndata st_else;
} NT_IF;

typedef struct {        /* computed goto statement */
    ndata type;
    ndata linenumber;
    ndata expr;
    ndata labels;
} NT_ONGOTO;

typedef struct {        /* computed call statement */
    ndata type;
    ndata linenumber;
    ndata expr;
    ndata calls;
} NT_ONCALL;

typedef struct {        /* labelled statement */
    ndata type;
    ndata linenumber;
    ndata label;
    ndata stmt;
} NT_LABELED;


typedef struct {        /* variable dimensioning */
    ndata type;
    ndata linenumber;
    ndata list;
    ndata id;
    ndata ref;
} NT_DIMLIST;

typedef struct {        /* function call expression */
    ndata type;
    ndata linenumber;
    ndata id;
    ndata args;
	ndata object_def;
} NT_FUNCEXPR;

typedef struct {        /* array reference expression */
    ndata type;
    ndata linenumber;
    ndata id;
    ndata ref;
	ndata objArgs;	/* if is really an object reference, parameters list */
} NT_AREF;

typedef struct {        /* arith/relate/bool operator expression */
    ndata type;
    ndata linenumber;
    ndata op;
    ndata lhs;
    ndata rhs;
} NT_OP;

typedef union {
    any_node *any;
    NT_FUNCEXPR *funcexpr;
    NT_STMTLIST *stmtlist;
    NT_NUMLIST *numlist;
    NT_ASSIGN *assign;
    NT_FOR *fr;
    NT_IF *ife;
    NT_ONGOTO *cgoto;
    NT_ONCALL *ccall;
    NT_OP *op;
    NT_AREF *aref;
    NT_LABELED *label;
    NT_FUNC *func;
	NT_FID *fid;
    long num;       /* for free listing */
} node_ptr;


long ExtNumber();       /* extract value from number node */
