/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elproc.c
   Applix ASC version: 25
   Date: Tue Feb  2 15:21:10 1999
*/
#if AXMS
/* need to make sure we include varargs not stdarg on NT,
 * so we fool any Windows defines that may include it with this
 * define.  eldata.h will include varargs.h     -donato
 */
#define _INC_STDARG
#endif

#include <elf_env.h>
#include <eldata.h>
#include <paripc.h>
#include <thimble.h>
#include <scac.h>
#include <getstr4.h>

#ifndef bool
#define bool int
#endif

 

int ElfCallFromC AXP_((char *name,data_ptr *retP,data_ptr args));

/*************************************************************************
function: ElfCallByName - Call an elf routine by name
The called routine is run as part of the caller's task... Not a new task!
Therefore, the called routine should never post a dialog!
It is the responsibilty of the caller to free any created args and the retval!
On non-AOK returns, retP is loaded wit han error info
array of {str,object,srcline,srcfile}.

Usage:  ElfCallByName( char *name, data_ptr *retP, int argcount, data_ptr arg1... argN )

*/
int ElfCallByName (va_alist)
va_dcl
{
    va_list ptr;
    register char *name;
    register data_ptr *retP;
    register int argcnt;
    register int ix;
    register int retCode;
    data_ptr retData, aP;
    int savedErrorBoxPref = Elf.noErrorBox;

    va_start(ptr);

/* get passed name, argcount, args */

    name = va_arg(ptr,char *);
    retP = va_arg(ptr,data_ptr *);
    argcnt = va_arg(ptr,int);
    aP = AxMakeArray(argcnt);
    for ( ix = 0; ix < argcnt; ix++ )
        AxWriteArray(aP,ix,va_arg(ptr,data_ptr));
    Elf.noErrorBox = TRUE;
    if ( retP )
        retCode = ElfCallFromC(name,retP,aP);
    else {
        retCode = ElfCallFromC(name,&retData,aP);
        AxFreeData(retData);
    }
    AxFreeArrayButNotContents(aP);
    Elf.noErrorBox = savedErrorBoxPref;
    return(retCode);
}

int ElfCallByNameArgArray (name,retP,argcnt,args)
register char *name;
register data_ptr *retP;
register int argcnt;
register data_ptr *args;
{
    register int ix;
    register int retCode;
    data_ptr retData, aP;
    int savedErrorBoxPref = Elf.noErrorBox;

/* It is the responsibilty of the caller to free any created args and the retval! */

    aP = AxMakeArray(argcnt);
    for ( ix = 0; ix < argcnt; ix++ )
        AxWriteArray(aP,ix,args[ix]);
    Elf.noErrorBox = TRUE;
    if ( retP )
        retCode = ElfCallFromC(name,retP,aP);
    else {
        retCode = ElfCallFromC(name,&retData,aP);
        AxFreeData(retData);
    }
    AxFreeArrayButNotContents(aP);
    Elf.noErrorBox = savedErrorBoxPref;
    return(retCode);
}

int ElfCallByNameElfArgArray (name,retP,args)
register char *name;
register data_ptr *retP;
register data_ptr args;
{
    register int argcnt;
    register int ix;
    register int retCode;
    data_ptr retData;
    int savedErrorBoxPref = Elf.noErrorBox;

    Elf.noErrorBox = TRUE;
    if ( retP )
        retCode = ElfCallFromC(name,retP,args);
    else {
        retCode = ElfCallFromC(name,&retData,args);
        AxFreeData(retData);
    }
    Elf.noErrorBox = savedErrorBoxPref;
    return(retCode);
}

/*************************************************************************
function: ElfStartNewTask - Start an independent new task
This must be called for all tasks started by the ATM's main logic.
This includes command_key_struck, initial_icons and low_memory.
Failure to work this way causes an IPCawait on a random THIMpid (really
the one that most recently did and IPC), thus trashing that pid's return
address.
*/
ElfStartNewTask (n)
int n;
{
    char *ElfSymbolNameFromId();

    char *func_name = ElfSymbolNameFromId(n,ElfLanguage);
    VfyContext(0);
    return(ElfbNewTask(func_name,""));
}
ElfStartNewTaskWithArgs (n,args)
int n;
char *args;
{
    char *ElfSymbolNameFromId();

    char *func_name = ElfSymbolNameFromId(n,ElfLanguage);
    VfyContext(0);
    return(ElfbNewTask(func_name,args));
}
TaskStartNewTaskWithArgs (n,args)
int n;
char *args;
{
    char *ElfSymbolNameFromId();

    char *func_name = ElfSymbolNameFromId(n,ElfLanguage);
    return(ElfbNewTask(func_name,args));
}

/*************************************************************************
function: ElfbNewTask - Start another Elf task
*/
ElfbNewTask (func_name,args)
char *func_name;
char *args;
{
    int pid;

    VERIFY(func_name[0] && (func_name[0] != ' '),ERR_ILLVAL);
    ACprocIt(APELFM,args,&pid,FALSE,TRUE,func_name,FALSE);
    if ( pid == -1 )
        return(0);
    return(TaskInfo[pid].uid);
}

/*************************************************************************
function: ElfbNewTaskWithReturn - Start another Elf task, output task retcode

Task is run UNPENDED.

- If task is started successfully, returns the uid of the task.
  Ignore the task return code ('ret') in this case.
- If task ran to completion, returns zero, and 'ret' represents
  the task return code: zero for success, non-zero for error.
*/
ElfbNewTaskWithReturn (func_name,args,ret)
char *func_name;
char *args;
int  *ret;      /* OUT: return code of task, if it completed */
{
    int pid;

    VERIFY(func_name[0] && (func_name[0] != ' '),ERR_ILLVAL);
    *ret = ACprocIt(APELFM,args,&pid,FALSE,TRUE,func_name,FALSE);
    if ( pid == -1 )
        return(0);                      /* Task done - error */
    else if ( pid == 0 )
        return(0);                      /* Task done - OK */
    else
        return(TaskInfo[pid].uid);      /* Task in progress */
}

/*************************************************************************
function: ElfbNewTaskWithData - Run another Elf task, output task retval

Task is run PENDED. Also outputs return code and uid of task

- If task ran to completion (usual outcome), returns task retval if any.
  In this case, 'ret' is return code (zero for success, non-zero for error)
  and 'uid' will be zero.
- If task is still running, returns NULL (shouldn't happen; we're pending)
  In this case, 'ret' should be ignored since the task is not complete,
  and 'uid' is the uid of the task.
*/
data_ptr
ElfbNewTaskWithData (func_name,args,ret,uid)
char *func_name;
char *args;
int  *ret;      /* OUT: return code of task, if it completed */
int  *uid;      /* OUT: uid code of task, if it is still running */
{
    int pid;

    VERIFY(func_name[0] && (func_name[0] != ' '),ERR_ILLVAL);
    *ret = ACprocIt(APELFM,args,&pid,TRUE,TRUE,func_name,FALSE);
    switch( pid) {
        case -1:                    /* Task done - error */
        case 0:                     /* Task done - OK */
            *uid = 0;
            if (TaskInfo[AxInTask].child_retdata)
                return (AxTaskCopyElfData(0, TaskInfo[AxInTask].child_retdata));
            else
                return (NULL);

        default:                    /* Task in progress */
            *uid = TaskInfo[pid].uid;
            return( NULL);
    }
}

/*************************************************************************
function: ElfbPendForNewTask - Start another Elf task and wait for it to complete
Returns child's exit code or -1 if could not start task (very rare).
*/
ElfbPendForNewTask (func_name,args)
char *func_name;
char *args;
{
    int pid, code;

    VERIFY(func_name[0] && (func_name[0] != ' '),ERR_ILLVAL);
    code = ACprocIt(APELFM,args,&pid,TRUE,FALSE,func_name,FALSE);
    return(code);
}


