/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elsock.c
   Applix ASC version: 124
   Date: Fri May 21 21:11:33 1999
*/
#if AUX
#include <X11/fd.h>
#endif

#if LINUX
#include <pwd.h>
#endif

#include <sys/types.h>
#if AXMS
#include <basicwin.h>
#include <aos.h>
#endif
#include <axsocket.h>

#if X386_NEW
#include <stdlib.h>
#endif

#include <math.h>
#include <elf_env.h>
#include <paripc.h>
#include <elf_strs.h>
#include <thimble.h>
#include <scac.h>
#include <elhelper.h>
#include <el_syms.h>
#include <getstr10.h>
#include <errno.h>
#include <fcntl.h>
#if SYSV4
/* for bzero */ 
#include <utils.h>
#endif
#if RIOS
#include <sys/select.h>
#endif
#include <eldata.h>
#include <axnet.h>
#include <arghdr.h>


#define SOCK_LONG 1
#define SOCK_STRING 2
#define SOCK_BINARY 3
#define SOCK_ARRAY 4
#define SOCK_NULL 5
#define SOCK_FLOAT 6        /* !!0 not tested yet */
#define SOCK_OBJECT 7

#define MSG_TYPE_NORMAL 0   /* normal peer messages */
#define MSG_TYPE_ERROR 1    /* error response message */

#ifndef SOCKET_ERROR
#define SOCKET_ERROR -1
#endif

static int WrtBufPos = 0;
static char WrtBuf[BUFSIZ];
static char *IncomingMessage;
static int MsgType = MSG_TYPE_NORMAL;
static int ReadTimer;       /* how long to wait for first byte */

extern int JavaMode;    /* TRUE for Java slaved axmain */
extern int AxSockFd;    /* LISTENER$ socket */
extern char *AxSocketName;
extern data_ptr NullDataPtr;


#include <elsock.p>



/*************************************************************************
function: ElfbServer - Start a server, return when a connection accepted
*/
ElfbServer (sockP,timeoutP,phaseP)
data_ptr sockP;
data_ptr timeoutP;
data_ptr phaseP;
{
    return(ElfServer(StrFromDataPtr(sockP),IntFromDataPtr(timeoutP),
            IntFromDataPtr(phaseP)));
}
static int ElfServer(portstr,timeout,phase)
char *portstr;
int timeout;
int phase;
{
    unsigned short socknum;
    static int sockfd;
    int clilen;
    struct sockaddr_in in_cli_addr, in_serv_addr;
#if !AXMS  
    struct sockaddr_un un_cli_addr, un_serv_addr;
#endif
    int uid, channel, code;
    int isUnix = FALSE;
    int createRealSocket = FALSE;
    char sockname[STRPATH+1];
    int i;
    long senderType, senderTask, size, msgType;
    char *msgP;
    int ret;

    SVERIFY((strlen(portstr) < 108),ERR_STRTOOLONG,portstr);
    strcpy(sockname,portstr);
    FullTrim(sockname,sockname);
    socknum = atoi(sockname);
    if ( !socknum )
       {
       /* we were passed some sort of string */
       if ( sockname[0] == '/' )
          isUnix = TRUE;
       else if ( (sockname[0] == '\\') || (sockname[1] == ':') ||
                 (strncmp(sockname, "remote_service", 14) == 0 ) )
          /* a windows style pathname or specific remote connection has been */
          /* requested.  In either case we create a real socket, as opposed  */
          /* to a UNIX named pipe (which NT does not support and which will  */
          /* not work across machine boundries.  So, we create a socket and  */
          /* bind here to get a port number.  The sockname just becomes a    */
          /* hook to look things up by.                                  */
          createRealSocket = TRUE;
       }

/* See if socket has already been created, if so, just goto accepter */
    uid = AxFindSocketByName(sockname);
    if ( uid )
        goto accepter;

/* Check to see if we have enough resources to create a new socket */

    if ( !AxCanCreateAnotherChannel() )
        ElfStrAbort(ERR_ILLVAL,XLT("Too many sockets"),NULL);
    
/* Create the socket if not already created */

    if ( isUnix ) {
#if AXMS 
        tempAbort(NXLT("UNIX sockets not supported on this platform") );
#else
        unlink(sockname);
        if ( (sockfd = socket(AF_UNIX, SOCK_STREAM, 0)) == INVALID_SOCKET )
            ElfStrAbort(ErrnoErr(),NULL,"socket");
#endif
    }
    else {
        if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET )
            {
#if AXMS
            ElfStrAbort(WSAGetLastError(), NULL,"socket");
#else
            ElfStrAbort(ErrnoErr(),NULL,"socket");
#endif
            }
    }

#if AX_X11
    /* under windows the select structure is an array of file descriptors */
    /* not a bitmask, so comparing a socket-descriptor number against the */
    /* max number of things one can select on is not meaningful          */
    if ( sockfd >= FD_SETSIZE ) {
        CLOSE_CHANNEL(sockfd);
        ElfStrAbort(ERR_ILLVAL,XLT("Too many sockets"),NULL);
    }
#endif

    if ( isUnix ) {
#if AXMS 
        tempAbort(NXLT("UNIX sockets not supported on this platform") );
#else
        bzero(&un_serv_addr, sizeof(un_serv_addr));
        un_serv_addr.sun_family = AF_UNIX;
        strcpy(un_serv_addr.sun_path,sockname);
    
        if ( bind( sockfd, (struct sockaddr*)&un_serv_addr,
                           sizeof(un_serv_addr)) == SOCKET_ERROR ) {
            CLOSE_CHANNEL(sockfd);
            ElfStrAbort(ErrnoErr(),NULL,"bind");
        }
#endif
    }
    else if ( createRealSocket )
        {
        /* do a bind with a default to get a port number */
        memset( &in_serv_addr, 0, sizeof(in_serv_addr) );
        in_serv_addr.sin_family = AF_INET;
        in_serv_addr.sin_addr.s_addr = INADDR_ANY;
        in_serv_addr.sin_port = 0;
        ret = bind( sockfd, (struct sockaddr*)&in_serv_addr,
                                                  sizeof(in_serv_addr));
        if ( ret == SOCKET_ERROR )
            {
#if AX_X11
            ret = ErrnoErr();
#else
            ret = WSAGetLastError();
#endif
            ElfStrAbort(ret,NULL,"bind");
            }
        i = sizeof(in_serv_addr);
        ret = getsockname( sockfd, (struct sockaddr*)&in_serv_addr, &i );
        if ( ret == SOCKET_ERROR )
            {
#if AX_X11
            ret = ErrnoErr();
#else
            ret = WSAGetLastError();
#endif
            ElfStrAbort(ret,NULL,"bind");
            }
         socknum = ntohs(in_serv_addr.sin_port); /* the assigned port number */
         }

    else {
        if ( !socknum ) {
            struct servent *serverInfo;
            serverInfo = getservbyname(sockname,NXLT("tcp"));
            if ( !serverInfo ) {
                CLOSE_CHANNEL(sockfd);
                ElfStrAbort(ERR_ILLVAL,XLT("Not a tcp service in /etc/services"),sockname);
            }
            socknum = ntohs(serverInfo->s_port);
        }
        bzero(&in_serv_addr, sizeof(in_serv_addr));
        in_serv_addr.sin_family = AF_INET;
        in_serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
        in_serv_addr.sin_port = htons(socknum);
    
        if ( bind( sockfd, (struct sockaddr*)&in_serv_addr,
                                      sizeof(in_serv_addr)) == SOCKET_ERROR ) {
            CLOSE_CHANNEL(sockfd);
            ElfStrAbort(ErrnoErr(), (char *)NULL,"bind");
        }
    }
#if AXMS
    {
    extern HWND WXDummyWnd;
    ret = WSAAsyncSelect( sockfd, WXDummyWnd, SOCKET_PSEUDO_EVENT, 
           (long)(FD_READ | FD_WRITE | FD_ACCEPT | FD_CONNECT) );
    axmsStartWindowsTimer();    /* fail safe "wake up" */
    if ( ret != 0 )
        tempAbort( NXLT("WSAAsyncSelect: bad return:") );
    }
#endif

    listen(sockfd,20);
    ACImOk();
    uid = AxNoteSocket(sockfd,sockname,AxInTask,TRUE);

/* Wait for a connection from a client */

    if ( phase == 1 )
        return( socknum );;
    if ( phase == 3 )
        return(uid);

   /*===================================================================*/
   /*This is the processing for phase 2 */
   /*===================================================================*/
accepter:

/* Wait for message */
msg_getter:
    if ( -1 == AxSockReadUp(uid) ) {
        AxLoadDeferedMessages();
        ElfStrAbort(ErrnoErr(),NULL,"accept");
    }
    TaskInfo[AxInTask].waiting_on_socket = TRUE;
    code = AxGetMessage(timeout,&msgP,&size,&senderType,&senderTask,&msgType);
    TaskInfo[AxInTask].waiting_on_socket = FALSE;
    switch ( code ) {
    case IPC_SOCKET:    /* expected response */
/*        AxMessageAbsorbed(); NO MESSAGE, don't discard it! */
        break;
    case IPC_TO:
        AxSockNoReadUp(uid);
        AxLoadDeferedMessages();
        return(-1);
    default:
        AxDeferMessage();
        goto msg_getter;
    }

/* Return client info */

    sockfd = AxSockFdFromUid(uid);
    if ( !sockfd )
        ElfStrAbort(ERR_ILLVAL,XLT("Lost channel"),XLT("server open"));

    if ( isUnix ) {
#if AXMS 
        tempAbort(NXLT("UNIX sockets not supported on this platform") );
#else
        clilen = sizeof(un_cli_addr);
        channel = accept( sockfd, (struct sockaddr *)(&un_cli_addr), &clilen);
        if ( channel == INVALID_SOCKET )
            ElfStrAbort(ErrnoErr(),NULL,"channel");
#endif
    }
    else {
        clilen = sizeof(in_cli_addr);
        channel = accept( sockfd, (struct sockaddr *)(&in_cli_addr), &clilen);
        if ( channel == INVALID_SOCKET )
            {
#if AXMS
            errno = WSAGetLastError();
#endif
            ElfStrAbort(ErrnoErr(),NULL,"channel");
            }
    }

    if ( !AxSockMakeChannelNonBlocking(channel) )
        ElfStrAbort(ErrnoErr(),NULL,"nonblock");

    AxLoadDeferedMessages();

#if AX_X11
    /* under windows the select structure is an array of file descriptors */
    /* not a bitmask, so comparing a socket-descriptor number against the */
    /* max number of things one can select on is not meaningful          */
    if ( channel >= FD_SETSIZE ) {
        CLOSE_CHANNEL(channel);
        ElfStrAbort(ERR_ILLVAL,XLT("Too many sockets"),NULL);
    }
#endif

    channel = AxNoteSocket(channel,NULL,AxInTask,FALSE);
    return(channel);
}

/*************************************************************************
function: AxSockMakeChannelNonBlocking - Read only as much as is there
*/
int AxSockMakeChannelNonBlocking (channel)
int channel;
{
#if AX_X11   /* {  */
#if SUN4 && !SUN4S  /* Posix style nonblock does not seem to work for SunOS */
    return(0 == fcntl(channel,F_SETFL,O_NDELAY));   /* bsd4.2 non-blocking only! */
#else
    return(0 == fcntl(channel, F_SETFL, (fcntl(channel,F_GETFL,0) | O_NONBLOCK)));
#endif
#else     /*  }  {  */
     /* MS-Windows,  NT  */
    /* setting up the use of async messages puts the socket in non-blocking */
    /*   mode.                                                   */
    SetSocketAsyncCallback( channel );

    return( TRUE );
#endif    /*  }  AX_X11 / AXMS  */
}

/*************************************************************************
function: ElfbClient - Connect to a server
*/
ElfbClient (sockP,timeoutP,hostnameP)
data_ptr sockP, timeoutP, hostnameP;
{
    short socknum;
    int timeout;
    char *passed_hostname;
    int sockfd;
    static struct sockaddr_in in_serv_addr;
#if !AXMS 
    static struct sockaddr_un un_serv_addr;
#endif  /* DAMNT */
    struct hostent *info;
    extern int errno;
    char hostname[101];
    int isUnix = FALSE;
    char sockname[STRPATH+1];
    char *p;
    int error_code;
#if AXMS
    FILE *fileP;
    char buff[40];
    int retry_count = 0;
#endif

    timeout = IntFromDataPtr(timeoutP);
    passed_hostname = StrFromDataPtr(hostnameP);

    p = StrFromDataPtr(sockP);
    SVERIFY((strlen(p) < 108),ERR_STRTOOLONG,p);
    strcpy(sockname,p);
    FullTrim(sockname,sockname);

    socknum = atoi(sockname);
#if AXMS
    if ( socknum == 0 )
        {
        /* the passed name is not a number.  We assume that it is a filename */
        /* which in the UNIX environment would mean a UNIX socket.  In the   */
        /* Windows/NT environment, the port number should have been written  */
        /* to this file when the server was started up.  We read the file,   */
        /* get the port number, and then proceed as per TCP sockets          */
        fileP = fopen( sockname, "r" );
        if ( fileP == (FILE *)NULL )
           ElfStrAbort( errno, XLT("Could not open socket file"), NULL);
        if ( fgets( buff, sizeof(buff), fileP ) == (char *)NULL )
           ElfStrAbort( errno, XLT("Could not read socket file"), NULL);
        fclose( fileP );
        unlink( sockname );        /* all done with the file */
        p = buff + strlen(buff) - 1;
        if ( *p <= ' ' )
           *p = 0;
        FullTrim( buff, buff );
        socknum = atoi( buff );
        if ( socknum == 0 )
            ElfStrAbort( errno, XLT("socket number is zero"), NULL);
        }
#endif

    if ( !socknum && ((sockname[0] == '/') || (sockname[0] == '\\')) )
        isUnix = TRUE;

    if ( isUnix ) {
#if AXMS
        tempAbort(NXLT("UNIX sockets not ported on this machine") );
#else
        if ( (sockfd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 )
            ElfStrAbort(ErrnoErr(),NULL,"socket");
#endif
    }
    else {
        if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET )
            {
#if AXMS
            ElfStrAbort( WSAGetLastError(), NULL, "socket");
#else
            ElfStrAbort(ErrnoErr(),NULL,"socket");
#endif
            }
    }

    ACImOk();

#if AX_X11
    /* under windows the select structure is an array of file descriptors */
    /* not a bitmask, so comparing a socket-descriptor number against the */
    /* max number of things one can select on is not meaningful          */
    if ( sockfd >= FD_SETSIZE ) {
        CLOSE_CHANNEL(sockfd);
        ElfStrAbort(ERR_ILLVAL,XLT("Too many sockets"),NULL);
    }
#endif

    if ( isUnix ) {
#if !AXMS   /* !!0 Hey, no intra-machine ipcs? */
        bzero(&un_serv_addr, sizeof(un_serv_addr));
        un_serv_addr.sun_family = AF_UNIX;
        strcpy(un_serv_addr.sun_path,sockname);
        AxSysDosUnixPath(un_serv_addr.sun_path);
        if ( connect( sockfd, (struct sockaddr *)(&un_serv_addr),
                                                 sizeof(un_serv_addr)) < 0 ) {
            CLOSE_CHANNEL(sockfd);
            ElfStrAbort(ErrnoErr(),NULL,"connect");
        }
#endif
    }
    else {
        if ( !socknum ) {
            struct servent *serverInfo;
            serverInfo = getservbyname(sockname,NXLT("tcp"));
            if ( !serverInfo )
                ElfStrAbort(ERR_ILLVAL,XLT("Not a tcp service in /etc/services"),sockname);
            socknum = ntohs(serverInfo->s_port);
        }
        /* If no hostname passed in, assume current host */
        strncpy(hostname,passed_hostname,100);
        hostname[100] = 0;
        FullTrim(hostname,hostname);
        if ( !hostname[0] )
            gethostname(hostname,100);
        info = gethostbyname(hostname);
        if ( !info ) {
            CLOSE_CHANNEL(sockfd);
            ElfStrAbort(ERR_ILLVAL,NULL,hostname);
        }
        bzero(&in_serv_addr, sizeof(in_serv_addr));
        in_serv_addr.sin_family = AF_INET;
        in_serv_addr.sin_addr.s_addr = *((int*)(info->h_addr));
        in_serv_addr.sin_port = htons(socknum);

retry_connect:
        if ( connect( sockfd, (struct sockaddr *)(&in_serv_addr),
                                    sizeof(in_serv_addr)) == SOCKET_ERROR ) {
#if AX_X11
            error_code = ErrnoErr();
#else
            error_code = WSAGetLastError();
            if ( (error_code == WSAECONNREFUSED) && (retry_count < 6)){
               /* because of some timing issues, where the socket may    */
               /* not yet be fully created by the other process, we will */
               /* retry a couple of time before giving up             	 */
               retry_count++;
               Sleep( 1000 );     /* sleep for 1 second */
               goto retry_connect;
            }      
#endif
            CLOSE_CHANNEL(sockfd);
            ElfStrAbort( error_code, (char *)NULL,"connect");
        }
    }

    if ( !AxSockMakeChannelNonBlocking(sockfd) )
        ElfStrAbort(ErrnoErr(),NULL,"nonblock");

    sockfd = AxNoteSocket(sockfd,NULL,AxInTask,FALSE);
    return(sockfd);
}

/*************************************************************************
function: ElfbSockWrite - Write to a socket
*/
ElfbSockWrite (uid,data)
int uid;
data_ptr data;
{
    int arrsize;

    if ( !AxSockFdFromUid(uid) )
        ElfStrAbort(ERR_ILLVAL,XLT("Cannot write, channel not open."),NULL);

    arrsize = ElfbArraySize(data);  /* can throw error */

/* send type - NORMAL unless this is called by SendErrorToElf */

    WrtBufPos = 0;
    if ( -1 == add_long(uid,MsgType) )
        ElfStrAbort(ErrnoErr(),NULL,"write socket");

    if ( -1 == SendSome(uid,data) )
        ElfStrAbort(ErrnoErr(),NULL,"write socket");

    return(flush_buf(uid));
}

/*************************************************************************
function: ElfbSockRPCWrite - Write to a socket using RPC style packing
*/
ElfbSockRPCWrite (uid,code,data)
int uid;
int code;
data_ptr data;
{
    char *bufP;
    int size;
    data_ptr aP;
    char resourceName[20];
    int taskid;
    int packedElfIO;    /* TRUE if channel set for packed IO */

    if ( !AxSockFdAndPackerFromUid(uid,&packedElfIO) )
        ElfStrAbort(ERR_ILLVAL,XLT("Cannot write, channel not open."),NULL);

    aP = AxMakeArray(2);
    AxAddIntToArray(aP,0,code);
    AxAddArrayToArray(aP,1,data);
    size = AxMemWritePacked(aP,0,&bufP,packedElfIO);
    WrtBufPos = 0;

    taskid = ThimUIDPid(AxInTask);
    sprintf(resourceName,"**rpc%d$",uid);
    ACImOk();
    AxLockResource(resourceName,taskid);

    code = AxSocketWriter(uid,bufP,size);

    AxUnlockResource(resourceName,taskid);

    ((data_array*)aP)->vals[1] = NULL;  /* so won't free return val */
    AxFreeData(aP);                 /* free old enclosing array */
    free(bufP);
    if ( code == -1 )
        ElfStrAbort(ErrnoErr(),NULL,"write socket");
    return(0);
}

/*************************************************************************
function: ElfbSockWrite - Write to a socket
*/
void ElfbSockWriteBinary (uid,data)
int uid;
data_ptr data;
{
    char *str;
    int arrsize;

    if ( !AxSockFdFromUid(uid) )
        ElfStrAbort(ERR_ILLVAL,XLT("Cannot write, channel not open."),NULL);

    arrsize = ElfbBinarySize(data); /* can throw error */
    str = ((data_binary*)data)->bytes;
    WrtBufPos = 0;
    if ( -1 == add_bytes(uid,str,arrsize) )
        ElfStrAbort(ErrnoErr(),NULL,"write socket");
    if ( -1 == flush_buf(uid) )
        ElfStrAbort(ErrnoErr(),NULL,"write socket");
    return;
}

/*************************************************************************
function: AxfSockWriteAscii - Write Ascii data to a socket
*/
data_ptr AxfSockWriteAscii (argDatum)
data_ptr argDatum;
{
    int uid;
    int null_term;
    char *str;
    int len;

    uid = IntFromArgArray(argDatum,0);
    str = StrFromArgArray(argDatum,1);
    null_term = IntFromArgArray(argDatum,2);

    if ( !AxSockFdFromUid(uid) )
        ElfStrAbort(ERR_ILLVAL,XLT("Cannot write, channel not open."),NULL);

    len = strlen(str);

    if (null_term)
        len++;

    WrtBufPos = 0;

    if ( -1 == add_bytes(uid,str,len) )
        ElfStrAbort(ErrnoErr(),NULL,"write socket ascii");
    if ( -1 == flush_buf(uid) )
        ElfStrAbort(ErrnoErr(),NULL,"write socket ascii");

    return(ELF_CREATE_INTEGER(len));
}

/*************************************************************************
function: SendSome - Recursive sender
*/
SendSome (uid,data)
int uid;
data_ptr data;
{
    char *str;
    int templong;
    double tempnum;
    int dex;
    int arrsize;
    data_ptr dP;
    static char zeros[3];
    char buf[50];

    arrsize = ElfbArraySize(data);  /* can throw error */
    if ( -1 == add_long(uid,SOCK_ARRAY) )
        return(-1);
    if ( -1 == add_long(uid,arrsize) )
        return(-1);

    for ( dex = 0; dex < arrsize; dex++ ) {
        dP = (data_ptr)AnyFromArray(data,dex);
        if ( !dP ) {
            if ( -1 == add_long(uid,SOCK_NULL) )
                return(-1);
        }
        else if ( IS_INTEGER(dP) ) {
            templong = GET_INTEGER(dP);
            if ( -1 == add_long(uid,SOCK_LONG) )
                return(-1);
            if ( -1 == add_long(uid,templong) )
                return(-1);
        }
        else if ( ElfbIsNumber(dP) ) {
            tempnum = NumFromDataPtr(dP);
            if ( -1 == add_long(uid,SOCK_FLOAT) )
                return(-1);
            sprintf(buf,"%.15g",tempnum);
            if ( -1 == add_c_string(uid,buf) )
                return(-1);
        }
        else if ( ElfbIsString(dP) ) {
            str = StrFromDataPtr(dP);
            if ( -1 == add_long(uid,SOCK_STRING) )
                return(-1);
            if ( -1 == add_c_string(uid,str) )
                return(-1);
        }
        else if ( ElfbIsArray(dP) )
            SendSome(uid,dP);
        else if ( ElfbIsBinary(dP) ) {
            if ( -1 == add_long(uid,SOCK_BINARY) )
                return(-1);
            templong = ElfbBinarySize(dP);
            if ( -1 == add_long(uid,templong) )
                return(-1);
            str = ((data_binary*)dP)->bytes;
            if ( -1 == add_bytes(uid,str,templong) )
                return(-1);
        }
        else if (ElfbIsObject(dP)) {
            if ( -1 == add_long(uid,SOCK_OBJECT) )
                return(-1);
            templong = OX(dP)->objDex;
            if ( -1 == add_long(uid,templong) )
                return(-1);
            templong = OX(dP)->objUid;
            if ( -1 == add_long(uid,templong) )
                return(-1);
        }
    }
    return(0);
}

/*************************************************************************
function: add_bytes - Write bytes to socket thru buffering
*/
static int add_bytes(fd,p,cnt)
int fd;
char *p;
int cnt;
{
    while ( cnt ) {
        while ( cnt && (WrtBufPos < BUFSIZ) ) {
            WrtBuf[WrtBufPos++] = *p++;
            cnt--;
        }
        if ( cnt ) {    /* buffer must have filled */
            if ( -1 == flush_buf(fd) )
                return(-1);
        }
    }
    return(0);
}

static int add_long(fd,l)
int fd;
int l;
{
    l = htonl(l);
    return(add_bytes(fd,(char*)&l,sizeof(int)));
}

static int add_c_string(fd,p)
int fd;
char *p;
{
    int len;
    static char zeros[3];

    len = strlen(p) + 1;
    add_long(fd,len);
    if ( -1 == add_bytes(fd,p,len) )
        return(-1);
    return(0);
}

static read_long(fd,l)
int fd;
int *l;
{
    int code;
    int buf;

    code = read_bytes(fd,(char*)&buf,sizeof(int));
    *l = ntohl(buf);
    return(code);
}


static read_bytes(fd,p,cnt)
int fd;
char *p;
int cnt;
{
    register int timer = ReadTimer;

    ReadTimer = -1;
    return(AxTimedReadBytes(fd,p,cnt,timer,TRUE));
}

static flush_buf(fd)
int fd;
{
    int code;

    code = AxSocketWriter(fd,WrtBuf,WrtBufPos);
    WrtBufPos = 0;
    return(code);
}

/*************************************************************************
function: ElfbSockReadBinary - Read message from socket into elf array
*/
void ElfbSockReadBinary (uid,timer,maxbytes,infinite_after_first)
int uid;
int timer;
int maxbytes;
int infinite_after_first;
{
    extern data_ptr ElfRetData;    /* Alis func generic return value */
    data_ptr aP;
    char *p;
    int did;

    if ( maxbytes == 0 )
        return;

    if ( !AxSockFdFromUid(uid) )
        ElfStrAbort(ERR_ILLVAL,XLT("Cannot read, channel not open."),NULL);

    VERIFY((maxbytes > 0) && (maxbytes < 10000000),ERR_ILLVAL);
    p = malloc(maxbytes);
    did = AxTimedReadBytes(uid,p,maxbytes,timer,infinite_after_first);
    SVERIFY(-1 != did,ErrnoErr(),NXLT("read socket"));
    if ( did != 0 ) {
        ElfRetData = aP = ELF_CREATE_BINARY(did);
        memcpy(((data_binary*)aP)->bytes,p,did);
    }
    free(p);
    AxRecycleIfTaskLockingActive();
    return;
}


/*************************************************************************
function: ElfbSockRead - Read message from socket into elf array
*/
void ElfbSockRead (uid,timer)
int uid;
int timer;  /* -1 for forever, else seconds */
{
    extern data_ptr ElfRetData;    /* Alis func generic return value */
    int type;
    data_ptr aP;

    if ( !AxSockFdFromUid(uid) )
        ElfStrAbort(ERR_ILLVAL,XLT("Cannot read, channel not open."),NULL);

    ReadTimer = timer;
    SVERIFY(-1 != read_long(uid,&type),ErrnoErr(),NXLT("read socket"));
    ElfRetData = aP = ReadSomeData(uid);
    AxRecycleIfTaskLockingActive();
    return;
}

/*************************************************************************
function: ElfbSockRPCRead - Read message from socket into elf array
*/
void ElfbSockRPCRead (uid,timer)
int uid;
int timer;  /* -2 for read whole buf as msg, -1 for forever, else seconds */
{
    extern data_ptr ElfRetData;    /* Alis func generic return value */
    int size;
    data_ptr aP, eP;
    int code;
    char *buf;
    int cnt;
    char errorStr[1001];
    char errorObj[1001];
    data_ptr AxRPCRead();

    if ( !AxSockFdFromUid(uid) )
        ElfStrAbort(ERR_ILLVAL,XLT("Cannot read, channel not open."),NULL);

    if ( timer < 0 )
        ReadTimer = -1;
    else
        ReadTimer = timer;
    cnt = read_long(uid,&size);
    if ( cnt == 0 ) {           /* timed out */
        ElfRetData = NULL;
        return;
    }
    SVERIFY(-1 != cnt,ErrnoErr(),NXLT("read socket"));
    buf = (char*)malloc(size);
    if ( -1 == read_bytes(uid,buf,size) ) {
        free(buf);
        ElfStrAbort(ErrnoErr(),NULL,"socket read");
    }
    aP = AxRPCRead(buf);
    if ( timer == -2 ) {
        ElfRetData = aP;
    }
    else if ( AxIntFromArray(aP,0) == 0 ) {
        ElfRetData = AxArrayFromArray(aP,1);
        ((data_array*)aP)->vals[1] = NULL;  /* so won't free return val */
        AxFreeData(aP);                 /* free old enclosing array */
    }
    else {
        eP = AxArrayFromArray(aP,1);
        code = AxIntFromArray(eP,0);
        strncpy(errorStr,AxStrPtrFromArray(eP,1),1000);
        errorStr[1000] = 0;
        strncpy(errorObj,AxStrPtrFromArray(eP,2),1000);
        errorObj[1000] = 0;
        AxFreeData(aP);
        free(buf);
        ElfStrAbort(code,errorStr,errorObj);
    }
    free(buf);
    AxRecycleIfTaskLockingActive();
    return;
}

/*************************************************************************
function: ElfbSockRPCCall - Merge of RPCWrite & RPCRead
*/
void ElfbSockRPCCall (uid,code,data)
    int uid;
    int code;
    data_ptr data;
{
    ElfbDoSockRPCCall(uid, code, data, TRUE);
}

data_ptr ElfbDoSockRPCCall (uid, code, data, throwError)
    int uid;
    int code;
    data_ptr data;
    int throwError; /* if TRUE, will throwError, if FALSE returns error */
{
    char *bufP;
    int size;
    data_ptr aP, eP, retData=NULL;
    char resourceName[20];
    int taskid;
    char *buf;
    int err;
    char errorStr[1001];
    char errorObj[1001];
    data_ptr AxRPCRead();
    int packedElfIO;    /* TRUE if channel set for packed IO */

    if ( !AxSockFdAndPackerFromUid(uid,&packedElfIO) )  {
        char *p = XLT("Cannot write, channel not open.");
        if(throwError)
            ElfStrAbort(ERR_ILLVAL, p, NULL);
        retData = AxMakeArray(3);
        AxAddIntToArray(retData, 0, ERR_ILLVAL);
        AxAddStrToArray(retData, 1, p);
        return(retData);
    }

    taskid = ThimUIDPid(AxInTask);
    sprintf(resourceName,"**rpc%d$",uid);
    ACImOk();
    AxLockResource(resourceName,taskid);

/* WRITE!!! */

    aP = AxMakeArray(2);
    AxAddIntToArray(aP,0,code);
    AxAddArrayToArray(aP,1,data);
    size = AxMemWritePacked(aP,0,&bufP,packedElfIO);
    WrtBufPos = 0;
    err = AxSocketWriter(uid,bufP,size);
    ((data_array*)aP)->vals[1] = NULL;  /* so won't free return val */
    AxFreeData(aP);                 /* free old enclosing array */
    free(bufP);
    if ( err == -1 ) {
        char *p = XLT("write socket");
        AxUnlockResource(resourceName,taskid);
        if(throwError)
            ElfStrAbort(ErrnoErr(),NULL,p);
        retData = AxMakeArray(3);
        AxAddIntToArray(retData, 0, ErrnoErr());
        AxAddStrToArray(retData, 2, p);
        return(retData);
    }

/* NOW READ!!! */

    ReadTimer = -1;
    if ( -1 == read_long(uid,&size) ) {
        char *p = XLT("read socket");
        AxUnlockResource(resourceName,taskid);
        if(throwError)
            ElfStrAbort(ErrnoErr(),NULL, p);
        retData = AxMakeArray(3);
        AxAddIntToArray(retData, 0, ErrnoErr());
        AxAddStrToArray(retData, 2, p);
        return(retData);
    }

    buf = (char*)malloc(size);
    if ( -1 == read_bytes(uid,buf,size) ) {
        char *p = XLT("read socket");
        AxUnlockResource(resourceName,taskid);
        free(buf);
        if(throwError)
            ElfStrAbort(ErrnoErr(),NULL,p);
        retData = AxMakeArray(3);
        AxAddIntToArray(retData, 0, ErrnoErr());
        AxAddStrToArray(retData, 2, p);
        return(retData);
    }
    aP = AxRPCRead(buf);
    if ( AxIntFromArray(aP,0) == 0 ) {
        ElfRetData = AxArrayFromArray(aP,1);
        ((data_array*)aP)->vals[1] = NULL;  /* so won't free return val */
        AxFreeData(aP);                 /* free old enclosing array */
    } else {
        eP = AxArrayFromArray(aP,1);
        err = AxIntFromArray(eP,0);
        strncpy(errorStr,AxStrPtrFromArray(eP,1),1000);
        errorStr[1000] = 0;
        strncpy(errorObj,AxStrPtrFromArray(eP,2),1000);
        errorObj[1000] = 0;
        AxFreeData(aP);
        AxUnlockResource(resourceName,taskid);
        free(buf);
        ElfStrAbort(err,errorStr,errorObj);
    }

    free(buf);
    AxRecycleIfTaskLockingActive();
    AxUnlockResource(resourceName,taskid);
    return(NULL);
}

extern data_ptr AxTaskCreateSizedElfString();

/*************************************************************************
function: ReadSomeData - Recursive Read array into elf
*/
static data_ptr ReadSomeData (uid)
int uid;
{
    int type;
    int templong;
    static char *errobj = NXLT("read socket");
    char str[501];
    char *buf;
    data_ptr dP;
    int ix;

    SVERIFY(-1 != read_long(uid,&type),ErrnoErr(),errobj);
    switch ( type ) {
        case SOCK_LONG:
            SVERIFY(-1 != read_long(uid,&templong),ErrnoErr(),errobj);
            return(ELF_CREATE_NUMBER((double)templong));
            break;

        case SOCK_STRING:
            SVERIFY(-1 != read_long(uid,&templong),ErrnoErr(),errobj);
            dP = AxTaskCreateSizedElfString(THIMpid,templong-1,"");
            SVERIFY(-1 != read_bytes(uid,SX(dP)->text,templong), ErrnoErr(),errobj);
            return(dP);
            break;

        case SOCK_FLOAT:
            SVERIFY(-1 != read_long(uid,&templong),ErrnoErr(),errobj);
            SVERIFY((templong < 500),ERR_STRTOOLONG,errobj);    /* never happen! */
            SVERIFY(-1 != read_bytes(uid,str,templong), ErrnoErr(),errobj);
            return(ELF_CREATE_NUMBER(atof(str)));
            break;

        case SOCK_BINARY:
            SVERIFY(-1 != read_long(uid,&templong),ErrnoErr(),errobj);
            SVERIFY(dP = ELF_CREATE_BINARY(templong),ErrnoErr(),errobj);
            buf = ((data_binary*)dP)->bytes;
            SVERIFY(-1 != read_bytes(uid,buf,templong), ErrnoErr(),errobj);
            return(dP);
            break;

        case SOCK_ARRAY:
            SVERIFY(-1 != read_long(uid,&templong),ErrnoErr(),errobj);
            dP = ELF_CREATE_ARRAY(templong);
            for ( ix = 0; ix < templong; ix++ )
                dP = WriteArray(dP,ix,ReadSomeData(uid));
            return(dP);
            break;
        case SOCK_OBJECT:
            dP = ELF_CREATE_REMOTE_OBJECT(0,0,NULL);
            SVERIFY(-1 != read_long(uid,&templong),ErrnoErr(),errobj); /* dex */
            OX(dP)->objDex = templong;
            SVERIFY(-1 != read_long(uid,&templong),ErrnoErr(),errobj); /* uid */
            OX(dP)->objUid = templong;
            /* OX(dP)->objHostname to be filled in later by caller */
            return(dP);
        case SOCK_NULL:
        default:
            return(0);
    }
}

/*************************************************************************
function: ElfbSockClose - Close a socket
*/
void ElfbSockClose (sockP)
char *sockP;
{
    int uid;
    char sockname[STRPATH+1];

    strcpy(sockname,sockP);
    FullTrim(sockname,sockname);
    uid = AxFindSocketByName(sockname);
    SVERIFY(AxSocketClose(uid),ERR_ILLVAL,sockname);
}

/*************************************************************************
function: ElfbChanClose - Close a socket channel
*/
void ElfbChanClose (uid)
int uid;
{
    AxSocketClose(uid); /* quietly ignore request for channel not currently open */
}

void ElfMsgPipe(){abort();}


/*************************************************************************
function: ElfbRPCAccept - Accept connection and return incoming message
*/
int ElfbRPCAccept (uid)
int uid;
{
    int sockfd;
    int code, clilen;
#if AXMS 
    struct sockaddr un_cli_addr;
#else
    struct sockaddr_un un_cli_addr;
#endif 
    long senderType, senderTask, size, msgType;
    char *msgP;
    int channel;
    extern int AxElfrtMode;

    if ( !uid ) {
        if ( !AxSockFd ) {  /* only if axmain started for debugging */
            if ( !AxElfrtMode && !JavaMode )
                ElfTaskProc(ELElfMacroId(ELFM_PROCESS_OPTIONS_));
            exit(0);    
        }
    }

/* Add socket to list looked for by IO handler */

    ACImOk();
    if ( !uid ) {
        uid = AxFindSocketByName(NXLT("**axnet_primary**"));
        if ( !uid )
            uid = AxNoteSocket(AxSockFd,NXLT("**axnet_primary**"),AxInTask,TRUE);
    }

/* Wait for a connection from a client */

msg_getter:
    if ( -1 == AxSockReadUp(uid) )
        ElfStrAbort(ErrnoErr(),NULL,"accept");
    TaskInfo[AxInTask].waiting_on_socket = TRUE;
    code = AxGetMessage(-1,&msgP,&size,&senderType,&senderTask,&msgType);
    TaskInfo[AxInTask].waiting_on_socket = FALSE;
    if ( code != IPC_SOCKET ) {
        AxDeferMessage();
        goto msg_getter;
    }
/*        AxMessageAbsorbed(); NO MESSAGE, don't discard it! */
    sockfd = AxSockFdFromUid(uid);
    if ( !sockfd )
        ElfStrAbort(ERR_ILLVAL,XLT("Lost channel"),NXLT("accept on socket"));

    clilen = sizeof(un_cli_addr);
    channel = accept( sockfd, (struct sockaddr *)(&un_cli_addr), &clilen);
    if ( channel == INVALID_SOCKET )
        ElfStrAbort(ErrnoErr(),NULL,"accept");
    if ( !AxSockMakeChannelNonBlocking(channel) )
        ElfStrAbort(ErrnoErr(),NULL,"nonblock");
    channel = AxNoteSocket(channel,NULL,AxInTask,FALSE);
    return(channel);
}

/*************************************************************************
function: ElfbRPCRespond - Send response to client
*/
void ElfbRPCRespond (channelP,response)
data_ptr channelP,response;
{
    data_ptr rP;        /* sent back to Aster*x */
    char *outBuf;   /* holds the message on the way out */
    int32 size;     /* size of socket message */
    int32 cnt;      /* counter */
    int32 uid;
    int packedElfIO;    /* TRUE if channel set for packed IO */

    uid = GET_INTEGER(channelP);
    if ( !AxSockFdAndPackerFromUid(uid,&packedElfIO) )
        ElfStrAbort(ERR_ILLVAL,XLT("Cannot write, channel not open."),NULL);

/* respond to Client */

    rP = AxMakeArray(2);
    AxAddIntToArray(rP,0,0);
    AxAddArrayToArray(rP,1,response);
    size = AxMemWritePacked(rP,0,&outBuf,packedElfIO);
    cnt = AxSocketWriter(uid,outBuf,size);
    AxCksWriter(uid,outBuf,size);

/* Release arrays and temporarily allocated data */

    ((data_array*)rP)->vals[1] = NULL;  /* so won't free return val */
    AxFreeData(rP);
    free(outBuf);
    if ( cnt == -1 )
        ElfStrAbort(ErrnoErr(),NULL,"rpc responding");
}

/*************************************************************************
function: ElfbRPCErrRespond - Throw error back to client
*/
void ElfbRPCErrRespond (channelP,codeP,strP,objP)
data_ptr channelP, codeP, strP, objP;
{
    data_ptr rP,        /* sent back to Aster*x */
            errP;   /* on error, sent back to Aster*x */
    char *outBuf;   /* holds the message on the way out */
    int32 size;     /* size of socket message */
    int32 cnt;      /* counter */
    char *errstr;
    int32 code;
    int32 uid;

    uid = GET_INTEGER(channelP);
    if ( !AxSockFdFromUid(uid) )
        ElfStrAbort(ERR_ILLVAL,XLT("Cannot write, channel not open."),NULL);

/* respond to Client */

    code = GET_INTEGER(codeP);
    errstr = StrFromDataPtr(strP);
    if ( !errstr[0] )
        errstr = AxGetStr(code);
    rP = AxMakeArray(2);
    AxAddIntToArray(rP,0,1);
    errP = AxMakeArray(3);
    AxAddIntToArray(errP,0,code);
    AxAddStrToArray(errP,1,errstr);
    AxAddStrToArray(errP,2,StrFromDataPtr(objP));
    AxAddArrayToArray(rP,1,errP);

    size = AxMemWrite(rP,0,&outBuf);
    cnt = AxSocketWriter(uid,outBuf,size);
    AxCksWriter(uid,outBuf,size);

/* Release arrays and temporarily allocated data */

    AxFreeData(rP);
    free(outBuf);
    if ( cnt == -1 )
        ElfStrAbort(ErrnoErr(),NULL,"rpc responding");
}

/*************************************************************************
function: AxRecycleIfTaskLockingActive - Keep this task from running
If another task has locked the schedular, this keeps the calling task
from executing.
*/
AxRecycleIfTaskLockingActive ()
{
    extern int AxLockingTask;

    while ( AxLockingTask && (TaskInfo[AxInTask].uid != AxLockingTask) )
        TaskRecycle();
}

/*************************************************************************
function: AxfAxNet - Wrapper for axnet calls
*/
data_ptr AxfAxNet (argDatum)
data_ptr argDatum;
{
    data_ptr cmdCodeP = AnyFromArgArray(argDatum,0),
            arg1 = AnyFromArgArray(argDatum,1),
            arg2 = AnyFromArgArray(argDatum,2),
            arg3 = AnyFromArgArray(argDatum,3),
            arg4 = AnyFromArgArray(argDatum,4),
            arg5 = AnyFromArgArray(argDatum,5),
            arg6 = AnyFromArgArray(argDatum,6);

    data_ptr ret_ptr;
    int code;
    char *str, *obj;
    char username[50];
    char *p;
#if AXMS
    int ret;
    DWORD size_username;
#endif

    switch ( GET_INTEGER(cmdCodeP) ) {

    case 1:         /* REGISTER SERVICE */
    {
        data_ptr nameP, privateP, passwdP;

        nameP = arg1;
        privateP = arg2;
        passwdP = arg3;
#if AXMS
        if ( !AxSocketName )
            {
            /* On NT, there is no starter program which registers a service
             * and starts the listener task, so we need to do it here.
             */
            int save_thimpid = THIMpid;
            AxSockFd = AxServerStartup(AxStrFromDataPtr(nameP));
            /* need to set this sysvar to emulate what the Unix side
             * does on startup.
             */
            ElfbSetSysVar(NXLT("first_listener_inited$"),
                                ELF_CREATE_INTEGER(BTRUE));
            /* this sets context to task zero, so need to reset after. */
            ElfStartNewTask(ELElfMacroId(ELFM_RPC_LISTENER_));
            VfyContext(save_thimpid);
            break;
            }
#endif
        AxNetRegisterService(AxSocketName,
                        AxStrFromDataPtr(nameP),
                        AxIntFromDataPtr(privateP),
                        AxStrFromDataPtr(passwdP),
                        &code, &str, &obj);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
        break;
    }
    case 2:             /* RPC CALL, channel was already opened */
    {
        int channel;
        data_ptr serviceArgs, response;
        int taskid;
        char resourceName[30];

        channel = AxSockFdFromUid(GET_INTEGER(arg1));
        if ( !channel )
            ElfStrAbort(ERR_ILLVAL,XLT("Lost channel"),NXLT("channel"));
        serviceArgs = arg3;

        taskid = ThimUIDPid(AxInTask);
        sprintf(resourceName,"**axnet%d$",channel);
        AxLockResource(resourceName,taskid);
        axSendIt(channel,arg2,serviceArgs,&code,&str,&obj);
        if ( code != 0 ) {
            AxUnlockResource(resourceName,taskid);
            ElfStrAbort(code,str,obj);
        }
        response = AxGetResponseFromServer(channel,&code,&str,&obj);
        /* if any object references are recvd from server, they should be noted
         * as remote objects (by setting hostname) otherwise they tend to ElfObjPtr()'d
         * real hostname should be set by rpc caller.
         */
        response = ChannelDeReferenceServerObjects(response, channel);
        AxUnlockResource(resourceName,taskid);
/*      AxReallocElfDataToAnotherTask(AxInTask,response); */
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
        return( response );
    }
    case 3:             /* Open channel to axnet service */
    {
        char *hostName, *serviceName;
        int channel;
        char sockName[100];
        data_ptr response, request;

        /* find service */

        hostName = StrFromDataPtr(arg1);
        serviceName = StrFromDataPtr(arg2);

        channel = AxNetConnect(hostName,&code,&str,&obj);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
        request = AxMakeArray(2);
        request = AxAddStrToArray(request,0,serviceName);
        request = AxAddIntToArray(request,1,hostName[0] != 0);
        AxSendMessageToServer(channel,AXNET_FIND_SERVICE,request,&code,&str,&obj);
        AxFreeData(request);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
        response = AxGetResponseFromServer(channel,&code,&str,&obj);
        if ( code != 0 ) {
            AxClientDisConnect(channel);
            ElfStrAbort(code,str,obj);
        }
        if ( !response )
            channel = channel;
        else {
/*          AxReallocElfDataToAnotherTask(AxInTask,response); */
            strcpy(sockName,AxStrPtrFromDataPtr(response));
            AxFreeData(response);
            AxClientDisConnect(channel);
            channel = AxClientConnect(0,sockName,hostName,&code,&str,&obj);
            if ( code != 0 )
                ElfStrAbort(code,str,obj);
        }

#if AX_X11
        /* under windows the select structure is an array of file descriptors */
        /* not a bitmask, so comparing a socket-descriptor number against the */
        /* max number of things one can select on is not meaningful          */
        if ( channel >= FD_SETSIZE ) {
            CLOSE_CHANNEL(channel);
            ElfStrAbort(ERR_ILLVAL,XLT("Too many sockets"),NULL);
        }
#endif

        channel = AxNoteSocket(channel,NULL,AxInTask,FALSE);
        ret_ptr = MK_INTEGER(channel);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
        return( ret_ptr );
    }
    case 4:             /* Close channel */
    {
        int channel;

        channel = AxSockFdFromUid(GET_INTEGER(arg1));
        if ( !channel )
            break;
        AxClientDisConnect(channel);
        break;
    }

    case 5:             /* List services on a host */
    {
        char *hostName;
        data_ptr response;

        hostName = StrFromDataPtr(arg1);

        response = AxNetServices(hostName,"",&code,&str,&obj);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
/*      AxReallocElfDataToAnotherTask(AxInTask,response); */
        return( response );
    }
    case 6:             /* Drop service */
    {
        char *name;

        name = StrFromDataPtr(arg1);
        AxNetUnRegisterService(name, &code, &str, &obj);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
        break;
    }
    case 7:             /* Start a server */
    {
        char *hostName;
		struct passwd *pwuid;
        data_ptr dataP;

        hostName = StrFromDataPtr(arg1);
        dataP = AxMakeArray(3);
        AxAddIntToArray(dataP,0,getuid());
        AxAddArrayToArray(dataP,1,AxCopyData(arg2));

        /* pass the username so that it can be verified on the other end */
        /* since the 'uid' is meaningless on NT                          */
#if AXMS
        size_username = sizeof( username );
        ret = MS32_GetUserName( username, &size_username );
        if ( !ret )
            /* do not know why this would fail, but ... */
            username[0] = 0;
#else
#if LINUX
		pwuid = getpwuid(geteuid());
		if (pwuid != NULL)
			strcpy(username, pwuid->pw_name);
#else
        cuserid( username );
#endif
#endif
        dataP = AxAddStrToArray( dataP,2,username );

        /* now, if the suer has specified it, add the encrypted password */
        /* for the remote machine.  THis is required only if the remote  */
        /* machine is an NT machine, since we must have both username and */
        /* password to run a process as the user on the remote NT machine */
        p = GetEncryptedPassword( hostName );
        dataP = AxAddStrToArray( dataP, 3, p );
        
        AxNetProc(hostName,dataP,&code,&str,&obj);
        AxFreeData(dataP);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
        break;
    }
    case 8:             /* Start a server as "zane" -- Applix internal ONLY */
    {
        char *hostName;
        data_ptr dataP;

        hostName = StrFromDataPtr(arg1);
        dataP = AxMakeArray(2);
        AxAddIntToArray(dataP,0,25);
        AxAddArrayToArray(dataP,1,AxCopyData(arg2));
        AxNetProc(hostName,dataP,&code,&str,&obj);
        AxFreeData(dataP);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
        break;
    }

    case 9:     /* disconnect channel from task */
#if BUILD_KANJI
       /* Ashisuto changed it back to ver3.2. The reason is seen in following
        * example:  1) Create a C dispatch function and create ELF macro
        * function such as AddNums(). 2) Use it in Spreadsheet. At this time,
        * the channel master is set. If you set it to ThimPid, after you close
        * this spreadsheet, the ElfCloseAllSockets() would be called. After
        * then, you open new spreadsheets, These spreadsheets would not be
        * able to use the macro Addnums() if you do not restart the applixware!
        * This change is without self-confident, at least for temperate
        * purpose.
        * 1996.06.03
        */
        if ( !AxChannelMaster(GET_INTEGER(arg1),GET_INTEGER(arg2)) )
#else
        if ( !AxChannelMaster(GET_INTEGER(arg1),
                                         ThimPidFromUID(GET_INTEGER(arg2))) )
#endif
            ElfStrAbort(ERR_ILLVAL,NXLT("Attempted to remaster a non-existant channel"),NULL);
        break;

    case 10:    /* lock channel */
        ACImOk();
        AxLockResource(StrFromDataPtr(arg1),ThimUIDPid(AxInTask));
        break;

    case 11:    /* unlock channel */
        AxUnlockResource(StrFromDataPtr(arg1),ThimUIDPid(AxInTask));
        break;

    case 12:    /* set "slaved" server */
        /* !!0 TODO */
        break;

    case 13: {    /* write to channel, get response */
        int channel = GET_INTEGER(arg1);
        ElfbSockRPCCall(channel,GET_INTEGER(arg2),arg3);
        ret_ptr = ChannelDeReferenceServerObjects(ElfRetData, channel);
        return( ret_ptr );
    }

    case 14:    /* define channel as being "slow" */
        AxSocketSetSlowness(GET_INTEGER(arg1),GET_INTEGER(arg2));
        break;

    case 15:                /* Start a slaved server */
    {
        char *hostName;
		struct passwd *pwuid;
        data_ptr dataP;
        int channel;

        hostName = StrFromDataPtr(arg1);
        dataP = AxMakeArray(2);
        AxAddIntToArray(dataP,0,getuid());
        AxAddArrayToArray(dataP,1,AxCopyData(arg2));
 
        /* pass the username so that it can be verified on the other end */
        /* since the 'uid' is meaningless on NT                          */
#if AXMS
        size_username = sizeof( username );
        ret = MS32_GetUserName( username, &size_username );
        if ( !ret )
            /* do not know why this would fail, but ... */
            username[0] = 0;
#else
#if LINUX
		pwuid = getpwuid(geteuid());
		if (pwuid != NULL)
			strcpy(username, pwuid->pw_name);
#else
		cuserid( username );  /*!!0 NOT A PORTABLE CALL!! */
#endif
#endif
        dataP = AxAddStrToArray( dataP,2,username );

 
        /* now, if the user has specified it, add the encrypted password */
        /* for the remote machine.  THis is required only if the remote  */
        /* machine is an NT machine, since we must have both username and */
        /* password to run a process as the user on the remote NT machine */
        p = GetEncryptedPassword( hostName );
        dataP = AxAddStrToArray( dataP, 3, p );
 

        channel = AxNetProcSlave(hostName,dataP,&code,&str,&obj);
        AxFreeData(dataP);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
        channel = AxNoteSocket(channel,NULL,AxInTask,FALSE);
        ret_ptr = MK_INTEGER(channel);
        return( ret_ptr );
    }
    case 16:
        ret_ptr = ELF_CREATE_STRING(AxSocketName);
        return( ret_ptr );

    case 17:
    {
        /* send command to axnet on host:
        arg1 = hostname
        arg2 = Axnet command code 'see axnet.h
        arg3 = args for command
        */
        char *hostName;
		struct passwd *pwuid;
        int channel, axnetCmdCode;
        data_ptr response;
        data_ptr dataP;

        hostName = StrFromDataPtr(arg1);
        axnetCmdCode = AxIntFromDataPtr(arg2);
        dataP = AxMakeArray(2);
        dataP = AxCopyData(arg3);
        channel = AxNetConnect(hostName,&code,&str,&obj);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);

        if ( axnetCmdCode == AXNET_SIGNAL_PROCESS )
            {
            /* pass the username so that it can be verified on the other end */
            /* since the 'uid' is meaningless on NT                          */
#if AXMS
            size_username = sizeof( username );
            ret = MS32_GetUserName( username, &size_username );
            if ( !ret )
                /* do not know why this would fail, but ... */
                username[0] = 0;
#else
#if LINUX
			pwuid = getpwuid(geteuid());
			if (pwuid != NULL)
				strcpy(username, pwuid->pw_name);
#else
            cuserid( username );
#endif
#endif
            dataP = AxAddStrToArray( dataP, AxArraySize(dataP), username );
            }
 
        /* now, if the user has specified it, add the encrypted password */
        /* for the remote machine.  This is required only if the remote  */
        /* machine is an NT machine, since we must have both username and */
        /* password to run a process as the user on the remote NT machine */
        p = GetEncryptedPassword( hostName );
        dataP = AxAddStrToArray( dataP, AxArraySize(dataP), p );
 
        AxSendMessageToServer(channel,axnetCmdCode,dataP,&code,&str,&obj);
        if ( code != 0 ) {
            AxClientDisConnect(channel);
            ElfStrAbort(code,str,obj);
        }

        response = AxGetResponseFromServer(channel,&code,&str,&obj);
        AxFreeData(dataP);
        AxClientDisConnect(channel);
        if ( code != 0 ) {
            ElfStrAbort(code,str,obj);
        }
        return( response );
    }   /* end case 17 */

    case 18:    /* Let aagwty know new axmain running */
    {
        int channel, uid;

        channel = JavaNetConnect(NULL,&code,&str,&obj);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
        AxSendMessageToServer(channel,JAVANET_ADD_LCL_SERVER,NULL,&code,&str,&obj);
        if ( code != 0 ) {
            AxClientDisConnect(channel);
            ElfStrAbort(code,str,obj);
        }
        AxGetResponseFromServer(channel,&code,&str,&obj);
        if ( code != 0 )
            ElfStrAbort(code,str,obj);
        uid = AxNoteSocket(channel,NXLT("**aagtwy**"),0,FALSE);
        /* for future 'connect-mode' support, axmain needs to
            close this when cnnection completed... */
        return(MK_INTEGER(uid));
    }   /* end case 18 */

    }

    return( NullDataPtr );
}





AxCallProcess(char *cmd){tempAbort("cannot find axnet");}
DLL_EXPORT elfData AxDispatch(int funcId, elfData funcData, int *quitNow)
{
    tempAbort("axnet routing error");
    return( (elfData)NULL );
}


void AxErrnoError(msg)
char *msg;
{
    char buf[1000];

    sprintf(buf,"%s: %s",msg,ErrnoTextFromUnix(errno));
    tempAbort(buf);
    return;
}
 
/*************************************************************************
function: AxfUserId - Return Unix user id
*/
data_ptr AxfUserId()
{
    return(ELF_CREATE_INTEGER(getuid()));
}

