/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elstrs.c
   Applix ASC version: 45
   Date: Tue Feb  2 15:21:10 1999
*/
#include <elf_env.h>
#include <getstr4.h>
#include <eltokens.h>
#include <wchars.h>

typedef struct keysrch {
    struct keysrch *next;
    int val;
    char text[1];
} keysrch;

#include <elstrs.p>

#define CHUNKSIZE 1000
static char *GetStorage;

#define assert(x) if ( !(x) ) abort();

#define MAX_NAMES 200

extern char **ElfProtected;    /* Elf names accesable only is sys mode */
extern int ElfNumProtected;
extern int ElfInitedStrs;	/* flag saying whether ElfInitStrs has run */

int ElfKeyCodes[] = {
	TK_RETURN, TK_FOR, TK_WHILE, TK_IF, TK_ELSE, TK_GOTO, TK_EXTERN,
	TK_ARRAYOF, TK_ARRAYOF2, TK_TRUE, TK_FALSE, TK_NULL, TK_ENDFUNC,
	TK_PROC, TK_UIPROC, TK_NRPROC, TK_FUNC, TK_COMPILE, TK_ENDCMD,
	TK_AND, TK_OR, TK_REM, TK_XOR, TK_NOT, TK_EQV, TK_IMP, TK_NEXT,
	TK_VAR, TK_TO, TK_NOTHING, TK_BY, TK_ON, TK_ERROR, TK_WEND, TK_NODEBUG,
	TK_DEFINE, TK_DEFINE, TK_INCLUDE, TK_LET, TK_INCLUDE, TK_THEN, TK_BEGIN,
	TK_ENDBLOCK, TK_CALL, TK_FORMAT, TK_SELECT, TK_CASE, TK_DEFAULT, TK_OBJECT, 
	TK_OF, TK_ENDCASE, TK_SYSVAR, TK_IN, TK_BREAK
};

int ElfObjectCodes[] = {
	TK_SET, TK_GET, TK_ENDFUNC, TK_ENDFUNC, TK_OBJVAR, TK_PUBLIC, TK_PRIVATE
};

int ElfSQLCodes[] = {
	TK_FROM, TK_GROUP, TK_ORDER, TK_BY, TK_CREATE, TK_TABLE, TK_VALUES,
	TK_INSERT, TK_DROP, TK_WHERE, TK_DISTINCT, TK_ALL, TK_IN, TK_SQL,
	TK_BETWEEN, TK_DECLARE, TK_PREPARE, TK_OPENWORD, TK_FETCH, TK_CLOSEWORD,
	TK_UPDATE, TK_EXECUTE, TK_ASC, TK_DESC, TK_UNION, TK_INTERSECT, TK_MINUS,
	TK_AT, TK_CURRENT, TK_UNPREPARE
};

int ElfOracleCodes[] = {
	TK_CONNECT, TK_START, TK_WITH
};

extern char *ElfKeyWords;   /* keywords list */
extern char *ElfSQLWords;   /* keywords list */
extern char *ElfOracleWords;   /* keywords list */
extern char *ElfRsvdWords;  /* future words list */
extern char *ElfObjectWords;	/* object mode keywords list */
extern int ElfcAllowSQL;	/* enabled by @@@SQL or @@@ORACLE */
extern int ElfcAllowOracle;	/* enabled by @@@ORACLE */
extern int ElfcObjectMode;		/* enabled by @@@OBJECT */
extern int ElfSysMode; /* TRUE if ELF compiled in system mode */

/* Token Separators */
/* Note \240 = 160 = 0xa0 = hard space in Applixware font */
static char *TokSeps =  " =><,!&|+-*/^\\(){}[].:;\n\t\"\240";
int IsElfTokSep[256];	/* quick lookup table */

/*************************************************************************
function: ElfInitStrs - Load standard strings into memory
*/
ElfInitStrs ()
{
    char storage[1024];     /* bigger than getStr's limit */
    char store[6000];   /* big enough for all keywords */
    char *protected[MAX_NAMES+1];
    char *tp;
    extern char *ElfOpChars;    /* operater chars */

    if ( ElfInitedStrs ) return;
    ElfInitedStrs++;

	for ( tp = TokSeps; *tp; tp++ )
		IsElfTokSep[((unsigned)*tp) & 0xFF] = TRUE;

    GetStorage = storage;
    *GetStorage = 0;

/* malloc the protected array */

    ElfProtected = (char**)TrueMalloc( (ElfNumProtected + 1)*sizeof(char*) );
    memcpy(ElfProtected,protected,ElfNumProtected*sizeof(char*));
    ElfProtected[ElfNumProtected] = NULL;

/* get the keywords list */

	strcpy(store,NXLT("return for while if else goto extern arrayof arrayof2 "));
	strcat(store,NXLT("true false null endfunction macro uimacro nrmacro function "));
	strcat(store,NXLT("@@@ endmacro and or mod xor not eqv imp next "));
	strcat(store,NXLT("var to nothing step on error wend nodebug "));
	strcat(store,NXLT("define #define include let #include then "));
	strcat(store,NXLT("begin end call format select case default object "));
	strcat(store,NXLT("of endcase sysvar in@ break  "));
    strcpy(ElfKeyWords = TrueMalloc(strlen(store)+1),store);
    ElfTokenize(ElfKeyWords);

/* get the reserved words list */

	strcpy(store,NXLT("number string integer struct array "));
	strcat(store,NXLT("select1 pointer control help  "));
    strcpy(ElfRsvdWords = TrueMalloc(strlen(store)+1),store);
    ElfTokenize(ElfRsvdWords);

/* get Object key words */

	strcpy(store,NXLT("set get endset endget objvar public private  "));
    strcpy(ElfObjectWords = TrueMalloc(strlen(store)+1),store);
    ElfTokenize(ElfObjectWords);

/* get the SQL key words */

	strcpy(store,NXLT("from group order by create table values insert drop "));
	strcat(store,NXLT("where distinct all in sql "));
	strcat(store,NXLT("between declare prepare open fetch close "));
	strcat(store,NXLT("update execute asc desc union intersect minus "));
	strcat(store,NXLT("at current unprepare  "));
    strcpy(ElfSQLWords = TrueMalloc(strlen(store)+1),store);
    ElfTokenize(ElfSQLWords);

/* get Oracle key words */

	strcpy(store,NXLT("connect start with  "));
    strcpy(ElfOracleWords = TrueMalloc(strlen(store)+1),store);
    ElfTokenize(ElfOracleWords);

/* get the operators list */

	strcpy(store, NXLT(". ! , : $ ? ; ( ) [ ] += -= >= <= <> = ^ - * / \\ >>= "));
	strcat(store, NXLT("<<= >> << ++ > < + { } "));
	strcpy(ElfOpChars = TrueMalloc(strlen(store)+1),store);
/*
    ElfOpChars = ". ! , : $ ? ; ( ) [ ] += -= >= <= <> = ^ - * / \\ >>= <<= >> << ++ > < + { } ";
*/
    ElfTokenize(ElfOpChars);
    ElfSetupQsrch();
}

/*************************************************************************
function: ElfIsProtected - Return whether a name is privleged
*/
ElfIsProtected (name)
register char *name;
{
    register char **p;

    for ( p = ElfProtected; *p; p++ )
        if ( streq(*p,name) )
            return(TRUE);
    return(FALSE);
}

/*************************************************************************
function: ElfValidNameBasedOnSysMode - Returns whether name can be created by compiler
This is used only by the compiler to determine whether it is free to
create a function/var with the passed name.  When running sysmode, names
ending in @_$ are allowed.
*/
typeErr ElfValidNameBasedOnSysMode (name)
register char *name;
{
    if ( strlen(name) > ELFIDSTR )
        return(ERR_STRTOOLONG);
    if ( !IsIdentifier(name) )
        return(ERR_BADNAME);
	if ( !ElfSysMode
			&& (strchr(name,'@')
				|| strchr(name,'$')
				|| (name[strlen(name)-1] == '_')) )
        return(ERR_BADNAME);
	if ( !ElfSysMode && ElfLookupSymbol(name,ElfLanguage) )
		return(ERR_MYNAME);
    if ( FindKeyword(name) )
        return(ERR_KEYWORD);
    if ( FindRsvdWord(name) )
        return(ERR_RSVDWORD);
    return(AOK);
}

/*************************************************************************
function: ElfValidName - Return whether user selected valid func name
*/
typeErr ElfValidName (name)
register char *name;
{
	return(ElfValidNameBasedOnSysMode(name));
}

keysrch *ElfKeysSrch[32];
keysrch *ElfRsvdSrch[32];
keysrch *ElfSQLSrch[32];
keysrch *ElfOracleSrch[32];
keysrch *ElfObjectSrch[32];
int ElfQsrchCalled;

static void doqsrch(arr,str)
register keysrch **arr;
register char *str;
{
    register char *p;
    register keysrch *pk;
    register int ix, dex;

    for ( ix = 0; *str; ix++ ) {
        for ( p = str; *str; str++ )
            *str = toupper(*str);
        str++;
        pk = (keysrch*)TrueMalloc(strlen(p)+sizeof(keysrch));
        strcpy(pk->text,p);
        pk->val = ix;
        dex = (pk->text[0] & 31);
        pk->next = arr[dex];
        arr[dex] = pk;
    }
}

/*************************************************************************
function: ElfSetupQsrch - Set up srchers for key and reserved words
*/
ElfSetupQsrch ()
{
    if ( !ElfQsrchCalled ) {
        doqsrch(ElfKeysSrch,ElfKeyWords);
        doqsrch(ElfObjectSrch,ElfObjectWords);
        doqsrch(ElfSQLSrch,ElfSQLWords);
        doqsrch(ElfOracleSrch,ElfOracleWords);
        doqsrch(ElfRsvdSrch,ElfRsvdWords);
		ElfQsrchCalled = TRUE;
    }
}

/*************************************************************************
function: FindRsvdWord - See whether string is a reserved word
*/
FindRsvdWord (p)
register char *p;
{
    register keysrch *pk;

    for ( pk = ElfRsvdSrch[*p & 31]; pk; pk = pk->next ) {
        if ( streq(pk->text,p) )
            return(TRUE);
    }
    return(FALSE);
}

/*
 * function:  FindKeyword - Compare string against keywords
*/
FindKeyword (p)
register char *p;
{
    register keysrch *pk;

    for ( pk = ElfKeysSrch[*p & 31]; pk; pk = pk->next ) {
        if ( streq(pk->text,p) )
            return(ElfKeyCodes[pk->val]);
    }
	if ( ElfcObjectMode ) {
		for ( pk = ElfObjectSrch[*p & 31]; pk; pk = pk->next ) {
			if ( streq(pk->text,p) )
				return(ElfObjectCodes[pk->val]);
		}
	}
	if ( ElfcAllowSQL ) {
		for ( pk = ElfSQLSrch[*p & 31]; pk; pk = pk->next ) {
			if ( streq(pk->text,p) )
				return(ElfSQLCodes[pk->val]);
		}
	}
	if ( ElfcAllowOracle ) {
		for ( pk = ElfOracleSrch[*p & 31]; pk; pk = pk->next ) {
			if ( streq(pk->text,p) )
				return(ElfOracleCodes[pk->val]);
		}
	}
    return(FALSE);
}

/*************************************************************************
function: ElfKeywordFromTokType - Return str of keyword
*/
char *ElfKeywordFromTokType (toktype)
int toktype;
{
	char *p;
	int ix;

    p = ElfKeyWords;
    for ( ix = 0; ix < sizeof(ElfKeyCodes)/sizeof(int); ix++ ) {
        if ( ElfKeyCodes[ix] == toktype )
			return(p);
        p += strlen(p)+1;
    }
	if ( ElfcObjectMode ) {
	    p = ElfObjectWords;
	    for ( ix = 0; ix < sizeof(ElfObjectCodes)/sizeof(int); ix++ ) {
	        if ( ElfObjectCodes[ix] == toktype )
				return(p);
	        p += strlen(p)+1;
	    }
	}
	if ( ElfcAllowSQL ) {
	    p = ElfSQLWords;
	    for ( ix = 0; ix < sizeof(ElfSQLCodes)/sizeof(int); ix++ ) {
	        if ( ElfSQLCodes[ix] == toktype )
				return(p);
	        p += strlen(p)+1;
	    }
	}
	if ( ElfcAllowOracle ) {
	    p = ElfOracleWords;
	    for ( ix = 0; ix < sizeof(ElfOracleCodes)/sizeof(int); ix++ ) {
	        if ( ElfOracleCodes[ix] == toktype )
				return(p);
	        p += strlen(p)+1;
	    }
	}
	return(NULL);
}

/*
 * function:  IsIdentifier - Return whether text OK for identifier
*/
IsIdentifier (p)
register char *p;
{
    register char ch;

    if ( !isalpha(*p) && *p != '@' )
        return(FALSE);

    p++;
    while ( ch = *p++ ) {
        if ( !isalnum(ch) && (NULL == strchr("_#@$",ch)) )
            return(FALSE);
    }
    if ( *(p-1) == '#' )    /* cannot have last char being a pound sign */
        return(FALSE);
    return(TRUE);
}

/*
 * function:  ElfTokenize - Convert space separated str to null seped
*/
ElfTokenize(p)
register char *p;
{
    while ( *p ) {
        if ( *p == ' ' )
            *p = 0;
        p++;
    }
}



/*
 * function:  IsOrientalSQLIdentifier - Return whether text OK for
 * SQL COLUMN/TABLE names. This is the complementation of IsIdentifier().
 * Ashisuto added.  95.10.06
 */
IsOrientalSQLIdentifier (p)
register char *p;
{
#if BUILD_ORIENT
    if (!ElfcAllowSQL)
        return(FALSE);
 
    if (!isalpha(*p) && !IS_MULTI_BYTE(p))
        return(FALSE);
 
    while (*p) {
        if (IS_MULTI_BYTE(p))
            p += 2;
        else if (isalnum(*p) || strchr("_#@$", *p))
            p++;
        else
            return(FALSE);
    }
 
    if ( *(p-1) == '#' )    /* cannot have last char being a pound sign */
        return(FALSE);
    return(TRUE);
#else
    return(FALSE);
#endif
}

