/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elsyms.c
   Applix ASC version: 10
   Date: Tue Feb  2 15:21:11 1999
*/
#include <stdio.h>
#include <aos.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <elf.h>
#include <errno.h>
#include <afile.h>

/*
	0 - Gothic
	1 - American
*/
#define MAX_ELF_LANGUAGES 2

#define MAX_ELF_MODULES 10
#define MODULE_NAME_LEN 15

#define BLTIN_ELF_MODULE "bltn_syms"
#define ELELF_MODULE "el_syms"
#define SSELF_MODULE "ss_syms"
#define WPELF_MODULE "wp_syms"
#define GEELF_MODULE "ge_syms"

/* Symbol table can hold 32767 entries */

#define MAX_SYMS 32767
typedef struct {
	short symval;		/* position in symbol table */
	short next;			/* thread to next in symbol table (or -1) */
	char *name;			/* ptr to symbol text, right from file */
} sym_t;

static char **NameList[MAX_ELF_LANGUAGES];		/* name mapping table */
 static char *SymText[MAX_ELF_LANGUAGES];			/* names of globals */
static short HashList[MAX_ELF_LANGUAGES][ELF_HASH_PRIME+1];	/* List heads */
static sym_t* HashData[MAX_ELF_LANGUAGES];						/* Lists */

/* Symbol table Module files list */
static struct elfmodules {
	char name[MODULE_NAME_LEN+1]; /* name of .dat files without lang. number */
	int symstart_ind;
} ElfModuleList[MAX_ELF_MODULES] = {  /* symstart_ind will be filled in during symbol reading */
	{ BLTIN_ELF_MODULE, 0 },
	{ ELELF_MODULE, 0 },
	{ SSELF_MODULE, 0 },	
	{ WPELF_MODULE, 0 },	
	{ GEELF_MODULE, 0 },	
	{ "", 0 }
};

/*************************************************************************
function: ElfAddModule - Add a new Symbol table module file name.
*/
int 
ElfAddModule(mod_name)
char* mod_name;
{
	char buf[500];
	int ix;
	if(mod_name == NULL || mod_name[0] == '\0' || strlen(mod_name) > MODULE_NAME_LEN)
		return(FALSE);
	for(ix=0; ix < MAX_ELF_MODULES && ElfModuleList[ix].name[0] != '\0'; ix++)
	    ; 
	if(ix >= MAX_ELF_MODULES)
	  return(FALSE);
	strcpy(ElfModuleList[ix].name,mod_name);
	ElfModuleList[ix].symstart_ind = -1;
	return(TRUE);
}

/*******************************************************************************
function: ElfGetModuleStartInd - Get start index of symbols for the given module.
*/
int 
ElfGetModuleStartInd(mod_name)
char* mod_name;
{
	int ix;
	if(mod_name == NULL)
		return(-1);
	for(ix=0; ix < MAX_ELF_MODULES; ix++) {
		if(streq(mod_name,ElfModuleList[ix].name)) {
			if(!HashData[0])
				ElfReadSymTab(0);
			return(ElfModuleList[ix].symstart_ind);
		}
	}
	return(-1);
}

/*************************************************************************
function: ElfReadSymTab - Read the symbol table file produced by elfsymtool.
*/
int
ElfReadSymTab (lang)
unsigned int lang;		/* language code */
{
	struct stat info;
	char *p, *modsym_ptr;
	char *PermAlloc();
	char *namelist[MAX_SYMS];
	int hashvals[MAX_SYMS];
	int symnum = 0;
	int fd, ix, hashval, size;
	sym_t *hashP;
	extern char *ThimApplixDir;	/* executables -> /usr/bin//thimble.data */
	char filename[STRPATH+1];
	char buf[500];
	int symtab_size=0;
	if ( lang >= MAX_ELF_LANGUAGES )	/* sanity */
		lang = 0;

	if ( HashData[lang] )	/* already called */
		return(0);
	/* compute symbol table size */
    for(ix=0; ix < MAX_ELF_MODULES && ElfModuleList[ix].name[0] != '\0'; ix++) {
		sprintf(filename,"%s/elf/%s%d.dat",ThimApplixDir, ElfModuleList[ix].name, lang); 
		AxSysDosUnixPath(filename);
		if ( -1 == stat(filename,&info) )	/* fetch size */
			goto fail;
		symtab_size += info.st_size;		
	}
	
	p = PermAlloc(symtab_size);
	SymText[lang] = p;
	namelist[0] = SymText[lang];
	modsym_ptr = SymText[lang];
	/* read symbols files into SymText */
    for(ix=0; ix < MAX_ELF_MODULES && ElfModuleList[ix].name[0] != '\0'; ix++) {
		ElfModuleList[ix].symstart_ind = symnum;
		sprintf(filename,"%s/elf/%s%d.dat",ThimApplixDir, ElfModuleList[ix].name, lang); 
		AxSysDosUnixPath(filename);
		if ( -1 == stat(filename,&info) )	/* fetch size */
			goto fail;
		fd = open(filename,O_RDONLY|O_BINARY,0);		/* open & read */
		if ( fd == -1 )
			goto fail;
		size = read(fd,modsym_ptr,info.st_size);
		if ( info.st_size != size )
			goto fail;
		close(fd);
		/* produce temp tables */
		for (p=modsym_ptr; p < modsym_ptr+info.st_size; p++ ) {
			if ( (*p == '\n') || (*p == 13) ) {	/* CR or newline */
				if ( *p == 13 ) {
					*p++ = 0;
				}
				else
					*p = 0;
				if ( (strlen(namelist[symnum]) == 0) ||	/* skip comments & blanks */
					(namelist[symnum][0] == '#') )
					continue;
				hashvals[symnum] = hash(namelist[symnum]);
				symnum++;
				namelist[symnum] = p+1;
			}
		}
		modsym_ptr += size;
	}

	/* allocate and fill out permanent tables */

	memset(HashList[lang],0,sizeof(HashList[lang]));
	hashP = (sym_t*)PermAlloc(sizeof(sym_t)*symnum);
	HashData[lang] = hashP;
	NameList[lang] = (char**)PermAlloc(sizeof(char*)*symnum);
	NameList[lang][0] = NXLT("unknown");

	hashP++;		/* skip 0'th */
	for ( ix = 1; ix < symnum; ix++,hashP++ ) {
		hashval = hashvals[ix];
		hashP->symval = ix;
		hashP->name = NameList[lang][ix] = namelist[ix];
		hashP->next = HashList[lang][hashval];
		HashList[lang][hashval] = ix;
	}
	return(0);
fail:
	sprintf(buf,"%s: initializing Elf",ErrnoTextFromUnix(errno));
	tempAbort(buf);
}

/*************************************************************************
function: ElfLookupSymbol - Returns sym id or 0 if not found
*/
ElfLookupSymbol (name,lang)
register char *name;
unsigned int lang;		/* language code */
{
	register short hashval = hash(name);
	register int ix;

	if ( lang >= MAX_ELF_LANGUAGES )	/* sanity */
		lang = 0;
	if ( !HashData[lang] )	/* load symtab if not here yet */
		ElfReadSymTab(lang);
	for ( ix = HashList[lang][hashval]; 
			ix && !streq(HashData[lang][ix].name,name); 
			ix = HashData[lang][ix].next ) {
	}
	return(ix);
}

/*************************************************************************
function: ElfCaselessLookupSymbol - Convert to upper case and lookup
*/
ElfCaselessLookupSymbol (name,lang)
register char *name;
unsigned int lang;		/* language code */
{
	char buf[STRLONG+1], *p;

	if ( strlen(name) > STRLONG )
		return(0);
	strcpy(buf,name);
	for ( p = buf; *p; p++ )
		*p = toupper(*p);
	return(ElfLookupSymbol(buf,lang));
}

/*************************************************************************
function: ElfSymbolNameFromId - Returns name from id
*/
char *ElfSymbolNameFromId (id,lang)
int id;
unsigned int lang;		/* language code */
{
	if ( lang >= MAX_ELF_LANGUAGES )	/* sanity */
		lang = 0;
	if ( !HashData[lang] )	/* load symtab if not here yet */
		ElfReadSymTab(lang);
	return(NameList[lang][id]);
}

/*************************************************************************
function: IsBltinSymCode - Return whether code within range of builtins
*/
IsBltinSymCode (n)
int n;
{
	if ( n == 0 )
		return(FALSE);
	return(n < ElfNumBuiltins());
}

