/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elwincom.c
   Applix ASC version: 175
   Date: Mon May 10 21:13:14 1999
*/
#include <stdio.h>
#include <aos.h>
#include <thimble.h>

#if AXMS
#include <basicwin.h>
#endif

#include <ax_xm.h>
#include <xmlib.h>
#include <elxwins.h>
#include <xmpixel.h>
#include <xmtable.h>
#include <xmcanvas.h>
#include <xmchoose.h>
#include <evtray.h>
#include <dwt2xm.h>
#include <adlgpoke.h>
#include <elf.h>
#include <elf_env.h>
#include <axtext.h>
#include <eltree.h>
#include <xmeditbx.h>
#include <getstr5.h>
#include <axfont.h>


#if AXMS
extern int Scrollbar_Thickness;
extern int Option_Pad;
extern int Entry_Height;
#endif
extern int ThimLogEvents;
extern int WidgetNeedsUpdated;
extern int AxAlternateDialogFont;
extern XFontStruct *AxDlgMainFont;
extern XFontStruct *AxDefaultReferenceFont;

#define XmCR_BUTTON_PRESS   10

#if 0  /* AX_X11 */
#define XWLOG(body) if ( ThimLogEvents ) XWINSlog body;
#else
#define XWLOG(body)
#endif

#if AXMS
extern int RunningWin95Look;
#endif

#include <elwincom.p>

/*************************************************************************
function: CallWidget - Send message to db_display@ or dbox main loop
*/
CallWidget (pid,widget,reason,info)
int pid;
AXWidget widget;
int reason;
data_ptr info;  /* !!0 Is this getting free'd anywhere? */
{
    data_ptr p, ap;
    data_ptr aP;
    AXWidget_handle w_handle;

    if ( p = ElfObjFromWidget(widget) ) {
        aP = ELF_CREATE_ARRAY(3);
        aP = WriteArray(aP,0,ElfDataCopy(p));
        aP = WriteArray(aP,1,AxMakeNumData(reason));
        aP = WriteArray(aP,2,info);
        SendElfData(pid,POKE_OBJECT_THROW,aP);
        ELF_DATA_FREE(aP);
    }
    else {
        w_handle = HANDLE_FROM_WIDGET(widget);
        aP = ELF_CREATE_ARRAY(3);
        aP = WriteArray(aP,0,AxMakeNumData(reason));
        ap = WriteArray(aP,1,AxMakeBinaryData(sizeof(w_handle),(char*)&w_handle));
        aP = WriteArray(aP,2,info);
        SendElfData(pid,POKE_WIDGET,aP);
        ELF_DATA_FREE(aP);
    }
}

/*************************************************************************
function: ElfObjFromWidget - Extract elfObject element
*/
data_ptr ElfObjFromWidget (w)
AXWidget w;
{
    return(w->elfObject);
}

/** get the widget pointer back out of the binary data object, into which */
/**  we stuffed it earlier so it could be carried around in the ELF layer */

AXWidget WidgetPointerFromHandleData( dataP )
data_ptr dataP;
{
    AXWidget_handle w_handle;
    AXWidget w;   
 
    if ( dataP == NULL )
        return( (AXWidget)NULL );
 
    if ( dataP->type != TYPE_BINARY )
        tempAbort("not binary data ");

    memcpy( &w_handle, BX(dataP)->bytes, sizeof(AXWidget_handle) );
    w = WIDGET_FROM_HANDLE(w_handle);
    return(w);
}

data_ptr
HandleDataFromWidgetPointer(w)
AXWidget w;
{
    AXWidget_handle w_handle;
    data_ptr retP;

    w_handle = HANDLE_FROM_WIDGET(w);
    retP = AxMakeBinaryData(sizeof(w_handle), (char*)&w_handle);
    return(retP);
}

int
readEventHandlerData(dp, dialog, list)
long dp;                 /* in: elf array */
int dialog;          /* in: bool */
eventHandlerList *list;  /* out: init this struct */
{
    int arrayIndex, size, ix;
    unsigned long *evMasks;
    data_ptr procnames, masks;
    char **nameStrings, *np;
    if (!list)
        return FALSE;
    if (!dp) {
        list->numProcs = 0;
        list->eventProcNames = (char **) NULL;
        list->eventMasks = (unsigned long *) NULL;
        return FALSE;
    }
            /* first get the event handler name strings */
    arrayIndex = dialog ? DB_event_handlers : DBDEXevent_handlers;
    procnames = AnyFromArray((data_ptr)dp, arrayIndex);
    size = ElfbArraySize(procnames);
    if (!size) {
        list->numProcs = 0;
        list->eventProcNames = (char **) NULL;
        list->eventMasks = (unsigned long *) NULL;
        return FALSE;
    }
    nameStrings = (char **) malloc (size * sizeof(char *));
    for (ix = 0; ix < size; ix++) {
        np = StrFromArray(procnames, ix);
        nameStrings[ix] = strcpy( malloc(strlen(np)+1), np);
    }
    list->numProcs = size;
    list->eventProcNames = nameStrings;
            /* now get the event masks; it's assumed that there
             * are the same # of them as the procnames */
    arrayIndex = dialog ? DB_event_masks : DBDEXevent_masks;
    masks = AnyFromArray((data_ptr)dp, arrayIndex);
    size = ElfbArraySize(masks); /* just in case, but better be the same.. */
    evMasks = (unsigned long *) malloc (size * sizeof(unsigned long));
    for (ix = 0; ix < size; ix++) {
        evMasks[ix] = (unsigned long) IntFromArray(masks, ix);
    }
    list->eventMasks = evMasks;
    return TRUE;
}

int
tossEventHandlerData(list)
eventHandlerList *list;
{
    int ix;

    if (list->eventProcNames) {
        for (ix = 0; ix < list->numProcs; ix++) {
            free( list->eventProcNames[ix]);
        }
        free(list->eventProcNames);
        list->eventProcNames = (char **)NULL;
    }
    if (list->eventMasks) {
        free(list->eventMasks);
        list->eventMasks = (unsigned long *)NULL;
    }
    list->numProcs = 0;
    return TRUE;
}

/*************************************************************************
function: ElfbUpdateDialogSize - Update dialog box height, width
*/
void ElfbUpdateDialogSize(dboxP, dataP, typeP)
data_ptr dboxP;  /* elf array */
data_ptr dataP;
data_ptr typeP;
{
    int x,y, wid,hyt, minwid,minhyt, maxwid,maxhyt, type;
    AXWidget id_widget;
    AXWidget_dialog dlg;
    extern short DpyWid; /* physical screen width */
    extern short DpyHt; /* physical screen height */
    Arg args[1];
    data_ptr pokeCodes;
 
 
    /* get the dialog box */
    id_widget = WidgetPointerFromHandleData(dataP);
    dlg = XmDlgInWidget((AXWidget_main)id_widget);
    if ( AxIntFromArray(dboxP, DB_in_edit) ) 
        deCopyDialog(dlg, dboxP);
 
    /* update poke information */
    pokeCodes = AnyFromArray(dboxP, DB_accept_poke);
    XtSetArg(args[0],XmNpokeCodes,pokeCodes);
    AxXtSetValues((AXWidget)dlg, args, 1);
 
    /* update size and position */
    type = typeP ? AxIntFromDataPtr(typeP) : 0;
    if ( type > 3 )
        type = 0;
    if ( type == 3 ) /* zoom to full screen */
    { 
        type = 2; 
        x = y = 0; 
        wid = DpyWid;
        hyt = DpyHt;
    } 
    else 
    {  
        x = AxIntFromArray(dboxP, DB_xpos);
        y = AxIntFromArray(dboxP, DB_ypos);
        AxScaleDialogFrame(dboxP, &wid,&hyt, &minwid,&minhyt, &maxwid,&maxhyt);
    } 
    AXmUpdateDialogSz((AXWidget_main)id_widget, x,y, wid,hyt, type);
}

/*************************************************************************
function: ElfbEntryFieldCreate - Create an enterable field                
Just a wrapper in order to get return type correct!                   
This returns an array!                                                    
*/                                                                        
void ElfbEntryFieldCreate (wtype,old,parent,info)                         
data_ptr wtype;     /* type of widget */                              
data_ptr old;   /* current widget id or 0 */                              
data_ptr parent;        /* parent window */                           
data_ptr info;          /* elf array of info */                       
{                                                                         
    ElfbCreateChangeWidget(wtype,old,parent,info);                    
}                                                                         

/*************************************************************
function: ElfbMenuStatus - Set status flags for a menu option
*/
void ElfbMenuStatus (list,w_dataP)
data_ptr list;          /* array of pairs of status */
data_ptr w_dataP;       /* of dialog box */
{
    int cnt, ix, val;
    data_ptr triplet;
    char *func;
    char *name;
    AXWidget w;

    w = WidgetPointerFromHandleData( w_dataP );
    cnt = ElfbArraySize(list);                            
    for ( ix = 0; ix < cnt; ix++ ) {
        triplet = AnyFromArray(list,ix);
        func = StrFromArray(triplet,0);
        val = IntFromArray(triplet,1);
        name = StrFromArray(triplet,2);
        XmMenuStatus((AXWidget_main)w,val,func);
        if (*name)
            XmMenuItemName((AXWidget_main)w,func,name);
    }
}

/*************************************************************************
function: ElfbUpdateDialog - Update dialog box title and icon
*/
void ElfbUpdateDialog(dbox, w_dataP, doExpress)
data_ptr dbox;  /* elf array */
data_ptr w_dataP;
data_ptr doExpress;
{
    char *title, *iconName;
    char *icon_title;
    AXWidget_main id_widget;
    AXWidget_dialog dlg;

    id_widget = (AXWidget_main)WidgetPointerFromHandleData( w_dataP );
    if ( id_widget == (AXWidget_main)NULL )
    {
       /* something is out of whack, but let's be safe */
       return;
    }

    dlg = XmDlgInWidget(id_widget);

    if (AxIntFromArray(dbox, DB_in_edit)) {
        deCopyDialog(dlg, dbox);
    }

    /* update dialog title */
    title = StrFromArray(dbox, DB_title);

    /* update set-aside title */
    icon_title = StrFromArray(dbox, DB_icon_title);

    if ( icon_title[0] )
        XmNewTitle(id_widget, title, icon_title);
    else
        XmNewTitle(id_widget, title, title);

    /* setaside icon */
    iconName = (char *)StrFromArray(dbox, DB_icon);
    AXmUpdateSetasideIcon(id_widget, iconName);

    /* poke expressLine */
    if ( IntFromDataPtr(doExpress) )
        UpdateToolBarDisplay((AXWidget)id_widget);
}

void ElfbSetActiveWidget( w_dataP, pending_delete )
data_ptr w_dataP;   /* widget to make active */
int pending_delete;
{
    AXWidget w;
 
    w = WidgetPointerFromHandleData( w_dataP );

    SetActiveWidget (w,pending_delete);
    return;
}

void ElfbGetActiveWidgetFromElf( w_dataP )
data_ptr w_dataP;
{
    AXWidget wid, active_widget;
    AXWidget_handle w_handle;
 
    wid = WidgetPointerFromHandleData( w_dataP );

    active_widget = (AXWidget)ElfbGetActiveWidget( wid );
    w_handle = HANDLE_FROM_WIDGET( active_widget );
    ElfRetData = AxMakeBinaryData(sizeof(w_handle), (char*)&w_handle);
    return;
}

int
ElfbGetDlgFontHeight()
{
    extern int AxCharHeight;
    return((int) AxCharHeight);
}

/*************************************************************************
function: ElfXCeboxFocusChanged - callback to notify cursor in new edit box
*/
ElfXCeboxFocusChanged (widget, pid)
AXWidget widget;
int pid;
{
    if ( !pid ) /* no task id, ignore event */
        return;
    if ( !widget )
        return;
    CallWidget(pid,widget,EV_FOCUS_IN_,NULL);
}


/*************************************************************************
function: ElfXCfieldChanged - Callback to notify changed to a new entry field
*/
ElfXCfieldChanged (widget,pid)
AXWidget widget;
int pid;
{
    if ( !pid ) /* no task id, ignore event */
        return;
    if ( !widget )
        return;
    CallWidget(pid,widget,EV_FOCUS_IN_,NULL);
}


/***********************************************************************
function: ElfXentryButtonPress - Report a (single) button press in the
entry box.  Used solely to clear up the "pending_delete" flag in the
elf layer.
*/
void
ElfXentryButtonPress(widget, pid, data)
AXWidget widget;
int pid;
{
    if (!pid)
        return;
    CallWidget(pid, widget, EV_BUTTON_PRESS_, NULL);
}

/*************************************************************************
function: ElfXCchoicePushed - Callback for option menus
Unlike radio boxes, the menu option selected is passed to this routine
directly as the pid arg (as was setup in the buttons list).
*/
void ElfXCchoicePushed (widget,pid,data)
AXWidget widget;
int pid;
XmRowColumnCallbackStruct *data;
{
    if ( !pid ) /* no task id, ignore event */
        return;
    XWLOG(("******* CHOICE %x pid=%d",widget,data->data));
    CallWidget(pid,widget,EV_CHANGED_,AxMakeNumData((int)data->data));
}

/*************************************************************************
function: ElfXCscale - Callback for scale/slider
*/
void ElfXCscale (widget,pid,data)
AXWidget widget;
int pid;
XmScaleCallbackStruct *data;
{
    int reason;

    if ( !pid ) /* no task id, ignore event */
        return;
    switch ( data->reason ) {
        case XmCR_DRAG:
            /* motion notify */
            reason = EV_MOTION_;
            break;
        case XmCR_VALUE_CHANGED:
            /* button up */
            reason = EV_CHANGED_;
            break;
        case XmCR_BUTTON_PRESS:
            /* button down */
            reason = EV_BUTTON_PRESS_;
            break;
        default:
            return;
    }
    CallWidget(pid,widget,reason,AxMakeNumData(data->value));
}

static AXWidget_table
axMakeTableWidget(info, old_widget, parent, name, 
				  xpos, ypos, width, height,
				  helpstr, helpid, multi_pick, 
				  thickness, elfObject, wflags, ehList,
				  textFont, textShadow, textColor,
				  headingFont, headingShadow, headingColor,
				  widgetColor, uid)

data_ptr info;          /* elf array of info */
AXWidget old_widget;
AXWidget parent;
mbchar *name;
int xpos, ypos;
int width, height;
mbchar *helpstr;
int helpid;
int multi_pick;
int thickness;
data_ptr elfObject;
int wflags; /* see elf_dlg.elf */
eventHandlerList *ehList;
data_ptr textFont;
data_ptr textShadow;
data_ptr textColor;
data_ptr headingFont;
data_ptr headingShadow;
data_ptr headingColor;
data_ptr widgetColor;
int uid;
{
    Arg args[60];
    AXWidget_table table_w;
    int marker_width = 0;
    int allow_hidden_columns = 0;
    int allow_column_resizing = 0;
    int suppress_row_markers = 0;
    int allow_column_reordering = 0;
    int use_data_request_model = 0;
    int alis_style_scroll = 0;
    int rowcolcreate      = 0;
    int hide_vertical_grid = 0;
    int hide_horizontal_grid = 0;
    int vscroll_origin = 0;
    int vscroll_length = 0;
    int hscroll_origin = 0;
    int hscroll_length = 0;
    int show_row_numbers = 0;
    int button_up_scroll = 0;
    int visible_line_count = 0;
    int hscr;       /* BTRUE if horiz scroll bar wanted */
    int hide_vscr;
    int an = 0;
    static XtCallbackRec Callback[2];   /* standard template */
    int old_x, old_y, old_width, old_height;
    data_ptr graphic, rminfo, cursP;
    char *cursor_name;

    hscr = IntFromArray(info, DBDEXhoriz_scroll);
    hide_vscr = IntFromArray(info, DBDEXhide_vscroll);
    marker_width = IntFromArray(info, DBDEXmarker_width);
    allow_hidden_columns = IntFromArray(info, DBDEXallow_hidden_columns);
    allow_column_resizing = IntFromArray(info, DBDEXallow_column_resizing);
    suppress_row_markers = IntFromArray(info, DBDEXsuppress_row_markers);
    allow_column_reordering = IntFromArray(info, DBDEXallow_column_reordering);
    use_data_request_model = IntFromArray(info, DBDEXuse_data_request_model);
    alis_style_scroll = IntFromArray(info, DBDEXalis_style_scroll);
    rowcolcreate = IntFromArray(info, DBDEXrowcolcreate);
    hide_vertical_grid = IntFromArray(info, DBDEXhide_vertical_grid);
    hide_horizontal_grid = IntFromArray(info, DBDEXhide_horizontal_grid);
    vscroll_origin = IntFromArray(info, DBDEXvscroll_origin);
    vscroll_length = IntFromArray(info, DBDEXvscroll_length);
    hscroll_origin = IntFromArray(info, DBDEXhscroll_origin);
    hscroll_length = IntFromArray(info, DBDEXhscroll_length);
    show_row_numbers = IntFromArray(info, DBDEXshow_row_numbers);
    button_up_scroll = IntFromArray(info, DBDEXbutton_up_scroll);
    visible_line_count = IntFromArray(info, DBDEXvisible_item_count);
    graphic = AnyFromArray(info, DBDEXgraphic);
    rminfo = AnyFromArray(info, DBDEXrminfo);
    cursP = AnyFromArray(info, DBDEXdefaultCursor);

    XtSetArg(args[an], XmNoldWidget, old_widget); an++;
    XtSetArg(args[an], XmNflags, wflags); an++;
    XtSetArg(args[an], XmNuid, uid); an++;
    XtSetArg(args[an], AXmNmarkerWidth, marker_width); an++;
    XtSetArg(args[an], AXmNallowHiddenColumns, allow_hidden_columns); an++;
    XtSetArg(args[an], AXmNallowColumnResizing, allow_column_resizing); an++;
    XtSetArg(args[an], AXmNsuppressRowMarkers, suppress_row_markers); an++;
    XtSetArg(args[an],AXmNallowColumnReordering, allow_column_reordering);an++;
    XtSetArg(args[an], AXmNuseDataRequestModel, use_data_request_model); an++;
    XtSetArg(args[an], AXmNalisStyleScroll, alis_style_scroll); an++;
    XtSetArg(args[an], AXmNhideVerticalGrid, hide_vertical_grid); an++;
    XtSetArg(args[an], AXmNhideHorizontalGrid, hide_horizontal_grid); an++;
    XtSetArg(args[an], AXmNvscrollOrigin, vscroll_origin); an++;
    XtSetArg(args[an], AXmNvscrollLength, vscroll_length); an++;
    XtSetArg(args[an], AXmNhscrollOrigin, hscroll_origin); an++;
    XtSetArg(args[an], AXmNhscrollLength, hscroll_length); an++;
    XtSetArg(args[an], AXmNshowRowNumbers, show_row_numbers); an++;
    XtSetArg(args[an], AXmNbuttonUpScroll, button_up_scroll); an++;
    XtSetArg(args[an], AXmNRowColCreate, rowcolcreate); an++;

    if (textColor) {
        XtSetArg(args[an], AXmNtextColor, textColor); an++;
    }
    if (widgetColor) {
        XtSetArg(args[an], AXmNwidgetColor, widgetColor); an++;
    }
    if (headingColor) {
        XtSetArg(args[an], AXmNlabelColor, headingColor); an++;
    }
    XtSetArg(args[an], AXmNlabelFont, headingFont); an++;
    /* for table, set NaxFont only after NlabelFont
     */
    XtSetArg(args[an], XmNaxFont, textFont); an++;
    
    /* set text shadow props only after setting font props
     */
    XtSetArg(args[an], AXmNtextShadow, textShadow); an++;
    XtSetArg(args[an], AXmNlabelShadow, headingShadow); an++;
    if (rminfo) {
        XtSetArg(args[an], AXmNrminfo, rminfo); an++;
    }
    if (graphic) {
        XtSetArg(args[an], AXmNgraphic, graphic); an++;
    }
    if (cursP) {
        cursor_name = StrFromArray(info, DBDEXdefaultCursor);
        XtSetArg(args[an], XmNDefaultCursor, cursor_name); an++;
    }

    if (old_widget) {
        old_x = old_widget->xpos;
        old_y = old_widget->ypos;
        old_width = old_widget->width;
        old_height = old_widget->height;
        if ((old_x != xpos) || (old_y != ypos)
            || (old_width != width) || (old_height != height)) {
            XtSetArg(args[an], XmNresize, TRUE); an++;
        }
    }
    XtSetArg(args[an], AXmNelfObject, elfObject); an++;
    XtSetArg(args[an], XmNx, xpos); an++;
    XtSetArg(args[an], XmNy, ypos); an++;
    XtSetArg(args[an], XmNwidth, width); an++;
    XtSetArg(args[an], XmNheight, height); an++;
    if (AnyFromArray(info, DBDEXthickness)) {
        XtSetArg(args[an], XmNthickness, thickness); an++;
    }
    if (visible_line_count) {
        XtSetArg(args[an], XmNvisibleItemCount, visible_line_count); an++;
    }

    /* Ashisuto changed to set multi-pick un-conditionly to disable the attrs.
     * by using ELF-macro function: db_ctrl_multi_select@().    95.10.06
     */
    if (multi_pick) {
        XtSetArg(args[an], XmNmultipleSelectionCallback, TRUE); an++;
    } else {
        XtSetArg(args[an], XmNmultipleSelectionCallback, FALSE); an++;
    }

    if (helpstr) {
        XtSetArg(args[an], XmNhelpStr, helpstr); an++;
        XtSetArg(args[an], XmNmenuHelpWidget, helpid); an++;
    }
    if (AxUseWorkColors(info)) {
        XtSetArg(args[an], AXmNuseWorkColors, TRUE); an++;
    }
    if (!old_widget) { /* only captured when created... */
        if (hide_vscr) {
            XtSetArg(args[an], AXmNhideListVScroll, TRUE); an++;
        } else {
            XtSetArg(args[an], AXmNhideListVScroll, FALSE); an++;
        }
        if (hscr) {
            XtSetArg(args[an], AXmNlistHScroll, TRUE); an++;
        } else {
            XtSetArg(args[an], AXmNlistHScroll, FALSE); an++;
        }
        CALLBACKER(Callback, ElfXCtable, (int)THIMpid);
        XtSetArg(args[an], XmNactivateCallback, Callback); an++;
    }
    if (ehList->numProcs) {
        XtSetArg(args[an], ZXmNeventHandlers, ehList); an++;
    }
    if (IntFromArray(info, DBDEXminor_event_notify)) {
        XtSetArg(args[an], AXmNminorEventNotify, TRUE); an++;
    }

	XtSetArg(args[an],XmNaccelerator_key,
			 AxArrayElement(info, DBDEXaccelerator_key)); an++;
    table_w = XmCreateTable(parent, name, args, an);
    if (wflags & WFLG_INACTIVE)
        AxXtSetSensitive((AXWidget)table_w, FALSE);
    return (table_w);
}

/*************************************************************************
function: ElfXCtable - Recieve a callback
*/
void ElfXCtable (widget,pid,data)
AXWidget widget;
int pid;
TblEventData *data;
{
    register data_ptr info = (data_ptr) NULL;
    register int ix;
    int row_num;
    
    XWLOG(("******* TABLE %x",widget));

    if ( !pid ) /* no task id, ignore event */
        return;
    if (!data)
        return;

    switch (data->any.type) {
    case EV_TBL_DBLCLICK_:
        if (data->double_click.col == -1) {
            info = AxMakeNumData(data->double_click.row);
        } else {
            info = ELF_CREATE_ARRAY(2);
            info = WriteArray(info, 0, AxMakeNumData(data->double_click.row));
            info = WriteArray(info, 1, AxMakeNumData(data->double_click.col));
        }
        break;

    case EV_TBL_SELECTION_CHANGE_:
        info = ELF_CREATE_ARRAY(data->selection_change.num_rows);
        for (ix = 0; ix < data->selection_change.num_rows; ix++) {
            row_num = data->selection_change.row_list[ix];
            info = WriteArray(info, ix, AxMakeNumData(row_num));
        }
        break;

    case EV_TBL_COLUMN_RESIZE_:
        info = ELF_CREATE_ARRAY(2);
        info = WriteArray(info, 0,
                          AxMakeNumData(data->column_resize.column));
        info = WriteArray(info, 1,
                          AxMakeNumData(data->column_resize.width));
        break;
    case EV_TBL_REQUEST_DATA_:
        info = ELF_CREATE_ARRAY(2);
        info = WriteArray(info, 0,
                          AxMakeNumData(data->request_data.starting_row));
        info = WriteArray(info, 1,
                          AxMakeNumData(data->request_data.num_rows));
        break;
    case EV_TBL_KEYPRESS_:
        info = ELF_CREATE_ARRAY(5);
        info = WriteArray(info, 0,
                          AxMakeNumData(data->keypress.elfkeyid));
        if (data->keypress.printable)
            info = WriteArray(info, 1, AxMakeNumData(BTRUE));
        else
            info = WriteArray(info, 1, AxMakeNumData(FALSE));
        info = WriteArray(info, 2,
                          AxMakeWideStrData(data->keypress.num_chars,
                                            data->keypress.wbuf));
        info = WriteArray(info, 3, AxMakeNumData(data->keypress.row));
        info = WriteArray(info, 4, AxMakeNumData(data->keypress.col));
        break;
    case EV_TBL_BUTTONPRESS_:
        info = ELF_CREATE_ARRAY(4);
        info = WriteArray(info, 0,
                          AxMakeNumData(data->buttonpress.row));
        info = WriteArray(info, 1,
                          AxMakeNumData(data->buttonpress.col));
        info = WriteArray(info, 2,
                          AxMakeNumData(data->buttonpress.start_char_pos));
        info = WriteArray(info, 3,
                          AxMakeNumData(data->buttonpress.end_char_pos));
        break;
    case EV_TBL_FOCUS_CHANGE_:
        info = ELF_CREATE_ARRAY(4);
        info = WriteArray(info, 0,
                          AxMakeNumData(data->focus_change.from_row));
        info = WriteArray(info, 1,
                          AxMakeNumData(data->focus_change.from_col));
        info = WriteArray(info, 2,
                          AxMakeNumData(data->focus_change.to_row));
        info = WriteArray(info, 3,
                          AxMakeNumData(data->focus_change.to_col));
        break;
    case EV_TBL_EDIT_BEGIN_:
        return; /* TODO */
    case EV_TBL_ROWCOLCREATE_:
        info = ELF_CREATE_ARRAY(4);
        info = WriteArray(info, 0,
                          AxMakeNumData(data->rowcolcreate.type));
        info = WriteArray(info, 1,
                          AxMakeNumData(data->rowcolcreate.totalCol));
        info = WriteArray(info, 2,
                          AxMakeNumData(data->rowcolcreate.totalRow));
        info = WriteArray(info, 3,
                          AxMakeNumData(data->rowcolcreate.delete));
        break;
    case EV_TYPING_:
        info = ELF_CREATE_ARRAY(2);
        info = WriteArray(info, 0, AxMakeNumData(data->cell_typing.row));
        info = WriteArray(info, 1, AxMakeNumData(data->cell_typing.col));
        break;
    default:
        tempAbort(NXLT("internal table error"));
        /*NOTREACHED*/
    }

    CallWidget(pid, widget, data->any.type, info);
    return;
}

/***********************************************************************
function: ElfbAllowTableEditing - Allow cell editing if flag is BTRUE,
Shut off cell editing otherwise.
*/
void ElfbAllowTableEditing(wid_dataP, flag)
data_ptr wid_dataP;
int flag;
{
    extern void AxXmAllowTableEditing();
    AXWidget wid;
 
    wid = WidgetPointerFromHandleData( wid_dataP );
 
    AxXmAllowTableEditing((AXWidget_table)wid, flag);
    return;
}

/***********************************************************************
function: ElfbSetTableSelections - Select rows in a table
*/
void ElfbSetTableSelections(selections, wid_dataP)
data_ptr selections;
data_ptr wid_dataP;
{
    AXWidget wid;
 
    wid = WidgetPointerFromHandleData( wid_dataP );
    AxXmSetTableSelections((AXWidget_table)wid, selections);
    return;
}

/************************************************************************/
/*  function: ElfbGetTableSelections - Get elf array of ints containing */
/*  selected rows.                                                      */
/************************************************************************/
void ElfbGetTableSelections(wid_dataP)
data_ptr wid_dataP;
{
    AXWidget wid;
    extern data_ptr AxXmGetTableSelections();
 
    wid = WidgetPointerFromHandleData( wid_dataP );
    ElfRetData = AxXmGetTableSelections((AXWidget_table)wid);
    return;
}

/***********************************************************************
function: ElfbGetTableData - Return the current contents of the table.
*/
void ElfbGetTableData(w_dataP)
data_ptr w_dataP;
{
    AXWidget wid;

    wid = WidgetPointerFromHandleData( w_dataP );
    ElfRetData = AxXmGetTableData((AXWidget_table)wid);
    return;
}

/***********************************************************************
function: ElfbDataRequestModel - Use data request model if flag is
BTRUE; do not use data request model otherwise.
*/
void
ElfbDataRequestModel(wid_dataP, flag)
data_ptr wid_dataP;
int flag;
{
    extern void AxXmDataRequestModel();
    AXWidget wid;

    wid = WidgetPointerFromHandleData(wid_dataP);
    AxXmDataRequestModel((AXWidget_table)wid, flag);
    return;
}

/*************************************************************************
function: ElfbInsEditBoxData - Add to contents of edit box
*/
void ElfbInsEditBoxData (w_id, passed_strs)
data_ptr  w_id;
data_ptr passed_strs;  /* elf array of strings */
{
    AXWidget wid;

    wid = WidgetPointerFromHandleData( w_id );
    InsEditBoxData(wid, passed_strs);
}

void InsEditBoxData (wid, passed_strs)
AXWidget wid;
data_ptr passed_strs;  /* elf array of strings */
{
    int ix, arraysize;
    int datasize;
    int bytes_copied;
    char *data = NULL;
    char **strs;  /* array of pointers to passed strings */
    int *len;     /* parallel array of string lengths */
    int num_strs, this_len;

    /*
     * Data is passed down as an array of strings, where each
     * element is symbolically separated by a newline.  While
     * assembling one long string for the editbox, ensure that
     * only a max of MAXTEXTMEM bytes are assembled.  Also, don't
     * add a newline to the end of the last element.
     */

    arraysize = ElfbArraySize(passed_strs);

    if ( arraysize > 0 ) {
        /* store string pointers and lengths locally:
         * better than using StrFromArray and strlen
         * everytime they're needed */
        strs = (char **) calloc(arraysize, sizeof(char *));
        len = (int *) calloc(arraysize, sizeof(int));
    }
    datasize = 0;
    num_strs = arraysize;
    /* get total size of passed data */
    for (ix = 0; ix < arraysize; ix++) {
        strs[ix] = StrFromArray(passed_strs, ix);
#if 0
        *** the old way - fixed 9/21/94 jon ****
        this_len = strlen(strs[ix]);
        if ( (datasize + this_len + 1) >= MAXTEXTMEM ) {
            /* passed data larger than max - ignore rest */
            len[ix] = MAXTEXTMEM - datasize;  /* no NL for last str */
            datasize += len[ix];
            num_strs = ix + 1;
            break;
        }
        len[ix] = this_len;
#else
        /* strings will get expanded to wide chars in edit-box code */
        /* and MAXTEXTMEM is a byte count, not a char count         */
        len[ix] = strlen(strs[ix]);         /* length in bytes */
#if BUILD_WIDE
        this_len = MbCharLen( strs[ix] ) * sizeof(wchar);
#else
        this_len = len[ix];
#endif
        if ( (datasize + this_len + 1) >= MAXTEXTMEM )
            {
            /* passed data larger than max - ignore this line and rest */
            this_len = 0;
            len[ix] = 0;
            num_strs = ix;
            break;
            }
#endif
        datasize += len[ix];
        if ( ix < (arraysize - 1) )
            datasize++;   /* add 1 for newline */
    }
    if ( datasize > 0 ) {
        /* malloc for size of data + number of newlines */
        data = (char *) malloc(datasize + 1); /* add null */
        /* copy data from passed array */
        bytes_copied = 0;
        for (ix = 0; ix < num_strs; ix++) {
            memcpy(data+bytes_copied, strs[ix], len[ix]);
            bytes_copied += len[ix];
            if ( ix != (arraysize - 1) )
                data[bytes_copied++] = '\n';  /* no NL for last str */
        }
        data[bytes_copied] = '\0';
    }
    if ( arraysize > 0 ) {
        free(strs);
        free(len);
    }

    /* insert data in editbox */
    AxXmInsEditBoxData(wid, data);

    if (data)
        free(data);
}

    /* some constants describing the default dialog font axsab-16 */
#define AXSAB_PIXHT  16 /* font height in screenpix */
#define AXSAB_AVGWID 65 /* average screen pixwid in decipix */
#define AXSAB_DIGWID  7 /* pixel width of digit in axsab-16 */

/*
 * : AxScaleDialogLayout - return bool indicating whether
 * to scale size of the dbox and the size/pos's of its
 * widgets.  such scaling is done when the dialog main font
 * is different than the dbox's 'reference' font
 */
static int    /* return whether to scale */
AxScaleDialogLayout(referenceFont, xnum, xdenom, ynum, ydenom)
XFontStruct *referenceFont;          /* in */
int *xnum, *xdenom, *ynum, *ydenom;  /* out: return scale factor */
{
    int refwid, refht, dwid, dht;
    
    if (referenceFont) {
            /* this is a font struct whose '0' width and
             * overall height match that of the AxDlgMainFont
             * that was in effect at the time the dbox was last edited
             */
        refwid = AXTextWidth(referenceFont, "0", 1);
#if BUILD_ORIENT
        /* Ashisuto add 2 lines to avoid Celluar font bounds a real font.
         * Because for orient build, we have forced the ->fid non-zero.
         * 1997.02.08 */
        if (referenceFont->min_bounds.width == referenceFont->max_bounds.width)

            refwid = referenceFont->max_bounds.width;
#endif

        refht = referenceFont->max_bounds.ascent +
            referenceFont->max_bounds.descent;
    }
    else {
        refwid = AXSAB_DIGWID;
        refht = AXSAB_PIXHT;
    }
    
    if (AxAlternateDialogFont) {
        dwid = AXTextWidth(AxDlgMainFont, "0", 1);
        dht = AxDlgMainFont->max_bounds.ascent +
            AxDlgMainFont->max_bounds.descent;
    }
    else {
        dwid = AXSAB_DIGWID;
        dht = AXSAB_PIXHT;
    }

#if AXMS
    if ( dwid == refwid )
       {
       /* the NT systemfont "MS Sans Serif" is the same width as the default */
       /* UNIX dialog font, but it is a little narrower.  In this case,      */
       /* suppress any vertical scaling                                      */
       dht = refht;
       }
    if ( RunningWin95Look )
        {
        /* HACK: just to get dialog box geometries to work on Win95 */
        if ( dwid == refwid - 1 )
            dht = refht;
        }
    if ( dht == refht )
       {
       /* the height is the same, so treat this as the same size */
       dwid = refwid;
       }
#endif

    *xnum = dwid;
    *xdenom = refwid;
    *ynum = dht;
    *ydenom = refht;
    
    if ((refwid == dwid) && (refht == dht)) {
        return FALSE;
    }
    else {
        return TRUE;
    }
}

/*
 * : AxScaleDialogFrame - do a dialog-layout scaling of a
 * dbox's height,width,min/max_height/width if appropriate
 */
void
AxScaleDialogFrame(dbox, wid, ht, minwid, minht, maxwid, maxht)
data_ptr dbox;            /* in: elf array in format dialog_box_ */
int *wid, *ht;            /* out */
int *minwid, *minht;      /* out */
int *maxwid, *maxht;      /* out */
{
    int different, itmp;
    XFontStruct *rfs;
    
    if (!dbox) {
        return;
    }

    *wid = AxMtOlVal(AxArrayElement(dbox, DB_width));
    *ht = AxMtOlVal(AxArrayElement(dbox, DB_height));
    *minwid = AxMtOlVal(AxArrayElement(dbox, DB_min_width));
    *minht = AxMtOlVal(AxArrayElement(dbox, DB_min_height));
    *maxwid = AxMtOlVal(AxArrayElement(dbox, DB_max_width));
    *maxht = AxMtOlVal(AxArrayElement(dbox, DB_max_height));
    
    if (AxIntFromArray(dbox, DB_no_expand)) {
        return;    /* don't scale regardless of reference font */
    }
    
    rfs = lookupReferenceFont(AxArrayElement(dbox, DB_reference_font),
                              &different);
    if (different) {
        itmp = AxMtOlVal(AxArrayElement(dbox, DB_width));
        *wid = axDlgScaleX(itmp, rfs);
        
        itmp = AxMtOlVal(AxArrayElement(dbox, DB_height));
        *ht = axDlgScaleY(itmp, rfs);
        
        itmp = AxMtOlVal(AxArrayElement(dbox, DB_max_width));
        *maxwid = axDlgScaleX(itmp, rfs);
        
        itmp = AxMtOlVal(AxArrayElement(dbox, DB_max_height));
        *maxht = axDlgScaleY(itmp, rfs);
        
        itmp = AxMtOlVal(AxArrayElement(dbox, DB_min_width));
        *minwid = axDlgScaleX(itmp, rfs);
        
        itmp = AxMtOlVal(AxArrayElement(dbox, DB_min_height));
        *minht = axDlgScaleY(itmp, rfs);
    }
}

/************************
  function: lookupNamedFont - return an XFontStruct for the screenfont
  which corresponds to the passed name.  the passed name may be one of:
  
  <xlfd name>                             (string datum == xlfd name)
  <<att1 type><att1 val>><<att2 type><att2 val>>, ...  (attr list)
  
  the <xlfd name> form is old; better to do something like:
  
      db_ctrl_label_font@(dlg, "my button", "Courier")
          which will call this function using the newer, att-list form

*/
XFontStruct *
lookupNamedFont(nameData)
data_ptr nameData;
{
    mbchar *family;
    int pointsize, weight, slant, scalable;
    int cellWidth, cellAscent, cellDescent;
    XFontStruct *fs = (XFontStruct *) NULL;
    
    if (!nameData) {
        return ((XFontStruct *) NULL);
    }
    
    if (AxIsArray(nameData)) {
        data_ptr atmp, vtmp;
        int ix, asize;

            /* init to hardcoded defaults
             */
        namedFontDefaults(&family, &pointsize, &weight, &slant);
        
            /* init vars associated with possible CELLSIZE form
             */
        cellWidth = cellAscent = cellDescent = 0;

            /* read any explicit attr data
             */
        asize = AxArraySize(nameData);
        for (ix = 0; ix < asize; ix++) {
            atmp = AxArrayElement(nameData, ix);
                /* interpret as <attr type><attr val>
                 */
            switch(AxIntFromArray(atmp, 0)) {
            case FONT_FAMILY:
                family = AxStrPtrFromArray(atmp, 1);
                break;
            case FONT_SIZE:
                pointsize = AxIntFromArray(atmp, 1);
                break;
            case FONT_WEIGHT:
                weight = AxIntFromArray(atmp, 1);
                if (weight == -1)
                    weight = 1;
                break;
            case FONT_SLANT:
                slant = AxIntFromArray(atmp, 1);
                if (slant == -1)
                    slant = 1;
                break;
            case FONT_CELLSIZE:
                    /* <attr val> :: <<cell width><ascent><descent>>
                     */
                vtmp = AxArrayElement(atmp, 1);
                if (AxIsArray(vtmp)) {
                    cellWidth = AxIntFromArray(vtmp, 0);
                    cellAscent = AxIntFromArray(vtmp, 1);
                    cellDescent = AxIntFromArray(vtmp, 2);
                }
                break;
            }
        }
        if (cellWidth) {
            char *errString;
            char cellName[40];   /* enough for 3 number strings plus spaces */
                /* the name data specifies cell size to use as a
                   monospace metric; this representation is used
                   only for a dialog box's referenceFont, which is
                   used to scale dialog layout; this form takes
                   precedence over the family,ptsize,wt,slant form 
                 */
            sprintf(cellName, "%d %d %d", cellWidth, cellAscent, cellDescent);
            if (errString = axGetReferenceCellStruct(cellName, &fs)) {
                freeErrString(errString);
                return ((XFontStruct *) NULL);
            }
        }
        else {
            if (axtWellFormedXLFD(family)) {
                    /* treat 'family' as an XLFD name - old
                     */
                fontMetrics *fm;
                char *errString;
                if (errString = axGetScreenMetrics(family, 
                                                   (fontTabEnt *) NULL, 
                                                   &fm)) {
                    freeErrString(errString);
                    return((XFontStruct *) NULL);
                }
                fs = fm->xfont;
            }
            else {
                if (!getXFontStruct(family, pointsize, weight, slant,
                                    &scalable, &fs)) {
                    return((XFontStruct *) NULL);
                }
            }
        }
    }
    else if (AxIsString(nameData)) {
        family = (mbchar *) AxStrFromDataPtr(nameData);
        if (axtWellFormedXLFD(family)) {
                /* treat 'family' as an XLFD name - old
                 */
            fontMetrics *fm;
            char *errString;
            if (errString = axGetScreenMetrics(family, (fontTabEnt *) NULL,
                                               &fm)) {
                freeErrString(errString);
                return( (XFontStruct *)NULL );
            }
            fs = fm->xfont;
        }
    }
    return (fs);
}

/* hardcoded defaults for font attributes on widget text */
void
namedFontDefaults(family, pointsize, weight, slant)
mbchar **family;    /* out */
int *pointsize; /* out */
int *weight;        /* out */
int *slant;     /* out */
{
#if AX_X11
        /* this is not quite the same as our axsab-16 font..
         */
    *family = (mbchar *)NXLT("Helvetica");
#else
    *family = (mbchar *)NXLT("Arial");
#endif
    *pointsize = 14;
    *weight = 1;
    *slant = 0;
}

/*
 * : lookupReferenceFont - return an XFontStruct pointer represening
 * a dialog box's reference font (the wkit dialog main font in effect
 * when this box was last edited).  also return whether that reference
 * font is different in size than the current wkit dialog main font.
 */
XFontStruct *    /* this may be null; if nonnull, don't free */
lookupReferenceFont(ref_fontP, different)
data_ptr ref_fontP; /* in: reference font info */
int *different;     /* out */
{
    int dwid, dht;     /* size numbers for current dialog main font */
    int rwid, rht;     /* size numbers for this dbox's reference font */
    XFontStruct *rfs, *lookupNamedFont();

    dwid = AXTextWidth(AxDlgMainFont, "0", 1);
    dht = AxDlgMainFont->max_bounds.ascent +
        AxDlgMainFont->max_bounds.descent;
    rfs = lookupNamedFont(ref_fontP);
    if (rfs) {
        rwid = AXTextWidth(rfs, "0", 1);
#if BUILD_ORIENT
        /* Ashisuto add 2 lines to avoid Celluar font bounds a real font.
         * Because for orient build, we have forced the ->fid non-zero.
         * 1997.02.08 */
        if (rfs->min_bounds.width > rfs->max_bounds.width)
            rwid = rfs->max_bounds.width;
#endif
        rht = rfs->max_bounds.ascent + rfs->max_bounds.descent;
    }
    else {
        rwid = AXSAB_DIGWID;
        rht = AXSAB_PIXHT;
    }
    if (rwid == dwid && rht == dht) {
        *different = FALSE;
    }
    else {
        *different = TRUE;
    }
    return rfs;
}

/*************************************************************************
function: axDlgScaleX - Scale X value based on dialog layout scale
*/
int
axDlgScaleX (x, referenceFont)
int x;
XFontStruct *referenceFont;    /* dialog box's reference font */
{
    int xnum, xdenom, ynum, ydenom;
    
    if ( AxScaleDialogLayout(referenceFont, &xnum, &xdenom, &ynum, &ydenom) )
        return ((int) (((float) x * xnum) / xdenom));
    else
        return(x);
}

int
axDlgUnscaleX (big, referenceFont)
int big;                       /* de-scale this x value */
XFontStruct *referenceFont;    /* dialog box's reference font */
{
    int xnum, xdenom, ynum, ydenom;
    
    if ( AxScaleDialogLayout(referenceFont, &xnum, &xdenom, &ynum, &ydenom) ) {
        
            /* return value which would forward-scale to the same number
             */
        return ((int) ((((float) big * xdenom) + xnum - 1) / xnum));
    }
    else {
        return(big);
    }
}

/*************************************************************************
function: axDlgScaleY - Scale Y value based on dialog layout scale
*/
int
axDlgScaleY (y, referenceFont)
int y;
XFontStruct *referenceFont;
{
    int xnum, xdenom, ynum, ydenom;
    
    if ( AxScaleDialogLayout(referenceFont, &xnum, &xdenom, &ynum, &ydenom) )
        return ((int) (((float) y * ynum) / ydenom));
    else
        return(y);
}

int
axDlgUnscaleY (big, referenceFont)
int big;                       /* de-scale this y value */
XFontStruct *referenceFont;
{
    int xnum, xdenom, ynum, ydenom;
    
    if ( AxScaleDialogLayout(referenceFont, &xnum, &xdenom, &ynum, &ydenom) ) {
        
            /* return value which would forward-scale to the same number
             */
        return ((int) ((((float) big * ydenom) + ynum - 1) / ynum));
    }
    else {
        return(big);
    }
}

/*************************************************************************
function: ElfbToggleToolBar - Display/hide dialog box tool bar
*/
void ElfbToggleToolBar (w_dataP)
data_ptr w_dataP;
{
    AXWidget wid;
 
    wid = WidgetPointerFromHandleData( w_dataP );
    AxXmToggleToolBar((AXWidget_main)wid);
}

/*************************************************************************
function: ElfbToolBarStatus - Return current display status of tool bar 
*/
int ElfbToolBarStatus (w_dataP)
data_ptr w_dataP;
{
    AXWidget wid;
    int ret;

    wid = WidgetPointerFromHandleData( w_dataP );
    ret = AxXmToolBarStatus((AXWidget_main)wid);
    return ( ret );
}

/*************************************************************************
function: ElfbRebuildDialog - Rebuild dialog after menubar poke
*/
ElfbRebuildDialog (dialog_dataP)
data_ptr dialog_dataP;
{
    AXWidget dialog;
 
    dialog = WidgetPointerFromHandleData( dialog_dataP );
    XmRestructureWindow((AXWidget_main)dialog);
}

/*************************************************************************
function: ElfXCmenuBarSelected - Return menu bar activity
*/
void ElfXCmenuBarSelected (widget,pid,mP)
AXWidget widget;
int pid;
menu_item_def *mP;
{
    if ( !pid ) /* no task id, ignore event */
        return;
    CallWidget(pid,widget,EV_CHANGED_,AxMakeStrData(-1,mP->word2));
}


int ElfbGetDpyWid()
{
    extern short DpyWid;    /* physical screen width */
    return( (int)DpyWid);
}
 
int ElfbGetDpyHt()
{
    extern short DpyHt; /* physical screen height */
    return( (int)DpyHt);
}

/*************************************************************************
function: ElfbGetEditBoxData - Return current contents of edit box
(Need to free the return val of EditBoxGetData)
*/
void ElfbGetEditBoxData (wid_dataP)
data_ptr wid_dataP;
{
    char *str;
    AXWidget wid;
    wid = WidgetPointerFromHandleData( wid_dataP );
    str = (char *)AxXmGetEditBoxData((AXWidget_editbox)wid);
    ElfReturnArrayOfStrings(str);
    if ( str )
        AXmFreeString(str);
}

void
ElfXRowColumnPulldown(widget, pid, data)
AXWidget widget;
int pid;
XmRowColumnCallbackStruct *data;
{
    if (!pid)
        return;
    CallWidget(pid, widget, EV_CHANGED_, AxMakeNumData((int)data->data));
    return;
}

/***********************************************************************
function: ElfXComboBoxPulldown - Callback for combo boxes.
*/
void
ElfXComboBoxPulldown(widget, pid, data)
AXWidget widget;
int pid;
XmRowColumnCallbackStruct *data;
{
    data_ptr value_data;
    AXWidget_combo_box cw = (AXWidget_combo_box) widget;
    
    if (!pid)
        return;

    if (data->reason) {
        /* data value is of string type; a "none of the above" value.
         * additional hack: callbackstruct member stores text length.
         */
        value_data = AxMakeStrData((int) data->callbackstruct, data->data);
    }
    else {
        /* data is of integer type; an index into fixed string array
         */
        value_data = AxMakeNumData((int)data->data);
    }
    CallWidget(pid, widget, EV_CHANGED_, value_data);
    
    return;
}

/***********************************************************************
function: ElfXTabControl - Callback for RECT_LAYERED-style W_RECT
*/
void
ElfXTabControl(widget, pid, data)
AXWidget widget;
int pid;
XmRowColumnCallbackStruct *data;
{
    if (!pid)
        return;
    CallWidget(pid, widget, EV_CHANGED_, AxMakeStrData(-1, data->data));
    return;
}

/*************************************************************************
function: ElfXCsplitter - Callback for splitters
*/
void ElfXCsplitter (widget,pid,data)
AXWidget widget;
int pid;
XmScaleCallbackStruct *data;
{
    if ( !pid ) /* no task id, ignore event */
        return;
    CallWidget(pid,widget,data->reason,AxMakeNumData(data->value));
}

void ebDoOptionCompositeLinks(option, buttons, numButtons)
AXWidget option, *buttons;
int numButtons;
{
    Arg args[50];
    int ix;
    for (ix = 0; ix < numButtons; ix++) {
        XtSetArg(args[0], ZXmNcompositeLinkDown, buttons[ix]);
        AxXtSetValues(option, args, 1);
        XtSetArg(args[0], ZXmNcompositeLinkUp, option);
        AxXtSetValues(buttons[ix], args, 1);
    }
}

void ebDoRadioCompositeLinks(box, label, buttons, numButtons)
AXWidget box, label, *buttons;
int numButtons;
{
    Arg args[50];
    int ix;
    if (label) {
        XtSetArg(args[0], ZXmNcompositeLinkDown, label);
        AxXtSetValues(box, args, 1);
        XtSetArg(args[0], ZXmNcompositeLinkUp, box);
        AxXtSetValues(label, args, 1);
    }
    for (ix = 0; ix < numButtons; ix++) {
        XtSetArg(args[0], ZXmNcompositeLinkDown, buttons[ix]);
        AxXtSetValues(box, args, 1);
        XtSetArg(args[0], ZXmNcompositeLinkUp, box);
        AxXtSetValues(buttons[ix], args, 1);
    }
}



/*************************************************************************
function: PositionGadget - In BigFont mode, position gadget, not label.
Used for entry fields and option menus (these gadgets have a problem because
their xpos is relative to the label, but they are usually aligned with
respect to the left edge of the gadget).
*/
static int
PositionGadget(parent, gadgetFont, str, xpos)
AXWidget_dialog parent;
data_ptr gadgetFont;
mbchar *str;
int xpos;
{
    int len;
    int gadget_xpos, new_xpos;
    int xnum, xdenom, ynum, ydenom;
    XFontStruct *refFont;
    
    if (gadgetFont)
        return (xpos);   /* scaleup doesn't change label size */
    
    if (!parent)
        return (xpos);   /* insufficient info */
    
    if (parent->noExpand ||
        !AxScaleDialogLayout(parent->referenceFont, &xnum, &xdenom,
                             &ynum, &ydenom)) {
        return (xpos);   /* no scaleup */
    }

    len = strlen(str);

        /* get gadget's position in the dialog's reference font */
    if (parent->referenceFont) {
        refFont = parent->referenceFont;
    }
    else {
            /* use the default reference font */
        refFont = AxDefaultReferenceFont;
    }
    gadget_xpos = xpos + AXTextWidth(refFont, str, len);

        /* position gadget xpos for big font */
    gadget_xpos = axDlgScaleX(gadget_xpos, refFont);

        /* return big font position relative to label */
    new_xpos = gadget_xpos - BasicStrlen(str);
    return(new_xpos);
}




/***********************************************************************
function: ElfXCtray - Receive a callback.
*/
void
ElfXCtray(widget, pid, data)
AXWidget widget;
int pid;
TrayEventData *data;
{
    data_ptr info = (data_ptr) NULL;

    XWLOG(("******* TRAY %x",widget));

    if (!pid) /* no task id, ignore event */
        return;
    if (!data)
        return;

    switch (data->any.type) {
    case EV_TRAY_DBLCLICK_:
        info = ELF_CREATE_ARRAY(ELF_TREE_NODE_MAX);
        info = WriteArray(info, ELF_TREE_NODE_title,
                         AxMakeStrData(-1,data->double_click.node_info.title));
        info = WriteArray(info, ELF_TREE_NODE_pixmap,
                      AxMakeStrData(-1,data->double_click.node_info.pixmap));
        info = WriteArray(info, ELF_TREE_NODE_value,
                          AxMakeNumData(data->double_click.node_info.value));
        info = WriteArray(info, ELF_TREE_NODE_handle,
                          AxMakeBinaryData(sizeof(long),
                                       (char*)&data->double_click.node_info.handle));
        info = WriteArray(info, ELF_TREE_NODE_subnodes, AxMakeNumData(0));
        break;
    case EV_TRAY_BUTTONPRESS_:
        info = ELF_CREATE_ARRAY(ELF_TREE_NODE_MAX);
        info = WriteArray(info, ELF_TREE_NODE_title,
                      AxMakeStrData(-1,data->button_press.node_info.title));
        info = WriteArray(info, ELF_TREE_NODE_pixmap,
                      AxMakeStrData(-1,data->button_press.node_info.pixmap));
        info = WriteArray(info, ELF_TREE_NODE_value,
                          AxMakeNumData(data->button_press.node_info.value));
        info = WriteArray(info, ELF_TREE_NODE_handle,
                          AxMakeBinaryData(sizeof(long),
                                       (char*)&data->button_press.node_info.handle));
        info = WriteArray(info, ELF_TREE_NODE_subnodes, AxMakeNumData(0));
        break;
    case EV_TRAY_SELECTION_CHANGE_:
        if (!data->selection_change.node_info_array) {
            info = (data_ptr) NULL;
            break;
        }
        info = ElfArrayOfTreeNodes(data->selection_change.node_info_array);
        break;
    default:
        return;
        /*NOTREACHED*/
        break;
    }

    CallWidget(pid, widget, data->any.type, info);
    return;
}

/***********************************************************************
function: ElfXCchooser - Receive a callback
*/
void
ElfXCchooser(widget, pid, data)
AXWidget_color_chooser widget;
int pid;
ChooserEventData *data;
{
    data_ptr info = NULL;
    int wcspace;
    wcolor_t color;
    data_ptr aP;
    int reason;

    XWLOG(("******* COLOR CHOOSER %x",widget));
    if (!widget)
        return;
    if (!pid) /* no task id, ignore event */
        return;
    if (!data)
        return;
    if (widget->widgetColor)
        wcspace = widget->widgetColor->any.type;
    else
        wcspace = WCSPACE_RGB;
    switch (wcspace) {
    case WCSPACE_RGB:
        color.any.type = wcspace;
        color.rgb.r = data->value.red;
        color.rgb.g = data->value.green;
        color.rgb.b = data->value.blue;
        break;
    case WCSPACE_HSB:
        color.any.type = wcspace;
        color.hsb.h = data->value.hue;
        color.hsb.s = data->value.saturation;
        color.hsb.b = data->value.brightness;
        break;
    default:
        return;
    }
    aP = AxCreateWColorData(&color);

    if (widget->widgetColor)
        AxFreeWColor(widget->widgetColor);
    widget->widgetColor = AxCreateWColor((AXWidget)widget, aP);

    reason = EV_CHANGED_;
    CallWidget(pid, (AXWidget)widget, reason, aP);
    return;
}

/*************************************************************************
function: ElfXCcanvas - Recieve a callback
*/
void ElfXCcanvas (widget,pid,data)
AXWidget widget;
int pid;
CanvasEventData *data;
{
    data_ptr pt1, pt2, aP;

    XWLOG(("******* CANVAS %x",widget));
    if ( !pid ) /* no task id, ignore event */
        return;
    switch ( data->any.type ) {
    case EV_EXPOSE_:
#if 0
        /* on Windows, we may be actively scrolling so we want to directly
         * call the expose event handler.          -donato
         */
        if ( widget->elfObject )
             AxWinDoCanvasExpose(widget, pid, data);
        else
#endif
        {
        aP = ELF_CREATE_ARRAY(3);
        pt1 = ELF_CREATE_ARRAY(2);
        pt2 = ELF_CREATE_ARRAY(2);
        ElfAddNumToArray(pt1,0,data->expose.x);
        ElfAddNumToArray(pt1,1,data->expose.y);
        ElfAddNumToArray(pt2,0,data->expose.x + data->expose.width);
        ElfAddNumToArray(pt2,1,data->expose.y + data->expose.height);
        WriteArray(aP,0,pt1);
        WriteArray(aP,1,pt2);
        ElfAddNumToArray(aP,2,data->expose.count);
        CallWidget(pid,widget,data->any.type,aP);
        }
        break;

    case EV_BUTTON_PRESS_:
    case EV_BUTTON_RELEASE_:
    case EV_MOTION_:
    case EV_DBLCLICK_:
        aP = ELF_CREATE_ARRAY(4);
        pt1 = ELF_CREATE_ARRAY(2);
        ElfAddNumToArray(pt1,0,data->button.x);
        ElfAddNumToArray(pt1,1,data->button.y);
        aP = WriteArray(aP,0,pt1);
        aP = ElfAddNumToArray(aP,1,data->button.shifted);
        aP = ElfAddNumToArray(aP,2,data->button.controlled);
        aP = ElfAddNumToArray(aP,3,data->button.which);
        CallWidget(pid,widget,data->any.type,aP);
        break;

    case EV_FOCUS_IN_:
    case EV_FOCUS_OUT_:
        CallWidget(pid,widget,data->any.type,NULL);
        break;
        
    case EV_NEW_SCROLL_POS_NOTIFY_:
#if 0
        /* on Windows, we may be actively scrolling so we want to directly
         * call the scroll event handler.          -donato
         */
        if ( widget->elfObject )
            AxWinDoCanvasScroll(widget, pid, data);
        else
#endif
        {
        aP = ELF_CREATE_ARRAY(2);
        ElfAddNumToArray(aP,0,data->scrolled.leftPoint);
        ElfAddNumToArray(aP,1,data->scrolled.topPoint);
        CallWidget(pid,widget,data->any.type,aP);
        }
        break;

    case EV_TYPING_:
        aP = ELF_CREATE_ARRAY(3);
        aP = ElfAddNumToArray(aP,0,data->key.keycode);
        aP = ElfAddNumToArray(aP,1,data->key.shifted);
        aP = ElfAddNumToArray(aP,2,data->key.controlled);
        CallWidget(pid,widget,data->key.type, aP);
        break;

    default:
        tempAbort(NXLT("internal canvas error"));
    }
}


/*************************************************************************
function: ElfXClistItem - Callback for lists
??1 The browse select stuff is probably gonna have a real bad effect
on command recording.  Need to think about this further.
*/
void ElfXClistItem (widget,pid,data)
AXWidget widget;
int pid;
XmListCallbackStruct *data;
{
    int mode;
    int pos;
    data_ptr info = (data_ptr) NULL;
    long *selections;
    int ix;
    int itm;
    int num_selections;
    char *drop_filename;

    if ( !pid ) /* no task id, ignore event */
        return;
    if (!data)
        return;

    switch (data->reason) {
    case AXmCR_SINGLE_SELECT:
    case AXmCR_BROWSE_SELECT:
    case AXmCR_DEFAULT_ACTION:
        if ((data->reason == AXmCR_SINGLE_SELECT) ||
            (data->reason == AXmCR_BROWSE_SELECT))
            mode = EV_CHANGED_;
        else
            mode = EV_DBLCLICK_;

        if (data->selected_items) { /* HACK!! */
            /* meaning the list box has multi-pick set to true.  For
               such list boxes, always return an array of selections. */
            selections = (long *) data->selected_items[0];
            num_selections = data->selected_item_count;
            if (num_selections != 1)
                return;
            info = ELF_CREATE_ARRAY(1);
            itm  = ReadIntsArray(selections, 0);
            info = WriteArray(info, 0, AxMakeNumData(itm));
        } else {
            pos = data->item_position;
            if (pos <= 0)
                return;
            info = AxMakeNumData(pos - 1);
        }
        break;

    case AXmCR_MULTIPLE_SELECT:
    case XmCR_EXTENDED_SELECT:
    case AXmCR_STROKE_SELECT:
        if (data->reason == AXmCR_MULTIPLE_SELECT)
            mode = EV_MULTI_SELECT_;
        else
            mode = EV_STROKE_SELECT_;
        pos = data->item_position;
        if (pos <= 0)
            return;
        selections = (long *) data->selected_items[0];
        num_selections = data->selected_item_count;
#if 0
        if (num_selections <= 0)
            return;
#endif
        /* fill info packet */
        info = ELF_CREATE_ARRAY(data->selected_item_count);
        for (ix = 0; ix < num_selections; ix++) {
            itm  = ReadIntsArray(selections, ix);
            info = WriteArray(info, ix, AxMakeNumData(itm));
        }
        break;

    case AXmCR_LIST_DROP:
        drop_filename = (char *) data->selected_items[0];
        if (!drop_filename)
            return;
        mode = EV_LIST_DROP_;
        /* fill info packet */
        info = ELF_CREATE_ARRAY(1);
        info = WriteArray(info, 0, AxMakeStrData(-1, drop_filename));
        break;

    default:
        return; /* ignore */
    }
    if (!info)
        return;
    XWLOG((NXLT("******* LIST %x pid=%d val=%d reason=%d"),
           widget,pid,pos,data->reason));
    CallWidget(pid,widget,mode,info);
    return;
}


/*************************************************************************
function: ElfXCbuttonPushed - Recieve a callback
*/
void ElfXCbuttonPushed (widget,pid,data)
AXWidget widget;
int pid;
XmAnyCallbackStruct *data;
{
    if ( !pid ) /* no task id, ignore event */
        return;

    XWLOG(("******* PUSHED %x",widget));
    CallWidget(pid,widget,data->reason,NULL);
}


/*************************************************************************
function: ElfXCtogglePushed - Callback for toggle buttons
*/
void ElfXCtogglePushed (widget,pid,data)
AXWidget widget;
int pid;
XmToggleButtonCallbackStruct *data;
{
    if ( !pid ) /* no task id, ignore event */
        return;
    XWLOG(("******* TOGGLE %x val=%d",widget,data->set));
    CallWidget(pid,widget,EV_CHANGED_,AxMakeNumData(data->set));
}

/*************************************************************************
function: ElfXCradioPushed - Callback for toggle buttons
*/
void ElfXCradioPushed (widget,pid,data)
AXWidget widget;
int pid;
XmRowColumnCallbackStruct *data;
{
    if ( !pid ) /* no task id, ignore event */
        return;
    if ( !data->widget )
        return;
    XWLOG(("******* RADIO %x val=%d",widget,data->data));
    CallWidget(pid,widget,EV_CHANGED_,AxMakeNumData((int)data->data));
}

/*************************************************************************
function: ElfXCentryChanged - Callback for entry box changes
*/
void ElfXCentryChanged (widget, pid)
AXWidget widget;
int pid;
{
    if ( !pid ) /* no task id, ignore event */
        return;
    CallWidget(pid,widget,EV_TYPING_,NULL);
}


/*************************************************************************
function: ElfbMenuCallback - Set elf callback for menubar
*/
void ElfbMenuCallback (widget_dataP,funcName)
data_ptr widget_dataP;
data_ptr funcName;
{
    AXWidget_main w;

    if ( !widget_dataP )
        return;
    w = (AXWidget_main)WidgetPointerFromHandleData( widget_dataP );
    w->menuCallback = ElfDataCopy(funcName);
}


/*************************************************************************
function: ElfbEntryValue - Return current string of entry field
??x Need to XtFree the return val of GetString!
*/
void ElfbEntryValue (w_dataP)
data_ptr w_dataP;
{
    AXWidget widget;

    if ( w_dataP->type == TYPE_ARRAY )
       {
       /* we were passed an array, use the first element */
       w_dataP = ((data_array *)w_dataP)->vals[0];
       }
       
    widget = WidgetPointerFromHandleData( w_dataP );
#if AXMS
    ElfStrRetter(WXGetWidgetString((AXWidget_entry)widget));
#else
    ElfStrRetter(GetWidgetString((AXWidget_entry)widget));
#endif
}


/*************************************************************************
function: ElfbCreateChangeWidget - Create/Change a widget
                returns value in "ElfRetData"
*/
void ElfbCreateChangeWidget (wtypeP,oldP,parentP,info)
data_ptr wtypeP;        /* type of widget */
data_ptr oldP;      /* current widget id or 0 */
                    /* Note this can be either a simple widget, */
                    /* or an array for some widget types */
data_ptr parentP;   /* parent widget */
data_ptr info;      /* elf array of info */
{
    int wtype;
    int style, size, ix;
    AXWidget_dialog parent;
    AXWidget_option w_option;
	AXWidget editbox, box;
    AXWidget button, *buttons;
    AXWidget label, rect, icon, rowcol, combo;
    AXWidget old_widget, old_label, stext;
    AXWidget retvalP, current;
    AXWidget_handle w_handle;
    char *str, *name;
    int xpos, ypos; /* where to place the button*/
    data_ptr xposP, yposP, widthP, heightP, thicknessP;
    int wflags; /* see elf_dlg.elf */
    int width;      /* optional width */
    int itmp, jtmp;
    int widest_width, pulldown_width, barwid, barspc, total_width;
    int old_x, old_y, old_size;
    int is_default;
    int isComboBox;
    int exists;
    int type;       /* 0-normal, 1-ok, 2-cancel, 3-pixmap */
                    /* 0 same as 2 */
    int height;
    int multi_pick;
    data_ptr listval;       /* item(s) selected is either int or array */
    data_ptr strs;  /* list of strings for buttons */
    data_ptr pixmaps; /* list of pixmap names for rowcolumn */
    int dbl_pick;   /* BTRUE if doubel click performs OK action */
    int hscr;       /* BTRUE if horiz scroll bar wanted */
    int hide_vscr;
    int scroll_pos; /* which line at top of window */
    int thickness;
    int helpid = 0;
    char *helpstr = (char *) 0;
    int value;
    Arg args[80], old_args[5];
    int an = 0;
    int n = 0;
    int fg_color;
    int bg_color;
    data_ptr text_color;
    data_ptr widget_color, label_color;
    data_ptr graphic, rminfo, cursP;
    data_ptr font, label_font, tablayers;
    static XtCallbackRec Callback[2];   /* standard template */
    int orig_xpos, orig_width, orig_height;
    int minValue, maxValue;
    int no_label;
    int three_state_toggle_flag;
    int ge_palette_btn_type;
    char *alternatePixmap;  /* for bitmap buttons */
    eventHandlerList ehList;
    int num_buttons;
    char **pixmap_names  = (char **) NULL;
    char **pixmap_labels = (char **) NULL;
    data_ptr elfObject;
    int label_wid, label_ht;
    char *valstr, *cursor_name;
    char *valid_chars;
    int title_type;
    int length;
    data_ptr aP;
    data_ptr shadow, textShadow, labelShadow;
    XFontStruct *labelFs, *fs;
    XFontStruct *refFont = (XFontStruct *) NULL;
    int box_marker_value;
    int marker_width = 0;
    int allow_hidden_columns  = 0;
    int allow_column_resizing = 0;
    int suppress_row_markers = 0;
    int allow_column_reordering = 0;
    int use_data_request_model  = 0;
    int alis_style_scroll    = 0;
    int hide_vertical_grid   = 0;
    int hide_horizontal_grid = 0;
    int vscroll_origin = 0;
    int vscroll_length = 0;
    int hscroll_origin = 0;
    int hscroll_length = 0;
    int show_row_numbers = 0;
    int button_up_scroll = 0;
    int visible_line_count = 0;
    int no_prim_seln = 0;
    int gadget_uid = 0;
    int expand = FALSE;
    int dbox_rev = 0;
    int do_label;

    wtype = IntFromDataPtr( wtypeP );  /* gadget type, e.g., DLG_ENTER */
    
    old_widget = (AXWidget)NULL;
    if ((DLG_ENTER == wtype) || (DLG_RADIO_BOX == wtype)) {
        if (oldP) {
            /* for these types, oldP arrives as a two-element array
             * of TYPE_BINARY. unpack 0th here.
             */
            aP = ((data_array *) oldP)->vals[0];
            old_widget = WidgetPointerFromHandleData( aP );
        }
    }
    else if (oldP) {
        /* a single elf datum of TYPE_BINARY
         */
        old_widget = WidgetPointerFromHandleData( oldP );
    }
    
    if (old_widget && old_widget->destroyed)   /* touched dead memory */
        return;

    parent = (AXWidget_dialog)WidgetPointerFromHandleData( parentP );
    if (parent = XmDlgInWidget((AXWidget_main)parent)) {
        refFont = parent->referenceFont;
        expand = !parent->noExpand;
        dbox_rev = parent->revstamp;
    }
    wflags = IntFromArray(info,DBDEXflags);
    str = StrFromArray(info,DBDEXtitle);
    font = getFontData(info);
    label_font = getLabelFontData(info);
    shadow = AxArrayElement(info, DBDEXshadow);
    if (AxIsInt(shadow) && !AxIntFromDataPtr(shadow)) {
        shadow = (data_ptr) NULL;    /* convert #0 to null array */
    }
    textShadow = AxArrayElement(info, DBDEXtextShadow);
    labelShadow = AxArrayElement(info, DBDEXlabelShadow);
    widget_color = AxArrayElement(info, DBDEXwidgetColor);
    label_color = AxArrayElement(info, DBDEXlabelColor);
    text_color = AxArrayElement(info, DBDEXtextColor);
    xposP = AnyFromArray(info,DBDEXxpos);
    yposP = AnyFromArray(info,DBDEXypos);
    xpos = AxMtOlVal(xposP);
    ypos = AxMtOlVal(yposP);
    widthP = AnyFromArray(info,DBDEXwidth);
    heightP = AnyFromArray(info,DBDEXheight);
    width = AxMtOlVal(widthP);
    height = AxMtOlVal(heightP);
    is_default = IntFromArray(info,DBDEXdflt);
    type = IntFromArray(info,DBDEXtype);
    multi_pick = IntFromArray(info,DBDEXmulti_pick);
    no_prim_seln = IntFromArray(info, DBDEXno_primary_seln);
    gadget_uid = IntFromArray(info, DBDEXuid);
    helpstr = StrFromArray(info,DBDEXhelpid);
    if ( helpstr[0] ) {
        helpid = atoi(helpstr);
    }
    thicknessP = AnyFromArray(info,DBDEXthickness);
    thickness = AxMtOlVal(thicknessP);
    orig_xpos = xpos;
    orig_width = width;
    orig_height = height;
    if (expand) {
        xpos = axDlgScaleX(xpos, refFont);
        ypos = axDlgScaleY(ypos, refFont);
        width = axDlgScaleX(width, refFont);
        height = axDlgScaleY(height, refFont);
    }
    if (AxArrayElement(info, DBDEXalignment)) {
        /* adjust widget position based on alignment
         */
        data_ptr geomAtts;
        p32 alignmentVector;
        
        geomAtts = wGetGeomAtts(info);
        wAlignmentOffset(dbox_rev, refFont, !expand, wtype,
                         geomAtts, &alignmentVector);
        
        xpos += alignmentVector.x;
        ypos += alignmentVector.y;
        
        AxFreeData(geomAtts);
    }
    elfObject = AnyFromArray(info,DBDEXelfObject);
    readEventHandlerData((long)info, FALSE, &ehList);

#if AXMS
    if (DLG_COMBO_BOX == wtype) {
        /* On Windows, the combo box is the same as an option menu - donato
         */
        wtype = DLG_CHOICE;
        isComboBox = TRUE;  /* but there is a difference in sizing */
    }
    else {
        isComboBox = FALSE;
    }
#endif
    
    switch ( wtype ) {

    case DLG_SPLITTER:
        rminfo = AnyFromArray(info, DBDEXrminfo);
        value = IntFromArray(info,DBDEXvalue);
        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        if ( width > 0 ) {
            XtSetArg(args[an],XmNwidth,width); an++;
        }
        if (rminfo) {
            XtSetArg(args[an], AXmNrminfo, rminfo); an++;
        }
        XtSetArg(args[an],XmNmaxHeight,height); an++;
        XtSetArg(args[an],XmNvalue,value); an++;
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        if (!old_widget) {
            CALLBACKER(Callback,ElfXCsplitter,(int)THIMpid);
            XtSetArg(args[an],XmNactivateCallback,Callback); an++;
        }
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
        rect = (AXWidget) XmCreateSplitter((AXWidget)parent,str,args,an);
        retvalP = rect;
        break;

    case DLG_ROWCOL: {
        register int ix;
        int num_cols;
        data_ptr btn_fg_color_datum;
        data_ptr btn_bg_color_datum;
        int *fg_color_vals;
        int *bg_color_vals;
        int have_colors = FALSE;

        num_buttons = 0;
        num_cols = 0;
        pixmap_labels = (char **) NULL;
        pixmap_names  = (char **) NULL;
        btn_fg_color_datum = (data_ptr) NULL;
        btn_bg_color_datum = (data_ptr) NULL;
        fg_color_vals = (int *) NULL;
        bg_color_vals = (int *) NULL;

        value = IntFromArray(info, DBDEXvalue);
        num_cols = IntFromArray(info, DBDEXlength);
        style = IntFromArray(info, DBDEXstyle);
        ge_palette_btn_type = IntFromArray(info, DBDEXge_palette_btn_type);
        btn_fg_color_datum = AnyFromArray(info, DBDEXfg_color);
        btn_bg_color_datum = AnyFromArray(info, DBDEXbg_color);
        alternatePixmap = StrFromArray(info, DBDEXreverseBitmap);
        rminfo = AnyFromArray(info, DBDEXrminfo);

        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        if (!old_widget) {
            CALLBACKER(Callback, ElfXRowColumnPulldown, (int)THIMpid);
            XtSetArg(args[an],XmNactivateCallback,Callback); an++;
        }
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],XmNwidth,width); an++;
        XtSetArg(args[an],XmNheight,height); an++;
        XtSetArg(args[an], XmNvalue, value); an++;
        XtSetArg(args[an], XmNgePaletteButtonType, ge_palette_btn_type); an++;
        if (alternatePixmap && !alternatePixmap[0])
            alternatePixmap = (char *) NULL;
        XtSetArg(args[an],XmNalternatePixmap,alternatePixmap); an++;

        /* Ashisuto changed to set multi-pick un-conditionly for disable the
         * attrs. by using ELF-macro function: db_ctrl_multi_select@(). 95.10.06
         */
        if (multi_pick) {
            /* Meaning this is a text-only rowcol. */
            XtSetArg(args[an], XmNmultipleSelectionCallback, TRUE); an++;
        } else {
            XtSetArg(args[an], XmNmultipleSelectionCallback, FALSE); an++;
        }

        if (IsValidRowColumnStyle(style)) {
            XtSetArg(args[an], ZXmNstyle, style); an++;
        } else {
            XtSetArg(args[an], ZXmNstyle, 0); an++; /* RC_TYPE_SQUARE */
        }
        if (num_cols > 0) {
            XtSetArg(args[an], XmNorientation, num_cols); an++;
        }
        if (rminfo) {
            XtSetArg(args[an], AXmNrminfo, rminfo); an++;
        }

        pixmaps = AnyFromArray(info, DBDEXpixmaps); /* pixmap files */
        num_buttons = ElfbArraySize(pixmaps);
        if ((ElfbArraySize(btn_fg_color_datum) >= num_buttons)
            && (ElfbArraySize(btn_bg_color_datum) >= num_buttons))
            have_colors = TRUE;

        XtSetArg(args[an], XmNitemCount, num_buttons); an++;
        if (pixmaps && num_buttons) {
            /* File names for Pixmapper() */
            pixmap_names = (char **) calloc(1, num_buttons * sizeof(char *));
            for (ix = 0; ix < num_buttons; ix++) {
                char *pixmap_name;

                pixmap_name = StrFromArray(pixmaps,ix);
                pixmap_names[ix] = (char *) malloc(strlen(pixmap_name) + 1);
                strcpy(pixmap_names[ix], pixmap_name);
            }
        }

        strs = AnyFromArray(info,DBDEXstrs); /* pixmap labels */
        if (strs && num_buttons) {
            pixmap_labels = (char **) calloc(1, num_buttons * sizeof(char *));
            for (ix = 0; ix < num_buttons; ix++) {
                char *label_string;

                label_string = StrFromArray(strs, ix);
                pixmap_labels[ix] = (char *) malloc(strlen(label_string) + 1);
                strcpy(pixmap_labels[ix], label_string);
            }
        }

        if (pixmap_names) {
            XtSetArg(args[an], AXmNbtnPixmapList, (char **)pixmap_names); an++;
        }

        if (pixmap_labels) {
            XtSetArg(args[an], XmNitems, (char **) pixmap_labels); an++;
        }
        /* form color vals. */
        if (have_colors && num_buttons) {
            fg_color_vals = (int *) calloc(1, num_buttons * sizeof(int));
            for (ix = 0; ix < num_buttons; ix++) {
                data_ptr this_btn_color;
                int red, green, blue;

                this_btn_color = AnyFromArray(btn_fg_color_datum, ix);
                red   = IntFromArray(this_btn_color, 0);
                green = IntFromArray(this_btn_color, 1);
                blue  = IntFromArray(this_btn_color, 2);

                if ((red == -1) && (green == -1) && (blue == -1)) {
                    /* special hack for the WP */
                    fg_color_vals[ix] = 0;
                } else {
                    fg_color_vals[ix]  = 0x01000000;
                    fg_color_vals[ix] |= 0x000000ff & (red << 0);
                    fg_color_vals[ix] |= 0x0000ff00 & (green << 8);
                    fg_color_vals[ix] |= 0x00ff0000 & (blue << 16);
                }
            }

            bg_color_vals = (int *) calloc(1, num_buttons * sizeof(int));
            for (ix = 0; ix < num_buttons; ix++) {
                data_ptr this_btn_color;
                int red, green, blue;

                this_btn_color = AnyFromArray(btn_bg_color_datum, ix);
                red   = IntFromArray(this_btn_color, 0);
                green = IntFromArray(this_btn_color, 1);
                blue  = IntFromArray(this_btn_color, 2);

                if ((red == -1) && (green == -1) && (blue == -1)) {
                    /* special hack for the WP */
                    bg_color_vals[ix] = 0;
                } else {
                    bg_color_vals[ix]  = 0x01000000;
                    bg_color_vals[ix] |= 0x000000ff & (red << 0);
                    bg_color_vals[ix] |= 0x0000ff00 & (green << 8);
                    bg_color_vals[ix] |= 0x00ff0000 & (blue << 16);
                }
            }
        }

        if (fg_color_vals) {
            XtSetArg(args[an], XmNfg_color, fg_color_vals); an++;
        }
        if (bg_color_vals) {
            XtSetArg(args[an], XmNbg_color, bg_color_vals); an++;
        }
        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        rowcol = (AXWidget)AXmCreateRowColumn((AXWidget)parent, str, args, an);
        retvalP = rowcol;

        if (pixmap_names) {
            for (ix = 0; ix < num_buttons; ix++)
                if (pixmap_names[ix])
                    free(pixmap_names[ix]);
            free(pixmap_names);
        }
        if (pixmap_labels) {
            for (ix = 0; ix < num_buttons; ix++)
                if (pixmap_labels[ix])
                    free(pixmap_labels[ix]);
            free(pixmap_labels);
        }
        if (fg_color_vals)
            free(fg_color_vals);
        if (bg_color_vals)
            free(bg_color_vals);

        pixmap_names  = (char **) NULL;
        pixmap_labels = (char **) NULL;
        fg_color_vals = (int *) NULL;
        bg_color_vals = (int *) NULL;
        
        break;
    }

    case DLG_COMBO_BOX: {
        char **option_strs = NULL;
        char *name;
        int  ix;
		data_ptr valP;

        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        if (!old_widget) {
            CALLBACKER(Callback, ElfXComboBoxPulldown, (int)THIMpid);
            XtSetArg(args[an],XmNactivateCallback,Callback); an++;
        }
        XtSetArg(args[an], AXmNelfObject, elfObject); an++;
        XtSetArg(args[an], XmNuid, gadget_uid); an++;
        XtSetArg(args[an], XmNflags, wflags); an++;
        strs = AnyFromArray(info, DBDEXstrs);
        num_buttons = ElfbArraySize(strs);
		valP = AxArrayElement(info, DBDEXvalue);
        XtSetArg(args[an], XmNx, xpos); an++;
        XtSetArg(args[an], XmNy, ypos); an++;
        XtSetArg(args[an], XmNwidth, width); an++;
        XtSetArg(args[an], XmNheight, height); an++;
		/* set num_buttons before value, as value gets range-checked
		 */
        XtSetArg(args[an], XmNitemCount, num_buttons); an++;
        XtSetArg(args[an], XmNvalue, valP); an++;
        XtSetArg(args[an], XmNaxFont, font); an++;
        XtSetArg(args[an], AXmNtextShadow, textShadow); an++;
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        if (text_color) {
            XtSetArg(args[an], AXmNtextColor, text_color); an++;
        }
        if (AxUseWorkColors(info)) {
            XtSetArg(args[an], AXmNuseWorkColors, TRUE); an++;
        }
        if (strs && num_buttons) {
            option_strs = (char **) calloc(1, sizeof(char *) * num_buttons);
            for (ix = 0; ix < num_buttons; ix++) {
                name = StrFromArray(strs, ix);
                option_strs[ix] = strcpy(malloc(strlen(name) + 1), name);
            }
        }
        XtSetArg(args[an], XmNitems, option_strs); an++;
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        combo = (AXWidget) AXmCreateComboBox(parent, str, args, an);
        retvalP = combo;
        if (num_buttons && option_strs) {
            for (ix = 0; ix < num_buttons; ix++) {
                if (option_strs[ix])
                    free(option_strs[ix]);
                option_strs[ix] = NULL;
            }
            free(option_strs);
            option_strs = NULL;
        }
        break;
    }
    case DLG_SCALE:
        value = IntFromArray(info,DBDEXvalue);
        minValue = IntFromArray(info,DBDEXminValue);
        maxValue = IntFromArray(info,DBDEXmaxValue);
        no_label = IntFromArray(info,DBDEXno_label);
        rminfo = AnyFromArray(info, DBDEXrminfo);
        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],XmNwidth,width); an++;
        XtSetArg(args[an],XmNheight,height); an++;
        XtSetArg(args[an],XmNvalue,value); an++;
        XtSetArg(args[an],AXmNminValue,minValue); an++;
        XtSetArg(args[an],AXmNmaxValue,maxValue); an++;
        XtSetArg(args[an],AXmNno_label,no_label); an++;
        if (!old_widget) {
            CALLBACKER(Callback,ElfXCscale,(int)THIMpid);
            XtSetArg(args[an],XmNactivateCallback,Callback); an++;
        }
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        if (rminfo) {
            XtSetArg(args[an], AXmNrminfo, rminfo); an++;
        }
        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        rect = (AXWidget) XmCreateScale((AXWidget)parent,str,args,an);
        retvalP = rect;
        break;

    case DLG_RECT:
        style = IntFromArray(info, DBDEXstyle);
        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],XmNwidth,width); an++;
        XtSetArg(args[an],XmNheight,height); an++;
        XtSetArg(args[an],XmNthickness,thickness); an++;
        XtSetArg(args[an],AXmNshadow, shadow); an++;
        XtSetArg(args[an],ZXmNstyle, style); an++;
        if (!old_widget) {
            CALLBACKER(Callback, ElfXTabControl, (int)THIMpid);
            XtSetArg(args[an], XmNactivateCallback, Callback); an++;
        }
        XtSetArg(args[an], XmNtablayers,
                 AxArrayElement(info, DBDEXtablayers)); an++;
        XtSetArg(args[an],XmNaxFont, label_font); an++;
        XtSetArg(args[an],AXmNtextShadow, labelShadow); an++;
        graphic = AnyFromArray(info, DBDEXgraphic);
        rminfo = AnyFromArray(info, DBDEXrminfo);
        cursP = AnyFromArray(info, DBDEXdefaultCursor);
        if (rminfo) {
            XtSetArg(args[an], AXmNrminfo, rminfo); an++;
        }
        if (graphic) {
            XtSetArg(args[an], AXmNgraphic, graphic); an++;
        }
        if (cursP) {
            cursor_name = StrFromArray(info, DBDEXdefaultCursor);
            XtSetArg(args[an], XmNDefaultCursor, cursor_name); an++;
        }
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        if (label_color) {
            XtSetArg(args[an], AXmNtextColor, label_color); an++;
        }
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        rect = (AXWidget) XmCreateRect((AXWidget)parent,str,args,an);
        retvalP = rect;
        break;

    case DLG_TEXT:
        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],XmNaxFont, label_font); an++;
        XtSetArg(args[an],AXmNtextShadow, labelShadow); an++;
        XtSetArg(args[an],ZXmNstyle, IntFromArray(info, DBDEXstyle)); an++;
        if (label_color) {
            XtSetArg(args[an], AXmNtextColor, label_color); an++;
        }
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        graphic = AnyFromArray(info, DBDEXgraphic);
        rminfo = AnyFromArray(info, DBDEXrminfo);
        if (rminfo) {
            XtSetArg(args[an], AXmNrminfo, rminfo); an++;
        }
        if (graphic) {
            XtSetArg(args[an], AXmNgraphic, graphic); an++;
        }
        if ( width > 0 ) {
            XtSetArg(args[an],XmNwidth,width); an++;
        }
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
        
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        label = AXmCreateLabelGadget((AXWidget)parent,str,args,an);
        retvalP = label;
        if ( wflags & WFLG_INACTIVE )
            AxXtSetSensitive(label,FALSE);
        break;

    case DLG_ICON:
        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        /* for DLG_ICON, width & height are maximum permissible values */
        XtSetArg(args[an],XmNmaxWidth,width); an++;
        XtSetArg(args[an],XmNmaxHeight,height); an++;
        XtSetArg(args[an],XmNvalue,str); an++;
        fg_color = IntFromArray(info,DBDEXfg_color);
        if ( fg_color ) {
            XtSetArg(args[an],XmNfg_color,fg_color); an++;
        }
        bg_color = IntFromArray(info,DBDEXbg_color);
        if ( bg_color ) {
            XtSetArg(args[an],XmNbg_color,bg_color); an++;
        }
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
        Pixmapper(str);
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;  /* ?? */
        icon = ZXmCreateIconImage((AXWidget)parent,NXLT("xxx"),args,an);
        retvalP = icon;
        break;

    case DLG_VERB_BUTTON:
        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],XmNbuttonType,type); an++;
        XtSetArg(args[an],XmNaxFont, label_font); an++;
        XtSetArg(args[an],AXmNshadow, shadow); an++;
        XtSetArg(args[an],AXmNtextShadow, labelShadow); an++;
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        if (label_color) {
            XtSetArg(args[an], AXmNtextColor, label_color); an++;
        }
        fg_color = IntFromArray(info,DBDEXfg_color);
        bg_color = IntFromArray(info,DBDEXbg_color);
        if ( fg_color ) {
            XtSetArg(args[an],XmNfg_color,fg_color); an++;
        }
        if ( bg_color ) {
            XtSetArg(args[an],XmNbg_color,bg_color); an++;
        }
        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
        if (!old_widget) {
            CALLBACKER(Callback,ElfXCbuttonPushed,(int)THIMpid);
            XtSetArg(args[an],XmNactivateCallback,Callback); an++;
        }
        if ( thickness != 0 ) {
            XtSetArg(args[an],XmNthickness,thickness); an++;
        }
        if ( type == DLG_ICON_BUTTON ) {
            XtSetArg(args[an],XmNlabelPixmap,TRUE); an++;
        }
        if (label_font) {
            labelFs = lookupNamedFont(label_font);
        }
        else {
            labelFs = (XFontStruct *) NULL;
        }
        wVerbButtonSize(str, type == DLG_ICON_BUTTON, labelFs,
                        thickness, shadow, textShadow,
                        &width, &height);
        XtSetArg(args[an],XmNwidth, width); an++;
        XtSetArg(args[an],XmNheight, height); an++;

        /* Ashisuto changed to set multi-pick un-conditionly for disable the
         * attrs by using ELF-macro function: db_ctrl_multi_select@(). 95.10.06
         */
        if (multi_pick) {
            XtSetArg(args[an], XmNmultipleSelectionCallback, TRUE); an++;
        } else {
            XtSetArg(args[an], XmNmultipleSelectionCallback, FALSE); an++;
        }

        XtSetArg(args[an],XmNwhichButton,Button1); an++;

        /* set value of button: FALSE - up, BTRUE - pressed */
        if ( IntFromArray(info, DBDEXvalue) ) {
            XtSetArg(args[an], XmNvalue, TRUE); an++;
        }
        else {
            XtSetArg(args[an], XmNvalue, FALSE); an++;
        }

        alternatePixmap = StrFromArray(info, DBDEXreverseBitmap);
        if ( alternatePixmap && !alternatePixmap[0] )
            alternatePixmap = (char *)NULL;
        XtSetArg(args[an],XmNalternatePixmap,alternatePixmap); an++;
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
        if ( IntFromArray(info,DBDEXminor_event_notify) ) {
            XtSetArg(args[an], AXmNminorEventNotify, TRUE); an++;
        }
        graphic = AnyFromArray(info, DBDEXgraphic);
        rminfo = AnyFromArray(info, DBDEXrminfo);
        if (rminfo) {
            XtSetArg(args[an], AXmNrminfo, rminfo); an++;
        }
        if (graphic) {
            XtSetArg(args[an], AXmNgraphic, graphic); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        button =(AXWidget)AXmCreatePushButtonGadget((AXWidget)parent, str,
		                                            args, an);
        retvalP = button;

        if ( is_default ) {
            XtSetArg(args[0],XmNdefaultButton,button);
            AxXtSetValues((AXWidget)parent,args,1);
        }
        else if (parent->defaultButton == (AXWidget_pushbutton)button) {
            /* turn off only if this button is currently the default
             */
            XtSetArg(args[0],XmNdefaultButton, (AXWidget_pushbutton) NULL);
            AxXtSetValues((AXWidget)parent,args,1);
        }
        if ( type == DLG_CANCEL_BUTTON ) {
            XtSetArg(args[0],XmNcancelButton,(AXWidget_pushbutton)button);
            AxXtSetValues((AXWidget)parent,args,1);
        }
        if ( wflags & WFLG_INACTIVE )
            AxXtSetSensitive(button,FALSE);
        break;

    case DLG_TOGGLE_BUTTON:
        value = IntFromArray(info,DBDEXvalue);
        three_state_toggle_flag = IntFromArray(info,DBDEXthree_state_toggle);
        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],XmNaxFont, label_font); an++;
        XtSetArg(args[an],AXmNshadow, shadow); an++;
        XtSetArg(args[an],AXmNtextShadow, labelShadow); an++;
        if (three_state_toggle_flag) {
            XtSetArg(args[an],XmNorientation, 3); an++;
        } else {
            XtSetArg(args[an],XmNorientation, 0); an++;
        }
        if (label_color) {
            XtSetArg(args[an], AXmNtextColor, label_color); an++;
        }
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
        if ( value == -1 )
            value = 1;
        XtSetArg(args[an],AXmNset,value); an++;
        if (!old_widget) {
            CALLBACKER(Callback,ElfXCtogglePushed,(int)THIMpid);
            XtSetArg(args[an],XmNvalueChangedCallback,Callback); an++;
        }
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        button = (AXWidget)
            AXmCreateToggleButtonGadget((AXWidget)parent, str, args, an);
        retvalP = button;
        if ( wflags & WFLG_INACTIVE )
            AxXtSetSensitive(button,FALSE);
        break;

    case DLG_TABLE:
        button = (AXWidget)axMakeTableWidget(info,old_widget,(AXWidget)parent,
		                                     str, xpos, ypos, width, height,
                                             helpstr, helpid, multi_pick,
                                             thickness, elfObject, wflags,
                                             &ehList, font, textShadow,
                                             text_color, label_font,
                                             labelShadow, label_color,
                                             widget_color, gadget_uid);

        retvalP = button;
        break;

    case DLG_CANVAS:
        hscr = IntFromArray(info,DBDEXhoriz_scroll);
        hide_vscr = IntFromArray(info,DBDEXhide_vscroll);
        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        if (!old_widget) {
            CALLBACKER(Callback,ElfXCcanvas,(int)THIMpid);
            XtSetArg(args[an],XmNvalueChangedCallback,Callback); an++;
        }
        XtSetArg(args[an],XmNaxFont,font); an++;

        if ( old_widget ) {
            int old_x, old_y, old_width, old_height;
            Arg old_args[4];

            XtSetArg(old_args[0],XmNx,&old_x);
            XtSetArg(old_args[1],XmNy,&old_y);
            XtSetArg(old_args[2],XmNwidth,&old_width);
            XtSetArg(old_args[3],XmNheight,&old_height);
            AxXtGetValues(old_widget,old_args,4);
            if ( (old_x != xpos) || (old_y != ypos)
                || (old_width != width) || (old_height != height) ) {
                XtSetArg(args[an],XmNresize,TRUE); an++;
            }
        }
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],XmNwidth,width); an++;
        XtSetArg(args[an],XmNheight,height); an++;
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        graphic = AnyFromArray(info, DBDEXgraphic);
        rminfo = AnyFromArray(info, DBDEXrminfo);
        cursP = AnyFromArray(info, DBDEXdefaultCursor);
        if (rminfo) {
            XtSetArg(args[an], AXmNrminfo, rminfo); an++;
        }
        if (graphic) {
            XtSetArg(args[an], AXmNgraphic, graphic); an++;
        }
        if (cursP) {
            cursor_name = StrFromArray(info, DBDEXdefaultCursor);
            XtSetArg(args[an], XmNDefaultCursor, cursor_name); an++;
        }


        /* Ashisuto changed to set multi-pick un-conditionly for disable the
         * attrs by using ELF-macro function: db_ctrl_multi_select@(). 95.10.06
         */
        if (multi_pick) {
            XtSetArg(args[an], XmNmultipleSelectionCallback, TRUE); an++;
        } else {
            XtSetArg(args[an], XmNmultipleSelectionCallback, FALSE); an++;
        }

        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
        if ( AxUseWorkColors(info) ) {
            XtSetArg(args[an],AXmNuseWorkColors,TRUE); an++;
        }
        if ( AnyFromArray(info,DBDEXthickness) ) {
            XtSetArg(args[an],XmNthickness,thickness); an++;
        }
		if ( hide_vscr ) {
			XtSetArg(args[an],AXmNhideListVScroll,TRUE); an++;
		}
		else {
			XtSetArg(args[an],AXmNhideListVScroll,FALSE); an++;
		}
		if ( hscr ) {
			XtSetArg(args[an],AXmNlistHScroll,TRUE); an++;
		}
		else {
			XtSetArg(args[an],AXmNlistHScroll,FALSE); an++;
		}
        XtSetArg(args[an],XmNaxFont, font); an++;
        XtSetArg(args[an],AXmNshadow, shadow); an++;
        XtSetArg(args[an],AXmNtextShadow, textShadow); an++;
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        button = (AXWidget) XmCreateCanvas(parent,str,args,an); an++;
        retvalP = button;
        if ( wflags & WFLG_INACTIVE )
            AxXtSetSensitive(button,FALSE);
        break;

    case DLG_CHOOSE: /* Color Chooser Widget */
        hscr = IntFromArray(info,DBDEXhoriz_scroll);
        hide_vscr = IntFromArray(info,DBDEXhide_vscroll);
        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        if (!old_widget) {
            CALLBACKER(Callback,ElfXCchooser,(int)THIMpid);
            XtSetArg(args[an],XmNvalueChangedCallback,Callback); an++;
        }
        XtSetArg(args[an],XmNaxFont,font); an++;

        if ( old_widget ) {
            int old_x, old_y, old_width, old_height;
            Arg old_args[4];

            XtSetArg(old_args[0],XmNx,&old_x);
            XtSetArg(old_args[1],XmNy,&old_y);
            XtSetArg(old_args[2],XmNwidth,&old_width);
            XtSetArg(old_args[3],XmNheight,&old_height);
            AxXtGetValues(old_widget,old_args,4);
            if ( (old_x != xpos) || (old_y != ypos)
                || (old_width != width) || (old_height != height) ) {
                XtSetArg(args[an],XmNresize,TRUE); an++;
            }
        }
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        width  = wColorChooserWidth(IntFromArray(info,DBDEXthickness));
        height = wColorChooserHeight(IntFromArray(info,DBDEXthickness));
        XtSetArg(args[an],XmNwidth,width); an++;
        XtSetArg(args[an],XmNheight,height); an++;
        graphic = AnyFromArray(info, DBDEXgraphic);
        rminfo = AnyFromArray(info, DBDEXrminfo);
        if (rminfo) {
            XtSetArg(args[an], AXmNrminfo, rminfo); an++;
        }
        if (graphic) {
            XtSetArg(args[an], AXmNgraphic, graphic); an++;
        }

        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
        if ( AnyFromArray(info,DBDEXthickness) ) {
            XtSetArg(args[an],XmNthickness,thickness); an++;
        }
        if ( !old_widget ) {    /* only captured when created... */
            if ( hide_vscr ) {
                XtSetArg(args[an],AXmNhideListVScroll,TRUE); an++;
            }
            else {
                XtSetArg(args[an],AXmNhideListVScroll,FALSE); an++;
            }
            if ( hscr ) {
                XtSetArg(args[an],AXmNlistHScroll,TRUE); an++;
            }
            else {
                XtSetArg(args[an],AXmNlistHScroll,FALSE); an++;
            }
        }
        XtSetArg(args[an],XmNaxFont, font); an++;
        XtSetArg(args[an],AXmNshadow, shadow); an++;
        XtSetArg(args[an],AXmNtextShadow, textShadow); an++;
        /* This serves as the value on the widget. */
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        button = (AXWidget) AXmCreateColorChooser(parent,str,args,an); an++;
        retvalP = button;
        if ( wflags & WFLG_INACTIVE )
            AxXtSetSensitive(button,FALSE);
        break;

    case DLG_TRAY:
        hscr = IntFromArray(info,DBDEXhoriz_scroll);
        hide_vscr = IntFromArray(info,DBDEXhide_vscroll);
        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        if (!old_widget) {
            CALLBACKER(Callback,ElfXCtray,(int)THIMpid);
            XtSetArg(args[an],XmNvalueChangedCallback,Callback); an++;
        }
        XtSetArg(args[an],XmNaxFont,font); an++;

        if ( old_widget ) {
            int old_x, old_y, old_width, old_height;
            Arg old_args[4];

            XtSetArg(old_args[0],XmNx,&old_x);
            XtSetArg(old_args[1],XmNy,&old_y);
            XtSetArg(old_args[2],XmNwidth,&old_width);
            XtSetArg(old_args[3],XmNheight,&old_height);
            AxXtGetValues(old_widget,old_args,4);
            if ( (old_x != xpos) || (old_y != ypos)
                || (old_width != width) || (old_height != height) ) {
                XtSetArg(args[an],XmNresize,TRUE); an++;
            }
        }
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],XmNwidth,width); an++;
        XtSetArg(args[an],XmNheight,height); an++;

        /* Ashisuto changed to set multi-pick un-conditionly for disable the
         * attrs by using ELF-macro function: db_ctrl_multi_select@(). 95.10.06
         */
        if (multi_pick) {
            XtSetArg(args[an], XmNmultipleSelectionCallback, TRUE); an++;
        } else {
            XtSetArg(args[an], XmNmultipleSelectionCallback, FALSE); an++;
        }

        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
        if ( AxUseWorkColors(info) ) {
            XtSetArg(args[an],AXmNuseWorkColors,TRUE); an++;
        }
        if ( AnyFromArray(info,DBDEXthickness) ) {
            XtSetArg(args[an],XmNthickness,thickness); an++;
        }
        if ( !old_widget ) {    /* only captured when created... */
            if ( hide_vscr ) {
                XtSetArg(args[an],AXmNhideListVScroll,TRUE); an++;
            }
            else {
                XtSetArg(args[an],AXmNhideListVScroll,FALSE); an++;
            }
            if ( hscr ) {
                XtSetArg(args[an],AXmNlistHScroll,TRUE); an++;
            }
            else {
                XtSetArg(args[an],AXmNlistHScroll,FALSE); an++;
            }
        }
        
        XtSetArg(args[an],XmNaxFont, font); an++;
        XtSetArg(args[an],AXmNshadow, shadow); an++;
        XtSetArg(args[an],AXmNtextShadow, textShadow); an++;
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        graphic = AnyFromArray(info, DBDEXgraphic);
        rminfo = AnyFromArray(info, DBDEXrminfo);
        cursP = AnyFromArray(info, DBDEXdefaultCursor);
        if (rminfo) {
            XtSetArg(args[an], AXmNrminfo, rminfo); an++;
        }
        if (graphic) {
            XtSetArg(args[an], AXmNgraphic, graphic); an++;
        }
        if (cursP) {
            cursor_name = StrFromArray(info, DBDEXdefaultCursor);
            XtSetArg(args[an], XmNDefaultCursor, cursor_name); an++;
        }
        
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        button = (AXWidget) AXmCreateTray(parent,str,args,an); an++;
        retvalP = button;
        if ( wflags & WFLG_INACTIVE )
            AxXtSetSensitive(button,FALSE);
        break;

    case DLG_ENTER:
    {
        int itmp, enterPixHeight;
        
        valstr = StrFromArray(info,DBDEXvalue);
        title_type = IntFromArray(info,DBDEXno_title);
        if ( !ReadArray(AX(info),DBDEXlength) )
            length = 256;   /* if length not spec'd, use 256 chars */
        else
            length = IntFromArray(info,DBDEXlength);
        valid_chars = StrFromArray(info,DBDEXvalid_chars);
            /* create/update the label */
        label = (AXWidget)NULL;
        if (oldP) {
            aP = ((data_array *)oldP)->vals[1];
            label = WidgetPointerFromHandleData( aP );
        }
        if (str && (title_type == LABEL_LEFT) &&
            !AxIntFromArray(AxArrayElement(info, DBDEXalignment), 0) ) {
            
            /* make an attempt to preserve visual alignment of left
             * edge of entry field; DBDEXalignment has no horizontal value.
             */
            xpos = PositionGadget(parent, label_font, (mbchar*)str, orig_xpos);
        }
        an = 0;
        XtSetArg(args[an],XmNoldWidget, label); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],ZXmNstyle, title_type); an++;
        XtSetArg(args[an],XmNaxFont, label_font); an++;
        XtSetArg(args[an],AXmNtextShadow, labelShadow); an++;
        if (label_color) {
            XtSetArg(args[an], AXmNtextColor, label_color); an++;
        }
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        label = AXmCreateLabelGadget((AXWidget)parent,str,args,an);
        if ( wflags & WFLG_INACTIVE )
            AxXtSetSensitive(label,FALSE);
        
            /* get the offset from the label origin to
             * the enterable field origin
             */
        fs = lookupNamedFont(font);
        if (!fs) {
            fs = AxDlgMainFont;
        }
        labelFs = lookupNamedFont(label_font);
        if (!labelFs) {
            labelFs = AxDlgMainFont;
        }
        wEnterFieldOffset(title_type, (mbchar *) str, labelFs, fs,
                          orig_width, textShadow, &itmp, &jtmp);
        xpos += itmp;
        ypos += jtmp;           /* move to origin of field */
        
            /* create/update the enterable field */
        an = 0;
        old_widget = (AXWidget)NULL;
        if (oldP) {
            aP = ((data_array *)oldP)->vals[0];
            old_widget = WidgetPointerFromHandleData( aP );
        }

        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        if (!old_widget) {
            CALLBACKER(Callback,ElfXCentryChanged,(int)THIMpid);
            XtSetArg(args[an],XmNentryCallback,Callback); an++;
        }
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        
        graphic = AnyFromArray(info, DBDEXgraphic);
        if (graphic) {
            XtSetArg(args[an], AXmNgraphic, graphic); an++;
        }
        rminfo = AnyFromArray(info, DBDEXrminfo);
        if (rminfo) {
            XtSetArg(args[an], AXmNrminfo, rminfo); an++;
        }
        wEnterPixHeight(fs, textShadow, &itmp, &enterPixHeight);
        XtSetArg(args[an],XmNheight, enterPixHeight); an++;
        
        /* set font and textShadow properties before setting
         * Ncolumns, since the latter will be interpreted in
         * terms of text size
         */
        XtSetArg(args[an],XmNaxFont, font); an++;
        XtSetArg(args[an],AXmNshadow, shadow); an++;
        XtSetArg(args[an],AXmNtextShadow, textShadow); an++;
        XtSetArg(args[an],XmNcolumns,orig_width); an++;
        XtSetArg(args[an],XmNmaxLength,length); an++;
        
        if ( !valid_chars[0] )
            valid_chars = NULL;
        XtSetArg(args[an],XmNtextOptions,valid_chars); an++;
		if (AxArrayElement(info, DBDEXedit_mask)) {
			XtSetArg(args[an], XmNedit_mask,
					 AxArrayElement(info, DBDEXedit_mask)); an++;
		}
		/* supply the field text only after valid_chars, edit_mask
		 */
        XtSetArg(args[an],XmNvalue,valstr); an++;
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        if (text_color) {
            XtSetArg(args[an], AXmNtextColor, text_color); an++;
        }
        if (AxUseWorkColors(info)) {
            XtSetArg(args[an], AXmNuseWorkColors, TRUE); an++;
        }
        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        stext = (AXWidget) AXmCreateText((AXWidget)parent,str,args,an);
        retvalP = stext;
        if ( wflags & WFLG_INACTIVE )
            AxXtSetSensitive(stext,FALSE);
        else
            AxXtSetSensitive(stext, TRUE);
        if (label) {
            /* tell label and text that they're part
             * of the same composite
             */
            XtSetArg(args[0], ZXmNcompositeLinkUp, stext);
            AxXtSetValues(label, args, 1);
            XtSetArg(args[0], ZXmNcompositeLinkDown, label);
            AxXtSetValues(stext, args, 1);
        }
        break;
    }

    case DLG_LIST_BOX:
    {
        AXWidget box;
        char *name;
        int ix, size;
        char **lstrs;
        int pos, num_selected, regen_list = FALSE;
        int use_old_strs = FALSE;   /* assumption of failure */
        int old_x, old_y, old_width, old_height;
        int item, rev;
        Arg revarg[1];

        listval = AnyFromArray(info,DBDEXvalue);
        strs = AnyFromArray(info,DBDEXstrs);
        dbl_pick = TRUE;
        hscr = IntFromArray(info,DBDEXhoriz_scroll);
        hide_vscr = IntFromArray(info,DBDEXhide_vscroll);
        box_marker_value = IntFromArray(info, DBDEXbox_marker_value);
        button_up_scroll = IntFromArray(info, DBDEXbutton_up_scroll);
        text_color = AnyFromArray(info, DBDEXtextColor);
        rminfo = AnyFromArray(info, DBDEXrminfo);

        if ( WFLG_RESET_LIST & wflags )
            regen_list = TRUE;

        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        
        /* NmonospaceFont is an order-dependent property;
         * it influences calculation of w->maxLen, and should be set early
         */
        if ( wflags & WFLG_MONOSPACE_FONT ) {
            XtSetArg(args[an],AXmNmonospaceFont,TRUE); an++;
        } else {
            XtSetArg(args[an], AXmNmonospaceFont, FALSE); an++;
        }
        
        XtSetArg(args[an], AXmNcurPos, box_marker_value); an++;
        XtSetArg(args[an], AXmNbuttonUpScroll, button_up_scroll); an++;
        if ( AxUseWorkColors(info) ) {
            XtSetArg(args[an],AXmNuseWorkColors,TRUE); an++;
        }
        if (rminfo) {
            XtSetArg(args[an], AXmNrminfo, rminfo); an++;
        }
        size = ElfbArraySize(strs);

/*
    On existing widgets, compare new strs list against old_widget.
    If they are the same, suppress sending strs list to widget again.
*/
        if ( old_widget ) {
            int old_count;
            char **old_strs;
            Arg old_args[7];

            XtSetArg(old_args[0],XmNitemCount,&old_count);
            XtSetArg(old_args[1],XmNitems,&old_strs);
            XtSetArg(old_args[2],XmNx,&old_x);
            XtSetArg(old_args[3],XmNy,&old_y);
            XtSetArg(old_args[4],XmNwidth,&old_width);
            XtSetArg(old_args[5],XmNvisibleItemCount,&old_height);
            AxXtGetValues(old_widget,old_args,6);
            if ( old_count == size ) {
                for ( ix = 0; ix < size; ix++ ) {
                    name = StrFromArray(strs,ix);
                    if ( !streq(name,old_strs[ix]) )
                        goto new_strs;
                }
                use_old_strs = TRUE;    /* if got here, they are the same */
            }
        }
new_strs:
        if ( !use_old_strs ) {
            lstrs = (char**)malloc(size * sizeof(char*));
            for ( ix = 0; ix < size; ix++ ) {
                name = StrFromArray(strs,ix);
                lstrs[ix] = strcpy(malloc(strlen(name)+1),name);
            }
        }
        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
        XtSetArg(args[an],AXmNlistSizePolicy,XmCONSTANT); an++;
        XtSetArg(args[an],AXmNscrollBarDisplayPolicy,XmAS_NEEDED); an++;
        XtSetArg(args[an],AXmNautomaticSelection,TRUE); an++;
        if (!old_widget) {
            CALLBACKER(Callback,ElfXClistItem,(int)THIMpid);
            XtSetArg(args[an],AXmNsingleSelectionCallback,Callback); an++;
            XtSetArg(args[an],XmNextendedSelectionCallback,Callback); an++;
        }
        if ( dbl_pick ) {   /* non-zero allows double click event */
            XtSetArg(args[an],AXmNdoubleClickInterval,1); an++;
        }
        if ( wflags & WFLG_DLG_NOT_PICKABLE ) {
            XtSetArg(args[an],AXmNno_picks,1); an++;
        }
        else {
            XtSetArg(args[an],AXmNno_picks,0); an++;
        }
        if ( regen_list ) {
            XtSetArg(args[an],XmNvalue,0); an++;
        }
        if ( multi_pick ) {
            XtSetArg(args[an],XmNmultipleSelectionCallback,TRUE); an++;
        }
        else {
            XtSetArg(args[an],XmNmultipleSelectionCallback,FALSE); an++;
        }

            /* scrollbar bools are set before Nwidth, Nheight (and also before
             * NvisibleItemCount) for the listbox since NlistHScroll can affect
             * height on AXMS
             */
        if ( hide_vscr ) {
            XtSetArg(args[an],AXmNhideListVScroll,TRUE); an++;
        }
        else {
            XtSetArg(args[an],AXmNhideListVScroll,FALSE); an++;
        }
        if ( hscr ) {
            XtSetArg(args[an],AXmNlistHScroll,TRUE); an++;
        }
        else {
            XtSetArg(args[an],AXmNlistHScroll,FALSE); an++;
        }

        /* set NaxFont before NvisibleItemCount, since the
         * latter depends on the value of the former
         */
        XtSetArg(args[an],XmNaxFont, font); an++;
        XtSetArg(args[an],XmNvisibleItemCount, orig_height); an++;

        if ( !use_old_strs ) {
            XtSetArg(args[an],XmNitemCount,size); an++;
            XtSetArg(args[an],XmNitems,lstrs); an++;
        }
        XtSetArg(revarg[0],XmNrevstamp,&rev);
        AxXtGetValues((AXWidget)parent,revarg,1);

        width = wListPixWidth(orig_width, rev, parent->referenceFont);

        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],XmNwidth,width); an++;
        XtSetArg(args[an],AXmNshadow, shadow); an++;
        XtSetArg(args[an],AXmNtextShadow, textShadow); an++;
        XtSetArg(args[an],AXmNelfAccess,TRUE); an++;
        if ( ElfbIsNull(listval) ) {
            if ( multi_pick ) {
                /* value for multi-select list boxes is null
                 * if nothing selected */
                XtSetArg(args[an],XmNselectedItemCount,0); an++;
            }
            else {
                /* for single select list boxes, null value
                 * equivalent to value=0 (1st item selected) */
                XtSetArg(args[an],XmNselectedItemCount,1); an++;
                XtSetArg(args[an],XmNselectedItems,0); an++;
            }
        }
        else if ( ElfbIsArray(listval) ) {
            /* value for multi-select list boxes is always array
             * if something selected */
            num_selected = ElfbArraySize(listval);
            XtSetArg(args[an],XmNselectedItemCount,num_selected); an++;
            XtSetArg(args[an],XmNselectedItems,listval); an++;
        }
        else if ( (0 <= (pos = IntFromDataPtr(listval))) && (pos < size) ) {
            XtSetArg(args[an],XmNselectedItemCount,1); an++;
            XtSetArg(args[an],XmNselectedItems,listval); an++;
        }
        else if ( regen_list ) {
            XtSetArg(args[an],XmNselectedItemCount,0); an++;
        }
        
        if ( IntFromArray(info,DBDEXforce_list_item_display) ) {
            XtSetArg(args[an],AXmNforce_list_item_display,TRUE); an++;
            item = IntFromArray(info,DBDEXlist_item_for_display);
            XtSetArg(args[an],XmNvalue,item); an++;  /* scroll_pos */
        }
        if (text_color) {
            XtSetArg(args[an], AXmNtextColor, text_color); an++;
        }
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        box = AXmCreateScrolledList((AXWidget)parent,str,args,an);
        retvalP = box;
        if ( wflags & WFLG_INACTIVE )
            AxXtSetSensitive(box,FALSE);
        if ( !use_old_strs ) {
            for ( ix = 0; ix < size; ix++ ) 
                free(lstrs[ix]);
            free(lstrs);
        }
        AxXtManageChild(box);
        break;
    }

    case DLG_EDIT_BOX:
        length = IntFromArray(info,DBDEXlength);
        XtSetArg(args[an],XmNoldWidget,old_widget); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        if ( wflags & WFLG_MONOSPACE_FONT ) {
            XtSetArg(args[an],AXmNmonospaceFont,TRUE); an++;
        }
        else {
            XtSetArg(args[an],AXmNmonospaceFont,FALSE); an++;
        }
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        XtSetArg(args[an],XmNwidth,width); an++;
        XtSetArg(args[an],XmNheight,height); an++;
        XtSetArg(args[an],XmNmaxLength,length); an++;
        XtSetArg(args[an], AXmNnoPrimarySeln, no_prim_seln); an++;
        if ( !old_widget ) {
            CALLBACKER(Callback,ElfXCentryChanged,(int)THIMpid);
            XtSetArg(args[an],XmNentryCallback,Callback); an++;
        }
        if ( AxUseWorkColors(info) ) {
            XtSetArg(args[an],AXmNuseWorkColors,TRUE); an++;
        }
        if ( wflags & WFLG_DLG_NOT_PICKABLE ) {
            XtSetArg(args[an],AXmNno_picks,1); an++;
        }
        else {
            XtSetArg(args[an],AXmNno_picks,0); an++;
        }
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        if (text_color) {
            XtSetArg(args[an], AXmNtextColor, text_color); an++;
        }
        XtSetArg(args[an],XmNaxFont, font); an++;
        XtSetArg(args[an],AXmNshadow, shadow); an++;
        XtSetArg(args[an],AXmNtextShadow, textShadow); an++;
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        editbox = (AXWidget) XmCreateEditBox((AXWidget)parent,str,args,an);
        retvalP = editbox;

        if ( IntFromArray(info,DBDEXval_changed) ) {
            strs = AnyFromArray(info,DBDEXvalue);
            if ( !strs || ElfbIsArray(strs) ) {
                data_ptr eb;
                AXWidget_handle eh;
                
                eh = HANDLE_FROM_WIDGET(editbox);
                eb = AxMakeBinaryData(sizeof(eh), (char*)&eh);
                if ( old_widget )
                    ElfbDelEditBoxData(eb);
                InsEditBoxData(editbox, strs);
                AxXmEditBoxResetEdited(eb);
            }
        }
		
		if (!old_widget) {
			aP = AnyFromArray(info, DBDEXselection);
			if (aP && strs && AxIsArray(strs)) {
				/* set initial cursor position using "start col" of
				 * db_editbox_selection info@
				 */
				char *stmp, seldex;
				AXWidget_editbox ebw = (AXWidget_editbox) editbox;
				for (itmp = 0, ix = 0; ix < AxArraySize(strs); ix++) {
					stmp = AxStrFromArray(strs, ix);
					if (stmp) {
						itmp += strlen(stmp);
					}
				}
				seldex = AxIntFromArray(aP, 1);  /* start col */
				if (seldex > itmp) {
					seldex = itmp;
				}
				ebw->editbox->sel.startIndex =
					ebw->editbox->sel.endIndex = seldex;
				ebw->editbox->tcursor.ac_index = seldex;
			}
		}
		
        rminfo = AnyFromArray(info, DBDEXrminfo);
        if (rminfo) {
            XtSetArg(args[0], AXmNrminfo, rminfo);
            AxXtSetValues(editbox, args, 1);
        }
        break;
        
    case DLG_CHOICE:
        strs = AnyFromArray(info, DBDEXstrs);
        value = IntFromArray(info, DBDEXvalue);
        XtSetArg(args[an], XmNoldWidget, old_widget); an++;
        XtSetArg(args[an], AXmNelfObject, elfObject); an++;
        XtSetArg(args[an], XmNuid, gadget_uid); an++;
        XtSetArg(args[an], XmNflags, wflags); an++;
        XtSetArg(args[an], XmNaxFont, label_font); an++;
        XtSetArg(args[an], AXmNshadow, shadow); an++;
        XtSetArg(args[an], AXmNtextShadow, labelShadow); an++;
        if ( !old_widget ) {
#if AXMS
			if ( isComboBox ) {
            	CALLBACKER(Callback, ElfXComboBoxPulldown, (int)THIMpid);
            }
			else
#endif
           	CALLBACKER(Callback, ElfXCchoicePushed, (int)THIMpid);
            XtSetArg(args[an], XmNentryCallback, Callback); an++;
        }
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
#if AXMS
    /* On Windows we need a separate label color on the option widget since
     * it's all one widget to Windows, as opposed to X where the choice
     * strings are separate buttons.        -donato
     */
        if (label_color) {
            XtSetArg(args[an], AXmNlabelColor, label_color); an++;
        }
        if (text_color) {
            XtSetArg(args[an], AXmNtextColor, text_color); an++;
        }
#else
        if (label_color) {
            XtSetArg(args[an], AXmNtextColor, label_color); an++;
        }
#endif
        if (graphic = AnyFromArray(info, DBDEXgraphic)) {
            XtSetArg(args[an], AXmNgraphic, graphic); an++;
        }
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
        if (helpstr) {
            XtSetArg(args[an], XmNhelpStr, helpstr); an++;
            XtSetArg(args[an], XmNmenuHelpWidget, helpid); an++;
        }
		XtSetArg(args[an],XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key)); an++;
        w_option = AXmCreatePulldownMenu((AXWidget)parent, str, args, an);
        retvalP = (AXWidget) w_option;
        
#if AXMS
{
#define STRALLOC(s) strcpy(malloc(strlen(s)+1),s)
        char **item_strs;
        int len, item_len, maxlen=0;
        static int option_pad_height;
		XFontStruct *fontP;

		if ( w_option->font )
			fontP = w_option->font;
		else
			fontP = AxDlgMainFont;

        size = ElfbArraySize(strs);
        item_strs = (char**)malloc(size * sizeof(char*));
        for ( ix = 0; ix < size; ix++ ) {
            item_strs[ix] = STRALLOC(StrFromArray(strs, ix));
            item_len = strlen(item_strs[ix]);
            len = AXTextWidth(fontP, item_strs[ix], item_len);
            if (len > maxlen)
                maxlen = len;
        }
        w_option->xpos = xpos;
        w_option->ypos = ypos;
        w_option->value = value;
        w_option->num_strs = w_option->num_buttons = size;
        if (isComboBox) {
            /* for combo boxes, always used the passed width. */
            if ( wflags & WFLG_EDITABLE )
                w_option->width = width - 8; /* account for space Windows puts
                                              * between edit area and pulldown
                                              * on editable combos.
                                              */
            else
                w_option->width = width;
            w_option->style = W_COMBO_BOX;
			w_option->maxLen = maxlen; /* so we can set the horizontal extent
										* to enable auto scrolling.
										*/
        }
        else if (maxlen == 0) {
            /* no strings, so set a default width. */
            w_option->width = ((Scrollbar_Thickness+8) * 3) + Option_Pad;
        }
        else {
            w_option->width = maxlen + Option_Pad;
        }
        /* compute height:  the AXmCreatePulldownMenu call will set the height
         * to the font height. multiply that by the number of strings in the
         * menu, and then add the font height times 2. this is to account for
         * the extra space needed by a Windows pulldown.
         */
        /* if number of items is greater than 10, let Windows provide the *
         * option menu scrollbar. Looks better, and is the "Windows way".
         */
        if ( size > 10 )
            size = 10;
        /* we need an "option pad height" to add to the pulldown height in 
         * order to account for the height of the option menu itself, that
         * is, in its non-pulled-down-state. On the new Win95 interface this
         * is the height of the pulldown-arrow bitmap, plus the pulldown button
         * borders(4), plus the option menu borders(4) (The extra 2 is to
         * account for the borders of the pulled-down menu). On the "old"
         * interface, we use the height of an entry field which is the same
         * as the height of a static option menu (plus 2 for the border).
         *                  -donato
         */
        if ( !option_pad_height ) {
            if ( RunningWin95Look )
                option_pad_height = MS_GetSystemMetrics(SM_CYVSCROLL) + 4+4+2;
            else
                option_pad_height = Entry_Height + 2;
        }
        w_option->pulldown_height = (w_option->height*size)+option_pad_height;
        SetOptionItems(w_option, item_strs); 
        for ( ix = 0; ix < size; ix++ )
			free(item_strs[ix]);
        free(item_strs);
}
#else
        /* create pulldown "buttons"; for already-existing
         * W_OPTIONS, AXmCreatePulldownMenu destroys them,
         * then they're re-created here -- this is because
         * they can change number/order
         */
        size = ElfbArraySize(strs);
        buttons = (AXWidget *) malloc (size * sizeof(AXWidget));
        for (current = (AXWidget) NULL, ix = 0; ix < size; ix++) {
            name = StrFromArray(strs, ix);
            an = 0;
            XtSetArg(args[an], XmNoldWidget, 0); an++;
            XtSetArg(args[an], XmNflags, wflags); an++;
            if (!old_widget) {
                CALLBACKER(Callback, ElfXCchoicePushed, ix);
                XtSetArg(args[an], XmNactivateCallback, Callback); an++;
            }
            XtSetArg(args[an], XmNaxFont, font); an++;
            XtSetArg(args[an], AXmNtextShadow, textShadow); an++;
            if (text_color) {
                XtSetArg(args[an], AXmNtextColor, text_color); an++;
            }
            if (ehList.numProcs) {
                XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
            }
            buttons[ix] = (AXWidget)
                AXmCreateToggleButtonGadget((AXWidget)w_option, name, args, an);
            if ( ix == value ) {
                current = buttons[ix];
            }
        }
    
        /* two properties can be set on the box only after the
         * buttons have been created: NmenuHistory, Nwidth
         */
        widest_width = widestOptionChoice(w_option);
        wOptionWidth(str, w_option->font, widest_width, &pulldown_width,
                     &barwid, &barspc, &total_width);
        if (old_widget && (w_option->width != total_width)) {
            /* the widget had already existed before entering this func,
             * and its width is now changing; unmap, and tag for repaint
             */
            UnMapWidget((AXWidget)w_option);
            w_option->update_widget = TRUE;
            WidgetNeedsUpdated = TRUE;
        }
        an = 0;
        XtSetArg(args[an], XmNmenuHistory, current); an++;
        XtSetArg(args[an], XmNwidth, total_width); an++;
        AxXtSetValues((AXWidget)w_option, args, an);
#endif
        
        if (wflags & WFLG_INACTIVE)
            AxXtSetSensitive((AXWidget)w_option,FALSE);

#if AX_X11
        /* tell optionmenu, buttons that they're all part of same composite */
        ebDoOptionCompositeLinks((AXWidget)w_option, buttons, size);
        free(buttons);
#endif
        break;
        
    case DLG_RADIO_BOX:
        label = (AXWidget) NULL;
        old_label = (AXWidget) NULL;
        do_label = TRUE;
        if (oldP) {
            aP = ((data_array *)oldP)->vals[1];
            old_label = WidgetPointerFromHandleData( aP );
        }
        if (old_label && old_label->destroyed) {
            do_label = FALSE;
        }
        value = IntFromArray(info, DBDEXvalue);
        strs = AnyFromArray(info, DBDEXstrs);
        if (do_label && str && *str) {
            XtSetArg(args[an],XmNoldWidget, old_label); an++;
            XtSetArg(args[an],XmNflags,wflags); an++;
            XtSetArg(args[an],XmNx,xpos); an++;
            XtSetArg(args[an],XmNy,ypos); an++;
            XtSetArg(args[an],XmNaxFont, label_font); an++;
            XtSetArg(args[an],AXmNshadow, shadow); an++;
            XtSetArg(args[an],AXmNtextShadow, labelShadow); an++;
            if (label_color) {
                XtSetArg(args[an], AXmNtextColor, label_color); an++;
            }
            if (graphic = AnyFromArray(info, DBDEXgraphic)) {
                XtSetArg(args[an], AXmNgraphic, graphic); an++;
            }
            if (ehList.numProcs) {
                XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
            }
            if ( helpstr ) {
                XtSetArg(args[an],XmNhelpStr,helpstr); an++;
                XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
            }
            label = AXmCreateLabelGadget((AXWidget)parent,str,args,an);
            if ( wflags & WFLG_INACTIVE )
                AxXtSetSensitive(label,FALSE);
            args[0].name = XmNheight;
            XtSetArg(args[0],XmNheight,&label_ht);
            AxXtGetValues(label,args,1);
            ypos += label_ht;
        }
        an = 0;
        size = ElfbArraySize(strs);
        XtSetArg(args[an],XmNoldWidget, old_widget); an++;
        XtSetArg(args[an],XmNflags,wflags); an++;
        if ( old_widget ) {
            XtSetArg(old_args[0],XmNx,&old_x);
            XtSetArg(old_args[1],XmNy,&old_y);
            XtSetArg(old_args[2],XmNvisibleItemCount,&old_size);
            AxXtGetValues(old_widget,old_args,3);
            if ( (old_x != xpos) || (old_y != ypos) || (old_size != size) ) {
                /* unmap radio box */
                XtSetArg(args[an],XmNupdate,size); an++;
            }
        }
        XtSetArg(args[an],AXmNelfObject,elfObject); an++;
        XtSetArg(args[an],XmNuid, gadget_uid); an++;
        XtSetArg(args[an],XmNx,xpos); an++;
        XtSetArg(args[an],XmNy,ypos); an++;
        if (!old_widget) {
            CALLBACKER(Callback,ElfXCradioPushed,(int)THIMpid);
            XtSetArg(args[an],XmNentryCallback,Callback); an++;
        }
        XtSetArg(args[an],XmNaxFont, label_font); an++;
        XtSetArg(args[an],AXmNtextShadow, labelShadow); an++;
        XtSetArg(args[an],AXmNshadow, shadow); an++;
        if (widget_color) {
            XtSetArg(args[an], AXmNwidgetColor, widget_color); an++;
        }
        if (ehList.numProcs) {
            XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
        }
        if ( helpstr ) {
            XtSetArg(args[an],XmNhelpStr,helpstr); an++;
            XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
        }
        retvalP = box = (AXWidget)AXmCreateRadioBox((AXWidget)parent,str,args,an);
#if AXMS
        box->winid = (HWND)-1;  /* this isn't a real widget. so give it a
                                 * non-null window handle so we can unmap
                                 * it when necessary .  -donato
                                 */
#endif
        buttons = (AXWidget *) malloc (size * sizeof(AXWidget));
        for ( ix = 0; ix < size; ix++ ) {
            name = StrFromArray(strs,ix);
            an = 0;
            XtSetArg(args[an],XmNoldWidget,old_widget ? ix+1 : 0); an++;
            XtSetArg(args[an],XmNflags,wflags); an++;
            XtSetArg(args[an],AXmNset,(value == ix ? TRUE : FALSE)); an++;
            if (!old_widget) {
                CALLBACKER(Callback,ElfXCradioPushed,ix);
                XtSetArg(args[an],XmNvalueChangedCallback,Callback); an++;
            }
            XtSetArg(args[an],XmNaxFont, font); an++;
            XtSetArg(args[an],AXmNshadow, shadow); an++;
            XtSetArg(args[an],AXmNtextShadow, textShadow); an++;
            if (text_color) {
                XtSetArg(args[an], AXmNtextColor, text_color); an++;
            }
            if (ehList.numProcs) {
                XtSetArg(args[an], ZXmNeventHandlers, &ehList); an++;
            }
            if ( helpstr ) {
                XtSetArg(args[an],XmNhelpStr,helpstr); an++;
                XtSetArg(args[an],XmNmenuHelpWidget,helpid); an++;
            }
            buttons[ix] = (AXWidget)
                AXmCreateToggleButtonGadget(box,name,args,an);
        }
        if ( wflags & WFLG_INACTIVE )
            AxXtSetSensitive(box,FALSE);

        /* tell label, box, buttons that they're
         * all part of the same composite
         */
        ebDoRadioCompositeLinks(box, label, buttons, size);
        free(buttons);

		/* apply accelerators only after adding composite links;
		 * the W_RADIO (box) holds the accelerator info, and applies
		 * it to the toggle buttons in SetVals.  The composite links
		 * allow the radio box to see its toggles.
		 */
		XtSetArg(args[0], XmNaccelerator_key,
				 AxArrayElement(info, DBDEXaccelerator_key));
		AxXtSetValues(box, args, 1);

        break;
    }

    tossEventHandlerData(&ehList);
    if (!old_widget) {
        appendNewWlayer(parent, retvalP); /* track back-to-front stacking */
    }
    if (wflags & WFLG_IN_EDIT) {
        deCopyGadget(retvalP, info);
    }
    
    /* finally, return the widget handle(s)
     */
    if (DLG_ENTER == wtype) {
#if AXMS
        AxWinAddWidgetToParent((AXWidget_handle) label, parent, is_default);
#endif
        aP = ELF_CREATE_ARRAY(2);
        w_handle = HANDLE_FROM_WIDGET(retvalP);
        aP = WriteArray(aP,0,AxMakeBinaryData(sizeof(w_handle),(char*)&w_handle));
        w_handle = HANDLE_FROM_WIDGET(label);
        aP = WriteArray(aP,1,AxMakeBinaryData(sizeof(w_handle),(char*)&w_handle));
        ElfRetData = aP;
    }
    else if (DLG_RADIO_BOX == wtype) {
#if AXMS
        AxWinAddWidgetToParent((AXWidget_handle) label, parent, is_default);
#endif
        aP = ELF_CREATE_ARRAY(2);
        w_handle = HANDLE_FROM_WIDGET(retvalP);
        aP = WriteArray(aP,0,AxMakeBinaryData(sizeof(w_handle),(char*)&w_handle));
        w_handle = HANDLE_FROM_WIDGET(label);
        aP = WriteArray(aP,1,AxMakeBinaryData(sizeof(w_handle),(char*)&w_handle));
        ElfRetData = aP;
    }
    else {
#if AXMS
        AxWinAddWidgetToParent((AXWidget_handle)retvalP, parent, is_default);
#endif
        w_handle = HANDLE_FROM_WIDGET(retvalP);
        ElfRetData = AxMakeBinaryData(sizeof(w_handle), (char*)&w_handle);
    }
    return;
}



/*************************************************************************
function: DialogPoked - Respond to an AxDialogPoke event
*/
static void DialogPoked (widget,pid,data)
AXWidget widget;
int pid;
XmRowColumnCallbackStruct *data;
{
    data_ptr aP;
    aP = ELF_CREATE_ARRAY(1);
    aP = WriteArray(aP, 0, AxMakeStrData(-1, data->data));
    SendElfData(pid, data->reason, aP);
    ELF_DATA_FREE(aP);
}

/*************************************************************************
function: ElfbCreateDialog - Fire up the box
          returns the widget handle in ElfRetData
*/
void ElfbCreateDialog (dbox)
data_ptr dbox;  /* elf array */
{
    char *name;
    char *background_image, *cursor_name;
    int width, height;
    int non_transient;
    int x, y;
    int icon;
    int cnt = 0;
    AXWidget dialog, parentWidget, dlgShell;
    Arg args[50];
    static XtCallbackRec Callback[2];   /* standard template */
    data_ptr pokeCodes;
    int rev, in_edit;
    data_ptr parent_dialog;
    int var_size, min_width, min_height, max_width, max_height;
    int no_record;
    int no_expand;
    int dbox_uid;
    eventHandlerList ehList;
    data_ptr elfObject, referenceFont, atmp, topP, cursP;
    AXWidget_handle w_handle;

    name = StrFromArray(dbox,DB_title);
    non_transient = IntFromArray(dbox,DB_non_transient);
    x = IntFromArray(dbox,DB_xpos);
    y = IntFromArray(dbox,DB_ypos);
    icon = IntFromArray(dbox,DB_icon);
    pokeCodes = AnyFromArray(dbox,DB_accept_poke);
    rev = IntFromArray(dbox,DB_revstamp);
    parent_dialog = AnyFromArray(dbox,DB_parent_id);
    var_size = IntFromArray(dbox,DB_var_size);
    no_record = IntFromArray(dbox,DB_disable_record);
    no_expand = IntFromArray(dbox,DB_no_expand);
    in_edit = IntFromArray(dbox,DB_in_edit);
    elfObject = AnyFromArray(dbox,DB_elfObject);
    referenceFont = AxArrayFromArray(dbox, DB_reference_font);
    atmp = AxArrayElement(dbox, DB_background_image);
    dbox_uid = IntFromArray(dbox, DB_uid);
    if (atmp) {
        background_image = StrFromArray(dbox, DB_background_image);
    }
    else {
        background_image = (char *) NULL;
    }

    XtSetArg(args[cnt],XmNresize,var_size); cnt++;
    XtSetArg(args[cnt],XmNautoUnmanage,FALSE); cnt++;
    XtSetArg(args[cnt],AXmNelfObject,elfObject); cnt++;
    XtSetArg(args[cnt],XmNx,x); cnt++;
    XtSetArg(args[cnt],XmNy,y); cnt++;
    XtSetArg(args[cnt],XmNpokeCodes,pokeCodes); cnt++;
    XtSetArg(args[cnt],AXmNIconId,icon); cnt++;
    XtSetArg(args[cnt],XmNrevstamp,rev); cnt++;
    CALLBACKER(Callback,DialogPoked,(int)THIMpid);
    XtSetArg(args[cnt],XmNvalueChangedCallback,Callback); cnt++;
    XtSetArg(args[cnt],ZXmNinEdit, in_edit); cnt++;
    XtSetArg(args[cnt],AXmNreferenceFont, referenceFont); cnt++;
    XtSetArg(args[cnt], AXmNbackgroundImage, background_image); cnt++;
    if ( dbox_uid ) {
        XtSetArg(args[cnt], XmNuid, dbox_uid); cnt++;
    }
    if ( no_expand ) {
        XtSetArg(args[cnt],XmNnoExpand,no_expand); cnt++;
    }
    topP = AxArrayElement(dbox, DB_topmost_window);
    if ( topP ) {
        XtSetArg(args[cnt], XmNtopmostwindow, AxIntFromDataPtr(topP)); cnt++;
    }
    cursP = AxArrayElement(dbox, DB_default_cursor);
    if ( cursP ) {
        cursor_name = StrFromArray(dbox, DB_default_cursor);
        XtSetArg(args[cnt], XmNDefaultCursor, cursor_name); cnt++;
    }
    AxScaleDialogFrame(dbox, &width, &height, &min_width,
                       &min_height, &max_width, &max_height);
    if ( width ) {
        XtSetArg(args[cnt],XmNwidth,width);
        cnt++;
    }
    if ( height ) {
        XtSetArg(args[cnt],XmNheight,height);
        cnt++;
    }
    if (var_size) {
        XtSetArg(args[cnt],XmNminWidth,min_width); cnt++;
        XtSetArg(args[cnt],XmNminHeight,min_height); cnt++;
        XtSetArg(args[cnt],XmNmaxWidth,max_width); cnt++;
        XtSetArg(args[cnt],XmNmaxHeight,max_height); cnt++;
    }

    parentWidget = WidgetPointerFromHandleData(parent_dialog);
    if ( !parentWidget && (non_transient || !TaskInfo[THIMpid].taskMaster) ) {
        XtSetArg(args[cnt],AXmNnotTransient,TRUE);
        cnt++;
        parentWidget = NULL;
    }
    else if (!parentWidget || !XmCanBeParent(parentWidget)) {
        int atmWid;
        atmWid = WindowIdFromPid(ThimPidFromUID(TaskInfo[THIMpid].taskMaster));
        if ( atmWid == -1 ) {
            parentWidget = (AXWidget)
                AXmMainWidgetOfTask(TaskInfo[THIMpid].taskMaster);
        }
        else {
            parentWidget = (AXWidget) WidgetOfWindow(atmWid);
        }
    }

    if ( no_record ) {
        XtSetArg(args[cnt],XmNdisableElfRecord,TRUE); cnt++;
    }
    else {
        XtSetArg(args[cnt],XmNdisableElfRecord,FALSE); cnt++;
    }

        /* see whether an event handler was specified for this dlg */
    readEventHandlerData((long)dbox, TRUE, &ehList);
    if (ehList.numProcs) {
        XtSetArg(args[cnt], ZXmNeventHandlers, &ehList); cnt++;
    }

    dlgShell = XmCreateDlgShell(parentWidget,name,name);
#if AXMS
    AxWinAddMenuBarToMain(dbox, (AXWidget_main)dlgShell);
#endif
    dialog = AXmCreateBulletinBoardDialog(dlgShell,name,args,cnt);

    tossEventHandlerData(&ehList);
    if (in_edit) {
        deCopyDialog((AXWidget_dialog)dialog, dbox);
    }

    w_handle = HANDLE_FROM_WIDGET( dlgShell );
    ElfRetData = AxMakeBinaryData(sizeof(w_handle), (char*)&w_handle);
    return;
}


/*************************************************************************
function: ElfbShowDialog - Map the dialog box
*/
ElfbShowDialog (dialog_dataP)
data_ptr dialog_dataP;
{
    AXWidget_main dialog;

    dialog = (AXWidget_main)WidgetPointerFromHandleData( dialog_dataP );
#if AXMS
    WXShowWindow(dialog);
#else
    AXtRealizeWidget(dialog);
#endif
    ThimIOSync();
}

extern int ThimExposuresMayBeComing;    /* Hint for command startup delaying */

/*************************************************************************
function: ElfbDialogDestroy - Shut down a dialog box
*/
ElfbDialogDestroy ( w_dataP )
data_ptr w_dataP;
{
    AXWidget id;        /*as returned by ElfbDialogCreate */

    id = WidgetPointerFromHandleData( w_dataP );
    AxXtDestroyWidget(id);
    ThimExposuresMayBeComing = TRUE;
    AxSleepSeconds(0);/* allow underlying windows to repaint */
}

/*************************************************************************
function: ElfbDestroyWidget - Destroy a widget in a dialog box
*/
void ElfbDestroyWidget ( w_dataP )
data_ptr w_dataP;
{
    AXWidget w; /*as returned by ElfbXXXCreate */

    if ( w_dataP == (data_ptr)NULL )
        return;

    if ( ElfbIsArray(w_dataP) ) {   /* only 2-object widgets supported! */
        w = WidgetPointerFromHandleData(ReadArray(AX(w_dataP),0));
        UnMapWidget(w);
        AxXtDestroyWidget(w);
        w = WidgetPointerFromHandleData(ReadArray(AX(w_dataP),1));
    }
    else
        w = WidgetPointerFromHandleData(w_dataP);
    UnMapWidget(w);
    AxXtDestroyWidget(w);
}

/*************************************************************************
function: ElfbDefaultButtons - Assign button widgets to RETURN/SHFT-RETURN keys
*/
void ElfbDefaultButtons (box_dataP, ok_dataP, cancel_dataP)
data_ptr box_dataP, ok_dataP, cancel_dataP;
{
    AXWidget box,ok,cancel;
    Arg args[2];
    int an = 0;

    if ( !ok_dataP )
        return;

    box = WidgetPointerFromHandleData( box_dataP );
    ok = WidgetPointerFromHandleData( ok_dataP );
    cancel = WidgetPointerFromHandleData( cancel_dataP );

    XtSetArg(args[an],XmNdefaultButton,ok); an++;
    if ( cancel ) {
        XtSetArg(args[an],XmNcancelButton,cancel); an++;
    }
    else
        AxXtSetValues(box,args,an);
    AxXtManageChild(box);
}

/*********************************************************************
 * function: getFontData - wrapper for reading font attribute data
 * from the gadget array.  DBDEXfont is the older attribute type,
 * and will supersede the newer if present
 */
data_ptr
getFontData(gadget)
data_ptr gadget;
{
    data_ptr f_str, f_att;
    
    f_str = AxArrayElement(gadget, DBDEXfont);     /* xlfd name */
    f_att = AxArrayElement(gadget, DBDEXfontAttr); /* family,size,weight,etc*/
    
    if (f_str)
        return (f_str);
    else
        return (f_att);
}

/*********************************************************************
 * function: getLabelFontData - wrapper for reading font attribute data
 * from the gadget array.  DBDEXfont is the older attribute type,
 * and will supersede the newer if present
 */
data_ptr
getLabelFontData(gadget)
data_ptr gadget;
{
    data_ptr f_str, f_att;
    
    f_str = AxArrayElement(gadget, DBDEXfont);     /* xlfd name */
    f_att = AxArrayElement(gadget, DBDEXlabelFont); /* like DBDEXfontAttr */
    
    if (f_str)
        return (f_str);
    else
        return (f_att);
}

/*
 * function: ElfbXTextSize - return size of passed text.
 * The return format is <width><height><ascent>.
 * The width is a simple sum of horizontal escapements.
 * The height and ascent are the overall font numbers.
 * if forPrinter is TRUE measurement is done using printer metrics, and
 * return is in mils.  if forPrinter is FALSE or null, measurement is done
 * against the screenfont, returned numbers are in 75dpi screenpix.
 */
int
ElfbTextSize(text, len, fontFamily, pointsize, weight, slant, forPrinterP)
data_ptr text;           /* in */
data_ptr len;            /* in: length of text */
data_ptr fontFamily;     /* in: typeface family name, e.g., "Courier" */
data_ptr pointsize;      /* in: in points */
data_ptr weight;         /* in: typical values: 1 - bold, 0 - regular */
data_ptr slant;          /* in: typical values: 1 - italic, 0 - upright */
data_ptr forPrinterP;    /* in: bool */ 
{
/* Ashisuto modified for wide chars. Attension that "len" is from ELF func
 * len@() or string_index@(), and these ELF functions are in CHARACTER unit,
 * not byte unit, for Oriental build, CHARACTER length are different from
 * byte length.	1998.02.24
 */
#if BUILD_WIDE
    wchar *textStr;
#else
    mbchar *textStr;
#endif
    int textLen, wid, ht, scalable, ret, ascent, ix;
    int forPrinter = FALSE;
    XFontStruct *fs;
    data_ptr sizeData;
    targetFontMetrics *tfm;
                
    
#if BUILD_WIDE
	textStr = AxWideStrPtrFromDataPtr(text);
#else
    textStr = AxStrPtrFromDataPtr(text);
#endif
    textLen = AxIntFromDataPtr(len);
    forPrinter = AxIntFromDataPtr(forPrinterP);
    
    if (forPrinter) {
        wid = 108;
        ht = 231;
        ascent = 187;     /* default valuse, from helvetica bold 14 */
    }
    else {
        wid = 7;
        ht = 16;
        ascent = 13;      /* default values, from axsab-16 */
    }
    ret = TRUE;

    sizeData = AxMakeArray(3);       /* return height, width, ascent */
    
    if (!textStr || !textLen) {
        ret = FALSE;
        goto all_done;
    }
    
    if (fontFamily) {
        mbchar *famStr;
        int psize;

        famStr = AxStrPtrFromDataPtr(fontFamily);
        psize = AxIntFromDataPtr(pointsize);
        if (!psize) {
            psize = 14;
        }
        
        if (famStr && psize) {
            
            if (forPrinter) {
                if (!axGetTargetFontMetrics(famStr, AxIntFromDataPtr(weight),
                                            AxIntFromDataPtr(slant),
                                            psize * 10000, &tfm)) {
                    ret = FALSE;
                    goto all_done;
                }
                fs = tfm->local->xfont;
            }
            else {
                if (!getXFontStruct(famStr, psize, AxIntFromDataPtr(weight),
                                    AxIntFromDataPtr(slant), &scalable, &fs)) {
                    ret = FALSE;
                    goto all_done;
                }
            }
        }
        else {
            ret = FALSE;
            goto all_done;
        }
    }
    else {
        if (forPrinter) {
                if (!axGetTargetFontMetrics(NXLT("Helvetica"), TRUE, FALSE,
                                            140000, &tfm)) {
                    ret = FALSE;
                    goto all_done;
                }
                fs = tfm->local->xfont;
        }
        else {
            fs = AxDlgMainFont;
        }
    }

    if (forPrinter) {
        XCharStruct *xc;
        for (wid = 0, ix = 0; ix < textLen; ix++) {
#if BUILD_WIDE
            if (xc = targetCharStruct(textStr[ix], tfm)) {
                wid += xc->width;
            }
#else
            if (xc = targetCharStruct((unsigned char)textStr[ix], tfm)) {
                wid += xc->width;
            }
#endif
        }
    }
    else {
#if BUILD_WIDE
        wid = AXWTextWidth(fs, textStr, textLen);
#else
        wid = AXTextWidth(fs, textStr, textLen);
#endif
    }
    ht = fs->max_bounds.ascent + fs->max_bounds.descent;
    ascent = fs->max_bounds.ascent;
    
all_done:    
    sizeData = AxAddIntToArray(sizeData, 0, wid);
    sizeData = AxAddIntToArray(sizeData, 1, ht);
    sizeData = AxAddIntToArray(sizeData, 2, ascent);
    ElfRetData = sizeData;
    return ret;
}




/***********************************************************************
function: ElfbMarkEditBoxRegion - Select a region in the edit-box;
given starting and ending row-column positions.
*/
void ElfbMarkEditBoxRegion(wid_dataP, dummy1, dummy2,start_rowP, start_colP,
                      end_rowP, end_colP)
data_ptr wid_dataP;
data_ptr dummy1;
data_ptr dummy2;
data_ptr start_rowP;
data_ptr start_colP;
data_ptr end_rowP;
data_ptr end_colP;
{
    AXWidget wid;

    int start_row, start_col, end_row, end_col;

    start_row = IntFromDataPtr( start_rowP );
    start_col = IntFromDataPtr( start_colP );
    end_row = IntFromDataPtr( end_rowP );
    end_col = IntFromDataPtr( end_colP );

    wid = WidgetPointerFromHandleData( wid_dataP );
    MarkEditBoxRegion((AXWidget_editbox)wid, start_row, start_col, end_row, end_col);
}

/*************************************************************************
function: ElfbEditBoxEdited - Return current contents of edit box
if edited since last queried (Need to free the return val of EditBoxGetData)
*/
void ElfbEditBoxEdited(wid_dataP)
data_ptr wid_dataP;
{
    char *str;
    AXWidget wid;
    wid = WidgetPointerFromHandleData( wid_dataP );
    str = (char *)AxXmEditBoxEdited((AXWidget_editbox)wid);
    ElfReturnArrayOfStrings(str);

    if ( str )
        AXmFreeString(str);
}
 
/*************************************************************************
function: ElfbCopyEditBoxSeln - Copy edit box selection to clipboard
*/
void ElfbCopyEditBoxSeln (w_dataP)
data_ptr w_dataP;
{
    AXWidget wid;
 
    wid = WidgetPointerFromHandleData( w_dataP );
    AxXmCopyEditBoxSeln(wid);
}
 
/*************************************************************************
function: ElfbCutEditBoxSeln - Cut edit box selection and save to clipboard
*/
void ElfbCutEditBoxSeln (w_dataP)
data_ptr w_dataP;
{
    AXWidget wid;
 
    wid = WidgetPointerFromHandleData( w_dataP );
    AxXmCutEditBoxSeln(wid);
}
 
 
/*************************************************************************
function: ElfbPasteEditBoxSeln - Paste selection to edit box
*/
void ElfbPasteEditBoxSeln (w_dataP)
data_ptr w_dataP;
{
    AXWidget wid;
 
    wid = WidgetPointerFromHandleData( w_dataP );
    AxXmPasteEditBoxSeln(wid);
}


/***********************************************************************
function: ElfbGetEditBoxSelectedRegion - Return the starting and
ending row-column positions of selected text in the edit-box.
If there is no cursor, same as before first char of first line.
Returned cursor values are 1-based for Elf consistancy.
*/
void ElfbGetEditBoxSelectedRegion( wid_dataP )
data_ptr wid_dataP;
{
    int start_row, start_col, end_row, end_col, ret;
    AXWidget wid;
 
    wid = WidgetPointerFromHandleData( wid_dataP );
 
    ret = AxXmGetEditBoxSelectedRegion((AXWidget_editbox)wid, &start_row,
	                                   &start_col, &end_row, &end_col);
    if ( ret == -1 ) {
        ElfReturnArrayOfFourNums(0, 1, 0, 1);
    }
    else {
        ElfReturnArrayOfFourNums(start_row, start_col+1, end_row, end_col+1);
    }
}
 
/* : AxDefaultButtonCenter - return bool whether passed dialog box
 * has a default button, and if so return the x and y coords of
 * its center.  Note that out params xoffset, yoffset will point
 * to the center of the dbox if no default button is found
 */
int
AxDefaultButtonCenter(dbox, xoffset, yoffset)
AXWidget_dialog dbox; /* in */
int *xoffset;         /* out: return default button center if found */
int *yoffset;
{
    int bx, by, bwid, bheight;
    AXWidget_main mw;

    *xoffset = dbox->width / 2;       /* default: center of box */
    *yoffset = dbox->height / 2;
    
    if (dbox->defaultButton) {
        bx = dbox->defaultButton->xpos;
        by = dbox->defaultButton->ypos;
        bwid = dbox->defaultButton->width;
        bheight = dbox->defaultButton->height;
        if (dbox->parent && dbox->parent->type == W_MAIN) {
            mw = dbox->parent;
            if (mw->menuBar) {
                by += mw->menuBar->height;
            }
            if (mw->toolBar) {
                by += AxToolBarHeight(mw, dbox->width);
            }
        }
        *xoffset = bx + bwid / 2;
        *yoffset = by + bheight / 2;
    }
    return (dbox->defaultButton != (AXWidget_pushbutton) NULL);
}

int
AxUseWorkColors(gadget)
data_ptr gadget;            /* elf data in format gadget_ */
{
    int gadget_val;
    static int useWorkAreaColors = -1;
    
    if (useWorkAreaColors == -1) {
        int itmp = UPnumVal(PROF_WIDGET_WORK_COLORS_);
        useWorkAreaColors = (itmp != 0);    /* force to boolean val */
    }
    if (useWorkAreaColors) {
        return (TRUE);        /* positive preference setting overrides */
    }
    gadget_val = FALSE;
    if (gadget) {
        /* for those widgets that support "use work area color",
         * the setting is carried in DBDEXfg_color
         */
        gadget_val = AxIntFromArray(gadget, DBDEXfg_color);
        gadget_val = (gadget_val != 0);
    }
    return (gadget_val);
}


#if AXMS    /* { */
/* this function directly calls the object method for exposes. */
/* needed for "active scrolling" on Windows NT. */
static void
AxWinDoCanvasExpose(AXWidget w, int pid, CanvasEventData *data)
{
    int save_THIMpid;
    data_ptr pt1, pt2, method_strP, countP;

    save_THIMpid = THIMpid;
    VfyContext(pid);
    pt1 = ELF_CREATE_ARRAY(2);
    pt2 = ELF_CREATE_ARRAY(2);
    ElfAddNumToArray(pt1, 0, data->expose.x);
    ElfAddNumToArray(pt1, 1, data->expose.y);
    ElfAddNumToArray(pt2, 0, data->expose.x + data->expose.width);
    ElfAddNumToArray(pt2, 1, data->expose.y + data->expose.height);

    method_strP = ELF_CREATE_STRING(NXLT("expose_event$"));
    countP = ELF_CREATE_INTEGER(data->expose.count);
    ElfCallByName(NXLT("call_obj_set_method$"), NULL, 5, ElfObjFromWidget(w),
                                            method_strP, pt1, pt2, countP);
    ELF_DATA_FREE(method_strP);
    ELF_DATA_FREE(countP);
    ELF_DATA_FREE(pt1);
    ELF_DATA_FREE(pt2);
    VfyContext(save_THIMpid);
}

static void
AxWinDoCanvasScroll(AXWidget w, int pid, CanvasEventData *data)
{
    int save_THIMpid;
    data_ptr aP, method_strP;

    save_THIMpid = THIMpid;
    VfyContext(pid);
    aP = ELF_CREATE_ARRAY(2);
    ElfAddNumToArray(aP, 0, data->scrolled.leftPoint);
    ElfAddNumToArray(aP, 1, data->scrolled.topPoint);

    method_strP = ELF_CREATE_STRING(NXLT("scroll_event$"));
    ElfCallByName(NXLT("call_obj_set_method$"), NULL, 3, ElfObjFromWidget(w),
                                                            method_strP, aP);
    ELF_DATA_FREE(method_strP);
    ELF_DATA_FREE(aP);
    VfyContext(save_THIMpid);
}
#endif    /* } AXMS */
