/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: mapiapp.c
   Applix ASC version: 10
   Date: Tue Feb  2 15:22:10 1999
*/
#include <aos.h>
#include <elf.h>
#include <getstr4.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>

#if AXMS	/* { */
#include "mmapp.h"
#include "mapi.h"
#define MAIN 1
#include "mminit.h"

/* temporary limits */

#define MAXFILES	10
#define MAXRECIPS	10

#define MAPI_LOGON		0
#define MAPI_LOGOFF		1
#define MAPI_SENDMSG	2
#define MAPI_MSGLIST	3
#define MAPI_ADDRESS	4
#define MAPI_CHKNAME	5
#define MAPI_DETAILS	6

extern int Running_Windows95;
extern HWND AxTopHWNDNeverMapped;
extern data_ptr ElfRetData;

data_ptr retData = NULL;

HANDLE hLibrary;
LHANDLE hMAPISession;

char szSeedMessageID[512];
char szMessageID[512];
char szSubject[512];
char szNoteText[4096];
char szBuf[128];

lpMapiMessage FAR *lppMessage;
lpMapiMessage lpMessage;

LPSTR lpszSeedMessageID = &szSeedMessageID[0];
LPSTR lpszMessageID = &szMessageID[0];

LPSTR pszSubject = &szSubject[0];
LPSTR pszNoteText = &szNoteText[0];
LPSTR pszDateReceived = NXLT("1991/03/23 12:00");

MapiFileDesc file[MAXFILES];
MapiRecipDesc recip[MAXRECIPS];
MapiRecipDesc rdOriginator = { 0L,MAPI_ORIG,(LPSTR)NXLT("ONE"),NULL};
MapiMessage msg;

void ElfStrAbort();
typeErr ErrnoErr();
data_ptr AxArrayFromArray(), AxAddStrToArray();
data_ptr AxMakeArray(), AxAddArrayToArray(), AxAddIntToArray();
data_ptr AxMakeIntData(), AxMakeStrData(), AxAddBinaryToArray();

int Astat(), ElfbArraySize(), AxIntFromArray(), AxBinaryFromArray();

int ElfbMapiSendMsg(), ElfbMapiMsgList(), ElfbMapiAddress();
int ElfbMapiLogon(), ElfbMapiLogoff(), AxIntFromDataPtr();
int ElfbMapiResolveName(), ElfbMapiDetails();

char *AxStrFromArray(), *strip_path();
void *THIMfree();

void InitMessage(lpMapiMessage pmmMessage)
{
  pmmMessage->ulReserved = 0L;
  pmmMessage->lpszSubject =pszSubject;
  pmmMessage->lpszNoteText =pszNoteText;
  pmmMessage->lpszMessageType = NULL;
  pmmMessage->lpszDateReceived = pszDateReceived;
  pmmMessage->flFlags = MAPI_UNREAD;
  pmmMessage->lpOriginator = &rdOriginator;
  pmmMessage->nRecipCount = 0L;
  pmmMessage->lpRecips = NULL;
  pmmMessage->nFileCount = 0L;
  pmmMessage->lpFiles = NULL;
}

/************************************************************************/
/*                                                                      */
/*   Put some stuff here...                                             */
/*                                                                      */
/*                                                                      */
/************************************************************************/

/*************************************************************************
function: ElfbMapiFuncs - the dispatcher for the Elf->Mapi interface
functions

- written 8/94 by Rob
*/

int ElfbMapiFuncs(funcId, funcData)
int funcId;
data_ptr funcData;
{
	int errnum;

	switch(funcId)
	{
		case MAPI_LOGON:

			errnum = ElfbMapiLogon(funcData);

			if (errnum != 0)
			{
				if ( (errnum == MAPI_USER_ABORT) ||
				  (errnum == MAPI_E_LOGIN_FAILURE) )
				{
					/*
					 * user cancelled logon attempt - don't
					 * throw an error, just treat it as a
					 * failed login attempt
					 */

					retData = AxMakeIntData(0);
				}
				else
				{
					ElfStrAbort(ERR_ILLVAL, szBuf, NULL);
				}
			}
			else
				retData = AxMakeIntData(hMAPISession);

			break;

		case MAPI_LOGOFF:

			errnum = ElfbMapiLogoff(funcData);

			if (errnum != 0)
			{
				ElfStrAbort(ERR_ILLVAL, szBuf, NULL);
			}
			else
				retData = AxMakeIntData(0);

			break;

		case MAPI_SENDMSG:

			errnum = ElfbMapiSendMsg(funcData);

			if (errnum != 0)
			{
				ElfStrAbort(ERR_ILLVAL, szBuf, NULL);
			}
			else
				retData = AxMakeIntData(0);

			break;

		case MAPI_MSGLIST:

			errnum = ElfbMapiMsgList(funcData);

			if (errnum != 0)
			{
				ElfStrAbort(ERR_ILLVAL, szBuf, NULL);
			}

			break;

		case MAPI_ADDRESS:

			errnum = ElfbMapiAddress(funcData);

			if (errnum != 0)
			{
				ElfStrAbort(ERR_ILLVAL, szBuf, NULL);
			}

			break;

		case MAPI_CHKNAME:

			errnum = ElfbMapiResolveName(funcData);

			if (errnum != 0)
				ElfStrAbort(ERR_ILLVAL, szBuf, NULL);

			break;

		case MAPI_DETAILS:

			errnum = ElfbMapiDetails(funcData);

			if (errnum != 0)
				ElfStrAbort(ERR_ILLVAL, szBuf, NULL);

			break;

		default:

			sprintf(szBuf,XLT("Error: Unknown MAPI function: %d"), funcId);

			ElfStrAbort(ERR_ILLVAL, szBuf, NULL);

			break;
	}

	ElfRetData = retData;

	return(0);
}

/*************************************************************************
function: ElfbMapiLogon - Initiate a MAPI session by prompting the user
for their name and password via a MAPI dialog box.

Returns an opaque session handle used in subsequent MAPI calls

- written 8/94 by Rob
*/

int ElfbMapiLogon(rData)
data_ptr rData;
{
	ULONG ulResult;
	FLAGS flFlag;
	HWND hWnd;
	char *username = StrFromArray(rData,0);
	char *password = StrFromArray(rData,1);

	/*
	 * initialize the MAPI DLL's - the DLL's are loaded here and
	 * unloaded in ElfbMapiLogoff
	 */

	ulResult = InitMAPI();

	if (ulResult != 0L)
	{
		sprintf(szBuf,XLT("Error: InitMAPI returned: %d"), ulResult);

		return(-1);
	}

	/* initiate the MAPI session */

	hWnd = AxTopHWNDNeverMapped;

	flFlag =MAPI_NEW_SESSION;

	ulResult = (*lpfnMAPILogon)(hWnd, username, password, flFlag, 0L,
	  (LPLHANDLE)&hMAPISession);

	if (ulResult != 0L)
	{
		if ( (ulResult == MAPI_USER_ABORT) ||
		  (ulResult == MAPI_E_LOGIN_FAILURE) )
		{
			return(ulResult);
		}
		else
		{
			sprintf(szBuf,XLT("Error: MAPILogon returned: %d"), ulResult);

			return(-1);
		}
	}
}

/*************************************************************************
function: ElfbMapiLogoff - Terminate a MAPI session and unload the DLLs

- written 8/94 by Rob
*/

int ElfbMapiLogoff(rData)
data_ptr rData;
{
	ULONG ulResult;
	HWND hWnd;
	LHANDLE hMAPISession;

	hMAPISession = AxIntFromDataPtr(rData);

	/* terminate the MAPI session */

	hWnd = AxTopHWNDNeverMapped;

	ulResult = (*lpfnMAPILogoff)(hMAPISession, hWnd, 0L, 0L);

	if (ulResult != 0)
	{
		sprintf(szBuf,XLT("Error: MAPILogoff returned: %d"), ulResult);

		return(-1);
	}

	DeInitMAPI();

	return(0);
}

/*************************************************************************
function: ElfbMapiSendDoc - generic interface to send documents via
the MAPI API

- written 8/94 by Rob
*/

int ElfbMapiSendDoc(filename)
char *filename;
{
	FLAGS flFlag = 0L;
	ULONG ulResult;
	char sepchar[2];
	char *fname;
	HWND hWnd = 0;
	struct stat lstat;

	if (Astat(filename, &lstat) != 0)
		ElfStrAbort(ErrnoErr(), NULL, filename);

	strcpy(sepchar,";");

	fname = strip_path(filename);

	ulResult = InitMAPI();

	if (ulResult != 0L)
	{
		sprintf(szBuf,XLT("Error: InitMAPI returned: %d"), ulResult);

		ElfStrAbort(ERR_ILLVAL, szBuf, NULL);
	}

	/* initiate the MAPI session */

	flFlag =MAPI_NEW_SESSION | MAPI_LOGON_UI;

	ulResult = (*lpfnMAPILogon)(hWnd, NULL, NULL, flFlag, 0L,
	  (LPLHANDLE)&hMAPISession);

	if (ulResult != 0L)
	{
		if ( (ulResult == MAPI_USER_ABORT) ||
		  (ulResult == MAPI_E_LOGIN_FAILURE) )
		{
			DeInitMAPI();
			return(0);
		}
		else
		{
			sprintf(szBuf,XLT("Error: MAPILogon returned: %d"), ulResult);

			ElfStrAbort(ERR_ILLVAL, szBuf, NULL);
		}
	}

	ulResult = (*lpfnMAPISendDocuments)(hWnd, sepchar, filename,
	  fname, 0L);

	if (ulResult != 0L)
	{
		if (ulResult != MAPI_USER_ABORT)
		{
			sprintf(szBuf,
			  XLT("Error: MAPISendDocuments returned: %d"), ulResult);

			ElfStrAbort(ERR_ILLVAL, szBuf, NULL);
		}
	}

	/* terminate the MAPI session */

	ulResult = (*lpfnMAPILogoff)(hMAPISession, hWnd, 0L, 0L);
	if (ulResult != 0)
	{
		sprintf(szBuf,XLT("Error: MAPILogoff returned: %d"), ulResult);

		ElfStrAbort(ERR_ILLVAL, szBuf, NULL);
	}

	DeInitMAPI();
}

/*************************************************************************
function: ElfbMapiSendMsg - generic interface to send Mail messages
via the MAPI API

The 'use_dialog' flag controls whether the user is prompted for
information with a MAPI Send dialog box;  if use_dialog is FALSE,
at least one recipient name MUST be specified, otherwise an error
is thrown. If use_dialog is TRUE, the dialog will be displayed with
all of the passed data pre-loaded.

- written 8/94 by Rob
*/

int ElfbMapiSendMsg(rData)
data_ptr rData;
{

	LHANDLE hMAPISession;
	HWND hWnd;
	data_ptr recips;
	data_ptr body;
	data_ptr files;
	data_ptr file_info;
	FLAGS flFlag = 0L;
	ULONG ulResult;
	char *recip_name;
	char *rname[MAXRECIPS];
	char *fname[MAXFILES];
	char *filename;
	char *subject;
	char *bodystr;
	char *psubj;
	char *tmpstr;
	int recip_type;
	int use_dialog;
	int nrecips;
	int nfiles;
	int fcnt;
	int bcnt;
	int bsize;
	int ix;
	int errnum;
	int errors;
	int is_certified;
	struct stat lstat;
	data_ptr recip_ent;

	errors = FALSE;

	hWnd = AxTopHWNDNeverMapped;

	/* take the passed args apart */

	hMAPISession = AxIntFromArray(rData,0);
	recips = AxArrayFromArray(rData,1);
	psubj = AxStrFromArray(rData,2);
	body = AxArrayFromArray(rData,3);
	files = AxArrayFromArray(rData,4);
	use_dialog = AxIntFromArray(rData,5);
	is_certified = AxIntFromArray(rData,6);

	/* process the message recipients */

	nrecips = ElfbArraySize(recips);

	if (nrecips == 0)
	{
		msg.lpRecips = NULL;
		msg.nRecipCount = 0;
	}
	else
	{
		for (ix = 0; ix < nrecips; ix++)
		{
			recip_ent = AxArrayFromArray(recips,ix);

			recip_name = AxStrFromArray(recip_ent,0);
			recip_type = AxIntFromArray(recip_ent,1);

			rname[ix] =
			  (char *)malloc( (strlen(recip_name)+1)*sizeof(char) );

			strcpy(rname[ix], recip_name);

			recip[ix].ulReserved = 0;
			recip[ix].ulRecipClass = recip_type;
			recip[ix].lpszName = rname[ix];
			recip[ix].lpszAddress = NULL;
			recip[ix].ulEIDSize = 0;
			recip[ix].lpEntryID = NULL;
		}

		msg.lpRecips = recip;
		msg.nRecipCount = nrecips;
	}

	/* process the attached files */

	nfiles = ElfbArraySize(files);

	fcnt = 0;

	for (ix = 0; ix < nfiles; ix++)
	{
		file_info = AxArrayFromArray(files,ix);
		filename = AxStrFromArray(file_info,0);

		if (Astat(filename, &lstat) == 0)
		{
			fname[fcnt] =
			  (char *)malloc( (strlen(filename)+1)*sizeof(char) );

			strcpy(fname[fcnt], filename);

			file[fcnt].ulReserved = 0;
			file[fcnt].flFlags = 0;
			file[fcnt].lpszPathName = fname[fcnt];
#if 0
			file[fcnt].lpszFileName = strip_path(filename);
#endif
			file[fcnt].lpFileType = NULL;

			fcnt++;
		}
	}

	msg.nFileCount = fcnt;

	if (fcnt != 0)
		msg.lpFiles = file;
	else
		msg.lpFiles = NULL;

	subject = (char *)malloc( (strlen(psubj)+1) * sizeof(char) );
	strcpy(subject, psubj);

	/*
	 * allocate enough space for the message body and the attached
	 * files which occupy a position in the message text
	 */

	bcnt = ElfbArraySize(body);

	bsize = 0;

	for (ix = 0; ix < bcnt; ix++)
	{
		tmpstr = AxStrFromArray(body,ix);

		bsize += strlen(tmpstr) + 1;
	}

	bodystr = (char *)malloc( (bsize + fcnt + 1)*sizeof(char) );

	*bodystr = '\0';

	for (ix = 0; ix < bcnt; ix++)
	{
		tmpstr = AxStrFromArray(body,ix);

		if (ix == 0)
			sprintf(bodystr,"%s\n", tmpstr);
		else
		{
			strcat(bodystr,tmpstr);
			strcat(bodystr,"\n");
		}
	}

	/*
	 * specify the location of any attached files at the end
	 * of the message text
	 */

	if (fcnt != 0)
	{
		for (ix = 0; ix < fcnt; ix++)
		{
			file[ix].nPosition = bsize + ix;

			strcat(bodystr,"x");
		}
	}
 
	/* assemble the message */

	msg.ulReserved = 0L;
	msg.lpszSubject = subject;
	msg.lpszNoteText = bodystr;
	msg.lpszMessageType = NULL;
	msg.lpszDateReceived = NULL;
	msg.lpszConversationID = NULL;
	msg.lpOriginator = &rdOriginator;

	if (is_certified)
		msg.flFlags = MAPI_RECEIPT_REQUESTED;
	else
		msg.flFlags = 0;

	if (use_dialog)
		flFlag = MAPI_DIALOG;
	else
		flFlag = 0;

	ulResult = (*lpfnMAPISendMail)(hMAPISession, hWnd, &msg, flFlag, 0L);

	if (ulResult != 0L)
	{
		if (ulResult == MAPI_USER_ABORT)
			errnum = 0;
		else
			errnum = ulResult;
	}
	else
		errnum = 0;

	if (errnum != 0)
		sprintf(szBuf,XLT("Error: MAPISendMail returned: %d"), errnum);

	for (ix = 0; ix < nrecips; ix++)
		free(rname[ix]);

	for (ix = 0; ix < fcnt; ix++)
		free(fname[ix]);

	free(bodystr);
	free(subject);

	if (errors)
		return(-1);
	else
		return(0);
}

/*************************************************************************
function: ElfbMapiMsgList - generic interface to list the contents of the
caller's Inbox via the MAPI API

- written 8/94 by Rob
*/

ElfbMapiMsgList(rData)
data_ptr rData;
{
	data_ptr sData;
	int ix = 0;
	ULONG ulResult;
	HWND hWnd;
	FLAGS flFlag;

	hWnd = AxTopHWNDNeverMapped;

	hMAPISession = AxIntFromDataPtr(rData);

	*lpszSeedMessageID = '\0';
	flFlag = 0;

	retData = AxMakeArray(0);

	while ( ulResult = (*lpfnMAPIFindNext)(hMAPISession, hWnd, NULL,
	  lpszSeedMessageID, flFlag, 0L, lpszMessageID) == 0 )
	{
		strcpy(lpszSeedMessageID,lpszMessageID);

		lppMessage = (lpMapiMessage FAR *) &lpMessage;

		ulResult = (*lpfnMAPIReadMail)(hMAPISession, hWnd, lpszMessageID,
		  flFlag, 0L, lppMessage);

		if (ulResult == 0)
		{
			sData = AxMakeArray(5);

			sData = AxAddStrToArray(sData,0,lpMessage->lpOriginator->lpszName);
			sData = AxAddStrToArray(sData,1,lpMessage->lpRecips[0].lpszName);
			sData = AxAddStrToArray(sData,2,lpMessage->lpszDateReceived);
			sData = AxAddStrToArray(sData,3,lpMessage->lpszSubject);
			sData = AxAddStrToArray(sData,4,lpMessage->lpszNoteText);

			retData = AxAddArrayToArray(retData,ix++,sData);

			ulResult = lpfnMAPIFreeBuffer((LPVOID)lpMessage);
		}
		else
		{
			sprintf(szBuf,
			  XLT("Error: MAPIReadMail returned: %d"), ulResult);

			return(-1);
		}
	}

	if ( (ulResult != SUCCESS_SUCCESS) && 
	  (ulResult != MAPI_E_NO_MESSAGES) )
	{
		sprintf(szBuf,XLT("Error: MAPIFindNext returned: %d"), ulResult);

		return(-1);
	}

	return(0);
}

/*************************************************************************
function: ElfbMapiAddress - display the MAPI dialog used to address msgs

- written 8/94 by Rob
*/

ElfbMapiAddress(rData)
data_ptr rData;
{
	data_ptr sData;
	data_ptr recips;
	data_ptr recip_ent;
	HWND hWnd;
	MapiRecipDesc *pRecipient;
	ULONG rcnt;
	ULONG ulResult;
	char *rname[MAXRECIPS];
	char *recip_name;
	int recip_type;
	int ix;
	int nEditFields;
	int flFlag;
	int errors;
	int errnum;
	int nrecips;

	errors = FALSE;

	nEditFields = 3;
	flFlag = MAPI_LOGON_UI;
	hWnd = AxTopHWNDNeverMapped;

	retData = AxMakeArray(0);

	hMAPISession = AxIntFromArray(rData,0);
	recips = AxArrayFromArray(rData,1);

	/* set up the initial recipients */

	nrecips = ElfbArraySize(recips);

	if (nrecips == 0)
		*recip = rdOriginator;
	else
	{
		for (ix = 0; ix < nrecips; ix++)
		{
			recip_ent = AxArrayFromArray(recips,ix);

			recip_name = AxStrFromArray(recip_ent,0);
			recip_type = AxIntFromArray(recip_ent,1);

			rname[ix] =
			  (char *)malloc( (strlen(recip_name)+1)*sizeof(char) );

			strcpy(rname[ix], recip_name);

			recip[ix].ulReserved = 0;
			recip[ix].ulRecipClass = recip_type;
			recip[ix].lpszName = rname[ix];
			recip[ix].lpszAddress = NULL;
			recip[ix].ulEIDSize = 0;
			recip[ix].lpEntryID = NULL;
		}
	}

	flFlag = 0;

	ulResult = (*lpfnMAPIAddress)(hMAPISession, hWnd, NULL, nEditFields,
	  NULL, nrecips, recip, flFlag, 0, &rcnt, &pRecipient);

	if (ulResult != 0L)
	{
		if (ulResult == MAPI_USER_ABORT)
			errnum = 0;
		else
			errnum = ulResult;

		retData = ElfDataCopy(recips);
#if 0
		retData = recips;
#endif
	}
	else
	{
		errnum = 0;

		for (ix=0; ix < (INT)rcnt; ix++)
		{
			sData = AxMakeArray(4);

			sData = AxAddStrToArray(sData,0,pRecipient[ix].lpszName);
			sData = AxAddIntToArray(sData,1,pRecipient[ix].ulRecipClass);
			sData = AxAddIntToArray(sData,2,pRecipient[ix].ulEIDSize);
			sData = AxAddBinaryToArray(sData,3,
			  pRecipient[ix].ulEIDSize,pRecipient[ix].lpEntryID);
#if 0
			sData = AxAddBinaryToArray(sData,3,
			  sizeof(LPVOID),&pRecipient[ix].lpEntryID);
#endif

			retData = AxAddArrayToArray(retData,ix,sData);
		}

		if (rcnt != 0)
			ulResult = (*lpfnMAPIFreeBuffer)(pRecipient);
	}

	for (ix = 0; ix < nrecips; ix++)
		free(rname[ix]);

	if (errors)
		return(-1);
	else
		return(0);
}

/*************************************************************************
function: strip_path - remove the path information from a filename
*/
char *strip_path(infile)
char *infile;
{
	static char filename[1024+1];
	char *name;

	strcpy(filename,infile);
	name = strrchr(filename,'/');

	if (!name)
		name = strrchr(filename,'\\');

	if (name)
		name = name + 1;
	else
		name = filename;

	return(name);
}

/*************************************************************************
function: ElfbMapiResolveName - validate the passed name against MAPI
address book(s)

- written 8/94 by Rob
*/

int ElfbMapiResolveName(rData)
data_ptr rData;
{
	char *name;
	LHANDLE hMAPISession;
	HWND hWnd;
	MapiRecipDesc *pchecked_name;
	ULONG ulResult;
	FLAGS flFlag;

	hWnd = AxTopHWNDNeverMapped;

	hMAPISession = AxIntFromArray(rData,0);
	name = AxStrFromArray(rData,1);

	flFlag = MAPI_DIALOG;

	ulResult = (*lpfnMAPIResolveName)(hMAPISession, hWnd, name,
	  flFlag, 0, &pchecked_name);

	if (ulResult != 0L)
	{
		if ( (ulResult == MAPI_USER_ABORT) ||
		  (ulResult == MAPI_E_AMBIGUOUS_RECIPIENT) ||
		  (ulResult == MAPI_E_UNKNOWN_RECIPIENT) )
		{
			retData = NULL;

			return(0);
		}
		else
		{
			sprintf(szBuf,
			  XLT("Error: MAPIResolveName returned: %d"), ulResult);

			retData = NULL;

			return(-1);
		}
	}
	else
	{
		retData = AxMakeArray(3);

		retData = AxAddStrToArray(retData,0,pchecked_name->lpszName);
		retData = AxAddIntToArray(retData,1,pchecked_name->ulEIDSize);
		retData = AxAddBinaryToArray(retData,2,pchecked_name->ulEIDSize,
		  pchecked_name->lpEntryID);

		ulResult = (*lpfnMAPIFreeBuffer)(pchecked_name);

		return(0);
	}
}

int ElfbMapiDetails(rData)
data_ptr rData;
{
	LHANDLE hMAPISession;
	HWND hWnd;
	MapiRecipDesc precip;
	ULONG ulResult;
	ULONG EntrySize;
	int size;
	FLAGS flFlag;
	LPVOID EntryID;

	hWnd = AxTopHWNDNeverMapped;

	hMAPISession = AxIntFromArray(rData,0);
	EntrySize = AxIntFromArray(rData,1);
	size = AxBinaryFromArray(rData,2,&EntryID);

	precip.ulReserved = 0;
	precip.ulRecipClass = 0;
	precip.lpszName = NULL;
	precip.lpszAddress = NULL;
	precip.ulEIDSize = EntrySize;
	precip.lpEntryID = (LPVOID)EntryID;

	flFlag = MAPI_AB_NOMODIFY;

	ulResult = (*lpfnMAPIDetails)(hMAPISession, hWnd, &precip,
	  flFlag, 0);

	retData = AxMakeIntData(ulResult);

	return(0);
}

#else	/* if AXMS } { */
int ElfbMapiFuncs(funcId, funcData)
int funcId;
data_ptr funcData;
{
	ElfStrAbort(ERR_ILLVAL,
	  XLT("MAPI functions only supported on NT platform"), NULL);

	return(-1);
}

#endif /* } AXMS */
