/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: recogniz.c
   Applix ASC version: 38
   Date: Thu Mar  4 21:13:43 1999
*/
#if CLIX || M_88K
#include <sys/types.h>
#endif

#if AXMS
#include <ctype.h>
#endif

#include <string.h>
#include <errno.h>
#include <varargs.h>
#include <vfont.h>

#include <aos.h>
#include <elf_env.h>
#include <recogniz.h>
#include <doctypes.h>
#include <of.h>  /* for OLE class ids */

#ifndef VFONT_MAGIC
#define VFONT_MAGIC 0436
#endif

#ifndef streq
#define streq(s1,s2) ((strcmp((s1),(s2)) == 0) ? 1 : 0)
#endif

#ifndef is_Latin_1
#define is_Latin_1(c) ((unsigned char)(c) != 0 && \
					   (((unsigned char)(c)<0x7F) || \
						(((unsigned char)(c)) >= 160)))
#endif
extern char *ThimApplixDir;
extern char *ElfStrPtrRet;

#if !RIOS
extern char *ElfMalloc();
extern char *TrueMalloc();
#endif

#define DYNAMIC 1

static int NumMagicEntries = 0;
#if DYNAMIC
static magic_table_t *MagicTable = (magic_table_t *) NULL;
#else
static magic_table_t MagicTable[MAXCACHEENT+1];
#endif

static time_t MagicFileDtLM = -1;/* dt and time magic file was last modified */

#define FS_LENGTH    16		/* length of the file system type string */

typedef struct {
    long            disk_magic;
    long            disk_version;
    char            disk_homefs[FS_LENGTH];
    short           disk_numentries;
} diskheader;

#define AX_NUM_CLSIDS 13
static struct clsidInfo {
	char *clsid_str;	/* OLE application identifier. */
	int doctype;       /* Applix application identifier */
} clsidInfo[AX_NUM_CLSIDS] = {

    { MSW_CLSID,	  AX_DOC_TYPE_MSWORD70 },
    { MSW8_CLSID,	  AX_DOC_TYPE_MSWORD97 },
    { MSPPT_CLSID,	  AX_DOC_TYPE_PPT },
    { MSXLS_CLSID,	  AX_DOC_TYPE_XLS4 },
    { MSXLS97_CLSID,  AX_DOC_TYPE_XLS4 },
    { MSXLS97A_CLSID, AX_DOC_TYPE_XLS4 },
    { MSXLS97B_CLSID, AX_DOC_TYPE_XLS4 },
    { MSXLS97C_CLSID, AX_DOC_TYPE_XLS4 },
    { MSXLS97D_CLSID, AX_DOC_TYPE_XLS4 },
    { MSPPT97_CLSID,  AX_DOC_TYPE_PPT },
    { MSPPT97A_CLSID, AX_DOC_TYPE_PPT },
    { WP7_CLSID,	  AX_DOC_TYPE_WP_70 },
    { WP8_CLSID,	  AX_DOC_TYPE_WP_80 }
};

#include <recogniz.p>

/* Return the file type id of FILENAME.
   The file type id returned is one of the ids defined in doctypes.h */
int
ElfbRecognizeFile(filename)
register char *filename;
{
    register char *filetype;
	int retval;

	if (!filename)
		return(AX_DOC_TYPE_UNKNOWN);

    ElfStrPtrRet = filetype = ElfMalloc(AXF_BUFSIZ);
	if (!filetype)
		return(AX_DOC_TYPE_UNKNOWN);
    XmPutHourglass();

	/* If it's an OLE file but we failed to find a file id, try the old way */
	if ( AxIsOleStorageFile(filename) &&
		 (retval = axf_do_olefile(filename, filetype)) == AOK )
	{
		;
	}
	else
	{
    	retval = axf_read_magic_file();
		if (retval != 0) {
			ElfStrAbort(ErrnoErr(), NULL, NXLT("axmagic"));
			/*NOTREACHED*/
		}

    	retval = axf_dofile(filename, filetype);

#if DYNAMIC
		axf_freeMagicTable();
#endif
	}

	if (retval != 0) {
		ElfStrAbort(ErrnoErr(), NULL, filename);
		/*NOTREACHED*/
	}
    return(retval); /* always 0 */
}

/* Return the file type id of FILENAME.
   The file type id returned is one of the ids defined in doctypes.h */
int
RecognizeFile(filename)
register char *filename;
{
	char filetype[1000];
	int retval;

	if (!filename)
		return(AX_DOC_TYPE_UNKNOWN);

	/* If it's an OLE file but we failed to find a file id, try the old way */
	if ( AxIsOleStorageFile(filename) &&
		 (retval = axf_do_olefile(filename, filetype)) == AOK )
	{
		;
	}
	else
	{
    	axf_read_magic_file();
    	retval = axf_dofile(filename, filetype);

#if DYNAMIC
		axf_freeMagicTable();
#endif
	}

	if (retval != 0)
		return(AX_DOC_TYPE_UNKNOWN);

	return(atoi(filetype));	
}

/* Read one line from FP into OUTSTR, zap trailing newline.
   Return 0 on end of file; 1 otherwise. */
static int
axf_rdstr(outstr, len, fp)
char *outstr;
int  len;
FILE *fp;
{
    int last;

    if (fgets((char *) outstr, len, fp)) {
        last = strlen(outstr) - 1;
        if (outstr[last] == '\n')
            outstr[last] = '\0';
        return(1);
    }
    return(0);
}

/* Extract the token followin BUF_P.  Place extracted token in TOK.
   On success return modified BUF_P; NULL otherwise. */
static
char *
axf_gettok(buf_p, tok)
char *buf_p;
char *tok;
{
    register int ii;

    SKIPWS((char *) buf_p);
    for (ii = 0; !(IS_DELIM((char) *buf_p)); ii++)
        tok[ii] = *buf_p++;
    tok[ii] = '\0';
    if (ii)
        return(buf_p);
    else
        return((char *) NULL);
}

/* Return TRUE if the string pointed to by STRP is has white space
   only; FALSE otherwise. */
static int
axf_isblankstr(strp)
char *strp;
{
    register char *p;

    for (p = strp; *p; p++)
        if (!isspace(*p))
            return(FALSE);
    return(TRUE);
}

/* Free storage for the magic-table.  Always return 0. */
static int
axf_freeMagicTable()
{
    register int cnt;
    register magic_table_t *entp;  /* table entry pointer */
    register int n_ents;
#if DYNAMIC
	if (!NumMagicEntries || !MagicTable) {
		NumMagicEntries = 0; /* sanity */
		MagicTable = (magic_table_t *) NULL; /* sanity */
		return(0);
	}

	for (cnt = 0; cnt < NumMagicEntries; cnt++) {
		entp = (magic_table_t *) &MagicTable[cnt];
		switch (entp->vtype) {
		case VT_STRING:
			if (entp->value.v_string)
				TrueFree(entp->value.v_string);
			entp->value.v_string = (char *) NULL;
			break;
		case VT_EXTEN:
			if (entp->value.v_exten)
				TrueFree(entp->value.v_exten);
			entp->value.v_exten = (char *) NULL;
			break;
		default:
			break; /* only VT_STRING and VT_EXTEN have allocated mem. */
		}
		if (entp->outstr)
			TrueFree(entp->outstr);
		entp->outstr = (char *) NULL;
	}
	free(MagicTable);
	MagicTable = (magic_table_t *) NULL;
	NumMagicEntries = 0;
	return(0);
#else
    if (NumMagicEntries < MAXCACHEENT)
        n_ents = NumMagicEntries;
    else
        n_ents = MAXCACHEENT;   /* bad NumMagicEntries */
    for (cnt=0; cnt < n_ents; cnt++) {
        entp = (magic_table_t *) &MagicTable[cnt];
        if (entp == (magic_table_t *) NULL)    /* bad count */
            continue;
        switch(entp->vtype) {
        case VT_STRING:
            if (entp->value.v_string)
                TrueFree(entp->value.v_string);
            entp->value.v_string = (char *) NULL;
            break;
        case VT_EXTEN:
            if (entp->value.v_exten)
                TrueFree(entp->value.v_exten);
            entp->value.v_exten = (char *) NULL;
            break;
        default: break; /* only VT_STRING and VT_EXTEN have allocated mem */
        }
        if (entp->outstr)
            TrueFree(entp->outstr);
        entp->outstr = (char *) NULL;
    }
    NumMagicEntries = 0;    /* reset entry count */
	return(0);
#endif
}

/* Scan in the Aster*x magic file "axmagic".
   Return 0 on success; -1 otherwise. */
static int
axf_read_magic_file()
{
    FILE *magicfp = NULL;
    char linebuf[AXF_BUFSIZ + 1];
    char magicfilename[STRPATH + 1];
    int  next_tok = TK_OFFSET;
    char token[TOKENLEN+1];      /* extracted token */
    char vtyp_str[TOKENLEN+1];   /* value type in string form */
    register char *bufp; /* line buffer pointer */
    magic_table_t *tabp;
    char *maskp;
    char *valuep;
    char *value_operp;
    int  lineno = 0;
    int  skip_this_entry;
    char *rev_p;
    int  ii;
    char revstr[10];
    struct stat m_statbuf;  /* stat buffer for axmagicfile */
    int rehash_magic = FALSE;
	int new_size;
	int read_only;

    sprintf(magicfilename, "%s/%s", ThimApplixDir, AXF_MAGIC);
#if !DYNAMIC
    if (stat(magicfilename, &m_statbuf) == -1) {
        rehash_magic = TRUE;
    } else {
        if (m_statbuf.st_mtime != MagicFileDtLM) {
            rehash_magic = TRUE;
            MagicFileDtLM = m_statbuf.st_mtime;
        } else
            rehash_magic = FALSE;
    }
    if (rehash_magic == FALSE)
        return(0);  /* No need to rehash magic table */

    if (MagicFileDtLM != (time_t) -1)   /* not the first time */
        axf_freeMagicTable();

    MagicFileDtLM = m_statbuf.st_mtime;
#endif
	read_only = TRUE;
	if ( AFTaskLockFile(magicfilename, 0, &read_only, FALSE) != AOK )
		return(-1);
    magicfp = (FILE *)AFTaskfopen(0, magicfilename, READ_STR);
	if (!magicfp)
		return(-1);

#if DYNAMIC
	/* This is needed because an SVERIFY() in some previous call to
	   ElfbRecognizeFile() would have already freed this memory, but
	   MagicTable will still be non-null */
	NumMagicEntries = 0; /* sanity */
	MagicTable = (magic_table_t *) NULL; /* sanity */
#endif

    while (axf_rdstr(linebuf, AXF_BUFSIZ, magicfp)) {
        lineno++;
        bufp = &linebuf[0];
        skip_this_entry = FALSE;
        if (*bufp == COMMENTCHAR)  /* '#' in the first column */
            continue;
        if (axf_isblankstr(bufp))
            continue;
#if DYNAMIC
		if (!NumMagicEntries || !MagicTable) {
			MagicTable = (magic_table_t *) malloc(sizeof(magic_table_t));
		} else {
			new_size = (NumMagicEntries + 1) * sizeof(magic_table_t);
			MagicTable = (magic_table_t *) realloc(MagicTable, new_size);
		}
		tabp = (magic_table_t *) &MagicTable[NumMagicEntries];
		memset((char *) tabp, 0, sizeof(magic_table_t));
		NumMagicEntries++;
#else
        tabp = (magic_table_t *) &MagicTable[NumMagicEntries++];
        if (NumMagicEntries >= MAXCACHEENT) { /* too many entries */
            AFfclose(magicfilename, read_only, magicfp);
			AFTaskUnlockFile(magicfilename, 0, read_only);
            return((int) -1);   /* bail out */
        }
#endif
        /* Filter out special case '>' in the first column */
        if (*bufp == '>') {    /* FIXME hardcoded */
            tabp->level = 1;
            bufp++;
        } else {
            tabp->level = 0;
        }

        /* Extract tokens. */
        next_tok = TK_OFFSET;
        while (bufp = axf_gettok(bufp, token)) {
            if (skip_this_entry == TRUE)
                break;
            if (next_tok == TK_NOMORE)
                break;
            switch (next_tok) {
            case TK_OFFSET:     /* offset */
                /* let strtol() figure out the base */
                tabp->offset = (long) strtol(token, (char **) NULL, 0);
                next_tok = TK_VALTYPE;
                break;

            case TK_VALTYPE:    /* value type */
                if ((maskp = strpbrk(token, MASK_LIST)) != (char *) NULL) {
                    tabp->mask_oper = *maskp;    /* mask operator */
                    *maskp = '\0';     /* step on it */
                    strcpy(vtyp_str, token);/* value type in str form */
                    maskp++;           /* to get to the actual mask */
                    if (streq(token, VTS_BYTE))
                        tabp->mask = (uint8) strtol(maskp, (char **) NULL, 0);
                    else if (streq(token, VTS_SHORT))
                        tabp->sh_mask = (unsigned short) strtol(maskp,
                                                           (char **) NULL, 0);
                } else {                /* no mask operator */
                    tabp->mask_oper = ' ';    /* no operator */
                    tabp->mask = 0;
                    strcpy(vtyp_str, token);
                }
                if (streq(token, VTS_STRING))
                    tabp->vtype = VT_STRING;
                else if (streq(token, VTS_LONG))
                    tabp->vtype = VT_LONG;
                else if (streq(token, VTS_SHORT))
                    tabp->vtype = VT_SHORT;
                else if (streq(token, VTS_BYTE))
                    tabp->vtype = VT_BYTE;
                else if (streq(token, VTS_EXTEN))
                    tabp->vtype = VT_EXTEN;
                else {    /* error */
                    skip_this_entry = TRUE;
                    NumMagicEntries--;
                }
                next_tok = TK_VALUE;
                break;

            case TK_VALUE:  /* value */
                value_operp = token;
                if (((*value_operp == OP_GT) || (*value_operp == OP_LT) ||
                    (*value_operp == OP_EQ) || (*value_operp == OP_ANY) ||
                    (*value_operp == OP_NOT))
                    && ((tabp->vtype != VT_STRING))) {
                    tabp->oper = *value_operp;
                    valuep = ++value_operp;
                } else {
                    tabp->oper = OP_EQ;   /* default */
                    valuep = token;
                }

                if (tabp->vtype == VT_STRING) {
                    tabp->v_str_len = strlen(valuep);
                    tabp->value.v_string = (char *) 
                        TrueMalloc(tabp->v_str_len+1);
/*                    SVERIFY(tabp->value.v_string != NULL,ErrnoErr(),NULL); */
                    strcpy(tabp->value.v_string, valuep);
                } else if (tabp->vtype == VT_EXTEN) {
                    tabp->v_exten_len = strlen(valuep);
                    tabp->value.v_exten = (char *)
                        TrueMalloc(tabp->v_exten_len+1);
/*                    SVERIFY(tabp->value.v_exten != NULL,ErrnoErr(),NULL); */
                    strcpy(tabp->value.v_exten, valuep);
                } else if (tabp->vtype == VT_LONG) {
                    tabp->value.v_fourb = 
                        (long) strtol(valuep, (char **) NULL, 0);
                } else if (tabp->vtype == VT_SHORT) {
                    tabp->value.v_twob = 
                        (uint16) strtol(valuep, (char **) NULL, 0);
                } else if (tabp->vtype == VT_BYTE) {
                    tabp->value.v_oneb =
                        (uint8) strtol(valuep, (char **) NULL, 0);
                } else {
                /* internal error!! vtype should have been set at this point */
                }
                next_tok = TK_MAJMIN;
                break;

            case TK_MAJMIN: /* major,minor revision numbers */
                rev_p = token;
                for (ii = 0; isdigit(*rev_p); ii++)
                    revstr[ii] = *rev_p++;
                revstr[ii] = '\0';
                tabp->major_rev = atoi(revstr);

                rev_p++;
                for (ii = 0; isdigit(*rev_p); ii++)
                    revstr[ii] = *rev_p++;
                revstr[ii] = '\0';
                tabp->minor_rev = atoi(revstr);

                SKIPWS(bufp);
                tabp->outstr = (char *) TrueMalloc(strlen(bufp)+1);
/*                SVERIFY(tabp->outstr != (char *) NULL, ErrnoErr(), NULL); */
                strcpy(tabp->outstr, bufp);
                next_tok = TK_NOMORE;
                break;

            default:    /* internal error */
                /* internal error - unknown token at line `lineno' */
                skip_this_entry = TRUE;
                NumMagicEntries--;
                break;
            }
        }
    }
    AFfclose(magicfilename, read_only, magicfp);
	AFTaskUnlockFile(magicfilename, 0, read_only);
    return(0);
}

/* Determine the file type of FILENAME; like file(1).
   Place the file type id in string form in FILETYPE.
   Return 0 on success, -1 on failure. */
static
int
axf_dofile(filename, filetype)
char *filename;
char *filetype;
{
    int  retval;
    register int listcnt;
    struct stat  f_statbuf;

    if (axf_dostat(filename, &f_statbuf) != 0)
        return(-1);
    if ((f_statbuf.st_mode & S_IFMT) == S_IFREG)
        retval = axf_filetype(filename, &f_statbuf, filetype);
    else
        retval = axf_prtspecial(filename, &f_statbuf, filetype);
    return(retval);
}

/* Determine the file type of FILENAME, assuming it's an OLE storage file.
   Place the file type id in string form in FILETYPE.
   Return 0 on success, -1 on failure. */
static int
axf_do_olefile(filename, filetype)
char *filename;
char *filetype;
{
    int ix, retval;
	char *clsid_str;

	clsid_str = (char*)malloc(80); /* plenty big, unless MS decides to change
	                                * their GUID format.
	                                */
	retval = AxGetClassIdFromOleFile(filename, clsid_str);
	if ( retval == AOK )
	{
		retval = !AOK;
    	for( ix=0; ix < AX_NUM_CLSIDS; ix++ )
		{
    	    if ( strcmp(clsid_str, clsidInfo[ix].clsid_str) == 0 )
			{
				retval = AOK;
				ax_itoa(filetype, 10, clsidInfo[ix].doctype);
    	        break;
			}
    	}
	}
	free(clsid_str);
    return retval;
}

/* stat a file, follow sym-links */
static int
axf_dostat(file, buf_p)
char  *file;
struct stat *buf_p;
{
    /* Will always follow symlinks */
    return((int) stat(file, buf_p));
}

/* Determine the file-type of a file */
/*ARGSUSED*/
static int
axf_filetype(file, fstat_buf, filetype)
char *file;
struct stat *fstat_buf;
char *filetype;
{
    FILE *fp;
    int  n_read;         /* # of chars read */
    char defaulttype[50];/* default file type */
    char tstbuf[AXF_BUFSIZ]; /* test buffer */
    register char *buf_p = tstbuf;/* pointer to the test buffer */
    int  n_ascii = 0;    /* # of ascii chars */
    int  cnt;            /* general purpose counter */
    register int ent_cnt = 0;    /* MagicTable[] entry counter */
    register int found_match = FALSE;/* found a match with a magic entry */
    int  f_depth;
    register int look_ahead;
    char matched_str[400];   /* matched str from axf_matchent  */
    int  is_addltest, has_addltests;
    int read_only;

    read_only = TRUE;
    if ( AFTaskLockFile(file, 0, &read_only, FALSE) != AOK )
        return(-1);
    fp = (FILE *)AFTaskfopen(0, file, READ_BINARY_STR);
    if (!fp)
	{
        AFTaskUnlockFile(file, 0, read_only);
        return(-1);
	}
    memset(tstbuf, '\0', AXF_BUFSIZ);
    errno = 0;
    SkipMacHeader (fp);
    if ((n_read = fread(tstbuf, 1, AXF_BUFSIZ, fp)) == 0) 
	{
        if (errno == 0)
            strcpy(filetype, FTIID_NIL);    /* empty file */
        else 
		{
	        AFfclose(file, read_only, fp);
	        AFTaskUnlockFile(file, 0, read_only);
			return(-1);
        }

        AFfclose(file, read_only, fp);
        AFTaskUnlockFile(file, 0, read_only);
        return(0);
    }
    AFfclose(file, read_only, fp);
    AFTaskUnlockFile(file, 0, read_only);

    if (axf_builtin(file, tstbuf, n_read, fstat_buf, filetype))
        return(0);

    f_depth = n_read;
    buf_p = tstbuf;
    n_ascii = 0;

    /* Do not use isascii(): based on discussion on Jan-17-94. -- sanjay */
    /* This test does not jive with UNIX file(1). */

#if BUILD_ORIENT   /* {  */
     /* Ashisuto added for judge multi-byte chars external files.    95.10.06 */
    for (cnt = 0; cnt < f_depth; buf_p++, cnt++)
        if (is_external_ascii_multi_byte(buf_p))
#if BUILD_KANJI
          if (ascii_is_shift_jis() && (mbchar)*buf_p >= 0xa1
                  && (mbchar)*buf_p <= 0xdf)  /* single byte case */
            n_ascii++;
          else
#endif
            {
            n_ascii += 2;
            buf_p++;
            cnt++;
            }
        else if (is_Latin_1(*buf_p))
            n_ascii++;
        else if (n_ascii == f_depth-1) {
            /* ignor judging of last unrecognized byte because
             * it might be half of a multi-byte char.
             */
            n_ascii++;
            break;
        } else
            break;
#else                     /*  }  {  */
    for (buf_p = tstbuf, cnt = 0; (cnt < f_depth) && (is_Latin_1(*buf_p));
		 buf_p++, cnt++)
       {  n_ascii++;  }
#endif                    /*  }  BUILD_ORIENT  */
	
    /* Determine the default. ascii/data */
    if (n_ascii == f_depth)
        strcpy(defaulttype, FTIID_ASC); /* ascii */
    else 
        strcpy(defaulttype, FTIID_DAT); /* data */

/* Look for matching strings */

/*
**
**
Cases involving additional tests:
l, m, n, o, p are arbitrary offsets.
Case #1:
                   found_match
l   string  ...      FALSE   No add'l tests required - skip all add'l entries
>m  string  ...      (skip)
>n  string  ...      (skip)
>o  string  ...      (skip)
p   string  ...

Case #2:
                   found_match
l   string  ...      TRUE
>m  string  ...      FALSE   Evaluates to a mismatch
>n  string  ...      (skip)
>o  string  ...      (skip)
p   string  ...
Case #3:
                   found_match
l   string  ...      TRUE
>m  string  ...      TRUE
>n  string  ...      TRUE
>o  string  ...      FALSE   Evaluates to a mismatch
p   string  ...
Case #4:
                   found_match
l   string  ...      TRUE
>m  string  ...      TRUE
>n  string  ...      TRUE
>o  string  ...      TRUE    Evaluates to a successful test.
p   string  ...
**
**
*/

    filetype[0] = '\0';
    found_match = FALSE;
    ent_cnt = 0;
    while ((ent_cnt < NumMagicEntries) && (found_match==FALSE)) {
        buf_p = tstbuf;
        if ((ent_cnt+1) < NumMagicEntries)
            has_addltests = (MagicTable[ent_cnt+1].level > 0) ? TRUE : FALSE;
        else
            has_addltests = FALSE;
        /* Is this entry an additional test itself?
           Note: ``is_addltest'' must *always* evaluate to FALSE at
           this stage.  ``is_addltest'' TRUE at this stage means that
           the first line in the magic file has a '>', which is not
           permitted */
        is_addltest = (MagicTable[ent_cnt].level>0) ? TRUE : FALSE;
        if (is_addltest) {  /* botch! Skip this entry */
            ent_cnt++;
            continue;
        }
        found_match = axf_matchent(file, buf_p, ent_cnt, matched_str, f_depth);
        if (!is_addltest && has_addltests && found_match == FALSE) {
            /* skip */
            ent_cnt++;
            while ((ent_cnt < NumMagicEntries) &&
                   (MagicTable[ent_cnt].level > 0))
                ent_cnt++;

            ent_cnt--;  /* ent_cnt is always incremented at the bottom */
        }
        if (found_match && has_addltests) {
            ent_cnt++;  /* this one *is* an addl test. guranteed */
            while (ent_cnt < NumMagicEntries) {
                found_match = axf_matchent(file, buf_p, ent_cnt,
                                           matched_str, f_depth);
                if (found_match == FALSE) {
                    /* skip the rest of addl tests(if any) */
                    while ((ent_cnt < NumMagicEntries)&&
                           (MagicTable[ent_cnt].level > 0))
                        ent_cnt++;

                    ent_cnt--;
                    break;
                }
                if ((ent_cnt + 1) < NumMagicEntries)
                    if (MagicTable[ent_cnt + 1].level == 0)
                        break;
                ent_cnt++;
                /* if found_match is TRUE, continue making addl tests */
            }
        }
        if (found_match)
            strcpy(filetype, matched_str);

        ent_cnt++;  /* Next entry */
    }
    if (found_match == FALSE)
        strcpy(filetype, defaulttype);
    return(0);
}

/* Match a buffer pointed to by `buf_p' with entries in the magic-file */
static int
axf_matchent(file, bufp, ent_cnt, matched_str, f_depth)
char *file;
register char *bufp;
int  ent_cnt;
char *matched_str;
int  f_depth;
{
    char prep_str[100];
    int  prepstr_len;
    int  file_len;   /* length of arg 'file' */
    int  cnt;        /* general purpose counter */
    long long_val;
    uint16 short_val;
    uint8  byte_val;
    int  found_match = FALSE;

    *matched_str = '\0';
    switch (MagicTable[ent_cnt].vtype) {
    case VT_STRING: /* look for a matching string */
        if ((MagicTable[ent_cnt].offset * sizeof(char)) > f_depth)
            break;
        bufp += (MagicTable[ent_cnt].offset * sizeof(char));
        axf_prepstr(prep_str, MagicTable[ent_cnt].value.v_string,
                    &prepstr_len);
        if (memcmp(bufp, prep_str, prepstr_len) == 0)
            found_match = TRUE;
        break;
    case VT_EXTEN:  /* look at file extension */
        file_len = strlen(file);
        found_match = TRUE;
        if (file_len > MagicTable[ent_cnt].v_exten_len) {
            for (cnt = MagicTable[ent_cnt].v_exten_len; cnt >= 0; cnt--)
                if (file[file_len--] != MagicTable[ent_cnt].value.v_exten[cnt])
                    found_match = FALSE;
        } else {
            found_match = FALSE;
        }
        break;
    case VT_LONG:
        if ((MagicTable[ent_cnt].offset * sizeof(long)) > f_depth)
            break;
            bufp += (MagicTable[ent_cnt].offset * sizeof(char));
        memcpy((long *) &long_val, bufp, sizeof(long));
        switch((char) MagicTable[ent_cnt].oper) {
        case OP_EQ:
            if (long_val == MagicTable[ent_cnt].value.v_fourb)
                found_match = TRUE;
            break;
        case OP_GT:
            if (long_val > MagicTable[ent_cnt].value.v_fourb)
                found_match = TRUE;
            break;
        case OP_LT:
            if (long_val < MagicTable[ent_cnt].value.v_fourb)
                found_match = TRUE;
            break;
        case OP_ANY:
            sprintf(matched_str, MagicTable[ent_cnt].outstr, long_val);
            found_match = TRUE;
            break;
        case OP_NOT:
#if 0
            if (long_val != MagicTable[ent_cnt].value.v_fourb)
                found_match = TRUE; /* FIXME, get this clarified */
#endif
            break;
        default:
            break;  /* silent */
        }
        break;

    case VT_SHORT:
        if ((MagicTable[ent_cnt].offset * sizeof(uint16)) > f_depth)
            break;
        bufp += (MagicTable[ent_cnt].offset * sizeof(char));
        memcpy((uint16 *) &short_val, bufp, sizeof(uint16));
        if (MagicTable[ent_cnt].mask_oper != ' ') {
            switch (MagicTable[ent_cnt].mask_oper) {
            case MASK_AND:
                short_val &= MagicTable[ent_cnt].sh_mask;
                break;
            default:
                break;  /* error cond */
            }
        }
        switch((char) MagicTable[ent_cnt].oper) {
        case OP_EQ:
            if (short_val == MagicTable[ent_cnt].value.v_twob)
                found_match = TRUE;
            break;
        case OP_GT:
            if (short_val > MagicTable[ent_cnt].value.v_twob)
                found_match = TRUE;
            break;
        case OP_LT:
            if (short_val < MagicTable[ent_cnt].value.v_twob)
                found_match = TRUE;
            break;
        case OP_ANY:
            sprintf(matched_str, MagicTable[ent_cnt].outstr, short_val);
            found_match = TRUE;
            break;
        case OP_NOT:
#if 0
            if (short_val != MagicTable[ent_cnt].value.v_twob)
                found_match = TRUE;
#endif
            break;
        default:
            break;
        }
        break;

    case VT_BYTE:
        if ((MagicTable[ent_cnt].offset * sizeof(uint8)) > f_depth)
            break;
        bufp += (int) (MagicTable[ent_cnt].offset *  sizeof(uint8));
        memcpy((uint8 *) &byte_val, bufp, sizeof(uint8));
        if (MagicTable[ent_cnt].mask_oper != ' ') {
            switch (MagicTable[ent_cnt].mask_oper) {
            case MASK_AND:
                byte_val &= MagicTable[ent_cnt].mask;
                break;
            default:
                break;  /* error cond */
            }
        }
        switch((char) MagicTable[ent_cnt].oper) {
        case OP_EQ:
            if (byte_val == MagicTable[ent_cnt].value.v_oneb)
                found_match = TRUE;
            break;
        case OP_GT:
            if (byte_val > MagicTable[ent_cnt].value.v_oneb)
                found_match = TRUE;
            break;
        case OP_LT:
            if (byte_val < MagicTable[ent_cnt].value.v_oneb)
                found_match = TRUE;
            break;
        case OP_ANY:
            sprintf(matched_str, MagicTable[ent_cnt].outstr, byte_val);
            found_match = TRUE;
            break;
        case OP_NOT:
#if 0
            if (byte_val != MagicTable[ent_cnt].value.v_oneb)
                found_match = TRUE;
#endif
            break;
        default:
            break;
        }
        break;

    default:    /* unknown VT_ type */
        break;
    }
    if (found_match && matched_str[0] == '\0')
        strcpy(matched_str, MagicTable[ent_cnt].outstr);
    return(found_match);
}

/* Determine if is a unix-special-file (block-device, socket, link...) */
static int
axf_prtspecial(file, f_buf, f_type)
char *file;
struct stat *f_buf;
char *f_type;
{
    unsigned short  fmode;
    char realname[STRPATH + 1];
    int  namelen;

    if (f_buf == (struct stat *) NULL)
        return(0);
    fmode = f_buf->st_mode;

    switch (fmode & S_IFMT) {
    case S_IFDIR:
        strcpy(f_type, FTIID_DIR);
        break;
    case S_IFCHR:
        strcpy(f_type, FTIID_CHR);
        break;

#if AX_X11         /*   {  these types do not exist on Windows/NT */
    case S_IFBLK:
        strcpy(f_type, FTIID_BLK);
        break;
#if !X386_NEW
    case S_IFLNK: 
        if ((namelen = readlink(file, realname, STRPATH)) == -1)
			return(-1);
        realname[namelen] = '\0';
        strcpy(f_type, FTIID_LNK);
        break;
    case S_IFSOCK:
        strcpy(f_type, FTIID_SOCK);
        break;
#endif
    case S_IFIFO: 
        strcpy(f_type, FTIID_FIFO);
        break;
#endif                  /*  }  AX_X11  */
    default:
        strcpy(f_type, FTIID_UKN);
        break;
    }
    return(0);
}

/* Determine if file-type is one of the builtin types.  (builtin types
   do not have entries in the magic-file) */
static int
axf_builtin(file, buf, buf_len, f_statbuf, f_type)
char *file;
int  buf_len;
char *buf;
struct stat *f_statbuf;
char *f_type;
{
	/* TODO, c-program text, English text, scripts, core-files,
	   machine exec files... */

    strcpy(f_type, " " );    /* a default, "not known" value */

    if (axf_isMiscDoc(file, buf, buf_len, f_type) == TRUE)
        return(TRUE);
    if (axf_isvfont(file, buf, buf_len, f_type) == TRUE)
        return(TRUE);
    if (axf_isAxImage(file, buf, buf_len, f_type, f_statbuf) == TRUE)
        return(TRUE);
/*    if (axf_isMacHdr(file, buf, buf_len, f_type) == TRUE)
        return(TRUE); */
    return(FALSE);
}

/* Determine if buffer is from a vfont description */
/*ARGSUSED*/
static int
axf_isvfont(file, buf, buf_len, f_type)
char *file;
int  buf_len;
char *buf;
char *f_type;
{
    struct header vf_hdr;
    int isvfont;

    if (buf_len < sizeof(struct header))
        return(FALSE);
    memcpy(&vf_hdr, buf, sizeof(struct header));
    isvfont = (vf_hdr.magic == VFONT_MAGIC);
    if (!isvfont)
        return(FALSE);
    strcpy(f_type, FTIID_VFONT);
    return(TRUE);
    /*NOTREACHED*/
}

/* Determine if buffer is from one of some misc types.
   Currently, wordperfect and msword */
/*ARGSUSED*/
static int
axf_isMiscDoc(file, buf, buf_len, f_type)
char *file;
int  buf_len;
char *buf;
char *f_type;
{
    int retval = FALSE;
    int WPversion;
#if 0
    if (buf_len >= 11) {    /* check if doc is WordPerfect 5.x */
        retval = (((unsigned char) buf[0] == 255) && (buf[1] == 'W') && 
                  (buf[2] == 'P') && (buf[3] == 'C') && (buf[8] == 1) && 
                  (buf[9] == 10));
        if (retval) {
            /* strcpy(f_type, DOC_WPERFECT); */
            WPversion = buf[11];    /* 0 for 5.0, 1 for 5.1 */
			if (WPversion == 0) 
				sprintf(f_type, "%d", AX_DOC_TYPE_WP_50);
			else if (WPversion == 1)
				sprintf(f_type, "%d", AX_DOC_TYPE_WP_51);
			else
				sprintf(f_type, "%d", AX_DOC_TYPE_WP_42);
            return(TRUE);
        }
    }

    if (buf_len >= 7) {     /* check if doc is Microsoft Word */
        retval = ((buf[0] == 49) && ((unsigned char) buf[1] == 190) && 
                  ((unsigned char) buf[5] == 171) && (buf[2] == 0) && 
                  (buf[6] == 0) && (buf[7] == 0));
        if (retval) {
			sprintf(f_type, "%d", AX_DOC_TYPE_MS_WORD); /* ms-word */
            return(TRUE);
        }
    }
#endif
    return(FALSE);
}

/* Return TRUE if buffer is from an Aster*x Image File; FALSE
   otherwise. */
static int
axf_isAxImage(file, buf, buf_len, f_type, f_sbuf)
char *file;
int  buf_len;
char *buf;
char *f_type;
struct stat *f_sbuf;
{
#define IM_HEADER_SIZE 6
    unsigned char usbyte0;
    unsigned char usbyte1;
    int magic;
    short height, width;
    short bytewidth;

    if (buf_len < IM_HEADER_SIZE)
        return(FALSE);
    usbyte0 = (unsigned char )buf[0];
    usbyte1 = (unsigned char )buf[1];
    magic   = (usbyte1 << 8) | (usbyte0);
    if ((magic != 0) && (magic != 1))
        return(FALSE);

    usbyte0 = (unsigned char) buf[2];
    usbyte1 = (unsigned char) buf[3];
    width   = (usbyte1 << 8) | (usbyte0);

    usbyte0 = (unsigned char) buf[4];
    usbyte1 = (unsigned char) buf[5];
    height  = (usbyte1 << 8) | (usbyte0);

    bytewidth = (width + 7) / 8;
    if (f_sbuf->st_size == ((bytewidth * height) + IM_HEADER_SIZE)) {
        strcpy(f_type, DOCIID_AXIMAGE);
        return(TRUE);
    }
    return(FALSE);
}

static int
rc_read32bits( buf, value )
char *buf;
int *value;
{
    /** we read 4 bytes of integer in 68000 order, and then put
    ** them into the return value correctly for this machine */

    *value = ((long)buf[0]) << 24;
    *value = *value + (((long)buf[1]) << 16);
    *value = *value + (((long)buf[2]) << 8);
    *value = *value + buf[3];
    return(0);
}

static int
rc_read_disk_header( buf, header )
char *buf;
diskheader *header;
{
    /*
	 * read the fields of a header off of disk and put them into
	 * the internal structure.  The in-memory structure may differ
	 * from the disk image due to byte swapping and structure padding
	 */
	int tmpint;

    rc_read32bits(buf, &tmpint);
	header->disk_magic = tmpint;
    return(0);
}

/* Return TRUE if buffer has a Mac header, else FALSE */
static int
axf_isMacHdr(file, buf, buf_len, f_type)
char *file;
int  buf_len;
char *buf;
char *f_type;
{
#define APPLESINGLE 0x00051600
#define APPLEDOUBLE 0x00051607

    diskheader      disk;   /* Header info for Apple single or double */
                            /*  format file */
    int             size;   /* Size returned by read. */
    int             status;

    if ((status = rc_read_disk_header(buf, &disk)) == -1)
        return(FALSE);

    if ( (disk.disk_magic == APPLESINGLE) ||
	  (disk.disk_magic == APPLEDOUBLE) )
	{
        strcpy(f_type, FTIID_MACHDR);
        return(TRUE);
	}
    else
		return(FALSE);
}

/* Substitute '%' chars from the magic-string with %% so that
   sprintf() does not trip on them.
   "xxx%dxx" -> "xxx%%dxx"
   "xxx%xxx" -> "xxx%xxx"
   "xxx\abc" -> "xxxabc"
   "xxx\\abc" -> "xxx\abc"
   "xxx\004xxx" -> "xxx<char equiv of octal 004>xxx"
   
   returns: length of `outstr' */
static int
axf_prepstr(outstr, fmt, outstrlen)
register char *outstr;/*str size should be (2*strlen(fmt))+1 for worst case*/
register char *fmt;
register int  *outstrlen;
{
    register char c;
    register char *p;
    register char *fmt_p;
    register char *buf_p;
    char buf[50];
    char nxtch;

    *outstrlen = 0;
    p = outstr;
    while (c = *fmt++) {
        nxtch = *fmt;
        switch (c) {
        case '%':
            switch (nxtch) {
            case 'd':/* printf style conversion char %[dco] */
            case 'c':
            case 'o':
                *p++ = '%'; *p++ = '%'; *p++ = nxtch; 
                fmt++;
                break;
            default:
                *p++ = '%';
                break;
            }
            break;
        case '\\':  /* Escape Sequences */
            switch (nxtch) {
            case 'n': *p++ = '\n'; fmt++; break; /* newline (line feed) */
            case 't': *p++ = '\t'; fmt++; break; /* tab */
            case 'b': *p++ = '\b'; fmt++; break; /* backspace */
            case 'f': *p++ = '\f'; fmt++; break; /* formfeed */
            case 'r': *p++ = '\r'; fmt++; break; /* carriage return */
            case '\\': *p++ = '\\'; fmt++; break; /* literal '\' */
            case '0':
                if (!isdigit(*(fmt+1))) { /* null char */
                    *p++ = '\0';
                    fmt++;
                } else {
                    for (buf_p = buf, fmt_p = fmt; isdigit(*fmt_p); fmt_p++)
                        *buf_p++ = *fmt_p;
                    *buf_p = '\0';
                    *p++ = (char)((unsigned short)strtol(buf,(char **)NULL,8));
                    fmt = fmt_p;
                }
                break;
            default:
                if (!isdigit(nxtch)) {  /* any other literal char 'c' */
                    *p++ = nxtch;
                    fmt++;
                } else {    /* octal value */
                    for (buf_p = buf, fmt_p = fmt; isdigit(*fmt_p); fmt_p++)
                        *buf_p++ = *fmt_p;
                    *buf_p = '\0';
                    *p++ = (char)((unsigned short)strtol(buf,(char **)NULL,8));
                    fmt = fmt_p;
                }
                break;
            }
            break;
        default: *p++ = c; break;
        }
    }
    *p = '\0';
    *outstrlen = p - outstr;
    return(*outstrlen);
}


/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *						in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * comment-column: 26
 * end:
 */
