/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: recogniz.h
   Applix ASC version: 9
   Date: Tue Feb  2 15:21:09 1999
*/
#ifndef _recogniz_h
#define _recogniz_h

#define COMMENTCHAR '#'
#define AXF_MAGIC   "axmagic"

#define TOKENLEN    100
#define AXF_BUFSIZ  2048 /* size of test buffer (and line buffers) */

/* #define MAXCACHEENT 300 */

#define SKIPWS(p) while ((*p == ' ') || (*p == '\t')) p++
#define IS_DELIM(c) (((c == ' ') || (c == '\t') || (c == '\0')) ? 1 : 0)

/* value types: VT_ */
#define VT_STRING   1   /* string */
#define VT_LONG     2   /* long */
#define VT_BYTE     3   /* byte */
#define VT_SHORT    4   /* short */
#define VT_EXTEN    5   /* file extension */

/* value type strings: VTS_ */
#define VTS_STRING  NXLT("string")
#define VTS_LONG    NXLT("long")
#define VTS_SHORT   NXLT("short")
#define VTS_BYTE    NXLT("byte")
#define VTS_EXTEN   NXLT("extension")

/* token types: TK_ */
#define TK_NOMORE   0
#define TK_OFFSET   1
#define TK_VALTYPE  2
#define TK_VALUE    3
#define TK_MAJMIN   4
#define TK_INFOSTR  5

/* operators: OP_ */
#define OPER_LIST   NXLT("=><x&^")
#define OP_EQ   '='
#define OP_GT   '>'
#define OP_LT   '<'
#define OP_ANY  'x'
#define OP_NOT  '^'

/* byte mask operators: MASK_ */
#define MASK_LIST   "&"
#define MASK_AND    '&'

/* file types: FT_ */
#define FT_DIR  NXLT("directory")
#define FT_CHR  NXLT("character special")
#define FT_BLK  NXLT("block special")
#define FT_LNK  NXLT("symbolic link")
#define FT_SOCK NXLT("socket")
#define FT_FIFO NXLT("fifo")
#define FT_UKN  NXLT("Unknown")
#define FT_NIL  NXLT("empty")
#define FT_ASC  NXLT("ascii text")
#define FT_DAT  NXLT("data")
#define FT_CMD  NXLT("command text")

/* file type internal-id in string form: FTIID_ */
#define FTIID_DIR   	"160"
#define FTIID_CHR   	"161"
#define FTIID_BLK   	"162"
#define FTIID_LNK   	"163"
#define FTIID_SOCK  	"164"
#define FTIID_FIFO  	"165"
#define FTIID_UKN   	"166"
#define FTIID_NIL   	"167"
#define FTIID_ASC   	"0"
#define FTIID_DAT   	"151"
#define FTIID_MACHDR    "170"	/* has Mac Finder header */
#define FTIID_CMD   	FTIID_ASC
#define FTIID_EXEC  	"171"   /* machine object/executable */
#define FTIID_CORE  	"172"   /* core file */
#define FTIID_VFONT 	"173"   /* vfont definition */

/* misc document types: DOC_ */
#define DOC_WPERFECT    NXLT("WordPerfect document")  /* 5.x */
#define DOC_MSWORD      NXLT("Microsoft Word document")
#define DOC_AXIMAGE     NXLT("Aster*x Image")

/* misc document type id: DOCIID_ */
#define DOCIID_WPERFECT5X   "5" /* word-perfect 5.[01] */
#define DOCIID_MSWORD       "6" /* ms-word */
#define DOCIID_WPERFECT     "8" /* word-perfect 4.2 */
#define DOCIID_AXIMAGE      "206"   /* Aster*x Image */

/* magic-file entry-structure type */
typedef struct  magic_table_t {
    int     level;          /* currently unused */
    long    offset;         /* offset from beginning of file */
    char    oper;           /* operator to apply '=' OP_ */
    int     vtype;          /* type of value to look for VT_ */
    union { /* depends on vtype */
        char    *v_string;  /* string value */
        char    *v_exten;   /* file extension */
        long    v_fourb;    /* long value (4 byte) */
        uint16 v_twob;     /* short value (2 byte) */
        uint8   v_oneb;     /* char value (1 byte) */
    } value;                /* value: depends on vtype */
    uint8  mask;           /* in case of byte values */
    unsigned short sh_mask; /* mask in case of short values */
    int     v_str_len;      /* length of value string */
    int     v_exten_len;    /* length of file extension */
    char    mask_oper;      /* mask operator MASK_ */
    int     major_rev;      /* major revision of file */
    int     minor_rev;      /* minor revision of file */
    char    *outstr;        /* string to output */
} magic_table_t;

/* Aster*x specific defines.
   The following defines are based on the defines found in
   "pgms/ax/thimio.c".*/
#define AX_MAGICSTR "2278791755874641370"   /* Asterix magic string */
#define AX_RESERVED "00000000000000000000"  /* All zeros as of this rev */

/* Supported machine architectures: AX_ARCH_: */
#define AX_ARCH_PMAX        "0001"
#define AX_ARCH_SUN4        "0002"
#define AX_ARCH_HP300       "0003"
#define AX_ARCH_SUN3        "0004"
#define AX_ARCH_MIPS        "0005"
#define AX_ARCH_X386_NEW    "0006"
#define AX_ARCH_SGI         "0007"
#define AX_ARCH_HP800       "0008"
#define AX_ARCH_M_88K       "0012"
#define AX_ARCH_UNIV        "0000"  /* universal (portable format) */
#define AX_ARCH_UNIV_2_1    "1000"  /* universal (portable format) 2.1 */
#define AX_ARCH_STRLEN      4   /* strlen of "AX_ARCH_"s */

/* Aster*x document header line specifics */
#define AX_HEADERLEN    78  /* includes '\n' char at the end */

/* AX_IDX_: index into the header string. */
#define AX_IDX_SYSREV   21  /* system revision stamp */
#define AX_IDX_DOCTYP   26  /* document type */
#define AX_IDX_APPREV   31  /* application revision stamp */
#define AX_IDX_FILCNT   52  /* file count */
#define AX_IDX_RSVD     57  /* reserved */
#define AX_IDX_DOCTYPE  29  /* document type */

/* Aster*x document types: AXDOCT_ */
#define AX_DOCT_WORDS       '6'
#define AX_DOCT_SS          '2'
#define AX_DOCT_GRAPHIC     '5'
#define AX_DOCT_ELFMACRO    '8'
#define AX_DOCT_S_AX        NXLT(" Asterix")
#define AX_DOCT_S_DOC       NXLT(" Document")
#define AX_DOCT_S_WORDS     NXLT(" Words")
#define AX_DOCT_S_SS        NXLT(" Spreadsheet")
#define AX_DOCT_S_GRAPHIC   NXLT(" Graphics")
#define AX_DOCT_S_BG        NXLT(" Business Graph")
#define AX_DOCT_S_ELFMACRO  NXLT(" Macro")

/* Supported machine architectures (in string form): AX_ARCH_S_: */
#define AX_ARCH_S_PMAX      NXLT(" dec-risc")
#define AX_ARCH_S_SUN4      NXLT(" sparc")
#define AX_ARCH_S_HP300     NXLT(" hp300")
#define AX_ARCH_S_SUN3      NXLT(" sun3")
#define AX_ARCH_S_MIPS      NXLT(" mips")
#define AX_ARCH_S_X386_NEW  NXLT(" 386")
#define AX_ARCH_S_SGI       NXLT(" sgi")
#define AX_ARCH_S_HP800     NXLT(" hp-risc")
#define AX_ARCH_S_M_88K     NXLT(" M_88K")
#define AX_ARCH_S_UNIV      NXLT(" portable")/* universal (portable format) */

#endif  /* !_recogniz_h */
