/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: aosxlate.c
   Applix ASC version: 88
   Date: Mon Jan 11 21:14:05 1999
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdlib.h>

#if AXMS
#include <windows.h>
#include <windowsx.h>
#include <tchar.h>
#endif


#ifdef __cplusplus
extern "C" {
    int LoadGlossary(char*, int);
    int AppendGlossary( char *filename );
    void XltExceptionsLog(int, char*, int);
    char *XLT( char*);
    char *XLT2(char*, char*);
    char *ReverseString(char*);
    int string_is_all_digits( char *str );
    void DoTranslateList( char *tag_str, char *fieldvalue,
                                             char *new_field, char seperator );
    void DoTranslateSQLStatement( char *tag_str, char *fieldvalue,
                                                   char *new_sql, int bad_ok );
    char *Trans_3( char *form_name, char *tag, char *tag2, char *str );
    char *Trans_2( char *form_name, char *tag, char *str );
    char *TransComplexThenSimple( char *tag, char *str );
    void PushLogOff();
    void PushLogOn();
    void PopLog();
}
#else
    int LoadGlossary(char*, int);
    int AppendGlossary( char *filename );
    void XltExceptionsLog(int, char*, int);    char *ReverseString(char*);
    int string_is_all_digits( char *str );
    void DoTranslateList( char *tag_str, char *fieldvalue,
                                             char *new_field, char seperator );
    void DoTranslateSQLStatement( char *tag_str, char *fieldvalue,
                                                   char *new_sql, int bad_ok );
    char *Trans_3( char *form_name, char *tag, char *tag2, char *str );
    char *Trans_2( char *form_name, char *tag, char *str );
    char *TransComplexThenSimple( char *tag, char *str );
    void PushLogOff();
    void PushLogOn();
    void PopLog();
#endif


#define BUILDING_AOSXLATE 1
#include <aos.h>

#define TR_HASH_PRIME 1009 /* Hash basis (must be a prime number) */


/* I am putting the following here, so that when things are built on */
/* a standalone basis (that is, not into axmain) we will use the "good" */
/* allocator on NT - jon                                                */
#if USE_GLOBAL_ALLOC       /*  { */
#define MALLOC(size) (char *)GlobalAlloc( GMEM_FIXED, size )
#define CALLOC(num,size) (char *)GlobalAlloc( (GMEM_FIXED | GMEM_ZEROINIT), size )
#define REALLOC( ptr, new_size ) (char *)GlobalReAlloc( ptr, new_size, GMEM_MOVEABLE );
#define FREE( ptr ) GlobalFree( ptr )
#else                   /* }  {  */
   /* use regular "malloc", which may have been defined to something else */
#define CALLOC calloc
#define MALLOC malloc
#define FREE free
#endif                  /* }  */




typedef struct name_entry {
	struct name_entry *next;
	unsigned char *name;		/* English */
	unsigned char *local;	/* Local */
} name_entry;


static name_entry *NameTab[TR_HASH_PRIME+1];
static int TrEnglishTrLocal_Enabled = FALSE;
static FILE *exceptions_FP = (FILE *)NULL;
static int Log_Exceptions = FALSE;
static void log_exception();
unsigned char *xlt_ReadWholeFile (char *filename, int *sizeP, int *linesP);
unsigned char *next_xlt_line( char *lineP );



int LoadGlossary( char *filename, int reload_gloss );
static void convert_escape_sequences( char *str );
static int xlHash ( unsigned char *s);
int TransType = 0;
int aos_Verbose = 0;
int ConvertEscapeSequences = TRUE;
static int save_log;
static unsigned char *locBuf = (unsigned char *)NULL;
static name_entry *StringPtrs = (name_entry *)NULL;


#define MAX_XLT_STR 2500
#define MAX_XLT_TAG 200


/* save the current log flag, and then turn it off temporarily */
void PushLogOff()
{
    save_log = Log_Exceptions;
    Log_Exceptions = FALSE;
    return;
}

/* save the current log flag, and then turn it on temporarily */
void PushLogOn()
{
    save_log = Log_Exceptions;
    Log_Exceptions = TRUE;
    return;
}



/* restore the verbose flag */
void PopLog()
{
    Log_Exceptions = save_log;
}


int string_is_all_digits( char *in_str )
{
    char *str;
    char buff[50];

    strncpy( buff, in_str, sizeof(buff)-1 );
    buff[sizeof(buff)-1] = '\0';

    str = buff;
    while ( *str != '\0' )
       {
       if ( (*str == ' ') || (*str == '-') )
          {
          str++;
          continue;
          }
       if ( (*str < '0') || (*str > '9') )
          return( FALSE );
       str++;
       }
    return( TRUE );
}


/*************************************************************************
function: XLTInit - Load English & Loal database
Returns 0 on success -1 on fail.
*/
void XLTInit (char *langDir, char *suffix)
{
	char filename[STRPATH+1];

    TrEnglishTrLocal_Enabled = FALSE;
    sprintf(filename,"%s/trLocal.%s", langDir, suffix );

    LoadGlossary( filename, 0 );
    return;
}


void CreateBasicGlossaryStructs()
{
    memset( NameTab, 0, sizeof(NameTab) );

    TrEnglishTrLocal_Enabled = TRUE;
}



/* Load a glossary file into memory.  This is used to drive the XLT  */
/* functionality.   The general form of a glosssary line is:          */
/* <English-string>||<translated-string>||<optional info>             */

int LoadGlossary( char *filename, int reload_gloss )
{
    int ret;

	if (reload_gloss)
	{
		/*Free this memory only if demanded. Should be used only during build */
		if ( locBuf != (unsigned char *)NULL )
			 /* we are re-loading, free the old one */
			 FREE( locBuf );
		if ( StringPtrs != (name_entry *)NULL )
			 FREE( StringPtrs );
		
		if (strcmp(filename, "") == 0)				/*	nothing to re-load */
			return( FALSE );
	}

    ret = AppendGlossary( filename );
    return( ret );
}


/* Append a glossary file to info already in memory.                  */

int AppendGlossary( char *filename )
{
	int ix;
	register unsigned char *p, *engP, *locP, *nextEng;
	int linecount = 0;
	int localLinecount = 0;
	register name_entry *vP;
	register int hashval;
	int locSize, locLines;
    int past_comments = FALSE;


    if ( !TrEnglishTrLocal_Enabled )
        {
        /* basic stuff not yet started up */
        CreateBasicGlossaryStructs();
        }

	locBuf = xlt_ReadWholeFile(filename,&locSize,&locLines);
    if ( locBuf == (unsigned char *)NULL )
       /* no "trLocal" ==> no translation */
       return( FALSE );

    /* create mem for hash StringPtrs */
    StringPtrs = (name_entry*)CALLOC(1, locLines*sizeof(name_entry));

    /* Load hash lists */
	engP = locBuf;
	for ( ix = 0; ix < locLines; ix++ ) 
        {
        if ( *engP == 0x02 )
           {
           /* we hit the special end marker */
           break;
           }

        nextEng = next_xlt_line( engP );      /* start of next line */
        if ( (*engP == '#') && !past_comments )
           {
           /* skip all leading lines which start with a pound-sign */
           /* as these are comment lines                           */
           engP = nextEng;
           continue;
           }
        past_comments = TRUE;

        /* find the end of the english portion, which is delimited by  */
        /*  a tab character                                            */
        p = engP;
        while ( (*p != '\t') && (*p != '\0') && (*p != 0x02) ) 
           { p++; }
        if ( (*p == '\0') || (*p == 0x02) )
           {
           /* hit the end of the string without finding a tab */
           /* an english string, with no translation, ignore */
           engP = nextEng;
           continue;
           }

        *p = '\0';   /* step on the tab */
        locP = p + 1;   /* start of translated string */
        /* find the end of the translated string, which is delimited by  */
        /*  another tab, or the end of the line                           */
        p = locP;
        while ( (*p != '\t') && (*p != '\0') && (*p != 0x02) )
           { p++; }
        if ( *p == '\t' )
           {
           /* there is some optional text at the end of the line */
           *p = '\0';    /* step on the tab */
           p = p + 1;    /* start of optional text */
           while ( *p != '\0' )
              {  p++;  }     /* find the end of the line */
           }

        if ( ConvertEscapeSequences )
           {
           /* defalut is to convert */
           convert_escape_sequences( engP );
           convert_escape_sequences( locP );
           }

		hashval = xlHash(engP);
		vP = &StringPtrs[ix];
		vP->name = engP;


        if ( strcmp( locP, NXLT("NULL-NULL") ) == 0 )
            /* special hook to allow translation to a null string  */
            locP = "";

		vP->local = locP;
		vP->next = NameTab[hashval];
		NameTab[hashval] = vP;
        engP = nextEng;
	    }
    return( TRUE );
}


/* Add a single string to the glossary */
void LoadStringIntoGlossary( char *tag_str, char *str, char *trans_str )
{
    char *space;
    char *strP;
    int tag_len, str_len, trans_str_len;
    int total_len;
    register name_entry *vP;
    register int hashval;
    
    if ( !TrEnglishTrLocal_Enabled )
        {
        /* no glossary yet, create it */
        CreateBasicGlossaryStructs();
        }

    /* allocate some space, held permenently, for the string data */
    if ( tag_str != (char *)NULL )
        tag_len = strlen( tag_str );
    else
        tag_len = 0;
    str_len = strlen( str );
    trans_str_len = strlen( trans_str );
    total_len = tag_len+4+str_len+1+trans_str_len+1;

    space = (char *)MALLOC( total_len );

    /* make up the tag string, if a tag was specified */
    if ( (tag_str != (char *)NULL) && (*tag_str != '\0') )
        /* tag specified, creat the compound string */
        sprintf( space, "%s|||%s", tag_str, str );
    else
        /* no tag, just use the basic string as the key */
        strcpy( space, str );

    /* put the translated version after the key-tag in the space allocated */
    strP = space + strlen( space ) + 1;
    strcpy( strP, trans_str );

    /* now add the string to the glossary */
    hashval = xlHash(space);
    vP = (name_entry *)MALLOC( sizeof(name_entry) );
    vP->name = space;                   /* original string */
    vP->local = strP;                   /* translated string */
    vP->next = NameTab[hashval];
    NameTab[hashval] = vP;

    return;
}



static void convert_escape_sequences( char *str )
{
    char *p, *p2;

    /* convert any 'C' style escape sequences to their real equivalents */
    p = p2 = str;
    while ( *p2 != (unsigned char)0 )
       {
       if ( (*p2 == '\\') && ( *(p2+1) == 'n' ) )
          {
          *p = '\n';
          p2++;
          }
       else if ( (*p2 == '\\') && ( *(p2+1) == 'r' ) )
          {
          *p = '\r';
          p2++;
          }
       else if ( (*p2 == '\\') && ( *(p2+1) == 't' ) )
          {
          *p = '\t';
          p2++;
          }
       else if ( (*p2 == '\\') && ( *(p2+1) == '\\' ) )
          {
          /* an escaped backslash */
          *p = '\\';
          p2++;
          }
       else if ( (*p2 == '\\') && ( *(p2+1) == '"' ) )
          {
          /* an escaped quote  (don't know why we have these - jon) */
          *p = '"';
          p2++;
          }
#if AXMS
       else if ( (*p2 == 0x0D) && !((*(p2+1) == '\\') && (*(p2+2) == 'n')) ) 
          {   
          /* found a ^M, which is not followed by an encoded newline */
          /* a leftover part of a CR/NL pair put on by a Windows editor */
          *p = 0;
          }
#endif
       else
          /* a normal character */
          *p = *p2;
       p++;
       p2++;
       }
    *p = (unsigned char)0;
    return;
}

/*************************************************************************
function: XLT - Given an English string, return local string
*/
#if AXMS_SETUP
/* if we are compiling this for Windows setup, we need to explicitly
 * declare the XLT funcs as being dll_export so that the InstallShield
 * script can resolve the functions (we can't use the .def file method
 * because we are dealing with a different language/compiler).
 */
DLL_EXPORT
#endif
char *XLT (str)
register char *str;
{
	register name_entry *vP;

    if ( MBstrlen(str) == 0 )
       return( (char *)str );

    if ( TrEnglishTrLocal_Enabled )
        {
	    for ( vP = NameTab[xlHash(str)]; vP; vP = vP->next ) 
            {
		    if ( MBstrcmp(str,vP->name) == 0)
                {
			    return( (char *)vP->local );
                }
	        }
        /** fell through , string not found "in trLocal"
         ** so things keep running, just return the original english  - jon  */
        if ( Log_Exceptions )
            log_exception( str );
        }


    return( (char *)str );
}

/*************************************************************************
function: XLT2 - Given an English str and a tag, return local string
TR2$(1,"Foo")  --> "1Foo" in english!
*/
#if AXMS_SETUP
/* if we are compiling this for Windows setup, we need to explicitly
 * declare the XLT funcs as being dll_export so that the InstallShield
 * script can resolve the functions (we can't use the .def file method
 * because we are dealing with a different language/compiler).
 */
DLL_EXPORT
#endif
char *XLT2 (tag,str)
char *tag;
char *str;
{
	register name_entry *vP;
	char buf[MAX_XLT_STR+MAX_XLT_TAG+1];	/* big enough? */
    int str_len, tag_len;


re_do_lookup:
    if ( TrEnglishTrLocal_Enabled )
        {
        str_len = MBstrlen(str);
        if ( str_len == 0 )
           return( str );

        tag_len = MBstrlen(tag);
        if ( str_len+tag_len+4 >= sizeof(buf) )
           {
           /* overflow condition */
           if ( Log_Exceptions )
              {
              strcpy( buf, NXLT("***XLT2 OVERFLOW*** <TAG>=") );
              strncat( buf, tag, 100 );
              strcat( buf, NXLT("\t<STR>=") );
              strncat( buf, str, 100 );
              log_exception( buf );
              }
           return( str );
           }

        sprintf(buf,"%s|||%s",tag,str);
    
        /** now look up this string  */
	    for ( vP = NameTab[xlHash(buf)]; vP; vP = vP->next ) 
            {
		    if ( MBstrcmp(buf,vP->name) == 0)
                {
                if ( strncmp( (char *)vP->local, "|||", 3 ) == 0 )
                    {
                    /* translated string begins with 3 pipe symbols */
                    /* this means that we are to take the rest of the */
                    /* line as a new tag, and actually get the string */
                    /* from that tag.                                 */
                    tag = (char *)vP->local + 3;
                    goto re_do_lookup;
                    }
			    return( (char *)vP->local );
                }
	        }

		/* fell through - since this is a contexted string, log exceptions
		 *	with context (chgd from "log_exception( str );" -- lhunt )*/

        if ( Log_Exceptions )
              log_exception( buf );
        }
    
	/** fell through , string not found "in trLocal"
	 ** so things keep running, just return the original english  - jon  */
    return( (char *)str );
}


#if 1
/*************************************************************************
function: XLT_SEQ - Given an English str, return local string
*/
char *XLT_SEQ (str)
char *str;
{
    return( (char *)XLT(str) );
}

/*************************************************************************
function: XLT_SEQ2 - Given an English str and a tag, return local string
XLT_SEQ2(1,"Foo",...)  --> search for "1|Foo" in english! --> return "Fuu" in german
*/
char *XLT_SEQ2 (tag,str)
char *tag;
char *str;
{
    return( XLT2(tag,str) );
}

#endif


/** take a "complex" string, which is composed of a tag, which  **
/**   is followed by 3 pipe symbols, which is followed by   **/
/**   the English equivalent, and return the translated string.  If   **/
/**   the input is not found in the translation files, just return the    **/
/**   English portion                                                     **/
/** Note: this does not change strings which do not contain the triple  **/
/**   pipe symbol delimiter between tag and string                   **/
/** Example strings:                                                 **/
/**   "abc|||xyz"   Returns translated version of "xyz", for the special */
/**                    case of "abc"                                 **/
/**   "efg"     returns  efg                    **/

/*	Search for '|||' from front of string, not back, because Equation Editor
 *	menu contains '|' and '||' for fences -- lhunt 12 aug 94
 */

unsigned char *XLT_Complex( str )
char *str;
{
    register unsigned char *p, *end;
	register name_entry *vP;

    if ( str == 0 )
        return( str );
    if ( MBstrlen(str) == 0 )
        return( str );

    /** see if this string contains the double pipe symbol */
	if( *str == '\0')
		return(str);	/* empty string */
    end = str + MBstrlen(str) - 3;     /* end of string */
    for ( p = str; p < end ; p++)
        {
        if ( (*p == '|') && (*(p+1) == '|') && (*(p+2) == '|') )
            {
            /** it does contain the double pipe, translate */
            if ( TrEnglishTrLocal_Enabled )
                {
                for ( vP = NameTab[xlHash(str)]; vP; vP = vP->next ) 
                    {
                    if ( MBstrcmp(str,vP->name) == 0)
                        {
                        /* found a translation */
                        return(vP->local);
                        }
                    }
                /* no translation found, just return the English string */
                if ( Log_Exceptions )
                    log_exception( str );
                }
            return( (p+3) );
            }
        }

    /* fell through, no triple pipe, just return original */
    return( str );
}


/** take a "complex" string, which is composed of a tag **
/**   is followed by 3 pipe symbols, followed by   **/
/**   the English equivalent, and return the English string.     **/
/** If the string does not contain the 3 pipe symbols in a row delimiter */
/**   then the input string is returned in its entirety        */

/*	Search for '|||' from front of string, not back, because Equation Editor
 *	menu contains '|' and '||' for fences -- lhunt 12 aug 94
 */

unsigned char *XLT_Complex_Strip( str )
char *str;
{
    register unsigned char *p, *end;

    if ( str == 0 )
        return( str );
    if ( *str == '\0' )
        return( str );

    end = str + MBstrlen(str) - 3;     /* end of string */
	for ( p = str; p < end ; p++)
        {
        if ( (*p == '|') && (*(p+1) == '|') && (*(p+2) == '|') )
           /** triple pipe found, return english */
           return( (p+3) );
        }
    
    /* fell through, no triple pipe, just return original */
    return( str );
}

/* move past this line to the next line in the buffer */
/* end of buffer is marked with a 0x02     */
unsigned char *next_xlt_line( char *lineP )
{
    if ( *lineP == 0x02 )
        /* already at the end of the buffer */
        return( lineP );

    /* point to char after trailing null */
    lineP = lineP + strlen(lineP) + 1;

    while ( (*lineP < ' ') && (*lineP != 0x02) )
        {
        /* skip any additional additional nulls left from stepping on */
        /* both CR and NL                                             */
        *lineP++;
         }

    return( lineP );
}





/*************************************************************************
function: xlt_ReadWholeFile - Load file into memory
*/
unsigned char *xlt_ReadWholeFile (filename,sizeP,linesP)
char *filename;
int *sizeP;	/* returned size */
int *linesP;	/* returned # lines */
{
	FILE *fp;
	struct stat status;
	char *retval;
	unsigned char *p, *endP;
	int lines = 0;

	errno = ENOENT;
#if AXMS	
	if ( NULL == (fp = fopen(filename,"rb")) ) 
#else
	if ( NULL == (fp = fopen(filename,"r")) ) 
#endif
        return( (unsigned char *)NULL );

    if ( -1 == stat(filename,&status) ) {
        fprintf(stderr,NXLT("%s: Cannot stat, errno=%d\n"),filename,errno);
        return( (unsigned char *)NULL );
    }

	if ( NULL == (retval = (unsigned char *)MALLOC(status.st_size+1)) ) 
        return( (unsigned char *)NULL );

	if ( 1 != fread(retval,status.st_size,1,fp) ) {
        fclose( fp );
		fprintf(stderr,NXLT("%s: Can not read, errno=%d\n"),filename,errno);
		return( (unsigned char *)NULL );
	}
	fclose(fp);

	*sizeP = status.st_size;
	p = retval;
	endP = p + status.st_size;
    *endP = 0x02;                  /* special end marker */

	while ( p < endP ) {
		if ( *p == '\n' ) {
            /* end of a line */
			lines++;
            while ( (*p < ' ') && (*p != '\t') )
               {
               /* trim any trailing CR/NL  */
               *p = 0;
               p--;
               }
		}
        else if ( *p == 1 )
            {
            /* we have encoded newlines in the glossary file as control A's */
            /*  (^A == 1).  Here we convert them back to newlines    */
            *p = '\n';
            }
		p++;
	}
	*linesP = lines;
	return(retval);
}

/*************************************************************************
function: xlHash - Comput a string's hash value

From Donald E. Knuth's (The Art of Programming) Searching & Sorting Algorithms
Uses only additions and subtractions - babu 12/5/94
*/
static int xlHash ( unsigned char *s)
{
	register int sum = 0;
	register int chr = 0;

	while (chr = *s++) {
		sum = sum + sum + (chr & 0xFF);
		while (sum >= TR_HASH_PRIME)
			sum -= TR_HASH_PRIME;
	}
	return(sum);
}




void XltExceptionsLog( int turn_on, char *filename, int append )
{
    if ( exceptions_FP != (FILE *)NULL )
        {
        /* already open, close this one */
        fflush( exceptions_FP );
        fclose( exceptions_FP );
        exceptions_FP = (FILE *)NULL;
        }

    if ( turn_on )
        {
		if ( append )
			exceptions_FP = fopen( filename, "a+" );
		else
			exceptions_FP = fopen( filename, "w+" );
        if ( exceptions_FP != (FILE *)NULL )
            Log_Exceptions = TRUE;
        }
    else 
        { /* close the exceptions file */
        if ( exceptions_FP != (FILE *)NULL )
            {
            fflush( exceptions_FP );
            fclose( exceptions_FP );
            exceptions_FP = (FILE *)NULL;
            }
        }
    return;
}


/*	Toggles flag WITHOUT truncating exceptions file
 *	Supports dialog box context translation, which does a double-pass
 *	by reporting failure only on the 2nd pass */

void XltExceptionsLogOnly( turn_on )
int turn_on;
{
	if ( turn_on )
		Log_Exceptions = TRUE;
	else
		Log_Exceptions = FALSE;
}

        
static void log_exception( str )
unsigned char *str;
{
    char outbuff[MAX_XLT_STR+MAX_XLT_TAG+100];
    char *outP, *inP;
    int out_size;


    if ( exceptions_FP == (FILE *)NULL )
        return;

    if ( (strlen( str ) <= 1) || string_is_all_digits( str ) )
        /* ignore 1 char strings and all numeric strings */
        return;

    /* if the exception string is composed entirely of punctuation, **/
    /*  we do not care about it                                     **/
    inP = str;
    while ( *inP != '\0' )
       {
       if ( ((*inP >= 'A') && (*inP <= 'z')) || ((*inP >= '0') && (*inP <=
	   '9')) )
           {
           if ( (*inP >= '[') && (*inP <= '`') )
               {
               /* there are a couple of punctuation chars in the middle */
               /*  of the alphabetic range which are ok                 */
               inP++;
               continue;
               }
           /* printable alphabetic, this is a real exception */
           break;
           }
       inP++;
       }

    if ( *inP == '\0' )
       /* hit end of string, string has no alphabetic, not exception */
       return;
 
    inP = str;
    outP = outbuff;
    out_size = 0;
 
    while ( *inP != '\0' )
        {
        /* convert newline and tab to the appropriate escape sequences */
        if ( *inP == '\n' )
            {
            *outP++ = '\\';
            *outP++ = 'n';
            out_size = out_size + 2;
            }
        else if ( *inP == '\t' )
            {
            *outP++ = '\\';
            *outP++ = 't';
            out_size = out_size + 2;
            }
        else
            {
            *outP++ = *inP;
            out_size++;
            }
        if ( out_size+1 >= sizeof(outbuff) )
            {
            /* overflow on output buffer, output OVERFLOW message, and */
            /* part of the string                                      */
            char temp_str[301];
            strncpy( temp_str, outbuff, 300 );
            temp_str[300] = 0;
            strcpy( outbuff, NXLT("***OVERFLOW*** <STR>=") );
            strcat( outbuff, temp_str );
            fprintf( exceptions_FP, "%s\n", outbuff );
            fflush( exceptions_FP );
            return;
            }
        inP++;
        }
    *outP = '\0';
 
    fprintf( exceptions_FP, "%s\n", outbuff );
    fflush( exceptions_FP );
    return;
}


char *ReverseString( char *old_string )
{
    static char new_string[300];
    char *p;

    if ( string_is_all_digits(old_string) )
       return( old_string );

   p = new_string + strlen(old_string);
   *p = '\0';
   p--;
   while ( *old_string != '\0' )
       {
       *p = *old_string;
       p--;
       old_string++;
       }

    return( new_string );
}




/* Translate a string, trying first for the fully qualified version, and */
/* if that is not found, try for the simple version without the tag      */

char *TransComplexThenSimple( char *tag, char *str )
{
    char *trans_str;

    if ( (str == (char *)NULL) || (*str == '\0') )
       return( str );
    if ( tag == (char *)NULL )
       tag = "";

    if ( Log_Exceptions && !aos_Verbose )
        {
        /* don't log fully qulified requests */
        XltExceptionsLogOnly( FALSE );
        trans_str = XLT2( tag, str );
        XltExceptionsLogOnly( TRUE );
        }
    else
        trans_str = XLT2( tag, str );

    if ( trans_str == str )
        {
        /* was not translated, look for a translation without the tag */
        trans_str = XLT( str );
        }

    if (  (TransType == 2) && (trans_str == str) )
        {
        /* no translation, reverse it */
        trans_str = ReverseString( str );
        }
    return( trans_str );
}



/* Break a string of delimiter seperated fields up into individual field, */
/* translate each part, and then re-build the string with the translated */
/* fields.  The input string may contain both the passed seperator, and  */
/* 0x02 characters as seperators.                                        */

void DoTranslateList( char *tag_str,     /* IN */
                      char *fieldvalue,  /* IN  string of fields */
                      char *new_field,   /* OUT */ 
                      char seperator  )
{
    char *p, *start_tok, *new_token;
    char *outP;
    char token[500];
    int done = FALSE;
    int tok_len;

    /* look for a seperator, translate the string, and ceate a new field string */
    *new_field = '\0';
    outP = new_field;
    p = start_tok = fieldvalue;
    
    while ( !done )
        {
        while ( (*p != '\0') && (*p != seperator) && (*p != 0x02) )
            {
#if AXMS
            p = _tcsinc(p);    /* correct increment on DBCS char string */
#else
            p++;
#endif
            }
        tok_len = p - start_tok;
        if ( tok_len > 0 )
            {
            /* get just the token and translate it */
            strncpy( token, start_tok, tok_len );
            token[tok_len] = '\0';
        
            new_token = TransComplexThenSimple( tag_str, token );
    
            strcpy( outP, new_token );
            outP = outP + strlen(outP);
            }
        *outP = *p;     /* the seperator, or 0x02 or null */
        if ( (*p != '\0') && (seperator == '\r') && (*(p+1) == '\n') )
           {
            /* seperator is a NL, move following CR as well */
           outP++;
           p++;
           *outP = *p;
           }
        outP++;

        if ( *p == '\0' )
            done = TRUE;
        else
            start_tok = p = p+1;
 
        if ( (*p != seperator) && (*p != 0x02) )
            p++;     /* next char past the seperator, test above handles */
        }            /* as of 2 seperators in a row                      */
    return;
}


void DoTranslateSQLStatement( char *tag_str, char *fieldvalue, 
                                   char *new_sql, int bad_ok )
{
    char *p, *start_tok, *new_token;
    char *outP, *strP;
    char token[500];
    char formname[500];
    int done = FALSE;
    int len, doing_token;

    /* look for quoted strings in the sql statement, and translate them */
    p = strP = fieldvalue;

    outP = new_sql;
    new_sql[0] = '\0';

    done = doing_token = FALSE;
    while ( !done )
        {
        while ( (*p != '\0') && (*p != '\'') )
           {
#if AXMS
           p = _tcsinc(p);    /* correct increment on DBCS char string */
#else
           p++;
#endif
           }

        /* we found a quote char, or the end of the string */
        if ( *p == '\0' )
            {
            /* end of string */
            if ( doing_token && !(bad_ok) )
                {
                /* hit the end of the string, but we were doing a quoted item */
                /* looks like broken, or wierd sql, bail out and just return  */
                /* original untranslated version                              */
                sprintf( new_sql, "BAD SQL <%s>", fieldvalue );
                if ( Log_Exceptions )
                   log_exception( new_sql );
                return; /*( fieldvalue ); causing compiler error */  
                }
            else
                {
                /* end of string, all done */
                if ( strP != (char *)NULL )
                   {
                   /* move any remaining non-quoted to the output */
                   len = p - strP;
                   strncpy( outP, strP, len );
                   outP = outP + len;
                   *outP = '\0';
                   }
                done = TRUE;
                }
            }
 
        else if ( *(p+1) == '\'' )
            {
            /* doubled single quote, this is an embedded quote, ignore it */
            p = p + 2;     /* next char past the double quote */
            }
        else if ( !doing_token )
            {
            /* start of a quoted token, note where the quoted string starts */
            /* and then move everything that has gone before to the output  */
            /* string.                                                      */
            doing_token = TRUE;
            start_tok = p+1;
            len = p - strP;
            strncpy( outP, strP, len );
            outP = outP + len;
            *outP = '\0';
            strP = (char *)NULL;    /* fail safe, should not use till reset */
            }
        else
            {
            /* we have found the end of the quoted token, translate it */
            /* and put the translated version into the output          */
            doing_token = FALSE;
            len = p - start_tok;
            strncpy( token, start_tok, len );
            token[len] = '\0';

			/*	set up the check for "form=".  If it matches, */
			/*	ignore this token; it shld be left in english. - lhunt  */
            strncpy( formname, start_tok - 6, 5 );
			formname[5] = '\0';

            if ( strcasecmp (formname, "form=") == 0 )
                {
                /* the token is an internal form name, no need to translate */
                new_token = token;
                }
            else if ( token[0] == '@' )
                {
                /* this is an "@sub" command, no need to translate */
                new_token = token;
                }
            else if ( (token[0] == '"') && (token[2] == '|') )
                {
                /* long special string for fulcrun, in AX_DB_SQL table */
                /* of the form " || EXTEND(x0.change_dt, YEAR TO DAY)"*/
                /* ignore it */
                new_token = token;
                }
            else if ( (token[1] == 0) &&
                      ((token[0] == 'd') || (token[0] == 'h') || 
                       (token[0] == 'q') || (token[0] == 'm')) )
                {
                /* single char string as part of date stuff, like:  */
                /* DatePart('h',LOGGED_DT) AS LOGGED_HOUR,   in table  */
                /* AX_DB_SQL,    ignore it                           */
                new_token = token;
                }
            else
                {
                new_token = TransComplexThenSimple( tag_str, token );
                }
            sprintf( outP, "'%s'", new_token );
            outP = outP + strlen(outP);
            strP = p+1;            /* start of next non-quoted stuff */
            doing_token = FALSE;
            }
        p++;     /* next char past the quote */
        }
           
     *outP = '\0';

}

#if AXMS && AX_DLL   /*  {   */

/************************************************************************/
/* These things are called from Visual Basic                            */
/* If you add any functions, or change the name of any functions, you   */
/* must add/change the name in "xlt_dll.def" to keep the names from     */
/*getting mangled, so VB can find things                                */
/************************************************************************/
#define STDCALL __stdcall





DLL_EXPORT BSTR STDCALL langdef( BSTR *in_strB )
{
    char *real_str;
    char *trans_str;
    BSTR ret_str;

    real_str = (char *)(*in_strB);
    if ( (real_str == (char *)NULL) || (*real_str == '\0') )
        return( *in_strB );


    trans_str = XLT( real_str );

    if ( (TransType == 2) && (trans_str == real_str) )
        {
        /* no translation, reverse it */
        trans_str = ReverseString( real_str );
        }
    ret_str = SysAllocStringByteLen( trans_str, strlen(trans_str) );
    return( ret_str );
}

    

/* translate first trying the complex form, and if that is not found in the */
/* glossary, go for the "simple" form                                       */
DLL_EXPORT BSTR STDCALL langdefEx( BSTR *tag_bstr, BSTR *in_bstr )
{
    char *tag, *str;
    char *trans_str;
    BSTR ret_str;

    tag = (char *)(*tag_bstr);
    str = (char *)(*in_bstr);

    if ( (str == (char *)NULL) || (*str == '\0') )
       return( *in_bstr );
    if ( tag == (char *)NULL )
       tag = "";

    trans_str = TransComplexThenSimple( tag, str );

    ret_str = SysAllocStringByteLen( trans_str, strlen(trans_str) );
    return( ret_str );
}



/* translate using only the complex form */
DLL_EXPORT BSTR STDCALL langdefComplex( BSTR *tag_bstr, 
                                        BSTR *in_bstr, 
                                        int *log )
{
    char *tag, *str;
    char *trans_str;
    BSTR ret_str;

    tag = (char *)(*tag_bstr);
    str = (char *)(*in_bstr);

    if ( (str == (char *)NULL) || (*str == '\0') )
       return( *in_bstr );
    if ( tag == (char *)NULL )
       tag = "";

    if ( Log_Exceptions && !aos_Verbose && !(*log) )
        {
        XltExceptionsLogOnly( FALSE );
        trans_str = XLT2( tag, str );
        XltExceptionsLogOnly( TRUE );
        }
    else
        trans_str = XLT2( tag, str );

    ret_str = SysAllocStringByteLen( trans_str, strlen(trans_str) );
    return( ret_str );
}



DLL_EXPORT int STDCALL LoadGlossaryFromVB( BSTR *gls_fileB )
{
    char *glossary;
    int ret;

    glossary = (char *)(*gls_fileB);

    ret = LoadGlossary( glossary, TRUE );

    return( ret );

}



DLL_EXPORT int STDCALL AppendGlossaryFromVB( BSTR *gls_fileB )
{
    char *glossary;

    glossary = (char *)(*gls_fileB);

    AppendGlossary( glossary );

    return( 0 );
}


DLL_EXPORT int STDCALL XltExceptionsLogFromVB( int *turn_on, 
                                           BSTR *filename_bstr, 
                                           int *append,
                                           int *verbose )
{
    char *filename;

    aos_Verbose = *verbose;
    ConvertEscapeSequences = FALSE;   /* leave "/t" as "/t"   */

    filename = (char *)(*filename_bstr);

    XltExceptionsLog( *turn_on, filename, *append );

    return( 0 );
}




    

/***********************************************************************/
/* translate each field in a string consisting of character seperated    */
/* tokens, and return a new string of character seperated translated tokens */
/* Note: in some fields, there are sets of character seperated tokens,      */
/*      seperated by ^B (0x02) characters                               */
/***********************************************************************/

DLL_EXPORT BSTR STDCALL TranslateList( BSTR *tag_bstr, 
                                       BSTR *field_bstr,
                                       BSTR *seperator_bstr )
{
    char *tag_str;
    char *fieldvalue;
    char *seperator_str;
    char seperator;
    char new_field[40000];
    BSTR ret_str;
    int len;

    tag_str = (char *)(*tag_bstr);
    fieldvalue = (char *)(*field_bstr);
    seperator_str = (char *)(*seperator_bstr);
    seperator = *seperator_str;

    len = strlen(fieldvalue);
    DoTranslateList( tag_str, fieldvalue, new_field, seperator );
    len = strlen(new_field);

    /* now make up the new string, and return it */
    ret_str = SysAllocStringByteLen( new_field, strlen(new_field) );
    return( ret_str );
}






/***********************************************************************/
/* given an SQL statement, go through it and translate all the quoted  */
/*    entries.                                                         */
/***********************************************************************/

DLL_EXPORT BSTR STDCALL TranslateSQLStatement( BSTR *tag_bstr, 
                                               BSTR *field_bstr,
                                               int *bad_ok )
{
    char *tag_str;
    char *fieldvalue;
    char new_sql[40000];
    BSTR ret_str;

    tag_str = (char *)(*tag_bstr);
    fieldvalue = (char *)(*field_bstr);


    DoTranslateSQLStatement( tag_str, fieldvalue, new_sql, *bad_ok );

    /* now make up the new string, and return it */
    ret_str = SysAllocStringByteLen( new_sql, strlen(new_sql) );
    return( ret_str );
}



DLL_EXPORT int STDCALL ByteLenOfString( BSTR *bstr )
{
    int str_len;
    char *str;

    str = (char *)(*bstr);
    
    str_len = strlen( str );
    return( str_len );
}



DLL_EXPORT int STDCALL TestIt(  BSTR *tag_bstr )
{
    char *str;

    str = (char *)(*tag_bstr);

    if ( *str = 'a' )
       return( 0 );
    else
       return( 1 );
}

#endif     /*  } */
