/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: atxint.h
   Applix ASC version: 11
   Date: Mon Nov 16 13:53:48 1998
*/
#ifndef _ATXINT_H
#define _ATXINT_H

typedef struct linetab_ent_t {
  int line_height; /* vertical offset of this line from prev, zoomed mils */
  int x_indent;    /* left x of this line's ink relative to whole.pos.x */
  int inkbottom;   /* vertical extent of this line's ink below baseline */
  int leftix;      /* index of 1st char this line */
} linetab_ent;

typedef struct whole_field_i {
  unsigned int zoomOn: 1;
  int nChars;
  int zoom;               /* zoom factor relative to 1000 */
  int width;              /* formatting width limit, unzoomed, mils */
  int formattedWidth;     /* of widest line, zoomed, mils (center/rt align)*/
  p32 pos;                /* setpos of upstream-most character */
  p32 fixedPos;           /* point to remain stationary during edit */
  r32 bbox;               /* ink bounder */
  int lineHeight;         /* vertical distance between baselines */
  int bounderMargin;      /* all-round fudge factor for ink bounder guess */
  int alignment;          /* left, right, center, justified, .. */
  int numlines;           /* # entries in line table */
  linetab_ent *linetab;   /* line table */
  int linetab_memsize;
} wholeFieldI;

typedef struct font_props_i {
  char *familyName; /* for example, "Times" */
  char *fontName; /* for example, "Times-Roman" */
  int weight;
  int slant;
  float ptsize;         /* in points */
  int ptsizeInt;        /* integer * 10000: e.g. 14.4 pts -> 144000 */
  int unzoomed_ptsize;  /* unzoomed pointsize, relative to 10000 */
  int underlinePosition; /* in milliems */
  int underlineThickness; /* in milliems */
  r32 bbox; /* font bounding box, mils */
} fontPropsI;

typedef struct tx_attr_i {
  unsigned int useFgPixel  :1,	/* if TRUE, member fgPixel meaningful */
               hilite      :1,
               strike_thru :1;
  int nChars;		   /* number of chars under this attr segment */
  int supy;		   /* non-accumulating y offset */
  int kern;		   /* accumulating x offset */
  int fkern;		   /* accumulating x offset, no zoom */
  int ruling;		   /* AXRULING_SINGLE, e.g., */
  fontContext fconx;	   /* font metrics and table reference */
  fontPropsI fontAttr;	   /* user-specified attr info */
  p32 offset;		   /* measured from current set position */
  int strokecolor;	   /* fg color of letterform outline stroke */
  int strokebcolor;
  attr_type *strokepattern;  /* fill information for outline stroke */
  attr_type *fillpattern;    /* fill information for letterform fill */
  char *screenFontName;	     /* resolved xlfd name */
  int fillcolor;	     /* packed cmyk letterform fill fg color */
  int fillbcolor;	     /* packed cmyk letterform fill bg color */
  unsigned long fgPixel;     /* direct pass-thru of fill fg color */
  axtIndent *indent;	     /* apply to galley width */
  axtDropShadow *dropShadow;
} txAttrI;

typedef struct per_char_i {
  unsigned int screenMetrics : 1; /* whether known */
  unsigned int targetMetrics : 1; /* whether known */
  unsigned int changed : 1;	/* since last redisplay */
  unsigned int deleted : 1;	/* since last redisplay */
  unsigned int damaged : 1;	/* since last redisplay */
  unsigned int inserted : 1;	/* since last redisplay */
  unsigned int moved : 1;		/* since last redisplay */
  unsigned int hidden : 1;	/* since last reFit */
  unsigned int temp : 1;		/* various temporary uses */
  unsigned int hyphen : 1;
  unsigned int nonbreakspace : 1;
  unsigned int linebreak : 1;	/* formatter started line with this */
  unsigned int hanging : 1;	/* formatter 'hung' this (space) char */
  unsigned int hilite : 1;	/* this char should be hilighted */
  tx_char charcode;
  p32 offset;	   /* formatting offset; affects downstream chars, mils */
  r32 tbox;					/* target letterform bounder, mils */
  r32 sbox;					/* source letterform bounder, screenpix */
  int tadvance;				/* target advance in mils */
  int sadvance;				/* screen advance in screenpix */
  int fadvance;				/* unzoomed target advance in mils */
  int kern;                   /* target kern, mils */
  int fkern;                  /* unzoomed target kern, mils */
  int sup;                    /* target subsup offset, mils */
} perCharI;

typedef struct e_state_i {
  unsigned int disp : 1; /* if true, text believed to be on-screen */
  int editPos;	         /* upstream-most index affected by edit */
  int displayLen;        /* redisplay this length after editPos */
  Pixmap background; /* for lifting text */
  unsigned long backgroundPixel; /* for simple color rather than pixmap */
  int javaPixel; /* for simple color rather than pixmap */
  r32 backRect;	   /* window-rel position, size of cleared background */

  int cursorPos;
  int cursorType;
  XPoint lastCursorPoints[4];
  int lastCursorPointcount;
  r32 lastCursorClip;
  r32 fieldBounder;
  int fullDisp;		 /* unknown damage, so draw in entirety */
	
  unsigned int
  cursorOn         : 1,	/* cursor belived to be on-screen */
    cursorActive     : 1,
    lastCursorActive : 1,
    inEdit           : 1,	/* display implies drawCursor */
    forceBGDraw      : 1;	/* draw the background, no matter what */
} eStateI;

typedef struct a_text {
  int pid;
  void *gContext;           /* handle of ax graphics context */
  unsigned int
  constructed   : 1,		/* att info and text string */
    measured      : 1,		/* target (printer) metrics */
    displayable   : 1,		/* screen metrics and resources */
    draw_shadow   : 1,		/* draw shadows */
    draw_emboss   : 1,		/* draw emboss */
    shadow_only   : 1,		/* draw only shadows */
    hide_colors   : 1;		/* hide colors */
    
  wholeFieldI whole;
  txAttrI  **atts;
  perCharI **items;
  txAttrI *userDefaults;
  unsigned int userDefaultFlags;

  eStateI editState;
  r32 clipRect;
  struct a_text *old;
} atext;

#endif /* _ATXINT_H */


