/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ax_inset.c
   Applix ASC version: 58
   Date: Tue Jun  1 21:12:11 1999
*/
#include <stdlib.h>
#include <math.h>

#define BOOL_DEFINED 1

#include <aos.h>
#include <elbltn.h>
#include <arghdr.h>
#include <ax_inset.h>
#include <eldata.h>
#include <bldrbltn.h>
#include <aos_xfer.h>
#if 0
#include <ax_inset.p>
#endif

extern data_ptr grPackageInsetObject();
extern data_ptr ssPackageInsetObject();
extern data_ptr wpPackageInsetObject();
extern data_ptr htmlPackageInsetObject();
extern data_ptr bldrPackageInsetObject();
extern data_ptr grCreateInsetImageGraphic();

extern Display *Dpy;
extern short DpyDepth;
extern Window RootWin;
extern GC PixClearGC;
extern data_ptr ElfRtData;
extern char *AxTempFile();

static inset_image **ins_images;
static int image_uid;
static int n_images;
Pixmap BackGroundPixmap = (Pixmap)NULL;
Pixmap FinalScreenPixmap = (Pixmap)NULL;

/****************************************************************************
 *
 * module: insetCreateImage
 */
static int
insetCreateImage()
{
	inset_image *im;
	int sz;

	sz = (n_images+1) * sizeof(inset_image *);
	ins_images = (inset_image **)TrueRealloc(ins_images, sz);
	im = (inset_image *)TaskAlloc(0, sizeof(inset_image));
	memset(im, 0, sizeof(inset_image));
	im->image_id = ++image_uid;
	ins_images[n_images++] = im;
	return(im->image_id);
}

/****************************************************************************
 *
 * module: insetRetrieveSlotIndex
 */
static int
insetRetrieveSlotIndex(image_id)
int image_id;
{
	int slot_ix = -1;
	int i;
	for ( i=0; i<n_images; i++ )
	{
		if ( ins_images[i]->image_id == image_id )
			slot_ix = i;
	}
	return(slot_ix);
}

/****************************************************************************
 *
 * module: insetRetrieveImage
 */
static int
insetRetrieveImage(image_id, slot_ix, im)
int image_id;
int *slot_ix;
inset_image **im;
{
	int err;

	*im = (inset_image *)0;
	*slot_ix = insetRetrieveSlotIndex(image_id);

	if ( *slot_ix >= 0 )
	{
		*im = ins_images[*slot_ix];
		err = AOK;
	}
	else
		err = !AOK;
	return(err);
}

static void 
destroy_final_info(im)
inset_image *im;
{
	if ( im->final != (inset_image *)0)
		TaskFree(0, im->final);
	im->final = (inset_image *)0;
}

static void 
destroy_dimbullet_info(im)
inset_image *im;
{
	if ( im->dimB != (inset_image *)0)
	{
		AxFreePixmap(im->dimB->pixmap);
		TaskFree(0, im->dimB);
	}
	im->dimB = (inset_image *)0;
}

static void 
destroy_flyin_info(im)
inset_image *im;
{
	if ( im->flyIn != (inset_image *)0)
	{
		AxFreePixmap(im->flyIn->pixmap);
		TaskFree(0, im->flyIn);
	}
	im->flyIn = (inset_image *)0;
}

static void 
destroy_txbx_info(im)
inset_image *im;
{
	if ( im->txbx != (inset_image *)0)
	{
		AxFreePixmap(im->txbx->pixmap);
		TaskFree(0, im->txbx);
	}
	im->txbx = (inset_image *)0;
}

/****************************************************************************
 *
 * module: freeInsetImage
 */
static void 
freeInsetImage(im)
inset_image *im;
{
	im->pixmap = (Pixmap)None;
	if ( im->final != (inset_image *)0)
		destroy_final_info(im);
	if ( im->flyIn != (inset_image *)0)
		destroy_flyin_info(im);
	if ( im->dimB != (inset_image *)0)
		destroy_dimbullet_info(im);
	if ( im->txbx != (inset_image *)0)
		destroy_txbx_info(im);
	TaskFree(0, im);
	im = (inset_image *)0;
}

/****************************************************************************
 *
 * module: insetFreeImage
 */
static void
insetFreeImage(im, slot_ix)
inset_image *im;
int slot_ix;
{
	int i, sz;

	/* compress list */
	for ( i=slot_ix+1; i<n_images; i++ )
		ins_images[i-1] = ins_images[i];
	n_images--;
	sz = (n_images +1)*sizeof(inset_image *);
	ins_images = (inset_image **)((sz>0)?TrueRealloc(ins_images, sz):0);

	/* free resources */
	freeInsetImage(im);
}

/****************************************************************************
 *
 * module: insetFreeImagebyImageId
 */
insetFreeImagebyImageId(image_id)
int image_id;
{
	inset_image *im;
	int err, slot_ix;

	err = insetRetrieveImage(image_id, &slot_ix, &im);
	if ( err == AOK )
		insetFreeImage(im, slot_ix);
	return(err);
}

/****************************************************************************
 *
 * module: createInsetImage
 */
createInsetImage(im)
inset_image **im;
{
	int image_id, slot_ix, err;

	image_id = insetCreateImage();
	err = insetRetrieveImage(image_id, &slot_ix, im);
	return(err);
}

/****************************************************************************
 *
 * module: extractInsetImageFromDataPtr
 */
extractInsetImageFromDataPtr(imP, im)
data_ptr imP;
inset_image **im;
{
	int image_id, slot_ix, err;
	*im = (inset_image *)0;
	if ( imP )
	{
		image_id = AxIntFromDataPtr(imP);
		err = insetRetrieveImage(image_id, &slot_ix, im);
	}
	return(err);
}

/****************************************************************************
 *
 * module: AxMakeInsetImageData
 */
data_ptr
AxMakeInsetImageData(im)
inset_image *im;
{
	if ( im )
		return(AxMakeIntData(im->image_id));
	else
		return(NULL);
}

static void
fill_inset_image(w, im , x1, y1, x2, y2, wid, hyt)
long w;
inset_image *im;
int x1, y1, x2, y2;
int *wid, *hyt;
{
	extern int JavaMode;
	extern Pixmap ScratchMap;

	im->widget = w;
	im->x1 = x1;
	im->y1 = y1;
	im->x2 = x2;
	im->y2 = y2;
	conv_i_p32(&x1, &y1, &x1, &y1, MILLS_TO_PIXEL);
	conv_i_p32(&x2, &y2, &x2, &y2, MILLS_TO_PIXEL);
	*wid = x2 - x1 + 1; 
	*hyt = y2 - y1 + 1;
	if ( im->pixmap == (Pixmap)None)
	{
		if ( JavaMode )
			im->pixmap = AxCreateQueryPixmap(Dpy, RootWin,
												  (*wid), (*hyt), DpyDepth);
		else if ( ScratchMap )
			im->pixmap = AxCreateQueryPixmap(Dpy, RootWin,
												  (*wid), (*hyt), DpyDepth);
	}
}

long
create_dimbullet_pmap(w, im, x1, y1, x2, y2)
long w;
inset_image *im;
int x1, y1, x2, y2;
{
	int wd, h;

	if ( im->dimB )
		destroy_dimbullet_info(im);
	im->dimB = (inset_image *)TaskAlloc(0, sizeof(inset_image));
	memset(im->dimB, 0, sizeof(inset_image));
	fill_inset_image(w, im->dimB , x1, y1, x2, y2, &wd, &h);
	return((long)im->dimB->pixmap);
}

long
create_flyin_pmap(w, im, x1, y1, x2, y2)
long w;
inset_image *im;
int x1, y1, x2, y2;
{
	int wd, h;

	if ( im->flyIn )
		destroy_flyin_info(im);
	im->flyIn = (inset_image *)TaskAlloc(0, sizeof(inset_image));
	memset(im->flyIn, 0, sizeof(inset_image));
	fill_inset_image(w, im->flyIn , x1, y1, x2, y2, &wd, &h);
	return((long)im->flyIn->pixmap);
}

long
create_txbx_pmap(w, im, x1, y1, x2, y2)
long w;
inset_image *im;
int x1, y1, x2, y2;
{
	int wd, h;

	if ( im->txbx )
		destroy_txbx_info(im);
	im->txbx = (inset_image *)TaskAlloc(0, sizeof(inset_image));
	memset(im->txbx, 0, sizeof(inset_image));
	fill_inset_image(w, im->txbx , x1, y1, x2, y2, &wd, &h);
	ax_clear_pixmap(im->txbx->pixmap, wd, h);
	return((long)im->txbx->pixmap);
}

long
create_final_pmap(w, im, x1, y1, x2, y2)
long w;
inset_image *im;
int x1, y1, x2, y2;
{
	int wd, h;

	if ( im->final )
		destroy_final_info(im);
	im->final = (inset_image *)TaskAlloc(0, sizeof(inset_image));
	memset(im->final, 0, sizeof(inset_image));
	if ( FinalScreenPixmap )
		im->final->pixmap = FinalScreenPixmap;
	fill_inset_image(w, im->final , x1, y1, x2, y2, &wd, &h);
	if ( FinalScreenPixmap == (Pixmap)None)
		FinalScreenPixmap = im->final->pixmap;
	XFillRectangle(Dpy, FinalScreenPixmap, PixClearGC, 0, 0, wd, h);
	return((long)im->final->pixmap);
}

/****************************************************************************
 *
 * module: create_inset_image
 */
inset_image *
create_inset_image(w, x1, y1, x2, y2, create_pmap)
long w;
int x1, y1, x2, y2;
int create_pmap;
{
	inset_image *im;
	int err, wd, h;

	im = (inset_image *)0;
	err = createInsetImage(&im);
	if ( err == AOK && create_pmap == TRUE )
	{
		if ( BackGroundPixmap )
			im->pixmap = BackGroundPixmap;
		fill_inset_image(w, im , x1, y1, x2, y2, &wd, &h);
		if ( BackGroundPixmap == (Pixmap)None)
			BackGroundPixmap = im->pixmap;
		XFillRectangle(Dpy, BackGroundPixmap, PixClearGC, 0, 0, wd, h);
	}
	create_final_pmap(w, im, x1, y1, x2, y2);
	return(im);
}

/****************************************************************************
 *
 * module: axExp10
 */
static int
axExp10(n)
int n;
{
    int i, ival;
    for ( i=0, ival=1; i<n; i++ )
        ival *= 10;
    return(ival);
}

/****************************************************************************
 *
 * module: AxUnitsToDots
 */
int
AxUnitsToDots(units, val)
ax_units *units;
double val;
{
    double mag, floor();
    int ival, tmag;

    mag = tmag = axExp10(units->precision);
	if (0.0 == mag)
		tmag = 1.0;
    ival = floor(val*mag + 0.5) * units->dpu / tmag;
    return(ival);
}

/****************************************************************************
 *
 * module: AxDotsToUnits
 */
double
AxDotsToUnits(units, ival)
ax_units *units;
int ival;
{
    double val, mag, floor();
	int tdpu = units->dpu;

    mag = axExp10(units->precision);
	if (0.0 == mag)
		mag = 1.0;
	if (0 == tdpu)
		tdpu = 1;
    val = floor((double)ival*mag/tdpu + 0.5) / mag;
    return(val);
}

/****************************************************************************
 *
 * module: AxExtractUnits - extract elf data into units struct
 */
void
AxExtractUnits(pid, unitsP, units)
    int pid;
    data_ptr unitsP;
    ax_units *units;
{   
    data_ptr valP;
    int n;
    char *p;

    memset((char *)units, '\0', sizeof(ax_units));

    n = (!unitsP || !AxIsArray(unitsP)) ? 0 : AxArraySize(unitsP);
    if ( n < 1 )
        units->dpi = 0;
    else
    {
        valP = AxArrayElement(unitsP, 0);
		if ( valP && AxIsInt(valP) )
			units->dpi = AxIntFromDataPtr(valP);
		else if ( valP && AxIsFloat(valP) )
			units->dpi = (int) AxFloatFromDataPtr(valP);
		else
			units->dpi = 0;
    }
    if ( n < 2 )
        units->dpu = 0;
    else
    {
        valP = AxArrayElement(unitsP, 1);
		if ( valP && AxIsInt(valP) )
			units->dpu = AxIntFromDataPtr(valP);
		else if ( valP && AxIsFloat(valP) )
			units->dpu = (int) AxFloatFromDataPtr(valP);
		else
			units->dpu = 0;
    }
    if ( n < 3 )
        units->precision = 0;
    else
    {
        valP = AxArrayElement(unitsP, 2);
		if ( valP && AxIsInt(valP) )
			units->precision = AxIntFromDataPtr(valP);
		else if ( valP && AxIsFloat(valP) )
			units->precision = (int) AxFloatFromDataPtr(valP);
		else
			units->precision = 0;
    }
    if ( n < 4 )
        units->name = NULL;
    else
    {
        valP = AxArrayElement(unitsP, 3);
        if(valP && AxIsString(valP) )    
        {
            p = AxStrPtrFromDataPtr(valP);
            units->name = (char *)TaskAlloc(pid, strlen(p)+1);
            strcpy(units->name, p);
        }
    }
}

/****************************************************************************
 *
 * module: AxMakeUnits - make elf data from units struct
 */
data_ptr
AxMakeUnits(pid, units)
    int pid;
    ax_units *units;
{   
    data_ptr unitsP;
    
    unitsP = AxTaskMakeArray(4);
    AxAddIntToArray(unitsP, 0, units->dpi);
    AxAddIntToArray(unitsP, 1, units->dpu);
    AxAddIntToArray(unitsP, 2, units->dpu);
    if(units->name) 
        AxAddStrToArray(unitsP, 3, units->name);
    return(unitsP);
}

/****************************************************************************
 *
 * module: extractTransFromELF - generic transition data_ptr reader
 */
int
extractTransFromELF(valP, val)
data_ptr valP;
int *val;
{
    int stat = !AOK;
    char *str;
    if ( valP )
    {
        if ( AxIsString(valP) )
        {
            str = (char *)AxStrFromDataPtr(valP);
            if ( strcmp(NXLT("none"), str) == 0 )
            {
                stat = AOK;
                *val = 0;
            }
            else if ( strcmp(NXLT("fade"), str) == 0 )
            {
                stat = AOK;
                *val = 1;
            }
            else if ( strcmp(NXLT("wipe right"), str) == 0 )
            {
                stat = AOK;
                *val = 2;
            }
            else if ( strcmp(NXLT("slide right"), str) == 0 )
            {
                stat = AOK;
                *val = 3;
            }
            else if ( strcmp(NXLT("wipe left"), str) == 0 )
            {
                stat = AOK;
                *val = 4;
            }
            else if ( strcmp(NXLT("slide left"), str) == 0 )
            {
                stat = AOK;
                *val = 5;
            }
            else if ( strcmp(NXLT("wipe down"), str) == 0 )
            {
                stat = AOK;
                *val = 6;
            }
            else if ( strcmp(NXLT("slide down"), str) == 0 )
            {
                stat = AOK;
                *val = 7;
            }
            else if ( strcmp(NXLT("wipe up"), str) == 0 )
            {
                stat = AOK;
                *val = 8;
            }
            else if ( strcmp(NXLT("slide up"), str) == 0 )
            {
                stat = AOK;
                *val = 9;
            }
            else if ( strcmp(NXLT("slide right up"), str) == 0 )
            {
                stat = AOK;
                *val = 10;
            }
            else if ( strcmp(NXLT("slide left down"), str) == 0 )
            {
                stat = AOK;
                *val = 11;
            }
            else if ( strcmp(NXLT("slide left up"), str) == 0 )
            {
                stat = AOK;
                *val = 12;
            }
            else if ( strcmp(NXLT("slide right down"), str) == 0 )
            {
                stat = AOK;
                *val = 13;
            }
            else if ( strcmp(NXLT("blinds right"), str) == 0 )
            {
                stat = AOK;
                *val = 14;
            }
            else if ( strcmp(NXLT("blinds down"), str) == 0 )
            {
                stat = AOK;
                *val = 15;
            }
            else if ( strcmp(NXLT("checkers right down"), str) == 0 )
            {
                stat = AOK;
                *val = 16;
            }
            else if ( strcmp(NXLT("checkers left up"), str) == 0 )
            {
                stat = AOK;
                *val = 17;
            }
            else if ( strcmp(NXLT("spread out"), str) == 0 )
            {
                stat = AOK;
                *val = 18;
            }
            else if ( strcmp(NXLT("split vertical in"), str) == 0 )
            {
                stat = AOK;
                *val = 19;
            }
            else if ( strcmp(NXLT("split vertical out"), str) == 0 )
            {
                stat = AOK;
                *val = 20;
            }
            else if ( strcmp(NXLT("split horizontal in"), str) == 0 )
            {
                stat = AOK;
                *val = 21;
            }
            else if ( strcmp(NXLT("split horizontal out"), str) == 0 )
            {
                stat = AOK;
                *val = 22;
            }
            else if ( strcmp(NXLT("implode"), str) == 0 )
            {
                stat = AOK;
                *val = 23;
            }
            else if ( strcmp(NXLT("meltdown"), str) == 0 )
            {
                stat = AOK;
                *val = 24;
            }
            else if ( strcmp(NXLT("blinds left"), str) == 0 )
            {
                stat = AOK;
                *val = 25;
            }
            else if ( strcmp(NXLT("blinds up"), str) == 0 )
            {
                stat = AOK;
                *val = 26;
            }
            else if ( strcmp(NXLT("wipe right down"), str) == 0 )
            {
                stat = AOK;
                *val = 27;
            }
            else if ( strcmp(NXLT("wipe left down"), str) == 0 )
            {
                stat = AOK;
                *val = 28;
            }
            else if ( strcmp(NXLT("wipe right up"), str) == 0 )
            {
                stat = AOK;
                *val = 29;
            }
            else if ( strcmp(NXLT("wipe left up"), str) == 0 )
            {
                stat = AOK;
                *val = 30;
            }
            else if ( strcmp(NXLT("checkers left down"), str) == 0 )
            {
                stat = AOK;
                *val = 31;
            }
            else if ( strcmp(NXLT("checkers right up"), str) == 0 )
            {
                stat = AOK;
                *val = 32;
            }
            else if ( strcmp(NXLT("checkers left"), str) == 0 )
            {
                stat = AOK;
                *val = 33;
            }
            else if ( strcmp(NXLT("checkers right"), str) == 0 )
            {
                stat = AOK;
                *val = 34;
            }
            else if ( strcmp(NXLT("checkers down"), str) == 0 )
            {
                stat = AOK;
                *val = 35;
            }
            else if ( strcmp(NXLT("checkers up"), str) == 0 )
            {
                stat = AOK;
                *val = 36;
            }
            else if ( strcmp(NXLT("uncover left"), str) == 0 )
            {
                stat = AOK;
                *val = 37;
            }
            else if ( strcmp(NXLT("uncover right"), str) == 0 )
            {
                stat = AOK;
                *val = 38;
            }
            else if ( strcmp(NXLT("uncover down"), str) == 0 )
            {
                stat = AOK;
                *val = 39;
            }
            else if ( strcmp(NXLT("uncover up"), str) == 0 )
            {
                stat = AOK;
                *val = 40;
            }
            else if ( strcmp(NXLT("explode"), str) == 0 )
            {
                stat = AOK;
                *val = 41;
            }
            else if ( strcmp(NXLT("uncover right up"), str) == 0 )
            {
                stat = AOK;
                *val = 42;
            }
            else if ( strcmp(NXLT("uncover left down"), str) == 0 )
            {
                stat = AOK;
                *val = 43;
            }
            else if ( strcmp(NXLT("uncover left up"), str) == 0 )
            {
                stat = AOK;
                *val = 44;
            }
            else if ( strcmp(NXLT("uncover right down"), str) == 0 )
            {
                stat = AOK;
                *val = 45;
            }
            else if ( strcmp(NXLT("spread in"), str) == 0 )
            {
                stat = AOK;
                *val = 46;
            }
        }
        else if ( AxIsInt(valP) )
        {
            *val = AxIntFromDataPtr(valP);
            if ( *val >= 0 && *val <= 46 )
                stat = AOK;
        }
    }
    return(stat);
}

/****************************************************************************
 *
 * module: createTransForELF - generic transition data_ptr creator
 */
data_ptr
createTransForELF(val)
int val;
{
    data_ptr stP;
    switch(val)
    {
        default:
        case  0:  stP = AxMakeStrData(-1, NXLT("none"));                 break;
        case  1:  stP = AxMakeStrData(-1, NXLT("fade"));                 break;
        case  2:  stP = AxMakeStrData(-1, NXLT("wipe right"));           break;
        case  3:  stP = AxMakeStrData(-1, NXLT("slide right"));          break;
        case  4:  stP = AxMakeStrData(-1, NXLT("wipe left"));            break;
        case  5:  stP = AxMakeStrData(-1, NXLT("slide left"));           break;
        case  6:  stP = AxMakeStrData(-1, NXLT("wipe down"));            break;
        case  7:  stP = AxMakeStrData(-1, NXLT("slide down"));           break;
        case  8:  stP = AxMakeStrData(-1, NXLT("wipe up"));              break;
        case  9:  stP = AxMakeStrData(-1, NXLT("slide up"));             break;
        case  10: stP = AxMakeStrData(-1, NXLT("slide right up"));       break;
        case  11: stP = AxMakeStrData(-1, NXLT("slide left down"));      break;
        case  12: stP = AxMakeStrData(-1, NXLT("slide left up"));        break;
        case  13: stP = AxMakeStrData(-1, NXLT("slide right down"));     break;
        case  14: stP = AxMakeStrData(-1, NXLT("blinds right"));         break;
        case  15: stP = AxMakeStrData(-1, NXLT("blinds down"));          break;
        case  16: stP = AxMakeStrData(-1, NXLT("checkers right down"));  break;
        case  17: stP = AxMakeStrData(-1, NXLT("checkers left up"));     break;
        case  18: stP = AxMakeStrData(-1, NXLT("spread out"));           break;
        case  19: stP = AxMakeStrData(-1, NXLT("split vertical in"));    break;
        case  20: stP = AxMakeStrData(-1, NXLT("split vertical out"));   break;
        case  21: stP = AxMakeStrData(-1, NXLT("split horizontal in"));  break;
        case  22: stP = AxMakeStrData(-1, NXLT("split horizontal out")); break;
        case  23: stP = AxMakeStrData(-1, NXLT("implode"));              break;
        case  24: stP = AxMakeStrData(-1, NXLT("meltdown"));             break;
        case  25: stP = AxMakeStrData(-1, NXLT("blinds left"));          break;
        case  26: stP = AxMakeStrData(-1, NXLT("blinds up"));            break;
        case  27: stP = AxMakeStrData(-1, NXLT("wipe right down"));      break;
        case  28: stP = AxMakeStrData(-1, NXLT("wipe left down"));       break;
        case  29: stP = AxMakeStrData(-1, NXLT("wipe right up"));        break;
        case  30: stP = AxMakeStrData(-1, NXLT("wipe left up"));         break;
        case  31: stP = AxMakeStrData(-1, NXLT("checkers left down"));   break;
        case  32: stP = AxMakeStrData(-1, NXLT("checkers right up"));    break;
        case  33: stP = AxMakeStrData(-1, NXLT("checkers left"));        break;
        case  34: stP = AxMakeStrData(-1, NXLT("checkers right"));       break;
        case  35: stP = AxMakeStrData(-1, NXLT("checkers down"));        break;
        case  36: stP = AxMakeStrData(-1, NXLT("checkers up"));          break;
        case  37: stP = AxMakeStrData(-1, NXLT("uncover left"));         break;
        case  38: stP = AxMakeStrData(-1, NXLT("uncover right"));        break;
        case  39: stP = AxMakeStrData(-1, NXLT("uncover down"));         break;
        case  40: stP = AxMakeStrData(-1, NXLT("uncover up"));           break;
        case  41: stP = AxMakeStrData(-1, NXLT("explode"));              break;
        case  42: stP = AxMakeStrData(-1, NXLT("uncover right up"));     break;
        case  43: stP = AxMakeStrData(-1, NXLT("uncover left down"));    break;
        case  44: stP = AxMakeStrData(-1, NXLT("uncover left up"));      break;
        case  45: stP = AxMakeStrData(-1, NXLT("uncover right down"));   break;
        case  46: stP = AxMakeStrData(-1, NXLT("spread in"));            break;
    }
    return(stP);
}

/****************************************************************************
 *
 * module: AxMakeTransData
 */
data_ptr
AxMakeTransData(inset)
ax_inset *inset;
{
    data_ptr transP, strP;

	/* create a placeholder */
    transP = AxMakeArray(8);

    if ( inset->use_video_enabled )
        AxAddIntToArray(transP, 0, inset->video_enabled ? -1 : 0);
    if ( inset->use_video_style )
    {
        strP = createTransForELF(inset->video_style);
        AxAddDataToArray(transP, 1, strP);
    }
    if ( inset->use_audio_enabled )
        AxAddIntToArray(transP, 2, inset->audio_enabled ? -1 : 0);
    if ( inset->use_audio_pathname )
        AxAddStrToArray(transP, 3, inset->audio_pathname);
    if ( inset->use_audio_pended )
        AxAddIntToArray(transP, 4, inset->audio_pended ? -1 : 0);
    if ( inset->use_audio_gain )
        AxAddIntToArray(transP, 5, inset->audio_gain);
    if ( inset->use_video_throttle )
        AxAddIntToArray(transP, 6, inset->video_throttle);
    if ( inset->use_build_level )
        AxAddIntToArray(transP, 7, inset->build_level);

    return(transP);
}

/****************************************************************************
 *
 * module: AxMakeInsetData
 */
data_ptr
AxMakeInsetData(inset)
ax_inset *inset;
{
    data_ptr insetP, unitsP, dstP, clipP, transP;
    ax_units units;

	/* create a placeholder */
    insetP = AxMakeArray(19);

    /* create units on the fly for clip/dest rect analysis */
    unitsP = AxMakeArray(4);
    if ( inset->use_unit_dpi == FALSE )
        units.dpi = 1000;
    else
    {
        AxAddIntToArray(unitsP, 0, inset->unit_dpi);
        units.dpi = inset->unit_dpi;
    }
    if ( inset->use_unit_dpu == FALSE)
        units.dpu = units.dpi;
    else
    {
        AxAddIntToArray(unitsP, 1, inset->unit_dpu);
        units.dpu = inset->unit_dpu;
    }
    if ( inset->use_unit_precision == FALSE )
        units.precision = 0;
    else
    {
        AxAddIntToArray(unitsP, 2, inset->unit_precision);
        units.precision = inset->unit_precision;
    }
    if ( inset->use_unit_name )
        AxAddStrToArray(unitsP, 3, inset->unit_name);

    dstP = AxMakeArray(4);
    if ( inset->use_dst_x )
        AxAddFloatToArray(dstP, 0, AxDotsToUnits(&units, inset->dst_x));
    if ( inset->use_dst_y )
        AxAddFloatToArray(dstP, 1, AxDotsToUnits(&units, inset->dst_y));
    if ( inset->use_dst_wid )
        AxAddFloatToArray(dstP, 2, AxDotsToUnits(&units, inset->dst_wid));
    if ( inset->use_dst_hyt )
        AxAddFloatToArray(dstP, 3, AxDotsToUnits(&units, inset->dst_hyt));

    clipP = AxMakeArray(4);
    if ( inset->use_clip_x )
        AxAddFloatToArray(clipP, 0, AxDotsToUnits(&units, inset->clip_x));
    if ( inset->use_clip_y )
        AxAddFloatToArray(clipP, 1, AxDotsToUnits(&units, inset->clip_y));
    if ( inset->use_clip_wid )
        AxAddFloatToArray(clipP, 2, AxDotsToUnits(&units, inset->clip_wid));
    if ( inset->use_clip_hyt )
        AxAddFloatToArray(clipP, 3, AxDotsToUnits(&units, inset->clip_hyt));

    /* fill out the inset data ptr */
    AxAddDataToArray(insetP, 0, unitsP);
    AxAddDataToArray(insetP, 1, dstP);
    AxAddDataToArray(insetP, 2, clipP);
    if ( inset->use_suppress_cursor )
        AxAddIntToArray(insetP,  3, inset->suppress_cursor ?  -1 : 0);
    if ( inset->use_direct_render )
        AxAddIntToArray(insetP,  4, inset->direct_render ?  -1 : 0);
    if ( inset->use_icon_only )
        AxAddIntToArray(insetP,  5, inset->icon_only ?  -1 : 0);
    if ( inset->use_x_zoom )
        AxAddFloatToArray(insetP,  6, (double)((double)inset->x_zoom/10.0));
    if ( inset->use_y_zoom )
        AxAddFloatToArray(insetP,  7, (double)((double)inset->y_zoom/10.0));
    if ( inset->use_scale_mode )
        AxAddIntToArray(insetP,  8, inset->scale_mode);
    if ( inset->use_proportional )
        AxAddIntToArray(insetP,  9, inset->proportional ? -1 : 0);
    if ( inset->use_centered )
        AxAddIntToArray(insetP,  10, inset->centered ? -1 : 0);
    if ( inset->use_widget_colors )
        AxAddIntToArray(insetP,  11, inset->widget_colors ? -1 : 0);
    if ( inset->use_invert_colors )
        AxAddIntToArray(insetP,  12, inset->invert_colors ? -1 : 0);
    if ( inset->use_clear )
        AxAddIntToArray(insetP,  13, inset->clear ? -1 : 0);
    if ( inset->use_fill )
        AxAddIntToArray(insetP,  14, inset->fill ? -1 : 0);
    if ( inset->use_border_style )
        AxAddIntToArray(insetP,  15, inset->border_style);
    if ( inset->use_inactive_content )
        AxAddIntToArray(insetP,  16, inset->inactive_content ? -1: 0);
    if ( inset->use_inactive_border )
        AxAddIntToArray(insetP,  17, inset->inactive_border ? -1: 0);

	transP = AxMakeTransData(inset);
    AxAddDataToArray(insetP, 18, transP);

    return(insetP);
}

/****************************************************************************
 *
 * module: AxDestroyAssociatedGraphic
 */
AxDestroyAssociatedGraphic(aP)
data_ptr aP;
{
    AxTaskFreeElfData(0, aP);
}

/****************************************************************************
 *
 * module: AxExtractTransitionFromELF
 */
int
AxExtractTransitionFromELF(transP, inset)
data_ptr transP;
ax_inset *inset;
{
    data_ptr valP;
    char *str;
    int n;

    if ( !transP || !AxIsArray(transP) )
        return;
    n = AxArraySize(transP);

    if ( n <= 0 )
        return;
    valP = AxArrayElement(transP, 0);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_video_enabled = TRUE;
        inset->video_enabled = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    if ( n <= 1 )
        return;
    valP = AxArrayElement(transP, 1);
    if ( extractTransFromELF(valP, &inset->video_style) == AOK )
        inset->use_video_style = TRUE;

    if ( n <= 2 )
        return;
    valP = AxArrayElement(transP, 2);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_audio_enabled = TRUE;
        inset->audio_enabled = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    if ( n <= 3 )
        return;
    valP = AxArrayElement(transP, 3);
    if ( valP && AxIsString(valP) )
    {
        inset->use_audio_pathname = TRUE;
        strcpy(inset->audio_pathname, (char *)AxStrFromDataPtr(valP));
    }

    if ( n <= 4 )
        return;
    valP = AxArrayElement(transP, 4);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_audio_pended = TRUE;
        inset->audio_pended = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    if ( n <= 5 )
        return;
    valP = AxArrayElement(transP, 5);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_audio_gain = TRUE;
        inset->audio_gain = AxIntFromDataPtr(valP);
        if ( inset->audio_gain < 0 || inset->audio_gain > 100 )
            inset->use_audio_gain = FALSE;
    }

    if ( n <= 6 )
        return;
    valP = AxArrayElement(transP, 6);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_video_throttle = TRUE;
        inset->video_throttle = AxIntFromDataPtr(valP);
        if ( inset->video_throttle < 0 || inset->video_throttle > 100 )
            inset->use_video_throttle = FALSE;
    }

    if ( n <= 7 )
        return;

    valP = AxArrayElement(transP, 7);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_build_level = TRUE;
        inset->build_level = AxIntFromDataPtr(valP);
    }
}

/****************************************************************************
 *
 * module: AxExtractInsetFromELF - read the ELF info onto the c structure
 */
AxExtractInsetFromELF(aP, inset)
data_ptr aP;
ax_inset *inset;
{
    data_ptr valP, valP2;
    ax_units units;
    double val;
    char *str;
    int n;

    /* sanity check */
    if ( !AxIsArray(aP) )
        return(!AOK);
    n = AxArraySize(aP);

    /* read the units */
    if ( n <= 0 )
        return(AOK);
    valP = AxArrayElement(aP, 0);
    if ( !valP ) /* assume mils */
    {
        units.dpi = 1000;
        units.dpu = 1000;
        units.precision = 0;
    }
    else
    {
        valP2 = AxArrayElement(valP, 0);
        if ( !valP2 )
            units.dpi = 1000;
        else
        {
			if ( valP2 && AxIsInt(valP2) )
			{
				inset->use_unit_dpi = TRUE;
				inset->unit_dpi = AxIntFromDataPtr(valP2);
				units.dpi = inset->unit_dpi;
			}
			else
				units.dpi = 1000;
        }
        valP2 = AxArrayElement(valP, 1);
        if ( !valP2 )
            units.dpu = units.dpi;
        else
        {
			if ( valP2 && AxIsInt(valP2) )
			{
				inset->use_unit_dpu = TRUE;
				inset->unit_dpu = AxIntFromDataPtr(valP2);
				units.dpu = inset->unit_dpu;
			}
			else
				units.dpu = units.dpi;
        }
        valP2 = AxArrayElement(valP, 2);
        if ( !valP2 )
            units.precision = 0;
        else
        {
			if ( valP2 && AxIsInt(valP2) )
			{
				inset->use_unit_precision = TRUE;
				inset->unit_precision = AxIntFromDataPtr(valP2);
				units.precision = inset->unit_precision;
			}
			else
				units.precision = 0;
        }
        valP2 = AxArrayElement(valP, 3);
        if ( valP2 && AxIsString(valP2) )
        {
            inset->use_unit_name = TRUE;
            strcpy(inset->unit_name, (char *)AxStrFromDataPtr(valP2));
        }
    }

    /* read the destintation rect */
    if ( n <= 1 )
        return(AOK);
    valP = AxArrayElement(aP, 1);
    if ( valP )
    {
        valP2 = AxArrayElement(valP, 0);
        if ( valP2 && AxIsFloatable(valP2) )
        {
            inset->use_dst_x = TRUE;
            inset->dst_x = AxUnitsToDots(&units, AxFloatFromDataPtr(valP2));
        }
        valP2 = AxArrayElement(valP, 1);
        if ( valP2 && AxIsFloatable(valP2) )
        {
            inset->use_dst_y = TRUE;
            inset->dst_y = AxUnitsToDots(&units, AxFloatFromDataPtr(valP2));
        }
        valP2 = AxArrayElement(valP, 2);
        if ( valP2 && AxIsFloatable(valP2) )
        {
            inset->use_dst_wid = TRUE;
            inset->dst_wid = AxUnitsToDots(&units, AxFloatFromDataPtr(valP2));
        }
        valP2 = AxArrayElement(valP, 3);
        if ( valP2 && AxIsFloatable(valP2) )
        {
            inset->use_dst_hyt = TRUE;
            inset->dst_hyt = AxUnitsToDots(&units, AxFloatFromDataPtr(valP2));
        }
    }

    /* read the clip rect */
    if ( n <= 2 )
        return(AOK);
    valP = AxArrayElement(aP, 2);
    if ( valP )
    {
        valP2 = AxArrayElement(valP, 0);
        if ( valP2 && AxIsFloatable(valP2) )
        {
            inset->use_clip_x = TRUE;
            inset->clip_x = AxUnitsToDots(&units, AxFloatFromDataPtr(valP2));
        }
        valP2 = AxArrayElement(valP, 1);
        if ( valP2 && AxIsFloatable(valP2) )
        {
            inset->use_clip_y = TRUE;
            inset->clip_y = AxUnitsToDots(&units, AxFloatFromDataPtr(valP2));
        }
        valP2 = AxArrayElement(valP, 2);
        if ( valP2 && AxIsFloatable(valP2) )
        {
            inset->use_clip_wid = TRUE;
            inset->clip_wid = AxUnitsToDots(&units, AxFloatFromDataPtr(valP2));
        }
        valP2 = AxArrayElement(valP, 3);
        if ( valP2 && AxIsFloatable(valP2) )
        {
            inset->use_clip_hyt = TRUE;
            inset->clip_hyt = AxUnitsToDots(&units, AxFloatFromDataPtr(valP2));
        }
    }

    /* read the cursor suppression flag */
    if ( n <= 3 )
        return(AOK);
    valP = AxArrayElement(aP, 3);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_suppress_cursor = TRUE;
        inset->suppress_cursor = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    /* read the direct rendering flag */
    if ( n <= 4 )
        return(AOK);
    valP = AxArrayElement(aP, 4);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_direct_render = TRUE;
        inset->direct_render = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    /* read the draw as icon flag */
    if ( n <= 5 )
        return(AOK);
    valP = AxArrayElement(aP, 5);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_icon_only = TRUE;
        inset->icon_only = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    /* read the zoom flags */
    if ( n <= 6 )
        return(AOK);
    valP = AxArrayElement(aP, 6);
    if ( valP && AxIsFloatable(valP) )
    {
        val = AxFloatFromDataPtr(valP) * 10;
        inset->use_x_zoom = TRUE;
        inset->x_zoom = MAX(30, val);
    }
    if ( n <= 7 )
        return(AOK);
    valP = AxArrayElement(aP, 7);
    if ( valP && AxIsFloatable(valP) )
    {
        val = AxFloatFromDataPtr(valP) * 10;
        inset->use_y_zoom = TRUE;
        inset->y_zoom = MAX(30, val);
    }

    /* read the scaling instructions */
    if ( n <= 8 )
        return(AOK);
    valP = AxArrayElement(aP, 8);
    if ( valP )
    {
        if ( AxIsString(valP) )
        {
            str = (char *)AxStrFromDataPtr(valP);
            if ( strcmp(NXLT("clip to fit"), str) == 0 )
            {
                inset->use_scale_mode = TRUE;
                inset->scale_mode = 0;
            }
            else if ( strcmp(NXLT("scale to fit"), str) == 0 )
            {
                inset->use_scale_mode = TRUE;
                inset->scale_mode = 1;
            }
            else if ( strcmp(NXLT("scale if clips"), str) == 0 )
            {
                inset->use_scale_mode = TRUE;
                inset->scale_mode = 2;
            }
        }
        else if ( AxIsInt(valP) )
        {
            inset->use_scale_mode = TRUE;
            inset->scale_mode = AxIntFromDataPtr(valP);
            if ( inset->scale_mode > 2 || inset->scale_mode < 0 )
                inset->use_scale_mode = FALSE;
        }
    }
    if ( n <= 9 )
        return(AOK);
    valP = AxArrayElement(aP, 9);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_proportional = TRUE;
        inset->proportional = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }
    if ( n <= 10 )
        return(AOK);
    valP = AxArrayElement(aP, 10);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_centered = TRUE;
        inset->centered = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    /* read the decorations */
    if ( n <= 11 )
        return(AOK);
    valP = AxArrayElement(aP, 11);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_widget_colors = TRUE;
        inset->widget_colors = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    if ( n <= 12 )
        return(AOK);
    valP = AxArrayElement(aP, 12);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_invert_colors = TRUE;
        inset->invert_colors = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    if ( n <= 13 )
        return(AOK);
    valP = AxArrayElement(aP, 13);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_clear = TRUE;
        inset->clear = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    if ( n <= 14 )
        return(AOK);
    valP = AxArrayElement(aP, 14);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_fill = TRUE;
        inset->fill = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    if ( n <= 15 )
        return(AOK);
    valP = AxArrayElement(aP, 15);
    if ( valP )
    {
        if ( AxIsString(valP) )
        {
            str = (char *)AxStrFromDataPtr(valP);
            if ( strcmp(NXLT("none"), str) == 0 )
            {
                inset->use_border_style = TRUE;
                inset->border_style = 0;
            }
            else if ( strcmp(NXLT("solid"), str) == 0 )
            {
                inset->use_border_style = TRUE;
                inset->border_style = 1;
            }
            else if ( strcmp(NXLT("dashed"), str) == 0 )
            {
                inset->use_border_style = TRUE;
                inset->border_style = 2;
            }
        }
        else if ( AxIsInt(valP) )
        {
            inset->use_border_style = TRUE;
            inset->border_style = AxIntFromDataPtr(valP);
            if ( inset->border_style > 2 || inset->border_style < 0 )
                inset->use_border_style = FALSE;
        }
    }
    if ( n <= 16 )
        return(AOK);
    valP = AxArrayElement(aP, 16);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_inactive_content = TRUE;
        inset->inactive_content = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }
    if ( n <= 17 )
        return(AOK);
    valP = AxArrayElement(aP, 17);
    if ( valP && AxIsInt(valP) )
    {
        inset->use_inactive_border = TRUE;
        inset->inactive_border = AxIntFromDataPtr(valP) ? TRUE : FALSE;
    }

    if ( n <= 18 )
        return(AOK);
    valP = AxArrayElement(aP, 18);
    if ( valP )
        AxExtractTransitionFromELF(valP, inset);

    /* done */
    return(AOK);
}

/****************************************************************************
 *
 * module: AxVerifyInset
 */
AxVerifyInset(inset, wid, hyt)
ax_inset *inset;
int wid, hyt;
{
    /* init the draw flags */
    if ( inset->use_suppress_cursor == FALSE )
        inset->suppress_cursor = TRUE;
    if ( inset->use_direct_render == FALSE )
        inset->direct_render = FALSE;
    if ( inset->use_icon_only == FALSE )
        inset->icon_only = FALSE;
    if ( inset->use_scale_mode == FALSE )
        inset->scale_mode = 1; /* scale to fit */
    if ( inset->use_proportional == FALSE )
        inset->proportional = TRUE;
    if ( inset->use_centered == FALSE )
        inset->centered = FALSE;
    if ( inset->use_widget_colors == FALSE )
        inset->widget_colors = FALSE;
    if ( inset->use_invert_colors == FALSE )
        inset->invert_colors = FALSE;
    if ( inset->use_clear == FALSE )
        inset->clear = TRUE;
    if ( inset->use_fill == FALSE )
        inset->fill = TRUE;
    if ( inset->use_border_style == FALSE )
        inset->border_style = 0;
    if ( inset->use_inactive_content == FALSE )
        inset->inactive_content = FALSE;
    if ( inset->use_inactive_border == FALSE )
        inset->inactive_border = FALSE;

    /* init the destination size and offset */
    if ( inset->use_unit_dpi == FALSE )
        inset->unit_dpi = 1000;
    if ( inset->use_unit_dpu == FALSE )
        inset->unit_dpu = 1;
    if ( inset->use_unit_precision == FALSE )
        inset->unit_precision = 0;
    if ( inset->use_unit_name == FALSE )
        inset->unit_name[0] = (char)0;

    /* init the destination size and offset */
    if ( inset->use_dst_x == FALSE )
        inset->dst_x = 0;
    if ( inset->use_dst_y == FALSE )
        inset->dst_y = 0;
    if ( inset->use_dst_wid == FALSE )
        inset->dst_wid = wid;
    if ( inset->use_dst_hyt == FALSE )
        inset->dst_hyt = hyt;

    /* init the clip */
    if ( inset->use_clip_x == FALSE )
        inset->clip_x = inset->dst_x;
    if ( inset->use_clip_y == FALSE )
        inset->clip_y = inset->dst_y;
    if ( inset->use_clip_wid == FALSE )
        inset->clip_wid = inset->dst_wid;
    if ( inset->use_clip_hyt == FALSE )
        inset->clip_hyt = inset->dst_hyt;

    /* init prescale factors */
    if ( inset->use_x_zoom == FALSE )
        inset->x_zoom = 1000;
    if ( inset->use_y_zoom == FALSE )
        inset->y_zoom = 1000;

    /* init the transition info */
    if ( inset->use_video_enabled == FALSE )
        inset->video_enabled = FALSE;
    if ( inset->use_audio_enabled == FALSE )
        inset->audio_enabled = FALSE;
    if ( inset->use_video_style == FALSE )
        inset->video_style = 0;
    if ( inset->use_audio_pended == FALSE )
        inset->audio_pended = FALSE;
    if ( inset->use_audio_gain == FALSE )
        inset->audio_gain = 50;
    if ( inset->use_video_throttle == FALSE )
        inset->video_throttle = 1;
    if ( inset->use_audio_pathname == FALSE )
    {
        inset->audio_pathname[0] = (char)0;
        inset->use_audio_enabled = FALSE;
        inset->audio_enabled = 0;
    }
    if ( inset->use_build_level == FALSE )
		inset->build_level = 0;
}

/****************************************************************************
 *
 *     module: ElfbCreateInsetObject
 */
typeErr
ElfbCreateInsetObject(typeP, uidP, srcP, unitsP, jnk4P, jnk5P, jnk6P)
data_ptr typeP;
data_ptr uidP;
data_ptr srcP;
data_ptr unitsP;
data_ptr jnk4P;
data_ptr jnk5P;
data_ptr jnk6P;
{
    int inset_type;

    if ( !typeP || !uidP )
        ElfRetData = (data_ptr)0;
    else
    {
        inset_type = AxIntFromDataPtr(typeP);
        switch (inset_type)
        {
            case IC_GE:
                ElfRetData = grPackageInsetObject(uidP, srcP, unitsP); break;
            case IC_SS:
                ElfRetData = ssPackageInsetObject(uidP, srcP, unitsP); break;
            case IC_DC:
                ElfRetData = wpPackageInsetObject(uidP, srcP, unitsP); break;
            case IC_HTML:
                ElfRetData = htmlPackageInsetObject(uidP,srcP,unitsP); break;
            case IC_AB:
                ElfRetData = bldrPackageInsetObject(uidP,srcP,unitsP); break;
            default:  
                ElfRetData = (data_ptr)0;                              break;
        }
    }
    return(AOK);
}

/****************************************************************************
 *
 *     module: ElfbAddIconInsetObject
 */
typeErr
ElfbAddIconInsetObject(objP, typeP, uidP, srcP, unitsP, jnk5P, jnk6P)
data_ptr objP;
data_ptr typeP;
data_ptr uidP;
data_ptr srcP;
data_ptr unitsP;
data_ptr jnk5P;
data_ptr jnk6P;
{
    int inset_type;
    data_ptr iconP;

    if ( !typeP || !uidP )
        iconP = (data_ptr)0;
    else
    {
        inset_type = AxIntFromDataPtr(typeP);
        switch (inset_type)
        {
            case IC_GE: iconP = grPackageInsetObject(uidP, srcP, unitsP); break;
            default:  iconP = (data_ptr)0;                                break;
        }
    }

    ElfRetData = AxTaskCopyData(THIMpid, objP);
    ElfRetData = AxAddArrayToArray(ElfRetData, 5, iconP); /* icon info */
    return(AOK);
}

/****************************************************************************
 *
 * module: AxSubRenderInsetGraphic
 */
static void
AxSubRenderInsetGraphic(w, objP, infP, imP)
AXWidget w;
data_ptr objP;
data_ptr infP;
data_ptr imP;
{
    if ( !objP || AxIsArray(objP) == FALSE )
        return;

    switch (AxIntFromArray(objP, 0))
    {
        case IC_GE:   grRenderAssociatedGraphic(w, objP, infP, imP); break;
        case IC_SS:   ssRenderAssociatedGraphic(w, objP, infP, imP); break;
        case IC_DC:   wpRenderAssociatedGraphic(w, objP, infP, imP); break;
        case IC_HTML: wpRenderAssociatedGraphic(w, objP, infP, imP); break;
        case IC_AB:   bldrRenderAssociatedGraphic(w,objP,infP, imP); break;
    }
}

/****************************************************************************
 *
 *     module: ElfbPaintInsetObject
 */
typeErr
ElfbPaintInsetObject(wP, objP, infP, imP, jnk4P, jnk5P, jnk6P)
data_ptr wP;
data_ptr objP;
data_ptr infP;
data_ptr imP;
data_ptr jnk4P;
data_ptr jnk5P;
data_ptr jnk6P;
{
	inset_image *im;
    AXWidget w;
	int err;

    if ( !wP )
        return(!AOK);
    w = (AXWidget)WidgetPointerFromHandleData(wP);
    if ( w == (AXWidget)0 )
        return(!AOK);
    if ( w->type == W_CANVAS && !w->window )
        PaintCanvas(w, NULL, TRUE, TRUE);
    AxSubRenderInsetGraphic(w, objP, infP, imP);
	err = extractInsetImageFromDataPtr(imP, &im);
	if ( im )
		destroy_flyin_info(im);
    return(AOK);
}

/****************************************************************************
 *
 *  module: AxSubCreateInsetImage
 */
static data_ptr
AxSubCreateInsetImage(w, objP, infP, imP)
AXWidget w;
data_ptr objP;
data_ptr infP;
data_ptr imP;
{
	inset_image *im;
    data_ptr retP;
	int wid, hyt;
	Pixmap pmap;

	retP = (data_ptr)0;
    if ( objP && AxIsArray(objP) )
	{
		switch (AxIntFromArray(objP, 0))
		{
			case IC_GE:
					retP = grCreateInsetImageGraphic(w, objP, infP, imP);
			break;
			case IC_SS:
			case IC_DC:
			case IC_HTML:
			case IC_AB:
					AxSubRenderInsetGraphic(w, objP, infP, (data_ptr)0);
					im = (inset_image *)create_inset_image(w, 0, 0, 0, 0,FALSE);
					if ( im != NULL )
					{
						im->pixmap = None;
						retP = (data_ptr)AxMakeInsetImageData(im);
					}
			break;
			default:  
			break;
		}
	}
    return(retP);
}

static data_ptr
get_java_bmap_from_pmap(pmap, wid, hyt, back_is_transparent)
Pixmap pmap;
int wid, hyt;
int back_is_transparent;
{
    XImage *ximg;
    data_ptr retP;
    extern data_ptr AxGetJavaBitmapFromXimage();

#if 0
	{
		extern GC Gc1;
		XCopyArea(Dpy, pmap, RootWin, Gc1, 0, 0,wid, hyt,0, 100);
	}
#endif
	/* capture the ximage from the pixmap, and free pixmap*/
	ximg = XGetImage(Dpy, pmap, 0, 0, wid, hyt, AllPlanes, ZPixmap);
#if AXMS
	axmsFlipXImage(ximg);
	/* On Windows there is a copy of the pixmap in the ximage.
	 * The subsequent destroy image call will free it. - so set it to NULL
	 */
	ximg->bitmap_handle = NULL;
#endif

	retP = AxGetJavaBitmapFromXimage(ximg, back_is_transparent);
	XDestroyImage(ximg);
    return(retP);
}

static data_ptr
create_inset_on_pmap(objP, infoP, imP, pmap, wid, hyt)
data_ptr objP;
data_ptr infoP;
data_ptr imP;
Pixmap pmap;
int wid, hyt;
{
    int err;
    AXWidget w;
    XImage *ximg;
    data_ptr retP;
	inset_image *im;

	w = (AXWidget)calloc(1, sizeof(axw_canvas));
	w->type = W_CANVAS; /* bogus widget type */
	w->parent = (AXWidget_main)w;
	w->width = wid;
	w->height = hyt;
	AxSetDrawableOfWidget(w, pmap);
	AxSubCreateInsetImage(w, objP, infoP, imP);
	free(w);
	retP = get_java_bmap_from_pmap(pmap, wid, hyt, FALSE);

    return(retP);
}

static data_ptr
render_inset_on_pmap(objP, infoP, imP, pmap, wid, hyt)
data_ptr objP;
data_ptr infoP;
data_ptr imP;
Pixmap pmap;
int wid, hyt;
{
    int err;
    AXWidget w;
    XImage *ximg;
    data_ptr retP;
	inset_image *im;

	w = (AXWidget)calloc(1, sizeof(axw_canvas));
	w->type = W_CANVAS; /* bogus widget type */
	w->parent = (AXWidget_main)w;
	w->width = wid;
	w->height = hyt;
	AxSetDrawableOfWidget(w, pmap);
	AxSubRenderInsetGraphic(w, objP, infoP, imP);
	free(w);
	retP = get_java_bmap_from_pmap(pmap, wid, hyt, FALSE);

    return(retP);
}

data_ptr
elfRenderInsetForJava (objP, infoP, sizeP, imP)
data_ptr objP;
data_ptr infoP;
data_ptr sizeP;
data_ptr imP;
{
	int wid, hyt, err;
	Pixmap pmap;
	data_ptr retP;
	inset_image *im;
	ax_inset ins_info;

	err = extractInsetImageFromDataPtr(imP, &im);
	if ( im )
	{
		if ( im->flyIn && im->flyIn->valP && im->dimB && im->dimB->valP )
		{
			retP = AxMakeArray(2);
			AxAddDataToArray(retP, 0, AxTaskCopyData(0, im->dimB->valP));
			AxAddDataToArray(retP, 1, AxTaskCopyData(0, im->flyIn->valP));
		}
		else
			retP = AxTaskCopyData(0, im->valP);
	}
	else
	{
		wid = AxIntFromArray(sizeP,0);
		hyt = AxIntFromArray(sizeP,1);

		pmap = AxCreateQueryPixmap(Dpy, RootWin, wid, hyt, DpyDepth);
		retP = render_inset_on_pmap(objP, infoP, NULL, pmap, wid, hyt);
		AxFreePixmap(pmap);
	}

	return retP;
}

/****************************************************************************
 *
 * module: AxRenderInsetGraphic - callback painter for widgets
 */
void
AxRenderInsetGraphic(w, insetP)
AXWidget w;
data_ptr insetP;
{
    data_ptr objP, infP, imP;

    if ( insetP && AxIsArray(insetP))
    {
        objP = AxArrayElement(insetP, 0);
        infP = AxArrayElement(insetP, 1);
        imP = AxArrayElement(insetP, 2);
        AxSubRenderInsetGraphic(w, objP, infP, imP);
    }
}

/****************************************************************************
 *
 *     module: elfCreateInsetImageForJava
 */
data_ptr
elfCreateInsetImageForJava(objP, infP, sizeP, imP)
data_ptr objP;
data_ptr infP;
data_ptr sizeP;
data_ptr imP;
{
    data_ptr valP, dataP, retP;
    int wid, hyt, err;
	ax_inset ins_info;
	inset_image *im;
    Pixmap pmap;
    AXWidget w;

    wid = AxIntFromArray(sizeP,0);
    hyt = AxIntFromArray(sizeP,1);
	w = (AXWidget)calloc(1, sizeof(axw_canvas));
	w->type = W_CANVAS; /* bogus widget type */
	w->parent = (AXWidget_main)w;
	w->width = wid;
	w->height = hyt;

	switch (AxIntFromArray(objP, 0))
	{
		default: 
		case IC_GE:
				pmap = (Pixmap)None;
				AxSetDrawableOfWidget(w, NULL);
		break;
		case IC_SS:
		case IC_DC:
		case IC_HTML:
		case IC_AB:
				pmap = AxCreateQueryPixmap(Dpy, RootWin, wid, hyt,DpyDepth);
				AxSetDrawableOfWidget(w, pmap);
		break;
	}

	valP = AxSubCreateInsetImage(w, objP, infP, imP);
	err = extractInsetImageFromDataPtr(valP, &im);
	if ( imP == (data_ptr)0 )	
	{
		if ( err == AOK )
			im->valP = get_java_bmap_from_pmap(im->pixmap, wid, hyt, FALSE);
		else if ( pmap != (Pixmap)None)
			AxFreePixmap(pmap);
		dataP = AxTaskCopyData(0, im->valP);
	}
	else
	{
		err = extractInsetImageFromDataPtr(imP, &im);
		if ( im == (inset_image *)0 )
		{
			err = extractInsetImageFromDataPtr(valP, &im);
			if ( err == AOK )
				im->valP = get_java_bmap_from_pmap(im->pixmap, wid, hyt, FALSE);
			else if ( pmap != (Pixmap)None)
				AxFreePixmap(pmap);
			dataP = AxTaskCopyData(0, im->valP);
		}
		else
		{
			if ( im->dimB && im->flyIn &&
				 im->dimB->pixmap && im->flyIn->pixmap )
			{
				dataP = AxMakeArray(2);
				/* update the dim bullet info */
				wid = im->dimB->x2 - im->dimB->x1;
				hyt = im->dimB->y2 - im->dimB->y1;
				convMilsToPix(wid, hyt, &wid, &hyt);
				im->dimB->valP = get_java_bmap_from_pmap(im->dimB->pixmap,
														wid, hyt, TRUE);
				/* update the flyin bullet info */
				wid = im->flyIn->x2 - im->flyIn->x1;
				hyt = im->flyIn->y2 - im->flyIn->y1;
				convMilsToPix(wid, hyt, &wid, &hyt);
				im->flyIn->valP = get_java_bmap_from_pmap(im->flyIn->pixmap,
														wid, hyt, TRUE);
				AxAddDataToArray(dataP, 0, AxTaskCopyData(0, im->dimB->valP));
				AxAddDataToArray(dataP, 1, AxTaskCopyData(0, im->flyIn->valP));
			}
			else if ( im->valP )
				dataP = AxTaskCopyData(0, im->valP);
			else
			{
				err = extractInsetImageFromDataPtr(valP, &im);
				if ( err == AOK )
					im->valP = get_java_bmap_from_pmap(im->pixmap,
													   wid, hyt, FALSE);
				else if ( pmap != (Pixmap)None)
					AxFreePixmap(pmap);
			}
		}
	}
	retP = AxMakeArray(2);
	AxAddDataToArray(retP, 0, valP);
	AxAddDataToArray(retP, 1, dataP);
	free(w);
    return retP;
}

/****************************************************************************
 *
 *     module: ElfbCreateInsetImage
 */
typeErr
ElfbCreateInsetImage(wP, objP, infP, imP, jnk4P, jnk5P, jnk6P)
data_ptr wP;
data_ptr objP;
data_ptr infP;
data_ptr imP;
data_ptr jnk4P;
data_ptr jnk5P;
data_ptr jnk6P;
{
    AXWidget w;

	ElfRetData = (data_ptr)0;
    if ( !wP )
        return(!AOK);

    w = (AXWidget)WidgetPointerFromHandleData(wP);
    if ( w == (AXWidget)0 )
        return(!AOK);
    if ( w->type == W_CANVAS && !w->window )
        PaintCanvas(w, NULL, TRUE, TRUE);
    if ( !objP || AxIsArray(objP) == FALSE )
        return(!AOK);

	ElfRetData = AxSubCreateInsetImage(w, objP, infP, imP);
    return(AOK);
}

/****************************************************************************
 *
 *     module: ElfbDestroyInsetImage
 */
typeErr
ElfbDestroyInsetImage(imP, resetP, jnk2P, jnk3P, jnk4P, jnk5P, jnk6P)
data_ptr imP;
data_ptr resetP;
data_ptr jnk2P;
data_ptr jnk3P;
data_ptr jnk4P;
data_ptr jnk5P;
data_ptr jnk6P;
{
	inset_image *im;
	int reset_pmaps, err;

	err = extractInsetImageFromDataPtr(imP, &im);

    if ( resetP && AxIsInt(resetP) )
        reset_pmaps = AxIntFromDataPtr(resetP) ? TRUE : FALSE;
	else
		reset_pmaps = FALSE;

	if ( im )
		insetFreeImagebyImageId(im->image_id);

	if ( reset_pmaps )
	{
		AxFreePixmap(BackGroundPixmap);
		BackGroundPixmap = (Pixmap)None;
		AxFreePixmap(FinalScreenPixmap);
		FinalScreenPixmap = (Pixmap)None;
	}
    return(AOK);
}

/****************************************************************************
 *
 * module: AxMeasureIconInsetInfo
 */

void
AxMeasureIconInsetInfo(objP, unitP, areaP)
data_ptr objP;
data_ptr *unitP;
data_ptr *areaP;
{
    /* get the units and area from inset object */
    if ( !objP || AxIsArray(objP) == FALSE )
    {
        *unitP = (data_ptr)0;
        *areaP = (data_ptr)0;
    }
    else switch (AxIntFromArray(objP, 0))
    {
        case IC_GE: grMeasureInsetObject(objP, unitP, areaP); break;
        default:
            *unitP = (data_ptr)0;
            *areaP = (data_ptr)0;
        break;
    }
}

/****************************************************************************
 *
 * module: ElfbWriteInsetToGraphic
 */
typeErr
ElfbWriteInsetToGraphic(argDatum)
data_ptr argDatum;
{
	data_ptr objP = AnyFromArgArray(argDatum,0);
	data_ptr infP = AnyFromArgArray(argDatum,1);
	mbchar *geFileName = StrFromArgArray(argDatum,2);
	
	AXWidget w;
	
    if ( !objP || AxIsArray(objP) == FALSE )
        return(!AOK);

    switch (AxIntFromArray(objP, 0))
    {
		case IC_DC:
		case IC_HTML:
#if 0
			wpWriteInsetToGraphic(objP, infP, geFileName);
			printf("geFileName=%s\n", geFileName);
#endif
		break;
		case IC_GE:
		case IC_SS:
		case IC_AB:
		break;
    }
	return(AOK);
}


/****************************************************************************
 *
 * module: ElfbMeasureInsetObject - measurement 
 *             wpMeasureInsetObject is in wpbltn.c
 *             grMeasureInsetObject is in gpainter.c
 *             ssMeasureInsetObject is in ssxobj.c
 *             bldrMeasureInsetObject is in bldrbltn.c
 */
data_ptr
ElfbMeasureInsetObject(objP)
data_ptr objP;
{
    data_ptr unitP, sizeP, retP, valP;

    /* get the units and size from inset object */
    if ( !objP || AxIsArray(objP) == FALSE )
    {
        unitP = (data_ptr)0;
        sizeP = (data_ptr)0;
    }
    else switch (AxIntFromArray(objP, 0))
    {
        case IC_GE:   grMeasureInsetObject(objP, &unitP, &sizeP);   break;
        case IC_AB:   bldrMeasureInsetObject(objP, &unitP, &sizeP); break;
        case IC_SS:   ssMeasureInsetObject(objP, &unitP, &sizeP);   break;
        case IC_DC:   wpMeasureInsetObject(objP, &unitP, &sizeP);   break;
        case IC_HTML: wpMeasureInsetObject(objP, &unitP, &sizeP);   break;
        default:
            unitP = (data_ptr)0;
            sizeP = (data_ptr)0;
        break;
    }

    /* package the return */
    retP = AxMakeArray(2);
    AxAddArrayToArray(retP, 0, unitP);
    AxAddArrayToArray(retP, 1, sizeP);

    /* check if there is icon info */
    valP = AxArrayElement(objP, 5);
    if ( valP )
    {
        AxMeasureIconInsetInfo(valP, &unitP, &sizeP);
        retP = AxAddArrayToArray(retP, 2, unitP);
        retP = AxAddArrayToArray(retP, 3, sizeP);
    }
    return(retP);
}

/****************************************************************************
 *
 * module: ElfbMeasureInsetObjectArea - measurement 
 *             AxfMeasureWpInsetObjectCoords is in wpbltn.c
 *             grMeasureInsetObjectArea is in gpainter.c
 *             SS: use ssMeasureInsetObject in ssxobj.c
 *             Bldr: use bldrMeasureInsetObject in bldrbltn.c
 */
data_ptr
ElfbMeasureInsetObjectArea(objP)
data_ptr objP;
{
    data_ptr unitP, sizeP, areaP, area2P, retP, valP;
	double wd, ht;

    /* get the units and area from inset object */
    if ( !objP || AxIsArray(objP) == FALSE )
    {
        unitP = (data_ptr)0;
        areaP = (data_ptr)0;
    }
    else switch (AxIntFromArray(objP, 0))
    {
        case IC_GE:
			grMeasureInsetObjectArea(objP, &unitP, &areaP);
			break;
        case IC_AB:
			bldrMeasureInsetObject(objP, &unitP, &sizeP); 
			areaP = AxMakeArray(4);
			AxAddFloatToArray(areaP, 0, 0);		/* x */
			AxAddFloatToArray(areaP, 1, 0);		/* y */
			wd = AxFloatFromArray(sizeP, 0);
			AxAddFloatToArray(areaP, 2, wd);	/* width */
			ht = AxFloatFromArray(sizeP, 1);
			AxAddFloatToArray(areaP, 3, ht);	/* height */
			break;
        case IC_SS:
			ssMeasureInsetObject(objP, &unitP, &sizeP);
			areaP = AxMakeArray(4);
			AxAddFloatToArray(areaP, 0, 0);		/* x */
			AxAddFloatToArray(areaP, 1, 0);		/* y */
			wd = AxFloatFromArray(sizeP, 0);
			AxAddFloatToArray(areaP, 2, wd);	/* width */
			ht = AxFloatFromArray(sizeP, 1);
			AxAddFloatToArray(areaP, 3, ht);	/* height */
			break;
        case IC_DC:
        case IC_HTML:
			wpMeasureInsetObjectCoords(objP, &unitP, &areaP);
			break;
        default:
            unitP = (data_ptr)0;
            areaP = (data_ptr)0;
        break;
    }

    /* package the return */
    retP = AxMakeArray(2);
    AxAddArrayToArray(retP, 0, unitP);
    AxAddArrayToArray(retP, 1, areaP);

    /* check if there is icon info */
    valP = AxArrayElement(objP, 5);
    if ( valP )
    {
        AxMeasureIconInsetInfo(valP, &unitP, &sizeP);
		area2P = AxMakeArray(4);
		AxAddFloatToArray(area2P, 0, 0);		/* x */
		AxAddFloatToArray(area2P, 1, 0);		/* y */
		wd = AxFloatFromArray(sizeP, 0);
		AxAddFloatToArray(area2P, 2, wd);		/* width */
		ht = AxFloatFromArray(sizeP, 1);
		AxAddFloatToArray(area2P, 3, ht);		/* height */
        retP = AxAddArrayToArray(retP, 2, unitP);
        retP = AxAddArrayToArray(retP, 3, area2P);
    }
    return(retP);
}


/*
*   AxWriteElfDataToClipboard - write elf data to clipboard
*/
typeErr AxWriteElfDataToClipboard(infoP, aP)
    xfer_data_convert_msg_t *infoP;
    data_ptr aP;
{
    typeErr err;
    char *p;
    data_ptr filenameP;

    p = AxTempFile(NULL, NULL);
    if(NULL == p)
        return(ErrnoErr());
    filenameP = AxMakeStrData(-1, p);
    err = AxElfbWriteDataFile(filenameP, aP, (data_ptr) NULL, TRUE);
    AxFreeData(filenameP);
    if(err == AOK)  
        err = FSReadFileIntoFileObj(p, infoP->binDoc, infoP->pathname);
    AxDeleteTempfile(p);    
    return (err);
}

/*
*   AxInsetElfDataFromClipboard - 
*   retDataP will contain the inset elf datum, Caller has to free the datum     
*   if inset is embedded, tempFileP will contain the name of the tempfile
*   that contains the embbedded object. Caller has to delete the tempfile.
*   caller supplies the memory to hold the tempfile name    
*   returns AOK on success, on failure returns error code and there is
*       nothing to free.            
*   
*/
typeErr AxInsetElfDataFromClipboard(infoP, retDataP, tempfile)
    xfer_data_convert_msg_t *infoP;
    data_ptr *retDataP;
    char tempfile[];
{
    char *p, *bufP;
    int bufSize;
    int isEmbedded;
    typeErr err;
    FILE *fp;
    data_ptr linkP, fileData;
    char temp[STRPATH+1];

    *retDataP = NULL;
    tempfile[0] = '\0';
    p = AxTempFile(NULL, NULL);
    if(NULL == p)
        return(ErrnoErr());
    strcpy(temp, p);
    err = FSWriteFileObjAsFile(temp, 0700, infoP->binDoc, infoP->pathname);
    if(err != AOK)
        return(err);
    ElfbReadDataFile(temp);
    AxDeleteTempfile(temp);
    *retDataP = ElfRetData;
    linkP = AxArrayElement(*retDataP, 0);

    isEmbedded = FALSE;
    if(AxIsString(linkP))   {
        p = AxStrPtrFromDataPtr(linkP); 
        if(0 == strcmp(p, NXLT("embedded")) )
            isEmbedded = TRUE;
    }

    fileData = AxArrayElement(*retDataP, 11);
    if(isEmbedded && fileData)  {
        /* write filedata to a tempfile */
        p = AxTempFile(NULL, NULL);
        if(p == NULL)   {
            err = ErrnoErr();
            goto handle_err;
        }
        strcpy(tempfile,p);
        if(AxIsBinary(fileData))    {
            bufSize = AxBinarySize(fileData);
            bufP = AxBinaryBytesPtr(fileData);
        }else if(AxIsString(fileData))  {
            bufP = AxStrPtrFromDataPtr(fileData);
            bufSize = strlen(bufP);
        }
        /* write it to file, no need for file-locking since tempfile */
        fp = fopen(tempfile, WRITE_STR);
        if(NULL == fp)  {
            err = ErrnoErr();
            goto handle_err;
        }

        if( fwrite(bufP, bufSize, 1, fp) < 1)   {
            err = ErrnoErr();
            fclose(fp);
            goto handle_err;
        }
        
        if(fclose(fp) != 0) {
            err = ErrnoErr();
            goto handle_err;
        }
    }
    return(AOK);

handle_err:
    if(tempfile[0] != '\0') {
        AxDeleteTempfile(tempfile);
        tempfile[0] = '\0';
    }

    if(*retDataP)   {
        AxFreeData(*retDataP);
        *retDataP = NULL;
    }

    return(err);
}

/***************************************************************************
function: AxFetchURL - caller supplies memory for holding tempfile's name 
*/
typeErr 
AxFetchURL (url, tempfile, errorP)
    mbchar *url;            /* URL to fetch */
    mbchar tempfile[];      /* Fetched tempfile of atleast STRPATH */
    data_ptr *errorP;       /* if failed, contains error information */
{
    data_ptr urlP, tmpP, throwErrorP;
    typeErr err;
    data_ptr retP=NULL;
    char *p;

    if(errorP)
        *errorP = NULL;
    tempfile[0] = '\0'; 
    p = AxTempFile(NULL, NULL);
    if(p == NULL)
        return(ErrnoErr());
    strcpy(tempfile, p);
    urlP = AxMakeStrData(-1, url);
    tmpP = AxMakeStrData(-1, tempfile);
    throwErrorP = AxMakeIntData(-1);    /* return error information */
    err = ElfCallByName(NXLT("WWW_FETCH_URL@"), &retP, 
                                        4, urlP, tmpP, NULL, throwErrorP);
    AxFreeData(urlP);
    AxFreeData(tmpP);
    AxFreeData(throwErrorP);
    if(err != AOK)      {
        AxDeleteTempfile(tempfile);
        tempfile[0] = '\0';
        if(errorP)
            *errorP = retP;     /* return error to caller */
        else
            AxFreeData(retP);   /* discard error message */
    }else
        AxFreeData(retP);
    return(err);
}

/*
*   AxIsURL - returns true if path is a URL, else returns false
*/
#define URL_STR NXLT("HTTP://")
AxIsURL(path)
    mbchar *path;
{
    if(0 == strncasecmp(path, URL_STR, strlen(URL_STR)) )   
        return(TRUE);
    return(FALSE);
}

/******************************************************************************/
/*																			  */
/*								"Drawable" Cache							  */
/*							------------------------						  */
/*																			  */
/* The cache is an array of entries of type ax_inset.h: drawablecache_t.	  */
/* The lookup key is a name.												  */
/*																			  */
/* To use:																	  */
/*																			  */
/*   To add an entry:														  */
/*																			  */
/*   serialnumber = AxDrawableCacheAddReplace(name, srcname,				  */
/*											  width, height, drawable);		  */
/*																			  */
/*		char *name 		  - ASCII name (e.g. Object name)  					  */
/*		char *srcname 	  - ASCII name (e.g. Source name)					  */
/*		int width, height - drawable's width, height						  */
/*		Drawable drawable - pixmap, ...										  */
/*																			  */
/*   AxDrawableCacheAddReplace adds a new entry, or overwrites an extant entry*/
/*   identified by name.													  */
/*																			  */
/*   To get info:															  */
/*																			  */
/*	 drawable = AxDrawableCacheGetDraw(name)								  */
/*																			  */
/*		char *name 		  - ASCII name (e.g. Object name)  					  */
/*		Drawable drawable - drawable field in cache entry					  */
/*																			  */
/*   cacheP = AxDrawableCacheGet(name)										  */
/*																			  */
/*      char *name		  		- ASCII name (e.g. Object name)				  */
/*	    drawablecache_t *cacheP - pointer to cache entry					  */
/*																			  */
/* Internals:																  */
/*																			  */
/* AxDrawableCacheInitEntry		Returns subscript of new entry, or of cleared */
/*								LRU											  */
/* AxDrawableCacheCleanLRU		Finds and clears LRU						  */
/* AxDrawableCacheAdd			AxDrawableCacheInitEntry, insert info		  */
/* AxDrawableCacheInsert		Inserts info in extant entry				  */
/* AxDrawableCacheFindName		Returns subscript of entry(name)			  */
/*																			  */
/* AxDrawableCacheFindSerial	Returns subscript of entry(serialnumber)	  */
/*								(Not used.)									  */
/*																			  */
/* (The serial number is not necessary.  It's a holdover.) 					  */
/*																			  */
/******************************************************************************/

/***************************************************************************
function: AxDrawableCacheAddReplace - Adds or replaces entry

Returns serial number of added entry.

If entry with specified name is already in cache, replaces.
*/

int
AxDrawableCacheAddReplace(name, srcname, width, height, drawable)
  char 	   *name;
  char	   *srcname;
  int 	   width, height;
  Drawable drawable;
{
	int	cacheix = 0;
	int newserialnumber;

	cacheix = AxDrawableCacheFindName(name);
	if(cacheix) {								/*If already in cache, replace*/
		newserialnumber = AxDrawableCacheInsert(cacheix, name, srcname,
												width, height, drawable);
	}else{										/* not in cache, create		  */
		newserialnumber = AxDrawableCacheAdd(name, srcname, width, height,
											 drawable);
	}
	return(newserialnumber);
}

/***************************************************************************
function: AxDrawableCacheGet - Seeks entry.  Key = name.  Sets access time.

Returns pointer to entry with name, NULL if not found.
*/
drawablecache_t *
AxDrawableCacheGet(name)
  char *name;
{
	int ix;
	ix = AxDrawableCacheFindName(name);
	if(ix){
		AxDrawableCache[ix].lastAccessTime = time(0);
		return(&AxDrawableCache[ix]);
	}
	return(NULL);							/* not in cache						  */
}

/***************************************************************************
function: AxDrawableCacheGetDraw - Seeks entry.  Key = name.  Sets access time.

Returns drawable field from cache entry, NULL if not found.
*/
Drawable
AxDrawableCacheGetDraw(name)
  char *name;
{
	int ix;
	ix = AxDrawableCacheFindName(name);
	if(ix){
		AxDrawableCache[ix].lastAccessTime = time(0);
		return(AxDrawableCache[ix].drawable);
	}
	return((int)NULL);							/* not in cache						  */
}

/***************************************************************************
function: AxDrawableCacheInitEntry - Creates entry in Drawable cache

If number cache entries > MAXNUMINDRAWABLECACHE, uses oldest.
Returns entry's subscript.
*/

int
AxDrawableCacheInitEntry()
{
	int cacheix;
												/* Find or create a useable   */
												/* cache entry.				  */
	if(DrawableCacheNumEntries >= MAXNUMINDRAWABLECACHE) { /*Room for another?*/
		cacheix = AxDrawableCacheCleanLRU();	/* Make LRU reuseable		  */
	}else{
		cacheix = DrawableCacheNumEntries;		/* subscript of new entry	  */
		DrawableCacheNumEntries++;
		AxDrawableCache = (drawablecache_t *)AxRealloc(AxDrawableCache,
													  DrawableCacheNumEntries *
													  sizeof(drawablecache_t));
	}

												/* This entry's id. 1st = 1   */
	AxDrawableCache[cacheix].serialnumber = ++DrawableCacheSerialNumber;

	AxDrawableCacheInsert(cacheix, NULL, 0, 0, NULL, NULL);

	return(cacheix);							/* return subscript			  */
}

/***************************************************************************
function: AxDrawableCacheCleanLRU - cleans oldest entry

Returns oldest's subscript.
*/

int
AxDrawableCacheCleanLRU()
{

	int ix, oldestix;
	
	oldestix = 0;									/* start...				  */
	if(DrawableCacheNumEntries == 0)return(0);		/* empty => use first	  */

	for (ix = 0; ix <= DrawableCacheNumEntries - 1; ix++) {
		if(AxDrawableCache[ix].lastAccessTime < 
		   AxDrawableCache[oldestix].lastAccessTime){
			oldestix = ix;							/* found older 			  */
		}
	}
	if(AxDrawableCache[oldestix].name) {			/*Free cache resources	  */
		AxFree(AxDrawableCache[oldestix].name);
		AxDrawableCache[oldestix].name = NULL;
	}
	if(AxDrawableCache[oldestix].srcname) {
		AxFree(AxDrawableCache[oldestix].srcname);
		AxDrawableCache[oldestix].srcname = NULL;
	}
	if(AxDrawableCache[oldestix].drawable) {
		XFreePixmap(Dpy, AxDrawableCache[oldestix].drawable);
		AxDrawableCache[oldestix].drawable = 0;
	}

	AxDrawableCache[oldestix].serialnumber = -1; /*Trouble if we ever see this*/

	return(oldestix);
}

/***************************************************************************
function: AxDrawableCacheAdd - Adds entry to Drawable Cache

Returns serial number of added entry.
Does not test whether entry is already in cache.
*/

int
AxDrawableCacheAdd(name, srcname, width, height, drawable)
  char 	   *name;
  char	   *srcname;
  int 	   width, height;
  Drawable drawable;
{
	int cacheix;

	cacheix = AxDrawableCacheInitEntry(name);	/* Find or make an entry	  */
												/* Move info into entry		  */
	AxDrawableCacheInsert(cacheix, name, srcname, width, height, drawable);
	return(AxDrawableCache[cacheix].serialnumber);
}

/***************************************************************************
function: AxDrawableCacheInsert - Inserts info in extant entry

Returns serial number.
*/

int
AxDrawableCacheInsert(cacheix, name, srcname, width, height, drawable)
  int	   cacheix;
  char 	   *name;
  char	   *srcname;
  int 	   width, height;
  Drawable drawable;
{

	if(name) {
		AxDrawableCache[cacheix].name = strcpy(AxMalloc(strlen(name)+1), name);
	}else{
		AxDrawableCache[cacheix].name = NULL;
	}

	if(srcname) {
		AxDrawableCache[cacheix].srcname = strcpy(AxMalloc(strlen(srcname)+1),
												  srcname);
	}else{
		AxDrawableCache[cacheix].srcname = NULL;
	}

	AxDrawableCache[cacheix].width = width;
	AxDrawableCache[cacheix].height = height;
	AxDrawableCache[cacheix].lastAccessTime = time(0);/*(Is seconds precise enough?)*/
	AxDrawableCache[cacheix].drawable = drawable;

	return(AxDrawableCache[cacheix].serialnumber);
}

/***************************************************************************
function: AxDrawableCacheFindName - Seeks entry.  Key = name

Returns subscript for entry with name, 0 if not found.
*/
int
AxDrawableCacheFindName(name)
  char *name;
{
	int ix;
	for(ix = 0; ix < DrawableCacheNumEntries; ix++){
		/* Need to look at srcname, too? */
		if(streq(AxDrawableCache[ix].name, name) ) return(ix);
	}
	return(0);							/* not in cache						  */
}

#if 0 /* Use name as lookup key... */
/***************************************************************************
function: AxDrawableCacheFindSerial - Seeks entry.  Key = Serial Number.

Returns subscript for entry with serialnumber, 0 if not found.
*/
int
AxDrawableCacheFindSerial(serialnumber)
  int serialnumber;
{
	int cacheix;
	for(cacheix = 0; cacheix < DrawableCacheNumEntries; cacheix++){
		if(AxDrawableCache[cacheix].serialnumber == serialnumber) {
			return(cacheix);
		}
	}
	return(0);							/* not in cache						  */
}
#endif
