/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ax_inset.h
   Applix ASC version: 15
   Date: Tue Jun  1 21:12:10 1999
*/
#ifndef AX_INSET_INCL	/* { */
#define AX_INSET_INCL

#define MAXNUMINDRAWABLECACHE 100	/* Max number entries in "Drawable" cache */

#if 0
#if __STDC__ || defined(__cplusplus)
#define AXP_(s) s
#else
#define AXP_(s) ()
#endif
#endif

typedef struct inset_image {
	long widget;
	int image_id;
	long pixmap;
	int x1;
	int y1;
	int x2;
	int y2;
	data_ptr valP;
	struct inset_image *txbx;
	struct inset_image *flyIn;
	struct inset_image *dimB;
	struct inset_image *final;
} inset_image;

typedef struct ax_units {
	int dpi;
	int dpu;
	int precision;
	char *name;
} ax_units;

typedef struct ax_inset_object {
	int type;
	char *ptr;
	char *segment_info;
	ax_units units;
	char **parents;
	struct ax_inset_object *icon_info;
} ax_inset_object;

typedef struct ax_inset {
	/* sparse usage indicators */
	unsigned int
	 use_unit_dpi         : 1, /* 0 */
	 use_unit_dpu         : 1,
	 use_unit_precision   : 1,
	 use_unit_name        : 1,
	 use_dst_x            : 1,
	 use_dst_y            : 1,
	 use_dst_wid          : 1,
	 use_dst_hyt          : 1,
	 use_clip_x           : 1,
	 use_clip_y           : 1,
	 use_clip_wid         : 1,
	 use_clip_hyt         : 1,
	 use_suppress_cursor  : 1,
	 use_direct_render    : 1,
	 use_icon_only        : 1,
	 use_x_zoom           : 1, /* 15 */
	 use_y_zoom           : 1,
	 use_scale_mode       : 1,
	 use_proportional     : 1,
	 use_centered         : 1,
	 use_widget_colors    : 1,
	 use_invert_colors    : 1,
	 use_clear            : 1,
	 use_fill             : 1,
	 use_border_style     : 1,
	 use_inactive_content : 1,
	 use_inactive_border  : 1,
	 use_video_enabled    : 1,
	 use_video_style      : 1,
	 use_audio_enabled    : 1,
	 use_audio_pathname   : 1,
	 use_audio_pended     : 1;/* 31 */

	unsigned int
	 use_video_throttle   : 1, /* 0 */
	 use_build_level      : 1;
 
	/* drawing flags */
	unsigned int
	 use_audio_gain       : 1,
	 suppress_cursor      : 1,
	 direct_render        : 1,
	 icon_only            : 1,
	 scale_mode           : 2,
	 proportional         : 1,
	 centered             : 1,
	 widget_colors        : 1,
	 invert_colors        : 1,
	 clear                : 1,
	 fill                 : 1,
	 border_style         : 2,
	 inactive_content     : 1,
	 inactive_border      : 1;

	/* units used by destination and clip */
	int unit_dpi;
	int unit_dpu;
	int unit_precision;
	char unit_name[STRPATH+1];

	/* destination rectangle */
	int dst_x;
	int dst_y;
	int dst_wid;
	int dst_hyt;
 
	/* destination clip */
	int clip_x;
	int clip_y;
	int clip_wid;
	int clip_hyt;

	/* zoom factors (1000 equals 100.0%) */
	int x_zoom;
	int y_zoom;

	/* transitional effects */
	int video_enabled;
	int audio_enabled;
	int video_style;
	int video_throttle;
	char audio_pathname[STRPATH+1];
	int audio_pended;
	int audio_gain;
	int build_level;
} ax_inset;


typedef struct drawablecache_t {				/* Drawable cache (rjd Apr 97)*/
	int	 serialnumber;							/* Creation count			  */
	char *name;
	char *srcname;
	int	 width, height;
	int	 lastAccessTime;						/* for LRU discard			  */
	Drawable drawable;							/* pixmap, window			  */
} drawablecache_t;

static int DrawableCacheNumEntries = 0;
static int DrawableCacheSerialNumber = 0;		/* 0 is invalid. 1st = 1.	  */
static drawablecache_t *AxDrawableCache = NULL;	/* array of ...Cache_t		  */
											/*(In what file does this belong?)*/

char *AxElfDataToInsetStr();
char *AxDoElfDataToInsetStr();
data_ptr AxInsetStrToElfData();
data_ptr AxDoInsetStrToElfData();
data_ptr createTransForELF();
int AxExtractTransitionFromELF();

#endif	/* } */
