/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ax_xm.h
   Applix ASC version: 6
   Date: Mon Nov 16 13:53:46 1998
*/
#ifndef _Xm_h    /*  {  */
#define _Xm_h


#if AX_X11 /* NT has its own method for accomplishing this */
#include <X11/Intrinsic.h>


/*
 * XtArgVal ought to be a union of XtPointer, char *, long, int *, and proc *
 * but casting to union types is not really supported.
 *
 * So the typedef for XtArgVal should be chosen such that
 *
 *      sizeof (XtArgVal) >=    sizeof(XtPointer)
 *                              sizeof(char *)
 *                              sizeof(long)
 *                              sizeof(int *)
 *                              sizeof(proc *)
 *
 * above comment from <X11/Intrinsic.h> -lance
 */

#if platform_where_long_is_not_largest__see_above_comment
typedef char* AxArgVal;
#else
typedef long AxArgVal;
#endif


#define Arg AxArg
#define ArgList AxArgList

typedef struct {
    int    name;
    AxArgVal    value;
} AxArg, *AxArgList;

#endif

#define XmVERSION	1
#define XmREVISION	0
#define XmVersion (XmVERSION * 1000 + XmREVISION)

extern int xmUseVersion;


/* define used to denote an unspecified pixmap  */

#define XmUNSPECIFIED_PIXMAP	2


/****************
 *
 * XmString structure defines. These must be here (at the start of the file) 
 * becaused they are used later on.
 *
 ****************/

#define XmSTRING_DIRECTION_L_TO_R	0
#define XmSTRING_DIRECTION_R_TO_L	1

typedef char *		XmString;		/* opaque to outside */
typedef XmString *	XmStringTable;		/* opaque to outside */
typedef char *		XmStringCharSet;	/* Null term string */
typedef unsigned char	XmStringComponentType;	/* component tags */
typedef unsigned char 	XmStringDirection;

typedef struct
{
    XFontStruct      *font;
    XmStringCharSet  charset;
}
    XmFontListRec, *XmFontList;

/*
 * internal context data block, for read-out
 */

typedef struct
{
    unsigned short	offset;			/* current place TCS */
    unsigned short	length;			/* max length */
    XmStringCharSet	charset;		/* last charset seen */
    unsigned short	charset_length;		/* and it's length */
    XmStringDirection	direction;		/* last direction */
    Boolean		error;			/* something bad */
}
    _XmStringContextRec, *_XmStringContext;

/*
 * external context data block
 */

typedef struct
{
    XmString		string;
    _XmStringContext	internal_context;
}
    XmStringContextRec, *XmStringContext;


#define XmSTRING_COMPONENT_UNKNOWN	0
#define XmSTRING_COMPONENT_CHARSET	1
#define XmSTRING_COMPONENT_TEXT		2
#define XmSTRING_COMPONENT_DIRECTION	3
#define XmSTRING_COMPONENT_SEPARATOR	4	/* 5-125 reserved */

#define XmSTRING_COMPONENT_END		126	/* no more comp in string */

#define XmSTRING_COMPOUND_STRING 	127	/* tag for whole TCS */

#define XmSTRING_COMPONENT_USER_BEGIN	128	/* 128-255 are user tags */
#define XmSTRING_COMPONENT_USER_END	255
/* Moved -1 to "". Problems with char_set. cc errors. SCM 05/03/93 */
#define XmSTRING_DEFAULT_CHARSET	""	/* (-1) */





typedef struct
{
   int reason;
   XEvent * event;
   int value;
   int pixel;
} XmScrollBarCallbackStruct;

typedef struct
{
    int     reason;
    XEvent  *event;
    Widget  widget;
    char    *data;
    char    *callbackstruct;
}XmRowColumnCallbackStruct;


typedef struct
{
   int reason;
   XEvent * event;
   int set;
} XmToggleButtonCallbackStruct;

typedef struct
{
    int     reason;
    XEvent  *event;
} XmAnyCallbackStruct;

typedef struct
{
   int 	     reason;
   XEvent    *event;
   XmString  item;
   int       item_length;
   int       item_position;
   XmString  *selected_items;
   int       selected_item_count;
   int       selection_type;
} XmListCallbackStruct;


typedef struct 
{
   int reason;
   XEvent * event;
   int value;
} XmScaleCallbackStruct;



#define	XmAS_NEEDED	1
#define XmCONSTANT	        1
#define XmDIALOG_WORK_AREA		0

#define XmCR_NONE               0
#define XmCR_INCREMENT			3
#define XmCR_DECREMENT			4
#define XmCR_TO_TOP             7
#define XmCR_TO_BOTTOM          8
#define XmCR_DRAG			    9

#define XmVERTICAL	       1
#define XmHORIZONTAL           2

#define XmCR_VALUE_CHANGED		2
#define XmCR_PAGE_INCREMENT		5
#define XmCR_PAGE_DECREMENT		6
#define XmCR_ARM			11
#define XmCR_DISARM			12

#define	XmCR_SINGLE_SELECT		23
#define	XmCR_MULTIPLE_SELECT		24
#define	XmCR_EXTENDED_SELECT		25
#define	XmCR_BROWSE_SELECT		26
#define	XmCR_DEFAULT_ACTION		27





/****************
 *
 *  Public functions defined in XmString.c
 *
 ****************/

extern XmString 		XmStringCreate ();
extern XmString 		XmStringDirectionCreate ();
extern XmString 		XmStringSeparatorCreate ();
extern XmString 		XmStringSegmentCreate ();
extern XmString 		XmStringLtoRCreate ();
extern XmString 		XmStringCreateLtoR ();

extern Boolean  		XmStringInitContext ();
extern void			XmStringFreeContext ();
extern XmStringComponentType	XmStringGetNextComponent ();
extern XmStringComponentType	XmStringPeekNextComponent ();
extern Boolean			XmStringGetNextSegment ();
extern Boolean			XmStringGetLtoR ();

extern XmFontList		XmFontListCreate ();
extern XmFontList		XmStringCreateFontList ();
extern void 			XmFontListFree ();
extern XmFontList		XmFontListAdd ();
extern XmFontList 		XmFontListCopy ();

extern XmString		XmStringConcat ();
extern XmString		XmStringNConcat ();
extern XmString		XmStringCopy ();
extern XmString		XmStringNCopy ();
extern Boolean			XmStringCompare ();
extern Boolean			XmStringByteCompare ();
extern int			XmStringLength ();
extern Boolean			XmStringEmpty ();

extern void 			XmStringFree ();

extern XmString		_XmStringCreateExternal ();

extern Dimension		XmStringWidth ();
extern Dimension		XmStringHeight ();
extern Dimension		XmStringBaseline ();
extern void			XmStringExtent ();
extern int			XmStringLineCount ();

extern void			XmStringDraw ();
extern void			XmStringDrawImage ();
extern void			XmStringDrawUnderline ();

/***********************************************************************

  The following workarounds have been suggested by licensees and have
  not been thoroughly tested. 

***********************************************************************/
#ifdef NO_REGEX
#define regex(a,b) re_exec(b)
#define regcmp(a,b) re_comp(a)
#endif

#ifdef NO_GETCWD
char *getcwd(buf,size)
char *buf;
int size;
{
extern char *malloc();
if (buf==(char*) NULL)
        buf=(char *) malloc (size);
return (char *)getwd(buf);
}
#endif

#ifdef NO_VFORK
#define vfork fork
#endif

#endif /* }   _Xm_h */
 /* DON'T ADD STUFF AFTER THIS #endif */
