/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axfutil.c
   Applix ASC version: 4
   Date: Fri Feb 26 21:14:42 1999
*/
#include <aos.h>
#include <eldata.h>
#include <afile.h>
#include <getstr20.h>
#include <getstr10.h>
#include <doctypes.h>
#include <axfutil.p>


/***********************************************************************/
/* Get access control info from a file by filename.                    */
/***********************************************************************/
int
AFGetAccessInfo(pathname, 
                read_only,      /* IN/OUT - true if only access is read     */
                group,          /* OUT    - 0=>no access, 1=>read, 2=>write */
                other,          /* OUT    - 0=>no access, 1=>read, 2=>write */
                exists)         /* OUT    - true if file exists             */
char *pathname;
int *read_only;
int *group;
int *other;
int *exists;
{
    struct stat stat_result;
#if AXMS
    int filemask;
#else
    mode_t filemask;
#endif
    int err;

    /* set the "file exists" flag and get permissions */
    *exists = (Astat(pathname, &stat_result) == 0) ? TRUE : FALSE;

    /* set the file permissions flags */
    if ( *exists )
        filemask = stat_result.st_mode;
    else
    {
        filemask = umask(0);
        umask(filemask);
        filemask = ~filemask; /* we want which bits are allowed */
    }
#if !AXMS
    if ( filemask&S_IWGRP )      /* read/write by group */
        *group = 2;
    else if ( filemask&S_IRGRP ) /* read by group */
        *group = 1;
    else                         /* no access by group */
#endif
        *group = 0;
#if !AXMS
    if ( filemask&S_IWOTH )      /* write by others */
        *other = 2;
    else if ( filemask&S_IXOTH ) /* read by others */
        *other = 1;
    else                         /* no access by others */
#endif
        *other = 0;

    /* if opening for read only, it's not writable by request */
    if ( *read_only != TRUE )
    {
        if ( *exists == FALSE ) /* not on disk, writable by definition */
            *read_only = FALSE;
        else /* get read permissions from file */
            *read_only = (Aaccess(pathname, W_OK) < 0) ? TRUE : FALSE;
    }

    /* check access for reading */
    if ( *read_only == FALSE )
        err = AOK;
    else
        err = (Aaccess(pathname, R_OK) != 0) ? ErrnoErr() : AOK;
    return(err);
}

/***************************************************************************/
/* Read the buffer, and if not an APPLIXware file convert it to HEX buffer */
/* Caller would free out_buf                                               */
/***************************************************************************/
int
HEXBinToHexBuffer(pid, in_buf, in_nbytes, out_buf, out_nbytes)
int pid;
char *in_buf, **out_buf;
int in_nbytes, *out_nbytes;
{
    char *hex_buf, *p1, *p2, c;
    char hdr[AREADBUFSIZE], end[AREADBUFSIZE];
    int hex_len, hdr_len, end_len, ix, extras_len;
    AFileInfo info;
    int hexify;
    int err;

    *out_buf = NULL;
    if ( in_buf[0] != '*' ) /* not an applix file */
        hexify = TRUE;
    else /* could be an applix file */
    {
        err = AFileInfoFromHeader(in_buf, &info);
        if ( err != AOK && err != FETooNew )
            hexify = TRUE;
        else if ( info.version <= 210 )
            hexify = TRUE;
        else /* ok to embed */
            hexify = FALSE;
    }
    if(!hexify)
        return(FEBadType);

    /* hexify it */
    memset((char *)&info, '\0', sizeof(AFileInfo));
    info.docType = AX_DOC_TYPE_HEX;
    info.version = info.min_version = AGetCurrentVersion();
    info.encoding = TRUE;
    AFileHeaderFromInfo(&info, hdr);
    AFileEnderFromInfo(&info, end);

    hdr_len = strlen(hdr);
    end_len = strlen(end);
    hex_len = in_nbytes * 2; /* 2 hex bytes per binary byte */

    /* compute len for space/newline for each line */
    extras_len =  2 * (hex_len/78 + ((hex_len % 78) ? 1 : 0));
    hex_len = hex_len + extras_len; /* add space/newline each line */
    *out_nbytes = hdr_len+1 + hex_len + end_len+1; /* +1 for new lines */
    hex_buf = (char *)TaskAlloc(pid, *out_nbytes + 1);  /* +1 for null char */

    /* copy the header */
    strcpy(hex_buf, hdr);
    hex_buf[hdr_len] = '\n';

    /* hex encode the data */
    p1 = &hex_buf[hdr_len+1];
    p2 = in_buf;
    ix = 0;
    while ( in_nbytes > 0 )
    {
        /* add a space at the beginning of each line */
        if ( ix == 0 )
        {
            *p1++ = ' ';
            ix = 1;
        }

        /* high nibble */
        c = ((*p2) >> 4) & 0xf;
        if ( c > 9 )
            *p1++ = '0' + 7 + c;
        else
            *p1++ = '0' + c;

        /* low nibble */
        c = (*p2) & 0xf;
        if ( c > 9 )
            *p1++ = '0' + 7 + c;
        else
            *p1++ = '0' + c;

        /* add a newline at the end of each line */
        ix += 2;
        if ( ix == 79 )
        {
            *p1++ = '\n';
            ix = 0;
        }

        /* next byte */
        in_nbytes--;
        p2++;
    }

    /* tag on the ender (and a null to make it a string) */
    if ( ix != 0 )
        *p1++ = '\n';
    strcpy(p1, end);
    p1[end_len+0] = '\n';
    p1[end_len+1] = (char)0;
    *out_buf = hex_buf;
    return(AOK);
}

/***************************************************************************/
/* Read the buffer, and if HEX file, remove header and de-HEXIFY it        */
/* Caller would free out_buf                                               */
/***************************************************************************/
int 
HEXHexToBinBuffer(pid, in_buf, in_n_bytes, out_buf, out_n_bytes)
int pid;
char *in_buf, **out_buf;
int in_n_bytes, *out_n_bytes;
{
    int err, read_only, count, lcount, ix, not_done;
    char b1, b2, *bin_buf;
    AFileInfo info;
    FILE *fp;

    /* init */
    *out_buf = (char *)0;
    *out_n_bytes = 0;
    err = AFileInfoFromHeader(in_buf, &info);
    if ( err != AOK)
        return(err);
    if ( info.docType != AX_DOC_TYPE_HEX )
        return(FEBadType);

    /* create an output buffer (half the size at most) */
    bin_buf = (char *)TaskAlloc(pid, in_n_bytes/2);

    /* scan past the newline (should be the header) */
    count = lcount = ix = 0;
    while ( in_buf[lcount] != '\n' && lcount < in_n_bytes )
        lcount++;
    count += lcount+1;
    lcount = 0;

    /* read the scanlines (start with space, end with newline) */
    not_done = TRUE;
    while ( count < (in_n_bytes-1) && not_done )
    {
        /* skip leading space */
        if ( lcount == 0 )
        {
            if ( in_buf[count+lcount] == ' ' )
                lcount++;
            else
            {
                if ( in_buf[count+lcount] != '*' )
                    err = !AOK;
                else /* assume *END HEX */
                {
                    count += lcount;
                    lcount = 0;
                }
                not_done = FALSE;
            }
        }

        /* skip trailing newline */
        else if ( lcount == 79 )
        {
            if ( in_buf[count+lcount] == '\n' )
            {
                count += lcount+1;
                lcount = 0;
            }
            else
            {
                not_done = FALSE;
                err = !AOK;
            }
        }

        else /* read the byte pair */
        {
            b1 = in_buf[count+lcount+0]; /* high nibble */
            b2 = in_buf[count+lcount+1]; /* low nibble */
            lcount += 2;
            if ( b1 == '\n' ) /* end of doc??? */
            {
                if ( b2 != '*' )
                    err = !AOK;
                else /* assume END HEX */
                {
                    count += lcount;
                    lcount = 0;
                }
                not_done = FALSE;
            }
            else
            {
                /* high nibble */
                if ( b1 >= 'A' )
                    bin_buf[ix] = ((b1 - 7 - '0') << 4) & 0xf0;
                else
                    bin_buf[ix] = ((b1 - '0') << 4) & 0xf0;

                /* low nibble */
                if ( b2 >= 'A' )
                    bin_buf[ix] |= (b2 - 7 - '0') & 0xf;
                else
                    bin_buf[ix] |= (b2 - '0') & 0xf;
                ix++;
            }
        }
    }

    if ( err != AOK )   
        TaskFree(pid, bin_buf);
    else
    {
        *out_n_bytes = ix;
        *out_buf = bin_buf;
    }
    return(err);
}
#if !FOR_FILTER
/***********************************************************************/
/* Read a file, and if not an APPLIXware file, convert it to HEX file  */
/***********************************************************************/
int
HEXFileToBuffer(pid, pathname, buf, n_bytes)
int pid;
char *pathname;
char **buf;
int *n_bytes;
{
    int err, read_only, group, other, exists, ix, in_n_bytes;
    char *in_buf;
    FILE *fp;

    /* init */
    read_only = TRUE;
    *buf = (char *)0;
    *n_bytes = 0;

    /* open the file for reading */
    err = AFGetAccessInfo(pathname, &read_only, &group, &other, &exists);
    if ( err )
        return(err);
    if ( exists == FALSE )
    {
        errno = ENOENT;
        err = ErrnoErr();
        return(err);
    }
    err = AFTaskLockFile(pathname, pid, &read_only, FALSE);
    if ( err )
        return(err);
    fp = AFTaskfopen(pid, pathname, READ_BINARY_STR);
    if ( !fp )
    {
        err = ErrnoErr();
        AFTaskUnlockFile(pathname, pid, read_only);
        return(err);
    }

    /* read the file into a buffer */
    fseek(fp, 0L, 2);
    in_n_bytes = ftell(fp);
    fseek(fp, 0L, 0);
    in_buf = (char *)TaskAlloc(pid, in_n_bytes+1);
    if ( fread(in_buf, in_n_bytes, 1, fp) != 1 )
    {
        AFfclose(pathname, read_only, fp);
        AFTaskUnlockFile(pathname, pid, read_only);
        err = ErrnoErr();
        return(err);
    }
    in_buf[in_n_bytes] = (char)0;

    /* close and unlock the file */
    AFfclose(pathname, read_only, fp);
    AFTaskUnlockFile(pathname, pid, read_only);

    /* convert the data into hex (including a hex header) */
    err = HEXBinToHexBuffer(pid, in_buf, in_n_bytes, buf, n_bytes);
    if ( err != FEBadType )
        TaskFree(pid, in_buf);  
    else /* not converted to HEX - return unmodified buffer */
    {   
        *buf = in_buf;
        *n_bytes = in_n_bytes;
        err = AOK;
    }
    return(err);
}

/***********************************************************************/
/* Write a buffer to disk, deHEXing it if needed                       */
/***********************************************************************/
int
HEXBufferToFile(pid, pathname, buf, n_bytes)
int pid;
char *pathname;
char *buf;
int n_bytes;
{
    int err, read_only, out_n_bytes, is_hex;
    char *out_buf;
    FILE *fp;

    /* convert the HEXified buffer to a binary buffer */
    err = HEXHexToBinBuffer(pid, buf, n_bytes, &out_buf, &out_n_bytes);
    if ( err != FEBadType )
        is_hex = TRUE;
    else /* not HEX, write raw data */
    {
        err = AOK;
        is_hex = FALSE;
    }

    /* write the buffer to disk */
    if ( err == AOK )
    {
        if ( is_hex )
        {
            buf = out_buf;
            n_bytes = out_n_bytes;
        }
        read_only = FALSE;
        err = AFTaskLockFile(pathname, pid, &read_only, FALSE);
        if ( err )
        {
            if ( out_buf )
                TaskFree(pid, out_buf);
            return(err);
        }
        fp = AFTaskfopen(pid, pathname, WRITE_STR);
        if ( !fp )
        {
            err = ErrnoErr();
            AFTaskUnlockFile(pathname, pid, read_only);
            if ( out_buf )
                TaskFree(pid, out_buf);
            return(err);
        }
        if ( fwrite(buf, n_bytes, 1, fp) != 1 )
        {
            err = ErrnoErr();
            AFfclose(pathname, read_only, fp);
            AFTaskUnlockFile(pathname, pid, read_only);
            if ( out_buf )
                TaskFree(pid, out_buf);
            return(err);
        }
        if ( AFfclose(pathname, read_only, fp) == EOF )
            err = ErrnoErr();
        else
            err = AOK;
        AFTaskUnlockFile(pathname, pid, read_only);
        if ( out_buf )
            TaskFree(pid, out_buf);
    }
    return(err);
}

#define NULL_ITEM_CHAR      '~'
#define ARRAY_START_CHAR    '<'
#define ARRAY_END_CHAR      '>'
#define TOKEN_DELIM_STR     NXLT(" ")
#define ARRAY_START_STR     NXLT("<")
#define ARRAY_END_STR       NXLT(">")
#define NULL_ITEM_STR       NXLT("~")

/******************************************************************************
* function: AxInsetStrToElfData - given a inset source string, convert it
*   to elf data - Caller is responsible for free'ing the string
*/
char *
AxDoElfDataToInsetStr(pid, aP, str)
int pid;
data_ptr aP;
char *str;
{
    char buf[STRPATH+1];
    char *p, *newP, *cP;
    int ix, jx, len, n;

    if(aP && AxIsArray(aP)) {

        /* add array start delimiter */
        len = strlen(str) + strlen(ARRAY_START_STR) + 1;
        str = (char *)TaskRealloc(pid, str, len);
        strcat(str, ARRAY_START_STR);

        /* tokenize items in the arrray */
        n = AxArraySize(aP);
        for ( ix=0; ix<n; ix++ ) {
            str = AxDoElfDataToInsetStr(pid, AxArrayElement(aP, ix), str);
        }

        /* remove extraneous delimiter */
        len = strlen(str);
        if(str[len-1] == ' ')
            str[len-1] = '\0';

        /* add array ending delimiter */
        len = strlen(str) + strlen(ARRAY_END_STR) + 1;
        str = (char *)TaskRealloc(pid, str, len);
        strcat(str, ARRAY_END_STR);

        return(str);    /* done */
    }

    newP = NULL;
    if(NULL == aP)  {
        newP = (char *)TaskAlloc(pid, strlen(NULL_ITEM_STR)+1);
        strcpy(newP, NULL_ITEM_STR);
    } else if(AxIsInt(aP))  {
        sprintf(buf,"%d", AxIntFromDataPtr(aP));
        newP = (char *)TaskAlloc(pid, strlen(buf)+1);
        strcpy(newP, buf);
    }else if(AxIsFloat(aP)) {
        sprintf(buf,"%lf", AxFloatFromDataPtr(aP));
        newP = (char *)TaskAlloc(pid, strlen(buf)+1);
        strcpy(newP, buf);
    }else if(AxIsString(aP))    {
        p = AxStrPtrFromDataPtr(aP);
        /* string len + enclosing quotes + null char */ 
        len = 2 * strlen(p) + 2 + 1;
        newP = cP = (char *)TaskAlloc(pid, len);
        *cP++ = '"';
        len = strlen(p);
        for(jx=0; jx < len; jx++)   {
            if((*p == '"') || (*p == '\\')) 
                *cP++ = '\\';   /* escape the delimiters and escape char */
            *cP++ = *p++;
        }
        *cP++ = '"';
        *cP = '\0';
        newP = (char *)TaskRealloc(pid, newP, strlen(newP)+1);
    }

    if(newP)    {
        /* len of existing str, new string, delimiter and null char */
        len = strlen(str) + strlen(newP) + 1 + 1;
        str = (char *)TaskRealloc(pid, str, len);
        strcat(str, newP);
        strcat(str, TOKEN_DELIM_STR);   /* delimiter */
    }
    free(newP);
    return(str);
}
char *
AxTaskElfDataToInsetStr(pid, aP)
int pid;
data_ptr aP;
{
    char *str;
    str = (char *)TaskAlloc(pid, 1);
    *str = '\0';    
    str = AxDoElfDataToInsetStr(pid, aP, str);
    return(str);
}
char *
AxElfDataToInsetStr(aP)
data_ptr aP;
{
    return(AxTaskElfDataToInsetStr(THIMpid, aP));
}


/******************************************************************************
* function: AxInsetStrToElfData - given a inset source string, convert it
*   to elf data - Caller is responsible for free'ing the elfdata
*/

data_ptr
AxDoInsetStrToElfData(pid, str, arrayP)
int pid;
char **str;
data_ptr arrayP;
{
    char *p, *cP, *newP;
    char buf[STRPATH+1];
    data_ptr aP, dataP;

    p = *str;
    while (*p)  {
        if(*p == ARRAY_START_CHAR)  {
            p++;    /* skip delimiter */
            aP = AxTaskMakeArray(pid, 0);
            aP = AxDoInsetStrToElfData(pid, &p, aP);    
            if(NULL == arrayP)
                return(aP);
            arrayP = AxTaskAddArrayToArray(pid,arrayP,AxArraySize(arrayP),aP);
            continue;
        }else if(*p == ARRAY_END_CHAR)  {
            p++;    /* skip the delimiter */
            *str = p;
            return(arrayP);
        } else if(*p == ' ')    {
            p++;    /* skip starting string delimiter */
            continue;
        }else if(*p == '"') {
            p++;    /* skip starting delimiter */
            cP = p;
            newP = buf;
            while (*p && (*p != '"'))   {
                if(*p == '\\')  
                    p++;    /* skip escape character */
                *newP++ = *p++;
            }
            *newP = '\0';
            if(*p == '"')   
                p++;    /* skip the ending string delimiter */
            dataP = AxTaskMakeStrData(pid, -1, buf);
        }else   {
            /* have a number or a null item */
            cP = p;
            while(*cP)  {
                if((*cP == ' ') || (*cP == ARRAY_END_CHAR))
                    break;
                cP++;
            }
            if(*cP == 0) {
                /* abort processing and return */
                p += strlen(p); 
                *str = p;
                return(arrayP);
            }
            strncpy(buf, p, cP - p);
            buf[cP-p] = '\0';
            if(buf[0] == NULL_ITEM_CHAR)    {
                /* null item */
                dataP = NULL;
            }else if(0 == strchr(buf, '.')) {
                /* no period, assume integer */
                dataP = AxTaskMakeIntData(pid, atoi(buf));
            }else   {
                dataP = AxTaskMakeFloatData(pid, atof(buf));
            }
            p = cP;
        }

        if(NULL == arrayP)  {
            /* simple item, no arrays in the string  */
            p += strlen(p); /* point p to end of the string */
            *str = p;
            return(dataP);
        }
        arrayP = AxTaskWriteArray(pid, arrayP, AxArraySize(arrayP), dataP);
    }
    *str = p;
    return(arrayP); 
}

data_ptr 
AxTaskInsetStrToElfData(pid, str)
int pid;
char *str;
{
    char *p;
    data_ptr dp;

    if ( str == (char *)0 )
        dp = (data_ptr)0;
    else
    {
        p = str;
        dp = AxDoInsetStrToElfData(pid, &p, NULL);
    }
    return(dp);
}

data_ptr 
AxInsetStrToElfData(str)
char *str;
{
    return(AxTaskInsetStrToElfData(THIMpid, str));
}
#endif

AxDocTypeIsNative(docType)
	int docType;
{
	if ( docType == AX_DOC_TYPE_ASTERX_WORDS )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_GRAPHICS )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_SS )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_MACROS )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_BINARY_SS )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_BITMAP )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_EQUATIONS )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_QUERYDATA )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_MACRO_OBJ )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_AUDIO )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASCII )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_HTML )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_SLIDESHOW )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_BUILDER )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_BUILDER_ABD )
		return(TRUE);
	if ( docType == AX_DOC_TYPE_ASTERX_BUILDER_TURBO )
		return(TRUE);
	return(FALSE);
}

#if FOR_FILTER
char *
TaskAlloc(pid, size)
int pid;
unsigned int size;
{
    return( malloc(size) );
}
#endif
