/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axgc.h
   Applix ASC version: 2
   Date: Mon Nov 16 13:53:46 1998
*/
#ifndef AXGC_H
#define AXGC_H

#if AX_X11
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif
#include <basicwin.h>

/* Types AxGCVinternal, AxGCVexternal are derived from XGCValues.
 * The internal struct holds XGCValues vals that will be used to
 * form a GC's lookup name.  AxGCVinternal also holds screen# and
 * drawable-depth info for partitioning the set of GCs according to
 * drawable type.
 * The AxGCVexternal type holds XGCValues vals that are considered
 * too 'variant' to be used as naming data.  AxGCVexternal also holds
 * info like clip-rects and clip-region that are logically part of the
 * GC but don't actually live on the real XGCValues type.
 */
#define DASH_MAX 7    /* we store this many dashlist elements tops */

typedef struct {
	int screen_num;           /* used for gc-creation */
	int depth;                /* used for gc-creation */
	int function;		/* logical operation */
	unsigned long plane_mask;/* plane mask */
	int line_width;		/* line width */
	int line_style;	 	/* LineSolid, LineOnOffDash, LineDoubleDash */
	int cap_style;	  	/* CapNotLast, CapButt, 
						   CapRound, CapProjecting */
	int join_style;	 	/* JoinMiter, JoinRound, JoinBevel */
	int fill_style;	 	/* FillSolid, FillTiled, 
						   FillStippled, FillOpaeueStippled */
	int fill_rule;	  	/* EvenOddRule, WindingRule */
	int arc_mode;		/* ArcChord, ArcPieSlice */
	Font font;	        /* default text font for text operations */
	int subwindow_mode;     /* ClipByChildren, IncludeInferiors */
	int graphics_exposures;/* boolean, should exposures be generated */
	int dash_offset;	/* patterned/dashed line information */
	char dash_list[DASH_MAX];
	int n_dashes;
} AxGCVinternal;

typedef struct {
	unsigned long foreground; /* foreground pixel */
	unsigned long background; /* background pixel */
	Pixmap tile;		/* tile pixmap for tiling operations */
	Pixmap stipple;		/* stipple 1 plane pixmap for stipping */
	int ts_x_origin;	/* offset for tile or stipple operations */
	int ts_y_origin;
	int clip_x_origin;	/* origin for clipping */
	int clip_y_origin;
	Pixmap clip_mask;	/* bitmap clipping; other calls for rects */
	XRectangle *clip_rects;
	Region region;
	int n_rects;
} AxGCVexternal;


/* type AxGC is used to remember GC information without
 * gc-writes actually hitting the X connection
 */
typedef struct axgc_t {
	AxGCVinternal loc;
	AxGCVexternal ext;
} AxGC;

/* type AxGCVals is used to pass vals & mask, et. al. to AxGC constructor
 */
typedef struct {
	int screen_num;
	int depth;
	XGCValues *vals;
	unsigned long mask;
} AxGCVals;

/* type AxRealGC is the item type in the AxRealGCtab array
 */
typedef struct ax_realgc_t {
	GC gc;                    /* real X gc handle */
	AxGCVinternal loc;
	AxGCVexternal ext;
} AxRealGC;

#endif /* AXGC_H */
