/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axgfs_com.c
   Applix ASC version: 5
   Date: Thu May  6 21:12:07 1999
*/
#include <stdlib.h>
#include <aos.h>
#include <axtext.h>
#include <axgfs.h>

#if AX_X11
AXFSServer *AxGFSServer;
#endif

static axhtab AxEncodingMaps;

extern int ScreenRes, TargetRes;
extern Display *Dpy;

int axgfsFreeExtension();
char *axgfsIntGetFid();



int axgfsFreeExtension(ours)
XExtData *ours;
{
    if (ours && ours->private_data) {
        axfs_ext *xf = (axfs_ext *) ours->private_data;
        if (xf->com.fids.width) {
            axhDeleteTable(&xf->com.fids);
        }
        if (xf->com.letters.width) {
            axhDeleteTable(&xf->com.letters);
        }
        if (xf->qletters) {
            int ix;
            for (ix = 0; ix < xf->numChars; ix++) {
                if (xf->qletters[ix]) {
                    if ((xf->qletters[ix])->bbox) {
                        free((xf->qletters[ix])->bbox);
                    }
                    if ((xf->qletters[ix])->bits) {
                        XFreePixmap(Dpy, (Pixmap) ((xf->qletters[ix])->bits));
                    }
                    free(xf->qletters[ix]);
                }
            }
            free(xf->qletters);
        }
        free(xf);
    }
}





/* : axgfsNewFid - constructs items in axfs_ext.com.fids
 */
static char *
axgfsNewFid(name, notUsed, extp)
void *name;
int notUsed;
char *extp;
{
    axfs_ext *ext;
    axctm *ctm;

    ctm = (axctm *) name;
    ext = (axfs_ext *) extp;

    return (axgfsIntGetFid(ext->com.xlfd, ext->pointsize, ScreenRes, ctm));
}




/* : axgfsNewPFid - constructs items in axpfs_ext.com.fids
 */
static char *
axgfsNewPFid(name, notUsed, extp)
void *name;
int notUsed;
char *extp;
{
    axfs_com *ext;
    axpfid *pfid;

    pfid = (axpfid *) name;
    ext = (axfs_com *) extp;

    return (axgfsIntGetFid(ext->xlfd, pfid->pointsize, TargetRes, &pfid->ctm));
}




static void
axgfsFreeFid(fidp, notUsed)
char *fidp;
char *notUsed;
{
    Font fid;

    if (fidp) {
        fid = (Font) fidp;
#if AX_X11
        FSCloseFont(AxGFSServer, fid);
#else
        MS_DeleteObject( fid );
#endif
    }
    return;
}



/* : axgfsNewLetter - constructs items in axfs_ext.com.letters
 */
static char *
axgfsNewLetter(name, notUsed, fsp)
void *name;
int notUsed;
char *fsp;
{
    Font fid;
    axformed_lettername *lname;
    XFontStruct *fs;
    axfs_ext *ext;


    lname = (axformed_lettername *) name;
    fs = (XFontStruct *) fsp;
    ext = (axfs_ext *) fs->ext_data->private_data;

    /* lookup fid associated with lname->ctm
     */
    fid = (Font) 0;
    axhGetItemI(&lname->ctm, sizeof(axctm),
                &ext->com.fids, (char *) ext, (char **) &fid);
    if (fid) {
        return ((char *) axgfsIntGetLetter((int) lname->charcode, fs, NULL,
                                           fid, TRUE, TRUE));
    }
    else {
        return ((char *) 0);
    }
}

char *
axgfsNewPLetter(name, notUsed, fsp)
void *name;
int notUsed;
char *fsp;
{
    Font fid;
    XFontStruct *fs;
    plettername *lname;
    axfs_com *ext;
    axpfid pfid;

    lname = (plettername *) name;
    fs = (XFontStruct *) fsp;
    ext = (axfs_com *) fs->ext_data->private_data;

    /* lookup fid associated with lname
     */
    fid = (Font) 0;
    axhGetItemI(&lname->pfid, sizeof(axpfid),
                &ext->fids, (char *) ext, (char **) &fid);

    if (fid) {
        return ((char *) axgfsIntGetLetter((int) lname->charcode, fs,
										   &(lname->pfid.ctm),
                                           fid, FALSE, TRUE));
    }
    else {
        return ((char *) 0);
    }
}






static void
axgfsFreeLetter(letter, exdata)
char *letter;
char *exdata;
{
    if (letter) {
        axletter *al = (axletter *) letter;

        if (al->bits) {
            XFreePixmap(Dpy, (Pixmap) al->bits);
        }
        if (al->bbox) {
            free(al->bbox);
        }
        free(letter);
    }
    return;
}





static int
axgfsFreePExtension(ours)
XExtData *ours;
{
    if (ours && ours->private_data) {
        axpfs_ext *xf = (axpfs_ext *) ours->private_data;
        if (xf->com.fids.width) {
            axhDeleteTable(&xf->com.fids, (char *) 0);
        }
        if (xf->com.letters.width) {
            axhDeleteTable(&xf->com.letters, (char *) 0);
        }
        free(xf);
    }
}




void
axgfsFreePLetter(letter, exdata)
char *letter;
char *exdata;
{
    if (letter) {
        free(letter);
    }
    return;
}





void
axgfsAddFontExtension(fs, resolution, pointsize, xlfd, ftab)
XFontStruct *fs;       /* in/out: add extension to this fontstruct */
int resolution;        /* in, ==ScreenRes or ==TargetRes */
int pointsize;         /* in, meaningful only when resolution==ScreenRes */
char *xlfd;            /* in: ref to name in AxGFSxlfdNames or fontcache */
fontTabEnt *ftab;      /* in: ref to associated fontcache record */
{
	int itmp, numChars;
	XExtData *ourExt;
	axfs_com *extcom;
	
	if (!fs) {
		return;
	}
	
	numChars = fs->max_char_or_byte2 - fs->min_char_or_byte2 + 1;
	numChars *= fs->max_byte1 - fs->min_byte1 + 1;
	ourExt = (XExtData *) malloc(sizeof(XExtData));
	memset((void *) ourExt, 0, sizeof(XExtData));
	ourExt->number = AXGFS_FONT_EXTENSION;
	
	if (resolution == ScreenRes) {
		/* adding to a screenfont struct
		 */
		axfs_ext *fx;
		ourExt->free_private = axgfsFreeExtension;
		ourExt->private_data = (char *) malloc (sizeof(axfs_ext));
		fx = (axfs_ext *) ourExt->private_data;
		memset((void *) fx, 0, sizeof(axfs_ext));
		fx->pointsize = pointsize;
		fx->numChars = numChars;
		fx->com.xlfd = xlfd;          /* copy reference only -- do not free */
 		fx->com.resolution = resolution;
/* %%%%%%% these table allocs should be done on demand!! */
		axhInitTable(&fx->com.fids, 11,
					 axgfsNewFid, axgfsFreeFid);
		axhInitTable(&fx->com.letters, 31,
					 axgfsNewLetter, axgfsFreeLetter);
	
		itmp = numChars * sizeof(axletter *);
		fx->qletters = (axletter **) malloc (itmp);
		memset((void *) fx->qletters, 0, itmp);

		extcom = &fx->com;
	}
	else {
		/* adding to a printerfont struct
		 */
		axpfs_ext *pfx;
		ourExt->free_private = axgfsFreePExtension;
		ourExt->private_data = (char *) malloc (sizeof(axpfs_ext));
		pfx = (axpfs_ext *) ourExt->private_data;
		memset((void *) pfx, 0, sizeof(axpfs_ext));
		pfx->com.xlfd = xlfd;     /* copy reference only -- do not free */
		pfx->com.resolution = resolution;
/* %%%%%%% these table allocs should be done on demand!! */
		axhInitTable(&pfx->com.fids, 11,
					 axgfsNewPFid, axgfsFreeFid);
		axhInitTable(&pfx->com.letters, 31,
					 axgfsNewPLetter, axgfsFreePLetter);
		
		extcom = &pfx->com;
	}

	if (ftab && ftab->internalEncodingName &&
		strcmp(ftab->internalEncodingName, NXLT("unspecified"))) {
		
		char *mapname, *externalEnc;
		int namelen;

		/* init the nativeCodes member of the extension, if external
		 * and internal encodings differ.  Note that if either encoding is
		 * "unspecified", we mustn't init nativeCodes -- no defined mapping.
		 */
		if (ftab->encodingFileName) {
			externalEnc = (char *) malloc(strlen(ftab->encodingFileName) + 1);
			strcpy(externalEnc, ftab->encodingFileName);
		}
		else {
			externalEnc = (char *) malloc(strlen(NXLT("latin1.enc")) + 1);
			strcpy(externalEnc, NXLT("latin1.enc"));
		}
		if (strcmp(ftab->internalEncodingName, externalEnc) &&
			strcmp(externalEnc, NXLT("unspecified")) ) {
			
			namelen = strlen(ftab->internalEncodingName) +
				strlen(externalEnc) + 5;
			mapname = (char *) malloc(namelen);
			strcpy(mapname, externalEnc);
			strcat(mapname, ",");
			strcat(mapname, ftab->internalEncodingName);
			axhGetItem(mapname, &AxEncodingMaps, (char *) NULL,
					   (char **) &extcom->nativeCodes);
			free(mapname);
		}
		free(externalEnc);
	}
	
#if AX_X11
	XAddToExtensionList(&fs->ext_data, ourExt);
#else
    fs->ext_data = ourExt;
#endif
}








int axgfsGetQLetterIndex(charcode, xfont)
int charcode;
XFontStruct *xfont;
{
	int row, col, rowWid, idx;
	
    if (xfont->max_byte1 == 0) {
        col = charcode & 0xff;
        if ((unsigned int) col >= xfont->min_char_or_byte2 &&
            (unsigned int) col <= xfont->max_char_or_byte2) {
			
			idx = col - xfont->min_char_or_byte2;
        }
        else {
			idx = 0;  /* indexes ascii space char's slot on qletters */
        }
    }
    else {
        row = (charcode >> 8) & 0xff;
        col = charcode & 0xff;
        if ((unsigned int) row >= xfont->min_byte1 &&
            (unsigned int) row <= xfont->max_byte1 &&
            (unsigned int) col >= xfont->min_char_or_byte2 &&
            (unsigned int) col <= xfont->max_char_or_byte2) {
            
            rowWid = xfont->max_char_or_byte2 -
                xfont->min_char_or_byte2 + 1;
            idx = ((row - xfont->min_byte1) * rowWid) +
                col - xfont->min_char_or_byte2;
            if (xfont->per_char == NULL) {
				idx = 0;
            }
        }
        else {
			idx = 0;
        }
    }
    return (idx);
}





void
axgfsDrawString(display, drawable, gc, font, x, y, in_str, length)
Display *display;         /* in */
Drawable *drawable;       /* in */
GC gc;                    /* in */
XFontStruct *font;        /* in */
int x, y;                 /* in */
mbchar *in_str;           /* in */
int length;               /* in */
{
	int ix, idx, txpos, tx, ty, charcode;
	mbchar *mbstr;
	XCharStruct *xc;
	Pixmap letterPix;
	axfs_ext *ext;

#if 1
	XSetFillStyle(display, gc, FillStippled);
#endif
	ext = (axfs_ext *) font->ext_data->private_data;
	
	for (ix = 0, txpos = x; ix < length; ix++) {
#if BUILD_WIDE		
		charcode = (int) in_str[ix];
#else
		charcode = (int) ((unsigned char) in_str[ix]);
#endif
		idx = axgfsGetQLetterIndex(charcode, font);
		if (ext->qletters[idx]) {
			letterPix = (Pixmap) (ext->qletters[idx])->bits;
		}
		else {
 			ext->qletters[idx] = axgfsIntGetLetter(charcode, font, NULL,
												   font->fid, TRUE, FALSE);
			if (!ext->qletters[idx]) {
				continue;
			}
			letterPix = (Pixmap) (ext->qletters[idx])->bits;
		}
		xc = getCharStruct(charcode, font);
		tx = txpos + xc->lbearing;
		ty = y - xc->ascent;

		XSetStipple(display, gc, letterPix);
		XSetTSOrigin(display, gc, tx, ty);
		XFillRectangle(display, drawable, gc, tx, ty,
					   xc->rbearing - xc->lbearing,
					   xc->ascent + xc->descent);
		
		txpos += xc->width;
	}
	
	XSetFillStyle(display, gc, FillSolid);
}

void
axgfsWDrawString(display, drawable, gc, font, x, y, in_str, length)
Display *display;         /* in */
Drawable *drawable;       /* in */
GC gc;                    /* in */
XFontStruct *font;        /* in */
int x, y;                 /* in */
wchar *in_str;            /* in */
int length;               /* in */
{
	int ix, idx, txpos, tx, ty, charcode;
	mbchar *mbstr;
	XCharStruct *xc;
	Pixmap letterPix;
	axfs_ext *ext;
	
#if 1
	XSetFillStyle(display, gc, FillStippled);
#endif
	ext = (axfs_ext *) font->ext_data->private_data;
	
	for (ix = 0, txpos = x; ix < length; ix++) {
#if BUILD_WIDE		
		charcode = (int) in_str[ix];
#else
		charcode = (int) ((unsigned char) in_str[ix]);
#endif
		idx = axgfsGetQLetterIndex(charcode, font);
		if (ext->qletters[idx]) {
			letterPix = (Pixmap) (ext->qletters[idx])->bits;
		}
		else {
			ext->qletters[idx] = axgfsIntGetLetter(charcode, font, NULL,
												   font->fid, TRUE, FALSE);
			if (!ext->qletters[idx]) {
				continue;
			}
			letterPix = (Pixmap) (ext->qletters[idx])->bits;
		}
		xc = getCharStruct(charcode, font);
		tx = txpos + xc->lbearing;
		ty = y - xc->ascent;

		XSetStipple(display, gc, letterPix);
		XSetTSOrigin(display, gc, tx, ty);
		XFillRectangle(display, drawable, gc, tx, ty,
					   xc->rbearing - xc->lbearing,
					   xc->ascent + xc->descent);
		
		txpos += xc->width;
	}
	
	XSetFillStyle(display, gc, FillSolid);
}

