/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axhtab.c
   Applix ASC version: 14
   Date: Sun Jan 31 21:12:25 1999
*/
#include <aos.h>
#include <axhtab.h>
#include <axhtab.p>

void
axhInitTable(tab, width, construct, destruct)
axhtab *tab;            /* in: points to caller's axhtab instance to init */
int width;              /* in */
char * (*construct)();  /* in: makes instances of types stored in table */
void (*destruct)();     /* in: tosses instances */
{
	tab->table = (hitem **) malloc (width * sizeof(hitem *));
	memset((char *) tab->table, 0, width * sizeof(hitem *));
	tab->fetch = construct;
	tab->toss = destruct;
	tab->width = width;
}

/* : axhDeleteTable - free a table's contents.  Note that
 * the axhtab instance itself is not freed; only its contents
 */
void
axhDeleteTable(tabl, idata)
axhtab *tabl;        /* in: free contents of this table */
void *idata;         /* in: extra data to pass to item destructor */
{
	int ix;
	hitem *list, *tmp;

	for (ix = 0; ix < tabl->width; ix++) {
		list = tmp = tabl->table[ix];
		while(list) {
			if (list->value && tabl->toss) {
				tabl->toss(list->value, (char *) idata);
			}
			free((char *) list->name);
			tmp = list->next;
			free((char *) list);
			list = tmp;
		}
	}
	free((char *) tabl->table);
}

/* : axhGetItem - look up named instance of some type.
 * type's constructor called if needed to complete the lookup
 */
void
axhGetItem(name, tabl, inp, result)
char *name;           /* in: name of item in passed table */
axhtab *tabl;         /* in: table */
char *inp;            /* in: optional data to pass to constructor */
char **result;        /* out: address of item */
{
	register int hash;
	register int wid = tabl->width;
	hitem *list, **adlist, *tail, *newItem;
	
	if (!tabl || !tabl->width) {
		*result = (char *) NULL;
		return;
	}

	hash = hashIndex(tabl, name);

	tail = list = tabl->table[hash];
	adlist = &tabl->table[hash];

	while(list) {
		if (!strcmp(name, list->name)) {
			*result = list->value;
			return;
		}
		tail = list;
		list = list->next;
		adlist = &list;
	}
	    /* not in there; fetch and add */
	newItem = (hitem *) malloc (sizeof(hitem));
	memset((char *) newItem, 0, sizeof(hitem));
	newItem->name = (char *) malloc(strlen(name) + 1);
	strcpy(newItem->name, name);
	if (tabl->fetch) {
		newItem->value = (char *) tabl->fetch(name, inp);
	}

	if (adlist == &tabl->table[hash]) {
		tabl->table[hash] = newItem;
	}
	else {
		tail->next = newItem;
	}
	*result = newItem->value;
}

void
axhTossItem(name, tabl, inp)
char *name;        /* in: name of item to destroy */
axhtab *tabl;      /* in: table believed to contain item */
char *inp;         /* in: optional data to pass to destructor */
{
	register int hash;
	register int wid = tabl->width;
	hitem *list, **adlist, *last, *newItem, *tmp;

	hash = hashIndex(tabl, name);

	last = list = tabl->table[hash];
	adlist = &tabl->table[hash];

	while(list) {
		if (!strcmp(name, list->name)) {
			tmp = list->next;
			if (tabl->toss) {
				tabl->toss(list->value, inp);
			}
			free((char *) list->name);
			free((char *) list);
			if (adlist == &tabl->table[hash]) {
				tabl->table[hash] = tmp;
			}
			else {
				last->next = tmp;
			}
			return;
		}
		last = list;
		list = list->next;
		adlist = &list;
	}
}

int    /* return true if found, false otherwise */
axhTestItem(name, tabl)
char *name;             /* in: name of item in table */
axhtab *tabl;           /* in: table believed to contain item */
{
	register int hash;
	register int wid = tabl->width;
	register int found = FALSE;
	hitem *list;

	hash = hashIndex(tabl, name);

	list = tabl->table[hash];

	while(list) {
		if (!strcmp(name, list->name)) {
			return TRUE;
		}
		list = list->next;
	}
	return FALSE;
}

int   /* return true if found, false (no-op) otherwise */
axhWriteItem(name, tabl, value)
char *name;             /* in: name of item in table */
axhtab *tabl;           /* in: table believed to contain item */
char *value;            /* in: addr of type instance to replace current with */
{
	register int hash;
	register int wid = tabl->width;
	register int found = FALSE;
	hitem *list;

	hash = hashIndex(tabl, name);

	list = tabl->table[hash];

	while(list) {
		if (!strcmp(name, list->name)) {
			if (tabl->toss) {
				tabl->toss(list->value, (char *) NULL);
			}
			list->value = value;
			return TRUE;
		}
		list = list->next;
	}
	return FALSE;
}

/* : axhListNames - for string-name axhtab only, produce
 * straight array of all name refs in the table.  Caller
 * must free returned array using axhFreeListNames.
 */
void
axhListNames(tabl, outlist, outlistsize)
axhtab *tabl;                           /* in */
char ***outlist;                        /* out */
int *outlistsize;                       /* out: array-size of outlist */
{
	int ix, jx;
	hitem *list;

	for (ix = 0, *outlistsize = 0; ix < tabl->width; ix++) {
		list = tabl->table[ix];
		while(list) {
			if (list->name) {
				(*outlistsize)++;
			}
			list = list->next;
		}
	}

	*outlist = (char **) malloc (*outlistsize * sizeof(char *));
	
	for (ix = 0, jx = 0; ix < tabl->width; ix++) {
		list = tabl->table[ix];
		while(list) {
			if (list->name) {
				(*outlist)[jx] = list->name;
				jx++;
			}
			list = list->next;
		}
	}
}

void
axhFreeListNames(names)
char **names;
{
	free(names);
}

static int
hashIndex(tabl, name)
register axhtab *tabl;
register char *name;
{
	register int sum = 0;
	register int chr = 0;
	register char *np = name;

	while (*np) {
		chr = *np++;
		sum = sum + sum + (chr & 0xFF);
		while (sum >= tabl->width)
			sum -= tabl->width;
	}
	return(sum);
}

/* : axhGetItemI - look up named instance of some type.
 * type's constructor called if needed to complete the lookup
 */
void
axhGetItemI(name, namelen, tabl, inp, result)
void *name;           /* in: name of item in passed table */
int namelen;          /* in: length of name in bytes */
axhtab *tabl;         /* in: table */
char *inp;            /* in: optional data to pass to constructor */
char **result;        /* out: address of item */
{
	register int hash;
	register int wid = tabl->width;
	hitem *list, **adlist, *tail, *newItem;
	
	if (!tabl || !tabl->width) {
		*result = (char *) NULL;
		return;
	}

	hash = hashIndexI(tabl, name, namelen);

	tail = list = tabl->table[hash];
	adlist = &tabl->table[hash];

	while(list) {
		if (!memcmp((void *) name, (void *) list->name, namelen)) {
			*result = list->value;
			return;
		}
		tail = list;
		list = list->next;
		adlist = &list;
	}
	    /* not in there; fetch and add */
	newItem = (hitem *) malloc (sizeof(hitem));
	memset((char *) newItem, 0, sizeof(hitem));
	newItem->name = (char *) malloc(namelen);
	memcpy((void *) newItem->name, (void *) name, namelen);
	if (tabl->fetch) {
		newItem->value = (char *) tabl->fetch(name, namelen, inp);
	}

	if (adlist == &tabl->table[hash]) {
		tabl->table[hash] = newItem;
	}
	else {
		tail->next = newItem;
	}
	*result = newItem->value;
}

void
axhTossItemI(name, namelen, tabl, inp)
void *name;        /* in: name of item to destroy */
int  namelen;      /* in: length of name in bytes */
axhtab *tabl;      /* in: table believed to contain item */
char *inp;         /* in: optional data to pass to destructor */
{
	register int hash;
	register int wid = tabl->width;
	hitem *list, **adlist, *last, *newItem, *tmp;

	hash = hashIndexI(tabl, name, namelen);

	last = list = tabl->table[hash];
	adlist = &tabl->table[hash];

	while(list) {
		if (!memcmp((void *) name, (void *) list->name, namelen)) {
			tmp = list->next;
			if (tabl->toss) {
				tabl->toss(list->value, inp);
			}
			free((char *) list->name);
			free((char *) list);
			if (adlist == &tabl->table[hash]) {
				tabl->table[hash] = tmp;
			}
			else {
				last->next = tmp;
			}
			return;
		}
		last = list;
		list = list->next;
		adlist = &list;
	}
}

int    /* return true if found, false otherwise */
axhTestItemI(name, namelen, tabl)
void *name;             /* in: name of item in table */
int namelen;            /* in: length of name in bytes */
axhtab *tabl;           /* in: table believed to contain item */
{
	register int hash;
	register int wid = tabl->width;
	register int found = FALSE;
	hitem *list;

	hash = hashIndexI(tabl, name, namelen);

	list = tabl->table[hash];

	while(list) {
		if (!memcmp((void *) name, (void *) list->name, namelen)) {
			return TRUE;
		}
		list = list->next;
	}
	return FALSE;
}

int   /* return true if found, false (no-op) otherwise */
axhWriteItemI(name, namelen, tabl, value)
void *name;             /* in: name of item in table */
int namelen;            /* in: length of name in bytes */
axhtab *tabl;           /* in: table believed to contain item */
char *value;            /* in: addr of type instance to replace current with */
{
	register int hash;
	register int wid = tabl->width;
	register int found = FALSE;
	hitem *list;

	hash = hashIndexI(tabl, name, namelen);

	list = tabl->table[hash];

	while(list) {
		if (!memcmp((void *) name, (void *) list->name, namelen)) {
			if (tabl->toss) {
				tabl->toss(list->value, (char *) NULL);
			}
			list->value = value;
			return TRUE;
		}
		list = list->next;
	}
	return FALSE;
}

/* : hashIndexI - adapted from Babu's elintr.c:hash(),
 * for use with binary-data names.  The axhtab param
 * *tabl should be initialized with a prime-numbered
 * width for reasonable results.
 */
static int 
hashIndexI (tabl, name, namelen)
register axhtab *tabl;
register void *name;
register int namelen;
{
	register int sum = 0;
	register int chr = 0;
	register char *np = name;

	while (namelen--) {
		chr = *np++;
		sum = sum + sum + (chr & 0xFF);
		while (sum >= tabl->width)
			sum -= tabl->width;
	}
	return(sum);
}
