/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axtext.h
   Applix ASC version: 26
   Date: Mon Nov 16 13:53:32 1998
*/
#ifndef _AXTEXT_H
#define _AXTEXT_H

#include <atm.h>				/* for attr_type */

typedef unsigned short tx_char;
typedef unsigned char attr_t;

typedef struct p_32 {
	int x;
	int y;
} p32;

typedef struct r_32 {
	p32 origin;
	unsigned int width;
	unsigned int height;
} r32;

/*
 * b32 can be used to describe the vertices of a rectangular
 * shape after some transform.  When used this way, the points
 * p0-p3 represent the transforms of the upper left, upper right,
 * lower right, lower left vertices of the untransformed rect.
 */
typedef struct b_32 {
	p32 p0; /* xform of upper left vertex */
	p32 p1; /* rest clockwise.. */
	p32 p2;
	p32 p3;
} b32;

typedef struct axt_bullet_t {
	unsigned int changed: 1;
	p32 bulletOffset;         /* vector from bullet origin to 1st text */
	r32 bounder;
	int type;
	int color;
	int scale;
	int font;
	int glyph;
	int page_ix;
	void *gContext;           /* handle of ax graphics context */
	void *gAttr;              /* handle of attribute to draw bullet */
	void *cvt;                /* handle of buffer to draw into */
	int (*bulletSize)();      /* func returning bounder of bullet */
	int (*bulletDraw)();      /* func that displays bullet */
} axtBullet;

typedef struct axt_indent_t {
	int firstLeft;            /* indent applied to 1st line in para */
	int left;                 /* indent applied to subsequent lines */
	int right;                /* applied to right edges of all */
	int topLineHeight;        /* line height for 1st line in para */
	int midLineHeight;        /* line height for other lines in para */
	int bottomLineOffset;     /* offset applied to next para's top lineht */
	int alignment;            /* CENTER_ALIGN, e.g. */
	axtBullet *bullet;
} axtIndent;

typedef struct axt_drop_shadow_t {
	p32 offset;               /* displacement from text */
	attr_type *attr;          /* color, pattern, etc.. */
} axtDropShadow;

/*
 * Text attribute description 
 */
#define TX_NARROW_STRING 1 /* all text chars 8 bits */
#define TX_WC_STRING 2		/* all text chars 16 bits */

	/* interest bits */
#define TXtextString (1L<<0) /* someday */
#define TXfontFamily (1L<<1)
#define TXweight (1L<<2)
#define TXslant (1L<<3)
#define TXpointsize (1L<<4)
#define TXruling (1L<<5)
#define TXhilite (1L<<6)
#define TXnoRuleChar (1L<<7)
#define TXangle (1L<<8)
#define TXxscale (1L<<9)
#define TXxshear (1L<<10)
#define TXfillcolor (1L<<11)
#define TXfillbcolor (1L<<12)
#define TXstrokecolor (1L<<13)
#define TXstrokebcolor (1L<<14)
#define TXfillpattern (1L<<15)
#define TXstrokepattern (1L<<16)
#define TXsubsup (1L<<17)
#define TXyscale (1L<<18)
#define TXsupx (1L<<19)
#define TXsupy (1L<<20)
#define TXkernx (1L<<21)
#define TXkerny (1L<<22)
#define TXfgPixel (1L<<23)
#define TXdropShadow (1L<<24)
#define TXbullet (1L<<25)
#define TXindent (1L<<26)
#define TXstrike (1L<<27)

#define TXAllAtts (0xffffffe) /* exclude textString attr for now */

	/* structure */
typedef struct text_attrs {
	unsigned int hilite     : 1,  /* screen-only; draw hilit */
	             useFgPixel : 1,  /* screen-only; use fgPixel for fg color */
	             strike_thru: 1;  /* strike thru info */
	int numChars;
#if 0 /* someday */
	int stringType;         /* narrow, mb, wc */
	unsigned char *string8; /* if narrow string */
	tx_char *string16;       /* if mb or wc string */
#endif
	char *fontFamily;       /* like "Times", or "Palatino" */
	int weight;             /* typ. 0 or 1 for nonbold, bold */
	int slant;              /* typ. 0 or 1 for nonital, ital */
	int pointsize;          /* in 10000ths of a point */
	int ruling;
	int supx;               /* superscript, x direction, mils */
	int supy;
	int kernx;              /* 'kern', x direction, mils */
	int kerny;
	tx_char noRuleChars[2];  /* e.g., space, tab */
	int angle;              /* in 1000ths of a radian */ 
	int xscale;             /* reference value 10000, so 10000==1.0 */
	int yscale;             /* reference value 10000, so 10000==1.0 */
	int xshear;             /* reference value 10000, so 10000==1.0 */
	int fillcolor;
	int fillbcolor;
	int strokecolor;
	int strokebcolor;
	unsigned long fgPixel;  /* screen-only alternate form for fillcolor */
	attr_type *fillpattern;
	attr_type *strokepattern;
	axtIndent *indent;
	axtBullet *bullet;
	axtDropShadow *shadow;
} textAttrs;

	/* text cursor attrs */
#define CURSOR_TRIANGLE 1
#define CURSOR_BAR 2
#define INACTIVE 0
#define ACTIVE 1

	/* text ruling attrs (bitwise) */
#define AXRULING_SINGLE 1
#define AXRULING_DOUBLE 2
#define AXRULING_STRIKETHRU 4

#ifndef _AXTXINT_H
#include <axtxint.h>
#endif

#endif /* _AXTEXT_H */

