/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: colorcom.c
   Applix ASC version: 79
   Date: Wed May 12 21:10:35 1999
*/
#include <stdio.h>
#include <math.h>
#include <aos.h>
#include <atm.h>
#include <basicwin.h>
#include <xmlib.h>
#include <tmi.h>
#include <getstr5.h>
#include <colors.h>
#include <axwinclr.h>

#define packedRGB16(r,g,b)  ((((b)&0xFE)<<16) | (((g)&0xFE)<<8) | ((r)&0xFE))
#define packedRGB(r,g,b)    ((((b)&0xFF)<<16) | (((g)&0xFF)<<8) | ((r)&0xFF))
#define packedBlue(rgb)     (((rgb)>>16)&0xFF)
#define packedGreen(rgb)    (((rgb)>>8)&0xFF)
#define packedRed(rgb)      ((rgb)&0xFF)

#define allocStr(s) strcpy(malloc(strlen(s)+1),s)

#define AX_XA_RGB_APPLIX_MAP_STR    NXLT("XA_RGB_APPLIX_MAP")
#define AX_XA_NUM_ASTERX_RUN_STR    NXLT("XA_NUM_ASTERX_RUN")
#define AX_WIDGET_PIXEL_BASE_POS    246
#define AX_APP_PIXEL_BASE_POS       30

char    *AxStdCmapName;
int     axPreDefinedStdmap;
int     AxNStdCmaps;
Atom    axXA_RGB_APPLIX_MAP;
Atom    axXA_NUM_ASTERX_RUN;
/**********************************************************************/
/* Here are some macros so one version of the code can be shared between */
/* X and MS-Windows which has similar color data item concepts, but  */
/* different structure names and layouts                             */
/**********************************************************************/
#if AX_X11    /*  {  */

#define MAX_GUN 0xFFFF
#define COLORVAL_TYPE unsigned short
#define COLOR_DESCRIPTOR XColor
#define GET_DEFAULT_COLOR_LIST DefaultColormap(Dpy, DefaultScreen(Dpy))
#define SET_RGB_VALS( base, r, g, b, f, p ) \
       { base red = (r << 8); \
         base green = (g << 8); \
         base blue = (b << 8); \
         base flags = f; \
         base pixel = p; }
#define SET_COLOR_VALS( base, r, g, b ) \
      { base red = r; \
         base green = g; \
         base blue = b; }
#define TOTAL_DENSITY(base) \
           (base red + base green + base blue)
#define SET_NAMED_PIXEL( base, p ) base pixel = p
#define SET_FLAGS( base, f ) base flags = f
#define GET_RED( base ) (base red)
#define GET_GREEN( base ) (base green)
#define GET_BLUE(  base ) (base blue)

extern Visual *AxVisual;

#endif  /*   }   AX_X11 */

#if AXMS     /*  {  */
#define MAX_GUN 0xFF
#define COLORVAL_TYPE unsigned char
#define COLOR_DESCRIPTOR PALETTEENTRY
#define GET_DEFAULT_COLOR_LIST (HPALETTE)DEFAULT_PALETTE
#define SET_RGB_VALS( base, r, g, b, f, p ) \
       { base peRed = r; \
         base peGreen = g; \
         base peBlue = b; \
         base peFlags = 0; }
#define SET_COLOR_VALS( base, r, g, b ) \
       { base peRed = r; \
         base peGreen = g; \
         base peBlue = b; }
#define TOTAL_DENSITY(base) \
           (base peRed + base peGreen + base peBlue)
#define SET_NAMED_PIXEL( base, pixel_val )  \
        pixel_val = (RGB( base peRed, base peGreen, base peBlue ))
#define SET_FLAGS( base, f ) base peFlags = 0
#define GET_RED( base ) (base peRed)
#define GET_GREEN( base ) (base peGreen)
#define GET_BLUE(  base ) (base peBlue)

#define N_AX_PIX 256
extern LOGPALETTE *AxWinPalette;
extern HPALETTE AxWinPaletteHandle;
#endif    /*   }   AXMS  */

#ifndef PIXEL_VALUE
#define PIXEL_VALUE unsigned long
#endif

/****************************************************************/
/* Data elements used by both X-Windows ans MS-Windows            */
/****************************************************************/
int UseWidgetColors;   /* color available for widgets and being used */


int AxLoadWidgetColors = TRUE;
int AxLoadApplicationColors = FALSE;

extern int JavaMode;    /* TRUE for Java slaved axmain */
extern int AxStandardColormapWasLoaded;
extern int AxNColorLevels, AxNColors;
extern int AxSlowDither;
extern PIXEL_VALUE BackPixel;  /* backround for control panels */
extern PIXEL_VALUE ForePixel;  /* foreground for control panels */
extern PIXEL_VALUE SunPixel, ShadowPixel; /*highlight and shadow for dialogs*/
extern PIXEL_VALUE ForeWorkPixel, BackWorkPixel;  /* for alt widget colors */
extern PIXEL_VALUE SunWorkPixel, ShadowWorkPixel, RecessWorkPixel;
extern PIXEL_VALUE ButtonFacePixel;
PIXEL_VALUE RecessPixel;
Colormap WorkColormap;
int Ax64Values[3][6];
COLOR_DESCRIPTOR Ax64Defs[242];
short ColorDisplay;

extern short DpyWid, DpyHt, DpyDepth;
extern Display *Dpy;
extern int AxNoWidgetColor;
int GotWorkAuxColorCells = FALSE;
extern short ToggleFG;
extern unsigned long AlisPixels[];
extern int ThimXErrorCondition;
XStandardColormap *AxStdCmapInfo;


extern GC Gc1; /* graphics context shared by all windows */
extern unsigned int FGpixel, BGpixel;
extern GC ForeGC;  /* label text (nominally black) */
extern GC RecessGC;
extern GC SmallFontGC; /* smaller sans-serif system font, forground mode */
extern GC DisabledGC;      /* "grayed" text (e.g. menu bar inactive) */
extern GC BackGC;  /* "white" */
extern GC DisabledGC;      /* "grayed" text (e.g. menu bar inactive) */
extern GC MonospaceGC; /* monspace system font, forground mode */
extern GC MonospaceBackGC; /* monspace system font, background mode */
extern GC MonospaceBoldGC;
extern GC MonospaceBoldBackGC;
extern GC MonospaceItalGC;
extern GC MonospaceItalBoldGC;
extern GC SmallFontBackGC;
extern GC MonospaceItalBackGC;
extern GC MonospaceItalBoldBackGC;
extern GC SunGC;
extern GC ColorSunGC;
extern GC ShadowGC;
extern GC PixmapperForeGC;
extern GC PixmapperBackGC;
extern GC PixmapperRecessGC;

int AxVisualIsReadOnly=FALSE; /* TRUE for StaticGray, StaticColor, TrueColor */
Colormap ApplixCmap;

extern short NoColor;      /* indicates mono apps - forced if set by profile */
extern unsigned long AxPixels[];
extern unsigned long AxColors[];
extern unsigned long WidgetPixels[];
extern int axStartWidgetPixel, axNumWidgetPixels;
extern int axStartAppPixel, axNumAppPixels;


extern AXWidget_dialog WindowList;
extern int AxOpenLook;


#define BRIGHTNESSFACTOR_SUN    0.12
#define BRIGHTNESSFACTOR_SHADOW 0.35
#define BRIGHTNESSFACTOR_RECESS 0.08



/*************************************************************************
function: WidgetColorsFromProfile - read profile and alloc/assign colors
*/
WidgetColorsFromProfile ()
{
    static int called, gotTheCells = FALSE;
    char wait_string[80];
    XGCValues values;
    int valueMask, invertFlag;

    /* HACK - this test is not sufficient to determine mono display */
    if ( AxNoWidgetColor && called )
    {
        sprintf(wait_string, "%s\n%s",
           XLT2("colorcom.c-[1]","On monochrome displays, color changes take"),
           XLT2("colorcom.c-[2]","effect when you next start Applixware."));
        AxInfoMessage(-1,1,NULL,wait_string);
        return;
    }

    /* get the window foreground and background colors */
    getFGandBGcolors ( );
    called = TRUE;

    /* get the widget colors */
    gotTheCells = WidgetColors(&invertFlag);

    /* assign the colors to the correct GC's */
    valueMask = GCForeground | GCBackground;
    values.foreground = ForePixel;
    values.background = BackPixel;
    XChangeGC(Dpy,ForeGC,valueMask,&values);
    XChangeGC(Dpy, PixmapperForeGC, valueMask, &values);
#if INVERTABLE
    XChangeGC(Dpy,InvertGC,valueMask,&values);
#endif
    XChangeGC(Dpy,SmallFontGC,valueMask,&values);
    XChangeGC(Dpy,DisabledGC,valueMask,&values);
    XChangeGC(Dpy,MonospaceGC,valueMask,&values);
    XChangeGC(Dpy,MonospaceBoldGC,valueMask,&values);
    XChangeGC(Dpy,MonospaceItalGC,valueMask,&values);
    XChangeGC(Dpy,MonospaceItalBoldGC,valueMask,&values);
    values.foreground = RecessPixel;
    XChangeGC(Dpy,RecessGC,valueMask,&values);
    XChangeGC(Dpy, PixmapperRecessGC, valueMask, &values);
    values.foreground = BackPixel;
    values.background = ForePixel;
    XChangeGC(Dpy,BackGC,valueMask,&values);
    XChangeGC(Dpy, PixmapperBackGC, valueMask, &values);
    XChangeGC(Dpy,SmallFontBackGC,valueMask,&values);
    XChangeGC(Dpy,MonospaceBackGC,valueMask,&values);
    XChangeGC(Dpy,MonospaceBoldBackGC,valueMask,&values);
    XChangeGC(Dpy,MonospaceItalBackGC,valueMask,&values);
    XChangeGC(Dpy,MonospaceItalBoldBackGC,valueMask,&values);
    if ( gotTheCells )
    {
        SunGC = ColorSunGC;
        values.foreground = SunPixel;
        values.background = ShadowPixel;
        XChangeGC(Dpy,SunGC,valueMask,&values);
        values.foreground = ShadowPixel;
        values.background = SunPixel;
        XChangeGC(Dpy,ShadowGC,valueMask,&values);
    }
    else /* once entered, cannot go to colors ever again */
    {
        AxNoWidgetColor = TRUE;
        if ( invertFlag )
        {
            ShadowGC = DisabledGC;
            SunGC = ForeGC;
        }
        else
        {
            SunGC = DisabledGC;
            ShadowGC = ForeGC;
        }
    }
}


/***************************************************************************
function: InitColormap - initializes colormap based on user profile switches and
        type of "default visual".  Note that for now we use only DefaultVisual,
        even if client can support better visual.
*/
void
InitColormap (cmapWindow)
Window cmapWindow;
{
    Colormap def_cmap;      /* default color map */
    char *profile, *GetProfileByCode();
    Visual *default_visual;     /* default visual */

    /* get profile of "axLoadWidgetColors" (default is TRUE) */
    AxLoadWidgetColors = TRUE;
    if ( !JavaMode ) {
        profile = GetProfileByCode(PROF_LOAD_WIDGET_COLORS_);
        if (profile && profile[0])
            AxLoadWidgetColors = (atoi(profile) == (-1));
    }

    /* get profile of "axLoadAppColors" (default is TRUE) */
    AxLoadApplicationColors = TRUE;
    if ( !JavaMode ) {
        profile = GetProfileByCode(PROF_LOAD_APP_COLORS_);
        if (profile && profile[0])
            AxLoadApplicationColors = (atoi(profile) == (-1));
    }

    /* get profile of "grFloydSteinbergDither" (default is FALSE) */
    AxSlowDither = FALSE;
    profile = GetProfileByCode(PROF_GE_SLOW_DITHER_);
    if (profile && profile[0])
        AxSlowDither = (atoi(profile) == (-1));

    /* get profile of "axColorCubeHeight" (default is 6) */
    AxNColorLevels = 6;
    if ( !JavaMode ) {
        profile = GetProfileByCode(PROF_COLOR_CUBE_HEIGHT_);
        if (profile && profile[0])
            AxNColorLevels = MAX(1, MIN(6, atoi(profile)));
    }

    /* set up color flags */
    AxNoWidgetColor = NoColor;
    NoColor = FALSE;

    if (AxNColorLevels == 1) {
        NoColor = TRUE;
        AxLoadApplicationColors = TRUE;
    }
    AxNColors = MAX(2, AxNColorLevels * AxNColorLevels * AxNColorLevels);
    ColorDisplay = (DpyDepth != 1);
    def_cmap = DefaultColormap(Dpy, DefaultScreen(Dpy));


    /* hammer these values so there is only one code path */
    if (DpyDepth == 1) {
        AxNColorLevels = 1;
        AxNColors = 2;
        NoColor = TRUE;
        AxNoWidgetColor = TRUE;
        AxLoadWidgetColors = TRUE;
        AxLoadApplicationColors = TRUE;
        AxVisualIsReadOnly = TRUE;
        ApplixCmap = def_cmap;
        return;
    }
    else if (DpyDepth == 4) {
        /* put a cap on number of colors */
        AxNColorLevels = MIN(2, AxNColorLevels);
        AxNColors = MAX(2, AxNColorLevels * AxNColorLevels * AxNColorLevels);
    }

#if AX_X11
    switch (AxVisual->class) {
        case StaticGray:
            NoColor = TRUE;
            AxNoWidgetColor = TRUE;
            AxLoadWidgetColors = TRUE;
            AxLoadApplicationColors = TRUE;
            ApplixCmap = def_cmap;
            AxVisualIsReadOnly = TRUE;
            return;

        case StaticColor:       /* read only colors */
        case TrueColor:
            NoColor = FALSE;
            AxNoWidgetColor = FALSE;
            ApplixCmap = def_cmap;
            AxVisualIsReadOnly = TRUE;
            return;

        case GrayScale:
        case PseudoColor:
        case DirectColor:       /* read/write colors */
        default:
            AxVisualIsReadOnly = FALSE;
            InitializePseudoColormap(cmapWindow);
            break;
    }
#endif
    return;
}

/*************************************************************************
function: InitApplicationColors - Attempt to allocate a NxNxN color table
*/
InitApplicationColors ()
{
    static unsigned long pmasks[8];
    int red, green, blue;
    int red_val, green_val, blue_val;
    int dex = 0;
    int min, max;
    Colormap def_cmap;
    XColor cmapDefs[256];


    def_cmap = GET_DEFAULT_COLOR_LIST;
    switch ( AxNColorLevels )
    {
        case 6:
            /* red */
            Ax64Values[0][0] = 0;
            Ax64Values[0][1] = 15;
            Ax64Values[0][2] = 24;
            Ax64Values[0][3] = 27;
            Ax64Values[0][4] = 29;
            Ax64Values[0][5] = 31;
            /* green */
            Ax64Values[1][0] = 0;
            Ax64Values[1][1] = 15;
            Ax64Values[1][2] = 24;
            Ax64Values[1][3] = 27;
            Ax64Values[1][4] = 29;
            Ax64Values[1][5] = 31;
            /* blue */
            Ax64Values[2][0] = 0;
            Ax64Values[2][1] = 15;
            Ax64Values[2][2] = 24;
            Ax64Values[2][3] = 27;
            Ax64Values[2][4] = 29;
            Ax64Values[2][5] = 31;
        break;
        case 5:
            /* red */
            Ax64Values[0][0] = 0;
            Ax64Values[0][1] = 15;
            Ax64Values[0][2] = 24;
            Ax64Values[0][3] = 27;
            Ax64Values[0][4] = 31;
            /* green */
            Ax64Values[1][0] = 0;
            Ax64Values[1][1] = 15;
            Ax64Values[1][2] = 24;
            Ax64Values[1][3] = 27;
            Ax64Values[1][4] = 31;
            /* blue */
            Ax64Values[2][0] = 0;
            Ax64Values[2][1] = 15;
            Ax64Values[2][2] = 24;
            Ax64Values[2][3] = 27;
            Ax64Values[2][4] = 31;
        break;
        case 4:
            /* red */
            Ax64Values[0][0] = 0;
            Ax64Values[0][1] = 15;
            Ax64Values[0][2] = 24;
            Ax64Values[0][3] = 31;
            /* green */
            Ax64Values[1][0] = 0;
            Ax64Values[1][1] = 15;
            Ax64Values[1][2] = 24;
            Ax64Values[1][3] = 31;
            /* blue */
            Ax64Values[2][0] = 0;
            Ax64Values[2][1] = 15;
            Ax64Values[2][2] = 24;
            Ax64Values[2][3] = 31;
        break;
        case 3:
            /* red */
            Ax64Values[0][0] = 0;
            Ax64Values[0][1] = 15;
            Ax64Values[0][2] = 31;
            /* green */
            Ax64Values[1][0] = 0;
            Ax64Values[1][1] = 15;
            Ax64Values[1][2] = 31;
            /* blue */
            Ax64Values[2][0] = 0;
            Ax64Values[2][1] = 15;
            Ax64Values[2][2] = 31;
        break;
        case 2:
            /* red */
            Ax64Values[0][0] = 0;
            Ax64Values[0][1] = 31;
            /* green */
            Ax64Values[1][0] = 0;
            Ax64Values[1][1] = 31;
            /* blue */
            Ax64Values[2][0] = 0;
            Ax64Values[2][1] = 31;
        break;
    }
 
    /* allocate the color cells and write them into the colormap */
    if ( NoColor == FALSE ) {
        for ( red = 0; red < AxNColorLevels; red++ ) {
            for ( green = 0; green < AxNColorLevels; green++ ) {
                for ( blue = 0; blue < AxNColorLevels; blue++ ) {
                    red_val =  (Ax64Values[0][red]   * 255 / 31);
                    blue_val = (Ax64Values[1][blue] * 255 / 31);
                    green_val = (Ax64Values[2][green]  * 255 / 31);
                    SET_RGB_VALS( Ax64Defs[dex]., red_val, green_val, blue_val, 
                                  (DoRed | DoGreen | DoBlue), AxPixels[dex] );
                    dex++;
                }
            }
        }

#if AX_X11    /* { */
        if (AxVisual->class == DirectColor) {
            XStoreColors(Dpy, ApplixCmap, Ax64Defs, AxNColors);
            AxInitDirectClrTable(Ax64Defs, AxNColors);
        }
        else if (AxVisualIsReadOnly) {      /* TrueColor, StaticColor */
            for (dex =0; dex < AxNColors; ++dex) {
                XAllocColor(Dpy, ApplixCmap, &Ax64Defs[dex]); 
                AxPixels[dex] = Ax64Defs[dex].pixel;
            }
        }
        else if (AxStandardColormapWasLoaded )
        {
            /* PsedoColor with stdmap */
            if ( axPreDefinedStdmap )
            {
                for (dex = 0; dex < 256; ++dex)
                    AxPixels[dex] =  dex;
            }
            else
            {
                for (dex = 0; dex < 256; ++dex)
                    cmapDefs[dex].pixel =  dex;
                XQueryColors(Dpy, ApplixCmap, cmapDefs, 
                                  MIN(AxVisual->map_entries,256));
                AxStdAllocColors(cmapDefs, Ax64Defs, AxNColors);
                for (dex = 0; dex < AxNColors; ++dex)
                    AxPixels[dex] =  Ax64Defs[dex].pixel;
            }
            if (AxLoadApplicationColors && axStartAppPixel > 0)
            {
                for (dex=0; Ax64Defs[dex].pixel < axStartAppPixel; ++dex);
                XStoreColors(Dpy,def_cmap,Ax64Defs+dex,MIN(AxNColors,axNumAppPixels));
            }
        }
        else {              /* PsedoColor with no stdmap */
            if (ApplixCmap != def_cmap && AxLoadApplicationColors && axStartAppPixel > 0) {
                for (dex=0; Ax64Defs[dex].pixel < axStartAppPixel; ++dex);
                XStoreColors(Dpy,def_cmap,Ax64Defs+dex,MIN(AxNColors,axNumAppPixels));
            }
            XStoreColors(Dpy, ApplixCmap, Ax64Defs, AxNColors);
        }

#else    /* } { */
        AxWinSetPalette();
        /* set up application color array */
        AxWinSetPixels();
#endif    /* }  AX_X11 */
    }
    else
    {
        AxPixels[0] = COLOR_BLACK;
        AxPixels[1] = COLOR_WHITE;
    }
    return ( NoColor == FALSE );
}


int
AxGetStdcmapAppColors (cdefs, base)
COLOR_DESCRIPTOR *cdefs;
int base;
{
    register int i,j,k,l;
    int red[6], green[6], blue[6];
    int red_val, green_val, blue_val;

    red[0] = green[0] = blue[0] = 0;
    red[1] = green[1] = blue[1] = 15;
    red[2] = green[2] = blue[2] = 24;
    red[3] = green[3] = blue[3] = 27;
    red[4] = green[4] = blue[4] = 29;
    red[5] = green[5] = blue[5] = 31;

    l = base;
    for (i = 0; i < 6; i++) {
        for (j = 0; j < 6; j++) {
            for (k = 0; k < 6; k++, l++) {
                   red_val =  (int)((float)red[i] * 255 / 31 + 0.5);
                   green_val = (int)((float)green[j] * 255 / 31 + 0.5);
                   blue_val = (int)((float)blue[k] * 255 / 31 + 0.5);
                   SET_RGB_VALS(cdefs[l]., red_val, green_val, blue_val, 
                                 (DoRed | DoGreen | DoBlue), l);
            }
        }
    }
    return(l);
}

AxGetApplicationColors (cdefs, ndefs)
COLOR_DESCRIPTOR *cdefs;
int *ndefs;
{
    register int i, j, k, l;
    int red_val, green_val, blue_val;
    int rgb_values[3][6];

    switch ( AxNColorLevels ) {
        case 6:
            /* red */
            rgb_values[0][0] = 0;
            rgb_values[0][1] = 15;
            rgb_values[0][2] = 24;
            rgb_values[0][3] = 27;
            rgb_values[0][4] = 29;
            rgb_values[0][5] = 31;
            /* green */
            rgb_values[1][0] = 0;
            rgb_values[1][1] = 15;
            rgb_values[1][2] = 24;
            rgb_values[1][3] = 27;
            rgb_values[1][4] = 29;
            rgb_values[1][5] = 31;
            /* blue */
            rgb_values[2][0] = 0;
            rgb_values[2][1] = 15;
            rgb_values[2][2] = 24;
            rgb_values[2][3] = 27;
            rgb_values[2][4] = 29;
            rgb_values[2][5] = 31;
            break;
        case 5:
            /* red */
            rgb_values[0][0] = 0;
            rgb_values[0][1] = 15;
            rgb_values[0][2] = 24;
            rgb_values[0][3] = 27;
            rgb_values[0][4] = 31;
            /* green */
            rgb_values[1][0] = 0;
            rgb_values[1][1] = 15;
            rgb_values[1][2] = 24;
            rgb_values[1][3] = 27;
            rgb_values[1][4] = 31;
            /* blue */
            rgb_values[2][0] = 0;
            rgb_values[2][1] = 15;
            rgb_values[2][2] = 24;
            rgb_values[2][3] = 27;
            rgb_values[2][4] = 31;
            break;
        case 4:
            /* red */
            rgb_values[0][0] = 0;
            rgb_values[0][1] = 15;
            rgb_values[0][2] = 24;
            rgb_values[0][3] = 31;
            /* green */
            rgb_values[1][0] = 0;
            rgb_values[1][1] = 15;
            rgb_values[1][2] = 24;
            rgb_values[1][3] = 31;
            /* blue */
            rgb_values[2][0] = 0;
            rgb_values[2][1] = 15;
            rgb_values[2][2] = 24;
            rgb_values[2][3] = 31;
            break;
        case 3:
            /* red */
            rgb_values[0][0] = 0;
            rgb_values[0][1] = 15;
            rgb_values[0][2] = 31;
            /* green */
            rgb_values[1][0] = 0;
            rgb_values[1][1] = 15;
            rgb_values[1][2] = 31;
            /* blue */
            rgb_values[2][0] = 0;
            rgb_values[2][1] = 15;
            rgb_values[2][2] = 31;
            break;
        case 2:
            /* red */
            rgb_values[0][0] = 0;
            rgb_values[0][1] = 31;
            /* green */
            rgb_values[1][0] = 0;
            rgb_values[1][1] = 31;
            /* blue */
            rgb_values[2][0] = 0;
            rgb_values[2][1] = 31;
            break;
        default:
            *ndefs = 0;
            return;
    }
    for (l = i = 0; i < AxNColorLevels; i++) {
        for (j = 0; j < AxNColorLevels; j++) {
            for (k = 0; k < AxNColorLevels; k++, l++) {
                   red_val = (rgb_values[0][i] * 255 / 31);
                   green_val = (rgb_values[1][j] * 255 / 31);
                   blue_val = (rgb_values[2][k] * 255 / 31);
                   SET_RGB_VALS(cdefs[l]., red_val, green_val, blue_val, 
                                 (DoRed | DoGreen | DoBlue), l);
            }
        }
    }
    *ndefs = AxNColors;
}


/*************************************************************************
function: RGBtoHSB - convert RGB values to HSB
Returns FALSE if hue cannot be calculated, i.e. if pure gray.
red, green, blue, hue, saturation, brightness are numbers in the range 0-1.
hue=0,hue=1 ==> red;  hue=1/3 ==> green;  hue=2/3 => blue
*/
RGBtoHSB(red,green,blue,hue,saturation,brightness)
double red, green, blue;
double *hue, *saturation, *brightness;
{
    double min;
    double hue_red, hue_green, hue_blue;

    *brightness = MAX(red, green);
    *brightness = MAX(*brightness, blue);

    min = MIN(red, green);
    min = MIN(min, blue);
    if ( (*brightness >= -1.0e-25) && (*brightness <= 1.0e-25) )
        *saturation = 0.0; /* brightness = 0 */
    else
        *saturation = (*brightness - min) / (*brightness);

    if ( *saturation == 0.0 ) 
        return(FALSE);
    
    hue_red   = ( *brightness - red )   / ( *brightness - min );
    hue_green = ( *brightness - green ) / ( *brightness - min );
    hue_blue  = ( *brightness - blue)   / ( *brightness - min );
    if (red == *brightness)
    {
        if (green == min)
            if (blue == min)
                *hue = 0.0;
            else
                *hue = 5 + hue_blue;
        else
            *hue = 1 - hue_green;
    }
    else if (green == *brightness)
    {
        if (blue == min)
            *hue = 1 + hue_red;
        else
            *hue = 3 - hue_blue;
    }
    else
    {
        if (red == min)
            *hue = 3 + hue_green;
        else
            *hue = 5 - hue_red;
    }
    *hue = *hue/6.0;
    return(TRUE);
}

/*************************************************************************
function: HSBtoRGB - convert HSB values to RGB
hue, saturation, brightness, red, green, blue are numbers in the range 0-1.
*/
HSBtoRGB(hue,saturation,brightness,red,green,blue)
double hue, saturation, brightness;
double *red, *green, *blue;
{
    int hue_i;
    double hue_f;
    double m, n, k;

    hue = hue*6.0;
    hue_i = (int)hue;
    hue_f = hue - hue_i;

    m = brightness*(1.0-saturation);
    n = brightness*(1.0-(saturation*hue_f));
    k = brightness*(1.0-(saturation*(1.0-hue_f)));

    switch (hue_i)
    {
        case 0:
            *red = brightness;
            *green = k;
            *blue = m;
            break;
        case 1:
            *red = n;
            *green = brightness;
            *blue = m;
            break;
        case 2:
            *red = m;
            *green = brightness;
            *blue = k;
            break;
        case 3:
            *red = m;
            *green = n;
            *blue = brightness;
            break;
        case 4:
            *red = k;
            *green = m;
            *blue = brightness;
            break;
        case 5:
            *red = brightness;
            *green = m;
            *blue = n;
            break;
    }
}

/**************************************************************************
function: DefineAuxColors - define auxiliary colors for drawing widgets
*/
void
DefineAuxColors(backcolor, suncolor, shadowcolor, recesscolor)
COLOR_DESCRIPTOR backcolor, *suncolor, *shadowcolor, *recesscolor;
{
    double red, green, blue;
    double backHue, backSaturation, backBrightness;
    double sunBrightness, shadowBrightness, recessBrightness;
    double sunSaturation;
    double sunBrightnessAdj, sunSaturationAdj;
    double sunFactor, shadowFactor, recessFactor;
    int profileContrast;

    /* normalize background r-g-b values to be in the range [0-1] */
    red   = (double)(GET_RED(backcolor.))   / (double)MAX_GUN;
    green = (double)(GET_GREEN(backcolor.)) / (double)MAX_GUN;
    blue  = (double)(GET_BLUE(backcolor.))  / (double)MAX_GUN;
 
    /* convert background [0-1]r-g-b values to [0-1]h-s-b values */
    if ( !RGBtoHSB(red,green,blue, &backHue,&backSaturation,&backBrightness) )
        backHue = 0.0;  /* hue doesn't matter because saturation is zero */

    /* get widget color contrast preference from profile file */
    profileContrast = UPnumVal(PROF_CONTRAST_);
#if 0
    if (profileContrast < 0)        /* NOTE: profileContrast can be */
        profileContrast = 0;        /* made non-boolean, giving */
    else if (profileContrast > 1)   /* the user more flexibility */
        profileContrast = 1;        /* in adjusting contrast */
#else
    /* untill contrast is changed to non-boolean in prefs editor */
    if (profileContrast != 0)
        profileContrast = 1;
    else 
        profileContrast = 0;
#endif

    /* adjust auxiliary color factors based on contrast preference */
    sunFactor    = BRIGHTNESSFACTOR_SUN    + profileContrast*0.04;
    shadowFactor = BRIGHTNESSFACTOR_SHADOW + profileContrast*0.15;
    recessFactor = BRIGHTNESSFACTOR_RECESS + profileContrast*0.02;;

    /* define auxiliary colors - maintain hue */
#if RIOS || M_88K || NCR || PMAX || BUILD_PURIFY || (LINUX && TURBO)
/* For reasons yet unknown, Asterx 3.0 dies with a segv inside
 * the pow() function at this point during initialization on RS6000. 
 * This workaround produces the exact same results without 
 * calling the pow() function. 
 * On the 88K this does not crash, but one gets really strange results!
*/
    {
    double tmparg1,tmparg2,tmparg3,tmp1,tmp2,tmp3;

    tmparg1 = 2.0*(backBrightness-0.5);
    tmparg2 = backBrightness-0.4;
    tmparg3 = backBrightness-0.6;

    tmp1 = tmparg1*tmparg1*tmparg1*tmparg1*tmparg1*tmparg1;
    tmp2 = 4.0*(tmparg2*tmparg2*tmparg2*tmparg2*tmparg2);
    tmp3 = 5.0*(tmparg3*tmparg3*tmparg3);

    sunBrightnessAdj =sunFactor*((tmp1 + tmp2) - tmp3);
    }
#else

/* If you make a change to the next 3 lines, please make the
   similar change in the if RIOS case above. */

    sunBrightnessAdj = sunFactor*( pow(2.0*(backBrightness-0.5),(double)6.0) +
                                   4.0*pow((backBrightness-0.4),(double)5.0) -
                                   5.0*pow((backBrightness-0.6),(double)3.0) );

#endif



    sunBrightness = backBrightness + sunFactor + sunBrightnessAdj;

    /* if we cannot increase brightness: decrease saturation */
    if (sunBrightness > 1.0)
        sunSaturationAdj = (exp(backSaturation)-0.75)*(sunBrightness-1.0);
    else
        sunSaturationAdj = 0.0;

    sunSaturation = backSaturation - sunSaturationAdj;
    shadowBrightness = backBrightness - shadowFactor;
    recessBrightness = backBrightness - recessFactor;

    /* "flip-around" auxiliary colors if necessary */
    if (!AxOpenLook)
    {
        if ( (sunBrightness > 1.15) && (sunSaturation < 0.0) )
        {
            /* background too bright to display sun color */
            sunBrightness = backBrightness - sunFactor;
            sunSaturation = backSaturation;
            shadowBrightness = sunBrightness - shadowFactor*1.15;
        }
        else if (backBrightness < 0.125)
        {
            /* background too dark to display shadow color */
            shadowBrightness = backBrightness + 
                               0.125*(1.125 - 2.2*backBrightness);
            sunBrightness = shadowBrightness + sunFactor*1.25;
            sunSaturation = backSaturation;
        }
    }

    if (sunBrightness > 1.0)
        sunBrightness = 1.0;
    if (sunSaturation < 0.0)
        sunSaturation = 0.0;
    if (shadowBrightness < 0.0)
        shadowBrightness = 0.0;
    if (recessBrightness < 0.0)
        recessBrightness = 0.0;

    /* convert sun color's h-s-b values back to r-g-b */
    HSBtoRGB(backHue,sunSaturation,sunBrightness,&red,&green,&blue);
    SET_COLOR_VALS( suncolor->, (COLORVAL_TYPE)(red*MAX_GUN),
                           (COLORVAL_TYPE)(green*MAX_GUN),
                           (COLORVAL_TYPE)(blue*MAX_GUN) );
#if 0
    suncolor->red   = (COLORVAL_TYPE)(red*MAX_GUN);
    suncolor->green = (COLORVAL_TYPE)(green*MAX_GUN);
    suncolor->blue  = (COLORVAL_TYPE)(blue*MAX_GUN);
#endif

    /* convert shadow color's h-s-b values back to r-g-b */
    HSBtoRGB(backHue,backSaturation,shadowBrightness,&red,&green,&blue);
    SET_COLOR_VALS( shadowcolor->, (COLORVAL_TYPE)(red*MAX_GUN),
                           (COLORVAL_TYPE)(green*MAX_GUN),
                           (COLORVAL_TYPE)(blue*MAX_GUN) );
#if 0
    shadowcolor->red   = (COLORVAL_TYPE)(red*MAX_GUN);
    shadowcolor->green = (COLORVAL_TYPE)(green*MAX_GUN);
    shadowcolor->blue  = (COLORVAL_TYPE)(blue*MAX_GUN);
#endif

    /* convert recess color's h-s-b values back to r-g-b */
    HSBtoRGB(backHue,backSaturation,recessBrightness,&red,&green,&blue);
    SET_COLOR_VALS(  recesscolor->, (COLORVAL_TYPE)(red*MAX_GUN),
                           (COLORVAL_TYPE)(green*MAX_GUN),
                           (COLORVAL_TYPE)(blue*MAX_GUN) );
#if 0
    recesscolor->red   = (COLORVAL_TYPE)(red*MAX_GUN);
    recesscolor->green = (COLORVAL_TYPE)(green*MAX_GUN);
    recesscolor->blue  = (COLORVAL_TYPE)(blue*MAX_GUN);
#endif
}







WidgetColors(invertFlag)
int *invertFlag;
{
    unsigned long plane_masks[1];
    COLOR_DESCRIPTOR defs[8], foredef, backdef, sundef;
    COLOR_DESCRIPTOR  shadowdef, winbackdef, recessdef;
    int foreRed, foreGreen, foreBlue, backRed, backGreen, backBlue;
    int foreDensity, backDensity, sunDensity, shadowDensity;
    int npixels, i;
    Colormap def_cmap;
    XColor cmapDefs[256];

#if AXMS
    WidgetPixels[ WIDGET_BACKGROUND ] = BackPixel;
    WidgetPixels[ WIDGET_FOREGROUND ] = ForePixel;
    WidgetPixels[ WIDGET_SUN ] = SunPixel;
    WidgetPixels[ WIDGET_SHADOW ] = ShadowPixel;
    WidgetPixels[ WIDGET_RECESS ] = RecessPixel;
    WidgetPixels[ WIDGET_SUN_WORK ] = SunPixel;
    WidgetPixels[ WIDGET_SHADOW_WORK ] = ShadowPixel;
    WidgetPixels[ WIDGET_RECESS_WORK ] = RecessPixel;

    return ( AxNoWidgetColor == FALSE );
#endif  

    /* allocate the widget colors */
    npixels = AxOpenLook ? (ColorDisplay ? 8 : 5) : 5;
    def_cmap = GET_DEFAULT_COLOR_LIST;

    /* read the color values */
    foreRed   = UPnumVal(PROF_FORE_RED_);
    foreGreen = UPnumVal(PROF_FORE_GREEN_);
    foreBlue  = UPnumVal(PROF_FORE_BLUE_);
    backRed   = UPnumVal(PROF_BACK_RED_);
    backGreen = UPnumVal(PROF_BACK_GREEN_);
    backBlue  = UPnumVal(PROF_BACK_BLUE_);
 
    foreRed = MIN(255, MAX(0, foreRed));
    foreGreen = MIN(255, MAX(0, foreGreen));
    foreBlue = MIN(255, MAX(0, foreBlue));
    backRed = MIN(255, MAX(0, backRed));
    backGreen = MIN(255, MAX(0, backGreen));
    backBlue = MIN(255, MAX(0, backBlue));

    SET_RGB_VALS( foredef., foreRed, foreGreen, foreBlue, 0, 0 );
    SET_RGB_VALS( backdef., backRed,  backGreen, backBlue, 0, 0 );
    foreDensity = TOTAL_DENSITY( foredef. );
    backDensity = TOTAL_DENSITY( backdef. );

    if ( !foreDensity && !backDensity )
    {
        /* fix entries from old profile file */
        SET_RGB_VALS( foredef., 0, 0, 0, 0, 0 );
        SET_RGB_VALS( backdef., 230, 230, 230, 0, 0 );
        foreDensity = TOTAL_DENSITY( foredef. );
        backDensity = TOTAL_DENSITY( backdef. );
    }

    /* load the colors into the colormap */
    DefineAuxColors(backdef, &sundef, &shadowdef, &recessdef);
    *invertFlag = (foreDensity > backDensity);
    if ( AxNoWidgetColor == FALSE )
    {
        defs[0] = backdef;
        BackPixel = WidgetPixels[WIDGET_BACKGROUND];
        SET_NAMED_PIXEL( defs[0]., BackPixel );
        SET_FLAGS( defs[0]., (DoRed | DoGreen | DoBlue) );

        defs[1] = foredef;
        ForePixel =  WidgetPixels[WIDGET_FOREGROUND];
        SET_NAMED_PIXEL( defs[1]., ForePixel );
        SET_FLAGS( defs[1]., (DoRed | DoGreen | DoBlue) );

        defs[2] = shadowdef;
        ShadowPixel = WidgetPixels[WIDGET_SHADOW];
        SET_NAMED_PIXEL( defs[2]., ShadowPixel );
        SET_FLAGS( defs[2]., (DoRed | DoGreen | DoBlue) );

        defs[3] = sundef;
        SunPixel = WidgetPixels[WIDGET_SUN];
        SET_NAMED_PIXEL( defs[3]., SunPixel );
        SET_FLAGS( defs[3]., (DoRed | DoGreen | DoBlue) );

        defs[4] = recessdef;
        RecessPixel = WidgetPixels[WIDGET_RECESS];
        SET_NAMED_PIXEL( defs[4]., RecessPixel );
        SET_FLAGS( defs[4]., (DoRed | DoGreen | DoBlue) );

        if ( AxOpenLook )
        {
            if ( UseWidgetColors )
            {
                /* read bg and fg colors, validating values */
                backRed   = UPnumVal(PROF_WK_BACK_RED_);
                backGreen = UPnumVal(PROF_WK_BACK_GREEN_);
                backBlue  = UPnumVal(PROF_WK_BACK_BLUE_);
                if ( !backRed && !backGreen && !backBlue )
                    backGreen = backBlue = backRed = 255;
                backRed   = MIN(255, MAX(0, backRed));
                backGreen = MIN(255, MAX(0, backGreen));
                backBlue  = MIN(255, MAX(0, backBlue));
                SET_RGB_VALS( winbackdef., backRed, backGreen, backBlue, 0, 0 );

                /* determine auxiliary colors for drawing widgets */
                DefineAuxColors(winbackdef, &defs[5], &defs[6], &defs[7]);

                SunWorkPixel = WidgetPixels[WIDGET_SUN_WORK];
                SET_NAMED_PIXEL( defs[5]., SunWorkPixel );
                SET_FLAGS( defs[5]., (DoRed | DoGreen | DoBlue) );

                ShadowWorkPixel = WidgetPixels[WIDGET_SHADOW_WORK];
                SET_NAMED_PIXEL( defs[6]., ShadowWorkPixel );
                SET_FLAGS( defs[6]., (DoRed | DoGreen | DoBlue) );

                RecessWorkPixel = WidgetPixels[WIDGET_RECESS_WORK];
                SET_NAMED_PIXEL( defs[7]., RecessWorkPixel );
                SET_FLAGS( defs[7]., (DoRed | DoGreen | DoBlue) );

                GotWorkAuxColorCells = TRUE;
            }
        }

#if AX_X11
        if (AxVisual->class == DirectColor)
            XStoreColors(Dpy, ApplixCmap, defs, npixels);
        else if (AxVisualIsReadOnly) {
            for (i =0; i < npixels; ++i) {
                XAllocColor(Dpy, ApplixCmap, &defs[i]);   
                WidgetPixels[i] = defs[i].pixel;
            }
            BackPixel = WidgetPixels[WIDGET_BACKGROUND];
            ForePixel =  WidgetPixels[WIDGET_FOREGROUND];
            ShadowPixel = WidgetPixels[WIDGET_SHADOW];
            SunPixel = WidgetPixels[WIDGET_SUN];
            RecessPixel = WidgetPixels[WIDGET_RECESS];
            SunWorkPixel = WidgetPixels[WIDGET_SUN_WORK];
            ShadowWorkPixel = WidgetPixels[WIDGET_SHADOW_WORK];
            RecessWorkPixel = WidgetPixels[WIDGET_RECESS_WORK];
        }
        else if (AxStandardColormapWasLoaded && 
                 ApplixCmap == AxStdCmapInfo->colormap) {
            if ( axPreDefinedStdmap )
            {
                /* predefined standard maps */
                for (i = 0; i < npixels; ++i)
                    WidgetPixels[i] =  convertStdRGBToPixel(&defs[i]);
            }
            else
            {
                for (i = 0; i < 256; i++)
                    cmapDefs[i].pixel = i;
                XQueryColors(Dpy, ApplixCmap, cmapDefs, 
                                  MIN(AxVisual->map_entries,256));
                AxStdAllocColors(cmapDefs, defs, npixels);
                for (i = 0; i < npixels; ++i)
                    WidgetPixels[i] =  defs[i].pixel;
            }
            BackPixel = WidgetPixels[WIDGET_BACKGROUND];
            ForePixel =  WidgetPixels[WIDGET_FOREGROUND];
            ShadowPixel = WidgetPixels[WIDGET_SHADOW];
            SunPixel = WidgetPixels[WIDGET_SUN];
            RecessPixel = WidgetPixels[WIDGET_RECESS];
            SunWorkPixel = WidgetPixels[WIDGET_SUN_WORK];
            ShadowWorkPixel = WidgetPixels[WIDGET_SHADOW_WORK];
            RecessWorkPixel = WidgetPixels[WIDGET_RECESS_WORK];
#if 0
            if (AxLoadWidgetColors && axStartWidgetPixel > 0) {
                for (i=0; defs[i].pixel < axStartWidgetPixel; ++i);
                XStoreColors(Dpy, def_cmap, defs+i, MIN(npixels, axNumWidgetPixels));
            }
#endif
        }
        else {
            if ( ApplixCmap != def_cmap && AxLoadWidgetColors && axStartWidgetPixel > 0) {
                for (i=0; defs[i].pixel < axStartWidgetPixel; ++i);
                XStoreColors(Dpy, def_cmap, defs+i, MIN(npixels, axNumWidgetPixels));
            }
            XStoreColors(Dpy, ApplixCmap, defs, npixels);
        }
#else
#if 0
        MS_SetPaletteEntries(ApplixCmap, 0,npixels, defs);
#endif
#endif
    }
    else /* color allocation failed - use system defaults */
    {
        if ( *invertFlag )
        {
            ForePixel = COLOR_WHITE;
            BackPixel = COLOR_BLACK;
        }
        else
        {
            BackPixel = COLOR_WHITE;
            ForePixel = COLOR_BLACK;
        }
        WidgetPixels[WIDGET_BACKGROUND] = BackPixel;
        WidgetPixels[WIDGET_FOREGROUND] = ForePixel;
    }
    return ( AxNoWidgetColor == FALSE );
}


getFGandBGcolors()
{
    Colormap def_cmap;
    COLOR_DESCRIPTOR defs[2];
    int foreR,foreG,foreB,backR,backG,backB;
    int i;
    XColor cmapDefs[256];


    def_cmap = GET_DEFAULT_COLOR_LIST;

    /* fetch GF/BG values from profile */
    getFGandBGfromProfile(&foreR,&foreG,&foreB,&backR,&backG,&backB);

#if AX_X11
    /* once NoColor is set, cannot go back */
    if ( !ColorDisplay || NoColor || DpyDepth == 1 )
    {
        NoColor = TRUE;
        if ( (foreR + foreG + foreB) > (backR + backG + backB) )
        {
            ToggleFG = TRUE;
            ForeWorkPixel = COLOR_WHITE;
            BackWorkPixel = COLOR_BLACK;
        }
        else
        {
            ToggleFG = FALSE;
            BackWorkPixel = COLOR_WHITE;
            ForeWorkPixel = COLOR_BLACK;
        }
        AlisPixels[0] = BackWorkPixel;
        AlisPixels[1] = ForeWorkPixel;
    }
    else /* we have color (at least win fg and bg) */
    {
        BackWorkPixel = AlisPixels[0];
        ForeWorkPixel = AlisPixels[1];

        /* set the bg color */
        SET_RGB_VALS( defs[0]., backR, backG, backB, 
                               (DoRed | DoGreen | DoBlue), AlisPixels[0] );

        /* set the fg color */
        SET_RGB_VALS( defs[1]., foreR, foreG, foreB, 
                               (DoRed | DoGreen | DoBlue), AlisPixels[1] );

        /* write fg and bg colors to colormap */
        if (AxVisual->class == DirectColor)
            XStoreColors(Dpy, ApplixCmap, defs, 2);
        else if (AxVisualIsReadOnly) {
            for (i =0; i < 2; ++i) {
                XAllocColor(Dpy, ApplixCmap, &defs[i]);   
                AlisPixels[i] = defs[i].pixel;
            }
            BackWorkPixel = AlisPixels[0];
            ForeWorkPixel = AlisPixels[1];
        }
        else if ( AxStandardColormapWasLoaded  &&
                 ApplixCmap == AxStdCmapInfo->colormap) {
            /* predefined standard maps */
            if ( axPreDefinedStdmap )
            {
                for (i = 0; i < 2; ++i)
                    AlisPixels[i] =  convertStdRGBToPixel(&defs[i]);
            }
            else
            {
                for (i = 0; i < 256; i++)
                    cmapDefs[i].pixel = i;
                XQueryColors(Dpy, AxStdCmapInfo->colormap, cmapDefs,
                                  MIN(AxVisual->map_entries,256));
                AxStdAllocColors(cmapDefs, defs, 2);
                for (i = 0; i < 2; ++i)
                    AlisPixels[i] =  defs[i].pixel;
            }
            BackWorkPixel = AlisPixels[0];
            ForeWorkPixel = AlisPixels[1];
        }
        else {
            if ( ApplixCmap != def_cmap && AxLoadWidgetColors )
                XStoreColors(Dpy, def_cmap, defs, 2);
            XStoreColors(Dpy, ApplixCmap, defs, 2);
        }
    }
#endif

    /* set initial colors to defaults (black on white) and return */
    WMsetColors(DEFAULT_FG, DEFAULT_BG);
}





getFGandBGfromProfile ( foreR, foreG, foreB, backR, backG, backB)
int *foreR, *foreG, *foreB, *backR, *backG, *backB;
{
    *foreR = UPnumVal(PROF_WK_FORE_RED_);
    *foreG = UPnumVal(PROF_WK_FORE_GREEN_);
    *foreB = UPnumVal(PROF_WK_FORE_BLUE_);
    *backR = UPnumVal(PROF_WK_BACK_RED_);
    *backG = UPnumVal(PROF_WK_BACK_GREEN_);
    *backB = UPnumVal(PROF_WK_BACK_BLUE_);

    /* fixup for empty profile (all vals 0) */
    if ( !*foreR && !*foreG && !*foreB && !*backR && !*backG && !*backB)
    {
        *backR = 255;
        *backG = 255;
        *backB = 255;
    }

    /* validate range */
    *foreR = MAX(0,*foreR);
    *foreR = MIN(255,*foreR);
    *foreG = MAX(0,*foreG);
    *foreG = MIN(255,*foreG);
    *foreB = MAX(0,*foreB);
    *foreB = MIN(255,*foreB);
    *backR = MAX(0,*backR);
    *backR = MIN(255,*backR);
    *backG = MAX(0,*backG);
    *backG = MIN(255,*backG);
    *backB = MAX(0,*backB);
    *backB = MIN(255,*backB);
}

/***************************************************************************
function: axGetPackedRGBFromColor - returns packed RGB color
*/
static int
axGetPackedRGBFromColor (type, v1, v2, v3, v4, max_v)
int type, v1, v2, v3, v4, max_v;
{
    int rgb;

    switch (type) {
        case 1: /* hsv */ 
            break;
        case 2: /* cmyk */
            /* cmyk to rgb */
            v1 = max_v - MIN(max_v, v1+v4); /* red */
            v2 = max_v - MIN(max_v, v2+v4); /* green */
            v3 = max_v - MIN(max_v, v3+v4); /* blue */
            break;
        case 0: /* rgb */
        default:
            break;
    }
    if (max_v != 255) {
        v1 = v1 * 255 / max_v;
        v2 = v2 * 255 / max_v;
        v3 = v3 * 255 / max_v;
    }
#if AXMS
    if ( DpyDepth == 15 || DpyDepth == 16 )
        rgb = packedRGB16(v1,v2,v3);
    else
#endif
    rgb = packedRGB(v1,v2,v3);
    return(rgb);
}

convertStdDefMapToPixel(r, g, b, max)
int r,g,b, max;
{
    unsigned long pixel;
    double val;

    if ( r == g && g == b )
    {
        val = (double)r / (double)max;
        pixel = AxStdCmapInfo->base_pixel +
         ((unsigned long)(val*AxStdCmapInfo->red_max)*
          AxStdCmapInfo->red_mult) +
         ((unsigned long)(val*AxStdCmapInfo->green_max)*
          AxStdCmapInfo->green_mult) +
         ((unsigned long)(val*AxStdCmapInfo->blue_max) *
          AxStdCmapInfo->blue_mult);
    }
    else
    {
        pixel = AxStdCmapInfo->base_pixel + RSTDCOMP(AxStdCmapInfo, r, max) +
                GSTDCOMP(AxStdCmapInfo, g, max)+BSTDCOMP(AxStdCmapInfo,b,max);
    }
    return((int)pixel);
}

axConvertStdToGray(red, green, blue, max_v) 
int red, green, blue, max_v;
{
    float val;

    val = 255/(float)max_v;
    red *= val;
    green *= val;
    blue *= val;
    return((int)MONO(red, green, blue));
}

convertStdGrayMapToPixel(color, color_max, color_mult)
int color, color_max, color_mult;
{
    int pixel;

    pixel = AxStdCmapInfo->base_pixel +
            (int)(0.5+((float)(color)/255)* color_max) * color_mult;
    return(pixel);
}

/*************************************************************************
function: axGetColormapIndex - get the nearest pixel to the requested color
*/
void axGetColormapIndex(type, v1, v2, v3, v4, max_v, cindex)
int type, v1, v2, v3, v4, max_v, *cindex;
{
    /* convert to rgb if needed */
    if (type != 0)
    {
        /* hsv => rgb */
        if (type == 1)
        {
            v1 = v1;
            v2 = v2;
            v3 = v3;
        }

        /* cmyk => rgb */
        else if (type == 2)
        {
            /* cyan to red */
            v1 += v4;
            if (v1 > max_v)
                v1 = 0;
            else
                v1 = max_v - v1;

            /* magenta to green */
            v2 += v4;
            if (v2 > max_v)
                v2 = 0;
            else
                v2 = max_v - v2;

            /* yellow to blue */
            v3 += v4;
            if (v3 > max_v)
                v3 = 0;
            else
                v3 = max_v - v3;
        }
    }
    if ( axPreDefinedStdmap ) switch ( axXA_RGB_APPLIX_MAP ) 
    {
        default:
            *cindex = convertStdDefMapToPixel(v1, v2, v3, max_v);
        break;

        case XA_RGB_RED_MAP:
            *cindex = convertStdGrayMapToPixel(axConvertStdToGray(v1, v2, v3, max_v),
                                           AxStdCmapInfo->red_max,
                                           AxStdCmapInfo->red_mult);
        break;

        case XA_RGB_GREEN_MAP:
            *cindex = convertStdGrayMapToPixel(axConvertStdToGray(v1, v2, v3, max_v),
                                           AxStdCmapInfo->green_max,
                                           AxStdCmapInfo->green_mult);
        break;

        case XA_RGB_BLUE_MAP:
            *cindex = convertStdGrayMapToPixel(axConvertStdToGray(v1, v2, v3, max_v),
                                           AxStdCmapInfo->blue_max,
                                           AxStdCmapInfo->blue_mult);
        break;
    }
    
    else switch (AxNColorLevels)
    {
        case 1:  /* black or white */
            if (v1 == max_v && v2 == max_v && v3 == max_v)
                *cindex = 1;
            else
                *cindex = 0;
        break;
        case 2:  /* 8 colors */
            get_2_level_color(&v1, max_v);
            get_2_level_color(&v2, max_v);
            get_2_level_color(&v3, max_v);
            *cindex = v1 * 4 + v2 * 2 + v3;
        break;
        case 3:  /* 27 colors */
            get_3_level_color(&v1, max_v);
            get_3_level_color(&v2, max_v);
            get_3_level_color(&v3, max_v);
            *cindex = v1 * 9 + v2 * 3 + v3;
        break;
        case 4:  /* 64 colors */
            get_4_level_color(&v1, max_v);
            get_4_level_color(&v2, max_v);
            get_4_level_color(&v3, max_v);
            *cindex = v1 * 16 + v2 * 4 + v3;
        break;
        case 5:  /* 125 colors */
            get_5_level_color(&v1, max_v);
            get_5_level_color(&v2, max_v);
            get_5_level_color(&v3, max_v);
            *cindex = v1 * 25 + v2 * 5 + v3;
        break;
        case 6:  /* 216 colors */
        default:
            get_6_level_color(&v1, max_v);
            get_6_level_color(&v2, max_v);
            get_6_level_color(&v3, max_v);
            *cindex = v1 * 36 + v2 * 6 + v3;
        break;
    }
}



/***************************************************************************
function: AxGetRealPixel - gets X pixel value of color in question depending on
            visual type.
    o   For first four visuals (StaticGray to PseudoColor) it returns pixel
            of nearest color in already loaded colors.
    o   For Truecolor visual it allocates a readonly color if necessary and
            returns pixel.
    o   For Directcolor visual it allocates a read/write color if necessary and
            returns pixel.
    
*/
unsigned long
AxGetRealPixel (type, c, m, y, k, maxValue, outval)
int type, c, m, y, k, maxValue;
unsigned long *outval;
{
    int index;
    int rgb;
    XColor cdef;

#if AXMS
    if ( DpyDepth > 8 )
        {
        rgb = axGetPackedRGBFromColor(type, c, m, y, k, maxValue);
        *outval = rgb;
        }
    else
        {
        axGetColormapIndex(type, c, m, y, k, maxValue, &index);
        *outval = AxPixels[index];
        }
#else
    switch (AxVisual->class) {
        case StaticGray:
        case GrayScale:
        case StaticColor:       /* read only colors */
        case PseudoColor:
            axGetColormapIndex(type, c, m, y, k, maxValue, &index);
            *outval = AxPixels[index];
            break;
        case TrueColor:
            rgb = axGetPackedRGBFromColor(type, c, m, y, k, maxValue);
            cdef.red = packedRed(rgb) << 8;
            cdef.green = packedGreen(rgb) << 8;
            cdef.blue = packedBlue(rgb) << 8;
            cdef.flags = (DoRed | DoGreen | DoBlue);
            axGetTrueColorPixel(&cdef, outval);
            break;
        case DirectColor:       /* read/write colors */
            rgb = axGetPackedRGBFromColor(type, c, m, y, k, maxValue);
            cdef.red = packedRed(rgb) << 8;
            cdef.green = packedGreen(rgb) << 8;
            cdef.blue = packedBlue(rgb) << 8;
            cdef.flags = (DoRed | DoGreen | DoBlue);
            axGetDirectColorPixel(&cdef, rgb, outval);
            break;
    }
#endif
    return (*outval);
}


convertRGBTograylevel(r, g, b, max_v)
int r, g, b, max_v;
{
    int pixel;

    switch ( axXA_RGB_APPLIX_MAP ) 
    {
        default:
            pixel = convertStdDefMapToPixel(r, g, b, max_v);
        break;
        case XA_RGB_RED_MAP:
            pixel = convertStdGrayMapToPixel(axConvertStdToGray(r, g, b,max_v),
                                                   AxStdCmapInfo->red_max,
                                                  AxStdCmapInfo->red_mult);
        break;
        case XA_RGB_GREEN_MAP:
            pixel = convertStdGrayMapToPixel(axConvertStdToGray(r,g,b,max_v),
                                                   AxStdCmapInfo->green_max,
                                                    AxStdCmapInfo->green_mult);
        break;
        case XA_RGB_BLUE_MAP:
            pixel = convertStdGrayMapToPixel(axConvertStdToGray(r,g,b,max_v),
                                                   AxStdCmapInfo->blue_max,
                                                   AxStdCmapInfo->blue_mult);
        break;
    }
    return(pixel);
}

convertStdRGBToPixel(cmapdefs)
XColor *cmapdefs;
{
    int pixel;
    pixel = convertRGBTograylevel(cmapdefs->red, cmapdefs->green,
                                  cmapdefs->blue, 0xFFFF);
    return(pixel);
}

axWhitePixel()
{
}

axBlackPixel()
{
}

/***************************************************************************
function: AxPixelFromCMYK - Return pixel closest to passed CMYK values
*/
unsigned long AxPixelFromCMYK(cyan, magenta, yellow, black, maxval)
int cyan, magenta, yellow, black;
int maxval;
{
    int ix;      /* colormap index */
    unsigned long pixel;
    unsigned long whitePixel;
    unsigned long blackPixel;

    AxGetRealPixel(2, cyan, magenta, yellow, black, maxval, &pixel);

    if (!ColorDisplay) {
        whitePixel = WhitePixel(Dpy, DefaultScreen(Dpy));
        blackPixel = BlackPixel(Dpy, DefaultScreen(Dpy));
        if (ToggleFG)
            pixel = (pixel == whitePixel) ? blackPixel : whitePixel;
        else
            pixel = (pixel != whitePixel) ? blackPixel : whitePixel;
    }

    return (pixel);
}

/* COPIED this from x2win.c */
/***********************************************************************/
/* We have lifed the "axrgb.txt" file from X, and read it here to allow  */
/* ELF code to set the colors on a widget by name, ie. "blue".  Rather */
/* than reading and storing the entire axrgb.txt file, we read it when we *
/* get a request for a new color, and then cache that entry.  While this */
/* is may be slower on startup, it saves quite a bit of memory, and it  */
/* is expected that any one user will use only a small fraction of the  */
/* color names in axrgb.txt (there are 738 colors defined)      - jon     */
/***********************************************************************/

/* Variables being used in AxXParseColor */
#define MAX_COLOR_NAME 21
typedef struct _color_list {
    unsigned short red;
    unsigned short green;
    unsigned short blue;
    char name[MAX_COLOR_NAME];
    } color_list;
 
extern char *ThimApplixDir;
extern char *DirSlash;

static
AxXParseColor( display, cmap, color_name, cdef )
void *display;
Colormap cmap;
char *color_name;
XColor *cdef;
{
    int i;
    char temp_name[MAX_COLOR_NAME];
    color_list *cP;
    FILE *fp;
    int red, green, blue;
    char filename[STRPATH+1];
	char *pfilename;	/* profile setting of alternate rgb.txt file */
	char *UPbyName();
    char buff[200];
    char *name, *p;
 
	static color_list *colorP=0;
	static int size_colorP=0;
 
    /* make sure name is not too long */
    strncpy( temp_name, color_name, MAX_COLOR_NAME-1 );
    temp_name[MAX_COLOR_NAME] = 0;
 
    /** se if the requested color is already in our cache  */
    cP = colorP;
    for ( i=0; i<size_colorP; i++ )
       {
       if ( streqi( temp_name, cP->name ) != 0 )
          {
          /* found it */
          cdef->red = cP->red;
          cdef->green = cP->green;
          cdef->blue = cP->blue;
          return( TRUE );
          }
       cP++;
       }
 
    /** fell through not in our cached list.  Read through the RGB file */
    /** to find the requested entry, and add it to the cache             */
	if (NULL == (pfilename = UPbyName(NXLT("axColorNamesTable"))) ) {
    	strcpy( filename, ThimApplixDir );
    	strcat( filename, DirSlash );
    	strcat( filename, "axrgb.txt" );
	}
	else
		strcpy(filename, pfilename);
 
    fp = fopen( filename, READ_STR );
    if ( fp == (FILE *)NULL )
        return( FALSE );
    while ( fgets( buff, sizeof(buff), fp ) != (char *)NULL )
       {
       if ( buff[0] == ';' )
          /* comment */
          continue;
 
       p = buff + strlen(buff);
       while ( *p < ' ' )
         {
         /* step on any trailing newlines */
         *p = '\0';
         p--;
         }
       p = buff;
       while ( *p <= ' ' )
         { p++; }
       red = atoi( buff );
       p = index( p, (int)' ' );
       p++;
       while ( *p <= ' ' )
         { p++; }
       green = atoi( p );
       p = index( p, (int)' ' );
       p++;
       while ( *p <= ' ' )
         { p++; }
       blue = atoi( p );
       while ( *p > ' ')
          { p++; }        /* skip over "blue" numbers */
       while ( *p <= ' ' )
          { p++; }        /* skip to color name */
       name = p;
 
       if ( streqi( temp_name, name ) != 0 )
           {
           /* found it, add to the cache */
           size_colorP++;
           colorP = (color_list *)realloc( colorP, (size_colorP * sizeof(color_list)) );
           cP = colorP + size_colorP - 1;
           /* X seems to store the RGB values in the upper byte of the short */
           cdef->red = cP->red = red << 8;
           cdef->green = cP->green = green << 8;
           cdef->blue = cP->blue = blue<<8;
           strcpy( cP->name, name );
           fclose( fp );
           return( TRUE );
           }
       }
 
    /* feel through, did not find the color */
    fclose( fp );
    return( FALSE );
}


/***************************************************************************
function: AxPixelFromColorname - return pixel closest to color description
This is done in two steps, ParseColor to get r,g,b values and then find closest
pixel for those r,g,b values
*/
unsigned long
AxPixelFromColorname (color_name, pixelP)
char *color_name;
unsigned long *pixelP;
{
    unsigned long pix;
    XColor cdef;

    pix = BlackPixel(Dpy, DefaultScreen(Dpy));
    if (AxXParseColor(Dpy, ApplixCmap, color_name, &cdef)) {
        AxGetRealPixel(0, (cdef.red>>8), (cdef.green>>8), (cdef.blue>>8), 
                                                                0, 255, &pix);
    }
    if (pixelP)
        *pixelP = pix;
    return(pix);
}

data_ptr
elfColorFromName (color_name)
char *color_name;
{
    data_ptr color = AxMakeArray(3);
    data_ptr elfColor = AxMakeArray(2);
    XColor cdef;
    if (AxXParseColor(Dpy, ApplixCmap, color_name, &cdef)) {
        AxAddIntToArray(color,0,(cdef.red>>8));
        AxAddIntToArray(color,1,(cdef.green>>8));
        AxAddIntToArray(color,2,(cdef.blue>>8));
    }
    AxAddIntToArray(elfColor,0,1);      /* color type WCSPACE_RGB -> xmwcolor.h */
    AxAddArrayToArray(elfColor,1,color);
    return elfColor;
}


/****************************************************************************/
double percent_levels[] = {0.0, 50.0, 75.0, 87.0, 95.0, 100.0};


/*****************************************************************************
 * module : create_cmap :- creates colormap used by all images/bitmaps
 *
 *   DO NOT CHANGE THE LAYOUT OF THIS COLORMAP TABLE.  THE REV3/REV4 ICON
 *   IMAGES WHICH USE THE 'DEFAULT' COLORMAP USE INDEXES RELATED TO THIS 
 *   TABLE.
 *
 * Also, the table built here is reflected in the lookup calculations in
 *   getColorIndex()
 */
create_cmap(pid, cmap)
int pid;
cmap_type *cmap;
{
    int ix, r,g,b, c,m,y,k;
    char *col_str;

    /* create the colormap */
    cmap->n_colors = 256;
    cmap->color = (color_type *)TaskAlloc(pid, 256 * sizeof(color_type));

    /* transparent in slot 0 */
    cmap->color[0].see_thru = TRUE;
    cmap->color[0].ink_type = 0;
    cmap->color[0].c = 0;
    cmap->color[0].m = 0;
    cmap->color[0].y = 0;
    cmap->color[0].k = 0;
    col_str = XLT("Transparent");
    cmap->color[0].name = (char *) TaskAlloc(pid, (strlen(col_str)+1));
    strcpy(cmap->color[0].name, col_str);

    /* black in slot 1 */
    cmap->color[1].see_thru = FALSE;
    cmap->color[1].ink_type = 0;
    cmap->color[1].c = 0;
    cmap->color[1].m = 0;
    cmap->color[1].y = 0;
    cmap->color[1].k = 255;
    col_str = XLT("Black");
    cmap->color[1].name = (char *) TaskAlloc(pid, (strlen(col_str)+1));
    strcpy(cmap->color[1].name, col_str);

    /* white in slot 2 */
    cmap->color[2].see_thru = FALSE;
    cmap->color[2].ink_type = 0;
    cmap->color[2].c = 0;
    cmap->color[2].m = 0;
    cmap->color[2].y = 0;
    cmap->color[2].k = 0;
    col_str = XLT("White");
    cmap->color[2].name = (char *) TaskAlloc(pid, (strlen(col_str)+1));
    strcpy(cmap->color[2].name, col_str);

    /* load the rest of the grey values */
    for ( ix=3, r=1; r<5; r++ )
    {
        cmap->color[ix].see_thru = FALSE;
        cmap->color[ix].ink_type = 0;
        cmap->color[ix].c = 0;
        cmap->color[ix].m = 0;
        cmap->color[ix].y = 0;
        cmap->color[ix].k = 255 - (int)percent_levels[r] *255/100.0;
        cmap->color[ix].name = (char *)TaskAlloc(pid,9);
        sprintf(cmap->color[ix].name, XLT("color%03d"), ix);
        ix++;
    }
    
    for ( r=0; r<6; r++ )
    {
        for ( g=0; g<6; g++ )
        {
            for ( b=0; b<6; b++ )
            {
                /* don't redo the grey values */
                if ( r == g && g == b )
                    continue;

                /* create a non-tranparent process color */
                cmap->color[ix].see_thru = FALSE;
                cmap->color[ix].ink_type = 0;
 
                /* calc the cmy values */
                c = 255.0 - percent_levels[r] * 255.0 / 100.0;
                m = 255.0 - percent_levels[g] * 255.0 / 100.0;
                y = 255.0 - percent_levels[b] * 255.0 / 100.0;

                /* calc the black value */
                if (c == 0 || m == 0 || y == 0)
                    k = 0;
                else
                {
                    if (c < m)
                        k = (c < y) ? c : y;
                    else
                        k = (m < y) ? m : y;
                    k = k/2;
                }

                /* set the cmyk values */
                cmap->color[ix].c = c - k;
                cmap->color[ix].m = m - k;
                cmap->color[ix].y = y - k;
                cmap->color[ix].k = k;

                /* name the color */
                cmap->color[ix].name = (char *)TaskAlloc(pid, 9);
                sprintf(cmap->color[ix].name, XLT("color%03d"), ix);
                ix++;
            }
        }
    }

    for ( ; ix<256; ix++ )
    {
        cmap->color[ix].see_thru = FALSE;
        cmap->color[ix].ink_type = 0;
        cmap->color[ix].c = 0;
        cmap->color[ix].m = 0;
        cmap->color[ix].y = 0;
        cmap->color[ix].k = 0;
        cmap->color[ix].name = (char *) TaskAlloc(pid, 9);
        sprintf(cmap->color[ix].name, XLT("color%03d"), ix);
    }
}

/* Compares c, m, y, k values of a pair of colormaps; returns TRUE if the
	colormaps are the same, FALSE if different. */
int
compare_cmaps ( cmap1, cmap2 )
cmap_type *cmap1, *cmap2;
{
    int ix, r,g,b, c,m,y,k;
    char *col_str;

    for ( ix=0; ix<256; ix++ ) {
		if (cmap1->color[ix].c != cmap1->color[ix].c)
			return( FALSE);
		if (cmap1->color[ix].m != cmap1->color[ix].m)
			return( FALSE);
		if (cmap1->color[ix].y != cmap1->color[ix].y)
			return( FALSE);
		/* If c, m, y are the same, so are the k's */
	}
	return( TRUE);
}

/****************************************************************************/
static int v_pat_table[] = {
     0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
     1,  1,  1,  1,  1,  1,  2,  2,  2,  2,
     2,  2,  3,  3,  3,  3,  3,  3,  4,  4,
     4,  4,  4,  4,  5,  5,  5,  5,  5,  5,
     6,  6,  6,  6,  6,  6,  7,  7,  7,  7,
     7,  7,  8,  8,  8,  8,  8,  8,  9,  9,
     9,  9,  9,  9, 10, 10, 10, 10, 10, 10,
    11, 11, 11, 11, 11, 11, 12, 12, 12, 12,
    12, 12, 13, 13, 13, 13, 13, 13, 14, 14,
    14, 14, 14, 14, 15, 15, 15, 15, 15, 15,
    15
};

/****************************************************************************/
find_nearest_6x6x6(v, v1, v2, v_pat,percent) /* 0,50,75,87,95,100 */
int v, *v1, *v2, *v_pat;
int percent;
{
    switch ( v )
    {
        case 0:
            *v1 = *v2 = 0;
            *v_pat = 0;
        break;

        case 50:
            *v1 = *v2 = 1;
            *v_pat = 0;
        break;

        case 75:
            *v1 = *v2 = 2;
            *v_pat = 0;
        break;

        case 87:
            *v1 = *v2 = 3;
            *v_pat = 0;
        break;

        case 95:
            *v1 = *v2 = 4;
            *v_pat = 0;
        break;

        case 100:
            *v1 = *v2 = 5;
            *v_pat = 0;
        break;

        /* index based on expression: 100 * ABS(v - val1) / ABS(val2 - val1) */
        default:
            if ( v < 25 ) /* 1-24 */
            {
                *v1 = 0;
                *v2 = 1;
                *v_pat = v_pat_table[100 * v / 50];
            }
                    
            else if ( v < 50 ) /* 25-49 */
            {
                *v1 = 1;
                *v2 = 0;
                *v_pat = v_pat_table[100 * (50 - v) / 50];
            }
            else if ( v < 63 ) /* 51-62 */
            {
                *v1 = 1;
                *v2 = 2;
                *v_pat = v_pat_table[100 * (v - 50) / 25];
            }
            else if ( v < 75 ) /* 63-74 */
            {
                *v1 = 2;
                *v2 = 1;
                *v_pat = v_pat_table[100 * (75 - v) / 25];
            }
            else if ( v < 81 ) /* 76-80 */
            {
                *v1 = 2;
                *v2 = 3;
                *v_pat = v_pat_table[100 * (v - 75) / 12];
            }
            else if ( v < 87 ) /* 81-86 */
            {
                *v1 = 3;
                *v2 = 2;
                *v_pat = v_pat_table[100 * (87 - v) / 12];
            }
            else if ( v < 91 ) /* 88-90 */
            {
                *v1 = 3;
                *v2 = 4;
                *v_pat = v_pat_table[100 * (v - 87) / 8];
            }
            else if ( v < 95 ) /* 91-94 */
            {
                *v1 = 4;
                *v2 = 3;
                *v_pat = v_pat_table[100 * (95 - v) / 8];
            }
            else if ( v < 98 ) /* 96-97 */
            {
                *v1 = 4;
                *v2 = 5;
                *v_pat = v_pat_table[100 * (v - 95) / 5];
            }
            else /* 98-99 */
            {
                *v1 = 5;
                *v2 = 4;
                *v_pat = v_pat_table[100 * (100 - v) / 5];
            }
        break;
    }
}

find_nearest_r_6x6x6(r, r1, r2, r_pat,percent)
int r, *r1, *r2, *r_pat,percent;
{
    find_nearest_6x6x6(r, r1, r2, r_pat,percent );
    *r1 *= 36; /* weight for red */
    *r2 *= 36;
}

find_nearest_g_6x6x6(g, g1, g2, g_pat,percent)
int g, *g1, *g2, *g_pat,percent;
{
    find_nearest_6x6x6(g, g1, g2, g_pat,percent );
    *g1 *= 6; /* weight for green */
    *g2 *= 6;
}

find_nearest_b_6x6x6(b, b1, b2, b_pat,percent)
int b, *b1, *b2, *b_pat,percent;
{
    find_nearest_6x6x6(b, b1, b2, b_pat,percent );
}

/****************************************************************************/
find_nearest_5x5x5(v, v1, v2, v_pat,percent) /* 0,50,75,87,100 */
int v, *v1, *v2, *v_pat;
int percent;
{
    switch ( v )
    {
        case 0:
            *v1 = *v2 = 0;
            *v_pat = 0;
        break;

        case 50:
            *v1 = *v2 = 1;
            *v_pat = 0;
        break;

        case 75:
            *v1 = *v2 = 2;
            *v_pat = 0;
        break;

        case 87:
            *v1 = *v2 = 3;
            *v_pat = 0;
        break;

        case 100:
            *v1 = *v2 = 4;
            *v_pat = 0;
        break;

        /* index based on expression: 100 * ABS(v - val1) / ABS(val2 - val1) */
        default:
            if ( v < 25 ) /* 1-24 */
            {
                *v1 = 0;
                *v2 = 1;
                *v_pat = v_pat_table[100 * v / 50];
            }
                    
            else if ( v < 50 ) /* 25-49 */
            {
                *v1 = 1;
                *v2 = 0;
                *v_pat = v_pat_table[100 * (50 - v) / 50];
            }
            else if ( v < 63 ) /* 51-62 */
            {
                *v1 = 1;
                *v2 = 2;
                *v_pat = v_pat_table[100 * (v - 50) / 25];
            }
            else if ( v < 75 ) /* 63-74 */
            {
                *v1 = 2;
                *v2 = 1;
                *v_pat = v_pat_table[100 * (75 - v) / 25];
            }
            else if ( v < 81 ) /* 76-80 */
            {
                *v1 = 2;
                *v2 = 3;
                *v_pat = v_pat_table[100 * (v - 75) / 12];
            }
            else if ( v < 87 ) /* 81-86 */
            {
                *v1 = 3;
                *v2 = 2;
                *v_pat = v_pat_table[100 * (87 - v) / 12];
            }
            else if ( v < 94 ) /* 88-93 */
            {
                *v1 = 3;
                *v2 = 4;
                *v_pat = v_pat_table[100 * (v - 87) / 13];
            }
            else /* 94-99 */
            {
                *v1 = 4;
                *v2 = 3;
                *v_pat = v_pat_table[100 * (100 - v) / 13];
            }
        break;
    }
}

find_nearest_r_5x5x5(r, r1, r2, r_pat,percent)
int r, *r1, *r2, *r_pat,percent;
{
    find_nearest_5x5x5(r, r1, r2, r_pat,percent );
    *r1 *= 25;
    *r2 *= 25;
}

find_nearest_g_5x5x5(g, g1, g2, g_pat,percent)
int g, *g1, *g2, *g_pat,percent;
{
    find_nearest_5x5x5(g, g1, g2, g_pat,percent );
    *g1 *= 5;
    *g2 *= 5;
}

find_nearest_b_5x5x5(b, b1, b2, b_pat,percent)
int b, *b1, *b2, *b_pat,percent;
{
    find_nearest_5x5x5(b, b1, b2, b_pat,percent );
}

/****************************************************************************/
find_nearest_4x4x4(v, v1, v2, v_pat,percent ) /* 0,50,75,100 */
int v, *v1, *v2, *v_pat;
int percent;
{
    switch ( v )
    {
        case 0:
            *v1 = *v2 = 0;
            *v_pat = 0;
        break;

        case 50:
            *v1 = *v2 = 1;
            *v_pat = 0;
        break;

        case 75:
            *v1 = *v2 = 2;
            *v_pat = 0;
        break;

        case 100:
            *v1 = *v2 = 3;
            *v_pat = 0;
        break;

        /* index based on expression: 100 * ABS(v - val1) / ABS(val2 - val1) */
        default:
            if ( v < 25 ) /* 1-24 */
            {
                *v1 = 0;
                *v2 = 1;
                *v_pat = v_pat_table[100 * v / 50];
            }
                    
            else if ( v < 50 ) /* 25-49 */
            {
                *v1 = 1;
                *v2 = 0;
                *v_pat = v_pat_table[100 * (50 - v) / 50];
            }
            else if ( v < 63 ) /* 51-62 */
            {
                *v1 = 1;
                *v2 = 2;
                *v_pat = v_pat_table[100 * (v - 50) / 25];
            }
            else if ( v < 75 ) /* 63-74 */
            {
                *v1 = 2;
                *v2 = 1;
                *v_pat = v_pat_table[100 * (75 - v) / 25];
            }
            else if ( v < 88 ) /* 76-87 */
            {
                *v1 = 2;
                *v2 = 3;
                *v_pat = v_pat_table[100 * (v - 75) / 25];
            }
            else /* 88-99 */
            {
                *v1 = 3;
                *v2 = 2;
                *v_pat = v_pat_table[100 * (100 - v) / 25];
            }
        break;
    }
}

find_nearest_r_4x4x4(r, r1, r2, r_pat,percent)
int r, *r1, *r2, *r_pat,percent;
{
    find_nearest_4x4x4(r, r1, r2, r_pat,percent );
    *r1 *= 16;
    *r2 *= 16;
}

find_nearest_g_4x4x4(g, g1, g2, g_pat,percent)
int g, *g1, *g2, *g_pat,percent;
{
    find_nearest_4x4x4(g, g1, g2, g_pat,percent );
    *g1 *= 4;
    *g2 *= 4;
}

find_nearest_b_4x4x4(b, b1, b2, b_pat,percent)
int b, *b1, *b2, *b_pat,percent;
{
    find_nearest_4x4x4(b, b1, b2, b_pat,percent );
}

/****************************************************************************/
find_nearest_3x3x3(v, v1, v2, v_pat,percent ) /* 0,50,100 */
int v, *v1, *v2, *v_pat;
int percent;
{
    switch ( v )
    {
        case 0:
            *v1 = *v2 = 0;
            *v_pat = 0;
        break;

        case 50:
            *v1 = *v2 = 1;
            *v_pat = 0;
        break;

        case 100:
            *v1 = *v2 = 2;
            *v_pat = 0;
        break;

        /* index based on expression: 100 * ABS(v - val1) / ABS(val2 - val1) */
        default:
            if ( v < 25 ) /* 1-24 */
            {
                *v1 = 0;
                *v2 = 1;
                *v_pat = v_pat_table[100 * v / 50];
            }
                    
            else if ( v < 50 ) /* 25-49 */
            {
                *v1 = 1;
                *v2 = 0;
                *v_pat = v_pat_table[100 * (50 - v) / 50];
            }
            else if ( v < 75 ) /* 51-74 */
            {
                *v1 = 1;
                *v2 = 2;
                *v_pat = v_pat_table[100 * (v - 50) / 50];
            }
            else /* 75-99 */
            {
                *v1 = 2;
                *v2 = 1;
                *v_pat = v_pat_table[100 * (100 - v) / 50];
            }
        break;
    }
}

find_nearest_r_3x3x3(r, r1, r2, r_pat,percent)
int r, *r1, *r2, *r_pat,percent;
{
    find_nearest_3x3x3(r, r1, r2, r_pat,percent );
    *r1 *= 9;
    *r2 *= 9;
}

find_nearest_g_3x3x3(g, g1, g2, g_pat,percent)
int g, *g1, *g2, *g_pat,percent;
{
    find_nearest_3x3x3(g, g1, g2, g_pat,percent );
    *g1 *= 3;
    *g2 *= 3;
}

find_nearest_b_3x3x3(b, b1, b2, b_pat,percent)
int b, *b1, *b2, *b_pat,percent;
{
    find_nearest_3x3x3(b, b1, b2, b_pat,percent );
}

/****************************************************************************/
find_nearest_2x2x2(v, v1, v2, v_pat,percent ) /* 0,100 */
int v, *v1, *v2, *v_pat;
int percent;
{
    switch ( v )
    {
        case 0:
            *v1 = *v2 = 0;
            *v_pat = 0;
        break;

        case 100:
            *v1 = *v2 = 1;
            *v_pat = 0;
        break;

        /* index based on expression: 100 * ABS(v - val1) / ABS(val2 - val1) */
        default:
            if ( v < 50 ) /* 1-49 */
            {
                *v1 = 0;
                *v2 = 1;
                *v_pat = v_pat_table[100 * v / 100];
            }
            else /* 51-99 */
            {
                *v1 = 1;
                *v2 = 0;
                *v_pat = v_pat_table[100 * (100 - v) / 100];
            }
        break;
    }
}

find_nearest_r_2x2x2(r, r1, r2, r_pat,percent)
int r, *r1, *r2, *r_pat,percent;
{
    find_nearest_2x2x2(r, r1, r2, r_pat,percent );
    *r1 *= 4;
    *r2 *= 4;
}

find_nearest_g_2x2x2(g, g1, g2, g_pat,percent )
int g, *g1, *g2, *g_pat;
{
    find_nearest_2x2x2(g, g1, g2, g_pat,percent );
    *g1 *= 2;
    *g2 *= 2;
}

find_nearest_b_2x2x2(b, b1, b2, b_pat,percent )
int b, *b1, *b2, *b_pat;
{
    find_nearest_2x2x2(b, b1, b2, b_pat,percent );
}

/****************************************************************************/
find_nearest_cmapindex(red,green,blue,col_tab,percent)
int red,green,blue;
col_table *col_tab;
int percent;
{
    int i, r,g,b, r1,g1,b1, r2,g2,b2, r_pat,g_pat,b_pat;
    int cindex;

    /* convert rgb to percentage values */
    r = 100 * red   / 255;
    g = 100 * green / 255;
    b = 100 * blue  / 255;

    switch ( AxNColorLevels )
    {
        case 2:
            /* get the nearest rgb values */
            find_nearest_r_2x2x2(r, &r1, &r2, &r_pat, percent);
            find_nearest_g_2x2x2(g, &g1, &g2, &g_pat, percent);
            find_nearest_b_2x2x2(b, &b1, &b2, &b_pat, percent);
        break;
        case 3:
            /* get the nearest rgb values */
            find_nearest_r_3x3x3(r, &r1, &r2, &r_pat, percent);
            find_nearest_g_3x3x3(g, &g1, &g2, &g_pat, percent);
            find_nearest_b_3x3x3(b, &b1, &b2, &b_pat, percent);
        break;
        case 4:
            /* get the nearest rgb values */
            find_nearest_r_4x4x4(r, &r1, &r2, &r_pat, percent);
            find_nearest_g_4x4x4(g, &g1, &g2, &g_pat, percent);
            find_nearest_b_4x4x4(b, &b1, &b2, &b_pat, percent);
        break;
        case 5:
            /* get the nearest rgb values */
            find_nearest_r_5x5x5(r, &r1, &r2, &r_pat, percent);
            find_nearest_g_5x5x5(g, &g1, &g2, &g_pat, percent);
            find_nearest_b_5x5x5(b, &b1, &b2, &b_pat, percent);
        break;
        case 6:
            /* get the nearest rgb values */
            find_nearest_r_6x6x6(r, &r1, &r2, &r_pat, percent);
            find_nearest_g_6x6x6(g, &g1, &g2, &g_pat, percent);
            find_nearest_b_6x6x6(b, &b1, &b2, &b_pat, percent);
        break;
        default:
            r1 = r2 = g1 = g2 = b1 = b2 = 0;
            r_pat = g_pat = b_pat = 0;
        break;
    }
    col_tab->rfg = r1;
    col_tab->rbg = r2;
    col_tab->rpat = r_pat;
    col_tab->gfg = g1;
    col_tab->gbg = g2;
    col_tab->gpat = g_pat;
    col_tab->bfg = b1;
    col_tab->bbg = b2;
    col_tab->bpat = b_pat;
    col_tab->fgcolor = AxPixels[(r1+g1+b1)];
    col_tab->bgcolor = AxPixels[(r2+g2+b2)];
}

/****************************************************************************/
int
white_pixel_from_table()
{
    col_table col_tab;
    find_nearest_cmapindex(255,255,255, &col_tab, FALSE);
    return(col_tab.fgcolor);
}

/****************************************************************************/
find_percent_6x6x6(v, v1, v2, v_pat,percent) /* 0,50,75,87,95,100 */
int v, *v1, *v2, *v_pat;
int percent;
{
    switch ( v )
    {
        case 0:
            *v1 = *v2 = 0;
            *v_pat = 0;
        break;

        case 50:
            *v1 = *v2 = 50;
            *v_pat = 0;
        break;

        case 75:
            *v1 = *v2 = 75;
            *v_pat = 0;
        break;

        case 87:
            *v1 = *v2 = 87;
            *v_pat = 0;
        break;

        case 95:
            *v1 = *v2 = 95;
            *v_pat = 0;
        break;

        case 100:
            *v1 = *v2 = 100;
            *v_pat = 0;
        break;

        /* index based on expression: 100 * ABS(v - val1) / ABS(val2 - val1) */
        default:
            if ( v < 25 ) /* 1-24 */
            {
                *v1 = 0;
                *v2 = 50;
                *v_pat = 0;
            }
                    
            else if ( v < 50 ) /* 25-49 */
            {
                *v1 = 50;
                *v2 = 0;
                *v_pat = 0;
            }
            else if ( v < 63 ) /* 51-62 */
            {
                *v1 = 50;
                *v2 = 75;
                *v_pat = 0;
            }
            else if ( v < 75 ) /* 63-74 */
            {
                *v1 = 75;
                *v2 = 50;
                *v_pat = 0;
            }
            else if ( v < 81 ) /* 76-80 */
            {
                *v1 = 75;
                *v2 = 87;
                *v_pat = 0;
            }
            else if ( v < 87 ) /* 81-86 */
            {
                *v1 = 87;
                *v2 = 75;
                *v_pat = 0;
            }
            else if ( v < 91 ) /* 88-90 */
            {
                *v1 = 87;
                *v2 = 95;
                *v_pat = 0;
            }
            else if ( v < 95 ) /* 91-94 */
            {
                *v1 = 95;
                *v2 = 87;
                *v_pat = 0;
            }
            else if ( v < 98 ) /* 96-97 */
            {
                *v1 = 95;
                *v2 = 100;
                *v_pat = 0;
            }
            else /* 98-99 */
            {
                *v1 = 100;
                *v2 = 95;
                *v_pat = 0;
            }
        break;
    }
}

find_nearest_rp_6x6x6(r, r1, r2, r_pat)
int r, *r1, *r2, *r_pat;
{
    find_percent_6x6x6(r, r1, r2, r_pat );
}

find_nearest_gp_6x6x6(g, g1, g2, g_pat)
int g, *g1, *g2, *g_pat;
{
    find_percent_6x6x6(g, g1, g2, g_pat );
}

find_nearest_bp_6x6x6(b, b1, b2, b_pat)
int b, *b1, *b2, *b_pat;
{
    find_percent_6x6x6(b, b1, b2, b_pat );
}

/****************************************************************************/
find_percent_5x5x5(v, v1, v2, v_pat) /* 0,50,75,87,100 */
int v, *v1, *v2, *v_pat;
{
    switch ( v )
    {
        case 0:
            *v1 = *v2 = 0;
            *v_pat = 0;
        break;

        case 50:
            *v1 = *v2 = 50;
            *v_pat = 0;
        break;

        case 75:
            *v1 = *v2 = 75;
            *v_pat = 0;
        break;

        case 87:
            *v1 = *v2 = 87;
            *v_pat = 0;
        break;

        case 100:
            *v1 = *v2 = 100;
            *v_pat = 0;
        break;

        /* index based on expression: 100 * ABS(v - val1) / ABS(val2 - val1) */
        default:
            if ( v < 25 ) /* 1-24 */
            {
                *v1 = 0;
                *v2 = 50;
                *v_pat = 0;
            }
                    
            else if ( v < 50 ) /* 25-49 */
            {
                *v1 = 50;
                *v2 = 0;
                *v_pat = 0;
            }
            else if ( v < 63 ) /* 51-62 */
            {
                *v1 = 50;
                *v2 = 75;
                *v_pat = 0;
            }
            else if ( v < 75 ) /* 63-74 */
            {
                *v1 = 75;
                *v2 = 50;
                *v_pat = 0;
            }
            else if ( v < 81 ) /* 76-80 */
            {
                *v1 = 75;
                *v2 = 87;
                *v_pat = 0;
            }
            else if ( v < 87 ) /* 81-86 */
            {
                *v1 = 87;
                *v2 = 75;
                *v_pat = 0;
            }
            else if ( v < 94 ) /* 88-93 */
            {
                *v1 = 87;
                *v2 = 100;
                *v_pat = 0;
            }
            else /* 94-99 */
            {
                *v1 = 100;
                *v2 = 87;
                *v_pat = 0;
            }
        break;
    }
}

find_nearest_rp_5x5x5(r, r1, r2, r_pat)
int r, *r1, *r2, *r_pat;
{
    find_percent_5x5x5(r, r1, r2, r_pat);
}

find_nearest_gp_5x5x5(g, g1, g2, g_pat)
int g, *g1, *g2, *g_pat;
{
    find_percent_5x5x5(g, g1, g2, g_pat);
}

find_nearest_bp_5x5x5(b, b1, b2, b_pat)
int b, *b1, *b2, *b_pat;
{
    find_percent_5x5x5(b, b1, b2, b_pat);
}

/****************************************************************************/
find_percent_4x4x4(v, v1, v2, v_pat) /* 0,50,75,100 */
int v, *v1, *v2, *v_pat;
{
    switch ( v )
    {
        case 0:
            *v1 = *v2 = 0;
            *v_pat = 0;
        break;

        case 50:
            *v1 = *v2 = 50;
            *v_pat = 0;
        break;

        case 75:
            *v1 = *v2 = 75;
            *v_pat = 0;
        break;

        case 100:
            *v1 = *v2 = 100;
            *v_pat = 0;
        break;

        /* index based on expression: 100 * ABS(v - val1) / ABS(val2 - val1) */
        default:
            if ( v < 25 ) /* 1-24 */
            {
                *v1 = 0;
                *v2 = 50;
                *v_pat = 0;
            }
                    
            else if ( v < 50 ) /* 25-49 */
            {
                *v1 = 50;
                *v2 = 0;
                *v_pat = 0;
            }
            else if ( v < 63 ) /* 51-62 */
            {
                *v1 = 50;
                *v2 = 75;
                *v_pat = 0;
            }
            else if ( v < 75 ) /* 63-74 */
            {
                *v1 = 75;
                *v2 = 50;
                *v_pat = 0;
            }
            else if ( v < 88 ) /* 76-87 */
            {
                *v1 = 75;
                *v2 = 100;
                *v_pat = 0;
            }
            else /* 88-99 */
            {
                *v1 = 100;
                *v2 = 75;
                *v_pat = 0;
            }
        break;
    }
}

find_nearest_rp_4x4x4(r, r1, r2, r_pat)
int r, *r1, *r2, *r_pat;
{
    find_percent_4x4x4(r, r1, r2, r_pat );
}

find_nearest_gp_4x4x4(g, g1, g2, g_pat)
int g, *g1, *g2, *g_pat;
{
    find_percent_4x4x4(g, g1, g2, g_pat );
}

find_nearest_bp_4x4x4(b, b1, b2, b_pat)
int b, *b1, *b2, *b_pat;
{
    find_percent_4x4x4(b, b1, b2, b_pat );
}

/****************************************************************************/
find_percent_3x3x3(v, v1, v2, v_pat) /* 0,50,100 */
int v, *v1, *v2, *v_pat;
{
    switch ( v )
    {
        case 0:
            *v1 = *v2 = 0;
            *v_pat = 0;
        break;

        case 50:
            *v1 = *v2 = 50;
            *v_pat = 0;
        break;

        case 100:
            *v1 = *v2 = 100;
            *v_pat = 0;
        break;

        /* index based on expression: 100 * ABS(v - val1) / ABS(val2 - val1) */
        default:
            if ( v < 25 ) /* 1-24 */
            {
                *v1 = 0;
                *v2 = 50;
                *v_pat = 0;
            }
                    
            else if ( v < 50 ) /* 25-49 */
            {
                *v1 = 50;
                *v2 = 0;
                *v_pat = 0;
            }
            else if ( v < 75 ) /* 51-74 */
            {
                *v1 = 50;
                *v2 = 100;
                *v_pat = 0;
            }
            else /* 75-99 */
            {
                *v1 = 100;
                *v2 = 50;
                *v_pat = 0;
            }
        break;
    }
}

find_nearest_rp_3x3x3(r, r1, r2, r_pat)
int r, *r1, *r2, *r_pat;
{
    find_percent_3x3x3(r, r1, r2, r_pat );
}

find_nearest_gp_3x3x3(g, g1, g2, g_pat)
int g, *g1, *g2, *g_pat;
{
    find_percent_3x3x3(g, g1, g2, g_pat);
}

find_nearest_bp_3x3x3(b, b1, b2, b_pat)
int b, *b1, *b2, *b_pat;
{
    find_percent_3x3x3(b, b1, b2, b_pat);
}

/****************************************************************************/
find_percent_2x2x2(v, v1, v2, v_pat) /* 0,100 */
int v, *v1, *v2, *v_pat;
{
    switch ( v )
    {
        case 0:
            *v1 = *v2 = 0;
            *v_pat = 0;
        break;

        case 100:
            *v1 = *v2 = 100;
            *v_pat = 0;
        break;

        /* index based on expression: 100 * ABS(v - val1) / ABS(val2 - val1) */
        default:
            if ( v < 50 ) /* 1-49 */
            {
                *v1 = 0;
                *v2 = 100;
                *v_pat = 0;
            }
            else /* 51-99 */
            {
                *v1 = 100;
                *v2 = 0;
                *v_pat = 0;
            }
        break;
    }
}

find_nearest_rp_2x2x2(r, r1, r2, r_pat)
int r, *r1, *r2, *r_pat;
{
    find_percent_2x2x2(r, r1, r2, r_pat );
}

find_nearest_gp_2x2x2(g, g1, g2, g_pat )
int g, *g1, *g2, *g_pat;
{
    find_percent_2x2x2(g, g1, g2, g_pat );
}

find_nearest_bp_2x2x2(b, b1, b2, b_pat)
int b, *b1, *b2, *b_pat;
{
    find_percent_2x2x2(b, b1, b2, b_pat);
}

/****************************************************************************/
find_nearest_percent(red,green,blue,col_tab)
int red,green,blue;
col_table *col_tab;
{
    int i, r,g,b, r1,g1,b1, r2,g2,b2, r_pat,g_pat,b_pat;

    /* convert rgb to percentage values */
    r = 100 * red   / 255;
    g = 100 * green / 255;
    b = 100 * blue  / 255;

    switch ( AxNColorLevels )
    {
        case 2:
            /* get the nearest rgb values */
            find_nearest_rp_2x2x2(r, &r1, &r2, &r_pat);
            find_nearest_gp_2x2x2(g, &g1, &g2, &g_pat);
            find_nearest_bp_2x2x2(b, &b1, &b2, &b_pat);
        break;
        case 3:
            /* get the nearest rgb values */
            find_nearest_rp_3x3x3(r, &r1, &r2, &r_pat);
            find_nearest_gp_3x3x3(g, &g1, &g2, &g_pat);
            find_nearest_bp_3x3x3(b, &b1, &b2, &b_pat);
        break;
        case 4:
            /* get the nearest rgb values */
            find_nearest_rp_4x4x4(r, &r1, &r2, &r_pat);
            find_nearest_gp_4x4x4(g, &g1, &g2, &g_pat);
            find_nearest_bp_4x4x4(b, &b1, &b2, &b_pat);
        break;
        case 5:
            /* get the nearest rgb values */
            find_nearest_rp_5x5x5(r, &r1, &r2, &r_pat);
            find_nearest_gp_5x5x5(g, &g1, &g2, &g_pat);
            find_nearest_bp_5x5x5(b, &b1, &b2, &b_pat);
        break;
        case 6:
            /* get the nearest rgb values */
            find_nearest_rp_6x6x6(r, &r1, &r2, &r_pat);
            find_nearest_gp_6x6x6(g, &g1, &g2, &g_pat);
            find_nearest_bp_6x6x6(b, &b1, &b2, &b_pat);
        break;
        default:
            r1 = r2 = g1 = g2 = b1 = b2 = 0;
            r_pat = g_pat = b_pat = 0;
        break;
    }

    r1 = (r1 == r2) ? red   : r1 * 255 / 100.0;
    g1 = (g1 == g2) ? green : g1 * 255 / 100.0;
    b1 = (b1 == b2) ? blue  : b1 * 255 / 100.0;

    col_tab->rfg = r1;
    col_tab->gfg = g1;
    col_tab->bfg = b1;
}

/****************************************************************************/
int
getWidgetRealPixel (wcolor, pixel)
int wcolor;
unsigned long *pixel;
{
    int r, g, b;

    if (wcolor & 0x80000000) {
        r = (wcolor & 0x000000ff);
        g = (wcolor & 0x0000ff00) >> 8;
        b = (wcolor & 0x00ff0000) >> 16;
        AxGetRealPixel(0, r, g, b, 0, 255, pixel);
        return (TRUE);
    }
    return (FALSE);
}

/****************************************************************************/
axConvertStdIndexToPixel(col_tab, cmap)
col_table *col_tab;
cmap_type *cmap;
{
    int i, k, r, g, b;
    color_type *tmp_color;
    double val;

    switch ( axXA_RGB_APPLIX_MAP ) 
    {
        default:
            for ( i=0; i<cmap->n_colors && i<256; i++ ) 
            {
                memset((char *)&col_tab[i], 0, sizeof(col_table));
                tmp_color = &cmap->color[i];
                if ( tmp_color->see_thru == TRUE )
                    col_tab[i].rfg = 255;
                else 
                {
                    /* Convert cmyk to rgb */
                    k = tmp_color->k;
                    r = 255 - MIN(255, tmp_color->c + k);
                    g = 255 - MIN(255, tmp_color->m + k);
                    b = 255 - MIN(255, tmp_color->y + k);
                    col_tab[i].rfg = AxStdCmapInfo->base_pixel;
                    if ( r == g && g == b )
                    {
                        val = (double)r / (double)255;
                        col_tab[i].rfg += ((unsigned long)
                                          (val*AxStdCmapInfo->red_max)*
                                          AxStdCmapInfo->red_mult);
                        col_tab[i].gfg = ((unsigned long)
                                          (val*AxStdCmapInfo->green_max)*
                                          AxStdCmapInfo->green_mult);
                        col_tab[i].bfg = ((unsigned long)
                                          (val*AxStdCmapInfo->blue_max) *
                                          AxStdCmapInfo->blue_mult);
                    }
                    else
                    {
                        col_tab[i].rfg += RSTDCOMP(AxStdCmapInfo, r, 255);
                        col_tab[i].gfg  = GSTDCOMP(AxStdCmapInfo, g, 255);
                        col_tab[i].bfg  = BSTDCOMP(AxStdCmapInfo, b, 255);
                    }
    
                }
            }   
        break;

        case XA_RGB_RED_MAP:
            for ( i=0; i<cmap->n_colors && i<256; i++ ) 
            {
                memset((char *)&col_tab[i], 0, sizeof(col_table));
                tmp_color = &cmap->color[i];
                if ( tmp_color->see_thru == TRUE )
                    col_tab[i].rfg = 255;
                else 
                {
                    /* Convert cmyk to rgb */
                    k = tmp_color->k;
                    r = 255 - MIN(255, tmp_color->c + k);
                    g = 255 - MIN(255, tmp_color->m + k);
                    b = 255 - MIN(255, tmp_color->y + k);
                    r = axConvertStdToGray(r, g, b, 255);
                    col_tab[i].rfg = AxStdCmapInfo->base_pixel;
                    col_tab[i].rfg += (int)(0.5+((float)(r)/255)*
                                               AxStdCmapInfo->red_max)*
                                               AxStdCmapInfo->red_mult;
                }
            }   
        break;
        case XA_RGB_GREEN_MAP:
            /* create an index to pixel table */
            for ( i=0; i<cmap->n_colors && i<256; i++ ) 
            {
                memset((char *)&col_tab[i], 0, sizeof(col_table));
                tmp_color = &cmap->color[i];
                if ( tmp_color->see_thru == TRUE )
                    col_tab[i].gfg = 255;
                else 
                {
                    /* Convert cmyk to rgb */
                    k = tmp_color->k;
                    r = 255 - MIN(255, tmp_color->c + k);
                    g = 255 - MIN(255, tmp_color->m + k);
                    b = 255 - MIN(255, tmp_color->y + k);
                    g = axConvertStdToGray(r, g, b, 255);
                    col_tab[i].gfg = AxStdCmapInfo->base_pixel;
                    col_tab[i].gfg += (int)(0.5+((float)(g)/255)*
                                               AxStdCmapInfo->green_max)*
                                               AxStdCmapInfo->green_mult;
                }
            }   
        break;
            case XA_RGB_BLUE_MAP:
            /* create an index to pixel table */
            for ( i=0; i<cmap->n_colors && i<256; i++ ) 
            {
                memset((char *)&col_tab[i], 0, sizeof(col_table));
                tmp_color = &cmap->color[i];
                if ( tmp_color->see_thru == TRUE )
                    col_tab[i].bfg = 255;
                else 
                {
                    /* Convert cmyk to rgb */
                    k = tmp_color->k;
                    r = 255 - MIN(255, tmp_color->c + k);
                    g = 255 - MIN(255, tmp_color->m + k);
                    b = 255 - MIN(255, tmp_color->y + k);
                    b = axConvertStdToGray(r, g, b, 255);
                    col_tab[i].bfg = AxStdCmapInfo->base_pixel;
                    col_tab[i].bfg += (int)(0.5+((float)(b)/255)*
                                               AxStdCmapInfo->blue_max)*
                                               AxStdCmapInfo->blue_mult;
                }
            }   
        break;
    }
}

/****************************************************************************/
