/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ddlfilter.c
   Applix ASC version: 2
   Date: Mon Nov 16 13:53:41 1998
*/
#include <stdio.h>
#include <DrgDrpLibP.h>
#include <X11/Xatom.h>


void DrgDrpSetupFilter (dpy, context, callback, closure)
Display *dpy;
DrgDrpContext *context;
void (*callback)();
void *closure;
{
    context->callback = callback;
    context->closure = closure;
    context->events = NULL;
}


int DrgDrpFilterEvent (e, context)
XEvent *e;
DrgDrpContext *context;
{
    unsigned int n_events, n_formats, n_styles, i;
    DrgDrpInfo *format;
    DrgDrpStyle *style;
    struct DrgDrp_Event *holder;
    struct DrgDrp_Event **pptr;
    struct DrgDrp_Event *ptr;

    if (e->type != ClientMessage || e->xclient.format != 32 ||
        e->xclient.message_type != DrgDrp_Atom (e->xclient.display))
        return True;

    /* Save the event at the end of the list. If it's not the last, return. */

    holder = (struct DrgDrp_Event *) malloc (sizeof (struct DrgDrp_Event));
    if (holder == NULL)
        return False;
    holder->next = NULL;
    holder->e = e->xclient;
    if (context->events == NULL)
        context->events = holder;
    else
        for (ptr = context->events; ; ptr = ptr->next)
            if (ptr->next == NULL)
            {
                ptr->next = holder;
                break;
            }

    if (e->xclient.data.l [2] & F_MORE)
        return False;

    /* Walk the event list, counting all the events of the sequence. */

    n_events = 0;
    ptr = context->events;
    while (ptr != NULL)
    {
        if (ptr->e.window == e->xclient.window &&
            ptr->e.data.l [0] == e->xclient.data.l [0] &&
            ptr->e.data.l [1] == e->xclient.data.l [1])
            n_events++;
        ptr = ptr->next;
    }

    /* Walk the event list, processing the events of the sequence. */

    format = (DrgDrpInfo *) malloc (n_events * sizeof (DrgDrpInfo));
    n_formats = 0;
    style = (DrgDrpStyle *) malloc (n_events * sizeof (DrgDrpStyle));
    n_styles = 0;
    for (pptr = &(context->events); ptr = *pptr, ptr != NULL; )
    {
        if (ptr->e.window != e->xclient.window ||
            ptr->e.data.l [0] != e->xclient.data.l [0] ||
            ptr->e.data.l [1] != e->xclient.data.l [1])
        {
            continue;
#if 0 /* NOTREACHED*/
            pptr = (struct DrgDrp_Event **) &(ptr->next);
#endif
        }

        /* Transfer the format. */

        if (ptr->e.data.l [3] != None)
        {
            unsigned long after;
            unsigned char *data;

            XGetWindowProperty (e->xclient.display, ptr->e.window,
                                ptr->e.data.l [4], 0L, 1048576L, True,
                                AnyPropertyType, &(format [n_formats].format),
                                &(format [n_formats].wordsize),
                                &(format [n_formats].length), &after, &data);
            format [n_formats].data = data;
            n_formats++;
        }

        /* Transfer the style info. */

        {
            Bool known = False;

            for (i = 0; i < n_styles; i++)
                if (ptr->e.data.l [2] == style [i].internal)
                    known = True;
            if (!known)
            {
                DrgDrpStyle *s = style + n_styles++;
                s->internal = ptr->e.data.l [2];
                switch (s->internal & F_STYLE)
                {
                case F_S_DEVICE:
                    s->type = DrgDrpDeviceStyle;
                    s->device_state = s->internal & F_DEVICE;
                    break;

                case F_S_LOGICAL:
                    s->type = DrgDrpLogicalStyle;
                    s->drop_type = s->internal & F_DATA;
                    break;

                default:
                    n_styles--;
                    break;
                }
            }
        }

        /* Delete the event from the list. */

        *pptr = ptr->next;
        free (ptr);
    }

    /* Now call the callback routine. */

    (context->callback)(context->closure, e->xclient.display,
        e->xclient.window, e->xclient.data.l [0],
        (e->xclient.data.l [1] >> 16) & 0xFFFF,
        (e->xclient.data.l [1]      ) & 0xFFFF,
        style, n_styles, format, n_formats);

    /* Finally, destroy all the data allocated here. */

    for (i = 0; i < n_formats; i++)
        XFree (format [i].data);
    free (format);
    free (style);
    return False;
}
