/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ddlget.c
   Applix ASC version: 2
   Date: Mon Nov 16 13:53:41 1998
*/
#include <stdio.h>
#include <DrgDrpLibP.h>
#include <X11/Xatom.h>


int DrgDrpGetTarget (dpy, win, x, y, n_formats, known, target, t_win)
Display *dpy;
Window win;
unsigned int x, y, n_formats;
Atom *known;
Atom *target;
Window *t_win;
{
    Window parent = None;

    /* Set parent to the correct root window. */
    {
        int x, y;
        unsigned int w, h, b, d;
        unsigned int s;

        for (s = 0; s < ScreenCount (dpy); s++)
            if (win == RootWindow (dpy, s))
            {
                parent = win;
                break;
            }

        if (parent == None &&
            !XGetGeometry (dpy, win, &parent, &x, &y, &w, &h, &b, &d))
            return -1;
    }

    /* Walk down the window tree, looking for the magic property. */

    for (;;)
    {
        Window child;
        int xx, yy;
        Atom type;
        int format;
        unsigned long n_items, after;
        unsigned char *data;

        if (!XTranslateCoordinates (dpy, win, parent, x, y, &xx, &yy, &child))
            return -2; /* We've somehow crossed screens. */
        if (child == None)
        {
            /* Didn't find the magic property. */

            *t_win = None;
            return 0;
        }
        x = xx; y = yy;
        XGetWindowProperty (dpy, child, DrgDrp_Atom (dpy), 0L, 1048576L,
                            False, XA_WINDOW, &type, &format, &n_items,
                            &after, &data);
        if (type == XA_WINDOW && format == 32 && n_items != 0)
        {
            Atom *atom = (Atom *) data + 1;

            *t_win = *(Window *)data;
            n_items--;
            if (known == NULL)
            {
                /* Copy up to n_formats entries and return. */

                unsigned int i;

                if (n_items > n_formats)
                    n_items = n_formats;
                for (i = 0; i < n_formats; i++)
                    target [i] = atom [i];
                return (n_items);
            }
            else
            {
                /* Copy entries that match those in known, and return.  */

                unsigned int i, j, k;

                for (i = j = 0; j < n_formats; j++)
                    for (k = 0; k < n_items; k++)
                        if (known [j] == atom [k])
                        {
                            target [i++] = atom [k];
                            break;
                        }
                return (i);
            }
        }
        win = parent;
        parent = child;
    }

    /* NOTREACHED */
}
