/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: editmask.C
   Applix ASC version: 3
   Date: Mon Nov 16 13:53:48 1998
*/
#include <stdlib.h>
#include <aos.h>
#include <AxEditMask.h>
#include <editmask.h>


// create a new edit mask.  axmain code will refer to the mask as void*
//
void *NewEditMask(char *inmask) {
	return ((void *) new AxEditMask(inmask));
}

void DeleteEditMask(void *mask) {
	if (mask) {
		delete (AxEditMask *) mask;
	}
}

// return a string representing the initial ("empty") value of
// an editmask-constrained text field.  Caller must free() returned string.
//
wchar *EditMaskInitialValue(void *mask) {
	AxEditMask *em = (AxEditMask *) mask;
	char *initval = em->initial_value();
	int masklen = em->mask_length();
	if (0 != initval && masklen > 0) {
		wchar *ret = (wchar *) malloc (masklen + 1);  // include a null
		int ix;
		for (ix = 0; ix < masklen; ix++) {
			ret[ix] = (wchar) initval[ix];
		}
		ret[masklen] = 0;
		delete initval;
		return ret;
	}
	else {
		return 0;
	}
}

// return just the "keystrokes" that produced a given masked-field's value
// caller must free() returned string
//
wchar *EditMaskKeystrokes(char *inmask, wchar *current_value) {
    void *em = NewEditMask(inmask);
	wchar *ival = EditMaskInitialValue(em);
	// "subtract" the initial value from the current value
	//
	int ix, nout = 0;
	int ival_len = EditMaskLength(em);
	for (ix = 0; ix < ival_len; ix++) {  // first get length of "difference"
		if (current_value[ix] != ival[ix]) {
			nout++;
		}
	}
	wchar *ret = (wchar *) malloc (nout + 1);
	for (nout = 0, ix = 0; ix < ival_len; ix++) {  // redo, tranfer this time
		if (current_value[ix] != ival[ix]) {
			ret[nout++] = current_value[ix];
		}
	}
	ret[nout] = 0;
	delete em;
	return (ret);
}

int EditMaskDiffers(void *mask, char *inmask) {
	AxEditMask *em = (AxEditMask *) mask;
	return em->mask_differs(inmask);
}

// Return the length of the edit mask -- this length does _not_ include
// a trailing null character, it's just the length of the mask itself.
//
int EditMaskLength(void *mask) {
	AxEditMask *em = (AxEditMask *) mask;
	return em->mask_length();
}

int EditMaskNextHigherIndex(void *mask, int curdex) {
	AxEditMask *em = (AxEditMask *) mask;
	return em->increment_cursor_position(curdex);
}

int EditMaskNextLowerIndex(void *mask, int curdex) {
	AxEditMask *em = (AxEditMask *) mask;
	return em->decrement_cursor_position(curdex);
}

int EditMaskSmallestIndex(void *mask) {
	AxEditMask *em = (AxEditMask *) mask;
	return em->home_cursor();
}

int EditMaskHighestIndex(void *mask) {
	AxEditMask *em = (AxEditMask *) mask;
	return em->end_cursor();
}

void EditMaskDeleteChars(void *mask, wchar *curstring, int dex, int len) {
	AxEditMask *em = (AxEditMask *) mask;
	int masklen = em->mask_length();
	if (masklen > 0) {
		int ix;
		char *curchars = new char[masklen];
		for (ix = 0; ix < masklen; ix++) {
			curchars[ix] = (char) curstring[ix];
		}
		em->deleteChars(curchars, dex, len);
		for (ix = 0; ix < masklen; ix++) {
			curstring[ix] = (wchar) curchars[ix];
		}
		delete curchars;
	}
}

void EditMaskOverwriteChars(void *mask, wchar *curstring, wchar *overstring,
							int overlen, int dex) {
	AxEditMask *em = (AxEditMask *) mask;
	int masklen = em->mask_length();
	if (masklen > 0 && overlen > 0) {
		int ix;
		char *curchars = new char[masklen];
		char *overchars = new char[overlen];
		for (ix = 0; ix < masklen; ix++) {
			curchars[ix] = (char) curstring[ix];
		}
		for (ix = 0; ix < overlen; ix++) {
			overchars[ix] = (char) overstring[ix];
		}
		em->overwriteChars(curchars, overchars, overlen, dex);
		for (ix = 0; ix < masklen; ix++) {
			curstring[ix] = (wchar) curchars[ix];
		}
		delete curchars;
		delete overchars;
	}
}

int EditMaskTestChars(void *mask, wchar *curstring, wchar *overstring,
					   int overlen, int dex) {
	AxEditMask *em = (AxEditMask *) mask;
	int masklen = em->mask_length();
	int num_accepted = 0;
	if (masklen > 0 && overlen > 0) {
		int ix;
		char *curchars = new char[masklen];
		char *overchars = new char[overlen];
		for (ix = 0; ix < masklen; ix++) {
			curchars[ix] = (char) curstring[ix];
		}
		for (ix = 0; ix < overlen; ix++) {
			overchars[ix] = (char) overstring[ix];
		}
		num_accepted = em->testChars(curchars, overchars, overlen, dex);
		delete curchars;
		delete overchars;
	}
	return num_accepted;
}

