/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: editwin.c
   Applix ASC version: 30
   Date: Tue Jun  1 21:12:11 1999
*/
#include <math.h>
#include <basicwin.h>
#if AX_X11
#include <X11/keysym.h>
#endif
#include <ax_xm.h>
#include <aos.h>
#include <dwt2xm.h>
#include <xmlib.h>
#include <elxwins.h>
#include <wchars.h>
#include <thimble.h>
#include <ctype.h>
#include <elf_env.h>
#include <eldata.h>
#include <evreason.h>
#include <xmpixel.h>
#include <intsarr.h>
#include <editmask.h>
#include <tmkeys.h>

typedef enum
{
  HIGHLIGHT,
  UNHIGHLIGHT
} highlight; /* also defined in xmcombo.c - should be in xwidgets.h? */
    
#include "editwin.p"

/* Standard Externs for Widget Modules */
extern Window AxCrtWindow();
extern AXWidget CrtWidget();
extern pixcache_t *Pixmapper();
extern void AxDrawBevel();
extern void xmSetWidgetDisplay();
extern XFontStruct *lookupNamedFont();
extern XCharStruct *getCharStruct();

extern int AxLanguage;
extern int WidgetNeedsUpdated; /* Set when a widget needs repainted */
extern int AxCharAbove, AxCharHeight, AxCharBelow;
extern int AxOpenLook;
extern int AxBigFont;
extern int UseWidgetColors;
extern Display *Dpy;
extern short DpyDepth;
extern int AxButtonDown;
extern short THIMpid;

extern long WinMask;
extern XSetWindowAttributes WinAtt;
extern Pixmap UnknownPixmap;
extern AXWidget ClickWidget;
extern AXWidget_menu AxPulldownActive;
extern unsigned long BackWorkPixel;
extern Colormap WorkColormap;
extern GC ForeGC, BackGC;
extern GC SunGC, ShadowGC, DisabledGC, GrayLineGC;
extern GC RecessGC;
extern GC olClrSunButtonGC, olClrShadowButtonGC, olClrRecessButtonGC;
extern GC MonospaceGC;		 /* monspace system font, forground mode */
extern GC MonospaceBackGC;	 /* monspace system font, background mode */
extern GC MonospaceBoldGC, MonospaceBoldBackGC;
extern GC MonospaceItalGC, MonospaceItalBackGC;
extern GC MonospaceItalBoldGC, MonospaceItalBoldBackGC;
extern XFontStruct *AxDlgMainFont;	/* Bolded dialog font */
extern XFontStruct *MonospaceFont, *MonospaceBoldFont;
extern XFontStruct *MonospaceItalFont, *MonospaceItalBoldFont;
extern int DoubleMillisec;
extern unsigned long ForePixel;
extern unsigned long BackPixel;
extern short ColorDisplay;
extern int ThimExposuresMayBeComing;
extern short DpyWid, DpyHt;
extern short NoColor;
extern Window RootWin;
extern Cursor ThimCrosshairCursor;
extern Cursor ThimPointerCursor;
extern Cursor ThimLeftRightCursor;
extern Cursor ThimTextCursor;
extern unsigned long AlisPixels[];
extern unsigned long WidgetPixels[];
extern Window AxDummyWin;       /* un-mapped root window */
extern AXWidget_combo_box ComboBoxInEditWindow; /* combo box w/ cursor on */
extern int AxPulldownStayup;
extern int AxExitPulldownStayupMode;
extern AXWidget_scroll ScrollBarArmed; /* currently armed W_SCROLL */
extern void AxAutoScrollHandler(); /* xmlib.c */

/* End of Standard Externs */

extern Atom AxXA_COMBOBOX_PROP;
extern Atom AxXA_ENTER_PROP;

static unsigned int last_click_time = 0;
static int last_button_clicked = 0;

#define MAX_XSTR 8000  /* length of seln to be inserted - same as in ATM */
/* IS_NGC adapted from ATM's tmin1.c definition */
#define BS ((unsigned) 0010)
#define DEL ((unsigned) 0177)  /* ascii code */
#define IS_NGC(c) ((((unsigned) (c)) < (unsigned) ' ') || \
                   (((unsigned) (c)) == DEL) || \
		   (((unsigned) (c)) == BS))

#if !BUILD_WIDE /* move to xmlib.h!! */
#define axDrawString16 axDrawString
#endif

#define BORDER_WIDTH 2 /* generic shadow borders */
#define TEXT_OFFSET  2 /* offset of text from border edge */
#define ENTRY_VOFFSET 1 /* vertical offset for entry field text */

/* to match arrow sizes in xmarrow.c */
#define down1_width 16
#define down1_height 16

#define SCROLLBAR_THICK 16 /* FIXME, from xmlibcom.c */

/* autoscroll vars EditWinArmed, editwinAutoscrollEdge
 */
AXWidget EditWinArmed;  /* the widget whose editwin is autoscrolling */
enum editwinEdge {
	noEdge = 0, leftEdge = 1, topEdge = 2, rightEdge = 3, bottomEdge = 4,
};
static enum editwinEdge editwinAutoscrollEdge;

/**************************************************************************
  function: EditWinIsPrintable - tells if string wbuf is printable
*/
	int
	EditWinIsPrintable(w, wbuf, num_chars)
    AXWidget w;
    wchar *wbuf;
    int num_chars;
{
    register int ix;
    register wchar *wbufp;
    int real_len;

    if (!w || ((w->type != W_COMBO_BOX) && (w->type != W_ENTER)))
		return(FALSE);

    for (ix = 0, wbufp = wbuf, real_len = 0; ix < num_chars; ix++) {
		if (!IS_NGC(*wbufp)) {
			/* char is printable */
			if (w->type == W_ENTER)
				if (((AXWidget_entry)w)->validChars &&
					!MbWstrchr(((AXWidget_entry)w)->validChars, *wbufp)) {
					/* invalid char */
					RingTheBell();
					return(FALSE);
				}
	    
			wbufp++;
			real_len++;
		}
    }
    if ((((unsigned) *wbuf) != DEL) && (((unsigned) *wbuf) != BS) && !real_len)
		return(FALSE);

    return(TRUE);
}

/**************************************************************************
  EditWinWidthHeight() - returns width and height of edit window
*/
void EditWinWidthHeight(w, editWidth, editHeight)
    AXWidget w;   /* in */
    int *editWidth;          /* out */
    int *editHeight;         /* out */
{
    if (w->type == W_COMBO_BOX) {
		*editWidth = w->width - (down1_width +
								 (4 * BORDER_WIDTH));
		*editHeight = w->height - (2 * BORDER_WIDTH);
    } else if (w->type == W_ENTER || w->type == W_KANJI_KANA) {
		*editWidth = w->width - (2 * BORDER_WIDTH);
		*editHeight = w->height - (2 * BORDER_WIDTH);
    }
}

void EditWinManageButtonPress(w, edit_win, eventP)
    AXWidget w;
    struct edit_window *edit_win;
    XEvent *eventP;
{
    seln_t *sel;
    int button_pressed;
    int x, y;

    if (!eventP)
		return;
    
    x = eventP->xbutton.x;
    y = eventP->xbutton.y;

    if ((w->type == W_COMBO_BOX) && (w->flags & WFLG_EDITABLE)) {
		/* we set ComboBoxInEditWindow here; if the button is released outside
		   of the edit window, then it will be reset to NULL */
		ComboBoxInEditWindow = (AXWidget_combo_box)w;
    }
		
    /* make sure nobody else has the text focus */
    ResetActiveTextsExceptFor(w);

    button_pressed = (int)eventP->xbutton.button;

    if (((eventP->xbutton.time - last_click_time) <= DoubleMillisec) &&
		(last_button_clicked == button_pressed))
		edit_win->num_clicks++;
    else
		edit_win->num_clicks = 1;

    last_button_clicked = button_pressed;
    last_click_time = eventP->xbutton.time;

    if ( button_pressed == Button1 ) {

		/* handle multi-click events in other functions */
		if (edit_win->num_clicks == 2) {
			EditWinDrawCursor(w, edit_win, TC_ERASE);
			EditWinHighlightWords(w, edit_win, x, x);
			return;
		} else if (edit_win->num_clicks == 3) {
			EditWinDrawCursor(w, edit_win, TC_ERASE);
			EditWinHighlightLine(w, edit_win, x);
			return;
		}
    
		sel = &(edit_win->sel);
		/* shift button is being held down; create a highlighted region */
		if ( ThimIsShifted()) {
			EditWinDrawCursor(w, edit_win, TC_ERASE);
			EditWinAdjustHighlight(w, edit_win,
								   MIN(sel->endIndex, sel->startIndex),
								   MAX(sel->endIndex, sel->startIndex),
								   UNHIGHLIGHT);
			sel->endIndex = EditWinGetNearestIndex(w, edit_win, x);
			if (sel->endIndex != sel->startIndex) {
				EditWinAdjustHighlight(w, edit_win,
									   MIN(sel->endIndex, sel->startIndex),
									   MAX(sel->endIndex, sel->startIndex),
									   HIGHLIGHT);
				w->pending_delete = TRUE;
			}
		} else {
			/* clear any existing selection */
			/*      EditWinClearSelection(wp, XA_PRIMARY, TRUE);*/
			EditWinDrawCursor(w, edit_win, TC_ERASE);
			w->pending_delete = FALSE;
			EditWinPaintWindow(w, edit_win);
			edit_win->sel.startIndex = edit_win->sel.endIndex =
				EditWinGetNearestIndex(w, edit_win,
									   eventP->xbutton.x);
			/* acquire selection atom (XA_PRIMARY) */
		}    

    } else if ( button_pressed == Button2 ) {
		edit_win->cursor_pos = EditWinGetNearestIndex(w, edit_win,
													  eventP->xbutton.x);
		if (w->type == W_ENTER) {
			((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
			XConvertSelection(Dpy, XA_PRIMARY, XA_STRING, AxXA_ENTER_PROP,
							  w->window, eventP->xbutton.time);
		} else {
			XConvertSelection(Dpy, XA_PRIMARY, XA_STRING, AxXA_COMBOBOX_PROP,
							  w->window, eventP->xbutton.time);
		}
    }
}

void
EditWinManageButtonRelease(w, edit_win, eventP)
    AXWidget w;
    struct edit_window *edit_win;
    XEvent *eventP;
{
    int button_released;
    seln_t *sel;

    if (!eventP)
		return;
    
    sel = &(edit_win->sel);
  
    button_released = (int) eventP->xbutton.button;
    if (button_released != Button1)
		return;

    if ((w->type == W_ENTER) ||
		((w->type == W_COMBO_BOX) && (w->flags & WFLG_EDITABLE))) {
		
		EditWinArmed = (AXWidget) NULL;

		/* if text is not already highlighted, then we set the new startIndex
		   at the current (new) cursor position */
		if (!w->pending_delete) {

			sel->startIndex = edit_win->cursor_pos =
				EditWinGetNearestIndex(w, edit_win,
									   eventP->xbutton.x);

		} else {
			edit_win->cursor_pos = sel->endIndex;
		}
    
		if (w->type == W_ENTER)
			((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
    
		EditWinDrawCursor(w, edit_win, TC_DRAW);
    }
}

void EditWinManageMouseMotion(w, edit_win, eventP)
    AXWidget w;
    struct edit_window *edit_win;
    XEvent *eventP;
{
    int oldEnd, editWidth, editHeight, stringWidth;

    int repaint_flag = FALSE;
    int highlight_flag = FALSE;
    seln_t *sel;

    if (!eventP) {
		return;
	}
    
    if (AxButtonDown != Button1)
		return;

	if (!EditWinArmed) {
		/* test whether to start autoscroll
		 */
		editwinAutoscrollEdge = mouseOutsideEditwin(w, eventP);
		if (noEdge != editwinAutoscrollEdge) {
			EditWinArmed = w;    /* start autoscroll */
		}
	}
	else if (EditWinArmed) {
		/* test whether to stop autoscroll
		 */
		editwinAutoscrollEdge = mouseOutsideEditwin(w, eventP);
		if (noEdge == editwinAutoscrollEdge) {
			EditWinArmed = (AXWidget) NULL;
		}
	}
	
    /* if the whole line is highlighted already, no need to do anything more */
    if (edit_win->num_clicks == 3)
		return;
  
    sel = &(edit_win->sel);

    oldEnd = sel->endIndex;
    EditWinWidthHeight(w, &editWidth, &editHeight);

    if (edit_win->num_clicks == 1) {
		if (eventP->xbutton.x < 0) {
			sel->endIndex = EditWinGetNearestIndex(w, edit_win, 0);
			if (sel->endIndex > 0) {
				edit_win->cursor_pos = --(sel->endIndex);
				repaint_flag = TRUE;
			}
			if ((sel->endIndex == 0) &&
				(edit_win->string_offset != TEXT_OFFSET)) {
				edit_win->cursor_pos = 0;
				repaint_flag = TRUE;
			}
		} else if (eventP->xbutton.x > editWidth+TEXT_OFFSET) {
			/* we feed EditWinGetNearestIndex dummy numbers to get it to
			   highlight to the end of the field (not the text) */
			sel->endIndex =
				EditWinGetNearestIndex(w, edit_win, editWidth+TEXT_OFFSET);
			if (sel->endIndex < edit_win->text_len) {
				edit_win->cursor_pos = ++(sel->endIndex);
				repaint_flag = TRUE;
			}
			if (sel->endIndex == edit_win->text_len) {
				edit_win->cursor_pos = edit_win->text_len;
				stringWidth = EditWinGetStringWidth(w, edit_win->text, 0,
													edit_win->text_len);
				if (edit_win->string_offset >
					-(stringWidth-editWidth+3*TEXT_OFFSET))
					repaint_flag = TRUE;
			}
		} else {
			sel->endIndex =
				EditWinGetNearestIndex(w, edit_win, eventP->xbutton.x);
		}
	
		if (w->type == W_ENTER)
			((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
		highlight_flag = (oldEnd != sel->endIndex);
    
    } else if (edit_win->num_clicks == 2) {
		int new_ind, start, end;

		start = MIN(sel->startIndex, sel->endIndex);
		end = MAX(sel->startIndex, sel->endIndex);
		new_ind = EditWinGetNearestIndex(w, edit_win, eventP->xbutton.x);

		if ((new_ind > start) && (new_ind < end))
			return;
		else if (new_ind < start) {
			edit_win->num_clicks = 1;
			sel->startIndex = end;
			sel->endIndex = start;
			EditWinManageMouseMotion(w, edit_win, eventP);
			return;
		} else if (new_ind > end) {
			edit_win->num_clicks = 1;
			sel->startIndex = start;
			sel->endIndex = end;
			EditWinManageMouseMotion(w, edit_win, eventP);
			return;
		}      
	
    }
    /* only change highlighting if endIndex has changed */
    if (highlight_flag) {

		if (repaint_flag) {
      
			/* we need to shift the string visibly, so repaint the whole
			   window */
			EditWinPaintWindow(w, edit_win);
      
		} else {

			/* just un-highlight originally highlighted text */
			/* make sure oldEnd is w/in the bounds of the new string */
			if (oldEnd > edit_win->text_len) {
				oldEnd = edit_win->text_len;
			}
    
			EditWinAdjustHighlight(w, edit_win, MIN(sel->startIndex, oldEnd),
								   MAX(sel->startIndex, oldEnd),
								   UNHIGHLIGHT);

		}
    
		/* only highlight if we're actually over > 0 characters */
		if ( sel->endIndex != sel->startIndex ) {
			EditWinAdjustHighlight(w, edit_win,
								   MIN(sel->startIndex, sel->endIndex),
								   MAX(sel->startIndex, sel->endIndex),
								   HIGHLIGHT);
			w->pending_delete = TRUE;
		} else
			w->pending_delete = FALSE;
    }
}

/**************************************************************************
  EditWinGetNearestIndex() -
        returns index of nearest text position in edit window
*/
int
EditWinGetNearestIndex(w, edit_win, x)
    AXWidget w;
    struct edit_window *edit_win;
    int x;
{
    int span;
    int editWidth, editHeight;
    wchar *p;

    if (!edit_win->text) {
		return (0);
    }
  
    EditWinWidthHeight(w, &editWidth, &editHeight);

    if (x < edit_win->string_offset)
		x = edit_win->string_offset;
  
    span = edit_win->string_offset;
    for (p = edit_win->text; *p; p++) {
		if (x < span)
			break;
        /* Ashisuto using the convinient existing function. 1998.02.25 */
        span += EditWinGetStringWidth(w, p, 0, 1);
    }
  
    if (!*p && (x >= span)) {
		return(p - edit_win->text);
    } else {
		return(p - edit_win->text - 1);
    }
}

/**************************************************************************
  EditWinReset() - reset state of edit window
*/  
void EditWinReset(w, edit_win)
    AXWidget w;
    struct edit_window *edit_win;
{
    seln_t *sel;

    sel = &(edit_win->sel);
  
    edit_win->edited = FALSE;
	if (edit_win->editmask) {
		edit_win->cursor_pos = sel->startIndex = sel->endIndex =
			EditMaskSmallestIndex(edit_win->editmask);
	}
	else {
		edit_win->cursor_pos = sel->startIndex = sel->endIndex =
			EditWinAssignCursorPos(edit_win, 0);
	}
    w->pending_delete = FALSE;

    if (w->type == W_ENTER) {
		edit_win->string_offset = TEXT_OFFSET;
		((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
    } else {
		edit_win->string_offset = TEXT_OFFSET;
		ComboBoxInEditWindow = NULL;
    }
}

/**************************************************************************
  EditWinDrawCursor() - takes three args.  mode should be either
       TC_DRAW or TC_ERASE (yes, this is poor style, but it remains for
       historical reasons (xmeditbx.c)
*/
void EditWinDrawCursor(w, edit_win, mode)
    AXWidget w;
    struct edit_window *edit_win;
    int mode;
{
    int cursor_height;
    int vert_offset = 0;
    GC gc;
    int restore_fg = FALSE;
    int width = 0;
    PIXEL_VALUE saved_fg;
    XFontStruct *fs;
    int editWidth, editHeight;
  
    AxSetWidgetWorkColors(w);

    if ( mode == TC_DRAW )
		gc = ForeGC;
    else
		gc = BackGC;
  
    if (w->widgetColor) {
		/* need to use a different color to draw the cursor */
		restore_fg = TRUE;
		saved_fg = AXGetForegroundPixel(Dpy, gc);
		if (TC_ERASE == mode) {
			/* use the widget's background pixel */
			XSetForeground(Dpy, gc, w->widgetColor->any.pixel);
		}
		else {
			/* mode is TC_DRAW */
			if (w->textColor) {
				/* use the text color for the cursor */
				XSetForeground(Dpy, gc, w->textColor->any.pixel);
			}
			else {
				/* use the wkit foreground pixel for the cursor */
				XSetForeground(Dpy, gc, AXGetForegroundPixel(Dpy, ForeGC));
			}
		}
    }

    width +=
		EditWinGetStringWidth(w, edit_win->text, 0,
							  edit_win->cursor_pos);

    EditWinWidthHeight(w, &editWidth, &editHeight);
    /*  width = MIN(width,editWidth-2*TEXT_OFFSET);*/

    width += edit_win->string_offset;
    width = MIN(width, editWidth);
  
    if (w->font)
		fs = w->font;
    else
		fs = AxDlgMainFont;
  
    cursor_height = fs->ascent + fs->descent;

    if (w->type == W_COMBO_BOX)
		vert_offset = TEXT_OFFSET;
    else {
		vert_offset = ENTRY_VOFFSET;
		width -= 1;
    }

    XDrawLine(Dpy, edit_win->win, gc, width,
			  vert_offset, width,
			  vert_offset + cursor_height);	/* vertical */
    XDrawLine(Dpy, edit_win->win, gc, width - 1,
			  vert_offset, width + 1, vert_offset);	/* top */
    XDrawLine(Dpy, edit_win->win, gc, width - 1,
			  vert_offset + cursor_height, width + 1,
			  vert_offset + cursor_height);	/* bottom */

    if (mode == TC_DRAW)
		edit_win->cursor_on = TRUE;
    else
		edit_win->cursor_on = FALSE;
  
    if (restore_fg) {
		XSetForeground(Dpy, gc, saved_fg);
    }

    AxUnsetWidgetWorkColors(w);
}

void EditWinAdjustHighlight(w, edit_win, start_index, end_index, mode)
    AXWidget w;
    struct edit_window *edit_win;
    int start_index;
    int end_index;
    highlight mode;
{
    int editWidth, editHeight, start_x, char_above, extra, char_height,
		itmp, cursor_height, hilite_width, hilite_x;
    int diff_x = 0;
    XFontStruct *fs;
    XRectangle rect[1];
    GC text_gc, back_gc;
	wcolor_t *savedTextColor, *savedWidgetColor;
	int restore_colors = FALSE;

	if (w->font) {
		char_above = w->font->max_bounds.ascent;
		char_height = char_above + w->font->max_bounds.descent;
		fs = w->font;
    } else {
		char_above = AxCharAbove;
		char_height = AxCharHeight;
		fs = AxDlgMainFont;
    }

    itmp = char_height + ((BORDER_WIDTH + TEXT_OFFSET) * 2);
    if (w->height > itmp) {
		extra = (w->height - itmp + 1) / 2;
    }
    else {
		extra = 0;
    }

    start_x = edit_win->string_offset +
		EditWinGetStringWidth(w, edit_win->text, 0, start_index);
    diff_x = EditWinGetStringWidth(w, edit_win->text, start_index,
								   end_index);
	
    EditWinWidthHeight(w, &editWidth, &editHeight);


	hilite_width = diff_x;
	hilite_x = start_x;
	if (0 == start_index) {
		hilite_x--;
		hilite_width++;
	}
	if (edit_win->text_len - 1 <= end_index) {
		hilite_width++;
	}
	
    AxSetWidgetWorkColors(w);
	
	savedWidgetColor = savedTextColor = (wcolor_t *) 0;
    if (mode == HIGHLIGHT) {
		text_gc = BackGC;
		back_gc = ForeGC;
		/* invert text, widget colors during hilite
		 */
		savedWidgetColor = w->widgetColor;
		savedTextColor = w->textColor;
		w->widgetColor = savedTextColor;
		w->textColor = savedWidgetColor;
		restore_colors = TRUE;
    }
	else {
		/* safety: if it's anything else, draw normal text */
		text_gc = ForeGC;
		back_gc = BackGC;
    }
    cursor_height = fs->ascent + fs->descent;
    if (w->type == W_COMBO_BOX) {
		xmColorFillRectangle(w, edit_win->win, back_gc, hilite_x,
							 TEXT_OFFSET - 1, hilite_width, cursor_height + 2);
	}
    else {
		xmColorFillRectangle(w, edit_win->win, back_gc, hilite_x - 1,
							 ENTRY_VOFFSET - 1, hilite_width,
							 cursor_height + 2);
	}
  
    /* set clipping rectangle appropriately */
    if (w->type == W_COMBO_BOX) {
		rect[0].x = TEXT_OFFSET;
    } else {
		rect[0].x = 0;
    }
    rect[0].y = 0;
    rect[0].width = editWidth - 2*TEXT_OFFSET;
    rect[0].height = editHeight;
    XSetClipOrigin(Dpy, text_gc, 0, 0);
    XSetClipRectangles(Dpy, text_gc, 0, 0, rect, 1, YXBanded);

	if (edit_win->text && edit_win->text_len) {
		if (w->type == W_COMBO_BOX) {
			xmDrawString16(w, edit_win->win, text_gc, start_x,
						 BORDER_WIDTH + char_above + extra,
						 edit_win->text + start_index,
						 end_index - start_index,
						 w->textShadow, w->textColor, w->widgetColor);
		}
		else if (w->type == W_ENTER) {
			wchar *outstr;
			int ix;
			
			if (!w->no_echo) {
				xmDrawString16(w, edit_win->win, text_gc, start_x-1,
							 char_above+ENTRY_VOFFSET,
							 edit_win->text+start_index,
							 end_index - start_index, w->textShadow,
							 w->textColor, w->widgetColor);
			} else {
				/* To echo passwords... */
				outstr = (wchar *)
					malloc((end_index-start_index + 1) * sizeof(wchar));
				for (ix = start_index; ix < end_index; ix++)
					outstr[ix] = '*';
				outstr[ix] = '\0';
				
				xmDrawString16(w, edit_win->win, text_gc, start_x-1,
							 char_above+ENTRY_VOFFSET, outstr,
							 end_index-start_index, w->textShadow,
							 w->textColor, w->widgetColor);
				free(outstr);
			}
		}
	}

    /* unset clipping rectangle */
    XSetClipMask(Dpy, text_gc, None);

	if (restore_colors) {
		w->widgetColor = savedWidgetColor;
		w->textColor = savedTextColor;
	}
	
    AxUnsetWidgetWorkColors(w);
}

int EditWinManageKeyPress(w, edit_win, ksym, wbuf, nbytes, eventP)
    AXWidget w;
    struct edit_window *edit_win;
    KeySym ksym;    /* keysym from key event */
    wchar *wbuf;    /* the displayable character */
    int nbytes;     /* number of chars (only looked at for zero)  */
    XEvent *eventP; /* the key event */
{
    int event_consumed = FALSE;

    if (!eventP)
		return FALSE;
    
    if (w->disabled)
		return FALSE;

    if ((w->type != W_ENTER) && (w->type != W_COMBO_BOX))
		return FALSE;
    
    if ((ComboBoxInEditWindow == NULL) && (w->type == W_COMBO_BOX))
		return FALSE;

    switch(ksym) {
    case XK_Left:
		if (edit_win->text_len > 0)
			EditWinBackwardChar(w, edit_win, (eventP->xkey.state & ShiftMask));
		event_consumed = TRUE;
		break;
    case XK_Right:
		if (edit_win->text_len > 0)
			EditWinForwardChar(w, edit_win, (eventP->xkey.state & ShiftMask));
		event_consumed = TRUE;
		break;
    case XK_Return:
    case XK_Tab:
    case XK_Execute:
		if (w->type == W_COMBO_BOX) {
			EditWinDrawCursor(w, edit_win, TC_ERASE);
			if (edit_win->edited || w->pending_delete) {
				CbFinish((AXWidget_combo_box)w, FALSE);
				EditWinPaintWindow(w, edit_win);
				ComboBoxInEditWindow = NULL;
				return (ksym != XK_Tab);  /* let dbox handle tabs, too */
			} else
				return (FALSE);
		} else /* w->type == W_ENTER */ {
			return (FALSE);
		}
    case XK_Delete:
    case XK_BackSpace:
		if (edit_win->text_len > 0) {
			EditWinManageDelete(w, edit_win);
			event_consumed = TRUE;
		}
		break;
    case XK_Escape:
		if (w->type == W_COMBO_BOX)
			EditWinDrawCursor(w, edit_win, TC_ERASE);
		if (edit_win->edited || w->pending_delete) {
			/* bail out of whatever edit was taking place, reshoot old val.
			 */
			w->value = edit_win->last_value;
			EditWinReset(w, edit_win);
			EditWinPaintWindow(w, edit_win);
		}
		/* let dbox handle escape as well for entry fields */
		event_consumed = (w->type == W_COMBO_BOX); 
		break;
    default:
		if ((eventP->xkey.state & ControlMask) &&
				!((nbytes == 1) && (wbuf[0] == EURO_SYMBOL_8BIT_VALUE)) ) {
			/* simple line editing commands */
			switch ((unsigned)*wbuf) {
			case (unsigned)1:	/* Ctrl-a */
				/* yes, this is a hacked way to do this, but it works well! */
				if (edit_win->text_len > 0) {
					EditWinDrawCursor(w, edit_win, TC_ERASE);
					edit_win->cursor_pos = EditWinAssignCursorPos(edit_win, 1);
					if (w->type == W_ENTER)
						((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
					EditWinBackwardChar(w, edit_win,
										(eventP->xkey.state & ShiftMask));
				}
				break;
			case (unsigned)2:	/* Ctrl-b */
				if (edit_win->text_len > 0)
					EditWinBackwardChar(w, edit_win,
										(eventP->xkey.state & ShiftMask));
				break;
			case (unsigned)4:	/* Ctrl-d */
				if ((edit_win->cursor_pos < edit_win->text_len) &&
					!(w->pending_delete) && (edit_win->text_len > 0)) {
					edit_win->cursor_pos++;
					if (w->type == W_ENTER)
						((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
					EditWinManageDelete(w, edit_win);
				} else if (w->pending_delete) {
					EditWinManageDelete(w, edit_win);
				}
				break;
			case (unsigned)5:	/* Ctrl-e */
				/* seem familiar?  same trick as for Ctrl-a */
				if (edit_win->text_len > 0) {
					EditWinDrawCursor(w, edit_win, TC_ERASE);
					edit_win->cursor_pos =
						EditWinAssignCursorPos(edit_win,
											   edit_win->text_len - 1);
					if (w->type == W_ENTER)
						((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
					EditWinForwardChar(w, edit_win,
									   (eventP->xkey.state & ShiftMask));
				}
				break;
			case (unsigned)6:	/* Ctrl-f */
				if (edit_win->text_len > 0)
					EditWinForwardChar(w, edit_win,
									   (eventP->xkey.state & ShiftMask));
				break;
			case (unsigned)11:	/* Ctrl-k */
				if (!(w->pending_delete) && (edit_win->text_len > 0)) {
					edit_win->sel.startIndex = edit_win->cursor_pos;
					edit_win->sel.endIndex = edit_win->text_len;
					w->pending_delete = TRUE;
				}
				EditWinManageDelete(w, edit_win);
				break;
			case (unsigned)21:	/* Ctrl-u */
				if (edit_win->text_len > 0) {
					edit_win->sel.startIndex = 0;
					edit_win->sel.endIndex = edit_win->text_len;
					w->pending_delete = TRUE;
					EditWinManageDelete(w, edit_win);
				}
				break;
			default:
				return(FALSE);
			}
		}
		if (!EditWinIsPrintable(w, wbuf, nbytes))
			return(FALSE);
		EditWinInsertText(w, edit_win, wbuf);

		event_consumed = TRUE;
    }

    return(event_consumed);
}

void EditWinDeleteSubstring(w, edit_win, start, end)
    AXWidget w;
    struct edit_window *edit_win;
    int start;
    int end;
{
    /* sanity check */
    if (!(edit_win->text) || (start < 0) || (end > edit_win->text_len)
		|| (start >= end)) {
		
		return;
	}

	if (edit_win->editmask) {
		EditMaskDeleteChars(edit_win->editmask, edit_win->text,
							start, end - start);
	}
	else {
		wchar *temp = (wchar *)malloc(sizeof(wchar) *
									  (edit_win->text_len - (end-start) + 1));
        Wstrncpy(temp, edit_win->text, start);
        Wstrncpy(&temp[start], &edit_win->text[end], edit_win->text_len - end);
		if (edit_win->text)
			free(edit_win->text);
		edit_win->text = temp;
		edit_win->text_len = edit_win->text_len - (end-start);
		edit_win->text[edit_win->text_len] = '\0';
	}
    if (w->type == W_ENTER) {
		((AXWidget_entry)w)->wstr =
			(wchar*)realloc(((AXWidget_entry)w)->wstr,
							sizeof(wchar)*(edit_win->text_len+1));
		Wstrcpy(((AXWidget_entry)w)->wstr, edit_win->text);
    }

    EditWinPaintWindow(w, edit_win);
    EditWinDrawCursor(w, edit_win, TC_DRAW);
    edit_win->edited = TRUE;
    if (w->value != -1) {
		edit_win->last_value = w->value;
		w->value = -1;
    }
}

void EditWinBackwardChar(w, edit_win, shifted)
AXWidget w;
struct edit_window *edit_win;
int shifted; /* whether or not shift button is depressed */
{
    seln_t *sel;

    sel = &(edit_win->sel);
  
    if (edit_win->cursor_pos > 0) {
		EditWinDrawCursor(w, edit_win, TC_ERASE);
		if (shifted) {
			if ( w->pending_delete ) { /* already highlighted */
				EditWinAdjustHighlight(w, edit_win,
									   MIN(sel->startIndex, sel->endIndex),
									   MAX(sel->startIndex, sel->endIndex),
									   UNHIGHLIGHT);
			}
			sel->endIndex = --edit_win->cursor_pos;
		}
		else {
			/* regular cursor movement
			 */
			if (edit_win->editmask) {
				edit_win->cursor_pos = 
					EditMaskNextLowerIndex(edit_win->editmask,
										   edit_win->cursor_pos);
			}
			else {
				edit_win->cursor_pos--;
			}
			sel->startIndex = edit_win->cursor_pos;
			sel->endIndex = sel->startIndex;
			w->pending_delete = FALSE;
		}
    
		if (w->type == W_ENTER)
			((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
		EditWinPaintWindow(w, edit_win);
		EditWinDrawCursor(w, edit_win, TC_DRAW);

		if (shifted) {
			if (sel->startIndex != sel->endIndex) {
				EditWinAdjustHighlight(w, edit_win,
									   MIN(sel->startIndex, sel->endIndex),
									   MAX(sel->startIndex, sel->endIndex),
									   HIGHLIGHT);
				w->pending_delete = TRUE;
			}
			else {
				/* if we're selecting over no text, turn off pending_delete
				 */
				w->pending_delete = FALSE;
			}
		}
    }
	else if (w->pending_delete && !shifted) {
		w->pending_delete = FALSE;
		EditWinPaintWindow(w, edit_win);
		EditWinDrawCursor(w, edit_win, TC_DRAW);
    }
}

void EditWinForwardChar(w, edit_win, shifted)
AXWidget w;
struct edit_window *edit_win;
int shifted; /* whether or not shift button is depressed */
{
    seln_t *sel;
  
    sel = &(edit_win->sel);
  
    if (edit_win->cursor_pos < edit_win->text_len) {
		EditWinDrawCursor(w, edit_win, TC_ERASE);
		if (shifted) {
			if (w->pending_delete) {
				EditWinAdjustHighlight(w, edit_win,
									   MIN(sel->startIndex, sel->endIndex),
									   MAX(sel->startIndex, sel->endIndex),
									   UNHIGHLIGHT);
			}
			sel->endIndex = ++edit_win->cursor_pos;

		}
		else {
			/* regular cursor movement
			 */
			if (edit_win->editmask) {
				edit_win->cursor_pos = 
					EditMaskNextHigherIndex(edit_win->editmask,
											edit_win->cursor_pos);
			}
			else {
				edit_win->cursor_pos++;
			}
			sel->startIndex = edit_win->cursor_pos;
			sel->endIndex = sel->startIndex;
			w->pending_delete = FALSE;
		}
    
		if (w->type == W_ENTER)
			((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
		EditWinPaintWindow(w, edit_win);
		EditWinDrawCursor(w, edit_win, TC_DRAW);

		if (shifted) {
			if (sel->startIndex != sel->endIndex) {
				EditWinAdjustHighlight(w, edit_win,
									   MIN(sel->startIndex, sel->endIndex),
									   MAX(sel->startIndex, sel->endIndex),
									   HIGHLIGHT);
				w->pending_delete = TRUE;
			}
			else {
				/* if we're selecting over no text, turn off pending_delete
				 */
				w->pending_delete = FALSE;
			}
		}
    }
	else if (w->pending_delete && !shifted) {
		w->pending_delete = FALSE;
		EditWinPaintWindow(w, edit_win);
		EditWinDrawCursor(w, edit_win, TC_DRAW);
    }
}

void EditWinManageDelete(w, edit_win)
    AXWidget w;
    struct edit_window *edit_win;
{
    seln_t *sel;
  
    /* sanity check */
    if (edit_win->cursor_pos < 0)
		return;

    sel = &(edit_win->sel);
    if (w->pending_delete) {
    
		EditWinDeleteSubstring(w, edit_win,
							   MIN(sel->startIndex, sel->endIndex),
							   MAX(sel->startIndex, sel->endIndex));
		w->pending_delete = FALSE;
		sel->startIndex = sel->endIndex = edit_win->cursor_pos =
			EditWinAssignCursorPos(edit_win,
								   MIN(sel->startIndex, sel->endIndex));
    } else if (edit_win->cursor_pos > 0) {
		EditWinDeleteSubstring(w, edit_win, edit_win->cursor_pos - 1,
							   edit_win->cursor_pos);
		EditWinAssignCursorPos(edit_win, --edit_win->cursor_pos);
		sel->startIndex = sel->endIndex = edit_win->cursor_pos;
    }
    if (w->type == W_ENTER)
		((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
    EditWinPaintWindow(w, edit_win);
    EditWinDrawCursor(w, edit_win, TC_DRAW);
    if (w->type == W_ENTER)
		if (w->callback && (w->flags & WFLG_TOUCHED_RETURN))
			AxCallback(w->callback, w, w->callbacktag, (caddr_t)NULL);
}

void EditWinInsertText(w, edit_win, wbuf)
AXWidget w;
struct edit_window *edit_win;
wchar *wbuf;
{
    seln_t *sel;
    wchar *cp;
    int start_index, end_index, text_len;
	int wbuf_length, ix, x_pos, editWidth, editHeight, wbuf_width;
	
    wbuf_length = Wstrlen(wbuf);
    sel = &(edit_win->sel);

    /* Ashisuto added the Entry widget maxLen validation check. - 1997.03.05 */
    if (w->type == W_ENTER) {
        text_len = edit_win->text_len + wbuf_length;
        if (w->pending_delete && edit_win->text) {
            text_len -= abs(sel->endIndex - sel->startIndex);
        }
        if (text_len > ((AXWidget_entry)w)->maxLen) {
            RingTheBell();
            return;
        }
    }
	
	if ((-1 == edit_win->cursor_pos) && !w->pending_delete) {
		/* need an in-range cursor index if field is not pending-delete
		 */
		edit_win->cursor_pos = 0;
	}
	
    EditWinDrawCursor(w, edit_win, TC_ERASE);
	
    if (w->pending_delete && edit_win->text) {
		EditWinDeleteSubstring(w, edit_win,
							   MIN(sel->startIndex, sel->endIndex),
							   MAX(sel->startIndex, sel->endIndex));
		w->pending_delete = FALSE;
		sel->startIndex = sel->endIndex = edit_win->cursor_pos =
			EditWinAssignCursorPos(edit_win,
								   MIN(sel->startIndex, sel->endIndex));
    }

	if (edit_win->editmask) {
		int num_accepted = EditMaskTestChars(edit_win->editmask,
											 edit_win->text,
											 wbuf, wbuf_length,
											 sel->startIndex);
		if (num_accepted != wbuf_length) {
			RingTheBell();
		}
		if (num_accepted > 0) {
			EditMaskOverwriteChars(edit_win->editmask, edit_win->text,
								   wbuf, wbuf_length, sel->startIndex);
			for (ix = 0; ix < num_accepted; ix++) {
				edit_win->cursor_pos =
					EditMaskNextHigherIndex(edit_win->editmask,
											edit_win->cursor_pos);
			}
		}
	}
	else {
		if (W_ENTER == w->type) {
			AXWidget_entry ew = (AXWidget_entry) w;
			if ((wbuf_length + edit_win->text_len) > ew->maxLen) {
				RingTheBell();
				return;
			}
		}
		
		/* allocate new space for text, move trailing text appropriately */
        edit_win->text =
            (wchar *)realloc(edit_win->text,
                             sizeof(wchar)*(edit_win->text_len+wbuf_length+1));
        if (edit_win->text_len != 0)
            memmove((void *) &edit_win->text[edit_win->cursor_pos +
                                             wbuf_length],
                    (void *) &edit_win->text[edit_win->cursor_pos],
                    (edit_win->text_len - edit_win->cursor_pos) *
                    sizeof(wchar));
 
        /* insert new text */
        for (ix = 0, cp = wbuf; ix < wbuf_length; ix++, cp++)
            edit_win->text[edit_win->cursor_pos + ix] = *cp;
 
        edit_win->text_len = edit_win->text_len + wbuf_length;
        edit_win->text[edit_win->text_len] = '\0';
        edit_win->cursor_pos += ix;
	}
		
	if (w->type == W_ENTER) {
		((AXWidget_entry)w)->wstr =
			(wchar *)realloc(((AXWidget_entry)w)->wstr,
					sizeof(wchar)*(edit_win->text_len + 1));
		Wstrcpy(((AXWidget_entry)w)->wstr, edit_win->text);
		((AXWidget_entry)w)->wstr[edit_win->text_len] = '\0';
	}

	sel->startIndex = edit_win->cursor_pos;
	sel->endIndex = sel->startIndex;
	if (w->type == W_ENTER)
		((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
	EditWinPaintWindow(w, edit_win);
	
    w->pending_delete = FALSE;
	
    EditWinDrawCursor(w, edit_win, TC_DRAW);
	
    if (w->type == W_ENTER)
		if (w->callback && (w->flags & WFLG_TOUCHED_RETURN))
			AxCallback(w->callback, w, w->callbacktag, (caddr_t)NULL);
	
    edit_win->edited = TRUE;
    if (w->value != -1) {
		edit_win->last_value = w->value;
		w->value = -1;
    }
}

/* this does (mostly) the same as ComboStringWidth, except that it
   doesn't necessarily count the width of the entire string */
int EditWinGetStringWidth(w, string, start, end)
    AXWidget w;
    wchar *string;
    int start, end;
{
    XFontStruct *fs;
    int width, i;
    wchar asterisk_char = '*';
	
	if (!string) {
		return 0;
	}
  
    if (start >= end)
		return(0);
  
    if (w->font)
		fs = w->font;
    else
		fs = AxDlgMainFont;

#if 1
	/* Ashisuto accelarated.    1998.02.25 */
    end = start + MIN( Wstrlen(string + start), (end - start) );
 
    if ( (w->type == W_ENTER) && ((AXWidget_entry)w)->no_echo )
        width = AXWTextWidth( fs, &asterisk_char, 1 ) * ( end - start );
    else
        width = AXWTextWidth( fs, (string + start), (end - start) );
 
#else
     ** above is ashisuto change
    for (width = 0, i = start; i < end; i++) {
		if (!(string[i]))
			break;
		if (!w->no_echo)
			width += GetAnyCharWidth(fs, string[i]);
		else
			width += GetAnyCharWidth(fs, '*');
    }
#endif

    return(width);
}

/* draw first LEN chars of string edit_win->text in edit window */
void EditWinDrawString(w, edit_win, gc, len)
    AXWidget w;
    struct edit_window *edit_win;
    GC gc;
    int len;
{
    int char_above, char_height, extra, itmp, sum, right_edge;
    int cursor_offset = 0;
    XRectangle rect[1];
    int editWidth, editHeight;
	
	if (!(edit_win->text && edit_win->text_len)) {
		edit_win->string_offset = TEXT_OFFSET;
		return;
	}
  
    if (w->font) {
		char_above = w->font->max_bounds.ascent;
		char_height = char_above + w->font->max_bounds.descent;
    } else {
		char_above = AxCharAbove;
		char_height = AxCharHeight;
    }
	
    itmp = char_height + ((BORDER_WIDTH + TEXT_OFFSET) * 2);
    if (w->height > itmp) {
		extra = (w->height - itmp + 1) / 2;
    }
    else {
		extra = 0;
    }

    cursor_offset += EditWinGetStringWidth(w, edit_win->text, 0,
										   edit_win->cursor_pos);

    EditWinWidthHeight(w, &editWidth, &editHeight);

    /* edit_win->string_offset holds the current position of the beginning
       of the string relative to the left edge of the edit window; so sum
       holds the position of the cursor relative to the window */
    sum = edit_win->string_offset + cursor_offset;

    if (w->type == W_COMBO_BOX)
		right_edge = editWidth - 3*TEXT_OFFSET;
    else
		right_edge = editWidth - 3*TEXT_OFFSET;
    
    /* if cursor is past right edge, pull entire string back a bit */
    if (sum > right_edge) {
		edit_win->string_offset = right_edge + TEXT_OFFSET - cursor_offset;
    }

    /* if cursor is past left edge, push string to the right */
    else if (sum < TEXT_OFFSET) {
		edit_win->string_offset = TEXT_OFFSET - cursor_offset;
    }
	
    AxSetWidgetWorkColors(w);

    /* set clipping rectangle appropriately */
    if (w->type == W_COMBO_BOX) {
		rect[0].x = TEXT_OFFSET;
		rect[0].width = right_edge;
    } else {
		rect[0].x = 0;
		rect[0].width = right_edge+TEXT_OFFSET;
    }
    rect[0].y = 0;
    rect[0].height = editHeight;
    XSetClipOrigin(Dpy, gc, 0, 0);
    XSetClipRectangles(Dpy, gc, 0, 0, rect, 1, YXBanded);

    if (w->type == W_COMBO_BOX)
		xmDrawString16(w, edit_win->win, gc, edit_win->string_offset,
					 BORDER_WIDTH + char_above + extra,
					 edit_win->text, len,
					 w->textShadow, w->textColor, w->widgetColor);
    else if (w->type == W_ENTER) {
		wchar *outstr;
		int ix;
	
		if (!w->no_echo) {
			xmDrawString16(w, edit_win->win, gc, edit_win->string_offset,
						 char_above+ENTRY_VOFFSET, edit_win->text, len,
						 w->textShadow,w->textColor, w->widgetColor);
		} else {
			/* To echo passwords... */
			outstr = (wchar *) malloc((len + 1) * sizeof(wchar));
			for (ix = 0; ix < len; ix++)
				outstr[ix] = '*';
			outstr[ix] = '\0';

			xmDrawString16(w, edit_win->win, gc, edit_win->string_offset,
						 char_above+ENTRY_VOFFSET, outstr, len,
						 w->textShadow, w->textColor, w->widgetColor);
			free(outstr);
		}
    } else if (w->type == W_KANJI_KANA) {
        xmDrawString16(w, edit_win->win, gc, edit_win->string_offset,
                     char_above+ENTRY_VOFFSET, edit_win->text, len,
                     w->textShadow,w->textColor, w->widgetColor);
    }

    /* unset clipping rectangle */
    XSetClipMask(Dpy, gc, None);
	
    AxUnsetWidgetWorkColors(w);
}

void EditWinFindWords(w, edit_win, x1, x2, start, end)
    AXWidget w;
    struct edit_window *edit_win;
    int x1, x2;
    int *start, *end; /* return values */
{
    int index1, index2, ix;
  
    /* Highlight word */
    index1 = EditWinGetNearestIndex(w, edit_win, x1);
    index2 = EditWinGetNearestIndex(w, edit_win, x2);

    *start = 0;
    *end = edit_win->text_len;

    /* find nearest word boundaries */
    for (ix = 0; ix < edit_win->text_len; ix++) {
		if ( *(edit_win->text + ix) != ' ' )
			continue;
		if (ix < index1) {
			if (ix > *start)
				*start = ix;
		}
		else if (ix > index2) {
			*end = ix;
			break;
		}
    }
}

/* returns index of (x1, 0) */
int EditWinHighlightWords(w, edit_win, x1, x2)
    AXWidget w;
    struct edit_window *edit_win;
    int x1; /* x-coord of left-most click */
    int x2; /* x-coord of right-most click */
{
    seln_t *sel;
    int index1, index2, ix;
	
	if (!(edit_win->text && edit_win->text_len)) {
		EditWinDrawCursor(w, edit_win, TC_DRAW);
		return 0;
	}
  
    sel = &(edit_win->sel);
    EditWinDrawCursor(w, edit_win, TC_ERASE);

    index1 = EditWinGetNearestIndex(w, edit_win, x1);
    index2 = EditWinGetNearestIndex(w, edit_win, x2);

    EditWinFindWords(w, edit_win, x1, x2,
					 &(sel->startIndex), &(sel->endIndex));
  
    /* if we have no words to highlight, get outta here */
    if (sel->startIndex == sel->endIndex) {
		w->pending_delete = FALSE;
		return (index1);
    }
  
    /* don't Highlight leading space */
    if ( *(edit_win->text + sel->startIndex) == ' ' )
		sel->startIndex++;
    EditWinAdjustHighlight(w, edit_win, sel->startIndex,
						   sel->endIndex, HIGHLIGHT);
    w->pending_delete = TRUE;

    /* place cursor in "correct" position, at closer end of word; this
       is done this way because EditWinManageButtonRelease sets cursor_pos
       to sel->endIndex */
    if ((index1 - sel->startIndex) < (sel->endIndex - index2)) {
		ix = sel->endIndex;
		sel->endIndex = sel->startIndex;
		sel->startIndex = ix;
    }
    return(index1);
}

void EditWinHighlightLine(w, edit_win, x)
    AXWidget w;
    struct edit_window *edit_win;
    int x;
{
    seln_t *sel;
    int char_index;
  
    /* if we have a zero-length string, don't highlight */
    if (edit_win->text_len == 0)
		return;
  
    sel = &(edit_win->sel);
    sel->startIndex = 0;
    sel->endIndex = edit_win->text_len;

    EditWinAdjustHighlight(w, edit_win, sel->startIndex,
						   sel->endIndex, HIGHLIGHT);
    w->pending_delete = TRUE;

    /* place cursor in "correct" position, at closer end of line */
    char_index = EditWinGetNearestIndex(w, edit_win, x);
    if (char_index < sel->endIndex - char_index)
		edit_win->cursor_pos = EditWinAssignCursorPos(edit_win, 0);
    else
		edit_win->cursor_pos = EditWinAssignCursorPos(edit_win, sel->endIndex);

    /* this is done because curPos = -1 in the old entry box code meant
       that the field was highlighted, so we have to keep this behavior */
    if (w->type == W_ENTER)
		((AXWidget_entry)w)->curPos = edit_win->cursor_pos;
}

void EditWinAutoScroll(w, edit_win)
AXWidget w;
struct edit_window *edit_win;
{
	int editWidth, editHeight;
	XEvent xev;
	Window root_return, child_return;
	struct edit_window *ee;

	ee = GetEditWin(w);
	if (ee) {
		if (XQueryPointer(Dpy, ee->win, &root_return, &child_return,
						  &xev.xbutton.x_root, &xev.xbutton.y_root,
						  &xev.xbutton.x, &xev.xbutton.y,
						  &xev.xbutton.state)) {
			if (xev.xbutton.state & Button1Mask) {
				editwinAutoscrollEdge = mouseOutsideEditwin(w, &xev);
				if (noEdge == editwinAutoscrollEdge) {
					EditWinArmed = (AXWidget) NULL;
					return;
				}
			}
		}
	}
	
	EditWinWidthHeight(w, &editWidth, &editHeight);
	
	/* fabricate a motion event that looks like it's
	 * off the appropriate edge, then pass to motion handler.
	 * (motion handler wants coordinate info, that's it.)
	 */
	memset((void *) &xev, 0, sizeof(XEvent));
	xev.xbutton.type = MotionNotify;
	xev.xbutton.display = Dpy;
	xev.xbutton.window = edit_win->win;
	xev.xbutton.y = editHeight / 2;
	xev.xbutton.x = (editwinAutoscrollEdge == leftEdge) ? -10 : editWidth + 10;
	
    EditWinManageMouseMotion(w, edit_win, &xev);
}

void EditWinPaintWindow(w, edit_win)
    AXWidget w;
    struct edit_window *edit_win;
{
    int editWidth;
    int editHeight;
  
    EditWinWidthHeight(w, &editWidth, &editHeight);
    AxSetWidgetWorkColors(w);
    xmColorFillRectangle(w, edit_win->win, BackGC, 0, 0, editWidth,
						 editHeight);
    AxUnsetWidgetWorkColors(w);
	
    EditWinDrawString(w, edit_win, ForeGC, edit_win->text_len);
}

/* : EditWinGetSelectedText - return wchar string containing
 * text currently selected.  Caller must free returned string.
 */
wchar *
EditWinGetSelectedText(edit_win)
struct edit_window *edit_win;
{
	wchar *ret;
	int sx, ex, nchars;
	
	if (!edit_win) {
		return ((wchar *) NULL);
	}
	if (edit_win->sel.startIndex == edit_win->sel.endIndex) {
		return ((wchar *) NULL);
	}
	
	if (edit_win->sel.startIndex < edit_win->sel.endIndex) {
		sx = edit_win->sel.startIndex;
		ex = edit_win->sel.endIndex;
	}
	else {
		sx = edit_win->sel.endIndex;
		ex = edit_win->sel.startIndex;
	}
	nchars = ex - sx;
	ret = (wchar *) malloc ((nchars + 1) * sizeof(wchar));
	
	Wstrncpy(ret, &edit_win->text[sx], nchars);
	ret[nchars] = (wchar) 0;
	
	return (ret);
}

/* : EditWinDoPendingDelete - delete selected text from
 * the passed edit_win struct.
 * 
 * Note that this func does only this deletion.  The
 * edit_win owner may have additional state that needs
 * to be kept consistent, or may need to send an event..
 */
void
EditWinDoPendingDelete(edit_win)
struct edit_window *edit_win;
{
    if (edit_win->sel.startIndex != edit_win->sel.endIndex) {
		/* do pending-delete of selected text
		 */
		int sx, ex, new_len;
		wchar *new_wstr;
		
		if (edit_win->sel.endIndex > edit_win->sel.startIndex) {
			ex = edit_win->sel.endIndex;
			sx = edit_win->sel.startIndex;
		}
		else {
			sx = edit_win->sel.endIndex;
			ex = edit_win->sel.startIndex;
		}
		if (edit_win->editmask) {
			EditMaskDeleteChars(edit_win->editmask, edit_win->text,
								sx, ex - sx);
		}
		else {
			new_len = Wstrlen(edit_win->text) - (ex - sx);
            /* Ashisuto modified to avoid crash in data func insert. */
            if (new_len < 0) {
                /* forgotten clear entry widget? do temporary changes */
                new_len = 0;
                sx = 0;
            }
			new_wstr = (wchar *) malloc((new_len + 1) * sizeof(wchar));
			if (sx > 0) {
				Wstrncpy(new_wstr, edit_win->text, sx);
			}
			Wstrncpy(&new_wstr[sx], &edit_win->text[ex], new_len - sx);
			new_wstr[new_len] = (wchar) 0;
			free(edit_win->text);
			edit_win->text = new_wstr;
			edit_win->text_len = Wstrlen(edit_win->text);
		}
        edit_win->cursor_pos = EditWinAssignCursorPos(edit_win, sx);
		edit_win->sel.startIndex = edit_win->sel.endIndex =
			edit_win->cursor_pos;
		edit_win->edited = TRUE;
    }
}

struct edit_window *
GetEditWin(ww)
AXWidget ww;
{
	struct edit_window *ee = 0;
	switch(ww->type) {
	case W_ENTER:
		ee = &(((AXWidget_entry) ww)->edit_win);
		break;
	case W_COMBO_BOX:
		ee = &(((AXWidget_combo_box) ww)->edit_win);
		break;
	}
	return (ee);
}

/* : mouseOutsideEditwin - return whether passed mouse
 * event, assumed to be for the editwin window, is outside
 * the editwin' footprint.  Can return "noEdge", meaning
 * mouse event is inside editwin footprint
 */
static enum editwinEdge
mouseOutsideEditwin(w, eventP)
AXWidget w;
XEvent *eventP;
{
	enum editwinEdge whichedge;
	int fudge = 5;
	
	if (eventP->xbutton.x < -fudge) {
		whichedge = leftEdge;
	}
	else if (eventP->xbutton.x > (int) w->width + fudge) {
		whichedge = rightEdge;
	}
	else {
		whichedge = noEdge;
	}

	return whichedge;
}

/* Not all assigments to cursor_pos come thru here; in particular,
 * mouse clicks and strokes may directly assign edit_win->cursor_pos.
 */
int EditWinAssignCursorPos(edit_win, newpos)
struct edit_window *edit_win;
int newpos;
{
	int mindex = 0;
	int maxdex = edit_win->text_len;
	if (edit_win->editmask) {
		mindex = EditMaskSmallestIndex(edit_win->editmask);
		maxdex = EditMaskHighestIndex(edit_win->editmask);
	}
	if (newpos < mindex) {
		newpos = mindex;
	}
	if (newpos > maxdex) {
		newpos = maxdex;
	}
	edit_win->cursor_pos = newpos;
	return edit_win->cursor_pos;
}

