/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: edtbxcom.c
   Applix ASC version: 23
   Date: Mon Nov 16 13:53:41 1998
*/
#include <basicwin.h>
#include <aos.h>
#include <xmlib.h>
#include <wchars.h>
#include <atmproto.h>

#ifndef MS_GetWindowText
#define MS_GetWindowText GetWindowText
#define MS_GetWindowTextLength GetWindowTextLength
#endif

/* IS_NGC adapted from ATM's tmin1.c definition */
#define DEL     0177        /* ascii code */
#define IS_NGC(c) ((((unsigned char)(c)) < ' ') || (((unsigned char)(c)) == DEL))


extern int WidgetNeedsUpdated;  /* Set when a widget needs repainted */
extern XFontStruct *MonospaceFont, *AxDlgMainFont;
extern int AxCharAbove, AxCharBelow, AxCharHeight;

extern int AXmNewFootprint();

void PaintEditBox();
void createNewObj();
void createLine();
void deleteText();
void getCursorInfo();
void setEditBoxSize();
void deleteSelection();
void manageDelete();
void setFormatInfo();
void xmEditboxSetLineHeight();
any_obj_t * initNewObj();
mbchar *WToMbOneStr();


static void
restoreSelectionInfo(w, op, start_ix, end_ix)
AXWidget_editbox w;
any_obj_t *op;
int start_ix, end_ix;
{
	w->editbox->sel.startObj = w->editbox->sel.endObj = op;
	w->editbox->sel.startIndex = start_ix;
	w->editbox->sel.endIndex = end_ix;
}

/*
 * getEditBoxData - supply edit box contents to requesting application.
 *                  Memory holding supplied contents should be freed
 *                  by application making request (since supplied data
 *                  is a copy, actual contents are not affected).
 */
mbchar *getEditBoxData(w)
AXWidget_editbox w;
{
    int size;
    mbchar *str, *tempP, *p1, *p2;
    int new_len;
    wchar *buffP;
	wchar *wstr = (wchar *) 0;
	wchar *wstr2 = (wchar *) 0;

    str = (mbchar *)NULL;
    if ( w->editbox == (editbox_t *)NULL )
        return( str );

#if AXMS                 /* {  */
    if ( !w->winid )
        return str;
    /* get the stuff in the editbox back from the Windows GDI */
    /* under Windows we do not maintain an edited version locally */
    size = MS_GetWindowTextLength( w->winid );
    
    if ( size < 0 )
        tempAbort( NXLT("bad return from MS_GetWindowTextLength") );
    size++;                            /* allow for null on end */
    buffP = w->editbox->objs->wdata;
    if ( buffP != (wchar *)NULL )
        buffP = (wchar *)realloc( w->editbox->objs->wdata, 
                                                     (size * sizeof(wchar)) );
    else
        buffP = (wchar *)malloc( size * sizeof(wchar) );
    tempP = malloc( size );
    new_len = MS_GetWindowText( w->winid, tempP, size );
    if ( new_len != size-1 )
        tempAbort( NXLT("bad return from MS_GetWindowText") );
    w->editbox->objs->wsize = MbToWStr( tempP, buffP );
    w->editbox->objs->wdata = buffP;
    free( tempP );

    /* Windows puts in CR/NL pairs at the end of hard-return lines.  Convert */
    /* these back to just NL's                                     */
    p1 = p2 = buffP;
    while ( *p2 != (mbchar)0 )
        {
        if ( *p2 == 0x0D )
           p2++;            /* skip CR */
        else
           *p1++ = *p2++;
        }
    *p1 = (mbchar) 0;
#endif                  /*   }  AXMS  */
       
#if AX_X11
	if (w->maxLen > 0) {
		/* formatting has been restricted to maxLen or fewer characters
		 * per line.  When "maxLen" linebreaks occur, the following line
		 * may need to be reported with an extra, leading newline.
		 *
		 * Some fuss occurs below because of tline_t quirks, in which
		 * newline characters are "skipped" by the pattern of tt->first_index,
		 * tt->len values.  (hence usage of soft_nl_indicies array)
		 */
		tline_t *tt;
		int *soft_nl_indicies;
		int num_soft_nl = 0;
		int prev_linelen_nonzero = FALSE;
		wchar *wc = w->editbox->objs->wdata;
		
		soft_nl_indicies = (int *) malloc(w->editbox->objs->wsize);
		for (tt = w->editbox->objs->lines; tt; tt = tt->next) {
			if (tt->len > 0) {
				if (prev_linelen_nonzero) {
					/* test whether to insert a newline at the start
					 * of this line.
					 */
					if ((wc[tt->first_index] != '\n') &&
						(tt->first_index > 0) &&
						(wc[tt->first_index - 1] != '\n')) {
					
						/* need to insert a newline
						 */
						soft_nl_indicies[num_soft_nl] = tt->first_index;
						num_soft_nl++;
					}
				}
				prev_linelen_nonzero = TRUE;
			}
			else {
				/* tt->len == 0 occurs for lines with a single, newline char.
				 */
				prev_linelen_nonzero = FALSE;
			}
		}
		
		if (num_soft_nl) {
			/* need to insert one or more newlines
			 */
			int ix, strsize, copylen, source_ix, dest_ix;
			
			strsize = w->editbox->objs->wsize + num_soft_nl;
			wstr2 = (wchar *) malloc((strsize + 1) * sizeof(wchar));
			memset((void *) wstr2, 0, strsize + 1);
			
			source_ix = dest_ix = 0;
			for (ix = 0; ix < num_soft_nl; ix++) {
				copylen = soft_nl_indicies[ix] - source_ix;
                Wstrncpy(&wstr2[dest_ix], &wc[source_ix], copylen);
				source_ix += copylen;
				dest_ix += copylen;
				wstr2[dest_ix++] = (wchar) '\n';
			}
			copylen = w->editbox->objs->wsize - source_ix;
			if (copylen > 0) {
                Wstrncpy(&wstr2[dest_ix], &wc[source_ix], copylen);
			}
			wstr = wstr2;
			size = strsize;
		}
		else {
			wstr = w->editbox->objs->wdata;
			size = w->editbox->objs->wsize;
		}
		free(soft_nl_indicies);
	}
	else {
		wstr = w->editbox->objs->wdata;
		size = w->editbox->objs->wsize;
	}
#else
	wstr = w->editbox->objs->wdata;
    size = w->editbox->objs->wsize;
#endif
    if (size) {
        str = (mbchar *)malloc( (size+1)*sizeof(wchar) );
        new_len = WnToMbStr( wstr, str, size );
        str[new_len] = '\0';
    }
	if (wstr2) {
		free(wstr2);
	}

    return(str);
}

void
xmEditboxSetLineHeight(w)
AXWidget_editbox w;
{
    XFontStruct *font;
    int fontHeight, fontAscent;
	int height_changed = FALSE;
	
#if AX_X11  
    if (w->font) {
        font = w->font;
    }
    else {
        font = AxDlgMainFont;
    }
#else
    font = AxDlgMainFont;
#endif  
    if (w->monospace_font) {
        font = MonospaceFont;    /* use mono variant of dlg main font */
    }
    fontAscent = font->ascent;
    fontHeight = font->ascent + font->descent;
    if (w->textShadow) {
        int xoffset, yoffset;
        xmTextShadowOffsets(w->font, &xoffset, &yoffset);
        yoffset++;                /* shadow may be diffused; add 1pix */
        fontHeight += yoffset;
    }

#if BUILD_KANJI
        fontAscent = fontAscent + 4;
        fontHeight =  fontHeight + 4;
#endif

    if (fontAscent != w->editbox->lineAscent ||
		fontHeight != w->editbox->lineHeight) 
		height_changed = TRUE;

    /* set up line ascent/height for text objs */
    w->editbox->lineAscent = fontAscent;
    w->editbox->lineHeight = fontHeight;

	if (height_changed && w->editbox->objs) {
		formatObjs(w, FMT_ALL);
	}

}

/* 
 * Create and initialize edit box. 
 * Attach a text object by default.
 */
void createEditBox(w)
AXWidget_editbox w;
{
    
    if (w->editbox)
        return;

    w->editbox = (editbox_t *) calloc(1, sizeof(editbox_t));

    setEditBoxSize(w, w->editbox);
	
	xmEditboxSetLineHeight(w);

    /* create and attach empty text object (default) */
    createNewObj(w, TEXTOBJ);

    /* initialize cursor */
    w->editbox->tcursor.obj = w->editbox->objs;
    w->editbox->tcursor.type = TC_ACTIVE;
}



/* 
 * Create, initialize, and position new object 
 */
void createNewObj(w, type)
AXWidget_editbox w;
int type;
{
    any_obj_t *op, *cur_op;
    tcursor_t *cursor;
    tline_t *lp;
    int objIndex, curIndex;

    /* get object containing cursor */
    getCursorInfo(w, &cursor, &cur_op, &lp);

    if (cur_op && cur_op->type==type)
        return;  /* no need to create new text object */

    /* create and initialize new object */
    op = initNewObj(w, type);
    
    /* position new object in current list of objects */
    if (cur_op == NULL) {  /* no objects exist */
        w->editbox->objs = op;
    }
    else {  /* other objects exist */
        int sizeofPart1, sizeofPart2;

        curIndex = w->editbox->tcursor.ac_index;
        sizeofPart1 = curIndex;
        sizeofPart2 = cur_op->wsize - curIndex;
        if (sizeofPart1 == 0) {
            /* put new obj before obj with cursor */
            op->next = cur_op;
            op->prev = cur_op->prev;
            cur_op->prev = op;
            if (op->prev)
                op->prev->next = op;
        }
        else if (sizeofPart2 == 0) {
            /* put new obj after obj with cursor */
            op->next = cur_op->next;
            op->prev = cur_op;
            cur_op->next = op;
            if (op->next)
                op->next->prev = op;
        }
        else {
            /* Split object with cursor (can only be text obj). 
             * Put cursor in second part of split obj.
             * Put new obj before obj with cursor. */
            any_obj_t *new_op;

            /* create new obj to hold first part of split data */
            new_op = initNewObj(w, cur_op->type);

            /* copy data of part1 to new obj */
            if ( !manageTextMem(w, new_op, sizeofPart1) ) {
                /* not enough memory */
                /* !!! check if anything else needs to be done here !!! */
                RingTheBell();
                return;
            }
            new_op->wsize = sizeofPart1;
            memcpy(new_op->wdata, cur_op->wdata, 
                                     (new_op->wsize*sizeof(wchar)) );

            /* delete data of part1 from obj with cursor */
            deleteText(w, sizeofPart1);

            op->next = cur_op;
            op->prev = new_op;
            new_op->next = op;
            new_op->prev = cur_op->prev;
            cur_op->prev = op;
            if (new_op->prev)
                new_op->prev->next = new_op;
            if (op->prev)
                op->prev->next = new_op;
        }
    }
}



/*
 * create and initialize new object - called ONLY by createNewObj().  
 * CAUTION: USE createNewObj() TO CREATE A NEW OBJECT
 */
any_obj_t *
initNewObj(w, type)
AXWidget_editbox w;
int type;
{
    any_obj_t *op;

    if ( !w || !w->editbox )
        return((any_obj_t *)NULL);

    op = (any_obj_t *) calloc(1, sizeof(any_obj_t));
    op->type = type;
    if (type == TEXTOBJ) {
        op->formatInfo = (tformat_t *) calloc(1, sizeof(tformat_t));
        op->allocsize = TEXTUNIT;
        op->wdata = (wchar *)calloc(1, TEXTUNIT );
        op->height = w->editbox->lineHeight;
        createLine(&op->lines);
        op->formatInfo->formatLine = op->lines;
    }
    return(op);
}


void deleteText(w, len)
AXWidget_editbox w;
int len;
{
    tcursor_t *cursor;
    any_obj_t *op;
    tline_t *lp;
    wchar *cp, *cp2;
    int curs_index, numChars;

    /* get object containing cursor */
    getCursorInfo(w, &cursor, &op, &lp);
    curs_index = cursor->ac_index;

    /* copy data "to the left" starting at cursor pos */
    cp = &op->wdata[curs_index]; /* char just after cursor */
    numChars = op->wsize - curs_index; /* #chars to right of deletion pt. */
    cp2 = cp - len;

    /* use bcopy instead of memcpy to handle overlap */
    bcopy((char *)cp, (char *)cp2, (numChars * sizeof(wchar)));

    /* update size and cursor pos */
    op->wsize -= len;
    cursor->ac_index -= len;
    w->changed = TRUE;
}


/*
 * manageTextMem - ensure object 'op' has enough storage for 
 *                 atleast 'len' additional bytes of "wchar" data.
 */
int
manageTextMem(w, op, len)
AXWidget_editbox w;
any_obj_t *op;
int len;
{
    wchar *cp;
    int newAllocsize;

    if ( ((op->wsize+len)*sizeof(wchar)) >= op->allocsize ) {
        newAllocsize = op->allocsize + 
                           ((TEXTUNIT*(1 + len/TEXTUNIT) * sizeof(wchar)));
        if ( newAllocsize > (MAXTEXTMEM + 2*TEXTUNIT) )
            return(FALSE);
        op->wdata = (wchar *)realloc( op->wdata, newAllocsize );
        *(op->wdata + op->wsize) = 0;  /* make sure old data is null-term */  
#if 0
        cp = (wchar *)calloc(1,newAllocsize);
        memcpy(cp, op->wdata, (op->wsize * sizeof(wchar)) );
        free((char *)op->wdata);
        op->wdata = cp;
#endif
        op->allocsize = newAllocsize;
    }
    return(TRUE);
}
   
void createLine(lineP)
tline_t **lineP;
{
    *lineP = (tline_t *) calloc(1, sizeof(tline_t));
}   
   
void updateLineIndex(w, objP, lineP, len)
AXWidget_editbox w;
any_obj_t *objP;
tline_t *lineP;
int len;
{
    tline_t *lp;

    for (lp=lineP->next; lp; lp=lp->next) {
        lp->first_index += len;
    }
}   
   


#define TEMP_BUF_SIZE 300

void manageInsert(w, wtext, len)
AXWidget_editbox w;
wchar *wtext;
int len;
{
#if AXMS    /*  {  */
    wchar *buffP;
    mbchar *tempP;

    /* we must store the text on the Applix widget for re-gens/refreshes */
    /* also, the text passed to Windows must be mixed-byte               */
    buffP = w->editbox->objs->wdata;
    if ( buffP != (wchar *)NULL )
        buffP = realloc( w->editbox->objs->wdata, ((len+1) * sizeof(wchar)) );
    else
        buffP = malloc( (len+1) * sizeof(wchar) );
    Wstrncpy( buffP, wtext, len );
    *(buffP+len) = 0;

    w->editbox->objs->wdata = buffP;
    w->editbox->objs->wsize = Wstrlen( buffP );
    
    /* now pass the text to the Windows/GDI  */
    tempP = WToMbOneStr( buffP );
    MS_SendMessage( w->winid, EM_REPLACESEL, (WPARAM)0, (LPARAM)tempP );
    return;
#else       /* }  {  */
    any_obj_t *op;
    tline_t *lp;
    int curs_index, start_ix, end_ix;
    tcursor_t *cursor;
    wchar temp_buf[TEMP_BUF_SIZE];
    int malloced_mem;

    start_ix = w->editbox->sel.startIndex;
    end_ix = w->editbox->sel.endIndex;

    if (w->pending_delete) {
		if (start_ix != end_ix) {
			if (start_ix < end_ix) {
				end_ix = start_ix;
			}
			else {
				start_ix = end_ix;
			}
		}
        deleteSelection(w);
	}
	start_ix += len;
	end_ix += len;

    /* get object containing cursor */
    getCursorInfo(w, &cursor, &op, &lp);
    curs_index = cursor->ac_index;

    /* lift old cursor */
    drawCursor(w, TC_ERASE);

    /* make sure allocated memory can hold new data */
    if ( !manageTextMem(w, op, len) ) {
        /* required memory exceed max editbox limit */
        RingTheBell();
        drawCursor(w, TC_DRAW);
		restoreSelectionInfo(w, op, start_ix - len, end_ix - len);
        return;
    }

    /* set formatting info for edited lines */
    if (lp->prev)
        op->formatInfo->formatLine = lp->prev;
    else
        op->formatInfo->formatLine = lp;
    op->formatInfo->numFormatLines = 2;
    setFormatInfo(w, op);

    insertText(w, wtext, len);

    /* update index */
    updateLineIndex(w, op, lp, len);
    /* reformat */
    formatObjs(w, FMT_LINE);
    /* draw new cursor */
    drawCursor(w, TC_DRAW);
	
	restoreSelectionInfo(w, op, start_ix, end_ix);

#endif  /*  } AXMS  */
}

insertText(w, text, len)
AXWidget_editbox w;
wchar *text;
int len;
{
    any_obj_t *op;
    tline_t *lp;
    int curs_index, numChars, i;
    wchar *cp, *cp2;
    tcursor_t *cursor;

    /* get object containing cursor */
    getCursorInfo(w, &cursor, &op, &lp);
    curs_index = cursor->ac_index;

    /* split data at cursor, then copy in new data */
    cp = &op->wdata[curs_index]; /* char just after cursor */
    numChars = op->wsize - curs_index; /* #chars to right of insertion pt. */
    cp2 = cp + len; /* new home for chars to right of insertion */

#if 0  /* too slow */
    for(i=numChars-1; i>=0; i--) {
        cp2[i] = cp[i];
    }
#else
    /* use bcopy instead of memcpy to handle overlap */
    bcopy((char *)cp, (char *)cp2, (numChars * sizeof(wchar)));
#endif
   
#if 1
    for ( i=0; i< len; i++ )
       {
       *cp = *text;
       cp++;
       text++;
       }
#else
    memcpy(cp, text, len );
#endif

    /* update size and cursor pos */
    op->wsize = op->wsize + len;
    cursor->ac_index =  cursor->ac_index + len;
    w->changed = TRUE;
}
   
   
   
void getLineFromIndex(w, objP,char_index, lineP)
AXWidget_editbox w;
any_obj_t *objP;
int char_index;
tline_t **lineP;  /* RETURN */
{
    tline_t *lp = NULL;

    if (!objP)
        return;

    for (lp=objP->lines; lp; lp=lp->next) {
        if (char_index>=lp->first_index && char_index<=lp->first_index+lp->len) 
            break;
    }

    *lineP = lp;
}

   
/*
 * getCursorInfo - returns cursor struct, and object & line
 *                 containing cursor
 */
void getCursorInfo(w, cursorP, objP, lineP)
AXWidget_editbox w;
tcursor_t **cursorP;  /* RETURN */
any_obj_t **objP;     /* RETURN */
tline_t **lineP;      /* RETURN */
{
    *cursorP = &w->editbox->tcursor;  /* cursor struct */
    *objP = (*cursorP)->obj;          /* object containing cursor */

    /* get line containing cursor */
    getLineFromIndex(w, *objP, (*cursorP)->ac_index, lineP);
}



/*
 * validateConvertChars - strip non-graphic characters.
 *                 Convert tabs to spaces, retain newlines.
 *                 Return valid string in caller-suplied buffer.
 */
void
validateConvertChars(buf, newbuf, len)
wchar *buf;     /* original buffer */
wchar *newbuf;  /* RETURN valid null-terminated string */
int len;       /* number of characters in original buffer */
{
    int ix, cnt;

    for ( ix = 0, cnt = 0; ix < len; ix++ ) {
        if ( !IS_NGC(buf[ix]) ) {
            /* printable character */
            newbuf[cnt] = buf[ix];
            cnt++;
        }
        else {
            /* non-graphic character */
            if ( buf[ix] == '\n' ) {
                /* keep newline */
#if AXMS
                /* we need to make the a CR-NL pair for the Windows widgets */
                newbuf[cnt] = 0x0D;
                cnt++;
#endif
                newbuf[cnt] = buf[ix];
                cnt++;
            }
            else if ( buf[ix] == '\t' ) {
                /* change tab to 4 spaces */
                newbuf[cnt++] = ' ';
                newbuf[cnt++] = ' ';
                newbuf[cnt++] = ' ';
                newbuf[cnt++] = ' ';
            }
        }
    }
    newbuf[cnt] = '\0';
}



/*
 * insertEditBoxData - add supplied data to existing contents.
 *                     (Should call PaintEditBox after this to
 *                     ensure that the contents are redisplayed
 *                     after the scroll adjustments).
 */
void insertEditBoxData(w, str)
AXWidget_editbox w;
mbchar *str;
{
    int new_size_bytes;
    int len;
    wchar *newstr, *wstr;

    if ( !w || !w->editbox || !str )
        return;

    len = (MBstrlen(str) + 1) * sizeof(wchar);    /* length in bytes */
    wstr = (wchar *)malloc( len );

    /*
     * validateConvertChars converts tabs to 4 spaces which at
     * worst will quadruple the length of the returned string;
     * add one more char for <CR> --> <CR><LF> conversion on NT
     */
    newstr = (wchar *)get_static_buffer( (len*4+2)*sizeof(wchar) );

    len = MbToWStr( str, wstr );    /* length in chars */

    /* strip non-graphic characters */
    validateConvertChars(wstr, newstr, len);

    len = Wstrlen(newstr);
    manageInsert(w, newstr, len);

    drawCursor(w, TC_ERASE);
    rewindCursor(w, len);
    dispCursorAtCenter(w);
}



/* assign editbox size to allow room for border and margin */
void
setEditBoxSize(w, ebx)
AXWidget_editbox w;
editbox_t *ebx;
{
    ebx->xpos = w->xpos + 4;
    ebx->ypos = w->ypos + 4;
    ebx->width = w->width - 8;
    ebx->height = w->height - 8;
}


/*
 * setFormatInfo - set old format state of edited lines.
 *                 Required to reduce flicker during reformatting.
 */
void setFormatInfo(w, objP)
AXWidget_editbox w;
any_obj_t *objP;
{
    any_obj_t *op;
    tformat_t *fp;
    tline_t *lp, *lastlp;
    int numlines, i, total_len;
    int start_index;

    if (objP->type != TEXTOBJ)
        return;

    op = objP;
    fp = op->formatInfo;
    start_index = fp->formatLine->first_index;
    numlines = fp->numFormatLines;

    /* calculate number of chars in lines being reformatted */
    for (lastlp=lp=fp->formatLine,i=0; lp && i<numlines; lp=lp->next,i++) {
        lastlp = lp;
    }
    total_len = (lastlp->first_index + lastlp->len) - start_index;

    /* reset fp->numformatLines (actual number of lines present
     * may be lesser than that previously set) */
    fp->numFormatLines = numlines = i;

    /* alloc memory to hold old info */
    fp->oldData = (wchar *) malloc( total_len*sizeof(wchar) );
    fp->oldIndex = (int *) calloc(numlines, sizeof(int));
    fp->oldLen = (int *) calloc(numlines, sizeof(int));
    fp->oldXpos = (int *) calloc(numlines, sizeof(int));
    fp->oldYpos = (int *) calloc(numlines, sizeof(int));

    /* copy old info to format struct */
    memcpy(fp->oldData, op->wdata+start_index, (total_len*sizeof(wchar)) );
    for (lp=fp->formatLine,i=0; lp && i<numlines; lp=lp->next,i++) {
        fp->oldIndex[i] = lp->first_index - start_index;
        fp->oldLen[i] = lp->len;
        fp->oldXpos[i] = w->editbox->xoffset + op->xpos + lp->xpos;
        fp->oldYpos[i] = w->editbox->yoffset + op->ypos + lp->ypos;
    }
}


void deleteSelection(w)
AXWidget_editbox w;
{
    /* !! manageDelete assumes that only one object exists !! */

    manageDelete(w, 0);
}

void resetSelectionInfo(w)
AXWidget_editbox w;
{
    seln_t *sel;

    sel = &w->editbox->sel;

    w->pending_delete = FALSE;
    sel->startObj = sel->endObj = NULL;
    sel->startIndex = sel->endIndex = 0;
}


void manageDelete(w, len)
AXWidget_editbox w;
int len;
{
    any_obj_t *op;
    tline_t *lp, *lp1, *tmplp;
    int curs_index, index1;
    tcursor_t *cursor;
    int start, end, orig_start, orig_end;

    /* lift old cursor */
    drawCursor(w, TC_ERASE);

	start = orig_start = w->editbox->sel.startIndex;
	end = orig_end = w->editbox->sel.endIndex;
	
    if (w->pending_delete) {
        /* !! this assumes that only one object exists !! */

        /* w->deleteMarked is FALSE only during middle button paste */
#if AX_X11
        if (!w->deleteMarked)
            return;
#endif
        if (start != end) {
            len = abs(end - start);
			if (start < end) {
				end = start;
			}
			else {
				start = end;
			}
		}
		
        resetSelectionInfo(w);
		
        /* change cursor.ac_index since it may not be at end of seln */
        w->editbox->tcursor.ac_index = MAX(orig_start, orig_end);
		
        /* unhilight before deleting */
        paintText(w, w->editbox->objs, orig_start, orig_end, GC_FORE);
    }
	else {
		end -= len;
		start -= len;
	}

    /* get object containing cursor */
    getCursorInfo(w, &cursor, &op, &lp);
    curs_index = cursor->ac_index;

    if (len <= 0) {
        /* nothing to delete */
        drawCursor(w, TC_DRAW);
		restoreSelectionInfo(w, op, orig_start, orig_end);
        return; 
    }

    if (curs_index == 0) {
        /* cannot delete past beginning of text obj (for now) */
        drawCursor(w, TC_DRAW);
		restoreSelectionInfo(w, op, orig_start, orig_end);
        return; 
    }

    /* set formatting info for edited lines */
    index1 = curs_index - len;
    getLineFromIndex(w, op, index1, &lp1);
    if (lp1->prev)
        op->formatInfo->formatLine = lp1->prev;
    else
        op->formatInfo->formatLine = lp1;

    /* calculate minimum # lines to reformat */
    op->formatInfo->numFormatLines = 1;
    tmplp = op->formatInfo->formatLine;
    while (tmplp && tmplp!=lp) {
        op->formatInfo->numFormatLines++;
        tmplp = tmplp->next;
    }
    setFormatInfo(w, op);

    deleteText(w, len);

    /* update index */
    updateLineIndex(w, op, lp, -len);
    /* reformat */
    formatObjs(w, FMT_LINE);
    /* draw new cursor */
    drawCursor(w, TC_DRAW);
	
	restoreSelectionInfo(w, op, start, end);
}


void checkEditBoxSize(w)
AXWidget w;
{
    /* make sure editbox is not too small
     * (very small widths imply a VERY large number of tline_t's) */
    if (w->width < 100)
        w->width = 100;
    if (w->height < 25)
        w->height = 25;
}




/*************************************************************************
function: XmCreateEditBox - Create edit box widget
*/
AXWidget_editbox
XmCreateEditBox(parentWidget, name, args, cnt)
register AXWidget parentWidget;
register char *name;
register Arg args[];
register int cnt;
{
    register AXWidget_editbox w;

    if ( args[0].value ) {
        w = (AXWidget_editbox)args[0].value;
        if (AXmNewFootprint(w, name, args, cnt)) {
            UnMapWidget(w);   /* clear area and destroy window */
#if AX_X11
            /* Windows editbox scrollbar handled by Windows */
            if ( w->vScrollBar ) {
                AxXtDestroyWidget(w->vScrollBar);
                w->vScrollBar = (AXWidget_scroll) NULL;
            }
#endif
        }
        w->update_widget = TRUE;
        WidgetNeedsUpdated = TRUE;
    }
    else 
        w = (AXWidget_editbox)CrtWidget(W_EDITBOX,parentWidget,name);

    w->painter = PaintEditBox;
    AxXtSetValues(w,args,cnt);
    createEditBox(w);
    checkEditBoxSize(w);
    setEditBoxSize(w, w->editbox);
    return((AXWidget_editbox)w);
}


/* 
 * destroyEditBox - free edit box memory (don't destroy window)
 */
void destroyEditBox(w)
AXWidget_editbox w;
{
    any_obj_t *op, *next_op;
    tline_t *lp, *next_lp;

    if ( !w->editbox )
        return;

    for(op = w->editbox->objs; op; op=next_op) {
        if (op->wdata)
            free((caddr_t)op->wdata);
        for (lp = op->lines; lp; lp = next_lp) {
            next_lp = lp->next;
            free(lp);
        }
        free(op->formatInfo);
        next_op = op->next;
        free(op);
    }

    free(w->editbox);
    w->editbox = NULL;
    w->pending_delete = FALSE;
}

