/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: entercom.c
   Applix ASC version: 27
   Date: Mon Nov 16 13:53:35 1998
*/
#include <stdlib.h>

#include <basicwin.h>
#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>
#include <wchars.h>
#include <elf_env.h>
#include <eldata.h>
#include <editmask.h>

extern int WidgetNeedsUpdated; /* need to repaint some widget */

#if AX_X11
extern void EditWinDoPendingDelete();
#endif

void EntAdjustCursorPos(/* widget_t *w */);
int  EntInsertText(/* widget_t *w, wchar *wbuf, int num_chars */);
void  EntDoPendingDelete(/* widget_t *w, int do_callback */);
extern AXWidget_main  TheMainWidget(/* AXWidget w */);

/*********************************************************************** 
function: PutCursorInEntry - Put cursor starting at START_CHAR and       
ending at END_CHAR in the entry W.                                       
NOTE: END_CHAR ignored (for now)                                         
Return TRUE if pending_delete turned on; FALSE otherwise.                
*/                                                                       
static int                                                               
PutCursorInEntry(w, start_char, end_char)                                
AXWidget_entry w;                                                             
int start_char; /* 0 based */                                            
int end_char;   /* 0 based */                                            
{                                                                        
    int selection_changed;
	int pending_delete = FALSE;
                                                                         
    if ((!w) || (w->type != W_ENTER))
        return(FALSE);
	
	pending_delete = (start_char < 0) || (end_char < 0) ||
		(end_char < start_char);
	
	if (!w->hasFocus) {
		xmTakeKeyFocus(w, (XButtonEvent *) 0, pending_delete);
	}
                                                                         
    if (pending_delete) {
        w->pending_delete = TRUE;
        w->curPos = -1;
    } else {
        w->pending_delete = FALSE;
        if (w->wstr)
            w->curPos = MIN(Wstrlen(w->wstr), start_char); /* clip */
        else
            w->curPos = 0; /* sanity */
    }

    /* some widget needs updating */
    WidgetNeedsUpdated = TRUE;
    /* this is the widget that needs the updating */
    w->update_widget = TRUE;
      
    return (w->pending_delete);
}

/*********************************************************************** 
function: ElfbEnterFuncs - Entry widget builtin functions entry point.
WVAL - widget handle
COMMAND_CODEP - entry widget builtin command code. (currently hardcoded)
DP1..DP5 - interpreted data depending on command code.
Note: Do not use command code value of 0.
*/
void
ElfbEnterFuncs(wval, command_codeP, dP1, dP2, dP3, dP4, dP5)
data_ptr wval;
data_ptr command_codeP;
data_ptr dP1, dP2, dP3, dP4, dP5;
{
    register AXWidget_entry w;
    int command_code;
    data_ptr retP = (data_ptr) NULL;
    int start_char;
    int end_char;
    int pending_delete;
	char  *insert_buf = (char *) NULL;
	wchar *insert_wbuf = (wchar *) NULL;
	int insert_num_chars;


    w = (AXWidget_entry)WidgetPointerFromHandleData(wval);
    if ((!w) || (w->type != W_ENTER))
        return; /* invalid args */

    command_code = IntFromDataPtr(command_codeP);
    switch (command_code) {
    case 1: /* db-cursor-in-entry */
        start_char = IntFromDataPtr(dP1); /* 1 based */
        end_char   = IntFromDataPtr(dP2); /* 1 based */

        /*
         *  NOTE: `start_char' can take the following values:
         *  -1 : (BTRUE) for compatiblity, means need pending_delete
         *  0  : (FALSE) means no pending_delete: treat 0 as a
         *        0 based char index.
         *  >0 : New usage: is a 1 based start char index.
         */
        if (start_char > 0)
            start_char--; /* make 0 based */
        if (end_char > 0)
            end_char--; /* make 0 based */

        pending_delete = PutCursorInEntry(w, start_char, end_char);
        if (pending_delete)
            retP = MK_INTEGER(BTRUE);
        else
            retP = MK_INTEGER(FALSE);
        break;

    case 2: /* no db wrapper: Returns state of the pending delete flag */
        pending_delete = w->pending_delete;
        if (pending_delete)
            retP = MK_INTEGER(BTRUE);
        else
            retP = MK_INTEGER(FALSE);
        break;
    case 3: /* db-get-cursor-start-pos */
        if (w->pending_delete)
            retP = MK_INTEGER(1);
        else
            retP = MK_INTEGER(w->curPos + 1);
        break;
    case 4: /* db-enter-insert-text */
		xmTakeKeyFocus(w, (XButtonEvent *) 0, w->pending_delete);
		insert_buf = StrFromDataPtr(dP1);
		insert_num_chars = strlen(insert_buf);
		if (insert_num_chars) {
			insert_wbuf = (wchar *) malloc((insert_num_chars + 1) *
										   sizeof(wchar));
            insert_num_chars = MbToWStr(insert_buf, insert_wbuf);
			(void) EntInsertText(w, insert_wbuf, insert_num_chars);
		}
		if (insert_wbuf)
			free(insert_wbuf);
        break;
    default:
        break; /* unimpl */
    }

    ElfRetData = retP;
    return;
}


/***********************************************************************
function: SetWidgetString - Set string on widget.
*/
SetWidgetString(w, str)
AXWidget_entry w;
mbchar *str;
{
    register int ix;
    register int real_len;
    register wchar *p;
    register wchar *p2;
    int len;
	wchar *tmpstr;

    if (!w)
        return(0);
    if (!str)
        str = (mbchar *) "";
	
	len = MBstrlen(str);
	tmpstr = (wchar *) malloc((len + 1) * sizeof(wchar));
	real_len = MbnToWStr(str, tmpstr, len);
	
	if (w->edit_win.editmask) {
		EditMaskOverwriteChars(w->edit_win.editmask, w->wstr, tmpstr,
							   real_len, 0);
	}
	else {
		if (w->wstr != (wchar *) NULL)
			free(w->wstr);
		w->wstr = tmpstr;
		/* eliminate any non-valid characters */
		p = w->wstr;
		p2 = w->wstr;
		for (ix = 0; ix < real_len; ix++) {
			if (w->validChars && !strchr(w->validChars, *p)) {
				p2++; /* skip this invalid char */
			}
			else {
				*p = *p2;
				p++;
				p2++;
			}
		}
		*p = 0;
	}

#if AX_X11
	/* validate cursorpos, selection indices against new text
	 */
	EditWinAssignCursorPos(&w->edit_win, w->edit_win.cursor_pos);
	w->edit_win.sel.startIndex = w->edit_win.sel.endIndex = len;
	if (w->edit_win.text) {
		w->edit_win.text_len = Wstrlen(w->edit_win.text);
	}
	else {
		w->edit_win.text_len = 0;
	}
#endif

    return(0);
}



/***********************************************************************
function: EntAdjustCursorPos - Adjust `w->curPos' to sane value (if necessary).
*/
void
EntAdjustCursorPos(w)
AXWidget_entry w;
{
    if (!w || w->type != W_ENTER)
        return;
    if (!w->wstr)
        return;

    if ((short)Wstrlen(w->wstr) > w->maxLen) {
        /* clip */
        w->wstr[w->maxLen] = 0;
        w->curPos = MIN(w->curPos, w->maxLen);
    }
    return;
}


/***********************************************************************
function: EntInsertText - Insert text WBUF at cursor position.
Return TRUE if entry text buffer changed; FALSE otherwise.
*/
int
EntInsertText(w, wbuf, num_chars)
AXWidget_entry w;
register wchar *wbuf;
register int num_chars;
{
    register int   ix;
    wchar *new_wstr = (wchar *) NULL;
    wchar *p = (wchar *) NULL;
    int start_to_cursor, chars_del_before_cursor;
    int real_len;
    int changed = FALSE;
    AXWidget_main main_w;
	
	if (!num_chars) {
		return (FALSE);
	}
	
	EntDoPendingDelete(w, FALSE);

    if ((num_chars + Wstrlen(w->wstr)) > w->maxLen) {
        RingTheBell();
        return(TRUE);
    }

#if BUILD_ORIENT
    main_w = TheMainWidget( w );
    if (main_w->conv_change_end > main_w->conv_change_start) {
        /* Take out the characters in the conversion zone, by copying
           the string back on itself.  These chars will be replaced in
           the string.  We adjust the cursor position to reflect the
           deletion */
        Wstrcpy((w->wstr + main_w->conv_change_start),
                (w->wstr + main_w->conv_change_end));
        start_to_cursor = w->curPos - main_w->conv_change_start;
        chars_del_before_cursor = MIN(start_to_cursor,
                                      (main_w->conv_change_end -
                                       main_w->conv_change_start));
        w->curPos = w->curPos - chars_del_before_cursor;
    }
#endif
	
	if (w->curPos < 0) {
		w->curPos = 0;
	}
	
	if (w->edit_win.editmask) {
		int num_accepted = EditMaskTestChars(w->edit_win.editmask,
											 w->wstr, wbuf, num_chars,
											 w->curPos);
		if (num_accepted != num_chars) {
			RingTheBell();
		}
		if (num_accepted > 0) {
			EditMaskOverwriteChars(w->edit_win.editmask, w->wstr,
								   wbuf, num_chars, w->curPos);
			for (ix = 0; ix < num_accepted; ix++) {
				w->curPos = 
					EditMaskNextHigherIndex(w->edit_win.editmask, w->curPos);
			}
			changed = TRUE;
		}
	}
	else {
		new_wstr = (wchar *) malloc((Wstrlen(w->wstr) + num_chars + 1) *
									sizeof(wchar));
		Wstrncpy(new_wstr, w->wstr, w->curPos); /* chars before cursor */
		p = new_wstr + w->curPos;
		real_len = 0;
		
		/* Add on the new characters */
		for (ix = 0; ix < num_chars; ix++) {
#if BUILD_ORIENT
			/* Ashisuto added for ORIENT build. 95.09.14 */
			if (w->validChars && !MbWstrchr(w->validChars, *wbuf)) 
#else
				/* this char is printable */
				if (w->validChars && !strchr(w->validChars, *wbuf)) 
#endif
					{
						RingTheBell();
						continue;
					}
			*p = *wbuf;
			p++;
			wbuf++;
			real_len++;
			changed = TRUE;
		}

		Wstrcpy(p, w->wstr + w->curPos); /* append chars after cursor */
		free(w->wstr);
		w->wstr = new_wstr;
		w->curPos = w->curPos + real_len;
	}
	
    if (changed) {
        EntAdjustCursorPos(w);
		w->update_widget = WidgetNeedsUpdated = TRUE;
        if (w->callback && (w->flags & WFLG_TOUCHED_RETURN))
            AxCallback(w->callback, w, w->callbacktag);
    }
    return(changed);
}

void
EntDoPendingDelete(w, do_callback)
AXWidget_entry w;
int do_callback;
{
#if AXMS
    DWORD start_char, end_char;
	
    MS_SendMessage(w->winid, EM_GETSEL, (WPARAM)&start_char,
				   (LPARAM)&end_char);
	if (start_char != end_char) {
		int sx, ex, new_len;
		wchar *new_wstr;
		
		if (start_char < end_char) {
			ex = end_char;
			sx = start_char;
		}
		else {
			sx = end_char;
			ex = start_char;
		}
		new_len = Wstrlen(w->wstr) - (ex - sx);
		new_wstr = (wchar *) malloc((new_len + 1) * sizeof(wchar));
		if (sx > 0) {
			Wstrncpy(new_wstr, w->wstr, sx);
		}
		Wstrncpy(&new_wstr[sx], &w->wstr[ex], new_len - sx);
		new_wstr[new_len] = (wchar) 0;
		free(w->wstr);
		w->wstr = new_wstr;
        w->curPos = sx;
		w->pending_delete = FALSE;
		
		if (do_callback) {
			w->update_widget = WidgetNeedsUpdated = TRUE;
			if (w->callback && (w->flags & WFLG_TOUCHED_RETURN)) {
				AxCallback(w->callback, w, w->callbacktag);
			}
		}
	}
#else
    if (w->edit_win.sel.startIndex != w->edit_win.sel.endIndex) {
		
		EditWinDoPendingDelete(&w->edit_win);
	
		w->pending_delete = FALSE;
		w->curPos = w->edit_win.sel.startIndex;
	
		w->wstr = (wchar *)
			realloc (w->wstr, (w->edit_win.text_len + 1) * sizeof(wchar));
		Wstrncpy(w->wstr, w->edit_win.text, w->edit_win.text_len);
		w->wstr[w->edit_win.text_len] = (wchar) 0;
	
		if (do_callback) {
			w->update_widget = WidgetNeedsUpdated = TRUE;
			if (w->callback && (w->flags & WFLG_TOUCHED_RETURN)) {
				AxCallback(w->callback, w, w->callbacktag);
			}
		}
	}
#endif
}

