/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: fontmap.c
   Applix ASC version: 43
   Date: Mon Nov 16 13:53:37 1998
*/
#include <stdio.h>
#include <afile.h>
#if ALPHA
#include <sys/access.h>
#endif
#include <axtext.h>
#include <fontmap.h>
#include <afile.h>

#define ESCCHAR ('\\')
#define QUOTECHAR ('"')
#define COMMACHAR (',')
#define EQUALCHAR ('=')
#define NEWLINECHAR ('\n')
#define OLD_COMMENTCHAR ('\'')
#define COMMENTCHAR ('%')       /* change comment char so we can run cpp */
                                /* over the fontmap.dir file */

#define WHITESPACE(c) (c <= ' ' ||  \
                       c == COMMACHAR || c == EQUALCHAR || \
                       c == OLD_COMMENTCHAR || c == COMMENTCHAR)
#define DONTCARE 0

    /* the "parse state" typedef used for reading fontmap.dir */
typedef struct pstate_t {
    unsigned int intext : 1;
    unsigned int incomment : 1;
    unsigned int inquotes : 1;
	/* defaultOnly means to read fontmap.dir only up to the
	 * first explicit FontRecord; that's enough to establish default record.
	 */
	unsigned int defaultOnly : 1;
    char lastchar;
    char *lastXblack;     /* @last transition into blackspace */
    void (*currState)();  /* handler called at enter-whitespace */
    fontTabEnt *currRec;
    fontTabEnt *defaultRec;
    int currAttr;
    char *filedata;
} pstate;

/* %%%%%%%%%%%%%%%%%%% put this global on the FontCache struct? */
#if AXMS
int AxtLetterformPixmaps = FALSE;
#else
int AxtLetterformPixmaps = TRUE;
#endif

extern FILE *getFontmetricsFile();
extern char *getErrString();
extern fontTabEnt *newFontRecord();
extern int freeFontRecord();
extern int copyFontRecord();
extern void loadFontRecord();
extern void setFontCacheLimit();
extern void axgfsInit();
extern FILE *axOpenFontFile();

extern char *DirSlash;
extern char *ThimApplixDir;    /* Where the THIMBLE executable lies */
 

void  completeCurrentRecord();
void  foldFontRecordDefaults();
char *MapFontFamily();

static fontTabEnt *hardDefaultFontRecord();
static void unionFontTabEntUse();
static void filteredDefaultRecord();

static int isBlack(), isAttr(), checkFontRecord();
static int usesAfmMetrics();
static char *isValue(), *strippedCopy();
static void toState(), lookingForAttr(), lookingForValue();
static void newValue();
static void loadDefaultFontRecord();
static void fontmapDirParse();


/* : axReadFontMapDirFile - initialize the FONTRECORD list in the
 * fontcache by reading the "fontmap.dir" file
 */
char *  /* OK if null; otherwise, caller must free return */
axReadFontMapDirFile(tablefile, screenRes, targetRes)
char *tablefile; /* typically a path to "fontmap.dir" */
int *screenRes, *targetRes;
{
    FILE *fmapdir;
    char *cp;
    int filesize;
    pstate ps;

	/* open the tablefile
	 */
    if (!(fmapdir = axOpenFontFile(tablefile))) {
        return(getErrString(XLT("Unable to open fontmap file named"),
                            tablefile));
    }

/* %%%%%%% supposed to read these from the fontmap.dir file, but.. */
    *screenRes = 75;
    *targetRes = 300;

	/* initialize pstate
	 */
    memset((char *) &ps, 0, sizeof(ps));
    ps.currState = lookingForAttr;
	
	/* read file contents into memory
	 */
    fseek(fmapdir, 0, 2);
    filesize = ftell(fmapdir);
    ps.filedata = (char *)malloc(filesize + 1);
    fseek(fmapdir, 0, 0);
    fread(ps.filedata, sizeof(char), (int)filesize, fmapdir);
    ps.filedata[filesize] = '\0';
    fclose(fmapdir);
	
	/* parse the file contents. This is done in two passes;
	 * the first pass is "defaultOnly", reading just enough
	 * to establish a default font record.  The second pass
	 * reads the rest.  Initialization of the Gallium fonts
	 * is done between passes.
	 */
	ps.defaultOnly = TRUE;
	cp = ps.filedata;
	fontmapDirParse(&ps, &cp, TRUE);
	axgfsInit();
	ps.defaultOnly = FALSE;
	fontmapDirParse(&ps, &cp, FALSE);
	
    if (ps.currRec) {
        fontTabEnt *tdefRecord;
        fontTabEntUse tdefUse;
		
		/* load last-completed font record into fontcache
		 */
        filteredDefaultRecord(ps.defaultRec, &ps.defaultRec->use,
                              &tdefRecord, &tdefUse);
        completeCurrentRecord(ps.currRec, &ps.currRec->use,
                              tdefRecord, &tdefUse);
        if (checkFontRecord(ps.currRec)) {
            loadFontRecord(ps.currRec);
        }
        if (tdefRecord) {
            freeFontRecord((char **) &tdefRecord);
        }
    }
    if (ps.defaultRec) {
        completeCurrentRecord(ps.defaultRec, &ps.defaultRec->use,
                              (fontTabEnt *) NULL,
                              (fontTabEntUse *) NULL);
    }
    loadDefaultFontRecord(ps.defaultRec, &ps.defaultRec->use);

	/* clean up pstate and return
	 */
    if (ps.currRec) {
        freeFontRecord((char **) &ps.currRec);
    }
    if (ps.defaultRec) {
        freeFontRecord((char **) &ps.defaultRec);
    }
    if (ps.filedata) {
        free((char *) ps.filedata);
    }
    return ((char *) NULL);
}

static void
fontmapDirParse(ps, cp_p, defaultOnly)
pstate *ps;                            /* in/out */
char **cp_p;                           /* in/out */
int defaultOnly;                       /* in */
{
	char *cp = *cp_p;
	
    while (*cp != '\0') {
        if (isBlack(ps, *cp)) {
            if (!ps->intext) {
				/* transition into blackspace
				 */
                ps->intext = TRUE;
                ps->lastXblack = cp;
            }
        }
        else {
            if (ps->intext) {
				/* transition into whitespace
				 */
                ps->currState(ps, cp);

				/* currState may have toggled defaultOnly,
				 * which ends the 1st pass. "back up" and return.
				 */
				if (ps->defaultOnly != defaultOnly) {
					*cp_p = ps->lastXblack;
					ps->intext = FALSE;
					ps->lastchar = *ps->lastXblack;
					return;
				}
				
                ps->intext = FALSE;
            }
        }
        if (*cp == QUOTECHAR && ps->lastchar != ESCCHAR) {
            ps->inquotes = !ps->inquotes;
        }
        ps->lastchar = *cp;
        cp++;
    }
	*cp_p = cp;
	return;
}
	
static void
lookingForAttr(p, cp)
pstate *p;
char *cp;
{
    int attr;
    if (isAttr(p, cp, &attr)) {
        p->currAttr = attr;
        toState(p, lookingForValue);
    }
}

static void
lookingForValue(p, cp)
pstate *p;
char *cp;
{
    char *value;
    int attr;
    if (isAttr(p, cp, &attr)) {
        p->currAttr = attr;
    }
    else if (value = isValue(p, cp)) {
        newValue(p, value);
    }
}

static int
isBlack(p, ch)
pstate *p;
char ch;
{
    if (WHITESPACE(ch)) {
        if (!p->incomment) {
            if (((ch == COMMENTCHAR) || (ch == OLD_COMMENTCHAR)) && 
                 (p->lastchar != ESCCHAR) ) {
                p->incomment = TRUE;
            }
        }
        else if (p->incomment) {
            if (ch == NEWLINECHAR && p->lastchar != ESCCHAR) {
                p->incomment = FALSE;
            }
        }
        if (p->intext) {
            return (p->lastchar == ESCCHAR || p->inquotes);
        }
        return FALSE;
    }
    else {
        return (!p->incomment);
    }
}

static int
isAttr(p, cp, attCode)
pstate *p;
char *cp;
int *attCode;  /* out */
{
    char *attstring;
    int found = FALSE;

    if (!(attstring = strippedCopy(p, cp))) {
        *attCode = 0;
        return FALSE;
    }

    if (!strcmp(attstring, FFontRecord)) {
        *attCode = ATT_FONT_RECORD;
        found = TRUE;
    }
    else if (!strcmp(attstring, FFamily)) {
        *attCode = ATT_FAMILY;
        found = TRUE;
    }
    else if (!strcmp(attstring, FAliasFor)) {
        *attCode = ATT_ALIAS_FOR;
        found = TRUE;
    }
    else if (!strcmp(attstring, FScreenName)) {
        *attCode = ATT_SCREENNAME;
        found = TRUE;
    }
    else if (!strcmp(attstring, FRasterTolerance)) {
        *attCode = ATT_RASTER_TOLERANCE;
        found = TRUE;
    }
    else if (!strcmp(attstring, FEncodingFile)) {
        *attCode = ATT_ENCODING_FILE;
        found = TRUE;
    }
    else if (!strcmp(attstring, FInternalEncoding)) {
        *attCode = ATT_INTERNAL_ENCODING;
        found = TRUE;
    }
    else if (!strcmp(attstring, FMetricsFile)) {
        *attCode = ATT_METRICS_FILE;
        found = TRUE;
    }
    else if (!strcmp(attstring, FPostScriptPrintName)) {
        *attCode = ATT_POSTSCRIPT_PRINT_NAME;
        found = TRUE;
    }
    else if (!strcmp(attstring, FPCLPrintName)) {
        *attCode = ATT_PCL_PRINT_NAME;
        found = TRUE;
    }
    else if (!strcmp(attstring, FPCLAlternate)) {
        *attCode = ATT_PCL_ALTERNATE;
        found = TRUE;
    }
    else if (!strcmp(attstring, FPostScriptAlternate)) {
        *attCode = ATT_POSTSCRIPT_ALTERNATE;
        found = TRUE;
    }
    else if (!strcmp(attstring, FType1FontFileName)) {
        *attCode = ATT_TYPE1_FONT_FILE_NAME;
        found = TRUE;
    }
    else if (!strcmp(attstring, FEncodingName)) {
        *attCode = ATT_ENCODING_NAME;
        found = TRUE;
    }
    else if (!strcmp(attstring, FEndFontRecord)) {
        *attCode = ATT_END_FONT_RECORD;
        found = TRUE;
    }
    else if (!strcmp(attstring, FLetterformCacheSize)) {
        *attCode = ATT_LETTERFORM_CACHESIZE;
        found = TRUE;
    }
    else if (!strcmp(attstring, FScreenfontCacheSize)) {
        *attCode = ATT_SCREENFONT_CACHESIZE;
        found = TRUE;
    }
    else if (!strcmp(attstring, FMetricsCacheSize)) {
        *attCode = ATT_METRICS_CACHESIZE;
        found = TRUE;
    }
    else if (!strcmp(attstring, FPrinterResolution)) {
        *attCode = ATT_PRINTER_RESOLUTION;
        found = TRUE;
    }
    else if (!strcmp(attstring, FScreenResolution)) {
        *attCode = ATT_SCREEN_RESOLUTION;
        found = TRUE;
    }
    else if (!strcmp(attstring, FWeight)) {
        *attCode = ATT_WEIGHT;
        found = TRUE;
    }
    else if (!strcmp(attstring, FSlant)) {
        *attCode = ATT_SLANT;
        found = TRUE;
    }
    else if (!strcmp(attstring, FWindowsAlternate)) {
        *attCode = ATT_WINDOWS_ALTERNATE;
        found = TRUE;
    }
    else if (!strcmp(attstring, FXAlternate)) {
        *attCode = ATT_X_ALTERNATE;
        found = TRUE;
    }
    else if (!strcmp(attstring, FLargeXAlternate)) {
        *attCode = ATT_LARGE_X_ALTERNATE;
        found = TRUE;
    }
    else if (!strcmp(attstring, FToleranceLowCutoff)) {
        *attCode = ATT_TOLERANCE_LOW_CUTOFF;
        found = TRUE;
    }
    else if (!strcmp(attstring, FToleranceHighCutoff)) {
        *attCode = ATT_TOLERANCE_HIGH_CUTOFF;
        found = TRUE;
    }
    else if (!strcmp(attstring, FFixedWidth)) {
        *attCode = ATT_FIXED_WIDTH;
        found = TRUE;
    }
    else if (!strcmp(attstring, FLetterformPixmaps)) {
        *attCode = ATT_LETTERFORM_PIXMAPS;
        found = TRUE;
    }
#if BUILD_KANJI
    else if (!strcmp(attstring, FSans)) {
        *attCode = ATT_SANS;
        found = TRUE;
    }
    else if (!strcmp(attstring, FLIPSPrintName)) {
        *attCode = ATT_LIPS_PRINT_NAME;
        found = TRUE;
    }
#endif

    else {
        *attCode = 0;
    }
    free((char *) attstring);
    return found;
}

static char *    /* caller must free nonnull return */
isValue(p, cp)
pstate *p;
char *cp;
{
    char *valstring;

    if (!(valstring = strippedCopy(p, cp))) {
        return ((char *) NULL);
    }
    else
        return valstring;
}

    /* : strippedCopy - make a copy of the string data between
     * p->lastXblack (inclusive) and cp, stripping out escape
     * chars, unescaped quotes, etc.
     */
static char *    /* caller must free nonnull return */
strippedCopy(p, cp)
pstate *p;
char *cp;
{
    char *scopy, *copyp, *tcp, lastchar;
    int len = cp - p->lastXblack;

    if (len <= 0)
        return ((char *) NULL);

        /* copy the input using these rules:
         *  - QUOTECHARs and ESCCHARs are stripped unless escaped.
         *  - newlines are stripped unconditionally..
         *  - nothing is expanded; memory alloc assumes this
         */
    scopy = (char *) malloc (len + 1);
    memset((char *) scopy, 0, len + 1);
    lastchar = '\0';
    for (tcp = p->lastXblack, copyp = scopy; tcp < cp; tcp++) {
        if ((*tcp == QUOTECHAR || *tcp == ESCCHAR) &&
            lastchar != ESCCHAR) {
            lastchar = *tcp;
            continue;
        }
        if (*tcp == '\n') {
            lastchar = *tcp;
            continue;
        }
        lastchar = *tcp;
        *copyp = *tcp;
        copyp++;
    }
    return scopy;
}

static void
toState(p, stateFunc)
pstate *p;
void (*stateFunc)();
{
    p->currState = stateFunc;
}

    /* put new value onto a font record */
static void
newValue(p, value)
pstate *p;
char *value;
{
    fontTabEnt *fontRec;
    fontTabEntUse *used;
    int tmp, freeValue = FALSE;
	
    if (p->currAttr == ATT_FONT_RECORD ||
        p->currAttr == ATT_END_FONT_RECORD) {
            /* these two atts pertain to the current font
             * record; either start a new one, or null it
             */
        if (p->currRec) {
            fontTabEnt *tdefRecord;
            fontTabEntUse tdefUse;
                /* load last-completed font record into fontcache
                 */
            filteredDefaultRecord(p->defaultRec, &p->defaultRec->use,
                                  &tdefRecord, &tdefUse);
            completeCurrentRecord(p->currRec, &p->currRec->use,
                                  tdefRecord, &tdefUse);
            if (checkFontRecord(p->currRec)) {
                loadFontRecord(p->currRec);
            }
            freeFontRecord((char **) &p->currRec);
            if (tdefRecord) {
                freeFontRecord((char **) &tdefRecord);
            }
			if (p->defaultOnly && p->currAttr == ATT_FONT_RECORD) {
				/* we've hit the first explicit font record.  end pass.
				 */
				p->defaultOnly = FALSE;
				p->currRec = (fontTabEnt *) NULL;
				return;
			}
        }
        if (p->currAttr == ATT_FONT_RECORD) {
            p->currRec = newFontRecord();
            p->currRec->globalName = value;
            p->currRec->use.globalName = TRUE;
        }
        else {
            p->currRec = (fontTabEnt *) NULL;
            free((char *) value);
        }
        return;
    }
    else if (p->currAttr == ATT_NEW_DEFAULT_RECORD) {
             /* do a 'reset' of the default record
              */
        freeFontRecord((char **) &p->defaultRec);
        p->defaultRec = (fontTabEnt *) NULL;
        free((char *) value);
        return;
    }

        /* the rest of the atts are either font record attributes,
           which will apply to the current or default record, or
           overall cache attributes */

        /* attrs are assigned either to the 'current record' if
         * there is one, or to the default record if no current
         */
    if (p->currRec) {
        fontRec = p->currRec;
        used = &p->currRec->use;
    }
    else {
        if (!p->defaultRec) {
            p->defaultRec = newFontRecord();
        }
        fontRec = p->defaultRec;
        used = &p->defaultRec->use;
    }

    switch(p->currAttr) {
        
    case ATT_FAMILY:
        if (fontRec == p->defaultRec) {
            break; /* defined as "axDefaultFontFamily" for the
                      default font record */
        }
        if (fontRec->familyName) {
            free((char *) fontRec->familyName);
        }
#if AXMS && !FOR_FILTER
        /* the Windows names, as enumerated from the postscript driver, */
        /* differ from the names in the fontmap.dir file.  So, to make  */
        /* things match up, we convert the family name to the Windows name. */
        {  
        char *new_value;
        new_value = MapFontFamily( value, UNIX_TO_WIN );
        if ( new_value != value )
           {
           /* name was changed.  Free the old and create a new name */
           /* since MapFontFamily returns a pointer to static storage */
           free( value );
           value = (char *)malloc( strlen(new_value)+1 );
           strcpy( value, new_value );
           }
        }
#endif



        fontRec->familyName = value;
        used->familyName = TRUE;
        break;
        
    case ATT_ALIAS_FOR:
        if (fontRec->aliasFor) {
            free((char *) fontRec->aliasFor);
        }
        fontRec->aliasFor = value;
        used->aliasFor = TRUE;
        break;
        
    case ATT_SCREENNAME:
            /* currently, the only supported name format is
             * the well-formed XLFD name, so go right onto xNames..
             */
        fontRec->xNameCount++;
        if (fontRec->xNames) {
            fontRec->xNames = (char **) realloc((char **) fontRec->xNames,
                    fontRec->xNameCount * sizeof(char *));
            fontRec->xNames[fontRec->xNameCount - 1] = value;
        }
        else {
            fontRec->xNames = (char **) malloc (sizeof(char *));
            fontRec->xNames[0] = value;
            fontRec->xNameCount = 1;
        }
        used->xNameCount = used->xNames = TRUE;
        break;
        
    case ATT_RASTER_TOLERANCE:
        fontRec->rasterTolerance = atoi(value);
        used->rasterTolerance = TRUE;
        freeValue = TRUE;
        break;
        
    case ATT_TOLERANCE_LOW_CUTOFF:
        fontRec->toleranceLowCutoff = atoi(value);
        used->toleranceLowCutoff = TRUE;
        freeValue = TRUE;
        break;
        
    case ATT_TOLERANCE_HIGH_CUTOFF:
        fontRec->toleranceHighCutoff = atoi(value);
        used->toleranceHighCutoff = TRUE;
        freeValue = TRUE;
        break;
        
    case ATT_ENCODING_FILE:
        if (fontRec->encodingFileName) {
            free((char *) fontRec->encodingFileName);
        }
        fontRec->encodingFileName = value;
        used->encodingFileName = TRUE;
        break;
        
    case ATT_INTERNAL_ENCODING:
        if (fontRec->internalEncodingName) {
            free((char *) fontRec->internalEncodingName);
        }
        fontRec->internalEncodingName = value;
        used->internalEncodingName = TRUE;
        break;
        
    case ATT_METRICS_FILE:
        tmp = fontRec->numTargetMetricsNames;
        fontRec->numTargetMetricsNames++;
        if (fontRec->targetMetricsNames) {
            fontRec->targetMetricsNames = (char **) realloc
                ((char **) fontRec->targetMetricsNames,
                 fontRec->numTargetMetricsNames * sizeof(char *));
            fontRec->targetMetricsNames[tmp] = value;
        }
        else {
            fontRec->targetMetricsNames = (char **) malloc
                (sizeof(char *));
            fontRec->targetMetricsNames[0] = value;
            fontRec->numTargetMetricsNames = 1;
        }
        used->numTargetMetricsNames = used->targetMetricsNames = TRUE;
        break;
        
    case ATT_POSTSCRIPT_PRINT_NAME:
        if (fontRec->PostScriptPrintName) {
            free((char *) fontRec->PostScriptPrintName);
        }
        fontRec->PostScriptPrintName = value;
        used->PostScriptPrintName = TRUE;
        break;
        
    case ATT_PCL_PRINT_NAME:
        if (fontRec->PCLPrintName) {
            free((char *) fontRec->PCLPrintName);
        }
        fontRec->PCLPrintName = value;
        used->PCLPrintName = TRUE;
        break;
        
    case ATT_PCL_ALTERNATE:
        if (fontRec->PCLAlternate) {
            free((char *) fontRec->PCLAlternate);
        }
        fontRec->PCLAlternate = value;
        used->PCLAlternate = TRUE;
        break;
        
    case ATT_POSTSCRIPT_ALTERNATE:
        if (fontRec->PostScriptAlternate) {
            free((char *) fontRec->PostScriptAlternate);
        }
        fontRec->PostScriptAlternate = value;
        used->PostScriptAlternate = TRUE;
        break;

#if BUILD_KANJI
    case ATT_SANS:
        fontRec->sans = atoi(value);
        used->sans = TRUE;
        freeValue = TRUE;
        break;

    case ATT_LIPS_PRINT_NAME:
        if (fontRec->LIPSPrintName) {
            free((char *) fontRec->LIPSPrintName);
        }
        fontRec->LIPSPrintName = value;
        used->LIPSPrintName = TRUE;
        break;
#endif
        
    case ATT_TYPE1_FONT_FILE_NAME:
        if (fontRec->Type1FontFileName) {
            free((char *) fontRec->Type1FontFileName);
        }
        fontRec->Type1FontFileName = value;
        used->Type1FontFileName = TRUE;
        break ;
        
    case ATT_WINDOWS_ALTERNATE:
        if (fontRec->WindowsAlternate) {
            free((char *) fontRec->WindowsAlternate);
        }
        fontRec->WindowsAlternate = value;
        used->WindowsAlternate = TRUE;
        break;
        
    case ATT_X_ALTERNATE:
        if (fontRec->XAlternate) {
            free((char *) fontRec->XAlternate);
        }
        fontRec->XAlternate = value;
        used->XAlternate = TRUE;
        break;
        
    case ATT_LARGE_X_ALTERNATE:
        if (fontRec->largeXAlternate) {
            free((char *) fontRec->largeXAlternate);
        }
        fontRec->largeXAlternate = value;
        used->largeXAlternate = TRUE;
        break;
        
    case ATT_ENCODING_NAME:
        if (fontRec->encodingName) {
            free((char *) fontRec->encodingName);
        }
        fontRec->encodingName = value;
        used->encodingName = TRUE;
        break;
        
    case ATT_LETTERFORM_CACHESIZE:
        setFontCacheLimit(LETTERFORM, atoi(value));
        freeValue = TRUE;
        break;
        
    case ATT_SCREENFONT_CACHESIZE:
        setFontCacheLimit(SCREEN_METRICS, atoi(value));
        freeValue = TRUE;
        break;
        
    case ATT_METRICS_CACHESIZE:
        setFontCacheLimit(TARGET_METRICS, atoi(value));
        freeValue = TRUE;
        break;
        
    case ATT_PRINTER_RESOLUTION:
/* %%%%%%%%%%%%%%% add support for this */
        break;
        
    case ATT_SCREEN_RESOLUTION:
/* %%%%%%%%%%%%%%% add support for this */
        break;
        
    case ATT_WEIGHT:
        fontRec->weight = atoi(value);
        used->weight = TRUE;
        freeValue = TRUE;
        break;
        
    case ATT_SLANT:
        fontRec->slant = atoi(value);
        used->slant = TRUE;
        freeValue = TRUE;
        break;
        
    case ATT_FIXED_WIDTH:
        fontRec->fixedWidth = atoi(value);
        used->fixedWidth = TRUE;
        freeValue = TRUE;
        break;
        
    case ATT_LETTERFORM_PIXMAPS:
/* %%%%%%%%%%%%%%% add this global to the FontCache struct */
#if AX_X11
        AxtLetterformPixmaps = atoi(value);
#endif
        freeValue = TRUE;
        break;
        
    }
    if (freeValue) {
        free((char *) value);
    }
}


    /* check the consistency of a to-be-loaded font record.
     */
static int
checkFontRecord(frec)
fontTabEnt *frec;            /* in/out */
{
    if (!frec) {
        return FALSE;
    }
    if (!frec->globalName || !frec->familyName) {
        return FALSE;    /* these two are needed for doing lookups */
    }
    if (!frec->aliasFor && !frec->targetMetricsNames) {
        
            /* supply default target metrics name (only records
             * declaring aliasFor are allowed to have no target
             * metrics). use Times-Roman....
             */
        frec->targetMetricsNames = (char **) malloc (sizeof(char *));
        frec->targetMetricsNames[0] = (char *)
            malloc (strlen(NXLT("tmr.afm")) + 1);
        strcpy(frec->targetMetricsNames[0], NXLT("tmr.afm"));
        frec->numTargetMetricsNames = 1;
        frec->use.targetMetricsNames = frec->use.numTargetMetricsNames = TRUE;
    }
	if ( (usesAfmMetrics(frec) || frec->PostScriptPrintName) &&
		(!frec->encodingFileName ||
		 !strcmp(frec->encodingFileName, NXLT("unspecified"))) ) {

		/* this record depends on an .afm file, and/or a non-downloaded
		 * postscript printer font, and yet the encodingFileName is null
		 * or unknown -- we must supply something; use latin1.enc as default.
		 */
        frec->encodingFileName = (char *)
            malloc (strlen(NXLT("latin1.enc")) + 1);
        strcpy(frec->encodingFileName, NXLT("latin1.enc"));
        frec->use.encodingFileName = TRUE;
    }
    return TRUE;
}

static int
usesAfmMetrics(fontRec)
fontTabEnt *fontRec;
{
    int ix, len;
    char *cp;
    
    for (ix = 0; ix < fontRec->numTargetMetricsNames; ix++) {
            /* look at the extension on the metrics file name
             */
        len = strlen(fontRec->targetMetricsNames[ix]);
        if (len >= 4) {
            cp = fontRec->targetMetricsNames[ix] + (len - 4);
            if (strncmp(cp, NXLT(".afm"), len) == 0) {
                return (TRUE);
                break;
            }
        }
    }
    return (FALSE);
}

static void
loadDefaultFontRecord(defRec, defUsed)
fontTabEnt *defRec;
fontTabEntUse *defUsed;
{
    fontTabEnt *hardDefault;
    fontTabEntUse hardUse;

        /* first write the hardcoded defaults */
    hardDefault = hardDefaultFontRecord(&hardUse);
    if (defRec) {
            /* fold hardcoded onto default and load completed default */
        foldFontRecordDefaults(hardDefault, &hardUse, defRec, defUsed);
        if (checkFontRecord(defRec)) {
            loadFontRecord(defRec);
        }
    }
    else {
            /* load hard defaults as the default font record */
        if (checkFontRecord(hardDefault)) {
            loadFontRecord(hardDefault);
        }
    }
    freeFontRecord((char **) &hardDefault);
}




static void
unionFontTabEntUse(in1, in2, outUse)
fontTabEntUse *in1, *in2; /* take the union of these two */
fontTabEntUse *outUse;    /* and write onto this */
{
        /* this has to be kept uptodate with definition of
           fontTabEntUse */
    memset((char *) outUse, 0, sizeof(fontTabEntUse));

    if (in1->xNamesKnown || in2->xNamesKnown)
        outUse->xNamesKnown = TRUE;
    
    if (in1->scalable || in2->scalable)
        outUse->scalable = TRUE;
    
    if (in1->fixedWidth || in2->fixedWidth)
        outUse->fixedWidth = TRUE;
    
    if (in1->isWindowsFont || in2->isWindowsFont)
        outUse->isWindowsFont = TRUE;

    if (in1->isPrinterFont || in2->isPrinterFont)
        outUse->isPrinterFont = TRUE;

    if (in1->globalName || in2->globalName)
        outUse->globalName = TRUE;
    
    if (in1->familyName || in2->familyName)
        outUse->familyName = TRUE;
    
    if (in1->aliasFor || in2->aliasFor)
        outUse->aliasFor = FALSE;
    
    if (in1->encodingWidth || in2->encodingWidth)
        outUse->encodingWidth = TRUE;
    
    if (in1->rasterTolerance || in2->rasterTolerance)
        outUse->rasterTolerance = TRUE;
    
    if (in1->toleranceHighCutoff || in2->toleranceHighCutoff)
        outUse->toleranceHighCutoff = TRUE;
    
    if (in1->toleranceLowCutoff || in2->toleranceLowCutoff)
        outUse->toleranceLowCutoff = TRUE;
    
    if (in1->weight || in2->weight)
        outUse->weight = TRUE;
    
    if (in1->slant || in2->slant)
        outUse->slant = TRUE;
    
    if (in1->numTargetMetricsNames || in2->numTargetMetricsNames) {
        outUse->targetMetricsNames = TRUE;
        outUse->numTargetMetricsNames = TRUE;
    }
    
    if (in1->xNameCount || in2->xNameCount) {
        outUse->xNameCount = TRUE;
        outUse->xNames = TRUE;
    }
    
    if (in1->scalableName || in2->scalableName)
        outUse->scalableName = TRUE;
    
    if (in1->PostScriptPrintName || in2->PostScriptPrintName)
        outUse->PostScriptPrintName = TRUE;
    
    if (in1->PCLPrintName || in2->PCLPrintName)
        outUse->PCLPrintName = TRUE;
    
    if (in1->PostScriptAlternate || in2->PostScriptAlternate)
        outUse->PostScriptAlternate = TRUE;
    
    if (in1->PCLAlternate || in2->PCLAlternate)
        outUse->PCLAlternate = TRUE;
    
#if BUILD_KANJI
    if (in1->sans || in2->sans)
        outUse->sans = TRUE;

    if (in1->LIPSPrintName || in2->LIPSPrintName)
        outUse->LIPSPrintName = TRUE;
#endif

    if (in1->WindowsAlternate || in2->WindowsAlternate)
        outUse->WindowsAlternate = TRUE;
    
    if (in1->XAlternate || in2->XAlternate)
        outUse->XAlternate = TRUE;
    
    if (in1->Type1FontFileName || in2->Type1FontFileName)
        outUse->Type1FontFileName = TRUE;
    
    if (in1->encodingFileName || in2->encodingFileName)
        outUse->encodingFileName = TRUE;
    
    if (in1->encodingName || in2->encodingName)
        outUse->encodingName = TRUE;

    if (in1->sizeEM || in2->sizeEM)
        outUse->sizeEM = TRUE;
    
    if (in1->avgWidth || in2->avgWidth)
        outUse->avgWidth = TRUE;
}


void
completeCurrentRecord(currRec, whichCurrent, defaultRec, whichDefault)
fontTabEnt *currRec;           /* out: currentRecord to be completed */
fontTabEntUse *whichCurrent;   /* in: which currRec members valid */
fontTabEnt *defaultRec;        /* in: defaults to fold if present */
fontTabEntUse *whichDefault;   /* in: which default members valid */
{
    fontTabEnt *tdef, *hard;
    fontTabEntUse hardUse, tmpUse;

        /* get a copy of the hardcoded defaults */
    hard = hardDefaultFontRecord(&hardUse);
    if (defaultRec) {
            /* get a copy of the default font record */
        tdef = (fontTabEnt *) NULL;  /* must do this prior to copy */
        copyFontRecord((char *) DONTCARE, (char **) &tdef,
                       (char *) DONTCARE, (char *) defaultRec);
            /* fold hardcoded onto this copy */
        foldFontRecordDefaults(hard, &hardUse, tdef,
                                   whichDefault);
        unionFontTabEntUse(&hardUse, whichDefault, &tmpUse);
            /* then fold the result onto the current record */
        foldFontRecordDefaults(tdef, &tmpUse,
                               currRec, whichCurrent);
        freeFontRecord((char **) &tdef);
    }
    else {
            /* fold hardcoded directly onto current */
        foldFontRecordDefaults(hard, &hardUse, currRec,
                               whichCurrent);
    }
    freeFontRecord((char **) &hard);
}




    /* fill out the possibly sparse fontTabEnt *onto,
     * using information from default fontTabEnt *defRec
     */
void
foldFontRecordDefaults(defRec, whichDef, onto, whichOnto)
fontTabEnt *defRec;           /* in: default font record */
fontTabEntUse *whichDef;      /* in: which default members set */
fontTabEnt *onto;             /* out: fill this out */
fontTabEntUse *whichOnto;     /* in: which onto members set explicitly */
{
    int len, ix;

    if (!defRec || !whichDef || !onto || !whichOnto)
        return;

        /* do the string members
         */
    
    if (defRec->globalName && whichDef->globalName &&
        !whichOnto->globalName) {
        
        if (onto->globalName) {
            free(onto->globalName);
        }
        len = strlen(defRec->globalName) + 1;
        onto->globalName = (char *) malloc(len);
        strcpy(onto->globalName, defRec->globalName);
    }
    
    if (defRec->familyName && whichDef->familyName &&
        !whichOnto->familyName) {
        
        if (onto->familyName) {
            free(onto->familyName);
        }
        len = strlen(defRec->familyName) + 1;
        onto->familyName = (char *) malloc(len);
        strcpy(onto->familyName, defRec->familyName);
    }
    
    if (defRec->aliasFor && whichDef->aliasFor &&
        !whichOnto->aliasFor) {
        
        if (onto->aliasFor) {
            free(onto->aliasFor);
        }
        len = strlen(defRec->aliasFor) + 1;
        onto->aliasFor = (char *) malloc(len);
        strcpy(onto->aliasFor, defRec->aliasFor);
    }

    if (!onto->gallium) {
/* %%%%%%% must preserve gallium-style reference to target metrics; this
 * %%%%%%% is because letterform access is merged with metrics access;
 * %%%%%%% this could be done differently..  for now, disallow overload
 */
		if (defRec->numTargetMetricsNames && whichDef->numTargetMetricsNames &&
			!whichOnto->numTargetMetricsNames) {
        
			if (onto->numTargetMetricsNames && onto->targetMetricsNames) {
				for (ix = 0; ix < onto->numTargetMetricsNames; ix++) {
					if (onto->targetMetricsNames[ix]) {
						free(onto->targetMetricsNames[ix]);
					}
				}
				free(onto->targetMetricsNames);
			}
			onto->numTargetMetricsNames = defRec->numTargetMetricsNames;
			onto->targetMetricsNames = (char **) malloc
				(onto->numTargetMetricsNames * sizeof(char *));
			for (ix = 0; ix < onto->numTargetMetricsNames; ix++) {
				len = strlen(defRec->targetMetricsNames[ix]) + 1;
            onto->targetMetricsNames[ix] = (char *) malloc(len);
				strcpy(onto->targetMetricsNames[ix],
					   defRec->targetMetricsNames[ix]);
			}
		}
	}
    
    if (defRec->xNameCount && whichDef->xNameCount &&
        !whichOnto->xNameCount) {

        if (onto->xNameCount && onto->xNames) {
            for (ix = 0; ix < onto->xNameCount; ix++) {
                if (onto->xNames[ix]) {
                    free(onto->xNames[ix]);
                }
            }
            free(onto->xNames);
        }
        onto->xNameCount = defRec->xNameCount;
        onto->xNames = (char **) malloc (onto->xNameCount * sizeof
                                         (char *));
        for (ix = 0; ix < onto->xNameCount; ix++) {
            len = strlen(defRec->xNames[ix]) + 1;
            onto->xNames[ix] = (char *) malloc(len);
            strcpy(onto->xNames[ix], defRec->xNames[ix]);
        }
    }
    
    if (defRec->scalableName && whichDef->scalableName &&
        !whichOnto->scalableName) {
        
        if (onto->scalableName) {
            free(onto->scalableName);
        }
        len = strlen(defRec->scalableName) + 1;
        onto->scalableName = (char *) malloc(len);
        strcpy(onto->scalableName, defRec->scalableName);
    }
    
    if (defRec->PostScriptPrintName && whichDef->PostScriptPrintName &&
        !whichOnto->PostScriptPrintName) {
        
        if (onto->PostScriptPrintName) {
            free(onto->PostScriptPrintName);
        }
        len = strlen(defRec->PostScriptPrintName) + 1;
        onto->PostScriptPrintName = (char *) malloc(len);
        strcpy(onto->PostScriptPrintName, defRec->PostScriptPrintName);
    }
    
    if (defRec->PCLPrintName && whichDef->PCLPrintName &&
        !whichOnto->PCLPrintName) {
        
        if (onto->PCLPrintName) {
            free(onto->PCLPrintName);
        }
        len = strlen(defRec->PCLPrintName) + 1;
        onto->PCLPrintName = (char *) malloc(len);
        strcpy(onto->PCLPrintName, defRec->PCLPrintName);
    }
    
    if (defRec->PostScriptAlternate && whichDef->PostScriptAlternate &&
        !whichOnto->PostScriptAlternate) {
        
        if (onto->PostScriptAlternate) {
            free(onto->PostScriptAlternate);
        }
        len = strlen(defRec->PostScriptAlternate) + 1;
        onto->PostScriptAlternate = (char *) malloc(len);
        strcpy(onto->PostScriptAlternate, defRec->PostScriptAlternate);
    }
    
    if (defRec->PCLAlternate && whichDef->PCLAlternate &&
        !whichOnto->PCLAlternate) {
        
        if (onto->PCLAlternate) {
            free(onto->PCLAlternate);
        }
        len = strlen(defRec->PCLAlternate) + 1;
        onto->PCLAlternate = (char *) malloc(len);
        strcpy(onto->PCLAlternate, defRec->PCLAlternate);
    }

#if BUILD_KANJI
    if (whichDef->sans && !whichOnto->sans) {
        onto->sans = defRec->sans;
    }

    if (defRec->LIPSPrintName && whichDef->LIPSPrintName && 
                                              !whichOnto->LIPSPrintName ) {
        if (onto->LIPSPrintName )
            free(onto->LIPSPrintName);

        len = strlen(defRec->LIPSPrintName) + 1;
        onto->LIPSPrintName = (char *) malloc(len);
        strcpy(onto->LIPSPrintName, defRec->LIPSPrintName);
    }
#endif
    
    if (defRec->WindowsAlternate && whichDef->WindowsAlternate &&
        !whichOnto->WindowsAlternate) {
        
        if (onto->WindowsAlternate) {
            free(onto->WindowsAlternate);
        }
        len = strlen(defRec->WindowsAlternate) + 1;
        onto->WindowsAlternate = (char *) malloc(len);
        strcpy(onto->WindowsAlternate, defRec->WindowsAlternate);
    }
    if (defRec->XAlternate && whichDef->XAlternate &&
        !whichOnto->XAlternate) {
        
        if (onto->XAlternate) {
            free(onto->XAlternate);
        }
        len = strlen(defRec->XAlternate) + 1;
        onto->XAlternate = (char *) malloc(len);
        strcpy(onto->XAlternate, defRec->XAlternate);
    }

    if (defRec->largeXAlternate && whichDef->largeXAlternate &&
        !whichOnto->largeXAlternate) {
        
        if (onto->largeXAlternate) {
            free(onto->largeXAlternate);
        }
        len = strlen(defRec->largeXAlternate) + 1;
        onto->largeXAlternate = (char *) malloc(len);
        strcpy(onto->largeXAlternate, defRec->largeXAlternate);
    }

    if (defRec->Type1FontFileName && whichDef->Type1FontFileName &&
        !whichOnto->Type1FontFileName) {
        
        if (onto->Type1FontFileName) {
            free(onto->Type1FontFileName);
        }
        len = strlen(defRec->Type1FontFileName) + 1;
        onto->Type1FontFileName = (char *) malloc(len);
        strcpy(onto->Type1FontFileName, defRec->Type1FontFileName);
    }
    
    if (defRec->encodingFileName && whichDef->encodingFileName &&
        !whichOnto->encodingFileName) {
        
        if (onto->encodingFileName) {
            free(onto->encodingFileName);
        }
        len = strlen(defRec->encodingFileName) + 1;
        onto->encodingFileName = (char *) malloc(len);
        strcpy(onto->encodingFileName, defRec->encodingFileName);
    }
    
    if (defRec->encodingName && whichDef->encodingName &&
        !whichOnto->encodingName) {
        
        if (onto->encodingName) {
            free(onto->encodingName);
        }
        len = strlen(defRec->encodingName) + 1;
        onto->encodingName = (char *) malloc(len);
        strcpy(onto->encodingName, defRec->encodingName);
    }
    
    if (defRec->internalEncodingName && whichDef->internalEncodingName &&
        !whichOnto->internalEncodingName) {
        
        if (onto->internalEncodingName) {
            free(onto->internalEncodingName);
        }
        len = strlen(defRec->internalEncodingName) + 1;
        onto->internalEncodingName = (char *) malloc(len);
        strcpy(onto->internalEncodingName, defRec->internalEncodingName);
    }

        /* now do the int members
         */
    
    if (whichDef->xNamesKnown && !whichOnto->xNamesKnown) {
        onto->xNamesKnown = defRec->xNamesKnown;
    }
    
    if (whichDef->scalable && !whichOnto->scalable) {
        onto->scalable = defRec->scalable;
    }
    
    if (whichDef->fixedWidth && !whichOnto->fixedWidth) {
        onto->fixedWidth = defRec->fixedWidth;
    }
    
    if (whichDef->isWindowsFont && !whichOnto->isWindowsFont) {
        onto->isWindowsFont = defRec->isWindowsFont;
    }

    if (whichDef->isPrinterFont && !whichOnto->isPrinterFont) {
        onto->isPrinterFont = defRec->isPrinterFont;
    }

    if (whichDef->gallium && !whichOnto->gallium) {
        onto->gallium = defRec->gallium;
    }
    
    if (whichDef->encodingWidth && !whichOnto->encodingWidth) {
        onto->encodingWidth = defRec->encodingWidth;
    }
    
    if (whichDef->rasterTolerance && !whichOnto->rasterTolerance) {
        onto->rasterTolerance = defRec->rasterTolerance;
    }
    
    if (whichDef->toleranceLowCutoff && !whichOnto->toleranceLowCutoff) {
        onto->toleranceLowCutoff = defRec->toleranceLowCutoff;
    }
    
    if (whichDef->toleranceHighCutoff && !whichOnto->toleranceHighCutoff) {
        onto->toleranceHighCutoff = defRec->toleranceHighCutoff;
    }
    
    if (whichDef->weight && !whichOnto->weight) {
        onto->weight = defRec->weight;
    }
    
    if (whichDef->slant && !whichOnto->slant) {
        onto->slant = defRec->slant;
    }
    
    if ( whichDef->sizeEM && !whichOnto->sizeEM) {
        onto->sizeEM = defRec->sizeEM;
    }
    
    if ( whichDef->avgWidth && !whichOnto->avgWidth) {
        onto->avgWidth = defRec->avgWidth;
    }

    if ( whichDef->charSet && !whichOnto->charSet) {
        onto->charSet = defRec->charSet;
    }

    if ( whichDef->pitchAndFamily && !whichOnto->pitchAndFamily) {
        onto->pitchAndFamily = defRec->pitchAndFamily;
    }
}


static fontTabEnt *   /* caller must free resultant font record */
hardDefaultFontRecord(whichUsed)
fontTabEntUse *whichUsed;  /* out: report which members set */
{
    fontTabEnt *hardDefault = newFontRecord();
    memset((char *)whichUsed, 0, sizeof(fontTabEntUse));

            /* the hardcoded defaults supply the names
             * "axDefault.." for fontFamilyName and globalName.
             * The screenfont and metrics names are for Times
             */
    hardDefault->globalName =
        (char *) malloc(strlen(NXLT("axDefaultFontRecord")) + 1);
	/* the existance of this check causes this function not to abort on ncr */
	/* I don't know why it just works -lance */
	if (hardDefault->globalName == (char *)NULL )
		 exit(-1);	
    strcpy(hardDefault->globalName, NXLT("axDefaultFontRecord"));
    whichUsed->globalName = TRUE;

    hardDefault->familyName =
        (char *) malloc(strlen(NXLT("axDefaultFontFamily")) + 1);
    strcpy(hardDefault->familyName, NXLT("axDefaultFontFamily"));
    whichUsed->familyName = TRUE;

    hardDefault->rasterTolerance = 25; /* decipoints */
    whichUsed->rasterTolerance = TRUE;
    hardDefault->toleranceLowCutoff = 6; /* points */
    whichUsed->toleranceLowCutoff = TRUE;
    hardDefault->toleranceHighCutoff = 36; /* points */
    whichUsed->toleranceHighCutoff = TRUE;

#if AX_X11  /*  {  */
    hardDefault->xNameCount = 6;  /* six times-roman sizes in X11R4 */
    whichUsed->xNameCount = TRUE;

        /* ideally, this should be the only place where
         * hardcoded xlfd strings are used..
         */
    hardDefault->xNames = (char **) malloc(6 * sizeof(char *));
    hardDefault->xNames[0] = (char *) malloc(strlen(AXDEFTIMES8) + 1);
    strcpy(hardDefault->xNames[0], AXDEFTIMES8);
    hardDefault->xNames[1] = (char *) malloc(strlen(AXDEFTIMES10) + 1);
    strcpy(hardDefault->xNames[1], AXDEFTIMES10);
    hardDefault->xNames[2] = (char *) malloc(strlen(AXDEFTIMES12) + 1);
    strcpy(hardDefault->xNames[2], AXDEFTIMES12);
    hardDefault->xNames[3] = (char *) malloc(strlen(AXDEFTIMES14) + 1);
    strcpy(hardDefault->xNames[3], AXDEFTIMES14);
    hardDefault->xNames[4] = (char *) malloc(strlen(AXDEFTIMES18) + 1);
    strcpy(hardDefault->xNames[4], AXDEFTIMES18);
    hardDefault->xNames[5] = (char *) malloc(strlen(AXDEFTIMES24) + 1);
    strcpy(hardDefault->xNames[5], AXDEFTIMES24);
    whichUsed->xNames = TRUE;

        /* supply Times metrics as a fallback
         */
    hardDefault->numTargetMetricsNames = 1;
    hardDefault->targetMetricsNames =
        (char **) malloc(sizeof(char *));
    hardDefault->targetMetricsNames[0] = 
                              (char *)malloc(strlen(NXLT("tmr.afm")) + 1);
    strcpy(hardDefault->targetMetricsNames[0], NXLT("tmr.afm"));
    whichUsed->targetMetricsNames = TRUE;

    
    hardDefault->encodingFileName =
        (char *) malloc(strlen(NXLT("latin1.enc")) + 1);
    strcpy(hardDefault->encodingFileName, NXLT("latin1.enc"));
    whichUsed->encodingFileName = TRUE;

#if BUILD_KANJI
    hardDefault->LIPSPrintName =
        (char *) malloc(strlen(NXLT("Dutch-Roman")) + 1);
    strcpy(hardDefault->LIPSPrintName, NXLT("Dutch-Roman"));
    whichUsed->LIPSPrintName = TRUE;
#endif

#else     /*  }  AX_X11  {  */

    /* Windows-NT  */
    hardDefault->numTargetMetricsNames = 1;
    hardDefault->targetMetricsNames =
        (char **) malloc(sizeof(char *));
    hardDefault->targetMetricsNames[0] = 
                               malloc(strlen(NXLT("Times New Roman.win"))+1);
    strcpy(hardDefault->targetMetricsNames[0], NXLT("Times New Roman.win"));
    whichUsed->targetMetricsNames = TRUE;
#endif    /*  }  */

    return hardDefault;
}

/* : filterDefaultRecord - return a filtered copy of the passed default
 * font record, plus usage bits.  this filtered copy is then passed to
 * completeCurrentRecord as part of the initialization of an explicit
 * font record.  we want to prevent certain attributes from being
 * transferred to an explicit record, e.g., alternate-font names.
 *
 * caller must free returned font record
 */
static void
filteredDefaultRecord(defaultRec, whichDefault, filtRec, whichFilt)
fontTabEnt *defaultRec;       /* in: pointer to default font record */
fontTabEntUse *whichDefault;  /* in: usage bits of default record */
fontTabEnt **filtRec;         /* out: alloc and return filtered copy */
fontTabEntUse *whichFilt;     /* out: return usage bits of default record */
{
    
    *filtRec = (fontTabEnt *) NULL;    /* init caller's pointer */
    
    if (defaultRec) {
            /* take straight copies at first
             */
        *whichFilt = *whichDefault;
        copyFontRecord((char *) DONTCARE, (char **) filtRec,
                       (char *) DONTCARE, (char *) defaultRec);

            /* remove certain att values from copy
             */
        
        if ((*filtRec)->PostScriptAlternate) {
            free((*filtRec)->PostScriptAlternate);
            (*filtRec)->PostScriptAlternate = (char *) NULL;
            whichFilt->PostScriptAlternate = FALSE;
        }
        if ((*filtRec)->PCLAlternate) {
            free((*filtRec)->PCLAlternate);
            (*filtRec)->PCLAlternate = (char *) NULL;
            whichFilt->PCLAlternate = FALSE;
        }
        if ((*filtRec)->XAlternate) {
            free((*filtRec)->XAlternate);
            (*filtRec)->XAlternate = (char *) NULL;
            whichFilt->XAlternate = FALSE;
        }
        if ((*filtRec)->WindowsAlternate) {
            free((*filtRec)->WindowsAlternate);
            (*filtRec)->WindowsAlternate = (char *) NULL;
            whichFilt->WindowsAlternate = FALSE;
        }
        if ((*filtRec)->PostScriptPrintName) {
            free((*filtRec)->PostScriptPrintName);
            (*filtRec)->PostScriptPrintName = (char *) NULL;
            whichFilt->PostScriptPrintName = FALSE;
        }
        if ((*filtRec)->PCLPrintName) {
            free((*filtRec)->PCLPrintName);
            (*filtRec)->PCLPrintName = (char *) NULL;
            whichFilt->PCLPrintName = FALSE;
        }
    }
}





#if !FOR_FILTER    /*   {  */


#define CACH_NAME_SIZE 40
typedef struct {
    char unix_name[CACH_NAME_SIZE+1];
    char win_name[CACH_NAME_SIZE+1];
} mapfont_struct;

static mapfont_struct *FontFamilyCache;
static int SizeFontFamilyCache;

static void StripWhiteSpace();
    
static void InitFontFamilyMap(  )
{

    static int cache_has_been_loaded = FALSE;
    char path[256], fontnames[256];
    int new_section_number;
    int size_cache;
    char *tempfont;
    FILE *fp;
    register mapfont_struct *cacheP = NULL;
    char *p;

    if ( cache_has_been_loaded )
        return;

    cache_has_been_loaded = TRUE;

    /* cache is empty; read winfont.map file and load cache. */
    /* first try the users local directory */
    sprintf( path, "%s%s%s", AGetAxHomeDir(), DirSlash, NXLT("winfont.map") );
    if ( Aaccess(path, F_OK) != AOK )
        {
        /* look in "axdata" */
        sprintf( path, "%s%s%s", ThimApplixDir, DirSlash, NXLT("winfont.map") );
        if ( Aaccess(path, F_OK) != AOK )            
            /* winfont file not found, no mapping */
            return;
        }
    
    fp = fopen( path, READ_STR );
    if ( fp == NULL )
        taskAbort( XLT("could not open winfont.map!") );

    /** cache will grow dynamically. **/
    size_cache = 4;          /* some starting value */
    FontFamilyCache = cacheP = (mapfont_struct *)TrueMalloc( 
                     size_cache * sizeof(*FontFamilyCache) );
    SizeFontFamilyCache = 0;

    while ( fgets(fontnames, sizeof(fontnames), fp) )
        {
        if ( fontnames[0] == '#' )
            continue;    /* ignore header */

        p = fontnames + strlen(fontnames);
        while ( (p >= fontnames) && ((unsigned char)(*p) <= ' ') )
             {
             *p = 0;          /* strip trailing blanks, NL, CR  */
             p--;
             }

        if ( fontnames[0] == '\0' )
             /* blank line */
             continue;

        if ( SizeFontFamilyCache >= size_cache )
             {
             /* need to grow the cach size */
             size_cache = size_cache + 4;
             FontFamilyCache = (mapfont_struct *)TrueRealloc( FontFamilyCache,
                            (size_cache * sizeof(*FontFamilyCache)) );
             cacheP = FontFamilyCache + SizeFontFamilyCache;
             }

        /* format of a file line is "<unix name>:<windows name>" */
        strcpy( cacheP->unix_name, "Times" );      /* failsafe defaults */
        strcpy( cacheP->win_name, "Times New Roman" );
        
        tempfont = strtok( fontnames, ":" );
        if ( tempfont != (char *)NULL )
           {
           StripWhiteSpace(tempfont);
           strncpy( cacheP->unix_name, tempfont, CACH_NAME_SIZE );
           }
        
        tempfont = strtok( NULL, "\n" ); 
        if ( tempfont != (char *)NULL )
           {
           StripWhiteSpace(tempfont);
           strncpy( cacheP->win_name, tempfont, CACH_NAME_SIZE );
           }
        SizeFontFamilyCache++;
        cacheP++;
        }

    fclose(fp);       
    return;
}




/**********************************************************************/
/** Since Windows and UNIX use different names for the same font.     */
/**   we need to convert family names as documents are read, or       */
/**   printing is done.  For example, on UNIX the user sees the font  */
/*    name "Bookman", but on NT the postscript driver enumerates      */
/*    this font as "ITC Bookman".  This function will map one name    */
/*    into another as documents get moved around.                     */
/*  Note: this name mapping ,must be done as a document is read, so   */
/*    that all the upper levels see only one font name.  Otherwise a  */
/*    use who edits a doc created on NT see in his font list both     */
/*    "Bookman" and "ITC Bookman".  The mappings in the font cache    */
/*    are not sufficient to avoid this problem, they only effect what */
/*    is eventually displayed, not what names the user sees           */
/*  The name returned is static, do not call free on it!              */
/**********************************************************************/

char *MapFontFamily( name_to_change, map_direction )
char *name_to_change;
int map_direction;
{
    int itemsize;
    mapfont_struct *cachP;
    int ix;

    InitFontFamilyMap( );

    if ( FontFamilyCache == (mapfont_struct *)NULL )
        /* no cache, no transformation */
        return( name_to_change );
    

    /* see if we have a mapping for the desired font */
    cachP = FontFamilyCache;
    for ( ix=0; ix < SizeFontFamilyCache; ix++ )
        {        
        if ( map_direction == UNIX_TO_WIN )
            {
            if ( *name_to_change == *(cachP->unix_name) )
                {
                /* first char is te same, check the whole name */
                if ( Astrcmp( cachP->unix_name, name_to_change) == 0 )
                    return( cachP->win_name );
                }
            }
        else
            {
            /* map UNIX name to windows name */
            if ( *name_to_change == *(cachP->win_name) )
                {
                /* first char is te same, check the whole name */
                if ( Astrcmp( cachP->win_name, name_to_change) == 0 )
                    return( cachP->unix_name );
                }
            }

        cachP++;
        }

    /* no match, no transformation */
    return( name_to_change );
}


/* called from FreeFontCache, when Applix is exited */
void FreeFontFamilyCache()
{
    if ( FontFamilyCache )
        TrueFree( FontFamilyCache );
}




/** strips unecessary white space from a start and end of string. *
 ** that is, this does not strip white space between text.*
 ** example:  "  Courier New  " would become "Courier New" **/
static void StripWhiteSpace(str)
char *str;
{
    char *p;
    
    /* strip leading white space */
    p = str;
    while ( ((unsigned char)(*p) <= ' ') && (*p != '\0') )
       {
       if ( (unsigned char)(*p) <= ' ' )
          strcpy( p, p+1 );
       }

    /* strip trailing white space */
    p = str + strlen(str);
    while ( (p >= str) && ((unsigned char)(*p) <= ' ') )
       {
       *p = '\0';
       p--;
       }
    return;
}



#endif        /* }   !FOR_FILTER  */
