/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: genop.c
   Applix ASC version: 5
   Date: Mon Nov 16 13:53:34 1998
*/
#include <stdio.h>
#include <aos.h>
#include <tmi.h>

#include <atmattr.h>
#include <oldwin.h>

#if SUN_G
#include <pixrect/pixrect_hs.h>
#endif


typedef struct { int x, y; } TmPoint32;
extern short ATMtrace;
extern short ATMdebug;
extern FILE *ATfp;
extern FILE *fd2;
extern short GrayBack;

/********** type definitions **********/

/***** Lrf() parameter packet *****/

typedef union
{
	struct
	{
		int funcode;
	} a;
	char dummy;
} LrfPacket;

/***** Lblt() parameter packet *****/

typedef struct
{
	A_Bitmap *abp;		/* -> argument bitmap */
	int f;			/* function code */
	short xoff;		/* x source-to-dest offset */
	short yoff;		/* y source-to-dest offset */
} LbltPack;

/***** globals *****/

extern Layer *FrontLayer, *BackLayer;

/********************************************************************/
/*
 * Lblt -- function applied by ATMlayerop() in lblt()
 */
Lblt(lp, rp, bp, ap, op)
	Layer *lp;
	AxRectangle *rp;
	A_Bitmap *bp;
	LbltPack *ap;
	Obscured *op;
{
	AxRectangle sr;		/* source rectangle */
	
	if (!getimage(bp, TRUE, FALSE)) return;	/* dest not available */
	/*
	 * - assume source image has been set up before the call
	 * - if source unavailable after getting dest, do nothing
	 */
	if (!ap->abp->inMemF) return;

	/* both are available */

	/* subtract offsets from dest rect to get source rect */

	sr.origin.x = rp->origin.x - ap->xoff;
	sr.origin.y = rp->origin.y - ap->yoff;
	sr.corner.x = rp->corner.x - ap->xoff;
	sr.corner.y = rp->corner.y - ap->yoff;
	ATMbitblt(ap->abp, &sr, bp, &rp->origin, ap->f, 0, 0);
}
/********************************************************************/
/*
 * lblt -- bitblt from off-screen Bitmap to a AxRectangle within a Layer
 */
lblt(lp, drp, sbp, sox, soy, f)
	Layer *lp;		/* -> destination Layer */
	A_Bitmap *sbp;		/* -> source Bitmap */
	AxRectangle *drp;		/* -> destination AxRectangle */
	short sox, soy;		/* source origin (x, y) */
	int f;			/* function code */
{
	LbltPack pack;
	/*
	 * Set up packet of "other args" for passing to Lblt by
	 * way of ATMlayerop().
	 * Compute source-to-destination offset.
	 */
	pack.abp = sbp;
	pack.f = f;
	pack.xoff = drp->origin.x - sox;	/* x offset */
	pack.yoff = drp->origin.y - soy;	/* y offset */

	ATMlayerop(lp, Lblt, drp, &pack);
}
/********************************************************************/
/*
 * LBblt -- function applied by ATMlayerop() in lbblt()
 */
LBblt(lp, rp, bp, ap, op)
	Layer *lp;
	AxRectangle *rp;		/* source rectangle */
	A_Bitmap *bp;			/* source bitmap */
	LbltPack *ap;
	Obscured *op;
{
	Point dest;		/* destination origin */

	/* subtract offsets from source rect to get dest rect */

	dest.x = rp->origin.x - ap->xoff;
	dest.y = rp->origin.y - ap->yoff;
	
	if (!getimage(bp, TRUE, FALSE)) return;	/* source unavailable */
	/*
	 * - assume dest image has been set up before the call
	 * - if dest unavailable after getting source, do nothing
	 */
	if (!ap->abp->inMemF) return;

	/* both are available */

	ATMbitblt(bp, rp, ap->abp, &dest, ap->f, 0, 0);
}
/********************************************************************/
/*
 * lbblt -- bitblt from a AxRectangle within a Layer to off-screen Bitmap 
 */
lbblt(lp, srp, dbp, dox, doy, f)
	Layer *lp;		/* -> source Layer */
	A_Bitmap *dbp;		/* -> destination Bitmap */
	AxRectangle *srp;		/* -> source AxRectangle */
	short dox, doy;		/* destination origin (x, y) */
	int f;			/* function code */
{
	LbltPack pack;
	/*
	 * Set up packet of "other args" for passing to Lblt by
	 * way of ATMlayerop().
	 * Compute destination-to-source offset.
	 */
	pack.abp = dbp;
	pack.f = f;
	pack.xoff = srp->origin.x - dox;	/* x offset */
	pack.yoff = srp->origin.y - doy;	/* y offset */

	ATMlayerop(lp, LBblt, srp, &pack);
}
/********************************************************************/
/*
 * Lpf() -- function applied by ATMlayerop() in lrectf()
 */
Lpf(lp, polyPoints, nPts, bp, ap, clip)
Layer *lp;
TmPoint32  *polyPoints;
int nPts;
A_Bitmap *bp;
LrfPacket *ap;
AxRectangle *clip;
{
	if (getimage(bp,TRUE,FALSE))
		polyf(bp,polyPoints,nPts,ap->a.funcode,clip);
}
/********************************************************************/
/*
 * lpolyf -- single-argument bit blitter (polyf) generalized for
 *		windowing
 *
 * Argument lp takes the place of the bitmap pointer in polyf().
 */

lpolyf(lp, polyPoints, nPts, funcode)
Layer *lp;
TmPoint32 *polyPoints;
int nPts;
int funcode;
{
	LrfPacket packet;

	packet.a.funcode = funcode;
	ATMnorectLayerop(lp, Lpf, polyPoints, nPts, &packet);
}

/********************************************************************/
/*
 * Lrf() -- function applied by ATMlayerop() in lrectf()
 */
Lrf(lp, rp, bp, ap, op)
	Layer *lp;
	AxRectangle *rp;
	A_Bitmap *bp;
	LrfPacket *ap;
	Obscured *op;

{
	if (getimage(bp, TRUE, FALSE)) rectf(bp, rp, ap->a.funcode);
}

/********************************************************************/

/*
 * lrectf -- single-argument bit blitter (rectf) generalized for
 *		windowing
 *
 * Argument lp takes the place of the bitmap pointer in rectf().
 */

lrectf(lp, rp, funcode)
	Layer *lp;
	AxRectangle *rp;
	int funcode;
{
	LrfPacket packet;
/*
 * Set up packet of "other args" for passing to Lrf by way of
 *	ATMlayerop().
 */
	packet.a.funcode = funcode;

	ATMlayerop(lp, Lrf, rp, &packet);
}
/********************************************************************/
/*
 * elrectf -- easy lrectf -- builds rectangle and calls lrectf()
 */
elrectf(lp, xo, yo, xc, yc, f)
	Layer *lp;
	short xo, yo, xc, yc, f;
{
	AxRectangle rr;

	rr.origin.x = xo;
	rr.corner.x = xc;
	rr.origin.y = yo;
	rr.corner.y = yc;
	lrectf(lp, &rr, f);
}

