/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: iconic.h
   Applix ASC version: 1
   Date: Mon Nov 16 13:53:34 1998
*/
#ifndef NDEBUG
#define LOGGER(a)	{ if (LOG) { fprintf a; fflush(LOG); } }
#else
#define LOGGER(a)
#endif

#ifndef FALSE
#define FALSE	0
#endif

#ifndef TRUE
#define TRUE	1
#endif

#define WINDOW_DB		"atm_icons"
#define OLD_WINDOW_DB	"atm_rev1_icons"

#define WHITE_GAP	1
#define DROP_SHADOW	2

#define T_EDGE	WHITE_GAP
#define L_EDGE	WHITE_GAP
#define B_EDGE	(DROP_SHADOW + WHITE_GAP)
#define R_EDGE	(DROP_SHADOW + WHITE_GAP)

/*** icon ids ***/
#define CLOSE		0
#define MOVE		1
#define UP_UNIT		2
#define DOWN_UNIT	3
#define RIGHT_UNIT	4
#define LEFT_UNIT	5
#define UP_PAGE		6
#define DOWN_PAGE	7
#define LEFT_PAGE	8
#define RIGHT_PAGE	9
#define FULL		10
#define FRAME		11
#define SET_ASIDE	12
#define EXECUTE		13
#define VERT_BAR	14
#define HORIZ_BAR	15
#define TITLE		16
#define SHRINK		17
#define MAX_ICON	18

#define INNER		18
#define OUTER		19
#define MAX_REL		20

/*** icon flags ***/
#define CLOSE_FLAG	0x00000001
#define MOVE_FLAG	0x00000002
#define UP_UNIT_FLAG	0x00000004
#define DOWN_UNIT_FLAG	0x00000008
#define RIGHT_UNIT_FLAG	0x00000010
#define LEFT_UNIT_FLAG	0x00000020
#define UP_PAGE_FLAG	0x00000040
#define DOWN_PAGE_FLAG	0x00000080
#define LEFT_PAGE_FLAG	0x00000100
#define RIGHT_PAGE_FLAG	0x00000200
#define FULL_FLAG	0x00000400
#define FRAME_FLAG	0x00000800
#define SET_ASIDE_FLAG	0x00001000
#define EXECUTE_FLAG	0x00002000
#define VERT_BAR_FLAG	0x00004000
#define HORIZ_BAR_FLAG	0x00008000
#define TITLE_FLAG	0x00010000	/* beyond range of border_control! */
#define SHRINK_FLAG	0x00020000

#define UNIT_FLAGS	0x0000003C
#define PAGE_FLAGS	0x000003C0
#define ALL_FLAGS	0x0003FFFF
#define DRAWN_FLAGS	0x0001C000			/* title & bars */
#define IMAGED_FLAGS	(ALL_FLAGS & ~DRAWN_FLAGS)
#define OPT_FLAGS	(MOVE_FLAG | PAGE_FLAGS)	/* not req'd */
#define REQ_FLAGS	(ALL_FLAGS & ~OPT_FLAGS)

#define RIGHT_FLAGS	(UP_UNIT_FLAG | DOWN_UNIT_FLAG | UP_PAGE_FLAG \
						| DOWN_PAGE_FLAG | FRAME_FLAG | VERT_BAR_FLAG)
#define BOT_FLAGS	(RIGHT_UNIT_FLAG | LEFT_UNIT_FLAG | LEFT_PAGE_FLAG \
						| RIGHT_PAGE_FLAG | FULL_FLAG | FRAME_FLAG \
						| HORIZ_BAR_FLAG)

typedef long flags;

#define FLAG_SET(n,m)	((icon_flag[n] & (m)) != 0)
#define FLAG_CLEAR(n,m)	((icon_flag[n] & (m)) == 0)
#define SET_FLAG(n,m)	(m |= icon_flag[n])
#define CLEAR_FLAG(n,m)	(m &= ~icon_flag[n])

#if !THIMBLE
/*** icon info ***/
typedef struct {
	short xx1, yy1;
	short rx1, ry1;
	union {
		struct {
			short uxsize, uysize;
			char *udata;
		} uimaged;
		struct {
			short uxx2, uyy2;
			short urx2, ury2;
		} udrawn;
	} uinfo;
} icon;
#define xsize	uinfo.uimaged.uxsize
#define ysize	uinfo.uimaged.uysize
#define data	uinfo.uimaged.udata
#define xx2	uinfo.udrawn.uxx2
#define yy2	uinfo.udrawn.uyy2
#define rx2	uinfo.udrawn.urx2
#define ry2	uinfo.udrawn.ury2
#endif	/* !THIMBLE */

#define BYTE_WIDTH(w)	(((w) + 15 >> 3) & ~1)

struct app_icon {
	struct app_icon *next;
	short icon_id;
	short appl_id;
#if THIMBLE
	char *pixmapP;	/* X-wins pixmap or NULL (really a Pixmap type) */
	short xsize, ysize;
#else
	short xpels, ypels;
#endif
	char *imdata;
	char *mapdata;
};

