/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: iswide.c
   Applix ASC version: 7
   Date: Mon Apr  5 21:13:10 1999
*/
#include <basicwin.h>
#include <aos.h>
#include <wide_extern.h>

#define NUM_WIDE_NAMES 5
#define NUM_NARROW_NAMES 15


#if BUILD_ORIENT /*  { */


/** return TRUE if the attribute structure is from a "wide" font */
int IsWideFontFace(name)
mbchar *name;
{
	XFontStruct	*xfont;
	int 		scalable;
    int 	wide;
    int i;
    static mbchar wide_names[NUM_WIDE_NAMES][20];    /* cache for performance */
    static mbchar narrow_names[NUM_NARROW_NAMES][20];
    static num_wide_names = 0;
    static num_narrow_names = 0;

    if ( name == (mbchar *)NULL )
       return( FALSE );           /* "nothing" is never wide! */
    
#if AXMS && BUILD_KANJI              /*  {  */
    /* The default orient faces are still in "Minchou" or "Gothic", which have
     * nothing to do with real font names. Pick these out as special cases.
     * Ashisuto added. 1996.12.03
     */
    if ( IsSimpleWideFontFace(name) )
        return( TRUE );
    /* on NT we determine if it is a Japanese UNICODE font by looking for */
    /* wide characters in the name.  This seems OK, but this test may have */
    /* to be made more better if some non-UNICODE fonts show up with       */
    /* Japanese chars in the name show up.                                 */
    if ( MBnum_wide_chars_n( name, strlen(name) ) > 0 )
        /* it is wide */
        return( TRUE );
    else
        return( FALSE );
#else                                /*  }  {  */

    /* first look through the wide cache */
    for ( i=0; i<num_wide_names; i++ )
       {
       if ( *name == wide_names[i][0] )
          {
          /* first char matches, do the work of the strcmp */
          if ( !MBstrcmp( name, wide_names[i] ) )
              /* found it, it is a wide face */
              return( TRUE );
          }
       }

    /* nope, look through the narrow cache */
    for ( i=0; i<num_narrow_names; i++ )
       {
       if ( *name == narrow_names[i][0] )
          {
          /* first char matches, do the work of the strcmp */
          if ( !MBstrcmp( name, narrow_names[i] ) )
              /* found it, it is a narrow face */
              return( FALSE );
          }
       }

    /* did not find the name in the cache, go the longer route */
	if (!getXFontStruct(name, 12, 0, 0, &scalable, &xfont))
		return(FALSE);

    if (xfont->max_byte1 > 0) {
        /* we selected a wide font */
        if (num_wide_names < NUM_WIDE_NAMES) {
           /* add to the cache */
           MBstrcpy(wide_names[num_wide_names], name);
           num_wide_names++;
		}
        return(TRUE);
	} else {
        if (num_narrow_names < NUM_NARROW_NAMES) {
           /* add to the cache */
           MBstrcpy(narrow_names[num_narrow_names], name);
           num_narrow_names++;
		}
        return( FALSE );
    }
#endif                       /*  }  AXMS vs X11 */

}

IsNarrowKanaFontFace(name)
mbchar	*name;
{
	mbchar *narrow_prefix = NXLT("Narrow");
	int	prefix_len = MBstrlen(narrow_prefix);

	if (MBstrncmp(name, narrow_prefix, prefix_len) == 0 
			&& IsWideFontFace(name+prefix_len))
		return(TRUE);
	else
		return(FALSE);
}

#endif   /*  } BUILD_ORIENT */

