/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: layline.c
   Applix ASC version: 8
   Date: Mon Nov 16 13:53:34 1998
*/
#include <stdio.h>
#include <tmi.h>
#include <aos.h>
#include <oldwin.h>
#include <atmattr.h>
#include <termdep.h>	/* Ltvec support (TMODE) */

typedef struct {
	short x;
	short y;
	} TmPoint;
typedef struct {
	int x;
	int y;
	} TmPoint32;

extern short ATMtrace;
extern FILE *ATfp;
extern short ATMdebug;
extern FILE *fd2;
extern short MousChecked;
extern short TermReleased;		/* 6/24/89 ega fix JJM */

/********** type definitions **********/

/***** Lline() parameter packet *****/

typedef struct
{
	Vector32 *strokep;	/* stroke definition (point pair) */
	AxRectangle *clipboxp;	/* clipping rectangle */
	short fn;		/* transfer mode */
	short esF;		/* extra segment flag */
	short interF;		/* interleaved organization flag */
	char *cbase;		/* coordinate base of bitmap image */
	short wid;		/* width of bitmap image */
	short line_style;	/* solid, dotted, etc. */
	short line_thickness;
} LlinePack;

#if SUN_G
#include <pixrect/pixrect_hs.h>
#endif

#define MAX_PEN	5
#define MAX_PATTERN	4

static unsigned char LpenThin[2] = {0xFF, 0xFF};

/* *** pen images *** */

static short LpenImage[MAX_PEN - 1][16] = {
	{0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
	 0x0180, 0x0180,
	 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0},

 	{0x0, 0x0, 0x0, 0x0, 0x0, 0x0,
	 0x0180, 0x03c0,
	 0x03c0, 0x0180,
	 0x0, 0x0, 0x0, 0x0, 0x0, 0x0},

	{0x0, 0x0, 0x0, 0x0, 0x0,
	 0x0180, 0x03c0,
	 0x07e0, 0x07e0,
	 0x03c0, 0x0180,
	 0x0, 0x0, 0x0, 0x0, 0x0},

	{0x0, 0x0, 0x0, 0x0,
	 0x01c0, 0x03e0,
	 0x07f0, 0x07f0,
	 0x07f0, 0x03e0,
	 0x01c0,
	 0x0, 0x0, 0x0, 0x0, 0x0}
};

/* *** line patterns (default version) *** */

static int dot_0[] = { 1,  2, 0};
static int dot_1[] = { 1,  5, 0};
static int dot_2[] = { 1,  9, 0};
static int dot_3[] = { 1, 13, 0};
static int dot_4[] = { 1, 17, 0};

static int dash_0[] = { 3,  2, 0};
static int dash_1[] = { 7,  5, 0};
static int dash_2[] = {11,  9, 0};
static int dash_3[] = {15, 13, 0};
static int dash_4[] = {19, 17, 0};

static int dotda_0[] = { 3,  2,  1,  2, 0};
static int dotda_1[] = { 7,  5,  1,  5, 0};
static int dotda_2[] = {11,  9,  1,  9, 0};
static int dotda_3[] = {15, 13,  1, 13, 0};
static int dotda_4[] = {19, 17,  1, 17, 0};

#if SUN_G
mpr_static(PEN0PR,  1,  1, 1, LpenThin);
mpr_static(PEN1PR, 16, 16, 1, LpenImage[0]);
mpr_static(PEN2PR, 16, 16, 1, LpenImage[1]);
mpr_static(PEN3PR, 16, 16, 1, LpenImage[2]);
mpr_static(PEN4PR, 16, 16, 1, LpenImage[3]);
A_Bitmap PEN0 = {(char *) &PEN0PR, 1, 0, 0,  1,  1};
A_Bitmap PEN1 = {(char *) &PEN1PR, 2, 0, 0, 16, 16};
A_Bitmap PEN2 = {(char *) &PEN2PR, 2, 0, 0, 16, 16};
A_Bitmap PEN3 = {(char *) &PEN3PR, 2, 0, 0, 16, 16};
A_Bitmap PEN4 = {(char *) &PEN4PR, 2, 0, 0, 16, 16};
#endif

A_Bitmap PEN0 = {(char *) LpenThin    , 1, 0, 0,  1,  1};
A_Bitmap PEN1 = {(char *) LpenImage[0], 2, 0, 0, 16, 16};
A_Bitmap PEN2 = {(char *) LpenImage[1], 2, 0, 0, 16, 16};
A_Bitmap PEN3 = {(char *) LpenImage[2], 2, 0, 0, 16, 16};
A_Bitmap PEN4 = {(char *) LpenImage[3], 2, 0, 0, 16, 16};

#if SUN_G
static struct penstruct {
	A_Bitmap *pr;			/* pentip bitmap */
	int x0, y0;					/* x, y origin of pen tip */
	int *Lpattern[3];			/* patterns */
} Lpens[MAX_PEN] = {
	{ &PEN0, 7, 7, dash_0, dot_0, dotda_0 },
	{ &PEN1, 7, 7, dash_1, dot_1, dotda_1 },
	{ &PEN2, 7, 7, dash_2, dot_2, dotda_2 },
	{ &PEN3, 7, 7, dash_3, dot_3, dotda_3 },
	{ &PEN4, 7, 7, dash_4, dot_4, dotda_4 }
};
#else
static struct penstruct {
	A_Bitmap *pr;					/* pentip bitmap */
	int x0, y0;					/* x, y origin of pen tip */
	int *Lpattern[3];			/* patterns */
} Lpens[MAX_PEN] = {
	{ &PEN0, 7, 7, dash_0, dot_0, dotda_0 },
	{ &PEN1, 7, 7, dash_1, dot_1, dotda_1 },
	{ &PEN2, 7, 7, dash_2, dot_2, dotda_2 },
	{ &PEN3, 7, 7, dash_3, dot_3, dotda_3 },
	{ &PEN4, 7, 7, dash_4, dot_4, dotda_4 }
};
#endif

static struct penstruct *Lpen;

/*******************************************************************/
/*
 * Lline -- line drawing function called by layline()
 */
Lline(lp, rp, bp, ap, op)
Layer *lp;
AxRectangle *rp;		/* clipping rectangle */
A_Bitmap *bp;
LlinePack *ap;
Obscured *op;
{
}

/*******************************************************************/
/*
 * layline -- draw a line, called with old atm defined line widths
 */
layline(lp, vp, fn, ldp)
Layer *lp;             /* -> Layer (can be NULL for Display) */
Vector32 *vp;          /* graphics vector (point pair)       */
short fn;              /* transfer mode                      */
TmLineDescriptor *ldp; /* can be NULL for default            */
{
}

/***********************************************************************/

#define MAX_POLY	100

#define BELOW(ty)	((clipArray[2].y < ty) << 3)
#define RIGHT(tx)	((clipArray[2].x < tx) << 2)
#define ABOVE(ty)	((ty < clipArray[0].y) << 1)
#define LEFT(tx)	(tx < clipArray[0].x)

typedef unsigned short bool;

static TmPoint clipArray[5], *clipdPtr, clipdPoly[MAX_POLY * 2];
static bool *edgePtr, edgePoly[MAX_POLY * 2];
static short edgeCnt;

/***********************************************************************/

LoadClip (clipRect)
AxRectangle *clipRect;
{
	int left, right, top, bottom;

	left = clipRect->origin.x;
	top = clipRect->origin.y;
	right = clipRect->corner.x - 1;
	bottom = clipRect->corner.y - 1;

	clipArray[0].x = left;	clipArray[0].y = top;
	clipArray[1].x = right;	clipArray[1].y = top;
	clipArray[2].x = right;	clipArray[2].y = bottom;
	clipArray[3].x = left;	clipArray[3].y = bottom;
	clipArray[4].x = left;	clipArray[4].y = top;
}

/***********************************************************************/

static bool interc(i, x1, y1, x2, y2, rx, ry)
short i, x1, y1, x2, y2, *rx, *ry;
{
	long dx, dy;

	dx = x2 - x1;
	dy = y2 - y1;
	if (i & 1) {
		if (!dy)
			return FALSE;
		*ry = clipArray[i].y;
		if ((y1 < *ry) != (y2 < *ry)) {
			*rx = x1 + dx * (*ry - y1) / dy;
			return TRUE;
		}
		return FALSE;
	}
	if (!dx)
		return FALSE;
	*rx = clipArray[i].x;
	if ((x1 < *rx) != (x2 < *rx)) {
		*ry = y1 + dy * (*rx - x1) / dx;
		return TRUE;
	}
	return FALSE;
}

/***********************************************************************/

static emit(code, x, y)
bool code;
short x, y;
{
	static short lx, ly;

	if (clipdPtr != clipdPoly && x == lx && y == ly)
		return;

	clipdPtr->x = lx = x;
	clipdPtr->y = ly = y;
	clipdPtr++;
	*edgePtr++ = code;
}

/***********************************************************************/

static clipSeg(lvl, x1, y1, x2, y2)
short lvl, x1, y1, x2, y2;
{
	short v, h, tmp, i, tx, ty;
	int bughack=0; /*usl54 bug hack so will build */

	for (i = lvl; i < 4; i++) {
		if (interc(i, x1, y1, x2, y2, &tx, &ty)) {
			clipSeg(i + 1, x1, y1, tx, ty);
			clipSeg(i + 1, tx, ty, x2, y2);
			return;
		}
	}

	if (x1 < x2)
		(h = LEFT(x1)) || (h = RIGHT(x2));
	else
		(h = LEFT(x2)) || (h = RIGHT(x1));
		
	if (y1 < y2)
		(v = ABOVE(y1)) || (v = BELOW(y2));
	else
		(v = ABOVE(y2)) || (v = BELOW(y1));

	tmp = v | h;
	if (!tmp) {
		emit(FALSE, x1, y1);
		emit(TRUE, x2, y2);
	}
	else if (v && h) {
		tmp >>= 2;
		if (1 < tmp)
			tmp = 5 - tmp;
		emit(FALSE, clipArray[tmp].x, clipArray[tmp].y);
	}
}

/***********************************************************************/

clipPoly(pts, cnt, fill, close)
TmPoint **pts;
int *cnt;
bool close, fill;
{
	register int i;
	register TmPoint *p1, *p2;

	p2 = *pts;
	p1 = p2++;
	clipdPtr = clipdPoly;
	*(edgePtr = edgePoly) = 0;
	for (i = 1; i < *cnt; p1++, p2++, i++)
		clipSeg(0, p1->x, p1->y, p2->x, p2->y);

	if (!close)
		edgeCnt = edgePtr - edgePoly;

	if (fill || close) {
		clipSeg(0, p1->x, p1->y, (*pts)->x, (*pts)->y);
		if (clipdPtr != clipdPoly)
			emit(*edgePoly, clipdPoly->x, clipdPoly->y);
	}

	if (close)
		edgeCnt = edgePtr - edgePoly;

	*pts = clipdPoly;
	*cnt = clipdPtr - clipdPoly;
}

bool borderLine(index)
short index;
{
	return index < edgeCnt && edgePoly[index];
}
/* ---------------- */
