/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: parachek.c
   Applix ASC version: 8
   Date: Mon Nov 16 13:53:42 1998
*/
#include <tmi.h>
#if CHECK
/*
 * Note: the includes alone can't be compiled under VENIX
 * without a dimension table overflow.
 */
#include <fcntl.h>
#include <stdio.h>
#include <aos.h>
#include <paripc.h>
#include <oldwin.h>
#include <atm.h>
#include <atmwin.h>
#include <atmreg.h>
#include <atmattr.h>
#include <atmtxtin.h>
#include <atmgraph.h>
#include <paratmgr.h>
#include <parwin.h>
#include <parreg.h>
#include <parout.h>
#include <paratmin.h>
#include <parsim.h>
#include <parattr.h>
#include <getstr13.h>


/* general defines */

#define OR	||
#define AND	&&
#define ZERO	0

#define MAX_TIT_LEN	STRPATH /* from aos.h */
#define MAX_OPT_LEN	80

extern short MaxwWid;
extern short MaxwHt;
extern short MinWWid, MinWHt;	/* min window size in mils */
extern char OldMouse;		/* compatibility with Tim's recorded demo */
extern short ATMtrace;
extern FILE *ATfp;

/* WINDOW MANAGEMENT */

typeErr						/* error status */
PCkCreate(pnt)
	char *pnt;
{
	struct SCreateData *cpnt;

	cpnt = (struct SCreateData *)pnt;

#if !THIMBLE
	/* first reserved value is used in tmfw.c */
	if ( !(OldMouse || cpnt->cr_pack.res_1 == 0 && cpnt->cr_pack.res_2 == 0) )
	{
		return TMR_RESVAL;			/* bad reserved value */
	}
	else 
#endif
	if ((cpnt->cr_pack.x_position > MaxwWid)
			OR (cpnt->cr_pack.y_position > MaxwHt)
			OR (cpnt->cr_pack.x_position < ZERO)
			OR (cpnt->cr_pack.y_position < ZERO))
	{
		return TMR_POSE;			/* position error */
	}
#if 0	/* removed on 8/30/91 ... let X handle this issue */
	else if ((cpnt->cr_pack.x_size > MaxwWid)
			OR (cpnt->cr_pack.y_size > MaxwHt)
			OR (cpnt->cr_pack.x_size < MinWWid)
			OR (cpnt->cr_pack.y_size < MinWHt) )
	{
		TLOGGER((ATfp, NXLT("size (%d %d) mils\n"), cpnt->cr_pack.x_size,
				cpnt->cr_pack.y_size));
		return TMR_SIZE;			/* size error */
	}
#endif
	else if ((cpnt->help_num > MAX_HELP_NUM)
			OR (cpnt->help_num < ZERO)
			OR (cpnt->option_len > MAX_OPT_LEN)
			OR (cpnt->option_len < ZERO))
	{
		return TMR_OPE;				/* too many selection line entries */
	}
	else if ((cpnt->title_len > MAX_TIT_LEN)
			OR (cpnt->title_len < ZERO))
	{
		return TMR_TITLE;			/* title too long */
	}
	else
	{
		return(TMR_OK);
	}
}
	
PCkSAside(pnt)
char *pnt;

{
}

PCkCurWin(pnt)
char *pnt;

{
}

/* REGION MANAGEMENT */

PCkDreg(pnt)
char *pnt;
{
	struct SDregData *cpnt;

	cpnt = (struct SDregData *)pnt;

	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMDREG, ERRWID);	/* "dreg : illegal window id") */
	}
	else
	{
	    if((cpnt->r_def_num > MAX_R_DEF_NUM)
			OR (cpnt->r_def_num < ZERO))
	    {
	 		ATMab1(RTTMDREG, ERREG);	/* "dreg : too many regions") */
	    }
	    else
	    {
		return(TMR_OK);
	    }
	}
}

PCkMreg(pnt)
char *pnt;
{
	struct SMregData *cpnt;

	cpnt = (struct SMregData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMMREG, ERRWID);	/* "mreg : illegal window id") */
	}
	else
	{
	    if((cpnt->r_mod_num > MAX_R_MOD_NUM)
			OR (cpnt->r_mod_num < ZERO))
	    {
	 	ATMab1(RTTMMREG, ERREG);	/* "mreg : too many regions") */
	    }
	    else
	    {
		return(TMR_OK);
	    }
	}
}

PCkRclose(pnt)
char *pnt;
{
	struct SRcloseData *cpnt;

	cpnt = (struct SRcloseData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMRCLOSE, ERRWID);	/* "rclose : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

PCkARclose(pnt)
char *pnt;
{
	struct SARcloseData *cpnt;

	cpnt = (struct SARcloseData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
		ATMab1(RTTMARCLOSE, ERRWID);	/* "arclose : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

/* TEXT AND GRAPHICS */

typeErr
PCkWriteTx(pnt)
char *pnt;
{
	struct SWriteTxData *cpnt;

	cpnt = (struct SWriteTxData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < TM_NULWIN))
	{
		return TMR_NOWIN;
	}
	else
	{
		if(cpnt->byte_count > MAX_TX_COUNT)
		{
			return TMR_LONG;
		}
		else
		{
		    return(TMR_OK);
		}
	}
}

PCkScroll(pnt)
char *pnt;
{
	struct SScrollData *cpnt;

	cpnt = (struct SScrollData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMSCROLL, ERRWID);	/* "scroll : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

PCkPMark(pnt)
char *pnt;
{
	struct SPMarkData *cpnt;

	cpnt = (struct SPMarkData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMPMARK , ERRWID);	/* "pmark : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

PCkPLine(pnt)
char *pnt;
{
	struct SPLineData *cpnt;

	cpnt = (struct SPLineData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMPLINE , ERRWID);	/* "pline : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

PCkArc(pnt)
char *pnt;
{
	struct SArcData *cpnt;

	cpnt = (struct SArcData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMARC   , ERRWID);	/* "arc : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

PCkBblt(pnt)
char *pnt;
{
	struct SBbltData *cpnt;

	cpnt = (struct SBbltData *) pnt;
	if (!winExists(cpnt->bblt_pack.window_id))
		tempAbort("TmBblt: nonexistent window");
	return(TMR_OK);
}

/* LOGICAL DEVICE INPUT */

PCkInitString(pnt)
char *pnt;
{
	struct SInitStrData *cpnt;

	cpnt = (struct SInitStrData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMINITSTR,ERRWID);	/* "init str : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

PCkMpick(pnt)
char *pnt;
{
	struct SInitPickData *cpnt;

	cpnt = (struct SInitPickData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMIPICK , ERRWID);	/* "init pick : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

PCkMstr(pnt)
char *pnt;
{
	struct SIStrokeData *cpnt;

	cpnt = (struct SIStrokeData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMISTROKE,ERRWID);	/* "init stroke : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

PCkProcessChar(pnt)
char *pnt;
{
	struct SProcessCharData *cpnt;

	cpnt = (struct SProcessCharData *)pnt;
}

/* ATTRIBUTES AND CHARACTERISTICS */

PCkInq (pnt)
char *pnt;

{
	struct SInqData *cpnt;

	cpnt = (struct SInqData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMINQ   , ERRWID);	/* "inq : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

PCkDchar(pnt)
char *pnt;

{
}

PCkDcursor(pnt)
char *pnt;

{
	struct SDcursorData *cpnt;

	cpnt = (struct SDcursorData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMDCURSOR , ERRWID);	/* "dcursor : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

PCkSetWin(pnt)
char *pnt;

{
	struct SSetWinData *cpnt;

	cpnt = (struct SSetWinData *)pnt;
	if((cpnt->window_id >= MAX_WIND_ID)
			OR (cpnt->window_id < ZERO))
	{
	    ATMab1(RTTMSETWIN, ERRWID);	/* "set win attr : illegal window id") */
	}
	else
	{
	    return(TMR_OK);
	}
}

PCkTextExt(pnt)
char *pnt;

{
/*
	struct STextExtData *cpnt;

	cpnt = (struct STextExtData *)pnt;
*/
}
#else			/* conditional compilation */
paradummy(){}
#endif
