/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmcolor.c
   Applix ASC version: 71
   Date: Tue May 11 21:11:16 1999
*/
#include <stdio.h>
#include <math.h>
#include <basicwin.h>
#include <errno.h>
#include <aos.h>
#include <tmi.h>        /* just for defines */
#include <atm.h>        /* just for defines */
#include <atmattr.h>    /* just for defines */
#include <getstr5.h>
#include <colors.h>

#define N_ALIS_PIX 16
#define N_AX_PIX 256
#define MAX_GUN 65535L      /* as in xmlib.c */
#define FILL_COLOR 0        /* as in tmout.c */
#define LINE_COLOR 1        /* as in tmout.c */

#undef allocStr
#define allocStr(s) strcpy(malloc(strlen(s)+1),s)
#define packedRGB(r,g,b)    ((((b)&0xFF)<<16) | (((g)&0xFF)<<8) | ((r)&0xFF))
#define packedBlue(rgb)     (((rgb)>>16)&0xFF)
#define packedGreen(rgb)    (((rgb)>>8)&0xFF)
#define packedRed(rgb)      ((rgb)&0xFF)
#define COLOR_LOG 1

/* profile color control flags */
extern int AxLoadApplicationColors; /* load app colors into default
                     * colormap */
extern int AxLoadWidgetColors;  /* load widget colors into def colormap */
extern short NoColor;           /* indicates mono apps - forced if set by
                                 * profile */
int AxCreateCmap;
int AxNoWidgetColor;        /* indicates mono widgets - forced if profiled */
extern int AxNColorLevels;  /* the height of the color cube */
extern int AxNColors;        /* total number of colors used by apps */
extern int AxSlowDither;        /* indicates slow dithering */ 
extern short ColorDisplay;  /* color available for apps and being used */
extern int UseWidgetColors; /* color available for widgets and being used */
extern int AxOpenLook;      /* running OpenLook mode */
extern int JavaMode;    /* TRUE for Java slaved axmain */

extern XColor Ax64Defs[];
extern short DpyWid, DpyHt, DpyDepth;
extern short ToggleFG;

extern Display *Dpy;
extern Visual *AxVisual;
extern Window RootWin;
extern char *AxDisplayName; /* Name of display from cmdline or getenv */
extern GC Gc1;          /* graphics context shared by all windows */
extern unsigned long FGpixel, BGpixel;

/* pixel indices allocated by Aster*x */
extern unsigned long ForeWorkPixel, BackWorkPixel;
extern unsigned long SunWorkPixel, ShadowWorkPixel, RecessWorkPixel;
extern unsigned long ForePixel, BackPixel;
extern unsigned long SunPixel, ShadowPixel, RecessPixel;
extern int Ax64Values[][6];
extern Colormap WorkColormap;
extern int GotWorkAuxColorCells;

unsigned long AlisPixels[N_ALIS_PIX];
unsigned long AxPixels[N_AX_PIX];
unsigned long AxColors[256];
unsigned long WidgetPixels[10];

typedef struct {
    int rgb;        /* color: packed RGB */
    int pixel;      /* pixel value */
} AXDIRECTCOLOR;

#define AX_XA_RGB_APPLIX_MAP_STR    NXLT("XA_RGB_APPLIX_MAP")
#define AX_XA_NUM_ASTERX_RUN_STR    NXLT("XA_NUM_ASTERX_RUN")

#define WDGT_BASE_POS    246
#define APP_BASE_POS       30
#define COLORMAP_CREATED       1 
#define COLORMAP_LOADED        2

static int  compare();
int     AxStandardColormapWasLoaded=0; /*0->notloaded,1->I loaded, 2->I found */
extern char  *AxStdCmapName;
extern int   axPreDefinedStdmap;     /* Built-in atom */
extern int   AxNStdCmaps;
extern Atom  axXA_RGB_APPLIX_MAP;        /* standard colormap */
extern Atom  axXA_NUM_ASTERX_RUN;        /* number of asterxs running */
Atom APPLIX_MAP;                         /* applix colormap */
int     axStartWidgetPixel, axNumWidgetPixels;
int     axStartAppPixel, axNumAppPixels;
extern int AxVisualIsReadOnly;
extern XStandardColormap   *AxStdCmapInfo;

static int AxChangeNumAsterxAtomData (/* int val, int delete */);
static int icbrt_with_bits(/* int, int */);

int     nAxDirectColors;
AXDIRECTCOLOR       *AxDirectClrTable;

unsigned long AxGetRealPixel();

/* table of old ALIS colors */
int PelR[] =     {
    0xffff, /* white */
    0x0000, /* black */
    0xffff, /* ~ desktop */
    0xa000, /* desktop color (Tina) */
    0xffff, /* red */
    0x0000, /* cyan */
    0x0000, /* green */
    0x8000, /* purple (Tina) */
    0x0000, /* blue (Tina) */
    0xffff, /* yellow */
    0x6000, /* dark gray */
    0x3000, /* olive */
    0xa000, /* light gray */
    0xa000, /* dark red */
    0xffff, /* orange */
    0x0000  /* dark blue (Tina) */
};
int PelG[] =     {
    0xffff, /* white */
    0x0000, /* black */
    0xffff, /* ~ desktop */
    0xc000, /* desktop color */
    0x0000, /* red */
    0xffff, /* cyan */
    0xc000, /* green */
    0x0000, /* purple (Tina) */
    0x6000, /* blue (Tina) */
    0xffff, /* yellow */
    0x6000, /* dark gray */
    0x8000, /* olive */
    0xa000, /* light gray */
    0x0000, /* dark red */
    0x8000, /* orange */
    0x0000  /* dark blue (Tina) */
};
int PelB[] =     {
    0xffff, /* white */
    0x0000, /* black */
    0xffff, /* ~ desktop */
    0xffff, /* desktop color */
    0x3000, /* red */
    0xffff, /* cyan */
    0x3000, /* green */
    0xffff, /* purple (Tina) */
    0xffff, /* blue (Tina) */
    0x0000, /* yellow */
    0x6000, /* dark gray */
    0x2000, /* olive */
    0xa000, /* light gray */
    0x0000, /* dark red */
    0x0000, /* orange */
    0xa000  /* dark blue (Tina) */
};

extern Colormap ApplixCmap;
int cmap_ix;

/***************************************************************************
function: axAllocAllColorCells - allocate all available color cells for read/write
*/
static int
axAllocAllColorCells (clrmap, cells, visual)
Colormap clrmap;
unsigned long *cells;
Visual *visual;
{
    unsigned long plane_masks[1];
    int ncells;

    ncells = 256;
    if (visual)
        ncells = visual->map_entries;
    while ((ncells > 0) &&
        (XAllocColorCells(Dpy, clrmap, TRUE, plane_masks, 0, cells, ncells) == False))
        ncells--;
    return(ncells);
}

/***************************************************************************
function: axGetUsedPixels - get what pixels are used out of 30-246 standard map pixels
*/
static void
axGetUsedPixels (usedCells, nUsedCells)
unsigned long *usedCells;
int *nUsedCells;
{
    register int i;
    XColor stdColors[256];
    XColor usedColors[256];

    AxGetStdcmapAppColors(stdColors, APP_BASE_POS);
    AxGetApplicationColors(usedColors, nUsedCells);
    AxStdAllocColors(stdColors, usedColors, *nUsedCells);
    for(i=0; i < *nUsedCells; ++i)
        usedCells[i] = usedColors[i].pixel;
}

/***************************************************************************
function: axFreeUnusedColors - free unused color cells
*/
static void
axFreeUnusedColors (color_map, colorCells, allocedCells, ncells)
Colormap color_map;
unsigned long *colorCells;
int allocedCells, ncells;
{
    register int i, j;
    unsigned long usedCells[256];
    int nUsedCells;

    if (allocedCells == 0)
        return;

    /* get used pixels */
    axGetUsedPixels(usedCells, &nUsedCells);

    /* free all the allocated pixels except used pixels */
    for (i=0; i < allocedCells; ++i) {
        if (colorCells[i] >= WDGT_BASE_POS)
            break;
        /* we do not use any pixels before APP_BASE_POS */
        if (colorCells[i] < APP_BASE_POS)
            XFreeColors(Dpy, color_map, &colorCells[i], 1, 0L);

        /* see if pixel colorCells[i] is used */
        for (j=0; j < nUsedCells; ++j)
            if (usedCells[j] == colorCells[i])
                break;
        /* free pixel if not used */
        if (j == nUsedCells)    /* not used */
            XFreeColors(Dpy, color_map, &colorCells[i], 1, 0L);
    }
}

Colormap
AxGetWindowColormap()
{
	if (AxVisual->class == PseudoColor)
	{
		if (axPreDefinedStdmap || axXA_RGB_APPLIX_MAP == APPLIX_MAP)
			return( AxStdCmapInfo->colormap);
		else
			return((Colormap)0);
	}
	else
		return((Colormap)0);
}


/***************************************************************************
* function: allocatePixels - allocate pixels for application colors 
*           and widget colors
*/
void
allocatePixels ()
{
	unsigned long plane_masks[1];
	Colormap def_cmap;
	int nwidgetpixels, i, j;
	unsigned long colorCells[256];
	int ncells, allocedCells;
	XColor cmapdefs[256];
	int black_pixel, white_pixel;
	
	def_cmap = DefaultColormap(Dpy, DefaultScreen(Dpy));
	WorkColormap = ApplixCmap;
	
	/* init the pixel arrays */
	if ( axPreDefinedStdmap == FALSE )
	{
		black_pixel = BlackPixel(Dpy, DefaultScreen(Dpy));
		white_pixel = WhitePixel(Dpy, DefaultScreen(Dpy));
	}
	else
	{
		black_pixel = 0;
		white_pixel = 255;
	}
	for ( i = 0;i<N_AX_PIX;i++ )
		AxPixels[i] = black_pixel;
	AxPixels[215] = white_pixel;
	
	WidgetPixels[0] = black_pixel;
	for (i = 1; i < 10; i++)
		WidgetPixels[i] = white_pixel;
	AlisPixels[0] = white_pixel;
	for (i = 1; i < N_ALIS_PIX; i++)
		AlisPixels[i] = black_pixel;
		
	/* allocate pixels for the widget kit and window fg/bg */
	nwidgetpixels = AxOpenLook ? (ColorDisplay ? 10 : 7) : 7;
	
	/* make sure widgets have pixels allocated first if only they are loaded */
	if (AxVisualIsReadOnly)
		goto done;
	
	if (AxVisual->class == DirectColor)
	{
		/* allocate pixels for widget colors */
		XAllocColorCells(Dpy, def_cmap, FALSE, plane_masks, 0,
							  WidgetPixels, nwidgetpixels);
		AlisPixels[0] = WidgetPixels[nwidgetpixels - 2];
		AlisPixels[1] = WidgetPixels[nwidgetpixels - 1];
		
		/* allocate pixels for application colors */
		XAllocColorCells(Dpy, def_cmap, FALSE, plane_masks, 0,
							  AxPixels, AxNColors);
		goto done;
	}
	
	if (axXA_RGB_APPLIX_MAP == (Atom)0)
	{
		/* dont use standard colormap */
		if (AxLoadWidgetColors && AxNoWidgetColor == FALSE)
		{
			if (XAllocColorCells(Dpy, def_cmap, False, plane_masks, 0,
									  WidgetPixels, nwidgetpixels))
			{
				AlisPixels[0] = WidgetPixels[nwidgetpixels - 2];
				AlisPixels[1] = WidgetPixels[nwidgetpixels - 1];
			}
			else
			{
				AxNoWidgetColor = TRUE;
			}
		}
		else if (ApplixCmap != def_cmap)
		{
			/* private map */
			cmap_ix = 0;
			for (i=0; i < nwidgetpixels; ++i)
				WidgetPixels[i] = cmap_ix++;
		}
		else
		{
			AxNoWidgetColor = TRUE;
		}
		if (AxLoadApplicationColors && NoColor == FALSE)
		{
			if (!XAllocColorCells(Dpy, def_cmap, False, plane_masks, 0,
									   AxPixels, AxNColors))
			{
				/* could not get application colors from default map */
				NoColor = TRUE;
				AxNColorLevels = 1;
				AxNColors = 2;
			}
		}
		else if (ApplixCmap != def_cmap)
		{
			/* private map */
			cmap_ix = 0;
			for (i=0; i < AxNColors; ++i)
			{
				if (cmap_ix >= WidgetPixels[0] &&
					cmap_ix <= WidgetPixels[nwidgetpixels-1])
                    cmap_ix = WidgetPixels[nwidgetpixels-1] + 1;
				AxPixels[i] = cmap_ix++;
			}
		}
		else 
		{
			NoColor = TRUE;
			AxNColorLevels = 1;
			AxNColors = 2;
		}
		goto done;
	}
	
	allocedCells = 0;
	ncells = 0;
	if (AxLoadWidgetColors || AxLoadApplicationColors)
	{
		if (AxNoWidgetColor == FALSE &&
			NoColor == FALSE && axPreDefinedStdmap == FALSE)
			ncells = axAllocAllColorCells(def_cmap, colorCells, AxVisual);
		allocedCells = ncells;
	}
	
	if (AxLoadWidgetColors && !AxLoadApplicationColors)
	{
		/* allocate colors for the widgets */
		if (AxNoWidgetColor == FALSE)
		{
			/* allocate widget colors in cells at top (from 255 to bottom) */
            cmap_ix = WDGT_BASE_POS;
            for (i=0; i < nwidgetpixels; ++i)
                WidgetPixels[i] = cmap_ix++;
            AlisPixels[0] = WidgetPixels[nwidgetpixels - 2];
            AlisPixels[1] = WidgetPixels[nwidgetpixels - 1];

            if (ncells >= nwidgetpixels &&
			   (colorCells[ncells-1] <= AxVisual->map_entries-1) &&
			   (colorCells[ncells-nwidgetpixels] >= WDGT_BASE_POS))
			{
                ncells -= nwidgetpixels;
                axStartWidgetPixel = WDGT_BASE_POS;
                axNumWidgetPixels = nwidgetpixels;
            }
            else
			{
				if (axPreDefinedStdmap || axXA_RGB_APPLIX_MAP == APPLIX_MAP)
				{
                    /* private map or standard colromap */
                    /* try to allocate as many as possible to def_cmap */
                    for (i=0; i < ncells; ++i)
                        if (colorCells[i] >= WDGT_BASE_POS)
                            break;
                    if (i == ncells)
                        axStartWidgetPixel = (-1);
                    else 
					{
                        axStartWidgetPixel = colorCells[i];
                        for (j=i; j < ncells; ++j)
						{
                            if (colorCells[j] == axStartWidgetPixel+j-i)
                                axNumWidgetPixels++;
                            else
                                break;
                        }
                        axNumWidgetPixels = MAX(nwidgetpixels,
												axNumWidgetPixels);
                        ncells -= axNumWidgetPixels;
                    }
                }
                else
				{
					if (ApplixCmap == def_cmap)
						AxNoWidgetColor = TRUE;
				}
            }
        }

        /* allocate colors for the applications in cells 30 - 245 */
        if (NoColor == FALSE) {
            cmap_ix = APP_BASE_POS;
            for (i=0; i < AxNColors; ++i)
                AxPixels[i] = cmap_ix++;
        }
    }
    else
	{
        /* allocate colors for the widgets in 246 - 255 cells */
        if (AxNoWidgetColor == FALSE )
		{
            cmap_ix = WDGT_BASE_POS;
            for (i = 0; i < nwidgetpixels; i++)
                WidgetPixels[i] = cmap_ix++;
            AlisPixels[0] = WidgetPixels[nwidgetpixels - 2];
            AlisPixels[1] = WidgetPixels[nwidgetpixels - 1];

            if (AxLoadWidgetColors == TRUE) 
			{
                if ( ncells >= nwidgetpixels &&
				   ( colorCells[ncells-nwidgetpixels] >= WDGT_BASE_POS) &&
				   (colorCells[ncells-1] == colorCells[ncells-nwidgetpixels]+
										 nwidgetpixels-1))
				{
                    axStartWidgetPixel = WDGT_BASE_POS;
                    axNumWidgetPixels = nwidgetpixels;
                    ncells -= nwidgetpixels;
                }
                else
				{
					if (axPreDefinedStdmap || axXA_RGB_APPLIX_MAP == APPLIX_MAP)
					{
                        /* private map or standard colromap */
                        /* try to allocate as many as possible to def_cmap */
                        for (i=0; i < ncells; ++i)
                            if (colorCells[i] >= WDGT_BASE_POS)
                                break;
                        if (i == ncells)
                            axStartWidgetPixel = (-1);
                        else {
                            axStartWidgetPixel = colorCells[i];
                            for (j=i; j < ncells; ++j) {
                                if (colorCells[j] == axStartWidgetPixel+j-i)
                                    axNumWidgetPixels++;
                                else
                                    break;
                            }
							axNumWidgetPixels = MAX(nwidgetpixels,
													axNumWidgetPixels);
                            ncells -= axNumWidgetPixels;
                        }
                    }
                    else
					{
                    	if (ApplixCmap == def_cmap)
                        	AxNoWidgetColor = TRUE;
					}
                }
            }
        }

		/* allocate colors for the applications in 30 - 245 cells */
		if (NoColor == FALSE)
		{
			axGetUsedPixels(AxPixels, &cmap_ix);
			
			while (ncells > 0 && colorCells[ncells-1] > WDGT_BASE_POS-1)
				ncells--;
				
			if (AxLoadApplicationColors == TRUE)
			{
				if (ncells >= AxNColors &&
				   colorCells[0] <= APP_BASE_POS &&
				   colorCells[ncells-1] == WDGT_BASE_POS-1)
				{
                    ncells -= AxNColors;
                    axStartAppPixel = APP_BASE_POS;
                    axNumAppPixels = AxNColors;
                }
                else  
				{
					if (axPreDefinedStdmap || axXA_RGB_APPLIX_MAP == APPLIX_MAP)
					{
                        /* private map or standard colromap */
                        /* try to allocate as many as possible to def_cmap */
                        for (i=ncells-1; i > 0; --i) {
                            if (colorCells[i] < WDGT_BASE_POS)
                                break;
                        }
                        if (i <= 0)
                            axStartAppPixel = (-1);
                        else {
                            axStartAppPixel = colorCells[i];
                            for (j=i; j >= 0; --j) {
                                if (colorCells[j] == axStartAppPixel+j-i)
                                    axNumAppPixels++;
                                else
                                    break;
                            }
                            axStartAppPixel -= (axNumAppPixels - 1);
                            ncells -= axNumAppPixels;
                        }
                    }
                    else
					{
						if (ApplixCmap == def_cmap)
							NoColor = TRUE;
					}
				}
			}
            if (NoColor)
			{
                AxNColorLevels = 1;
                AxNColors = 2;
            }
        }
    }

    if (allocedCells > 0)
        axFreeUnusedColors(def_cmap, colorCells, allocedCells, ncells);
#if 0
    /* now create an array that contains colors based on pixel values */
    for (i = 0; i < N_AX_PIX; i++)
        AxColors[AxPixels[i]] = i;
#endif
done:

    /* set the global application color indicators */
    UseWidgetColors = (DpyDepth != 1 && AxNoWidgetColor == FALSE);
    ColorDisplay = (DpyDepth != 1 && NoColor == FALSE);
}

void initAlisPixels ()
{
    AxGetRealPixel(0, PelR[2], PelG[2], PelB[2], 0, 0xffff, &AlisPixels[2]);
    AxGetRealPixel(0, PelR[3], PelG[3], PelB[3], 0, 0xffff, &AlisPixels[3]);
    AxGetRealPixel(0, PelR[4], PelG[4], PelB[4], 0, 0xffff, &AlisPixels[4]);
    AxGetRealPixel(0, PelR[5], PelG[5], PelB[5], 0, 0xffff, &AlisPixels[5]);
    AxGetRealPixel(0, PelR[6], PelG[6], PelB[6], 0, 0xffff, &AlisPixels[6]);
    AxGetRealPixel(0, PelR[7], PelG[7], PelB[7], 0, 0xffff, &AlisPixels[7]);
    AxGetRealPixel(0, PelR[8], PelG[8], PelB[8], 0, 0xffff, &AlisPixels[8]);
    AxGetRealPixel(0, PelR[9], PelG[9], PelB[9], 0, 0xffff, &AlisPixels[9]);
    AxGetRealPixel(0, PelR[10], PelG[10], PelB[10], 0, 0xffff, &AlisPixels[10]);
    AxGetRealPixel(0, PelR[11], PelG[11], PelB[11], 0, 0xffff, &AlisPixels[11]);
    AxGetRealPixel(0, PelR[12], PelG[12], PelB[12], 0, 0xffff, &AlisPixels[12]);
    AxGetRealPixel(0, PelR[13], PelG[13], PelB[13], 0, 0xffff, &AlisPixels[13]);
    AxGetRealPixel(0, PelR[14], PelG[14], PelB[14], 0, 0xffff, &AlisPixels[14]);
    AxGetRealPixel(0, PelR[15], PelG[15], PelB[15], 0, 0xffff, &AlisPixels[15]);
}

get_2_level_color(val, max) /* 0, 100 */
int *val, max;
{
    int temp_color;

    temp_color = 100 * *val / max;
    if ( temp_color < 25 )
        *val = 0;
    else
        *val = 1;
}

get_3_level_color(val, max) /* 0, 50, 100 */
int *val, max;
{
    int temp_color;

    temp_color = 100 * *val / max;
    if ( temp_color < 25 )
        *val = 0;
    else if ( temp_color < 82 )
        *val = 1;
    else
        *val = 2;
}

get_4_level_color(val, max) /* 0, 50, 75, 100 */
int *val, max;
{
    int temp_color;

    temp_color = 100 * *val / max;
    if (temp_color < 25)
        *val = 0;
    else if (temp_color < 61)
        *val = 1;
    else if (temp_color < 82)
        *val = 2;
    else
        *val = 3;
}

get_5_level_color(val, max) /* 0, 50, 75, 87, 100 */
int *val, max;
{
    int temp_color;

    temp_color = 100 * *val / max;
    if (temp_color < 25)
        *val = 0;
    else if (temp_color < 61)
        *val = 1;
    else if (temp_color < 82)
        *val = 2;
    else if (temp_color < 92)
        *val = 3;
    else
        *val = 4;
}

get_6_level_color(val, max) /* 0, 50, 75, 87, 95, 100 */
int *val, max;
{
    int temp_color;

    temp_color = 100 * *val / max;
    if (temp_color < 25)
        *val = 0;
    else if (temp_color < 61)
        *val = 1;
    else if (temp_color < 82)
        *val = 2;
    else if (temp_color < 92)
        *val = 3;
    else if (temp_color < 99)
        *val = 4;
    else
        *val = 5;
}



/*************************************************************************
function: WMsetColors - set system pixel values by id
*/
void WMsetColors (fg, bg)
int fg, bg;
{
    WMsetPixels(fg, bg);
    if ( Gc1 ) {
        XSetForeground(Dpy, Gc1, FGpixel);
        XSetBackground(Dpy, Gc1, BGpixel);
    }
}
 
/*************************************************************************
function: WMsetPixels - set system pixel values by id
*/
WMsetPixels(fg, bg)
int fg, bg;
{
    unsigned long black, white;

    if (fg == TMA_DEF)
        fg = TMA_BLACK;

    if (bg == TMA_DEF)
        bg = TMA_WHITE;

    /* the ALIS colors end with color 15 */
    FGpixel =  AlisPixels[fg];
    BGpixel =  AlisPixels[bg];
}


#define colorEQ(c1, c2)     ((abs((int)c1 - (int)c2) < 512) ? 1 : 0)

int
AxStdAllocColors (stdDefs, targetDefs, num)
XColor *stdDefs;
XColor *targetDefs;
int num;
{
    register int i, l;

	if ( axPreDefinedStdmap ) {
		for (i=0; i < num; ++i) 
			targetDefs[i].pixel = convertStdRGBToPixel(targetDefs);
    }
    else {
        for (i=0; i < num; ++i) {
            targetDefs[i].pixel = 0;
            for (l=APP_BASE_POS; l < 216+APP_BASE_POS; ++l) {
                if (colorEQ(targetDefs[i].red, stdDefs[l].red) &&
                        colorEQ(targetDefs[i].green, stdDefs[l].green) &&
                        colorEQ(targetDefs[i].blue, stdDefs[l].blue))  {
                    targetDefs[i].pixel = stdDefs[l].pixel;
                    break;
                }
            }
        }
    }
}

/***************************************************************************
function: AxGetStandardColormap - returns true if a valid standard colormap is
                                  found 
*/
static int
AxGetStandardColormap(dpy, vis, stdcmaps)
Display *dpy;
Visual *vis;
XStandardColormap  **stdcmaps;
{
	register int i;
	XStandardColormap  *s;
	/* is a colormap atom defined */
	if ( axXA_RGB_APPLIX_MAP == (Atom)0 )
		return(FALSE);

	/* are we PseudoColor (we only use colormap on PseudoColor displays) */
	if ( vis->class != PseudoColor )
		return(FALSE);

	/* is the atom undefined */
	if ( !XGetRGBColormaps(dpy, RootWindow(dpy, DefaultScreen(dpy)),
	                       stdcmaps,
	                       &AxNStdCmaps,
	                       axXA_RGB_APPLIX_MAP))
		return(FALSE);

	if ( axXA_RGB_APPLIX_MAP == XA_RGB_DEFAULT_MAP )
	{
		for (i=0, s = (*stdcmaps);
		    (i < AxNStdCmaps) && (s->visualid != vis->visualid); i++, s++)
		{
			;
		}
		if ( i == AxNStdCmaps )
			return(FALSE);
		AxStdCmapInfo = s;
	}
	else
	{
		AxStdCmapInfo = *stdcmaps;
		/* does the atom contain a colormap */
		if ( !AxStdCmapInfo->colormap )
			return(FALSE);

		/* do the visuals match */
		if (AxStdCmapInfo->visualid != vis->visualid )
			return (FALSE); /* registered for different visual */
	}
	/* done */
	return (TRUE);
}


/***************************************************************************
function: AxDeleteStandardColormap - free resources of standard colormap and
                    delete standard colormap property
*/
void
AxDeleteStandardColormap (dpy, win, property)
Display *dpy;
Window win;
Atom property;
{
    XStandardColormap *stdcmaps, *s;
    int count=0;

    stdcmaps = NULL;
    if (XGetRGBColormaps(dpy, win, &stdcmaps, &count, property)) {
        for(s=stdcmaps; count > 0; --count, ++s) {
            if ((s->killid == ReleaseByFreeingColormap) &&
				(s->colormap != None) &&
				(s->colormap != DefaultColormap(dpy, DefaultScreen(dpy))))
                XFreeColormap(dpy, s->colormap);
            else if (s->killid != None)
                XKillClient(dpy, s->killid);
        }
        if ( stdcmaps )
            XFree((char *)stdcmaps);
		XDeleteProperty(dpy, win, property);
		XSync(dpy, False);
    }
}

/***************************************************************************
function: AxChangeNumAsterxAtomData - returns updated number of asterxs running
*/
static int
AxChangeNumAsterxAtomData (val, delete)
int val;
int delete;
{
    Display *newDpy;
    Atom actual_type;
    int actual_format;
    unsigned long nitems, leftover;
    int *data;
    int retval;
	register int status;
    int freeit=FALSE;

    if ( axXA_RGB_APPLIX_MAP == (Atom)0 || axXA_RGB_APPLIX_MAP != APPLIX_MAP )
        return 0;

	retval = 0;
	newDpy = XOpenDisplay(AxDisplayName);
	if ( newDpy )
	{
    	status =  XGetWindowProperty(newDpy,
									 RootWindow(newDpy, DefaultScreen(newDpy)),
									 axXA_NUM_ASTERX_RUN, 0L, 1000000L, FALSE,
									 XA_ATOM, &actual_type, &actual_format,
									 &nitems, &leftover, (unsigned char **)&data);
		if ( status == Success )
		{
			if (!data)
			{
				data = (int *)malloc(sizeof(int));
				*data = 0;
				freeit = TRUE;
			}
    		if ( actual_format != 32 || actual_type != XA_ATOM || nitems != 1)
        		*data = 0;
    		if (*data + val >= 0)
        		*data += val;
    		retval = *data;
    		if (retval || !delete) 
			{
				XChangeProperty(newDpy,
								RootWindow(newDpy, DefaultScreen(newDpy)),
								axXA_NUM_ASTERX_RUN, XA_ATOM, 32,
								PropModeReplace, (char *)data, 1);
			}
			else if ( delete ) 
			{
				/* delete the properties */
				AxDeleteStandardColormap(newDpy,
									 RootWindow(newDpy, DefaultScreen(newDpy)),
									 axXA_RGB_APPLIX_MAP);
				XDeleteProperty(newDpy,
								RootWindow(newDpy, DefaultScreen(newDpy)),
								axXA_NUM_ASTERX_RUN);
			}

			if (freeit)
				free(data);
			else
				XFree((char*)data);
		}
		XCloseDisplay(newDpy);
	}
    return retval;
}

/***************************************************************************
function: AxDestroyStandardColormap - deletes standard colormap atom if caller
            is last instance of asterx
*/
void
AxDestroyStandardColormap ()
{
    if (!AxVisual || AxVisual->class != PseudoColor ||
            axXA_RGB_APPLIX_MAP == (Atom)0)
        return;
    AxChangeNumAsterxAtomData(-1, TRUE);
    return;
}

/***************************************************************************
function: InitializePseudoColormap - allocates read/write color cells
*/
void
InitializePseudoColormap (cmapWindow)
Window cmapWindow;
{
    register int i;
    Colormap def_cmap;      /* default color map */
    XColor AxDefs[256];
    int nwidgetColors;
	XStandardColormap  *stdcmaps;

    def_cmap = DefaultColormap(Dpy, DefaultScreen(Dpy));
    ApplixCmap = def_cmap;

    /* put a cap on available cube size */
    if (DpyDepth == 4) {
        AxNColorLevels = MIN(2, AxNColorLevels);
        AxNColors = MAX(2, AxNColorLevels * AxNColorLevels * AxNColorLevels);
    }

    /* see if a standard colormap is already loaded */
	stdcmaps = NULL;
    if ( AxGetStandardColormap(Dpy, AxVisual, &stdcmaps) )
	{
        if ( !AxStandardColormapWasLoaded )
            AxStandardColormapWasLoaded = COLORMAP_LOADED;

		if (axPreDefinedStdmap )
            ApplixCmap = AxStdCmapInfo->colormap;
        else
		{
            /*  copy standard colormap to a private colormap, this mechanism
                allows to change colors of widgetpixels change dynamically
                Warning: this causes flashing
            */
            for (i = 0; i < 256; i++)
                AxDefs[i].pixel = i;
            XQueryColors(Dpy, AxStdCmapInfo->colormap,
							  AxDefs, MIN(AxVisual->map_entries,256));

            ApplixCmap = XCreateColormap(Dpy, cmapWindow, AxVisual, AllocAll);
            XStoreColors(Dpy, ApplixCmap, AxDefs, MIN(AxVisual->map_entries,256));
        }
        return;
    }

    /* now init the colormap based on what's loaded into the default map */
    if (AxLoadWidgetColors && AxLoadApplicationColors && ApplixCmap == def_cmap) {
        /* allocates color cells off default cmap in 'allocatePixels()' */
        /* stores application colors in default cmap in 'InitApplicationColors()' */
        /* stores widget colors in 'WidgetColors()' */

        if (axXA_RGB_APPLIX_MAP == (Atom)0)
            return;

        /* copy the colors from the default colormap */
        for (i = 0; i < 256; i++)
            AxDefs[i].pixel = i;
        XQueryColors(Dpy, def_cmap, AxDefs, MIN(AxVisual->map_entries, 256));

        /* create a private colormap: note using default visual */
        ApplixCmap = XCreateColormap(Dpy, cmapWindow, AxVisual, AllocAll);
        XStoreColors(Dpy, ApplixCmap, AxDefs, MIN(256, AxVisual->map_entries));
        return;
    }
    if (DpyDepth >= 8) { /* DirectColor or PseudoColor */
        /* copy the colors from the default colormap */
        for (i = 0; i < 256; i++)
            AxDefs[i].pixel = i;
        XQueryColors(Dpy, def_cmap, AxDefs, MIN(AxVisual->map_entries, 256));
        if (AxVisual->map_entries < 256) {
            for (i=AxVisual->map_entries; i < 256; ++i) {
                AxDefs[i].red = 0xFFFF;
                AxDefs[i].green = 0xFFFF;
                AxDefs[i].blue = 0xFFFF;
            }
        }

        /* create a private colormap: note using default visual */
        ApplixCmap = XCreateColormap(Dpy, cmapWindow, AxVisual, AllocAll);
        XStoreColors(Dpy, ApplixCmap, AxDefs, MIN(AxVisual->map_entries,256));
    }
    else if (DpyDepth == 4) {
        /* copy the colors from the default colormap */
        for (i = 0; i < 16; i++)
            AxDefs[i].pixel = i;
        XQueryColors(Dpy, def_cmap, AxDefs, MIN(AxVisual->map_entries, 16));
        if (AxVisual->map_entries < 16) {
            for (i=AxVisual->map_entries; i < 16; ++i) {
                AxDefs[i].red = 0xFFFF;
                AxDefs[i].green = 0xFFFF;
                AxDefs[i].blue = 0xFFFF;
            }
        }

        /* create a private colormap: note using default visual */
        ApplixCmap = XCreateColormap(Dpy, cmapWindow, AxVisual, AllocAll);
        XStoreColors(Dpy, ApplixCmap, AxDefs, MIN(AxVisual->map_entries,16));
    }
}


/***************************************************************************
function: axGetTrueColorPixel - returns pixel value for given color for 
                    TrueColor visual class
*/
unsigned long
axGetTrueColorPixel (cdef, pixel)
XColor *cdef;
unsigned long *pixel;
{
    XAllocColor(Dpy, ApplixCmap, cdef);
    *pixel = cdef->pixel;
#if 0
    else    /* XAllocColor failed why --> serious error */
        *pixel = BlackPixel(Dpy, DefaultScreen(Dpy));
#endif
    return (*pixel);
}

/***************************************************************************
function: cmpDirClrs - compare two colors for "qsort"
*/
static int
cmpDirClrs (c1, c2)
AXDIRECTCOLOR *c1, *c2;
{
    return(c1->rgb - c2->rgb);
}

/***************************************************************************
function: AxInitDirectClrTable - initialise DirectClrTable and sort it
*/
void
AxInitDirectClrTable (clrDefs, ncolors)
XColor *clrDefs;
int ncolors;
{
    register int i;
    static int inited = FALSE;

    if (AxVisual->class != DirectColor)
        return;

    if (!inited) {
        nAxDirectColors = ncolors;
        AxDirectClrTable = (AXDIRECTCOLOR *)malloc(ncolors * sizeof(AXDIRECTCOLOR));
        for (i=0; i < ncolors; ++i) {
            AxDirectClrTable[i].rgb =
                packedRGB(clrDefs[i].red, clrDefs[i].green, clrDefs[i].blue);
            AxDirectClrTable[i].pixel = clrDefs[i].pixel;
        }
        qsort(AxDirectClrTable, nAxDirectColors, sizeof(AXDIRECTCOLOR), cmpDirClrs);
    }
    inited = TRUE;
}

/***************************************************************************
function: axAddColorToDirectClrTable - add a color to DirectClrTable if
            it is not there
*/
static unsigned long
axAddColorToDirectClrTable (cdef, rgb)
XColor *cdef;
int rgb;
{
    int min, mid, max;
    unsigned long pixel;
    unsigned long plane_masks[1];

    min = 0;
    max = nAxDirectColors-1;
    while (min < max) {
        mid = (min + max)/2;
        if (rgb < AxDirectClrTable[mid].rgb)
            max = mid - 1;
        else if (rgb > AxDirectClrTable[mid].rgb)
            min = mid + 1;
        else
            break;
    }
    if (min > max) {        /* not found */
        XAllocColorCells(Dpy, ApplixCmap, FALSE, plane_masks, 0, &pixel, 1);
        cdef->pixel = pixel;
        XStoreColor(Dpy, ApplixCmap, cdef);

        AxDirectClrTable = (AXDIRECTCOLOR *)realloc(AxDirectClrTable,
                                (nAxDirectColors+1) * sizeof(AXDIRECTCOLOR));
        /* overlapped memcpy */
        bcopy(&AxDirectClrTable[min], &AxDirectClrTable[min+1],
                            (nAxDirectColors-min) * sizeof(AXDIRECTCOLOR));
        AxDirectClrTable[min].rgb = rgb;
        AxDirectClrTable[min].pixel = pixel;
        nAxDirectColors++;
    }
    else {  /* found */
        pixel = AxDirectClrTable[mid].pixel;
        cdef->pixel = pixel;
    }

    return (pixel);
}

/***************************************************************************
function: axGetDirectColorPixel - returns pixel value for a given color
                                for Direct visual class
*/
unsigned long
axGetDirectColorPixel (cdef, rgb, pixel)
XColor *cdef;
int rgb;
unsigned long *pixel;
{
    /* Some servers might allocate readonly colors on Directcolor visual */
    /* see if that works */
    if (!XAllocColor(Dpy, ApplixCmap, cdef)) {
        /* X couldnt allocate readonly color (some servers fail)
         * now, search AxDirectClrTable[] table (already allocated colors) and
         * see if this color already exists, otherwise allocate a read/write
         * color add that to AxDirectClrTable[] table and return pixel value
         */
        /* AxDirectClrTable[] is a sorted in ascending order on pixel */
        axAddColorToDirectClrTable(cdef, rgb);
    }
    *pixel = cdef->pixel;
}

/***************************************************************************
function: ROmap- allocate a read only map
*/
static int
ROmap(dpy, colormap, pixels, cmap_size, ncolors)
Display *dpy;
Colormap colormap;
unsigned long pixels[];
int cmap_size;
int ncolors;
{
	register int p;

	/* try to allocate the entire colormap */
	if (XAllocColorCells(dpy, colormap, 1, (unsigned long *) NULL,
							  (unsigned) 0, pixels, (unsigned) cmap_size))
	return cmap_size;

	/*
	 * Allocate all available cells in the colormap, using a binary
	 * algorithm to discover how many cells we can allocate in the colormap.
	 */

	cmap_size--;
	while ( ncolors <= cmap_size )
	{
		p = ncolors + ((cmap_size - ncolors + 1) / 2);
		if ( XAllocColorCells(dpy, colormap, 1, (unsigned long *) NULL,
								   (unsigned) 0, pixels, (unsigned) p))
		{
			if ( p == cmap_size )
				return p;
			else
			{
				XFreeColors(dpy, colormap, pixels, p, (unsigned long) 0);
				ncolors = p;
			}
		}
		else
			cmap_size = p -1;
	}
	return 0;
}

/***************************************************************************
function: contiguous- 
*/
static int
contiguous(pixels, npixels, ncolors, first, rem)
unsigned long pixels[];
int npixels;
int ncolors;
int *first;
int *rem;
{
	register int i = 1; /* walking index into the pixel array */
	register int count = 1; /* length of sequence discovered so far */

	*rem = npixels - 1;
	*first = 0;
	while ( count < ncolors && ncolors - count <= *rem )
	{
		if (pixels[i-1] + 1 == pixels[i])
			count++;
		else
		{
			count = 1;
			*first = i;
		}
		i++;
		(*rem)--;
	}
	if ( count != ncolors )
		return FALSE;

	return TRUE;
}

/***************************************************************************
function: RWcell - allocate read/write cell
*/
static int
RWcell(dpy, colormap, color, request, pixel, n)
Display *dpy;
Colormap colormap;
XColor *color;
XColor *request;
unsigned long *pixel;
int n;
{
	XFreeColors(dpy, colormap, &(color->pixel), 1, (unsigned long)0);
	if (! XAllocColorCells(dpy, colormap, (Bool) 0, (unsigned long *) NULL,
								(unsigned) 0, pixel, (unsigned) 1))
		return FALSE;
	
	if ( *pixel  != n )
	{
		XFreeColors(dpy, colormap, pixel, 1, (unsigned long) 0);
		return FALSE;
	}
	color->pixel = *pixel;
	color->flags = DoRed | DoGreen | DoBlue;
	color->red = request->red;
	color->green = request->green;
	color->blue = request->blue;
	XStoreColors(dpy, colormap, color, 1);

	return TRUE;
}

static int
free_cells(dpy, colormap, pixels, npixels, p)
Display *dpy;
Colormap colormap;
unsigned long pixels[];
int npixels;
int p;
{
	XFreeColors(dpy, colormap, pixels, p, (unsigned long) 0);
	XFreeColors(dpy, colormap, &(pixels[p+1]),
					 (npixels - p - 1),(unsigned long) 0);
	free((char *) pixels);
}

static int
ROorRWcell(dpy, colormap, pixels, npixels, color, p, n)
Display *dpy;
Colormap colormap;
unsigned long pixels[];
int npixels;
XColor *color;
int p;
int n;
{
	unsigned long   pixel;
	XColor request;

	pixel = (unsigned long) n;
	request.red = color->red;
	request.green = color->green;
	request.blue = color->blue;

	XFreeColors(dpy, colormap, &pixel, 1, (unsigned long) 0);
	if (! XAllocColor(dpy, colormap, color) ||
		  (color->pixel != pixel &&
			(!RWcell(dpy, colormap, color, &request, &pixel, n))))
	{
		free_cells(dpy, colormap, pixels, npixels, p);
		return FALSE;
	}
	return TRUE;
}

static int
compare(e1, e2)
unsigned long   *e1, *e2;
{
	if ( *e1 < *e2 ) return -1;
	if ( *e1 > *e2 ) return 1;
	return 0;
}

static void
gray_allocation(n, red_max, green_max, blue_max)
int n;
unsigned long *red_max, *green_max, *blue_max;
{
	*red_max = (n * 30) / 100;
	*green_max = (n * 59) / 100;
	*blue_max = (n * 11) / 100;
	*green_max += ((n - 1) - (*red_max + *green_max + *blue_max));
}

static int
default_allocation(visual, red, green, blue)
Visual *visual;
unsigned long *red, *green, *blue;
{
	if ( visual->map_entries < 250 )
		return FALSE;
	if ( visual->map_entries > 500000 )
	{
		/* intended for displays with 24 planes */
		*red = *green = *blue = (unsigned long) 63;
	}
	else if ( visual->map_entries > 4000 )
	{
		/* intended for displays with 12 planes */
		*red = *green = *blue = (unsigned long) 12;
	}
	else
	{
		/* intended for displays with 8 planes */
		*red = *green = *blue = (unsigned long)4;
	}
	return TRUE;
}

static void
best_allocation(visual, red, green, blue)
Visual *visual;
unsigned long *red, *green, *blue;
{
	register int bits, n;

	n = 1;
	bits =0;

	while ( visual->map_entries > n )
	{
		n = n << 1;
		bits++;
	}
	if ( n == visual->map_entries )
	{
		register int r, g, b;
		b = bits / 3;
		g = b + ((bits % 3) ? 1 : 0);
		r = b + (((bits % 3) == 2) ? 1 : 0);
		*red = 1 << r;
		*green = 1 << g;
		*blue = 1 << b;
	}
	else
	{
		*red = icbrt_with_bits(visual->map_entries, bits);
		*blue = *red;
		*green = (visual->map_entries / ((*red) * (*blue)));
	}
	(*red)--;
	(*green)--;
	(*blue)--;
}

static
int icbrt_with_guess(a, guess)
{
	register int delta;

	if ( a <= 0 )
		return 0;
	if ( guess < 1 )
		guess = 1;
	
	do {
		delta = (guess - a/(guess*guess))/3;
		guess -= delta;
	} while (delta != 0);

	if ( guess*guess*guess > a )
		guess--;

	return guess;
}

static int 
icbrt_with_bits(a, bits)
{
	return icbrt_with_guess(a, a>>2*bits/3);
}

static int
icbrt(a)
{
	register int bits = 0;
	register unsigned n = a;

	while ( n )
	{
		bits++;
		n >>= 1;
	}
	return icbrt_with_bits(a, bits);
}

static int
AXmuCreateColormap(dpy, stdcmap)
Display *dpy;
XStandardColormap *stdcmap;
{
	int r, g, gmax, max_v, i, n;
	XColor color;
    Visual *visual;
	int first_index;
	int remainder;
	int npixels;
	int ncolors;
	unsigned long *pixels;

	ncolors = stdcmap->red_max * stdcmap->red_mult +
			  stdcmap->green_max * stdcmap->green_mult +
			  stdcmap->blue_max * stdcmap->blue_mult + 1;

    visual = DefaultVisual(dpy, DefaultScreen(dpy));

	if ( ncolors <= 1 || ncolors > visual->map_entries )
		return FALSE;
	pixels = (unsigned long *)calloc(1,visual->map_entries*sizeof(unsigned long));
	npixels = ROmap(dpy, stdcmap->colormap, pixels,
								  visual->map_entries, ncolors);
	if ( npixels == 0 )
	{
		free((char *) pixels);
		return FALSE;
	}
	qsort((char *) pixels, npixels, sizeof(unsigned long), compare);

	if (! contiguous(pixels, npixels, ncolors, &first_index, &remainder))
	{
		/* can't find enough contiguous cells, give up */
		XFreeColors(dpy, stdcmap->colormap, pixels, npixels, (unsigned long) 0);
		free((char *) pixels);
		return FALSE;
	}
	stdcmap->base_pixel = pixels[first_index];

	/* construct a gray map */
	if ( stdcmap->red_mult == 1 && stdcmap->green_mult == 1 &&
		 stdcmap->blue_mult == 1 )
	{
		max_v = stdcmap->red_max + stdcmap->green_max + stdcmap->blue_max;
		for ( n = stdcmap->base_pixel,i=0; i< ncolors; i++, n++ )
		{
			color.pixel = (unsigned long) n;
			color.blue = color.green = color.red =
			(unsigned short) ((i * 65535) / max_v);
			if ( ! ROorRWcell(dpy, stdcmap->colormap, pixels, 
									  npixels, &color, first_index + i, n))
				return FALSE;
		}
	}
	/* construct a red ramp map */
	else if ( stdcmap->green_max == 0 && stdcmap->blue_max == 0 )
	{
		for ( n = stdcmap->base_pixel,i=0; i< ncolors; i++, n++ )
		{
			color.pixel = (unsigned long) n;
			color.red = (unsigned short) ((i * 65535) / stdcmap->red_max);
			color.blue = color.green = 0;
			if ( ! ROorRWcell(dpy, stdcmap->colormap, pixels, 
									  npixels, &color, first_index + i, n))
				return FALSE;
		}
	}
	/* construct a green ramp map */
	else if ( stdcmap->red_max == 0 && stdcmap->blue_max == 0 )
	{
		for ( n = stdcmap->base_pixel,i=0; i< ncolors; i++, n++ )
		{
			color.pixel = (unsigned long) n;
			color.green = (unsigned short) ((i * 65535) / stdcmap->green_max);
			color.red = color.blue = 0;
			if ( ! ROorRWcell(dpy, stdcmap->colormap, pixels, 
									  npixels, &color, first_index + i, n))
				return FALSE;
		}
	}
	/* construct a blue ramp map */
	else if ( stdcmap->red_max == 0 && stdcmap->green_max == 0 )
	{
		for ( n = stdcmap->base_pixel,i=0; i< ncolors; i++, n++ )
		{
			color.pixel = (unsigned long) n;
			color.blue = (unsigned short) ((i * 65535) / stdcmap->blue_max);
			color.red = color.green = 0;
			if ( ! ROorRWcell(dpy, stdcmap->colormap, pixels, 
									  npixels, &color, first_index + i, n))
				return FALSE;
		}
	}
	/* construct a standard red green blue cube map */
	else
	{
		r = stdcmap->red_mult;
		g = stdcmap->green_mult;
		gmax = stdcmap->green_max + 1;

		for ( n = stdcmap->base_pixel,i=0; i< ncolors; i++, n++ )
		{
			color.pixel = (unsigned long) n;
			color.red = (unsigned short) (((i/r) * 65535) / stdcmap->red_max);
			color.green = (unsigned short) ((((i/g)%gmax) * 65535) /
											stdcmap->green_max);
			color.blue = (unsigned short) (((i%g) * 65535) /
											stdcmap->blue_max);
			if ( ! ROorRWcell(dpy, stdcmap->colormap, pixels, 
									  npixels, &color, first_index + i, n))
				return FALSE;
		}
	}

	if ( first_index )
		XFreeColors(dpy, stdcmap->colormap, pixels,
						 first_index, (unsigned long) 0);
	if ( remainder )
		XFreeColors(dpy, stdcmap->colormap, &(pixels[first_index + ncolors]),
						 remainder, (unsigned long) 0);
	free((char *) pixels);
	return TRUE;
}

XStandardColormap *AXmuLookupStandardColormap(newDpy, property)
Display *newDpy;
Atom property;
{
	XStandardColormap  *stdcmap;
	Window win;
    Visual *visual;
	int screen, defDepth;
	int status, rmax, gmax, bmax;

	screen = DefaultScreen(newDpy);
    visual = DefaultVisual(newDpy, screen);
    defDepth = DisplayPlanes(newDpy, screen);


	if ( visual->map_entries <= 2 )
		return FALSE;
	
	switch ( property )
	{
		case XA_RGB_DEFAULT_MAP: 
			status = default_allocation(visual, &rmax, &gmax, &bmax);
		break;
		case XA_RGB_BEST_MAP:
			best_allocation(visual, &rmax, &gmax, &bmax);
		break;
		case XA_RGB_GRAY_MAP:
			gray_allocation(visual->map_entries, &rmax, &gmax, &bmax);
		break;
		case XA_RGB_RED_MAP:
			rmax = visual->map_entries -1;
			gmax = bmax = 0;
		break;
		case XA_RGB_GREEN_MAP:
			gmax = visual->map_entries -1;
			rmax = bmax = 0;
		break;
		case XA_RGB_BLUE_MAP:
			bmax = visual->map_entries -1;
			rmax = gmax = 0;
		break;
	}
	stdcmap = XAllocStandardColormap();
	if ( stdcmap == NULL )
		return FALSE;

	if ( property == XA_RGB_DEFAULT_MAP )
	{
		win = XCreateWindow(newDpy, RootWindow(newDpy, screen), 1, 1, 1, 1,
								 0, 0, InputOnly, visual,
								 (unsigned long) 0,
								 (XSetWindowAttributes *)NULL);
		stdcmap->killid  = (XID) XCreatePixmap(newDpy, win, 1, 1, defDepth);
		XDestroyWindow(newDpy, win);
		stdcmap->colormap = DefaultColormap(newDpy, screen);
	}
	else
	{
		stdcmap->killid = ReleaseByFreeingColormap;
		stdcmap->colormap = XCreateColormap(newDpy, RootWindow(newDpy, screen),
													visual, AllocNone);
	}
	stdcmap->visualid = visual->visualid;
	stdcmap->red_max = rmax;
	stdcmap->green_max = gmax;
	stdcmap->blue_max = bmax;
	stdcmap->base_pixel = 0;

	if ( property == XA_RGB_GRAY_MAP )
	{
		stdcmap->red_mult = 1;
		stdcmap->green_mult = 1;
		stdcmap->blue_mult = 1;
	}
	else
	{
		stdcmap->red_mult = (rmax > 0) ? (gmax +1) *(bmax +1) : 0;
		stdcmap->green_mult = (gmax > 0) ? bmax + 1 : 0;
		stdcmap->blue_mult = (bmax > 0) ? 1 : 0;
	}

	/* make the colormap */
	status = AXmuCreateColormap(newDpy, stdcmap);

	if ( !status )
	{
		if ( stdcmap->killid == ReleaseByFreeingColormap )
			XFreeColormap(newDpy, stdcmap->colormap);
		else if ( stdcmap->killid != None )
			XFreePixmap(newDpy, stdcmap->killid);
		XFree((char *) stdcmap);
		return (XStandardColormap *) NULL;
	}
	return stdcmap;
}

/***************************************************************************
function: AxSetRGBColormaps - register the colormap
*/
static void
AxSetRGBColormaps(newDpy, visual, stdcmaps)
Display *newDpy;
Visual *visual;
XStandardColormap  *stdcmaps;
{
	XStandardColormap  *s, *m, *maps;
	register int i, status;

	if ( axXA_RGB_APPLIX_MAP != XA_RGB_DEFAULT_MAP )
		XSetRGBColormaps(newDpy, RootWindow(newDpy, DefaultScreen(newDpy)),
						 AxStdCmapInfo, 1, axXA_RGB_APPLIX_MAP);
	else
	{

		for (i=0, s=stdcmaps;
				  (i< AxNStdCmaps) && (s->visualid != visual->visualid);
				  i++, s++)
		{
			;
		}
		/* only default map can have more than one definition */
		if ( i == AxNStdCmaps )
		{
			s = (XStandardColormap *) malloc((unsigned)
								 ((AxNStdCmaps+1) * sizeof(XStandardColormap)));
			for (i = 0, m = s,maps = stdcmaps; i<AxNStdCmaps;i++,m++,maps++)
			{ 
				m->colormap   = maps->colormap;
				m->red_max    = maps->red_max;
				m->red_mult   = maps->red_mult;
				m->green_max  = maps->green_max;
				m->green_mult = maps->green_mult;
				m->blue_max   = maps->blue_max;
				m->blue_mult  = maps->blue_mult;
				m->base_pixel = maps->base_pixel;
				m->visualid   = maps->visualid;
				m->killid     = maps->killid;
			}
			m->colormap   = AxStdCmapInfo->colormap;
			m->red_max    = AxStdCmapInfo->red_max;
			m->red_mult   = AxStdCmapInfo->red_mult;
			m->green_max  = AxStdCmapInfo->green_max;
			m->green_mult = AxStdCmapInfo->green_mult;
			m->blue_max   = AxStdCmapInfo->blue_max;
			m->blue_mult  = AxStdCmapInfo->blue_mult;
			m->base_pixel = AxStdCmapInfo->base_pixel;
			m->visualid   = AxStdCmapInfo->visualid;
			m->killid     = AxStdCmapInfo->killid;
			XSetRGBColormaps(newDpy, RootWindow(newDpy, DefaultScreen(newDpy)),
									 s, ++AxNStdCmaps, XA_RGB_DEFAULT_MAP);
		}
	}
}

/***************************************************************************
function: AxCreateStandardColormap - the cmap MUST NOT exist before this call
*/
static int
AxCreateStandardColormap (newDpy, property)
Display *newDpy;
Atom property;
{
    Visual *visual;
    Colormap def_cmap;
    XColor clrDefs[256];
    int ndefs, i, defDepth, status;

	if ( AxCreateCmap == FALSE )
	{
		axXA_RGB_APPLIX_MAP = (Atom)0;
		axPreDefinedStdmap = FALSE;
		return;
	}

	if ( axPreDefinedStdmap )
	{
		AxStdCmapInfo = AXmuLookupStandardColormap(newDpy, property);
		if ( AxStdCmapInfo == (XStandardColormap *)NULL )
			return FALSE;
	}
	else /* xa_rgb_applix_map  */
	{
		visual = DefaultVisual(newDpy, DefaultScreen(newDpy));
		def_cmap = DefaultColormap(newDpy, DefaultScreen(newDpy));
		defDepth = DisplayPlanes(newDpy, DefaultScreen(newDpy));

		AxStdCmapInfo=(XStandardColormap *)calloc(1,sizeof(XStandardColormap));
		if ( AxStdCmapInfo == (XStandardColormap *)NULL )
			return FALSE;
    	if ( defDepth >= 8 )   /* 3/3/2 model */
		{
			AxStdCmapInfo->red_max = 7;
			AxStdCmapInfo->red_mult = 32;
			AxStdCmapInfo->green_max = 7;
			AxStdCmapInfo->green_mult = 4;
			AxStdCmapInfo->blue_max = 3;
			AxStdCmapInfo->blue_mult = 1;
			ndefs = 256;

            for (i = 0; i < ndefs; i++ )
                clrDefs[i].pixel = i;
            XQueryColors(newDpy, def_cmap, clrDefs,
								 MIN(256, visual->map_entries));
            AxGetStdcmapAppColors(clrDefs, APP_BASE_POS);

        	/* create a new colormap: note using default visual */
        	AxStdCmapInfo->colormap = XCreateColormap(newDpy,
                                    RootWindow(newDpy, DefaultScreen(newDpy)),
                                    visual, AllocAll);
        	if (AxStdCmapInfo->colormap == None) 
			{
            	free((char*)AxStdCmapInfo);
				AxStdCmapInfo = NULL;
            	return FALSE;
        	}
        	XStoreColors(newDpy, AxStdCmapInfo->colormap, clrDefs,
							 	MIN(ndefs, visual->map_entries));
        }
    	else if (defDepth == 4)  /* 2/2/2 */
		{
        	/* copy the colors from the cmap */
        	ndefs = 16;
        	for (i = 0; i < ndefs; i++)
            	clrDefs[i].pixel = i;
        	XQueryColors(newDpy, def_cmap, clrDefs,
							 	MIN(ndefs, visual->map_entries));

        	/* create a new colormap: note using default visual */
        	AxStdCmapInfo->colormap = XCreateColormap(newDpy,
									RootWindow(newDpy, DefaultScreen(newDpy)),
									visual, AllocAll);
	
        	if (AxStdCmapInfo->colormap == None)
			{
            	free((char*)AxStdCmapInfo);
				AxStdCmapInfo = NULL;
            	return FALSE;
        	}
	
        	XStoreColors(newDpy, AxStdCmapInfo->colormap, clrDefs,
							 	MIN(ndefs, visual->map_entries));
	
        	AxStdCmapInfo->red_max = 1;
        	AxStdCmapInfo->red_mult = 4;
        	AxStdCmapInfo->green_max = 1;
        	AxStdCmapInfo->green_mult = 2;
        	AxStdCmapInfo->blue_max = 1;
        	AxStdCmapInfo->blue_mult = 1;
    	}
		/* set visual id of standard colormap */
		AxStdCmapInfo->visualid = visual->visualid;

		/* standard colormap should be deleted by freeing colormap */
		AxStdCmapInfo->killid = ReleaseByFreeingColormap;
    }
	return TRUE;
}

/***************************************************************************
function: AxVerifyStandardColormap - if it doesn't exist, create it
*/
static void
AxVerifyStandardColormap(newDpy, visual)
Display *newDpy;
Visual *visual;
{
	XStandardColormap *stdcmaps;

	/* sanity check */
	if ( visual->class != PseudoColor || axXA_RGB_APPLIX_MAP == (Atom)0 )
		return;

	/* increment the number of users running against APPLIX map */
	AxChangeNumAsterxAtomData(1, FALSE);

    /* grab the server, to avoid race for registering standard colormaps */
    XGrabServer(newDpy);

	/* see if already registered */
	stdcmaps = NULL;
	if ( AxGetStandardColormap(newDpy, visual, &stdcmaps) )
	{
		XFree((char *)stdcmaps);
		stdcmaps = NULL;
	}
	else /* create a new standard colormap */
	{
		/* register standard colormap on successful create */
		if ( AxCreateStandardColormap(newDpy, axXA_RGB_APPLIX_MAP) )
		{
			AxSetRGBColormaps(newDpy, visual, stdcmaps);
			XSetCloseDownMode(newDpy, RetainPermanent);
			AxStandardColormapWasLoaded = COLORMAP_CREATED;
		}
	}

    /* ungrab the server and close display */
    XUngrabServer(newDpy);

}

/***************************************************************************
function: AxReadColormapName - read the profile for std cmap
*/
static void
AxReadColormapName(newDpy)
Display *newDpy;
{
	char *profile = NULL, *GetProfileByCode();

	/* init */
	axXA_RGB_APPLIX_MAP = (Atom)0;
	APPLIX_MAP = (Atom)-1;

	/* read "axStandardColormap" (default is AX_XA_RGB_APPLIX_MAP_STR) */
	if ( !JavaMode )
		profile = GetProfileByCode(PROF_STANDARD_COLORMAP_);
	if ( !profile || !profile[0] )
		AxStdCmapName = allocStr(AX_XA_RGB_APPLIX_MAP_STR);
	else
	{
		/* use the atom name specified in the profile */
		AxStdCmapName = allocStr(profile);

		/* no colormap specified - don't use a colormap */
		if ( !(*AxStdCmapName) )
			axPreDefinedStdmap = FALSE;
		else if ( strcasecmp(AxStdCmapName, NXLT("null")) == 0 )
			axPreDefinedStdmap = FALSE;
		else
		{ 
			axPreDefinedStdmap = TRUE;
			AxCreateCmap = TRUE;

			/* check for standard atom names */
			if (strcmp(AxStdCmapName, NXLT("XA_RGB_DEFAULT_MAP")) == 0)
				axXA_RGB_APPLIX_MAP = XA_RGB_DEFAULT_MAP;
			else if (strcmp(AxStdCmapName, NXLT("XA_RGB_BEST_MAP")) == 0)
				axXA_RGB_APPLIX_MAP = XA_RGB_BEST_MAP;
			else if (strcmp(AxStdCmapName, NXLT("XA_RGB_GRAY_MAP")) == 0)
				axXA_RGB_APPLIX_MAP = XA_RGB_GRAY_MAP;
			else if (strcmp(AxStdCmapName, NXLT("XA_RGB_RED_MAP")) == 0)
				axXA_RGB_APPLIX_MAP = XA_RGB_RED_MAP;
			else if (strcmp(AxStdCmapName, NXLT("XA_RGB_GREEN_MAP")) == 0)
				axXA_RGB_APPLIX_MAP = XA_RGB_GREEN_MAP;
			else if (strcmp(AxStdCmapName, NXLT("XA_RGB_BLUE_MAP")) == 0)
				axXA_RGB_APPLIX_MAP = XA_RGB_BLUE_MAP;
			else
			{
				/* get the atom from the string */
				axXA_RGB_APPLIX_MAP = XInternAtom(newDpy, AxStdCmapName, TRUE);
				AxCreateCmap = FALSE;
			}

			if ( axXA_RGB_APPLIX_MAP == None )
			{
				/* not found, revert to APPLIX map */
				free(AxStdCmapName);
				AxStdCmapName = allocStr(AX_XA_RGB_APPLIX_MAP_STR);
			}
		}
	}

	/* the APPLIX default colormap? */
	if ( strcmp(AxStdCmapName, AX_XA_RGB_APPLIX_MAP_STR) == 0 )
	{
		axXA_RGB_APPLIX_MAP = XInternAtom(newDpy, AxStdCmapName, FALSE);
		axPreDefinedStdmap = FALSE;
		AxCreateCmap = TRUE;
		APPLIX_MAP = axXA_RGB_APPLIX_MAP;
	}
}

/***************************************************************************
function: AxRegisterStandardColormap - register a named colormap
*/
void
AxRegisterStandardColormap ()
{
	Display *newDpy;
	Visual *visual;

	/* get the display info */
	newDpy = XOpenDisplay(AxDisplayName);
	if ( newDpy )
	{
		/* read the axStandardColormap string in the  profile */
		AxReadColormapName(newDpy);

		visual = DefaultVisual(newDpy, DefaultScreen(newDpy));
		if ( visual->class != PseudoColor )
			axXA_NUM_ASTERX_RUN = None;
		else
			axXA_NUM_ASTERX_RUN = XInternAtom(newDpy, AX_XA_NUM_ASTERX_RUN_STR,
			                                  FALSE);

		if ( DisplayPlanes(newDpy, DefaultScreen(newDpy)) >= 4 )
			AxVerifyStandardColormap(newDpy, visual);

		XCloseDisplay(newDpy);
	}
}
