/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmkeys.c
   Applix ASC version: 66
   Date: Thu Apr 29 21:14:32 1999
*/
#include <stdio.h>
#include <ctype.h>
#include <basicwin.h>

#if AX_X11
#include <X11/keysym.h>
#if PMAX || DEC_ALPHA
#include <X11/DECkeysym.h>
#endif
#endif

#include <aos.h>
#include <xim.h>
#include <amenudef.h>
#include <elf_env.h>
#include <tmkeys.h>
#include <getstr4.h>
#include <ximcomm.h>

/* This is a hack define for recognizing the HP-keyboard shifted tab.
 * There's also a keypad flavor of the back tab, but it's bad enuf to
 * define even this -- it's tested for generically and could have the
 * effect of stealing function keys on some keyboard somewhere..
 */
#define HP_BackTab (0x1000FF74)

#define STRALLOC(p) strcpy(malloc(strlen(p)+1),p)
#define SIZEOF_EMPTY_TABLE (sizeof(key_table_type) - sizeof(key_table_entry))

static int Mode;        /* id of current table_type (prefix key) */
unsigned short CurrentMods = 0;     /* current modifier mask */

KeySym xmPreprocessKeysym();
char *xmKeysymInMenubar();

static char *keyInMenuBar();  /* sub of xmKeysymInMenubar */

static key_table_type *ModeList[MAX_MODES];
static KeySym Shifters[MAX_MODIFIERS];
static char *ShifterNames[MAX_MODIFIERS];
static unsigned short ShifterValues[MAX_MODIFIERS];
static int keys_translations_inited = 0;
extern int WeThinkItsSunX;


#define KEY_SHIFT_SHIFT     0x01
#define KEY_SHIFT_CONTROL   0x02
#define KEY_SHIFT_ALT       0x04
#define KEY_SHIFT_META      0x08




/* Asterix key modifier masks - set in xmlib.c ( AxModifierMapping() ) 
 * based on current modifier map.
 * (ShiftMask, LockMask, & ControlMask usually correspond to
 * Shift, CapsLock, & Control keys, but modifier masks for Alt,
 * Meta, etc. are server-dependent. This mapping from X modmap
 * to Asterix modmap eliminates the server-dependency.  The only
 * drawback is that the meaning of the Alt keys cannot be 
 * changed by the user.) */
/* On NT, these are fixed values since the user can not do a 
 * "XModMap" type of keyboard change */
int AxShiftMask = ShiftMask;
int AxLockMask = LockMask;
int AxControlMask = ControlMask;
int AxAltMask = Mod1Mask;      /* good default guess?? */
int AxMetaMask = Mod2Mask;     /* good default guess?? */
int AxNumLockMask = Mod4Mask;  /* good default guess?? */


/* Each app type has its own slot in mode list at high end */

#define APPMODE(id) (MAX_MODES - id - 1)

/*************************************************************************
function: ThimIsShifted - Return whether shift key is down
*/
ThimIsShifted ()
{
    register int ret = FALSE;

#if AXMS
    if ( (MS_GetKeyState(VK_SHIFT)) < 0 )
        ret = TRUE;
#else
    ret = (CurrentMods & KEY_SHIFT_SHIFT) != 0;
#endif
    return( ret );
}

/*************************************************************************
function: ThimIsControled - Return whether control key is down
*/
ThimIsControled ()
{
    register int ret = FALSE;

#if AXMS
    if ( (MS_GetKeyState(VK_CONTROL)) < 0 )
        ret = TRUE;
#else
    ret = (CurrentMods & KEY_SHIFT_CONTROL) != 0;
#endif
    return( ret );
}

/*************************************************************************
function: ThimIsAlted - Return whether alt key is down
*/
ThimIsAlted ()
{
    register int ret = FALSE;

#if AXMS
    if ( (MS_GetKeyState(VK_MENU)) < 0 )
        ret = TRUE;
#else
    ret = (CurrentMods & KEY_SHIFT_ALT) != 0;
#endif
    return( ret );
}


KeySym
xmPreprocessKeysym(ksym)
KeySym ksym;
{
#if AX_X11
    /*******************************************************************\
    * Convert upper case keysyms XK_A...XK_Z to lower case XK_a...XK_z. *
    * This is needed to handle shift-lock, ^x because the shift lock    *
    * causes ^X to be reported.                                         *
    \*******************************************************************/
    if ( (ksym >= XK_A) && (ksym <= XK_Z) ) {
        ksym = XK_a + (ksym - XK_A);
	}
#endif
	return ksym;
}

char *
xmKeysymInMenubar(mP, eventP, ksym)
menu_item_def *mP;
XEvent *eventP;
KeySym ksym;
{
	if ( (CurrentMods == 0) &&
		((ksym >= XK_space) && (ksym <= XK_asciitilde)) ) {
			
		/* simple displayable character, with no modifier bits set */
		/* these keys are never in the menu bar */
		return (char *) NULL;
	}
	return keyInMenuBar(mP,ksym,CurrentMods);
}

extern KeySym AxEuroDollarKeysym;							/* method 1 */
extern int AxEuroDollarModifiers;							/* method 1 */
extern int AxEuroDollarKeyState, AxEuroDollarKeyCode ;		/* method 2 */

/*************************************************************************
function: ProcessKeyEvent - Handle a keyboard activity
Lookup keysym in current table.
 If the key is found and binds to an activity, return TRUE.
 If the key is not found, return FALSE.
 If the key is found and binds to a non-activity (mode change) return FALSE.
*/
int ProcessKeyEvent (mP,eventP,is_text,bufP,bellP)
register menu_item_def *mP;
XEvent *eventP;
int *is_text;   /* returned flag for staright text entry */
char **bufP;
int *bellP; /* ring the bell */
{
    register key_table_entry *kp = ModeList[Mode]->entries;
    register int ix;
    register int size,mode;
    register int nbytes;
    static char buf[XLS_BUF_LEN+1];
    KeySym ksym, ksym_virgin;
    extern int ThimbleSaysToReturnNow; /* hack to break out of XtAppNextEvent */
    extern int THIMhelpMode;

    *bellP = FALSE;

    /* Only interested in downstroke of key */
    if ( eventP->type != KeyPress )
        return(FALSE);

	/* Low-level Euro encoding - keyboard specific */
	if (AxEuroDollarKeyCode && (AxEuroDollarKeyCode == eventP->xkey.keycode) && 
							(AxEuroDollarKeyState == eventP->xkey.state)) {
		/* we have found the euro $ key sequence, return back Euro$ ascii */
		buf[0] = EURO_SYMBOL_8BIT_VALUE; /* We've added the Euro glyph to this
									/* location in the Latin1 TrueType fonts */
		buf[1] = 0;
        *is_text = TRUE;
        *bufP = buf;
        return(1);
	}


    /* Decode event */
    nbytes = AXLookupString(eventP, buf, XLS_BUF_LEN, &ksym);/* OK non-wide*/
    buf[nbytes] = 0;
    XWINSlog("----KEYCODE %x",ksym);
    if ( (nbytes == 0) && (ksym == 0) )
        return(FALSE);

	/* High-level Euro keysym - hardware-independent */
	if (AxEuroDollarKeysym && (AxEuroDollarKeysym == ksym)) {
		int mods = AxDecodeModifiers(eventP->xkey.state);
		if (mods == AxEuroDollarModifiers) {
			buf[0] = EURO_SYMBOL_8BIT_VALUE; 
			buf[1] = 0;
        	*is_text = TRUE;
        	*bufP = buf;
        	return(1);
		}
	}


#if PROFILE_BUILD
    if ( (ksym == XK_F12) || (ksym == 0x1005ff11) )
       /* A hard wired function key to turn profiling on and off */
       /* Some Sun keyboards sure have a weird idea of F12!
            (the hardwired number represents XK_SunXK_F37) */
       toggle_profiling(TRUE);
#endif

    if ( (nbytes == 0) && AxKeysymIsModifier(ksym) )
        /* Ignore non-keys like caps lock */
        return(FALSE);

#if HP300 || HP800
    /** the HP machines return a propritary keysym for the shift-tab **/
    /** sequence.  Make it a tab, since we look for a shifted tab    **/
    if ( ksym == XK_BackTab )
        ksym = XK_Tab; /* make it a tab */
#endif
	if ( ksym == HP_BackTab )
		ksym = XK_Tab; /* same thing, but for the hp-keyboard-non-hp-axmain */
	
#if HP800
    /** since we are now using the HP specific form of XLokupString **/
    /** and associated libraries (see changes in xmlib.c) in the **/
    /** "middle" of a compse sequence the intermediate character **/
    /** comes in as something like "0x100000ab", which we nned to **/
    /** screen out and ignore so that the upper levels do not generate */
    /** a 'beep' because they get an unknown key. - jon               **/
    if ( (ksym >= 0x10000000) && (ksym <= 0x100000FF) )
        return ( FALSE );
#endif

    /* Note modifiers, converting lowercase to uppercase if LockMask is set */
    if((eventP->xkey.state & AxLockMask) || (eventP->xkey.state & AxShiftMask))
    {
        if ( (ksym >= XK_a) && (ksym <= XK_z) )
            ksym = XK_A + (ksym - XK_a);
    }

    ksym_virgin = ksym;
	ksym = xmPreprocessKeysym(ksym_virgin);

#if AX_X11
    if ( ksym == 0 )
        {
        /* Hack for MIPS X-server, ignore kysm 0 */
        return(FALSE);
        }
#endif
	

    AxSetKeyModifiers(eventP->xkey.state);
    if ( mP ) {
        /* We were passed a menu-bar, search the it for the (modified) key */
        /*   if this is not a simple displayable keystroke                */
 		*bufP = xmKeysymInMenubar(mP, eventP, ksym);
		if ( *bufP ) {
			*is_text = FALSE;
			return(TRUE);
		}
    }

    /* pass only non-modified or Capitals modified keys */
    if ( nbytes && !(CurrentMods & ~KEY_SHIFT_SHIFT) )
    {
        *is_text = TRUE;
        *bufP = buf;
        return(nbytes);
    }

#if RIOS
    else if ( ksym_virgin == XK_Mode_switch )
        /* ignore the "alt" key used to generate different keys */
        return(FALSE);
#endif

    else if ( (nbytes == 0) && (CurrentMods == 0) &&
             ((ksym_virgin >= XK_space) && (ksym_virgin  <= XK_asciitilde)) )
        {
        /* keysym is for a printable char, but no bytes were returned, and no */
        /* state bits (like control) are set.                                 */
        /* assume that we are in the middle of a compose sequence, and ignore */
        /* this keystroke.                                             */
        return(FALSE);
        }

#if AX_X11
    else if ( ksym_virgin == XK_Multi_key )
        /* ignore the "compose" key on a SUN keyboard */
        return(FALSE);
#endif

    return(FALSE);
}

/*************************************************************************
function: AxSetKeyModifiers - Set CurrentMods var based on event info
*/
AxSetKeyModifiers (state)
int state;  /* from X event */
{
    CurrentMods = AxDecodeModifiers(state);
}

/*************************************************************************
function: AxDecodeModifiers - Analyze state info for modifier flags
*/
int
AxDecodeModifiers (state)
int state;  /* from X event */
{
    int modifiers = 0;

    if ( state & AxShiftMask )
        modifiers |= KEY_SHIFT_SHIFT;
    if ( state & AxControlMask )
        modifiers |= KEY_SHIFT_CONTROL;
    if ( state & AxAltMask )
        modifiers |= KEY_SHIFT_ALT;

    return( modifiers);
}

/*************************************************************************
function: AxKeysymIsModifier - Return TRUE if ksym is a modifier
*/
AxKeysymIsModifier (ksym)
KeySym ksym;
{
    switch ( ksym ) {
    case XK_Num_Lock:
    case XK_Shift_L:
    case XK_Shift_R:
    case XK_Control_L:
    case XK_Control_R:
    case XK_Caps_Lock:
    case XK_Shift_Lock:
    case XK_Meta_L:
    case XK_Meta_R:
    case XK_Alt_L:
    case XK_Alt_R:
    case XK_Super_L:
    case XK_Super_R:
    case XK_Hyper_L:
    case XK_Hyper_R:
        return(TRUE);
    }
    return(FALSE);
}



/*************************************************************************
function: keyInMenuBar - Return menu bar func bound to key
*/
static char *keyInMenuBar (mP,ksym,modifiers)
register menu_item_def *mP;
register KeySym ksym;
register int modifiers;
{
    char *retval;

    mP++;   /* get past leading mark */
    while ( TRUE ) {
        if ( (mP->type == ITEM_FUNC) && (mP->accel_key == ksym) &&
             (mP->accel_modifiers == modifiers) ) {
            return(mP->word2);
        }
        else if ( mP->type == ITEM_MENU ) {
            retval = keyInMenuBar(mP->word2,ksym,modifiers);
            if ( retval )
                return(retval);
        }
        else if ( mP->type == ITEM_MARK )
            break;
        mP++;
    }
    return(NULL);
}




/*************************************************************************
function: ParseKeySequence - Convert menu bar key sequence to key & mod
This does not yet, but may eventually support multi-key sequences.
In multi-key sequences, it should automatically create sub-modes.
*/
ParseKeySequence (str,mP)
char *str;
menu_item_def *mP;
{
    char *p;
    char ch,keyname[30];
    int namelen = 0;
    int key;
    int got_shift = FALSE, got_ctrl = FALSE, got_alt = FALSE, got_char = FALSE;

    mP->accel_modifiers = 0;
    mP->accel_key = 0;
    if ( !str || !str[0] )
        return(FALSE);

    for ( p = str; *p; p++ ) {
        switch ( *p ) {
        case '!':           /* shift */
            if ( got_shift || got_char )
                return(FALSE);
            got_shift = TRUE;
            mP->accel_modifiers |= KEY_SHIFT_SHIFT;
            break;
        case '^':           /* control */
            if ( got_ctrl || got_char )
                return(FALSE);
            got_ctrl = TRUE;
            mP->accel_modifiers |= KEY_SHIFT_CONTROL;
            break;
        case '*':           /* alt */
            if ( got_alt || got_char )
                return(FALSE);
            got_alt = TRUE;
            mP->accel_modifiers |= KEY_SHIFT_ALT;
            break;
        default:
            if ( namelen >= 20 )
                return(FALSE);
            got_char = TRUE;
            ch = *p;
            keyname[namelen++] = ch;
        }
    }
    if ( !got_char )
        return(FALSE);

    keyname[namelen++] = 0;
    key = KeyStrToKeysym(keyname);
    if ( key == 0 )
        return(FALSE);

/* Don't allow bindings to un-modified typewriter keys */

    if ( !got_ctrl && !got_alt ) {
        if ( (strlen(keyname)==1) || (!got_shift && (key == XK_space)) )
            return(FALSE);
    }
    mP->accel_key = key;
#if 0
   *** took this hack out, see "20000" above */
    if ( key & 0x10000000 )
        mP->accel_key |= 0x20000;
#endif
    return(TRUE);
}






typedef struct {
    char *name;
    int keycode;
} name_and_str;

typedef struct user_key {
    char *name;
    int keycode;
    struct user_key *next;
} user_key;




static user_key *UserKeys;

/*************************************************************************
function: ElfbSetXKey - Set X function key name & code
*/
ElfbSetXKey (name,datum)
char *name;
data_ptr datum;
{
    int val;
    user_key *p;

    val = IntFromDataPtr(datum);
    for ( p = UserKeys; p; p = p->next ) {
        if ( streq(p->name,name) ) {
            p->keycode = val;
            return;
        }
    }
    p = (user_key*)malloc(sizeof(user_key));
    p->next = UserKeys;
    p->name = STRALLOC(name);
    p->keycode = val;
    UserKeys = p;
}


static name_and_str OneCharGroup[] = {
#if AXMS    /* {  use uppercase codes for NT because the lowercase codes *
             * are the same as the codes for function keys. -donato      */
    {"A",XK_A},
    {"B",XK_B},
    {"C",XK_C},
    {"D",XK_D},
    {"E",XK_E},
    {"F",XK_F},
    {"G",XK_G},
    {"H",XK_H},
    {"I",XK_I},
    {"J",XK_J},
    {"K",XK_K},
    {"L",XK_L},
    {"M",XK_M},
    {"N",XK_N},
    {"O",XK_O},
    {"P",XK_P},
    {"Q",XK_Q},
    {"R",XK_R},
    {"S",XK_S},
    {"T",XK_T},
    {"U",XK_U},
    {"V",XK_V},
    {"W",XK_W},
    {"X",XK_X},
    {"Y",XK_Y},
    {"Z",XK_Z},
#else    /* } { */
    {"A",XK_a},
    {"B",XK_b},
    {"C",XK_c},
    {"D",XK_d},
    {"E",XK_e},
    {"F",XK_f},
    {"G",XK_g},
    {"H",XK_h},
    {"I",XK_i},
    {"J",XK_j},
    {"K",XK_k},
    {"L",XK_l},
    {"M",XK_m},
    {"N",XK_n},
    {"O",XK_o},
    {"P",XK_p},
    {"Q",XK_q},
    {"R",XK_r},
    {"S",XK_s},
    {"T",XK_t},
    {"U",XK_u},
    {"V",XK_v},
    {"W",XK_w},
    {"X",XK_x},
    {"Y",XK_y},
    {"Z",XK_z},
#endif    /* AXMS  } */
    {"0",XK_0},
    {"1",XK_1},
    {"2",XK_2},
    {"3",XK_3},
    {"4",XK_4},
    {"5",XK_5},
    {"6",XK_6},
    {"7",XK_7},
    {"8",XK_8},
    {"9",XK_9},
    {"!",XK_exclam},
    {"\"",XK_quotedbl},
    {"#",XK_numbersign},
    {"$",XK_dollar},
    {"%",XK_percent},
    {"&",XK_ampersand},
    {"\"",XK_quoteright},
    {"(",XK_parenleft},
    {")",XK_parenright},
    {"*",XK_asterisk},
    {"+",XK_plus},
    {",",XK_comma},
    {"-",XK_minus},
    {".",XK_period},
    {"/",XK_slash},
    {":",XK_colon},
    {";",XK_semicolon},
    {"<",XK_less},
    {"=",XK_equal},
    {">",XK_greater},
    {"?",XK_question},
    {"@",XK_at},
    {"[",XK_bracketleft},
    {"\\",XK_backslash},
    {"]",XK_bracketright},
    {"^",XK_asciicircum},
    {"_",XK_underscore},
    {"'",XK_quoteleft},
    {"{",XK_braceleft},
    {"|",XK_bar},
    {"}",XK_braceright},
    {"~",XK_asciitilde},
    {NULL,0}
};

/* Define a dummy version of XLT2 which does nothing, but will cause the */
/* key names to be picked up during the scan of the sources for strings */
/* to be translated - jon                                               */
/*  Added NXLT's to the rest of them so they won't be flagged as naked strings.
 *  -- lhunt 4/4/94
 */
#ifdef XLT2
#undef XLT2
#endif
#define XLT2( tag, string ) string

static name_and_str MultiCharGroup[] = {
    { XLT2("key name","BackSpace"), XK_BackSpace},
    { XLT2("key name","Tab"), XK_Tab},
    { XLT2("key name","Linefeed"), XK_Linefeed},
    { XLT2("key name","Clear"), XK_Clear},
    { XLT2("key name","Return"), XK_Return},
    { XLT2("key name","Pause"), XK_Pause},
    { XLT2("key name","Escape"), XK_Escape},
    { XLT2("key name","Delete"), XK_Delete},
    { XLT2("key name","Left"), XK_Left},
    { XLT2("key name","Up"), XK_Up},
    { XLT2("key name","Right"), XK_Right},
    { XLT2("key name","Down"), XK_Down},
    { XLT2("key name","KP_Tab"), XK_KP_Tab},
    { XLT2("key name","KP_Enter"), XK_KP_Enter},
    { XLT2("key name","Prev"), XK_Prior},
    { XLT2("key name","Prior"), XK_Prior},
    { XLT2("key name","Space"), XK_space},
    { NXLT("F1"),XK_F1},
    { NXLT("F2"),XK_F2},
    { NXLT("F3"),XK_F3},
    { NXLT("F4"),XK_F4},
    { NXLT("F5"),XK_F5},
    { NXLT("F6"),XK_F6},
    { NXLT("F7"),XK_F7},
    { NXLT("F8"),XK_F8},
    { NXLT("F9"),XK_F9},
    { NXLT("F10"),XK_F10},
    { NXLT("F11"),XK_F11},
    { NXLT("F12"),XK_F12},
    { NXLT("F13"),XK_F13},
    { NXLT("F14"),XK_F14},
    { NXLT("F15"),XK_F15},
    { NXLT("F16"),XK_F16},
    { NXLT("F17"),XK_F17},
    { NXLT("F18"),XK_F18},
    { NXLT("F19"),XK_F19},
    { NXLT("F20"),XK_F20},
    { NXLT("F21"),XK_F21},
    { NXLT("F22"),XK_F22},
    { NXLT("F23"),XK_F23},
    { NXLT("F24"),XK_F24},
    { NXLT("F25"),XK_F25},
    { NXLT("F26"),XK_F26},
    { NXLT("F27"),XK_F27},
    { NXLT("F28"),XK_F28},
    { NXLT("F29"),XK_F29},
    { NXLT("F30"),XK_F30},
    { NXLT("F31"),XK_F31},
    { NXLT("F32"),XK_F32},
    { NXLT("F33"),XK_F33},
    { NXLT("F34"),XK_F34},
    { NXLT("F35"),XK_F35},
    { XLT2("key name","Home"), XK_Home},
    { XLT2("key name","Next"), XK_Next},
    { XLT2("key name","End"), XK_End},
    { XLT2("key name","Begin"), XK_Begin},
    { XLT2("key name","Select"), XK_Select},
    { XLT2("key name","Print"), XK_Print},
    { XLT2("key name","Execute"), XK_Execute},
    { XLT2("key name","Insert"), XK_Insert},
    { XLT2("key name","Undo"), XK_Undo},
    { XLT2("key name","Redo"), XK_Redo},
    { XLT2("key name","Menu"), XK_Menu},
    { XLT2("key name","Find"), XK_Find},
    { XLT2("key name","Cancel"), XK_Cancel},
    { XLT2("key name","Help"), XK_Help},
    { XLT2("key name","Break"), XK_Break},
    { NXLT("KP_Home"),XK_KP_Home},
    { NXLT("KP_End"),XK_KP_End},
    { NXLT("KP_Space"),XK_KP_Space},
    { NXLT("PF1"),XK_KP_F1},
    { NXLT("PF2"),XK_KP_F2},
    { NXLT("PF3"),XK_KP_F3},
    { NXLT("PF4"),XK_KP_F4},
    { NXLT("KP_Equal"),XK_KP_Equal},
    { NXLT("KP_Multiply"),XK_KP_Multiply},
    { NXLT("KP_Add"),XK_KP_Add},
    { NXLT("KP_Separator"),XK_KP_Separator},
    { NXLT("KP_Subtract"),XK_KP_Subtract},
    { NXLT("KP_Decimal"),XK_KP_Decimal},
    { NXLT("KP_Divide"),XK_KP_Divide},
    { NXLT("KP_0"),XK_KP_0},
    { NXLT("KP_1"),XK_KP_1},
    { NXLT("KP_2"),XK_KP_2},
    { NXLT("KP_3"),XK_KP_3},
    { NXLT("KP_4"),XK_KP_4},
    { NXLT("KP_5"),XK_KP_5},
    { NXLT("KP_6"),XK_KP_6},
    { NXLT("KP_7"),XK_KP_7},
    { NXLT("KP_8"),XK_KP_8},
    { NXLT("KP_9"),XK_KP_9},
    { NXLT("L1"),XK_L1},
    { NXLT("L2"),XK_L2},
    { NXLT("L3"),XK_L3},
    { NXLT("L4"),XK_L4},
    { NXLT("L5"),XK_L5},
    { NXLT("L6"),XK_L6},
    { NXLT("L7"),XK_L7},
    { NXLT("L8"),XK_L8},
    { NXLT("L9"),XK_L9},
    { NXLT("L10"),XK_L10},
    { NXLT("R1"),XK_R1},
    { NXLT("R2"),XK_R2},
    { NXLT("R3"),XK_R3},
    { NXLT("R4"),XK_R4},
    { NXLT("R5"),XK_R5},
    { NXLT("R6"),XK_R6},
    { NXLT("R7"),XK_R7},
    { NXLT("R8"),XK_R8},
    { NXLT("R9"),XK_R9},
    { NXLT("R10"),XK_R10},
    { NXLT("R11"),XK_R11},
    { NXLT("R12"),XK_R12},
    { NXLT("R13"),XK_R13},
    { NXLT("R14"),XK_R14},
    { NXLT("R15"),XK_R15},
#if PMAX || DEC_ALPHA
    {NXLT("Remove"),DXK_Remove},  /* DEC key */
#endif
#ifdef XK_S1    /* solbourne keys */
    { NXLT("S1"),XK_S1},
    { NXLT("F36"),XK_F36},
    { NXLT("S2"),XK_S2},
    { NXLT("F37"),XK_F37},
    { NXLT("S3"),XK_S3},
    { NXLT("F38"),XK_F38},
    { NXLT("S4"),XK_S4},
    { NXLT("F39"),XK_F39},
    { NXLT("S5"),XK_S5},
    { NXLT("F40"),XK_F40},
    { NXLT("S6"),XK_S6},
    { NXLT("F41"),XK_F41},
    { NXLT("S7"),XK_S7},
    { NXLT("F42"),XK_F42},
    { NXLT("S8"),XK_S8},
    { NXLT("F43"),XK_F43},
    { NXLT("S9"),XK_S9},
    { NXLT("F44"),XK_F44},
    { NXLT("S10"),XK_S10},
    { NXLT("F45"),XK_F45},
    { NXLT("F46"),XK_F46},
    { NXLT("F47"),XK_F47},
#endif
#if (HP300 || HP800) && !WE2
#if HP800V3
    {NXLT("InsertChar"), hpXK_InsertChar},
    {NXLT("DeleteChar"), hpXK_DeleteChar},
    {NXLT("ClearLine"), hpXK_ClearLine},
    {NXLT("InsertLine"), hpXK_InsertLine},
    {NXLT("DeleteLine"), hpXK_DeleteLine},
    {NXLT("Reset"), hpXK_Reset},            /* shift break */
    {NXLT("System"), hpXK_System},
    {NXLT("User"), hpXK_User},
#else
    {NXLT("InsertChar"), XK_InsertChar},
    {NXLT("DeleteChar"), XK_DeleteChar},
    {NXLT("ClearLine"), XK_ClearLine},
    {NXLT("InsertLine"), XK_InsertLine},
    {NXLT("DeleteLine"), XK_DeleteLine},
    {NXLT("Reset"), XK_Reset},            /* shift break */
    {NXLT("System"), XK_System},
    {NXLT("User"), XK_User},
#endif
#endif
    {NULL,0}
};

#undef XLT2    /* go back to the real thing */



/*************************************************************************
function: KeyStrToKeysym - Convert string to name
*/
KeyStrToKeysym (str)
char *str;
{
    name_and_str *p;
    user_key *up;

    if ( !keys_translations_inited )
        {
        SetupKeyTranslations();
        }


    if ( !str[0] )
        return(0);
    if ( strlen(str) == 1 ) {
        for ( p = OneCharGroup; p->name; p++ )
            if ( *str == p->name[0] )
                return(p->keycode);
    }
    else {
        for ( up = UserKeys; up; up = up->next )
            if ( !Astrcmp(str,up->name) )
                return(up->keycode);
        for ( p = MultiCharGroup; p->name; p++ ) {
            if ( !Astrcmp(str,p->name) )
                return(p->keycode);
        }
    }
    return(0);
}

/*************************************************************************
function: ElfMakeAccelStr - Return a string from accels data
*/
ElfMakeAccelStr (mP,buf,bufsiz)
register menu_item_def *mP;
char *buf;
{
    register char *p = buf;
    register name_and_str *pp;
    register user_key *up;
    register int keycode = mP->accel_key;

    if ( !keycode ) {
        *p = 0;
        return;
    }

    if ( !keys_translations_inited )
        {
        SetupKeyTranslations();
        }

#if 0
    *** took hack out, see "20000" above
    if ( keycode & 0x20000 )
        keycode = 0x10000000 + (keycode & ~0x20000);
#endif
    if ( mP->accel_modifiers & KEY_SHIFT_SHIFT )
        *p++ = '!';
    if ( mP->accel_modifiers & KEY_SHIFT_CONTROL )
        *p++ = '^';
    if ( mP->accel_modifiers & KEY_SHIFT_ALT )
        *p++ = '*';
    for ( pp = OneCharGroup; pp->name; pp++ ) {
        if ( keycode == pp->keycode ) {
            strcpy(p,pp->name);
            return;
        }
    }
    for ( up = UserKeys; up; up = up->next ) {
        if ( keycode == up->keycode ) {
            strcpy(p,up->name);
            return;
        }
    }

    for ( pp = MultiCharGroup; pp->name; pp++ ) {
        if ( keycode == pp->keycode ) {
            strcpy(p,pp->name);
            return;
        }
    }
}

/*************************************************************************
function: ElfbGetXKeys - Return list of current X keybindings
*/
ElfbGetXKeys ()
{
    extern data_ptr ElfRetData;
    data_ptr toparray;
    data_ptr n2array;
    int cnt = 0;
    name_and_str *p;
    user_key *up;

    if ( !keys_translations_inited )
        {
        SetupKeyTranslations();
        }


    for ( p = OneCharGroup; p->name; p++ ) cnt++;
    for ( up = UserKeys; up; up = up->next ) cnt++;
    for ( p = MultiCharGroup; p->name; p++ ) cnt++;

    toparray = ElfRetData = MkArray(cnt);
    cnt = 0;
    for ( p = OneCharGroup; p->name; p++ ) {
        n2array = MkArray(2);
        WriteArray(toparray,cnt++,n2array);
        ElfAddStrToArray(n2array,0,p->name);
        ElfAddDblToArray(n2array,1,(double)(p->keycode));
    }
    for ( up = UserKeys; up; up = up->next ) {
        n2array = MkArray(2);
        WriteArray(toparray,cnt++,n2array);
        ElfAddStrToArray(n2array,0,up->name);
        ElfAddDblToArray(n2array,1,(double)(up->keycode));
    }
    for ( p = MultiCharGroup; p->name; p++ ) {
        n2array = MkArray(2);
        WriteArray(toparray,cnt++,n2array);
        ElfAddStrToArray(n2array,0,p->name);
        ElfAddDblToArray(n2array,1,(double)(p->keycode));
    }   
}




SetupKeyTranslations()
{
    name_and_str *p;

     FILE *ff;

    if ( !keys_translations_inited )
        {
        /** since the key names appear in the menu bars, and are **/
        /** visible to the user, the first time through we go through **/
        /** the hard coded english values, and substitute the translated **/
        /** strings so that the user can enter the translated value      **/
        /** on the menu bar                                              **/
        /** Note: this relies on the fact that "XLT2()" returns either a  **/
        /**       pointer to an otherwise static string which will not  **/
        /**       move or disappear, or it returns its input if there   **/
        /**       is no translated value.  ( for such things as "F10")   **/
        for ( p = MultiCharGroup; p->name; p++ )
            {
            p->name = XLT2( "key name", p->name );
            }
        keys_translations_inited = 1;
        }
}




#if 0
ResetShiftState()
{
;
}

#endif
