/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmlayer.c
   Applix ASC version: 10
   Date: Mon Nov 16 13:53:34 1998
*/
#include <stdio.h>
#include <aos.h>
#include <paripc.h>
#include <tmi.h>
#include <tmassert.h>

#include <atm.h>
#include <akeys.h>
#include <atmwin.h>
#include <atmreg.h>
#include <atmattr.h>
#include <atmtxtin.h>

#include <paratmgr.h>
#include <parwin.h>
#include <parint.h>

char *ATMmalloc();

extern short ATMtrace;
extern FILE *ATfp;

static Obscured *Obs;
#if AX_X11
extern Layer *FrontLayer, *BackLayer;
#endif

/*****************************************************************/

/*
 * Test if two rectangles overlap.
 */
rectXrect(r, s)
	AxRectangle *r, *s;
{
	return(	r->origin.x < s->corner.x &&
		s->origin.x < r->corner.x &&
		r->origin.y < s->corner.y &&
		s->origin.y < r->corner.y  );
}

/*****************************************************************/

/*
 * ATMlayerop --  apply function `fn' within rectangle `rp'
 *					of Layer `lp'
 *
 *	(Subdivide `rp' into rectangles that do not cross
 *	bitmap boundaries and apply `fn' to each rectangle.)
 *
 *	Clip `rp' to layer, then call Rlayerop().
 *
 *	Note: `lp' can be NULL, in which case
 *	bitmap operator is applied directly to Display.
 */

ATMlayerop(lp, fn, rp, otherargs)
	Layer *lp;
	int (*fn)();		/* "bitmap operator" */
	AxRectangle *rp;		/* area to which operator applied */
	char *otherargs;	/* pointer to packet of
				   			other args used by (*fn)() */
{
	AxRectangle rr, *clipRp;
	Obscured *op;

	if (lp == NULL)							/* special case */
	{
		clipRp = &A_Display->maprect;
		op = NULL;
	}
	else									/* normal case */
	{
		clipRp = &lp->curClip;
		op = lp->obs;
	}
	if ( sectRect(rp, clipRp, &rr) ) 			/* clip */
		Rlayerop(lp, fn, &rr, otherargs, op);
}

ATMnorectLayerop(lp, fn, polyPoints, nPts, otherargs)
Layer *lp;
int (*fn)();           /* "bitmap operator" */
TmPoint32 *polyPoints; /* area to which operator applied */
int nPts;
char *otherargs;       /* packet ptr of other args used by (*fn)() */
{
	AxRectangle rr, *clipRp;
	Obscured *op;
	int nClippedPts;
	TmPoint32 clippedPoints[10];

	if (lp == NULL)							/* special case */
	{
		clipRp = &A_Display->maprect;
		op = NULL;
	}
	else									/* normal case */
	{
		clipRp = &lp->curClip;
		op = lp->obs;
	}
	(*fn)(lp, polyPoints, nPts, A_Display, otherargs, clipRp);
}

/*****************************************************************/

/*
 * Rlayerop -- recursively subdivide and intersect
 *		rectangle with list of obscured-piece bitmaps
 *
 *		If list is empty, `fn' is called immediately for the Bitmap
 *		Display, with the other parameters passed through.
 */
Rlayerop(lp, fn, rp, otherargs, op)
	Layer *lp;
	int (*fn)();
	AxRectangle *rp;
	char *otherargs;
	Obscured *op;	/* element of obscured list with which
						to intersect rp */
{
AxRectangle temp, rr;

if (op == NULL)								/* rectangle not obscured */
	(*fn)(lp, rp, A_Display, otherargs, op);			/* draw on screen */
else if (rectXrect(rp, &op->bmap->maprect) == 0)			/* they miss */
	Rlayerop(lp, fn, rp, otherargs, op->next);				/* chain */
else 											/* they must intersect */
{
	rr = *rp;									/* copy input rectangle */
	if ( rr.origin.x < op->bmap->maprect.origin.x )
	{							/* rr extends left of obscured rectangle */
		temp = rr;
		temp.corner.x = op->bmap->maprect.origin.x;
					/* temp = left-extending piece */
		Rlayerop(lp, fn, &temp, otherargs, op->next);
					/* remove piece from rr */
		rr.origin.x = op->bmap->maprect.origin.x;
	}
	if ( rr.corner.y > op->bmap->maprect.corner.y )
	{							/* rr extends below obscured rectangle */
		temp = rr;
		temp.origin.y = op->bmap->maprect.corner.y;
				/* temp = below-extending piece */
		Rlayerop(lp, fn, &temp, otherargs, op->next);
					/* remove piece from rr */
		rr.corner.y = op->bmap->maprect.corner.y;
	}
	if ( rr.corner.x > op->bmap->maprect.corner.x )
	{		 				/* rr extends right of obscured rectangle */
		temp = rr;
		temp.origin.x = op->bmap->maprect.corner.x;
				/* temp = right-extending piece */
		Rlayerop(lp, fn, &temp, otherargs, op->next);
					/* remove piece from rr */
		rr.corner.x = op->bmap->maprect.corner.x;
	}
	if ( rr.origin.y < op->bmap->maprect.origin.y )
	{							/* rr extends above obscured rectangle */
		temp = rr;
		temp.corner.y = op->bmap->maprect.origin.y;
				/* temp = above-extending piece */
		Rlayerop(lp, fn, &temp, otherargs, op->next);
					/* remove piece from rr */
		rr.origin.y = op->bmap->maprect.origin.y;
	}
						/* "What's left goes in this obscured bitmap" */

	(*fn)(lp, &rr, op->bmap, otherargs, op);
}

} /* end Rlayerop */

/*
 * setclip -- set current clipping rectangle of Layer
 */
setclip(lp, rp)
	Layer *lp;
	AxRectangle *rp;
{
	if (lp == NULL
			|| rp->origin.x == lp->curClip.origin.x
				&& rp->origin.y == lp->curClip.origin.y
				&& rp->corner.x == lp->curClip.corner.x
				&& rp->corner.y == lp->curClip.corner.y )
	{
		return;
	}
	lp->curClip = *rp;
	delClipList(lp, TRUE);	/* invalidate Layer's cliplist */
}
/*************************************************************/
/*
 * delClipList -- delete clip list for a Layer
 *
 *		must be called whenever clipping state of Layer changes
 */
delClipList(lp, doWin)
	Layer *lp;
	int doWin;			/* delete window manager resource */
{
	ClipList *curEnt, *nxEnt;

	curEnt = lp->clp;
	while (curEnt != NULL)
	{
		nxEnt = curEnt->link;
		ATMfree(curEnt);
		curEnt = nxEnt;
	}
	lp->clp = NULL;
}


#if AX_X11    /* { */
/*************************************************************/
delAllCL(doWin)
	int doWin;		/* whether or not to delete window manager resource */
{
	Layer *lp;

	for (lp = FrontLayer; lp != NULL; lp = lp->back) delClipList(lp, FALSE);
}
/*************************************************************/
/*
 * getCList -- get clip list for a Layer
 *
 *		construct list if necessary and save for reuse
 */
ClipList *
getCList(lp)
	Layer *lp;			/* can be NULL */
{
	ClipList *getClip();

	if (lp == NULL)
	{
		return getClip(NULL);
	}
	else
	{
		if (lp->clp == NULL)
			return ( lp->clp = getClip(lp) );
		else
			return lp->clp;
	}
}
#endif    /* } AX_X11 */
