/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmout2.c
   Applix ASC version: 8
   Date: Mon Jan 11 16:28:32 1999
*/
#include <stdio.h>
#include <fcntl.h>
#include <errno.h>

#if RIOS /* math.h won't include stdlib.h (redefs malloc and free) */
#	define _BSD
#endif

#if AXMS
#define hypot _hypot
#endif

#include <math.h>
#include <aos.h>
#include <tmi.h>
#include <paripc.h>
#include <paratm.h>
#include <paratmgr.h>
#include <parout.h>
#include <parint.h>
#include <bounds.h>
#include <scac.h>
#include <getstr13.h>
#include <laytext.h>
#include <devsetup.h>

/*****************************************************************
 *
 * atm_blend_func - determine affect of a point on the current point
 */
double
atm_blend_func(currIx, finalIx, parameter)
int currIx, finalIx;
double parameter;
{
	double positionAffect;
	int i, k, g;

	k = g = 1;

	for (i=finalIx; i>currIx; i--)
		k *= i;
	for (i=finalIx-currIx; i>1; i--)
		g *= i;

	positionAffect = (double)k / (double)g;

	for (i=1; i<=currIx; i++)
		positionAffect *= parameter;
	for (i=1; i<=finalIx-currIx; i++)
		positionAffect *= (1-parameter);

	return ( positionAffect );
}

/*****************************************************************
 *
 * AxBezierCurve - TmPolyline/bezier curve utility
 * TmPolyline utility function for converting bezier control
 * points into a simple poly. Given n points, this function
 * constructs a new polyline consisting of n segments. The curve
 * interpolates the end points passed in. The caller is reponsible
 * for providing space for the return points.
 */
AxBezierCurve(nSegments, nPts, pts, returnPts)
int nSegments;
int nPts;
TmPoint32 *pts, *returnPts;
{
	double atm_blend_func(), affectOnCurrentPoint;
	double parameter;
	double x, y;
	int i,j;

	for (i=0; i<=nSegments; i++)
	{
		parameter = (double)i / (double)nSegments;
		for (x=y=j=0; j<nPts; j++)
		{
			affectOnCurrentPoint = atm_blend_func(j, nPts-1, parameter);
			x += pts[j].x * affectOnCurrentPoint;
			y += pts[j].y * affectOnCurrentPoint;
		}
		returnPts[i].x = x;
		returnPts[i].y = y;
	}
}

/*****************************************************************
 *
 * AxFindPointOnCurve - TmPolyline/bezier curve utility
 */
AxFindPointOnCurve(pid, param_num, param_den, n_pts, pts, return_pt, cosAng, sinAng)
int pid;
int param_num, param_den, n_pts;
TmPoint32 *pts, *return_pt;
double *cosAng, *sinAng;
{
	double parameter, scaler, p_num, p_den, x, y, dx, dy, d;
	double atm_blend_func(), affectOnCurrentPoint;
	double p1_x, p2_x, *tan_pt_x;
	double p1_y, p2_y, *tan_pt_y;
	int j;

	/* get the initial parameter */
	if ( param_num > param_den )
		param_num = param_den;
	parameter = (double)param_num / (double)param_den;
	if ( parameter < 0.0 )
		parameter = 0.0;

	/* find the point */
	for (x=y=j=0; j<n_pts; j++)
	{
		affectOnCurrentPoint = atm_blend_func(j, n_pts-1, parameter);
		x += pts[j].x * affectOnCurrentPoint;
		y += pts[j].y * affectOnCurrentPoint;
	}
	return_pt->x = x;
	return_pt->y = y;

	/* increase the precision of the passed in points */
#if FOR_FILTER
	tan_pt_x = (double *)malloc(n_pts * sizeof(double));
	tan_pt_y = (double *)malloc(n_pts * sizeof(double));
#else
	tan_pt_x = (double *)TaskAlloc(pid, n_pts * sizeof(double));
	tan_pt_y = (double *)TaskAlloc(pid, n_pts * sizeof(double));
#endif
	scaler = 1000000.0;
	for ( j=0; j<n_pts; j++ )
	{
		tan_pt_x[j] = (double)pts[j].x * scaler;
		tan_pt_y[j] = (double)pts[j].y * scaler;
	}
	scaler = 100000000.0;
	p_num = (double)param_num * scaler;
	p_den = (double)param_den * scaler;

	/* find the point in the enhanced coords */
	parameter = p_num / p_den;
	if ( parameter < 0.0 )
		parameter = 0.0;
	for (p2_x=p2_y=j=0; j<n_pts; j++)
	{
		affectOnCurrentPoint = atm_blend_func(j, n_pts-1, parameter);
		p2_x += tan_pt_x[j] * affectOnCurrentPoint;
		p2_y += tan_pt_y[j] * affectOnCurrentPoint;
	}

	/* find a point just prior to the point (sort of a limit) */
	parameter -= 1.0 / p_den;
	if ( parameter < 0.0 )
		parameter = 0.0;
	for (p1_x=p1_y=j=0; j<n_pts; j++)
	{
		affectOnCurrentPoint = atm_blend_func(j, n_pts-1, parameter);
		p1_x += tan_pt_x[j] * affectOnCurrentPoint;
		p1_y += tan_pt_y[j] * affectOnCurrentPoint;
	}

	/* now use the deltas to determine the sine and cosine */
	*cosAng = 0.0;
	*sinAng = 0.0;
	dx = p2_x - p1_x;
	dy = p2_y - p1_y;
	d = (dx == 0 && dy == 0) ? 0 : hypot(dx, dy);
	if ( d )
	{
		*cosAng = dx / d;
		*sinAng = dy / d;
	}
	else /* at pt1, so get from pt1-ctl1 chord */
	{
		dx = tan_pt_x[1] - tan_pt_x[0];
		dy = tan_pt_y[1] - tan_pt_y[0];
		d = (dx == 0 && dy == 0) ? 0 : hypot(dx, dy);
		if ( d )
		{
			*cosAng = dx / d;
			*sinAng = dy / d;
		}
	}

	/* clean up */
	TaskFree(pid, tan_pt_x);
	TaskFree(pid, tan_pt_y);
}

/*****************************************************************
 *
 * TmPolyline utility for converting complex polygons into simple
 * polygons. Also converts from mills to screen pels.
 */
convertPols(nPols, pol, nPoints, points, to_pix)
int nPols;
poly_type *pol;
int *nPoints;
TmPoint32 **points;  /* caller must free the elements of this array */
int to_pix;
{
	TmPoint32 bezierPts[4];
	int i, j, maxPts, byte_size, count;

	/* convert each mil based bezier/poly to a simple screen pel based poly */
	for ( i=0; i<nPols; i++ ) if ( pol[i].nPoints >= 1 )
	{
		/* if it's a list of beziers, convert to a polyline for display */
		if ( pol[i].nCtrls >= 1 )
		{
			maxPts = pol[i].nPoints / 2;
			byte_size = (maxPts*26 + 1) * sizeof(TmPoint32);
			points[i] = (TmPoint32 *)malloc(byte_size);
			for(count=0, j=0; j<maxPts; j++)
			{
				if ( pol[i].points[j*2+0].x == pol[i].controls[j*2+0].x &&
				     pol[i].points[j*2+0].y == pol[i].controls[j*2+0].y &&
				     pol[i].points[j*2+1].y == pol[i].controls[j*2+1].y &&
				     pol[i].points[j*2+1].y == pol[i].controls[j*2+1].y )
				{
					points[i][count+0] = pol[i].points[j*2+0];
					points[i][count+1] = pol[i].points[j*2+1];
					count += 2;
				}
				else
				{
					/* endpoints */
					bezierPts[0].x = pol[i].points[j*2+0].x;
					bezierPts[0].y = pol[i].points[j*2+0].y;
					bezierPts[3].x = pol[i].points[j*2+1].x;
					bezierPts[3].y = pol[i].points[j*2+1].y;

					/* first control point */
					bezierPts[1].x = pol[i].controls[j*2+0].x;
					bezierPts[1].y = pol[i].controls[j*2+0].y;

					/* second control point */
					bezierPts[2].x = pol[i].controls[j*2+1].x;
					bezierPts[2].y = pol[i].controls[j*2+1].y;

					AxBezierCurve(25, 4,bezierPts, &points[i][count]);
					count += 26;
				}
			}
			nPoints[i] = count;
		}
		else /* just point at the passed in data */
		{
			nPoints[i] = pol[i].nPoints;
			byte_size = (nPoints[i]+1) * sizeof(TmPoint32);
			points[i] = (TmPoint32 *)malloc(byte_size);
			for(j=0; j<nPoints[i]; j++)
				points[i][j] = pol[i].points[j];
		}

#if !FOR_FILTER
		/* transform the poly from mills to screen pels */
		if ( to_pix ) for ( j=0; j<nPoints[i]; j++ )
			conv_i_p32(&points[i][j].x, &points[i][j].y,
					   &points[i][j].x, &points[i][j].y, MILLS_TO_PIXEL);
#endif

	}
	else
	{
		nPoints[i] = 0;
		points[i] = (TmPoint32 *)0;
	}
}

    /* free the poly info allocated by convertPols */
void
axFreePolylinePoints(points, numPols)
TmPoint32 **points;
int numPols;
{
	int ix;
	if (points) {
		for (ix = 0; ix < numPols; ix++) {
			if (points[ix]) {
				free((char *) points[ix]);
			}
		}
	}
}
