/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmsup.c
   Applix ASC version: 26
   Date: Tue Apr  6 21:51:30 1999
*/
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <aos.h>
#include <dia.h>
#include <userp.h>
#include <tmi.h>
#include <paratm.h>
#include <parint.h>
#include <paratmgr.h>
#include <paripc.h>
#include <scac.h>
#include <tmassert.h>
#include <getstr13.h>
#include <mincom.h>
#include <pargmod.h>
#include <devsetup.h>
#include <amemtrak.h>

#if CLIX
#define NSIG _NSIG
#endif

#define USE_MOUSE   IBM_PHM == 0
#define NULCHAR     '\0'

short TermReleased = TRUE;      /* initialize to deal with early aborts */
short OldBorders = FALSE;
extern struct wind_struct *StatWp;
extern short CellWidDC, CellHtDC, CellUnder;
extern short UWidPels, UHtPels, UWidMils, UHtMils;
short Xmpc = X_MPC; /* mills pers syschar cell */
short Ympc = Y_MPC; /* mills pers syschar cell */
extern short RastWid, RastHt;
extern short ToggleFG;

char *DpyBase, *OffBase;
STATIC ShStruct *Sh;
#define DpyBase DPY_BASE
#define OffBase OFF_BASE

extern int MapMode;


        /* general */

short cur_win = TM_NULWIN;  /* currently active window ID */
short cur_flag;     /* cursor on/off indicator */

short ACwin = TM_NULWIN;    /* ID of window created by APAC */
short IntWin = TM_NULWIN;   /* ID of window belonging to
                                handler of INTERRUPT */
        /* global strings */
char IntString[64];     /* "interrupt" option */
short IntLen;           /* IntString length */
char HelpTx0[64];       /* "HELP" icon */
char HelpTx1[128];      /* help picked message */

        /* array of window control blocks */

struct wind_struct window[MAX_WIND_ID], *wind_pnt;

        /* request dispatching globals */

short mes_len;

        /* mouse variables */

short mous_x;
short mous_y;

int TermFD; /* file descriptor for terminal output */

FILE *fd2;      /* log file pointer */
FILE *ATfp;     /* OBS */
extern FILE *ATfp;      /* ATM trace file pointer */
short ATMdebug = 0, ATMtrace = 0;
char *ATMcurMes;    /* -> current request message (for core dumps) */

char *AOSpathname();

int MINpid;         /* pid of min process - needed in termination */

/* global to this module only */

/********************************************************************
 *
 * Initialize Main ATM Process
 */
void init_atm()
{
    char *getenv();

    /* low level inits */
    WMinit();           /* get system resources and characteristics */

    /* higher level inits */
    TermReleased = FALSE;
    bmapinit();     /* initialize bitmap management and Display Bitmap */
    fontinit();     /* initialize font related database */
    cur_flag = 1;   /* initialize flag to indicate that cursor is visible */

    /* make no window the active window */
    make_curwin(TM_NULWIN);
}

        /* centralized malloc, realloc, free calls */


static int Nmalloc, Nfree;  /* malloc/free use counters assume initally 0 */
int Balmalloc;          /* malloc balance: current total allocated */

char *ATMmalloc(size)
unsigned size;
{
    RETADDR(size);
    return((char *)malloc(size));
}

void ATMfree(addr)
char *addr;
{
    free(addr);
}

char *ATMrealloc(pointer, size)
int *pointer;
unsigned size;
{
    return((char *)realloc((char *)pointer, size));
}

/*******************************************************************/
/*
 * ATMabort -- abort on fatal error reporting one string file message
 */
void ATMabort(errNo)
int errNo;

{
    char abortMes[256];

    getStr(errNo, abortMes);
    tempAbort(abortMes);
}

/*******************************************************************/
/*
 * ATMab1 -- abort on fatal error reporting two string file messages
 */
void ATMab1(errNo1, errNo2)
int errNo1;
int errNo2;

{
    char abortMes[512];

    sprintf(abortMes,"%s %s",AxGetStr(errNo1),AxGetStr(errNo2));
    tempAbort(abortMes);
}
/*******************************************************************/
/*
 * ATMab2 -- abort reporting two messages and two decimal int values
 */
void ATMab2(errNo1,errNo2, v1, v2)
    int errNo1;
    int errNo2;
    int v1, v2;
{
    char abortMes[512];

    sprintf(abortMes,"%s %s\nvalue 1: %d, value 2: %d", 
                AxGetStr(errNo1),AxGetStr(errNo2),v1, v2 );
    tempAbort(abortMes);
}
/*******************************************************************/
/*
 * ATMab3 -- abort reporting string followed by string file message
 */
void ATMab3(abort1Mes, errNo2)
char *abort1Mes;
int errNo2;

{
    char abortMes[512];
    char *mesP;

    strcpy(abortMes, abort1Mes);
    if (errNo2 != -1)
    {
        mesP = abortMes + strlen(abortMes);
        *mesP++ = ':';
        *mesP++ = ' ';
        getStr(errNo2, mesP);
    }
    tempAbort(abortMes);
}
/*******************************************************************/
/*
 * ATMerr -- abort with two messages, decimal int value
 */
void ATMerr(errNo1,errNo2, n)
    int errNo1;
    int errNo2;
    int n;
{
    char abortMes[512];

    sprintf(abortMes,"%s %s\nvalue : %d", 
                AxGetStr(errNo1),AxGetStr(errNo2),n );
    tempAbort(abortMes);
}

/***************************************************************/
/*
 * widthDC -- width to device coordinates -- convert a width in
 *      world coordinates (mils) to device coordinates
 *      (cells, pixels)
 *
 *          obsolete routine to be transitioned out
 */
short
widthDC(wcWidth)
    short wcWidth;
{
    return wcWidth / Xmpc;
}

/*****************************************************************/
/*
 * coordinate transformations
 */
/*****************************************************************/

void conv_i_p(i1, i2, p1, p2, direction)
    short *i1,*i2;
    short *p1,*p2;
    short direction;
{
    if (direction == MILLS_TO_PIXEL)
    {
        *p1 = ( (long) *i1 * UWidPels ) / UWidMils;
        *p2 = ( (long) *i2 * UHtPels ) / UHtMils;
    }
    else
    {
        int round_up;
        /*
         * Value is being converted to mils for sending to
         * client application.
         * Make sure mil value, if later returned to ATM, will
         * convert back to the original pixel value.
         */
        round_up = UWidPels - 1;
        if (*p1 < 0) round_up = (-round_up);
        *i1 = ( (long) (*p1) * UWidMils + round_up ) / UWidPels;
        round_up = UHtPels - 1;
        if (*p2 < 0) round_up = (-round_up);
        *i2 = ( (long) (*p2) * UHtMils + round_up ) / UHtPels;
    }
}

void conv_i_p32(i1, i2, p1, p2, direction)
    int *i1,*i2;
    int *p1,*p2;
    int direction;
{
    if (direction == MILLS_TO_PIXEL)
    {
        *p1 = ( (long) *i1 * UWidPels ) / UWidMils;
        *p2 = ( (long) *i2 * UHtPels ) / UHtMils;
    }
    else
    {
        int round_up;
        /*
         * Value is being converted to mils for sending to
         * client application.
         * Make sure mil value, if later returned to ATM, will
         * convert back to the original pixel value.
         */
        round_up = UWidPels - 1;
        if (*p1 < 0) round_up = (-round_up);
        *i1 = ( (long) (*p1) * UWidMils + round_up ) / UWidPels;
        round_up = UHtPels - 1;
        if (*p2 < 0) round_up = (-round_up);
        *i2 = ( (long) (*p2) * UHtMils + round_up ) / UHtPels;
    }
}

void convMilsToPix(inx, iny, outx, outy)
int inx,iny;      /* in */
int *outx,*outy;    /* out */
{
    conv_i_p32(&inx, &iny, outx, outy, MILLS_TO_PIXEL);
}

void convPixToMils(inx, iny, outx, outy)
int inx, iny;    /* in */
int *outx, *outy;  /* out */
{
    conv_i_p32(outx, outy, &inx, &iny, PIXEL_TO_MILLS);
}

/* : axGridForScroll - adjust a mil displacement
 * into the next higher value which is a clean
 * multiple of device pixels.
 * This function uses the fact that UMils is
 * UMils is defined as equal in size to some perfect
 * multiple of UPels, for instance 10:3 for a 300dpi device)
 */
void
axGridForScroll(milx, mily, gmilx, gmily)
int milx, mily;      /* input displacement in mils */
int *gmilx, *gmily;  /* output displacement in mils */
{
    int tx, ty, tmpx, tmpy;

    *gmilx = milx - (milx % UWidMils);
    if (*gmilx < milx)
        *gmilx += UWidMils;  /* go to next higher multiple */

    *gmily = mily - (mily % UHtMils);
    if (*gmily < mily)
        *gmily += UHtMils;
}

/*****************************************************************/

/*****************************************************************/
/*
 * mils_dc -- convert from mils to device coordinates.
 *      In MapMode, return pixel coordinates;
 *      otherwise, return cell coordinates.
 *
 *  Values can be converted in place.
 */
void mils_dc(xMp, yMp, xDCp, yDCp)
    short *xMp, *yMp, *xDCp, *yDCp;
{
    *xDCp = (long) *xMp * UWidPels / UWidMils;
    *yDCp = (long) *yMp * UHtPels / UHtMils;
}

void mils_dc32(xMp, yMp, xDCp, yDCp)
    int *xMp, *yMp, *xDCp, *yDCp;
{
    *xDCp = (long) *xMp * UWidPels / UWidMils;
    *yDCp = (long) *yMp * UHtPels / UHtMils;
}
/*****************************************************************/
/*
 * milPix -- convert a size (or displacement) from mils to pixels
 *
 *      Convert (dx, dy) components in place.
 *      A "pixel" is the smallest addressable screen unit.
 *
 *      replaces mils_dc()
 */
void milPix(dxp, dyp)
    short *dxp, *dyp;
{
    mils_dc(dxp, dyp, dxp, dyp);
}

void milPix32(dxp, dyp)
    int *dxp, *dyp;
{
    mils_dc32(dxp, dyp, dxp, dyp);
}
/*****************************************************************/
/*
 * WmilSpix -- convert from window-relative mils to screen pixels
 *
 *      Converts an (x, y) position in application coordinates
 *      to device coordinates.
 */
void WmilSpix(wp, xp, yp)
    struct wind_struct *wp;
    short *xp, *yp;
{
    milPix(xp, yp);
    *xp += wp->inner.origin.x;
    *yp += wp->inner.origin.y;
}

void WmilSpix32(wp, xp, yp)
    struct wind_struct *wp;
    int *xp, *yp;
{
    milPix32(xp, yp);
    *xp += (int)wp->inner.origin.x;
    *yp += (int)wp->inner.origin.y;
}

/*****************************************************************/
/*
 * ccRect -- convert a call interface rectangle to internal form
 *
 *      - bypass the conversions for arguments already in screen pixels
 *      - uses inner clipping rectangle of window (wp)
 */
AxRectangle *
ccRect(wp, inPixels, ox, oy, xsiz, ysiz, rp)
    struct wind_struct *wp;
    int inPixels;               /* screen pixels vs. window mils */
    int ox, oy, xsiz, ysiz;
    AxRectangle *rp;
{
    short orgx, orgy, corsizx, corsizy;
    int posPreference;

    /* temp method for trial */
    posPreference = wp->ap_id == APGE || wp->ap_id == APBG
                    || wp->ap_id == APPIXED;
    orgx = ox;
    orgy = oy;
    corsizx = xsiz;
    corsizy = ysiz;
    if (posPreference) {
        /* recover application's positions before converting */
        corsizx += orgx;                
        corsizy += orgy;                
    }
    if (!inPixels) {
        milPix(&orgx, &orgy);
        milPix(&corsizx, &corsizy);     /* corners or sizes */
    }
    if (!posPreference) {       /* in pixels, now compute corners */
        corsizx += orgx;
        corsizy += orgy;
    }
    rp->origin.x = orgx;
    rp->origin.y = orgy;
    rp->corner.x = corsizx;
    rp->corner.y = corsizy;
    if (!inPixels)
    {
        rp->origin.x += wp->inner.origin.x;
        rp->origin.y += wp->inner.origin.y;
        rp->corner.x += wp->inner.origin.x;
        rp->corner.y += wp->inner.origin.y;
    }
    return rp;
}
/*****************************************************************/
/*
 * cTmRect -- convert a TmRect to a AxRectangle
 *
 *      should eventually replace ccRect()
 */
AxRectangle *
cTmRect(wp, tmRp, rp)
    struct wind_struct *wp;
    TmRect *tmRp;
    AxRectangle *rp;
{
    ccRect(wp, FALSE, tmRp->x_position, tmRp->y_position,
               tmRp->x_size, tmRp->y_size, rp);
}

/* This function moves IPC message to a local buffer */
void ATMmesMov(mes, buf, mesLen, bufLen, flag)
char *mes;
char *buf;
short mesLen;
short bufLen;
short flag;
{
    /*
     * flag TRUE(or 2): (mes, mesLen) is source; FALSE: use IPCevMes
     *
     * (mes, mesLen) does not include IPC header
     */
    switch(flag)
    {
        case(TRUE):
        case(FALSE):
            tempAbort("ATMmesMov");
            break;
        case 2: {
            extern short THIMpid;
            memcpy(buf, mes, mesLen);
            ATMcurMes = buf;        /* for debugging */
            return;
            }
    }
    ATMcurMes = buf + IPC_HEAD_LEN;     /* for debugging */
}

/***************************************************************/
/* calculations with rectangles */
/***************************************************************/
/*
 * relRect -- convert absolute AxRectangle to window-relative
 */
AxRectangle *
relRect(absRp, relRp)
    AxRectangle *absRp, *relRp;
{
    relRp->origin.x = absRp->origin.x - Wp->inner.origin.x;
    relRp->corner.x = absRp->corner.x - Wp->inner.origin.x;
    relRp->origin.y = absRp->origin.y - Wp->inner.origin.y;
    relRp->corner.y = absRp->corner.y - Wp->inner.origin.y;
    return relRp;
}
/***************************************************************/
/*
 * absRect -- convert window-relative AxRectangle to "absolute"
 */
AxRectangle *
absRect(inRectp, outRectp)
    AxRectangle *inRectp, *outRectp;
{
    outRectp->origin.x = inRectp->origin.x + Wp->inner.origin.x;
    outRectp->origin.y = inRectp->origin.y + Wp->inner.origin.y;
    outRectp->corner.x = inRectp->corner.x + Wp->inner.origin.x;
    outRectp->corner.y = inRectp->corner.y + Wp->inner.origin.y;
    return outRectp;
}

/***************************************************************/
/* 
 * absRectOuter -- convert window-relative AxRectangle to "absolute"
 */
AxRectangle *
absRectOuter(inRectp, outRectp)
    AxRectangle *inRectp, *outRectp;
{
    outRectp->origin.x = inRectp->origin.x + Wp->outer.origin.x;
    outRectp->origin.y = inRectp->origin.y + Wp->outer.origin.y;
    outRectp->corner.x = inRectp->corner.x + Wp->outer.origin.x;
    outRectp->corner.y = inRectp->corner.y + Wp->outer.origin.y;
    return outRectp;
}

/***************************************************************/
/*
 * inRect -- Is pixel (xx, yy) in rectangle rp ?
 */
int
inRect(xx, yy, rp)
    short xx, yy;
    AxRectangle *rp;
{
    return ( xx >= rp->origin.x
            && xx < rp->corner.x
            && yy >= rp->origin.y
            && yy < rp->corner.y );
}
/***************************************************************/
/*
 * rectInRect -- Is rectangle 1 is contained within rectangle 2 ?
 */
int
rectInRect(rp1, rp2)
    AxRectangle *rp1, *rp2;
{
    return( rp1->origin.x >= rp2->origin.x &&
        rp1->origin.y >= rp2->origin.y &&
        rp1->corner.x <= rp2->corner.x &&
        rp1->corner.y <= rp2->corner.y );
}
/***************************************************************/
int
rectEq(rp1, rp2)
    AxRectangle *rp1, *rp2;
{
    return( rp1->origin.x == rp2->origin.x &&
        rp1->origin.y == rp2->origin.y &&
        rp1->corner.x == rp2->corner.x &&
        rp1->corner.y == rp2->corner.y );
}



/* Return a pointer to the window stucture for a particular window */
/* This is kept as a small function as part of a movement away from the */
/*    notion of a global window structure, and also to ease inclusing of */
/*    files which get linked into elfrt - jon                            */

struct wind_struct *GetPointerToGLobalWinStruct( wid )
int wid;
{
    return( &window[wid] );
}


/* OBS STUBS */

void atm_rls(){}
void atm_rst(){}
void atm_rpl(){}
void atm_rec(){}
void mesRec(){}
void qaPost(){}
void CursVis(){}
char QAplayback;
char QAdelay;
short AppPlayback;
short LocTmR;
short LocWid;
void regenEvent(){}
void reqRegen(){}
