/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmtabwid.c
   Applix ASC version: 2
   Date: Mon Nov 16 13:53:37 1998
*/
#include	<aos.h>
#include	<paratm.h>

#define	TAB	'\t'
/*
 *
 */
int
getTabWid(tex, lenb, txlen, x, tabx, tabtype,
			to1, to2, mintab, decimal)
char		*tex;		/* text string */
short		*lenb;		/* corresp. width array ptr */
int			txlen;		/* length of string */
int			x;			/* current x pix position */
int			tabx;		/* x pos of tab stop */
int			tabtype;
short		*to1, *to2; /* formatting clues */
short		mintab;		/* minimum width for adjusted tab */
char		decimal;	/* charcode of decimal char */
{
	char	*tptr;
	int		wid, i, twid, dwid, dpt, ttwid;

	if(tabtype != TM_LEFT_TAB)
	{
		/*
		 * If the tab type != LEFT, then we must
		 * compute the length of the substring which
		 * starts just after the tab and extends to
		 * just before the next tab, or to the end of
		 * the textstring if no more tabs are found.
		 *
		 * Use the term "tabtext" to refer to this
		 * substring, which is thought of as "attatched to"
		 * or "under the influence of" the current tab.
		 *
		 */
		tptr = tex + 1;
		i = 1;
		twid = 0;
		dwid = -1;	/* dwid==0 could mean ".123" e.g. */
		dpt = FALSE;
		while(*tptr != TAB && i < txlen)
		{
			if (*tptr==decimal && tabtype==TM_DECIMAL_TAB) 
			{
				dwid = twid;
				dpt = TRUE;
			}
			twid += lenb[i++];
			tptr++;
		}
	}
	wid = tabx - x;		/* tentative - "simple width" */
	/*
	 * Record the worst overlap.
	 * (Encroaching on the mintab limit
	 * is also considered overlap.)
	 */
	if (wid < mintab)
	{
		if (to1 && mintab-wid > *to1)
		{
			*to1 = mintab - wid;
		}
		/*
		 * Limit tabwidths to +mintab.  If another
		 * tab is to be found, the intervening
		 * text width may produce a larger expression
		 * of overlap error.  The largest such error
		 * can be used to help guess how far to bump the
		 * expansion factor (dc.)  Intended just to
		 * avoid gross wysiwyg errors on non-fontmatched
		 * screen-printer views; less severe errors will
		 * remain.
		 *
		 * At display/print time (or at any time for
		 * fontmatched pairs,) this condition better
		 * not occur.
		 */
		wid = mintab;
	}
	ttwid = wid - mintab;

	switch(tabtype)
	{
		case TM_LEFT_TAB:
			break;
		case TM_CENTER_TAB:
			/*
			 * Center the tabtext on the stop.  If the
			 * tabtext is too long for this, then the
			 * tab width is mintab.
			 */
			if ((twid>>1) <= (ttwid))
			{
				wid -= (twid>>1);
			}
			else
			{
				/*
				 * Push.  Record max on *to2.
				 */
				if (to2 && (twid>>1)-ttwid > *to2)
				{
					*to2 = (twid>>1)-ttwid;
				}
				wid = mintab;
			}
			break;
		case TM_RIGHT_TAB:
		case TM_DECIMAL_TAB:
			/*
			 * For decimal tabtext that contains a
			 * decimal point, adjust the effective
			 * tabtext width.
			 */
			if (dpt==TRUE)
			{
				twid = dwid;
			}
			/*
			 * The tabtext should be right-justified
			 * against the stop if there's room.
			 */
			if (twid <= ttwid)
			{
				wid -= twid;
			}
			else
			{
				/*
				 * Push.  Record max on *to2.
				 */
				if (to2 && twid-ttwid > *to2)
				{
					*to2 = twid-ttwid;
				}
				wid = mintab;
			}
			break;
	}
	return wid;
}

