/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmximage.c
   Applix ASC version: 36
   Date: Mon Nov 16 13:53:34 1998
*/
#include <basicwin.h>
#include <stdio.h>
#include <assert.h>
#include <aos.h>
#include <paripc.h>
#include <atm.h>
#include <atmimage.h>
#include <atmattr.h>
#include <parimage.h>
#include <paratmgr.h>
#include <parcint.h>

static int WidthInMils;    /* unit x-size in mils */
static int HeightInMils;   /* unit y-size in mils */
static int WidthInPixels;  /* unit x-size in pixels */
static int HeightInPixels; /* unit y-size in pixels */

extern int TmAmSE;

/******************************************************************************
 * old calls - unused
 */
typeErr TmPutColormap(ix, n, r, g, b) int ix, n; char r[], g[], b[]; { }
typeErr TmGetColormap(ix, n, r, g, b) int ix, n; char r[], g[], b[]; { }
typeErr TmPutMask(wID, mask) short wID; int *mask; { }
typeErr TmGetMask(wID, mask) short wID; int *mask; { }
typeErr TmInclude() {}
typeErr TmPutImage() {}
typeErr TmGetImage() {}

/******************************************************************************
 * Utility functions for coordinate conversions
 */
getCoords()
{
    short discard;
    struct TmFeatures features;

    TmDchar ( &discard, &discard, &discard, &discard, &features, -1 );
    HeightInPixels   = features.grPHt;
    WidthInPixels  = features.grPWid;
    HeightInMils   = features.grMHt;
    WidthInMils  = features.grMWid;
}
typeErr TmMilPix(xp, yp)
short *xp, *yp;
{
    if ( WidthInMils == 0 )
        getCoords();
    *xp = ( (long) (*xp) * WidthInPixels) / WidthInMils;
    *yp = ( (long) (*yp) * HeightInPixels) / HeightInMils;
}
typeErr TmPixMil(xp, yp)
short *xp, *yp;
{
    if ( WidthInMils == 0 )
        getCoords();
    *xp = ( (long) (*xp) * WidthInMils + WidthInPixels - 1 ) / WidthInPixels;
    *yp = ( (long) (*yp) * HeightInMils + HeightInPixels - 1 ) / HeightInPixels;
}
short
TmXMilPix(xx)
short xx;
{
    if ( WidthInMils == 0 )
        getCoords();
    return (long) xx * WidthInPixels / WidthInMils;
}
short
TmYMilPix(yy)
short yy;
{
    if ( WidthInMils == 0 )
        getCoords();
    return (long) yy * HeightInPixels / HeightInMils;
}
short
TmXPixMil(xx)
short xx;
{
    if ( WidthInMils == 0 )
        getCoords();
    return ( (long) xx * WidthInMils + WidthInPixels - 1 ) / WidthInPixels;
}
short
TmYPixMil(yy)
short yy;
{
    if ( WidthInMils == 0 )
        getCoords();
    return ( (long) yy * HeightInMils + HeightInPixels - 1 ) / HeightInPixels;
}
mlPxRect(rp)
TmRect *rp;
{
    /* convert a TmRect from mils to pixels */
    TmMilPix ( &rp->x_position, &rp->y_position );
    TmMilPix ( &rp->x_size, &rp->y_size );
}

/****************************************************************************/
TmCreateImage(w, pid, pts, use_color, imP, cmap)
long w;
int pid;
TmPoint32 pts[2];
int use_color;
ge_image_type *imP;
cmap_type *cmap;
{
    /* if no window, don't do anything */
    if ( !w )
        return TMR_NOWIN;

    /* grab the image */
    if ( TmAmSE == FALSE )
        ax_tm_create_image(w, pid, pts, use_color, imP, cmap );

    /* done */
    return AOK;
}
 
TmGetImage2(w, pid, pts, use_color, imP, cmap)
long w;
int pid;
TmPoint32 pts[2];
int use_color;
ge_image_type *imP;
cmap_type *cmap;
{
    /* if no window, don't do anything */
    if ( !w )
        return TMR_NOWIN;

    /* grab the image */
    if ( TmAmSE == FALSE )
        ax_tm_gimage(w, pid, pts, use_color, imP, cmap );

    /* done */
    return AOK;
}

AxScaleImage(w, pid, imP, cmap, xform, x_res, y_res)
long w;
int pid;
ge_image_type *imP;
cmap_type *cmap;
poly_type *xform;
int x_res,y_res;
{
	double x_scale,y_scale;
	int status;

	/* if no window, don't do anything */
	if ( !w )
		status = TMR_NOWIN;
	else /* scale the image */
	{
		if ( TmAmSE == FALSE ) /* don't scale for printing */
		{
			x_scale = (double)((double)x_res/1000.0);
			y_scale = (double)((double)y_res/1000.0);
			ax_tm_simage(pid, imP, cmap, xform, x_scale, y_scale);
		}
		status = AOK;
	}
	return(status);
}

AxScaleJavaImage(w, pid, imP, cmap, xform, x_res, y_res)
long w;
int pid;
ge_image_type *imP;
cmap_type *cmap;
poly_type *xform;
int x_res,y_res;
{
	double x_scale,y_scale;
	int status;

	/* if no window, don't do anything */
	if ( !w )
		status = TMR_NOWIN;
	else /* scale the image */
	{
		if ( TmAmSE == FALSE ) /* don't scale for printing */
		{
			x_scale = (double)((double)x_res/1000.0);
			y_scale = (double)((double)y_res/1000.0);
			ax_tm_sjimage(pid, imP, cmap, xform, x_scale, y_scale);
		}
		status = AOK;
	}
	return(status);
}

AxPutImage(w, pid, pnt, im, attr, xform, x_res, y_res, cmap, clip)
long w;
int pid;
TmPoint32 *pnt;
ge_image_type *im;
attr_type *attr;
poly_type *xform;
int x_res,y_res;
cmap_type *cmap;
TmRect32 *clip;
{
    double x_scale, y_scale;
    int status;

    /* render the image */
    x_scale = (double)((double)x_res/1000.0);
    y_scale = (double)((double)y_res/1000.0);

    if ( TmAmSE ) /* printing */
    {
        ax_se_pimage(w, pnt, im, attr, xform, x_scale, y_scale, cmap, clip);
        status = AOK; /* treat as OK on success or failure */
    }
    else /* displaying (could be client or server display) */
    {
        if ( !w ) /* if no window, don't do anything */
            status = TMR_NOWIN;
        else
        {
            ax_tm_pimage(w, pid, pnt, im, attr, xform,
                         x_scale, y_scale,
                         cmap,
                         clip);
            status = AOK;
        }
    }    

    /* done */
    return(status);
}
