/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: toolbcom.c
   Applix ASC version: 45
   Date: Mon Jan 11 21:14:06 1999
*/
#include <stdio.h>
#include <basicwin.h>
#include <aos.h>
#include <xmlib.h>
#include <xmpixel.h>
#include <elxwins.h>


#define TM_MAXWIN 21            /* copied atmwin.h!! */


extern AXWidget WindowList;
extern int MenuUID;             /* counter of menu starts */

extern AXWidget AxPulldownActive; /* non-zero while menubar pulldown active */
extern int WidgetNeedsUpdated;   /* Set when a widget needs repainted */
extern Display *Dpy;
extern int AxNoMapWindows;
extern int ToolbarExtraSep; /* nominally 25 */
extern int ToolbarButtonSep;    /* nominally 2 */
extern int AxHideDisabledTools; /* TRUE==> undisplay disabled toolbar buttons */
extern GC olButtonGC;
extern GC  SunGC, ShadowGC, BackGC;
extern int AxOpenLook;

#if AX_X11
XSetWindowAttributes WinAtt;
#endif

#ifndef PALETTE_WIDTH    /* temp till it gets into xmlib.h */
#define PALETTE_WIDTH 101
#endif

#include <toolbcom.p>


/*************************************************************************
function: PaintToolBar - Paint a toolbar
*/
void PaintToolBar (w, exP, regen, layered)
register AXWidget_tool w;  /* W_TOOLBAR */
XExposeEvent *exP;     /* NULL or exposure event.  If null, do all */
int regen;
int layered;
{
    int ix;
    AXWidget childP;
    menu_item_def *mP;

    if ( !w || (w->type != W_TOOLBAR) )
        return;
    if ( w->parent && w->parent->dialog && w->parent->dialog->topmostwindow )
        return;

    if ((w->flags & WFLG_IN_EDIT) && layered)
        {
        if (dePaintWidget((AXWidget)w))
            return;
    }

    if ( !w->window )
        {
        w->xpos = w->parent->palette && w->parent->palette->display ?
                                                     PALETTE_WIDTH : 0;
        w->ypos = w->parent->menuBar->height;
        if ( w->height && w->display )
            {  /* toolbar needed */
#if AX_X11    /* toolbar window created elsewhere on Windows */
            w->window = AxCrtWindow(Dpy,w->parent->window,
                                    w->xpos, w->ypos, w->width, w->height,
                                    0,CopyFromParent,InputOutput,
                                    CopyFromParent, 
                                    (CWBackPixel | CWWinGravity | CWBitGravity),
                                    &WinAtt);
            XSelectInput(Dpy, w->window, TM_ALL_INPUTS_MASK);
            AxMapWindow((AXWidget)w, Dpy, w->window, w->name);
#endif
            w->mapped = TRUE;
        }
    }

    if ( (w->height == 0) || !w->display )
        {
#if AXMS
        if ( w->winid )
            {
            MS_DestroyWindow( w->winid );
            w->winid = (HWND)NULL;
            }
#else
        if ( w->window )
            {    /* take away old one */
            AxDestroyWindow(Dpy, (AXWidget)w, w->window);
            w->window = 0;
            }
#endif
        return;
    }

/* toolbar buttons are either text or bitmap.
 * if they are text, MS Windows will draw them on the
 * CreateWindow call.  if they are bitmaps then they
 * will have the Windows OWNERDRAW style and are painted
 * elsewhere.
*/  
#if AX_X11    
    /* paint toolbar buttons */
    for (ix = 0; ix < MAX_TOOLBAR_ITEMS; ix++)
        {
        mP = w->toolBarInfo->menuP[ix];
        if (!mP)
            break;
        childP = w->toolBarInfo->widgets[ix];
        
        if ( childP && childP->painter )
            (*childP->painter)( childP, NULL, regen, layered );
    }
#endif

    if ( AxOpenLook )
        XDrawLine(Dpy, w->window, olButtonGC, 0, 0, w->width, 0);   /* ??T */
    else
        {
        AxDrawShadow(w->window, SunGC,ShadowGC, w->thickness, 0, 0, 
							w->width, w->height, OLSHADOW_NODRAW);
                                                             
    }
}




/*************************************************************************
function: AxToolButtonPushed - Respond to push of a toolbar button
*/
void AxToolButtonPushed (widget, tp, unused_data)
AXWidget widget;
AXWidget_tool tp;
mbchar *unused_data;
{
    AXWidget w, childP;
    menu_item_def *mP, menudata;
    int ix;

    w = widget;
    if ( !w )
        return;

    if ( !tp || (tp->type != W_TOOLBAR) )
        return;

    for ( ix = 0; ix < MAX_TOOLBAR_ITEMS; ix++ )
        {
        mP = tp->toolBarInfo->menuP[ix];
        if ( !mP )
            break;     /* end of list */
        if ( mP->type == ITEM_LINE )
            continue;  /* separator */
        childP = tp->toolBarInfo->widgets[ix];
        if ( w == childP )
            {
            /* set ATM's cur_win */
            SetCurWinFromMenuWidget(tp->parent->menuBar);
            /* make callback */
            memset(&menudata,0,sizeof(menudata));
            menudata.word = (char *)w->name;
            menudata.word2 = tp->toolBarInfo->menuP[ix]->word2;
            /* use the menubar's widget id for callback
             * (because menubar & toolbar items are the "same") */
            AxCallback(tp->callback,(AXWidget)tp->parent->menuBar,
                                  tp->parent->menuBar->callbacktag, 
							   	  (caddr_t)&menudata);
            break;
        }
    }
}

/*************************************************************************
function: ZXmCreateToolBar - Create a toolbar associated with menubar
Called only by ZXmCreateMenuBar!
*/
void
ZXmCreateToolBar(parentWidget, mP, flags, elf_obj, callback_func, callback_pid)
register AXWidget_main parentWidget;
register menu_item_def *mP;     /* the elf menu bar data */
int flags;
data_ptr elf_obj;
void (*callback_func)();
caddr_t callback_pid;             /* a bad type, but that is how it is in the */
{
    register AXWidget_tool w;
    int infodex = 0;
    int status;
    short pid;

    if ( !mP || !parentWidget->display || AxNoMapWindows )
        return;
   
    AxDestroyWidget((AXWidget)parentWidget->toolBar);
    w = (AXWidget_tool)CrtWidget(W_TOOLBAR,(AXWidget)parentWidget,"");
    parentWidget->toolBar = w;
    w->painter = PaintToolBar;
    AddToWindowList((AXWidget)w);

    /* now set the various input flags and values into the widget */
    SetWidgetFlags( (AXWidget)w, flags );
    w->elfObject = AxTaskCopyElfData(0,elf_obj);
    w->callback = callback_func;
    ((AXWidget_dialog)w)->callbacktag = callback_pid;

#if AXMS
	w->thickness = 1;
#else
	w->thickness = 2;
#endif

    w->update_widget = TRUE;
    WidgetNeedsUpdated = TRUE;

    w->toolBarInfo = (tool_bar_t *)calloc(1, sizeof(tool_bar_t));

    MenuUID++;
    pid = THIMpid;
    MenuBarToToolBar(w, mP, &infodex);
    VfyContext(pid);

    UpdateToolBarDisplay((AXWidget)w);

    if ( parentWidget->dialog )
        {
        status = AxXmToolBarStatus(parentWidget);
        XmMenuStatus(parentWidget, status, NXLT("DLG.VIEW_EXPRESSLINE"));
    }   
}

/*************************************************************************
function: MenuBarToToolBar - Recursively add stuff from menu struct to toolbar
*/
void MenuBarToToolBar (w, mP, infodexP)
register AXWidget_tool w;
register menu_item_def *mP;     /* the elf menu bar data */
int *infodexP;
{
    AXWidget_pushbutton bp;
    char *name;
    int labelPixmap;

    while ( mP )
    {
        labelPixmap = FALSE;
        if ( (mP->type == ITEM_MARK) && mP->ender )
            break;
        if ( mP->type == ITEM_MENU )
            MenuBarToToolBar(w, (menu_item_def*)mP->word2, infodexP);
        if ( (mP->type == ITEM_LINE) && mP->toolshow )
        {
            /* separator */
            if ( *infodexP >= MAX_TOOLBAR_ITEMS )
                break;
            w->toolBarInfo->menuP[*infodexP] = mP;
            w->toolBarInfo->widgets[*infodexP] = NULL;
            *infodexP = *infodexP + 1;
        }
        if ( (mP->type == ITEM_FUNC) && mP->toolshow )
        {
            if ( *infodexP >= MAX_TOOLBAR_ITEMS )
                break;
            if ( (mP->tooltype == TOOLTYPE_BITMAP) &&
                  mP->bitmap && mP->bitmap[0] && Pixmapper(mP->bitmap) )
            {                                     
                /* use bitmap */
                labelPixmap = TRUE;
                name = mP->bitmap;
            }
            else if ( (mP->tooltype == TOOLTYPE_TOOLWORD) &&
                       mP->toolword && mP->toolword[0] )
            {
                /* use toolbar mnemonic */
                name = mP->toolword;
            }
            else
            {
                /* use menubar word (dflt: mP->tooltype==TOOLTYPE_NAME) */
                name = mP->word;
            }
            /* Use toolbar as callbacktag - easier to locate menu 
            * struct when callback occurs.  Use the corresponding
            * menubar's callbacktag to send msg to calling app. */
            bp =  AxCreatePushButtonGadget((AXWidget)w, name, 0, 0,
                                   AxToolButtonPushed, (caddr_t)w,
                                   Button1, labelPixmap, NULL,
                                    0,0,0,0);
            bp->display = FALSE;  /* will be set in BuildToolBar */
            w->toolBarInfo->menuP[*infodexP]   = mP;
            w->toolBarInfo->widgets[*infodexP] = (AXWidget)bp;
            *infodexP = *infodexP + 1;
        }
        mP++;
    }
}

/*************************************************************************
function: BuildToolBar - compute tool button positions, toolbar height
*/
void BuildToolBar(w, tool_width)
AXWidget_tool w;     /* W_TOOLBAR */
int tool_width;  /* new toolbar width */
{
    AXWidget bp;
    int xpos = TOOLBAR_B1_XPOS;
    int ypos = 4;
    int rowNum = 0;
    int rowHeights[MAX_TOOLBAR_ROWS];
    int ix, this_height;
    menu_item_def *mP;

    if ( !w || (w->type != W_TOOLBAR) )
        return;

    /* toolbar width */
    w->width = tool_width;

    /* compute button xpos & ypos, row heights */
    memset(rowHeights, 0, sizeof(rowHeights));
    for ( ix = 0; ix < MAX_TOOLBAR_ITEMS; ix++ )
        {
        mP = w->toolBarInfo->menuP[ix];
        if ( !mP )  /* end of list, done */
            break;
        if ( mP->type == ITEM_LINE )
            {
            /* separator: add space */
            xpos += ToolbarExtraSep;
            continue;
        }
        bp = w->toolBarInfo->widgets[ix];
        if ( !bp )
            continue;   /* sanity */
        if ( bp->display == FALSE )
            continue;   /* if not displayed don`t bother */
        if ( (xpos + bp->width) >= w->width )
            {
            /* button won't fit entirely on this row */
            if ( xpos > TOOLBAR_B1_XPOS )
                {
                /* not 1st row: move button to next row */
                if ( rowNum < MAX_TOOLBAR_ROWS )
                    {
                    /* button is visible */
                    xpos = TOOLBAR_B1_XPOS;
                    ypos += rowHeights[rowNum];
                    rowNum++;
            }
        }
        }
        bp->xpos = xpos;
        bp->ypos = ypos;
        if ( rowNum >= MAX_TOOLBAR_ROWS )
            bp->display = FALSE;
        else
            bp->display = TRUE;
        xpos += bp->width + ToolbarButtonSep;
        this_height = bp->height + 2;  /* 2 pix padding between rows */
        if ( (rowNum < MAX_TOOLBAR_ROWS) && (this_height > rowHeights[rowNum]) )
            rowHeights[rowNum] = this_height;
    }

    /* compute toolbar height, center buttons in row */
    ypos = 0;/* < initial so "if (bp->ypos > ypos)" executed even for 1st row */
    rowNum = -1;  /* -1 so that will become 0 after first increment */
    w->height = 0;
    for ( ix = 0; ix < MAX_TOOLBAR_ITEMS; ix++ )
        {
        mP = w->toolBarInfo->menuP[ix];
        if ( !mP )  /* end of list, done */
            break;
        if ( mP->type == ITEM_LINE )
            continue;  /* separator */
        bp = w->toolBarInfo->widgets[ix];
        if ( !bp )
            continue;   /* sanity */
        if ( bp->display == FALSE )
            continue;   /* if not displayed don`t bother */
        if ( bp->ypos > ypos )
            {
            ypos = bp->ypos;
            rowNum++;
            if ( rowNum >= MAX_TOOLBAR_ROWS )
                break;
            w->height += rowHeights[rowNum];
        }
        bp->ypos = ypos + (rowHeights[rowNum] - 2 - bp->height)/2;
    }
    if ( w->height )
        w->height += 6;  /* top & bottom padding */

    /* since button display has been forced here, old
     * menu status has been over-written if profile
     * PROF_EL_HIDE_DISABLED_ has been set */
    if ( AxHideDisabledTools )
        {
        /* inactive buttons should not be displayed: 
         * update button status */
        UpdateToolBarDisplay((AXWidget)w);
    }
}

/*************************************************************************
function: MenuItemFromTool - Return menu_item struct from toolbar button
*/
menu_item_def *
MenuItemFromTool(w)
AXWidget w;
{
    int ix;
    AXWidget_tool tp;
    menu_item_def *mP = NULL;

    if ( !w )
        return(NULL);

    tp = (AXWidget_tool) w->parent;
    if ( !tp || (tp->type != W_TOOLBAR) )
        return(NULL);

    for ( ix = 0; ix < MAX_TOOLBAR_ITEMS; ix++ )
        {
        mP = tp->toolBarInfo->menuP[ix];
        if ( !mP )  /* end of list, done */
            break;
        if ( tp->toolBarInfo->widgets[ix] == w )
            break;
    }
    return(mP);
}


/*************************************************************************
function: AxToolBarHeight - recompute toolbar geometry and return height 
*/
AxToolBarHeight (w, tool_width)
AXWidget_main w;    /* W_MAIN containing toolbar */
int tool_width;
{
    if ( !w || !w->toolBar || !w->menuBar )
        return(0);
    if ( !w->toolBar->display )
        return(0);

    BuildToolBar(w->toolBar, tool_width);
    return(w->toolBar->height);
}


/*************************************************************************
function: AxXmToolBarStatus - return current display status of toolbar
*/
AxXmToolBarStatus(w)
AXWidget_main w;    /* W_MAIN containing toolbar */
{
    int status = MENUSTAT_NO_SHOW;

    if (w && (w->type == W_MAIN) && w->toolBar) {
        if (!w->toolBar->toolBarInfo->menuP[0]) {
            /* nothing in toolbar - gray out choice */
            status = MENUSTAT_INACTIVE;
        } else {
            status = w->toolBar->display ?
                MENUSTAT_TOGGLE_ON : MENUSTAT_TOGGLE_OFF;
        }
    }
    return (status);
}


/***************************************************************************
function: TmUpdateToolBar
*/
void
TmUpdateToolBar(w)
long w;
{
    if (w)
        UpdateToolBarDisplay((AXWidget)w);
}


/****************************************************************************
function: UpdateToolStatus - Update menu and display status of toolbar button
*/
void UpdateToolStatus(w, mP)
AXWidget_pushbutton w;          /* W_BUTTON */
menu_item_def *mP;
{
    int flag, display_it, val;

    if (!w || (w->type != W_BUTTON))
        return;
    if (!w->parent || (w->parent->type != W_TOOLBAR))
        return;
    if (!mP)
        return;

    flag = FlagsFromApp(w->parent, mP, TRUE);
    if (flag == MENUSTAT_NO_SHOW) {
        if (w->display) {
            w->display = FALSE;
            w->update_widget = TRUE;
            WidgetNeedsUpdated = TRUE;
        }
        return;
    } else {
        display_it = (flag == MENUSTAT_INACTIVE ? FALSE : TRUE);
    }
    if (AxHideDisabledTools) {
        if (w->display != (unsigned int) display_it) {
            if (w->display == FALSE) {
                /* don't change status if button not in visible row */
                if ((w->ypos + w->height) > w->parent->height) {
                    return;
                }
            }
            w->display = display_it;
            w->update_widget = TRUE;
            WidgetNeedsUpdated = TRUE;
        }
    } else {
        if (!w->display) {
            /* may have been undisplayed earlier */
            w->display = TRUE;
            w->update_widget = TRUE;
            WidgetNeedsUpdated = TRUE;
        }
        if (w->disabled == (unsigned int) display_it) {
            w->disabled = !display_it;
            w->update_widget = TRUE;
            WidgetNeedsUpdated = TRUE;
        }
    }
    if (w->display && !w->disabled) {
        if ((flag == MENUSTAT_TOGGLE_ON) || (flag == MENUSTAT_RADIO_ON))
            val = TRUE;
        else
            val = FALSE;
        if (w->value != val) {
            w->value = val;
            w->update_widget = TRUE;
            WidgetNeedsUpdated = TRUE;
        }
    }
}


/*************************************************************************
function: UpdateAllToolBars - Update menu and display status of ALL toolbars
*/
void UpdateAllToolBars()
{
    AXWidget w;

    for ( w = WindowList; w; w = w->nextWin ) {
        if ( (w->type == W_MAIN) && ((AXWidget_main)w)->toolBar )
            UpdateToolBarDisplay(w);
    }
}

/*************************************************************************
function: UpdateToolBarDisplay - Update menu and display status of
toolbar entries
*/
void UpdateToolBarDisplay(w)
AXWidget w;   /* either W_MAIN or W_TOOLBAR */
{
    short pid;
    menu_item_def *mP;
    AXWidget childP;
    AXWidget_tool w_toolbar;
    int ix;

    if ( !w )
        return;
    if ( w->type == W_MAIN )
        w_toolbar = ((AXWidget_main)w)->toolBar;
    else
        w_toolbar = (AXWidget_tool)w;

    if ( !w_toolbar || w_toolbar->type != W_TOOLBAR )
        return;

    MenuUID++;
    pid = THIMpid;
    for ( ix = 0; ix < MAX_TOOLBAR_ITEMS; ix++ ) {
        mP = w_toolbar->toolBarInfo->menuP[ix];
        if ( !mP )  /* end of list, done */
            break;
        if ( mP->type == ITEM_LINE )
            continue;  /* separator */
        childP = w_toolbar->toolBarInfo->widgets[ix];
        if ( !childP )
            continue;       /* sanity */
        UpdateToolStatus((AXWidget_pushbutton)childP, mP);
    }
    VfyContext(pid);
}


/*************************************************************************
function: AxXmToggleToolBar - display/hide toolbar
*/
void AxXmToggleToolBar(w)
AXWidget_main w;   /* W_MAIN containing toolbar */
{
    if ( w && (w->type == W_MAIN) && w->toolBar ) {
        w->toolBar->display = !w->toolBar->display;
        XmRestructureWindow(w);
        if ( w->dialog )
            XmMenuStatus(w, AxXmToolBarStatus(w), NXLT("DLG.VIEW_EXPRESSLINE"));
    }
}

AxGetToolBarHeight(w)
AXWidget_main w;
{
    if ( !w || !w->toolBar || !w->menuBar )
        return(0);
    if ( !w->toolBar->display )
        return(0);
    return(w->toolBar->height);
}

AxGetToolBarWidth(w)
AXWidget_main w;
{
    int bar_width=0, palt_width = 0;

    if ( w->menuBar )
        bar_width = MAX(30, w->width);
    if ( w->palette && w->palette->display )
        palt_width = PALETTE_WIDTH;

    return(bar_width - palt_width);
}

/*************************************************************************
function: AxRebuildToolBar - recompute toolbar geometry and return true/false 
*/
AxRebuildToolBar(w)
AXWidget_main w;    /* W_MAIN containing toolbar */
{
    int tool_wid, old_hyt;
    if ( !w || !w->toolBar || !w->menuBar )
        return(FALSE);
    if ( !w->toolBar->display )
        return(FALSE);
    tool_wid = AxGetToolBarWidth(w);
    old_hyt  = AxGetToolBarHeight(w);
    UpdateToolBarDisplay((AXWidget)w);
    BuildToolBar(w->toolBar, tool_wid);
    if ( old_hyt != w->toolBar->height )
    {
        XmRestructureWindow(w);
        return(TRUE);
    }
    else
    {
        XFillRectangle(Dpy,w->toolBar->window,BackGC,0,0,
                       w->toolBar->width, w->toolBar->height);
        PaintToolBar(w->toolBar, 0, 0, FALSE);
    }
    return(FALSE);
}
