/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tree.c
   Applix ASC version: 22
   Date: Mon Nov 16 13:53:37 1998
*/
#if AX_X11
/* don't do directory tree testing on NT for now. */
#define TESTING 1
#endif

#if TESTING
#include <stdio.h> /* for dd */
#include <sys/types.h> /* for dd */
#include <sys/stat.h> /* for dd */
#include <dirent.h> /* for dd */
#endif

#include <aos.h>
#include <elf_env.h>
#include <eldata.h>
#include <eltree.h>
#include <tree.h>
#include <intsarr.h>

#if !AXMS && !LINUX
extern char *sys_errlist[]; /* for dd */
#endif

#if TREE_DEBUG
#define TLOG(m) printf m
#else
#define TLOG(m)
#endif

#if 0 /* FIXME, move this to intsarr.h */
#define FAST_INTS_ARRAY 1

#if FAST_INTS_ARRAY
#define WriteIntsArray(p,n,val) ((p)[(n)+1] = (val))
#define ReadIntsArray(p,n) ((long) (p)[(n)+1])
#define IntsArraySize(p) ((p) ? *(p) : 0)
#endif
#endif

static void AddCache AXP_((treeNode *nodeP));
static void DelCache AXP_((treeNode *nodeP));
static treeNode *AllocTreeNode AXP_((void));
static void FreeTreeNode AXP_((treeNode *nodeP, int fast));
static void IDestroyTree AXP_((treeNode *nodeP));

/* begin for dd */
#if TESTING
static int IsADirectory AXP_((char *name));
static treeNode *ListDir AXP_((char *dirname, int recurse));
static void IListDir AXP_((char *dirname, treeNode *parent_node, int recurse));
#endif
/* end for dd */

/* Add NODEP to it's top level node's cache.
   Silently return if top node is not found. */
static void
AddCache(nodeP)
register treeNode *nodeP;
{
    register treeNode *topP;
    register long *newCache;

    if (!nodeP)
        return;
    topP = FindTopNode(nodeP);
    if (!topP) {
        TLOG(("*** AddCache: did not find topP\n"));
        return;
    }
    if (InIntsArray(topP->u.cache, NODE_ID(nodeP))) {
        TLOG(("*** AddCache: 0x%lx already in cache\n"));
        return; /* alread there. ERROR */
    }

    newCache = AddIntsArray(topP->u.cache, NODE_ID(nodeP));
    if (topP->u.cache)
        free(topP->u.cache);
    topP->u.cache = newCache;
    TLOG(("+++ AddCache: new cache size %d\n", IntsArraySize(topP->u.cache)));
    return;
}

/* Delete NODEP from it's top level node's cache.
   Silently return if top node is not found. */
static void
DelCache(nodeP)
register treeNode *nodeP;
{
    register treeNode *topP;
    long *newCache;

    if (!nodeP)
        return;

    topP = FindTopNode(nodeP);
    if (!topP)
        return;

    newCache = DeleteIntsArray(topP->u.cache, NODE_ID(nodeP));
    if (topP->u.cache)
        free(topP->u.cache);
    topP->u.cache = newCache;
}

/* Set the display title of NODEP to TITLE */
void
SetNodeTitle(nodeP, title)
treeNode *nodeP;
char *title;
{
    if (!nodeP || !title)
        return;
    if (nodeP->info.title)
        free(nodeP->info.title);
    nodeP->info.title = (char *) malloc(strlen(title) + 1);
    strcpy(nodeP->info.title, title);
    return;
}

/* Set the display icon pixmap of NODEP to ICON_PIXMAP. */
void
SetNodeIconPixmap(nodeP, icon_pixmap)
treeNode *nodeP;
char *icon_pixmap;
{
    if (!nodeP || !icon_pixmap)
        return;
    if (nodeP->info.icon_pixmap)
        free(nodeP->info.icon_pixmap);
    nodeP->info.icon_pixmap = (char *) malloc(strlen(icon_pixmap) + 1);
    strcpy(nodeP->info.icon_pixmap, icon_pixmap);
    return;
}

/* Set the value of NODEP to VALUE */
void
SetNodeValue(nodeP, value)
treeNode *nodeP;
int value;
{
    if (nodeP)
        nodeP->value = value;
    return;
}

/* Expand the title of a node. Return allocated expanded title on
   success, NULL otherwise. */
char *
ExpandNodeTitle(nodeP, sepStr)
treeNode *nodeP;
char *sepStr;
{
    char *expanded_title = (char *) NULL;
    int sep_len;

    if (!nodeP || !sepStr)
        return((char *) NULL);

    sep_len = strlen(sepStr);

    if (nodeP->level == 0) {
        if (nodeP->info.title) {
            expanded_title = (char *) malloc(strlen(nodeP->info.title) + 1);
            strcpy(expanded_title, nodeP->info.title);
        } /* else, return NULL */
    } else {
        char *p = (char *) NULL;
        int len;
        treeNode *tmpP = (treeNode *) NULL;

        tmpP = nodeP;
        while (1) { /* walk up to top */
            if (!tmpP)
                break;
            if (!tmpP->info.title)
                break;
            if (!expanded_title) {
                expanded_title = (char *)
                    malloc(strlen(tmpP->info.title) + 1);
                strcpy(expanded_title, tmpP->info.title);
            } else {
                len = (strlen(expanded_title) + sep_len +
                       strlen(tmpP->info.title));
                p = (char *) malloc(len + 1);
                strcpy(p, tmpP->info.title);
                strcat(p, sepStr);
                strcat(p, expanded_title);
                free(expanded_title);
                expanded_title = p;
            }
            if (tmpP->level == 0)
                break; /* reached top */
            tmpP = tmpP->u.parent;
        }
    }
    return(expanded_title);
}

/* Expand the `value' of a node.  Return allocated expanded values on
   success; NULL otherwise.  Used only by openmail. */
char *
ExpandNodeValue(nodeP, sepStr)
treeNode *nodeP;
char *sepStr;
{
    char *expanded_value = (char *) NULL;
    int sep_len;
    char this_valstr[STRNORMAL + 1];
    treeNode *tmpP;
    int len;
    char *p;

    if (!nodeP || !sepStr)
        return((char *) NULL);

    sep_len = strlen(sepStr);
    
    if (nodeP->level == 0) {
        sprintf(this_valstr, "%d", nodeP->value);
        expanded_value = (char *) malloc(strlen(this_valstr) + 1);
        strcpy(expanded_value, this_valstr);
        goto finish;
    }

    tmpP = nodeP;
    while (TRUE) { /* walk up to top */
        if (!tmpP)
            break;
        if (!expanded_value) {
            sprintf(this_valstr, "%d%s", tmpP->value, sepStr);
            expanded_value = (char *) malloc(strlen(this_valstr) + 1);
            strcpy(expanded_value, this_valstr);
        } else {
            sprintf(this_valstr, "%d%s", tmpP->value, sepStr);
            len = strlen(expanded_value) + strlen(this_valstr);
            p = (char *) malloc(len + 1);
            strcpy(p, this_valstr);
            strcat(p, expanded_value);
            free(expanded_value);
            expanded_value = p;
        }
        if (tmpP->level == 0)
            break; /* reached top */
        tmpP = tmpP->u.parent;
    }

finish:
    return(expanded_value);
}

/* Return the node pointer of a node whose display title matches
   TITLE.  ANYP is a pointer to the top node of the tree or any node
   in the tree (the top can be found). */
treeNode *
FindNodeByTitle(anyP, title)
treeNode *anyP;
char *title;
{
    treeNode *topP;
    int size;
    int ix;

    if (!anyP || !title)
        return((treeNode *) NULL);
    if (anyP->level == 0)
        topP = anyP;
    else
        topP = FindTopNode(anyP);
    size = IntsArraySize(topP->u.cache);
    for (ix = 0; ix < size; ix++) {
        treeNode *thisP;

        thisP = (treeNode *) ReadIntsArray(topP->u.cache, ix);
        if (!strcmp(thisP->info.title, title))
            return(thisP);
    }
    return((treeNode *) NULL);
}

/* Return the node pointer of a node whose display title matches
   VALUE.  ANYP is a pointer to the top node of the tree or any node
   in the tree (the top can be found). */
treeNode *
FindNodeByValue(anyP, value)
treeNode *anyP;
int value;
{
    treeNode *topP;
    int size;
    int ix;

    if (!anyP)
        return((treeNode *) NULL);
    if (anyP->level == 0)
        topP = anyP;
    else
        topP = FindTopNode(anyP);
    size = IntsArraySize(topP->u.cache);
    for (ix = 0; ix < size; ix++) {
        treeNode *thisP;

        thisP = (treeNode *) ReadIntsArray(topP->u.cache, ix);
		if (thisP->value == value)
            return(thisP);
    }
    return((treeNode *) NULL);
}

/* Create a tree node whose parent node is PARENT.  If PARENT is NULL,
   the created node is marked as a top node.  Return a pointer to the
   created node. */
treeNode *
CrtTreeNode(parent)
treeNode *parent;
{
    treeNode *treeP;
    long *new_sub_nodes;

    TLOG(("CrtTreeNode: parent 0x%lx\n", parent));
    treeP = AllocTreeNode();
    if (!parent) {
        treeP->level = 0;
        treeP->u.cache = CreateIntsArray(0);
        AddCache(treeP); /* add self */
    } else {
        treeP->level = parent->level + 1;
        treeP->u.parent = parent;
        new_sub_nodes = AddIntsArray(parent->sub_nodes, NODE_ID(treeP));
        if (parent->sub_nodes)
            free(parent->sub_nodes);
        parent->sub_nodes = new_sub_nodes;
        AddCache(treeP);
    }
    treeP->sub_nodes = CreateIntsArray(0);
    return(treeP);
}

/* Allocate storage for a tree node.  Return pointer to the allocated
   node.*/
static treeNode *
AllocTreeNode()
{
    treeNode *treeP;

    TLOG(("AllocTreeNode...\n"));
    treeP = (treeNode *) calloc(1, sizeof(treeNode));
    treeP->info.title = (char *) calloc(1, sizeof(char));
    return(treeP);
}

/* Free storage for a tree node, and also remove it from the cache.
   If FAST it TRUE, do not delete from cache (takes a long time). */
static void
FreeTreeNode(nodeP, fast)
treeNode *nodeP;
int fast;
{
    TLOG(("FreeTreeNode...\n"));
    if (!nodeP)
        return;
    if (nodeP->info.icon_pixmap)
        free(nodeP->info.icon_pixmap);
    if (nodeP->info.title)
        free(nodeP->info.title);
    if (nodeP->sub_nodes)
        free(nodeP->sub_nodes);
    if (nodeP->level == 0) {
        if (nodeP->u.cache)
            free(nodeP->u.cache);
    } else {
        if (!fast)
            DelCache(nodeP);
    }

    free(nodeP);
}

/* Return a pointer to the top node of NODEP. */
treeNode *
FindTopNode(nodeP)
register treeNode *nodeP;
{
    register treeNode *tmpP;

    TLOG(("FindTopNode...\n"));
    if (!nodeP)
        return((treeNode *) NULL);
    tmpP = nodeP;
    while (1) {
        if (tmpP->level == 0)
            break;
        tmpP = tmpP->u.parent;
    }
    return(tmpP);
}

/* Return TRUE if NODEP is contained in the tree TOPP; FALSE
   otherwise. */
int
FindNode(topP, nodeP)
register treeNode *topP;
register treeNode *nodeP;
{
    TLOG(("FindNode...\n"));
    if (topP->level != 0)
        return(FALSE); /* a top node is level 0 */

    if (InIntsArray(topP->u.cache, NODE_ID(nodeP)))
        return(TRUE);

    return(FALSE);
}

/* Completely delete a node; free it's storage */
void
DelTreeNode(nodeP)
treeNode *nodeP;
{
    long *parent_sub_nodes = (long *) NULL;
    long *new_parent_sub_nodes = (long *) NULL;
    int ix, num_subnodes;
	long *sub_nodes = (long *) NULL;

    TLOG(("DelTreeNode...\n"));
    if (!nodeP)
        return; /* invalid args */

/* is a top node, destroy the whole tree. */
    if (!nodeP->level) {
        FastDestroyTree(nodeP);
        return;
    }

    if (!nodeP->u.parent) /* sanity */
        return; /* ERROR */

/* first delete all nodes underneath. */
	sub_nodes = CopyIntsArray(nodeP->sub_nodes);
    num_subnodes = IntsArraySize(sub_nodes);
    for (ix = 0; ix < num_subnodes; ix++)
        DelTreeNode((treeNode *) ReadIntsArray(sub_nodes, ix));
	if (sub_nodes)
		free(sub_nodes);
	sub_nodes = NULL;

/* detach from parent */
    parent_sub_nodes = nodeP->u.parent->sub_nodes;
    if (!InIntsArray(parent_sub_nodes, NODE_ID(nodeP))) {
        FreeTreeNode(nodeP, FALSE);
        return; /* not in parent's list. ERROR */
    }
    new_parent_sub_nodes = DeleteIntsArray(parent_sub_nodes, NODE_ID(nodeP));
    if (parent_sub_nodes)
        free(parent_sub_nodes);
    nodeP->u.parent->sub_nodes = new_parent_sub_nodes;

/* free it */
    FreeTreeNode(nodeP, FALSE);
    return;
}

/* Completely delete all subnodes of a node; free it's storage */
void
DelTreeSubnodes(nodeP)
treeNode *nodeP;
{
    long *parent_sub_nodes = (long *) NULL;
    long *new_parent_sub_nodes = (long *) NULL;
    int ix, num_subnodes;
	long *sub_nodes = (long *) NULL;

    TLOG(("DelTreeSubnodes...\n"));
    if (!nodeP)
        return; /* invalid args */

/* delete all nodes underneath */
	sub_nodes = CopyIntsArray(nodeP->sub_nodes);
    num_subnodes = IntsArraySize(sub_nodes);
    for (ix = 0; ix < num_subnodes; ix++)
        DelTreeNode((treeNode *) ReadIntsArray(sub_nodes, ix));
	if (sub_nodes)
		free(sub_nodes);
    return;
}

/* Reset all display information (x, y positions and dimensions) to 0. */
void
ResetTreeDispInfo(topP)
register treeNode *topP;
{
    register int size;
    register int ix;
    register treeNode *thisP;

    if (!topP)
        return;
    if (topP->level != 0)
        return; /* not the top node. */

    size = IntsArraySize(topP->u.cache);
    for (ix = 0; ix < size; ix++) {
        thisP = (treeNode *) ReadIntsArray(topP->u.cache, ix);
        if (!thisP)
            continue;
        thisP->info.xpos = thisP->info.ypos = 0;
        thisP->info.width = thisP->info.height = 0;
        thisP->info.title_width = thisP->info.title_height = 0;
        thisP->info.pixmap_width = thisP->info.pixmap_height = 0;
    }
    return;
}

/* Completely destroy a tree whose top node is TOPP.  Use recursion to
   destroy. (slow) */
void
DestroyTree(topP)
register treeNode *topP;
{
    register int size;
    register int ix;
    register treeNode *nodeP;

    if (!topP)
        return;
    if (topP->level != 0)
        return;
    size = IntsArraySize(topP->sub_nodes);
    for (ix = 0; ix < size; ix++)
        IDestroyTree((treeNode *) ReadIntsArray(topP->sub_nodes, ix));
    FreeTreeNode(topP, FALSE);
}

/* Completely destroy a tree whose top node is TOPP. Use iteration to
   destory. */
void
FastDestroyTree(topP)
treeNode *topP;
{
    register int size;
    register int ix;
    register treeNode *nodeP;
    register long *tmp_cache = (long *) NULL;

    if (!topP)
        return;
    if (topP->level != 0)
        return;

    size = IntsArraySize(topP->u.cache);
    if (size)
        tmp_cache = CopyIntsArray(topP->u.cache);
    else
        return;

    /* Note: the cache has `topP' in it. */
    for (ix = 0; ix < size; ix++) {
        nodeP = (treeNode *) ReadIntsArray(tmp_cache, ix);
        if (!nodeP)
            continue;
        FreeTreeNode(nodeP, TRUE);
    }
    if (tmp_cache)
        free(tmp_cache);
    return;
}

/* Only `DestroyTree ()' must call this function. */
static void
IDestroyTree(nodeP)
treeNode *nodeP;
{
    int size;
    int ix;

    if (!nodeP)
        return;
    size = IntsArraySize(nodeP->sub_nodes);
    for (ix = 0; ix < size; ix++)
        IDestroyTree((treeNode *) ReadIntsArray(nodeP->sub_nodes, ix));
    FreeTreeNode(nodeP, FALSE);
    return;
}

/* Print the tree cache of a tree whose top node is TOPP to stdout. */
void
DumpTreeCache(topP)
treeNode *topP;
{
    int ix, size;
    treeNode *nodeP;

    TLOG(("DumpCache...\n"));
    if (!topP) {
        printf("topP NULL\n");
        return;
    }
    if (topP->level != 0) {
        printf("DumpCache: 0x%lx: not a top node\n", topP);
        return;
    }
    size = IntsArraySize(topP->u.cache);
    printf("Dumping %d cache entries:\n", size);
    for (ix = 0; ix < size; ix++) {
        nodeP = (treeNode *) ReadIntsArray(topP->u.cache, ix);
        printf(NXLT("    #%d <0x%lx> <%s> parent <0x%lx> level = %d\n"),
               ix, nodeP, nodeP->info.title, nodeP->u.parent, nodeP->level);
    }
    fflush(stdout);
    return;
}

/***********************************************************************
                 List Current Directory (for testing)
 ***********************************************************************/
#if TESTING

/* Return TRUE if NAME is a directory; FALSE otherwise. */
static int
IsADirectory(name)
char *name;
{
    struct stat sbuf;

    if (!name)
        return(FALSE);
    if (lstat(name, &sbuf) != 0) {
        fprintf(stderr, "%s: %s\n", name, sys_errlist[errno]);
        return(FALSE);
    }
    if ((sbuf.st_mode & S_IFMT) == S_IFDIR)
        return(TRUE);
    return(FALSE);
}

/* List entries in the directory DIRNAME and form a tree.  Return the top
   node of the formed tree.  If RECURSE is TRUE, form a recursive
   listing. */
static treeNode *
ListDir(dirname, recurse)
char *dirname;
int recurse;
{
    treeNode *topP;
    char resolved_dirname[(3 * STRPATH) + 1];
    char dirname_title[(3 * STRPATH) + 1];
    struct stat sbuf;
    int len;
    int seen_error = FALSE;

    if (!dirname)
        return((treeNode *) NULL);

    dirname_title[0] = '\0';
    /* Top dir gets special processing.  Read trough the link to
       resolve symlink, fail on error */
    if (lstat(dirname, &sbuf) != 0) {
        sprintf(resolved_dirname, "%s: %s", dirname, sys_errlist[errno]);
        seen_error = TRUE;
    }
    if (!seen_error) {
        if ((sbuf.st_mode & S_IFMT) == S_IFLNK) {
            len = readlink(dirname, resolved_dirname, (2 * STRPATH));
            if (len == -1) {
                sprintf(resolved_dirname, "%s: %s", dirname,
                        sys_errlist[errno]);
                seen_error = TRUE;
            } else {
                resolved_dirname[len] = '\0';
/*
                sprintf(dirname_title, "%s -> %s", dirname, resolved_dirname);
*/
            }
        } else {
            strcpy(resolved_dirname, dirname);
        }
    }

    topP = CrtTreeNode(NULL);
    if (dirname_title[0])
        SetNodeTitle(topP, dirname_title);
    else
        SetNodeTitle(topP, resolved_dirname);

#if 0
    SetNodeIconPixmap(topP, NXLT("gptxtbox")); /* test */
#else
    SetNodeIconPixmap(topP, NXLT("gr_aln61")); /* test */
#endif
    if (seen_error)
        return(topP);

    if (IsADirectory(resolved_dirname))
        IListDir(resolved_dirname, topP, recurse);
    return(topP);
}

/* Only called by `ListDir ()' */
static void
IListDir(dirname, parent_node, recurse)
char *dirname; /* *must* be a directory */
treeNode *parent_node;
int recurse;
{
    treeNode *nodeP = (treeNode *) NULL;
    DIR *dirp = (DIR *) NULL;
    struct dirent *dp = (struct dirent *) NULL;
    char *p;
#if 1
    char *tmpp;
    int  known;
#endif

    dirp = opendir(dirname);
    if (!dirp) {
        fprintf(stderr, "%s: %s\n", dirname, sys_errlist[errno]);
        return;
    }

    for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
        if (!strcmp(dp->d_name, ".") || !strcmp(dp->d_name, ".."))
            continue;
        nodeP = CrtTreeNode(parent_node);
        p = (char *) malloc(strlen(dirname) + strlen(dp->d_name) + 2);
        strcpy(p, dirname);
        strcat(p, "/");
        strcat(p, dp->d_name);
        SetNodeTitle(nodeP, dp->d_name);
#if 0
        SetNodeIconPixmap(nodeP, NXLT("gptxtbox"));
#else
        known = FALSE;
        if (tmpp = strrchr(dp->d_name, '.')) {
            tmpp++;
            if (!*tmpp || !*(tmpp + 1))
                goto not_known;
            if ((*tmpp == 'a') && (*(tmpp + 1) == 'w')) {
                SetNodeIconPixmap(nodeP, NXLT("dc50x50"));
                known = TRUE;
            }
            else if ((*tmpp == 'a') && (*(tmpp + 1) == 'g')) {
                SetNodeIconPixmap(nodeP, NXLT("ge50x50"));
                known = TRUE;
            }
            else if ((*tmpp == 'a') && (*(tmpp + 1) == 's')) {
                SetNodeIconPixmap(nodeP, NXLT("ss50x50"));
                known = TRUE;
            }
            else if ((*tmpp == 'a') && (*(tmpp + 1) == 'q')) {
                SetNodeIconPixmap(nodeP, NXLT("sq50x50"));
                known = TRUE;
            }
        }
not_known:
        if (!known) {
            /* Special effects -- for testing */
            switch (dp->d_name[0]) {
            case 'a':
            case 'b':
            case 'c':
            case 'd':
            case 'e':
            case 'A':
            case 'B':
            case 'C':
            case 'D':
            case 'E':
                SetNodeIconPixmap(nodeP, NXLT("wpframe0"));
                break;
            case 'f':
            case 'g':
            case 'h':
            case 'F':
            case 'G':
            case 'H':
                SetNodeIconPixmap(nodeP, NXLT("gr_aln52"));
                break;
            case '.':
                SetNodeIconPixmap(nodeP, NXLT("gptxtbox"));
                break;
            default:
                SetNodeIconPixmap(nodeP, NXLT("gplayers"));
                break;
            }
            /*      SetNodeIconPixmap(nodeP, NXLT("gr_aln61")); */
        }
#endif
        if (recurse) {
            if (IsADirectory(p))
                IListDir(p, nodeP);
        }
        free(p);
    }
    closedir(dirp);
    return;
}
#endif

/***********************************************************************
                         Elf Layer Interface
 ***********************************************************************/

/* Return a tree node pointer from an elf BINARY_DATA datum DP. */
treeNode *
NodePtrFromBinaryData(dP)
data_binary *dP;
{
    treeNode *nodeP = (treeNode *) NULL;

    if (dP == NULL)
        return((treeNode *) NULL);
    if (dP->type != TYPE_BINARY)
        return(NULL); /* FIXME, error message here */
    memcpy(&nodeP, dP->bytes, sizeof(treeNode *));
    return(nodeP);
}

/* Return an elf array of tree-node-formats. NODEP_LIST is an
   IntsArray of tree node pointers. */
data_ptr
ElfArrayOfTreeNodes(nodep_list)
long *nodep_list;
{
    int num_nodes = IntsArraySize(nodep_list);
    data_ptr retP = (data_ptr) NULL;
    treeNode *nodeP;
    int ix;

    if (!num_nodes)
        return((data_ptr) NULL);

    retP = ELF_CREATE_ARRAY(num_nodes);
    for (ix = 0; ix < num_nodes; ix++) {
        data_ptr thisP;

        nodeP = (treeNode *) ReadIntsArray(nodep_list, ix);
        if (!nodeP)
            continue;
        thisP = ElfFormatOfTreeNode(nodeP);
        retP = WriteArray(retP, ix, thisP);
    }
    return(retP);
}

/* Return an elf array of tree-node-formats (array) of immediate
   subnodes of a node NODEP. (Returns an elf array of arrays). */
data_ptr
ElfArrayOfSubnodes(nodeP)
treeNode *nodeP;
{
    int num_subnodes;
    int ix;
    data_ptr retP = (data_ptr) NULL;

    if (!nodeP)
        return((data_ptr) NULL);

    num_subnodes = IntsArraySize(nodeP->sub_nodes);
    if (!num_subnodes)
        return((data_ptr) NULL);

    retP = ELF_CREATE_ARRAY(num_subnodes);
    for (ix = 0; ix < num_subnodes; ix++) {
        data_ptr thisNodeP;

        thisNodeP = ElfFormatOfTreeNode((treeNode *)
                                        ReadIntsArray(nodeP->sub_nodes, ix));
        retP = WriteArray(retP, ix, thisNodeP);
    }
    return(retP);
}

/* Return a tree-node-format of node NODEP.
   (Returns an elf array of size ELF_TREE_NODE_MAX) */
data_ptr
ElfFormatOfTreeNode(nodeP)
treeNode *nodeP;
{
    data_ptr retP = (data_ptr) NULL;

    if (!nodeP)
        return((data_ptr) NULL);
    retP = ELF_CREATE_ARRAY(ELF_TREE_NODE_MAX);
    retP = ElfAddStrToArray(retP, ELF_TREE_NODE_title,
                            nodeP->info.title);
    retP = ElfAddStrToArray(retP, ELF_TREE_NODE_pixmap,
                            nodeP->info.icon_pixmap);
    retP = ElfAddNumToArray(retP, ELF_TREE_NODE_value, nodeP->value);
    retP = ElfAddBinaryToArray(retP, ELF_TREE_NODE_handle,
                               sizeof(treeNode *), &nodeP);
    /* don't set ELF_TREE_NODE_subnodes information */

    return(retP);
}

/* Recursively expand 'nodeP' and create a ElfFormatOfTreeNode.
   This is like ElfFormatOfTreeNode(), but also fills in
   ELF_TREE_NODE_subnodes information */
data_ptr
TreeDataToElfArray(nodeP)
treeNode *nodeP;
{
	data_ptr nodeDP = NULL;
	data_ptr subNodesDP = NULL;
	int ix, num_subnodes;

	if (!nodeP)
		return(NULL);
	nodeDP = ElfFormatOfTreeNode(nodeP);
	num_subnodes = IntsArraySize(nodeP->sub_nodes);
	subNodesDP = ELF_CREATE_ARRAY(num_subnodes);
	for (ix = 0; ix < num_subnodes; ix++) {
		data_ptr dP = NULL;
		dP = TreeDataToElfArray((treeNode *) ReadIntsArray(nodeP->sub_nodes,
														   ix));
		subNodesDP = WriteArray(subNodesDP, ix, dP);
	}
	nodeDP = WriteArray(nodeDP, ELF_TREE_NODE_subnodes, subNodesDP);
	return(nodeDP);
}

/* Generic entry point for tree builtin functions. */
void
ElfbTreeFuncs(cmd_codep, dP1, dP2, dP3, dP4, dP5, dP6)
data_ptr cmd_codep;
data_ptr dP1, dP2, dP3, dP4, dP5, dP6;
{
    int command_code;
    data_ptr retP = (data_ptr) NULL;

    command_code = IntFromDataPtr(cmd_codep);
    switch (command_code) {
    case 0: /* invalid */
        break;
    case 1: { /* CMD#CRT_TREE_TOP */
        treeNode *topP;

        topP = CrtTreeNode(NULL);
        retP = AxMakeBinaryData(sizeof(treeNode *), &topP);
    }
        break;
    case 2: { /* CMD#CRT_TREE_NODE */
        treeNode *nodeP;
        treeNode *parentP;

        parentP = NodePtrFromBinaryData(dP1);
        if (!parentP)
            goto finish;
        nodeP = CrtTreeNode(parentP);
        retP = AxMakeBinaryData(sizeof(treeNode *), &nodeP);
    }
        break;
    case 3: { /* CMD#SET_NODE_VALUE */
        treeNode *nodeP;
        int value;

        nodeP = NodePtrFromBinaryData(dP1);
        if (!nodeP)
            goto finish;
        value = IntFromDataPtr(dP2);
        SetNodeValue(nodeP, value);
    }
        break;
    case 4: { /* CMD#SET_NODE_INFO */
        treeNode *nodeP;
        char *pixmap;
        char *title;
        int  value;

        nodeP  = NodePtrFromBinaryData(dP1);
        if (!nodeP)
            goto finish;
        pixmap = StrFromArray(dP2, ELF_TREE_NODE_pixmap);
        title  = StrFromArray(dP2, ELF_TREE_NODE_title);
        value  = IntFromArray(dP2, ELF_TREE_NODE_value);
        SetNodeTitle(nodeP, title);
        SetNodeIconPixmap(nodeP, pixmap);
        SetNodeValue(nodeP, value);
    }
        break;
    case 5: { /* CMD#GET_NODE_INFO */
        treeNode *nodeP = NodePtrFromBinaryData(dP1);
        if (!nodeP)
            goto finish;
        retP = ElfFormatOfTreeNode(nodeP);
    }
        break;
    case 6: { /* CMD#TREE_TO_ARR */
        treeNode *topP;

        topP = NodePtrFromBinaryData(dP1);
        if (!topP)
            goto finish;
		retP = TreeDataToElfArray(topP);
    }
        break;
    case 7: { /* CMD#DESTROY_TREE */
        treeNode *topP;
        int fast = FALSE;

        topP = NodePtrFromBinaryData(dP1);
        if (!topP)
            goto finish;
        if (dP2) {
            if (IntFromDataPtr(dP2) == BTRUE)
                fast = TRUE;
        }
        if (fast == TRUE)
            FastDestroyTree(topP);
        else
            DestroyTree(topP);
    }
        break;
    case 8: { /* CMD#DELETE_NODE */
        treeNode *nodeP;

        nodeP = NodePtrFromBinaryData(dP1);
        DelTreeNode(nodeP);
    }
        break;
    case 9: { /* CMD#DUMP_TREE */
        treeNode *topP;

        topP = NodePtrFromBinaryData(dP1);
        DumpTreeCache(topP);
    }
        break;
    case 10: { /* CMD#DIRECTORY_LIST */
        treeNode *topP;
        int recurse;
        char *dirname;

        dirname = StrFromDataPtr(dP1);
        recurse = IntFromDataPtr(dP2);
#if TESTING
        topP = ListDir(((dirname) ? dirname : "."), recurse);
        retP = AxMakeBinaryData(sizeof(treeNode *), &topP);
#else
        goto finish;
#endif
    }
    case 11: { /* CMD#EXPAND_TITLE */
        treeNode *nodeP;
        char *sepStr;
        char *expanded_title;

        nodeP = NodePtrFromBinaryData(dP1);
        sepStr = StrFromDataPtr(dP2);
        if (!nodeP || !sepStr)
            goto finish;
        expanded_title = ExpandNodeTitle(nodeP, sepStr);
        if (!expanded_title) {
            retP = AxMakeStrData(-1, "");
        } else {
            retP = AxMakeStrData(-1, expanded_title);
            free(expanded_title);
        }
    }
        break;
    case 12: { /* CMD#LIST_SUBNODES */
        treeNode *nodeP;

        nodeP = NodePtrFromBinaryData(dP1);
        retP = ElfArrayOfSubnodes(nodeP);
    }
        break;
    case 13: { /* CMD#EXPAND_VALUE */
        treeNode *nodeP;
        char *sepStr;
        char *expanded_value;

        nodeP = NodePtrFromBinaryData(dP1);
        sepStr = StrFromDataPtr(dP2);
        if (!nodeP || !sepStr)
            goto finish;
        expanded_value = ExpandNodeValue(nodeP, sepStr);
        if (!expanded_value) {
            retP = AxMakeStrData(-1, "");
        } else {
            retP = AxMakeStrData(-1, expanded_value);
            free(expanded_value);
        }
    }
        break;
	case 14: { /* CMD#FIND_BY_VALUE */
		treeNode *anyNodeP;
		treeNode *retNodeP;
		int value;
		anyNodeP = NodePtrFromBinaryData(dP1);
		value = IntFromDataPtr(dP2);
		retNodeP = FindNodeByValue(anyNodeP, value);
		if (retNodeP)
			retP = ElfFormatOfTreeNode(retNodeP);
		else
			retP = NULL;
		break;
	}
	case 15: { /* CMD#FIND_BY_TITLE */
		treeNode *anyNodeP;
		treeNode *retNodeP;
		char *title;
		anyNodeP = NodePtrFromBinaryData(dP1);
		title = StrFromDataPtr(dP2);
		retNodeP = FindNodeByTitle(anyNodeP, title);
		if (retNodeP)
			retP = ElfFormatOfTreeNode(retNodeP);
		else
			retP = NULL;
		break;
	}
	case 16: { /* CMD#DELETE_SUBNODES */
		treeNode *nodeP;
		nodeP = NodePtrFromBinaryData(dP1);
		DelTreeSubnodes(nodeP);
		retP = NULL;
		break;
	}
    default: /* unimpl */
        break;
    }

finish:
    ElfRetData = retP;
    return;
}


/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
