/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: wideutil.c
   Applix ASC version: 16
   Date: Mon Nov 16 13:53:40 1998
*/
#include <stdlib.h>
#include <aos.h>

#include <wide_extern.h>

#include <wideutil.p>

#if BUILD_ORIENT    /*  {  */


void InitAsianData();

extern char *ThimApplixDir;
extern char *ThimUserDir;
extern char *AxLangDir;

mbchar *OrientFace[2];
#if BUILD_KANJI
mbchar *OrientNarrowKanaFace[2];
#endif

static wchar *NoLineEndChars = (wchar *)NULL;
static wchar *NoLineStartChars = (wchar *)NULL;
static char *DefaultDialogFont = (char *)NULL;
static unsigned int MinNoLineStart = 0xFFFF;
static unsigned int MaxNoLineStart = 0;
static unsigned int MinNoLineEnd = 0xFFFF;
static unsigned int MaxNoLineEnd = 0;





/* Is it ok to break a line between these two characters.  There are */
/* certain characters which can not be left on the end of one line or */
/* on the begining of another                                         */
 
 
IsOKtoBreakOriental( int char1, int char2 )
{
    wchar *wP;
    int breakable1, breakable2;

    if ( char2 == (wchar)0 )
        /* at the end of a line */
        return(TRUE);

    breakable1 = breakable2 = FALSE;

    if ( (char1 >= MinNoLineEnd) && (char1 <= MaxNoLineEnd) )
        {
        /* We may not be able to leave the first character at the */
        /*     end of the line */
        wP = NoLineEndChars;
        while ( *wP != (wchar)NULL )
            {
            if ( *wP == char1 )
               {
               /* We can not leave character at the end of the line */
               breakable1 = TRUE;
               break;
               }
            wP++;
            }
        }

    if ( breakable1 && (char2 >= MinNoLineEnd) && (char2 <= MaxNoLineEnd) )
        {
        /* We can may be able to leave the second character at the */
        /*     end of the line */
        wP = NoLineEndChars;
        while ( *wP != (wchar)NULL )
            {
            if ( *wP == char2 )
               /* We can not leave character at the end of the line */
               {
               breakable2 = TRUE; 
               break; 
               }
            wP++;
            }
        }
    if (breakable1 && !breakable2)
        return(FALSE);

    breakable1 = breakable2 = FALSE;

    if ( (char2 >= MinNoLineStart) && (char2 <= MaxNoLineStart ) )
        {
        /* We may be able to start a line with the second char */
        wP = NoLineStartChars;
        while ( *wP != (wchar)NULL )
            {
            if ( *wP == char2 )
               {
               /* We can not start a line with the second char */
               breakable1 = TRUE;
               break;
               }
            wP++;
            }
        }

    if ( breakable1 && (char1 >= MinNoLineStart) && (char1 <= MaxNoLineStart ) )
        {
        /* We may be able to start a line with the second char */
        wP = NoLineStartChars;
        while ( *wP != (wchar)NULL )
            {
            if ( *wP == char1 )
               { 
               /* We can not start a line with the second char */
               /*return( FALSE );*/
               breakable2 = TRUE; 
               break; 
               }
            wP++;
            }
        }
    if (breakable1 && !breakable2)
        return(FALSE);

    return( TRUE );
}




/** return an integer indicating which "group" a character falls into.  */
/** For example, in Kanji, the groups might be "symbols", "Hiragana",   */
/*     "Katakana", and "Kanji"                                          */
int OrientalCharGroup( int ch )
{
    int ret_val = 0;

#if BUILD_ORIENT
    if ( ch < 0xFF )
       /* normal ascii */
       ret_val = 0;
    else if ( ch < 0x227F )
       /* a symbol */
       ret_val = 1;
    else if ( ch < 0x237F )
       /* wide ascii */
       ret_val = 2;
    else if ( ch < 0x247F )
       /* Hiragana */
       ret_val = 3;
    else if ( ch < 0x257F )
       /* Katakana */
       ret_val = 4;
    else if ( ch < 0x3000 )
       /* greek, russian, and some more symbols */
       ret_val = 5;
#if BUILD_KANJI
    else if ( (ch >= 0x8E00) && (ch <= 0x8EFF) )
       /* narrow kana */
       ret_val = 6;
#endif
    else
       /* kanji */
       ret_val = 7;
#endif
 
    return( ret_val );
}


/* Get the name of the default which char face */
/* the very first time this is called we return the opposite of later */
/* times so that the default for the dialog boxes is opposite of that */
/* for text documents,  at least in the Japanese case where we want to */
/*  to use gothic for menus and Minchou for documents    */

#ifdef getenv
#undef getenv   /* so we can link this with filters and not get "THIMgetenv" */
#endif









get_default_face()
{
    static default_face = -1;

    if ( default_face == -1 )
      {
#if BUILD_KANJI
      OrientFace[0] = (mbchar *)NXLT("Minchou");
      OrientFace[1] = (mbchar *)NXLT("Gothic");
      OrientNarrowKanaFace[0] = (mbchar *)NXLT("NarrowMinchou");
      OrientNarrowKanaFace[1] = (mbchar *)NXLT("NarrowGothic");
#endif
#if BUILD_KOREAN
      OrientFace[0] = (mbchar *)NXLT("Myeongjo");
      OrientFace[1] = (mbchar *)NXLT("Gothic");
#endif
#if BUILD_CHINA
      OrientFace[0] = (mbchar *)NXLT("Song");
      OrientFace[1] = (mbchar *)NXLT("Song");
#endif

      /* the very first time */
      if ( getenv("AX_DEFAULT_GOTHIC") != (char *)NULL )
          {
          /* force gothic default for everything */
          default_face = 1;
          }
      else
          {
          default_face = 0;     /* default is minchou */
          }
      }
    return( default_face );
}








mbchar *get_default_orient_face()
{
    return( OrientFace[get_default_face()] );
}

#if BUILD_KANJI   /*   {  */
mbchar *get_default_narrow_kana_face()
{
    return( OrientNarrowKanaFace[get_default_face()] );
}

mbchar  *serif_narrow_kana_face()
{
    get_default_face();     /* buffer init assurence call. */
    return(OrientNarrowKanaFace[0]);
}
mbchar  *sansserif_narrow_kana_face()
{
    get_default_face();     /* buffer init assurence call. */
    return(OrientNarrowKanaFace[1]);    /* Orient Gothic */
}

#endif            /* }  BUILD_KANJI  */




int IsSimpleWideFontFace( char *face_name )
{
    if ( face_name == (char *)NULL )
       /* a null is never a wide font face */
       return( FALSE );
 
    get_default_face();    /* force loading of face names */
 
    if ( (MBstrcmp( OrientFace[0], face_name ) == 0 ) ||
         (MBstrcmp( OrientFace[1], face_name ) == 0 ) )
       return( TRUE );
    else
       return( FALSE );
}
 



/** return the xlfd for the default asian font face, if there is one */

char *get_default_dialog_xlfd()
{
    InitAsianData();

    return( DefaultDialogFont );
}



/** Return whether the char is one of the set of chars which may not */
/*    start a line.  These are the wide equivalents of things like:  { */
/**   "] } )" etc.                                                   */

IsAsiaLineStartChar( int char1 )
{
    wchar *wP;
    if ( (char1 >= MinNoLineStart) && (char1 <= MaxNoLineStart ) )
        {
        /* We may be able to start a line with the second char */
        wP = NoLineStartChars;
        while ( *wP != (wchar)NULL )
            {
            if ( *wP == char1 )
               /* We can not start a line with the second char */
               return( TRUE );
            wP++;
            }
        }
    return( FALSE );
}


/************************************************************************/
/* The following function reads a configureation file "asia_dat.sp" to  */
/*    set various things for an Asian language environment.             */
/************************************************************************/

#define NO_LINE_END NXLT("no_line_end:")     /* chars illegal at end of line */
#define NO_LINE_START NXLT("no_line_start:") /* chars illegal at start of line */
#define DEFAULT_FONT NXLT("default_font:")   /* default dialog font          */


void InitAsianData()
{
    static int InitAsianData_init = FALSE;
    char filename[ 1025 ];
    mbchar buff[2000];
    FILE *ff;
    mbchar *p;
    wchar *wP;
    int size;

    if ( InitAsianData_init )
       return;

    InitAsianData_init = TRUE; 

    /* first look for the config file in the user's home dir */
    sprintf( filename, "%s/%s", ThimUserDir, "asia_dat.sp" );
    ff = fopen( filename, "r" );
    if ( ff == (FILE *)NULL )
        {
        /* try in the .../axexec/axdata/<lang> place */
        sprintf( filename, "%s/%s", AxLangDir, "asia_dat.sp" );
        ff = fopen( filename, "r" );
        if ( ff == (FILE *)NULL )
           return;
        }


    while ( fgets( (char *)buff, 200, ff) != (char *)NULL )
       {
       if ( buff[0] == '#' )
           /* comment line */
           continue;

       p = buff + MBstrlen(buff) - 1;
       if ( *p == '\n' )
           *p = '\0';

       if ( MBstrncmp( buff, NO_LINE_END, strlen(NO_LINE_END)) == 0 )
           {
           /* this is a string composed of the characters which may not be */
           /*   left at the end of a line by the formatter                 */
           p = buff + strlen(NO_LINE_END);
           size = (MBstrlen( p )+1) * sizeof(wchar);
           NoLineEndChars = (wchar *)calloc( 1,size );
           MbToWStr( p, NoLineEndChars );

           wP = NoLineEndChars;
           while ( *wP != (wchar)NULL )
             {
             if ( *wP < MinNoLineEnd )
                MinNoLineEnd = *wP;
             else if ( *wP > MaxNoLineEnd )
                MaxNoLineEnd = *wP;
             wP++;
             }
           }

       if ( MBstrncmp( buff, NO_LINE_START, strlen(NO_LINE_START)) == 0 )
           {
           /* this is a string composed of the characters which may not be */
           /*   left at the start of a line by the formatter                 */
           p = buff + strlen(NO_LINE_START);
           size = (MBstrlen( p )+1) * sizeof(wchar);
           NoLineStartChars = (wchar *)calloc( 1,size );
           MbToWStr( p, NoLineStartChars );

           wP = NoLineStartChars;
           while ( *wP != (wchar)NULL )
             {
             if ( *wP < MinNoLineStart )
                MinNoLineStart = *wP;
             else if ( *wP > MaxNoLineStart )
                MaxNoLineStart = *wP;
             wP++;
             }
           }


       if ( MBstrncmp( buff, DEFAULT_FONT, strlen(DEFAULT_FONT)) == 0 )
           {
           /* use this font as the default dialog font */
           p = buff + strlen(DEFAULT_FONT);
           size = MBstrlen( p )+1;
           DefaultDialogFont = (char *)malloc( size );

           MBstrcpy( DefaultDialogFont, p );
           }

       }

    fclose( ff );
    return;
}

mbchar  *serif_orient_face()
{
    get_default_face();     /* buffer init assurence call. */
    return(OrientFace[0]);
}
mbchar  *sansserif_orient_face()
{
    get_default_face();     /* buffer init assurence call. */
    return(OrientFace[1]);  /* Orient Gothic */
}

#endif   /* }  BUILD_ORIENT */
