/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: widgfmt.c
   Applix ASC version: 49
   Date: Wed Feb 24 21:13:22 1999
*/
#include <limits.h>
#include <basicwin.h>
#include <aos.h>
#include <ax_xm.h>
#include <xmdlged.h>
#include <elxwins.h>
#include <axconn.h>
#include <db_util.h>
#include <widgfmt.h>
#include <intsarr.h>

extern XFontStruct *AxDlgMainFont;
extern data_ptr ElfRetData;
extern int WidgetNeedsUpdated;

extern int AxIntFromDataPtr(), AxArraySize(), wMinimumGadgetSize();
extern int wElfGadgetSize(), dlgReadSizeValue();
extern int xmWidgetIsHidden();
extern int axDlgScaleX(), axDlgScaleY();
extern int xmTabctlTopInset();
extern int Atoupper();
extern mbchar *AxStrPtrFromArray(), *AxStrPtrFromDataPtr();
extern wchar *EditWinGetSelectedText();
extern data_ptr AxArrayElement(), AxMakeStrData(), AxMakeIntData();
extern data_ptr AxMakeArray(), AxAddStrToArray(), AxAddIntToArray();
extern axconnC *axNewConnScope();
extern void dlgWriteSizeValue(), wActualGadgetSize();
extern void xmRectHideOrRevealLayer();
extern void xmFormatRectTabs();
extern void SetTabLayers();
extern void EntDoPendingDelete(), EditWinDoPendingDelete();
extern void manageInsert();
extern struct edit_window *GetEditWin();
XFontStruct *lookupReferenceFont();
XFontStruct *lookupNamedFont();
extern AXWidget xmKeyFocusWidget();

static int    AXDialogUIDS = 1;      /* dbox unique-within-session id */
static axhtab AXDialogGadgetSets;

void xmtcSetTabctlInfo();
void wInitDialogGadgetSets();
void wfDeleteGadgetSet();
AXWidget xmkfHasLink();

static int gadgetSizeConnLocator();
static int isDeformable();
static int isContainer();
static int linkId();
static int xmkfTotalLinkCount(), xmkfLinkCount();

static data_ptr wfInsertNewSizes(), wfDeleteSizes();
static data_ptr wfApplyLinks(), wfBreakLinks(), wfApplySizechanges();
static data_ptr wfReadSizes(), wfNewUid(), wfAlignLinkCount();
static data_ptr wfReadElfSizes(), wfReadActualSizes();
static data_ptr wfInsertDlgSize(), wfApplyDlgSizechange();
static data_ptr xmClassNameFromGType(), wfNewDboxUid();
static data_ptr xmtcApplyTabctrlLinks();
static data_ptr xmtcApplyTabctrlLink();
static data_ptr xmtcBreakTabctrlLinks();
static data_ptr xmtcGetActiveTabctrlLayer();
static data_ptr xmtcSetActiveTabctrlLayer();
static data_ptr xmtcSetTablayers();
static data_ptr wfActualSize();
static data_ptr xmkfSetKeyFocus(), xmkfGetKeyFocusWidget();
static data_ptr xmtcReadTabctlInset();
static data_ptr xmtcDistinctLayernames();
static data_ptr xmLayoutStringWidth();
static data_ptr xmDlgClipboardOp();
static data_ptr xmReadWidgetSubAccelerators();

#if AXMS
static XFontStruct *xmInitAxsabMetrics();
#endif

static char *newGadgSize(), *doAlignLink(), *copyAlignLinkData();
static char *newDlgGadgetSet(), *newGadgetSize();

static void freeAlignLinkData();
static void tossDlgGadgetSet(), tossGadgetSize();
static void selfRelativeSizeChange(), doRowcolAlign();
static void takeChange();
static void callSizeChange();
static void wfDeleteConn();
static void containerSize();
static void initChange();
static void insertContainerEdges();
static void normalizePositions();
static void addContainerEdgeLink();
static void xmTabctlLinkContainees();

static void initKeyFocusLinks();

int
ElfbGadgetSizesFuncs( functype, dbox_handle, info, d4, d5, d6, d7 )
data_ptr functype;
data_ptr dbox_handle;      /* dbox widget handle */
data_ptr info;             /* additional functype-specific input */
data_ptr d4, d5, d6, d7;   /* unused so far */
{
	int opType;

	if (!AXDialogGadgetSets.width) {
		wInitDialogGadgetSets();
	}
	
	opType = AxIntFromDataPtr(functype);
	switch(opType) {
		
	case MAKE_DB_UID:
		ElfRetData = wfNewUid(dbox_handle, info);
		return TRUE;
		
	case MAKE_DBOX_UID:
		ElfRetData = wfNewDboxUid();
		return TRUE;
		
	case COUNT_ALIGN_LINKS:
		ElfRetData = wfAlignLinkCount(dbox_handle, info);
		return TRUE;
		
	case INSERT_NEW_SIZES:
		ElfRetData = wfInsertNewSizes(dbox_handle, info);
		return TRUE;
		
	case DELETE_SIZES:
		ElfRetData = wfDeleteSizes(dbox_handle, info);
		return TRUE;
		
	case APPLY_LINKS:
		ElfRetData = wfApplyLinks(dbox_handle, info);
		return TRUE;
		
	case BREAK_LINKS:
		ElfRetData = wfBreakLinks(dbox_handle, info);
		return TRUE;
		
	case APPLY_SIZECHANGES:
		ElfRetData = wfApplySizechanges(dbox_handle, info);
		return TRUE;
		
	case READ_SIZES:
		ElfRetData = wfReadSizes(dbox_handle, info);
		return TRUE;
		
	case READ_ELF_SIZES:
		ElfRetData = wfReadElfSizes(dbox_handle, info);
		return TRUE;
		
	case READ_ACTUAL_SIZES:
		ElfRetData = wfReadActualSizes(dbox_handle, info);
		return TRUE;
		
	case INSERT_DLG_SIZE:
		ElfRetData = wfInsertDlgSize(dbox_handle, info);
		return TRUE;
		
	case APPLY_DLG_SIZECHANGE:
		ElfRetData = wfApplyDlgSizechange(dbox_handle, info);
		return TRUE;

	case CLASSNAME_FROM_GTYPE:
		ElfRetData = xmClassNameFromGType(info);
		return TRUE;
		
	case APPLY_TABCTRL_LINKS:
		ElfRetData = xmtcApplyTabctrlLinks(info);
		return TRUE;
		
	case ADD_TABCTRL_LINK:
		ElfRetData = xmtcApplyTabctrlLink(dbox_handle, info);
		return TRUE;
		
	case BREAK_TABCTRL_LINKS:
		ElfRetData = xmtcBreakTabctrlLinks(info);
		return TRUE;
		
	case GET_ACTIVE_TABCTRL_LAYER:
		ElfRetData = xmtcGetActiveTabctrlLayer(info);
		return TRUE;
		
	case SET_ACTIVE_TABCTRL_LAYER:
		ElfRetData = xmtcSetActiveTabctrlLayer(info);
		return TRUE;
		
	case SET_TABLAYERS:
		ElfRetData = xmtcSetTablayers(info);
		return TRUE;
		
	case DB_CTRL_GET_PIXEL_SIZE:
		ElfRetData = wfActualSize(info);
		return TRUE;

	case APPLY_KEYFOCUS_LINKS:
		initKeyFocusLinks(info, dbox_handle);
		/* TODO */
		return TRUE;

	case SET_KEYFOCUS:
		ElfRetData = xmkfSetKeyFocus(info);
		return TRUE;
		
	case GET_KEYFOCUS:
		ElfRetData = xmkfGetKeyFocusWidget(info);
		return TRUE;
		
	case READ_TABCTL_INSET:
		ElfRetData = xmtcReadTabctlInset(info);
		return TRUE;
		
	case DISTINCT_LAYERNAMES:
		ElfRetData = xmtcDistinctLayernames(info);
		return TRUE;
		
	case LAYOUT_STR_WIDTH:
		ElfRetData = xmLayoutStringWidth(info);
		return TRUE;
		
	case CUTCOPYPASTE:
		ElfRetData = xmDlgClipboardOp(info);
		return TRUE;
		
	case READ_SUB_ACCELERATORS:
		ElfRetData = xmReadWidgetSubAccelerators(info);
		return TRUE;
		
	}
}

/* utility which gens uid for a particular gadget
 * (these uids must be nonzero and positive).
 * this db_uid is unique within its dialog box.
 */
static data_ptr       /* return dlg-unique uid, or return 0 on failure */
wfNewUid(not_used, info)
data_ptr not_used;
data_ptr info;
{
	int ix, jx, kx, asize, maxID, thisID;
	data_ptr garray, gadg, dbox;
	
	if (dbox = AxArrayElement(info, 0)) {
	
		for (ix = 0, maxID = 0; ix < DLG_NUM_GADGETS; ix++) {
			garray = AxArrayElement(dbox, ix);
			asize = AxArraySize(garray);
			for (jx = 0; jx < asize; jx++) {
				gadg = AxArrayElement(garray, jx);
				thisID = AxIntFromArray(gadg, DBDEXuid);
				if (thisID > maxID) {
					maxID = thisID;
				}
			}
		}
		
		maxID++;      /* tentative new id */
		
		if (maxID >= MAX_GADGET_UID) {
			int uniq = TRUE;
			int found = FALSE;
			
			/* wrap the id around to 1st legal value of 1, then
			 * find the first unused id val in the box.  (give up at 10000)
			 */
			for (jx = 0, maxID = 1; jx < 10000; jx++) {
				for (ix = 0; ix < DLG_NUM_GADGETS; ix++) {
					garray = AxArrayElement(dbox, ix);
					asize = AxArraySize(garray);
					for (kx = 0; kx < asize; kx++) {
						gadg = AxArrayElement(garray, kx);
						thisID = AxIntFromArray(gadg, DBDEXuid);
						if (thisID == maxID) {
							uniq = FALSE;
							maxID++;
							break;
						}
					}
				}
				if (uniq) {
					found = TRUE;
					break;
				}
				else {
					maxID++;
				}
			}
			if (!found) {
				maxID = 0;
			}
		}
	}
	return (AxMakeIntData(maxID));
}

/* utility which gens uid for a particular dialog box.
 * (these uids must be nonzero and positive).
 * this dbox_uid is unique within an axmain session.
 */
static data_ptr
wfNewDboxUid()
{
	int new_db_uid = AXDialogUIDS++;
	return (AxMakeIntData(new_db_uid));
}

/* uncovered - intended for link-deletion elf code, so that the
 * DBDEX#uid gadget_member (a clue to exert oneself looking for
 * link info) can be removed from a gadget if the last align_link
 * is removed
 */
static data_ptr
wfAlignLinkCount(dbox_handle, info)
data_ptr dbox_handle;
data_ptr info;
{
	int ix, jx, linkCount, uid, dbox_uid;
	axhtab *tabl;
	hitem *list;
	gadget_size *gSize = (gadget_size *) NULL;
	gadget_set *gadgSet = (gadget_set *) NULL;
	
	if (!dbox_handle || !info)
		return (AxMakeIntData(0));
	
	uid = AxIntFromDataPtr(info);
	if (!uid)
		return (AxMakeIntData(0));

	/* lookup gadget set for this dialog
	 */
	dbox_uid = AxIntFromDataPtr(dbox_handle);
	if (!axhTestItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets)) {
		return (AxMakeIntData(0));  /* no gadget set for this dbox */
	}
	axhGetItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets,
				(char *) 0, (char **) &gadgSet);
	
	/* lookup gadget_size instance for gadget 'uid'
	 */
	axhGetItemI((void *) &uid, sizeof(int), &gadgSet->gadgets,
				(char *) NULL, (char **) &gSize);
	
	if (!gSize || !gSize->conn)
		return (AxMakeIntData(0));
	
	/* count links of all types
	 */
	for (ix = 0, linkCount = 0; ix < gSize->conn->numLinkTypes; ix++) {
		tabl = &gSize->conn->linkTables[ix];
		for (jx = 0; jx < tabl->width; jx++) {
			list = tabl->table[jx];
			while(list) {
				if (list->value && list->name) {
					linkCount++;
				}
				list = list->next;
			}
		}
	}
	if (isContainer(gSize)) {
		/* include link counts of edges
		 */
		int ex;
		gadget_size *edge;
		for (ex = 0; ex < 4; ex++) {
			if (gSize->edges[ex]) {
				edge = gSize->edges[ex];
				if (edge->conn) {
					for (ix = 0; ix < edge->conn->numLinkTypes; ix++) {
						tabl = &edge->conn->linkTables[ix];
						for (jx = 0; jx < tabl->width; jx++) {
							list = tabl->table[jx];
							while(list) {
								if (list->value && list->name) {
									linkCount++;
								}
								list = list->next;
							}
						}
					}
				}
			}
		}
	}
	return (AxMakeIntData(linkCount));
}

/*
 * : wfInsertNewSizes - create gadget_size instances for new widgets
 *
 * info is in the format: <<dbox rev><dbox reference font><dbox expand bool>
 *                        <sizeArray>>
 * sizeArray is in this format: <<g1 uid><g1 gtype><g1 "size" atts array>>
 *                              <<g2 uid><g2 gtype><g2 "size" atts array>> ..
 */
static data_ptr
wfInsertNewSizes(dbox_handle, info)
data_ptr dbox_handle;
data_ptr info;
{
	int numGadgets, gx, gType, gadgId, dboxRev, dboxNoExpand;
	int fixedSize, dbox_uid, different;
	data_ptr dtmp, sizeArray, sizeAtts;
	new_gsize newGadg;
	gadget_size *gsize = (gadget_size *) 0;
	gadget_set *gadgSet;
	XFontStruct *refFont;
	
	dtmp = AxArrayElement(info, 0);
	dboxRev = AxIntFromArray(dtmp, 0);
	refFont = lookupReferenceFont(AxArrayElement(dtmp, 1), &different);
	dboxNoExpand = AxIntFromArray(dtmp, 2);
	sizeArray = AxArrayElement(info, 1);
	
	/* lookup gadget set for this dialog
	 */
	dbox_uid = AxIntFromDataPtr(dbox_handle);
	if (!axhTestItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets)) {
		return ((data_ptr) NULL);  /* no gadget set for this dbox */
	}
	axhGetItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets,
				(char *) 0, (char **) &gadgSet);
	
	memset((char *) &newGadg, 0, sizeof(new_gsize));
	newGadg.scope = gadgSet->scope;
	
	numGadgets = AxArraySize(sizeArray);
	for (gx = 0; gx < numGadgets; gx++) {
		
		dtmp = AxArrayElement(sizeArray, gx);
		gadgId = AxIntFromArray(dtmp, 0);
		gType = AxIntFromArray(dtmp, 1);
		sizeAtts = AxArrayElement(dtmp, 2);
		
		wActualGadgetSize(dboxRev, refFont, dboxNoExpand,
						  gType, sizeAtts, &newGadg.sizeData.proposedsize);

		fixedSize = !wMinimumGadgetSize(dboxRev, refFont, dboxNoExpand,
										gType, sizeAtts,
										&newGadg.sizeData.minsize);
		if (fixedSize) {
			newGadg.sizeData.fixedWidth = newGadg.sizeData.fixedHeight = TRUE;
		}
		
		/* The current and proposed sizes for the new instance are
		 * set equal, which will have the effect of 'quietly' inserting.
		 * That is, no link call will be stimulated because the instance
		 * will look up-to-date.  Not really what we want, but easier.
		 * (assumes that instances are installed already meeting whatever
		 * layout constraints are required.. yeah right!)
		 */
		newGadg.sizeData.currentsize = newGadg.sizeData.proposedsize;
		
		newGadg.sizeData.gadgetType = gType;
		newGadg.sizeData.gadgetMinorType = gType;
		newGadg.sizeData.gadgId = gadgId;
		axhGetItemI((void *) &gadgId, sizeof(int), &gadgSet->gadgets,
					(char *) &newGadg, (char **) &gsize);  /* store */
		
		if (gsize && isContainer(&newGadg.sizeData)) {
			insertContainerEdges(gadgSet, &newGadg, gsize);
		}
	}
	return((data_ptr) NULL);
}

/* uncovered
 * : wfDeleteSizes - delete gadget_size instances
 *
 * delArray is in this format: <g1 uid> <g2 uid> ..
 */
static data_ptr
wfDeleteSizes(dbox_handle, delArray)
data_ptr dbox_handle;         /* in */
data_ptr delArray;
{
	int ix, numGadgets, gadgId;
	int dbox_uid;
	gadget_set *gadgSet = (gadget_set *) NULL;
	gadget_size *tsize = (gadget_size *) NULL;
	
	/* lookup gadget set for this dialog
	 */
	dbox_uid = AxIntFromDataPtr(dbox_handle);
	if (!axhTestItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets)) {
		return ((data_ptr) NULL);  /* no gadget set for this dbox */
	}
	axhGetItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets,
				(char *) 0, (char **) &gadgSet);
	
	if (!delArray) {
		hitem *list;
		
		/* no delArray; delete the entire set
		 */
		for (ix = 0; ix < gadgSet->gadgets.width; ix++) {
			list = gadgSet->gadgets.table[ix];
			while(list) {
				tsize = (gadget_size *) list->value;
				if (tsize && isContainer(tsize)) {
					/* delete edges too
					 */
					if (tsize->edges[0] && tsize->edges[0]->gadgId) {
						axhTossItemI((void *) &tsize->edges[0]->gadgId,
									 sizeof(int), &gadgSet->gadgets,
									 (char *) gadgSet->scope);
					}
					if (tsize->edges[1] && tsize->edges[1]->gadgId) {
						axhTossItemI((void *) &tsize->edges[1]->gadgId,
									 sizeof(int), &gadgSet->gadgets,
									 (char *) gadgSet->scope);
					}
					if (tsize->edges[2] && tsize->edges[2]->gadgId) {
						axhTossItemI((void *) &tsize->edges[2]->gadgId,
									 sizeof(int), &gadgSet->gadgets,
									 (char *) gadgSet->scope);
					}
					if (tsize->edges[3] && tsize->edges[3]->gadgId) {
						axhTossItemI((void *) &tsize->edges[3]->gadgId,
									 sizeof(int), &gadgSet->gadgets, 
									 (char *) gadgSet->scope);
					}
				}
				axhTossItemI((void *) &tsize->gadgId, sizeof(int),
							 &gadgSet->gadgets, (char *) gadgSet->scope);
				list = list->next;
			}
		}
		return ((data_ptr) NULL);
	}
	
	numGadgets = AxArraySize(delArray);
	for (ix = 0; ix < numGadgets; ix++) {
		gadgId = AxIntFromArray(delArray, ix);
		tsize = (gadget_size *) NULL;
		axhGetItemI((void *) &gadgId, sizeof(int),
					&gadgSet->gadgets, (char *) NULL, (char **) &tsize);
		if (tsize && isContainer(tsize)) {
			/* delete edges too
			 */
			if (tsize->edges[0] && tsize->edges[0]->gadgId) {
				axhTossItemI((void *) &tsize->edges[0]->gadgId, sizeof(int),
							 &gadgSet->gadgets, (char *) gadgSet->scope);
			}
			if (tsize->edges[1] && tsize->edges[1]->gadgId) {
				axhTossItemI((void *) &tsize->edges[1]->gadgId, sizeof(int),
							 &gadgSet->gadgets, (char *) gadgSet->scope);
			}
			if (tsize->edges[2] && tsize->edges[2]->gadgId) {
				axhTossItemI((void *) &tsize->edges[2]->gadgId, sizeof(int),
							 &gadgSet->gadgets, (char *) gadgSet->scope);
			}
			if (tsize->edges[3] && tsize->edges[3]->gadgId) {
				axhTossItemI((void *) &tsize->edges[3]->gadgId, sizeof(int),
							 &gadgSet->gadgets, (char *) gadgSet->scope);
			}
		}
		axhTossItemI((void *) &gadgId, sizeof(int), &gadgSet->gadgets,
					 (char *) gadgSet->scope);
	}
	return((data_ptr) NULL);
}

/*
 * : wfApplyLinks - declare links between gadget_size instances
 *
 * linksArray is in format:  <link1 data> <link2 data>, ...
 *
 * <link data> is in format: <type>
 *                           <fromname>
 *                           <toname(s)>
 *                           <min distance>
 *                           <max distance>
 *                           <toedge>
 *                           <fromedge>
 */
static data_ptr
wfApplyLinks(dbox_handle, linksArray)
data_ptr dbox_handle;
data_ptr linksArray;
{
	int numLinks, ix, jx, numToNames;
	int dbox_uid;
	int fromname, toname;
	gadget_set *gadgSet;
	gadget_size *gSize, *dlgSize;
	data_ptr thisLink, tonameList, tmp;
	align_link_data linkData;
	linkname linkFrom, linkTo;
	
	if (!linksArray) {
		return ((data_ptr) NULL);
	}
	
	numLinks = AxArraySize(linksArray);
	
	/* lookup gadget set for this dialog
	 */
	dbox_uid = AxIntFromDataPtr(dbox_handle);
	if (!axhTestItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets)) {
		return ((data_ptr) NULL);  /* no gadget set for this dbox */
	}
	axhGetItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets,
				(char *) 0, (char **) &gadgSet);
	
	dlgSize = (gadget_size *) NULL;
	for (ix = 0; ix < numLinks; ix++) {
		/* read the data for this link
		 */
		thisLink = AxArrayElement(linksArray, ix);
		linkData.type = AxIntFromArray(thisLink, 0);
		fromname = AxIntFromArray(thisLink, 1);
		tmp = AxArrayElement(thisLink, 2);
		if (AxIsArray(tmp)) {
			/* a list of 'to' names
			 */
			tonameList = tmp;
			toname = 0;
		}
		else {
			/* single 'to' name
			 */
			tonameList = (data_ptr) NULL;
			toname = AxIntFromDataPtr(tmp);
		}
		linkData.minDistance = AxIntFromArray(thisLink, 3);
		linkData.maxDistance = AxIntFromArray(thisLink, 4);
		linkData.toEdge = AxIntFromArray(thisLink, 5);
		linkData.fromEdge = AxIntFromArray(thisLink, 6);
		
		/* declare the link(s)
		 */
		if (linkData.type == FIXED_WIDTH ||
			linkData.type == FIXED_HEIGHT) {
			
			/* this really isn't a link, but a hint to mark
			 * the gadget_size instance named by "fromname"
			 * as either fixed width or height during layout adjust
			 */
			axhGetItemI((void *) &fromname, sizeof(int),
						&gadgSet->gadgets, (char *) NULL, (char **) &gSize);
			
			if (linkData.type == FIXED_WIDTH) {
				gSize->fixedWidth = TRUE;
			}
			else if (linkData.type == FIXED_HEIGHT) {
				gSize->fixedHeight = TRUE;
			}
		}
		else {
			/* real align link data
			 */
			linkFrom.type = linkData.type;
			linkFrom.toedge = linkData.fromEdge;
			linkFrom.toid = linkId(gadgSet, &linkFrom, fromname);
			if (tonameList) {
				numToNames = AxArraySize(tonameList);
				for (jx = 0; jx < numToNames; jx++) {
					toname = AxIntFromArray(tonameList, jx);
					linkTo.type = linkData.type;
					linkTo.toedge = linkData.toEdge;
					linkTo.toid = linkId(gadgSet, &linkTo, toname);
					gadgSet->scope->link(gadgSet->scope, ALIGN_LINK_TYPE,
										 &linkFrom, &linkTo, sizeof(linkname),
										 (char *) &linkData, TRUE);
				}
			}
			else {
				linkTo.type = linkData.type;
				linkTo.toedge = linkData.toEdge;
				linkTo.toid = linkId(gadgSet, &linkTo, toname);
				gadgSet->scope->link(gadgSet->scope, ALIGN_LINK_TYPE,
									 &linkFrom, &linkTo, sizeof(linkname),
									 (char *) &linkData, TRUE);
			}
		}
	}
	return ((data_ptr) NULL);
}

/* uncovered, not to mentioned unfinished
 * : wfBreakLinks - delete links between gadget_size instances
 * 
 * deleteArray is in format: <link1 data><link2 data>,...
 *
 * <link data> is in format: <fromname><toname(s)>
 *
 * %%%%%%%%%% this is insufficient as coded:
 *            1) need to pass further info (probably same as passed to wfApply)
 *               in order to select which of several possible links
 *               exist between from and to; this probably affects axconn.c
 *               dlink definition too (pass void *data)
 */
static data_ptr
wfBreakLinks(dbox_handle, deleteArray)
data_ptr dbox_handle;
data_ptr deleteArray;
{
	int numLinks, ix, jx, numToNames;
	int dbox_uid;
	int fromname, toname;
	gadget_set *gadgSet;
	data_ptr thisLink, tonameList, tmp;
	linkname linkFrom, linkTo;
	
	if (!deleteArray)
		return ((data_ptr) NULL);
	
	numLinks = AxArraySize(deleteArray);
	
	/* lookup gadget set for this dialog
	 */
	dbox_uid = AxIntFromDataPtr(dbox_handle);
	if (!axhTestItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets)) {
		return ((data_ptr) NULL);  /* no gadget set for this dbox */
	}
	axhGetItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets,
				(char *) 0, (char **) &gadgSet);
	
	for (ix = 0; ix < numLinks; ix++) {
		/* read the data for this link
		 */
		thisLink = AxArrayElement(deleteArray, ix);
		fromname = AxIntFromArray(thisLink, 0);
		tmp = AxArrayElement(thisLink, 1);
		if (AxIsArray(tmp)) {
			/* a list of 'to' names
			 */
			tonameList = tmp;
			toname = 0;
		}
		else {
			/* single 'to' name
			 */
			tonameList = (data_ptr) NULL;
			toname = AxIntFromDataPtr(tmp);
		}

#if 0 /* %%%%%%%%%%%% need to get the identifying info type, to/from edge */
		/* delete the link(s)
		 */
		linkFrom.type = linkData.type;
		linkFrom.toname = fromname;
		linkFrom.toedge = linkData.fromEdge;
		if (tonameList) {
			numToNames = AxArraySize(tonameList);
			for (jx = 0; jx < numToNames; jx++) {
				toname = AxIntFromArray(tonameList, jx);
				linkTo.type = linkData.type;
				linkTo.toname = toname;
				linkTo.toedge = linkData.toEdge;
				gadgSet->scope->dlink(gadgSet->scope, ALIGN_LINK_TYPE,
									  &linkFrom, &linkTo,
									  sizeof(linkname), TRUE);
			}
		}
		else {
			linkTo.type = linkData.type;
			linkTo.toname = toname;
			linkTo.toedge = linkData.toEdge;
			gadgSet->scope->dlink(gadgSet->scope, ALIGN_LINK_TYPE,
								  &linkFrom, &linkTo, sizeof(linkname), TRUE);
		}
#endif
	}
	return ((data_ptr) NULL);
}

/* : wfApplySizechanges - resolve changes in size among
 * the gadget_size set.  If gadgList is present, apply
 * its size information to the gadget_size->currentsize
 * first.
 *
 * info is in the format: <<dbox rev><dbox reference font><dbox expand bool>>
 *                        <<sizeArray>>
 *
 * sizeArray is in this format: <<g1 uid><g1 gtype><g1 "size" atts array>>
 *                              <<g2 uid><g2 gtype><g2 "size" atts array>> ..
 */
static data_ptr
wfApplySizechanges(dbox_handle, info)
data_ptr dbox_handle;
data_ptr info;
{
	int dbox_uid;
	gadget_set *gadgSet;
	gadget_size *gsize;
	data_ptr dtmp, sizeArray, sizeAtts;
	int dboxRev, dboxNoExpand, numGadgets, gx;
	int gType, gadgId, different;
	r32 newSize;
	XFontStruct *refFont;
	
	dtmp = AxArrayElement(info, 0);
	dboxRev = AxIntFromArray(dtmp, 0);
	refFont = lookupReferenceFont(AxArrayElement(dtmp, 1), &different);
	dboxNoExpand = AxIntFromArray(dtmp, 2);
	sizeArray = AxArrayElement(info, 1);
	
	/* lookup gadget set for this dialog
	 */
	dbox_uid = AxIntFromDataPtr(dbox_handle);
	if (!axhTestItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets)) {
		return ((data_ptr) NULL);  /* no gadget set for this dbox */
	}
	axhGetItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets,
				(char *) 0, (char **) &gadgSet);
	
	numGadgets = AxArraySize(sizeArray);
	for (gx = 0; gx < numGadgets; gx++) {
		
		dtmp = AxArrayElement(sizeArray, gx);
		gadgId = AxIntFromArray(dtmp, 0);
		gType = AxIntFromArray(dtmp, 1);
		sizeAtts = AxArrayElement(dtmp, 2);
		
		axhGetItemI((void *) &gadgId, sizeof(int), &gadgSet->gadgets,
					(char *) NULL, (char **) &gsize); /* get gadgsize */
			
		wActualGadgetSize(dboxRev, refFont, dboxNoExpand,
						  gType, sizeAtts, &gsize->proposedsize); /* update */
			
		callSizeChange(gadgSet, gsize); /* propagate */
	}
	normalizePositions(gadgSet);
	return((data_ptr) NULL);
}

/*
 * : wfReadSizes - read the dbox array's gadgets, and return an
 * array describing the new sizes of those that have changed.
 * The return numbers are in as-realized (screenpix, not elf units)
 * numbers, so caller must convert result.
 */
static data_ptr
wfReadSizes(dbox_handle, notUsed)
data_ptr dbox_handle;             /* in */
data_ptr notUsed;
{
	int dbox_uid;
	gadget_set *gadgSet;
	gadget_size *gsize;
	data_ptr sizeList, dtmp;
	hitem *list;
	int any, ix, gx, cx;
  
	if (!dbox_handle)
		return ((data_ptr) NULL);
  
	/* lookup gadget set for this dialog
	 */
	dbox_uid = AxIntFromDataPtr(dbox_handle);
	if (!axhTestItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets)) {
		return ((data_ptr) NULL);  /* no gadget set for this dbox */
	}
	axhGetItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets,
				(char *) 0, (char **) &gadgSet);
	
	sizeList = AxMakeArray(0);
	for (ix = 0, cx = 0, any = FALSE; ix < gadgSet->gadgets.width; ix++) {
		list = gadgSet->gadgets.table[ix];
		while(list) {
			gsize = (gadget_size *) list->value;
			if (gsize->updateElf && !gsize->containerTop) {
				if (isContainer(gsize)) {
					containerSize(gsize); /* update from edges */
				}
				dtmp = AxMakeArray(4);
				dtmp = AxAddIntToArray(dtmp, 0, gsize->gadgId);
				dtmp = AxAddIntToArray(dtmp, 1, gsize->currentsize.origin.x);
				dtmp = AxAddIntToArray(dtmp, 2, gsize->currentsize.origin.y);
				dtmp = AxAddIntToArray(dtmp, 3, gsize->currentsize.width);
				dtmp = AxAddIntToArray(dtmp, 4, gsize->currentsize.height);
				sizeList = AxAddArrayToArray(sizeList, cx, dtmp);
				cx++;
				any = TRUE;
				gsize->updateElf = FALSE;
			}
			list = list->next;
		}
	}
	if (!any) {
		AxFreeData(sizeList);
		sizeList = (data_ptr) NULL;
	}
	return (sizeList);
}

/*
 * : wfReadElfSizes - convert passed as-realized sizes to elf-unit sizes
 * return array in form: <<g1 uid><g1 x><g1 y><g1 wid><g1 ht>><<g2....
 *
 * info is in the format: <<dbox rev><dbox reference font><dbox expand bool>
 *                        <sizeArray>
 * sizeArray is in this format: <<g1 uid><g1 gtype><g1 "size" atts array>
 *                               <g1 as-realized size>>
 *                              <<g2 uid><g2 gtype><g2 "size" atts array>
 *                               <g2 as-realized size>>....
 * as-realized size is: <x><y><wid><ht>
 */
static data_ptr
wfReadElfSizes(dbox_handle, info)
data_ptr dbox_handle;
data_ptr info;
{
	data_ptr dtmp, refFont, sizeArray, sizeAtts, dsize;
	data_ptr retSizes, rtmp;
	int dboxRev, numGadgets, gx, gType, gadgId;
	int dboxNoExpand, itmp;
	r32 csize, elfSize;
  
	dtmp = AxArrayElement(info, 0);
	dboxRev = AxIntFromArray(dtmp, 0);
	refFont = AxArrayElement(dtmp, 1);
	dboxNoExpand = AxIntFromArray(dtmp, 2);
	sizeArray = AxArrayElement(info, 1);
  
	retSizes = (data_ptr) NULL;
	numGadgets = AxArraySize(sizeArray);
	if (numGadgets) {
		retSizes = AxMakeArray(0);
		for (gx = 0; gx < numGadgets; gx++) {
			dtmp = AxArrayElement(sizeArray, gx);
			gadgId = AxIntFromArray(dtmp, 0);
			gType = AxIntFromArray(dtmp, 1);
			sizeAtts = AxArrayElement(dtmp, 2);
			dsize = AxArrayElement(dtmp, 3);
			
			csize.origin.x = AxIntFromArray(dsize, 0);
			csize.origin.y = AxIntFromArray(dsize, 1);
			itmp = AxIntFromArray(dsize, 2);
			if (itmp < 0) {
				itmp = 0;
			}
			csize.width = itmp;
			itmp = AxIntFromArray(dsize, 3);
			if (itmp < 0) {
				itmp = 0;
			}
			csize.height = itmp;
			
			wElfGadgetSize(dboxRev, refFont, dboxNoExpand,
						   gType, sizeAtts, &csize, &elfSize);
			
			rtmp = AxMakeArray(5);
			rtmp = AxAddIntToArray(rtmp, 0, gadgId);
			rtmp = AxAddIntToArray(rtmp, 1, elfSize.origin.x);
			rtmp = AxAddIntToArray(rtmp, 2, elfSize.origin.y);
			rtmp = AxAddIntToArray(rtmp, 3, elfSize.width);
			rtmp = AxAddIntToArray(rtmp, 4, elfSize.height);
			retSizes = AxAddArrayToArray(retSizes, gx, rtmp);
		}
	}
	return (retSizes);
}

/* uncovered
 * : wfReadActualSizes - convert passed elf sizes to c-kit sizes
 * return array in form: <<g1 uid><g1 x><g1 y><g1 wid><g1 ht>><<g2....
 *
 * info is in the format: <<dbox rev><dbox reference font><dbox expand bool>
 *                        <sizeArray>
 * sizeArray is in this format: <<g1 uid><g1 gtype><g1 "size" atts array>>
 *                              <<g2 uid><g2 gtype><g2 "size" atts array>>....
 */
static data_ptr
wfReadActualSizes(dbox_handle, info)
data_ptr dbox_handle;
data_ptr info;
{
	data_ptr dtmp, sizeArray, sizeAtts;
	data_ptr retSizes, rtmp;
	int dboxRev, numGadgets, gx, gType, gadgId;
	int dboxNoExpand, different;
	r32 csize;
	XFontStruct *refFont;
	
	dtmp = AxArrayElement(info, 0);
	dboxRev = AxIntFromArray(dtmp, 0);
	refFont = lookupReferenceFont(AxArrayElement(dtmp, 1), &different);
	dboxNoExpand = AxIntFromArray(dtmp, 2);
	sizeArray = AxArrayElement(info, 1);

	retSizes = (data_ptr) NULL;
	numGadgets = AxArraySize(sizeArray);
	if (numGadgets) {
		retSizes = AxMakeArray(0);
		for (gx = 0; gx < numGadgets; gx++) {
			dtmp = AxArrayElement(sizeArray, gx);
			gadgId = AxIntFromArray(dtmp, 0);
			gType = AxIntFromArray(dtmp, 1);
			sizeAtts = AxArrayElement(dtmp, 2);
			
			wActualGadgetSize(dboxRev, refFont, dboxNoExpand,
							  gType, sizeAtts, &csize);
			
			rtmp = AxMakeArray(5);
			rtmp = AxAddIntToArray(rtmp, 0, gadgId);
			rtmp = AxAddIntToArray(rtmp, 1, csize.origin.x);
			rtmp = AxAddIntToArray(rtmp, 2, csize.origin.y);
			rtmp = AxAddIntToArray(rtmp, 3, csize.width);
			rtmp = AxAddIntToArray(rtmp, 4, csize.height);
			retSizes = AxAddArrayToArray(retSizes, gx, rtmp);
		}
	}
	return (retSizes);
}

static data_ptr
wfInsertDlgSize(dbox_handle, info)
data_ptr dbox_handle;     /* in */
data_ptr info;            /* in, format: <xpos><ypos> */
{
	int dbox_uid;
	gadget_set *gadgSet;
	gadget_size *gs;
	new_gsize ns;
	int height, width;
	
	/* lookup gadget set for this dialog
	 */
	dbox_uid = AxIntFromDataPtr(dbox_handle);
	if (!axhTestItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets)) {
		return ((data_ptr) NULL);  /* no gadget set for this dbox */
	}
	axhGetItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets,
				(char *) 0, (char **) &gadgSet);
	
	width = AxIntFromArray(info, 0);
	if (width < 0) {
		width = 0;
	}
	height = AxIntFromArray(info, 1);
	if (height < 0) {
		height = 0;
	}
	
	memset((void *) &ns, 0, sizeof(new_gsize));
	ns.scope = gadgSet->scope;
	ns.sizeData.gadgId = DBOX_GSIZE_ID;
	ns.sizeData.gadgetType = DBOX_GTYPE;
	ns.sizeData.minsize.width = ns.sizeData.minsize.height = 10;
	ns.sizeData.proposedsize.width = width;
	ns.sizeData.proposedsize.height = height;
	ns.sizeData.currentsize = ns.sizeData.proposedsize;
	
	axhGetItemI((void *) &ns.sizeData.gadgId, sizeof(int), &gadgSet->gadgets,
				(char *) &ns, (char **) &gs); /* store */
	if (gs) {
		insertContainerEdges(gadgSet, &ns, gs);
	}
	
	return((data_ptr) NULL);
}

static data_ptr
wfApplyDlgSizechange(dbox_handle, info)
data_ptr dbox_handle;     /* in */
data_ptr info;            /* in, format: <xpos><ypos> */
{
	int dbox_uid;
	gadget_set *gadgSet;
	gadget_size *dlgSize;
	int gid, itmp;
	
	/* lookup gadget set for this dialog
	 */
	dbox_uid = AxIntFromDataPtr(dbox_handle);
	if (!axhTestItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets)) {
		return ((data_ptr) NULL);  /* no gadget set for this dbox */
	}
	axhGetItemI((void *) &dbox_uid, sizeof(int), &AXDialogGadgetSets,
				(char *) 0, (char **) &gadgSet);
	
	gid = DBOX_GSIZE_ID;
	axhGetItemI((void *) &gid, sizeof(int), &gadgSet->gadgets,
				(char *) NULL, (char **) &dlgSize);
	
	itmp = AxIntFromArray(info, 0);
	if (itmp < 0) {
		itmp = 0;
	}
	dlgSize->proposedsize.width = itmp;
	itmp = AxIntFromArray(info, 1);
	if (itmp < 0) {
		itmp = 0;
	}
	dlgSize->proposedsize.height = itmp;
	
	callSizeChange(gadgSet, dlgSize);
	normalizePositions(gadgSet);
	
	return ((data_ptr) NULL);
}

void
wfDeleteGadgetSet(dbox_uid)
int dbox_uid;
{
	if (AXDialogGadgetSets.width) {
		axhTossItemI((void *) &dbox_uid, sizeof(int),
					 &AXDialogGadgetSets, (char *) 0);
	}
}

/* : xmClassNameFromGType - for the passed dialog_box_ gtype
 * (e.g., DLG#TABLE), return the Builder classname for the
 * corresponding direct descendent of ControlClass (e.g., "TableClass")
 */
static data_ptr
xmClassNameFromGType(gP)
data_ptr gP;
{
	mbchar *cname = (mbchar *) NULL;
	int gtype = AxIntFromDataPtr(gP);
	
	switch (gtype) {
		
	case DLG_RADIO_BOX:
		cname = NXLT("RadioBoxClass");
		break;
		
	case DLG_TOGGLE_BUTTON:
		cname = NXLT("ToggleButtonClass");
		break;
		
	case DLG_CHOICE:
		cname = NXLT("OptionMenuClass");
		break;
		
	case DLG_VERB_BUTTON:
		cname = NXLT("ButtonClass");
		break;
		
	case DLG_ENTER:
		cname = NXLT("EntryFieldClass");
		break;
		
	case DLG_TEXT:
		cname = NXLT("LabelClass");
		break;
		
	case DLG_ICON:
		cname = NXLT("IconClass");
		break;
		
	case DLG_LIST_BOX:
		cname = NXLT("ListBoxClass");
		break;
		
	case DLG_RECT:
		cname = NXLT("PanelClass");
		break;
		
	case DLG_EDIT_BOX:
		cname = NXLT("EditBoxClass");
		break;
		
	case DLG_SPLITTER:
		cname = NXLT("SplitterClass");
		break;
		
	case DLG_SCALE:
		cname = NXLT("ScalerClass");
		break;
		
	case DLG_TABLE:
		cname = NXLT("TableClass");
		break;
		
	case DLG_ROWCOL:
		cname = NXLT("RowColClass");
		break;
		
	case DLG_CANVAS:
		cname = NXLT("CanvasClass");
		break;
		
	case DLG_COMBO_BOX:
		cname = NXLT("ComboBoxClass");
		break;
		
	}
	return (AxMakeStrData(-1, cname));
}

/* : xmTabctlLinkContainees - sub used by xmtcSetTabctlInfo,
 * xmtcApplyTabctrlLinks.
 */
static void
xmTabctlLinkContainees(rw, layer_ix, make_links, create_time)
AXWidget_rect rw;
int layer_ix;
int make_links;   /* bool: TRUE, make links, FALSE, break them */
int create_time;  /* bool: TRUE, we're prior to box 1st display */
{
	int asize, ix, this_uid;
	tablayer *tl;
	AXWidget_dialog dw;
	AXWidget tw;

	if (!(rw && rw->tabctl.num_layers &&
		  (layer_ix < rw->tabctl.num_layers))) {
		
		return;
	}

	dw = (AXWidget_dialog) rw->parent;
	if (!(dw->type == W_DIALOG) && dw->uidWidgets) {
		return;
	}
	
	tl = &rw->tabctl.layerlist[layer_ix];
	asize = AxArraySize(tl->gadget_uids);
	for (ix = 0; ix < asize; ix++) {
		this_uid = AxIntFromArray(tl->gadget_uids, ix);
		axhGetItemI(&this_uid, sizeof(int),
					dw->uidWidgets,
					(char *) NULL, (char **) &tw);
		if (tw) {
			/* link/unlink tw into tabctl rw, in layer layer_ix
			 */
			if (make_links) {
				xmAddWidgetLink(WIDGLINK_TABLAYER,
								rw, tw, (void *) layer_ix);
				
				if (create_time  && (layer_ix != rw->tabctl.active_layer)) {
					tw->saved_display = tw->display;
					tw->display = FALSE;
					tw->update_widget =
						WidgetNeedsUpdated = TRUE;
				}
			}
			else {
				xmDeleteWidgetLink(WIDGLINK_TABLAYER, rw, tw);
			}
		}
	}
}

/* : xmtcSetTabctlInfo - assign the passed elf tabcontrol
 * description onto the C widget substruct *tabctl_info
 */
void
xmtcSetTabctlInfo(rw, tlist, tablayer_info, full_update)
AXWidget_rect rw;                       /* in, possibly null */
tabctl_info *tlist;                     /* in/out */
data_ptr tablayer_info;                 /* in */
int full_update;       /* bool: whether to reset and reapply containee links */
{
    int ix, itmp, num_uids;
    data_ptr layerlist_info;
    data_ptr thislayer, uidlist;
    mbchar *stmp;
	
	if (rw && full_update) {
		/* hide the currently active layer
		 */
		xmRectHideOrRevealLayer(rw, rw->tabctl.active_layer, TRUE);
	}
    tlist->layer_style = AxIntFromArray(tablayer_info, 0);
    tlist->active_layer = AxIntFromArray(tablayer_info, 1);
    layerlist_info = AxArrayElement(tablayer_info, 2);
    if (tlist->layerlist) {
        /* toss old tablayer list
         */
        for (ix = 0; ix < tlist->num_layers; ix++) {
            if (tlist->layerlist[ix].layerName)
                free(tlist->layerlist[ix].layerName);
            if (tlist->layerlist[ix].gadget_uids) {
				if (rw && full_update) {
					/* unlink any containees in this layer
					 */
					xmTabctlLinkContainees(rw, ix, FALSE, FALSE);
				}
				AxTaskFreeElfData(0, tlist->layerlist[ix].gadget_uids);
			}
        }
        free(tlist->layerlist);
    }
    /* copy in new tablayer list info
     */
    tlist->num_layers = AxArraySize(layerlist_info);
    itmp = tlist->num_layers * sizeof(tablayer);
    tlist->layerlist = (tablayer *) malloc(itmp);
    memset((void *) tlist->layerlist, 0, itmp);
    for (ix = 0; ix < tlist->num_layers; ix++) {
        thislayer = AxArrayElement(layerlist_info, ix);
        stmp = AxStrFromArray(thislayer, 0);
        tlist->layerlist[ix].layerName =
            (mbchar *) malloc(MBstrlen(stmp) + 1);
        MBstrcpy(tlist->layerlist[ix].layerName, stmp);
        tlist->layerlist[ix].layerNumber = ix;
        
        uidlist = AxArrayElement(thislayer, 1);
        num_uids = AxArraySize(uidlist);
        if (num_uids) {
            tlist->layerlist[ix].gadget_uids =
                AxTaskCopyElfData(0, uidlist);
			if (rw && full_update) {
				/* link any containees in this layer
				 */
				xmTabctlLinkContainees(rw, ix, TRUE, FALSE);
			}
        }
		else {
			tlist->layerlist[ix].gadget_uids = (data_ptr) 0;
		}
    }
	if (rw && full_update) {
		/* reveal the active layer
		 */
		xmRectHideOrRevealLayer(rw, rw->tabctl.active_layer, FALSE);
	}
}

/* apply the full DBDEX#tablayers info for all tabcontrol-style rects
 * in the passed dbox.  The dbox/children AXWidget structs are assumed
 * to be constructed at this point.
 * this should become generic
 */
static data_ptr       /* return dlg-unique uid, or return 0 on failure */
xmtcApplyTabctrlLinks(dbox_handle)
data_ptr dbox_handle;
{
	int ix, jx, kx, style, asize;
	int this_uid;
	AXWidget_dialog dw;
	AXWidget_rect rw;
	AXWidget tw;
	tablayer *tl;
	hitem *list;

	dw = (AXWidget_dialog) NULL;
	tw = (AXWidget) WidgetPointerFromHandleData(dbox_handle);
	if (!tw) {
		return ((data_ptr) NULL);
	}
	if (W_DIALOG == tw->type) {
		dw = (AXWidget_dialog) tw;
	}
	else if (W_MAIN == tw->type && ((AXWidget_main)tw)->dialog) {
		dw = (AXWidget_dialog) ((AXWidget_main)tw)->dialog;
	}
	if (!(dw && dw->uidWidgets)) {
		return ((data_ptr) NULL);
	}

	/* for each RECT_LAYERED W_RECT in dw, make links to containees
	 */
	for (ix = 0; ix < dw->uidWidgets->width; ix++) {
		list = dw->uidWidgets->table[ix];
		while (list) {
			if (list->value) {
				tw = (AXWidget) list->value;
				if ((W_RECT == tw->type) && (RECT_LAYERED == tw->style)) {
					rw = (AXWidget_rect) tw;
					if (rw->tabctl.num_layers) {
						for (jx = 0; jx < rw->tabctl.num_layers; jx++) {
							xmTabctlLinkContainees(rw, jx, TRUE, TRUE);
						}
					}
				}
			}
			list = list->next;
		}
	}
	
	/* rescan rects, this time hiding the active layers
	 * of any tabcontrols that themselves are hidden.
	 */
	for (ix = 0; ix < dw->uidWidgets->width; ix++) {
		list = dw->uidWidgets->table[ix];
		while (list) {
			if (list->value) {
				tw = (AXWidget) list->value;
				if ((W_RECT == tw->type) && (RECT_LAYERED == tw->style)) {
					rw = (AXWidget_rect) tw;
					if (rw->tabctl_linktab) {
						if (xmWidgetIsHidden(rw)) {
							xmRectHideOrRevealLayer(rw,
													rw->tabctl.active_layer,
													TRUE);
						}
					}
				}
			}
			list = list->next;
		}
	}
	
	return ((data_ptr) NULL);
}

static data_ptr
xmtcApplyTabctrlLink(not_used, info)
data_ptr not_used;
data_ptr info;
{
	int ix, layer_number;
	data_ptr dtmp;
	mbchar *layername;
	AXWidget_dialog dw;
	AXWidget_rect rw;
	AXWidget insertee;

	dtmp = AxArrayElement(info, 0);   /* dialog handle */
	if (!dtmp) {
		return ((data_ptr) NULL);
	}
	dw = (AXWidget_dialog) WidgetPointerFromHandleData(dtmp);
	if (!dw) {
		return ((data_ptr) NULL);
	}
	dtmp = AxArrayElement(info, 1);   /* rect handle */
	if (!dtmp) {
		return ((data_ptr) NULL);
	}
	rw = (AXWidget_rect) WidgetPointerFromHandleData(dtmp);
	if (!(rw && rw->tabctl.num_layers && rw->tabctl.layerlist)) {
		return ((data_ptr) NULL);
	}
	dtmp = AxArrayElement(info, 2);   /* layer name */
	if (!dtmp) {
		return ((data_ptr) NULL);
	}
	layername = (mbchar *) AxStrFromDataPtr(dtmp);
	if (!(layername && *layername)) {
		return ((data_ptr) NULL);
	}
	dtmp = AxArrayElement(info, 3);   /* insertee handle */
	if (!dtmp) {
		return ((data_ptr) NULL);
	}
	insertee = (AXWidget) WidgetPointerFromHandleData(dtmp);
	if (!insertee) {
		return ((data_ptr) NULL);
	}
	
	for (ix = 0, layer_number = -1; ix < rw->tabctl.num_layers; ix++) {
		if (rw->tabctl.layerlist[ix].layerName &&
			!MBstrcmp(rw->tabctl.layerlist[ix].layerName, layername)) {
			
			layer_number = ix;
			break;
		}
	}
	if (layer_number >= 0) {
		xmAddWidgetLink(WIDGLINK_TABLAYER, rw, insertee,
						(void *) layer_number);
	}
	return ((data_ptr) NULL);
}

/* : xmtcBreakTabctrlLinks - for the passed widget,
 * break any tablayer uplinks currently present.
 */
static data_ptr
xmtcBreakTabctrlLinks(wP)
data_ptr wP;
{
	int lx;
	hitem *list, *ltmp;
	widglink *wl;
	AXWidget w;
	AXWidget_rect rw;

	w = (AXWidget) WidgetPointerFromHandleData(wP);
	if (!w) {
		return ((data_ptr) NULL);
	}
	if (w->tabctl_linktab) {
		for (lx = 0; lx < w->tabctl_linktab->width; lx++) {
			list = w->tabctl_linktab->table[lx];
			while (list) {
				ltmp = list->next;      /* (*list) may be deleted below */
				wl = (widglink *) list->value;
				if (!wl->direction) {
					/* break this uplink (to tabcontrol W_RECT)
					 */
					xmDeleteWidgetLink(WIDGLINK_TABLAYER, w, wl->to);
					rw = (AXWidget_rect) wl->to;
					if ((long) wl->link_data != rw->tabctl.active_layer) {
						w->display = w->saved_display; /* un-hide */
					}
				}
				list = ltmp;
			}
		}
	}
	return ((data_ptr) NULL);
}

static data_ptr
xmtcGetActiveTabctrlLayer(wP)
data_ptr wP;
{
	AXWidget_rect rw;
	int ix;
	
	rw = (AXWidget_rect) WidgetPointerFromHandleData(wP);
	if (!(rw && RECT_LAYERED == rw->style)) {
		return ((data_ptr) NULL);
	}
	ix = rw->tabctl.active_layer;
	return (AxMakeStrData(-1, rw->tabctl.layerlist[ix].layerName));
}

static data_ptr
xmtcSetActiveTabctrlLayer(dP)
data_ptr dP;
{
	AXWidget_rect rw;
	mbchar *layerName;
	int ix;
	
	rw = (AXWidget_rect) WidgetPointerFromHandleData(AxArrayElement(dP, 0));
	if (!(rw && RECT_LAYERED == rw->style)) {
		return ((data_ptr) NULL);
	}
	layerName = AxStrFromArray(dP, 1);
	if (!(layerName && *layerName)) {
		return ((data_ptr) NULL);
	}
	for (ix = 0; ix < rw->tabctl.num_layers; ix++) {
		if (rw->tabctl.layerlist[ix].layerName &&
			!MBstrcmp(rw->tabctl.layerlist[ix].layerName, layerName)) {
			
			if (rw->tabctl.active_layer !=
				rw->tabctl.layerlist[ix].layerNumber) {

				/* activate this layer
				 */
				xmRectHideOrRevealLayer(rw, rw->tabctl.active_layer, TRUE);
				rw->tabctl.active_layer = ix;
				rw->update_widget = WidgetNeedsUpdated = TRUE;
				xmRectHideOrRevealLayer(rw, ix, FALSE);
			}
		}
	}
	return ((data_ptr) NULL);
}

static data_ptr
xmtcSetTablayers(dP)
data_ptr dP;
{
	AXWidget_rect rw;
	data_ptr tablayer_info;
	
	rw = (AXWidget_rect) WidgetPointerFromHandleData(AxArrayElement(dP, 0));
	if (!(rw && RECT_LAYERED == rw->style)) {
		return ((data_ptr) NULL);
	}
	tablayer_info = AxArrayFromArray(dP, 1);

	if (!(rw->flags & WFLG_IN_EDIT)) {
		UnMapWidget(rw);                           /* this is heavy-handed */
	}
	rw->update_widget = WidgetNeedsUpdated = TRUE; /* this is necessary */
	
	SetTabLayers(rw, tablayer_info, TRUE);
	
	return ((data_ptr) NULL);
}

/* : xmtcReadTabctlInset - given an elf tabcontrol description,
 * return the top-inset value (in pixels) that would result on
 * the realized tabctl widget.  The top inset is the pixel
 * height of the tabs part of the widget.
 */
static data_ptr
xmtcReadTabctlInset(info)
data_ptr info;
{
	tabctl_info tt;
	data_ptr tablayer_info;
	data_ptr dbox_ref_font;
	data_ptr tabctl_label_font;
	XFontStruct *tabfont;
	XFontStruct *refFont;
	int tabctl_width, itmp, ix, inset;
	
	tabctl_width = AxIntFromArray(info, 0);
	tablayer_info = AxArrayElement(info, 1);
	tabctl_label_font = AxArrayElement(info, 2);
	dbox_ref_font = AxArrayElement(info, 3);
	
	refFont = lookupReferenceFont(dbox_ref_font, &itmp);
	tabfont = lookupNamedFont(tabctl_label_font);
	if (!tabfont) {
		tabfont = AxDlgMainFont;
	}
	tabctl_width = axDlgScaleX(tabctl_width, refFont);

	/* fill out tt
	 */
	memset((void *) &tt, 0, sizeof(tabctl_info));  /* must do */
	xmtcSetTabctlInfo((AXWidget_rect) NULL, &tt, tablayer_info, FALSE);
	
	if (!tt.num_layers) {
		return (AxMakeIntData(0));
	}
	
	inset = xmTabctlTopInset(&tt, tabctl_width, tabfont);

	for (ix = 0; ix < tt.num_layers; ix++) {
		if (tt.layerlist[ix].layerName) {
			free(tt.layerlist[ix].layerName);
		}
	}
	if (tt.layerlist) {
		free(tt.layerlist);
	}
	
	return (AxMakeIntData(inset));
}

/* : xmtcDistinctLayernames - return whether passed array
 * of tabcontrol layernames are all different from each other.
 */
static data_ptr
xmtcDistinctLayernames(info)
data_ptr info;               /* array of layer name strings */
{
	int ix, asize;
	int distinct = -1;       /* elf-flavored TRUE */
	mbchar *thisname;
	char *ctmp;
	axhtab layertab;

	asize = AxArraySize(info);
	if (!asize) {
		return (AxMakeIntData(0));
	}
	axhInitTable(&layertab, 7, 0, 0);
	for (ix = 0; ix < asize; ix++) {
		thisname = AxStrPtrFromArray(info, ix);
		if (thisname) {
			if (axhTestItem((char *) thisname, &layertab)) {
				distinct = 0;                           /* already there */
				break;
			}
			axhGetItem(thisname, &layertab, 0, &ctmp);  /* add in */
		}
	}
	axhDeleteTable(&layertab, 0);
	return (AxMakeIntData(distinct));
}	
	
/* : wfActualSize - return pixel footprint corresponding
 * to the passed gadget's elf data.
 * Param 'info' arrives in format:
 *   <dbox revstamp><dbox reference font><dbox no_expand bool>
 *     <gadget type><gadget "size atts">
 * The "size atts" are defined in db_ctrl_get_geom_atts$
 */
static data_ptr
wfActualSize(info)
data_ptr info;
{
	int dboxNoExpand, dboxRev, gType, different;
	data_ptr ret_data, sizeAtts;
	r32 csize;
	XFontStruct *refFont;
	
	dboxRev = AxIntFromArray(info, 0);
	refFont = lookupReferenceFont(AxArrayFromArray(info, 1), &different);
	dboxNoExpand = AxIntFromArray(info, 2);
	gType = AxIntFromArray(info, 3);
	sizeAtts = AxArrayFromArray(info, 4);
	
	wActualGadgetSize(dboxRev, refFont, dboxNoExpand,
					  gType, sizeAtts, &csize);

	ret_data = AxMakeArray(4);
	ret_data = AxAddIntToArray(ret_data, 0, csize.origin.x);
	ret_data = AxAddIntToArray(ret_data, 1, csize.origin.y);
	ret_data = AxAddIntToArray(ret_data, 2, csize.width);
	ret_data = AxAddIntToArray(ret_data, 3, csize.height);
	
	return (ret_data);
}

static void
initKeyFocusLinks(dbox, dbox_handle_data)
data_ptr dbox;
data_ptr dbox_handle_data;
{
	AXWidget_dialog dboxP;
	AXWidget w;
	data_ptr linksP;
	int cx, vx, size, asize, kx, ax;

	if (!dbox)
		return;
	w = (AXWidget) WidgetPointerFromHandleData(dbox_handle_data);
	if (!w)
		return;
	if (W_DIALOG == w->type) {
		dboxP = (AXWidget_dialog) w;
	} else if (W_MAIN == w->type && ((AXWidget_main) w)->dialog) {
		dboxP = (AXWidget_dialog) ((AXWidget_main) w)->dialog;
	} else
		return;
	linksP = AxArrayElement(dbox, DB_keyfocus_links);
	if (dboxP->keyFocusLinks) {
		free(dboxP->keyFocusLinks);
		dboxP->keyFocusLinks = NULL;
	}
	if (!linksP || AxArraySize(linksP) == 0)
		return;
	size = AxArraySize(linksP);
	for (cx = 0, asize = 0; cx < size; cx++) {
		data_ptr arr;
		arr = AxArrayElement(linksP, cx);
		for (vx = 0; vx < AxArraySize(arr); vx++)
			asize++;
		asize++; /* 0 */
	}
	dboxP->keyFocusLinks = CreateIntsArray(asize);
	size = AxArraySize(linksP);
	kx = 0;
	for (cx = 0, ax = 0; cx < size; cx++) {
		data_ptr arr;
		arr = AxArrayElement(linksP, cx);
		for (vx = 0; vx < AxArraySize(arr); vx++) {
			WriteIntsArray(dboxP->keyFocusLinks, ax,
					   WidgetPointerFromHandleData(AxArrayElement(arr, vx)));
			ax++;
		}
		WriteIntsArray(dboxP->keyFocusLinks, ax, 0);
		ax++;
	}
	return;
}

static data_ptr
xmkfSetKeyFocus(info)
data_ptr info;
{
	AXWidget ww = WidgetPointerFromHandleData(info);
	xmTakeKeyFocus(ww, (XButtonEvent *) NULL, FALSE);	
	return ((data_ptr) NULL);
}

static data_ptr
xmkfGetKeyFocusWidget(info)
data_ptr info;
{
	AXWidget kw;
	AXWidget_dialog dw;
	AXWidget_handle wh;
	
	dw = (AXWidget_dialog) WidgetPointerFromHandleData(info);
	if (!dw) {
		return ((data_ptr) NULL);
	}
	kw = xmKeyFocusWidget(dw);
	if (!kw) {
		return ((data_ptr) NULL);
	}
	wh = HANDLE_FROM_WIDGET(kw);
	return (AxMakeBinaryData(sizeof(wh), &wh));
}

/**************************/

static char *
doAlignLink(scope, evnum, from, to, namelen, linkType, callData)
axconnC *scope;
int evnum;
linkname *from;
linkname *to;
int namelen;
int linkType;
char *callData;
{
	int typeIndex, ix, direction;
	int doCall;
	gadget_size *fromG, *toG;
	axconn *toconn, **ctmp;
	axhtab *mylinks;
	hitem *list;
	axlink *fromlink, *tlink;
	align_link_data *tdata;
	size_change *fromCh, toCh, saveFromCh;
	linkname toLink, *tlname;
	enum edgeType myEdge;
	
	fromCh = (size_change *) callData; /* caller's sizechange */
  
	if (!fromCh->rightChange && !fromCh->leftChange &&
		!fromCh->topChange && !fromCh->bottomChange) {
		return ((char *) NULL);	/* no propagation past this point */
	}
  
	/* locate myself (*toG)
	 */
	if (!scope->locate(scope, linkType, to, (char **) &toG, &ctmp))
		return ((char *) NULL);
	toconn = *ctmp;
	if (!toG || !toconn)
		return ((char *) NULL);
  
	/* get a reference to my links
	 */
	axhGetItemI((void *) &linkType, sizeof(int),
				&toconn->typeTable, (char *) 0, (char **) &typeIndex);
	mylinks = &toconn->linkTables[typeIndex];
  
	/* save a copy of the caller's sizechange request
	 */
	saveFromCh = *fromCh;
  
	/* make a copy of linkname param "to", which we'll fully init
	 * for calls over links
	 */
	toLink = *to;
  
	/* make a self-relative copy of the sizechange request
	 */
	if (from) {
		/* compute self-relative change sum against
		 * all align links between my caller and me.
		 */
		memset((void *) &toCh, 0, sizeof(size_change));
		toCh.callerType = toG->gadgetType;
		toCh.callerMinorType = toG->gadgetMinorType;
		toCh.currentsize = toG->currentsize;
		
		for (ix = 0; ix < mylinks->width; ix++) {
			list = mylinks->table[ix];
			while(list) {
				tlname = (linkname *) list->name;
				if (tlname->toid == from->toid) {
					tlink = (axlink *) list->value;
					tdata = (align_link_data *) tlink->data;
					direction = tlink->direction;

					selfRelativeSizeChange(scope, evnum, from, fromCh,
										   tdata, direction, mylinks,
										   toG, &toCh);
				}
				list = list->next;
			}
		}
	}
	else {
		/* no "from" denotes a direct call rather than a
		 * link call; in this case the size change request
		 * is already self-relative
		 */
		toCh = *fromCh;
	}
  
	if (!toCh.rightChange && !toCh.leftChange &&
		!toCh.topChange && !toCh.bottomChange) {
		return ((char *) NULL);	/* no propagation past this point */
	}
  
	/* replace the call data with the self-relative copy;
	 * must use the same memory as passed by the caller.
	 */
	*fromCh = toCh;

	/* before calling links, accept change on self
	 */
	takeChange(fromCh, toG);
	
	/* call my links with this sizechange request.
	 * first call uplinks to containers, then call edge-edge links.
	 */
  
	/* call uplinks to rowcol containers
	 */
	for (ix = 0; ix < mylinks->width; ix++) {
		list = mylinks->table[ix];
		while(list) {
			tlink = (axlink *) list->value;
			tdata = (align_link_data *) tlink->data;
			if (tdata->type == ROW_COL && tlink->direction) {
				tlname = (linkname *) list->name;
				if (!from || from->toid != tlname->toid) {
					toLink.type = tdata->type;
					toLink.toedge = tdata->fromEdge;
					scope->linkCall(scope, evnum, &toLink, list->name,
									sizeof(linkname), ALIGN_LINK_TYPE,
									(char *) fromCh);
				}
			}
			list = list->next;
		}
	}
	
	/* call edge-edge links
	 */
	for (ix = 0; ix < mylinks->width; ix++) {
		list = mylinks->table[ix];
		while(list) {
			tlink = (axlink *) list->value;
			tdata = (align_link_data *) tlink->data;
			if (tdata->type == EDGE_TO_EDGE) {
				tlname = (linkname *) list->name;
				
				if (from && from->toid == tlname->toid) {
					doCall = FALSE; /* don't call my caller */
				}
				else {
					myEdge = tlink->direction ?
						tdata->fromEdge : tdata->toEdge;
					
					switch (myEdge) {
					case INNER_TOP:
					case OUTER_TOP:
						doCall = (fromCh->topChange != 0);
						break;
						
					case INNER_RIGHT:
					case OUTER_RIGHT:
						doCall = (fromCh->rightChange != 0);
						break;
						
					case INNER_BOTTOM:
					case OUTER_BOTTOM:
						doCall = (fromCh->bottomChange != 0);
						break;
						
					case INNER_LEFT:
					case OUTER_LEFT:
						doCall = (fromCh->leftChange != 0);
						break;
					}
				}
				if (doCall) {
					
					toLink.type = tdata->type;
					if (tlink->direction) {
						toLink.toedge = tdata->fromEdge;
					}
					else {
						toLink.toedge = tdata->toEdge;
					}
					scope->linkCall(scope, evnum, &toLink, list->name,
									sizeof(linkname), ALIGN_LINK_TYPE,
									(char *) fromCh);
				}
			}
			list = list->next;
		}
	}
  
	/* done calling my links
	 */
  
	/* convert the sizechange request
	 * back into caller-relative form
	 */
	saveFromCh.leftChange += (fromCh->leftChange - toCh.leftChange);
	saveFromCh.rightChange += (fromCh->rightChange - toCh.rightChange);
	saveFromCh.topChange += (fromCh->topChange - toCh.topChange);
	saveFromCh.bottomChange += (fromCh->bottomChange - toCh.bottomChange);
	*fromCh = saveFromCh; /* this updates passed callData */
  
	return ((char *) NULL);
}

/*
 * : selfRelativeSizeChange - convert the passed caller-relative
 * sizechange request into a self-relative sizechange request.
 *
 * This function may perform link calls into any containees of *toSize.
 */
static void
selfRelativeSizeChange(scope, evnum, from, fromCh, linkData,
					   linkDirection, tolinks, toSize, toCh)
axconnC *scope;
int evnum;
linkname *from;             /* in: caller's id */
size_change *fromCh;        /* in: caller's sizechange request */
align_link_data *linkData;  /* in: data for link between caller and me */
int linkDirection;          /* in: bool */
axhtab *tolinks;            /* in */
gadget_size *toSize;        /* in */
size_change *toCh;          /* out */
{
	int tmin, tmax, itmp, callerInner, ix;
	int fixed;
	enum edgeType myEdge, callerEdge;
	r32 callerRect;
	p32 ptmp;

	fixed = !isDeformable(toSize);

	/* the rest of this function is spent determining the
	 * left,right,top,bottom change numbers to write onto toCh
	 */

	/* init the callerRect, expressing the caller's size
	 * including its proposed changes
	 */
	callerRect = fromCh->currentsize;
	callerRect.origin.x += fromCh->leftChange;
	callerRect.origin.y += fromCh->topChange;
	ptmp.x = fromCh->currentsize.origin.x + fromCh->currentsize.width +
		fromCh->rightChange;
	ptmp.y = fromCh->currentsize.origin.y + fromCh->currentsize.height +
		fromCh->bottomChange;
	itmp = ptmp.x - callerRect.origin.x;
	if (itmp < 0) {
		itmp = 0;
	}
	callerRect.width = itmp;
	itmp = ptmp.y - callerRect.origin.y;
	if (itmp < 0) {
		itmp = 0;
	}
	callerRect.height = itmp;
	
	switch(linkData->type) {
		
	case ROW_COL:
		if (linkDirection) {
			/* this is an 'up' link.  this means i'm being
			 * called by a ROW_COL aligner.  The sizechange
			 * "request" passed by the rowCol already describes
			 * me -- it's a command to assume this new size
 * %%%%%%% currently not used, since doRowcolAlign in its simple form
 *         just directly hammers the positions of its containees...
 */
			toCh->currentsize = fromCh->currentsize;
			toCh->leftChange = fromCh->leftChange;
			toCh->rightChange = fromCh->rightChange;
			toCh->topChange = fromCh->topChange;
			toCh->bottomChange = fromCh->bottomChange;
		}
		else {
			/* i'm a rowCol aligner over the caller; i would
			 * prefer to remain at my current size, but must
			 * do the rowcol alignment over my containees now
			 * in order to check whether my size must change
			 */
			doRowcolAlign(scope, tolinks, toSize,
						  TRUE, from, &callerRect, toCh);
			return;
		}
		break;
		
	case EDGE_TO_EDGE:
		if (toSize->edges[0]) {
			/* I'm a container top, and the call is coming
			 * in from one of my edges. all I care about at
			 * this point is handling any rowcol-linked containees.
			 */
			doRowcolAlign(scope, tolinks, toSize,
						  FALSE, from, &callerRect, toCh);
		}
		else {
			/* regular edge-edge link handling
			 */
			if (linkDirection) {
				myEdge = linkData->fromEdge;
				callerEdge = linkData->toEdge;
			}
			else {
				myEdge = linkData->toEdge;
				callerEdge = linkData->fromEdge;
			}
			
			/* init tmin and tmax, representing distance
			 * limits compliant with the edge-to-edge link
			 */
			if (callerEdge == OUTER_RIGHT ||
				callerEdge == OUTER_LEFT ||
				callerEdge == OUTER_TOP ||
				callerEdge == OUTER_BOTTOM) {
				
				/* these limits are relative to the outside of callerRect.
				 */
				callerInner = FALSE;
			}
			else {
				/* these limits are relative to the inside of callerRect.
				 */
				callerInner = TRUE;
			}
			switch (myEdge) {
				
			case OUTER_RIGHT:
				if (callerInner) {
					/* we're attached to caller's inner right
					 */
					tmin = callerRect.origin.x + callerRect.width - 1 -
						linkData->maxDistance;
					tmax = callerRect.origin.x + callerRect.width - 1 -
						linkData->minDistance;
				}
				else {
					/* we're attached to caller's outer left
					 */
					tmin = callerRect.origin.x - linkData->maxDistance;
					tmax = callerRect.origin.x - linkData->minDistance;
				}
				
				itmp = toCh->currentsize.origin.x + toCh->leftChange +
					toCh->currentsize.width + toCh->rightChange - 1;
				
				if (itmp > tmax) { 
					toCh->rightChange += tmax - itmp;
				}
				else if (itmp < tmin) {
					toCh->rightChange += tmin - itmp;
				}
				if (fixed || toSize->fixedWidth) {
					toCh->leftChange = toCh->rightChange; /* fixed-size move */
				}
				else {
					/* keep to's width above minimum
					 */
					itmp = toCh->currentsize.width + toCh->rightChange;
					if (itmp < toSize->minsize.width) {
						toCh->leftChange += itmp - toSize->minsize.width;
					}
				}
				break;
				
			case OUTER_LEFT:
				if (callerInner) {
					/* we're attached to caller's inner left
					 */
					tmin = callerRect.origin.x + linkData->minDistance;
					tmax = callerRect.origin.x + linkData->maxDistance;
				}
				else {
					/* we're attached to caller's outer right
					 */
					tmin = callerRect.origin.x + callerRect.width - 1 +
						linkData->minDistance;
					tmax = callerRect.origin.x + callerRect.width - 1 +
						linkData->maxDistance;
				}
				
				itmp = toCh->currentsize.origin.x + toCh->leftChange;
				
				if (itmp > tmax) { 
					toCh->leftChange += tmax - itmp;
				}
				else if (itmp < tmin) {
					toCh->leftChange += tmin - itmp;
				}
				if (fixed || toSize->fixedWidth) {
					toCh->rightChange = toCh->leftChange; /* fixed-size move */
				}
				else {
					/* keep to's width above minimum
					 */
					itmp = toCh->currentsize.width + toCh->rightChange;
					if (itmp < toSize->minsize.width) {
						toCh->rightChange += itmp - toSize->minsize.width;
					}
				}
				break;
				
			case OUTER_TOP:
				if (callerInner) {
					/* we're attached to caller's inner top
					 */
					tmin = callerRect.origin.y + linkData->minDistance;
					tmax = callerRect.origin.y + linkData->maxDistance;
				}
				else {
					/* we're attached to caller's outer bottom
					 */
					tmin = callerRect.origin.y + callerRect.height - 1 +
						linkData->minDistance;
					tmax = callerRect.origin.y + callerRect.height - 1 +
						linkData->maxDistance;
				}
				
				itmp = toCh->currentsize.origin.y + toCh->topChange;
				
				if (itmp > tmax) { 
					toCh->topChange += tmax - itmp;
				}
				else if (itmp < tmin) {
					toCh->topChange += tmin - itmp;
				}
				if (fixed || toSize->fixedHeight) {
					toCh->bottomChange = toCh->topChange; /* fixed-size move */
				}
				else {
					/* keep to's height above minimum
					 */
					itmp = toCh->currentsize.height - toCh->topChange;
					if (itmp < toSize->minsize.height) {
						toCh->bottomChange -= itmp - toSize->minsize.height;
					}
				}
				break;
				
			case OUTER_BOTTOM:
				if (callerInner) {
					/* we're attached to caller's inner bottom
					 */
					tmin = callerRect.origin.y + callerRect.height - 1 -
						linkData->maxDistance;
					tmax = callerRect.origin.y + callerRect.height - 1 -
						linkData->minDistance;
				}
				else {
					/* we're attached to caller's outer top
					 */
					tmin = callerRect.origin.y - linkData->maxDistance;
					tmax = callerRect.origin.y - linkData->minDistance;
				}
				
				itmp = toCh->currentsize.origin.y + toCh->topChange +
					toCh->currentsize.height + toCh->bottomChange - 1;
				
				if (itmp > tmax) { 
					toCh->bottomChange += tmax - itmp;
				}
				else if (itmp < tmin) {
					toCh->bottomChange += tmin - itmp;
				}
				if (fixed || toSize->fixedHeight) {
					toCh->topChange = toCh->bottomChange; /* fixed-size move */
				}
				else {
					/* keep to's height above minimum
					 */
					itmp = toCh->currentsize.height - toCh->topChange;
					if (itmp < toSize->minsize.height) {
						toCh->topChange -= itmp - toSize->minsize.height;
					}
				}
				break;
				
			case INNER_RIGHT:
				tmin = callerRect.origin.x + callerRect.width - 1 +
					linkData->minDistance;
				tmax = callerRect.origin.x + callerRect.width - 1 +
					linkData->maxDistance;
				
				itmp = toCh->currentsize.origin.x + toCh->leftChange +
					toCh->currentsize.width + toCh->rightChange - 1;
				
				if (itmp > tmax) {
					toCh->rightChange += tmax - itmp;
				}
				else if (itmp < tmin) {
					toCh->rightChange += tmin - itmp;
				}
				if (fixed || toSize->fixedWidth) {
					toCh->leftChange = toCh->rightChange;
				}
				else {
					/* keep to's width above minimum
					 */
					itmp = toCh->currentsize.width + toCh->rightChange;
					if (itmp < toSize->minsize.width) {
						toCh->leftChange += itmp - toSize->minsize.width;
					}
				}
				break;
				
			case INNER_LEFT:
				tmin = callerRect.origin.x - linkData->maxDistance;
				tmax = callerRect.origin.x - linkData->minDistance;
				
				itmp = toCh->currentsize.origin.x + toCh->leftChange;
				
				if (itmp > tmax) {
					toCh->leftChange += tmax - itmp;
				}
				else if (itmp < tmin) {
					toCh->leftChange += tmin - itmp;
				}
				if (fixed || toSize->fixedWidth) {
					toCh->rightChange = toCh->leftChange;
				}
				else {
					/* keep to's width above minimum
					 */
					itmp = toCh->currentsize.width - toCh->leftChange;
					if (itmp < toSize->minsize.width) {
						toCh->rightChange -= itmp - toSize->minsize.width;
					}
				}
				break;
				
			case INNER_TOP:
				tmin = callerRect.origin.y - linkData->maxDistance;
				tmax = callerRect.origin.y - linkData->minDistance;
				
				itmp = toCh->currentsize.origin.y + toCh->topChange;
				
				if (itmp > tmax) {
					toCh->topChange += tmax - itmp;
				}
				else if (itmp < tmin) {
					toCh->topChange += tmin - itmp;
				}
				if (fixed || toSize->fixedHeight) {
					toCh->bottomChange = toCh->topChange;
				}
				else {
					/* keep to's height above minimum
					 */
					itmp = toCh->currentsize.height - toCh->topChange;
					if (itmp < toSize->minsize.height) {
						toCh->bottomChange -= itmp - toSize->minsize.height;
					}
				}
				break;
				
			case INNER_BOTTOM:
				tmin = callerRect.origin.y + callerRect.height - 1 +
					linkData->minDistance;
				tmax = callerRect.origin.y + callerRect.height - 1 +
					linkData->maxDistance;
				
				itmp = toCh->currentsize.origin.y + toCh->topChange +
					toCh->currentsize.height + toCh->bottomChange - 1;
				
				if (itmp > tmax) {
					toCh->bottomChange += tmax - itmp;
				}
				else if (itmp < tmin) {
					toCh->bottomChange += tmin - itmp;
				}
				if (fixed || toSize->fixedHeight) {
					toCh->topChange = toCh->bottomChange;
				}
				else {
					/* keep to's height above minimum
					 */
					itmp = toCh->currentsize.height + toCh->bottomChange;
					if (itmp < toSize->minsize.height) {
						toCh->topChange += itmp - toSize->minsize.height;
					}
				}
				break;
				
			case CENTER:
				/* move in parallel with the caller.  Define this
				 * as taking the caller's origin change
				 */
				toCh->leftChange = fromCh->leftChange;
				toCh->topChange = fromCh->topChange;
				toCh->rightChange = toCh->leftChange;
				toCh->bottomChange = toCh->topChange;
				break;
				
			}
		}
		break;
	}
}

/* : doRowcolAlign - do a simple, single-line distribution of 
 * the set of widgets connected to myself as downlinks of type ROW_COL.
 * No attempt to partition the set into connected groups, no attempt to
 * matrix-arrange.... intended primarily for bottom-panel o' buttons
 */
static void
doRowcolAlign(scope, tolinks, toSize,
			  callerIsContainee, callerLinkName, callerRect, toCh)
axconnC *scope;
axhtab *tolinks;            /* in */
gadget_size *toSize;        /* in: my current size */
int callerIsContainee;      /* in */
linkname *callerLinkName;   /* in */
r32 *callerRect;            /* in */
size_change *toCh;    /* out: my own sizechange, possibly updated here */
{
	int numContainees, itmp, numGaps;
	int horiz, ix, cx, minGap, tpos;
	int occupiedExtent, gapExtent, tmax, sorted;
	gadget_size *cG;
	axconn **ctmp;
	axlink *tlink;
	align_link_data *tdata;
	hitem *list;
	r32 myrect;
	p32 ptmp;
	gadget_size **cons, *gtmp;
	
	/* init myrect as my current size plus whatever change is present
	 */
	myrect = toCh->currentsize;
	myrect.origin.x += toCh->leftChange;
	myrect.origin.y += toCh->topChange;
	ptmp.x = toCh->currentsize.origin.x + toCh->currentsize.width +
		toCh->rightChange;
	ptmp.y = toCh->currentsize.origin.y + toCh->currentsize.height +
		toCh->bottomChange;
	itmp = ptmp.x - myrect.origin.x;
	if (itmp < 0) {
		itmp = 0;
	}
	myrect.width = itmp;
	itmp = ptmp.y - myrect.origin.y;
	if (itmp < 0) {
		itmp = 0;
	}
	myrect.height = itmp;
	
	/* count the number of ROW_COL-linked containees.  The toSize->edges
	 * comparisons are done to exclude links to my own edges; we want only
	 * the real containees.
	 */
	for (ix = 0, numContainees = 0, minGap = 0; ix < tolinks->width; ix++) {
		list = tolinks->table[ix];
		while(list) {
			tlink = (axlink *) list->value;
			tdata = (align_link_data *) tlink->data;
			if (tdata->type == ROW_COL && !tlink->direction) {
				minGap = tdata->minDistance; /* (yes, last one wins) */
				numContainees++;
			}
			list = list->next;
		}
	}
	
	if (!numContainees)
		return;
	
	numGaps = numContainees + 1;
	
	/* whether to arrange vertically or horizontally is taken simply
	 * from the major extent of myself
	 */
	horiz = myrect.width > myrect.height;
	
	/* allocate an array of rect instances just to remember
	 * each containee's current size; the rowcol-align process
	 * will possibly change the origins of the containees' currentsizes
	 */
	itmp = sizeof(gadget_size *) * numContainees;
	cons = (gadget_size **) malloc(itmp);
	memset((void *) cons, 0, itmp);
	for (ix = 0, cx = 0; ix < tolinks->width; ix++) {
		list = tolinks->table[ix];
		while(list) {
			tlink = (axlink *) list->value;
			tdata = (align_link_data *) tlink->data;
			if (tdata->type == ROW_COL && !tlink->direction) {
				if (list->name) {
					if (scope->locate(scope, ALIGN_LINK_TYPE, list->name,
									  (char **) &cG, &ctmp)) {
						cons[cx] = cG;
						cx++;
					}
				}
			}
			list = list->next;
		}
	}
	
	/* sort the cons array elements by increasing major coord value
	 */
	sorted = FALSE;
	if (horiz) {
		while (!sorted) {
			for (ix = 0, sorted = TRUE; ix < numContainees; ix++) {
				if (ix < numContainees - 1) {
					if (cons[ix]->currentsize.origin.x >
						cons[ix + 1]->currentsize.origin.x) {
						
						gtmp = cons[ix];
						cons[ix] = cons[ix + 1];
						cons[ix + 1] = gtmp;
						sorted = FALSE;
						continue;
					}
				}
			}
		}
	}
	else {
		while (!sorted) {
			for (ix = 0, sorted = TRUE; ix < numContainees; ix++) {
				if (ix < numContainees - 1) {
					if (cons[ix]->currentsize.origin.y >
						cons[ix + 1]->currentsize.origin.y) {
						
						gtmp = cons[ix];
						cons[ix] = cons[ix + 1];
						cons[ix + 1] = gtmp;
						sorted = FALSE;
						continue;
					}
				}
			}
		}
	}
	
	/* arrangement is decided by evenly distributing the gaps; find
	 * out how much gap space there is
	 */
	for (ix = 0, occupiedExtent = 0; ix < numContainees; ix++) {
		if (cons[ix]->currentsize.width && cons[ix]->currentsize.height) {
			occupiedExtent +=
				horiz ? cons[ix]->currentsize.width :
					cons[ix]->currentsize.height;
		}
	}
	itmp = horiz ? myrect.width : myrect.height;
	gapExtent = itmp - occupiedExtent;
	
	if (gapExtent < (numGaps * minGap)) {
		/* i need to stretch a bit to fit these guys
		 */
		itmp = (numGaps * minGap) - gapExtent; /* by this amount */
		if (horiz) {
			if (!callerIsContainee && 
				callerLinkName && callerLinkName->toedge == OUTER_RIGHT) {
				
				/* stretch to the left
				 */
				toCh->leftChange -= itmp;
			}
			else {
				/* just punch out my own right edge
				 */
				toCh->rightChange += itmp;
			}
			myrect.width += itmp;
		}
		else {
			if (!callerIsContainee &&
				callerLinkName && callerLinkName->toedge == OUTER_BOTTOM) {
				
				/* stretch upwards
				 */
				toCh->topChange -= itmp;
			}
			else {
				/* just punch out my own bottom edge
				 */
				toCh->bottomChange += itmp;
			}
			myrect.height += itmp;
		}
		gapExtent += itmp;
	}
	
	if (horiz) {
		/* now check vertical extents of containees, and keep
		 * myself taller than the tallest by (2 * minGap)
		 */
		for (ix = 0, tmax = 0; ix < numContainees; ix++) {
			if (cons[ix]->currentsize.height > tmax) {
				tmax = cons[ix]->currentsize.height;
			}
		}
		itmp = (myrect.height - tmax) - (2 * minGap);
		if (!callerIsContainee &&
			callerLinkName && callerLinkName->toedge == OUTER_BOTTOM) {
			
			toCh->topChange += itmp;
		}
		else {
			toCh->bottomChange -= itmp;
		}
		myrect.height -= itmp;
	}
	else {
		/* check horiz extents of containees, and adjust as needed.
		 */
		for (ix = 0, tmax = 0; ix < numContainees; ix++) {
			if (cons[ix]->currentsize.width > tmax) {
				tmax = cons[ix]->currentsize.width;
			}
		}
		itmp = (myrect.width - tmax) - (2 * minGap);
		if (!callerIsContainee &&
			callerLinkName && callerLinkName->toedge == OUTER_RIGHT) {
			
			toCh->leftChange += itmp;
		}
		else {
			toCh->rightChange -= itmp;
		}
		myrect.width -= itmp;
	}
	
	/* convert gapExtent from the sum of all gaps, to the size of
	 * an individual gap.
	 */
	if (numGaps > 2) {
		/* minGap is used on the ends, slack distributed in the middle
		 */
		gapExtent = (gapExtent - (2 * minGap)) / (numGaps - 2);
	}
	else {
		gapExtent = gapExtent / numGaps;  /* just distribute evenly */
	}
	
	/* express arrangement by rewriting the origins of the size copies.
	 */
	if (horiz) {
		tpos = myrect.origin.x + minGap;
		for (ix = 0; ix < numContainees; ix++) {
			if (cons[ix]->currentsize.width && cons[ix]->currentsize.height) {
				cons[ix]->currentsize.origin.x = tpos;
				itmp = (myrect.height - cons[ix]->currentsize.height) / 2;
				cons[ix]->currentsize.origin.y = myrect.origin.y + itmp;
				cons[ix]->updateElf = TRUE;
				tpos += cons[ix]->currentsize.width + gapExtent;
			}
		}
	}
	else {
		tpos = myrect.origin.y + minGap;
		for (ix = 0; ix < numContainees; ix++) {
			if (cons[ix]->currentsize.width && cons[ix]->currentsize.height) {
				cons[ix]->currentsize.origin.y = tpos;
				itmp = (myrect.width - cons[ix]->currentsize.width) / 2;
				cons[ix]->currentsize.origin.x = myrect.origin.x + itmp;
				cons[ix]->updateElf = TRUE;
				tpos += cons[ix]->currentsize.height + gapExtent;
			}
		}
	}
	
	free(cons);
}

/* : wInitDialogGadgetSets - the AXDialogGadgetSets table pairs
 * dbox menu id's with gadget_set instances.  Each gadget_set
 * stores gadget_size instances representing that dbox's widgets
 */
void
wInitDialogGadgetSets()
{
	axhInitTable(&AXDialogGadgetSets, 31, newDlgGadgetSet, tossDlgGadgetSet);
}

static char *
newDlgGadgetSet(scopeName, namelen, notUsed)
void *scopeName;
int namelen;
void *notUsed;
{
	gadget_set *newset;
	
	newset = (gadget_set *) malloc(sizeof(gadget_set));
	memset((char *) newset, 0, sizeof(gadget_set));
	newset->newEdgeId = MAX_EDGE_UID;
	axhInitTable(&newset->gadgets, 53, newGadgetSize, tossGadgetSize);
	newset->scope = axNewConnScope(scopeName, namelen);
	newset->scope->dconn = wfDeleteConn;
	axRegisterLinkFunc(newset->scope, ALIGN_LINK_TYPE, doAlignLink,
					   copyAlignLinkData, freeAlignLinkData);
	axRegisterConnLocator(newset->scope, ALIGN_LINK_TYPE,
						  gadgetSizeConnLocator);
	
	return((char *) newset);
}

static void
tossDlgGadgetSet(gadgSet, notUsed)
gadget_set *gadgSet;
void *notUsed;
{
	if (gadgSet) {
		axhDeleteTable(&gadgSet->gadgets, (void *) gadgSet->scope);
		axDeleteConnScope(gadgSet->scope);
		free((char *) gadgSet);
	}
}

static char *
newGadgetSize(name, namelen, sizeInfo)
void *name;            /* in: name of new gadget_size instance */
int namelen;
void *sizeInfo;        /* in: data of type (new_gsize *) */
{
	gadget_size *to;
	gadget_size *from;
	new_gsize *gs = (new_gsize *) sizeInfo;
	
	to = (gadget_size *) malloc(sizeof(gadget_size));
	from = &gs->sizeData;
	memset((char *) to, 0, sizeof(gadget_size));
	*to = *from;
	memcpy((void *) &to->gadgId, name, namelen);

	return((char *) to);
}

static void
tossGadgetSize(gadgSize, namescope)
gadget_size *gadgSize;
char *namescope;
{
	axconnC *scope = (axconnC *) namescope;
	if (gadgSize) {
		if (scope && gadgSize->conn) {
			/* delete this' axconn instance
			 */
			scope->dconn(scope, &gadgSize->gadgId, sizeof(linkname),
						 gadgSize->conn);
		}
		free((char *) gadgSize);
	}
}

/* this is the "conn locator" func for use with
 * gadget_size_t.  Report my address, and the
 * address of my conn instance
 */
static int
gadgetSizeConnLocator(scope, name, ref)
axconnC *scope;   /* scope of dialog box */
linkname *name;   /* the toid member of *name identifies me */
remoteRef *ref;   /* out */
{
	gadget_set *gadgSet;
	gadget_size *gadg;

	/* find the appropriate gadget_set
	 */
	axhGetItemI((void *) scope->name, scope->namelen, &AXDialogGadgetSets,
				(char *) 0, (char **) &gadgSet);

	/* find "name" within that gadget set
	 */
	axhGetItemI((void *) &name->toid, sizeof(int), &gadgSet->gadgets,
				(char *) 0, (char **) &gadg);
	
	ref->rHandle = (char *) gadg;
	ref->rConn = &gadg->conn;
	return TRUE;
}

static char *
copyAlignLinkData(from, to)
char *from;
char **to;
{
	align_link_data *fromdata = (align_link_data *) from;
	align_link_data **todata = (align_link_data **) to;
	
	*todata = (align_link_data *) malloc(sizeof(align_link_data));
	memcpy((char *) (*todata), fromdata, sizeof(align_link_data));
	return ((char *) NULL);
}

static void
freeAlignLinkData(ldata)
char *ldata;    /* (align_link_data *) */
{
	if (ldata) {
		free((char *) ldata);
	}
}

/* %%%%%%%%%% UN-static the xmdlged.c version for use here */
static int
pointInsideRect(pt, r)
p32 *pt;
r32 *r;
{
	if (pt->x > r->origin.x &&
		pt->y > r->origin.y &&
		pt->x < r->origin.x + r->width &&
		pt->y < r->origin.y + r->height) {

		return TRUE;
	}
	else {
		return FALSE;
	}
}

/* %%%%%%%%%% UN-static the xmdlged.c version for use here */
/* %%%%%%%%%% include equality in last test in that version, too */
static int
rectanglesIntersect(r1, r2)
r32 *r1, *r2;
{
	int tmp1, tmp2;
	int maxOx, minCx; /* limits in x direction; max origin, min corner */
	int maxOy, minCy; /* limits in y direction */

	if (r1->origin.x > r2->origin.x) {
		maxOx = r1->origin.x;
	}
	else {
		maxOx = r2->origin.x;
	}
	if ((tmp1 = r1->origin.x + r1->width - 1) <
		(tmp2 = r2->origin.x + r2->width - 1)) {
		minCx = tmp1;
	}
	else {
		minCx = tmp2;
	}
	if (r1->origin.y > r2->origin.y) {
		maxOy = r1->origin.y;
	}
	else {
		maxOy = r2->origin.y;
	}
	if ((tmp1 = r1->origin.y + r1->height - 1) <
		(tmp2 = r2->origin.y + r2->height - 1)) {
		minCy = tmp1;
	}
	else {
		minCy = tmp2;
	}
	return (maxOx <= minCx && maxOy <= minCy);
}

/*
 * : takeChange - update a gadget_size's currentsize
 *                using the passed size_change reference
 */
static void
takeChange(fromCh, toG)
size_change *fromCh;   /* in */
gadget_size *toG;      /* in/out */
{
	p32 corner;
	int itmp;
  
	/* apply whatever's left of the sizechange request
	 */
	corner.x = toG->currentsize.origin.x + toG->currentsize.width;
	corner.y = toG->currentsize.origin.y + toG->currentsize.height;
	toG->currentsize.origin.x += fromCh->leftChange;
	toG->currentsize.origin.y += fromCh->topChange;
	corner.x += fromCh->rightChange;
	corner.y += fromCh->bottomChange;
	itmp = corner.x - toG->currentsize.origin.x;
	if (itmp < 0) {
		itmp = 0;
	}
	toG->currentsize.width = itmp;
	itmp = corner.y - toG->currentsize.origin.y;
	if (itmp < 0) {
		itmp = 0;
	}
	toG->currentsize.height = itmp;
  
	/* check limits; change size only, not position
	 */
	if (toG->currentsize.width < toG->minsize.width)
		toG->currentsize.width = toG->minsize.width;
	if (toG->currentsize.height < toG->minsize.height)
		toG->currentsize.height = toG->minsize.height;

	toG->updateElf = TRUE;
	if (toG->containerTop) {
		/* update container top, too
		 */
		toG->containerTop->updateElf = TRUE;
	}
}

static void
callSizeChange(gadgSet, thisSize)
gadget_set *gadgSet;
gadget_size *thisSize;
{
	size_change change;
	linkname thisLink;
	int container = FALSE;
  
	if (isContainer(thisSize)) {
		containerSize(thisSize); /* update size from edges */
		container = TRUE;
	}
  
	initChange(thisSize, &change);
  
	if (change.topChange || change.bottomChange ||
		change.leftChange || change.rightChange) {
    
		/* the type and edge info is unknown, since this
		 * call is not originating from a link
		 */
		thisLink.type = 0;
		thisLink.toedge = 0;
    
		if (container) {
			/* call the top, that is, the gadget_size representing the
			 * container as a whole.
			 */
			initChange(thisSize, &change);
			thisLink.toid = thisSize->gadgId;
			gadgSet->scope->linkCall(gadgSet->scope, 0, 0,
									 &thisLink, sizeof(linkname),
									 ALIGN_LINK_TYPE, &change);
			
			/* call individual edges
			 */
			if (change.topChange && thisSize->edges[CONTAINER_TOP_EDGE] &&
				thisSize->edges[CONTAINER_TOP_EDGE]->gadgId) {
				
				change.currentsize.height = 1;
				change.leftChange = change.rightChange = 0;
				change.bottomChange = change.topChange;
				if (thisSize->edges[CONTAINER_TOP_EDGE]->conn) {
					thisLink.toid =
						thisSize->edges[CONTAINER_TOP_EDGE]->gadgId;
					gadgSet->scope->linkCall(gadgSet->scope, 0, 0,
											 &thisLink, sizeof(linkname),
											 ALIGN_LINK_TYPE, &change);
				}
				else {
					/* this edge is unconnected; just take change directly
					 */
					takeChange(&change, thisSize->edges[CONTAINER_TOP_EDGE]);
				}
			}
			if (thisSize->edges[CONTAINER_LEFT_EDGE] &&
				thisSize->edges[CONTAINER_LEFT_EDGE]->gadgId) {
				
				containerSize(thisSize); /* re-update size */
				initChange(thisSize, &change);
				if (change.leftChange) {
					change.currentsize.width = 1;
					change.topChange = change.bottomChange = 0;
					change.rightChange = change.leftChange;
					if (thisSize->edges[CONTAINER_LEFT_EDGE]->conn) {
						thisLink.toid =
							thisSize->edges[CONTAINER_LEFT_EDGE]->gadgId;
						gadgSet->scope->linkCall(gadgSet->scope, 0, 0,
												 &thisLink, sizeof(linkname),
												 ALIGN_LINK_TYPE, &change);
					}
					else {
						takeChange(&change,
								   thisSize->edges[CONTAINER_LEFT_EDGE]);
					}
				}
			}
			if (thisSize->edges[CONTAINER_RIGHT_EDGE] &&
				thisSize->edges[CONTAINER_RIGHT_EDGE]->gadgId) {
				
				containerSize(thisSize); /* re-update size */
				initChange(thisSize, &change);
				if (change.rightChange) {
					change.currentsize.origin.x +=
						change.currentsize.width - 1;
					change.currentsize.width = 1;
					change.bottomChange = change.topChange = 0;
					change.leftChange = change.rightChange;
					if (thisSize->edges[CONTAINER_RIGHT_EDGE]->conn) {
						thisLink.toid =
							thisSize->edges[CONTAINER_RIGHT_EDGE]->gadgId;
						gadgSet->scope->linkCall(gadgSet->scope, 0, 0,
												 &thisLink, sizeof(linkname),
												 ALIGN_LINK_TYPE, &change);
					}
					else {
						takeChange(&change,
								   thisSize->edges[CONTAINER_RIGHT_EDGE]);
					}
				}
			}
			if (thisSize->edges[CONTAINER_BOTTOM_EDGE] &&
				thisSize->edges[CONTAINER_BOTTOM_EDGE]->gadgId) {
				
				containerSize(thisSize); /* re-update size */
				initChange(thisSize, &change);
				if (change.bottomChange) {
					change.currentsize.origin.y +=
						change.currentsize.height - 1;
					change.currentsize.height = 1;
					change.leftChange = change.rightChange = 0;
					change.topChange = change.bottomChange;
					if (thisSize->edges[CONTAINER_BOTTOM_EDGE]->conn) {
						thisLink.toid =
							thisSize->edges[CONTAINER_BOTTOM_EDGE]->gadgId;
						gadgSet->scope->linkCall(gadgSet->scope, 0, 0,
												 &thisLink, sizeof(linkname),
												 ALIGN_LINK_TYPE, &change);
					}
					else {
						takeChange(&change,
								   thisSize->edges[CONTAINER_BOTTOM_EDGE]);
					}
				}
			}
		}
		else {
			thisLink.toid = thisSize->gadgId;
			gadgSet->scope->linkCall(gadgSet->scope, 0, 0,
									 &thisLink, sizeof(linkname),
									 ALIGN_LINK_TYPE, &change);
		}
	}
}

static int
isDeformable(gsize)
gadget_size *gsize;
{
	int gtype = gsize->gadgetType;
	
	if (gtype == DLG_LIST_BOX ||
		gtype == DLG_EDIT_BOX ||
		gtype == DLG_TABLE ||
		gtype == DLG_CANVAS ||
		gtype == DLG_TRAY) {
		
		return TRUE;
	}
	else if (gtype == DLG_RECT || gtype == DBOX_GTYPE) {
		if (gsize->containerTop) {
			/* this is an individual edge, which behaves as not deformable
			 */
			return FALSE;
		}
		return TRUE;
	}
	else {
		return FALSE;
	}
}

static int
isContainer(gSize)
gadget_size *gSize;
{
	if (gSize->containerTop) {
		return (FALSE);			/* gSize represents a single container edge */
	}
	return (gSize->gadgetType == DLG_RECT ||
			gSize->gadgId == DBOX_GSIZE_ID);
}

/* : linkId - given the passed linkname, return the
 * gadget_size id appropriate for linking.
 * This func exsists because for container-type gadget_size's,
 * edge-to-edge links are applied not to the container
 * gadget_size itself, but to one of the edge-specific
 * gadget_size's associated with the container
 */
static int                                 /* return id */
linkId(gadgSet, lname, simpleId)
gadget_set *gadgSet;                       /* in */
linkname *lname;                           /* in */
int simpleId;                              /* in */
{
	gadget_size *gsize;
	int uid = 0;

	/* lookup gadget_size named by simpleId
	 */
	axhGetItemI((void *) &simpleId, sizeof(int),
				&gadgSet->gadgets, (char *) NULL, (char **) &gsize);
	
	if (gsize && lname->type == EDGE_TO_EDGE && isContainer(gsize)) {
		/* convert simpleId to that of one of the edge-specific
		 * gadget_size's stored on gsize->edges.
		 */
		switch (lname->toedge) {
			
		case OUTER_LEFT:
		case INNER_LEFT:
			uid = gsize->edges[CONTAINER_LEFT_EDGE]->gadgId;
			break;
			
		case OUTER_RIGHT:
		case INNER_RIGHT:
			uid = gsize->edges[CONTAINER_RIGHT_EDGE]->gadgId;
			break;
			
		case OUTER_TOP:
		case INNER_TOP:
			uid = gsize->edges[CONTAINER_TOP_EDGE]->gadgId;
			break;
			
		case OUTER_BOTTOM:
		case INNER_BOTTOM:
			uid = gsize->edges[CONTAINER_BOTTOM_EDGE]->gadgId;
			break;
			
		}
		if (!uid)
			uid = simpleId;
	}
	else {
		uid = simpleId;
	}
	return (uid);
}

/* : containerSize - update the passed container-type
 * gadget_size's currentsize member, by reading the
 * current sizes of the edge-specific gadget_size's on gsize->edges
 */
static void
containerSize(gsize)
gadget_size *gsize;  /* in, expected to be of container type */
{
	gadget_size *esize;
	
	if (!gsize)
		return;
	
	if (gsize->edges[CONTAINER_LEFT_EDGE]) {
		esize = gsize->edges[CONTAINER_LEFT_EDGE];
		gsize->currentsize.origin.x = esize->currentsize.origin.x;
	}
	if (gsize->edges[CONTAINER_TOP_EDGE]) {
		esize = gsize->edges[CONTAINER_TOP_EDGE];
		gsize->currentsize.origin.y = esize->currentsize.origin.y;
	}
	if (gsize->edges[CONTAINER_RIGHT_EDGE]) {
		esize = gsize->edges[CONTAINER_RIGHT_EDGE];
		gsize->currentsize.width = esize->currentsize.origin.x -
			gsize->currentsize.origin.x + 1;
	}
	if (gsize->edges[CONTAINER_BOTTOM_EDGE]) {
		esize = gsize->edges[CONTAINER_BOTTOM_EDGE];
		gsize->currentsize.height = esize->currentsize.origin.y -
			gsize->currentsize.origin.y + 1;
	}
}

static void
initChange(gsize, change)
gadget_size *gsize;          /* in */
size_change *change;         /* out: init caller's memory */
{
	memset((char *) change, 0, sizeof(size_change));
	change->callerType = gsize->gadgetType;
	change->callerMinorType = gsize->gadgetMinorType;
	change->leftChange = gsize->proposedsize.origin.x -
		gsize->currentsize.origin.x;
	change->topChange = gsize->proposedsize.origin.y -
		gsize->currentsize.origin.y;
	change->rightChange = (gsize->proposedsize.origin.x +
						   gsize->proposedsize.width) -
							   (gsize->currentsize.origin.x +
								gsize->currentsize.width);
	change->bottomChange = (gsize->proposedsize.origin.y +
							gsize->proposedsize.height) -
								(gsize->currentsize.origin.y +
								 gsize->currentsize.height);
	change->currentsize = gsize->currentsize;
}

/* : insertContainerEdges - this sub is used for inserting into gadgSet
 * the gadget_size's representing a container's edges.  this is done
 * after first inserting the "top", or overall container gadget_size.
 */
static void
insertContainerEdges(gadgSet, newGadg, gsize)
gadget_set *gadgSet;
new_gsize *newGadg;        /* info which was used to insert the container */
gadget_size *gsize;        /* ref to newly inserted container */
{
	/* add four more gadget_sizes to gadgSet, representing
	 * the edges of the container.  these edge gadget_size's aren't
	 * known to the elf layer, so they use out-of-band names above
	 * MAX_GADGET_UID.
	 */
	new_gsize edge;
	gadget_size *gtmp;
	
	edge = *newGadg;
	
	/* left edge
	 */
	edge.sizeData.gadgId = gadgSet->newEdgeId--; /* get new name */
	edge.sizeData.currentsize.width = 1;
	edge.sizeData.proposedsize = edge.sizeData.currentsize;
	edge.sizeData.minsize.width = 1;
	edge.sizeData.fixedWidth = TRUE;
	axhGetItemI((void *) &edge.sizeData.gadgId, sizeof(int), &gadgSet->gadgets,
				(char *) &edge, (char **) &gtmp); /* store */
	gsize->edges[CONTAINER_LEFT_EDGE] = gtmp;  /* point from top to edge */
	gtmp->containerTop = gsize;                /* point from edge to top */
	addContainerEdgeLink(gadgSet->scope,
						 gtmp, CONTAINER_LEFT_EDGE); /* link edge to top */
	
	/* top edge
	 */
	edge.sizeData = newGadg->sizeData;
	edge.sizeData.gadgId = gadgSet->newEdgeId--;
	edge.sizeData.currentsize.height = 1;
	edge.sizeData.proposedsize = edge.sizeData.currentsize;
	edge.sizeData.minsize.height = 1;
	edge.sizeData.fixedHeight = TRUE;
	axhGetItemI((void *) &edge.sizeData.gadgId, sizeof(int), &gadgSet->gadgets,
				(char *) &edge, (char **) &gtmp);
	gsize->edges[CONTAINER_TOP_EDGE] = gtmp;
	gtmp->containerTop = gsize;
	addContainerEdgeLink(gadgSet->scope, gtmp, CONTAINER_TOP_EDGE);
	
	/* right edge
	 */
	edge.sizeData = newGadg->sizeData;
	edge.sizeData.gadgId = gadgSet->newEdgeId--;
	edge.sizeData.currentsize.origin.x += gsize->currentsize.width - 1;
	edge.sizeData.currentsize.width = 1;
	edge.sizeData.proposedsize = edge.sizeData.currentsize;
	edge.sizeData.minsize.width = 1;
	edge.sizeData.fixedWidth = TRUE;
	axhGetItemI((void *) &edge.sizeData.gadgId, sizeof(int), &gadgSet->gadgets,
				(char *) &edge, (char **) &gtmp);
	gsize->edges[CONTAINER_RIGHT_EDGE] = gtmp;
	gtmp->containerTop = gsize;
	addContainerEdgeLink(gadgSet->scope, gtmp, CONTAINER_RIGHT_EDGE);
	
	/* bottom edge
	 */
	edge.sizeData = newGadg->sizeData;
	edge.sizeData.gadgId = gadgSet->newEdgeId--;
	edge.sizeData.currentsize.origin.y += gsize->currentsize.height - 1;
	edge.sizeData.currentsize.height = 1;
	edge.sizeData.proposedsize = edge.sizeData.currentsize;
	edge.sizeData.minsize.height = 1;
	edge.sizeData.fixedHeight = TRUE;
	axhGetItemI((void *) &edge.sizeData.gadgId, sizeof(int), &gadgSet->gadgets,
				(char *) &edge, (char **) &gtmp);
	gsize->edges[CONTAINER_BOTTOM_EDGE] = gtmp;
	gtmp->containerTop = gsize;
	addContainerEdgeLink(gadgSet->scope, gsize, CONTAINER_BOTTOM_EDGE);
}

/* : normalizePositions - test whether the dialog box's
 * gadget_size origin has gone nonzero; if so, apply
 * opposite to all gadget_sizes
 */
static void
normalizePositions(gadgSet)
gadget_set *gadgSet;
{
	int gid, ix;
	hitem *list;
	gadget_size *gsize;
	p32 offset;
	
	gid = DBOX_GSIZE_ID;
	gsize = (gadget_size *) NULL;
	axhGetItemI((void *) &gid, sizeof(int), &gadgSet->gadgets,
				(char *) NULL, (char **) &gsize);
	if (!gsize)
		return;
	
	containerSize(gsize);                /* update from edges */
	
	if (gsize && (gsize->currentsize.origin.x ||
				  gsize->currentsize.origin.y)) {
		offset = gsize->currentsize.origin;
		for (ix = 0; ix < gadgSet->gadgets.width; ix++) {
			list = gadgSet->gadgets.table[ix];
			while(list) {
				gsize = (gadget_size *) list->value;
				gsize->currentsize.origin.x -= offset.x;
				gsize->currentsize.origin.y -= offset.y;
				gsize->updateElf = TRUE;
				list = list->next;
			}
		}
	}
}

/* : addContainerEdgeLink - make an edge-to-edge link from passed edge
 * gadget_size to its container top gadget_size.  These links
 * carry external size changes on the edges to the top, in case
 * the top holds rowcol links from actual containees.  These are the
 * only edge-to-edge links attached to container tops, so the tops
 * may assume that calls across these links come from their own edges.
 */
static void
addContainerEdgeLink(scope, edge, which_edge)
axconnC *scope;
gadget_size *edge;
int which_edge;
{
	linkname linkFrom, linkTo;
	align_link_data linkData;
	
	if (!(edge && edge->containerTop)) {
		return;
	}

	linkFrom.type = linkTo.type = EDGE_TO_EDGE;
	linkFrom.toid = edge->gadgId;
	linkTo.toid = edge->containerTop->gadgId;
	switch (which_edge) {
	case CONTAINER_LEFT_EDGE:
		linkFrom.toedge = linkTo.toedge = OUTER_LEFT;
		linkData.toEdge = linkData.fromEdge = OUTER_LEFT;
		break;
	case CONTAINER_TOP_EDGE:
		linkFrom.toedge = linkTo.toedge = OUTER_TOP;
		linkData.toEdge = linkData.fromEdge = OUTER_TOP;
		break;
	case CONTAINER_RIGHT_EDGE:
		linkFrom.toedge = linkTo.toedge = OUTER_RIGHT;
		linkData.toEdge = linkData.fromEdge = OUTER_RIGHT;
		break;
	case CONTAINER_BOTTOM_EDGE:
		linkFrom.toedge = linkTo.toedge = OUTER_BOTTOM;
		linkData.toEdge = linkData.fromEdge = OUTER_BOTTOM;
		break;
	}
	
	linkData.type = EDGE_TO_EDGE;
	linkData.minDistance = linkData.maxDistance = 0;
	
	scope->link(scope, ALIGN_LINK_TYPE,
				&linkFrom, &linkTo, sizeof(linkname),
				(char *) &linkData, TRUE);
}

/* : wfDeleteConn - this is a hack-overload of
 * (axconn *)->dconn(), done because for align
 * links, the link names aren't one-to-one with
 * the linked-object names; instead, they're
 * derived from the linked-object names by
 * including align-type and to-edge info. We
 * assume here that gadget_size's are linked
 * using only one link type: ALIGN_LINK_TYPE
 */
static void
wfDeleteConn(scope, name, namelen, tconn)
axconnC *scope;
void *name;      /* name of containing object */
int namelen;     /* length of name in bytes */
axconn *tconn;   /* object's conn handle */
{
	int ix, jx;
	axhtab *tabl;
	hitem *list;
	linkname lname;
	axlink *tlink;
	align_link_data *tdata;

	if (tconn) {
		/* destruct links first
		 */
		tabl = &tconn->linkTables[0];
		for (jx = 0; jx < tabl->width; jx++) {
			list = tabl->table[jx];
			while(list) {
				if (list->value && list->name) {
					
					tlink = (axlink *) list->value;
					tdata = (align_link_data *) tlink->data;
					lname.type = tdata->type;
					memcpy((void *) &lname.toid, name, sizeof(int));
					if (tlink->direction) {
						lname.toedge = tdata->fromEdge;
					}
					else {
						lname.toedge = tdata->toEdge;
					}
					
					/* use the (reciprocal) deleteLink here
					 */
					scope->dlink(scope, ALIGN_LINK_TYPE, (void *) &lname,
								 list->name, namelen, TRUE);
					
					list->value = (char *) NULL;
				}
				list = list->next;
			}
		}
		axhDeleteTable(&tconn->linkTables[0], (void *) NULL);
		free((char *) tconn->linkTables);

		axhDeleteTable(&tconn->typeTable, (void *) NULL);
		axhDeleteTable(&tconn->evTable, (void *) NULL);
		free((char *) tconn);
	}
}

/* The intent of this function is to measure the passed
 * text string using the font that its dialog box was layed
 * out in.  The passed info arrives as <<reference font><text string>>
 *
 * There are several holes here.... 1) the string being measured may
 * have been layed out in some specific font attribute, not the overall
 * dialog box font (represented by the reference font).  This is ignored..
 * 2)  The reference font is just a character cell description, and so
 * doesn't really have enough information to do a string measurement..
 * In place of reference_font should be a font attribute description
 * of family, size, weight, slant.  But even then, often the font that
 * was used for layout isn't available on the system that's displaying
 * the dbox, the typical case of this being UNIX axsab vs. NT MS_SansSerif --
 * Neither of these fonts will look-up on the other OS.
 *
 * Currently....
 * This func implements the NT-case of the java-dialog$ sub "layout_width$",
 * used for adjusting the alignment of various widgets so that they hit the
 * java client in much the same alignment as occurs on the (unix) workstation.
 * For the anyware-axmain running on NT, it uses cribbed axsab widths for the
 * measurement.
 */
#if AXMS
static XFontStruct *AxsabMetrics = 0;   /* the axsab metrics go here */
#endif

static data_ptr
xmLayoutStringWidth(info)
data_ptr info;
{
#if AXMS
	mbchar *tstr;
	if (!info) {
		return AxMakeIntData(0);
	}
	tstr = AxStrFromDataPtr(info);
	if (!tstr) {
		return AxMakeIntData(0);
	}
	if (0 == AxsabMetrics) {
		AxsabMetrics = xmInitAxsabMetrics();
	}
	return (AxMakeIntData(AXTextWidth(AxsabMetrics, tstr, MBstrlen(tstr))));
#endif
}

#if AXMS

static unsigned char axsab_widths[] = {
	 4,  4,  5,  7,  7, 12,  9,  3,  4,  4,  5,  8,  4,  4,  4,  4,  /* / */
	 7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  4,  4,  8,  8,  8,  7,  /* ? */
	14,  9,  9, 10, 10,  9,  8, 10, 10,  4,  7,  9,  7, 11, 10, 10,  /* O */
	 9, 10, 10,  9,  8, 10, 10, 13,  9,  9,  8,  5,  4,  5,  6,  7,  /* _ */
	 3,  7,  7,  7,  7,  7,  4,  7,  7,  3,  3,  7,  3, 11,  7,  7,  /* o */
	 7,  7,  5,  7,  4,  7,  8, 10,  7,  7,  7,  4,  3,  4,  8,  4,  /* ~ */
	 4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  /* err */
	 4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  /* err */
	 4,  4,  7,  7,  7,  9,  4,  7,  5, 12,  5,  7,  9,  9, 12,  5,  /* 8bit */
	 6,  8,  5,  5,  5,  9,  7,  4,  5,  4,  5,  7, 12, 12, 12,  7,
	 9,  9,  9,  9,  9,  9, 13, 10,  9,  9,  9,  9,  4,  4,  4,  4,
	11, 10, 10, 10, 10, 10, 10,  9, 10, 10, 10, 10, 10, 10, 10,  8,
	 7,  7,  7,  7,  7,  7, 13,  7,  7,  7,  7,  7,  4,  4,  4,  4,
	 9,  7,  7,  7,  7,  7,  7,  9,  8,  7,  7,  7,  7,  8,  9,  7,
};

static XFontStruct *
xmInitAxsabMetrics()
{
	int ix;
	XFontStruct *fs;
    fs = (XFontStruct *) malloc(sizeof(XFontStruct));
    memset((char *) fs, 0, sizeof(XFontStruct));
    fs->min_char_or_byte2 = (unsigned int) 32;
    fs->max_char_or_byte2 = (unsigned int) 255;
    fs->default_char = (unsigned int) 32;
    fs->min_bounds.width = 3;
    fs->min_bounds.ascent = 13;
    fs->min_bounds.descent = 3;
	fs->min_bounds.rbearing = 3;
    fs->max_bounds.width = 14;
    fs->max_bounds.ascent = 13;
    fs->max_bounds.descent = 3;
	fs->max_bounds.rbearing = 14;
    fs->ascent = 13;
    fs->descent = 3;
	fs->per_char = (XCharStruct *) malloc (224 * sizeof(XCharStruct));
	memset((void *) fs->per_char, 0, (224 * sizeof(XCharStruct)));
	for (ix = 0; ix < 224; ix++) {
		
		fs->per_char[ix].width = fs->per_char[ix].rbearing = axsab_widths[ix];
		
		/* the per_char ascent, descent, sidebearings are fictional;
		 * the critical numbers are the widths.
		 */
		fs->per_char[ix].ascent = 13;
		fs->per_char[ix].descent = 3;
	}
	return (fs);
}
#endif

/* This sub used by xmDlgClipboardOp
 */
static void
xmTextInsert(ww, strP, prepend_space)
AXWidget ww;
data_ptr strP;
int prepend_space;
{
#if AX_X11
	mbchar *mtmp;
	wchar *wtmp;
	int textlen, ix;
	wchar nl[2];
	
	if (!strP)
		return;
	
	mtmp = AxStrFromDataPtr(strP);
	textlen = MBstrlen(mtmp);
	wtmp = (wchar *) malloc((textlen + 1) * sizeof(wchar));
	textlen = MbToWStr(mtmp, wtmp);

	/* do a cheapie in-place conversion of non-graphic characters
	 * to spaces.  'cheap' means high-range NGCs not tested for.
	 */
	if (W_EDITBOX == ww->type) {           /* for editboxes, newlines ok */
		for (ix = 0; ix < textlen; ix++) {
			if (wtmp[ix] < 20 && wtmp[ix] != '\n') {
				wtmp[ix] = (wchar) ' ';
			}
		}
	}
	else {
		for (ix = 0; ix < textlen; ix++) {
			if (wtmp[ix] < 20) {
				wtmp[ix] = (wchar) ' ';
			}
		}
	}

	/* do the insertion
	 */
	if (prepend_space) {
		nl[0] = (wchar ) ' ';
		nl[1] = (wchar ) 0;
	}
	switch (ww->type) {
	case W_ENTER:
		if (prepend_space) {
			EntInsertText(ww, nl, 1);
		}
		EntInsertText(ww, wtmp, textlen);
		break;
	case W_COMBO_BOX:
		if (prepend_space) {
			EditWinInsertText(ww, GetEditWin(ww), nl);
		}
		EditWinInsertText(ww, GetEditWin(ww), wtmp);
		break;
	case W_EDITBOX:
		if (prepend_space) {
			manageInsert(ww, nl, 1);
		}
		manageInsert(ww, wtmp, textlen);
		break;
	}
	free(wtmp);
#endif
}

void
xmInsertElfSelection(tw, strsP)
AXWidget tw;
data_ptr strsP;
{
	int ix, jx, asize, bsize;
	data_ptr subP;
	
	if (!strsP) {
		return;
	}
	
	/* can arrive as: 1) a plain string datum,
	 * 2) an array of strings, or 3) an array
	 * of arrays of strings.
	 *
	 * W_EDITBOX is treated differently than the
	 * single-line widgets: form 2) appears with
	 * trailing newlines, which editbox consumes
	 * rather than prepending a space on the next line.
	 * form 3) lacks trailing newlines, so for
	 * editbox, they're appended at each non-last row-end.
	 */
	if (AxIsString(strsP)) {
		xmTextInsert(tw, strsP, FALSE);      /* form 1) */
	}
	else {
		bsize = AxArraySize(strsP);
		for (jx = 0; jx < bsize; jx++) {
			if (AxIsString(AxArrayElement(strsP, jx))) {
				xmTextInsert(tw, AxArrayElement(strsP, jx),
							 tw->type != W_EDITBOX && jx > 0);   /* form 2) */
			}
			else {
				subP = AxArrayElement(strsP, jx);
				asize = AxArraySize(subP);
				for (ix = 0; ix < asize; ix++) {
					if (AxIsString(AxArrayElement(subP, ix))) {
						xmTextInsert(tw, AxArrayElement(subP, ix),
									 (tw->type != W_EDITBOX && jx > 0) ||
									 ix > 0);
						if (W_EDITBOX == tw->type && ix == asize - 1 &&
							jx < bsize - 1) {
							xmTextInsert(tw, AxMakeStrData(-1, "\n"), FALSE);
						}
					}
				}
			}
		}
	}
}

static data_ptr
xmDlgClipboardOp(info)
data_ptr info;
{
#if AX_X11
	AXWidget tw;
	AXWidget_dialog dw;
	struct edit_window *ee;
	data_ptr strsP;
	int opcode;
				
	if (!info) {
		return ((data_ptr) NULL);
	}
	
	dw = (AXWidget_dialog) NULL;
	tw = (AXWidget) WidgetPointerFromHandleData(AxArrayElement(info, 1));
	if (!tw) {
		return ((data_ptr) NULL);
	}
	if (W_DIALOG == tw->type) {
		dw = (AXWidget_dialog) tw;
	}
	else if (W_MAIN == tw->type && ((AXWidget_main)tw)->dialog) {
		dw = (AXWidget_dialog) ((AXWidget_main)tw)->dialog;
	}
	
	opcode = AxIntFromArray(info, 0);
	
	tw = xmKeyFocusWidget(dw);
	if (!tw) {
		return ((data_ptr) NULL);
	}
	else {
		switch (opcode) {
		case CUTCOPYPASTE_CUT:
			if (W_ENTER == tw->type) {
				EntDoPendingDelete(tw, TRUE);
			}
			else {
				ee = GetEditWin(tw);
				if (ee) {
					EditWinDoPendingDelete(ee);
					tw->update_widget = WidgetNeedsUpdated = TRUE;
				}
			}
			break;
		case CUTCOPYPASTE_GET_TEXT:
			ee = GetEditWin(tw);
			if (ee) {
				wchar *wtmp = EditWinGetSelectedText(ee);
				if (wtmp) {
					data_ptr retstr = AxMakeStrData(-1, WToMbOneStr(wtmp));
					free(wtmp);
					return retstr;
				}
			}
			else {
				return ((data_ptr) NULL);
			}
			break;
		case CUTCOPYPASTE_PASTE:
			strsP = AxArrayFromArray(info, 2);  /* selection data */
			xmInsertElfSelection(tw, strsP);
			break;
		}
	}
#endif
	
	return ((data_ptr) NULL);
}

/* : xmReadWidgetSubAccelerators - report the list of "sub" accelerator
 * characters for the widget.  This is applicable only to widgets of type
 * W_RADIO and tabcontrol-style W_RECT.  Both widget types appear at the
 * elf level as single entities, but contain parts that can be individually
 * activated.  For W_RADIO, it's the set of radio-toggle widgets.  For
 * tabcontrol-style W_RECT, it's the set of tabbed layers.
 */
static data_ptr                      /* <<id1><accel1>><<id2><accel2>, ... */
xmReadWidgetSubAccelerators (info)
data_ptr info;                       /* <widget handle> */
{
	AXWidget tw;
	mbchar accel_str[2];
	data_ptr tpair;
	data_ptr ret = (data_ptr) NULL;
	
	if (!info) {
		return ((data_ptr) NULL);
	}
	tw = (AXWidget) WidgetPointerFromHandleData(AxArrayElement(info, 0));
	if (tw) {
		switch (tw->type) {
		case W_RADIO:
			if (tw->auxData) {
				wComposite *cp;
				data_ptr accels = AxMakeArray(1);
				int ax = 0;
				int any_toggles = FALSE;
				accel_str[1] = 0;
				for (cp = tw->auxData->compositeList; cp; cp = cp->next) {
					if (cp->type == WLINKDOWN && W_TOGGLE == cp->w->type) {
						any_toggles = TRUE;
						tpair = AxMakeArray(2);
						tpair = AxAddStrToArray(tpair, 0, cp->w->name);
						accel_str[0] = (mbchar) cp->w->accelerator_key;
						tpair = AxAddStrToArray(tpair, 1, accel_str);
						accels = AxAddArrayToArray(accels, ax++, tpair);
					}
				}
				if (!any_toggles) {
					AxFreeData(accels);
				}
				else {
					ret = accels;
				}
			}
			break;
		case W_RECT:
			if (RECT_LAYERED == tw->style) {
				AXWidget_rect rw = (AXWidget_rect) tw;
				if (rw->tabctl.num_layers > 0) {
					int rx;
					mbchar *stmp;
					data_ptr accels = AxMakeArray(1);
					accel_str[1] = 0;
					for (rx = 0; rx < rw->tabctl.num_layers; rx++) {
						tpair = AxMakeArray(2);
						stmp = rw->tabctl.layerlist[rx].layerName;
						accel_str[0] = 
							(mbchar) rw->tabctl.layerlist[rx].accelerator_key;
						tpair = AxAddStrToArray(tpair, 0, stmp);
						tpair = AxAddStrToArray(tpair, 1, accel_str);
						accels = AxAddArrayToArray(accels, rx, tpair);
					}
					ret = accels;
				}
			}
			break;
		}
	}
	return ret;
}


