/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: widgsiz.c
   Applix ASC version: 33
   Date: Mon Nov 16 13:53:45 1998
*/
#include <aos.h>
#include <xmdlged.h>
#include <elxwins.h>

#if AXMS
extern int Scrollbar_Thickness;
#define SCROLLBAR_THICKNESS Scrollbar_Thickness   /* variable on Windows */
#else
#define SCROLLBAR_THICKNESS 16  /* motif */
extern int CbComboBaselineOffset();
#endif

extern XFontStruct *AxDlgMainFont;

extern data_ptr AxArrayElement(), AxAddIntToArray();
extern mbchar *AxStrPtrFromArray();
extern int AxArraySize(), AxIntFromArray();
extern int wColorButtonDefaultSize(), wOptionHeight();
extern int wListPixWidth(), wListPixHeight();
extern int wEnterPixWidth(), BasicStrlen();
extern int wListColWidth(), wListVisibleItemCount();
extern int AXTextWidth(), wComboHeight();
extern int centerButtonText();
extern void wLabelSize(), wToggleButtonSize(), wEnterPixHeight();
extern void wPixButtonSize(), wVerbButtonSize();
extern void wRadioButtonSize(), wOptionSize(), wIconSize();
extern XFontStruct *lookupNamedFont(), *lookupReferenceFont();

int dlgReadSizeValue(), wElfGadgetSize(), wMinimumGadgetSize();
void dlgWriteSizeValue(), wActualGadgetSize();

static void scrollbarSizes();
static void gadgetGeomAtts();


/* : wActualGadgetSize - predict the size and position
 * in screenpix of an elf-visible gadget, using only the
 * elf-level information (i.e., without widget_t data)
 */
void
wActualGadgetSize(dboxRev, dboxRefFont, dboxNoExpand,
                  gadgetType, gadgetSizeAtts, actualSize)

int dboxRev;               /* in: revstamp of containing dbox */
XFontStruct *dboxRefFont;  /* in: reference font */
int dboxNoExpand;          /* in: !whether box does "bigfont" expansion */
int gadgetType;            /* in: elf gadget type, e.g, DLG_TABLE */
data_ptr gadgetSizeAtts;   /* in: ordered array of gadget atts re: geometry */
r32 *actualSize;           /* out: size-as-realized by widgetkit */
{
    data_ptr widgetShadow, labelShadow, textShadow;
    data_ptr atmp, strs, fontAttr, labelFont, widgetAlignment;
    int expos, eypos, ewidth, eheight, size, horiz_scroll;
    int iwidth, iheight, ixtmp, iytmp, ix, hide_vscroll;
    int thickness, pixType, titleType, buttonType, expand;
    int widest, pulldownWidth, barwid, barspc, rev;
    XFontStruct *fs, *labelFs, *refFont, *tfs;
    mbchar *title, *stmp;

    expand = !dboxNoExpand;
    
    gadgetGeomAtts(gadgetSizeAtts, &title, &expos, &eypos, &ewidth,
                   &eheight, &strs, &titleType, &horiz_scroll, &thickness,
                   &buttonType, &hide_vscroll, &widgetShadow, &textShadow,
                   &fontAttr, &labelFont, &labelShadow, &widgetAlignment);
    
    fs = lookupNamedFont(fontAttr);
    labelFs = lookupNamedFont(labelFont);
    
    /* this is the dialog main font at the time the dbox was last saved
     */
    refFont = dboxRefFont;
    
    /* the elf and c positions are identical for most types
     */
    actualSize->origin.x = expos;
    actualSize->origin.y = eypos;

    /* fill out the display width/height of the gadget,
     * and modify position if it differs from elf's
     */
    switch(gadgetType) {
        
    case DLG_TEXT:
        wLabelSize(title, labelFs, labelShadow, &iwidth, &iheight);
        actualSize->width = (unsigned int) iwidth;
        actualSize->height = (unsigned int) iheight;
        break;
        
    case DLG_TOGGLE_BUTTON:
        wToggleButtonSize(title, labelFs, &iwidth, &iheight);
        actualSize->width = (unsigned int) iwidth;
        actualSize->height = (unsigned int) iheight;
        break;
        
    case DLG_RADIO_BOX:
        /* the radio group is sized as a stack of
         * radio-flavored toggles, plus an optional label
         */
        iwidth = iheight = 0;
        if (title && *title) {
            wLabelSize(title, labelFs, labelShadow, &iwidth, &iheight);
        }
        size = AxArraySize(strs);
        for (ix = 0; ix < size; ix++) {
            stmp = AxStrPtrFromArray(strs, ix);  /* button label */
            wRadioButtonSize(stmp, fs, &ixtmp, &iytmp);
            if (ixtmp > iwidth) {
                iwidth = ixtmp;   /* widest string sets overall width */
            }
            iheight += iytmp;
        }
        actualSize->width = (unsigned int) iwidth;
        actualSize->height = (unsigned int) iheight;
        break;
        
    case DLG_VERB_BUTTON:
        pixType = buttonType == DLG_ICON_BUTTON;
        
        /* the following section reflects the wkit's sizing of
         * the various verb button types.  The basic theme is to
         * compute height and width only when the elf sizes are
         * zero.  But: 1) pix buttons always ignore elf-stored
         * size info, and depend on pixmapper-lookup to determine
         * size -- the pixmapper should be fixed so that this
         * lookup is reliable
         */
        if (ewidth > 0) {
            if (expand) {
                iwidth = axDlgScaleX(ewidth, refFont);
            }
        }
        if (eheight > 0) {
            if (expand) {
                iheight = axDlgScaleY(eheight, refFont);
            }
        }
        wVerbButtonSize(title, pixType, labelFs, thickness,
                        widgetShadow, textShadow, &iwidth, &iheight);
        actualSize->width = (unsigned int) iwidth;
        actualSize->height = (unsigned int) iheight;
        break;
        
    case DLG_CHOICE:
        size = AxArraySize(strs);
        for (ix = 0, widest = 0; ix < size; ix++) {
            stmp = AxStrPtrFromArray(strs, ix);  /* button label */
            tfs = fs ? fs : AxDlgMainFont;
            ixtmp = AXTextWidth(tfs, stmp, strlen(stmp));
            if (ixtmp > widest) {
                widest = ixtmp;
            }
        }
        wOptionWidth(title, labelFs, widest, &pulldownWidth,
                     &barwid, &barspc, &iwidth);
        actualSize->width = (unsigned int) iwidth;
        actualSize->height =
            (unsigned int) wOptionHeight(labelFs, fs, labelShadow,
                                         textShadow, &ixtmp, &iytmp);
        break;
        
    case DLG_LIST_BOX:
        iwidth = wListPixWidth(ewidth, dboxRev, refFont);
        iheight = wListPixHeight(eheight, fs, textShadow, horiz_scroll);
        scrollbarSizes(horiz_scroll, hide_vscroll, &ixtmp, &iytmp);
        actualSize->width = (unsigned int) iwidth + ixtmp;
        actualSize->height = (unsigned int) iheight + iytmp;
        break;
        
    case DLG_ENTER:
        iwidth = wEnterPixWidth(ewidth, fs, textShadow);
        wEnterPixHeight(fs, textShadow, &ixtmp, &iheight);
        wLabelSize(title, labelFs, labelShadow, &ixtmp, &iytmp);
        if (titleType == LABEL_LEFT) {
            /* total height is same as enterfield;
             * total width includes title width plus 3pix pad
             * (cribbed from elwincom.c)
             */
            iwidth += ixtmp + 3;
        }
        else if (titleType == LABEL_TOP) {
            /* total height includes title plus 4pix pad;
             * total width reflects title if longer than field
             */
            iheight += iytmp + 4;
            if (ixtmp > iwidth)
                iwidth = ixtmp;
        }
        actualSize->width = (unsigned int) iwidth;
        actualSize->height = (unsigned int) iheight;
        break;
        
    case DLG_ICON:
        wIconSize(title, ewidth, eheight, &iwidth, &iheight);
        actualSize->width = (unsigned int) iwidth;
        actualSize->height = (unsigned int) iheight;
        break;
        
    case DLG_RECT:
    case DLG_SCALE:
        actualSize->width = (unsigned int) ewidth;
        actualSize->height = (unsigned int) eheight;
        break;
        
    case DLG_TABLE:
    case DLG_CANVAS:
    case DLG_TRAY:
    case DLG_EDIT_BOX:
        scrollbarSizes(horiz_scroll, hide_vscroll, &ixtmp, &iytmp);
        actualSize->width = (unsigned int) ewidth + SCROLLBAR_THICKNESS;
        actualSize->height = (unsigned int) eheight + iytmp;
        break;
        
    case DLG_SPLITTER:
        iwidth = ewidth;
        if (iwidth < 4)
            iwidth = 4;
        actualSize->width = (unsigned int) iwidth;
        actualSize->height = (unsigned int) iwidth;
        break;
        
    case DLG_ROWCOL:
        /* for now, just use common default of 20x20
         */
        actualSize->width = (unsigned int) 20;
        actualSize->height = (unsigned int) 20;
        break;
        
    case DLG_COMBO_BOX:
        actualSize->width = (unsigned int) iwidth;
        actualSize->height = (unsigned int) wComboHeight(fs);
        break;
    }
}

/* : wAlignmentOffset - return a vector representing the offset
 * between the widget's "alignment origin" and the upper-left
 * corner of the widget's footprint, in pixels.  The vector points
 * from the alignment origin to the upper-left corner.
 *
 * The aligment offset is very widget-type-specific, and relies on
 * knowledge of specific painter-code offsets, so you'll see some of
 * that code replicated here..
 */
void
wAlignmentOffset(dboxRev, dboxRefFont, dboxNoExpand,
                 gadgetType, gadgetSizeAtts, offset)

int dboxRev;               /* in: revstamp of containing dbox */
XFontStruct *dboxRefFont;  /* in: reference font */
int dboxNoExpand;          /* in: !whether box does "bigfont" expansion */
int gadgetType;            /* in: elf gadget type, e.g, DLG_TABLE */
data_ptr gadgetSizeAtts;   /* in: ordered array of gadget atts re: geometry */
p32 *offset;               /* out */
{
    data_ptr widgetShadow, labelShadow, textShadow;
    data_ptr atmp, strs, fontAttr, labelFont, widgetAlignment;
    int expos, eypos, ewidth, eheight, size, horiz_scroll;
    int iwidth, iheight, ixtmp, iytmp, ix, hide_vscroll;
    int horiz_align, vert_align, charHeight, charAscent;
    int thickness, pixType, titleType, buttonType;
    int widest, pulldownWidth, barwid, barspc, rev;
    XFontStruct *fs, *labelFs, *refFont, *tfs;
    mbchar *title, *stmp;
    r32 actualSize;

    gadgetGeomAtts(gadgetSizeAtts, &title, &expos, &eypos, &ewidth,
                   &eheight, &strs, &titleType, &horiz_scroll, &thickness,
                   &buttonType, &hide_vscroll, &widgetShadow, &textShadow,
                   &fontAttr, &labelFont, &labelShadow, &widgetAlignment);
    
    fs = lookupNamedFont(fontAttr);
    labelFs = lookupNamedFont(labelFont);
    
    tfs = labelFs ? labelFs : AxDlgMainFont;
    charHeight = tfs->ascent + tfs->descent;
    charAscent = tfs->ascent;
    
    offset->x = offset->y = 0;
    horiz_align = AxIntFromArray(widgetAlignment, 0);
    vert_align = AxIntFromArray(widgetAlignment, 1);

    /* start with overall pix footprint, stored on actualSize
     */
    wActualGadgetSize(dboxRev, dboxRefFont, dboxNoExpand,
                      gadgetType, gadgetSizeAtts, &actualSize);
    
    /* now init offset based on alignment flavor
     */
    switch(gadgetType) {
        
    case DLG_TEXT:
    case DLG_TOGGLE_BUTTON:
        switch(horiz_align) {
        case ALIGN_RIGHT:
            offset->x -= actualSize.width;
            break;
        case ALIGN_CENTER:
            offset->x -= actualSize.width / 2;
            break;
        }
        switch(vert_align) {
        case ALIGN_BOTTOM:
            offset->y -= charHeight;
            break;
        case ALIGN_CENTER:
            offset->y -= charHeight / 2;
            break;
        case ALIGN_BASELINE:
            offset->y -= charAscent;
            break;
        }
        break;
        
    case DLG_RADIO_BOX:
        /* the radio group is sized as a stack of
         * radio-flavored toggles, plus an optional label
         */
        switch(horiz_align) {
        case ALIGN_RIGHT:
            offset->x -= actualSize.width;
            break;
        case ALIGN_CENTER:
            offset->x -= actualSize.width / 2;
            break;
        }
        switch(vert_align) {
        case ALIGN_BOTTOM:
            offset->y -= charHeight;
            break;
        case ALIGN_CENTER:
            offset->y -= charHeight / 2;
            break;
        case ALIGN_BASELINE:
            offset->y -= charAscent;  /* just use baseline of label */
            break;
        }
        break;
        
    case DLG_VERB_BUTTON:
        switch(horiz_align) {
        case ALIGN_RIGHT:
            offset->x -= actualSize.width;
            break;
        case ALIGN_CENTER:
            offset->x -= actualSize.width / 2;
            break;
        }
        switch(vert_align) {
        case ALIGN_BOTTOM:
            offset->y -= actualSize.height;
            break;
        case ALIGN_CENTER:
            offset->y -= actualSize.height / 2;
            break;
        case ALIGN_BASELINE:
            centerButtonText(actualSize.width, actualSize.height,
                             title, labelFs, FALSE, &ixtmp, &iytmp);
            offset->y -= iytmp - actualSize.origin.y;
            break;
        }
        break;

    case DLG_CHOICE:
#if AXMS
    case DLG_COMBO_BOX:
#endif
        switch(horiz_align) {
        case ALIGN_RIGHT:
            offset->x -= actualSize.width;
            break;
        case ALIGN_CENTER:
            offset->x -= actualSize.width / 2;
            break;
        case ALIGN_LEFT:
            size = AxArraySize(strs);
            for (ix = 0, widest = 0; ix < size; ix++) {
                stmp = AxStrPtrFromArray(strs, ix); /* button label */
                tfs = fs ? fs : AxDlgMainFont;
                ixtmp = AXTextWidth(tfs, stmp, strlen(stmp));
                if (ixtmp > widest) {
                    widest = ixtmp;
                }
            }
            wOptionWidth(title, labelFs, widest, &pulldownWidth,
                         &barwid, &barspc, &iwidth);
            offset->x -= iwidth - pulldownWidth;
            break;
        }
        switch(vert_align) {
        case ALIGN_BOTTOM:
            offset->y -= actualSize.height;
            break;
        case ALIGN_CENTER:
            offset->y -= actualSize.height / 2;
            break;
        case ALIGN_BASELINE:
            wOptionHeight(labelFs, fs, labelShadow,
                          textShadow, &iytmp, &ixtmp);
            offset->x -= iytmp;
            break;
        }
        break;
        
    case DLG_TABLE:
    case DLG_CANVAS:
    case DLG_TRAY:
    case DLG_EDIT_BOX:
    case DLG_LIST_BOX:
        scrollbarSizes(horiz_scroll, hide_vscroll, &ixtmp, &iytmp);
        iwidth = (int) actualSize.width;
        iheight = (int) actualSize.height;
        iwidth -= ixtmp;
        iheight -= iytmp;
        switch(horiz_align) {
        case ALIGN_RIGHT:
            offset->x -= iwidth;
            break;
        case ALIGN_CENTER:
            offset->x -= iwidth / 2;
            break;
        }
        switch(vert_align) {
        case ALIGN_BOTTOM:
            offset->y -= iheight;
            break;
        case ALIGN_CENTER:
            offset->y -= iheight / 2;
            break;
        }
        break;
        
    case DLG_ENTER:
        switch(horiz_align) {
        case ALIGN_RIGHT:
            offset->x -= actualSize.width;
            break;
        case ALIGN_CENTER:
            offset->x -= actualSize.width / 2;
            break;
        case ALIGN_LEFT:   /* align to left edge of field portion */
            if (LABEL_LEFT == titleType) {
                wEnterFieldOffset(LABEL_LEFT, title, labelFs, fs, ewidth,
                                  textShadow, &ixtmp, &iytmp);
                offset->x -= ixtmp;
            }
            break;
        }
        switch(vert_align) {
        case ALIGN_BOTTOM:
            offset->y -= actualSize.height;
            break;
        case ALIGN_CENTER:
            offset->y -= actualSize.height / 2;
            break;
        case ALIGN_BASELINE:
            offset->y -= charAscent;
            break;
        }
        break;
        
    case DLG_ROWCOL:
    case DLG_ICON:
    case DLG_RECT:
    case DLG_SCALE:
        switch(horiz_align) {
        case ALIGN_RIGHT:
            offset->x -= actualSize.width;
            break;
        case ALIGN_CENTER:
            offset->x -= actualSize.width / 2;
            break;
        }
        switch(vert_align) {
        case ALIGN_BOTTOM:
            offset->y -= actualSize.height;
            break;
        case ALIGN_CENTER:
            offset->y -= actualSize.height / 2;
            break;
        }
        break;

#if AX_X11
/*** Windows case handled under DLG_CHOICE, see above.  -donato  ***/
    case DLG_COMBO_BOX:
        switch(horiz_align) {
        case ALIGN_RIGHT:
            offset->x -= actualSize.width;
            break;
        case ALIGN_CENTER:
            offset->x -= actualSize.width / 2;
            break;
        }
        switch(vert_align) {
        case ALIGN_BOTTOM:
            offset->y -= actualSize.height;
            break;
        case ALIGN_CENTER:
            offset->y -= actualSize.height / 2;
            break;
        case ALIGN_BASELINE:
            offset->y -= CbComboBaselineOffset(fs, wComboHeight(fs));
            break;
        }
        break;
#endif
    }
}

/* : wElfGadgetSize - given the current size-as-realized
 * in screenpix, return the corresponding size/pos as seen
 * by elf.  A FALSE return indicates that the gadget's size
 * is fixed under dialog-layout operation; zero size is
 * written for these gadgets.  All gadgets' positions are
 * considered meaningful in this output, however.
 *
 * This function is just a units-converter, and doesn't
 * try to validate the passed actualSize data; for instance,
 * it doesn't check whether a verb button is too small for
 * its text.
 */
int   /* return bool, whether size output is meaningful */
wElfGadgetSize(dboxRev, dboxRefFont, dboxNoExpand,
               gadgetType, gadgetSizeAtts, actualSize, elfSize)

int dboxRev;               /* in: revstamp of containing dbox */
data_ptr dboxRefFont;      /* in: reference font */
int dboxNoExpand;          /* in: !whether box does "bigfont" expansion */
int gadgetType;            /* in: elf gadget type, e.g, DLG_TABLE */
data_ptr gadgetSizeAtts;   /* in: ordered array of gadget atts re: geometry */
r32 *actualSize;           /* in: size-as-realized by widgetkit */
r32 *elfSize;              /* out: corresponding elf size */
{
    int iwidth, iheight, ixtmp, iytmp, different, expand;
    int expos, eypos, ewidth, eheight, titleType, horiz_scroll;
    int thickness, buttonType, hide_vscroll, horiz_align, vert_align;
    XFontStruct *fs, *labelFs, *refFont;
    data_ptr textShadow, labelShadow, widgetShadow;
    data_ptr strs, fontAttr, labelFont, widgetAlignment;
    mbchar *title;
    
    expand = !dboxNoExpand;
    
    gadgetGeomAtts(gadgetSizeAtts, &title, &expos, &eypos, &ewidth,
                   &eheight, &strs, &titleType, &horiz_scroll, &thickness,
                   &buttonType, &hide_vscroll, &widgetShadow, &textShadow,
                   &fontAttr, &labelFont, &labelShadow, &widgetAlignment);
    
    /* the elf and c positions are identical for most types
     */
    elfSize->origin.x = actualSize->origin.x;
    elfSize->origin.y = actualSize->origin.y;
    
    /* many of the widgets look at the local font attr
     */
    fs = lookupNamedFont(fontAttr);
    
    /* this is the dialog main font at the time the dbox was last saved
     */
    refFont = lookupReferenceFont(dboxRefFont, &different);
    
    switch(gadgetType) {
        
    case DLG_TEXT:
    case DLG_TOGGLE_BUTTON:
    case DLG_RADIO_BOX:
    case DLG_CHOICE:
    case DLG_ICON:
    case DLG_SPLITTER:
    case DLG_ROWCOL:
        /* these types are fixed-size under layout
         * adjustment
         */
        elfSize->width = elfSize->height = 0;
        return FALSE;

    case DLG_RECT:
    case DLG_SCALE:
    case DLG_VERB_BUTTON:
    case DLG_COMBO_BOX:
        /* these types are resizeable under layout
         * adjustment, and have the same elf and c
         * sizes.
         */
        elfSize->width = actualSize->width;
        elfSize->height = actualSize->height;
        return TRUE;

    case DLG_LIST_BOX:
        /* remove the size contributions of any
         * scrollbars, then convert to #columns, #items
         */
        scrollbarSizes(horiz_scroll, hide_vscroll, &ixtmp, &iytmp);
        iwidth = (int) (actualSize->width - ixtmp);
        iheight = (int) (actualSize->height - iytmp);
        elfSize->width = (unsigned int) wListColWidth(iwidth, dboxRev,
                                                      refFont);
        elfSize->height = (unsigned int)
            wListVisibleItemCount(iheight, fs, textShadow, horiz_scroll);
        return TRUE;
        
    case DLG_TABLE:
    case DLG_CANVAS:
    case DLG_TRAY:
    case DLG_EDIT_BOX:
        /* remove the size contributions of any scrollbars
         */
        scrollbarSizes(horiz_scroll, hide_vscroll, &ixtmp, &iytmp);
        elfSize->width = (unsigned int) (actualSize->width -
                                         SCROLLBAR_THICKNESS);
        elfSize->height = (unsigned int) (actualSize->height - iytmp);
        return TRUE;
        
    case DLG_ENTER:
        /* remove the size contribution of any label, then
         * convert width to #columns.
         */
        elfSize->height = 0;   /* elf ignores height of this gadget */
        labelFs = lookupNamedFont(labelFont);
        wLabelSize(title, labelFs, labelShadow, &ixtmp, &iytmp);
        if (titleType == LABEL_LEFT) {
            /* total height is same as enterfield;
             * total width includes title width plus 3pix pad
             * (cribbed from elxwins.c)
             */
            iwidth = (int) actualSize->width - (ixtmp + 3);
            elfSize->width = (unsigned int)
                wEnterColWidth(iwidth, fs, textShadow);
            return TRUE;
        }
        else if (titleType == LABEL_TOP) {
            /* total height includes title plus 4pix pad;
             * total width reflects title if longer than field
             */
            iheight = (int) actualSize->height - (iytmp + 4);
            iwidth = (int) actualSize->width;
            if (ixtmp > iwidth) {
                /* title is wider than entryfield; assume
                 * that we wouldn't have allowed dialog-layout
                 * adjustment of this gadget's width
                 */
                elfSize->width = (unsigned int) 0;
                return FALSE;
            }
            else {
                iwidth = (int) actualSize->width;
                elfSize->width = (unsigned int)
                    wEnterColWidth(iwidth, fs, textShadow);
                return TRUE;
            }
        }
        else {
            /* no label
             */
            iwidth = (int) actualSize->width;
            elfSize->width = (unsigned int)
                wEnterColWidth(iwidth, fs, textShadow);
            return TRUE;
        }
        break;
        
    default:
        elfSize->width = elfSize->height = 0;
        break;
    }
}

/* : wMinimumGadgetSize - report the minimum allowable realized size
 * for the passed gadget.  return whether the gadget is deformable
 * under dialog layout adjustment.  reported min size is meaningful
 * only if the gadget can be deformed, i.e., only if TRUE is returned.
 */
int   /* return whether gadgetType is deformable under layout adjust */
wMinimumGadgetSize(dboxRev, dboxRefFont, dboxNoExpand,
                   gadgetType, gadgetSizeAtts, minSize)

int dboxRev;               /* in: revstamp of containing dbox */
XFontStruct *dboxRefFont;  /* in: reference font */
int dboxNoExpand;          /* in: !whether box does "bigfont" expansion */
int gadgetType;            /* in: elf gadget type, e.g, DLG_TABLE */
data_ptr gadgetSizeAtts;   /* in: ordered array of gadget atts re: geometry */
r32 *minSize;              /* out: the c-level minimum size */
{
    int iwidth, iheight, ixtmp, iytmp, titleType;
    int thickness, pixType, different;
    int expos, eypos, ewidth, eheight, horiz_scroll;
    int buttonType, hide_vscroll;
    data_ptr strs, widgetShadow, textShadow;
    data_ptr fontAttr, labelFont, labelShadow, widgetAlignment;
    XFontStruct *fs, *labelFs, *refFont, *tfs;
    mbchar *title;
    r32 trect;
    
    gadgetGeomAtts(gadgetSizeAtts, &title, &expos, &eypos, &ewidth,
                   &eheight, &strs, &titleType, &horiz_scroll, &thickness,
                   &buttonType, &hide_vscroll, &widgetShadow, &textShadow,
                   &fontAttr, &labelFont, &labelShadow, &widgetAlignment);
    
    minSize->origin.x = minSize->origin.y = 0;
    
    fs = lookupNamedFont(fontAttr);
    labelFs = lookupNamedFont(labelFont);
    
    switch(gadgetType) {
        
    case DLG_TEXT:
    case DLG_TOGGLE_BUTTON:
    case DLG_RADIO_BOX:
    case DLG_CHOICE:
    case DLG_SPLITTER:
    case DLG_ROWCOL:
    case DLG_ICON:
        /* these types are fixed-size under dialog
         * layout adjustment
         */
        return FALSE;
        
    case DLG_VERB_BUTTON:
        /* the button should be able to fit its text,
         * if any.  treat pix buttons as fixed-size
         * under dlg layout
         */
        pixType = buttonType == DLG_ICON_BUTTON;
        if (pixType) {
            return FALSE;
        }
        else {
            iheight = iwidth = 0;
            wVerbButtonSize(title, pixType, labelFs, thickness,
                            widgetShadow, textShadow, &iwidth, &iheight);
            minSize->width = iwidth;
            minSize->height = iheight;
            return TRUE;
        }
        
    case DLG_RECT:
        /* it'd be nice to prevent the panel-form from
         * being squeezed into an etch, but just accept
         */
        minSize->width = minSize->height = 0;
        return TRUE;
        
    case DLG_CANVAS:
    case DLG_TRAY:
    case DLG_TABLE:
    case DLG_EDIT_BOX:
        if (gadgetType == DLG_CANVAS ||
            gadgetType == DLG_TRAY) {
            /* use a minimum of 40pix sqaure, not
             * counting any scrollbars
             */
            iwidth = iheight = 40;
        }
        else {
            iwidth = iheight = 100;
        }
        scrollbarSizes(horiz_scroll, hide_vscroll, &ixtmp, &iytmp);
        minSize->width = (unsigned int) (iwidth + SCROLLBAR_THICKNESS);
        minSize->height = (unsigned int) (iheight + iytmp);
        return TRUE;
        
    case DLG_SCALE:
        /* limit to no smaller than 60x16
         */
        minSize->width = 60;
        minSize->height = 16;
        return TRUE;
        
    case DLG_LIST_BOX:
        /* use minimum of 6 columns by 2 items
         */
        refFont = dboxRefFont;
        minSize->width = (unsigned int)
            wListPixHeight(2, fs, textShadow, horiz_scroll);
        iheight = wListPixWidth(6, dboxRev, refFont);   /* six columns */
        minSize->height = (unsigned int) iheight;
        return TRUE;
        
    case DLG_ENTER:
        /* the height is unchanged under layout adjust;
         * the width should allow an entryfield fitting
         * at least 2 columns
         */
        /* get width of entryfield alone
         */
        ixtmp = wEnterPixWidth(2, fs, textShadow);
        tfs = labelFs ? labelFs : AxDlgMainFont;
        iwidth = AXTextWidth(labelFs, title, strlen(title));
        
        if (titleType == LABEL_LEFT) {
            iwidth = iwidth + 3 + ixtmp;
        }
        else if (titleType == LABEL_TOP) {
            if (iwidth < ixtmp) {
                iwidth = ixtmp;
            }
        }
        else {
            iwidth = ixtmp;   /* no label */
        }
        minSize->height = 0;
        minSize->width = (unsigned int) iwidth;
        return TRUE;
        break;
        
    case DLG_COMBO_BOX:
        minSize->height = 24;  /* based on arrow bitmap size + bevels */
        minSize->width = 24;
        break;
    }
}

/* : dlgReadSizeValue - read one of xpos, ypos,
 * width, or height from a gadget- or dbox-format array
 */
int
dlgReadSizeValue(array, offset)
data_ptr array;
int offset;
{
    data_ptr atmp;
    
    atmp = AxArrayElement(array, offset);
    if (AxIsArray(atmp)) {
        return(AxIntFromArray(atmp, 0));    /* "motif" val */
    }
    else {
        return(AxIntFromDataPtr(atmp));
    }
}

/* : dlgReadSizeValue - write one of xpos, ypos,
 * width, or height from a gadget- or dbox-format array
 */
void
dlgWriteSizeValue(array, offset, val)
data_ptr array;
int offset;
int val;
{
    data_ptr atmp;
    
    atmp = AxArrayElement(array, offset);
    if (AxIsArray(atmp)) {
        AxAddIntToArray(atmp, 0, val);    /* "motif" val */
    }
    else {
        AxAddIntToArray(array, offset, val);    /* "motif" val */
    }
}

/* : scrollbarSizes - return the size contributions of
 * any scrollbars that may be present
 */
static void
scrollbarSizes(horiz_scroll, hide_vscroll, widthDelta, heightDelta)
int horiz_scroll;                 /* in */
int hide_vscroll;                 /* in */
int *widthDelta, *heightDelta;    /* out */
{
    *widthDelta = *heightDelta = 0;
    if(horiz_scroll) {
        *widthDelta = SCROLLBAR_THICKNESS;
    }
    if(!hide_vscroll) {
        *heightDelta = SCROLLBAR_THICKNESS;
    }
}

/* : wGetGeomAtts - create a gadget geom atts array from
 * the passed gadget_-format array.  This corresponds to
 * dialog.els:db_ctrl_get_geom_atts$
 */
data_ptr
wGetGeomAtts(gadget)
data_ptr gadget;       /* in format gadget_ */
{
    data_ptr ret;
    
    ret = AxMakeArray(17);

    ret = AxAddStrToArray(ret, 0, AxStrPtrFromArray(gadget, DBDEXtitle));
    ret = AxAddIntToArray(ret, 1, dlgReadSizeValue(gadget, DBDEXxpos));
    ret = AxAddIntToArray(ret, 2, dlgReadSizeValue(gadget, DBDEXypos));
    ret = AxAddIntToArray(ret, 3, dlgReadSizeValue(gadget, DBDEXwidth));
    ret = AxAddIntToArray(ret, 4, dlgReadSizeValue(gadget, DBDEXheight));
    ret = AxAddArrayToArray
        (ret, 5, ElfDataCopy(AxArrayElement(gadget, DBDEXstrs)));
    ret = AxAddIntToArray(ret, 6, AxIntFromArray(gadget, DBDEXno_title));
    ret = AxAddIntToArray(ret, 7, AxIntFromArray(gadget, DBDEXhoriz_scroll));
    ret = AxAddIntToArray(ret, 8, dlgReadSizeValue(gadget, DBDEXthickness));
    ret = AxAddIntToArray(ret, 9, AxIntFromArray(gadget, DBDEXtype));
    ret = AxAddIntToArray(ret, 10, AxIntFromArray(gadget, DBDEXhide_vscroll));
    ret = AxAddArrayToArray
        (ret, 11, ElfDataCopy(AxArrayElement(gadget, DBDEXshadow)));
    ret = AxAddArrayToArray
        (ret, 12, ElfDataCopy(AxArrayElement(gadget, DBDEXtextShadow)));
    ret = AxAddArrayToArray
        (ret, 13, ElfDataCopy(AxArrayElement(gadget, DBDEXfontAttr)));
    ret = AxAddArrayToArray
        (ret, 14, ElfDataCopy(AxArrayElement(gadget, DBDEXlabelFont)));
    ret = AxAddArrayToArray
        (ret, 15, ElfDataCopy(AxArrayElement(gadget, DBDEXlabelShadow)));
    ret = AxAddArrayToArray
        (ret, 16, ElfDataCopy(AxArrayElement(gadget, DBDEXalignment)));
    
    return(ret);
}

/* : gadgetGeomAtts - return values of a set of
 * geometry-affecting gadget atts.  The passed attsArray
 * param is prepared by dialog.els:db_ctrl_get_geom_atts$,
 * and by owidget.els:get_gadget_geom_atts$
 */
static void
gadgetGeomAtts(attsArray,
               title, xpos, ypos, width, height,
               strs, no_title, horiz_scroll, thickness,
               type, hide_vscroll, shadow, textShadow, fontAttr,
               labelFont, labelShadow, widgetAlignment)

data_ptr attsArray;                    /* in */
mbchar **title;                        /* rest of params out */
int *xpos, *ypos, *width, *height;
data_ptr *strs;
int *no_title, *horiz_scroll;
int *thickness, *type, *hide_vscroll;
data_ptr *shadow, *textShadow;
data_ptr *fontAttr, *labelFont;
data_ptr *labelShadow;
data_ptr *widgetAlignment;
{
    *title = AxStrPtrFromArray(attsArray, 0);
    *xpos = AxIntFromArray(attsArray, 1);
    *ypos = AxIntFromArray(attsArray, 2);
    *width = AxIntFromArray(attsArray, 3);
    *height = AxIntFromArray(attsArray, 4);
    *strs = AxArrayElement(attsArray, 5);
    *no_title = AxIntFromArray(attsArray, 6);
    *horiz_scroll = AxIntFromArray(attsArray, 7);
    *thickness = AxIntFromArray(attsArray, 8);
    *type = AxIntFromArray(attsArray, 9);
    *hide_vscroll = AxIntFromArray(attsArray, 10);
    *shadow = AxArrayElement(attsArray, 11);
    *textShadow = AxArrayElement(attsArray, 12);
    *fontAttr = AxArrayElement(attsArray, 13);
    *labelFont = AxArrayElement(attsArray, 14);
    *labelShadow = AxArrayElement(attsArray, 15);
    *widgetAlignment = AxArrayElement(attsArray, 16);
}
