/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: win_axgfs.c
   Applix ASC version: 4
   Date: Mon Nov 16 13:53:48 1998
*/
#include <stdlib.h>
#include <aos.h>
#include <axtext.h>
#include <axgfs.h>




axletter *
axgfsIntGetLetter(charcode, fs, ctm, fid, pix, bbox)
int charcode;     /* in */
XFontStruct *fs;  /* in */
axctm *ctm;		  /* in */
Font fid;         /* in */
int pix;          /* in: bool; whether to return pixmap or just bits */
int bbox;         /* in: bool; whether to include a bbox in the return */
{

    int haveLetter, shift_up;
    axletter *thisLetter = (axletter *)0;
    GLYPHMETRICS g_metrics;
    int ret, buff_size;
    char *raw_bits;
    MAT2 mat;
    HDC hdc;
    HFONT old_font;
    int ms_bytes_per_row, gallium_bytes_per_row;
    char *p, *p2;
    int i;

    if (!fs || !fid)
        return ((axletter *) NULL);

    haveLetter = TRUE;
    if ( charcode < 0)
        haveLetter = FALSE;

    /* return the letterform bits directly */
    thisLetter = (axletter *)calloc(1, sizeof(axletter));
    thisLetter->bbox = (r32 *)calloc(1, sizeof(r32));
    
    if (!haveLetter) 
        return( thisLetter );    /* return a zero structure */

    /*****************************************/
    /* first find out how much space we need */
    /*****************************************/
    hdc = MS_GetDC( NULL );
    old_font = MS_SelectObject( hdc, fid );

    /* set up the matrix for the GlyphOutline call */
	if (ctm)
	    MakeCtmMatrix( &mat, ctm );
	else
	    MakeIdentityMatrix( &mat );

    buff_size = MS_GetGlyphOutline( hdc, (UINT)charcode, GGO_BITMAP, &g_metrics,
                0, 0, &mat );

    if ( !buff_size )
       {
       ret = MS_GetLastError();
       goto return_letter;
       }

    raw_bits = (char *)malloc( buff_size );

    /*****************************************/
    /* now actually get the bits */
    /*****************************************/
    ret = MS_GetGlyphOutline( hdc, (UINT)charcode, GGO_BITMAP, &g_metrics,
            buff_size, raw_bits, &mat );

    thisLetter->bbox->width = g_metrics.gmBlackBoxX;
    thisLetter->bbox->height = g_metrics.gmBlackBoxY;
    thisLetter->bbox->origin.x = g_metrics.gmptGlyphOrigin.x;
    thisLetter->bbox->origin.y = -(g_metrics.gmptGlyphOrigin.y);

    shift_up = 0;
    if ( (g_metrics.gmptGlyphOrigin.y > 0) &&
         (g_metrics.gmBlackBoxY < g_metrics.gmptGlyphOrigin.y) )
       {
       /* the character appears above the baseline, but there are only */
       /* bits for the axtual character.  Padd the image with more bits */
       /* to get the character shifted up to the correct position       */
       shift_up = g_metrics.gmptGlyphOrigin.y - g_metrics.gmBlackBoxY;
       }

    /* the bits returned from the MS_GetGlyphOutline() call are padded to   */
    /* 32 bits,  but the upper levels of our code expect things to be only  */
    /* byte alligned, since that is the way the Galium server returns them  */
    /* To compensate, we copy the bits (if need be)                         */
    ms_bytes_per_row = buff_size/g_metrics.gmBlackBoxY;
    gallium_bytes_per_row = (g_metrics.gmBlackBoxX + 7)/8;
    if ( (ms_bytes_per_row == gallium_bytes_per_row) && (shift_up == 0) )
       {
       /* the fast path */
       thisLetter->bits = raw_bits;
       goto return_letter;
       }

    /* allocate a new buffer */
    thisLetter->bits = (char *)calloc( 1, 
                  gallium_bytes_per_row * (thisLetter->bbox->height+shift_up));
    for ( i=0; i<thisLetter->bbox->height; i++ )
       {
       p = (char *)(thisLetter->bits) + (i * gallium_bytes_per_row );
       p2 = raw_bits + (i * ms_bytes_per_row);
       memcpy( p, p2, gallium_bytes_per_row );
       }
    thisLetter->bbox->height = thisLetter->bbox->height + shift_up;    
    free( raw_bits );

return_letter:
    MS_SelectObject(hdc, old_font);
    MS_ReleaseDC(NULL, hdc);
    return (thisLetter);
}



/* create a font to yield the desired pointsize at the specified resolution  */
/* returns a font handle                                                     */
char *axgfsIntGetFid( char *xlfd_template,   /* should be "fontname.win"     */
                      int pointsize, 
                      int resolution, 
                      axctm *ctm)
{
    HFONT hfont;
    LOGFONT logfont;
    char char_face_name[50];
    char *p;
    fontTabEnt *trec, *screenRecord;
    char *screenFamily, *err;
    int bold, italic, effective_size;

    /* first, the input "xlfs" will be of the form "FONTNAME[ Bold][ Italic].win" */
    /* so take off any "Bold" and "Italic", rermember them, and leave just        */
    /* the font name                                                              */
    memset( &logfont, 0, sizeof(logfont) );

    strcpy( char_face_name, xlfd_template );
    p = char_face_name + strlen(char_face_name) - 4;
    *p = '\0';                                          /* clip off ".win"  */
    bold = italic = 0;
    while ( p > char_face_name )
       {
       if ( *p == ' ' )
           {
           /* possible "Bold" or "Italic" */
           if ( strcmp( (p+1), "Italic" ) == 0 )
              {
              italic = TRUE;
              *p = '\0';                              /* clip it off */
              }
           else if ( strcmp( (p+1), "Bold" ) == 0 )
              {
              bold = TRUE;
              *p = '\0';                              /* clip it off */
              }
           else
              /* must be a blank in the middle of the face name, all done */
              break;
           }
       p--;
       }

    if ( bold )
       logfont.lfWeight = FW_BOLD;
    else
       logfont.lfWeight = FW_NORMAL;
    if ( italic )
       logfont.lfItalic = 1;
       

    /* since we want the font bits to come out at the input resolution, but */
    /* we will actually be drawing the font at the screen resolution, we compute */
    /* an effective point size, to yield the desired bits at the input resolution */
    /* make the size negative so that Windows gives us a font of this height */
    /*   not a font within a line spacing of this height           */
    effective_size = -((pointsize * resolution) / 72);
    logfont.lfHeight = effective_size;

    /* lookup charset information based on actual screenfont */
    err = axGetFontRecordForScreen(char_face_name, bold, italic, pointsize,
                                   &trec, &screenRecord);
    if ( err )
        {
        freeErrString ( err );
        }
    else
        {
        logfont.lfCharSet = screenRecord->charSet;
        logfont.lfPitchAndFamily = screenRecord->pitchAndFamily;
        }

    strcpy( logfont.lfFaceName, char_face_name );

    logfont.lfOutPrecision = OUT_TT_ONLY_PRECIS;
    
    hfont = MS_CreateFontIndirect( &logfont );
    if ( hfont == (HFONT)NULL )
        {
        /*did not get the font, try with everything default except the size */
        memset( &logfont, 0, sizeof(logfont) );
        logfont.lfHeight = effective_size;

        hfont = MS_CreateFontIndirect( &logfont );
        if ( hfont == (HFONT)NULL );
            return( (char *)NULL );
        }

    return ((char *) hfont);
}





void axgfsInit()
{
    return;
}


int isGFSxlfdName()
{
    return( FALSE );
}


XFontStruct *
axgfsOpenFont(xlfd, resolution, pointsize, ftab)
char *xlfd;
int resolution;
int pointsize;
fontTabEnt *ftab;
{
    return ((XFontStruct *) NULL);
}


void
axgfsInitCharStruct(){}


