/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winbuttn.c
   Applix ASC version: 126
   Date: Mon Nov 16 13:53:36 1998
*/
#include <windows.h>
#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>
#include <xmpixel.h>
#include <axwinclr.h>


#if 0
/* The following constants are used so that the code will draw asymetric */
/* sun and shadow lines around a button.  In Windows, small buttons for  */
/* things like the xpress line, and scrool bar buttons. actually have a  */
/* sun width of 1 pixel, and a shadow width of 2 pixels                  */
#define SMALL_BUTTON_HEIGHT 26
#define SMALL_BUTTON_WIDTH 50
#endif


#define IS_DEFAULT_PUSHBUTTON(w) \
            (w->parent->type == W_DIALOG && \
            ((AXWidget_dialog)w->parent)->currentDefaultButton && \
            w->uid == ((AXWidget_dialog)w->parent)->currentDefaultButton->uid)

extern Display *Dpy;
extern GC BackGC;
extern HANDLE hInst;
extern PIXEL_VALUE BackPixelRGB;
extern PIXEL_VALUE ButtonFacePixel, ForePixel, BackPixel;
extern PIXEL_VALUE SunPixel, ShadowPixel;
extern AXWidget_pushbutton ButtonArmed;
extern HFONT NormalDialogFont;
extern HPALETTE AxWinPaletteHandle;
extern int Scrollbar_Thickness;

#include <winbuttn.p>




/***************************************************************************/
/* THIS FUNCTION IS FOR WINDOWS ONLY!  the UNIX/X version is in XMBUTTON.C */
/* ALL COMMON BUTTON FUNCTIONS SHOULD GO IN BUTTNCOM.C   - donato          */
/*************************************************************************/
/* function: PaintButton - Paint a button                                */
/*************************************************************************/
void
PaintButton(AXWidget_pushbutton w, int not_used, int not_used2, int layered)
{
	int save_thickness;
    HDC hDC, pkgDC;
    HBITMAP bitmap_handle;
    RECT rect;
    int border_width, border_space;
    int xpos, ypos, base_x, base_y, pad_x, pad_y;
    AXWidget_main parentP;
    AXWidget_pushbutton childP;
    int remove_widget_dc = FALSE,
        remove_parent_dc = FALSE,
        release_dc = FALSE;
    int button_is_down;
    int asymetric;
    pixcache_t *pix_cacheP;
    HFONT hFont, old_font = NULL;
    int old_bk_mode;
    COLORREF old_bk_color, old_text_color;
    HPALETTE old_palette;
    char *image_name;
    POINT point;
    int curr_wid, curr_hyt;
    BITMAP bm;
	char *disp_name;

    WIDGET_PARANOID(w, W_BUTTON);
    if ( !w->parent->winid && (w->parent->type != W_CORNER) )
        /* parent wasn't created, probably a windowless app. */
        return;

	save_thickness = w->thickness;

    /* first, check for buttons that don't get painted here... */
    if( w->parent->type == W_PLATES && layered /*hack for now!*/ )
        {
        if (ssDrawPlatesPushButton(w))
            {
            goto cleanup;
            }
        }

    if ( !w->display )
        return;

    /* check for layering */
    if ( layered )
        {
        if ( w->flags & WFLG_IN_EDIT )
            {
            if ( dePaintWidget((AXWidget)w) )
                return;
            }
        else
            {
            if ( (AXWidget_pushbutton)topWidget((AXWidget)w) == w )
                {
                paintCompositeWidget((AXWidget)w, (XExposeEvent*)not_used,
                                                                not_used2);
                return;
                }
            }
        }

    if ( w->winid && !w->owner_draw )
        {
        /* this is a "normal" button, which Windows will paint for us. *
         * However, we still need to update its display; on, off, etc..*/

        /* move or re-size as necessary. */
        MS_GetWindowRect(w->winid, &rect);
        point.x = rect.left;
        point.y = rect.top;
        MS_ScreenToClient(w->parent->winid, &point);
        curr_wid = rect.right - rect.left;
        curr_hyt = rect.bottom - rect.top;
        if ( (w->xpos != point.x) || (w->ypos != point.y) ||
                (w->width != curr_wid) || (w->height != curr_hyt) )
            /* button has changed, move/resize it. */
            {
            MS_MoveWindow(w->winid, w->xpos, w->ypos,
                            w->width, w->height, TRUE);
            /* we explicitly invalidate the whole button window *
             * to make sure all of it gets repainted.           */
            MS_InvalidateRect(w->winid, (CONST RECT*)NULL, FALSE);
            }

        MS_InvalidateRect(w->winid, (CONST RECT*)NULL, FALSE);
        if ( IS_DEFAULT_PUSHBUTTON(w) )
            {
            /* pushbutton with "default" style border. */
            MS_SendMessage(w->winid, BM_SETSTYLE,
                            (WPARAM)LOWORD(BS_DEFPUSHBUTTON),
                            MAKELPARAM(TRUE, 0));
            }
        else
            {
            /* normal pushbutton. */
            MS_SendMessage(w->winid, BM_SETSTYLE,
                            (WPARAM)LOWORD(BS_PUSHBUTTON),
                            MAKELPARAM(TRUE, 0));
            }
        if ( w->disabled )
            {
            MS_EnableWindow(w->winid, FALSE);    /* gray out button */
            }
        else
            {
            MS_EnableWindow(w->winid, TRUE);
            MS_ShowWindow(w->winid, SW_SHOW);
            /* This block we are in deals with "normal" Windows pushbuttons.
             * But if the value is set to 'down' we need to reset the
             * button style to owner-draw because that is the only way
             * to draw a button in the down state on Windows.    -donato
             */
            if ( w->value )
                {
                MS_SendMessage(w->winid, BM_SETSTYLE,
                                (WPARAM)LOWORD(BS_OWNERDRAW),
                                MAKELPARAM(TRUE, 0));
                w->owner_draw = TRUE;
                }
            }
		/* set name in case it changed. */
		if ( !w->pixmapLabel && !w->pixmap )
			{
			axmsPutMnemonicInString(w->name, w->accelerator_key, &disp_name);
			MS_SendMessage(w->winid, WM_SETTEXT, 0, (LPARAM)disp_name);
			free(disp_name);
			}
        goto cleanup;
        }

    if ( w->parent->type == W_CORNER && !w->winid )
        {
        w->xpos += w->parent->xpos;
        w->ypos += w->parent->ypos;
		w->width += Scrollbar_Thickness - w->width - 1;
		w->height += Scrollbar_Thickness - w->height - 1;
        w->parent->winid = w->parent->parent->winid;
        }

    if ( (w->style != W_POPUP) && !w->winid )
        {
        /* widget not created yet; this case occurs when     *
         * creating buttons for the ruler and macro debugger.*
         * the "tab" buttons are not actually buttons so     *
         * don't create them.     - donato                   */
        if ( IsTabRulerButton(w) )
            w->thickness = 1;   /* just draw a button-like object */
        else
            {
            /* if not in DE, realize button. (DE buttons painted below) */
            if ( !(w->flags & WFLG_IN_EDIT) )
                {
                RealizePushButton(w, FALSE);
                goto cleanup;
                }
            else
                w->thickness = 1;    /* make sure thickness is set for DE */
            }
        }

    parentP = w->parent;
    /* if the button widget has a window (HDC) (ie OWNERDRAW buttons) 
     * then use it, otherwise it's a RECT or ICON, so get the parent's
     * HDC, because these are drawn on the parents' surface.
     */
    /* bit of a HACK here: we use 2 DCs, one "real" and one "packaged".
     * we send the packaged one to the X calls because they assume that any
     * UNpackaged handle is a bitmap. the "real" DC we pass to the MS calls.
     *      -donato 
     */
    if ( w->window )
        {    /* draw relative to button window */
        base_x = 0;
        base_y = 0;
        pkgDC = w->window;
        hDC = InsureDrawable(w->window, &remove_widget_dc);
        }
    else if ( parentP->window > (HDC)1 )
        {
        /* the dialog editor specifies another hdc, i.e. a background pixmap */
        /* (the test "> 1" is because when we release an HDC we set the value */
        /* to 1 as a marker in case someone tries to use it after release)  */
        pkgDC = parentP->window;
        hDC = InsureDrawable(parentP->window, &remove_parent_dc);
        base_x = w->xpos;
        base_y = w->ypos;
        }
    else
        {
        /* draw relative to dialog window */
        base_x = w->xpos;
        base_y = w->ypos;
        hDC = MS_GetDC(parentP->winid);
        pkgDC = AxPackageHandle(hDC, AX_DC_TYPE);
        release_dc = TRUE;
        }


    if ( (w->flags & WFLG_IN_EDIT) && layered )
        {
        if ( dePaintWidget((AXWidget)w) )
            return;
        }

#if 0
	if ( w->style == W_POPUP )
		w->thickness = 2;
	else
#endif
		w->thickness = 1;	/* force all buttons to a thickness of one, 
							 * ala Win95/NT4.0
							 */

    /* if button is a bitmap, but hasn't been realized yet, *
     * or if it is a W_POPUP, we need to get the pixmap.    */    
    if ( w->pixmapLabel || w->style == W_POPUP )
        {
        if ( w->alternatePixmap && w->value )
            image_name = w->alternatePixmap;
        else if ( !w->pixmap || (w->alternatePixmap && !w->value) )
            image_name = w->name;
        else
            image_name = NULL;

        if ( image_name )
            {
            pix_cacheP = (pixcache_t*) Pixmapper(image_name);
            if ( pix_cacheP )
                w->pixmap = pix_cacheP->pix;
            else
                w->pixmap = NULL;
            }
        }

    bitmap_handle = (HBITMAP) NULL;
    if ( w->pixmap )
        {
        /* a button with a pixmap, as opposed to text, on it */
        bitmap_handle = w->pixmap;
        if ( w->value )
            button_is_down = TRUE;
        else
            /* not inverse, normal mode */
            button_is_down = FALSE;
        }
    else
        {
        /* we may have a text button that needs to be owner-draw,
         * so check for that case.
         */
        if ( w->owner_draw && w->value )
            button_is_down = TRUE;
        else
            /* if here, button is in DE; therefore always "up" */
            button_is_down = FALSE;
        }


    /* check for shadow */
    if ( wShadowIsOn(w->shadow) )
        {
        int shadow_base_x, shadow_base_y;
        HDC shadow_hDC, shadow_pkgDC;
        int xoff, yoff;

        shadow_base_x = w->xpos;
        shadow_base_y = w->ypos;
        shadow_hDC = MS_GetDC(parentP->winid);
        shadow_pkgDC = AxPackageHandle(shadow_hDC, AX_DC_TYPE);

        if ( w->value && !w->disabled )
            {
            wWidgetShadowOffsets(w->type, w->shadow, &xoff, &yoff);
            XFillRectangle(Dpy, shadow_pkgDC, BackGC,
                           shadow_base_x, shadow_base_y,
                             w->width + xoff + 2, w->height + yoff + 2);
            }
        else
            {
            xmDrawRectShadow((AXWidget)w, shadow_base_x, shadow_base_y,
                                                      w->width, w->height);
            }
        if ( !w->value )              
            {
            xmDrawRectAntiShadow((AXWidget)w, shadow_base_x, shadow_base_y,
                                                      w->width, w->height);
            }
        MS_ReleaseDC(parentP->winid, shadow_hDC);
        REMOVE_DRAWABLE(shadow_pkgDC);
        }


    /* paint the border */
    /* Note: the border is a 'thickness' shaded border and a single *
     * pixel width black box around the whole thing.                */
    if ( w->disabled && w->pixmapLabel )
        {
        /* we want a borderless "disabled" button, but we may *
         * be updating an already existing button, so "clear" *
         * it first, so the new state can be drawn.           */
        FillButtonFace(hDC, base_x, base_y,
                   w->width, w->height, ButtonFacePixel);
        /* undraw the "focus" rectangle */
        axmsUnDrawButtonFocusRect(w);
        if ( MS_IsWindowEnabled(w->winid) )
            MS_EnableWindow(w->winid, FALSE);    /* disable button */
        }
    else
        {
        axmsDrawButtonBorder(hDC, base_x, base_y, w->width, w->height,
								w->thickness, TRUE, button_is_down);

        if ( IS_DEFAULT_PUSHBUTTON(w) )
            axmsDrawButtonFocusRect(w);

        if ( w->disabled && MS_IsWindowEnabled(w->winid) )
            MS_EnableWindow(w->winid, FALSE);    /* disable button */
        else if ( !w->disabled && !MS_IsWindowEnabled(w->winid) )
            MS_EnableWindow(w->winid, TRUE);    /* enable button */
        }


    if ( bitmap_handle != (HBITMAP) NULL )
        {
        /* put down the pixmap */
#if 0
        if ( (w->height < SMALL_BUTTON_HEIGHT) && 
             (w->width < SMALL_BUTTON_WIDTH) )
            /* small buttons are drawn with one less sun line */
            asymetric = -1;
        else
#endif
            asymetric = 0;

        /* the "+1" below is because the black border is part of the button */
        xpos = base_x + w->thickness + asymetric;
        ypos = base_y + w->thickness + asymetric;

        border_width = 0;
        if ( !w->noBorder )
            border_width = 1;

        /* leave a possible 1 pixel border around the icon */
        xpos = xpos + border_width;
        ypos = ypos + border_width;

        /* here is a little code to make sure that a pixmap is centered */
        /*on the available drawing space of the button.  This was put in */
        /* originally to handle the small pixmap in the "corner" window, */
        /* because the size of the corner changes with large/small font  */
        /* mode.  This should have no effect on buttons which "fit" - jon */
        MS_GetObject(bitmap_handle, sizeof(BITMAP), (LPSTR)&bm);
        border_space = 2*w->thickness + 2*border_width + asymetric;
        if ( (int)(w->width - bm.bmWidth - border_space) < 0 )
            pad_x = 0;
        else
            pad_x = (w->width - bm.bmWidth - border_space)/2;
        if ( (int)(w->height - bm.bmHeight - border_space) < 0 )
            pad_y = 0;
        else
            pad_y = (w->height - bm.bmHeight - border_space)/2;
        xpos = xpos + pad_x;
        ypos = ypos + pad_y;
        PaintButtonPixmap(bitmap_handle,hDC, xpos,ypos, w->pix_mask,0,0,TRUE);

        /* simulate toggle effect for palette buttons */
        /* (taken from PaintButton in xmbutton.c)     */
        if ( ButtonArmed && w->parent->type == W_PALETTE )
            {
            for ( childP = (AXWidget_pushbutton)w->parent->widgets; childP;
                            childP = (AXWidget_pushbutton)childP->nextChild )
                {
                if ( w->gePaletteButtonType == childP->gePaletteButtonType )
                    {
                    if ( childP->value && (w->name != childP->name) &&
                                                    childP->display )
                        {
                        childP->value = w->value ? FALSE : TRUE;
                        ButtonArmed = FALSE;

                        PaintButton(childP, 0, 0, FALSE);
                        childP->value = TRUE;
                        break;
                        }
                    }
                }
            }                                                
        }
    else
        {
        int new_x, new_y;
        int use_gray_string = FALSE;
        COLORREF gray_color;

        old_palette = SelectAndRealizeAxWinPalette(hDC);
        /* (-+) 1 just to make sure text does not over-write border */
        rect.left   = base_x + w->thickness + 1;
        rect.top    = base_y + w->thickness + 1;
        rect.right  = base_x + w->width - 1;
        rect.bottom = base_y + w->height - 1;
        old_bk_mode = MS_SetBkMode(hDC, TRANSPARENT);
        if ( w->widgetColor )
            {
            old_bk_color = MS_SetBkColor(hDC,
                        GET_COLOR_VALUE(w->widgetColor->any.pixel));
            /* paint button face with user defined widget color. */
            FillButtonFace(hDC, base_x + 3, base_y + 3,
                            w->width - 5, w->height - 5,
                            w->widgetColor->any.pixel);
            }
        else
            old_bk_color = MS_SetBkColor(hDC,GET_COLOR_VALUE(ButtonFacePixel));
        if ( w->disabled )
            {
            if ( (gray_color = (COLORREF)MS_GetSysColor(COLOR_GRAYTEXT)) > 0 )
                old_text_color = MS_SetTextColor(hDC, gray_color);
            else
                use_gray_string = TRUE;
            }
        else if ( w->textColor )
            {
            old_text_color = MS_SetTextColor(hDC,
                        GET_COLOR_VALUE(w->textColor->any.pixel));
            }
        else
            old_text_color = MS_SetTextColor(hDC, GET_COLOR_VALUE(ForePixel));
        if ( w->font )
            {
            /* use specified font (from widget) */        
            hFont = w->font->fid;
            if ( !hFont )
                hFont = NormalDialogFont;    /* paranoia */
            }
        else
            hFont = NormalDialogFont;
        old_font = MS_SelectObject(hDC, hFont);
		axmsPutMnemonicInString(w->name, w->accelerator_key, &disp_name);
        if ( use_gray_string )
            {
            GetCenteredButtonTextPos(hDC, w, &new_x, &new_y);
            MS_GrayString(hDC, (HBRUSH)NULL, (GRAYSTRINGPROC)NULL,
                                (LPARAM)disp_name, strlen(disp_name),
                                new_x, new_y, 0, 0);
            }
        else
            {
            MS_DrawText(hDC, disp_name, strlen(disp_name), &rect,
                     DT_CENTER | DT_VCENTER | DT_SINGLELINE);
            }
		free(disp_name);
        /* cleanup */
        MS_SelectObject(hDC, old_font);
        MS_SetBkMode(hDC, old_bk_mode);
        MS_SetBkColor(hDC, old_bk_color);
        MS_SetTextColor(hDC, old_text_color);
        UnSelectPalette(hDC, old_palette);
        }

cleanup:
    if ( release_dc )
        {
        MS_ReleaseDC(parentP->winid, hDC);
        REMOVE_DRAWABLE(pkgDC);
        }
    else if ( remove_widget_dc )
        RemoveDrawable(w->window, hDC);
    else if ( remove_parent_dc )
        RemoveDrawable(parentP->window, hDC);

	w->thickness = save_thickness;
    w->mapped = TRUE;   /* flag to signify display */
    w->update_widget = FALSE;    /* it has been drawn */
}


void
axmsDrawButtonBorder(HDC hdc, int x, int y, int width, int height,
						int shadow_width, int fill_face, int down)
/* int down;    TRUE ==> the button is down */
/*** the hDC must be unpackaged! ***/
{
    int asymetric;
	RECT rect;
	HBRUSH hbrush;
    POINT points[2];
	HPALETTE old_palette;

	if ( width == 0 && height == 0 )
		return;

	/* first clear the area. */
	rect.top = y;
	rect.left = x;
	rect.right = x + width;
	rect.bottom = y + height;
	hbrush = MS_CreateSolidBrush(GET_COLOR_VALUE(ButtonFacePixel));
	old_palette = SelectAndRealizeAxWinPalette(hdc);
	FrameRect(hdc, &rect, hbrush);	/* GDI call */
	InflateRect(&rect, -1, -1);     /* GDI call */
	FrameRect(hdc, &rect, hbrush);
	MS_DeleteObject(hbrush);
	UnSelectPalette(hdc, old_palette);

	/* then draw a partial black border, depending on whether the button
	 * is up or down (ala the Win95 interface). 
	 */
    MS_SelectObject(hdc, MS_GetStockObject(BLACK_PEN));
    points[0].x = x;// + 1;
    points[0].y = points[1].y = y;
	if ( down )
		{
    	points[1].x = x + width;
    	MS_Polyline(hdc, points, 2);	/* top edge */
		}
	else
		{
    	/* bottom edge */
    	points[1].x = x + width - 1;
    	points[0].y = points[1].y = y + height - 1;
    	MS_Polyline(hdc, points, 2);
		}

	if ( down )
		{
    	points[0].y = y;
    	points[1].y = y + height;
    	points[0].x = points[1].x = x;
    	MS_Polyline(hdc, points, 2);	/* left edge */
		}
	else
		{
    	/* right edge */
    	points[0].y = y;// + 1;
    	points[1].y = y + height;
    	points[0].x = points[1].x = x + width - 1;
    	MS_Polyline(hdc, points, 2);
		}

	if ( down )
    	DrawShadow(hdc, shadow_width, x+1, y+1, width - 1, height - 1, down);
	else	
    	DrawShadow(hdc, shadow_width, x, y, width - 1, height - 1, down);

#if 0
    if ( ((height-2) < SMALL_BUTTON_HEIGHT) && 
         ((width-2) < SMALL_BUTTON_WIDTH) )
        /* small buttons are drawn with one less sun line */
        asymetric = -1;
    else
#endif
        asymetric = 0;

    if (fill_face)
        FillButtonFace(hdc, (x+shadow_width+1+asymetric), 
                            (y+shadow_width+1+asymetric), 
                            (width-(shadow_width*2)-2-asymetric), 
                            (height-(shadow_width*2)-2-asymetric), 
                            ButtonFacePixel);
    return;
}


void
DrawShadow(HDC in_hdc, register int size, register int x, register int y,
                    register int width, register int height, int state)
/* int state;    up, down, or erase the shadow */
{
    static HPEN hSunPen = (HPEN) - 1;
    static HPEN hShadowPen;
    HPEN upper_left_pen, lower_right_pen ;
    HDC hdc;
    int need_to_delete_hdc = FALSE;
    HPALETTE old_palette;


    if ( in_hdc == (HDC)NULL )
        return;

    if ( hSunPen == (HPEN) - 1 )
        {
        /* initialize the pens */
        hSunPen    = MS_CreatePen(PS_SOLID, 1, GET_COLOR_VALUE(SunPixel));
        hShadowPen = MS_CreatePen(PS_SOLID, 1, GET_COLOR_VALUE(ShadowPixel));
        }

    if ( state == DRAW_DOWN_SHADOW )
        {
        /* reverse sun and shadow */
        upper_left_pen = hShadowPen;
        lower_right_pen = hSunPen;
        }
    else if ( state == DRAW_UP_SHADOW )
        {
        /* normal sun and shadow directions */
        upper_left_pen = hSunPen;
        lower_right_pen = hShadowPen;
        }
    else if ( state == DRAW_NO_SHADOW )
        {
        /* no shading */
        upper_left_pen = lower_right_pen =
                    MS_CreatePen(PS_SOLID,1,GET_COLOR_VALUE(ButtonFacePixel));
        }

    if ( size <= 0 )
        return;
    if ( size > width / 2 )
        size = width / 2;
    if ( size > height / 2 )
        size = height / 2;
    if ( size <= 0 )
        return;

#if 0
    if ( (height < SMALL_BUTTON_HEIGHT) && 
            (width < SMALL_BUTTON_WIDTH) )
        /* "small" buttons are asymetric, with one less sun line */
        sun_size = size - 1;
    else
        sun_size = size;
#endif

    if ( IS_PACKAGED_DRAWABLE(in_hdc) )
        hdc = (HDC)InsureDrawable(in_hdc, &need_to_delete_hdc);
    else
        hdc = in_hdc;
    old_palette = SelectAndRealizeAxWinPalette(hdc);
	if ( state == DRAW_DOWN_SHADOW )
		{
		axmsDrawTopLeftRect(hdc, upper_left_pen, x,y,width,height, size);
		axmsDrawBottomRightRect(hdc, lower_right_pen, x,y,width,height, size);
		}
	else
		{
		axmsDrawBottomRightRect(hdc, lower_right_pen, x,y,width,height, size);
		axmsDrawTopLeftRect(hdc, upper_left_pen, x,y,width,height, size);
		}

    /* cleanup */
    UnSelectPalette(hdc, old_palette);
	if ( state == DRAW_NO_SHADOW )
		MS_DeleteObject(upper_left_pen);
    if ( need_to_delete_hdc )
        RemoveDrawable(in_hdc, hdc);
    return;
}

void
axmsDrawBottomRightRect(HDC hdc, HPEN hpen, int x, int y, 
						int width, int height, int thickness)
												
{
	int i;
	HPEN old_pen;
	POINT points[3];

   	old_pen = MS_SelectObject(hdc, hpen);
    for ( i = 0; i < thickness; i++ )
        {
        /* Bottom segments  */
        /* we draw across the bottom, then up the right side */
        points[0].x = x + i;
        points[0].y = y + height - i - 1;
        points[1].x = x + width - i - 1;
        points[1].y = points[0].y;
        points[2].x = points[1].x;
        points[2].y = y + i - 1;
		Polyline(hdc, points, 3);
		}
	MS_SelectObject(hdc, old_pen);
}

void
axmsDrawTopLeftRect(HDC hdc, HPEN hpen, int x, int y, int width, int height,
															int thickness)
{
	int i;
	HPEN old_pen;
	POINT points[3];

   	old_pen = MS_SelectObject(hdc, hpen);
    for ( i = 0; i < thickness; i++ )
        {
        /* Left and Top segments  */
        /* we draw up the left side, then across the top */
        points[0].x = x + i;
        points[0].y = y + height - i - 1;
        points[1].x = points[0].x;
        points[1].y = y + i;
        points[2].x = x + width - i;
        points[2].y = points[1].y;
		Polyline(hdc, points, 3);
		}
	MS_SelectObject(hdc, old_pen);
}

/* returns whether the tab button widget that was pressed. */
/* HACK for now! */
static BOOL
IsTabRulerButton(AXWidget_pushbutton w)
{
    if ( (strcmp(w->name, NXLT("tableft"))) != 0 &&
         (strcmp(w->name, NXLT("tabright"))) != 0 &&
         (strcmp(w->name, NXLT("tabcentr"))) != 0 &&
         (strcmp(w->name, NXLT("tabdec"))) != 0 )
        return FALSE;
    else
        return TRUE;
}



void
FillButtonFace(HDC hdc, int x, int y, int width, int height,
                                    PIXEL_VALUE pixel_value)
/*** the hdc must be unpackaged! ***/
{
#if 1
    HBRUSH hFaceBrush, old_brush;
    HPALETTE old_palette;

    old_palette = SelectAndRealizeAxWinPalette(hdc);

    /* fill the face of the button */
    hFaceBrush = MS_CreateSolidBrush(GET_COLOR_VALUE(pixel_value));
    old_brush = MS_SelectObject(hdc, hFaceBrush);
    MS_PatBlt(hdc, x, y, width, height, PATCOPY);

    MS_SelectObject(hdc, old_brush);
    MS_DeleteObject(hFaceBrush);
    UnSelectPalette(hdc, old_palette);
#else
    RECT rect;
    HBRUSH hFaceBrush;
    HPALETTE old_palette;

    /* fill the face of the button */
    hFaceBrush  = MS_CreateSolidBrush(GET_COLOR_VALUE(pixel_value));

    rect.left   = x;
    rect.top    = y;
    rect.right  = x + width;
    rect.bottom = y + height;

    old_palette = SelectAndRealizeAxWinPalette(hdc);
    MS_FillRect(hdc, &rect, hFaceBrush);
    MS_DeleteObject(hFaceBrush);
    UnSelectPalette(hdc, old_palette);
#endif
}



/*****************************************************************
function: DrawButtonEvent - called in response to WM_DRAWITEM.
***/
BOOL 
DrawButtonEvent(WPARAM wParam, LPARAM lParam)
{
    int save_val, save_dc;
    LPDRAWITEMSTRUCT lpds;        /* ownerdraw structure   */
    AXWidget_pushbutton w_button;

    if ( wParam >= FIRST_WIDGET_ID )
        {
        /* wParam identifies widget */
        w_button = (AXWidget_pushbutton) WIDGET_FROM_UID(wParam);
        }
    else
        return (FALSE);
    if ( !w_button )
        taskAbort("DrawButtonEvent: no widget!");

    if ( !w_button->display )
        return TRUE;

    lpds = (LPDRAWITEMSTRUCT) lParam;
    save_dc = MS_SaveDC(lpds->hDC);
    w_button->window = AxPackageHandle(lpds->hDC, AX_DC_TYPE);

    switch (lpds->itemAction)
        {
        case ODA_DRAWENTIRE:
            PaintButton(w_button, 0, 0, FALSE);
            /* PaintButton will draw a "rounded" rect (i.e. one without *
             * corners), but Windows will have set the background to a  *
             * default of white. So make sure corners are painted with  *
             * the widget background color.                             */
            break;


        case ODA_FOCUS:
            if ( lpds->itemState == (UINT)ODS_FOCUS )
                {
                /* draw "focus" rectangle around button only if we *
                 * are NOT drawing onto a toolbar or GE palette.   */
                if ( w_button->parent->type != W_TOOLBAR &&
                        w_button->parent->type != W_PALETTE &&
                        w_button->parent->type != W_CORNER )
                    {
                    axmsDrawButtonFocusRect(w_button);
                    }
                }
            else if ( lpds->itemState == (UINT)0 )
                {
                save_val = w_button->value;
                /* button has lost focus: remove rect if need to   *
                 * and paint the button up (only if NOT in toolbar *
                 * or GE palette; keep button down in those cases).*/
                if ( w_button->parent->type != W_TOOLBAR &&
                        w_button->parent->type != W_PALETTE &&
                        w_button->parent->type != W_CORNER )
                    {
                    axmsUnDrawButtonFocusRect(w_button);
                    w_button->value = 0;    /* button up */
                    }

                if ( !w_button->disabled )
                    PaintButton(w_button, 0, 0, FALSE);
                w_button->value = save_val;
                }
            break;


        case ODA_SELECT:
            if ( lpds->itemState == (UINT)ODS_FOCUS )
                {
                /* draw button UP so we get the proper effect if *
                 * the user moves the mouse off the button while *
                 * still holding the mouse button down.          */
                save_val = w_button->value;
                w_button->value = 0;
                PaintButton(w_button, 0, 0, FALSE);
#if 0
                if ( w_button->parent->type != W_TOOLBAR &&
                        w_button->parent->type != W_PALETTE &&
                        w_button->parent->type != W_CORNER )
                    {
                    axmsUnDrawButtonFocusRect(w_button);
                    }
#endif
                w_button->value = save_val;
                }

            if ( lpds->itemState == (UINT)(ODS_SELECTED | ODS_FOCUS) )
                {
				if ( xmCanTakeKeyFocus(w_button->parent, w_button, TRUE) )
					{
					xmTakeKeyFocus((AXWidget)w_button,
								   (XButtonEvent*)NULL,
								   FALSE);
					}
                save_val = w_button->value;
                /* button down */
                if ( w_button->parent->type == W_CORNER )
                    {
                    /* if corner button presently up, paint it down. */
                    if ( w_button->value == 0 )
                        {
                        w_button->value = 1;    /* button down */
                        }
                    else
                        {
                        /* otherwise we are about to paint the corner *
                         * button down, so set its flag.              */
                        w_button->value = 0;    /* button up */
                        }
                    }
                else
                    w_button->value = 1;    /* button down */

                PaintButton(w_button, 0, 0, TRUE);
                /* draw "focus" rectangle around button only if we *
                 * are NOT drawing onto a toolbar or GE palette.   */
                if ( w_button->parent->type != W_TOOLBAR &&
                        w_button->parent->type != W_PALETTE &&
                        w_button->parent->type != W_CORNER )
                    {
                    axmsDrawButtonFocusRect(w_button);
                    }

                /* if this is the corner button, we don't want to restore *
                 * the value. It must stay in its toggled state.          */
                if ( w_button->parent->type != W_CORNER )
                    w_button->value = save_val;

                if ( w_button->parent->type == W_PALETTE )
                    {
                    /* to make sure GE palette buttons toggle properly. */
                    ButtonArmed = w_button;
                    }
                }
            break;
        }

    MS_RestoreDC(lpds->hDC, save_dc);
    if ( w_button->window != NULL )
        {
        REMOVE_DRAWABLE(w_button->window);
        w_button->window = NULL;
        }
    return TRUE;
}


/*************************************
function: msDrawBevel - draw a rectangular bevel, typically using
sun and shadow colors.  This is the ms subroutine for AxDrawBevel
*/
void
msDrawBevel(HDC in_hdc, PIXEL_VALUE sun_pixel, PIXEL_VALUE shadow_pixel,
            int x, int y, int width, int height, int thickness, int style)
{
    register int i;
    register int thickness2;
    register int thickness3;
    POINT points[3];
    HPEN hSunPen;
    HPEN hShadowPen;
    HPEN upper_left_pen, lower_right_pen, old_pen;
    int sun_thickness;
    HDC hdc;
    int need_to_delete_hdc;
    int delete_single_pen = FALSE;
    HPALETTE old_palette;


    if ( in_hdc == (HDC)NULL )
        return;

    hSunPen    = MS_CreatePen(PS_SOLID, 1, GET_COLOR_VALUE(sun_pixel));
    hShadowPen = MS_CreatePen(PS_SOLID, 1, GET_COLOR_VALUE(shadow_pixel));

    switch (style)
        {
        case BEVEL_IN:
            upper_left_pen = hShadowPen;
            lower_right_pen = hSunPen;
            break;

        case BEVEL_OUT:
            upper_left_pen = hSunPen;
            lower_right_pen = hShadowPen;
            break;

        case BEVEL_SIMPLE:
            /* a solid line rectangle */
            upper_left_pen = lower_right_pen =
                MS_CreatePen(PS_SOLID, 1, GET_COLOR_VALUE(ForePixel));
            delete_single_pen = TRUE;
            break;

        case BEVEL_NONE:
            /* a lift, in effect */
            upper_left_pen = lower_right_pen =
                MS_CreatePen(PS_SOLID, 1, GET_COLOR_VALUE(ButtonFacePixel));
            delete_single_pen = TRUE;
            break;
        }

    if ( thickness <= 0 )
        return;
    if ( thickness > width / 2 )
        thickness = width / 2;
    if ( thickness > height / 2 )
        thickness = height / 2;
    if ( thickness <= 0 )
        return;

    thickness2 = thickness + thickness;
    thickness3 = thickness2 + thickness;
    sun_thickness = thickness;

    hdc = (HDC)InsureDrawable(in_hdc, &need_to_delete_hdc);
    old_palette = SelectAndRealizeAxWinPalette(hdc);
    old_pen = MS_SelectObject(hdc, upper_left_pen);
	
	/* Left and Top segments  */
	/* we draw up the left side, then across the top */
    for ( i = 0; i < sun_thickness; i++ )
        {
        points[0].x = x + i;
        points[0].y = y + height - i - 1;
        points[1].x = points[0].x;
        points[1].y = y + i;
        points[2].x = x + width - i - 1;
        points[2].y = points[1].y;
        MS_Polyline(hdc, points, 3);
        }
	
   	/* Bottom segments  */
   	/* we draw across the bottom, then up the right side */
	if ( BEVEL_OUT == style && thickness > 1 )
		MS_SelectObject(hdc, MS_GetStockObject(BLACK_PEN));
	else
		MS_SelectObject(hdc, lower_right_pen);
	for ( i = 0; i < thickness; i++ )
		{
		if ( BEVEL_OUT == style && 1 == i )
			MS_SelectObject(hdc, lower_right_pen);
		points[0].x = x + i;
		points[0].y = y + height - i - 1;
		points[1].x = x + width - i - 1;
		points[1].y = points[0].y;
		points[2].x = points[1].x;
		points[2].y = y + i - 1;
		MS_Polyline(hdc, points, 3);
		}

    /* cleanup */
    UnSelectPalette(hdc, old_palette);
    MS_SelectObject(hdc, old_pen);
    if ( delete_single_pen )
        MS_DeleteObject(upper_left_pen);
    else
        {
        MS_DeleteObject(upper_left_pen);
        MS_DeleteObject(lower_right_pen);
        }
    if ( need_to_delete_hdc )
        RemoveDrawable(in_hdc, hdc);

    return;
}



static void
axmsDrawButtonFocusRect(AXWidget_pushbutton w)
{
	HBRUSH brush;
	brush = MS_GetStockObject(BLACK_BRUSH);
	axmsDrawButtonRect(w, brush);
}

static void
axmsUnDrawButtonFocusRect(AXWidget_pushbutton w)
{
	HBRUSH brush;
	brush = MS_CreateSolidBrush(GET_COLOR_VALUE(ButtonFacePixel));
	axmsDrawButtonRect(w, brush);
	MS_DeleteObject(brush);
}

static void
axmsDrawButtonRect(AXWidget_pushbutton w, HBRUSH brush)
{
	int delete_dc;
	HDC hdc;
	Window win;
	RECT rect;
	HPALETTE old_palette;

	rect.left = w->xpos - 1;
	rect.top = w->ypos - 1;
	rect.right = w->xpos + w->width + 1;
	rect.bottom = w->ypos + w->height + 1;

	if ( w->parent->type == W_CORNER )
		win = w->parent->parent->window;
	else
		win = w->parent->window;

	hdc = InsureDrawable(win, &delete_dc);
	if ( !hdc )
		return;
	old_palette = SelectAndRealizeAxWinPalette(hdc);
	FrameRect(hdc, &rect, brush);   /* GDI call */
	UnSelectPalette(hdc, old_palette);
	if ( delete_dc )
		RemoveDrawable(win, hdc);
}



void
PaintButtonPixmap(HBITMAP bitmap_handle, HDC hDC, int xpos, int ypos,
                    HBITMAP mask, unsigned long fg, unsigned long bg,
                    BOOL use_widget_colors)
/*** the hDC must be unpackaged! ***/
{
    BITMAP bm;
    HDC hdcMem;
    POINT point, points[3];
    DWORD rop_code;
    HGDIOBJ old_obj;
    HPALETTE old_palette;


    hdcMem = MS_CreateCompatibleDC((HDC)NULL);
    old_obj = MS_SelectObject(hdcMem, bitmap_handle);

    MS_GetObject(bitmap_handle, sizeof(BITMAP), (LPSTR) &bm);
    point.x = bm.bmWidth;
    point.y = bm.bmHeight;
    MS_DPtoLP(hDC, &point, 1);    /* device to logical coordinates */

    old_palette = SelectAndRealizeAxWinPalette(hDC);
    if ( bm.bmBitsPixel == 1 )
        {
        /* we have a mono image, so set the fore and back colors. */
        if ( use_widget_colors )
            {
            MS_SetBkColor(hDC, GET_COLOR_VALUE(ForePixel));
            MS_SetTextColor(hDC, GET_COLOR_VALUE(BackPixel));
            }
        else
            {
            /* use colors that were passed in. */
            MS_SetBkColor(hDC, GET_COLOR_VALUE(fg));
            MS_SetTextColor(hDC, GET_COLOR_VALUE(bg));
            }
        }

    if ( mask == (HBITMAP)NULL )
        {
        /* no mask, a simple transfer */
        rop_code = SRCCOPY;
        MS_BitBlt(hDC, xpos, ypos, point.x, point.y, hdcMem, 0, 0, rop_code);
        }
    else
        {
        /* a mask, draw the bitmap through the mask */
        points[0].x = xpos;              /* upper left corner */
        points[0].y = ypos;
        points[1].x = xpos + point.x;    /* upper right corner */
        points[1].y = ypos;
        points[2].x = xpos;
        points[2].y = ypos + point.y;    /* lower left corner */
        AxWinPlgBlt(hDC, points, hdcMem,0,0, point.x,point.y, mask,0,0);
        }

    /* cleanup */
    UnSelectPalette(hDC, old_palette);
    MS_SelectObject(hdcMem, old_obj);
    MS_DeleteDC(hdcMem);
}


#if 0
static void
DrawButtonCorners(HDC hdc, RECT *rect)
/*** the hdc must be unpackaged! ***/
{
    COLORREF color;
    color = BackPixelRGB;

    MS_SetPixel(hdc, rect->left, rect->top, color);
    MS_SetPixel(hdc, rect->right - 1, rect->top, color);
    MS_SetPixel(hdc, rect->right - 1, rect->bottom - 1, color);
    MS_SetPixel(hdc, rect->left, rect->bottom - 1, color);
}
#endif


static void
GetCenteredButtonTextPos(HDC hdc, AXWidget_pushbutton w,
                                int * new_x, int *new_y)
/*** the hdc must be unpackaged! ***/
{
    SIZE size;
    int ypos, xpos;

    MS_GetTextExtentPoint32(hdc, w->name, strlen(w->name), &size);
    xpos = (w->width - size.cx) / 2;
    ypos = (w->height - size.cy) / 2;
    *new_x = w->xpos + xpos;
    *new_y = w->ypos + ypos;
}
