/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winclip.c
   Applix ASC version: 10
   Date: Mon Nov 16 13:53:46 1998
*/
#include <aos.h>
#include <basicwin.h>
#include <ax_xm.h>
#include <doctypes.h>
#include <aos_xfer.h>
#include <xmlib.h>
#include <winclip.h>
#include <scac.h>
#include <fsrts.h>
#include <paratm.h>
#include <parint.h>
#include <adocio.h>
#include <thimble.h>
#include <wp.h>

#include <winclip.p>

extern int Running_Windows95;
extern int Ax_Clear_Selection; /* whether to process WM_DESTROYCLIPBOARD */
//extern Atom AxXA_EXCEL_DATA; /* Atom for registering "XlTable" data. */

extern DocInfoType XferDoc; /* transfer bin document wrapper */
extern HWND WXDummyWnd;     /* window handle of AX_ROOT_WINDOW_ID */
extern char *ThimApplixDir;




/******************************************************************
 ***   Windows  CLIPBOARD EVENT FUNCTIONS                       ***
 ******************************************************************/

/************************************************************
 * called during WM_RENDERFORMAT message. for more info see
 * comments there (WorkClassProc).
 */
void
DoTaskXfer(struct wind_struct *winP, UINT cb_format, AXWidget w)
{
    int winID;
    int save_THIMpid;
    XEvent xevent;
    xfer_data_convert_msg_t xferMsgP;
    static char pathname[STRPATH+1];
    Atom x_atom;
    Window owner_window;


    if ( WINDOWS_CLIPBOARD_FORMAT(cb_format) )
        x_atom = cb_format;
    else
        /* get one of the registered formats. */
        x_atom = AxWinGetAtomId((int)cb_format);

    owner_window = w->window;
    /* fake a SelectionRequest event. we do this so the xfer code *
     * can set the proper selection pointer. we set the target to *
     * zero so that no messages are sent. we will call the actual *
     * task xfer function directly.        -donato                */
    xevent.type = SelectionRequest;
    xevent.xselectionrequest.requestor = owner_window;
    xevent.xselectionrequest.selection = XA_CLIPBOARD;
    xevent.xselectionrequest.target    = x_atom;
    xevent.xselectionrequest.property  = x_atom;
    AxSelectionEvent(&xevent);

    /* file object for data */
    xferMsgP.code = XFER_CODE_COPY;    /* ask for clipboard material. */
    xferMsgP.binType = XFER_BIN_CLIPBOARD;
    if ( cb_format == CF_TEXT )
        xferMsgP.type = XA_STRING;
    else
        xferMsgP.type = x_atom;
    xferMsgP.binDoc = (char*)(&XferDoc);
    /* XA_CLIPBOARD selection code = "C" */
    sprintf(pathname, NXLT("%s/%d%s"), XferDoc.pathname, (int)x_atom, "C");
    strcpy(xferMsgP.pathname, pathname);

    /* if the format type is non-applix, let's change it to an *
     * equivalent applix format (e.g. RTF = WP) so the XferOut *
     * code can get the proper data.                           */
    change_xfer_type(&xferMsgP);

    /* set the per-task context, and then call the *
     * interface function in that task.            */
    save_THIMpid = THIMpid;
    VfyContext(winP->p_id);
    winID = WindowIdFromShellWidget((long)w->parent);
    if ( winP->ap_id == APGE )
        {
        doXferOut(winID, &xferMsgP);
        }
    else if ( winP->ap_id == APDC || winP->ap_id == APUCM ||
                                        winP->ap_id == APHTML )
        {
		/* Words or macro editor or HTML author */
        WpXferOut(winID, &xferMsgP);
        }
    else if ( winP->ap_id == APSS )
        {
        SSnewXferOut(winID, &xferMsgP);
        } 
    VfyContext(save_THIMpid);
}

#if 0   /*  {  */
void
AxWinSendRequestEvent(int winID, Atom property)
{
    XEvent xevent;
    Window win;

    win = XWinFromATMWin(winID);
    /* fake a SelectionRequest event. we do this so the xfer code *
     * can set the proper selection pointer. we set the target to *
     * zero so that no messages are sent. we will call the actual *
     * task xfer function directly.        -donato                */
    xevent.type = SelectionRequest;
    xevent.xselectionrequest.requestor = win;
    xevent.xselectionrequest.selection = XA_CLIPBOARD;
    xevent.xselectionrequest.target    = 0;
    xevent.xselectionrequest.property  = property;
    AxSelectionEvent(&xevent);
}
#endif    /*  }  */





/***********************************************************************
 *** this is called in response to a copy or cut. Open the clipboard
 *** and establish which formats we know.
 ***********************************************************************/
void
SetWinClipboardOwner(int winID, int *dataTypes)
{
    int ix;
    AXATOMLIST *aP;
    HWND hwnd;
    Window win;
    UINT win_format_id;


    if ( winID == AX_ROOT_WINDOW_ID )
        {
        /* clipboard owner will be root window. *
         * that is, the ever present unmapped window. */
        hwnd = WXDummyWnd;
        }
    else
        {
        /* find the window (device context) associated with current app. */
        win = XWinFromATMWin(winID);
        /* Get hwnd so we can make it the current clipboard owner */
        /* and Windows knows who to call to give up ownership */
        hwnd = AxGetHWNDFromHDC(win);
        if ( !hwnd )
            return;
        }

    if ( !MS_OpenClipboard(hwnd) )
        return;
    Ax_Clear_Selection = FALSE;
    MS_EmptyClipboard();    /* this will send a WM_DESTROYCLIPBOARD       *
                             * message to WorkClassProc and make hwnd the *
                             * new clipboard owner.                       */

    for ( ix=0; dataTypes[ix] != None; ix++ )
        {
        aP = AXGetAtomElement(dataTypes[ix]);
        if ( aP )
            {
            MS_SetClipboardData(aP->winFormatID, (HANDLE)NULL);

            /* register some compatible formats. */
            if ( aP->atom == XA_STRING )
                {
                /* if we can do X ascii text, we can do Windows ascii text. */
                MS_SetClipboardData(CF_TEXT, (HANDLE)NULL);
                }
#if 0
            /* the WP rtf filter will not run on the partial-wp doc which */
            /* the WP actually puts on the clipboard                      */
            if ( aP->atom == AxXA_WP_FILE )
                {
                /* if we can do WP data, we can do RTF data. */
                win_format_id = AxWinGetWinFormatId(AxXA_RTF_DATA);
                MS_SetClipboardData(win_format_id, (HANDLE)NULL);
                }
#endif
            if ( aP->atom == AxXA_GEGE_FILE )
                {
                /* if we can do GE data, we can do Windows Metafile data. */
                MS_SetClipboardData(CF_METAFILEPICT, (HANDLE)NULL);
                }
            }
        else
            taskAbort(NXLT("SetWinClipboardOwner: Unknown data type"));
        }

    MS_CloseClipboard();
    Ax_Clear_Selection = TRUE;
}


void
SupplyExitClipboardData(UINT cb_format)
{
    int data_type;
    int fd, nbytes, created;
    xfer_data_convert_msg_t info;
    char *buf;


    if ( WINDOWS_CLIPBOARD_FORMAT(cb_format) )
        {
        if ( cb_format == CF_TEXT )
            data_type = XA_STRING;
        else
            data_type = cb_format;
        }
    else
        {
        /* some registered format. */
        data_type = AxWinGetAtomId(cb_format);
        }

    if ( !GetExitClipboardSelection(data_type, &info) )
        return;    /* can't get this type of data */

    /* how much data is there to send */
    fd = OpenFileObject(&XferDoc, info.pathname, DFM_READ, &created, &nbytes);
    if ( fd == -1 || created )
        { /* no data available */
        if ( created )
            {
            CloseFileObject(fd);
            DeleteFileObject((DocInfoType*)info.binDoc, info.pathname);
            }
        }

    buf = malloc(nbytes);
    nbytes = DocSegRead(fd, buf, nbytes, -1);
    CloseFileObject(fd);
    buf[nbytes] = '\0';

    /* put data in global memory. */
    XChangeProperty((Display*)NULL, (Window)NULL, cb_format, 0, 0, 0,
                                                        buf, nbytes);
    free(buf);
}

void
RenderExitClipboardData(void)
{
    //SetupWinClipboardForExit(winId);
}



/**************************************************************************
 a debug function: tells amount of formats on clipboard and what they are.
                           ***********************************************/
void
ShowClipboardFormats(void)
{
    int numcbf;
    UINT cbf;
    char buf[100], cbbuf[80];
    HWND non_ax_wnd;


    non_ax_wnd = MS_GetClipboardOwner();
    MS_OpenClipboard(non_ax_wnd);
    cbf = 0;
    numcbf = MS_CountClipboardFormats();
    sprintf(buf, "number of formats on the clipboard: %d", numcbf);
    DebugMessage(buf);

    while ( (cbf = MS_EnumClipboardFormats(cbf)) )
        {
        MS_GetClipboardFormatName(cbf, cbbuf, 80);
        sprintf(buf, "cb format id: %d, name: %s", cbf, cbbuf);
        DebugMessage(buf);
        } 
    MS_CloseClipboard();
}


/****************** clipboard conversion functions ***************/
extern data_ptr AxfEncryptTime(void);
extern char *AxStrFromDataPtr(register data_ptr);


static void
convertRTFtoWP(xfer_data_convert_msg_t *infoP)
{
    char cmdbuf[STRPATH+1];
    char rtf_tempfile[STRPATH+1], wp_tempfile[STRPATH+1];
    char encrypt_time[80];
    typeErr wrt_err, rd_err;
    data_ptr dataP;


    AxTempFile(NULL, rtf_tempfile);
    /* write the file object containing the RTF data to a temp file, *
     * so we can pass it to the rtf filter.                          */
    wrt_err = FSWriteFileObjAsFile(rtf_tempfile,
                        AxChmodValueFromCodes(rtf_tempfile, 2, 2),
                        (DocInfoType*)infoP->binDoc, infoP->pathname);
    if ( wrt_err == AOK )
        {
        AxTempFile(NULL, wp_tempfile);
        dataP = AxfEncryptTime();
        strcpy(encrypt_time, AxStrFromDataPtr(dataP));
        /* where is the filter on a customer's machine? */
        sprintf(cmdbuf, "%s\\filters\\wpimport rtf2wp %s %s -s %s %s",
                            ThimApplixDir, rtf_tempfile, wp_tempfile,
                            ThimApplixDir, encrypt_time);
        AxRunProgram(cmdbuf, NULL, NULL, NULL, FALSE);    /* run the filter. */

        /* read the now converted temp file back into a file obj ptr. */
        rd_err = FSReadFileIntoFileObj(wp_tempfile,
                                        (DocInfoType*)infoP->binDoc,
                                        infoP->pathname);
        if ( rd_err != AOK )
            taskAbort(NXLT("convertRTFtoWP: could not read into file object!"));

        infoP->docType = AX_DOC_TYPE_ASTERX_WORDS;

        AxDeleteTempfile(wp_tempfile);
        AxDeleteTempfile(rtf_tempfile);
        }
    else
        taskAbort(NXLT("convertRTFtoWP: could not write data file!"));
}

static void
convertWPtoRTF(xfer_data_convert_msg_t *infoP)
{
    char cmdbuf[STRPATH+1];
    char rtf_tempfile[STRPATH+1], wp_tempfile[STRPATH+1];
    typeErr wrt_err, rd_err;
    char encrypt_time[80];
    data_ptr dataP;


    AxTempFile(NULL, wp_tempfile);
    /* write the file object containing the Applix WP data to a temp file, *
     * so we can pass it to the rtf filter.                          */
    wrt_err = FSWriteFileObjAsFile(wp_tempfile,
                        AxChmodValueFromCodes(wp_tempfile, 2, 2),
                        (DocInfoType*)infoP->binDoc, infoP->pathname);
    if ( wrt_err == AOK )
        {
        AxTempFile(NULL, rtf_tempfile);
        dataP = AxfEncryptTime();
        strcpy(encrypt_time, AxStrFromDataPtr(dataP));
        sprintf(cmdbuf,"%s\\filters\\wp2rtf.exe %s %s -s %s %s",ThimApplixDir,
                    wp_tempfile, rtf_tempfile, ThimApplixDir, encrypt_time);
        AxRunProgram(cmdbuf, NULL, NULL, NULL, FALSE);    /* run the filter. */

        /* read the now converted temp file back into a file obj ptr. */
        rd_err = FSReadFileIntoFileObj(rtf_tempfile,
                                        (DocInfoType*)infoP->binDoc,
                                        infoP->pathname);
        if ( rd_err != AOK )
            taskAbort(NXLT("convertWPtoRTF: could not read into file object!"));

        AxDeleteTempfile(wp_tempfile);
        AxDeleteTempfile(rtf_tempfile);
        }
    else
        taskAbort(NXLT("convertRTFtoWP: could not write data file!"));
}


static void
convertWMFtoGE(xfer_data_convert_msg_t *infoP)
{
    char *wmf_tempfile;
    char cmdbuf[STRPATH+1], ge_tempfile[STRPATH+1];
    typeErr rd_err;
    int buf_size;


/* the WMF file was created during XGetWindowProperty (that is, when  *
 * the metafile handle was retrieved from the Clipboard). the name of *
 * that file is in the file object pointer.                           */

    /* get the WMF file name from the file object. */
    wmf_tempfile = FSBufferFromFileObject(THIMpid, (DocInfoType*)infoP->binDoc,
                                            infoP->pathname, &buf_size);
    wmf_tempfile[buf_size] = 0;

    AxTempFile(NULL, ge_tempfile);
    sprintf(cmdbuf, "%s\\filters\\geinout.exe %s %s -i wmf -s %s",
            ThimApplixDir, wmf_tempfile, ge_tempfile, ThimApplixDir);
    AxRunProgram(cmdbuf, NULL, NULL, NULL, FALSE);    /* run the filter. */

    /* read the now converted temp file back into a file obj ptr. */
    rd_err = FSReadFileIntoFileObj(ge_tempfile, (DocInfoType*)infoP->binDoc,
                                                        infoP->pathname);
    if ( rd_err != AOK )
        taskAbort(NXLT("convertWMFtoGE: could not read into file object!"));

    /* selected material is now a GE file. */
    infoP->type = AxXA_GEGE_FILE;
    infoP->docType = AX_DOC_TYPE_ASTERX_GRAPHICS;

    AxDeleteTempfile(ge_tempfile);
    AxDeleteTempfile(wmf_tempfile);
    THIMfree(wmf_tempfile);
}

static void
convertGEtoWMF(xfer_data_convert_msg_t *infoP)
{
    int fd;
    char cmdbuf[STRPATH+1];
    char wmf_tempfile[STRPATH+1], ge_tempfile[STRPATH+1];
    typeErr wrt_err;


    AxTempFile(NULL, ge_tempfile);
    /* write the file object containing the Applix GE data to a temp file, *
     * so we can pass it to the wmf filter.                                */
    wrt_err = FSWriteFileObjAsFile(ge_tempfile,
                        AxChmodValueFromCodes(ge_tempfile, 2, 2),
                        (DocInfoType*)infoP->binDoc, infoP->pathname);
    if ( wrt_err == AOK )
        {
        AxTempFile(NULL, wmf_tempfile);
        /* Note: the '-f' tells the filter not to write out an Aldus file
		 * header because the Clipboard MetaFile format does not have a header.
         */
        sprintf(cmdbuf, "%s\\filters\\geinout.exe %s %s -o wmf -s %s -f",
                ThimApplixDir, ge_tempfile, wmf_tempfile, ThimApplixDir);
        AxRunProgram(cmdbuf, NULL, NULL, NULL, FALSE);    /* run the filter. */

        /* write out the name of the newly created metafile.  *
         * XChangeProperty will read the file, and delete it. */
        fd = (int)OpenFileObjFmode(&XferDoc, infoP->pathname, "w");
        if ( fd )
            {
            DocSegWrite(fd, wmf_tempfile, strlen(wmf_tempfile) + 1, -1);
            CloseFileObject(fd);
            }
        }
    AxDeleteTempfile(ge_tempfile);
}


static void
convertBMPtoGE(xfer_data_convert_msg_t *infoP)
{
    char *bmp_tempfile;
    char cmdbuf[STRPATH+1], ge_tempfile[STRPATH+1];
    typeErr rd_err;
    int buf_size;


/* the BMP file was created during XGetWindowProperty (that is, when *
 * the bitmap handle was retrieved from the Clipboard). the name of  *
 * that file is in the file object pointer.                          */

    /* get the BMP file name from the file object. */
    bmp_tempfile = FSBufferFromFileObject(THIMpid, (DocInfoType*)infoP->binDoc,
                                                infoP->pathname, &buf_size);
    bmp_tempfile[buf_size] = 0;

    AxTempFile(NULL, ge_tempfile);
    sprintf(cmdbuf, "%s\\filters\\geinout.exe %s %s -i bmp -s %s",
            ThimApplixDir, bmp_tempfile, ge_tempfile, ThimApplixDir);
    AxRunProgram(cmdbuf, NULL, NULL, NULL, FALSE);    /* run the filter. */

    /* read the now converted temp file back into a file obj ptr. */
    rd_err = FSReadFileIntoFileObj(ge_tempfile, (DocInfoType*)infoP->binDoc,
                                                        infoP->pathname);
    if ( rd_err != AOK )
        taskAbort(NXLT("convertBMPtoGE: could not read into file object!"));

    /* selected material is now a GE file. */
    infoP->type = AxXA_GE_FILE;

    AxDeleteTempfile(ge_tempfile);
    AxDeleteTempfile(bmp_tempfile);
    THIMfree(bmp_tempfile);
}

static void
convertXLtoSS(xfer_data_convert_msg_t *infoP)
{
}


/* needed for applix-->non-applix transfers. */
void
change_xfer_type(xfer_data_convert_msg_t *infoP)
{
    if ( infoP->type > 0 )
        {
        if ( infoP->type == (int)AxXA_RTF_DATA )
            infoP->type = AxXA_WP_FILE;

        //else if ( infoP->type == (int)AxXA_EXCEL_DATA )
        //    infoP->type = AxXA_SS_FILE;

        else if ( infoP->type == CF_METAFILEPICT ||
                        infoP->type == CF_BITMAP )
            infoP->type = AxXA_GEGE_FILE;
        }
}


int
convert_to_applix_format(xfer_data_convert_msg_t *infoP)
{
    int retval = FALSE;
    if ( infoP->type > 0 )
        {
        if ( infoP->type == (int)AxXA_RTF_DATA )
            {
            convertRTFtoWP(infoP);
            retval = TRUE;
            }
#if 0
        else if ( infoP->type == (int)AxXA_EXCEL_DATA )
            {
            convertXLtoSS(infoP);
            retval = TRUE;
            }
#endif
        else if ( infoP->type == CF_METAFILEPICT )
            {
            convertWMFtoGE(infoP);
            retval = TRUE;
            }

        else if ( infoP->type == CF_BITMAP )
            {
            convertBMPtoGE(infoP);
            retval = TRUE;
            }
        }
    return retval;
}

void
convert_to_non_applix_format(xfer_data_convert_msg_t *infoP, Atom target)
{
    /* make sure we have something to transfer. */
    if ( infoP->type > 0 )
        {
        /* xfer transfer data type should be same as target data type.   *
         * if they are not, then xfer type was previously changed (using *
         * change_xfer_type) to do a applix-->non_applix conversion.     *
         * so set xfer type back to original target.                     */
        if ( infoP->type != (int)target )
            infoP->type = target;
 
        if ( infoP->type == (int)AxXA_RTF_DATA )
            convertWPtoRTF(infoP);

        else if ( infoP->type == CF_METAFILEPICT )
            convertGEtoWMF(infoP);
        }
}
/****************** end clipboard conversion functions ***************/


#if 0
/************************* DDE ***************************/
int
ax_dde_request(HDC hdc, UINT cb_format)
{
    int retval;
    HWND hwnd, ax_hwnd;
    AXWidget w;
    ATOM atom, topic;

#if 0
    hwnd = MS_GetClipboardOwner();
    w = GetWidgetFromDC(hdc);
    if ( w )
        ax_hwnd = w->winid;
    else
        {
        /* search total windows list first. */
        ax_hwnd = GetHWNDFromTotalWindowsList(hdc);
        if ( !ax_hwnd )
            /* now search WindowList (widgets that have windows). */
            ax_hwnd = AxGetHWNDFromHDC(hdc);
        }
    if ( !ax_hwnd )
        /* else abort... */
        taskAbort(NXLT("ax_dde_request: cannot find window!"));

    /* first, initiate conversation. */
    //topic = GlobalAddAtom(NXLT("DataObject"));
    MS_SendMessage(hwnd, WM_DDE_INITIATE, (WPARAM)ax_hwnd,
                                MAKELPARAM(NULL, NULL));

    /* request the data. */
    atom = GlobalAddAtom(NXLT("AX_DDE_ATOM"));
    retval = MS_PostMessage(hwnd, WM_DDE_REQUEST, (WPARAM)ax_hwnd,
                                    MAKELPARAM(cb_format, atom));
#endif
    if ( retval )
        return Success;
    else
        return !Success;
}
#endif


/*****************************************************************************
	Explanation of the Windows 95 "Clipboard hack":  for some undiscovered
reason, if we try to proc a new task from a task that has Clipboard ownership
the longjmp (SuspendTask) back to the scheduler fails; that is, we crash
somewhere in the Windows "unwind" code. So here we give up the clipboard
contents as if we had exited the app that did the cut/copy.  This seems
to "fix" the problem.  It would be nice to know what's really going on
with longjmp on Win95.					-donato
(NOTE: this function <<must>> be called by the task that has Clipboard
ownership!)
*****************************************************************************/
void
AxWin95ClipboardHack(int binType)
{
	if ( Running_Windows95 && (binType != XFER_BIN_SELECTION) )
		{
		int win_id = WindowIdFromPid(THIMpid);
		int app_id = AppIdFromATMWin(win_id);
	    if ( app_id == APGE )
			{
			AxMoveClipboardToExitClipboard(win_id, doXferOut);
			}
		else if ( app_id == APDC || app_id == APUCM || app_id == APHTML )
        	{
			AxMoveClipboardToExitClipboard(win_id, WpXferOut);
    	    }
		else if ( app_id == APSS )
			{
			AxMoveClipboardToExitClipboard(win_id, SSnewXferOut);
			}
		}
}
/****************************************************************************/
