/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: windlg.c
   Applix ASC version: 203
   Date: Mon May 10 21:13:06 1999
*/
#define NEED_X_INFO 1
#include <stdlib.h>
#include <string.h>
#include <basicwin.h>
#include <ax_xm.h>
#include <aos.h>
#include <xmlib.h>
#include <xmdlged.h>
#include <elxwins.h>
#include <thimble.h>
#include <dwt2xm.h>
#include <axwinclr.h>



extern int THIMhelpMode;
extern HWND CurrentDialogHwnd; /* currently active dialog's window handle */
extern HANDLE hInst;
extern AXWidget WindowList;
extern AXWidget_scale ScaleArmed;
extern AXWidget_splitter SplitterArmed;
extern AXWidget_pushbutton ToolBarButtonWidget;
extern AXWidget_pushbutton ButtonArmed;
extern eventHandlerList ehList;
extern int NtLineHeight;
extern int AxCharHeight;

extern PIXEL_VALUE BackPixel, BackWorkPixel, ForePixel, ForeWorkPixel;
/******************************* 
 * the DialogClassProc will never receive a WM_DESTROY message,
 * so we can't delete any static brushes there. The destroy message
 * will instead go to the MainClassProc, so we keep this global here,
 * and a destoyer function below called DeleteDlgBrush to be called
 * from MainClassProc.               -donato    ***********************/
static HBRUSH DlgWidgetBrush = NULL;
static HBRUSH DlgEditBrush = NULL;
static HPALETTE OldDialogPalette;

extern HFONT NormalDialogFont;
extern Cursor ThimPointerCursor;
extern int RunningWin95Look;

#include <windlg.p>




/****************************************************************************
 FUNCTION: DialogClassProc(HWND, UINT, WPARAM, LPARAM)
 PURPOSE:  callback for dialog class windows
****************************************************************************/
LRESULT CALLBACK
DialogClassProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    HDC hDC;           /* display-context variable  */
    HWND curr_hwnd;
    BOOL bEvent;
    PAINTSTRUCT ps;    /* paint structure */
    int widgetId;
    //POINT point;
    XEvent xevent;
    AXWidget widget;
    AXWidget_main w_main;
    AXWidget_dialog w_dlg;
    XmAnyCallbackStruct data;
    LRESULT retval;
    eventRec *ep;
    LPMEASUREITEMSTRUCT lpms;
    unsigned long pixel;    /* for use during the WM_CTLCOLOR events. */
    static int *consumed;
    //mbchar *strP;
    HDC DlgHdc;


    /* Find the dialog widget associated with the event */
    w_dlg = (AXWidget_dialog) GetWidgetFromHWND(hWnd, W_DIALOG);
    if ( !w_dlg )
        {
        /* hasn't been painted yet, ignore event. */
        return MS_DefWindowProc(hWnd, message, wParam, lParam); 
        }
    else
        {
        if ( w_dlg->parent )
            w_main = w_dlg->parent;
        else
            taskAbort(NXLT("DialogClassProc: no main widget!"));
        }

    if ( w_main->dialog->winid != hWnd )
       taskAbort("DlgEvent window mismatch");

    memset(&xevent, 0, sizeof(XEvent));
    bEvent = WinToXEvent(hWnd,message, wParam,lParam, &xevent, (AXWidget)NULL);
    if ( bEvent && w_dlg->event_table )
        {
        /* this "dialog" is really a dialog editor window  */
        if ( bEvent == -2 )
           /* special flag meaning that we doo not want to do anything with */
           /* this message since the mouse seems to be being dragged        */
           return( FALSE );
 
        if ( message == WM_PAINT )
            MS_BeginPaint(hWnd, &ps);

        /* dialog editor event handling */
        ep = w_dlg->event_table;
        (*ep->proc)(w_main, ep->clientData, &xevent, &consumed); 

        if ( message == WM_PAINT )
            MS_EndPaint(hWnd, &ps);
        return 0;
        }

    retval = 0;
    switch (message)
        {
    case WM_INITDIALOG:
        /* set the font to our dialog font, "MS Sans Serif" like other MS */
        /* apps seem to do.  They do NOT use the SYSTEM_FONT for dialog text */
        MS_SendMessage( hWnd, WM_SETFONT, (WPARAM)NormalDialogFont, (LPARAM)0 );
        break;

    case WM_LBUTTONUP:
    case WM_MOUSEMOVE:
        /* only process scale and splitter widgets if mouse *
         * button is released, or if mouse is moving while  *
         * left button is pressed.                          */
        if ( message == WM_LBUTTONUP ||
                (message == WM_MOUSEMOVE && wParam == MK_LBUTTON) )
            {
            if ( ScaleArmed )
                {
                ArmScale(ScaleArmed, &xevent);
                }
            else if ( SplitterArmed )
                {
                ArmSplitter((AXWidget_splitter)SplitterArmed, &xevent);
                }
            }
        break;


    case WM_CTLCOLORSTATIC:
    case WM_CTLCOLORBTN:
    case WM_CTLCOLOREDIT:
    case WM_CTLCOLORLISTBOX:
        {
        DlgHdc = (HDC)wParam;
        /* Find the entry widget associated with the event */
        widget = GetGadgetFromHWND((AXWidget)w_dlg, (HWND)lParam);
#if 0
        if ( !widget || (widget && !widget->textColor && !widget->widgetColor) )
            {
            retval = MS_DefWindowProc(hWnd, message, wParam, lParam);
            break;
            }
#endif
        if ( DlgEditBrush )
            {
            /* PARANOIA */
            MS_DeleteObject(DlgEditBrush);
            DlgEditBrush = NULL;
            }
        OldDialogPalette = SelectAndRealizeAxWinPalette(DlgHdc);

        /* set the foreground color. */
        if ( widget && (widget->type == W_OPTION) &&
                ((AXWidget_option)widget)->labelColor )
            pixel = ((AXWidget_option)widget)->labelColor->any.pixel;
        else if ( widget && widget->textColor )
            pixel = widget->textColor->any.pixel;
        else if ( RunningWin95Look && widget && widget->disabled )
            pixel = ForePixel;
        else
            pixel = ForeWorkPixel;
        MS_SetTextColor(DlgHdc, GET_COLOR_VALUE(pixel));

        /* if we have a widget with a custom color on it, and we are
         * NOT painting the "static label" portion of it, then use
         * the custom Applix color. if we are dealing with an edit widget,
         * use the background window color; all other widgets use no color,
         * that is transparent.     -donato
         */
        if ( widget && widget->widgetColor && message != WM_CTLCOLORSTATIC )
            pixel = widget->widgetColor->any.pixel;
        else if ( RunningWin95Look && widget && widget->disabled )
            pixel = BackPixel;
        else if ( message == WM_CTLCOLOREDIT || message == WM_CTLCOLORLISTBOX )
            pixel = BackWorkPixel;
        else if ( widget && widget->type == W_TOGGLE )
            pixel = BackPixel;
        else
            {
            MS_SetBkMode(DlgHdc, TRANSPARENT);
            /* return now with the transparent brush (a stock object)
             * and don't create the DlgEditBrush so we don't try to delete
             * it later.            -donato
             */
            return (DWORD)MS_GetStockObject(NULL_BRUSH);
            }

        DlgEditBrush = MS_CreateSolidBrush(GET_COLOR_VALUE(pixel));
        MS_SetBkColor(DlgHdc, GET_COLOR_VALUE(pixel));

        retval = (DWORD)DlgEditBrush;
        break;
        }

    case WM_COMMAND:
        widgetId = LOWORD(wParam);
        if ( widgetId < FIRST_WIDGET_ID )
            {
            if ( wParam == IDOK )
                {
                /* the RETURN key was pressed in the   *
                 * dialog, so activate default button. */
                if ( w_main->dialog->currentDefaultButton )
                    {
                    /* get current window with focus. */
                    curr_hwnd = MS_GetFocus();
                    /* do the callback. */
                    MS_SendMessage(w_main->dialog->winid, (UINT)WM_COMMAND,
                    	(WPARAM)w_main->dialog->currentDefaultButton->uid, 0L);
                    /* simulate button press, which fill give focus *
                     * to the default button.                       */
                    MS_SendMessage(w_main->dialog->currentDefaultButton->winid,
                                                WM_LBUTTONDOWN, 0, 0L);
                    /* set focus back to window that had it originally. */
                    MS_SetFocus(curr_hwnd);
                    }
				else if ( (widget = xmKeyFocusWidget(w_main->dialog)) &&
											widget->type == W_BUTTON )
					{
					/* no default button but a button has key focus so
					 * we still need to activate it.
					 */
                    MS_SendMessage(w_main->dialog->winid, (UINT)WM_COMMAND,
                    								(WPARAM)widget->uid, 0L);
                    /* simulate button press, which fill give focus *
                     * to the default button.                       */
                    MS_SendMessage(widget->winid, WM_LBUTTONDOWN, 0, 0L);
                                                
					}
                }
            else if ( wParam == IDCANCEL )
                {
                /* the ESCAPE key was pressed in the      *
                 * dialog, so activate the cancel button. */
                if ( w_main->dialog->cancelButton )
                    {
                    MS_SendMessage(w_main->dialog->winid, (UINT)WM_COMMAND,
                            (WPARAM)w_main->dialog->cancelButton->uid, 0L);
                    }
                }                
            else
                {
                /* this must be an event on the menu bar */
                taskAbort("DialogClassProc: called with menu bar event");
                }
            }
        else
            {
            /* an event on the widget */
            DialogWidgetCallback(wParam);                
            }
        break;
                             

    case WM_VKEYTOITEM:
        /* a key was pressed while a listbox had focus */
        if ( bEvent )
            {
            /* the key is one we are interested in. */
            DialogKeyEvent(w_main, hWnd, message, wParam, lParam);
            retval = -2;     /* indicates that we've done everything,  *
                              * the Windows listbox has to do nothing. */
            }
        else
            retval = -1;    /* do Windows default action. */
        break;


    case WM_DRAWITEM:
        retval = DoOwnerdrawEvent(wParam, lParam);
        break;

    case WM_MEASUREITEM:
        {
        AXWidget w = (AXWidget)WIDGET_FROM_UID(wParam);
        if ( w && w->type == W_LIST )
            {
            /* fill out measure-item struct for the listbox. */
            lpms = (LPMEASUREITEMSTRUCT)lParam;
            lpms->itemHeight = xmListBoxLineHeight((AXWidget_list)w);
            lpms->itemData = 0;
            retval = TRUE;
            }
        else if ( w && w->type == W_OPTION )
            {
            /* fill out measure-item struct for the listbox. */
            lpms = (LPMEASUREITEMSTRUCT)lParam;
            lpms->itemHeight = xmOptionMenuLineHeight((AXWidget_option)w);
            lpms->itemData = 0;
            retval = TRUE;
            }
        break;
        }

    case WM_LBUTTONDOWN:
		w_dlg->hasFocus = TRUE;   /* force true on click */
        widget = (AXWidget)FindGadget(w_dlg->winid, xevent.xbutton.x,
                                                     xevent.xbutton.y);
        if ( widget )
            { 
            if ( xmCanTakeKeyFocus(w_dlg, widget, TRUE) ) 
				{
                xmTakeKeyFocus(widget, &(xevent.xbutton), FALSE);
            	}

            if ( widget->type == W_SCALE )
				{
                ArmScale((AXWidget_scale)widget, &xevent);
                MS_SetFocus(w_main->winid);
				}

            else if ( widget->type == W_SPLITTER )
				{
                ArmSplitter((AXWidget_splitter)widget, &xevent);
                MS_SetFocus(w_main->winid);
				}

            else if ( widget->type == W_ROWCOL )
                {
                PaintRowColumnPulldown((AXWidget_rowcol)widget, 
                                          OPTION_NEW, &xevent);
                MS_SetFocus(w_main->winid);
                }
            else if ( widget->type == W_POPUP )
                {
                data.reason = 0;
                data.event = &xevent;
                (*widget->callback) (widget, widget->callbacktag, &data);
                }
            else if ( widget->type == W_RECT )
                {
                MS_SetFocus(w_main->winid);
                RectEvent((AXWidget_rect)widget, &xevent); 
                }
            }
        else
            {
			/* we didn't click on a specific widget, but still need to
			 * give focus to the one that has it.
			 */
			widget = xmKeyFocusWidget(w_dlg);
			if ( widget )
				MS_SetFocus(widget->winid);
			else
            	MS_SetFocus(w_main->winid);
            }
        break;


    case WM_PAINT:
        hDC = MS_BeginPaint(hWnd, &ps);
        PaintDialog(w_dlg, hDC);
        MS_EndPaint(hWnd, &ps);
        break;


    case WM_VSCROLL:
    case WM_HSCROLL:
        {
        int sub_command;

        sub_command = LOWORD(wParam);
        if ( sub_command == SB_ENDSCROLL )
            break;
        if ( sub_command == SB_THUMBPOSITION )
            sub_command = LOWORD(wParam);  /* handy place to stop debugger */

        DialogScrollCallback(w_dlg, wParam, lParam);
        break;
        }


    default:
        retval = MS_DefWindowProc(hWnd, message, wParam, lParam);
        break;
        }
    return retval;
}


static void
PaintDialog(AXWidget_dialog w_dlg, HDC hDC)
/** HDC hDC;      use this for painting if not null **/
{
    AXWidget childP;
    wLayer *ly;
    HDC save_hdc;


    WIDGET_PARANOID(w_dlg, W_DIALOG);

    save_hdc = w_dlg->window;
    if ( hDC != (HDC)NULL )
        w_dlg->window = AxPackageHandle(hDC, AX_DC_TYPE);

    /* create any widgets that haven't been created yet */
    if ( w_dlg->winid )
        {
        if ( w_dlg->layerList )
            {
            /* expose in back-to-front order */
            for ( ly = w_dlg->layerList; ly; ly = ly->next)
                {
                childP = ly->w;
                /* no winid means we have never realized this widget or  *
                 * it is not a real window (ie. in the dialog editor).   *
                 * a winid that is less than zero indicates a non-Windows*
                 * widget and therefore needs to be re-painted here.     */
                if ( ((int)childP->winid <= 0) &&
                        ((w_dlg->flags & WFLG_IN_EDIT) && 
                            (childP->painter != NULL)) )
                    {
                    childP->display = TRUE;
                    (*childP->painter) (childP, 0, 0, TRUE);
                    }
                else
                    {
                    RealizeDialogWidget(w_dlg, childP);
                    }
                }
            }
        else
            {
            /* expose in widgets-chain order */
            for ( childP = w_dlg->widgets; childP; childP = childP->nextChild )
                {
                if ( ((int)childP->winid <= 0) &&
                        ((w_dlg->flags & WFLG_IN_EDIT) && 
                            (childP->painter != NULL)) )
                    {
                    childP->display = TRUE;
                    (*childP->painter) (childP, 0, 0, TRUE);
                    }
                else
                    {
                    RealizeDialogWidget(w_dlg, childP);
                    }
                }
            }
        }

    /* get active entry field if one exists and force focus on it. */
    //SetActiveEntryWidget(w_dlg);

    REMOVE_DRAWABLE(w_dlg->window);
    w_dlg->window = save_hdc;
    return;
}



/*************************************************************************
function: FindDlgPtrFromHWND - Return axw_dialog ptr
*/
AXWidget_dialog
FindDlgPtrFromHWND(HWND hwnd)
{
    register AXWidget mw;

    /* Find dlg */
    for ( mw = WindowList; mw; mw = mw->nextWin )
        {
        if ( mw->winid == hwnd )
            return ( (AXWidget_dialog)mw );
        }
    return( (AXWidget_dialog)NULL );
}



/*********************************************************
function: RealizeDialog - create dialog window
*/
void
RealizeDialog(AXWidget_dialog w_dlg)
{
    int ix;


    WIDGET_PARANOID(w_dlg, W_DIALOG);
	if ( !w_dlg->display )
		return;

    if ( w_dlg->winid )
        MS_DestroyWindow(w_dlg->winid);

    w_dlg->winid = MS_CreateWindow(NXLT("DialogClass"), w_dlg->name,
                                    WS_CHILD | WS_VISIBLE,
                                    w_dlg->xpos, w_dlg->ypos,
                                    w_dlg->width, w_dlg->height,
                                    w_dlg->parent->winid,
                                    (HMENU)NULL, (HANDLE)NULL, (LPVOID)NULL);

    if ( w_dlg->winid )
        {
        w_dlg->exposed = TRUE;
        w_dlg->window = AxPackageHandle(w_dlg->winid, AX_HWND_TYPE);
		/* see comments in winwidgt.c for
                                        * why we do this.   -donato
                                        */
        MS_SetClassLong(w_dlg->winid, GCL_HCURSOR, (LONG)ThimPointerCursor);
        }
    else
        taskAbort(NXLT("RealizeDialog: could not create window!"));

    /*** subclass window if need to, i.e. for Dialog Editor ***/
    for ( ix = 0; ix < ehList.numProcs; ix++ )
        {
        InsertEventHandler((AXWidget)w_dlg, ehList.eventMasks[ix], FALSE,
                              ehList.eventProcNames[ix], (char*)NULL, 0);
        }
}


/* Dialog box key bindings */
static menu_item_def *AxDialogKeyBindings = NULL;


/**************************************************************
function: DialogKeyEvent  - processes key events in a dialog.
             this ONLY processes accelerators, because the
             dialog widgets that allow key events are all
             Windows controls, which means Windows will
             handle all key press events.
	The above is not true anymore. We now also process the key events
	meant for widgets that don't have windows (e.g. W_RECT, W_SCALE, etc)
	See xmWidgetKeyPressHandler().
***/
int
DialogKeyEvent(AXWidget_main mw, HWND hwnd, UINT message,
                            WPARAM wParam, LPARAM lParam)
{
    BOOL bEvent;
    XEvent xevent;
    char *name;
    int is_text, bell;
    menu_item_def md, *mP = NULL; /* the elf menu bar data */
    AXWidget_dialog dialog_w;
    KeySym ksym;
    int num_chars;
    wchar *buf;
	AXWidget widget;
	AXWidget_togglebutton w_toggle;

    if ( AxDialogKeyBindings == NULL )
       AxDialogKeyBindings = ElfLookupSlider(0);

    WIDGET_PARANOID(mw, W_MAIN);
    dialog_w = mw->dialog;

    bEvent = WinToXEvent(hwnd,message, wParam,lParam, &xevent, (AXWidget)NULL);
    if ( bEvent )
        {
		/* give the key-focus widget a chance to consume the keystroke */
        num_chars = AXLookupString16(&xevent, &buf, &ksym);
		widget = xmKeyFocusWidget(dialog_w);
		if ( widget )
			{
			if ( widget->type == W_TOGGLE )
				{
				w_toggle = (AXWidget_togglebutton)widget;
				if ( w_toggle->in_radio &&
						RadioKeyPressEvent(
							(AXWidget_togglebutton)w_toggle->in_radio, &xevent,
														ksym, buf, num_chars) )
					{
					return TRUE;
					}
				}
			else if ( xmWidgetKeyPressHandler(widget, &xevent, ksym, buf,
											num_chars, FALSE) )
				return TRUE;
			}
        /*
         * If this dialog has a menu bar, see if key event is an
         * accelerator in the menu bar
         */
        if ( mw->dialog->flags & WFLG_IN_EDIT )
            {
            /* use de's menubar for key accels */
            if( deMenuBarKey((AXWidget)mw->dialog, &xevent) )
                return TRUE;
            }
        else
            {
            if ( mw->menuBar )              /* fetch menu bar */
                mP = mw->menuBar->popups;
            if ( ProcessKeyEvent(mP,&xevent,&is_text,&name,&bell) && !is_text )
                {
                if ( !AxIsCommandAvailable((AXWidget)mw, name) )
                    return FALSE;
                if ( mw->menuBar->callback )
                    {
                    md.word2 = name;    /* md for ElfXCmenuBarSelected() */
                    AxCallback(mw->menuBar->callback, (AXWidget)mw->menuBar,
                                    mw->menuBar->callbacktag, (caddr_t)&md);
					return TRUE;
                    }
                //return TRUE;
                }

             /* See if key event is an accelerator in dialog key bindings */
			mP = AxDialogKeyBindings;   /* dialog key bindings */
            if ( ProcessKeyEvent(mP,&xevent,&is_text,&name,&bell) && !is_text )
            	{
				/* if it's a dialog execute or cancel, consume the event. */
                if ( CallDlgFuncFromName(dialog_w, name) )
					return TRUE;
                }
             }
        }

    return FALSE;
}



/*************************************************************************
function: CallBackToWidget - Respond to widget action
If the widget has a callback tied to it (usually does!), this
will call back to the elf layer.
This is called when there is a WM_COMMAND sent to the widget.

return values:
	FALSE:  either we're ignoring the notification or we're not actually
			processing a callback (e.g. just setting focus).
	TRUE: 	we did the callback.
*/
static int
DialogWidgetCallback(WPARAM wParam)
/** WPARAM wParam;      widget in low word, subcommand in high word **/
{
	int retval;
    AXWidget w;
    AXWidget_entry w_enter;
	AXWidget_option w_option;
    AXWidget_togglebutton w_toggle;
    int widgetId;
    int widgetNotification;
    int newValue;
    int sel_changed, cur_sel;
    XmRowColumnCallbackStruct data;    /* for W_OPTION and W_RADIO */
    char *bufP;
    int num_bytes, buf_len;

    widgetId = LOWORD(wParam);
    widgetNotification = HIWORD(wParam);
    w = WIDGET_FROM_UID(widgetId);

	retval = TRUE;
    switch (w->type)
        {
        case W_OPTION:
            if ( widgetNotification == CBN_EDITCHANGE && w->winid &&
                                            !w->update_widget )
                {     /* user has typed in editable portion of combobox. */
				w_option = (AXWidget_option)w;
                /* make sure entry field has actually changed. */
                buf_len = MS_GetWindowTextLength(w_option->winid);
                bufP = (char*)malloc((buf_len+1)*sizeof(char));
                num_bytes = MS_GetWindowText(w_option->winid, bufP, buf_len+1);
                bufP[num_bytes] = 0;

				cur_sel = MS_SendMessage(w_option->winid, CB_GETCURSEL, 0, 0);
                if ( w_option->strs[cur_sel] == NULL && num_bytes == 0 )
                    /* nothing equals nothing */
                    break;

                if ( (w_option->strs[cur_sel] != (mbchar *)NULL) &&
                            !strcmp(w_option->strs[cur_sel], bufP) )
                    /* what windows has and what we have are still the same */
                    break;

                /* something really did change, so tell the elf layer */
				if ( w_option->flags & WFLG_EDITABLE )
				    {
#if BUILD_WIDE
					mbstr = (mbchar *)malloc((buf_len+1)*sizeof(wchar));
					data.data = (char *)mbstr;
					data.reason = TRUE;	/* hack, means string data & len */
					data.callbackstruct = (char*)WnToMbStr(bufP,mbstr,buf_len);
#else
					data.data = (char *)bufP;
					data.reason = TRUE;	/* hack, means string data & len */
					data.callbackstruct = (char *)buf_len;
#endif
                    if ( w_option->callback )
                        {
                        AxCallback(w_option->callback, (AXWidget)w_option,
                                   w_option->callbacktag, (caddr_t)&data);
                        }
					}
                free(bufP);
                }
            else if ( widgetNotification == CBN_SELCHANGE )
                {
                /* user has selected an item, do callback. */
                newValue = MS_SendMessage(w->winid, CB_GETCURSEL, 0, 0L);
                if ( newValue == LB_ERR )
                    break;

				data.reason = FALSE;	/* hack, means integer data */
                data.data = (char*)w->value = (char*)newValue;
                if ( w->callback )
                    AxCallback(w->callback, w, w->callbacktag, (caddr_t)&data);
                }
            else if ( widgetNotification == CBN_SETFOCUS )
                {
				xmTakeKeyFocus(w, (XButtonEvent*)NULL, FALSE);
                if ( THIMhelpMode )
                    {
                    AxPostHelp(w);
                    return retval;
                    }
                }
            break;


        case W_LIST:
            {
            int ignore_notification;
            XmListCallbackStruct list_data;

            if ( THIMhelpMode && widgetNotification == LBN_SETFOCUS )
                {
                AxPostHelp(w);
                return retval;
                }

            memset((char*) &list_data, '\0', sizeof(XmListCallbackStruct));

            if ( w->multi_pick )
                /* must send LB_GETCARETINDEX to multiple selection boxes */
                newValue = MS_SendMessage(w->winid, LB_GETCARETINDEX, 0, 0L);
            else
                newValue = MS_SendMessage(w->winid, LB_GETCURSEL, 0, 0L);            

            newValue++;    /* zero to one relative */
            if ( newValue == LB_ERR )
                break;

            ignore_notification = FALSE;
            list_data.item_position = w->value = newValue;

            switch (widgetNotification)
                {
                case LBN_SELCHANGE:
                    if ( w->multi_pick )
                        DoMultiPick((AXWidget_list)w, &list_data);
                    else
                        DoSinglePick((AXWidget_list)w);

                    if ( w->multi_pick && list_data.selected_item_count > 1 )
                        list_data.reason = XmCR_EXTENDED_SELECT;
                    else if ( w->multi_pick && list_data.selected_item_count == 0 )
                        list_data.reason = AXmCR_MULTIPLE_SELECT;
                    else
                        list_data.reason = AXmCR_SINGLE_SELECT;

                    /* save index of the 1st visible item in the listbox. */
                    w->scroll_pos = MS_SendMessage(w->winid, LB_GETTOPINDEX,
                                                        (WPARAM)0, (LPARAM)0);
                    break;
                case LBN_DBLCLK:
                    if ( w->multi_pick )
                        {
                        DoMultiPick((AXWidget_list)w, &list_data);
                        if ( list_data.selected_item_count != 1 )
                            ignore_notification = TRUE;
                        }

                    list_data.reason = AXmCR_DEFAULT_ACTION;
                    break;
				case LBN_SETFOCUS:
					xmTakeKeyFocus(w, (XButtonEvent*)NULL, FALSE);
                    ignore_notification = TRUE;
					break;
                default:
                    ignore_notification = TRUE;
                    break;
                }
            if ( !ignore_notification && w->callback )
                AxCallback(w->callback,w, w->callbacktag, (caddr_t)&list_data);
			else if ( ignore_notification )
				retval = FALSE;

            /* may have been allocated in DoMultiPick */
            if ( list_data.selected_items )
                free(list_data.selected_items);
            break;
            }


        case W_TOGGLE:
            {
            AXWidget_radio w_radio;
            XmToggleButtonCallbackStruct toggle_data;

			if ( widgetNotification == BN_SETFOCUS )
				{
				xmTakeKeyFocus(w, (XButtonEvent*)NULL, FALSE);
				retval = FALSE;
				break;
				}
			else if ( widgetNotification == BN_KILLFOCUS )
				{
				retval = FALSE;
				break;
				}

            if ( THIMhelpMode )
                {
                AxPostHelp(w);
                return retval;
                }

            w_toggle = (AXWidget_togglebutton)w;
            if ( w_toggle->in_radio )
                {
                /* toggle button is in a radio group. */
                w_radio = w_toggle->in_radio;
                data.widget = (Widget)w_radio;
                if ( SetRadioValue(w_toggle, TRUE) && w_radio->callback )
                    {
                    data.data = (char*)w_radio->value;
                    if ( w_radio->callback )
                        {
                        AxCallback(w_radio->callback, (AXWidget)w_radio,
                                    w_radio->callbacktag, (caddr_t)&data);
                        }
                    }
                }
            else
                {
                w_toggle->value = (short)MS_SendMessage(w_toggle->winid,
                                            BM_GETCHECK, (WPARAM)0, 0L);
                toggle_data.set = w_toggle->value;
                if ( w_toggle->callback )
                    {
                    AxCallback(w_toggle->callback, (AXWidget)w_toggle,
                            w_toggle->callbacktag, (caddr_t)&toggle_data);
                    }
                }

            break;
            }


        case W_BUTTON:
            {
            int save_value = w->value;
            XmAnyCallbackStruct data;

			if ( widgetNotification == BN_SETFOCUS )
				{
				xmTakeKeyFocus(w, (XButtonEvent*)NULL, FALSE);
				retval = FALSE;
				break;
				}
			else if ( widgetNotification == BN_KILLFOCUS )
				{
				retval = FALSE;
				break;
				}

            if ( THIMhelpMode )
                {
                AxPostHelp(w);
                return retval;
                }

            data.reason = EV_CHANGED_;
            /* button has been pushed so toggle its value. */
            w->value = !w->value;
            if ( w->callback )
                AxCallback(w->callback, w, w->callbacktag, (caddr_t)&data);
            w->value = save_value;
            break;
            }


        case W_ENTER:
            {

            w_enter = (AXWidget_entry)w;
            if ( widgetNotification == EN_UPDATE && w_enter->winid &&
                                            !w_enter->update_widget )
                {
                /* make sure entry field has actually changed. */
                buf_len = MS_GetWindowTextLength(w_enter->winid);
                bufP = (char*)malloc((buf_len+1)*sizeof(char));
                num_bytes = MS_GetWindowText(w_enter->winid, bufP, buf_len+1);
                bufP[num_bytes] = 0;

                if ( (w_enter->wstr == (mbchar *)NULL) && (num_bytes == 0) )
                    /* nothing equals nothing */
                    break;

                if ( (w_enter->wstr != (mbchar *)NULL) &&
                            !strcmp(w_enter->wstr, bufP) )
                    /* what windows has and what we have are still the same */
                    break;

                /* something really did change, so tell the elf layer */
                if ( w_enter->callback  &&
                    (w_enter->flags & WFLG_TOUCHED_RETURN) )
                    {
                    AxCallback(w_enter->callback, (AXWidget)w_enter,
                                w_enter->callbacktag, (caddr_t)NULL);
                    }
                free(bufP);
                w_enter->curPos = num_bytes;
#if 0
                /* set "caret" (cursor) in entry field */
                if ( w_enter->curPos > 0 )
                    {
                    MS_SendMessage(w_enter->winid, EM_SETSEL, w_enter->curPos,
                                                            w_enter->curPos);
                    }
#endif
                }
            else if ( widgetNotification == EN_SETFOCUS )
                {
                if ( THIMhelpMode )
                    {
                    AxPostHelp(w);
                    return retval;
                    }
                /* entry widget has gained focus. update the cursor pos *
                 * (-1 signifies lack of focus).                        */
                if ( w_enter->wstr[0] )
                    {
                    /* this should set curpos to wherever the cursor is. */
                    w_enter->curPos = (short)MS_SendMessage(w_enter->winid,
                                                        EM_GETSEL, 0, 0L);
                    }
                else
                    w_enter->curPos = 0;

                sel_changed = ResetActiveTextsExceptFor((AXWidget)w_enter);
                if ( sel_changed || w_enter->flags & WFLG_TOUCHED_RETURN )
                    {
                    CallWidget(w_enter->taskOwner, (AXWidget)w_enter,
                                               EV_FOCUS_IN_, NULL);
                    }
				xmTakeKeyFocus(w, (XButtonEvent*)NULL, FALSE);
                }
            else if ( widgetNotification == EN_KILLFOCUS )
                {
                /* first get any data out of the widget */
                WXGetWidgetString(w_enter);
                if ( w_enter->flags & WFLG_TOUCHED_RETURN )
                    {
                    CallWidget(w_enter->taskOwner, (AXWidget)w_enter,
                                                EV_FOCUS_OUT_, NULL);
                    }
                }
            break;
            }


        case W_EDITBOX:
            if ( widgetNotification == EN_UPDATE )
                {
                /* do we need to make the same check as above? */
                /* something really did change, so tell the elf layer */
                if ( w->callback && (w->flags & WFLG_TOUCHED_RETURN) )
                    AxCallback(w->callback, w, w->callbacktag, (caddr_t)NULL);
                }
            else if ( widgetNotification == EN_SETFOCUS )
                {
                if ( THIMhelpMode )
                    {
                    AxPostHelp(w);
                    return retval;
                    }
                /* editbox has gained focus. make sure other *
                 * text widgets are "de-selected".           */
                sel_changed = ResetActiveTextsExceptFor(w);
                if ( sel_changed || w->flags & WFLG_TOUCHED_RETURN )
                    CallWidget(w->taskOwner, w, EV_FOCUS_IN_, NULL);
				xmTakeKeyFocus(w, (XButtonEvent*)NULL, FALSE);
                }       
            else if ( widgetNotification == EN_KILLFOCUS )
                {
                /* first get any data out of the widget */
                AxWinGetEditBoxData((AXWidget_editbox)w);
                if ( w->flags & WFLG_TOUCHED_RETURN )
                    CallWidget(w->taskOwner, w, EV_FOCUS_OUT_, NULL);
                }
            break;


        default:
            DebugMessage(NXLT("DialogWidgetCallback: bad widget type!"));
			retval = FALSE;
        }

	return retval;
}

/*** wrapper; called from PaletteClassProc in winevent.c ***/
int
PaletteWidgetCallback(WPARAM wParam)
{
    return DialogWidgetCallback(wParam);
}

/*** wrapper; called from WorkClassProc and MEDebuggerClassProc ***/
int
WorkWidgetCallback(WPARAM wParam)
{
    return DialogWidgetCallback(wParam);
}
/**************************************************************/


static void
RealizeDialogWidget(AXWidget_dialog w_dlg, AXWidget w)
{
    long wstyle = WS_VISIBLE | WS_CHILD;
    static int entry_width;


    WIDGET_PARANOID(w_dlg, W_DIALOG);
    if ( w_dlg->parent->flags & WFLG_IN_EDIT )
        /* don't create windows if in Dialog Editor */
        return;

    /* make sure widget needs to be displayed */
    if ( !w->display || (w->flags & WFLG_NODISPLAY) )
        return;

    switch (w->type)
        {
        case W_BUTTON:
            if ( (int)w->winid <= 0 )
                {
                RealizePushButton((AXWidget_pushbutton)w, FALSE);
                }
            break;


        case W_ICON:
        case W_RECT:
        case W_SCALE:
        case W_SPLITTER:
            if ( !w->parent->window )
                w->parent->window = w_dlg->window;
            (*w->painter)(w, 0, TRUE, TRUE);
            w->winid = (HWND)-1;    /* just to have a non-NULL value */
            break;


        case W_LABEL:
        case W_ROWCOL:
        case W_CHOOSE:
            if ( !w->parent->window )
                w->parent->window = w_dlg->window;
            (*w->painter)(w, NULL, TRUE, TRUE);
            break;


        case W_LIST:
        case W_ENTER:
        case W_RADIO:
        case W_OPTION:
        case W_TOGGLE:
        case W_EDITBOX:
            /* paint if needs updating (also, we must always paint
             * the radio widget's label).
             */
            if ( w->update_widget || w->type == W_RADIO )
                (*w->painter)(w, 0, TRUE, TRUE);
            /* if entry field has a label, we _must_ paint it! */
            if ( (w->type == W_ENTER) && (w->next->type == W_LABEL) )
                (*w->next->painter)(w->next, 0, TRUE, TRUE);
            break;


        case W_TOOLBAR:
            /* these are created as part of WM_PAINT of the main shell. this *
             * case is here just so this function can be called as part of   *
             * WM_PAINT of a widget, whether or not it has a toolbar.        */
            break;


        /* Tables, Trays and Canvases have borders *outside* their windows,
         * so make sure we draw those borders on a dialog expose.
         * The individual class procs will handle full widget exposes.
         */
        case W_TABLE:
            if ( !w->winid )
                /* widget hasn't been created yet; make sure it does. */
                (*w->painter)(w, NULL, FALSE, TRUE);
            else
                PaintTableBorder((AXWidget_table)w);
            break;

        case W_TRAY:
            if ( !w->winid )
                /* widget hasn't been created yet; make sure it does. */
                (*w->painter)(w, NULL, FALSE, TRUE);
            else
                DrawTrayBorder((AXWidget_tray)w);
            break;

        case W_CANVAS:
            {
            int thickness;
            /* if canvas hasn't been created, do so. otherwise
             * just paint the border (see comment above).
             */
            if ( !w->winid )
                (*w->painter)(w, NULL, TRUE, TRUE);
            else
                {
                if ( w->thickness < 0 )
                    {
                    thickness = -w->thickness;
                    AxDrawBevel(w, WIN_OF(w), w->xpos, w->ypos,
                                w->width, w->height,
                                thickness, BEVEL_IN);
                    }
                else if ( w->thickness )
                    {
                    thickness = w->thickness;
                    AxDrawBevel(w, WIN_OF(w), w->xpos, w->ypos,
                                w->width, w->height,
                                thickness, BEVEL_OUT);
                    }
                }
            }
            break;

        default:
            taskAbort("RealizeDialogWidget: bad widget type");
        }

    w->mapped = TRUE;
    w->update_widget = FALSE;
}


/***********************************************/
void
DeleteDlgBrush(void)
{
    if ( DlgWidgetBrush )
        {
        MS_DeleteObject(DlgWidgetBrush);
        DlgWidgetBrush = NULL;
        }
    if ( DlgEditBrush )
        {
        MS_DeleteObject(DlgEditBrush);
        DlgEditBrush = NULL;
        }
}



static int
CallDlgFuncFromName(AXWidget_dialog w, char *name)
{
	int retval;
    AXWidget_pushbutton button_w;
    XmAnyCallbackStruct data;
 
    WIDGET_PARANOID(w, W_DIALOG);

	retval = TRUE;
	button_w = NULL;
    if ( streq(name, NXLT("DLG.CANCEL")) )
        button_w = w->cancelButton;
    else if ( streq(name, NXLT("DLG.EXECUTE")) )
        button_w = w->currentDefaultButton;
    else
        retval = FALSE;

    if ( button_w )
    {
        data.reason = EV_CHANGED_;
        if ( button_w->callback )
        {
            AxCallback(button_w->callback, (AXWidget)button_w,
                      button_w->callbacktag, (caddr_t)&data);
        }
    }

    return retval;
}


static BOOL
WidgetIsScrollable(AXWidget w)
{
    if ( w->type == W_CANVAS )
        return FALSE;
    else
        return TRUE;
}

static void
DialogScrollCallback(AXWidget_dialog w_dlg, WPARAM wParam, LPARAM lParam)
{
    int scroll_code, win_scroll_pos, app_scroll_pos;
    AXWidget_scroll w_scroll;
    int min, max, user_space;
    XmScrollBarCallbackStruct sbData;


    WIDGET_PARANOID(w_dlg, W_DIALOG);
    min = max = -1;
    w_scroll =(AXWidget_scroll)GetGadgetFromHWND((AXWidget)w_dlg,(HWND)lParam);
    MS_GetScrollRange(w_scroll->winid, SB_CTL, &min, &max);

    scroll_code = LOWORD(wParam);
    win_scroll_pos  = HIWORD(wParam);

    /* Since the Windows scroll position report is only 16 bits we may */
    /* need to scale the reported measurements back up to the user range */
    if ( w_scroll->scroll_max > 32767 )
       {
       /* scale up to max position */
       app_scroll_pos = (float)win_scroll_pos * 
                             (((float)w_scroll->scroll_max) / 32767.0);
       }
    else
       app_scroll_pos = win_scroll_pos;

    user_space = w_scroll->scroll_max - w_scroll->scroll_per;
    if ( user_space < 0 || w_scroll->scroll_max == 0 )
       return;
    if ( !RunningWin95Look )
       {
       /* Since we can not set the size of the thumb (yet, until Chicago) */
       /* we need to emulate that here.  Basically, the position of the   */
       /* thumb is always the left/top edge.  We keep the value returned  */
       /* to the application within the range that the user could have    */
       /* created if he had a "fat" thumb which could not be moved much   */
       /* by mapping the user movable space (the area not occupied by the */
       /* fat scroll bar) to the Windows space, which is the entire bar   */
       app_scroll_pos = ((float)user_space * (float)app_scroll_pos) / 
                                             (float)w_scroll->scroll_max;
       }

    memset(&sbData, 0, sizeof(sbData));
    switch (scroll_code)
        {
        default:
            sbData.reason = XmCR_NONE;
            break;
        case SB_TOP:
            sbData.reason = XmCR_TO_TOP;
            sbData.value = 0;
            break;
        case SB_BOTTOM:
            sbData.reason = XmCR_TO_BOTTOM;
            sbData.value = w_scroll->scroll_max;
            break;
        case SB_PAGEDOWN:
            sbData.reason = XmCR_PAGE_INCREMENT;
            break;
        case SB_LINEDOWN:
            sbData.reason = XmCR_INCREMENT;
            break;
        case SB_PAGEUP:
            sbData.reason = XmCR_PAGE_DECREMENT;
            break;
        case SB_LINEUP:
            sbData.reason = XmCR_DECREMENT;
            break;
        case SB_THUMBPOSITION:
        case SB_THUMBTRACK:
            sbData.reason = XmCR_VALUE_CHANGED;
            sbData.value = app_scroll_pos;
            /* tell windows to move the thumb.  */
            if ( user_space < 5 )
               /* thumb is same size as scroll bar, user can not move */
               win_scroll_pos = 0;

            MS_SetScrollPos(w_scroll->winid, SB_CTL, win_scroll_pos, TRUE);
            break;
        }

	/* Why we don't dynamically scroll the canvas: Windows directly calls
	 * our window proc with scroll messages, which bypasses the Applixware
	 * schedular, which means we need to handle the scroll event with a 
	 * direct call to the application (rather than adding a message to the
	 * Applixware queue).  The problem arises in Builder, where someone may
	 * have set a callback on the scroll event which activates a dialog box.
	 * We are then in a situation where task zero (the "kernel") is trying
	 * to start a new task, which is of course forbidden.
	 */ 
    if ( scroll_code == SB_THUMBTRACK &&
	     !WidgetIsScrollable((AXWidget)w_scroll->callbacktag) )
        ;    /* don't "active scroll" the canvas. */
    else
        /* do the dialog widget scroll callback. */
        AxCallback(w_scroll->callback,(AXWidget)w_scroll,w_scroll->callbacktag,
                                                        (caddr_t)&sbData);
}

