/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winfont.c
   Applix ASC version: 76
   Date: Mon Nov 16 13:53:37 1998
*/
#include <stdlib.h>
#include <basicwin.h>
#include <aos.h>
#include <thimble.h>
#include <axfont.h>
#include <wp.h>


extern XFontStruct *AxDlgMainFont; /* main dialog font */
extern XFontStruct *AxDlgMiniFont; /* small version */
extern XFontStruct *Grafont;
XFontStruct *Txfont, *olCursorfont, *BoldTxfont;
XFontStruct *MiniTxfont, *MiniLabelTxfont, *MiniButtonFont;
XFontStruct *AxStatusLineFont;
XFontStruct *AxDefaultReferenceFont;     /* widths only, based on axsab-16 */
int AxAlternateDialogFont = TRUE;  /* bool: FALSE when X11 axsab-16 used */
data_ptr AxDialogFontName;   /* <face><pointsize><weight><slant> (or null) */

#if FOR_FILTER      /*  {  */
//#define TrueRealloc( ptr, size ) realloc( ptr, size )
//#define TrueMalloc( size )   malloc( size )
//#define TrueFree( ptr )      free( ptr )

XFontStruct *AxDlgMainFont; /* main dialog font */
XFontStruct *AxDlgMiniFont; /* small version */
int AxCharAbove, AxCharBelow, AxCharHeight;
int SmallCharAbove, SmallCharBelow, SmallCharHeight;
XFontStruct *MonospaceFont, *MonospaceBoldFont;
XFontStruct *MonospaceItalFont, *MonospaceItalBoldFont;

#else               /* }  {  */

extern XFontStruct *AxDlgMainFont; /* main dialog font */
extern XFontStruct *AxDlgMiniFont; /* small version */
extern int AxCharAbove, AxCharBelow, AxCharHeight;
extern int SmallCharAbove, SmallCharBelow, SmallCharHeight;
extern XFontStruct *MonospaceFont, *MonospaceBoldFont;
extern XFontStruct *MonospaceItalFont, *MonospaceItalBoldFont;

extern int RunningWin95Look;
#endif               /*   }  */

int AxBigFont = FALSE;   /* If true, use larger font set */
                         /* never true in Windows/NT */

int CharWids[256];      /* of dialog font */
int SmallCharWids[256]; /* of reduced dialog font */

extern int ScreenRes;

#include <winfont.p>



HDC InitScreenHDC()
{
    static HDC screen_hdc = (HDC)NULL;

    if ( screen_hdc == (HDC)NULL )
       {
       /* create the hdc we will keep around and use for measuring text.  */
       /* Timing test have shown that things are much faster if we create */
       /* and then re-use a hdc, rather than creating it anew every time  */
       /* we want to measure text                                         */
       screen_hdc = MS_CreateDC( NXLT("DISPLAY"), NULL, NULL, NULL );
       if ( screen_hdc == (HDC)NULL )
           {
           tempAbort("Could not create screen_hdc");
           }

       }
    return( screen_hdc );;
}


#if 0    /* don't think we need this. */
HDC InitScratchHDC()
{
    static HDC scratch_hdc = (HDC)NULL;

    if ( scratch_hdc == (HDC)NULL )
       {
       /* create the hdc we will keep around and use for measuring text.  */
       /* Timing test have shown that things are much faster if we create */
       /* and then re-use a hdc, rather than creating it anew every time  */
       /* we want to measure text                                         */
       scratch_hdc = MS_CreateDC( NXLT("DISPLAY"), NULL, NULL, NULL );
       if ( scratch_hdc == (HDC)NULL )
           {
           tempAbort("Could not create scratch_hdc");
           }

       }
    return( scratch_hdc );
}
#endif

HDC InitPrinterHDC()
{
    char *device;
    char *driver;
    char *output_port;
    char printer_string[200];
    char empty_string[2];
    static HDC printer_hdc = (HDC)NULL;
    
    if ( printer_hdc == (HDC)NULL )
       {
       /* create the hdc we will keep around and use for measuring text.  */
       /* Timing test have shown that things are much faster if we create */
       /* and then re-use a hdc, rather than creating it anew every time  */
       /* we want to measure text                                         */
       printer_string[0] = '\0';
       MS_GetProfileString (NXLT("windows"), NXLT("device"), "", printer_string, 200 );

       printer_hdc = (HDC)NULL;
       if ((printer_string[0] != '\0') &&
           (NULL != (device = strtok (printer_string, "," ))) &&
           (NULL != (driver = strtok (NULL,      ", "))) &&
           (NULL != (output_port = strtok (NULL,      ", "))))
           {
           empty_string[0] = 0;
           printer_hdc = CreateDC( driver, device, 
                                          empty_string, (DEVMODE *)NULL );
           }
       if ( printer_hdc == (HDC)NULL )
           {
#if 0
           DebugMessage("Could not create printer_hdc");
#endif
           return( (HDC)NULL );
           }

       MS_SetMapMode( printer_hdc, MM_HIENGLISH );   /* results in mills */
       }
    return( printer_hdc );
}


HFONT NormalDialogFont, SmallDialogFont, StatusDialogFont;

void
InitFontInfo(void)
{
    int ptsize, junk;
    char *msg_path = NULL;
    int lineNumber;
    LOGFONT font_info;
    HDC hdc;
    HFONT old_font;
    TEXTMETRIC tm;
    int large_fonts_mode = FALSE;

    /* read in the font set config file "fontmap.dir" */
    if(!axtInitFonts(&lineNumber))
        {
        printf(XLT("Problem with fontmap.dir file around line %d\n"),
                lineNumber);
        }

#if !FOR_FILTER    /*  {  */
    AxDlgMainFont    = (XFontStruct *)calloc( 1, sizeof(XFontStruct) );
    AxDlgMiniFont    = (XFontStruct *)calloc( 1, sizeof(XFontStruct) );
    AxStatusLineFont = (XFontStruct *)calloc( 1, sizeof(XFontStruct) );

    AxDefaultReferenceFont = initDefaultReferenceFont();

    /* We start by getting the metrics for the dialog box font, which */
    /*   is the Microsoft default                                     */
    memset( &font_info, 0, sizeof(LOGFONT) );
    strcpy( font_info.lfFaceName, NXLT("MS Sans Serif") );
    font_info.lfHeight = -((8 * 75)/72);  /* "8" point */
    if ( !RunningWin95Look )
        font_info.lfWeight = FW_BOLD;
    NormalDialogFont = MS_CreateFontIndirect( &font_info );
    /* somewhat of a hack here: we check to see the height of the font we just
     * created. If it is more than 13, we assume the user is running in
     * "Large Fonts" mode and use a non-bold font instead (We do this on 
     * NT 3.51 only because other platforms already use a non-bold font).
     */
    if ( !RunningWin95Look )
        {
        hdc = MS_GetDC(NULL);
        old_font = MS_SelectObject(hdc, NormalDialogFont);
        MS_GetTextMetrics(hdc, &tm);
        MS_SelectObject(hdc, old_font);
        if ( tm.tmHeight > 13 )
            {
            MS_DeleteObject(NormalDialogFont);
            font_info.lfWeight = FW_NORMAL;
            NormalDialogFont = MS_CreateFontIndirect( &font_info );
            large_fonts_mode = TRUE;
            }
        MS_ReleaseDC(NULL, hdc);
        }

    if ( !WinRasterFontToXFontMetrics(NormalDialogFont, AxDlgMainFont, TRUE ))
        {
        /* could not get dialog font info, this is serious */
        WXErrorMessage( XLT("Could not open dialog font") );
        exit( 1 );
        }

    memset( &font_info, 0, sizeof(LOGFONT) );
    strcpy( font_info.lfFaceName, NXLT("MS Sans Serif") );
    font_info.lfHeight = -((5 * 75)/72);  /* "5" point */
    if ( !RunningWin95Look && !large_fonts_mode )
        font_info.lfWeight = FW_BOLD;

    SmallDialogFont = MS_CreateFontIndirect(&font_info);
    if ( SmallDialogFont == NULL )
       /* could not get it, use the 'normal' one */
       SmallDialogFont = NormalDialogFont;

    if ( !WinRasterFontToXFontMetrics(SmallDialogFont, AxDlgMiniFont, TRUE) )
        {
        /* could not get dialog font info, this is serious */
        WXErrorMessage( XLT("Could not open small dialog font") );
        exit( 1 );
        }

    memset( &font_info, 0, sizeof(LOGFONT) );
    strcpy( font_info.lfFaceName, "MS Sans Serif" );
    font_info.lfHeight = -((5 * 75)/72);  /* "5" point */
    if ( !RunningWin95Look && !large_fonts_mode )
        font_info.lfWeight = FW_BOLD;

    StatusDialogFont = MS_CreateFontIndirect(&font_info);
    if ( StatusDialogFont == NULL )
       /* could not get it, use the 'normal' one */
       StatusDialogFont = NormalDialogFont;
    if ( !WinRasterFontToXFontMetrics(StatusDialogFont, AxStatusLineFont, TRUE) )
        {
        /* could not get dialog font info, this is serious */
        WXErrorMessage( XLT("Could not open small dialog font") );
        exit( 1 );
        }

    if ( AxAlternateDialogFont && ((AxDlgMainFont->max_bounds.ascent +
                                  AxDlgMainFont->max_bounds.descent) > 16) )
        ptsize = 14;
    else
        ptsize = 12;

    getXFontStruct("Courier New", ptsize, 0, 0, &junk, &MonospaceFont);
    getXFontStruct("Courier New", ptsize, 1, 0, &junk, &MonospaceBoldFont);
    getXFontStruct("Courier New", ptsize, 0, 1, &junk, &MonospaceItalFont);
    getXFontStruct("Courier New", ptsize, 1, 1, &junk, &MonospaceItalBoldFont);

    memset(CharWids, 0, sizeof(CharWids));
    memset(SmallCharWids, 0, sizeof(SmallCharWids));
    initCharWids(AxDlgMainFont, 1, 1000, CharWids, &AxCharAbove,
                                    &AxCharBelow, &AxCharHeight);
    initCharWids(AxDlgMiniFont, 1, 1000, SmallCharWids, &SmallCharAbove,
                                    &SmallCharBelow, &SmallCharHeight);
#endif    /*    }  !FOR_FILTER  */

    return;
}


AxUseBigFont()
{
    return(AxBigFont);
}


/** The following is for the screen only, not the printer */
static int WinRasterFontToXFontMetrics(win_font, x_info, do_per_char)
HFONT win_font;
XFontStruct *x_info;
int do_per_char;
{
    TEXTMETRIC tm;
    HGDIOBJ old_font;
    XCharStruct *charP;
    int i, tmp;
    char str[2];
    SIZE size;
    HDC hDC;


    hDC = MS_GetDC(NULL);
    /* now select the font into the hDC so that we can get information */
    /*   about it, and can measure the characters */
    old_font = MS_SelectObject(hDC, win_font);

    MS_SetTextAlign(hDC, TA_LEFT | TA_BASELINE);

    if ( !MS_GetTextMetrics(hDC, &tm) )
        {
        MS_SelectObject(hDC, old_font);   /* restor the original */
        return (FALSE);
        }

    x_info->fid = win_font;
    x_info->min_byte1 = x_info->max_byte1 = 0;
    x_info->min_char_or_byte2 = tm.tmFirstChar;
    x_info->max_char_or_byte2 = tm.tmLastChar;
    x_info->all_chars_exist = TRUE;
    x_info->default_char = ' ';
    x_info->min_bounds.lbearing = 0;
    x_info->min_bounds.rbearing = (short)tm.tmMaxCharWidth;
    x_info->min_bounds.width = (short)tm.tmMaxCharWidth;
    x_info->min_bounds.ascent = (short)tm.tmAscent;
    x_info->min_bounds.descent = (short)tm.tmDescent;

    x_info->max_bounds = x_info->min_bounds;

    x_info->ascent = tm.tmAscent;
    x_info->descent = tm.tmDescent;

    if ( do_per_char )
       {
       /* now fill out the per-char structure */
       tmp = tm.tmLastChar - tm.tmFirstChar + 1;
       tmp *= sizeof(XCharStruct);
       charP = x_info->per_char = (XCharStruct*)malloc(tmp);
       memset(charP, 0, tmp);

       for ( i=tm.tmFirstChar; i <= tm.tmLastChar; i++ )
          {
          /* measure each character for width.  Windows does not return */
          /* any per-character ascent/descent info, just the overall height */
          /*   so we just use the  global font values.                     */
          str[0] = i;
          if ( MS_GetTextExtentPoint32( hDC, str, 1, &size ) )
             {
            charP->rbearing = charP->width = (short)size.cx;
            charP->ascent = (short)tm.tmAscent;
            charP->descent = (short)tm.tmDescent;
             }
          charP++;
          }
       }

    MS_ReleaseDC(NULL, hDC);
    return (TRUE);
}





/************************************************************************/
/* Herin follow a set of functions to find, and cache logical fonts     */
/* The intent is that when a task requests a font, we look in the       */
/*   cache to find that font, and if it exists, we note the fact that   */
/*   this task is using this font and will increment the usage count    */
/*   When the task goes away, we scan the cache list, and decrment the  */
/*   usage count for all fonts which were used by this task.  If the    */
/*   count goes to zero, we free the font resource and delete the       */
/*   element from the cache                                             */
/************************************************************************/



/* returns TRUE if OK,   FALSE otherwise */

int axGetWinScreenMetrics(familyName, weight, slant, pointsize, 
                          targetRec, metricsP )
char *familyName;       /* in: e.g., "Times" */
int weight, slant;      /* in */
int pointsize;          /* in: pointsize in points */
fontTabEnt *targetRec;  /* in: reference to complete font record */
fontMetrics *metricsP;  /* in: this structure should exist, the "xfont" */
                        /* sub-structure will be malloced if call is    */
                        /* successful                                   */
{

    LOGFONT logfont;
    HFONT hfont;
    int ret;
    char *screenFamily, *err;
    fontTabEnt *trec, *screenRecord;
    int twidth;

    memset( &logfont, 0, sizeof(logfont) );
    /* we want to read the screen metrics back in screen pixels, so we */
    /* express the height in screen resolution pixels.  'ScreenRes' was */
    /* loaded by the startup code to be the number of pixels per logical */
    /* inch on the display. (a logical inch may be larger or smaller     */
    /* than a physical inch.  see Petzold    */
#if 0    /* ? ? ScreenRes should work! */
    logfont.lfHeight = (pointsize * ScreenRes)/72;
#else 
    logfont.lfHeight = (pointsize * 75)/72;
#endif

#if 0
    /* we shouldn't need this block of code, because in the case where
     * we have a printer font we can let Windows give us a substitute
     * simply by passing it to the CreateFont call.    -donato
     */
#if !FOR_FILTER
    if ( targetRec && targetRec->isPrinterFont && !targetRec->isWindowsFont )
        {
        /* specify a lfWidth value based on the target font information. */
        /* Get the real value now by creating an instance of the font
         * in the "em" size.
         */
        GetWinPrinterAverageWidth(targetRec);
        twidth = (targetRec->avgWidth * 1000) / targetRec->sizeEM;
        
        /* convert to screenpix. should use ScreenRes, not 75.. */
        logfont.lfWidth = (LONG) mToPix(twidth, (float) pointsize, 75);
        }
#endif
#endif

    if ( targetRec && targetRec->WindowsAlternate && 
                        !targetRec->isWindowsFont &&
                        !targetRec->isPrinterFont)
        {
        /* an example of familyName in this context might be "Palatino".  */
        /* the alternate is provided for purposes of successfully finding */
        /* a screenfont to use; an example might be "Book Antigua"        */
        /* we'll create an instance of the alternate's screen font,       */
        /* but width-modified based on the target's average width         */

        screenFamily = targetRec->WindowsAlternate;
        if ( !IsWinFont(targetRec->WindowsAlternate) )
            screenFamily = NXLT("Arial");
        
        /* specify a lfWidth value based on the target and alternate */
        /*     font information. */
        twidth = axGetAlternateAverageWidth(targetRec);
        /* convert to screenpix. should use ScreenRes, not 75.. */
        
        logfont.lfWidth = (LONG) mToPix(twidth, (float) pointsize, 75);
        }
    else
        {
        screenFamily = familyName;
        }

#if 0
    /* see comment at rev 74  */
    if ( targetRec && targetRec->isPrinterFont && !targetRec->isWindowsFont )
        logfont.lfOutPrecision = OUT_DEVICE_PRECIS;
    else
#endif
    logfont.lfOutPrecision = OUT_TT_ONLY_PRECIS;  /* always get TrueType */

    /* make the size negative so that Windows gives us a font of this height */
    /*   not a font within a line spacing of this height           */
    logfont.lfHeight = -logfont.lfHeight;  

    if ( weight > 0 )
       logfont.lfWeight = FW_BOLD;   /* bold */
    else
       logfont.lfWeight = FW_NORMAL;   /* normal */

    if ( slant > 0 )
       logfont.lfItalic = 1; 

    /* lookup charset information based on actual screenfont */
    err = axGetFontRecordForScreen(screenFamily, weight, slant, pointsize,
                                   &trec, &screenRecord);
    if ( err )
        {
        freeErrString ( err );
        }
    else
        {
        logfont.lfCharSet = screenRecord->charSet;
        logfont.lfPitchAndFamily = screenRecord->pitchAndFamily;
        }

    strcpy( logfont.lfFaceName, screenFamily );
#if 1
    if ( strcmp(logfont.lfFaceName, NXLT("Courier") ) == 0 )
        {
        /* for some reason Windows doesn't map Courier the same way it maps
         * other printer fonts (e.g. "Times" becomes "Times New Roman", etc.)
         * If we set the true-type precision flag, however, we get a better
         * mapping ("Courier New").                -donato
         */
        logfont.lfOutPrecision = OUT_TT_ONLY_PRECIS;
        }
#else
    /* let Windows do the substitution.   -donato */
    if ( strcmp( logfont.lfFaceName, NXLT("Times") ) == 0 )
       {
       /* fail safe for all that X11 code which defaults to "Times"  */
       strcpy( logfont.lfFaceName, NXLT("Times New Roman") );
       }
    else if ( strcmp( logfont.lfFaceName, NXLT("Courier") ) == 0 )
       {
       /* fail safe for all that X11 code which defaults to "Courier"  */
       strcpy( logfont.lfFaceName, NXLT("Courier New") );
       }
#endif

    hfont = MS_CreateFontIndirect( &logfont );
    if ( hfont == (HFONT)NULL )
        {
        /*did not get the font, try with everything default except the size */
        memset( &logfont, 0, sizeof(logfont) );
        logfont.lfHeight = pointsize;
 
        hfont = MS_CreateFontIndirect( &logfont );
        if ( hfont == (HFONT)NULL )
            {
            /* still no luck, bail out */
            taskAbort( "Could not get requested font" );
            return( FALSE );
            }
        }

    /**............................................................*/
    /* get metrics 
    /**............................................................*/
    /* allocate the space to hold all the font metrics info */
    metricsP->xfont = (XFontStruct *)calloc(1, sizeof(XFontStruct));

    /* this adds the font handle (hfont) to the XFontStruct in metricsP,   *
     * so don't delete the object! (we hopefully delete it later). -donato */
    ret = WinTrueTypeToXFontMetrics(hfont, metricsP, FALSE);

    return( ret );
}

/* : axGetTargetAverageWidth - return the average width of the
 * target metrics for the passed fontTabEnt.  The return value
 * is in milliems
 */
int axGetTargetAverageWidth(targetRecord)
fontTabEnt *targetRecord;
{
    fontMetrics *fm;
    XCharStruct *xc;
    XFontStruct *fs;
    int avgWidth, sumWidth, ix, numChars, numActual;
    char *err;
    
    if (!targetRecord || !targetRecord->numTargetMetricsNames ||
        !targetRecord->targetMetricsNames[0])
        {
        return (500);         /* return .5 em by default */
        }
    
    err = axGetTargetMetrics(targetRecord->targetMetricsNames[0],
                             targetRecord, &fm);
    if (err)
        {
        freeErrString(err);
        return (500);         /* return .5 em by default */
        }

    fs = fm->xfont;
    
    if (targetRecord->fixedWidth || !fs->per_char)
        {
        avgWidth = fs->min_bounds.width;
        }
    else if (fs->max_byte1 > 0)
        {
        /* some sort of wide font; haven't a clue
         * how to determine average width.. but,
         * use max_width in the hopes of hitting the
         * width of one of the ideographic chars..
         */
        avgWidth = fs->max_bounds.width;
        }
    else
        {
        /* use a "scrabble" average.  This assumes knowledge
         * of encoding, which we don't really have..  should
         * work OK for iso646 based encodings like ascii,
         * iso8859 and unicode, though..
         */
        static char scrabble[32] = "aaeeiioouuAEIOUstndglmwSTNDGLMW";
        
        numChars = 31;   /* size of scrabble */

        for (ix = 0, sumWidth = 0, numActual = 0; ix < numChars; ix++)
            {
                xc = getCharStruct((int)((unsigned char) scrabble[ix]), fs);
                if (xc->width > 0)
                    {
                        sumWidth += xc->width;
                        numActual ++;
                    }
            }
        
        if (numActual)
            {
            avgWidth = sumWidth / numActual;
            }
        else
            {
            avgWidth = 500;
            }
        }
    
    return (avgWidth);
}


/************************************************************************/
/* End of a set of functions to find, and cache logical fonts     */
/************************************************************************/



/* Measure the widt of a string.  The font handle we are passed is assumed  */
/*   to ba a handle to a font which was created at the point size specified */
/*   in mills, so that we can read mill based dimensions directly from      */
/*   Windows.   See axGetWinScreenMetrics() for where the font handle       */
/*   is created                                                             */

int
MeasureWinText(wchar *wtext, int len, HFONT hFont)   /* returns length in mills */
{
    SIZE size;
    HDC screen_hdc;
    HFONT old_font;

    screen_hdc = MS_GetDC(NULL);

    old_font = MS_SelectObject( screen_hdc, hFont );

    MS_GetTextExtentPoint32( screen_hdc, wtext, len, &size );

    MS_SelectObject(screen_hdc, old_font);
    MS_ReleaseDC(NULL, screen_hdc);
    return( size.cx );
}


/* The two functions "axMakeCompoundWinName" and "ParseCompoundWinName" are */
/*  used to mash the attributes of a font into a single string so that it */
/*  can be passed around ala an xlfd in the font-cache code and elsewhere */
/*  NOTE: caller must free returned string                                */
    
char* axMakeCompoundWinName( family, point_size, weight, slant)
char *family;                    /* in */
int point_size, weight, slant;   /* in */
{
    int len;
    char *buffer;

    len = strlen(family);
    /* add 10 chars ( max ascii-decimal for INT_MAX) for each
     * of pointsize, weight, slant.  then add 4 for commas and null.
     */
    len += 34;
    buffer = (char *) malloc(len);
    
    sprintf( buffer, "%d,%d,%d,%s", point_size, weight, slant, family );

    return (buffer);    /* caller must free returned memory */
}

void ParseCompoundWinName( family, point_size, weight, slant, buffer )
char *family;
int *point_size, *weight, *slant;
char *buffer;
{
    char *p;

    /* we pick the buffer apart manually, rather than with sscanf since */
    /*   the font name may contain blanks, which causes sscanf to stop  */
    /*   prematurely.  The string should be of the form:                */
    /*      <point_size>,<weight>,<slant>,<face name>                   */
    /*      "12,0,1,Times New Roman"                                    */
    p = buffer;
    *point_size = atoi( p );

    while ( *p && (*p != ',') )
       p++;
    p++;                /* one past the comma */
    *weight = atoi( p );

    while ( *p && (*p != ',') )
       p++;
    p++;                /* one past the comma */
    *slant = atoi( p );

    while ( *p && (*p != ',') )
       p++;
    p++;                /* one past the comma */
    strcpy(family, p );

    return;
}

    

#if 0        /*  {  */

int ntGetFileLineCount(fp)
FILE *fp;
{
    int count = 0;
    char temp_buf[256];
    
    while ( !feof(fp) ) 
        {
        fgets( temp_buf, 255, fp );
        count++;
        }
        
    /* restore file pointer to beginning of file */
    fseek( fp, 0L, SEEK_SET );     
    return( count );
}

#endif               /*  }  */



/* widths of unix applixware default menu font axsab-16 */
static int axsabWidths[224] = {
     4,  4,  5,  7,  7, 12,  9,  3,  4,  4,  5,  8,  4,  4,  4,  4,
     7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  4,  4,  8,  8,  8,  7,
    14,  9,  9, 10, 10,  9,  8, 10, 10,  4,  7,  9,  7, 11, 10, 10, 
     9, 10, 10,  9,  8, 10, 10, 13,  9,  9,  8,  5,  4,  5,  6,  7,
     3,  7,  7,  7,  7,  7,  4,  7,  7,  3,  3,  7,  3, 11,  7,  7,
     7,  7,  5,  7,  4,  7,  8, 10,  7,  7,  7,  4,  3,  4,  8,  4,
     4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,
     4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,
     4,  4,  7,  7,  7,  9,  4,  7,  5, 12,  5,  7,  9,  9, 12,  5,
     6,  8,  5,  5,  5,  9,  7,  4,  5,  4,  5,  7, 12, 12, 12,  7,
     9,  9,  9,  9,  9,  9, 13, 10,  9,  9,  9,  9,  4,  4,  4,  4,
    11, 10, 10, 10, 10, 10, 10,  9, 10, 10, 10, 10, 10, 10, 10,  8,
     7,  7,  7,  7,  7,  7, 13,  7,  7,  7,  7,  7,  4,  4,  4,  4,
     9,  7,  7,  7,  7,  7,  7,  9,  8,  7,  7,  7,  7,  8,  9,  7,
};

/* : initDefaultReferenceFont - provide widths-only struct
 * describing axsab-16 (default menu font used w/ unix applixware)
 * This uses the axsabWidths array defined above.
 */
static XFontStruct *initDefaultReferenceFont()
{
    XFontStruct *fs;
    int ix;
    
    fs = (XFontStruct *) malloc(sizeof(XFontStruct));
    memset((char *) fs, 0, sizeof(XFontStruct));
    
    /* this font structure has little interesting on it
       except overall ascent and descent, min and max bounds,
       and per_char array which stores widths only */

    fs->min_char_or_byte2 = (unsigned int) 32;
    fs->max_char_or_byte2 = (unsigned int) 255;
    fs->default_char = (unsigned int) 20;
    fs->all_chars_exist = 1;
    
    fs->min_bounds.width = 3;
    fs->min_bounds.ascent = 13;
    fs->min_bounds.descent = 3;
    
    fs->max_bounds.width = 14;
    fs->max_bounds.ascent = 13;
    fs->max_bounds.descent = 3;

    fs->per_char = (XCharStruct *) malloc(224 * sizeof(XCharStruct));
    memset((char *) fs->per_char, 0, 224 * sizeof(XCharStruct));
    
    for (ix = 0; ix < 224; ix++)
        {
            fs->per_char[ix].width = axsabWidths[ix];
        }
    
    fs->ascent = 13;
    fs->descent = 3;

    return (fs);
}

/* : axGetAlternateAverageWidth - return an average-width
 * figure for use in creating a screenfont based on the
 * WindowsAlternate (which names a typeface family believed
 * to be readily available on win32).  width is returned in milliems
 */
int
axGetAlternateAverageWidth(fontTabEnt *targetRec)
/* fontTabEnt *targetRec;    in: font being represented */
{
    fontTabEnt *altRec;
    int altWidth, targetWidth;
    int avgWidth = 500;              /* half an em by default */
    char *recName = (char *) NULL;
    char *err = (char *) NULL;
    
    if (!targetRec || !targetRec->WindowsAlternate)
        goto all_done;

    /* get the font record for the alternate */

    recName = (char *) NULL;
    err = axGetFontRecordName(targetRec->WindowsAlternate,
                              targetRec->weight,
                              targetRec->slant,
                              &recName);
    if (err)
        {
        goto all_done;
        }
        
    err = axGetFontRecord(recName, &altRec);
    if (err)
        {
        goto all_done;
        }
#if !FOR_FILTER
    if ( altRec->avgWidth < 0 )
        {
        /* a temporary flag value.  Get the real valu7e now by creating */
        /* an instance of the font in the "em" size                     */
        GetWinPrinterAverageWidth( altRec );
        }
#endif
    
    altWidth = axGetTargetAverageWidth(altRec);
    targetWidth = axGetTargetAverageWidth(targetRec);

    if (altWidth)
        {
        avgWidth = (int) (((float) targetWidth *
                           (float) altRec->avgWidth) / altWidth);
        }
    else
        {
        avgWidth = 500;    /* half an em by default, rather fat */
        }

 all_done:
    if (err)
        free(err);
    if (recName)
        free(recName);

    return (avgWidth);
}


/* is a font one of the fonts we enumerated from Windows ? */
/***** Only call these 2 functions if you don't have a font record. *****/
int 
IsWinFont(char *familyName)
{
    int retval;
    char *errString;
    fontTabEnt *fontRecord;

    retval = FALSE;
    if ( errString = axGetFontRecord(familyName, &fontRecord) )
        freeErrString(errString);
    else if ( fontRecord->isWindowsFont )
        retval = TRUE;
    return retval;
}

int
IsPrinterFont(char *familyName)
{
    int retval;
    char *errString;
    fontTabEnt *fontRecord;

    retval = FALSE;
    if ( errString = axGetFontRecord(familyName, &fontRecord) )
        freeErrString(errString);
    else if ( fontRecord->isPrinterFont )
        retval = TRUE;
    return retval;
}
/************************************************************************/
