/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winicon.c
   Applix ASC version: 10
   Date: Mon Nov 16 13:53:39 1998
*/
#include <stdlib.h>
#include <basicwin.h>
#include <aos.h>
#include <xmlib.h>
#include <xmpixel.h>
#include <aiconids.h>

typedef struct {
    char *name;
    HICON icon_handle;
} win_icon_cache_t;

static int WinIconCacheSize = 0;
static win_icon_cache_t *WinIconCache;


extern HANDLE hInst;
extern HDC RootWin;
extern unsigned int AxWinIconWidth, AxWinIconHeight;

#include <winicon.p>


static void
PaintIconPixmap(HDC hdc, HBITMAP pixmap, HBITMAP mask)
/*** hdc must be unpackaged! ***/
{
    HDC mem_hdc;
    HBITMAP old_hbm;
    BITMAP bm;


    mem_hdc = MS_CreateCompatibleDC((HDC)NULL);
    old_hbm = MS_SelectObject(mem_hdc, pixmap);
    MS_GetObject(pixmap, sizeof(BITMAP), (LPSTR) &bm);

    /* for now:  if we have a mono image, paint it black on white. */
    if ( bm.bmBitsPixel == 1 )
        {
        //mask = pixmap;
        MS_SetBkColor(hdc, COLOR_BLACK);
        MS_SetTextColor(hdc, COLOR_WHITE);
        }

    /* center the bitmap (assuming a 32x32 image to be placed
     * into a 36x36 space).
     */
    MS_BitBlt(hdc, 2, 2, 32, 32, mem_hdc, 0, 0, SRCCOPY);

    /* cleanup */
    MS_SelectObject(mem_hdc, old_hbm);
    MS_DeleteDC(mem_hdc);
}

void
PaintWindowsIcon(HDC hdc, int icon_id)
/*** hdc must be unpackaged! ***/
{
    HDC pkg_dc;
    char name[100];
    pixcache_t *p;
    HICON hicon;

/* As of now (01-DEC-95), all icons are Applix bitmaps (.im),
 * and not Windows resource icons (.ico).
 * This means 2 things: 1) we blt the image, centered. 2) we need
 * to convert the image to an icon handle (HICON) for icon dragging
 * and Alt-Tab switching.             -donato
 */

    /* we "package" the DC for the generic drawing calls. */
    pkg_dc = AxPackageHandle(hdc, AX_DC_TYPE);

    /* We draw a bevel around the 36x36 Windows icon space.
     * The icon image will later be centered in the space.
     */
    AxDrawBevel((AXWidget)NULL, pkg_dc, 0, 0, 36, 36, 2, BEVEL_OUT);
    /* get and draw resource icon. */
    /* assume we have an Applixware icon image or user-defined icon. */
    sprintf((char*)name, "%d-%s", icon_id, NXLT("32x32"));
    if ( !(p = Pixmapper(name)) )
        {
        /* use default icon as last resort. */
        sprintf((char*)name, "%s", NXLT("windefault"));
        p = Pixmapper(name);
        }
    /* we won't have a DC during icon drag. */
    if ( hdc && p )
        PaintIconPixmap(hdc, p->pix, p->clip_mask);
    else
        {
        /* during icon-drag (or Alt-Tab Window switching) we need
         * to convert the bitmap handle to an icon handle.
         * We are also keeping a cache so we don't have to keep 
         * recreating icon handles.
         */
        hicon = IconInCache(name);
        if ( p && !hicon )
            {
            hicon = BitmapToIconHandle(p->pix, p->clip_mask, name, TRUE);
            AddIconToCache(hicon, name);
            }
        }

    REMOVE_DRAWABLE(pkg_dc);

#if 0
/* we only need DrawIcon for Windows resource icons. */
    if ( (int)hicon > 0 )
        MS_DrawIcon(hdc, 0, 0, hicon);
#endif
}


#if 1    /* { */
HICON
GetIconHandleFromId(int icon_id, BOOL normal_icon)
{
    char name[100];
    pixcache_t *p;
    HICON hicon;

    if ( normal_icon )
        sprintf((char*)name, "%d-%s", icon_id, NXLT("32x32"));
    else
        /* should only be true on Windows 95! */
        sprintf((char*)name, "%d-%s", icon_id, NXLT("16x16"));

    if ( !(p = Pixmapper(name)) )
        {
        /* use default icon as last resort. */
        sprintf((char*)name, "%s", NXLT("windefault"));
        p = Pixmapper(name);
        }

    hicon = IconInCache(name);
    if ( p && !hicon )
        {
        hicon = BitmapToIconHandle(p->pix, p->clip_mask, name, normal_icon);
        AddIconToCache(hicon, name);
        }
    return hicon;
}

#else    /* } { */
HICON
GetIconHandleFromId(int icon_id, HDC hdc)
{
    HICON hicon;

    switch (icon_id)
        {
        case IC_UNKNOWN:
            /* no icon specified, use a default to emulate X. */
            hicon = MS_LoadIcon(hInst, MAKEINTRESOURCE(1) );
            break;
        case IC_MAIN:      /* main menu */
            hicon = MS_LoadIcon(hInst, MAKEINTRESOURCE(1) );
            break;
        case IC_DC:        /* Words */
            hicon = MS_LoadIcon(hInst, MAKEINTRESOURCE(3) );
            break;
        case IC_GE:
            hicon = MS_LoadIcon(hInst, MAKEINTRESOURCE(4) );
            break;
        case IC_SS:
            hicon = MS_LoadIcon(hInst, MAKEINTRESOURCE(5) );
            break;
        case IC_CE:        /* macro editor */
            hicon = MS_LoadIcon(hInst, MAKEINTRESOURCE(6) );
            break;
        case IC_DIALOG:    /* dialog editor */
            hicon = MS_LoadIcon(hInst, MAKEINTRESOURCE(7) );
            break;
#if 0
        case IC_FILING:
        case IC_HELP:
        case IC_PIXED:
        case IC_SQL:
        case IC_TUTOR:
        case IC_EQN:
        case IC_SSPR:
        case IC_SSPG:
#endif
        case IC_AB:
            /* builder */
            hicon = MS_LoadIcon(hInst, MAKEINTRESOURCE(2) );
            break;
        default:
            {
            char name[100];
            pixcache_t *p;
            /* assume we have an Applixware icon image or user-defined icon. */
            sprintf((char*)name, "%d-%s", icon_id, "32x32");
            p = Pixmapper(name);
            /* we won't have a DC during icon drag. */
            if ( hdc && p )
                {
                PaintIconPixmap(hdc, p->pix, p->clip_mask);
                /* a flag to indicate we have already painted the icon. */
                hicon = (HICON)-1;
                }
            else
                {
                /* during icon-drag (or Alt-Tab Window switching) we need
                 * to convert the bitmap handle to an icon handle.
                 * We are also keeping a cache so we don't have to keep 
                 * recreating icon handles.
                 */
                hicon = IconInCache(name);
                if ( p && !hicon )
                    {
                    hicon = BitmapToIconHandle(p->pix, p->clip_mask, name, TRUE);
                    AddIconToCache(hicon, name);
                    }
                else if ( !hicon )
                    {
                    /* last resort */
                    hicon = MS_LoadIcon(hInst, NXLT("axmain"));
                    }
                }
            break;
            }
        }
    return hicon;
}
#endif    /* } */



/************************* ICON CACHE FUNCTIONS *****************************/
static HICON
IconInCache(char *icon_name)
{
    int ix;
    /* search for the icon (by name) in the icon cache */
    for ( ix=0; ix < WinIconCacheSize; ix++ )
        {
        if ( streq(WinIconCache[ix].name, icon_name) )
            return WinIconCache[ix].icon_handle;
        }
    return NULL;
}

static void
AddIconToCache(HICON hicon, char *icon_name)
{
    int c_ix = WinIconCacheSize;
    WinIconCacheSize = c_ix + 1;
    WinIconCache = (win_icon_cache_t*)realloc(WinIconCache,
                    sizeof(win_icon_cache_t)*WinIconCacheSize);
    WinIconCache[c_ix].name = strcpy(malloc(strlen(icon_name)+1), icon_name);
    WinIconCache[c_ix].icon_handle = hicon;
}

static HICON
BitmapToIconHandle(HBITMAP hbm, HBITMAP mask_hbm, char *icon_name,
                                                BOOL normal_icon)
{
    HDC hdc;
    HICON hicon;
    int i, ret;
    int bytewidth, win_byte_wid;
    int width, height; 
    BYTE depth, planes;
    int scan_lines, nColors;
    int raw_bytes_per_line;
    unsigned char *image_data, *sourceP, *destP;
    PBITMAPINFO  bm_infoP;
    ULONG        sizeBMI;
    int icon_wid, icon_hyt, icon_image_size;
    BYTE *and_bits, *xor_bits;


    hdc = MS_GetDC(NULL);
    if ( normal_icon )
        {
        icon_wid = AxWinIconWidth;
        icon_hyt = AxWinIconHeight;
        }
    else
        {
        /* small (16x16) icon (Windows 95 only). The actual values may differ
         * depending on resolution and/or font mode, but always set to 16
         * and let windows do the stretch/shrink.
         */
        icon_wid = 16;
        icon_hyt = 16;
        }

    /* First call GetDIBits with a zero'ed structure (particularly
     * the bm_infoP->bmiHeader.biBitCount field) to get the image size.
     */
    bm_infoP = (BITMAPINFO *)calloc( 1, sizeof(BITMAPINFO) );
    bm_infoP->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    ret = MS_GetDIBits(hdc, hbm, (UINT)0, (UINT)0, (LPBYTE)NULL,
                   (LPBITMAPINFO)bm_infoP, (DWORD)DIB_RGB_COLORS);
    if ( ret == 0 )
        {
        /* some GDI error, just return. */
        MS_ReleaseDC(NULL, hdc);
        return NULL;
        }

    /* Now that we know the size, allocate space for the image data */
    scan_lines = bm_infoP->bmiHeader.biHeight;
    icon_image_size = bm_infoP->bmiHeader.biSizeImage;
    image_data = malloc(icon_image_size);

    /** Allocate space for the color table (if necessary) */
    /** from \mstools\samples\mandel\dibmp.c **/
    /*
     * Note: 24 bits per pixel has no color table.  So, we don't have to
     * allocate memory for retrieving that.  Otherwise, we do.
     */
    nColors = GetNumDIBColors(&bm_infoP->bmiHeader);
    if ( nColors )
        {
        /* has color table */
        sizeBMI = sizeof(BITMAPINFO) + sizeof(RGBQUAD) * nColors;
        /* I need more memory for the color table */
        bm_infoP = (PBITMAPINFO)realloc( bm_infoP, sizeBMI );
        }
    if ( bm_infoP->bmiHeader.biCompression == BI_BITFIELDS )
        {
        /* need more space for color mask values (3 double words). */
        sizeBMI = sizeof(BITMAPINFO) + (sizeof(DWORD) * 3);
        bm_infoP = (PBITMAPINFO)realloc( bm_infoP, sizeBMI );
        }
    /* get the bits and the color table */
    ret = MS_GetDIBits(hdc, hbm, 0, scan_lines, (LPSTR)image_data,
                                        bm_infoP, DIB_RGB_COLORS);
    if ( ret != scan_lines )
        {
        /* some GDI error, just return. */
        MS_ReleaseDC(NULL, hdc);
        return NULL;
        }

    width = bm_infoP->bmiHeader.biWidth;
    height = bm_infoP->bmiHeader.biHeight; 
    depth = bm_infoP->bmiHeader.biBitCount;
    planes = bm_infoP->bmiHeader.biPlanes;
    raw_bytes_per_line = ((width * depth) + 7)/8;
    if ( depth == 1 )
        {
        bytewidth = icon_image_size/height;
        win_byte_wid = ((width + 15) & ~0xF)/8;
        }
    else 
        bytewidth = win_byte_wid = ((raw_bytes_per_line + 3)/4)*4;

    if ( depth != 8 )
        icon_image_size = raw_bytes_per_line * height;

    and_bits = malloc(icon_image_size);
    memset(and_bits, 0x00, icon_image_size);
    xor_bits = malloc(icon_image_size);
    memset(xor_bits, 0x00, icon_image_size);

    /* flip the image. */
    for ( i=0; i < height; i++ )
        {
        destP = xor_bits + ((height - i - 1) * win_byte_wid);
        sourceP = (unsigned char *)(image_data + (i * bytewidth));
        memcpy(destP, sourceP, win_byte_wid);
        }

    if ( depth == 1 )
        {
        /* I'm not sure why, but with monochrome images we need 
         * to invert the bits.        -donato
         */
        destP = (unsigned char *)xor_bits;
        for ( i=0; i < icon_image_size; i++ )
            {
            *destP = ~(*destP);
            destP++;
            } 
        }

    /* the Windows CreateIcon call requires AND and XOR bits.
     * we set the AND bits to zero and use the actual image data
     * as the XOR bits (therefore, there is no transparency).
     * the bitmasks are combined using the following truth table:
     *   AND   XOR   Display
     *    0     0     Black
     *    0     1     White
     *    1     0     Screen
     *    1     1     Reverse of screen
     */
    /* create the actual icon handle (GDI object). */
    hicon = MS_CreateIcon(hInst, icon_wid, icon_hyt, planes, depth,
                                            and_bits, xor_bits);

    /* cleanup */
    free(and_bits);
    free(xor_bits);
    free(image_data);
    free(bm_infoP);
    MS_ReleaseDC(NULL, hdc);
    return hicon;
}


void
DeleteIconFromCache(int icon_id)
{
    int ix;
    char icon_name[100];

    sprintf((char*)icon_name, "%d-%s", icon_id, NXLT("32x32"));
    for ( ix=0; ix < WinIconCacheSize; ix++ )
        {
        if ( streq(WinIconCache[ix].name, icon_name) )
            {
            /* if the icon handle is in the cache, then it was created
             * using MS_CreateIcon, so we must destroy it.
             */
            MS_DestroyIcon(WinIconCache[ix].icon_handle);
            free(WinIconCache[ix].name);
            if ( ix < WinIconCacheSize )
                {
                memcpy((char*)&WinIconCache[ix],
                        (char*)&WinIconCache[ix+1],
                        sizeof(win_icon_cache_t)*(WinIconCacheSize-ix));
                }
            WinIconCacheSize--;
            WinIconCache = (win_icon_cache_t*)realloc(WinIconCache,
                            sizeof(win_icon_cache_t)*WinIconCacheSize);
            return;
            }
        }
}
/********************** END ICON CACHE FUNCTIONS ****************************/

