/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winimage.c
   Applix ASC version: 35
   Date: Fri May 28 21:11:49 1999
*/
#include <basicwin.h>
#include <stdio.h>
#include <math.h>
#include <tmi.h>
#include <atm.h>
#include <atmattr.h>
#include <axwinclr.h>
#include <pfpat.h>

#define MILLS_TO_PIXEL 1


extern HPALETTE   AxWinPaletteHandle;
extern LOGPALETTE *AxWinPalette;
extern unsigned long AlisPixels[];
extern int Running_Windows95;

#include <winimage.p>

/*** TEMP: move to winmap.h! ***/
#define MS_FindResource		FindResource
#define MS_FreeResource		FreeResource
#define MS_LoadResource		LoadResource
#define MS_LockResource		LockResource
#define MS_UnlockResource	UnlockResource
/*******************************/

/*****************************************************************
* function: AxCreateWinColorBitmap - creates a Windows DIB handle
*                                    from an XImage.
*/
static HBITMAP
AxCreateWinColorBitmap(XImage *xim, cmap_type *cmap, int pid, int depth)
{
    int ix, i, j, width, height, nColors;
    unsigned int bytewidth, win_bytewidth, win_imgSize;
    HDC hdc;
    HBITMAP hbm;
    ULONG   sizeBMI;
    RGBQUAD *colorP;
    PBITMAPINFO  bm_infoP;
    HPALETTE old_palette;
    unsigned char *new_data, *sourceP, *destP;


    hdc = MS_GetDC((HWND)NULL);
    width  = xim->width;
    height = xim->height;

    bm_infoP = (BITMAPINFO *)calloc( 1, sizeof(BITMAPINFO) );
    bm_infoP->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bm_infoP->bmiHeader.biWidth = width;
    if ( depth > 8 )
        bm_infoP->bmiHeader.biHeight = -(height);
    else
        bm_infoP->bmiHeader.biHeight = height;
    bm_infoP->bmiHeader.biPlanes = 1;
    bm_infoP->bmiHeader.biCompression = BI_RGB;
    bm_infoP->bmiHeader.biXPelsPerMeter = 0;
    bm_infoP->bmiHeader.biYPelsPerMeter = 0;
    bm_infoP->bmiHeader.biClrUsed = 0;
    bm_infoP->bmiHeader.biClrImportant = 0;
    bm_infoP->bmiHeader.biBitCount = depth;
    bm_infoP->bmiHeader.biSizeImage = (width + 7) / 8 * height * depth;

    if ( depth <= 8 )
        {
        bytewidth = width + (width%2);
        win_bytewidth = (unsigned int)((width + 3) / 4) * 4;
        win_imgSize = win_bytewidth * height;

        nColors = GetNumDIBColors(&bm_infoP->bmiHeader);
        sizeBMI = sizeof(BITMAPINFO) + sizeof(RGBQUAD) * nColors;
        /* I need more memory for the color table */
        bm_infoP = (BITMAPINFO *)realloc(bm_infoP, sizeBMI);

        colorP = &bm_infoP->bmiColors[0];
        for ( ix=0; ix < nColors; ix++ )
            {
            colorP->rgbRed   = AxWinPalette->palPalEntry[ix].peRed;
            colorP->rgbGreen = AxWinPalette->palPalEntry[ix].peGreen;
            colorP->rgbBlue  = AxWinPalette->palPalEntry[ix].peBlue;
            colorP->rgbReserved = 0;
            colorP++;
            }
        old_palette = SelectAndRealizeAxWinPalette(hdc);

        /* pad to 4-byte boundary and flip. */
        sourceP = xim->data;
        new_data = (char*)TaskAlloc(pid, win_imgSize*sizeof(char) );
        for ( i = 0; i < height; i++ )
            {
            destP = new_data + ((height - i - 1) * win_bytewidth);
            memset(destP, 0, win_bytewidth);
            for ( j = 0; j < (int)bytewidth; j++ )
                *destP++ = *sourceP++;
            }
        }
    else
        {
        /* greater than 8 bit image: no color table, and no need to
         * flip and pad the data.
         */
        new_data = xim->data;
        }

    hbm = MS_CreateDIBitmap(hdc, &bm_infoP->bmiHeader, CBM_INIT,
                            new_data, bm_infoP, DIB_RGB_COLORS);

    /* cleanup and return */
    if ( depth <= 8 )
        {
        TaskFree(pid, new_data);
        UnSelectPalette(hdc, old_palette);
        }
    MS_ReleaseDC((HWND)NULL, hdc);
    return hbm;
}



/*****************************************************************
* function: AxCreateWinMonoBitmap - creates a Windows bitmap handle
*                                   from an XImage.
*/
static HBITMAP
AxCreateWinMonoBitmap(XImage *image, int flip)
{
    int width, height;
    HBITMAP hbm;

    width = image->width;
    height = image->height;
    hbm = MS_CreateBitmap(width, height, 1, 1, image->data);
    if ( flip )
        {
        HBITMAP old_hbm;
        HDC mem_dc = MS_CreateCompatibleDC((HDC)NULL);
        old_hbm = MS_SelectObject(mem_dc, hbm);
        MS_PatBlt(mem_dc, 0, 0, width, height, DSTINVERT);
        MS_SelectObject(mem_dc, old_hbm);
        MS_DeleteDC(mem_dc);
        }
    return hbm;
}

void
display_win_image(dst_dc, pid, imP, cmap, x1, y1, xP, mP, c_rect, attr)
HDC dst_dc;
int pid;
ge_image_type *imP;
cmap_type *cmap;
int x1, y1;
XImage *xP, *mP;
XRectangle *c_rect;
attr_type *attr;
{
    unsigned long fore_pixel, back_pixel;
    int delete_dc;
    HBITMAP data_hbm, mask_hbm;
    HPALETTE old_palette;
    HDC src_hdc;
    HRGN clip_region;
    HGDIOBJ old_obj;
    DWORD rop_code;
    POINT pnts[3];

	if ( !xP || xP->width <= 0 || xP->height <= 0 )
		return;

    /* get the drawing dc and set it's palette */
    dst_dc = InsureDrawable(dst_dc, &delete_dc);
    if ( xP->palette_handle )
    {
        old_palette = MS_SelectPalette(dst_dc, xP->palette_handle,FALSE);
        MS_RealizePalette(dst_dc);
    }
    else
        old_palette = SelectAndRealizeAxWinPalette(dst_dc);

    /* create mask and image bitmap handles */
    if ( imP->dep == 1 )
    {
        data_hbm = AxCreateWinMonoBitmap(xP, attr->ffg.see_thru);
        mask_hbm = NULL;

        /* set the pixel colors */
        if ( attr->use_ffg )
            fore_pixel = AlisPixels[TMA_BLACK];
        else
        {
            AxGetRealPixel(2, attr->ffg.c, attr->ffg.m, attr->ffg.y,
                                        attr->ffg.k, 255, &fore_pixel);
        }
        if ( attr->use_fbg )
            back_pixel = AlisPixels[TMA_WHITE];
        else
        {
            AxGetRealPixel(2, attr->fbg.c, attr->fbg.m, attr->fbg.y,
                                        attr->fbg.k, 255, &back_pixel);
        } 
        MS_SetBkColor  (dst_dc, GET_COLOR_VALUE(fore_pixel));
        MS_SetTextColor(dst_dc, GET_COLOR_VALUE(back_pixel));

        /* create a memory dc of what we need to display */
        if ( attr->ffg.see_thru || attr->fbg.see_thru )
        {
            /* the mask becomes the image we want to display */
            mask_hbm = data_hbm;

            /* create a memory dc of ones or zeros based on transparent color */
            src_hdc = MS_CreateCompatibleDC((HDC)NULL);
            data_hbm = MS_CreateCompatibleBitmap(src_hdc, xP->width,xP->height);
            old_obj = MS_SelectObject(src_hdc, data_hbm);
            if ( attr->ffg.see_thru )
                rop_code = BLACKNESS;
            else
                rop_code = WHITENESS;
            MS_PatBlt(src_hdc, 0, 0, xP->width, xP->height, rop_code);
        }
        else
        {
            src_hdc = MS_CreateCompatibleDC((HDC)NULL);
            old_obj = MS_SelectObject(src_hdc, data_hbm);
        }
    }
    else
    {
        data_hbm = AxCreateWinColorBitmap(xP, cmap, pid, xP->depth);
        mask_hbm = mP ? AxCreateWinMonoBitmap(mP, FALSE) : NULL;

        /* create a memory dc of what we need to display */
        src_hdc = MS_CreateCompatibleDC((HDC)NULL);
        old_obj = MS_SelectObject(src_hdc, data_hbm);
    }

    /* set clip region. */
    if ( c_rect )
    {
		clip_region = MS_CreateRectRgn(c_rect->x, c_rect->y,
									   c_rect->x + c_rect->width  - 1,
									   c_rect->y + c_rect->height - 1);
		MS_SelectClipRgn(dst_dc, clip_region);
    }

	/* do it */
    pnts[0].x = x1;             /* upper left corner */
    pnts[0].y = y1;
    pnts[1].x = x1 + xP->width;     /* upper right corner */
    pnts[1].y = y1;
    pnts[2].x = x1;
    pnts[2].y = y1 + xP->height;    /* lower left corner */
    AxWinPlgBlt(dst_dc, pnts, src_hdc, 0,0, xP->width,xP->height, mask_hbm,0,0);

    /* clean up */
    UnSelectPalette(dst_dc, old_palette);
    MS_SelectObject(src_hdc, old_obj);
    MS_DeleteDC(src_hdc);
    MS_DeleteObject(data_hbm);
    MS_SelectClipRgn(dst_dc, (HRGN)NULL);
    MS_DeleteObject(clip_region);
    if ( mask_hbm )
        MS_DeleteObject(mask_hbm);
    if ( delete_dc )
        MS_DeleteDC(dst_dc);
}

/**** called from ax_tm_gimage only (tmimage.c) ****/
/* this is called after a call to XGetImage with an HDC (as opposed    *
 * to a HBITMAP). This means the XImage struct returned from XGetImage *
 * was not completely filled, so we need to correctly fill it here.    */
void
GetWinImageData(XImage *ximage)
{
    HDC hdc;
    HBITMAP hbm;
    int i, j, nColors;
    int raw_bytes_per_line, byte_pad;
    int width, height, depth, bytewidth, win_bytewidth;
    int scan_lines, img_size;
    unsigned char *win_im_data, *sourceP, *destP;
    PBITMAPINFO  bm_infoP;
    ULONG        sizeBMI;
    int ret;
    HPALETTE old_palette;
    LPLOGPALETTE paletteP;
    RGBQUAD *colorP;
    BYTE red, green, blue;


    hdc = MS_GetDC((HWND)NULL);
    if ( ximage->bitmap_handle )
        hbm = ximage->bitmap_handle;
    else
        taskAbort("GetWinImageData: no bitmap handle!");

    old_palette = SelectAndRealizeAxWinPalette(hdc);

    /* First call GetDIBits with a zero'ed structure (particularly     *
     * the bm_infoP->bmiHeader.biBitCount field) to get the image size */
    bm_infoP = (BITMAPINFO *)calloc( 1, sizeof(BITMAPINFO) );
    bm_infoP->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);

    ret = MS_GetDIBits(hdc, hbm, (UINT)0, (UINT)0, (LPBYTE)NULL,
                   (LPBITMAPINFO)bm_infoP, (DWORD)DIB_RGB_COLORS);
    if ( !ret )
        taskAbort("GetWinImageData - GetDIBits returns zero");

    /* Now that we know the size, allocate space for the image data */
    scan_lines = bm_infoP->bmiHeader.biHeight;
    img_size   = bm_infoP->bmiHeader.biSizeImage;

    win_im_data = (char*)malloc(img_size);
    ximage->data = (char*)malloc(img_size);

    /** Allocate space for the color table (if necessary) */
    /** from \mstools\samples\mandel\dibmp.c **/
    /*
     * Note: 24 bits per pixel has no color table.  So, we don't have to
     * allocate memory for retrieving that.  Otherwise, we do.
     */
    nColors = GetNumDIBColors(&bm_infoP->bmiHeader);
    if ( nColors )
        {
        /* has color table */
        sizeBMI = sizeof(BITMAPINFO) + sizeof(RGBQUAD) * nColors;
        /* I need more memory for the color table */
        bm_infoP = (BITMAPINFO *)realloc(bm_infoP, sizeBMI);
        }
    if ( bm_infoP->bmiHeader.biCompression == BI_BITFIELDS )
        {
        /* need more space for color mask values (3 double words). */
        sizeBMI = sizeof(BITMAPINFO) + (sizeof(DWORD) * 3);
        bm_infoP = (BITMAPINFO *)realloc( bm_infoP, sizeBMI );
        }

    /* get the bits and the color table */
    ret = MS_GetDIBits(hdc, hbm, 0, scan_lines, (LPSTR)win_im_data,
                                        bm_infoP, DIB_RGB_COLORS);
    if ( ret != scan_lines )
        taskAbort("GetWinImageData - GetDIBits failed");

    width = bm_infoP->bmiHeader.biWidth;
    height = bm_infoP->bmiHeader.biHeight;
    depth = bm_infoP->bmiHeader.biBitCount;
    bytewidth = width + (width%2);
    win_bytewidth = ((width + 3) / 4) * 4;

    /* fill XImage struct */
    ximage->width   = width;
    ximage->height  = height;
    ximage->xoffset = 0;
    ximage->format  = XYBitmap; 
    if ( bm_infoP->bmiHeader.biBitCount == 16 )
        {
        ximage->byte_order  = LSBFirst;
        ximage->bitmap_bit_order = LSBFirst;
        }
    else
        {
        ximage->byte_order  = MSBFirst;
        ximage->bitmap_bit_order = MSBFirst;
        }
    if ( depth == 1 )
       ximage->bitmap_pad = 16;    /* monochrome is padded to even bytes */
    else
       ximage->bitmap_pad = 32;    /* color is padded to 4 byte boundries */ 
    ximage->bitmap_unit = 8;
    ximage->depth = ximage->bits_per_pixel = depth;

    raw_bytes_per_line = ((width * depth) + 7)/8;
    byte_pad = ximage->bitmap_pad / 8;
    ximage->bytes_per_line =
                       ((raw_bytes_per_line + byte_pad - 1)/byte_pad)*byte_pad;
    ximage->bitmap_info = bm_infoP;

    /* GetDIBits will retrieve the bits in Windows format (origin at *
     * bottom left).  So flip the bits here so the GE gets them in a *
     * "normal" format (origin at top left).                         */
    sourceP = win_im_data;
    for ( i = 0; i < height; i++ )
        {
        destP = ximage->data + ((height - i - 1) * ximage->bytes_per_line);
        memset(destP, 0, ximage->bytes_per_line);
        for ( j = 0; j < ximage->bytes_per_line; j++ )
            {
            /* HACK HACK HACK: force GDI white (255) to Applix
             * white (AlisPixels[0]) so we get transparency during
             * digitize with a white background.    -donato
             */
            if ( DpyDepth == 8 && *sourceP == 0xFF )
                *sourceP = (unsigned char)AlisPixels[0];
            *destP++ = *sourceP++;
            }
        }
    free(win_im_data);

    /* Create a palette as necessary. */
    if ( nColors )
        {
        paletteP = (LOGPALETTE*)malloc(sizeof(LOGPALETTE) +
                            (nColors*sizeof(PALETTEENTRY)));
        paletteP->palNumEntries = nColors;
        paletteP->palVersion = PALVERSION;

        colorP = &ximage->bitmap_info->bmiColors[0];
        /* don't insert the Windows static colors (0-9, 246-255),
         * because they conflict with Applix colors.  -donato
         */
        for ( i=0; i < nColors; i++ )
            {
            paletteP->palPalEntry[i].peRed   = colorP->rgbRed;
            paletteP->palPalEntry[i].peGreen = colorP->rgbGreen;
            paletteP->palPalEntry[i].peBlue  = colorP->rgbBlue;
            paletteP->palPalEntry[i].peFlags = 0;
            colorP++;
            }
        for ( i=1; i < 10; i++ )
            {
            paletteP->palPalEntry[i].peRed =
                paletteP->palPalEntry[i].peGreen =
                    paletteP->palPalEntry[i].peBlue = 1;
                        paletteP->palPalEntry[i].peFlags = 0;
            }
        for ( i=246; i < 249; i++ )
            {
            paletteP->palPalEntry[i].peRed =
                paletteP->palPalEntry[i].peGreen =
                    paletteP->palPalEntry[i].peBlue = 1;
                        paletteP->palPalEntry[i].peFlags = 0;
            }
        ximage->palette_handle = MS_CreatePalette(paletteP);
        free(paletteP);
        }
    else if ( bm_infoP->bmiHeader.biBitCount == 24 )
        {
        /* A 24 bitcount DIB has no color table entries so, set the number of
         * to the maximum value (256).
         */
        nColors = MAXPALETTE;
        paletteP = (LOGPALETTE*)malloc(sizeof(LOGPALETTE) +
                            (nColors*sizeof(PALETTEENTRY)));
        if ( paletteP )
            {
            paletteP->palNumEntries = nColors;
            paletteP->palVersion    = PALVERSION;

            red = green = blue = 0;
            /* Generate 256 (= 8*8*4) RGB combinations to fill the palette
             * entries.
             */
            for ( i = 0; i < paletteP->palNumEntries; i++ )
                {
                paletteP->palPalEntry[i].peRed   = red;
                paletteP->palPalEntry[i].peGreen = green;
                paletteP->palPalEntry[i].peBlue  = blue;
                paletteP->palPalEntry[i].peFlags = (BYTE)0;

                if ( !(red += 32) )
                    if ( !(green += 32) )
                        blue += 64;
                }
            ximage->palette_handle = MS_CreatePalette(paletteP);
            free(paletteP);
            }
        }
    SetXColorMasks(bm_infoP, ximage);

    UnSelectPalette(hdc, old_palette);
    MS_ReleaseDC((HWND)NULL, hdc);
}



/*************************************************************************
        box[0] = LLC of rect
        box[1] = URC of rect
        box[2] = LLC of rotated rect
        box[3] = width, height of rotated rect
************************************************************************/
#define TX(x, y)    ((x-orgx)*cosa - (y-orgy)*sina)
#define TY(x, y)    ((x-orgx)*sina + (y-orgy)*cosa)
static void
AxWinGetImageBox(attr_type *attr, TmPoint32 *box)
{
    double minx, miny, maxx, maxy, xr, yr;
    double angle, sina, cosa;
    double orgx, orgy;

    if ( attr->f_ang == 0 )
        {
        box[2].x = 0;
        box[2].y = 0;
        box[3].x = ABS(box[1].x - box[0].x);
        box[3].y = ABS(box[1].y - box[0].y);
        return;
        }

    angle = ((double)(-attr->f_ang) * M_PI) / 180.0;
    sina = sin(angle);
    cosa = cos(angle);
    orgx = box[0].x;
    orgy = box[0].y;

    /* init boundary */
    minx = maxx = TX(box[0].x, box[0].y);
    miny = maxy = TY(box[0].x, box[0].y);

    xr = TX(box[0].x, box[1].y);
    yr = TY(box[0].x, box[1].y);

    if (xr < minx)
        minx = xr;
    else if (xr > maxx)
        maxx = xr;
    if (yr < miny)
        miny = yr;
    else if (yr > maxy)
        maxy = yr;

    xr = TX(box[1].x, box[1].y);
    yr = TY(box[1].x, box[1].y);

    if (xr < minx)
        minx = xr;
    else if (xr > maxx)
        maxx = xr;
    if (yr < miny)
        miny = yr;
    else if (yr > maxy)
        maxy = yr;

    xr = TX(box[1].x, box[0].y);
    yr = TY(box[1].x, box[0].y);

    if (xr < minx)
        minx = xr;
    else if (xr > maxx)
        maxx = xr;
    if (yr < miny)
        miny = yr;
    else if (yr > maxy)
        maxy = yr;

    box[2].x = (int)minx;
    box[2].y = (int)miny;
    box[3].x = (int)maxx;
    box[3].y = (int)maxy;
}

void
GetForegroundPixelFromSubattr(sub_attr_type *subattr, int *pixel)
{
    double fg_intensity, bg_intensity;

    fg_intensity = 255.0 * IntensFromCMYKColor(subattr->fg.c,
                                            subattr->fg.m,
                                            subattr->fg.y,
                                            subattr->fg.k, 255);
    bg_intensity = 255.0 * IntensFromCMYKColor(subattr->bg.c,
                                            subattr->bg.m,
                                            subattr->bg.y,
                                            subattr->bg.k, 255);
    if ( fg_intensity < bg_intensity )
        *pixel = 0;
    else
        *pixel = 1;
}

void
CreateCmapFor2ColorDIB(cmap_type *cmap, sub_attr_type *subattr, int fg_pixel)
{
    cmap->n_colors = 2;

    /*** CALLER MUST FREE color_type!!! ***/
    cmap->color = (color_type *)malloc(2 * sizeof(color_type));
    if ( fg_pixel == 1 )
        {
        /* foreground color goes in slot 1. */
        cmap->color[1].c = subattr->fg.c;
        cmap->color[1].m = subattr->fg.m;
        cmap->color[1].y = subattr->fg.y;
        cmap->color[1].k = subattr->fg.k;
        /* background color in slot 0 */
        cmap->color[0].c = subattr->bg.c;
        cmap->color[0].m = subattr->bg.m;
        cmap->color[0].y = subattr->bg.y;
        cmap->color[0].k = subattr->bg.k;
        }
    else
        {
        cmap->color[0].c = subattr->fg.c;
        cmap->color[0].m = subattr->fg.m;
        cmap->color[0].y = subattr->fg.y;
        cmap->color[0].k = subattr->fg.k;
        /* background color in slot 1 */
        cmap->color[1].c = subattr->bg.c;
        cmap->color[1].m = subattr->bg.m;
        cmap->color[1].y = subattr->bg.y;
        cmap->color[1].k = subattr->bg.k;
        }
}


/* the following is "stolen" from the applixware pcl and gdi driver code. */
void
AxWinDrawGradientImage(HGDIOBJ drawable, int xpos, int ypos, int wid, int hyt,
                                attr_type *attr, HBITMAP mask, BOOL for_lines)
{
    int fg_pixel;
    XImage *xim, *xim_intens;
    BITMAPINFO *bitmap_info;
    cmap_type cmap;    /* to hold the 2 colors to create a DIB with */
    int i, j, win_bytewidth, remove_drawable;
    unsigned char *sourceP, *destP, *new_data;
    sub_attr_type subattr;
    TmPoint32 boundingBox[4];
    HDC draw_dc, src_dc, mask_dc;
    HBITMAP hbm, old_hbm, old_mask_hbm;
    COLORREF old_bk_color, old_text_color;
    DWORD rop_code;
    HPALETTE old_palette;


    if ( (hyt == 0) || (wid == 0) )
        return;

    draw_dc = InsureDrawable(drawable, &remove_drawable);
    old_palette = SelectAndRealizeAxWinPalette(draw_dc);

    boundingBox[0].x = xpos;
    boundingBox[0].y = ypos;
    boundingBox[1].x = xpos + wid;
    boundingBox[1].y = ypos + hyt;
    AxWinGetImageBox(attr, boundingBox);

    memset(&subattr, 0, sizeof(subattr));
    if ( for_lines )
        {
        subattr.type = attr->l_type;
        subattr.fg   = attr->lfg;
        subattr.bg   = attr->lbg;
        subattr.ang  = attr->l_ang;
        subattr.xoff = attr->l_xoff;
        subattr.yoff = attr->l_yoff;
        }
    else
        {
        subattr.type = attr->f_type;
        subattr.fg   = attr->ffg;
        subattr.bg   = attr->fbg;
        subattr.ang  = attr->f_ang;
        subattr.xoff = attr->f_xoff;
        subattr.yoff = attr->f_yoff;
        }

    /* first create an ximage of intensity values. */
    switch (subattr.type)
        {
        case LINEAR_FILL_TYPE:
            xim_intens = LinearGradientXImage(&subattr, boundingBox);
            break;
        case BILINEAR_FILL_TYPE:
            xim_intens = BilinearGradientXImage(&subattr, boundingBox);
            break;
        case RADIAL_FILL_TYPE:
            xim_intens = RadialGradientXImage(&subattr, boundingBox);
            break;
        case RECT_FILL_TYPE:
            xim_intens = RectGradientXImage(&subattr, boundingBox);
            break;
        default:
            xim_intens = NULL;
        }
    if ( !xim_intens )
        return;    /* uh oh */

    /* dither down to a monochrome image. */
    xim = FS_HalfTone(xim_intens);
    pfFreeXImage(xim_intens);

    /* get colors */
    GetForegroundPixelFromSubattr(&subattr, &fg_pixel);
    CreateCmapFor2ColorDIB(&cmap, &subattr, fg_pixel);
    /* create a 2-color DIB */
    bitmap_info = AxGetDIBInfo(xim->width, xim->height, &cmap, AX_DIB_NORMAL);

    /* pad and flip the data for the DIB call. */
    win_bytewidth = ((xim->bytes_per_line + 3) / 4) * 4;
    bitmap_info->bmiHeader.biSizeImage = win_bytewidth * xim->height;
    sourceP = xim->data;
    new_data = (unsigned char*)malloc(win_bytewidth*xim->height);
    for ( i = 0; i < xim->height; i++ )
        {
        destP = new_data + ((xim->height - i - 1) * win_bytewidth);
        memset(destP, 0, win_bytewidth);
        for ( j = 0; j < xim->bytes_per_line; j++ )
            *destP++ = *sourceP++;
        }

    /* create the source bitmap */
    hbm = MS_CreateDIBitmap(draw_dc, &bitmap_info->bmiHeader, CBM_INIT,
                                new_data, bitmap_info, DIB_RGB_COLORS);
    src_dc = MS_CreateCompatibleDC(draw_dc);
    old_hbm = MS_SelectObject(src_dc, hbm);

    /* setup the mask dc */
    mask_dc = MS_CreateCompatibleDC(draw_dc);
    old_mask_hbm = MS_SelectObject(mask_dc, mask);

    /* "True Mask Method" of doing transparency. */
#if 0
    MS_StretchBlt(draw_dc, xpos, ypos, wid, hyt, src_dc, 0, 0,
                            xim->width, xim->height, SRCINVERT);
    /* set up the destination correctly */
    old_bk_color = MS_SetBkColor(draw_dc, PALETTERGB(255, 255, 255));
    old_text_color = MS_SetTextColor(draw_dc, PALETTERGB(0,0,0));
    rop_code = 0x220326;   /* not source and destination */
    MS_BitBlt(draw_dc, xpos, ypos, wid, hyt, mask_dc, 0, 0, rop_code);
    /* restore destination */
    MS_SetBkColor(draw_dc, old_bk_color);
    MS_SetTextColor(draw_dc, old_text_color);
    MS_StretchBlt(draw_dc, xpos, ypos, wid, hyt, src_dc, 0, 0,
                            xim->width, xim->height, SRCINVERT);
#else
    MS_BitBlt(draw_dc, xpos, ypos, wid, hyt, src_dc, 0, 0, SRCINVERT);
    /* set up the destination correctly */
    old_bk_color = MS_SetBkColor(draw_dc, PALETTERGB(255, 255, 255));
    old_text_color = MS_SetTextColor(draw_dc, PALETTERGB(0,0,0));
    rop_code = 0x220326;   /* not source and destination */
    MS_BitBlt(draw_dc, xpos, ypos, wid, hyt, mask_dc, 0, 0, rop_code);
    /* restore destination */
    MS_SetBkColor(draw_dc, old_bk_color);
    MS_SetTextColor(draw_dc, old_text_color);
    MS_BitBlt(draw_dc, xpos, ypos, wid, hyt, src_dc, 0, 0, SRCINVERT);
#endif

    /* cleanup */
    MS_SelectObject(mask_dc, old_mask_hbm);
    /* the caller will delete the mask bitmap that was passed in. */
    MS_DeleteDC(mask_dc);
    MS_SelectObject(src_dc, old_hbm);
    MS_DeleteObject(hbm);
    MS_DeleteDC(src_dc);

    pfFreeXImage(xim);
    free(bitmap_info);
    free(cmap.color);
    free(new_data);
    UnSelectPalette(draw_dc, old_palette);
    if ( remove_drawable )
        RemoveDrawable(drawable, draw_dc);
}


/*** CALLER must FREE the pointer that this returns!! ***/
BITMAPINFO *
AxGetDIBInfo(int pix_width, int pix_height, cmap_type *cmap, int dib_type)
{
    int k, ix;
    int nColors, bmi_size;
    int bytewidth, image_size;
    RGBQUAD *colorP;
    BITMAPINFO *bitmap_info;
    BITMAPINFOHEADER *headerP;

    nColors = cmap->n_colors;
    bytewidth = ((pix_width + 3) / 4) * 4;
    image_size = bytewidth * pix_height;

    bmi_size = sizeof(BITMAPINFO) + (nColors*sizeof(RGBQUAD));
    bitmap_info = (BITMAPINFO *)calloc(1, bmi_size);
    memset(bitmap_info, 0, bmi_size);

    headerP = &bitmap_info->bmiHeader;
    headerP->biSize = sizeof(BITMAPINFOHEADER);
    headerP->biWidth = pix_width;
    headerP->biHeight = pix_height;  /* negative puts corner at upper left */
    headerP->biPlanes = 1;
    if ( nColors > 2 && dib_type != AX_DIB_MASK )
        headerP->biBitCount = 8;
    else
        headerP->biBitCount = 1;
    headerP->biCompression = BI_RGB;    /* no compression */
    headerP->biSizeImage = image_size;  /* size in bytes */
	if ( dib_type == AX_DIB_MASK)
		{
    	headerP->biClrUsed = 2;
    	headerP->biClrImportant = 2;
		}
	else
		{
    	headerP->biClrUsed = nColors;       /* full color table is used */
    	headerP->biClrImportant = nColors;  /* all colors are important */
		}

    colorP = &bitmap_info->bmiColors[0];
    for ( ix=0; ix < nColors; ix++ )
        {
        k = cmap->color[ix].k;
        colorP->rgbRed   = 255 - MIN(255, cmap->color[ix].c + k);
        colorP->rgbGreen = 255 - MIN(255, cmap->color[ix].m + k);
        colorP->rgbBlue  = 255 - MIN(255, cmap->color[ix].y + k);
		if ( dib_type == AX_DIB_TRANSPARENT && cmap->color[ix].see_thru )
			{
			colorP->rgbRed = colorP->rgbGreen = colorP->rgbBlue = 0;
			}
		else if ( dib_type == AX_DIB_MASK && cmap->color[ix].see_thru )
			{
			colorP->rgbRed = colorP->rgbGreen = colorP->rgbBlue = 255;
			}
		else if ( dib_type == AX_DIB_MASK )
			{
			colorP->rgbRed = colorP->rgbGreen = colorP->rgbBlue = 0;
			}
        colorP->rgbReserved = 0;
        colorP++;
        }
    return bitmap_info;
}


#if 0
/*** CALLER MUST FREE pointer that this returns!! ***/
static POINT *
AxGetPlgPointsFromXform(poly_type *xform, int xpos, int ypos,
                                double x_scale, double y_scale)
{
    int i, x, y, minx, miny, npoints;
    POINT *points;

    /* convert to pixels */
    npoints = xform->nPoints;
    points = (POINT*)malloc(npoints*sizeof(POINT));
    for ( i=0; i < npoints; i++ )
        {
        conv_i_p32(&xform->points[i].x, &xform->points[i].y, &x, &y,
                                                    MILLS_TO_PIXEL);
        points[i].x = x;
        points[i].y = y;
        if ( i == 0 )
            {
            minx = x;
            miny = y;
            }
        else
            {
            if ( x < minx )
                minx = x;
            if ( y < miny )
                miny = y;
            }
        }
    /* zoom the size and set the position */
    for ( i=0; i < npoints; i++ )
        {
        points[i].x = (points[i].x - minx) * x_scale + xpos;
        points[i].y = (points[i].y - miny) * y_scale + ypos;
        }
    return points;
}
#endif


/*************************************************************************
	The following three functions are taken from
	Microsoft Knowledge Base article: Q124947

	Caller must delete both the bitmap and the palette.
 *************************************************************************/
HBITMAP 
axmsLoadBitmapWithPalette(HINSTANCE hInstance, char *bitmap_name, 
										HPALETTE *palette_handle)
{
    HRSRC  hRsrc;
    HGLOBAL hGlobal;
    HBITMAP hBitmapFinal = NULL;
    BITMAPINFOHEADER *lpbi;
    HDC hdc;
    int num_colors;

	if ( hRsrc = MS_FindResource(hInstance, bitmap_name, RT_BITMAP) )
    	{
		hGlobal = MS_LoadResource(hInstance, hRsrc);
		lpbi = (BITMAPINFOHEADER*)MS_LockResource(hGlobal);

		hdc = MS_GetDC(NULL);
		*palette_handle =  axmsCreateDIBPalette((BITMAPINFO*)lpbi,&num_colors);
		if ( *palette_handle )
			{
			MS_SelectPalette(hdc, *palette_handle, FALSE);
			MS_RealizePalette(hdc);
			}

		hBitmapFinal = MS_CreateDIBitmap(hdc, (BITMAPINFOHEADER*)lpbi, CBM_INIT,
                   (LPSTR)lpbi + lpbi->biSize + num_colors * sizeof(RGBQUAD),
                   (LPBITMAPINFO)lpbi, DIB_RGB_COLORS);

		MS_ReleaseDC(NULL, hdc);
		MS_UnlockResource(hGlobal);
		MS_FreeResource(hGlobal);
		}

	return hBitmapFinal;
}

HPALETTE 
axmsCreateDIBPalette(LPBITMAPINFO lpbmi, LPINT lpiNumColors) 
{ 
	LPBITMAPINFOHEADER  lpbi;
	LPLOGPALETTE     lpPal;
	HANDLE           hLogPal;
	HPALETTE         hPal = NULL;
	int              i;

	lpbi = (LPBITMAPINFOHEADER)lpbmi;
	if ( lpbi->biBitCount <= 8 )
		*lpiNumColors = (1 << lpbi->biBitCount);
	else
		*lpiNumColors = 0;  // No palette needed for 24 BPP DIB

	if ( *lpiNumColors )
		{
		hLogPal = MS_GlobalAlloc (GHND, sizeof (LOGPALETTE) +
                             sizeof (PALETTEENTRY) * (*lpiNumColors));
		lpPal = (LPLOGPALETTE) MS_GlobalLock (hLogPal);
		lpPal->palVersion    = 0x300;
		lpPal->palNumEntries = *lpiNumColors;

		for ( i = 0;  i < *lpiNumColors;  i++ )
			{
			lpPal->palPalEntry[i].peRed   = lpbmi->bmiColors[i].rgbRed;
			lpPal->palPalEntry[i].peGreen = lpbmi->bmiColors[i].rgbGreen;
			lpPal->palPalEntry[i].peBlue  = lpbmi->bmiColors[i].rgbBlue;
			lpPal->palPalEntry[i].peFlags = 0;
			}

		hPal = MS_CreatePalette (lpPal);
		MS_GlobalUnlock (hLogPal);
		MS_GlobalFree   (hLogPal);
		}

	return hPal;
}

void
axmsDrawWindowsBitmapWithPalette(HDC hdc, HBITMAP hbm, HPALETTE hpalette,
													int xpos, int ypos)
{
	HDC mem_dc;
	HBITMAP old_hbm;
	HPALETTE old_palette, old_mem_palette;
	BITMAP bm;

	MS_GetObject(hbm, sizeof(BITMAP), (LPSTR)&bm);

	old_palette = MS_SelectPalette(hdc, hpalette, FALSE);
	MS_RealizePalette(hdc);

	mem_dc = MS_CreateCompatibleDC(hdc);
	old_mem_palette = MS_SelectPalette(mem_dc, hpalette, FALSE);
	MS_RealizePalette(mem_dc);
	old_hbm = MS_SelectObject(mem_dc, hbm);

	MS_BitBlt(hdc, xpos, ypos, bm.bmWidth, bm.bmHeight, mem_dc, 0,0,SRCCOPY);

	MS_SelectObject(mem_dc, old_hbm);
	MS_SelectPalette(mem_dc, old_mem_palette, FALSE);
	MS_DeleteDC(mem_dc);
	MS_SelectPalette(hdc, old_palette, FALSE);
}
/****************************************************************************/
