/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winprintf.c
   Applix ASC version: 4
   Date: Mon Nov 16 13:53:46 1998
*/
#include <stdlib.h>
#include <math.h>
#include <basicwin.h>
#include <aos.h>

static HWND  printf_hwnd, listbox_hwnd;
static int num_list_entries = 0;
static int listbox_scroll_width = 0;


extern HANDLE hInst;


#define DEFAULT_PRINTF_WIDTH 600
#define DEFAULT_PRINTF_HEIGHT 250


#include <winprintf.p>



long FAR PASCAL PrintfProc( HWND hwnd, UINT message, UINT wParam, LONG lParam)
{
    switch( message )
      {
      case WM_DESTROY:
          printf_hwnd = (HWND)NULL;
          return(0);
          break;

      case WM_SYSCOMMAND:
          if ( wParam == SC_CLOSE )
             {
             MS_DestroyWindow( printf_hwnd );
             }
          else
             /* some other system command */
             return( MS_DefWindowProc (hwnd, message, wParam, lParam) );
          break;
      default:
          return( MS_DefWindowProc (hwnd, message, wParam, lParam) );
      }
}

void
CreateWinPrintfWindow(void)
{
    static class_registered = 0;
    WNDCLASS    wndclass;
    int border, width, height, menu_width;
    RECT rect;

    if ( printf_hwnd != (HWND)NULL )
        /* already created */
        return;

    memset( &wndclass, 0, sizeof(wndclass) );
    if ( !class_registered )
        {
         wndclass.style         = CS_HREDRAW | CS_VREDRAW;
         wndclass.lpfnWndProc   = PrintfProc;
         wndclass.cbClsExtra    = 0 ;
         wndclass.cbWndExtra    = 0 ;
         wndclass.hInstance     = hInst;
         wndclass.hIcon         = NULL ;
         wndclass.hbrBackground = MS_GetStockObject (WHITE_BRUSH);
         wndclass.lpszClassName = NXLT("ElfPrintfClass");

         if ( !MS_RegisterClass(&wndclass) )
             return;  /* error! */
         else
             class_registered = TRUE;
         }

     /* create the container simple window */
     printf_hwnd = MS_CreateWindow( NXLT("ElfPrintfClass"), 
               XLT("Applixware printf output"),
               (WS_OVERLAPPED | WS_CAPTION | WS_MINIMIZEBOX | 
                                   WS_SYSMENU | WS_MAXIMIZEBOX),
               CW_USEDEFAULT, CW_USEDEFAULT,
               DEFAULT_PRINTF_WIDTH, DEFAULT_PRINTF_HEIGHT,
               NULL, NULL, hInst, NULL);


    /* create a listbox which fills the container window */
    border = MS_GetSystemMetrics(SM_CXBORDER);
    menu_width = MS_GetSystemMetrics(SM_CYMENU);

    width = DEFAULT_PRINTF_WIDTH - ( 2 * border );
    height = DEFAULT_PRINTF_HEIGHT - menu_width - (2*border);
    listbox_hwnd = MS_CreateWindow ("LISTBOX", "",
                          (WS_CHILD | WS_VSCROLL | WS_HSCROLL ),
                          0,0,
                          width, height,
                          printf_hwnd, NULL, hInst, NULL) ;

    /* Before we actually show the window,                              */
    /* we must get the height of the listbox just created, and make */
    /* the container window a multiple of this height since listboxes are  */
    /* actually allocated in multiples of the lne height and we do not want */
    /* to have an odd white border at the bottom of the printf window      */
    MS_GetClientRect( listbox_hwnd, &rect );  /* returns window-relative */
    height = rect.bottom;

    /* now adjust the container window */
    MS_GetWindowRect( printf_hwnd, &rect );   /* returns screen relative */
    height = height +  menu_width + (2*border);
    MS_MoveWindow( printf_hwnd, rect.top, rect.left, DEFAULT_PRINTF_WIDTH, 
                                                       height, FALSE );

    MS_ShowWindow (printf_hwnd, SW_SHOW) ;
    MS_UpdateWindow (printf_hwnd) ;

    MS_ShowWindow (listbox_hwnd, SW_SHOW) ;
    MS_UpdateWindow (listbox_hwnd) ;

    num_list_entries = -1;
    listbox_scroll_width = 0;
    return;
}



void
WinPrintf(char *str)
{
    int pixel_str_width, width;
    char *p, *messP;
    char temp_str[500];


    if ( printf_hwnd == (HWND)NULL )
        CreateWinPrintfWindow();

    pixel_str_width = -1;
    strcpy( temp_str, str );
    messP = p = temp_str;

    while ( *p != '\0' )
       {
       if ( *p == '\n' )
          {
          *p = '\0';
          /* post the text to the end of the listbox */
          MS_SendMessage( listbox_hwnd, LB_ADDSTRING, (WPARAM)NULL, 
                                                            (LPARAM)messP );       
          messP = p+1;
          num_list_entries++;
          width = GetListBoxStrLen( listbox_hwnd, str );
          if ( width > pixel_str_width )
              pixel_str_width = width;
          }
       p++;
       }

    if ( *messP != '\0' )
       {
       MS_SendMessage( listbox_hwnd, LB_ADDSTRING, (WPARAM)NULL,(LPARAM)messP );
       num_list_entries++;
       width = GetListBoxStrLen( listbox_hwnd, str );
       if ( width > pixel_str_width )
           pixel_str_width = width;
       }

    /* make sure the string is displayed, by setting the index of the to */
    /* entry to what we just put in.  This will force the display in the */
    /* lsitbox if it has been scrolled                                   */
    MS_SendMessage( listbox_hwnd, LB_SETTOPINDEX, (WPARAM)num_list_entries, 
                                                                 (LPARAM)0);

    /* now see if we need to update the horizontal scroll extent because */
    /* the string we just added is longer than anything we have added before */
    if ( pixel_str_width > listbox_scroll_width )
        {
        /* reset the horizontal scroll extent */
        MS_SendMessage(listbox_hwnd, LB_SETHORIZONTALEXTENT,
                                               (WPARAM)pixel_str_width, 0L);
        listbox_scroll_width = pixel_str_width;
        }

    return;
}


void 
WinPrintfToForeground(void)
{
    /* force the printf window to the foreground.  Used in error cases */
    /* when we want the user to have this in his face. - jon           */
    if ( printf_hwnd == (HWND)NULL )
       return;

    MS_SetForegroundWindow( printf_hwnd );
    return;
}
