/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winproc.c
   Applix ASC version: 4
   Date: Thu Apr 22 21:16:53 1999
*/
#include <stdio.h>
#include <windows.h>
#include <winmap.h>


int
axmsStartProcessNopend(char *cmd_line, int no_hide, int return_pid)
{
    int ret;
    STARTUPINFO startup;
    PROCESS_INFORMATION process_info;

    memset(&startup, 0, sizeof(STARTUPINFO));
    memset(&process_info, 0, sizeof(PROCESS_INFORMATION));
    startup.cb = sizeof(STARTUPINFO);

    /* set the wierd Windows global so things get passed along */
#if !_DLL
    _fileinfo = TRUE;
#endif

    if ( !no_hide )
        {
        startup.dwFlags = STARTF_USESHOWWINDOW;
        startup.wShowWindow = SW_HIDE; 
        }

    ret = MS_CreateProcess(NULL, cmd_line,
                           NULL,     /* process security attributes */
                           NULL,     /* thread security attributes */
                           TRUE,     /* process inherits handles   */
                           (DWORD)0, /* creation flags */
                           (LPVOID)0, /* address of new environment block */
                           (LPCSTR)NULL, /* address of current directory */
                           &startup,
                           &process_info);
    if ( !ret )
        {
        /* error */
        return -1;
        }

    MS_CloseHandle(process_info.hThread);
    MS_CloseHandle(process_info.hProcess);
	if ( return_pid )
    	return (int)process_info.dwProcessId;
	else
		return 0;
}

int
axmsStartProcessAndWaitForEvent(char *cmd_line)
{
    int ret;
    STARTUPINFO startup;
    PROCESS_INFORMATION process_info;
	SECURITY_ATTRIBUTES sec_attrs;
	DWORD exit_code;
	HANDLE hEvent, handles[2];
	char *new_cmd;

    memset(&startup, 0, sizeof(STARTUPINFO));
    memset(&process_info, 0, sizeof(PROCESS_INFORMATION));
    startup.cb = sizeof(STARTUPINFO);
    startup.dwFlags = STARTF_USESHOWWINDOW;
    startup.wShowWindow = SW_HIDE;

    /* set the wierd Windows global so things get passed along */
#if !_DLL
    _fileinfo = TRUE;
#endif

	exit_code = -1;

	/* create an (inheritable) event handle on which to wait. the child process
	 * will receive the handle and will need to signal the event as appropriate
	 * so we can return the exit code.
	 */
	sec_attrs.nLength = sizeof(SECURITY_ATTRIBUTES);
	sec_attrs.lpSecurityDescriptor = NULL;
	sec_attrs.bInheritHandle = TRUE;
	hEvent = MS_CreateEvent(&sec_attrs, FALSE, FALSE, (LPCTSTR)NULL);
	if ( hEvent )
	{
		new_cmd = malloc(strlen(cmd_line) + 20);
		sprintf(new_cmd, "%s -event %d", cmd_line, (int)hEvent);

    	ret = MS_CreateProcess(NULL, new_cmd,
    	                       NULL,     /* process security attributes */
    	                       NULL,     /* thread security attributes */
    	                       TRUE,     /* process inherits handles   */
    	                       (DWORD)0, /* creation flags */
    	                       (LPVOID)0, /* address of new environment block */
    	                       (LPCSTR)NULL, /* address of current directory */
    	                       &startup,
    	                       &process_info);
    	if ( ret )
		{
			/*wait for either the event to be set or the process to terminate*/
			handles[0] = process_info.hProcess;
			handles[1] = hEvent;
			ret = WaitForMultipleObjects(2, handles, FALSE, INFINITE);
			if ( ret == WAIT_OBJECT_0 )  /* process terminated */
			{
				MS_GetExitCodeProcess(process_info.hProcess, &exit_code);
			}
			else if ( ret == WAIT_OBJECT_0 + 1 )  /* event signalled */
			{
				exit_code = 0;
			}

		    MS_CloseHandle(process_info.hThread);
		    MS_CloseHandle(process_info.hProcess);
		}

		MS_CloseHandle(hEvent);
		free(new_cmd);
	}

	return (int) exit_code;
}
