/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winsup.c
   Applix ASC version: 8
   Date: Tue Apr  6 22:45:20 1999
*/
#include <stdio.h>
#include <aos.h>
#include <tmi.h>
#include <tmassert.h>
#include <paratm.h>
#include <parint.h>
#include <getstr13.h>

		/* globals */

int BmapCount = 0;		/* # Bitmaps currently allocated */

extern int MapMode;
extern short ATMtrace;
extern short CursX;
extern short RastWid;
extern short RastHt;
extern short DpyDepth;
extern short SupOutput;

Layer *FrontLayer = NULL, *BackLayer = NULL;

extern A_Bitmap *A_Display;

short WinTrace = 0;
extern FILE *Tfp;				/* trace file pointer */
extern FILE *ATfp;

/*******************************************************************/
/*
 * refCmap -- return a pointer to the byte in a character cell map
 *		for coordinates (x, y)
 */
char *
refCmap(bp, x, y)
	A_Bitmap *bp;
	short x, y;
{
	char *cp;

	cp = bp->mapdata + bp->mapwid * (y - bp->maprect.origin.y)
			+ x - ( ( bp->maprect.origin.x >> 1) << 1 ) ;
	return cp;
}

/*******************************************************************/
/*
 * refCAmap -- return a pointer to the byte in a character attribute map
 *		for coordinates (x, y)
 */
char *
refCAmap(bp, x, y)
	A_Bitmap *bp;
	short x, y;
{
	char *cp;

	cp = bp->mapdata + (bp->imsiz >> 1)
			+ bp->mapwid * (y - bp->maprect.origin.y)
			+ x - ( ( bp->maprect.origin.x >> 1) << 1 ) ;
	return cp;
}

/*******************************************************************/
/*
 * ATMbitblt -- character cell (and initial RGMODE) version
 *
 * general notes:
 * - fast scroll causes exit from deferred output mode
 * - fast scroll calls ATMbitblt recursively with SupOutput set for clearing
 * - special handling of highlighting applies to R_CLEAR function only
 */

/*******************************************************************/

/*
 * rectf -- single-operand bit blitter
 *
 *		candidate for future optimization
 */
rectf(bp, rp, f)
	A_Bitmap *bp;
	AxRectangle *rp;
	int f;			/* function code */
{
	extern int halftoneMode;

	if(halftoneMode)
		ATMbitblt((A_Bitmap *)NULL,rp,bp,&rp->origin,f,halftoneMode,0);
	else
		ATMbitblt((A_Bitmap *)NULL, rp, bp, &rp->origin, f, 0, 0);
}
/*******************************************************************/
/*
 * erectf -- easy rectf -- takes 4 separate arguments instead of
 *				a pointer to a rectangle
 */
erectf(bp, ox, oy, cx, cy, f)
	A_Bitmap *bp;
	short ox, oy, cx, cy;
	int f;
{
AxRectangle rr;

rr.origin.x = ox;
rr.corner.x = cx;
rr.origin.y = oy;
rr.corner.y = cy;
rectf(bp, &rr, f);
}
/*******************************************************************/
/*
 * balloc -- allocate a bitmap
 *
 * In character cell mode (MapMode FALSE), each unit in device
 * coordinates (each cell) corresponds to a byte in the "bitmap".
 *
 * Bitmaps are aligned so that the origin in device
 * coordinates is on a word boundary. (The origin of the bitmap
 * rectangle is not necessarily word-aligned.)
 * This is done for ATMbitblt() performance, and in the case of
 * cell bitmaps, for consistency.
 *
 * By convention, if a bitmap is not associated with a particular
 * screen location, the origin of the rectangle is (0, 0).
 *
 * This routine interfaces with the image memory manager, which
 * determines where an image is allocated (esF).
 *
 * The first call to balloc() allocates the Display bitmap, which
 * refers to the hardware display buffer.
 *
 * Note:  In SUN graphics mode, mapdata points to a pixrect, instead
 * of an image, if the data is in memory (inMemF).
 */
A_Bitmap *
balloc(rp)
	AxRectangle *rp;
{
	char *ATMmalloc();
	static short isDisplay = TRUE;
	A_Bitmap *bmap;
	int ht, wid, rsh, i, discard;
	int bytecount;

#if 1
	if (ATMtrace)
	{
		fputs("\nentering balloc()\n", ATfp);
		ATMshoRect(ATfp, NXLT("  rp"), rp);
	}
#endif

	/* height in scanlines */

	ht = rp->corner.y - rp->origin.y;

	/* width in bytes, taking alignment into account */

	if (MapMode) rsh = 4; else rsh = 1;
	wid = (
			( (rp->corner.x - 1) >> rsh )
			+ 1
			- (rp->origin.x >> rsh)
		) << 1;
	if (MapMode)
	{
		bytecount = wid * ht * DpyDepth;
	}
	else
	{
		/* double size bitmap to store the attributes on a dumb terminal */
		bytecount = wid * ht * 2;
	}
	bmap = (A_Bitmap *) ATMmalloc(sizeof(A_Bitmap));

	/* set up A_Bitmap structure */

	bmap->maprect = *rp;
	bmap->mapwid = wid;
	bmap->newF = !isDisplay;	/* new image -- no content */
	bmap->depthF = TRUE;		/* same depth as display */
	bmap->wmF = TRUE;			/* a window manager resource */

	/* memory management extension */

	bmap->interF = isDisplay; /* only Display is interleaved */
	bmap->imsiz = bytecount;
	bmap->inMemF = isDisplay;	/* image in memory */
	bmap->mapLayer = NULL;	/* initialize to default */
	if (isDisplay)			/* display buffer */
	{
		bmap->esF = MapMode;
		if (MapMode)
		{
#if GMODE
			WMfbInit(bmap);		/* can alter mapwid */
#endif
		}
		else
		{
#if TMODE || RGMODE
		extern short ATMinvEmb;
		extern char *HLlist;
#endif
			/* permanently allocate text mode display buffer */

			bmap->mapdata = ATMmalloc(bytecount);
			for (i=0; i<bytecount/2; i++) bmap->mapdata[i] = ' ';

			/* set the attributes to null */
			for (i=bytecount/2; i<bytecount; i++) bmap->mapdata[i] = 0;

#if TMODE || RGMODE
		/* brought here from tmsup.c -- raj May 30, 86 */
			/* allocate buffer and Initialise hilighting list */
			if (ATMinvEmb)
			{
			register short j;
			char *Ptr;

				HLlist = Ptr = ATMmalloc(RastHt*RastWid);
				for(i=0;i<RastHt;++i)
					for(j=0;j<RastWid;++j)
						*(Ptr+i*RastWid+j) = 0;
			}
#endif
		}
	}
	else				/* ordinary bitmap */
	{
		bmap->mapdata = NULL;		/* for debugging */
	}
	if (!isDisplay) BmapCount++;	/* one more allocated */
	isDisplay = FALSE;				/* for subsequent calls */

#if TRACECOMP
	if (ATMtrace) fprintf(ATfp,
			NXLT("\nballoc returning bmap=%x\n"), bmap);
#endif
	return bmap;

} /* end balloc() */
/*******************************************************************/
#if GMODE  /* code assumes GMODE */
/*
 * mballoc -- malloc balloc
 *
 *		- allocate pixel image in regular malloc storage
 *		- bitmap cannot be used with getimage or bfree (mbfree instead)
 */
A_Bitmap *
mballoc(rp)
	AxRectangle *rp;
{
	char *ATMmalloc();
	A_Bitmap *bmap;
	int ht, wid, i;
	int bytecount;

#ifndef NDEBUG
	if (ATMtrace)
	{
		fputs("\nentering mballoc()\n", ATfp);
		ATMshoRect(ATfp, NXLT("  rp"), rp);
	}
#endif

	/* height in scanlines */

	ht = rp->corner.y - rp->origin.y;

	/* width in bytes, taking alignment into account */

	wid = (
			( (rp->corner.x - 1) >> 4 )
			+ 1
			- (rp->origin.x >> 4)
		) << 1;
	bytecount = wid * ht;			/* single plane maps only */
	bmap = (A_Bitmap *) ATMmalloc(sizeof(A_Bitmap));

	/* set up A_Bitmap structure */

	bmap->maprect = *rp;
	bmap->mapwid = wid;
	bmap->newF = TRUE;	/* new image -- no content */

	/* memory management extension */

	bmap->interF = FALSE; 	/* only Display is interleaved */
	bmap->imsiz = bytecount;
	bmap->inMemF = TRUE;	/* image in memory */
	bmap->imaddr = NULL;	/* indicate noncache bitmap */
	bmap->mapLayer = NULL;	/* not applicable */
	bmap->mapdata = ATMmalloc(bytecount);	/* pixel image memory */
#ifndef NDEBUG
	if (ATMtrace) fprintf(ATfp,
			NXLT("\nmballoc returning bmap=%x\n"), bmap);
#endif
	return bmap;

} /* mballoc */
#endif  /* #if GMODE */
/*******************************************************************/
/*
 * bfree - free a bitmap that was allocated by balloc() (except Display)
 */
bfree(bmap)
	A_Bitmap *bmap;
{
	ATMfree(bmap);					/* the A_Bitmap structure */
	BmapCount--;					/* one less currently allocated */
}
/*******************************************************************/
#if GMODE  /* goes with mballoc */
/*
 * mbfree - free a bitmap that was allocated by mballoc
 */
mbfree(bmap)
	A_Bitmap *bmap;
{
	ATMfree(bmap->mapdata);			/* pixel image memory */
	ATMfree(bmap);					/* the A_Bitmap structure */
}
#endif  /* #if GMODE */
/*******************************************************************/
/*
 * ccbase - compute coordinate system base address for a bitmap
 *
 *		Possible tradeoff: compute once and keep in Bitmap?
 *
 *		Note: not used and does not work in SUN MapMode
 */
char *
ccbase(bp)
	A_Bitmap *bp;
{
	if (bp == A_Display)
		return bp->mapdata;		/* express path */
	else
		return bp->mapdata - ((bp->maprect.origin.x >> 4) << 1)
				- bp->mapwid * bp->maprect.origin.y;
}
/*******************************************************************/
/*
 * bmapinit -- initialize bitmap management and Display Bitmap
 *
 *		formerly called wmnginit
 */
bmapinit()
{
	extern short DpyWid, DpyHt;
	AxRectangle dispRect;
	A_Bitmap *discard;


	/* build display bitmap, set global A_Display */

	dispRect.origin.x = dispRect.origin.y = 0;

#if TMODE || RGMODE
	dispRect.corner.x = RastWid;
	dispRect.corner.y = RastHt;
#else
	dispRect.corner.x = DpyWid;
	dispRect.corner.y = DpyHt;
#endif

	A_Display = balloc(&dispRect);	/* first call special case */
	WMresInit();		/* create window manager bitmap resources */
}  /* bmapinit */
/*******************************************************************/
/*
 * sectRect -- Calculate intersection of two rectangles
 *
 *	If rectangles do not intersect, return FALSE without
 *	modifying destination rectangle.
 *
 *	Note: If an argument has zero area, the rectangles may still
 *	intersect (overlap), even though the result will have zero area.
 *
 *	Last parameter points to destination, which can also be a
 *	source (to be clipped in place).
 */
int
sectRect(r1, r2, rr)
	AxRectangle *r1, *r2, *rr;
	
{
short maxOx, minCx;	/* limits in x dimension */
short maxOy, minCy;	/* limits in y dimension */


maxOx = r1->origin.x > r2->origin.x ? r1->origin.x : r2->origin.x;
minCx = r1->corner.x < r2->corner.x ? r1->corner.x : r2->corner.x;

maxOy = r1->origin.y > r2->origin.y ? r1->origin.y : r2->origin.y;
minCy = r1->corner.y < r2->corner.y ? r1->corner.y : r2->corner.y;

if ( minCx <= maxOx || minCy <= maxOy ) return(FALSE);

rr->origin.x = maxOx;
rr->corner.x = minCx;

rr->origin.y = maxOy;
rr->corner.y = minCy;


return(TRUE);

} /* end sectRect */

sectPolyWithRect(inPoly, nPtsIn, clipRectangle, outPoly, nPtsOut)
TmPoint32 *inPoly, *outPoly;
int nPtsIn, *nPtsOut;
AxRectangle *clipRectangle;
{
 int amInside;
 int i, stat;
 TmPoint32 intPt1, intPt2;

 *nPtsOut = 0;

 /* determine the state of the first point, inside or outside */
 if( amInside = pointInsideClip(inPoly[0], clipRectangle) )
 {
	/* if it's inside, the add it to the clipped polygon list */
	outPoly[(*nPtsOut)++] = inPoly[0];
 }

 /* for all the remaining vertices in the unclipped polygon... */
 for(i=1; i<nPtsIn; i++)
 {
	/* if the line segment crosses a clip edge... */
	if( lineCrossesClip(inPoly[i-1],inPoly[i],clipRectangle,TRUE) )
	{
		/* save the point of intersection... */
		intPt1 = inPoly[i-1];
		intPt2 = inPoly[i];
		stat = intersectPt(&inPoly[i-1], &inPoly[i], clipRectangle);

		if( stat == AOK )
		{
			/* then the second point will be clipped */
			if( amInside )
			{
				outPoly[(*nPtsOut)++] = inPoly[i];
				amInside = FALSE;
			}

			/* outside to inside, both may be clipped */
			else
			{
				/* save the point where it reenters the clip rect */
				outPoly[(*nPtsOut)++] = inPoly[i-1];

				/* does the second point clip? */
				if( inPoly[i].x == intPt2.x && inPoly[i].y == intPt2.y )
				{
					/* if not, save it */
					amInside = TRUE;
					outPoly[(*nPtsOut)++] = inPoly[i];
				}
	
				/* otherwise save the point where it exits */
				else
				{
					outPoly[(*nPtsOut)++] = inPoly[i];
				}
			}
		}
		inPoly[i-1] = intPt1;
		inPoly[i] = intPt2;
	}

	/* if this is another inside point, just add it to the list */
	else if ( amInside )
	{
		outPoly[(*nPtsOut)++] = inPoly[i];
	}
 }

 /* return TRUE if there are any unclipped segments left to display */
 if( *nPtsOut == 0 )
 {
	AxRectangle polyClip;
	TmPoint32 clipPoint;
	int i;

	/* if the clipped rect is inside the poly, return clip rect */
	clipPoint.x = clipRectangle->origin.x;
	clipPoint.y = clipRectangle->origin.y;
	polyClip.origin.x = polyClip.corner.x = inPoly[0].x;
	polyClip.origin.y = polyClip.corner.y = inPoly[0].y;
	for(i=1; i<nPtsIn; i++)
	{
		if(inPoly[i].x > polyClip.corner.x)
			polyClip.corner.x = inPoly[i].x;
		else if(inPoly[i].x < polyClip.origin.x)
			polyClip.origin.x = inPoly[i].x;

		if(inPoly[i].y > polyClip.corner.y)
			polyClip.corner.y = inPoly[i].y;
		else if(inPoly[i].y < polyClip.origin.y)
			polyClip.origin.y = inPoly[i].y;
	}
	if( pointInsideClip( clipPoint, &polyClip ) )
	{
		outPoly[0].x = clipRectangle->origin.x;
		outPoly[0].y = clipRectangle->origin.y;
		outPoly[1].x = clipRectangle->corner.x;
		outPoly[1].y = clipRectangle->origin.y;
		outPoly[2].x = clipRectangle->corner.x;
		outPoly[2].y = clipRectangle->corner.y;
		outPoly[3].x = clipRectangle->origin.x;
		outPoly[3].y = clipRectangle->corner.y;
		outPoly[4].x = clipRectangle->origin.x;
		outPoly[4].y = clipRectangle->origin.y;
		*nPtsOut = 5;
		return( TRUE );
	}

	/* else, no intersection at all */
	else
	{
		return( FALSE );
	}
 }
 else
 {

	/* return the closed polygon */
	return( TRUE );
 }
} /* end sectPolyWithRect */

int
pointInsideClip(point, clipRect)
TmPoint32 point;
AxRectangle *clipRect;
{
	if(( point.x >= clipRect->origin.x && point.x <= clipRect->corner.x ) &&
	   ( point.y >= clipRect->origin.y && point.y <= clipRect->corner.y ))
		return( TRUE );
	else
		return( FALSE );
}

typedef struct { int top, bottom, left, right; } quadrantMarker;

int
setQuadrantMarker(point, rect, marker)
TmPoint32 point;
AxRectangle *rect;
quadrantMarker *marker;
{
	/* set the left/right flags */
	if( point.x < rect->origin.x )
	{
		marker->left  = TRUE;
		marker->right = FALSE;
	}
	else if( point.x > rect->corner.x )
	{
		marker->left  = FALSE;
		marker->right = TRUE;
	}
	else
	{
		marker->left  = FALSE;
		marker->right = FALSE;
	}

	/* set the top/bottom flags */
	if( point.y < rect->origin.y )
	{
		marker->top    = TRUE;
		marker->bottom = FALSE;
	}
	else if( point.y > rect->corner.y )
	{
		marker->top    = FALSE;
		marker->bottom = TRUE;
	}
	else
	{
		marker->top    = FALSE;
		marker->bottom = FALSE;
	}
}

int
lineCrossesClip(point1, point2, clipRect, allCases)
TmPoint32 point1, point2;
AxRectangle *clipRect;
int allCases;
{
	quadrantMarker point1_pos, point2_pos;
	int crossFlag;

	/* determine the quadrant each point is located in */
	setQuadrantMarker(point1, clipRect, &point1_pos);
	setQuadrantMarker(point2, clipRect, &point2_pos);

	/* determine if the line will clip */
	if( point1_pos.bottom && point1_pos.bottom == point2_pos.bottom )
		/* line lies below clip rect */
		crossFlag = FALSE;
	else if( point1_pos.top && point1_pos.top == point2_pos.top )
		/* line lies above clip rect */
		crossFlag = FALSE;
	else if( point1_pos.left && point1_pos.left == point2_pos.left )
		/* line lies to the left of clip rect */
		crossFlag = FALSE;
	else if( point1_pos.right && point1_pos.right == point2_pos.right )
		/* line lies to the right of clip rect */
		crossFlag = FALSE;
	else if( allCases &&
	         point1_pos.top == point2_pos.top &&
	         point1_pos.bottom == point2_pos.bottom &&
	         point1_pos.left == point2_pos.left &&
	         point1_pos.right == point2_pos.right )
		/* line never crosses quadrant boundry */
		crossFlag = FALSE;
	else
		crossFlag = TRUE;
	return( crossFlag );
}

intersectPt(point1, point2, clipRect)
TmPoint32 *point1, *point2;
AxRectangle *clipRect;
{
 TmPoint32 *outsidePt, *insidePt;
 quadrantMarker p1Pos, p2Pos, *outsidePos, *insidePos; 
 float slope;
 int status, notDone;

 notDone = TRUE;
 while( notDone )
 {
	/* determine the quadrant each point lies within */
	setQuadrantMarker(*point1, clipRect, &p1Pos);
	setQuadrantMarker(*point2, clipRect, &p2Pos);
		
	/* if the points lie within the clip rect, we're done */
	if((!p1Pos.top && !p1Pos.bottom && !p1Pos.left && !p1Pos.right) &&
	   (!p2Pos.top && !p2Pos.bottom && !p2Pos.left && !p2Pos.right))
	{
		notDone = FALSE;
		status  = AOK;
	}

	/* find the intersection */
	else
	{
		if(lineCrossesClip(*point1,*point2,clipRect,0))
		{
			/* make sure that the first point is outside the rect */
			if( pointInsideClip(*point1,clipRect) )
			{
				outsidePt  = point2;
				outsidePos = &p2Pos;
				insidePt   = point1;
				insidePos  = &p1Pos;
			}
			else
			{
				outsidePt  = point1;
				outsidePos = &p1Pos;
				insidePt   = point2;
				insidePos  = &p2Pos;
			}

			/* check for a vertical line */
			if(	insidePt->x == outsidePt->x )
			{
				notDone = FALSE;
				if( outsidePos->left || outsidePos->right )
				{
					status = !AOK;
				}
				else
				{
					if( outsidePos->top )
						outsidePt->y = clipRect->origin.y;
					else if( outsidePos->bottom )
						outsidePt->y = clipRect->corner.y;
					if( insidePos->top )
						insidePt->y = clipRect->origin.y;
					else if( insidePos->bottom )
						insidePt->y = clipRect->corner.y;
					status = AOK;
				}
			}

			/* intersect 1 dimension at a time */
			else
			{
				slope = (float)(insidePt->y-outsidePt->y)/
				        (float)(insidePt->x-outsidePt->x);
				if( outsidePos->left )
				{
					outsidePt->y += (clipRect->origin.x-outsidePt->x)*slope;
					outsidePt->x = clipRect->origin.x;
				}
				else if( outsidePos->right )
				{
					outsidePt->y += (clipRect->corner.x-outsidePt->x)*slope;
					outsidePt->x = clipRect->corner.x;
				}
				else if( outsidePos->top )
				{
					outsidePt->x += (clipRect->origin.y-outsidePt->y) / slope;
					outsidePt->y = clipRect->origin.y;
				}
				else if( outsidePos->bottom )
				{
					outsidePt->x += (clipRect->corner.y - outsidePt->y) / slope;
					outsidePt->y = clipRect->corner.y;
				}
			}
		}

		/* if line doesn't cross clip boundries, no intersection */
		else
		{
			notDone = FALSE;
			status  = !AOK;
		}
	}
 }
 return( status );
}

/***** debugging support *****/


ATMshoRect(fp, s, r)
	FILE *fp;
	char *s;
	AxRectangle *r;
{
	fprintf(fp, NXLT("%s: ((%d %d) (%d %d))\n"), s,
		r->origin.x, r->origin.y, r->corner.x, r->corner.y);
	fflush(fp);
}

