/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xfntwrap.c
   Applix ASC version: 12
   Date: Fri Jan 22 21:12:20 1999
*/
#if AX_X11
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif
#include <aos.h>
#include <basicwin.h>
#include <axhtab.h>
#include <wchars.h>
#include <xim.h>
#include <axtext.h>
#include <axgfs.h>

extern Display *Dpy;

extern GC AxXCreateGC();
extern void axgfsDrawString();
extern void axgfsFreeFontStruct();

#if BUILD_ORIENT
extern int GetNarrowKanaFont();
extern FontSet FindFontSet();
extern Font FontSetFontId();
#endif

void axDrawString();
void axWDrawString();
void axDrawString16();
XFontStruct *AXLoadQueryFont();
int AXFreeFont();
int axSetFont();
int axChangeGC();
GC axCreateGC();

/*  define a data type which always represents a wide character since this */
/*  code is built whether the rest of the product is wide or narrow   */
#define always_wchar unsigned short

void
axDrawString(display, drawable, gc, font, x, y, in_str, length)
Display *display;
Drawable drawable;
GC  gc;
XFontStruct *font;
int x, y;
mbchar *in_str;
int length;
{
#if BUILD_WIDE
    /* this will be the only choice in the future */
    always_wchar temp_str[2000];
    int real_length;

    /* expand the string to wide characters */
    real_length = MbnToWStr(in_str, temp_str, length);

    /* now call the "wide" flavor to do the real work */
    axDrawString16(display, drawable, gc, font, x, y, temp_str, real_length);

#else
	if (font && IS_AXGFS_FONT(font)) {
		axgfsDrawString(display, drawable, gc, font, x, y, in_str, length);
	}
	else {
		/* for the rest of the world, do the normal thing for now */
		XDrawString(display, drawable, gc, x, y, in_str, length);
	}
#endif

    return;
}


/* This function should eventually go away.  It is here to allow an easy */
/* transiton between wide and narrow builds.  An application can call */
/* this function with a string of type "wchar" and it will work in both */
/* wide and narrow builds                                               */

void
axWDrawString(display, drawable, gc, font, x, y, in_str, length)
Display *display;
Drawable drawable;
GC  gc;
XFontStruct *font;
int x, y;
wchar *in_str;
int length;
{
#if BUILD_WIDE
    axDrawString16(display, drawable, gc, font, x, y, in_str, length);
#else
	if (font && IS_AXGFS_FONT(font)) {
		axgfsWDrawString(display, drawable, gc, font, x, y, in_str, length);
	}
	else {
		/* for the rest of the world, do the normal thing for now */
		XDrawString(display, drawable, gc, x, y, in_str, length);
	}
#endif
    return;
}

#if SUNIS || (LINUX && __i386__)        /*  {  */
IntelWrapChar2b(str2b, nchar)
XChar2b *str2b;
int nchar;
{
    mbchar  byte;
    int n;
 
    for (n = 0; n < nchar; n++) {
        byte = str2b[n].byte1;
        str2b[n].byte1 = str2b[n].byte2;
        str2b[n].byte2 = byte;
    }
}
 
IntelWrapTextItem16(items, nitem)
XTextItem16 *items;
int nitem;
{
    int segn, chn;
 
    for (segn = 0; segn < nitem; segn++)
        IntelWrapChar2b(items[segn].chars, items[segn].nchars);
}
#endif      /*  }  SUNIS || LINUX  */


#define DOING_ASCII 0
#define DOING_NARROW 1
#define DOING_KANJI 2

void
axDrawString16(display, drawable, gc, font, x, y, in_str, length)
Display *display;
Drawable drawable;
GC  gc;
XFontStruct *font;
int x, y;
always_wchar *in_str;
int length;
{
	static XChar2b buff[1024];
	XChar2b *tmpbuff;
	int ix;

#if !BUILD_ORIENT
	if (font && IS_AXGFS_FONT(font)) {
		/* Gallium font */
		axgfsWDrawString(display, drawable, gc, font, x, y, in_str, length);
	}
	else {
		/* X bitmapped font */
		/* no mixed text, just do the simple draw string */

		/* The function XDrawString16 wants an array of XChar2b,
			which is a structure consisting of a pair of unsigned
			chars.  Each wide character (integer) is broken into an 
			upper and lower half (byte).  On Big-endian platforms, 
			the integer that represents the wchar is already laid out 
			properly. On little-endian platforms, we have to swap the 
			bytes so the high-order part comes first in the structure. */
#ifdef BIGENDIAN
		XDrawString16(display, drawable, gc, x, y, (XChar2b *) in_str, length);
#else
		if (length >= 1024) {
			tmpbuff = (XChar2b *)malloc(length * sizeof(XChar2b));
		}
		else
			tmpbuff = &(buff[0]);
		for (ix=0;ix<length;++ix) {
			tmpbuff[ix].byte1 = (in_str[ix] >> 8) & 0xFF;
			tmpbuff[ix].byte2 = in_str[ix] & 0xFF;
		}
		XDrawString16(display, drawable, gc, x, y, tmpbuff, length);
		if (length >= 1024) {
			free( tmpbuff);
		}
#endif
	}
#else               /**         {  */
    XTextItem16 segments[732];  /* This limit is from MAX_CELLSTR(300)       */
                                /* Where every 3 bytes may become 2 segments */
                                /* in worst cases. (This is really happened  */
                                /* to our user!)                             */
    int next_seg;
    always_wchar *p;
    int i, j;
    int current_item;
    int doing_kanji;
    Font passed_gc_font;
    int found_kanji;
    int did_narrow_kana = FALSE;
    XGCValues gc_values;
    FontSet font_set;
    int junk;
    wchar blank_str[2];

    if ( in_str == NULL || *in_str == 0 || length == 0 )
         return;

    /* get the default font out of the passed graphics context */
    XGetGCValues(display, gc, GCFont, &gc_values);
    passed_gc_font = gc_values.font;

    font_set = FindFontSet( passed_gc_font, &junk, (XFontStruct *)NULL );
   
    /* go through the string looking for kanji characters, and if there */
    /* are any, break the string up into multi-font pieces              */
    p = in_str;
    doing_kanji = DOING_ASCII;
    found_kanji = FALSE;

    /* initialize the first segment */
    next_seg = 0;
    segments[next_seg].chars = (XChar2b *) in_str;    /* bad cast!! */
    segments[next_seg].nchars = 0;
    segments[next_seg].delta = 0;
    if (IS_WIDE_CHAR(*p))
        {
#if HP800 && !BUILD_KANJI
        /* HP DrawText16() call is broken.  If the first font drawn is wide */
        /* then wide spacing is used for all char sets, even the */
        /* narrow ones, so we fake it out by writing a narrow blank, and */
        /* specifying a negative offset so the blank gets overwritten */
        blank_str[0] = ' ';
        blank_str[1] = 0;
        segments[next_seg].chars = blank_str;
        segments[next_seg].nchars = 1;
        segments[next_seg].font = FontSetFontId( font_set, NARROW_FONT );
        next_seg++;

        segments[next_seg].chars = (XChar2b *) in_str;
        segments[next_seg].nchars = 0;
         /* back up width of blank*/
         segments[next_seg].delta = -XTextWidth(font_set->narrow_font, " ", 1);
#endif
        segments[next_seg].font = FontSetFontId( font_set, WIDE_FONT );
        doing_kanji = DOING_KANJI;
        found_kanji = TRUE;
        }
    else if (IS_NARROW_KANA(*p))
        {
        /* this is a "narrow" kana character, display from the narrow
         * font */
        segments[next_seg].font = FontSetFontId( font_set, NARROW_KANA_FONT );
        did_narrow_kana = TRUE;
        doing_kanji = DOING_NARROW;
        found_kanji = TRUE;
        }
    else
        {
        segments[next_seg].font = FontSetFontId( font_set, NARROW_FONT );
        doing_kanji = FALSE;
        }

    for (i = 0; i < length; i++)
        {
        if ((IS_WIDE_CHAR(*p) && (doing_kanji == DOING_KANJI)) ||
            (IS_NARROW_KANA(*p) && (doing_kanji == DOING_NARROW)) ||
            (!IS_WIDE_CHAR(*p) && !IS_NARROW_KANA(*p) && doing_kanji == DOING_ASCII))
            {
            /* we are continuing with the same type of string */
            segments[next_seg].nchars++;
            if (IS_NARROW_KANA(*p))
                {
#if BUILD_KANJI
                /* strip off the 0x8E in the high byte since
                 * this is really a character out of a small font */
                *p = *p & 0xFF;
#else
                /* Korean, just a wide blank to be treated as a narrow blank */
                *p = ' ';
#endif
                }
            }
        else
            {
            /* switch to the other type of segment */
            next_seg++;
            segments[next_seg].chars = (XChar2b *) (in_str + i);
            segments[next_seg].nchars = 1;
            segments[next_seg].delta = 0;
            if (IS_WIDE_CHAR(*p))
                {
                segments[next_seg].font =  FontSetFontId( font_set, WIDE_FONT );
                doing_kanji = DOING_KANJI;
                found_kanji = TRUE;
                }
            else if (IS_NARROW_KANA(*p))
                {
                segments[next_seg].font = 
                                    FontSetFontId( font_set, NARROW_KANA_FONT );
#if BUILD_KANJI
                *p = *p & 0xFF;
#else
                /* Korean, just a wide blank to be treated as a narrow blank */
                *p = ' ';
#endif
                did_narrow_kana = TRUE;
                doing_kanji = DOING_NARROW;
                found_kanji = TRUE;
                }
            else
                {
                segments[next_seg].font = FontSetFontId( font_set, NARROW_FONT);
                doing_kanji = DOING_ASCII;
                }
            }
        p++;
        }

end_of_story:

#if SUNIS || (LINUX && __i386__)
   IntelWrapTextItem16(segments, (next_seg + 1) );
#endif

    XDrawText16(display, drawable, gc, x, y,
            (XTextItem16 *) segments, (next_seg + 1));

#if SUNIS || (LINUX && __i386__)
   IntelWrapTextItem16(segments, (next_seg + 1));
#endif

    XSetFont(display, gc, passed_gc_font);  /* restore original gc font */

#if !BUILD_CHINA       /*  {  */
    if (did_narrow_kana)
        {
        /* we must put back the 0x8E bytes we clipped out for
         * display */
        for (i = 0; i <= next_seg; i++)
            {
            if (segments[i].font == 
                            FontSetFontId( font_set, NARROW_KANA_FONT) )
                {
                p = (always_wchar *) segments[i].chars;
                for (j = 0; j < segments[i].nchars; j++)
                    {
#if BUILD_KANJI
                    *p = 0x8E00 | *p;
#else
                    *p = KOREAN_WIDE_BLANK;
#endif
                    p++;
                    }
                }
            }
        }
#endif        /* }  !BUILD_CHINA */
#endif     /** } BUILD_ORIENT**/
}

#if !BUILD_ORIENT  /*  {   */

/* An interface to XLoadQueryFont where we cache some info about the */
/* font so that later we can get the info with just a font-id        */

XFontStruct *AXLoadQueryFont(Dpy, name)
Display *Dpy;
char *name;
{
    return( XLoadQueryFont(Dpy, name) );
}

#else    /*   BUILD_ORIENT} {  */


XFontStruct *AXLoadQueryFont(Dpy, name)
Display *Dpy;
char *name;
{
	XFontStruct *font;
	FontSet info;
	int font_type;
   
	if ( (font = XLoadQueryFont(Dpy, name)) == NULL) {
		return( (XFontStruct *)NULL );
	}
   
	/* create a fontset for the new font */
	info = FindFontSet( font->fid, &font_type, font );

    /* Set the fontset height here. Latter it may be enlarged by adding new
     * font to this font set in case of using GetNarrowFontFromSize(),
     * GetWideFontFromSize() and GetNarrowKanaFontFromSize().
     */
    info->height = font->ascent + font->descent;
 
    /* Set the fontset pointsize (in deci point unit) and fontset style (sans)
     * here so that to find more harmonic font couple or font triple latter.
     * The setting is done only once in fontset life cycle, and in here.
     */
    info->pointsize = axtDecipointSize(name);
    info->sans = IsSansXFontName(name);
 



#if SUN4S   /*  {  */
    /* the new Solaris 2.3 revisions of Japanese and Korean */
    /* return "broken" font metrics.  The ascent/descent values */
    /* do not match the max-cahr bounds numbers, and are real   */
    /* bogus.  As a saftey net, we here make sure that the ascent */
    /* and descent numbers match the max-bounds numbers         */
    if ( font->ascent > font->max_bounds.ascent )
       font->ascent = font->max_bounds.ascent;
    if ( font->descent > font->max_bounds.descent )
       font->descent = font->max_bounds.descent;
#endif    /*  } SUN4S */
    return( font );
}

#endif    /*  } BUILD_ORIENT    */

AXFreeFont(dpy, fs)
Display *dpy;
XFontStruct *fs;
{
#if AX_X11	
	if (fs && IS_AXGFS_FONT(fs)) {
		axgfsFreeFontStruct(fs);
		return;
	}
	else {
		XFreeFont(dpy, fs);
	}
#endif
}

axSetFont(dpy, gc, fs)
Display *dpy;
GC gc;
XFontStruct *fs;
{
	if (fs && IS_AXGFS_FONT(fs)) {
		return;
	}
	else {
		XSetFont(dpy, gc, fs->fid);
	}
}

axChangeGC(dpy, gc, fs, valuemask, values)
Display *dpy;
GC gc;
XFontStruct *fs;
unsigned long valuemask;
XGCValues *values;
{
	if (fs && IS_AXGFS_FONT(fs)) {
		if (valuemask & GCFont) {
			valuemask &= ~(GCFont);  /* protect server from fid */
		}
	}
	XChangeGC(dpy, gc, valuemask, values);
}

GC
axCreateGC(dpy, drw, screen_number, depth, fs, valuemask, values)
Display *dpy;
Drawable drw;
int screen_number;          /* screen# appropriate for gc */
int depth;                  /* drawable depth appropriate for gc */
XFontStruct *fs;
unsigned long valuemask;	/* which ones to set initially */
XGCValues *values;		/* the values themselves */
{
	if (fs && IS_AXGFS_FONT(fs)) {
		if (valuemask & GCFont) {
			valuemask &= ~(GCFont);  /* protect server from fid */
		}
	}
	return (AxXCreateGC(dpy, drw, screen_number, depth, valuemask, values));
}

