/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ximfuncs.c
   Applix ASC version: 70
   Date: Tue May 25 21:13:47 1999
*/
#if BUILD_ORIENT    /*  {  */

#include <basicwin.h>
#include <X11/Shell.h>
#include <X11/cursorfont.h>
#include <X11/keysym.h>

#include <aos.h>
#include <xim.h>
#include <xmlib.h>
#include <wchars.h>
#include <ximcomm.h>
#include <widefont.h>

#include <locale.h>
#include <axtime.h>
#include <aiconids.h> /* Just want to use the definition of "IC_MAIN" */

#if BUILD_KANJI && USE_XIM
#include <sys/utsname.h>      /* Ashisuto added to correspond HP-UX' spec of
                               * Japanese Input Method. 1998.03.16 */
#endif

int ActiveConversion;
XIC ICWhichHasFocus = (XIC)NULL;
Window WinOfICFocus = -1;
XIM input_method = (XIM)NULL;

extern font_name_struct wide_serif_names[];
extern font_name_struct wide_sansserif_bold_names[];
extern font_name_struct wide_sansserif_names[];

#ifndef KOREAN_WIDE_BLANK
#if OPENWIN
#define KOREAN_WIDE_BLANK 0x2121
#else
#define KOREAN_WIDE_BLANK 0xA1A1
#endif
#endif

char *getenv();

XFontStruct *AXLoadQueryFont();
AXWidget_main FindMainWidgetFromWindow();
AXWidget_main TheMainWidget();
win_list_t *FindWindow();
mbchar *get_default_orient_face();
char *get_default_dialog_xlfd();
XFontStruct *FontSetFontStruct();
int CallbackGlobalRetValue;     /* used to pass return status back from */
                                /* a callbak function */


extern Display *Dpy;
extern XFontStruct *AxDlgMainFont;      /* basic dialog font */
extern PIXEL_VALUE ForePixel;   /* foreground for control panels */
extern PIXEL_VALUE BackPixel;   /* backround for control panels */


XFontStruct *orient_dlg_font;
FontSet FontInfoFromIdList;
char *current_locale;


    
XIM AXOpenIM( display, database, resource_name, resource_class, status )
Display *display;
char *resource_name;
char *resource_class;
int *status;
{
    int ret;
    XIM xim_base = (XIM)NULL;
    char *lang_env;
#if BUILD_KANJI && LINUX_LIBC5
    extern char *_Xsetlocale();
#endif
#if BUILD_KANJI && USE_XIM
    struct utsname os;
#endif

    /*----------------------------------------------------------------*/
    /* make sure that the oriental dialogue font is loaded */
    /*----------------------------------------------------------------*/
    ret = LoadOrientalDialogueFont( Dpy );
    if ( ret != AOK )
       {
       *status = NO_KANJI_FONT;
       return( FALSE );
       }

#if USE_XIM
    lang_env = getenv("LANG");
    if ( (lang_env == (char *)NULL) || (strcmp(lang_env,"C") == 0) ||
	 (strcmp(lang_env,"") == 0) )
        {
        /* we are running the 'C' local, exit now because XOpenIM() will */
        /* succeed, but the input mechansim, in a 'C' local, eats various */
        /* western characters                                             */
        *status = NOT_LANG_LOCALE;
        return( (XIM)NULL );
        }
#endif

#if HYUNDAI && HP800
#elif BUILD_KANJI && LINUX_LIBC5
    if ( _Xsetlocale( LC_CTYPE, "") == NULL )
        {
        *status = INVALID_LOCALE;
        return( (XIM)NULL );
        }
#else
    if ( setlocale( LC_CTYPE, "") == NULL )
        {
        *status = INVALID_LOCALE;
        return( (XIM)NULL );
        }
#endif

 
#if BUILD_KANJI && USE_XIM
  /* Ashisuto extended HITACHI case to all platforms to correspond prural
   * Japanese Input Servers running on the same machine simoutenously,
   * except for HP-UX
   * 1996.11.12
   * 1998.03.16 absorbed HP-UX's speciality by judging uname.
   */

   /* Ashisuto added for JAPANESE build.  95.09.14 */
   uname(&os);
   if (!XSupportsLocale()) {
       if (strcmp(os.sysname, "HP-UX")) {
          *status = INVALID_LOCALE;
          return((XIM)NULL);
       }
   }
   if (XSetLocaleModifiers("") == NULL) {
       if (strcmp(os.sysname, "HP-UX")) {
          *status = NO_KANJI_CONVERTER;
          return((XIM)NULL);
	   }
   }
#endif


    xim_base = XOpenIM( Dpy, NULL, NULL, NULL);
    if ( xim_base == (XIM)NULL )
        {
        *status = NO_KANJI_CONVERTER;
        return( (XIM)NULL );
        }
    return( xim_base );
}


void AXCloseIM( )
{
#if !HP800
    /* on HP there is an input-method per-window, not a global one */
    if ( input_method != (XIM)NULL )
        {
        /* close the input mechanism */
        XCloseIM( input_method );
        }
#endif

    return;
}




XIC AXCreateIC( win, call_data, Pstart, Pdone, Pdraw, Pcaret, 
                                Sstart, Sdone, Sdraw )
Window win;
XPointer call_data;
XIMProc Pstart, Pdone, Pdraw, Pcaret, Sstart, Sdone, Sdraw;
{
    XIC ic;
    XIMCallback Pstart_c, Pdone_c, Pdraw_c, Pcaret_c;
    XIMCallback Sstart_c, Sdone_c, Sdraw_c;
    XVaNestedList  preedit_attr, status_attr;
    app_call_data *call_dataP;
    XRectangle s_rect;
    AXWidget_main main_w;
	AXWidget_kanji status_w;

    Pstart_c.client_data = call_data;
    Pstart_c.callback = Pstart;
    Pdone_c.client_data = call_data;
    Pdone_c.callback = Pdone;
    Pdraw_c.client_data = call_data;
    Pdraw_c.callback = Pdraw;
    Pcaret_c.client_data = call_data;
    Pcaret_c.callback = Pcaret;
    preedit_attr = XVaCreateNestedList( NULL, 
        XNPreeditStartCallback, &Pstart_c,
        XNPreeditDoneCallback, &Pdone_c,
        XNPreeditDrawCallback, &Pdraw_c,
        XNPreeditCaretCallback, &Pcaret_c,
        NULL );

    call_dataP = (app_call_data *)call_data;
    main_w = call_dataP->xim_widget;
    status_w = main_w->kanji_status;

#if HYUNDAI && (SUN4||SUN4S)   /*   {  */
    /*-----------------------------------------------------------------*/
    /* the X11R5 server (not Solaris) has many problems with the */
    /*  'StatusCallback' status style, so we will use a 'StatusArea' style */
    /*-----------------------------------------------------------------*/
    s_rect.x = 0;
    s_rect.y = 0;
    s_rect.width = 60;    /* force status disolay to left of status area */
    s_rect.height = status_w->height;

    status_attr = XVaCreateNestedList(NULL,
                      XNArea, &s_rect,
                      XNForeground, ForePixel,
                      XNBackground, BackPixel,
                      NULL);

    ic = XCreateIC( input_method, XNInputStyle,
                    (XIMPreeditCallbacks | XIMStatusArea ),
                    XNClientWindow, status_w->window,
                    XNFocusWindow, win,
                    XNPreeditAttributes, preedit_attr,
                    XNStatusAttributes, status_attr,
                    NULL );

#else          /* }  HYUNDAI X11R5   {  */

    Sstart_c.client_data = call_data;
    Sstart_c.callback = Sstart;
    Sdone_c.client_data = call_data;
    Sdone_c.callback = Sdone;
    Sdraw_c.client_data = call_data;
    Sdraw_c.callback = Sdraw;

    status_attr = XVaCreateNestedList( NULL, 
        XNStatusStartCallback, &Sstart_c,
        XNStatusDoneCallback, &Sdone_c,
        XNStatusDrawCallback, &Sdraw_c,
        NULL );

    ic = XCreateIC( input_method, XNInputStyle, 
                    (XIMPreeditCallbacks | XIMStatusCallbacks), 
                    XNClientWindow, win,
                    XNPreeditAttributes, preedit_attr,
                    XNStatusAttributes, status_attr,
                    NULL );
#endif   /*  }  HYUNDAI X11R5  */

    XFree( preedit_attr );
    XFree( status_attr );

    if ( ic == (XIC)NULL )
       {
       printf("Unable to create Input Context (IC)\n");
       return( (XIC)NULL );
       }

    return( ic );
}



/* process an event with regard to the input mechanism */
/* returns TRUE => event was handled here, caller should disregard */
/*        FALSE ==> caller should handle this event */
AXFilterEvent( eventP )
XEvent *eventP;
{
    Window win;
    static Window old_window_value;
    AXWidget_main main_w;
    XIC ic;
    int ret;
    int i;
    app_call_data *callback_data;
    Window client_window;
    Window focus_window;
    XFontSet fs;
    int reset_packet;

    if (input_method == NULL)
         return(FALSE);

    win = eventP->xany.window;
    main_w = FindMainWidgetFromWindow( win );

    /* Ashisuto also ignores the IC_MAIN case because from version 4.3, the
     * IC_MAIN begins to accept KeyPress events, since the InputContext is
     * not available for IC_MAIN, don't filter out any event to
     * InputMethod/InputContext.
     * 1996.11.13
     */
    if ( main_w == NULL || main_w->style == IC_MAIN )
        {
        return( FALSE );
        }
    else
        {
        /* a widget window */
        ic = main_w->ic;
        callback_data = (app_call_data *)main_w->callback_data;
        }


    ActiveConversion = FALSE;
    if ( (callback_data != (app_call_data *)NULL) && (ic != (XIC)NULL)  )
        {
        callback_data->event_win = win;
        if ( callback_data->preedit_buf[0] != (wchar)0 )
           {
           /* we are in the middle of doing a conversion, we ignore certain */
           /* event types to keep the user from moving the focus            */
           ActiveConversion = TRUE;
           if ( eventP->type == ButtonPress )
              {
              RingTheBell();
              return( TRUE );
              }
		  
           if ( eventP->type == ButtonRelease )
              return( TRUE );
           }
        }

if ( eventP->type == KeyPress )
    {
  /** a handy debug breakpoint for keypress events */
    i = i + 1;
    }

    if ( (ic != ICWhichHasFocus) && (ICWhichHasFocus == (XIC)NULL) )
        {
        /* The ICWhichHasFocus is null, which probably results from */
        /* a focus out of a window going away.  We do have an input */
        /* context as part of this widget, so install it */
        ICWhichHasFocus = ic;
        /* Avoid duplicate setting of IC focus. */
        if ( main_w->window == WinOfICFocus )
            XSetICFocus( ICWhichHasFocus );
        }

    if ((ic != (XIC)NULL) && (ic == ICWhichHasFocus) && ( main_w->window != 0))
        {
        /* we normalize all input to the main window */
        if ( main_w->window != WinOfICFocus )
           {
           /* the ic is the current ic, but the curent window of focus seems */
           /* to have gotten lost.  Set it to the current window */
           /* change the input window on a "significant" event */
           WinOfICFocus =  main_w->window;
           XSetICValues( ic, XNFocusWindow, WinOfICFocus, NULL );
           XSetICFocus(ic);
           }

#if 0
        /* debug code */
        XGetICValues( ic, XNClientWindow, &client_window,
                      XNFocusWindow, &focus_window,
                      XNFontSet, &fs, NULL );
#endif

        /*------------------------------------------------------------------*/
        /* Note: we pass the window of the W_MAIN widget which normalizes */
        /*  all input activity for keystrokes anywhere in this dialog or */
        /*  application to a single window ID.  IF we do not do this the */
        /*  layering of various windows means that we get Focus events on */
        /*  one window (say W_MAIN) but the window ID returned with a */
        /*  keystroke is the window ID of the W_DIALOG so XFilterEvent below */
        /*  disreards events it should be processing because the */
        /*  window ID from the keystroke is not the last one set in */
        /*  SetXIMFocus() via  the focus event    */
        /*------------------------------------------------------------------*/
#if HYUNDAI
        reset_packet = FALSE;
        if ( (eventP->type == KeyPress) && (eventP->xkey.keycode == 0) &&
             (eventP->xkey.x == 0) && (eventP->xkey.y == 0) )
           reset_packet = TRUE;
        else if ( eventP->type == KeyPress )
           /*  see big explination below */
           old_window_value = eventP->xkey.window;
#endif

        ret = XFilterEvent( eventP, main_w->window );

#if HYUNDAI
           /* Discovery:    December 21, 1994  by Jon Dale                  */
           /* The Korean X11R5 input server seems to operate in a strange   */
           /* way when conversion mode is OFF.  When a keypress event is    */
           /* passed to XFilterEvent() the input server returns TRUE showing */
           /* that it processed the event (even though input is in English!)  */
           /* and also changes the 'window' field in the input 'eventP'       */
           /* structure to the parent window of the dialog/application window. */
           /* Since the return value is TRUE from XFilterEvent() Applixware   */
           /* will ignore the keypress event.                                 */
           /* As part of the input server processing of the English KeyPress  */
           /* event, the input server generates another KeyPress event, which */
           /* has only a couple of fields in the event packet initialized.    */
           /* The fields which are initialized are:                           */
           /*    type, serial, send_event, display, window                    */
           /* The 'time' field seems to be set to a hard coded (and invalid) */
           /*   value of "101".                                               */
           /* The 'window' field is set to the "main" window-id used to       */
           /*   overwrite the real window-id on the first 'real' keypress event */
           /* When this second "artificial" KeyPress event is passed to       */
           /* XFilterEvent() the input server sets the 'keycode' field to the */
           /* value from the first event, and returns FALSE so that the       */
           /* application will proces the event. This second event is the event */
           /* which Applixware applications arde processing when doing English */
           /* typing.
           /* Also note: there is not following KeyRelease event for the second */
           /*   artificial KeyPress event.  This I think leads to some problems */
           /*   in the widgets where a button will go down (the KeyPress) and   */
           /*   not come back up (should happen on KeyRelese)                   */
           /*                                                                  */
           /* Since the input server destroys the original window-id on the    */
           /* original KeyPress event, English typing is never passed to       */
           /* widgets which are processed specially, and are addressed by the  */
           /* window-id in the event packet.  see tests for W_TABLE, W_TRAY,   */
           /* W_CANVAS, W_LIST under KeyPress case in ZXtDispatchEvent.        */
           /* If Korean input mode is ON, the window-id is not changed!        */
           /*                                                                  */
           /* The real fix should be in the input server which should ignore   */
           /* typing events if input mode is OFF. The input server should:     */
           /*    1) never change anything in the input caller's packet         */
           /*    2) return FALSE so the application can process the            */
           /*       English KeyPress event in a normal mode.                   */
           /*    3) should not need to generate an "artificial" KeyPress event */
           /*       which has no matching KeyRelease                           */
           /*                                                                  */
           /* For now, we will put in a hack to detect this situation, and     */
           /*   do the following:                                              */
           /*  1) Remember the window-id from the "normal" keypress events     */
           /*  2) Detect the "artifical" KeyPress events generated by the      */
           /*     input server, and reset the window-id to the value from the  */
           /*     previous "real" KeyPress event                               */
        if ( reset_packet )
           eventP->xkey.window = old_window_value;
#endif
        return( ret );
        }
    else
        return( FALSE );    /* application should process this event */

}

    





/** find the width of a set of text characters.  The input is:
**   wstr - the text as wid characters
**   char_len - number of characters, not bytes
**   font_id  - the Font to be used as the base font 
**                    - or - 
**   gc       - a graphics context, out of which to get the font if the 
**              font_id == 0 , otherwise this can be NULL
**/

AXTextExtents16( font_id, wstr, char_len, gc, ascent, descent )
Font font_id;
wchar *wstr;
int char_len;
GC gc;
int *ascent, *descent;
{
    FontSet font_set;
    XGCValues gc_values;
    int junk;
    int result = 0;

    if ( font_id == 0 )
       {
       XGetGCValues( Dpy, gc, GCFont, &gc_values);
       font_id = gc_values.font;
       }

    font_set =  FindFontSet( font_id, &junk, NULL );

    result = AXwcTextEscapement( font_set, wstr, char_len, ascent, descent );

    return( result );
}




/** The followig is a temp replacement for the X11 R5 XwcTextEscapement()
**  call. It will measure the characters in a string given a font set */
         
AXwcTextEscapement( font_set, wstr, num_chars, ascent, descent )
FontSet font_set;
wchar *wstr;
int num_chars;
int *ascent;
int *descent;
{
    register XFontStruct *font;
    register int seg_len;
    register wchar *segP;
    register int i, j, char_type, next_char_type;
    register int result = 0;

    *ascent = *descent = 0;
    if ( num_chars == 0 )
       return( 0 );

    segP = wstr;
    seg_len = 0;

    if ( IS_WIDE_CHAR( *wstr ) )
       char_type = WIDE_FONT;
    else if ( IS_NARROW_KANA( *wstr ) )
       char_type = NARROW_KANA_FONT;
    else 
       char_type = NARROW_FONT;
    
    for ( i=0; i<=num_chars; i++ )
       {
       if ( i == num_chars )
          {
          /* we are one beyond the end of the string, force a char change */
          /* to get the last segment measured */
          next_char_type = -1;
          }
       else if ( IS_WIDE_CHAR( *wstr ) )
          next_char_type = WIDE_FONT;
#if BUILD_KANJI || HYUNDAI
       else if ( IS_NARROW_KANA( *wstr ) )
          {
          next_char_type = NARROW_KANA_FONT;
#if  BUILD_KANJI
          /* The narrow kana fonts are encoded internally with a "0x8E" */
          /*  in the first byte, but the font is a simple 256 char font */
          /*  therefore, we need to get rid of the 0x8E, and putit back */
          /*  in after we have measured the string                      */
          *wstr = *wstr & 0xFF;
#endif
#if BUILD_KOREAN
          /* We treat wide-blanks specially to yield narrow blanks when */
          /* displayed */
          *wstr = ' ';
#endif
          }
#endif
       else
          next_char_type = NARROW_FONT;

       if ( char_type != next_char_type )
          {
          /* character type changes, measure what we have */
          font = FontSetFontStruct( font_set, char_type );
          if ( char_type == NARROW_FONT )
              result = result + AXWTextWidthWestern( font, segP, seg_len );
          else
              {
              /* a wide font */
#if SUNIS || (LINUX && __i386__)
              IntelWrapChar2b(segP, seg_len);
#endif
              result = result + XTextWidth16( font, segP, seg_len );
#if SUNIS || (LINUX && __i386__)
              IntelWrapChar2b(segP, seg_len);
#endif
              }
          if ( font->ascent > *ascent )
             *ascent = font->ascent;
          if ( font->descent > *descent )
             *descent = font->descent;
#if BUILD_KANJI || HYUNDAI
          if ( char_type == NARROW_KANA_FONT )
             {
             /* Put back the 0x8E we stripped out above */
             for ( j=0; j<seg_len; j++ )
                {
#if BUILD_KANJI
                *segP = *segP | 0x8E00;
#endif
#if HYUNDAI
                /* korean */
                *segP = KOREAN_WIDE_BLANK;     /* a wide-blank */
#endif
                segP++;
                }
             }
#endif
          segP = wstr;
          seg_len = 1;
          char_type = next_char_type;
          }
       else
          { 
          /* a continuting sequece of the same type of characters */
          seg_len++;
          }
       wstr++;
       }
    return( result );
}



LoadOrientalDialogueFont(display)
    Display *display;
{
    static int kanji_inited = -1;
    FontSet dlg_set;
    int type;
    char *default_face;
    extern int  AxAlternateDialogFont;
    extern int  AxCharAbove, AxCharHeight;
    int AxCharBelow;


    if ( kanji_inited != -1 )
        /* we have been here before */
        return( kanji_inited );

    /* start by looking for a user specified choice */
    default_face = get_default_dialog_xlfd();
    

#if BUILD_KANJI              /*  {   */
 	if (default_face == NULL) {
 		int point = 14, weight = FALSE, slant, sans;
 		char *family;
 		if (AxAlternateDialogFont) {
 			if (getXFontAttr(AxDlgMainFont, NULL, &point, &weight, &slant, &sans)) {
 				if (sans)
 					family = sansserif_orient_face();
 				else
 					family = serif_orient_face();
 			} else {
 				family = get_default_orient_face();
 				weight = 0;
 				point = 14;
 			}
 		} else { /* using axsab which do not exist in font cach */
 			family = sansserif_orient_face();
#if AT_APPLIX
            weight = 0;  /* I do not have the bold sans-serif kanji font-jon*/
#else
 			weight = 1;
#endif
 			point = 14;
 		}
 		if (!getXFontStruct(family, point, weight, 0, &type, &orient_dlg_font)){
 			/* use the hard coded face */
 			if (AxAlternateDialogFont)
 				default_face = wide_serif_names[2].font_name;
 			else
 				default_face = wide_sansserif_bold_names[2].font_name;
 			orient_dlg_font = AXLoadQueryFont(display, default_face);
 			if (orient_dlg_font == (XFontStruct *) NULL)
 				return (!AOK);
 		}
 	} else {
 		orient_dlg_font = AXLoadQueryFont(display, default_face);
 		if (orient_dlg_font == (XFontStruct *) NULL)
 			return (!AOK);
 	}
    /* put the selected font into the font set cache structure */
    dlg_set = FindFontSet( AxDlgMainFont->fid, &type, (XFontStruct *)NULL );
    dlg_set->wide_font_id = orient_dlg_font->fid;
    dlg_set->wide_font = orient_dlg_font;

#else                  /*  }  BUILD_KANJI  {  */

    if ( default_face == (char *)NULL )
        {
        /* no user specified value                                           */
        /* Use the standard function to find a font to match the dialog font */
        dlg_set = FindFontSet( AxDlgMainFont->fid, &type, (XFontStruct *)NULL );
        orient_dlg_font = FontSetFontStruct( dlg_set, WIDE_FONT );
        }
    else
        {
        /* use the user specified value */
        orient_dlg_font = AXLoadQueryFont(display, default_face );
        }
        
    if (orient_dlg_font == (XFontStruct *) NULL)
        return (!AOK);

    /* put the selected font into the font set cache structure */
    dlg_set = FindFontSet( AxDlgMainFont->fid, &type, (XFontStruct *)NULL );
    dlg_set->wide_font_id = orient_dlg_font->fid;
    dlg_set->wide_font = orient_dlg_font;
#endif               /*  }  else BUILD_KAJI  */

#if 0
    AxCharAbove = MAX(dlg_set->narrow_font->ascent, orient_dlg_font->ascent);
    AxCharBelow = MAX(dlg_set->narrow_font->descent, orient_dlg_font->descent);
#else
    AxCharAbove =  dlg_set->narrow_font->ascent;
    AxCharBelow =  dlg_set->narrow_font->descent;
#endif
    dlg_set->height = AxCharHeight = AxCharAbove + AxCharBelow;

    kanji_inited = 1;  /* all is OK */

    return (AOK);
}



/***************************************************************************/
/* Here we set the focus of the X Input Method so that it follows the */
/* input focus The real difference is that we normalize all activity to the */
/* window of the "W_MAIN" widget.  You can find the corresponding other half */
/* of this is "AXFilterEvent()" and most particularly the window id passed */
/* to the call to XFilterEvent()                                           */
/***************************************************************************/

SetXIMFocus( wp, xwin )
widget_t *wp;  /* widget of focus, or NULL and we will find widget from xwin */
Window xwin;   /* X window of focus */
{
    AXWidget_main main_w;

#if !HP800
    if ( input_method == (XIM)NULL )
       return;
#endif

    if ( wp == (widget_t *)NULL )
       {
       /* get the widget from the X window id */
       main_w = FindMainWidgetFromWindow( xwin );
       }
    else
        main_w = TheMainWidget( wp );
 
    if ( main_w == NULL )
        return(0);
 
    if ( (main_w->ic != (XIC)NULL) && (main_w->ic != ICWhichHasFocus) )
        {
        if ( ICWhichHasFocus != (XIC)NULL )
            XUnsetICFocus( ICWhichHasFocus );
        ICWhichHasFocus = main_w->ic;
        /* Avoid duplicate setting of IC focus */
        if ( ICWhichHasFocus != (XIC)NULL && main_w->window == WinOfICFocus )
            XSetICFocus( ICWhichHasFocus );
        }
    else if ( main_w->ic == (XIC)NULL )
        ICWhichHasFocus = (XIC)NULL;
 
    if ( (ICWhichHasFocus != (XIC)NULL) && (main_w->window != WinOfICFocus) &&
         (main_w->window != 0) )
       {
       /* change the input window on a "significant" event */
       WinOfICFocus = main_w->window;
       XSetICValues( ICWhichHasFocus, XNFocusWindow, WinOfICFocus, NULL );
       XSetICFocus( ICWhichHasFocus );
       }
 
    return;
}


UnSetXIMFocus( wp, xwin )
widget_t *wp;
Window xwin;
{
#if !HP800
    if ( input_method == (XIM)NULL )
       return;
#endif

    if ( ICWhichHasFocus != (XIC)NULL )
       {
       XUnsetICFocus( ICWhichHasFocus );
       ICWhichHasFocus = (XIC)NULL;
       WinOfICFocus = (Window)NULL;
       }
    return;
}

#endif /* }  BUILD_ORIENT */
