/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xinicurs.c
   Applix ASC version: 28
   Date: Mon Nov 16 13:53:41 1998
*/
#include <aos.h>
#include <eldata.h>
#include <basicwin.h>
#include <afile.h>

#if AX_X11    /* { */
#include <X11/cursorfont.h>
#if ALPHA
#include <ax_xm.h>
#endif
#if AUX
#include <X11/fd.h>
#endif
#else     /* } AXMS  { */
#include <x2wincur.h>
#endif    /*  } AX_X11 */

#include <aiconids.h>
#include <getstr4.h>
#include <getstr5.h>
#include <getstr7.h>

typedef struct ax_cursor {
    char *name;
    Cursor data;
} ax_cursor;
typedef struct udcursor {
    int count;
    struct ax_cursor *cursor;
} udcursor;
typedef struct oldudcursor {
    int old_count;
    struct ax_cursor *olddata;
} oldudcursor;

static char file_str[STRPATH+1];
extern char *DirSlash;
extern char *ThimUserElfDir;
extern char *ThimCustomDir;
extern char *ThimApplixDir;
extern long WinMask;
extern data_ptr ElfRetData;
extern Cursor ThimTextCursor;
extern Display *Dpy;
Cursor ThimHelpCursor;
Cursor ThimPointerCursor = None;
Cursor ThimBusyCursor = None;
Cursor ThimRecordCursor = None;
Cursor ThimNoRecordCursor = None;
Cursor ThimFleurCursor = None;
Cursor ThimDownCursor = None;
Cursor ThimLeftRightCursor = None;
Cursor ThimUpDownCursor = None;
Cursor ThimTopLeftCursor = None;
Cursor ThimTopRightCursor = None;
Cursor ThimBottomLeftCursor = None;
Cursor ThimBottomRightCursor = None;
Cursor ThimRotateCursor = None;
Cursor ThimCrosshairCursor = None;
Cursor ThimWPDefaultCursor = None;
Cursor ThimMEDefaultCursor = None;
Cursor ThimEQNDefaultCursor = None;
Cursor ThimZoomInCursor = None;
Cursor ThimZoomOutCursor = None;
Cursor ThimAddCursor = None;
Cursor ThimSelectCursor = None;
Cursor ThimDeleteCursor = None;
Cursor ThimNoopCursor = None;
Cursor ThimBoltCursor = None;
Cursor ThimHyperlinkCursor = None;
Cursor ThimChartCursor = None;
Cursor ThimRightMouseCursor = None;

struct udcursor cursors;
struct udcursor old_cursors;

#if AX_X11
extern XSetWindowAttributes WinAtt;
#endif

#if AXMS
extern HANDLE hInst;
#endif  


void InitCursors()
{
    /* basic cursor init */
    readCustomCursors();
    ThimTextCursor = XCreateFontCursor(Dpy,XC_xterm);
#if AXMS
    ThimHelpCursor = MS_LoadCursor(hInst, "question_arrow");
#else
    ThimHelpCursor = XCreateFontCursor(Dpy,XC_question_arrow);
#endif

    /* customizable system cursors */
    ZXmBindCursor(UPbyCode(PROF_CURS_POINTER_),&ThimPointerCursor);
    ZXmBindCursor(UPbyCode(PROF_CURS_BUSY_),&ThimBusyCursor);
    ZXmBindCursor(UPbyCode(PROF_CURS_ELFREC_),&ThimRecordCursor);
    ZXmBindCursor(UPbyCode(PROF_CURS_NOELFREC_),&ThimNoRecordCursor);
    ZXmBindCursor(UPbyCode(PROF_WP_DEFAULT_CURSOR), &ThimWPDefaultCursor);
    ZXmBindCursor(UPbyCode(PROF_ME_DEFAULT_CURSOR), &ThimMEDefaultCursor);
    ZXmBindCursor(UPbyCode(PROF_EQN_DEFAULT_CURSOR), &ThimEQNDefaultCursor);
 
    /* non-customizable system cursors */
    ZXmBindCursor(NXLT("fleur"), &ThimFleurCursor);
    ZXmBindCursor(NXLT("sb down arrow"), &ThimDownCursor);
    ZXmBindCursor(NXLT("sb h double arrow"), &ThimLeftRightCursor);
    ZXmBindCursor(NXLT("sb v double arrow"), &ThimUpDownCursor);
    ZXmBindCursor(NXLT("top left corner"), &ThimTopLeftCursor);
    ZXmBindCursor(NXLT("top right corner"), &ThimTopRightCursor);
    ZXmBindCursor(NXLT("bottom left corner"), &ThimBottomLeftCursor);
    ZXmBindCursor(NXLT("bottom right corner"), &ThimBottomRightCursor);
    ZXmBindCursor(NXLT("exchange"), &ThimRotateCursor);
    ZXmBindCursor(NXLT("crosshair"), &ThimCrosshairCursor);
    ZXmBindCursor(NXLT("hand2"), &ThimHyperlinkCursor);
    ZXmCreateCursor(NXLT("chrt_dat.im"), NXLT("chrt_msk.im"),  5,  5,
                    &ThimChartCursor);
    ZXmCreateCursor(NXLT("zm_i_dat.im"), NXLT("zm_i_msk.im"),  7,  7,
                    &ThimZoomInCursor);
    ZXmCreateCursor(NXLT("zm_o_dat.im"), NXLT("zm_o_msk.im"),  7,  7,
                    &ThimZoomOutCursor);
    ZXmCreateCursor(NXLT("selt_dat.im"), NXLT("selt_msk.im"),  2,  3,
                    &ThimSelectCursor);
    ZXmCreateCursor(NXLT("add_dat.im"),  NXLT("add_msk.im"),   2,  3,
                    &ThimAddCursor);
    ZXmCreateCursor(NXLT("del_dat.im"),  NXLT("del_msk.im"),   2,  3,
                    &ThimDeleteCursor);
    ZXmCreateCursor(NXLT("noop_dat.im"), NXLT("noop_msk.im"), 10, 10,
                    &ThimNoopCursor);
    ZXmCreateCursor(NXLT("bolt_dat.im"), NXLT("bolt_msk.im"),  2,  3,
                    &ThimBoltCursor);
    ZXmCreateCursor(NXLT("rtms_dat.im"), NXLT("rtms_msk.im"), 10, 10,
                    &ThimRightMouseCursor);

    /* default application cursors */
    ElfbMapCursor(IC_DC, UPbyCode(PROF_CURS_WP_));
    ElfbMapCursor(IC_GE, UPbyCode(PROF_CURS_GE_));
    ElfbMapCursor(IC_SS, UPbyCode(PROF_CURS_SS_));
    ElfbMapCursor(IC_CE, UPbyCode(PROF_CURS_CE_));
    ElfbMapCursor(IC_EQN,UPbyCode(PROF_CURS_EQN_));
 
    /* sane values */
    if ( ThimPointerCursor == (Cursor)None)
        ZXmBindCursor(NXLT("left ptr"), &ThimPointerCursor);
    if ( ThimBusyCursor == (Cursor)None)
        ZXmBindCursor(NXLT("left ptr"), &ThimBusyCursor);
    if ( ThimRecordCursor == (Cursor)None)
        ZXmBindCursor(NXLT("left ptr"), &ThimRecordCursor);
    if ( ThimNoRecordCursor == (Cursor)None)
        ZXmBindCursor(NXLT("left ptr"), &ThimNoRecordCursor);
    if ( ThimWPDefaultCursor == (Cursor)None)
        ZXmBindCursor(NXLT("left ptr"), &ThimWPDefaultCursor);
    if ( ThimMEDefaultCursor == (Cursor)None)
        ZXmBindCursor(NXLT("left ptr"), &ThimMEDefaultCursor);
    if ( ThimEQNDefaultCursor == (Cursor)None)
        ZXmBindCursor(NXLT("left ptr"), &ThimEQNDefaultCursor);

    WinMask |= CWCursor;
#if AX_X11
    WinAtt.cursor = ThimPointerCursor;
#endif /* AX_X11 */
}

char *
ElfbReadCustomFilename()
{
    int found;
    char *file;
    
    found = FALSE;
    file = NXLT("cursor.map");

    /* check in macros dir */
    strcpy(file_str, ThimUserElfDir);
    strcat(file_str, DirSlash);
    strcat(file_str, file);
    if ( Aaccess(file_str, F_OK) == 0 )
        found = TRUE;

    /* check in custom_dir dir */
    if ( found == FALSE )
    {
        strcpy(file_str, ThimCustomDir);
        strcat(file_str, DirSlash);
        strcat(file_str, file);
        if ( Aaccess(file_str, F_OK) == 0 )
            found = TRUE;
    }

    /* check in system dir */
    if ( found == FALSE )
    {
        strcpy(file_str, ThimApplixDir);
        strcat(file_str, DirSlash);
        strcat(file_str, file);
        if ( Aaccess(file_str, F_OK) == 0 )
            found = TRUE;
    }
    if ( found )
        return(file_str);
    else
        return((char *)0);
}

check_name_in_use(cursor_name, cursor)
char *cursor_name;
Cursor cursor;
{
    char *name;

    name = (char *)UPbyCode(PROF_CURS_POINTER_);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ThimPointerCursor = cursor;
    }
    name = (char *)UPbyCode(PROF_CURS_BUSY_);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ThimBusyCursor = cursor;
    }
    name = (char *)UPbyCode(PROF_CURS_ELFREC_);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ThimRecordCursor = cursor;
    }
    name = (char *)UPbyCode(PROF_CURS_NOELFREC_);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ThimNoRecordCursor = cursor;
    }
    name = (char *)UPbyCode(PROF_WP_DEFAULT_CURSOR);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ThimWPDefaultCursor = cursor;
    }
    name = (char *)UPbyCode(PROF_ME_DEFAULT_CURSOR);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ThimMEDefaultCursor = cursor;
    }
    name = (char *)UPbyCode(PROF_EQN_DEFAULT_CURSOR);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ThimEQNDefaultCursor = cursor;
    }

    name = (char *)UPbyCode(PROF_CURS_WP_);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ElfbMapCursor(IC_DC, name);
    }

    name = (char *)UPbyCode(PROF_CURS_GE_);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ElfbMapCursor(IC_GE, name);
    }

    name = (char *)UPbyCode(PROF_CURS_SS_);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ElfbMapCursor(IC_SS, name);
    }

    name = (char *)UPbyCode(PROF_CURS_CE_);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ElfbMapCursor(IC_CE, name);
    }

    name = (char *)UPbyCode(PROF_CURS_EQN_);
    if (name)
    {
        if ( strcmp(name , cursor_name) == 0 )
            ElfbMapCursor(IC_EQN, name);
    }
}

create_cursor(name, datfile,mskfile, check_name)
char *name;
char *datfile;
char *mskfile;
int check_name;
{
    ax_cursor *cursor;
    int len, sz, i;

    i = cursors.count;
    cursors.count++;
    sz = cursors.count * sizeof(ax_cursor);
    cursors.cursor = (ax_cursor *)TrueRealloc(cursors.cursor, sz);
    cursor = &cursors.cursor[i];

    len = strlen(name) + 1;
    cursor->name = (char *)TrueMalloc(len);
    strcpy(cursor->name, name);

    /* create the cursor */
    cursor->data = None;
    ZXmCreateCursor(datfile, mskfile, 0,0, &cursor->data);

    /* compare to all cursors - if in use, replace with new Cursor */
    if ( check_name )
        check_name_in_use(cursor->name, cursor->data);
}

backup_custom_cursors()
{
    old_cursors = cursors;
    cursors.cursor = (ax_cursor *)0;
    cursors.count = 0;
}

restore_backup_cursors()
{
    cursors = old_cursors;
    old_cursors.cursor = (ax_cursor *)0;
    old_cursors.count = 0;
}

free_backup_cursors()
{
    ax_cursor *cursor;
    int i;

    for ( i=0; i<old_cursors.count; i++ )
    {
        cursor = &old_cursors.cursor[i];
        XFreeCursor(Dpy, cursor->data);
        TaskFree(0, cursor->name);
    }

    TaskFree(0, old_cursors.cursor);
    old_cursors.cursor = (ax_cursor *)0;
    old_cursors.count = 0;
}

Cursor
get_custom_cursor(cursor_name)
char *cursor_name;
{
    Cursor curs;
    int i;

    curs = None;
    for ( i=0; i<cursors.count; i++ )
    {
        if ( strcmp(cursors.cursor[i].name , cursor_name) == 0 )
        {
            curs = cursors.cursor[i].data;
            break;
        }
    }
    return(curs);
}

check_window_cursors()
{
    ax_cursor *cursor, *old_cursor;
    int i;

    for ( i=0; i<old_cursors.count; i++ )
    {
        old_cursor = &old_cursors.cursor[i];
        cursor = &cursors.cursor[i];
        if ( old_cursor && cursor )
            VerifyCursor(old_cursor->data, cursor->data);
    }
#if AX_X11
    WinAtt.cursor = ThimPointerCursor;
#endif /* AX_X11 */
}

int createCustomCursors(cursrP)
data_ptr cursrP;
{
    char *name, str[STRPATH], datfile[STRPATH], mskfile[STRPATH];
    data_ptr aP, elmP, nameP, datP, mskP;
    int i, n_elm;

    if ( !cursrP || !AxIsArray(cursrP) )
        return (!AOK);
    
    /* init */
    cursors.cursor = NULL;
    cursors.count = 0;

    n_elm = AxArraySize(cursrP);
    for ( i=0; i<n_elm; i++ )
    {
        elmP = AxArrayElement(cursrP, i);
        if ( !elmP || !AxIsArray(elmP) || AxArraySize(elmP) < 2 )
            continue;

        /* extract the name */
        nameP = AxArrayElement(elmP, 0);    
        if ( !nameP || !AxIsString(nameP) )
            continue;
        else
            name = (char *)AxStrFromDataPtr(nameP);

        /* write the data file */
        datP = AxArrayElement(elmP, 1); 
        if ( !datP )
            continue;
        else
        {
            AxTempFile(NXLT("crt%%%%%%.im"), datfile);
            ElfbWriteBinaryFile(datfile, datP);
        }

        /* write the mask file (if any) and create the cursor */
        mskP = AxArrayElement(elmP, 2); 
        if ( !mskP )
        {
            create_cursor(name, datfile, (char*)0, FALSE);
            AxDeleteTempfile(datfile);
        }
        else
        {
            AxTempFile(NXLT("crt%%%%%%.im"), mskfile);
            ElfbWriteBinaryFile(mskfile, mskP);

            create_cursor(name, datfile, mskfile, FALSE);
            AxDeleteTempfile(datfile);
            AxDeleteTempfile(mskfile);
        }
    }

    return( AOK );

}

ElfbCreateCursors(cursrP)
data_ptr cursrP;
{
    char *name, str[STRPATH], datfile[STRPATH], mskfile[STRPATH];
    data_ptr aP, elmP, nameP, datP, mskP;
    int i, n_elm;

    if ( !cursrP || !AxIsArray(cursrP) )
        return (!AOK);
    
    n_elm = AxArraySize(cursrP);
    for ( i=0; i<n_elm; i++ )
    {
        elmP = AxArrayElement(cursrP, i);
        if ( !elmP || !AxIsArray(elmP) || AxArraySize(elmP) < 2 )
            ElfStrAbort(ERR_ILLVAL, (char *)0, (char *)0);

        /* extract the name */
        nameP = AxArrayElement(elmP, 0);    
        if ( !nameP || !AxIsString(nameP) )
            ElfStrAbort(ERR_ILLVAL, (char *)0, (char *)0);
        else
            name = (char *)AxStrFromDataPtr(nameP);

        /* write the data file */
        datP = AxArrayElement(elmP, 1); 
        if ( !datP )
            ElfStrAbort(ERR_ILLVAL, (char *)0, (char *)0);
        else
        {
            AxTempFile(NXLT("crt%%%%%%.im"), datfile);
            ElfbWriteBinaryFile(datfile, datP);
        }

        /* write the mask file (if any) and create the cursor */
        mskP = AxArrayElement(elmP, 2); 
        if ( !mskP )
        {
            create_cursor(name, datfile, (char*)0, TRUE);
            AxDeleteTempfile(datfile);
        }
        else
        {
            AxTempFile(NXLT("crt%%%%%%.im"), mskfile);
            ElfbWriteBinaryFile(mskfile, mskP);

            create_cursor(name, datfile, mskfile, TRUE);
            AxDeleteTempfile(datfile);
            AxDeleteTempfile(mskfile);
        }
    }
    check_window_cursors();
}

readCustomCursors()
{
    char *filename;

    filename = ElfbReadCustomFilename();
    if ( filename != (char *)0 )
    {
        ElfbReadDataFile(filename);
        createCustomCursors(ElfRetData);
        AxFreeData(ElfRetData);
        ElfRetData = NULL;
    }
}

ElfbReadCustomCursors()
{
    char *filename;

    /* save the existing custom cursors */
    backup_custom_cursors();

    /* read the cursor file */
    filename = ElfbReadCustomFilename();
    if ( filename == (char *)0 )
        restore_backup_cursors();
    else
    {
        /* apply the new cursors */
        ElfbReadDataFile(filename);
        ElfbCreateCursors(ElfRetData);
        AxFreeData(ElfRetData);

        /* delete the backup cursors */
        free_backup_cursors();
    }
    ElfRetData = NULL;
}

AxFreeCursor(cursor)
Cursor cursor;
{
    int i;

    if ( cursor != None )
    {
#if AX_X11 /* don't need to free this cursor on Windows/NT */
        for ( i=0; i<cursors.count; i++ )
        {
            if ( cursors.cursor[i].data == cursor )
                break;
        }
        if ( i >= cursors.count )
            XFreeCursor(Dpy, cursor);
#endif
    }
}
