/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xinifont.c
   Applix ASC version: 27
   Date: Tue Jan 26 21:13:23 1999
*/
#include <aos.h>
#include <eldata.h>
#include <basicwin.h>
#include <elxwins.h>
#include <axol.h>
#include <getstr4.h>
#include <getstr5.h>
#include <getstr7.h>


extern XFontStruct *AxDlgMainFont; /* Bolded dialog font */
extern XFontStruct *AxDlgMiniFont; /* small version */
extern XFontStruct *MonospaceFont, *MonospaceBoldFont;
extern XFontStruct *MonospaceItalFont, *MonospaceItalBoldFont;
extern XFontStruct *Grafont;
XFontStruct *Txfont, *olCursorfont, *BoldTxfont;
XFontStruct *MiniTxfont, *MiniLabelTxfont, *MiniButtonFont;
XFontStruct *AxStatusLineFont;
XFontStruct *AxDefaultReferenceFont;

int AxBigFont = FALSE;       /* obsolete */
int AxAlternateDialogFont;   /* bool: FALSE when venerable axsab-16 used */
data_ptr AxDialogFontName;   /* <face><pointsize><weight><slant> (or null) */

char *AxFontDir;    /* where fonts are found */
int CharWids[256];      /* of dialog font */
int SmallCharWids[256]; /* of reduced dialog font */

extern char *ThimApplixDir;    /* Where the THIMBLE executable lies */
extern int AxFontShift; /* Hack to force all fonts selects up in size */

extern int AxOpenLook;
extern int AxHighRes;       /* If 1, use 100 dpi.  If 2, use 125 dpi */
extern Display *Dpy;

extern int AxCharAbove, AxCharBelow, AxCharHeight;
extern int SmallCharAbove, SmallCharBelow, SmallCharHeight;
extern char *AxComlFontPath;
extern int JavaMode;
extern int AxElfrtMode;     /* TRUE if running as elfrt rather than axmain */
int AxElfrtHasNoFonts;	/* if elfrt mode nad cannot find applix fonts */

extern char *GetProfileByStr();

extern XCharStruct *getCharStruct();

void tempAbort(), initCharWids();
char *UPneedByCode(), *AxProfileValue();
XFontStruct *AxProfXLoadQueryFont();
XFontStruct *AXLoadQueryFont();

static XFontStruct *AXLoadDialogFont();


void InitFontInfo()
{
    XFontStruct *main_fsP;

    char *path, *prof_path, *name;
    char primary[STRPATH + 1], secondary[STRPATH + 1];
    char *msg_path = NULL;
    int ix, j, lineNumber, canFree;

    /* read in the font set config file "fontmap.dir" */
    if(!axtInitFonts(&lineNumber)) {
        printf(XLT("Problem with fontmap.dir file around line %d\n"),
                lineNumber);
    }

    /* ============================================================= */
    /* Default font name for menus/dialog boxes */
    /* ============================================================= */
    name = NXLT("-applix-axsab-bold-r-normal--16-160-75-75-p-65-iso8859-1");
    
    AxBigFont = FALSE;
    AxHighRes = FALSE;
    
    /* ============================================================= */
    /* Priority of font access path:
        1. Command line
        2. Profile
        3. Device primary
        4. Device secondary
    */
    /* ============================================================= */

    path = prof_path = UPneedByCode(PROF_FONTS_DIR_);

    /* Don't bother asking about paths if a sample font can be opened now */
        /* ASTERIX_FONT should be unique to us */
    main_fsP = AXLoadQueryFont(Dpy, ASTERIX_FONT);
    if (main_fsP != NULL)
        {
        AXFreeFont(Dpy, main_fsP);    /* toss sample */
        main_fsP = AXLoadDialogFont(Dpy, name, &canFree);  /* open real one */
        if (main_fsP != NULL)
            {
            goto got_font_path;
            }
        }

    /* Try command line specified if there was one */
    if (AxComlFontPath)
        {
        msg_path = AxComlFontPath;
        AxSetFontPath(msg_path, path = msg_path);
		XSync(Dpy, 0);
        main_fsP = AXLoadDialogFont(Dpy, name, &canFree);
        if (main_fsP != NULL)
            goto got_font_path;
        }

    /* Nope, try from profile if there is one */
    if (prof_path && prof_path[0])
        {
        if (!msg_path)
            msg_path = prof_path;
        AxSetFontPath(msg_path, path = prof_path);
		XSync(Dpy, 0);
        main_fsP = AXLoadDialogFont(Dpy, name, &canFree);
        if (main_fsP != NULL)
            goto got_font_path;
        }

    /* Nope, try based on device's 'old' style */
    AxGetDevicePaths(primary, secondary);
    if (!msg_path)
        msg_path = primary;
    AxSetFontPath(msg_path, path = primary);
	XSync(Dpy, 0);
    main_fsP = AXLoadDialogFont(Dpy, name, &canFree);
    if (main_fsP != NULL)
        goto got_font_path;
    
    /* Nope, try based on device's 'new' style */
    if (!msg_path)
        msg_path = secondary;
    AxSetFontPath(msg_path, path = secondary);
	XSync(Dpy, 0);
    main_fsP = AXLoadDialogFont(Dpy, name, &canFree);
    if (main_fsP != NULL)
        goto got_font_path;

	/* If opensrc elfmain, setle for system fonts! */
	if ( AxElfrtMode ) {
		AxElfrtHasNoFonts = TRUE;
	    main_fsP = AXLoadQueryFont(Dpy, "fixed");
		if (main_fsP != NULL)
	        goto got_font_path;
	}

/* Oh well, punt! */

font_punt:
    if (!msg_path || !(*msg_path)) {
        msg_path = XLT("<null>");
    }
    printf(XLT2("xinifont.c-[1]","X-Server problem - unable to access fonts\n"));
    printf(XLT2("xinifont.c-[2]","The font path being used is:\n\n"));
    printf(XLT2("xinifont.c-[3]","    '%s'\n\n"), msg_path);
    printf(XLT2("xinifont.c-[4]","If this is not appropriate for your display, restart\n"));
    printf(XLT2("xinifont.c-[5]","Applixware specifying '-fp your_font_path' on the\n"));
    printf(XLT2("xinifont.c-[6]","command line.\n\n"));
    printf(XLT2("xinifont.c-[7]","For further info, see the 'Fonts' section\n"));
    printf(XLT2("xinifont.c-[8]","of the installation notice.\n"));
	AxRemovePipe();
	exit(1);

got_font_path:

    AxDlgMainFont = main_fsP;

    if (AxAlternateDialogFont)
        {
            /* should open mini font based on acutal size of alternate.. */
        AxDlgMiniFont = AXLoadQueryFont(Dpy,
          NXLT("-applix-axsab-bold-r-normal--16-160-75-75-p-65-iso8859-1"));
        AxDefaultReferenceFont = AxDlgMiniFont;  /* points to axsab-16 */
        }
    else
        {
        AxDlgMiniFont = AxProfXLoadQueryFont(Dpy, NXLT("axFontMini"),
          NXLT("-applix-axhvb-bold-r-normal--13-120-75-75-p-68-iso8859-1"));
        AxDefaultReferenceFont = AxDlgMainFont;  /* points to axsab-16 */
        }

    AxStatusLineFont = AXLoadQueryFont
        (Dpy,NXLT("-applix-axhvb-bold-r-normal--13-120-75-75-p-68-iso8859-1"));
                                       

    if (AxAlternateDialogFont &&
        ((AxDlgMainFont->max_bounds.ascent +
          AxDlgMainFont->max_bounds.descent) > 16))
        {
            /* we're scaling dialog layout using a main dialog font
               taller than axsab-16 */
        MonospaceFont = AXLoadQueryFont(Dpy,
         NXLT("-applix-axcor-medium-r-normal--15-140-75-75-p-90-iso8859-1"));
        MonospaceBoldFont = AXLoadQueryFont(Dpy,
           NXLT("-applix-axcob-bold-r-normal--15-140-75-75-p-90-iso8859-1"));
        MonospaceItalFont = AXLoadQueryFont(Dpy,
         NXLT("-applix-axcoi-medium-o-normal--15-140-75-75-p-90-iso8859-1"));
        MonospaceItalBoldFont = AXLoadQueryFont(Dpy,
           NXLT("-applix-axcoj-bold-o-normal--15-140-75-75-p-90-iso8859-1"));
        }
    else
        {
        MonospaceFont = AxProfXLoadQueryFont(Dpy, NXLT("axFontMono"),
         NXLT("-applix-axcor-medium-r-normal--13-120-75-75-p-70-iso8859-1"));
        MonospaceBoldFont = AxProfXLoadQueryFont(Dpy, NXLT("axFontMonoBold"),
           NXLT("-applix-axcob-bold-r-normal--13-120-75-75-p-70-iso8859-1"));
        MonospaceItalFont = AxProfXLoadQueryFont(Dpy, NXLT("axFontMonoItalic"),
         NXLT("-applix-axcoi-medium-o-normal--13-120-75-75-p-70-iso8859-1"));
        MonospaceItalBoldFont = AxProfXLoadQueryFont(Dpy, NXLT("axFontMonoBoldItalic"),
           NXLT("-applix-axcoj-bold-o-normal--13-120-75-75-p-70-iso8859-1"));
        }


	if (!AxDlgMiniFont || !MonospaceFont) {
		if ( !AxElfrtMode )
			goto font_punt;
		if ( !AxDlgMiniFont ) AxDlgMiniFont = AxDlgMainFont;
		if ( !AxDefaultReferenceFont ) AxDefaultReferenceFont = AxDlgMainFont;
		if ( !MonospaceFont ) MonospaceFont = AxDlgMainFont;
		if ( !MonospaceBoldFont ) MonospaceBoldFont = AxDlgMainFont;
		if ( !MonospaceItalFont ) MonospaceItalFont = AxDlgMainFont;
		if ( !MonospaceItalBoldFont ) MonospaceItalBoldFont = AxDlgMainFont;
	}

    /* map current server-dependent modifier key masks to ours */
    AxModifierMapping();

    if (!AxElfrtMode || AxOpenLook)
        {
        if ((olCursorfont = AXLoadQueryFont(Dpy, OL_CURSOR_FONT)) == NULL)
            {
            printf(NXLT("Couldn't find the cursor font: %s\n"), OL_CURSOR_FONT);
			AxRemovePipe();
			exit(1);
            }
        }

    /*=======================================================================*/
    /* If the ultimately selected font path is not the same as the profile,  */
    /* update the profile. */
    /*=======================================================================*/
    AxFontDir = strcpy(malloc(strlen(path) + 1), path);
    if (!prof_path || !prof_path[0] || !streq(prof_path, path))
        {
        AddProfileByCode(PROF_FONTS_DIR_, path);
        ProfileSave();  /* update disk file */
        }
    
    initCharWids(main_fsP, 1, 1000, CharWids, &AxCharAbove,
                 &AxCharBelow, &AxCharHeight);

    initCharWids(AxDlgMiniFont, 1, 1000, SmallCharWids, &SmallCharAbove,
                 &SmallCharBelow, &SmallCharHeight);

    return;
}


AxUseBigFont()
{
    return(AxBigFont);
}

char *
AxProfileValue(profName, defaultName, foundProfName)
char *profName;      /* in */
char *defaultName;   /* in */
int *foundProfName;  /* out: bool */
{
    char *GetProfileByStr(), *pName;
    
    pName = GetProfileByStr(profName);
    if (pName && *pName) {
        *foundProfName = TRUE;
        return (pName);
    }
    else {
        *foundProfName = FALSE;
        return (defaultName);
    }
}

/*************************************************************************
function: AxProfXLoadQueryFont - Load font based on profile
*/
XFontStruct *AxProfXLoadQueryFont (dpy, profName, defaultName)
Display *dpy;
char *profName;
char *defaultName;
{
    char *p;
    int tmp;
    XFontStruct *fs = NULL;

    p = AxProfileValue(profName, defaultName, &tmp);
    fs = AXLoadQueryFont(dpy, p);
    return(fs);
}

static XFontStruct *
AXLoadDialogFont(dpy, defaultName, freeThis)
Display *dpy;
char *defaultName;
int  *freeThis;        /* out: bool; whether struct is "freeable" */
{
    data_ptr atmp;
    XFontStruct *fs = (XFontStruct *) NULL;
    char *profFontName = GetProfileByStr(XLT("axDialogMainFont"));
    char *profFontSize = GetProfileByStr(XLT("axDialogMainFontSize"));
    char *profFontWeight = GetProfileByStr(XLT("axDialogMainFontWeight"));
    char *profFontSlant = GetProfileByStr(XLT("axDialogMainFontSlant"));
    
    *freeThis = FALSE;

    if (!JavaMode && (profFontName && *profFontName) && 
		(strcmp(profFontName,NXLT("AxDialog")) != 0)) {
		
        int weight, slant, pointsize, scalable;
        
        if (axtWellFormedXLFD(profFontName)) {
                /* try to open directly as an XLFD name */
            fs = AXLoadQueryFont(dpy, profFontName);
            if (fs) {
                AxAlternateDialogFont = TRUE;
                *freeThis = TRUE;
            }
            else {
                AxAlternateDialogFont = FALSE;
            }
            return (fs);
        }
        
        if (profFontSize && *profFontSize) {
            pointsize = atoi(profFontSize);
        }
        else {
            pointsize = 14;
        }
        
        if (profFontWeight && *profFontWeight) {
            weight = atoi(profFontWeight);
            if (weight == -1)
                weight = 1;     /* convert elf true to weight value of 1 */
        }
        else {
            weight = 0;
        }
        
        if (profFontSlant && *profFontSlant) {
            slant = atoi(profFontSlant);
            if (slant == -1)
                slant = 1;     /* convert elf true to slant value of 1 */
        }
        else {
            slant = 0;
        }
        
        if (getXFontStruct(profFontName, pointsize, weight, slant,
                           &scalable, &fs)) {
            AxAlternateDialogFont = TRUE;
#if BUILD_ORIENT
            if (fs->max_byte1 > 0) {
                /* The Dialog font must be narrow, since the wide font bonded
                 * to this font will be loaded in LoadOrientalDialogueFont()
                 * latter. Ashisuto added on preventing crash of program when
                 * user selected wide font as menu&dialog font face.    96.02.14
                 */
                if (IsSansFontFace(profFontName))
                    profFontName = NXLT("Helvetica");
                else
                    profFontName = NXLT("Times");
                if (!getXFontStruct(profFontName, pointsize, weight, slant,
                        &scalable, &fs))
                    AxAlternateDialogFont = FALSE;
            }
#endif
        }
        else {
            AxAlternateDialogFont = FALSE;
        }
        
            /* write the dialog font name represented by
               the profile data. */
        AxDialogFontName = AxMakeArray(4);
        
        atmp = AxMakeArray(2);
        AxAddIntToArray(atmp, 0, FONT_FAMILY);
        AxAddStrToArray(atmp, 1, (mbchar *) profFontName);
        AxAddArrayToArray(AxDialogFontName, 0, atmp);
        
        atmp = AxMakeArray(2);
        AxAddIntToArray(atmp, 0, FONT_SIZE);
        AxAddIntToArray(atmp, 1, pointsize);
        AxAddArrayToArray(AxDialogFontName, 1, atmp);
        
        atmp = AxMakeArray(2);
        AxAddIntToArray(atmp, 0, FONT_WEIGHT);
        AxAddIntToArray(atmp, 1, weight);
        AxAddArrayToArray(AxDialogFontName, 2, atmp);
        
        atmp = AxMakeArray(2);
        AxAddIntToArray(atmp, 0, FONT_SLANT);
        AxAddIntToArray(atmp, 1, slant);
        AxAddArrayToArray(AxDialogFontName, 3, atmp);
        
        return (fs);
    }
    else {
        AxDialogFontName = (data_ptr) NULL;
        fs = AXLoadQueryFont(dpy, defaultName);
        if (fs) {
            *freeThis = TRUE;
        }
        return (fs);
    }
}


