/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmarrows.c
   Applix ASC version: 27
   Date: Mon Nov 16 13:53:42 1998
*/
#define OEMRESOURCE     /* needed for Windows built-in bitmap defines */

#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>

extern Display *Dpy;
extern Window RootWin;
extern int UseWidgetColors;
extern GC ForeGC, ShadowGC, BackGC, SunGC, RecessGC;

extern Pixmap AxCreateQueryPixmap();
extern GC axCreateGC();

#if AXMS
extern HDC InsureDrawable(HGDIOBJ, int*);    /* get or create a DC */
extern void RemoveDrawable(HGDIOBJ, HDC);    /* delete or release a DC */
#endif

Pixmap AXCreatePixmapFromBitmapData();

static int bitVal();

/* These pixmap names will be referred to by the widget kit */

static struct {
    Pixmap 
#if AX_X11
           xmTopArrowSun,
           xmTopArrowFore,
           xmTopArrowShadow,
           xmBotArrowSun,
           xmBotArrowFore,
           xmBotArrowShadow,
#endif
           xmRightArrowSun,
           xmRightArrowFore,
           xmRightArrowShadow,
           xmLeftArrowSun,
           xmLeftArrowFore,
           xmLeftArrowShadow;
}xmarrowsStr;

#if AX_X11       /*  {  */
static unsigned char topArrowSunBits[] = {
   0x04, 0x00,
   0x04, 0x00,
   0x0c, 0x00,
   0x0c, 0x00,
   0x18, 0x00,
   0x18, 0x00,
   0x30, 0x00,
   0x30, 0x00,
   0x60, 0x00,
   0x60, 0x00,
   0xc0, 0x00,
   0x00, 0x00,
};

static unsigned char topArrowForeBits[] = {
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x06, 0x00,
   0x06, 0x00,
   0x0f, 0x00,
   0x0f, 0x00,
   0x1f, 0x80,
   0x1f, 0x80,
   0x00, 0x00,
   0x00, 0x00,
};

static unsigned char topArrowShadowBits[] = {
   0x02, 0x00,
   0x02, 0x00,
   0x03, 0x00,
   0x03, 0x00,
   0x01, 0x80,
   0x01, 0x80,
   0x00, 0xc0,
   0x00, 0xc0,
   0x00, 0x60,
   0x00, 0x60,
   0x7f, 0xf0,
   0xff, 0xf0,
};

static unsigned char botArrowSunBits[] = {
   0xff, 0xe0,
   0xff, 0xc0,
   0x60, 0x00,
   0x60, 0x00,
   0x30, 0x00,
   0x30, 0x00,
   0x18, 0x00,
   0x18, 0x00,
   0x0c, 0x00,
   0x0c, 0x00,
   0x04, 0x00,
   0x04, 0x00,
};

static unsigned char botArrowForeBits[] = {
   0x00, 0x00,
   0x00, 0x00,
   0x1f, 0x80,
   0x1f, 0x80,
   0x0f, 0x00,
   0x0f, 0x00,
   0x06, 0x00,
   0x06, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
};

static unsigned char botArrowShadowBits[] = {
   0x00, 0x10,
   0x00, 0x30,
   0x00, 0x60,
   0x00, 0x60,
   0x00, 0xc0,
   0x00, 0xc0,
   0x01, 0x80,
   0x01, 0x80,
   0x03, 0x00,
   0x03, 0x00,
   0x02, 0x00,
   0x02, 0x00,
};
#endif             /* } AX_X11 */  

static unsigned char rightArrowSunBits[] = {
   0xc0, 0x00,
   0xf0, 0x00,
   0xfc, 0x00,
   0xcf, 0x00,
   0xc3, 0xc0,
   0xc0, 0xf0,
   0xc0, 0x00,
   0xc0, 0x00,
   0xc0, 0x00,
   0xc0, 0x00,
   0x80, 0x00,
   0x00, 0x00,
};

static unsigned char rightArrowForeBits[] = {
   0x00, 0x00, 
   0x00, 0x00,
   0x00, 0x00,
   0x30, 0x00,
   0x3c, 0x00,
   0x3f, 0x00,
   0x3f, 0x00,
   0x3c, 0x00,
   0x30, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
};

static unsigned char rightArrowShadowBits[] = {
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0xf0,
   0x03, 0xc0,
   0x0f, 0x00,
   0x3c, 0x00,
   0x70, 0x00,
   0xc0, 0x00,
};

static unsigned char leftArrowSunBits[] = {
   0x00, 0x20,
   0x00, 0xe0,
   0x03, 0xc0,
   0x0f, 0x00,
   0x3c, 0x00,
   0xf0, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
};

static unsigned char leftArrowForeBits[] = {
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0xc0,
   0x03, 0xc0,
   0x0f, 0xc0,
   0x0f, 0xc0,
   0x03, 0xc0,
   0x00, 0xc0,
   0x00, 0x00,
   0x00, 0x00,
   0x00, 0x00,
};

static unsigned char leftArrowShadowBits[32] = {
   0x00, 0x10,
   0x00, 0x10,
   0x00, 0x30,
   0x00, 0x30,
   0x00, 0x30,
   0x00, 0x30,
   0xf0, 0x30,
   0x3c, 0x30,
   0x0f, 0x30,
   0x03, 0xf0,
   0x00, 0xf0,
   0x00, 0x30,
};

#define ARROW_WIDTH 16
#define ARROW_HEIGHT 12

void
axmCreateArrowStipples()
{
#if AX_X11      /*  {  */
    xmarrowsStr.xmTopArrowSun =
		AXCreatePixmapFromBitmapData(Dpy, RootWin, topArrowSunBits,
									 ARROW_WIDTH, ARROW_HEIGHT,
									 AllPlanes, ~AllPlanes, 1);

    xmarrowsStr.xmTopArrowFore =
		AXCreatePixmapFromBitmapData(Dpy, RootWin, topArrowForeBits,
									 ARROW_WIDTH, ARROW_HEIGHT,
									 AllPlanes, ~AllPlanes, 1);

    xmarrowsStr.xmTopArrowShadow =
		AXCreatePixmapFromBitmapData(Dpy, RootWin, topArrowShadowBits,
									 ARROW_WIDTH, ARROW_HEIGHT,
									 AllPlanes, ~AllPlanes, 1);

    xmarrowsStr.xmBotArrowSun =
		AXCreatePixmapFromBitmapData(Dpy, RootWin, botArrowSunBits,
									 ARROW_WIDTH, ARROW_HEIGHT,
									 AllPlanes, ~AllPlanes, 1);

    xmarrowsStr.xmBotArrowFore =
		AXCreatePixmapFromBitmapData (Dpy, RootWin, botArrowForeBits,
									  ARROW_WIDTH, ARROW_HEIGHT,
									  AllPlanes, ~AllPlanes, 1);

    xmarrowsStr.xmBotArrowShadow =
		AXCreatePixmapFromBitmapData(Dpy, RootWin, botArrowShadowBits,
									 ARROW_WIDTH, ARROW_HEIGHT,
									 AllPlanes, ~AllPlanes, 1);
#endif    /*  } AX_X11   */

    xmarrowsStr.xmRightArrowSun =
		AXCreatePixmapFromBitmapData(Dpy, RootWin, rightArrowSunBits,
									 ARROW_WIDTH, ARROW_HEIGHT,
									 AllPlanes, ~AllPlanes, 1);

    xmarrowsStr.xmRightArrowFore =
		AXCreatePixmapFromBitmapData(Dpy, RootWin, rightArrowForeBits,
									 ARROW_WIDTH, ARROW_HEIGHT,
									 AllPlanes, ~AllPlanes, 1);

    xmarrowsStr.xmRightArrowShadow =
		AXCreatePixmapFromBitmapData(Dpy, RootWin, rightArrowShadowBits,
									  ARROW_WIDTH, ARROW_HEIGHT,
									  AllPlanes, ~AllPlanes, 1);

    xmarrowsStr.xmLeftArrowSun =
		AXCreatePixmapFromBitmapData(Dpy, RootWin, leftArrowSunBits,
									 ARROW_WIDTH, ARROW_HEIGHT,
									 AllPlanes, ~AllPlanes, 1);

    xmarrowsStr.xmLeftArrowFore =
		AXCreatePixmapFromBitmapData(Dpy, RootWin, leftArrowForeBits,
									 ARROW_WIDTH, ARROW_HEIGHT,
									 AllPlanes, ~AllPlanes, 1);

    xmarrowsStr.xmLeftArrowShadow =
		AXCreatePixmapFromBitmapData(Dpy, RootWin, leftArrowShadowBits,
									 ARROW_WIDTH, ARROW_HEIGHT, AllPlanes,
									 ~AllPlanes, 1);
}
	
Pixmap
AXCreatePixmapFromBitmapData(dpy, win, bits, width, height,
                             fpix, bpix, depth)
Display *dpy;
Drawable win;
unsigned char *bits;
int width, height;
unsigned long fpix, bpix;
int depth;
{
    int ix, iy, bytewidth;
    XImage *img;
    Visual *vis;
    Pixmap pix;
    GC gc;
    XGCValues gcv;

        /* first get the pixmap */
    if( !(pix = AxCreateQueryPixmap(dpy, win, width, height, depth))) {
        return ((Pixmap) NULL);
    }
        /* get a same-depth GC */
    gcv.foreground = fpix;
    gcv.background = bpix;
    gc = axCreateGC(dpy, pix, DefaultScreen(dpy), depth, (XFontStruct *) 0,
					(unsigned long)(GCForeground | GCBackground), &gcv);

    /* create an XImage and draw the bitmap data onto it */
    bytewidth = (width + 7) / 8;                    /* for image data */
    vis = DefaultVisual(dpy, DefaultScreen(dpy));
    img = XCreateImage(dpy, vis, depth, XYPixmap, 0, (char *)NULL,
                       width, height, 8, 0);
    img->data = (char *)calloc(1, (img->bytes_per_line * img->height) );

    for (iy = 0; iy < height; iy++) {
        for (ix = 0; ix < width; ix++) {
            if (bitVal(bits, ix, iy, bytewidth)) {
                XPutPixel(img, ix, iy, fpix);
            }
            else {
                XPutPixel(img, ix, iy, bpix);
            }
        }
    }
        /* transfer to the pixmap */
    XPutImage(dpy, pix, gc, img, 0, 0, 0, 0, width, height);

    free(img->data);
    img->data = (char *) NULL;
    XDestroyImage(img);
    XFreeGC(dpy, gc);
    return (pix);
}

static int masks[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

static int
bitVal(data, x, y, bytewidth)
char *data;
int x, y, bytewidth;
{
    int container, rem, tmp;
    tmp = (y * bytewidth) + (x / 8);
    rem = x % 8;
    container = data[ tmp ];    /* byte containing bit at x, y */
    return (container & masks[rem]);
}



#define AXTOPARROW 1
#define AXBOTARROW 2
#define AXRIGHTARROW 3
#define AXLEFTARROW 4


#if AX_X11      /*  {  */

static void
axmPaintArrow(w, drw, xoffset, yoffset, arrow_up, whichArr)
AXWidget w;
Drawable drw;
int xoffset, yoffset;
int arrow_up, whichArr;
{
    Pixmap sunpix, shadowpix, forepix;
    int xoff, yoff;

	if (UseWidgetColors) {
		switch(whichArr) {
		case AXTOPARROW:
			xoff = xoffset + 2;
			yoff = yoffset + 2;
			sunpix = xmarrowsStr.xmTopArrowSun;
			shadowpix = xmarrowsStr.xmTopArrowShadow;
			forepix = xmarrowsStr.xmTopArrowFore;
			break;
		case AXBOTARROW:
			xoff = xoffset + 2;
			yoff = yoffset - ARROW_HEIGHT - 2;
			sunpix = xmarrowsStr.xmBotArrowSun;
			shadowpix = xmarrowsStr.xmBotArrowShadow;
			forepix = xmarrowsStr.xmBotArrowFore;
			break;

		case AXRIGHTARROW:
			xoff = xoffset - ARROW_WIDTH + 2;
			yoff = 2;
			sunpix = xmarrowsStr.xmRightArrowSun;
			shadowpix = xmarrowsStr.xmRightArrowShadow;
			forepix = xmarrowsStr.xmRightArrowFore;
			break;
		case AXLEFTARROW:
			yoff = 2;
			xoff = xoffset + 2;
			sunpix = xmarrowsStr.xmLeftArrowSun;
			shadowpix = xmarrowsStr.xmLeftArrowShadow;
			forepix = xmarrowsStr.xmLeftArrowFore;
			break;
		}
		
		/* prepare GCs, then draw arrows as stippled fill-rects
		 */
		if (arrow_up ) {
			if (getenv("EWS_XFILL_HACK")) {
				drawStipple(Dpy, drw, SunGC, sunpix, 12, 12, xoff, yoff);
				drawStipple(Dpy, drw, BackGC, forepix, 12, 12, xoff, yoff);
				drawStipple(Dpy, drw, ShadowGC, shadowpix, 12, 12, xoff, yoff);
			} else {
				XSetFillStyle(Dpy, SunGC, FillStippled);
				XSetStipple(Dpy, SunGC, sunpix);
				XSetTSOrigin(Dpy, SunGC, xoff, yoff);
				XFillRectangle(Dpy, drw, SunGC, xoff, yoff, 12, 12);

				XSetFillStyle(Dpy, BackGC, FillStippled);
				XSetStipple(Dpy, BackGC, forepix);
				XSetTSOrigin(Dpy, BackGC, xoff, yoff);
				XFillRectangle(Dpy, drw, BackGC, xoff, yoff, 12, 12);

				XSetFillStyle(Dpy, ShadowGC, FillStippled);
				XSetStipple(Dpy, ShadowGC, shadowpix);
				XSetTSOrigin(Dpy, ShadowGC, xoff, yoff);
				XFillRectangle(Dpy, drw, ShadowGC, xoff, yoff, 12, 12);
			}
		}
		else {
			if (getenv("EWS_XFILL_HACK")) {
				drawStipple(Dpy, drw, SunGC, shadowpix, 12, 12, xoff, yoff);
				drawStipple(Dpy, drw, RecessGC, forepix, 12, 12, xoff, yoff);
				drawStipple(Dpy, drw, ShadowGC, sunpix, 12, 12, xoff, yoff);
			} else {
				XSetFillStyle(Dpy, SunGC, FillStippled);
				XSetStipple(Dpy, SunGC, shadowpix);
				XSetTSOrigin(Dpy, SunGC, xoff, yoff);
				XFillRectangle(Dpy, drw, SunGC, xoff, yoff, 12, 12);
		
				XSetFillStyle(Dpy, RecessGC, FillStippled);
				XSetStipple(Dpy, RecessGC, forepix);
				XSetTSOrigin(Dpy, RecessGC, xoff, yoff);
				XFillRectangle(Dpy, drw, RecessGC, xoff, yoff, 12, 12);
				
				XSetFillStyle(Dpy, ShadowGC, FillStippled);
				XSetStipple(Dpy, ShadowGC, sunpix);
				XSetTSOrigin(Dpy, ShadowGC, xoff, yoff);
				XFillRectangle(Dpy, drw, ShadowGC, xoff, yoff, 12, 12);
			}
		}
		
		/* restore the GCs */
		XSetFillStyle(Dpy, SunGC, FillSolid);
		XSetFillStyle(Dpy, BackGC, FillSolid);
		XSetFillStyle(Dpy, RecessGC, FillSolid);
		XSetFillStyle(Dpy, ShadowGC, FillSolid);
	}
	else {
		/* just do solid black arrows, identical 12x12 footprint
		 */
		switch(whichArr) {

		case AXTOPARROW:
			xoff = xoffset + 2;
			yoff = yoffset + 2;
			XFillRectangle(Dpy, drw, ForeGC, xoff + 1, yoff + 8, 10, 4);
			XFillRectangle(Dpy, drw, ForeGC, xoff + 3, yoff + 4, 6, 4);
			XFillRectangle(Dpy, drw, ForeGC, xoff + 5, yoff, 2, 4);
			XDrawLine(Dpy, drw, ForeGC, xoff, yoff + 10, xoff, yoff + 11);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 11, yoff + 10, xoff + 11, yoff + 11);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 2, yoff + 6, xoff + 2, yoff + 7);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 9, yoff + 6, xoff + 9, yoff + 7);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 4, yoff + 2, xoff + 4, yoff + 3);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 7, yoff + 2, xoff + 7, yoff + 3);
			break;
			
		case AXBOTARROW:
			xoff = xoffset + 2;
			yoff = yoffset - ARROW_HEIGHT - 2;
			XFillRectangle(Dpy, drw, ForeGC, xoff + 1, yoff, 10, 4);
			XFillRectangle(Dpy, drw, ForeGC, xoff + 3, yoff + 4, 6, 4);
			XFillRectangle(Dpy, drw, ForeGC, xoff + 5, yoff + 8, 2, 4);
			XDrawLine(Dpy, drw, ForeGC, xoff, yoff, xoff, yoff + 1);
			XDrawLine(Dpy, drw, ForeGC, xoff + 11, yoff, xoff + 11, yoff + 1);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 2, yoff + 4, xoff + 2, yoff + 5);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 9, yoff + 4, xoff + 9, yoff + 5);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 4, yoff + 8, xoff + 4, yoff + 9);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 7, yoff + 8, xoff + 7, yoff + 9);
			break;
			
		case AXRIGHTARROW:
			xoff = xoffset - ARROW_WIDTH + 2;
			yoff = 2;
			XFillRectangle(Dpy, drw, ForeGC, xoff, yoff + 1, 4, 10);
			XFillRectangle(Dpy, drw, ForeGC, xoff + 4, yoff + 3, 4, 6);
			XFillRectangle(Dpy, drw, ForeGC, xoff + 8, yoff + 5, 4, 2);
			
			XDrawLine(Dpy, drw, ForeGC, xoff, yoff, xoff + 1, yoff);
			XDrawLine(Dpy, drw, ForeGC, xoff, yoff + 11, xoff + 1, yoff + 11);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 4, yoff + 2, xoff + 5, yoff + 2);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 4, yoff + 9, xoff + 5, yoff + 9);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 8, yoff + 4, xoff + 9, yoff + 4);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 8, yoff + 7, xoff + 9, yoff + 7);
			break;
			
		case AXLEFTARROW:
			yoff = 2;
			xoff = xoffset + 2;
			XFillRectangle(Dpy, drw, ForeGC, xoff + 8, yoff + 1, 4, 10);
			XFillRectangle(Dpy, drw, ForeGC, xoff + 4, yoff + 3, 4, 6);
			XFillRectangle(Dpy, drw, ForeGC, xoff, yoff + 5, 4, 2);
			
			XDrawLine(Dpy, drw, ForeGC, xoff + 10, yoff, xoff + 11, yoff);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 10, yoff + 11, xoff + 11, yoff + 11);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 6, yoff + 2, xoff + 7, yoff + 2);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 6, yoff + 9, xoff + 7, yoff + 9);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 2, yoff + 4, xoff + 3, yoff + 4);
			XDrawLine(Dpy, drw, ForeGC,
					  xoff + 2, yoff + 7, xoff + 3, yoff + 7);
			break;
			
		}
	}
}

/* : axmPaintTopArrow - do the top triangle of a vertical scrollbar */
void
axmPaintTopArrow(w, drw, xoffset, yoffset, arrow_up)
AXWidget w;
Drawable drw;
int xoffset, yoffset;
int arrow_up;
{
    axmPaintArrow(w, drw, xoffset, yoffset, arrow_up, AXTOPARROW);
}

/* : axmPaintBottomArrow - do the bottom triangle of a vertical scrollbar */
void
axmPaintBottomArrow(w, drw, xoffset, yoffset, arrow_up)
AXWidget w;
Drawable drw;
int xoffset, yoffset;
int arrow_up;
{
    axmPaintArrow(w, drw, xoffset, yoffset, arrow_up, AXBOTARROW);
}

#endif     /*  }   AX_X11  */


#if AXMS    /* { */
/** On Windows, we use the pre-defined arrow bitmaps. **/

static void
axwinPaintArrow(AXWidget w, int xoffset, int yoffset,
                            int arrow_up, int whichArr)
{
    int x, y;
    HDC hdc, mem_hdc;
    HBITMAP hbm, old_hbm;
    int remove_dc;
    BITMAP bm;
 
    hdc = InsureDrawable(w->window, &remove_dc);
    switch (whichArr)
        {
        case AXLEFTARROW:
            x = 7;
            y = 0;
            hbm = MS_LoadBitmap((HINSTANCE)NULL, (LPCTSTR)OBM_LFARROW);
            MS_GetObject(hbm, sizeof(BITMAP), &bm);
            break;
 
        case AXRIGHTARROW:
            x = w->width - 45 - 16;
            y = 0;
            hbm = MS_LoadBitmap((HINSTANCE)NULL, (LPCTSTR)OBM_RGARROW);
            MS_GetObject(hbm, sizeof(BITMAP), &bm);
            break;
        }
 
    mem_hdc = MS_CreateCompatibleDC(NULL);
    old_hbm = MS_SelectObject(mem_hdc, hbm);
    MS_BitBlt(hdc, x, y, bm.bmWidth, bm.bmHeight, mem_hdc, 0, 0, SRCCOPY);
    MS_SelectObject(mem_hdc, old_hbm);
    MS_DeleteObject(hbm);
    MS_DeleteDC(mem_hdc);
    if ( remove_dc )
        RemoveDrawable(w->window, hdc);
}
#endif    /* } AXMS */




/* : axmPaintRightArrow - do the right triangle of a vertical scrollbar */
void
axmPaintRightArrow(w, drw, xoffset, yoffset, arrow_up)
AXWidget w;
Drawable drw;
int xoffset, yoffset;
int arrow_up;
{
#if AXMS
    axwinPaintArrow(w, xoffset, yoffset, arrow_up, AXRIGHTARROW);
#else
    axmPaintArrow(w, drw, xoffset, yoffset, arrow_up, AXRIGHTARROW);
#endif
}

/* : axmPaintLeftArrow - do the left triangle of a vertical scrollbar */
void
axmPaintLeftArrow(w, drw, xoffset, yoffset, arrow_up)
AXWidget w;
Drawable drw;
int xoffset, yoffset;
int arrow_up;
{
#if AXMS
    axwinPaintArrow(w, xoffset, yoffset, arrow_up, AXLEFTARROW);
#else
    axmPaintArrow(w, drw, xoffset, yoffset, arrow_up, AXLEFTARROW);
#endif
}

