/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmcanvas.c
   Applix ASC version: 56
   Date: Wed May 12 21:10:33 1999
*/
#if AX_X11
#include <X11/keysym.h>
#else
#include <basicwin.h>
#endif

#include <ax_xm.h>
#include <aos.h>
#include <dwt2xm.h>
#include <xmlib.h>
#include <elxwins.h>
#include <xmpixel.h>
#include <thimble.h>
#include <ctype.h>
#include <elf_env.h>
#include <eldata.h>
#include <evreason.h>
#include <xmcanvas.h>

extern int WidgetNeedsUpdated;  /* Set when a widget needs repainted */
extern Window AxCrtWindow();
extern AXWidget CrtWidget();
extern int AxCharAbove, AxCharHeight;

extern int UseWidgetColors;
extern Display *Dpy;
extern short DpyDepth;
extern int AxButtonDown;

extern long WinMask;
extern XSetWindowAttributes WinAtt;
extern Pixmap UnknownPixmap;
extern AXWidget ClickWidget;
extern unsigned long BackWorkPixel;
extern Colormap WorkColormap;
extern GC ForeGC, BackGC;
extern GC SunGC, ShadowGC, DisabledGC, GrayLineGC;
extern GC MonospaceGC;       /* monspace system font, forground mode */
extern GC MonospaceBackGC;   /* monspace system font, background mode */
extern GC MonospaceBoldGC, MonospaceBoldBackGC;
extern GC MonospaceItalGC, MonospaceItalBackGC;
extern GC MonospaceItalBoldGC, MonospaceItalBoldBackGC;
extern XFontStruct *AxDlgMainFont;  /* Bolded dialog font */
extern XFontStruct *MonospaceFont, *MonospaceBoldFont;
extern XFontStruct *MonospaceItalFont, *MonospaceItalBoldFont;
extern int DoubleMillisec;
extern Window AxCrtWindow();
extern int TextWidth();
extern int *CreateIntsArray(), *CopyIntsArray();
extern void AxXtDestroyWidget();
extern void paintCompositeWidget();
extern void AxDrawBevel();
extern AXWidget topWidget();
extern pixcache_t *Pixmapper();
#if AX_X11
extern KeySym AxControlKeyDown;
#endif

extern void *AxRenderInsetGraphic();
extern void xmSetWidgetDisplay();
extern int AXmNewFootprint();
AXWidget WidgetPointerFromHandleData();
void PaintCanvas();
int CanvasKeyStroke();
void CanvasUpdateScrollbars();
static void RealizeCanvas();
static void CanvasButtonEventCB();

/* autoscroll vars CanvasArmed, canvasAutoscrollEdge
 */
extern AXWidget_canvas CanvasArmed;
enum canvasEdge {
	noEdge = 0, leftEdge = 1, topEdge = 2, rightEdge = 3, bottomEdge = 4,
};
static enum canvasEdge canvasAutoscrollEdge;
static enum canvasEdge mouseOutsideCanvas();

/*************************************************************************
function: XmCreateCanvas - Create rows/column table
*/
AXWidget_canvas XmCreateCanvas (parentWidget,name,args,cnt)
register AXWidget_dialog parentWidget;
register char *name;
register Arg args[];
register int cnt;
{
    register AXWidget_canvas w;
    register canvas_info *cP;

    if ( args[0].value ) {
        w = (AXWidget_canvas)args[0].value;
        if (AXmNewFootprint(w, name, args, cnt)) {
            UnMapWidget(w);
            if ( w->vScrollBar ) {
                UnMapWidget(w->vScrollBar);
                xmSetWidgetDisplay(w->vScrollBar, FALSE);
            }
            if ( w->hScrollBar ) {
                UnMapWidget(w->hScrollBar);
                xmSetWidgetDisplay(w->hScrollBar, FALSE);
            }
        }
        else {
            if ( w->vScrollBar ) {
                xmSetWidgetDisplay(w->vScrollBar, TRUE);
			}
            if ( w->hScrollBar ) {
                xmSetWidgetDisplay(w->hScrollBar, TRUE);
			}
        }
        w->update_widget = TRUE;
        WidgetNeedsUpdated = TRUE;
    }
    else {
        w = (AXWidget_canvas)CrtWidget(W_CANVAS,parentWidget,name);
        w->info.canvas = (dummy_info *)calloc(1,sizeof(canvas_info));
        cP = (canvas_info *) w->info.canvas;
        cP->paintWidth = 7500;  /* 100 inch default */
        cP->paintHeight = 7500; /* 100 inch default */
        cP->incX = 25;  /* default */
        cP->incY = 25;  /* default */
        cP->use_grid = FALSE;  /* default */
        cP->supress_copy = FALSE;  /* default */
        w->thickness = -2;     /* recessed */
    }

    w->painter = PaintCanvas;
    AxXtSetValues(w,args,cnt);
    return((AXWidget_canvas)w);
}

/*************************************************************************
function: PaintCanvas - Respond to an exposure
*/
void PaintCanvas (w, exP, regen, layered)
register AXWidget_canvas w;
XEvent *exP;    /* NULL or exposure event */
int regen;      /* what does this flag do? */
int layered;    /* what does this mean */
{
    int thickness;
    canvas_info *cP;

    if ( !w->display || !WIN_OF(w) )
        return;

    if (w->parent && (w->parent->type == W_DIALOG) &&
        !w->parent->window) {
        return;
    }

    if (layered) {
        if (w->flags & WFLG_IN_EDIT) {
            if (dePaintWidget(w)) {
                return;
            }
        }
        else {
            if (topWidget(w) == (AXWidget)w) {
                paintCompositeWidget(w, exP, regen);
                return;
            }
        }
    }

    if ( !w->mapped )
        w->mapped = TRUE;

    if ( !w->window )
        RealizeCanvas(w,layered);   /* create windows, scrolls... */
    
    CanvasUpdateScrollbars(w, layered);
	
	if (w->graphic)
		AxRenderInsetGraphic(w, w->graphic);

	if (w->def_cursor)
		AXDefineDefaultCursor(w, w->def_cursor);

    if ( !(exP || regen)) {
            /* when no real expose or full regen, the only things
             * we might've needed to do in this painter were to call
             * the scrollbars' painters and/or do resource creation
             */
        return;
    }

	if (regen) {
/* Draw Canvas border !!1 Draws border on ALL exposes! */
		if ( w->thickness < 0 ) {
			thickness = -w->thickness;
			AxDrawBevel(w, WIN_OF(w), w->xpos, w->ypos, w->width, w->height,
						thickness, BEVEL_IN);
		}
		else if ( w->thickness ) {
			thickness = w->thickness;
			AxDrawBevel(w, WIN_OF(w), w->xpos, w->ypos, w->width, w->height,
						thickness, BEVEL_OUT);
		}
	}

	if (exP) {
/* Paint in children */
		AXmExposeChildren(w, exP, regen, layered);
		
		if ( w->callback && !(w->flags&WFLG_IN_EDIT) ) {
			CanvasEventData data;
			data.expose.type = EV_EXPOSE_;
			data.expose.x = exP->xexpose.x;
			data.expose.y = exP->xexpose.y;
			data.expose.width = exP->xexpose.width;
			data.expose.height = exP->xexpose.height;
			data.expose.count = exP->xexpose.count;
			AxCallback(w->callback,w,w->callbacktag,&data);
		}
	}
}



#if AXMS   /*   {  */
static void PaintCanvasNow( AXWidget w, int xpos, int ypos, 
                                                       int width, int height )
{
    XExposeEvent expose;

    expose.type = Expose;
    expose.x = xpos;
    expose.y = ypos;
    expose.width =  width;
    expose.height =  height;
    PaintCanvas( w, &expose, 0, TRUE );
    return;
}
#endif   /*  }  */

getCanvasXPos(wc, newleft, direction)
canvas_info *wc;
int newleft;
int direction;
{
	int x, n, idx, ix;
	int grid;

	grid = wc->hgrid_size/2;
	if ( direction ) {
		n = wc->paintWidth /wc->hgrid_size;
		idx = wc->leftPoint/wc->hgrid_size;
		x = idx*wc->hgrid_size;

		for (ix = idx; ix < n; ix++ ) {
			if ( (x + grid) > newleft )
				return(x);
			x += wc->hgrid_size;
			if ( x >= wc->paintWidth )
				break;
		}
	} else {
		idx = wc->leftPoint/wc->hgrid_size;
		x = idx*wc->hgrid_size;

		for (ix = idx; ix > 0; ix-- ) {
			if ( (x - grid ) < newleft )
				return(x);
			x -= wc->hgrid_size;
			if ( x <= 0 )
				break;
		}
	}
	return(x);
}

getCanvasYPos(wc, newtop, direction)
canvas_info *wc;
int newtop;
int direction;
{
	int y, n, idx, ix;
	int grid;

	grid = wc->vgrid_size/2;
	if ( direction ) {
		n = wc->paintHeight /wc->vgrid_size;
		idx = wc->topPoint/wc->vgrid_size;
		y = idx*wc->vgrid_size;

		for (ix = idx; ix < n; ix++ ) {
			if ( (y + grid) > newtop )
				return(y);
			y += wc->vgrid_size;
			if ( y >= wc->paintHeight )
				break;
		}
	} else {
		idx = wc->topPoint/wc->vgrid_size;
		y = idx*wc->vgrid_size;

		for (ix = idx; ix > 0; ix-- ) {
			if ( (y - grid ) < newtop )
				return(y);
			y -= wc->vgrid_size;
			if ( y <= 0 )
				break;
		}
	}
	return(y);
}

CanvasScrollNewLeftPos(widget, w, newleft, do_expose)
AXWidget widget;
AXWidget_canvas w;
int newleft;
int do_expose;
{
    register canvas_info *wc = (canvas_info *) w->info.canvas;
	int oldleft, src_x, src_y, dst_x, dst_y;
	int wid, hyt, xpos, ypos;

	if ( wc->leftPoint == newleft )
		return(!AOK);
	oldleft = wc->leftPoint;

	if ((newleft + wc->page_incX) > wc->paintWidth )
		newleft = wc->paintWidth - wc->page_incX;
	else if ( newleft < 0 )
		newleft = 0;
	if (ABS(newleft - oldleft) >= wc->page_incX ) {
		/* results in a page scroll */
		wc->leftPoint = newleft;
		if ( wc->supress_copy == FALSE )
			XClearArea(Dpy,w->window,0,0,wc->winWidth,wc->winHeight,True);
	} else {
		if ( newleft > oldleft ) { /* shift left */
			src_x = getCanvasXPos(wc, newleft, TRUE);
			src_y = 0;
			dst_x = 0;
			dst_y = src_y;
			wid   = wc->winWidth -src_x;
			hyt   = wc->winHeight;

			if ( wc->leftPoint != src_x )
			{
				xpos = wc->winWidth-src_x;
				ypos = 0;
				wid = src_x;
				hyt = wc->winHeight;
				if ( wc->supress_copy == FALSE )
				{
					XCopyArea(Dpy, w->window, w->window, ForeGC, src_x, src_y,
								   wid, hyt, dst_x, dst_y);
					XClearArea(Dpy,w->window,xpos, ypos, wid, hyt, do_expose);
				}
			}
			wc->leftPoint = newleft;
		} else { /* shift right */
			src_x = 0;
			src_y = 0;
			dst_y = src_y;
			dst_x = getCanvasXPos(wc, newleft, FALSE);
			wid   = wc->winWidth -dst_x;
			hyt   = wc->winHeight;
			if ( wc->leftPoint != dst_x )
			{
				xpos = 0;
				ypos = 0;
				wid = dst_x;
				hyt = wc->winHeight;
				if ( wc->supress_copy == FALSE )
				{
					XCopyArea(Dpy, w->window, w->window, ForeGC, src_x, src_y,
								   wid, hyt, dst_x, dst_y);
					XClearArea(Dpy,w->window,xpos, ypos, wid, hyt, do_expose);
				}
			}
            wc->leftPoint = newleft;
		}
	}
	return(AOK);
}

CanvasScrollNewTopPos(widget, w, newtop, do_expose)
AXWidget widget;
AXWidget_canvas w;
int newtop;
int do_expose;
{
    register canvas_info *wc = (canvas_info *) w->info.canvas;
	int oldtop, old_pos, src_x, src_y, dst_x, dst_y;
	int wid, hyt, xpos, ypos, ix;

	if ( wc->topPoint == newtop )
		return(!AOK);
	oldtop = wc->topPoint;

	if ((newtop + wc->page_incY) > wc->paintHeight )
		newtop = wc->paintHeight - wc->page_incY;
	else if ( newtop < 0 )
		newtop = 0;
	if (ABS(newtop - oldtop) >= wc->page_incY ) {
		/* results in a page scroll */
		wc->topPoint = newtop;
		if ( wc->supress_copy == FALSE )
			XClearArea(Dpy,w->window,0,0,wc->winWidth,wc->winHeight,True);
	} else {
		if ( newtop > oldtop ) { /* shift up */
			src_x = 0;
			src_y = getCanvasYPos(wc, newtop, TRUE);
			dst_x = src_x;
			dst_y = 0;
			wid   = wc->winWidth;
			hyt   = wc->winHeight- src_y;
			ix    = (oldtop+wc->vgrid_size/2)/wc->vgrid_size;
			old_pos = ix*wc->vgrid_size;

			if ( old_pos != src_y )
			{
				xpos = 0;
				ypos = wc->winHeight-src_y;
				wid = wc->winWidth;
				hyt = src_y;
				if ( wc->supress_copy == FALSE )
				{
					XCopyArea(Dpy, w->window, w->window, ForeGC, src_x, src_y,
								   wid, hyt, dst_x, dst_y);
					XClearArea(Dpy,w->window,xpos, ypos, wid, hyt, do_expose);
				}
			}
			wc->topPoint = newtop;
		} else { /* shift down */
			src_x = 0;
			src_y = 0;
			dst_x = src_x;
			dst_y = getCanvasYPos(wc, newtop, FALSE);
			wid   = wc->winWidth;
			hyt   = wc->winHeight- dst_y;
			ix    = oldtop/wc->vgrid_size;
			if ( oldtop < (ix*wc->vgrid_size + wc->vgrid_size/2 ) &&
			   ( oldtop > (ix*wc->vgrid_size - wc->vgrid_size/2 )))
				old_pos = ix*wc->vgrid_size;
		   else
		   {
				ix    = (oldtop + wc->vgrid_size/2)/wc->vgrid_size;
				old_pos = ix*wc->vgrid_size;
		   }

			if ( old_pos != dst_y )
			{
				xpos = 0;
				ypos = 0;
				wid = wc->winWidth;
				hyt = dst_y;
				if ( wc->supress_copy == FALSE )
				{
					XCopyArea(Dpy, w->window, w->window, ForeGC, src_x, src_y,
								   wid, hyt, dst_x, dst_y);
					XClearArea(Dpy,w->window,xpos, ypos, wid, hyt, do_expose);
				}
			}
            wc->topPoint = newtop;
		}
	}
	return(AOK);
}

/*************************************************************************
function: CanvasBarChanged - Callback routine for Canvass
*/
static void CanvasBarChanged (widget,w,data)
AXWidget widget;
AXWidget_canvas w;
XmScrollBarCallbackStruct *data;
{
    register canvas_info *wc = (canvas_info *) w->info.canvas;
    register int oldTop = wc->topPoint, oldLeft = wc->leftPoint;
    register int span;
    int do_expose = TRUE;
    int xpos, ypos, width, height, newtop, err;
#if AXMS
    int paint_now = FALSE;
#endif

    if ( w->vScrollBar == (AXWidget_scroll)widget ) {
		if ( wc->use_grid ) {
			switch ( data->reason ) {
			case XmCR_INCREMENT:
				newtop = wc->topPoint + wc->incY;
				newtop = MIN(newtop,wc->paintHeight-wc->winHeight);
				break;
			case XmCR_DECREMENT:
				newtop = wc->topPoint - wc->incY;
				newtop = MAX(0,newtop);
				break;
			case XmCR_PAGE_INCREMENT:
				newtop = wc->topPoint + wc->page_incY;
				newtop = MIN(newtop,wc->paintHeight-wc->winHeight);
				break;
			case XmCR_PAGE_DECREMENT:
				newtop = wc->topPoint - wc->page_incY;
				newtop = MAX(0,newtop);
				break;
			case XmCR_VALUE_CHANGED:
				newtop = data->value;
				break;
			}
			err = CanvasScrollNewTopPos(widget, w, newtop, do_expose);
			if (err != AOK )
				return;
#if AXMS
			else
			{
				/* paint now to get active scroll */
				paint_now = TRUE;
			}
#endif
		}
		else {
			switch ( data->reason ) {
			case XmCR_INCREMENT:
				wc->topPoint += wc->incY;
				wc->topPoint = MIN(wc->topPoint,wc->paintHeight-wc->winHeight);
				wc->topPoint = MAX(0,wc->topPoint);
				span = wc->topPoint - oldTop;
				if ( span == 0 )
					return;
				xpos = 0;
				ypos = wc->winHeight-span;
				width = wc->winWidth;
				height = span;
				if ( wc->supress_copy == FALSE )
				{
					XCopyArea(Dpy,w->window,w->window,ForeGC,
							0,span,wc->winWidth,wc->winHeight-span,0,0);
					XClearArea(Dpy, w->window, xpos, ypos,width, height, do_expose);
				}
#if AXMS
				/* paint now to get active scroll */
				paint_now = TRUE;
#endif
				break;
			case XmCR_DECREMENT:
				wc->topPoint -= wc->incY;
				wc->topPoint = MAX(0,wc->topPoint);
				span = oldTop - wc->topPoint;
				if ( span == 0 )
					return;
				xpos = ypos = 0;
				width = wc->winWidth;
				height = span;
				if ( wc->supress_copy == FALSE )
				{
					XCopyArea(Dpy,w->window,w->window,ForeGC,
							0,0,wc->winWidth,wc->winHeight-span,0,span);
					XClearArea(Dpy,w->window, xpos, ypos, width, height, do_expose);
				}
#if AXMS
				/* paint now to get active scroll */
				paint_now = TRUE;
#endif
				break;
			case XmCR_PAGE_INCREMENT:
				wc->topPoint += wc->page_incY;
				wc->topPoint = MIN(wc->topPoint,wc->paintHeight-wc->winHeight);
				wc->topPoint = MAX(0,wc->topPoint);
				span = oldTop - wc->topPoint;
				if ( span == 0 )
					return;
				if ( wc->supress_copy == FALSE )
					XClearArea(Dpy,w->window,0,0,wc->winWidth, wc->winHeight,True);
				break;
			case XmCR_PAGE_DECREMENT:
				wc->topPoint -= wc->page_incY;
				wc->topPoint = MAX(0,wc->topPoint);
				span = oldTop - wc->topPoint;
				if ( span == 0 )
					return;
				if ( wc->supress_copy == FALSE )
					XClearArea(Dpy,w->window,0,0,wc->winWidth, wc->winHeight,True);
				break;
			case XmCR_VALUE_CHANGED:
				/* scroll bar was dragged */
				wc->topPoint = data->value;
				span = wc->topPoint - oldTop;
				if ( span > 0 ) {
					if ( wc->supress_copy == FALSE )
						XCopyArea(Dpy,w->window,w->window,ForeGC,
								0,span,wc->winWidth,wc->winHeight-span,0,0);
					xpos = 0;
					ypos = wc->winHeight-span;
					width = wc->winWidth;
					height = span;
					if ( wc->supress_copy == FALSE )
						XClearArea(Dpy,w->window,xpos,
								   ypos,width,height,do_expose);
#if AXMS
					/* paint now to get active scroll */
					paint_now = TRUE;
#endif
				}
				else if ( span < 0 ) {
					span = -span;
					if ( wc->supress_copy == FALSE )
						XCopyArea(Dpy,w->window,w->window,ForeGC,
								0,0,wc->winWidth,wc->winHeight-span,0,span);
					xpos = ypos = 0;
					width = wc->winWidth;
					height = span;
					if ( wc->supress_copy == FALSE )
						XClearArea(Dpy,w->window,xpos,ypos,width,height,do_expose);
#if AXMS
					/* paint now to get active scroll */
					paint_now = TRUE;
#endif
				}
				else if ( span == 0 )
					return;
				break;
			}
        }
    }
    else if ( w->hScrollBar == (AXWidget_scroll)widget ) {
		if ( wc->use_grid ) {
			switch ( data->reason ) {
			case XmCR_INCREMENT:
				newtop = wc->leftPoint + wc->incX;
				break;
			case XmCR_DECREMENT:
				newtop = wc->leftPoint - wc->incX;
				break;
			case XmCR_PAGE_INCREMENT:
				newtop = wc->leftPoint + wc->page_incX;
				break;
			case XmCR_PAGE_DECREMENT:
				newtop = wc->leftPoint - wc->page_incX;
				break;
			case XmCR_VALUE_CHANGED:
				newtop = data->value;
				break;
			}
			err = CanvasScrollNewLeftPos(widget, w, newtop, do_expose);
			if (err != AOK )
				return;
#if AXMS
			else
			{
				/* paint now to get active scroll */
				paint_now = TRUE;
			}
#endif
		} else {
			switch ( data->reason ) {
			case XmCR_INCREMENT:
				wc->leftPoint += wc->incX;
				wc->leftPoint = MIN(wc->leftPoint,wc->paintWidth-wc->winWidth);
				span = wc->leftPoint - oldLeft;
				if ( span == 0 )
					return;
				xpos = wc->winWidth-span;
				ypos = 0;
				width = span;
				height = wc->winHeight;
				if ( wc->supress_copy == FALSE )
				{
					XCopyArea(Dpy,w->window,w->window,ForeGC,
							span,0,wc->winWidth-span,wc->winHeight,0,0);
					XClearArea(Dpy, w->window, xpos,
									ypos, width,height, do_expose);
				}
#if AXMS
				/* paint now to get active scroll */
				paint_now = TRUE;
#endif
				break;
			case XmCR_DECREMENT:
				wc->leftPoint -= wc->incX;
				wc->leftPoint = MAX(0,wc->leftPoint);
				span = oldLeft - wc->leftPoint;
				if ( span == 0 )
					return;
				xpos = ypos = 0;
				width = span;
				height = wc->winHeight;
				if ( wc->supress_copy == FALSE )
				{
					XCopyArea(Dpy,w->window,w->window,ForeGC,
								0,0,wc->winWidth-span,wc->winHeight,span,0);
					XClearArea(Dpy,w->window, xpos,ypos,width,height,do_expose);
				}
#if AXMS
				/* paint now to get active scroll */
				paint_now = TRUE;
#endif
				break;
			case XmCR_PAGE_INCREMENT:
				wc->leftPoint += wc->page_incX;
				wc->leftPoint = MIN(wc->leftPoint,wc->paintWidth-wc->winWidth);
				if ( wc->supress_copy == FALSE )
					XClearArea(Dpy,w->window,0,0,wc->winWidth,
							   wc->winHeight,True);
				break;
			case XmCR_PAGE_DECREMENT:
				wc->leftPoint -= wc->page_incX;
				wc->leftPoint = MAX(0,wc->leftPoint);
				if ( wc->supress_copy == FALSE )
					XClearArea(Dpy,w->window,0,0,wc->winWidth,
							   wc->winHeight,True);
				break;
			case XmCR_VALUE_CHANGED:
				wc->leftPoint = data->value;
				span = wc->leftPoint - oldLeft;
				if ( span > 0 ) {
					if ( wc->supress_copy == FALSE )
						XCopyArea(Dpy,w->window,w->window,ForeGC,
								span,0,wc->winWidth-span,wc->winHeight,0,0);
					xpos = wc->winWidth-span;
					ypos = 0;
					width = span;
					height = wc->winHeight;
					if ( wc->supress_copy == FALSE )
						XClearArea(Dpy,w->window,xpos,
								   ypos,width,height,do_expose);
#if AXMS
					/* paint now to get active scroll */
					paint_now = TRUE;
#endif
				}
				else if ( span < 0 ) {
					span = -span;
					if ( wc->supress_copy == FALSE )
						XCopyArea(Dpy,w->window,w->window,ForeGC,
								0,0,wc->winWidth-span,wc->winHeight,span,0);
					xpos = ypos = 0;
					width = span;
					height = wc->winHeight;
					if ( wc->supress_copy == FALSE )
						XClearArea(Dpy,w->window,xpos,
								   ypos,width,height,do_expose);
#if AXMS
					/* paint now to get active scroll */
					paint_now = TRUE;
#endif
				}
				else if ( span == 0 )
					return;
				break;
			}
        }
    }
    if ( w->callback ) {
        CanvasEventData data;
        data.scrolled.type = EV_NEW_SCROLL_POS_NOTIFY_;
        data.scrolled.topPoint = wc->topPoint;
        data.scrolled.leftPoint = wc->leftPoint;
        AxCallback(w->callback,w,w->callbacktag,&data);
    }

#if AXMS
    if ( paint_now )
        PaintCanvasNow( w, xpos, ypos, width, height );
#endif
}

/*************************************************************************
function: RealizeCanvas - Create window & scroll bars
*/
static void RealizeCanvas (w,layered)
register AXWidget_canvas w;
int layered;    /* what does this mean */
{
    register canvas_info *wc = (canvas_info *) w->info.canvas;
    int x,y,win_xpos,win_ypos,width,height;
    unsigned long mask;
    XSetWindowAttributes atts;
    GC gc,backgc;
    int thickness = w->thickness;

    if ( w->thickness < 0 )
        thickness = -thickness;

    x = w->xpos;
    y = w->ypos;
    win_xpos = x + thickness;
    win_ypos = y + thickness;
    width = w->width;
    height = w->height;
    wc->winWidth = width - 2*thickness;
    wc->winHeight = height - 2*thickness;
    wc->page_incX = wc->winWidth;
    wc->page_incY = wc->winHeight;

    mask = WinMask;
    atts = WinAtt;
    if ( w->useWorkColors ) {
        mask |= CWBackPixel | CWColormap;
        atts.background_pixel = BackWorkPixel;
        atts.colormap = WorkColormap;
    }
    if (w->flags & WFLG_IN_EDIT) { /* dialog editor */
        w->window = AxCreateQueryPixmap(Dpy, w->parent->window,
                        wc->winWidth, wc->winHeight, DpyDepth);
        XFillRectangle(Dpy, w->window, BackGC, 0, 0,
                        wc->winWidth, wc->winHeight);
    }
    else {
#if AXMS
        AxWinCreateCanvasWindow(w, win_xpos, win_ypos,
                            wc->winWidth, wc->winHeight);
#else
        w->window = AxCrtWindow(Dpy,WIN_OF(w),
            win_xpos, win_ypos, wc->winWidth, wc->winHeight,
                    0,CopyFromParent,InputOutput,
                    CopyFromParent,mask,&atts);

        XSelectInput(Dpy, w->window, TM_ALL_INPUTS_MASK);
        XMapWindow(Dpy,w->window);
#endif
        AddToWindowList(w);
    }

    if ( w->useWorkColors ) {
        SetWorkColors(BackGC, 0);
        SetWorkColors(ForeGC, 1);
        SetWorkColors(DisabledGC, 1);
    }

/* clear gap between border and window
    (should be AFTER GCs changed for work area colors) */

    XFillRectangle(Dpy,w->window,BackGC,0,0,wc->winWidth,wc->winHeight);

    if ( w->useWorkColors ) {
        UnSetWorkColors(BackGC, 0);
        UnSetWorkColors(ForeGC, 1);
        UnSetWorkColors(DisabledGC, 1);
    }

    CanvasUpdateScrollbars(w, layered);

    if (w->flags & WFLG_IN_EDIT) {
        XCopyArea(Dpy, w->window, w->parent->window, ForeGC, 0, 0,
                    wc->winWidth, wc->winHeight, win_xpos, win_ypos);
    }
}

/**************************************************************
 * function: CanvasUpdateScrollbars - create/destroy/paint scroll bars
 */
void
CanvasUpdateScrollbars (w, layered)
AXWidget_canvas w;
int layered;
{
    register canvas_info *wc = (canvas_info *) w->info.canvas;

    /* NOTE: scrollbar parent MUST be dialog to ensure proper
     *       behavior when task of listbox is pended */
    if ( !w->hide_vscroll) {
            /* widget should have vertical scrollbar */
        Arg args[10];
        int n = 0;
        int offset = 2;
        static XtCallbackRec callback_list[2];

        if ( w->thickness == 0 )
            offset = 0;

        CALLBACKER(callback_list,CanvasBarChanged,w);
        XtSetArg(args[n],XmNx,w->xpos+w->width+offset); n++;
        XtSetArg(args[n],XmNy,w->ypos); n++;
        XtSetArg(args[n],XmNheight,w->height); n++;
        XtSetArg(args[n],XmNorientation,XmVERTICAL); n++;
        XtSetArg(args[n],XmNvalueChangedCallback,callback_list); n++;
        if ( w->vScrollBar ) {
            AxXtSetValues(w->vScrollBar, args, n);
        }
        else {
            w->vScrollBar = (AXWidget_scroll) AXmCreateScrollBar(w->parent,
                                "x",args,n);
            if (w->event_table) {
                    /* give the newly created scrollbar an event table
                     * which matches that of the list widget */
                copyEventHandlerList(w->event_table,
                                     &w->vScrollBar->event_table);
            }
                /* dlg editor support */
            XtSetArg(args[0], ZXmNcompositeLinkUp, w);
            XtSetArg(args[1], ZXmNinEdit, w->flags & WFLG_IN_EDIT);
            AxXtSetValues(w->vScrollBar, args, 2);
            XtSetArg(args[0], ZXmNcompositeLinkDown, w->vScrollBar);
            AxXtSetValues(w, args, 1);
            xmSetWidgetDisplay(w->vScrollBar, TRUE);
        }
        w->vScrollBar->scroll_per = wc->winHeight;
        w->vScrollBar->scroll_max = wc->paintHeight;
        w->vScrollBar->scroll_pos = wc->topPoint;
        PaintScrollBar(w->vScrollBar, NULL, FALSE, layered);
    }
    else if (w->vScrollBar) {
        AxXtDestroyWidget(w->vScrollBar);
        w->vScrollBar = NULL;
    }

    if ( w->list_hscroll ) {
            /* widget should have horizontal scrollbar */
        Arg args[10];
        int n = 0;
        static XtCallbackRec hcallback_list[2];
        int offset = 2;

        if ( w->thickness == 0 )
            offset = 0;
        CALLBACKER(hcallback_list,CanvasBarChanged,w);
        XtSetArg(args[n],XmNx,w->xpos); n++;
        XtSetArg(args[n],XmNy,w->ypos+w->height+offset); n++;
        XtSetArg(args[n],XmNwidth,w->width); n++;
        XtSetArg(args[n],XmNorientation,XmHORIZONTAL); n++;
        XtSetArg(args[n],XmNvalueChangedCallback,hcallback_list); n++;
        if ( w->hScrollBar ) {
            AxXtSetValues(w->hScrollBar, args, n);
        }
        else {
            w->hScrollBar = (AXWidget_scroll)AXmCreateScrollBar(w->parent,
                                "y",args,n);
            if (w->event_table) {
                    /* give the newly created scrollbar an event table
                     * which matches that of the list widget */
                copyEventHandlerList(w->event_table,
                                     &w->hScrollBar->event_table);
            }
                /* dlg editor support */
            XtSetArg(args[0], ZXmNcompositeLinkUp, w);
            XtSetArg(args[1], ZXmNinEdit, w->flags & WFLG_IN_EDIT);
            AxXtSetValues(w->hScrollBar, args, 2);
            XtSetArg(args[0], ZXmNcompositeLinkDown, w->hScrollBar);
            AxXtSetValues(w, args, 1);
            xmSetWidgetDisplay(w->hScrollBar, TRUE);
        }
        w->hScrollBar->scroll_per = wc->winWidth;
        w->hScrollBar->scroll_max = wc->paintWidth;
        w->hScrollBar->scroll_pos = wc->leftPoint;
        PaintScrollBar(w->hScrollBar,NULL, FALSE, layered);
    }
    else if (w->hScrollBar) {
        AxXtDestroyWidget(w->hScrollBar);
        w->hScrollBar = NULL;
    }
}

/* Handle all types of XEvents in the table widget.
   Return 1 if event was consumed, 0 otherwise. */
CanvasEvent(w, evP)
register AXWidget_canvas w;
register XEvent *evP;
{
    int event_consumed = 0;
    int keysym;
    CanvasEventData data;

    if (!w || !evP)
        return(0); /* invalid args */
    if (w->type != W_CANVAS)
        return(0); /* invalid args */
    if (!w->info.canvas)
        return(0); /* invalid args */

    switch (evP->xany.type) {
    case Expose:
        if ( w->callback ) {
            data.expose.type = EV_EXPOSE_;
            data.expose.x = evP->xexpose.x;
            data.expose.y = evP->xexpose.y;
            data.expose.width = evP->xexpose.width;
            data.expose.height = evP->xexpose.height;
            data.expose.count = evP->xexpose.count;
            AxCallback(w->callback,w,w->callbacktag,&data);
        }
        event_consumed = TRUE;
        break;
	case FocusIn:
        if ( w->callback ) {
            data.any.type = EV_FOCUS_IN_;
            AxCallback(w->callback,w,w->callbacktag,&data);
        }
        event_consumed = TRUE;
		break;
	case FocusOut:
        if ( w->callback ) {
            data.any.type = EV_FOCUS_OUT_;
            AxCallback(w->callback,w,w->callbacktag,&data);
        }
        event_consumed = TRUE;
		break;
    case ButtonPress:
        CanvasButtonEventCB(w,evP,EV_BUTTON_PRESS_);
        event_consumed = TRUE;
        break;
    case ButtonRelease:
		CanvasButtonEventCB(w,evP,EV_BUTTON_RELEASE_);
        event_consumed = TRUE;
        break;
    case MotionNotify:
		CanvasButtonEventCB(w,evP,EV_MOTION_);
        event_consumed = TRUE;
        break;
    case KeyPress:
        event_consumed = TRUE;
        if (keysym = CanvasKeyStroke(w, evP)) {
            if (keysym == XK_Escape)        /* ESC is never consumed */
                event_consumed = FALSE;
            else if ( w->callback ) {
                data.key.type = EV_TYPING_;
                data.key.keycode = keysym;
                data.key.shifted = (ThimIsShifted()? BTRUE:BFALSE);
                data.key.controlled = (ThimIsControled()? BTRUE:BFALSE);
                AxCallback(w->callback,w,w->callbacktag,&data);
            }
        }
        break;
    case KeyRelease:
        event_consumed = FALSE; /* !! */
        break;
    default:
        event_consumed = FALSE;
        break;
    }
    return(event_consumed);
}


/* : CanvasKeyStroke - filter XKeyEvent; check whether
 *                     key should invoke a menubar action,
 *                     and filter into X keysym form
 */
int   /* return nonzero ksym if a real key, and not consumed by mb */
CanvasKeyStroke(w, evP)
AXWidget w;
XEvent *evP;
{
    int is_text, bell, nbytes;
    char *name;
    KeySym ksym;
    wchar *kstring;
    menu_item_def md;
    menu_item_def *mP = (menu_item_def *) NULL;

    /* check menubar first */
    if ( w->parent->parent->menuBar )
        mP = w->parent->parent->menuBar->popups;

    if (ProcessKeyEvent(mP, evP ,&is_text, &name, &bell) && !is_text) {

        if ( !AxIsCommandAvailable(w->parent->parent, name) ) {
            return FALSE;  /* "consumed" by menubar */
        }
        
        if ( w->parent->parent->menuBar->callback ) {
            md.word2 = name;    /* md for ElfXCmenuBarSelected() */
            AxCallback(w->parent->parent->menuBar->callback,
                       w->parent->parent->menuBar,
                       w->parent->parent->menuBar->callbacktag,&md);
        }
        return FALSE;   /* consumed by menu bar */
    }

        /* convert into x keysym form */
    nbytes = AXLookupString16(evP, &kstring, &ksym);
    
#if AX_X11
    if ( (ksym == XK_Control_L) || (ksym == XK_Control_R) ) {
        if ( !AxControlKeyDown )
            AxControlKeyDown = ksym;
    }
#endif
    if ( (nbytes == 0) && AxKeysymIsModifier(ksym) ) {
        /* Ignore non-keys like caps lock */
        return(FALSE);
    }

    return ((int) ksym);
}

static void
CanvasButtonEventCB (w, evP, type)
register AXWidget_canvas w;
register XEvent *evP;
int type;
{
    unsigned long time;
    CanvasEventData data;

	if ( !w || !w->callback )
		return;

	data.button.type = type;
	if (type == EV_BUTTON_PRESS_) {
#if LINUX
	if (evP->xbutton.button == Button4 || evP->xbutton.button == Button5) {
		char *str = (char *)UPbyName(NXLT("axMouseWheelScroll"));
		XmScrollBarCallbackStruct cbdata;
		if (!str || (atoi(str) == 0)) {
			return;
		}
		
		if (evP->xbutton.button == Button4) { /* mouse scroll up */
			if (ThimIsShifted())
				cbdata.reason = XmCR_PAGE_DECREMENT;
			else
				cbdata.reason = XmCR_DECREMENT;
		} else if (evP->xbutton.button == Button5) { /* mouse scroll down */
			if (ThimIsShifted())
				cbdata.reason = XmCR_PAGE_INCREMENT;
			else
				cbdata.reason = XmCR_INCREMENT;
		}
		CanvasBarChanged(w->vScrollBar, w, &cbdata);
		return;
	}
#endif
		time = evP->xbutton.time;
		if (time - w->last_click_time <= DoubleMillisec)
			data.button.type = EV_DBLCLICK_;
		w->last_click_time = time;
	}
	else if (type == EV_BUTTON_RELEASE_) {
		CanvasArmed = (AXWidget_canvas) NULL;
	}

	if (type == EV_MOTION_) {
		data.button.which = AxButtonDown;
		if (Button1 == AxButtonDown) {
			if (!CanvasArmed) {
				/* test whether to start autoscroll
				 */
				canvasAutoscrollEdge = mouseOutsideCanvas(w, evP);
				if (noEdge != canvasAutoscrollEdge) {
					CanvasArmed = w;    /* start autoscroll */
				}
			}
			else if (CanvasArmed) {
				/* test whether to stop autoscroll
				 */
				canvasAutoscrollEdge = mouseOutsideCanvas(w, evP);
				if (noEdge == canvasAutoscrollEdge) {
					CanvasArmed = (AXWidget_canvas) NULL;
				}
			}
		}
	}
	else {
		data.button.which = (((int)evP->xbutton.button) & 0xFF);
	}

	data.button.x = evP->xbutton.x;
	data.button.y = evP->xbutton.y;
	data.button.shifted = (ThimIsShifted()? BTRUE:BFALSE);
	data.button.controlled = (ThimIsControled()? BTRUE:BFALSE);
	AxCallback(w->callback,w,w->callbacktag,&data);
}

void
CanvasAutoScroll(w)
AXWidget_canvas w;
{
	XmScrollBarCallbackStruct cbdata;
	XEvent fake_event;
	Window root_return, child_return;
	struct canvas_info *cp;
	
	cp = (canvas_info *) w->info.canvas;
	memset((void *) &cbdata, 0, sizeof(XmScrollBarCallbackStruct));
	
	if (XQueryPointer(Dpy, w->window, &root_return, &child_return,
					  &fake_event.xbutton.x_root,
					  &fake_event.xbutton.y_root,
					  &fake_event.xbutton.x, &fake_event.xbutton.y,
					  &fake_event.xbutton.state)) {
		if (fake_event.xbutton.state & Button1Mask) {
			canvasAutoscrollEdge = mouseOutsideCanvas(w, &fake_event);
			if (noEdge == canvasAutoscrollEdge) {
				CanvasArmed = (AXWidget_canvas) NULL;
				return;
			}
		}
	}
	
	switch(canvasAutoscrollEdge) {
	case leftEdge:
		if (w->hScrollBar) {
			if (cp && cp->leftPoint > 0) {
				cbdata.reason = XmCR_DECREMENT;
				CanvasBarChanged(w->hScrollBar, w, &cbdata);
			}
		}
		break;
	case rightEdge:
		if (w->hScrollBar) {
			if (cp && cp->leftPoint + cp->winWidth < cp->paintWidth) {
				cbdata.reason = XmCR_INCREMENT;
				CanvasBarChanged(w->hScrollBar, w, &cbdata);
			}
		}
		break;
	case topEdge:
		if (w->vScrollBar) {
			if (cp && cp->topPoint > 0) {
				cbdata.reason = XmCR_DECREMENT;
				CanvasBarChanged(w->vScrollBar, w, &cbdata);
			}
		}
		break;
	case bottomEdge:
		if (w->vScrollBar) {
			if (cp && cp->topPoint + cp->winHeight < cp->paintHeight) {
				cbdata.reason = XmCR_INCREMENT;
				CanvasBarChanged(w->vScrollBar, w, &cbdata);
			}
		}
		break;
	}
}

/* : mouseOutsideCanvas - return whether passed mouse
 * event, assumed to be for the canvas window, is outside
 * the canvas' footprint.  Can return "noEdge", meaning
 * mouse event is inside canvas footprint
 */
static enum canvasEdge
mouseOutsideCanvas(w, eventP)
AXWidget_canvas w;
XEvent *eventP;
{
	enum canvasEdge whichedge;
	int fudge = 5;
	
	if (eventP->xbutton.x < -fudge) {
		whichedge = leftEdge;
	}
	else if (eventP->xbutton.x > (int) w->width + fudge) {
		whichedge = rightEdge;
	}
	else if (eventP->xbutton.y < -fudge) {
		whichedge = topEdge;
	}
	else if (eventP->xbutton.y > (int) w->height + fudge) {
		whichedge = bottomEdge;
	}
	else {
		whichedge = noEdge;
	}

	return whichedge;
}

