/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmchoose.c
   Applix ASC version: 22
   Date: Fri Feb  5 21:11:57 1999
*/
#include <math.h>
#include <basicwin.h>
#if AX_X11
#include <X11/keysym.h>
#endif
#include <ax_xm.h>
#include <aos.h>
#include <dwt2xm.h>
#include <xmlib.h>
#include <elxwins.h>
#include <wchars.h>
#include <thimble.h>
#include <ctype.h>
#include <elf_env.h>
#include <eldata.h>
#include <evreason.h>
#include <xmpixel.h>
#include <xmwcolor.h>
#include <xmchoose.h> /* Private color chooser info. */
#include <intsarr.h>

/* Standard Externs for Widget Modules */
extern Window AxCrtWindow();
extern AXWidget CrtWidget();
extern pixcache_t *Pixmapper();
extern void AxDrawBevel();

extern int WidgetNeedsUpdated; /* Set when a widget needs repainted */
extern int AxCharAbove, AxCharHeight, AxCharBelow;
extern int AxOpenLook;
extern int AxBigFont;
extern int UseWidgetColors;
extern Display *Dpy;
extern short DpyDepth;
extern int AxButtonDown;
extern short THIMpid;

extern long WinMask;
extern XSetWindowAttributes WinAtt;
extern Pixmap UnknownPixmap;
extern AXWidget ClickWidget;
extern unsigned long BackWorkPixel;
extern Colormap WorkColormap;
extern GC ForeGC, BackGC;
extern GC SunGC, ShadowGC, DisabledGC, GrayLineGC;
extern GC RecessGC;
extern GC olClrSunButtonGC, olClrShadowButtonGC, olClrRecessButtonGC;
extern GC MonospaceGC;		 /* monspace system font, forground mode */
extern GC MonospaceBackGC;	 /* monspace system font, background mode */
extern GC MonospaceBoldGC, MonospaceBoldBackGC;
extern GC MonospaceItalGC, MonospaceItalBackGC;
extern GC MonospaceItalBoldGC, MonospaceItalBoldBackGC;
extern XFontStruct *AxDlgMainFont;	/* Bolded dialog font */
extern XFontStruct *MonospaceFont, *MonospaceBoldFont;
extern XFontStruct *MonospaceItalFont, *MonospaceItalBoldFont;
extern int DoubleMillisec;
extern unsigned long ForePixel;
extern unsigned long BackPixel;
extern short ColorDisplay;
extern short NoColor;
extern Window RootWin;
extern Cursor ThimCrosshairCursor;
extern Cursor ThimPointerCursor;
extern Cursor ThimLeftRightCursor;
extern unsigned long AlisPixels[];
extern unsigned long WidgetPixels[];
/* End of Standard Externs */

#if !BUILD_WIDE /* move to xmlib.h!! */
#define axDrawString16 axDrawString
#endif

/* Switch this ON for testing colored shadows -- The sample rectangle
   is painted with computed sun and shadow pixels. */
#define TEST_COLOR_SHADOW 0

/* Static decls. */
static void CcSetWorkColors(/* AXWidget_color_chooser w */);
static void CcUnsetWorkColors(/* AXWidget_color_chooser w */);
static void CcInitColors(/* AXWidget_color_chooser w */);
static void CcChangeBrightness(/* AXWidget_color_chooser w, float pct */);
static void CcColorButton(/* AXWidget w, Window win, GC gc, int x, int y
							 int width, int height, int thickness,
							 unsigned long pixel */);
static void CcColorRect(/* Window win, GC gc, int x, int y, int width,
						   int height, int thickness,
						   unsigned long pixel */);
static void CcSendValues();
static void CcColorButtonAndShadow();
static void CcChangeHue();
static void CcChangeSaturation();
static void CcGetCellPosAndDim();
static void CcPaintSampleRect();
static void CcPaintPadSelector();
static void CcPaintPad();
static void CcPadMotionNotify();
static void CcPadButtonPress();
static void CcPadButtonRelease();
static void CcPaintPadRegion();
static void CcDrawSliderBarBevels();
static void CcSliderExpose();
static void CcSliderSetNewValue();
static void CcSliderButtonPress();
static void CcSliderButtonRelease();
static void CcSliderMotionNotify();
static void CcFreeSliderData();
static void CcSliderXlateCoords();
static void CcInitLabelFont();
static void CcDrawString();
static int wCcSliderWidth();
static int CcPointInSlider();
static int CcFindCell(), CcGetWcolor();
static int CcRectIntersect();
static int CcPointInPad();
static CcSlider *CcNewSlider();
/* End of Static decls. */

typedef struct {
	int x, y, xcor, ycor;
} CcRect;

#define REGOVERLAP(r1, r2) (((r1)->xcor > (r2)->x) && \
							((r1)->x    < (r2)->xcor) && \
							((r1)->ycor > (r2)->y) && \
							((r1)->y    < (r2)->ycor))

static float Hvalues[MAX_H_LEV] = {
	0.000000,
	10.000000,
	20.000000,
	30.000000,
	40.000000,
	50.000000,
	60.000000,
	70.000000,
	80.000000,
	90.000000,
	100.000000,
	110.000000,
	120.000000,
	130.000000,
	140.000000,
	150.000000,
	160.000000,
	170.000000,
	180.000000,
	190.000000,
	200.000000,
	210.000000,
	220.000000,
	230.000000,
	240.000000,
	250.000000,
	260.000000,
	270.000000,
	280.000000,
	290.000000,
	300.000000,
	310.000000,
	320.000000,
	330.000000,
	340.000000,
	350.000000,
};

/* Color pad (leftmost) */
static int pad_left_offset	= 4;
static int pad_right_offset = 4;
static int pad_top_offset = 16;
static int pad_bottom_offset = 4;
/* pad xpos and ypos are left_offset and top_offset */

/* HSB sliders (to the right of the pad) */
static int hsb_sld_left_offset = 4;
static int hsb_sld_right_offset = 4;
static int hsb_sld_top_offset = 4;
static int hsb_sld_bottom_offset = 4;
static int hsb_sld_xpos; /* computed */
static int hsb_sld_ypos; /* computed */

static XFontStruct *LabelFont;

/* Sample rect. (below the hsb sliders) */
#define SAMPLE_RECT_HEIGHT 40 /* sample button has const height;
								 computed x, y, and width */
static int sample_rect_top_offset = 4;
static int sample_rect_bottom_offset = 4;

static int cell_dim = 12; /* dimension of color cell */

/***********************************************************************
function: AXmCreateColorChooser - Create a color chooser widget.
*/
Widget
AXmCreateColorChooser(parentWidget, name, args, cnt)
register AXWidget_dialog parentWidget;
register char *name;
register Arg args[];
register int cnt;
{
	register AXWidget_color_chooser w;
	register int already_exists = FALSE;

	if (args[0].value) {
		w = (AXWidget_color_chooser) args[0].value;
		if (args[1].value & WFLG_NODISPLAY)
			UnMapWidget(w);
		w->update_widget = TRUE;
		WidgetNeedsUpdated = TRUE;
		already_exists = TRUE;
	} else {
		w = (AXWidget_color_chooser)CrtWidget(W_CHOOSE, parentWidget, name);
		w->info.color_chooser = (dummy_info *) calloc(1, sizeof(chooser_data));
	}
	w->painter = PaintColorChooser;
	AxXtSetValues(w, args, cnt);
	if (already_exists) {
		if (w->widgetColor && WCSPACE_HSB == w->widgetColor->any.type) {
			chooser_data *cP = (chooser_data *) w->info.color_chooser;
			if (cP) {
				/* update state of exisiting chooser based on current colorval
				 * (should convert for other wcspace types, however lossy..)
				 */
				CcSliderSetNewValue(w, Dpy, w->window, cP->v_slider,
									(int) w->widgetColor->hsb.b, FALSE);
				CcSliderSetNewValue(w, Dpy, w->window, cP->h_slider,
									(int) w->widgetColor->hsb.h, FALSE);
				CcSliderSetNewValue(w, Dpy, w->window, cP->s_slider,
									(int) w->widgetColor->hsb.s, FALSE);
				
				CcChangeBrightness(w, w->widgetColor->hsb.b);
				CcChangeHue(w, w->widgetColor->hsb.h);
				CcChangeSaturation(w, w->widgetColor->hsb.s);
			}
		}
	}
	else {
		CcInitColors(w);
		CcInitLabelFont();
	}
	return((Widget) w);
}

/***********************************************************************
function: FreeChooserData - Free all data inside CP and then CP itself
NOTE: Caller must set CP to NULL for sanity.
*/
void
FreeChooserData(cP)
chooser_data *cP;
{
	if (!cP)
		return;
	if (cP->h_slider)
		CcFreeSliderData(cP->h_slider);
	cP->h_slider = NULL;
	if (cP->s_slider)
		CcFreeSliderData(cP->s_slider);
	cP->s_slider = NULL;
	if (cP->v_slider)
		CcFreeSliderData(cP->v_slider);
	cP->v_slider = NULL;
	free(cP);
	return;
}

/***********************************************************************
function: CcKeyPressCB - Handle a KeyPress event in the color chooser.
Return TRUE if event was consumed;	FALSE otherwise.
*/
static int
CcKeyPressCB(w, ev)
AXWidget_color_chooser w;
XEvent *ev;
{
	register int ix;
	KeySym ksym;
	XComposeStatus *status = (XComposeStatus *) NULL;
	int nb;
	char buf[100];
	int event_consumed = FALSE;

	nb = XLookupString(&ev->xkey, buf, 20, &ksym, status);
	switch (ksym) {
	case XK_Up:
	case XK_Down:
	default:
		event_consumed = FALSE;
		break;
	}
	if (event_consumed)
		PaintColorChooser(w, ev, TRUE, TRUE);
	return(event_consumed);
}

/***********************************************************************
function: CcPaintSampleRect - Paint the sample color-rectangle.
*/
static void
CcPaintSampleRect(w)
register AXWidget_color_chooser w;
{
	chooser_data *cP = NULL;
	unsigned long pixel;
	unsigned long saved_fg, saved_bg;
	float h, s, v;
	int r, g, b;

	if (!w)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;

	h = cP->h_slider->value;
	s = cP->s_slider->value;
	v = cP->v_slider->value;

	HSBToRGB(&r, &g, &b, h, s, v);
	AxGetRealPixel(0, r, g, b, 0, 255, &pixel);
	saved_fg = AXGetForegroundPixel(Dpy, ForeGC);
	saved_bg = AXGetBackgroundPixel(Dpy, ForeGC);

#if TEST_COLOR_SHADOW
	XSetForeground(Dpy, ForeGC, pixel);
	XSetBackground(Dpy, ForeGC, pixel);
	XFillRectangle(Dpy, w->window, ForeGC, cP->sample_x, cP->sample_y,
				   cP->sample_width, cP->sample_height, 2, pixel);
#else
	CcColorButton(w, w->window, ForeGC, cP->sample_x, cP->sample_y,
				  cP->sample_width, cP->sample_height, 2, pixel);
#endif

	XSetForeground(Dpy, ForeGC, saved_fg);
	XSetBackground(Dpy, ForeGC, saved_bg);

	
#if TEST_COLOR_SHADOW
	{
		wcolor_t color;
		extern void AxDrawColorShadow();

		color.any.type = WCSPACE_HSB;
		color.hsb.h = h;
		color.hsb.s = s;
		color.hsb.b = v;
		AxDrawColorShadow(w, w->window, SunGC, ShadowGC, 4,cP->sample_x,
						  cP->sample_y, cP->sample_width,
						  cP->sample_height, OLSHADOW_DRAW, &color,
						  NULL, NULL);
	}
#endif
	return;
}

/***********************************************************************
function: CcPaintPadSelector - Paint the selector.
*/
static void
CcPaintPadSelector(w)
register AXWidget_color_chooser w;
{
	chooser_data *cP = NULL;
	int x, y, width, height;
	unsigned long pixel;
	unsigned long saved_fg, saved_bg;

	if (!w)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;

	CcGetCellPosAndDim(w, cP->selected_cell_idx, &x, &y, &width, &height);
	AxDrawBevel((AXWidget) NULL, w->window, x, y, width, height, -2, BEVEL_IN);
	
	/* update sample color rectangle */
	CcPaintSampleRect(w);
	return;
}

/***********************************************************************
function: CcSendValues - Send new values to the elf layer.
*/
static void
CcSendValues(w)
register AXWidget_color_chooser w;
{
	ChooserEventData data;
	chooser_data *cP = NULL;
	float hue, saturation, brightness;
	int red, green, blue;
	if (!w)
		return;
	if (!w->callback)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;
	hue = cP->h_slider->value;
	saturation = cP->s_slider->value;
	brightness = cP->v_slider->value;
	HSBToRGB(&red, &green, &blue, hue, saturation, brightness);
	data.value.red = red;
	data.value.green = green;
	data.value.blue = blue;
	data.value.hue = hue;
	data.value.saturation = saturation;
	data.value.brightness = brightness;
	AxCallback(w->callback, w, w->callbacktag, &data);
	return;
}

/***********************************************************************
function: CcPadButtonPress - Handle a buttonpress in the color pad.
*/
static void
CcPadButtonPress(w, evp)
register AXWidget_color_chooser w;
register XEvent *evp;
{
	register chooser_data *cP = NULL;
	int idx;
	int bx, by;
	int new_value;

	bx = evp->xbutton.x;
	by = evp->xbutton.y;
	if (!w || !evp)
		return;
	if (w->type != W_CHOOSE)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;
	idx = CcFindCell(w, bx, by);
	if (idx == -1)
		goto finish;
	cP->selected_cell_idx = idx;

	/* set new value on 'H' bar. */
	new_value = (int) cP->colors[idx].h;
	CcSliderSetNewValue(w, Dpy, w->window, cP->h_slider, new_value, TRUE);
	/* set new value on 'S' bar. */
	new_value = (int) cP->colors[idx].c;
	CcSliderSetNewValue(w, Dpy, w->window, cP->s_slider, new_value, TRUE);

	CcPaintPad(w); /* FIXME, lazy paint */
finish:
	cP->pad_armed = TRUE;
	return;
}

/***********************************************************************
function: CcPadButtonRelease - Handle a buttonrelease in the color pad.
*/
static void
CcPadButtonRelease(w, evp)
register AXWidget_color_chooser w;
register XEvent *evp;
{
	register chooser_data *cP = NULL;
	if (!w || !evp)
		return;
	if (w->type != W_CHOOSE)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;
	cP->pad_armed = FALSE;
	return;
}

/***********************************************************************
function: CcPadMotionNotify - Handle a motionnotify in the color pad.
*/
static void
CcPadMotionNotify(w, evp)
register AXWidget_color_chooser w;
register XEvent *evp;
{
	register chooser_data *cP = NULL;
	int idx;
	int mx, my;
	int new_value;

	mx = evp->xmotion.x;
	my = evp->xmotion.y;
	if (!w || !evp)
		return;
	if (w->type != W_CHOOSE)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;
	if (!cP->pad_armed)
		return;
	idx = CcFindCell(w, mx, my);
	if (idx == -1)
		goto finish;
	if (idx == cP->selected_cell_idx)
		goto finish;
	cP->selected_cell_idx = idx;

	/* set new value on 'H' bar. */
	new_value = (int) cP->colors[idx].h;
	CcSliderSetNewValue(w, Dpy, w->window, cP->h_slider, new_value, TRUE);
	/* set new value on 'S' bar. */
	new_value = (int) cP->colors[idx].c;
	CcSliderSetNewValue(w, Dpy, w->window, cP->s_slider, new_value, TRUE);

	CcPaintPad(w); /* FIXME, lazy paint */
finish:
	return;
}

/***********************************************************************
function: ColorChooserEvent - Handle all types of XEvents in the color
chooser widget.  Return TRUE if event was consumed;  FALSE otherwise
(or on error).
*/
int
ColorChooserEvent(w, evp)
register AXWidget_color_chooser w;
register XEvent *evp;
{
	int event_consumed = FALSE;

	if (!w || !evp)
		return(FALSE);
	if (w->type != W_CHOOSE)
		return(FALSE);

	switch (evp->xany.type) {
	case ButtonPress: {
		chooser_data *cP = NULL;
		int x, y;

		cP = (chooser_data *) w->info.color_chooser;
		if (!cP)
			return;
		if (!evp)
			return;

		x = evp->xbutton.x;
		y = evp->xbutton.y;

		if (AxButtonDown != Button1)
			goto finish;
		if (CcPointInPad(w, x, y)) {
			cP->pad_armed = TRUE;
			CcPadButtonPress(w, evp);
			event_consumed = TRUE;
		} else if (CcPointInSlider(cP->h_slider, x, y, FALSE)) {
			CcSliderButtonPress(w, Dpy, w->window, evp, cP->h_slider);
			/* button-press was in the recessed area of bar. */
			if (!cP->h_slider->armed)
				CcSendValues(w);
			event_consumed = TRUE;
		} else if (CcPointInSlider(cP->s_slider, x, y, FALSE)) {
			CcSliderButtonPress(w, Dpy, w->window, evp, cP->s_slider);
			/* button-press was in the recessed area of bar. */
			if (!cP->s_slider->armed)
				CcSendValues(w);
			event_consumed = TRUE;
		} else if (CcPointInSlider(cP->v_slider, x, y, FALSE)) {
			CcSliderButtonPress(w, Dpy, w->window, evp, cP->v_slider);
			/* button-press was in the recessed area of bar. */
			if (!cP->v_slider->armed)
				CcSendValues(w);
			event_consumed = TRUE;
		}
		break;
	}
	case ButtonRelease: {
		chooser_data *cP = NULL;

		cP = (chooser_data *) w->info.color_chooser;
		if (!cP)
			return;
		if (!evp)
			return;

		if (cP->pad_armed) {
			CcPadButtonRelease(w, evp);
			event_consumed = TRUE;
		} else if (cP->h_slider && cP->h_slider->armed) {
			CcSliderButtonRelease(w, Dpy, w->window, evp, cP->h_slider);
			event_consumed = TRUE;
		} else if (cP->s_slider && cP->s_slider->armed) {
			CcSliderButtonRelease(w, Dpy, w->window, evp, cP->s_slider);
			event_consumed = TRUE;
		} else if (cP->v_slider && cP->v_slider->armed) {
			CcSliderButtonRelease(w, Dpy, w->window, evp, cP->v_slider);
			event_consumed = TRUE;
		}
		if (event_consumed)
			CcSendValues(w);
		break;
	}
	case KeyPress:
		CcKeyPressCB(w, evp);
		break;
	case MotionNotify: {
		chooser_data *cP = NULL;

		cP = (chooser_data *) w->info.color_chooser;
		if (!cP)
			return;
		if (!evp)
			return;

		if (AxButtonDown != Button1)
			goto finish;
		if (cP->pad_armed) {
			CcPadMotionNotify(w, evp);
			event_consumed = TRUE;
		} else if (cP->h_slider && cP->h_slider->armed) {
			CcSliderMotionNotify(w, Dpy, w->window, evp, cP->h_slider);
			event_consumed = TRUE;
		} else if (cP->s_slider && cP->s_slider->armed) {
			CcSliderMotionNotify(w, Dpy, w->window, evp, cP->s_slider);
			event_consumed = TRUE;
			goto finish;
		} else if (cP->v_slider && cP->v_slider->armed) {
			CcSliderMotionNotify(w, Dpy, w->window, evp, cP->v_slider);
			event_consumed = TRUE;
		}
		break;
	}
	default:
		break;
	}
finish:
	return(event_consumed);
}


/***********************************************************************
function: RealizeColorChooser - Create color chooser window.
*/
static void
RealizeColorChooser(w, layered)
register AXWidget_color_chooser w;
int layered;
{
	int xpos, ypos, win_xpos, win_ypos;
	int width, height, win_width, win_height;
	unsigned long mask;
	XSetWindowAttributes atts;
	int thickness;
	chooser_data *cP = NULL;
	int sliders_start_x, sliders_end_x;

	if (!w)
		return;
	if (w->type != W_CHOOSE)
		return;
	if (!WIN_OF(w))
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;

	/* Set widget position/dimension */
	thickness = w->thickness;
	if (w->thickness < 0)
		thickness = -thickness;
	xpos = w->xpos;
	ypos = w->ypos;
	win_xpos = xpos + thickness;
	win_ypos = ypos + thickness;
	width  = w->width;
	height = w->height;
	win_width  = width - (2 * thickness);
	win_height = height - (2 * thickness);
	cP->win_width  = win_width;
	cP->win_height = win_height;

	mask = WinMask;
	atts = WinAtt;
	/* Create window */
	if (w->useWorkColors) {
		mask |= (CWBackPixel | CWColormap);
		atts.background_pixel = BackWorkPixel;
		atts.colormap = WorkColormap;
	}
	if (w->flags & WFLG_IN_EDIT) { /* dialog editor */
		w->window = AxCreateQueryPixmap(Dpy, w->parent->window,
										win_width, win_height, DpyDepth);
		XFillRectangle(Dpy, w->window, BackGC, 0, 0, win_width, win_height);
	} else {
#if AXMS
		AxCreateWinChooserWindow(w);
#else
		w->window = AxCrtWindow(Dpy, WIN_OF(w), win_xpos, win_ypos,
								win_width, win_height, 0, CopyFromParent,
								InputOutput, CopyFromParent, mask, &atts);
		XSelectInput(Dpy, w->window, (TM_ALL_INPUTS_MASK) | Button1MotionMask);
		XMapWindow(Dpy, w->window);
#endif
		AddToWindowList(w);
	}

	/* Clear gap between border and window */
	CcSetWorkColors(w);
	XFillRectangle(Dpy, w->window, BackGC, 0, 0, win_width, win_height);
	CcUnsetWorkColors(w);

	/* Set default cursor */
	AXDefineDefaultCursor(w, ThimPointerCursor);

	{
		int x, y;
		int min_value, max_value, step_value, value;
		float hue_value, saturation_value, brightness_value;

		AxGetHSBFromWColor(w->widgetColor, &hue_value, &saturation_value,
						   &brightness_value);
		/* H bar */
		x = pad_left_offset + (MAX_H_LEV * cell_dim) + 
			hsb_sld_left_offset;
		sliders_start_x = x;
		y = pad_top_offset;
		min_value = 0;
		max_value = 360;
		step_value = 1;
		value = (int) hue_value;
		cP->h_slider = CcNewSlider(w, "H", SLIDER_VERT, x, y,
								   (MAX_C_LEV * cell_dim),
								   min_value, max_value, step_value, value);
		cP->h_slider->value_changed_cb = CcChangeHue;
		cP->h_slider->closure = (caddr_t) w;
		CcChangeHue(w, hue_value);
		x += wCcSliderWidth();

		/* S bar */
		min_value = 0;
		max_value = 100;
		step_value = 1;
		value = saturation_value;
		cP->s_slider = CcNewSlider(w, "S", SLIDER_VERT, x, y,
								   (MAX_C_LEV * cell_dim),
								   min_value, max_value, step_value, value);
		cP->s_slider->value_changed_cb = CcChangeSaturation;
		cP->s_slider->closure = (caddr_t) w;
		CcChangeSaturation(w, saturation_value);
		x += wCcSliderWidth();

		/* V bar */
		min_value = 0;
		max_value = 100;
		step_value = 1;
		value = brightness_value;
		cP->v_slider = CcNewSlider(w, "B", SLIDER_VERT, x, y,
								   (MAX_C_LEV * cell_dim),
								   min_value, max_value, step_value, value);
		CcChangeBrightness(w, brightness_value);
		cP->v_slider->value_changed_cb = CcChangeBrightness;
		cP->v_slider->closure = (caddr_t) w;
		x += wCcSliderWidth();

		/* more bars... */
		sliders_end_x = x;
	}
#if 0 /* small sample rect */
	cP->sample_x = sliders_start_x;
	cP->sample_y = cP->h_slider->y + cP->h_slider->height
		+ sample_rect_top_offset + 20; /* FIXME, fudge for value strings. */
	cP->sample_height = SAMPLE_RECT_HEIGHT;
	cP->sample_width = sliders_end_x - sliders_start_x;
	cP->sample_width -= hsb_sld_right_offset; /* so that it does not
												 touch the border */
#else /* long sample rect */
	cP->sample_x = pad_left_offset;
	cP->sample_y = cP->h_slider->y + cP->h_slider->height
		+ sample_rect_top_offset + 20; /* FIXME, fudge for value strings. */
	cP->sample_height = SAMPLE_RECT_HEIGHT;
	cP->sample_width = sliders_end_x - pad_left_offset;
	cP->sample_width -= hsb_sld_right_offset; /* so that it does not
												 touch the border */
#endif
	return;
}

/***********************************************************************
function: CcSetWorkColors - Set work colors on GCs (if necessary).
*/
static void
CcSetWorkColors(w)
register AXWidget_color_chooser w;
{
	if (!w)
		return;
	if (w->useWorkColors) {
		SetWorkColors(BackGC, 0);
		SetWorkColors(ForeGC, 1);
		SetWorkColors(DisabledGC, 1);
	}
	return;
}

/***********************************************************************
function: CcUnsetWorkColors - Unset work colors on GCs.
*/
static void
CcUnsetWorkColors(w)
register AXWidget_color_chooser w;
{
	if (!w)
		return;
	if (w->useWorkColors) {
		UnSetWorkColors(BackGC, 0);
		UnSetWorkColors(ForeGC, 1);
		UnSetWorkColors(DisabledGC, 1);
	}
	return;
}

/***********************************************************************
function: CcColorButton - Paint a color button.
Note: This clobbers foreground and background pixel on GC.	It is the
caller's responsibility to restore.
*/
static void
CcColorButton(w, win, gc, x, y, width, height, thickness, pixel)
AXWidget w;
Window win;
GC gc;
int x, y;
int width, height;
int thickness;
unsigned long pixel;
{
	wcolor_t *saveColor, currentColor;
	
	XSetForeground(Dpy, gc, pixel);
	XSetBackground(Dpy, gc, pixel);
	XFillRectangle(Dpy, win, gc, x, y, width, height);
	
	/* set the w->widgetColor member temporarily in order
	 * to paint the button's bevel
	 */
	saveColor = w->widgetColor;
	CcGetWcolor(w, &currentColor);
	w->widgetColor = &currentColor;
	
	AxDrawBevel(w, win, x, y, width, height, thickness, BEVEL_OUT);
	
	w->widgetColor = saveColor;    /* restore */
	return;
}

/***********************************************************************
function: CcColorRect - Fill a rectangle with color
Note: This clobbers foreground and background pixel on GC.	It is the
caller's responsibility to restore.
*/
static void
CcColorRect(win, gc, x, y, width, height, thickness, pixel)
Window win;
GC gc;
int x, y;
int width, height;
int thickness;
unsigned long pixel;
{
	XSetForeground(Dpy, gc, pixel);
	XSetBackground(Dpy, gc, pixel);
	XFillRectangle(Dpy, win, gc, x, y, width, height);
	return;
}

/***********************************************************************
function: CcInitColors - Initialize the color table from template.
Also init pad dimensions.
*/
static void
CcInitColors(w)
AXWidget_color_chooser w;
{
	chooser_data *cp = NULL;
	unsigned long pixel;
	register int h, c;
	float thisH;
	float maxC;
	float thisC;
	float stepC;
	float thisV;
	int ix;

	if (!w)
		return;
	cp = (chooser_data *) w->info.color_chooser;
	if (!cp)
		return;
	for (ix = 0, h = 0; h < MAX_H_LEV; h++) {
		thisH = Hvalues[h];
		maxC = 100.0;
		stepC = maxC / MAX_C_LEV;
		thisC = 0.0; /* thisC = stepC; */
		thisV = 100.0;
		for (c = 0; c < MAX_C_LEV; c++) {
			int r, g, b;

			cp->colors[ix].h = thisH;
			cp->colors[ix].v = thisV;
			cp->colors[ix].c = thisC;
			HSBToRGB(&r, &g, &b, thisH, thisC, thisV);
			cp->colors[ix].r = r;
			cp->colors[ix].g = g;
			cp->colors[ix].b = b;
			AxGetRealPixel(0, r, g, b, 0, 255, &pixel);
			cp->colors[ix].pixel = pixel;
			ix++;
			thisC += stepC;
		}
	}
	cp->pad_x = pad_left_offset;
	cp->pad_y = pad_top_offset;
	cp->pad_width = (MAX_H_LEV * cell_dim);
	cp->pad_height = (MAX_C_LEV * cell_dim);
	return;
}

/***********************************************************************
function: CcChangeHue - Change Hue.
*/
static void
CcChangeHue(w, new_hue)
AXWidget_color_chooser w;
float new_hue;
{
	register int ix;
	register int max = (MAX_H_LEV * MAX_C_LEV);
	int idx;
	chooser_data *cP = NULL;
	float cur_sat;

	if (!w)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;
	idx = -1;
	cur_sat = cP->colors[cP->selected_cell_idx].c;
	for (ix = 0; ix < max; ix++) {
		if (cP->colors[ix].c != cur_sat)
			continue;
		if (cP->colors[ix].h > new_hue)
			continue;
		if (idx == -1) {
			idx = ix;
			continue;
		}
		if (cP->colors[ix].h > cP->colors[idx].h)
			idx = ix;
	}
	if (idx == -1)
		return;
	cP->selected_cell_idx = idx;
	return;
}

/***********************************************************************
function: CcChangeSaturation - Change Saturation
*/
static void
CcChangeSaturation(w, new_sat)
AXWidget_color_chooser w;
float new_sat;
{
	register int ix;
	register int max = (MAX_H_LEV * MAX_C_LEV);
	int idx;
	chooser_data *cP = NULL;
	float cur_hue;

	if (!w)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;
	idx = -1;
	cur_hue = cP->colors[cP->selected_cell_idx].h;
	for (ix = 0; ix < max; ix++) {
		if (cP->colors[ix].h != cur_hue)
			continue;
		if (cP->colors[ix].c > new_sat)
			continue;
		if (idx == -1) {
			idx = ix;
			continue;
		}
		if (cP->colors[ix].c > cP->colors[idx].c)
			idx = ix;
	}
	if (idx == -1)
		return; /* no change */
	cP->selected_cell_idx = idx;
	return;
}

/***********************************************************************
function: CcChangeBrightness - Apply NEW_B to all colors.
*/
static void
CcChangeBrightness(w, new_b)
AXWidget_color_chooser w;
float new_b;
{
	chooser_data *cp = NULL;
	float maxV = 100.0;
	float pct;
	unsigned long pixel;
	register int ix;
	register int max = (MAX_H_LEV * MAX_C_LEV);

	pct = 0.01 * new_b;
	if (!w)
		return;
	if (pct < 0.0)
		pct = 0.0;
	if (pct > 1.0)
		pct = 1.0;
	cp = (chooser_data *) w->info.color_chooser;
	if (!cp)
		return;
	for (ix = 0; ix < max; ix++) {
		float thisV, thisH, thisC;
		int r, g, b;

		thisV = maxV * pct;
		thisH = cp->colors[ix].h;
		thisC = cp->colors[ix].c;
		HSBToRGB(&r, &g, &b, thisH, thisC, thisV);
		cp->colors[ix].r = r;
		cp->colors[ix].g = g;
		cp->colors[ix].b = b;

		AxGetRealPixel(0, r, g, b, 0, 255, &pixel);
		cp->colors[ix].pixel = pixel;
	}
	return;
}

/***********************************************************************
function: CcFindCell - Return the cell idx that contains the point (x, y);
-1 on failure.
*/
static int
CcFindCell(w, x, y)
AXWidget_color_chooser w;
int x, y;
{
	register chooser_data *cP = NULL;
	register int ix, c, h;
	int cx, cy, cw, ch;

	if (!w)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;
	cy = cP->pad_y;
	cx = cP->pad_x;
	cw = cell_dim;
	ch = cell_dim;

	for (ix = 0, h = 0; h < MAX_H_LEV; h++) {
		for (c = 0; c < MAX_C_LEV; c++) {
			if (cx <= x && 
				(cx + cw) >= x &&
				cy <= y &&
				(cy + ch) >= y)
				return(ix);
			cy += ch;
			ix++;
		}
		cy = cP->pad_y;
		cx += cw;
	}
	return(-1);
}

/***********************************************************************
function: CcGetCellPosAndDim - Get the position and dimension of cell
at IDX.
*/
static void
CcGetCellPosAndDim(w, idx, xp, yp, wp, hp)
AXWidget_color_chooser w;
int idx;
int *xp, *yp, *wp, *hp;
{
	int x, y, width, height;
	chooser_data *cP = NULL;
	register int ix, c, h;

	if (!w || !xp || !yp || !wp || !hp)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;
	x = y = width = height = 0;
	x = cP->pad_x;
	y = cP->pad_y;
	width = cell_dim;
	height = cell_dim;
	for (ix = 0, h = 0; h < MAX_H_LEV; h++) {
		for (c = 0; c < MAX_C_LEV; c++) {
			y += height;
			if (ix == idx) {
				*xp = x;
				*yp = y;
				*wp = width;
				*hp = height;
				return;
			}
			ix++;
		}
		y = cP->pad_y;
		x += width;
	}
	/* err */
	*xp = 0;
	*yp = 0;
	*wp = 0;
	*hp = 0;
	return;
}

/***********************************************************************
function: CcPaintPad - Paint the color pad.
*/
static void
CcPaintPad(w)
AXWidget_color_chooser w;
{
	chooser_data *cP = NULL;
	unsigned long saved_fg;
	unsigned long saved_bg;
	register int ix, h, c;
	int cx, cy, cw, ch;

	if (!w)
		return;
	if (w->type != W_CHOOSE)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;

	cy = cP->pad_y;
	cx = cP->pad_x;
	cw = cell_dim;
	ch = cell_dim;
	saved_fg = AXGetForegroundPixel(Dpy, ForeGC);
	saved_bg = AXGetBackgroundPixel(Dpy, ForeGC);
	for (ix = 0, h = 0; h < MAX_H_LEV; h++) {
		for (c = 0; c < MAX_C_LEV; c++) {
			unsigned long pixel;

			pixel = cP->colors[ix].pixel;
			CcColorRect(w->window, ForeGC, cx, cy, cw, ch, 2, pixel);
			if (ix == cP->selected_cell_idx) {
				AxDrawBevel((AXWidget) NULL, w->window, cx, cy, cw, ch,
							2, BEVEL_IN);
			}
			cy += ch;
			ix++;
		}
		cy = cP->pad_y;
		cx += cw;
	}
	XSetForeground(Dpy, ForeGC, saved_fg);
	XSetBackground(Dpy, ForeGC, saved_bg);
	CcPaintSampleRect(w);
	return;
}

/***********************************************************************
function: CcPaintPadRegion - Paint a region enclosed in (x, y, width, height)
of the color pad.
*/
static void
CcPaintPadRegion(w, x, y, width, height)
AXWidget_color_chooser w;
int x, y, width, height;
{
	chooser_data *cP = NULL;
	unsigned long saved_fg;
	unsigned long saved_bg;
	register int ix, h, c;
	int cx, cy, cw, ch;
	CcRect reg_rect;

	if (!w)
		return;
	if (w->type != W_CHOOSE)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;

	cy = cP->pad_y;
	cx = cP->pad_x;
	cw = cell_dim;
	ch = cell_dim;

	reg_rect.x = x;
	reg_rect.y = y;
	reg_rect.xcor = x + width;
	reg_rect.ycor = y + height;

	saved_fg = AXGetForegroundPixel(Dpy, ForeGC);
	saved_bg = AXGetBackgroundPixel(Dpy, ForeGC);

	for (ix = 0, h = 0; h < MAX_H_LEV; h++) {
		for (c = 0; c < MAX_C_LEV; c++) {
			unsigned long pixel;
			int r, g, b;
			CcRect cell_rect;

			cell_rect.x = cx;
			cell_rect.y = cy;
			cell_rect.xcor = cx + cw;
			cell_rect.ycor = cy + ch;
			if (CcRectIntersect(&cell_rect, &reg_rect)) {
				pixel = 0;
				r = cP->colors[ix].r;
				g = cP->colors[ix].g;
				b = cP->colors[ix].b;
				AxGetRealPixel(0, r, g, b, 0, 255, &pixel);
				CcColorRect(w->window, ForeGC, cx, cy, cw, ch, 2, pixel);
			}
			cy += ch;
			ix++;
		}
		cy = cP->pad_y;
		cx += cw;
	}
	XSetForeground(Dpy, ForeGC, saved_fg);
	XSetBackground(Dpy, ForeGC, saved_bg);
	return;
}

/***********************************************************************
function: CcRectIntersect - Return TRUE if rect 1 intersects rect 2;
FALSE otherwise.
*/
static int
CcRectIntersect(r1, r2)
CcRect *r1, *r2;
{
	if (REGOVERLAP(r1, r2) || REGOVERLAP(r2, r1))
		return(TRUE);
	return(FALSE);
}

/***********************************************************************
function: CcPointInPad - Return TRUE if point (x, y) lies in the color pad;
FALSE otherwise.
*/
static int
CcPointInPad(w, x, y)
AXWidget_color_chooser w;
int x, y;
{
	chooser_data *cP = NULL;
	if (!w)
		return;
	if (w->type != W_CHOOSE)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;

	if (x >= cP->pad_x &&
		x <= (cP->pad_x + cP->pad_width) &&
		y >= cP->pad_y &&
		y <= (cP->pad_y + cP->pad_height))
		return(TRUE);
	return(FALSE);
}

/***********************************************************************
function: DrawCcBorder - Draw color chooser border on parent window.
*/
void
DrawCcBorder(w)
register AXWidget_color_chooser w;
{
	int thickness;

	if (!w)
		return;
	if (w->type != W_CHOOSE)
		return;
	if (!WIN_OF(w))
		return;

	if (w->thickness < 0) {
		thickness = -w->thickness;
		AxDrawBevel((AXWidget) NULL, WIN_OF(w), w->xpos, w->ypos,
					w->width, w->height, thickness, BEVEL_IN);
	} else if (w->thickness) {
		thickness = w->thickness;
		AxDrawBevel((AXWidget) NULL, WIN_OF(w), w->xpos, w->ypos,
					w->width, w->height, thickness, BEVEL_OUT);
	}
	return;
}

/***********************************************************************
function: PaintColorChooser - Paint the color chooser.
*/
void
PaintColorChooser(w, exP, regen, layered)
register AXWidget_color_chooser w;
XExposeEvent *exP;
int regen;
int layered;
{
	chooser_data *cP = NULL;

	if (!w)
		return;
	if (!WIN_OF(w))
		return;
	if (w->type != W_CHOOSE)
		return;
	cP = (chooser_data *) w->info.color_chooser;
	if (!cP)
		return;
	if ((w->flags & WFLG_IN_EDIT) && layered) {
		if (dePaintWidget(w))
			return;
	}
	if (!w->mapped)
		w->mapped = TRUE;
	if (!w->window)
		RealizeColorChooser(w, layered);

	DrawCcBorder(w);
	CcPaintPad(w);
/* Draw sliders (if present) */
	if (cP->h_slider)
		CcSliderExpose(w, Dpy, w->window, exP, cP->h_slider);
	if (cP->s_slider)
		CcSliderExpose(w, Dpy, w->window, exP, cP->s_slider);
	if (cP->v_slider)
		CcSliderExpose(w, Dpy, w->window, exP, cP->v_slider);
	return;
}

/***********************************************************************
function: wColorChooserWidth - Return the standard width of a color
chooser widget.
*/
int
wColorChooserWidth(thickness)
int thickness;
{
	int width;
	int pad_width;
	int sliders_width;
	int single_slider_width;

	if (thickness < 0)
		thickness = -thickness;

	single_slider_width = wCcSliderWidth();
	sliders_width = 3 * single_slider_width; /* H, S, V */
	pad_width = pad_left_offset + (MAX_H_LEV * cell_dim) + pad_right_offset;
	width = pad_width + sliders_width + (2 * thickness);
	return(width);
}

/***********************************************************************
function: wColorChooserHeight - Return the standard width of a color chooser.
*/
int
wColorChooserHeight(thickness)
int thickness;
{
	int height;
	int pad_height;
	int sample_height;
	if (thickness < 0)
		thickness = -thickness;
	pad_height = pad_top_offset + (MAX_C_LEV * cell_dim) + pad_bottom_offset;
	pad_height += 20; /* FIXME, fudge so that slider values appear. */
	sample_height =  sample_rect_top_offset + SAMPLE_RECT_HEIGHT +
		sample_rect_bottom_offset;
	height = pad_height + sample_height + (2 * thickness);
	return(height);
}

/***********************************************************************
			SLIDER GADGET -- PRIVATE TO THE COLOR CHOOSER
 This implementation is a `gadgetized' variation of the slider widget in the
 tiny-toolkit.
 See "dixit:/home/sanjay/misc/app/src/stk/slider.c" for implementation.

 The scale widget (W_SCALE) could have been used here.	But W_SCALE
 can only display horizontal-scales :-(  It seemed to be much easier to
 use the `slider.c' implementation here instead of adding vertical-scales
 to W_SCALE.
*/

/***********************************************************************
function: pos_to_value - Return the VALUE corresponding to position P
in the slider SP.
*/
static int
pos_to_value(sp, p)
register CcSlider *sp;
register int p;
{
	register int v;
	register int d, l;

	d = sp->max_value - sp->min_value;
	l = sp->length - SLIDER_BARLEN - 2;
	if (p >= l) {
		v = sp->max_value;
		goto finish;
	}
	if (p <= 2) {
		v = sp->min_value;
		goto finish;
	}
	p = p - 2;
	v = (p * d) / l;
	v += sp->min_value;
	if (v > sp->max_value)
		v = sp->max_value;
	if (v < sp->min_value)
		v = sp->min_value;
finish:
	return(v);
}

/***********************************************************************
function: value_to_pos - Return the POS corresponding to value V in
the slider SP.
*/
static int
value_to_pos(sp, v)
register CcSlider *sp;
register int v;
{
	register int p;
	register int d, l;

	d = sp->max_value - sp->min_value;
	l = sp->length - SLIDER_BARLEN - 2;
	v -= sp->min_value;
	p = (l * v) / d;
	p += 2;
	if (p < 2)
		p = 0;
	if (p > l)
		p = l;
	return(p);
}

/***********************************************************************
function: wCcSliderWidth - Return width of a slider bar.
*/
static int
wCcSliderWidth()
{
	int width;
	width = hsb_sld_left_offset + SLIDER_THICKNESS + hsb_sld_right_offset;
	return(width);
}

/***********************************************************************
function: CcDrawSliderValue - Paint the slider value-string of slider
SP on window XWIN.  The string is painted on the right side for the
horizontal slider, and at the bottom for the vertical slider.  If
CLEAR_ONLY is TRUE, just blank out the value string.
*/
void
CcDrawSliderValue(dpy, xwin, sp, clear_only)
Display *dpy;
Window xwin;
CcSlider *sp;
int clear_only;
{
	int x, y;
	char str[100];
	static clear_width = 0;
	static clear_height = 0;

	if (!clear_width || !clear_height) {
		clear_width = AXTextWidth(LabelFont, NXLT("000"), 3);
		clear_height = LabelFont->max_bounds.ascent +
			LabelFont->max_bounds.descent;
	}
	if (sp->style == SLIDER_HORIZ) {
		x = sp->x + sp->width + 4;
		y = sp->y + LabelFont->max_bounds.ascent + 2;
		sprintf(str, NXLT("%d"), sp->value);
		str[4] = '\0';
		XFillRectangle(dpy, xwin, BackGC, x, sp->y, clear_width, clear_height);
	} else { /* vert */
		x = sp->x;
		y = sp->y + sp->height + 4 + LabelFont->max_bounds.ascent;
		sprintf(str, NXLT("%d"), sp->value);
		str[4] = '\0';
		XFillRectangle(dpy, xwin, BackGC, x, sp->y + sp->height + 4,
					   clear_width, clear_height);
	}
	if (clear_only)
		return;
	CcDrawString(dpy, xwin, ForeGC, x, y, str, strlen(str));
	return;
}

/***********************************************************************
function: CcDrawSliderName - Paint the slider name-string of slider SP
on window XWIN.  The string is painted on the left side for the
horizontal slider, and at the top for the vertical slider.  If
CLEAR_ONLY is TRUE, just blank out the name-string.
*/
void
CcDrawSliderName(dpy, xwin, sp, clear_only)
Display *dpy;
Window xwin;
CcSlider *sp;
int clear_only;
{
	int x, y;
	char str[50];
	static int name_width = 0;
	static int name_height = 0;

	if (!name_width || !name_height) {
		name_width = AXTextWidth(LabelFont, NXLT("X"), 1);
		name_height = LabelFont->max_bounds.ascent +
			LabelFont->max_bounds.descent;
	}
	if (sp->style == SLIDER_HORIZ) {
		x = sp->x - name_width - 2;
		y = sp->y + LabelFont->max_bounds.ascent + 2;
		XFillRectangle(dpy, xwin, BackGC, x, sp->y, name_width, name_height);
	} else { /* vert */
		x = sp->x + ((SLIDER_THICKNESS - name_width) / 2);
		y = sp->y - LabelFont->max_bounds.descent;
		XFillRectangle(dpy, xwin, BackGC, x, sp->y - name_height,
					   name_width, name_height);
	}
	if (clear_only)
		return;
	if (!sp->name || !strlen(sp->name))
		return;
	str[0] = sp->name[0];
	str[1] = '\0';
	CcDrawString(dpy, xwin, ForeGC, x, y, str, strlen(str));
	return;
}

/***********************************************************************
function: CcSliderExpose - Handle an Expose event in the slider. --
Repaint whole slider.
Note: Does not look at EV.
*/
static void
CcSliderExpose(w, dpy, win, ev, sp)
AXWidget w;
Display *dpy;
Window win;
XEvent *ev;
CcSlider *sp;
{
	int x, y, width, height;

	/* trough */
	XFillRectangle(dpy, win, RecessGC, sp->x, sp->y, sp->width,
				   sp->height);
	if (sp->style == SLIDER_HORIZ) {
		x = sp->cur_pos;
		y = 2;
		width = SLIDER_BARLEN;
		height = SLIDER_THICKNESS - 4;
	} else {
		y = sp->cur_pos;
		x = 2;
		width = SLIDER_THICKNESS - 4;
		height = SLIDER_BARLEN;
	}

	/* bar */
	XFillRectangle(dpy, win, BackGC, sp->x + x, sp->y + y, width, height);
	CcDrawSliderBarBevels(win, sp->x + x, sp->y + y,
						  width, height, sp->style);
	/* border */
	AxDrawBevel((AXWidget) NULL, win, sp->x, sp->y, sp->width,
				sp->height, 2, BEVEL_IN);
	CcDrawSliderValue(dpy, win, sp, FALSE);
	CcDrawSliderName(dpy, win, sp, FALSE);
	return;
}

/***********************************************************************
function: CcSliderSetNewValue - Set new value on slider, update cur_pos.
If REPAINT is not FALSE, repaint the slider.
*/
static void
CcSliderSetNewValue(w, dpy, win, sp, new_value, repaint)
AXWidget w;
Display *dpy;
Window win;
CcSlider *sp;
int new_value;
int repaint;
{
	if (!sp)
		return;
	if (new_value > sp->max_value)
		new_value = sp->max_value;
	if (new_value < sp->min_value)
		new_value = sp->min_value;
	sp->value = new_value;
	sp->cur_pos = value_to_pos(sp, new_value);
	if (repaint)
		CcSliderExpose(w, dpy, win, NULL, sp);
	return;
}

/***********************************************************************
function: CcSliderButtonRelease - Handle a ButtonRelease event in the slider.
*/
static void
CcSliderButtonRelease(w, dpy, win, ev, sp)
AXWidget w;
Display *dpy;
Window win;
XEvent *ev;
CcSlider *sp;
{
	int x, y, width, height;
	int bx, by;
	int old_value;

	if (!sp->armed)
		return;
	sp->armed = FALSE;
	old_value = sp->value;
	CcSliderXlateCoords(sp, ev->xbutton.x, ev->xbutton.y, &bx, &by);
	if (sp->style == SLIDER_HORIZ) {
		x = sp->cur_pos;
		y = 2;
		width = SLIDER_BARLEN;
		height = SLIDER_THICKNESS - 4;
		if (bx >= x &&
			bx <= (x + width) &&
			by >= y &&
			by <= y + height) {
			return;
		}
		bx -= (SLIDER_BARLEN / 2);
		bx = MAX(2, bx);
		if (bx > (sp->length - SLIDER_BARLEN - 2))
			bx = sp->length - SLIDER_BARLEN - 2;
		sp->cur_pos = bx;
		sp->value = pos_to_value(sp, bx);
	} else {
		y = sp->cur_pos;
		x = 2;
		width = SLIDER_THICKNESS - 4;
		height = SLIDER_BARLEN;
		if (by >= y &&
			by <= (y + height) &&
			bx >= x &&
			bx <= (x + width))
			return;
		by -= (SLIDER_BARLEN / 2);
		by = MAX(2, by);
		if (by > (sp->length - SLIDER_BARLEN - 2))
			by = sp->length - SLIDER_BARLEN - 2;
		sp->cur_pos = by;
		sp->value = pos_to_value(sp, by);
	}
	CcSliderExpose(w, dpy, win, NULL, sp);
	if (old_value != sp->value) {
		float new_value;
		new_value = sp->value * 1.0;
		CcDrawSliderValue(dpy, win, sp, FALSE);
		if (sp->value_changed_cb)
			(sp->value_changed_cb)(sp->closure, sp->value);
		CcPaintPad(sp->parent);
	}
	return;
}

/***********************************************************************
function: CcSliderMotionNotify - Handle a MotionNotify in the slider.
*/
static void
CcSliderMotionNotify(w, dpy, win, ev, sp)
AXWidget w;
Display *dpy;
Window win;
XEvent *ev;
CcSlider *sp;
{
	int x, y, width, height;
	int mx, my;
	XEvent nev;
	int old_value;

	if (!sp->armed)
		return;
	old_value = sp->value;
	CcSliderXlateCoords(sp, ev->xmotion.x , ev->xmotion.y, &mx, &my);
	if (sp->style == SLIDER_HORIZ) {
		x = sp->cur_pos;
		y = 2;
		width = SLIDER_BARLEN;
		height = SLIDER_THICKNESS - 4;
		XFillRectangle(dpy, win, RecessGC, sp->x + x, sp->y + y,
					   width, height);
		/* Draw new bar */
		mx -= (SLIDER_BARLEN / 2);
		x = MAX(2, mx);
		if (x > (sp->length - SLIDER_BARLEN - 2))
			x = sp->length - SLIDER_BARLEN - 2;
		y = 2;
		width = SLIDER_BARLEN;
		height = SLIDER_THICKNESS - 4;
		XFillRectangle(dpy, win, BackGC, sp->x + x, sp->y + y, width, height);
		CcDrawSliderBarBevels(win, sp->x + x, sp->y + y,
							  width, height, sp->style);
		sp->cur_pos = x;
		sp->value = pos_to_value(sp, x);
	} else {
		y = sp->cur_pos;
		x = 2;
		width = SLIDER_THICKNESS - 4;
		height = SLIDER_BARLEN;
		XFillRectangle(dpy, win, RecessGC, sp->x + x, sp->y + y,
					   width, height);
		/* Draw new bar */
		my -= (SLIDER_BARLEN / 2);
		y = MAX(2, my);
		if (y > (sp->length - SLIDER_BARLEN - 2))
			y = sp->length - SLIDER_BARLEN - 2;
		XFillRectangle(dpy, win, BackGC, sp->x + x, sp->y + y, width, height);
		CcDrawSliderBarBevels(win, sp->x + x, sp->y + y,
							  width, height, sp->style);
		sp->cur_pos = y;
		sp->value = pos_to_value(sp, y);
	}
	if (old_value != sp->value) {
		float new_value;
		new_value = sp->value * 1.0;
		CcDrawSliderValue(dpy, win, sp, FALSE);
		if (sp->value_changed_cb)
			(sp->value_changed_cb)(sp->closure, new_value);
		CcPaintPad(sp->parent);
	}
	return;
}

/***********************************************************************
function: CcSliderButtonPress - Handle a ButtonPress in the slider.
*/
static void
CcSliderButtonPress(w, dpy, win, ev, sp)
AXWidget w;
Display *dpy;
Window win;
XEvent *ev;
CcSlider *sp;
{
	int x, y, width, height;
	int bx, by;
	int new_value;
	int old_value;

	/* make bx and by relative to the slider */
	CcSliderXlateCoords(sp, ev->xbutton.x, ev->xbutton.y, &bx, &by);
	old_value = sp->value;
	if (sp->style == SLIDER_HORIZ) {
		x = sp->cur_pos;
		y = 2;
		width = SLIDER_BARLEN;
		height = SLIDER_THICKNESS - 4;
		if (bx >= x &&
			bx <= (x + width) &&
			by >= y &&
			by <= y + height) {
			sp->armed = TRUE;
			return;
		}
		if (bx < x) {
			new_value = sp->value - sp->step_value;
		} else {
			new_value = sp->value + sp->step_value;
		}
	} else {
		y = sp->cur_pos;
		x = 2;
		width = SLIDER_THICKNESS - 4;
		height = SLIDER_BARLEN;
		if (by >= y &&
			by <= (y + height) &&
			bx >= x &&
			bx <= (x + width)) {
			sp->armed = TRUE;
			return;
		}
		if (by < y)
			new_value = sp->value - sp->step_value;
		else
			new_value = sp->value + sp->step_value;
	}
	if (new_value > sp->max_value)
		new_value = sp->max_value;
	if (new_value < sp->min_value)
		new_value = sp->min_value;
	sp->value = new_value;
	sp->cur_pos = value_to_pos(sp, new_value);
	if (old_value != sp->value) {
		float new_value;
		new_value = sp->value * 1.0;
		CcDrawSliderValue(dpy, win, sp, FALSE);
		if (sp->value_changed_cb)
			(sp->value_changed_cb)(sp->closure, new_value);
		CcPaintPad(sp->parent);
	}
	CcSliderExpose(w, dpy, win, NULL, sp);
	return;
}

/***********************************************************************
function: CcNewSlider - Create a new slider.  Return pointer to new
slider on success;  NULL on failure.
*/
static CcSlider *
CcNewSlider(parent, name, style, x, y, length, min_value, max_value,
			step_value, value)
AXWidget_color_chooser parent;
char *name;
int style;
int x, y, length;
int min_value, max_value, step_value, value;
{
	CcSlider *sp = NULL;

	if (!name)
		return(sp);
	sp = (CcSlider *) calloc(1, sizeof(CcSlider));
	sp->parent = (AXWidget) parent;
	sp->name = strcpy(malloc(strlen(name) + 1), name);
	if (style == SLIDER_HORIZ || style == SLIDER_VERT)
		sp->style = style;
	else
		sp->style = SLIDER_VERT; /* default */
	sp->x = x;
	sp->y = y;
	sp->length = MAX(SLIDER_MINLEN, length);
	sp->min_value = MIN(min_value, max_value);
	sp->max_value = MAX(min_value, max_value);
	sp->step_value = step_value;
	if (value > sp->max_value)
		value = sp->max_value;
	if (value < sp->min_value)
		value = sp->min_value;
	sp->value = value;
	if (sp->style == SLIDER_HORIZ) {
		sp->width = length;
		sp->height = SLIDER_THICKNESS;
	} else {
		sp->width = SLIDER_THICKNESS;
		sp->height = length;
	}
	sp->cur_pos = value_to_pos(sp, value);
	return(sp);
}

/***********************************************************************
function: CcSliderXlateCoords - Translate X and Y (chooser relative)
to SX, SY (slider relative).
*/
static void
CcSliderXlateCoords(sp, x, y, sxp, syp)
register CcSlider *sp;
register int x, y;
int *sxp;
int *syp;
{
	if (!sp) { /* err */
		*sxp = x;
		*syp = y;
		return;
	}
	*sxp = x - sp->x;
	*syp = y - sp->y;
	return;
}

/***********************************************************************
function: CcPointInSlider - Return TRUE if point (x, y) lies in the
slider SP;  FALSE otherwise.
*/
/*ARGSUSED*/
static int
CcPointInSlider(sp, x, y, str_too)
CcSlider *sp;
int x, y;
int str_too;
{
	/* FIXME, STR_TOO ignored */
	if (!sp)
		return(FALSE);
	if ((x > sp->x) &&
		(x < (sp->x + sp->width)) &&
		(y > sp->y) &&
		(y < (sp->y + sp->height)))
		return(TRUE);
	return(FALSE);
}

/***********************************************************************
function: CcFreeSliderData - Free slider data in SP, then free SP
itself.
NOTE: Caller must set SP to NULL for sanity.
*/
static void
CcFreeSliderData(sp)
CcSlider *sp;
{
	if (!sp)
		return;
	if (sp->name)
		free(sp->name);
	sp->name = NULL;
	free(sp);
	return;
}

/***********************************************************************
function: CcUnmapSlider - Unmap slider SP by clearing window area used
by the slider.
*/
static void
CcUnmapSlider(dpy, win, sp)
Display *dpy;
Window win;
CcSlider *sp;
{
	if (!sp)
		return;
	XFillRectangle(dpy, win, BackGC, sp->x, sp->y, sp->width, sp->height);
	return;
}

/* : CcGetWcolor - init the passed wcolor_t instance to
 * reflect the current value of the passed color chooser widget
 */
static int
CcGetWcolor(w, wcolor)
AXWidget_color_chooser w;  /* in */
wcolor_t *wcolor;          /* out: init caller's storage */
{
	chooser_data *cP;
	
	cP = (chooser_data *) w->info.color_chooser;
	
	if (!cP)
		return FALSE;
	
	wcolor->any.type = WCSPACE_HSB;
	wcolor->hsb.h = cP->h_slider->value;
	wcolor->hsb.s = cP->s_slider->value;
	wcolor->hsb.b = cP->v_slider->value;
	
	return TRUE;
}

static void
CcDrawSliderBarBevels(win, x, y, width, height, style)
Drawable win;              /* in */
int x, y, width, height;   /* in: dimensions of bar */
int style;                 /* in: vert or horiz */
{
	int halflen = SLIDER_BARLEN / 2;   /* put a notch at midpoint of bar */
	
	if (SLIDER_HORIZ == style) {
		AxDrawBevel((AXWidget) NULL, win, x, y, halflen, height,
				    2, BEVEL_OUT);
		AxDrawBevel((AXWidget) NULL, win, x + halflen, y, halflen, height,
					2, BEVEL_OUT);
		AxDrawBevel((AXWidget) NULL, win, x, y, width, height,
					1, BEVEL_OUT); /* grind the outer edge a bit */
	}
	else {
		AxDrawBevel((AXWidget) NULL, win, x, y, width, halflen,
					2, BEVEL_OUT);
		AxDrawBevel((AXWidget) NULL, win, x, y + halflen, width, halflen,
					2, BEVEL_OUT);
		AxDrawBevel((AXWidget) NULL, win, x, y, width, height,
					1, BEVEL_OUT);
	}
}

static void
CcInitLabelFont()
{
	XFontStruct *fs;
	int scalable;
	
	if (!getXFontStruct(NXLT("Times"), 12, 1, 0, &scalable, &fs)) {
		LabelFont = AxDlgMainFont;
	}
	else {
		LabelFont = fs;
	}
}

static void
CcDrawString(dpy, win, gc, x, y, str, len)
Display *dpy;
Drawable win;
GC gc;          /* not used, we assume ForeGC */
int x, y;
char *str;
int len;
{
	axSetFont(dpy, ForeGC, LabelFont);
	axDrawString(dpy, win, ForeGC, LabelFont, x, y, str, len);
	axSetFont(dpy, ForeGC, AxDlgMainFont);
}

#if TEST_COLOR_SHADOW
/***********************************************************************
function: AxDrawColorShadow - 
FIXME, move to xmlibcom.c
*/
void
AxDrawColorShadow(w, win, top_gc, bottom_gc, thickness, x, y, width,
				  height, olshadow, base_color, top_color, bottom_color)
AXWidget w;
Window win;
GC top_gc;
GC bottom_gc;
int thickness;
int x, y, width, height;
int olshadow;
wcolor_t *base_color;
wcolor_t *top_color;
wcolor_t *bottom_color;
{
	float base_h, base_s, base_b;
	float top_h, top_s, top_b;
	float bottom_h, bottom_s, bottom_b;
	int red, green, blue;
	unsigned long top_pixel, bottom_pixel, base_pixel;
	unsigned long saved_top_fg, saved_top_bg;
	unsigned long saved_bottom_fg, saved_bottom_bg;

	if (!base_color)
		goto normal_shadow;

	AxGetHSBFromWColor(base_color, &base_h, &base_s, &base_b);

	if (!top_color)
		compute_sun(&top_h, &top_s, &top_b, base_h, base_s, base_b);
	else
		AxGetHSBFromWColor(top_color, &top_h, &top_s, &top_b);

	if (!bottom_color)
		compute_shade(&bottom_h, &bottom_s, &bottom_b, base_h,
					   base_s, base_b);
	else
		AxGetHSBFromWColor(bottom_color, &bottom_h, &bottom_s, &bottom_b);

	HSBToRGB(&red, &green, &blue, base_h, base_s, base_b);
	AxGetRealPixel(0, red, green, blue, 0, 255, &base_pixel);

	HSBToRGB(&red, &green, &blue, top_h, top_s, top_b);
	AxGetRealPixel(0, red, green, blue, 0, 255, &top_pixel);

	HSBToRGB(&red, &green, &blue, bottom_h, bottom_s, bottom_b);
	AxGetRealPixel(0, red, green, blue, 0, 255, &bottom_pixel);

	if (base_pixel == top_pixel)
		goto normal_shadow; /* punt */
	if (base_pixel == bottom_pixel)
		goto normal_shadow; /* punt */
	if (top_pixel == bottom_pixel)
		goto normal_shadow; /* punt */

	saved_top_fg = AXGetForegroundPixel(Dpy, top_gc);
	saved_top_bg = AXGetBackgroundPixel(Dpy, top_gc);
	saved_bottom_fg = AXGetForegroundPixel(Dpy, bottom_gc);
	saved_bottom_bg = AXGetForegroundPixel(Dpy, bottom_gc);

	XSetForeground(Dpy, top_gc, top_pixel);
	XSetBackground(Dpy, top_gc, top_pixel);
	XSetForeground(Dpy, bottom_gc, bottom_pixel);
	XSetBackground(Dpy, bottom_gc, bottom_pixel);

	/* (don't really know whether it's BEVEL_OUT, oh well) */
	AxDrawBevel(w, win, x, y, width, height, thickness, BEVEL_OUT);

	XSetForeground(Dpy, top_gc, saved_top_fg);
	XSetBackground(Dpy, top_gc, saved_top_bg);
	XSetForeground(Dpy, bottom_gc, saved_bottom_fg);
	XSetBackground(Dpy, bottom_gc, saved_bottom_bg);
	return;
normal_shadow:
#if TEST_COLOR_SHADOW
	printf("PUNT...\n");
#endif
	/* (don't really know whether it's BEVEL_OUT, oh well) */
	AxDrawBevel(w, win, x, y, width, height, thickness, BEVEL_OUT);
	return;
}
#endif /* TEST_COLOR_SHADOW */




/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *						in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
