/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmdecolr.c
   Applix ASC version: 8
   Date: Mon Nov 16 13:53:46 1998
*/
#include <stdio.h>
#include <aos.h>
#include <elxwins.h>
#include <xmdlged.h>
#include <axhtab.h>

    /* these colorspace identifiers are meant to fit
	 * with the current graphics editor file format, colormap section
	 * not to be confused with xmwcolor.h
	 */
#define COLORSPACE_CMYK               0
#define COLORSPACE_TRANSPARENT        1
#define COLORSPACE_SPOT               2
#define COLORSPACE_HSB                3
#define COLORSPACE_RGB                4

#define NEWLINECH(ch)  (((ch) == '\n') || ((ch) == 0x0d))
#define WHITESPACE(ch) (((ch) == ' ')  || ((ch) == '\t') || NEWLINECH(ch))

#define DE_DEFAULT_COLORFILE_NAME NXLT("axDEdefaultColors")

#define DE_MAX_COLORNAME 16          /* to fit: ddd,ddd,ddd */

#define DE_NUM_DEFAULT_COLORS 16

    /* this should be plenty; note that dechoose.d entry field
	 * for color name allows 100 chars max
	 */
#define COLORFILE_LINE_MAX 258

    /* Some colors in HSB; each entry: <name, 0 0 h s b 3>
	 */
static char deDefaultColors[DE_NUM_DEFAULT_COLORS][40] = {
	"<\"White\" 0 0 0 0 100 3>",
	"<\"Black\" 0 0 0 0 0 3>",
	"<\"Gray\" 0 0 0 0 70 3>",
	"<\"\" 0 0 0 50 66 3>",
	"<\"\" 0 0 10 90 100 3>",
	"<\"\" 0 0 30 50 73 3>",
	"<\"\" 0 0 50 90 100 3>",
	"<\"\" 0 0 60 50 80 3>",
	"<\"\" 0 0 80 50 64 3>",
	"<\"\" 0 0 190 90 100 3>",
	"<\"\" 0 0 190 50 80 3>",
	"<\"\" 0 0 210 50 78 3>",
	"<\"\" 0 0 210 30 100 3>",
	"<\"\" 0 0 270 50 64 3>",
	"<\"\" 0 0 280 90 100 3>",
};


typedef struct deColorFile_t {
	axhtab colors;                   /* colors contained in named file */
} deColorFile;

    /* deColorVal_t is used for the value of items
	 * cached in DE.colorTable; the items are named
	 * by their hsb triples, e.g., "278,100,100"
	 */
typedef struct deColorVal_t {
	char *nickName;
	int hue;
	int saturation;
	int brightness;
} deColorVal;

extern edWidget DE;

extern int AxIsArray(), AxArraySize();
extern void axhGetItem(), axhInitTable(), axhDeleteTable();
extern void axhTossItem();
extern void TaskFree();
extern void RGBToHSB(), HSBToRGB(), AxFreeData();
extern void deInitColorTable();
extern void deSendUpdatePoke();
extern char *TaskAlloc();
extern mbchar *AxStrPtrFromArray();
extern data_ptr AxMakeArray(), AxAddArrayToArray();
extern data_ptr AxAddStrToArray(), AxAddIntToArray();
extern data_ptr AxMakeIntData();

char *deNewColorFile(), *deNewColor();
void deTossColorFile(), deTossColor();
void deIncludeDlgColors();
void deAddColors(), deDeleteColors(), deClearColorset();
data_ptr deReadColorset(), deReadColorsetFile();
data_ptr deListColorsUniqueToFile();

static int isAcolor(), convertToHSB(), deReadHSB(), boundValue();
static int deMakeColorsArray();
static char *deDefaultColorFile();
static void deHueSort(), includeThisColor(), writeHSBname();
static void deBrightnessSort();
static data_ptr deGetColorListElement();

/*
 * builtins called by ElfbDEreadData, ElfbDEeditWidget
 */


/*
 * : deReadColorsetFile - cache the contents of a
 * colorset file.  A very simple file might look like:
 *
COLORMAP
<"Black" 0 0 0 0 0 3>
<"White" 0 0 0 0 100 3>
<"Purple" 0 0 278 100 100 3>
END COLORMAP
 *
 */
data_ptr
deReadColorsetFile(info)
data_ptr info;
{
	mbchar *fileName, *result;

	if (!DE.colorFiles.width) {
		deInitColorTable();
	}
	
	fileName = (mbchar *) NULL;
	if (info && AxIsArray(info) && AxIsString(AxArrayElement(info, 1))) {
		    /* this should be an absolute pathname
			 */
		fileName = AxStrPtrFromArray(info, 1);
	}
	else {
		    /* invoke the loading of the hardcoded
			 * default color set, under a name hopefully
			 * odd enough not to collide with a real file name
			 */
		axhGetItem(DE_DEFAULT_COLORFILE_NAME, &DE.colorFiles,
				   (char *) NULL, &result);
		return ((data_ptr) NULL);
	}
	
	    /* hash this file name into DE.colorFiles
		 */
	axhGetItem(fileName, &DE.colorFiles, (char *) NULL, &result);
	deSendUpdatePoke(DE_COLORSET_CHANGED, FALSE);
	
	return ((data_ptr) NULL);
}

/* : deReadColorset - produce a hue-sorted list of all
 * the colors in DE.colorTable
 */
data_ptr
deReadColorset()
{
	    /* the returned data is in this form:
		 * <<color1 name><color1 rgb><color1 hsb triple>>,
		 * <<color2 name><color2 rgb><color2 hsb triple>>, ...
		 *
		 * the color names are for the rowcol label field
		 * the rgbs are for the rowcol values
		 * the hsb triples are used as color identifiers, not
		 *   by the rowcols but by the dlged elf code
		 */
	int ix, numColors;
	hitem **orderedColors;
	data_ptr ret, elptr;
	
	numColors = deMakeColorsArray(&orderedColors);
	if (!numColors) {
		return ((data_ptr) NULL);
	}

	    /* produce the return elf array
		 */
	ret = AxMakeArray(numColors);
	for (ix = 0; ix < numColors; ix++) {
		elptr = deGetColorListElement(orderedColors[ix]);
		if (!elptr) {
			AxFreeData(ret);
			TaskFree(0, (char *) orderedColors);
			return ((data_ptr) NULL);
		}
		ret = AxAddArrayToArray(ret, ix, elptr);
	}
	TaskFree(0, (char *) orderedColors);
	return (ret);
}

/* : deAddColors - add a list of colors to DE.colorTable.
 * the list is expressed on element 1 of info, in format:
 * <doPoke><<c1 entry><c2 entry>...>  each entry looks
 * like a line from a color set file
 */
void
deAddColors(info)
data_ptr info;
{
	int size, ix, doPoke, len;
	data_ptr list;
	char *result, *nickName, colorName[DE_MAX_COLORNAME];
	mbchar *thisline, *cp;
	wcolor_t thisColor;

	if (!info || !AxIsArray(info)) {
		return;
	}

	if (!DE.colorTable.width) {
		deInitColorTable();
	}

	doPoke = AxIntFromArray(info, 1);
	list = AxArrayElement(info, 2);
	size = AxArraySize(list);
	
	for (ix = 0; ix < size; ix++) {
		cp = AxStrPtrFromArray(list, ix);
		len = strlen(cp) + 1;                 /* pass copy to isAcolor */
		thisline = (char *) TaskAlloc(0, len);
		strcpy(thisline, cp);
		if (isAcolor(thisline, &thisColor, &nickName)) {
			    /* looks like a valid colormap entry
				 */
			if (convertToHSB(&thisColor, colorName)) {
					    /* hash this colorname into DE.colorTable
						 */
				axhGetItem(colorName, &DE.colorTable, nickName, &result);
			}
		}
		TaskFree(0, thisline);
	}
	if (doPoke) {
		deSendUpdatePoke(DE_COLORSET_CHANGED, FALSE);
	}
}

/* : deDeleteColors - delete a list of colors from DE.colorTable.
 * the list is expressed on element 1 of info, in format:
 * <<c1 hsbName><c2 hsbName> ...>
 */
void
deDeleteColors(info)
data_ptr info;
{
	int size, ix;
	data_ptr list;
	char *hsbName;
	char *result;

	if (!info || !AxIsArray(info)) {
		return;
	}

	list = AxArrayElement(info, 1);
	size = AxArraySize(list);
	
	for (ix = 0; ix < size; ix++) {
		hsbName = (char *) AxStrPtrFromArray(list, ix);
		axhTossItem(hsbName, &DE.colorTable, (char *) NULL);
	}
	deSendUpdatePoke(DE_COLORSET_CHANGED, FALSE);
}

void
deClearColorset()
{
	axhDeleteTable(&DE.colorTable, (void *) NULL);
	axhDeleteTable(&DE.colorFiles, (void *) NULL);
	axhInitTable(&DE.colorTable, 256, deNewColor, deTossColor);
	axhInitTable(&DE.colorFiles, 32, deNewColorFile, deTossColorFile);
	deSendUpdatePoke(DE_COLORSET_CHANGED, FALSE);
}

/* : deListColorsUniqueFile - return a list of colors that
 * are unique to the named color set file, as compared with
 * a passed list of colors.  the passed list's format is the
 * same as that returned by deReadColorset:
 * <<color1 name><color1 rgb><color1 hsb triple>>,
 * <<color2 name><color2 rgb><color2 hsb triple>>, ...
 *
 */
data_ptr
deListColorsUniqueToFile(info)
data_ptr info;
{
	int ix, listx, anything, listsize;
	char *dontcare;
	data_ptr retList, passedList, element;
	mbchar *fileName, *hsb;
	hitem *list;
	axhtab loctab, *fileColors;
	deColorFile *namedFile;
	
	if (!DE.colorFiles.width) {
		return ((data_ptr) NULL);
	}

	    /* get the table storing the named file's colors
		 */
	fileName = (mbchar *) NULL;
	if (info && AxIsArray(info) && AxIsString(AxArrayElement(info, 1))) {
		    /* this should be an absolute pathname
			 */
		fileName = AxStrPtrFromArray(info, 1);
	}
	axhGetItem(fileName, &DE.colorFiles, (char *) NULL,
			   (char **) &namedFile);
	
	if (!namedFile) {
		return ((data_ptr) NULL);
	}
	fileColors = &namedFile->colors;
	
	    /* record the passed list's hsbnames onto a local table
		 */
	axhInitTable(&loctab, 256, NULL, NULL);
	passedList = AxArrayElement(info, 2);
	if (!passedList || !AxIsArray(passedList)) {
		return ((data_ptr) NULL);
	}
	listsize = AxArraySize(passedList);
	for (ix = 0; ix < listsize; ix++) {
		element = AxArrayElement(passedList, ix);
		hsb = AxStrPtrFromArray(element, 2);
		axhGetItem(hsb, &loctab, (char *) NULL, &dontcare);  /* store */
	}
	
	    /* return hsbnames of those items in namedFile which don't
		 * appear in loctab.
		 * the returned list format is:
		 *   <<c1 wcolor><c1 nickname>><c2 wcolor><c2 nickname>>, ..
		 * c<n> wcolor:
		 *   <WCSPACE_HSB><<hue><sat><bri>>
		 */
	retList = AxMakeArray(1);
	for (ix = 0, listx = 0, anything = FALSE; ix < fileColors->width; ix++) {
		list = fileColors->table[ix];
		while(list) {
			if (list->name && list->value) {
				if (!axhTestItem(list->name, &loctab)) {
					    /* color not in passed list; add to return list
						 */
					data_ptr dwcolor, dhsb;
					deColorVal *cv = (deColorVal *) list->value;
					
					dhsb = AxMakeArray(3);
					dhsb = AxAddIntToArray(dhsb, 0, cv->hue);
					dhsb = AxAddIntToArray(dhsb, 1, cv->saturation);
					dhsb = AxAddIntToArray(dhsb, 2, cv->brightness);
					
					dwcolor = AxMakeArray(2);
					dwcolor = AxAddIntToArray(dwcolor, 0, WCSPACE_HSB);
					dwcolor = AxAddArrayToArray(dwcolor, 1, dhsb);
					
					element = AxMakeArray(2);
					element = AxAddArrayToArray(element, 0, dwcolor);
					element = AxAddStrToArray(element, 1,
											  (mbchar *) cv->nickName);
					
					retList = AxAddArrayToArray(retList, listx, element);
					
					listx++;
					anything = TRUE;
				}
			}
			list = list->next;
		}
	}
	axhDeleteTable(&loctab, (void *) NULL);
	if (!anything) {
		AxFreeData(retList);
		return ((data_ptr) NULL);
	}
	return (retList);
}

data_ptr
deReadColorsetIndex(info)
data_ptr info;
{
	data_ptr colorAtt, hsb;
	data_ptr idx;
	int hue, sat, bri, numColors, ix, colorIx;
	hitem **orderedColors;
	deColorVal *cv;

	idx = AxMakeIntData(-1);
	if (!info) {
		return (idx);
	}
	colorAtt = AxArrayElement(info, 1);
	if (!AxIsArray(colorAtt)) {
		return (idx);
	}
	hsb = AxArrayElement(colorAtt, 1);
	if (!AxIsArray(hsb)) {
		return (idx);
	}
	hue = AxIntFromArray(hsb, 0);
	sat = AxIntFromArray(hsb, 1);
	bri = AxIntFromArray(hsb, 2);

	numColors = deMakeColorsArray(&orderedColors);
	if (!numColors) {
		return (idx);
	}
	
	for (ix = 0, colorIx = -1; ix < numColors; ix++) {
		if (orderedColors[ix]) {
			cv = (deColorVal *) (orderedColors[ix])->value;
			if (cv && hue == cv->hue &&
				sat == cv->saturation &&
				bri == cv->brightness) {

				colorIx = ix;
				break;
			}
		}
	}
	AxFreeData(idx);
	TaskFree(0, orderedColors);
	
	return (AxMakeIntData(colorIx));
}

data_ptr
deHSBToRGB(info)
data_ptr info;
{
	int hue, saturation, brightness;
	int red, green, blue;
	data_ptr hsb, rgb;
	
	if (!info) {
		return ((data_ptr) NULL);
	}
	
	hsb = AxArrayElement(info, 1);
	
	if (!AxIsArray(hsb)) {
		return ((data_ptr) NULL);
	}
	hue = AxIntFromArray(hsb, 0);
	saturation = AxIntFromArray(hsb, 1);
	brightness = AxIntFromArray(hsb, 2);
	
	HSBToRGB(&red, &green, &blue,
			 (float) hue, (float) saturation, (float) brightness);
	
	rgb = AxMakeArray(3);
	rgb = AxAddIntToArray(rgb, 0, red);
	rgb = AxAddIntToArray(rgb, 1, green);
	rgb = AxAddIntToArray(rgb, 2, blue);
	return (rgb);
}

void
deIncludeDlgColors(info)
data_ptr info;
{
	data_ptr dbox, gadgArray, thisGadg, color;
	int ix, gx, arraySize, includedColors;

	dbox = AxArrayElement(info, 1);
	if (!dbox) {
		return;
	}

	    /* index thru the dialog box array, looking
		 * for gadget color info; include any colors
		 * in the session color palette (DE.colorTable)
		 */
	for (ix = 0, includedColors = FALSE; ix < DLG_NUM_GADGETS; ix++) {
		gadgArray = AxArrayElement(dbox, ix);
		arraySize = AxArraySize(gadgArray);
		for (gx = 0; gx < arraySize; gx++) {
			thisGadg = AxArrayElement(gadgArray, gx);
			
			    /* include any of widgetColor, labelColor,
				 * or textColor attributes
				 */
			color = AxArrayElement(thisGadg, DBDEXwidgetColor);
			if (color && AxIsArray(color)) {
				includeThisColor(color);
			}
			color = AxArrayElement(thisGadg, DBDEXlabelColor);
			if (color && AxIsArray(color)) {
				includeThisColor(color);
			}
			color = AxArrayElement(thisGadg, DBDEXtextColor);
			if (color && AxIsArray(color)) {
				includeThisColor(color);
			}
		}
	}
	if (includedColors) {
		deSendUpdatePoke(DE_COLORSET_CHANGED, FALSE);
	}
}


/*
 * support stuff
 *
 */


/* : deGetColorListElement - return an elf array in the form:
 * <nick name><<hue><sat><bri>><hsbName> from the passed deColorVal
 */
static data_ptr
deGetColorListElement(listElement)
hitem *listElement;                      /* in */
{
	data_ptr ret, hsb_ptr;
	deColorVal *thisColor;

	thisColor = (deColorVal *) listElement->value;
	
	/* init array element in form:
	 * <nick name><<hue><sat><bri>><hsbName>
	 */
	ret = AxMakeArray(3);
	ret = AxAddStrToArray(ret, 0, thisColor->nickName);
	hsb_ptr = AxMakeArray(3);
	hsb_ptr = AxAddIntToArray(hsb_ptr, 0, thisColor->hue);
	hsb_ptr = AxAddIntToArray(hsb_ptr, 1, thisColor->saturation);
	hsb_ptr = AxAddIntToArray(hsb_ptr, 2, thisColor->brightness);
	ret = AxAddArrayToArray(ret, 1, hsb_ptr);
	ret = AxAddStrToArray(ret, 2, listElement->name);
	return (ret);
}

static void
includeThisColor(colorAtt)
data_ptr colorAtt;
{
	data_ptr hsb;
	int hue, sat, bri, result;
	mbchar *nickName;
	char hsbName[DE_MAX_COLORNAME];
	
	if (WCSPACE_HSB == AxIntFromArray(colorAtt, 0)) {
		hsb = AxArrayElement(colorAtt, 1);
		hue = AxIntFromArray(hsb, 0);
		sat = AxIntFromArray(hsb, 1);
		bri = AxIntFromArray(hsb, 2);
		nickName = AxStrPtrFromArray(colorAtt, 2);
		if (nickName && !(*nickName)) {
			nickName = (mbchar *) NULL;
		}
		
		    /* form the lookup name of this color, and hash into set
			 */
		writeHSBname(hsbName, hue, sat, bri);
		axhGetItem(hsbName, &DE.colorTable, (char *) nickName, &result);
	}
}


/* there are two hash tables on the DE struct; one (DE.colorFiles)
 * is keyed by colorset filename, and stores the contents of those
 * files.  The other (DE.colorTable) is keyed by HSB triple, and
 * stores the union of all the colorset files' colors, plus whatever
 * additional colors have been added during an edit session
 *
 * Here are the constructors and destructors for the item types
 * in these tables:
 */
 
/* : deNewColorFile - constructor for items in DE.colorFiles.  Each
 * item's constructed value is a deColorFile type instance
 */
char *
deNewColorFile(filename, notUsed)
char *filename;                    /* an absolute pathname */
char *notUsed;
{
	char thisline[COLORFILE_LINE_MAX], *result, *nickName;
	int inColors;
	char colorName[DE_MAX_COLORNAME];    /* max name: "ddd,ddd,ddd" */
	FILE *inf;
	wcolor_t thisColor;
	deColorFile *newSet = (deColorFile *) NULL;
	
	if (!DE.colorFiles.width || !DE.colorTable.width) {
		deInitColorTable();
	}
	
	if (!filename) {
		return ((char *) deDefaultColorFile());
	}
	
	if (!strcmp(filename, DE_DEFAULT_COLORFILE_NAME)) {
		return ((char *) deDefaultColorFile());
	}
	
	if ((inf = fopen(filename, "r")) == (FILE *) NULL) {
		return ((char *) deDefaultColorFile());
	}

	inColors = FALSE;
	
	while (fgets(thisline, COLORFILE_LINE_MAX, inf) != (char *) NULL) {
		
		if (!strncmp(thisline, NXLT("END COLORMAP"), 12)) {
			break;
		}
		
		if (inColors) {
			if (isAcolor(thisline, &thisColor, &nickName)) {
				    /* looks like a valid colormap entry
					 */
				if (convertToHSB(&thisColor, colorName)) {
					    /* hash this colorname both into newSet->colors
						 * and into DE.colorTable
						 */
					if (!newSet) {
						newSet = (deColorFile *) TaskAlloc
							(0, sizeof(deColorFile));
						axhInitTable(&newSet->colors, 128,
									 deNewColor, deTossColor);
					}

					axhGetItem(colorName, &newSet->colors, nickName, &result);
					axhGetItem(colorName, &DE.colorTable, nickName, &result);
				}
			}
		}
		
		if (!strncmp(thisline, NXLT("COLORMAP"), 8)) {
			inColors = TRUE;
		}
		
	}
	return ((char *) newSet);
}

/* : deTossColorFile - destructor for DE.colorFiles
 */
void
deTossColorFile(colorfile, notUsed)
char *colorfile;    /* interpret as (deColorFile *) */
char *notUsed;
{
	int ix;
	hitem *list;
	deColorFile *cFile;
	
	cFile = (deColorFile *) colorfile;
	if (cFile) {
		if (cFile->colors.width && cFile->colors.table) {
			axhDeleteTable(&cFile->colors, (void *) NULL);
		}
		TaskFree(0, (char *) cFile);
	}
}

/* : deNewColor - constructor for individual color items
 */
char *
deNewColor(hsbName, nickName)
char *hsbName;                 /* e.g., "278,100,100" */
char *nickName;                /* e.g., "Purple" */
{
	deColorVal *cv; 
	int ihue, isat, ibri;
	char *nn;

	if (!deReadHSB(hsbName, &ihue, &isat, &ibri)) {
		    /* the passed hsbName doesn't look right
			 */
		return ((char *) NULL);
	}
	cv = (deColorVal *) TaskAlloc(0, sizeof(deColorVal));
	if (nickName) {
		nn = (char *) TaskAlloc(0, strlen(nickName) + 1);
		strcpy(nn, nickName);
		cv->nickName = nn;
	}
	else {
		cv->nickName = (char *) NULL;
	}
	cv->hue = ihue;
	cv->saturation = isat;
	cv->brightness = ibri;
	return ((char *) cv);
}

/* : deTossColor - destructor for individual color items
 */
void
deTossColor(colorVal, notUsed)
char *colorVal;                 /* interpreted as (deColorVal *) */
char *notUsed;
{
	deColorVal *cv = (deColorVal *) colorVal;
	if (cv) {
		if (cv->nickName) {
			TaskFree(0, (char *) cv->nickName);
		}
		TaskFree(0, (char *) cv);
	}
}

/* : isAcolor - read the passed color file line, and
 * fill out the passed wcolor_t instance if possible.
 * caller beware -- returned nickName pointer points
 * into passed entry argument
 */
static int             /* return whether conversion to hsb occurred */
isAcolor(entry, color, nickName)
char *entry;           /* in/out: line copy from colormap portion of
						  color file; will be null-stamped at nickname end */
wcolor_t *color;       /* out: rewrite as HSB wcolor_t instance if possible */
char **nickName;       /* out: return nick-name of color (ptr into entry) */
{
	int i1, i2, i3, i4, i5, colorType, itmp;
	char *cp, *tcp;
	
	    /* the expected format for the input line is:
		 * <name 0 0 m n p colorType>, for instance:
		 *
		 * <"Red" 0 0 7 100 98 3>, representing a
		 * red shade in hsb space (type 3), where h,s,b = 7,100,98
		 */
	
	    /* first test whether bounded by < > characters
		 */
	        /* look at left edge */
	cp = entry;
	while (*cp && WHITESPACE(*cp)) {
		cp++;
	}
	if (*cp != '<') {
		return FALSE;    /* 1st blackspace char not '<' */
	}
	        /* look at right edge */
	itmp = strlen(entry);
	cp = &entry[itmp - 1];
	while ((cp > entry) && WHITESPACE(*cp)) {
		cp--;
	}
	if (*cp != '>') {
		return FALSE;    /* last blackspace char not '<' */
	}
	*cp = '\0';          /* convert the closing '>' to a null */
	
	    /* find the name field; it must be bounded by double quotes
		 */
	cp = entry;
	while (*cp && WHITESPACE(*cp)) {
		cp++;
	}
	cp++;    /* skip past '<' char tested for above */
	
	while(*cp && (*cp != '\"')) {    /* look for start quote */
		cp++;
	}
	if (!(*cp)) {
		return FALSE;
	}
	cp++;    /* skip leading quote */
	tcp = cp;
	while(*cp && (*cp != '\"')) {    /* look for close quote */
		cp++;
	}
	if (!(*cp)) {
		return FALSE;
	}
	*cp = '\0';          /* we're at the close quote; convert to null */
	itmp = strlen(tcp);  /* length of name string */
	if (itmp) {
		*nickName = tcp;  /* caller beware, this points into passed entry */
	}
	else {
		*nickName = (char *) NULL;   /* couldn't find a nickname */
	}

	    /* we expect exactly six more fields expressing integer
		 * numbers
		 */
	cp++;                /* skip the nulled closing quote of the name */
	itmp = sscanf(cp, "%d %d %d %d %d %d",
				  &i1, &i2, &i3, &i4, &i5, &colorType);
	if (itmp != 6) {
		return FALSE;
	}
	
	    /* init the passed wcolor_t instance
		 */
	if (!color) {
		return FALSE;
	}
	memset((char *) color, 0, sizeof(wcolor_t));
	color->any.type = colorType;
	
	switch(colorType) {
		
	case COLORSPACE_HSB:
		color->hsb.h = (float) i3;
		color->hsb.s = (float) i4;
		color->hsb.b = (float) i5;
		break;
		
	case COLORSPACE_RGB:
		color->rgb.r = i3;
		color->rgb.g = i4;
		color->rgb.b = i5;
		break;
		
	default:
		return FALSE;
		break;
		
	}
	return TRUE;
}

/* : convertToHSB - produce a name string representing
 * a color in HSB space, for instance, "7,100,98" represents
 * a shade of red.  The convention is for the name to be formed
 * as three decimal numbers in order h,s,b, comma separated,
 * with no leading zeros or whitespace
 */
static int
convertToHSB(color, hsbName)
wcolor_t *color;             /* in */
char *hsbName;               /* out: write name onto caller's storage */
{
	float fhue, fsat, fbri;
	int ihue, isat, ibri;
	
	switch(color->any.type) {
		
	case COLORSPACE_HSB:
		ihue = (int) color->hsb.h;
		isat = (int) color->hsb.s;
		ibri = (int) color->hsb.b;
		break;
		
	case COLORSPACE_RGB:
		RGBToHSB(&fhue, &fsat, &fbri, color->rgb.r,
				 color->rgb.g, color->rgb.b);
		ihue = (int) fhue;
		isat = (int) fsat;
		ibri = (int) fbri;
		break;
		
	default:
		return FALSE;
		break;
		
	}
	
	    /* write the name onto *hsbName
		 */
	writeHSBname(hsbName, ihue, isat, ibri);
	
	return TRUE;
}

static char *
deDefaultColorFile()
{
	int ix, len;
	deColorFile *defValue;
	wcolor_t thisColor;
	char colorName[DE_MAX_COLORNAME];      /* to fit max: "ddd,ddd,ddd */
	char *result, *nickName, *entry;
	
	defValue = (deColorFile *) TaskAlloc(0, sizeof(deColorFile));
	axhInitTable(&defValue->colors, 128, deNewColor, deTossColor);
	
	for (ix = 0; ix < DE_NUM_DEFAULT_COLORS; ix++) {
		    /* copy the deDefaultColors element onto entry and
			 * pass that to isAcolor (isAcolor will null-stamp onto entry)
			 */
		len = strlen(deDefaultColors[ix]) + 1;
		entry = (char *) TaskAlloc(0, len);
		strcpy(entry, deDefaultColors[ix]);
		if (isAcolor(entry, &thisColor, &nickName)) {
			convertToHSB(&thisColor, colorName);
			if (colorName) {
				    /* store this colorName in defValue->colors,
					 * and in DE.colorTable
					 */
				axhGetItem(colorName, &defValue->colors, nickName, &result);
				axhGetItem(colorName, &DE.colorTable, nickName, &result);
			}
		}
		TaskFree(0, (char *) entry);
	}
	
	return ((char *) defValue);
}
		
static int
boundValue(val, min, max)
int val;
int min;
int max;
{
	if (val < min) {
		return (min);
	}
	else if (val > max) {
		return (max);
	}
	else {
		return val;
	}
}

/* : deReadHSB - extract the hue, saturation, and
 * brightness values from the passed hsbName
 * (e.g., "278,100,100"). This function will return
 * FALSE if the hsbName doesn't doesn't look like a
 * comma-separated triple of ascii decimal integers,
 * no whitespace, no leading zeros
 */
static int
deReadHSB(hsbName, hue, saturation, brightness)
char *hsbName;              /* in */
int *hue;                   /* out */
int *saturation;            /* out */
int *brightness;            /* out */
{
	int ihue, isat, ibri;
	data_ptr ret, rgb_ptr;
	char scratch[DE_MAX_COLORNAME];   /* big enough for "ddd,ddd,ddd" */
	char *comma1, *comma2, *cp;
	
	if ((int) strlen(hsbName) >= DE_MAX_COLORNAME) {
		    /* the color caching code should have
			 * written all hsb names in the form: "ddd,ddd,ddd"
			 * something is wrong..
			 */
		return (FALSE);
	}
	strcpy(scratch, hsbName);

	    /* count commas as a second check
		 */
	for (cp = scratch, comma1 = comma2 = (char *) NULL; *cp; cp++) {
		if (',' == *cp) {
			if (!comma1) {
				comma1 = cp;                /* must be 1st comma */
			}
			else if (!comma2) {
				comma2 = cp;                /* must be 2nd comma */
			}
			else {
				return (FALSE);             /* too many commas */
			}
		}
	}
	if ('\0' == *(comma2+1)) {
		return (FALSE);                     /* nothing after 2nd comma */
	}
	*comma1 = *comma2 = '\0';       /* null-stamp the commas */
	comma1++;                       /* position at start of saturation */
	comma2++;                       /* position at start of brightness */

	    /* return hsb integers
		 */
	sscanf(scratch, "%d", hue);
	sscanf(comma1, "%d", saturation);
	sscanf(comma2, "%d", brightness);

	return (TRUE);
}


/* : deMakeColorsArray - produce a straight, hue-sorted
 * array of pointers to the deColorVal items in DE.colorTable.
 * Caller must free the returned array
 */
static int
deMakeColorsArray(orderedColors)
hitem ***orderedColors;          /* out: alloc and init; caller frees */
{
	int ix, cx, numColors;
	hitem *list;
	
	    /* count the colors in DE.colorTable
		 */
	for (ix = 0, numColors = 0; ix < DE.colorTable.width; ix++) {
		list = DE.colorTable.table[ix];
		while(list) {
			if (list->name && list->value) {
				numColors++;
			}
			list = list->next;
		}
	}
	if (!numColors) {
		return 0;
	}
	
	    /* write a straight array storing pointers to the table items
		 */
	ix = numColors * sizeof(hitem *);
	(*orderedColors) = (hitem **) TaskAlloc(0, ix);
	for (ix = 0, cx = 0; ix < DE.colorTable.width; ix++) {
		list = DE.colorTable.table[ix];
		while(list) {
			if (list->name && list->value) {
				(*orderedColors)[cx] = list;
				cx++;
			}
			list = list->next;
		}
	}
	
	deHueSort((*orderedColors), numColors);    /* sort by hue, increasing */
	
	return (numColors);
}
	

/* : deHueSort - sort the passed array of
 * (deColorVal *) hitems by hue, smaller->larger
 * no provision for tie-breaking
 */
static void
deHueSort(orderedColors, numColors)
register hitem **orderedColors;
register int numColors;
{
	register int ix;
	register int nx;
	register int redo;
	register hitem *itmp;
	register deColorVal *iv;
	register deColorVal *nv;
	register int rx;
	register int inRange;
	register int rangeLen;
	
again:	
	for (ix = 0, redo = FALSE; ix < numColors; ix++) {
		if (ix < (numColors - 1)) {
			nx = ix + 1;
			iv = (deColorVal *) (orderedColors[ix])->value;
			nv = (deColorVal *) (orderedColors[nx])->value;
			if (iv->hue > nv->hue) {
				itmp = orderedColors[nx];
				orderedColors[nx] = orderedColors[ix];
				orderedColors[ix] = itmp;
				redo = TRUE;
			}
		}
	}
	if (redo) {
		goto again;
	}
	
	    /* break hue-ties with a brightness sort over the tied ranges
		 */
	inRange = FALSE;
	rx = 0;
	rangeLen = 0;
	for (ix = 0; ix < numColors; ix++) {
		if (ix < (numColors - 1)) {
			nx = ix + 1;
			iv = (deColorVal *) (orderedColors[ix])->value;
			nv = (deColorVal *) (orderedColors[nx])->value;
			if (inRange) {
				if (nv->hue != iv->hue) {
					    /* end of tied range
						 */
					deBrightnessSort(&orderedColors[rx], rangeLen);
					inRange = FALSE;
					rangeLen = 0;
				}
				else {
					rangeLen++;
				}
			}
			else if (nv->hue == iv->hue) {
				inRange = TRUE;
				rangeLen = 2;
				rx = ix;
			}
		}
	}
}

static void
deBrightnessSort(orderedColors, numColors)
register hitem **orderedColors;
register int numColors;
{
	register int ix;
	register int nx;
	register int redo;
	register hitem *itmp;
	register deColorVal *iv;
	register deColorVal *nv;
	
again:	
	for (ix = 0, redo = FALSE; ix < numColors; ix++) {
		if (ix < (numColors - 1)) {
			nx = ix + 1;
			iv = (deColorVal *) (orderedColors[ix])->value;
			nv = (deColorVal *) (orderedColors[nx])->value;
			if (iv->brightness > nv->brightness) {
				itmp = orderedColors[nx];
				orderedColors[nx] = orderedColors[ix];
				orderedColors[ix] = itmp;
				redo = TRUE;
			}
		}
	}
	if (redo) {
		goto again;
	}
}

static void
writeHSBname(hsbName, ihue, isat, ibri)
char *hsbName;            /* out: write name onto caller's memory */	
int ihue, isat, ibri;     /* in */
{	
	char chue[32];        /* more than enough for ascii MAX_INT */
	char csat[32];
	char cbri[32];
	
	sprintf(chue, "%d", boundValue(ihue, 0, 360));
	sprintf(csat, "%d", boundValue(isat, 0, 100));
	sprintf(cbri, "%d", boundValue(ibri, 0, 100));
	strcpy(hsbName, chue);
	strcat(hsbName, ",");
	strcat(hsbName, csat);
	strcat(hsbName, ",");
	strcat(hsbName, cbri);
}

