/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmdlged.h
   Applix ASC version: 35
   Date: Mon Nov 16 13:53:42 1998
*/
#ifndef _XMDLGED_H
#define _XMDLGED_H

#include <xmlib.h>
#include <axhtab.h>
#include <axtext.h>
#include <elf.h>


#define DONTCARE 0
#define ELFTRUE (-1)

#define NUM_GADGET_ATTS 40 /* #atts expressed to user interface */

#define MAX_PICK_DISTANCE 2 /* strokes shorter than this are picks */
#define HANDLESIZE 7
#define HANDLEHALF 3

/* 'link' types associating groups of widgets
 */
#define WLINKUP 1		/* toward composite or group top */
#define WLINKDOWN 2		/* toward composite or group bot */

/* alignment directive types - bitwise
 */
#define WALIGN_LEFT		(0x1)
#define WALIGN_RIGHT	(0x2)
#define WALIGN_TOP		(0x4)
#define WALIGN_BOT		(0x8)
#define WALIGN_CENTER	(0x10)
#define WALIGN_WIDTH	(0x20)
#define WALIGN_HEIGHT	(0x40)
#define WALIGN_RECT		(0x80)

/* alignment object types
 */
#define ALIGN_LINE 1
#define ALIGN_RECT 2
#define ALIGN_GRID 3

/* DE message codes
 */
#define DE_WIDGET_UPDATE 1		/* update an existing widget */
#define DE_NEW_WIDGET 2			/* create a new widget */
#define DE_DELETE_WIDGET 3		/* toss a widget */
#define DE_DELETE_SELECTED 4	/* delete those widgets curr. selected */
#define DE_WIDGET_ATTS 5		/* set atts, or read atts */
#define DE_EDIT_PREFS 6
#define DE_READ_PREFS 7
#define DE_READ_ATTS 8
#define DE_SET_DIALOG_ATTS 9
#define DE_UPDATE_NOTIFY 10		/* broadcast poke to editor's dialogs */
#define DE_READ_WIDGET_ATTS 11 /* read atts for particular type of widget */
#define DE_READ_DIALOG_ATTS 12
#define DE_READ_VERB_BUTTON_SIZE 13
#define DE_WIDGETS_LIST 14
#define DE_MENU_BAR 15		/* placeholder on c side */
#define DE_OPEN_EDIT_SESSION 16
#define DE_CLOSE_EDIT_SESSION 17
#define DE_NEW_LAYER 18
#define DE_DELETE_LAYER 19
#define DE_READ_DIALOGS_LIST 20
#define DE_READ_MM 21
#define DE_WRITE_MM 22
#define DE_DISPLAY_SELECTED_AS_TABLE 23
#define DE_UNIQUE_DLG_NAME 24  /* compute a unique title for elf */
#define DE_GENERIC_ATTS 25     /* express generic widget atts */
#define DE_WRITE_CLIPBOARD  26
#define DE_READ_CLIPBOARD   27
#define DE_UNIQUE_GADGET_ID 28
#define DE_READ_SELECTION_DLG 29
#define DE_READ_SELECTED_WTYPES 30
#define DE_READ_GADGET_ORDER 31
#define DE_READ_COLOR_SET 32
#define DE_READ_COLOR_FILE 33
#define DE_ADD_COLORS 34
#define DE_DELETE_COLORS 35
#define DE_READ_COLORSET_INDEX 36
#define DE_HSB_TO_RGB 37
#define DE_INCLUDE_DLG_COLORS 38
#define DE_CLEAR_COLORSET 39
#define DE_LIST_COLORS_UNIQUE_TO_FILE 40
#define DE_SEND_UPDATE_POKE 41
#define DE_FLUSH_UPDATE_POKES 42
#define DE_DEFER_UPDATE_POKES 43
#define DE_WIDGET_HANDLES_MATCH 44
#define DE_LIST_FONT_FAMILIES 45
#define DE_ALIGN_SELECTED_WIDGETS 46
#define DE_ACCELERATOR_KEYS 47


	/* some "datatype" defines for use with the messages
	 */
#define DTYPE_INT 1
#define DTYPE_STR 2
#define DTYPE_STRS_ARRAY 3
#define DTYPE_POLY 4
#define DTYPE_NONE 5
#define DTYPE_FONT_ATTR 6

/* more DE message modifiers
 */
#define DE_APPLYSET_WTYPE 1
#define DE_APPLYSET_ALL_SELECTED 2
#define DE_APPLYSET_HANDLE 3

/* reasons for DE_UPDATE_NOTIFY; these start at a
 * large number because some of the message codes
 * are used as update-notify reasons
 */
#define DE_SELSET_CHANGED 101
#define DE_POSITION_CHANGED 102
#define DE_SIZE_CHANGED 103
#define DE_SELSET_CHANGED_BY_MENU 104
#define DE_COLORSET_CHANGED 105
#define DE_WIDGET_ATTS_CHANGED 106

#define DE_STROKE_START 1
#define DE_STROKE_MID 2
#define DE_STROKE_END 3

/* this pixel amount estimates the maximum
 * extent to which any widget can be larger
 * than its w->width, w->height
 */
#define WMARGIN 5 /* too small, ?? */

enum handleType {
	topLeft, topRight, botRight, botLeft,
	topEdge, rightEdge, botEdge, leftEdge,
};

enum stretch_constraint {
	noConstraint, widthOnly, heightOnly, noStretch,
};

/* wEditType's values are initialized as bitwise,
 * to allow them to coincide.
 */
enum wEditType {
	wMove = 0x1, wResize = 0x2, wAtts = 0x4,
};

/* some typedefs to help with editing
 */
typedef struct text_rect_t {
	struct text_rect_t *next;
	AXWidget ownWidget;
	int whichMember;
	int whichIndex;
	r32 txrect;
	p32 txpos;
} textRect;

typedef struct wcomposite_t {
	struct wcomposite_t *next;
	AXWidget w;
	int type; /* WLINKUP or WLINKDOWN */
} wComposite;

typedef struct wgroup_t {
	struct wgroup_t *next;
	AXWidget w;
	int type; /* WLINKUP or WLINKDOWN */
} wGroup;

typedef struct walign_t {
	unsigned int persistent : 1;
	struct walign_t *next;
	AXWidget w;
	int type;
} wAlign;

typedef struct wlayer_t {
	struct wlayer_t *next; /* toward front */
	struct wlayer_t *prev; /* toward back */
	AXWidget w; /* widget represented by this layer */
} wLayer;

typedef struct wcreate_t {
	unsigned int doCreate : 1;
	AXWidget_dialog whichDialog;
	int wType;		/* widget type of widget-to-be */
	data_ptr wInfo;	/* gadget_ format elf data for widget-to-be */
	AXWidget wWidget;
} wCreate;

typedef struct w_edit_state_t {
	int marked : 1;
	int measureText : 1;
	int damaged : 1;
	int changed : 1;
	p32 pixPosAdjust;
	AXWidget old;
	AXWidget gridWidget;
} wEditState;

typedef struct w_aux_t {
	wEditState wEdit;
	wComposite *compositeList;
	data_ptr gadget;
	r32 damageRect;
} wAux;

typedef struct prefs_t {
	unsigned int grid : 1;       /* whether to display grid */
	unsigned int snapToGrid : 1;
	unsigned int borders : 1;    /* whether to display borders */
	unsigned int ghost : 1;      /* whether to do ghost move,resize */
	int gridsize;	             /* in pix */
} wPrefs;

typedef struct intab_t {
	int numInts;              /* number of distinct integer values */
	int *table;               /* array holding integer values */
} intab;

typedef struct de_message_holder_t {
	int num_lists;
	data_ptr *lists;          /* lists of return of deAddToUpdate */
} de_message_holder;

typedef struct de_dialog_t {
	unsigned int doUpdate: 1; /* whether to bother sending update data */
	AXWidget dlg;             /* handle of dialog under edit */
	axhtab widgetUpdates;     /* saved DE_WIDGET_UPDATE message data */
} de_dialog;

typedef struct edwidget_t {
	unsigned int possiblePick : 1;
	unsigned int possibleDblPick : 1;
	unsigned int stretching : 1;
	unsigned int dragSelecting;
	unsigned int init : 1;
	unsigned int main_menu : 1;
	unsigned int deferUpdatePokes : 1;
	int numDlgs;                 /* #dialogs being edited */
	de_dialog *dialogs;          /* their handles and other info */
	AXWidget *selset;
	AXWidget gridWidget;
	AXWidget *selectedSet;       /* handles of selected widgets */
	int selectedSetSize;         /* #widgets in the selection */
	wLayer **layerLists;
	wCreate newWidget;
	AXWidget textEditWidget;
	axtext *txt;
	wPrefs prefs;
	XExposeEvent fakeExpose;
	AXWidget dragWidget;
	p32 dragpt, startpt;         /* for mouse moves and resizes */
	r32 origSize;
	r32 *selsetOrigSizes;
	r32 *lastGhost;
	enum handleType stretchHandle;
	Time lastReleaseTime, lastPressTime;
	GC alignGC;
	GC borderGC;
	GC rubberGC;
	GC ghostGC;
	int typeTable[NUM_GADGET_ATTS][2];
	data_ptr clipboard;
	axhtab colorTable;        /* session colorset */
	axhtab colorFiles;        /* colorset files' contents */
	intab updatePokes;        /* deferred update pokes */
	AXWidget *deferredSelection;
	int numDeferredSelected;
} edWidget;

#endif /* XMDLGED_H */

