/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmeditbx.c
   Applix ASC version: 97
   Date: Wed May 12 21:10:32 1999
*/
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>
#include <wchars.h>
#include <tmkeys.h>
#include <aos_xfer.h>
#include <doctypes.h>


/*  defines  */
#define MAX_XSTR 8000  /* length of seln to be inserted - same as in ATM */
/* IS_NGC adapted from ATM's tmin1.c definition */
#define DEL     0177        /* ascii code */
#define IS_NGC(c) ((((unsigned char)(c)) < ' ') || (((unsigned char)(c)) == DEL))

#define SPACE ((wchar) ' ')
#define WHITESPACE(ch) (((ch) == (int) SPACE) || ((ch) == (int) '\t') ||\
						((ch) == (int) '\n'))

void displayTextObj();
void (*EB_ObjDisplayer[])() = {
    NULL,               /* 0 - undefined */ 
    displayTextObj,     /* 1 - TEXTOBJ */
    NULL
};

/*  externs  */
extern int UseWidgetColors;
extern XFontStruct *AxDlgMainFont;  /* Bolded dialog font */
extern int AxCharAbove, AxCharHeight;
extern int AxButtonDown, AxControlKeyDown;
extern int DoubleMillisec;
extern AXWidget WindowList;
extern Display *Dpy;
extern GC MonospaceGC, MonospaceBackGC;
extern GC ForeGC, BackGC;
extern GC GXForeGC;
extern GC olButtonGC, olHilitButtonGC;
extern GC olTextGC, olHilitTextGC;
extern XFontStruct *MonospaceFont;
extern Atom AxXA_CLIPBOARD;
extern Atom AxXA_LENGTH;
extern Atom AxXA_EDITBOX_PROP;
extern data_ptr ElfRetData;
#define XA_CLIPBOARD AxXA_CLIPBOARD
#define XA_LENGTH AxXA_LENGTH
#define XA_EDITBOX_PROP AxXA_EDITBOX_PROP

#include "xmeditbx.p"

/* autoscroll vars EditBoxArmed, editboxAutoscrollEdge
 */
extern AXWidget_editbox EditBoxArmed;
enum editboxEdge {
	noEdge = 0, leftEdge = 1, topEdge = 2, rightEdge = 3, bottomEdge = 4,
};
static enum editboxEdge editboxAutoscrollEdge;

/**********************************************************************
                            Edit Box Widget
***********************************************************************/




/*
 * clearEditBoxData - destroy & clear contents, re-initialize editbox
 */
clearEditBoxData(w)
AXWidget_editbox w;
{
    destroyEditBox(w);
    createEditBox(w);
    clearWindow(w);
    resetScrollInfo(w);
}


clearWindow(w)
AXWidget_editbox w;
{
    GC gc;

    if ( !w->window )
        return;

    gc = w->monospace_font ? MonospaceBackGC : BackGC;

    if ( w->useWorkColors )
        SetWorkColors(gc, 0);

    xmColorFillRectangle((AXWidget)w, w->window, gc, 0, 0, w->editbox->width,
			 w->editbox->height);

    if ( w->useWorkColors )
        UnSetWorkColors(gc, 0);
}


int
editboxMouseEvent(w, eventP)
AXWidget_editbox w;
XEvent *eventP;
{
    int x, y;
	int consumed = FALSE;

    switch (eventP->type) {
        case ButtonPress:
            x = eventP->xbutton.x;
            y = eventP->xbutton.y;
            manageButtonPress(w, eventP, x, y);
			consumed = TRUE;
            break;
        case ButtonRelease:
            x = eventP->xbutton.x;
            y = eventP->xbutton.y;
            manageButtonRelease(w, eventP, x, y);
			consumed = TRUE;
            break;
        case MotionNotify:
            x = eventP->xmotion.x;
            y = eventP->xmotion.y;
            manageMouseMotion(w, eventP, x, y);
			consumed = TRUE;
            break;
    }
	return (consumed);
}

manageButtonPress(w, eventP, x, y)
AXWidget_editbox w;
XEvent *eventP;
int x, y;
{
    any_obj_t *op;
    tcursor_t *cursor;
    tline_t *lp;
    seln_t *sel;
    int char_index;
    int button_pressed;

    /* unfocus msg current entry box, other edit box */
    AxUnFocusCurrent((AXWidget)w);

    sel = &w->editbox->sel;
    cursor = &w->editbox->tcursor;
    button_pressed = (int)eventP->xbutton.button;

#if LINUX
	if (button_pressed == Button4) { /* mouse scroll up */
		char *str = (char *)UPbyName(NXLT("axMouseWheelScroll"));
		if (str && (atoi(str) != 0)) {
			if (ThimIsShifted())
				w->scroll_pos = MAX(0, w->scroll_pos - w->editbox->height +
									w->editbox->lineHeight);
			else
				w->scroll_pos = MAX(0,w->scroll_pos - 3*w->editbox->lineHeight);
			if (w->scroll_pos != (-w->editbox->yoffset)) {
				PaintEditBox(w, NULL, TRUE, TRUE);
			}
		}
	} else if (button_pressed == Button5) { /* mouse scroll down */
		char *str = (char *)UPbyName(NXLT("axMouseWheelScroll"));
		if (str && (atoi(str) != 0)) {
			if (ThimIsShifted())
				w->scroll_pos = MIN(w->scroll_max - w->editbox->lineHeight,
									w->scroll_pos + w->editbox->height -
									w->editbox->lineHeight);
			else
				w->scroll_pos = MIN(w->scroll_max - w->editbox->lineHeight,
									w->scroll_pos + 3*w->editbox->lineHeight);
			if (w->scroll_pos != (-w->editbox->yoffset)) {
				PaintEditBox(w, NULL, TRUE, TRUE);
			}
		}
	}
#endif

    if ( (eventP->xbutton.time-w->last_click_time)<=DoubleMillisec
                && w->last_button_clicked==button_pressed )
        w->num_multi_clicks++;
    else
        w->num_multi_clicks = 1;
    w->last_button_clicked = button_pressed;
    w->last_click_time = eventP->xbutton.time;

    if ( button_pressed == Button1 ) {

	EditBoxArmed = w;

        if (w->num_multi_clicks == 2) {
            hilightWord(w, x, y);
            return;
        }
        else if (w->num_multi_clicks == 3) {
            hilightLine(w, x, y);
            return;
        }
        else if (w->num_multi_clicks == 4) {
            hilightPara(w, x, y);
            return;
        }
        else if (w->num_multi_clicks == 5) {
            hilightDoc(w, x, y);
            return;
        }

        w->num_multi_clicks = 1;
        if ( ThimIsShifted() && w->pending_delete ) {
            /* Motif: adjust mark */
            drawCursor(w, TC_ERASE);
            adjustHilight(w, x, y);
        }
        else {
            /* clear any existing selection */
            clearSelection(w, XA_PRIMARY, TRUE);
            drawCursor(w, TC_ERASE);
            /* acquire selection atom (XA_PRIMARY) */
            if ( initSelection(w, XA_PRIMARY) ) {
                w->pending_delete = TRUE;
                w->deleteMarked = TRUE;
                getNearestObj(w, x, y, &op, &lp);
                getNearestIndex(w, op, lp, x, y, &char_index);
                sel->startObj = sel->endObj = op;
                sel->startIndex = sel->endIndex = char_index;
            }
        }
    }
    else if ( button_pressed == Button2 ) {
		/* reposition cursor and paste selection */
		if (w->no_picks)
			return;
		
		/* change cursor position */
		drawCursor(w, TC_ERASE);
		getNearestVisibleIndex(w, x, y, &op, &lp, &char_index);
		cursor->obj = op;
		cursor->ac_index = char_index;
		drawCursor(w, TC_DRAW);
		
		/* get XA_PRIMARY selection */
		w->deleteMarked = FALSE;  /* don't delete any local marked text */
		xmEditboxRequestSelection(w, XFER_BIN_SELECTION);
		w->deleteMarked = TRUE;
    }
}

manageButtonRelease(w, eventP, x, y)
AXWidget_editbox w;
XEvent *eventP;
int x, y;
{
    tcursor_t *cursor;
    seln_t *sel;
    any_obj_t *op;
    tline_t *lp;
    int char_index;
    int button_released;

    button_released = (int)eventP->xbutton.button;
    if (button_released != Button1)
	return;

    EditBoxArmed = (AXWidget_editbox)NULL;
	
    /* set new cursor position */
    cursor = &w->editbox->tcursor;
    sel = &w->editbox->sel;
    if (w->pending_delete) {
        /* if something is hilit, cursor is at end of stroke */
        cursor->obj = sel->endObj;
        cursor->ac_index = sel->endIndex;
    }
    else {
        /* make sure cursor is visible in window 
         * (ButtonRelease may be outside window if 
         * XA_PRIMARY could not be acquired) */
        getNearestVisibleIndex(w, x, y, &op, &lp, &char_index);
        cursor->obj = op;
        cursor->ac_index = char_index;
    }
    drawCursor(w, TC_DRAW);
}

manageMouseMotion(w, eventP, x, y)
AXWidget_editbox w;
XEvent *eventP;
int x, y;
{
    any_obj_t *op;
    tline_t *lp;
    int char_index;
    tcursor_t *cursor, *dummy;
    seln_t *sel;
	
    if (AxButtonDown != Button1)
		return;
	
    cursor = &w->editbox->tcursor;
    sel = &w->editbox->sel;

	if (eventP) {
		if (!EditBoxArmed) {
			/* test whether to start autoscroll
			 */
			editboxAutoscrollEdge = mouseOutsideEditbox(w, eventP);
			if (noEdge != editboxAutoscrollEdge) {
				EditBoxArmed = w;    /* start autoscroll */
			}
		}
		else if (EditBoxArmed) {
			/* test whether to stop autoscroll
			 */
			editboxAutoscrollEdge = mouseOutsideEditbox(w, eventP);
			if (noEdge == editboxAutoscrollEdge) {
				EditBoxArmed = (AXWidget_editbox) NULL;
			}
		}
	}
	else if (EditBoxArmed) {
		/* no real event available, so use a query-pointer
		 * to test whether to stop autoscroll
		 */
		Window root_return, child_return;
		XEvent fake_event;
		if (XQueryPointer(Dpy, w->window, &root_return, &child_return,
						  &fake_event.xbutton.x_root,
						  &fake_event.xbutton.y_root,
						  &fake_event.xbutton.x, &fake_event.xbutton.y,
						  &fake_event.xbutton.state)) {
			if (fake_event.xbutton.state & Button1Mask) {
				editboxAutoscrollEdge = mouseOutsideEditbox(w, &fake_event);
				if (noEdge == editboxAutoscrollEdge) {
					EditBoxArmed = (AXWidget_editbox) NULL;
				}
			}
		}
	}
    
    /*  hilight only if seln atom (XA_PRIMARY) owner */
    if (w->pending_delete) {

        if ( w->num_multi_clicks>1 && eventP &&
			(eventP->xmotion.time-w->last_click_time)<=DoubleMillisec )
            return;
		if ((y + w->editbox->lineHeight < 0) || (y > w->editbox->height)) {
			int ypos, old_scroll_pos;
			old_scroll_pos = w->scroll_pos;
			if (y + w->editbox->lineHeight < 0) {

				/* if we're already at the top, just exit */
				if (w->scroll_pos <= 0)
					return;
			
				cursor->obj = sel->endObj;
				cursor->ac_index = sel->endIndex;
				moveCursorVert(w, TC_UP, 1);
				if ((w->scroll_pos -= w->editbox->lineHeight) < 0)
					w->scroll_pos = 0;
			}
			else {
				cursor->obj = sel->endObj;
				cursor->ac_index = sel->endIndex;
				moveCursorVert(w, TC_DOWN, 1);
				getLineFromIndex(w, cursor->obj, cursor->ac_index, &lp);
				ypos = w->editbox->yoffset + cursor->obj->ypos + lp->ypos;
				if ((w->scroll_pos += w->editbox->lineHeight) >
					(w->scroll_max - w->editbox->height +
					 w->editbox->lineHeight))
					w->scroll_pos =
						w->scroll_max - w->editbox->height +
							w->editbox->lineHeight;
			}
			if (old_scroll_pos == w->scroll_pos)
				return;
	    
			/* dispCursorAtCenter(w);*/
			drawCursor(w, TC_ERASE);
		
			PaintEditBox(w, NULL, TRUE, TRUE);
	    
			if (y < 0)
				hilightForMotion(w, 0, 0);
			else
				hilightForMotion(w, w->editbox->width, w->editbox->height);
	    
		}
		else {
			hilightForMotion(w, x, y);
		}
    }
}

void EditBoxAutoScroll(w)
AXWidget_editbox w;
{
	int xx = 0, yy = 0;
	
	/* fabricate a coordinate that looks like it's off
	 * the appropriate edge, then pass to motion handler.
	 */
	xx = w->width / 2;
	switch (editboxAutoscrollEdge) {
	case topEdge:
		yy = -20;
		manageMouseMotion(w, NULL, xx, yy);
		break;
	case bottomEdge:
		yy = w->height + 20;
		manageMouseMotion(w, NULL, xx, yy);
		break;
	}
}

hilightForMotion(w, x, y)
AXWidget_editbox w;
int x, y;
{
    any_obj_t *op;
    tline_t *lp;
    seln_t *sel;
    int char_index, start, end;

    sel = &w->editbox->sel;
    getNearestVisibleIndex(w, x, y, &op, &lp, &char_index);

    if (sel->startObj == sel->endObj) {  
        /* !! this needs to be generalized so that !!
         * !! start and end objs can be different  !! */
        start = sel->startIndex;
        end = sel->endIndex;
        sel->endIndex = char_index;
        if (char_index > end) {
            if (end > start)
                paintText(w, op, end, char_index, GC_HILIT);
            else {
                paintText(w, op, end, char_index, GC_FORE);
                if (char_index > start)
                    paintText(w, op, start, char_index, GC_HILIT);
            }
        }
        else {
            if (end > start) {
                paintText(w, op, char_index, end, GC_FORE);
                if (char_index < start)
                    paintText(w, op, char_index, start, GC_HILIT);
            }
            else
                paintText(w, op, char_index, end, GC_HILIT);
        }
    }
}

adjustHilight(w, x, y)
AXWidget_editbox w;
int x, y;
{
    any_obj_t *op;
    tline_t *lp;
    seln_t *sel;
    int char_index, start, end;

    sel = &w->editbox->sel;
    getNearestVisibleIndex(w, x, y, &op, &lp, &char_index);

    start = sel->startIndex;
    end = sel->endIndex;
    if (start == end) {
        drawCursor(w, TC_ERASE);
    }
    else if (start > end) {
        /* exchange start & end */
        sel->startIndex = end;
        sel->endIndex = start;
        start = sel->startIndex;
        end = sel->endIndex;
    }

    if ( abs(end - char_index) < abs(start - char_index) ) {
        /* new index is closer to endIndex */
        if (end == start) {
            paintText(w, op, char_index, end, GC_HILIT);
        }
        else {
            if (char_index < end)
                paintText(w, op, char_index, end, GC_FORE);
            else
                paintText(w, op, end, char_index, GC_HILIT);
        }
    }
    else {
        /* new char_index is closer to startIndex */
        if (end == start) {
            paintText(w, op, start, char_index, GC_HILIT);
        }
        else {
            if (char_index > start)
                paintText(w, op, start, char_index, GC_FORE);
            else
                paintText(w, op, char_index, start, GC_HILIT);
        }
        sel->startIndex = end;
    }
    /* reposition end of hilight to index of button-press 
     * so that future motion will result in proper 
     * adjustment of hilight */
    sel->endIndex = char_index;
}

hilightWord(w, x, y)
AXWidget_editbox w;
int x, y;
{
    any_obj_t *op;
    tline_t *lp;
    seln_t *sel;
    int char_index, ix;

    if (!w->pending_delete)
        return;

    sel = &w->editbox->sel;
    drawCursor(w, TC_ERASE);
    /* unhilight existing mark (start and end index should be same 'cos
     * first button click resets selection info - unhilight for sanity) */
    paintText(w, sel->startObj, sel->startIndex, sel->endIndex, GC_FORE);

    /* hilight word */
    getNearestVisibleIndex(w, x, y, &op, &lp, &char_index);
    if (!lp) {
        sel->startIndex = sel->endIndex = w->editbox->tcursor.ac_index;
        return;
    }
    sel->startIndex = lp->first_index;
    sel->endIndex = lp->first_index + lp->len;
    /* find nearest word boundaries */
    for (ix = lp->first_index; ix < lp->first_index+lp->len; ix++) {
        if ( *(op->wdata + ix) != ' ' )
            continue;
        if (ix < char_index) {
            if (ix > sel->startIndex)
                sel->startIndex = ix;
        }
        else {
            sel->endIndex = ix;
            break;
        }
    }
    /* don't hilight leading space */
    if ( *(op->wdata + sel->startIndex) == ' ' )
        sel->startIndex++;
    paintText(w, op, sel->startIndex, sel->endIndex, GC_HILIT);
}

hilightLine(w, x, y)
AXWidget_editbox w;
int x, y;
{
    any_obj_t *op;
    tline_t *lp;
    seln_t *sel;
    int char_index;

    if (!w->pending_delete)
        return;

    sel = &w->editbox->sel;
    drawCursor(w, TC_ERASE);
    /* unhilight existing mark */
    paintText(w, sel->startObj, sel->startIndex, sel->endIndex, GC_FORE);

    /* hilight line */
    getNearestVisibleIndex(w, x, y, &op, &lp, &char_index);
    if (!lp) {
        sel->startIndex = sel->endIndex = w->editbox->tcursor.ac_index;
        return;
    }
    sel->startIndex = lp->first_index;
    sel->endIndex = lp->first_index + lp->len;
    paintText(w, op, sel->startIndex, sel->endIndex, GC_HILIT);
}

hilightPara(w, x, y)
AXWidget_editbox w;
int x, y;
{
    any_obj_t *op;
    tline_t *lp, *tmplp;
    seln_t *sel;
    int char_index, ix, end;

    if (!w->pending_delete)
        return;

    op = w->editbox->objs;
    sel = &w->editbox->sel;
    drawCursor(w, TC_ERASE);
    /* unhilight existing mark */
    paintText(w, sel->startObj, sel->startIndex, sel->endIndex, GC_FORE);

    getNearestVisibleIndex(w, x, y, &op, &lp, &char_index);
    if (!lp) {
        sel->startIndex = sel->endIndex = w->editbox->tcursor.ac_index;
        return;
    }
    sel->startIndex = 0;
    sel->endIndex = op->wsize;
    /* find nearest newlines */
    for (tmplp = op->lines; tmplp; tmplp = tmplp->next) {
        end = tmplp->first_index + tmplp->len;
        if ( *(op->wdata + end) != '\n' )
            continue;
        ix = end + 1;
        if (ix < char_index) {
            if (ix > sel->startIndex)
                sel->startIndex = ix;
        }
        else {
            sel->endIndex = end;
            break;
        }
    }
    paintText(w, op, sel->startIndex, sel->endIndex, GC_HILIT);
}

hilightDoc(w, x, y)
AXWidget_editbox w;
int x, y;
{
    any_obj_t *op;
    seln_t *sel;

    if (!w->pending_delete)
        return;

    op = w->editbox->objs;
    sel = &w->editbox->sel;
    drawCursor(w, TC_ERASE);

    sel->startIndex = 0;
    sel->endIndex = op->wsize;
    paintText(w, op, sel->startIndex, sel->endIndex, GC_HILIT);
}

int
editboxSelectionEvent(w, eventP)
AXWidget_editbox w;
XEvent *eventP;
{
	int consumed = FALSE;
    switch (eventP->type) {
        case SelectionRequest:
            supplySelection(w, eventP);
			consumed = TRUE;
            break;
        case SelectionNotify:
            if (!w->no_picks) {
                insertSelection(w, eventP);
				consumed = TRUE;
			}
            break;
        case SelectionClear:
            clearSelection(w, eventP->xselectionclear.selection, FALSE);
			consumed = TRUE;
            break;
    }
	return (consumed);
}

/*
 * initSelection - acquire selection atom (XA_PRIMARY/XA_CLIPBOARD)
 *                 Directly notify ATM or editbox window of
 *                 ownership transfer if it was previous owner.
 */
int
initSelection(w, selection)
AXWidget_editbox w;
Atom selection;
{
    AXWidget wp;
    Window win;

    if ( !w->window )
        return(FALSE);
    if (w->no_primary_seln)
        return(TRUE); /* don't grab primary selection, but allow
                         hilighting. */

    if ((win=XGetSelectionOwner(Dpy, selection)) != w->window) {
        XSetSelectionOwner(Dpy, selection, w->window, CurrentTime);
        if (XGetSelectionOwner(Dpy, selection) == w->window) {
            if (ATMWindowFromXWindow(win)) {
                /* previous owner was ATM win */

                /* this call relies on the fact that editbox
                 * windows are not ATM windows ( AxClearSelection
                 * calls editboxClearSelection if 'win' is not ATM
                 * window - will nullify this acquisition!! ) */

                AxClearSelection(selection, win);
            }
            else {
                /* check if previous owner was an editbox */
                for ( wp = WindowList; wp; wp = wp->nextWin ) {
                    if ( wp->type==W_EDITBOX && wp->window==win ) {
                        clearSelection((AXWidget_editbox)wp, selection, FALSE);
                        break;
                    }
                }
            }
            return(TRUE);
        }
        else {
            return(FALSE);
        }
    }
    else {
        return(TRUE);
    }
}

/*
 * editboxClearSelection - called when a ATM window acquires a seln
 *                         If owner, unhilight / destroy clipboard.
 *                         (this call relies on the fact that 
 *                         editbox windows are not ATM windows).
 */
editboxClearSelection(win, selection)
Window win;      /* previous owner of 'selection' */
Atom selection;  /* XA_PRIMARY, XA_CLIPBOARD */
{
    AXWidget wp;

    for ( wp = WindowList; wp; wp = wp->nextWin ) {
        if ( wp->type==W_EDITBOX && wp->window==win ) {
            clearSelection((AXWidget_editbox)wp, selection, FALSE);
        }
    }
}

/*
 * clearSelection - unhilight selection / destroy clipboard without 
 *                  changing contents of editbox
 */
clearSelection(w, selection, clear_if_select_mode)
AXWidget_editbox w;
Atom selection;
int clear_if_select_mode;
{
    seln_t *sel;

    if (!w->pending_delete)
        return;
    if (w->no_primary_seln && clear_if_select_mode == FALSE)
        return;

    /* unhilight existing selection */
    /* !! currently assumes selection is within one text obj !! */
    sel = &w->editbox->sel;
    paintText(w, sel->startObj, sel->startIndex, sel->endIndex, GC_FORE);

    /* restore selection info */
    resetSelectionInfo(w);

    drawCursor(w, TC_DRAW);
}

/*
 * changeSelection - change existing selection and update display
 */
changeSelection(w, startObj, endObj, startIndex, endIndex)
AXWidget_editbox w;
any_obj_t *startObj;
any_obj_t *endObj;
int startIndex;
int endIndex;
{
    any_obj_t *op;
    tline_t *lp;
    tcursor_t *cursor;
    seln_t *sel;
    
    sel = &w->editbox->sel;

    /* !! currently assumes selection is within one text obj !! */

    /* clear old cursor/selection from display */
    if ( sel->startIndex != sel->endIndex ) {
        /* clear old selection */
        paintText(w, sel->startObj, sel->startIndex, sel->endIndex, GC_FORE);
    }
    else {
        /* erase cursor */
        drawCursor(w, TC_ERASE);
        initSelection(w, XA_PRIMARY);
    }

    /* set new cursor/selection and update display */
    if ( startObj && (startObj == endObj) && (startIndex == endIndex) ) {
        /* set cursor */
        resetSelectionInfo(w);
        getCursorInfo(w, &cursor, &op, &lp);
        cursor->obj = startObj;
        cursor->ac_index = MIN(startIndex, (startObj->wsize-1));
        cursor->ac_index = MAX(startIndex, 0);
        drawCursor(w, TC_DRAW);
    }
    else {
        /* set new selection */
        w->pending_delete = TRUE;
        w->deleteMarked = TRUE;
        sel->startObj = startObj;
        sel->endObj = endObj;
        sel->startIndex = MIN(startIndex, (startObj->wsize-1));
        sel->startIndex = MAX(startIndex, 0);
        sel->endIndex = MIN(endIndex, (endObj->wsize-1));
        sel->endIndex = MAX(endIndex, 0);
        /* hilight new selection */
        paintText(w, sel->startObj, sel->startIndex, sel->endIndex, GC_HILIT);
    }
}

/*
 * saveSelection - copy selection to clipboard
 */
saveSelection(w)
AXWidget_editbox w;
{
    seln_t *sel;
    mbchar *p;
    int length, start, new_len;

    /* save only if selection active */
    if (!w->pending_delete)
        return;

    /* save selection contents in XA_STRING format
	 */
    sel = &w->editbox->sel;
    start = MIN(sel->startIndex, sel->endIndex);
    length = abs(sel->endIndex - sel->startIndex);
    p = (mbchar *)malloc( (length+1) * sizeof(wchar) );
    new_len = WnToMbStr( sel->startObj->wdata+start, p, length );
    p[new_len] = '\0';

	/* put data on clipboard
	 */
	ElfbSetSelection(AxMakeStrData(-1, (mbchar *)(NXLT("CLIPBOARD")) ),
					 AxMakeStrData(-1, p),
					 AxMakeNumData(AX_DOC_TYPE_ASTERX_ELFARRAY));

}

/*
 * supplySelection - supply selection data in response to 
 *                   SelectionRequest event
 */
static int
supplySelection(w, eventP)
AXWidget_editbox w;
XEvent *eventP;
{
	Window requestor;
	Atom selection, target, property;
	XSelectionEvent event;
	seln_t *sel;
	int length, new_len;
	
	requestor = eventP->xselectionrequest.requestor;
	selection = eventP->xselectionrequest.selection;
	target = eventP->xselectionrequest.target;
	property = eventP->xselectionrequest.property;
	
	event.type = SelectionNotify;
	event.serial = eventP->xselectionrequest.serial;
	event.send_event = eventP->xselectionrequest.send_event;
	event.display = eventP->xselectionrequest.display;
	event.requestor = requestor;
	event.selection = selection;
	event.target = target;
	event.property = None; /* reset to supplied prop if successful */
	event.time = eventP->xselectionrequest.time;
		
    if (selection == XA_PRIMARY && w->pending_delete) {
        if (target == XA_STRING) {
            mbchar *p;
            mbchar *sel_buf;
            int start;
            sel = &w->editbox->sel;
            start = MIN(sel->startIndex, sel->endIndex);
            length = abs(sel->endIndex - sel->startIndex);
            p = (mbchar *) malloc( (length+1)*sizeof(wchar) ); 
            new_len = WnToMbStr( sel->startObj->wdata+start, p, length );
            p[new_len] = '\0';
            sel_buf = p;
            XChangeProperty(Dpy, requestor, property, target, 8,
                             PropModeReplace, sel_buf, MBstrlen(sel_buf));
            event.property = eventP->xselectionrequest.property;
        }
		XSendEvent(Dpy, requestor, True, 0, (XEvent *)&event);
    }
}

/*
 * xmEditboxRequestSelection - request X clients for selection data.
 */
xmEditboxRequestSelection(w, selname)
AXWidget_editbox w;
int selname;              /* e.g., XFER_BIN_CLIPBOARD */
{
    any_obj_t *op;
    seln_t *sel;
    wchar *buf;
    int min_index, len, new_len;

    if ( !w->window )
        return;

	if (XFER_BIN_SELECTION == selname) {
		/* try for XA_PRIMARY selection
		 */
		op = w->editbox->objs;
		sel = &w->editbox->sel;

		if (w->pending_delete) {
			/* local XA_PRIMARY */
			if (sel->startIndex != sel->endIndex) {
				/* something is hilit locally */
				min_index = MIN(sel->startIndex, sel->endIndex);
				len = abs(sel->endIndex - sel->startIndex);
				buf = (wchar *)get_static_buffer( (len+1)*sizeof(wchar) );
				memcpy(buf, op->wdata+min_index, (len*sizeof(wchar)) );
				clearSelection(w, XA_PRIMARY, FALSE);
				manageInsert(w, buf, len);
				return;
			}
		}
		else {
			XConvertSelection(Dpy, XA_PRIMARY, XA_STRING, 
							  XA_EDITBOX_PROP, w->window, CurrentTime);
			return;
		}
	}
	else if (XFER_BIN_CLIPBOARD == selname) {
		data_ptr strsP;
		ElfbGetSelection(AxMakeStrData(-1, (mbchar *)(NXLT("CLIPBOARD")) ),
						 (data_ptr) NULL,
						 AxMakeNumData(AX_DOC_TYPE_ASTERX_ELFARRAY));
		strsP = ElfRetData;    /* ancient global return plumbing, yuk.... */
		xmInsertElfSelection((AXWidget) w, strsP);
	}
}

/*
 * insertSelection - insert selection data provided by X client
 */
static int
insertSelection(w, eventP)
AXWidget_editbox w;
XEvent *eventP;
{
    Atom property, actual_type;
    long long_size;
    int retval, actual_format;
    unsigned long nitems, bytes_after, *data;
    wchar *buf, *newbuf;
    int len;

    if ( !w->window )
        return;

    property = eventP->xselection.property;
    if (property==None) {
		int seltype;
		if (XA_PRIMARY == eventP->xselection.selection) {
			seltype = XFER_BIN_SELECTION;
		}
		else {
			seltype = XFER_BIN_CLIPBOARD;
		}
        xmEditboxRequestSelection(w, seltype);
        return;
    }
    if (property!=XA_EDITBOX_PROP)
        return;

    /* make initial guess for data size */
    long_size = MAX_XSTR/sizeof(long);

    /* If we use XA_STRING as req_type but actual_type 
     * is someting else, a BadMatch eror is generated. To
     * avoid the error, we use AnyPropertyType and then 
     * confirm that actual_type is XA_STRING */
    retval = XGetWindowProperty(Dpy, w->window, property, 0, 
                    long_size, False, AnyPropertyType, &actual_type, 
                    &actual_format, &nitems, &bytes_after, 
                    (unsigned char **)&data);

    if (retval != Success)
        return;
    if (actual_type != XA_STRING) {
		XDeleteProperty(Dpy, w->window, property);
        return;
	}

    /* insert data at cursor */
insert_stuff:
    buf = (wchar *)get_static_buffer( (nitems+1)*sizeof(wchar) );
    len = MbnToWStr( (mbchar *)data, buf, nitems );    /* len is in chars */

    /*
	 * validateConvertChars converts tabs to 4 spaces which at
	 * worst will quadruple the length of the returned string;
	 * add one more char for <CR> --> <CR><LF> conversion on NT
	 */

    newbuf = (wchar *) calloc(1, ((len*4+2)*sizeof(wchar)));

	validateConvertChars(buf, newbuf, len);
	len = Wstrlen(newbuf);

    manageInsert(w, newbuf, len);
	free(newbuf);
    /* free data */
    XFree((caddr_t)data);

    /* read remaining data if partial read performed */
    if (bytes_after) {
        /* repeat above process */
        retval = XGetWindowProperty(Dpy, w->window, property, 
                    nitems*actual_format/32, bytes_after, False, 
                    AnyPropertyType, &actual_type, &actual_format, 
                    &nitems, &bytes_after, (unsigned char **)&data);
        if (retval != Success)
            return;
        goto insert_stuff;
    }
	XDeleteProperty(Dpy, w->window, property);
}



/***********************************************************************
function: GetEditBoxSelectedRegion - get co-ordinates of selected
region in the edit box in row-number, column-number form.
returns: number of characters selected.
*/
int
GetEditBoxSelectedRegion(w, start_rowp, start_colp, end_rowp, end_colp)
AXWidget_editbox w;
int *start_rowp;
int *start_colp;
int *end_rowp;
int *end_colp;
{
    any_obj_t *op;
    tline_t *lp;
    tcursor_t *cursor;
    seln_t *sel;
    register int start_idx;
    register int end_idx;
    
    if (!w || (w->type != W_EDITBOX) || !w->editbox->objs)
        goto errout;
    
    sel = &w->editbox->sel;
    
    if ((sel->startObj == NULL) || (sel->endObj == NULL)) {
        /* no selection */
        if ( w != (AXWidget_editbox)ElfbGetActiveWidget((AXWidget)w->parent) )
	{
            /* !! HACK: shouldn't use ElfbGetActiveWidget */
            /* active cursor not in this editbox */
            start_idx = 0;
            end_idx = 0;
        }
        else {
            /* this editbox is active widget */
            getCursorInfo(w, &cursor, &op, &lp);
            start_idx = cursor->ac_index;
            end_idx = cursor->ac_index;
        }
    }
    else {
        /* have selection */
        start_idx = MIN(sel->startIndex, sel->endIndex);
        end_idx   = MAX(sel->startIndex, sel->endIndex);
    }
    if (GetRowColAddrOfIndex(w, start_idx, start_rowp, start_colp) != 0)
        goto errout;
    if (GetRowColAddrOfIndex(w, end_idx, end_rowp, end_colp) != 0)
        goto errout;
    return(abs(end_idx - start_idx));
errout:
    return(-1);
}

/***********************************************************************
function: GetRowColAddrOfIndex - given an index into the the
edit-box's data, compute the corresponding row-column numbers
by virtually breaking the data into an elf-array-of-strings.
See Also: ElfReturnArrayOfStrings() in axelf/elbltn3.c
returns: 0 on success, -1 on failure (bad index...)
*/
static int
GetRowColAddrOfIndex(w, idx, rowp, colp)
AXWidget_editbox w;
int idx;
int *rowp;
int *colp;
{
    register wchar *dp;
    register wchar *point;
    register int rownum;
    register int colnum;
    register int incr_row;
    register wchar *data; 
	tline_t *lp;

    if (!w || (w->type != W_EDITBOX) || !w->editbox->objs)
        return(-1); /* bad widget */

    if ((idx < 0) || (idx > w->editbox->objs->wsize))
        return(-1); /* index out of range */

    data = w->editbox->objs->wdata;

    if (!data)
        return(-1); /* no data */

	lp = w->editbox->objs->lines;
    for (dp = data, rownum = 0, colnum = 0, point = &data[idx];
         *dp != (wchar)NULL; dp++) {
		
        if (dp == point)
            break;
	
		incr_row = TRUE;	
        if ( *dp != '\n') {
			incr_row = FALSE;
           	colnum++;
			if (w->maxLen > 0 && lp && colnum > lp->len) 
				incr_row = TRUE;
		}

        if (incr_row) {
			lp = lp->next;
            rownum++;
            colnum = 0;
		}
    }
	
	if (dp == point) {
		*rowp = rownum;
		*colp = colnum;
		return (0);
	}
	else {
		return(-1); /* point not in data */
	}
}

/***********************************************************************
function: MarkEditBoxRegion - mark a region in the edit-box given
starting and ending address in row-column form.
returns: 0 on success, -1 on failure.
*/
int
MarkEditBoxRegion(w, start_row, start_col, end_row, end_col)
AXWidget_editbox w;
int start_row;
int start_col;
int end_row;
int end_col;
{
    register wchar *datap;
    register int startix, endix;
	register int ypos;
    tline_t *lineP;
    
    if (!w || (w->type != W_EDITBOX) || !w->editbox->objs)
        return(0);
    
    startix = GetIndexFromRowColAddr(w, start_row, start_col);
    endix   = GetIndexFromRowColAddr(w, end_row, end_col);
	if ((startix == -1) || (endix == -1)) {
        return(-1); /* row-col address out of bounds */
	}

    changeSelection(w, w->editbox->objs, w->editbox->objs, startix, endix);

	getLineFromIndex(w, w->editbox->objs, startix, &lineP);
	ypos = w->editbox->yoffset + w->editbox->objs->ypos + lineP->ypos;

	if ((ypos < 0) || (ypos + w->editbox->lineHeight > w->editbox->height)) {
		if (startix != endix) {
			w->editbox->tcursor.ac_index = startix;
			w->editbox->tcursor.obj = w->editbox->objs;
		}
		dispCursorAtCenter(w);
	}
	
    PaintEditBox(w, NULL, TRUE, TRUE);
    return(0);
}

/***********************************************************************
function: GetIndexFromRowColAddr - return the index in w->editbox->objs->wdata
corresponding to the given row-column address in the edit-box.
returns: index, or -1 if row-col addr is out of range.  The index which refers
to one position past the end of w->editbox->objs->wdata is considered in-range.

Note the assumption that non-null w->editbox->objs is always
a list containing a single any_obj_t instance which holds all the data
*/
static int GetIndexFromRowColAddr(w, row, col)
AXWidget_editbox w;
register int row;
register int col;
{
	register int ix;
    register int rownum;
    register int colnum;
    register int incr_row;
	tline_t *lp;
    
    if (!w || (w->type != W_EDITBOX) || !w->editbox->objs) {
        return( -1 ); /* bad widget */
	}
    
    if (!w->editbox->objs->wdata) {
        return( -1 ); /* no data */
	}

	lp = w->editbox->objs->lines;
	rownum = 0;
	colnum = 0;
    for (ix = 0; ix <= w->editbox->objs->wsize; ix++) {
        if ((rownum == row) && (colnum == col)) {
            return( ix );
		}
		incr_row = TRUE;
        if ( w->editbox->objs->wdata[ix] != '\n' ) {
			incr_row = FALSE;
            colnum++;
			if (w->maxLen && lp && colnum > lp->len)
				incr_row = TRUE;
        }
        if (incr_row) {
            /* start a new line */
			lp = lp->next;
            rownum++;
            colnum = 0;
        }
	}
	
	if (ix > w->editbox->objs->wsize) {
		return ( -1 );
	}
	
	return (ix);
}

extern KeySym AxEuroDollarKeysym;							/* method 1 */
extern int AxEuroDollarModifiers;							/* method 1 */
extern int AxEuroDollarKeyState, AxEuroDollarKeyCode ;		/* method 2 */

/* ---------------- */

int
editboxKeyEvent(w, eventP, in_menu_bar)
AXWidget_editbox w;
XEvent *eventP;
int in_menu_bar;
{
    KeySym ksym;
    wchar *str; 
    register int num_chars;
    register wchar *p1, *p2;
    register int cnt;
    wchar buff[5];
    AXWidget_main main_w;
    XKeyEvent *kep;
    int have_changed_it = FALSE;
	int consumed = FALSE;
	int euro = FALSE;

	if (!w || !eventP)
		return( FALSE);

    if ( eventP->type != KeyPress )
        return (FALSE);

	if (w->type != W_EDITBOX)
        return (FALSE);

#if BUILD_ORIENT
    main_w = TheMainWidget( (AXWidget)w );
#endif

	if (AxEuroDollarKeyCode && (AxEuroDollarKeyState == eventP->xkey.state) && 
								(AxEuroDollarKeyCode == eventP->xkey.keycode)) {
		/* we have found the euro dollar key sequence, return back EuroDollar
			ascii */
		buff[0] = EURO_SYMBOL_8BIT_VALUE;
		buff[1] = 0;
		str = buff;
		num_chars = 1;
		ksym = 0;
		euro = TRUE;
	} else {
		num_chars = AXLookupString16(eventP, &str, &ksym);    
    	if ( (num_chars == 0) && ((ksym == 0) || AxKeysymIsModifier(ksym)) )
        	return (FALSE);
	}

	/* High-level Euro keysym - hardware-independent */
	if (AxEuroDollarKeysym && (AxEuroDollarKeysym == ksym)) {
		int mods = AxDecodeModifiers(eventP->xkey.state);
		if (mods == AxEuroDollarModifiers) {
			buff[0] = EURO_SYMBOL_8BIT_VALUE; 
			buff[1] = 0;
			str = buff;
			num_chars = 1;
			ksym = 0;
			euro = TRUE;
		}
	}

	/* If we didn't find the Euro, and the keystroke is or may be in
		a menu bar, reject the event */
	if (in_menu_bar && !euro && 
							!(XK_Return == ksym || XK_Execute == ksym))
		return( FALSE);

    if ( w->no_picks ) {
		switch (ksym) {
		case XK_Down:
			w->scroll_pos = MIN(w->scroll_max - w->editbox->lineHeight,
								w->scroll_pos + w->editbox->lineHeight);
			if (w->scroll_pos != (-w->editbox->yoffset)) {
				PaintEditBox(w, NULL, TRUE, TRUE);
				consumed = TRUE;
			}
			break;
		case XK_Up:
			w->scroll_pos = MAX(0, w->scroll_pos - w->editbox->lineHeight);
			if (w->scroll_pos != (-w->editbox->yoffset)) {
				PaintEditBox(w, NULL, TRUE, TRUE);
				consumed = TRUE;
			}
			break;
		case XK_Next:
			w->scroll_pos = MIN(w->scroll_max - w->editbox->lineHeight,
								w->scroll_pos + w->editbox->height -
								w->editbox->lineHeight);
			if (w->scroll_pos != (-w->editbox->yoffset)) {
				PaintEditBox(w, NULL, TRUE, TRUE);
				consumed = TRUE;
			}
			break;
		case XK_Prior:
			w->scroll_pos = MAX(0, w->scroll_pos - w->editbox->height +
								w->editbox->lineHeight);
			if (w->scroll_pos != (-w->editbox->yoffset)) {
				PaintEditBox(w, NULL, TRUE, TRUE);
				consumed = TRUE;
			}
			break;
		}
        return (consumed);
	}

    /* handle all special keys first */
    switch (ksym) {
	case XK_Next:
		moveCursorVert(w, TC_DOWN,
					   (w->editbox->height/w->editbox->lineHeight)-1);
		w->scroll_pos = MIN(w->scroll_max - w->editbox->lineHeight,
							w->scroll_pos + w->editbox->height -
							w->editbox->lineHeight);
		PaintEditBox(w, NULL, TRUE, TRUE);
		consumed = TRUE;
		break;
	case XK_Prior:
		moveCursorVert(w, TC_UP, (w->editbox->height/w->editbox->lineHeight)-1);
		w->scroll_pos = MAX(0, w->scroll_pos - w->editbox->height +
							w->editbox->lineHeight);
		PaintEditBox(w, NULL, TRUE, TRUE);
		consumed = TRUE;
		break;
	case XK_Left:
		moveCursor(w, TC_LEFT, 1);
		consumed = TRUE;
		break;
	case XK_Up:
		moveCursor(w, TC_UP, 1);
		consumed = TRUE;
		break;
	case XK_Right:
		moveCursor(w, TC_RIGHT, 1);
		consumed = TRUE;
		break;
	case XK_Down:
		moveCursor(w, TC_DOWN, 1);
		consumed = TRUE;
		break;
	case XK_Delete:
	case XK_BackSpace:
		manageDelete(w, 1);
		have_changed_it = TRUE;
		consumed = TRUE;
		break;
	case XK_Tab:
		return (FALSE);
		break;
	case XK_Return:
		buff[0] = '\n';
		buff[1] = 0;
		manageInsert(w, buff, 1);
		have_changed_it = TRUE;
		consumed = TRUE;
		break;
#if BUILD_ORIENT && 0
	case CONVERSION_TOGGLE:
		ToggleDialogConversionState( main_w );
		consumed = TRUE;
		return;
		
	case LEARN_TOGGLE:
	case CODE_INPUT_TOGGLE:
	case MANUAL_AOTO_TOGGLE:
	case INPUT_MODE_CHANGE:
		if ( ((XKeyEvent *)eventP)->state & ControlMask )
			/* update the conversion status */
			UpdateWidgetStatus(  main_w, (EGC_JMODE *)NULL,
								 !main_w->inWideInput );
		consumed = TRUE;
		return;
#endif
    }

    /* strip non-graphic characters */
    p1 = p2 = str;
    while ( *p2 != 0 )
       {
       if ( IS_NGC( *p2 ) )
          p2++;
       else
          *p1++ = *p2++;
       }
    *p1 = 0;

    cnt = p1 - str;
    if ( cnt > 0 ) {  /* we have at least one printable char, insert */
        manageInsert(w, str, cnt);
        have_changed_it = TRUE;
    }
    if ( have_changed_it && w->callback && (w->flags & WFLG_TOUCHED_RETURN) )
        AxCallback(w->callback,(AXWidget)w,w->callbacktag,NULL);

    return (consumed);
}

/***************************************************************************
function: EditBoxBackwardDeleteChar - Delete char to left of cursor
*/
EditBoxBackwardDeleteChar (w)
AXWidget_editbox w;
{
    if ( !w )
        return;
    if ( w->no_picks )
        return;

    manageDelete(w, 1);
}

/***************************************************************************
function: EditBoxForwardDeleteChar - Delete char to right of cursor
*/
EditBoxForwardDeleteChar (w)
AXWidget_editbox w;
{
    if ( !w )
        return;
    if ( w->no_picks )
        return;

    /* move cursor right by one char */
	if (w->editbox && w->editbox->sel.startIndex == w->editbox->sel.endIndex) {
        if ( !moveCursor(w, TC_RIGHT, 1) )
            return;  /* cursor did not move */
    }

    manageDelete(w, 1);
}




/*
 * moveCursor - move and redraw cursor
 *              Return whether cursor moved
 *              (cursor is always in a text object)
 */
int
moveCursor(w, direction, num)
AXWidget_editbox w;
int direction;
int num; /* #chars to move cursor */
{
    tcursor_t *cursor;
    any_obj_t *op;
    tline_t *lp;
    int ypos;
    int start_index;

    /* clear selection (if any) */
    clearSelection(w, XA_PRIMARY, FALSE);

    /* make sure cursor is visible before moving cursor */
    getCursorInfo(w, &cursor, &op, &lp);
    start_index = cursor->ac_index;
    ypos = w->editbox->yoffset + op->ypos + lp->ypos;
	if (ypos + w->editbox->lineHeight < 0 || ypos > w->editbox->height) {
        dispCursorAtCenter(w);
        PaintEditBox(w, NULL, TRUE, TRUE);
    }

    /* erase old cursor */
    drawCursor(w, TC_ERASE);

    /* update cursor position */
    if (direction==TC_RIGHT || direction==TC_LEFT) { /* horz movement */
        moveCursorHorz(w, direction, num);
    }
    else { /* vertical movement */
        moveCursorVert(w, direction, num);
    }

    /* draw new cursor */
    drawCursor(w, TC_DRAW);

    /* check if scroll is needed */
    /* !! currently assumes that for down movement num=1 !! */
    getCursorInfo(w, &cursor, &op, &lp);
    ypos = w->editbox->yoffset + op->ypos + lp->ypos;
    if ( ypos + w->editbox->lineHeight > w->editbox->height ) {
        w->scroll_pos += w->editbox->lineHeight; /* according to assumption */
        PaintEditBox(w, NULL, TRUE, TRUE);
    }
    else if ( ypos < 0 ) {
        w->scroll_pos = op->ypos + lp->ypos;
        PaintEditBox(w, NULL, TRUE, TRUE);
    }

    /* return whether cursor moved */
    return(start_index != cursor->ac_index);
}

moveCursorHorz(w, direction, num)
AXWidget_editbox w;
int direction;
int num; /* #chars to move cursor */
{
    any_obj_t *op;
    tline_t *lp;
    tcursor_t *cursor;
    int char_index;
    int inc, newindex;

    /* get object containing cursor */
    getCursorInfo(w, &cursor, &op, &lp);
    char_index = cursor->ac_index;

    inc = direction==TC_RIGHT ? num : -num;
    newindex = char_index + inc;
    if (newindex < 0) { /* prev line or obj */
        if (op->prev) {
        }
        else {  /* hit top of doc */
            cursor->ac_index = 0;
        }
    }
    else if (newindex > op->wsize) { /* next line or obj */
        if (op->next) {
        }
        else {  /* hit bottom of doc */
            cursor->ac_index = op->wsize;
        }
    }
    else {  /* cursor stays in current object */
        cursor->ac_index = newindex;
    }
}

moveCursorVert(w, direction, num)
AXWidget_editbox w;
int direction;
int num; /* #lines to move cursor */
{
    any_obj_t *op;
    tline_t *lp;
    tcursor_t *cursor;
    int x, y, ix, lbear, rbear;
    int char_index;

    /* get object containing cursor */
    getCursorInfo(w, &cursor, &op, &lp);

    /* set current cursor xpos (in window coordinates) */
    x = getTextWid(w, op->wdata+lp->first_index,
				   cursor->ac_index-lp->first_index, &lbear, &rbear);
    x += w->editbox->xoffset + op->xpos + lp->xpos;

    /* set new line based on direction */
	for (ix = 0; ix < num; ix++) {
		if (direction == TC_UP) {
			if (lp->prev)
				lp = lp->prev;
			else
				break;
		} else {
			if (lp->next)
				lp = lp->next;
			else
				break;
		}
	}

	if (ix == 0) /* couldn't find any prev/next line */
		return;

    /* set ypos of new line (in window coordinates) */
    y = w->editbox->yoffset + op->ypos + lp->ypos;  

    /* update cursor info based on new position */
    getNearestIndex(w, op, lp, x, y, &char_index);
    cursor->obj = op;
    cursor->ac_index = char_index;
}

/*
 * getNearestObj - find nearest object and line 
 *                 (x,y are in window coordinates)
 */
getNearestObj(w, x, y, objP, lineP)
AXWidget_editbox w;
int x, y;
any_obj_t **objP;  /* RETURN */
tline_t **lineP;   /* RETURN */
{
    any_obj_t *op;
    tline_t *lp = NULL;
    int next_ypos;

    for (op=w->editbox->objs; op; op=op->next) {
        if (y>w->editbox->yoffset+op->ypos+op->height && op->next)
            continue;
        if (op->type == TEXTOBJ) {
            /* ButtonPress lies within this obj */
            for (lp=op->lines; lp; lp=lp->next) {
                if (lp->next)
                    next_ypos = w->editbox->yoffset+op->ypos+lp->next->ypos;
                else
                    break;  /* last line in current text obj */
                if (y < next_ypos) 
                    break;
            }
            break;
        }
    }
    *objP = op;
    *lineP = lp;
}

/*
 * getNearestIndex - find nearest inter-char spacing on line
 *                   (x,y are in window coordinates)
 */
getNearestIndex(w, objP, lineP, x, y, char_index)
AXWidget_editbox w;
any_obj_t *objP;
tline_t *lineP;  /* line on which nearest index is to be calculated */
int x, y;
int *char_index;  /* RETURN */
{
    wchar *cp, *p;
    int wid, len, prevwid, lbear, rbear;

    if (!objP || !lineP) {
        *char_index = 0;
        return;
    }

    /* make x,y relative to obj */
    x -= (w->editbox->xoffset + objP->xpos);
    y -= (w->editbox->yoffset + objP->ypos);

    if (y > lineP->ypos+w->editbox->lineHeight) {
        *char_index = lineP->first_index + lineP->len;
        return;
    }
    else if (y < lineP->ypos) {
        *char_index = lineP->first_index;
        return;
    }

    cp = objP->wdata + lineP->first_index;
    prevwid = wid = 0;
    len = lineP->len;
    for (p=cp; len; p++, len--) {
        wid = wid + getTextWid( w, p, 1, &lbear, &rbear );
        if (wid >= x)
            break;
        prevwid = wid;
    }
    if ( len && (wid-x)<(x-prevwid) )
        p++;

    *char_index = p - objP->wdata;
}

/*
 * getNearestVisibleIndex - performs work of getNearestObj and
 *                          getNearestIndex, but also ensures that
 *                          the index is inside the window.
 *                          (x,y are in window coordinates)
 */
getNearestVisibleIndex(w, x, y, objP, lineP, char_index)
AXWidget_editbox w;
int x, y;
any_obj_t **objP;  /* RETURN */
tline_t **lineP;   /* RETURN */
int *char_index;        /* RETURN */
{
    any_obj_t *op;
    tline_t *lp = NULL;
    int ypos, next_ypos;

    for (op=w->editbox->objs; op; op=op->next) {
        if (y>w->editbox->yoffset+op->ypos+op->height && op->next)
            continue;
        if (op->type == TEXTOBJ) {
            /* ButtonPress lies within this obj */
            for (lp=op->lines; lp; lp=lp->next) {
                ypos = w->editbox->yoffset + op->ypos + lp->ypos;
                if ( ypos < 0 )
                    continue;  /* line is not visible */
                if ( y < 0 ) {
                    /* first visible line */
                    *objP = op;
                    *lineP = lp;
                    *char_index = lp->first_index;
                    return;
                }
                if (lp->next)
                    next_ypos = w->editbox->yoffset+op->ypos+lp->next->ypos;
                else
                    break;  /* last line in current text obj */
                if (y < next_ypos) 
                    break;
                if ((next_ypos)>w->editbox->height) {
                    /* last visible line */
                    *objP = op;
                    *lineP = lp;
                    *char_index = lp->first_index + lp->len;
                    return;
                }
            }
            break;
        }
    }
    /* x,y is visible in window - get nearest index */
    getNearestIndex(w, op, lp, x, y,char_index);
    *objP = op;
    *lineP = lp;
}


changeCursorType(w, type)
AXWidget_editbox w;
int type;
{
    tcursor_t *cursor;
    any_obj_t *op;
    tline_t *lp;

    getCursorInfo(w, &cursor, &op, &lp);
    drawCursor(w, TC_ERASE);
    cursor->type = type;
    drawCursor(w, TC_DRAW);
}

drawCursor(w, mode)
AXWidget_editbox w;
int mode;  /* TC_DRAW, TC_ERASE */
{
    any_obj_t *op;
    tline_t *lp;
    seln_t *sel;
    tcursor_t *cursor;
    int cur_off;      /* # chars between cursor & beginning of line */
    int xoff, yoff;   /* # pix between cursor & origin of line */
    int x, y, lbear, rbear;
    GC gc, repair_gc;
	PIXEL_VALUE saved_fg;
	int restore_fg = FALSE;

    if ( !w->window )
        return;

    /* don't draw cursor if editbox is non-pickable or inactive 
     * (w->hasFocus is used to indicate if editbox is active widget)
	 */
    if (w->no_picks || !w->hasFocus)
        return;

    /* don't draw cursor if something is hilit */
    sel = &w->editbox->sel;
    if (w->pending_delete && sel->startIndex!=sel->endIndex)
        return; 

    /* get obj with cursor */
    /* get line within obj with cursor */
    getCursorInfo(w, &cursor, &op, &lp);

    /* don't draw cursor if line not visible */
    y = w->editbox->yoffset + op->ypos + lp->ypos;
    if ( (y + w->editbox->lineHeight < 0) || (y > w->editbox->height) )
        return;

    cur_off = cursor->ac_index - lp->first_index;

    /* draw cursor */
	xoff = getTextWid(w, op->wdata+lp->first_index, cur_off, &lbear, &rbear);
	x = w->editbox->xoffset + op->xpos + lp->xpos + xoff;
	y = w->editbox->yoffset + op->ypos + lp->ypos;
	
	if ( mode == TC_DRAW ) {
		gc = ForeGC;
		if ( w->useWorkColors )
			SetWorkColors(gc, 1);
	}
	else {
		gc = BackGC;
		if ( w->useWorkColors )
			SetWorkColors(gc, 0);
	}
	
	if (w->widgetColor) {
		/* need to use a different color to draw the cursor */
		restore_fg = TRUE;
		saved_fg = AXGetForegroundPixel(Dpy, gc);
		if (TC_ERASE == mode) {
			/* use the widget's background pixel */
			XSetForeground(Dpy, gc, w->widgetColor->any.pixel);
		}
		else {
			/* mode is TC_DRAW */
			if (w->textColor) {
				/* use the text color for the cursor */
				XSetForeground(Dpy, gc, w->textColor->any.pixel);
			}
			else {
				/* use the wkit foreground pixel for the cursor */
				XSetForeground(Dpy, gc, AXGetForegroundPixel(Dpy, ForeGC));
			}
		}
	}
	XDrawLine(Dpy,w->window,gc,x,y,x,y+w->editbox->lineHeight);
	XDrawLine(Dpy,w->window,gc,x-1,y,x+1,y);
	XDrawLine(Dpy,w->window,gc,x-1,y+w->editbox->lineHeight,x+1,
			  y+w->editbox->lineHeight);
	if (restore_fg) {
		XSetForeground(Dpy, gc, saved_fg);    /* restore */
	}
	
	if (mode == TC_ERASE) {
		/* redisplay chars on each side of cleared cursor */
		wchar regencp[2];
		int gcType, workColorType;
		
		x = w->editbox->xoffset + op->xpos + lp->xpos;
		y = w->editbox->yoffset + op->ypos + lp->ypos;
		if (cur_off != 0) {
			/* cursor not at beginning of line; repair char to left
			 */
			Wstrncpy( regencp, op->wdata+cursor->ac_index-1, 1 );
			regencp[1] = '\0';
			xoff = getTextWid(w, op->wdata+lp->first_index,
							  cur_off-1, &lbear, &rbear);
			if (w->editbox->sel.endIndex==cursor->ac_index &&
				w->editbox->sel.endIndex>w->editbox->sel.startIndex) {
				gcType = GC_HILIT;
				workColorType = 0;
			}
			else {
				gcType = GC_FORE;
				workColorType = 1;
			}
			repair_gc = getTextGC(w, gcType);
			if (w->useWorkColors) {
				SetWorkColors(repair_gc, workColorType);
			}
			xmDrawString16((AXWidget)w, w->window, repair_gc,
				       x + xoff, y + w->editbox->lineAscent,
				       regencp, 1, w->textShadow,
				       ebGetTextColor((AXWidget)w, gcType),
				       w->widgetColor);
			if (w->useWorkColors) {
				UnSetWorkColors(repair_gc, workColorType);
			}
		}
		if (cur_off != lp->len) {
			/* cursor not after last char on line; repair char to right
			 */
			Wstrncpy( regencp, op->wdata+cursor->ac_index, 1 );
			regencp[1] = '\0';
			xoff = getTextWid(w, op->wdata+lp->first_index,
							  cur_off, &lbear, &rbear);
			if (w->editbox->sel.endIndex==cursor->ac_index &&
				w->editbox->sel.endIndex<w->editbox->sel.startIndex) {
				gcType = GC_HILIT;
				workColorType = 0;
			}
			else {
				gcType = GC_FORE;
				workColorType = 1;
			}
			repair_gc = getTextGC(w, gcType);
			if (w->useWorkColors) {
				SetWorkColors(repair_gc, workColorType);
			}
			xmDrawString16((AXWidget)w, w->window, repair_gc,
				       x + xoff, y + w->editbox->lineAscent,
				       regencp, 1, w->textShadow,
				       ebGetTextColor((AXWidget)w, gcType),
				       w->widgetColor);
			if (w->useWorkColors) {
				UnSetWorkColors(repair_gc, workColorType);
			}
		}
	}
	
	if ( w->useWorkColors ) {
		if ( mode == TC_DRAW )
			UnSetWorkColors(gc, 1);
		else
			UnSetWorkColors(gc, 0);
	}
}


/* 
 * displayObjs - call each object's displayer 
 *               (expy & expht are in window coordinates)
 */
displayObjs(w, expy, expht)
AXWidget_editbox w;
int expy, expht;
{
    any_obj_t *op;
    int y;
    void (*displayer)();

    for ( op=w->editbox->objs; op; op=op->next ) {
        /* call obj displayer if any part of obj 
         * is visible in viewport */
        y = w->editbox->yoffset + op->ypos;
        if (y<=(expy+expht) && (y+op->height)>=expy) {
            displayer = EB_ObjDisplayer[op->type];
            if (displayer)
                (*displayer) (w, op, &expy, &expht);
        }
    }
}

/*
 * displayTextObj - displayer for text objects
 * ( index 1 in EB_ObjDisplayer[] )
 */
void
displayTextObj(w, objP, expyP, exphtP)
AXWidget_editbox w;
any_obj_t *objP;
int *expyP, *exphtP;  /* MODIFY and RETURN ypos & ht of exposed area */
{
    int size, x, y;
    wchar *str;
    tline_t *lp;
    AXWidget_main main_w;

    /* draw each line in text object */
    for (lp=objP->lines; lp; lp=lp->next) {
        x = w->editbox->xoffset + objP->xpos + lp->xpos;
        y = w->editbox->yoffset + objP->ypos + lp->ypos;
        /* draw line only if it intersects exposed area 
         * and is visible in viewport
		 */
        if ( (y<=(*expyP+*exphtP) && (y+w->editbox->lineHeight)>=*expyP) &&
                (y+w->editbox->lineHeight>=0 && (y <= w->editbox->height))) {
            size = lp->len;
            str = (wchar *)malloc( (size + 1)*sizeof(wchar) );
            Wstrncpy( str, objP->wdata+lp->first_index, size );
            str[size] = '\0';
            localDrawString(w, GC_FORE, x, y, str, size, SPACE, SPACE,
							GC_FORE, GC_FORE);
            free(str);
            *expyP += w->editbox->lineHeight;
            *exphtP -= w->editbox->lineHeight;
        }
    }

    if (w->pending_delete && objP==w->editbox->sel.startObj) {
        paintText(w, objP, w->editbox->sel.startIndex,
                       w->editbox->sel.endIndex, GC_HILIT);
    }

#if BUILD_ORIENT
    main_w = TheMainWidget((AXWidget)w);
    if ( objP == w->editbox->conv_info.obj )
       {
       /* this is the object in which input conversion is taking place */
       if ( main_w->conversion_start >= 0 )
          {
          /* some conversion active */
          paintText(w, objP, main_w->conversion_start, 
                              (main_w->conversion_end+1), GC_UNDERLINE );
          }
       if ( main_w->conv_phrase_start >= 0 )
          {
          /* sub-pharsese get highlighted */
          paintText(w, objP, main_w->conv_phrase_start,
                              (main_w->conv_phrase_end+1), GC_HILIT);
          }
       }
#endif
}

formatObjs(w, reason)
AXWidget_editbox w;
int reason;
{
    int old_scroll_max;
    tcursor_t *cursor;
    any_obj_t *op;
    tline_t *lp;
    int ypos, tmpyoff;

    old_scroll_max = w->scroll_max;
    formatTextObj(w, w->editbox->objs, reason);

    /* make sure cursor is visible */
    getCursorInfo(w, &cursor, &op, &lp);
    ypos = w->editbox->yoffset + op->ypos + lp->ypos;
    if (old_scroll_max == 0) {
        /* entire contents being formatted (called from PaintEditBox) */
        dispCursorAtCenter(w);
    }
    else {
        /* (not being called from PaintEditBox) */
        if ( (w->scroll_max <= w->scroll_per) && (w->scroll_pos != 0) ) {
            w->scroll_pos = 0;
            PaintEditBox(w, NULL, TRUE, TRUE);
        }
        else if ( (ypos < 0) || ((ypos+w->editbox->lineHeight) >
                    w->editbox->height) ) {
            /* make cursor visible */
            w->scroll_pos += (w->scroll_max - old_scroll_max);
            /* force cursor to center of screen if still not visible */
            tmpyoff = -w->scroll_pos;
            ypos = tmpyoff + op->ypos + lp->ypos;
/*            if ( (ypos < 0) || ((ypos+w->editbox->lineHeight) >
                                w->editbox->height) )*/
                dispCursorAtCenter(w);
            PaintEditBox(w, NULL, TRUE, TRUE);
        }
        else {
            updateScroll(w, NULL);
        }
    }
}

formatTextObj(w, objP, reason)
AXWidget_editbox w;
any_obj_t *objP;
int reason;  /* FMT_ALL, FMT_LINE, FMT_YPOS */
{
    any_obj_t *op;
    tline_t *lp, *prevlp;
    tformat_t *fp;         /* formatting info */
    wchar *cp, *nextcp, *origcp;
    wchar *buf;             /* data for current line (new format) */
    wchar *oldbuf;          /* data for current line (old format) */
    wchar *diff, *olddiff;  /* ptrs from where buf and oldbuf differ */
	wchar leftChar, rightChar;
	int leftgcType;
    int size, fmtWid;
    int len, curr_ypos, ht;
    int newindex, linenum;
    int numFmtLines;        /* minimum number of lines to reformat */
    int x, y, lbear, rbear;
    int oldx, oldy;         /* x,y of current line wrt win (old format) */
    int difflen;            /* # chars different from old format */
    int difference;

    if (!objP)
        return;

    op = objP;
    fp = op->formatInfo;
    lp = prevlp = fp->formatLine;
    numFmtLines = fp->numFormatLines;
    cp = op->wdata + lp->first_index;
    origcp = op->wdata;
    size = (origcp + op->wsize) - cp;
    fmtWid = w->editbox->width;
    curr_ypos = lp->ypos;
    ht = 0;
    linenum = 1;

    /* malloc memory just once to hold max possible formatting data */
    buf = (wchar *)malloc( (size + 1)*sizeof(wchar) );

    while ( size >= 0 ) {
        newindex = cp - origcp;
        curr_ypos += ht;  /* incr curr line's ypos by prev line's ht */

        /* check for convergence */
        if (reason!=FMT_ALL && linenum>numFmtLines) {  
            if (reason == FMT_LINE) {
                if (lp && lp->first_index==newindex)
                    reason = FMT_YPOS;
            }
            if (reason == FMT_YPOS) {
                if (!lp || lp->ypos==curr_ypos)
                    break;
            }
            /* clear old line */
            if (lp) {
                Wstrncpy( buf, op->wdata+lp->first_index, lp->len );
                buf[lp->len] = '\0';
                x = w->editbox->xoffset + op->xpos + lp->xpos;
                y = w->editbox->yoffset + op->ypos + lp->ypos;
                localDrawString(w, GC_BACK, x, y, buf, lp->len,
								SPACE, SPACE, GC_FORE, GC_FORE);
            }
        }


        /* do necessary formatting */
        if (reason != FMT_YPOS) {
            formatLine(w, cp, size, fmtWid, &len, &ht, &nextcp);
            if (!lp) {
                createLine(&lp);
                lp->prev = prevlp;
                lp->next = prevlp->next;
                prevlp->next = lp;
                if (lp->next)
                    lp->next->prev = lp;
            }
            lp->first_index = newindex;
            lp->len = len;
        }
        else {
            ht = w->editbox->lineHeight;
            nextcp = origcp + lp->next->first_index;
        }
        lp->ypos = curr_ypos;

        /* display new line */
        if (lp) {
            Wstrncpy( buf, op->wdata+lp->first_index, lp->len);
            buf[lp->len] = '\0';
            x = w->editbox->xoffset + op->xpos + lp->xpos;
            y = w->editbox->yoffset + op->ypos + lp->ypos;
            if (linenum <= numFmtLines) {
                /* before drawing new line, check if new format 
                 * of line is different from old */
                getFormatInfo(w, op, linenum, &oldbuf, &oldx, &oldy);
                if ( getStrDiff(buf, oldbuf, &diff, &olddiff) ||
                               (oldx != x) || (oldy != y) ) {

                    /* line has changed - clear changed part of old line */
                    oldx = oldx + getTextWid(w, oldbuf, olddiff-oldbuf,
											 &lbear, &rbear);
					if (olddiff > oldbuf) {
						/* char just to left of range to be displayed */
						leftChar = *(olddiff - 1);
					}
					else {
						/* range includes leftmost char on line */
						leftChar = SPACE;
					}
                    difflen = Wstrlen(olddiff);
					localDrawString(w, GC_BACK, oldx, oldy,
									olddiff, difflen, leftChar,
									SPACE, GC_FORE, GC_FORE);

                    /* draw new line */
                    if (y + w->editbox->lineHeight >= 0 &&
						y <= w->editbox->height) {
						
                        /* draw changed part of new line */
                        x = x + getTextWid(w, buf, diff-buf, &lbear, &rbear);
						if (diff > buf) {
							/* char just to left of range to be displayed */
							leftChar = *(diff - 1);
						}
						else {
							/* range includes leftmost char on line */
							leftChar = SPACE;
						}
                        difflen = Wstrlen(diff);
						localDrawString(w, GC_FORE, x, y,
										diff, difflen, leftChar,
										SPACE, GC_FORE, GC_FORE);
                    }
                }
                free(oldbuf);
            }
            else {
                /* draw only if completely visible */
				if (y + w->editbox->lineHeight >= 0 &&
					y <= w->editbox->height) {
					
					localDrawString(w, GC_FORE, x, y, buf,
									lp->len, SPACE, SPACE,
									GC_FORE, GC_FORE);
                }
            }
            op->height = lp->ypos + w->editbox->lineHeight;
            w->scroll_max = op->ypos + lp->ypos + w->editbox->lineHeight;
        }

        linenum++;
        prevlp = lp;
        lp = lp->next;
        difference = (nextcp - cp);
        size = size - difference;
        cp = nextcp;
    }

    free(buf);

    /* clear and free remaining lines (only if did not converge midway) */
    if (size < 0) {
        if (lp && lp->prev)
            lp->prev->next = NULL;
        while (lp) {
            tline_t *nextlp;
            buf = (wchar *) malloc( (lp->len + 1)*sizeof(wchar) );
            Wstrncpy( buf, op->wdata+lp->first_index, lp->len );
            buf[lp->len] = '\0';
            x = w->editbox->xoffset + op->xpos + lp->xpos;
            y = w->editbox->yoffset + op->ypos + lp->ypos;

            if (y + w->editbox->lineHeight >= 0 &&
				(y <= w->editbox->height)) {
				
                if (!lp->next) {
                    clearLine(w, x, y);
                } else {
                    localDrawString(w, GC_BACK, x, y,
									buf, lp->len, SPACE,
									SPACE, GC_FORE, GC_FORE);
                }
            }
            nextlp = lp->next;
            free(buf);
            free(lp);
            lp = nextlp;
        }
    }
    else {  /* converged! */
        while (lp) {
            op->height = lp->ypos + w->editbox->lineHeight;
            w->scroll_max = op->ypos + lp->ypos + w->editbox->lineHeight;
            lp = lp->next;
        }
    }

    resetFormatInfo(w, op);
}

formatLine(w, data, dataLen, fmtWid, len, ht, next)
AXWidget_editbox w;
wchar *data;    /* data to be formatted */
int dataLen;   /* size of data to be formatted */
int fmtWid;    /* max width of line */
int *len, *ht; /* RETURN  computed length and height of formatted line */
wchar **next;   /* RETURN  ptr to start of next line (not always data+*len) */
{
    int wid, ptrForwarded, lbear, rbear;
	int line_full;
    wchar *p, *p1;

    wid = 0;
    ptrForwarded = FALSE;
    /* gather enough data for one line */
    for ( p = data; p < (data+dataLen); p++ ) {
        wid = wid + getTextWid( w, p, 1, &lbear, &rbear );
		if (w->maxLen > 0) {
			line_full = (wid > fmtWid) ||
				(((p - data) >= w->maxLen) && (*p != (wchar) '\n'));
		}
		else {
			line_full = wid > fmtWid;
		}
        if (line_full) {
            /* got enough chars to fit on line */
            ptrForwarded = TRUE;
            if ( p == data ) {
                p++; /* HACK: avoid infinite loop if 1st char width > fmtWid */
                break;
            }
            /* break at last space on line (if any) */
            p1 = p;
            while ( --p1 >= data ) {
                if (*p1==' ') {
                    p = p1;  /* don't include space on line */
                    ptrForwarded = FALSE; /* don't draw space on next line */
                    break;
                }
#if BUILD_KANJI
                /* Ashisuto added to let the long line wraps at any wide */
                /* chars in Applixware EditBox widget. 95.09.14          */
                if (IS_WIDE_CHAR(*p1) || IS_NARROW_KANA(*p1)) {
                    p = p1+1;
                    break;
                }
#endif
            }
            break;
        }
        if (*p == '\n') 
            break;
    }
    *len = p - data;
    *ht = w->editbox->lineHeight;
    *next = data + *len;  /* point to data for next line */
    if (!ptrForwarded)
        (*next)++;  /* advance ptr past last char on curr line */
}



/*
 * getFormatInfo - get old state (data, xpos, ypos) of edited line 
 *                 (x,y are in window coordinates)
 */
getFormatInfo(w, objP, linenum, bufP, oldXP, oldYP)
AXWidget_editbox w;
any_obj_t *objP;
int linenum;
wchar **bufP;         /* RETURN */
int *oldXP, *oldYP;  /* RETURN */
{
    tformat_t *fp;
    int start, len;
    wchar *buf;

    fp = objP->formatInfo;
    start = fp->oldIndex[linenum-1];
    len = fp->oldLen[linenum-1];
    buf = (wchar *)malloc( (len+1)*sizeof(wchar) );
    Wstrncpy( buf, fp->oldData+start, len );
    *(buf+len) = '\0';

    *bufP = buf;
    *oldXP = fp->oldXpos[linenum-1];
    *oldYP = fp->oldYpos[linenum-1];
}

/*
 * resetFormatInfo - reset format info struct
 */
resetFormatInfo(w, objP)
AXWidget_editbox w;
any_obj_t *objP;
{
    tformat_t *fp;

    fp = objP->formatInfo;

    free(fp->oldData);
    free(fp->oldIndex);
    free(fp->oldLen);
    free(fp->oldXpos);
    free(fp->oldYpos);

    fp->formatLine = objP->lines;
    fp->numFormatLines = 0;
    fp->oldData = NULL;
    fp->oldIndex = NULL;
    fp->oldLen = NULL;
    fp->oldXpos = NULL;
    fp->oldYpos = NULL;
}

/*
 * dispCursorAtCenter - adjust scroll_pos so that line with cursor
 *                      is approximately at center of window
 */
dispCursorAtCenter(w)
AXWidget_editbox w;
{
    tcursor_t *cursor;
    any_obj_t *op;
    tline_t *lp, *scrlp;
    int tmpyoffset, ypos;

    getCursorInfo(w, &cursor, &op, &lp);
    tmpyoffset = w->editbox->height/2 - op->ypos - lp->ypos;
    for (; lp; lp=lp->prev) {
        scrlp = lp;
        ypos = tmpyoffset + op->ypos + lp->ypos;
        if (ypos < 0) {
            if (lp->next)
                scrlp = lp->next;
            break;
        }
    }
    w->scroll_pos = op->ypos + scrlp->ypos;
}

/*
 * rewindCursor - move cursor pos back by 'len' chars
 */
rewindCursor(w, len)
AXWidget_editbox w;
int len;
{
    tcursor_t *cursor;

    /* !! assumes only one obj exists (TEXT) !! */
    cursor = &w->editbox->tcursor;
    cursor->ac_index -= len;
    if (cursor->ac_index < 0)
        cursor->ac_index = 0;
}

/*
 * resetScrollInfo - reset scroll info to default
 */
resetScrollInfo(w)
AXWidget_editbox w;
{
    w->scroll_per = w->editbox->height;
    w->scroll_max = 0;
    w->scroll_pos = 0;
    w->editbox->yoffset = 0;
}

/*
 *
 */
updateScroll(w, exP)
AXWidget_editbox w;
XExposeEvent *exP;
{
    xmSetWidgetDisplay((AXWidget)(w->vScrollBar), TRUE);
    w->vScrollBar->scroll_per = w->scroll_per;
    w->vScrollBar->scroll_pos = w->scroll_pos;
    w->vScrollBar->scroll_max = MAX(w->scroll_max, w->scroll_per);
    PaintScrollBar(w->vScrollBar,exP,TRUE,TRUE);
}

/*
 * manageScroll - update display to conform to new scroll pos.
 *                When this function is called, w->scroll_pos
 *                is the new scroll pos, and -w->editbox->yoffset
 *                is the old scroll pos.
 */
manageScroll(w, y, ht)
AXWidget_editbox w;
int *y, *ht;  /* RETURN */
/* ypos (in win coord) & ht of area not handled by this routine */
{
    any_obj_t *op;
    tline_t *lp, *prevlp;
    int src_x, src_y, dest_x, dest_y;
    int fillytop, fillybot;
    int ly;      /* ypos of line (in window coordinates) */
    GC foregc, backgc;
    unsigned long valuemask;
    XGCValues values;
    int monospacegcChanged;

    if ( !w->window )
        return;

    if ( w->scroll_pos == -w->editbox->yoffset )
        return;

    if ( w->monospace_font ) {
        monospacegcChanged = FALSE;
        valuemask = GCGraphicsExposures;
        XGetGCValues(Dpy, MonospaceGC, valuemask, &values);
        if ( values.graphics_exposures == FALSE ) {
            /* temporarily enable graphics exposures in monospace fore gc */
            monospacegcChanged = TRUE;
            values.graphics_exposures = TRUE;
            axChangeGC(Dpy, MonospaceGC, (XFontStruct *) 0,
					   valuemask, &values);
        }
        foregc = MonospaceGC;
        backgc = MonospaceBackGC;
    }
    else {
        foregc = GXForeGC;
        backgc = BackGC;
    }
    if ( w->useWorkColors ) {
        SetWorkColors(foregc, 1);
        SetWorkColors(backgc, 0);
    }

    /* copy common area */
    src_x = dest_x = -w->editbox->xoffset;
    src_y = 0;
    dest_y = -(w->editbox->yoffset + w->scroll_pos);
    if ( (dest_y <= w->editbox->height) &&
                    (dest_y + w->editbox->height >= 0) ) {
        /* copy only if part of currently visible text
         * is still visible after scroll */
        XCopyArea(Dpy, w->window, w->window, foregc, src_x, src_y,
                                    w->editbox->width, w->editbox->height,
                                    dest_x, dest_y);
    }

    /* clear area exposed by scroll */
    if ( dest_y <= 0 ) {
        fillytop = MAX(0, dest_y + w->editbox->height);
        fillybot = w->editbox->height;
    }
    else {
        fillytop = MIN(0, dest_y);
        fillybot = MIN(w->editbox->height, dest_y);
    }
    xmColorFillRectangle((AXWidget)w, w->window, backgc, 0, fillytop,
			 w->editbox->width, fillybot - fillytop);

    /* return ypos of first line to be displayed by displayer */
    if (w->scroll_pos <= -w->editbox->yoffset) {
        /* top of window to be redisplayed */
        w->editbox->yoffset = -w->scroll_pos;
        /* Last line in window after XCopyArea may only be 
         * partially visible.  If so, clear because only
         * completely visible lines are drawn. */
        op = w->editbox->objs;
        for (lp=op->lines; lp; lp=lp->next) {
            ly = w->editbox->yoffset + op->ypos + lp->ypos;
            if ( ly < w->editbox->height ) {
                if ( (ly+w->editbox->lineHeight) > w->editbox->height ) {
                    /* line not fully displayed at bottom of window */
/*					xmColorFillRectangle(w, w->window, backgc, 0, ly,
										 w->editbox->width,
										 w->editbox->lineHeight);*/
                    break;
                }
            }
            else {
                break;
            }
        }
    }
    else {
        /* bottom of window to be redisplayed */
        /* !! assuming only one text object !! */
        op = w->editbox->objs;
        if (op->type == TEXTOBJ) {
            for (lp=op->lines; lp; lp=lp->next) {
                prevlp = lp;
                ly = w->editbox->yoffset+op->ypos+lp->ypos;
                if ((ly+w->editbox->lineHeight) > w->editbox->height)
                    /* first line which is NOT displayed */
                    break;
            }
            if (!lp)
                lp = prevlp;
        }
        w->editbox->yoffset = -w->scroll_pos;
    }
    *y = fillytop;
    *ht = fillybot - fillytop;

    if ( w->useWorkColors ) {
        UnSetWorkColors(foregc, 1);
        UnSetWorkColors(backgc, 0);
    }
    if ( w->monospace_font ) {
        if ( monospacegcChanged ) {
            /* revert monospace fore gc to original state */
            valuemask = GCGraphicsExposures;
            values.graphics_exposures = FALSE;
            axChangeGC(Dpy, MonospaceGC, (XFontStruct *) 0,
					   valuemask, &values);
        }
    }
}

/*
 * getStrDiff - compare two strings.
 *              Return TRUE if strings differ, FALSE if they are same.
 *              Also return pointers to substrings from where the two 
 *              strings differ, or pointers to the end of each string
 *              if strings are the same.
 */
getStrDiff(s1, s2, diff1P, diff2P)
wchar *s1, *s2;
wchar **diff1P, **diff2P;  /* RETURN */
{
    int retval = TRUE;

    for ( ; *s1 == *s2; s1++,s2++ ) {
        if ( *s1 == '\0' ) {
            /* strings don't differ */
            retval = FALSE;  
            break;
        }
    }
    *diff1P = s1;
    *diff2P = s2;
    return(retval);
}


/*
 * getTextWid - return width of 'len' chars in 'p'
 */
static int
getTextWid(in_widget, wstrP, len, lbearing, rbearing)
AXWidget_editbox in_widget;
wchar *wstrP;
int len;
int *lbearing;   /* out: report leftmost char's left sidebearing */
int *rbearing;   /* out: report rightmost char's 'excess' right sidebearing */
{
    int wid, tmp;
    int i;
    int sub_len;
    wchar *wsubP;
	XFontStruct *font;
	XCharStruct *ch;
	
	if (!len) {
		*lbearing = *rbearing = 0;
		return (0);
	}

    if ( (in_widget != (AXWidget_editbox)NULL) && in_widget->monospace_font ) {
		ch = getCharStruct((int) wstrP[0], MonospaceFont);
		if ( WHITESPACE((int) wstrP[0]) ) {
			*lbearing = 0;
		}
		else {
			*lbearing = (int) ch->lbearing;
		}
		
		ch = getCharStruct((int) wstrP[len - 1], MonospaceFont);
		if ( WHITESPACE((int) wstrP[len - 1]) ) {
			*rbearing = 0;
		}
		else if (ch->rbearing > ch->width) {
			*rbearing = (int) (ch->rbearing - ch->width);
		}
		else {
			*rbearing = 0;
		}
		
		if (wShadowIsOn(in_widget->textShadow)) {
			    /* report shadow extent by increasing right sidebearing */
			int xoff, yoff;
			xmTextShadowOffsets(in_widget->font, &xoff, &yoff);
			*rbearing += xoff + 1;
		}
        return(AXWTextWidth(MonospaceFont, wstrP, len));
    }
    else {
		if (in_widget->font) {
			font = in_widget->font;
		}
		else {
			font = AxDlgMainFont;
		}
		
		ch = getCharStruct((int) wstrP[0], font);
		if ( WHITESPACE((int) wstrP[0]) ) {
			*lbearing = 0;
		}
		else {
			*lbearing = (int) ch->lbearing;
		}
		
		ch = getCharStruct((int) wstrP[len - 1], font);
		if ( WHITESPACE((int) wstrP[len - 1]) ) {
			*rbearing = 0;
		}
		else if (ch->rbearing > ch->width) {
			*rbearing = (int) (ch->rbearing - ch->width);
		}
		else {
			*rbearing = 0;
		}
		
		if (wShadowIsOn(in_widget->textShadow)) {
			    /* report shadow extent by increasing right sidebearing */
			int xoff, yoff;
			xmTextShadowOffsets(in_widget->font, &xoff, &yoff);
			*rbearing += xoff + 1;
		}
		wid = AXWTextWidth(font, wstrP, len);
		return(wid);
	}

}

/*
 * paintText - draw/clear/hilight text (specified by 'start'
 *             and 'end' indices)
 * Note: the drawing is exclusive of the end position, not inclusive, so
 *       paittext(w, objP, 1, 2 ) only draws 1 character
 */
paintText(w, objP, start, end, inGCtype)
AXWidget_editbox w;
any_obj_t *objP;
int start, end;  /* indices indicating text to be drawn */
int inGCtype;    /* GC_FORE, GC_HILIT, GC_UNDERLINE  */
{
    tline_t *lp;
    int newstart, newend;
	int selstart, selend;
    int x, y, lbear, rbear, ytmp;
    int tmp, len;
    wchar *str_start, leftChar, rightChar;
    int gcType, leftgcType, rightgcType;
    int under_len, ypos;

    if (start == end)
        return;
    if (start > end) {
        tmp = start;
        start = end;
        end = tmp;
    }

    /* draw each line */
    for (lp=objP->lines; lp; lp=lp->next) {
        if (lp->first_index+lp->len < start) {
            continue;
        }
        if (lp->first_index > end) {
            break;
        }
        y = w->editbox->yoffset + objP->ypos + lp->ypos;
		ytmp = y + w->editbox->lineHeight;
        if (ytmp < 0) {
            /* line completely above window top */
            continue;
        }
        if (y > w->editbox->height) {
            /* line completely below window bottom */
            break; 
        }
        newstart = MAX(start, lp->first_index);
        newend = MIN(lp->first_index+lp->len, end);
        x = w->editbox->xoffset + objP->xpos + lp->xpos;
        if (newstart > lp->first_index) {
            x += getTextWid(w, objP->wdata+lp->first_index,
							newstart-lp->first_index, &lbear, &rbear);
		}
        len = newend - newstart;
        str_start = objP->wdata + newstart;

        gcType = inGCtype;
        if ( inGCtype == GC_UNDERLINE ) {
            /* a conveninent "pseudo-type" */
            gcType = GC_FORE;
		}

		/* pass right and left neighboring chars, if any */
		    /* determine whether these neighbors are selected;
			   as in hilightForMotion, we assume that the selection
			   endpoints fall within the same any_obj_t instance */
		selstart = w->editbox->sel.startIndex;
		selend = w->editbox->sel.endIndex;
		leftgcType = rightgcType = GC_FORE;
		if (selstart != selend) {
			int leftix = newstart - 1;
			if (selstart > selend) {
				tmp = selstart;
				selstart = selend;
				selend = tmp;
			}
			if ((leftix >= selstart) && (leftix < selend)) {
				leftgcType = GC_HILIT;
			}
			if ((newend >= selstart) && (newend < selend)) {
				rightgcType = GC_HILIT;
			}
		}
		if (newstart > lp->first_index) {
			leftChar = *(objP->wdata + newstart - 1);
		}
		else {
			leftChar = SPACE;
		}
		if (newend < (lp->first_index + lp->len)) {
			rightChar = *(objP->wdata + newend);
		}
		else {
			rightChar = SPACE;
		}
        localDrawString(w, gcType, x, y, str_start, len,
						leftChar, rightChar, leftgcType, rightgcType);

        if ( inGCtype == GC_UNDERLINE ) {
           /* underline the string we just drew */
           under_len = getTextWid(w, str_start, len, &lbear, &rbear );
           ypos = y + w->editbox->lineHeight - 1;
           XDrawLine( Dpy, w->window, ForeGC, x, ypos, (x+under_len), ypos );
		}
    }
}

static void
clearLine(w, x, y)
AXWidget_editbox w;
int x, y;
{
    GC fillgc;

    if (!w)
        return;
    if (!w->window)
        return;

    if (w->monospace_font)
        fillgc = MonospaceBackGC;
    else
        fillgc = BackGC;

    if (w->useWorkColors)
        SetWorkColors(fillgc, 0);

    xmColorFillRectangle((AXWidget)w, w->window, fillgc, x, y,
			 w->editbox->width - x, w->editbox->lineHeight);

    if (w->useWorkColors)
        UnSetWorkColors(fillgc, 0);

    return;
}

static void
localDrawString(w, gcType, x, y, wstr, size, leftChar, rightChar,
				leftgcType, rightgcType)
AXWidget_editbox w;
int gcType;  /* GC_FORE, GC_BACK, GC_HILIT */
int x, y;
wchar *wstr;
int size;
wchar leftChar;  /* char on same line just to left of passed range */
wchar rightChar;  /* char on same line just to right of passed range */
int leftgcType;   /* same meaning as gcType, applies to left char */
int rightgcType;  /* same meaning as gcType, applies to right char */
{
    GC fillgc, textgc;
	GC rightgc, leftgc;
    int wid, lbear, rbear, twid;
	int unsetRightGC = FALSE;
	int unsetLeftGC = FALSE;

    if ( !w->window || !size)
        return;
	
	fillgc = getFillGC(w, gcType);
	textgc = getTextGC(w, gcType);
	rightgc = getTextGC(w, rightgcType);
	leftgc = getTextGC(w, leftgcType);

    if ( w->useWorkColors ) {
        if ( gcType == GC_FORE ) {
            SetWorkColors(fillgc, 0);
            SetWorkColors(textgc, 1);
        }
        else if ( gcType == GC_BACK ) {
            SetWorkColors(fillgc, 0);
        }
        else {
            SetWorkColors(fillgc, 1);
            SetWorkColors(textgc, 0);
        }
    }
	
	/* pre-clear using fillgc */
	wid = getTextWid(w, wstr, size, &lbear, &rbear);
	if (lbear > 0) {
		lbear = 0;    /* use left sidebearing only if it protrudes */
	}
	if (GC_HILIT == gcType && w->widgetColor) {
		/* pre-clear to hilite color, using the
		 * non-hilit text color as the hilite color
		 */
		PIXEL_VALUE savPix, hilitePix;
		savPix = AXGetForegroundPixel(Dpy, fillgc);
		if (w->textColor) {
			hilitePix = w->textColor->any.pixel;
		}
		else {
			hilitePix = AXGetForegroundPixel(Dpy, ForeGC);
		}
		XSetForeground(Dpy, fillgc, hilitePix);
		XFillRectangle(Dpy, w->window, fillgc, x + lbear, y,
					   wid - lbear + rbear,
					   w->editbox->lineHeight);
		XSetForeground(Dpy, fillgc, savPix);
	}
	else {
		/* fill to background color */
		xmColorFillRectangle((AXWidget)w, w->window, fillgc, x + lbear,
				     y, wid - lbear + rbear,
				     w->editbox->lineHeight);
	}
	
	/* set the text */
	if (gcType != GC_BACK) {
		xmDrawString16((AXWidget)w, w->window, textgc, x,
			       y+w->editbox->lineAscent, wstr, size,
			       w->textShadow,
			       ebGetTextColor((AXWidget)w, gcType),
			       w->widgetColor);
	}

	/* repair damage to neighboring text if neighbors are non-SPACE */
 	if (leftChar != SPACE) {
		if (w->useWorkColors &&
			((leftgc != textgc) || (leftgcType != gcType))) {
				
			SetWorkColors(leftgc, leftgcType == GC_FORE);
			unsetLeftGC = TRUE;
		}
		twid = getTextWid(w, &leftChar, 1, &lbear, &rbear);
		xmDrawString16((AXWidget)w, w->window, leftgc, x - twid,
			       y+w->editbox->lineAscent, &leftChar, 1, 
			       w->textShadow,
			       ebGetTextColor((AXWidget)w, leftgcType),
			       w->widgetColor);
	}
	if (rightChar != SPACE) {
		if (w->useWorkColors &&
			((rightgc != leftgc) || (rightgcType != leftgcType))) {
			
			SetWorkColors(rightgc, rightgcType == GC_FORE);
			unsetRightGC = TRUE;
		}
		xmDrawString16((AXWidget)w, w->window, rightgc, x + wid,
			       y+w->editbox->lineAscent, &rightChar, 1, 
			       w->textShadow,
			       ebGetTextColor((AXWidget)w, rightgcType),
			       w->widgetColor);
	}

    if ( w->useWorkColors ) {
        if ( gcType == GC_FORE ) {
            UnSetWorkColors(fillgc, 0);
            UnSetWorkColors(textgc, 1);
        }
        else if ( gcType == GC_BACK ) {
            UnSetWorkColors(fillgc, 0);
        }
        else {
            UnSetWorkColors(fillgc, 1);
            UnSetWorkColors(textgc, 0);
        }
		if (unsetRightGC) {
			UnSetWorkColors(rightgc, rightgcType == GC_FORE);
		}
		if (unsetLeftGC) {
			UnSetWorkColors(leftgc, leftgcType == GC_FORE);
		}
    }
}

static GC
getFillGC(w, gcType)
AXWidget_editbox w;
int gcType;
{
	GC fillgc;
	
    if ( w->monospace_font ) {
        if ( gcType == GC_FORE ) {
            fillgc = MonospaceBackGC;
        }
        else if ( gcType == GC_BACK ) {
            fillgc = MonospaceBackGC;
        }
        else {   /* GC_HILIT */
            fillgc = MonospaceGC;
        }
    }
    else {
        if ( gcType == GC_FORE ) {
            fillgc = BackGC;
        }
        else if ( gcType == GC_BACK ) {
            fillgc = BackGC;
        }
        else {   /* GC_HILIT */
            fillgc = ForeGC;
        }
    }
	return (fillgc);
}

static GC
getTextGC(w, gcType)
AXWidget_editbox w;
int gcType;
{
	GC textgc;
	
    if ( w->monospace_font ) {
        if ( gcType == GC_FORE ) {
            textgc = MonospaceGC;
        }
        else if ( gcType == GC_BACK ) {
            textgc = MonospaceGC;
        }
        else {   /* GC_HILIT */
            textgc = MonospaceBackGC;
        }
    }
    else {
        if ( gcType == GC_FORE ) {
            textgc = ForeGC;
        }
        else if ( gcType == GC_BACK ) {
            textgc = ForeGC;
        }
        else {   /* GC_HILIT */
            textgc = BackGC;
        }
    }
	return (textgc);
}

static wcolor_t *
ebGetTextColor(w, gcType)
AXWidget w;
int gcType;             /* GC_HILIT, e.g. */
{
	if (GC_HILIT == gcType && w->widgetColor) {
		return (w->widgetColor);  /* invert widget,entry colors */
	}
	else {
		return (w->textColor);
	}
}

/* : mouseOutsideEditbox - return whether passed mouse
 * event, assumed to be for the editbox window, is outside
 * the editbox' footprint.  Can return "noEdge", meaning
 * mouse event is inside editbox footprint
 */
static enum editboxEdge
mouseOutsideEditbox(w, eventP)
AXWidget_editbox w;
XEvent *eventP;
{
	enum editboxEdge whichedge;
	int fudge = 5;
	
	if (eventP->xbutton.y < -fudge) {
		whichedge = topEdge;
	}
	else if (eventP->xbutton.y > (int) w->height + fudge) {
		whichedge = bottomEdge;
	}
	else {
		whichedge = noEdge;
	}

	return whichedge;
}

/***********************************************************************

    NOTES:

    *   w->hasFocus indicates that the editbox is the active widget in
        the parent dialog.  The cursor is drawn only if w->hasFocus 
        is TRUE.  The cursor type - active/inactive - depends on whether
        the parent dialog has InputFocus (OpenLook only).

    *   w->pending_delete indicates that the editbox owns the XA_PRIMARY
        selection atom.  The editbox may own XA_PRIMARY, but not have 
        anything hilit (w->editbox->sel.startIndex & endIndex indicate 
        the actual extents of the hilit mark).

    *   w->deleteMarked indicates that the hilit text should be deleted
        during an editing action.  This is FALSE only during middle
        mouse button paste.

***********************************************************************/



#if BUILD_ORIENT    /** {  **/

ReplaceEntryText(w, text, add_len)
AXWidget_editbox w;
wchar *text;
int add_len;
{
    any_obj_t *op;
    wchar *start_replace;
    int change_len;
    int i;
    AXWidget_main main_w;

    main_w = TheMainWidget((AXWidget)w);

    /* lift old cursor */
    drawCursor(w, TC_ERASE);

    if (w->pending_delete) 
        {
        /* something is selected already, this must be the first character */
        /* that was entered, so we never have to delete any old stuff      */
        deleteSelection(w);
        main_w->conversion_start = w->editbox->tcursor.ac_index;
        main_w->conversion_end = main_w->conversion_start + add_len;
        manageInsert( w, text, add_len );
        return;
        }
    

    /* conversion input is always in the same object as the text cursor */
    op = w->editbox->tcursor.obj;
                    
    change_len = main_w->conv_change_end - main_w->conv_change_start;
    
#if !BUILD_KANJI
   /* If the wide kanji is replaced by narrow kana, the number of bytes are
    * not changing, but the width of the new string is shorter than old one.
    * The old staring text will be leave behind on the screen, if we do string
    * replace as follows.
    */
    if ( change_len == add_len )
       {
       /* we are doing a character for character replacement, take the */
       /* fast route.  This type of operation happens fairly frequently */
       /* during the keyboard entry process                             */
       start_replace = op->wdata + main_w->conv_change_start;
       for ( i=0; i<add_len; i++ )
          { *start_replace++ = *text++; }
       /* get the display updated */
       displayObjs( w, (w->editbox->yoffset + op->ypos), op->height );
       drawCursor(w, TC_DRAW);   /* draw the cursor */
       return;
       }
#endif

    /* Now we start the harder case, where the text we are to put in is, */
    /* not the same length as the test iwe are to take out.  We accomplish */
    /* this by the reliable (but not necessisarily fastest) way by emulating */
    /* the selection mechanism.  We make it look like the text to be deleted */
    /* has been selected, and then call the insertion code which will    */
    /* delete the "current selection" and add in the new characters      */

    if ( change_len > 0 )
       {
       w->pending_delete = TRUE;
       w->deleteMarked = TRUE;
       w->editbox->sel.startObj = w->editbox->sel.endObj = op;
       w->editbox->sel.startIndex = main_w->conv_change_start;
       w->editbox->sel.endIndex = main_w->conv_change_end;
       }
    
    manageInsert( w, text, add_len );

    resetSelectionInfo( w );

    displayObjs( w, (w->editbox->yoffset + op->ypos), op->height );
    drawCursor(w, TC_DRAW);   /* draw the cursor */

    return;

}

#endif     /*  }   BUILD_ORIENT  */



/* ---------------- */
