/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmeditbx.h
   Applix ASC version: 9
   Date: Mon Nov 16 13:53:35 1998
*/
#ifndef INC_XMEDITBX
#define INC_XMEDITBX
/*
 * For describing contents of a multiline text widget.
 * The contents are considered as a single-strand ordered
 * list of objects. (typ. special case: 1 object, type text.)
 */
#define TEXTOBJ 1     /* obj type */
#define TEXTUNIT 192 /* alloc unit as edit-box buffer is "grown"  */

/* 
 * Max editbox data size (512k - 2*TEXTUNIT - 64).  Subtract 2*TEXTUNIT
 * because TEXTUNIT initially allocated and TEXTUNIT extra always 
 * added when text memory is managed in manageTextMem.  Subtract 64
 * because malloc adds sizeof mem_header to requested size, and then looks
 * for nearest power of 2.  All this calculation so that if malloc needs 
 * to sbrk, max size requested from system is 512k, not next power of 2.
 * Resulting size allows max of MAXTEXTMEM  characters to be passed
 * down from the elf layer, and a max of MAXTEXTMEM+2*TEXTUNIT 
 * characters totally. 
 */
                 /* 0x80000 == 512K  */
#define MAXTEXTMEM (0x80000 - (2*TEXTUNIT) - 64)
 
#define TC_RIGHT 1 /* text cursor movement direction */
#define TC_LEFT 2
#define TC_UP 3
#define TC_DOWN 4
#define TC_DRAW 5
#define TC_ERASE 6
#define TC_ACTIVE 7
#define TC_INACTIVE 8

/* This widget uses only the "gui system font".
 * For our product's Motif interface, that's:
 * Motif regular:
 * -applix-axsab-bold-r-normal-a16-16-150-75-75-p-65-iso8859-1
 * Motif big:
 * -applix-axhvb-bold-r-normal--19-180-75-75-p-104-iso8859-1
 * For our OpenLook interface:
 * -b&h-lucida-medium-r-normal-sans-12-120-75-75-p-71-iso8859-1
 * These fonts' basic height numbers are:
 */
#define MO_FONT_ASCENT 13
#define MO_FONT_DESCENT 3
#define MO_FONT_HT ((MO_FONT_ASCENT) + (MO_FONT_DESCENT))
#define MO_BIG_FONT_ASCENT 14
#define MO_BIG_FONT_DESCENT 4
#define MO_BIG_FONT_HT ((MO_BIG_FONT_ASCENT) + (MO_BIG_FONT_DESCENT))
#define OL_FONT_ASCENT 11
#define OL_FONT_DESCENT 2
#define OL_FONT_HT ((OL_FONT_ASCENT) + (OL_FONT_DESCENT))

/* gc types for drawing text */
#define GC_FORE 1
#define GC_BACK 2
#define GC_HILIT 3
#define GC_UNDERLINE 4

/* format reasons */
#define FMT_ALL 1
#define FMT_LINE 2
#define FMT_YPOS 3

typedef struct _tline_t {
	struct _tline_t *next;
	struct _tline_t *prev;
	int first_index;     /* index within obj-data of first char on line */
	int len;       /* size of formatted line of text */
	int xpos;      /* xpos of line relative to start of text object */
	int ypos;      /* ypos of line relative to start of text object */
} tline_t;

typedef struct {
	tline_t *formatLine;  /* line for start of reformatting */
	int numFormatLines;   /* min number of lines to reformat */
	wchar *oldData;        /* data corresponding to numFormatLines */
	int *oldIndex;        /* array of old line indices */
	int *oldLen;          /* array of old line lengths */
	int *oldXpos;         /* array of old xpos wrt win for each line */
	int *oldYpos;         /* array of old ypos wrt win for each line */
} tformat_t;

typedef struct _any_obj_t {
	struct _any_obj_t *next;	  
	struct _any_obj_t *prev;	  
	int type;
	int wsize;        /* actual size of data in wchar units */
	int  allocsize;  /* size of allocated memory for data */
	wchar *wdata;      /* single buffer holding all object's text */
	tline_t *lines;  /* linked list describing formatted state */
	int xpos;        /* xpos of obj relative to 0,0 of editbox */
	int ypos;        /* ypos of obj relative to 0,0 of editbox */
	int height;      /* formatted total height of this obj */
	tformat_t *formatInfo;  /* info required for reformatting text */
} any_obj_t;

typedef struct {
	int type;       /* active or inactive */
	any_obj_t *obj;	/* ptr to obj containing cursor (always text obj) */
	int ac_index;      /* index within obj-data of char just after cursor */
} tcursor_t;

typedef struct {
	any_obj_t *startObj; /* obj containing start of selection */
	int startIndex;      /* index within startObj->data */
	any_obj_t *endObj;   /* obj containing end of selection */
	int endIndex;        /* index within endObj->data */
	mbchar *clipboard;     /* saved selection (in XA_STRING format) */
} seln_t;

typedef struct {
	int xpos;
	int ypos;
	int width;
	int height;
	int xoffset;       /* horz offset of editbox into viewport  */
	int yoffset;       /* vert offset of editbox into viewport */
	int lineHeight;    /* for text */
	int lineAscent;    /* for text */
	seln_t sel;         /* selection info */
	tcursor_t tcursor; /* cursor info */
#if BUILD_ORIENT
    tcursor_t conv_info; /* converstion information */
#endif
	any_obj_t *objs;   /* linked list of all objects in editbox */
} editbox_t;

/* ---------------- */

#endif    /* end of total include */
