/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmenter.c
   Applix ASC version: 86
   Date: Thu Apr 29 21:14:32 1999
*/
#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>
#include <X11/keysym.h>
#include <wchars.h>
#include <elf_env.h>
#include <eldata.h>
#include <xmdlged.h>
#include <tmkeys.h>

#if !BUILD_WIDE
#define xmDrawString16 xmDrawString
#endif

#define MAX_XSTR 8000  /* length of seln to be inserted - same as in ATM */
/* IS_NGC adapted from ATM's tmin1.c definition */
#define DEL ((unsigned) 0177)  /* ascii code */
#define IS_NGC(c) ((((unsigned) (c)) < (unsigned) ' ') || \
                   (((unsigned) (c)) == DEL) || \
				   (((unsigned) (c)) == BS))

#define BORDER_WIDTH 2 /* generic shadow borders */
#define TEXT_OFFSET  2 /* offset of text from border edge */	
	
extern int CharWids[256];
extern int AxCharAbove;
extern Cursor ThimRecordCursor, ThimNoRecordCursor;

extern int AxBigFont;
extern Display *Dpy;
extern short DpyDepth;

extern GC ForeGC, BackGC;
extern GC SunGC,  ShadowGC;
extern GC DisabledGC;

extern Cursor ThimPointerCursor;
extern Cursor ThimTextCursor;

extern int DoubleMillisec;
extern AXWidget ClickWidget;

extern unsigned long BackWorkPixel;
extern Colormap WorkColormap;
extern long WinMask;
extern XSetWindowAttributes WinAtt;
extern int WidgetNeedsUpdated;  /* Set when a widget needs repainted */
extern int AxButtonDown;

extern Atom AxXA_ENTER_PROP;

extern AXWidget CrtWidget();
extern AXWidget topWidget();

extern void wEnterPixHeight();
extern void paintCompositeWidget();
extern void AxDrawBevel();
extern void xmColorFillRectangle();
extern int  EntInsertText(/* widget_t *w, wchar *wbuf, int num_chars */);
extern void EntAdjustCursorPos(/* widget_t *w */);
extern AXWidget_main TheMainWidget(/* AXWidget w */);

#include "xmenter.p"

/* Interesting widget_t members used by W_ENTER:
   MEMBER         USAGE                                    RESOURCE
   w->window      Window of entry box                      -----
   w->display                                              -----
   w->mapped                                               -----
   w->xpos        xpos of entry window                     XmNx
   w->ypos        ypos of entry window                     XmNy
   w->width       width of entry window                    XmNwidth
   w->height      height of entry window                   XmNheight
   w->wstr        text in the entry box.                   XmNvalue
   w->curPos      0 based cursor position in
                  `w->wstr[]'                              ------
   w->maxLen      maximum # of chars allowed in the box    XmNmaxLength

   NOTE:
   The entry box's parent needs to be a W_DIALOG, or else, things
   won't work.

   This widget is now free of OpenLook code.
   */

/***********************************************************************
function: PaintEnter - Paint an entry box.
*/
void
PaintEnter(w, ev, full_regen, layered)
    AXWidget_entry w;
    XEvent *ev;
    int full_regen;
    int layered;
{
    int win_xpos;
    int win_ypos;
    unsigned int win_height;
    unsigned int win_width;
    int xpos, winHt;
    int ypos;
    int xtext, ix, xcurs;
    AXWidget_main main_w;
    wchar *charP;
    int under_start, highlight_start; /* in pix from beginning of string */
    int under_len, highlight_len;      /* in pix */
    XSetWindowAttributes atts;
    int char_width, fontAscent, lpending_delete;
    int new_win = FALSE;
    GC lforeGC;
    long mask;
    
    if (!w->display)
        return;

    if (w->parent && (w->parent->type == W_DIALOG) &&
		!((AXWidget_dialog)w->parent)->exposed)
        return;
    
    if (layered) {
        if (w->flags & WFLG_IN_EDIT) {
            if (dePaintWidget((AXWidget)w)) {
                return;
            }
        }
        else if (w->parent && w->parent->layerList && w->auxData) {
            wComposite *cp;
			/* instead of unconditional call to paintCompositeWidget,
			 * paint composite kid (W_LABEL in this case) only if
			 * it's in the update_widget state, or if full_regen
			 */
            for (cp = w->auxData->compositeList; cp; cp = cp->next) {
                if (cp->type == WLINKDOWN && cp->w->painter &&
                    (cp->w->update_widget || full_regen)) {
                    
                    cp->w->painter(cp->w, ev, full_regen, FALSE);
                    cp->w->update_widget = FALSE;
                }
            }
        }
		if (w->parent && (W_DIALOG == w->parent->type) && w->hasFocus) {
			xmPaintFocusSelector((AXWidget)w);
		}
    }
    
    w->mapped = TRUE;

    wEnterPixHeight(w->font, w->textShadow, &winHt, (int *)(&w->height));
    win_height = (unsigned int) winHt;
    ypos = w->ypos - 4;
    win_ypos = ypos + 4;
    win_width = w->width - 8;
    xpos = w->xpos + 2;
    win_xpos = xpos + 4;
    xtext = 0;
    
    under_len = highlight_len = 0;
    under_start = highlight_start = -1;
    
    if (w->font) {
        fontAscent = w->font->ascent;
    }
    else {
        fontAscent = AxCharAbove;
    }


    
#if BUILD_ORIENT
    main_w = TheMainWidget((AXWidget)w); /* get to W_MAIN */
    if ((main_w->inWideInput) &&
        ((main_w->conversion_start >= 0) || (main_w->conv_phrase_start >= 0))) {
        xcurs = 0;
        charP = w->wstr;
        ix = -1;
        while (*charP != 0) {
            ix++;
            if (main_w->conversion_start == ix)
                under_start = xcurs;
            if (main_w->conv_phrase_start == ix)
                highlight_start = xcurs;

            char_width = GetAnyCharWidth(w->font, *charP);
            xcurs += char_width;

            if ((ix >= main_w->conversion_start) && (ix <= main_w->conversion_end))
                under_len = under_len + char_width;
            if ((ix >= main_w->conv_phrase_start) && (ix <= main_w->conv_phrase_end))
                highlight_len = highlight_len + char_width;

            charP++;
        }
    }
    if ( !(w->flags & WFLG_IN_EDIT) && (main_w->allow_wide_input == FALSE) )
    {
        /* Ashisuto added for BUILD_ORIENT: the XIC may be not enabled yet
         * because the enter-able (typing-able) at the first time of creating
         * might be disabled. If the XIC is disabled, it will be never be 
         * rebuild because of the XCreateIC is called by re-paint callback.
         * Here is the nice place(?) to enable the XIC if it is necessary.
         * 1996.12.24
         */
        main_w->allow_wide_input = TRUE;
        PaintOrientalStatus( main_w->kanji_status, NULL, FALSE, FALSE );
    }
#endif
    
    AxSetWidgetWorkColors((AXWidget)w);

    if (!w->window) {
        mask = WinMask | CWCursor;

        atts = WinAtt;
        if (ElfInLearnMode) {
            /* if parent W_MAIN has disabled elf recordings, set
               cursor type to `ThimNoRecordCursor' */
#if !BUILD_ORIENT
            main_w = (AXWidget_main)w->parent;
            while (main_w) {
                if (main_w->type == W_MAIN)
                    break;
                main_w = (AXWidget_main)main_w->parent;
            }
#endif
            if (!main_w) {
                AxUnsetWidgetWorkColors((AXWidget)w);
                return; /* no main?? */
            }
            if (main_w && main_w->disable_record) {
                atts.cursor = ThimNoRecordCursor;
			} else {
                atts.cursor = ThimRecordCursor;
			}
        } else {
            atts.cursor = ThimTextCursor;
        }

        full_regen = TRUE;
        if (w->flags & WFLG_IN_EDIT) { /* dialog editor */
			/* create main window */
            w->window = AxCreateQueryPixmap(Dpy, w->parent->window,
                                            win_width, win_height, DpyDepth);
			xmColorFillRectangle((AXWidget)w, w->window, BackGC, 0, 0,
								 win_width, win_height);
			/* create edit window */
			w->edit_win.win = AxCreateQueryPixmap(Dpy, w->parent->window,
												  win_width, win_height,
												  DpyDepth);
			xmColorFillRectangle((AXWidget)w, w->edit_win.win, BackGC, 0, 0,
								 win_width, win_height);
        } else {
			/* we don't need to do the cursor twice */
			mask |= CWBackPixel | CWColormap;
			atts.background_pixel = BackWorkPixel;
			atts.colormap = WorkColormap;

			/* create main window */
            w->window = AxCrtWindow(Dpy, WIN_OF(w), win_xpos, win_ypos,
                                    win_width, win_height, 0, CopyFromParent,
                                    InputOutput, CopyFromParent, mask, &atts);
            SetWindowHints((AXWidget_main)w, 0);
            XMapWindow(Dpy, w->window);
            AddToWindowList((AXWidget)w);
			xmColorFillRectangle((AXWidget)w, w->window, BackGC, 0, 0,
								 win_width, win_height);

			/* create edit window */
			w->edit_win.win = AxCrtWindow(Dpy, w->window, 0, 0,
										  win_width, win_height,
										  0, CopyFromParent, InputOutput,
										  CopyFromParent, mask, &atts);
			XSelectInput(Dpy, w->edit_win.win,
						 (TM_ALL_INPUTS_MASK | Button1MotionMask));
			XMapWindow(Dpy, w->edit_win.win);

			AXDefineDefaultCursor((AXWidget)w, atts.cursor);
			new_win = TRUE;
		}
    }
    
    if (w->pending_delete) {
        if (!w->wstr || !w->wstr[0]) {
            w->pending_delete = FALSE;
        }
		w->curPos = w->edit_win.cursor_pos;
    }

	if (w->widgetColor) {
		/* fill the 2pix margin which surrounds the
		 * enterable field window, but which is
		 * inside the enterable field's bevel
		 */
		xmColorFillRectangle((AXWidget)w, WIN_OF(w), BackGC, win_xpos - 2,
							 win_ypos - 2, win_width + 4, win_height + 2);
		
		/* fill inside the enterable field window too */
		xmColorFillRectangle((AXWidget)w, w->window, BackGC, 0, 0,
							 win_width, win_height);
	}
	else {
		/* fill the 2pix margin which surrounds the
		 * enterable field window, but which is
		 * inside the enterable field's bevel
		 */
		XFillRectangle(Dpy, WIN_OF(w), BackGC, win_xpos - 2,
					   win_ypos - 2, win_width + 4, win_height + 2);
	}
	
	AxDrawBevel((AXWidget)w, WIN_OF(w), xpos, ypos, w->width, w->height,
				2, BEVEL_IN);

    if ((w->disabled) && (w->type != W_KANJI_KANA)) {
        lforeGC = DisabledGC;
        lpending_delete = FALSE;
    }
    else {
        lforeGC = ForeGC;
        lpending_delete = w->pending_delete;
    }

#if BUILD_ORIENT
    if (highlight_start >= 0) {
		/* we are to highlight a conversion zone. */
		if ((xtext + highlight_start) < 0) {
			/* highlight region starts to the left of field, shift
			   field */
			xtext = -highlight_start;
		}
		if ((xtext + highlight_start + highlight_len) > w->width) {
			/* highlight extends off right end. */
			xtext = xtext - (highlight_start + highlight_len - w->width);
		}
    }
#endif

#if BUILD_ORIENT
    if (w->type == W_KANJI_KANA) {
        EntActivateEditWin(w);
		if (under_start >= 0) {
			/* draw a line under the conversion zone. */
			XDrawLine(Dpy, w->edit_win.win, lforeGC, (under_start + xtext),
					  fontAscent + 3, (under_start + xtext + under_len),
					  fontAscent + 3);
		}
		
		if (highlight_start >= 0) {
			/* reverse the active phrase */
			XFillRectangle(Dpy, w->edit_win.win, lforeGC, 
						   (highlight_start + xtext),
						   0, highlight_len, win_height - 3);
			xmDrawString16((AXWidget)w, w->edit_win.win, BackGC,
					(highlight_start + xtext + 1), /* xpos */
					(fontAscent + 1), /* ypos */
					(w->wstr + main_w->conv_phrase_start), /* str start */
					(main_w->conv_phrase_end - main_w->conv_phrase_start +1),
					w->textShadow, w->textColor, w->widgetColor);
		}
		return;
	}
#endif

    if (!(w->flags & WFLG_IN_EDIT)) {
		EntActivateEditWin(w);
		
		if (w->hasFocus) {
			if (new_win || (w->edit_win.text_len == 0) || w->pending_delete)
				w->edit_win.cursor_pos = 0;
			EditWinDrawCursor((AXWidget)w, &(w->edit_win), TC_DRAW);
		}
    }
    
    AxUnsetWidgetWorkColors((AXWidget)w);
    
    if (w->flags & WFLG_IN_EDIT) {
		XCopyArea(Dpy, w->edit_win.win, w->window, ForeGC, 0, 0,
				  win_width - (2*BORDER_WIDTH), win_height - (2*BORDER_WIDTH),
				  BORDER_WIDTH, BORDER_WIDTH);
        XCopyArea(Dpy, w->window, w->parent->window, ForeGC, 0, 0,
                  win_width, win_height, win_xpos, win_ypos);
    }

    return;
}

void EntActivateEditWin(w)
    AXWidget_entry w;
{
    int text_width;
    w->edit_win.text = (wchar *)realloc(w->edit_win.text,
							          (sizeof(wchar) * Wstrlen(w->wstr)) + 1);
    Wstrcpy(w->edit_win.text, w->wstr);
	w->edit_win.cursor_pos = w->curPos;
	w->edit_win.text_len = Wstrlen(w->edit_win.text);
    EditWinPaintWindow((AXWidget)w, &(w->edit_win));

    if (w->hasFocus) {
		if (w->pending_delete || (w->curPos == -1)) {
            text_width = EditWinGetStringWidth((AXWidget)w, w->edit_win.text, 0,
                     w->edit_win.text_len) + w->edit_win.string_offset;
			EditWinHighlightLine((AXWidget)w, &(w->edit_win), text_width);
		}
    }
}


/* The following function has been absorbed by EditWinGetStringWidth() in which
  * the password case "no_echo" was considered.
  * Ashisuto, 1996.11.22
  */
#if 0   /*  {  */

/***********************************************************************
function: WEnterTextLen - Return the pixel length of LENGTH number of
(wide)chars in the entry box starting at STARTIX.
*/
static int
WEnterTextLen(w, startix, length)
    AXWidget_entry w;
    int startix;
    register int length;
{
    register int result;
    register int ix;
    register wchar *p;
    wchar asterisk_char = '*'; /* Nothing to do with the product name. */

    if (w->no_echo) {
        result = GetAnyCharWidth(w->font, (int) asterisk_char) * length;
    } else {
        p = w->wstr + startix; /* FIXME, what if `p' is beyond the end
                                  of the string? */
        for (result = 0, ix = 0; ix < length; ix++) {
            result += GetAnyCharWidth(w->font, *p);
            p++;
        }
    }
    return(result);
}

/***********************************************************************
function: DrawEnterString - Draw string in entry box.  Draw CHAR_LEN
number of chars starting from CHAR_START in `w->wstr'.
*/
static void
DrawEnterString(w, gc, x, y, char_start, char_len)
    AXWidget_entry w;
    GC gc;
    int x;
    int y;
    int char_start;
    int char_len;
{
    wchar *outstr;
    int ix;

    if (char_len < 0)
        char_len = Wstrlen(w->wstr);

    if (!w->no_echo) {
        xmDrawString16((AXWidget)w, w->window, gc, x, y,
					   (char *)(w->wstr + char_start),
                       char_len, w->textShadow, w->textColor, w->widgetColor);
    } else {
        /* To echo passwords... */
        outstr = (wchar *) malloc((char_len + 1) * sizeof(wchar));
        for (ix = 0; ix < char_len; ix++)
            outstr[ix] = '*';
        outstr[ix] = 0;

        xmDrawString16((AXWidget)w, w->window, gc, x, y, (char *)outstr,
					   char_len, w->textShadow, w->textColor, w->widgetColor);
        free(outstr);
    }
    return;
}
#endif     /* #if 0  }  */

/***********************************************************************
function: axolInTextScrollButtons - Compat -- No more OpenLook.
*/
axolInTextScrollButtons(w, eventP)
    AXWidget_entry w;
    XEvent *eventP;
{
    return(FALSE);
}

/***********************************************************************
function: EntryBoxEvent - Handle all types of XEvents in the entry box
widget.
Return TRUE if event was consumed; FALSE otherwise.
*/
int
EntryBoxEvent(w, eventP)
    AXWidget_entry w;
    XEvent *eventP;
{
    int event_consumed = FALSE;

    if (!w || !eventP)
        return(FALSE); /* invalid args */
    if (w->type != W_ENTER)
        return(FALSE); /* invalid args */

    if (w->disabled || (w->type != W_ENTER))
        return(TRUE); /* trap out */


    switch (eventP->xany.type) {
    case ButtonPress:
		EditWinManageButtonPress((AXWidget)w, &(w->edit_win), eventP);
        event_consumed = TRUE;
        break;
    case ButtonRelease:
		EditWinManageButtonRelease((AXWidget)w, &(w->edit_win), eventP);
        event_consumed = TRUE;
        break;
    case MotionNotify:
		EditWinManageMouseMotion((AXWidget)w, &(w->edit_win), eventP);
        event_consumed = TRUE;
        break;
    case KeyPress:
        event_consumed = EntKeyPressCB(w, eventP, FALSE);
        break;
    case SelectionNotify:
        event_consumed = EntSelectionNotifyCB(w, eventP);
        break;
    case SelectionClear:
    case SelectionRequest:
    default:
        event_consumed = FALSE;
        break;
    }
    return(event_consumed);

ignore_event:
    return(TRUE);
}

/***********************************************************************
function: EntSelectionNotifyCB - Handle a SelectionNotify event in the
entry box.  Return TRUE if event was consumed; FALSE otherwise.
*/
static int
EntSelectionNotifyCB(w, ev)
    AXWidget_entry w;
    XEvent *ev;
{
    Atom prop, actual_type;
    long long_size;
    unsigned long nitems, bytes_after, *data;
    int len, actual_format;
    int changed = FALSE;
    int ix;
    wchar *wbuf = NULL;
    mbchar *buf = NULL;
    mbchar *p;

    if (!w)
        return(FALSE);
    if (w->type != W_ENTER)
        return(FALSE);
    if (!ev)
        return(FALSE);
    if (ev->xany.type != SelectionNotify)
        return(FALSE);

    prop = ev->xselection.property;
    if (prop != AxXA_ENTER_PROP)
        return;
    long_size = MAX_XSTR/ sizeof(long);
    if (XGetWindowProperty(Dpy, w->window, prop, 0, long_size, False,
                           AnyPropertyType, &actual_type,
                           &actual_format, &nitems, &bytes_after,
                           (unsigned char **) &data) != Success) {
        return(TRUE);
	}
	XDeleteProperty(Dpy, w->window, prop);

    if (actual_type != XA_STRING)
        return(TRUE);

    buf = (mbchar *)malloc(MBstrlen((mbchar *)data) + 1);
	
    /* Accept buffer, ignoring newline chars  */
    for (ix = 0, p = (mbchar *)data; *p; p++) {
		if (*p == '\n')
			continue;
		buf[ix] = *p;
		ix++;
    }
    buf[ix] = '\0';

#if BUILD_ORIENT
    wbuf = (wchar *)malloc(sizeof(wchar) * (MBstrlen(buf) + 1));

    MbToWStr((mbchar *)buf, wbuf); 
    EditWinInsertText((AXWidget)w, &(w->edit_win), wbuf);
    free(wbuf);
#else
    EditWinInsertText((AXWidget)w, &(w->edit_win), buf);
#endif
	
    free( buf );

    return(TRUE);
}

extern KeySym AxEuroDollarKeysym;							/* method 1 */
extern int AxEuroDollarModifiers;							/* method 1 */
extern int AxEuroDollarKeyState, AxEuroDollarKeyCode ;		/* method 2 */

/***********************************************************************
function: EntKeyPressCB - Handle a KeyPress event in the entry box.
Return TRUE if key-press event was consumed, FALSE otherwise.
*/
int
EntKeyPressCB(w, ev, in_menu_bar)
    register AXWidget_entry w;
    XEvent *ev;
	int in_menu_bar;
{
    int event_consumed = FALSE;
    int changed = FALSE;
    int num_chars = 0;
    wchar *wbuf, buf[100];
    KeySym ksym;
    AXWidget_main w_main   = (AXWidget_main) NULL;
    AXWidget_dialog w_dialog = (AXWidget_dialog) NULL;
    AXWidget next_w   = (AXWidget) NULL;
    AXWidget wp;
	int euro = FALSE;

    if (!w || !ev)
        return(FALSE);

    if (ev->xany.type != KeyPress)
        return(FALSE);

    if (w->disabled || (w->type != W_ENTER) || !w->hasFocus)
		return(FALSE);  /* ignore inactive widget */
	
    w_dialog = (AXWidget_dialog)w;
    if (w->parent && w->parent->type == W_DIALOG)
        w_dialog = (AXWidget_dialog)w->parent;

    if (!w_dialog)
        return(FALSE);

    w_main = w_dialog->parent;
    if (w_main->type != W_MAIN)
        return(FALSE); /* trap out */

	if (AxEuroDollarKeyCode && (AxEuroDollarKeyCode == ev->xkey.keycode) && 
								(AxEuroDollarKeyState == ev->xkey.state)) {
		/* we have found the euro dollar key sequence, return back EuroDollar
			ascii */
		buf[0] = EURO_SYMBOL_8BIT_VALUE;
		buf[1] = 0;
		wbuf = buf;
		num_chars = 1;
		ksym = 0;
		euro = TRUE;
	} else {
		num_chars = AXLookupString16(ev, &wbuf, &ksym);    
	}

	/* High-level Euro keysym - hardware-independent */
	if (AxEuroDollarKeysym && (AxEuroDollarKeysym == ksym)) {
		int mods = AxDecodeModifiers(ev->xkey.state);
		if (mods == AxEuroDollarModifiers) {
			buf[0] = EURO_SYMBOL_8BIT_VALUE; 
			buf[1] = 0;
			wbuf = buf;
			num_chars = 1;
			ksym = 0;
			euro = TRUE;
		}
	}

	/* If this keystroke is or might be in the menubar, and it wasn't
		a Euro, reject it */
	if (in_menu_bar && !euro)
    	return(FALSE);

    event_consumed = EditWinManageKeyPress((AXWidget)w, &(w->edit_win), ksym,
										   wbuf, num_chars, ev);

    return(event_consumed);
}


/***********************************************************************
  function: GetWidgetString - Return the string `w->wstr' of entry widget.
  This is kept for compatibility reasons.
*/
mbchar *
GetWidgetString(w)
    AXWidget_entry w;
{
    if (w && (w->type == W_ENTER)) {
#if !BUILD_WIDE
        return((mbchar *) w->wstr);
#else
        return(WToMbOneStr(w->wstr));
#endif
    }

    return((mbchar *)NULL);
}







#if BUILD_ORIENT  /*   {   */

EntUpdateEditWinString( w, edit_win, wstr, curPos )
AXWidget w;
struct edit_window *edit_win;
wchar *wstr;
int curPos;
{
    int win_height;
    int win_width;
    int xtext, i, xcurs;
    AXWidget_main main_w;
    wchar *charP;
    int under_start, highlight_start; /* in pix from beginning of string */
    int under_len, highlight_len;      /* in pix */
    int char_width, fontAscent;

    main_w = TheMainWidget( w );

    wEnterPixHeight( w->font, w->textShadow, &win_height, (int *)(&w->height) );
    win_width = w->width - 8;
    xtext = 0;
    if (w->font)
        fontAscent = w->font->ascent;
    else
        fontAscent = AxCharAbove;
    under_len = highlight_len = 0;
    under_start = highlight_start = -1;

    /* Calculate underline, highlight data */
    if ( (main_w->conversion_start >= 0) || (main_w->conv_phrase_start >= 0) )
    {
        xcurs = 0;
        charP = wstr;
        i = -1;
        while (*charP != 0)
        {
            i++;
            if ( main_w->conversion_start == i )
                under_start = xcurs;
            if ( main_w->conv_phrase_start == i )
                highlight_start = xcurs;
            char_width = GetAnyCharWidth( w->font, *charP );
            xcurs += char_width;
            if ( i >= main_w->conversion_start && i <= main_w->conversion_end )
                under_len = under_len + char_width;
            if (i >= main_w->conv_phrase_start && i <= main_w->conv_phrase_end)
                highlight_len = highlight_len + char_width;
            charP++;
        }
        if ( under_start > 0 )
        {
            /* we have a cursor to display in the file, make sure it is in
               the portion visible on the screen. */
            xcurs = EditWinGetStringWidth( w, edit_win->text, 0, curPos );
            if ( xcurs > win_width )
            {
                xtext = win_width - xcurs - 4;
                xcurs = xcurs + xtext;
            }
        }
        if (highlight_start >= 0)
        {
            /* we are to highlight a conversion zone. */
            if ((xtext + highlight_start) < 0)
                /* highlight region starts to the left of field, shift field */
                xtext = -highlight_start;
            if ((xtext + highlight_start + highlight_len) > w->width)
                /* highlight extends off right end. */
                xtext = xtext - (highlight_start + highlight_len - w->width);
        }
        if (curPos == 0)
        {
            /* show whole cursor */
            xtext = 1;
            xcurs = 1;
        }
    }

    AxSetWidgetWorkColors( w );

    xmColorFillRectangle(w, edit_win->win, BackGC, 0, 0, win_width, win_height);

    xmDrawString16( w, edit_win->win, ForeGC, xtext+1, fontAscent+1,
            wstr, Wstrlen(wstr), w->textShadow, w->textColor, w->widgetColor );

    if ( under_len > 0 )
    {
        /* draw a line under the conversion zone. */
        XDrawLine(Dpy, edit_win->win, ForeGC, (under_start + xtext),
                  fontAscent + 3, (under_start + xtext + under_len),
                  fontAscent + 3);
    }
    if ( highlight_len > 0 )
    {
        /* reverse the active phrase */
        XFillRectangle(Dpy, edit_win->win, ForeGC, (highlight_start + xtext),
                       0, highlight_len, win_height - 3);
        xmDrawString16(w, edit_win->win, BackGC,
                       (highlight_start + xtext + 1), /* xpos */
                       (fontAscent + 1), /* ypos */
                       (wstr + main_w->conv_phrase_start), /* str start */
                       (main_w->conv_phrase_end - main_w->conv_phrase_start+1),
                       w->textShadow, w->textColor, w->widgetColor);
    }

    AxUnsetWidgetWorkColors( w );

    edit_win->text = (wchar *) realloc( edit_win->text, 
                                       (Wstrlen(wstr) + 1) * sizeof(wchar) );
    Wstrcpy( edit_win->text, wstr );
    edit_win->text_len = Wstrlen( edit_win->text );
    edit_win->cursor_pos = curPos;

    if ( (under_len == 0) && (highlight_len == 0) )
    {
        EditWinDrawCursor( (AXWidget)w, edit_win, TC_DRAW );
    }
    return;

}


/***********************************************************************
function: EntReplaceText - Insert input text WBUF at cursor position.
          Copied from entercom.c:EntInsertText(), ref ReplaceEntryText()
          for Editbox (That func is unadequatedly named). Only be called
          at On-The-Spot PreeditDrawCallback on EntryField widgets. 
          Removed the char validation checks because here is not the
          final commit place.
Ashisuto inited, 1996.11.20
*/
int
EntReplaceText(w, wbuf, num_chars)
AXWidget_entry w;
register wchar *wbuf;
register int num_chars;
{
    wchar *new_wstr = (wchar *) NULL;
    int start_to_cursor, chars_del_before_cursor;
    int change_len;
    AXWidget_main main_w;
    seln_t *sel;

    main_w = TheMainWidget( (AXWidget)w );

    if ( w->pending_delete )
    {
        /* do pending delete, no right on display */
        sel = &(w->edit_win.sel);
        chars_del_before_cursor = sel->endIndex - sel->startIndex;
        Wstrcpy( (w->edit_win.text + w->edit_win.sel.startIndex),
                 (w->edit_win.text + w->edit_win.sel.endIndex) );
        w->edit_win.text_len -= chars_del_before_cursor;
        w->edit_win.text[w->edit_win.text_len] = 0;
        w->edit_win.cursor_pos -= chars_del_before_cursor;
        sel->endIndex = sel->startIndex;
        w->pending_delete = FALSE;

        /* reflect changes upward */
        Wstrcpy( w->wstr, w->edit_win.text );
        w->curPos = w->edit_win.cursor_pos;

        if ( chars_del_before_cursor > 0 )
        {
            if ( main_w->conversion_start >= 0 )
                main_w->conversion_start -= chars_del_before_cursor;
            if ( main_w->conv_change_start >= 0 )
                main_w->conv_change_start -= chars_del_before_cursor;
            if ( main_w->conv_phrase_start >= 0 )
                main_w->conv_phrase_start -= chars_del_before_cursor;
            if ( main_w->conversion_end > 0 )
                main_w->conversion_end -= chars_del_before_cursor;
            if ( main_w->conv_change_end >= 0 )
                main_w->conv_change_end -= chars_del_before_cursor;
            if ( main_w->conv_phrase_end >= 0 )
                main_w->conv_phrase_end -= chars_del_before_cursor;
        }
    }

    change_len = main_w->conv_change_end - main_w->conv_change_start;

    if ( (num_chars + Wstrlen(w->wstr) - change_len) >= w->maxLen ) {
        RingTheBell();
        return(TRUE);
    }

    if ( change_len > 0 ) {
        Wstrcpy((w->wstr + main_w->conv_change_start),
                (w->wstr + main_w->conv_change_end));
        start_to_cursor = w->curPos - main_w->conv_change_start;
        chars_del_before_cursor = MIN(start_to_cursor, change_len);
        w->curPos = w->curPos - chars_del_before_cursor;
    }

    new_wstr = (wchar *) malloc((Wstrlen(w->wstr)+num_chars+1) * sizeof(wchar));
    Wstrncpy(new_wstr, w->wstr, w->curPos); /* chars before cursor */
    Wstrcpy( new_wstr + w->curPos, wbuf );
    Wstrcat( new_wstr, w->wstr + w->curPos); /* append chars after cursor */

    if ( w->wstr )
        free( w->wstr );
    w->wstr = new_wstr;
    w->curPos = w->curPos + num_chars;

    EntAdjustCursorPos(w);
    EntUpdateEditWinString( (AXWidget)w, &w->edit_win, w->wstr, w->curPos );

    return;

}

#endif   /*  BUILD_ORIENT }   */





/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * end:
 */
