/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmlib.h
   Applix ASC version: 97
   Date: Mon Nov 16 13:53:34 1998
*/
#ifndef _XMLIB_H
#define _XMLIB_H

#include <basicwin.h>   /* used to be <X11/Intrinsic.h> ... basicwin.h includes
                           the appropriate file(s) for UNIX or NT port */
#include <amenudef.h>  /* menu defs */
#include <axol.h>         /* OpenLook def's */
#include <xmeditbx.h>    /* Editbox defs */
#include <xmevent.h>    /* eventRec def */
#include <xmres.h>      /* extra resource name defines */
#include <eldata.h>
#include <xim.h>
#include <evreason.h>
#include <xmdefs.h>

#define USE_SYS_MAP 0 /* use system resource for now */
#if USE_SYS_MAP
extern DLL_IMPORT Pixmap ScratchMap;
#endif

#ifndef MIN
#define MIN(a,b)    (((a)<(b))?(a):(b))
#endif
#ifndef MAX
#define MAX(a,b)    (((a)>(b))?(a):(b))
#endif
#ifndef ABS
#define ABS(x)      (((x) < 0) ? -(x) : (x))
#endif

#define MEMWIN_WID 310
#define MEMWIN_HT 95

#define MAX_NAME 512
#define MAX_MENU_ITEMS 51   /* actually 50, this is for array allocs */

/* for filling transparent shapes */
#define FILL_TRANS_POLY 0
#define FILL_TRANS_RECT 1
#define FILL_TRANS_ARC  2


/* Row-Column types */
#define RC_TYPE_SQUARE  0 /* default. must be 0 */
#define RC_TYPE_COLUMN  1 /* single column */
#define RC_TYPE_ROW     2 /* single row */
#define RC_TYPE_HORIZ   3 /* row major */
#define RC_TYPE_VERT    4 /* column major */

typedef struct menu_opt_status {    /* menu item status for dbox menus */
    char *name;     /* the "invoked" function name */
    int status;     /* from amenudef.h */
    char *item_name;  /* name of the menu item */
    struct menu_opt_status *next;
} menu_opt_status;

typedef struct scroll_history {
    int   scroll_max, scroll_per, scroll_pos;
    short forepix, pospix, availpix;
    int   num_buttons;
} scroll_history;

/***************************************************************************
                       meaning in W_SCROLL        meaning in W_LIST
                       -------------------        -----------------
   w->scroll_pos       ypos of top line in win    line # of top line in win
   w->scroll_max       total # pix in doc         max # lines visible
   w->scroll_per       max # pix visible                    -
   w->num_buttons             -                   total # lines in doc
***************************************************************************/

/* Widget types */

#define W_DIALOG 1
#define W_LABEL 2
#define W_TOGGLE 3
#define W_RADIO 4
#define W_BUTTON 5
#define W_OPTION 6
#define W_LIST 7
#define W_SCROLL 8
#define W_ENTER 9
#define W_ICON 10
#define W_HGROOVE 11
#define W_POPUP 12
#define W_RECT 13
#define W_MAIN 14
#define W_WORK 15
#define W_MENUBAR 16
#define W_PALETTE 17
#define W_RULER 18
#define W_STATUS 19
#define W_DEBUG 20
#define W_CORNER 21
#define W_TOOLBAR 22
#define W_EDITBOX 23
#define W_SPLITTER 24
#define W_SCALE 25
#define W_COLORPAL 26
#define W_RIBBON 27
#define W_GE_XRULER 28
#define W_GE_YRULER 29
#define W_ALIGN 30
#define W_TABLE 31
#define W_KANJI_KANA 32
#define W_ROWCOL 33
#define W_KANJI_STATUS 34
#define W_CANVAS 35
#define W_TRAY 36   /* Tree display widget. */
#define W_PLATES 37     /* plate select in SS */
#define W_CHOOSE 38 /* Color chooser widget. */
#define W_PAGEBAR 39 /* page info widget in GE */
#define W_BUTBAR 40 /* button widget in GE  */
#define W_COMBO_BOX 41 /* Combo box widget */
#define W_DUMMY 42 /* dummy widget */
#define W_FRAME 43 /* frame widget on dlg box */

/************************************************************************/
/* defines for the positions of the various colors in the "WidgetPixels[]" */
/*   array                                                                */
/************************************************************************/
#define WIDGET_BACKGROUND 0
#define WIDGET_FOREGROUND 1
#define WIDGET_SHADOW 2
#define WIDGET_SUN 3
#define WIDGET_RECESS 4
#define WIDGET_SUN_WORK 5
#define WIDGET_SHADOW_WORK 6
#define WIDGET_RECESS_WORK 7

/************************************************************************/
/* below is the include for the widget structure(s), and appropriate    */
/*   typedefs.  Any changes to either the windows or the X side should  */
/*   be reflected in the other side                                     */
/************************************************************************/
#if AXMS   /* {  */
#include <xmwcolor.h>
#include <winwdgts.h>     /* Windows/NT */
#define WIDGET_PARANOID( wP, w_type ) if ( wP && wP->type != w_type ) \
             tempAbort("Widget type mismatch");

#define PALVERSION 0x300    /* current pallette version.  There does not */
                            /* seem to be a real windows define for this! */

#else      /* }  {  */
#include <xmwcolor.h>
#include <xwidgets.h>    /* X windows */
#define WIDGET_PARANOID
#endif    /*  }  */


#define MAX_TOOLBAR_ITEMS 200
#define MAX_TOOLBAR_ROWS 25
#define TOOLBAR_B1_XPOS 10    /* xpos of first button on row (pixels) */

typedef struct tool_bar_t {
    menu_item_def *menuP[MAX_TOOLBAR_ITEMS];  /* ptrs to menu items */
    AXWidget widgets[MAX_TOOLBAR_ITEMS];/* parallel array of child buttons */
    } tool_bar_t;



/* window creation masks */

#define TM_MOST_INPUTS_MASK \
        KeyPressMask | KeyReleaseMask |  ResizeRedirectMask | \
        FocusChangeMask | SubstructureNotifyMask | StructureNotifyMask | \
        ButtonPressMask | ButtonReleaseMask | \
        ButtonMotionMask | ExposureMask
#define TM_KEYS_INPUT_MASK KeyPressMask | KeyReleaseMask
#define TM_ALL_INPUTS_MASK (TM_MOST_INPUTS_MASK | TM_KEYS_INPUT_MASK)

/* menu pulldown and cascade modes */
#define OPTION_NEW              0
#define OPTION_MOVE             1
#define OPTION_DONE             2
#define OPTION_STAYUP_SELECT    3
#define OPTION_EXPOSE           4
#define OPTION_SETDEFAULT_NEW   5
#define OPTION_SETDEFAULT_MOVE  6
#define OPTION_SETDEFAULT_DONE  7
#define OPTION_EXECUTE_DEFAULT  8


#define ASTER_MENU_WID      24
#define ASTER_MENU_HT       16
#define SAVE_UNDER          TRUE
#define MENU_BAR_REL_Y      (AxOpenLook ? (OLBUTTON_HT/5) : 7)
#define LISTBOX_ICON_HT     16

#define AX_WINDOW_UID_BASE 100000  /* start of AxWindowCreationCounter */

#define WIN_OF(w) (w->parent->window)

/* bevel-style constants used in calls to AxDrawBevel */
#define BEVEL_OUT            1  /* representing a raised surface edge */
#define BEVEL_IN             2  /* pushed in */
#define BEVEL_NONE           3  /* used for erasing bevels */
#define BEVEL_OUT_ASYMMETRIC 4  /* used for windows buttons where the */
                                /* sun is 1pix narrower than the shadow */
#define BEVEL_ETCH           5  /* a ditch */
#define BEVEL_RIDGE          6  /* inverted etch */
#define BEVEL_SIMPLE         7  /* a plain, solid line */

typedef struct win_list_t {
    struct win_list_t *next;
    struct win_list_t *prev;
    Window win;
#if AXMS
    HWND hwnd, parentHwnd;
    int xOffset;
#endif
} win_list_t;


unsigned long AXGetForegroundPixel(/* Display *dpy, GC gc */);
unsigned long AXGetBackgroundPixel(/* Display *dpy, GC gc */);

#endif /* _XMLIB_H */
