/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmoption.c
   Applix ASC version: 43
   Date: Mon Nov 16 13:53:36 1998
*/
#include <ax_xm.h>
#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>
#include <xmdlged.h>


extern AXWidget MenuArmed;

extern int CharWids[256];
extern int AxCharAbove, AxCharBelow, AxCharHeight;

extern int AxButtonDown;
extern unsigned long BackPixel;
extern GC ForeGC, BackGC;
extern GC DisabledGC, SunGC, ShadowGC;
extern XFontStruct *Txfont, *BoldTxfont, *Grafont;
extern Window RootWin;
extern Display *Dpy;
extern short DpyDepth;

extern int WidgetNeedsUpdated;
extern int ThimExposuresMayBeComing;

extern int  widestOptionChoice();
extern XFontStruct *AxDlgMainFont;
extern void wWidgetShadowOffsets();
extern void xmDrawRectShadow(), xmDrawRectAntiShadow();
extern void xmColorFillRectangle();
extern void paintCompositeWidget();
extern void AxDrawBevel();

extern wAux *newAuxData();

extern AXWidget topWidget();
extern AXWidget backgroundWidget();

void PaintOption();
void PaintOptionPulldown();
void wOptionWidth();

/*************************************************************************
function: PaintOption - Paint static choices gadget
*/
void PaintOption (w, exP, regen, layered)
register AXWidget_option w;
XEvent *exP;
int regen;
int layered;
{
    register GC gc;
    register int strWid = 0;
    register char *p, *str;
    register int wid, ix, jx, this, x, y, max;
    int width, labelYoffset, boxYoffset, totWid;
    int spacewid;
    XFontStruct *ofont, *choicesFont;
	data_ptr oShadow, choicesShadow;
	AXWidget_togglebutton choiceW;
	int box_x, box_y, box_width, box_height, box_baseline;
    int barwid = 0;    /* width of graphic bar (Motif) */
    int barht = 0;     /* height of graphic bar (Motif) */
    int barspc = 0;    /* space between graphic bar and end */

    if ( !w->display )
        return;

	if (w->parent && (w->parent->type == W_DIALOG) &&
				   !((AXWidget_dialog)w->parent)->exposed)
		return;

	if (layered) {
		if (w->flags & WFLG_IN_EDIT) {
			if (dePaintWidget(w)) {
				return;
			}
		}
		else {
			if (topWidget(w) == (AXWidget)w) {
				paintCompositeWidget(w, exP, regen);
				return;
			}
		}
	}

    if (!WIN_OF(w))
        return;
    w->mapped = TRUE;
    x = w->xpos;
    y = w->ypos;
    
    if (w->font) {
        ofont = w->font;
		oShadow = w->textShadow;
    }
    else {
        ofont = AxDlgMainFont;
		oShadow = (data_ptr) NULL;
    }
    spacewid = AXTextWidth(ofont, " ", 1);

	if (w->num_buttons) {
		if ((w->buttons[0])->font) {
			choicesFont = (w->buttons[0])->font;
			choicesShadow = (w->buttons[0])->textShadow;
		}
		else {
			choicesFont = AxDlgMainFont;
			choicesShadow = (data_ptr) NULL;
		}
	}
	else {
		choicesFont = AxDlgMainFont;
		choicesShadow = (data_ptr) NULL;
	}
	w->height = wOptionHeight(ofont, choicesFont, oShadow, choicesShadow,
							  &labelYoffset, &boxYoffset);
	
/* paint the title */

    gc = w->disabled ? DisabledGC : ForeGC;
    if ( MBstrlen( w->name ) > 0 ) {
		AXWidget bw;
		wcolor_t *backColor;

		backColor = (wcolor_t *) NULL;
		if (w->disabled && (bw = backgroundWidget(w))) {
			backColor = bw->widgetColor;
		}
        xmDrawStringAndAccelerator(w, WIN_OF(w), gc, x,
								   y + wOptionFieldBaseline(ofont) +
								   labelYoffset,
								   w->name, MBstrlen(w->name), w->textShadow,
								   w->textColor, backColor,
								   w->accelerator_key);
		
        strWid = AXTextWidth(ofont, w->name, MBstrlen(w->name));
	}
    else
        strWid = 0;

    x += strWid + spacewid;

    w->pulldown_xpos = x;

    if ( w->scrolltext < 0 ) {
		    /* pixwidth of widest pulldown choice string is
			 * stored on member scrolltext; this width affects
			 * widget's box width
			 */
		w->scrolltext = widestOptionChoice(w);
	}

    wOptionWidth(w->name, w->font, w->scrolltext, &width,
				 &barwid, &barspc, &totWid);
    barht = 7;
    
	box_x = x;
	box_y = y + boxYoffset;
	box_width = width + barwid + barspc;
	box_height = wOptionFieldHeight(choicesFont, choicesShadow);
	box_baseline = wOptionFieldBaseline(choicesFont);
	
    /* draw box outline */
	AxDrawBevel(w, WIN_OF(w), box_x, box_y, box_width, box_height,
				2, BEVEL_OUT);
	
    if (wShadowIsOn(w->shadow)) {
		/* draw drop shadow */
        xmDrawRectShadow(w, box_x, box_y, box_width, box_height);
        xmDrawRectAntiShadow(w, box_x, box_y, box_width, box_height);
    }

    /* clear inside */
	if (w->widgetColor) {
		xmColorFillRectangle(w, WIN_OF(w), BackGC, box_x + 2, box_y + 2,
							 box_width - 4, box_height - 4);
	}
	else {
		XFillRectangle (Dpy, WIN_OF(w), BackGC, box_x+2, box_y+2,
						box_width-4, box_height-4);
	}
	
    /* draw graphic bar */
	AxDrawBevel(w, WIN_OF(w), box_x + width,
				box_y + (box_height - barht) / 2,
				barwid, barht, 2, BEVEL_OUT);

/* put current value into box */

    str = " ";
    for ( ix = 0; ix < w->num_buttons; ix++ ) {
        if ( w->buttons[ix] == ((AXWidget_togglebutton)(w->optionValue)) ) {
            str = w->buttons[ix]->name;
            w->value = ix;
            break;
        }
    }
    wid = AXTextWidth(choicesFont, str, strlen(str));
    x += width/2;
    x -= wid/2;
	
	if (w->num_buttons) {
		choiceW = w->buttons[0];
		xmDrawString(choiceW, WIN_OF(w), gc, x,
					 box_y + box_baseline, str, strlen(str),
					 choiceW->textShadow, choiceW->textColor,
					 w->widgetColor);
	}
	
    w->pulldown_width = width;
    w->width = totWid;
}

/*************************************************************************
function: PaintOptionPulldown - Place option menu over option box area
Mode:
    0 - OPTION_NEW - Initial display
    1 - OPTION_MOVE - Pointer motion
    2 - OPTION_DONE - Button up
*/
void
PaintOptionPulldown (w,mode,eventP)
register AXWidget_option w;
int mode;
XEvent *eventP;
{
    XSetWindowAttributes atts;
    int height = 0;
    int width, labelYoffset, boxYoffset;
    long att_mask;
    int ix;
    int real_x, real_y;
    XmRowColumnCallbackStruct data;
    int borderwidth;
    int choiceHeight, xadj, yadj, radioWid;
    char menubutton, *p;
    GC gc;
    GC topGC, bottomGC, fillGC;
    static int buttonDown=0;
    static int doPulldown = 0;
    static int old_value; /*initially w->value, changed only when doPulldown */
    int new_value;
    static int defaultIsShowing;
    static int option_default;
    int button_xpos;  /* xpos of OpenLook abbr button */
    int inEdit = FALSE;
    Window childWin;
    int spacewid, fontHeight, fontAscent;
    XFontStruct *ofont, *pfont;
	data_ptr oShadow, pShadow;
    

    /*
     *  When OPTION_NEW: w->optionValue is used to determine the
     *  active w->value, old_value is initialized to w->value.
     *
     *  If doPulldown: old_value and new_value keep track of new choices;
     *  w->value and w->optionValue are changed only when callback occurs 
     *  on OPTION_DONE.
     *
     *  If !doPulldown: w->value, w->optionValue, and defaultIsShowing are 
     *  changed to keep track of whether the pointer has selected the default 
     *  or not; old_value is unchanged.  If button-up occurs outside the 
     *  widget, w->value is reset to old_value and w->optionValue is 
     *  reset accordingly.
     */

    if ( !w->display )
            return;
    w->mapped = TRUE;

/* ignore activity on empty lists */

    if ( w->num_buttons == 0 ) {
        ResetActive();
        return;
    }

    if (w->flags & WFLG_IN_EDIT) {
            /* being edited by dialog editor; in here for
             * display only, mode OPTION_NEW */
        inEdit = TRUE;
    }

    if (mode == OPTION_NEW)
        buttonDown = AxButtonDown;

    doPulldown = 1;

	/* set height, ascent used by option pulldown text */
    if (w->num_buttons && (w->buttons[0])->font) {
		pfont = (w->buttons[0])->font;
        fontHeight = pfont->max_bounds.ascent +
            pfont->max_bounds.descent;
        fontAscent = pfont->max_bounds.ascent;
		pShadow = (w->buttons[0])->textShadow;
    }
    else {
		pfont = AxDlgMainFont;
        fontHeight = AxCharHeight;
        fontAscent = AxCharAbove;
		pShadow = (data_ptr) NULL;
    }
	
	/* set ofont, used for option label */
	if (w->font) {
		ofont = w->font;
		oShadow = w->textShadow;
	}
	else {
		ofont = AxDlgMainFont;
		oShadow = (data_ptr) NULL;
	}
    spacewid = AXTextWidth(ofont, " ", 1);

/* compute pulldown x position */
    w->pulldown_xpos = w->xpos;
    w->pulldown_xpos = w->pulldown_xpos + AXTextWidth(ofont, w->name,
													  MBstrlen(w->name));
    w->pulldown_xpos += spacewid;

/* compute height, 3above,3below,1sep */

    width = w->pulldown_width;
    height += w->num_buttons * (fontHeight + 7);
    if ( height == 0 )
        return;
    height += 7;    /* area below */
    w->pulldown_height = height;

/* Handle button up case */

    if ( mode == OPTION_DONE ) {
        if ( doPulldown &&
            (eventP->xbutton.y >= w->pulldown_ypos) &&
            (eventP->xbutton.y <= w->pulldown_ypos + height) &&
            (eventP->xbutton.x >= w->pulldown_xpos) &&
            (eventP->xbutton.x <= w->pulldown_xpos + width) ) {
            /* button-up inside pulldown */
            if (!inEdit) {
                if (old_value != -1) {
                    w->value = old_value;
                    w->optionValue = (AXWidget)w->buttons[w->value];
                }
                data.data = (char*)w->value;
                if ( w->callback )
                    AxCallback(w->callback,w,w->callbacktag,&data);
                w->update_widget = TRUE;
                WidgetNeedsUpdated = TRUE;
            }
        }
        else  {
            if (!inEdit) {
                /* button-up outside pulldown: restore old value */
                w->optionValue = (AXWidget)w->buttons[w->value];
            }
        }

        buttonDown = 0;
        if (!inEdit) {
            ResetActive();
            ThimExposuresMayBeComing = TRUE;
        }
        return;
    }

    if ( mode == OPTION_NEW) {
        /* determine which one is currently selected */
        old_value = -1;
        for ( ix = 0; ix < w->num_buttons; ix++ ) {
            if ( w->optionValue == (AXWidget)w->buttons[ix] ) {
                w->value = ix;
                old_value = w->value;
                break;
            }
        }
        if ( old_value == -1 ) {  /* box intially empty, now becomes default */
            old_value = option_default;
            w->value = option_default;
            w->optionValue = (AXWidget)w->buttons[option_default];
        }
    }
    
    if ( mode != OPTION_NEW)
        goto have_window;

    att_mask = CWOverrideRedirect | CWSaveUnder | CWBackPixel |
                CWEventMask | CWDontPropagate;
    atts.override_redirect = TRUE;
    atts.save_under = FALSE; /* SAVE_UNDER; */
    atts.background_pixel = BackPixel;
    atts.event_mask = atts.do_not_propagate_mask =
        (KeyPressMask | KeyReleaseMask | ButtonPressMask |
                ButtonReleaseMask | PointerMotionMask);

	wOptionHeight(ofont, pfont, oShadow, pShadow, &labelYoffset, &boxYoffset);
    w->pulldown_ypos = w->ypos + boxYoffset -
		(old_value * (fontHeight + 7) + 5);

            /* get root-relative values of the tentative position,
             * adjust to stay on-screen, and convert back to parent-relative */
    XTranslateCoordinates(Dpy,WIN_OF(w), RootWin,
                          w->pulldown_xpos,w->pulldown_ypos,
                          &real_x, &real_y, &childWin);
    AdjustPulldownOrigin( &real_x, &real_y, width, height, &xadj, &yadj);
    w->pulldown_ypos += yadj;
    w->pulldown_xpos += xadj;
    borderwidth = 0;
    if (inEdit) { /* dialog editor */
        w->window = AxCreateQueryPixmap(Dpy, w->parent->window,
                                        width, height, DpyDepth);
		if (w->widgetColor) {
			xmColorFillRectangle(w, w->window, BackGC, 0, 0, width, height);
		}
		else {
			XFillRectangle(Dpy, w->window, BackGC, 0, 0, width, height);
		}
    }
    else {
        w->window = AxCrtWindow(Dpy, RootWin, 
                                real_x, real_y,
                                width,height,borderwidth,CopyFromParent,
                                InputOutput,CopyFromParent,att_mask,&atts);
        XMapWindow(Dpy,w->window);
		if (w->widgetColor) {
			xmColorFillRectangle(w, w->window, BackGC, 0, 0, width, height);
		}
    }
    if (!inEdit) {
        MenuArmed = (AXWidget)w;
    }

 /* paint contents */

 have_window:
     if ( mode == OPTION_NEW ) {
         if ( doPulldown ) {
			 AxDrawBevel(w, w->window, 0, 0, width, height, 2, BEVEL_OUT);
			 AxDrawBevel(w, w->window, 5, 
						 old_value * (fontHeight + 7) + 5,
						 width - 10, fontHeight + 6, 2, BEVEL_OUT);
        }
    }
    else if ( doPulldown && mode == OPTION_MOVE ) {
        int bx,by;  /* button pos within the menu */

        bx = eventP->xmotion.x - w->pulldown_xpos;
        by = eventP->xmotion.y - w->pulldown_ypos;

        if (by < 0)
            new_value = -1;
        else
            new_value = (by - 5) / (fontHeight+7);
        if ( (bx <= 0) || (bx >= width) || (new_value >= w->num_buttons) )
            new_value = -1;

        if (new_value == old_value)
            return; /* motion in same value. */
        
        if ( (new_value != old_value) && (new_value != -1) ) {
            /* valid new selection has been made */
            if (old_value == -1) {
                /* currently active (default) selection is showing */
                old_value = w->value;
            }
			AxDrawBevel(w, w->window, 5,
						old_value * (fontHeight+7) + 5,
						width - 10, fontHeight + 6,
						2, BEVEL_NONE);
			AxDrawBevel(w, w->window, 5,
						new_value * (fontHeight+7) + 5,
						width - 10, fontHeight + 6,
						2, BEVEL_OUT);
        }
        if ( (new_value == -1) && (old_value != -1) ) {
            /* display currently active choice, shut off displayed choice */
			AxDrawBevel(w, w->window, 5,
						old_value * (fontHeight+7) + 5,
						width - 10, fontHeight + 6,
						2, BEVEL_NONE);
			AxDrawBevel(w, w->window, 5,
						w->value * (fontHeight+7) + 5,
						width - 10, fontHeight + 6,
						2, BEVEL_OUT);
        }
        if (new_value == -1)
            old_value = w->value;
        else
            old_value = new_value;
        return;
    }

    if ( mode == OPTION_NEW ) {

        if (doPulldown) {
            int y = 0;
            unsigned long saved_fg;
			AXWidget_togglebutton choiceW;

            for ( ix = 0; ix < w->num_buttons; ix++ ) {
				choiceW = w->buttons[ix];
                xmDrawString(choiceW, w->window, ForeGC,
                             7 + spacewid, y + fontAscent + 8,
                             choiceW->name,
                             MBstrlen(choiceW->name),
                             choiceW->textShadow, choiceW->textColor,
							 w->widgetColor);
                y += fontHeight + 7;
            }
        }
    }
}
