/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmpixel.c
   Applix ASC version: 160.8
   Date: Wed May 19 21:14:53 1999
*/
#include <stdio.h>
#include <errno.h>

#include <aos.h>
#include <thimble.h>
#include <basicwin.h>
#include <xmlib.h>
#include <xmpixel.h>
#include <getstr5.h>
#include <getstr4.h>
#include <elf_env.h>
#include <eldata.h>
#include <axtext.h>
#include <atm.h>
#include <axbitmap.h>
#include <afile.h>
#include <doctypes.h>
#include <colors.h>

static int PixMapCacheSize = 0;
static pixcache_t *PixMapCache;

static unsigned long cmapper_truecolor[256];	/* See FSRev3ImToImage */
static cmap_type local_cmap_truecolor;
static int got_cmapper_truecolor=FALSE;

typedef struct java_image_t {
    int type;                  /* data format:                   */
                               /*    0=1bit                      */
                               /*    1=8bit, single transparency */
                               /*    2=8bit, multi transparency  */
    int compression_type;      /* 0=uncompressed, for future use */
    struct {
        int n_colors;          /* number of colors in colormap   */
        unsigned char *r;      /* red values                     */
        unsigned char *g;      /* green values                   */
        unsigned char *b;      /* blue values                    */
        unsigned char *alpha;  /* opacity - 255=opaque           */
        int transparent_index; /* (-1 <= lut entry < size)       */
    } colormap;
    int scan_wid;              /* width of a scanline in bytes   */
    int wid;                   /* width in pixels                */
    int hyt;                   /* height in pixels               */
    unsigned char *data;       /* 1 byte per pixel, no alignment */
} java_image_t;

typedef struct jmap_cache_t {
    char *name;
    java_image_t *jmap;
} jmap_cache_t;

#if 0
#include <xmpixel.p>
#endif

static jmap_cache_t *jmap_cache = NULL;
static int jmap_cache_size = 0;

#define byte3(x)    (((x) >> 24) & 0xFF)
#define byte2(x)    (((x) >> 16) & 0xFF)
#define byte1(x)    (((x) >>  8) & 0xFF)
#define byte0(x)    ((x) & 0xFF)
#define lowbit(x) ((x) & (~(x) + 1))

pixcache_t *Pixmapper();
extern FILE *AFTaskfopen();

Pixmap UnknownPixmap;   /* 16x16 pixmap of unknown image */
extern char *DirSlash;
extern Display *Dpy;
extern short DpyWid, DpyHt, DpyDepth;
extern short THIMpid;
extern Visual *AxVisual;
extern Window RootWin;
extern Window AxTopWinNeverMapped;  /* created by AxGetDpyAndGCs */
extern AXWidget WindowList;
extern int UseWidgetColors;

#define MAX_IM_BYTES 100000  /* !! ALSO REDEFINED IN WP */
#define X_MILLS_TO_PIXELS(x) (((long)(x) * UWidPels)/UWidMils)
#define Y_MILLS_TO_PIXELS(y) (((long)(y) * UHtPels)/UHtMils)

#if AX_X11   /*  {  */
extern GC PixmapperForeGC;
extern GC PixmapperBackGC;
extern GC PixmapperRecessGC;

extern GC olClrSunButtonGC, olClrShadowButtonGC;
extern GC olClrForeButtonGC, olClrBackButtonGC, olClrRecessButtonGC;
extern GC olClrRecessPixGC, olClrInvertedRecessPixGC, olInvertedPixGC;
extern GC olClrSunMiniButtonGC, olClrShadowMiniButtonGC;
extern GC olClrForeMiniButtonGC, olClrBackMiniButtonGC, olClrRecessMiniButtonGC;
extern GC olBusyButtonGC;
#endif    /*  }  */

static unsigned char Bitmask[] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};



FILE *AxFindInGlom();
static void RepaintPixmap();
static int WidgetUsesPixmap();
char *XLT2();
extern char *AxTempFile();
extern GC axCreateGC();
void get_color_name();
static Pixmap createClipMask();
extern short UWidMils, UHtMils, UWidPels, UHtPels;

static int FSRev3ImToImage(/* bitmap_struct *bmap, FILE *fp, AFileInfo *info */);
static int FSRev1ImToImage(/* bitmap_struct *bmap, FILE *fp */);

#if 0
The following are dummies so that these XLT2_s will be found by the source
scan tool, and will therefore end up in the glossary and 'trLocal' file.
This is a list of all of the images which are translated.
XLT2("translated image file","dir-16x16.im" )
XLT2("translated image file","bar_graph.im ")
XLT2("translated image file","clicker.im")
XLT2("translated image file","de_tmpim.im")
XLT2("translated image file","el_bold.im")
XLT2("translated image file","el_uline.im")
XLT2("translated image file","fillNone16.im")
XLT2("translated image file","gp_backc.im")
XLT2("translated image file","gp_backci.im")
XLT2("translated image file","gp_forec.im")
XLT2("translated image file","gp_foreci.im")
XLT2("translated image file","histgrph.im")
XLT2("translated image file","line_graph.im")
XLT2("translated image file","pie_graph.im")
XLT2("translated image file","sqnot.im")
XLT2("translated image file","sqnotgry.im")
XLT2("translated image file","stratgrf.im")
XLT2("translated image file","tmpimage.im")
XLT2("translated image file","wpnofram.im ")
#endif


/*****************************************************************************
 *
 * ExtractShift_from_mask - TmScaleImage utility function
 *
 * Return the index needed to shift color bits to the mask position.
 */
ExtractShift_from_mask(mask)
int mask;
{
    register int i;
    for ( i=0; i<32; i++ )
    {
        if ( mask & (0x01<<i) )
            return(i);
    }
    return(0);
}

#if AXMS
void
normalize_rgb_to_555(int in_r, int in_g, int in_b,
                int *out_r, int *out_g, int *out_b)
{
    /* The maximum value is 5 bit, 0x1F.  Compute the R/G/B values as a */
    /* fraction of the max possible value of 255, and then use that     */
    /* fraction to scale the max possible 5-bit value of 0x1F.  Finally */
    /* shift the resultant integer into the default position in the     */
    /* 16 bit color word. - jon                                         */
    register int r_5, g_5, b_5;
    r_5 = ((double)in_r / 255.0) * (double)0x1F;
    g_5 = ((double)in_g / 255.0) * (double)0x1F;
    b_5 = ((double)in_b / 255.0) * (double)0x1F;

    *out_r = r_5 << 10;
    *out_g = g_5 << 5;
    *out_b = b_5;
}

#if 0
void
normalize_555_to_rgb(unsigned long pixel, int red_mask, int green_mask,
                    int blue_mask, int *out_r, int *out_g, int *out_b)
{
    int r16, g16, b16;
    int red_shift, green_shift, blue_shift;

    /* get the shift values.  These may not always be the same, have seen */
    /* the masks come back from Windows as 0xF800, 0x7E0, ox1F.  There    */
    /* are 6 bits alloczated for "green"!                           - jon */

    red_shift = ExtractShift_from_mask( red_mask );
    green_shift = ExtractShift_from_mask( green_mask );
    blue_shift = ExtractShift_from_mask( blue_mask );

    /* get the 16 bit r, g, and b */
    r16 = (pixel & red_mask) >> red_shift;
    g16 = (pixel & green_mask) >> green_shift;
    b16 = (pixel & blue_mask) >> blue_shift;

    *out_r = ((double)r16/(double)0x1F) * 255.0;
    *out_g = ((double)g16/(double)0x1F) * 255.0;
    *out_b = ((double)b16/(double)0x1F) * 255.0;
}
#endif

#endif


data_ptr ElfGetSysVar();

/***************************************************************************
function: file_exists
*/
static int
file_exists(pathname)
char *pathname;
{
    struct stat buf;

    if ( pathname == (char *)0 )
        return(FALSE);
    if ( stat(pathname, &buf) )
        return(FALSE);
    if ( ((buf.st_mode)&S_IFMT) == S_IFDIR )
        return(FALSE);
    return(TRUE);
}

/***************************************************************************
function: initPixmapCacheEntry 
*/
initPixmapCacheEntry(name, ix)
char *name;
int ix;
{
    int sz;
    PixMapCacheSize = ix + 1;
    sz = PixMapCacheSize * sizeof(pixcache_t);
    PixMapCache = (pixcache_t *)realloc(PixMapCache, sz);
    PixMapCache[ix].name = strcpy(malloc(strlen(name)+1), name);
    PixMapCache[ix].width = 0;
    PixMapCache[ix].height = 0;
    PixMapCache[ix].pix = UnknownPixmap;
    PixMapCache[ix].invert_pix = UnknownPixmap;
    PixMapCache[ix].clip_mask = 0;
}

/***************************************************************************
function: AxReadIMtoXImage - read the image from the file into the AX bmp
*/
static int
AxReadIMtoXImage(fp, bmap)
FILE *fp;
bitmap_struct *bmap;
{
    char header[ALINELENGTH+1];
    unsigned char magic[2];
    AFileInfo info;
    int retval;

    if ( !fp )
        retval = !AOK;
    else
    {
        /* evaluate the header and process the data */
        magic[0] = fgetc(fp); /* read the first character in the header */
        if ( magic[0] == '*' ) /* post rev3.2 file */
        {
            /* read the header */
            ungetc(magic[0], fp);
            fgets(header, ALINELENGTH, fp);
            retval = AFileInfoFromHeader(header, &info);
            if ( retval == AOK )
            {
                if ( info.docType != AX_DOC_TYPE_ASTERX_BITMAP )
                    retval = !AOK;
                else
                {
                    /* convert the image data into an image */
                    retval = FSRev3ImToImage(bmap, fp, &info);
                    bmap->version = info.version;
                }
            }
        }
        else /* pre rev3.2 */
        {
            /* read the magic "number" */
            ungetc(magic[0],fp);
            fread(magic,2,1,fp);

            /* convert the data */
            if ( (magic[0] == 'A') && (magic[1] == 'X') ) /* 3.0 magic number */
            {
                retval = FSRev2ImToImage(bmap, fp);
                bmap->version = 300;
            }
            else if ( !magic[0] && !magic[1] ) /* pre 3.0 magic number */
            {
                bmap->hasMask = FALSE;
                retval = FSRev1ImToImage(bmap, fp);
                bmap->version = 210;
            }
        }
    }
    return(retval);
}

/***************************************************************************
function: AxIMtoXImage 
*/
int 
AxIMtoXImage(pathname, name, bmap)
char *pathname;
char *name;
bitmap_struct *bmap;
{
    int inGlom, retval, jnk, err, read_only;
    FILE *fp;

    /* open the image source file (already open if it's in the glom file) */
    fp = AxFindInGlom(name, &jnk);
    if ( fp )
        retval = AxReadIMtoXImage(fp, bmap);
    else if ( file_exists(pathname) )
    {
        read_only = TRUE;
        err = AFTaskLockFile(pathname, bmap->pid, &read_only, FALSE);
        if ( err == AOK )
        {
            fp = AFTaskfopen(bmap->pid, pathname, READ_BINARY_STR);
            if ( fp )
            {
                retval = AxReadIMtoXImage(fp, bmap);
                AFfclose(pathname, read_only, fp);
            }
            else
                retval = !AOK;
            AFTaskUnlockFile(pathname, bmap->pid, read_only);
        }
    }
    return(retval);
}

/***************************************************************************
function: FSIMtoXImage 
!! THIS FUNCTION ALSO CALLED FROM WP
*/
FSIMtoXImage(pathname, name, bmap)
char *pathname;
char *name;
bitmap_struct *bmap;
{
    char glomName[STRPATH], imageName[STRPATH];
    int retval;

    sprintf(imageName, "%s.im", name);
    strcpy(glomName, XLT2("translated image file", imageName));
    retval = AxIMtoXImage(pathname, glomName, bmap);
    return(retval);
}

/*************************************************************************
function: AXBmpToCacheMap - Create X Pixmap / MS Bitmap from AX Bmap
*/
static int
AXBmpToCacheMap(bmap, name, ix)
bitmap_struct *bmap;
char *name;
int ix;
{
#if AX_X11
    XGCValues recessvals, forevals;
    Pixmap pixmap, inv_pixmap;
    unsigned long valuemask;
    XImage *image;

    /* create the pixmaps */
    image = &bmap->image;
    pixmap = XCreatePixmap(Dpy, RootWin,
                           image->width,
                           image->height,
                           DpyDepth);
    inv_pixmap = XCreatePixmap(Dpy, RootWin,
                           image->width, 
                           image->height,
                           DpyDepth);

    /* copy pixmaps onto the cache */
    PixMapCache[ix].pix = pixmap;
    PixMapCache[ix].invert_pix = inv_pixmap;

    /* render the pixmaps */
    if ( UseWidgetColors )
    {
        /* render the image onto the pixmap */
        XPutImage(Dpy, pixmap, PixmapperForeGC,
                  image,
                  0, 0, 0, 0,
                  image->width,
                  image->height);

        /* render the inverted pixmap (depressed buttons) */
        valuemask = (GCForeground | GCBackground);
        XGetGCValues(Dpy, PixmapperForeGC, valuemask, &forevals);
        XGetGCValues(Dpy, PixmapperRecessGC, valuemask, &recessvals);
        XSetForeground(Dpy, PixmapperRecessGC, forevals.foreground);
        XSetBackground(Dpy, PixmapperRecessGC, recessvals.foreground);
        XPutImage(Dpy, inv_pixmap, PixmapperRecessGC,
                  image,
                  0, 0, 0, 0,
                  image->width,
                  image->height);

        /* restore the real RecessGC */
        XSetForeground(Dpy, PixmapperRecessGC, recessvals.foreground);
        XSetBackground(Dpy, PixmapperRecessGC, recessvals.background);
    }
    else
    {
        /* render the image onto the pixmap */
        XPutImage(Dpy, pixmap, PixmapperForeGC,
                  image,
                  0, 0, 0, 0,
                  image->width,
                  image->height);

        /* render the inverted pixmap (depressed buttons) */
        XPutImage(Dpy, inv_pixmap, PixmapperBackGC,
                  image,
                  0, 0, 0, 0,
                  image->width,
                  image->height);
    }

    /* free up image data allocated in the FSIM functions */
    TaskFree(bmap->pid, image->data);
#else /* Windows/NT  */
    XImage *image;

    image = &bmap->image;
    if ( image->bitmap_handle )
        PixMapCache[ix].pix = image->bitmap_handle;

    /* free up image data allocated in the FSIM functions */
    TaskFree(bmap->pid, image->data);
    TaskFree(bmap->pid, image->bitmap_info);
    if ( image->palette_handle )
        MS_DeleteObject(image->palette_handle);
#endif

    /* set image size */
    PixMapCache[ix].width  = image->width;
    PixMapCache[ix].height = image->height;
    return(AOK);
}

/*************************************************************************
function: AXBmpToCacheMask - Create clipmask from AX Bmp
*/
static int
AXBmpToCacheMask(bmap, name, ix)
bitmap_struct *bmap;
char *name;
int ix;
{
#if AX_X11
    unsigned long valuemask;
    XGCValues values;
    XImage *image;
    Pixmap mask;
    GC gc;

    image = &bmap->mask;
    mask = XCreatePixmap(Dpy, RootWin, image->width, image->height, 1);
    valuemask = GCForeground | GCBackground;
    values.foreground = 1;
    values.background = 0;
    gc = axCreateGC(Dpy, mask, DefaultScreen(Dpy), 1, (XFontStruct *)0,
                    valuemask, &values);
    XPutImage(Dpy, mask, gc, image, 0,0, 0,0, image->width,image->height);
    XFreeGC(Dpy, gc);

    PixMapCache[ix].clip_mask = mask;
#else /* Windows/NT  */
    XImage *image;
    image = &bmap->mask;
    if ( !image->bitmap_handle )
    {
        image->bitmap_handle = MS_CreateBitmap(image->width, 
                                                  image->height,
                                                  1, 1, 
                                                  image->data);
    }

    PixMapCache[ix].clip_mask = image->bitmap_handle;
    TaskFree(bmap->pid, image->bitmap_info);
    if ( image->palette_handle )
        MS_DeleteObject(image->palette_handle);
#endif

    TaskFree(bmap->pid, image->data);
    return(AOK);
}

/*************************************************************************
function: FSIMtoPixmap - Create a Pixmapper cache entry from .im file
*/
FSIMtoPixmap(pathname, name, ix)
char *pathname, *name;
int ix;
{
    bitmap_struct bmap;
    int err;

    /* read the im file, convert to an AX Bmp */
    memset(&bmap, 0, sizeof(bitmap_struct));
    bmap.pid        = THIMpid;
    bmap.display    = TRUE;
    err = FSIMtoXImage(pathname, name, &bmap);
    if ( err == AOK )
    {
        /* init the cache */
        initPixmapCacheEntry(name, ix);

        /* copy the map onto the cache */
        err = AXBmpToCacheMap(&bmap, name, ix);

        /* copy the mask onto the cache */
        if ( err == AOK && bmap.hasMask )
            err = AXBmpToCacheMask(&bmap, name, ix);
    }
    return(err);
}

/*************************************************************************
function: AXtoXImage
*/
AXtoXImage(pathname, doctype, bmap)
char *pathname;
int doctype;
bitmap_struct *bmap;
{
    char tempname[STRPATH+1];
    int retval;
    FILE *fp;

    if ( file_exists(pathname) == FALSE )
        retval = !AOK;
    else
    {
        /* convert the file from .ag to .im */
        if ( !AxTempFile(NXLT("%%%%%%%%%%.im"), tempname) )
            retval = !AOK;
        else
        {
            if ( doctype == AX_DOC_TYPE_UNKNOWN )
                doctype = RecognizeFile(pathname);

            if ( doctype == AX_DOC_TYPE_UNKNOWN )
                retval = !AOK;
            else
            {
                retval = AxToIm(doctype, pathname, tempname);
                if ( retval == AOK )
                {
                    fp = fopen(tempname, "r");
                    if ( !fp )
                        retval = !AOK;
                    else
                    {
                        retval = AxReadIMtoXImage(fp, bmap);
                        fclose(fp);
                    }
                }
            }
            AxDeleteTempfile(tempname);
        }
    }
    return(retval);
}

/*************************************************************************
function: FSGEtoXImage
*/
FSGEtoXImage(pathname, bmap)
char *pathname;
bitmap_struct *bmap;
{
    return(AXtoXImage(pathname, AX_DOC_TYPE_ASTERX_GRAPHICS, bmap));
}

/*************************************************************************
function: FStoXImage
*/
FStoXImage(pathname, bmap)
char *pathname;
bitmap_struct *bmap;
{
    return(AXtoXImage(pathname, AX_DOC_TYPE_UNKNOWN, bmap));
}

/************************************************************************/
    /* cache pixmap derived from GE document */
addGEPixmapToCache(name, map, width, height, transparentPel)
char *name;
Pixmap map;
int width,height;
unsigned long transparentPel;
{
    int cacheIndex, foundEntry;

    /* find the correct entry in the cache */
    foundEntry = FALSE;
    for ( cacheIndex = 0; cacheIndex < PixMapCacheSize; cacheIndex++ )
    {
        if ( streq(PixMapCache[cacheIndex].name, name) )
        {
            foundEntry = TRUE;
            break;
        }
    }
    if ( foundEntry == FALSE )
        return(!AOK);

    /* update the entry */
    PixMapCache[cacheIndex].pix    = map;
    PixMapCache[cacheIndex].width  = width;
    PixMapCache[cacheIndex].height = height;
    PixMapCache[cacheIndex].clip_mask = createClipMask(map, width,
                height, transparentPel);
    return(AOK);
}

/*************************************************************************
function: FSGEtoPixmap - Create a Pixmapper cache entry from .ag file
*/
FSGEtoPixmap(pathname, name, ix)
char *pathname, *name;
int ix;
{
    data_ptr geP, fileP, insetP, wP, srcP, retData;
    int err;
    int x,y, wid,hyt, width, height;
    char *gePtr;
    Pixmap pmap;
    AXWidget w;

    /* sanity check - does the file exist */
    if ( Aaccess(pathname, 0) != 0 )
        return(!AOK);

    /* create the graphic */
    srcP = AxMakeArray(1);
    err = ElfCallByName(NXLT("create_graphic$"),&retData, 1, srcP);
    AxFreeData(srcP);
    if ( err != AOK )
        AxFreeData(retData);
    else
    {
        gePtr = (void *)AxGFXFromData(retData);
        geP = (data_ptr)AxMakeGFXData(gePtr);
        AxFreeData(retData);

        /* read the file */
        fileP = AxMakeStrData(-1, pathname);
        wP = AxMakeIntData(TRUE);
        srcP = AxMakeArray(3);
        AxAddDataToArray(srcP, 0, geP);
        AxAddDataToArray(srcP, 1, fileP);
        AxAddDataToArray(srcP, 2, wP);
        err = ElfCallByName(NXLT("read_graphic_file$"),&retData, 1, srcP);
        AxFreeData(srcP);
        if ( err != AOK )
        {
            AxFreeData(retData);
            return(err);
        }

        /* get the size of the graphic */
        geP = (data_ptr)AxMakeGFXData(gePtr);
        err = ElfCallByName(NXLT("gfx_get_bounder@"),&retData, 1, geP);
        if ( err != AOK )
        {
            AxFreeData(retData);
            return(err);
        }
        else
        {
            /* create a pixmap */
            x = grUnitsToDots(gePtr, AxFloatFromArray(retData, 0));
            y = grUnitsToDots(gePtr, AxFloatFromArray(retData, 1));
            wid = grUnitsToDots(gePtr, AxFloatFromArray(retData, 2));
            hyt = grUnitsToDots(gePtr, AxFloatFromArray(retData, 3));
            AxFreeData(retData);

            /* add the offset of origin so that image is always at 0,0 */
            wid += x;
            hyt += y;
            /* conv to pixels */
            width = X_MILLS_TO_PIXELS(wid);
            height = Y_MILLS_TO_PIXELS(hyt);
            pmap = AxCreateQueryPixmap(Dpy, RootWin, width, height, DpyDepth);

            /* create a temp widget to load the pixmap  */
            w = (AXWidget) calloc(1, sizeof(axw_any));
            w->type = W_CANVAS; /* bogus widget type */
            w->parent = (AXWidget_main)w;
            w->width = width;
            w->height = height;
            AxSetDrawableOfWidget(w, pmap);
            wP = AxMakeBinaryData(sizeof(AXWidget), &w);

            srcP = AxMakeArray(4);
            AxAddIntToArray(srcP, 0, 0);
            AxAddIntToArray(srcP, 1, 0);

            /* render the graphic onto the pixmap */
            insetP = AxMakeArray(12);
            AxAddDataToArray(insetP, 11, wP);
            AxAddDataToArray(insetP, 5, srcP);
            err = ElfCallByName(NXLT("gfx_render_graphic@"),&retData,
                                    2, geP, insetP);
            AxFreeData(insetP);
    
            /* add the pixmap to the cache */
            if ( err != AOK )
                AxFreePixmap(pmap);
            else
            {
                initPixmapCacheEntry(name, ix);
                cacheGEPixmap(w, name, width, height);
            }
            free(w);
        }

        /* Destroy grafix handle */
        srcP = AxMakeArray(1);
        AxAddDataToArray(srcP, 0, geP);
        ElfCallByName(NXLT("destroy_graphic$"), &retData, 1, srcP);
        AxFreeData(retData);
        AxFreeData(srcP);
    }

    /* done */
    return(err);
}

/*************************************************************************
function: Pixmapper
*/
pixcache_t *
Pixmapper (name)
char *name;
{
    char pathname[STRPATH+1], tempname[STRPATH+1], *path;
    int doctype, ix, c_ix, n, jnk;
    data_ptr searchP;

    /* search for the pixmap (by name) in the pixmap cache */
    for ( c_ix=0; c_ix<PixMapCacheSize; c_ix++ )
    {
        if ( streq(PixMapCache[c_ix].name, name) )
            return(&PixMapCache[c_ix]);
    }

    /* if there is a full pathname, check there */
    if ( name[0] == *DirSlash )
    {
        strcpy(pathname, name);
        AxSysDosUnixPath(pathname);
        if ( file_exists(pathname) )
        {
            /* an .im file */
            if ( FSIMtoPixmap(pathname, name, c_ix) == AOK )
                return(&PixMapCache[c_ix]);

            /* an .ag file */
            else if ( FSGEtoPixmap(pathname, name, c_ix) == AOK )
            {
                PixMapCache[c_ix].invert_pix = PixMapCache[c_ix].pix;
                return(&PixMapCache[c_ix]);
            }

            /* a foreign file format */
            else /* try to autoconvert, return NULL on failure */
            {
                /* recognize type */
                doctype = RecognizeFile(pathname);
                if ( doctype == AX_DOC_TYPE_UNKNOWN )
                    return((pixcache_t *)0);

                /* convert to .ag file */
                if ( !AxTempFile(NXLT("%%%%%%%%%%.ag"), tempname) )
                    return((pixcache_t *)0);
                else if ( AxToGr(doctype, pathname, tempname) != AOK )
                {
                    AxDeleteTempfile(tempname);
                    return((pixcache_t *)0);
                }

                /* convert to a pixmap */
                if ( FSGEtoPixmap(tempname, name, c_ix) != AOK )
                {
                    AxDeleteTempfile(tempname);
                    return((pixcache_t *)0);
                }
                else
                {
                    AxDeleteTempfile(tempname);
                    PixMapCache[c_ix].invert_pix = PixMapCache[c_ix].pix;
                    return(&PixMapCache[c_ix]);
                }
            }
        }
    }

    else /* not a full path - look in the usual places */
    {
        /* check the glom file */
        sprintf(pathname, "%s.im", name);
        strcpy(tempname, XLT2("translated image file", pathname));
        if ( AxFindInGlom(tempname, &jnk) )
        {
            if ( FSIMtoPixmap(name, name, c_ix) == AOK )
                return(&PixMapCache[c_ix]);
        }

        /* init the ELF search path */
        searchP = ElfGetSysVar(NXLT("elf_search_list@"));
        if ( !searchP )
        {
            ElfSetStandardSearchList();
            searchP = ElfGetSysVar(NXLT("elf_search_list@"));
            if ( !searchP )
                return((pixcache_t *)0);
        }

        /* check the ELF search path */
        n = AxArraySize(searchP);
        for ( ix=0; ix<n; ix++ )
        {
            path = (char *)AxStrFromArray(searchP, ix);
            if ( !path )
                continue;

            /* look for images first */
            sprintf(pathname, "%s/%s.im", path, name);
            AxSysDosUnixPath(pathname);
            if ( Aaccess(pathname, 0) == 0 )
            {
                if ( FSIMtoPixmap(pathname, name, c_ix) != AOK )
                    return((pixcache_t *)0);
                else
                    return(&PixMapCache[c_ix]);
            }

            /* then look for graphic files */
            sprintf(pathname, "%s/%s.ag", path, name);
            AxSysDosUnixPath(pathname);
            if ( Aaccess(pathname, 0) == 0 )
            {
                if ( FSGEtoPixmap(pathname, name, c_ix) != AOK )
                    return((pixcache_t *)0);
                else
                {
                    PixMapCache[c_ix].invert_pix = PixMapCache[c_ix].pix;
                    return(&PixMapCache[c_ix]);
                }
            }
        }
    }
    return((pixcache_t *)0);
}

/*************************************************************************
function: LookupPixmap - Add and/or Get pixmap from cache
Returns the Unknown Pixmap (16x16) if name not found.
Looks in user's macro dir, custom macro dir & finally system macro dir
*/
Pixmap
LookupPixmap(name)
char *name;
{
    pixcache_t *p;
    p = Pixmapper(name);
    if ( !p )
        return(UnknownPixmap);
    else
        return(p->pix);
}

/*************************************************************************
function: free_pcache_resources - never free "UnknownPixmap"
*/
static void
free_pcache_resources(ix)
int ix;
{
    Pixmap pmap1, pmap2;

    pmap1 = PixMapCache[ix].pix;
    if ( pmap1 && pmap1 != UnknownPixmap )
        XFreePixmap(Dpy, pmap1);

    pmap2 = PixMapCache[ix].invert_pix;
    if ( pmap2 && pmap1 != pmap2 && pmap2 != UnknownPixmap )
        XFreePixmap(Dpy, pmap2);

    pmap2 = PixMapCache[ix].clip_mask;
    if ( pmap2 && pmap1 != pmap2 && pmap2 != UnknownPixmap )
        XFreePixmap(Dpy, pmap2);

    free(PixMapCache[ix].name);
}

/*************************************************************************
function: deletePixmapFromCache - free named entry and all associated reources
*/
int
deletePixmapFromCache(name)
char *name;
{
    int ix, sz, stat;

    /* find the correct entry in the cache */
    for ( ix=0; ix<PixMapCacheSize; ix++ )
    {
        if (streq(PixMapCache[ix].name, name) )
            break;
    }
    if ( ix >= PixMapCacheSize )
        stat = !AOK;
    else
    {
        free_pcache_resources(ix);

        if ( ix < (PixMapCacheSize-1) )
        {
            sz = sizeof(pixcache_t) * (PixMapCacheSize-ix);
            memcpy((char *)&PixMapCache[ix], (char *)&PixMapCache[ix+1], sz);
        }
        PixMapCacheSize--;
        sz = sizeof(pixcache_t) * PixMapCacheSize;
        PixMapCache = (pixcache_t *)realloc(PixMapCache, sz);

        stat = AOK;
    }
    return(stat);
}

/*************************************************************************
function: deleteAllPixmapsFromCache - free ALL entries and associated reources
*/
void
deleteAllPixmapsFromCache()
{
    register int ix;
    if ( PixMapCacheSize > 0 )
    {
        if ( PixMapCache )
        {
            for ( ix=0; ix<PixMapCacheSize; ix++ )
                free_pcache_resources(ix);
            free(PixMapCache);
            PixMapCache = (pixcache_t *)0;
        }
        PixMapCacheSize = 0;
    }
}

/*************************************************************************
function: AXtoJAVA - convert AX bmp to AX java image
*/
static java_image_t *
AXtoJAVA(bmap,alignment)
bitmap_struct *bmap;
int alignment;       /* if false, byte aligned */
{
    int sz, bwid, wid, hyt, i, n, x, y;
    java_image_t *jmap;
    color_type *color;
    cmap_type *cmap;
    XImage *image;
    int c1,c2;

    /* sanity check */
    if ( !bmap )
        return((java_image_t *)0);

    /* create a blank java image */
    sz = sizeof(java_image_t);
    jmap = (java_image_t *)malloc(sz);
    memset((char *)jmap, 0, sz);
    jmap->compression_type = 0;
    cmap = &(bmap->cmap);

    image = &(bmap->image);
    if ( image->depth == 1 )
    {
        int ix;
        /* create a type 0 image (mono) */
        jmap->type = 0;
        jmap->colormap.n_colors = 2;
        
        /* load the colors */
        sz = 2 * sizeof(unsigned char);
        jmap->colormap.r = (unsigned char *)malloc(sz);
        jmap->colormap.g = (unsigned char *)malloc(sz);
        jmap->colormap.b = (unsigned char *)malloc(sz);
        jmap->colormap.alpha = (unsigned char *)malloc(sz);
        
        jmap->colormap.r[0] = 255;
        jmap->colormap.g[0] = 255;
        jmap->colormap.b[0] = 255;
        jmap->colormap.alpha[0] = 0;
        jmap->colormap.r[1] = 0;
        jmap->colormap.g[1] = 0;
        jmap->colormap.b[1] = 0;
        jmap->colormap.alpha[1] = 255;

        /* load the data */
        wid = jmap->wid = image->width;
        hyt = jmap->hyt = image->height;
        bwid = jmap->scan_wid = ((wid+15)/16) * 2;
        sz = bwid * hyt * sizeof(unsigned char);
        jmap->data = (unsigned char *)malloc(sz);
        if ( bmap->version > 210 ) {        /* I didn't write this */
            for ( ix = 0; ix < sz; ix++ ) {
                unsigned char old = image->data[ix];
                image->data[ix] = 
                    ((old >> 7) & 0x01) |
                    ((old >> 5) & 0x02) |
                    ((old >> 3) & 0x04) |
                    ((old >> 1) & 0x08) |
                    ((old << 7) & 0x80) |
                    ((old << 5) & 0x40) |
                    ((old << 3) & 0x20) |
                    ((old << 1) & 0x10);
            }
        }
        memcpy((char *)jmap->data, (char *)image->data, sz);

        /* set the transparent index */
        jmap->colormap.transparent_index = -1;
    }
    else /* assume 8 bit image */
    {
        /* create a type 2 image (color, multi transparency) */

        wid = jmap->wid = image->width;
        hyt = jmap->hyt = image->height;
        switch (alignment) {
        case 0: /* quad word alignment */
            bwid = jmap->scan_wid = ((wid+3)>>2)<<2;
            break;
        case 1: /* word-alignment */
            bwid = jmap->scan_wid = wid + wid%2;
            break;
        case 2: /* no alignment */
            bwid = jmap->scan_wid = wid;
            break;
        }
        sz = bwid * hyt * sizeof(unsigned char);
        /** Detect, if the image uses only 2 colors **/
        for(i=1, c1 = image->data[0]; i < sz && image->data[i] == c1; i++)
          /* empty */ ;
        if(i < sz) {
           for( c2=image->data[i]; i < sz && 
                  (image->data[i] == c1 || image->data[i] == c2); i++)
             /* empty */ ;
        }
        else  /* only 1 color used ?? */
          c2 = c1; 
        if(i == sz) { /* yes, the image uses only 2 colors */ 
            unsigned char val;
            int j,ind,w,ix,bx,fgcol_ind;
            int org_bwid, row_start_ind;
            org_bwid = bwid;
            fgcol_ind = c2; /* assume c2 as fg color and set bit for it */
            
            /* load the colors */
            sz = 2 * sizeof(unsigned char);
            jmap->colormap.r = (unsigned char *)malloc(sz);
            jmap->colormap.g = (unsigned char *)malloc(sz);
            jmap->colormap.b = (unsigned char *)malloc(sz);
            jmap->colormap.alpha = (unsigned char *)malloc(sz);
          
			c1 = c1 & 0xFF;
            color = &cmap->color[c1];
            jmap->colormap.r[0] = 255 - MIN(255, color->c+color->k);
            jmap->colormap.g[0] = 255 - MIN(255, color->m+color->k);
            jmap->colormap.b[0] = 255 - MIN(255, color->y+color->k);
            if ( color->see_thru )
                jmap->colormap.alpha[0] = 0;
            else
                jmap->colormap.alpha[0] = 255;

			c2 = c2 & 0xFF;
            color = &cmap->color[c2];
            jmap->colormap.r[1] = 255 - MIN(255, color->c+color->k);
            jmap->colormap.g[1] = 255 - MIN(255, color->m+color->k);
            jmap->colormap.b[1] = 255 - MIN(255, color->y+color->k);
            if ( color->see_thru )
                jmap->colormap.alpha[1] = 0;
            else
                jmap->colormap.alpha[1] = 255;
          
            bwid = jmap->scan_wid = (wid+15)/16*2;
            jmap->colormap.n_colors = 2;
            jmap->type = 0;
            sz = bwid * hyt * sizeof(unsigned char);
            jmap->data = (unsigned char *)malloc(sz);
            row_start_ind = 0;
            for( i=0,bx=0,ind=0; i < hyt; i++ ) {
                ix = 0;
                ind = row_start_ind;
                for(w = 0; w < bwid; w++) {
                    for( j=0, val=0; ix < wid && j < 8; j++,ix++)
                        if(image->data[ind++]==fgcol_ind) val |= Bitmask[7-j];
                    jmap->data[bx++] = val;
                }
                row_start_ind += org_bwid;
            }
            /* set the transparent index */
            jmap->colormap.transparent_index = -1;
        }
		else {
			n = cmap->n_colors;
			if ( n > 256 )
				n = 256;
			jmap->type = 2;
			jmap->colormap.n_colors = n;
			sz = n * sizeof(unsigned char);
			jmap->colormap.r = (unsigned char *)malloc(sz);
			memset((char *)jmap->colormap.r, 0, sz);
			jmap->colormap.g = (unsigned char *)malloc(sz);
			memset((char *)jmap->colormap.g, 0, sz);
			jmap->colormap.b = (unsigned char *)malloc(sz);
			memset((char *)jmap->colormap.b, 0, sz);
			jmap->colormap.alpha = (unsigned char *)malloc(sz);
			memset((char *)jmap->colormap.alpha, 0, sz);

			/* load the colors */
			for ( i=0; i<n; i++ )
			{
				color = &cmap->color[i];
				jmap->colormap.r[i] = 255 - MIN(255, color->c+color->k);
				jmap->colormap.g[i] = 255 - MIN(255, color->m+color->k);
				jmap->colormap.b[i] = 255 - MIN(255, color->y+color->k);
				if ( color->see_thru )
					jmap->colormap.alpha[i] = 0;
				else
					jmap->colormap.alpha[i] = 255;
			}

			/* load the data */
			sz = bwid * hyt * sizeof(unsigned char);
			jmap->data = (unsigned char *)malloc(sz);
			memcpy((char *)jmap->data, (char *)image->data, sz);

			/* set the transparent index */
			jmap->colormap.transparent_index = -1;
		}
    }
    return(jmap);
}

/*************************************************************************
function: AxJavaMapper - like pixmapper, with no cache
*/
java_image_t *
AxJavaMapper(name)
char *name;
{
    char pathname[STRPATH+1], tempname[STRPATH+1], *path;
    bitmap_struct bmap;
    java_image_t *jmap;
    data_ptr searchP;
    int ix, n, jnk;

    /* read the im file, convert to an AX Bmp */
    memset(&bmap, 0, sizeof(bitmap_struct));
    bmap.needcolor = TRUE;
    bmap.pid = THIMpid;
    jmap = (java_image_t *)0;

    /* if there is a full pathname, check there */
    if ( name[0] == *DirSlash )
    {
        strcpy(pathname, name);
        AxSysDosUnixPath(pathname);
        if ( file_exists(pathname) )
        {
            /* an .im file */
            if ( FSIMtoXImage(pathname, name, &bmap) == AOK )
                jmap = AXtoJAVA(&bmap,1); /* convert to a JAVAmap */

            /* some other format - try to convert */
            else if ( FStoXImage(pathname, &bmap) == AOK )
                jmap = AXtoJAVA(&bmap,1); /* convert to a JAVAmap */
        }
    }

    else /* not a full path - look in the usual places */
    {
        /* check the glom file */
        sprintf(pathname, "%s.im", name);
        strcpy(tempname, XLT2("translated image file", pathname));
        if ( AxFindInGlom(tempname, &jnk) )
        {
            if ( FSIMtoXImage(pathname, name, &bmap) == AOK )
                jmap = AXtoJAVA(&bmap,1);
        }

        else /* not in the glom file - check the ELF searchpath */
        {
            /* init the ELF search path */
            searchP = ElfGetSysVar(NXLT("elf_search_list@"));
            if ( !searchP )
            {
                ElfSetStandardSearchList();
                searchP = ElfGetSysVar(NXLT("elf_search_list@"));
                if ( !searchP )
                    return((java_image_t *)0);
            }

            /* check the ELF search path */
            n = AxArraySize(searchP);
            for ( ix=0; ix<n; ix++ )
            {
                path = (char *)AxStrFromArray(searchP, ix);
                if ( !path )
                    continue;

                /* look for images first */
                sprintf(pathname, "%s/%s.im", path, name);
                AxSysDosUnixPath(pathname);
                if ( Aaccess(pathname, 0) == 0 )
                {
                    if ( FSIMtoXImage(pathname, name, &bmap) == AOK )
                        jmap = AXtoJAVA(&bmap,1);
					break;
                }

                else /* then look for graphic files */
                {
                    sprintf(pathname, "%s/%s.ag", path, name);
                    AxSysDosUnixPath(pathname);
                    if ( Aaccess(pathname, 0) == 0 )
                    {
                        if ( FSGEtoXImage(pathname, &bmap) == AOK )
                            jmap = AXtoJAVA(&bmap,1);
						break;
                    }
                }
            }
        }
    }
    return(jmap);
}

/*************************************************************************
function: JAVAtoELF
*/
data_ptr
JAVAtoELF(jmap)
java_image_t *jmap;
{
    int packet_size, data_size, cmap_size, rgb_size, bx;
    data_ptr packP;
    char *bytes;
    int *data;

	/* sanity */
	if ( jmap == (java_image_t *)0 )
		return((data_ptr)0);

    /* create a packet for the colormap and data */
    rgb_size  = jmap->colormap.n_colors * sizeof(unsigned char);
    cmap_size = 2 + 4*rgb_size + 2;
    switch ( jmap->type )
    {
        case 0:
        case 2:
            cmap_size = 2 + 4 * rgb_size;
        break;
        case 1:
            cmap_size = 2 + 3 * rgb_size + 2;
        break;
    }
    data_size = jmap->scan_wid * jmap->hyt;
    packet_size = 2 + 2 + cmap_size + 2 + 2 + 2 + data_size;

    /* make the elf data */
    packP = AxMakeBinaryData(packet_size, (char *)0);
    bytes = AxBinaryBytesPtr(packP);
    bx = 0;

    /* 2 bytes for jmap type */
    bytes[bx++] = byte1(jmap->type);
    bytes[bx++] = byte0(jmap->type);

    /* 2 bytes for compression format */
    bytes[bx++] = byte1(jmap->compression_type);
    bytes[bx++] = byte0(jmap->compression_type);

    /* 2 bytes for number of colors in colormap */
    bytes[bx++] = byte1(jmap->colormap.n_colors);
    bytes[bx++] = byte0(jmap->colormap.n_colors);

    /* colormap */
    if ( rgb_size > 0 ) {
        /* 4 bytes * number of colors in colormap for rgb values */
        memcpy(bytes+bx, jmap->colormap.r, rgb_size);
        bx += rgb_size;
        memcpy(bytes+bx, jmap->colormap.g, rgb_size);
        bx += rgb_size;
        memcpy(bytes+bx, jmap->colormap.b, rgb_size);
        bx += rgb_size;
        if ( jmap->type != 1 ) {
            memcpy(bytes+bx, jmap->colormap.alpha, rgb_size);
            bx += rgb_size;
        }
        else {
            /* 2 bytes for transparent value in colormap */
            bytes[bx++] = byte1(jmap->colormap.transparent_index);
            bytes[bx++] = byte0(jmap->colormap.transparent_index);
        }
    }
        
    /* 2 bytes for scanline width */
    bytes[bx++] = byte1(jmap->scan_wid);
    bytes[bx++] = byte0(jmap->scan_wid);
        
    /* 2 bytes for width */
    bytes[bx++] = byte1(jmap->wid);
    bytes[bx++] = byte0(jmap->wid);

    /* 2 bytes for height */
    bytes[bx++] = byte1(jmap->hyt);
    bytes[bx++] = byte0(jmap->hyt);

    /* width * height bytes for jmap data */
    memcpy(bytes+bx, jmap->data, data_size);
    bx += data_size;

    /* done */
    return(packP);
}

/*************************************************************************
function: AxDestroyJAVA
*/
AxDestroyJAVA(jmap)
java_image_t *jmap;
{
    if ( jmap->colormap.r )
        free(jmap->colormap.r);
    if ( jmap->colormap.g )
        free(jmap->colormap.g);
    if ( jmap->colormap.b )
        free(jmap->colormap.b);
    if ( jmap->colormap.alpha )
        free(jmap->colormap.alpha);
    if ( jmap->data )
        free(jmap->data);
    free(jmap);
}

static int
get_java_rgb(jmap, i)
java_image_t *jmap;
int i;
{
	int r, g, b, rgb;

	r = jmap->colormap.r[i];
	g = jmap->colormap.g[i];
	b = jmap->colormap.b[i];
	rgb = 0xff000000 + 
		  ((r<<16)&0x00ff0000) +
		  ((g<<8)&0x0000ff00) +
		  (b&0x000000ff);
	return(rgb);
}

/*************************************************************************
function: jmapToElF
*/
data_ptr
jmapToElF(jmap, x, y)
java_image_t *jmap;
int x, y;
{
    int i, ix, sz, pel;
    data_ptr dataP, bP;
    unsigned char *colors, *mask;

	/* sanity */
	if ( jmap == (java_image_t *)0 )
		return((data_ptr)0);

    /* create a packet for the colormap and data */
	dataP = AxMakeArray(7);
	AxAddIntToArray(dataP, 0, x);
	AxAddIntToArray(dataP, 1, y);
	AxAddIntToArray(dataP, 2, jmap->scan_wid);
	AxAddIntToArray(dataP, 3, jmap->hyt);
	sz = ((jmap->wid+7)/8)*jmap->hyt;
	mask = (unsigned char *)TaskAlloc(0, sz);
	memset(mask, 0xFF, sz);
	AxAddBinaryToArray(dataP, 4, sz, (char *)mask);
	TaskFree(0, mask);
	AxAddBinaryToArray(dataP, 5, jmap->scan_wid*jmap->hyt, (char *)jmap->data);
	bP = AxMakeBinaryData(1024, (char *)0); /* 256*4*/
	colors = (unsigned char *)AxBinaryBytesPtr(bP);
	for ( i=ix=0; i < 256; i++ )
	{
		pel = get_java_rgb(jmap, i);
		colors[ix++] = byte3(pel);
		colors[ix++] = byte2(pel);
		colors[ix++] = byte1(pel);
		colors[ix++] = byte0(pel);
	}
	AxAddDataToArray(dataP, 6, bP);

    /* done */
    return(dataP);
}

/*************************************************************************
function: AxGetJavaImagePacket
*/
data_ptr
AxGetJavaImagePacket(iconname)
char *iconname;
{
    java_image_t *jmap;
    data_ptr packP;
    int ix;

    /* look in the cache */
    for (ix = 0; ix < jmap_cache_size; ++ix)
        if (streq(jmap_cache[ix].name,iconname))
            return JAVAtoELF(jmap_cache[ix].jmap);

    /* read the image file */
    jmap = (java_image_t *)AxJavaMapper(iconname);
	ix = jmap_cache_size++;
	jmap_cache = (jmap_cache_t *)realloc(jmap_cache,
								jmap_cache_size*sizeof(jmap_cache_t));
	jmap_cache[ix].name = strcpy(malloc(strlen(iconname)+1), iconname);
	jmap_cache[ix].jmap = jmap;     /* update with new content */
	packP = JAVAtoELF(jmap);
    return(packP);
}

/************************************************************************/
static Pixmap
createClipMask(deepMap, mapwid, mapht, transparentPel)
Pixmap deepMap;
int mapwid, mapht;
unsigned long transparentPel;
{
    XImage *deepImage, *onePlaneImage;
    Pixmap maskMap;
    int ix, iy, paddedWid, size;
    unsigned long mask, tpix;
    XGCValues gcvals;
    GC tgc = NULL;              /* initialize! */
    unsigned char *imdata;
    int flip_dex;

    if (!deepMap)
        return ((Pixmap) NULL);

        /* take the image of the deepMap */
    deepImage = XGetImage(Dpy, deepMap, 0, 0, (unsigned int) mapwid,
                    (unsigned int) mapht, AllPlanes, ZPixmap);
    if (!deepImage)
        return ((Pixmap) NULL);
#if AXMS
    /* for now, this function is only called when creating a clip_mask for   *
     * the pixmap cache, so deepImage->bitmap_handle will contain the pixmap *
     * from the cache. null it out so we don't delete it later.  -donato     */
    deepImage->bitmap_handle = NULL;
#endif

        /* create a 1-deep image for the mask */
    paddedWid = 0;
    while( (paddedWid += deepImage->bitmap_pad) < mapwid) ; 
    size = (paddedWid / 8) * mapht;
    imdata = (unsigned char *) calloc (1,size);
    if (!imdata) {
        XDestroyImage(deepImage);
        return ((Pixmap) NULL);
    }

    onePlaneImage = XCreateImage(Dpy, AxVisual,
                (unsigned int) 1, ZPixmap, 0, imdata, (unsigned int) mapwid,
                (unsigned int) mapht, deepImage->bitmap_pad, 0);
    if (!onePlaneImage) {
        free((unsigned char *) imdata);
        XDestroyImage(deepImage);
        return ((Pixmap) NULL);
    }

        /* index thru deep image, and set bits on one plane image for
         * all nontransparent pixels */
#if AXMS
    /* on Windows, the bits in deepImage are flipped (they were obtained *
     * with MS_GetDIBits), so we need to write out the one-plane image   *
     * correctly by flipping it.           -donato                       */
	/* since XGetPixel will return the DIB color value, we need to make	
	 * sure we have the DIB value, not the RGB value.
	 */
	transparentPel = axmsRetrieveDIBValue(transparentPel);
    for ( iy=0, flip_dex=(mapht-1); iy < mapht; iy++, flip_dex-- )
        {
        for ( ix = 0; ix < mapwid; ix++ )
            {
            tpix = XGetPixel(deepImage, ix, iy);
            if ( tpix != transparentPel )
                XPutPixel(onePlaneImage, ix, flip_dex, AllPlanes);
            }
        }
#else
    for (ix = 0; ix < mapwid; ix++) {
        for (iy = 0; iy < mapht; iy++) {
            tpix = XGetPixel(deepImage, ix, iy);
            if (tpix != transparentPel) {
                XPutPixel(onePlaneImage, ix, iy, AllPlanes);
            }
        }
    }
#endif
        /* create a one-deep pixmap for use as the clipmask */
    maskMap = AxCreateQueryPixmap(Dpy, AxTopWinNeverMapped, mapwid, mapht, 1);
    if (maskMap) {
            /* shoot one-plane image onto maskMap */
                /* need a 1-plane gc to do this.. */
        mask = GCForeground | GCBackground;
        gcvals.foreground = AllPlanes;
        gcvals.background = ~AllPlanes;
        tgc = axCreateGC(Dpy, maskMap, DefaultScreen(Dpy), 1,
                         (XFontStruct *) 0, mask, &gcvals);
        XPutImage(Dpy, maskMap, tgc, onePlaneImage, 0, 0, 0, 0,
                    (unsigned int) mapwid, (unsigned int ) mapht);
    }

        /* cleanup */
    XFreeGC(Dpy, tgc);
    free((unsigned char *) imdata);
    onePlaneImage->data = (char *) NULL;
    XDestroyImage(onePlaneImage);
    XDestroyImage(deepImage);
    return(maskMap);
}

/***********************************************************************
function: ElfbListImages - Return a list of all Glom file image names.
*/
void
ElfbListImages()
{
    register long *image_list = (long *) NULL;
    register char *p;
    register int len;
    register int ix;
    data_ptr retP = (data_ptr) NULL;
    int size;

    image_list = (long *)AxCreateGlomList(); /* context.c */

/* stuff image list into an elf array */

    size = IntsArraySize(image_list);
    if (!size) {
        retP = ELF_CREATE_ARRAY(0);
        goto finish;
    }

    retP = ELF_CREATE_ARRAY(size);
    for (ix = 0; ix < size; ix++) {
        p = (char *) ReadIntsArray(image_list, ix);
        if (!p) {
            retP = ElfAddStrToArray(retP, ix, "");
            continue;
        }

        /* strip out trailing ".im" from the image name */
        len = strlen(p);
        if (len > 3) {
            if ((p[len - 1] == 'm') && (p[len - 2] == 'i') &&
                (p[len - 3] == '.'))
                p[len - 3] = '\0';
        } /* else, just add the full name from the image list */

        retP = ElfAddStrToArray(retP, ix, p);
    }

finish:
    AxFreeGlomList(image_list);
    ElfRetData = retP;
    return;
}

static void 
AxLoadJavaBitmap (name, imDataP)
char *name;
data_ptr imDataP;
{
    int size, err;
    unsigned char *buf;
    FILE *fp;
    java_image_t *jmap;
    int ix;
    char path[STRPATH];

    /* Create pixmap from imDataP */
    /* get the size of image */
    size = AxBinaryFromDataPtr(imDataP, &buf);

    /* create a temp file */
    if (AxTempFile(NXLT("%%%%%%%%%%"), path)) {

        /* write the buffer to file */
        fp = fopen(path, WRITE_TRUNC_STR);
        if (!fp)
            ElfAbort(ERR_ILLVAL);
        if (fwrite(buf, size, 1, fp) != 1)
            ElfAbort(ERR_ILLVAL);
        err = fclose(fp);
        if (err != AOK)
            ElfAbort(err);

        jmap = AxJavaMapper(path);      /* convert to jmap format */
        AxDeleteTempfile(path);         /* done with temp file */

        if (NULL != jmap) {
            /* look if already cached, if yes replace old content with new */
            for (ix = 0; ix < jmap_cache_size; ++ix) {
                if (streq(name, jmap_cache[ix].name))
                    break;
            }
            if (ix == jmap_cache_size) {    /* not in the cache */
                jmap_cache_size++;
                jmap_cache = (jmap_cache_t *)realloc(jmap_cache,
                                            jmap_cache_size*sizeof(jmap_cache_t));
                jmap_cache[ix].name = strcpy(malloc(strlen(name)+1), name);
            }
            else {
                AxDestroyJAVA(jmap_cache[ix].jmap);     /* destroy old one */
            }
            jmap_cache[ix].jmap = jmap;     /* update with new content */
        }
    }
}


/*************************************************************************
function: ElfbLoadBitmap - Load a named image
For now, creates a pixmap.
*/
void ElfbLoadBitmap (name, imDataP)
char *name;
data_ptr imDataP;
{
    unsigned char *buf;
    int size, err;
    char path[STRPATH], tempname[STRPATH];
    int docType, c_ix;
    FILE *fp;

    /* if task is Java based, cache it in jmap_cache */
    if (TaskInfo[AxInTask].TaskUsesJavaDisplay) {
        AxLoadJavaBitmap(name,imDataP);
        return;
    }

    /* Create pixmap from imDataP */
    /* get the size of image */
    size = AxBinaryFromDataPtr(imDataP, &buf);

    /* create a temp file */
    if ( AxTempFile(NXLT("%%%%%%%%%%"), path) )
    {
        /* write the buffer to file */
        fp = fopen(path, WRITE_TRUNC_STR);
        if (!fp )
            ElfAbort(ERR_ILLVAL);
        if ( fwrite(buf, size, 1, fp) != 1 )
            ElfAbort(ERR_ILLVAL);
        err = fclose(fp);
        if ( err != AOK )
            ElfAbort(err);
    
        c_ix = PixMapCacheSize;
        /* convert the image to a pixmap */
        /* an .im file */
        if ( FSIMtoPixmap(path, name, c_ix) == AOK )
        {
            AxDeleteTempfile(path);
        }
        /* an .ag file */
        else if ( FSGEtoPixmap(path, name, c_ix) == AOK )
        {
            PixMapCache[c_ix].invert_pix = PixMapCache[c_ix].pix;
            AxDeleteTempfile(path);
        }
        /* a foreign file format */
        else /* try to autoconvert, return NULL on failure */
        {
            /* recognize type */
            docType = RecognizeFile(path);
            if ( docType == AX_DOC_TYPE_UNKNOWN )
                ElfAbort(!AOK);

            /* convert to .ag file */
            if ( AxTempFile(NXLT("%%%%%%%%%%.ag"), tempname) )
            {
                if ( AxToGr(docType, path, tempname) == AOK )
                {
                    /* convert to a pixmap */
                    if ( FSGEtoPixmap(tempname, name, c_ix) == AOK )
                    {
                        PixMapCache[c_ix].invert_pix = PixMapCache[c_ix].pix;
                    }
                }
            }
            AxDeleteTempfile(tempname);
            AxDeleteTempfile(path);
        }
    }
}

/*************************************************************************
function: WidgetUsesPixmap - Return TRUE if pixmap "name" used by widget
*/
static int 
WidgetUsesPixmap(w, name)
AXWidget w;
char *name;
{
    if ( w->type == W_BUTTON ) {
        if ( (((AXWidget_pushbutton)w)->pixmapLabel && streq(name, w->name)) ||
             (((AXWidget_pushbutton)w)->alternatePixmap && 
                    streq(name, ((AXWidget_pushbutton)w)->alternatePixmap)) ) {
            return(TRUE);
        }
    }
    else if ( w->type == W_ICON ) {
        if ( streq(name, w->name) ) {
            return(TRUE);
        }
    }
    else if (w->type == W_ROWCOL) {
        return(TRUE);
    }
    return(FALSE);
}

/*************************************************************************
function: RepaintPixmap - Redraw all widgets which use pixmap "name"
*/
static void 
RepaintPixmap(name)
char *name;
{
    AXWidget w, childP;
 
    if ( !name || !name[0] )
        return;
 
    for ( w = WindowList; w; w = w->nextWin ) {
        for ( childP = w->widgets; childP; childP = childP->nextChild ) {
            if ( WidgetUsesPixmap(childP, name) ) {
                if ( childP->painter )
                    (*childP->painter)(childP, NULL, TRUE, TRUE);
            }
        }
    }
}

/*************************************************************************
function: updateJavaMapCache
*/
static void
updateJavaMapCache(name)
char *name;
{
    int ix, foundJavamap;
    java_image_t *jmap;
	data_ptr strP;

    /* search for the Javamap (by name) in the Java cache */
    foundJavamap = FALSE;
    for ( ix = 0; ix < jmap_cache_size; ix++ )
    {
        if ( foundJavamap == FALSE )
        {
            if ( streq(jmap_cache[ix].name, name) )
            {
                if ( jmap_cache[ix].jmap != (java_image_t *)0 )
					AxDestroyJAVA(jmap_cache[ix].jmap);  /* destroy old one */
                foundJavamap = TRUE;
				break;
            }
        }
    }

    /* update display and cache size */
    if ( foundJavamap )
    {
		/* read the image file */
		jmap_cache[ix].jmap = (java_image_t *)AxJavaMapper(name);
		strP = AxTaskMakeStrData(0, -1, name);
		ElfCallByName("POST_JAVA_IMAGE$", NULL, 1, strP);
    }
}

/*************************************************************************
function: updatePixMapCache
*/
static void 
updatePixMapCache(name)
char *name;
{
    int ix, foundPixmap, newCacheSize;
    /* search for the pixmap (by name) in the pixmap cache */
    foundPixmap = FALSE;
    for ( ix = 0; ix < PixMapCacheSize; ix++ )
    {
        if ( foundPixmap == FALSE )
        {
            if ( streq(PixMapCache[ix].name, name) )
            {
                /* NOTE: `UnknownPixmap' is a permanent pixmap used under
                   default conditions, used as a window border pixmap in
                   `AxCrtWindow ()'... -- Do not free the resource. */
                if ((PixMapCache[ix].pix == UnknownPixmap) ||
                    (PixMapCache[ix].invert_pix == UnknownPixmap))
                    return; /* The GE is in the process of putting a
                               real pixmap in this slot. -- just
                               return. */

                if ( PixMapCache[ix].invert_pix != PixMapCache[ix].pix )
                    XFreePixmap(Dpy, PixMapCache[ix].invert_pix);
                XFreePixmap(Dpy, PixMapCache[ix].pix);
                foundPixmap = TRUE;
            }
        }
        else
            PixMapCache[ix-1] = PixMapCache[ix];
    }

    /* update display and cache size */
    if ( foundPixmap )
    {
        /* compress the cache */
        newCacheSize = --PixMapCacheSize * sizeof(pixcache_t);
        PixMapCache = (pixcache_t *)realloc(PixMapCache, newCacheSize);

        /* redraw all instances of this pixmap (will re-create pixmap) */
        RepaintPixmap(name);
    }
}

/*************************************************************************
function: AxUpdatePixCache
*/
void AxUpdatePixCache(name)
char *name;
{
    if (TaskInfo[AxInTask].TaskUsesJavaDisplay)
		updateJavaMapCache(name);
	else
		updatePixMapCache(name);
}

#if AX_X11   /*  {  */

/*************************************************************************
function: FSRev2ImToXImage - read a new style (magic number AX) .im file
*/
int
FSRev2ImToXImage(bmap, fp)
bitmap_struct *bmap;
FILE *fp;
{
    unsigned char ch, localMap[768], *lmap;
    unsigned long cmapper[256];
    unsigned char *mask_dat, *data_dat;
    unsigned char *temp_data, *inDataP,*outDataP;
    unsigned char *clr,*clrP,*def_name;
    unsigned char *imgData, *cDataP, *mdataP, *maskP;
    int width, height, depth, ix, rm, gm, bm, rs, gs, bs, x, y;
    int ii,jj,len,size,nColors,byteWid,monoByteWid, data_size, mask_size;
    unsigned int r, g, b,c,m,k;
    unsigned long cindex;
    color_type *tmp_color;
    int pid,line_ix,mask_ix;
    XImage *xP, *imageP, *mask;

    /* get the width */
    width = fgetc(fp);
    ch = fgetc(fp);
    width = (width << 8) | ch;

    /* get the height */
    height = fgetc(fp);
    ch = fgetc(fp);
    height = (height << 8) | ch;

    /* get the depth */
    depth = fgetc(fp);
    ch = fgetc(fp);
    depth = (depth << 8) | ch;

    pid = bmap->pid;
    /* get the mask flag */
    bmap->hasMask = fgetc(fp);
    if ( (bmap->hasMask) && (depth == 1) )
        bmap->hasMask = FALSE;

    /* get the number of colors */
    nColors = fgetc(fp);
    ch = fgetc(fp);
    nColors = (nColors << 8) | ch;

    /* if it's a color image, get the colormap */
    if ( depth == 8 )
        fread(localMap,nColors,3,fp);

    /* mono image */
    if ( depth == 1 )
    {
        /* get the image size, allocate a buffer, and read the image data */
        byteWid = (width + 7) / 8;
        data_size = byteWid*height;
        temp_data = (unsigned char *)TaskAlloc(pid,data_size);
        if ( fread(temp_data, 1, data_size, fp) != data_size )
        {
            if ( temp_data )
                TaskFree(pid,data_dat);
            return(!AOK);
        }
        monoByteWid = (width + 7)/8 + ((width + 7)/8)%2;
        data_size = monoByteWid*height;
        data_dat = (unsigned char *)TaskAlloc(pid,data_size);
        inDataP = temp_data;
        outDataP = data_dat;
        for (ix=0; ix<height; ix++)
        {
            memcpy(outDataP,inDataP,byteWid);
            inDataP += byteWid;
            outDataP += monoByteWid;
        }
        TaskFree(pid,temp_data);

        /* fake an XImage */
        imageP = &bmap->image;
        memset(imageP, 0, sizeof(*imageP));
        imageP->width = width;
        imageP->height = height;
        imageP->depth = depth;
        imageP->bits_per_pixel = depth;
        imageP->xoffset = 0;
        imageP->bitmap_pad = imageP->bitmap_unit = 16;
        imageP->bytes_per_line = monoByteWid;
        imageP->byte_order = MSBFirst;
        imageP->bitmap_bit_order = MSBFirst;
        imageP->obdata = NULL;
        imageP->format = XYBitmap;
        imageP->data = (char *)data_dat;
    }
    /* color image */
    else /* assume depth == 8 */
    {
        imageP = &bmap->image;
        memset(imageP, 0, sizeof(*imageP));
        mask = &bmap->mask;
        memset(mask, 0, sizeof(*mask));
        /* get the image size, allocate a buffer, and read the image data */
        byteWid = width;
        data_size = byteWid*height;
        temp_data = (unsigned char *)TaskAlloc(pid,data_size);
        if ( fread(temp_data, 1, data_size, fp) != data_size )
        {
            if ( temp_data )
                TaskFree(pid,temp_data);
            return(!AOK);
        }
        byteWid = width + width%2;
        data_size = byteWid*height;
        data_dat = (unsigned char *)TaskAlloc(pid,data_size);
        inDataP = temp_data;
        outDataP = data_dat;
        for (ix=0; ix<height; ix++)
        {
            memcpy(outDataP,inDataP,width);
            inDataP += width;
            outDataP += byteWid;
        }
        TaskFree(pid,temp_data);

        /* get the mask size, allocate a buffer, and read the mask data */
        if ( bmap->hasMask )
        {
            byteWid = (width + 7) /8;
            mask_size = byteWid*height;
            temp_data = (unsigned char *)TaskAlloc(pid,mask_size);
            if ( fread(temp_data, 1, mask_size, fp) != mask_size )
            {
                if ( data_dat )
                    TaskFree(pid,data_dat);
                if ( temp_data )
                    TaskFree(pid,temp_data);
                return(!AOK);
            }
            monoByteWid = (width + 7)/8 + ((width + 7)/8)%2;
            mask_size = monoByteWid*height;
            mask_dat = (unsigned char *)TaskAlloc(pid,mask_size);
            inDataP = temp_data;
            outDataP = mask_dat;
            for (ix=0; ix<height; ix++)
            {
                memcpy(outDataP,inDataP,byteWid);
                inDataP += byteWid;
                outDataP += monoByteWid;
            }
            TaskFree(pid,temp_data);
        }

        /* create an 8 bit XImage and a 1 bit mask to return to the caller */
        if ( bmap->display == FALSE )
        {
            /* copy the colormap for external use */
            byteWid = width + width%2;
            if ( bmap->needcolor )
            {
                lmap = localMap;
                bmap->cmap.n_colors = 256;
                size = 256 *sizeof(color_type);
                bmap->cmap.color = (color_type *)TaskAlloc(pid,size);
                def_name = (unsigned char *)XLT("Color");
                len = strlen((char *)def_name);
                clr = (unsigned char *)malloc((len+3) * sizeof(char));
                strcpy((char *)clr, (char *)def_name);
                clrP = &clr[len];
                for (ii=0; ii<nColors && ii<256; ii++)
                {
                    r = lmap[0];
                    g = lmap[1];
                    b = lmap[2];
                    lmap += 3;
                    sprintf((char *)clrP,"%3d",ii);
                    rgb2cmyk(r,g,b, &c,&m,&y,&k, FALSE);
                    bmap->cmap.color[ii].see_thru = FALSE;
                    bmap->cmap.color[ii].ink_type = 0;
                    bmap->cmap.color[ii].c = c;
                    bmap->cmap.color[ii].m = m;
                    bmap->cmap.color[ii].y = y;
                    bmap->cmap.color[ii].k = k;
                    bmap->cmap.color[ii].name = (char *)TaskAlloc(pid,len+3);
                    strcpy(bmap->cmap.color[ii].name, (char *)clr);
                }
                for (; ii<256; ii++ )
                {
                    sprintf((char *)clrP,"%3d",ii);
                    bmap->cmap.color[ii].see_thru = FALSE;
                    bmap->cmap.color[ii].ink_type = 0;
                    bmap->cmap.color[ii].c = 0;
                    bmap->cmap.color[ii].m = 0;
                    bmap->cmap.color[ii].y = 0;
                    bmap->cmap.color[ii].k = 0;
                    bmap->cmap.color[ii].name = (char *)TaskAlloc(pid,len+3);
                    strcpy(bmap->cmap.color[ii].name, (char *)clr);
                }
            }
            /* fake an XImage to transfer the image data */
            imageP->width = width;
            imageP->height = height;
            imageP->depth = depth;
            imageP->bits_per_pixel = depth;
            imageP->xoffset = 0;
            imageP->bitmap_pad = imageP->bitmap_unit = 16;
            imageP->bytes_per_line = byteWid;
            imageP->byte_order = MSBFirst;
            imageP->bitmap_bit_order = MSBFirst;
            imageP->obdata = NULL;
            imageP->format = ZPixmap;
            imageP->data = (char *)data_dat;

            /* fake an image for the mask, too (if there is one) */
            if ( bmap->hasMask )
            {
                *mask = *imageP;
                mask->depth = mask->bits_per_pixel = 1;
                mask->bytes_per_line = monoByteWid;
                mask->format = XYBitmap;
                mask->data = (char *)mask_dat;

                inDataP = mask_dat;
                outDataP = data_dat;
                monoByteWid = (width + 7)/8 + ((width + 7)/8)%2;
                /* check to see if there is a transparent in the colormap */
                for (y=0; y<height; y++)
                {
                    line_ix = y*byteWid;
                    mask_ix = y*monoByteWid;
                    for (ix=0; ix<width; ix++)
                    {
                        cindex = outDataP[line_ix+ix];
                        if ( !(inDataP[mask_ix+(ix>>3)] & Bitmask[ix&7]) )
                            bmap->cmap.color[cindex].see_thru = TRUE;
                    }
                }
            }
        }
        /* need a real image and mask, formatted for display */
        else
        {
            byteWid = width + width%2;
            /* the mask is the same on all displays */
            if ( bmap->hasMask )
            {
                mask->width = width;
                mask->height = height;
                mask->depth = 1;
                mask->bits_per_pixel = 1;
                mask->xoffset = 0;
                mask->bitmap_pad = imageP->bitmap_unit = 16;
                mask->bytes_per_line = monoByteWid;
                mask->byte_order = MSBFirst;
                mask->bitmap_bit_order = MSBFirst;
                mask->obdata = NULL;
                mask->format = XYBitmap;
                mask->data = (char *)mask_dat;
            }
            /* create a TrueColor/DirectColor DpyDepth image */
            if ( DpyDepth > 8 )
            {
                /* create and copy the image */
                byteWid = 4*width;
                imageP->data = (char *)TaskAlloc(pid,height * byteWid);
                xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                            DpyDepth, ZPixmap,
                                            0, imageP->data,
                                            width, height,
                                            32, byteWid);
                *imageP = *xP;
                xP->width = xP->height = 0;
                xP->data = (char *)0;
                XDestroyImage(xP);

                /* create an "pixel to index" table */
                rm = imageP->red_mask;
                gm = imageP->green_mask;
                bm = imageP->blue_mask;
        if ( 0 == rm && 0 == gm && 0 == bm ) {
          /* cheat, and read masks from display directly */
          Screen *ss = XScreenOfDisplay(Dpy, 0);
          Visual *vv = ss->root_visual;
            
          rm = vv->red_mask;
          gm = vv->green_mask;
          bm = vv->blue_mask;
            
          imageP->red_mask = rm;
          imageP->green_mask = gm;
          imageP->blue_mask = bm;
        }
                rs = ExtractShift_from_mask(rm);
                gs = ExtractShift_from_mask(gm);
                bs = ExtractShift_from_mask(bm);
                lmap = localMap;
                for ( ii=0; ii<nColors; ii++ )
                {
          r = lmap[0];
          g = lmap[1];
          b = lmap[2];
          if (16 == DpyDepth) {
            /* first normalize the 255-rgb to <maskdepth>-rgb
             */
            r = (int) (((double) r * (rm >> rs)) / 255.0);
            g = (int) (((double) g * (gm >> gs)) / 255.0);
            b = (int) (((double) b * (bm >> bs)) / 255.0);
          }
          cmapper[ii] = ((r << rs) & rm) | (( g << gs) & gm) |
            ((b << bs ) & bm);
          lmap += 3;
                }
                for ( ix=y=0; y<height; y++ )
                {
                    ix += ix%2;
                    for ( x=0; x<width; x++ )
                    {
                        XPutPixel(imageP, x, y, cmapper[data_dat[ix]]);
                        ix++;
                    }
                }
            }
            /* create an 8 bit image */
            else if ( DpyDepth == 8 )
            {
                /* create an "pixel to index" table */
                lmap = localMap;
                for ( ii=0; ii<nColors; ii++ )
                {
                    r = lmap[0];
                    g = lmap[1];
                    b = lmap[2];
                    AxGetRealPixel(0, r,g,b,0, 255, &cmapper[ii]);
                    lmap += 3;
                }
                /* fake an XImage for display */
                imageP->width = width;
                imageP->height = height;
                imageP->depth = depth;
                imageP->bits_per_pixel = depth;
                imageP->xoffset = 0;
                imageP->bitmap_pad = imageP->bitmap_unit = 16;
                imageP->bytes_per_line = byteWid;
                imageP->byte_order = MSBFirst;
                imageP->bitmap_bit_order = MSBFirst;
                imageP->obdata = NULL;
                imageP->format = ZPixmap;
                imageP->data = (char *)data_dat;
                imgData = (unsigned char*)imageP->data;
                while ( data_size-- )
                {
                    ix = (int) *imgData;
                    *imgData = cmapper[ix];
                    imgData++;
                }
            }

            /* dither down to a 1 bit image */
            else
            {
                /* create an "pixel to index" table */
                lmap = localMap;
                for ( ii=0; ii<nColors; ii++ )
                {
                    r = lmap[0];
                    g = lmap[1];
                    b = lmap[2];
                    AxGetRealPixel(0, r,g,b,0, 255, &cmapper[ii]);
                    lmap += 3;
                }

                /* fake an XImage to create a source for the dither */
                imageP->width = width;
                imageP->height = height;
                imageP->depth = depth;
                imageP->bits_per_pixel = depth;
                imageP->xoffset = 0;
                imageP->bitmap_pad = imageP->bitmap_unit = 16;
                imageP->bytes_per_line = byteWid;
                imageP->byte_order = MSBFirst;
                imageP->bitmap_bit_order = MSBFirst;
                imageP->obdata = NULL;
                imageP->format = ZPixmap;
                imageP->data = (char *)data_dat;
                imgData = (unsigned char*)imageP->data;
                while ( data_size-- )
                {
                    ix = (int) *imgData;
                    *imgData = cmapper[ix];
                    imgData++;
                }

                /* get the 'white' pixel. */
                AxGetRealPixel(0,255,255,255,0,255,&cindex);

                /* dither the color image in place */
                imgData = mdataP = (unsigned char*)TaskCalloc( pid,
                                                     mask_size,sizeof(char));
                cDataP = (unsigned char*)imageP->data;
                if ( bmap->hasMask )
                {
                    maskP = (unsigned char*)mask->data;
                    for (ii=0; ii<height; ii++)
                    {
                        /* if not white and mask on, set to black */
                        for (jj=0; jj<width; jj++)
                        {
                            if (((cDataP[jj]) != cindex) &&
                                 (maskP[(jj>>3)] & (1 << (7 - (jj%8)))))
                                    mdataP[(jj>>3)] |= (1 << (7 - (jj%8)));
                        }
                        cDataP += byteWid;
                        mdataP += monoByteWid;
                        maskP += monoByteWid;
                    }
                    TaskFree(pid,mask->data);
                    bmap->hasMask = FALSE;
                }
                else
                {
                    for (ii=0; ii<height; ii++)
                    {
                        /* if not white, set to black */
                        for (jj=0; jj<width; jj++)
                        {
                            if ((cDataP[jj]) != cindex)
                                mdataP[(jj>>3)] |= (1 << (7 - (jj%8)));
                        }
                        cDataP += byteWid;
                        mdataP += monoByteWid;
                    }
                }

                /* the color image is replaced */
                TaskFree(pid,imageP->data);
                imageP->data = (char*)imgData;
                imageP->format = XYBitmap;
                imageP->bits_per_pixel = imageP->depth = 1;
                imageP->bytes_per_line = monoByteWid;
            }
        }
    }
    return(AOK);
}

/*************************************************************************
function: FSRev1ImToXImage - read an old style (magic number 00) .im file
*/
static int
FSRev1ImToXImage(bmap, fp)
bitmap_struct *bmap;
FILE *fp;
{
    unsigned char shortbuf[2];
    unsigned short width, height;
    unsigned int bytewidth, monoByteWid, imgSize;
    unsigned char *temp_data, *inDataP,*outDataP;
    int pid, ix;
    XImage *imageP;

    /* read the width */
    fread(shortbuf,2,1,fp);
    width = (short) ((0x00ff & shortbuf[0]) | (0xff00 & (shortbuf[1] << 8)));

    /* read the height */
    fread(shortbuf,2,1,fp);
    height = (short) ((0x00ff & shortbuf[0]) | (0xff00 & (shortbuf[1] << 8)));

    /* allocate a buffer and read the data */
    pid = bmap->pid;
    imageP = &bmap->image;
    bytewidth = (unsigned int)(width+7)/8;
    imgSize = bytewidth*height;
    temp_data = (unsigned char *)TaskAlloc(pid,imgSize);
    if ( fread(temp_data,1,imgSize,fp) != imgSize )
        return(!AOK);
    memset(imageP, 0, sizeof(*imageP));
    monoByteWid = (width + 7)/8 + ((width + 7)/8)%2;
    imgSize = monoByteWid*height;
    imageP->data = (char*)TaskAlloc(pid,imgSize);
    inDataP = temp_data;
    outDataP = (unsigned char *)imageP->data;
    for (ix=0; ix<height; ix++)
    {
        memcpy(outDataP,inDataP,bytewidth);
        inDataP += bytewidth;
        outDataP += monoByteWid;
    }
    TaskFree(pid,temp_data);

    /* create an image */
    imageP->width = width;
    imageP->height = height;
    imageP->depth = 1;
    imageP->xoffset = 0;
    imageP->format = XYBitmap;
    imageP->bits_per_pixel = 1;
    imageP->bitmap_pad = imageP->bitmap_unit = 8;
    imageP->bytes_per_line = monoByteWid;
    imageP->byte_order = MSBFirst;
    imageP->bitmap_bit_order = LSBFirst;
    imageP->obdata = NULL;

    /* done */
    return(AOK);
}
#else   /*   } AX_X11 { AXMS  */

/*************************************************************************
function: FSRev1ImToWinImage - read an old style .im file
*/
static int
FSRev1ImToWinImage(bmap, fp)
bitmap_struct *bmap;
FILE *fp;
{
    int height, width;
    int retval;
    char shortbuf[10];
    XImage *infoP;


    infoP = &bmap->image;

    fread(shortbuf,2,1,fp);
    width = ((0x00ff & shortbuf[0]) | (0xff00 & (shortbuf[1] << 8)));

    fread(shortbuf,2,1,fp);
    height = ((0x00ff & shortbuf[0]) | (0xff00 & (shortbuf[1] << 8)));

    retval = MakeMonoWinPixmap(infoP, fp, width,height, bmap->pid, FALSE);
    return(retval);
}


/***************************************************************************/
static int
MakeMonoWinPixmap(infoP, fp, width, height, pid, new_style_image)
XImage *infoP;
FILE *fp;
int width, height;
int pid;
int new_style_image;    /* "old" or "new" style image. */
{
    unsigned int err;
    unsigned char *sourceP, *destP;
    unsigned int i, j, k;
    unsigned char *image_data, *win_image_data;
    unsigned char *temp_data, *inDataP, *outDataP;
    int data_size, bytewidth, mono_bytewidth;
    register unsigned char in, out;
    register unsigned char *byteP;

    infoP->width  = width;
    infoP->height = height;
    infoP->depth  = 1;

    /* get and pad the mono image data for Applix. */
    bytewidth = (width + 7) / 8;
    data_size = bytewidth * height;
    temp_data = (unsigned char *)TaskAlloc(pid, data_size);
    if ( fread(temp_data, 1, data_size, fp) != data_size )
        {
        if ( temp_data )
            TaskFree(pid, temp_data);
        return(!AOK);
        }
    mono_bytewidth = (width + 7)/8 + ((width + 7)/8)%2;
    data_size = mono_bytewidth * height;
    image_data = (unsigned char *)TaskAlloc(pid, data_size);
    inDataP = temp_data;
    outDataP = image_data;
    for ( i=0; i < height; i++ )
        {
        memset(outDataP, 0, mono_bytewidth);
        memcpy(outDataP, inDataP, bytewidth);
        inDataP += bytewidth;
        outDataP += mono_bytewidth;
        }
    infoP->data = image_data;
    infoP->bytes_per_line = mono_bytewidth;
    infoP->byte_order = MSBFirst;

    /* create the Windows monochrome bitmap handle. */
    if ( new_style_image )
        {
        /* Rev 2 and greater. */
        infoP->bitmap_bit_order = MSBFirst;
        infoP->bitmap_handle = MS_CreateBitmap(width, height, 1,1, image_data);
        }
    else
        {
        /* old images (rev 1) need to invert the order *
         * of the bits in the byte.                    */
        sourceP = temp_data;
        win_image_data = (unsigned char*)TaskAlloc(pid, data_size);
        for ( i=0; i < height; i++ )
            {
            destP = win_image_data + (i * mono_bytewidth);
            memset( destP, 0, mono_bytewidth );
            for ( j=0; j < bytewidth; j++ )
                {
                in = *sourceP++;
                out = 0;
                for ( k=0; k<8; k++ )
                    {
                    out = out << 1;
                    out = out | (in & 0x01);
                    in = in >> 1;
                    }
                *destP++ = out;
                }
            }

        infoP->bitmap_handle = MS_CreateBitmap(width, height, 1, 1,
                                                    win_image_data);
        infoP->bitmap_bit_order = LSBFirst;
        TaskFree(pid, win_image_data);
        }

    TaskFree(pid, temp_data);
    return AOK;
}



/*************************************************************************
function: FSRev2ImToWinImage - read a new style .im file
*/
static int
FSRev2ImToWinImage(bmap, fp)
bitmap_struct *bmap;
FILE *fp;
{
    HDC hdc;
    XImage *xP, *infoP, *im_maskP;
    int monoByteWid, monoImgsize;
    int win_mono_size;
    unsigned long cindex;    
    unsigned long cmapper[256];
    unsigned char *mask_dat, *data_dat;    
    unsigned char *inDataP, *outDataP;
    char *clr, *clrP, *def_name;    
    unsigned char *mask_bits;
    unsigned char *image_data, *ax_byte_data, *ax_mask_data;
    unsigned char localMap[768], *lmap;
    unsigned int file_imgsize;
    unsigned int win_bytewidth, win_imgSize;
    unsigned int size;
    int width, height, depth;
    int i, j, len;
    int nColors;
    int ii, jj, ix, x, y, byteWid, mask_size;    
    int ch, pid, line_ix, mask_ix;
    unsigned int r, g, b,c,m,k;    
    int rm, gm, bm, rs, gs, bs;
    unsigned int retval;
    unsigned char *sourceP, *destP;
    int bytewidth, data_size;
    int not_used = 0;


    /* get the width */
    width = fgetc(fp);
    ch = fgetc(fp);
    width = (width << 8) | ch;

    /* get the height */
    height = fgetc(fp);
    ch = fgetc(fp);
    height = (height << 8) | ch;

    /* get the depth */
    depth = fgetc(fp);
    ch = fgetc(fp);
    depth = (depth << 8) | ch;

    pid = bmap->pid;
    /* image_mask existence flag */
    bmap->hasMask = fgetc(fp);
    if ( (bmap->hasMask) && (depth == 1) )
        bmap->hasMask = FALSE;

    /* Number of colors */
    nColors = fgetc(fp);
    ch = fgetc(fp);
    nColors = (nColors << 8) | ch;

    infoP = &bmap->image;
    im_maskP = &bmap->mask;

    if ( depth == 1 )
        {
        /* just a bitmap */
        retval = MakeMonoWinPixmap(infoP, fp, width, height, pid, TRUE);
        return (retval);
        }

    if ( depth == 8 )
        fread(localMap, nColors, 3, fp);

    /* the data on disk is not padded on a per-line basis   */
    file_imgsize = width * height;
    win_bytewidth = ((width + 3) / 4) * 4;
    win_imgSize = win_bytewidth * height;

    /* a buffer for the file data */
    memset(infoP, 0, sizeof(*infoP));
    ax_byte_data = (unsigned char *) TaskAlloc(pid, file_imgsize);

    /* read the raw data */
    retval = fread(ax_byte_data, 1, file_imgsize, fp);
    if ( retval != file_imgsize )
        return (!AOK);

    if ( bmap->hasMask )
        {
        /* create a monochrome bitmap to use as a mask */
        /* asterix monochrome images are written to disk as straight bytes */
        /* and not padded to any particular byte boundries.  On the other  */
        /* hand, Windows monochrome images are padded to 2 byte boundries */
        monoByteWid = (width + 7)/8;
        monoImgsize = monoByteWid*height;

        ax_mask_data = (unsigned char *) TaskAlloc(pid, monoImgsize);        
        if ( fread( ax_mask_data, 1,monoImgsize,fp) != monoImgsize )
            {
            TaskFree(pid, ax_mask_data);
            return(!AOK);
            }

        /* windows rounds all image lines to 4 byte multiples */
        win_mono_size = ((monoByteWid + 1) & ~0x1 );  /* round up to 2-byte */
        mask_size = win_mono_size * height;
        mask_bits = (unsigned char *) TaskAlloc(pid, mask_size);
 
        sourceP = ax_mask_data;
        for ( i = 0; i < height; i++ )
            {
            /* pad the rows of the mask image to 4 byte boundries */
            destP = mask_bits + (i * win_mono_size);
            memset(destP, 0, win_mono_size );
            for (j = 0; j < monoByteWid; j++)
                {
                *destP++ = *sourceP++;
                }
            }

        im_maskP->bitmap_handle = MS_CreateBitmap(width, height,
                                                1, 1, mask_bits);
        if ( im_maskP->bitmap_handle == (HBITMAP)NULL )
            {
            TaskFree(pid, ax_mask_data);
            TaskFree(pid, mask_bits);
            return(!AOK);
            }
        }

    /* create an 8 bit XImage and a 1 bit mask to return to the caller */
    if ( bmap->display == FALSE )
        {
        /* copy the colormap for external use */
        if ( bmap->needcolor )
            {
            lmap = localMap;
            bmap->cmap.n_colors = 256;
            size = 256 *sizeof(color_type);
            bmap->cmap.color = (color_type *)TaskAlloc(pid, size);
            def_name = XLT("Color");
            len = strlen(def_name);
            clr = (char *)malloc((len+3) * sizeof(char));
            strcpy(clr, def_name);
            clrP = &clr[len];
            for (ii=0; ii<nColors && ii<256; ii++)
                {
                r = lmap[0];
                g = lmap[1];
                b = lmap[2];
                lmap += 3;
                sprintf(clrP,"%3d",ii);
                rgb2cmyk(r,g,b, &c,&m,&y,&k, FALSE);
                bmap->cmap.color[ii].see_thru = FALSE;
                bmap->cmap.color[ii].ink_type = 0;
                bmap->cmap.color[ii].c = c;
                bmap->cmap.color[ii].m = m;
                bmap->cmap.color[ii].y = y;
                bmap->cmap.color[ii].k = k;
                bmap->cmap.color[ii].name = (char *)TaskAlloc(pid,len+3);
                strcpy(bmap->cmap.color[ii].name, clr);
                }
            for (; ii<256; ii++ )
                {
                sprintf(clrP,"%3d",ii);
                bmap->cmap.color[ii].see_thru = FALSE;
                bmap->cmap.color[ii].ink_type = 0;
                bmap->cmap.color[ii].c = 0;
                bmap->cmap.color[ii].m = 0;
                bmap->cmap.color[ii].y = 0;
                bmap->cmap.color[ii].k = 0;
                bmap->cmap.color[ii].name = (char *)TaskAlloc(pid,len+3);
                strcpy(bmap->cmap.color[ii].name, clr);
                }
            free(clr);
            }

        /* fake an XImage to transfer the image data */
        infoP->width = width;
        infoP->height = height;
        infoP->depth = depth;
        infoP->bits_per_pixel = depth;
        infoP->xoffset = 0;
        infoP->bitmap_pad = infoP->bitmap_unit = 32;
        infoP->bytes_per_line = win_bytewidth;
        infoP->byte_order = MSBFirst;
        infoP->bitmap_bit_order = MSBFirst;
        infoP->obdata = NULL;
        infoP->format = ZPixmap;

        /* pad the data for Applix. */
        bytewidth = width + width%2;
        data_size = bytewidth * height;
        infoP->data = (char*)TaskAlloc(pid, data_size);
        sourceP = ax_byte_data;
        destP = infoP->data;
        for ( ix=0; ix < height; ix++ )
            {
            memcpy(destP, sourceP, width);
            sourceP += width;
            destP += bytewidth;
            }
        AxWinCreate8BitDIB(infoP, infoP->data, data_size);

        /* fake an image for the mask, too (if there is one) */
        if ( bmap->hasMask )
            {
            im_maskP->depth = im_maskP->bits_per_pixel = 1;
            im_maskP->bytes_per_line = monoByteWid;
            im_maskP->format = XYBitmap;
            im_maskP->data = mask_bits;

            inDataP = mask_bits;
            outDataP = infoP->data;

            /* check to see if there is a transparent in the colormap */
            for (y=0; y<height; y++)
                {
                line_ix = y*bytewidth;
                mask_ix = y*win_mono_size;
                for (ix=0; ix<width; ix++)
                    {
                    cindex = outDataP[line_ix+ix];
                    if ( !(inDataP[mask_ix+(ix>>3)] & Bitmask[ix&7]) )
                        bmap->cmap.color[cindex].see_thru = TRUE;
                    }
                }
            }
        TaskFree(pid, ax_byte_data);
        }
    else
        {
        /*------------------------------------------------------------------*/
        /* get the image.   Note: the rows of pixels for Windows */
        /* must be a multiple of 4 bytes.  We are relying on "XCreateImage"
        /* to set up the bytes_per_line value to reflect this padding     */
        /* Also, Windows treats the origin of a bitmap as lower left corner */
        /* so we also flip the image over here so the rest of the  world */
        /* can treat the upper left corner as the (0,0) position */
        /*------------------------------------------------------------------*/
        if ( DpyDepth > 8 )
            {
            /* create a dummy X image structure to get a bunch of parameters */
            /* filled in a consistant way.                                   */
            int tmp_wid = width + width%2;
            if ( DpyDepth == 15 || DpyDepth == 16 )
                byteWid = ((tmp_wid * DpyDepth) + 7)/8;
            else
                byteWid = width * 4;

            xP = XCreateImage(Dpy, AxVisual, DpyDepth, ZPixmap, 0,
                            (char *)NULL, width, height, not_used, not_used );
            /* copy the new parameters to our struct and then throw away     */
            /* the dummy struct                                              */
            *infoP = *xP;
            XDestroyImage(xP);

            /* allocate space for the image data */
            infoP->data = (char *)TaskAlloc(pid, 
                                     (height * infoP->bytes_per_line) );

            /* create an "pixel to index" table */
            rm = infoP->red_mask;
            gm = infoP->green_mask;
            bm = infoP->blue_mask;
            rs = ExtractShift_from_mask(rm);
            gs = ExtractShift_from_mask(gm);
            bs = ExtractShift_from_mask(bm);
            lmap = localMap;
            for ( ii=0; ii < nColors; ii++ )
                {
                if ( DpyDepth == 15 || DpyDepth == 16 )
                    {
                    normalize_rgb_to_555(lmap[0],lmap[1],lmap[2], &r,&g,&b);
                    cmapper[ii] = (r & rm) + (g & gm) + (b & bm);
                    }
                else
                    {
                    r = lmap[0];
                    g = lmap[1];
                    b = lmap[2];
                    cmapper[ii] = ((r << rs) & rm) | (( g << gs) & gm) |
                                                ((b << bs ) & bm);
                    }
                lmap += 3;
                }
            
            /* put the pixels into the image data area                  */
            /* make sure we flip the data because the "negative height" */
            /* method doesn't seem to be reliable (thanks Microsoft!).  */
            /* and otherwise Windows treats the lower left corner as the */
            /* image origin                                              */
            ix = 0;                     /* index into image data from disk */
            for ( y=height-1; y >= 0; y-- )
                {
                for ( x=0; x<width; x++ )
                    {
                    XPutPixel(infoP, x, y, cmapper[ax_byte_data[ix]]);
                    ix++;
                    }
                }

            AxWinCreate16BitDIB(infoP, height * win_bytewidth);
            }
        else if ( DpyDepth == 8 )
            {
            /* a buffer for the padded data, for windows */
            image_data = (unsigned char *) TaskAlloc(pid, win_imgSize);

            infoP->width = width;
            infoP->height = height;

            /* create an "pixel to index" table */
            lmap = localMap;
            for ( ii=0; ii<nColors; ii++ )
                {
                r = lmap[0];
                g = lmap[1];
                b = lmap[2];
                AxGetRealPixel(0, r,g,b,0, 255, &cmapper[ii]);
                lmap += 3;
                }

            /* now copy the bytes to the real image so that we can */
            /* padd the byes out to 4 byte boundry per line, and
            /* invert the image from top to bottom */
            sourceP = ax_byte_data;
            for (i = 0; i < height; i++)
                {
                destP = image_data + ((height - i - 1) * win_bytewidth);
                memset(destP, 0, win_bytewidth);
                for (j = 0; j < width; j++)
                    *destP++ = cmapper[(int)*sourceP++];
                }

            infoP->depth = depth;
            infoP->bits_per_pixel = depth;
            infoP->xoffset = 0;
            infoP->bitmap_pad = infoP->bitmap_unit = 32;
            infoP->bytes_per_line = win_bytewidth;
            infoP->byte_order = MSBFirst;
            infoP->bitmap_bit_order = MSBFirst;
            infoP->obdata = NULL;
            infoP->format = ZPixmap;
            infoP->data = ax_byte_data;        

            if ( bmap->hasMask )
                {
                im_maskP->depth = im_maskP->bits_per_pixel = 1;
                im_maskP->bytes_per_line = monoByteWid;
                im_maskP->format = XYBitmap;
                im_maskP->data = mask_bits;
                }

            AxWinCreate8BitDIB(infoP, image_data, win_imgSize);
            TaskFree(pid, image_data);
            }
        }

    if ( bmap->hasMask )
        TaskFree(pid, ax_mask_data);
    return (AOK);
}

#endif   /* } AXMS  */

static int
doReadNum(buf, fp)
char *buf;
FILE *fp;
{
    char ch, *bufP;

    ch = fgetc(fp);
    if ( ch == EOF )
        return EOF;
    while ( isspace(ch) )
    {
        ch = fgetc(fp);
        if ( ch == EOF )
            return EOF;
    }
    while ( isdigit(ch) )
    {
        if ( ch == EOF )
            return EOF;
        *buf++ = ch;
        ch = fgetc(fp);
    }
    *buf = '\0';
    return(AOK);
}

static int
doReadHexNum(fp,n, result)
FILE *fp;
int n;
int *result;
{
    char ch,  *posP;
    static char *chars = NXLT("0123456789ABCDEF");

    ch = fgetc(fp);
    if ( ch == EOF )
        return EOF;
    if ( !isxdigit(ch) )
        return !AOK;
    posP = strchr(chars,ch);

    if ( n == 1 )
    {
        *result = ((posP-chars) & 15 );
        return AOK;
    }

    *result = ((posP-chars) << 4 );

    ch = fgetc(fp);
    if ( ch == EOF )
        return EOF;
    if ( !isxdigit(ch) )
        return !AOK;
    posP = strchr(chars,ch);
    *result = *result | ((posP-chars) & 15);

    return AOK;
}

static int
doReadString(bmap, buf, fp)
bitmap_struct *bmap;
char *buf;
FILE *fp;
{
    char ch;
    int ignore;

    ch = fgetc(fp);
    ignore = FALSE;
    while ( TRUE )
    {
        if ( ch == '\n' )
        {
            ch = fgetc(fp);
            if ( ch != ' ' )
                *buf++ = ch;
        }
        if ( ignore )
        {
            *buf++ = ch;
            ignore = FALSE;
            continue;
        }
        else if (  ch == '"' )
            break;
        else if (  ch == '\\' )
        {
            *buf++ = ch;
            ignore = TRUE;
        }
        else
            *buf++ = ch;
        ch = fgetc(fp);
    }
    *buf = '\0';
}

static int
dobitsread(data_dat, data_size, fp)
char *data_dat;
int data_size;
FILE *fp;
{
    int i;
    char ch, *datP;

    datP = data_dat;
    for ( i=0;i<data_size;i++ )
    {
        if( fetchbyte( &ch,fp))
            *datP++ = ch;
        else
            return(!AOK);
    }
    while (fgetc(fp) != '\n');
    return(AOK);
}

static int
doReadMask(bmap, imageP, fp)
bitmap_struct *bmap;
XImage *imageP;
FILE *fp;
{
    char *data_dat;
    int err, byteWid, data_size;

    byteWid = (bmap->image.width + 7)/8 + ((bmap->image.width + 7)/8)%2;

    /* get the image size, allocate a buffer, and read the image data */
    data_size = byteWid*bmap->image.height;
    data_dat = (char *)TaskAlloc(bmap->pid, data_size);
    err = dobitsread(data_dat, data_size, fp);
    imageP->width = bmap->image.width;
    imageP->height = bmap->image.height;
    imageP->depth = 1;
    imageP->data = data_dat;
    imageP->bytes_per_line = byteWid;
    return(err);
}

static int
doReadData(bmap, fp)
bitmap_struct *bmap;
FILE *fp;
{
    unsigned char *data_dat;
    int err, byteWid, data_size;

    if ( bmap->image.depth == 1 )
    {
        bmap->hasMask = FALSE;
        return( doReadMask(bmap, &bmap->image, fp));
    }
    else
    {
        byteWid = bmap->image.width + bmap->image.width%2;
        data_size = byteWid*bmap->image.height;
        data_dat = (unsigned char *)TaskAlloc(bmap->pid, data_size);
        err = dobitsread(data_dat, data_size, fp);
        bmap->image.data = (char *)data_dat;
        bmap->image.bytes_per_line = byteWid;
        return(err);
    }
}

static int
doReadColormap(bmap, local_cmap, fp, info)
bitmap_struct *bmap;
cmap_type *local_cmap;
FILE *fp;
AFileInfo *info;
{
    color_type *tmp_color;
    char name[STRPATH+1];
    char buf[80];
    unsigned char  outval;
    int err, val, i, size, cindex;

    memset( local_cmap, 0, sizeof(cmap_type) );
    local_cmap->n_colors = 256;
    size = 256 *sizeof(color_type);
    local_cmap->color = (color_type *)TaskAlloc(bmap->pid,size);

    while (fgetc(fp) != '\n');
    for ( i=0;i<256;i++ )
    {
        cindex = 0;
        outval = fgetc(fp);
        if (outval == '"') /* begin of name */
        {
            doReadString(bmap, name, fp);
            cindex = strlen(name);
            tmp_color = &local_cmap->color[i];
            get_color_name(bmap->pid, tmp_color, name, cindex, info);
        }
        err = doReadHexNum(fp, 2, &val);
        if ( err == AOK )
            tmp_color->c = val;
        err = doReadHexNum(fp, 2, &val);
        if ( err == AOK )
            tmp_color->m = val;
        err = doReadHexNum(fp, 2, &val);
        if ( err == AOK )
            tmp_color->y = val;
        err = doReadHexNum(fp, 2, &val);
        if ( err == AOK )
            tmp_color->k = val;
        err = doReadHexNum(fp, 1, &val);
        if ( err == AOK )
            tmp_color->ink_type = val;
        err = doReadHexNum(fp, 1, &val);
        if ( err == AOK )
            tmp_color->see_thru = (val) ? TRUE: FALSE;

        while (fgetc(fp) != '\n');
    }
    while (fgetc(fp) != '\n'); /*end colormap */
    bmap->hascmap = TRUE;
    return (AOK);
}

static int
doReadEnd(bmap, done, fp)
bitmap_struct *bmap;
int *done;
FILE *fp;
{
    *done = TRUE;
    return (AOK);
}

static int
doReadHotSpot(bmap, fp)
bitmap_struct *bmap;
FILE *fp;
{
    char buf[80];
    fgetc(fp);
    doReadNum(buf, fp);
    bmap->x_spot = atoi(buf);
    doReadNum(buf, fp);
    bmap->y_spot = atoi(buf);
    while (fgetc(fp) != '\n');
    return (AOK);
}

static int
doReadDepth(bmap, fp)
bitmap_struct *bmap;
FILE *fp;
{
    char buf[80];
    doReadNum(buf, fp);
    bmap->image.depth = atoi(buf);
    return (AOK);
}

static int
doReadHeight(bmap, fp)
bitmap_struct *bmap;
FILE *fp;
{
    char buf[80];
    doReadNum(buf, fp);
    bmap->image.height = atoi(buf);
    return (AOK);
}

static int
doReadWidth(bmap, fp)
bitmap_struct *bmap;
FILE *fp;
{
    char buf[80];
    doReadNum(buf, fp);
    bmap->image.width = atoi(buf);
    return (AOK);
}

static int
doCommand(bmap, info, local_cmap, buf, fp, done)
bitmap_struct *bmap;
AFileInfo *info;
cmap_type *local_cmap;
char *buf;
FILE *fp;
int *done;
{
    int err;

    err = !AOK;
    if ( buf )
    {
        if (strncmp(buf, NXLT("WIDTH"), 5) == 0 )
            err = doReadWidth(bmap, fp);
        else if (strncmp(buf, NXLT("HEIGHT"), 6) == 0 )
            err = doReadHeight(bmap, fp);
        else if (strncmp(buf, NXLT("DEPTH"), 5) == 0 )
            err = doReadDepth(bmap, fp);
        else if (strncmp(buf, NXLT("HOTSPOT"), 8) == 0 )
            err = doReadHotSpot(bmap, fp);
        else if ( strncmp(buf, NXLT("COLORMAP"), 8) == 0 )
            err = doReadColormap(bmap, local_cmap, fp, info);
        else if (strncmp(buf, NXLT("DATA"), 4) == 0 )
            err = doReadData(bmap, fp);
        else if (strncmp(buf, NXLT("MASK"), 4) == 0 )
        {
            err = doReadMask(bmap, &bmap->mask, fp);
            bmap->hasMask = TRUE;
        }
        else if (strcmp(buf, ATOK_END) == 0 )
            err = doReadEnd(bmap, done, fp);
        else
            err = !AOK;
    }
    return(err);
}

static int
getCommand(buf, fp)
char *buf;
FILE *fp;
{
    char ch;

    ch = fgetc(fp);
    if ( ch == EOF )
        return EOF;
    if ( ch == '\n' )
        ch = fgetc(fp);
    if ( ch == EOF )
        return EOF;
    else if ( ch == '*' )
    {
        *buf++ = ch;
        ch = fgetc(fp);
        while ( isalnum(ch) )
        {
            *buf++ = ch;
            ch = fgetc(fp);
        }
        *buf = '\0';
        return  AOK;
    }

    while ( isalnum(ch) )
    {
        if ( ch == EOF )
            return EOF;
        *buf++ = ch;
        ch = fgetc(fp);
    }

    *buf = '\0';
    return AOK;
}

/*************************************************************************
*function: FSRev3ImToImage - read a new style im file
*/
static int
FSRev3ImToImage(bmap, fp, info)
bitmap_struct *bmap;
FILE *fp;
AFileInfo *info;
{
    unsigned char  localMap[768 /* 256*3 */], *lmap;
    color_type *t_c;
    cmap_type local_cmap;
    unsigned char *data_dat, buf[80];
    unsigned char *imgData, *cDataP, *mdataP, *maskP;
    int wid, hyt, ix, rm, gm, bm, rs, gs, bs, x, y;
    int ii,jj,nColors, byteWid, monoByteWid;
    int err, data_size, mask_size, j;
    unsigned int r, g, b;
    unsigned long cindex, cmapper[256];
    int c,m,k,image_byte_width,  width_on_disk, done;
    XImage *xP,  *imageP, *mask;
#if AXMS
    unsigned char *destP, *sourceP, *win_data_dat;
    int red, green, blue, win_data_size;
#endif


    /* get the width,height,depth, etc */
    err = AOK;
    done = FALSE;
    while (done == FALSE )
    {
        err = getCommand(buf, fp);
        if ( err )
            return(err);
        err = doCommand(bmap, info, &local_cmap, buf, fp, &done);
        if ( err )
            return(err);
    }

    hyt = bmap->image.height;
    wid = bmap->image.width;
    data_dat = (unsigned char *)bmap->image.data;

    /* mono image */
    if ( bmap->image.depth == 1 )
    {
        /* fake an XImage */
        imageP = &bmap->image;
        imageP->bits_per_pixel = 1;
        imageP->xoffset = 0;
#if AX_X11
        imageP->bitmap_pad = imageP->bitmap_unit = 16;
#else
        imageP->bitmap_pad = imageP->bitmap_unit = 32;
#endif
        imageP->byte_order = MSBFirst;
        imageP->bitmap_bit_order = MSBFirst;
        imageP->obdata = NULL;
        imageP->format = XYBitmap;


#if AXMS    /* { */
        /* Note: for color images, the rows of pixels for Windows must be *
         * a multiple of 4 bytes, so we need to read the image into one   *
         * buffer, and then copy it to another, padding each row out to   *
         * a multiple of 4 bytes.                                         */
        /* Also,Windows treats the origin of a bitmap as the lower   *
         * left corner so we also flip the image over here so the    *
         * rest of the aster*x world can treat the upper left corner *
         * as the (0,0) position.                                    */
        if ( bmap->needcolor && DpyDepth == 8 )
            {
            byteWid = (wid + 7)/8 + ((wid + 7)/8)%2;
            image_byte_width = ((byteWid + 3)/4 ) * 4;
            data_size = image_byte_width * hyt;
            win_data_dat = (unsigned char *)TaskAlloc(bmap->pid, data_size);

            sourceP = data_dat;
            for ( ix=hyt-1; ix >= 0; ix-- )
                {
                destP = win_data_dat + (ix * image_byte_width);
                for ( ii=0; ii < byteWid; ii++ )
                    *destP++ = *sourceP++;
                }
            AxWinCreate1BitDIB(imageP, win_data_dat);
            TaskFree(bmap->pid, win_data_dat);
            }
        else
            {
            /* mono, black & white image. don't need to manipulate the data. */
            imageP->bitmap_handle = MS_CreateBitmap(wid, hyt, 1, 1, data_dat);
            }
#endif    /* } AXMS */

        return( AOK );
    }
    /* color image */
    else /* assume depth == 8 */
    {
		/* Note: if nColors is changed, a number of declarations will also
			have to change, including local_map, cmapper, cmapper_truecolor */
        nColors = 256;

        if ( bmap->hascmap == FALSE )
        {
			/* Attention future programmers: if our standard icons 
				begin to include a standard colormap, please update 
				'create_cmap' to generate that same colormap */
            local_cmap.color = (color_type *)0;
            create_cmap(bmap->pid, &local_cmap);

#if AX_X11
			/* In 8-bit Truecolor mode, convert the standard colormap
				to an index table and cache the results. This is
				expensive to do (at least on EDJ terminals), so
				we will use the cached results wherever possible */
			if ((AxVisual->class == TrueColor) && 
            							( DpyDepth == 8 ) &&
										!got_cmapper_truecolor) {

            	local_cmap_truecolor.color = (color_type *)0;
            	create_cmap(bmap->pid, &local_cmap_truecolor);

               	for ( ii=0; ii<nColors; ii++ )
               	{
                   	t_c = &local_cmap_truecolor.color[ii];
                   	c = t_c->c;
                   	m = t_c->m;
                   	y = t_c->y;
                   	k = t_c->k;
                   	AxGetRealPixel(2, c,m,y,k, 255, 
										&cmapper_truecolor[ii]);
               	}
				got_cmapper_truecolor = TRUE;
			}
#endif
        }
        /* all images are padded to even byte boundries when written to disk  */
        width_on_disk = wid + wid%2;
        image_byte_width = width_on_disk;


        imageP = &bmap->image;
        mask = &bmap->mask;
        monoByteWid = (imageP->width + 7)/8 + ((imageP->width + 7)/8)%2;

        /* create an 8 bit XImage and a 1 bit mask to return to the caller */
        if ( bmap->display == FALSE )
        {
            /* copy the colormap for external use */
            if ( bmap->needcolor )
            {
                bmap->cmap.n_colors =local_cmap.n_colors;
                bmap->cmap.color = local_cmap.color;
                local_cmap.color = (color_type *)0;
            }

            /* fake an XImage to transfer the image data */
            imageP->xoffset = 0;
            imageP->bits_per_pixel = 8;
#if AX_X11
            imageP->bitmap_pad = imageP->bitmap_unit = 16;
#else
            imageP->bitmap_pad = imageP->bitmap_unit = 32;
#endif
            imageP->byte_order = MSBFirst;
            imageP->bitmap_bit_order = MSBFirst;
            imageP->obdata = NULL;
            imageP->format = ZPixmap;

            /* fake an image for the mask, too (if there is one) */
            if ( bmap->hasMask )
            {
#if AX_X11
                mask->bitmap_pad = mask->bitmap_unit = 16;
#else
                mask->bitmap_pad = mask->bitmap_unit = 32;
#endif
                mask->depth = mask->bits_per_pixel = 1;
                mask->bytes_per_line = monoByteWid;
                mask->xoffset = 0;
                mask->byte_order = MSBFirst;
                mask->bitmap_bit_order = MSBFirst;
                mask->obdata = NULL;
                mask->format = XYBitmap;
            }
        }
        /* need a real image and mask, formatted for display */
        else
        {
            /* the mask is the same on all displays */
            if ( bmap->hasMask )
            {
                mask->bits_per_pixel = 1;
                mask->xoffset = 0;
                mask->bitmap_pad = imageP->bitmap_unit = 16;
                mask->bytes_per_line = monoByteWid;
                mask->byte_order = MSBFirst;
                mask->bitmap_bit_order = MSBFirst;
                mask->obdata = NULL;
                mask->format = XYBitmap;
            }

            /* create a TrueColor/DirectColor DpyDepth image */
            if ( DpyDepth > 8 )
            {
                /* create and copy the image */
#if AXMS
                if ( DpyDepth >= 15 )
                    {
                    /* compute the number of bytes, rounded up to 4 */
                    /* byte boundries. To do this we compute number of words */
                    /* and multiply that by 4                                */
                    width_on_disk = ((width_on_disk * DpyDepth) + 31)/32;
                    width_on_disk = width_on_disk * 4;
                    }
                else
#endif
                width_on_disk = 4*wid;

                data_size = width_on_disk * hyt;
                imageP->data = (char *)TaskAlloc(bmap->pid,data_size);
                xP = (XImage *)XCreateImage(Dpy, AxVisual,
                                            DpyDepth, ZPixmap,
                                            0, imageP->data,
                                            wid, hyt,
                                            32, width_on_disk);
                *imageP = *xP;
                xP->width = xP->height = 0;
                xP->data = (char *)0;
                XDestroyImage(xP);

                /* create an "pixel to index" table */
                rm = imageP->red_mask;
                gm = imageP->green_mask;
                bm = imageP->blue_mask;
#if AX_X11
        if ( 0 == rm && 0 == gm && 0 == bm ) {
          /* cheat, and read masks from display directly */
          Screen *ss = XScreenOfDisplay(Dpy, 0);
          Visual *vv = ss->root_visual;
            
          rm = vv->red_mask;
          gm = vv->green_mask;
          bm = vv->blue_mask;
            
          imageP->red_mask = rm;
          imageP->green_mask = gm;
          imageP->blue_mask = bm;
        }
#endif
                rs = ExtractShift_from_mask(rm);
                gs = ExtractShift_from_mask(gm);
                bs = ExtractShift_from_mask(bm);
                for ( ii=0; ii<nColors; ii++ )
                {
                    t_c = &local_cmap.color[ii];
                    cmyk2rgb(t_c->c, t_c->m, t_c->y, t_c->k, &r, &g, &b);
#if AXMS
                    if ( DpyDepth == 15 || DpyDepth == 16 )
                    {
                        normalize_rgb_to_555(r, g, b, &r, &g, &b);
                        cmapper[ii] = (r & rm) + (g & gm) + (b & bm);
                    }
                    else
                    {
                        cmapper[ii] = ((r << rs) & rm) | (( g << gs) & gm) |
                                                            ((b << bs ) & bm);
                    }
#else
            if (16 == DpyDepth) {
              /* first normalize the 256-rgb to <maskdepth>-rgb
               */
              r = (int) (((double) r * (rm >> rs)) / 255.0);
              g = (int) (((double) g * (gm >> gs)) / 255.0);
              b = (int) (((double) b * (bm >> bs)) / 255.0);
            }
                    
                    cmapper[ii] = ((r << rs) & rm) | (( g << gs) & gm) |
              ((b << bs ) & bm);
#endif
                }
                for ( ix=y=0; y<hyt; y++ )
                {
                    ix += ix%2;
                    for ( x=0; x<wid; x++ )
                    {
                        XPutPixel(imageP, x, y, cmapper[data_dat[ix]]);
                        ix++;
                    }
                }
            }

            /* create an 8 bit image */
			/* EDJ TrueColor terminals come thru here too */
            else if ( DpyDepth == 8 )
            {
                /* Create an "pixel to index" table */
#if AX_X11
				/* In TrueColor mode, see if we can use the cached one,
					because the required XAllocColor call (in 
					axGetTrueColorPixel) is too slow to do 256 times 
					for each icon. We can used the cached one if
					the icon has not specified a bitmap, or if the
					bitmap is the same as the 'standard' one. */
				if ((AxVisual->class == TrueColor) && 
							got_cmapper_truecolor &&
        					(( bmap->hascmap == FALSE ) ||
							 ( compare_cmaps( &local_cmap, 
												&local_cmap_truecolor))) ) {
                	for ( ii=0; ii<nColors; ii++ ) {
						cmapper[ii] = cmapper_truecolor[ii];
					}
				}
				else {
#endif
                	for ( ii=0; ii<nColors; ii++ )
                	{
                    	/* Note: on windows we build the palette directly 
							from the local colormap table, so the mapping 
							is identity   */
                    	t_c = &local_cmap.color[ii];
                    	c = t_c->c;
                    	m = t_c->m;
                    	y = t_c->y;
                    	k = t_c->k;
                    	AxGetRealPixel(2, c,m,y,k, 255, &cmapper[ii]);
                	}
#if AX_X11
                }
#endif

                /* fake an XImage for display */
                imageP->bits_per_pixel = 8;;
                imageP->xoffset = 0;
                imageP->byte_order = MSBFirst;
                imageP->bitmap_bit_order = MSBFirst;
                imageP->obdata = NULL;
                imageP->format = ZPixmap;
                imgData = (unsigned char*)imageP->data;
                data_size = width_on_disk * hyt;
#if AXMS
                imageP->bitmap_pad = imageP->bitmap_unit = 32;
#else
                imageP->bitmap_pad = imageP->bitmap_unit = 16;
#endif
                for ( ix=0; ix<data_size; ix++ )
                   {
                   *imgData = cmapper[  (int) *imgData ];
                   imgData++;
                   }
#if AXMS    /* { */
        /*-------------------------------------------------------------- */
        /*  Remember, that Windows 8-bit images are padded               */
        /*   to 4-byte boundries, so we need to account for this when we */
        /*   allocate space for the buffer, and then we pad as we read the */
        /*   applix image  */  
        /* Windows treats the lower left corner of bitmap as the origin */
        /* so we also flip the image over here so the rest of applixware */
        /* can treat the upper left corner as the (0,0) position */
        /*------------------------------------------------------------------*/
                image_byte_width = ((width_on_disk + 3) & ~0x3);
                win_data_size = image_byte_width * hyt;
                win_data_dat = (unsigned char *)TaskAlloc(bmap->pid, win_data_size);

                sourceP = data_dat;
                for ( ix = 0; ix < hyt; ix++ )
                    {
                    destP = win_data_dat + ((hyt - ix - 1) * image_byte_width);
                    memset(destP, 0, image_byte_width);
                    for ( j = 0; j < width_on_disk; j++ )
                       *destP++ = *sourceP++;
                    }
#endif    /*  } AXMS  */
            }
            /* dither down to a 1 bit image */
            else
            {
                /* create an "pixel to index" table */
                /* not needed for windows, see coment above */
                for ( ii=0; ii<nColors; ii++ )
                {
                    t_c = &local_cmap.color[ii];
                    c = t_c->c;
                    m = t_c->m;
                    y = t_c->y;
                    k = t_c->k;
                    AxGetRealPixel(2, c,m,y,k, 255, &cmapper[ii]);
                }

                /* fake an XImage to create a source for the dither */
                imageP->bits_per_pixel = bmap->image.depth;
                imageP->xoffset = 0;
                imageP->bitmap_pad = imageP->bitmap_unit = 16;
                imageP->byte_order = MSBFirst;
                imageP->bitmap_bit_order = MSBFirst;
                imageP->obdata = NULL;
                imageP->format = ZPixmap;
                imgData = (unsigned char*)imageP->data;
                data_size = width_on_disk * hyt;
                data_size = width_on_disk * hyt;
                data_size = width_on_disk * hyt;
                for ( ix=0; ix<data_size; ix++ )
                   {
                   *imgData = cmapper[  (int) *imgData ];
                   imgData++;
                   }

                /* get the 'white' pixel. */
                AxGetRealPixel(0,255,255,255,0,255,&cindex);

                /* dither the color image in place */
                mask_size = monoByteWid*hyt;
                imgData = mdataP = (unsigned char*)TaskCalloc(bmap->pid,
                                               mask_size,sizeof(char));
                cDataP = (unsigned char*)imageP->data;
                if ( bmap->hasMask )
                {
                    maskP = (unsigned char*)mask->data;
                    for (ii=0; ii<hyt; ii++)
                    {
                        /* if not white and mask on, set to black */
                        for (jj=0; jj<wid; jj++)
                        {
                            if (((cDataP[jj]) != cindex) &&
                                 (maskP[(jj>>3)] & (1 << (7 - (jj%8)))))
                                    mdataP[(jj>>3)] |= (1 << (7 - (jj%8)));
                        }
                        cDataP += width_on_disk;
                        mdataP += monoByteWid;
                        maskP += monoByteWid;
                    }
                    TaskFree(bmap->pid,mask->data);
                    bmap->hasMask = FALSE;
                }
                else
                {
                    for (ii=0; ii<hyt; ii++)
                    {
                        /* if not white, set to black */
                        for (jj=0; jj<wid; jj++)
                        {
                            if ((cDataP[jj]) != cindex)
                                mdataP[(jj>>3)] |= (1 << (7 - (jj%8)));
                        }
                        cDataP += width_on_disk;
                        mdataP += monoByteWid;
                    }
                }

                /* the color image is replaced */
                TaskFree(bmap->pid,imageP->data);
                imageP->data = (char*)imgData;
                imageP->format = XYBitmap;
                imageP->bits_per_pixel = imageP->depth = 1;
                imageP->bytes_per_line = monoByteWid;
#if AXMS
            /* create the monochrome bitmap handle for Windows. */
                imageP->bitmap_handle =
                        MS_CreateBitmap(wid, hyt, 1 ,1, imageP->data);
                mask->bitmap_handle =
                        MS_CreateBitmap(wid, hyt, 1 ,1, imageP->data);
                bmap->hasMask = TRUE;
#endif
            }
        }
    }

#if AXMS
    /* we need to build a color table from the cmyk colors in the image */
    if ( bmap->display && imageP->depth == 8 )
        {
        AxWinCreate8BitDIB(imageP, win_data_dat, win_data_size);
        TaskFree(bmap->pid, win_data_dat);
        }
    else if ( bmap->display && imageP->depth > 8 )
        {
        /* make sure we flip (and pad) the data because the "negative height"
         * method doesn't seem to be reliable (thanks Microsoft!). -donato
         */
        data_size = hyt * imageP->bytes_per_line;
        win_data_dat = (unsigned char *)TaskAlloc(bmap->pid, data_size);
        sourceP = imageP->data;
        for ( ii = 0; ii < hyt; ii++ )
            {
            destP = win_data_dat + ((hyt - ii - 1) * imageP->bytes_per_line);
            memset(destP, 0, imageP->bytes_per_line);
            memcpy( destP, sourceP, imageP->bytes_per_line );
            sourceP = sourceP + imageP->bytes_per_line;
            }
        TaskFree(bmap->pid, imageP->data);
        imageP->data = win_data_dat;
        AxWinCreate16BitDIB(imageP, data_size);
        }
#endif

    if ( local_cmap.n_colors >0 && local_cmap.color )
        {
        for ( ii=0; ii<local_cmap.n_colors; ii++ )
            {
            t_c = &local_cmap.color[ii];
            if (t_c->name )
                TaskFree(bmap->pid,t_c->name);
            }
        TaskFree(bmap->pid,local_cmap.color);
        }

    return(AOK);
}

/************************************************************************/
fetchbyte( ch,fp)
char *ch;
FILE *fp;
{
    static char *chars = NXLT("0123456789ABCDEF");
    int p;
    char *posP;
    int result;

    p = fgetc(fp);
    if (p =='\n' )
        p = fgetc(fp);
    if ( p == EOF )
        return(FALSE);
    posP = strchr(chars,p);
    while (posP == NULL)
    {
        p = fgetc(fp);
        if ( p == EOF )
            return(FALSE);
        posP = strchr(chars,p);
    }
    result = ((posP-chars)<< 4 );
    p = fgetc(fp);
    if ( p == EOF )
        return(FALSE);
    posP = strchr(chars,p);
    result = (result | ((posP-chars) & 0xF ));
    *ch = result;
    return( TRUE);
}

/************************************************************************/
void WideToMbStr(in_str, out_str)
tx_char *in_str;
mbchar  *out_str;
{
    tx_char *ip;
    mbchar *op;

#if BUILD_WIDE
    WToMbStr(in_str, out_str);
#else
    ip = in_str;
    op = out_str;
    while (*ip)
        *op++ = (mbchar)(*ip++);
    *op = 0;
#endif
}

/************************************************************************/
get_string_length(wstr,numcharsIn)
tx_char *wstr;
int numcharsIn;
{
    tx_char *inP, *outP;
    int ii, numcharsOut;

    numcharsOut = numcharsIn;
    inP = outP = wstr;
    for (ii=0; ii<numcharsIn; ii++)
    {
        if (*inP == '\\')
        {
            *outP = *++inP;
            if (*outP == 'n')
                *outP = '\n';
            numcharsOut--;
        }
        else
            *outP = *inP;

        inP++;outP++;
    }
    wstr[numcharsOut] = 0;
    return(numcharsOut);
}

/************************************************************************/
void get_color_name(pid,color,buf,len,info)
int pid;
color_type *color;
char *buf;
int len;
AFileInfo *info;
{
    int  size, numchars;
    tx_char *wstr;

    size = (len+1)*sizeof(tx_char);
    wstr = (tx_char *)TaskAlloc(pid,size);
    numchars = NormalizeIn16((mbchar *)buf, wstr, (len+1), info);
    numchars = get_string_length(wstr, numchars);
    color->name = (char *)TaskAlloc(pid, size);
    WideToMbStr(wstr, color->name);

    /* done */
    TaskFree(pid, wstr);
}

/************************************************************************/

int
FSRev2ImToImage(bmap, fp)
bitmap_struct *bmap;
FILE *fp;
{
#if AX_X11
    return(FSRev2ImToXImage(bmap,fp));
#else
    return(FSRev2ImToWinImage(bmap,fp));
#endif
}

static int
FSRev1ImToImage(bmap, fp)
bitmap_struct *bmap;
FILE *fp;
{
#if AX_X11
    return(FSRev1ImToXImage(bmap, fp));
#else
    return(FSRev1ImToWinImage(bmap, fp));
#endif
}

ElfbReadGlomBitmap(name, path)
char *name;
char *path;
{
    FILE *fp, *fout;
    char *buf;
    int sz;

    /* open the image source file (already open if it's in the glom file) */
    fp = AxFindInGlom(name, &sz);
    if ( fp == NULL )
        return ( !AOK );
    buf = (char *)TrueMalloc(sz);
    if ( fread(buf, sz, 1,fp) != 1 )
    {
        TrueFree(buf);
        return ( !AOK );
    }

    /* write out */
    fout = fopen(path, WRITE_TRUNC_STR);
    if ( fout == NULL )
    {
        TrueFree(buf);
        return ( !AOK );
    }
    if ( fwrite(buf, sz, 1, fout)  != 1 )
    {
        TrueFree(buf);
        return(!AOK );
    }
    if ( fclose(fout) != 0 )
    {
        TrueFree(buf);
        return(!AOK );
    }

    /* done */
    TrueFree(buf);

    return(AOK);
}



AxDestroyBitmap(bmap)
bitmap_struct *bmap;
{
    int ii;
    color_type *t_c;

    if ( bmap->cmap.n_colors >0 &&  bmap->cmap.color )
    {
        for ( ii=0; ii< bmap->cmap.n_colors; ii++ )
        {
            t_c = & bmap->cmap.color[ii];
            if (t_c->name )
                TaskFree(bmap->pid,t_c->name);
        }
        TaskFree(bmap->pid, bmap->cmap.color);
    }
    if (bmap->image.data)
        TaskFree(bmap->pid,bmap->image.data);
    if (bmap->mask.data)
        TaskFree(bmap->pid,bmap->mask.data);
}


get_cmap_from_image(bmap)
bitmap_struct *bmap;
{
#if AX_X11
    XWindowAttributes win_info;
    cmap_type cmap;
    int i, c, m, y, k, nc, sz;
    XColor *colors;
    Colormap wcmap;
    Pixel r, g, b, r1, g1, b1;

    /* see if a strandard cmap is being used */
    wcmap = AxGetWindowColormap();
    if ( wcmap )
    {
        nc = AxVisual->map_entries;
        sz = sizeof(XColor) * nc;
        colors = (XColor *) TaskAlloc(bmap->pid,sz);

        for ( i=0; i<nc; i++ )
            colors[i].pixel = i;
    }
    else
    {
        XGetWindowAttributes (Dpy, DefaultRootWindow(Dpy), &win_info);

        if (!win_info.colormap)
            return(!AOK);
        if (win_info.visual->class == TrueColor )
            return(!AOK);
        nc = win_info.visual->map_entries;
        wcmap = win_info.colormap;

        sz = sizeof(XColor) * nc;
        colors = (XColor *) TaskAlloc(bmap->pid,sz);

        if ( win_info.visual->class == DirectColor )
        {
            r = g = b = 0;
            r1 = lowbit(win_info.visual->red_mask);
            g1 = lowbit(win_info.visual->green_mask);
            b1 = lowbit(win_info.visual->blue_mask);
            for ( i=0; i<nc; i++ )
            {
                colors[i].pixel =  r|g|b;
                r += r1;
                if ( r > win_info.visual->red_mask )
                    r = 0;
                g += g1;
                if ( g > win_info.visual->green_mask )
                    g = 0;
                b += b1;
                if ( b > win_info.visual->blue_mask )
                    b = 0;
            }
        }
        else
        {
            for ( i=0; i<nc; i++ )
                colors[i].pixel = i;
        }
    }

    if ( !wcmap )
        return(!AOK);

    XQueryColors(Dpy, wcmap, colors, nc);

    cmap.n_colors = nc;
    sz = nc * sizeof(color_type);
    cmap.color = (color_type *)TaskAlloc(bmap->pid, sz);
    for (i=0; i<nc; i++)
    {
        rgb2cmyk(colors[i].red, colors[i].green, colors[i].blue,
                 &c, &m, &y, &k, 8);
        cmap.color[i].see_thru = TRUE;
        cmap.color[i].ink_type = 0;
        cmap.color[i].c = c;
        cmap.color[i].m = m;
        cmap.color[i].y = y;
        cmap.color[i].k = k;
        cmap.color[i].name = (char *)TaskAlloc(bmap->pid,9);
        sprintf(cmap.color[i].name, NXLT("color%03d"), i);
    }
    bmap->cmap.n_colors = cmap.n_colors;
    bmap->cmap.color    = cmap.color;
    TaskFree(bmap->pid, colors);
#endif
    return(AOK);
}

static java_image_t *
getJmapFromXimage(ximg, back_is_transparent)
XImage *ximg;
int back_is_transparent;
{
	bitmap_struct bmap;
	java_image_t *jmap;
	register int ix;
	int size;

	/* copy image data to task memory */
	MEMSET(&bmap, 0, sizeof(bitmap_struct));
	bmap.pid     = ThimMyPid();
	bmap.display = TRUE;
	size = ximg->bytes_per_line*ximg->height;
	bmap.image = *ximg;
	bmap.image.data = (char*)TaskAlloc(bmap.pid,size);
	memcpy((void*)bmap.image.data,(void*)ximg->data,size);
	if ( ximg->depth > 1 ) /* create a colormap */
		create_cmap(bmap.pid, &(bmap.cmap));
	update_image_from_cmap(&(bmap.image), &(bmap.cmap), bmap.pid);
    
	/* convert the bmap to a javamap */
	jmap = AXtoJAVA(&bmap, 1); 

	/* force transparency to opacity (on request) */
	if ( back_is_transparent == FALSE )
	{
		if ( jmap->colormap.alpha )
		{
			for ( ix=jmap->colormap.n_colors-1; ix>=0; ix-- )
				if ( jmap->colormap.alpha[ix] != 255 )
					jmap->colormap.alpha[ix] = 255;
		}
		jmap->colormap.transparent_index = -1;
	}

	/* done */
	AxDestroyBitmap(&bmap);
	return(jmap);
}

data_ptr
AxGetJavaBitmapFromXimage(ximg, back_is_transparent)
XImage *ximg;
int back_is_transparent;
{
    java_image_t *jmap;
    data_ptr mapP;

	jmap = getJmapFromXimage(ximg, back_is_transparent);
#if 0
    printJmapData(jmap);
#endif

    /* convert the javamap to an ELF datum */
    mapP = JAVAtoELF(jmap);

	/* done */
    AxDestroyJAVA(jmap);

    return mapP;
}

data_ptr
getJavaBitmapForMaskBits(ximg, x, y)
XImage *ximg;
int x, y;
{
    java_image_t *jmap;
    data_ptr mapP;

	jmap = getJmapFromXimage(ximg, FALSE);

    /* convert the javamap to an ELF datum */
    mapP = jmapToElF(jmap, x, y);

	/* done */
    AxDestroyJAVA(jmap);

    return(mapP);
}

data_ptr
getJavaBitmap (ximg)
XImage *ximg;
{
	return(AxGetJavaBitmapFromXimage(ximg, FALSE));
}

#if AXMS
void
axmsFlipXImage(XImage *imageP)
{
    int i, j, bwid, image_size;
    unsigned char *new_data, *sourceP, *destP;

    bwid = imageP->bytes_per_line;
    image_size = bwid * imageP->height;
    new_data = (unsigned char *)TaskAlloc(ThimMyPid(), image_size);

    /* flip the data from DIB format to "normal" bitmap format. */
    sourceP = imageP->data;
    for ( i=imageP->height-1; i >= 0; i-- )
        {
        destP = new_data + (i * bwid);
        for ( j=0; j < bwid; j++ )
            *destP++ = *sourceP++;
        }

    TaskFree(ThimMyPid(), imageP->data);
    imageP->data = new_data;
}
#endif


printJmapData (jmap)
java_image_t *jmap;
{
    static int counter = 0;
    char filename[128];
    FILE *fp;
    int ix, jx;

    sprintf(filename,"/tmp/jmap-%d",counter++);
    fp = fopen(filename,"w");
    if (NULL == fp)
        return;
    fprintf(fp, "type = %d\n", jmap->type);
    fprintf(fp, "compression type = %d\n", jmap->compression_type);
    fprintf(fp, "ncolors = %d\n", jmap->colormap.n_colors);
    for (ix = 0; ix < jmap->colormap.n_colors; ++ix)
        fprintf(fp, "color[%d] = %d, %d, %d, %d\n", ix,
                jmap->colormap.r[ix], jmap->colormap.g[ix], jmap->colormap.b[ix],
                jmap->colormap.alpha?jmap->colormap.alpha[ix]:255);
    fprintf(fp, "transparent index = %d\n", jmap->colormap.transparent_index);
    fprintf(fp, "scan width = %d\n", jmap->scan_wid);
    fprintf(fp, "image width in pixels = %d\n", jmap->wid);
    fprintf(fp, "image height in pixels = %d\n", jmap->hyt);
    fprintf(fp, "image data: \n");
    for (ix = 0; ix < jmap->hyt; ++ix) {
        for (jx = 0; jx < jmap->scan_wid; ++jx)
            fprintf(fp, "%-2x ", jmap->data[ix*jmap->scan_wid+jx]);
        fprintf(fp, "\n");
    }
    fclose(fp);
}

