/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmpopup.c
   Applix ASC version: 43
   Date: Mon Nov 16 13:53:42 1998
*/
#include <basicwin.h>
#include <ax_xm.h>
#include <aos.h>
#include <xmlib.h>
#include <xmpixel.h>
#include <eldata.h>
#include <scac.h>
#include <thimble.h>

extern AXWidget_popup RightButtonArmed;
extern AXWidget_popup PopupArmed;

extern int CharWids[256];
extern int AxCharAbove, AxCharBelow, AxCharHeight;
extern int NWidth, N2Width, N3Width;

extern int AxOpenLook;
extern int AxBigFont;
extern unsigned long BackPixel;
extern GC ForeGC, BackGC, DisabledGC;
extern GC SunGC, ShadowGC;
extern GC olTextGC, olButtonGC, olDisabledButtonGC;
extern GC olClrSunButtonGC, olClrShadowButtonGC;
extern GC olClrBackButtonGC, olClrRecessButtonGC;
extern Pixmap MenuCheckPixmap;
extern XFontStruct *Txfont, *MiniTxfont, *Grafont;
extern Window RootWin;
extern Display *Dpy;

extern data_ptr AxAddDataToArray();
extern void AdjustPulldownOrigin();

extern int ThimExposuresMayBeComing;

void PaintPopupPulldown();
void PaintPixmapPopupPulldown();
void PaintRowColumnPulldown();
extern AXWidget WidgetOfXWindow();

extern pixcache_t *Pixmapper();

typedef struct rm_info {
    int n;
    char **strings;
    char **macros;
    char **argArray;
    int  *ungrey;
} rm_info;


void
setButtonDimensions(w, map_wid, map_hyt, margin)
AXWidget_popup w;
int *map_wid, *map_hyt, *margin;
{
    pixcache_t *pix;
    int i;

    /* get the margin size based on window manager emulation mode */
    *margin = 3;

    /* get the largest button dimensions */
    pix = Pixmapper(w->buttons[0]->name);
    if ( !pix )
        *map_wid = *map_hyt = 0;
    else
    {
        *map_wid = pix->width;
        *map_hyt = pix->height;
    }
    for ( i=1; i<w->num_buttons; i++ )
    {
        pix = Pixmapper(w->buttons[0]->name);
        if ( !pix )
            continue;

        if ( *map_wid < pix->width )
            *map_wid = pix->width;
        if ( *map_hyt < pix->height )
            *map_hyt = pix->height;
    }
    *map_wid += *margin * 2;
    *map_hyt += *margin * 2;
}

void
PaintPixmapPopupPulldown (w,mode,eventP)
register AXWidget_popup w;
int mode;
XEvent *eventP;
{
    static int map_hyt, map_wid, margin;
    XSetWindowAttributes atts;
    Window child;
    int by,bx;  /* button pos within the menu */
    long att_mask;
    int y, ix, tmp, x_adj, y_adj;
    int real_x, real_y;
    static int show_checkmark = FALSE;
    static int check_at;
    XmRowColumnCallbackStruct data;
    static char **strs;
    int borderwidth;
    static int old_value;
    int new_value;
    GC topGC, bottomGC, fillGC, gc;
    char menubutton;
    char bitmapName[256];
    pixcache_t *pxc;

    switch ( mode )
    {
        /* post the popup */
        case OPTION_NEW:
            /* determine the location of the checkmark (if any) */
            if ( (w->value != -1) && (w->value < w->num_buttons) )
            {
                show_checkmark = TRUE;
                check_at = w->value;
            }
            else
                show_checkmark = FALSE;
            old_value = w->value;

            /* get the dimensions of the pixmap elements */
            setButtonDimensions(w, &map_wid, &map_hyt, &margin);

            /* get the total width and height of the pulldown */
            w->height = 2*margin + w->num_buttons * map_hyt;
            w->pulldown_width = 2*margin + map_wid;

            /* set the popup flags and position */
            att_mask = CWOverrideRedirect | CWSaveUnder | CWBackPixel;
            atts.override_redirect = TRUE;
            atts.save_under = FALSE; /* SAVE_UNDER; */
            atts.background_pixel = BackPixel;
            XTranslateCoordinates(Dpy,WIN_OF(w),RootWin,
                                  w->xpos,w->ypos,
                                  &real_x, &real_y,
                                  &child);

#if AX_X11
			/* Adjust real_x, real_y to make sure pulldown is on-screen */
			AdjustPulldownOrigin(&real_x, &real_y,
								 w->pulldown_width, w->height,
								 &x_adj, &y_adj);
			bx += x_adj;
			by += y_adj;
			w->xpos += x_adj;
			w->ypos += y_adj;
#endif

            borderwidth = 0;

#if AXMS
            AxCreateWinPopupWindow(w, real_x, real_y);
#else            
            /* create and map the window */
            w->window = AxCrtWindow(Dpy, RootWin,
                                    real_x, real_y,
                                    w->pulldown_width, w->height,
                                    borderwidth,
                                    CopyFromParent, InputOutput,
                                    CopyFromParent,
                                    att_mask, &atts);
            XMapWindow(Dpy,w->window);
#endif
            PopupArmed = w;

            /* draw the shadow/border on the popup */
            AxDrawShadow(w->window,SunGC,ShadowGC,
                         2,0,0,
                         w->pulldown_width,w->height,
                         OLSHADOW_DRAW);

            /* draw the options on the pulldown */
            for ( ix=0; ix<w->num_buttons; ix++ )
            {
                tmp = show_checkmark && ix==check_at;
                /* draw the pixmap */
                strcpy(bitmapName, w->buttons[ix]->name);
                pxc = Pixmapper(bitmapName);
                if ( !pxc )
                    return;
                XCopyArea(Dpy, pxc->pix, w->window, ForeGC,
                          0, 0, pxc->width, pxc->height,
                          2*margin, 2*margin + ix*map_hyt);
                
                if ( ix == w->value ) {
                    AxDrawShadow(w->window,SunGC,ShadowGC,2,
                                 margin, margin+ix*map_hyt,
                                 map_wid, map_hyt,
                                 OLSHADOW_NODRAW);
                }
            }
        break;

        case OPTION_MOVE:
#if AXMS
            /* correct position already in xevent struct */
            bx = eventP->xmotion.x;
            by = eventP->xmotion.y;
#else            
            bx = eventP->xmotion.x - w->xpos;
            by = eventP->xmotion.y - w->ypos;
#endif

            /* convert the position to an index */
            if ( bx < margin || bx > w->pulldown_width-margin )
                new_value = -1;
            else if ( by < margin || by > w->height-margin )
                new_value = -1;
            else
            {
                new_value = (by - margin)/map_hyt;
                if ( new_value >= w->num_buttons )
                    new_value = -1;
            }

            /* new value valid? show it */
            if ( (new_value != old_value) && (new_value != -1) )
            {
                /* valid new selection has been made */
                if (old_value == -1)
                {
                    /* currently active (default) selection is showing */
                    old_value = w->value;
                }

                if (old_value != -1)
                {
                    AxDrawShadow(w->window,BackGC,BackGC,2,
                                 margin, margin + map_hyt*old_value,
                                 map_wid, map_hyt,
                                 OLSHADOW_NODRAW);
                }
                
                AxDrawShadow(w->window,SunGC,ShadowGC,2,
                             margin, margin + map_hyt*new_value,
                             map_wid, map_hyt,
                             OLSHADOW_NODRAW);
            }

            /* new value illegal? display default & shut off choice pending */
            if ( (new_value == -1) && (old_value != -1) )
            {
                /* display default & shut off current "choice pending" */
                AxDrawShadow(w->window,BackGC,BackGC,2,
                             margin, margin + old_value*map_hyt,
                             map_wid, map_hyt,
                             OLSHADOW_NODRAW);
                
                if (w->value != -1)
                {
                    AxDrawShadow(w->window,SunGC,ShadowGC,2,
                                 margin, margin + w->value*map_hyt,
                                 map_wid, map_hyt,
                                 OLSHADOW_NODRAW);
                }
            }
            old_value = new_value;
        break;

        /* button up */
        case OPTION_DONE:
            /* call the handling routine */
            if (old_value != -1)
                w->value = old_value;
            data.data = (char*)w->value;
            if ( w->callback )
                AxCallback(w->callback,w,w->callbacktag,&data);

            /* reset state */
            ThimExposuresMayBeComing = TRUE;
            ResetActive();

            /* destroy the transient widget */           
            AxXtDestroyWidget(w);
        break;
    }
}

/*************************************************************************
function: PaintPopupPulldown - Create and manage popup pulldowns
Currently used only on DC ruler and GE palette.
Mode:
    0 - OPTION_NEW - Initial display
    1 - OPTION_MOVE - Pointer motion
    2 - OPTION_DONE - Button up
*/
void
PaintPopupPulldown (w,mode,eventP)
register AXWidget_popup w;
int mode;
XEvent *eventP;
{
    XSetWindowAttributes atts;
	Window child;
	int height = 0, width;
	int by,bx;  /* button pos within the menu */
    long att_mask;
    int y, ix, tmp;
    int real_x, real_y, x_adj, y_adj;
    static int show_checkmark = FALSE;
    static int check_at;
    XmRowColumnCallbackStruct data;
    static char **strs;
    int borderwidth;
    static int old_value;
    int new_value;
    static int radioWid = 0;
    static int choiceHeight = 0;
    GC topGC, bottomGC, fillGC, gc;
    char menubutton;
    int GEarrowsPulldown;

    if ( !w->display )
        return;

    w->mapped = TRUE;

    /* ignore activity on empty lists */
    if ( w->num_buttons == 0 )
    {
        ResetActive();
        return;
    }

    /* special 1D pulldowns for GE palette */
    if ( w->gePaletteButtonType == 7 )
    {
        PaintPixmapPopupPulldown(w, mode, eventP);
        return;
    }

/* compute height, 3above,3below,1sep */

    height += w->num_buttons * (AxCharHeight + 7);
    if ( height == 0 )
        return;
    height += 7;    /* area below */

/* Handle button up case */

    if ( mode == OPTION_DONE ) {
        if (old_value != -1)
            w->value = old_value;
        data.data = (char*)w->value;
        if ( w->callback )
            AxCallback(w->callback,w,w->callbacktag,&data);
        ThimExposuresMayBeComing = TRUE;
        ResetActive();       
        AxXtDestroyWidget(w); /* transient widget, take it away */
                return;
    }


/* determine which one is currently selected */

    if ( mode != OPTION_NEW )
        goto have_window;

/* Compute width required, never less than 7n wide */

    if ( (w->value != -1) && (w->value < w->num_buttons) ) {
        show_checkmark = TRUE;
        check_at = w->value;
    }
    else
        show_checkmark = FALSE;
    old_value = w->value;
    width = BasicStrlen(NXLT("nnnnnnn"));
    for ( ix = 0; ix < w->num_buttons; ix++ ) {
        tmp = BasicStrlen(w->buttons[ix]->name);
        width = MAX(width,tmp);
    }
    width += 2*(NWidth + 2);

    if ( show_checkmark )
        width += 13;    /* room for checkmark */

    w->pulldown_width = width;

    att_mask = CWOverrideRedirect | CWSaveUnder | CWBackPixel;
    atts.override_redirect = TRUE;
    atts.save_under = FALSE; /* SAVE_UNDER; */
    atts.background_pixel = BackPixel;

    XTranslateCoordinates(Dpy,WIN_OF(w),RootWin,
                            w->xpos,w->ypos,
                            &real_x, &real_y,
                            &child);

#if AX_X11
	/* Adjust real_x, real_y to make sure pulldown is on-screen */
	AdjustPulldownOrigin(&real_x, &real_y,
						 w->pulldown_width, height,
						 &x_adj, &y_adj);
	bx += x_adj;
	by += y_adj;
	w->xpos += x_adj;
	w->ypos += y_adj;
#endif

    borderwidth = 0;

#if AXMS
    AxCreateWinPopupWindow(w, real_x, real_y);
#else
    w->window = AxCrtWindow(Dpy,RootWin,
                    real_x, real_y,
                    w->pulldown_width,height,borderwidth,CopyFromParent,
                    InputOutput,CopyFromParent,att_mask,&atts);
    XMapWindow(Dpy,w->window);
#endif
    PopupArmed = w;

/* paint contents */

have_window:
    if ( mode == OPTION_NEW ) {
        AxDrawShadow(w->window,SunGC,ShadowGC,2,0,0,
                w->pulldown_width,height,OLSHADOW_DRAW);
    }
    else if ( mode == OPTION_MOVE ) {

        bx = eventP->xmotion.x - w->xpos;
        by = eventP->xmotion.y - w->ypos;

        if ( by < 0 )
            new_value = -1;
        else
            new_value = by / (AxCharHeight+7);
        if ( (bx <= 0) || (bx >= w->pulldown_width) ||
                            (new_value >= w->num_buttons) )
            new_value = -1;

        if ( (new_value != old_value) && (new_value != -1) ) {
            /* valid new selection has been made */
            if (old_value == -1) {
                /* currently active (default) selection is showing */
                old_value = w->value;
                /* if old_value is still -1, nothing is currently active */
            }
            if (old_value != -1) {
                AxDrawShadow(w->window,BackGC,BackGC,2,5,
                             old_value * (AxCharHeight+7) + 5,
                             w->pulldown_width-10,AxCharHeight+6,
                             OLSHADOW_NODRAW);
            }
            AxDrawShadow(w->window,SunGC,ShadowGC,2,5,
                         new_value * (AxCharHeight+7) + 5,
                         w->pulldown_width-10,AxCharHeight+6,OLSHADOW_NODRAW);
        }
        if ( (new_value == -1) && (old_value != -1) ) {
            /* need to display default & shut off current "choice pending" */
            AxDrawShadow(w->window,BackGC,BackGC,2,5,
                         old_value * (AxCharHeight+7) + 5,
                         w->pulldown_width-10,AxCharHeight+6,OLSHADOW_NODRAW);
            if (w->value != -1) {
                AxDrawShadow(w->window,SunGC,ShadowGC,2,5,
                             w->value * (AxCharHeight+7) + 5,
                             w->pulldown_width-10,AxCharHeight+6,
                             OLSHADOW_NODRAW);
            }
        }
        old_value = new_value;
        return;
    }

    /* OPTION_NEW */
    y = 0;
    for ( ix = 0; ix < w->num_buttons; ix++ ) {
        if ( show_checkmark ) {
			tmp = 13;
            if ( ix == check_at ) {
                XCopyArea(Dpy,MenuCheckPixmap,w->window,ForeGC,
                          0,0,16,16,CharWids['n']-2,y+5);
            }
		}
		else {
			tmp = 0;
		}
		xmDrawString(w, w->window, ForeGC, 7 + CharWids[' '] + tmp,
					 y + AxCharAbove + 8, w->buttons[ix]->name,
					 strlen(w->buttons[ix]->name), w->textShadow,
					 w->textColor, w->widgetColor);
		
        if ( ix == w->value ) {
            AxDrawShadow(w->window,SunGC,ShadowGC,2,5,
                         w->value * (AxCharHeight+7) + 5,
                         w->pulldown_width-10,AxCharHeight+6,OLSHADOW_NODRAW);
        }
        y += AxCharHeight + 7;
    }
}

/*
function: axolAdjPopupPos - openlook, obsolete
*/

axolAdjPopupPos(xp, yp, palette, palettestr)
int *xp, *yp;
int palette; /* bool */
char *palettestr;
{
    *xp = 0;
    *yp = 0;
    return;
}

/****************************************************************************
 *
 * module:  
 */
xmDestroyRmInfo(info)
rm_info *info;
{
    int i;
    for ( i=0; i<info->n; i++ )
    {
        free(info->strings[i]);
        free(info->macros[i]);
        if (info->argArray[i])
            free(info->argArray[i]);
    }
    free(info->strings);
    free(info->macros);
    if (info->argArray)
        free(info->argArray);
    free(info->ungrey);
    free(info);
}

/****************************************************************************
 *
 * module:  
 */
xmSetRmInfo(w, n, strings, macros, argArray, ungrey)
register AXWidget_popup w;
int n;
char **strings;
char **macros;
char **argArray;
int *ungrey;
{
    rm_info *info;
	if ( w == (AXWidget_popup)NULL )
		return;
    /* get rid of old rminfo */
    if ( w->rminfo )
        xmDestroyRmInfo(w->rminfo);

    /* copy new rminfo onto widget */
    if ( n <= 0 )
        info = (rm_info *)0;
    else
    {
        info = (rm_info *)malloc(sizeof(rm_info));
        info->n = n;
        info->strings = strings;
        info->macros = macros;
        info->argArray = argArray;
        info->ungrey = ungrey;
    }
    w->rminfo = ((char *)info);
}

/****************************************************************************
 *
 * module:  
 */
SetRmInfoFrmDatPtr(w, arrayP)
register AXWidget_popup w;
data_ptr arrayP;
{
    data_ptr infoP, valP;
    int i, n_elm;
    char *str;
    char **strings;
    char **macros;
    char **argArray;
    int *ungrey;

    if ( AxIsArray(arrayP) == FALSE )
        return(!AOK);

    n_elm = AxArraySize(arrayP);
    strings = (char **)TaskAlloc(0, n_elm * sizeof(char *));
    macros  = (char **)TaskAlloc(0, n_elm * sizeof(char *));
    argArray  = (char **)TaskAlloc(0, n_elm * sizeof(char *));
    ungrey  = (int *)  TaskAlloc(0, n_elm * sizeof(int));
    for ( i=0;i<n_elm;i++ )
    {
        infoP = AxArrayElement(arrayP, i );
        if (infoP )
        {
            valP = AxArrayElement(infoP, 0);            /* displayed name */
            if (valP && AxIsString(valP))
                str = (char *)AxStrFromDataPtr(valP);
            else
                str = (char *)0;
            if ( !str)
                strings[i] = (char *)0;
            else
            {
                strings[i] = (char *)TaskAlloc(0, strlen(str)+1);
                strcpy(strings[i], str);
            }
            valP = AxArrayElement(infoP, 1);            /* macro name */
            if (valP && AxIsString(valP))
                str = (char *)AxStrFromDataPtr(valP);
            else
                str = (char *)0;
            if ( !str)
                macros[i] = (char *)0;
            else
            {
                macros[i] = (char *)TaskAlloc(0, strlen(str)+1);
                strcpy(macros[i], str);
            }
            valP = AxArrayElement(infoP, 2);            /* command line */
            if (valP && AxIsString(valP))
                str = (char *)AxStrFromDataPtr(valP);
            else
                str = (char *)0;
            if ( !str)
                argArray[i] = (char *)0;
            else
            {
                argArray[i] = (char *)TaskAlloc(0, strlen(str)+1);
                strcpy(argArray[i], str);
            }
            valP = AxArrayElement(infoP, 3);            /* ungrayed item? */
            if (valP)
            ungrey[i] = valP ? (AxIntFromDataPtr(valP)?TRUE:FALSE) : FALSE;
        }
    }
    xmSetRmInfo(w, n_elm, strings, macros, argArray, ungrey);
}

static void
xmObjSetRmInfo (w)
register AXWidget_popup w;
{
	int retCode, pid;
	data_ptr retP=NULL, item;
	int num, ix;
	rm_info *rminfo;

	if (!w || !w->rminfo)
		return;
	rminfo = (rm_info *)w->rminfo;

	pid = w->taskOwner;
	AxActiveTask = TaskInfo[pid].uid;
	VfyContext(pid);
	ElfbSetTaskMaster(AxActiveTask);    /* connect called macro to self */

	num = (rminfo) ? rminfo->n : 0;

	for (ix = 0; ix < num; ++ix) {		/* get active status of menu items */
		item = ELF_CREATE_STRING(rminfo->macros[ix]);
		retP = NULL;
		retCode = ElfCallByName(NXLT("BUTTON3_MENU_STATE_RQST$"),
								&retP, 2, w->elfObject, item);
		if (retCode == 0 && retP)
			rminfo->ungrey[ix] = AxBoolFromDataPtr(retP);
		AxFreeData(retP);
		ELF_DATA_FREE(item);
	}

	return;
}

/****************************************************************************
 *
 * module:  
 */
char *
xmGetRmInfo(w)
register AXWidget_popup w;
{
    Drawable xwin;
    int atm_win, pid, x_off, y_off;
    int x, y;

    /* get the window */
    if ( w->type == W_WORK )
    {
        xwin = AxGetDrawableFromWidget(w->parent, &x_off, &y_off);
        atm_win = ATMWindowFromXWindow(xwin);
        if ( atm_win )
        {
            pid = PidFromATMWin(atm_win);
            VfyContext(pid);
            switch ( AppIdFromATMWin(atm_win) )
            {
                case APGE:
                    GrSetRmInfo(w, pid);
                break;

                case APDC:          /* WP */
                case APUCM:         /* Macro Editor */
                case APEQN:         /* Equation Editor */
                case APHTML:        /* HTML Authoring System */
                    WpSetRmInfo(w, pid);
                break;

                case APSS:
                    SSSetRmInfo(w, pid);
                break;
            }
        }
    }

	if (w->elfObject) {		/* in object model ask object to supply
							   right mouse button menu info */
		xmObjSetRmInfo(w);
	}

    /* return the (modified?) right mouse info */
    return( (char *) w->rminfo);
}

/****************************************************************************
 *
 * module:  
 */
PaintRightButtonMenu(wp, w, mode, eventP)
register AXWidget_popup wp;
register AXWidget_popup w;
int mode;
XEvent *eventP;
{
    int retval, x_adj, y_adj;
    int an, value, real_x, real_y, ix, err, pid;
	int bx, by;
    static int n, map_wid, map_hyt;
    XSetWindowAttributes atts;
    long att_mask;
    data_ptr retP, itemP;
	int len;
    Window child;
    Arg args[7];
    static rm_info *info = (rm_info *) NULL;
    data_ptr macroArgs;
	GC tgc;

    switch ( mode )
    {
        /* post the popup */
        case OPTION_NEW:
            info = (rm_info *)xmGetRmInfo(wp);
            if ( !info || info->n <= 0 )
                return(FALSE);

            map_hyt = info->n * (AxCharHeight + 7) + 7;
            map_wid = BasicStrlen(NXLT("nnnnnnn")) + 2*(NWidth+2);
            for ( ix=0; ix<info->n; ix++ ) {
                map_wid = MAX(map_wid, 
							  BasicStrlen(info->strings[ix])+2*(NWidth+2));
			}

            an = 0;
            XtSetArg(args[an], XmNx, eventP->xbutton.x+1); an++;
            XtSetArg(args[an], XmNy, eventP->xbutton.y+1); an++;
            XtSetArg(args[an], XmNheight, map_hyt); an++;
            XtSetArg(args[an], XmNwidth, map_wid); an++;
            XtSetArg(args[an], XmNitemCount, info->n); an++;
            XtSetArg(args[an], XmNitems, info->strings); an++;
            XtSetArg(args[an], XXmNvalue, -1); an++;
		/*	XtSetArg(args[an], AXmNelfObject, wp->elfObject); an++; */
            w = (AXWidget_popup)CrtWidget(W_POPUP, wp, NXLT("rmpopup"));
            AddToWindowList(w);
            AxXtSetValues(w,args,an);
            RightButtonArmed = (AXWidget_popup)w;
            w->pulldown_width = map_wid;

		/*	w->taskOwner = wp->taskOwner; */
			
            /* set the popup flags and position */
            att_mask = CWOverrideRedirect | CWSaveUnder | CWBackPixel;
            atts.override_redirect = TRUE;
            atts.save_under = TRUE; /* SAVE_UNDER; */
            atts.background_pixel = BackPixel;
            retval = XTranslateCoordinates(Dpy,WIN_OF(w),RootWin,
                                            w->xpos,w->ypos,
                                            &real_x, &real_y,
                                            &child);

#if AX_X11
			/* Adjust real_x, real_y to make sure pulldown is on-screen */
			AdjustPulldownOrigin(&real_x, &real_y,
								 map_wid, map_hyt,
								 &x_adj, &y_adj);
			bx += x_adj;
			by += y_adj;
			w->xpos += x_adj;
			w->ypos += y_adj;
#endif
			
            /* create and map the window */
#if AXMS
        /* on Windows, check the return value of XTranslateCoordinates
         * because it may have failed if the window we are translating
         * from is not "findable", that is, not on some global WindowList.
         * the reason we need to "find" the window is that the equivalent
         * Windows call takes type HWND while XTranslateCoordinates is
         * passed type HDC (X Window). (see comments in x2win.c) -donato
         */
            if ( !retval )
                {
                AxWinTranslateCoordinates(w->parent->winid, HWND_DESKTOP,
                                    w->xpos, w->ypos, &real_x, &real_y);
			    }
            AxCreateWinPopupWindow(w, real_x, real_y);
#else
            w->window = AxCrtWindow(Dpy, RootWin,
                                    real_x, real_y,
                                    map_wid, map_hyt,
                                    0,
                                    CopyFromParent, InputOutput,
                                    CopyFromParent,
                                    att_mask, &atts);
            XMapWindow(Dpy,w->window);
			/* to prevent virtual window managers from switching screens */
			XGrabKeyboard(Dpy,w->window,False,GrabModeAsync,GrabModeAsync,
						  CurrentTime);
#endif
            /* draw the shadow/border on the popup */
            AxDrawShadow(w->window,SunGC,ShadowGC,
                         2,0,0,
                         map_wid,map_hyt,
                         OLSHADOW_DRAW);

            /* draw the strings */
            for ( by = ix = 0; ix < info->n; ix++ )
            {
				if (info->ungrey[ix]) {
					tgc = ForeGC;
				}
				else {
					tgc = DisabledGC;
					w->disabled = TRUE;     /* just for xmDrawString call..*/
				}
				xmDrawString(w, w->window, tgc, 7 + CharWids[' '],
							 by + AxCharAbove + 8, info->strings[ix],
							 strlen(info->strings[ix]), w->textShadow,
							 w->textColor, w->widgetColor);
				w->disabled = FALSE;
                by += AxCharHeight + 7;
            }
        break;

        case OPTION_MOVE:
#if AXMS
            /* correct position already in xevent struct */
            bx = eventP->xmotion.x;
            by = eventP->xmotion.y;
#else        
            bx = eventP->xmotion.x - w->xpos;
            by = eventP->xmotion.y - w->ypos;
#endif

            /* convert the position to an index */
            if ( bx < 0 || bx > w->pulldown_width )
                value = -1;
            else if ( by < 0 || by > w->height )
                value = -1;
            else
            {
                value = MAX(0, by - 5) / (AxCharHeight + 7);
                if ( value == w->num_buttons )
                    value = w->num_buttons-1;
                else if ( value > w->num_buttons )
                    value = -1;
            }

            /* if the new index is in range, update the selection rect */
            if ( value != w->value && value != (-1) )
            {
                /* if the last value was valid, erase the select rect */
                if ( w->value != -1 )
                {
                    AxDrawShadow(w->window,BackGC,BackGC,2,
                              5, w->value*(AxCharHeight+7)+5,
                              map_wid-10, AxCharHeight+6,
                              OLSHADOW_NODRAW);
                }

				if (info && info->ungrey[value]) {
					/* draw the new selection rect */
					AxDrawShadow(w->window,SunGC,ShadowGC,2,
								 5, value*(AxCharHeight+7)+5,
								 map_wid-10, AxCharHeight+6,
								 OLSHADOW_NODRAW);
				}
            }

            /* if the new index is not legal, remove the select rect */
            if ( value == (-1) && w->value != (-1) )
            {
                AxDrawShadow(w->window,BackGC,BackGC,2,
                             5, w->value*(AxCharHeight+7)+5,
                             map_wid-10, AxCharHeight+6,
                             OLSHADOW_NODRAW);
            }

            w->value = value;
        break;

        /* button up */
        case OPTION_DONE:
            /* call the handling routine */
            pid = THIMpid;
            if ( w->value != (-1) && info &&
                  info->macros[w->value] && 
                  info->ungrey[w->value] )
            {
				char method_name[256];
				
				if (w->parent && w->parent->elfObject) {
					strcpy(method_name, info->macros[w->value]);
					if (info->argArray[w->value]) {
						strcat(method_name, " ");
						strcat(method_name, info->argArray[w->value]);
					}
					itemP = ELF_CREATE_STRING(method_name);
					/* send event to its parent */
					CallWidget(w->parent->callbacktag, w->parent,
							   EV_BUT3_MENU_SELECT_,itemP);
					/* itemP is freed by CallWidget */
				}
				else {
					/* If specified, pass arguments (NULL is OK) to the macro. */
					if ( w->parent )
						AxSetThisPidAsActiveTask(w->parent->taskOwner);
					if (info->argArray[w->value])
						ElfbNewTask(info->macros[w->value], 
									info->argArray[w->value]);
					else
						ElfbNewTask(info->macros[w->value], ""); 
				}
				VfyContext(pid);
            }

            /* reset state */
            ThimExposuresMayBeComing = TRUE;
#if AX_X11
			XUngrabKeyboard(Dpy, CurrentTime);
#endif
            AxXtDestroyWidget(w);
            RightButtonArmed = NULL;
			info = (rm_info *) NULL;
        break;
    }
    return(TRUE);
}



/****************************************************************************
 *
 * module:  
 */
int
RightButtonEvent( w, mode, eventP)
register AXWidget_popup w;
int mode;
XEvent *eventP;
{
    int err;

    err = FALSE;
    if ( !w )
        return (FALSE);

    err = PaintRightButtonMenu(w, RightButtonArmed, mode, eventP);

    return(err);
}

/***************************************************************************
function: xmEventNeededForRightButtonPopup 
*/
int xmEventNeededForRightButtonPopup (eventP)
XEvent *eventP;
{
    AXWidget w;

    /* !! CURRENTLY WORKS ONLY FOR WP/ME/EQN/SS */

    if ( (eventP->type == ButtonPress) ||
                                    (eventP->type == MotionNotify) ||
                                    (eventP->type == ButtonRelease) ) {
        if ( RightButtonArmed ) {
            /* popup already up */
            return(TRUE);
        }

        w = WidgetOfXWindow(eventP->xany.window);

        /* !! The following line assumes that work windows will enable
         * !! right button popup. Note that w->rminfo is not set till
         * !! the right button event handler is called (*after* this). */
        if ( w && (w->type == W_WORK) &&
                                    (eventP->type == ButtonPress) &&
                                    (eventP->xbutton.button == Button3) ) {
            /* right button press in work window */
            return(TRUE);
        }
    }

    return(FALSE);
}

