/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmpulldn.c
   Applix ASC version: 47
   Date: Fri Jan  8 21:11:31 1999
*/
#include <ctype.h>
#include <X11/keysym.h>
#include <ax_xm.h>
#include <aos.h>
#include <xmlib.h>
#include <eldata.h>
#include <paripc.h>
#include <thimble.h>


extern AXWidget AxPulldownActive;/* non-zero while menubar pulldown active */
extern int AxPulldownStayup; /* if TRUE: menu pulldown is in stayup mode*/
extern int PointerHasMoved; /* TRUE if pointer has moved since last set to FALSE */
extern int AxExitPulldownStayupMode; /* if TRUE: ResetActive() ungrabs pointer */

extern int AxMenuItem;      /* current item in menuBar */
extern int AxPulldownItem;  /* current item in pulldown */
extern int AxCascadeItem;   /* current item in cascade */

extern int CharWids[256];
extern int AxCharAbove, AxCharBelow, AxCharHeight;
extern int NWidth, N2Width, N3Width;

extern int AxBigFont;
extern int AxButtonDown;
extern KeySym AxControlKeyDown;
extern unsigned long BackPixel;
extern GC ForeGC, BackGC, DisabledGC;
extern GC SunGC, ShadowGC, RecessGC;
extern GC olButtonGC;
extern XFontStruct *Txfont, *Grafont;
extern XFontStruct *AxDlgMainFont;
extern Window RootWin;
extern Display *Dpy;

extern int MenuUID;
extern int ShowAccelerators;
extern int AlwaysWide;
extern int AxMarkMenuMnemonics;
extern int UseChecksForToggles;
extern Pixmap CascadePixmap, MenuCheckPixmap;
extern int AxShiftMask, AxLockMask, AxControlMask, AxAltMask;
extern int AxNoWidgetColor;

extern int THIMhelpMode;
extern int Thim2dMode;
extern int ThimExposuresMayBeComing;
extern AXWidget WindowList;

extern void DrawHorizGroove();
extern void AdjustPulldownOrigin();
extern void xmDrawString();
extern mbchar *ElfFetchMenuWord();
extern char *AccelStr();


/* active box in menu bar */
int BarBoxX,BarBoxY,BarBoxWidth,BarBoxHeight;
int BarBoxMenuBarIndex;

#define PROF_START_STR NXLT("axMB")
static char AxMenuBarName[31];
static int AxMenuBarParent;
static int AxLastCascadeItem = -1;

#define CASCADE_TIMEOUT 300 /* in milliseconds */
struct timeval TimeCascadeShouldPopup;

static menu_item_def *PulldownMenu();
static menu_item_def *CascadeMenu();
static void axPaintPulldownContents();
static int axPulldownItemInactive();
static int axNearestActiveItem();

static MenuItemNumFromKey(/* AXWidget_menu w, menu_item_def *mP, wchar key */);
static PullDownItemNumFromKey(/* AXWidget_menu w, menu_item_def *mP, wchar key */);
static CascadeItemNumFromKey(/* AXWidget_menu w, menu_item_def *mP, wchar key */);
static XYFromMenuItem(/* AXWidget_menu w, int menu_item, int *xP, *yP */);
static XYFromPulldownItem(/* AXWidget_menu w, int pulldown_item, int *xP, *yP */);
static XYFromCascadeItem(/* AXWidget_menu w, int cascade_item, int *xP, *yP */);
static XYForCascade(/* AXWidget_menu w, int pulldown_item, int *xP, *yP */);
static PulldownKeyModeNew(/* AXWidget_menu w, int x, y */);
static PulldownKeyModeMove(/* AXWidget_menu w, int x, y */);
static PulldownKeyModeDone(/* AXWidget_menu w, int x, y */);
static PulldownItemHasCascade(/* AXWidget_menu w, menu_item_def *mP, int item */);
static HandleCascade (/* register AXWidget_menu w, int pulldown_item, int mode, XEvent *eventP, register menu_item_def *mP */);
static int ActivateCascade(/* AXWidget_menu w, register int x */);
static ClearAccelsList (/* char **pp, int cnt */);
static ClearAllocStrsList (/* char **pp, int *isAlloced, int cnt */);

void PaintMenuPulldown();
char *MnemonicInString();

/*************************************************************************
  function: TmPulldownKeyEvent
  */
TmPulldownKeyEvent(eventP)
XEvent *eventP;
{
	AXWidget w;
	KeySym ksym;
	int nbytes;
	wchar *buf;
	int ret_val;

	if (!eventP)
		return(FALSE);
	if ( (eventP->type != KeyPress) && (eventP->type != KeyRelease) )
		return(FALSE);

	for ( w = WindowList; w; w = w->nextWin ) {
		if ( w->window == eventP->xany.window )
			break;
	}
	if ( !w )
		return(FALSE);

	nbytes = AXLookupString16(eventP, &buf, &ksym);

	ret_val = PullDownKeyEvent(w, ksym, buf[0], nbytes, eventP);

	return( ret_val );
}


/*************************************************************************
  function: PullDownKeyEvent - Keyboard operation of pulldowns.
  Returns TRUE if event handled by this routine.
  */
int
PullDownKeyEvent(w, ksym, key_char, nbytes, eventP)
AXWidget w;
KeySym ksym;     /* keysym from key event */
wchar key_char;    /* the displayable character */
int nbytes;      /* number of chars (only looked at for zero)  */
XEvent *eventP;  /* the key event */
{
	menu_item_def *mP = NULL;
	menu_item_def *pulldown_mP;
	int item;
	int cur_x, cur_y;
	XEvent dummy_ev;
	unsigned int state;
	AXWidget_menu m_w;
	AXWidget wp;
	
	if (!w || !eventP)
		return(FALSE);
	
	if ( eventP->type != KeyPress )
		return(FALSE);			/* operate pulldown off KeyPress events */
	
	if ( nbytes == 0 ) {
		if( (ksym == 0) || AxKeysymIsModifier(ksym) )
			/* bogus key  or shift/control...  which we ignore */
			return(FALSE);
	}
	
	if ( (w->type != W_MENUBAR) && (w->type != W_MAIN) )
		wp = (AXWidget)w->parent;
	else
		wp = w;
	
	/* try again for events reported to listbox, etc. */
	if ( wp && (wp->type == W_DIALOG) )
		wp = (AXWidget)wp->parent;
	
	if ( wp && wp->type == W_MAIN ) {
		m_w = ((AXWidget_main)wp)->menuBar;
	}
	else
		m_w = (AXWidget_menu)wp;
	
	if ( !m_w || m_w->type != W_MENUBAR )
		return(FALSE);
	
	mP = m_w->popups;
	if ( !mP )
		return(FALSE);
	
	/*
	 * !! HACK !!
	 * Key operation of pulldowns implemented by sending "logical"
	 * pointer events to PaintMenuPulldown.  The event x & y are
	 * calculated based on the menuBar/pulldown/cascade geometries
	 * and the key's logical action.
	 *
	 * Implemented this way because PaintMenuPulldown is fairly
	 * large AND complex, and makes toooo many assumptions about
	 * the event x & y.  Should be re-written???  If time permits,
	 * maybe.  But then, this falls out fairly cleanly too.....
	 */
	
	state = eventP->xkey.state;
	if ( AxPulldownStayup ) {
		switch ( ksym ) {
			
		case XK_Escape:
			ResetActive();
			AxLastCascadeItem = -1;
			return(TRUE);
			
		case XK_Return:
			if ( m_w->cascade ) {
				if ( AxCascadeItem != -1 ) {
					XYFromCascadeItem(m_w, AxCascadeItem, &cur_x, &cur_y);
					PulldownKeyModeDone(m_w, cur_x, cur_y);
				}
				else {
					return(FALSE);
				}
			}
			else if ( AxPulldownItem != -1 ) {
				if ( CascadeMenu(m_w) ) {
					XYFromPulldownItem(m_w, AxPulldownItem, &cur_x, &cur_y);
					PulldownKeyModeMove(m_w, cur_x, cur_y);
					XYForCascade(m_w, AxPulldownItem, &cur_x, &cur_y);
					PulldownKeyModeDone(m_w, cur_x, cur_y);
				}
				else {
					XYFromPulldownItem(m_w, AxPulldownItem, &cur_x, &cur_y);
					PulldownKeyModeDone(m_w, cur_x, cur_y);
				}
			}
			else {
				return(FALSE);
			}
			return(TRUE);
			
		case XK_Left:
			if ( m_w->cascade ) {
				HandleCascade(m_w, AxPulldownItem, OPTION_DONE, NULL, NULL);
				XYFromPulldownItem(m_w, AxPulldownItem, &cur_x, &cur_y);
				TimeCascadeShouldPopup.tv_sec = 0;  /* reset popup timer */
				PulldownKeyModeMove(m_w, cur_x, cur_y);
				
			}
			else {
				item = PrevMenuItem(m_w, AxMenuItem);
				XYFromMenuItem(m_w, item, &cur_x, &cur_y);
				PulldownKeyModeMove(m_w, cur_x, cur_y);
				PulldownKeyModeDone(m_w, cur_x, cur_y);
			}
			return(TRUE);
			
		case XK_Right:
			if ( CascadeMenu(m_w) && !m_w->cascade ) {
				XYFromPulldownItem(m_w, AxPulldownItem, &cur_x, &cur_y);
				PulldownKeyModeMove(m_w, cur_x, cur_y);
				XYForCascade(m_w, AxPulldownItem, &cur_x, &cur_y);
				PulldownKeyModeDone(m_w, cur_x, cur_y);
			}
			else {
				item = NextMenuItem(m_w, AxMenuItem);
				XYFromMenuItem(m_w, item, &cur_x, &cur_y);
				PulldownKeyModeMove(m_w, cur_x, cur_y);
				PulldownKeyModeDone(m_w, cur_x, cur_y);
			}
			return(TRUE);
			
		case XK_Up:
		case XK_Down:
			if ( m_w->cascade ) {
				item = axNearestActiveItem(m_w, AxCascadeItem,
										   m_w->num_cascade_opts,
										   TRUE, ksym == XK_Up);
				if (item >= 0) {
					XYFromCascadeItem(m_w, item, &cur_x, &cur_y);
					PulldownKeyModeMove(m_w, cur_x, cur_y);
					return(TRUE);
				}
				else {
					return (FALSE);
				}
			}
			item = axNearestActiveItem(m_w, AxPulldownItem,
									   m_w->num_pulldown_opts,
									   FALSE, ksym == XK_Up);
			if (item >= 0) {
				XYFromPulldownItem(m_w, item, &cur_x, &cur_y);
				PulldownKeyModeMove(m_w, cur_x, cur_y);
				return(TRUE);
			}
			else {
				return (FALSE);
			}
			
		}
		
		if ( state & ~(AxShiftMask | AxLockMask | AxAltMask) )
			return(FALSE);		/* only Shift, CapsLock and Alt allowed */
		/* check in menubar */
		if ( state & AxAltMask ) {
			item = MenuItemNumFromKey(m_w, mP, key_char);
			if ( item != -1 ) {
				/* current menuBar item chosen - exit pulldown */
				if ( item == AxMenuItem ) {
					ResetActive();
				}
				else {
					/* new item chosen in menuBar - bring up new pulldown */
					XYFromMenuItem(m_w, item, &cur_x, &cur_y);
					PulldownKeyModeMove(m_w, cur_x, cur_y);
					PulldownKeyModeDone(m_w, cur_x, cur_y);
				}
				return(TRUE);
			}
			return(FALSE);
		}
		
		/* first check in cascade */
		if ( m_w->cascade ) {
			item = CascadeItemNumFromKey(m_w, CascadeMenu(m_w), key_char);
			if ( item != -1 ) {
				XYFromCascadeItem(m_w, item, &cur_x, &cur_y);
				PulldownKeyModeMove(m_w, cur_x, cur_y);
				PulldownKeyModeDone(m_w, cur_x, cur_y);
				return(TRUE);
			}
		}
		
		/* check in pulldown */
		if ( m_w->pulldown ) {
			pulldown_mP = PulldownMenu(m_w);
			item = PullDownItemNumFromKey(m_w, pulldown_mP, key_char);
			if ( item != -1 ) {
				if ( PulldownItemHasCascade(m_w, pulldown_mP, item) ) {
					XYFromPulldownItem(m_w, item, &cur_x, &cur_y);
					PulldownKeyModeMove(m_w, cur_x, cur_y);
					XYForCascade(m_w, item, &cur_x, &cur_y);
					PulldownKeyModeDone(m_w, cur_x, cur_y);
					return(TRUE);
				}
				else {
					if ( m_w->cascade ) {
						HandleCascade(m_w,AxPulldownItem,OPTION_DONE,NULL,NULL);
					}
					XYFromPulldownItem(m_w, item, &cur_x, &cur_y);
					PulldownKeyModeMove(m_w, cur_x, cur_y);
					PulldownKeyModeDone(m_w, cur_x, cur_y);
					return(TRUE);
				}
			}
		}
	}
	else if ( !AxPulldownActive ) {
		if ( state & ~(AxShiftMask | AxLockMask | AxAltMask) )
			return(FALSE);		/* only Shift, CapsLock and Alt allowed */
		/* check in menubar */
		if ( state & AxAltMask ) {
			if ( key_char == 0 )
				{
					/* a keystroke with no printable equivalent, can not be */
					/* a menu-bar mnemonic                   */
					return( FALSE );
				}
			item = MenuItemNumFromKey(m_w, mP, key_char);
			if ( item != -1 ) {
				/* activate new pulldown */
				XYFromMenuItem(m_w, item, &cur_x, &cur_y);
				if ( PulldownKeyModeNew(m_w, cur_x, cur_y) )
					return(TRUE);
			}
		}
	}
	
	return(FALSE);
}

/*************************************************************************
function: MenuItemNumFromKey - Return item num in menuBar having 
keyboard mnemonic "key", or -1 if it doesn't exist
*/
static
MenuItemNumFromKey(w, mP, key)
     AXWidget_menu w;
     menu_item_def *mP;   /* the elf menu data (menuBar) */
     wchar key;            /* key which was pressed */
{
  int ix = 0;
  wchar menu_key;

  if ( islower(key) )
    key = (char) toupper(key);
  while ( mP ) {
    if ( (mP->type == ITEM_MARK) && mP->ender )
      break;
    if ( mP->menu_key_str != (unsigned char *)NULL )
      {
	menu_key = *(mP->menu_key_str);
	if ( islower(menu_key) )
	  menu_key = (char) toupper(menu_key);
	if ( menu_key == key )
	  return(ix);
      }
    if ( mP->type != ITEM_MARK )
      ix++;
    mP++;
  }
  return(-1);
}

/*************************************************************************
function: PullDownItemNumFromKey - Return item num in pulldown having 
keyboard mnemonic "key", or -1 if it doesn't exist
*/
static
PullDownItemNumFromKey(w, mP, key)
     AXWidget_menu w;
     menu_item_def *mP;   /* the elf menu data (pulldown) */
     wchar key;            /* key which was pressed */
{
  int ix = 0;
  int flag;
  char menu_key;

  if ( islower(key) )
    key = (char) toupper(key);
  while ( mP ) {
    if ( ix == MAX_MENU_ITEMS-1 )
      break;
    if ( mP->type == ITEM_MARK && mP->ender )
      break;
    if ( mP->menu_key_str != (unsigned char *)NULL )
      {
	menu_key = *(mP->menu_key_str);
	if ( islower(menu_key) )
	  menu_key = (char) toupper(menu_key);
	if ( menu_key == key )
	  return(ix);
      }
    if ( mP->type == ITEM_FUNC ) {
      flag = FlagsFromApp(w, mP, TRUE);
      if ( (flag != MENUSTAT_NO_SHOW) && (flag != MENUSTAT_KEYONLY) )
	ix++;
    }
    else if ( mP->type == ITEM_MENU ) {
      flag = FlagsOfSubMenu(w, mP);
      if ( flag != MENUSTAT_NO_SHOW )
	ix++;
    }
    mP++;
  }
  return(-1);
}

/*************************************************************************
function: CascadeItemNumFromKey - Return item num in cascade having 
keyboard mnemonic "key", or -1 if it doesn't exist
*/
static
CascadeItemNumFromKey(w, mP, key)
     AXWidget_menu w;
     menu_item_def *mP;   /* the elf menu data (cascade) */
     wchar key;            /* key which was pressed */
{
  int ix = 0;
  int flag;
  wchar menu_key;

  if ( islower(key) )
    key = (char) toupper(key);
  while ( mP ) {
    if ( ix == MAX_MENU_ITEMS-1 )
      break;
    if ( mP->type == ITEM_MARK && mP->ender )
      break;

    /* Ashisuto changed the order of source program to avoid the mistake
     * when Graphics->View->Zoom->"6"(Keypress) is happened. The result
     * was the Zoom of 800%.        95.09.18
     */
    if ( mP->type == ITEM_FUNC ) {
      flag = FlagsFromApp(w, mP, TRUE);
      if ( (flag != MENUSTAT_NO_SHOW) && (flag != MENUSTAT_KEYONLY) ) 
	{
	  if ( mP->menu_key_str != (unsigned char *)NULL )
	    {
	      menu_key = *(mP->menu_key_str);
	      if ( islower(menu_key) )
		menu_key = (char) toupper(menu_key);
	      if ( menu_key == key )
		return(ix);
	    }
	  ix++;
	}
    }
    mP++;
  }
  return(-1);
}

NextMenuItem(w, item)
     AXWidget_menu w;
     int item;
{
  int i;

  /* search to right for visible menu item */
  for ( i = item + 1; i < w->num_buttons; i++ ) {
    if ( w->menu_strs[i] && w->menu_strs[i][0] )
      return(i);
  }
  /* wrap around */
  for ( i = 0; i < item; i++ ) {
    if ( w->menu_strs[i] && w->menu_strs[i][0] )
      return(i);
  }
  return(item);  /* sanity */
}

PrevMenuItem(w, item)
     AXWidget_menu w;
     int item;
{
  int i;

  /* search to left for visible menu item */
  for ( i = item - 1; i >= 0; i-- ) {
    if ( w->menu_strs[i] && w->menu_strs[i][0] )
      return(i);
  }
  /* wrap around */
  for ( i = w->num_buttons - 1; i > item; i-- ) {
    if ( w->menu_strs[i] && w->menu_strs[i][0] )
      return(i);
  }
  return(item);  /* sanity */
}

static
XYFromMenuItem(w, menu_item, xP, yP)
     AXWidget_menu w;
     int menu_item;
     int *xP, *yP;     /* RETURN */
{
  int x, y;

  x = w->menu_starts[menu_item] - NWidth - 2;
  if ( x < 0 )
    x = 0;
  x += ( streq(w->menu_strs[menu_item],"*") ? ASTER_MENU_WID+N2Width :
	 BasicStrlen(w->menu_strs[menu_item])+N2Width ) / 2;
  y = 2 + (w->height / 2);

  *xP = x;
  *yP = y;
}

static
XYFromPulldownItem(w, pulldown_item, xP, yP)
     AXWidget_menu w;
     int pulldown_item;
     int *xP, *yP;     /* RETURN */
{
  int x, y;

  x = w->pulldown_xpos + (w->pulldown_width/2);
  y = w->pulldown_ypos;
  y += pulldown_item * (AxCharHeight + 8);
  y += (AxCharHeight + 8) / 2;

  *xP = x;
  *yP = y;
}

static
XYFromCascadeItem(w, cascade_item, xP, yP)
     AXWidget_menu w;
     int cascade_item;
     int *xP, *yP;     /* RETURN */
{
  int x, y;

  x = w->cascade_xpos + (w->cascade_width/2);
  y = w->cascade_ypos;
  y += cascade_item * (AxCharHeight + 8);
  y += (AxCharHeight + 8) / 2;

  *xP = x;
  *yP = y;
}

static
XYForCascade(w, pulldown_item, xP, yP)
     AXWidget_menu w;
     int pulldown_item;
     int *xP, *yP;     /* RETURN */
{
  int x, y;

  x = w->pulldown_xpos + w->pulldown_width - 6;
  y = w->pulldown_ypos;
  y += pulldown_item * (AxCharHeight + 8);
  y += (AxCharHeight + 8) / 2;

  *xP = x;
  *yP = y;
}

static
PulldownKeyModeNew(w, x, y)
     AXWidget_menu w;
     int x, y;
{
  XEvent dummy_ev;

  dummy_ev.type = ButtonPress;
  dummy_ev.xbutton.x = x;
  dummy_ev.xbutton.y = y;
  dummy_ev.xbutton.button = Button1;
  /* create pulldown */
  PaintMenuPulldown(w, OPTION_NEW, &dummy_ev);

  dummy_ev.type = ButtonRelease;
  /* enter stayup mode */
  PaintMenuPulldown(w, OPTION_DONE, &dummy_ev);

  if ( !AxPulldownStayup ) {
    ResetActive();  /* just in case we didn't grab pointer focus */
    return(FALSE);
  }

#if 0
  dummy_ev.type = ButtonPress;
  PaintMenuPulldown(w, OPTION_STAYUP_SELECT, &dummy_ev);
#endif
  return(TRUE);
}

static
PulldownKeyModeMove(w, x, y)
     AXWidget_menu w;
     int x, y;
{
  XEvent dummy_ev;

  dummy_ev.type = MotionNotify;
  dummy_ev.xmotion.x = x;
  dummy_ev.xmotion.y = y;
  PaintMenuPulldown(w, OPTION_MOVE, &dummy_ev);
}

static
PulldownKeyModeDone(w, x, y)
     AXWidget_menu w;
     int x, y;
{
  XEvent dummy_ev;

  dummy_ev.type = ButtonRelease;
  dummy_ev.xbutton.x = x;
  dummy_ev.xbutton.y = y;
  PaintMenuPulldown(w, OPTION_DONE, &dummy_ev);
}

static menu_item_def *
PulldownMenu(w)
     AXWidget_menu w;
{
  menu_item_def *mP;
  int ix = 0;

  if ( !AxPulldownActive || (AxMenuItem == -1) )
    return(NULL);

  for ( mP = w->popups; mP; mP++ ) {
    if ( (mP->type == ITEM_MARK) && mP->ender )
      break;
    if ( mP->type == ITEM_MENU ) {
      if ( ix == AxMenuItem )
	return( ((menu_item_def *)(mP->word2)) + 1 );
      ix++;
    }
  }
  return(NULL);
}

static menu_item_def *
CascadeMenu(w)
     AXWidget_menu w;
{
  menu_item_def *mP;
  int ix = 0;
  int flag;

  if ( !AxPulldownActive || (AxMenuItem == -1) || (AxPulldownItem == -1) ) {
    return(NULL);
  }

  for ( mP = PulldownMenu(w); mP; mP++ ) {
    if ( mP->type == ITEM_MARK && mP->ender ) {
      break;
    }
    if ( ix == MAX_MENU_ITEMS-1 ) {
      break;
    }
    if ( mP->type == ITEM_FUNC ) {
      flag = FlagsFromApp(w, mP, TRUE);
      if ( flag == MENUSTAT_KEYONLY )
	continue;
      if ( flag == MENUSTAT_NO_SHOW )
	continue;
      ix++;
    }
    else if ( mP->type == ITEM_MENU ) {
      flag = FlagsOfSubMenu(w, mP);
      if (flag == MENUSTAT_INACTIVE) {
	ix++;
	continue;
      }
      if ( flag == MENUSTAT_NO_SHOW )
	continue;
      if ( ix == AxPulldownItem ) {
	return( ((menu_item_def *)(mP->word2)) + 1 );
      }
      ix++;
    }
  }
  return(NULL);
}

static
PulldownItemHasCascade(w, mP, item)
     AXWidget_menu w;
     menu_item_def *mP;
     int item;
{
  int ix = 0;
  int flag;

  if ( !AxPulldownActive || (AxMenuItem == -1) || (item == -1) )
    return(FALSE);

  for ( ; mP; mP++ ) {
    if ( mP->type == ITEM_MARK && mP->ender )
      break;
    if ( ix == MAX_MENU_ITEMS-1 )
      break;
    if ( mP->type == ITEM_FUNC ) {
      flag = FlagsFromApp(w, mP, TRUE);
      if ( flag == MENUSTAT_KEYONLY )
	continue;
      if ( flag == MENUSTAT_NO_SHOW )
	continue;
      ix++;
    }
    else if ( mP->type == ITEM_MENU ) {
      flag = FlagsOfSubMenu(w, mP);
      if (flag == MENUSTAT_INACTIVE)
	return (FALSE);
      if ( flag == MENUSTAT_NO_SHOW )
	continue;
      if ( ix == item )
	return(TRUE);
      ix++;
    }
  }
  return(FALSE);
}


/*************************************************************************
function: EnterPulldownStayupMode - grab pointer and begin menu pulldown
stayup mode.
*/
static int
EnterPulldownStayupMode(w)
     AXWidget_menu w;
{
  unsigned int event_mask;

  event_mask = ButtonPressMask | ButtonReleaseMask | ButtonMotionMask;
  if (XGrabPointer(Dpy,w->window,False,event_mask,GrabModeAsync, 
		   GrabModeAsync,None,None,CurrentTime) == GrabSuccess &&
	  XGrabKeyboard(Dpy,w->window,False,GrabModeAsync, 
					GrabModeAsync,CurrentTime) == GrabSuccess) {
	  AxPulldownStayup = TRUE;
	  return(TRUE);
  }
  /* could not grab pointer: cannot enter stayup mode */
  return(FALSE);
}

/*************************************************************************
function: PaintMenuPulldown - Menu Bar menu
Mode:
    0 - OPTION_NEW - Initial display
    1 - OPTION_MOVE - Pointer motion
    2 - OPTION_DONE - Button up
    3 - OPTION_STAYUP_SELECT - button down in stayup mode
    4 - OPTION_EXPOSE - expose event generated by destroyed cascade: redraw
*/
void
PaintMenuPulldown (w,mode,eventP)
register AXWidget_menu w;
int mode;
XEvent *eventP;
{
	XSetWindowAttributes atts;
	long att_mask;
	int y, ix;
	int real_x, real_y;
	XmAnyCallbackStruct data;
	register menu_item_def *mP;
	static mbchar *strs[MAX_MENU_ITEMS];    /* display str */
	static char seps[MAX_MENU_ITEMS];   /* bool of whether separator here */
	static char menus[MAX_MENU_ITEMS];  /* bool if whether cascade here */
	static char flags[MAX_MENU_ITEMS];  /* disable/toggle/radio... */
	static menu_item_def *funcs[MAX_MENU_ITEMS];
	static char *accels[MAX_MENU_ITEMS];
	static int isAlloced[MAX_MENU_ITEMS];
	static int height;
	static int width;
	static int ypos, xpos;  /* relative to menuBar */
	static int num_items;
	static int make_room_for_toggles;
	int pad = N2Width;
	int by,bx;  /* button pos within the menu */
	GC gc;
	int flag;
	int cascade_pad = 20;
	int xadj=0, yadj=0;
	axolButton button;
	static int defaultIsShowing=0, choicePending=0;
	static int textBoxWid;
	static char *defaultText;
	char asterixSymbol[2], *cp, *tcp;
	static int txwidth;
	static int buttonDown=0;
	static int stayup_old_item, stayup_new_item;
	unsigned long event_mask;
	Window child, fromwin;
	AXWidget_main mw;
	
	if (mode == OPTION_EXPOSE) {
		/* destroyed cascade has generated expose event - redraw pulldown */
		DrawShadow(w->pulldown,SunGC,ShadowGC,2,0,0,width,height,
				   OLSHADOW_NODRAW);
		
		axPaintPulldownContents(w, w->pulldown, num_items,
								strs, seps, menus, flags, funcs, accels,
								width, make_room_for_toggles);
		return;
	}
	
	if ( eventP->xbutton.y<=0 || eventP->xbutton.y>=w->height ) {
		/* HACK!!
		 * This test eliminates (partly) strange behavior
		 * when mouse button is VERY quickly released and pressed 
		 * again. */
		if (AxPulldownActive)
			ix = AxMenuItem;
		else
			return;
	}
	else {
		if (eventP->type==MotionNotify || eventP->type==ButtonPress) {
			/* event verticaly within menubar strip, but check if event 
			 * is outside window (required because InMenuopt will return 
			 * the menu position instead of -1 in this case even if the 
			 * menubar button is not displayed within the window) */
			int checkx;
			XWindowAttributes checkWinAttr;
			if (eventP->type==MotionNotify)
				checkx = eventP->xmotion.x;
			else
				checkx = eventP->xbutton.x;
			XGetWindowAttributes(Dpy,w->window,&checkWinAttr);
			if (checkx<0 || checkx>checkWinAttr.width)
				ix = -1;
			else
				ix = InMenuOpt(w,eventP->xbutton.x);
		}
		else
			ix = InMenuOpt(w,eventP->xbutton.x);
	}
	
	buttonDown = Button3; /* that is, regular pulldown behavior for motif */
	if (mode==OPTION_DONE) {
		if ( (eventP->xbutton.y > 0) &&
			(eventP->xbutton.y < w->height) &&
			(ix == AxMenuItem) && (AxPulldownItem == -1) ) {
			if ( AxPulldownStayup || EnterPulldownStayupMode(w) ) {
				AxPulldownItem = axNearestActiveItem(w, -1, num_items,
													 FALSE, FALSE);
				if (AxPulldownItem >= 0) {
					/* starting stayup mode: hilight first active item
					 */
					DrawShadow(w->pulldown,SunGC,ShadowGC,2,4,
							   AxPulldownItem * (AxCharHeight+8) + 5,
							   width-8,AxCharHeight+6,OLSHADOW_NODRAW);
				}
				if (AxPulldownItem < 0) {
					AxPulldownItem = 0;
				}
				return;
			}
		}
	}
	
	if (mode==OPTION_STAYUP_SELECT){
		/* button down in stayup mode */
		if ( ix != AxMenuItem ) {
			/* new menubar item has been selected */
			HandleCascade(w,-1,OPTION_DONE,eventP,NULL);
			AxExitPulldownStayupMode = FALSE;
			ResetActive();
			AxPulldownItem = -1;
			if ( ix == -1 )
				return;
			mode = OPTION_NEW;
		}
		else {
			if (HandleCascade(w,AxPulldownItem,mode,eventP,NULL)) {
				/* cascade has handled event */
				return;
			}
			by = eventP->xbutton.y - ypos;
			bx = eventP->xbutton.x - xpos;
			if (bx<0 || bx>width || by<0 || by>height) {
				/* pointer outside pulldown window */
				if (eventP->xbutton.y<=0 || 
					eventP->xbutton.y>=w->height) {
					/* pointer outside menubar too */
					HandleCascade(w,-1,OPTION_DONE,eventP,NULL);
					AxExitPulldownStayupMode = FALSE;
					ResetActive();
					return;
				}  
				/* pointer on current menubar option */
				HandleCascade(w,-1,OPTION_DONE,eventP,NULL);
				/* unhighlight current puldown choice */
				if ( AxPulldownItem != -1 ) {
					DrawShadow(w->pulldown,BackGC,BackGC,2,4,
							   AxPulldownItem * (AxCharHeight+8) + 5,
							   width-8,AxCharHeight+6,OLSHADOW_NODRAW);
					AxPulldownItem = -1;
				}
				return;
			}
			/* pointer inside pulldown window */
			mode = OPTION_MOVE;  
		}
	}
	else {
		/* if cascade is up, see if cascade handles event */
		if (HandleCascade(w,AxPulldownItem,mode,eventP,NULL)) {
			if (ix==AxMenuItem) {
				/* new menubar item not selected, 
				 * cascade has handled event, return */
				if (mode == OPTION_DONE)
					buttonDown = 0;
				return;
			}
		}
		/* detect lateral move along menu bar */
		if (mode == OPTION_MOVE) {
			if ( ix == -1 )
				return;
			if ( ix != AxMenuItem ) {
				HandleCascade(w,-1,OPTION_DONE,eventP,NULL);
				AxExitPulldownStayupMode = FALSE;
				ResetActive();
				mode = OPTION_NEW;
			}
		}
	}

	if ( mode == OPTION_NEW ) {
		
		SetCurWinFromMenuWidget(w);
		MenuUID++;
		
		/* figure out which item is selected */
		AxMenuItem = InMenuOpt(w,eventP->xbutton.x);
		if ( AxMenuItem == -1 ) {
			mode = OPTION_DONE;
			AxPulldownItem = -1;
			goto toss_menuPulldown;
		}
		
		/* compute heights and strs array for the pulldown */
		ix = 0;
		for ( mP = w->popups; TRUE; mP++ ) { /* first find it! */
			if ( mP->type == ITEM_MENU ) {
				if ( ix == AxMenuItem )
					break;
				ix++;
			}
		}
		num_items = 0;
		width = 0;
		memset(seps,0,sizeof(seps));
		memset(menus,0,sizeof(menus));
		memset(flags,0,sizeof(flags));
		ClearAccelsList(accels,MAX_MENU_ITEMS);
		ClearAllocStrsList(strs,isAlloced,MAX_MENU_ITEMS);
		make_room_for_toggles = FALSE;
		strncpy(AxMenuBarName,mP->word,30);
		AxMenuBarParent = w->parent->style;
		AxMenuBarName[30] = 0;
		
		for ( mP = ((menu_item_def*)(mP->word2))+1; TRUE; mP++ ) {
			if ( num_items == MAX_MENU_ITEMS-1 )
				break;
			if ( mP->type == ITEM_FUNC ) {
				strs[num_items] = (mbchar *)mP->word;
				if ( *strs[num_items] == '@' ) {
					strs[num_items] =
						ElfFetchMenuWord(w,strs[num_items]+1,mP->word2);
					isAlloced[num_items] = TRUE;
				}
				else
					isAlloced[num_items] = FALSE;
				if ( flags[num_items] = flag = FlagsFromApp(w,mP,TRUE) ) {
					if ( flag == MENUSTAT_NO_SHOW )
						continue;
					if ( flag == MENUSTAT_KEYONLY )
						continue;
					if ( flag != MENUSTAT_INACTIVE )
						make_room_for_toggles = TRUE;
				}
				if ( ShowAccelerators && (accels[num_items] = AccelStr(mP)) ) {
					int len = BasicStrlen(accels[num_items]);
					ix = BasicStrlen(strs[num_items]) + N3Width + len;
					cascade_pad = MIN(cascade_pad,cascade_pad-len);
				}
				else
					ix = BasicStrlen(strs[num_items]);
				if ( ix > width )
					width = ix;
				funcs[num_items] = mP;
				num_items++;
			}
			else if ( mP->type == ITEM_MENU ) {
				strs[num_items] = (mbchar *)mP->word;
				if ( flags[num_items] = flag = FlagsOfSubMenu(w,mP) ) {
					if ( flag == MENUSTAT_NO_SHOW )
						continue;
					if ( flag != MENUSTAT_INACTIVE )
						make_room_for_toggles = TRUE;
				}
				ix = BasicStrlen(mP->word);
				ix += 20;
				cascade_pad = 0;
				if ( ix > width )
					width = ix;
				funcs[num_items] = mP;
				menus[num_items] = TRUE;
				num_items++;
			}
			else if ( (mP->type == ITEM_LINE) && !mP->noshow ) {
				if ( num_items != 0 )
					seps[num_items-1] = TRUE;
			}
			else if ( mP->type == ITEM_MARK ) {
				break;
			}
		}
		w->num_pulldown_opts = num_items;
		if ( AlwaysWide ) {
			make_room_for_toggles = TRUE;
			width += MAX(0,cascade_pad);
		}
		if ( make_room_for_toggles )
			width += 15;    /* seems like a good number! */
		
		/* compute height, 3above,3below,2sep */
		height = num_items * (AxCharHeight+8);
		height += 7;    /* area below */
		
		AxPulldownItem = -1;
	}
	
	/* Handle button up case */
done_case:

	/* Switch to pulldown menu based coordinates */
	by = eventP->xbutton.y - ypos;
	bx = eventP->xbutton.x - xpos;
	
	if ( mode == OPTION_DONE ) {
		if ( (AxPulldownItem != -1) &&
			(by >= 0) && (by <= height) &&
			(bx >= 0) && (bx <= width) ) {
			if (!menus[AxPulldownItem]) {
				if (flags[AxPulldownItem] != MENUSTAT_INACTIVE) {
					if ( THIMhelpMode ) {
						ElfPostHelp(funcs[AxPulldownItem]->word2);
						ElfbSetHelpMode(FALSE);
					}
					else if ( w->callback )
						AxCallback(w->callback,w,
								   w->callbacktag,funcs[AxPulldownItem]);
				}
			}
			else if (flags[AxPulldownItem] != MENUSTAT_INACTIVE) {
				/* selected choice has cascade */
				/* button up on choice having cascade: set pulldown and 
				 * cascade in stayup mode if not already so */
				if (!AxPulldownStayup) {
					if (EnterPulldownStayupMode(w)) {
						/* draw cascade */
						HandleCascade(w,AxPulldownItem,OPTION_NEW,
									  eventP,funcs[AxPulldownItem]);
						return;
					}
					else
						/* pointer not grabbed: destroy cascade & pulldown */
						HandleCascade(w,-1,OPTION_DONE,eventP,NULL);
				}
				else {
					/* already in stayup mode: draw cascade */
					HandleCascade(w,AxPulldownItem,OPTION_NEW,
								  eventP,funcs[AxPulldownItem]);
					return;
				}
			}
		}
  toss_menuPulldown:
		ResetActive();
		ThimExposuresMayBeComing = TRUE;
		buttonDown = 0;
		ClearAccelsList(accels,num_items);
		ClearAllocStrsList(strs,isAlloced,num_items);
		return;
	}
	
	if ( mode != OPTION_NEW )
		goto have_window;
	
	att_mask = CWOverrideRedirect | CWSaveUnder | CWBackPixel;
	atts.override_redirect = TRUE;
	atts.save_under = SAVE_UNDER;
	atts.background_pixel = BackPixel;
	
	ypos = w->height;
	xpos = w->menu_starts[AxMenuItem] - NWidth;
	xpos += 2;
	width += 2*(NWidth + 2);
	
	BarBoxY = 2;
	BarBoxWidth = streq(w->menu_strs[AxMenuItem],"*") ? 
		ASTER_MENU_WID+pad :
			BasicStrlen(w->menu_strs[AxMenuItem])+pad;
	BarBoxHeight = w->height-4;
	BarBoxX = xpos;
	BarBoxWidth -= 4;
	
	if ( Thim2dMode ) {
		DrawShadow(w->window,ForeGC,ForeGC,1,
				   BarBoxX,BarBoxY,BarBoxWidth-1,BarBoxHeight,OLSHADOW_NODRAW);
	}
	else
		DrawShadow(w->window,SunGC,ShadowGC,2,
				   BarBoxX,BarBoxY,BarBoxWidth,BarBoxHeight,OLSHADOW_NODRAW);

	mw = (AXWidget_main) w->parent;
	if (mw && mw->frame && mw->dialog && mw->menuBar && mw->menuBar->window) {
		/* take coords with respect to menubar window, which may be xscrolled
		 */
		fromwin = mw->menuBar->window;
	}
	else {
		fromwin = WIN_OF(w);
	}
	XTranslateCoordinates(Dpy, fromwin, RootWin, xpos, ypos,
						  &real_x, &real_y, &child);
	
	/* Adjust real_x, real_y to make sure pulldown is on-screen */
	AdjustPulldownOrigin( &real_x, &real_y, width, height, &xadj, &yadj);
	xpos += xadj;
	ypos += yadj;
	bx += xadj;
	by += yadj;
	
	w->pulldown = AxCrtWindow(Dpy,RootWin,
							  real_x, real_y, width, height,
							  0, CopyFromParent, InputOutput,
							  CopyFromParent, att_mask, &atts);
	w->pulldown_width = width;
	w->pulldown_height = height;
	w->pulldown_xpos = xpos;
	w->pulldown_ypos = ypos;
	AxPulldownActive = (AXWidget)w;
	event_mask = ExposureMask;
	XSelectInput(Dpy, w->pulldown, event_mask);
	XMapWindow(Dpy,w->pulldown);
	
	/*  XSetInputFocus(Dpy,w->pulldown,RevertToParent,CurrentTime); */
	DrawShadow(w->pulldown,SunGC,ShadowGC,2,0,0,width,height,
			   OLSHADOW_NODRAW);
	
	/* paint contents */
 have_window:
	if ( mode == OPTION_MOVE ) {
		int new_item;
		/* if within current choice.... */
		y = AxPulldownItem * (AxCharHeight+8) + 5;
		if (AxPulldownItem != -1 && by>=y && by<=y+AxCharHeight+8 && 
			bx>=0 && bx<=width) {
			if ( menus[AxPulldownItem] && ActivateCascade(w,bx) &&
				flags[AxPulldownItem] != MENUSTAT_INACTIVE ) {
				HandleCascade(w,AxPulldownItem,OPTION_NEW,
							  eventP,funcs[AxPulldownItem]);
			}
			return;
		}
		
		/* remove shadow for items that don't have cascades */
		if ( AxPulldownItem != -1 && !(w->cascade)) {
			DrawShadow(w->pulldown,BackGC,BackGC,2,4,
					   AxPulldownItem * (AxCharHeight+8) + 5,
					   width-8,AxCharHeight+6,OLSHADOW_NODRAW);
		}
		
		/* if we move outside the pulldown, reset the pulldown timer */
		if (by<0 || by>height || bx<0 || bx>width) {
			if (!w->cascade)
				AxPulldownItem = -1;
			TimeCascadeShouldPopup.tv_sec = 0;
			return;
		}
		new_item = (by - 5) / (AxCharHeight+8);
		
		if ( new_item < 0 )
			new_item = 0;
		if ( new_item >= num_items )
			new_item = num_items - 1;
		/* remove shadow and cascade for items with cascades */
		if ( AxPulldownItem!=new_item ) {
			DrawShadow(w->pulldown,BackGC,BackGC,2,4,
					   AxPulldownItem * (AxCharHeight+8) + 5,
					   width-8,AxCharHeight+6,OLSHADOW_NODRAW);
			HandleCascade(w,-1,OPTION_DONE,eventP,NULL);
		}
		AxPulldownItem = new_item;
		/* reset cascade popup timer */
		TimeCascadeShouldPopup.tv_sec = 0;
		if ( menus[AxPulldownItem] && ActivateCascade(w,bx) &&
			buttonDown==Button3 &&
			flags[AxPulldownItem] != MENUSTAT_INACTIVE) {
			
			HandleCascade(w,AxPulldownItem,OPTION_NEW,
						  eventP,funcs[AxPulldownItem]);
		}
	}
	if ( AxPulldownItem != -1 ) {
		if ( flags[AxPulldownItem] != MENUSTAT_INACTIVE ) {
			DrawShadow(w->pulldown,SunGC,ShadowGC,2,4,
					   AxPulldownItem * (AxCharHeight+8) + 5,
					   width-8,AxCharHeight+6,OLSHADOW_NODRAW);
		}
	}
	else
		choicePending = 0;
	
	if ( mode == OPTION_NEW ) {
		/* reset cascade popup timer */
		TimeCascadeShouldPopup.tv_sec = 0;
		axPaintPulldownContents(w, w->pulldown, num_items,
								strs, seps, menus, flags, funcs, accels,
								width, make_room_for_toggles);
	}
}

/*************************************************************************
function: HandleCascade - Deal with cascading menus
Motion rules:
    Movement within cascade responded to, rtn TRUE
    Movement outside cascade:
        In menu bar - toss cascade
        In pulldown:
            In same option - keep
            In other option - toss
        Else - keep

*/
static
HandleCascade (w,pulldown_item,mode,eventP,mP)
register AXWidget_menu w;
int pulldown_item;      /* pos in pulldown */
int mode;           /* hint on what to do */
XEvent *eventP;
register menu_item_def *mP;
{
	XSetWindowAttributes atts;
	long att_mask;
	Window child, fromwin;
	static int ypos, xpos;  /* relative to start of menuBar */
	int real_y, real_x;
	static int width = 100, height = 100;
	int x, y, ix, flag;
	XmAnyCallbackStruct data;
	static char *strs[MAX_MENU_ITEMS];
	static char seps[MAX_MENU_ITEMS];
	static char menus[MAX_MENU_ITEMS];
	static menu_item_def *funcs[MAX_MENU_ITEMS];
	static char *accels[MAX_MENU_ITEMS];
	static int isAlloced[MAX_MENU_ITEMS];
	static char flags[MAX_MENU_ITEMS];  /* disable/toggle/radio... */
	static int num_items;
	static int make_room_for_toggles;
	int pad = N2Width;
	int by,bx;  /* button pos within the menu */
	int did_call;
	GC gc;
	static int xadj=0, yadj=0;
	static int defaultIsShowing=0, choicePending=0;
	int option_stayup_select;
	static int cascade_default;
	AXWidget_main mw;
	
	if ( !w->pulldown ) {
		return(FALSE);
	}
	
	if (mode==OPTION_STAYUP_SELECT){
		option_stayup_select = TRUE;
		mode = OPTION_MOVE;
	}
	else
		option_stayup_select = FALSE;
	
	if ( mode == OPTION_NEW ) {
		if ( pulldown_item == -1 ) {
			return(FALSE);
		}
		if ( pulldown_item == AxLastCascadeItem ) {
			return(FALSE);
		}
		AxLastCascadeItem = pulldown_item;
		if ( w->cascade ) {		/* should move! */
			AxDestroyWindow(Dpy, w, w->cascade);
			ClearAccelsList(accels,num_items);
			ClearAllocStrsList(strs,isAlloced,num_items);
		}
		att_mask = CWOverrideRedirect | CWSaveUnder | CWBackPixel;
		atts.override_redirect = TRUE;
		atts.save_under = SAVE_UNDER;
		atts.background_pixel = BackPixel;
		xpos = w->pulldown_xpos + w->pulldown_width - 4;
		ypos = pulldown_item * (AxCharHeight+8);
		ypos += w->pulldown_ypos;   /* area below */
		/* compute heights and strs array for the window */
		num_items = 0;
		width = 0;
		memset(seps,0,sizeof(seps));
		memset(flags,0,sizeof(flags));
		ClearAccelsList(accels,MAX_MENU_ITEMS);
		ClearAllocStrsList(strs,isAlloced,MAX_MENU_ITEMS);
		make_room_for_toggles = FALSE;
		
		for ( mP = ((menu_item_def*)(mP->word2))+1; TRUE; mP++ ) {
			if ( num_items == MAX_MENU_ITEMS-1 )
				break;
			if ( mP->type == ITEM_FUNC ) {
				strs[num_items] = mP->word;
				if ( *strs[num_items] == '@' ) {
					strs[num_items] = (char *)
						ElfFetchMenuWord(w,strs[num_items]+1,mP->word2);
					isAlloced[num_items] = TRUE;
				}
				else
					isAlloced[num_items] = FALSE;
				if ( flags[num_items] = flag = FlagsFromApp(w,mP,TRUE) ) {
					if ( flag == MENUSTAT_NO_SHOW )
						continue;
					if ( flag == MENUSTAT_KEYONLY )
						continue;
					if ( flag != MENUSTAT_INACTIVE )
						make_room_for_toggles = TRUE;
				}
				if ( ShowAccelerators && (accels[num_items] = AccelStr(mP)) ) {
					int len = BasicStrlen(accels[num_items]);
					ix = BasicStrlen(mP->word) + N3Width + len;
				}
				else
					ix = BasicStrlen(mP->word);
				if ( ix > width )
					width = ix;
				funcs[num_items] = mP;
				num_items++;
			}
			else if ( (mP->type == ITEM_LINE) && !mP->noshow ) {
				if ( num_items != 0 )
					seps[num_items-1] = TRUE;
			}
			else if ( mP->type == ITEM_MARK ) {
				break;
			}
		}
		w->num_cascade_opts = num_items;
		if ( AlwaysWide ) {
			width += 15 + 20;
			make_room_for_toggles = TRUE;
		}
		else if ( make_room_for_toggles )
			width += 15;    /* seems like a good number! */
		
		/* compute height, 3above,3below,2sep */
		height = num_items * (AxCharHeight+8);
		height += 7;    /* area below */
		width += 2*(NWidth + 2);
		
		AxCascadeItem = -1;
		
		mw = (AXWidget_main) w->parent;
		if (mw && mw->frame && mw->dialog &&
			mw->menuBar && mw->menuBar->window) {
			
			/* take coords with respect to menubar window,
			 * which may be scrolled horizontally (dialogs
			 * may be scrollable if mw->frame present)
			 */
			fromwin = mw->menuBar->window;
		}
		else {
			fromwin = WIN_OF(w);
		}
		XTranslateCoordinates(Dpy,fromwin,RootWin,
							  xpos,ypos,
							  &real_x, &real_y,
							  &child);
		
		/* adjust pulldown origin if necessary to keep it on-screen */
		AdjustPulldownOrigin( &real_x, &real_y, width, height, &xadj, &yadj );
		xpos += xadj;
		ypos += yadj;
		
		w->cascade = AxCrtWindow(Dpy, RootWin,
								 real_x, real_y, width, height,
								 0, CopyFromParent, InputOutput,
								 CopyFromParent, att_mask, &atts);
		w->cascade_width = width;
		w->cascade_height = height;
		w->cascade_xpos = xpos;
		w->cascade_ypos = ypos;
		XMapWindow(Dpy,w->cascade);
		
		DrawShadow(w->cascade,SunGC,ShadowGC,2,0,0,width,height,
				   OLSHADOW_NODRAW);
		
		axPaintPulldownContents(w, w->cascade, num_items,
								strs, seps, (char *) NULL,
								flags, funcs, accels,
								width, make_room_for_toggles);
		
		if ( AxPulldownStayup && eventP && 
			(eventP->type == ButtonRelease) ) {
			
			AxCascadeItem = axNearestActiveItem(w, -1, num_items, TRUE, FALSE);
			if (AxCascadeItem >= 0) {
				/* starting stayup mode: hilight first active item
				 */
				DrawShadow(w->cascade,SunGC,ShadowGC,2,4,
						   AxCascadeItem * (AxCharHeight+8) + 5,
						   width-8,AxCharHeight+6,OLSHADOW_NODRAW);
			}
			if (AxCascadeItem < 0) {
				AxCascadeItem = 0;
			}
		}
		return(FALSE);
	}
	if ( !w->cascade )
		return(FALSE);
	
	did_call = FALSE;
	
	if ( mode == OPTION_DONE ) {
		if ( eventP ) {
			x = eventP->xbutton.x - xpos;
			y = eventP->xbutton.y - ypos;
		}
		else {
			x = y = -1;
		}
		if ( (pulldown_item != -1) && (AxCascadeItem != -1) &&
			(y >= 0) && (y <= height) &&
			(x >= 0) && (x <= width) &&
			(flags[AxCascadeItem] != MENUSTAT_INACTIVE) ) {
			if ( THIMhelpMode ) {
				ElfPostHelp(funcs[AxCascadeItem]->word2);
				ElfbSetHelpMode(FALSE);
			}
			else if ( w->callback )
				AxCallback(w->callback,w,w->callbacktag,funcs[AxCascadeItem]);
			did_call = TRUE;
		}
		
	toss:
		AxDestroyWindow(Dpy, w, w->cascade);
		w->cascade = 0;
		ClearAccelsList(accels,num_items);
		ClearAllocStrsList(strs,isAlloced,num_items);
		AxLastCascadeItem = -1;
		AxCascadeItem = -1;
		ThimExposuresMayBeComing = TRUE;
		if ( !did_call ) {
			XFillRectangle(Dpy,w->pulldown,BackGC, /* restore overlap */
						   w->pulldown_width-4,0,2,w->pulldown_height);
			DrawShadow(w->pulldown,SunGC,ShadowGC,2,
					   0,0,w->pulldown_width,
					   w->pulldown_height,OLSHADOW_NODRAW);
			return(FALSE);
		}
		else {
			ResetActive();
			return(TRUE);
		}
	}
	
	if ( mode == OPTION_MOVE ) {
		y = eventP->xmotion.y;
		x = eventP->xmotion.x;
		if ((x < xpos) && (AxCascadeItem != -1)) {
			DrawShadow(w->cascade,BackGC,BackGC,2,4,
					   AxCascadeItem * (AxCharHeight+8) + 5,
					   width-8,AxCharHeight+6,OLSHADOW_NODRAW);
			AxCascadeItem = -1;
			return(FALSE);
		}
		if ( x > xpos )
			goto handle_self;	/* in cascade window */
		if ( (x > 0) && (x < w->width) && (y > 0) && (y < w->height) )
			goto toss;			/* in menu bar */
		if ( pulldown_item != AxLastCascadeItem )
			goto toss;
	}
	return(FALSE);
 handle_self:
	if ( eventP ) {
		by = eventP->xmotion.y - ypos;
		bx = eventP->xmotion.x - xpos;
	}
	else
		by = bx = 0;
	
	/* return if within current choice */
	y = AxCascadeItem * (AxCharHeight+8) + 5;
	if (by>=y && by<=y+AxCharHeight+8 && bx>=0 && bx<=width)
		return(TRUE);
	
	if ( AxCascadeItem != -1 ) { /* lift old */
		DrawShadow(w->cascade,BackGC,BackGC,2,4,
				   AxCascadeItem * (AxCharHeight+8) + 5,
				   width-8,AxCharHeight+6,OLSHADOW_NODRAW);
	}
	if (by<0 || by>height || bx<0 || bx>width) {
		/* actually, bx<0 check not necessary since handled above */
		/* pointer outside cascade */
		AxCascadeItem = -1;
		if (option_stayup_select==TRUE)
			/* this OPTION_MOVE is actually button down in stayup mode */
			return(FALSE);
		else if (xadj!=0 || yadj!=0) {
			/* cascade position has been adjusted: pointer may be 
			 * inside pulldown although bx>0 */
			y = eventP->xmotion.y;
			x = eventP->xmotion.x;
			if (x>=w->pulldown_xpos 
				&& x<=w->pulldown_xpos+w->pulldown_width
				&& y>=w->pulldown_ypos
				&& y<=w->pulldown_ypos+w->pulldown_height)
				/* pointer outside cascade but inside pulldown */
				return(FALSE);
		}
		return(TRUE);
	}
	AxCascadeItem = (by - 5) / (AxCharHeight+8);
	
	if ( AxCascadeItem < 0 )
		AxCascadeItem = 0;
	if ( AxCascadeItem >= num_items )
		AxCascadeItem = num_items - 1;
	if ( AxCascadeItem != -1 && flags[AxCascadeItem] != MENUSTAT_INACTIVE ) {
		DrawShadow(w->cascade,SunGC,ShadowGC,2,4,
				   AxCascadeItem * (AxCharHeight+8) + 5,
				   width-8,AxCharHeight+6,OLSHADOW_NODRAW);
	}
	return(TRUE);
}

void UpdateCascadeTimeout()
{
  struct timeval tp;
  struct timezone tzp;
  gettimeofday(&tp, &tzp);

  if ((tp.tv_usec + (CASCADE_TIMEOUT*1000)) > 1000000) {
    TimeCascadeShouldPopup.tv_sec =
      tp.tv_sec + (tp.tv_usec + (CASCADE_TIMEOUT*1000))/1000000;
    TimeCascadeShouldPopup.tv_usec = 
      (tp.tv_usec + (CASCADE_TIMEOUT*1000))%1000000;
  }
  else {
    TimeCascadeShouldPopup.tv_sec = tp.tv_sec;
    TimeCascadeShouldPopup.tv_usec = 
      tp.tv_usec + (CASCADE_TIMEOUT*1000);
  }
}

static int ActivateCascade(w,x)
     AXWidget_menu w;
     register int x;
{
  struct timeval tp;
  struct timezone tzp;
  long cascade_interval;

  gettimeofday(&tp, &tzp);

  if (TimeCascadeShouldPopup.tv_sec == 0) {
    UpdateCascadeTimeout();
    return(FALSE);
  }
  
  cascade_interval = AxSubtractTime(&tp, &TimeCascadeShouldPopup);
  if (cascade_interval > 0) {
    return(TRUE);
  }
}


/*************************************************************************
function: ClearAccelsList - Free alloced accelerator strings
*/
static ClearAccelsList (pp,cnt)
     char **pp;
     int cnt;
{
  register int ix;

  for ( ix = 0; ix < cnt; ix++ ) {
    if ( pp[ix] ) {
      free(pp[ix]);
      pp[ix] = NULL;
    }
  }
}

/*************************************************************************
function: ClearAllocStrsList - Free alloced menu strings
*/
static ClearAllocStrsList (pp,isAlloced,cnt)
     char **pp;
     int *isAlloced;
     int cnt;
{
  register int ix;

  for ( ix = 0; ix < cnt; ix++ ) {
    if ( isAlloced[ix] && pp[ix] ) {
      free(pp[ix]);
      pp[ix] = NULL;
      isAlloced[ix] = FALSE;
    }
  }
}

/*
 * MnemonicInString - Return pointer to mnemonic char in "str".
 *                    Used for underlining mnemonic char on menu bar,
 *                    pulldown, cascade.
 */
char *
MnemonicInString(mnemonic, str)
     char mnemonic;
     char *str;
{
  char *p, *caseless_p;
  char check_char, caseless_mnemonic;

  if ( !str )
    return(NULL);

  caseless_p = NULL;
  caseless_mnemonic = mnemonic;
  if ( islower(mnemonic) )
    caseless_mnemonic = (char) toupper(mnemonic);

  /* First look for mnemonic character having same case.
   * If not found, return case insensitive match if found */
  for ( p = str; *p; p++ ) {
    check_char = *p;
    if ( mnemonic == check_char )
      return(p);  /* case sensitive match found */
    if ( !caseless_p ) {
      /* not yet found a case insensitive match */
      if ( islower(check_char) )
	check_char = (char) toupper(check_char);
      if ( caseless_mnemonic == check_char )
	caseless_p = p;
    }
  }
  return(caseless_p);  /* return case insensitive match, if any */
}


static void
axPaintPulldownContents(w, window, num_items,
			strs, seps, menus, flags, funcs, accels,
			width, make_room_for_toggles)

     AXWidget_menu w;
     Drawable window;           /* in */
     int num_items;             /* in: size of arrays: strs, seps, menus, .. */
     mbchar **strs;             /* in: pulldown item text strings */
     char *seps;                /* in: flags indicating separator positions */
     char *menus;               /* in: whether each item points to cascade */
     char *flags;               /* in: state per item */
     menu_item_def **funcs;     /* in: callbacks per item */
     char **accels;             /* in: accelerator text per item */
     int width;                 /* in: width of pulldown window */
     int make_room_for_toggles; /* in: bool; whether toggle-show pref is on */
{
  int ypos, ix, itmp;
  int key_char_wid, key_start_x;
  char *key_start;
  GC gc;
  int etch = !AxNoWidgetColor;
  static XFontStruct *cascadeFont = (XFontStruct *) NULL;
  static mbchar cascadeChar = (mbchar) 0xdc;   /* an arrow */
	
  if (!cascadeFont) {
    getXFontStruct(NXLT("Zapf Dingbats"), 14, 0, 0,
		   &itmp, &cascadeFont);
  }

  ypos = 8;
  for ( ix = 0; ix < num_items; ix++ ) {
		
    if (flags[ix] == MENUSTAT_INACTIVE){
      gc = DisabledGC;
      w->disabled = TRUE;    /* just for xmDrawString call.. */
    }
    else {
      gc = ForeGC;
    }
		
    if ( make_room_for_toggles ) {
      itmp = 15;
    }
    else {
      itmp = 0;
    }
    xmDrawString(w, window, gc, 2 + CharWids['n'] + itmp,
		 ypos + AxCharAbove, strs[ix], MBstrlen(strs[ix]),
		 w->textShadow, w->textColor, w->widgetColor);
		
    if ( menus && menus[ix] && cascadeFont ) {
      /* draw cascade indicator (right-pointing arrow)
       */
      if (etch && flags[ix] == MENUSTAT_INACTIVE) {
	axSetFont(Dpy, SunGC, cascadeFont);
	axSetFont(Dpy, ShadowGC, cascadeFont);
	axDrawString(Dpy, window, SunGC, cascadeFont, width - 21,
		     ypos + AxCharAbove + 1, &cascadeChar, 1);
	axDrawString(Dpy, window, ShadowGC, cascadeFont, width - 22,
		     ypos + AxCharAbove, &cascadeChar, 1);
      }
      else {
	axSetFont(Dpy, gc, cascadeFont);
	axDrawString(Dpy, window, gc, cascadeFont, width - 22,
		     ypos + AxCharAbove, &cascadeChar, 1);
	axSetFont(Dpy, gc, AxDlgMainFont);
      }
    }
    if ( accels[ix] ) {
      xmDrawString(w, window, gc,
		   width - BasicStrlen(accels[ix]) - NWidth,
		   ypos + AxCharAbove, accels[ix], strlen(accels[ix]),
		   w->textShadow, w->textColor, w->widgetColor);
    }
    w->disabled = FALSE;
		
    if (AxMarkMenuMnemonics &&
	(funcs[ix]->menu_key_str != (unsigned char *)NULL) ) {
      /* underline key mnemonic */
      key_start = MnemonicInString(*(funcs[ix]->menu_key_str),
				   strs[ix]);
      if ( key_start && key_start[0] ) {
	key_start_x = 2 + CharWids['n'];
	key_start_x += (make_room_for_toggles ? 15 : 0);
	key_start_x += BasicStrlen(strs[ix]) -
	  BasicStrlen(key_start);
	key_char_wid = CharWids[*key_start];
	if (etch && flags[ix] == MENUSTAT_INACTIVE ) {
	  XDrawLine(Dpy, window, ShadowGC, key_start_x,
		    ypos+AxCharAbove+2,
		    key_start_x+key_char_wid, ypos+AxCharAbove+2);
	  XDrawLine(Dpy, window, SunGC, key_start_x + 1,
		    ypos+AxCharAbove+3,
		    key_start_x+key_char_wid + 1,
		    ypos+AxCharAbove+3);
	}
	else {
	  XDrawLine(Dpy, window, gc, key_start_x,
		    ypos+AxCharAbove+2,
		    key_start_x+key_char_wid, ypos+AxCharAbove+2);
	}
      }
    }
    if (flags[ix] == MENUSTAT_TOGGLE_ON) {
      if ( UseChecksForToggles )
	XCopyArea(Dpy,MenuCheckPixmap,window,gc,
		  0,0,16,16,CharWids['n']-2,ypos);
      else {
	if ( Thim2dMode )
	  DrawShadow(window,ForeGC,ForeGC,1,
		     CharWids['n'],ypos+2,13,13,OLSHADOW_DRAW);
	else
	  DrawShadow(window,ShadowGC,SunGC,2,
		     CharWids['n'],ypos+2,13,13,OLSHADOW_DRAW);
	XFillRectangle(Dpy,window,ShadowGC,
		       CharWids['n']+3,ypos+5,7,7);
      }
    }
    else if (flags[ix] == MENUSTAT_TOGGLE_OFF) {
      if ( !UseChecksForToggles ) {
	if ( Thim2dMode )
	  DrawShadow(window,ForeGC,ForeGC,1,
		     CharWids['n'],ypos+2,13,13,OLSHADOW_DRAW);
	else
	  DrawShadow(window,SunGC,ShadowGC,2,
		     CharWids['n'],ypos+2,13,13,OLSHADOW_DRAW);
      }
    }
    else if (flags[ix] == MENUSTAT_RADIO_ON) {
      if ( UseChecksForToggles )
	XCopyArea(Dpy,MenuCheckPixmap,window,gc,
		  0,0,16,16,CharWids['n']-2,ypos);
      else {
	if ( Thim2dMode )
	  AxDrawDiamond(window,CharWids['n']-2,ypos,16,
			ForeGC,ForeGC,ForeGC);
	else
	  AxDrawDiamond(window,CharWids['n']-2,ypos,16,
			ShadowGC,SunGC,ShadowGC);
      }
    }
    else if (flags[ix] == MENUSTAT_RADIO_OFF) {
      if ( !UseChecksForToggles ) {
	if ( Thim2dMode )
	  AxDrawDiamond(window,CharWids['n']-2,ypos,16,
			ForeGC,ForeGC,BackGC);
	else
	  AxDrawDiamond(window,CharWids['n']-2,ypos,16,
			SunGC,ShadowGC,BackGC);
      }
    }
    ypos += AxCharHeight + 8;
    if ( seps[ix] )
      DrawHorizGroove(window,5,ypos-5,width-11);
  }
}


/* : axPulldownItemInactive - test whether for the open pulldown on
 * menubar w, the choice located at position dex is grayed out or missing
 */
static int
axPulldownItemInactive(w, dex, cascade)
AXWidget_menu w;
int dex;                   /* in */
int cascade;               /* in: bool; TRUE if dealing with cascade */
{
	int ix, jx, tflags;
	menu_item_def *mP = PulldownMenu(w);

	if (!mP) {
		return (TRUE);
	}
	
	if (cascade) {
		mP = CascadeMenu(w);    /* descend to cascade menu */
		if (!mP) {
			return (TRUE);
		}
	}
	
	/* find dex'th item on this pulldown/cascade, test for grayedness
	 */
	ix = 0;
	for ( jx = 0; jx < 1000; mP++, jx++ ) {
		
		if (ix == MAX_MENU_ITEMS - 1) {
			break;
		}
		if ( (mP->type == ITEM_MARK) && mP->ender ) {
			break;
		}
		if (mP->type == ITEM_FUNC) {
			tflags = FlagsFromApp(w, mP, TRUE);
			if (tflags == MENUSTAT_NO_SHOW || tflags == MENUSTAT_KEYONLY) {
				continue;
			}
			if (ix == dex) {
				if (tflags == MENUSTAT_INACTIVE) {
					return (TRUE);
				}
				else {
					return (FALSE);
				}
			}
			ix++;
		}
		else if (mP->type == ITEM_MENU) {
			tflags = FlagsOfSubMenu(w, mP);
			if (tflags == MENUSTAT_NO_SHOW) {
				continue;
			}
			if (ix == dex) {
				if (tflags == MENUSTAT_INACTIVE) {
					return (TRUE);
				}
				else {
					return (FALSE);
				}
			}
			ix++;
		}
	}
	return (TRUE);
}

/* : axNearestActiveItem - for the open pulldown on menubar w,
 * report the index of the nearest active choice to current_ix.
 * If no active choices found (including current_ix), return (-1),
 * therefore caller must test for >=0 return before using as index.
 */
static int
axNearestActiveItem(w, current_ix, len, cascade, up)
AXWidget_menu w;
int current_ix;         /* in: current index on pulldown */
int len;                /* in: total# choices on pulldown */
int cascade;            /* in: bool; TRUE if dealing with cascade */
int up;                 /* in: bool; TRUE if looking upwards, else downwards */
{
	int incr, wrap, ix;
	int firstdex, wrapped;
	
	if (len <= 0) {
		return (-1);            /* definitely no active items */
	}
	
	if (up) {
		incr = (-1);
		wrap = 0;				/* index at which wraparound occurs */
	}
	else {
		incr = 1;
		wrap = len - 1;
	}
	
	/* init firstdex to the index of immediate neighbor
	 */
	if (current_ix == -1 || current_ix == wrap) {
		firstdex = up ? len - 1 : 0;
	}
	else {
		firstdex = current_ix + incr;
	}

	for (ix = firstdex, wrapped = FALSE ; TRUE; ix += incr) {
		if (wrapped && ix == firstdex) {
			/* we wrapped around without finding any active
			 */
			if (axPulldownItemInactive(w, ix, cascade)) {
				return (-1);
			}
			else {
				return (ix);
			}
		}
		if (!axPulldownItemInactive(w, ix, cascade)) {
			return (ix);
		}
		if (ix == wrap) {
			ix = up ? len : -1;	/* (that's len-1 : 0 after loop incr) */
			wrapped = TRUE;
		}
	}
}
