/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmribbon.c
   Applix ASC version: 29
   Date: Fri Feb  5 21:11:57 1999
*/
#include <basicwin.h>
#include <ax_xm.h>
#include <dwt2xm.h>
#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>
#include <xmdlged.h>
#include <ge_syms.h>

#include <xmribbon.p>


#define RIBB_SZ  15
#define RIBB_OFF 0
#define RIBB_OP_SZ 85

AXWidget_ribbon RibbonArmed = NULL;   /* currently armed ribbon */

extern Display *Dpy;
extern long WinMask;
extern XSetWindowAttributes WinAtt;
extern XFontStruct *MonospaceBoldFont;
extern int WidgetNeedsUpdated; /* Set when a widget needs repainted */


extern GC ForeGC;      /* label text (nominally black) */
extern GC BackGC;      /* "white" */
extern GC SunGC;       /* sunlit side */
extern GC ShadowGC;    /* shaded side */
extern GC MonospaceBoldGC;
extern int Thim2dMode;
extern int DoubleMillisec;
extern unsigned long TimeOfButtonPress;    /* event time of button down */


/***********************************************************************\
*  Color-Ribbon management code                                         *
\***********************************************************************/
#define RIBBON_STRIP 0
#define RIBBON_B1 1
#define RIBBON_B2 2
#define RIBBON_B3 3
#define RIBBON_B4 4
#define RIBBON_B5 5
#define RIBBON_OP 6

/***********************************************************************\
*  GE color palette management code                                     *
\***********************************************************************/

getColorpalMode(ribbon, mode)
AXWidget ribbon;
int *mode;
{
    if ( ribbon && ribbon->type == W_RIBBON )
    {
        if ( streq(ribbon->name, XLT("Fill fg")) )
            *mode = 0;
        else if ( streq(ribbon->name, XLT("Brush fg")) )
            *mode = 0;
        else if ( streq(ribbon->name, XLT("Fill bg")) )
            *mode = 1;
        else if ( streq(ribbon->name, XLT("Brush bg")) )
            *mode = 1;
        else if ( streq(ribbon->name, XLT("Line fg")) )
            *mode = 2;
        else if ( streq(ribbon->name, XLT("Line bg")) )
            *mode = 3;
        else if ( streq(ribbon->name, XLT("Shadow")) )
            *mode = 4;
    }
    else
        *mode = 0;
}

AxGetColorpalMode(w, mode)
AXWidget_main w;
int *mode;
{

    if ( w && w->type == W_MAIN && w->colorpal )
        getColorpalMode(w->colorpal->widgets, mode);
    else
        *mode = 0;
}

AxResetRibbonMode(w, pixed)
AXWidget_main w;
int pixed;
{
    if ( !w || w->type != W_MAIN )
        return;

    if ( w->colorpal && w->colorpal->widgets )
	{
		if ( pixed )
			SetWidgetName(w->colorpal->widgets, XLT("Brush fg"));
		else
			SetWidgetName(w->colorpal->widgets, XLT("Fill fg"));
	}
}

AxResetRibbonPosition(w)
AXWidget_main w;
{
    void PaintRibbon();

    if ( !w )
        return;

    if ( w && w->type == W_MAIN && w->colorpal && w->colorpal->widgets )
        w->colorpal->widgets->scroll_pos = 0;
}

/* color palette option menu callback routine - must match loadRibbons */
void GeRibbonOptionChanged(widget, tag, data)
AXWidget_ribbon widget;
caddr_t tag;
XmRowColumnCallbackStruct *data;
{
    char elfArgString[STRPATH], *ElfSymbolNameFromId();
    AXWidget_main w;
    AXWidget_option childP;
    AXWidget ribbon;
    int id;

    if ( !widget )
        return;

    /* change the ribbon name and request colormap info */
    id = (int)data->data;
    w = (AXWidget_main) WidgetOfWindow((int)tag);
    if ( w && w->colorpal && w->colorpal->widgets )
    {
        /* find the child option menu */
        ribbon = w->colorpal->widgets;
        childP = (AXWidget_option)ribbon->widgets;
        while ( childP )
        {
            if ( childP->type != W_OPTION )
                childP = (AXWidget_option)childP->nextChild;
            else
            { /* FIXME, Use SetWidgetName() */
                if ( childP->num_buttons == 2 ) switch ( id )
                {
                    default:
                    case 0:
                    SetWidgetName(ribbon, XLT("Brush fg"));
                    break;
                    case 1:
                    SetWidgetName(ribbon, XLT("Brush bg"));
                    break;
                }
                else switch ( id )
                {
                    default:
                    case 0:
                    SetWidgetName(ribbon, XLT("Fill fg"));
                    break;
                    case 1:
                    SetWidgetName(ribbon, XLT("Fill bg"));
                    break;
                    case 2:
                    SetWidgetName(ribbon, XLT("Line fg"));
                    break;
                    case 3:
                    SetWidgetName(ribbon, XLT("Line bg"));
                    break;
                    case 4:
                    SetWidgetName(ribbon, XLT("Shadow"));
                    break;
                }
                break;
            }
        }

        /* request for colormap info */
        sprintf(elfArgString, "%d %s %d",
                tag,
                ElfSymbolNameFromId(GEElfMacroId(ELFM_GE_SET_RIBBON_INFO), ElfLanguage),
                id);
        ElfStartNewTaskWithArgs(GEElfMacroId(ELFM_GE_PALETTE_LAUNCHER_), elfArgString);
    }
}

/* color palette generic callback routine */
void AxColorRibbonModified (widget,winID,data)
AXWidget widget;
int winID;
XmRowColumnCallbackStruct *data;
{
    int mode, val;
    AXWidget w;
    char elfArgString[STRPATH], *ElfSymbolNameFromId();

    w = (AXWidget ) widget;
    if ( !w )
        return;

    /* launch the custom color dialog box */
    getColorpalMode(w, &mode);
    if ( data->reason == RIBBON_B5 )
    {
        sprintf(elfArgString, "%d %s %d",
                winID,
                ElfSymbolNameFromId(GEElfMacroId(ELFM_GE_COLOR_BOX_DLG), ElfLanguage),
                mode);
        ElfStartNewTaskWithArgs(GEElfMacroId(ELFM_GE_PALETTE_LAUNCHER_), elfArgString);
    }

    else /* must be a selection, update ge */
    {
		val = *(int *)(data->data);

        sprintf(elfArgString, "%d %s %d %d",
                winID,
                ElfSymbolNameFromId(GEElfMacroId(ELFM_GE_SET_COLOR_BY_CMAP_INDEX_),
                                    ElfLanguage),
                val, mode);
        ElfStartNewTaskWithArgs(GEElfMacroId(ELFM_GE_PALETTE_LAUNCHER_), elfArgString);
    }
}

void sizeRibbons(parentWid, wid, per)
int parentWid, *wid, *per;
{
    *wid = parentWid;
    *per = (parentWid - 155 - RIBB_OP_SZ - RIBB_OFF - 3*RIBB_SZ) / RIBB_SZ;
    *per *= RIBB_SZ;
}

loadRibbons(parent, winID, n_colors, ix,str, ribbon_pixmap, pixed)
AXWidget parent;
int winID;
int n_colors;
int ix;
char *str;
Pixmap ribbon_pixmap;
int pixed;
{
    Arg args[20];
    int stripWid, n;
    int ribbon_width, x, y, total_wid, total_hyt;
    static XtCallbackRec Callback[2];
    extern GC Gc1;

    /* calculate the new size */
    x = y = 0;
    total_hyt = parent->height;
    sizeRibbons(parent->width, &total_wid, &ribbon_width);
    stripWid = RIBB_SZ*n_colors;

    /* create (or update) the color ribbons */
    n = 0;
    XtSetArg(args[n],XmNoldWidget,(caddr_t)parent->widgets); n++;
    XtSetArg(args[n],XmNx,x); n++;
    XtSetArg(args[n],XmNy,y); n++;
    XtSetArg(args[n],XmNwidth, total_wid); n++;
    XtSetArg(args[n],XmNheight,total_hyt); n++;
    XtSetArg(args[n],XmNselectedItemCount,1); n++; /* location of marker */
    XtSetArg(args[n],XmNselectedItems,ix); n++;
    XtSetArg(args[n],XmNmaximum,stripWid); n++;/* width of pixmap */
    XtSetArg(args[n],XmNsliderSize,ribbon_width);n++;/* amt of max displayed */
    CALLBACKER(Callback,AxColorRibbonModified,(caddr_t)winID);
    XtSetArg(args[n],XmNactivateCallback,Callback); n++;
    XtSetArg(args[n],XmNmessageString,str); n++;
    XtSetArg(args[n], XmNpixmap, ribbon_pixmap); n++;
	XtSetArg(args[n],XmNmenuHelpWidget,3571); n++;
    XmCreateRibbon (parent,XLT("Fill fg"),args,n, winID, pixed);
}


changeRibbons(parent, winID, ix,str, pixed)
AXWidget_ribbon parent;
int winID;
int ix;
char *str;
int pixed;
{
    Arg args[20];
    int n;
    static XtCallbackRec Callback[2];
    extern GC Gc1;

    /* update the color ribbons */
    n = 0;
    XtSetArg(args[n],XmNoldWidget,(caddr_t)parent->widgets); n++;
    XtSetArg(args[n],XmNselectedItemCount,1); n++;
    XtSetArg(args[n],XmNselectedItems,ix); n++; /* location of marker */
    CALLBACKER(Callback,AxColorRibbonModified,(caddr_t)winID);
    XtSetArg(args[n],XmNactivateCallback,Callback); n++;
    XtSetArg(args[n],XmNmessageString,str); n++;
    XmCreateRibbon((AXWidget)parent, XLT("Fill fg"), args, n, winID, pixed);
}


/***********************************************************************\
*  Generic ribbon management code                                       *
\***********************************************************************/
/* paint a ribbon style button */
PaintRibbonButton(w, buttonID)
AXWidget_ribbon w;
int buttonID;
{
    XPoint arrow[7];
    GC *gc;
    int x, y;
    unsigned int wid, hyt;
    int strip_wid, x_offset;
    int disabled;

    y = 5;
    disabled = FALSE;
    if ( buttonID == RIBBON_B1 )
    {
        /* left button arrow */
        x_offset = 3 + RIBB_OFF + RIBB_SZ;
        arrow[0].x = x_offset +  4; arrow[0].y = y + 7;
        arrow[1].x = x_offset + 10; arrow[1].y = y + 4;
        arrow[2].x = x_offset + 10; arrow[2].y = y + 10;
        x = x_offset;
        wid = hyt = RIBB_SZ;
        if ( w->scroll_per + w->scroll_pos >= w->scroll_max )
            disabled = TRUE;
    }
    else if ( buttonID == RIBBON_B2 )
    {
        /* right button arrow */
        x_offset = 3 + RIBB_OFF + RIBB_SZ + RIBB_SZ;
        strip_wid = w->scroll_per<w->scroll_max ? w->scroll_per : w->scroll_max;
        arrow[0].x = x_offset + strip_wid + 10; arrow[0].y = y + 7;
        arrow[1].x = x_offset + strip_wid +  4; arrow[1].y = y + 4;
        arrow[2].x = x_offset + strip_wid +  4; arrow[2].y = y + 10;
        x = x_offset + strip_wid;
        wid = hyt = RIBB_SZ;
        if ( w->scroll_pos <= 0 )
            disabled = TRUE;
    }
    else if ( buttonID == RIBBON_B3 )
    {
        /* left button double arrow */
        x_offset = 3 + RIBB_OFF;
        arrow[0].x = x_offset +  2; arrow[0].y = y +  7;
        arrow[1].x = x_offset +  7; arrow[1].y = y +  4;
        arrow[2].x = x_offset +  7; arrow[2].y = y +  7;
        arrow[3].x = x_offset + 12; arrow[3].y = y +  4;
        arrow[4].x = x_offset + 12; arrow[4].y = y + 10;
        arrow[5].x = x_offset +  7; arrow[5].y = y +  7;
        arrow[6].x = x_offset +  7; arrow[6].y = y + 10;
        x = x_offset;
        wid = hyt = RIBB_SZ;
        if ( w->scroll_per+w->scroll_pos >= w->scroll_max )
            disabled = TRUE;
    }
    else if ( buttonID == RIBBON_B4 )
    {
        /* right button double arrow */
        disabled = w->scroll_per > 0 ? FALSE : TRUE;
        x_offset = 3 + RIBB_OFF + RIBB_SZ + RIBB_SZ + RIBB_SZ;
        strip_wid = w->scroll_per<w->scroll_max ? w->scroll_per : w->scroll_max;
        arrow[0].x = x_offset + strip_wid + 12; arrow[0].y = y +  7;
        arrow[1].x = x_offset + strip_wid +  7; arrow[1].y = y +  4;
        arrow[2].x = x_offset + strip_wid +  7; arrow[2].y = y +  7;
        arrow[3].x = x_offset + strip_wid +  2; arrow[3].y = y +  4;
        arrow[4].x = x_offset + strip_wid +  2; arrow[4].y = y + 10;
        arrow[5].x = x_offset + strip_wid +  7; arrow[5].y = y +  7;
        arrow[6].x = x_offset + strip_wid +  7; arrow[6].y = y + 10;
        x = x_offset + strip_wid;
        wid = hyt = RIBB_SZ;
        if ( w->scroll_pos <= 0 )
            disabled = TRUE;
    }

    /* pushed */
    if ( w->value && !w->disabled && !disabled )
        XFillRectangle(Dpy,w->window,ForeGC, x+3,y+3, wid-6,hyt-6);

    /* not pushed */
    else
        XFillRectangle(Dpy,w->window,BackGC, x,y, wid,hyt);

    /* draw the button content */
    if ( !w->disabled ) switch ( buttonID )
    {
        case RIBBON_B1:
            if ( w->value )
                gc = &BackGC;
            else
                gc = &ForeGC;
            XFillPolygon(Dpy,w->window,*gc,arrow,3,Complex,CoordModeOrigin);
        break;
        case RIBBON_B2:
            if ( w->value )
                gc = &BackGC;
            else
                gc = &ForeGC;
            XFillPolygon(Dpy,w->window,*gc,arrow,3,Complex,CoordModeOrigin);
        break;
        case RIBBON_B3:
            if ( w->value )
                gc = &BackGC;
            else
                gc = &ForeGC;
            XFillPolygon(Dpy,w->window,*gc,arrow,7,Complex,CoordModeOrigin);
        break;
        case RIBBON_B4:
            if ( w->value )
                gc = &BackGC;
            else
                gc = &ForeGC;
            XFillPolygon(Dpy,w->window,*gc,arrow,7,Complex,CoordModeOrigin);
        break;
    }

    /* draw button outline */
    if ( w->value )
    {
        if ( !w->disabled && !disabled )
        {
            if ( w->use2d || Thim2dMode )
                DrawRect(w->window,ForeGC,2,x,y,wid,hyt);
            else
            {
                AxDrawShadow(w->window,ShadowGC,SunGC,w->thickness, x,y,wid,hyt,
                           OLSHADOW_NODRAW);
            }
        }
    }
    else
    {
        if ( !w->disabled && !disabled )
        {
            if ( w->use2d || Thim2dMode )
                DrawRect(w->window,ForeGC,1,x,y,wid,hyt);
            else
            {
                AxDrawShadow(w->window,SunGC,ShadowGC,w->thickness, x,y,wid,hyt,
                               OLSHADOW_NODRAW);
            }
        }
    }
}

PaintRibbonStrip(w)
AXWidget_ribbon w;
{
    int stripWid, stripPos, y;

    y = 5;
    if ( w->scroll_per < w->scroll_max )
    {
        stripWid = w->scroll_per;
        stripPos = w->scroll_pos;
    }
    else
    {
        stripWid = w->scroll_max;
        stripPos = 0;
    }

    stripWid = w->scroll_per<w->scroll_max ? w->scroll_per : w->scroll_max;
    XCopyArea(Dpy, w->pixmap, w->window, ForeGC,
              stripPos, 0,
              stripWid, RIBB_SZ,
              3+RIBB_OFF+RIBB_SZ+RIBB_SZ, y);
}

PaintRibbonMarker(w, end_x)
AXWidget_ribbon w;
int end_x;
{
    int leftSlot, rightSlot, x,y, stripWid, stripPos;
    XRectangle r[1];

    y = 5;
    if ( w->scroll_per < w->scroll_max )
    {
        stripWid = w->scroll_per;
        stripPos = w->scroll_pos;
    }
    else
    {
        stripWid = w->scroll_max;
        stripPos = 0;
    }
    leftSlot  = stripPos / RIBB_SZ;
    rightSlot = stripPos / RIBB_SZ + stripWid / RIBB_SZ;
    if ( w->selections[1] >= leftSlot && w->selections[1] < rightSlot )
    {
        x = 3 + RIBB_OFF + 2*RIBB_SZ + ((w->selections[1] - leftSlot)*RIBB_SZ);

        /* outline the cell */
        XDrawRectangle(Dpy,w->window,ForeGC, x+0,y+0, RIBB_SZ-1, RIBB_SZ-1);
        XDrawRectangle(Dpy,w->window,ForeGC, x+1,y+1, RIBB_SZ-3, RIBB_SZ-3);
        XDrawRectangle(Dpy,w->window,BackGC, x+2,y+2, RIBB_SZ-5, RIBB_SZ-5);
    }

    /* clear the area to draw the string on */
    x = 3 + RIBB_OFF + 2*RIBB_SZ + stripWid + 2*RIBB_SZ + RIBB_SZ/2;
    XFillRectangle(Dpy,w->window,BackGC, x, y, end_x-x-2, RIBB_SZ);

    /* draw string associated with selection */
    r[0].x = x;
    r[0].y = y;
    r[0].width = end_x-x-2;
    r[0].height = RIBB_SZ;
    XSetClipRectangles(Dpy, MonospaceBoldGC, 0, 0, r, 1, YXBanded);
#if BUILD_WIDE
    if ( w->selections[1] >= 0 )
    {
        axDrawString16(Dpy, w->window, MonospaceBoldGC, MonospaceBoldFont,
					   x+RIBB_SZ, y+11,
					   w->wstr, Wstrlen(w->wstr) );
    } else {
       /* Ashisuto changed for doubled label on ribbon marker. 95.11.23 */
       mbchar *mbstr = XLT("Multiple Colors Selected");
       axDrawString(Dpy, w->window, MonospaceBoldGC, MonospaceBoldFont,
                    x+RIBB_SZ, y+11,
                    mbstr, MBstrlen(mbstr) );

    }
#else
    if ( w->selections[1] >= 0 )
    {
        axDrawString(Dpy, w->window, MonospaceBoldGC, MonospaceBoldFont,
					 x+RIBB_SZ, y+11, w->wstr, strlen(w->wstr) );
    }
    else
    {
        axDrawString(Dpy, w->window, MonospaceBoldGC, MonospaceBoldFont,
					 x+RIBB_SZ, y+11,
					 XLT("Multiple Colors Selected"), 15);
    }
#endif
    XSetClipMask(Dpy, MonospaceBoldGC, None);
}

void PaintRibbon(w, exP, full_regen, layered)
AXWidget_ribbon w;
XExposeEvent *exP;
int full_regen;
int layered;
{
    AXWidget childP;
    int x_offset;
    int strip_wid;

    if ( !w || (w->type != W_RIBBON) || !w->parent->display )
        return;

    if ( (w->flags & WFLG_IN_EDIT) && layered && dePaintWidget((AXWidget)w) )
        return;

    /* if the widget window has not been created, do so now */
    if ( !w->window )
    {
#if AX_X11
        w->window = AxCrtWindow(Dpy,w->parent->window,    
                                w->xpos, w->ypos,
                                w->width, w->height,
                                0,CopyFromParent,InputOutput,
                                CopyFromParent,WinMask,&WinAtt);
        XSelectInput(Dpy, w->window, TM_ALL_INPUTS_MASK);
        AxMapWindow((AXWidget)w, Dpy, w->window, NULL);
#else
    if ( !w->winid )
        RealizeRibbon(w);
#endif
        w->mapped = TRUE;
    }

    /* Clear area */
    XFillRectangle(Dpy, w->window, BackGC, 0, 0, w->width, w->height);

    /* draw pixmap */
    PaintRibbonStrip(w);

    /* draw the "scroll 1 cell" left button */
    PaintRibbonButton(w, RIBBON_B1);

    /* draw the "scroll 1 cell" right button */
    PaintRibbonButton(w, RIBBON_B2);

    /* draw the "scroll 1 field" left button */
    PaintRibbonButton(w, RIBBON_B3);

    /* draw the "scroll 1 field" right button */
    PaintRibbonButton(w, RIBBON_B4);

    /* draw option menu */
    x_offset = w->width;
    childP = w->widgets;
    while ( childP )
    {
        if ( childP->type != W_OPTION )
            childP = childP->nextChild;
        else
        {
            strip_wid = w->scroll_per < w->scroll_max ?
                w->scroll_per : w->scroll_max;
            x_offset = w->width - childP->width;
            childP->xpos = x_offset;
            PaintOption((AXWidget_option)childP,(XEvent *)exP, full_regen, layered);
            break;
        }
    }

    /* draw marker on selected cell, and draw string associated with cell */
    PaintRibbonMarker(w, x_offset+2);

#if AXMS
    /* extend ribbon's shadow border on NT. for now? */
    x_offset = w->width - 4;        
#endif

    /* draw a shadow around the non-option part */
    AxDrawShadow(w->window, SunGC, ShadowGC, w->thickness,
               0, 1, x_offset+4, w->height-1,
               OLSHADOW_NODRAW );
}

ArmRibbon(w, eventP)
AXWidget_ribbon w;
XEvent *eventP;
{
    int pos, select, stripWid, stripPos, un_disable, x, y, x_offset;
    XmRowColumnCallbackStruct cb_data;
    AXWidget childP;

    if ( !w )
        return;

    if ( w->scroll_per < w->scroll_max )
    {
        stripWid = w->scroll_per;
        stripPos = w->scroll_pos;
    }
    else
    {
        stripWid = w->scroll_max;
        stripPos = 0;
    }

    x = eventP->xbutton.x - 3;
    y = eventP->xbutton.y;
    if ( eventP->type == ButtonPress )
    {
        if ( RibbonArmed )
            return;

        /* an inactive area */
        if ( x < RIBB_OFF )
            return;

        /* the left buttons */
        else if ( x < RIBB_OFF + 1*RIBB_SZ )
            pos = RIBBON_B3;
        else if ( x < RIBB_OFF + 2*RIBB_SZ )
            pos = RIBBON_B1;

        /* the ribbon strip */
        else if ( x < RIBB_OFF + 2*RIBB_SZ + stripWid )
        {
            pos = RIBBON_STRIP;
            select = stripPos/RIBB_SZ + (x-(RIBB_OFF+2*RIBB_SZ))/RIBB_SZ;
        }

        /* the right buttons */
        else if ( x < RIBB_OFF + 3*RIBB_SZ + stripWid )
            pos = RIBBON_B2;
        else if ( x < RIBB_OFF + 4*RIBB_SZ + stripWid )
            pos = RIBBON_B4;

        /* the option list */
        else if ( x > w->width - RIBB_OP_SZ )
            pos = RIBBON_OP;

        else /* an inactive area */
            return;

        /* an inactive area for all but the option menu */
        if ( pos != RIBBON_OP && (y < 5 || y > RIBB_SZ+5) )
            return;

        switch ( pos )
        {
            /* left buttons */
            case RIBBON_B3:
                /* depress button, scroll 1 block of cells to the left */
                w->value = TRUE;
                PaintRibbonButton(w, RIBBON_B3);

                RibbonArmed = w;
                w->armed = RIBBON_B3;

                un_disable = w->scroll_pos <= 0 ? TRUE : FALSE;

                /* draw scrolled ribbon */
                w->scroll_pos+=stripWid;
                if ( w->scroll_pos > w->scroll_max - stripWid )
                    w->scroll_pos = w->scroll_max - stripWid;
                PaintRibbonStrip(w);

                /* draw option menu */
                x_offset = w->width;
                childP = w->widgets;
                while ( childP )
                {
                    if ( childP->type != W_OPTION )
                        childP = childP->nextChild;
                    else
                    {
                        x_offset = w->width - childP->width;
                        childP->xpos = x_offset;
                        PaintOption((AXWidget_option)childP, (XEvent *)0, TRUE, FALSE);
                        break;
                    }
                }

                /* draw the marker and the color string */
                PaintRibbonMarker(w, x_offset+2);

                /* draw newly active buttons */
                if ( un_disable )
                {
                    w->value = FALSE;
                    PaintRibbonButton(w, RIBBON_B4);
                    PaintRibbonButton(w, RIBBON_B2);
                    w->value = TRUE;
                }
            break;
            case RIBBON_B1:
                /* depress button, scroll 1 cell left */
                w->value = TRUE;
                PaintRibbonButton(w, RIBBON_B1);

                RibbonArmed = w;
                w->armed = RIBBON_B1;

                un_disable = w->scroll_pos <= 0 ? TRUE : FALSE;

                /* draw scrolled ribbon */
                w->scroll_pos+=RIBB_SZ;
                if ( w->scroll_pos > w->scroll_max - stripWid )
                    w->scroll_pos = w->scroll_max - stripWid;
                PaintRibbonStrip(w);

                /* draw option menu */
                x_offset = w->width;
                childP = w->widgets;
                while ( childP )
                {
                    if ( childP->type != W_OPTION )
                        childP = childP->nextChild;
                    else
                    {
                        x_offset = w->width - childP->width;
                        childP->xpos = x_offset;
                        PaintOption((AXWidget_option)childP, (XEvent *)0, TRUE, FALSE);
                        break;
                    }
                }

                /* draw the marker and the color string */
                PaintRibbonMarker(w, x_offset+2);

                /* draw newly active buttons */
                if ( un_disable )
                {
                    w->value = FALSE;
                    PaintRibbonButton(w, RIBBON_B4);
                    PaintRibbonButton(w, RIBBON_B2);
                    w->value = TRUE;
                }
            break;

            /* right buttons */
            case RIBBON_B2:
                /* depress button, scroll 1 cell right */
                w->value = TRUE;
                PaintRibbonButton(w, RIBBON_B2);

                RibbonArmed = w;
                w->armed = RIBBON_B2;

                un_disable = FALSE;
                if ( w->scroll_per+w->scroll_pos >= w->scroll_max )
                    un_disable = TRUE;

                /* draw scrolled ribbon */
                w->scroll_pos-=RIBB_SZ;
                if ( w->scroll_pos < 0 )
                    w->scroll_pos = 0;
                PaintRibbonStrip(w);

                /* draw option menu */
                x_offset = w->width;
                childP = w->widgets;
                while ( childP )
                {
                    if ( childP->type != W_OPTION )
                        childP = childP->nextChild;
                    else
                    {
                        x_offset = w->width - childP->width;
                        childP->xpos = x_offset;
                        PaintOption((AXWidget_option)childP, (XEvent *)0, TRUE, FALSE);
                        break;
                    }
                }

                /* draw the marker and the color string */
                PaintRibbonMarker(w, x_offset+2);

                /* draw newly active buttons */
                if ( un_disable )
                {
                    w->value = FALSE;
                    PaintRibbonButton(w, RIBBON_B3);
                    PaintRibbonButton(w, RIBBON_B1);
                    w->value = TRUE;
                }
            break;
            case RIBBON_B4:
                /* depress button, scroll 1 block of cells to the right */
                w->value = TRUE;
                PaintRibbonButton(w, RIBBON_B4);

                RibbonArmed = w;
                w->armed = RIBBON_B4;

                un_disable = FALSE;
                if ( w->scroll_per+w->scroll_pos >= w->scroll_max )
                    un_disable = TRUE;

                /* draw scrolled ribbon */
                w->scroll_pos-=stripWid;
                if ( w->scroll_pos < 0 )
                    w->scroll_pos = 0;
                PaintRibbonStrip(w);

                /* draw option menu */
                x_offset = w->width;
                childP = w->widgets;
                while ( childP )
                {
                    if ( childP->type != W_OPTION )
                        childP = childP->nextChild;
                    else
                    {
                        x_offset = w->width - childP->width;
                        childP->xpos = x_offset;
                        PaintOption((AXWidget_option)childP, (XEvent *)0, TRUE, FALSE);
                        break;
                    }
                }

                /* draw the marker and the color string */
                PaintRibbonMarker(w, x_offset+2);

                /* draw newly active buttons */
                if ( un_disable )
                {
                    w->value = FALSE;
                    PaintRibbonButton(w, RIBBON_B3);
                    PaintRibbonButton(w, RIBBON_B1);
                    w->value = TRUE;
                }
            break;

            /* check strip for double click */
            case RIBBON_STRIP:
                w->selections[1] = select;
                PaintRibbonStrip(w);

                /* draw option menu */
                x_offset = w->width;
                childP = w->widgets;
                while ( childP )
                {
                    if ( childP->type != W_OPTION )
                        childP = childP->nextChild;
                    else
                    {
                        x_offset = w->width - childP->width;
                        childP->xpos = x_offset;
                        PaintOption((AXWidget_option)childP, (XEvent *)0, TRUE, FALSE);
                        break;
                    }
                }

                /* draw the marker and the color string */
                PaintRibbonMarker(w, x_offset+2);

                /* do the single click callback */
                if ( w->callback )
                {
                    cb_data.data   = (char *)&select;
                    cb_data.reason = RIBBON_STRIP;
                    cb_data.event  = eventP;
                    AxCallback(w->callback,(AXWidget)w,w->callbacktag,
													(caddr_t)&cb_data);
                }

                /* check for a double click callback */
                if ((TimeOfButtonPress - w->last_click_time) <= DoubleMillisec )
                {
                    RibbonArmed = w;
                    w->armed = RIBBON_STRIP;
                }
                w->last_click_time = TimeOfButtonPress;
            break;

            /* option menu */
            case RIBBON_OP:
                childP = w->widgets;
                while ( childP )
                {
                    if ( childP->type != W_OPTION )
                        childP = childP->nextChild;
                    else
                    {
                        PaintOptionPulldown((AXWidget_option)childP,OPTION_NEW,eventP);
                        break;
                    }
                }
            break;
        }
    }
    else if ( eventP->type == ButtonRelease )
    {
        switch(RibbonArmed->armed)
        {
            /* left buttons */
            case RIBBON_B3:
                w->value = FALSE;
                PaintRibbonButton(w, RIBBON_B3);
                if ( w->scroll_per+w->scroll_pos >= w->scroll_max )
                    PaintRibbonButton(w, RIBBON_B1);
            break;
            case RIBBON_B1:
                w->value = FALSE;
                PaintRibbonButton(w, RIBBON_B1);
                if ( w->scroll_per+w->scroll_pos >= w->scroll_max )
                    PaintRibbonButton(w, RIBBON_B3);
            break;

            /* right buttons */
            case RIBBON_B2:
                w->value = FALSE;
                PaintRibbonButton(w, RIBBON_B2);
                if ( w->scroll_pos <= 0 )
                    PaintRibbonButton(w, RIBBON_B4);
            break;
            case RIBBON_B4:
                w->value = FALSE;
                PaintRibbonButton(w, RIBBON_B4);
                if ( w->scroll_pos <= 0 )
                    PaintRibbonButton(w, RIBBON_B2);
            break;

            /* must be a double click */
            case RIBBON_STRIP:
                if ( w->callback )
                {
                    cb_data.reason = RIBBON_B5;
                    cb_data.event  = eventP;
                    AxCallback(w->callback,(AXWidget)w,w->callbacktag,
												(caddr_t)&cb_data);
                }
            break;

            /* option menu (never seen - handled in PaintOptionPulldown) */
            case RIBBON_OP:
            break;
        }
        w->armed = 0;
        RibbonArmed = NULL;
    }
}

AxXmToggleColorpal(w)
AXWidget_main w;   /* W_MAIN containing color palette */
{
    if ( w && w->type == W_MAIN && w->colorpal )
    {
        w->colorpal->display = !w->colorpal->display;
#if AXMS
        RebuildGDIWindow( w, NULL );
#else
        RebuildMainWindow(w, NULL);
#endif
        if ( w->colorpal->display ) {
            PaintColorpal(w->colorpal,(XExposeEvent *)NULL, TRUE, TRUE);
            if (w->colorpal->widgets) {
                PaintRibbon((AXWidget_ribbon)w->colorpal->widgets, 
												NULL, TRUE, TRUE);
			}
        }
    }
}





#define NUM_RIBBON_NAMES 5
/*************************************************************************
function: XmCreateRibbon - Create color ribbon for GE
*/
AXWidget_ribbon XmCreateRibbon (parentWidget,name,args,cnt, winID, pixed)
register AXWidget parentWidget;
register char *name;
register Arg args[];
register int cnt;
int winID;
int pixed;
{
    register AXWidget_ribbon w;
    int ix, len;
    char *names[NUM_RIBBON_NAMES];

    if ( args[0].value )
    {
        w = (AXWidget_ribbon)args[0].value;
        w->update_widget = TRUE;
        WidgetNeedsUpdated = TRUE;
    }
    else
    {
        w = (AXWidget_ribbon)CrtWidget(W_RIBBON,parentWidget,"");
        AddToWindowList((AXWidget)w);
        SetWidgetName((AXWidget)w, name);
    }
    AxXtSetValues((AXWidget)w,args,cnt);
    w->painter = PaintRibbon;
#if AXMS
	w->thickness = 1;
#else
	w->thickness = 2;
#endif

    for (ix = 0; ix < NUM_RIBBON_NAMES; ix++) {
        names[ix] = (char *) 0;
    }
    
    if ( pixed )
    {
        len = strlen(XLT("Brush fg")) + 1;
        names[0] = (char *) malloc(len);
        strcpy(names[0], XLT("Brush fg"));
        
        len = strlen(XLT("Brush bg")) + 1;
        names[1] = (char *) malloc(len);
        strcpy(names[1], XLT("Brush bg"));
        
        AttachOptionMenuToRibbon(w, names, 2, winID);
    }
    else
    {
        len = strlen(XLT("Fill fg")) + 1;
        names[0] = (char *) malloc(len);
        strcpy(names[0], XLT("Fill fg"));
        
        len = strlen(XLT("Fill bg")) + 1;
        names[1] = (char *) malloc(len);
        strcpy(names[1], XLT("Fill bg"));
        
        len = strlen(XLT("Line fg")) + 1;
        names[2] = (char *) malloc(len);
        strcpy(names[2], XLT("Line fg"));
        
        len = strlen(XLT("Line bg")) + 1;
        names[3] = (char *) malloc(len);
        strcpy(names[3], XLT("Line bg"));
        
        len = strlen(XLT("Shadow")) + 1;
        names[4] = (char *) malloc(len);
        strcpy(names[4], XLT("Shadow"));
        
        AttachOptionMenuToRibbon(w, names, 5, winID);
    }
    
    for (ix = 0; ix < NUM_RIBBON_NAMES; ix++) {
        if (names[ix])
            free((char *) names[ix]);
    }
    
    return( (AXWidget_ribbon)w );
}

