/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmscale.c
   Applix ASC version: 27
   Date: Mon Nov 16 13:53:42 1998
*/
#include <basicwin.h>
#if AX_X11
#include <X11/keysym.h>
#endif
#include <ax_xm.h>
#include <aos.h>
#include <xmlib.h>
#include <elxwins.h>


struct XMlibStr {
    PIXEL_VALUE FGpixel;    /* foreground color for app work areas */
    PIXEL_VALUE BGpixel;    /* backround color for app work areas */
    int FullRegenForUpdate;  /* Set by UnMapWidget,
                                            hint for WidgetNeedsUpdated */
};

struct XMlibStr xmlibStr;

#define XmCR_BUTTON_PRESS	10


/* externs */
extern Display *Dpy;
extern int AxCharAbove, AxCharBelow, AxCharHeight;
extern AXWidget_scale ScaleArmed;
extern XFontStruct *Grafont, *BoldTxfont;
extern GC ForeGC, BackGC;
extern GC SunGC, ShadowGC;
extern GC GXForeGC;
extern int WidgetNeedsUpdated;  /* Set when a widget needs repainted */
extern int UseWidgetColors;

#if AX_X11    /* { don't need OpenLook stuff on Windows NT */
extern GC olButtonGC, olHilitButtonGC;
extern GC olClrBackButtonGC, olClrSunButtonGC, olClrShadowButtonGC;
extern GC olClrRecessButtonGC;
#endif


#include <xmscale.p>



/*************************************************************************
function: ArmScale
*/
void
ArmScale(w, eventP)
AXWidget_scale w;
XEvent *eventP;
{
    int xpos, val;
    XmScaleCallbackStruct data;

    if ( !w )
        return;

    if ( eventP->type == ButtonPress ) {
        if ( ScaleArmed )
            return;
        ScaleArmed = w;
        xpos = ScaleXpos(w, eventP->xbutton.x);
        w->armed = ScaleArmType(w, xpos);
        if ( w->armed == XmCR_VALUE_CHANGED )
		{
			if ( w->flags & WFLG_MOD_RETURN )
			{
				data.value = w->value;
				data.reason = XmCR_BUTTON_PRESS;
				if ( w->callback )
					(*w->callback)(w, w->callbacktag, &data);
			}

			w->changed = FALSE;
        }
        else {
            val = w->value;
            if ( w->armed == XmCR_DECREMENT ) {
                val = MAX(w->minValue, val-1);
                val = MIN(w->maxValue, val);
            }
            else if ( w->armed == XmCR_INCREMENT ) {
                val = MAX(w->minValue, val+1);
                val = MIN(w->maxValue, val);
            }
            if ( val != w->value ) {
                MoveScale(w, val);  /* this updates w->value */
#if 0
                if ( w->flags & WFLG_MOD_RETURN ) {
                    data.value = w->value;
                    data.reason = w->armed;
                    if ( w->callback )
                        (*w->callback)(w, w->callbacktag, &data);
                }
#endif
            }
        }
    }
    else if ( eventP->type == MotionNotify ) {
        xpos = ScaleXpos(w, eventP->xmotion.x);
        if ( (w->armed == XmCR_VALUE_CHANGED) && (w->elevator_pos != xpos) ) {
            w->changed = TRUE;
            val = ScaleXposToVal(w, xpos);
            MoveScale(w, val);
            if ( w->flags & WFLG_MOD_RETURN ) {
                data.value = w->value;
                data.reason = XmCR_DRAG;
                if ( w->callback ) {
                    AxCallback(w->callback,(AXWidget)w, w->callbacktag, 
														(caddr_t)&data);
				}
            }
        }
    }
    else if ( eventP->type == ButtonRelease ) {
        if ( (w->armed == XmCR_VALUE_CHANGED) && w->changed ) {
            w->changed = FALSE;
            xpos = ScaleXpos(w, eventP->xbutton.x);
            if ( w->elevator_pos != xpos ) {
                val = ScaleXposToVal(w, xpos);
                MoveScale(w, val);
            }
        }
        data.value = w->value;
        data.reason = XmCR_VALUE_CHANGED;
        if ( w->callback )
            AxCallback(w->callback,(AXWidget)w, w->callbacktag,(caddr_t)&data);
        w->armed = 0;
        ScaleArmed = NULL;
    }
}

/*************************************************************************
function: MoveScale - Update widget and display based on new value
*/
void
MoveScale(w, new_val)
AXWidget_scale w;
int new_val;
{
    int new_pos;
    int old_pos, old_val;
    int x, y, width, height;
    int dest_x, dest_y;
    int fill_x, fill_wid;

    old_pos = w->elevator_pos;
    old_val = w->value;
    w->value = new_val;
    new_pos = ScaleValToXpos(w);
    w->elevator_pos = new_pos;

    x = w->xpos + old_pos - 12;
    y = w->ypos + 2;
    width = 25;
    height = w->height - 4;
    dest_x = w->xpos + new_pos - 12;
    dest_y = y;
    fill_x = x;
    fill_wid = width;
    if ( ABS(old_pos-new_pos) <= 25 ) {
        if ( old_pos > new_pos ) {
            fill_x = w->xpos + new_pos + 13;
            fill_wid = old_pos - new_pos;
        }
        else {
            fill_wid = new_pos - old_pos;
        }
    }

    XCopyArea(Dpy, WIN_OF(w), WIN_OF(w), GXForeGC, x, y, width, height,
              dest_x, dest_y);
    
    xmColorFillRectangle((AXWidget)w, WIN_OF(w), BackGC, fill_x, y,
                         fill_wid, height);

    UpdateScaleLabel(w);
}

/*************************************************************************
function: PaintScale
*/
void
PaintScale(w, exP, full_regen, layered)
AXWidget_scale w;
XExposeEvent *exP;
int full_regen;
int layered;
{
    int x, y, width, height;

    if ( !w || (w->type != W_SCALE) )
        return;

    if ( !w->display )
        return;

    if ( w->parent && (w->parent->type == W_DIALOG) &&
                !((AXWidget_dialog) w->parent)->exposed )
        return;

    if (layered) {
        if (w->flags & WFLG_IN_EDIT) {
            if (dePaintWidget((AXWidget)w)) {
                return;
            }
        }
        else {
            if ((AXWidget_scale)topWidget((AXWidget)w) == w) {
                paintCompositeWidget((AXWidget)w, exP, full_regen);
                return;
            }
        }
    }

    w->mapped = TRUE;

    x = w->xpos;
    y = w->ypos;
    width = w->width;
    height = w->height;
    w->elevator_pos = ScaleValToXpos(w); /* position relative to w->xpos */

    if ( full_regen || exP ) {
        xmColorFillRectangle((AXWidget)w, WIN_OF(w), BackGC, x,y,width,height);
    }

    if ( full_regen || exP ) {
        AxDrawBevel((AXWidget)w, WIN_OF(w), x, y, width, height, 2, BEVEL_IN);
    }
        /* pass null to DrawBevel to force usage of default wkit colors */
    AxDrawBevel((AXWidget) NULL, WIN_OF(w), x+w->elevator_pos-12,
                y+2, 25, height-4, 2, BEVEL_OUT);
    XFillRectangle(Dpy, WIN_OF(w), ForeGC, x+w->elevator_pos-10,
                   y+4, 21, height-8);
    XDrawLine(Dpy, WIN_OF(w), SunGC, x+w->elevator_pos, y+3,
              x+w->elevator_pos, y+height-3);

    UpdateScaleLabel(w);
}

/*************************************************************************
function: ScaleXposToVal - Returns value corresponding to scale-relative xpos
*/
static int
ScaleXposToVal(w, xpos)
AXWidget_scale w;
int xpos;     /* relative to w->xpos */
{
    int val = 0;

    xpos -= 14;
    xpos = MAX(0, xpos);    /* make sure position isn't negative */
    val = w->minValue + (w->maxValue-w->minValue)*xpos/(w->width-28);
    val = MAX(w->minValue, val);
    val = MIN(w->maxValue, val);
    return(val);
}

/*************************************************************************
function: ScaleValToXpos - Returns scale-relative xpos corresponding to value
*/
static int
ScaleValToXpos(w)
AXWidget_scale w;
{
    int xpos = 0;

    /* Motif: xpos is the x-position of the center of the elevator.
     * It's actual xpos is w->xpos+14 (14 is 2+2+10).  Thus, the
     * value range (w->maxValue-w->minValue) is mapped to the width
     * (w->width-28).  The xpos returned is relative to the left edge
     * of the scale; hence the +14. */
    
    /* get xpos corresponding to value.  This xpos is relative
     * to the width range (w->width-28)
	 */
	if (0 == (w->maxValue - w->minValue)) {
		xpos = 0;
	}
	else {
		xpos = (w->value - w->minValue) * (w->width - 28) /
			(w->maxValue - w->minValue);
	}
    xpos += 14;              /* make xpos relative to left edge of scale */
    xpos = MIN(xpos, w->width-15);  /* trim any overflow in calculations */
    return(xpos);
}

/*************************************************************************
function: ScaleXpos - Returns scale-relative xpos from window-relative xpos
*/
static int
ScaleXpos(w, x)
AXWidget_scale w;
int x;
{
    int xpos = 0;

    xpos = x - w->xpos;
    xpos = MAX(0, xpos);
    xpos = MIN(w->width, xpos);

    return(xpos);
}


/*************************************************************************
function: ScaleArmType - Returns callback type based on scale-relative xpos
*/
static int
ScaleArmType(w, xpos)
AXWidget_scale w;
int xpos;    /* relative to w->xpos */
{
    int armed = 0;

    /* w->elevator_pos is center of elevator */
    if ( xpos < (w->elevator_pos-12) )
        armed = XmCR_DECREMENT;
    else if ( xpos > (w->elevator_pos+12) )
        armed = XmCR_INCREMENT;
    else
        armed = XmCR_VALUE_CHANGED;
    return(armed);
}

/*************************************************************************
function: UpdateScaleLabel - Lift old label and draw new
*/
static void
UpdateScaleLabel(w)
AXWidget_scale w;  /* W_SCALE */
{
    char valstr[100];
    int valwid;
    Arg args[10];
    int n;
    int label_x, label_y;
    int actualCreate = FALSE;

#if 0
/* !!0 - this stuff needs to be fully supported, meaning either the
 * the scale widget needs to optionally create&manage its own entry field,
 * or the entry widget needs to be generalized to include an optional
 * slider for numeric fields.. until then, so much for labels.. */

    if ( w->no_label ) {
        if ( w->label && w->label->mapped )
            UnMapWidget(w->label);
        return;
    }

    sprintf(valstr, "%d", w->value);
    valwid = BasicStrlen(valstr);
    label_x = w->xpos + w->elevator_pos - valwid/2;
    label_y = w->ypos - AxCharHeight - 2;

    if ( w->label ) {
        /* clear if pos changed; XmCreateLabel clears if string changed */
        if ( (w->label->xpos != label_x) || (w->label->ypos != label_y) )
            UnMapWidget(w->label);
    }

    n = 0;
    XtSetArg(args[n], XmNoldWidget, w->label); n++;
    XtSetArg(args[n], XmNflags, w->flags); n++;
    XtSetArg(args[n], XmNx, label_x); n++;
    XtSetArg(args[n], XmNy, label_y); n++;
    if (!w->label) {
        actualCreate = TRUE;
    }
    w->label = (AXWidget_Scale) XmCreateLabelGadget(w->parent, valstr, args, n);
    if (actualCreate) {
            /* dlg editor support */
        if (w->event_table) {
                /* give the newly created scrollbar an event table
                 * which matches that of the list widget */
            copyEventHandlerList(w->event_table,
                        &w->label->event_table);
        }
        XtSetArg(args[0], ZXmNcompositeLinkUp, w);
        XtSetArg(args[1], ZXmNinEdit, w->flags & WFLG_IN_EDIT);
        AxXtSetValues(w->label, args, 2);
        XtSetArg(args[0], ZXmNcompositeLinkDown, w->label);
        AxXtSetValues(w, args, 1);
    }
#endif /* suppress until upgraded */
}

int
xmScaleWidth(w)
AXWidget w;
{
    int width;
    
    /* 2-pix thick border around scale; 2-pix thick border around
     * elevator; elevator 11 wide and 8 high (excluding border) */
    width = MAX((2+2+10+1+10+2+2), w->width);
    return (width);
}

int
xmScaleHeight(w)
AXWidget w;
{
    int height;
    
    /* 2-pix thick border around scale; 2-pix thick border around
     * elevator; elevator 11 wide and 8 high (excluding border) */
    height = MAX((2+2+8+2+2), w->height);
    return (height);
}

/************************************************************
function: XmCreateScale - Create scale/slider
*/
AXWidget_scale
XmCreateScale(parentWidget, name, args, cnt)
register AXWidget parentWidget;
register char *name;
register Arg args[];
register int cnt;
{
    AXWidget_scale w;
    AXWidget backW;
    int temp;
    int oldx, oldy, oldwid, oldht;
    int oldminval, oldmaxval, oldval, oldnolabel;

    if ( args[0].value ) {
        w = (AXWidget_scale)args[0].value;
        SetWidgetName((AXWidget)w, name);
        oldx = w->xpos;
        oldy = w->ypos;
        oldwid = w->width;
        oldht = w->height;
        oldminval = w->minValue;
        oldmaxval = w->maxValue;
        oldval = w->value;
        oldnolabel = w->no_label;
    }
    else
        w = (AXWidget_scale)CrtWidget(W_SCALE, parentWidget, name);

    w->painter = PaintScale;
    AxXtSetValues((AXWidget)w, args, cnt);
    w->width = xmScaleWidth((AXWidget)w);
    w->height = xmScaleHeight((AXWidget)w);
    if ( w->maxValue < w->minValue ) {
        /* swap min and max */
        temp = w->maxValue;
        w->maxValue = w->minValue;
        w->minValue = temp;
    }
    else if ( w->maxValue == w->minValue ) {
        w->maxValue += 100;  /* !! */
    }
    w->value = MAX(w->value, w->minValue);
    w->value = MIN(w->value, w->maxValue);

    /* HACK: manage widget from here to avoid flicker
     * if pos/size/value changed compared to old. */
    if ( args[0].value && w->mapped ) {
        if ( (oldx != w->xpos) || (oldy != w->ypos) ||
            (oldwid != w->width) || (oldht != w->height) ||
            (oldminval != w->minValue) || (oldmaxval != w->maxValue) ||
            (oldnolabel != w->no_label) ) {
            
            /* unmap widget
             * (can't use UnMapWidget because pos/size updated..)
             */
            backW = (AXWidget)backgroundWidget((AXWidget)w);
            xmColorFillRectangle(backW, WIN_OF(w), BackGC,
                                 oldx, oldy, oldwid + 1, oldht + 1);
            w->mapped = FALSE;
            w->update_widget = TRUE;
            WidgetNeedsUpdated = TRUE;
            xmlibStr.FullRegenForUpdate = TRUE;/* need to regen full widget */
        }
        else if ( oldval != w->value ) {
            temp = w->value;
            w->value = oldval;
            MoveScale(w, temp);
        }
    }

    return((AXWidget_scale)w);
}    
 

int
ScaleKeyPressEvent(w, eventP, ksym, kbuf, num_chars)
     AXWidget_scale w;  /* in: widget getting the keypress */
     XEvent *eventP;    /* in: keypress event */
     KeySym ksym;
     wchar *kbuf;       /* in: key string associated with ksym, from LookupString */
     int num_chars;     /* num_chars in kbuf */
{
  /* Just do the right and left arrow keys.  The scale widget
   * is horizontal-only at present, so don't consume vertical arrows.
   */
  int do_callback = FALSE;
  int new_val = w->value;
    
  if (XK_Right == ksym) {
    new_val += (w->maxValue - w->minValue) / 10;
    if (new_val > w->maxValue) {
      new_val = w->maxValue;
    }
    MoveScale(w, new_val);
    w->update_widget = TRUE;
    WidgetNeedsUpdated = TRUE;
    do_callback = TRUE;
  }
  else if (XK_Left == ksym) {
    new_val -= (w->maxValue - w->minValue) / 10;
    if (new_val < w->minValue) {
      new_val = w->minValue;
    }
    MoveScale(w, new_val);
    w->update_widget = TRUE;
    WidgetNeedsUpdated = TRUE;
    do_callback = TRUE;
  }
  if (do_callback) {
    XmScaleCallbackStruct data;
    data.value = w->value;
    data.reason = XmCR_VALUE_CHANGED;
    if ( w->callback )
      AxCallback(w->callback,(AXWidget)w, w->callbacktag, (caddr_t)&data);
    return (TRUE);
  }
  return (FALSE);
}

