/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmtable.h
   Applix ASC version: 20
   Date: Mon May 10 21:13:06 1999
*/
#ifndef _ax_xmtable_h_
#define _ax_xmtable_h_

#include <xmwcolor.h>

typedef struct {
	int type; /* one of the types defined in ``evreason.h''. */
} TblAnyEvent;

typedef struct {
	int type; /* `EV_TBL_DBLCLICK_' */
	int row;
	int col;
} TblDoubleClickEvent;

typedef struct {
	int type; /* `EV_COLUMN_RESIZE_' */
	int column;
	int width;
} TblColumnResizeEvent;

/* When user clicks into a cell or arrows into a cell */
typedef struct {
	int type; /* `EV_TBL_FOCUS_CHANGE_' */
	int from_row;
	int from_col;
	int to_row;
	int to_col;
} TblFocusChangeEvent;

typedef struct {
	int type; /* `EV_TBL_KEYPRESS_' */
	int elfkeyid;
	int printable;
	wchar *wbuf;
	int num_chars;
	int row;
	int col;
} TblKeyPressEvent;

typedef struct {
	int type; /* `EV_TBL_EDIT_BEGIN_' */
	int row;
	int col;
} TblEditBeginEvent;

typedef struct {
	int type; /* `EV_TBL_SELECTION_CHANGE_' */
	int *row_list;
	int num_rows;
} TblSelectionChangeEvent;

typedef struct {
	int type; /* `EV_TBL_REQUEST_DATA_' */
	int starting_row;
	int num_rows;
} TblRequestDataEvent;

typedef struct {
	int type; /* `EV_TBL_BUTTONPRESS_' */
	int row;
	int col;
	int start_char_pos;
	int end_char_pos;
} TblButtonPressEvent;

typedef struct {
     int type;
     int totalCol, totalRow;
	int delete;
} TblRowColCreateEvent;

typedef struct {
	int type; /* `EV_TYPING' */
	int row;
	int col;
} TblTypingEvent;

typedef union {
	TblAnyEvent any;
	TblDoubleClickEvent double_click;
	TblColumnResizeEvent   column_resize;
	TblFocusChangeEvent    focus_change;
	TblKeyPressEvent   keypress;
	TblEditBeginEvent  edit_begin;
	TblSelectionChangeEvent    selection_change;
	TblRequestDataEvent  request_data;
	TblButtonPressEvent  buttonpress;
	TblRowColCreateEvent rowcolcreate;
	TblTypingEvent cell_typing;
} TblEventData;

typedef struct tbl_cursor_cell_t {
	int row;  /* cursor row */
	int col;  /* cursor column */
	int char_pos;     /* start of cursor - 0 based character position */
	int end_char_pos; /* end of cursor   - 0 based character position */
	int cursor_x;     /* window-relative i-bar xpos, most recent display */
	int cursor_y;     /* i-bar ypos */
	int cell_xpos;   /* cursor cell ypos */
	int cell_ypos;   /* cursor cell xpos */
	int cell_width;  /* cursor cell width */
	int cell_height; /* cursor cell height */
	int left_char_pos;  /* left most char of cell */
	int right_char_pos; /* right most char of cell */
	wchar *wtext;
	int text_len; /* number of chars in cursor text */
	unsigned int partial_left_char: 1;
	unsigned int partial_right_char: 1;
	unsigned int text_cursor_on: 1;
	unsigned int cursor_spare3: 1;
	unsigned int cursor_spare4: 1;
	unsigned int curosr_spare5: 1;
	unsigned int cursor_spare6: 1;
	unsigned int cursor_spare7: 1;
} tbl_cursor_cell_t;

typedef struct table_info {
	int topRow;	/* row to display at top of table */
	short leftCol;	/* col to display at left of table */
	data_ptr colText;	/* array of heading strs, one per column */
	data_ptr colWidth;	/* in pixels */
	data_ptr colType;	/* justification, number style... */
	short showRowNumbers;	/* If TRUE, use numeric row ids */
	data_ptr rows;		/* 2-d array of info */
	data_ptr marker_pixmaps; /* 2-d array of info */
	data_ptr corner_pixmaps; /* 1-d array of pixmap names */
	data_ptr marker_strings; /* 1-d array of marker strings */
	int headingHeight;	/* pixel height of column headings */
	int rowHeight;		/* pixel height of each row */
	int markerWidth;	/* pixel width of row markers */
	int winHeight;	/* slightly smaller than widget height */
	int winWidth;	/* slightly smaller than widget width */
	int lineHeight;		/* height of each line */
	int numLines;	/* number of lines in window, rounded up */
	int totalRowCount; /* total number of rows in data set */
	tbl_cursor_cell_t cursor;
	long *selected_rows; /* "IntsArray" of selected (marked) rows */
	long *selected_columns; /* "IntsArray" of selected (marked) columns */
	int colToResize; /* column number that is about to be resized */
	Cursor pointer_shape;
	GC grid_gc;  /* gc for drawing grid. */
	GC mark_gc;  /* gc for drawing marks and cursor. */
	XFontStruct *headingFont; /* font for headings. */
	int heading_char_above; /* for heading font */
	int heading_char_below; /* for heading font */
	int heading_char_height;/* for heading font */
	wcolor_t *headingColor;
	data_ptr headingShadow;
	XFontStruct *font; /* table text font */
	int char_above; /* for table text font */
	int char_below; /* for table text font */
	int char_height;/* for table text font */
	int hScrollOrigin;
	int hScrollLength;
	int vScrollOrigin;
	int vScrollLength;
	char *fontFamily;      /* for table text font */
	int  fontSize;         /* for table text font */
	int  fontWeight;       /* for table text font */
	int  fontSlant;        /* for table text font */
	int right_col;         /* last valid data cell pos */
	int bottom_row;        /* last valid data cell pos */
/* Set'able table attributes.  All default values are FALSE. */
	unsigned int allow_cell_editing : 1;
	unsigned int allow_column_resizing : 1;
	unsigned int suppress_row_markers : 1;
	unsigned int allow_hidden_columns : 1;
	unsigned int allow_column_reordering : 1;
	unsigned int use_data_request_model : 1;
	unsigned int alis_style_scroll : 1;
	unsigned int hide_vertical_grid   : 1;
	unsigned int hide_horizontal_grid : 1;
	unsigned int spare_attr_flag1 : 1;
	unsigned int spare_attr_flag2 : 1;
	unsigned int spare_attr_flag3 : 1;
	unsigned int spare_attr_flag4 : 1;
	unsigned int spare_attr_flag5 : 1;
	unsigned int spare_attr_flag6 : 1;
	unsigned int spare_attr_flag7 : 1;
/* Table state information.  All default values are FALSE. */
	unsigned int seeing_button1_motion : 1;
	unsigned int doing_column_resizing : 1;
	unsigned int doing_cell_editing : 1;
	unsigned int doing_row_marking : 1;
	unsigned int doing_column_marking : 1;
	unsigned int vscroll_dim_changed : 1; /* vscroll{origin,length} changed */
	unsigned int hscroll_dim_changed : 1; /* hscroll{origin,length} changed */
	unsigned int spare_state_flag7 : 1;
	unsigned int rowcolcreate : 1;  /* for optimized data handling */
} table_info;

#endif /* !_ax_xm_table_h_ */


/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
