/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmtray.c
   Applix ASC version: 56
   Date: Tue May 18 21:12:14 1999
*/
#include <basicwin.h>
#include <ax_xm.h>
#include <aos.h>
#include <dwt2xm.h>
#include <xmlib.h>
#include <elxwins.h>
#if AX_X11
#include <X11/keysym.h>
#endif
#include <wchars.h>
#include <thimble.h>
#include <ctype.h>
#include <elf_env.h>
#include <eldata.h>
#include <evreason.h>
#include <xmpixel.h>
#include <xmtray.h> /* Private tray info. */
#include <evtray.h> /* Elf event packaging */
#include <intsarr.h>

/* Standard Externs for Widget Modules */
extern Window AxCrtWindow();
extern AXWidget CrtWidget();
extern pixcache_t *Pixmapper();
extern GC axCreateGC();

extern int AxLanguage;
extern int WidgetNeedsUpdated; /* Set when a widget needs repainted */
extern int AxCharAbove, AxCharHeight, AxCharBelow;
extern int AxOpenLook;
extern int AxBigFont;
extern int UseWidgetColors;
extern Display *Dpy;
extern short DpyDepth;
extern int AxButtonDown;
extern short THIMpid;

extern long WinMask;
extern XSetWindowAttributes WinAtt;
extern Pixmap UnknownPixmap;
extern AXWidget ClickWidget;
extern unsigned long BackWorkPixel;
extern Colormap WorkColormap;
extern GC ForeGC, BackGC;
extern GC SunGC, ShadowGC, DisabledGC, GrayLineGC;
extern GC RecessGC;
extern GC olClrSunButtonGC, olClrShadowButtonGC, olClrRecessButtonGC;
extern GC MonospaceGC;       /* monspace system font, forground mode */
extern GC MonospaceBackGC;   /* monspace system font, background mode */
extern GC MonospaceBoldGC, MonospaceBoldBackGC;
extern GC MonospaceItalGC, MonospaceItalBackGC;
extern GC MonospaceItalBoldGC, MonospaceItalBoldBackGC;
extern XFontStruct *AxDlgMainFont;  /* Bolded dialog font */
extern XFontStruct *MonospaceFont, *MonospaceBoldFont;
extern XFontStruct *MonospaceItalFont, *MonospaceItalBoldFont;
extern int DoubleMillisec;
extern unsigned long ForePixel;
extern unsigned long BackPixel;
extern short ColorDisplay;
extern short NoColor;
extern Window RootWin;
extern Cursor ThimCrosshairCursor;
extern Cursor ThimPointerCursor;
extern Cursor ThimLeftRightCursor;
extern unsigned long AlisPixels[];
extern unsigned long WidgetPixels[];
extern void *AxRenderInsetGraphic();
extern void xmSetWidgetDisplay();
/* End of Standard Externs */

#if !BUILD_WIDE /* move to xmlib.h!! */
#define axDrawString16 axDrawString
#endif

extern AXWidget_tray TrayArmed;

/* Event mask used when the table widget window does not have focus.
   Prevent unnecessary MotionNotify events. */
#define TRAY_NONFOCUS_MASK (TM_ALL_INPUTS_MASK)

/* Event mask used when the table widget window has focus.  Requests
   pointer motion. */
#define TRAY_INFOCUS_MASK  ((TM_ALL_INPUTS_MASK) | PointerMotionMask)

/* Default X and Y pixel increments for unit scrolling. */
#define UNIT_SCROLL_X_INC 38
#define UNIT_SCROLL_Y_INC 38

#define BTN_MARGIN_DIM 1

#define NODE_HEIGHT_EXTENSION 2 /* pixels */

#define NODE_WIDTH_EXTENSION 30 /* pixels */

#define BTN_TXT_SPACING 2 /* used to be 4 */

/* Node spacing in pixels (same for X and Y directions) */
#define NODE_SPACING_DIM  1 /* used to be 2 */

#ifndef DIFF
#define DIFF(x,y) (((x)>(y)) ? ((x)-(y)) : ((y)-(x)))
#endif

#define TESTING 1

/* Forward decls of static functions. */
static void UnmapTrayScrollBars AXP_((AXWidget_tray w, int vscroll, int hscroll));
static void TrayExposeCB AXP_((AXWidget_tray w, XEvent *evp));
static void TrayButtonPressCB AXP_((AXWidget_tray w, XEvent *evp));
static void TrayButtonReleaseCB AXP_((AXWidget_tray w, XEvent *evp));
static void TrayMotionNotifyCB AXP_((AXWidget_tray w, XEvent *evp));
static int  TrayKeyPressCB AXP_((AXWidget_tray w, XEvent *evp));
static void TrayKeyReleaseCB AXP_((AXWidget_tray w, XEvent *evp));
static void TraySelectionRequestCB AXP_((AXWidget_tray w, XEvent *evp));
static void TraySelectionNotifyCB  AXP_((AXWidget_tray w, XEvent *evp));
static void TraySelectionClearCB AXP_((AXWidget_tray w, XEvent *evp));
static void TrayGraphicsExposeCB AXP_((AXWidget_tray w, XEvent *evp));
static void RealizeTray AXP_((AXWidget_tray w, int layered));
static void CreateTrayScrollBars AXP_((AXWidget_tray w, int layered));
static void TrayBarChanged AXP_((Widget widget, AXWidget_tray w,
                                 XmScrollBarCallbackStruct *data));
static void TrayFocusOutCB AXP_((AXWidget_tray w, XEvent *evp));
static void TrayFocusInCB AXP_((AXWidget_tray w, XEvent *evp));
static void SetTrayWorkColors AXP_((AXWidget_tray w));
static void UnsetTrayWorkColors AXP_((AXWidget_tray w));
static void UpdateTrayScrollBars AXP_((AXWidget_tray w));
static void SetClipAllGCs AXP_((AXWidget_tray w, int clip_x_origin,
                                int clip_y_origin, XRectangle *rectangles,
                                int n, int ordering));
static void UnsetClipAllGCs AXP_((AXWidget_tray w));
static void DrawTrayMaterial AXP_((AXWidget_tray w, int x, int y,
                                   unsigned int width, unsigned int height));
static void SetNodeDim AXP_((tray_data *trayP, treeNode *nodeP));
static void FormatNode AXP_((tray_data *trayP, treeNode *nodeP, int x, int y));
static void IFormatNode AXP_((tray_data *trayP, treeNode *nodeP,int x, int y));
static void DrawNodeWiring AXP_((AXWidget_tray w, treeNode *nodeP));
static int  IdxInIntsArray AXP_((long *p, long val));
static void SendDoubleClickEvent AXP_((AXWidget_tray w, treeNode *nodeP));
static void SendButtonPressEvent AXP_((AXWidget_tray w, treeNode *nodeP));
static void SendSelectionChangeEvent AXP_((AXWidget_tray w));
static void CreateTrayGCs AXP_((AXWidget_tray w));
static void FreeTrayGCs AXP_((tray_data *trayP));
static int  SetTrayFont AXP_((AXWidget_tray w, char *family, int size, int
                              weight, int slant));
static void DrawNode AXP_((AXWidget_tray w, treeNode *nodeP, int full_regen));
static void TrayResetAll AXP_((AXWidget_tray w));
static int  PointInNode AXP_((AXWidget_tray w, treeNode *nodeP, int x, int y));
static treeNode *FindNodeUnderPoint AXP_((AXWidget_tray w, int x, int y));
static void TrayXCopyArea AXP_((Display *dpy, Drawable src_drawable,
                                Drawable dst_drawable, GC gc, int src_x,
                                int src_y, unsigned int width,
                                unsigned int height, int dst_x, int dst_y));
static int SelectTrayNode AXP_((AXWidget_tray w, treeNode *nodeP));
static int UnselectTrayNode AXP_((AXWidget_tray w, treeNode *nodeP));
static int NodeIsVisible AXP_((AXWidget_tray w, treeNode *nodeP));
static int NodeIsFullyVisible AXP_((AXWidget_tray w, treeNode *nodeP));
static int NodeInRegion AXP_((AXWidget_tray w, treeNode *nodeP, int x, int y,
                              int width, int height));
static long *FindNodesUnderRegion AXP_((AXWidget_tray w, int x, int y, int width,
                                        int height));
static int FormatTray(AXWidget_tray w);

typedef struct tray_region_t {
    int x, y;
    int xcor, ycor;
} tray_region_t;

static struct btn1_info_t {
    int x, y;
    treeNode *node;
} btn1_info;

static struct tray_motion_rect_t {
    int x, y, width, height;
} motion_rect;

/* Must say (TRAY_REGOVERLAP(r1, r2) || TRAY_REGOVERLAP(r2, r2))
   to test if r1 and r2 overlap. */
#define TRAY_REGOVERLAP(r1, r2) (((r1)->xcor > (r2)->x) && \
                                 ((r1)->x    < (r2)->xcor) && \
                                 ((r1)->ycor > (r2)->y) && \
                                 ((r1)->y    < (r2)->ycor))
/* If r1 fully consumes r2 */
#define TRAY_REGCONSUME(r1, r2) (((r1)->x <= (r2)->x) && \
                                 ((r1)->xcor >= (r2)->xcor) && \
                                 ((r1)->y <= (r2)->y) && \
                                 ((r1)->ycor >= (r2)->ycor))

/***********************************************************************
function: AXmCreateTray - Create a tray widget.
*/
Widget
AXmCreateTray(parentWidget, name, args, cnt)
register AXWidget_dialog parentWidget;
register char *name;
register Arg args[];
register int cnt;
{
    register AXWidget_tray w;

    if (args[0].value) {
        w = (AXWidget_tray) args[0].value;
        if (args[1].value & WFLG_NODISPLAY) {
            UnMapWidget(w);
            if (w->vScrollBar) {
                UnMapWidget(w->vScrollBar);
                xmSetWidgetDisplay(w->vScrollBar, FALSE);
            }
            if (w->hScrollBar) {
                UnMapWidget(w->hScrollBar);
                xmSetWidgetDisplay(w->hScrollBar, FALSE);
            }
            
        } else {
            if (w->vScrollBar) {
                xmSetWidgetDisplay(w->vScrollBar, TRUE);
			}
            if (w->hScrollBar) {
                xmSetWidgetDisplay(w->hScrollBar, TRUE);
			}
					
        }
        w->update_widget = TRUE;
        WidgetNeedsUpdated = TRUE;
    } else {
        w = (AXWidget_tray)CrtWidget(W_TRAY, parentWidget, name);
        w->info.tray = (dummy_info *) calloc(1, sizeof(tray_data));
    }

    w->painter = PaintTray;
    AxXtSetValues(w, args, cnt);
    CreateTrayGCs(w);
    return((Widget)w);
}

/***********************************************************************
function: CreateTrayGCs - Create GCs for text and drawing.
*/
static void
CreateTrayGCs(w)
AXWidget_tray w;
{
    tray_data *trayP;
    int ret = FALSE;
    XFontStruct *xfont;
    XGCValues gcv;
    int scalable;

    if (!w)
        return;
    if (w->type != W_TRAY)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    if (trayP->font.family) {
        ret = getXFontStruct(trayP->font.family, trayP->font.size,
                             trayP->font.weight, trayP->font.slant,
                             &scalable, &trayP->font.font_desc);
    }
    if ((ret != TRUE) || (!trayP->font.family)) {
        /* use default -- dialog font */
        trayP->font.char_above  = AxCharAbove;
        trayP->font.char_height = AxCharHeight;
        trayP->font.char_below  = AxCharBelow;
        trayP->font.font_desc   = AxDlgMainFont;
        trayP->text_gc = ForeGC;
    } else {
        /* use user specified font */
        xfont = trayP->font.font_desc;
        trayP->font.char_above  = xfont->max_bounds.ascent;
        trayP->font.char_below  = xfont->max_bounds.descent;
        trayP->font.char_height = (trayP->font.char_above +
                                   trayP->font.char_below);

        memset((char *) &gcv, 0, sizeof(XGCValues));
        if (!UseWidgetColors || w->useWorkColors) {
            gcv.foreground = ForePixel;
            gcv.background = BackPixel;
        } else {
            /* is a color display and need widget colors */
            gcv.foreground = WidgetPixels[1 /* TMA_BLACK */];
            gcv.background = WidgetPixels[0 /* TMA_WHITE */];
        }
        gcv.fill_style = FillSolid;
        gcv.function = GXcopy;
        gcv.plane_mask = AllPlanes;
        gcv.font = xfont->fid;
        trayP->text_gc = axCreateGC(Dpy, RootWin, DefaultScreen(Dpy),
                                    DisplayPlanes(Dpy, DefaultScreen(Dpy)),
                                    xfont,
                                    (GCForeground  | GCBackground
                                     | GCPlaneMask | GCFunction
                                     | GCFont), &gcv);
        if (!trayP->text_gc) {
            /* create failed -- use default */
            trayP->font.char_above  = AxCharAbove;
            trayP->font.char_height = AxCharHeight;
            trayP->font.char_below  = AxCharBelow;
            trayP->font.font_desc   = AxDlgMainFont;
            trayP->text_gc = ForeGC;
        }
    }

/* Create gc for motion rectangle */
    if (!trayP->motion_gc) {
        XGCValues gcv;

        memset((char *) &gcv, 0, sizeof(XGCValues));
        if (!UseWidgetColors /* !ColorDisplay || NoColor */) {
            gcv.foreground = ForePixel^BackPixel;
            gcv.background = ForePixel^BackPixel;
        } else {
            if (w->useWorkColors) {
                gcv.foreground = AlisPixels[0]^AlisPixels[1];
                gcv.background = AlisPixels[0]^AlisPixels[1];
            } else {
                gcv.foreground = WidgetPixels[0]^WidgetPixels[1];
                gcv.background = WidgetPixels[0]^WidgetPixels[1];
            }
        }
        gcv.fill_style = FillSolid;
        gcv.function = GXxor;
        gcv.plane_mask = AllPlanes;
        gcv.line_width = 0;
        gcv.join_style = JoinRound; /* makes no diff */
        gcv.cap_style  = CapRound;  /* makes no diff */
        trayP->motion_gc = axCreateGC(Dpy, RootWin, DefaultScreen(Dpy),
                                      DisplayPlanes(Dpy, DefaultScreen(Dpy)),
                                      (XFontStruct *) 0,
                                      (GCForeground | GCBackground |
                                       GCPlaneMask | GCFunction |
                                       GCLineWidth | GCJoinStyle |
                                       GCCapStyle), &gcv);
        XSetLineAttributes(Dpy, trayP->motion_gc, 0, LineOnOffDash, CapRound,
                           JoinMiter);
    }

    return;
}

/***********************************************************************
function: FreeTrayGCs - Free all tray related gcs.
*/
static void
FreeTrayGCs(trayP)
tray_data *trayP;
{
    if (!trayP)
        return;
    if (trayP->text_gc) {
        if (trayP->text_gc != ForeGC)
            XFreeGC(Dpy, trayP->text_gc);
        trayP->text_gc = (GC) NULL;
    }
    if (trayP->motion_gc) {
        XFreeGC(Dpy, trayP->motion_gc);
        trayP->motion_gc = (GC) NULL;
    }
    return;
}

/***********************************************************************
function: TrayResetAll - Reset stuff.
*/
static void
TrayResetAll(w)
register AXWidget_tray w;
{
    register tray_data *trayP;
    int num_selections;
    int ix;
    treeNode *thisP;

    if (!w)
        return;
    if (w->type != W_TRAY)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    trayP->last_node_selected = NULL;
	if (trayP->selected_nodes) {
		num_selections = IntsArraySize(trayP->selected_nodes);
	}
	else {
		num_selections = 0;
	}
    for (ix = 0; ix < num_selections; ix++) {
        thisP = (treeNode *) ReadIntsArray(trayP->selected_nodes, ix);
        if (!thisP)
            continue;
        thisP->attr.selected = FALSE;
    }
    if (trayP->selected_nodes)
        free(trayP->selected_nodes);
    trayP->selected_nodes = CreateIntsArray(0);
    trayP->left_x = trayP->top_y = 0;
    trayP->state.btn1_down = FALSE;
    memset((char *) &motion_rect, 0, sizeof(struct tray_motion_rect_t));
    memset((char *) &btn1_info, 0, sizeof(struct btn1_info_t));
    AXDefineDefaultCursor(w, ThimPointerCursor);
    return;
}

/***********************************************************************
function: SetTrayFont - Set font on tray widget.
Will repaint tray if it exists.
Return 0 on success; -1 otherwise
*/
static int
SetTrayFont(w, family, size, weight, slant)
AXWidget_tray w;
char *family;
int size;
int weight;
int slant;
{
    tray_data *trayP;

    if (!w)
        return(-1);
    if (w->type != W_TRAY)
        return(-1);
    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(-1);
    if (VerifyFontSpec(family, size, weight, slant) == -1)
        return(-1);
    if (w->window) { /* tray has been realized */
        FreeTrayGCs(trayP);
        if (trayP->font.family)
            free(trayP->font.family);
        trayP->font.family = (char *) malloc(strlen(family) + 1);
        strcpy(trayP->font.family, family);
        trayP->font.size   = size;
        trayP->font.weight = weight;
        trayP->font.slant  = slant;
        CreateTrayGCs(w);
        trayP->state.needs_formatting = TRUE;
        PaintTray(w, NULL, TRUE, FALSE);
    } else {
        if (trayP->font.family)
            free(trayP->font.family);
        trayP->font.family = (char *) malloc(strlen(family) + 1);
        strcpy(trayP->font.family, family);
        trayP->font.size   = size;
        trayP->font.weight = weight;
        trayP->font.slant  = slant;
    }
    return(0);
}

/***********************************************************************
function: UnmapTrayScrollBars - Unmap scrollbars.
*/
static void
UnmapTrayScrollBars(w, vscroll, hscroll)
AXWidget_tray w;
int vscroll;
int hscroll;
{
    if (!w)
        return;
    if (w->type != W_TRAY)
        return;

    return;
}
        
/***********************************************************************
function: TrayEvent - Handle all types of XEvents in the tray widget.
Return TRUE if event was consumed; FALSE otherwise (or on error).
*/
int
TrayEvent(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    int event_consumed = FALSE;

    if (!w || !evp)
        return(FALSE);
    if (w->type != W_TRAY)
        return(FALSE);

    switch (evp->xany.type) {
    case FocusIn:
        TrayFocusInCB(w, evp);
        event_consumed = TRUE;
        break;
    case FocusOut:
        TrayFocusOutCB(w, evp);
        event_consumed = TRUE;
        break;
    case Expose:
        TrayExposeCB(w, evp);
        event_consumed = TRUE;
        break;
    case GraphicsExpose:
        TrayGraphicsExposeCB(w, evp);
        event_consumed = TRUE;
        break;
    case ButtonPress:
        TrayButtonPressCB(w, evp);
        event_consumed = TRUE;
        break;
    case ButtonRelease:
        TrayButtonReleaseCB(w, evp);
        event_consumed = TRUE;
        break;
    case MotionNotify:
        TrayMotionNotifyCB(w, evp);
        event_consumed = TRUE;
        break;
    case KeyPress:
        if (AxButtonDown) {
            RingTheBell();
            event_consumed = TRUE;
        } else
            event_consumed = TrayKeyPressCB(w, evp);
        break;
    case KeyRelease:
        TrayKeyReleaseCB(w, evp);
        event_consumed = TRUE;
        break;
    case SelectionRequest:
        TraySelectionRequestCB(w, evp);
        event_consumed = FALSE;
        break;
    case SelectionNotify:
        TraySelectionNotifyCB(w, evp);
        event_consumed = FALSE;
        break;
    case SelectionClear:
        TraySelectionClearCB(w, evp);
        event_consumed = FALSE;
        break;
    default:
        event_consumed = FALSE;
        break;
    }
    return(event_consumed);
}

/***********************************************************************
function: TrayExposeCB - Handle an exposure event in the tray widget
*/
static void
TrayExposeCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    if (!w || !evp)
        return;
    if (w->type != W_TRAY)
        return;

    return;
}

/***********************************************************************
function: TrayGraphicsExposeCB - Handle a graphics expose event in the
tray widget.
*/
static void
TrayGraphicsExposeCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    XExposeEvent exp;

    if (!w || !evp)
        return;
    if (w->type != W_TRAY)
        return;
    if (evp->xany.type != GraphicsExpose)
        return;

    /* fake an expose event */
    memset((char *) &exp, 0, sizeof(XExposeEvent));
    exp.type = Expose;
    exp.x = evp->xgraphicsexpose.x;
    exp.y = evp->xgraphicsexpose.y;
    exp.width  = evp->xgraphicsexpose.width;
    exp.height = evp->xgraphicsexpose.height;
    PaintTray(w, &exp, FALSE, TRUE);
    return;
}

/***********************************************************************
function: TrayButtonPressCB - Handle a ButtonPress in the tray.
*/
static void
TrayButtonPressCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    register tray_data *trayP = (tray_data *) NULL;
    register treeNode  *click_node = (treeNode *) NULL;
    register unsigned long time;
    int selection_changed = FALSE;
    int num_selections = 0;
    int work_colors_set = FALSE;
#if LINUX
	char *str;
	XmScrollBarCallbackStruct cbdata;
#endif
	
    if (!w || !evp)
        return;
    if (w->type != W_TRAY)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    memset((char *) &btn1_info, 0, sizeof(struct btn1_info_t));
    memset((char *) &motion_rect, 0, sizeof(struct tray_motion_rect_t));
    motion_rect.x = evp->xbutton.x;
    motion_rect.y = evp->xbutton.y;

    switch (AxButtonDown) {
    case Button1:
        trayP->state.btn1_down = TRUE;
        btn1_info.x = evp->xbutton.x;
        btn1_info.y = evp->xbutton.y;
        time = evp->xbutton.time;
        if (time - w->last_click_time <= DoubleMillisec) {
            if (ThimIsControled() || ThimIsShifted())
                return; /* no double clicks when controled or shifted */

            if (!trayP->last_node_selected)
                return;
            if (!PointInNode(w, trayP->last_node_selected, evp->xbutton.x,
                             evp->xbutton.y))
                goto single_click;

            /* successful double click */
            if (w->callback)
                SendDoubleClickEvent(w, trayP->last_node_selected);

            w->last_click_time = time;
        } else {
single_click:
            click_node = FindNodeUnderPoint(w, evp->xbutton.x, evp->xbutton.y);
            btn1_info.node = click_node;
            if (ThimIsControled()) {
                if (!w->multi_pick)
                    return;
                if (!click_node)
                    return;

                SetTrayWorkColors(w);
                work_colors_set = TRUE;
                if (trayP->selected_nodes &&
					InIntsArray(trayP->selected_nodes, NODE_ID(click_node))) {
					
                    /* unselect it */
                    if (UnselectTrayNode(w, click_node))
                        selection_changed = TRUE;
                    DrawNode(w, click_node, TRUE);
                } else {
                    if (SelectTrayNode(w, click_node))
                        selection_changed = TRUE;
                    DrawNode(w, click_node, TRUE);
                }
                UnsetTrayWorkColors(w);
                work_colors_set = FALSE;
                goto finish_button_press;
            } else if (ThimIsShifted()) {
                if (!w->multi_pick)
                    return;
                if (!click_node)
                    return;
                return; /* no selection extend */
            } else {
                /* clear all selections */

				if (trayP->selected_nodes) {
					long *old_selections;
					int num_selections;
					int ix;
					
					old_selections = CopyIntsArray(trayP->selected_nodes);
					num_selections = IntsArraySize(old_selections);
					SetTrayWorkColors(w);
					work_colors_set = TRUE;
					if (UnselectTrayNode(w, NULL)) /* clear all selections */
						selection_changed = TRUE;
					for (ix = 0; ix < num_selections; ix++) {
						treeNode *thisP;

						thisP = (treeNode *) ReadIntsArray(old_selections, ix);
						if (!thisP)
							continue;
						if (NodeIsVisible(w, thisP))
							DrawNode(w, thisP, TRUE);
					}
					free(old_selections);
					UnsetTrayWorkColors(w);
					work_colors_set = FALSE;
				}

                if (!click_node)
                    goto finish_button_press;

                if (work_colors_set == FALSE) {
                    SetTrayWorkColors(w);
                    work_colors_set = TRUE;
                }
                SelectTrayNode(w, click_node);
                trayP->last_node_selected = click_node;
                DrawNode(w, click_node, TRUE);
                UnsetTrayWorkColors(w);
                work_colors_set = FALSE;
                selection_changed = TRUE;

finish_button_press:
                if (work_colors_set)
                    UnsetTrayWorkColors(w);
                work_colors_set = FALSE;
                if (w->callback && selection_changed)
                    SendSelectionChangeEvent(w);
            }
            w->last_click_time = time;
        }
        break;
#if LINUX
	case Button4:
	case Button5:
		str = (char *)UPbyName(NXLT("axMouseWheelScroll"));
		if (str && (atoi(str) != 0)) {
			if (AxButtonDown == Button4) { /* mouse scroll up */
				if (ThimIsShifted())
					cbdata.reason = XmCR_PAGE_DECREMENT;
				else
					cbdata.reason = XmCR_DECREMENT;
			} else if (AxButtonDown == Button5) { /* mouse scroll down */
				if (ThimIsShifted())
					cbdata.reason = XmCR_PAGE_INCREMENT;
				else
					cbdata.reason = XmCR_INCREMENT;
			}
			TrayBarChanged(w->vScrollBar, w, &cbdata);
		}
		break;
#endif
    default:
        break;
    }
    if (work_colors_set)
        UnsetTrayWorkColors(w);
    return;
}

/***********************************************************************
function: TrayButtonReleaseCB - Handle a button release in the tray.
*/
static void
TrayButtonReleaseCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    tray_data *trayP;
    long *node_list = (long *) NULL;
    int num_nodes;
    int ix;
    int selections_changed = FALSE;

    if (!w || !evp)
        return;
    if (w->type != W_TRAY)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    if (trayP->state.btn1_down) {
        if (motion_rect.width || motion_rect.height)
            XDrawRectangle(Dpy, w->window, trayP->motion_gc, motion_rect.x,
                           motion_rect.y, motion_rect.width,
                           motion_rect.height);
        node_list = FindNodesUnderRegion(w, motion_rect.x, motion_rect.y,
                                         motion_rect.width,
                                         motion_rect.height);
        num_nodes = IntsArraySize(node_list);
        if (num_nodes) {
            SetTrayWorkColors(w);
            for (ix = 0; ix < num_nodes; ix++) {
                treeNode *thisP;

                thisP = (treeNode *) ReadIntsArray(node_list, ix);
                if (!thisP)
                    continue;
                SelectTrayNode(w, thisP);
                DrawNode(w, thisP, TRUE);
            }
            UnsetTrayWorkColors(w);
            selections_changed = TRUE;
        }
    }
    if (node_list)
        free(node_list);
    trayP->state.btn1_down = FALSE;
    memset((char *) &motion_rect, 0, sizeof(struct tray_motion_rect_t));
    memset((char *) &btn1_info, 0, sizeof(struct btn1_info_t));

    if (selections_changed && w->callback)
        SendSelectionChangeEvent(w);

    return;
}

/***********************************************************************
function: TrayMotionNotifyCB - Handle a motion notify in the tray.
*/
static void
TrayMotionNotifyCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    tray_data *trayP;
    int new_x, new_y, new_width, new_height;
    int ev_x, ev_y;

    if (!w || !evp)
        return;
    if (w->type != W_TRAY)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    if (!trayP->state.btn1_down)
        return;

/* button 1 is down */
    if (btn1_info.node)
        return; /* button press had happened on a node.  don't do
                   motion-rect stuff.
                   TODO, for this case, drag the button (for drag-n-drop) */
    if (evp->xmotion.x > (trayP->win_width - 2))
        ev_x = trayP->win_width - 2;
    else if (evp->xmotion.x < 2)
        ev_x = 2;
    else
        ev_x = evp->xmotion.x;

    if (evp->xmotion.y > (trayP->win_height - 2))
        ev_y = trayP->win_height - 2;
    else if (evp->xmotion.y < 2)
        ev_y = 2;
    else
        ev_y = evp->xmotion.y;

    if (motion_rect.width || motion_rect.height)
        XDrawRectangle(Dpy, w->window, trayP->motion_gc, motion_rect.x,
                       motion_rect.y, motion_rect.width,
                       motion_rect.height);
    /* compute new x, y, width, height */
    new_x = MIN(btn1_info.x, ev_x);
    new_y = MIN(btn1_info.y, ev_y);
    new_width = DIFF(btn1_info.x, ev_x);
    new_height = DIFF(btn1_info.y, ev_y);
    motion_rect.x = new_x;
    motion_rect.y = new_y;
    motion_rect.width = new_width;
    motion_rect.height = new_height;
    XDrawRectangle(Dpy, w->window, trayP->motion_gc, motion_rect.x,
                   motion_rect.y, motion_rect.width,
                   motion_rect.height);
    return;
}

/***********************************************************************
function: TrayKeyPressCB - Handle a key press in the tray.
Return TRUE is event was consumed; FALSE otherwise
*/
static int
TrayKeyPressCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    int event_consumed = FALSE;
    int num_chars = 0;
    wchar *wbuf;
    KeySym ksym;
    XKeyEvent *kev;
    tray_data *trayP = (tray_data *) NULL;
    treeNode *topP = (treeNode *) NULL;
	treeNode *newSelNode = NULL;
	treeNode *nodeP = NULL;
	int idx;
	int navigator_key = FALSE;

    if (!w || !evp)
        return(event_consumed);
    if (w->type != W_TRAY)
        return(event_consumed);
    if (evp->xany.type != KeyPress)
        return(event_consumed);

    num_chars = AXLookupString16(evp, &wbuf, &ksym);
    if ((num_chars == 0) && ((ksym == 0) || AxKeysymIsModifier(ksym)))
      return(event_consumed); /* ignore shift/control... */

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(event_consumed);
	topP = (treeNode *) ReadIntsArray(trayP->top_node_list, 0);
	if (!topP)
		return(event_consumed);
	if (!topP->u.cache)
		return(event_consumed); /* empty tray */
	if (IntsArraySize(topP->u.cache) <= 0)
		return(event_consumed); /* empty tray */

	switch (ksym) {
	case XK_Up:
		event_consumed = TRUE;
		navigator_key = TRUE;
		if (!trayP->selected_nodes) {
			if (topP->u.cache)
				newSelNode = (treeNode *) ReadIntsArray(topP->u.cache, 0);
			else
				newSelNode = NULL;
			break;
		}
		if (IntsArraySize(trayP->selected_nodes) <= 0) {
			newSelNode = NULL;
			newSelNode = (treeNode *) ReadIntsArray(topP->u.cache, 0);
			break;
		}
		newSelNode = (treeNode *) ReadIntsArray(trayP->selected_nodes, 0);
		idx = IdxInIntsArray(topP->u.cache, newSelNode);
		if (idx < 0) {
			newSelNode = NULL;
			break;
		}
		if (idx > 0)
			newSelNode = (treeNode *) ReadIntsArray(topP->u.cache, idx - 1);
		else
			newSelNode = NULL;
		break;
	case XK_Down:
		event_consumed = TRUE;
		navigator_key = TRUE;
		if (!trayP->selected_nodes) {
			if (topP->u.cache)
				newSelNode = (treeNode *) ReadIntsArray(topP->u.cache, 0);
			else
				newSelNode = NULL;
			break;
		}
		if (IntsArraySize(trayP->selected_nodes) <= 0) {
			newSelNode = NULL;
			newSelNode = (treeNode *) ReadIntsArray(topP->u.cache, 0);
			break;
		}
		newSelNode = (treeNode *) ReadIntsArray(trayP->selected_nodes, 0);
		idx = IdxInIntsArray(topP->u.cache, newSelNode);
		if (idx < 0) {
			newSelNode = NULL;
			break;
		}
		if (idx >= (IntsArraySize(topP->u.cache) - 1))
			newSelNode = NULL;
		else
			newSelNode = (treeNode *) ReadIntsArray(topP->u.cache, idx + 1);
		break;
	default:
		break;
	}

	if (!navigator_key)
		return(event_consumed);
/* navigation stuff... */
	if (newSelNode != NULL) {
		long *old_selections;
		int nx, num_selections;

		if (!trayP->selected_nodes) {
			num_selections = 0;
			old_selections = NULL;
		} else {
			old_selections = CopyIntsArray(trayP->selected_nodes);
			num_selections = IntsArraySize(old_selections);
		}

		UnselectTrayNode(w, NULL); /* clear all selections */

		SetTrayWorkColors(w);
		for (nx = 0; nx < num_selections; nx++) {
			treeNode *np;
			np = (treeNode *) ReadIntsArray(old_selections, nx);
			if (!np)
				continue;
			if (NodeIsVisible(w, np))
				DrawNode(w, np, TRUE);
		}
		UnsetTrayWorkColors(w);
		if (old_selections)
			free(old_selections);

		SelectTrayNode(w, newSelNode);

		if (NodeIsFullyVisible(w, newSelNode)) {
			SetTrayWorkColors(w);
			DrawNode(w, newSelNode, TRUE);
			UnsetTrayWorkColors(w);
		} else { /* node is not visible -- make it visible */
			XmScrollBarCallbackStruct data;

			data.event = NULL; /* unused */
			data.pixel = 0; /* unused */
			data.reason = XmCR_VALUE_CHANGED;
			data.value = newSelNode->info.ypos - (trayP->win_height / 2);
			TrayBarChanged(w->vScrollBar, w, &data);
		}
		SendSelectionChangeEvent(w);
	}
    return(event_consumed);
}

/***********************************************************************
function: TrayKeyReleaseCB - Handle a key release in the tray.
*/
static void
TrayKeyReleaseCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    if (!w || !evp)
        return;
    if (w->type != W_TRAY)
        return;

    return;
}

/***********************************************************************
function: TraySelectionRequestCB - Handle a selection request in the tray
*/
static void
TraySelectionRequestCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    if (!w || !evp)
        return;
    if (w->type != W_TRAY)
        return;

    return;
}

/***********************************************************************
function: TraySelectionNotifyCB - Handle a selection notify request in
the tray
*/
static void
TraySelectionNotifyCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    if (!w || !evp)
        return;
    if (w->type != W_TRAY)
        return;

    return;
}

/***********************************************************************
function: TraySelectionClearCB - Handle a selection clear in the tray
*/
static void
TraySelectionClearCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    if (!w || !evp)
        return;
    if (w->type != W_TRAY)
        return;

    return;
}

/***********************************************************************
function: TrayFocusInCB - Handle a FocusIn in the tray.
*/
static void
TrayFocusInCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    if (!w || !evp)
        return;
    if (w->type != W_TRAY)
        return;

    return;
}

/***********************************************************************
function: TrayFocusOutCB - Handle a FocusOut in the tray.
*/
static void
TrayFocusOutCB(w, evp)
register AXWidget_tray w;
register XEvent *evp;
{
    if (!w || !evp)
        return;
    if (w->type != W_TRAY)
        return;

    return;
}

/***********************************************************************
function: DrawTrayBorder - Draw tray etch border around tray.
*/
void
DrawTrayBorder(w)
register AXWidget_tray w;
{
    int thickness;

    if (!w)
        return;
    if (w->type != W_TRAY)
        return;
    if (!WIN_OF(w))
        return;

    if (w->thickness < 0) {
        thickness = -w->thickness;
        AxDrawShadow(WIN_OF(w), ShadowGC, SunGC, thickness, w->xpos,
                   w->ypos, w->width, w->height, OLSHADOW_NODRAW);
    } else if (w->thickness) {
        thickness = w->thickness;
        AxDrawShadow(WIN_OF(w), SunGC, ShadowGC, thickness, w->xpos, w->ypos,
                   w->width, w->height, OLSHADOW_NODRAW);
    }
    return;
}

/***********************************************************************
function: PaintTray - Paint the tray.
*/
static void
PaintTray(w, exP, regen, layered)
register AXWidget_tray w;
XExposeEvent *exP;
int regen;
int layered;
{
    tray_data *trayP;
    int realize_tray = FALSE;
    int sb_dims_changed = FALSE;

    if (!w)
        return;
    if (w->type != W_TRAY)
        return;

    if (!w->display)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    if ((w->flags & WFLG_IN_EDIT) && layered) {
        if (dePaintWidget(w))
            return;
    }

    if (!w->mapped)
        w->mapped = TRUE;

    if (!w->window) {
        RealizeTray(w, layered); /* create window, scrolls... */
        realize_tray = TRUE;
    }

    /* Recreate scrollbars if dimensions of scrollbars changed.
       Don't bother if realizing tray (gets taken care of...). */
    if (!realize_tray &&
        (trayP->state.vscroll_dim_changed ||
         trayP->state.hscroll_dim_changed)) {
        UnmapTrayScrollBars(w, trayP->state.vscroll_dim_changed ? TRUE : FALSE,
                            trayP->state.hscroll_dim_changed ? TRUE : FALSE);
        CreateTrayScrollBars(w, layered);
        sb_dims_changed = TRUE;
    } else
        sb_dims_changed = FALSE;

    if (exP) {
        XRectangle crect[1];

        crect[0].x = 0;
        crect[0].y = 0;
        crect[0].width  = exP->width;
        crect[0].height = exP->height;
        SetClipAllGCs(w, exP->x, exP->y, crect, 1, YXBanded);
    }

    DrawTrayBorder(w);

    SetTrayWorkColors(w);
    if (regen && !exP)
        XFillRectangle(Dpy, w->window, BackGC, 0, 0, trayP->win_width,
                       trayP->win_height); /* Clear whole surface */

/* Draw in material */
    if (!regen && exP)
        DrawTrayMaterial(w, exP->x, exP->y, exP->width, exP->height);
    else
        DrawTrayMaterial(w, 0, 0, 0, 0); /* draw whole tray */

    if (w->graphic)
        AxRenderInsetGraphic(w, w->graphic);

/* Reset clips if necessary */
    if (trayP->state.clips_are_on)
        UnsetClipAllGCs(w);

    UnsetTrayWorkColors(w);

/* Update scrollbars if necessary */
    if (regen || sb_dims_changed)
        UpdateTrayScrollBars(w);

    return;
}

/***********************************************************************
function: ComputeTrayMaxHeight - Compute max height of tray.  Assumes tray
has been formatted.
*/
ComputeTrayMaxHeight(w)
AXWidget_tray w;
{
    tray_data *trayP = (tray_data *) NULL;
    int cnt;
    treeNode *topP;

    if (!w)
        goto use_default;
    if (w->type != W_TRAY)
        goto use_default;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        goto use_default;
    cnt = IntsArraySize(trayP->top_node_list);
    if (!cnt)
        goto use_default;
    topP = (treeNode *) ReadIntsArray(trayP->top_node_list, 0);
    if (!topP)
        goto use_default;

    return(topP->total_height + (2 * NODE_SPACING_DIM));

use_default:
    return(2 * NODE_SPACING_DIM);
}

/***********************************************************************
function: ComputeTrayMaxWidth - Compute max width of tray.  Assumes
tray has been formatted.
*/
ComputeTrayMaxWidth(w)
AXWidget_tray w;
{
    tray_data *trayP = (tray_data *) NULL;
    int cnt;
    treeNode *topP;

    if (!w)
        goto use_default;
    if (w->type != W_TRAY)
        goto use_default;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        goto use_default;
    cnt = IntsArraySize(trayP->top_node_list);
    if (!cnt)
        goto use_default;

    /* Assuming one top node */
    topP = (treeNode *) ReadIntsArray(trayP->top_node_list, 0);
    return(topP->total_width + (4 * NODE_SPACING_DIM));

use_default:
    return(2 * NODE_SPACING_DIM);
}

/***********************************************************************
function: ComputeTrayUnitY - Compute pixel amount for vertical unit scroll.
Assumes tray has been formatted.
*/
ComputeTrayUnitY(w)
AXWidget_tray w;
{
    tray_data *trayP = (tray_data *) NULL;
    treeNode *topP;
    treeNode *thisP;

    if (!w)
        return(UNIT_SCROLL_Y_INC);
    if (w->type != W_TRAY)
        return(UNIT_SCROLL_Y_INC);

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(UNIT_SCROLL_Y_INC);
    if (!IntsArraySize(trayP->top_node_list))
        goto use_default;
    topP = (treeNode *) ReadIntsArray(trayP->top_node_list, 0);
    if (!topP)
        goto use_default;
    thisP = (treeNode *) ReadIntsArray(topP->u.cache, 0);
    if (!thisP)
        goto use_default;
    if (thisP->info.height == 0)
        goto use_default; /* tray not formatted */
    return(thisP->info.height);

use_default:
    return(UNIT_SCROLL_Y_INC);
}

/***********************************************************************
function: CompareNodesAlpha - Compare two nodes. Called by qsort(3).
*/
CompareNodesAlpha(node1, node2)
treeNode **node1;
treeNode **node2;
{
    int case_sensitive;

    /* may be passed in the future... */

    case_sensitive = FALSE;

    if (!node1 || !node2)
        return(0);
    if (!*node1 || !*node2)
        return(0);

#if 0
    return(strcmp((*node1)->info.title, (*node2)->info.title));
#else
    return(Istrcmp((unsigned char *)(*node1)->info.title,
      (unsigned char *)(*node2)->info.title, AxLanguage, case_sensitive));
#endif
}

/***********************************************************************
function: CompareNodesPosition - Compare two nodes, using y-pos for
comparison.  Called by qsort(3).
*/
CompareNodesPosition(node1, node2)
treeNode **node1;
treeNode **node2;
{
    if (!node1 || !node2)
        return(0);
    if (!*node1 || !*node2)
        return(0);
    return(((*node1)->info.ypos) - ((*node2)->info.ypos));
}

/***********************************************************************
function: CompareNodesFtype - Compare two nodes using file-type for
comparison.
*/
CompareNodesFtype(node1, node2)
treeNode **node1;
treeNode **node2;
{
    return(0); /* unimpl */
}

/***********************************************************************
function: SortNode - Recursively sort all subnodes of a node.
*/
SortNode(nodeP, compar)
treeNode *nodeP;
int (*compar)();
{
    treeNode *thisP = (treeNode *) NULL;
    int cnt;
    int ix;
    long *sub_nodes;

    if (!compar)
        return(0);
    cnt = IntsArraySize(nodeP->sub_nodes);
    if (cnt < 2)
        return(0);
    sub_nodes = (long *) &nodeP->sub_nodes[1];

    qsort((char *) sub_nodes, cnt, sizeof(long), compar);
    for (ix = 0; ix < cnt; ix++)
        SortNode((treeNode *) ReadIntsArray(nodeP->sub_nodes, ix), compar);
    return(0);
}

/***********************************************************************
function: SortTrayNodes - Sort all nodes in the tray.
*/
SortTrayNodes(w)
AXWidget_tray w;
{
    tray_data *trayP = (tray_data *) NULL;
    treeNode *topP;

    if (!w)
        return(-1);
    if (w->type != W_TRAY)
        return(-1);
    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(-1);
    if (!IntsArraySize(trayP->top_node_list))
        return(0);
    /* assuming 1 top node */
    topP = (treeNode *) ReadIntsArray(trayP->top_node_list, 0);
    if (!topP)
        return(0);
    
    if (trayP->attr.skip_top_node)
    {
        int num_topnodes;
        int nx;

        num_topnodes = IntsArraySize(topP->sub_nodes);
        for (nx = 0; nx < num_topnodes; nx++)
            SortNode((treeNode *) ReadIntsArray(topP->sub_nodes, nx),
                     CompareNodesAlpha);
    }
    else
        SortNode(topP, CompareNodesAlpha);
    
    return(0);
}

/***********************************************************************
function: SortCache - Sort caches of all top nodes.
Note: The cache must not be in use!! Nodes must already be formatted.
*/
SortCache(w)
AXWidget_tray w;
{
    tray_data *trayP = (tray_data *) NULL;
    treeNode *topP = (treeNode *) NULL;
    int ntop;
    int ix;
    long *cache;
    int size;

    if (!w)
        return(-1);
    if (w->type != W_TRAY)
        return(-1);
    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(-1);
    ntop = IntsArraySize(trayP->top_node_list);
    if (!ntop)
        return(0);
    for (ix = 0; ix < ntop; ix++) {
        topP = (treeNode *) ReadIntsArray(trayP->top_node_list, ix);
        if (!topP)
            continue;
        if (topP->level != 0)
            continue; /* huh? */
        size = IntsArraySize(topP->u.cache);
        if (size < 2)
            continue;
        cache = (long *) &topP->u.cache[1];
        qsort((char *) cache, size, sizeof(long), CompareNodesPosition);
    }
    return(0);
}

/***********************************************************************
function: ComputeTrayUnitX - Compute pixel amount for horiz unit scroll.
Assumes tray has been formatted.
*/
ComputeTrayUnitX(w)
AXWidget_tray w;
{
    tray_data *trayP = (tray_data *) NULL;
    treeNode *topP;
    treeNode *thisP;

    if (!w)
        return(UNIT_SCROLL_X_INC);
    if (w->type != W_TRAY)
        return(UNIT_SCROLL_X_INC);

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(UNIT_SCROLL_X_INC);
    if (!IntsArraySize(trayP->top_node_list))
        goto use_default;
    topP = (treeNode *) ReadIntsArray(trayP->top_node_list, 0);
    if (!topP)
        goto use_default;
    thisP = (treeNode *) ReadIntsArray(topP->u.cache, 0);
    if (!thisP)
        goto use_default;
    if (thisP->info.width == 0)
        goto use_default; /* tray not formatted */
    return(thisP->info.width);

use_default:
    return(UNIT_SCROLL_X_INC);
}

/***********************************************************************
function: IdxInIntsArray - Find the position of VAL in the IntsArray P.
Return -1 on failure or if not found.
*/
static int
IdxInIntsArray(p, val)
register long *p;
register long val;
{
    register int idx;
	register int size;

    if (!p)
        return(-1);
    size = *p++;
	for (idx = 0; idx < size; idx++)
		if (*p++ == val)
			return(idx);
    return(-1);
}

/***********************************************************************
function: FindPrevNode - Return the previous node (not necessarily the
parent).  Return NULL on error.
FIXME, this function must go into `tree.c'.
*/
treeNode *
FindPrevNode(nodeP)
register treeNode *nodeP;
{
    register treeNode *parentP;
    int idx;

    if (!nodeP)
        return((treeNode *) NULL);
    if (nodeP->level == 0)
        return((treeNode *) NULL); /* top node has no previous */

    parentP = nodeP->u.parent;
    if (!parentP)
        return((treeNode *) NULL); /* err */

    idx = IdxInIntsArray(parentP->sub_nodes, nodeP);
    if (idx == -1)
        return((treeNode *) NULL); /* err */
    if (idx == 0)
        return(parentP);
    return((treeNode *) ReadIntsArray(parentP->sub_nodes, idx - 1));
}

/***********************************************************************
function: DrawNodeWiring - Draw all wiring associated with this node NODEP.
*/
static void
DrawNodeWiring(w, nodeP)
AXWidget_tray w;
treeNode *nodeP;
{
    treeNode *lastNodeP = (treeNode *) NULL;
    tray_data *trayP = (tray_data *) NULL;
    int num_subnodes;
    unsigned long saved_fg;
    static unsigned long blue_pixel = (unsigned long) -1;

    if (!nodeP)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    num_subnodes = IntsArraySize(nodeP->sub_nodes);
    if (num_subnodes)
        lastNodeP = (treeNode *) ReadIntsArray(nodeP->sub_nodes,
                                               num_subnodes - 1);

/* Note, must reset saved_fg back into ForeGC.  Do 'goto finish;' to bailout */
    saved_fg = AXGetForegroundPixel(Dpy, ForeGC);
    if (blue_pixel == -1)
        AxGetRealPixel(0, 0, 0, 200, 0, 255, &blue_pixel);
    XSetForeground(Dpy, ForeGC, blue_pixel);

    if (num_subnodes && lastNodeP) {
        int x1, y1, x2, y2;

        x1 = (nodeP->info.xpos - trayP->left_x)+
            (int) (nodeP->info.pixmap_width / 2);
        if (x1 > (int) trayP->win_width)
            goto do_horiz; /* won't see it */
        y1 = (nodeP->info.ypos - trayP->top_y) + (int) nodeP->info.height;
        if (y1 > (int) trayP->win_height)
            goto do_horiz; /* won't see it */
        x2 = x1;
        y2 = (lastNodeP->info.ypos - trayP->top_y) +
            (int) (lastNodeP->info.height / 2);
        if (y2 < 0)
            goto do_horiz; /* won't see it */
        /* thick vertical line */
        if (y2 > (int) trayP->win_height)
            y2 = (int) trayP->win_height;
        if (y1 < 0)
            y1 = 0;
/*        XDrawLine(Dpy, w->window, DisabledGC, x1 - 1, y1, x2 - 1, y2); */
        XDrawLine(Dpy, w->window, ForeGC, x1, y1, x2, y2);
/*        XDrawLine(Dpy, w->window, DisabledGC, x1 + 1, y1, x2 + 1, y2); */
    }

do_horiz:
    if ((nodeP->level != 0) && (nodeP->u.parent)) {
        treeNode *parentP;
        int x1, y1, x2, y2;

        parentP = nodeP->u.parent;
        x1 = (parentP->info.xpos - trayP->left_x) +
            (int) (parentP->info.pixmap_width / 2);
        if (x1 > (int) trayP->win_width)
            goto finish; /* won't see it */
        x2 = (nodeP->info.xpos - trayP->left_x - (2 * BTN_MARGIN_DIM));
        y1 = y2 = (nodeP->info.ypos - trayP->top_y) +
            (int) (nodeP->info.height / 2);
        if (y1 > (int) trayP->win_height)
            goto finish; /* won't see it */
        if (y1 < 0)
            goto finish; /* won't see it */

        /* thick horizontal line */
        if (x1 < 0)
            x1 = 0;
        if (x1 > (int) trayP->win_width)
            goto finish;
        if (x2 > (int) trayP->win_width)
            x2 = trayP->win_width;
/*        XDrawLine(Dpy, w->window, DisabledGC, x1, y1 - 1, x2, y2 - 1); */
        XDrawLine(Dpy, w->window, ForeGC, x1, y1, x2, y2);
/*        XDrawLine(Dpy, w->window, DisabledGC, x1, y1 + 1, x2, y2 + 1); */
    }

finish:
    XSetForeground(Dpy, ForeGC, saved_fg);
    return;
}

/***********************************************************************
function: PointInNode - Return TRUE if point is inside a node NODEP;
FALSE otherwise or on error.
Assumes tray has been formatted.
*/
static int
PointInNode(w, nodeP, x, y)
register AXWidget_tray w;
register treeNode *nodeP;
register int x;
register int y;
{
    tray_data *trayP;
    int node_x, node_y;

    if (!w)
        return(FALSE);
    if (w->type != W_TRAY)
        return(FALSE);
    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(FALSE);
    node_x = nodeP->info.xpos - trayP->left_x;
    node_y = nodeP->info.ypos - trayP->top_y;

    if (x < node_x)
        return(FALSE);
    if (y < node_y)
        return(FALSE);
    if (x < (int) (node_x + nodeP->info.width))
        if (y < (int) (node_y + nodeP->info.height))
            return(TRUE);
    return(FALSE);
}

/***********************************************************************
function: NodeInRegion - Return TRUE if node NODEP and the specified
region overlap; FALSE otherwise.
Assumes tray has been formatted.
*/
static int
NodeInRegion(w, nodeP, x, y, width, height)
AXWidget_tray w;
treeNode *nodeP;
int x;
int y;
int width;
int height;
{
    tray_data *trayP;
    tray_region_t r1, r2;

    if (!w)
        return(FALSE);
    if (w->type != W_TRAY)
        return(FALSE);
    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(FALSE);

    /* node */
    r1.x = nodeP->info.xpos - trayP->left_x;
    r1.y = nodeP->info.ypos - trayP->top_y;
    r1.xcor = r1.x + nodeP->info.width;
    r1.ycor = r1.y + nodeP->info.height;

    /* box */
    r2.x = x;
    r2.y = y;
    r2.xcor = x + width;
    r2.ycor = y + height;

#if 0
    if (TRAY_REGOVERLAP(&r1, &r2) || TRAY_REGOVERLAP(&r2, &r1))
        return(TRUE);
#else
    if (TRAY_REGCONSUME(&r2, &r1))
        return(TRUE);
#endif
    return(FALSE);
}

/***********************************************************************
function: FindNodeUnderPoint - Return (treeNode *) of the node under
the point; NULL if there is'nt any node.
Assumes tray has been formatted.
*/
static treeNode *
FindNodeUnderPoint(w, x, y)
register AXWidget_tray w;
register int x;
register int y;
{
    register tray_data *trayP;
    register int node_x, node_y;
    register treeNode *nodeP = (treeNode *) NULL;
    register int tx;
    int num_top;

    if (!w)
        return((treeNode *) NULL);
    if (w->type != W_TRAY)
        return((treeNode *) NULL);
    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return((treeNode *) NULL);

    num_top = IntsArraySize(trayP->top_node_list);
    for (tx = 0; tx < num_top; tx++) {
        treeNode *this_top;
        int num_nodes;
        int nx;

        this_top = (treeNode *) ReadIntsArray(trayP->top_node_list, tx);
        if (!this_top)
            continue;

        num_nodes = IntsArraySize(this_top->u.cache);
        for (nx = 0; nx < num_nodes; nx++) {
            treeNode *this_node;

            this_node = (treeNode *) ReadIntsArray(this_top->u.cache, nx);
            if (!this_node)
                continue;

            if (PointInNode(w, this_node, x, y))
                return((treeNode *) this_node);
        }
    }
    return((treeNode *) NULL);
}

/***********************************************************************
function: FindNodesUnderRegion - 
*/
static long *
FindNodesUnderRegion(w, x, y, width, height)
AXWidget_tray w;
int x;
int y;
int width;
int height;
{
    tray_data *trayP;
    long *ret_node_list = (long *) NULL;
    treeNode *nodeP = (treeNode *) NULL;
    int tx;
    int num_top;

    if (!w)
        goto no_nodes;
    if (w->type != W_TRAY)
        goto no_nodes;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        goto no_nodes;

    num_top = IntsArraySize(trayP->top_node_list);
    for (tx = 0; tx < num_top; tx++) {
        treeNode *this_top;
        int num_nodes;
        int nx;

        this_top = (treeNode *) ReadIntsArray(trayP->top_node_list, tx);
        if (!this_top)
            continue;

        num_nodes = IntsArraySize(this_top->u.cache);
        for (nx = 0; nx < num_nodes; nx++) {
            treeNode *this_node;

            this_node = (treeNode *) ReadIntsArray(this_top->u.cache, nx);
            if (!this_node)
                continue;

            if (NodeInRegion(w, this_node, x, y, width, height)) {
                long *new_node_list;

                new_node_list = AddIntsArray(ret_node_list,
                                             NODE_ID(this_node));
                if (ret_node_list)
                    free(ret_node_list);
                ret_node_list = new_node_list;
            }
        }
    }

    if (!ret_node_list)
        goto no_nodes;

    return(ret_node_list);

no_nodes:
    return(CreateIntsArray(0));
}

/***********************************************************************
function: NodeIsVisible - Return TRUE if NODEP is visible; FALSE otherwise.
*/
static int
NodeIsVisible(w, nodeP)
AXWidget_tray w;
treeNode *nodeP;
{
    tray_data *trayP;
    int real_x, real_y;

    if (!w)
        return(FALSE);
    if (w->type != W_TRAY)
        return(FALSE);

	if (!nodeP)
		return(FALSE);
    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(FALSE);

    real_x = nodeP->info.xpos - trayP->left_x;
    real_y = nodeP->info.ypos - trayP->top_y;
#if 0 /* does not account for partially visible nodes. e.g. nodes that have
         their left halves clipped when scrollbar is dragged to the right */
    if (real_x < 0)
        return(FALSE);
    if (real_y < 0)
        return(FALSE);
    if (real_x > trayP->win_width)
        return(FALSE);
    if (real_y > trayP->win_height)
        return(FALSE);
#else
    if ((real_x < 0) && ((real_x + (int) nodeP->info.width) < 0))
        return(FALSE);
    if ((real_y < 0) && ((real_y + (int) nodeP->info.height) < 0))
        return(FALSE);
    if (real_x > trayP->win_width)
        return(FALSE);
    if (real_y > trayP->win_height)
        return(FALSE);
#endif
    return(TRUE);
}

/***********************************************************************
function: NodeIsFullyVisible - Return TRUE if NODEP is visible in *full*;
FALSE otherwise. **Only check against the height. If the node is only too
wide, that's ok.**
*/
static int
NodeIsFullyVisible(w, nodeP)
AXWidget_tray w;
treeNode *nodeP;
{
	tray_data *trayP;
	int realx, realy;

	if (!w)
		return(FALSE);
	if (w->type != W_TRAY)
		return(FALSE);
	if (!nodeP)
		return(FALSE);
    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(FALSE);
	realx = nodeP->info.xpos - trayP->left_x;
	realy =  nodeP->info.ypos - trayP->top_y;
	if (realx < 0)
		return(FALSE);
	if (realy < 0)
		return(FALSE);
	if (realy > trayP->win_height)
		return(FALSE);
	if ((realy + (int) nodeP->info.height) > trayP->win_height)
		return(FALSE);
	return(TRUE);
}

/***********************************************************************
function: DrawTrayMaterial - Draw material in tray.
*/
static void
DrawTrayMaterial(w, x, y, width, height)
register AXWidget_tray w;
int x;
int y;
unsigned int width;
unsigned int height;
{
    tray_data *trayP = (tray_data *) NULL;
    treeNode *thisTop;
    treeNode *thisNode;
    int numTopNodes;
    int ix;
    int size;

    if (!w)
        return;
    if (w->type != W_TRAY)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    if ((width <= 0) || (width > trayP->win_width))
        width = trayP->win_width;
    if ((height <= 0) || (height > trayP->win_height))
        height = trayP->win_height;

    if (!trayP->top_node_list) {
        /* tray is empty -- just clear the whole work area */
        XFillRectangle(Dpy, w->window, BackGC, 0, 0, trayP->win_width,
                       trayP->win_height); /* Clear whole surface */
        return;
    }
    if (trayP->state.needs_formatting) {
        if (trayP->attr.sort_nodes)
            SortTrayNodes(w);
        FormatTray(w);
        SortCache(w);
        trayP->max_height = ComputeTrayMaxHeight(w);
        trayP->max_width  = ComputeTrayMaxWidth(w);
        trayP->incY = ComputeTrayUnitY(w);
        trayP->incX = ComputeTrayUnitX(w);
        if (trayP->max_width <= trayP->win_width)
            trayP->left_x = 0;
        if (trayP->max_height <= trayP->win_height)
            trayP->top_y = 0;
        UpdateTrayScrollBars(w);
        trayP->state.needs_formatting = FALSE;
    }
    /* Just iterate over the cache and paint nodes */
    thisTop = (treeNode *) ReadIntsArray(trayP->top_node_list, 0);
    if (!thisTop) {
        /* tray is empty -- just clear the whole work area */
        XFillRectangle(Dpy, w->window, BackGC, 0, 0, trayP->win_width,
                       trayP->win_height); /* Clear whole surface */
        return;
    }
    size = IntsArraySize(thisTop->u.cache);
    for (ix = 0; ix < size; ix++) {
        int real_x, real_y;
        pixcache_t *pixent;
        thisNode = (treeNode *) ReadIntsArray(thisTop->u.cache, ix);
        if (!thisNode)
            continue;
        real_x = thisNode->info.xpos - trayP->left_x;
        real_y = thisNode->info.ypos - trayP->top_y;
        if (((int) (real_y + thisNode->info.height)) < y)
            continue;
        if (real_y >= ((int) (y + height)))
            break;
        /* node is visible -- draw it. */
        DrawNode(w, thisNode, FALSE);
    }

    /* draw node wiring */
    if (trayP->attr.show_node_wiring) {
        for (ix = 0; ix < size; ix++) {
            int real_x, real_y;

            thisNode = (treeNode *) ReadIntsArray(thisTop->u.cache, ix);
            real_x = thisNode->info.xpos - trayP->left_x;
            real_y = thisNode->info.ypos - trayP->top_y;

            if (thisNode->level == 0) {
                if (((int) (real_y + thisNode->total_height)) < y)
                    continue;
                if (real_y >= ((int) (y + thisNode->total_height)))
                    break;
            }
            DrawNodeWiring(w, thisNode);
        }
    }
        
    return;
}

/***********************************************************************
function: DrawNode - Draw a node.  Assumes caller has made sure that
node is visible and formatting has been done.
*/
static void
DrawNode(w, nodeP, full_regen)
register AXWidget_tray w;
register treeNode *nodeP;
int full_regen;
{
    int real_x, real_y;
    pixcache_t *pixent;
    tray_data  *trayP;
    int pixy, stry;
    unsigned long save_fg;
    unsigned long save_bg;

    if (!w)
        return;
    if (!nodeP)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    real_x = nodeP->info.xpos - trayP->left_x;
    real_y = nodeP->info.ypos - trayP->top_y;

    if (full_regen)
        XFillRectangle(Dpy, w->window, BackGC, real_x, real_y,
                       nodeP->info.width, nodeP->info.height);

    if (nodeP->attr.selected) {
        int hilit_x = real_x + nodeP->info.pixmap_width + BTN_TXT_SPACING;
        int hilit_width = (nodeP->info.width - 1 - nodeP->info.pixmap_width +
                           BTN_TXT_SPACING);
#if 0
        XFillRectangle(Dpy, w->window, ForeGC,
                       nodeP->info.xpos - trayP->left_x,
                       nodeP->info.ypos - trayP->top_y,
                       nodeP->info.width - 1,
                       nodeP->info.height - 1);
#else
        XFillRectangle(Dpy, w->window, ForeGC,
                       (nodeP->info.xpos - trayP->left_x +
                        nodeP->info.pixmap_width + BTN_TXT_SPACING),
                       nodeP->info.ypos - trayP->top_y,
                       (nodeP->info.width - 1 - nodeP->info.pixmap_width -
                        BTN_TXT_SPACING),
                       nodeP->info.height - 1);
#endif 
    }
    /* draw icon pixmap */
    if (nodeP->info.icon_pixmap &&
        strlen(nodeP->info.icon_pixmap) &&
        (pixent = Pixmapper(nodeP->info.icon_pixmap))) {
        pixy = (real_y + (nodeP->info.height / 2)) - (pixent->height / 2);
        if (pixent->clip_mask) {
            XSetClipOrigin(Dpy, ForeGC, real_x + BTN_MARGIN_DIM, pixy);
            XSetClipMask(Dpy, ForeGC, pixent->clip_mask);
        }

        XCopyArea(Dpy, pixent->pix, w->window, ForeGC, 0, 0,
                  pixent->width, pixent->height,
                  real_x + BTN_MARGIN_DIM, pixy);

        if (pixent->clip_mask)
            XSetClipMask(Dpy, ForeGC, None);

/*      AxDrawShadow(w->window, SunGC, ShadowGC, 2, real_x , real_y,
                   nodeP->info.pixmap_width,
                   nodeP->info.pixmap_height, OLSHADOW_NODRAW); */
        real_x += nodeP->info.pixmap_width + BTN_TXT_SPACING;
    }

    stry = (real_y + (nodeP->info.height / 2)) - (trayP->font.char_height / 2);
    if (nodeP->attr.selected) { /* set fg and bg colors */
        save_fg = AXGetForegroundPixel(Dpy, trayP->text_gc);
        save_bg = AXGetBackgroundPixel(Dpy, trayP->text_gc);
        XSetForeground(Dpy, trayP->text_gc, AXGetForegroundPixel(Dpy, BackGC));
        XSetBackground(Dpy, trayP->text_gc, AXGetBackgroundPixel(Dpy, BackGC));
    }
    axDrawString(Dpy, w->window, trayP->text_gc, trayP->font.font_desc,
                 real_x, (stry + trayP->font.char_above),
                 nodeP->info.title, strlen(nodeP->info.title));
    if (nodeP->attr.selected) { /* restore fg and bg colors */
        XSetForeground(Dpy, trayP->text_gc, save_fg);
        XSetBackground(Dpy, trayP->text_gc, save_bg);
    }
    /* draw node border */
    if (trayP->attr.show_borders && !nodeP->attr.selected)
        XDrawRectangle(Dpy, w->window, DisabledGC,
                       nodeP->info.xpos - trayP->left_x,
                       nodeP->info.ypos - trayP->top_y,
                       nodeP->info.width,
                       nodeP->info.height);

    return;
}

/***********************************************************************
function: TrayBarChanged - Callback routine for Tray bar change.
*/
static void
TrayBarChanged(widget, w, data)
Widget widget;
AXWidget_tray w;
XmScrollBarCallbackStruct *data;
{
    tray_data *trayP = (tray_data *) NULL;
    int oldTop;
    int oldLeft;
    int span;
    int send_expose = FALSE;
    XExposeEvent fake_expose;

    if (!w)
        return;
    if (w->type != W_TRAY)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    oldTop  = trayP->top_y;
    oldLeft = trayP->left_x;
    memset((char *) &fake_expose, 0, sizeof(XExposeEvent));
    if (w->vScrollBar == (AXWidget_scroll) widget) {
        switch (data->reason) {
        case XmCR_INCREMENT:
            trayP->top_y += trayP->incY;
            trayP->top_y = MIN(trayP->top_y,
                              trayP->max_height - trayP->win_height);
            span = trayP->top_y - oldTop;
            if (span == 0)
                return;
            SetTrayWorkColors(w);
            TrayXCopyArea(Dpy, w->window, w->window, ForeGC, 0, span,
                          trayP->win_width, trayP->win_height - span, 0, 0);
            XFillRectangle(Dpy, w->window, BackGC, 0,
                           trayP->win_height - span, trayP->win_width, span);
            UnsetTrayWorkColors(w);
            send_expose = TRUE;
            fake_expose.x = 0;
            fake_expose.y = trayP->win_height - span;
            fake_expose.width = trayP->win_width;
            fake_expose.height = span;
            break;
        case XmCR_DECREMENT:
            trayP->top_y -= trayP->incY;
            trayP->top_y = MAX(0, trayP->top_y);
            span = oldTop - trayP->top_y;
            if (span == 0)
                return;
            SetTrayWorkColors(w);
            TrayXCopyArea(Dpy, w->window, w->window, ForeGC, 0, 0,
                          trayP->win_width, trayP->win_height - span, 0, span);
            XFillRectangle(Dpy, w->window, BackGC, 0, 0,
                           trayP->win_width, span);
            UnsetTrayWorkColors(w);
            send_expose = TRUE;
            fake_expose.x = 0;
            fake_expose.y = 0;
            fake_expose.width = trayP->win_width;
            fake_expose.height = span;
            break;
        case XmCR_PAGE_INCREMENT:
            trayP->top_y += trayP->win_height;
            trayP->top_y = MIN(trayP->top_y,
                              trayP->max_height - trayP->win_height);
            break;
        case XmCR_PAGE_DECREMENT:
            trayP->top_y -= trayP->win_height;
            trayP->top_y = MAX(0, trayP->top_y);
            break;
        case XmCR_VALUE_CHANGED:
            if (data) {
                trayP->top_y = data->value;
				if (trayP->top_y < 0)
					trayP->top_y = 0;
				else
					trayP->top_y = MIN(trayP->top_y,
									   trayP->max_height - trayP->win_height);
			}
            span = trayP->top_y - oldTop;
            if (span == 0)
                return;
            if (span > 0) {
                if (span > trayP->win_height)
                    break;
                SetTrayWorkColors(w);
                TrayXCopyArea(Dpy, w->window, w->window, ForeGC, 0, span,
                              trayP->win_width, trayP->win_height - span,
                              0, 0);
                XFillRectangle(Dpy, w->window, BackGC, 0,
                               trayP->win_height - span, trayP->win_width,
                               span);
                UnsetTrayWorkColors(w);
                fake_expose.x = 0;
                fake_expose.y = trayP->win_height - span;
                fake_expose.width = trayP->win_width;
                fake_expose.height = span;
            } else {
                span = -span;
                if (span > trayP->win_height)
                    break;
                SetTrayWorkColors(w);
                TrayXCopyArea(Dpy, w->window, w->window, ForeGC, 0, 0,
                              trayP->win_width, trayP->win_height - span,
                              0, span);
                XFillRectangle(Dpy, w->window, BackGC, 0, 0,
                               trayP->win_width, span);
                UnsetTrayWorkColors(w);
                fake_expose.x = 0;
                fake_expose.y = 0;
                fake_expose.width = trayP->win_width;
                fake_expose.height = span;
            }
            send_expose = TRUE;
            break;
        default:
            break;
        }
    } else if (w->hScrollBar == (AXWidget_scroll) widget) {
        switch (data->reason) {
        case XmCR_INCREMENT:
            trayP->left_x += trayP->incX;
            trayP->left_x = MIN(trayP->left_x,
                               trayP->max_width - trayP->win_width);
            break;
        case XmCR_DECREMENT:
            trayP->left_x -= trayP->incX;
            trayP->left_x = MAX(0, trayP->left_x);
            break;
        case XmCR_PAGE_INCREMENT:
            trayP->left_x += trayP->win_width;
            trayP->left_x = MIN(trayP->left_x,
                               trayP->max_width - trayP->win_width);
            break;
        case XmCR_PAGE_DECREMENT:
            trayP->left_x -= trayP->win_width;
            trayP->left_x = MAX(0, trayP->left_x);
            break;
        case XmCR_VALUE_CHANGED:
            if (data)
                trayP->left_x = data->value;
            break;
        default:
            break;
        }
    }
    UpdateTrayScrollBars(w);

    if (send_expose)
        PaintTray(w, fake_expose, FALSE, FALSE);
    else
        PaintTray(w, NULL, TRUE, TRUE);
    return;
}

/***********************************************************************
function: UpdateTrayScrollBars - Get scrollbars upto sync.
*/
static void
UpdateTrayScrollBars(w)
register AXWidget_tray w;
{
    tray_data *trayP;

    if (!w)
        return;
    if (w->type != W_TRAY)
        return;
    if (!w->vScrollBar && !w->hScrollBar)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;
    
/* Adjust vertical scrollbar */
    if (w->vScrollBar) {
        w->vScrollBar->scroll_per = trayP->win_height;
        w->vScrollBar->scroll_max = trayP->max_height;
        w->vScrollBar->scroll_pos = trayP->top_y;
        PaintScrollBar(w->vScrollBar, NULL, TRUE, FALSE);
    }

/* Adjust horizontal scrollbar */
    if (w->hScrollBar) {
        w->hScrollBar->scroll_per = trayP->win_width;
        w->hScrollBar->scroll_max = trayP->max_width;
        w->hScrollBar->scroll_pos = trayP->left_x;
        PaintScrollBar(w->hScrollBar, NULL, TRUE, FALSE);
    }
    return;
}

/***********************************************************************
function: RealizeTray - Create tray window & scroll bars.
*/
static void
RealizeTray(w, layered)
register AXWidget_tray w;
int layered;
{
    int xpos, ypos, win_xpos, win_ypos;
    int width, height, win_width, win_height;
    unsigned long mask;
    XSetWindowAttributes atts;
    int thickness = w->thickness;
    tray_data *trayP = (tray_data *) NULL;

    if (!w)
        return;
    if (w->type != W_TRAY)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    if (w->thickness < 0)
        thickness = -thickness;

    xpos = w->xpos;
    ypos = w->ypos;
    width  = w->width;
    height = w->height;
    win_xpos = xpos + thickness;
    win_ypos = ypos + thickness;
    win_width  = width - (2 * thickness);
    win_height = height - (2 * thickness);

    trayP->win_width  = win_width;
    trayP->win_height = win_height;

    mask = WinMask;
    atts = WinAtt;
    if (w->useWorkColors) {
        mask |= (CWBackPixel | CWColormap);
        atts.background_pixel = BackWorkPixel;
        atts.colormap = WorkColormap;
    }

    if (w->flags & WFLG_IN_EDIT) { /* dialog editor */
        w->window = AxCreateQueryPixmap(Dpy, w->parent->window,
                                        win_width, win_height, DpyDepth);
        XFillRectangle(Dpy, w->window, BackGC, 0, 0, win_width, win_height);
    } else {
#if AXMS
        AxWinCreateTrayWindow(w, win_xpos, win_ypos, win_width, win_height);
#else
        w->window = AxCrtWindow(Dpy, WIN_OF(w), win_xpos, win_ypos,
                                win_width, win_height, 0,
                                CopyFromParent, InputOutput,
                                CopyFromParent, mask, &atts);
        XSelectInput(Dpy, w->window, TRAY_NONFOCUS_MASK);
        XMapWindow(Dpy, w->window);
#endif
        AddToWindowList(w);
    }

    SetTrayWorkColors(w);

/* clear gap between border and window (should be AFTER GCs changed
   for work area colors) */
    XFillRectangle(Dpy, w->window, BackGC, 0, 0, win_width, win_height);
    UnsetTrayWorkColors(w);

    CreateTrayScrollBars(w, layered);

/* Adjust top_y and leftY if necessary -- need to do this if tray gets
   resized. */
    if (trayP->top_y) {
        if (trayP->win_height >= trayP->top_y)
            trayP->top_y = 0;
    }
    if (trayP->left_x) {
        if (trayP->win_width >= trayP->left_x)
            trayP->left_x = 0;
    }

    if (w->flags & WFLG_IN_EDIT) {
        XCopyArea(Dpy, w->window, w->parent->window, ForeGC, 0, 0,
                  win_width, win_height, win_xpos, win_ypos);
    }
    AXDefineDefaultCursor(w, ThimPointerCursor);
    return;
}

/***********************************************************************
function: SetTrayWorkColors - Set colors on tray GCs.
*/
static void
SetTrayWorkColors(w)
register AXWidget_tray w;
{
    if (!w)
        return;
    if (w->type != W_TRAY)
        return;
    if (w->useWorkColors) {
        SetWorkColors(BackGC, 0);
        SetWorkColors(ForeGC, 1);
        SetWorkColors(DisabledGC, 1);
    }
    return;
}

/***********************************************************************
function: UnsetTrayWorkColors - Unset work colors on tray GCs.
*/
static void
UnsetTrayWorkColors(w)
register AXWidget_tray w;
{
    if (!w)
        return;
    if (w->type != W_TRAY)
        return;
    if (w->useWorkColors) {
        UnSetWorkColors(BackGC, 0);
        UnSetWorkColors(ForeGC, 1);
        UnSetWorkColors(DisabledGC, 1);
    }
    return;
}

/***********************************************************************
function: TrayXCopyArea - Wrapper for XCopyArea () for tray widget.
*/
static void
TrayXCopyArea(dpy, src_drawable, dst_drawable, gc, src_x, src_y, width, height,
              dst_x, dst_y)
register Display *dpy;
Drawable src_drawable, dst_drawable;
GC  gc;
int src_x, src_y;
unsigned int width, height;
int  dst_x, dst_y;
{
    XGCValues    orig_values;
    unsigned int valuemask;
    int changed = FALSE;

    valuemask = GCGraphicsExposures;
    XGetGCValues(dpy, gc, valuemask, &orig_values);
    if (orig_values.graphics_exposures == False) {
        XSetGraphicsExposures(dpy, gc, True);
        changed = TRUE;
    }
    XCopyArea(dpy, src_drawable, dst_drawable, gc, src_x, src_y,
              width, height, dst_x, dst_y);
    if (changed)
        XSetGraphicsExposures(dpy, gc, orig_values.graphics_exposures);
    return;
}

/***********************************************************************
function: SetClipAllGCs - Set clip on all tray related GCs.
*/
static void
SetClipAllGCs(w, clip_x_origin, clip_y_origin, rectangles, n, ordering)
register AXWidget_tray w;
int clip_x_origin;
int clip_y_origin;
XRectangle *rectangles;
int n;
int ordering;
{
    tray_data *trayP;

    if (!w)
        return;
    if (w->type != W_TRAY)
        return;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    trayP->state.clips_are_on = TRUE;
    return;
}

/***********************************************************************
function: UnsetClipClipAllGCs - Unset clip on all tray related GCs.
*/
static void
UnsetClipAllGCs(w)
register AXWidget_tray w;
{
    tray_data *trayP;
    if (!w)
        return;
    if (w->type != W_TRAY)
        return;
    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    trayP->state.clips_are_on = FALSE;
    return;
}

/***********************************************************************
function: CreateTrayScrollBars - Create tray scrollbars.
*/
static void
CreateTrayScrollBars(w, layered)
AXWidget_tray w;
int layered;
{
    tray_data *trayP;
    int vscroll_origin = 0;
    int vscroll_length = 0;
    int hscroll_origin = 0;
    int hscroll_length = 0;

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;
    if (trayP) {
        vscroll_origin = trayP->vScrollOrigin;
        vscroll_length = trayP->vScrollLength;
        hscroll_origin = trayP->hScrollOrigin;
        hscroll_length = trayP->hScrollLength;
    }

    /* NOTE: scrollbar parent MUST be dialog to ensure proper
       behaviour with task of listbox is pended */
    if (!w->hide_vscroll) {
        Arg args[10];
        int n = 0;
        int offset = 2;
        static XtCallbackRec callback_list[2];

        if (w->thickness == 0)
            offset = 0;
        CALLBACKER(callback_list, TrayBarChanged, w);
        XtSetArg(args[n], XmNx, w->xpos + w->width + offset); n++;
        XtSetArg(args[n], XmNy, w->ypos + vscroll_origin); n++;
        if (vscroll_length) {
            XtSetArg(args[n], XmNheight, vscroll_length); n++;
        } else {
            XtSetArg(args[n], XmNheight, w->height); n++;
        }
        XtSetArg(args[n], XmNorientation, XmVERTICAL); n++;
        XtSetArg(args[n], XmNvalueChangedCallback, callback_list); n++;
        if (w->vScrollBar)
            AxXtSetValues(w->vScrollBar, args, n);
        else
            w->vScrollBar = (AXWidget_scroll) AXmCreateScrollBar(w->parent,
                                                            "x", args, n);
        if (w->event_table) {
            /* give the newly create scrollbar an event table which
               matches that of the list widget. */
            copyEventHandlerList(w->event_table, &w->vScrollBar->event_table);
        }

        /* dialog editor support */
        XtSetArg(args[0], ZXmNcompositeLinkUp, w);
        XtSetArg(args[1], ZXmNinEdit, w->flags & WFLG_IN_EDIT);
        AxXtSetValues(w->vScrollBar, args, 2);
        XtSetArg(args[0], ZXmNcompositeLinkDown, w->vScrollBar);
        AxXtSetValues(w, args, 1);

        xmSetWidgetDisplay(w->vScrollBar, TRUE);
        w->vScrollBar->scroll_per = 100;
        w->vScrollBar->scroll_max = 100;
        w->vScrollBar->scroll_pos = 0;
        PaintScrollBar(w->vScrollBar, NULL, FALSE, layered);
    }

    if (w->list_hscroll) {
        Arg args[10];
        int n = 0;
        static XtCallbackRec hcallback_list[2];
        int offset = 2;

        if (w->thickness == 0)
            offset = 0;
        CALLBACKER(hcallback_list, TrayBarChanged, w);
        XtSetArg(args[n], XmNx, w->xpos + hscroll_origin); n++;
        XtSetArg(args[n], XmNy, w->ypos + w->height + offset); n++;
        if (hscroll_length) {
            XtSetArg(args[n], XmNwidth, hscroll_length); n++;
        } else {
            XtSetArg(args[n], XmNwidth, w->width); n++;
        }
        XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
        XtSetArg(args[n], XmNvalueChangedCallback, hcallback_list); n++;
        if (w->hScrollBar)
            AxXtSetValues(w->hScrollBar, args, n);
        else
            w->hScrollBar = (AXWidget_scroll)AXmCreateScrollBar(w->parent,
                                                            "y", args, n);

        if (w->event_table) {
            /* give the newly created scrollbar an event table which
               matches that of the list widget. */
            copyEventHandlerList(w->event_table, &w->hScrollBar->event_table);
        }

        /* dialog editor support */
        XtSetArg(args[0], ZXmNcompositeLinkUp, w);
        XtSetArg(args[1], ZXmNinEdit, w->flags & WFLG_IN_EDIT);
        AxXtSetValues(w->hScrollBar, args, 2);
        XtSetArg(args[0], ZXmNcompositeLinkDown, w->hScrollBar);
        AxXtSetValues(w, args, 1);

        xmSetWidgetDisplay(w->hScrollBar, TRUE);
        w->hScrollBar->scroll_per = 100;
        w->hScrollBar->scroll_max = 100;
        w->hScrollBar->scroll_pos = 0;
        PaintScrollBar(w->hScrollBar, NULL, FALSE, layered);
    }

    /* Propogate `buttonUpScroll' value here.  This is done here
       because we need valid handles to the scrollbar widgets of this
       tray.  This attribute might not have been set at the
       AxXtSetValues() layer because the tray might not have scrollbars
       at that point. */
    if (w->hScrollBar)
        w->hScrollBar->buttonUpScroll = w->buttonUpScroll;
    if (w->vScrollBar)
        w->vScrollBar->buttonUpScroll = w->buttonUpScroll;

    if (trayP) {
        trayP->state.vscroll_dim_changed = FALSE;
        trayP->state.hscroll_dim_changed = FALSE;
    }

    return;
}

/***********************************************************************
function: FormatTray - Format all nodes in a tray.
Return TRUE if there was a change; FALSE otherwise (or on error).
*/
static int
FormatTray(w)
AXWidget_tray w;
{
    tray_data *trayP = (tray_data *) NULL;
    int numTopNodes;
    treeNode *thisTopNode;
    treeNode *thisNode;
    int ix;
    int x, y;
    unsigned int width, height;

    if (!w)
        return(FALSE);
    if (w->type != W_TRAY)
        return(FALSE);

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(FALSE);

    numTopNodes = IntsArraySize(trayP->top_node_list);
    if (numTopNodes <= 0)
        return(FALSE);

    for (ix = 0; ix < numTopNodes; ix++) {
        thisTopNode = (treeNode *) ReadIntsArray(trayP->top_node_list, ix);
        if (!thisTopNode)
            continue;
        ResetTreeDispInfo(thisTopNode);
    }

    x = NODE_SPACING_DIM;
    y = NODE_SPACING_DIM;
    
    for (ix = 0; ix < numTopNodes; ix++) {
        thisTopNode = (treeNode *) ReadIntsArray(trayP->top_node_list, ix);
        if (!thisTopNode)
            continue;
        FormatNode(trayP, thisTopNode, x, y);
    }
}

/***********************************************************************
function: FormatNode - Format a top node.
*/
static void
FormatNode(trayP, topP, x, y)
tray_data *trayP;
treeNode *topP;
int x;
int y;
{
    int max_x = 0;
    int max_y = 0;

    if (!topP || !trayP)
        return;
    if (topP->level != 0)
        return;
    SetNodeDim(trayP, topP);
    IFormatNode(trayP, topP, x, y, &max_x, &max_y);
    topP->total_width  = max_x - x;
    topP->total_height = max_y - y;
    return;
}

/***********************************************************************
function: IFormatNode - Recursively format a node (and it's subnodes...).
This function must only be called by `FormatNode ()'.
*/
static void
IFormatNode(trayP, nodeP, x, y, max_xp, max_yp)
tray_data *trayP;
treeNode *nodeP;
int x;
int y;
int *max_xp;
int *max_yp;
{
    int ix = 0;
    int num_subnodes = 0;
    treeNode *thisP = (treeNode *) NULL;

    if (!nodeP || !trayP)
        return;
    nodeP->info.xpos = x;
    nodeP->info.ypos = y;
    num_subnodes = IntsArraySize(nodeP->sub_nodes);
    if (num_subnodes) {
        /* Set `*max_xp' before changing `x'. */
        if ((x + nodeP->info.width) > *max_xp)
            *max_xp = x + nodeP->info.width;

        x += (int) (nodeP->info.pixmap_width + BTN_TXT_SPACING);
        y += nodeP->info.height;
        if (y > *max_yp)
            *max_yp = y;
    }
    for (ix = 0; ix < num_subnodes; ix++) {
        thisP = (treeNode *) ReadIntsArray(nodeP->sub_nodes, ix);
        if (!thisP)
            continue;
        SetNodeDim(trayP, thisP);
        if (IntsArraySize(thisP->sub_nodes) > 0) {
            IFormatNode(trayP, thisP, x, y, max_xp, max_yp);
            thisP->total_width  = *max_xp - x;
            thisP->total_height = *max_yp - y;
            y = *max_yp;
        } else {
            thisP->info.xpos = x;
            thisP->info.ypos = y;
            y += thisP->info.height;
            if ((x + thisP->info.width) > *max_xp)
                *max_xp = x + thisP->info.width;
            if (y > *max_yp)
                *max_yp = y;
        }
    }
}

/***********************************************************************
function: SetNodeDim - Set `nodeP->info.width' and `nodeP->info.height'
of node by looking at pixmap dimension.
*/
static void
SetNodeDim(trayP, nodeP)
tray_data *trayP;
treeNode *nodeP;
{
    pixcache_t *pixent = (pixcache_t *) NULL;
    int width  = 16; /* default width */
    int height = 16; /* default height */
    int title_height = 0;
    int title_width  = 0;

    if (!nodeP || !trayP)
        return;

    if (nodeP->info.icon_pixmap && strlen(nodeP->info.icon_pixmap))
        pixent = Pixmapper(nodeP->info.icon_pixmap);

    if (pixent) {
        if ((int) pixent->width > width)
            width = pixent->width;
        if ((int) pixent->height > height)
            height = pixent->height;

        nodeP->info.pixmap_width  = width  + (BTN_MARGIN_DIM * 2);
        nodeP->info.pixmap_height = height + (BTN_MARGIN_DIM * 2);
    } else {
        nodeP->info.pixmap_width  = width  + (BTN_MARGIN_DIM * 2);
        nodeP->info.pixmap_height = height + (BTN_MARGIN_DIM * 2);
    }

    if (nodeP->info.title && strlen(nodeP->info.title)) {
        title_width  = AXTextWidth(trayP->font.font_desc,
                                   nodeP->info.title,
                                   strlen(nodeP->info.title));
        title_height = trayP->font.char_height;
        nodeP->info.title_width  = title_width;
        nodeP->info.title_height = title_height;
    }

    height = (MAX(nodeP->info.pixmap_height, title_height) +
              NODE_HEIGHT_EXTENSION);
    width  = nodeP->info.pixmap_width + title_width + BTN_TXT_SPACING;

    /* Extend the right edge of the node by NODE_WIDTH_EXTENSION so it looks
       neat when the node is marked. */
    nodeP->info.width  = width + NODE_WIDTH_EXTENSION;
    nodeP->info.height = height;
    return;
}

/***********************************************************************
function: FreeTrayData - Free all (tray_data).
*/
void
FreeTrayData(trayP)
tray_data *trayP;
{
    int numTopNodes;
    int ix;

    if (!trayP)
        return;
    numTopNodes = IntsArraySize(trayP->top_node_list);
    if (numTopNodes) {
        for (ix = 0; ix < numTopNodes; ix++)
            FastDestroyTree((treeNode *)ReadIntsArray(trayP->top_node_list,
                                                      ix));
    }
    trayP->top_node_list = NULL;
    FreeTrayGCs(trayP);
    if (trayP->font.family)
        free(trayP->font.family);
    return;
}

/***********************************************************************
function: UnselectTrayNode - Clear selection on node NODEP.
If NODEP is NULL, clear all selections in tray.
Return TRUE if selections changed; FALSE otherwise.
*/
static int
UnselectTrayNode(w, nodeP)
AXWidget_tray w;
treeNode *nodeP;
{
    tray_data *trayP;
    int num_selections;
    long *new_selected_nodes;
    int ix;

    if (!w)
        return(FALSE);
    if (w->type != W_TRAY)
        return(FALSE);

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(FALSE);

    if (trayP->selected_nodes) {
		num_selections = IntsArraySize(trayP->selected_nodes);
	}
	else {
		num_selections = 0;
	}
	
    if (!num_selections)
        return(FALSE);

    if (!nodeP) {
        treeNode *thisP;

        for (ix = 0; ix < num_selections; ix++) {
            thisP = (treeNode *) ReadIntsArray(trayP->selected_nodes, ix);
            if (!thisP)
                continue;
            thisP->attr.selected = FALSE;
        }
        free(trayP->selected_nodes);
        trayP->selected_nodes = CreateIntsArray(0);
    } else {
        if (!InIntsArray(trayP->selected_nodes, nodeP))
            return(FALSE);

        nodeP->attr.selected = FALSE;
        new_selected_nodes = DeleteIntsArray(trayP->selected_nodes,
                                             NODE_ID(nodeP));
        free(trayP->selected_nodes);
        trayP->selected_nodes = new_selected_nodes;
    }
    return(TRUE);
}

/***********************************************************************
function: SelectTrayNode - Add NODEP to the list of selected nodes.
Return TRUE if selected nodes list changed; FALSE otherwise.
*/
static int
SelectTrayNode(w, nodeP)
AXWidget_tray w;
treeNode *nodeP;
{
    tray_data *trayP;
    long *new_selected_nodes;
    int ix;

    if (!w)
        return(FALSE);
    if (w->type != W_TRAY)
        return(FALSE);

    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return(FALSE);
	
	if (!trayP->selected_nodes) {
		trayP->selected_nodes = CreateIntsArray(0);
	}

    if (InIntsArray(trayP->selected_nodes, NODE_ID(nodeP)))
        return(FALSE); /* already there */

    new_selected_nodes = AddIntsArray(trayP->selected_nodes, NODE_ID(nodeP));
    if (trayP->selected_nodes)
        free(trayP->selected_nodes);
    trayP->selected_nodes = new_selected_nodes;
    nodeP->attr.selected = TRUE;
    return(TRUE);
}

/***********************************************************************
function: PrintWindowAttributes - For debugging...
*/
PrintWindowAttributes(win)
Window win;
{
    XWindowAttributes attr;
    int status;

#if AX_X11
    status = XGetWindowAttributes(Dpy, win, &attr);
    if (!status) {
        printf("XGetWindowAttributes for 0x%lx FAILED\n", win);
        return -1;
    }
    printf("    backing_store = ");

    switch (attr.backing_store) {
    case NotUseful:
        printf("NotUseful\n");
        break;
    case WhenMapped:
        printf("WhenMapped\n");
        break;
    case Always:
        printf("Always\n");
        break;
    default:
        printf("UNKNOWN\n");
        break;
    }

    printf("    backing_planes = %ld\n", attr.backing_planes);
    printf("    backing_pixel  = %ld\n", attr.backing_pixel);
    return 0;
#endif
}

/***********************************************************************
                   Sending up events from the tray.
 ***********************************************************************/

/***********************************************************************
function: SendSelectionChangeEvent - Send back new selections in tray.
*/
static void
SendSelectionChangeEvent(w)
register AXWidget_tray w;
{
    TrayEventData data;
    tray_data *trayP;
    int num_selections;
    int ix;

    if (!w)
        return;
    if (w->type != W_TRAY)
        return;
    if (!w->callback)
        return;
    
    trayP = (tray_data *) w->info.tray;
    if (!trayP)
        return;

    data.any.type = EV_TRAY_SELECTION_CHANGE_;
    if (trayP->selected_nodes && IntsArraySize(trayP->selected_nodes))
        data.selection_change.node_info_array =
            CopyIntsArray(trayP->selected_nodes);
    else
        data.selection_change.node_info_array = NULL;
    AxCallback(w->callback, w, w->callbacktag, &data);
    return;
}

/***********************************************************************
function: SendButtonPressEvent - Report button press in one of the
tray items
*/
static void
SendButtonPressEvent(w, nodeP)
register AXWidget_tray w;
register treeNode *nodeP;
{
    TrayEventData data;

    if (!w)
        return;
    if (!nodeP)
        return;
    if (!w->callback)
        return;

    data.any.type = EV_TRAY_BUTTONPRESS_;

    data.button_press.node_info.title  = nodeP->info.title;
    data.button_press.node_info.pixmap = nodeP->info.icon_pixmap;
    data.button_press.node_info.value  = nodeP->value;
    data.button_press.node_info.handle = NODE_ID(nodeP);
    AxCallback(w->callback, w, w->callbacktag, &data);
    return;
}

/***********************************************************************
function: SendDoubleClickEvent - Report a double-click on one of the
tray items
*/
static void
SendDoubleClickEvent(w, nodeP)
register AXWidget_tray w;
treeNode *nodeP;
{
    TrayEventData data;

    if (!w)
        return;
    if (!nodeP)
        return;
    if (!w->callback)
        return;

    data.any.type = EV_TRAY_DBLCLICK_;

    data.double_click.node_info.title  = nodeP->info.title;
    data.double_click.node_info.pixmap = nodeP->info.icon_pixmap;
    data.double_click.node_info.value  = nodeP->value;
    data.double_click.node_info.handle = NODE_ID(nodeP);
    AxCallback(w->callback, w, w->callbacktag, &data);
    return;
}

/***********************************************************************
                         Elf Layer Interface
 ***********************************************************************/

/***********************************************************************
function: ElfbTrayFuncs - Generic tray builtin functions entry point.
WVAL     - widget handle
CMDCODEP - tray builtin command code. (currently hardcoded)
DP1..DP5 - interpreted data depending on command code.
*/
void
ElfbTrayFuncs(wval, cmdCodeP, dP1, dP2, dP3, dP4, dP5)
data_ptr wval;
data_ptr cmdCodeP;
data_ptr dP1, dP2, dP3, dP4, dP5;
{
    AXWidget_tray w;
    tray_data *trayP = (tray_data *) NULL;
    data_ptr retP = (data_ptr) NULL;
    int command_code;

    w = (AXWidget_tray)WidgetPointerFromHandleData(wval);
    trayP = (tray_data *) w->info.tray;
    if (!w)
        goto finish;
    if (w->type != W_TRAY)
        goto finish;

    command_code = IntFromDataPtr(cmdCodeP);
    switch (command_code) {
    case 0:
        break;
    case 1: { /* TRAY#CMD_SET_DATA */
        treeNode *topNode;
        int size;
        int ix;
        int update_mode = FALSE;
        long *top_list = NULL;

        if (!trayP)
            goto finish;
        topNode = NodePtrFromBinaryData(dP1);
        size = IntsArraySize(trayP->top_node_list);
        if (!topNode) {
            /* destroy whole tree */
            for (ix = 0; ix < size; ix++) {
                treeNode *this_top;

                this_top = (treeNode *)ReadIntsArray(trayP->top_node_list, ix);
                if (!this_top)
                    continue;
                FastDestroyTree(this_top);
            }
            if (trayP->top_node_list)
                free(trayP->top_node_list);
            trayP->top_node_list = NULL;
            goto reset_and_finish;
        }
        /* FIXME, the tray can't handle multiple top nodes */
        if (size) {
            treeNode *top_node;

            /* handle multiple top nodes */
            top_node = (treeNode *) ReadIntsArray(trayP->top_node_list, 0);
            if (top_node == topNode)
                update_mode = TRUE;
        }
        if (size && !update_mode) {
            /* Destroy existing data */
            top_list = (long *) CopyIntsArray(trayP->top_node_list);
            for (ix = 0; ix < size; ix++) {
                treeNode *this_top;

                this_top = (treeNode *) ReadIntsArray(top_list, ix);
                if (!this_top)
                    continue;
                FastDestroyTree(this_top);
            }
            if (top_list)
                free(top_list);
            if (trayP->top_node_list)
                free(trayP->top_node_list);
            trayP->top_node_list = NULL;
        }
        if (!update_mode) {
            if (trayP->top_node_list)
                free(trayP->top_node_list);
            /* tray can't handle multiple top nodes */
            trayP->top_node_list = CreateIntsArray(1);
            WriteIntsArray(trayP->top_node_list, 0, NODE_ID(topNode));
        }
reset_and_finish:
        if (update_mode) {
            /* clear all selections the proper way. */
            UnselectTrayNode(w, NULL);
        }
        TrayResetAll(w);
        trayP->state.needs_formatting = TRUE;
        PaintTray(w, NULL, TRUE, FALSE);
    }
        break;
    case 2: { /* TRAY#CMD_SET_FONT */
        char *font_family;
        int  font_size;
        int  font_weight;
        int  font_slant;

        font_family = StrFromDataPtr(dP1);
        font_size   = IntFromDataPtr(dP2);
        if (IntFromDataPtr(dP3) == BTRUE)
            font_weight = TRUE; /* bold */
        else
            font_weight = FALSE; /* not bold */
        if (IntFromDataPtr(dP4) == BTRUE)
            font_slant = TRUE; /* italic */
        else
            font_slant = FALSE; /* not italic */
        (void) SetTrayFont(w, font_family, font_size, font_weight, font_slant);
    }
        break;
    case 3: { /* TRAY#CMD_SET_ATTRS */
        int show_borders = FALSE;
        int show_node_wiring = FALSE;
        int sort_nodes = FALSE;
        int skip_top_node = FALSE;
        int repaint = FALSE;

        if (!trayP)
            goto finish;

        if (dP1) {
            show_borders = (IntFromDataPtr(dP1) == BTRUE) ? TRUE : FALSE;
            if (show_borders != trayP->attr.show_borders)
                repaint = TRUE;
            trayP->attr.show_borders = show_borders;
        }

        if (dP2) {
            show_node_wiring = (IntFromDataPtr(dP2) == BTRUE) ? TRUE : FALSE;
            if (show_node_wiring != trayP->attr.show_node_wiring)
                repaint = TRUE;
            trayP->attr.show_node_wiring = show_node_wiring;
        }

        if (dP3) {
            sort_nodes = (IntFromDataPtr(dP3) == BTRUE) ? TRUE : FALSE;
            if (sort_nodes != trayP->attr.sort_nodes)
                repaint = TRUE;
            trayP->attr.sort_nodes = sort_nodes;
        }
        if (dP4) {
            if (sort_nodes)
            {
                skip_top_node = (IntFromDataPtr(dP4) == BTRUE) ? TRUE : FALSE;
                if (skip_top_node != trayP->attr.skip_top_node)
                    repaint = TRUE;
                trayP->attr.skip_top_node = skip_top_node;
            }
        }

        if (repaint)
            PaintTray(w, NULL, TRUE, FALSE);
    }
        break;
    case 4: { /* TRAY#CMD_GET_SELECTIONS */
        long *selected_nodes_array = NULL;

        if (!trayP)
            goto finish;
        if (!(trayP->selected_nodes && IntsArraySize(trayP->selected_nodes)))
            goto finish;
        selected_nodes_array = CopyIntsArray(trayP->selected_nodes);
        retP = ElfArrayOfTreeNodes(selected_nodes_array);
        if (selected_nodes_array)
            free(selected_nodes_array);
    }
        break;
    case 5: { /* TRAY#CMD_SET_SELECTIONS */
        int num_selections = 0;
        int ix;
        int repaint = FALSE;

        if (!dP1) {
            if (UnselectTrayNode(w, NULL))
                repaint = TRUE;
        } else if (!ElfbIsArray(dP1)) {
            if (UnselectTrayNode(w, NULL))
                repaint = TRUE;
        } else {
            num_selections = ElfbArraySize(dP1);
            if (UnselectTrayNode(w, NULL))
                repaint = TRUE;
            for (ix = 0; ix < num_selections; ix++) {
                treeNode *thisP;
                data_ptr node;

                node = ReadArray(dP1, ix);
                if (!node)
                    continue;
                thisP = NodePtrFromBinaryData(node);
                if (!thisP)
                    continue;
                if (SelectTrayNode(w, thisP))
                    repaint = TRUE;
            }
        }
        if (repaint)
            PaintTray(w, NULL, TRUE, FALSE);
    }
        break;
    case 6: { /* TRAY#CMD_DELETE_NODE */
        treeNode *nodeP;
        int deleting_tree = FALSE;
        int left_x, top_y;

        nodeP = NodePtrFromBinaryData(dP1);
        if (!nodeP)
            goto finish;
        if (!trayP)
            goto finish;
        if (nodeP->level == 0)
            deleting_tree = TRUE;
        UnselectTrayNode(w, nodeP); /* unselect if selected */
        DelTreeNode(nodeP);
        if (deleting_tree) {
            if (trayP->top_node_list) /* FIXME, assumes one top node */
                free(trayP->top_node_list);
            trayP->top_node_list = NULL;
        }
#if 0 /* Don't do this -- Openmail will call db_tray_refresh@() to
         correct display */
        left_x = trayP->left_x;
        top_y = trayP->top_y;
        TrayResetAll(w);
        trayP->left_x = left_x;
        trayP->top_y = top_y;
        trayP->state.needs_formatting = TRUE;
        PaintTray(w, NULL, TRUE, FALSE);
#endif
    }
        break;
    case 7: { /* TRAY#CMD_DELETE_SUBNODES */
        treeNode *nodeP;
        int deleting_tree = FALSE;
        int left_x, top_y;

        nodeP = NodePtrFromBinaryData(dP1);
        if (!nodeP)
            goto finish;
        if (!trayP)
            goto finish;
        UnselectTrayNode(w, nodeP); /* unselect if selected */
        DelTreeSubnodes(nodeP);
        left_x = trayP->left_x;
        top_y = trayP->top_y;
        TrayResetAll(w);
        trayP->left_x = left_x;
        trayP->top_y = top_y;
        trayP->state.needs_formatting = TRUE;
        PaintTray(w, NULL, TRUE, FALSE);        
        break;
    }
    case 8: { /* TRAY#CMD_REFRESH */
        int left_x, top_y;
        left_x = trayP->left_x;
        top_y = trayP->top_y;
        TrayResetAll(w);
        trayP->left_x = left_x;
        trayP->top_y = top_y;
        trayP->state.needs_formatting = TRUE;
        PaintTray(w, NULL, TRUE, FALSE);
        break;
    }
    default: /* unimpl */
        break;
    }

finish:
    ElfRetData = retP;
    return;
}


/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
