/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xmtray.h
   Applix ASC version: 12
   Date: Fri Feb  5 21:11:56 1999
*/
#ifndef _xmtray_h_
#define _xmtray_h_

#include <tree.h>

/* Tray state information -- All default values are FALSE */
typedef struct {
	unsigned int vscroll_dim_changed: 1;
	unsigned int hscroll_dim_changed: 1;
	unsigned int clips_are_on: 1;
	unsigned int realized: 1; /* if tray has been realized. */
	unsigned int needs_formatting: 1; /* TRUE if tray needs formatting */
	unsigned int btn1_down: 1;
	unsigned int spare6: 1;
	unsigned int spare7: 1;
} tray_state_t;

/* Tray user settable attributes/hints -- All default values are FALSE */
typedef struct {
	unsigned int show_borders: 1; /* show border rect around all nodes. */
	unsigned int fs_display: 1; /* hint: displaying file system files */
	unsigned int show_fullname: 1; /* if `fs_display', show full names
									  of files (not basename) */
	unsigned int no_tree: 1; /* not a tree displayer; just a flat icon
								displayer. */
	unsigned int show_node_wiring: 1; /* draw node wiring */
	unsigned int sort_nodes: 1;		/* sort tray nodes */
	unsigned int skip_top_node: 1;	/* skip top node when sorting */
	unsigned int spare7: 1;
} tray_attr_t;

/* Tray font info */
typedef struct {
	XFontStruct *font_desc;
	char *family;
	int size;
	int weight;
	int slant;
	int char_above;
	int char_below;
	int char_height;
} tray_font_t;

/* Tray data (Private to xmtray.c and xmlib.c::XtSetValues()) */
typedef struct tray_data {
	int win_width;
	int win_height;
	int hScrollOrigin;
	int vScrollOrigin;
	int hScrollLength;
	int vScrollLength;
	int left_x;
	int top_y;
	int max_width;
	int max_height;
	int incX; /* for unit scroll */
	int incY; /* for unit scroll */
	/* IntsArray of top nodes.  The internal nodes themselves hold
	   display information like x, y, width, height of each node. */
	long *top_node_list;
	treeNode *last_node_selected;
	long *selected_nodes; /* IntsArray of selected node ids */
	GC text_gc;
	GC motion_gc;
	tray_font_t  font;  /* display font info */
	tray_state_t state; /* internal state */
	tray_attr_t  attr; /* user settable attributes */
} tray_data;

Widget AXmCreateTray AXP_((widget_t *parentWidget, char *name,
						   Arg args[], int cnt));
int  TrayEvent AXP_((widget_t *w, XEvent *evp));
static void PaintTray AXP_((widget_t *w, XExposeEvent *exP, int regen, int layered));
void DrawTrayBorder AXP_((widget_t *w));

#endif


/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
