/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: atmwin.h
   Applix ASC version: 15
   Date: Mon Nov 16 13:52:52 1998
*/
#include <atmicon.h>	/* flags to enable window border controls */

/*
 * max number of windows for any ATM (max window ID + 1)
 * - some system-dependent versions may allow fewer
 */
#define TM_MAXWIN	21

/* window create packet */

struct TmCrePacket
{
	struct
	{
		unsigned int rel_pos : 2,
				sub_window	 : 1,
				cur_window	 : 1,
				gr_window	 : 1, /* graphics mode window (RGMODE ATM) */
				promptable	 : 1, /* form of window depends on prompt mode */
				set_aside	 : 1, /* set window aside */
				save_image	 : 1, /* save on resize; else clear window */
				prompt		 : 1, /* prompt window (ignored if 'promptable') */
				borderless	 : 1, /* no space for borders, special text lines */
				mouse_xchg	 : 1, /* app wants to own mouse/keyboard */
				near_mouse   : 1, /* overrides window placement values */
				regen_rect   : 1, /* able to regen arbitrary rectangle */
				borrow_shell : 1, /* use parent's win instead of create new */
				ruler        : 1, /* for DC only - create ruler at crt time */
				plates       : 1, /* for SS only - create 3-d at crt time */
				windowless	 : 1, /* should be zero */
				zoomButton   : 1, /* zoom button at bottom-right */
				px_mode      : 1, /* true if ge is in bitmap mode */
				geruler      : 1, /* to create ge ruler */
				gecolorpal   : 1, /* to create ge color palette */
				gepalette    : 1, /* to create ge tool palette */
				pagebar      : 1, /* to create pagebar widget */
				presentation : 1, /* true if ge is in presentation mode */
				no_outliner  : 1; /* true if no outliner in presentation */
	} modifiers;
	short x_position;
	short y_position;
	short x_size;
	short y_size;
	short x_min_size;
	short y_min_size;
	short x_max_size;				/* maximum size (default -1) */
	short y_max_size;
	short sa_x_position;			/* set-aside position (default -1) */
	short sa_y_position;
	long *help_id;
	long border_control;			/* border control enable flags */
	short icon_id;					/* icon identifer (0: use app ID) */
	short menubar_id;               /* menubar id number */
	short zoom_state;				/* zoom / 300 gives zoom ratio */
	short grid_size;				/* number of grid per 1000 mills */
	int h_sc_pos;
	int h_sc_per;
	int v_sc_pos;
	int v_sc_per;
	short parent_id;
	char *title_line;	/* name to appear on title bar */
	char *icon_name;	/* name of window in set-aside form */
	short appid;		/* application ID (e.g. APGE, APDC...) */
	long docId;         /* 0 or doc owned by window */
	int page_ix; 
	int n_pages;
};

struct TmGridRulerInfo
{
	int x_resolution;    /* x zoom (100 means 100%, or unzoomed) */
	int y_resolution;    /* y zoom */
	int x_zero_point;    /* grid x origin */
	int y_zero_point;    /* grid y origin */
	int x_scroll_origin; /* origin of scrollable area */
	int y_scroll_origin;
	int grid_dpi;        /* typically 1000 */
	int grid_size;       /* size of grid cell in grid_dpi units */
	int unit_size;       /* size of ruler unit in grid_dpi units */
	int unit_prec;       /* precision of the units               */
	char *unit_name;     /* "in.", "cm.", ... */
};

/* flag definitions */

/* modifiers.rel_pos: positioning option */

/* do not change window position (TmModify) */
#define TM_NO_CHANGE	0

#define TM_DESKTOP_REL	1	/* origin is position of flush topleft window */
#define TM_BEST_FIT		1	/* implemented same as desktop rel */
#define TM_WINDOW_REL	2	/* not implemented */
#define TM_SCREEN_REL	3	/* origin is screen topleft (for future use) */

/* modifiers.sub_window: subordinate window */

#define TM_TOP_WINDOW	0	/* (or FALSE) */
#define TM_SUB_WINDOW	1	/* (or TRUE) */

/* modifiers.cur_window: candidate for active window */

#define TM_NO_CANDIDATE	0	/* (or FALSE) */
#define TM_BE_CANDIDATE	1	/* (or TRUE) */

/* window status structure returned by TmNextWin */

#define TMW_TIMAX	STRPATH		/* max title length returned */

/* stat values */

#define TMW_OPEN		0
#define TMW_ACTIVE		1
#define TMW_SETASIDE	2

struct TmWinStat
{
	short winID;		/* window ID */
	short appID;		/* application ID of owner */
	short pid;			/* process ID of owner */
	short stat;			/* active, open, set aside */
	char title[TMW_TIMAX];		/* title */
};

/* calls */

typeErr TmCreate();
typeErr TmTerm();
typeErr TmModify();
typeErr TmSasideAll();
typeErr TmCurWin();
typeErr TmNextWin();
typeErr TmQmIds();
typeErr TmQmSelect();
typeErr TmInqWin();

/* SYNOPSIS */
#if 0    /*  {  SYNOPSIS  */
/* ************************************************
* 
* typeErr TmCreate(window_id,cre_ptr)
* short *window_id;
* struct TmCrePacket *cre_ptr;
* 
*  return status
* 
*  TMR_OK
*  TMR_TOOM
*  TMR_POSE
*  TMR_SIZE
*  TMR_TITLE
*  TMR_OPE
*  TMR_SUB
*  TMR_DDEV
*  TMR_RERR
* 
* *************************************************
* 
* typeErr TmTerm(window_id,suspend_ptr) 
* short window_id;
* struct TmCrePacket *suspend_ptr;
* 
*  return status
* 
*  TMR_OK
*  TMR_OWN
* 
* *************************************************
* 
* typeErr TmModify(window_id,modify_ptr, ruler_ptr)
* short window_id;
* struct TmCrePacket *modify_ptr;
* strcut TmGridRulerInfo *ruler_ptr;
* 
*  return status
* 
*  TMR_OK
*  TMR_POSE
*  TMR_SIZE
*  TMR_TITLE
*  TMR_OPE
*  TMR_SUB
* 
* *************************************************
* 
* typeErr TmSasideAll()
* 
*  return status
* 
*  TMR_OK
*  TMR_OWN
* 
* *************************************************
* 
* typeErr TmCurWin(window_id)
* short *window_id;
* 
*  return status
* 
*  TMR_OK
*  TMR_NOWIN
* 
* *************************************************
* 
* typeErr TmMkCurWin(window_id)
* short window_id;
* 
*  return status
* 
*  TMR_OK
*  TMR_NOWIN
* 
* ************************************************
* 
* typeErr TmNextWin(winID, ptr)
* short winID;		* initially TM_NULWIN *
* struct TmWinStat *ptr;
* 
* return status
* 
* TMR_OK
* TMR_NOWIN			* window does not exist *
* 
* ************************************************
* 
* * inquire for window status in the form of a create packet *
* 
* typeErr TmInqWin(window_id, crePack, selBuf, titleBuf, helpBuf) 
* 	short window_id;
* 	struct TmCrePacket *crePack;	* space for create packet *
* 	char *selBuf;			* space for selection line text or NULL *
* 	char *titleBuf;			* space for title line text or NULL *
* 	long *helpBuf;			* space for array of help IDs or NULL *
* 
*  return status
* 
*  TMR_OK
*  TMR_OWN
*  TMR_NOWIN
* 
* *************************************************
*/
#endif   /*  } SYNOPSIS */
