/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axthread.h
   Applix ASC version: 4
   Date: Mon Nov 16 13:52:54 1998
*/
#if SUN4S /* {-{ */

#include <pthread.h>

#define MUTEX_LOCK pthread_mutex_t
#define INIT_MUTEX( mutex ) mutex = PTHREAD_MUTEX_INITIALIZER
#define LOCK_MUTEX( mutex ) pthread_mutex_lock( &mutex )
#define UNLOCK_MUTEX( mutex) pthread_mutex_unlock( &mutex )

#elif AXMS  /*  }-{  */

#ifndef MS_InitializeCriticalSection
#define MS_InitializeCriticalSection InitializeCriticalSection
#define MS_LeaveCriticalSection LeaveCriticalSection
#define MS_EnterCriticalSection EnterCriticalSection
#endif

#define MUTEX_LOCK CRITICAL_SECTION
#define INIT_MUTEX( mutex ) MS_InitializeCriticalSection( &mutex );
#define LOCK_MUTEX( mutex ) \
          {                                                              \
          if ( (mutex.LockCount == 0) && (mutex.RecursionCount == 0) ) \
              /* must initialize before first use                     */ \
              MS_InitializeCriticalSection( &mutex );                    \
          MS_EnterCriticalSection( &mutex );                             \
          }
#define UNLOCK_MUTEX( mutex) MS_LeaveCriticalSection( &mutex )

#else /* }-{ */

/* disable threads */
#define MUTEX_LOCK int
#define INIT_MUTEX( mutex ) 
#define LOCK_MUTEX( mutex ) 
#define UNLOCK_MUTEX( mutex) 

#endif /* }-} */
