/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: boom.h
   Applix ASC version: 9
   Date: Mon Nov 16 13:52:51 1998
*/
#ifndef OBJMGRdef
#define OBJMGRdef 1

#include <otparams.h>

/* ERROR codes */

#define  M_INVPMTR    1     /* invalid parameters passed - memInit    */
#define  M_SEQERR     2     /* mmgr not initialized               */
#define  M_INVOBJH    3     /* invalid object handle              */
#define  M_MAXOBJC    4     /* max obj count has reached              */
#define  M_MAXLOCK    5     /* max lock count has reached             */
#define  M_NO_SP_MEM  6     /* no space in memory                 */
#define  M_INVLOCK    7     /* obj not found in the lock frame        */
#define  M_INVTYPE    8     /* invalid object type specified          */
#define  M_ZEROSIZE   9     /* OBSOLETE */
#define  M_RSZ_LK_OBJ 10    /* resizing invalid for locked object     */
#define  M_INVSPID    11    /* invalid objSpaceId  specified          */
#define  M_SPCOUNT    12    /* max obj space count has reached        */
#define  M_INVREQ     13    /* performed a write activity on a READONLY space */
#define  M_STACKERR   14    /* OBSOLETE */
#define  M_OPENFILE   15    /* unable to open objmgr file */
#define  M_DSKFULL    16    /* filled disk */
#define  M_FILEFULL   17    /* filled file */
#define  M_READERR    18    /* a file read returned a bad code */
#define  M_WRITERR    19    /* a file write returned a bad code */
#define  M_FREELOCKED 20    /** attempt to free a locked object */
#define  M_BADREV     21    /* swap file from incorrect objmgr revision */
#define  M_BADMACHINE 22    /* swap file not created on this machine type */
#define  M_BADUSER    23    /* swap file not created by this user type */
#define  M_LOCKERR    24    /** unlocked object not at top of lock frame */
#define  M_DEADLOCK   25    /** recursive thrashing during swap */
#define  M_CONFIGED   26    /* Boom already objConfiged */
#define  M_NOCONFIG   27    /* Boom never objConfiged */
#define  M_NOMALLOC   28    /* Unable to perform critical malloc */
#define  M_TOOBIG     29    /* object exceeds OM_MAX_OBJSIZE */
#define  M_NOTINMEM   30    /* OBSOLETE */
#define  M_CORRUPT    31    /* aborted because of memory/file corruption */
#define  M_PARAM      32    /* invalid call or parameter to call */

typedef unsigned int Handle;
#define ID Handle

typedef short objSpId;   /* object space id's */
typedef unsigned short sizeType;    /* used by DC */
typedef unsigned short flagsType;   /* change to long if more than 16 flags */

/* MemInit structure, things with ** are not used by ZOOM */

typedef struct {
    char *filename;         /* object file name (without ".sw" extension) */
    short openMode;         /* 2 - Owner, 1 - Writer, 0 - Reader */
    long userType;          /* supply an application id */
    short newObjFile;       /* TRUE if fresh object file created */
    short minPages;         /** number of object table pages in core */
    long maxObjects;        /* in Kobjects */
    short blockSize;        /** size of each malloced block */
    short numBlocks;        /** max number of malloc blocks */
    short blkIdShift;       /** to get block id in lsbs */
    short blkOfsMask;       /** get get intra-block offset */
    short dskGranularity;   /** in bytes */
    short pageLen;          /** how many entries in an object table */
    short pageShift;        /** to get pageno on right shift of id */
    short pageMask;         /** to get offset into page when and'ed */
    long lockMask;          /** */
    long writMask;          /** */
    long diskMask;          /** */
    long secnMask;          /** */
    long flagMask;          /** */
    long obitMask;          /** */
    short addrShift;        /** */
} OmMemInit;

/* 64 Meg bytes is the maximum object size */
#define OM_MAX_OBJSIZE (64000000-sizeof(ObjHeader)+4)/* num from undexed array*/

#define WRT_OBJECT 1
#define NWRT_OBJECT 0

#define OMODE_READER    0   /* open BOOM for reading only */
#define OMODE_WRITER    1   /* obsolete, same as OWNER */
#define OMODE_OWNER     2   /* allow object writing and memSaves */

#define NULL_OBJ 0

/* MACROS for unlock with and without write */

#define  unlObj(x,y,z)
#define  objUnlock(x,y)    
#define  objWUnlock(x,y)  

#define  UVRETURN(y)  {return(y);}
#define  URETURN      {return;}
#define  WVRETURN(y)  {return(y);}
#define  WRETURN      {return;}

/* Unlock calls using pointers and no lock frames */
#define ObjPUnlock(p)       unlPObj((char *)(p),FALSE)
#define ObjPWUnlock(p)      unlPObj((char *)(p),TRUE)
#define UnlPObj(p, write)   unlPObj((char *)(p), (write))
#define ObjPSetMod(p)       objPSetMod((char *)(p))

/* GLOBAL VARIABLES */

extern short memErr;            /* contains the error code on err_return */
extern int OMlogfile;       /* obsolete - here for code compatability */

/* Brief summary of calls  (more info in boom.info.h)

objDefaults(&meminit)         Return the default objmgr OmMemInit structure
objConfig(&meminit)           Configure all-space variables
objInit(&meminit)             Initialize object manager or just a new space
objSize(id)                   Return the size of an object
getFlags(id)                  return an objects's flag word
setFlags(id,flags)            Set an object's flags word
getObjSpace(&space)           Return current spavce number
memClose(delete)              Close current object space
objShutDown()                 Get rid of temp file ...
objAlloc(size,flags)          Create an object, return its id
objCopy(id)                   Make a duplicate of an object, return new id
objCurRev(id)                 Return whether object created this session
objInMem(id)                  Return in-memory status of an object
objAddr(id)                   Return the addr of an object without locking it
objLock(id,&frame)            Lock object and return its address
setObjSpace(space)            Switch to another object space
unlObj(id,&frame,write)       Unlock passed object
objFree(id,&frame)            Eliminate passed object
objResize(id,sz)              Change the size of passed object
memSave(close)                Save object space to disk
objWrite(id)                  Write a user visible object
unlStack(mode,&frame)         Unlock all object in local lock frame
objSwap(id)                   Swap an object to disk, free the core
AllcCore(size)                Get some core, swap if needed
FreeCore(addr)                Free core gotten from NeedCore

Definitions for object calls:                     */

int objDefaults();
int objConfig();
int objInit();
int getObjSpace();
int memClose();
int objShutDown();
Handle objAlloc();
Handle objCopy();
int objCurRev();
int objInMem();
int setObjSpace();
int objFree();
int objResize();
int memSave();
int objWrite();
int objSwap();
char *AllcCore();
int FreeCore();

#endif


