/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: datetime.h
   Applix ASC version: 9
   Date: Mon Nov 16 13:52:52 1998
*/
#ifndef DTTDEF
#define DTTDEF 1

/*------------------------- Header/Footer Defines ---------------------------*/
/*---------------------------- (used by ge/ss) ------------------------------*/

#define HDR_CODE_CHAR '#'

#define	NO_HF_CODE		-1
#define	HF_DOCNAME		0
#define	HF_PATHNAME		1
#define	HF_PAGENUM		2
#define	HF_SEQ_PAGENUM	3
#define	HF_PAGECOUNT	4
#define	HF_VIEWNAME		5
#define	HF_SHEETNAME	6
#define	HF_LASTMOD		7
#define	HF_DATE			8
#define	HF_TIME			9

/*-------------------------- Date and Time Format ----------------------------

-------> A memo summarizing Alis Rev2.0 changes to Date and Time has been
appended to the end of this file.  -Paul Dale 4/22/87  <------

On 12/14/87 DTSPAN was changed from 236 to 250.  This means that dates 
can now go thru Dec 31, 2100 insead of only 2086.  This is because Lotus
supports dates thru Dec 31, 2099.  It also means that about the last 14
years of dates are negative long ints unless the date values are correctly
declared as unsigned long as they should be!  -Paul

Applix has its own internal representation format for dates and times as 
defined here.  There are two parameters which can be varied for a given
customer if absolutely necessary; the granularity and the base year.
The minimum granularity is 1 second.
Currently the granularity is set to "quad-seconds",
i.e. the nearest even multiple of 4.  The base year is currently set to
1850.  This gives a span of over 472 years ranging from Jan 1, 1850 00:00:00 to
sometime after year 2322.  Today's dates are still positive numbers. 
*/
#define DTTIMEGRANULARITY	4	/* must divide evenly into 60. */
#define DTTIMEBASEYEAR		1850
#define DTSPAN			350  
#define DTNULL			((unsigned long)1382400)
/*
A special value, DTNULL,  is reserved to mean "null", "void", "not specified".
(This value represents day 0 of Jan, 1850 00:00, a non-existant time! )

If an application wishes to keep just a time value, it should use the standard
format with the year, month and day values set to zero.  

If an application wishes to keep just a date value, it should use the standard
format with the hour, minute, and bi-second fields set to zero, i.e. record 
the date as the "date-time" of the first second of the given day.  Thus it is
possible to combine a date only value and a time only value simply by adding
the two numbers together.

A more complex algorithm is needed to add a duration to a date-time value and
produce a new date-time value.  (We will probably not need to implement this
algorithm until we do a project management system.)

The ParseDate function is provided which parses an external date string, time
string, or date and time string and produces the internal representation.
Similarly, the Formatdt function has been provided to convert an internal 
date-time value into one of several standard external representations.

A date-time value is computed straightforwardly as follows:

f = 60/DTTIMEGRANULARITY	(assumes DTTIMEGRANULARITY is selected such 
				 that there is no remainder)

date_time = (sec/granularity)  +  	[Note: range(sec) is [0-59] ]
	    (min    * f)       +	[Note: range(min) is [0-59] ]
	    (hour   * f * 60)  +	[Note: range(hours) is [0-23] ]
	    (day    * f * 60 * 24)  +	[Note: range(days) is [1-31] ]
	    (month  * f * 60 * 24 * 32) + [Note: range(months) is [1-12] ]
    ((year-DTTIMEBASEYEAR) * f * 60 * 24 * 32 * 13) 
				[Note: range(years) is [DTTIMEBASEYEAR-????]]

 -------------------------- Formatdt() Options ----------------------------
The options for Formatdt() are described below.  The date,
time, and day-of-week formats can be specified separately and then 
combined (by adding) to form a single formatopt as required by Formatdt:

Date Formats:
 */
#define DTDFNONE	((int)0)	/*no date text, used to get time only*/
#define	DTDFLETTER	((int)1)	/*  e.g.  February 14, 1984  */
#define DTDFABBREV	((int)2)	/*  e.g.  Feb 14, 1984  */
#define DTDFMIL		((int)3)	/*  e.g.  14 Feb 84   */
#define DTDFSLASH	((int)4)	/*  e.g.  2/14/84        */
#define DTDFDOT		((int)5)	/*  e.g.  14.2.84  (used in Europe)*/
#define DTDFY4DASH	((int)6)	/* yyyy-mm-dd e.g. 1985-02-14 */
#define DTDFY2DASH	((int)7)	/*   yy-mm-dd e.g.   85-02-14 */
#define DTDFY4SPACE	((int)8)	/* yyyy mm dd e.g. 1985 02 14 */
#define DTDFY2SPACE	((int)9)	/*   yy mm dd e.g.   85 02 14 */
#define DTDFY4  	((int)10)	/* yyyymmdd   e.g. 19850214 */
#define DTDFY2  	((int)11)	/*   yymmdd   e.g.   850214 */
#define DTDFEURSLASH  	((int)12)	/*  dd/mm/yy  e.g. 14/02/85 */
#define DTDFDOTY4	((int)13)	/*  e.g.  14.2.1984  (used in Europe)*/

#define DTDFMAILABRV	((int)14)	/* Feb 09, 1984 */

#define DTDFLETYR       ((int)15)	/* February 1984 */
#define DTDFABBYR       ((int)16)	/* Feb 1984 */
#define DTDFABB2YR      ((int)17)	/* Feb 84 */
#define DTDFSLYR        ((int)18)	/*  2/84 (1 leadng space if mo <10) */
#define DTDFSP2YR       ((int)19)	/* 84 02 */
#define DTDFSP4YR       ((int)20)	/* 1984 02 */

/* Time formats */
#define DTTFNONE	((int)0)	/*no time text, used to get just date*/
#define DTTFBASE	((int)100)
#define DTTF12HR	((int)100)	/*  e.g.  9:34 PM        */
#define DTTF12HB	((int)200)	/*  e.g.  9:34PM         */
#define DTTF24HR	((int)300)	/*  e.g.  21:34          */
#define DTTFDOT 	((int)400)	/*  e.g.  09.34          */
#define DTTF4DIGIT 	((int)500)	/*  e.g.  0934          */
#define DTTF24HRSS 	((int)600)	/*  e.g.  21:34:00 		*/

/* Day of the Week formats */
#define DTWFNONE	((int)0)	/* Don't show day of week.  */
#define DTWFBASE	((int)1000)	
#define DTWFBEG		((int)1000)	/*  e.g.  Tue <date and/or time> */
#define DTWFEND		((int)2000)	/*  e.g.  <date and/or time> Tue */
#define DTWFFBEG	((int)3000)	/*  e.g.  Tuesday <date and/or time> */
#define DTWFFEND	((int)4000)	/*  e.g.  <date and/or time> Tuesday*/

/* For example, DTDFSLASH+DTTFNONE+DTWFBEG might produce the string,
		"Tue 2/14/84"
*/

/*------------------ ParseDate() Options ---------------------------------*/
#define PDNORMAL 1	/*- requests normal operation; the string can 
				contain either a date, time or both. If a time is 
				present, the AM/PM rules stated below for PDTIMEAMPM
				apply here also.*/
#define PDNORMEUROPE 65 /* Just like PDNORMAL except that ##/##/## and
				##-##-## are interpreted as dd/mm/yy instead of the 
				American mm/dd/yy interpretation. */
#define PDDATE   2	/* only accepts a date.  If a date and time or 
				just a time is presented, an error is reported
				One subtle point: Under this option, if the 
				string is a single number, it is interpreted 
				as a day value in the current month of the 
				current year. */
#define PDDATEEUROPE 66 /* Just like PDDATE except for ##/##/## and
				##-##-## interpretation. */
#define PDTIMEAMPM 3	/* only accepts a time.  If neither AM or PM 
				is specified and the hour is between 12:01 and 6:59 
				(inclusinve) PM is assumed.  Subtle point: If 
				the string is a single number, it is treated 
				as an hour value. */
#define PDTIME     4 	/* only accepts a time.  Just like PDTIMEAMPM 
				except 24 hr clock assumed if necessary (i.e. 
				AM/PM not given). */
#define PDCURRDATE 5	/* give back system date. The input string 
				is ignored; no parsing is done. */
#define PDCURRTIME 6	/* give back the system time. The input 
				string is ignored. */
#define PDCURRBOTH 7	/* give back the system date and time The 
				input string is ignored. */
#define PDUNIXBOTH 8	/* given a Unix date/time value, the Applix
				internal representation is returned. */
#define PDUNIXLONG 9	/* given a date-time string, the UNIX long
				is returned. */

/*------------------ ParseMonth() Options ---------------------------------*/
#define	PMNULLOK  1  /*Parse a month and year string, return DTNULL if
				     string is null or empty.*/
#define	PMNULLCURR 2 /*Parse a month and year string, return current
					  month and year if string is null or empty.*/
#define	PMCURR 3     /*Return the current month and year, stringP ignored.*/
#define	PMSYS  4     /*Return the Applix format for the month and year for the 
					system time provided by dtvalueP. (stringP ignored.) */
#endif

/*----------------------------------------------------------------------*/
/* In Rev 2.0, I have done the following:

(A) The granularity of system times converted to Applix format by PDCURRBOTH,
PDCURRTIME, or PDUNIXBOTH ParseDate() options was only accurate to 1 minute.  
It is now accurate to quad seconds.  I am somewhat concerned that this may 
have some unanticipated affect.  Here's a review of the issues as I know them.
If a document (or other entity) create/modify date is obtained from unix and 
converted to Applix format, when it is displayed to the user, a seconds value 
is never shown and therefore two such entities created within the same minute
should compare as equal.  This in fact happens if datecmp() is used since it 
always ignores the seconds value.  (datecmp() has other interesting properties 
such as handling DTNULL in comparisons correctly.)  If the intent is to see if
the entity has changed, then the two Applix dates should be compared
directly so that a file rewritten within the last minute will be recognized as 
such.  [It used to be the case that opening a pdb inset, changing, and 
closing quickly (within one minute) would not cause the DC display to be 
updated.  This is the problem that this change is supposed to fix.]

(B)  When formatting dates, if the year is to be displayed as a 2
digit value (as in mm/dd/yy) but the century is not the current
century, you will actually get a 4 digit year (as in 1/31/1897 or
1/31/2002).  This means that such date strings may be 2 characters
longer than in the past.  The PDB must change its idea of the width of
the affected date display formats.  [The motivation for this change is that
transfered date data in UDF ascii format can cause the century to
change in transit; i.e. a 2002 date represented in mm/dd/yy form when
pasted is parsed as a 1902 date.  ParseDate() always assumes the
current century.]

(C) The date format DTDFMIL has been changed so that it only shows a 2
digit year (for the current century).  This change was requested by
EDS to have Alis properly adhere to military standards.  Note that (B)
above involves no change to DTDFMIL since it already assumes a 4 digit year.

(D) When Formatdt() is requested to display the day of the week at the
end of the string it no longer puts the value in parens.  This makes
Formatdt() match the examples on the Personal Profile menu and makes
the leading and trailing placements look the same.  Examples:
   Old                 New
12/14/84(Tuesday)       12/14/84 Tuesday
12/14/84(Tue)           12/14/84 Tue
Tue 12/14/84            Tue 12/14/84      <-- unchanged.

(E) There is a new ParseMonth() procedure which will only allow entry of a
month and year value.  The day value is implicitly 1.  [This was invented for 
use in month-at-a-time calendar work, but is generally available and
would be useful for "spreadsheeters".]

(F)  There are corresponding new options to cause a date value to be
formatted as only a month and year value.  These are 
DTDFLETYR       e.g. February 1984
DTDFABBYR       e.g. Feb 1984
DTDFABB2YR      e.g. Feb 84
DTDFSLYR        e.g.  2/84       (one leading space if month <10)
DTDFSP4YR       e.g. 1984 02
DTDFSP2YR       e.g. 84 02

(G) ParseDate() has been fixed so that it will parse any date produced
by Formatdt (other than those in (F) above).  In particular it now
ignores the day-of-the-week.  Previously this caused
a problem in transfer in the PDB when day-of-the-week was displayed.

(H) A bug has been fixed which allowed ParseDate to parse the string
"jan - july" as a valid date.

(I) The sources have been reorganized so that ParseDate and Formatdt
can share common routines, but such routines are in a separate source
file so use of one routine does not force the other to be loaded from
libaos.a.  Also ParseMonth() is in a separate source file.
 
(J) There are 2 new options to ParseDate() which allow you to direct
it to interpret ##/##/## and ##-##-## as dd/mm/yy values (British/European 
style) rather than American mm/dd/yy style.  [The existing options
continue to base the decision on the separator used.  Slash and hyphen are
considered American, period is considered European.]  The new options
are PDNORMEUROPE and PDDATEEUROPE.  It is not clear yet how these
options are to be employed.

(K) There is a corresponding new Formatdt() option which allows
formatting with slashes European style; i.e. dd/mm/yy.  This is
DTDFEURSLASH.
*/

typedef struct {
	float fraction; /* fraction of one second */
	unsigned int second:6; /* 0-59 */
	unsigned int minute:6; /* 0-59 */
	unsigned int hour:6; /* 0-23 */
	unsigned int day:6; /* 1-31 */
	unsigned int month:6; /* 1-12 */
	int year; /* Anno Domini */
} AxTimeStruct;
