/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pageform.h
   Applix ASC version: 3
   Date: Mon Nov 16 13:52:52 1998
*/
#ifndef PAGEFORM_INCLUDED
#define PAGEFORM_INCLUDED 1

#define PFREV	3	/* Indicates a Rev. 2.1 .pf file */

/* The following definitions constitute the code which indicates
   the type of page the current record defines. Headers and Footers
   are packed together into a single record. Since records are
   self identifying, they do not need to be in any order, and not
   all need be present. The format of the file is as follows:

	A revision stamp, the first word of which is a -1
		followed by the revision.

	n pageInfo structures, each followed by a text buffer containing
		null-terminated strings which constitute the text for
		each of the text pieces that compose the headers and footers.
		Each of these records is fixed length.  The length of each
		record is sizeof( struct pageInfo) + MAXTEXT

	Note that pre-Rev1.20 files begin with a pageInfo structure, the
		first word of which contains a pageType - never -1.
*/

/* MODIFIED: 8/89 Added mControl field in pageInfo structure to make
 * room for new header and footer option Alternating margins - tomd
 */

/* MODIFIED: 1/10/89: Changed define PFREV from 2 to 3	*/
/* this was done so that a conversion of the .pf file	*/
/* would occur for the SLAN to rev2.1A upgrade.			*/
/*	jan-16-92 pmh:	changed/added DEF_HF_LINES and XTRA_HF_LINES to
					represent the ability to handle up to 10 lines of
					headers and footers. (see hdrftr.elfs for more info).
*/

#define frstInitial	0
#define frstContin	1
#define stdInitial	2
#define stdContin	3
#define evInitial	4
#define evContin	5
#define oddInitial	6
#define oddContin	7
#define MAXTYPES	8

/* The total amount of text for a page is limited to a total of
   MAXTEXT characters, which, currently, is equal to 400. */

#define MAXTEXT		800


/* the linePiece structure defines where in the text following
   the record structure the text for the specific piece will
   be found. A pointer of -1 indicates that the text is not present.
   The pAtt word contains the text attributes as defined in atmattr.h
*/

struct linePiece {
	short	pText;	/* pointer to the text */
	long	pAtt;	/* attribute flags */
};

/* a line is composed of a left piece, a center piece and a right
   piece.
*/

struct lineInfo {
	struct linePiece leftText;
	struct linePiece centText;
	struct linePiece rightText;
};

/* a header or footer is composed of a distance from the content
   area of the page, and an array of (0, 1 or 2) lines for the
   header or footer. The line count is redundant, because the
   pointer into the text, specified above, will be set to -1 if
   the text is not present.
*/

#define DEF_HF_LINES	4	/* allows up to 4 of each h/f lines */
							/* DEF_HF_LINES was MAXLINES		*/
#define XTRA_HF_LINES	10	/* allows up to 10 of each h/f lines */

struct hfStruct {
	short lineCount;
	struct lineInfo	hfLines [XTRA_HF_LINES];
};

/* a page is composed of version information, a page type,
	a header specification and a footer specification.
*/

struct pageInfo {
	short	pageType;	/* see above */
	short	lMargin;    /* The distance from the left edge of the paper
				to the left edge of the content area.  If
				revision marks are to appear on the left, 
				this value must be at least equal to the
				width of the revision mark characters. */
	short	rMargin;    /* The distance from the right edge of the 
				content area to the right edge of the paper.
				If revision marks are to appear on the right
				this value must be at least the width of the
				revision mark characters. */
	short	tMargin;    /* The distance from the top of the paper to the
				top of the first body line (the content area). */
	short	bMargin;    /* The distance from the bottom of the last 
				body line (the end of the content area) to the bottom edge of
				the paper.*/
	short	hDistance;  /* The distance from the top to the paper to the
				first header line. */
	short	fDistance;  /* The distance from the bottom of the 
				paper to the bottom of the last footer line. */
	short	revBars;    /* THIS FIELD IS OBSOLETE: cch Oct '90 */
	long	dAtt;		/* Default global text attributes */
	struct	hfStruct pHeader;
	struct	hfStruct pFooter;
				/* the MAXTEXT number of chars follows this
                    structure when recorded on disc. */
	short   mControl;   /* This is the field which determines if left and
                and right page margin control is set for this document.
                If it is this flag will cause the values of the left
                and right margins to swap depending on odd or even pages.
                This will then make the document printable in book style
                format. */
};
/*
    ----------------------------------------------------
    |       ||              |                   |       |
    |-------||tMargin       | hDistance         |rMargin|
    |lMargin||          Header Text             |______ |
    |       ||__________________________________|       |
    |       |   Body Text                       |       |
    |       |   Body Text                       |       |
    |       |___________________________________|       |
    |       ||          Footer Text             |       |
    |       ||bMargin       | fDistance         |       |
    |       ||              |                   |       |
    -----------------------------------------------------
*/
struct pageVersion {
	short	flag;
	short	version;
	short	filler[10];
};


#endif   /*  PAGEFORM_INCLUDED   */
