/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: qmenu.h
   Applix ASC version: 2
   Date: Mon Nov 16 13:52:53 1998
*/
#ifndef qmenuDef
#define MNONEVENT	-2
#define MCANCEL		-1
#define QTEXTMAX	TM_TAASIZE

/* the following are used in the ATM system font */ 
#define L_PARENTHESES '\033'
#define R_PARENTHESES '\035' 
 
/* The following are the routines which are provided by the qmenu
   mechanism. These are callable by the application. */

int MenuInit ();		/* Initialize the menu environment */
int MenuEvent ();		/* An event has occurred for the qmenu window.
						Process the event */
int MenuNewDefault ();	/* Reinitialize the specified menu item, and
						ask the application for the new defaut value */
int QMInCreate ();	/* Window creation component of qmenu
						initialization */
int QMInDisplay ();			/* Display component of qmenu initialization */
int QMenuInit ();		/* Parent routine to call QMInCreate and
						QMInDisplay */
int QMenuEvent ();	/* An event has occurred for the qmenu window.
						Process the event. Accepts all TmGetEvent arguments,
						not just subset as with MenuEvent */



/* The following are the routines expected by the qmenu
   mechanism. Default routines are provided in qmenu.a, but the
   application should provide them itself */

int MenuCommand ();			/* A pick has occurred on the "command"
						   line. (AKA "option" or "selection" line
						*/

char *MenuCurrent ();	/* Return a pointer to the current value
						   to be displayed in the enter field */

int MenuDefault ();			/* The application should, at this time,
						   take the opportunity to reinitialize
						   the specified item to its default value
						*/

int MenuError ();			/* Invalid data was entered into the specified
						   field. Perform any error processing.  */

int MenuIndex ();			/* The index key was struck on the specified
						   field. Perform the appropriate index
						   processing. */

int MenuPick ();			/* A pick has occurred on the specified
						   item which does not have an enterable
						   field associated with it. The application
						   should do whatever processing is
						   appropriate. */

int MenuText ();			/* A field read has successfully completed.
						   The application is called for the specified
						   field and should save both the entered
						   text and the validated text. */

int MenuValid ();		/* A listcheck string was not provided
						   for the specified filed. The application
						   should do the appropriate validity
						   checking */

int MenuHelp ();			/* The help key was pressed. Call the application
							to perform the help processing */

char *MenuItem ();		/* If the field is an autoindex field, with a zero
							length text strings and a valid item number,
							return a pointer to the text to display */

int MenuLogkey ();		/* The user has pressed a logical key the qmenu
							runtimes are not prepared to handle. Pass the key
							on to the application. */

/* The following structure is used to define the communication between
the atm and the qmenu routine QMenuFR. It is also used by the IMP to
define the information passed on to the atm for specifying the
preparation of running scripts at field read time. */

struct QMselect {
	char *ckString;
	short count;
	short menuid [1];
};

/* The atm expects the following routines to be present in qmenu.a:

QMenuSelect (parentID, menuID);
short parentID;		window id of parent window
short menuID;		menu id to be run on behalf of the user

QMenuFR (parentID, menuStruct, input)
short parentID;					window id of parent window
struct QMselect *menuStruct;	QMselect structure pointer
char *input;					input as typed by user 


*/

/*	The following structure is provided to allow the application to change
	the routine names that get called for each of the routines described above.
	By default, the routines named above are called. By calling the routine
	QMenuFunctions, the application can change the names. This also will allow
	for future expansion for shared libraries where the application must
	initialize the entry points.
*/

struct QMfunctions {
	int (*MenuCommand) ();
	char *(*MenuCurrent) ();
	int (*MenuDefault) ();
	int (*MenuError) ();
	int (*MenuHelp) ();
	int (*MenuIndex) ();
	char *(*MenuItem) ();
	int (*MenuLogkey) ();
	int (*MenuPick) ();
	int (*MenuText) ();
	short (*MenuValid) ();
};

struct QMIStruct {
	int initDisplay;
};

#endif
#define qmenuDef 1
