/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: thimble.h
   Applix ASC version: 69
   Date: Wed May  5 21:13:54 1999
*/
#ifndef _THIMBLE_H
#define _THIMBLE_H

/* Commentary on THIMpid and AxInTask.

    THIMpid and AxInTask area globals that describe the state of the
data system and the "current task".  They are indexes in the TaskInfo table.

THIMpid tells which set of globals are currently in memory.  (e.g which
task's WP, SS, GE, or Elf structure.) It also specifies the ownership
of data to be malloced/freed.

AxInTask describes the current machine stack (what you would get doing a 
"where" in dbx).

Usually, THIMpid and AxInTask are the same, but when an Elf task calls a WP,
the AxInTask is the Elf task but the THIMpid is the WP task.  Mallocs and Frees
go to the WP task.

This is different than when the WP calls Elf code.  In that case, the current
task (both THIMpid and AxInTask) remain  the WP, but the WP executing elf code.

When Elf code calls a WP, the WP can later pend for some reason, such as
proc'ing another task.  When that WP unblocks, we have to make sure it can
get no further events while its a subroutine of the Elf task.
*/

#if AX_NT
#include <axsocket.h>
#include <axtimev.h>
#endif

#include <eldata.h>
#include <axtime.h>
#include <setjmp.h>

#if VAX               /* special Applix setjmp/longjmp for VAX */
typedef int ax_jmp_buf[14];   /* see axjmp.s for more info */
#else
#define axsetjmp setjmp
#define axlongjmp longjmp
#define ax_jmp_buf jmp_buf
#endif

#include <memblk.h>
#include <paripc.h>

/*************** DIRECTORIES *************************************/

extern DLL_IMPORT char *ThimUserDir;   /* per-user -> $(HOME)/thimble.user */
extern DLL_IMPORT char *ThimCustomDir; /* per-group -> /usr/bin/thimble.custom */
extern DLL_IMPORT char *ThimUserElfDir;    /* optional user's commands directory */
extern DLL_IMPORT char *ThimUserModelsDir; /* optional user's model docs dir */
extern DLL_IMPORT char *AxLangDir;     /* subdir for lang-dependent files, default eng */
extern DLL_IMPORT char *ThimApplixDir; /* executables -> /usr/bin//thimble.data */
extern DLL_IMPORT char ThimTempDir[];  /* dir to put allemp files in */
extern DLL_IMPORT char *AxLocalLangDir;    /* subdir in axlocal for lang-dependent files*/

#define THIM_FIND_USER  0   /* find file in user dir */
#define THIM_FIND_CUSTOM 1  /* find in customizable dir */
#define THIM_FIND_APPLIX 2      /* find in applix place */
#define THIM_FIND_USERPLUS  3   /* find file in user dir or beyond */
#define THIM_FIND_CUSTOMPLUS 4  /* find in customizable dir or beyond */
#define THIM_FIND_ELF 5     /* Find Elf file in user's cmds dir & beyond */
#define THIM_FIND_MODEL 6   /* Find model doc in user's models */
#define THIM_FIND_AXLANG 7  /* Find lang-dependent directory */
#define THIM_FIND_AXLOCALLANG 8 /* Find axlocal's lang-dependent directory */

/*************** ATTRIBUTES/CONTROL OF TASKS *********************/

#define MAX_PIDS 100     /* Max number of simultanious tasks */
#define MAX_OPEN_FILES  20  /* per task */


/**************** ERROR RETURNS - ?? SHOULD BE IN STRFL!!! **********/

#define THIM_NO_PID 1       /* Could not create task, too many tasks running */
#define THIM_NO_MEMORY 2    /* Could not create task, not enough memory */

/****************** MACROS ******************************************/

#define TASK_ASSERT(n) {if (!(n)) taskLineAbort((char*)0,__LINE__);}

/****************** TYPEDEFS *************************************/

typedef struct task_info_type {
    char *data;
    void (*delete_func)();
    int (*pid_func)();
} task_info_type;

/* Java client channel info */

#define IOBUFSIZE 8192

typedef struct jchannel_t {
    struct jchannel_t *next;    /* next in list */
    int uid;                    /* sequencially assigned, non-repeating */
    int fd;                     /* channel */
    int logoutOnDisconnect;     /* if TRUE, exits on channel error */
    int showingHourglass;       /* client is showing hourglass in all wins */ 
    char **rsrcCacheMap;	/* array of resource cache bitmaps */
    /* message buffering: */
    struct jmsg_q *FirstQMsg;
    struct jmsg_q *LastQMsg;
    int BytesInMainQ;
    int BytesInLowPriQ;
    int BytesInBuffer;
    int QPos;
    /* base IO buffering: */
    int bufPos;  /* where to write next bytes from */
    int bufSize; /* how much of buffer has stuff in it */
    char ioBuf[IOBUFSIZE];
    /* statistics */
    char *username;
    int bytes_read;
    int msgs_read;
    int min_msg_read;
    int max_msg_read;
    int bytes_written;
    int msgs_written;
    int min_msg_written;
    int max_msg_written;
} jchannel_t;

typedef struct jmsg_q {
    int size;
    char *outBuf;
    struct jmsg_q *next;
} jmsg_q;


/* The context_block contains the major info about a task or, at the
least, is the root for access to all major task info */

typedef struct context_block {
    short in_use;   /* whether there is a task here */
    long uid;       /* task creation order (THIMpidUID) */
    short appid;    /* app type of the task */
    task_info_type *gfx; /* current graphic data associated with this task */
    task_info_type *ssx; /* current ss data associated with the task */
    task_info_type *wpx; /* current WP data associated with the task */
    task_info_type *mcx; /* MC data associated with the task         */
    ax_mem_block_info_t **block_types;  /* null termed array to Task's block desc's */
    char **my_blocks;   /* null termed array to my copies of block_types */
    struct backup_block *data;      /* head of context data */
    int start_me;   /* TRUE if task needs started now */
    int kill_me;    /* TRUE if task should be shot down */
    int unpend_me;  /* TRUE if wakeup for a task that just did a proc */
    int (*func)();  /* task's main routine if start_me */
    int usesElfArgs;    /* if TRUE, process elfArgv rather than cArgv */
    int argc;           /* argc if start_me */
    char **argv;        /* argv if start_me */
    data_ptr elfArgv;       /* elf args if startme */
    char *stack;        /* ptr to suspended stack area */
    int stack_size;     /* size of suspended stack area */
    ax_jmp_buf jump;        /* longjump addr for suspension return */
    int helper_type;        /* used by windowless Elf servers */
    int fds[MAX_OPEN_FILES];    /* filedes of all open files */
    int in_busy_loop;       /* TRUE only for suspended ELF commands */
    int ipc_response_waiting;   /* TRUE if waiting for IPC response */
    int ipc_timed_wait;         /* TRUE if doing a timed await */
    int ipc_true_timer;         /* TRUE if not shared */
    struct timeval ipc_run_time;    /* time when this task times out */
    int ipc_milliseconds;   /* time when this timer began */
    int recycle_only;   /* if TRUE, give timeslice only to active task, no keyboard poll */
    int have_polled;        /* set by kernel if have done at least one poll */
    int timed_out;          /* set by kernel if IPCawait timed out */
    int pended; /* I called ACproc/ACpproc (I am parent) */
    int ok_pended;  /* pended just till ACImOk (ACproc mode) */
    int elfCallBlocked; /* pended until called Elf routine returns */
    int interruptable;  /* TRUE if allow pre-emption */
    int have_okayed;    /* set by child while it does the OK */
    int parentTask;     /* task that started this one */
    int child_retcode;      /* returned code of a pended proc (ACPPROC) */
    data_ptr child_retdata; /* NULL or returned child Elf data */
    int haveSocketInput;    /* set by select() code, notifies schedular */
    struct msg_info *msgs;  /* head of incoming messages list */
    struct msg_info *rsp;   /* head of incoming response */
    struct msg_info **deferList;    /* deferred messages by AxDeferMessage() */
    char **allocPools;      /* dense allocation support */
    int numPools;           /* size of allocPools array */
    int numDefered;             /* count of deferred messages */
    long *doc_interest;     /* array docs this is interested in */
    char **profile_interest;    /* array of profile strs this is interested in */
    long fscurrdoc;         /* Task's FSCurrDoc.DocId */
    int taskMaster;     /* only relavent for Elf tasks */
    int priorTaskMaster;        /* only relavent for Elf tasks */
    int parasite_of;        /* 0 or ID of DC that proced this local inset */
    int (*sigfpehandler)(); /* task's SIGFPE handler if any */
    void (*elfAbortHandler)(); /* task's elf abort handler */
    int activationData; /* Id of task to VfyContext() on activation */
    char *ElfIpcMessage;
    data_ptr ElfIpcElfArray;
    int ElfSenderPid;       /* in fs_private.c */
    int ElfRetCode; /* HACK to ElfRetErr returns to use same mechanism as ElfRespond normal returns */
    int alarm_mode; /* see ALARM_MODE_XXX below */
    int alarm_millisec; /* relavent for ALARM_MODE_RECYCLE only */
    struct timeval alarm_wakeup_time;    /* time when this task gets notified */
    int alarm_has_fired;        /* gets set when time above has elapsed */
    int ss_func_call_counter;   /* tracks when SS in middle of func call */
    int ss_task_calling_bltin_func; /* AxInTask at builtin calling time */
    char ldata[16]; 
    data_ptr pokeCodes;         /* poke codes that this task accepts */
    unsigned int
        topLevel : 1,       /* set for top-level tasks */
        deamon : 1,     /* Uses "task-0" storage areas rather than HIMEM */
        defer_acimok : 1,   /* if TRUE, hold off on ACImOK calls */
        showing_hourglass : 1,  /* TRUE if task's windows hourglassed */
        slaved_hourglass : 1,/* TRUE if task hg'ed beacause waiting on hg'ed task */
        breakpointed : 1, /* Can only be true for ELF tasks */
        created_tempfiles : 1,  /* TRUE if this task has created temps */
        loaded_elf_images : 1,  /* TRUE if loaded task transient elf code */
        waiting_on_socket : 1,  /* prevents RPC interleave */
        TaskUsesJavaDisplay : 1, /* remote display */
        suppressExitErrorBox : 1; /* if FALSE, task posts box on error exit
                    even if exit occured before ACImOK, but not for fully
                    pended boxes (understand this and get a dime) */


        /* tasks_interested is a list of tasks that will be informed when this
           task goes away (AxDoExit()) and if this task is an application task
           then all messages going via FSPokeAll() are also sent to this task 
           For more info look at the above 2 functions. 
           To set tasks_interested the elf layer bltin is NOTE_TASK_INTERESTED$()
        */         
        int *tasks_interested; /* NULL or ptr to tasks lists (null termed) */
    jchannel_t *javaChannel; /* if TaskUsesJavaDisplay, this is default channel */
    int javaWinId;  /* window id on client side (only used by SS for multi-window) */
    struct info_box_t *infoBox; /* NULL or ptr if task has a blocking info box */
} context_block;

#define ALARM_MODE_NOP 0        /* task not doing alarms */
#define ALARM_MODE_FLAG 1       /* set alarm_has_fired in TaskInfo */
#define ALARM_MODE_MSG 2        /* send task a message on alarm */
#define ALARM_MODE_RECYCLE 3    /* return to schedular */

/* The IPC chain is read from the head.  Messages are
appended to at the end of list */

typedef struct msg_info {   /* list of incoming "ipc" messages */
    struct msg_info *next;  /* "ipc" message in chain (must be first element) */
    int sender_pid;         /* THIMpid that sent message */
    int sender_app;         /* app_id of sender */
    int type;               /* IPC_SRT ... */
    int discard_me;         /* msg already read, to be freed on next read */
    int gotten;             /* has been absorbed */
    int bytecount;          /* Number of bytes in msg[] */
    char msg[1];            /* varying extent */
} msg_info;

/* Format of "SCAC" messages.  The union in the buffer is not called out
here to be able to avoid excessive global include file requirements. SIGH */

typedef struct {
    char header[IPC_HEAD_LEN];
    int type;       /* see scac.h for list */
    char buf[800];  /* apply interesting stuff here */
} scac_message_info;

extern DLL_IMPORT int AxInTask;        /* Current task */
extern DLL_IMPORT short THIMpid;       /* Current task's data id, usually same as AxInTask */
extern DLL_IMPORT context_block TaskInfo[];    /* Array of task descriptions */
extern DLL_IMPORT short THIMappid;     /* Current tasks application type */
extern DLL_IMPORT ax_jmp_buf KernelJump;   /* jump to ATM's main loop */
extern DLL_IMPORT char *StackBase;     /* Stack base of ATM's main loop */
extern DLL_IMPORT char THIMsysDir[];   /* Where the THIMBLE executable lies */

/************* RANDOM STUFF ************************************/

/* The following codes are used the by POPUP logic to get commands
    from the personal profile.
    This is only tmpopup.c - thim_fs.c usage.
*/

#define POP_L_BEG   0
#define POP_L_END   1
#define POP_M_BEG   2
#define POP_M_END   3
#define POP_R_BEG   4
#define POP_R_END   5


#endif
