/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: x2win.h
   Applix ASC version: 46
   Date: Mon Nov 16 13:52:53 1998
*/
#ifndef _X2WIN_H
#define _X2WIN_H

#ifndef True
# ifdef TRUE
#  define True TRUE
#  define False FALSE
# else
#  define True 1
#  define False 0
# endif
#endif

#define AllPlanes       (~0)
#define BitmapUnit(dpy)  0


typedef long            XtArgVal;
typedef unsigned char   XtEnum;
typedef char *String;
typedef char *XPointer;

typedef struct {
    XtArgVal    name;
    XtArgVal    value;
} Arg, *ArgList;

#define XtSetArg(arg, n, d) \
    ((void)( (arg).name = (n), (arg).value = (XtArgVal)(d) ))

#define AxArgVal XtArgVal

/*
 * To allow arbitrary information with fonts, there are additional properties
 * returned.
 */
typedef struct {
    Atom name;
    unsigned long card32;
} XFontProp;
 

/*
 * Extensions need a way to hang private data on some structures.
 */
typedef struct _XExtData {
    int number;     /* number returned by XRegisterExtension */
    struct _XExtData *next; /* next item on list of data for structure */
    int (*free_private)();  /* called to free private storage */
    XPointer private_data;  /* data private to this extension. */
} XExtData;


typedef void *XtPointer;
typedef void (*XtCallbackProc)();

typedef struct _XtCallbackRec {
    XtCallbackProc      callback;
    XtPointer           closure;
} XtCallbackRec, *XtCallbackList; 



/*
 * Display datatype maintaining display specific data.
 */
typedef struct _XDisplay {
        int dummy_for_windows;
    int qlen;   /* for now, to get QLength macro to work in DE */
} Display;                   


/* GC components: masks used in CreateGC, CopyGC, ChangeGC, OR'ed into
   GC.stateChanges */
 
#define GCFunction              (1L<<0)
#define GCPlaneMask             (1L<<1)
#define GCForeground            (1L<<2)
#define GCBackground            (1L<<3)
#define GCLineWidth             (1L<<4)
#define GCLineStyle             (1L<<5)
#define GCCapStyle              (1L<<6)
#define GCJoinStyle     (1L<<7)
#define GCFillStyle     (1L<<8)
#define GCFillRule      (1L<<9)
#define GCTile          (1L<<10)
#define GCStipple       (1L<<11)
#define GCTileStipXOrigin   (1L<<12)
#define GCTileStipYOrigin   (1L<<13)
#define GCFont          (1L<<14)
#define GCSubwindowMode     (1L<<15)
#define GCGraphicsExposures     (1L<<16)
#define GCClipXOrigin       (1L<<17)
#define GCClipYOrigin       (1L<<18)
#define GCClipMask      (1L<<19)
#define GCDashOffset        (1L<<20)
#define GCDashList      (1L<<21)
#define GCArcMode       (1L<<22)
 
#define GCLastBit       22

typedef struct {
        int function;           /* logical operation */
        unsigned long plane_mask;/* plane mask */
        unsigned long foreground;/* foreground pixel */
        unsigned long background;/* background pixel */
        int line_width;         /* line width */
        int line_style;         /* LineSolid, LineOnOffDash, LineDoubleDash */
        int cap_style;          /* CapNotLast, CapButt,
                                   CapRound, CapProjecting */
        int join_style;         /* JoinMiter, JoinRound, JoinBevel */
        int fill_style;         /* FillSolid, FillTiled,
                                   FillStippled, FillOpaeueStippled */
        int fill_rule;          /* EvenOddRule, WindingRule */
        int arc_mode;           /* ArcChord, ArcPieSlice */
        Pixmap tile;            /* tile pixmap for tiling operations */
        Pixmap stipple;         /* stipple 1 plane pixmap for stipping */
        int ts_x_origin;        /* offset for tile or stipple operations */
        int ts_y_origin;
        HFONT font;     /* default text font for text operations */
        int subwindow_mode;     /* ClipByChildren, IncludeInferiors */
        int graphics_exposures; /* boolean, should exposures be generated */
        int clip_x_origin;  /* origin for clipping */
        int clip_y_origin;
        Pixmap clip_mask;       /* bitmap clipping; other calls for rects */
        int dash_offset;        /* patterned/dashed line information */
        int dash_list_size;
        DWORD dashes[10];
} XGCValues; 
        
/*
 * Graphics context.  All Xlib routines deal in this rather than
 * in raw protocol GContext ID's.  This is so that the library can keep
 * a "shadow" set of values, and thus avoid passing values over the
 * wire which are not in fact changing.
 */
/** explanation behind 'clip_mask_is_hregion' :
    on X, the XSetClipRectangles call will set the GC so that future calls
    with that GC use the given rect(s) as a clipping mask. No objects are
    created so none need to be free'd. On Windows, however, we set the
    clip_mask element of the GC to be a GDI object of type HRGN. This means
    it has to be free'd (deleted). When XSetClipMask or XSetClipRectangles
    is called the type of the current clip mask is unknown and we can't do
    a MS32_GetObjectType because the handle may be invalid. The only way to
    make sure we delete the HRGNs is to set this flag on creation of a Windows
    region handle.                     -donato
    ****************************************************************/
typedef struct _XGC {
    unsigned int clip_mask_is_hregion : 2;
    XGCValues   values;
} *GC;


#if 0
/* LineStyle */
#define LineSolid       0
#define LineOnOffDash       1
#define LineDoubleDash      2
#endif

/* capStyle */
#define CapNotLast      0
#define CapButt         1
#define CapRound        2
#define CapProjecting       3
 
/* joinStyle */
#define JoinMiter       0
#define JoinRound       1
#define JoinBevel       2

/* fillStyle */
#define FillSolid       0
#define FillTiled       1
#define FillStippled        2
#define FillOpaqueStippled  3
 
/* fillRule */
#define EvenOddRule     0
#define WindingRule     1
 

typedef struct {
    short x, y;
    unsigned short width, height;
    short angle1, angle2;
} XArc;


typedef struct {
    short x, y;
} XPoint;
 
/*
 * per character font metric information.
 */
typedef struct {
    short       lbearing;       /* origin to left edge of raster */
    short       rbearing;       /* origin to right edge of raster */
    short       width;          /* advance to next char's origin */
    short       ascent;         /* baseline to top edge of raster */
    short       descent;        /* baseline to bottom edge of raster */
    unsigned short attributes;  /* per char flags (not predefined) */
} XCharStruct;

typedef struct {
    HFONT   fid;
    unsigned    min_char_or_byte2;/* first character */
    unsigned    max_char_or_byte2;/* last character */
    unsigned    min_byte1;      /* first row that exists */
    unsigned    max_byte1;      /* last row that exists */
    int         all_chars_exist;/* flag if all characters have non-zero size*/
    unsigned    default_char;   /* char to print for undefined character */
    XCharStruct min_bounds;     /* minimum bounds over all existing char*/
    XCharStruct max_bounds;     /* maximum bounds over all existing char*/
    XCharStruct *per_char;      /* first_char to last_char information */
    int         ascent;         /* log. extent above baseline for spacing */
    int         descent;        /* log. descent below baseline for spacing */
    XExtData    *ext_data;      /* hook for extension to hang data */
    int         n_properties;   /* how many properties there are */
    XFontProp   *properties;    /* pointer to array of additional properties*/
} XFontStruct; 

#define EVENT_HEAD           \
        int type;            /* of event */                                   \
        unsigned long serial;/* # of last request processed by server */      \
        BOOL send_event;      /* true if this came from a SendEvent request */ \
        Display *display;    /* Display the event was read from */            \
        Window window;       /* "event" window it is reported relative to */  \
        Window root;         /* root window that the event occured on */      \
        Window subwindow;    /* child window */                               \
        Time time;           /* milliseconds */                               \
        int x, y;            /* pointer x, y coordinates in event window */   \
        int x_root, y_root;  /* coordinates relative to root */               \
        unsigned int state;  /* key or button mask */                         \
        int same_screen;     /* same screen flag */                           

typedef struct XKeyEvent {
        EVENT_HEAD
        unsigned int keycode;   /* detail */
} XKeyEvent;

typedef struct XButtonEvent {
        EVENT_HEAD
        unsigned int button;    /* detail */
} XButtonEvent; 

typedef struct {
        EVENT_HEAD
        char is_hint;           /* detail */
} XMotionEvent;

typedef struct {
        EVENT_HEAD
        int mode;
        int detail;
} XFocusChangeEvent;
typedef XFocusChangeEvent XFocusInEvent;
typedef XFocusChangeEvent XFocusOutEvent;

typedef struct {
        int type;
        unsigned long serial;  /* # of last request processed by server */
        BOOL send_event;        /* true if this came from a SendEvent request */
        Display *display;      /*Display the event was read from */
        Window window;
        int x, y;
        int width, height;
        int count;             /* if non-zero, at least this many more */
} XExposeEvent;  


typedef struct XConfigureEvent {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    BOOL send_event;    /* true if this came from a SendEvent request */
    Display *display;   /* Display the event was read from */
    Window event;
    Window window;
    int x, y;
    int width, height;
    int border_width;
    Window above;
    BOOL override_redirect;
} XConfigureEvent;

           
typedef struct {
    int type;
    unsigned long serial;    /* # of last request processed by server */
    BOOL send_event;         /* true if this came from a SendEvent request */
    Display *display;        /* Display the event was read from */
    Drawable drawable;
    int x, y;
    int width, height;
    int count;         /* if non-zero, at least this many more */
    int major_code;    /* core is CopyArea or CopyPlane */
    int minor_code;    /* not defined in the core */
} XGraphicsExposeEvent; 

typedef struct {
    int type;
    unsigned long serial;    /* # of last request processed by server */
    BOOL send_event;         /* true if this came from a SendEvent request */
    Display *display;        /* Display the event was read from */
    Window window;    /* window on which event was requested in event mask */
} XAnyEvent;


typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    BOOL send_event;    /* true if this came from a SendEvent request */
    Display *display;   /* Display the event was read from */
    Window window;
    Atom atom;
    Time time;
    int state;      /* NewValue, Deleted */
} XPropertyEvent;

typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    BOOL send_event;    /* true if this came from a SendEvent request */
    Display *display;   /* Display the event was read from */
    Window window;
    Atom selection;
    Time time;
} XSelectionClearEvent;

typedef struct {
        int type;
        unsigned long serial;   /* # of last request processed by server */
        BOOL send_event;        /* true if this came from a SendEvent request */

        Display *display;       /* Display the event was read from */
        Window owner;
        Window requestor;
        Atom selection;
        Atom target;
        Atom property;
        Time time;
} XSelectionRequestEvent;     


typedef struct {
    int type;
    unsigned long serial;   /* # of last request processed by server */
    BOOL send_event;    /* true if this came from a SendEvent request */
    Display *display;   /* Display the event was read from */
    Window requestor;
    Atom selection;
    Atom target;
    Atom property;      /* ATOM or None */
    Time time;
} XSelectionEvent;


/*
 * this union is defined so Xlib can always use the same sized
 * event structure internally, to avoid memory fragmentation.
 */
typedef union _XEvent {
        int             type;     /* must not be changed; first element */
        MSG             msg;
        XAnyEvent       xany;
        XKeyEvent       xkey;
        XButtonEvent    xbutton;
        XExposeEvent    xexpose;
        XMotionEvent    xmotion;
        XConfigureEvent xconfigure;  
        XGraphicsExposeEvent xgraphicsexpose;         
        XFocusChangeEvent    xfocus;
        XSelectionClearEvent xselectionclear;
        XSelectionRequestEvent xselectionrequest;
        XSelectionEvent xselection;
        XPropertyEvent xproperty;
        long            pad[24];
} XEvent;                    


typedef unsigned short  Dimension;  /* Size in pixels */
typedef int     Widget;
typedef int     Boolean;
typedef int     WidgetClass;  /* can't find reference in UNIX includes! */

#define QLength(dpy)    (dpy ? ((dpy)->qlen) : dpy)

/*
 * Data structure used by color operations
 */
typedef struct {
    unsigned long pixel;
    unsigned short red, green, blue;
    char flags;  /* do_red, do_green, do_blue */
    char pad;
} XColor;


typedef struct {
    short x1, x2, y1, y2;
} REGION_BOX;

typedef struct _XRegion {
    long size;
    long numPoints;
    XPoint *points;
    REGION_BOX extents;
    int fill_rule;
    int xoffset, yoffset;
    HRGN region_handle;
} REGION;

typedef struct _XRegion *Region;


typedef struct {
    long your_event_mask;
} XWindowAttributes; 


/* A structure for storing,retrieving the results */
/* of XSetSelectionOwner,XGetSelectionOwner calls */
typedef struct axselectionowner {
    Atom selection;
    Window ownerWindow;
    Time time;
    struct axselectionowner *nextSel, *prevSel;
} AXSelectionOwner;

#define XA_LAST_PREDEFINED ((Atom) 68)
typedef struct AxAtomList {
    Atom atom;
    char *propertyName;
    int  winFormatID;
    struct AxAtomList *nextItem;
} AXATOMLIST;

/*****************************************************************
 *  IMAGING
 *****************************************************************/
 
/*
 * Data structure for "image" data, used by image manipulation routines.
 * We use the Windows device Independent Bitmap (DIB) structure.
 */

typedef struct _XImage {
    int width, height;      /* size of image */
    int xoffset;        /* number of pixels offset in X direction */
    int format;         /* XYBitmap, XYPixmap, ZPixmap */
    char *data;         /* pointer to image data */
    int byte_order;     /* data byte order, LSBFirst, MSBFirst */
    int bitmap_unit;        /* quant. of scanline 8, 16, 32 */
    int bitmap_bit_order;   /* LSBFirst, MSBFirst */
    int bitmap_pad;     /* 8, 16, 32 either XY or ZPixmap */
    int depth;          /* depth of image */
    int bytes_per_line;     /* accelarator to next line */
    int bits_per_pixel;     /* bits per pixel (ZPixmap) */
    unsigned long red_mask; /* bits in z arrangment */
    unsigned long green_mask;
    unsigned long blue_mask;
    char *obdata;        /* hook for the object routines to hang on */
    struct _XImage *(*create_image)();

    LPBITMAPINFO bitmap_info;
    HPALETTE palette_handle;
    HBITMAP  bitmap_handle;
    unsigned char *image_data;
    } XImage;
        
/* ImageFormat -- PutImage, GetImage */
#define XYBitmap        0   /* depth 1, XYFormat */
#define XYPixmap        1   /* depth == drawable depth */
#define ZPixmap         2   /* depth == drawable depth */
 
 
typedef struct {
    short x1, y1, x2, y2;
} XSegment;

typedef struct {
    short x, y;
    unsigned short width, height;
} XRectangle;

typedef struct _XComposeStatus {
    char *compose_ptr;          /* state table pointer */
    int chars_matched;          /* match state */
} XComposeStatus;


/*
 * PolyText routines take these as arguments.
 */
typedef struct {
    char *chars;                /* pointer to string */
    int nchars;                 /* number of characters */
    int delta;                  /* delta between strings */
    Font font;                  /* font to print it in, None don't change */
} XTextItem;

typedef struct {                /* normal 16 bit characters are two bytes */
    unsigned char byte1;
    unsigned char byte2;
} XChar2b;

typedef struct {
    XChar2b *chars;             /* two byte characters */
    int nchars;                 /* number of characters */
    int delta;                  /* delta between strings */
    HFONT font;                 /* font to print it in, None don't change */
} XTextItem16;


typedef struct {
        Colormap colormap;
        unsigned long red_max;
        unsigned long red_mult;
        unsigned long green_max;
        unsigned long green_mult;
        unsigned long blue_max;
        unsigned long blue_mult;
        unsigned long base_pixel;
        VisualID visualid;              /* added by ICCCM version 1 */
        XID killid;                     /* added by ICCCM version 1 */
} XStandardColormap; 


/*
 * Visual structure; contains information about colormapping possible.
 */
typedef struct {
//        XExtData *ext_data;     /* hook for extension to hang data */
        VisualID visualid;      /* visual id of this visual */
#if defined(__cplusplus) || defined(c_plusplus)
        int c_class;            /* C++ class of screen (monochrome, etc.) */
#else
        int class;              /* class of screen (monochrome, etc.) */
#endif
        unsigned long red_mask, green_mask, blue_mask;  /* mask values */
        int bits_per_rgb;       /* log base 2 of distinct color values */
        int map_entries;        /* color map entries */
} Visual;


typedef struct {
        char *res_name;
        char *res_class;
} XClassHint;


/*
 * Data structure for setting window attributes.
 */
typedef struct {
    Pixmap background_pixmap;   /* background or None or ParentRelative */
    unsigned long background_pixel; /* background pixel */
    Pixmap border_pixmap;   /* border of the window */
    unsigned long border_pixel; /* border pixel value */
    int bit_gravity;        /* one of bit gravity values */
    int win_gravity;        /* one of the window gravity values */
    int backing_store;      /* NotUseful, WhenMapped, Always */
    unsigned long backing_planes;/* planes to be preseved if possible */
    unsigned long backing_pixel;/* value to use in restoring planes */
    int save_under;        /* should bits under be saved? (popups) */
    long event_mask;        /* set of events that should be saved */
    long do_not_propagate_mask; /* set of events that should not propagate */
    int override_redirect; /* boolean value for override-redirect */
    Colormap colormap;      /* color map to be associated with window */
    Cursor cursor;      /* cursor to be displayed (or None) */
} XSetWindowAttributes;


/* 
 * X function declarations.
 */
extern XImage *XCreateImage(Display*, Visual*, unsigned int, int, int, char*,
                                        unsigned int, unsigned int, int, int);
extern XImage *XGetImage(Display*, Drawable, int, int, unsigned int,
                                    unsigned int, unsigned long, int);
extern void XDestroyImage(XImage*);
extern unsigned long XGetPixel(XImage*, int, int);
extern int XPutPixel(XImage*, int, int, unsigned long);

Display *XOpenDisplay(const char*);
Atom XInternAtom(Display*, const char*, BOOL);
/*** Not actually ported, but defined to abort (see x2win.c) ***/
Cursor XCreatePixmapCursor(Display*, Pixmap, Pixmap, XColor*, XColor*,
                                        unsigned int, unsigned int);
Cursor XCreateGlyphCursor(Display*, Font, Font, unsigned int, unsigned int,
                                                        XColor*, XColor*);
Cursor XCreateFontCursor(Display*, unsigned int);
GC XCreateGC(Display*, Drawable, unsigned long, XGCValues*);
Pixmap XCreatePixmap(Display*,Drawable,unsigned int,unsigned int,unsigned int);
Pixmap XCreatePixmapFromBitmapData(Display*, Drawable, char*, unsigned int,
                    unsigned int, unsigned long, unsigned long, unsigned int);
Window XGetSelectionOwner(Display*, Atom);
long XMaxRequestSize(Display*);
int XIconifyWindow(Display*, Window, int);
int XSetWMProtocols(Display*, Window, Atom*, int);
XFontStruct *XLoadQueryFont(Display*, const char*);
void XSetTransientForHint(Display*, Window, Window);

void XChangeGC(Display*, GC, unsigned long, XGCValues*);
int XChangeProperty(Display*, Window, Atom, Atom, int, int,
                                const unsigned char*, int);
void XChangeWindowAttributes(Display*, Window, unsigned long,
                                        XSetWindowAttributes*);
void XClearArea(Display*, Window, int, int, unsigned int, unsigned int, BOOL);
void XCloseDisplay(Display*);
void XConvertSelection(Display*, Atom, Atom, Atom, Window, Time);
BOOL XCopyArea(Display*, Drawable, Drawable, GC, int, int, unsigned int,
                                                unsigned int, int, int);
void XCopyGC(Display*, GC, unsigned long, GC);
BOOL XCopyPlane(Display*, Drawable, Drawable, GC, int, int, unsigned int,
                                    unsigned int, int, int, unsigned long);
void XDefineCursor(Display*, Window, Cursor);
void XDeleteProperty(Display*, Window, Atom);
void XDestroyWindow(Display*, Window);
BOOL XDrawArc(Display*, Drawable, GC, int, int, unsigned int, unsigned int,
                                                                int, int);
void XDrawArcs(Display*, Drawable, GC, XArc*, int);
void XDrawImageString(Display*, Drawable, GC, int, int, const char*, int);
void XDrawImageString16(Display*, Drawable, GC, int, int, const XChar2b*, int);
BOOL XDrawLine(Display*, Drawable, GC, int, int, int, int);
BOOL XDrawLines(Display*, Drawable, GC, XPoint*, int, int);
void XDrawPoint(Display*, Drawable, GC, int, int);
void XDrawPoints(Display*, Drawable, GC, XPoint*, int, int);
BOOL XDrawRectangle(Display*, Drawable, GC, int,int,unsigned int,unsigned int);
void XDrawRectangles(Display*, Drawable, GC, XRectangle*, int);
void XDrawSegments(Display*, Drawable, GC, XSegment*, int);
BOOL XDrawString(Display*, Drawable, GC, int, int, const char*, int);
BOOL XDrawString16(Display*, Drawable, GC, int, int, const XChar2b*, int);
BOOL XDrawText(Display*, Drawable, GC, int, int, XTextItem*, int);
BOOL XDrawText16(Display*, Drawable, GC, int, int, XTextItem16*, int);
BOOL XFillArc(Display*, Drawable, GC, int, int, unsigned int, unsigned int,
                                                                int, int);
void XFillArcs(Display*, Drawable, GC, XArc*, int);
BOOL XFillPolygon(Display*, Drawable, GC, XPoint*, int, int, int);
void XFillRectangles(Display*, Drawable, GC, XRectangle*, int);
void XFlush(Display*);
void XFree(void*);      /* Unix/X declares char* */
void XFreeCursor(Display*, Cursor);
void XFreeGC(Display*, GC);
void XFreePixmap(Display*, Pixmap);
BOOL XGetGCValues(Display*, GC, unsigned long, XGCValues*);
BOOL XGetGeometry(Display*, Drawable, Window*, int*, int*, unsigned int*,
                            unsigned int*, unsigned int*, unsigned int*);
int XGetWindowProperty(Display*, Window, Atom, long, long, BOOL, Atom, Atom*,
                        int*, unsigned long*, unsigned long*, unsigned char**);
int XGetWindowAttributes(Display*, Window, XWindowAttributes*);
void XMapWindow(Display*, Window);
void XMoveResizeWindow(Display*, Window, int, int, unsigned int, unsigned int);
void XMoveWindow(Display*, Window, int, int);
void XNextEvent(Display*, XEvent*);
void XPeekEvent(Display*, XEvent*);
void XPutImage(Display*, Drawable, GC, XImage*, int, int, int, int,
                                        unsigned int, unsigned int);
BOOL XQueryPointer(Display*, Window, Window*, Window*, int*, int*, int*, int*,
                                                                unsigned int*);
void XRaiseWindow(Display*, Window);
void XResizeWindow(Display*, Window, unsigned int, unsigned int);
void XRestackWindows(Display*, Window*, int);
void XSelectInput(Display*, Window, long);
void XSendEvent(Display*, Window, BOOL, long, XEvent*);
void XSetWindowBackground(Display*, Window, unsigned long);
void XSetWindowColormap(Display*, Window, Colormap);
void XSync(Display*, BOOL);
int XTextWidth(XFontStruct*, const char*, int);
BOOL XTranslateCoordinates(Display*, Window, Window, int, int, int*, int*,
                                                                Window*);
void XUnmapWindow(Display*, Window);
void XWarpPointer(Display*, Window, Window, int, int, unsigned int,
                                            unsigned int, int, int);


#endif /* _X2WIN_H */
