/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: gtwycomm.c
   Applix ASC version: 8
   Date: Fri May 14 21:14:59 1999
*/
/* STRIP LINES THROUGH HERE */

/*
  gtwycomm.c - Utility funcs to communicate to 'aagtwy' in HTTP protocol.

	  A typical comm session might like this:
  
	  sockfd = aagtwyConnect("babu", 80);
	  aagtwySendf(sockfd,"GET /index.html HTTP/1.0\r\n");
	  aagtwySendf(sockfd,"User-agent: Applix Anyware/4.41\r\n");
	  aagtwySendf(sockfd,"\r\n");
	  response = aagtwyGetResponse(sockfd,&responselen,&contentlen);
		  // process response
	  aagtwyFreeResponse(response);
	  aagtwyDisconnect(sockfd);
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <fcntl.h>

#if AXMS
#include <process.h>
#include <io.h>
#include <winsock.h>
#define open _open
#define read _read
#define strncasecmp strnicmp
#define CLOSE_CHANNEL closesocket
#else
#include <unistd.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <netdb.h>
#define CLOSE_CHANNEL close
#endif


#if !BUILD_STANDALONE
#include <aos.h>
void TaskFree(int pid, void *ptr);
char *TaskAlloc(int pid, unsigned int size);
char *TaskRealloc(int pid, void *inP, unsigned int newsize);
#endif


/**
 * Get hostinfo from passed "ip-addr" or "hostname"
 */
static struct hostent *aagethostinfo (char *host) {
	u_long addr;
	return((-1 != (int)(addr = inet_addr(host))) ?
		   gethostbyaddr((char*)&addr,sizeof(addr),AF_INET):gethostbyname(host));
}

/**
 * Connect to 'aagtwy' and return socket fd
 */
int aagtwyConnect (char *host, int port) {
	short sport = port;
    struct sockaddr_in in_saddr;
	struct hostent *hostinfo = aagethostinfo(host);
	int sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (-1 == sockfd)
		return -1;
	if (NULL == hostinfo) {
		CLOSE_CHANNEL(sockfd);
		return -1;
	}
	memset(&in_saddr, 0, sizeof(in_saddr));
	in_saddr.sin_family = AF_INET;
	in_saddr.sin_addr.s_addr = *((long*)(hostinfo->h_addr));
	in_saddr.sin_port = htons(sport);
	if (-1 == connect(sockfd,(struct sockaddr *)(&in_saddr),sizeof(in_saddr))) {
		CLOSE_CHANNEL(sockfd);
		return -1;
	}
	return sockfd;
}

/**
 * Disconnect 'aagtwy' socket connection
 */
void aagtwyDisconnect (int sockfd) {
	shutdown(sockfd,2);
	CLOSE_CHANNEL(sockfd);
}

/**
 * Send bytes to 'aagtwy'
 */
int aagtwySend (int sockfd, char *buf, int count) {
	int offset = 0;
#if !AXMS
	while (count > 0) {
		int wbytes = write(sockfd,buf+offset,count);
		if (wbytes < 0) {
            if ((EINTR == errno) || (EAGAIN == errno) || (EWOULDBLOCK == errno))
                errno = 0;
            else
                return -1;
		}
		else {
			count -= wbytes;
			offset += wbytes;
		}
	}
#else
    DWORD last_err;
    int wbytes;
    while (count > 0) {
        wbytes = send(sockfd,buf+offset,count,0);
        if (wbytes < 0) {
            last_err = WSAGetLastError();
            if ( (last_err == WSAEINTR) ||
                (last_err == WSAEWOULDBLOCK) )
                errno = 0;
            else
                return -1;
        }
        else {
            count -= wbytes;
            offset += wbytes;
        }
    }
#endif
	return offset;
}

/**
 * Send formatted printf like buffer to 'aagtwy'
 */
int aagtwySendf (int sockfd, char *fmt, ...) {
	char bigbuf[8192];
	int len;
	va_list args;
	va_start(args,fmt);
	vsprintf(bigbuf,fmt,args);
	va_end(args);
	len = strlen(bigbuf);
	return (len == aagtwySend(sockfd,bigbuf,len));
}

/**
 * Recieve bytes from 'aagtwy'
 */
int aagtwyRecv (int sockfd, char *buf, int count) {
	int offset = 0;
#if !AXMS
	while (count > 0) {
		int rbytes = read(sockfd,buf+offset,count);
		if (rbytes < 0) {
            if ((EINTR == errno) || (EAGAIN == errno) || (EWOULDBLOCK == errno))
                errno = 0;
            else
                return -1;
		}
		else if (0 == rbytes)
			break;
		else {
			count -= rbytes;
			offset += rbytes;
		}
	}
#else
	DWORD last_err;
	int rbytes;
    while (count > 0) {
        rbytes = recv(sockfd,buf+offset,count,0);
        if (rbytes < 0) {
			last_err = WSAGetLastError();
            if ( (last_err == WSAEINTR) || 
				(last_err == WSAEWOULDBLOCK) )
                errno = 0;
            else
                return -1;
        }
        else if (0 == rbytes)
            break;
        else {
            count -= rbytes;
            offset += rbytes;
        }
    }
#endif
	return offset;
}

/**
 * Returns TRUE if bufP points to full response string (blank line ender)
 */
int aagtwyIsResponseComplete (char *response, int responselen, int *clen) {
	int headerlen = 0;
	int contentlen = 0;
	char *blankline = "\r\n\r\n";
	char *strP = strstr(response,blankline);
	int ix;
	if (NULL == strP) {
		blankline = "\n\n";
		strP = strstr(response,blankline);
		if (NULL == strP)
			return 0;
	}
	headerlen = strP-response;
	if ((NULL != (strP = strstr(response,"Content-length:"))) ||
		(NULL != (strP = strstr(response,"Content-Length:"))) ||
		(NULL != (strP = strstr(response,"content-length:")))) {
		strP += strlen("Content-length:");
		while (!isdigit(*strP))
			strP++;
		contentlen = atoi(strP);
	}
	else {
		char *hdr = "Content-length:";
		int hlen = strlen(hdr);
		for (ix = 0; ix < headerlen; ++ix) {
			if (0 == strncasecmp(hdr,response+ix,hlen)) {
				strP = response+ix+hlen;
				while (!isdigit(*strP))
					strP++;
				contentlen = atoi(strP);
				break;
			}
		}
	}
	*clen = contentlen;
	return (responselen >= headerlen+strlen(blankline)+contentlen);
}

/**
 * Free the response buffer allocated in 'aagtwyGetResponse' func
 */
void aagtwyFreeResponse (char *response) {
	if (NULL != response)
		free(response);
}
	
/**
 * Read HTTP response for the request send
 * Returns pointer to alloced buffer, call 'aagtwyFreeResponse' to release it.
 * Also supplies responselen and contentlen.
 */
char *aagtwyGetResponse (int sockfd, int *responselen, int *contentlen) {
	int respbufsize = 8192;
	char *response = (char *)malloc(respbufsize);
	int offset = 0;
	int clen = 0;
	while (1) {
		char rbuf[4096];
		int rbufsize = 4096;
		int rbytes = aagtwyRecv(sockfd,rbuf,rbufsize);
		if (rbytes < 0) {
			free(response);
			response = NULL;
			offset = 0;
			clen = 0;
			goto windup;
		}
		while (offset+rbytes+1 > respbufsize) {
			respbufsize += respbufsize;
			response = (char *)realloc(response,respbufsize);
		}
		memcpy(response+offset,rbuf,rbytes);
		offset += rbytes;
		response[offset] = 0;
		/* break out when we have full response buffer */
		if ((rbytes < rbufsize) && aagtwyIsResponseComplete(response,offset,&clen))
			break;
	}
  windup:
	if (NULL != responselen)
		*responselen = offset;
	if (NULL != contentlen)
		*contentlen = clen;
	return response;
}
