/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: javacomm.c
   Applix ASC version: 55
   Date: Thu Feb 25 21:12:36 1999
*/
#include <stdio.h>
#include <errno.h>
#if !M_88K && !PMAX
#include <sys/stat.h>
#endif
#include <axtime.h>
#if RIOS
#include <sys/select.h>
#endif
#include <aos.h>
#include <setjmp.h>
#include <thimble.h>
#include <java_syms.h>
#include <ssjsyms.h>
#include <elfjsyms.h>

#if AXMS
#include <stdlib.h>
#include <process.h>    /* for getpid() */
#include <io.h>         /* for open, read, etc... */
#include <fcntl.h>
#define open _open
#define read _read
#else
#include <sys/types.h>                  /* caddr_t */
#include <sys/file.h>
#include <fcntl.h>
#include <unistd.h>
#endif

#define UidFromJavaWinId(jwid) (jwid & 0xffff)

extern context_block *ActiveTasks[]; /* array of tasks .in_use */
extern char *ThimApplixDir;

jchannel_t *JavaChannels;	/* list of client channels */
int AxJavaShuttingDown = FALSE; /* If true, stop queuing messages */

/* Resource Lists:

At the top level, /applix/axdata/aarsrc/all.list, is a list of applications.
This is a simple ascii file of application names.  For each application name,
there must exist a subdirectory of the same name.

Within the application directory, there is a file "aarsrc.lst".  This simple
ascii file has a version number (0-64K) as its first line.  Subsequent
lines are the names of resources.  The resources are numerically mapped
to the java client based on their line number in this file +15. Thus,
the first named resource, which is the second line in this file, is mapped
to the client as resource #16.  The next line of this file is resource #17.
*/
typedef struct {
	char *name;
	int bitpos;
} appRsrcItem_t;

typedef struct {
	char *name;
	int version;
	int numMembers;
	appRsrcItem_t *members;	/* sorted array of members */
} appRsrcInfo_t;


appRsrcInfo_t *AppRsrcInfo;
int NumAppRsrcInfo;

#include <javacomm.p>

extern char *aagtwyGetResponse();

/*************************************************************************
function: AxJavaGetCurwinId - get current java window id for the given task
*/
int AxJavaGetCurwinId(taskid)
	int taskid;
{
	int pid = ThimPidFromUID(taskid);
	int jwid = taskid;
	if (AxIsSSTask(pid)) {
		return (SSJVGetCurwinId(pid));
	}
	return (jwid);
}

/*************************************************************************
function: AxJavaNotifier - Send message to task's java client
*/
void AxJavaNotifier (mainP,cmdCode,infoP)
    genericAppStruct *mainP;    /* any of wpP/ssP/geP */
    int cmdCode;    /* visible to the Java swicth */
    data_ptr infoP; /* the data, FREE'd by THIS routine */
{
    data_ptr msgP;
	jchannel_t *jc = javaChanFromAppP(mainP);

    if ( !mainP->javaClientId ) /* no connection established, just return */
        return;
    msgP = AxMakeArray(3);
    AxAddIntToArray(msgP,0,mainP->javaClientId);
    AxAddIntToArray(msgP,1,cmdCode);
    AxAddArrayToArray(msgP,2,infoP);
    QueueMsgToJavaClient(jc,0,msgP);
    AxFreeData(msgP);
}

/*************************************************************************
function: AxJavaUnQueuedNotifier - Send message to task's java client
*/
void AxJavaUnQueuedNotifier (mainP,cmdCode,infoP)
    genericAppStruct *mainP;    /* any of wpP/ssP/geP */
    int cmdCode;    /* visible to the Java swicth */
    data_ptr infoP; /* the data, FREE'd by THIS routine */
{
    data_ptr msgP;
	jchannel_t *jc = javaChanFromAppP(mainP);

    if ( !mainP->javaClientId ) /* no connection established, just return */
        return;
    msgP = AxMakeArray(3);
    AxAddIntToArray(msgP,0,mainP->javaClientId);
    AxAddIntToArray(msgP,1,cmdCode);
    AxAddArrayToArray(msgP,2,infoP);
    QueueMsgToJavaClient(jc,2,msgP);
    AxFreeData(msgP);
}

/*************************************************************************
function: AxJavaChannelNotifier - Send notifier via jchannel_t*
*/
void AxJavaChannelNotifier (jc,clientId,cmdCode,infoP)
jchannel_t *jc;
int clientId;
int cmdCode;    /* visible to the Java swicth */
data_ptr infoP; /* the data, FREE'd by THIS routine */
{
	data_ptr msgP;

	msgP = AxMakeArray(3);
	AxAddIntToArray(msgP,0,clientId);
	AxAddIntToArray(msgP,1,cmdCode);
	AxAddArrayToArray(msgP,2,infoP);
	QueueMsgToJavaClient(jc,0,msgP);
	AxFreeData(msgP);
}

void AxJavaPidNotifier (pid,cmdCode,infoP)  /* TEMP!!0 */
int pid;
int cmdCode;    /* visible to the Java switch */
data_ptr infoP; /* the data, FREE'd by THIS routine */
{
	data_ptr msgP;
	int jwId = AxJavaGetCurwinId(TaskInfo[pid].uid);
	jchannel_t *jc = javaChanFromUid(TaskInfo[pid].uid);

	msgP = AxMakeArray(3);
	AxAddIntToArray(msgP,0,jwId);
	AxAddIntToArray(msgP,1,cmdCode);
	AxAddArrayToArray(msgP,2,infoP);
	QueueMsgToJavaClient(jc,0,msgP);
	AxFreeData(msgP);
}
 
void AxJavaWinNotifier (jwId,cmdCode,infoP)
int jwId;
int cmdCode;    /* visible to the Java switch */
data_ptr infoP; /* the data, FREE'd by THIS routine */
{
	data_ptr msgP;
	jchannel_t *jc;
	
	jc = javaChanFromUid(UidFromJavaWinId(jwId));
	msgP = AxMakeArray(3);
	AxAddIntToArray(msgP,0,jwId);
	AxAddIntToArray(msgP,1,cmdCode);
	AxAddArrayToArray(msgP,2,infoP);
	QueueMsgToJavaClient(jc,0,msgP);
	AxFreeData(msgP);
}

/*************************************************************************
function: AxJavaCoreNotifier - Send message to ALL java client's IASMain
*/
void AxJavaCoreNotifier (cmdCode,infoP)
	int cmdCode;	/* visible to the Java swicth */
	data_ptr infoP; /* the data, FREE'd by THIS routine */
{
	data_ptr msgP;

	msgP = AxMakeArray(3);
	AxAddIntToArray(msgP,0,0);
	AxAddIntToArray(msgP,1,cmdCode);
	AxAddArrayToArray(msgP,2,infoP);
	QueueMsgToJavaClient(NULL,0,msgP);
	AxFreeData(msgP);
}

/*************************************************************************
function: AxJavaInitOrReloadNotify - Tell client about basic state stuff
This is called when a window is to be changed or File->Open...
*/
void AxJavaInitOrReloadNotify (mainP,name,menubar_id,moduleStartInd,
            width,height,moreP,cmdCode,isReload)
    genericAppStruct *mainP;    /* any of wpP/ssP/geP */
    char *name;
    int menubar_id,moduleStartInd,width,height;
    data_ptr moreP; /* !FREED!HERE! application specific state to xmit */
    int cmdCode, isReload;
{
    data_ptr msgP, infoP;
    int jwid = isReload ? mainP->javaClientId : 0;
	jchannel_t *jc = javaChanFromUid(UidFromJavaWinId(mainP->javaClientId));

    infoP = AxMakeArray(7);
    AxAddStrToArray(infoP,0,name);
    AxAddIntToArray(infoP,1,width);
    AxAddIntToArray(infoP,2,height);
    AxAddIntToArray(infoP,3,menubar_id);
    AxAddIntToArray(infoP,4,mainP->javaClientId);
    AxAddArrayToArray(infoP,5,moreP);
    AxAddIntToArray(infoP,6,moduleStartInd);
    baseNotify(jc,JVITX_NOTIFY,jwid,cmdCode,infoP);
    AxFreeData(infoP);
}

/*************************************************************************
function: AxJavaPostMessage - Post an info/error/statusline message
*/
void AxJavaPostMessage (jc,mainP,buf,is_status,error_mode)
jchannel_t *jc;
void *mainP;    /* any of wpP/ssP/geP */
char *buf;
int is_status, error_mode;
{
    genericAppStruct *p = (genericAppStruct *)mainP;    /* NOT USED YET!!0 */
    data_ptr msgP, infoP;
	int jwid = 0;

	if ( !jc && !mainP ) {	/* special case - 'wall' messages */
        int cmdCode = INFO_MSG;
        infoP = AxMakeArray(4);
        AxAddIntToArray(infoP,0,0);
        AxAddStrToArray(infoP,1,buf);
        AxAddIntToArray(infoP,2,is_status);
        AxAddIntToArray(infoP,3,error_mode);
		for ( jc = JavaChannels; jc; jc = jc->next )
	        baseNotify(jc,JVITX_NOTIFY,jwid,cmdCode,infoP);
        AxFreeData(infoP);
		return;
	}
	else {
/***	if ( is_status || !THIMpid || !buf[0] ) { ****/

        int cmdCode = INFO_MSG;
        int taskid = TaskInfo[THIMpid].taskMaster;
		if ( !jc )
			jc = javaChanFromUid(UidFromJavaWinId(p->javaClientId));
		if (is_status) {
			if (p != NULL)
				jwid = p->javaClientId;
			else
				jwid = AxJavaGetCurwinId(taskid);
		}
        infoP = AxMakeArray(4);
        AxAddIntToArray(infoP,0,taskid);
        AxAddStrToArray(infoP,1,buf);
        AxAddIntToArray(infoP,2,is_status);
        AxAddIntToArray(infoP,3,error_mode);
        baseNotify(jc,JVITX_NOTIFY,jwid,cmdCode,infoP);
        AxFreeData(infoP);
    }
/****
	else {
        msgP = AxMakeStrData(-1,buf);
        ElfbSetSysVar("JAVA$INFO$MSG$",msgP);
        AxFreeData(msgP);
        if ( error_mode )
            ElfbPendForNewTask("JAVA_POST_INFO_MSG$","error");
        else
            ElfbPendForNewTask("JAVA_POST_INFO_MSG$","");
    }
****/
    return;
}

/*************************************************************************
function: AxSetJavaDisplayInfo - Set linkage info in wp/ss/ge structs for java clients
This sets the javaClientId to the task's uid - not required, but a convenvient
value so that we can do set_macro_parent_task@() on messages from client.
Otherwise, we will need javaClientId/taskId mapping array for axjava.els.
SS- changes the javaClientId to (win_atmId<<16+uid).
*/
void AxSetJavaDisplayInfo (task, mainP)
int task;
genericAppStruct *mainP;    /* any of wpP/ssP/geP */
{
    if ( !TaskIsJavaBased(task) )
        mainP->javaClientId = 0;
    else
        mainP->javaClientId = TaskInfo[task].uid;
}

/*************************************************************************
function: TaskIsJavaBased - Returns whether the task was spawned for a java client
*/
int TaskIsJavaBased(pid)
int pid;
{
    return (TaskInfo[pid].in_use && TaskInfo[pid].TaskUsesJavaDisplay);
}


/*************************************************************************
function: JavaXmitMenuBar - Transmit menubar to java if not already there
*/
void JavaXmitMenuBar (num, uidP)
int num;
data_ptr uidP;
{
    int err;
    data_ptr menuP = NULL;
    int jwid = 0;
    int cmdCode = JV_NEW_MENUBAR;
	jchannel_t *jc = javaChanFromUid(TaskInfo[AxInTask].uid);

    err = ElfCallByName(NXLT("GET_JAVA_MENUBAR$"),&menuP,1,AxMakeIntData(num));
    if ( err == AOK && menuP )
        baseNotify(jc,JVITX_NOTIFY,jwid,cmdCode,menuP);
    AxFreeData(menuP);
}

/*
*   JavaSetTaskCursor - display/remove hour glass
*   
*/
void
JavaSetTaskCursor(jwid,cursor_type) 
    int jwid;
    int cursor_type;
{
    data_ptr msgP, infoP;
    int hourGlassMode;
	jchannel_t *jc;
	if (jwid) {
		 jc = javaChanFromUid(UidFromJavaWinId(jwid));
	 }
	else {
		if (cursor_type == AX_HOURGLASS_CURSOR) {
			/* hourglass all windows of the current task's client */
			jc = javaChanFromUid(TaskInfo[AxInTask].uid);
			if (cursor_type == AX_HOURGLASS_CURSOR && jc->showingHourglass)
				return;
			jc->showingHourglass = TRUE;
		}
		else { /* remove hourglass for all clients */
			msgP = AxMakeArray(3);
			AxAddIntToArray(msgP,0,0);
			AxAddIntToArray(msgP,1,SET_ACTIVE_CURSOR);
			AxAddIntToArray(msgP,2,cursor_type);
			for (jc=JavaChannels; jc != NULL; jc = jc->next) {
				if (jc->showingHourglass) {
					jc->showingHourglass = FALSE;
					QueueMsgToJavaClient(jc,0,msgP);
				}
			}
			AxFreeData(msgP);
			return;
		}
    }
    msgP = AxMakeArray(3);
    AxAddIntToArray(msgP,0,jwid);
    AxAddIntToArray(msgP,1,SET_ACTIVE_CURSOR);
    AxAddIntToArray(msgP,2,cursor_type);
    QueueMsgToJavaClient(jc,0,msgP);
    AxFreeData(msgP);
}

/*
 *  JavaSetMenuItem - set menuItem name and status for the task's main Window
 *
 */
void
JVSetMenuItem(mainP,macroName,itemName, itemStatus, bubblehelp)
    genericAppStruct* mainP;
    char* macroName;
    char* itemName;
    int itemStatus;
    char* bubblehelp;
{
    data_ptr msgP,infoP;
	int macroid;
	jchannel_t *jc = javaChanFromAppP(mainP);
    infoP = AxMakeArray(4);
	macroid = ElfLookupSymbol(macroName,1);
	if (macroid != 0)
		AxAddIntToArray(infoP,0,macroid);
	else
		AxAddStrToArray(infoP,0,macroName);
    AxAddStrToArray(infoP,1,itemName);
    AxAddIntToArray(infoP,2,itemStatus);
	if ( bubblehelp != (char *)0)
		AxAddStrToArray(infoP, 3, bubblehelp);
    msgP = AxMakeArray(3);
    AxAddIntToArray(msgP,0,mainP->javaClientId);
    AxAddIntToArray(msgP,1,JV_SET_MENUITEM);
    AxAddArrayToArray(msgP,2,infoP);
    QueueMsgToJavaClient(jc,0,msgP);
    AxFreeData(msgP);
}   

typedef struct java_dlg_item {
	struct java_dlg_item *next;
	int uid;
	char *title;
	int parentTask;
	int icon;
	int clientId;	/* which connected client (needed for parentTask==0) */
} java_dlg_item;
java_dlg_item *JavaDlgList;

/*************************************************************************
function: JVAddDlgToList - Note new dialog into 'promote-able' list
Note that passed uid is both winId and task uid.
*/
void JVAddDlgToList (int uid, char *title, int parentTask, int icon)
{
	java_dlg_item *p;

	JVRemoveDlgFromList(uid);/* paranoia... task can have only one dlg */
	p = (java_dlg_item*)TaskAlloc(0,sizeof(java_dlg_item));
	p->uid = uid;
	p->title = TaskAlloc(0,strlen(title)+1);
	p->icon = icon;
	strcpy(p->title,title);
	p->parentTask = parentTask;
	p->clientId = AxJavaClientId(AxInTask);
	p->next = JavaDlgList;
	JavaDlgList = p;
}

/*************************************************************************
function: JVRemoveDlgFromList - Remove task's dialog
*/
void JVRemoveDlgFromList (int uid)
{
	java_dlg_item *p, *prev = NULL;

	for ( p = JavaDlgList; p; prev = p, p = p->next ) {
		if ( p->uid == uid ) {
			if ( !prev )
				JavaDlgList = p->next;
			else
				prev->next = p->next;
			TaskFree(0,p->title);
			TaskFree(0,p);
			return;
		}
	}
}

void JVDialogsList(dlist, numDlgs)
	data_ptr *dlist;
	int *numDlgs;
{
	java_dlg_item *p;
	register data_ptr element;
		
	for ( p = JavaDlgList; p; p = p->next ) {
		element = MkArray(4);
		AxAddIntToArray(element, 0, p->uid);
		AxAddStrToArray(element, 1, p->title);
		AxAddIntToArray(element, 2, p->icon);
		AxAddIntToArray(element, 3, p->uid);
		*dlist = AxAddArrayToArray(*dlist, *numDlgs, element);
		(*numDlgs)++;
	}
}

/*************************************************************************
function: JVPromoteDlg - Promote to top (or just return existance)
*/
int JVPromoteDlg (clientId, name, parentTask, no_promote)
int clientId;
char *name; /* name of dialog box */
int parentTask; /* UID of task this is slaved to */
int no_promote;
{
	java_dlg_item *p, *prev = NULL;

	for ( p = JavaDlgList; p; p = p->next ) {
		if ( (!parentTask || (p->parentTask == parentTask))
			&& p->clientId == clientId
			&& streq(p->title,name) ) {
			if (!no_promote) /* promote the dlg to top */
				AxJavaPidNotifier(ThimPidFromUID(parentTask),DB_PROMOTE,
								  AxMakeIntData(p->uid));
			return(BTRUE);
		}
	}
	return(BFALSE);
}

int JVSelectWinByWinUid(uid, grab)
    int uid; /* window uid */
    int grab;
{
	java_dlg_item *p, *prev = NULL;

	for ( p = JavaDlgList; p; p = p->next ) {
		if (p->uid == uid) {
			if (grab) {
				AxJavaPidNotifier(ThimPidFromUID(p->parentTask), DB_PROMOTE,
								  AxMakeIntData(p->uid));
			}
			AxSetTaskMaster(THIMpid, p->parentTask);
			return(TRUE);
		}
	}
	return(FALSE);
}

/*************************************************************************
function: KillJavaWindows - Close any windows associated with this task
Errors sending message assumed harmless
*/
void KillJavaWindows (pid)
int pid;
{
    data_ptr channelP, response;
    int taskuid = TaskInfo[pid].uid;
	jchannel_t *jc = javaChanFromUid(taskuid);

	/* in case window was not explicitly closed with close_dialog$ */
	JVRemoveDlgFromList(taskuid);
	
    response = AxMakeArray(3);
    AxAddIntToArray(response,0,0);
    AxAddIntToArray(response,1,DB_CLOSE);
/* NOTE: here taskuid is sent instead of java win id, as at this point
   windows list is already cleared 
   client closes all java windows belonging to this taskuid */
    AxAddIntToArray(response,2,taskuid); 
    QueueMsgToJavaClient(jc,0,response);
    AxFreeData(response);
}

#define JV_RSRC_IMAGE 1

/*************************************************************************
function: AxfPostJavaImage - Verify named image on client
macro post_java_image$(name)
    var sysvarname

    sysvarname = "jv$im$"++name
    if not system_var@(sysvarname) {
        set_system_var@(sysvarname,true)
        java_notify$({JV#NEW#IMAGE,{name,get_java_image_packet$(name)}})
    }
endmacro
*/
data_ptr AxfPostJavaImage(data_ptr argDatum) {
    char *name = AxStrFromArray(argDatum,0);
    char buf[1000];
    data_ptr aP;
    int jwid = 0;
	jchannel_t *jc = javaChanFromUid(TaskInfo[AxInTask].uid);
	int uid;
	int xmitted;
	int rsrcId;
	static int appId = -1;

	if ( !jc )
		return;

	if ( appId == -1 )
		appId = AxJavaAppDexFromName("aximg");
	rsrcId = AxJavaVerifyClientResource(jc,appId,name,JV_RSRC_IMAGE,&xmitted);
	if ( rsrcId != 0 ) {
		if ( !xmitted ) {	/* image was there, let client know name vs. id */
		    sprintf(buf,"jv$im$%d$",jc->uid);
		    strcat(buf,name);
		    if ( !ElfGetSysVar(buf) ) {
		        ElfbSetSysVar(buf,AxMakeIntData(-1));
		        aP = AxMakeArray(3);
		        AxAddStrToArray(aP,0,name);
		        AxAddIntToArray(aP,1,rsrcId);
		        AxAddStrToArray(aP,2,"aximg");
		        baseNotify(jc,JVITX_NOTIFY,jwid,JV_NEW_IMAGE,aP);
		        AxFreeData(aP);
			}
		}
	}
	else {	/* not a standard resource, use old scheme */
/*		uid = (jc != NULL) ? jc->uid : -1;   made no sense - zane removed */
	    sprintf(buf,"jv$im$%d$",jc->uid);
	    strcat(buf,name);
	    if ( !ElfGetSysVar(buf) ) {
#if 0
			printf("Posting non-resource image, %s, to java client.\n",name);
#endif
	        ElfbSetSysVar(buf,AxMakeIntData(-1));
	        aP = AxMakeArray(2);
	        AxAddStrToArray(aP,0,name);
	        AxAddDataToArray(aP,1,AxGetJavaImagePacket(name));
	        baseNotify(jc,JVITX_NOTIFY,jwid,JV_NEW_IMAGE,aP);
	        AxFreeData(aP);
	    }
	}
    return(NULL);
}

/*************************************************************************
function: AxfPostJavaImagePacket - Send resolved data to client
macro post_java_image_packet$(name, packet)
    java_notify$({JV#NEW#IMAGE,{name,packet}})
endmacro
*/
data_ptr AxfPostJavaImagePacket(data_ptr argDatum) {
    int jwid = 0;
	jchannel_t *jc = javaChanFromUid(TaskInfo[AxInTask].uid);
    baseNotify(jc,JVITX_NOTIFY,jwid,JV_NEW_IMAGE,argDatum);
    return(NULL);
}

/*************************************************************************
function: AxfJavaMenubarNotify - Send menubar to client
macro java_menubar_notify$(bardata)
    send_message@(java_sender_task$,JVITX#NOTIFY,{0,JV#NEW#MENUBAR,bardata})
endmacro
*/
data_ptr AxfJavaMenubarNotify(data_ptr argDatum) {
    data_ptr cmdData = AxArrayElement(argDatum,0);
    int jwid = 0;
    int cmdCode = JV_NEW_MENUBAR;
	jchannel_t *jc = javaChanFromUid(TaskInfo[AxInTask].uid);
    baseNotify(jc,JVITX_NOTIFY,jwid,cmdCode,cmdData);
    return(NULL);
}

/*************************************************************************
function: AxfJavaTaskSyncMesNotify - Respond to function call
macro java_task_sync_mes_notify$(info)
    send_message@(java_sender_task$,JVITX#RESPONSE,info)
endmacro
*/
data_ptr AxfJavaTaskSyncMesNotify(data_ptr argDatum) {
    data_ptr infoP = AxArrayElement(argDatum,0);
	jchannel_t *jc = javaChanFromUid(TaskInfo[AxInTask].uid);

    QueueMsgToJavaClient(jc,1,infoP);
    return(NULL);
}
void AxfJavaTaskSyncMesAcknowledge() {
    data_ptr infoP = AxMakeArray(2);
	jchannel_t *jc = javaChanFromUid(TaskInfo[AxInTask].uid);

    AxAddIntToArray(infoP,0,0);
    QueueMsgToJavaClient(jc,1,infoP);
    AxFreeData(infoP);
}

/*************************************************************************
function: AxfJavaTaskNotify - Send message from task to client part
macro java_task_notify$(info)
    send_message@(java_sender_task$,JVITX#NOTIFY,info)
endmacro
*/
data_ptr AxfJavaTaskNotify(data_ptr argDatum) {
    data_ptr infoP = AxArrayElement(argDatum,0);
    int jwid = AxIntFromArray(infoP,0);
    int cmdCode = AxIntFromArray(infoP,1);
	int taskid = UidFromJavaWinId(jwid);
    data_ptr cmdData = AxArrayElement(infoP,2);
	jchannel_t *jc = javaChanFromUid(taskid);
    baseNotify(jc,JVITX_NOTIFY,jwid,cmdCode,cmdData);
    return(NULL);
}

/*************************************************************************
function: AxfJavaNotify - Send message to core client code
macro java_notify$(info)
    send_message@(java_sender_task$,JVITX#NOTIFY,{0,info[0],info[1]})
endmacro
*/
data_ptr AxfJavaNotify(data_ptr argDatum) {
    data_ptr infoP = AxArrayElement(argDatum,0);
    int jwid = 0;
    int cmdCode = AxIntFromArray(infoP,0);
    data_ptr cmdData = AxArrayElement(infoP,1);
	jchannel_t *jc = javaChanFromUid(TaskInfo[AxInTask].uid);
    baseNotify(jc,JVITX_NOTIFY,jwid,cmdCode,cmdData);
    return(NULL);
}

/*************************************************************************
function: baseNotify - Send message to java client window
*/
static
void baseNotify (jc,style,jwid,cmdCode,cmdData)
	jchannel_t *jc;	/* which channel to send message on */
	int jwid;		/*  client window id */
    int cmdCode;    /* visible to the Java swicth */
    data_ptr cmdData;   /* the data, FREE'd by THIS routine */
{
    data_ptr msgP;
    msgP = AxMakeArray(3);
    AxAddIntToArray(msgP,0,jwid);
    AxAddIntToArray(msgP,1,cmdCode);
    AxAddArrayToArray(msgP,2,cmdData);
    QueueMsgToJavaClient(jc,0,msgP);
    AxFreeArrayButNotContents(msgP);
}

/*************************************************************************
function: QueueMsgToJavaClient - Apply msg to sender queue
*/
int QueueMsgToJavaClient(jc, type, aP)
jchannel_t *jc;
int type;   /* 0-notify, 1-response, 2-unqueued_notify 3-background */
data_ptr aP;
{
    int retval = FALSE;
    data_ptr rP;        /* sent back to Aster*x */
    char *outBuf;   /* holds the message on the way out */
    int32 cnt;      /* counter */
    data_ptr infoP,msgP;
    jmsg_q *q;
    int qs;
    int putItNowIfPossible = FALSE;

    if ( AxJavaShuttingDown )   /* Don't Q more while quitting! */
        return(FALSE);

	if ( !jc ) {	/* if no channel - ALL channels */
		for ( jc = JavaChannels; jc; jc = jc->next )
			QueueMsgToJavaClient(jc, type, aP);
		return(TRUE);
	}

	DumpOutgoingMessages(aP);

    /* Overpack due to current client expectations */
    infoP = AxMakeArray(2);
    AxAddIntToArray(infoP,0,type);
    AxAddArrayToArray(infoP,1,aP);
    /* Overpack due to current client expectations */
    msgP = AxMakeArray(2);
    AxAddIntToArray(msgP,0,0);
    AxAddArrayToArray(msgP,1,infoP);
    q = (jmsg_q*)TaskAlloc(0,sizeof(jmsg_q));
    q->size = AxMemWritePacked(msgP,0,&outBuf,TRUE);
    AxReallocToAnotherTask(outBuf, 0);
    q->outBuf = outBuf;
    q->next = NULL;
    if ( !jc->FirstQMsg ) {
        jc->FirstQMsg = q;
        jc->LastQMsg = q;
    }
    else {
        jc->LastQMsg->next = q;
        jc->LastQMsg = q;
    }
    qs = AxJavaOutputBytesQueued(jc); /* note qsize before this msg */
    if ( type == 1 || type == 0 || type == 2 )
        jc->BytesInMainQ += q->size;
    else
        jc->BytesInLowPriQ += q->size;

	jc->msgs_written++;
	jc->min_msg_written = MIN(jc->min_msg_written,q->size);
	jc->max_msg_written = MAX(jc->max_msg_written,q->size);

    AxFreeArrayButNotContents(infoP);
    AxFreeArrayButNotContents(msgP);
    if ( qs/1024 != (qs-q->size)/1024 )
        AxJavaWriteSomeStuff(-1,jc); /* attempt to write on crossing 1K bounds */
    return(TRUE);
}


AxJavaOutputBytesQueued(jchannel_t *jc) {
	int ix, total = 0;
	if ( !jc ) {
	    for ( ix = 0; ActiveTasks[ix]; ix++ ) {
			if ( ActiveTasks[ix]->TaskUsesJavaDisplay
				&& ActiveTasks[ix]->javaChannel )
				total += AxJavaOutputBytesQueued(ActiveTasks[ix]->javaChannel);
		}
		return(total);
	}
    return(jc->BytesInMainQ+jc->BytesInLowPriQ+jc->BytesInBuffer);
}

static int FillBufFromQueue(jchannel_t *jc, char *ioBuf,int curSize, int bufSize) {
    int availBytes, cnt;
    jmsg_q *q;

    while ( TRUE ) {
        availBytes = bufSize - curSize;
        if ( availBytes == 0 || !jc->FirstQMsg )
            return(curSize);
        q = jc->FirstQMsg;
        cnt = q->size - jc->QPos;
        if ( cnt > availBytes ) {   /* cannot de-queue */
            memcpy(ioBuf+curSize,q->outBuf+jc->QPos,availBytes);
            curSize += availBytes;
            jc->QPos += availBytes;
            jc->BytesInMainQ -= availBytes;
        }
        else {                      /* de-queue and go for next one */
            memcpy(ioBuf+curSize,q->outBuf+jc->QPos,cnt);
            curSize += cnt;
            jc->BytesInMainQ -= cnt;
            jc->QPos = 0;
            if ( jc->LastQMsg == q ) {
                jc->LastQMsg = NULL;
                jc->FirstQMsg = NULL;
            }
            else
                jc->FirstQMsg = q->next;
            TaskFree(0,q->outBuf);
            TaskFree(0,q);
        }
    }
}

void AxJavaWriteSomeStuff(int events, jchannel_t *jc) {
    int cnt, did;
    jmsg_q *q;
    static fd_set read_fds, wtfds, nofds;
    static struct timeval timeout;
    struct timeval *timeP = &timeout;
    int go_for_more = FALSE;

while ( TRUE ) {
    if ( events == -1       /* possible channel error */
        || go_for_more ){   /* have drained buf, doing more */
        FD_ZERO( &wtfds );
        FD_SET(jc->fd,&wtfds);
        events = select(FD_SETSIZE, NULL, &wtfds, &nofds, timeP);
        if ( events == -1 ) {
            TossJavaSenderFd(jc,TRUE);
			return;
        }
        if ( !FD_ISSET(jc->fd,&wtfds) )   /* not possible */
            return;
    }

    jc->bufSize = FillBufFromQueue(jc,jc->ioBuf,jc->bufSize,IOBUFSIZE);
    cnt = jc->bufSize - jc->bufPos;
    jc->BytesInBuffer = cnt;
    if ( cnt == 0 )
        return;

#if AX_X11
    did = write(jc->fd,jc->ioBuf+jc->bufPos,cnt);
    if ( -1 == did ) {
        if ( errno == EWOULDBLOCK || errno == EINTR || errno == EAGAIN )
            return;
        TossJavaSenderFd(jc,TRUE);
		return;
    }
#else
    /*  Windows-NT */
    did = send(jc->fd,jc->ioBuf+jc->bufPos,cnt,0);
    if ( -1 == did ) {
        DWORD err = WSAGetLastError();
        if ( err == WSAEWOULDBLOCK || err == WSAEINTR )
            return;
        TossJavaSenderFd(jc,TRUE);
		return;
    }
#endif
    else if ( did == 0 ) {      /* Pipe must be broken */
#if AX_X11
        close(jc->fd);
#else
        closesocket(jc->fd);
#endif
        TossJavaSenderFd(jc,TRUE);
		return;
    }

	jc->bytes_written += did;

    if ( did < cnt ) {
        jc->bufPos += did;
        jc->BytesInBuffer -= did;
        return;
    }
    else {
        jc->bufPos = 0;
        jc->bufSize = 0;
        jc->BytesInBuffer = 0;
        go_for_more = TRUE;
    }
}
}

/*************************************************************************
function: TossJavaSenderFd - Remove channel from JavaChannels, May LOGOUT.
If slaved axmain (this channel has logoutOnDisconnect) and this function
is called in errorMode, this shuts down the axmain.
*/
void TossJavaSenderFd(jchannel_t *jc, int errorMode) {
	jchannel_t *p;
    jmsg_q *q, *nextQ;
	int ix,clients;
	int is_shared_server;

	is_shared_server = AxIntFromDataPtr(ElfGetSysVar("JAVA_SHARED_SERVER$"));
	if ( errorMode && jc->logoutOnDisconnect && !is_shared_server )
		AxExit(0);

	if ( JavaChannels == jc )
		JavaChannels = jc->next;
	else {
		for ( p = JavaChannels; p->next != jc; p = p->next );
		p->next = jc->next;
	}
	for ( q = jc->FirstQMsg; q; q = nextQ ) {
		nextQ = q->next;
		TaskFree(0,q);
	}

	AxSocketClose(jc->uid);

	for ( ix = 0; ActiveTasks[ix]; ix++ ) {
		if ( jc == ActiveTasks[ix]->javaChannel ) {
			ActiveTasks[ix]->javaChannel = 0;
			KillTask(ThimPidFromUID(ActiveTasks[ix]->uid));
		}
	}

	free(jc->username);
	if ( jc->rsrcCacheMap ) {
		for ( ix = 0; ix < NumAppRsrcInfo; ix++ ) {
			if ( jc->rsrcCacheMap[ix] )
				free(jc->rsrcCacheMap[ix]);
		}
		free(jc->rsrcCacheMap);
	}
	free(jc);

	/* If running in shared_server mode, notify aagtwy a client went away */
	if (is_shared_server) {
		for ( clients = 0, p = JavaChannels; p; p = p->next )
			clients++;
		/* the next line safe because error throw will never happen here */
		ElfbRPCRespond(ElfGetSysVar("AA_GTWY_CHANNEL$"),AxMakeIntData(clients));
	}
}

jchannel_t *
javaChanFromAppP(mainP)
    genericAppStruct *mainP;    /* any of wpP/ssP/geP */
{
	return(javaChanFromUid(UidFromJavaWinId(mainP->javaClientId)));	/* !!0 TEMP HACK!!!! */
}


jchannel_t *javaChanFromPid(int pid) {
	jchannel_t *jc = TaskInfo[pid].javaChannel;	/* !!0 TEMP HACK!!!! */
	if ( !jc ) jc = findAnyJavaChannel();	/* HACK paranoia! */
	return(jc);
}

jchannel_t *javaChanFromUid(int uid) {
	int pid = ThimPidFromUID(uid);
	if (pid == 0 && uid != 0) { 
        /* Error: probably java winid passed, instead of uid */
		char buf[STRPATH+1];
        sprintf(buf,
                NXLT("Invalid task-uid %d, in javaChanFromUid\n"),uid);
		tempAbort(buf);
	}
	return(javaChanFromPid(pid));
}

jchannel_t *findAnyJavaChannel() {
	int ix;
    for ( ix = 0; ActiveTasks[ix]; ix++ ) {
		if ( ActiveTasks[ix]->TaskUsesJavaDisplay && ActiveTasks[ix]->javaChannel )
			return(ActiveTasks[ix]->javaChannel);
	}
	return(NULL);
}

/*************************************************************************
function: AxJavaNoteClientChannel - Register client against this task
All tasks that are children of this task will inherit this channel.
*/
void AxJavaNoteClientChannel(int pid, int uid, char *username) {
	jchannel_t *jc = (jchannel_t*)calloc(1,sizeof(jchannel_t));
	jc->uid = uid;
	jc->fd = AxSockFdFromUid(uid);
	jc->next = JavaChannels;
	/* using strcpy/malloc instead of strdup to guarantee that we
	 * use the Applix memory allocator.  -donato
	 */
	jc->username = strcpy(malloc(strlen(username)+1), username);
	JavaChannels = jc;
	TaskInfo[pid].javaChannel = jc;
}

/*************************************************************************
function: AxJavaClientShut - Called when elf layer's read_socket fails
Sicne the shutdown sequence may get called during client side shutdown,
we are paranoid and always de-reference TaskInfo[AxInTask].javaChannel.
*/
void AxJavaClientShut() {
	jchannel_t *jc = TaskInfo[AxInTask].javaChannel;
	data_ptr msgP, infoP;
	long old_time;

	if ( !jc )
		return;
	AxJavaChannelNotifier(jc,0,JV_SHUTDOWN_NOTIFY,NULL);
	AxJavaWriteSomeStuff(-1,jc); /* see if fd 'alive' - attempt to write */
	jc = TaskInfo[AxInTask].javaChannel;
	if ( !jc )		/* fd gone, now shutdown to send */
		return;
	if ( AxJavaOutputBytesQueued(jc) ) { /* shutdown nor sent keep trying */
		AxSleepMilliseconds(0);	/* with luck, message will go out now */
		old_time = time(0L);
		while ( TRUE ) {
			jc = TaskInfo[AxInTask].javaChannel;
			if ( !jc || !AxJavaOutputBytesQueued(jc) )
				break;
			AxSleepMilliseconds(500);	/* wait till it gets out */
			if ( time(0L) > old_time+30 )	/* regardless, never wait more than 30 secs */
				return;
		}
	}
	jc = TaskInfo[AxInTask].javaChannel;
	if ( jc )
		TossJavaSenderFd(jc, TRUE);
}

/*************************************************************************
function: AxJavaClientId - Return connection client channel id
*/
int AxJavaClientId (int pid)
{
	return(TaskInfo[pid].javaChannel ? TaskInfo[pid].javaChannel->uid : 0);
}

/*************************************************************************
function: AxJavaChannelStats - IO channle static list for perf evals
*/
data_ptr AxJavaChannelStats ()
{
	jchannel_t *jc;
	data_ptr rData, rec;
	int dex = 0;

	rData = AxMakeArray(0);
	for ( jc = JavaChannels; jc; jc = jc->next ) {
		rec = AxMakeArray(9);
		AxAddStrToArray(rec,0,jc->username);
		AxAddIntToArray(rec,1,jc->bytes_written);
		AxAddIntToArray(rec,2,jc->msgs_written);
		AxAddIntToArray(rec,3,jc->min_msg_written);
		AxAddIntToArray(rec,4,jc->max_msg_written);
		AxAddIntToArray(rec,5,jc->bytes_read);
		AxAddIntToArray(rec,6,jc->msgs_read);
		AxAddIntToArray(rec,7,jc->min_msg_read);
		AxAddIntToArray(rec,8,jc->max_msg_read);
		rData = AxAddArrayToArray(rData,dex++,rec);
	}
	return(rData);
}

/*************************************************************************
function: AxJavaChannelStatsReset - Reset counters
*/
void AxJavaChannelStatsReset ()
{
	jchannel_t *jc;

	for ( jc = JavaChannels; jc; jc = jc->next ) {
		jc->bytes_written = 0;
		jc->msgs_written = 0;
		jc->min_msg_written = 0;
		jc->max_msg_written = 0;
		jc->bytes_read = 0;
		jc->msgs_read = 0;
		jc->min_msg_read = 0;
		jc->max_msg_read = 0;
	}
}

int AADumpCounter = 0;	/* if >0, dumping messages */

/*************************************************************************
function: DumpOutgoingMessages - Put out both human & machine readable message logs
writes aaouth.123 and aaoutm.123
*/
void DumpOutgoingMessages (data_ptr aP)
{
	FILE *fp;
	char filename[STRPATH+1];
	data_ptr np;

	if ( !AADumpCounter )
		return;

	sprintf(filename,"/tmp/aamsgh.%03d",AADumpCounter);
	fp = fopen(filename,"w");
	AxDumpDatum((elfData)aP,fp);
	fclose(fp);
	sprintf(filename,"/tmp/aamsgm.%03d",AADumpCounter);
	np = AxMakeStrData(-1,filename);
	ElfbWriteDataFile(np, aP, NULL);
	AxFreeData(np);
	AADumpCounter++;
}

/*************************************************************************
function: SetAADumpCounter - Started dumping messages (from specied count value)
*/
void SetAADumpCounter (int val)
{
	if ( val < 0 )
		val = 1;	/* sanity */
	AADumpCounter = val;
}

/*************************************************************************
function: JVAddRightMouseItem
*/
data_ptr JVAddRightMouseItem(mbchar *name, mbchar *macro, int state)
{
	data_ptr aP;
	aP = AxMakeArray(4);
	AxAddStrToArray(aP, 0, name);
	AxAddStrToArray(aP, 1, macro);
	AxAddStrToArray(aP, 2, ""); /* args to macro */
	AxAddIntToArray(aP, 3, state);
	return(aP);
}

/*************************************************************************
function: JVGetMappedFontName
*/
char *JVGetMappedFontName(char *orig_name) {
	if (!strcmp(orig_name, NXLT("Avant Garde")) ||
		!strcmp(orig_name, NXLT("Helvetica")) ||
		!strcmp(orig_name, NXLT("Helvetica Narrow")) ||
		!strcmp(orig_name, NXLT("AvantGarde")) ||
		!strcmp(orig_name, NXLT("Helvetica-Narrow")) ||
		!strcmp(orig_name, NXLT("Alburtus")) ||
		!strcmp(orig_name, NXLT("Antique Olive")) ||
		!strcmp(orig_name, NXLT("Arial")) ||
		!strcmp(orig_name, NXLT("CG Omega")) ||
		!strcmp(orig_name, NXLT("Snowcap")) ||
		!strcmp(orig_name, NXLT("Swiss Narrow")) ||
		!strcmp(orig_name, NXLT("Swiss")) ||
		!strcmp(orig_name, NXLT("Univers Condensed")) ||
		!strcmp(orig_name, NXLT("Univers")))
		return(NXLT("Helvetica"));

	else if (!strcmp(orig_name, NXLT("Courier")) ||
			 !strcmp(orig_name, NXLT("Courier New")) ||
			 !strcmp(orig_name, NXLT("Courier Swa")))
		return(NXLT("Courier"));

	else if (!strcmp(orig_name, NXLT("Symbol")) ||
			 !strcmp(orig_name, NXLT("Zapf Dingbats")) ||
			 !strcmp(orig_name, NXLT("ZapfDingbats")) ||
			 !strcmp(orig_name, NXLT("Holidaypi")) ||
			 !strcmp(orig_name, NXLT("Morewingbats")) ||
			 !strcmp(orig_name, NXLT("Symbol Swa")) ||
			 !strcmp(orig_name, NXLT("Wingdings")))
		return(NXLT("Symbol"));

	else if (!strcmp(orig_name, NXLT("Bookman")) ||
			 !strcmp(orig_name, NXLT("Century Schoolbook")) ||
			 !strcmp(orig_name, NXLT("New Century Schoolbook")) ||
			 !strcmp(orig_name, NXLT("Palatino")) ||
			 !strcmp(orig_name, NXLT("Times")) ||
			 !strcmp(orig_name, NXLT("Zapf Chancery")) ||
			 !strcmp(orig_name, NXLT("Century Schoolbook")) ||
			 !strcmp(orig_name, NXLT("NewCenturySchlbk")) ||
			 !strcmp(orig_name, NXLT("ZapfChancery")) ||
			 !strcmp(orig_name, NXLT("CG Times")) ||
			 !strcmp(orig_name, NXLT("Clarendon Condensed")) ||
			 !strcmp(orig_name, NXLT("Cloisterblack")) ||
			 !strcmp(orig_name, NXLT("Coronet")) ||
			 !strcmp(orig_name, NXLT("Dutch")) ||
			 !strcmp(orig_name, NXLT("Freeform")) ||
			 !strcmp(orig_name, NXLT("Garamond")) ||
			 !strcmp(orig_name, NXLT("ITC Bookman")) ||
			 !strcmp(orig_name, NXLT("Letter Gothic")) ||
			 !strcmp(orig_name, NXLT("Marigold")) ||
			 !strcmp(orig_name, NXLT("Shelley")) ||
			 !strcmp(orig_name, NXLT("Times New Roman")) ||
			 !strcmp(orig_name, NXLT("Times Roman")) ||
			 !strcmp(orig_name, NXLT("Zapf Calligraphic")))
		return(NXLT("TimesRoman"));

	else if (!strcmp(orig_name, NXLT("Lucida Console")))
		return(NXLT("Dialog"));

	else
		return(NXLT("Dialog"));
}

/*************************************************************************
function: AxJavaVerifyClientResource - Update client if needed
Returns rsrcId as known on the client.

THIS MUST BE CALLED IN AN ELF CONTEXT - CAN THROW ELF ERRORS

*/
int AxJavaVerifyClientResource (
	jchannel_t *jc,		/* channel info */
	int appId,		/* "axss"... */
	char *rsrcName,	/* name in table /applix/axdata/aarsrc/lang/appName/appName */
	int rsrcType,	/* some resources xmit differently... */
	int *xmittedNow)	/* set true if sent up by this call */
{
	appRsrcInfo_t *app;
	appRsrcItem_t key, *rsrcInfo;
	int rsrcId, dex, mask;
	data_ptr aP;
	char fname[STRPATH+1];
	int fileSize;
	char *fileBuf;

	*xmittedNow = FALSE;
return(0);	/* just for now!!!, still debugging, but want to get sources IN */
	if ( !AppRsrcInfo )
		loadAppRsrcInfo();
	if ( (appId < 0) || (appId >= NumAppRsrcInfo) )
		return(0);
	app = &AppRsrcInfo[appId];
	if ( !app->members )
		loadAppRsrcTable(appId);
	key.name = rsrcName;
	rsrcInfo = (appRsrcItem_t*)bsearch((char*)&key,(char*)app->members,
				app->numMembers,sizeof(appRsrcItem_t),cmpRsrc);
	if ( !rsrcInfo )
		return(0);
	rsrcId = rsrcInfo->bitpos;
	dex = rsrcId / 8;
	mask = 1 << (rsrcId % 8);
	if ( !jc->rsrcCacheMap )
		jc->rsrcCacheMap = (char**)TaskCalloc(0,NumAppRsrcInfo,sizeof(char*));
	if ( !jc->rsrcCacheMap[appId] )
		jc->rsrcCacheMap[appId] = (char*)TaskCalloc(0,1,(app->numMembers+15+7)/8);
	if ( !(jc->rsrcCacheMap[appId][dex] & mask) ) {
		jc->rsrcCacheMap[appId][dex] |= mask;
		if ( rsrcType == JV_RSRC_IMAGE ) {
printf("Downloading rsrcId %d image %s\n",rsrcId,rsrcName);
	        aP = AxMakeArray(4);
	        AxAddStrToArray(aP,0,rsrcName);
	        AxAddDataToArray(aP,1,AxGetJavaImagePacket(rsrcName));
			AxAddStrToArray(aP,2,AppRsrcInfo[appId].name);
			AxAddIntToArray(aP,3,rsrcId);
	        baseNotify(jc,JVITX_NOTIFY,0,JV_NEW_IMAGE,aP);
	        AxFreeData(aP);
		}
		else {
printf("Downloading rsrcId %d\n",rsrcId);
	        aP = AxMakeArray(4);
			AxAddStrToArray(aP,0,AppRsrcInfo[appId].name);
			AxAddIntToArray(aP,1,rsrcId);
			sprintf(fname,"%s/aarsrc/eng/%s/%s",ThimApplixDir,app->name,rsrcName);
			fileBuf = AxReadFileIntoTaskBuf(0,fname,&fileSize);
			if ( !fileBuf )
				ElfStrAbort(1,"Could not open file",fname);
	        AxAddDataToArray(aP,2,AxMakeBinaryData(fileSize,fileBuf));
			AxAddIntToArray(aP,3,TRUE);	/* memory cache */
	        baseNotify(jc,JVITX_NOTIFY,0,DB_CACHE_ADD,aP);
	        AxFreeData(aP);
			TaskFree(0,fileBuf);
		}
		*xmittedNow = TRUE;
	}
	return(rsrcId);
}

static data_ptr loadAppRsrcTable(int appId) {
	int fileSize;
	char *fileBuf;
	char filename[STRPATH+1];
	int ix;
	char *p, *start;
	appRsrcInfo_t *app = &AppRsrcInfo[appId];

	sprintf(filename,"%s/aarsrc/eng/%s/%s.lst",
			ThimApplixDir,app->name,app->name);
	fileBuf = AxReadFileIntoTaskBuf(0,filename,&fileSize);
	if ( !fileBuf )
		ElfStrAbort(1,"Could not open file",filename);
	for ( ix = 0, p = fileBuf; *p; p++ )
		if ( *p == '\n' )
			ix++;
	app->numMembers = ix-1;
	app->members = (appRsrcItem_t*)TaskCalloc(0,ix-1,sizeof(appRsrcItem_t));
	for ( p = fileBuf; *p != '\n'; p++ );
	if ( *(p-1) == 13 )	/* DOS CR-LF */
		*(p-1) = 0;
	*p++ = 0;
	app->version = atoi(fileBuf);
	for ( ix = 0, start = p; *p; p++ ) {
		if ( *p == '\n' ) {
			app->members[ix].name = start;
			app->members[ix].bitpos = ix + 16;
			ix++;
			*p = 0;
			if ( *(p-1) == 13 )	/* DOS CR-LF */
				*(p-1) = 0;
			start = p+1;
		}
	}
	qsort(app->members,app->numMembers,sizeof(appRsrcItem_t),cmpRsrc);
}

#ifndef ACONST
#define ACONST const
#endif

static int
compSubArrayElem(arg1, arg2)
ACONST void *arg1, *arg2;
{
	data_ptr *e1, *e2;
	e1 = (data_ptr *)arg1;
	e2 = (data_ptr *)arg2;
	if (ElfbArraySize(*e1) != ElfbArraySize(*e2))
		tempAbort("Elements are of different size");
	return(IntFromArray(*e1,1)-IntFromArray(*e2,1));
}

static int cmpRsrc(arg1, arg2)
ACONST void *arg1, *arg2;
{
	appRsrcItem_t *rc1 = (appRsrcItem_t*)arg1;
	appRsrcItem_t *rc2 = (appRsrcItem_t*)arg2;
	return(strcmp(rc1->name,rc2->name));
}	


static void loadAppRsrcInfo(void) {
	int fileSize;
	char *fileBuf;
	char filename[STRPATH+1];
	int ix;
	char *p, *start;

	sprintf(filename,"%s/aarsrc/eng/aarsrc.lst",ThimApplixDir);
	fileBuf = AxReadFileIntoTaskBuf(0,filename,&fileSize);
	if ( !fileBuf )
		ElfStrAbort(1,"Could not open file",filename);
	for ( ix = 0, p = fileBuf; *p; p++ )
		if ( *p == '\n' )
			ix++;
	AppRsrcInfo = (appRsrcInfo_t*)TaskCalloc(0,ix,sizeof(appRsrcInfo_t));
	start = fileBuf;
	for ( ix = 0, p = fileBuf; *p; p++ ) {
		if ( *p == '\n' ) {
			if ( start[0] != '#' && start[1] != '\n' )
				AppRsrcInfo[ix++].name = start;
			*p = 0;
			if ( *(p-1) == 13 )	/* DOS CR-LF */
				*(p-1) = 0;
			start = p+1;
		}
	}
	NumAppRsrcInfo = ix;
}

/*************************************************************************
function: AxJavaAppDexFromName - Return application id from name string
*/
int AxJavaAppDexFromName (char *name)
{
	int ix;
	for ( ix = 0; ix < NumAppRsrcInfo; ix++ )
		if ( streq(AppRsrcInfo[ix].name,name) )
			return(ix);
	return(-1);
}

/*************************************************************************
function: AxJavaNoteClientResources - Message from client about its resource cache
*/
void AxJavaNoteClientResources(data_ptr info) {
	data_ptr allLists, thisList;
	int numLists, numBytes, appId, ix, allocBytes;
	char *bytes, *name;
	jchannel_t *jc = javaChanFromUid(TaskInfo[AxInTask].uid);
	appRsrcInfo_t *app;
	int err;

/*	AxPrintfDatum(info); */

	allLists = AxArrayElement(info,4);
	numLists = AxArraySize(allLists);
	for ( ix = 0; ix < numLists; ix++ ) {
		thisList = AxArrayElement(allLists,ix);
		name = AxStrFromArray(thisList,0);
		numBytes = AxBinaryFromArray(thisList, 1, &bytes);
		if ( !AppRsrcInfo )
			loadAppRsrcInfo();
		appId = AxJavaAppDexFromName(name);
		if ( (appId < 0) || (appId >= NumAppRsrcInfo) )
			continue;
		app = &AppRsrcInfo[appId];
		if ( !app->members )
			loadAppRsrcTable(appId);
		if ( jc->rsrcCacheMap )
			tempAbort(NXLT("anyware init timing error, resources pre-sent"));
		jc->rsrcCacheMap = (char**)TaskCalloc(0,NumAppRsrcInfo,sizeof(char*));
		allocBytes = (app->numMembers+15+7)/8;
		jc->rsrcCacheMap[appId] = (char*)TaskCalloc(0,1,allocBytes);
		memcpy(jc->rsrcCacheMap[appId],bytes,MIN(numBytes,allocBytes));
	}

/* Post initial resources */

	err = ElfCallByName(NXLT("POST_ALL_IMAGES$"),NULL,0);
	if ( err != AOK )
		tempAbort("Init error - could not open resource files (javacomm)");
}

void JvIASMainUtil (data_ptr cmdData)
{
	data_ptr aP = AxMakeArray(1);
	data_ptr bellP = AxMakeArray(2);
	bellP = AxAddIntToArray(bellP,0,JV_RING_BELL);
	bellP = AxAddDataToArray(bellP,1,cmdData);
	aP = AxAddDataToArray(aP,0,bellP);
	AxfJavaNotify(aP);
    AxFreeArrayButNotContents(aP);
    AxFreeArrayButNotContents(bellP);
}

void RingJavaBell ()
{
	data_ptr cmdData = AxMakeArray(1);
	AxAddIntToArray(cmdData,0,0);		/* subcode for star menu */
	JvIASMainUtil(cmdData);
	AxFreeData(cmdData);
}

void JvNewStarMenu (int newmail)
{
	data_ptr cmdData = AxMakeArray(2);
	AxAddIntToArray(cmdData,0,1);		/* subcode for star menu */
	AxAddIntToArray(cmdData,1,newmail);
	JvIASMainUtil(cmdData);
	AxFreeData(cmdData);
}

/**
  macro aagtwy_post_file$(host,port,user,filename)
  endmacro
 */
data_ptr AxfAAgtwyPostFile (data_ptr argDatum) {
	char *host = AxStrFromArray(argDatum,0);
	int port = AxIntFromArray(argDatum,1);
	char *uri = AxStrFromArray(argDatum,2);
	char *filename =AxStrFromArray(argDatum,3);
	int sockfd = -1;
	int rcode = -1;
	int fd = -1;
	char *filebuf = NULL;
	struct stat stbuf;

	if ((NULL == filename) || (0 == *filename) || (-1 == stat(filename,&stbuf))) {
		return AxMakeIntData(-1);
	}
	if (0 == port)
		port = 80;
	sockfd = aagtwyConnect(host,port);
	if (sockfd < 0)
		return AxMakeIntData(-2);

#if AXMS
    fd = open(filename, _O_RDONLY | _O_BINARY);
#else
	fd = open(filename, O_RDONLY);
#endif
	if (-1 == fd) {
		goto windup;
	}
	filebuf = (char*)malloc(stbuf.st_size+1);
	if (NULL == filebuf)
		goto windup;
#if AXMS
	if (read(fd,filebuf,stbuf.st_size) <= 0 )
#else
	if (stbuf.st_size != read(fd,filebuf,stbuf.st_size))
#endif
	{
		goto windup;
	}
	/* Send POST request and headers */
	aagtwySendf(sockfd,"POST %s HTTP/1.0\r\n",uri);
	aagtwySendf(sockfd,"User-agent: Applix Anyware/4.4\r\n");
	aagtwySendf(sockfd,"Content-length: %d\r\n",stbuf.st_size);
	aagtwySendf(sockfd,"\r\n");
	/* send content */
	aagtwySend(sockfd,filebuf,stbuf.st_size);

	/* get response */
	if (TRUE) {
		char *response = aagtwyGetResponse(sockfd,NULL,NULL);
		if (NULL != response) {
			char *rcodestr = strchr(response,' ');
			if (NULL != rcodestr)
				rcode = atoi(rcodestr+1);
		}
		aagtwyFreeResponse(response);
	}

  windup:
	if (fd > 0)
		close(fd);
	if (NULL != filebuf)
		free(filebuf);
	/* close socket */
	aagtwyDisconnect(sockfd);

	return AxMakeIntData(rcode);
}
