/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: jvsymgen.c
   Applix ASC version: 11.2
   Date: Wed Nov 18 16:31:18 1998
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#if AX_X11
#include <strings.h>
#endif
#include <malloc.h>

/********************/
/* PROTOTYPES */
/********************/

void process_syms();
char *fileRead();
void parseLine();
char *pounder();
void dumpLine();

/********************/
/* DEFINES AND MACROS */
/********************/

/********************/
/***** TYPDEFS AND STRUCTS *****/
/********************/

/********************/
/***** GLOBALS *****/
/********************/

extern int errno;
#if !LINUX
extern char *sys_errlist[];
#endif

/********************/
/***** CODE *****/
/********************/


int main(int argc, char *argv[]) 
{
    char *buf;
    int bufsz;
    int i;
	int len;
    char *elf_inc = NULL;
    char *c_inc = NULL;
    char *mc_class = NULL;
    char *symbase = NULL;
    char *class_name = NULL;
    char *temp_name = NULL;
    char *p = NULL;
    FILE *elf_inc_fp = NULL;
    FILE *c_inc_fp = NULL;
    FILE *mc_class_fp = NULL;

    symbase = argv[2];
    if( strstr(argv[1],".elh") ){
        elf_inc = argv[1];
    } else if( strstr(argv[1],".h") ){
        c_inc = argv[1];
    } else if( strstr(argv[1],".java") ){
        mc_class = argv[1];
    } else if( strstr(argv[1],".jv") ){
        mc_class = argv[1];
    } if( strstr(argv[1],".dat") ){
        symbase = argv[1];
    } 
    if( !elf_inc && !c_inc && !mc_class ) {
        fprintf(stderr,"jvsymgen: missing c_inc/elf_inc/mc_class parameter\n");
        exit(1);
    }
    if( symbase == NULL ){
        fprintf(stderr,"jvsymgen: missing symbase parameter\n");
        exit(1);
    }

    buf = fileRead(symbase,&bufsz);
    if( bufsz < 1 || buf == NULL ){
        /* empty file */
        fprintf(stderr,"jvsymgen: symbase file %s is empty\n",symbase);
        exit(1);
    }

    if ( elf_inc ) {
        elf_inc_fp = fopen(elf_inc,"w");
        if(elf_inc_fp == NULL){
            perror(elf_inc);
            exit(1);
        }
        if( 0 > fprintf(elf_inc_fp,
            "/* Machine generated file hack for manifest constants */\n")){
            perror(elf_inc);
            exit(1);
        }
    }

    if ( c_inc ) {
        c_inc_fp = fopen(c_inc,"w");
        if(c_inc_fp == NULL){
            perror(c_inc);
            exit(1);
        }
        if( 0 > fprintf(c_inc_fp,
            "/* Machine generated file hack for manifest constants */\n")){
            perror(c_inc);
            exit(1);
        }
    }

    if ( mc_class ) {
        mc_class_fp = fopen(mc_class,"w");
        if(mc_class_fp == NULL){
            perror(mc_class);
            exit(1);
        }
        fprintf(mc_class_fp,"//  Machine generated file\n");
        fprintf(mc_class_fp,"//  Hack for manifest constants\n");
        fprintf(mc_class_fp,"//\n");
		temp_name = strstr(mc_class,"mc");  
		class_name = strtok(temp_name,".");
        fprintf(mc_class_fp,"class %s {\n",class_name);
    }
    process_syms(buf,bufsz,elf_inc_fp,c_inc_fp,mc_class_fp);

    if ( elf_inc ) {    
        if( fclose(elf_inc_fp) != 0 ){
            perror(elf_inc);
            exit(1);
        }
    }

    if ( c_inc ) {
        if( fclose(c_inc_fp) != 0 ){
            perror(c_inc);
            exit(1);
        }
    }

    if ( mc_class ) {
        fprintf(mc_class_fp,"}\n");
        if( fclose(mc_class_fp) != 0 ){
            perror(mc_class);
            exit(1);
        }
    }

    free(buf); /* not really necessary since we're about to exit, but... */
    exit(0);
}

void process_syms(buf,bufsz,elf_inc_fp,c_inc_fp,mc_class_fp)
char *buf;
int bufsz;
FILE *elf_inc_fp;
FILE *c_inc_fp;
FILE *mc_class_fp;
{
    int chars;
    char *lineBegin;
    char *val,*var,*comment;

    chars = 0;
    lineBegin = buf;
    while( chars < bufsz ){
        
        /* have we come to the end of a line? */
        if( buf[chars] == '\n' ){

            /* null terminate it */
            buf[chars] = '\0';

            /* skip comment and empty lines */
            if( *lineBegin != '#' && *lineBegin != '\0'){

                /* split line up */
                parseLine(lineBegin,&val,&var,&comment);

                /* dump output to header files */
                dumpLine(elf_inc_fp,c_inc_fp,mc_class_fp,val,var,comment);
            }

            /* the next line begins after the current character */
            lineBegin = buf+chars+1;
        }
        chars++;
    }

    /* ideally '\n' is last char in file */
    if( lineBegin == buf+bufsz )
        return;
    
    /* if last line is basically void that's fine too */
    if( *lineBegin == '#' )
        return;

    fprintf(stderr,"Warning: jvsymgen input missing newline on last line\n");

    /* this char is in the comment field, so no real damage possible */
    buf[bufsz-1] = '\0';

    parseLine(lineBegin,&val,&var,&comment);
    dumpLine(elf_inc_fp,c_inc_fp,mc_class_fp,val,var,comment);

}

void dumpLine(elf_inc_fp,c_inc_fp,mc_class_fp,val,var,comment)
FILE *elf_inc_fp;
FILE *c_inc_fp;
FILE *mc_class_fp;
char *val;
char *var;
char *comment;
{
    /* input lines beginning with "." are only for mc.class */
    if( val[0] == '.' ){
        /* +1 strips off the leading . from val */
        if ( mc_class_fp )
            fprintf(mc_class_fp,
                "public final static int %s = %s; // %s\n",
                var,val+1,comment);
    } else {
        if ( mc_class_fp )
            fprintf(mc_class_fp,
                "public final static int %s = %s; // %s\n",
                var,val,comment);
        if ( c_inc_fp )
            fprintf(c_inc_fp,"#define %s %s /* %s */\n",
                var,val,comment);
        /* this must come last, since pounder modifies var */
        if ( elf_inc_fp ) 
            fprintf(elf_inc_fp,"#define %s %s /* %s */\n",
                pounder(var),val,comment);
    }
}

/* return contents of a file */
char *fileRead(char *fPath,int *size)
{
    struct stat statBuf;
    FILE *fP;
    char *fBuf;

    if( stat( fPath, &statBuf) == -1 ){
        fprintf(stderr,"jvsymgen: cannot stat file <%s>\n",fPath);
        exit(1);
    }

    fP = fopen(fPath, "r");
    if( fP == NULL ){
        perror(fPath);
        exit(1);
    }
        
    fBuf = (char *) malloc( sizeof(char)*statBuf.st_size );
    if( !fBuf ){
        fprintf(stderr,"jvsymgen: cannot allocate buffer for file %s",fPath);
        exit(1);
    }

    fread( fBuf, sizeof(char), statBuf.st_size, fP);
    if( fBuf == (char *) 0)
        if( ferror(fP) ){
            perror(fPath);
            exit(1);
        }
    if( fclose(fP) != 0 ){
        perror(fPath);
        exit(1);
    }

    *size = statBuf.st_size;
    return(fBuf);
}

/* split up colon separated line with three fields */
void parseLine(line,val,var,comment)
char *line;
char **val;
char **var;
char **comment;
{
    int i;
    char *endP;
    char lineBak[300];
    
    /* save a copy for error reporting. */
    sprintf(lineBak,"%s",line);
        
    *val = line;
    endP = strchr(*val,':');
    if( !endP ){
        fprintf(stderr,"jvsymgen: badly formatted line <%s>\n",lineBak);
        exit(1);
    }
    *endP = '\0';

    *var = endP+1;
    endP = strchr(*var,':');
    if( !endP ){
        fprintf(stderr,"jvsymgen: badly formatted line <%s>\n",lineBak);
        exit(1);
    }
    *endP = '\0';

    *comment = endP+1;

    return;
}


/* convert underscores in name to pound signs */
char *pounder( name )
char *name;
{
    char *poundP;

    poundP = strchr(name,'_');
    while( poundP ){
        *poundP = '#';
        poundP = strchr(name,'_');
    }

    return(name);
}

