/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pf04prim.c
   Applix ASC version: 52
   Date: Mon Nov 16 13:56:50 1998
*/
#define _WINMAP_H    1

#include <stdio.h>
#include <math.h>
#include <basicwin.h>
#include <winspool.h>
#include <aos.h>
#include <pf.h>
#include <paratm.h> 
#include <pfint.h>
#include <pfextra.h>
#include <pfpat.h>
#include <xmlib.h>
#include <axtext.h>
#include <axbitmap.h>
#include <axprint.h>
#include <getstr2.h>
#include <colors.h>
#include <xmpixel.h>
#include <axwinclr.h>


/****************************************************************************/
/* ADD ANY NEW Win32 API CALLS HERE RATHER THAN IN winmap.h
 * KEEP READING FOR REASON WHY...
 */
/*     It seems that when a user cancels a print job, the software may
 * still be in the process of making a GDI call. Since the HDC to the
 * printer becomes invalid, the GDI call fails (that is, it quietly
 * fails, simply returning zero or whatever, depending on the call).
 * This would seem to indicate that any GDI call used for printing
 * should NOT check its error return (or i should say, should not
 * cause a software abort). Is this true? I have no idea. I cannot find 
 * any documentation on this subject.
 *     Since we have mapped many of the API calls to a macro that checks
 * for an error return (and aborts!), I have made sure we do not include
 * the file that does this (winmap.h). And just so we are consistent with
 * the rest of the code, I have defined the API calls used here as simple
 * MS_ tokens, rather than macros which may cause an abort. This, of course,
 * means we'll need to use other means (debugger or tools like Bounds-Checker)
 * to catch any _real_ GDI errors.             -donato
 */
#define MS_PatBlt                  PatBlt
#define MS_RestoreDC               RestoreDC
#define MS_GetLastError            GetLastError
#define MS_AbortDoc                AbortDoc
#define MS_MessageBox              MessageBox
#define MS_GetFocus                GetFocus
#define MS_GetParent               GetParent
#define MS_GetDeviceCaps           GetDeviceCaps
#define MS_DPtoLP                  DPtoLP
#define MS_LPtoDP                  LPtoDP
#define MS_EnableWindow            EnableWindow
#define MS_CreateDialog            CreateDialog
#define MS_SetAbortProc            SetAbortProc
#define MS_StartDoc                StartDoc
#define MS_EndDoc                  EndDoc
#define MS_DestroyWindow           DestroyWindow
#define MS_DeleteDC                DeleteDC
#define MS_EndPage                 EndPage
#define MS_StartPage               StartPage
#define MS_SendMessage             SendMessage
#define MS_ResetDC                 ResetDC
#define MS_SaveDC                  SaveDC
#define MS_SetTextAlign            SetTextAlign
#define MS_SetBkMode               SetBkMode
#define MS_SetBkColor              SetBkColor
#define MS_SetTextColor            SetTextColor
#define MS_TextOut                 TextOut
#define MS_SelectObject            SelectObject
#define MS_DeleteObject            DeleteObject
#define MS_CreateSolidBrush        CreateSolidBrush
#define MS_Polygon                 Polygon
#define MS_Polyline                Polyline
#define MS_GetTextMetrics          GetTextMetrics
#define MS_CreateFontIndirect      CreateFontIndirect
#define MS_CreateRectRgn           CreateRectRgn
#define MS_SelectClipRgn           SelectClipRgn
#define MS_GetStockObject          GetStockObject
#define MS_CreatePatternBrush      CreatePatternBrush
#define MS_CreateBitmap            CreateBitmap
#define MS_CreateCompatibleDC      CreateCompatibleDC
#define MS_CreateCompatibleBitmap  CreateCompatibleBitmap
#define MS_StretchBlt              StretchBlt
#define MS_StretchDIBits           StretchDIBits
#define MS_CreatePen               CreatePen
#define MS_BitBlt                  BitBlt
#define MS_CreateDIBitmap          CreateDIBitmap
#define MS_GetObject               GetObject
#define MS_SetPolyFillMode         SetPolyFillMode
#define MS_GetTextExtentPoint32    GetTextExtentPoint32
#define MS_PeekMessage             PeekMessage
#define MS_IsDialogMessage         IsDialogMessage
#define MS_TranslateMessage        TranslateMessage
#define MS_DispatchMessage         DispatchMessage
#define MS_EnableMenuItem          EnableMenuItem
#define MS_GetSystemMenu           GetSystemMenu
#define MS_SetWindowText           SetWindowText
#define MS_SetDlgItemText          SetDlgItemText
#define MS_BeginPaint              BeginPaint
#define MS_EndPaint                EndPaint
#define MS_ClientToScreen          ClientToScreen
#define MS_SetBrushOrgEx           SetBrushOrgEx
#define MS_LineTo                  LineTo
#define MS_MoveToEx                MoveToEx
#define MS_Rectangle               Rectangle
#define MS_Arc                     Arc
#define MS_Pie                     Pie
#define MS_Escape                  Escape

#define MS32_BeginPath             BeginPath
#define MS32_EndPath               EndPath
#define MS32_FillPath              FillPath
#define MS32_PolyBezier            PolyBezier
#define MS32_CloseFigure           CloseFigure
#define MS32_SelectClipPath        SelectClipPath
#define MS32_ExtCreatePen          ExtCreatePen
#define MS32_WidenPath             WidenPath
#define MS32_StrokePath            StrokePath
#define MS32_StrokeAndFillPath     StrokeAndFillPath
#define MS32_DocumentProperties    DocumentProperties

#define MSNT_SetGraphicsMode       SetGraphicsMode
#define MSNT_SetWorldTransform     SetWorldTransform
#define MSNT_ModifyWorldTransform  ModifyWorldTransform
/****************************************************************************/


#define MINIMUM_LINE_WIDTH      1
#define MARK_MARGIN             125

/* these are tied to axPrintDitherStyle */
/* (used for gradients) */
#define PF_DITHER_HT4       0
#define PF_DITHER_FS        -1

/* our own custom message to be sent to the print dialog window. */
/* see PrintDialogProc */
#define WM_PAGECOUNT WM_USER + 4

/* dialog box control IDs for the print cancel dialog box on NT. */
/* These are not in an include file right now! so make sure they *
 * are exactly what is defined in axmain.rc          -donato     */
#define IDD_PRINTTEXT 100
#define IDD_PAGETEXT  101
#define IDD_FILENAME  102
#define IDD_PAGECOUNT 103
/*****************************************************************/

#define BANNER_POINTSIZE  18
#define BANNER_LEFTMARGIN 500

#define D04_POLY_FILL   2
#define D04_LINE_FILL   3

#define D04_300_DPI  0
#define D04_600_DPI  1
static int PrinterResolution;
static int PrinterIsPostScript;
static short D04_UWidMils, D04_UHtMils, D04_UWidPels, D04_UHtPels;
static int XPixelsPerInch, YPixelsPerInch;

#define GRADIENT_FILL(fill) (fill == RECT_FILL_TYPE || \
                            fill == LINEAR_FILL_TYPE || \
                            fill == RADIAL_FILL_TYPE || \
                            fill == BILINEAR_FILL_TYPE)

#define SOLID_FILL(attr) (attr->f_type == GRAY_FILL_TYPE && \
						 (attr->f_pat == 1000 || attr->f_pat == 0))

extern int Running_WindowsNT, Running_Windows95;
extern short THIMpid;
extern HANDLE hInst;
extern unsigned long AlisPixels[];
extern LOGPALETTE *AxWinPalette;
extern PIXEL_VALUE BackPixelRGB, ForeWorkPixelRGB;


static DWORD basicdash[2] = {  70, 50 };
static DWORD fat__dott[2] = {  30, 50 };
static DWORD dot__dash[4] = {  70, 50, 30, 50 };
static DWORD thin_dott[2] = {  10, 10 };
static DWORD dtdt_dash[6] = {  70, 50, 30, 50, 30, 50 };
static DWORD long_dash[2] = { 140, 50 };

static int dash_length[] = { 2, 2, 4, 2, 6, 2 };
static DWORD *dash_list[] = {
    basicdash, fat__dott, dot__dash, thin_dott, dtdt_dash, long_dash
};

static unsigned char Bitmask[] = {0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};

static DWORD gray_values[10] = { RGB(0,0,0), RGB(50,50,50), RGB(80,80,80),
                           RGB(105,105,105),RGB(128,128,128),RGB(155,155,155),
                           RGB(180,180,180),RGB(205,205,205),RGB(233,233,233),
                           RGB(255,255,255)
                           };


static HWND Dlg_Hwnd, Parent_Hwnd;
BOOL UserPrintAbort = FALSE;

static int Duplex_global;

/* should we put these in the global PF struct? */
static int Page_X_Offset;
static int Page_Y_Offset;
static int PrinterSupportsPalette;
static int D04_Printing_Color;
/************************************************/

#include <pf04prim.p>


static void
D04PathAbort(void)
{
    int ret;
    DWORD last_error;

    last_error = MS_GetLastError();
    D04DebugMessage("Path error!  Printing aborted.");
    ret = MS_AbortDoc(PF.PfPrinterHDC);
}

static void
D04DebugMessage(char *str)
{
#if WIN_DEBUG
    if ( !UserPrintAbort )
        MS_MessageBox(NULL, str, NXLT("debug"), MB_OK);
#endif
}



static HWND
D04GetParentWindow(void)
{
    HWND focus, parent;

    focus = MS_GetFocus();
    parent = MS_GetParent(focus);

    return parent;
}


static void
D04InitPrinter(void)
{
    int prn_code;
    HDC hdc = PF.PfPrinterHDC;

    Page_X_Offset = MS_GetDeviceCaps(hdc, PHYSICALOFFSETX);
    Page_Y_Offset = MS_GetDeviceCaps(hdc, PHYSICALOFFSETY);

    XPixelsPerInch = MS_GetDeviceCaps(hdc, LOGPIXELSX);
    YPixelsPerInch = PrinterResolution = MS_GetDeviceCaps(hdc, LOGPIXELSY);

    /* compute conversion factors for printer resolution <--> 75 dpi */
    D04_UWidMils = D04_UHtMils = PrinterResolution / 25;
    D04_UWidPels = D04_UHtPels = 3;

    if ( MS_GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE )
        PrinterSupportsPalette = TRUE;
    else
        PrinterSupportsPalette = FALSE;

    /* set the global duplex flag to -1. we will be changing it to the
     * requested duplex settings as they change.
     */
    Duplex_global = -1;

    /* are we printing to a PostScript printer? (this method taken from
     * Knowledge Base article Q124135. I'm assuming it's reliable!).
     * We only care about the printer driver on Win95 (due to strange
     * differences in GDI interpretation).  The default assumption is
     * that PostScript is true because it's the PCL case where things
     * go strange.
     */
    PrinterIsPostScript = TRUE;
    prn_code = POSTSCRIPT_PASSTHROUGH;
    if ( !MS_Escape(hdc, QUERYESCSUPPORT, sizeof(int), (LPCSTR)&prn_code,
                                                            (VOID*)NULL) )
        PrinterIsPostScript = FALSE;

    /* setup the "print cancel" dialog box and abort procedure. */
    UserPrintAbort = FALSE;    
    Parent_Hwnd = D04GetParentWindow();
    MS_EnableWindow(Parent_Hwnd, FALSE);
    Dlg_Hwnd = MS_CreateDialog(hInst, "PrintDialogBox", Parent_Hwnd,
                                        (DLGPROC)PrintDialogProc);
    MS_SetAbortProc(hdc, (ABORTPROC)PrinterAbortProc);
}

static void
D04PixToPrinterRes(int pix_wid, int pix_hyt, int *res_wid, int *res_hyt)
{
    int roundup;

    roundup = D04_UWidPels - 1;
    if ( pix_wid < 0 )
        roundup = (-roundup);
    *res_wid = ((long)(pix_wid) * D04_UWidMils + roundup) / D04_UWidPels;
    roundup = D04_UHtPels - 1;
    if ( pix_hyt < 0 )
        roundup = (-roundup);
    *res_hyt = ((long)(pix_hyt) * D04_UHtMils + roundup) / D04_UHtPels;
}

static void
D04ConvertMilsToPrinterRes(POINT *points, int npoints, BOOL use_page_offsets)
{
    int i;
    int x_offset = 0, y_offset = 0;

    if ( use_page_offsets )
        {
        x_offset = Page_X_Offset;
        y_offset = Page_Y_Offset;
        }

    for ( i=0; i < npoints; i++ )
        {
        points[i].x = ((points[i].x * XPixelsPerInch) / 1000) - x_offset;
        points[i].y = ((points[i].y * YPixelsPerInch) / 1000) - y_offset;
        }
}


int
D04StartJob(void)
{
    int retval;
    DWORD pf_error;
    DOCINFO docinfo;


    PF.PfClipRgn = NULL;
    if ( PF.PfPrinterName && PF.PfPrinterName[0] )
        PF.PfPrinterHDC = GetPrinterHDCfromName(PF.PfPrinterName);
    else
        {
        PF.PfPrinterHDC = GetDefaultPrinterHDC(PF.pi.print_to_file,
                                                &PF.PfPrinterName);
        }

    D04InitPrinter();

    memset(&docinfo, 0, sizeof(DOCINFO));
    docinfo.cbSize = sizeof(DOCINFO);
    docinfo.lpszDocName = PF.PfDoc_name;
    docinfo.lpszOutput = NULL;
    retval = MS_StartDoc(PF.PfPrinterHDC, &docinfo);
    if ( retval < 0 )
        {
        pf_error = MS_GetLastError();
        D04DebugMessage(NXLT("D04StartJob: could not start printer!"));
        }

    pfClearPattern();
    /* Load pattern definitions. Since we don't do a pre-scan for *
     * GDI printing, the following will load all patterns.        */
    //D04PreLoadPatterns();

    if ( PF.PfBanner )
        StdBanner(NULL, BANNER_POINTSIZE, BANNER_LEFTMARGIN + Page_X_Offset); 
    return retval;
}

int
D04EndJob(void)
{
    int retval;
    DWORD pf_error;

    /* if user did not cancel print job, we need to destroy the dialog. */
    if ( !UserPrintAbort )
        {
        retval = MS_EndDoc(PF.PfPrinterHDC);
        if ( retval < 0 )
            {
            pf_error = MS_GetLastError();
            D04DebugMessage(NXLT("D04EndJob: could not end job!"));
            }
        MS_EnableWindow(Parent_Hwnd, TRUE);
        if ( Dlg_Hwnd )
            MS_DestroyWindow(Dlg_Hwnd);
        }

    D04DeletePatterns();
    MS_DeleteDC(PF.PfPrinterHDC);
    /* free what was allocated in GetDefaultPrinterHDC when
     * printing to file.
     */
    if ( PF.pi.print_to_file )
        {
        free(PF.PfPrinterName);
        ClosePrinter(PF.PfPrinterHandle);
        PF.PfPrinterHandle = NULL;
        }
    return retval;
}


int
D04NewPage(int firstPage, TmRect32 *clip)
/* int firstPage;    bool: true for 1st physical page of printed output */
{
    int retval;
    DWORD pf_error;
    static int page_count;

    retval = 0;
    /* if this isn't the first page, let's "end" the *
     * last one we printed.           -donato        */
    if ( !firstPage || (PF.PfJob_done && !firstPage) )
        {
        retval = MS_EndPage(PF.PfPrinterHDC);
        if ( retval < 0 )
            {
            pf_error = MS_GetLastError();
            D04DebugMessage(NXLT("D04NewPage: could not end page!"));
            }
        }

    /* only start a new page if we haven't finished printing. */
    if ( !PF.PfJob_done && !UserPrintAbort )
        {
        page_count++;
        /* The following strange looking 'if' statement means:
         * If duplex hasn't been set, or it is the "simplex" default,
         * then we can set page modes on the driver. If a non-simplex
         * mode has been previously set, and has not yet changed, do 
         * not set any page modes (as this would, in effect, remove any
         * previous duplexing). If we are changing duplex mode, then we also
         * want to set page modes.         -donato
         */
        if ( Duplex_global <= 0 || (PF.duplex >= 0 &&
                       Duplex_global != PF.duplex) )
            D04SetDevicePageModes();
        retval = MS_StartPage(PF.PfPrinterHDC);
        if ( retval < 0 )
            {
            pf_error = MS_GetLastError();
            D04DebugMessage(NXLT("D04NewPage: could not start page!"));
            }
        /* tell the "print cancel" dialog box which page we are printing. */
        MS_SendMessage(Dlg_Hwnd, WM_PAGECOUNT, page_count, 0L);
        }
    else
        {
        page_count = 0;   /* print job done, reset counter */
        if ( UserPrintAbort )
            MS_AbortDoc(PF.PfPrinterHDC);
        }

    return retval;
}


int
D04SetPos(int new_x, int new_y)
{
    /* For GDI printing, we'll just set these globals. */
    PF.PfHardX = new_x;
    PF.PfHardY = new_y;
    return 0;
}



/***********************************************************************
  D04SetDevicePageModes - set orientation, color, collation, etc...
 ***/
static void
D04SetDevicePageModes(void)
{
    char *form_name;
    int wid, hyt, mil_wid, mil_hyt;
    LONG ret;
    DWORD bytes_needed;
    DEVMODE *dev_modeP = NULL;

    /* can we assume we have a printer handle? */
    /* yes, because one is always created except when printing to file, *
     * which is never the case if we are doing GDI printing.  -donato   */
    /* Right now (10.3.95), we may be printing to file thru the GDI
     * if the user so specified in the ELF print macro. In that case,
     * the printer handle will be null, so we "open" one here. We "close"
     * it later in D04EndJob.                   -donato
     */
    if ( !PF.PfPrinterHandle && PF.pi.print_to_file )
        OpenPrinter(PF.PfPrinterName, &PF.PfPrinterHandle, NULL);
    bytes_needed = MS32_DocumentProperties(NULL, PF.PfPrinterHandle,
                                            PF.PfPrinterName,
                                            NULL, NULL, 0);
    if ( bytes_needed > 0 )
        {
        dev_modeP = (DEVMODE*)malloc(bytes_needed);
        ret = MS32_DocumentProperties(NULL, PF.PfPrinterHandle,
                                            PF.PfPrinterName, dev_modeP,
                                            NULL, DM_OUT_BUFFER);
        }
    else
        D04DebugMessage("DocumentProperties error in D04SetDevicePageModes");

    if ( dev_modeP && ret >= 0 ) 
        {
        /* we check the dmFields member to make sure the printer
         * actually supports the mode we are attempting to set.
         */
        mil_wid = PF.PfPageMilWidth;
        mil_hyt = PF.PfPageMilHeight;
        if ( PF.PfLandscape ) //&& (dev_modeP->dmFields & DM_ORIENTATION) )
            dev_modeP->dmOrientation = DMORIENT_LANDSCAPE;
        else
            dev_modeP->dmOrientation = DMORIENT_PORTRAIT;
        dev_modeP->dmFields |= DM_ORIENTATION;

        //if ( dev_modeP->dmFields & DM_DUPLEX )
        //    {
            /* "PF.pi.duplex" is the duplex setting from the Print Options
             * dialog box, and overrides any other settings (e.g. document).
             * "PF.duplex" is the duplex setting from the document.
             */
            if ( PF.pi.duplex >= 0 )
                {
                dev_modeP->dmDuplex = select_plex(PF.pi.duplex);
                Duplex_global = PF.pi.duplex;
                dev_modeP->dmFields |= DM_DUPLEX;
                }
            else if ( PF.duplex >= 0 )
                {
                dev_modeP->dmDuplex = select_plex(PF.duplex);
                Duplex_global = PF.duplex;
                dev_modeP->dmFields |= DM_DUPLEX;
                }
        //    }

        if ( PF.pi.collate )//&& (dev_modeP->dmFields & DM_COLLATE) )
            {
            dev_modeP->dmFields |= DM_COLLATE;
            dev_modeP->dmCollate = DMCOLLATE_TRUE;
            }

        D04_Printing_Color = FALSE;
        dev_modeP->dmFields |= DM_COLOR;
        if ( PF.PfColordev ) //&& (dev_modeP->dmFields & DM_COLOR) )
            {
            dev_modeP->dmColor = DMCOLOR_COLOR;
            D04_Printing_Color = TRUE;
            }
        else
            /* this will set even color printers to monochrome. */
            dev_modeP->dmColor = DMCOLOR_MONOCHROME;

        if ( PF.PfPrintMultiCopiesInLang > 1 )//&&
                //(dev_modeP->dmFields & DM_COPIES) )
            {
            dev_modeP->dmFields |= DM_COPIES;
            dev_modeP->dmCopies = PF.PfPrintMultiCopiesInLang;
            }

        if ( PF.pi.size ) //&& (dev_modeP->dmFields & DM_PAPERSIZE) )
            {
            dev_modeP->dmFields |= DM_PAPERSIZE;
            dev_modeP->dmPaperSize = select_paper(mil_wid, mil_hyt);
            }
        else if ( PF.pi.size )//&& (dev_modeP->dmFields & DM_PAPERWIDTH &&
                              //  dev_modeP->dmFields & DM_PAPERLENGTH) )
            {
            select_size(mil_wid, mil_hyt, &wid, &hyt);
            /* convert back to 10ths of millimeters for GDI */
            dev_modeP->dmPaperWidth  = (wid/1000) * 254;
            dev_modeP->dmPaperLength = (hyt/1000) * 254;
            dev_modeP->dmFields |= DM_PAPERWIDTH;
            dev_modeP->dmFields |= DM_PAPERLENGTH;
            }
        else if ( PF.pi.size )//&& (dev_modeP->dmFields & DM_FORMNAME) )
            {
            form_name = select_form(mil_wid, mil_hyt);
            if ( form_name )
                strcpy(dev_modeP->dmFormName, form_name);
            dev_modeP->dmFields |= DM_FORMNAME;
            }

        ret = MS32_DocumentProperties(NULL,PF.PfPrinterHandle,PF.PfPrinterName,
                                                dev_modeP, dev_modeP,
                                                DM_IN_BUFFER | DM_OUT_BUFFER);
        if ( ret >= 0 )
            if ( !MS_ResetDC(PF.PfPrinterHDC, dev_modeP) )
                DebugMessage("Oops, ResetDC failed!");
        free(dev_modeP);
        }
    else
        {
        char buf[100];
        sprintf(buf, "Couldn't set page mode on: %s", PF.PfPrinterName);
        D04DebugMessage(buf);
        }
}
#if 0
static void
D04SetInitialDuplex(void)
{
    DWORD ret, bytes_needed;
    DEVMODE *dev_modeP = NULL;

    if ( !PF.PfPrinterHandle && PF.pi.print_to_file )
        OpenPrinter(PF.PfPrinterName, &PF.PfPrinterHandle, NULL);
    bytes_needed = MS32_DocumentProperties(NULL, PF.PfPrinterHandle,
                                            PF.PfPrinterName,
                                            NULL, NULL, 0);
    if ( bytes_needed > 0 )
        {
        dev_modeP = (DEVMODE*)malloc(bytes_needed);
        bytes_needed = MS32_DocumentProperties(NULL, PF.PfPrinterHandle,
                                            PF.PfPrinterName, dev_modeP,
                                            NULL, DM_OUT_BUFFER);
        }
    else
        D04DebugMessage("DocumentProperties error in D04SetDevicePageModes");

    if ( dev_modeP ) 
        {
        if ( dev_modeP->dmFields & DM_DUPLEX )
            {
            if ( PF.pi.duplex >= 0 )
                dev_modeP->dmDuplex = select_plex(PF.pi.duplex);
            else if ( PF.duplex >= 0 )
                dev_modeP->dmDuplex = select_plex(PF.duplex);
            }
        //MS_ResetDC(PF.PfPrinterHDC, dev_modeP);
        ret = MS32_DocumentProperties(NULL,PF.PfPrinterHandle,PF.PfPrinterName,
                                            NULL, dev_modeP, DM_MODIFY);
        free(dev_modeP);
        }
    else
        {
        char buf[100];
        sprintf(buf, "Couldn't set duplex mode on: %s", PF.PfPrinterName);
        D04DebugMessage(buf);
        }
}
#endif



static int
D04DrawFilledText(tx_char *textstr, int txlen, int tx, int ty,
					attr_type *attr, int unity, HFONT hfont, 
					int transparent)
{
	int tx_wid, tx_hyt;
    int retval, stock_brush; 
	int xpos, ypos, y_offset;
    HDC hdc, mask_dc;
	HFONT old_font;
	HBRUSH hbrush, old_brush;
	HBITMAP mask_hbm, old_mask_hbm;
    char *out_text;
    COLORREF fore_color, back_color, old_bk_color, old_text_color;
    unsigned long fore_pixel, back_pixel;
	TEXTMETRIC tm;
	SIZE size;
	DWORD rop_code;

    /* font handle should already be selected into the printer hdc. */
    hdc = PF.PfPrinterHDC;
    if ( PF.PfColordev && PrinterSupportsPalette )
        {
        D04GetForeAndBackPixels(attr, &fore_pixel, &back_pixel);
        fore_color = PALETTEINDEX(fore_pixel);
        back_color = PALETTEINDEX(back_pixel);
        }
    else
        D04GetForeAndBackColors(attr, &fore_color, &back_color);

	/* get the ascent because we'll be drawing a memory dc to the
	 * printer dc rather than the actual text, which comes in with
	 * baseline coordinates.
	 */
	MS_GetTextMetrics(hdc, &tm);
	y_offset = tm.tmAscent;

    out_text = D04GetOutputText(textstr, txlen);
    MS_GetTextExtentPoint32(hdc, out_text, txlen, &size);
    tx_wid = size.cx;
    tx_hyt = size.cy;
    if ( unity || Running_Windows95 )
        {
        xpos = ((tx * XPixelsPerInch) / 1000) - Page_X_Offset;
        ypos = ((ty * YPixelsPerInch) / 1000) - Page_Y_Offset;
        }
    else
        {
        /* if we're here, a world transform is in effect, which has
         * already taken the offsets into account.
         */
        xpos = ((tx * XPixelsPerInch) / 1000);
        ypos = ((ty * YPixelsPerInch) / 1000);
        }

    mask_dc  = MS_CreateCompatibleDC(hdc);
    mask_hbm = MS_CreateCompatibleBitmap(mask_dc, tx_wid, tx_hyt);
    old_mask_hbm = MS_SelectObject(mask_dc, mask_hbm);
    MS_PatBlt(mask_dc, 0, 0, tx_wid, tx_hyt, WHITENESS);
	MS_SetBkMode(mask_dc, TRANSPARENT);
	MS_SetBkColor(mask_dc, RGB_WHITE);
	MS_SetTextColor(mask_dc, RGB_BLACK);
	old_font = MS_SelectObject(mask_dc, hfont);
    retval = MS_TextOut(mask_dc, 0, 0, out_text, txlen);

    old_bk_color = MS_SetBkColor(hdc, fore_color);
    old_text_color = MS_SetTextColor(hdc, back_color);
    hbrush = D04GetPatternBrush(attr, fore_color, back_color, &stock_brush);
    old_brush = MS_SelectObject(hdc, hbrush);
	if ( transparent )
    	rop_code = 0x00A20E09;
	else
    	rop_code = 0x00E20746;
    MS_BitBlt(hdc, xpos,ypos-y_offset, tx_wid,tx_hyt, mask_dc, 0,0, rop_code);
    MS_SelectObject(hdc, old_brush);
	MS_SetBkColor(hdc, old_bk_color);
	MS_SetTextColor(hdc, old_text_color);

    /* cleanup */
	MS_SelectObject(mask_dc, old_font);
    MS_SelectObject(mask_dc, old_mask_hbm);
    MS_DeleteObject(mask_hbm);
    MS_DeleteDC(mask_dc);
    if ( !stock_brush )
        MS_DeleteObject(hbrush);
    free(out_text);
    return retval;
}

static int
D04DrawSimpleText(tx_char *textstr, int txlen, int tx, int ty,
                                COLORREF fore_color, int unity)
{
    int retval;
    HDC hdc;
    char *out_text;
    int xpos, ypos;

/* Do not use the Window Path API for simple text because the driver
 * will download raster fonts!
 */

    /* font handle should already be selected into the printer hdc. */
    hdc = PF.PfPrinterHDC;
    MS_SaveDC(hdc);

    /* only draw foreground text color, ignore background */
    MS_SetTextAlign(hdc, TA_LEFT | TA_BASELINE);
    MS_SetBkMode(hdc, TRANSPARENT);
    MS_SetTextColor(hdc, fore_color);
    out_text = D04GetOutputText(textstr, txlen);

    if ( unity || Running_Windows95 )
        {
        xpos = ((tx * XPixelsPerInch) / 1000) - Page_X_Offset;
        ypos = ((ty * YPixelsPerInch) / 1000) - Page_Y_Offset;
        }
    else
        {
        /* if we're here, a world transform is in effect, which has
         * already taken the offsets into account.
         */
        xpos = ((tx * XPixelsPerInch) / 1000);
        ypos = ((ty * YPixelsPerInch) / 1000);
        }

    retval = MS_TextOut(hdc, xpos, ypos, out_text, txlen);

    /* cleanup */
    free(out_text);
    MS_RestoreDC(hdc, -1);
    return retval;
}


/************************ TEXT PRIMITIVE *********************************/
int
D04AxText(tx_char *textstr, int txlen, p32 *pos, int width, char *fontName,
          int underPos, int underThick, int strikePos, textAttrs *txProps,
          textAttrs *fieldProps, int xZoom, int yZoom, TmRect32 *clip)
/* p32 *pos;                 mils */
/* int width;                mils */
/* int underPos;             in milliems */
/* int underThick;           in milliems */
/* int strikePos;            in mils */
/* textAttrs *txProps;       properties of this segment of text */
/* textAttrs *fieldProps;    props of the overall field of text */
/* int xZoom, yZoom;         zoom factor relative to ref val 100 */
{
    HDC hdc;
    BOOL retval;
    HFONT hfont, old_font;
    int tx, ty, unity, new_tx, new_ty;
    int font_error = FALSE, did_special_fill;
    fontTabEnt *ftab;
    textAttrs txprops, fldprops, *txp, *fld;
    axform xform, rulxform;
    axformTerms terms;
    BOOL xformed = FALSE;
    attr_type *attr;
    char *errString;
    HPALETTE old_palette;
    COLORREF fore_color;
    int red, green, blue;
    unsigned long pixel;
    XFORM ms_xform;
    float pointsize;


    if ( txlen <= 0 || width <= 0 )
        return FALSE;

    D04DoClip(clip);

    tx = pos->x;
    ty = pos->y;
    pointsize = (float) txProps->pointsize / 10000.0;

    D04SetPos(tx, ty);

    /* look up the font record; select the font */
    if ( errString = axGetFontRecord(fontName, &ftab) )
        {
        freeErrString(errString);
        font_error = TRUE;
        }

    retval = FALSE;
    hdc = PF.PfPrinterHDC;
    if ( PF.PfColordev && PrinterSupportsPalette )
        {
        old_palette = SelectAndRealizeAxWinPalette(hdc);
        pixel = D04GetPixelFromCMYK(axCyanComponent(txProps->fillcolor),
                                    axMagentaComponent(txProps->fillcolor),
                                    axYellowComponent(txProps->fillcolor),
                                    axBlackComponent(txProps->fillcolor), 255);
        fore_color = PALETTEINDEX(pixel);
        }
    else
        {
        D04GetRGBFromCMYK(axCyanComponent(txProps->fillcolor),
                        axMagentaComponent(txProps->fillcolor),
                        axYellowComponent(txProps->fillcolor),
                        axBlackComponent(txProps->fillcolor),
                        &red, &green, &blue);
        fore_color = RGB(red, green, blue);
        }
    /* Since some drivers won't give us gray values for colors when printing
     * black & white, use black as a sane default so we at least get
     * some text.               -donato
     */
    if ( !D04_Printing_Color && (fore_color != RGB_WHITE) )
        fore_color = RGB_BLACK;

    /* the transform terms arrive in a form relative to our internal,
       y-increasing-downwards coord sys.  Postscript's sys is normal,
       so account for the differences on the transform terms before
       converting to ctm form.  (that is, negate sines and shears) */
    if ( txProps )
        {
        txprops = *txProps;
        txprops.angle = txprops.angle;
        txprops.xshear = txprops.xshear;
        txp = &txprops;
        }
    else
        txp = (textAttrs *) NULL;

    if ( fieldProps )
        {
        fldprops = *fieldProps;
        fldprops.angle = fldprops.angle;
        fldprops.xshear = fldprops.xshear;
        fld = &fldprops;
        }
    else
        fld = (textAttrs *) NULL;
    
    /* should test for zero scale factors..
     * (to guard against a concat of  [0 0 0 0 0 0]) ....
     */
    if ( txp != NULL )
        {
        if ( txp->xscale == 0 )
            txp->xscale = 10000;
        if ( txp->yscale == 0 )
            txp->yscale = 10000;
        }
    if ( fld != NULL )
        {
        if ( fld->xscale == 0 )
            fld->xscale = 10000;
        if ( fld->yscale == 0 )
            fld->yscale = 10000;
        }

    unity = pfGetTextSegTransform(txp, fld, xZoom, yZoom, &xform);
    if ( txProps->ruling || txProps->strike_thru )
        {
        /* get an unmirrored xform for use in drawing underlines;
           underlines won't be drawn under a modified ctm */
        pfGetTextSegTransform(txProps, fieldProps, xZoom, yZoom, &rulxform);
        }

    attr = txProps->fillpattern;
    if ( !unity && Running_WindowsNT )
        {
        /*** Setting world transforms only works on NT ***/
        static int x_mil_offset, y_mil_offset;
        if ( !x_mil_offset )
            {
            x_mil_offset = (1000*Page_X_Offset)/XPixelsPerInch;
            y_mil_offset = (1000*Page_Y_Offset)/YPixelsPerInch;
            }
        /* change the current DC transform. */
        xformed = TRUE;
        /* don't use "pixel resolution" (MM_TEXT mapping mode)
         * when rotating (transforming) text.
         */
        ms_xform.eM11 = xform.forward.ca;
        ms_xform.eM12 = xform.forward.cb;
        ms_xform.eM21 = xform.forward.cc;
        ms_xform.eM22 = xform.forward.cd;
        ms_xform.eDx  = xform.forward.ctx;
        ms_xform.eDy  = xform.forward.cty;

        MSNT_SetGraphicsMode(hdc, GM_ADVANCED);
        MSNT_SetWorldTransform(hdc, &ms_xform);
        /* since the world transform setting will translate  *
         * the (x,y) position, we need to compute new values *
         * to get proper positioning.         -donato        */
        /** (equation comes from documentation on MSNT_SetWorldTransform) **/
        new_tx = ((tx-x_mil_offset) * xform.reverse.ca) +
                    ((ty-y_mil_offset) * xform.reverse.cc) +
                    xform.reverse.ctx;
        new_ty = ((tx-x_mil_offset) * xform.reverse.cb) +
                    ((ty-y_mil_offset) * xform.reverse.cd) +
                    xform.reverse.cty;

        tx = new_tx;
        ty = new_ty;
        }
    else
        pfGetTransformTerms(&xform, &terms);

    if ( !font_error )
        {
        hfont = D04GetFontHandleFromFontRecord(ftab,(int)pointsize,terms.angle);
        if ( hfont )
            {
            did_special_fill = FALSE;
            old_font = MS_SelectObject(hdc, hfont);
            if ( pfEasyTextFill(attr) || D04IsDeviceFont() )
                {
                /* we cannot use the Windows Path API with a device font,
                 * so we have this compromise for now (the Path API gives
                 * us color dithering and pattern fills). This compromise
                 * will give us plain black or white text on non-color
                 * printers, and color text on color printers.    -donato
                 */ 
                retval = D04DrawSimpleText(textstr, txlen, tx, ty,
                                                fore_color, unity);
                }
            else
                {
                if ( (attr->fill_flag && attr->f_pat != 1000) &&
                    (attr->fbg.see_thru || attr->ffg.see_thru) )
                    {
                    if ( attr->f_type == USERDEF_FILL_TYPE &&
                                            attr->f_ddep == 8 )
                        /* if we have a color custom fill pattern, we will
                         * never have a transparency, so skip and do
                         * "normal" fill.
                         */
                        ;
                    else
                        {
                        D04DoTransparentText(attr,textstr,txlen,tx,ty, 
														unity, hfont);
                        did_special_fill = TRUE;
                        }
                    }
                else if ( GRADIENT_FILL(attr->f_type) )
                    {
                    D04DoGradientText(attr, textstr,txlen, tx,ty, unity,hfont);
                    did_special_fill = TRUE;
                    }
                if ( !did_special_fill )
                    retval = D04DrawFilledText(textstr,txlen,tx,ty,attr,unity,
																hfont, FALSE);
                }
            MS_SelectObject(hdc, old_font);
            MS_DeleteObject(hfont);
            }
        else
            {
            D04DebugMessage("D04AxText: could not get font handle!");
            retval = FALSE;
            }
        }
    else
        retval = FALSE;

    /* cleanup. */
    D04UnSetClip();
    if ( PF.PfColordev && PrinterSupportsPalette )
        UnSelectPalette(hdc, old_palette);
    /* reset the DC if it was previously transformed. */
    if ( xformed )
        {
        MSNT_ModifyWorldTransform(hdc, (XFORM*)NULL, MWT_IDENTITY);
        MSNT_SetGraphicsMode(hdc, GM_COMPATIBLE);
        }
    return (int)retval;
}
/*********************** END TEXT PRIMITIVE ****************************/



/************************** LINE PRIMITIVE *****************************/
int
D04Line(int nPolys, attr_type *attr, poly_type *pol, TmRect32 *clip)
{
    HDC hdc;
    BOOL did_special_fill, did_special_line;
    COLORREF lfg_color, lbg_color;

    hdc = PF.PfPrinterHDC;
    D04DoClip(clip);

    D04GetLineColors(attr, &lfg_color, &lbg_color);

    //MS_SetBkMode(hdc, TRANSPARENT);

    /*** do the poly objects, as appropriate. ***/
    /* "special" means gradient or transparent.  *
     * "normal" means everything else.   -donato */
    did_special_fill = FALSE;
    did_special_line = FALSE;
    /* check for special lines */
    if ( GRADIENT_FILL(attr->l_type) )
        {
        D04DoGradientLine(attr, pol, nPolys, lbg_color);
        did_special_line = TRUE;
        }
    else if ( (attr->l_pat != 1000 && attr->l_pat != -1) &&
                (attr->lbg.see_thru || attr->lfg.see_thru) )
        {
        if ( attr->l_type == USERDEF_FILL_TYPE && attr->l_ddep == 8 )
            /* if we have a color custom fill pattern, we will never
             * have a transparency, so skip and do "normal" fill.
             */
            ;
        else
            {
            D04DoTransparentLine(attr, pol, nPolys, lfg_color, lbg_color);
            did_special_line = TRUE;
            }
        }
    /* check for special poly fills. */
    if ( GRADIENT_FILL(attr->f_type) )
        {
        D04DoGradientFill(attr, pol, nPolys);
        did_special_fill = TRUE;
        }
    else if ( !D04IsSolidFill(attr) &&
			(attr->fill_flag && attr->f_pat != 1000) &&
            (attr->fbg.see_thru || attr->ffg.see_thru) )
        {
        if ( attr->f_type == USERDEF_FILL_TYPE && attr->f_ddep == 8 )
            /* if we have a color custom fill pattern, we will never
             * have a transparency, so skip and do "normal" fill.
             */
            ;
        else
            {
            D04DoTransparentPoly(attr, pol, nPolys);
            did_special_fill = TRUE;
            }
        }

    /* do the rest ("normal") as necessary. */
    if ( did_special_fill && !did_special_line )
        D04DoNormalLine(hdc, attr, pol, nPolys);
    else if ( did_special_line && !did_special_fill )
        D04DoNormalFill(hdc, attr, pol, nPolys);
    else if ( !did_special_fill && !did_special_line )
        D04DoNormalLineAndFill(hdc, attr, pol, nPolys);
    /* else both fill and line are "special" and we are done. */


    /* cleanup */
    D04UnSetClip();
    /* finish */
    PF.cur_thick = MINIMUM_LINE_WIDTH;
    PF.cur_style = TM_SOLID;
    return 0;
}
/********************** END LINE PRIMITIVE ****************************/



static void
D04DrawPoly(HDC hdc, poly_type *pol, int nPolys, int drawing_to_mem_dc,
                                                    int xpos, int ypos)
{
    int retval;
    int a, b, c, i, j;
    int nPoints, nCtrls;
    POINT *msPoints;

    /* do polygons. */
    for ( i=0; i < nPolys; i++ )
        {
        nPoints = pol[i].nPoints;
        nCtrls = pol[i].nCtrls;
        if ( nCtrls == 0 )    /* draw a simple poly */
            {
            /* make y-axis values negative. */
            msPoints = (POINT*)malloc(nPoints*sizeof(POINT));
            /* convert XPoints to POINTs */
            for ( j=0; j < nPoints; j++ )
                {
                msPoints[j].x = pol[i].points[j].x - xpos;
                msPoints[j].y = pol[i].points[j].y - ypos;
                }
            if ( drawing_to_mem_dc )
                {
				if ( xpos == 0 && ypos == 0 )
                	D04ConvertMilsToPrinterRes(msPoints, nPoints, TRUE);
				else
                	D04ConvertMilsToPrinterRes(msPoints, nPoints, FALSE);
                /* Polygon uses the currently selected brush, which
                 * functionality we want in the mem_dc case.
                 */
                retval = MS_Polygon(hdc, msPoints, nPoints);
                }
            else
                {
                D04ConvertMilsToPrinterRes(msPoints, nPoints, TRUE);
                retval = MS_Polyline(hdc, msPoints, nPoints);
                }
            }
        else    /* draw a curve */
            {
            /* draw a bezier spline. */
            msPoints = (POINT*)malloc((nCtrls+nPoints)*sizeof(POINT));
            nPoints = (nPoints/2) + 1;
            if ( nPoints == pol[i].nPoints )
                nPoints--;
            /* start point */
            msPoints[0].x = pol[i].points[0].x;
            msPoints[0].y = pol[i].points[0].y;
            /* end points are at every 3rd point in the array, *
             * with 2 control points between each.             */
            for ( a=0, b=1, c=0; c < nPoints; a+=3, b+=2, c++ )
                {
                /* control points. */
                msPoints[a+1].x = pol[i].controls[b-1].x;
                msPoints[a+1].y = pol[i].controls[b-1].y;
                msPoints[a+2].x = pol[i].controls[b].x;
                msPoints[a+2].y = pol[i].controls[b].y;
                /* end point. */
                msPoints[a+3].x = pol[i].points[b].x;
                msPoints[a+3].y = pol[i].points[b].y;
                }
            /* final points array should look like this:
             * [start pt] [ctl pt] [ctl pt] [end pt]
             * [ctl pt] [ctl pt] [end pt] ...
             */
            D04ConvertMilsToPrinterRes(msPoints, nCtrls + nPoints, TRUE);
            retval = MS32_PolyBezier(hdc, msPoints, nCtrls + nPoints);
            }
        /* loop cleanup. */
        free(msPoints);
        }
}


/***************************************************************/
int D04ClrBox()    { D04DebugMessage("ClrBox not ported!"); return 0; }
int D04SetColor()  { D04DebugMessage("SetColor not ported!"); return 0; }
int D04Save()      { /* NO OP for GDI printing? */ return 0; }
int D04Restore()   { /* NO OP for GDI printing? */ return 0; }
int D04Landscape() { D04DebugMessage("Landscape not ported!"); return 0; }
/***************************************************************/




int
D04Printmarks(int use_crops, int use_regis, int use_bounds, TmRect32 *bounds)
{
    HDC hdc;
    HPEN black_pen, white_pen, old_pen;
    HBRUSH color_brush, old_color_brush;
    HBRUSH null_brush, black_brush, old_brush;
    int i, edge_x1, edge_y1, edge_x2, edge_y2, page_wid, page_hyt;
    int margin, radius;
    TmRect32 pf_bounds;
    POINT pts[4];


    hdc = PF.PfPrinterHDC;
    if ( use_crops || use_regis || use_bounds )
        {
        black_pen = MS_CreatePen(PS_SOLID, 3, RGB_BLACK);
        white_pen = MS_CreatePen(PS_SOLID, 3, RGB_WHITE);
        null_brush = MS_GetStockObject(NULL_BRUSH);
        black_brush = MS_GetStockObject(BLACK_BRUSH);
        /* set the line color to black */
        old_pen = MS_SelectObject(hdc, black_pen);
        old_brush = MS_SelectObject(hdc, null_brush);

        pts[0].x = bounds->x_position;
        pts[0].y = bounds->y_position;
        pts[1].x = bounds->x_size;
        pts[1].y = bounds->y_size;
        pts[2].x = PF.PfPageMilWidth;
        pts[2].y = PF.PfPageMilHeight;
        pts[3].x = MARK_MARGIN;
        pts[3].y = 0;
        D04ConvertMilsToPrinterRes(pts, 4, FALSE);
        pf_bounds.x_position = pts[0].x - Page_X_Offset;
        pf_bounds.y_position = pts[0].y - Page_Y_Offset;
        pf_bounds.x_size = pts[1].x;
        pf_bounds.y_size = pts[1].y;
        page_wid = pts[2].x;
        page_hyt = pts[2].y;
        margin = pts[3].x;

        if ( use_bounds )
            {
            edge_x1 = pf_bounds.x_position;
            edge_y1 = pf_bounds.y_position;
            edge_x2 = edge_x1 + pf_bounds.x_size;
            edge_y2 = edge_y1 + pf_bounds.y_size;
            MS_Rectangle(hdc, edge_x1, edge_y1, edge_x2, edge_y2);
            }

        if ( use_crops )
            {
            /* top crops */
            if ( bounds->y_position > 250 )
                {
                edge_x1 = pf_bounds.x_position;
                edge_y1 = margin;
                edge_x2 = pf_bounds.x_position + pf_bounds.x_size;  
                edge_y2 = pf_bounds.y_position - margin;
                /* left */
                MS_MoveToEx(hdc, edge_x1, edge_y1, NULL);
                MS_LineTo(hdc, edge_x1, edge_y2);
                /* right */
                MS_MoveToEx(hdc, edge_x2, edge_y1, NULL);
                MS_LineTo(hdc, edge_x2, edge_y2);
                }

            /* right crops */
            if ( PF.PfPageMilWidth-(bounds->x_position+bounds->x_size) > 250 )
                {
                edge_x1 = pf_bounds.x_position + pf_bounds.x_size + margin;
                edge_y1 = pf_bounds.y_position;
                edge_x2 = page_wid - margin;
                edge_y2 = pf_bounds.y_position + pf_bounds.y_size;
                /* top */
                MS_MoveToEx(hdc, edge_x1, edge_y1, NULL);
                MS_LineTo(hdc, edge_x2, edge_y1);
                /* right */
                MS_MoveToEx(hdc, edge_x1, edge_y2, NULL);
                MS_LineTo(hdc, edge_x2, edge_y2);
                }
 
            /* bottom crops */
            if ( PF.PfPageMilHeight-(bounds->y_position+bounds->y_size) > 250 )
                {
                edge_x1 = pf_bounds.x_position;
                edge_y1 = pf_bounds.y_position + pf_bounds.y_size + margin;
                edge_x2 = pf_bounds.x_position + pf_bounds.x_size;
                edge_y2 = page_hyt - margin;
                /* right */
                MS_MoveToEx(hdc, edge_x2, edge_y1, NULL);
                MS_LineTo(hdc, edge_x2, edge_y2);
                /* left */
                MS_MoveToEx(hdc, edge_x1, edge_y1, NULL);
                MS_LineTo(hdc, edge_x1, edge_y2);
                }
 
            /* left crops */
            if ( bounds->x_position > 250 )
                {
                edge_x1 = margin;
                edge_y1 = pf_bounds.y_position;
                edge_x2 = pf_bounds.x_position - margin;
                edge_y2 = pf_bounds.y_position + pf_bounds.y_size;
                /* bottom */
                MS_MoveToEx(hdc, edge_x1, edge_y2, NULL);
                MS_LineTo(hdc, edge_x2, edge_y2);
                /* top */
                MS_MoveToEx(hdc, edge_x1, edge_y1, NULL);
                MS_LineTo(hdc, edge_x2, edge_y1);
                }
            }

        if ( use_regis )
            {
            /* left mark */
            if ( bounds->x_position > 250 )
                {
                edge_x1 = pf_bounds.x_position / 2;
                edge_y1 = page_hyt / 2;

                /* draw the registration circle */
                /* we have center(x,y) and radius, so bounding box is:
                 * x1 = cx - r, y1 = cy - r, x2 = cx + r, y2 = cy + r
                 */
                radius = margin;
                MS_Arc(hdc, (edge_x1-radius), edge_y1-radius,
                        (edge_x1+radius), edge_y1+radius, 0,0,0,0);

                MS_MoveToEx(hdc, edge_x1 - (margin+10), edge_y1, NULL);
                MS_LineTo(hdc, edge_x1 + (margin+10), edge_y1);

                MS_MoveToEx(hdc, edge_x1, edge_y1 - (margin+10), NULL);
                MS_LineTo(hdc, edge_x1, edge_y1 + (margin+10));

                radius = (int)(margin/2.0);
                MS_SelectObject(hdc, black_brush);
                /* filled arc */
                MS_Pie(hdc, (edge_x1-radius), edge_y1-radius,
                        (edge_x1+radius), edge_y1+radius, 0,0,0,0);

                MS_SelectObject(hdc, white_pen);
                MS_MoveToEx(hdc,(int)(edge_x1 - margin/2.0), edge_y1, NULL);
                MS_LineTo(hdc, (int)(edge_x1 + margin/2.0), edge_y1);
                MS_MoveToEx(hdc,edge_x1, (int)(edge_y1 - margin/2.0), NULL);
                MS_LineTo(hdc, edge_x1, (int)(edge_y1 + margin/2.0));
                }

            /* right mark */
            if ( PF.PfPageMilWidth-(bounds->x_position+bounds->x_size) > 250 )
                {
                edge_x2 = (page_wid-(pf_bounds.x_position+pf_bounds.x_size))/2.0;
                edge_x1 = pf_bounds.x_position + pf_bounds.x_size + edge_x2;
                edge_y1 = page_hyt / 2;

                /* draw the registration circle */
                radius = margin;
                MS_SelectObject(hdc, black_pen);
                MS_SelectObject(hdc, null_brush);
                MS_Arc(hdc, (edge_x1-radius), edge_y1-radius,
                        (edge_x1+radius), edge_y1+radius, 0,0,0,0);

                MS_MoveToEx(hdc, edge_x1 - (margin+10), edge_y1, NULL);
                MS_LineTo(hdc, edge_x1 + (margin+10), edge_y1);

                MS_MoveToEx(hdc, edge_x1, edge_y1 - (margin+10), NULL);
                MS_LineTo(hdc, edge_x1, edge_y1 + (margin+10));

                radius = (int)(margin/2.0);
                MS_SelectObject(hdc, black_brush);
                MS_Pie(hdc, (edge_x1-radius), edge_y1-radius,
                        (edge_x1+radius), edge_y1+radius, 0,0,0,0);

                MS_SelectObject(hdc, white_pen);
                MS_MoveToEx(hdc,(int)(edge_x1 - margin/2.0), edge_y1, NULL);
                MS_LineTo(hdc, (int)(edge_x1 + margin/2.0), edge_y1);
                MS_MoveToEx(hdc,edge_x1, (int)(edge_y1 - margin/2.0), NULL);
                MS_LineTo(hdc, edge_x1, (int)(edge_y1 + margin/2.0));

                /* draw the color boxes */
                margin = margin * 2;
                edge_x1 -= margin/2;
                edge_y1 -= margin * 2;
                MS_SelectObject(hdc, black_pen);
                for ( i=0; i < 10; i++ )
                    {
                    color_brush = MS_CreateSolidBrush(gray_values[i]);
                    old_color_brush = MS_SelectObject(hdc, color_brush);
                    MS_Rectangle(hdc, edge_x1, edge_y1, edge_x1 + margin,
                                                    edge_y1 + margin);
                    edge_y1 -= margin;
                    MS_SelectObject(hdc, old_color_brush);
                    MS_DeleteObject(color_brush);
                    }

                /* draw the registration string */
                edge_x2 = PF.PfPageMilWidth - (bounds->x_position +
                                                    bounds->x_size);
                edge_x1 = bounds->x_position + bounds->x_size +
                                                    (edge_x2 / 2.5);
                edge_y1 = (PF.PfPageMilHeight / 2.0) + (3 * MARK_MARGIN);
          		pf_display_reg_string(NXLT("Arial"), edge_x1, edge_y1);
                }
            }

        /* cleanup */
        MS_SelectObject(hdc, old_pen);
        MS_SelectObject(hdc, old_brush);
        MS_DeleteObject(black_pen);
        MS_DeleteObject(white_pen);
        /* the brushes are stock objects, so don't delete. */
        }
    return 0;
}





/*************************************************************************
 *
 *  D01SetClip - sets the clipping in the print graphics context 
 * 
************************************************************************/
int
D04SetClip(AxRectangle *clipRect, int x, int y, int last_x, int last_y)
/* AxRectangle *clipRect;       current app clip */
/* int x, y, last_x, last_y;    new appclip requested */
{
    int retval;
    TmRect32 clip;

    retval = FALSE;
    /* set app clip */
    if ( (x != clipRect->origin.x) ||
         (y != clipRect->origin.y) ||
         (last_x != clipRect->corner.x) ||
         (last_y != clipRect->corner.y) )
        {
        PF.PfPage.appClip.origin.x = x;
        PF.PfPage.appClip.origin.y = y;
        PF.PfPage.appClip.corner.x = last_x;
        PF.PfPage.appClip.corner.y = last_y;

        PF.left = x;
        PF.top = y;
        PF.right = last_x;
        PF.bottom = last_y;

        if ( PF.ClipGSaveLevel )    /* should this be a while() ? */
            PF.ClipGSaveLevel--;

        PF.ClipGSaveLevel++;
        PF.cur_thick = MINIMUM_LINE_WIDTH;          /* default page setup */
        PF.cur_style = TM_SOLID;                    /* default page setup */

        clip.x_position = PF.left;
        clip.y_position = PF.top;
        clip.x_size = PF.right - PF.left;
        clip.y_size = PF.bottom - PF.top;
        D04DoClip(&clip);
        retval = TRUE;
        }
    return retval;
}

static void
D04SetClipFromPoly(HDC hdc, poly_type *pol, int nPolys)
{
    MS32_BeginPath(hdc);
    D04DrawPoly(hdc, pol, nPolys, FALSE, 0, 0);
    /* make sure shape is closed, otherwise we'll try to set an open
     * clipping region which causes BAD things to happen.
     * (on the DEC Alpha i was forced to reboot!)   -donato
     */
    MS32_CloseFigure(hdc);
    MS32_EndPath(hdc);
    if ( !MS32_SelectClipPath(hdc, RGN_COPY) )
        D04PathAbort();
}


static int
D04IsDeviceFont(void)
{
    int retval;
    HDC hdc;
    TEXTMETRIC tm;

    /** ASSUMPTION:  a font is currently selected into the printer hdc. */
    hdc = PF.PfPrinterHDC;
    retval = FALSE;
    MS_GetTextMetrics(hdc, &tm);
#if 0
    if ( !(tm.tmPitchAndFamily & TMPF_VECTOR) &&
            !(tm.tmPitchAndFamily & TMPF_TRUETYPE) &&
            (tm.tmPitchAndFamily & TMPF_DEVICE) )
#else
    if ( !(tm.tmPitchAndFamily & TMPF_TRUETYPE) &&
            (tm.tmPitchAndFamily & TMPF_DEVICE) )
#endif
        retval = TRUE;

    return retval;
}


/*** CALLER must DELETE font handle!!! ***/
static HFONT
D04GetFontHandleFromFontRecord(fontTabEnt *ftab, int pointsize, int angle)
{
    LOGFONT logfont;
    HFONT win_font;
    char *family_name, *err;
    fontTabEnt *trec, *screenRecord;


    memset(&logfont, 0, sizeof(logfont));
    /* Note: a negative value for 'lfHeight' forces "real" size as opposed */
    /*       to a size within a line spacing of the requested size      */
    logfont.lfHeight = (LONG) -((pointsize * YPixelsPerInch)/72);
    if ( ftab && ftab->WindowsAlternate && !ftab->isWindowsFont &&
                                            !ftab->isPrinterFont )
        {
        int twidth;
        /* an example of familyName in this context might be "Palatino".  */
        /* the alternate is provided for purposes of successfully finding */
        /* a screenfont to use; an example might be "Book Antiqua"        */
        /* we'll create an instance of the alternate's screen font,       */
        /* but width-modified based on the target's average width         */
        /* specify a lfWidth value based on the target font information. */
#if 0
        if ( PrinterIsPostScript )
            {
            twidth = axGetAlternateAverageWidth(ftab, PS_CLASS_NUM);
            logfont.lfWidth = (LONG) mToPix(twidth, (float)pointsize,
                                                    XPixelsPerInch);
            family_name = ftab->PostScriptAlternate;
            }
        else
            {
            twidth = axGetAlternateAverageWidth(ftab, PCL_CLASS_NUM);
            logfont.lfWidth = (LONG) mToPix(twidth, (float)pointsize,
                                                    XPixelsPerInch);
            family_name = ftab->PCLAlternate;
            }
#endif

        //if ( !family_name )
        //    {
            twidth = axGetAlternateAverageWidth(ftab);
            logfont.lfWidth = (LONG) mToPix(twidth, (float)pointsize,
                                                    XPixelsPerInch);
            family_name = ftab->WindowsAlternate;
            if ( !IsWinFont(ftab->WindowsAlternate) )
                family_name = NXLT("Arial");
         //   }
        }
    else
        family_name = ftab->familyName;

    if ( ftab->weight )
       logfont.lfWeight = FW_BOLD;   /* bold */
    if ( ftab->slant )
       logfont.lfItalic = 1;     /* italic */

    /* On NT, we set a world transform to get scaling, shearing, rotation
     * (see D04AxText). On Win95, we set the orientation on the font here.
     */
    if ( !Running_WindowsNT )
        logfont.lfEscapement = logfont.lfOrientation = (long)(angle / 1000);

    /* lookup charset information based on actual screenfont */
    err = axGetFontRecordForScreen(family_name, ftab->weight, ftab->slant,
                                   pointsize, &trec, &screenRecord);
    if ( err )
        freeErrString(err);
    else
        {
        logfont.lfCharSet = screenRecord->charSet;
        logfont.lfPitchAndFamily = screenRecord->pitchAndFamily;
        }
    strcpy(logfont.lfFaceName, family_name);

    win_font = MS_CreateFontIndirect(&logfont);
    return win_font;
}


static char *
D04GetOutputText(register tx_char *instr, register int txlen)
{
    int malloc_len = 0;
    register int tx;
    register int out_x;
    register unsigned char *outstr = NULL;

    /* scan the string looking for special chars that will
       require the escape char. These will not be accounted for in the
       txlen value */
    malloc_len = txlen * 8;   /* prepare for worst case: all 16-bit chars*/
    outstr = (unsigned char *) malloc (malloc_len + 1);
    for ( tx = 0, out_x=0; tx < txlen; tx++, out_x++ )
        {
        outstr[out_x] = instr[tx];
        }
    outstr[out_x] = '\0';
    return outstr;
}


/*************************************************************************
 * 
 *  accumulate an enclosing bounder for a poly_type instance 
 *  and also the rectangle for the rotated poly
 
    returns:
        box[0] = LLC of rect
        box[1] = URC of rect
        box[2] = LLC of rotated rect
        box[3] = width, height pf rotated rect
************************************************************************/
#define TX(x, y)    ((x-orgx)*cosa - (y-orgy)*sina)
#define TY(x, y)    ((x-orgx)*sina + (y-orgy)*cosa)
#define MIL_TO_DOT(mil) (((mil) * (double)300) / 1000.0)

static void
D04GetBoundingBox(int nPolys, attr_type *attr, poly_type *pol, TmPoint32 *box)
/* int nPolys;         in: */
/* attr_type *attr;    in: need width, ends */
/* poly_type *pol;     in: polygon info */
/* TmPoint32 *box;     out: set this two-element array to enclose poly */
{
    int ix, jx, half_line_width;
    double minx, miny, maxx, maxy, xr, yr;
    double angle, sina, cosa;
    double orgx, orgy;

    /* init boundary */
    box[0].x = box[1].x = (int) (pol[0].points[0].x);
    box[0].y = box[1].y = (int) (pol[0].points[0].y);
    /* scan poly info */
    for ( ix = 0; ix < nPolys; ix++ )
        {
        /* scan points */
        for ( jx = 0; jx < pol[ix].nPoints; jx++ )
            {
            if ( pol[ix].points[jx].x < box[0].x )
                {
                box[0].x = pol[ix].points[jx].x;
                }
            else if ( pol[ix].points[jx].x > box[1].x )
                {
                box[1].x = pol[ix].points[jx].x;
                }
            if ( pol[ix].points[jx].y < box[0].y )
                {
                box[0].y = pol[ix].points[jx].y;
                }
            else if ( pol[ix].points[jx].y > box[1].y )
                {
                box[1].y = pol[ix].points[jx].y;
                }
            }

        /* this is the case for round ends, need to do square, butt */
        half_line_width = attr->l_wid/2;
        box[0].x -= half_line_width;
        box[0].y -= half_line_width;
        box[1].x += half_line_width;
        box[1].y += half_line_width;

            /* scan control points also */
        for ( jx = 0; jx < pol[ix].nCtrls; jx++ )
            {
            if ( pol[ix].controls[jx].x < box[0].x )
                {
                box[0].x = pol[ix].controls[jx].x;
                }
            else if ( pol[ix].controls[jx].x > box[1].x )
                {
                box[1].x = pol[ix].controls[jx].x;
                }
            if ( pol[ix].controls[jx].y < box[0].y )
                {
                box[0].y = pol[ix].controls[jx].y;
                }
            else if ( pol[ix].controls[jx].y > box[1].y )
                {
                box[1].y = pol[ix].controls[jx].y;
                }
            }
        }

    /* now find the rotated box */
    if ( attr->f_ang == 0 )
        {
        box[2].x = 0;
        box[2].y = 0;
        box[3].x = ABS(box[1].x - box[0].x);
        box[3].y = ABS(box[1].y - box[0].y);
        return;
        }

    angle = ((double)(-attr->f_ang) * M_PI) / 180.0;
    sina = sin(angle);
    cosa = cos(angle);
    orgx = box[0].x;
    orgy = box[0].y;

    /* init boundary */
    minx = maxx = TX(box[0].x, box[0].y) ;
    miny = maxy = TY(box[0].x, box[0].y) ;

    xr = TX(box[0].x, box[1].y) ;
    yr = TY(box[0].x, box[1].y) ;

    if (xr < minx)
        minx = xr ;
    else if (xr > maxx)
        maxx = xr ;
    if (yr < miny)
        miny = yr ;
    else if (yr > maxy)
        maxy = yr ;

    xr = TX(box[1].x, box[1].y) ;
    yr = TY(box[1].x, box[1].y) ;

    if (xr < minx)
        minx = xr ;
    else if (xr > maxx)
        maxx = xr ;
    if (yr < miny)
        miny = yr ;
    else if (yr > maxy)
        maxy = yr ;

    xr = TX(box[1].x, box[0].y) ;
    yr = TY(box[1].x, box[0].y) ;

    if (xr < minx)
        minx = xr ;
    else if (xr > maxx)
        maxx = xr ;
    if (yr < miny)
        miny = yr ;
    else if (yr > maxy)
        maxy = yr ;

    box[2].x = minx;
    box[2].y = miny;
    box[3].x = maxx + 0.5;
    box[3].y = maxy + 0.5;
}

#if 0   /* { */
static void
D04GetTextBoundingBox(attr_type *attr, int xpos, int ypos, char *str, int len,
                                                            TmPoint32 *box)
{
    HDC hdc = PF.PfPrinterHDC;
    SIZE size;
    TEXTMETRIC tm;
    int ty;    /* tmp val */
    double minx, miny, maxx, maxy, xr, yr;
    double angle, sina, cosa;
    double orgx, orgy;

    /* assume there is a font selected into the printer HDC;
     * get the dimensions.
     */
    MS_GetTextMetrics(hdc, &tm);
    MS_GetTextExtentPoint32(hdc, str, len, &size);

    box[0].x = xpos;
    box[0].y = ypos - tm.tmDescent;
    box[1].x = xpos + size.cx;// + Page_X_Offset;
    box[1].y = ypos + tm.tmAscent;

    if ( attr->f_ang == 0 )
        {
        ty = box[0].y;
        box[0].y = ty;
        ty = box[1].y;
        box[1].y = ty;
#if 0
        box[0].x -= Page_X_Offset;
        box[1].x -= Page_X_Offset;
        ty = box[0].y;
        box[0].y = Y(ty - Page_Y_Offset);
        ty = box[1].y;
        box[1].y = Y(ty - Page_Y_Offset);
#endif
        box[2].x = 0;
        box[2].y = 0;
        box[3].x = ABS(box[1].x - box[0].x);
        box[3].y = ABS(box[1].y - box[0].y);
        return;
        }

    /* find the rotated box */
    angle = ((double)(-attr->f_ang) * M_PI) / 180.0;
    sina = sin(angle);
    cosa = cos(angle);
    orgx = box[0].x;
    orgy = box[0].y;

    /* init boundary */
    minx = maxx = TX(box[0].x, box[0].y) ;
    miny = maxy = TY(box[0].x, box[0].y) ;

    xr = TX(box[0].x, box[1].y) ;
    yr = TY(box[0].x, box[1].y) ;

    if (xr < minx)
        minx = xr ;
    else if (xr > maxx)
        maxx = xr ;
    if (yr < miny)
        miny = yr ;
    else if (yr > maxy)
        maxy = yr ;

    xr = TX(box[1].x, box[1].y) ;
    yr = TY(box[1].x, box[1].y) ;

    if (xr < minx)
        minx = xr ;
    else if (xr > maxx)
        maxx = xr ;
    if (yr < miny)
        miny = yr ;
    else if (yr > maxy)
        maxy = yr ;

    xr = TX(box[1].x, box[0].y) ;
    yr = TY(box[1].x, box[0].y) ;

    if (xr < minx)
        minx = xr ;
    else if (xr > maxx)
        maxx = xr ;
    if (yr < miny)
        miny = yr ;
    else if (yr > maxy)
        maxy = yr ;

    box[2].x = minx;
    box[2].y = miny;
    box[3].x = ceil(maxx);
    box[3].y = ceil(maxy);
}
#endif  /* } #if 0 */


/********************************************************************
 * Creates a GDI clip region and selects it into the printer HDC.
 * the object handle is saved so the caller can delete it.
 ********************************************************************/
int
D04DoClip(TmRect32 *clip)
{
    HRGN clip_rgn;
    POINT points[2];

    if ( PF.PfClipRgn )
        D04UnSetClip();
    if ( clip && (clip->x_size != 0) && (clip->y_size != 0) )
        {
        points[0].x = clip->x_position;
        points[0].y = clip->y_position;
        points[1].x = points[0].x + clip->x_size;
        points[1].y = points[0].y + clip->y_size;
        D04ConvertMilsToPrinterRes(points, 2, TRUE);
        clip_rgn = MS_CreateRectRgn(points[0].x, points[0].y,
                                    points[1].x, points[1].y);

        if ( clip_rgn )
            {
            MS_SelectClipRgn(PF.PfPrinterHDC, clip_rgn);
            PF.PfClipRgn = clip_rgn;
            }
        else
            D04DebugMessage("D04DoClip: clip failed!");
        }
    return 0;
}

static void
D04UnSetClip(void)
{
    MS_SelectClipRgn(PF.PfPrinterHDC, (HRGN)NULL);
    if ( PF.PfClipRgn )
        {
        MS_DeleteObject(PF.PfClipRgn);
        PF.PfClipRgn = NULL;
        }
}


/*************************************************************************
 *
 * DoFillPattern fills a bounded region. Any clipping is setup prior to
 * calling the function.
 
 * On Windows we obtain a brush handle to perform the fill with.
 * 
************************************************************************/
static HBRUSH
D04GetPatternBrush(attr_type *attr, COLORREF fore_color, COLORREF back_color,
                                                            int *stock_brush)
{
    HBRUSH hbrush;
	int gray_percent;

    /* this does not do gradient fill types. 
     * must call separate function for those (see D04DoGradientFill).
     */
    *stock_brush = FALSE;
    switch (attr->f_type)
        {
        case BUILTIN_FILL_TYPE:
        case USERDEF_FILL_TYPE:
            if ( attr->f_type == BUILTIN_FILL_TYPE && attr->f_pat == -1 )
                {
                hbrush = (HBRUSH)MS_GetStockObject(NULL_BRUSH);
                *stock_brush = TRUE;
                }
            else
                {
				switch (attr->f_pat)
					{
        			case 0:
        			case 12:
        			case 1:
        			case 10:
        			case 2:
        			case 21:
        			case 3:
        			case 13:
        			case 4:
						if ( attr->f_type == BUILTIN_FILL_TYPE )
							{
        			    	gray_percent = pfGetPermilFill(attr->f_pat);
							hbrush = D04GetGrayFillBrush(attr, gray_percent);
							break;
							}
						/* otherwise fall through */
					default:
                		hbrush = D04GetFillBrush(attr);
					}
                }
            break;

        case GRAY_FILL_TYPE:
            if ( attr->f_pat == 1000 )
                /* 100% foreground fill. */
                hbrush = MS_CreateSolidBrush(fore_color);
            else if ( attr->f_pat == 0 )
                /* 0% foreground fill. */
                hbrush = MS_CreateSolidBrush(back_color);
            else
                {
                int ix;
                /* get percentage bitmap and create brush from that. */
                ix = getIXfromPercent(attr->f_pat);
                hbrush = MS_CreatePatternBrush(D04GetStippleMap(ix, 0));
                }
            break;

        default:
            /* unknown fill type, throw a fatal?... */
            hbrush = (HBRUSH)MS_GetStockObject(NULL_BRUSH);
            *stock_brush = TRUE;
            break;
        }

    return hbrush;
}

static HBRUSH
D04GetLinePatternBrush(attr_type *attr, COLORREF fg, COLORREF bg,
                                                int *stock_brush)
{
    int inv, delete_hbm;
    HBRUSH hbrush;
    HBITMAP hbm;


    *stock_brush = FALSE;
    delete_hbm = FALSE;
    if ( (attr->l_type == BUILTIN_FILL_TYPE) &&
        (attr->l_pat != 4 && attr->l_pat != 0) )
        {
        /* set the map to fill the poly */
        inv = attr->lfg.see_thru ? TRUE : FALSE;
        hbm = D04GetStippleMap(attr->l_pat, inv);
        }
    else if ( attr->l_type == USERDEF_FILL_TYPE )
        {
        delete_hbm = TRUE;
        /* we need to create a bitmap handle from the image data. */
        if ( attr->l_ddep == 8 )
            {
            hbm = D04GetDIBFromImageData(attr->l_data, attr->l_dwid,
                                        attr->l_dhyt, attr->l_cmap);
            }
        else
            {
            hbm = D04GetMonoBitmapFromImageData(attr->l_data, attr->l_dwid,
                                                    attr->l_dhyt, fg, bg);
            }
        }
    else if ( attr->l_type == GRAY_FILL_TYPE )
        {
        if ( attr->l_pat != 1000 && attr->l_pat != 0 && attr->l_pat != -1 )
            {
            int tmp_ix;
            /* get percentage bitmap and create brush from that. */
            tmp_ix = getIXfromPercent(attr->l_pat);
            hbm = D04GetStippleMap(tmp_ix, 0);
            }
        }
    hbrush = MS_CreatePatternBrush(hbm);
    if ( delete_hbm )
        MS_DeleteObject(hbm);
    return hbrush;
}


static HBITMAP
D04GetPatternBitmap(attr_type *attr)
{
    int inv;
    HBITMAP fill_map;

    inv = attr->ffg.see_thru ? TRUE : FALSE;
    if ( attr->f_type == BUILTIN_FILL_TYPE )
        {
        /* get the pre-loaded bitmap handle. */
        fill_map = D04GetStippleMap(attr->f_pat, inv);
        }
    else if ( attr->f_type == USERDEF_FILL_TYPE )
        {
        /* we need to create a bitmap handle from the image data. */
        if ( attr->f_ddep == 8 )
            {
            fill_map = D04GetDIBFromImageData(attr->f_data, attr->f_dwid,
                                            attr->f_dhyt, attr->f_cmap);
            }
        else
            {
            COLORREF fg, bg;
            D04GetForeAndBackColors(attr, &fg, &bg);
            fill_map = D04GetMonoBitmapFromImageData(attr->f_data,
                                        attr->f_dwid, attr->f_dhyt,
                                        fg, bg);
            }
        }
    else if ( attr->f_type == GRAY_FILL_TYPE )
        {
        /*** There is an assumption here: that we have called this function
            * knowing we need a pattern bitmap. In the case where we have a
            * 100 or 0 percent "gray" fill pattern, we should not be calling
            * this function.                -donato
         ***/
        int ix;
        /* return a "percentage" bitmap. */
        ix = getIXfromPercent(attr->f_pat);
        fill_map = D04GetStippleMap(ix, 0);
        }
    return fill_map;
}

/** this gives us the printer pattern, i.e. a bitmap with printer resolution.
 **/
static HBRUSH
D04GetFillBrush(attr_type *attr)
{ 
    HBRUSH hbrush;
    HBITMAP fill_map;

    fill_map = D04GetPatternBitmap(attr);
    hbrush = MS_CreatePatternBrush(fill_map);
    if ( attr->f_type == USERDEF_FILL_TYPE )
        MS_DeleteObject(fill_map);
    return hbrush;
}

static HBRUSH
D04GetGrayFillBrush(attr_type *attr, int gray_percent)
{ 
    HBRUSH hbrush;
	COLORREF gray_color;

	gray_color = D04GetGrayColor(attr, gray_percent);
	hbrush = MS_CreateSolidBrush(gray_color);
    return hbrush;
}

static COLORREF
D04GetGrayColor(attr_type *attr, int gray_percent)
{
	int gray_factor;
	int red, green, blue;
	int c_diff, m_diff, y_diff, k_diff;

    c_diff = attr->fbg.c - attr->ffg.c;
    m_diff = attr->fbg.m - attr->ffg.m;
    y_diff = attr->fbg.y - attr->ffg.y;
    k_diff = attr->fbg.k - attr->ffg.k;

    gray_factor = 1000 - gray_percent;
    D04GetRGBFromCMYK(attr->ffg.c + (c_diff * gray_factor)/1000,
                    attr->ffg.m + (m_diff * gray_factor)/1000,
                    attr->ffg.y + (y_diff * gray_factor)/1000,
                    attr->ffg.k + (k_diff * gray_factor)/1000,
					&red, &green, &blue);
	return RGB(red, green, blue);
}


static HBITMAP
D04GetMonoBitmapFromImageData(unsigned char *attr_data, int wid, int hyt,
                                                COLORREF fg, COLORREF bg)
{
    HDC hdc, pix_dc, src_dc;
    HBITMAP pix, hbm;
    int new_wid, new_hyt;


    hdc = PF.PfPrinterHDC;
    new_wid = wid * (PrinterResolution/75);
    new_hyt = hyt * (PrinterResolution/75);

    /* create bitmap handle from given (screen) data. */
    hbm = MS_CreateBitmap(wid, hyt, 1, 1, attr_data);

    pix_dc = MS_CreateCompatibleDC(hdc);
    pix = MS_CreateCompatibleBitmap(hdc, new_wid, new_hyt);
    MS_SelectObject(pix_dc, pix);

    src_dc = MS_CreateCompatibleDC(hdc);
    MS_SetBkColor(src_dc, fg);
    MS_SetTextColor(src_dc, bg);
    MS_SelectObject(src_dc, hbm);

    MS_SetBkColor(pix_dc, fg);
    MS_SetTextColor(pix_dc, bg);
    /* stretch the bitmap for printer resolution. */
    MS_StretchBlt(pix_dc, 0,0,new_wid,new_hyt, src_dc,0,0,wid,hyt, SRCCOPY);

    /* cleanup */
    MS_DeleteObject(hbm);
    MS_DeleteDC(src_dc);
    MS_DeleteDC(pix_dc);

    return pix;
}

static HBITMAP
D04GetDIBFromImageData(unsigned char *attr_data, int wid, int hyt,
                                                    cmap_type *cmap)
{
    HDC hdc, pix_dc;
    HBITMAP pix;
    int new_wid, new_hyt, scan_lines;
    unsigned char *data, *sourceP, *destP;
    int i, j, bytewid, win_bytewid, data_size;
    BITMAPINFO *bitmap_info;


    hdc = PF.PfPrinterHDC;
    new_wid = wid * (PrinterResolution/75);// * (PrinterResolution/300);
    new_hyt = hyt * (PrinterResolution/75);// * (PrinterResolution/300);

    bytewid = wid + wid%2;
    win_bytewid = ((wid + 3) / 4) * 4;
    data_size = win_bytewid * hyt;
    data = (unsigned char *)malloc(data_size*sizeof(unsigned char));
    sourceP = attr_data;
    /* flip the data for the MS_StretchDIBits call below. */
    for ( i = 0; i < hyt; i++ )
        {
        destP = data + ((hyt - i - 1) * win_bytewid);
        memset(destP, 0, win_bytewid);
        for ( j = 0; j < bytewid; j++ )
            *destP++ = *sourceP++;
        }

    pix_dc = MS_CreateCompatibleDC(hdc);
    pix = MS_CreateCompatibleBitmap(hdc, new_wid, new_hyt);
    MS_SelectObject(pix_dc, pix);

    bitmap_info = AxGetDIBInfo(wid, hyt, cmap, AX_DIB_NORMAL);
    /* stretch the image data for printer resolution. */
    scan_lines = MS_StretchDIBits(pix_dc, 0, 0, new_wid, new_hyt,
                                0, 0, wid, hyt, data, bitmap_info,
                                DIB_RGB_COLORS, SRCCOPY);
    if ( scan_lines < hyt )
        D04DebugMessage("D04GetDIBFromImageData: error drawing image");

    /* cleanup */
    free(data);
    free(bitmap_info);
    MS_DeleteDC(pix_dc);
    return pix;
}


static HPEN
D04GetDrawingPen(attr_type *attr, COLORREF fg, COLORREF bg)
{
    int ix, dash_list_size;
    int inv, width, pen_width, pen_style;
    BOOL use_geometric_pen = FALSE;
    BOOL delete_fill_map = FALSE;
    HPEN hpen;
    LOGBRUSH lbrush;
    HBITMAP fill_map;
    DWORD *dashes;
    COLORREF color;


    /* line width is in mils; convert to printer resolution,
     * or set to minimum.
     */
    if ( attr->l_wid > 10 )
        width = (attr->l_wid * XPixelsPerInch) / 1000;
    else
        width = MINIMUM_LINE_WIDTH;
    fill_map = NULL;

    /* normal solid pens have a default cap style of CapRound.   *
     * to get anything other than that, we need to use a Windows *
     * GEOMETRIC style pen.               -donato                */
    if ( attr->l_cap != TM_CAP_ROUND )
        use_geometric_pen = TRUE;

    /* see if we have some pattern fills. */
    if ( (attr->l_type == BUILTIN_FILL_TYPE) &&
        (attr->l_pat != 4 && attr->l_pat != 0) )
        {
        use_geometric_pen = TRUE;
        /* set the map to fill the poly */
        inv = attr->lfg.see_thru ? TRUE : FALSE;
        fill_map = D04GetStippleMap(attr->l_pat, inv);
        }
    else if ( attr->l_type == USERDEF_FILL_TYPE )
        {
        use_geometric_pen = TRUE;
        delete_fill_map = TRUE;
        /* we need to create a bitmap handle from the image data. */
        if ( attr->l_ddep == 8 )
            {
            fill_map = D04GetDIBFromImageData(attr->l_data, attr->l_dwid,
                                            attr->l_dhyt, attr->l_cmap);
            }
        else
            {
            fill_map = D04GetMonoBitmapFromImageData(attr->l_data,attr->l_dwid,
                                                         attr->l_dhyt, fg, bg);
            }
        }
    else if ( attr->l_type == GRAY_FILL_TYPE )
        {
        if ( attr->l_pat != 1000 && attr->l_pat != 0 && attr->l_pat != -1 )
            {
            int tmp_ix;
            /* get percentage bitmap and create brush from that. */
            tmp_ix = getIXfromPercent(attr->l_pat);
            fill_map = D04GetStippleMap(tmp_ix, 0);
            use_geometric_pen = TRUE;
            }
        }

    /* if "l_pat" is zero, it means we need to use 0% percent foreground,
     * and therefore 100% background  -donato
     */
    if ( attr->l_pat == 0 )
        color = bg;
    else
        color = fg;

    if ( attr->l_style == TM_SOLID )
        {
        /* SOLID, non-dashed line. */
        if ( fill_map )
            {
            /* solid line with a stipple pattern. */
            /* we will be using a geometric pen with a pattern */
            lbrush.lbStyle = BS_PATTERN;
            lbrush.lbColor = 0;    /* this field ignored when BS_PATTERN */
            lbrush.lbHatch = (LONG)fill_map;

            pen_style = PS_GEOMETRIC;
            pen_width = width;
            hpen = (HPEN)MS32_ExtCreatePen((DWORD)pen_style, (DWORD)pen_width,
                                        &lbrush, (DWORD)0, (CONST DWORD*)NULL);
            }
        else
            {
            /* normal solid line. */
            pen_style = PS_SOLID;
            pen_width = width;
            if ( use_geometric_pen )
                {
                pen_style |= PS_GEOMETRIC | PS_ENDCAP_FLAT;
                lbrush.lbStyle = BS_SOLID;
                lbrush.lbColor = color;
                lbrush.lbHatch = 0;
                hpen = MS32_ExtCreatePen(pen_style, pen_width, &lbrush,
                                        (DWORD)0, (CONST DWORD*)NULL);
                }
            else
                {
                /* if width is greater than one pixel, use GEOMTERIC pen. */
                if ( pen_width > 10 )
                    {
                    pen_style |= PS_GEOMETRIC;
                    lbrush.lbStyle = BS_SOLID;
                    lbrush.lbColor = color;
                    lbrush.lbHatch = 0;
                    hpen = MS32_ExtCreatePen(pen_style, pen_width, &lbrush,
                                            (DWORD)0, (CONST DWORD*)NULL);
                    }
                else
                    hpen = MS_CreatePen(pen_style, 0, color);
                }
            }
        }
    else
        {
        /* DASHED line. */
        pen_width = width;
        /* default pen style for all non solid lines. */
        pen_style = PS_USERSTYLE | PS_GEOMETRIC | PS_ENDCAP_FLAT | PS_JOIN_BEVEL;
        if ( fill_map )
            {
            /* special line with a stipple pattern. */
            /* we will be using a geometric pen with a pattern */
            lbrush.lbStyle = BS_PATTERN;
            lbrush.lbColor = 0;    /* this field ignored when BS_PATTERN */
            lbrush.lbHatch = (LONG)fill_map;
            }
        else
            {
            /* non-stippled special pen. */
            lbrush.lbStyle = BS_SOLID;
            lbrush.lbColor = color;
            lbrush.lbHatch = 0;
            }

        dash_list_size = dash_length[attr->l_style-1];
        if ( dash_list_size > 0 )
            {
            /* create special 'dashed' pen */
            dashes = (DWORD*)malloc(dash_list_size*sizeof(DWORD));
            for ( ix=0; ix < dash_list_size; ix++ )
                dashes[ix] = dash_list[attr->l_style-1][ix];
            hpen = AxWinCreateDashedPen(dashes, dash_list_size, pen_style,
                                                    pen_width, &lbrush);
            free(dashes);
            }
        else
            {
            /* does this ever happen? */
            D04DebugMessage("dashed pen requested without a dash list!");
            }                       
        }

    if ( delete_fill_map )
        MS_DeleteObject(fill_map);
    return hpen;
}



static void
D04InitPatternPixmap(int pat_ix)
{
    HDC hdc, pix_dc, mem_hdc;
    HBITMAP pix, old_pix, old_hbm;
    int wid, hyt, ix, j;
    int new_wid, new_hyt;
    int retval;
    int bytesize = 0, bytewidth;
    bitmap_struct Bitmap;
    extern char *printer_fills[];
    BOOL stretch_bitmap = FALSE;

    if ( pat_ix >= MAX_PATS )
        {
        char buf[200];
        sprintf(buf,
            NXLT("invalid number passed to D04InitPatternPixmap=%d"),pat_ix);
        taskAbort(buf);
        }

    if ( pattern[pat_ix].pixmap != (Pixmap)NULL )
        {
        /* already created */
        return;
        } 


    hdc = PF.PfPrinterHDC;

    memset((char *)&Bitmap, '\0', sizeof(bitmap_struct));
    /* get XImage. */
    retval = AxIMtoXImage(printer_fills[pat_ix],printer_fills[pat_ix],&Bitmap);
    /* we currently load every builtin pattern because we don't *
     * do a pre-scan like we do for PS and PCL.    -donato      */
    if ( retval == AOK )
        {
        wid = pattern[pat_ix].wid = Bitmap.image.width;
        hyt = pattern[pat_ix].hyt = Bitmap.image.height;
        bytewidth = Bitmap.image.bytes_per_line;
        bytesize = bytewidth * Bitmap.image.height;

        /* loop twice: once for normal image, second for inverted image */
        for ( ix=0; ix < 2; ix++ )
            {
            pix_dc = MS_CreateCompatibleDC(hdc);
            /* create monochrome bitmap (to use as mask). */
            if ( PrinterResolution == 300 )
                pix = MS_CreateCompatibleBitmap(pix_dc, wid, hyt);
            else
                {
                stretch_bitmap = TRUE;
                new_wid = wid * (PrinterResolution/300);
                new_hyt = hyt * (PrinterResolution/300);
                pix = MS_CreateCompatibleBitmap(pix_dc, new_wid, new_hyt);
                }
            old_pix = MS_SelectObject(pix_dc, pix);

            if ( ix == 0 )
                pattern[pat_ix].pixmap = pix;
            else    /* inverted */
                pattern[pat_ix].imap = pix;

            mem_hdc = MS_CreateCompatibleDC(hdc);

            old_hbm = MS_SelectObject(mem_hdc, Bitmap.image.bitmap_handle);
            /* stretch the bitmap for printer resolution. */
            if ( stretch_bitmap )
                {
                MS_StretchBlt(pix_dc, 0, 0, new_wid, new_hyt,
                            mem_hdc, 0, 0, wid, hyt, SRCCOPY);
                }
            else
                MS_BitBlt(pix_dc, 0,0, wid, hyt, mem_hdc, 0,0, SRCCOPY);
            /* cleanup */
            MS_SelectObject(mem_hdc, old_hbm);
            MS_SelectObject(pix_dc, old_pix);
            MS_DeleteDC(pix_dc);
            MS_DeleteDC(mem_hdc);

            /* invert the bits in the image */
            if ( ix == 0 )
                {
                for ( j=0; j < bytesize; j++ )
                    Bitmap.image.data[j] = ~Bitmap.image.data[j];
                }
            }

        if ( Bitmap.image.data )
            TaskFree(Bitmap.pid, Bitmap.image.data);
        if ( Bitmap.mask.data )
            TaskFree(Bitmap.pid, Bitmap.mask.data);
        }
    else
        {
        /* Didn't create XImage. Set to null. */
        pattern[pat_ix].pixmap = pattern[pat_ix].imap = (Pixmap)NULL;
        pattern[pat_ix].wid = pattern[pat_ix].hyt = 0;
        }

    /* user defined patterns will be created as needed. */
}

static Pixmap
D04GetStippleMap(int pat_ix, int inv)
{
    if ( pat_ix >= MAX_PATS )
        {
        char buf[200];
        sprintf(buf,
            NXLT("invalid number passed to D04GetStippleMap=%d"), pat_ix);
        taskAbort(buf);
        }

    if ( pattern[pat_ix].pixmap == (Pixmap)NULL )
       {
       /* not yet created */
       D04InitPatternPixmap(pat_ix);
       }

    if ( inv )
        return pattern[pat_ix].imap;
    else
        return pattern[pat_ix].pixmap;
}  

static void
D04DeletePatterns(void)
{
    int pat;

    for ( pat=0; pat < MAX_PATS; pat++ )
        {
        if ( pattern[pat].pixmap )
            MS_DeleteObject(pattern[pat].pixmap);
        if ( pattern[pat].imap )
            MS_DeleteObject(pattern[pat].imap);
        }
}


static void
D04GetFillColors(attr_type *attr, COLORREF *fg, COLORREF *bg)
{
	unsigned long fore_pixel, back_pixel;

    /*** get the colors ***/
    if ( PF.PfColordev && PrinterSupportsPalette )
        {
        //old_palette = SelectAndRealizeAxWinPalette(hdc);
        D04GetForeAndBackPixels(attr, &fore_pixel, &back_pixel);
        *fg = PALETTEINDEX(fore_pixel);
        *bg = PALETTEINDEX(back_pixel);
        }
    else
        D04GetForeAndBackColors(attr, fg, bg);
}

/* this returns the color pixel indices necessary when we are dealing *
 * with a palette (i.e. when we have a color printer).                */
static void
D04GetForeAndBackPixels(attr_type *attr, unsigned long *fore_pixel,
                                        unsigned long *back_pixel)
{
    int c, m, y, k;

    if ( attr->use_ffg )
        *fore_pixel = AlisPixels[TMA_BLACK];
    else
        {
        c = attr->ffg.c;
        m = attr->ffg.m;
        y = attr->ffg.y;
        k = attr->ffg.k;
        AxGetRealPixel(2, c,m,y,k, 255, fore_pixel);
        }

    if ( attr->use_fbg )
        *back_pixel = AlisPixels[TMA_WHITE];
    else
        {
        c = attr->fbg.c;
        m = attr->fbg.m;
        y = attr->fbg.y;
        k = attr->fbg.k;
        AxGetRealPixel(2, c,m,y,k, 255, back_pixel);
        }
}

/************************************************************************
 * this returns actual RGB color reference values for use with 
 * non palette devices.
 ************************************************************************/
static void
D04GetForeAndBackColors(attr_type *attr, COLORREF *fore_color,
                                        COLORREF *back_color)
{
    int c, m, y, k, red, green, blue;

    if ( attr->use_ffg )
        *fore_color = RGB(0, 0, 0);    /* TMA_BLACK */
    else
        {
        c = attr->ffg.c;
        m = attr->ffg.m;
        y = attr->ffg.y;
        k = attr->ffg.k;
        D04GetRGBFromCMYK(c, m, y, k, &red, &green, &blue);
        *fore_color = RGB(red, green, blue);
        }

    if ( attr->use_fbg )
        *back_color = RGB(255, 255, 255);    /* TMA_WHITE */
    else
        {
        c = attr->fbg.c;
        m = attr->fbg.m;
        y = attr->fbg.y;
        k = attr->fbg.k;
        D04GetRGBFromCMYK(c, m, y, k, &red, &green, &blue);
        *back_color = RGB(red, green, blue);
        }
}

static void
D04GetLineColors(attr_type *attr, COLORREF *fore_color, COLORREF *back_color)
{
    int c, m, y, k, red, green, blue;

    if ( attr->use_lfg )
        *fore_color = RGB(0, 0, 0);    /* TMA_BLACK */
    else
        {
        c = attr->lfg.c;
        m = attr->lfg.m;
        y = attr->lfg.y;
        k = attr->lfg.k;
        D04GetRGBFromCMYK(c, m, y, k, &red, &green, &blue);
        *fore_color = RGB(red, green, blue);
        }

    if ( attr->use_lbg )
        *back_color = RGB(255, 255, 255);    /* TMA_WHITE */
    else
        {
        c = attr->lbg.c;
        m = attr->lbg.m;
        y = attr->lbg.y;
        k = attr->lbg.k;
        D04GetRGBFromCMYK(c, m, y, k, &red, &green, &blue);
        *back_color = RGB(red, green, blue);
        }
}




/******************************* IMAGE PRIMITIVE ************************/
int
D04PutImage(int xpos_mil, int ypos_mil, ge_image_type *imgP, int mode,
        poly_type *xform, cmap_type *cmap, attr_type *attr, TmRect32 *clip)
{
    int mil_wid, mil_hyt;
    int boxwid=0, boxht=0;              /* width & height of bounding box */
    int transx_mil, transy_mil;         /* x,y translation */
    double scalex, scaley;              /* x,y scale factors */
    double cosAng, sinAng;              /* sin, cos of rotation angle */
    double xShear;                      /* x shear */
    HPALETTE old_palette;


    D04DoClip(clip);
    /* for now, we always set the palette at the beginning of each *
     * "primitive".  Make sure it's un-selected at the end.        */
    if ( PF.PfColordev && PrinterSupportsPalette )
        old_palette = SelectAndRealizeAxWinPalette(PF.PfPrinterHDC);

    /* any units returned by extract_terms() should be in mils */
    scalex = 1.0;
    scaley = 1.0;
    extract_terms_for_device(imgP, xform, &transx_mil, &transy_mil,
                        &scalex, &scaley, &sinAng, &cosAng, &xShear,
                        &mil_wid, &mil_hyt, 1);

    if ( imgP->dep == 1 )
        {
        D04DrawMonoImage(mode, xpos_mil,ypos_mil, mil_wid, mil_hyt,
                                                imgP, attr, xform);
        }
    else
        {
        /* Windows will handle any color conversion necessary. */
        D04DrawColorImage(xpos_mil,ypos_mil,mil_wid,mil_hyt, imgP,cmap,xform);
        }

    if ( PF.PfColordev && PrinterSupportsPalette )
        UnSelectPalette(PF.PfPrinterHDC, old_palette);
    D04UnSetClip();
    return 0;
}
/************************************************************************/




static void
d04_convert_xform_to_pixel(poly_type *xform)
{
    int i;
    int res = 150;   /* pixel resolution */
    /* The above is a bit of a hack. We are doing images the "PCL way",
     * which means the highest res we can get an XImage of is 300 dpi.
     * This turns out to be pretty slow. At 75 dpi we get fast imaging,
     * but less than satisfactory image quality. So we use 150 dpi
     * as an acceptable compromise.             -donato
     */
    for ( i=0; i < xform->nPoints; i++ )
        {
        xform->points[i].x = (xform->points[i].x * res) / 1000;
        xform->points[i].y = (xform->points[i].y * res) / 1000;
        }
    for ( i=0; i < xform->nCtrls; i++ )
        {
        xform->controls[i].x = (xform->controls[i].x * res) / 1000;
        xform->controls[i].y = (xform->controls[i].y * res) / 1000;
        }
}

static void
D04DrawMonoImage(int mode, int xpos, int ypos, int mil_wid, int mil_hyt,
                ge_image_type *imgP, attr_type *attr, poly_type *xform)
{
    int old_mode;
    HDC hdc, src_dc, mask_dc;
    HBITMAP hbm, mask_hbm=NULL, old_hbm, old_mask_hbm;
    XImage *ximP;
    ge_image_transform_type tr; 
    int best_res = 300;
    BITMAPINFO *bitmap_info;
    int i, j, win_bytewidth, wid, hyt, prn_wid, prn_hyt;
    unsigned char *sourceP, *destP, *new_data;
    cmap_type cmap;    /* to hold the 2 colors to create a DIB with */
    COLORREF old_bk_color, old_text_color;
    DWORD rop_code;


    hdc = PF.PfPrinterHDC;
    /* convert dimensions to pixel values, for faster imaging */
    tr.x_scale = 1.0;
    tr.y_scale = 1.0;
    d04_convert_xform_to_pixel(xform);
    extract_factors_for_device(imgP, xform, &tr, GDI_CLASS_NUM);
    ximP = pf_tr_sys_mono_image(THIMpid, imgP, 0, &tr, &best_res);
    wid = ximP->width;
    hyt = ximP->height;

    /* see if we need a transparency mask */
    if ( mode == TMS_O_D )
        mask_hbm = MS_CreateBitmap(wid, hyt, 1, 1, ximP->data);

    /* create a 2-color DIB */
    cmap.n_colors = 2;
    cmap.color = (color_type *)malloc(cmap.n_colors * sizeof(color_type));
    /* foreground color in slot 1 */
    cmap.color[1].c = attr->ffg.c;
    cmap.color[1].m = attr->ffg.m;
    cmap.color[1].y = attr->ffg.y;
    cmap.color[1].k = attr->ffg.k;
    /* background color in slot 0 */
    cmap.color[0].c = attr->fbg.c;
    cmap.color[0].m = attr->fbg.m;
    cmap.color[0].y = attr->fbg.y;
    cmap.color[0].k = attr->fbg.k;
    bitmap_info = AxGetDIBInfo(wid, hyt, &cmap, AX_DIB_NORMAL);

    /* pad the data for the DIB call. */
    win_bytewidth = ((ximP->bytes_per_line + 3) / 4) * 4;
    sourceP = ximP->data;
    new_data = (unsigned char*)malloc(win_bytewidth * hyt);
    for ( i = 0; i < hyt; i++ )
        {
        destP = new_data + ((hyt - i - 1) * win_bytewidth);
        memset(destP, 0, win_bytewidth);
        for ( j = 0; j < ximP->bytes_per_line; j++ )
            *destP++ = *sourceP++;
        }

    /* convert mil values to printer resolution values. */
    xpos = ((xpos * XPixelsPerInch) / 1000) - Page_X_Offset;
    ypos = ((ypos * YPixelsPerInch) / 1000) - Page_Y_Offset;
    prn_wid = (mil_wid * XPixelsPerInch) / 1000;
    prn_hyt = (mil_hyt * YPixelsPerInch) / 1000;

    old_mode = SetStretchBltMode(hdc, COLORONCOLOR);

    if ( mask_hbm )
        {
        /* create the source bitmap */
        hbm = MS_CreateDIBitmap(hdc, &bitmap_info->bmiHeader, CBM_INIT,
                            new_data, bitmap_info, DIB_RGB_COLORS);
        src_dc = MS_CreateCompatibleDC(hdc);
        old_hbm = MS_SelectObject(src_dc, hbm);

        /* setup the mask dc */
        mask_dc = MS_CreateCompatibleDC(hdc);
        old_mask_hbm = MS_SelectObject(mask_dc, mask_hbm);
#if 1
		rop_code = 0x00E20746;
		MS_StretchBlt(hdc, xpos, ypos, prn_wid, prn_hyt, 
						mask_dc, 0, 0, wid, hyt, rop_code);
#else
        /* "True Mask Method" of doing transparency. */
        MS_StretchBlt(hdc, xpos, ypos, prn_wid, prn_hyt,
                        src_dc, 0, 0, wid, hyt, SRCINVERT);
        /* set up the destination correctly */
        old_bk_color = MS_SetBkColor(hdc, RGB_WHITE);
        old_text_color = MS_SetTextColor(hdc, RGB_BLACK);
        rop_code = 0x220326;   /* not source and destination */
        MS_StretchBlt(hdc, xpos, ypos, prn_wid, prn_hyt,
                        mask_dc, 0, 0, wid, hyt, rop_code);
        /* restore destination */
        MS_StretchBlt(hdc, xpos, ypos, prn_wid, prn_hyt,
                        src_dc, 0, 0, wid, hyt, SRCINVERT);
#endif
        /* local cleanup */
        MS_SelectObject(mask_dc, old_mask_hbm);
        MS_DeleteObject(mask_hbm);
        MS_DeleteDC(mask_dc);
        MS_SelectObject(src_dc, old_hbm);
        MS_DeleteObject(hbm);
        MS_DeleteDC(src_dc);
        MS_SetBkColor(hdc, old_bk_color);
        MS_SetTextColor(hdc, old_text_color);
        }
    else
        {
        int scan_lines;
        /* StretchDIBits draws the image data at the proper dpi and *
         * and with proper color conversions.                       */
        scan_lines = MS_StretchDIBits(hdc, xpos, ypos, prn_wid, prn_hyt,
                                        0, 0, wid, hyt, new_data, bitmap_info,
                                        DIB_RGB_COLORS, SRCCOPY);
        if ( scan_lines < hyt )
            D04DebugMessage("D04DrawMonoImage: error drawing image");
        }

    /* cleanup */
    pfFreeXImage(ximP);
    free(bitmap_info);
    free(cmap.color);
    free(new_data);
    SetStretchBltMode(hdc, old_mode);
}


static void
D04DrawColorImage(int xpos, int ypos, int mil_wid, int mil_hyt,
                ge_image_type *imgP, cmap_type *cmap, poly_type *xform)
{
    int old_mode, scan_lines;
    HDC hdc;
    int i, j, win_bytewidth, wid, hyt, prn_wid, prn_hyt;
    BITMAPINFO *bitmap_info, *mask_bitmap_info;
    unsigned char *new_data, *destP, *sourceP;
    XImage *ximP;
    ge_image_transform_type tr; 
    int best_res = 300;
	COLORREF old_bk_color, old_text_color;

    hdc = PF.PfPrinterHDC;
    /* convert dimensions to pixel values, for faster imaging */
    tr.x_scale = 1.0;
    tr.y_scale = 1.0;
    d04_convert_xform_to_pixel(xform);
    extract_factors_for_device(imgP, xform, &tr, GDI_CLASS_NUM);

    /* get the color XImage */
    ximP = pf_tr_sys_color_image(THIMpid, imgP, cmap, &tr, &best_res); 
    wid = ximP->width;
    hyt = ximP->height;

    /* see if we need a transparency mask */
    if ( D04IsTransparentImage(imgP, cmap) )
		mask_bitmap_info = AxGetDIBInfo(wid, hyt, cmap, AX_DIB_MASK);
	else
		mask_bitmap_info = NULL;

    /* pad and flip the data for the DIB call. */
    win_bytewidth = ((ximP->bytes_per_line + 3) / 4) * 4;
    sourceP = ximP->data;
    new_data = (unsigned char*)malloc(win_bytewidth * hyt);
    for ( i = 0; i < hyt; i++ )
        {
        destP = new_data + ((hyt - i - 1) * win_bytewidth);
        memset(destP, 0, win_bytewidth);
        for ( j = 0; j < ximP->bytes_per_line; j++ )
            *destP++ = *sourceP++;
        }

    xpos = ((xpos * XPixelsPerInch) / 1000) - Page_X_Offset;
    ypos = ((ypos * YPixelsPerInch) / 1000) - Page_Y_Offset;
    prn_wid = (mil_wid * XPixelsPerInch) / 1000;
    prn_hyt = (mil_hyt * YPixelsPerInch) / 1000;

    old_mode = SetStretchBltMode(hdc, HALFTONE);//COLORONCOLOR);

    if ( mask_bitmap_info )
        {
	    old_bk_color = MS_SetBkColor(hdc, RGB_BLACK);
    	old_text_color = MS_SetTextColor(hdc, RGB_WHITE);
        scan_lines = MS_StretchDIBits(hdc, xpos, ypos,
                                        prn_wid, prn_hyt, 0, 0,
                                        wid, hyt, new_data, mask_bitmap_info,
                                        DIB_RGB_COLORS, SRCAND);
        if ( scan_lines < hyt )
            D04DebugMessage("D04DrawColorImage: error drawing image");

    	bitmap_info = AxGetDIBInfo(wid, hyt, cmap, AX_DIB_TRANSPARENT);
        scan_lines = MS_StretchDIBits(hdc, xpos, ypos,
                                        prn_wid, prn_hyt, 0, 0,
                                        wid, hyt, new_data, bitmap_info,
                                        DIB_RGB_COLORS, SRCPAINT);
        if ( scan_lines < hyt )
            D04DebugMessage("D04DrawColorImage: error drawing image");

		free(mask_bitmap_info);
		MS_SetBkColor(hdc, old_bk_color);
		MS_SetTextColor(hdc, old_text_color);
        }
    else
        {
        /* StretchDIBits draws the image data at the proper dpi and *
         * and with proper color conversions.                       */
    	bitmap_info = AxGetDIBInfo(wid, hyt, cmap, AX_DIB_NORMAL);
        scan_lines = MS_StretchDIBits(hdc, xpos, ypos,
                                        prn_wid, prn_hyt, 0, 0,
                                        wid, hyt, new_data, bitmap_info,
                                        DIB_RGB_COLORS, SRCCOPY);
        if ( scan_lines < hyt )
            D04DebugMessage("D04DrawColorImage: error drawing image");
        }

    /* cleanup */
    pfFreeXImage(ximP);
    free(bitmap_info);
    free(new_data);
    SetStretchBltMode(hdc, old_mode);
}

#if 0
static HBITMAP
D04GetMaskBitmapFromColorXImage(XImage *ximP, cmap_type *cmap)
{
    HBITMAP hbm;
    int i, j, bytewidth, mask_bytewidth;
    int line_ix, mask_ix, fg_color, mask_size;
    unsigned char *new_data;


    bytewidth = ximP->bytes_per_line;
    mask_bytewidth = (bytewidth + 15) / 16 * 2;
    mask_size = mask_bytewidth * ximP->height;

    /* following taken from PXOpenDoc(), ge/add.c */
    new_data = (unsigned char*)malloc(mask_size);
    memset(new_data, 0, mask_size);
    for ( i = 0; i < ximP->height; i++ )
        {
        line_ix = i * bytewidth;
        mask_ix = i * mask_bytewidth;
        for ( j=0; j < bytewidth; j++ )
            {
            fg_color = ximP->data[line_ix + j];
            if ( !(cmap->color[fg_color].see_thru) )
                new_data[mask_ix+(j>>3)] |= Bitmask[j&7];
            }
        }  
    hbm = MS_CreateBitmap(ximP->width, ximP->height, 1, 1, new_data);
    free(new_data);
    return hbm;    /*** CALLER MUST DELETE THIS OBJECT! ***/
}
#endif



static unsigned long
D04GetPixelFromCMYK(int c, int m, int y, int k, int max)
{
    unsigned long pixel;

    /* this is only called if printer supports palette. */
    AxGetRealPixel(2, c, m, y, k, max, &pixel);
    return pixel;
}


static void
D04GetRGBFromCMYK(int c, int m, int y, int k, int *red, int *green, int *blue)
{
    *red   = 255 - MIN(255, c + k);
    *green = 255 - MIN(255, m + k);
    *blue  = 255 - MIN(255, y + k);
}



static void
D04GetColorsFromSubAttr(sub_attr_type *subattr, COLORREF *fore_color,
                                                COLORREF *back_color)
{
    int c, m, y, k, red, green, blue;

    c = subattr->fg.c;
    m = subattr->fg.m;
    y = subattr->fg.y;
    k = subattr->fg.k;
    D04GetRGBFromCMYK(c, m, y, k, &red, &green, &blue);
    *fore_color = RGB(red, green, blue);

    c = subattr->bg.c;
    m = subattr->bg.m;
    y = subattr->bg.y;
    k = subattr->bg.k;
    D04GetRGBFromCMYK(c, m, y, k, &red, &green, &blue);
    *back_color = RGB(red, green, blue);
}


/***    How the gradient fill is implemented:
    First, we convert the bounding box points from mils to 300 dpi (because
    we are using the PCL method, and because a higher resolution would use too
    much memory). The next two steps are identical to the Applix PCL driver:
    we create an XImage of intensity values and then a monochrome
    dithered XImage where 1 represents the foreground color and 0 the
    background color. We then create a 2-color DIB and blt it to the page.
***/
static void
D04DrawGradient(sub_attr_type *subattr, TmPoint32 *boundingBox)
{
    HDC hdc;
    int fg_pixel;
    int ix, wid, hyt, old_mode;
    XImage *xim, *xim_intens;
    TmPoint32 pcl_bbox[4];
    BITMAPINFO *bitmap_info;
    cmap_type cmap;    /* to hold the 2 colors to create a DIB with */
    int i, j, win_bytewidth;
    unsigned char *sourceP, *destP, *new_data;
    POINT pts[2];


    hdc = PF.PfPrinterHDC;
    wid = ABS(boundingBox[1].x - boundingBox[0].x);
    hyt = ABS(boundingBox[1].y - boundingBox[0].y);
    if ( (hyt == 0) || (wid == 0) )
        return;

    /* convert box values to 300 dpi */
    for ( ix=0; ix < 4; ix++ )
        {
        pcl_bbox[ix].x = (boundingBox[ix].x * 300) / 1000;
        pcl_bbox[ix].y = (boundingBox[ix].y * 300) / 1000;
        }

    /* first create an ximage of intensity values. */
    switch (subattr->type)
        {
        case LINEAR_FILL_TYPE:
            xim_intens = LinearGradientXImage(subattr, pcl_bbox);
            break;
        case BILINEAR_FILL_TYPE:
            xim_intens = BilinearGradientXImage(subattr, pcl_bbox);
            break;
        case RADIAL_FILL_TYPE:
            xim_intens = RadialGradientXImage(subattr, pcl_bbox);
            break;
        case RECT_FILL_TYPE:
            xim_intens = RectGradientXImage(subattr, pcl_bbox);
            break;
        default:
            xim_intens = NULL;
        }
    if ( !xim_intens )
        return;    /* uh oh */

    /* dither it */
    xim = FS_HalfTone(xim_intens);  /* always use diffusion dither. */
    pfFreeXImage(xim_intens);

    /* get colors */
    GetForegroundPixelFromSubattr(subattr, &fg_pixel);
    CreateCmapFor2ColorDIB(&cmap, subattr, fg_pixel);
    /* create a 2-color DIB */
    bitmap_info = AxGetDIBInfo(xim->width, xim->height, &cmap, AX_DIB_NORMAL);

    /* pad and flip the data for the DIB call. */
    win_bytewidth = ((xim->bytes_per_line + 3) / 4) * 4;
    bitmap_info->bmiHeader.biSizeImage = win_bytewidth * xim->height;
    sourceP = xim->data;
    new_data = (unsigned char*)malloc(win_bytewidth*xim->height);
    for ( i = 0; i < xim->height; i++ )
        {
        destP = new_data + ((xim->height - i - 1) * win_bytewidth);
        memset(destP, 0, win_bytewidth);
        for ( j = 0; j < xim->bytes_per_line; j++ )
            *destP++ = *sourceP++;
        }

    pts[0].x = boundingBox[0].x;
    pts[0].y = boundingBox[0].y;
    pts[1].x = wid;
    pts[1].y = hyt;
    D04ConvertMilsToPrinterRes(pts, 2, FALSE);
    pts[0].x -= Page_X_Offset;
    pts[0].y -= Page_Y_Offset;

    old_mode = SetStretchBltMode(hdc, COLORONCOLOR);
    MS_StretchDIBits(hdc, pts[0].x, pts[0].y, pts[1].x, pts[1].y,
                            0, 0, xim->width, xim->height, new_data,
                            bitmap_info, DIB_RGB_COLORS, SRCCOPY);
    SetStretchBltMode(hdc, old_mode);

    pfFreeXImage(xim);
    free(bitmap_info);
    free(cmap.color);
    free(new_data);
}

static void
D04DoGradientFill(attr_type *attr, poly_type *pol, int nPolys)
{
    HDC hdc;
    int save_dc;
    TmPoint32 boundingBox[4];
    sub_attr_type subattr;


    hdc = PF.PfPrinterHDC;
    /* we save the DC here (and restore it later), because we
     * will be setting a new clipping region that we only want to be
     * valid during this function.
     */
    save_dc = MS_SaveDC(hdc);

    D04SetClipFromPoly(hdc, pol, nPolys);

    D04GetBoundingBox(nPolys, attr, pol, boundingBox);

    memset(&subattr, 0, sizeof(subattr));
    subattr.type = attr->f_type;
    subattr.fg   = attr->ffg;
    subattr.bg   = attr->fbg;
    subattr.ang  = attr->f_ang;
    subattr.xoff = attr->f_xoff;
    subattr.yoff = attr->f_yoff;
    D04DrawGradient(&subattr, boundingBox);

    MS_RestoreDC(hdc, save_dc);
}

static void
D04DoGradientLine(attr_type *attr, poly_type *pol, int nPolys, COLORREF bg)
{
    HDC hdc;
    HPEN hpen, old_pen;
    int save_dc, stock_pen;
    TmPoint32 boundingBox[4];
    sub_attr_type subattr;


    hdc = PF.PfPrinterHDC;
    /* we save the DC here (and restore it later), because we
     * will be setting a new clipping region that we only want to be
     * valid during this function.
     */
    save_dc = MS_SaveDC(hdc);

    hpen = D04GetPenFromAttr(attr, &stock_pen);

    MS_SetPolyFillMode(hdc, WINDING);
    MS32_BeginPath(hdc);
    D04DrawPoly(hdc, pol, nPolys, FALSE, 0, 0);
    MS32_EndPath(hdc);
    old_pen = MS_SelectObject(hdc, hpen);
    /* "redefine" the path as if the current pen was used to paint it,
     * but don't actually render it (this is what WidenPath does).
     * and then set the path as the new clipping region.
     */
    MS32_WidenPath(hdc);
    if ( !MS32_SelectClipPath(hdc, RGN_AND) )
        D04PathAbort();
    MS_SelectObject(hdc, old_pen);

    D04GetBoundingBox(nPolys, attr, pol, boundingBox);

    memset(&subattr, 0, sizeof(subattr));
    subattr.type = attr->l_type;
    subattr.fg   = attr->lfg;
    subattr.bg   = attr->lbg;
    subattr.ang  = attr->l_ang;
    subattr.xoff = attr->l_xoff;
    subattr.yoff = attr->l_yoff;
    D04DrawGradient(&subattr, boundingBox);

    if ( !stock_pen )
        MS_DeleteObject(hpen);
    MS_RestoreDC(hdc, save_dc);
}

static void
D04DoGradientText(attr_type *attr, tx_char *txstr, int txlen, int tx, int ty,
													int unity, HFONT hfont)
{
#if 1
    /* Gradient text not implemented yet; just do a fill. */
    attr->f_pat  = 1000;
    attr->f_type = GRAY_FILL_TYPE;
    D04DrawFilledText(txstr, txlen, tx, ty, attr, unity, hfont, FALSE);

#else
    int save_dc;
    char *out_text;
    HDC hdc;
    BOOL retval;
    TmPoint32 txBox[4];    /* text bounding box */
    sub_attr_type subattr;

    hdc = PF.PfPrinterHDC;
    /* we explicitly save the DC here (and restore it later), because we
     * will be setting a new clipping region that we only want to be
     * valid during this function.
     */
    save_dc = MS_SaveDC(hdc);

    MS32_BeginPath(hdc);
    MS_SetTextAlign(hdc, TA_LEFT | TA_BASELINE);
    MS_SetBkMode(hdc, TRANSPARENT);
    out_text = D04GetOutputText(txstr, txlen);
    retval = MS_TextOut(hdc, tx, ty, out_text, txlen);
    MS32_EndPath(hdc);
    if ( !MS32_SelectClipPath(hdc, RGN_AND) )
        D04PathAbort();

    memset(&subattr, 0, sizeof(subattr));
    subattr.type = attr->f_type;
    subattr.fg   = attr->ffg;
    subattr.bg   = attr->fbg;
    subattr.ang  = attr->f_ang;
    subattr.xoff = attr->f_xoff;
    subattr.yoff = attr->f_yoff;
    D04GetTextBoundingBox(attr, tx, ty, out_text, txlen, txBox);
    D04DrawGradient(&subattr, txBox);

    free(out_text);
    MS_RestoreDC(hdc, save_dc);
#endif
}


static BOOL
IsPatternFill(attr_type *attr)
{
    int retval = FALSE;
    switch (attr->f_type)
        {
        case BUILTIN_FILL_TYPE:
        case USERDEF_FILL_TYPE:
            if ( attr->f_type == BUILTIN_FILL_TYPE && attr->f_pat == -1 )
                retval = FALSE;
            else
                retval = TRUE;
            break;

        case GRAY_FILL_TYPE:
            if ( attr->f_pat == 1000 || attr->f_pat == 0 )
                retval = FALSE;
            else
                retval = TRUE;
            break;
        }
    return retval;
}

/*****************************************************************************
    The following two functions are used on Windows 95 to manually tile
    bitmap images. We need to do this because Win95 does not support
    greater than 8x8 pattern bitmaps.       -donato
 ***/
static void
D04DrawPatternFill(HDC hdc, HBITMAP hbm, attr_type *attr, poly_type *pol,
							int nPolys, int memory_draw, DWORD rop_code)
{
    int i, j, x, y, tile_x, tile_y;
    int xpos, ypos, poly_wid, poly_hyt;
    HDC mem_hdc;
    HBITMAP old_hbm;
    BITMAP bm;
    TmPoint32 boundingBox[4];
    POINT pts[2];

    if ( !hbm )
        return; /* uh oh! */

	/* if we were passed a rop code, use it. otherwise do a straight copy. */
	if ( !rop_code )
		rop_code = SRCCOPY;

    /* Get bounding box of object in mils, then convert to printer res */ 
    D04GetBoundingBox(nPolys, attr, pol, boundingBox);
    pts[0].x = boundingBox[0].x;
    pts[0].y = boundingBox[0].y;
    pts[1].x = ABS(boundingBox[1].x - boundingBox[0].x);  /* poly width */
    pts[1].y = ABS(boundingBox[1].y - boundingBox[0].y);  /* poly height */
    D04ConvertMilsToPrinterRes(pts, 2, FALSE);
    if ( memory_draw )
        {
        /* we are drawing to a memory dc and need to zero the coordinates */
        xpos = ypos = 0;
        }
    else
        {
        xpos = pts[0].x - Page_X_Offset;
        ypos = pts[0].y - Page_Y_Offset;
        }
    poly_wid = pts[1].x;
    poly_hyt = pts[1].y;

    mem_hdc = MS_CreateCompatibleDC(hdc);
    old_hbm = MS_SelectObject(mem_hdc, hbm);

    MS_GetObject(hbm, sizeof(BITMAP), (LPSTR)&bm);
    if ( bm.bmBitsPixel == 1 )
        {
        /* we have a mono image, so set the fore and back colors. */ 
        COLORREF fg, bg;
        D04GetForeAndBackColors(attr, &fg, &bg);
        MS_SetBkColor(hdc, fg);
        MS_SetTextColor(hdc, bg);
        }

    x = xpos;
    y = ypos;
    tile_x = (poly_wid / bm.bmWidth) + 1;
    tile_y = (poly_hyt / bm.bmHeight) + 1;
    for ( i=0; i < tile_y; i++ )
        {
        for ( j=0; j < tile_x; j++ )
            {
            MS_BitBlt(hdc,x,y,bm.bmWidth,bm.bmHeight, mem_hdc,0,0, rop_code);
            x += bm.bmWidth;
            }
        y += bm.bmHeight;
        x = xpos;
        }

    MS_SelectObject(mem_hdc, old_hbm);
    MS_DeleteDC(mem_hdc);
}

static void
D04DrawTextPatternFill(HDC hdc, HBITMAP hbm, attr_type *attr,
                        int xpos, int ypos, int wid, int hyt)
{
    int i, j, x, y, tile_x, tile_y;
    HDC mem_hdc;
    HBITMAP old_hbm;
    BITMAP bm;

    if ( !hbm )
        return; /* uh oh! */

    mem_hdc = MS_CreateCompatibleDC(hdc);
    old_hbm = MS_SelectObject(mem_hdc, hbm);

    MS_GetObject(hbm, sizeof(BITMAP), (LPSTR)&bm);
    if ( bm.bmBitsPixel == 1 )
        {
        /* we have a mono image, so set the fore and back colors. */ 
        COLORREF fg, bg;
        D04GetForeAndBackColors(attr, &fg, &bg);
        MS_SetBkColor(hdc, fg);
        MS_SetTextColor(hdc, bg);
        }

    x = xpos;
    y = ypos;
    tile_x = (wid / bm.bmWidth) + 1;
    tile_y = (hyt / bm.bmHeight) + 1;
    for ( i=0; i < tile_y; i++ )
        {
        for ( j=0; j < tile_x; j++ )
            {
            MS_BitBlt(hdc,x,y,bm.bmWidth,bm.bmHeight, mem_hdc,0,0, SRCCOPY);
            x += bm.bmWidth;
            }
        y += bm.bmHeight;
        x = xpos;
        }

    MS_SelectObject(mem_hdc, old_hbm);
    MS_DeleteDC(mem_hdc);
}
/****************************************************************************/




static HPEN
D04GetPenFromAttr(attr_type *attr, int *stock_pen)
{
    HPEN hpen;
    COLORREF fg, bg;

    *stock_pen = FALSE;
    if ( attr->l_wid )
        {
        /* get the pen for the line. */
        if ( attr->l_type == BUILTIN_FILL_TYPE && attr->l_pat == -1 )
            {
            hpen = (HPEN)MS_GetStockObject(NULL_PEN);
            *stock_pen = TRUE;
            }
        else
            {
            D04GetLineColors(attr, &fg, &bg);
            hpen = D04GetDrawingPen(attr, fg, bg);
            if ( !hpen )
                {
                D04DebugMessage("D04Line: could not create pen!");
                hpen = (HPEN) MS_GetStockObject(BLACK_PEN);
                *stock_pen = TRUE;
                }
            }
        }
    else
        {
        hpen = (HPEN)MS_GetStockObject(NULL_PEN);
        *stock_pen = TRUE;
        }

    return hpen;
}

static HBRUSH
D04GetBrushFromAttr(attr_type *attr, int *stock_brush)
{
    HBRUSH hbrush;
    COLORREF fg, bg;
    unsigned long fore_pixel, back_pixel;

    *stock_brush = FALSE;
    if ( attr->fill_flag )
        {
        if ( PF.PfColordev && PrinterSupportsPalette )
            {
            D04GetForeAndBackPixels(attr, &fore_pixel, &back_pixel);
            fg = PALETTEINDEX(fore_pixel);
            bg = PALETTEINDEX(back_pixel);
            }
        else
            D04GetForeAndBackColors(attr, &fg, &bg);
        hbrush = D04GetPatternBrush(attr, fg,bg, stock_brush);
        }
    else
        {
        hbrush = (HBRUSH)MS_GetStockObject(NULL_BRUSH);
        *stock_brush = TRUE;
        }
    return hbrush;
}

static void
D04DoNormalLineAndFill(HDC hdc, attr_type *attr, poly_type *pol, int nPolys)
{
    /* render the path */
    if ( attr->fill_flag && attr->l_wid )
        {
        /* Stroke and Fill */
        /* we do separate strokes and fills so that open polys 
         * don't become closed.
         */
        D04DoNormalFill(hdc, attr, pol, nPolys);
        D04DoNormalLine(hdc, attr, pol, nPolys);
        }
    else if ( attr->fill_flag )
        {
        /* Fill */
        D04DoNormalFill(hdc, attr, pol, nPolys);
        }
    else
        /* Stroke */
        D04DoNormalLine(hdc, attr, pol, nPolys);
}

static void
D04DoNormalLine(HDC hdc, attr_type *attr, poly_type *pol, int nPolys)
{
    HPEN hpen, old_pen;
    int stock_pen;

    hpen = D04GetPenFromAttr(attr, &stock_pen);
    old_pen = MS_SelectObject(hdc, hpen);
    D04DrawPoly(hdc, pol, nPolys, FALSE, 0, 0);
    MS_SelectObject(hdc, old_pen);
    if ( !stock_pen )
        MS_DeleteObject(hpen);
}

static void
D04DoNormalFill(HDC hdc, attr_type *attr, poly_type *pol, int nPolys)
{
	D04DoPolyFill(attr, pol, nPolys, FALSE);
}

void
D04DoPolyFill(attr_type *attr, poly_type *pol, int nPolys, int transparent)
{
    if ( IsBezierCurve(pol, nPolys) )
        D04DrawBezierFill(attr, pol, nPolys, transparent);
    else if ( !transparent && D04IsSolidFill(attr) )
		/* 'Solid' means non-transparent, non-patterned. */
		D04DrawSolidFill(attr, pol, nPolys);
	else
        D04DrawPolyPatternFill(attr, pol, nPolys, transparent);
}

int
IsBezierCurve(poly_type *pol, int nPolys)
{
	int ix;
	for ( ix=0; ix < nPolys; ix++ )
		{
		if ( pol[ix].nCtrls > 0 )
			return TRUE;
		}
	return FALSE;
}

static int
D04IsSolidFill(attr_type *attr)
{
	int retval;

	retval = FALSE;
	if ( SOLID_FILL(attr) )
		retval = TRUE;
	else
		{
		switch (attr->f_pat)
			{
   			case 0:
   			case 12:
   			case 1:
   			case 10:
   			case 2:
   			case 21:
   			case 3:
   			case 13:
   			case 4:
				if ( attr->f_type == BUILTIN_FILL_TYPE )
					retval = TRUE;
			}
		}
	return retval;
}

static void
D04DoTransparentPoly(attr_type *attr, poly_type *pol, int nPolys)
{
	D04DoPolyFill(attr, pol, nPolys, TRUE);
}

static void
D04DrawSolidFill(attr_type *attr, poly_type *pol, int nPolys)
{
	int stock_brush;
	HDC hdc;
	HPEN old_pen;
	HBRUSH hbrush, old_brush;
	HPALETTE old_palette;
	COLORREF fg, bg;

	hdc = PF.PfPrinterHDC;
	D04GetFillColors(attr, &fg, &bg);
    if ( PF.PfColordev && PrinterSupportsPalette )
        old_palette = SelectAndRealizeAxWinPalette(hdc);
	/* since we are in 'DrawSolidFill' we know the following function
     * will return a solid brush that we need to delete, so we can ignore
	 * the 'stock_brush' value.
	 */
   	hbrush = D04GetPatternBrush(attr, fg, bg, &stock_brush);
	old_brush = MS_SelectObject(hdc, hbrush);
	old_pen = MS_SelectObject(hdc, MS_GetStockObject(NULL_PEN));
	D04DrawPoly(hdc, pol, nPolys, TRUE, 0, 0);	
	MS_SelectObject(hdc, old_brush);
	MS_SelectObject(hdc, old_pen);
	MS_DeleteObject(hbrush);
    if ( PF.PfColordev && PrinterSupportsPalette )
        UnSelectPalette(hdc, old_palette);
}

static void
D04DrawPolyPatternFill(attr_type *attr, poly_type *pol, int nPolys, 
													int transparent)
{
    int stock_brush;
    int xpos, ypos, poly_wid, poly_hyt;
    HDC hdc, mask_dc;
    HBITMAP mask_hbm, old_mask_hbm;
    TmPoint32 boundingBox[4];
    HBRUSH old_brush, hbrush, old_mask_brush;
    HPEN old_pen;
    POINT pts[2];
    DWORD rop_code;
    int min_xpos, min_ypos;
	COLORREF fg, bg, old_bk_color, old_text_color;
	HPALETTE old_palette;

    hdc = PF.PfPrinterHDC;
	stock_brush = FALSE;
	D04GetFillColors(attr, &fg, &bg);
    if ( PF.PfColordev && PrinterSupportsPalette )
        old_palette = SelectAndRealizeAxWinPalette(hdc);

    D04GetBoundingBox(nPolys, attr, pol, boundingBox);
    pts[0].x = boundingBox[0].x;
    pts[0].y = boundingBox[0].y;
    pts[1].x = ABS(boundingBox[1].x - boundingBox[0].x);  /* poly width */
    pts[1].y = ABS(boundingBox[1].y - boundingBox[0].y);  /* poly height */
    D04ConvertMilsToPrinterRes(pts, 2, FALSE);
    xpos = pts[0].x - Page_X_Offset;
    ypos = pts[0].y - Page_Y_Offset;
    poly_wid = pts[1].x;
    poly_hyt = pts[1].y;

    min_xpos = boundingBox[0].x;
    min_ypos = boundingBox[0].y;

    mask_dc  = MS_CreateCompatibleDC(hdc);
    mask_hbm = MS_CreateCompatibleBitmap(mask_dc, poly_wid, poly_hyt);
    old_mask_hbm = MS_SelectObject(mask_dc, mask_hbm);
    MS_PatBlt(mask_dc, 0, 0, poly_wid, poly_hyt, WHITENESS);
    old_mask_brush = MS_SelectObject(mask_dc, MS_GetStockObject(BLACK_BRUSH));
    old_pen = MS_SelectObject(mask_dc, MS_GetStockObject(NULL_PEN));
    D04DrawPoly(mask_dc, pol, nPolys, TRUE, min_xpos, min_ypos);

	if ( Running_Windows95 )
		{
		/* Windows 95 doesn't support greater than 8x8 bitmaps for brushes. */
		/* so we need to manually tile the bitmap pattern to a color 
		 * compatible dc.
		 */
		D04Win95FillObject(hdc, mask_dc, attr, pol, nPolys, xpos, ypos,
							poly_wid, poly_hyt, transparent);
		}
	else 
		{
   		old_bk_color = MS_SetBkColor(hdc, fg);
   		old_text_color = MS_SetTextColor(hdc, bg);
   		hbrush = D04GetPatternBrush(attr, fg, bg, &stock_brush);
    	old_brush = MS_SelectObject(hdc, hbrush);
		if ( transparent )
    		rop_code = 0x00A20E09;
		else
			rop_code = 0x00E20746;
		MS_BitBlt(hdc, xpos,ypos, poly_wid,poly_hyt, mask_dc, 0,0, rop_code);
		MS_SelectObject(hdc, old_brush);
		MS_SetBkColor(hdc, old_bk_color);
		MS_SetTextColor(hdc, old_text_color);
		}

    /* cleanup */
    MS_SelectObject(mask_dc, old_mask_brush);
    MS_SelectObject(mask_dc, old_pen);
    MS_SelectObject(mask_dc, old_mask_hbm);
    MS_DeleteObject(mask_hbm);
    MS_DeleteDC(mask_dc);
    if ( !Running_Windows95 && !stock_brush )
        MS_DeleteObject(hbrush);
    if ( PF.PfColordev && PrinterSupportsPalette )
        UnSelectPalette(hdc, old_palette);
}

static void
D04DrawBezierFill(attr_type *attr, poly_type *pol, int nPolys, int transparent)
{
    int stock_brush, save_dc;
    int xpos, ypos, poly_wid, poly_hyt;
    HDC hdc, src_dc;
    HBITMAP src_hbm, old_src_hbm;
    TmPoint32 boundingBox[4];
    HBRUSH old_brush, hbrush, solid_brush, old_solid_brush;
    COLORREF fg, bg, old_bk_color, old_text_color;
    POINT pts[2];
    DWORD rop_code;
	HPALETTE old_palette;

    hdc = PF.PfPrinterHDC;
	save_dc = MS_SaveDC(hdc);
	D04GetFillColors(attr, &fg, &bg);
    if ( PF.PfColordev && PrinterSupportsPalette )
        old_palette = SelectAndRealizeAxWinPalette(hdc);

    D04GetBoundingBox(nPolys, attr, pol, boundingBox);
    pts[0].x = boundingBox[0].x;
    pts[0].y = boundingBox[0].y;
    pts[1].x = ABS(boundingBox[1].x - boundingBox[0].x);  /* poly width */
    pts[1].y = ABS(boundingBox[1].y - boundingBox[0].y);  /* poly height */
    D04ConvertMilsToPrinterRes(pts, 2, FALSE);
    xpos = pts[0].x - Page_X_Offset;
    ypos = pts[0].y - Page_Y_Offset;
    poly_wid = pts[1].x;
    poly_hyt = pts[1].y;

    MS32_BeginPath(hdc);
    D04DrawPoly(hdc, pol, nPolys, FALSE, 0, 0);
    /* make sure shape is closed, otherwise we'll try to set an open
     * clipping region which causes BAD things to happen.
     * (on the DEC Alpha i was forced to reboot!)   -donato
     */
    MS32_CloseFigure(hdc);
    MS32_EndPath(hdc);
    if ( !MS32_SelectClipPath(hdc, RGN_AND) )
        D04PathAbort();
 
	if ( Running_Windows95 )
		{
		/* Windows 95 doesn't support greater than 8x8 bitmaps for brushes,
		 * so we need to manually tile the bitmap pattern. 
		 */
		HBITMAP fill_map = D04GetPatternBitmap(attr);
		if ( transparent )
			{
		   	rop_code = 0x00B8074A;
    		solid_brush = MS_CreateSolidBrush(fg);
    		old_solid_brush = MS_SelectObject(hdc, solid_brush);
			}
		else
			rop_code = SRCCOPY;

		D04DrawPatternFill(hdc, fill_map, attr, pol, nPolys, FALSE, rop_code);

		if ( attr->f_type == USERDEF_FILL_TYPE )
			MS_DeleteObject(fill_map);
		if ( transparent )
			{
    		MS_SelectObject(src_dc, old_solid_brush);
    		MS_DeleteObject(solid_brush);
			}
    	if ( PF.PfColordev && PrinterSupportsPalette )
    	    UnSelectPalette(hdc, old_palette);
		MS_RestoreDC(hdc, save_dc);
		return;
		}

    /* create source dc and fill with the foreground color. */
    src_dc  = MS_CreateCompatibleDC(hdc);
    src_hbm = MS_CreateCompatibleBitmap(hdc, poly_wid, poly_hyt);
    old_src_hbm = MS_SelectObject(src_dc, src_hbm);
    solid_brush = MS_CreateSolidBrush(fg);
    old_solid_brush = MS_SelectObject(src_dc, solid_brush);
    MS_PatBlt(src_dc, 0, 0, poly_wid, poly_hyt, PATCOPY);
    MS_SelectObject(src_dc, old_solid_brush);
    MS_DeleteObject(solid_brush);

	if ( transparent )
   		rop_code = 0x00AC0744;
	else
   		rop_code = 0x00B8074A;
   	hbrush = D04GetPatternBrush(attr, fg, bg, &stock_brush);
   	old_brush = MS_SelectObject(hdc, hbrush);
   	old_bk_color = MS_SetBkColor(hdc, fg);
   	old_text_color = MS_SetTextColor(hdc, bg);
	if ( !transparent && D04IsSolidFill(attr) )
		MS_PatBlt(hdc, xpos, ypos, poly_wid, poly_hyt, PATCOPY);
	else
   		MS_BitBlt(hdc, xpos,ypos, poly_wid,poly_hyt, src_dc,0,0, rop_code);
   	MS_SelectObject(hdc, old_brush);
   	MS_SetBkColor(hdc, old_bk_color);
	MS_SetTextColor(hdc, old_text_color);

    /* cleanup */
    MS_SelectObject(src_dc, old_src_hbm);
    MS_DeleteObject(src_hbm);
    MS_DeleteDC(src_dc);
    if ( !Running_Windows95 && !stock_brush )
        MS_DeleteObject(hbrush);
    if ( PF.PfColordev && PrinterSupportsPalette )
        UnSelectPalette(hdc, old_palette);
	MS_RestoreDC(hdc, save_dc);
}

static void
D04DoTransparentLine(attr_type *attr, poly_type *pol, int nPolys,
                                        COLORREF fg, COLORREF bg)
{
    int stock_brush, save_dc;
    int line_width;
    int xpos, ypos, poly_wid, poly_hyt;
    HDC hdc, src_dc, mask_dc;
    HBITMAP src_hbm, mask_hbm, old_src_hbm, old_mask_hbm, fill_map;
    TmPoint32 boundingBox[4];
    HBRUSH hbrush, old_brush;
    HPEN hpen, old_pen;
    COLORREF old_bk_color, old_text_color;
    POINT pts[2];


    hdc = PF.PfPrinterHDC;
    /* we explicitly save the DC here (and restore it later), because we
     * will be setting a new clipping region that we only want to be
     * valid during this function.
     */
    save_dc = MS_SaveDC(hdc);

    line_width = attr->l_wid;
    D04GetBoundingBox(nPolys, attr, pol, boundingBox);
    pts[0].x = boundingBox[0].x;
    pts[0].y = boundingBox[0].y;
    pts[1].x = ABS(boundingBox[1].x - boundingBox[0].x);  /* poly width */
    pts[1].y = ABS(boundingBox[1].y - boundingBox[0].y);  /* poly height */
    D04ConvertMilsToPrinterRes(pts, 2, FALSE);
    xpos = pts[0].x - Page_X_Offset;
    ypos = pts[0].y - Page_Y_Offset;
    poly_wid = pts[1].x;
    poly_hyt = pts[1].y;

    hpen = D04GetDrawingPen(attr, fg, bg);

    MS_SetPolyFillMode(hdc, WINDING);
    MS32_BeginPath(hdc);
    D04DrawPoly(hdc, pol, nPolys, FALSE, 0, 0);
    MS32_EndPath(hdc);
    old_pen = MS_SelectObject(hdc, hpen);
    /* "redefine" the path as if the current pen was used to paint it,
     * but don't actually render it (this is what WidenPath does).
     * and then set the path as the new clipping region.
     */
    MS32_WidenPath(hdc);
    if ( !MS32_SelectClipPath(hdc, RGN_AND) )
        D04PathAbort();

    /* draw fill to a color-compatible memory DC using the proper colors. */
    src_dc  = MS_CreateCompatibleDC(hdc);
    src_hbm = MS_CreateCompatibleBitmap(hdc, poly_wid, poly_hyt);
    old_src_hbm = MS_SelectObject(src_dc, src_hbm);
    MS_SetBkColor(src_dc, fg);
    MS_SetTextColor(src_dc, bg);
    if ( Running_WindowsNT )
        {
        hbrush = D04GetLinePatternBrush(attr, fg, bg, &stock_brush);
        old_brush = MS_SelectObject(src_dc, hbrush);
        MS_PatBlt(src_dc, 0, 0, poly_wid, poly_hyt, PATCOPY);
        MS_SelectObject(src_dc, old_brush);
        }
    else
        {
        /* need to do manual pattern tiling on Windows 95 */
        fill_map = D04GetPatternBitmap(attr);
        D04DrawPatternFill(src_dc, fill_map, attr, pol, nPolys, TRUE, 0);
        }

    /* draw fill to a mono memory DC, to be used as the mask. */
    fg = RGB_WHITE;
    bg = RGB_BLACK;
    mask_dc  = MS_CreateCompatibleDC(hdc);
    mask_hbm = MS_CreateCompatibleBitmap(mask_dc, poly_wid, poly_hyt);
    old_mask_hbm = MS_SelectObject(mask_dc, mask_hbm);
    MS_PatBlt(mask_dc, 0, 0, poly_wid, poly_hyt, BLACKNESS);
    MS_SetBkColor(mask_dc, fg);
    MS_SetTextColor(mask_dc, bg);
    if ( Running_WindowsNT )
        {
        old_brush = MS_SelectObject(mask_dc, hbrush);
        MS_PatBlt(mask_dc, 0, 0, poly_wid, poly_hyt, PATCOPY);
        MS_SelectObject(mask_dc, old_brush);
        }
    else
        {
        /* need to do manual pattern tiling on Windows 95 */
        D04DrawPatternFill(mask_dc, fill_map, attr, pol, nPolys, TRUE, 0);
        /* we only need to delete the bitmap if we created one as a user-defined
         * fill. Otherwise, the bitmap is part of our cache.
         */
        if ( attr->f_type == USERDEF_FILL_TYPE )
            MS_DeleteObject(fill_map);
        }

    /* "True Mask Method" of doing transparency. */
    MS_BitBlt(hdc, xpos, ypos, poly_wid, poly_hyt, src_dc, 0, 0, SRCINVERT);
    old_bk_color = MS_SetBkColor(hdc, RGB_WHITE);
    old_text_color = MS_SetTextColor(hdc, RGB_BLACK);
    MS_BitBlt(hdc, xpos, ypos, poly_wid, poly_hyt, mask_dc, 0, 0, 0x00220326);
    MS_BitBlt(hdc, xpos, ypos, poly_wid, poly_hyt, src_dc, 0, 0, SRCINVERT);
    MS_SetBkColor(hdc, old_bk_color);
    MS_SetTextColor(hdc, old_text_color);

    MS_SelectObject(src_dc, old_src_hbm);
    MS_SelectObject(mask_dc, old_mask_hbm);
    MS_DeleteObject(src_hbm);
    MS_DeleteObject(mask_hbm);
    MS_DeleteDC(src_dc);
    MS_DeleteDC(mask_dc);
    if ( Running_WindowsNT && !stock_brush )
        MS_DeleteObject(hbrush);
    MS_SelectObject(hdc, old_pen);
    MS_DeleteObject(hpen);
    MS_RestoreDC(hdc, save_dc);
}

static void
D04DoTransparentText(attr_type *attr, tx_char *txstr, int txlen, int tx,int ty,
														int unity, HFONT hfont)
{
#if 1
	D04DrawFilledText(txstr, txlen, tx, ty, attr, unity, hfont, TRUE);
#else
    int stock_brush;
    int xpos, ypos, tx_wid, tx_hyt, y_offset;
    char *out_text;
    HDC hdc, mask_dc;
    HBITMAP mask_hbm, old_mask_hbm;
    HBRUSH old_brush, hbrush;
	HFONT old_font;
    SIZE size;
    BOOL retval;
    COLORREF fg, bg, old_bk_color, old_text_color;
	DWORD rop_code;
	TEXTMETRIC tm;

    hdc = PF.PfPrinterHDC;

	/* get the ascent because we'll be drawing a memory dc to the
	 * printer dc rather than the actual text, which comes in with
	 * baseline coordinates.
	 */
	MS_GetTextMetrics(hdc, &tm);
	y_offset = tm.tmAscent;

    out_text = D04GetOutputText(txstr, txlen);
    MS_GetTextExtentPoint32(hdc, out_text, txlen, &size);
    tx_wid = size.cx;
    tx_hyt = size.cy;
    if ( unity || Running_Windows95 )
        {
        xpos = ((tx * XPixelsPerInch) / 1000) - Page_X_Offset;
        ypos = ((ty * YPixelsPerInch) / 1000) - Page_Y_Offset;
        }
    else
        {
        /* if we're here, a world transform is in effect, which has
         * already taken the offsets into account.
         */
        xpos = ((tx * XPixelsPerInch) / 1000);
        ypos = ((ty * YPixelsPerInch) / 1000);
        }

    mask_dc  = MS_CreateCompatibleDC(hdc);
    mask_hbm = MS_CreateCompatibleBitmap(mask_dc, tx_wid, tx_hyt);
    old_mask_hbm = MS_SelectObject(mask_dc, mask_hbm);
    MS_PatBlt(mask_dc, 0, 0, tx_wid, tx_hyt, WHITENESS);
    MS_SetBkMode(mask_dc, TRANSPARENT);
	MS_SetTextColor(mask_dc, RGB_BLACK);
	old_font = MS_SelectObject(mask_dc, hfont);
    retval = MS_TextOut(mask_dc, 0, 0, out_text, txlen);
    free(out_text);

    D04GetForeAndBackColors(attr, &fg, &bg);
    old_bk_color = MS_SetBkColor(hdc, fg);
    old_text_color = MS_SetTextColor(hdc, bg);
    hbrush = D04GetPatternBrush(attr, fg, bg, &stock_brush);
    old_brush = MS_SelectObject(hdc, hbrush);
	MS_SetBkColor(hdc, old_bk_color);
	MS_SetTextColor(hdc, old_text_color);

    rop_code = 0x00A20E09;
    MS_BitBlt(hdc, xpos,ypos-y_offset, tx_wid,tx_hyt, mask_dc, 0,0, rop_code);
    MS_SelectObject(hdc, old_brush);

    /* cleanup */
	MS_SelectObject(mask_dc, old_font);
    MS_SelectObject(mask_dc, old_mask_hbm);
    MS_DeleteObject(mask_hbm);
    MS_DeleteDC(mask_dc);
    if ( Running_WindowsNT && !stock_brush )
        MS_DeleteObject(hbrush);
#endif
}


static BOOL
D04IsTransparentImage(ge_image_type *imgP, cmap_type *cmap)
{
    int i, j, x;
    unsigned char *dataP;

    for ( i = 0, dataP = imgP->data; i < imgP->hyt;
                    i++, dataP += imgP->data_byte_wid )
        {
        for ( j = 0; j < imgP->wid; j++ )
            {
            x = dataP[j] ;
            if ( cmap->color[x].see_thru )
                return TRUE;
            }
        }
    return FALSE;
}

static void
D04Win95FillObject(HDC dest_dc, HDC mask_dc, attr_type *attr, poly_type *pol,
					int nPolys, int xpos, int ypos, int poly_wid, int poly_hyt, 
					int transparent) 
{
	HDC src_dc, new_mask_dc;
	HBITMAP fill_map, src_hbm, old_src_hbm, new_mask_hbm, old_mask_hbm;
	DWORD rop_code;

	if ( transparent )
		{
		new_mask_dc = MS_CreateCompatibleDC(dest_dc);
		new_mask_hbm = MS_CreateCompatibleBitmap(new_mask_dc,poly_wid,poly_hyt);
		old_mask_hbm = MS_SelectObject(new_mask_dc, new_mask_hbm);
    	MS_PatBlt(new_mask_dc, 0, 0, poly_wid, poly_hyt, WHITENESS);
		}

	fill_map = D04GetPatternBitmap(attr);
	src_dc = MS_CreateCompatibleDC(dest_dc);
	src_hbm = MS_CreateCompatibleBitmap(dest_dc, poly_wid, poly_hyt);
	old_src_hbm = MS_SelectObject(src_dc, src_hbm);
	D04DrawPatternFill(src_dc, fill_map, attr, pol, nPolys, TRUE, 0);

	if ( transparent )
		{
		MS_BitBlt(new_mask_dc, 0,0, poly_wid,poly_hyt, src_dc, 0,0, SRCINVERT);
		MS_BitBlt(new_mask_dc, 0,0, poly_wid,poly_hyt, mask_dc, 0,0, SRCAND);
		MS_BitBlt(new_mask_dc, 0,0, poly_wid,poly_hyt, src_dc, 0,0, SRCINVERT);
		}

	MS_BitBlt(dest_dc, xpos, ypos, poly_wid, poly_hyt, src_dc, 0,0, SRCINVERT);
	if ( transparent )
		MS_BitBlt(dest_dc, xpos,ypos,poly_wid,poly_hyt, new_mask_dc,0,0,SRCAND);
	else
		MS_BitBlt(dest_dc, xpos, ypos, poly_wid, poly_hyt, mask_dc,0,0,SRCAND);
	MS_BitBlt(dest_dc, xpos, ypos, poly_wid, poly_hyt, src_dc, 0,0, SRCINVERT);

	MS_SelectObject(src_dc, old_src_hbm);
	MS_DeleteObject(src_hbm);
	MS_DeleteDC(src_dc);
	if ( attr->f_type == USERDEF_FILL_TYPE )
		MS_DeleteObject(fill_map);

	if ( transparent )
		{
		MS_SelectObject(new_mask_dc, old_mask_hbm);
		MS_DeleteObject(new_mask_hbm);
		MS_DeleteDC(new_mask_dc);
		}
}




#if 0
static HBRUSH
D04CreateMonoBrushFromColor(char *data, cmap_type *cmap, int wid, int hyt)
{
    HBRUSH hbrush;
    HBITMAP hbm;
    int i, j, bytewidth, mask_bytewidth;
    int line_ix, mask_ix, fg_color, mask_size;
    unsigned char *new_data;


    //bytewidth = wid + wid%2;
    bytewidth = ((wid + 3) / 4) * 4;
    mask_bytewidth = (wid + 7)/8;
    mask_size = mask_bytewidth * hyt;

    /* following taken from PXOpenDoc(), ge/add.c */
    new_data = (unsigned char*)malloc(mask_size);
    memset(new_data, 0, mask_size);
    for ( i = 0; i < hyt; i++ )
        {
        line_ix = i * bytewidth;
        mask_ix = i * mask_bytewidth;
        for ( j=0; j < bytewidth; j++ )
            {
            fg_color = data[line_ix + j];
            if ( !(cmap->color[fg_color].see_thru) )
                new_data[mask_ix+(j>>3)] |= Bitmask[j&7];
            }
        }  
    hbm = MS_CreateBitmap(wid, hyt, 1, 1, new_data);
    hbrush = MS_CreatePatternBrush(hbm);
    free(new_data);
    MS_DeleteObject(hbm);
    return hbrush;
}
#endif


BOOL CALLBACK
PrinterAbortProc(HDC hdc, int printer_error)
{
    MSG msg;

    while ( !UserPrintAbort && MS_PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) )
        {
        if ( !MS_IsDialogMessage(Dlg_Hwnd, &msg) )
            {
            MS_TranslateMessage(&msg);
            MS_DispatchMessage(&msg);
            }
        }
    return !UserPrintAbort;
}

LRESULT CALLBACK
PrintDialogProc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
{
    LRESULT retval;
    static HBRUSH fill_brush = NULL;

    retval = FALSE;
    switch (message)
        {
        case WM_INITDIALOG:
            MS_EnableMenuItem(MS_GetSystemMenu(hwnd, FALSE),
                                        SC_CLOSE, MF_GRAYED);
            MS_SetWindowText(hwnd, XLT("Applixware Print Cancel"));
            MS_SetDlgItemText(hwnd, IDD_FILENAME, PF.PfDoc_name);
            retval = TRUE;
            break;

        case WM_PAINT:
            {
            HDC hdc;
            PAINTSTRUCT ps;

            MS_BeginPaint(hwnd, &ps);
            hdc = AxPackageHandle(ps.hdc, AX_DC_TYPE);
            AxDrawBevel((AXWidget)NULL, hdc, ps.rcPaint.left, ps.rcPaint.top,
                            ps.rcPaint.right - ps.rcPaint.left,  /* width */
                            ps.rcPaint.bottom - ps.rcPaint.top,  /* height */
                            2, BEVEL_OUT);
            REMOVE_DRAWABLE(hdc);
            MS_EndPaint(hwnd, &ps);
            retval = TRUE;
            break;
            }

        case WM_CTLCOLORDLG:
        case WM_CTLCOLORSTATIC:
            {
            POINT point;
            /* we use RGB colors here because we must use the Windows default*
             * palette anyway (the only colors the user can choose from).    */
            MS_SetBkColor((HDC)wparam, BackPixelRGB);
            MS_SetTextColor((HDC)wparam, ForeWorkPixelRGB);
            if ( !fill_brush )
                fill_brush = MS_CreateSolidBrush(BackPixelRGB);
            point.x = point.y = 0;
            MS_ClientToScreen(hwnd, &point);
            MS_LPtoDP((HDC)wparam, &point, 1);
            MS_SetBrushOrgEx((HDC)wparam, point.x, point.y, (LPPOINT)NULL);
            retval = (DWORD)fill_brush;
            break;
            }

        case WM_COMMAND:
            UserPrintAbort = TRUE;
            MS_EnableWindow(Parent_Hwnd, TRUE);
            MS_DestroyWindow(hwnd);
            Dlg_Hwnd = NULL;
            retval = TRUE;
            break;

        case WM_PAGECOUNT:    /* our own custom message */
            {
            /* wparam holds the page count. */
            char buf[5];
            sprintf(buf, "%d", (int)wparam);
            MS_SetDlgItemText(hwnd, IDD_PAGECOUNT, buf);
            retval = TRUE;
            break;
            }
        }
    return retval;
}


/**************************************************************************/
/***                    PRINT OPTION FUNCTIONS                          ***/
/**************************************************************************/
static int
close_enough(int a, int b)
{
    return a >= b-4 && a <= b+4;
}

/*-------------------------------------------------------------------------
    select_plex - Win GDI version
 -------------------------------------------------------------------------*/
static short
select_plex(int duplex_code)
{
    switch ( duplex_code )
        {
        case 0:
            return DMDUP_SIMPLEX;
        case 1:
            return DMDUP_VERTICAL;    /* "short-edge" */
        case 2:
            return DMDUP_HORIZONTAL;  /* "long-edge" */
        default:
            return DMDUP_SIMPLEX;
        }
}

/**************************************************************************/
/*** Functions to retrieve paper size information from elf array **********/
/**************************************************************************/
/*** retrieve GDI paper value, e.g. DMPAPER_LEGAL ***/
static short
select_paper(int in_w, int in_h)
{
    data_ptr pointP;
    int ix, size, wid, hyt;
    short paper;

    size = AxArraySize(PF.pi.size);
    for ( ix=0; ix < size; ix++ )
        {
        pointP = AxArrayFromArray(PF.pi.size, ix);
        wid = AxIntFromArray(pointP, 0);
        hyt = AxIntFromArray(pointP, 1);
        if ( close_enough(wid, in_w) && close_enough(hyt, in_h) )
            {
            paper = AxIntFromArray(pointP, 3);
            return paper;
            }
        }
    return DMPAPER_LETTER;    /* a default */
}
/*------------------------------------------------------------------------
    select_size - Win GDI version - retrieves paper dimensions
 -------------------------------------------------------------------------*/
static void
select_size(int in_w, int in_h, int *out_w, int *out_h)
{
    data_ptr pointP;
    int ix, size, wid, hyt;

    *out_w = 0;
    *out_h = 0;
    size = AxArraySize(PF.pi.size);
    for ( ix=0; ix < size; ix++ )
        {
        pointP = AxArrayFromArray(PF.pi.size, ix);
        wid = AxIntFromArray(pointP, 0);
        hyt = AxIntFromArray(pointP, 1);
        if ( close_enough(wid, in_w) && close_enough(hyt, in_h) )
            {
            *out_w = wid;
            *out_h = hyt;
            return;
            }
        }
}
/*** Retrieve "form name", e.g. Letter, Legal, A4, etc... ***/
static char *
select_form(int in_w, int in_h)
{
    data_ptr pointP;
    int ix, size, wid, hyt;
    char *form_str;


    size = AxArraySize(PF.pi.size);
    for ( ix=0; ix < size; ix++ )
        {
        pointP = AxArrayFromArray(PF.pi.size, ix);
        wid = AxIntFromArray(pointP, 0);
        hyt = AxIntFromArray(pointP, 1);
        if ( close_enough(wid, in_w) && close_enough(hyt, in_h) )
            {
            form_str = AxStrFromArray(pointP, 2);
            return form_str;
            }
        }
    return NULL;
}
/************************ END PRINT OPTION FUNCTIONS **********************/

