/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfarcell.c
   Applix ASC version: 6
   Date: Mon Nov 16 13:56:50 1998
*/
#include <math.h>
#include <aos.h>

typedef struct {
	short x, y;
} TmPoint;

typedef struct {
	short x, y, dx, dy;
} TmRect;

#include <pfarcell.p>

#define M	11
#define RADCVT	0.00174533

int b0[10] = {2000,3315,3979,4069,3703,3024,2188,1347,637,165};
int b1[10] = {32269,30804,28460,25378,21740,17755,13647,9634,5916,2664};

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static pushq(q, p)
short *q, *p;
{
	int i, x, y;

	x = *p++;
	y = *p;
	for (i = 0; i < 6; i++, q++)
		*q = *(q + 2);
	*q++ = x;
	*q = y;
}


/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
static int cdraw(q, stp, enp, p, nr)
TmPoint q[4];
int stp, enp;
TmPoint p[];
int *nr;
{
	int m, i;
	long n;

	for (i = stp; i < enp; i++) {
		m = M - 2 - i;
		n = (long) b1[i] * q[1].x - (long) b0[i] * q[0].x -
		    (long) b0[m] * q[3].x + (long) b1[m] * q[2].x;
		p[*nr].x = (n + 16384) >> 15;
		n = (long) b1[i] * q[1].y - (long) b0[i] * q[0].y -
		    (long) b0[m] * q[3].y + (long) b1[m] * q[2].y;
		p[*nr].y = (n + 16384) >> 15;
		(*nr)++;
	}
}


/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
int PfEllipse(rect, stang, arcang, p, nr)
TmRect *rect;
int stang;				/* start angle */
int arcang;				/* arc angle */
TmPoint p[];
int *nr;
{
	TmPoint q[4];
	int stp, arp, cx, cy, hx, hy, sx, sy, ex, ey;
	double rad;

	hx = rect->dx >> 1;
	hy = rect->dy >> 1;
	cx = rect->x + hx;
	cy = rect->y + hy;

/* fill up point queue */
	q[0].x = rect->x;
	q[0].y = cy;
	q[1].x = cx;
	q[1].y = rect->y;
	q[2].x = cx + hx;
	q[2].y = cy;
	q[3].x = cx;
	q[3].y = cy + hy;

	if (arcang < 0) {
		stang += arcang;
		arcang = -arcang;
	}

	if (arcang >= 3600) {
		sx = ex = q[1].x;
		sy = ey = q[1].y;
		stp = 0;
		arp = M << 2;		/* 4 times M */
		*nr = 0;
	}
	else {

/* compute endpoints with trig functions */
		rad = stang * RADCVT;
		sx = cx + (sin(rad) * hx);
		sy = cy + (-cos(rad) * hy);
		rad = (stang + arcang) * RADCVT;
		ex = cx + (sin(rad) * hx);
		ey = cy + (-cos(rad) * hy);

/* what quadrant to start int? */
		stp = (stang / 900) % 4;

		while (stp--)
			pushq((short *) q, (short *) q);
				/* rotate around until */

		stang %= 900;		/* where in this quadrant */
		while (stang < 0)	/* in case stang starts as neg. */
			stang += 900;

/* find start part in this quadrant */
		stp = ((unsigned int) stang * M) / 900;

/* find total parts to be drawn */
		arp = ((unsigned int) arcang * M) / 900;

		p[0].x = cx;
		p[0].y = cy;
		*nr = 1;
	}

	p[*nr].x = sx;
	p[*nr].y = sy;
	(*nr)++;

	while (arp > 0) {
		if (stp + arp > M - 1) {
			cdraw(q, stp, M - 1, p, nr);
			p[*nr].x = q[2].x;
			p[*nr].y = q[2].y;
			(*nr)++;
			pushq((short *) q, (short *) q);
		}
		else {
			cdraw(q, stp, stp + arp, p, nr);
		}
		arp += (stp - M);
		stp = 0;
	}

	p[*nr].x = ex;
	p[*nr].y = ey;
	(*nr)++;

	if (arcang < 3600) {
		p[*nr].x = cx;
		p[*nr].y = cy;
		(*nr)++;
		return (1);
	}
	return (0);
}
