/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfimage.c
   Applix ASC version: 59
   Date: Mon Nov 16 13:56:49 1998
*/
#include	<stdio.h>
#if AXMS
#include	<process.h>
#endif
#include	<pf.h>
#include	<math.h>
#include	<aos.h>
#include	<paratm.h>
#include	<axtext.h>
#include	<devsetup.h>
#include	<pfint.h>
#include	<pfpat.h>

#include	<pfimage.p>

/*-----------------------------------------------------------------------------
	define
-----------------------------------------------------------------------------*/

#define	max(x, y)	(((x) >= (y)) ? (x) : (y))
#define	min(x, y)	(((x) <= (y)) ? (x) : (y))


/*-----------------------------------------------------------------------------
	local data
-----------------------------------------------------------------------------*/
static unsigned char bitmask[] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};

/*-----------------------------------------------------------------------------
	pf_image
	
	Taken from TmPutImage function
-----------------------------------------------------------------------------*/
int pf_image()
{
	int image_mode, x_pos, y_pos, x_size, y_size, vert_scan;
	register int horiz;
	register unsigned char *d_ptr, *m_ptr, *t_ptr;
	int tmp_int = 0;
	attr_type attr;
	poly_type xform;
	TmPoint32 xform_points[5];
	cmap_type cmap;
	int cmap_ix;
	TmRect32 clip ;
	ge_image_type img;
	int see_thru ;
	unsigned char m, pix ;
	int i ;
			
	Debugs ("pf_image()\n");
	
	memset((char *)&img, 0, sizeof(ge_image_type));
	img.data = NULL;
	img.mask = NULL;
	img.eps_seg = NULL;
	
	xform.points = xform_points;
	xform.nPoints = 5;
	xform.nCtrls = 0;
	xform.controls = (TmPoint32 *)0;
	
	InWord ();		 /* read out pixel flag only to stay in sync */
	image_mode = InWord ();
	
	memset((char *)&attr, 0, sizeof(attr_type));
	/* color stuff mainly for mono-images */
		attr.ffg.c = InByte();
	attr.ffg.m = InByte();
	attr.ffg.y = InByte();
	attr.ffg.k = InByte();
	
	attr.fbg.c = InByte();
	attr.fbg.m = InByte();
	attr.fbg.y = InByte();
	attr.fbg.k = InByte();
	
	x_pos = InInt();
	y_pos = InInt();
	x_size = InInt();
	y_size = InInt();
	
	img.dep = InWord ();
	img.wid = x_size;
	img.hyt = y_size;
	
	xform.points[0].x = InLong ();
	xform.points[0].y = InLong ();
	xform.points[1].x = InLong ();
	xform.points[1].y = InLong ();
	xform.points[2].x = InLong ();
	xform.points[2].y = InLong ();
	xform.points[3].x = InLong ();
	xform.points[3].y = InLong ();
	xform.points[4].x = InLong ();
	xform.points[4].y = InLong ();
	
#if BYTE_WID_IN_SE
	img.data_byte_wid = InWord ();  /* obsolete... */
#else
	img.data_byte_wid = EVEN_BYTES_FROM_BITS(x_size * img.dep);
#endif
	
	Debug4f(NXLT("pix values: x_pos=%d y_pos=%d x_size=%d y_size=%d\n"),
			x_pos,y_pos,x_size,y_size);
	
	img.data = (unsigned char *)malloc(y_size * img.data_byte_wid);
	d_ptr = img.data;
	
	for (vert_scan=0; vert_scan < y_size; vert_scan++)
		for (horiz=0; horiz < img.data_byte_wid; horiz++)
			*d_ptr++ = (unsigned char)(InByte() & 0xFF);

	/* MASK_IN_STREAM */
	InAlign();
	img.mask_byte_wid = InWord();	/* BYTE wid; not bit! */
		if (img.mask_byte_wid != 0)
		{
			img.mask = (unsigned char *)malloc(y_size *
												img.mask_byte_wid);
			m_ptr = img.mask;
			
			for (vert_scan=0; vert_scan < y_size; vert_scan++)
				for (horiz=0; horiz < img.mask_byte_wid; horiz++)
					*m_ptr++ = (unsigned char)(InByte() & 0xFF);
		}
	
	cmap.n_colors = InLong();
	cmap.color =
		(color_type *)malloc(cmap.n_colors * sizeof(color_type));
	memset((char *)cmap.color, 0, (cmap.n_colors * sizeof(color_type)));
	
	/* read the color map and identify a transparent entry */
	see_thru = -1 ;
	for (cmap_ix=0; cmap_ix<cmap.n_colors; cmap_ix++)
	{
		tmp_int = InInt();
		if (tmp_int)
		{
			if (see_thru == -1)
				see_thru = cmap_ix ;
			cmap.color[cmap_ix].see_thru = 1 ;
		}
		else
			cmap.color[cmap_ix].see_thru = 0 ;

		cmap.color[cmap_ix].c = InByte();
		cmap.color[cmap_ix].m = InByte();
		cmap.color[cmap_ix].y = InByte();
		cmap.color[cmap_ix].k = InByte();
	}

	/* any masked out bits in color image are set to "see thru" color and vice versa */
	if (img.dep == 8 && img.mask != NULL && see_thru != -1)
	{
		for (vert_scan=0, d_ptr = img.data, m_ptr = img.mask;
				vert_scan < y_size;
				vert_scan++, m_ptr += img.mask_byte_wid)
		{
			t_ptr = m_ptr ;
			m = 0x80 ;

			for (horiz=0; horiz < img.data_byte_wid; horiz++, d_ptr++)
			{
				/* if no bit, pixel is see_thru */
				if ((*t_ptr & m) == 0)
					*d_ptr = (unsigned char)see_thru ;

				/* else, pixel is not see_thru - find another cmap entry, if necessary */
				else
				{
					pix = *d_ptr ;
					if (cmap.color[pix].see_thru)
					{
						for (i = 0 ; i < cmap.n_colors ; i++)
						{
							if (!cmap.color[i].see_thru &&
								cmap.color[pix].c == cmap.color[i].c &&
								cmap.color[pix].m == cmap.color[i].m &&
								cmap.color[pix].y == cmap.color[i].y &&
								cmap.color[pix].k == cmap.color[i].k)
							{
								*d_ptr = (unsigned char)i ;
								break ;
							}
						}
					}
				}
					
				m = m >> 1 ;
				if (m == 0)
				{
					m = 0x80 ;
					t_ptr++ ;
				}
			}
		}
	}

	/* clipping rectangle */
	InBlock((char *)&clip, sizeof(clip)) ;

	/* Now for the EPS data */
		img.eps_seg_size = InLong();
	if (img.eps_seg_size != 0)
	{
		img.eps_seg = (char *)malloc(img.eps_seg_size + 1);
		InCString(img.eps_seg, img.eps_seg_size);
	}
	
	PutImage (x_pos, y_pos, &img, image_mode, &xform, &cmap, &attr, &clip);
	
	if (cmap.color != NULL)
		free(cmap.color);
	if (img.data != NULL)
		free(img.data);
	if (img.mask != NULL)
		free(img.mask);
	if (img.eps_seg != NULL)
		free(img.eps_seg);
	
	return(0);
}



/*-----------------------------------------------------------------------------
	ColorToGray

	converts the inBuf from 8 bit color to 8 bit intensities
	(each in a 16 bit array entry).
-----------------------------------------------------------------------------*/
void ColorToGray(inBuf, inWid, inHeight, cmap, outBuf, Align)
char *inBuf;
unsigned int inWid;
unsigned int inHeight;
cmap_type *cmap;
short *outBuf;
int Align;  /* how the input and output arrays are aligned... */
{
	register int WidAligned;
	register int line;
	register int col;
	unsigned char tmp_char;
	register unsigned int cmap_ix;

	int c, m, y, k;
	short tmp;
	short cmapix_to_intens[256];	/* a shadow to the ge cmap */

	if (Align == 16)
		WidAligned = (inWid + 1) & ~1;
	else /* 8 */
		WidAligned = inWid;

	for (tmp=0; tmp<256; tmp++)
		cmapix_to_intens[tmp] = (short)-1;  /* something invalid */

	for (line=0; line < inHeight; line++)
	{
		for (col=0; col < WidAligned; col++)
		{
			tmp_char = inBuf[(line * WidAligned) + col];
			cmap_ix = (unsigned int)tmp_char;
			/* cache the brightness values */
			if (cmapix_to_intens[cmap_ix] == (short)-1)
			{
				/* following code based on ExtractRGBCmapEntry() */
				if (cmap->color[cmap_ix].see_thru)
				{
								/* Take BG color... */
					c = 0;
					m = 0;
					y = 0;
					k = 0;
					cmapix_to_intens[cmap_ix] = 255;
				}
				else
				{
					c = (int)cmap->color[cmap_ix].c;
					m = (int)cmap->color[cmap_ix].m;
					y = (int)cmap->color[cmap_ix].y;
					k = (int)cmap->color[cmap_ix].k;

					/* This gives the BLACKness (0..1) */
					cmapix_to_intens[cmap_ix] =
						(short)(255.0 - (255.0 * IntensFromCMYKColor(c, m, y, k, 255)));
				}
			}
			outBuf[(line * WidAligned) + col] = cmapix_to_intens[cmap_ix];
		}
	}
}

/*-----------------------------------------------------------------------------
	IntensFromCMYKColor

	returns the BLACKness 0.0 - 1.0 
	based on the c, m, y, k values, each 0..max
-----------------------------------------------------------------------------*/
double IntensFromCMYKColor(cyan, magenta, yellow, black, xmax)
int cyan, magenta, yellow, black, xmax;
{
#if 0
	int red, green, blue;
#endif
	register int tmp_c, tmp_m, tmp_y;
	double gray_f;

	if ( cyan || magenta || yellow || black)
	{
#if 1
		tmp_c = MIN(cyan+black, xmax);
		tmp_m = MIN(magenta+black, xmax);
		tmp_y = MIN(yellow+black, xmax);

		gray_f = (double)((30*tmp_c) + (59*tmp_m) + (11*tmp_y))/(xmax*100.0);
#else
		red =	0xFF - MIN(0xFF, c+k);  /* 255..0 */
		green = 0xFF - MIN(0xFF, m+k);
		blue =  0xFF - MIN(0xFF, y+k);
		gray_f = (double)((xmax*100.0) - ((red*30) + (green*59) + (blue*11))) / 25500.0;
#endif
		return(gray_f);
	}
	else
		return(0.0); /* white */
}


/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
#define CLR_XY_BIT(ptr,byte_wid,row,col)	\
	(ptr)[((byte_wid)*(row))+((col)/8)] &= ~(((unsigned char)0x80 >> ((col)%8) & 0xFF))

#if 1
#define SET_XY_BIT(ptr,byte_wid,row,col)	\
	(ptr)[((byte_wid)*(row))+((col)/8)] |= (((unsigned char)0x80 >> (col)%8) & 0xFF)

#else	/* only use this for debugging... OK? */

static void
SET_XY_BIT(ptr, byte_wid, row, col)
unsigned char *ptr;
int byte_wid;
int col;
int row;
{
	register unsigned char tmp;
	register unsigned char tmp2;
	register int ptr_index;

	ptr_index = (row * byte_wid) + (col/8);
	tmp = ptr[ ptr_index ]; /* access the byte which contain the bit */
	tmp2 = (unsigned char)((unsigned char)(0x80 >> (col %8)));
	ptr[ ptr_index ] = tmp | tmp2;
}
#endif

/*-----------------------------------------------------------------------------
	There are various options when given a color (or gray) image and a
	monochrome output device.  Most of these options have been explored
	here.  Points to note: (1) When do we do the color quantization,
	(2) when do we scale and (3) when do we "dither" (halftone or
	error-difuse).  What seems most successful at the mo' is to scale
	-> color convert -> error-diffusion dither
	(FloydSteinbergDither()).
	We should combine the scale and color conversion stages...
-----------------------------------------------------------------------------*/
#define FS_ERROR_DIFF 1
#define FOUR_X_FOUR_45 1

#if FOUR_X_FOUR
#define HT_CELL_HYT 4
#define HT_CELL_WID 4
static
unsigned char HT_Cell[HT_CELL_HYT][HT_CELL_WID] =
{
	{ 13, 11, 12, 15},
	{  4,  3,  2,  9},
	{  5,  0,  1, 10},
	{  8,  6,  7, 14}
};
#endif



/*-----------------------------------------------------------------------------
	pf_tr_sys_color_image
-----------------------------------------------------------------------------*/
XImage *pf_tr_sys_color_image(pid, imP, cmap, tr, best_res)
	int pid;
	ge_image_type *imP;
	cmap_type *cmap;
	ge_image_transform_type *tr;
	int *best_res;
{
	register int i, j, x, y, ix, line_ix;
	int wid, hyt;
	register double x_xScale, x_Trans;
	register double y_xScale, y_Trans;
	double y_yScale, x_yScale;
	register double i_y_yScale, i_x_yScale;
	XImage *xP;


	/* get the width and height of the new, unmasked raster */
	wid = tr->wid;
	hyt = tr->hyt;
	
	/* create the header */
	xP = (XImage *)TaskAlloc(pid, sizeof(XImage));
	xP->format = ZPixmap;
	xP->bits_per_pixel = 8;
	xP->width  = wid;
	xP->height = hyt;
	xP->depth  = 8;
	xP->bytes_per_line = wid + wid%2;
	xP->xoffset = 0;
	xP->byte_order = MSBFirst;
	xP->bitmap_bit_order = MSBFirst;
	xP->bitmap_unit = 16;
	xP->bitmap_pad = 16;

	/* don't do anything if there is nothing to do */
	if ( xP->width == 0 || xP->height == 0 )
		xP->data = (char *)0;

	else /* transform data into a "deep" sys image */
	{
		/* allocate a block of memory */
		xP->data = (char *)TaskCalloc(pid, 1, xP->height * xP->bytes_per_line);

		/* set up the transfrom factors */
		x_xScale = (double)tr->x_x;
		x_yScale = (double)tr->x_y;
		y_xScale = (double)tr->y_x;
		y_yScale = (double)tr->y_y;
		x_Trans  = (double)tr->x_t;
		y_Trans  = (double)tr->y_t;

		/* scan thru the dest space, getting the pel from the source image */
		for ( i=0; i<xP->height; i++ )
		{
			line_ix = i*xP->bytes_per_line;
			i_x_yScale = i * x_yScale;
			i_y_yScale = i * y_yScale;
			for (j=0; j<xP->width; j++ )
			{
				/* get the index for the pixel in the source image */
				x = j*x_xScale + i_x_yScale + x_Trans;
				y = i_y_yScale - j*y_xScale + y_Trans;

				/* if this value is valid update the pixel */
				if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt )
				{
					ix = y*imP->data_byte_wid + x;
					xP->data[line_ix+j] = imP->data[ix];
				}
				else
					xP->data[line_ix+j] = 0;
			}
		}
	}

	return ( xP );
}


/*-----------------------------------------------------------------------------
	pf_tr_sys_mono_image

	based on tr_sys_mono_image()
-----------------------------------------------------------------------------*/
XImage *pf_tr_sys_mono_image(pid, imP, cmap, tr, best_res)
	int pid;
	ge_image_type *imP;
	cmap_type *cmap;
	ge_image_transform_type *tr;
	int *best_res;
{
	register int i, j, x, y, ix, line_ix;
	int wid, hyt;
	double x_xScale, x_yScale, x_Trans;
	double y_xScale, y_yScale, y_Trans;
	XImage *xP;



	/* get the width and height of the new, unmasked raster */
	wid = tr->wid;
	hyt = tr->hyt;

	/* create the header */
	xP = (XImage *)TaskAlloc(pid, sizeof(XImage));
	xP->format = XYBitmap;
	xP->bits_per_pixel = 1;
	xP->width  = wid;
	xP->height = hyt;
	xP->depth  = 1;
	xP->bytes_per_line = (xP->width+7)/8 + ((xP->width+7)/8)%2;
	xP->xoffset = 0;
	xP->byte_order = MSBFirst;
	xP->bitmap_bit_order = MSBFirst;
	xP->bitmap_unit = 16;
	xP->bitmap_pad = 16;

	/* don't do anything if there is nothing to do */
	if ( xP->width == 0 || xP->height == 0 )
		xP->data = (char *)0;

	else /* transform data into a "deep" sys image */
	{
		/* allocate a block of memory */
		xP->data = (char *)TaskCalloc(pid, 1, xP->height * xP->bytes_per_line);

		/* set up the transfrom factors */
		x_xScale = (double)tr->x_x;
		x_yScale = (double)tr->x_y;
		y_xScale = (double)tr->y_x;
		y_yScale = (double)tr->y_y;
		x_Trans  = (double)tr->x_t;
		y_Trans  = (double)tr->y_t;

		/* scan thru the dest space, getting the pel from the source image */
		for ( i=0; i<xP->height; i++ )
		{
			line_ix = i*xP->bytes_per_line;
			for (j=0; j<xP->width; j++ )
			{
				/* get the index for the pixel in the source image */
				x = j*x_xScale + i*x_yScale + x_Trans;
				y = i*y_yScale - j*y_xScale + y_Trans;

				/* if this value is valid update the pixel */
				if ( x >= 0 && x < imP->wid && y >= 0 && y < imP->hyt )
				{
					ix = y*imP->data_byte_wid + (x>>3);
					if ( imP->data[ix] & bitmask[x&7] )
						xP->data[line_ix+(j>>3)] |= bitmask[j&7];
					else
						xP->data[line_ix+(j>>3)] &= ~bitmask[j&7];
				}
				else
					xP->data[line_ix+(j>>3)] &= ~bitmask[j&7];
			}
		}
	}

	return ( xP );
}



#define HT_CELL_WID		4
#define HT_CELL_HYT		4
static int HT_Cell_even[HT_CELL_HYT][HT_CELL_WID] = 
{
	{ 14, 12, 13, 16},
	{  5,  4,  3, 10},
	{  6,  1,  2, 11},
	{  9,  7,  8, 15}
};


#define INIT_ERR_ROW	1
#define ERROR_PROP		1
#define SERPENTINE		1

#define FS_SCALE		1  /* 1024 */
#define HALF_FS_SCALE	0 /* 512 */


/*-----------------------------------------------------------------------------
	FSdither
-----------------------------------------------------------------------------*/
static void FSdither(inBuffer, inWidth, inHeight, alignSize, outBuffer)
unsigned char *inBuffer;
register int inWidth, inHeight;
int alignSize;
unsigned char *outBuffer;
{
	int outByteWidth, inByteWidth;
	register int row;
	register int col;
	int endstop;
	int in_index;
	int fs_direction = 1;	/* 1=> left to right, 0=> right to left */
	register int sum;
	int sum_forward, sum_d_forward, sum_d, sum_d_backward;
	short *inbuf;
	short *this_err;
	short *next_err;
	short *tmp_err; /* used when switching direction */
	register int threshold = 127 * FS_SCALE;


	/* How is the output buffer aligned? (Add other values as necessary.) */
	if (alignSize == 16)
		outByteWidth = (((inWidth + 15) / 8) & ~1);
	else /* 8 */
		outByteWidth = (inWidth + 7) / 8;

	inByteWidth = inWidth;

	/* time/space tradeoff; this way we dont have to check at the end
		of each col */
	this_err = (short *)malloc((inByteWidth +2) * sizeof(short));
	next_err = (short *)malloc((inByteWidth +2) * sizeof(short));
	inbuf = (short *)malloc(inByteWidth * sizeof(short));

	memset(next_err, '\0', (inByteWidth +2) * sizeof(short));
#if INIT_ERR_ROW
	srand((unsigned int)(time(0) ^ getpid()));  /* seed */

	for (col=0; col<inByteWidth+2; col++)
		this_err[col] = 63 - (rand() & 127);	/* fill array with
												+/-0..FS_SCALE/8 */
#else

	memset(this_err, '\0', (inByteWidth +2) * sizeof(short));
#endif
	memset(outBuffer, '\0', inHeight * outByteWidth);

	for (row=0; row<inHeight; row++)
	{
		for (col=0; col<inByteWidth; col++)
		{
			/* refill the inbuf[] */
			inbuf[col] =
				(unsigned short)(0xFF - (unsigned char)inBuffer[(row * inWidth) + col]);
		}
		memset(next_err, 0, sizeof(short) * (inWidth +2));

		if ( fs_direction )
		{
			col = 0;		/* start at left */
			endstop = inWidth; /* end at right */
		}
		else
		{
			col = inWidth -1;	/* start at right */
			endstop = -1;	/* end at left */
		}

		do
		{
			/* note the col+1; col means different things when refering
				to inbuf than to this_err[]/next_err[].
				*/
			sum = (inbuf[col] * FS_SCALE) + this_err[col +1];

			if ( sum > threshold) /* becomes white */
			{
#if 0	/* do nothing, already white (paper) */
				CLR_XY_BIT(outBuffer, outByteWidth, row, col);
#endif
#if 1
				sum = sum - threshold - HALF_FS_SCALE - 127;
#else
				sum = sum - threshold;
#endif
			}
			else
			{
				SET_XY_BIT(outBuffer, outByteWidth, row, col);
			}

			sum_forward =	(sum * 7) / 16;
			sum_d_forward =  (sum	) / 16;
			sum_d =		(sum * 5) / 16;
			sum_d_backward = (sum * 3) / 16;

			if ( fs_direction )
			{
				this_err[col + 2] += sum_forward;
				next_err[col	] += sum_d_backward;
				next_err[col + 1] += sum_d;
				next_err[col + 2] += sum_d_forward;
				col++;
			}
			else
			{
				this_err[col	] += sum_forward;
				next_err[col + 2] += sum_d_backward;
				next_err[col + 1] += sum_d;
				next_err[col	] += sum_d_forward;
				col--;
			}
		}
		while ( col != endstop);

		tmp_err = this_err; /* advance to the 'next' error row */
		this_err = next_err;
		next_err = tmp_err;
#if SERPENTINE
		fs_direction = !fs_direction;	/* change direction */
#endif
	}
	free(this_err);
	free(next_err);
	free(inbuf);
}


/*-----------------------------------------------------------------------------
	FS_HalfTone()
-----------------------------------------------------------------------------*/
XImage *FS_HalfTone(ximg_p)
	XImage *ximg_p;
{
	XImage *monoimg_p;

	monoimg_p = (XImage *)malloc(sizeof(XImage));
	memcpy(monoimg_p, ximg_p, sizeof(XImage));

	monoimg_p->bytes_per_line = EVEN_BYTES_FROM_BITS(monoimg_p->width);
	monoimg_p->depth = 1;		/* Sinn der Sache */
	monoimg_p->bits_per_pixel = 1;
	monoimg_p->data = malloc(monoimg_p->bytes_per_line * monoimg_p->height);

	FSdither((unsigned char *)ximg_p->data,
				monoimg_p->width,
				monoimg_p->height,
				16,
				(unsigned char *)monoimg_p->data);

	return(monoimg_p);
}

/*-----------------------------------------------------------------------------
	HalfTone()
-----------------------------------------------------------------------------*/
XImage *HalfTone(ximg)
	XImage *ximg;				 /* each pixel is 0..255 */
{
	register int super_row, super_col;
	register int x, y;
	int x_offset_in_row;
	int super_row_max, super_col_max, row_max, col_max;
	int this_cell_avg;
	XImage *out_img;
	register int *avg_buf;
	register char *out_buf;
	int out_buf_byte_wid;
	int in_buf_byte_wid;
	int x_index, y_index, data_ix;

	super_row_max = (ximg->height +(HT_CELL_HYT-1))/HT_CELL_HYT;
	super_col_max = (ximg->width  +(HT_CELL_WID-1))/HT_CELL_WID;
	row_max = ximg->height;
	col_max = ximg->width;
	in_buf_byte_wid = ximg->bytes_per_line;
	avg_buf = (int *)malloc((super_col_max +2) * sizeof(int));

	out_img = (XImage *)calloc(1, sizeof(XImage));
	out_img->width = ximg->width;
	out_img->height = ximg->height;
	out_img->depth = 1;			/* by definition... */
	out_img->bytes_per_line = EVEN_BYTES_FROM_BITS(ximg->width);
	out_buf_byte_wid = out_img->bytes_per_line;
	out_img->data = malloc(out_buf_byte_wid * ximg->height);
	memset(out_img->data, 0, (out_buf_byte_wid * ximg->height));
	out_buf = out_img->data;

	for (super_row=0; super_row<row_max; super_row++)
	{
		for(super_col=0; super_col<super_col_max+2; super_col++)
			avg_buf[super_col] = 0;

/*
				| super_col 0
				V
super_row n 1111 2222 3333 4444 5555 ....
				1111 2222 3333 4444 5555 ....
				1111 2222 3333 4444 5555 ....
				1111 2222 3333 4444 5555 ....

super_row n+1  1111 2222 3333 4444 5555 ....
				1111 2222 3333 4444 5555 ....
				1111 2222 3333 4444 5555 ....
				1111 2222 3333 4444 5555 ....

*/
		/* sum each of the cells from the data source */
		if (super_row % 2)
			x_offset_in_row = HT_CELL_WID/2;
		else
			x_offset_in_row = 0;

		for(super_col=0; super_col<super_col_max; super_col++)
		{
			for (y=0; y<HT_CELL_HYT; y++)
			{
				y_index = (super_row * HT_CELL_HYT) +y;
				if ((y_index >= 0) && (y_index < ximg->height))
				{
					for (x=0; x<HT_CELL_WID; x++)
					{
						x_index = (super_col *HT_CELL_WID) +x + x_offset_in_row;
						if ((x_index >= 0) && (x_index < ximg->width))
						{
							data_ix = (y_index * in_buf_byte_wid) + x_index;
							avg_buf[super_col] +=
								(int)(ximg->data[data_ix] & 0xFF);
						}
					}
				}
			}
		}

		/* now scan through the averages in this row setting the
			HT bits in the output */

		for(super_col=0; super_col<super_col_max; super_col++)
		{
/* there are 16 pixels each containing 0..255 values.  We can
	represent this group of 16 with one cell representing 0..16 (i.e.
	17 intensity levels => 16*256 = 4096, divide (equally) amongst 17
	levels; means dividing the sum by... 241 */
			this_cell_avg = (avg_buf[super_col] +7)/ 241;
			for (y=0; y<HT_CELL_HYT; y++)
			{
				y_index = (super_row *HT_CELL_HYT) +y;
				if ((y_index >= 0) && (y_index < out_img->height))
				{
					for (x=0; x<HT_CELL_HYT; x++)
					{
						x_index = (super_col *HT_CELL_WID) +x +x_offset_in_row;
						if ((x_index >= 0) && (x_index < out_img->width))
						{
							if (HT_Cell_even[x][y] < this_cell_avg)
							{
								/* turn it on! */
								SET_XY_BIT(out_buf, out_buf_byte_wid,
											y_index, x_index);
							}
							else
							{
								/* turn it off */
								CLR_XY_BIT(out_buf, out_buf_byte_wid,
											y_index, x_index);
							}
						}
					}
				}
			}
		}
	}

	free(avg_buf);
	return(out_img);
}


/*-----------------------------------------------------------------------------
	LinearGradientXImage()
-----------------------------------------------------------------------------*/
XImage *LinearGradientXImage(subattr, box)
	sub_attr_type *subattr;
	TmPoint32 box[];
{
	int wid, hyt, rhyt ;
	double fg_intensity, bg_intensity;
	register int col, row;
	XImage *X;
	double delta, r ;
	unsigned char intensity;
	int yoff = subattr->yoff ;
	int y1, y2 ;
	unsigned char *p ;
	double angle, x, y, xr, yr ;
	double sina, cosa ;

	wid = box[1].x - box[0].x ;
	hyt = box[1].y - box[0].y ;
	rhyt = box[3].y - box[2].y ;

	/* create an XImage rectangle */
	X = (XImage *)malloc(sizeof(XImage));
	X->height = hyt;
	X->width = wid;
	X->depth = 8;
	X->bytes_per_line = wid;
	X->format = ZPixmap;
	X->xoffset = 0;

	X->data = malloc(X->bytes_per_line * X->height);

	/* check use_black/use_white */
	Intens(subattr, &fg_intensity, &bg_intensity) ;

	/* Fill rectangle from top to bottom:
		0  to y1:  fg
		y1 to y2:  gradient fg to bg
		y2 to end: bg
	*/

	y1 = box[2].y ;
	y2 = box[3].y ;
	if (yoff > 0)
	{
		if (yoff > 100)
			yoff = 100 ;
		y1 += (yoff * rhyt) / 100 ;
	}
	else if (yoff < 0)
	{
		if (yoff < -100)
			yoff = -100 ;
		y2 -= (-yoff * rhyt) / 100 ;
	}

	if (y1 != y2)
		delta = ( bg_intensity - fg_intensity ) / (double)(y2-y1);
	else
		delta = 0.0 ;
	angle = ((double)(-subattr->ang) * M_PI) / 180.0 ;
	sina = sin(angle) ;
	cosa = cos(angle) ;


	for (row = 0 ; row < hyt ; row++)
	{
		p = (unsigned char *)&X->data[row*wid] ;
		for (col = 0 ; col < wid; col++)
		{
			x = (double)col;
			y = (double)(hyt - (row+1)) ;
			yr = x*sina + y*cosa ;
			if (yr > y2)
				intensity = (unsigned char)((int)bg_intensity) ;
			else if (yr < y1)
				intensity = (unsigned char)((int)fg_intensity) ;
			else
			{
				r = (yr-y1) ;
				intensity = (unsigned char)((int)(fg_intensity + (delta * r))) ;
			}
			*p++ = intensity ;
		}
	}
						 
	return X ;
}

/*-----------------------------------------------------------------------------
	BilinearGradientXImage()
-----------------------------------------------------------------------------*/
XImage *BilinearGradientXImage(subattr, box)
	sub_attr_type *subattr;
	TmPoint32 box[];
{
	int wid, hyt, rhyt ;
	double fg_intensity, bg_intensity;
	register int col, row;
	XImage *X;
	double delta, r;
	double delta1, delta2;
	unsigned char intensity;
	int yoff = subattr->yoff ;
	int y1, y2, y3, y4 ;
	unsigned char *p ;
	double angle, x, y, xr, yr ;
	double sina, cosa ;

	wid = box[1].x - box[0].x ;
	hyt = box[1].y - box[0].y ;
	rhyt = box[3].y - box[2].y ;

	/* create an XImage rectangle */
	X = (XImage *)malloc(sizeof(XImage));
	X->height = hyt;
	X->width = wid;
	X->depth = 8;
	X->bytes_per_line = wid;
	X->format = ZPixmap;
	X->xoffset = 0;

	X->data = malloc(X->bytes_per_line * X->height);

	/* check use_black/use_white */
	Intens(subattr, &fg_intensity, &bg_intensity) ;

	/* yr =
	    y1 to y2: gradient bg to fg
		y2 to y3: gradient fg to bg
	*/

	y1 = box[2].y ;
	y2 = y1 + rhyt/2 ;
	y3 = y1 + rhyt ;

	if (yoff > 0)
	{
		if (yoff > 100)
			yoff = 100 ;
		y2 -= (yoff*rhyt/2)/100 ;
	}
	else if (yoff < 0)
	{
		if (yoff < -100)
			yoff = -100 ;
		y2 += (-yoff*rhyt/2)/100 ;
	}

	angle = ((double)(-subattr->ang) * M_PI) / 180.0 ;
	sina = sin(angle) ;
	cosa = cos(angle) ;
	if (y2 != y1)
		delta1 = ( fg_intensity - bg_intensity ) / (double)(y2-y1) ;
	else
		delta1 = 0.0 ;
	if (y2 != y3)
		delta2 = ( bg_intensity - fg_intensity ) / (double)(y3-y2) ;
	else
		delta2 = 0.0 ;

	for (row = 0 ; row < hyt ; row++)
	{
		p = (unsigned char *)&X->data[row*wid] ;
		for (col = 0 ; col < wid; col++)
		{
			x = (double)col;
			y = (double)(hyt - (row+1)) ;
			yr = x*sina + y*cosa ;
			if (yr < y2)
			{
				r = yr - y1 ;
				intensity = (unsigned char)((int)(bg_intensity + (delta1 * r))) ;
			}
			else
			{
				r = yr - y2 ;
				intensity = (unsigned char)((int)(fg_intensity + (delta2 * r))) ;
			}
			*p++ = intensity ;
		}
	}
	return X;
}

/*-----------------------------------------------------------------------------
	RectGradientXImage()
-----------------------------------------------------------------------------*/
XImage *RectGradientXImage(subattr, box)
	sub_attr_type *subattr;
	TmPoint32 box[];
{
	int wid, hyt, rwid, rhyt ;
	double fg_intensity, bg_intensity;
	register int col, row;
	XImage *X;
	unsigned char *Y ;
	int v_dist, h_dist;
	int startx, endx;
	int starty, endy;
	int xoff, yoff ;
	int xc, yc, k ;
	double delta_x, delta_y ;
	double t1, t2;
	unsigned char intensity ;
	unsigned char *p ;
	double angle, x, y, xr, yr ;
	double sina, cosa ;

	wid = box[1].x - box[0].x ;
	hyt = box[1].y - box[0].y ;
	rwid = box[3].x - box[2].x ;
	rhyt = box[3].y - box[2].y ;

	X = (XImage *)malloc(sizeof(XImage));
	X->height = hyt;
	X->width = wid;
	X->depth = 8;
	X->bits_per_pixel = 8;
	X->bytes_per_line = wid;
	X->format = ZPixmap;
	X->xoffset = 0;

	X->data = malloc(X->bytes_per_line * X->height);
	memset(X->data, 0, (X->bytes_per_line * X->height));
	
	xoff = subattr->xoff ;
	yoff = subattr->yoff ;

	/* check use_black/use_white */
	Intens(subattr, &fg_intensity, &bg_intensity) ;

	/* build the rectangle in the rotated bounding box */
	Y = (unsigned char *)malloc(rwid * rhyt);
	memset(Y, (unsigned char)bg_intensity, rwid * rhyt);
	
	/* (xc, yc) is the center of the rectangle */
	xc = wid / 2 ;
	yc = hyt / 2 ;

	/* translate (xc, yc) to rotated rectangle (now use (0,0) for ULH corner) */
	angle = ((double)(-subattr->ang) * M_PI) / 180.0 ;
	sina = sin(angle) ;
	cosa = cos(angle) ;
	xr = xc*cosa - yc*sina - box[2].x ;
	yr = xc*sina + yc*cosa - box[2].y ;
	xc = xr ;
#if 1
	yc = rhyt - yr ;
#else
	yc = yr ;
#endif

	/* now apply offset to (xc, yc) */
	if (xoff > 0)
	{
		if (xoff > 200)
			xoff = 200 ;
		xc += (xoff*wid/2)/100 ;
	}
	else if (xoff < 0)
	{
		if (xoff < -200)
			xoff = -200 ;
		xc -= (-xoff*wid/2)/100 ;
	}

	if (yoff > 0)
	{
		if (yoff > 200)
			yoff = 200 ;
		yc += (yoff*hyt/2)/100 ;
	}
	else if (yoff < 0)
	{
		if (yoff < -200)
			yoff = -200 ;
		yc -= (-yoff*hyt/2)/100 ;
	}

#if 0
	/* compute intensity deltas in x and y direction */
	if (xc < rwid/2)
		delta_x = ( fg_intensity - bg_intensity ) / (double)(rwid - xc);
	else
		delta_x = ( fg_intensity - bg_intensity ) / (double)(xc);

	if (yc < rhyt/2)
		delta_y = ( fg_intensity - bg_intensity ) / (double)(rhyt - yc);
	else
		delta_y = ( fg_intensity - bg_intensity ) / (double)(yc);
#endif

	/* now fill the 4 quadrants defined by lines from (xc, yc) to corners*/

	/* UPPER */
	if (yc > 0)
	{
		t1 = (double)xc / yc;
		t2 = (double)(rwid-xc) / yc ;
#if 1
		delta_y = ( fg_intensity - bg_intensity ) / (double)(yc);
#endif
		starty = 0 ;
		endy = min(yc, rhyt) ;
		for (row = starty ; row < endy ; row++)
		{
			intensity = (unsigned char)((int)(bg_intensity + (yc - row) * delta_y));
			startx = (int)(row * t1) ;
			if (startx < 0)
				startx = 0 ;
			endx = rwid - (int)(row * t2) ;
			if (endx > rwid)
				endx = rwid ;
			for (col = startx ; col < endx ; col++)
				Y[(row * rwid) + col] = intensity ;
		}
	}

	/* LOWER */
	if (yc < rhyt)
	{
		t1 = (double)(xc) / (rhyt - yc);
		t2 = (double)(rwid-xc) / (rhyt - yc) ;
#if 1
		delta_y = ( fg_intensity - bg_intensity ) / (double)(rhyt-yc);
#endif
		starty = max(0, yc) ;
		endy = rhyt ;
		for (row = starty ; row < endy ; row++)
		{
			v_dist = rhyt - (row+1) ;
			intensity = (unsigned char)((int)(bg_intensity + (row - yc) * delta_y)) ;
			startx = (int)(v_dist * t1) ;
			if (startx < 0)
				startx = 0 ;
			endx = rwid - (int)(v_dist * t2) ;
			if (endx > rwid)
				endx = rwid  ;
			for (col = startx ; col < endx ; col++)
				Y[(row * rwid) + col] = intensity ;
		}
	}

	/* LEFT */
	if (xc > 0)
	{
		t1 = (double)yc / xc;
		t2 = (double)(rhyt - yc) / xc ;
#if 1
		delta_x = ( fg_intensity - bg_intensity ) / (double)(xc);
#endif
		startx = 0 ;
		endx = min(xc, rwid) ;
		for (col = startx ; col < endx ; col++)
		{
			intensity = (unsigned char)((int)(bg_intensity + (xc - col) * delta_x)) ;
			starty = (int)(t1 * col) ;
			if (starty < 0)
				starty = 0 ;
			endy = rhyt - (int)(t2 * col) ;
			if (endy > rhyt)
				endy = rhyt ;
			for (row = starty ; row < endy ; row++)
				Y[(row * rwid) + col] = intensity ;
		}
	}

	/* RIGHT */
	if (xc < rwid)
	{
		t1 = (double)(yc) / (rwid-xc) ;
		t2 = (double)(rhyt - yc) / (rwid-xc) ;
#if 1
		delta_x = ( fg_intensity - bg_intensity ) / (double)(rwid - xc);
#endif
		startx = max(xc, 0) ;
		endx = rwid ;
		for (col = startx ; col < endx ; col++)
		{
			h_dist = rwid - (col+1) ;
			intensity = (unsigned char)((int)(bg_intensity + (col - xc) * delta_x)) ;
			starty = (int)(t1 * h_dist);
			if (starty < 0)
				starty = 0 ;
			endy = rhyt - (int)(t2 * h_dist);
			if (endy > rhyt)
				endy = rhyt ;
			for (row = starty ; row < endy ; row++)
				Y[(row * rwid) + col] = intensity ;
		}
	}

#if 0
	{
	XImage *Z;

	Z = (XImage *)malloc(sizeof(XImage));
	Z->height = rhyt;
	Z->width = rwid;
	Z->depth = 8;
	Z->bits_per_pixel = 8;
	Z->bytes_per_line = rwid;
	Z->format = ZPixmap;
	Z->xoffset = 0;
	Z->data = Y ;
	return Z ;
	}
#endif
	
	/* now transfer that to the original bounding box */
	for (row = 0 ; row < hyt ; row++)
	{
		y = (double)(hyt - (row+1)) ;
		p = (unsigned char *)&X->data[row*wid] ;
		for (col = 0 ; col < wid; col++)
		{
			x = (double)col;
			xr = x*cosa - y*sina - box[2].x ;
			yr = x*sina + y*cosa - box[2].y ;
			if ( yr < 0 )
				yr = 0;
			else if ( yr >= rhyt )
				yr = rhyt - 1;
			if ( xr < 0 )
				xr = 0;
			else if ( xr >= rwid )
				xr = rwid - 1;
			k = (int)(rhyt - (yr+1)) * rwid + (int)xr ;
			*p++ = Y[k] ;
		}
	}
    free(Y);
	return(X);
}

/*-----------------------------------------------------------------------------
	RadialGradientXImage()
-----------------------------------------------------------------------------*/
XImage *RadialGradientXImage(subattr, box)
	sub_attr_type *subattr;
	TmPoint32 box[];
{
	int wid, hyt ;
	double fg_intensity, bg_intensity;
	register int col, row;
	XImage *X;
	unsigned char intensity ;
	int xc, yc ;
	int xoff, yoff ;
	int yd2, xd2 ;
	int xc2, yc2, pxc2, pyc2 ;
	double d, td ;
	unsigned char *p ;

	wid = box[1].x - box[0].x ;
	hyt = box[1].y - box[0].y ;
	X = (XImage *)malloc(sizeof(XImage));
	X->height = hyt;
	X->width = wid;
	X->depth = 8;
	X->bits_per_pixel = 8;
	X->bytes_per_line = wid;
	X->format = ZPixmap;
	X->xoffset = 0;

	X->data = malloc(X->bytes_per_line * X->height);

	/* check use_black/use_white */
	Intens(subattr, &fg_intensity, &bg_intensity) ;
	xoff = subattr->xoff ;
	yoff = subattr->yoff ;

	/* compute (xc, yc), center point */
	xc = wid / 2 ;
	yc = hyt / 2 ;

	if (xoff > 0)
	{
		if (xoff > 100)
			xoff = 100 ;
		xc += (xoff*xc)/100 ;
	}
	else if (xoff < 0)
	{
		if (xoff < -100)
			xoff = -100 ;
		xc -= (-xoff*xc)/100 ;
	}

	if (yoff > 0)
	{
		if (yoff > 100)
			yoff = 100 ;
		yc += (yoff*yc)/100 ;
	}
	else if (yoff < 0)
	{
		if (yoff < -100)
			yoff = -100 ;
		yc -= (-yoff*yc)/100 ;
	}

	if (xoff >= 0)
	{
		if (yoff >= 0)
			td = sqrt((double)(xc*xc + yc*yc)) ;
		else
			td = sqrt((double)((xc*xc) + (hyt-yc)*(hyt-yc))) ;
	}
	else
	{
		if (yoff >= 0)
			td = sqrt((double)((wid-xc)*(wid-xc) + (yc*yc))) ;
		else
			td = sqrt((double)((wid-xc)*(wid-xc) + (hyt-yc)*(hyt-yc))) ;
	}

	p = (unsigned char *)X->data ;

	/* fill with a gradient from bg color to fg color */
	for (row = 0; row < hyt; row++)
	{
		yd2 = (row - yc) * (row - yc) ;

		for (col = 0 ; col < wid ; col++)
		{
			xd2 = (col - xc) * (col - xc) ;

			/* d = distance from center */
			d = (double)sqrt((double)(xd2 + yd2));

			intensity = (unsigned char)((int)(bg_intensity +
							(fg_intensity - bg_intensity) * (d/td))) ;
			*p++ = intensity ;
		}
	}
	return(X);
}


/*-----------------------------------------------------------------------------
	pfMaskXImage()

	masks Xp (1 or 8 bit deep) against Xmaskp (1 bit) and returns a new
	XImage.
-----------------------------------------------------------------------------*/
XImage *pfMaskXImage(Xp, Xmaskp)
		XImage *Xp;
		XImage *Xmaskp;
{
	XImage *Xmaskdp;
	int data_byt_width, y;
	register int x;
	register int mask_byte;
	int mask_byt_width;
	unsigned char mask_char;
	register char *src_line_data;
	register char *dest_line_data;
	register char *mask_line_data;


	Xmaskdp = (XImage *)malloc(sizeof(XImage));
	memcpy((char *)Xmaskdp, (char *)Xp, sizeof(XImage));
	Xmaskdp->data = malloc(Xp->bytes_per_line *
							Xp->height);

	data_byt_width = Xp->bytes_per_line;
	mask_byt_width = Xmaskp->bytes_per_line;
	src_line_data = Xp->data;
	dest_line_data = Xmaskdp->data;
	mask_line_data = Xmaskp->data;

	if ((Xmaskp != NULL) && (Xmaskp->data != NULL))
	{
		if (Xp->bits_per_pixel == 8)
		{
			for (y=0; y<Xmaskp->height; y++)
			{
								/* for each byte in the mask */
				for (mask_byte=0; mask_byte < Xmaskp->bytes_per_line; mask_byte++)
				{
					mask_char =
						(Xmaskp->data[(y*mask_byt_width) + mask_byte] & 0xFF);
					if (mask_char == (char)0)
					{
								/* mask is clear, 8 xxxx of nothing */
						for (x=0;
								(x<8) && ((mask_byte + x) < Xp->width);
								x++)
						{
							dest_line_data[mask_byte + x] = (char)0;
						}
					}
					else
					{
						if (mask_char == (unsigned char)0xFF)
						{
								/* mask is 11111111, 8 xxxx copied */
							for (x=0;
									(x<8) && ((mask_byte + x) < Xp->width);
									x++)
							{
								dest_line_data[mask_byte + x] =
									src_line_data[mask_byte + x];
							}
						}
						else
						{
								/* some copied, some cleared */
							for (x=0;
									(x<8) && ((mask_byte + x) < Xp->width);
									x++)
							{
								if ((mask_char & (unsigned char)0x80 >> x))
								{
									dest_line_data[mask_byte + x] =
										src_line_data[mask_byte + x];
								}
								else
									dest_line_data[mask_byte + x] = (char)0;
							}
						}
					}
				}
				dest_line_data += data_byt_width;
				src_line_data += data_byt_width;
				mask_line_data += mask_byt_width;
			}
		}
		else
		{
			for (y=0; y<Xmaskp->height; y++)
			{
								/* for each byte in the mask */
				for (mask_byte=0; mask_byte < Xmaskp->bytes_per_line; mask_byte++)
				{
					mask_char = mask_line_data[mask_byte] & 0xFF;
					dest_line_data[mask_byte] =
						src_line_data[mask_byte] & mask_char;
				}
				dest_line_data += data_byt_width;
				src_line_data += data_byt_width;
				mask_line_data += mask_byt_width;
			}
		}
	}
	else
	{
		memcpy(Xmaskdp->data, Xp->data, (Xp->height * Xp->bytes_per_line));
	}
	return(Xmaskdp);
}

/*-----------------------------------------------------------------------------
	pfFreeXImage()
-----------------------------------------------------------------------------*/
void pfFreeXImage(Xp)
		XImage *Xp;
{

	if (Xp != NULL)
	{
		if (Xp->data != NULL)
		{
			free(Xp->data);
			Xp->data = NULL;
		}
		free(Xp);
	}
}

/*-----------------------------------------------------------------------------
	Intens()

	computes fg and bg intensity in rang 0..255
-----------------------------------------------------------------------------*/
static void Intens(subattr, fg_intensity, bg_intensity)
	sub_attr_type *subattr;
	double *fg_intensity, *bg_intensity ;
{

	if (PF.PfColordev)
	{
		*fg_intensity = 255.0 ;
		*bg_intensity = 0.0 ;
	}
	else
	{
		*fg_intensity = 255.0 * IntensFromCMYKColor(subattr->fg.c,
											subattr->fg.m,
											subattr->fg.y,
											subattr->fg.k, 255);
		*bg_intensity = 255.0 * IntensFromCMYKColor(subattr->bg.c,
											subattr->bg.m,
											subattr->bg.y,
											subattr->bg.k, 255);
	}
}
