/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfint.h
   Applix ASC version: 37
   Date: Mon Nov 16 13:56:49 1998
*/
#include	<basicwin.h>
#include	<pfdef.h>


#define TXPrintAtts	(TXAllAtts ^ (TXbullet | TXindent))

#define NEWLINE ('\n')
#define TABCHAR ('\t')
#define CHAR_IS_DIMENSIONLESS(cc) (( cc == NEWLINE ) || ( cc == TABCHAR ))

/*-----------------------------------------------------------------------------
	Print information structure
-----------------------------------------------------------------------------*/
typedef struct {
	char	*pclass ;
	char	*printer ;
	char	*infile ;
	char	*outfile ;
	char	*docname ;
	char	*command ;
	int		device ;
	int		print_to_file ;
	int		banner ;
	int		copies ;
	int		collate ;
	int		color ;
	int		reverse ;
	int		duplex ;
	char	*source_all ;
	char	*source_1st ;
	char	*source_last ;
	char	*dest_all ;
	char	*dest_1st ;
	char	*dest_last ;
	data_ptr source ;
	data_ptr dest ;
	data_ptr size ;
	char	*PPDfile ;
	data_ptr duplex_code ;
	long	date ;
	char	*user ;
	char	*host ;
	} PF_info ;

/*-----------------------------------------------------------------------------
	PF structure
-----------------------------------------------------------------------------*/

typedef struct {
	struct wind_struct PfPage;

#if AXMS
	HDC		PfPrinterHDC;			/* printer device context. */
	HRGN	PfClipRgn;				/* handle to clipping region. */
	HANDLE	PfPrinterHandle;		/* handle to printer device */
#endif

	TabInfo PfTabInfo;

/* graphics and text coordinate system conversion factors */

	int		PfGXPixFactor, PfGYPixFactor;
	int		PfGXMilFactor, PfGYMilFactor;
	int		PfTXPixFactor, PfTYPixFactor;
	int		PfTXMilFactor, PfTYMilFactor;

	int		PfDotsPerInch;

	AX_BOOLEAN PfLandscape;
	int		PfPageMilWidth;
	int		PfPageMilHeight;
	int		PfPagePixWidth;
	int		PfPagePixHeight;
	int		PfXPixOffset;
	int		PfYPixOffset;

	int		PfLeftMarginMil;

#ifdef DEBUGCOMP
	FILE	*PFdfp;				/* debug file pointer */
	AX_BOOLEAN PFdebug;			/* debug switch */
#endif

	AX_BOOLEAN PfUnderAC;		/* true if running under AC */
	int		PfDev;					/* device setup number */
	int		PfCopies;				/* number of copies to print		*/
	AX_BOOLEAN PfReverse;		/* true if printing pages in reverse order */
	AX_BOOLEAN PfJob_done;		/* true when end of file and done with repeats */
	AX_BOOLEAN PfAbortInProgress;	/* True if SIGTERM received */
	AX_BOOLEAN DevDirect;		/* writing to hardware port */

	int		PfHardX, PfHardY;		/* Hardware print		position */

	float	PfPatXMul;			/* Pattern multipliers */
	float	PfPatYMul;

	int		PfDev_align;			/* device char positioning alignment */

	int		inputTray ;
	int		outputTray ;
	int		duplex ;

	char	*PfUser_name;
	char	*PfHost_name;
	char	*PfDoc_name;
	char	*PfDoc_date;
	char	*PfShell_command;
	char	*PfPrinterName;

	AX_BOOLEAN PfBanner;		/* true if banner is to be printed. */

	int		PfSe_version;
	long	PfFileMark;			/* Saved location in input file. */

	int		PfPageCat;
	int		PfPageNum;

	int		PfWin_fg_color;
	int		PfWin_bg_color;
	int		PfColordev;
	int		PfUlForColor;

/* text delay globals (text delay for mixed text and graphic modes) */

	int		PfTx_ynext;			/* next y row of inset text */
	int		PfDelay_all;		/* TRUE if all text is delayed */
	int		PfIn_text;			/* FALSE at NewPage and GrfEnd */
	int		PfText_ok;			/* FALSE during graphics on non-graphic devs */


/* Polygon shape variables for use with plotters */

	int		PfPoly_shape;
	int		PfPoly_radius;
	int		PfStart_angle;
	int		PfArc_angle;
	TmPoint	PfWedge_center;


/* pf70prim global vars. declared here so that they can
	be initialized at start of subroutine */
	int		ClipGSaveLevel;
	int		D70FontNumber;

	char	*InsertSpec;

	int		small;
	int		pagetop;
	int		eightbit;
	int		eightbitDingbats;
	int		PfNumDocPages;
	int		PfNumTotalPages;
	int		left; 
	int		right;
	int		top;
	int		bottom;
	int		cur_thick;
	int		cur_style;
	int		monospace;
	int		PrintFontNameIndex;
	int		PrintFontEncodingIndex;

	int		fontNeeded;

	int		PfNumEPSIncluded;
	int		PfNumEPSAddedFonts;
	int		PfNumEPSProcessed;

/* pfpol.c externs */
	int		CurPat;

/* pfinf.c externs */
	FILE	*PFifdes;

/* pfdispat */
	AX_BOOLEAN envset;
	AX_BOOLEAN abend;
	AX_BOOLEAN in_remove;
	AX_BOOLEAN out_remove;
	int		PfPrintMultiCopiesInLang; /* if TRUE: use #copies with showpage */
	int		PfSEByteAlignment;		/* to keep track of se file byte alignment */


/* pfwin stuff */
	int		FirstPage;
	int		LastPage;
	int		FirstCreate;

	int		Pferrno;
	char	*Pferror_message ;


/* for pfline.c */
	char	*PfDolinePatStart, *PfDolinePatPtr;
	int		PfDolineBlack, PfDolineCount;

	PF_info	pi ;

	int		scount ;
	void	*slist ;
	
} PF_Global;

extern PF_Global PF;

