/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfout.c
   Applix ASC version: 29
   Date: Mon Nov 16 13:56:49 1998
*/
#include <stdio.h>
#include <pf.h>
#include <aos.h>
#include <paratm.h>
#include <paratmgr.h>
#include <devsetup.h>
#include <pfint.h>
#include <sff.h>

typedef struct {
	short x, y;
} short_point;

#include <pfout.p>

/*-----------------------------------------------------------------------------
	read the polyline attributes from the se file
-----------------------------------------------------------------------------*/
pf_read_poly_attr(attr)
attr_type *attr;
{
	int i, j, scan_bytes, ncolors, tmp;

	/* read in the fill attributes */
	memset((char *)attr, 0, sizeof(attr_type));
	attr->fill_flag = InWord();
	attr->f_type = InWord();
	attr->f_pat = InWord();
	attr->f_ang = InWord();
	attr->f_xoff = InWord();
	attr->f_yoff = InWord();

	attr->s_type = InWord();
	attr->s_xoff = InWord();
	attr->s_yoff = InWord();

	/* read any passed in fill fill pattern */
	attr->f_dhyt = InWord();
	attr->f_dwid = InWord();
	attr->f_ddep = InWord();
	scan_bytes = 2*((attr->f_dwid * attr->f_ddep + 15) / 16);
	if ((tmp = attr->f_dhyt * scan_bytes) > 0)
	{
		attr->f_data = (char *)malloc(tmp);
#if 0
		for ( i=0; i<attr->f_dhyt; i++ )
			for ( j=0; j<scan_bytes; j++ )
				attr->f_data[i*scan_bytes + j] = (char)InByte();
#endif
		InCString(attr->f_data, tmp) ;
	}
	InAlign();
	ncolors = InWord() ;
	if (ncolors > 0) {
		attr->f_cmap = (cmap_type *)malloc(sizeof(cmap_type)) ;
		attr->f_cmap->n_colors = ncolors ;
		tmp = ncolors * sizeof(color_type) ; 
		attr->f_cmap->color = (color_type *)malloc(tmp) ;
		InCString((char *)attr->f_cmap->color, tmp) ;
		}
	InAlign();

	/* read the line style */
	attr->l_style = InWord ();
	attr->l_wid = InWord ();
#if 0	/* leave it in mils */
	if ( attr->l_wid > 1 )
	{
		attr->l_wid = xmiltopix(attr->l_wid); /* rounds DOWN! */
		if (attr->l_wid < 1)
			attr->l_wid = 1;
	}
#endif
	attr->l_cap = InWord ();

	/* read the line attributes */
	attr->line_flag = InWord();
	attr->l_type = InWord ();
	attr->l_pat = InWord ();
	attr->l_ang = InWord ();
	attr->l_xoff = InWord ();
	attr->l_yoff = InWord ();
 
	/* read any passed in line fill pattern */
	attr->l_dhyt = InWord ();
	attr->l_dwid = InWord ();
	attr->l_ddep = InWord ();
	scan_bytes = 2*((attr->l_dwid * attr->l_ddep + 15) / 16);
	if ((tmp = attr->l_dhyt * scan_bytes) > 0)
	{
		attr->l_data = (char *)malloc(tmp);
#if 0
		for ( i=0; i<attr->l_dhyt; i++ )
			for ( j=0; j<scan_bytes; j++ )
				attr->l_data[i*scan_bytes + j] = (char)InByte();
#endif
		InCString(attr->l_data, tmp) ;
	}
	InAlign();
	ncolors = InWord() ;
	if (ncolors > 0) {
		attr->l_cmap = (cmap_type *)malloc(sizeof(cmap_type)) ;
		attr->l_cmap->n_colors = ncolors ;
		tmp = ncolors * sizeof(color_type) ; 
		attr->l_cmap->color = (color_type *)malloc(tmp) ;
		InCString((char *)attr->l_cmap->color, tmp) ;
		}
 
	/* read the color attributes */
	attr->ffg.c = InWord ();
	attr->ffg.m = InWord ();
	attr->ffg.y = InWord ();
	attr->ffg.k = InWord ();
	attr->ffg.see_thru = InWord ();

	attr->fbg.c = InWord ();
	attr->fbg.m = InWord ();
	attr->fbg.y = InWord ();
	attr->fbg.k = InWord ();
	attr->fbg.see_thru = InWord ();

	attr->lfg.c = InWord ();
	attr->lfg.m = InWord ();
	attr->lfg.y = InWord ();
	attr->lfg.k = InWord ();
	attr->lfg.see_thru = InWord ();

	attr->lbg.c = InWord ();
	attr->lbg.m = InWord ();
	attr->lbg.y = InWord ();
	attr->lbg.k = InWord ();
	attr->lbg.see_thru = InWord ();

	attr->sfg.c = InWord ();
	attr->sfg.m = InWord ();
	attr->sfg.y = InWord ();
	attr->sfg.k = InWord ();
	attr->sfg.see_thru = InWord ();
 
	/* read the override transfer mode */
	attr->mode = InWord ();

	/* read the absolute fill info */
	attr->fill_abs = InWord ();
	attr->line_abs = InWord ();
	attr->x_off = InWord ();
	attr->y_off = InWord ();

	/* read the other flags */
	attr->line_first = InWord ();
	attr->use_ffg = InWord ();
	attr->use_fbg = InWord ();
	attr->use_lfg = InWord ();
	attr->use_lbg = InWord ();
}

/*-----------------------------------------------------------------------------
	TmPolyline function
-----------------------------------------------------------------------------*/
pf_pline()
{
	int i, j, found, x, y, cmaxX, cminX, cmaxY, cminY;
	int nPolys, nClips;
	attr_type attr;
	poly_type *pol = (poly_type *) 0;
	poly_type *cpol = (poly_type *) 0;
	TmRect32 clip ;

	/* read in the poly data (converting from mills to printer pels) */
	nPolys = InWord ();
	pol = (poly_type *)malloc(nPolys * sizeof(poly_type));
	for ( i=0; i<nPolys; i++ )
	{
		/* read in the poly points */
		pol[i].nPoints = InWord ();
		pol[i].points = (TmPoint32 *)malloc(pol[i].nPoints*sizeof(TmPoint32));
		for ( j=0; j<pol[i].nPoints; j++ )
		{
			pol[i].points[j].x = xmiltopix(InInt ());
			pol[i].points[j].y = ymiltopix(InInt ());
		}

		/* read in the control points */
		pol[i].nCtrls = InWord ();
		pol[i].controls = (TmPoint32 *)malloc(pol[i].nCtrls*sizeof(TmPoint32));
		for ( j=0; j<pol[i].nCtrls; j++ )
		{
			pol[i].controls[j].x = xmiltopix(InInt ());
			pol[i].controls[j].y = ymiltopix(InInt ());
		}
	}

	/* read in the fill attributes */
	pf_read_poly_attr(&attr);

	/* clipping rectangle */
	InBlock((char *)&clip, sizeof(clip)) ;

	/* get the clipping info */
	nClips = InWord ();
	if ( nClips > 0 )
	{
		cpol = (poly_type *)malloc(nClips * sizeof(poly_type));
		for ( found =FALSE, i=0; i<nClips; i++ )
		{
			/* read in the poly points */
			cpol[i].nPoints = InWord ();
			cpol[i].points = (TmPoint32 *)malloc(cpol[i].nPoints*sizeof(TmPoint32));
			for ( j=0; j<cpol[i].nPoints; j++ )
			{
				x = InInt ();
				y = InInt ();
				if ( found == FALSE )
				{
					found = TRUE;
					cmaxX = cminX = x;
					cmaxY = cminY = y;
				}
				if ( x < cminX )
					cminX = x;
				else if ( x > cmaxX )
					cmaxX = x;
				if ( y < cminY )
					cminY = y;
				else if ( y > cmaxY )
					cmaxY = y;

				cpol[i].points[j].x = x;
				cpol[i].points[j].y = y;
			}

			/* read in the control points */
			cpol[i].nCtrls = InWord ();
			cpol[i].controls = (TmPoint32 *)malloc(pol[i].nCtrls*sizeof(TmPoint32));
			for ( j=0; j<cpol[i].nCtrls; j++ )
			{
				cpol[i].controls[j].x = InInt ();
				cpol[i].controls[j].y = InInt ();
			}
		}

		/* take the intersection of clip and clip path */
		if ( clip.x_position > cminX ) 
			cminX = clip.x_position;
		if ( clip.y_position > cminY ) 
			cminY = clip.y_position;
		if ( cmaxX > (clip.x_position+clip.x_size))
			cmaxX = clip.x_position+clip.x_size;
		if ( cmaxY > (clip.y_position+clip.y_size))
			cmaxY = clip.y_position+clip.y_size;
		clip.x_position = cminX;
		clip.y_position = cminY;
		clip.x_size = cmaxX - cminX;
		clip.y_size = cmaxY - cminY;
	}

	/* draw the poly */
	DrawLine(nPolys, &attr, pol, &clip);

	/* free local data and return */
	if (pol) {
		for ( i=0; i<nPolys; i++ )
		{
			if ( pol[i].nPoints )
				free ( pol[i].points );
			if ( pol[i].nCtrls )
				free ( pol[i].controls );
		}
		free(pol);
	}

	if (cpol) {
		for ( i=0; i<nClips; i++ )
		{
			if ( cpol[i].nPoints )
				free ( cpol[i].points );
			if ( cpol[i].nCtrls )
				free ( cpol[i].controls );
		}
		free(cpol);
	}
	if (attr.f_data) {
		free(attr.f_data);
	}
	if (attr.l_data) {
		free(attr.l_data);
	}
}


/*-----------------------------------------------------------------------------
	TmArc function
-----------------------------------------------------------------------------*/
pf_Arc()
{
	TmRect bound_box;
	TmRect32 bbox;
	int start_angle;
	int arc_angle;
	int fill_pat;
	int point_count;
	int cpF;
	TmPoint32 *pp, points[65];
#if 0
	short_point local_p[65];
#else
	TmPoint local_p[65];
#endif
	attr_type attr;
	poly_type pol;

	int indx;

	/* read the ellipse data */
	bound_box.x_position = xmiltopix (InWord ());
	bound_box.y_position = ymiltopix (InWord ());
	bound_box.x_size = xmiltopix (InWord ());
	bound_box.y_size = ymiltopix (InWord ());
	start_angle = InWord ();
	arc_angle = InWord ();

	/* read in the fill attributes */
	pf_read_poly_attr(&attr);

	/* express as curves (not lines) for full ellipses/circles */
	if ((arc_angle == 3600 || bound_box.x_size == bound_box.y_size) &&
		(PF.PfDev == 45))
	{
		bbox.x_position = (int) bound_box.x_position;
		bbox.y_position = (int) bound_box.y_position;
		bbox.x_size = (int) bound_box.x_size;
		bbox.y_size = (int) bound_box.y_size;
		start_angle = convertAngle(start_angle + arc_angle);
		DrawArc(&bbox, start_angle, arc_angle, &attr);
	}
	else /* reduce the arc to a polyline */
	{
		cpF = PfEllipse(&bound_box, start_angle, arc_angle, local_p,
						&point_count);
		for (indx = 0; indx < point_count ; indx ++) {
			points[indx].x = (int)local_p[indx].x;
			points[indx].y = (int)local_p[indx].y;
		}

		pp = points;
		if (cpF && !attr.fill_flag)
		{
			/* PfEllipse() returned center points, ignore if not filled */
			pp = &points[1];
			point_count -= 2;
		}

		/* do the polyline */
		pol.nPoints = point_count;
		pol.points = pp;
		pol.nCtrls = 0;
		pol.controls = (TmPoint32 *)NULL;
		DrawLine(1, &attr, &pol, NULL);
	}
	if (attr.l_data) {
		free(attr.l_data);
	}
	if (attr.f_data) {
		free(attr.f_data);
	}
} /* pf_Arc */

/*-----------------------------------------------------------------------------
	TmPolyline function

	convert our backwards, 90 shifted angles to a more standard definition

-----------------------------------------------------------------------------*/
static int
convertAngle(ang)
int ang;
{
	ang = -ang + 900;
	while (ang < 0) {
		ang += 3600;
	}
	return ang;
}

