/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfoutf.c
   Applix ASC version: 19
   Date: Mon Nov 16 13:56:50 1998
*/
#include <stdio.h>
#include <errno.h>
#if AXMS
#include <basicwin.h>
#include <winspool.h>
#endif
#include <aos.h>
#include <afile.h>
#include <getstr10.h>
#include <paratm.h>
#include <pfint.h>

#include <pfoutf.p>

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

static	int	PFofdes = -1;
static	int	bufindex;
static	char *PfBuf = (char *)0;

#if AXMS
#ifndef MS32_WritePrinter
#define MS32_OpenPrinter                OpenPrinter
#define MS32_StartDocPrinter            StartDocPrinter
#define MS32_WritePrinter               WritePrinter
#define MS32_ClosePrinter               ClosePrinter
#endif
#define MS32_EndDocPrinter              EndDocPrinter
#endif

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
int
OutOpen (fname)
char *fname;
{
#if AXMS
    PFofdes = open (fname, (_O_BINARY | _O_WRONLY) );
#else
    /* open a file for writing; mode should be ignored. */
    PFofdes = open (fname, 1, 0222);
#endif
    if (PFofdes == -1)
		return FALSE;
	getBuf();
    bufindex = 0;
	return TRUE;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
int
OutCreate (fname)
char *fname;
{
#if AXMS    /*  {  */
    /* create the handle to the printer device */
    int ret;
    DOC_INFO_1 doc_info;

    ret = MS32_OpenPrinter( PF.PfPrinterName, &PF.PfPrinterHandle, NULL );
    if ( ret != TRUE )
       return( FALSE );
    doc_info.pDocName = PF.PfDoc_name;
    doc_info.pOutputFile = "";
    doc_info.pDatatype = NXLT("RAW");
    ret = MS32_StartDocPrinter( PF.PfPrinterHandle, 1, 
                                           (unsigned char *)(&doc_info) );
    if ( ret == 0 )
       {
       MS32_ClosePrinter( PF.PfPrinterHandle );
       return( FALSE );
       }

    PFofdes = -1;     /* fail safe, this should never be used */

#else    /*  }  AXMS  {  */

    PFofdes = creat (fname, 0666);
    /* create a file with read and write privileges for all */
    if (PFofdes == -1)
		return FALSE;
#endif   /*  } !AXMS  */

	getBuf();
    bufindex = 0;
	return TRUE;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void
OutJustClose()
{
#if AXMS
    int ret;

    if ( PF.PfPrinterHandle != (HANDLE)NULL )
       {
       ret = MS32_EndDocPrinter( PF.PfPrinterHandle );
       if ( !ret )
          tempAbort("invalid return from closing printer");
       ret = MS32_ClosePrinter(PF.PfPrinterHandle);
       if ( !ret )
          tempAbort("invalid return from closing printer");
       PF.PfPrinterHandle = (HANDLE)NULL;
       }
    else if ( PFofdes != -1 )
       {
       close( PFofdes );
       PFofdes = -1;
       }
#else
    close(PFofdes);
#endif
	if (PfBuf)
	{
		free(PfBuf);
		PfBuf = (char *) 0;
	}
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void
OutClose ()
{
    PfWriteIt ();
	OutJustClose();
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void
PfOutByte (b)
int b;
{
    if (bufindex>=BUFSIZ)
		PfWriteIt ();

	if(PfBuf)
    	PfBuf[bufindex++]=b;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void
PfOutString (s)
register char *s;
{
    while (*s)
	{
		if (bufindex>=BUFSIZ)
			PfWriteIt ();
		if(PfBuf)
			PfBuf[bufindex++] = *s++;
	}
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void
PfOutCString (s, n)
register char *s;
short n;
{
	register int i;

	for (i=0;i<n;i++)
	{
		if (bufindex>=BUFSIZ)
			PfWriteIt ();
		if(PfBuf)
			PfBuf[bufindex++] = *s++;
	}
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void
PfOutAlign ()
{
    if (bufindex&1) PfOutByte (0);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void
PfWriteIt ()
{
	int written = 0;
	char error_str[STRPATH];
    int ret;
    int last_error;

#if AXMS    /*  {  */
    if ( !PF.DevDirect )
       {
       /* write the buffer to the print spooler */
       ret = MS32_WritePrinter( PF.PfPrinterHandle, PfBuf, bufindex, &written );
       if ( (bufindex != written) || (ret != TRUE) )
          {
          if ( bufindex != written )
             last_error = -1;
          else
             last_error = GetLastError();
          sprintf(error_str,
             XLT("Error writing printer output, error_code = %d"), last_error );
   
          Abandon(error_str);
          }
       bufindex = 0;
       return;
       }
#endif    /*  }  AXMS  */


	while (TRUE)
	{
		if(PfBuf)
			written = write (PFofdes, PfBuf, bufindex);
		if (written == bufindex)
			break;
		if (errno == EINTR)
		{
			/* Try again. */
			errno = 0;
			continue;
		}
		if (!PF.PfAbortInProgress)
		{
			PF.Pferrno = errno; /* save errno for storage on exitval */

			sprintf(error_str,
					XLT("Error writing printer output, %s"),
					AxGetStr(ErrnoErr()));
			Abandon(error_str);
#if 0
			PFabort1(NXLT("Error writing printer output.  errno: %d"), errno);
#endif
		}
	}

    bufindex = 0;
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void
getBuf()
{
	char	*PFmalloc();

	if (PfBuf == (char *)0)
	{
		PfBuf = PFmalloc(BUFSIZ);
	}
}


