/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfpat.c
   Applix ASC version: 11
   Date: Mon Nov 16 13:56:50 1998
*/
#include <stdio.h>
#include <pf.h>
#include <aos.h>
#include <getstr5.h>
#include <atm.h>
#include <pfpat.h>
#include <thimble.h>

#include <pfpat.p>


pattern_type pattern[MAX_PATS*2] ;		/* 0 - MAX_PATS-1 are built-in */
									/* MAX_PATS - 2*MAX_PATS-1 are user defined */

/*************************************************************************
 *
 *	Finf user-def fill pattern
 *	
 *	Return index to pattern array
*************************************************************************/
int pfFindPattern(pat)
int pat ;
{
	int ix ;

	for (ix = MAX_PATS ; ix < 2*MAX_PATS ; ix++)
		if (pattern[ix].needed && pattern[ix].pat == pat)
			return ix ;
	return -1 ;
}

/*************************************************************************
 *
 *	Set fill pattern needed (either user-defined or built-in)
 *	
 *	Return index to pattern array
*************************************************************************/
int pfSetPattern(type, pat, hyt, wid, dep, data, cmap)
int type;
int hyt, wid;
void *data;
cmap_type *cmap ;
{
	int ix, dsize, csize;

	switch (type)
	{
	case BUILTIN_FILL_TYPE:
		ix = pat ;
		pattern[ix].needed = TRUE ;
		pattern[ix].pat = pat ;
		return ix ;

	case USERDEF_FILL_TYPE:
		dsize = hyt * 2 * ((wid*dep+15)/16);
		for (ix = MAX_PATS ; ix < 2*MAX_PATS ; ix++)
			if (pattern[ix].needed &&
				pattern[ix].hyt == hyt &&
				pattern[ix].wid == wid &&
				pattern[ix].dep == dep &&
				memcmp(data, pattern[ix].data, dsize) == 0)
				return ix ;
		for (ix = MAX_PATS ; ix < 2*MAX_PATS ; ix++)
			if (!pattern[ix].needed)
				break ;
		if (ix >= 2*MAX_PATS)
			return -1 ;
		pattern[ix].needed = TRUE ;
		pattern[ix].ref_index = -1 ;
		pattern[ix].pat = pat ;

		/* copy the data */
		dsize = hyt * 2 * ((wid*dep+15)/16);
		pattern[ix].hyt = hyt;
		pattern[ix].wid = wid;
		pattern[ix].dep = dep;
		pattern[ix].data = malloc(dsize);
		memcpy(pattern[ix].data, data, dsize);

		/* copy the color map */
		if (cmap)
		{
			pattern[ix].cmap.n_colors = cmap->n_colors ;
			csize = cmap->n_colors * sizeof(color_type) ;
			pattern[ix].cmap.color = (color_type *)malloc(csize);
			memcpy(pattern[ix].cmap.color, cmap->color, csize);
			}
		else
		{
			pattern[ix].cmap.n_colors = 0 ;
			pattern[ix].cmap.color = NULL ;
		}
		return ix ;

	default:
		return -1 ;
	}
}



/*************************************************************************
 *
 *	free memory used by pattern fills
 *	
*************************************************************************/
void
pfResetPattern()
{
	int ix;

	for (ix = 0 ; ix < MAX_PATS * 2 ; ix++)
	{
		pattern[ix].needed = FALSE ;
		if (pattern[ix].data)
			free(pattern[ix].data);
		if (pattern[ix].cmap.color)
			free(pattern[ix].cmap.color);
		pattern[ix].data = NULL ;
		pattern[ix].cmap.color = NULL ;
	}
}




/*************************************************************************
 *
 *	clear memory used by pattern fills
 *	
*************************************************************************/
void
pfClearPattern()
{
	int ix;

	memset(pattern, 0, MAX_PATS*2*sizeof(pattern_type)) ;

	for (ix = 0 ; ix < MAX_PATS * 2 ; ix++)
		pattern[ix].ref_index = -1 ;
}
