/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pfsup.c
   Applix ASC version: 30
   Date: Mon Nov 16 13:56:50 1998
*/
#include <stdio.h>
#include <setjmp.h>
#include <errno.h>
#include <math.h>
#include <pf.h>
#include <aos.h>
#include <paripc.h>
#include <scac.h>
#include <paratm.h>
#include <devsetup.h>
#include <getstr10.h>
#include <pfextra.h>
#include <pfint.h>

#if AXMS
#include <winspool.h>
#define hypot _hypot
#endif

#include <pfsup.p>

PF_Global PF; /* main pf structure for global data */

/* bit tables */

unsigned char PfBit_rmask[8] = { 128, 64, 32, 16, 8, 4, 2, 1 };
unsigned char PfRight_mask[8] = { 1, 3, 7, 15, 31, 63, 127, 255 };

extern int TargetRes;

extern jmp_buf PfAbortEnv;

#if THIM_STATIC
char *malloc(), *realloc();
#endif


/*-----------------------------------------------------------------------------
	Initialize Main PF Process

	init_pf(d) initializes the pf for device d
-----------------------------------------------------------------------------*/
void init_pf(d)
int d;
{
    short h_desktop,w_desktop,d_height,d_width;
    int err;
    struct TmFeatures features;
    /* initialize conversion factors globals (300dpi square) */

    err = TmDchar(&h_desktop, &w_desktop, &d_height, &d_width, &features, d);
    if (err == AOK)
    {
        PF.PfDotsPerInch = (1000 * features.grMWid) / features.grMHt;

        PF.PfGXPixFactor = features.grPWid;
        PF.PfGYPixFactor = features.grPHt;
        PF.PfTXPixFactor = features.txPWid;
        PF.PfTYPixFactor = features.txPHt;

        PF.PfGXMilFactor = features.grMWid;
        PF.PfTXMilFactor = features.grMHt;
        PF.PfGYMilFactor = features.txMWid;
        PF.PfTYMilFactor = features.txMHt;
    }
    else
    {
        PF.PfDotsPerInch = TargetRes;

        /* Change these to match! */
        PF.PfGXPixFactor = PF.PfTXPixFactor = 3;
        PF.PfGYPixFactor = PF.PfTYPixFactor = 3;

        PF.PfGXMilFactor = PF.PfTXMilFactor = 10;
        PF.PfGYMilFactor = PF.PfTYMilFactor = 10;

        d_width = 8500;
        d_height = 11000;
    }
    if (PF.PfPageMilWidth == 0)
        PF.PfPageMilWidth = d_width;
    if (PF.PfPageMilHeight == 0)
        PF.PfPageMilHeight = d_height;

    PF.PfPagePixWidth = xmiltogpix(PF.PfPageMilWidth);
    PF.PfPagePixHeight = ymiltogpix(PF.PfPageMilHeight);

    PfPage_init();  /* initialize base window management */
    PF.Pferrno = -1; /* for saving fatal errno val 'till exit time */
    PF.Pferror_message = (char *)NULL ;
}

/*-----------------------------------------------------------------------------
	centralized malloc
	guaranteed to return nonzero pointer (if it returns)
-----------------------------------------------------------------------------*/
char *PFmalloc(size)
unsigned size;
{
    char *p;

    if (size>0)
    {
        if ((p = malloc(size)) == CNULL)
            /* malloc failed */
            PFabort("malloc failure");
        return p;
    }
    else
        return NULL;

}  /* end PFmalloc */

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
char *PFcalloc(num,size)
unsigned size;
unsigned num;
{
    char *p;

    if (size>0)
    {
        if ((p = calloc(num,size)) == CNULL)
            /* calloc failed */
            PFabort("calloc failure");
        return p;
    }
    else
        return NULL;
 
}  /* end PFcalloc */



/*-----------------------------------------------------------------------------
	centralized free

	 only place free is called
-----------------------------------------------------------------------------*/
PFfree(pointer)
    char *pointer;

{
    if (pointer != NULL) { 
        free(pointer);
    }
}


/*-----------------------------------------------------------------------------
	PFabort -- abort on fatal error
-----------------------------------------------------------------------------*/
PFabort3(format, s1, s2, s3)
char *format, *s1, *s2, *s3;
{
    char buf[256];
    void pfFreeData();

    Debugf("errno = %d\n", errno);
    sprintf (buf, format, s1, s2, s3);
    PF.Pferror_message = malloc(strlen(buf) + 1) ;
    strcpy(PF.Pferror_message, buf) ;
    pf_unlink(FALSE);
    pfFreeData();
#if 0
    taskAbort(buf);
#else
    PF.PfAbortInProgress = TRUE;
#if AXMS
    if ( PF.PfPrinterHandle != (HANDLE)NULL )
        {
        MS_AbortPrinter( PF.PfPrinterHandle );
        MS32_ClosePrinter( PF.PfPrinterHandle );
        PF.PfPrinterHandle = (HANDLE)NULL;
        }
#endif
    longjmp(PfAbortEnv, 1);
#endif
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
PFabort2(format, s1, s2)
char *format, *s1, *s2;
{
    PFabort3(format, s1, s2, CNULL);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
PFabort1(format, s1)
char *format, *s1;
{
    PFabort3(format, s1, CNULL, CNULL);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
PFabort(format)
char *format;
{
    PFabort3(format, CNULL, CNULL, CNULL);
}


/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
void
pf_asserter (file,line)
char *file;
int line;
{
    char buf[300];

    sprintf(buf,XLT("assertion failed at line %d of file %s\n"),line,file);
    tempAbort(buf);
}

/*-----------------------------------------------------------------------------
/*
 * coordinate transformations
 *
 *  Force to long for better intermediate accuracy
 *  on short int machines.
 *
 *  There is only one external coordinate system: Applix inches, measured
 *  in mils (thousandths of an Applix inch).
 *
 *  There are three internal coordinate systems: graphic, text measuring,
 *  and text positioning.  On most printers these work out to be the same.
 *  But the HP Laserjet, for example, can position text to a 720th of an inch,
 *  while the text is measured in 300ths.  The graphic resolution ranges
 *  from 75 dpi to 300 dpi, with several intermediate resolutions.
 *
 *  function:                   from:        to:
 *
 *  xmiltopix(),ymiltopix()     mils         text measure
 *  xpixtomil(),ypixtomil()     text measure mils
 *
 *  xmiltogpix(),ymiltogpix()   mils         graphic pos
 *  xgpixtomil(),ygpixtomil()   graphic pos  mils
 *
 *  xmiltotpix(),ymiltotpix()   mils         text pos
 *  xtpixtomil(),ytpixtomil()   text pos     mils
 *
 *  The mil to pix conversions truncate, while the pix to mil conversions
 *  round up.  Hopefully, this all works out.
-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
xmiltopix(xmils)
int xmils;
{
    return (int) (((long) xmils*PF.PfGXPixFactor) / PF.PfGXMilFactor);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
ymiltopix(ymils)
int ymils;
{
    return (int) (((long) ymils*PF.PfGYPixFactor) / PF.PfGYMilFactor);
}

/* The text measure to mil conversions.  Note the rounding up */

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
xpixtomil(pix)
int pix;
{
    return (int) (((long) pix*PF.PfGXMilFactor + PF.PfGXPixFactor - 1) / PF.PfGXPixFactor);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
ypixtomil(pix)
int pix;
{
    return (int) (((long) pix*PF.PfGYMilFactor + PF.PfGYPixFactor - 1) / PF.PfGYPixFactor);
}

/* The mil to graphic position conversions */

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
xmiltogpix(xmils)
int xmils;
{
    return (int) (((long) xmils*PF.PfGXPixFactor) / PF.PfGXMilFactor);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
ymiltogpix(ymils)
int ymils;
{
    return (int) (((long) ymils*PF.PfGYPixFactor) / PF.PfGYMilFactor);
}

/* The graphic position to mil conversions.  Note the rounding up */

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
xgpixtomil(pix)
int pix;
{
    return (int) (((long) pix*PF.PfGXMilFactor + PF.PfGXPixFactor - 1) / PF.PfGXPixFactor);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
ygpixtomil(pix)
int pix;
{
    return (int) (((long) pix*PF.PfGYMilFactor + PF.PfGYPixFactor - 1) / PF.PfGYPixFactor);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
xtpixtogpix(tpix)
int tpix;
{
    return xmiltogpix(xpixtomil(tpix));
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
ytpixtogpix(tpix)
int tpix;
{
    return ymiltogpix(ypixtomil(tpix));
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
xgpixtotpix(gpix)
int gpix;
{
    return xmiltopix(xgpixtomil(gpix));
} 

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
ygpixtotpix(gpix)
int gpix;
{
    return ymiltopix(ygpixtomil(gpix));
}

/*-----------------------------------------------------------------------------
 * Postscript support.
 * Use the four points describing the on-paper extent of
 * the transformed text object (wp->imageBox,) plus the
 * rect describing the untransformed objects (uheight/uwidth)
 * to derive the transform info for the object (translation,
 * rotation, scale factor.)  Both boxes are in device pixels
 * at this point.
-----------------------------------------------------------------------------*/
int
pfGetTextXform(wp, uwidth, uheight, tx, ty, sx, sy, sinAng, cosAng,xshear)
/* inputs */
struct wind_struct *wp;
int uwidth, uheight;
/* outputs */
int *tx, *ty;
double *sx, *sy, *sinAng, *cosAng, *xshear;
{
    Point32 pt[4], tpt[4];
    double destwid, destht;
    double srcwid, srcht;
    int i, minx, miny, maxx, maxy, ttx, tty;
    int transformed; /* bool */
    int zeroSize;

    /* axatmrt functions */
    void findMinMax(),  scaleInvR();
    int scaleSXRT(), badImageBox();
    
    transformed = TRUE;
    if (badImageBox(wp->imageBox)) {
        /* For example, zero area.  Handle this by setting
         * the imageBox to the default dimensions for the
         * text object (as if untransformed.)
         */
        wp->imageBox[0].x = wp->imageBox[0].y = 0;
        wp->imageBox[1].x = uwidth;
        wp->imageBox[1].y = 0;
        wp->imageBox[2].x = uwidth;
        wp->imageBox[2].y = uheight;
        wp->imageBox[3].x = 0;
        wp->imageBox[3].y = uheight;
        transformed = FALSE;
    }

    pt[0] = wp->imageBox[0];
    pt[1] = wp->imageBox[1];
    pt[2] = wp->imageBox[2];
    pt[3] = wp->imageBox[3];

    findMinMax(pt, 4, &minx, &miny, &maxx, &maxy);

    /* translation term, by making 1st point relative to upper left */
    ttx = pt[0].x - minx;
    tty = pt[0].y - miny;

    /* no yshear by definition, so read object's width directly */
    destwid = hypot((double)(pt[0].x - pt[1].x),
                    (double)(pt[0].y - pt[1].y));

    /* remove translation */
    pt[0].x -= ttx;
    pt[0].y -= tty;
    pt[1].x -= ttx;
    pt[1].y -= tty;
    pt[2].x -= ttx;
    pt[2].y -= tty;
    pt[3].x -= ttx;
    pt[3].y -= tty;

    /* get angle info */
    if (destwid != 0) {
        *sinAng = (-1) * (pt[0].y - pt[1].y) / destwid;
        *cosAng = (pt[1].x - pt[0].x) / destwid;
    }
    else {
        *sinAng = 0.0;
        *cosAng = 1.0;
    }

    /* de-rotate pts 0, 3 and compare to get xshear and yscale */
    scaleInvR(pt[0].x, pt[0].y, &tpt[0].x, &tpt[0].y,
                *cosAng, *sinAng);
    scaleInvR(pt[3].x, pt[3].y, &tpt[3].x, &tpt[3].y,
                *cosAng, *sinAng);
    if((i = tpt[3].y - tpt[0].y) == 0)
        *xshear = 0.0;
    else
        *xshear = ((double)(tpt[3].x - tpt[0].x) /
                  (double)(i));
    destht = (double)(i);
    if (uwidth != 0)
        *sx = (destwid) / (double) uwidth;
    else
        *sx = 1.0;
    if (uheight != 0)
        *sy = (destht) / (double) uheight;
    else
        *sy = 1.0;

    /* translation terms of transformation matrix for translating
     * top left corner of text/image (after SXR) from origin of
     * bounding box coordinate system (origin at top left of
     * untransformed text/image, y increasing downwards) to its
     * final transformed position */
    *tx = ttx;
    *ty = tty;

    if (*sx==1.0 && *sy==1.0 && *cosAng==1.0 && *sinAng==0.0)
        transformed = FALSE;

    return(transformed);
}

/*-----------------------------------------------------------------------------
 * pfXformTextPoint - Given an untransformed point in the bounding box
 *              coordinate system (origin at top left of untransformed
 *              text, y increasing downwards), this routine returns
 *              the transformed position in the same coordinate system.
-----------------------------------------------------------------------------*/
pfXformTextPoint(pt, tx, ty, sx, sy, sinAng, cosAng, xShear, pt1)
TmPoint32 pt;            /* point to be transformed */
int tx, ty;              /* translation terms in transformation matrix */
double sx, sy;           /* scale terms in transformation matrix */
double sinAng, cosAng;   /* sin and cos of rotation angle */
double xShear;           /* x-shear term in transformation matrix */
TmPoint32 *pt1;          /* RETURN - transformed point */
{
    pt1->x = (pt.x * sx * cosAng)
                + (-pt.y * ((-xShear * sy * cosAng) + (sy * sinAng)))
                + tx;
    pt1->y = (pt.x * sx * sinAng)
                + (pt.y * ((xShear * sy * sinAng) + (sy * cosAng)))
                + ty;
}


/*-----------------------------------------------------------------------------
 * Write onto 'off' the offset between the upper-left corner of the
 * imageBox bounding box and the position of the text object as specified
 * by the printing application.  (This doesn't actually read the app-
 * specified position; rather, we assume that the app has positioned the
 * object so that the untransformed text bounding box is to be centered
 * on the imageBox bounding box.
-----------------------------------------------------------------------------*/
pfImageBoxOffset(off, uwidth, uheight, ascent, bbwidth, bbheight)
TmPoint32 *off;
int uwidth, uheight; /* untransformed object height, width */
int ascent; /* untransformed text object total ascent */
int bbwidth, bbheight; /* imageBox bounding box width, height */
{
    off->x = (bbwidth>>1) - (uwidth>>1);
    off->y = (bbheight>>1) + (ascent - (uheight>>1));
}


/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
int
pfEasyTextFill(patt)
attr_type *patt;
{
    /* NOTE that line patterns aren't currently supported, so
       we won't bother testing them in here.. */

    if (!patt)
        return(TRUE);

    if (patt->f_type == 5 && patt->f_pat == 1000)
        return(TRUE);           /* fill pat is solid foreground */

    if (patt->f_pat == -1)
        return(TRUE);           /* no fill; NOTE this is treated as solid */

    if (patt->f_type == 0 && patt->f_pat == 4)
        return(TRUE);           /* fill pat is solid foreground */

    if ((patt->ffg.c == patt->fbg.c) &&
        (patt->ffg.m == patt->fbg.m) &&
        (patt->ffg.y == patt->fbg.y) &&
        (patt->ffg.k == patt->fbg.k) &&
        (patt->ffg.see_thru == patt->fbg.see_thru ))
        return(TRUE);           /* foreground and background are the same! */

    return (FALSE);
}

/*-----------------------------------------------------------------------------
-----------------------------------------------------------------------------*/
int
pfGetPermilFill(pat)
     int pat;
{
    int ret_val = 0;
    switch(pat)
    {
      case 0:                   /* 0% */
          ret_val = 0;
          break;
      case 12:                  /* 12.5% */
          ret_val = 125;
          break;
      case 1:                   /* 25% */
          ret_val = 250;
          break;
      case 10:                  /* 37.5% */
          ret_val = 375;
          break;
      case 2:                   /* 50% */
          ret_val = 500;
          break;
      case 21:                  /* 62.5% */
          ret_val = 625;
          break;
      case 3:                   /* 75% */
          ret_val = 750;
          break;
      case 13:                  /* 87.5% */
          ret_val = 875;
          break;
      default:
      case 4:                   /* 100% */
          ret_val = 1000;
          break;
      }
    return(ret_val);
}


/*-----------------------------------------------------------------------------
	incoming  range  pattern
    percent
              -    0
         0         :       0
              -   62
        125        :      12
              -  187
        250        :       1
              -  312
        375        :      10
              -  437
        500        :       2
              -  562
        625        :      21
              -  687
        750        :       3
              -  812
        875        :      13
              -  937
       1000        :       4
              - 1000
-----------------------------------------------------------------------------*/
int
getIXfromPercent(percent)
int percent;
{
    if (percent < 62)
        return(0);
    else if (percent < 187)
        return(12);
    else if (percent < 312)
        return(1);
    else if (percent < 437)
        return(10);
    else if (percent < 562)
        return(2);
    else if (percent < 687)
        return(21);
    else if (percent < 812)
        return(3);
    else if (percent < 937)
        return(13);
    else
        return(4);
}

/*-----------------------------------------------------------------------------
	pfIsTransparent

	return true if fill pattern is transparent
-----------------------------------------------------------------------------*/
int pfIsTransparent(attr)
	attr_type *attr ;

	{
	if (!attr)
		return FALSE ;

    if (attr->ffg.see_thru && attr->fbg.see_thru)
        return TRUE ;

    if (attr->ffg.see_thru)
	{
		if (attr->f_type == GRAY_FILL_TYPE && attr->f_pat == 1000)
	        return TRUE ;
		if (attr->f_type == BUILTIN_FILL_TYPE && attr->f_pat == -1)
	        return TRUE ;
	}

    if (attr->fbg.see_thru)
	{
		if (attr->f_type == GRAY_FILL_TYPE && attr->f_pat == 0)
        return TRUE ;
	}

	return FALSE ;
	}

/*-----------------------------------------------------------------------------
 *
 * extract_terms - extract_factors utility function
 *
 * expects the imP->wid,hyt terms to be in pixels, everything else in mils.
-----------------------------------------------------------------------------*/
extract_terms_for_device(imP, xform, TrXp, TrYp,
              ScXp, ScYp, sinAngp, cosAngp, ShXp, widp, hytp, device)
ge_image_type *imP;             /* IN */
poly_type *xform;               /* IN */
int *TrXp;                      /* OUT */
int *TrYp;                      /* OUT */
double *ScXp;                   /* IN/OUT */
double *ScYp;                   /* IN/OUT */
double *sinAngp;                /* OUT */
double *cosAngp;                /* OUT */
double *ShXp;                   /* OUT */
int *widp;                      /* OUT */
int *hytp;                      /* OUT */
int device;                     /* IN */

{
    int swid_mils,shyt_mils, i, min_x,min_y, max_x,max_y;
    double width, delta_x, delta_y, orig_ScX, orig_ScY;
    TmPoint32 p1, p2, p3, p4;

    /* no trans means don't scale */
    if ( !xform )
    {
        *TrXp = 0.0;
        *TrYp = 0.0;
        *ScXp = 1.0;
        *ScYp = 1.0;
        *sinAngp = 0.0;
        *cosAngp = 1.0;
        *ShXp = 0.0;
        swid_mils = xpixtomil(imP->wid);
        shyt_mils = ypixtomil(imP->hyt);

        *widp = swid_mils;
        *hytp = shyt_mils;
    }
    else
    {
        orig_ScX = *ScXp;
        orig_ScY = *ScYp;
        swid_mils = xpixtomil(imP->wid);
        shyt_mils = ypixtomil(imP->hyt);

        /* set the first outline point on the origin */
        min_x = max_x = xform->points[0].x;
        min_y = max_y = xform->points[0].y;

        /* for points[1..4], find the bounding box */
        for ( i=1; i<5; i++ )
        {
            if ( min_x > xform->points[i].x )
                min_x = xform->points[i].x;
            else if ( max_x < xform->points[i].x )
                max_x = xform->points[i].x;
            if ( min_y > xform->points[i].y )
                min_y = xform->points[i].y;
            else if ( max_y < xform->points[i].y )
                max_y = xform->points[i].y;
        }
        min_x = min_x * orig_ScX;
        max_x = max_x * orig_ScX;
        min_y = min_y * orig_ScY;
        max_y = max_y * orig_ScY;
        *widp = max_x - min_x;
        *hytp = max_y - min_y;
        p4.x = (xform->points[3].x - xform->points[0].x) * orig_ScX;
        p4.y = (xform->points[3].y - xform->points[0].y) * orig_ScY;
        p3.x = (xform->points[2].x - xform->points[0].x) * orig_ScX;
        p3.y = (xform->points[2].y - xform->points[0].y) * orig_ScY;
        p2.x = (xform->points[1].x - xform->points[0].x) * orig_ScX;
        p2.y = (xform->points[1].y - xform->points[0].y) * orig_ScY;
        p1.x = (xform->points[0].x - xform->points[0].x) * orig_ScX;
        p1.y = (xform->points[0].y - xform->points[0].y) * orig_ScY;

        /* determine the sin/cos, scale factors, and shear factor */
        width = fabs(hypot((double)(p1.x - p2.x), (double)(p1.y - p2.y)));
        if ( width == 0.0 ) /* an empty field? */
        {
            *ScXp    = 1.0; *ScYp    = 1.0;
            *sinAngp = 0.0; *cosAngp = 1.0;
            *ShXp    = 0.0;
            *TrXp    =   0; *TrYp    =   0;
        }
        else
        {
            /* calc the angle info (note: y increases downward) */
            *sinAngp = (double)(p2.y - p1.y) / width;
            *cosAngp = (double)(p2.x - p1.x) / width;
 
            /* de-rotate last point and calc x shear relative to origin */
            delta_x = (double)p4.x * (*cosAngp) + (double)p4.y * (*sinAngp);
            delta_y = (double)p4.y * (*cosAngp) - (double)p4.x * (*sinAngp);
            if ( delta_y < 0.01 && delta_y > -0.01 )
                *ShXp = 0.0;
            else
                *ShXp = delta_x / delta_y;

            /* the offset of the origin in the image to be created */
            *TrXp = (-1.0) * (min_x - (xform->points[0].x * orig_ScX));
            *TrYp = (-1.0) * (min_y - (xform->points[0].y * orig_ScY));

            /* calc scaling */
            *ScXp = width   / (double)swid_mils;
            *ScYp = delta_y / (double)shyt_mils;
        }
    }
}

/*-----------------------------------------------------------------------------
 *
 * extract_factors - TmScaleImage utility function
-----------------------------------------------------------------------------*/
extract_factors_for_device(imP, xform, trans, device)
ge_image_type *imP;
poly_type *xform;
ge_image_transform_type *trans;
int device;
{
    double ShX, ScX, ScY, sinAng, cosAng, delta_x, delta_y;
    int src_wid_mils, src_hyt_mils, TrX_mils, TrY_mils;
    int wid, hyt, TrX, TrY;

    /* get translation, scaling, rotation and shearing from imP into xform */
    ScY = trans->y_scale;
    ScX = trans->x_scale;
    extract_terms_for_device(imP, xform,
                             &TrX_mils, &TrY_mils,
                             &ScX, &ScY,
                             &sinAng, &cosAng,
                             &ShX,
                             &src_wid_mils, &src_hyt_mils,
                             device);

    wid = xmiltopix(src_wid_mils);
    hyt = ymiltopix(src_hyt_mils);
    TrX = xmiltopix(TrX_mils);
    TrY = ymiltopix(TrY_mils);

    trans->wid = wid;
    trans->hyt = hyt;
    trans->x_scale = ScX;
    trans->y_scale = ScY;

    /* set up matrix that projects xform onto the source rect */
    if ( ScX == 0.0 )
    {
        trans->x_x = trans->x_y = 0.0;
        trans->x_t = 0;
    }
    else
    {
        trans->x_x = (cosAng + ShX*sinAng)/ScX;
        trans->x_y = (sinAng - ShX*cosAng)/ScX;
        trans->x_t = ShX*(TrY*cosAng - TrX*sinAng)/ScX -
            (TrX*cosAng + TrY*sinAng)/ScX;
    }
    if ( ScY == 0.0 )
    {
        trans->y_x = trans->y_y = 0.0;
        trans->y_t = 0;
    }
    else
    {
        trans->y_x = (sinAng / ScY);
        trans->y_y = (cosAng / ScY);
        trans->y_t = (TrX*sinAng - TrY*cosAng)/ScY;
    }
}
