/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: fstructs.h
   Applix ASC version: 63.1.4
   Date: Tue Jan  5 21:14:45 1999
*/
#ifndef SSFSTRUCTS_INCL	/* { */
#define SSFSTRUCTS_INCL	

#include <getargs.h>

typedef struct {
    char *name;
    double (*the_func)();   /* addr of the function */
    unsigned arg1 : 4,      /* arg types */
         arg2 : 4,
         arg3 : 4,
         arg4 : 4,
         type : 4,          /* the function type */
         optional_args:1;   /* true if can have optional args */
} bltin_value;

typedef struct  {
    char *fname;
    char *arg_str;
} func_table_type;


extern bltin_value Zbuiltins[];
extern func_table_type func_table[];

#define NUM_BUILTINS 164
#define ELF_BUILTIN_IDX NUM_BUILTINS-1
#define RPC_BUILTIN_IDX ELF_BUILTIN_IDX-1
#define LAST_BUILTIN_IDX RPC_BUILTIN_IDX-1

#define SS_FUNC_TYPE_NONE    0
#define SS_FUNC_TYPE_NUMERIC 1
#define SS_FUNC_TYPE_STRING  2
#define SS_FUNC_TYPE_ARRAY   3
#define SS_FUNC_TYPE_MACRO   4
#define SS_FUNC_TYPE_SPECIAL 5
#define SS_FUNC_TYPE_EXTERNAL   6

#define SS_FINANCIAL    1
#define SS_MATH         2
#define SS_LOGICAL      3
#define SS_LOOKUP       4
#define SS_STRING       5
#define SS_DB           6
#define SS_DATETIME     7
#define SS_SPECIAL      8
#define SS_REALTIME     9
#define SS_STATISTICAL  10  
#define SS_OBSOLETE     11
#define SS_NA           12

/* function subtypes */
#define SS_FUNC_DATE	1
#define SS_FUNC_TIME 	2
#define SS_FUNC_BOOL 	3

#define ARG_NUM 1
#define ARG_LIST 2
#define ARG_STR 3
#define ARG_ADR 4



/*-----------------------------------------------------------------*/
/* the following define the flag argument to the "form_mgr" routine*/
/* they should appear in this order and agree with the "check_first*/
/* argument passed to the "add_bkptr" routine                      */
/*-----------------------------------------------------------------*/
#define NUM_TOKTYPES 38


#define DELETE_FORM 0
#define ADD_FORM    1
#define ADD_BKPTRS  2

#define MAX_TOKENS 1000

#define END_PTR 0
#define STOPPER 1

#define STOPPER_VAL 0x00010001


/*====================================================================*/
/* NOTE:  all of the following structures MUST be an even multiple of */
/*         4 bytes to run correctly on the RISC machines              */
/*         8 bytes to run correctly on the HP-RISC, M88K              */
/*                    and Intergraph (CLIX) machines.                 */
/*                                                                    */
/*====================================================================*/
typedef struct {
    short type;
    short where;
#if (LINUX && __alpha__)
	int pad;
#endif
    double value;
} 
tok_any;

#define TK_ADD 2
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_add;

#define TK_SUB 3
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_sub;

#define TK_MULT 4
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_mult;

#define TK_DIV 5
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_div;

#define TK_OPEN 6
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_open;

#define TK_CLOSE 7
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_close;

#define TK_UNARY 8
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_unary;

#define TK_LITERAL 9
typedef struct {
    short type;
    short where;
    short num_chars;
/* why wasn't this pad already in here? -- ericding */
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
	short pad;
#endif
    double value;
} 
tok_literal;

#define TK_CELL 10
typedef struct {
    short type;
    short where;
    short colnum;
    short rownum;
    short platenum;
    boolean abs_col;
    boolean abs_row;
    boolean abs_plate;
    short quoted;
    short pad;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad2;
#endif
} 
tok_cell;

#define TK_SPACE 11
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_space;

#define TK_FUNC 12
typedef struct {
    short type;     /* type of token */
    short where;    /* lexographic position */  
    short fcode;    /* index into Zbuiltins array */
    short no_args;  /* for TRUE/FALSE/ERROR ... */
    short col_idx;  /* Collision Index for Function HashTable */
    short fdex;     /* index into SSFuncTable */ 
    int pad2;
}
tok_func;

#define TK_FLAG 13
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
}
tok_flag;

#define TK_COMMA 14
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
}
tok_comma;

#define TK_RANGE 15
typedef struct {
    short type;
    short where;
    short tl_colnum;
    short tl_rownum;
    short tl_platenum;
    boolean tl_col_abs;
    boolean tl_row_abs;
    boolean tl_plt_abs;
    short br_colnum;
    short br_rownum;
    short br_platenum;
    boolean br_col_abs;
    boolean br_row_abs;
    boolean br_plt_abs;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_range;

#define TK_EQUAL 16
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_equal;

#define TK_GREATER 17
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_greater;

#define TK_GTEQUAL 18
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_gtequal;

#define TK_LESS 19
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_less;

#define TK_LTEQUAL 20
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_ltequal;

#define TK_NOTEQUAL 21
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_notequal;

#define TK_AND 22
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_and;

#define TK_OR 23
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_or;

#define TK_NOT 24
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_not;

#define TK_POWER 25
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_power;

#define TK_COMMENT 26
typedef struct {
    short type;
    short where;
    ID strid;
} 
tok_cmt;

/* I have created "old_tok_name" sice it was found that this
*  structure was not padded out to a 4 byte boundry, which 
*  causes problems on the RISC machines.  It was fixed locally
*  for the SATURN but nowhere else. See also the usage of this
*  in getdspform.c  GetDspForm() - jon
*/
#define TK_NAME 27
typedef struct {
    short type;
    short where;
    short quoted;
#if SATURN || SGI || HP800 || M_88K || CLIX || ALPHA || (LINUX && __alpha__)
	short pad2; /* added later cuz int needs to be on 4-byte boundary;
				   I guess jon's change was when ID was a short? */
#endif
    ID id;
#if SATURN || SGI || HP800 || M_88K || CLIX || ALPHA || (LINUX && __alpha__)
    int pad;
#endif
} 
old_tok_name;

typedef struct {
    short type;
    short where;
    short quoted;
#if SATURN || SGI || HP800 || M_88K || CLIX || ALPHA || (LINUX && __alpha__)
	short pad2; /* added later cuz int needs to be on 4-byte boundary;
				   I guess jon's change was when ID was a short? */
#endif
    ID id;
#if SATURN || SGI || HP800 || M_88K || CLIX || ALPHA || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_name;

#define TK_GORCH 28
typedef struct {
    short type;
    short where;
    ID strid;
#if 0
/* ID is now an int, so padding with a short causes 
   alignment problems; padding no longer needed, so
   removed - Colin
 */
    short pad;
#endif
} tok_gorch;

#define TK_XOR 29
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_xor;

#define TK_ECELL 30
typedef struct {
    short type;
    short where;
    short colnum;
    short rownum;
    short platenum;
    boolean abs_col;
    boolean abs_row;
    boolean abs_plate;
    short quoted;
    short pad;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad1;
#endif
} 
tok_ecell;

#define TK_ERANGE 31
typedef struct {
    short type;
    short where;
    short tl_colnum;
    short tl_rownum;
    short tl_platenum;
    boolean tl_col_abs;
    boolean tl_row_abs;
    boolean tl_plt_abs;
    short br_colnum;
    short br_rownum;
    short br_platenum;
    boolean br_col_abs;
    boolean br_row_abs;
    boolean br_plt_abs;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
} 
tok_erange;

#define TK_NULL 32
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
}
tok_null;

#define TK_STRING 33
typedef struct {
    short type;
    short where;
    ID strid;
}
tok_string;

/* The strcat operator '&' */
#define TK_STRCAT 34
typedef struct {
    short type;
    short where;
#if SGI || HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    int pad;
#endif
}
tok_strcat;

#define TK_REFERR 35
/* uses the same struct as TK_ECELL or TK_ERANGE */

#define TK_FLAG_IF 36
/* uses the same struct as TK_FLAG */
#define TK_FLAG_QUOTED 37
/* uses the same struct as TK_FLAG */






typedef short tok_short;

typedef union {
    tok_literal *l;
    tok_cell *c;

    tok_open *op;
    tok_close *cl;

    tok_unary *u;
    tok_any *a;

    tok_add *add;
    tok_sub *sub;

    tok_mult *mult;
    tok_div *div;

    tok_space *space;
    tok_func *func;

    tok_flag *flag;
    tok_comma *comma;

    tok_range *range;
    tok_cmt *cmt;

    tok_name *name;
    tok_gorch *gorch;

    tok_short *as_short;
    tok_ecell *ecell;

    tok_erange *erange;
#if SGI ||  HP800 || M_88K || CLIX || ALPHA || NT_PPC || (LINUX && __alpha__)
    tok_erange *pad;
#endif
    tok_null   *tknull;
    tok_string *tkstr; /* literal strings */
    tok_strcat  *tkstrcat; /* concat oper */
} 
tok_ptr;


#define FUNCTABLESIZE 2999

typedef struct ss_func_type {
    char *name; /* actual translated version of the function name */
    char *engName;  /* english name of the function for writing to files */
    struct ss_func_type *next;
    int id;/* index into Zbuiltins[];  NUM_BUILTINS-1 for Remote Function */
    int type;
    int collisionIdx;  /* index of func name collision in func hash table */
	int	callMode;	/* calling mode for external functions */
	short calcOnDemand;/* used for calculating external functions individually
						primary usage when someone does Edit->Recalc */
	short calcAlways;/* used for calculating external functions individually
							primary usage when anything changes in the SS */
} ss_func_info_type;


typedef struct {
int fdex;
int col_idx;
}ss_elf_macro_func_type;


extern short GetFuncSlotIndex();
extern struct FuncInfo *GetFuncInfo();
extern struct FuncInfo *GetFuncInfoFromIndex();

#define AX_SSBLTIN_NUM_FUNC 0x01000000
#define AX_SSBLTIN_STR_FUNC 0x02000000
#define AX_SSBLTIN_MASK     (AX_SSBLTIN_NUM_FUNC | AX_SSBLTIN_STR_FUNC)


#define AX_SLOT_MASK        0x8000
#define AX_REMOTE_FCODE     0x8000
#define AX_MACRO_FCODE      0x4000

#define NUM_SPL_FUNCS 6 /* number of arg-less functions */
#define NUM_RSVD_NAMES 7    /* number of keywords not to be used as names */
#endif	/* } SSFSTRUCTS_INCL */
