/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: hdrftr.h
   Applix ASC version: 17
   Date: Wed Nov 18 16:29:14 1998
*/
#ifndef __HDRFTR_H
#define __HDRFTR_H

#define	MAX_PAGE_TYPES	5	/* first page, odd page, even page, (last page) */
#define MAX_LINES		4	/* max #of header/footer lines per page */
#define MAX_LINE_PIECES	3	/* left, center, right */


typedef	struct	{
char *text;		/* the text for this header footer itself */
int facedex;		/* the typeface */
int colordex;	/* the name of color */
int fsize;		/* the font size in points */
unsigned int 	bold		:1;	/* TRUE if bold , else FALSE */
unsigned int 	italic		:1;	/* TRUE if italic on, else FALSE */
unsigned int strike_thru	:1;	/* TRUE if strike_thru on, else FALSE */
unsigned int 	underline	:2;	/* 0-none, 1-single, 2-double */
unsigned uline_words_only	:1;	/* TRUE if underlining words only */
}hfline_str_type;

typedef struct 	{
hfline_str_type	inner;
hfline_str_type center;
hfline_str_type outer;
}hfline_type;


typedef struct	{
int lMargin;	/* left margin in mills */
int rMargin;	/* right margin in mills */
int	tMargin;	/* top margin in mills */
int bMargin;	/* bottom margin in mills */
int pageWidth;	/* width of page in mills */
int pageHeight;	/* height of page in mills */
int printDevice;	
uint16	numPagesWide;	/* number of pages wide for print to fit */
uint16  numPagesTall;	/* number of pages tall for print to fit */
uint16	scaleFactor;	/* scaling to be used when printing */
uint8  	paperType;		/* the paper type, eg "US Letter" */
unsigned int printToFit:1,	/* TRUE if print to option is selected */
		landscape: 1,	/* TRUE if printing in landscape mode */
		chalign:	1,	/* TRUE if material to be centered horizontally */
		cvalign:	1,	/* TRUE if material to be centered vertically */
		dsphdrs:	1,	/* TRUE if row/col numbers to be printed */
		mControl:	1,	/* TRUE if swap margins for double sided printing*/
		scaledPrint:1, /* TRUE if printing with a scale factor */
		noPrintBeyondLastCell:1, /* TRUE if not to print beyond */
								/* last cell containing data even if*/
											/* they have borders/shading */
		printGridLines:1,		/* TRUE to print grid lines */
		printFormulas:1;		/* TRUE to print formulas */
}ss_print_setup_type;

typedef struct	{
int hlineCount;	/* # of header lines in this page */
int flineCount;	/* # of footer lines in this page */
int hMargin;	/* distance to headers from top of page in mills */
int fMargin;	/* distance to footers from bot of page in mills */
hfline_type	*hlines;		/* the headers lines in this page */
hfline_type	*flines;		/* the footer lines in this page */
}hdrftr_page_type;	




typedef struct{
unsigned int use_even_odd	: 1; /* TRUE if there is even page special */
unsigned int use_first   	: 1; /* TRUE if there is a first page special */
unsigned int use_last    	: 1; /* TRUE if there is a last page special */
hdrftr_page_type base;
hdrftr_page_type even;
hdrftr_page_type odd;
hdrftr_page_type first;
hdrftr_page_type last;
}hdrftr_type;

/* tokens written to the file (file format ) */
#define HLEFT_TEXT	"HleftText"
#define HCENT_TEXT	"HcentText"
#define HRIGHT_TEXT	"HrightText"
#define FLEFT_TEXT	"FleftText"
#define FCENT_TEXT	"FcentText"
#define FRIGHT_TEXT	"FrightText"


#define LMARGIN		"lMargin"
#define RMARGIN		"rMargin"
#define TMARGIN		"tMargin"
#define BMARGIN		"bMargin"
#define FDISTANCE	"fDistance"
#define HDISTANCE	"hDistance"

#define PAGE_TYPE	"pageType"

#define HLINE_COUNT	"HlineCount"
#define FLINE_COUNT	"FlineCount"

#define REV_BARS	"revBars"
#define MCONTROL	"mControl"
#define GLOBAL_ATTR	"dAtt"
#define PFA_REV	"printFormatRev"


/* these defines are based on rev2 format */
#define	FIRST_PAGE	0
#define REG_PAGE	2
#define ODD_PAGE	4
#define EVEN_PAGE	6
#define LAST_PAGE	7


#define HLEFT 	0
#define HCENTER	1
#define HRIGHT	2

#define FLEFT	3
#define FCENTER	4
#define FRIGHT	5


#define MAX_TEXT_PIECE	800
#define MAX_SS_TEXT		300


#define	LINESPAN	3	/* 0,1,2 	inner, center,outer */
#define NUMLINES	4	/* 4 lines for hdr/ftr */
#define PAGESPAN	12	/* LINESPAN * NUMLINES */
#define HFSPAN		36	/* 	PAGESPAN*NUMLINES*LINESPAN */


#define SS_MAX_TMARGIN 	4000	/* 4000 mils, 4 inches */
#define SS_MAX_BMARGIN  4000	/* 4000 mils, 4 inches */	

typedef struct {
/* the strings in each hdr/ftr line */
	wchar          *wp1;	/* left piece */
	wchar          *wp2;	/* center piece */
	wchar          *wp3;	/* right piece */
/* the respective attrs */
	hfline_str_type *lP1, *lP2, *lP3;
/* the respective descents */
	int             d1, d2, d3;
}               spl_hfattrs;

#endif		/*  __HDRFTR_H */
