/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ss.h
   Applix ASC version: 308.5
   Date: Wed Nov 18 16:29:06 1998
*/
#ifndef SS_H_INCL
#define SS_H_INCL

#include <setjmp.h>

#ifdef boolean
#undef boolean
#endif

#include <basicwin.h>
#include <xmlib.h>

#if RPC
#define MAX_RPC_SERVERS 10 /* should agree with calc.h */
#endif

#define OFFSET_MASK ((TOUCHED_MODULO/8 - 1) << 3)
#define SP_AR_ZEROS 50
#define MAX_ARRAY_FORMULAS	64001	/* the array index in cell is ushort */
#define MAX_SS_WINDOWS 4	/* maxium #of multiple window instances */

#define MANUAL_CALC 0
#define AUTO_CALC	1
#define INTERVAL_CALC	2
#define MAX_INTERVAL	3600

#include <elf.h>
#include <eldata.h>
#include <thimble.h>
#include <ssdefs.h>
#include <ssedit.h>
#include <hdrftr.h>

#define SSJV_UID_MASK   0xffff
#define SSJV_UID(id)    (id & SSJV_UID_MASK)
#define SSJV_ATM_ID(id) (id >> 16)
#define COL_LONGS ((NUM_COLS+7)/8)	/* 22 32bitters */
#define SSJV_MAX_OBJ_CACHE  10

typedef struct ssjvRowList {
	int NumRows;
	struct ssjvColList **info;
} ssjvRowList;

typedef struct ssjvColList {
	int low_col;
	int high_col;	/* highest col used */
	unsigned char tags[COL_LONGS];
} ssjvColList;

typedef struct {
	int flush_mode;	
	int attrPoolSize;
	ssjvRowList *linkage[LAST_USER_PLATE+1];
	int high_sheet;
	boolean backgroundFlush; /* TRUE, if background flush requested
								from client */
	int undoType;
	int nranges;
	range_desc *dispRange;
	int initing; /* TRUE while loading/reloading file */
	int flushAllObjects;
	GPHOBJ* objCache[SSJV_MAX_OBJ_CACHE];
	int objImageChanged[SSJV_MAX_OBJ_CACHE];
	jchannel_t* javaChannel;
} SSJvState;


typedef struct {
  genericAppStruct appInfo;
  SSJvState  *javaInfo;
  short ATM_id;				/* ATM window id */
  Window xwin;
  ID view;				/* current view being displayed */
  coordinate top_left;	/* top left displayed cell */
  SSRECT Doc;			/* Doc coords in pixels */
  SSRECT Win;			/* This Window's coords in pixels */
  uint32 titleWid;		/* width of title area in pixels */
  uint32 titleHyt;		/* height of title area in pixels */
  int32 num_rows;		/* how many rows are being displayed */
  int32 num_cols;		/* how many columns are being displayed */
  int32 page_num;		/* the pagenum where the cell cursor is */
  int32 page_ext;		/* the pageext where the cell cursor is */
  int32 RgnOpenCell;	/* The region the grid cursor is at */
  short y_valform;		/* where the valu/formula line is */
  short y_coldesc;		/* where the column descriptor is */
  short y_grid;			/* where the y-grid begins  in pixels */
  short x_grid;			/* where the x-grid begins  in pixels */
  int 	x_end;			/* where the SS window ends in pixels */
  int	y_end;			/* where the SS window ends in pixels */
  short unitGwidth;		/* how wide is grid in character units */
  short unitGheight;	/* how high the grid in character units */
  range_desc PointOrg;	/* where cursor was when point mode entered */
  short OpenEdit;		/* index to start of editable area */
  short ValForSize;		/* strlen of OpenStr */
  unsigned int Changed:1;		/* TRUE if cell contents have been edited */
  struct TmCrePacket cp;/* ATM window create packet */
  int Gpixel_height;    /* height of grid area in pixels */
  int pixel_height;     /* height of grid area in pixels */
  int Gpixel_width;     /* width of grid area in pixels */
  coordinate startgridcell;	/* used during editmode if screen scrolls */
  char OpenStr[MAX_CELLSTR+20];   /* text in data entry field */
  coordinate MarkOrg;	/* first pos of mark sequence */
  coordinate open_cell;	/* where the open cell is */
  coordinate bkup_opencell; /** to back up open cell when changing plates in 
							the edit mode ***/
  ID mark_array;        /* Independent Marks are stored for each SS Window */ 
  ID old_marks;
  TmPoint32 OCpoints[5];
  ID *cells;			/* what cells are displayed on the screen */
  char *marked;         /* what regions currently highlighted */
  char *row_d_ok;       /* whether the row descriptors are ok */
  short *cols;          /* the cols being displayed */
  short *rows;          /* The rows currently being displayed */
  int plate;			/* plate being displayed by this window */
  unsigned char *widths;/* how wide they are */
  unsigned short *heights; /* heights of rows */
  uint16  	*borders;  /* the borders for cells being displayed */
  long	grid_top;       /* how many pixels the screen scrolled vertically*/
  long grid_left;       /* how many pixels the screen scrolled horizontally*/
  int maxCols;          /* max number of cols based on window size */
  int maxRegions;       /* max number of regions based on window size */
  int maxRows;          /* max number of rows based on window size */
  int IsWindow;         /* Window has been created */
  SS_EDITLINE	edit_struct;
  SS_EDITLINE *el; /*  = &edit_struct; */
  SS_EDITLINE	incell_edit; /** stores information for incell_editing **/
  descr_win	rwin;	/* row window */
  descr_win cwin; 	/* col window */
  unsigned char *rowPageBrks; /* TRUE if row is a pagebreak, FALSE if not */
  unsigned char *colPageBrks; /* TRUE if col is a pagebreak, FALSE if not */
  mark_rect	mR;	/* the rectangle around marks */
  int x_old; /* = -1 */
  int y_old; /* = -1 */
  char lastName[MAX_NAMELEN+1];
  SSRECT ExposeRt;
  Region winRgn;	/* a X-Region describing an entire window */
  Region workRgn;	/* window - Graphic Objects area */
  Region tmpRgn;
  coordinate viewTopLeft;	/* top left used for graphic object positioning */
  short old_col;	/* for optimization in calcing new thumbing bars */
  int32 old_row;	
  int firstPlateDex;
  int lastPlateDex;
  uint16 numRowsTitled;	/* the number of row titles in this window */
  uint16 numColsTitled;	/* the number of col titles in this window */
  AXWidget platesWidget;	/* plate select widget */
  AXWidget ssWidget;		/* SS window widget  */
  char *platesWidgetStuff;	/* maintained entirely by ssplates.c */
  SSRECT dragDropRect;
  dragDropSelectionInfo dragDropInfo;
  grobj_list objList;		/* objects visible on screen */
  char *rowsOnScreen;	/* bit array to keep track of rows on screen */
  char *colsOnScreen;	/* bit array to track cols on screen */
  uint16	dspIntRow;	/* the row where display was interrupted */
  XRectangle	clipRectForZoom[1];/*clip rect filled for zooming purpose only*/
  unsigned int clipRgnInUse:1,	/* the region enclosing workarea 
												minus graphic objects*/
	displayInterrupted:1,	/* TRUE if display was interrupted */
  	new_open:1,		/* TRUE if should repaint the open cell */
  	zoomClipInUse:1, /** set when clipRectForZoom is modified ***/
  	fullRedisplay:1,	/* true a complete window refresh */
  	noGridAddr:1,	/* TRUE if not to display row/col addresses */
  	noPlates:1,				/* TRUE if plates are visible */
  	set_aside : 1,
  	blockCursor:1,	/* TRUE if cursor is in highlighted mode */
	checkBordersShades:1,	/* TRUE if borders/shades be checked to be 
									checked for repair */
	loadObjPoses:1,			/* TRUE to recompute object doc poses */
	screen_changed:1,	 /* TRUE if Vfywindow() has to be called in RedispAll */
	reformatObjects:1,	/* TRUE if objects need to be formatted */
	callDspMarked:1,	/*TRUE if DspMarked() need to be called from RedispAll*/
	isInCellEditOn:1,	/* if incell editing is on or regular editline **/
	highlightCurRowDescr:1,/* 1 if needs to highlight else FALSE for row/col*/
	highlightCurColDescr:1;/* 1 if needs to highlight else FALSE for row/col*/
  } window_value;


typedef struct {
  char 	*dblClickCallbackP,	/* this document's local double click callback*/
  		*loadCellCallbackP,	/* this document's local load click callback */
		*crCallbackP,		/* this document's local cr callback */
  		*selectCallbackP,	/* this document's selection callback */
		*uiLoadCellCallbackP; /* similar to loadcell except only called when
								user types in any new data */
  double revision;
  short calc_depth;		/* Size of calc_array */
  ID plates[LAST_TOTAL_PLATE+1];
  ID view_array;
  ID tview_array;		/** temp views are stored in this array **/
  ID name_array;
  ID clipboard;
  ID calc_array;		/* ID of CalcCell "Stack" array */
  ID calcStack;			/* Calc's execution machine stack of doubles */
  ID calcStyleStack;	/* Parallel list of styles info for cell */
  ID graph_array;		/* to array of graph objects */
  ID DoingMenu;			/* current menu script object ID */
  ID sparseArray;       /* array of non-Dense array ids */
  ID PrdctArray;		/* List of cells that use the PREDICT function */
  ID RandArray;			/* List of cells that use the RAND function */
  ID TodayArray;		/* List of cells that use the TODAY function */
  ID NowArray;			/* List of cells that use the NOW 	function */
  ID IndirectArray;		/* List of cells that use the Indirect Array */
  ID OffsetArray;		/* List of cells that use the OFFSET function */
  ID RowFuncArray;		/* List of cells that use the Row function */
  ID ColFuncArray;		/* List of cells that use the Col function */
  ID SheetFuncArray;	/* List of cells that use the Col function */
  ID RefFuncArray;		/* List of cells that use the Col function */
  ID ExtFuncsArray;		/* List of cells that use the Col function */
  ID calcOnDemandExtFuncsArray;	/* List of external funcs cells which uses */
								/*	different calculation mode */
  ID calcAlwaysExtFuncsArray;	/* List of external funcs cells which uses */
								/*	different calculation mode */
  ID GotoStack;			/* stack for goto Undos */
  ID strHash;			/* hash array for shared string objects */
  ID rtsHash;			/* hash array for shared realtime links */
  ID formHash;			/* hash array for shared formula objects */
  ID GoalArray;			/* Array of goal cells (#8398 for move commands) */
  ID Symbols;			/* symbols, e.g., decimal point, separator char*/
  ID arrayFormTable;	/* used by regression functions to maintain */
						/* info about the output range etc */
  ID webLinkAddressTable;/* stores the information regarding the html link */
  						/* address. Stores the string ids */
  ID db_array;			/* database range */
  long guid;			/* unique id for 'next' graph refdoc */
  long nuid;			/* unique id for 'next' ISR refdoc */
  int zoomFactor;		/* the zoom factor */
  short calcdex;		/* execution stack index */
  short scan_depth;		/* Size of scan_array */
  short calcMode;		/* Normal, Row, Col */
  int32 calcInterval;	/* interval in number of seconds */
  short calcMethod;		/* auto, manual or interval */
  short cnt_iterate;	/* 0 for display "Circular" */
  short goal_iterate;	/* Iterate till Percentage change */
  short plateBound;		/* Min bounding plate */
  int32 rowBound;		/* Min bounding row */
  int16 colBound;		/* Min bounding col */
  int32 dataRowBound;	/* Min bounding row containing data */
  int16 dataColBound;	/* Min bounding col containing data */
  int32 grObjRowBound;	/* max row of spanned by graphic objects */
  int16 grObjColBound;	/* max row of spanned by graphic objects */
  short invalid;		/* A cell was "put" since last display */
  short numAlign;		/* default number style */
  short textAlign;		/* default text style */
  short GotoStkPtr;		/* index into the goto stack, -1 means empty */
  short ClpStatus;		/* status of clipboard (empty, copy, cut) */
  short ReadOnly;
  short PrintFont;      /* Print font size */
  short UndoType;		/* type of undo noted or 0 for nothing there */
  short FontControl;    /* indicator of current font in use      */
  short GridsOn;        /* whether grid line display is on or off */
  int fgColor, bgColor;
  range_desc ClpRange;	/* range noted for copy/cut */
  uint16 numExtLinks;	/* # of external links in this ss */
  hdrftr_type	hf;
  ss_print_setup_type ps;
  cmap_type	cmap;	/* array of color definitions */
  fmap_type fmap;	/* table of font faces */
  csmap_type csMap;	/* custom style map */
  GPHOBJ *noRegenObj;	 
  RCFRMT *Formats; 	/* Linked List of Row/Col Formats in the SS */
  comment_table_type cmtTable;
  hook_map_type hookMap;
  ID RTSCONTRIBArray;		/* List of cells that use the RTSCONTRIB function */
  plate_list_type plateList;
  int numWin;
  unsigned int zoom_sign:1, 	/* stores the zoom sign currently displayed **/
  	regraph:1,		/* force the updating of all graphs now */
  	autoGraph:1,		/* update graphs at each redisplay */
	SyncWin:1,			/* TRUE if Synchronus Scrolling multiple windows */
	showPageBrks:1,	/* TRUE if page breaks are to be displayed */
	minimalReCalc:1, /* TRUE if calc only obsolete cells on Re-Calc */
	backCalcEnabled:1,	/* TRUE if background calc is enabled */
	calcOnDisplay:1,		/* TRUE if display driven calc is enabled */
	rtInsertOnDisplay:1,	/* TRUE if to RtInsert after display pass */
	mvCalcNow:1,		/* TRUE if calc after every tik model */
	calcTypeConversion:1,	/* TRUE if convert text to numbers and
										vice versa during calc */
	noAutoRtStart:1,	/* TRUE if not to active rt stuff automatically*/ 
	noAutoLinkLoad:1,	/* TRUE if not to load links automatically */
	rtShowOldValues:1,	/* TRUE to show old values on rt register */
	calcOnOpen:1,		/* FALSE if no calc on file open */
  	calcOnSave:1,		/* TRUE to calc on save in manual mode */
  	calcVisibleCellsOnly:1,		/* TRUE to calc visible cells only */
	viewFormulas:1,		/* TRUE to view formulas */
	optimalCalc:1,		/* TRUE if not to calculate the dependent 
							formuals */
	calcBeforeSave:1,	/* TRUE if should recalc before saving default
							is true*/
						/* if cell value hasn't changed */
	noODRegisterOnOpen:1, /* TRUE if u do not want to calc OD RTS formula */
	noRTIRegister:1; 	/* TRUE if u do not want to calc OD RTS formula */
 } ss_boom_value;


/* 
 * START OF THE GLOBAL SS STRUCT
 */
#include <ss_style.h>
#include <fmtnum.h>
#include <fstructs.h>
#include <tk_array.h>
#include <xfer.h>
#include <fsrts.h>
#include <scac.h>
#include <paripc.h>
#include <aos_xfer.h>
#include <axtime.h>

#include <zmglobal.h>
#include <ominited.h>

#define GOTO_STRLEN 32
#define MARKWIDTH 6         /* width of mark column */
#define MAX_BKS_LEVEL 10    /* width of mark column */


typedef struct 
  {
  int uid;
  int pid;                     /* task id                               */

  boolean javaSelNotifyOn;
  int javaHammerSuspended;	/* !!0 TEMP 120578 */
  otParams OM;	/* current object space parameters */
  OmMemInit MemStruct;
  ss_boom_value *Obj;
  ss_calc_list calcList;
  ss_calc_list screenCalcList;
  pr_page_info 	page_info ;	/* stores static info used by the printing ***/
  pr_preview_info preview_info ; /** stores all print preview info ***/
  ss_print_var_info pfv;
  ss_cmd_status cmd;	/* Used by SSCmdStatus() */
  struct highlight_block h_block1;/*  = {-1, -1, -1, -1 }; */
  struct highlight_block h_block2;/* = {-1, -1, -1, -1 }; */
  ss_click_type butt_click; /** keeps track of the button press click typea **/
  attr_type	line_attr;	 /* used by the TmPolyline for lines and filling */
  ss_file_buf_type wBuf;	 /* struct to manage writing to a buffer */
  ss_file_buf_type rBuf;	 /* struct to manage reading from a buffer */	
  bltin_num_format numFmt;

  double lastfillincr;
  double lastfillstart;
  double lastfillstop;
  double Zero;
  double DblZero;
  double DblOne;

  int psFactorX;
  int psFactorY ; 
  char tempdbview[STRNORMAL+1];

  long id;                     /* widget id of drawing surface          */
  boolean invalid;				/* TRUE if this is invalid */	

  /* uimain.c */
  int autoSavePeriod;		/* # of seconds before doing an auto save */
  int IpcControl;
  int linkPollPeriod;		/* Check links every # seconds */
  int IpcCounter;			/* Timer counter */
  int IpcCounterResetVal;	/* Timer counter resets at this interval */
  int prevZoomFactor;
  todo_list_item_t *todoList;   /* Head of timer based action list */
  
/* printrange.c  */
/* putcell.c */
  char *cellPool;
  char *calcCellPool;

  char **attrStrs;	/* the attr table for source document during copy/paste */
  int  numAttrStrs;	/* the of attr strs */

  int BkptrAdding;
  int PutCerrno;
  long CellsAdjusted;

/* readax.c */
  int loadingAscii;/* = FALSE; TRUE while loading ascii dumpfile, see parse.c */

/* readvc.c */


/* regions */
  struct highlight_block *back_h_ptr; /* = &h_block2; */
  struct highlight_block *cur_h_ptr;  /*  = &h_block1 */

/* NEEDS INVESTIGATION LEFT THEM IN THERE */
/* ss_array */
  int first_call;/* = 0; */
  char FFArray[SP_AR_ZEROS];       /* an array of FF's */

/* ss_elftalk.c */
  int StoreLastCell;
  int currentwidth;
  int currentheight;
  int gosdirection;
  int gospldirection;
  int lastdatestyle;
  int lastnumprec;
  int lastnumstyle;
  int lasttime; /*  = -1,  */
  int SaveErrorForElf;
  int CmdState; /* =AOK; */
  char *CmdErrStr; /* ="" */
  char *errorObj;
  int UserCallback; /*bitmap if user has installed an Elf callback function */

  char *GotoStr;
  char *gosearch;
  char *gospecial;
  char *gorange;
  char *gocell;
  char *namechange;
  char *currtitles;

  int LastPos, NewPos, StartPos, MaxPos, MinPos;
  range_desc editRng;
  int32 EditRelRowNum, EditRelColNum;

  XFontStruct *Grafont;
  XFontStruct *Colfont;
  Cursor DefWinCursor;

  int OMflush;
  int OMlogging;
  int backOne ; 		/*= TRUE;    only process this document while */
  int doingBackISR; 
  last_cell_value LastCell;
  short *UndoBuf;		/* buffer used by undo code "undo.c" to keep from  */
  int NewIconId;

  int QuietMode;		/* if TRUE, don't complain about conversion errors */

  int	cvWP ;      			/* convert SS to WP Beads */
  char *bkup_viewname;

/* ssmain.c */
  ID objId;         	/* Object number of spreadsheet main object */
  ID defSrcViewID;		/* view for default source  insetting */
  calc_error calcErr;   /* Recalc error information for HELP on status */
  char func_sepChar;    /* seperates function arguments, "," or ";" */
  foreignChars *symbols_ptr;
  int BksRecordLevel;	/* Which array we are considering at this moment */
  int BksSupLevel;		/* Which array we are working with at this moment */
  int BscanControl;      /* whether suppressing scanning of some backpointers */
  int colPartial;        /* Is there a partially displayed column? */
  int rowPartial;        /* Is there a partially displayed row? */
  int initCode;
  int numGraphs;         /* how many graphs are currently active */

  jmp_buf WhatIfErr; /* if a PREDICT/GOAL function fails */
  jmp_buf floatErr;



  int defaulterRngCnt;  /* = 0;  Used to get specific cell/rng when multiple  */

/* ssundo.c */
  FILE *UndoFp;               /* file ptr for reading undo file */
  ID undoShareList;          /* array of objs whose refcounts to be dec'd */
  char *UndoFile;
  int SeekFailed;
  int UndoFd;                 /* filedesc for writing undo file */
  int UndoPosBuf;             /* write position in buffer */
  int undoAble;              /* Whether anything can get added to undo file */
  int undoCnt;               /* used for row/col ins/del */
  int undoPause;             /* Pause undo support (used by ins/del cols/rows) */
  int undoing;               /* TRUE while undoing */
  long FileSize;              /* how much has been written */
  range_desc undoRange;       /* used for row/col ins/del */
  int UndoInited; /* = FALSE */
  unsigned int UndoCounter;	/* Just for posting a "working" message */
  char *tempname;			/* undo.c */

/* ssxfer.c */
  int LoadingISR ;/* = FALSE; */
  int over_msg;
  int sameLengthRequest; /*  = 0; */
  char *sel_data_buf; /*  = 0; */

/* uimain.c */
  int ElfRecordMode; /*  = SSABSOLUTE; */
  range_desc markingRange;

/* verify.c */
  FILE *VerFile;

  arg_desc ArgDesc;

  int numExposeEvents;	
 
  coordinate bltCell;
  int bltRngDiver;		/* Extract individual items from passed ranges */
  int bltTxtOk;
  int bltArrayIsAcceptable;
  boolean bltArrayGetFirstValue;
  int CalcError;

/* display.c */
  int GridTop;
  int GridLeft;
  int ROWHEAD_WIDTH; /* width of the header in front of a row */
  int COLDESC_HEIGHT; /* height of the header in front of a row */


  int RowSpaces;      /* Space padding below bottom row */
  int ColSpaces;      /* Space padding beside rightmost col */

  int patternfactor;
  SSRECT	DamagedRect;	/*  = 0, 0, 32000, 32000 */
  SSRECT	DispClipRect;	/*   = 0, 0, 32000, 32000 */
  SSRECT	lClip;	/*  = 0, 0, 32000, 32000 */
  SSRECT	gClip;	/*   = 0, 0, 32000, 32000 */

/* calc.c */
  coordinate calcCellCoord;
  ID calcCid;
  int circled;            /* circularity detected this pass */
  ID  circ_head;          /* cell id where circ detected */
  jmp_buf CalcExit;
  int CalcRow;
  int CalcCol;
  int CalcPlate;
  ID  CalcCid;
  range_desc calcArrRange;	/* array range of the formula currenly calc'ing */

/* getBlock.c */
  int msg_code;


  char *GetStorage;
  int CurString;
  int pos;      	/* SS.used to keep track of position in buffer */

/* copy.c */
  int del_action_code;
  range_desc Rng;


/* graph.c */

/* dispsetup.c */
  int uiEK_BASE; /*= EK_BASE;       start of internal GOTO GOTO key nums */
  int uiHK_BASE; /*= HK_BASE;       start of internal homeline key nums */
  int uiZK_BASE; /*= ZK_BASE;      * start of internal selection line nums */

  int uiMAX_ATM_KEY; /*= MAX_ATM_KEY;    ** max "real" key from the atm  */
                                    /** we handle, in "dispatch.h"   */

  int uiMAX_KEY;/* = MAX_KEY;      ** largest internal key number supported */
                              /**  in "dispatch.h"                      */


/* SSsort.c */
  tuple_ptr sort_blocks[6];   /* assumes min block size of 32K */


/* move.c */
  int scroll_factor;    /* what portion of the screen to scroll if  */

/* movec.c */
  range_desc *MvSrc, *MvDst, *MvDel;
  short DidGorch;
  int cellsRngAction;
  char Cols_gorched[TOUCHED_MODULO/8];   
  char Rows_gorched[TOUCHED_MODULO/8];
  char Plates_gorched[TOUCHED_MODULO/8];



/* fault.c */

  jmp_buf pow_env;
  int matherr;   /* TRUE if matherr() called. */
  int doLongJmp; /* Tested in matherr().  If on, matherr() does
                          a LongJmp exit. */



/* dumpax.c */
  ID FidArray;     /* Array of already dumped formula IDs */
  int NumFids;     /* How many are in the above array */



/* SSsort.c */
  int num_sort_keys;
  int num_sort_entries;
 
/* 
 * the following are defined as globals so they 
 * can be accessed by the qsort comparison routines  
 */

  int key1_ascending;	/* TRUE for key1 ascending sort */
  int key2_ascending;	/* TRUE for key1 ascending sort */
  int key3_ascending;	/* TRUE for key1 ascending sort */

  int row_wise;
  int32 key1, key2, key3;
/*********/
  jmp_buf prs_env;
  tok_ptr *ToksP;	/* addr of array of token pointers */
  tok_ptr *ExtrasP;	/* addr of array of extra-tokens pointers */
  char *LexP;		/* current addr while lexing */
  char *InStr;		/* addr of string to lex */
  short OperLast;	/* TRUE if last lexed token was operator */
  short NumToks;	/* Number of tokens in token chain */
  short NumExtras;	/* Number of tokens in "extras" chain */
  short ColBias;	/* col to de-bias all relative specs */
  int32 RowBias;	/* row to de-bias all relative specs */
  short PlateBias;	/* plate to de-bias all relative specs */
  short LexNum;		/* location counter for tokens */
  char *TokStore;	/* addr of scratchpad storage for tokens */
  char *CurTokP;	/* currently lexed token */
  short err_col;
  short *PosToks;	/* string pos of each token */

/*****/
  int rand_initialized;

/* graph.c */
  int didGroup; 

  
  ID NAform;
  range_desc *yrange; /* range consisting of y -  values */
  range_desc *xrange; /* range consisting of x - values */
  int rowinc;	/* if TRUE, data is arranged in rowwise */
  int colinc; 	/* if TRUE , data is arranged in colwise */
  int regerror; /* if TRUE, error occurred during calcuation */
  int DbError;	/* True if error occured */
  range_desc DbRange;      /* The database range */
  int DbRows;              /* Number of data rows in the database */
  bks_list	*bksArray[MAX_BKS_LEVEL+1];

  long timeOfButtonClick;	/* time when button was clicked */
  int menuId;			/* used by elf for running menubar macros */

  GC WorkGC;

  window_value *win;
  window_value *w;
  short CurWinIdx;		/* The Current Active SS win[] Index */
  short NumWinOpen;		/* Number of Windows Open for a single SS Obj */
  int warn_count;
  double nextAutoSaveTime;
  double nextLinkPollTime;
  double nextCalcTime;
  double nextBackCalcTime;
  int	EditModeIdx;		/** set to the window idx where editmode is 	
							currently on. Used for controlling the editmode 
							functionality **/
  char printerClass[STRNORMAL+1]; /* eg "PostScript" / "PCL5" */
  char printerType[STRNORMAL+1];  /* eg "HP3SI523.PPD" / "LaserJet4" */
  int 	elfCaller;	/* THIMpid of ss task being called */
  int ssTaskBeingCalledByElf;	/* THIMpid of elf task calling SS */
  int oldParent;				/* relavent only during ElfbSSCall() */
  int oldPended;				/* relavent only during ElfbSSCall() */
  int doingElfCommand;			/* we are processing an elf request, */
								/* relevent only during ElfbSSCall() */

  GPHOBJ Select;    /* doc based selected graphic obj */
  GPHOBJ *CurGphObj;/* Current Graphic Object Active in the SS */
  GPHOBJ RngObj;    /* Current Cell/Range Object Active in the SS */
  GPHOBJ OrgObj;    /* Temporary Object */
  GC InvertGC;      /* temp for resizing Graphic Object */
  short dragDirection;  /* used when resizing cols/rows */
  short dragStatus;     /* for Object dragging & drag & drop */
  short dragMode;       /*for Object dragging & drag & drop or NewObject Size */
  int dragXStart; /* button down location in doc pixels */
  int dragYStart; /* button down location in doc pixels */

  /* file io parameters */
	int fileSaveMode;		/* see fs_save.h */
	int fileGrpAccess;	/* Unix group enabled */
	int fileAllAccess;	/* Unix all access enabled */
	int fileWritable;	/* whether saved/read in writeable mode */
	char *docname;	/* just the document name */
	char *pokedSelectionStr;	/* the full pathname */
	char *pathname;	/* the full pathname */
	char *parentPathName;
	char *iconTitle;	/* the name to be put up for the icon */
	char *windowTitle;
	int hdrSize;				/* size of headers in mills */
	int ftrSize;				/* size of footers in mills */
								/* window we are loading into */
	short *rowPos;
	short *colPos;
	uint8 curDispWin;			/* the index of the multiple windows, the */
								/* the display is currently going on */
	uint8	pasteInfo;
	char *selectedObj;
	char *editText;	/* if not null string, selection is editline */
	pointMode pointmode;
										/* for file already open, else ignore */
	ss_stroke_buf stkBuf;

	attrStrEntry *attrStrHashTable;
	ss_attr_dump_type *ssaP;

	char *attrPoolattrsBuf;		/* style allocation pool */
	ss_attr_pool_t **attrPool;/* alloced array of ptrs into attrPoolattrsBuf */
	int attrPoolSize;		/* how many items in pool */
	ss_pool_item_t **attrHash;	/* hashed access to attributes pool */

	char *attrPoolbrdrsBuf;		/* style allocation pool... 2nd level borders */
	ss_border_pool_t **brdrPool;/* alloced array of ptrs into attrPoolattrsBuf*/
	int brdrPoolSize;		/* how many items in 2nd level border pool */
	ss_brdr_pool_item_t **brdrHash;	/* hashed access to attributes pool */

	char *attrPoolfontBuf;		/* style allocation pool... 2nd level borders */
	ss_font_pool_t **fontPool;/* alloced array of ptrs into attrPoolattrsBuf*/
	int fontPoolSize;		/* how many items in 2nd level border pool */
	ss_font_pool_item_t **fontHash;	/* hashed access to attributes pool */

	SS_FONT_ENTRY	*fontPtrArr;	/* array of font pointer info */

	char *attrPoolstyleBuf;		/* style allocation pool... 2nd level borders */
	ss_style_pool_t **stylePool;/* alloced array of ptrs into attrPoolattrsBuf*/
	int stylePoolSize;		/* how many items in 2nd level border pool */
	ss_style_pool_item_t **styleHash;	/* hashed access to attributes pool */
  	cmap_type	extCmap;	/* array of color definitions */
  	fmap_type extFmap;		/* table of font faces */




	ss_mem_desc calcMem;	/* builtin functions allocated memory */
	data_ptr	calcDataP;	/* data created by builtin functions, that */
							/* need to be freed incase of math/FPE longjmps */

	/* some display check vars */
	GPHOBJ *noRegenObj;		/* the object that should'nt be displayed */
							/* during display cleanup */
	range_desc	autoResizeRowsRng;	/* the range thats needs auto row resize*/
    caddr_t rtRegisterList;    /* an elf data_ptr casted out by caddr_t */
    caddr_t mvDataList;    		/* an elf data_ptr casted out by caddr_t */
	coordinate lastCalcedCell;	/* last background re-calced cell */
	short defcolwidth;	/** for default col width used in document 
				setting ***/
	view_plate_str_type *viewStrs;		/* array of view strings */
	int		numViewStrs;			/* number of current view strs */
	win_info_type	*winArr;
	int		elfMacroWin;
	int		saveCurWinIdx;
	char    *nmBadChars;	
	int		numObsCells;	 /* number of obsolete cells remaining */
	int		numCellsCalced;	 /* number of cells calc'ed */

	range_desc arrayRng;	/* contains the array range to be put */	
	ss_cur_func_calc_dex curIfDexes;	/* struct to hold if function's calc */
										/* dexes to be skipped in the */
										/* current function */	
	data_ptr helpErrorObj;
	int originalVersion;	/* the original version file was created in */
	int buttonPressX; /** stores the current x at the button press */
	int buttonPressY; /** stores the current y at the button press */
	int yetToFixForm ; /* used as a flag when compute cell during the 
							copy/paste  in manual mode**/
	int eventCounter;	/* the ss event counter */
	int lastIndirectUpdate; /* the count at which last update occured */
	rng_hash_type rngHash;
	name_range_hash_type nameHash;
	int defAttrDex;	/* the attr dex for default doc setting attr */
	unsigned int	elfCallCounter;	/* ElfCallByName nested counter */
	double CalcBgEndTime;
	double 	curDateTime;
	long curTimeOfDay;
    todo_list_item_t *nowList;   /* Head of now function update list */
	ID curView;
	ID editView;	/* the view in which editmode was initiated */
	double rtEngineRev;			/* the rt gateway rev */
	int	(*axSSMathErr)();		/* the matherr handler for addin functions */
	data_ptr	pasteP;			/* the paste info */	
	void		*elf_bld_obj;	/* elf_obj set by builder **/
	refDepEntry	*referencesList;
	int			numReferencesList;
	refDepEntry	*dependentsList;
	int			numDependentsList;
	unsigned short EditMode;		

  short CHREGOFS;          /* offset used to space the column descriptor line */
  short COL_WIDTH_PAD;    /* how many pixels to pad between cols on grid */
  short CUR_CELL_HEIGHT;  /* how hight the cell area inside the grid lines is */
  short CUR_ROW_HEIGHT;
  short DEF_MIL_HEIGHT;      /* height of a character cell in mills */
  short DEF_MIL_WIDTH;       /* width of a character cell in mills */
  short DEF_PIXEL_HEIGHT;      /* width of a character cell in pixels */
  short DEF_PIXEL_WIDTH;       /* width of a character cell in pixels */
  short MIL_HEIGHT;      /* height of a character in mills */
  short MIL_WIDTH;       /* width of a character in mills */
  short PIXEL_HEIGHT;      /* width of a character in pixels */
  short PIXEL_WIDTH;       /* width of a character in pixels */
  short NON_ZOOM_PIXEL_WIDTH ; /** actual widht of a character and not ZOOMED*/
  short ParentWid;
  short Print_Close; /*  whether we are doing a print-and-close */
  short ROW_HEIGHT_PAD;   /* how many pixels to pad each row on the grid */
  short TEXT_MIL_HEIGHT; 
  short TEXT_MIL_WIDTH;
  short TEXT_PIXEL_HEIGHT;
  short TEXT_PIXEL_WIDTH;
 
/* 
 * the following variables are used for performance improvements
 * when fixing a formula during an insert.  Putcell looks at
 * them when determining whether the empty hash needs recalc. 
 */
  short InsRowColPlate; /* set when inserting column or row */
  short InsRow; /* set when inserting rows */
  short InsCol; /* set when inserting columns */
  short DelRowColPlate;    /* set when deleting columns/rows */
  short DelRow; /* set when deleting rows */
  short DelCol; /* set when deleting columns */
  short InsPlate;
  short DelPlate;
 
  short skipBackPtrMash; 
  short rangeSpans; 

	/* BOOLEANS START HERE */
	unsigned int 	forceRtUpdate:1,	/* TRUE if force rt update */	
					memErr:1,
					strsInited:1,		/* strings allocated */
  					chkERRinForm:1,
  					ERRinForm:1,
  					mvFixPut:1,
					calcCellIsArrayForm:1,/* 1 if cell being calc'ed 
												is array formula*/
  				docMod:2,/* SS_SAVE_AND_BACKUP for saving and autoback
							SS_SAVE_ONLY means since the last autoback there
								is no new modification so don't do autobackup
							SS_SAVE_NOTHING means nothing has changed since
								the last save */
  				hdnAffected:1,   /* Hidden material affected by command */
  				moving:1,        /* TRUE during PasteCut */
				duringPasteOnly:1,/* set only during Cut/Paste, Copy/Paste */
  				newBars:1,
  				newBoom:1,
  				newBounds:1,         /* re-compute bounds */
  				noOverPut:1, /* prevent PutCell from overwriting material */
  				protOver:1,      /* Protected Cells where overwritten by  */
									/* PutCell */
  				putOver:1,       /* Cells where overwritten by PutCell */
  				undoLogging:1,   /* Noting cells for undo plate */
	rtStuffRegistered:1,	/* TRUE if rt stuff is registered */
	hasRtStuff:1,			/* TRUE if there are rt formulas */

	ignoreProt:1,	/* TRUE to ignore protection */
	calcing:1,
	skipGridMapUpdate:1,	/* TRUE to skip grid map update, for insert/
									delete sheets only */
	calcInterruptNow:1,	/* TRUE to interrupt calc immediately */
	backCalcIntrptd:1, /* TRUE if back calc is interrupted */
	backCalcNow:1,	 /*  TRUE if back calc should be resumed */
	calcDisplayNow:1,	 /*  TRUE needs to do display calc now */
	viewNameStrsOutOfDate:1,	/* TRUE if need re-compute viewnames */
    pageBrksOutOfDate :1, /** TRUE if page Breaks are out of date **/
	doingFileReadWrite:1,	/* TRUE if reading or writing a document */
	haveTriedMVConnect:1,
	checkAutoResize:1,	/* TRUE if rows are to be autoresized */
	redrawCharts:1,		/* TRUE if charts need to be redraw */
	rowsNeedResize:1,
	objectsRedisplay:1,	/* TRUE if objects need display */
	arrayFormulaMode:1,		/* true if should create array formula*/
	checkForFileAlreadyOpen:1,	/* if true we would check for */
	useLinkedSSAttrs:1,	/* use attrs from the linked external ss */
	curWinNoGridAddr:1,	/* TRUE if currently no grid address in the */
	fileOnDisk:1,		/* saved/read-from above name */
	childDoc:1,	/* if this is an inset */	
	hdrFtrSizesCalced:1,	/* TRUE if hdrs/ftrs size calculated */
  	statusLineDisabled:1,
  	cursorOnDisplay:1,   /* FALSE after UnDraw, TRUE after Draw */
  	optimize_display:1,	/* TRUE if should clear before write text */
  printing:1,	/* TRUE if SS is printing */
  ZUnmarkOnRead:1,	/* = FALSE */
  BadColDesc:1,     /* Column Descriptor String Valid */
  BadRegions:1,     /* Grid regions are foobared */
  clickIsDouble:1,	/* TRUE if double clicked */
  CircularISRs:1,
  BadRowDesc:1,
  BadRowHeight:1,
  printRange:1, /* FALSE if regular print  */
  writingToClipBoard:1,	/* TRUE if writing to clipboard */
  linkPollingEnabled:1, /* TRUE means check links periodically */
  dbViewColMode:1,	/* TRUE if in view column mode */
  autoSaveEnabled:1,	/* TRUE if autoSave is enabled */
  LocalizeLinks:1, /* Spreadsheet opened from mail */
  ForMail :1, /* TRUE if spreadsheet was opened from mail inbox */
  NewDoc:1,
  NewWindowCreated :1, /* TRUE if a new SS window is create */
  UnmarkOnRead:1,  /* if TRUE, ReadPrompt unmarks items */
  copyfile:1,
  delfile:1,
  fastOpen:1,
  marking:1,
  noRedisp:1,
  quitting:1,
  restart:1,
  restartingNewFile:1,	/* TRUE for Open... (new file) */
  savefile:1,
  NeedsRedisplay:1, /* TRUE if SS needs redisplay, as a result */
  mouseButtonIsDown:1,
  needs_redisplay:1, /* if we should go thru main loop and RedispAll */
  AutoResizeRows:1,	 /* if rows to be resized automatically */
  doing_regen:1,	 /* TRUE if we are regening an exposed area */
  arrow_key_pressed:1, /* TRUE  if an arrow key was pressed in work area */
  In3dMode:1,
  awaiting_resize_expose:1,	/* TRUE if waiting for an expose after resize */
  paste_to_editwin:1, /* TRUE if paste destination is editwindow */
	haveTriedForRealtimeLicense:1,	/* set after first try */
	busyReadingFile:1,	/* ss is busy reading file */
	busyDisplaying:1,		/* ss is busy displaying */
	shrinkArrayFormTable:1,	/* TRUE if need to shrink array form table*/
	needsRecalc:1,		/* TRUE if there are obsolete cells in grid */
	keyShifted:1,		/* TRUE if shift key was pressed with key event */	
	keyStrokeEvent:1,	/** TRUE for key events and FALSE for mouse EVENTS **/
	converting:1,		/* TRUE if in process of rev filtering */
	dataChanged:1,		/* TRUE if data changed in SS */		
	noDisplay:1,		/* TRUE if no display at all, make no attmept */
						/* to call the displayer */
	noWindows:1,		/* TRUE if no windows were created, make no attempt */
						/* to create or destroy windows */
	hasLocalClip:1,		/* TRUE if has local clip */
	hasGlobalClip:1,	/* TRUE if has global clip */
	readOnly:1,			/* TRUE if SS is in readonly mode */
	elClicked:1,		/* TRUE if clicked in editline, valid only in editmode*/
	hasEditWindow:1,	/* TRUE if displayed window has editline */
	inCellEditWinUp:1,	/* TRUE if in-cell edit window is up */		
	readFromBuf:1,		/* TRUE to read from buffer */
	writeToBuf:1,		/* TRUE to write to buffer */
	insetInited:1,		/* TRUE if inset def source is inited */
	forceClear:1,		/* TRUE if delete cells even if they have row-col 
							formats*/
	caseSort:1,			/* TRUE if this is case-sensitive sort*/
	redoRTSOnLiveFeedEnable:1,	/* TRUE to re subscribe on enabling live feed */
	dblClickCallback:1,		/* TRUE of dblClickCallback is set **/
	saveDblClickCallback:1, /* save to the file */
	loadCellCallback:1,		/* TRUE of load cell callback  is set **/
	saveLoadCellCallback:1, /* save to the file */
	selectCallback:1,		/* TRUE of select callback is set **/
	saveSelectCallback:1, 	/* save to the file */
	crCallback:1,			/* TRUE of carriage return callback is set **/
	saveCRCallback:1, 		/* save to the file */
	uiLoadCellCallback:1,	/* TRUE of carriage return callback is set **/
	saveUILoadCellCallback:1, /* save to the file */
	globalDbCallback:1,		/* TRUE if global double callback set */
	globalCrCallback:1,		/* TRUE if global carriage return callback set */
	globalLoadCellCallback:1,/* TRUE if global load cell callback set */
	globalSelectCallback:1,	/* TRUE if global selection callback set */
	globalUILoadCellCallback:1,/* TRUE if global ui load cell callback set */
	openFileInManualMode:1, /* TRUE if want to open file in manual mode 
								irrespective of what is saved in the file */
    noMergeCellsForWP:1, /*TRUE do not merge adjacent cells for bleed over */
	noAttrDexAddOnCellDel:1; /* TRUE when using DoCutPaste during features like
							delete cells. The problem is when cells are deleted
							the attr dex gets added to the row list if there
							exist a row/col format. In such cases it should not
							have to worry about it. */
  } ss_value;



#endif /* SS_H_INCL */
