/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ssbltins.c
   Applix ASC version: 25
   Date: Mon Jun 21 21:11:13 1999
*/
#include <stdio.h>
#include <math.h>
#include <axtime.h>
#include <global.h>
#include <ssfunc.h>
#include <gstyle.h>
#include <ssbltins.p>


/*============================================================================*/
/*      SS Date utilities                                                     */
/*============================================================================*/
extern DLL_IMPORT foreignChars UTprofsym;
static unsigned char decimalChar = 0xFF;
double          AxSSPi = 3.1415926535898;
extern double   axpow();
#define ss_value void

double 
ssToday( isError)       /* return todays date as a SS julian date */
    int            *isError;
{
    struct tm      *tmp;
    long            clock;
    double          toDay;

    *isError = FALSE;
    clock = time((long *) 0);
    tmp = localtime(&clock);
    toDay = DateBuild(tmp->tm_year, tmp->tm_mon, tmp->tm_mday); /* LOK la */
    return (toDay);
}

double 
ssNow(isError)
    int            *isError;/* return current time */
{
    struct tm      *tmp;
    long            clock;
    double          the_date, the_time;
    SS_TIME_STRUCT  ts;
    double          val;

    clock = time((long *) 0);
    tmp = localtime(&clock);
    the_date = DateBuild(tmp->tm_year, tmp->tm_mon, tmp->tm_mday);
    ts.hour = tmp->tm_hour;
    ts.min = tmp->tm_min;
    ts.sec = tmp->tm_sec;
    ts.tsec = 0;
    the_time = TimeBuild(&ts);
    *isError = FALSE;
    return (the_date + the_time);
}

double 
ssYearAbs(date, isError)
	double date;
	int *isError;
{
double yr;
	yr = ssYear(date, isError);
	if (*isError != TRUE) {
	/* since DateExtract subtracts 1900 from the number we should add it here*/
		yr = 1900.0 + yr ;
	}
	return(yr);
}

double 
ssYear(date, isError)
    double          date;
    int            *isError;
{
    double          aval;
    short           year, month, day;

    if (DateExtract((unsigned long) date, &year, &month, &day) < 0) {
        *isError = TRUE;
        return (0.0);
    }
    *isError = FALSE;
    return ((double) year);
}

double 
ssMonth(date, isError)
    double          date;
    int            *isError;
{
    double          aval;
    short           year, month, day;

    if (DateExtract((unsigned long) date, &year, &month, &day) < 0) {
        *isError = TRUE;
        return (0.0);
    }
    *isError = FALSE;
    return ((double) month + 1);
}

double 
ssTime(hour, min, sec, isError)
    double          hour, min, sec;
    int            *isError;
{
    double          the_time;
    SS_TIME_STRUCT  ts;

    *isError = TRUE;
    if ((hour > 23.0) || (hour < 0.0) ||
        (min > 59.0) || (min < 0.0) ||
        (sec > 59.0) || (sec < 0.0)) {
        return (0.0);
    }
    ts.hour = hour;
    ts.min = min;
    ts.sec = sec;
    ts.tsec = 0;
    the_time = TimeBuild(&ts);
    *isError = FALSE;
    return (the_time);  /* LOK la */
}

double 
ssHour(timeValue, isError)
    double          timeValue;
    int            *isError;
{
    short           hour, min, sec;
    double          the_time;
    SS_TIME_STRUCT  ts;

    if (timeValue < 0.0) {
        *isError = TRUE;
        return (0.0);
    }
    timeValue -= int_part(timeValue);
    TimeExtract(timeValue, &ts);
    *isError = FALSE;
    return ((double) ts.hour);
}

double 
ssMinute(timeValue, isError)
    double          timeValue;
    int            *isError;
{
    short           hour, min, sec;
    double          the_time;
    SS_TIME_STRUCT  ts;

    if (timeValue < 0.0) {
        *isError = TRUE;
        return (0.0);
    }
    timeValue -= int_part(timeValue);
    TimeExtract(timeValue, &ts);
    *isError = FALSE;
    return ((double) ts.min);
}

double 
ssSecond(timeValue, isError)
    double          timeValue;
    int            *isError;
{
    short           hour, min, sec;
    double          the_time;
    SS_TIME_STRUCT  ts;

    if (timeValue < 0.0) {
        *isError = TRUE;
        return (0.0);
    }
    timeValue -= int_part(timeValue);
    TimeExtract(timeValue, &ts);
    *isError = FALSE;
    return ((double) ts.sec);
}




double 
ssDay(date, isError)
    double          date;
    int            *isError;
{
    double          aval;
    short           year, month, day;

    if (DateExtract((unsigned long) date, &year, &month, &day) < 0) {
        *isError = TRUE;
        return (0.0);
    }
    *isError = FALSE;
    return ((double) day + 1);
}



double 
ssDate(year, month, day, isError)
    double          year, month, day;
    int            *isError;
{
    short           yr, mo, dy;
    unsigned long   the_date;
    short           yr2, mo2, dy2;

    *isError = TRUE;    /* assume error */
    if ((int)year >= 1900)
        year -= 1900;
	else if ((int)year >= 100) {
		/* three digit year passed is taken as absolute from 1900 and do not
			do anything */
	}
	else { /* just two digits are passed and we try to get the year since 1900*/
		year = axGetBaseYear((int)year) + year - 1900;
	}

    /*
     * The spreadsheet now supports dates until 2200,Dec 31 so the year
     * can be now checked against 200 ,instead of 199 - Prem. Nov-2-1990.
     */
    if ((day > 31) || (day < 1) ||
        (month > 12) || (month < 1) ||
        (year > 300) || (year < 0)) {
        *isError = TRUE;
        return (0.0);
    }
    yr = year;
    mo = month;
    dy = day;
    the_date = DateBuild( yr, mo - 1, dy);

    if (DateExtract(the_date, &yr2, &mo2, &dy2) < 0)
        return (0.0);
    /* rnm 3/9/86 #4973 */

    if ((yr != yr2) || (mo != mo2 + 1) || (dy != dy2 + 1))
        return (0.0);

    *isError = FALSE;
    return ((double) the_date);
}

double 
ssAddDays(datenum, days, isError)
    double          datenum, days;
    int            *isError;
{
    double          the_date;

    the_date = datenum + days;
    if (the_date <= 0) {
        *isError = TRUE;
        return (0.0);
    }
    *isError = FALSE;
    return (the_date);
}

/*
*   function ssAddMonths
*   add the given months and update the date
*/
double 
ssAddMonths(datenum, num_months, isError)
    double          datenum, num_months;
    int            *isError;
{
    unsigned long   the_date;
    short           year, month, day, new_month;
    int             tmonth, tyear, int_months;
    double          the_time;

    /* get the args */
    the_date = floor(datenum);
    the_time = datenum - the_date;
    *isError = TRUE;

    /* extract the years, months and days from the date number */
    if (DateExtract(the_date, &year, &month, &day) < 0)
        return (0.0);

    if (num_months != 0) {
        int_months = (int) num_months;
        tyear = int_months / 12;
        tmonth = int_months % 12;
        year += tyear;
        month += tmonth;
        if (month < 0) {
            /** decrease the year by 1, and roll over the month **/
            year--;
            month += 12;
        } else if (month > 11) {
            /** increate the year by 1, and roll over the month **/
            year++;
            month -= 12;
        }
    }
    the_date = DateBuild(year, month, day + 1);
    *isError = FALSE;
    return ((double) (the_date + the_time));
}

/*
*   function bltAddYears -
*   add the given number of years and update the date
*/
double 
ssAddYears(datenum, num_years, isError)
    double          datenum, num_years;
    int            *isError;
{
    double          the_time;
    unsigned long   the_date;
    short           year, month, day;

    /* get the args */
    *isError = TRUE;
    the_date = floor(datenum);
    the_time = datenum - the_date;

    /* extract the years, months and days from the date number */
    if (DateExtract(the_date, &year, &month, &day) < 0)
        return (0.0);
    year += num_years;
    the_date = DateBuild( year, month, day + 1);
    *isError = FALSE;
    return ((double) (the_time + the_date));
}

/*
*   function ssAddHours
*   add the given number of hours and update the time and also the
*   date if necessary
*/
double 
ssAddHours(datenum, num_hours, isError)
    double          datenum, num_hours;
    int            *isError;
{
    SS_TIME_STRUCT  ts;
    double          this_time, this_date;
    double          new_hour;
    double          delta_time;

    this_time = datenum - floor(datenum);
    this_date = datenum - this_time;
    delta_time = num_hours * (1.0 / 24.0);
    NewTimeDate(&this_date, &this_time, delta_time);
    *isError = FALSE;
    return (this_date + this_time);
}


/*
*   function bldAddMinutes
*   add the given number of minutes and update the time and also the
*   date if necessary
*/

double 
ssAddMinutes(datenum, num_minutes, isError)
    double          datenum, num_minutes;
    int            *isError;
{
    double          delta_time;
    double          the_date, the_time;

    the_date = floor(datenum);
    the_time = datenum - the_date;
    /* calculate the total change in time given the minutes */
    delta_time = (1.0 / (24 * 60)) * num_minutes;
    NewTimeDate(&the_date, &the_time, delta_time);
    *isError = FALSE;
    return (the_date + the_time);
}

/*
*   function bldAddSeconds
*   add the given number of seconds and update the time and also the
*   date if necessary
*/
double 
ssAddSeconds( datenum, num_seconds, isError)
    double          datenum, num_seconds;
    int            *isError;
{
    double          delta_time;
    double          the_date, the_time;
    double          aval;

    the_date = floor(datenum);
    the_time = datenum - the_date;
    /* calculate the total change in time given the minutes */
    delta_time = (1.0 / (24 * 60 * 60)) * num_seconds;
    NewTimeDate(&the_date, &the_time, delta_time);
    *isError = FALSE;
    return (the_date + the_time);
}

DLL_EXPORT double 
ssExp(val, isError)
    double          val;
    int            *isError;
{
    double          retval;
	errno = 0;
    retval = exp(val);
    if ((errno != 0) || (SPL_ERROR == retval)) {
        *isError = TRUE;
        retval = 0.0;
    } else
        *isError = FALSE;
    return (retval);
}

double 
ssPower(value, exponent, isError)
    double          value, exponent;
    int            *isError;
{
    double          z;

    *isError = FALSE;
	errno = 0;
    if (value < 0.0) {
        /* for negative values we only allow integer exponents */
        z = floor(exponent);
        if (exponent != z) {
            /* non-integer */
            *isError = TRUE;
            return (0.0);
        }
    }
    errno = 0;      /* set global math error variable */
    z = axpow(value, exponent);
    /* check for exception */
    if (z == 0.0) {
        /*
         * we probably have trouble, its ok if value is 0 and
         * exponent is > 0, otherwise exception condition occured.
         */
        if ((value == 0.0) && (exponent > 0.0)) {
            ;   /* ok - so continue */
        } else {
            /* result is zero and input base value is not */
            *isError = TRUE;
        }
    } else if (errno > 0) {
        /* math routines set the global variable */
        /* do_ss_pow - errno set  */
        *isError = TRUE;
        z = 0.0;
    }
    /* passed all the tests, all is well */
    return (z);
}

double 
ssRound(number, prec, isError)
    double          number, prec;
    int            *isError;
{
    int             negative = FALSE;
    double          shift;

    *isError = FALSE;
	errno = 0;
    prec = int_part( prec);
    shift = ssPower(10.0, prec, isError);
    if (*isError)
        return (0.0);
    if (number < 0) {
        negative = TRUE;
        number = fabs(number);
    }
    /* Division by zero should be impossible */
    number = floor(number * shift + 0.5000000001) / shift;
    if (negative)
        number = -number;
    return (number);
}

double 
ssPi()
{
    return (AxSSPi);
}


/*
*   Function NewTimeDate - Give the change in time, updates the time
*   and also date if necessary
*/
NewTimeDate( the_date, the_time, delta_time)
    double         *the_date, *the_time, delta_time;
{
    double          new_time;
    double          delta_date;
    new_time = *the_time + delta_time;
    if (new_time >= 1.0 || new_time < 0.0) {
        /* the date will change */
        /*
         * the integral part will give the number of days the date
         * will be
         */
        /* affected by and the fractional part will give the time */
        if (new_time >= 1.0) {
            /* get the integral part */
            delta_date = floor(new_time);
            *the_date += delta_date;
            /* set the time to the fractional part */
            *the_time = new_time - delta_date;
        } else {    /* subtracting time is more subtle */
            /*
             * eg: if new_time is -3.82, we want -4 days to be
             * subtracted
             */
            /* from date and the time should be (1.0-82) ==> 0.18 */
            /* delta_date ==> floor(-3.82) ==> -4.0 */
            /*
             * time = -delta_date+new_time ==> -(-4.0) + -3.82
             * ==> 0.18
             */
            delta_date = floor(new_time);
            *the_date += delta_date;
            *the_time = -delta_date + new_time;
        }
    } else          /* the date will not change */
        *the_time = new_time;
}


ssTimeStr(tvalue, fmt, timeStr, isError)
    double          tvalue;
    int             fmt;
    char           *timeStr;
    int            *isError;
{
    SS_TIME_STRUCT  ts;
    int             retval;

    TimeExtract(tvalue, &ts);
    *timeStr = '\0';
    if (0xFF == decimalChar) {
        /* initialize the decimal char */
        UTgetsym();
        decimalChar = UTprofsym.decimalChar;
    }
    retval = ssGetTimeStr((ss_value *)NULL, tvalue, fmt, timeStr, ':', decimalChar,
                  NXLT("AM"), NXLT("PM"));
    if (!retval)
        *isError = TRUE;
    else
        *isError = FALSE;
}

ssDateStr(dvalue, fmt, fullYear, dateStr, isError)
    double          dvalue;
    int             fmt;
    char           *dateStr;
    int            *isError;
{
    if (!ssGetDateStr((ss_value *)NULL, dvalue, fmt, dateStr, fullYear))
        *isError = TRUE;
    else
        *isError = FALSE;
}

double 
int_part( x)
    double          x;
{
    if (x >= 0.0)
        return (floor(x));
    else
        return (ceil(x));
}


/*
*   ssFuncMathError - general matherr function for spreadsheet functions
*
*/

#if !(LINUX && LIBC5)
ssFuncMathError(ssP, mathP, isMathError)
ss_value    *ssP;
#if AXMS
    struct _exception *mathP;
#else
    struct exception *mathP;
#endif
    int            *isMathError;
{
    *isMathError = FALSE;
    switch (mathP->type) {
    case OVERFLOW:
        *isMathError = TRUE;
        mathP->retval = SPL_ERROR;  /* for all cases */
        break;
    case UNDERFLOW:
        mathP->retval = 0.0;    /* for all cases */
        break;
    case DOMAIN:
        *isMathError = TRUE;
        if (!strcmp(mathP->name, NXLT("pow")) ||
            !strcmp(mathP->name, NXLT("sqrt")))
            mathP->retval = 0.0;
        else
            mathP->retval = SPL_ERROR;
        break;
    case SING:
        mathP->retval = SPL_ERROR;
        *isMathError = TRUE;
        break;
    case TLOSS:
        *isMathError = TRUE;
        if (!strcmp(mathP->name, NXLT("bessel")))
            mathP->retval = 0.0;
        break;
    case PLOSS:
        *isMathError = TRUE;
        break;
    }
    return (1);
}
#else
ssFuncMathError()
{
    printf("ss/ssbltins.c: MATHERR\n");
}
#endif

