/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ssdate.c
   Applix ASC version: 28
   Date: Thu Feb 25 21:11:31 1999
*/
#include <math.h>
#include <aos.h>
#include <datetime.h>
#include <global.h>
#include <gstyle.h>
#include <ss_style.h>
#include <axtime.h>
#include <trim.h>
#include <ssfunc.h>
#include <ssdate.p>

/*
*	functions in this module
*
*	(1) long DateBuild(ssP, short year, short month, short day)
*		- given year,month,day build a 123 date number
*	(2) Explodedt(ssP, unsigend long datetimevalue, DateStruct *datestruct)
*		- given an alis datenumber break it into components
*	(3) DateExtract(ssP, unsigned long date, short *year, short *month, short *day)
*		- given a 123 date number return year, month and day
*
*/



/* number of days array */
short           ZDays[12] = {
	 /* Jan */ 0,
	 /* Feb */ 31,
	 /* Mar */ 31 + 28,
	 /* Apr */ 31 + 28 + 31,
	 /* May */ 31 + 28 + 31 + 30,
	 /* Jun */ 31 + 28 + 31 + 30 + 31,
	 /* Jul */ 31 + 28 + 31 + 30 + 31 + 30,
	 /* Aug */ 31 + 28 + 31 + 30 + 31 + 30 + 31,
	 /* Sep */ 31 + 28 + 31 + 30 + 31 + 30 + 31 + 31,
	 /* Oct */ 31 + 28 + 31 + 30 + 31 + 30 + 31 + 31 + 30,
	 /* Nov */ 31 + 28 + 31 + 30 + 31 + 30 + 31 + 31 + 30 + 31,
	 /* Dec */ 31 + 28 + 31 + 30 + 31 + 30 + 31 + 31 + 30 + 31 + 30
};

ConvertTmToLocalStruct(ssP, tm, as_struct)
ss_value *ssP;
struct tm *tm;
DateStruct *as_struct;
{
	as_struct->year = tm->tm_year - 1900;
	as_struct->month = tm->tm_mon;
	as_struct->day = tm->tm_mday;
	as_struct->hour = tm->tm_hour;
	as_struct->minute = tm->tm_min;
	as_struct->second = tm->tm_sec;
}


/************************  explodedt.c  *************************************
#define ThisSources explodedtFILE
NAME
	Explodedt - Convert a date number to a structure of components

SYNOPSIS

Explodedt(ssP,  datetimevalue, datestruct )
	ss_value *ssP;
unsigned long datetimevalue;	/* input parm, see datetime.h
DateStruct *datestruct;			/* returned filled in date structure

DESCRIPTION

This function writes all the components of the date structure.
They include year, month, day, hour, minute and second.

RETURN VALUE
 None.
*/


Explodedt(ssP, dtvalue, as_struct)
	ss_value *ssP;
	unsigned long   dtvalue;
	DateStruct     *as_struct;
{

#define lc   (60/DTTIMEGRANULARITY)
#define lcyr (lc*(unsigned long)599040)	/* 60*24*32*13   */
#define lcmo (lc*46080)		/* 60*24*32      */
#define lcdy (lc*1440)		/* 60*24         */
#define lchr (lc*60)

	dtvalue -= (as_struct->year = dtvalue / lcyr) * lcyr;	/* LOK la  */
	dtvalue -= (as_struct->month = dtvalue / lcmo) * lcmo;	/* LOK la  */
	dtvalue -= (as_struct->day = dtvalue / lcdy) * lcdy;	/* LOK la  */
	dtvalue -= (as_struct->hour = dtvalue / lchr) * lchr;	/* LOK la  */
	dtvalue -= (as_struct->minute = dtvalue / lc) * lc;	/* LOK la  */
	as_struct->second = dtvalue*DTTIMEGRANULARITY;	/* LOK la  */
}

/*
 * module:              DateExtract - Given a 1-2-3 date, compute year/month/day
 * input:               long date
 *                              short *year - where to put year (0-199)
 *                              short *month - where to put month (0-11)
 *                              short *day - where to put day of month (0-30)
 * algorithm:
 * side effects:
 */
DateExtract(date, year, month, day)
    register unsigned long date;
    register short *year;
    register short *month;
    register short *day;
{
    register long   did;
    register short  yr, mo, num_days, leap_year;
    int             rc;         /* #4973 */

    rc = 0;                     /* set return code *//* #4973 */

    /* verify that date is not too big or negative *//* #4973 */
    if ((date > MAX_DATE) || (date < 1)) {
        /* Dec 31, 2200, i.e., max date *//* #4973 */
        rc = -1;                /* rnm 3/9/86  #4973 */
        goto goback;            /* #4973 */
    }                           /* #4973 */
    date--;                     /* normalize for Jan 1, 1900 == 0 */
    did = 0;
	/* 
	*	EXCEL and LOTUS both maintain that 2/29/00 existed, which is
	*	wrong but to comply with their format, lets treat year 1900
	*	as a leap year.
	*/
	
	did = 366;	/* for the year 1900 */
	num_days = 366;	/* for 1900 */
	leap_year = TRUE;

    for (yr = 1901; did <= date; yr++) {
		/* NOTE - Date numbers are now stored according to the standard
		   set by Lotus and Excel, which maintain that the day 2/29/00
		   DID exist (and was date number 60) - this the comments in
		   the line below */
		leap_year = (((yr % 4) == 0) && ((yr% 100)!= 0)) || 
			(0 == (yr % 400)) || (yr == 0);
        num_days = 365 + leap_year;
        did += num_days;
    }
    date -= (did -= num_days);
    *year = (yr - 1) - 1900;
    for (mo = 11;; mo--) {
        num_days = ZDays[mo] + (leap_year && (mo > 1));
        if (num_days <= date)   /* LOK la */
            break;
    }
    *month = mo;
    *day = date - num_days;     /* LOK la */

goback:;                        /* #4973 */
    return (rc);                /* #4973 */

}


/*
 * module:  DateBuild - Given year month day, build a 1-2-3 date
 * input:   short year - 0-199
 *          short month - 0-11
 *          short day - 1-31
 * returns: long
 * algorithm:
 * side effects:
 */

long 
DateBuild(year,month,day)
	short year;
	short month;
	short day;
{
	long date;
	short num_leaps, leap_year;
	int actual_year;
	int add_num, sub_num;


	/* NOTE - Date numbers are now stored according to the standard
	   set by Lotus and Excel, which maintain that the day 2/29/00
	   DID exist (and was date number 60) - this the comments in
	   the line below */

	actual_year = year + 1900;
	/* one should not be added for every century accept when the century is
	a leap year for example it should not add 1 for years 2100, 2200, 2300 but
	add one for years 1900, 2000, 2400 */
	leap_year = (((actual_year % 4) == 0) && ((actual_year% 100)!= 0)) || 
		(0 == (actual_year % 400)) || (actual_year == 1900);
#if 1
	/* one should not be added for every century accept when the century is
	a leap year for example it should not add 1 for years 2100, 2200, 2300 
	but add one for years 1900, 2000, 2400 */
	if (year > 0) {
		/* According to the Lotus and Excel standard, 1900 and 2000 are 
			BOTH leap years */
		num_leaps = (year-1)/4 - ((actual_year-1)/100 - 19 - 1) + 
					((actual_year-1)/400 - 4);
	} else {
		num_leaps = 0;
	}
#else
	if (year > 0)	{
		/* According to the Lotus and Excel standard, 1900 and 2000 are 
			BOTH leap years */
		num_leaps = (year-1) / 4 + 1;	
	}	else
		num_leaps = 0;
#endif

    date = (long) year * 365 + num_leaps;

		
    if((month < 11) &&   ((ZDays[month] + day)  > ZDays[month+1]) ) {
    	leap_year = leap_year && (month >= 1);
    	date += ZDays[month+1] +leap_year;
	} else {
    	leap_year = leap_year && (month > 1);
		date += ZDays[month] +leap_year;
	    date += day;
	}

    return(date);
}

GetBldDate(ssP)
	ss_value *ssP;
{
}


double 
TimeBuild( ts)
	SS_TIME_STRUCT *ts;
{
	double          timevalue;
	timevalue = ts->hour * HOURFACTOR +
		ts->min * MINFACTOR +
		ts->sec * SECFACTOR +
		ts->tsec;
	return (timevalue * SS_TIME_GRANULE);
}


TimeExtract(tvalue, ts)
	double          tvalue;
	SS_TIME_STRUCT *ts;
{
	int             granules;

	tvalue -= floor(tvalue);
	if (tvalue < 0.0) {
		/* lotus way of treating negative values for times */
		tvalue += 1.0;
	}
	if ((tvalue < 0.0) || (tvalue > 1.0))
		return (-1);

	granules = (int) ceil(tvalue / SS_TIME_GRANULE);
	ts->hour = granules / HOURFACTOR;
	granules -= ts->hour * HOURFACTOR;
	ts->min = granules / MINFACTOR;
	granules -= ts->min * MINFACTOR;
	ts->sec = granules / SECFACTOR;
	ts->tsec = granules - ts->sec * SECFACTOR;

	/** set the hour from 0-23 and for anthing greater set it to 0 **/
	if (ts->hour > 23)
		ts->hour = 0;
	return (AOK);
}




/*
 * module:              trim - remove either leading or trailing spaces (or both)
 * input:               char *str - the untrimmed string and the result location
 *              int code
 * algorithm:
 * side effects:
 */

trim(ssP, str, code)
	ss_value *ssP;
	register char  *str;
	int             code;
{
	register char  *p;

	switch (code) {
	case TRIM_RIGHT:
		for (p = str; *p != '\0'; p++);
		p--;
		while ((p > str) && (*p == ' '))
			*p-- = '\0';
		break;
	case TRIM_BOTH:
		for (p = str; *p != '\0'; p++);
		p--;
		while ((p > str) && (*p == ' '))
			*p-- = '\0';
	case TRIM_LEFT:
		for (p = str; *p == ' '; p++);
		MBstrcpy(str, p);
		break;
	}
}

wtrim(ssP, str, code)
	ss_value *ssP;
	register wchar *str;
	int             code;
{
	register wchar *p;

	switch (code) {
	case TRIM_RIGHT:
		for (p = str; *p != '\0'; p++);
		p--;
		while ((p > str) && (*p == ' '))
			*p-- = '\0';
		break;
	case TRIM_BOTH:
		for (p = str; *p != '\0'; p++);
		p--;
		while ((p > str) && (*p == ' '))
			*p-- = '\0';
	case TRIM_LEFT:
		for (p = str; *p == ' '; p++);
		Wstrcpy(str, p);
		break;
	}
}

/************************  builddt.c  *************************************
NAME
	Builddt - Convert a date structure to a date number

SYNOPSIS

unsigned long Explodedt(ssP,  datestruct )
DateStruct *datestruct;			/* returned filled in date structure

DESCRIPTION

This function writes all the components of the date structure.
They include year, month, day, hour, minute and second.

RETURN VALUE
  Date value - see datetime.h
*/

unsigned long 
Builddt(as_struct)
	DateStruct     *as_struct;
{

#define lc   (60/DTTIMEGRANULARITY)
#define lcyr (lc*(unsigned long)599040)	/* 60*24*32*13   */
#define lcmo (lc*46080)		/* 60*24*32      */
#define lcdy (lc*1440)		/* 60*24         */
#define lchr (lc*60)

	return ((long) as_struct->year * lcyr +
		(long) as_struct->month * lcmo +
		(long) as_struct->day * lcdy +
		(long) as_struct->hour * lchr +
		(long) as_struct->minute * lc +
		(long) as_struct->second/DTTIMEGRANULARITY);
}
