/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ssdate2.c
   Applix ASC version: 9
   Date: Thu Jan 21 21:10:22 1999
*/
#include <math.h>
#include <aos.h>
#include <datetime.h>
#include <global.h>
#include <gstyle.h>
#include <ss_style.h>
#include <axtime.h>
#include <trim.h>
#include <ssfunc.h>
#include <ssdate2.p>
/*===========================================================================*/
/* Time Format utilities													 */
/*===========================================================================*/

#define NUM_MAX_TOKENS	10
#define	MAX_TIME_TOKENS	6

/*
* Tokens return when parsing a string for time
*/

#define TOKEN_EOF	0	/* end of string */
#define TOKEN_INT	1	/* an integer token */
#define	TOKEN_DOUBLE	2	/* a double token */
#define TOKEN_SEP	3	/* token separator */
#define	TOKEN_AM	4	/* am string */
#define TOKEN_PM	5	/* pm string */
#define	TOKEN_UNKNOWN	-1	/* unknown token */

extern double   int_part();

/*
*	ss_parsetime - takes a string a parses it to see if it represents a
*					valid time format.
*	returns TRUE if successful parsing
*			FALSE if the string is not a time.
*
*	Time format.
*	<hour>:<minutes>:<seconds>[.tseconds] [AM|PM|am|pm]
*				where tseconds represents thousands of a second.
*
*	supported formats.
*	for eg:	the time 01:25:36 PM in each format will look like this.
*
*	format code					time string
*-------------------------------------------------
*	0					01:25:36 PM
*	1					01:25 PM
*	2					13:25:36 PM
*	3					13:25 PM
*	4					13:25:36.676 PM
*	5					13:25:36.676
*/

ss_parsetime(ssP, str, format, ts, listsep, decimalChar, amstr, pmstr)
	ss_value *ssP;
	char           *str;	/* the string to be parsed for time */
	short          *format;	/* OUT,	the format of the time */
	SS_TIME_STRUCT *ts;	/* the structure containing various
				 * components */
 /* time */
	char            listsep, decimalChar;
	char           *amstr, *pmstr;
{
	char           *p;
	int             token;
	int             i;
	double          tokval, tmp;
	int             pm = FALSE;
	int             gotampm = FALSE;
	int             numtoks = 0;
	char            local_str[STRPATH + 1];

	strncpy(local_str, str, STRPATH);
	local_str[STRPATH] = '\0';
	trim(ssP, local_str, TRIM_BOTH);
	/* Initialize the components */
	ts->hour = ts->min = ts->tsec = 0;
	ts->sec = -1;		/* a special flag, to find out if a seconds
				 * field was */
	/* specified at all or not */
	ts->tsec = -1;		/* a special flag to find out if tsecs was
				 * specified */
	p = local_str;
	for (i = 0; i < NUM_MAX_TOKENS; i++) {
		if (i > MAX_TIME_TOKENS)	/* more tokens than expected */
			return (FALSE);	/* invalid format */

		/* get next token */
		token = get_time_token(ssP, p, &tokval, &p, listsep, decimalChar,
				       amstr, pmstr);
		if (token == TOKEN_UNKNOWN)
			return (FALSE);	/* unregnizable token */

		if (token == TOKEN_EOF)
			break;	/* no more tokens */

		numtoks++;	/* count of tokens */

		/* find out what kind of token */
		switch (i) {
		case 0:
			if (token != TOKEN_INT)
				return (FALSE);
			ts->hour = (short) tokval;
			break;
		case 1:
			if (token != TOKEN_SEP)
				return (FALSE);
			break;
		case 2:
			if (token != TOKEN_INT)
				return (FALSE);
			ts->min = (short) tokval;
			break;
		case 3:
			if ((token == TOKEN_AM) || (token == TOKEN_PM) ||
			    (token == TOKEN_SEP)) {
				if (token == TOKEN_SEP)
					continue;
				gotampm = TRUE;
				if (token == TOKEN_PM)
					pm = TRUE;
				continue;
			}
			return (FALSE);
		case 4:
			if (gotampm || ((token != TOKEN_INT) && (token != TOKEN_DOUBLE)))
				return (FALSE);
			if (token == TOKEN_INT) {
				ts->sec = (short) tokval;
			} else {/* thousands of a second was specified */
				ts->sec = (short) int_part(tokval);
				tmp = tokval - ts->sec;
				ts->tsec = (short) int_part((tmp * 1000.0) + 0.5);
			}
			break;
		case 5:
			if ((token != TOKEN_PM) && (token != TOKEN_AM))
				return (FALSE);
			gotampm = TRUE;
			if (token == TOKEN_PM)
				pm = TRUE;
			break;
		}
	}

	if (i == 0) {
		return (FALSE);	/* null string or string will all spaces */
	}
	if ((ts->hour > 23 || ts->hour < 0) ||
	    (ts->min > 59 || ts->min < 0) ||
	    ((numtoks > 4) &&
	     (ts->sec > 60.0 || ts->sec < 0.0)))
		return (FALSE);

	if (gotampm) {		/* convert to 24 hours clock */
		if (ts->hour > 12)
			return (FALSE);
		if (ts->sec != -1) {
			if (ts->tsec != -1)
				*format = 4;	/* thousands of seconds
						 * SPECFIED */
			else {
				*format = 0;	/* thousands of seconds NOT
						 * specified */
				ts->tsec = 0;
			}
		} else {
			ts->sec = 0;
			ts->tsec = 0;
			*format = 1;
		}

		if (pm) {
			if (ts->hour != 12)
				ts->hour += 12;
		} else {
			if (ts->hour == 12)
				ts->hour = 0;
		}
	} else {		/* already in 24 hour clock */
		if (ts->sec != -1) {	/* seconds was specified */
			if (ts->tsec != -1)
				*format = 5;	/* thousands of sec was
						 * SPECIFIED */
			else {
				*format = 2;	/* thousands of sec was NOT
						 * specified */
				ts->tsec = 0;
			}
		} else {
			ts->sec = 0;
			ts->tsec = 0;
			*format = 3;
		}
	}

	return (TRUE);
}

get_time_token(ssP, str, tokval, p, listsep, decimalChar, amstr, pmstr)
	ss_value *ssP;
	char           *str, **p;
	double         *tokval;
	char            listsep;
	char            decimalChar;
	char           *amstr, *pmstr;
{
	char            buf[1000];	/* a lot more than what is needed */
	char           *cptr;

	*tokval = 0;
	if (*str == '\0')
		return (FALSE);

	/* skip spaces */
	while (isspace(*str))
		str++;

	if (*str == listsep) {
		*p = ++str;
		return (TOKEN_SEP);
	}
	if (isdigit(*str)) {
		int             gotdot = FALSE;
		cptr = buf;
		/* collect all digits */
		while (isdigit(*str))
			*cptr++ = *str++;
		if (*str == decimalChar) {
			gotdot = TRUE;
			/* collect the dot also */
			*cptr = '.';
			/* skip the dot */
			cptr++;
			str++;
			if (!isdigit(*str))
				return (TOKEN_UNKNOWN);
			while (isdigit(*str))
				*cptr++ = *str++;
		}
		*cptr = '\0';
		*p = str;

		if (gotdot) {
			*tokval = atof(buf);
			return (TOKEN_DOUBLE);
		} else {
			*tokval = atoi(buf);
			return (TOKEN_INT);
		}
	}
	if (isalpha(*str)) {	/* collect all alphabets */
		cptr = buf;
		while (isalpha(*str))
			*cptr++ = toupper(*str++);
		*cptr = '\0';
		*p = str;
		if (0 == strcmp(buf, amstr))
			return (TOKEN_AM);
		else if (0 == strcmp(buf, pmstr))
			return (TOKEN_PM);
	}
	return (TOKEN_UNKNOWN);
}

ss_format_time(ssP, ts, format, str, listsep, decimalChar, amstr, pmstr)
	ss_value *ssP;
	SS_TIME_STRUCT *ts;
	short           format;
	char           *str;
	char            listsep;
	char            decimalChar;
	char           *amstr, *pmstr;
{
	int             ispm = FALSE;
	char           *ampmstr;

	if (format == 0 || format == 1 || format == 4) {
		/* times with am pm */
		if (ts->hour > 11) {	/* a pm time */
			ispm = TRUE;
			/*
			 * adjust for pm times, since internally everything
			 * is
			 */
			/* stored in 24 hour clock */
			if (ts->hour != 12)
				ts->hour -= 12;
		}
		if (ispm)
			ampmstr = pmstr;
		else
			ampmstr = amstr;
		if (!ispm && ts->hour == 0)
			ts->hour = 12;

	}
	sprintf(str, "%02d%c%02d", ts->hour, listsep, ts->min);

	if (format == 1 || format == 3) {
		/* short formats, ie without the seconds display */
		if (format == 1) {
			strcat(str, " ");
			strcat(str, ampmstr);
		}
		return;
	}
	/* we need to display the seconds */
	sprintf(str, "%s%c%02d", str, listsep, ts->sec);


	if (format == 4 || format == 5) {
		/* the formats which display thousands of a second */
		sprintf(str, "%s%c%03d", str, decimalChar, ts->tsec);
	}
	if (format == 0 || format == 4) {
		/* am pm time */
		strcat(str, " ");
		strcat(str, ampmstr);
	}
}

ssGetDateStr(ssP, value, fmt, workstr, fullYear)
	ss_value *ssP;
	double          value;
	int             fmt;
	char           *workstr;
	int             fullYear;	/* TRUE if for displaying in editline */
{
	DateStruct      date;
	unsigned long   date_long;
	int             rc, len;
	int             retval = TRUE;
	char           *p;
	double          timeVal;
	int		lang;
	SS_TIME_STRUCT  ts;

	/* make sure you discard the fractional part */
	timeVal = value - floor(value);
	value = floor(value);

	if (value <= MAX_DATE) {
		date_long = value;
		MEMSET(&date, 0, sizeof(DateStruct));
		rc = DateExtract(date_long, &date.year, &date.month, &date.day);
		TimeExtract(timeVal, &ts);
		date.hour = ts.hour;
		date.minute = ts.min;
		date.second = ts.sec;
	} else {
		/* force the error condition here */
		rc = -1;
		retval = FALSE;
	}

	if (rc != 0)
		return (FALSE);	/* error condition */

	date.day++;
	date.year += 1900;
	date.month++;
	if (fullYear){
		lang = id_lang();
		Formatdt_lang_components(date.day, date.month, date.year, date.hour, date.minute, date.second, fmt, workstr, lang, fullYear);
	}
	else{
		lang = id_lang();
		Formatdt_lang_components(date.day, date.month, date.year, date.hour, date.minute, date.second, fmt, workstr, lang, fullYear);
	}
	trim(ssP, workstr, TRIM_BOTH);
	len = strlen(workstr);
	workstr[len] = ' ';
	workstr[++len] = '\0';
	return (TRUE);
}

/*
* 	ssGetTimeStr - get the time represented by the given number.
*/
ssGetTimeStr(ssP, value, fmt, workstr, listsep, decimalChar, amStr, pmStr)
	ss_value *ssP;
	double          value;
	int             fmt;
	char           *workstr;
	char            listsep, decimalChar;
	char           *amStr, *pmStr;
{
	SS_TIME_STRUCT  ts;
	int             rc, len;
	int             retval = TRUE;
	DateStruct      date;


	/* just take the fractional part */
	rc = TimeExtract(value, &ts);
	if (rc != 0) {
		workstr[0] = '\0';
		return (FALSE);
	}
	ss_format_time(ssP, &ts, fmt, workstr, listsep, decimalChar, amStr, pmStr);
	trim(ssP, workstr, TRIM_BOTH);
	len = strlen(workstr);
	workstr[len] = ' ';
	workstr[++len] = '\0';
	return (TRUE);
}
