/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: ssdefs.h
   Applix ASC version: 284.3.3
   Date: Mon Mar 15 21:12:52 1999
*/
#ifndef SSDEFS_H
#define SSDEFS_H

#include <brdrdefs.h>

#define SS_CURRENT_REV 4.0

#ifdef SMALL_DOC_DEMO /* { */
#define NUM_COLS     20
#define NUM_ROWS   	20
#else /* } { */
#define NUM_COLS     702
#define NUM_ROWS   	32767
#endif /* } */

/* Better memory usage for potencially long strings */
#define SS_STR_ALLOC(str) strcpy(TaskAlloc(ssP->pid, strlen(str)+1),str)
#define SS_STR_REALLOC(p,str) strcpy(TaskRealloc(ssP->pid, p,strlen(str)+1),str)


#define DRAW_AHEAD 1            /* how many lines displayer draws before
								 * haveinp check */

#define NORM_CALC 0
#define ROW_CALC 1
#define COL_CALC 2

/*
 * Drag States
 */
#define SS_NOT_DRAGGING                         0       /* normal state */
#define SS_DRAG_MOVE                            1       /* frame being moved */
#define SS_DRAG_RESIZE_TOP_LEFT         2       /* resize from top-left corner */
#define SS_DRAG_RESIZE_TOP_RIGHT        3       /* resize from top-right corner */
#define SS_DRAG_RESIZE_BOT_LEFT         4       /* resize from bottom-left
												 * corner */
#define SS_DRAG_RESIZE_BOT_RIGHT        5       /* resize from bottom-right
												 * corner */
#define SS_DRAG_RESIZE_TOP                      6       /* resize from top edge */
#define SS_DRAG_RESIZE_BOTTOM           7       /* resize from bottom edge */
#define SS_DRAG_RESIZE_LEFT                     8       /* resize from left edge */
#define SS_DRAG_RESIZE_RIGHT            9       /* resize from right edge */
#define SS_DRAG_NEW_FRAME                       10      /* user bounding box for
														 * new frame */
#define SS_DRAG_COPY                11  /* Object being copied */

/*
 * Cell States:
 * Note: All formula types must be less than CS_CONSTANT.
 * Type 0 is reserved by the object manager.
 */


/*
 * (1) 1 to 99 all object types other than cell types
 * (2) 100 to 199 all cell formula types
 * (3) From 200, all other cell types
 */

#define CS_FREE     999
#define CS_DELETE   998
#define CS_CLEAR    997

#define SS_TYPE_CELL            1       /* cell type */
/*
 * all other object types come after cell type, STRING must be first
 */
#define SS_TYPE_STRING  2
#define SS_TYPE_FORMULA 3
#define SS_TYPE_ARRAY   4
#define SS_TYPE_GRAPH   5
#define SS_TYPE_NAME    6
#define SS_TYPE_NMBAGGAGE       7       /* named range carried as baggage in
										 * xfer file */
#define SS_TYPE_VIEW    8       /* named views */

/*
 * GRAPH GROUP DEFINITIONS
 * In Pre-Rev3.0 GROUP# 6 was the Start of the Data Groups for BG's
 * In Rev3.0 Each Chart Defines Where the Data Groups Begin. (most likely #8)
 * Refer to chartParm indices in bg.h
 */
#define SS_MAIN_TITLE    0
#define SS_SUB_TITLE     1
#define SS_FOOTER        2
#define SS_LEGEND_TITLE  3
#define SS_LEGEND        4

#define SS_X_AXIS_TITLE  5
#define SS_X_AXIS_LABELS 6

#define SS_Y_AXIS_TITLE  7
#define SS_Y_AXIS_LABELS 8

#define SS_BEGIN_DATA    9
/*
 * special types that are used only in programs, e.g., SS_TYPE_PAGE_BRK
 *      is actually a SS_TYPE_ARRAY. have to be less than FMSK_TYPE
 */
#define SS_TYPE_LAST      SS_TYPE_VIEW
#define SS_TYPE_PAGE_BRK  (SS_TYPE_LAST + 1)
#define SS_TYPE_NMCREATE  (SS_TYPE_LAST + 2)    /* name creation for undo */
#define SS_TYPE_NMDELETE  (SS_TYPE_LAST + 3)    /* name deletion for undo */
#define SS_TYPE_NMCHANGE  (SS_TYPE_LAST + 4)    /* name change for undo */
#define SS_TYPE_NMSPECIAL (SS_TYPE_LAST + 5)    /* name chg during
												 * move/copy/ins/del */
#define SS_TYPE_COORD     (SS_TYPE_LAST + 6)    /* undoing a coordinate 
					* type,used for restoring open * cell */
#define SS_TYPE_ARRAY_FORM      (SS_TYPE_LAST + 7)      /* only for undo */
#define SS_TYPE_DB                      (SS_TYPE_LAST + 8)/* only for undo */
#define SS_TYPE_CHART_GROUP     (SS_TYPE_LAST + 9) /* only for undo */
#define SS_TYPE_ROWCOL_FORMAT 	(SS_TYPE_LAST+10) /* only for undo */
#define SS_TYPE_INSERT_DELETE 	(SS_TYPE_LAST+11) /* only for undo */


/*
 * Danger!! Danger!! Danger!!
 * All error types must be between CS_ERROR and CS_LASTERR. and CS_LASTERR
 *      must be updated when new ERROR types are added.
 * These are the types which exist as a field in the cell structure
 * Note that all the formula types should be less than CS_CONSTANT
 *
 *
 */
#define CS_NORMAL          100  /* numeric formula */
#define CS_TEXT_NORMAL 101      /* text formula */
#define CS_BOOLEAN         102  /* formula returning a   boolean value */
#define CS_INVALID         103  /* needs to recalced */
#define CS_NA              105  /* no applicable */
#define CS_PENDING         106  /* formula references error cells */
#define CS_CIRCULAR        107  /* formula referencing itself in someway */
#define CS_ERROR           104  /* formula produces error */
/* the following types upto CS_LASTERR are sub classes of CS_ERROR, */
/* they only differ from CS_ERROR for displaying better error statues , */
/* for other than displaying purposes, these are identical to CS_ERROR */
#define CS_DIVZERO         108  /* division by zero error */
#define CS_UDEFNAME        109  /* undefined named range specified */
#define CS_NUMERR          110  /* invalid argument for eg: sqrt(-1) */
#define CS_TYPERR          111  /* wrong type, eg: entering text where number */
#define CS_REFERR          112  /* reference to cells deleted etc */
#define CS_ARGERR          113  /* invalid #of arguments */
#define CS_FAILERR         114  /* function couldn't evaluate */
#define CS_DBERR1          115  /* database invalid */
#define CS_DBERR2          116  /* criteria invalid */
#define CS_DBERR3          117  /* type mismatch in data of a dbcol */
#define CS_LASTERR         117  /* type of last error */

/* !!! any new types added go before CS_NULL. CS_NULL and CS_MOVETAG */
/* must be pushed down */
#define CS_NULL            118  /* any other type of errors in future */
#define CS_MOVETAG         119  /* ephemermal - only during move command */

/*
 * left plenty of room between the formula types and constant types
 * to allow for future expansion
 */
#define CS_CONSTANT     200     /* Numeric constant */
#define CS_TEXT         201     /* Text cell */
#define CS_DOCUMENT     202     /* cell containing a document name */
#define CS_PICTURE      203     /* cell containing a picture */
#define CS_EMPTY        204     /* empty cell, which has no value but has style
								 * etc. */
#define CS_NA_VAL       205     /* just NA will be displayed in the cell */
#define CS_ERROR_VAL    206     /* ERROR will be displayed in the cell */
#define CS_GORCH        207     /* ephemermal - only during move command */


/*
 * Clipboard Status
 */
#define CLP_EMPTY 0
#define CLP_COPY 1
#define CLP_CUT 2


#define ALIGNMENT 16            /* total size of header */

#define MAX_PASTE 	(NUM_ROWS*NUM_COLS)	/* maximum cells that can be pasted at a time */
#define MAX_FILL  	(NUM_ROWS*NUM_COLS)	/* maximum cells that can be filled at a time */
#define MAX_RANGE 	(NUM_ROWS*NUM_COLS)	/* max number of cells in a range reference */

#define DFW          0xFF       /* use default columns width */
#define DEFHEIGHT    14         /* default height is 14 point */
#define MAX_OBJSIZE  10000
#define MAX_CELLSTR  1096       /* max len of label or formula string */
#define MAX_WINDOWS  16         /* 1st is active, only first 4 may be open. */
#define DESK_WIDTH   320        /* Worst case desktop in char units */
#define MARKLISTLEN  160        /* how many chars in a displayed mark list */
#define DBLSTRLEN    32         /* how many chars in a float str */
#define LEN_DOCNAME  80         /* how many chars allowed in document name */
#define LEN_UNIXPATH 255        /* how many chars in a unix file name */
#define MAXCOLWIDTH  254        /* maximum width of a column */
#define MINCOLWIDTH  1          /* minimum column width */
#define MAXROWHEIGHT 200        /* maximum row height in points */
#define MINROWHEIGHT 6          /* minimum row height in points */
#define MAXITRCOUNT      10     /* maximum iteration count for calculation */
#define MINITRCOUNT  1          /* minimum iteration count for calculation */

#define SS_HASH_PRIME 1009      /* for hashed stuff */

#define FIRST_PLATE 0        /* Plate "A" */
#define USER_PLATE 0         /* For now!!! */

#ifdef SMALL_DOC_DEMO /* { */
#define NUM_USER_PLATES 4
#define NUM_TOTAL_PLATES 8    /* total max number of plates */
#else /* } { */
#define NUM_USER_PLATES 256
#define NUM_TOTAL_PLATES 260    /* total max number of plates */
#endif /* } */

#define LAST_USER_PLATE (FIRST_PLATE+NUM_USER_PLATES-1)
#define LAST_TOTAL_PLATE (FIRST_PLATE+NUM_TOTAL_PLATES-1)

#define GRPH_PLATE (LAST_USER_PLATE+1)
#define RTIN_PLATE (LAST_USER_PLATE+2)  /* non-visible cells for Real-Time
										 * links */
#define RTOUT_PLATE (LAST_USER_PLATE+3) /* non-visible cells for Real-Time
										 * export links */

#define TEMP_PLATE	(LAST_USER_PLATE+4)	/* used for sorting */

#define ONE_PLATE(r) ((r)->tl_plate)    /* grepable cubic string */
#define IS_USERS_PLATES(plate) (((plate) >= FIRST_PLATE) && ((plate) <= LAST_USER_PLATE))

#define ROWS_TITLED 10          /* maximum number of title rows */
#define COLS_TITLED 10          /* maximum number of title columns */
#define COL_D_WIDTH 100         /* max chars in column descriptor line */
#define ROW_D_WIDTH 4           /* max chars in each row descriptor */
#define MAX_NESTING 8           /* max formula nesting (e.g. stack depth) */
#define NUM_MENUS 20            /* see menu.h */
#define MAX_NAMELEN STRNORMAL   /* max length of named range or view  */
 /* STRNORMAL in "aos.h"             */
 /* should agree with RANGELEN in "menu_gen.h" */
#define MAX_DOCLEN 255          /* max length of Document Names */
#define MAX_BKS_CONTAINED 1     /* max backptrs a cell can hold */

#define NUMXROWS DCL_PLATE_ARRAY
#define NUMXCOLS DCL_ROW_ARRAY


/*** defines pertaining to document postion ***/
#define DOC_POSITION_DAMAGED_MASK       0x80000000 /* the MSB bit of the
                                        array element is set if any rows/cols
                                        which it represents is damaged ****/

#define DOC_COL_BUNDLED                 20 /* this many col widths are bundled
                                                in one element of array */
#define DOC_ROW_BUNDLED                 72 /* this many rows heights are bundled
                                                 into one element of array */

/* the (TOTAL_DOC_COL_ARRAY+TOTAL_DOC_ROW_ARRAY)*sizeof(int) should be <= 2k */
#define TOTAL_DOC_COL_ARRAY             50 /* total element in the col array */
#define TOTAL_DOC_ROW_ARRAY             462/* total element in the row array */

	/***the following 2 defines are to be used as args to getBlocks***/
#define USA_STRINGS 0	/*dont translate*/
#define TRANS_STRINGS 1	/*translate*/
typedef unsigned long tm_text_descr;    /* should match TmTextDescriptor in */

typedef unsigned short hash_type;
typedef struct {
	int             x, y;       /* upper-left corner */
	int             wid, hyt;
}               SSRECT;

typedef uint16  bline_type;
typedef struct {
	short           num_pairs;
	int             fd[20];
}               fd_info;

typedef struct {
	int32          	row;
	short           col;
	short           plate;
}               coordinate;

typedef struct {
	int32           tl_row;
	int32           br_row;
	short           tl_col;
	short           tl_plate;
	short           br_col;
	short           br_plate;
}               range_desc;


typedef struct {
	int             num;
	range_desc     *ranges;
}               rng_list;

typedef struct {
	int32           start;
	int32           end;
}               view_pair;


typedef struct {
	unsigned int rows[TOTAL_DOC_ROW_ARRAY];
	unsigned int cols[TOTAL_DOC_COL_ARRAY];
}		doc_position; 

#ifndef BOOLEAN_DEFINED
typedef short   boolean;
#define BOOLEAN_DEFINED 1
#endif

#define SS_OBJ_NUM_PROPS 3
#define SS_OBJ_MOVE_RESIZE	0
#define SS_OBJ_MOVE_ONLY	1
#define SS_OBJ_NO_MOVE_RESIZE	2


typedef	struct	{
	char *name;
	SSRECT doc;
}grobj_docpos_type;

typedef struct	{
	int numObjects;
	grobj_docpos_type *objs;
} grobj_docpos_list;

typedef struct {
char *name;	/* name of the object */
SSRECT rect;	/* screen rectangle describing the object */
uint8	needsRedisplay;	/* needs to redisplay the object */
uint8	marked;	/* TRUE if object marks are visible */
}grobj_type;

typedef struct {
int numObjects;
grobj_type *objs;
} grobj_list;

/* the number of bits for ROWS_RESIZED */
/* 1 bit per row, for 10000 rows( make it 10k), the number of bytes is */
/* 1280, a nice figure divisible by 2,4,8,16,32 64 and so on */
#define ROWS_RESIZED    1280    /* number of bytes required for 10000 rows */

/*
*       !!!Note that the view_pair element should be the last in the view_value
*               struct since this structure is resized based on number of pairs
*               in the view
*/
typedef struct {
	uint16 rownum;
	uint16 height;
	boolean         user;       /* if TRUE, user set height, else system did */
}               height_pair;

typedef struct {
	coordinate      top_left;   /* when removed from window */
	coordinate	open_cell;
	unsigned char   width_default;      /* 4 - 254 */
	short           height_default;     /* 17 ?? */
	unsigned char   width[NUM_COLS];
	int             plate;
	ID              heightsList;
	uint32			titleWidth;
	uint32			titleHeight;
	uint16          rows_titled;/* how many 0-ROWS_TITLED */
	uint16          title_rows[ROWS_TITLED];
	uint16          cols_titled;
	uint16          title_cols[COLS_TITLED];
	ID              name;
	ID              page_brks;
	unsigned int    landscape:1,
					propagate_width:1,  /* recieves all col width changes */
					propagate_height:1, /* recieves all row height changes */
					propagate_page_break:1,     /* recieves all page break
												 * changes */
					is_hidden:1,	/* TRUE if not to show in the plates bar */
					is_temp_name:1,     /* for printing... */
					center_hallign:1,
					center_vallign:1;
	short           num_row_pairs;
	short           num_col_pairs;
	uint16          *rowPageBrks;/* list of row page breaks */
	uint16          *colPageBrks;/* list of col page breaks  */
	short           last_col;   /* last col whose width is different from
								 * default */
	uint16          numRowPageBrks;     /* total number of pagebreaks until
										 * last_row */
	uint16          numColPageBrks;     /* total number of pagebreaks until
										 * last_col */
	uint16          rowsPerPage;/* number of rows per page after last_row */
	uint16          colsPerPage;/* number of cols per page after last_col */

	doc_position	*docPos ;
	grobj_docpos_list	objList;		/* list of objects in the view */	
	unsigned	int row_doc_update_flag:1,
					col_doc_update_flag:1,
					is_dirty:1;	
	view_pair       pairs[2];
}               view_value;


#define VIEW_NAME_LEN 30

typedef struct {
	int32 tl_rownum;
	int32 br_rownum;
	short br_colnum;
	short tl_colnum;
	short tl_platenum;
	boolean tl_col_abs;
	boolean tl_row_abs;
	boolean tl_plt_abs;
	short br_platenum;
	boolean br_col_abs;
	boolean br_row_abs;
	boolean br_plt_abs;
}r_range_desc;	/* relative range */


typedef struct {
	int32 rownum;
	short colnum;
	short platenum;
	boolean abs_col;
	boolean abs_row;
	boolean abs_plate;
}r_cell; 	/* relative cell */

/**
 * Named Range Scope types (bit flags)
 *  if name_value.RefName is NULL, type is NAMES_LOCAL
 *  else type is NAMES_EXTERNAL (it's an ISR)
 **/
#define NAMES_LOCAL     1
#define NAMES_EXTERNAL  2
#define NAMES_ALL       (NAMES_LOCAL|NAMES_EXTERNAL)

typedef struct {
	ID              backptrs;   /* formula objects that use me */
	r_range_desc     oldRng;        /* save original, for move command */
	r_range_desc      rng;        /* the range token */
	long            doc_id;     	/* NOT USED */
	char           *Name;       /* varying extent */
	char           *NameThere;  /* Range name in other Reference SS */
	char           *RefName;    /* name of the Reference SS, NULL if not an 
								ISR */
	unsigned long   dtmod;      /* only menaingful if not current ss */
	unsigned long   HashId;     /* used to compared names with Name, much 
									faster */
	unsigned int    useLinkedSSAttrs:1, /* TRUE if use attrs from external ss */
					isRelative:1,	/* TRUE if relative named range */
					notInUse:1,	/* TRUE if not in use */
					changed: 1,	/* TRUE if named changed during move,transient*/
					foundLink:1; /* if the link is found when created else 
									abandon */
}               name_value;

typedef struct string_value {
	unsigned   int refcount; /*refcount must be first word forever and ever*/
	int     bytesize;   /** size of the obj in bytes ***/
	char    text[1];
}   string_value;

typedef struct id_array {
	ID              count;
	ID              ids[1];
}               id_array;

typedef struct formula_value {
	unsigned   int refcount; /* refcount must be first word forever and ever */
	uint16     isExternal; 	/* TRUE if this is a macro or an addin function */
	uint16     rtCount; 	/* position in xfer file */
	unsigned int  bytesize;   /** size of the obj in bytes ***/
	int			junk ;		/** added this junk for the byte allignment problem
							caused on the HP machine **/
	ID         formula[1];
}              formula_value;

typedef struct {
	boolean    short_special;      /* see array.c for meaning */
	boolean	   sorted;				/* to have a sorted array_value */
	int32      num_groups; /* highest group number written */
	int32     group_size; /* size of each group in chars */
	int32      avail_groups;       /* groups available after 'num_groups' */
	unsigned int 	changed0:1,
					changed1:1,
					changed2:1,
					changed3:1,
					junk:28;
	char       storage[1]; /* boundary */
}               array_value;

typedef struct {                /* "stack" for sensitivity scanner recursion */
	ID              cid;
	uint32           num_bkptrs;
}               scan_stack;

typedef struct {                /* "stack" for CalCell recursion */
	ID              cid;        /* cell being worked on */
	uint16           row;        /* its row pos */
	uint16           col;        /* its column pos */
	uint16           plate;      /* its plate pos */
	ID              cid2;       /* cell to recurse into */
	uint16           tp_offset;  /* to re-gen formula token ptr */
	uint16           stackbase;  /* for guaranteed clean stack on exit */
}               calc_stack;


typedef	struct	{
	range_desc rng;
	unsigned int 	refErr:1,	/* TRUE if this reference is an error */	
					nullRange:1;	/* not range associated with this series */
}graph_rng;

typedef	struct	{
	int numRngs;
	graph_rng *gr;
}series_value;

/* group info */
typedef struct {
	int    type;  	/* group type, data, titles, axes etc */
	int	   group_num;	/* the column in the graph row */
	char   *name;   /* data0, data1 for datagroups, axis x0, axis */
	int	delThisGroup;/*currently only used for pie charts for row/col ins/del */
					/*TRUE if som series in the group is err */
	int    numSeries;
	series_value *series;
} group_value;


typedef struct {
	int            numGroups;  /* number of groups in this graph */
	group_value    *groups;     /* array of all groups except axes */
}               gmap_type;

typedef struct {
	char           *type;
	char           *name;
}               ss_hook_type;

typedef struct {
	int             nHooks;     /* num of hooks */
	ss_hook_type   *hook;
}   hook_map_type;


/* main graph object struct */
typedef struct {
	boolean         invalid;    /* graph needs updated */
	int             pid;        /* Rev3.0 process id or 0 if not active */
	long            doc_id;     /* OBSOLETE in Rev3.0 document id */
	SSRECT          Loc;
	int             refErr;     /* TRUE if group ranges refs errored out
								 * because */
	/* of a move operation (insert/delete row/cols)  */
	char            Mgphtype[32];
	char           *name;       /* name of the chart */
	char           *path;       /* path of the graph */
	gmap_type       gmap;       /* array of groups */
	int				numSeries;	/* number of high level series */
	series_value	*svP;		/* array  of high level series */
}               graph_value;







typedef struct {
	range_desc      rng;        /* description of current range */
	ID              strFreeArray;       /* list of strings to be freed */
	data_ptr        arrP;       /* the elf array */
	int32           rowpos;     /* row position if doing range */
	int32           arrRowPos;/*the row pos of current item in the array */
	int32           numArrRows; /* number of rows in the array */
	short           control;    /* usually used as an arg index */
	short           range;      /* TRUE if processing a range spec */
	short           colpos;     /* col position if doing range */
	short           plate;      /* plate position if doing range */
	short       	skip_empty; /* TRUE if should skip empty items in ranges */
	short           base;       /* where in calcStack the start flag is */
	short           FuncCScode; /* CS_XXX returned by function */
	short           numArrCols; /* number of cols in the array */
	short           numArrPlates;    /* number of plates in the array */
	short           inArray;    /*TRUE if processing array */
	short       	arrPlatePos;/*the plate pos of current item in the array */
	short           arrColPos;/* the col pos of current item in the array */
	short           needsPerfectPlane;/* TRUE if needs a perfect 2d array */
	short           colDirection;  /* if true will get args column wise */
	unsigned		int getCellAddr:1; /* TRUE to get ARG_ADDR, else get its 
												cell contents */
}               arg_desc;

typedef struct {
	int             got_bad;    /* found the problem */
	ID              cid;        /* cell user requested help on */
	int             status;   /* reason why recalc failed, see calc_err.h */
	int             fcode;      /* builtin function that detected error */
	coordinate      cell;       /* cell where error began */
}               calc_error;

typedef struct {
	ID              gridID;     /* id of GridMap */
	ID              rowsID;     /* id of rows */
	ID              rowdID;     /* id of row_d_ok */
	ID              colsID;     /* id of cols */
	ID              widthsID;   /* id of widths */
	ID              cellsID;    /* id of cells */
	ID              markedID;   /* id of marked */
	ID              heightsID;  /* id of heights */
}               heap_stuff;


typedef struct {
XFontStruct *xf;	/* pointer to x-font */
int kickOutCounter;	/* cache key, used by atm to pull in a font */
}SS_FONT_ENTRY;

#define SS_BOLD_MIXED	0x3
#define SS_ITALIC_MIXED 0x3
#define SS_ULINE_MIXED 	0x7
#define	SS_PTSIZE_MIXED 0xFF
#define SS_FACE_MIXED 0xFF
typedef struct	{
	unsigned int faceDex:8,		/* index to type-faces */
					fsize:8,	/* point size */
					bold:2,		/* TRUE if bold */
					italic:2,	/* TRUE if italic */
					uline:3;	/* 0 no underline, 1 - single, 2 -double, */
								/* 3 - single word, 4 - double word */
}ss_font_pool_t;

typedef struct	{
	int16 faceDex;
	int8 bold;
	int8 italic;
	int8 uline;
	int8 fsize;
}ss_font_attr_t;

#define SS_STYLE_MIXED 0xFF
#define SS_PREC_MIXED 0xFF
typedef struct	{
	unsigned int 	type:8,		/* 0 - 254 number styles */
					prec:8,		/* 0 - 254 sub-types, or precision */
					custom:1;	/* 0 if builtin style, 1 if custom style */
}ss_style_pool_t;

#define SS_VALIGN_MIXED		0x7
#define SS_HALIGN_MIXED		0x7
#define SS_PROT_MIXED		0x3
#define SS_VISIB_MIXED		0x3
#define SS_COLOR_MIXED		0xFF
#define SS_WRAP_TEXT_MIXED	0x3

typedef struct {                
	uint16			fontDex;
	uint16			styleDex;	
	uint16          brdrDex;    /* dex into ssP->brdrPool */
	uint16 			defStyleDex;	/* global style dex */
	uint8           color;		/* foreground color for cell */
	unsigned int	valign:3,	/* vertical align, 0 for default, 1 for top */
								/* 2 for centered, 3 for bottom */
	halign:3,					/* horizontal align, 0 for default, 1 for left*/
								/* 2 for right, 3 for center, 4 for repeat */
	wrap_text:2,				/* 0 if no wrap text, 1 if wrap text */
	cas:2,						/* center across selection */
	justify:2,					/* justify */
	invisible:2, 				/* 0 if visible, 1 if invisible */
	protected:2,				/* 0 if not-protected, 1 if protected */
	useFont:1,					/* 0 to use default, 1 to use local */
	useStyle:1,					/* 0 to use default, 1 to use local */
	useBorder:1,				/* 0 to use default, 1 to use local */
	useAlign:1,					/* 0 to use default, 1 to use local */
	useProt:1,					/* 0 to use default, 1 to use local */
	useVisib:1,					/* 0 to use default, 1 to use local */
	useColor:1,					/* 0 to use default, 1 to use local */
	styleRecord:1;				/* 0 if this is cell record, 1 if */
									/* style record */
}              ss_attr_pool_t;

typedef struct ss_pool_item_t {
	struct ss_pool_item_t *next;/* thread */
	int             poolDex;    /* pos in ssP->attrPool ptr array */
	ss_attr_pool_t value;      /* the stuff! */
}               ss_pool_item_t;

typedef struct {                /* the actual attributes */
	border_type     borders;
	shade_type      shades;
}               ss_border_pool_t;

typedef struct ss_brdr_pool_item_t {
	struct ss_brdr_pool_item_t *next;   /* thread */
	int             poolDex;    /* pos in ssP->brdrPool ptr array */
	ss_border_pool_t value;     /* the stuff! */
}               ss_brdr_pool_item_t;

typedef struct ss_font_pool_item_t {
	struct ss_font_pool_item_t *next;   /* thread */
	int             poolDex;    /* pos in ssP->brdrPool ptr array */
	ss_font_pool_t value;     /* the stuff! */
}           ss_font_pool_item_t;

typedef struct ss_style_pool_item_t {
	struct ss_style_pool_item_t *next;   /* thread */
	int             poolDex;    /* pos in ssP->brdrPool ptr array */
	ss_style_pool_t value;     /* the stuff! */
}           ss_style_pool_item_t;


/* all the attributes for a cell */
typedef	struct	{
	ss_font_pool_t	f;
	ss_style_pool_t s;
	ss_border_pool_t b;
	uint8           color;		/* foreground color for cell */
	unsigned int	valign:3,	/* vertical align, 0 for default, 1 for top */
								/* 2 for centered, 3 for bottom */
	halign:3,					/* horizontal align, 0 for default, 1 for left*/
								/* 2 for right, 3 for center, 4 for repeat */
	wrap_text:2,				/* 0 if no wrap text, 1 if wrap text */
	cas:2,						/* center across selection */
	justify:2,					/* justify */
	invisible:2, 				/* 0 if visible, 1 if invisible */
	protected:2,				/* 0 if not-protected, 1 if protected */
	fontChanged:1,				/* TRUE if font changed*/
	bordersChanged:1,			/* TRUE if borders/shading changed */
	styleChanged:1,				/* TRUE if style changed */
	colorChanged:1,				/* TRUE if color changed */
	alignChanged:1,				/* TRUE if align changed */
	visibChanged:1,				/* TRUE if visibility changed */
	protChanged:1;				/* TRUE if protection changed */
}cell_attrs;





/*
*       Note: the definitions of cell_value, cell_text_value and
*                       cell_formula_value should match exactly
*
*/
typedef struct {
	double          value;      /* cell value */
	ID              id;         /* unused for constants */
	ID              backptrs[MAX_BKS_CONTAINED];/* number of bkptrs here */
	uint16          attrDex;    /* index into ssP->attrs table */
	uint16          array_index;/* index of the array formula */
	uint16          type;       /* cell type */
	unsigned int	rtRow:16,
					row:16,
					col:10,
					plate_idx:10,
					reComputeHashRow:1,	/* TRUE to recompute rt hash row */
					invalid:1,	/* TRUE if this cells need to be re-calced */
					bks_contained:1, 
					bkptrs_removed:1,  /* cell has ERR tokens and 
										backptrs are removed */
					needs_recalc:1, /* array formula, needs to be recalced */
					resolving:1,/* for CIRC control, true while calcing */
					changed0:1, /* flags for multiple windows redisplay */
					changed1:1, 
					changed2:1, 
					changed3:1,
					isRtODCell:1,		/* TRUE if this is a OD Cell */
					isRtInsertCell:1,	/* TRUE if this is a RTInsert Cell */
					timerUpdate:1,		/* TRUE if timer based calc in process*/
					timerInitiated:1,	/* TRUE if initiated by timer logic */
					obsListCount:2,		/* TRUE if cell on calc obs list */
					processed:1,		/* true if already processed, used in */
										/* finding dependents */
                    suppressDisplay:1;  /* to suppress initial display of
											formula on file load until the 
											first calc is done */
}               cell_value;

typedef struct {
	double          value;
	ID              text;       /* string object */
	ID              backptrs[MAX_BKS_CONTAINED];
	uint16          attrDex;    /* index into ssP->attrs table */
	uint16          array_index;/* index of the array formula */
	uint16          type;       /* cell type */
	unsigned int	rtRow:16,
					row:16,
					col:10,
					plate_idx:10,	
					reComputeHashRow:1,	/* TRUE to recompute rt hash row */
					invalid:1,	/* TRUE if this cells need to be re-calced */
					bks_contained:1, 
					bkptrs_removed:1,  /* cell has ERR tokens backptrs are 
										removed */
					needs_recalc:1,/* array formula, needs to be recalced */
					resolving:1,/* for CIRC control, true while calcing */
					changed0:1, /* flags for multiple windows redisplay */
					changed1:1, 
					changed2:1, 
					changed3:1,
					isRtODCell:1,		/* TRUE if this OD cell */
					isRtInsertCell:1,	/* TRUE if this is a RTInsert Cell */
					timerUpdate:1,		/* TRUE if timer based calc in process*/
					timerInitiated:1,	/* TRUE if initiated by timer logic */
					obsListCount:2,		/* TRUE if cell on calc obs list */
					processed:1,		/* true if already processed, used in */
										/* finding dependents */
                    suppressDisplay:1;  /* to suppress initial display of
											formula on file load until the 
											first calc is done */
}               	cell_text_value;

typedef struct {
	double          value;
	ID              formula;    /* formula object */
	ID              backptrs[MAX_BKS_CONTAINED];
	uint16          attrDex;    /* index into ssP->attrs table */
	uint16          array_index;/* index of the array formula */
	uint16          type;       /* cell type */
	unsigned int	rtRow:16,
					row:16,
					col:10,
					plate_idx:10,	
					reComputeHashRow:1,	/* TRUE to recompute rt hash row */
					invalid:1,	/* TRUE if this cells need to be re-calced */
					bks_contained:1, 
					bkptrs_removed:1,  /* cell has ERR tokens and backptrs 
										are  removed */
					needs_recalc:1,/* array formula, needs to be recalced */
					resolving:1,/* for CIRC control, true while calcing */
					changed0:1, /* flags for multiple windows redisplay */
					changed1:1, 
					changed2:1, 
					changed3:1,
					isRtODCell:1,		/* TRUE if this is a OD Cell */
					isRtInsertCell:1,	/* TRUE if this is a RTInsert Cell */
					timerUpdate:1,		/* TRUE if timer based calc in process*/
					timerInitiated:1,	/* TRUE if initiated by timer logic */
					obsListCount:2,		/* TRUE if cell on calc obs list */
					processed:1,		/* true if already processed, used in */
										/* finding dependents */
                    suppressDisplay:1;  /* to suppress initial display of
											formula on file load until the 
											first calc is done */
}               cell_formula_value;

typedef struct {
	int             col;
	int             row;
	int             plate;
	double          value;
	int             cell_type;
	char            display_str[2 * MAX_CELLSTR + 1];
	char            entry_str[2 * MAX_CELLSTR + 1];
	border_type		borders;
	shade_type		shades;
	uint8           fgcolor;
	uint8			style;
	uint8			prec;
	uint16			fsize;
	unsigned int	faceDex:8,
					size:8,
					halign:3,
					valign:3,
					wrap_text:3,
					cas:2,
					justify:2,
					protected:2,
					invisible:2,
					bold:2,
					italic:2,
					uline:2;
}               last_cell_value;

#define MAXATTRIBS  9           /* the number of attributes we are looking for */
typedef struct {
	border_type		borders;
	shade_type		shades;
	int             bold;
	int             italic;
	int             uline;
	int             faceDex;
	int             fsize;
	int           	style;
	int				prec;
	int             halign;
	int				valign;
	int				cas;
	int				justify;
	int             wrap_text;
	int             color;
	int             protected;
	int             invisible;
}               spl_cell_attrs;


typedef union {
	cell_value     *v;
	cell_text_value *t;
	cell_formula_value *f;
}               cell_ptr;

/* structures for the Undo file */
typedef struct {
	ID              type;       /* this has to be the first item */
	ID              id;
	coordinate      x;
	double          value;
	unsigned int    cellType:16;
	uint16          attrDex;
	uint16          array_index;
}               cell_info;

typedef struct {
	ID              type;
	coordinate      x;
}               empty_info;

typedef struct {
	ID              type;       /* this is not really an ID, but it should
								 * match size of ID */
	ID              id;
}               obj_info;

typedef struct {
	range_desc      arr_rng;    /* the output range for this array formula */
}               array_output;

typedef struct {
	int32           tag;
	int16			plate;
	double          value;
}               sort1_tuple;

typedef struct {
	int32           tag;
	int16			plate;
	double          value1;
	double          value2;
}               sort2_tuple;

typedef struct {
	int32           tag;
	int16			plate;
	double          value1;
	double          value2;
	double          value3;
}               sort3_tuple;

typedef union {
	sort1_tuple    *t1;
	sort2_tuple    *t2;
	sort3_tuple    *t3;
}               tuple_ptr;

/* MAY BE TRUE STATIC */
struct highlight_block {
	int             xpos;
	int             ypos;
	int             xsize;
	int             ysize;
};

typedef struct {
	int             retval;     /* = MENUSTAT_TOGGLE_OFF; */
	char            *rngStr;
	spl_cell_attrs	sattrs;
}               ss_cmd_status;

typedef struct {
	int             strokestate;
	boolean         entire_grid;
	boolean         hzscrollenabled;
	boolean         vtscrollenabled;
	int             xl, xr, yt, yb;
	int             hotx, hoty;
	int             clippedx, clippedy, clipped, waitState;
	int32           FirstColStroke, FirstRowStroke, FirstPlateStroke;
	int32           LastColStroke, LastRowStroke, LastPlateStroke, hotspot;
	int32           FirstPoint, LastPoint;
	boolean         in_autoscroll;
	boolean         firstpoint, lastpoint, stroking;
	short           key;
	boolean         ok_to_scroll;
	boolean         sync_scroll;/* TRUE if we have to syncronize our scrolls
								 * with */
	/* with the NoExpose/GraphicsExpose flag */
	XEvent          stroke_event;
	coordinate      open_cell;
	range_desc		markingRange;	/* the stroke range */
	int				markDex;		/* the index of stroke range in the */
									/* mark array */
}   ss_stroke_buf;

typedef struct {
	bline_type      left;
	bline_type      top;
	bline_type      right;
	bline_type      bot;
}               border_info;



#define MAX_MARK_RECTS  6
typedef struct {
	int             nrects;
	XRectangle      rects[MAX_MARK_RECTS];
}               mark_rect;


/* Object Types , !!! should match with elf_ss.elf defs */
#define SS_GRAPHIC_OBJ        	0x01
#define SS_CHART_OBJ          	0x02
#define SS_EQN_OBJ		  	  	0x04	
#define SS_AUDIO_INSET_OBJ    	0x08
#define SS_BUTTON_INSET_OBJ   	0x10
#define SS_SS_OBJ			  	0x20			
#define SS_WP_OBJ			  	0x40	
#define SS_DATA_OBJ			  	0x80
#define SS_HTML_OBJ			  	0x100
#define SS_BUILDER_OBJ    		0x200
#define SS_UNKNOWN_OBJ        	0x400
#define SS_MACRO_INSET_OBJ	  	0x800
#define SS_CELL_RANGE_OBJ     	0x1000
#define SS_ALL_OBJS    			0xFFFF

#define SS_GRAPH_NAMES  0x1     /* these match Graphic Object Types in
								 * elf_ss.elf */
#define SS_CHART_NAMES  0x2
#define SS_AUDIO_NAMES  0x8
#define SS_BUTTON_NAMES 0x10

#define SS_RANGE_NAMES  0x100
#define SS_VIEW_NAMES   0x200

/* both fields == 0 gets ALL external and internal References */
#define SS_EXT_NAMES    0x400   /* if set get ONLY external References */
#define SS_INT_NAMES    0x800   /* if set get ONLY internal References */

#define SS_ALL_NAMES (SS_RANGE_NAMES | SS_VIEW_NAMES | SS_CHART_OBJ | \
										  SS_GRAPHIC_OBJ | SS_BUTTON_INSET_OBJ | SS_AUDIO_INSET_OBJ)

#define SS_RNG_EDGE 3


/* Object Properties  (for safty, don't use MSB)*/
#define SS_OBJ_INDEPENDENT    0x00

#define SS_OBJ_FOLLOW_MOVE    0x01
#define SS_OBJ_FOLLOW_RESIZE  0x02
#define SS_OBJ_ANCHOR_TO_CELL 0x04
#define SS_OBJ_PRINT          0x08
#define SS_OBJ_LOCKED         0x10
#define SS_OBJ_HIDDEN         0x20

#define SS_OBJ_FOLLOW_ALL     (SS_OBJ_FOLLOW_MOVE | SS_OBJ_FOLLOW_RESIZE)
#define SS_OBJ_DEFAULT        (SS_OBJ_FOLLOW_ALL | SS_OBJ_PRINT)

#define SS_COL_MOVED 0
#define SS_ROW_MOVED 1

/** Button bitmap kinds **/
#define	BUTTON_DEFAULT		0		/** regular button **/
#define BUTTON_BITMAP_OBJ	1		/** button is a bitmap (*.im) file **/
#define	BUTTON_GRAPHIC_OBJ	2		/** button is a graphic image(*.ag) file */


/* the row/col window info */

typedef struct SSOBJECT {
	unsigned int    type;       /* Type of Object */
			
	long            x, y;       /* Absolute Doc Coordinate mils */
	long            wid, hyt;   /* Absolute Doc Coordinate mils */
	range_desc      rng;        /* Object Located in Range/Cell */
	double			insetScaleX, insetScaleY;
	double          xoff1, yoff1,	/* percentage offsets from top and */
					xoff2, yoff2;	/* bottom cells */
	char            name[STRPATH + 1];  /* objects name */
	char            path[STRPATH + 1];  /* path/FileName to External Object */
	char 			Title[STRPATH + 1];/*if the object is a button then This is
										is set to title of the object **/
	char            elfMacro[STRPATH + 1];    /* elf macro if button */
	char			*filterMacro;	/* macro to convert this doc */
	int8			borderStyle;	/* line style for inset border */
	long            dateStamp;  /* used for rev for objects, transient */
	/* than save time */
	SSRECT          Rt;         /* used when dragging */
	int				char_width, char_height; /* for button text */
	int             XStart, YStart;     /* Start of Drag location */
	int             insetTaskId;
	int             sizeData;
	int             srcx, srcy; /* position within the ge document */
	long            srcWid, srcHt;      /* size of the inset */
	long			dstWid_mil, dstHyt_mil;	/* display size of obj in mils*/

	/* position and size of object in print document mils ,ephemeral only */
	int             prtdocx, prtdocy, prtdocwid, prtdocht;
	int             damagedRectInUse;
	SSRECT          damagedRect;
	unsigned int    needsToFormat:1,
			property:4,		/* move/resize properties */
			locked:1,		/* TRUE if object is position locked */
			anchorToCell:1,	/* true if object is anchored to cell */
			hidden:1,		/* true if object is hidden */
			print:1,		/* true if object is printable */
			savedLocked:1,  /* save these temporarily */
			isButtonBitmap:2,/*if the object is a button. and the button
									has bitmap associated with it then it sets 
									this flag to 
									BUTTON_BITMAP_OBJ/BUTTON_GRAPHIC_OBJ 
									& the object->path would be set to the 
									filename of the bitmap*/
			objInCache:1,	/* true if to update chart cache */
			dialog_locked:1,/** TRUE->prevent the chart from updating
										when the dialog box is up ***/
			needsConversion:1, /* TRUE if needs conversion at save 
										time */
			isTitleObj:1,	/* True it is a title object else false **/
			noBorder:1, /* if TRUE, don't draw border around object */
			proportional:1, /* for clip mode scale */
			centered:1,		/* TRUE if inset material is centered */
			clipMode:4,		/* scale mode */	
			partial:1,		/* TRUE if object is not created yet */
			newObj:1,		/* TRUE if new object */
			valsHaveChanged:1,		/* TRUE if chart doesn't match vals in SS*/
			fileHasChanged:1,		/* need to re-save chart */
			needsRedraw:1,			/* Object is damaged, redraw */
			needsUpdate:1,			/* Object is updated, redraw */
			selected:1,   		/* Object is selected for Grouping/Marking */
			extLink:1,   	/* TRUE if link */
			iconOnly:1,		/* TRUE if inset is displayed as icon */
			loaded:1;		/* TRUE if loaded, FALSE if not loaded */
	ID				cacheGid;		/* the graph object thats in the cache */
	int				uid;			/* the objects uid */
	data_ptr		axInsetObjP;		/* inset source information */
	int				javaRevStamp;	/* incremented each time image produced */
	uint16				appType;	/*the app type of this object elf_apps.elf*/
	uint16				docType;	/*the doc type of this object doclinks.h*/
	char            data[1];    /* this should be at the end */
}               GPHOBJ;

#define CLIP_TO_FIT	0
#define SCALE_TO_FIT 1
#define SCALE_IF_CLIPS 2
#define CLIP_MODE_UNDEFINED	15U



typedef struct {
	Window          xwin;       /* The X-window handle */
	int             width, height,      /* the dimensions of the window */
					x, y;       /* the position they are placed in the SS
								 * window */
}               descr_win;

/*
*       the reason dbrngstr is used instead  of the range itself is that the
*       range can be a named range and it should be transparent to this database
*       entity as to what happens to the named range (shrunk or expanded).
*/
typedef struct {
	char            dbrngstr[MAX_NAMELEN + 1];  /* db range str */
	char            critrngstr[MAX_NAMELEN + 1];        /* critrange str */
	char            extrngstr[MAX_NAMELEN + 1]; /* ext range str */
	char            dbname[MAX_NAMELEN + 1];    /* name of this database entity
												 * itself */
	char            lastExtractRngStr[MAX_NAMELEN + 1]; /* the last extract
														 * range */
}               db_value;


/*
*       The table of font faces.
*               The table actually contains 256 entries with the zeroth entry for
*               default typeface.
*/
#define SS_MAX_FACES    255     /* maximum # of faces the table can hold */

typedef struct {
	char           *name;       /* the name of the fontFamily */
}               font_face_type;

typedef struct {
	int             numFaces;   /* number of faces in the table */
	font_face_type *faces;
}               fmap_type;

typedef	struct	{
	char *styleStr;
	void *styleTok;
}custom_style_type;

typedef	struct	{
	int numStyles;
	custom_style_type *cStyles;
} csmap_type;


#define SS_NUM_FONT_SIZES       8
#define DEFAULT_FACE    0
#define DEFAULT_FSIZE   14
#define SS_MAX_FONT_SIZE 254
#define SS_MAX_COLOR_VAL        255


#define SSFMFONT        255     /* we have 8 bits for typeface */
#define SSFPFONT        24      /* we have to shift 24 bits to get to typeface */

#define DATABASE        "Database"
#define CRITERIA        "Criteria"
#define EXTRACT         "Extract"


#define COL 0
#define ROW 1
#define RCFORMAT_INIT   -1
#define RCFORMAT_ATTR   0
#define RCFORMAT_STYLE  1
#define RCFORMAT_PTSIZE 2
#define RCFORMAT_PROT   3
#define RCFORMAT_INV    4
#define RCFORMAT_SHADE  5
#define RCFORMAT_BORDERS  6


/** struct for the row/col format. This now has only the two elements 
the row/col number and the attrbute index number into the attribute pool **/
typedef struct SS_RC_FORMAT {
	uint16 RowCol;  	/* 	column/row number (0-701)/(0-32766) */
	uint16 attrDex; 	/*  index into the attribute pool */
	int16	numCells;	/*  number of col formats that intersect with this row*/
	uint16 *cells;		/*  the attrs at the above intersections */
}               RCFRMT;

typedef	struct	{
	char *name;
	char *value;
}comment_type;

typedef struct {
	int             numComments;
	comment_type    *comments;
}               comment_table_type;

typedef struct {
	long            docId;
	char           *pathname;
	int16           docType;    /* per doctypes.h, -2 also used internally for
								 * errors  */
}               ss_doc_info_type;

typedef struct {
	char           *pathname;   /* name of the file */
	int16           docType;    /* recognize_file's doctype , -2 if couldn't
								 * access the file */
}               ss_doc_type;

typedef struct {
	int             numDocs;
	ss_doc_type    *docList;
}               ss_doc_list;

typedef struct {                /* used to return doc name info */
	char            name[STRPATH + 1];
	char            path[STRPATH + 1];
}               obj_value;

typedef struct todo_list_item_t {       /* used by the realtime logic */
	struct todo_list_item_t *next;      /* NULL or next in todo list */
	double          triggerTime;
	int             itemType;
	caddr_t         itemInfo;   /* an elf data_ptr casted out by caddr_t */
}               todo_list_item_t;

typedef struct rts_link_value {
	ID              text;       /* the linkages path as a string */
	int             rowNum;     /* row of rtin_plate this occupies */
}               rts_link_value;

#define PASTEALL    0
#define PASTEFORMS  1
#define PASTESTYLE  2
#define PASTEVALUE  3

#define SPARSE_VAL      0xFFFFFFFF      /* all bits set for a 32 bit quantity */
#define SSZOOMSIZE(x)   ( (int)( x * ((double)ssP->Obj->zoomFactor/100.0)) )



/* plate object, for now contains only graphic object lists and also the
row/col format information */

#define MAXCOLFORMATS	253
#define MAXROWFORMATS	253

typedef	struct	{
	RCFRMT *list;
	unsigned int last:16;	/* NUM_ROWS/NUM_COLS for grep */
	unsigned int num:15;	/* NUM_ROWS/NUM_COLS for grep */	
	unsigned int sorted:1;
}rowColEntry;

typedef	struct	{
	range_desc	rng;
	ID *bkptrs;	/* array containing the backpointers to formulas referencing*/
				/* this range */
	int numBkPtrs;	/* number of backpointers */
	unsigned int deleted:1;
} rng_entry_type;

typedef	struct	{
	int num;	/* number of entries in this block */
	rng_entry_type **list;
}rng_hash_block_type;

#define SS_RANGE_PRIME	253
typedef	struct	{
	int numRanges;
	int numDeleted;
	rng_hash_block_type **rngHashBlocks;
}rng_hash_type;

typedef struct	{
	char *name;
	rng_entry_type rngb;
}name_entry_type;

typedef	struct	{
	int num;
	name_entry_type **list;
} name_hash_block_type;

typedef struct	{
	int numNames;
	int numDeleted;
	name_hash_block_type **nameHashBlocks;
	unsigned int sorted:1;
	unsigned int numNamesDirty:31;	/* each time a name is changed this is */
									/* set */
} name_range_hash_type;

typedef struct	{
	char *name;		/* alias for the plate */
	char passwd[20];	/* alias for the plate */
	int numObjects;	/* total number of objects hanging off this plate */
	GPHOBJ **obj;	/* pointers to objects */
	int	attrDex;	/* non-zero if sheet format exists */
	rowColEntry *rowFormat[MAXROWFORMATS];
	rowColEntry *colFormat[MAXCOLFORMATS];
	uint16	lastRowFormat;
	uint16	lastColFormat;
	unsigned int numRowFormats;	/* total number of row formats in this plate */
	unsigned int numColFormats;	/* total number of col formats in this plate */
	rng_entry_type ***gridMap;
	uint32	numCells;		/* total number of cells in this plate */
	uint16  numGridCols;	/* number of cols in the grid map */
	uint16	numGridRows;	/* number of rows in the grid map */
	uint16  colBound;		/* column bound in this plate */
	uint16  rowBound;		/* row bound in this plate */
	uint16  dataColBound;	/* column bound in this plate containing data */
	uint16  dataRowBound;	/* row bound in this plate containing data */
	unsigned int gridIsDirty:1;	/* TRUE if this plate references deleted */
								/* ranges */
	unsigned int newBars:1;	/* if this plates row/col bounds need be calced */
	unsigned int protected:1;	/* if this sheet is protected or not */
	uint16  maxJVCols;      /* max cols on Java Client */
	uint16  maxJVRows;      /* max rows on Java Client */
}plate_value;

/* list of plates */
typedef struct	{
	int numPlates;
	plate_value *plates;
}plate_list_type;

extern GPHOBJ **ssGetObjList();

#define MAXRNGSTR	512	/* maximum size of buffers holding cell and rng strs */
#define MAXPLATENAME 32	/* maximum size for a plate name */


typedef struct {
	range_desc rng;	 /** to store the laste range used **/
	coordinate editcell; /** to store the cell cursor info in pointmode */
	char	str[MAXRNGSTR+1];	/* to hold a range or a string */
	int 	cur_pos; /** cursor position in point mode **/
	int		winIdx;	/* window in which pointmode is active */
	boolean marksused ; /** if we have used the shift range selection  **/
	boolean Mode; 
	boolean editline_clicked; /* this is set when the editline window is clicked					using the mouse **/
} pointMode ;

typedef	struct	{
	int plate;
	char *str;
}view_plate_str_type;

typedef struct	{
	coordinate open_cell;
	coordinate top_left;
	ID view;
	int x, y, width, height;
	unsigned int noGridAddr:1;
	unsigned int noPlates:1;
	unsigned int editline_mapped;
	unsigned int incell_edit_mapped;
}win_info_type;


/** common type of structure used by the printing routines and also the 
	print preview routines ***/

typedef struct {
	int		disp_x;				/*** stores the x value for the displaying **/
	int		disp_y;				/*** stores the y value for the displaying **/
	int		pagenum;			/*** stores the y value for the displaying **/
	int		extnum;				/*** stores the ext number of the page ***/
	int		seq_pagenum;		/*** stores the seq page num for the whole doc*/
	int		hoffset;			/** horrizontal offset ***/
	int		voffset;			/** vertical offset ***/
} 	pr_format_info ;


/* page description during printing for ss */
typedef struct {        /* Page id for paginated mode */
    long xpos;          /* from start of spreadsheet, in pixels */
    long ypos;          /* from start of spreadsheet, in pixels */
	int startcol, endcol;	/* what cols get printed on this page */ 
	int startrow, endrow;	/* what rows get printed on this page */
	int width, height;	/* dimensions of the page in pixels */
	uint16 pagenum;
	boolean is_lateral;	/* TRUE if this is a lateral page */
	uint16 cpairnum;
	uint16 rpairnum;
	int numVisibleRows;	/* visible rows in this page */
	int numVisibleCols;	/* visible cols in this page */
	pr_format_info *pg_fmt ;/*stores the fmt info for the print preview */
} ss_page_desc;


typedef struct {
	int numPages;	/* total num of pages including lateral pages */
	int lastPage;	/* 0 based last vertical page */
	ss_page_desc **pages;
}ss_page_info;


typedef struct {
	ss_page_info *SSPageArr;
	int		ss_rwidth ; /* same as SS_rwidth, copied into this during preview 
						creation **/
	int		ss_cheight ; /* same as SS_cheight, copied into this during preview 
						creation **/
	union page_breaks {
		long	(*arrayP)[];
		char	*array_charP;
	}	*pg_break;
} pr_page_info ;


typedef struct	{
	char *name;
	int	plate;
	ID viewId;
} view_info;

typedef struct	{
	view_info      *views;
	int			   numViews;
	int             color;
	short			device_setup;
	char			*print_class ;
	ss_font_pool_t  char_atts;  /* attributes from document settings */
	short           spacing;    /* amount of spacing between lines */
	int             startpage, endpage; /* starting and ending pages to be
										 * printed */
	int             ext1, extN; /* page extensions */
	int             landscape;  /* TRUE if landscape, FALE for potrait */
	char           *printer;    /* name of printer */
	char           *pf_file;
	int             copies;
	int             banner;
	int             hallign,    /* TRUE if material to be centered horizontally
								 * on page */
					vallign,    /* TRUE if material to be centered vertically
								 * on page */
					dsphdrs;    /* TRUE, if row/col headers are to be printed */
	char 			rngStr[MAX_CELLSTR+1];
	data_ptr		pi ;		/* ELF array of info for print formatter */
}ss_print_info;


/** defines used to find what is being dragged when & with whatdragged of 
pr_rowcol structure (in short used for the print preview  ***/
#define PR_LEFT_MARGIN		1
#define PR_RIGHT_MARGIN		2
#define PR_TOP_MARGIN		3
#define PR_BOT_MARGIN		4
#define PR_ROWS				5
#define PR_COLS				6

typedef struct {
	int		rowcol;
	int		value ;
} pr_draginfo  ;

typedef struct {
	int32 nRows;			/** num of Rows */
	short lMargin;			/** left Margin in mils for display */
	short rMargin;			/** right Margin in mils for display */
	short tMargin;			/** top Margin in mils for display */
	short bMargin;			/** bot Margin in mils for display */
	pr_draginfo *Rows;		/** list of rows on the current page **/
	short nCols;			/** num of Cols */
	pr_draginfo *Cols;		/** list of rows on the current page **/
	short rowDraged;		/** which row is being draged */
	short colDraged;		/** which col is being draged */
	unsigned int  isRowDragged:1,/** TRUE if row is being dragged else false */
		isColDragged:1,	/** TRUE if col is being dragged else false */
		isLeftDragged:1,	/** TRUE if leftmargin being dragged else false */
		isRightDragged:1,	/** TRUE if rightmargin being dragged else false */
		isTopDragged:1,	/** TRUE if topmargin being dragged else false */
		isBotDragged:1,	/** TRUE if botmargin being dragged else false */
		bPressed:1 ;	/** button pressed when moving a row or a cols */
} pr_rowcol ;

/** for storing print preview structure ***/
typedef struct {
	short 	Wid;		/*** Window id of the preview window ***/
	AXWidget	widgetId;
	Window	xwin ;
	int	  	Mode;	/*** TRUE -> on else FALSE->off ***/
	int		sync_flag ; /** TRUE if mode is TRUE and also if any module 
							shared by the preview and display is called by
							print preview code like shade_block, 
							drawImgText() etc ...**/
	int		scale_factor;
	int		zoomFlag;
	int		printDone; 	/*** set to true if print was called **/ 
						/***when print_preview was up ***/
	struct 	TmCrePacket *Cp;	/*** pointer to the packet information **/
	int		x;			/** display x and y location **/
	int		y;			/** display x and y location **/
	int		width;		/** display width of the window **/
	int		height;		/** display height of the window **/
	int		white_x_off;/** white area start at this x offset **/
	int		white_y_off;/** white area start at this y offset **/
	int		white_width;/** white area width **/
	int		white_height;/** white area height **/
	int		act_width ; /** act_width see comment for act_height **/
	int		act_height ; /** actual height which would be same as the
							ssP->Obj->ps.pageHeight when preview started since
							this ssP->Obj->ps.pageHeight can change after
							preview starts ***/
	int		printToFit ;	/** same reason as the act_height **/
	int	  	curPage;		/** current page being displayed **/
	int		curView ;	/** stores the currnet view number **/
	int		left_margin;	/** stores the current left margin ***/
	int 	right_margin;	/** stores the current right margin ***/
	int		top_margin ;	/** stores the current top margin ***/
	int		bottom_margin ;	/** stores the current bottom margin ***/
	pr_page_info *pg_info ;	/** stores the static values of the pages and 
								page brks **/
	ss_print_info	ssPrInfo; 	/** stores and updates the important parameters 
								used by the print preview functions***/
	SSRECT		clipRect ;	/** this is the clipping variable which is set 
				durning the true zoom mode ***/
	pr_rowcol	pr_viewMargin; /** structure for the row/col/margins dragging */
	boolean		pr_rng	;	/** TRUE if range is getting preview else FALSE **/
	boolean		showMargins; /** TRUE if margins are on else false useful 
				also for row and col dragging ***/
	boolean		printFormulas;	/* TRUE if printing formulas */
} pr_preview_info ;

#define SS_ZOOM_BUTTON_EVENT    1
#define SS_PLATES_PUSHED 2
typedef struct  {
	 int shifted;
	 char name[STRPATH+1];
 }ss_plates_msg_t;

 
typedef struct {
	int type;
	AXWidget widget;
	char data[1];
}ss_msg_t;

/* flag set if was processed during a dependent/precedent scan */
#define CELL_WAS_PROCESSED(x) ((xobjptr(cell_value,x))->processed)

/* macro to return the obsolete state of a cell */
#define		CELL_IS_OBS(x)	((xobjptr(cell_value,x))->invalid)
#define	MAKE_CELL_OBS(a1, a2)	{	\
	if((xobjptr(cell_value,a1))->invalid != a2)	{	\
		if(a2) ssP->numObsCells++; else ssP->numObsCells--;\
		(xobjptr(cell_value,a1))->invalid = a2; \
	}\
	if(a2)	ssP->needsRecalc = TRUE;\
}


#define		SET_TIMER_MODE(x, y) { (xobjptr(cell_value,x))->timerUpdate = y; }
#define		IS_TIMER_MODE(x, y) ((xobjptr(cell_value,x))->timerUpdate)


#define		SS_SET_VIEW_DIRTY_VP(x,y){ (x)->is_dirty = y;}
#define		SS_SET_VIEW_DIRTY_VID(x,y){ (xobjptr(view_value,x))->is_dirty = y;}
#define		SS_VIEW_DIRTY_VP(x) ((x)->is_dirty)
#define		SS_VIEW_DIRTY_VID(x) ((xobjptr(view_value,x))->is_dirty)



#define SS_CLICK_TYPE_NONE	0
#define SS_CLICK_TYPE_OBJ	1
#define SS_CLICK_TYPE_ROW	2
#define SS_CLICK_TYPE_COL	3
#define SS_CLICK_TYPE_CELL	4

typedef struct {
	int type;
	union {
		GPHOBJ *obj_ptr;
		int rowcol ;
		coordinate cell;
	} u;
} ss_click_type;

typedef	struct	calc_block_ {
	struct	calc_block_ *next;	
	ID cid;
}ss_calc_block;

typedef	struct	{
	ss_calc_block *head;	
	ss_calc_block *tail;
	int numInUse;	/* actual number of links in use */
	int numFree;	/* number of links on free list */
	ss_calc_block *freeHead;
}ss_calc_list;

/* this is used for storing calculation dexes for compatability */
/* with 4.1 Current Function */
typedef	struct	{
	int num;
	int dex;
	int *calcDexes;
}ss_cur_func_calc_dex;

/* used for spreadsheet calc memory allocation */
typedef	struct	{
	int num;
	void **ptrs;
}ss_mem_desc;



#define SS_SCREEN_ROW_BYTES	(((NUM_ROWS/8)+1) * sizeof(char))
#define SS_SCREEN_COL_BYTES	(((NUM_COLS/8)+1) * sizeof(char))

border_type *SS_GET_CELL_BORDERS_CP();
shade_type  *SS_GET_CELL_SHADES_CP();
border_type *SS_GET_CELL_BORDERS();
shade_type  *SS_GET_CELL_SHADES();
ss_font_pool_t *SS_GET_CELL_FONT_INFO_CP();
ss_font_pool_t *SS_GET_CELL_FONT_INFO();
ss_font_pool_t *SS_GET_DEF_FONT_INFO();
ss_style_pool_t *SS_GET_CELL_STYLE_INFO_CP();
ss_style_pool_t *SS_GET_CELL_STYLE_INFO();
ss_border_pool_t	*SS_GET_CELL_BORDERS_INFO_CP();


typedef struct {
	int				type;	/* cell type */
	int           	style;
	int				prec;
	int             align;
	TmTextDescriptor attrs;
	ss_font_pool_t  fattrs;
	int             wrap_text;
	int             color;
	int				valign;
	int				protected;
	int				invisible;
	int				fsize;
	border_type		borders;
	shade_type		shades;
} prs_cell_attrs;	/* returned by ParseStyle */

#define	SS_DEF_ATTR_DEX	1	/* where document settings are stored */

#define SS_FONT_NAME(x, y)	((x)->Obj->fmap.faces[(y)->faceDex].name)
#define SS_FONT_SIZE(x, y)	xxx

typedef	struct {
	AXWidget widgetId;
	SSRECT	lClip;
	SSRECT 	gClip;
	unsigned int 	noColor:1,	/* TRUE to ignore color */
					hasGlobalClip:1,	/* TRUE if has a page clip */
					hasLocalClip:1,	/* TRUE if has local clip */
					noBorder:1,	/* TRUE if not to draw border on page */
					objPrint:1;	/* TRUE to print a object */
	int psFactorX;
	int psFactorY;

	short           prtUnitHt;/* height of a character cell in target */
 										/* device pixels                      */
	short           prtUnitWid;	/* width of a character cell in target  */
 									/* device pixels                      */
short           prtColPad;	/* pixels to pad between columns on
					 						* target output device          */
short           prtRowHt;	/* pixel height of row including
					 * inter-row spacing on target device  */
short           blankWid;	/* pixel width of a blank on target device */
short           prtRowPad;	/* amount of space to pad between rows to allow */ 
							/* printing of grid */
short           prtCellPad;	/* space between grid line and cell text area
				 * on the left */
short           printDevice;	/* printer device setup number */
short           scnToPrtFactor;	/* factor to convert from display pixels to
				 * printer */

int             minLineWid_mil;	/* minimum line width in mils */

ID              viewId;	/* object id of view to work with */
view_value     *vp;		/* addr of view object */

long            prtWidth;	/* width of entire ss in target pixels */
long            prtHeight;	/* height of entire ss in target pixels */
long            prtTitleWid;	/* width of titles in target pixels */
long            prtTitleHt;	/* height of titles in target pixels */

long            objWidth;
long            objHeight;

long            rWidth;	/* width of row headers */
long            cHeight;	/* height of col headers */

long            widestCol;	/* width of widest col in target pixels */
long            tallestRow;	/* height of tallest row in targe pixels */

int             lMargin;	/* Clipping rect - left */
int             rMargin;/* Clipping rect - right */
int             tMargin;	/* Clipping rect - top */
int             bMargin;	/* Clipping rect - bottom */
int				rowBound;
int				colBound;
	ss_font_pool_t	defFont;		/* default font for printing */
	union page_breaks *pg_break;
	short    break_count, rowBreakCount, colBreakCount;
	short   *colNums;	/* ptr to array of cols to do */
	short   *widthCols;	/* ptr to array of col widths (as above) */
	short    numCols;	/* How many are in the cols array */
	short   *rowNums;	/* ptr to array of rows */
	short   *heightRows;	/* ptr to array of row height in pixels */
	short    numRows;	/* Number of rows to display on this page */
	char    *cellsP;		/* array to keep a note of bleedover cells */
	int hPos, vPos;	/* top left corner of the page where the grid */
	GPHOBJ *objP;	/* object to be printed */
}ss_print_var_info;


typedef	struct	{
	SSRECT	dstRect;
	SSRECT	clipRect;
	double zoomX;
	double zoomY;
	int clipMode;
	int proportional;
	int clear;
	int fill;
	int gray;
	int centered;
	int border;
	unsigned int 	useClipMode:1,
					useProportional:1,
					useClear:1,
					useFill:1,
					useGray:1,
					useCentered:1,
					useClip:1,
					useBorder:1,
					useZoomX:1,
					useZoomY:1;
}ss_inset_type;

typedef	struct	{
	char *bufP;			/* buffer to write data to */
	uint32 pos;			/* position in buf to write */
	int  bufSize;		/* size of allocated buffer */
}ss_file_buf_type;

typedef struct	{
	range_desc	rng;
	double xoff1, xoff2;
	double yoff1, yoff2;
}ss_obj_pos_type;

#define CELL_ROW(id)	(((cell_value *)ssP->OM.ptrs[id])->row)	
#define CELL_COL(id)	(((cell_value *)ssP->OM.ptrs[id])->col)	
#define CELL_PLATE(id)	((((cell_value *)ssP->OM.ptrs[id])->plate_idx) + FIRST_PLATE)

typedef	struct	{
	char *str;
	unsigned int inUse:1,
				 idx:16;	 
}ss_attr_dump_type;

typedef	struct		{
	char *str;		/* the actual format string from the file */
	uint16 dex;		/* attrDex */	
	uint16 type;	/* cell type */
} attrStrType;

/* struct used in breaking up ranges for border outline */
typedef	struct {
	int num;
	cell_attrs	*cattrs;
	range_desc *rng_arr;
}ss_range_attr_type;

typedef	struct	{
	uint16	num;
	attrStrType *arr;
}attrStrEntry;

typedef	struct	bks_block_{
	double 	val;
	ID		cid;
	struct bks_block_ *next;
	unsigned int type:16,
				invalid:1,
				 isText:1;
	uint16	attrDex;	
}bks_block;

typedef	struct	{
	char *bksPool;
	bks_block *listP;
	ID endCid;
}bks_list;

typedef struct {
	union {
		coordinate cell;
		range_desc rng;
	} u;
	unsigned int type:16; /* 1  for cell and 2 for rng */
	unsigned int onScreen:16; /* TRUE for cell onScreen else FALSE, for opt */
} coordRange;

/* reference and dependents List */
typedef struct {
	coordinate	cell; /* the cell for which the ref/dep is turned on*/
	coordRange	*cellList;/*keeps track of all the reference/dependents cells */
	int	numInCellList;
	unsigned int onScreen:1;	/*keeps track to see if this cell is on screen*/
	unsigned int referencesOn:1; /* references for this cell turned on */
	unsigned int dependentsOn:1; /* dependents for this cell turned on */
} refDepEntry;
	

/* for drag and drop information */
#define SS_NOTHING_DRAGGED 0
#define SS_LEFT_DRAGGED 1
#define SS_RIGHT_DRAGGED 2
#define SS_TOP_DRAGGED 3
#define SS_BOTTOM_DRAGGED 4

typedef struct {
	coordinate cell; /* where the current cell is located */
	range_desc rng; /* range where the data will be dropped for future display*/
	int leftOfCell;/* how many cells to the left of the cur cell selected*/
	int	rightOfCell; /* how many cell to the right of the current cell */
	int	topOfCell; /* how many cells to the top of the cell */
	int	bottomOfCell; /* how many cells to the bottom of the cell */
	int whichSide; /* stores information on what side is being dragged */
} dragDropSelectionInfo;

#define SS_SAVE_NOTHING 0 /* file does not need to be saved */
#define SS_SAVE_AND_BACKUP 1 /* save and autobackup */
#define SS_SAVE_ONLY	2 /* file just needs to be save when explicit call to 								save is done */


#ifndef SS_FILTER	/* { */
/* undefine these for the spreadsheet, since Taskbased allocs are used */
#undef malloc
#undef calloc
#undef realloc
#undef free

/* causes a build error if any of malloc,calloc,realloc or free used */
#define malloc	(XXX defined)
#define calloc	(XXX defined)
#define realloc	(XXX defined)
#define free	(XXX defined)
#endif	/* } */
#endif /* SSDEFS_H */
