/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: genincl.c
   Applix ASC version: 7
   Date: Mon Nov 16 13:57:09 1998
*/
#include <stdio.h>

#define NO_APX_INTERNATIONAL    /* turn off aos.h stuff */

#include <utloc.h>
#include <exitcode.h>

main(argc,argv)
int argc;
char *argv[];
{

    char buff[1024];
    char *p1, *p2;
    int category;
    int in_comment = 0;
    int value;
    char identifier[80];
    FILE *fd;
    int have_category = 0;

    if(argc < 3) {
        fprintf(stderr,"%soutput file not specified",
            (argc < 2) ? "input and " : "");
        exit(EXIT_FAILURE);
    }
    if(!freopen(argv[1],"r",stdin)) {
        fprintf(stderr,"not able to open input file <%s>",argv[1]);
        exit(EXIT_FAILURE);
    }
#if AXMS
    if(!(fd = fopen(argv[2],"wb")))  /* open in binary mode */
#else
    if(!(fd = fopen(argv[2],"w"))) 
#endif
        {
        fprintf(stderr,"not able to open output file <%s>",argv[2]);
        exit(EXIT_FAILURE);
        }

    while( fgets( buff, 1024, stdin ) != NULL )
       {
       if ( strlen( buff ) <= (unsigned)1 )
          {
          /* blank line, just output a blank line */  
          fprintf( fd, "\n" );
          continue;
          }

       if ( (buff[0] == '#') || (buff[0] <= ' ') )
          {
          /* a comment line */
          if ( !in_comment )
             {
             /* output a start-of-comment   */
             fprintf( fd, "/*\n" );
             in_comment = 1;
             }
          fprintf( fd, "%s", buff );
          continue;
          }

      if ( (buff[0] >= '0') && (buff[0] <= '9') && !have_category )
          {
          /* line starts with a digit, may be the "-CATEGORY-" line */
          p1 = buff;
          while ( *p1 != '-' )
             {
             if ( *p1 == '\n' )
                {
                /* hit newline, not "-CATEGORY-" line */
                goto regular_line;
                }
             p1++;
             }
          /* just to make sure */
          if ( strncmp( p1, "-CATEGORY-", 9 ) != 0 )
              {
              /* did not match "-CATEGORY-" */
              goto regular_line;
              }

          /* we have the category line, get the category number */
          p1 = buff;
          while ( (*p1 >= '0') && (*p1 <= '9') )
             {
             p1++;
             }
          *p1 = '\0';    /* step on char after category number */
          p1++;          /* start of the category name */
          p2 = p1 + strlen(p1) - 1;
          if ( *p2 == '\n' )
             *p2 = '\0';               /* step on the newline */
          category = atoi( buff );
          have_category = 1;
          if ( !in_comment )
             {
             /* output a start-of-comment   */
             fprintf( fd, "/*\n" );
             in_comment = 1;
             }
          fprintf( fd, "\n --------------- CATEGORY %d %s --------------\n",
                                                        category, p1);
          continue;
          }

regular_line:
      /* this should be a line composed of an identifier, a value, and
      *    a string.  We want to split the line up and generate a #define
      *    statement with the identifier, value, and the text string as
      *    comment                                                     */
      if ( in_comment )
         {
         /* output an end-of-commnet    */
         fprintf( fd, "*/\n" ); 
         in_comment = 0;
         }

      p1 = buff;
      while ( *p1 > ' ' )
         {
         /* find the end of the identifier string */
         p1++;
         }
      *p1 = '\0';
      strcpy( identifier, buff );

      while ( *p1 <= ' ' )
         {
         /* look for the start of the value  */
         p1++;
         }

      p2 = p1;           /* p1 points to start of the value  */
      while ( (*p2 >= '0') && (*p2 <= '9') )
         {
         /* look for the end of the value */
         p2++;
         }
      *p2 = '\0';
      value = atoi( p1 );

      while ( *p2 <= ' ' ) 
         { 
         /* look for the start of the text string */
         p2++;
         } 
      p1 = p2 + strlen(p2) - 1;
      if ( *p1 == '\n' )
         {
         *p1 = '\0';    /* step on the newline */
         }

      /* now output the line */
      value = category * ERRCOFFSET + value;
      fprintf( fd, "#define\t%s\t\t%u  /* %s */ \n",
                    identifier, value, p2 );
      }

/*  This will add an ending commnet to the outupt file if needed.
        Before a blank line at the end of the input file generate an
        open-ended commnet on the output file.     -lance       */


   if ( in_comment )
         {
         /* output an end-of-commnet    */
         fprintf( fd, "*/\n" );
         in_comment = 0;
         }




   fclose( fd );
   exit(0);
}

