/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: abspath.c
   Applix ASC version: 13
   Date: Mon Nov 16 13:53:26 1998
*/
#if SUN3 || SYSV4
#include <unistd.h>
#endif    /*  defines for ?_OK  used with access command  -lance  */
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#if AX_NT
#include <fcntl.h>
#define ECONNREFUSED    61  /* not defined in NT's errno.h, DAMNT */
#endif
#if PMAX
#include <Alib.h>
#endif
#include <aos.h>
#include <afile.h>
#include <getstr10.h>

extern int AxNoSymbolicFileLinks;
extern char AxCurrentDir[];
extern char *DirSlash;
extern char *TrueMalloc();

/*************************************************************************
function: ExpandLink - Expand a symblic link
*/
static int
expand_link(name)
char *name;
{
#if AXMS
    return(FALSE);
#else
#if X386_NEW
    return(FALSE);
#else /* all other X environs */
    char new_name[STRPATH+1], *p;
    int len, link_was_expanded;

    /* sanity check - are links allowed, and if so, is this file a link */
    if ( AxNoSymbolicFileLinks ) /* For HP CD/ROM */
        return(FALSE);
    link_was_expanded = FALSE;
    while ( (len = readlink(name, new_name, STRPATH)) != (-1) )
    {
        /* clip trailing /'s from link */
        new_name[len] = (char)0; /* null terminate the string */
        while ( new_name[(len = strlen(new_name))-1] == '/' )
            new_name[len-1] = (char)0;

        /* replace whole name if leading /, else replace last element only */
        if ( new_name[0] == '/' )
            strcpy(name, new_name);
        else
        {
            p = strrchr(name,'/');  /* get rid of old entryname */
            if ( !p )
                strcpy(name,new_name);  /* best guess */
            else
                strcpy(p+1,new_name);
        }
        if ( link_was_expanded == FALSE )
            link_was_expanded = TRUE;
    }
    return(link_was_expanded);
#endif
#endif  /* DAMNT */
}

/*************************************************************************
function: true_path - Convert soft link to full pathnames
*/
static void
true_path (dst,src,expand_last)
char *dst, *src;
int expand_last;    /* TRUE/FALSE */
{
    char inbuf[STRPATH+1], *dstbuf;
    int expanded, len;

    expanded = TRUE;
    dstbuf = dst;
    while(expanded)
    {
        /* init */
        expanded = FALSE;
        strcpy(inbuf, src);
        src = inbuf;

        /* make relative references absolute - WARNING - uses current dir */
#if AXMS
        AxSysDosUnixPath(src);
        if ( *src == *DirSlash )
        {
            dstbuf[0] = AxCurrentDir[0];
            dstbuf[1] = ':';
            dstbuf[2] = *DirSlash;
            dstbuf[3] = (char)0;
            src++;
        }
        else
        {
            len = strlen(src);
            if ( len < 2 )
                sprintf(dstbuf, "%s%s", AxCurrentDir, DirSlash);
            else
            {
                if ( src[1] != ':' )
                    sprintf(dstbuf, "%s%s", AxCurrentDir, DirSlash);
                else /* device prefix */
                {
                    dstbuf[0] = src[0];
                    dstbuf[1] = ':';
                    dstbuf[2] = *DirSlash;
                    dstbuf[3] = (char)0;
                    if ( len < 3 || src[2] != *DirSlash )
                        src += 2;
                    else
                        src += 3;
                }
            }
        }
	if ( islower(dstbuf[0]) )
		dstbuf[0] = toupper(dstbuf[0]);
#else
        if ( *src != *DirSlash )
            sprintf(dstbuf, "%s/", AxCurrentDir);
        else
        {
            strcpy(dstbuf, "/");
            src++;
        }
#endif
    
        /* trim off trailing slashes */
        while ( inbuf[(len = strlen(inbuf))-1] == *DirSlash )
        {
            /* avoid leading slashes */
            if ( len == 1 || (len == 3 && inbuf[1] == ':') )
                return;
            inbuf[len-1] = 0;
        }

        /* build up path */
        len = strlen(dstbuf);
        while ( *src )
        {
            if ( *src != *DirSlash )
                dst[len++] = *src++;
            else
            {
                dst[len] = 0; /* null term the partial path */
                if ( expand_link(dstbuf) )
                {
                    len = strlen(dstbuf);
                    expanded = TRUE;
                }
                dst[len++] = *DirSlash;
                src++;
            }
        }
        dst[len] = 0;

        /* final check for expansion */
        if ( expand_last )
        {
            if ( expand_link(dstbuf) )
                expanded = TRUE;
        }

        /* keep re-checking until it expands no more! */
        if ( expanded )
            src = dstbuf;
    }
    return;
}

/*************************************************************************/
/* function: MakeAbsolutePathname Fix relative and symbolic paths        */
/* Src/Dst allowed to be same address.                                   */
/*************************************************************************/
void
MakeAbsolutePathname(dst,src,expand_last)
char *dst;  /* returned fully absolute path */
char *src;
int expand_last;    /* TRUE if should expand entryname */
{
    int got_slash, dot_count;
    char *dstbegins, *buf;

    /* make sure we don't overwrite source when writing destination */
    if ( src != dst )
        buf = (char *)0;
    else
    {
        buf = TrueMalloc(STRPATH+1);
        strcpy(buf, src);
        src = buf;
    }

    /* expand relative paths and resolves links (WARNING - uses current dir) */
    true_path(dst, src, expand_last);

    /* remove directory dots and associated directories */
    got_slash = FALSE;
    dot_count = 0;
    for ( dstbegins=src=dst; *src; src++ )
    {
        if ( *src == *DirSlash )
        {
            if ( got_slash )
            {
                if ( dot_count == 0 )   /* remove "/" */
                    dst--;
                else if ( dot_count == 1 )  /* remove "/." */
                    dst -= 2;
                else if ( dot_count == 2 )  /* remove "/dir/.." */
                {
                    dst -= 4;
                    while ( (dst > dstbegins) && (*dst != *DirSlash) )
                        dst--;
                    if ( dst < dstbegins )  /* sanity */
                        dst = dstbegins;
                }
            }
            got_slash = TRUE;
            dot_count = 0;
        }
        else if ( got_slash && *src == '.' )
            dot_count++;
        else
        {
            got_slash = FALSE;
            dot_count = 0;
        }
        *dst++ = *src;
    }
    *dst = 0;

    /* make sure slashes occur singly (eg. remove // and replace with /) */

    /* done */
    if ( buf )
        TrueFree(buf);
    return;
}

/*************************************************************************
function: ParsePath - Split a pathname into dir & entry names
The pathname and the dirname portion are allowed to be the same address.
*/
void ParsePath (path,dir,entryname)
IN char *path;
OUT char *dir;
OUT char *entryname;
{
    register char *p;
    char temp_path[ STRPATH+1 ];  /* AxSysDosUnixPath may change input */

    strcpy( temp_path, path );
    AxSysDosUnixPath( temp_path );   /* normalize the slashes */

    p = strrchr(temp_path,*DirSlash);
    if ( !p ) {
        /* no slash, just a simple filename */
        strcpy(entryname,temp_path);
        *dir = 0;
    }
    else {
        strcpy(entryname,p+1);   /* copy the filename portion              */

        *p = 0;                  /* step on slash between dir and filename */
        if ( p == temp_path ) 
            {
            /* a file in the root directory */
            *dir = *DirSlash;
            *(dir+1) = 0;
            }
        else
            strcpy( dir, temp_path );   /* copy the directory portion */
    }

   return;
}
