/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: acstubs.c
   Applix ASC version: 52
   Date: Mon Jun 21 21:11:07 1999
*/
/*

Launch and manage tasks.

This file is called acstubs.c because it was originally derived
from the Alis Activity Controller (AC) process.  It should be renamed
to something meaningful...

*/

#include <setjmp.h>
#include <math.h>
#include <ctype.h>
#include <aos.h>
#include <fsrts.h>
#include <paripc.h>
#include <scac.h>
#include <getstr15.h>
#include <thimble.h>
#include <getstr4.h>
#include <getstr5.h>
#include <aprofpok.h>
#include <eldata.h>
#include <axtrace.h>
#include <acstubs.p>
#include <arghdr.h>

/* Note: All mallocs here belong to the Task, not the system */

char *ACoptarg;     /* getopt's optarg */
int ACoptind;       /* getopt's optind */
int ACopterr;
int ACtimeout;
DocInfoType XferDoc;	/* transfer bin document wrapper */

extern char ThimTempDir[];
extern short THIMpid;
extern int AxInTask;	/* 0 or task that owns the current stack */

void AxSyncAgainstProfile();

/*************************************************************************
function: ACgetEvent - Decode an APAC event
*/
ACgetEvent (typeP,dstP)
int *typeP;		/* as listed in scac.h */
AC_IND *dstP;
{
	static scac_message_info msg;
	int code;
	short junk;

    code = IPCevMes(&msg,sizeof(scac_message_info),&junk,&junk,&junk);
	if ( code != IPC_OK ) taskAbort(NXLT("IPC message get failed"));
	*typeP = msg.type;
	switch(msg.type)	{
	case AC_OPEN:
	case AC_DOCMOD:
	case AC_TERM:
	case AC_RTSTARTED:
	case AC_NEWFONTS:
		dstP->docid = msg.buf;
		break;
	default:
		taskAbort(XLT("Garbled AC IPC msg"));
	}
	return(AOK);
}

/*************************************************************************
function: ACpproc - Start a task, pend till it returns
*/
ACpproc (appid,commandline)
int appid;		/* APDC.. */
char *commandline;	/* argv as a single string */
{
	int pid;
	int exitcode;

	exitcode = ACprocIt(appid,commandline,&pid,TRUE,FALSE,(char*)NULL,FALSE);
	if ( pid == -1 )
		return(127);
	else
		return(exitcode);
}

/*************************************************************************
function: ACpprocIt - Pended proc of a routine, setting argv0
*/
ACpprocIt (appid,commandline,name)
int appid;		/* APDC.. */
char *commandline;	/* argv as a single string */
char *name;		/* argv0 */
{
	int pid;
	int exitcode;

	exitcode = ACprocIt(appid,commandline,&pid,TRUE,FALSE,name,FALSE);
	if ( pid == -1 )
		return(127);
	else
		return(exitcode);
}

/*************************************************************************
function: ACproc - Start a process compatable to real ACproc
*/
ACproc (appid,commandline,pidP)
int appid;		/* APDC.. */
char *commandline;	/* argv as a single string */
int *pidP;		/* returned pid > 0 if success */
{
	return(ACprocIt(appid,commandline,pidP,FALSE,FALSE,(char*)NULL,FALSE));
}

/*************************************************************************
function: ACprocWithError - Proc task, throw error if task died before ACImOK
*/
int ACprocWithError (appid, cmdline, pidP)
int appid;
char *cmdline;
int *pidP;
{
    int             code;

	code = ACprocIt(appid,cmdline,pidP,FALSE,FALSE,(char*)NULL,TRUE);
    if (*pidP == -1) {
        if (code < 100)
            ElfStrAbort(ERR_ILLVAL, NXLT("Could not start task."), NULL);
        else
            ElfAbort(code);
    }
	return(code);
}

extern char *ElfSymbolNameFromId();
extern int ElfLanguage;

/*************************************************************************
function: ElfTaskProc - Start an elf process, can be called by any app
*/
ElfTaskProc (n)
int n;
{
	char *func = ElfSymbolNameFromId(n,ElfLanguage);
	int pid;
	return(ACprocIt(APELFM,"",&pid,FALSE,FALSE,func,FALSE));
}

/*************************************************************************
function: ElfTaskProcArgs - Start another Elf task
*/
ElfTaskProcArgs (n,args)
int n;
char *args;
{
	int pid;
	char *func_name = ElfSymbolNameFromId(n,ElfLanguage);

	ACprocIt(APELFM,args,&pid,FALSE,TRUE,func_name,FALSE);
	return(pid);
}

/*************************************************************************
function: ElfTaskPProc - Start and pend on an ELF process
*/
ElfTaskPProc (n)
int n;
{
	char *func = ElfSymbolNameFromId(n,ElfLanguage);
	return(ACpprocIt(APELFM,"",func));
}

/* ?? all of the proc failure cases should generate a message here! */

ACprocIt (appid,commandline,pidP,pended,quick_return,name,suppressErrBox)
int appid;		/* APDC.. */
char *commandline;	/* argv as a single string */
int *pidP;		/* returned pid > 0 if success */
int pended;		/* If TRUE, should pend the caller */
int quick_return;	/* IGNORED - OBSOLETE */
char *name;		/* NULL or override value for argv[0] */
int suppressErrBox;	/* see comment in thimble.h */
{
	register int i;
	register char *c,*ccmd;
	int len;
#define MAX_ARGC 10000
	char *argv[MAX_ARGC+1];
	extern char *strtok();
    int pid;
    extern short THIMpid;
    int mypid, argc;
	char **argv2, **ArgvAlloc();
	int (*func)();
	int proctype;
    typeErr code, THIMstart();
	int MyActivationData = THIMpid;

	AppidToFunc(appid,&func,&argv[0],&proctype);
	if ( name )
		argv[0] = name;
	if ( func == NULL ) {
		char buf[100];
		sprintf(buf,NXLT("!Sorry, this proc not implemented yet (id %d)"),appid);
		SystemMessage(buf);
		sleep(2);
		SystemMessage("");
		return(*pidP = -1);
	}

/* re-package command line as arg vector */

	i = strlen(commandline) + 1;
	if ((c = ccmd = malloc(i)) == NULL)
		return(*pidP = -1);
	memcpy(ccmd,commandline,i);
	for (argc = 1; argc < MAX_ARGC && (argv[argc] = strtok(c," ")); argc++) {
		if ( *argv[argc] != '\"' ) { 
			c = NULL; 
			continue; 
		}    
		argv[argc]++; 

		/* The following fix was copied from shcmd() in thim.c */
		/* if the last char is a quote, we're almost done */
		len = strlen(argv[argc]);
		if (argv[argc][len-1] == '\"') {
			argv[argc][len-1] = 0;
			c = NULL;
			continue;
		}

		c = argv[argc] + strlen(argv[argc]);
		*c++ = ' '; 
		c = argv[argc];
		c = strtok(c,"\"");
		c = NULL;
	}
	argv[argc] = NULL;

/* Create the THIMBLE task */

	argv2 = ArgvAlloc(argc,argv);
	free(ccmd);
	if ( argv2 == NULL )
		return(*pidP = -1);

    mypid = AxInTask;
    code = THIMstart(appid,&pid);	/* create task's BSS & DATA */
	if ( code != 0 )
		return(*pidP = -1);
	TaskInfo[pid].func = func;
	TaskInfo[pid].argc = argc;
	TaskInfo[pid].argv = argv2;
	TaskInfo[pid].start_me = TRUE;
	TaskInfo[pid].child_retcode = 1;	/* assume generic failure */
	if ( !AxInTask )		/* kernel startup */
		return(*pidP = pid);		/* total success */
    VfyContext(MyActivationData);
	TaskInfo[mypid].pended = pid;	/* ?? if StartTask fails, reset ths */
	TaskInfo[pid].parentTask = ThimUIDPid(mypid);
	TaskInfo[pid].suppressExitErrorBox = suppressErrBox;
	TaskInfo[pid].TaskUsesJavaDisplay = TaskInfo[mypid].TaskUsesJavaDisplay;
	TaskInfo[pid].javaChannel = TaskInfo[mypid].javaChannel;
	TaskInfo[pid].javaWinId = TaskInfo[mypid].javaWinId;
	TaskInfo[pid].alarm_mode = TaskInfo[mypid].alarm_mode;
    TaskInfo[pid].alarm_millisec = TaskInfo[mypid].alarm_millisec;
	TaskInfo[pid].alarm_has_fired = TRUE;  /* enable time slicing now */
	TaskInfo[mypid].ok_pended = !pended;
	if ( mypid != 0 )
		SuspendTask(0);	/* to get back to ATM main loop for child to start */
	if ( pended ) {
		*pidP = 0;
		return(TaskInfo[mypid].child_retcode);
	}
	else if ( TaskInfo[pid].in_use )	/* child up & running, return pid */
		return(*pidP = pid);
	else {
		if ( TaskInfo[mypid].child_retcode == 0 )
			*pidP = 0;	/* came up and completed */
		else
			*pidP = -1;
		return(TaskInfo[mypid].child_retcode);
	}
}

/*************************************************************************
function: AxfPendForNewTaskWithArgs - Start another Elf task and wait for it to complete
*/
data_ptr AxfPendForNewTaskWithArgs (data_ptr argDatum)
{
	int args_passed;
	int pid;
	char *name;
	data_ptr elfArgv,aP;
	int mypid;
	typeErr code;
	int (*func)();
	int proctype;
	char *args[2];
	int ix;
	int MyActivationData = THIMpid;

    name = StrFromArgArray(argDatum,0);
	if ( !name[0] || (name[0] == ' ') )
		ElfAbort(ERR_ILLVAL);

	AppidToFunc(APELFM,&func,&args[0],&proctype);
	args[0] = name;

	/* Allocate args in task 0 */
	args_passed = ElfbArraySize(argDatum)-1;
	elfArgv = AxTaskCreateElfArray(0,args_passed);
	for ( ix = 0; ix < args_passed; ix++ )
		AxAddArrayToArray(elfArgv,ix,
				AxTaskCopyElfData(0,AxArrayElement(argDatum,ix+1)));

    mypid = AxInTask;
	if ( !mypid )
		tempAbort(NXLT("ElfbPendForNewTaskWithArgs called from kernel"));

    code = THIMstart(APELFM,&pid);	/* create task's BSS & DATA */
	if ( code != 0 )
		ElfAbort(ERR_ILLVAL);		/* !!1 ?? better code? */

	TaskInfo[pid].func = func;
	TaskInfo[pid].argc = 1;
	TaskInfo[pid].argv = ArgvAlloc(1,args);
	TaskInfo[pid].elfArgv = elfArgv;
	TaskInfo[pid].usesElfArgs = TRUE;
	TaskInfo[pid].start_me = TRUE;
	TaskInfo[pid].child_retcode = 1;	/* assume generic failure */
    VfyContext(MyActivationData);
	TaskInfo[mypid].pended = pid;	/* ?? if StartTask fails, reset ths */
	TaskInfo[pid].parentTask = ThimUIDPid(mypid);
	TaskInfo[pid].TaskUsesJavaDisplay = TaskInfo[mypid].TaskUsesJavaDisplay;
	TaskInfo[pid].javaChannel = TaskInfo[mypid].javaChannel;
	TaskInfo[pid].javaWinId = TaskInfo[mypid].javaWinId;
	TaskInfo[pid].alarm_mode = TaskInfo[mypid].alarm_mode;
    TaskInfo[pid].alarm_millisec = TaskInfo[mypid].alarm_millisec;
	TaskInfo[pid].alarm_has_fired = TRUE;  /* enable time slicing now */
	TaskInfo[mypid].ok_pended = FALSE;
	SuspendTask(0);	/* to get back to ATM main loop for child to start */

/* when it gets here, the child has been completed */

	if ( TaskInfo[mypid].child_retcode == 0 )
		return(ElfDataCopy(TaskInfo[mypid].child_retdata));
	else {
		char *str, *obj;
		int code;
		code = TaskInfo[mypid].child_retcode;
		str = AxStrPtrFromArray(TaskInfo[mypid].child_retdata,0);
		obj = AxStrPtrFromArray(TaskInfo[mypid].child_retdata,1);
		ElfStrAbort(code,str,obj);
		return(NULL);	/* just to prevent compiler warning */
	}
}

/*************************************************************************
function: AxfNewTaskWithArgs - Start another Elf task
*/
data_ptr AxfNewTaskWithArgs (data_ptr argDatum)
{
	return(AxfDoNewTaskWithArgs(argDatum,FALSE,TRUE));
}
/*************************************************************************
function: AxfUnpendedNewTaskWithArgs - Add a new Elf task to Active tasks 
          and continue (do not pend) current task
*/
data_ptr AxfUnpendedNewTaskWithArgs (data_ptr argDatum)
{
	return(AxfDoNewTaskWithArgs(argDatum,FALSE,FALSE));
}
/*************************************************************************
function: AxfNewDeamonWithArgs - Start another Elf task
*/
data_ptr AxfNewDeamonWithArgs (data_ptr argDatum)
{
	return(AxfDoNewTaskWithArgs(argDatum,TRUE,TRUE));
}
data_ptr AxfDoNewTaskWithArgs (argDatum,is_deamon, suspend_flag)
data_ptr argDatum;
int suspend_flag;
{
	int args_passed;
	int pid;
	char *name;
	data_ptr elfArgv;
	int mypid;
	typeErr code;
	int (*func)();
	int proctype;
	char *args[2];
	int ix;
	int MyActivationData = THIMpid;

    name = StrFromArgArray(argDatum,0);
	if ( !name[0] || (name[0] == ' ') )
		ElfAbort(ERR_ILLVAL);

	AppidToFunc(APELFM,&func,&args[0],&proctype);
	args[0] = name;

	/* Allocate args in task 0 */
	args_passed = ElfbArraySize(argDatum)-1;
	elfArgv = AxTaskCreateElfArray(0,args_passed);
	for ( ix = 0; ix < args_passed; ix++ )
		AxAddArrayToArray(elfArgv,ix,
				AxTaskCopyElfData(0,AxArrayElement(argDatum,ix+1)));

    mypid = AxInTask;
	if ( !mypid )
		tempAbort(NXLT("ElfbNewTaskWithArgs called from kernel"));

    code = THIMstart(APELFM,&pid);	/* create task's BSS & DATA */
	if ( code != 0 )
		ElfAbort(ERR_ILLVAL);		/* !!1 ?? better code? */

	TaskInfo[pid].func = func;
	TaskInfo[pid].deamon = is_deamon;
	TaskInfo[pid].argc = 1;
	TaskInfo[pid].argv = ArgvAlloc(1,args);
	TaskInfo[pid].elfArgv = elfArgv;
	TaskInfo[pid].usesElfArgs = TRUE;
	TaskInfo[pid].start_me = TRUE;
	TaskInfo[pid].child_retcode = 1;	/* assume generic failure */
	TaskInfo[pid].parentTask = ThimUIDPid(mypid);
	TaskInfo[pid].TaskUsesJavaDisplay = TaskInfo[mypid].TaskUsesJavaDisplay;
	TaskInfo[pid].javaChannel = TaskInfo[mypid].javaChannel;
	TaskInfo[pid].javaWinId = TaskInfo[mypid].javaWinId;
	TaskInfo[pid].alarm_mode = TaskInfo[mypid].alarm_mode;
    TaskInfo[pid].alarm_millisec = TaskInfo[mypid].alarm_millisec;
	TaskInfo[pid].alarm_has_fired = TRUE;  /* enable time slicing now */
	if(suspend_flag) {
		VfyContext(MyActivationData);
		TaskInfo[mypid].pended = pid;
		TaskInfo[mypid].ok_pended = TRUE;
		SuspendTask(0);	/* to get back to ATM main loop for child to start */
		if ( TaskInfo[pid].in_use )	/* child up & running, return pid */
			return(MK_INTEGER(TaskInfo[pid].uid));
		else
			return(MK_INTEGER(0));	/* came up and completed */
	}
	else {
		/* continue current Task, child will start later
		   when the parent task suspend's itself  */
		VfyContext(MyActivationData);
		TaskInfo[pid].child_retcode = 0;
		TaskInfo[pid].pended = 0;
		TaskInfo[pid].ok_pended = FALSE;
		/* Find my task master now!. When I start if my parent is not alive,
		   you can never find my taskMaster */
		AxFindTaskMaster(pid); 
		return(MK_INTEGER(TaskInfo[pid].uid));
	}
}


/*************************************************************************
function: ACImOk - Unpend caller of ACproc
*/
ACImOk ()
{
	return(ACDoImOk(AxInTask));
}

ACDoImOk (task)
int task;
{
	context_block *p;
	int parent = 0;

	p = &TaskInfo[task];
	if ( p->have_okayed )
		return(0);
	if ( p->defer_acimok )
		return(0);
	p->have_okayed = TRUE;

	/* re-activate parent if it is of ok_pended (ACproc) form */

	if ( p->parentTask )
		parent = ThimPidFromUID(p->parentTask);
	if ( parent
		&& TaskInfo[parent].ok_pended
		&& (TaskInfo[parent].pended == task) ) {
		p->have_okayed = TRUE;
		TaskInfo[parent].pended = 0;
		TaskInfo[parent].child_retcode = 0;
		TaskInfo[parent].ok_pended = FALSE;
		TaskInfo[parent].unpend_me = TRUE;
	}
	return(0);
}

/*************************************************************************
function: AxDelay - sleep for n milli seconds
*/
AxDelay(msecs)
int msecs;
{
	ACImOk();   /* calling task should not block on sleeping
                 * task */
    AxSleepMilliseconds(MAX(0, msecs));
}

/*************************************************************************
function: ACTask1PendedOnTask2(task1,task2) - check if task1 is pending 
on task2 to finish.
Note that task1==task2 has no meaning.
*/
ACTask1PendedOnTask2(task1, task2, rls_ok_pend)
	int task1;	/* host task */
	int task2;
	int rls_ok_pend;	/* If true, unblock ok_pended tasks */
{  
	register context_block *p = &TaskInfo[task1];

	if ( task1 == task2 )
		return(TRUE);		/* The only return-true point */

    if ( p->pended && p->ok_pended && rls_ok_pend )
		if ( p->pended == task2 )
			ACDoImOk(task2);	/* will full thru & return FALSE */

	if ( p->pended )
		return(ACTask1PendedOnTask2(p->pended,task2,rls_ok_pend));
	if ( p->elfCallBlocked )
		return(ACTask1PendedOnTask2(p->elfCallBlocked,task2,rls_ok_pend));
	if ( p->ipc_response_waiting )
		return(ACTask1PendedOnTask2(p->ipc_response_waiting,task2,rls_ok_pend));
	return(FALSE);
}

/*************************************************************************
function: ACLOGTask1PendedOnTask2 - identical to ACTask1PendedOnTask2,
but just logs the pended chain
*/
ACLOGTask1PendedOnTask2(task1, task2)
	int task1;	/* host task */
	int task2;
{  
	register context_block *p = &TaskInfo[task1];

	if ( task1 == task2 )
		return(TRUE);		/* The only return-true point */

	if ( p->pended )	{
		AXTRACE(("<%s> is pended on <%s>\n", TaskInfo[task1].argv[0],
			TaskInfo[task2].argv[0]));
		return(ACLOGTask1PendedOnTask2(p->pended,task2));
	}	
	if ( p->elfCallBlocked )	{
		AXTRACE(("<%s> is elfCallBlocked on <%s>\n", TaskInfo[task1].argv[0],
			TaskInfo[task2].argv[0]));
		return(ACLOGTask1PendedOnTask2(p->elfCallBlocked,task2));
	}
	if ( p->ipc_response_waiting )	{
		AXTRACE(("<%s> is ipc_response_waiting on <%s>\n", 
			TaskInfo[task1].argv[0], TaskInfo[task2].argv[0]));
		return(ACLOGTask1PendedOnTask2(p->ipc_response_waiting,task2));
	}
	return(FALSE);
}


/*****************************************************************************
function: matherr - function to be called when there is a matherr in the 
standard math library.
*/

#if !(LINUX && LIBC5)
int matherr(mathP)
struct exception *mathP;
{

        switch(TaskInfo[THIMpid].appid)
                {
                case APGE:
                case APBG:
                case APGEOCVT:
                case APGEICVT:
                case APBGOCVT:
                case APBGICVT:
                case APHPGL:
                case APBGBACK:
                                 return(gematherr(mathP,NULL));
                case APSS:
                case APSSCONV:
                case APSSEXP:
                case APSSIMP:
                case APSSBACK:
                                return(SStmatherr(mathP));
                default:
                           return(TRUE); /* Ignore the error and don't print any error messages */


                }
}
#endif /* !(LINUX && LIBC5) */

/* Stubs: */

ACerrmsg(x,y)
  char *x;
  short y;
{
  TmErrmsg(x,y);
  return(0);
}
int IPCdebug;
void ACerrlog(char *msg) {}



