/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axaccess.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:53:26 1998
*/
#include <aos.h>

extern char **ThimAccessDirPath;  /* context.c */
extern char *ThimExecDir;    /* context.c */
extern char *DirSlash;

extern void MakeAbsolutePathname();

/*
Restricted Document Access Feature:

There is a need at some of our customer sites to restrict access of
documents in certain UNIX directories.  Rather than choosing UNIX
file permissions, these customers needed Applixware to restrict document
access.  The method needs to be simple, as these users do not have
access to the shell...

The restricted document access feature will be enabled via the
presence of an environment variable 'AX_ACCESS_DIRPATH'.  This
environment variable if set, must point to one or more directories.
Multiple directories must be separated with the ':' separator just
like in the shell PATH variable.  This environment variable is looked
at in context.c, where 'ThimAccessDirPath' list of directories is created.

Note, Documents contained in the install dir 'ThimExecDir' are
critical to the running of the product.  Access to documents contained
there is allowed.

EXAMPLE:
Assume that Applixware needs to be configured in such a way that user 'joe'
must have access to Documents in his home directory and '/tmp' only.
Assume the home directory is '/home/joe'.  The following startup shell
script will produce the desired effect:
#!/bin/sh
AX_ACCESS_DIRPATH=/home/joe:/tmp
export AX_ACCESS_DIRPATH
# Let INSTALL_DIR be the Applixware install directory
$INSTALL_DIR/applix $*
*/

int
AxAccessAllowed(file)
register char *file;
{
	char abspath[STRPATH + 1];
	char allowed_dirname[STRPATH + 1];
	char **path_list, *path;
    char *p;

	if (!file)
		return(FALSE);
	if (!ThimAccessDirPath)
		return(TRUE); /* no access control */

	MakeAbsolutePathname(abspath, file, TRUE);
    strcat(abspath, DirSlash);

/* Allow access to ThimExecDir */
	sprintf(allowed_dirname, "%s%s", ThimExecDir, DirSlash);
	if (strncmp(allowed_dirname, abspath, strlen(allowed_dirname)) == 0)
		return(TRUE);

/* Allow access to path(s) pointed to by 'ThimAccessDirPath' */
	for (path_list = ThimAccessDirPath; *path_list; path_list++) {
		path = *path_list;
		if (!path)
			break;

        sprintf(allowed_dirname, "%s%s", path, DirSlash);

#if AXMS
		if (strnicmp(allowed_dirname, abspath, strlen(allowed_dirname)) == 0)
#else
		if (strncmp(allowed_dirname, abspath, strlen(allowed_dirname)) == 0)
#endif
			return(TRUE);
	}

/* Access Denied */
	return(FALSE);
}



char *GetFirstAccessPath( char *buff )
{
    buff[0] = '\0';

    if ( !ThimAccessDirPath )
        /* no restricted paths set */
        return;

    strcpy( buff, *ThimAccessDirPath );    /* return the first one */
}




/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
