/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axfile.c
   Applix ASC version: 77
   Date: Tue Jun  1 21:12:09 1999
*/
#if !AXMS && !HP800V3 && !(SUN4 && !SUN4S) && !SGI && !ALPHA && !RIOS && !LINUX && !PMAX && !SCOUNIX && !NEC_EWS
#define USE_ACL 1
#else
#define USE_ACL 0
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <errno.h>
#if AXMS
#include <io.h>
#include <sys/locking.h>
#define S_IWGRP _S_IREAD | _S_IWRITE
#define S_IRGRP _S_IREAD
#define S_IWOTH _S_IWRITE
#define S_IXOTH _S_IREAD
#else
#include <unistd.h>
#endif

#if USE_ACL
#include <sys/acl.h>
#include <pwd.h>
#include <grp.h>
#endif

#include <aos.h>
#include <elf.h>
#include <fsrts.h>
#include <afile.h>
#include <getstr20.h>
#include <getstr10.h>
#include <doctypes.h>
#include <arghdr.h>

/* a node in the list of open files */
typedef struct ax_file_info {
    struct ax_file_info *prev;
    struct ax_file_info *next;

    /* identity */
    char pathname[STRPATH+1];  /* pathname */

    /* lock info */
    unsigned int locked;       /* file is globally locked with "lockf'*/
    FILE *writer_fp;
    FILE *reader_fp;
    int writer_fd;
    int reader_fd;
    int n_readers;
    int n_writers;
    int *writer_pid;
    int *reader_pid;
} ax_file_info;
static ax_file_info *FileInfoList = (ax_file_info *)NULL;

#if USE_ACL /* ACLs in use { */
#define ENTRY_IS_GRP 0
#define ENTRY_IS_PWD 1

typedef struct ax_pwd_t {
    char *login_name;
    char *real_name;
    int uid;
    int gid;
} ax_pwd_t;

typedef struct tmp_grp_t {
    char *name;
    int gid;

    int n_secondary_member_strs;
    char **secondary_member_str;
} tmp_grp_t;

typedef struct ax_grp_t {
    char *name;
    int gid;

    int n_primary_members;
    struct ax_pwd_t *primary_member;
    int n_secondary_members;
    struct ax_pwd_t *secondary_member;
} ax_grp_t;

typedef union ax_subentry_t {
    struct ax_grp_t grp;
    struct ax_pwd_t pwd;
} ax_subentry_t;
 
typedef struct ax_entry_t {
    int type;
    ax_subentry_t u;
} ax_entry_t;
 
typedef struct entry_list_t {
    int n;
    ax_entry_t *element;
} entry_list_t;
#endif /* } ACLs in use */
 
static int listing_pwds;
static int AxUseLockDemon;
static int AxLockOnOpen;
static int AxACL;

extern data_ptr AxMakeArray();
extern data_ptr AxMakeIntData();
extern char *GetProfileByStr();
extern char *TrueRealloc();
extern char *TrueMalloc();
extern char *TaskAlloc();
extern char *TaskRealloc();

extern data_ptr NullDataPtr;
extern char *DirSlash;

/***********************************************************************/
/*  read profile to determine if we should lock files                  */
/***********************************************************************/
int
AxReadLockState()
{
#if AXMS
    AxUseLockDemon = FALSE;
#else
    char *p;
    p = GetProfileByStr(NXLT("axLockFiles"));
    if ( *p != '\0' )
        AxUseLockDemon = atoi(p) ? TRUE : FALSE;
    else
        AxUseLockDemon = FALSE; /* default is locking OFF  */
#endif
    AxLockOnOpen = TRUE;
    return(AxUseLockDemon);
}

/***********************************************************************/
/* print out current file nodes                                        */
/***********************************************************************/
void
DumpFileNodes()
{
    int n_nodes, n_rlocks, n_wlocks;
    ax_file_info *nodeP;

    n_nodes = n_rlocks = n_wlocks = 0;
    nodeP = FileInfoList;
    while ( nodeP )
    {
        n_nodes++;
        printf("NODE %d\n", n_nodes);
        printf(" pathname <%s>\n", nodeP->pathname);
        printf(" n_readers %d\n", nodeP->n_readers);
        printf(" n_writers %d\n", nodeP->n_writers);
        if ( nodeP->n_readers )
            n_rlocks++;
        if ( nodeP->n_writers )
            n_wlocks++;
        nodeP = nodeP->next;
    }
    if ( n_nodes == 0 )
        printf("No files locked.\n");
    else
    {
        printf("%d files locked:", n_nodes);
        printf(" %d read locks,", n_rlocks);
        printf(" %d write locks.\n", n_wlocks);
    }
}

/***********************************************************************/
/* Add an owner to an existing node in the list of open files.         */
/***********************************************************************/
static int
AddNodeOwner(infoP, pid, read_only)
ax_file_info *infoP;
int pid;
int read_only;
{
    int ix, sz, fd, err;
    FILE *fp;

    /* compatability mode (no locking for read only) */
    if ( AxUseLockDemon == FALSE && read_only == TRUE )
        return(AOK);

    /* set the owner */
    if ( read_only )
    {
        /* create pointers (if we don't already have 'em) */
        if ( !infoP->reader_fp )
        {
            fp = fopen(infoP->pathname, READ_BINARY_STR);
            if ( fp )
            {
                infoP->reader_fp = fp;
                infoP->reader_fd = fileno(fp);
            }
        }

        /* add another ref */
        if ( !infoP->reader_fp )
            err = errno;
        else
        {
            ix = infoP->n_readers++;
            sz = infoP->n_readers * sizeof(int);
            if ( infoP->reader_pid )
                infoP->reader_pid = (int *)TrueRealloc(infoP->reader_pid, sz);
            else
                infoP->reader_pid = (int *)TrueMalloc(sz);
            infoP->reader_pid[ix] = pid;
            err = AOK;
        }
    }
    else if ( infoP->n_writers == 0 || infoP->writer_pid[0] == pid )
    {
        /* create pointers (if we don't already have 'em) */
        if ( !infoP->writer_fp && AxUseLockDemon )
        {
            fp = fopen(infoP->pathname, READ_TRUNC_STR);
            if ( fp )
            {
                fd = fileno(fp);
#if AXMS
                if ( locking(fd, _LK_NBLCK, 0) != 0 )
#else
                if ( lockf(fd, F_TLOCK, 0) != 0 )
#endif
                {
                    if ( errno == EACCES )
                        errno = EAGAIN;
                    fclose(fp);
                }
                else
                {
                    infoP->writer_fp = fp;
                    infoP->writer_fd = fd;
                    infoP->locked = TRUE;
                }
            }
        }

        /* add a ref */
        if ( !infoP->writer_fp && AxUseLockDemon )
            err = errno;
        else
        {
            ix = infoP->n_writers++;
            sz = infoP->n_writers * sizeof(int);
            if ( infoP->writer_pid )
                infoP->writer_pid = (int *)TrueRealloc(infoP->writer_pid, sz);
            else
                infoP->writer_pid = (int *)TrueMalloc(sz);
            infoP->writer_pid[ix] = pid;
            err = AOK;
        }
    }
    else /* tried to get a read/write lock on an already locked file */
    {
        err = EAGAIN;
    }
    return(err);
}

/***********************************************************************/
/* Remove an owner from an existing node in the list of open files.    */
/***********************************************************************/
static void
RemoveNodeOwner(infoP, pid, read_only)
ax_file_info *infoP;
int pid;
int read_only;
{
    int ix, sz, task;

    /* compatability mode (no locking for read only) */
    if ( AxUseLockDemon == FALSE && read_only == TRUE )
        return;

    /* remove the owner */
    if ( read_only )
    {
        for ( ix=0; ix<infoP->n_readers; ix++ )
        {
            if ( pid == infoP->reader_pid[ix] )
                break;
        }
        if ( ix < infoP->n_readers )
        {
            if ( infoP->n_readers < 1 )
                infoP->n_readers = 0;
            else
                infoP->n_readers--;
            for ( ; ix<infoP->n_readers; ix++ )
                infoP->reader_pid[ix] = infoP->reader_pid[ix+1]; 
            sz = infoP->n_readers * sizeof(int);
            if ( infoP->n_readers == 0 )
                infoP->reader_pid = (int *)0;
            else if ( infoP->reader_pid )
                infoP->reader_pid = (int *)TrueRealloc(infoP->reader_pid, sz);
            else
                infoP->reader_pid = (int *)TrueMalloc(sz);
        }
        if ( infoP->n_writers == 0 )
        {
            if ( infoP->n_readers == 0 )
            {
                if ( infoP->reader_fp )
                {
                    fclose(infoP->reader_fp);
                    infoP->reader_fp = (FILE *)0;
                    infoP->reader_fd = -1;
                }
            }
        }
    }
    else /* writer */
    {
        for ( ix=0; ix<infoP->n_writers; ix++ )
        {
            if ( pid == infoP->writer_pid[ix] )
                break;
        }
        if ( ix < infoP->n_writers )
        {
            if ( infoP->n_writers < 1 )
                infoP->n_writers = 0;
            else
                infoP->n_writers--;
            for ( ; ix<infoP->n_writers; ix++ )
                infoP->writer_pid[ix] = infoP->writer_pid[ix+1]; 
            sz = infoP->n_writers * sizeof(int);
            if ( infoP->n_writers == 0 )
                infoP->writer_pid = (int *)0;
            else if ( infoP->writer_pid )
                infoP->writer_pid = (int *)TrueRealloc(infoP->writer_pid, sz);
            else
                infoP->writer_pid = (int *)TrueMalloc(sz);
        }
        if ( infoP->n_writers == 0 )
        {
            if ( infoP->writer_fp )
            {
                fclose(infoP->writer_fp);
                infoP->writer_fp = (FILE *)0;
                infoP->writer_fd = -1;
                infoP->locked = FALSE;
            }
            if ( infoP->n_readers == 0 )
            {
                if ( infoP->reader_fp )
                {
                    fclose(infoP->reader_fp);
                    infoP->reader_fp = (FILE *)0;
                    infoP->reader_fd = -1;
                }
            }
        }
    }
}

/***********************************************************************/
/* Add a node to the list of open files.                               */
/***********************************************************************/
static ax_file_info *
create_node(pathname)
char *pathname;
{
    ax_file_info *infoP;
    int ix, sz;

    /* add an empty node to the front of the chain */
    infoP = (ax_file_info *)TrueMalloc(sizeof(ax_file_info));
    memset(infoP, 0, sizeof(ax_file_info));
    infoP->writer_fd = -1;
    infoP->reader_fd = -1;
    infoP->prev = (ax_file_info *)0;
    infoP->next = FileInfoList;
    if ( FileInfoList != (ax_file_info *)0 )
        FileInfoList->prev = infoP; /* point old head at new head */
    FileInfoList = infoP;

    /* set the identity of the node */
    MakeAbsolutePathname(infoP->pathname, pathname, TRUE);
    return(infoP);
}

/***********************************************************************/
/* Remove a node from the open file list.                              */
/***********************************************************************/
static void
destroy_node(infoP)
ax_file_info *infoP;
{
    ax_file_info *prev_element;
    ax_file_info *next_element;
    
    if ( infoP && infoP->n_writers == 0 && infoP->n_readers == 0 )
    {
        /* element to delete is the first thing in the chain */
        if ( infoP == FileInfoList )
        {
            next_element = infoP->next;
            FileInfoList = next_element;
            if ( next_element != (ax_file_info *)0 )
                next_element->prev = (ax_file_info *)0;
        }
        else /* unlink this entry out of the chain */
        {
            prev_element = infoP->prev;
            next_element = infoP->next;

            prev_element->next = next_element;
            if ( next_element != (ax_file_info *)0 )
                next_element->prev = prev_element;
        }
 
        /* If we have a file handle, close it and unlock the file */
        if ( infoP->reader_pid )
            TrueFree(infoP->reader_pid);
        if ( infoP->writer_pid )
            TrueFree(infoP->writer_pid);
        if ( infoP->reader_fp )
            fclose(infoP->reader_fp);
        if ( infoP->writer_fp )
            fclose(infoP->writer_fp);
        memset(infoP, 0, sizeof(ax_file_info));
        TrueFree(infoP);
    }
    return;
}

/***********************************************************************/
/* find the node of the pathname on the FileInfoList. If not on the    */
/* list, optionally create a node for it.                              */
/***********************************************************************/
static ax_file_info *
find_node(pathname)
char *pathname;
{
    ax_file_info *infoP, *nodeP;
    int err, read_only, pid;
    char abspath[STRPATH+1];

    /* look this file up in the list of files we know about */
    MakeAbsolutePathname(abspath, pathname, TRUE);
    infoP = (ax_file_info *)0;
    nodeP = FileInfoList;
    while ( nodeP && infoP == (ax_file_info *)0 )
    {
        if ( streq(abspath, nodeP->pathname) )
            infoP = nodeP; /* we found it, it is already open */
        else
            nodeP = nodeP->next;
    }
    return(infoP);
}

/***********************************************************************/
/* clean up any file entries used by a particular task                 */
/***********************************************************************/
void
AFCleanFilesByTask(pid)
int pid;
{
    ax_file_info *infoP, *nextP;
    int i;

    infoP = FileInfoList;
    while ( infoP )
    {
        nextP = infoP->next;  /* record now, since infoP may get free'd */
        for ( i=0; i<infoP->n_readers; i++ )
        {
            if ( infoP->reader_pid[i] == pid )
                RemoveNodeOwner(infoP, pid, TRUE);
        }
        for ( i=0; i<infoP->n_writers; i++ )
        {
            if ( infoP->writer_pid[i] == pid )
                RemoveNodeOwner(infoP, pid, FALSE);
        }
        destroy_node(infoP);
        infoP = nextP;
    }
    return;
}

/***********************************************************************/
/* return whether or not a file is already locked                      */
/* This returns:   TRUE  - if the file is already open                 */
/*                 FALSE - if the file is not open, or does not exist  */
/***********************************************************************/
int
AFIsLocked(pathname, read_only, local_lock)
mbchar *pathname;
int read_only;
int local_lock;
{
    ax_file_info *infoP;
    int is_locked, fd;
    FILE *fp;

    infoP = find_node(pathname);
    if ( AxUseLockDemon == FALSE )
    {
        if ( infoP == (ax_file_info *)0 )
            is_locked = FALSE;
        else
            is_locked = TRUE;
    }
    else
    {
        if ( infoP == (ax_file_info *)0 )
        {
            if ( local_lock || read_only == TRUE )
                is_locked = FALSE;
            else /* not locked in this process - check the lock demon */
            {
                fp = fopen(pathname, READ_BINARY_STR);
                if ( !fp )
                    is_locked = FALSE;
                else
                {
                    fd = fileno(fp);
#if AXMS
                    if ( locking(fd, _LK_NBLCK, 0) == 0 )
#else
                    if ( lockf(fd, F_TEST, 0) == 0 )
#endif
                        is_locked = FALSE;
                    else if ( errno == EAGAIN )
                        is_locked = TRUE;
                    else if ( errno == EACCES )
                        is_locked = TRUE;
                    else
                        is_locked = FALSE;
                    fclose(fp);
                }
            }
        }
        else if ( read_only == -1 ) /* read or write lock */
            is_locked = TRUE;
        else if ( read_only ) /* read "lock" */
        {
            if ( infoP->n_readers > 0 )
                is_locked = TRUE;
            else
                is_locked = FALSE;
        }
        else /* write lock */
        {
            if ( infoP->n_writers > 0 )
                is_locked = TRUE;
            else
                is_locked = FALSE;
        }
    }
    return(is_locked);
}

/***********************************************************************/
/* return whether or not a file is already open for writing            */
/* This returns:   TRUE  - if the file is already open                 */
/*                 FALSE - if the file is not open, or does not exist  */
/***********************************************************************/
int
AFIsFileOpen(pathname)
mbchar *pathname;
{
    if ( AFIsLocked(pathname, FALSE, TRUE) ) /* writeable locally */
        return(TRUE);
    else
        return(FALSE);
}

/***********************************************************************/
/* return whether or not a file is already open for writing            */
/* This returns:   BTRUE - if the file is already open                 */
/*                 FALSE - if the file is not open, or does not exist  */
/***********************************************************************/
int
ElfbIsFileOpen(pathname)
mbchar *pathname;
{
    if ( AFIsFileOpen(pathname) ) /* write */
        return(BTRUE);
    else
        return(FALSE);
}

/***********************************************************************/
/* return whether or not a file is already locked  for writing            */
/* This returns:   BTRUE - if the file is already locked                 */
/*                 FALSE - if the file is not locked, or does not exist  */
/***********************************************************************/
int
ElfbIsFileLocked(pathname)
mbchar *pathname;
{
    if ( AFIsLocked(pathname, FALSE, FALSE) ) /* locked */
        return(BTRUE);
    else
        return(FALSE);
}

/***********************************************************************/
/* Lock file according to access and profile.                          */
/***********************************************************************/
static int
cache_lock(pathname, pid, force_read_only, retry)
char *pathname;
int pid;
int *force_read_only;
int retry;
{
    int fd, locked, read_only, err;
    ax_file_info *infoP;
    struct stat buf;
    FILE *fp;

    read_only = *force_read_only;
    infoP = find_node(pathname);
    if ( infoP ) /* file and node already exist */
    {
        err = AddNodeOwner(infoP, pid, read_only);
        if ( err != AOK && read_only == FALSE && retry )
        {
            read_only = TRUE;
            err = AddNodeOwner(infoP, pid, read_only);
        }
    }
    else /* create a node */
    {
        /* open/create the file... */
        if ( AxUseLockDemon == FALSE ) /* simple pre4.3 lock */
        {
            if ( read_only )
                err = AOK;
            else
            {
                /* check to see if the file exists and is writable */
                if ( stat(pathname, &buf) == 0 )    
                {
                    err = (Aaccess(pathname, W_OK) < 0) ? errno : AOK;
                    if ( err && retry )
                    {
                        if ( Aaccess(pathname, R_OK) == 0 )
                        {
                            read_only = TRUE;
                            err = AOK;
                        }
                    }
                }   
                else /* create the file if it doesn't exist... */
                {
                    if ( errno != ENOENT )
                        err = errno;
                    else
                    {
                        fp = fopen(pathname, WRITE_TRUNC_STR);
                        if ( !fp )
                            err = errno;
                        else
                        {
                            fclose(fp);
                            err = AOK;
                        }
                    }
                }

                /* if no errors, create the node */
                 if ( err == AOK )
                 {
                     infoP = create_node(pathname);
                     err = AddNodeOwner(infoP, pid, read_only);
                     if ( err != AOK )
                        destroy_node(infoP);
                 }
            }
        }
        else /* full blown lock - use lock deamon */
        {
            if ( read_only )
                fp = fopen(pathname, READ_BINARY_STR);
            else
            {
                /* create the fp and lock it */
                fp = fopen(pathname, READ_TRUNC_STR); /* read/write */
                if ( !fp )
                {
                    /* if the file doesn't exist, create for read/write */
                    err = errno;
                    if ( stat(pathname, &buf) == 0 )
                        errno = err;
                    else
                        fp = fopen(pathname, WRITE_TRUNC_STR);
                    err = AOK;
                }
                if ( fp )
                {
                    fd = fileno(fp);
#if AXMS
                    if ( locking(fd, _LK_NBLCK, 0) != 0 )
#else
                    if ( lockf(fd, F_TLOCK, 0) != 0 )
#endif
                    {
                        if ( errno == EACCES )
                            errno = EAGAIN;
                        fclose(fp);
                        fp = (FILE *)0;
                        fd = -1;
                    }
                }
                 
                /* create or lock failed - try to open it read only */
                if ( !fp && retry )
                {
                    fp = fopen(pathname, READ_BINARY_STR);
                    read_only = TRUE;
                }
            }    

            /* ...handle any errors... */
            if ( fp )
            { 
                err = AOK;
                fd = fileno(fp);
            }
            else
            { 
                err = errno;
                fd = -1;
            }

            /* ...then if no errors, create the node */
            if ( err == AOK )
            {
                if ( stat(pathname, &buf) != 0 )
                {
                    err = errno;
                    fclose(fp);
                    fd = -1;
                }
                else
                {
                    /* lock the file if locked for writing */
                    infoP = create_node(pathname);
                    if ( read_only )
                    {
                        infoP->reader_fp = fp;
                        infoP->reader_fd = fd;
                    }
                    else
                    {
                        infoP->writer_fp = fp;
                        infoP->writer_fd = fd;
                        infoP->locked = TRUE;
                    }
                    err = AddNodeOwner(infoP, pid, read_only);
                    if ( err != AOK )
                        destroy_node(infoP);
                }
            }
        }
    }
    *force_read_only = read_only;
    return(err);
}

/***********************************************************************/
/* Lock file according to access and profile.                          */
/***********************************************************************/
int
AFTaskLockFile(pathname, pid, force_read_only, retry)
char *pathname;
int pid;
int *force_read_only;
int retry;
{
    int err;
    err = cache_lock(pathname, pid, force_read_only, retry);
    if ( err != AOK )
    {
        if ( err == EAGAIN )
            err = ISLOCKED;
        else
        {
            errno = err;
            err = ErrnoErr();
        }
    }
    return(err);
}

/***********************************************************************/
/* Lock file according to access and profile.                          */
/***********************************************************************/
int
AFLockFile(pathname, force_read_only, retry)
char *pathname;
int *force_read_only;
int retry;
{
    int pid = ThimTopPidFromPid(THIMpid);
    return(AFTaskLockFile(pathname,pid,force_read_only,retry));
}

/***********************************************************************/
/* Note the fact that a particular file is in use                      */
/***********************************************************************/
void
AFNoteFileInUse(pathname)
char *pathname;
{
    int read_only;
    read_only = FALSE;
    AFLockFile(pathname, &read_only, FALSE);
    return;
}

/***********************************************************************/
/* Same as AFNoteFileInUse - takes a readonly flag as arg              */
/***********************************************************************/
data_ptr
AxfNoteFileInUse(argDatum)
data_ptr argDatum;
{
    
    char *pathname = (char *)StrFromArgArray(argDatum,0);
    int read_only = IntFromArgArray(argDatum,1);
    int pid = ThimPidFromUID(IntFromArgArray(argDatum,2));
    int err;

    /* If Elf TRUE, change to C TRUE --Kiran 10/4/96 */
    if (read_only == -1)
        read_only = TRUE;
    else
        read_only = FALSE;

    if (!pid)
        err = AFLockFile(pathname,&read_only, FALSE);
    else 
        err = AFTaskLockFile(pathname,pid,&read_only,FALSE);
    if ( err )
        ElfStrAbort(err, NULL, pathname);
    return(NullDataPtr);
}

/***********************************************************************/
/* Determine read_only using mode string.                              */
/*  r  - open for reading                                              */
/*  w  - truncate (not supported) or create for writing                */
/*  a  - append (not supported) or create for writing                  */
/*  r+ - open for update (read/write)                                  */
/*  w+ - truncate (not supported) or create for update                 */
/*  a+ - append (not supported) or create for update                   */
/* modes not supported are opened read/write, instead                  */
/***********************************************************************/
int
af_read_only_mode(mode_str)
char *mode_str;
{
    int read_only;

    if ( streq(mode_str, READ_BINARY_STR) ) /* r */
        read_only = TRUE;
    else if ( streq(mode_str, READ_STR) ) /* r */
        read_only = TRUE;
    else
        read_only = FALSE;
    return(read_only);
}
int
af_append_mode(mode_str)
char *mode_str;
{
    int append;

    if ( streq(mode_str, NXLT("a")) )
        append = TRUE;
    else if ( streq(mode_str, NXLT("ab")) )
        append = TRUE;
    else
        append = FALSE;
    return(append);
}

/***********************************************************************/
/* Search FileInfoList and return locked file pointer. If not found,   */
/* do a simple fopen.                                                  */
/***********************************************************************/
static FILE *
cache_fopen(pathname, mode_str, pid_check, pid)
char *pathname;
char *mode_str;
int pid_check;
int pid;
{
    int read_only, fd, i, err;
    ax_file_info *infoP;
    FILE *fp;

    /* read or write? */
    read_only = af_read_only_mode(mode_str);

    /* lock demon disabled */
    if ( AxUseLockDemon == FALSE )
    {
        if ( read_only )
            fp = fopen(pathname, mode_str);
        else
        {
            /* find the node on the list */
            infoP = find_node(pathname);

            /* no open node - try to create one */
            if ( infoP == (ax_file_info *)0 )
            {
                /* need a real pid to create a new node */
                if ( pid_check == FALSE || AxLockOnOpen == FALSE )
                    errno = EIO;
                else /* open with a real pid - create the node */
                {
                    errno = cache_lock(pathname, pid, &read_only, FALSE);
                    if ( errno == AOK )
                        infoP = find_node(pathname);
                }
            }
            if ( infoP == (ax_file_info *)0 )
                fp = (FILE *)0; /* errno set above, during node create */

            else /* already open */
            {
                if ( pid_check == FALSE )
                    fp = fopen(pathname, mode_str);
                else
                {
                    if ( infoP->n_writers == 0 )
                        AddNodeOwner(infoP, pid, read_only);
                    for ( i=0; i<infoP->n_writers; i++ )
                    {
                        if ( pid == infoP->writer_pid[i] )
                            break;
                    }
                    if ( i < infoP->n_writers )
                        fp = fopen(pathname, mode_str);
                    else
                    {
                        fp = (FILE *)0;
                        errno = EIO;
                    }
                }
            }
        }
    }
    else
    {
        /* find the node on the list */
        infoP = find_node(pathname);

        /* no open node - try to create one */
        if ( infoP == (ax_file_info *)0 )
        {
            /* need a real pid to create a new node */
            if ( pid_check == FALSE || AxLockOnOpen == FALSE )
                errno = EIO;
            else /* open with a real pid - create the node */
            {
                errno = cache_lock(pathname, pid, &read_only, FALSE);
                if ( errno == AOK )
                    infoP = find_node(pathname);
            }
        }
        if ( infoP == (ax_file_info *)0 )
            fp = (FILE *)0; /* errno set above, during node create */

        else /* already open */
        {
            if ( read_only )
            {
                if ( pid_check == FALSE )
                    fp = infoP->reader_fp;
                else
                {
                    if ( !infoP->reader_fp || infoP->n_readers == 0 )
                        AddNodeOwner(infoP, pid, read_only);
                    for ( i=0; i<infoP->n_readers; i++ )
                    {
                        if ( pid == infoP->reader_pid[i] )
                            break;
                    }
                    if ( i < infoP->n_readers )
                        fp = infoP->reader_fp;
                    else
                    {
                        fp = (FILE *)0;
                        if ( AxLockOnOpen == FALSE )
                            errno = EIO;
                        else /* try to establish a lock */
                        {
                            errno = cache_lock(pathname, pid, &read_only, FALSE);
                            if ( errno == AOK )
                            {
                                infoP = find_node(pathname);
                                if ( infoP == (ax_file_info *)0 )
                                    errno = EIO;
                                else
                                    fp = infoP->reader_fp;
                            }
                        }
                    }
                }
            }
            else /* writable */
            {
                if ( pid_check == FALSE )
                    fp = infoP->writer_fp;
                else
                {
                    if ( !infoP->writer_fp || infoP->n_writers == 0 )
                        AddNodeOwner(infoP, pid, read_only);
                    for ( i=0; i<infoP->n_writers; i++ )
                    {
                        if ( pid == infoP->writer_pid[i] )
                            break;
                    }
                    if ( i < infoP->n_writers )
                        fp = infoP->writer_fp;
                    else
                    {
                        fp = (FILE *)0;
                        if ( AxLockOnOpen == FALSE )
                            errno = EIO;
                        else /* try to establish a lock */
                        {
                            errno = cache_lock(pathname, pid, &read_only, FALSE);
                            if ( errno == AOK )
    						{
                                infoP = find_node(pathname);
                                if ( infoP == (ax_file_info *)0 )
                                    errno = EIO;
                                else
                                    fp = infoP->writer_fp;
                            }
                        }
                    }
                }
            }
            if ( fp )
            {
                if ( af_append_mode(mode_str) == FALSE )
                    rewind(fp);
                else
                    fseek(fp, 0, 2);
            }
        }
    }
    return(fp);
}

/***********************************************************************/
/* Search FileInfoList and return locked file pointer. If not found,   */
/* do a simple fopen.                                                  */
/***********************************************************************/
FILE *
AFTaskfopen(pid, pathname, mode_str)
int pid;
char *pathname;
char *mode_str;
{
#if AX_X11
    return(cache_fopen(pathname, mode_str, TRUE, pid));
#else
/* On Windows, file open mode "a" does *not* set the file pointer to the end
 * of the file until you actually write to it. So we explicitly set it here.
 */
    FILE *fp=cache_fopen(pathname, mode_str, TRUE, pid);
    if ( streq(mode_str, NXLT("a")) )
        fseek(fp, 0L, SEEK_END);
    return fp;
#endif
}

/***********************************************************************/
/* Search FileInfoList and return locked file pointer. If not found,   */
/* do a simple fopen.                                                  */
/***********************************************************************/
FILE *
AFfopen(pathname, mode_str)
char *pathname;
char *mode_str;
{
#if AX_X11
    return(cache_fopen(pathname, mode_str, FALSE, -1));
#else
/* On Windows, file open mode "a" does *not* set the file pointer to the end
 * of the file until you actually write to it. So we explicitly set it here.
 */
    FILE *fp=cache_fopen(pathname, mode_str, FALSE, -1);
    if ( streq(mode_str, NXLT("a")) )
        fseek(fp, 0L, SEEK_END);
    return fp;
#endif
}

/***********************************************************************/
/* set file length to current position in file                         */
/***********************************************************************/
static void
ax_truncate(fp, fd)
FILE *fp;
int fd;
{
    unsigned long curr_pos = ftell(fp);
#if AXMS
    chsize(fd, curr_pos);
#else
    fflush(fp);
    ftruncate(fd, curr_pos);
#endif
}

/***********************************************************************/
/* Find the node of the pathname on the FileInfoList. If not on the    */
/* list, close the file pointer. Otherwise, optionally truncate file   */
/* following current file pointer position, then sync file to disk.    */
/***********************************************************************/
static int
cache_fclose(pathname, fp, set_eof, read_only)
char *pathname;
FILE *fp;
int set_eof;   /* if TRUE, will set end-of-file to current file position */
int read_only;
{
    ax_file_info *infoP;
    int ret;

    if ( AxUseLockDemon == FALSE )
    {
        if ( set_eof && (read_only == FALSE))
            ax_truncate(fp, fileno(fp));
        ret = fclose(fp);
    }
    else
    {
        /* find the file's node on the open file list */
        infoP = find_node(pathname);
        if ( (infoP == (ax_file_info *)0) )
            ret = fclose(fp); /* no record of this file, close it directly */
        if ( read_only ) {
			if (infoP->reader_fp) {
				fflush( infoP->reader_fp );  /* ignore errors */
					 /*will force 'real' read next time file is accessed*/
			}
            ret = AOK;
		}
        else /* read_only == FALSE */
        {
            /* "delete" material following file pointer position */
            if ( set_eof )
                ax_truncate(infoP->writer_fp, infoP->writer_fd);

            /* update */
            ret = fflush(infoP->writer_fp);
            if ( ret < 0 )
                ret = EOF; /* error code from bad "fclose" */
        }
    }
    return(ret);
}

/***********************************************************************/
/* Find the node of the pathname on the FileInfoList. If not on the    */
/* list, close the file. Otherwise, truncate material after current    */
/* file position, then write the file to disk.                         */
/***********************************************************************/
int
AFfclose(pathname, read_only, fp)
char *pathname;
int read_only;
FILE *fp;
{
    return(cache_fclose(pathname, fp, TRUE, read_only));
}

/***********************************************************************/
/* As AFfclose, but no file truncation takes place. Use this function  */
/* if you open a file with mode 'r+', and then decide to close it      */
/* without making any changes after all.                               */
/***********************************************************************/
int
AFfcloseNoTruncate(pathname, read_only, fp)
char *pathname;
int read_only;
FILE *fp;
{
    return(cache_fclose(pathname, fp, FALSE, read_only));
}

/***********************************************************************/
/* delete the file from our internal list of files which are open    */
/***********************************************************************/
void
AFTaskUnlockFile(pathname, pid, read_only)
char *pathname;
int pid;
int read_only;
{
    ax_file_info *infoP;
    if ( AxUseLockDemon == FALSE && read_only )
        return;

    infoP = find_node(pathname);
    if ( infoP )
    {
        RemoveNodeOwner(infoP, pid, read_only);
        destroy_node(infoP);
    }
    return;
}

/***********************************************************************/
/* delete the file from our internal list of files which are open    */
/***********************************************************************/
void
AFUnlockFile(pathname, read_only)
char *pathname;
{
    AFTaskUnlockFile(pathname, ThimTopPidFromPid(THIMpid), read_only);
}

/*********************************************************************/
/* delete the file from our internal list of files which are open    */
/*********************************************************************/
void
AFDoneWithFile(pathname)
char *pathname;
{
    AFUnlockFile(pathname, FALSE);
    return;
}

/***********************************************************************/
/* Same as AFDoneWithFile - takes a readOnlyFlag as its 2nd arg        */
/***********************************************************************/
data_ptr
AxfDoneWithFile(argDatum)
data_ptr argDatum;
{
    char *pathname = (char *)StrFromArgArray(argDatum,0);
    int read_only = IntFromArgArray(argDatum,1);
    int pid = ThimPidFromUID(IntFromArgArray(argDatum,2));

    /* If Elf TRUE, change to C TRUE --Kiran 10/4/96 */
    if (read_only == -1)
        read_only = TRUE;
    else
        read_only = FALSE;

    if (!pid)
        AFUnlockFile(pathname, read_only);
    else
        AFTaskUnlockFile(pathname, pid, read_only);
    return(NullDataPtr);
}

/***********************************************************************/
/* return the true path to a documents link. Note: Caller is           */
/* reponsible for allocating memory for trueLinkName.                  */
/* OBSOLETE - use AxMergePaths instead!                                */
/***********************************************************************/
typeErr
AFGetLinkPath(trueLinkName, parentDocName, linkName)
char *trueLinkName;     /* RET'd the true path to the link */
char *parentDocName;    /* IN the parent document containing the link */
char *linkName;         /* IN the link name as known to the parent doc */
{
    char directory[STRPATH+1];
    char pathname[STRPATH+1];

    /* if linkName is an absolute pathname, then return without modifications */
#if AX_NT
    if ( (*linkName == *DirSlash) || (NULL != strchr(linkName, ':')) )
#else
    if ( *linkName == *DirSlash )   
#endif
    {
        strcpy(trueLinkName, linkName);
    }
    else
    {
        /* get the directory where the parentDoc exits */
        AxParsePath(parentDocName, directory, pathname);

        /* make it absolute */
        MakeAbsolutePathname(trueLinkName, directory, TRUE);

        /* concat a slash */
        strcat(trueLinkName, DirSlash);

        /* concat the linkName to this new path */
        strcat(trueLinkName, linkName);
    }
    return(AOK);
}

/***********************************************************************/
/* Return whether passed filenames are the same file.  The passed      */
/* filenames must contain the path (i.e., must not be relative), but   */
/* but may contain links.                                              */
/***********************************************************************/
int
AFFilesSame (filename1, filename2)
mbchar *filename1;
mbchar *filename2;
{
    mbchar absFilename1[STRPATH+1];
    mbchar absFilename2[STRPATH+1];

    MakeAbsolutePathname(absFilename1, filename1, TRUE);  /* resolve links */
    MakeAbsolutePathname(absFilename2, filename2, TRUE);  /* resolve links */
    return(MBstreq(absFilename1, absFilename2));
}

/***********************************************************************/
/***********************************************************************/


/***********************************************************************/
/***********************************************************************/


#if USE_ACL /* ACLs active { */
/***********************************************************************/
/*  read profile to determine if we should enable access control lists */
/***********************************************************************/
int
AxReadACLState()
{
    char *p;
    p = GetProfileByStr(NXLT("axACL"));
    if ( *p != '\0' )
        AxACL = atoi(p) ? TRUE : FALSE;
    else
        AxACL = TRUE; /* default is access control disabled */
    return(AxACL);
}

/***********************************************************************/
/*  determine if we support changes to acl entries of this type        */
/***********************************************************************/
static int
validate_acl_set_type(type)
int type;
{
    switch(type)
    {
        case USER_OBJ:  return(AOK);
        case USER:      return(AOK);
        case GROUP_OBJ: return(AOK);
        case GROUP:     return(AOK);
        case OTHER_OBJ: return(AOK);
    }
    return(!AOK);
}

/***********************************************************************/
/*  determine if we support changes to acl entries of this type        */
/***********************************************************************/
static int
validate_acl_clear_type(type)
int type;
{
    switch(type)
    {
        case USER:      return(AOK);
        case GROUP:     return(AOK);
    }
    return(!AOK);
}

/***********************************************************************/
/*  determine if type:id match acl entry                               */
/***********************************************************************/
static int
acl_match(acl_entry, type, id, exact)
aclent_t *acl_entry;
int type, id, exact;
{
    switch(type)
    {
        case USER_OBJ:
        case USER:     
            if ( acl_entry->a_id == id )
            {
                if ( exact )
                {
                    if ( acl_entry->a_type == type )
                        return(AOK);
                }
                else
                {
                    if ( acl_entry->a_type == USER_OBJ ||
                         acl_entry->a_type == USER )
                        return(AOK);
                }
            }
        break;
        case GROUP_OBJ:
        case GROUP:    
            if ( acl_entry->a_id == id )
            {
                if ( exact )
                {
                    if ( acl_entry->a_type == type )
                        return(AOK);
                }
                else
                {
                    if ( acl_entry->a_type == GROUP_OBJ ||
                         acl_entry->a_type == GROUP )
                        return(AOK);
                }
            }
        break;

        case OTHER_OBJ:
            if ( acl_entry->a_type == type )
                return(AOK);
        break;
    }
    return(!AOK);
}

/***********************************************************************/
/*  create string for type:id                                          */
/***********************************************************************/
static void
make_acl_str(type, id, str)
int type, id;
char *str;
{
    struct passwd *pwd;
    struct group *grp;

    switch(type)
    {
        case USER_OBJ:
            pwd = getpwuid(id);
            if ( !pwd )
                str[0] = (char)0;
            else
                sprintf(str, "Owning user: %s", pwd->pw_name);
        break;
        case USER:
            pwd = getpwuid(id);
            if ( !pwd )
                str[0] = (char)0;
            else
                sprintf(str, "Additional user: %s", pwd->pw_name);
        break;
        case GROUP_OBJ:
            grp = getgrgid(id);
            if ( !grp )
                str[0] = (char)0;
            else
                sprintf(str, "Owning group: %s", grp->gr_name);
        break;
        case GROUP:
            grp = getgrgid(id);
            if ( grp )
                str[0] = (char)0;
            else
                sprintf(str, "Additional group: %s", grp->gr_name);
        break;
        case OTHER_OBJ:
            sprintf(str, "All Others");
        break;

        default:
            str[0] = (char)0;
        break;
    }
}

/***********************************************************************/
/*  set permissions for type:id on file <path>                         */
/***********************************************************************/
static int
set_acl_entry(path, type, id, perm)
char *path;
int type;
int id;
int perm;
{
    aclent_t acl_buf[MAX_ACL_ENTRIES];
    int n_entries, err, i;

    /* sanity */
    if ( validate_acl_set_type(type) != AOK )
    {
        errno = EINVAL;
        err = ErrnoErr();
        return(err);
    }

    /* get the current acl */
    n_entries = acl(path, GETACL, MAX_ACL_ENTRIES, acl_buf);
    if ( n_entries == (-1) )
        err = ErrnoErr();
    else
    {
        /* search the current acl for the type:id */
        for ( i=0; i<n_entries; i++ )
        {
            if ( acl_match(&acl_buf[i], type, id, FALSE) == AOK )
            {
                type = acl_buf[i].a_type; /* don't alter old type */
                break;
            }
        }

        /* update/add the new type:id with passed perm */
        if ( i >= n_entries )
            n_entries++;
        if ( n_entries > MAX_ACL_ENTRIES )
        {
            errno = EINVAL;
            err = ErrnoErr();
        }
        else
        {
            acl_buf[i].a_type = type;
            acl_buf[i].a_id = id;
            acl_buf[i].a_perm = perm;
            if ( acl(path, SETACL, n_entries, acl_buf) == (-1) )
                err = ErrnoErr();
            else
                err = AOK;
        }
    }
    return(err);
}

/***********************************************************************/
/*  remove type:id form acl                                            */
/***********************************************************************/
static int
clear_acl_entry(path, type, id)
char *path;
int type;
int id;
{
    aclent_t acl_buf[MAX_ACL_ENTRIES];
    int n_entries, err, i, j;

    /* sanity */
    if ( validate_acl_clear_type(type) != AOK )
    {
        errno = EINVAL;
        err = ErrnoErr();
        return(err);
    }

    /* get the current acl */
    n_entries = acl(path, GETACL, MAX_ACL_ENTRIES, acl_buf);
    if ( n_entries == (-1) )
        err = ErrnoErr();
    else
    {
        /* search the current acl for the type:id */
        for ( i=0; i<n_entries; i++ )
        {
            if ( acl_match(&acl_buf[i], type, id, TRUE) == AOK )
                break;
        }

        if ( i >= n_entries )
            err = AOK;
        else /* remove type:id */
        {
            for ( j=i+1; j<n_entries; j++ )
                acl_buf[j-1] = acl_buf[j];
            n_entries--;
            if ( acl(path, SETACL, n_entries, acl_buf) == (-1) )
                err = ErrnoErr();
            else
                err = AOK;
        }
    }
    return(err);
}

/***********************************************************************/
/* convert secondary group strings to user entries                     */
/***********************************************************************/
static void
convert_grp(pid, ent, tmp_ent)
int pid;
struct group *ent;
tmp_grp_t *tmp_ent;
{
    int i, count, ix, sz;
    char **member;

    /* get name and group id */
    tmp_ent->name = (char *)TaskAlloc(pid, strlen(ent->gr_name)+1);
    strcpy(tmp_ent->name, ent->gr_name);
    tmp_ent->gid = ent->gr_gid;

    /* get list of group members */
    member = (char **)0;
    for ( i=count=ix=0; ent->gr_mem[i]; i++ )
    {
        if ( ix >= count )
        {
            count += 10;
            if ( member )
                member = (char **)TaskRealloc(pid, member, count*sizeof(char *));
            else
                member = (char **)TaskAlloc(pid, count*sizeof(char *));
        }
        member[ix] = (char *)TaskAlloc(pid, strlen(ent->gr_mem[i])+1);
        strcpy(member[ix], ent->gr_mem[i]);
        ix++;
    }
    tmp_ent->n_secondary_member_strs = ix;
    if ( member && count != tmp_ent->n_secondary_member_strs )
    {
        sz = tmp_ent->n_secondary_member_strs * sizeof(char *);
        member = (char **)TaskRealloc(pid, member, sz);
    }
    tmp_ent->secondary_member_str = member;
}

/***********************************************************************/
/* generate a list of all groups                                       */
/***********************************************************************/
static void
list_gps(pid, n_gps, gp)
int pid;
int *n_gps;
tmp_grp_t **gp;
{
    int n_ents, curr_ents, sz, ix;
    tmp_grp_t *ent_list, tmp_ent;
    struct group *ent;

    /* enumerate the groups */
    ent_list = (tmp_grp_t *)0;
    curr_ents = n_ents = 0;
    setgrent();
    while ( ent = getgrent() )
    {
        /* copy the interesting fields of the group entry */
        convert_grp(pid, ent, &tmp_ent);

        /* add the group entry to the list */
        ix = n_ents++;
        if ( n_ents > curr_ents )
        {
            curr_ents += 10;
            sz = sizeof(tmp_grp_t);
            if ( ent_list )
                ent_list = (tmp_grp_t *)TaskRealloc(pid, ent_list, curr_ents*sz);
            else
                ent_list = (tmp_grp_t *)TaskAlloc(pid, curr_ents*sz);
        }
        memcpy((char *)&(ent_list[ix]), (char *)&tmp_ent, sz);
    }
    endgrent();

    /* done */
    *n_gps = n_ents;
    if ( ent_list && n_ents != curr_ents )
        ent_list = (tmp_grp_t *)TaskRealloc(pid, ent_list, n_ents*sz);
    *gp = ent_list;
    return;
}

/***********************************************************************/
/* copy interesting fields from system password data structure         */
/***********************************************************************/
static void
convert_pwd(pid, ent, ax_ent)
int pid;
struct passwd *ent;
ax_pwd_t *ax_ent;
{
    /* login name */
    ax_ent->login_name = (char *)TaskAlloc(pid, strlen(ent->pw_name)+1);
    strcpy(ax_ent->login_name, ent->pw_name);

    /* actual name */
    if ( !ent->pw_gecos )
        ax_ent->real_name = (char *)0;
    else
    {
        ax_ent->real_name = (char *)TaskAlloc(pid, strlen(ent->pw_gecos)+1);
        strcpy(ax_ent->real_name, ent->pw_gecos);
    }

    /* user and group id */
    ax_ent->uid = ent->pw_uid;
    ax_ent->gid = ent->pw_gid;
    return;
}

/***********************************************************************/
/* copy interesting fields from system group data structure            */
/***********************************************************************/
static void
sort_pwds(pid, n_gps, gp, list)
int pid;
int n_gps;
tmp_grp_t *gp;
entry_list_t *list;
{
    int n_ents, i, j, ix, sz;
    ax_entry_t *ent;
    struct passwd *pwd;
    ax_pwd_t *member;

    /* reset pw entry pointer (and allocate system resources?) */
    setpwent();

    /* move groups into entry list */
    list->n = n_gps;
    list->element = ent = (ax_entry_t *)TaskAlloc(pid, list->n * sizeof(ax_entry_t));
    for ( i=0; i<list->n; i++ )
    {
        ent[i].type = 0;
        ent[i].u.grp.name = (char *)TaskAlloc(pid, strlen(gp[i].name)+1);
        strcpy(ent[i].u.grp.name, gp[i].name);
        ent[i].u.grp.gid = gp[i].gid;
        ent[i].u.grp.n_primary_members = 0;
        ent[i].u.grp.primary_member = (ax_pwd_t *)0;
        if ( gp[i].n_secondary_member_strs <= 0 )
        {
            ent[i].u.grp.n_secondary_members = 0;
            ent[i].u.grp.secondary_member = (ax_pwd_t *)0;
        }
        else
        {
            sz = gp[i].n_secondary_member_strs * sizeof(ax_pwd_t);
            member = (ax_pwd_t *)TaskAlloc(pid, sz);
            for ( ix=j=0; j<gp[i].n_secondary_member_strs; j++ )
            {
                pwd = getpwnam(gp[i].secondary_member_str[j]);
                if ( pwd )
                {
                    convert_pwd(pid, pwd, &member[ix]);
                    ix++;
                }
            }
            member = (ax_pwd_t *)TaskRealloc(pid, member, ix*sizeof(ax_pwd_t));
            ent[i].u.grp.n_secondary_members = ix;
            ent[i].u.grp.secondary_member = member;
        }
    }
}

/***********************************************************************/
/* attach user lists to group lists                                    */
/***********************************************************************/
static void
list_pwds(pid, n_gps, gp, list)
int pid;
int n_gps;
tmp_grp_t *gp;
entry_list_t *list;
{
    ax_pwd_t ax_ent, *member;
    struct passwd *ent;
    ax_grp_t *grp;
    int sz, ix, i;

    /* move tmp groups to permanent list, expanding passwords strings */
    sort_pwds(pid, n_gps, gp, list);

    /* enumerate the passwords */
    setpwent();
    while ( ent = getpwent() )
    {
        /* copy the interesting fields of the password entry */
        convert_pwd(pid, ent, &ax_ent);

        /* add the new password entry to the appropriate list */
        for ( i=0; i<list->n; i++ )
        {
            if ( list->element[i].type != ENTRY_IS_GRP )
                continue;
            if ( list->element[i].u.grp.gid != ax_ent.gid )
                continue;

            /* add pwd to grp */
            grp = &(list->element[i].u.grp);
            member = list->element[i].u.grp.primary_member;
            ix = grp->n_primary_members++;
            sz = sizeof(ax_pwd_t) * grp->n_primary_members;
            if ( member )
                member = (ax_pwd_t *)TaskRealloc(pid, member, sz);
            else
                member = (ax_pwd_t *)TaskAlloc(pid, sz);
            grp->primary_member = member;
            grp->primary_member[ix] = ax_ent;
            break;
        }
        if ( i >= list->n ) /* didn't find an owning group */
        {
            ix = list->n++;
            sz = sizeof(ax_entry_t) * list->n;
            if ( list->element )
                list->element = (ax_entry_t *)TaskRealloc(pid, list->element, sz);
            else
                list->element = (ax_entry_t *)TaskAlloc(pid, sz);
            list->element[ix].type = ENTRY_IS_PWD;
            list->element[ix].u.pwd = ax_ent;
        }
    }

    /* done */
    endpwent();
    return;
}

/***********************************************************************/
/* free password entry                                                 */
/***********************************************************************/
static void
free_pwd(pid, pwd)
int pid;
ax_pwd_t *pwd;
{
    if ( pwd->login_name )
        TaskFree(pid, pwd->login_name);
    if ( pwd->real_name )
        TaskFree(pid, pwd->real_name);
}

/***********************************************************************/
/* free group entry                                                    */
/***********************************************************************/
static void
free_grp(pid, grp)
int pid;
ax_grp_t *grp;
{
    int i;
    if ( grp->name )
        TaskFree(pid, grp->name);
    if ( grp->n_primary_members > 0 && grp->primary_member )
    {
        for ( i=0; i<grp->n_primary_members; i++ )
            free_pwd(pid, &grp->primary_member[i]);
        TaskFree(pid, grp->primary_member);
    }
    if ( grp->n_secondary_members > 0 && grp->secondary_member )
    {
        for ( i=0; i<grp->n_secondary_members; i++ )
            free_pwd(pid, &grp->secondary_member[i]);
        TaskFree(pid, grp->secondary_member);
    }
}

/***********************************************************************/
/* free list of groups                                                 */
/***********************************************************************/
void
AxTaskFreePasswords(pid, list)
int pid;
entry_list_t *list;
{
    int i;
    if ( list->n > 0 && list->element )
    {
        for ( i=0; i<list->n; i++ )
        {
            if ( list->element[i].type == ENTRY_IS_GRP )
                free_grp(pid, &list->element[i].u.grp);
            else
                free_pwd(pid, &list->element[i].u.pwd);
        }
        TaskFree(pid, list->element);
    }
}

/***********************************************************************/
/* create list of groups, with primary and secondary users attached    */
/***********************************************************************/
int
AxTaskCreatePasswords(pid, list)
int pid;
entry_list_t *list;
{
    int n_gps, i, j;
    tmp_grp_t *gp;

    /* can only be called one at a time */
    if ( listing_pwds )
    {
        list->n = 0;
        list->element = (ax_entry_t *)0;
        return(!AOK);
    }
    listing_pwds = TRUE;

    /* generate a list of all groups */
    list_gps(pid, &n_gps, &gp);

    /* attach password entries to groups */
    list_pwds(pid, n_gps, gp, list);

    /* done */
    for ( i=0; i<n_gps; i++ )
    {
        if ( gp[i].name )
            TaskFree(pid, gp[i].name);
        if ( gp[i].n_secondary_member_strs > 0 && gp[i].secondary_member_str )
        {
            for ( j=0; j<gp[i].n_secondary_member_strs; j++ )
            {
                if ( gp[i].secondary_member_str[j] )
                    TaskFree(pid, gp[i].secondary_member_str[j]);
            }
            TaskFree(pid, gp[i].secondary_member_str);
        }
    }
    TaskFree(pid, gp);
    listing_pwds = FALSE;
    return(AOK);
}

/***********************************************************************/
/* ELF code for reading ACL enabled flag                               */
/***********************************************************************/
data_ptr
AxfACLEnabled()
{
    data_ptr retP;
    retP = AxMakeIntData(AxACL?(-1):0);
    return(retP);
}

/***********************************************************************/
/* ELF code for setting ACL entries                                    */
/***********************************************************************/
data_ptr
AxfSetACLEntry(argDatum)
data_ptr argDatum;
{
    char *path = (char *)StrFromArgArray(argDatum,0);
    int type   = IntFromArgArray(argDatum,1);
    int id     = IntFromArgArray(argDatum,2);
    int perm   = IntFromArgArray(argDatum,3);
    int err;

    err = set_acl_entry(path, type, id, perm);
    if ( err )
        ElfStrAbort(err, NULL, path);
    return(NullDataPtr);
}

/***********************************************************************/
/* ELF code for removing ACL entries                                   */
/***********************************************************************/
data_ptr
AxfClearACLEntry(argDatum)
data_ptr argDatum;
{
    char *path = (char *)StrFromArgArray(argDatum,0);
    int type   = IntFromArgArray(argDatum,1);
    int id     = IntFromArgArray(argDatum,2);
    int err;

    err = clear_acl_entry(path, type, id);
    if ( err )
        ElfStrAbort(err, NULL, path);
    return(NullDataPtr);
}

/***********************************************************************/
/* ELF code for getting ACL                                            */
/***********************************************************************/
data_ptr
AxfGetACL(argDatum)
data_ptr argDatum;
{
    char *path = (char *)StrFromArgArray(argDatum,0);
    aclent_t acl_buf[MAX_ACL_ENTRIES];
    int i, ix, n_entries, err;
    data_ptr retP, entP;
    char str[STRPATH+1];

    /* get the current acl */
    n_entries = acl(path, GETACL, MAX_ACL_ENTRIES, acl_buf);
    if ( n_entries == (-1) )
        err = ErrnoErr();
    else
    {
        err = AOK;
        retP = AxMakeArray(n_entries);
        {
            for ( i=ix=0; i<n_entries; i++ )
            {
                if ( validate_acl_set_type(acl_buf[i].a_type) == AOK )
                {
                    entP = AxMakeArray(4);
                    AxAddIntToArray(entP, 0, acl_buf[i].a_type);
                    AxAddIntToArray(entP, 1, acl_buf[i].a_id);
                    AxAddIntToArray(entP, 2, acl_buf[i].a_perm);
                    make_acl_str(acl_buf[i].a_type, acl_buf[i].a_id, str);
                    AxAddStrToArray(entP, 3, str);

                    AxAddArrayToArray(retP, ix, entP);
                    ix++;
                }
            }
        }
    }

    /* done */
    if ( err == AOK )
        return(retP);
    else
    {
        ElfStrAbort(err, NULL, path);
        return(NullDataPtr);
    }
}

/***********************************************************************/
/* ELF code for listing a group or user                                */
/***********************************************************************/
static int
insert_pwd_entry(aclP, strP, ix, entry, nested)
data_ptr aclP, strP;
int ix;
ax_pwd_t *entry;
int nested;
{
    char str[STRPATH+1];
    data_ptr entP;

    /* entry type:id */
    entP = AxMakeArray(2);
    AxAddIntToArray(entP, 0, USER);
    AxAddIntToArray(entP, 1, entry->uid);
    AxAddArrayToArray(aclP, ix, entP);

    /* string */
    if ( nested )
    {
        if ( !entry->real_name )
            sprintf(str, "->User: %s", entry->login_name);
        else
            sprintf(str, "->User: %s (%s)", entry->login_name, entry->real_name);
    }
    else
    {
        if ( !entry->real_name )
            sprintf(str, "User: %s", entry->login_name);
        else
            sprintf(str, "User: %s (%s)", entry->login_name, entry->real_name);
    }
    AxAddStrToArray(strP, ix, str);
    ix++;
    return(ix);
}
static int
insert_list_entry(aclP, strP, ix, entry)
data_ptr aclP, strP;
int ix;
ax_entry_t *entry;
{
    char str[STRPATH+1];
    data_ptr entP;
    int i;

    if ( entry->type != ENTRY_IS_GRP )
        ix = insert_pwd_entry(aclP, strP, ix, &(entry->u.pwd), FALSE);
    else
    {
        /* entry type:id */
        entP = AxMakeArray(2);
        AxAddIntToArray(entP, 0, GROUP);
        AxAddIntToArray(entP, 1, entry->u.grp.gid);
        AxAddArrayToArray(aclP, ix, entP);

        /* string */
        sprintf(str, "Group: %s", entry->u.grp.name);
        AxAddStrToArray(strP, ix, str);
        ix++;

        /* add primary users in group */
        for ( i=0; i<entry->u.grp.n_primary_members; i++ )
        {
            ix = insert_pwd_entry(aclP, strP, ix,
                                  &entry->u.grp.primary_member[i],
                                  TRUE);
        }

        /* add secondary users in group */
        for ( i=0; i<entry->u.grp.n_secondary_members; i++ )
        {
            ix = insert_pwd_entry(aclP, strP, ix,
                                  &entry->u.grp.secondary_member[i],
                                  TRUE);
        }
    }
    return(ix);
}

/***********************************************************************/
/* ELF code for listing ALL groups and users                           */
/***********************************************************************/
data_ptr
AxfListGroupsAndUsers(argDatum)
data_ptr argDatum;
{
    int stat, i, ix, err, count;
    data_ptr retP, aclP, strP;
    entry_list_t list;

    stat = AxTaskCreatePasswords(THIMpid, &list);
    if ( stat == (-1) || list.n <= 0 )
    {
        errno = EIO;
        err = ErrnoErr();
    }
    else
    {
        /* get the number of entries */
        for ( i=count=0; i<list.n; i++ )
        {
            count++;
            if ( list.element[i].type == ENTRY_IS_GRP )
            {
                count += list.element[i].u.grp.n_primary_members;
                count += list.element[i].u.grp.n_secondary_members;
            }
        }

        /* create ELF data */
        retP = AxMakeArray(2);
        aclP = AxMakeArray(count);
        strP = AxMakeArray(count);
        for ( i=ix=0; i<list.n; i++ )
            ix = insert_list_entry(aclP, strP, ix, &list.element[i]);
        AxAddArrayToArray(retP, 0, aclP);
        AxAddArrayToArray(retP, 1, strP);

        /* free local resources */
        AxTaskFreePasswords(THIMpid, &list);
        err = AOK;
    }

    if ( err == AOK )
        return(retP);
    else
    {
        ElfStrAbort(err, NULL, XLT("Group and password lists unavailable"));
        return(NullDataPtr);
    }
}
#else /* } Don't use ACLs { */
int
AxReadACLState()
{
    return(FALSE);
}
data_ptr
AxfACLEnabled()
{
    data_ptr retP;
    retP = AxMakeIntData(FALSE);
    return(retP);
}
data_ptr
AxfSetACLEntry(argDatum)
data_ptr argDatum;
{
    return(NullDataPtr);
}
data_ptr
AxfClearACLEntry(argDatum)
data_ptr argDatum;
{
    return(NullDataPtr);
}
data_ptr
AxfGetACL(argDatum)
data_ptr argDatum;
{
    return(NullDataPtr);
}
data_ptr
AxfListGroupsAndUsers(argDatum)
data_ptr argDatum;
{
    return(NullDataPtr);
}
#endif /* } Don't use ACL's */

/***********************************************************************/
/***********************************************************************/
