/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axio.c
   Applix ASC version: 72
   Date: Sat Apr 17 21:15:34 1999
*/
#if SUN3 || SYSV4
#include <unistd.h>
#endif    /*  defines for ?_OK  used with access command  -lance  */


#define DOS_SLASH '\\'
#define UNIX_SLASH '/'
#if AX_NT
#define SLASH_CHAR DOS_SLASH
#else
#define SLASH_CHAR UNIX_SLASH
#endif

#include <stdio.h>
#include <errno.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/stat.h>
#if AX_NT
#include <fcntl.h>
#endif

#if PMAX
#include <Alib.h>
#endif

#if HP800
#include <audio/Alib.h>
#if BUILD_KANJI
#include <sys/utsname.h>
#endif
#endif

#include <aos.h>
#include <afile.h>
#include <adocio.h>
#include <fsrts.h>
#include <getstr10.h>
#include <paripc.h>

#if AX_NT
#undef Status      /* for now! */
#endif  /* DAMNT */

#include <thimble.h>
#include <adlgpoke.h>
#include <axproto.h>


#define TASKASSERT(x,y) if ( !(x) ) taskAbort(QuickGetStr(y))

#define MAGIC_STRING    "2278791755874641370"
#define RESERVED_STR    "00000000000000000000"

#define REV1_GEN_INFO   "0000"      /* generic info, portable */
#define REV2_GEN_INFO   "1000"      /* generic info, portable */

#define HEADER_LINE_LEN 78  /* includes the newline char at end */
#define FPOS_SYSREV 21
#define FPOS_DOCTYP 26
#define FPOS_APPREV 31
#define FPOS_FILCNT 52
#define FPOS_RSVD   57

#define MAX_OPEN_SEGS   200 /* Maximum dircetly accessable file segments */
#define FILE_ABSURD 100000000   /* 100 Mbyte file */
#define MAX_TOFC_ENTRIES 1000   /* max number of file objects */

#if !AXMS
/* NT declares this is stdlib.h */
#if !LINUX
extern char *sys_errlist[];
#endif
#endif
extern char *ThimUserBackupDir;
extern char *DirSlash;
extern short THIMpid;

tofcEntry *DocSegPtrs[MAX_OPEN_SEGS];   /* FD->info table */
int DocSegTaskOpeners[MAX_OPEN_SEGS];   /* task that opened the FD */

int AxIOblocks = 8100;  /* IO block size (should not be exact power of 2) */
char *TaskAlloc(), *TaskRealloc(), *TaskCalloc();

/* FORWARD REFS */

static typeErr ReadFileObject();
static Logger();
char *AxBufGets AXP_((
                    char *buf,          /* result buffer */
                    int size,   /* size of buf including room for null-term */
                    char *inBuf,        /* buffer from AxReadFileIntoTaskBuffer */
                    int *bufPosP,       /* IN/OUT buffer position */
                    int bufSize         /* total size of buf */
                    ));
char *AxReadFileIntoTaskBuf AXP_((
                    int task,
                    char *filename,
                    int *sizeP
                    ));
static void makeBackupFileName AXP_((char *filename, char *backupname));
static docFileData *CreateNewDP();
typeErr CheckAudioWare();

#define ALC_STR(task,p) strcpy(TaskAlloc(task,strlen(p)+1),p)

extern int AxDocFromPriorRev;   /* Set for rev1 docs on file opens */
extern int AxDocInBinaryForm;   /* Set if machineType not 1000 or 0000 */
int SupportsAudio=FALSE;

char *ThimSfxFromType();

static DeleteFileObjByDP (/* register docFileData *dp;int docpos; */); 
/*************************************************************************
function: AxClearVirtualFiles - Close any of task's fds
*/
AxClearVirtualFiles (pid)
register int pid;
{
    register int ix;

    for ( ix = 0; ix < MAX_OPEN_SEGS; ix++ ) {
        if ( DocSegTaskOpeners[ix] == pid ) {
            DocSegTaskOpeners[ix] = 0;
            DocSegPtrs[ix] = NULL;
        }
    }
}

/*************************************************************************
function: OpenDocumentFile - Load file into memory
Note that this does NOT actually create the file, unless so told.
It just creates a writable structure.  The file itself is created
when the output buffer overflows or the file is closed.
*/
typeErr OpenDocumentFile (task,id,name,mode,typeP,revstampP,createdP,fileP)
int task;       /* Normally THIMpid */
char *name;     /* name of file to open */
long id;        /* IN doc_id to assciate with the file */
int mode;       /* IN DFM_READ... */
int *typeP;     /* INOUT type of doc */
char **revstampP;   /* INOUT revstamp */
int *createdP;  /* OUT TRUE if file does not already exist */
docFileData **fileP;
{
#define OPEN_FAIL(x) {err = x; goto fail;}

    register docFileData *dp = NULL;
    register tofcEntry *sp;
    int fd = -1;        /* file desc */
    char truepath[STRPATH+1];
    typeErr err;        /* returned error code */
    int code, bytes_left, rqst, cnt;
    int num_files;  /* number of entries in tofc */
    char hdr[HEADER_LINE_LEN];
    register int ix, blockno;

    int size, rsvd;
    char str[71], entryname[72];
    int fileOk = FALSE;         /* Used by ascii editor */
    int triedAsFile = FALSE;    /* flag while attempting to open as AED file */
    int typeGotten;
    char *hdrMachine;

/* open file & create docFileData struct */

    fileOk = mode & DFM_FILEOK;
    strcpy(truepath,name);
    AxSysDosUnixPath(truepath);
    AXLOG(("OpenDocumentFile '%s'\n",truepath));
    fd = open(truepath,O_RDONLY+O_BINARY);
    AXLOG(("OpenDocumentFile fd %d\n",fd));
    if ( fd == -1 ) {       /* file opened OK */
        if ( !(DFM_CREATE & mode) || (DFM_ACCESS & mode) || (errno != ENOENT) ){
            err = ErrnoErr();
            goto do_fail;
        }
        *createdP = TRUE;
        dp = CreateNewDP(task,truepath,mode,id,0);
        strcpy(dp->revstamp,*revstampP);
        *fileP = dp;
        return(AOK);
    }
    if ( DFM_FRESH & mode ) /* Must be freshly created? */
        OPEN_FAIL(FEEEXIST);

    if ( DFM_ACCESS & mode )    /* access check? just return AOK now */
        OPEN_FAIL(AOK);

/* read header */

    *createdP = FALSE;
    code = read(fd,hdr,HEADER_LINE_LEN);
    if ( code == -1 )
        OPEN_FAIL(ErrnoErr());

    if ( code != HEADER_LINE_LEN )
        OPEN_FAIL(FEnotDoc);    /* generic i/o error */

/* null term each header segment */

    hdr[FPOS_SYSREV-1] = hdr[FPOS_DOCTYP-1] = hdr[FPOS_APPREV-1] =
            hdr[FPOS_FILCNT-1] = hdr[FPOS_RSVD-1] =
            hdr[HEADER_LINE_LEN-1] = 0;

/* validate header */

    typeGotten = atoi(&hdr[FPOS_SYSREV-2]);
    if ( (typeGotten != 0) && (*typeP != typeGotten) )
        OPEN_FAIL(FEnotRightSfx);
    hdr[FPOS_SYSREV-2] = 0;
    if ( !streq(MAGIC_STRING,&hdr[0]) || !streq(RESERVED_STR,&hdr[FPOS_RSVD]) )
        OPEN_FAIL(FEnotDoc);

    hdrMachine = &hdr[FPOS_SYSREV];
    if ( !streq(REV2_GEN_INFO+1,hdrMachine+1) && !streq(SYS_INFO+1,hdrMachine+1) )
        OPEN_FAIL(FEbadMachine);
    AxDocFromPriorRev = FALSE;
    AxDocInBinaryForm = FALSE;
    if ( !streq(REV2_GEN_INFO,hdrMachine) && !streq(SYS_INFO,hdrMachine) )
        AxDocFromPriorRev = TRUE;
    if ( !streq(REV2_GEN_INFO+1,hdrMachine+1) )
        AxDocInBinaryForm = TRUE;

    typeGotten = atoi(&hdr[FPOS_DOCTYP]);
    if ( *typeP != typeGotten )
        OPEN_FAIL(FEnotRightSfx);

    num_files = atoi(&hdr[FPOS_FILCNT]);
    dp = CreateNewDP(task,truepath,mode,id,num_files);
    strcpy(dp->revstamp,&hdr[FPOS_APPREV]);
    strcpy(*revstampP,dp->revstamp);
    *fileP = dp;
    fileOk = FALSE;     /* if valid header, MUST be a document */

/* Read in each TOFC line */

    for ( ix = 0; ix < num_files; ix++ ) {
        code = read(fd,str,71);     /* read TOFC info */
        if ( code != 71 )
            OPEN_FAIL(FEnotDoc);
        if ( str[70] != '\n' )
            OPEN_FAIL(FEnotDoc);
        sscanf(str,"%d %s %d",&size,entryname,&rsvd);
        if ( (strlen(entryname) > 50) || (strlen(entryname) == 0) ||
            (size < 0) || (size > FILE_ABSURD) ||
            (rsvd != 0) )
            OPEN_FAIL(FEnotDoc);

        sp = (tofcEntry*)TaskCalloc(dp->task,1,sizeof(tofcEntry));
        sp->num_blocks = (size+AxIOblocks-1)/ AxIOblocks;
        sp->blocks = (char**)TaskAlloc(dp->task,sp->num_blocks * sizeof(char*));
        dp->segP[ix] = sp;
        sp->name = ALC_STR(dp->task,entryname);
        sp->size = size;
    }

/* read in each file object */

read_files:
    for ( ix = 0; ix < num_files; ix++ ) {
        sp = dp->segP[ix];
        bytes_left = sp->size;
        for ( blockno = 0; blockno < sp->num_blocks; blockno++ ) {
            rqst = MIN(AxIOblocks,bytes_left);
            sp->blocks[blockno] = TaskAlloc(dp->task,rqst);
            errno = 0;
            cnt = read(fd,sp->blocks[blockno],rqst);
            if ( cnt == -1 )
                OPEN_FAIL(ErrnoErr());
            if ( cnt != rqst ) {
                if ( errno != 0 )
                    OPEN_FAIL(ErrnoErr());
                OPEN_FAIL(FEbadIO); /* generic IO error */
            }
            bytes_left -= cnt;
        }
    }
    close(fd);
    return(AOK);

fail:
    if ( (fd > 0) && fileOk && !triedAsFile ) {
        struct stat status;
        triedAsFile = TRUE;
        if ( -1 == lseek(fd,0L,0) )
            OPEN_FAIL(ErrnoErr());
        if ( -1 == stat(truepath,&status) )
            return(ErrnoErr());  /* no such dir! */
        size = status.st_size;
        dp = CreateNewDP(task,truepath,mode,id,1);
        strcpy(dp->revstamp,*revstampP);
        *fileP = dp;
        sp = (tofcEntry*)TaskCalloc(dp->task,1,sizeof(tofcEntry));
        sp->num_blocks = (size+AxIOblocks-1)/ AxIOblocks;
        sp->blocks = (char**)TaskAlloc(dp->task,sp->num_blocks * sizeof(char*));
        dp->segP[0] = sp;
        if ( (*typeP == MUserCommand) || (*typeP == MNone) )
            sp->name = ALC_STR(dp->task,"1.elf");   /* HACK */
        else {
            sp->name = ALC_STR(dp->task,"1.mm");    /* HACK */
        }
        sp->size = size;
        num_files = 1;
        goto read_files;
    }
do_fail:
    if ( fd != -1 )
        close(fd);
    if ( dp )
        AbandonDocumentFile(dp);
    *fileP = NULL;
    return(err);
}

/*************************************************************************
function: CreateNewDP - Alloc a DP
*/
static docFileData *CreateNewDP (task,truename,mode,id,subfiles)
int task;       /* normally THIMpid */
char *truename;
int mode;
long id;
int subfiles;   /* number of file objects to make room for */
{
    register docFileData *dp;

    dp = (docFileData*)TaskCalloc(task,1,sizeof(docFileData));
    dp->segP = (tofcEntry**)TaskCalloc(task,1,(1+subfiles) * sizeof(tofcEntry*));
    dp->name = ALC_STR(task,truename);
    dp->mode = mode;
    dp->id = id;
    dp->task = task;
    return(dp);
}

/*************************************************************************
function: FilePermsFromDocP - Return "chmod" value
*/
FilePermsFromDocP (DocP,pathname)
DocInfoType *DocP;  /* access modes used from this */
char *pathname;     /* override path (nominally DocP->pathname) */
{
    return(AxChmodValueFromCodes(pathname,DocP->grpAccess,DocP->allAccess));
}


/*************************************************************************
function: SaveDocumentFile - Save file, transferring old stuff to new file
*/
typeErr SaveDocumentFile (DocP, revID)
DocInfoType *DocP;
int revID;
{
#define VWRITE(fd,p,n) {cnt = write(fd,p,n); if (cnt != n) goto ferr;}

    register docFileData *dp;
    tofcEntry *sp;
    char tempname[STRPATH+20];
    char junk[STRPATH+2];
    char hdr[HEADER_LINE_LEN];
    int ix, cnt, rqst, num_files, blockno;
    long bytes_left;
    int fd;
    int mode;
    int is_binary;

/* open file */

    dp = DocP->docFileId;
    ParsePath(dp->name,tempname,junk);
    strcat(tempname,"/tmpXXXXXX");
    mktemp(tempname);
    mode = FilePermsFromDocP(DocP,dp->name);
    fd = open(tempname,O_CREAT+O_WRONLY+O_EXCL+O_BINARY,mode);
    if ( fd == -1 )
        return(ErrnoErr());


    if ( DocP->DocMatType == MNone )
        goto write_blocks;

/* write header */

/* Count number of files, use hueristic hack to determine whether
    this is a binary form write */

    num_files = 0;
    is_binary = FALSE;
    for ( ix = 0; dp->segP[ix]; ix++ ) {
        char *p;
        if ( !dp->segP[ix]->orphaned ) {
            num_files++;
            p = dp->segP[ix]->name;
            if ( streq(p,"1.elfo") || streq(".sw",p+strlen(p)-3) ||
                 streq(p,"1.elfx") )
                is_binary = TRUE;
        }
    }

    if ( revID == 1 )
    {
        sprintf(hdr,"%s%d %s %4d %20s %4d %s\n",
                MAGIC_STRING,
                DocP->DocMatType,
                is_binary ? SYS_INFO : REV1_GEN_INFO,
                DocP->DocMatType,
                dp->revstamp,
                num_files,
                RESERVED_STR);
    }
    else
    {
        sprintf(hdr,"%s%d %s %4d %20s %4d %s\n",
                MAGIC_STRING,
                DocP->DocMatType,
                is_binary ? SYS_INFO : REV2_GEN_INFO,
                DocP->DocMatType,
                dp->revstamp,
                num_files,
                RESERVED_STR);
    }
    rqst = HEADER_LINE_LEN;
    VWRITE(fd,hdr,rqst);

/* write the table of contents */

    for ( ix = 0; sp = dp->segP[ix]; ix++ ) {
        char str[72];
        if ( sp->orphaned ) continue;
        sprintf(str,"%8d %50s %10d\n",sp->size,sp->name,0);
        rqst = 71;
        VWRITE(fd,str,rqst);
    }

/* write the file objects */

write_blocks:
    for ( ix = 0; sp = dp->segP[ix]; ix++ ) {
        if ( sp->orphaned ) continue;
        if ( (DocP->DocMatType == MNone) && !streq(sp->name,"1.elf") )
            continue;
        bytes_left = sp->size;
        for ( blockno = 0; bytes_left; blockno++ ) {
            rqst = MIN(AxIOblocks,bytes_left);
            VWRITE(fd,sp->blocks[blockno],rqst);
            bytes_left -= rqst;
        }
    }
    if ( -1 == close(fd) )
        goto ferr;

/* rename tempfile to true file */

    if ( -1 == Arename(tempname,dp->name) ) {
        unlink(tempname);
        return(ErrnoErr());
    }
    else {
        chmod(dp->name,mode);
        TmPokeDialogs(0, POKE_WRITE_FILE,dp->name);
        return(AOK);
    }

ferr:
    unlink(tempname);
    close(fd);
    return(ErrnoErr());
}



/*************************************************************************
function: FSWriteFileObjAsFile - Write a file object out to a true file
*/
typeErr FSWriteFileObjAsFile (filename,fmode,docP,objname)
char *filename;     /* result file name */
int fmode;          /* chmod of result file */
DocInfoType *docP;
char *objname;
{
    int created, fd, ofd;
    long size, cnt;
    char buf[8096];
    typeErr err;

    ofd = creat(filename,fmode);
    if ( ofd == -1 )
        return(ErrnoErr());
    fd = OpenFileObject(docP,objname,DFM_READ,&created,&size);
    if ( fd == -1 )
        return(ErrnoErr());
    while ( cnt = DocSegRead(fd,buf,8096,-1) ) {
        if ( cnt != write(ofd,buf,cnt) ) {
            err = ErrnoErr();
            close(ofd);
fail:       unlink(filename);
            CloseFileObject(fd);
            return(err);
        }
    }
    if ( -1 == close(ofd) ) {
        err = ErrnoErr();
        goto fail;
    }
    CloseFileObject(fd);
    return(AOK);
}

/*************************************************************************
function: FSBufferFromFileObject - Allocate a buffer and copy file object into it
*/
char *FSBufferFromFileObject (task,docP,objname, size)
int task;   /* IN */
DocInfoType *docP;  /* IN */
char *objname;  /* IN */
int *size;  /* OUT */
{
    int fd;
    int created;
    char *retP;

    fd = OpenFileObject(docP,objname,DFM_READ,&created, size);
    if ( fd == -1 )
        return(NULL);
    retP = TaskAlloc(task, *size);
    DocSegRead(fd,retP, *size,-1);
    return(retP);
}

/*************************************************************************
function: FSWriteBufAsDoc - Write a memory buffer as a full doc
*/
typeErr FSWriteBufAsDoc (filename,fmode,MatType,buf,size)
char *filename;     /* result file name... must end with .aw/.ag/.as... */
int fmode;          /* chmod of result file */
int MatType;        /* MPicture... */
char *buf;          /* the buffer */
int size;           /* bytes in buffer */
{
    int ofd;
    char revstamp[21];
    char str[72];
    char hdr[HEADER_LINE_LEN];
    typeErr err;
    char *objname = NULL;

    ofd = creat(filename,fmode);
    if ( ofd == -1 )
        return(ErrnoErr());
/*
 * build the doc header
 */
    strcpy(revstamp,"12345678901234567890");

    sprintf(hdr,"%s%d %s %4d %20s %4d %s\n",
            MAGIC_STRING,MatType, REV2_GEN_INFO, MatType,
                        revstamp,2,RESERVED_STR);
    write(ofd,hdr,HEADER_LINE_LEN);

/*
 * write out the table of contents
 */

    switch ( MatType ) {
    case MSpreadsheet:
        objname = NXLT("1.sa");
        break; 
    case MDrawing:
        objname = NXLT("1.gc");
        break;
    default:
        tempAbort(NXLT("nonsense in FSWriteBufAsDoc"));
    }
    sprintf(str,"%8d %50s %10d\n",0,"doc.info",0);
    write(ofd,str,71);
    sprintf(str,"%8d %50s %10d\n",size,objname,0);
    write(ofd,str,71);


    if ( size != write(ofd,buf,size) ) {
        err = ErrnoErr();
        close(ofd);
fail:   unlink(filename);
        return(err);
    }
    if ( -1 == close(ofd) ) {
        err = ErrnoErr();
        goto fail;
    }
    return(AOK);
}

/*************************************************************************
function: FSWriteFileObjAsDoc - Write a file object out as a document
*/
typeErr FSWriteFileObjAsDoc (filename,fmode,docP,objname)
char *filename;     /* result file name... must end with .aw/.ag/.as... */
int fmode;          /* chmod of result file */
DocInfoType *docP;
char *objname;
{
    int created, fd, ofd;
    int MatType;
    long size, cnt;
    char buf[8096];
    char revstamp[21];
    char str[72];
    char hdr[HEADER_LINE_LEN];
    typeErr err;

    ofd = creat(filename,fmode);
    if ( ofd == -1 )
        return(ErrnoErr());
    fd = OpenFileObject(docP,objname,DFM_READ,&created,&size);
    if ( fd == -1 )
        return(ErrnoErr());
/*
 * build the doc header
 */
    strcpy(revstamp,"12345678901234567890");

    MatType = DocTypeFromName(filename);
    sprintf(hdr,"%s%d %s %4d %20s %4d %s\n",
            MAGIC_STRING,MatType, REV2_GEN_INFO, MatType,
                        revstamp,2,RESERVED_STR);
    write(ofd,hdr,HEADER_LINE_LEN);

/*
 * write out the table of contents
 */

    sprintf(str,"%8d %50s %10d\n",0,"doc.info",0);
    write(ofd,str,71);
    sprintf(str,"%8d %50s %10d\n",size,strrchr(objname,'/')+1,0);
    write(ofd,str,71);


    while ( cnt = DocSegRead(fd,buf,8096,-1) ) {
        if ( cnt != write(ofd,buf,cnt) ) {
            err = ErrnoErr();
            close(ofd);
fail:       unlink(filename);
            CloseFileObject(fd);
            return(err);
        }
    }
    if ( -1 == close(ofd) ) {
        err = ErrnoErr();
        goto fail;
    }
    CloseFileObject(fd);
    return(AOK);
}

/*************************************************************************
function: FSHackedUpWriteFileObjAsDoc - Write a file object out as a document
*/
typeErr FSHackedUpWriteFileObjAsDoc (filename,fmode,docP,objname)
char *filename;     /* result file name */
int fmode;          /* chmod of result file */
DocInfoType *docP;
char *objname;
{
    int created, fd, ofd;
    int MatType, num_files, xcnt;
    long size, cnt;
    char buf[8096];
    char revstamp[21];
    char str[72];
    char fname[128];
    char hdrent[128];
    char hdr[HEADER_LINE_LEN];
    char newextn[4];
    char *p, *extn, *hdrnam;
    typeErr err;

    ofd = creat(filename,fmode);
    if ( ofd == -1 )
        return(ErrnoErr());
    fd = OpenFileObject(docP,objname,DFM_READ,&created,&size);
    if ( fd == -1 )
        return(ErrnoErr());
/*
 * build the doc header
 */
    xcnt = size;
    num_files = 2;
    strcpy(revstamp,"12345678901234567890");

/*
 * determine the new doc extension and read MatType
 */
    strcpy(fname,filename);
    p = strchr(fname,'.');
    extn = p+1;

    if (streq(extn,"mm"))
    {
        strcpy(newextn,".aw");
        MatType = MText;
    }
    else if (streq(extn,"sa"))
    {
        strcpy(newextn,".as");
        MatType = MSpreadsheet;
    }
    else if (streq(extn,"gc") || streq(extn,"bc"))
    {
        strcpy(newextn,".ag");
        MatType = MDrawing;
    }
    else
    {
        sprintf(newextn,".%s",extn);
        MatType = MUserCommand;
    }

    *p = 0;
    strcat(fname,newextn);

     sprintf(hdr,"%s%d %s %4d %20s %4d %s\n",MAGIC_STRING,MatType, REV2_GEN_INFO, MatType,
                        revstamp,num_files,RESERVED_STR);
    write(ofd,hdr,HEADER_LINE_LEN);

/*
 * write out the table of contents
 */

    strcpy(hdrent,filename);
    p = strrchr(hdrent,'/');
    hdrnam = p+1;

    sprintf(str,"%8d %50s %10d\n",0,"doc.info",0);
    write(ofd,str,71);
    sprintf(str,"%8d %50s %10d\n",xcnt,hdrnam,0);
    write(ofd,str,71);


    while ( cnt = DocSegRead(fd,buf,8096,-1) ) {
        if ( cnt != write(ofd,buf,cnt) ) {
            err = ErrnoErr();
            close(ofd);
fail:       unlink(filename);
            CloseFileObject(fd);
            return(err);
        }
    }
    if ( -1 == close(ofd) ) {
        err = ErrnoErr();
        goto fail;
    }
    CloseFileObject(fd);

/*
 * rename the file as a document
 */
    Arename(filename, fname);

    return(AOK);
}

/*************************************************************************
function: FSReadFileIntoFileObj - Read a unix file in as a file object
*/
typeErr FSReadFileIntoFileObj (filename,docP,objname)
char *filename;     /* disk file to read */
DocInfoType *docP;
char *objname;
{
    int created, fd, ifd;
    long size, cnt;
    char buf[8096];
    typeErr err;

    ifd = open(filename,O_RDONLY+O_BINARY);
    if ( ifd == -1 )
        return(ErrnoErr());
    fd = OpenFileObject(docP,objname,DFM_CREATE+DFM_FRESH,&created,&size);
    if ( fd == -1 )
        return(ErrnoErr());
    while ( cnt = read(ifd,buf,8096) ) {
        if ( cnt == -1 ) {
            err = ErrnoErr();
            close(ifd);
            CloseFileObject(fd);
            DeleteFileObject(docP,filename);
            return(err);
        }
        DocSegWrite(fd,buf,cnt,-1);
    }
    close(ifd);
    CloseFileObject(fd);
    return(AOK);
}

/*************************************************************************
function: RenameDocumentFile - Chnage the name entry
*/
RenameDocumentFile (DocP,name)
DocInfoType *DocP;
char *name;
{
    register docFileData *dp;

    dp = DocP->docFileId;
    if(!dp)
        return;
    TaskFree(dp->task,dp->name);
    dp->name = ALC_STR(dp->task,name);
}

/*************************************************************************
function: FSCopyFileObject - Copy virtual file from one doc to another
*/
typeErr FSCopyFileObject (dstDocP,dst,srcDocP,src)
DocInfoType *dstDocP, *srcDocP;
char *dst, *src;
{
    long crt,size;
    char buf[1000];
    int dfd,sfd,n;

    DeleteFileObject(dstDocP,dst);
    dfd = OpenFileObject(dstDocP,dst,DFM_OWNER+DFM_CREATE+DFM_FRESH,&crt,&size);
    if ( dfd <= 0 )
        return(ErrnoErr());
    sfd = OpenFileObject(srcDocP,src,DFM_READ,&crt,&size);
    if ( sfd <= 0 ) {
        CloseFileObject(dfd);
        return(ErrnoErr());
    }
    while ( n = DocSegRead(sfd,buf,1000,-1) )
        DocSegWrite(dfd,buf,n,-1);
    CloseFileObject(dfd);
    CloseFileObject(sfd);
    return(AOK);
}

/*************************************************************************
function: FSCopyKSObject - Copy kitchen sink object from one doc to another
*/
FSCopyKSObject(dstDocP,srcDocP)
DocInfoType *dstDocP,*srcDocP;
{
    register ks_entry *sks = srcDocP->ksData;

    AbandonKSObject(dstDocP);
    while ( sks ) {
        FSKSAdd(dstDocP,sks->name,sks->value);
        sks = sks->next;
    }
    return(0);
}

/*************************************************************************
function: AbandonDocumentFile - Called by AbandonTaskDocuments
*/
AbandonDocumentFile (dp)
docFileData *dp;
{
    int count, ix;

    if ( !dp )
        return;

/* count number of contained file objects */

    for ( count = 0; dp->segP[count]; count++ );

/* for each, delete first, next becomes first */

    for ( ix = 0; ix < count; ix++ )
        DeleteFileObjByDP(dp,0);
}

/*************************************************************************
function: FindFileStruct - Utility to find file object struct in doc struct
Returns NULL and sets errno on failure.
*/
static FindFileStruct (dp,pathname)
register docFileData *dp;
register char *pathname;        /* segment name */
{
    register int ix;
    register tofcEntry *sp;
    char name[STRPATH+1];

    if ( !dp ) {
fail:   errno = ENOENT;
        return(-1);
    }
    GetFileObjectName(dp->name,pathname,name);
    for ( sp = dp->segP[0], ix = 0; sp = dp->segP[ix]; ix++ )
        if ( streq(sp->name,name) )
            break;
    if ( !sp )
        goto fail;
    return(ix);
}

/*************************************************************************
function: GetFileObjectName - Extract object name piece
If the pathname begins with "*", the portion after the "*" is the object
name piece.
*/
GetFileObjectName (docname,pathname,objname)
register char *docname;     /* pathname of containing doc */
register char *pathname;        /* pathname of file object */
register char *objname;     /* entryname of file object */
{
    register int len;
    typeErr FSGDInfoGivenID();

    if ( *pathname == '?' ) {   /* embeded file, produce expanded name */
        register char *p;
        char idstr[10];
        typeErr err;
        typeErr FSXPathOfDocId();
        char pathnameBuf[STRPATH+100];

        pathname++;
        p = strchr(pathname,'?');   /* find matching hook */
        if ( !p || (p - pathname > 7) )
            goto punt;
        memset(idstr,0,10); /* FORCES NULL term string !! */
        strncpy(idstr,pathname,p-pathname);
        err = FSXPathOfDocId(atoi(idstr),pathnameBuf);
        if ( err != AOK )
            goto punt;
        strcat(pathnameBuf,p+1);
        pathname = pathnameBuf;
        if ( strchr(pathname,'?') )
            taskAbort(NXLT("document container error"));
    }

    if ( *pathname == '*' ) {   /* never encountered? - historical? */
        strcpy(objname,pathname+1);
        return;
    }

/* make sure pathname within docname */

    len = strlen(docname);
    if ( (len >= strlen(pathname)) || strncmp(pathname,docname,len) ) 
        {
#if 0       /* ??x debug! */
        char buf[STRPATH*3];
        sprintf(buf,NXLT("%s not in \n%s"),pathname,docname);
        SuperMessage(buf);
#endif
punt:   
        /* 
         * in the case of the pathname & docname 
         * are NOT the same (one may be a LINK to the other) 
         *sso at least try returning the objname
         */
        {
        char dirname[2*STRPATH+1];
        ParsePath(pathname,dirname,objname); 
        }
        return;
    }
/* return portion after the docname */
    strcpy(objname,pathname+len+1); /* +1 to get past the slash */
}

/*************************************************************************
function: OpenFileObject - Open a document file segment
If DFM_CREATE, creates a new file if one does not already exist.
If DFM_FRESH, deletes old and creates a new one (DFM_CREATE must be true).
Note that DFM_CREATE without DFM_FRESH is sort of silly (just deletes).
*/
OpenFileObject (DocP, pathname, mode, createdP, sizeP)
DocInfoType *DocP;
register char *pathname;        /* segment name */
register int mode;      /* DFM_... */
register int *createdP; /* TRUE returned TRUE if file object has been created */
register long *sizeP;   /* returned initial size */
{
    register docFileData *dp;
    register tofcEntry *sp;
    register int ix;
    char name[2*STRPATH+1];

    dp = DocP->docFileId;
    if ( !dp )
        goto fail;
    GetFileObjectName(dp->name,pathname,name);
start_over:
    *createdP = FALSE;      /* assumption */

/* find file file object struct if it exists */

    for ( sp = dp->segP[0], ix = 0; sp = dp->segP[ix]; ix++ )
        if ( streq(sp->name,name) )
            break;
set_fd:             /* assign file desc & return */
    if ( sp ) {
        sp->task = dp->task;
        if ( mode & DFM_FRESH ) {
            mode &= ~DFM_FRESH;
            DeleteFileObject(DocP,pathname);
            goto start_over;
        }
        for ( ix = 1; (ix < MAX_OPEN_SEGS) && DocSegPtrs[ix]; ix++ );
        *sizeP = sp->size;
        if ( ix == MAX_OPEN_SEGS ) {
            errno = EMFILE;
            return(-1);
        }
        DocSegPtrs[ix] = sp;
        DocSegTaskOpeners[ix] = THIMpid;
        sp->filepos = 0;
        return(ix);
    }

/* If create mode, realloc to extend doc file info, create and init the
    entry info. */

    else if ( mode & DFM_CREATE ) {
        if ( mode & DFM_FRESH )
            mode &= ~DFM_FRESH;
        *createdP = TRUE;
        dp->segP = (tofcEntry**)
                        TaskRealloc(dp->task,dp->segP,(ix+2)*sizeof(tofcEntry*));
        dp->segP[ix] = sp = (tofcEntry*)TaskCalloc(dp->task,1,sizeof(tofcEntry));
        dp->segP[ix+1] = 0; /* re-null term */
        sp->task = dp->task;
        sp->blocks = (char**)TaskAlloc(dp->task,0);
        sp->name = ALC_STR(dp->task,name);
        goto set_fd;
    }
    else {
fail:   errno = ENOENT;
        return(-1);
    }
}

/*************************************************************************
function: CloseFileObject - Shut virtual file desc
*/
CloseFileObject (fd)
int fd;
{
    DocSegPtrs[fd] = NULL;
}

/*************************************************************************
function: OrphanFileObject - Remove accessability, but leave data in memory
Retaining its name but orphaning it means that open apps see the data, but
the data is not stored at file save time.  This feature is needed for GE
image files when saving (but not closing) to machine independent format so
that later image references are still found.
*/
OrphanFileObject (DocP,name,make_nameless)
DocInfoType *DocP;
char *name; /* name of file object */
int make_nameless;
{
    register docFileData *dp;
    register int docpos;

    dp = DocP->docFileId;
    if ( -1 == (docpos = FindFileStruct(dp,name)) )
        return(-1);
    if ( dp->segP[docpos]->orphaned && !make_nameless )
        return(-1);
    dp->segP[docpos]->orphaned = TRUE;
    if ( make_nameless )
        dp->segP[docpos]->name[0] = 0;
    return(0);
}

/*************************************************************************
function: AdoptOrphanedFileObject - Restore accessability removed above
*/
AdoptOrphanedFileObject(DocP, name)
DocInfoType *DocP;
char *name;
{
    register docFileData *dp;
    register int docpos;

    dp = DocP->docFileId;
    if ( -1 == (docpos = FindFileStruct(dp,name)) )
        return(-1);
    dp->segP[docpos]->orphaned = FALSE;
    return(0);
}

/*************************************************************************
function: RenameFileObject - Change referenced name
*/
RenameFileObject (DocP,src,dst)
DocInfoType *DocP;
char *src, *dst;
{
    register docFileData *dp;
    register tofcEntry *sp;
    register int docpos;
    char objname[STRPATH+1];

    DeleteFileObject(DocP,dst);
    dp = DocP->docFileId;
    if ( -1 == (docpos = FindFileStruct(dp,src)) )
        return(-1);
    sp = dp->segP[docpos];
    sp->orphaned = FALSE;   /* renaming orphaned file makes it visible */
    TaskFree(dp->task,sp->name);
    GetFileObjectName(dp->name,dst,objname);
    sp->name = ALC_STR(dp->task,objname);
    return(0);
}

/*************************************************************************
function: ScanForNthFileObject - Return name of Nth file or NULL
*/
char *ScanForNthFileObject (DocP,n)
DocInfoType *DocP;
int n;
{
    register int ix;
    register docFileData *dp;

    dp = DocP->docFileId;
    if (!dp) return(NULL);
    for ( ix = 0; (ix != n) && dp->segP[ix]; ix++ );
    if ( dp->segP[ix] )
        return(dp->segP[ix]->name);
    else
        return(NULL);
}


/*************************************************************************
function: DeleteFileObject - Remove object from table of contents
*/
DeleteFileObject (DocP,name)
DocInfoType *DocP;
char *name; /* name of file object */
{
    register docFileData *dp;
    register int docpos;

    dp = DocP->docFileId;
    if ( -1 == (docpos = FindFileStruct(dp,name)) )
        return(-1);
    DeleteFileObjByDP(dp,docpos);
    return(0);
}

/*************************************************************************
function: DeleteFileObjByDP - Low level file object deleter
*/
static DeleteFileObjByDP (dp,docpos)
register docFileData *dp;
int docpos;             /* in segP list of above */
{
    register tofcEntry *sp;
    register int jx, ix;
    int left_in_last_block;

    sp = dp->segP[docpos];

/* for debug purposes, trash the data image, hopefully this will help
    us find invalid references sooner! */

    for ( ix = 0; ix < sp->num_blocks-1; ix++ )
        memset(sp->blocks[ix],-1,AxIOblocks);
    left_in_last_block = sp->size % AxIOblocks;
    if ( left_in_last_block != 0 )
        memset(sp->blocks[ix],-1,left_in_last_block);

/* release file's data */

    TaskFree(dp->task,sp->name);
    for ( jx = 0; jx < sp->num_blocks; jx++ )
        TaskFree(dp->task,sp->blocks[jx]);

/* close file if open, scans thru all segptrs because a task (or child)
    may have opend the file more than once */

    for ( jx = 0; jx < MAX_OPEN_SEGS; jx++ ) {
        if ( DocSegPtrs[jx] == sp )
            DocSegPtrs[jx] = NULL;
    }

/* remove from doc object */

    TaskFree(dp->task,sp->blocks);
    memset(sp,0,sizeof(tofcEntry));
    TaskFree(dp->task,sp);
    while ( dp->segP[docpos+1] != NULL ) {      /* shift rest in */
        dp->segP[docpos] = dp->segP[docpos+1];
        docpos++;
    }
    dp->segP[docpos] = NULL;        /* re-term the list */
    dp->segP = (tofcEntry**)
                    TaskRealloc(dp->task,dp->segP,(docpos+1) * sizeof(tofcEntry*));
}

/*************************************************************************
function: TestForFileObject - Return whether a file object exists in file
*/
TestForFileObject (DocP,name)
DocInfoType *DocP;
char *name; /* name of file object */
{
    return(-1 != FindFileStruct(DocP->docFileId,name));
}

DocHardRead(fd,p,cnt)
int fd;
char *p;  /* place to deposit data  */ 
int cnt;
{
    return(DocSegRead(fd,p,cnt,-1));
}
DocHardWrite(fd,p,cnt)
int fd;
char *p;
int cnt;
{
    return(DocSegWrite(fd,p,cnt,-1));
}
/*************************************************************************
function: DocSegRead - Get data from file image
*/
DocSegRead (fd, p,cnt,pos)
int fd;         /* file descriptor */
char *p;        /* where to deposit data */
int cnt;        /* how many bytes */
long pos;       /* seek position */
{
    register char *memP;            /* current block desc */
    register int read_here;         /* bytes read in this block */
    register int bytes_read = 0;    /* work performed */
    register int seek_here;         /* current seek pos within block */
    register tofcEntry *te;

    TASKASSERT((fd > 0) && (fd < MAX_OPEN_SEGS) && 
                (te = DocSegPtrs[fd]) &&    /* assignment */
                (pos >= -1) && (pos < FILE_ABSURD) && 
                (cnt >= 0) && (cnt < FILE_ABSURD), FEbadIO);

    if ( pos == -1 )
        pos = te->filepos;
    if ( (cnt + pos) > te->size )    /* adjust for EOF reads */
        cnt = te->size - pos;
    while ( cnt > 0 ) {
        memP = te->blocks[pos / AxIOblocks];  /* select the block */
        TASKASSERT(memP != NULL,FEbadIO);
        seek_here = pos % AxIOblocks;                /* intra-block seek pos */
        read_here = MIN(cnt,AxIOblocks-seek_here);/* intra-block read amount */
        memcpy(p,memP+seek_here,read_here);     /* doit! */
        cnt -= read_here;                       /* adjust counters */
        bytes_read += read_here;
        p += read_here;
        pos += read_here;
    }
    te->filepos += bytes_read;
    return(bytes_read);
}

/*************************************************************************
function: FSfeof - Return whether at end of file
*/
FSfeof (fd)
int fd;
{
    register tofcEntry *te;

    te = DocSegPtrs[fd];
    if ( te->filepos == te->size )
        return(EOF);
    return(FALSE);
}

/*************************************************************************
function: FSgetc - Read one char from a file object or return EOF
*/
FSgetc (fd)
int fd;
{
    register tofcEntry *te;
    register char *p;
    register int pos;

    te = DocSegPtrs[fd];
    if ( (pos = te->filepos) == te->size )
        return(EOF);
    p = te->blocks[pos/AxIOblocks] + (pos % AxIOblocks);
    te->filepos++;
    return(*p & 0x00FF);
}

/*************************************************************************
function: FSungetc - Un-read a char from a file
*/
FSungetc (ch,fd)
char ch;
int fd;
{
    register tofcEntry *te;

    te = DocSegPtrs[fd];
    if ( (ch == EOF) || (te->filepos == 0) )
        return(EOF);
    te->filepos--;
    FSfputc(ch,fd);
    te->filepos--;
    return(ch);
}

/*************************************************************************
function: FSfgets - Read a line from a file object
*/
char *FSfgets (passed_buf,size,fd)
char *passed_buf;   /* result buffer */
register int size;  /* size of buf including room for null-term */
int fd;     /* logical file desc */
{
    register tofcEntry *te;
    register char *buf = passed_buf;
    register long pos;
    register long left;     /* bytes left in file from seek pos */
    register char *p;       /* read pointer */
    register int read_here = -1;
    register int seek_here;         /* current seek pos within block */

    size--;                 /* leave room for trailing null */
    te = DocSegPtrs[fd];
again:
    pos = te->filepos;
    left = te->size - pos;
    if ( left == 0 ) {
        if ( !read_here )   /* -1 for time, 0-for block spans */
            goto wrapup;
        return(NULL);
    }
    seek_here = pos % AxIOblocks;                /* intra-block seek pos */
    read_here = MIN(left,AxIOblocks-seek_here);/* intra-block read amount */
    p = te->blocks[pos/AxIOblocks] + seek_here;
    pos += read_here;
    while ( read_here && size ) {
        te->filepos++;
        read_here--;
        if ( *p == '\n' ) {
            {           /* append newline and return */
                *buf++ = *p;
                *buf = 0;
                return(passed_buf);
            }
        }
#if AXMS
        /* handle possible CR/NL pair from DOS  */
        if ( (*p == 0x0d) && (*(p+1) == '\n') )
            {
            te->filepos++;
            *buf++ = '\n';
            *buf = 0;
            return(passed_buf);
            }
#endif
        else {
            *buf++ = *p++;
            size--;
        }
    }
    if ( size )
        goto again;
wrapup:
    *buf = 0;
    return(passed_buf);
}

/*************************************************************************
function: DocSegSeek - Set the filepos
*/
DocSegSeek (fd,pos,mode)
int fd;
long pos;
int mode;
{
    register tofcEntry *te;

    TASKASSERT((fd > 0) && (fd < MAX_OPEN_SEGS) && 
                (te = DocSegPtrs[fd]), FEbadIO);
    if ( mode == 0 )
        te->filepos = pos;
    else if ( mode == 1 )
        te->filepos += pos;
    else if ( mode == 2 )
        te->filepos = te->size + pos;
    else
        taskAbort(NXLT("I/O error (file object seek)"));
    TASKASSERT((te->filepos >= 0) && (te->filepos < FILE_ABSURD), FEbadIO);
    return(0);
}

/*************************************************************************
function: FSftell - Like ftell
*/
long FSftell (fd)
int fd;
{
    register tofcEntry *te;

    TASKASSERT((fd > 0) && (fd < MAX_OPEN_SEGS) && 
                (te = DocSegPtrs[fd]), FEbadIO);
    return(te->filepos);
}

/*************************************************************************
function: FSfsize - returns the current size of the file object.
*/
int FSfsize( fd)
int fd;
{
    register tofcEntry *te;

    TASKASSERT((fd > 0) && (fd < MAX_OPEN_SEGS) && 
                (te = DocSegPtrs[fd]), FEbadIO);
    return(te->size);
}

/*************************************************************************
function: FSfputc - Write one char to a file object
*/
FSfputc (ch,fd)
char ch;
int fd;
{
    register tofcEntry *te;
    register long pos;

    TASKASSERT((fd > 0) && (fd < MAX_OPEN_SEGS) && 
                (te = DocSegPtrs[fd]),FEbadIO); /* assignment */

    pos = te->filepos;

    if ( !te->have_written || ((pos+1)/AxIOblocks >= te->num_blocks) ) {
        DocSegWrite(fd,&ch,1,-1);
        return(ch);
    }
    *(te->blocks[pos/AxIOblocks] + (pos%AxIOblocks)) = ch;
    te->filepos++;
    te->size = MAX(te->size,te->filepos);
    return(ch);
}

/*************************************************************************
function: FSfputs - Write a line and append a newline
*/
FSfputs (s,fd)
char *s;
int fd;
{
    DocSegWrite(fd,s,strlen(s),-1);
}

/*************************************************************************
function: DocSegWrite - Write data to a file image
Relies on THIMalloc shooting down task on out-of-memory.
*/
DocSegWrite (fd,p,cnt,pos)
int fd;         /* file descriptor */
char *p;        /* where to get data from */
int cnt;        /* how many bytes */
long pos;       /* seek position */
{
    register char *memP;            /* current block desc */
    register int write_here;        /* bytes read in this block */
    register int seek_here;         /* current seek pos within block */
    register tofcEntry *te;
    register int ix, jx;
    register passed_cnt = cnt;

    TASKASSERT((fd > 0) && (fd < MAX_OPEN_SEGS) && 
                (te = DocSegPtrs[fd]) &&    /* assignment */
                (pos >= -1) && (pos < FILE_ABSURD) && 
                (cnt >= 0) && (cnt < FILE_ABSURD), FEbadIO);

    if ( pos == -1 )
        pos = te->filepos;

/* Roundup last block to blocksize on first write */

    if ( !te->have_written ) {
        if ( te->size != 0 ) {          /* only if there is a block */
            ix = te->size/AxIOblocks;
            te->blocks[ix] = TaskRealloc(te->task,te->blocks[ix],AxIOblocks);
        }
        te->have_written = TRUE;
    }

/* extend file if needed */

    ix = (pos+cnt)/AxIOblocks;
    if ( ix >= te->num_blocks ) {
        te->blocks = (char**)TaskRealloc(te->task,te->blocks,(ix+1)*sizeof(char*));
        for ( jx = te->num_blocks; jx <= ix; jx++ )
            te->blocks[jx] = memP = TaskCalloc(te->task,1,AxIOblocks);
        te->num_blocks = ix+1;
    }
    while ( cnt != 0 ) {
        memP = te->blocks[pos/AxIOblocks];      /* select the block */
        seek_here = pos % AxIOblocks;                /* intra-block seek pos */
        write_here = MIN(cnt,AxIOblocks-seek_here);/*intra-block write amount */
        memcpy(memP+seek_here,p,write_here);   /* doit! */
        cnt -= write_here;                       /* adjust counters */
        p += write_here;
        pos += write_here;
    }
    te->filepos = pos;
    if ( pos > te->size )
        te->size = pos;
    return(passed_cnt);
}

/*************************************************************************
function: FSKSAdd - Add a line to the FS KS file
*/
FSKSAdd (infoP,name,value)
register DocInfoType *infoP;
register char *name, *value;
{
    register ks_entry *p, *kshead;
    register char *valp, *namep;
    register docFileData *dp = infoP->docFileId;

    if ( !dp )
        return;

/* Try to re-use an existing field */

    valp = ALC_STR(dp->task,value);
    kshead = infoP->ksData;
    for ( p = kshead; p; p = p->next ) {
        if ( streq(p->name,name) ) {
            TaskFree(dp->task,p->value);
            p->value = valp;
            return;
        }
    }

/* allocate a new line */

    namep = ALC_STR(dp->task,name);
    p = (ks_entry*)TaskAlloc(dp->task,sizeof(ks_entry));
    p->next = kshead;
    p->value = valp;
    p->name = namep;
    infoP->ksData = p;
    return;
}

/*************************************************************************
function: FSKSDelete - Remove a line from the KS file
*/
FSKSDelete (infoP,name)
register DocInfoType *infoP;
char *name;
{
    register ks_entry *p, *next, **prevP;
    register docFileData *dp = infoP->docFileId;

    prevP = &infoP->ksData;
    for ( p = *prevP; p; p = next ) {
        next = p->next;
        if ( streq(name,p->name) ) {
            *prevP = p->next;
            TaskFree(dp->task,p->name);
            TaskFree(dp->task,p->value);
            TaskFree(dp->task,p);
        }
        else
            prevP = &p->next;
    }
}

/*************************************************************************
function: FSKSLookup - Lookup a name in this doc's KS file
Returns NULL if entry not found.
*/
char *FSKSLookup (infoP,name)
DocInfoType *infoP;
char *name;     /* name of field */
{
    register ks_entry *ks = infoP->ksData;

    while ( ks ) {
        if ( streq(name,ks->name) )
            return(ks->value);
        ks = ks->next;
    }
    return(NULL);
}

/*************************************************************************
function: ReadKSObject - Read kitchen sink format file into memory list
This returns the head of the list or NULL if there were any read
errors.
*/
typeErr ReadKSObject (DocP)
DocInfoType *DocP;
{
    register char *p;
    register ks_entry *kshead = NULL, *ksP;
    register int len;
    char buf[2000];     /* plenty big */
    long created, size;
    register int fd;
    char filename[STRPATH+1];
    register docFileData *dp = DocP->docFileId;

    AbandonKSObject(DocP);
    sprintf(filename,"%s/doc.info",DocP->pathname);
    fd = OpenFileObject(DocP,filename,DFM_CREATE,(int *)&created,&size);
    if ( fd == -1 )
        return(ErrnoErr());
    while ( FSfgets(buf,2000,fd) ) {
        len = strlen(buf);
        if ( NULL == (p = strchr(buf,':')) )
            return(FEnotDoc);
        buf[--len] = '\0';  /* clip newline */
        *p++ = 0;       /* null term name, get beyond marker */
        ksP = (ks_entry*)TaskAlloc(dp->task,sizeof(ks_entry));
        ksP->next = kshead;
        ksP->name = ALC_STR(dp->task,buf);
        ksP->value = ALC_STR(dp->task,p);
        kshead = ksP;           /* thread to head of list */
    }
    CloseFileObject(fd);
    DocP->ksData = kshead;
    return(AOK);
}

/*************************************************************************
function: WriteKSObject - Write kitchen sink file object
*/
WriteKSObject (infoP)
DocInfoType *infoP;
{
    register ks_entry *ks = infoP->ksData;
    char ksname[STRPATH+1];
    char str[2000];
    int crt, size, fd;
    register docFileData *dp = infoP->docFileId;

#if 0           /* all files need KS info, even if empty 10/23/89 */
    if ( !ks )
        return;
#endif
    sprintf(ksname,"%s/doc.info",infoP->pathname);
    fd = OpenFileObject(infoP,ksname,DFM_CREATE+DFM_FRESH,&crt,(long *)&size);
    while ( dp && ks ) {
        sprintf(str,"%s:%s\n",ks->name,ks->value);
        DocSegWrite(fd,str,strlen(str),(long)-1);
        ks = ks->next;
    }
    CloseFileObject(fd);
    return(0);
}

/*************************************************************************
function: AbandonKSObject - Toss the KS info
*/
AbandonKSObject (infoP)
DocInfoType *infoP;
{
    register ks_entry *ks = infoP->ksData, *next;
    register docFileData *dp = infoP->docFileId;

    while (dp && ks ) {
        TaskFree(dp->task,ks->name);
        TaskFree(dp->task,ks->value);
        next = ks->next;
        TaskFree(dp->task,ks);
        ks = next;
    }
    infoP->ksData = NULL;
}





#if 0
static Logger(str,a,b,c,d,e,f,g,h)
int str,a,b,c,d,e,f,g,h;
{
    static FILE *fp;
    if ( !fp )
        fp = fopen(NXLT("zzio.log"),"w");
    fprintf(fp,str,a,b,c,d,e,f,g,h);
    fprintf(fp,"\n");
    fflush(fp);
}
#endif

/*************************************************************************
function: OpenFileObjFmode - Fopen style of doc open
*/
FILE *OpenFileObjFmode (docP,name,smode)
DocInfoType *docP;
char *name;
char *smode;
{
    int mode, created, fd;
    long size;

#if 0
    tempAbort("call to OpenFileObjFmode is bogus - jon");
#else
/* WHO THE HELL WROTE CODE WHICH CASTS AN INT AS A "FILE *" FOR RETURN- JON
   I did - Zane
*/

    if ( streq(smode,"r") )
        mode = DFM_READ;
    else if ( streq(smode,"r+") )
        mode = DFM_READ;
    else if ( streq(smode,"w") || streq(smode,"w+") )
        mode = DFM_CREATE+DFM_FRESH;
    else if ( streq(smode, "wb") )    /* Windows binary mode. */
        mode = DFM_CREATE + DFM_FRESH;
    else
        taskAbort(NXLT("Illegal file open mode"));

    if ( docP->DocMatType == MNone )    /* for inset displayer */
        mode |= DFM_FILEOK;

    fd = OpenFileObject(docP,name,mode,&created,&size);
    if ( fd == -1 )
        return(NULL);
    else
        return((FILE*)fd);
#endif
}

/*************************************************************************
function: FileObjAccess - Like Aaccess
*/
FileObjAccess (DocP,name,mode)
DocInfoType *DocP;
char *name;
int mode;
{
    int n;
    
    if ( mode && mode != R_OK )
        taskAbort(NXLT("Illegal file access"));
    n = FindFileStruct(DocP->docFileId,name);
    if ( n == -1 )
        return(n);
    else
        return(0);
}


void taskAbortf(s,a,b,c,d,e,f,g,h)
char *s;
int a,b,c,d,e,f,g,h;
{
    char buf[2000];
    sprintf(buf,s,a,b,c,d,e,f,g,h);
    taskAbort(buf);
    return;
}

/*************************************************************************
function: ElfbMachine - Return machine type
*/
ElfbMachine ()
{
    return(atoi(SYS_INFO+1));   /* ignore first char of sysinfo */
}

/*************************************************************************
function: ElfbSupportsAudio - Return current state of audio support for
                this machine.
*/
ElfbSupportsAudio()
{
    GetSupportsAudio();

    if (SupportsAudio)
        return(-1);		/* ELF true */
    else
        return(0);
}

/*************************************************************************
function: ElfbCheckAudioSupport - Determine if this machine supports audio 
In order to meete that requirement, 3 conditions must be met:
- a profile option has not been set to manually override our determination
- our noise making executable, "player" must exist in the filters directory
- a platform-specific attempt to actually talk to the audio device/server
  must succeed.  This is intended to eliminate machines with the right parts
  (e.g. a /dev/audio file) but missing drivers, etc.
*/
ElfbCheckAudioSupport()
{
    GetSupportsAudio();
}

GetSupportsAudio()
{
    char buf[STRPATH];

    if (UPnumVal(FSNOAUDIO))
	{
        /* the audio option is overridden to be off */

        SupportsAudio = FALSE;
    }
    else
	{
		/*
		 * for all platforms, the first test is the existance of our
		 * player executable in the filters directory
		 */

    	sprintf(buf,NXLT("%s%sfilters%splayer"),ThimApplixDir,DirSlash,
		  DirSlash);

#if AXMS
		strcat(buf,".exe");
#endif

		if ( access(buf,0) != F_OK )
			SupportsAudio = FALSE;
		else
        	CheckAudioWare();
	}
}


#if HP800
long
myAudioHandler(audio,err_event)
Audio  *audio;
AErrorEvent  *err_event;
{
	char errorbuf[132];

	if (err_event->error_code != AENoError)
	{
#if BUILD_KANJI
        struct utsname name;
        int main_release_num;
        char *p, *q, *rel;
        char *digits = "0123456789";
        char lang_buf[128];
 
        uname(&name);
        rel = name.release;
        p = rel + strcspn(rel, digits);
        q = p + strspn(p, digits);
        *q = '\0';
        main_release_num = atoi(p);
        /* HP-UX.09.07 /usr/lib/libAlib.a was in bug for Japanese version, which
         * forcees environment variable LANG to "C" within following function
         * call AGetErrorText(). This bug is fixed at HP-UX B.10.01, in new
         * audio library /opt/audio/lib/libAlib.sl. We edit it here to enable
         * the later AXOpenIM().
         * Ashisuto, 1996.06.12
         */
        if (main_release_num < 10)
            /* Save the LANG variable */
            sprintf(lang_buf, "LANG=%s", (char*)getenv("LANG"));
#endif
		AGetErrorText(audio,err_event->error_code,errorbuf,131);

		fprintf(stderr,"Audio error: %s\n", errorbuf);

#if BUILD_KANJI
        /* Restore the LANG variable */
        if ( (main_release_num < 10) &&
             (strcmp(lang_buf+5, (char*)getenv("LANG")) != 0) )
            putenv(lang_buf);
#endif

		SupportsAudio = FALSE;
	}

	return(err_event->error_code); 
}

#endif

typeErr
CheckAudioWare()
{
	int err;

#if SUN4
	struct stat st;
	char *Audio_dev = "/dev/audio", *a_file;
	int Audio_fd;

	err = 0;
	Audio_fd = -1;

	err = stat(Audio_dev,&st);

	if (err < 0)
	{
		SupportsAudio = FALSE;
	}
	else
	{
		if (!S_ISCHR(st.st_mode))
		{
			SupportsAudio = FALSE;
		}
		else
		{
			Audio_fd = open(Audio_dev, O_WRONLY | O_NDELAY);

			if (Audio_fd < 0)
			{
				if (errno == EBUSY)
					SupportsAudio = TRUE;
				else
					SupportsAudio = FALSE;
			}
			else
			{
				SupportsAudio = TRUE;
				close(Audio_fd);
			}
		}
	}
#else 
#if SUN4S || SUNIS
	char *Audio_dev = "/dev/audio", *a_file;
	int Audio_fd;

	err = 0;
	Audio_fd = -1;

	a_file = (char *) strdup("/dev/audio");
	Audio_fd = open(a_file, O_WRONLY);

	if (Audio_fd < 0)
	{
		if (errno == EBUSY)
			SupportsAudio = TRUE;
		else
			SupportsAudio = FALSE;
	}
	else
	{
		SupportsAudio = TRUE;
		close(Audio_fd);
	}
#else 
#if SGI
	char        *Audio_dev = "/usr/sbin/sfplay";
	struct stat st;

	err = 0;
	err = stat(Audio_dev, &st);
	if (err < 0)
		SupportsAudio = FALSE;
	else
		SupportsAudio = TRUE;
#else
#if PMAX
    AudioServer *pAserver = NULL;
    char xxx;

    /* try kicking the audio server */
    xxx = *sys_errlist[ECONNREFUSED]; /* suppress err msg */
    *sys_errlist[ECONNREFUSED] = 0;
    pAserver = AOpenAudio(NULL);
    if(pAserver == NULL)
        SupportsAudio = FALSE;
    else {
        SupportsAudio = TRUE;
        ACloseAudio(pAserver);
    }
    *sys_errlist[ECONNREFUSED] = xxx;
#else
#if HP800
	Audio	*audio;
	AErrorHandler prevHandler;
	long status_return;
	char server[2];

	/*
	 * replace the current audio error handler with one
	 * that doesn't exit...
	 */

	prevHandler = ASetErrorHandler(myAudioHandler);

	/* use default audio device */

	server[0] = '\0';

	audio = AOpenAudio(server,&status_return);

	if (audio == NULL)
	{
		if (status_return == AEOpenFailed)
		{
        	SupportsAudio = FALSE;
		}
		else
		{
			/* device probably supports audio, but some
			 * kind of transient problem occurred...
			 */

        	SupportsAudio = TRUE;
		}
	}

	ACloseAudio(audio,NULL);

	SupportsAudio = TRUE;
#else
	/*
	 * all other platforms will return TRUE, in order to allow
	 * audio capability to be added on site, simply by providing
	 * a player executable
	 */

	SupportsAudio = TRUE;

#endif
#endif
#endif
#endif
#endif

}

/*************************************************************************
function: AutoSaveDocumentFile - Save file, transferring old stuff to new file
*/
typeErr AutoSaveDocumentFile (DocP)
DocInfoType *DocP;
{
int DemoMode=FALSE;
#define AVWRITE(fd,p,n) {cnt = write(fd,p,n); if (cnt != n) goto bkferr;}

    register docFileData *dp;
    tofcEntry *sp;
    char bkname[STRPATH+20];
    char hdr[HEADER_LINE_LEN];
    int ix, cnt, rqst, num_files, blockno;
    long bytes_left;
    int fd;
    int mode;
    int is_binary;


    if(mkdir(ThimUserBackupDir,0755) == -1)
        if(errno != EEXIST)
            return(ErrnoErr());

    sprintf(bkname,"%s/%s.%s",ThimUserBackupDir,DocP->DocName,
                        ThimSfxFromType(DocP->DocMatType));

    dp = DocP->docFileId;
    mode = FilePermsFromDocP(DocP,dp->name);
    fd = open(bkname,O_CREAT+O_WRONLY+O_TRUNC+O_BINARY,mode);
    if ( fd == -1 )
        return(ErrnoErr());


    if ( DocP->DocMatType == MNone )
        goto write_blocks;

/* write header */

/* Count number of files, use hueristic hack to determine whether
    this is a binary form write */

    num_files = 0;
    is_binary = FALSE;
    for ( ix = 0; dp->segP[ix]; ix++ ) {
        char *p;
        if ( !dp->segP[ix]->orphaned ) {
            num_files++;
            p = dp->segP[ix]->name;
            if ( streq(p,"1.elfo") || streq(".sw",p+strlen(p)-3) ||
                 streq(p,"1.elfx") )
                is_binary = TRUE;
        }
    }

    sprintf(hdr,"%s%d %s %4d %20s %4d %s\n",MAGIC_STRING,DocP->DocMatType,
                        is_binary ? SYS_INFO : REV2_GEN_INFO,
                        DocP->DocMatType,dp->revstamp,num_files,RESERVED_STR);
    rqst = HEADER_LINE_LEN;
    AVWRITE(fd,hdr,rqst);

/* write the table of contents */

    for ( ix = 0; sp = dp->segP[ix]; ix++ ) {
        char str[72];
        if ( sp->orphaned ) continue;
        sprintf(str,"%8d %50s %10d\n",sp->size,sp->name,0);
        rqst = 71;
        AVWRITE(fd,str,rqst);
    }

/* write the file objects */

write_blocks:
    for ( ix = 0; sp = dp->segP[ix]; ix++ ) {
        if ( sp->orphaned ) continue;
        if ( (DocP->DocMatType == MNone) && !streq(sp->name,"1.elf") )
            continue;
        bytes_left = sp->size;
        for ( blockno = 0; bytes_left; blockno++ ) {
            rqst = MIN(AxIOblocks,bytes_left);
            AVWRITE(fd,sp->blocks[blockno],rqst);
            bytes_left -= rqst;
        }
    }
    if ( -1 == close(fd) )
        goto bkferr;
    return(AOK);

bkferr:
    close(fd);
    return(ErrnoErr());
}

/*************************************************************************
function: BackupDocExists - return whether the backup copy exists !!
*/
typeErr BackupDocExists (DocP)
DocInfoType *DocP;
{
char bkname[STRPATH+20];
struct stat buf;

    sprintf(bkname,"%s/%s.%s",ThimUserBackupDir,DocP->DocName,
                        ThimSfxFromType(DocP->DocMatType));

    if(Astat(DocP->pathname,&buf) != 0)
        {
        if(errno == ENOTDIR || errno == ENOENT) /* was ; */
            return(FALSE);
        }
    return(TRUE);
}
/*************************************************************************
function: RmBackupDoc - remove the backup copy !!
*/
typeErr RmBackupDoc(DocP)
DocInfoType *DocP;
{
char bkname[STRPATH+20];

    sprintf(bkname,"%s/%s.%s",ThimUserBackupDir,DocP->DocName,
                        ThimSfxFromType(DocP->DocMatType));
    unlink(bkname);
}



RenameRefDocSegs(DocP,Chrt,oldname,newname)
DocInfoType *DocP, *Chrt;
char *oldname,*newname;
{
    char old_docname[STRPATH+1],new_docname[STRPATH+1];
    char objname[STRPATH+1];
    char tmp[30],*p;
    tofcEntry **segP;
    tofcEntry *sp;  
    int i;

    /* find where the slash is */
    p = strchr(Chrt->pathname,'/'); 
    strncpy(tmp,Chrt->pathname,(p-Chrt->pathname));
    tmp[p-Chrt->pathname] = '\0';

    /* Check if a doc already exists with the newname */
    sprintf(new_docname,"%s/%s.ag/doc.info",tmp,newname);
    if(TestForFileObject(DocP, new_docname) )
        return(FEsubNameUsed);



    /* walk through the doc segs and rename all the segments which */
    /* belong to this graph */
    segP = DocP->docFileId->segP;
    i=0;
    for(;;)
        {
        sp = segP[i];   
        if(sp == NULL)
            break;
        /*
         * get the objname .
         * for eg: if sp->name is *one.ag/1.bc, then  objname is *one 
         */
        p = strchr(sp->name,'.');   
        strncpy(objname,sp->name, (p-sp->name));
        objname[p-sp->name] = '\0';

        if(!Astrcmp(objname, oldname) )
            { /* this object has to be renamed */
            /* if sp->name is *one.ag/1.bc, p contains .ag/1.bc */
            sprintf(old_docname,"%s/%s%s",tmp,oldname,p);
            sprintf(new_docname,"%s/%s%s",tmp,newname,p);
            RenameFileObject(DocP,old_docname,new_docname);
            }
        i++;
        }
    return(AOK);
}

/*************************************************************************
function: AxReadFileObjIntoBuf - Read FS file object into a single malloced buf
Returns the buf.
The buffer has been alloced into the task space of the calling task.
*/
char *AxReadFileObjIntoBuf (fd,sizeP)
int fd;     /* file object desc */
int *sizeP; /* returned size of buffer */
{
    register char *buf;
    register tofcEntry *te;

    te = DocSegPtrs[fd];
    *sizeP = te->size;
    buf = (char*)TaskAlloc(THIMpid,te->size);
    if ( !buf )
        return(NULL);
    DocSegRead(fd,buf,*sizeP,(long)0); /* arg #3 should be int, arg #4 long */
    return(buf);
}

/****************************************************************************
 *
 * function: dir_exists - Checks if the dir Exists. Checks if the passed name
 *           is a directory, also checks if dir is writable to the user
 */
static typeErr
dir_exists (name, exists, writable)
char *name;
int *exists;
int *writable;
{
	struct stat lstat;
	typeErr err;

	if ( 0 != Astat(name, &lstat) )
	{
		*exists = FALSE;
		*writable = FALSE;
		err = AOK;
	}
	else
	{
		*exists = TRUE;
		/* check if it`s a directory */
		if ( (lstat.st_mode & S_IFMT) == S_IFDIR )
		{
			err = AOK;
			if ( (lstat.st_mode & 0200) == 0200 )
				*writable = TRUE;
			else
				*writable = FALSE;
		}
		else
		{
			/* passed name is not a directory */
			err = FEnoDir;
			*writable = FALSE;
		}
	}
	return(err);
}

/*************************************************************************
function: AxBackupFile - Construct name of backup file. Create directory
    it if it isn't found. returns AOK or error code.
*/
typeErr 
AxMakeBackupFile(filename, backupfile)
char *filename;     /* in: real file pathname */
char *backupfile;   /* out: backup file (should be size STRPATH) */
{
	typeErr err;
	int exists, writable;

	err = dir_exists(ThimUserBackupDir, &exists, &writable);
	if ( err == AOK )
	{
		if ( exists == FALSE )
		{
			if ( 0 != mkdir(ThimUserBackupDir, 0755) )
				err = ErrnoErr();
			else
				err = AOK;
		}
		else if ( writable == FALSE )
			err = FEEACCES;
	}
	if ( err == AOK )
		makeBackupFileName(filename, backupfile);
	return(err);
}

/*************************************************************************
function: AxRmBackupFile - Remove backup file, if it exists, given original
    file name
*/
AxRmBackupFile(filename)
char *filename; /* name of real file */
{
    char backup[STRPATH+1];

    makeBackupFileName(filename, backup);
    /* if the backup is not the same as the original, remove the backup */
    if (strcmp(backup,filename))
        unlink(backup);
}


static void 
makeBackupFileName(pathname, backupfile)
char *pathname;     /* in: real file pathname */
char *backupfile;   /* out: backup file (should be size STRPATH+1) */
{
    char dirname[STRPATH+1], filename[STRPATH+1];
    AxParsePath(pathname,dirname,filename);
    sprintf(backupfile,"%s/%s",ThimUserBackupDir,filename);
}


/***************************************************************************
function: AxReadDataFromFileObject - Read data from a file object. Used
                                    for copy/paste of elf data. returns
                                    NULL on error.
*/
data_ptr AxReadDataFromFileObject(task,docP, objname)
int task;
DocInfoType *docP;
char *objname;
{
    int fd;
    int created, size;
    char *bufP;
    data_ptr dataP, ElfReadDataFromBuffer();

    fd = OpenFileObject(docP,objname,DFM_READ,&created,(long *)&size);

    if (fd == -1) {
        return NULL;
    }
    bufP = TaskAlloc(task,size);
    DocSegRead(fd,bufP,size,(long)-1);  /* arg #4 should be long */
    CloseFileObject(fd);
    dataP = ElfReadDataFromBuffer(task,bufP, size);
    return dataP;
}
