/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axipc.c
   Applix ASC version: 36
   Date: Sat Jan 23 21:14:18 1999
*/
#include <stdio.h>
#include <setjmp.h>
#include <aos.h>
#include <paripc.h>
#include <scac.h>
#include <thimble.h>

extern int AxInTask;	/* 0 or task that owns the current stack */

typedef struct {
	char buf[IPC_HEAD_LEN];
	int type;		/* type of IPC_IPC mes */
} ipc_ipc_mes;

#include <axipc.p>

/* Special hack for Motif is that whenever an "ipc" message gets sent,
	The XtAppNextEvent should return rather than wait for next event.
	This is crucial for solving the return == OK on dialog boxes.
	Before this fix, XtAppNextEvent would gobble up the KeyUp event,
	send the OK callback and block waiting for a Thimble visible event!
	Thus, Thimble did not negin processing the OK until next user event.
*/
extern int ThimbleSaysToReturnNow; /* hack for breaking out of XtAppNextEvent */

/*************************************************************************
function: IPCsendMes - Add a message to someone's message queue
*/
int IPCsendMes (type,pid,msg,bytecount)
int type;		/* IPC_SRT... */
int pid;		/* Task id to send it to */
void *msg;		/* The message itself */
int bytecount;		/* how many bytes in msg including IPC_HEAD */
{
	return(IPCXsendMes(type,pid,msg,bytecount,THIMappid,AxInTask));
}

/*************************************************************************
function: AxSendMes - Send a Message to a task... new API
This largely replaces IPCXsendMes whihc is also passed the taskid
of the caller... rarely useful.
*/
int AxSendMes (taskid,msg,size,whoami)
int taskid, size, whoami;
void *msg;
{
	return(IPCXsendMes(IPC_SRT,taskid,msg,size,whoami,AxInTask));
}

/*************************************************************************
function: IPCXsendMes - Add a message to someone's message queue
Threads event style messages to end of msgs list.
Places Response style messages at the only place they can be.
*/
int IPCXsendMes (type,pid,msg,bytecount,my_appid,my_pid)
int type;		/* IPC_SRT... */
int pid;		/* Task id to send it to */
void *msg;		/* The message itself */
int bytecount;		/* how many bytes in msg */
int my_appid;	/* as per scac */
int my_pid;		/* task number or 0 for main task */
{
	register msg_info *p, **parentP;
	register context_block *dstP;

	p = (msg_info*)malloc(sizeof(msg_info)+bytecount-1);
	p->type = type;
	p->sender_pid = my_pid;
	p->sender_app = my_appid;
	p->bytecount = bytecount;
	p->discard_me = FALSE;
	p->gotten = FALSE;
	memcpy(p->msg,msg,bytecount);

	ThimbleSaysToReturnNow = 1;
	dstP = &TaskInfo[pid];
	if ( type == IPC_RES ) {	/* only one response can be active */
#if 1	/* 0 is just for debug! */
		if ( !TaskInfo[pid].in_use ||	/* no listener */
			 !dstP->ipc_response_waiting || /* listener not waiting */
			 dstP->rsp )				/* listener already has a response */
			return(0);		/* quietly proceed */
#else
		if ( dstP->rsp )
			tempAbort("response channel overflow");
		if ( !dstP->ipc_response_waiting )
			tempAbort("responded to hyperspace");
#endif
		dstP->rsp = p;
		return(0);
	}
	if ( (type != IPC_MES) && (type != IPC_SRT) && (type != IPC_IPC) )
		taskAbort("Illegal IPC message");

	for ( parentP = &dstP->msgs; *parentP; parentP = (msg_info**)*parentP );
	*parentP = p;
	p->next = NULL;
	return(0);
}



#if AXMS   /*  {  */

/*************************************************************************
function: AxSendUniqueMes - Add a message to someone's message queue
          replacing any xevent message of the same type which may already
          be there.  THis is used to handle things on Windows which generate
          a serires of "related" messages, and we really only want/need to
          process the last one.  For example, the "ConfigureNotify: messages
          generated in full-drag mode as a window is re-sized - jon
*/
int AxSendUniqueMes( pid, msg, bytecount, my_appid, check_func )
int pid;			/* Task id to send it to */
void *msg;		/* The message itself */
int bytecount;		/* how many bytes in msg */
int my_appid;		/* as per scac */
int (*check_func)();    /* a callback, supplied by the caller, returns TRUE */
                        /* if a message on the queue "matches" this message */
{
	register msg_info *p, **parentP, *existingP, *oldP;
	register context_block *dstP;

	p = (msg_info*)malloc(sizeof(msg_info)+bytecount-1);
	p->type = IPC_SRT;
	p->sender_pid = AxInTask;
	p->sender_app = my_appid;
	p->bytecount = bytecount;
	p->discard_me = FALSE;
	p->gotten = FALSE;
	memcpy(p->msg,msg,bytecount);

	ThimbleSaysToReturnNow = 1;
	dstP = &TaskInfo[pid];

	for ( parentP = &dstP->msgs; *parentP; parentP = (msg_info**)*parentP )
       {
       existingP = *parentP;
       if ( (*check_func)( existingP->msg, p->msg ) )
          {
          /* found the same type of message, replace the one which is there */
          p->next = existingP->next;
          *parentP = p;
          free( existingP );
          return(0);
          }
       }

    /* fell through, add to the end */
	*parentP = p;
	p->next = NULL;
	return(0);
}

#endif  /*  }  AXMS  */





/*************************************************************************
function: AxSleepSeconds - Sleep N seconds
This removes the incoming messages from the to-be-delayed task.
This is because THIMawait would otherwise immediately return
rather than go to schedular.  Even if THIMawait went to the schedular,
the schedular would immediately fire this task up anyhow.

The old message state is removed before the THIMawait.
If there are new messages that occur during the THIMawait,
the old messages are append after the THIMawait.  If there
are no new messages, the old ones are simply restored.
*/
void AxSleepSeconds (timer)
int timer;
{
	AxSleepMilliseconds(timer*1000);
}


void AxSleepMilliseconds (timer)
int timer;
{
	register msg_info *mesP, *newMesP;
	register context_block *p = &TaskInfo[AxInTask];
	char *msgP;
	long size, senderType, senderTask, ljunk;

	timer = MAX(0,timer);
	p->ipc_true_timer = TRUE;

/* note old messages */

retry:
	mesP = p->msgs;
	if ( mesP && mesP->discard_me ) {
		p->msgs = mesP->next;
		free(mesP);
		goto retry;
	}
	p->msgs = NULL;

/* sleep */

	AxTimedGetMessage(timer,&msgP,&size,&senderType,&senderTask,&ljunk);

/* restore messages */

	if ( mesP ) {			/* was old msg */
		if ( p->msgs ) {	/* new msg arrived */
			newMesP = p->msgs;
			p->msgs = mesP;
			while ( mesP->next )
				mesP = mesP->next;
			mesP->next = newMesP;
		}
		else
			p->msgs = mesP;
	}

	p->ipc_true_timer = FALSE;	/* reset for next call */
    return;
}

/*************************************************************************
function: TaskRecycle - Lower tasks priority to let another run
This is effectively a no-op if there are no other tasks waiting to run.
These tasks have lower precedence than tasks polling the keyboard.
*/
void TaskRecycle ()
{
	AxSleepSeconds(0);
}

/*************************************************************************
function: AxAllowReschecule - List this task a recycle-only
This is used by the WP to allow itself to be woken up from
an infinite AxGetMessage because its internal state has changed
and it may no longer match the displayed state.
*/
void AxAllowReschedule (pid)
int pid;
{
    register context_block *p = &TaskInfo[pid];
	struct timezone cur_timezone;

	if ( !p->ipc_timed_wait || (0 == p->ipc_milliseconds) )
		return;
	p->ipc_timed_wait = TRUE;
	p->timed_out = FALSE;
	p->recycle_only = FALSE;
	gettimeofday (&p->ipc_run_time, &cur_timezone);
	p->ipc_milliseconds = 0;
    return;
}

/*************************************************************************
function: AxMessageAbsorbed - Message has been processed by app, discard it
*/
void AxMessageAbsorbed ()
{
	register msg_info *mp = TaskInfo[AxInTask].msgs;

	if ( mp )
		mp->discard_me = TRUE;
    return;
}

/*************************************************************************
function: AxHaveQueuedMessage - Return whether message in task queue
Note that this does not cuase a context switch.
It is just usefull for checking for "high-speed" interactions
between tasks.
*/
int AxHaveQueuedMessage ()
{
	register context_block *p = &TaskInfo[AxInTask];
	register msg_info *mesP;


	while ( p->msgs ) {
		mesP = p->msgs;
		if ( mesP->discard_me ) {
			p->msgs = mesP->next;
			free(mesP);
		}
		else
			return(TRUE);
	}
	return(FALSE);
}

/*************************************************************************
function: AxGetMessage - Wait for a message, return message & status
*/
int AxGetMessage (timeout,msgP,sizeP,senderTypeP,senderTaskP,messageTypeP)
long timeout;		/* for now, seconds or -1 for infinite */
char **msgP;		/* ptr to message itself */
long *sizeP;		/* size of incoming message */
long *senderTypeP;	/* type of task that sent mesage */
long *senderTaskP;	/* task id if sent vy a task rather than system */
long *messageTypeP;
{
	if ( timeout > 0 )
		timeout = timeout * 1000;
	return(AxTimedGetMessage(
				timeout,msgP,sizeP,senderTypeP,senderTaskP,messageTypeP));
}

int AxTimedGetMessage (timeout,msgP,sizeP,senderTypeP,senderTaskP,messageTypeP)
long timeout;		/* milliseconds or -1 for infinite */
char **msgP;		/* ptr to message itself */
long *sizeP;		/* size of incoming message */
long *senderTypeP;	/* type of task that sent mesage */
long *senderTaskP;	/* task id if sent vy a task rather than system */
long *messageTypeP;
{
	register context_block *p = &TaskInfo[AxInTask];
	register msg_info *mesP;
	long GetTime();
	struct timezone cur_timezone;

retry:
	p->timed_out = FALSE;
	if ( p->msgs ) {
		mesP = p->msgs;
		if ( mesP->discard_me ) {
			p->msgs = mesP->next;
			free(mesP);
			goto retry;
		}
		goto return_event;
	}
	if ( p->haveSocketInput )
		goto return_event;
	p->recycle_only = FALSE;
	if ( timeout < 0 ) {
		p->ipc_timed_wait = FALSE;
	}
	else {
		p->ipc_timed_wait = TRUE;
		gettimeofday (&p->ipc_run_time, &cur_timezone);
		p->ipc_run_time.tv_sec += timeout / 1000;
		p->ipc_run_time.tv_usec += (timeout %1000) * 1000;
		p->ipc_milliseconds = timeout;
	}
	p->have_polled = FALSE;
sp:	SuspendTask(0);
	if ( p->pended )
		goto sp;
	if ( p->timed_out ) {
		if ( p->haveSocketInput )
			goto return_event;
		return(IPC_TO);
	}
return_event:
	if ( p->haveSocketInput ) {
		p->haveSocketInput = FALSE;
		return(IPC_SOCKET);
	}
	mesP = p->msgs;
	if ( !mesP )
		return(IPC_TO);
	*sizeP = (long) mesP->bytecount;
	*senderTypeP = (long) mesP->sender_app;
	*senderTaskP = (long) mesP->sender_pid;
	*msgP = mesP->msg;
	*messageTypeP = (long) mesP->type;
	return(IPC_OK);
}

/*************************************************************************
function: IPCgetHdr - Get ipc message
*/
int IPCgetHdr (mesP, wholeMesP, lenP, hidP, pidP, aidP)
char **mesP;
short *wholeMesP;
short *lenP;
long *hidP;
short *pidP;
short *aidP;
{
	register msg_info *mp = TaskInfo[AxInTask].msgs;

	if ( mp ) {
		*mesP = mp->msg;
		*wholeMesP = TRUE;
		*lenP = mp->bytecount;
		*hidP = 0;
		*pidP = mp->sender_pid;
		*aidP = mp->sender_app;
		mp->discard_me = TRUE;
	} 
	else
		*mesP = NULL;
}

/*************************************************************************
function: IPCignore - Ignore incoming message
*/
void IPCignore ()
{
	register msg_info *mp = TaskInfo[AxInTask].msgs;

	if ( mp )
		mp->discard_me = TRUE;
    return;
}

/*************************************************************************
function: AxDiscardMessage - toss current message
*/
void AxDiscardMessage ()
{
	register msg_info *mp = TaskInfo[AxInTask].msgs;

	if ( mp )
		mp->discard_me = TRUE;
    return;
}

/*************************************************************************
function: AxGetResponse - Get response message from a task
*/
int AxGetResponse (pid, mesBuf, bufLen)
int pid;
void *mesBuf;
int bufLen;
{
	register context_block *p = &TaskInfo[AxInTask];
	register msg_info *mp;

	if ( TaskInfo[AxInTask].haveSocketInput )
		tempAbort("IO conflict on get-response");
	while ( !TaskInfo[AxInTask].rsp ) {
		if ( TaskInfo[AxInTask].haveSocketInput )
			tempAbort("IO conflict on rsp looping");
		TaskInfo[AxInTask].ipc_response_waiting = pid;
		SuspendTask(0);
	}
	p = &TaskInfo[AxInTask];
	mp = p->rsp;
	p->ipc_response_waiting = FALSE;
	if ( !mp )	/* possible? schedular kills of any waiters */
		return(IPC_NOS);
	if ( mp->gotten )
		return(IPC_OK);
	if ( bufLen < mp->bytecount )
		tempAbort("IPC buffer too small");
	memcpy(mesBuf, mp->msg, MIN(bufLen, mp->bytecount));
	free(p->rsp);
	p->rsp = NULL;
	return(IPC_OK);
}

/*************************************************************************
function: IPCresponseSize - Wait for a response, return message size
*/
int IPCresponseSize (for_task)
int for_task;
{
	register context_block *p = &TaskInfo[AxInTask];
	register msg_info *mp;

	while ( !TaskInfo[AxInTask].rsp ) {
		TaskInfo[AxInTask].ipc_response_waiting = for_task;
		SuspendTask(0);
	}
	p = &TaskInfo[AxInTask];
	mp = p->rsp;
	if ( !mp )
		return(0);
	return(mp->bytecount);
}

/*************************************************************************
function: IPCresponseSizeAndSender - Return response size and sender appid
*/
int IPCresponseSizeAndSender (for_task, appidP)
int for_task;
int *appidP;		/* OUT: Appid of sender */ 
{
	register context_block *p = &TaskInfo[AxInTask];
	register msg_info *mp;

	while ( !TaskInfo[AxInTask].rsp ) {
		TaskInfo[AxInTask].ipc_response_waiting = for_task;
		SuspendTask(0);
	}
	p = &TaskInfo[AxInTask];
	mp = p->rsp;
	if ( !mp ) {
		*appidP = 0;
		return(0);
	}
	*appidP = mp->sender_app;
	return(mp->bytecount);
}

/*************************************************************************
function: IPCcopyMes - Copy a message back to caller's buffer
Unlike the real IPCcopyMes, this REQUIRES that the passed buffer is big enough
*/
int IPCcopyMes (mesBuf, bufLen, lenP, hidP, pidP, aidP)
register void *mesBuf;
int bufLen;
short *lenP;
long *hidP;
short *pidP;
short *aidP;
{
	register msg_info *mp = TaskInfo[AxInTask].msgs;

	if ( mp->gotten )
		return(IPC_OK);
	if ( bufLen < mp->bytecount )
		tempAbort("IPC buffer too small");
	mp->gotten = TRUE;
	mp->discard_me = TRUE;
	*pidP = mp->sender_pid;
	*aidP = mp->sender_app;
	*lenP = mp->bytecount;
	*hidP = 0;
	memcpy(mesBuf, mp->msg, MIN(bufLen, mp->bytecount));
	return(IPC_OK);
}

/*************************************************************************
function: IPCevMes - Get an event
See pgms/ipc/sources/oawait.c for info
*/
int IPCevMes (mesBuf, bufLen, lenP, pidP, aidP)
void *mesBuf;
int bufLen;
short *lenP;
short *pidP;
short *aidP;
{
	long hid;

	return (IPCcopyMes(mesBuf, bufLen, lenP, &hid, pidP, aidP));
}

/*************************************************************************
function: IPCevLocate - Locate and return info from the event
See pgms/ipc/sources/oawait.c for real info
*/
int IPCevLocate (mesP, wholeMesP, lenP, pidP, aidP)
char **mesP;	/* points to message or NULL */
short *wholeMesP;	/* always ret'd TRUE */
short *lenP;	/* length of the message obtained from message header */
short *pidP;	/* sender's pid */
short *aidP;	/* sender's application id */
{
	long hid;

	return (IPCgetHdr(mesP, wholeMesP, lenP, &hid, pidP, aidP));
}

/*************************************************************************
function: IPCcliOpen - Accept connection from a process
This must be called on a IPCconnect request as the next action!
*/
int IPCcliOpen (evChanP)
short *evChanP;	/* AxInTask of client */
{
	short len,aid;	/* ignored */
	long hid;
	char buf[100];

	return(IPCcopyMes(buf,100,&len,&hid,evChanP,&aid));
}

/*************************************************************************
function: IPCcliClose - A Thimble no-op
*/
int IPCcliClose ()
{
	return(IPC_OK);
}


long GetTime()
{
	return(time(0L));
}

#if 0
IPCterm(){}
int IPCreset() {taskAbort("IPCreset");}
#endif


/*************************************************************************
function: AxDeferMessage - Defer passed message onto a queue maintained by caller
*/
void AxDeferMessage()
{
	register context_block *p = &TaskInfo[AxInTask];
	msg_info *mesP;

	while ( p->msgs && p->msgs->discard_me ) {
		mesP = p->msgs;
		p->msgs = p->msgs->next;
		free(mesP);
	}

	p->deferList = (msg_info**)
			realloc(p->deferList,(p->numDefered+1)*sizeof(msg_info*));
	p->deferList[p->numDefered] = p->msgs;
	p->numDefered++;
	p->msgs->discard_me = FALSE;	/* sanity */
	p->msgs = p->msgs->next;
}

/*************************************************************************
function: AxLoadDeferedMessages - Release defered messages
*/
void AxLoadDeferedMessages ()
{
	register context_block *p = &TaskInfo[AxInTask];
	register int ix;
	msg_info *mesP;

	while ( p->msgs && p->msgs->discard_me ) {
		mesP = p->msgs;
		p->msgs = p->msgs->next;
		free(mesP);
	}

	for ( ix = p->numDefered-1; ix >= 0; ix-- ) {
		p->deferList[ix]->next = p->msgs;
		p->msgs = p->deferList[ix];
	}
	free(p->deferList);
	p->deferList = NULL;
	p->numDefered = 0;
}

/*************************************************************************
function: AxCreateMsgP
*/
data_ptr
AxCreateMsgP(msg_info *p)
{
	data_ptr msgP;
	if ( p == (msg_info *)0 )
		msgP = (data_ptr)0;
	else
	{
		msgP = AxMakeArray(7);
		AxAddIntToArray   (msgP, 0, p->sender_pid);
		AxAddIntToArray   (msgP, 1, p->sender_app);
		AxAddIntToArray   (msgP, 2, p->type);
		AxAddIntToArray   (msgP, 3, p->discard_me);
		AxAddIntToArray   (msgP, 4, p->gotten);
		AxAddIntToArray   (msgP, 5, p->bytecount);
		AxAddBinaryToArray(msgP, 6, p->bytecount, p->msg);
	}
	return(msgP);
}

/*************************************************************************
function: AxCreateMsgListP
*/
data_ptr
AxCreateMsgListP(msg_info *msgs)
{
	data_ptr msgP;
	msg_info *p;
	int n_msgs;

	if ( msgs == (msg_info *)0 )
		msgP = (data_ptr)0;
	else
	{
		/* count the messages */
		n_msgs = 0;
		p = msgs;
		while ( p )
		{
			p = p->next;
			n_msgs++;
		}

		/* create an array */
		msgP = AxMakeArray(n_msgs);
		p = msgs;
		n_msgs = 0;
		while ( p )
		{
			AxAddArrayToArray(msgP, n_msgs, AxCreateMsgP(p));
			p = p->next;
			n_msgs++;
		}
	}
	return(msgP);
}

/*************************************************************************
function: AxExtractMsgP
*/
msg_info *
AxExtractMsgP(data_ptr msgP)
{
	int bytecount;
	char *bin_dat;
	msg_info *p;

	if ( msgP == (data_ptr)0 || !AxIsArray(msgP) || AxArraySize(msgP) != 7 )
		p = (msg_info *)0;
	else
	{
		bytecount = AxIntFromArray(msgP, 5);
		p = (msg_info *)malloc(sizeof(msg_info)+bytecount-1);
		p->sender_pid = AxIntFromArray(msgP, 0);
		p->sender_app = AxIntFromArray(msgP, 1);
		p->type       = AxIntFromArray(msgP, 2);
		p->discard_me = AxIntFromArray(msgP, 3);
		p->gotten     = AxIntFromArray(msgP, 4);
		p->bytecount  = bytecount;
		AxBinaryFromArray(msgP, 6, &bin_dat);
		memcpy(p->msg, bin_dat, bytecount);
	}
	return(p);
}

/*************************************************************************
function: AxExtractMsgListP
*/
msg_info *
AxExtractMsgListP(data_ptr msgP)
{
	msg_info *msg, *curr_msg, *last_msg;
	int ix, n_msgs;

	last_msg = msg = (msg_info *)0;
	if ( msgP && AxIsArray(msgP) )
	{
		n_msgs = AxArraySize(msgP);
		for ( ix=0; ix<n_msgs; ix++ )
		{
			curr_msg = AxExtractMsgP(AxArrayElement(msgP, ix));
			if ( curr_msg == (msg_info *)0 )
				continue;

			if ( msg == (msg_info *)0 )
				msg = last_msg = curr_msg;
			else
			{
				last_msg->next = curr_msg;
				last_msg = curr_msg;
			}
		}
		if ( last_msg )
			last_msg->next = (msg_info *)0;
	}
	return(msg);
}

/*************************************************************************
function: AxGetMsgP
*/
data_ptr
AxGetMsgP()
{
	register context_block *p = &TaskInfo[AxInTask];
	data_ptr msgP = AxCreateMsgListP(p->msgs);
	return(msgP);
}

/*************************************************************************
function: AxSetMsgP
*/
data_ptr
AxSetMsgP(data_ptr msgP)
{
	register context_block *p = &TaskInfo[AxInTask];
	msg_info *msg, *last_msg;

	if ( msgP == (data_ptr)0 )
	{
		while ( p->msgs )
		{
			msg = p->msgs;
			p->msgs = p->msgs->next;
			free(msg);
		}
		p->msgs = (msg_info *)0;
	}
	else
	{
		if ( p->msgs == (msg_info *)0 )
			last_msg = (msg_info *)0;
		else
		{
			msg = p->msgs;
			while ( msg->next )
				msg = msg->next;
			last_msg = msg;
		}
		if ( last_msg )
			last_msg->next = AxExtractMsgListP(msgP);
		else
			p->msgs = AxExtractMsgListP(msgP);
	}
	return((data_ptr)0);
}

/*************************************************************************
function: AxPopMsgP
*/
data_ptr
AxPopMsgP()
{
	data_ptr msgP = AxGetMsgP();
	AxSetMsgP((data_ptr)0);
	return(msgP);
}
