/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axshell.c
   Applix ASC version: 33
   Date: Sun Jan 31 21:12:24 1999
*/
#if AXMS
#include <process.h>
#include <stdlib.h>
#include <winmap.h>
#endif

#include <aos.h>
#include <ashell.h>
#include <thimble.h>
#include <paripc.h>
#include <scac.h>
#include <getstr4.h>
#include <getstr10.h>
#include <signal.h>
#include <axshell.p>

extern short THIMpid;
extern int AxShellWaiting;
extern int AxHaveShellServer;   /* true if can pipe to out command server */

int THIMwtHelper, THIMrdHelper; /* pipe fds to thimble helper */
#if PMAX
int ShellPid=0; /* If we need the pid of a shelled process */
#else
static int ShellPid=0; /* If we need the pid of a shelled process */
#endif

#include <axshell.p>


/*************************************************************************
function: AxHelperExit - Tell helper process to exit
Called only during process shutdown.
*/
void AxHelperExit (exitCode)
int exitCode;
{
    int code;

    if ( AxHaveShellServer ) { /* invoked from thim program */
	code = SHC_EXIT;
	DoWrite(THIMwtHelper,(char*)&code,sizeof(code));
	DoWrite(THIMwtHelper,(char*)&exitCode,sizeof(exitCode));
	DoRead(THIMrdHelper,(char*)&code,sizeof(code));
    }
}

#if !AXMS    /* { */
/*************************************************************************
function: system - Do a remote call to subroutine system
??x check return codes

* MODIFIED 7/2892 luke; initialized code to 0
*/
system (cmd)
const char *cmd;
{
    int len;
    int code = 0;
    int task;

    if ( 4080 <= (len = strlen(cmd)+1) )
	return(XXcmdTooLong);
    if ( AxHaveShellServer && !AxForceFork() ) { /* invoked from thim program */
		LockShellServerToThisTask();
	code = SHC_RETSYS;
	DoWrite(THIMwtHelper,(char*)&code,sizeof(code));
	task = ThimUIDPid(THIMpid);
	DoWrite(THIMwtHelper,(char*)&task,sizeof(task));
	DoWrite(THIMwtHelper,(char*)&len,sizeof(len));
	DoWrite(THIMwtHelper,(char*)cmd,len);
	DoRead(THIMrdHelper,(char*)&code,sizeof(code));
		UnLockShellServerFromThisTask();
    }
    else {                                  /* invoked from dbx */
	int waitHere=1;

	if (waitHere)
	    code = shcmdLocalReturningChildsExitCode((char*)cmd, NULL, NULL, NULL);
	else
	    code = shcmdLocalReturningChildsPid((char*)cmd, NULL, NULL, NULL);
    }
    return(code);
}
#endif    /* } !AXMS */

/*************************************************************************
function: qsystem - Do a remote unpended call to subroutine system
*/
qsystem (cmd)
char *cmd;
{
    int len, code;
    int task;

    if ( 4080 <= (len = strlen(cmd)+1) )
	return(XXcmdTooLong);
#if AXMS
	axmsStartProcessNopend(cmd, FALSE, FALSE);
#else
    if (AxHaveShellServer) {  /* invoked from thim program */
	code = SHC_SYSTEM;
	DoWrite(THIMwtHelper,(char*)&code,sizeof(code));
	task = ThimUIDPid(THIMpid);
	DoWrite(THIMwtHelper,(char*)&task,sizeof(task));
	DoWrite(THIMwtHelper,(char*)&len,sizeof(len));
	DoWrite(THIMwtHelper,cmd,len);
	}
    else {                                  /* invoked from dbx */
	int waitHere=1;

	if (waitHere)
	    shcmdLocalReturningChildsExitCode(cmd, NULL, NULL, NULL);
	else
	    shcmdLocalReturningChildsPid(cmd, NULL, NULL, NULL);
    }
#endif
    return(0);
}

/*************************************************************************
function: THIMexec - Unpended Exec a command
*/
THIMexec (cmd)
char *cmd;
{
    int len, code;
    int task;

    if ( 4080 <= (len = strlen(cmd)+1) )
	return(XXcmdTooLong);
#if AXMS
	code = axmsStartProcessNopend(cmd, FALSE, FALSE);
#else
    if (AxHaveShellServer) {    /* invoked from thim program */
	code = SHC_FORK;
	DoWrite(THIMwtHelper,(char*)&code,sizeof(code));
	task = ThimUIDPid(THIMpid);
	DoWrite(THIMwtHelper,(char*)&task,sizeof(task));
	DoWrite(THIMwtHelper,(char*)&len,sizeof(len));
	DoWrite(THIMwtHelper,cmd,len);
    }
    else {                                  /* invoked from dbx */
	int waitHere=1;

	if (waitHere)
	    code = shcmdLocalReturningChildsExitCode(cmd, NULL, NULL, NULL);
	else
	    code = shcmdLocalReturningChildsPid(cmd, NULL, NULL, NULL);
    }
#endif
    return(code);
}

/*************************************************************************
function: THIMpexec - Pended Exec a command
*/
THIMpexec (cmd)
char *cmd;
{
    int len, code;
    int task;

    if ( 4080 <= (len = strlen(cmd)+1) )
	return(XXcmdTooLong);
    if (AxHaveShellServer && !AxForceFork()) { /* invoked from thim program */
		LockShellServerToThisTask();
	code = SHC_RETCODE;
	DoWrite(THIMwtHelper,(char*)&code,sizeof(code));
	task = ThimUIDPid(THIMpid);
	DoWrite(THIMwtHelper,(char*)&task,sizeof(task));
	DoWrite(THIMwtHelper,(char*)&len,sizeof(len));
	DoWrite(THIMwtHelper,cmd,len);
	DoRead(THIMrdHelper,(char*)&code,sizeof(code));
		UnLockShellServerFromThisTask();
    }
    else {                                  /* invoked from dbx */
	int waitHere=1;

	if (waitHere)
	    code = shcmdLocalReturningChildsExitCode(cmd, NULL, NULL, NULL);
	else
	    code = shcmdLocalReturningChildsPid(cmd, NULL, NULL, NULL);
    }
    return(code);
}

/*************************************************************************
function: THIMexecRet - Unpended Exec a command, getting child pid
*/
THIMexecRet (cmd, no_hide)
char *cmd;
int no_hide;
{
    int len, code;
    int task;

    if ( (len = strlen(cmd)+1) >= MAX_SHELL_CMD )
	return(CMD_TOO_LONG);
#if AXMS
	ShellPid = axmsStartProcessNopend(cmd, no_hide, TRUE);
#else
    if (AxHaveShellServer && !AxForceFork() ) {
		/* invoked from thim program */
		LockShellServerToThisTask();
	code = SHC_FORK_RET;
	DoWrite(THIMwtHelper,(char*)&code,sizeof(code));
	task = ThimUIDPid(THIMpid);
	DoWrite(THIMwtHelper,(char*)&task,sizeof(task));
	DoWrite(THIMwtHelper,(char*)&len,sizeof(len));
	DoWrite(THIMwtHelper,cmd,len);
	DoRead(THIMrdHelper,(char*)&ShellPid,sizeof(ShellPid));
		UnLockShellServerFromThisTask();
    }
    else {                                  /* invoked from dbx */
	int waitHere=1;

	if (waitHere)
	    code = shcmdLocalReturningChildsExitCode(cmd, NULL, NULL, NULL);
	else
	    code = shcmdLocalReturningChildsPid(cmd, NULL, NULL, NULL);
    }
#endif
    return(ShellPid);
}

/*************************************************************************
function: AxRunProgram - Execute shell command, rtn exit code
If no stdout file is specified, all ouput goes to dev/null.
If a stdout file is specified but no stderr, all output
goes to the stdout file.

* MODIFIED 7/28/92 luke; initialized code to 0
*
*/
AxRunProgram (progname,file_stdin,file_stdout,file_stderr,no_hide)
char *progname;
char *file_stdin;   /* name of stdin input file */
char *file_stdout;  /* NULL or "" for /dev/null */
char *file_stderr;  /* NULL or "" for /dev/null */
int no_hide;        /* NT ONLY! tell winshell not to hide the window of
					 * program being started.
					 */
{
    int code = 0;
    int task;
    exec_request msg;
#if AXMS
    int need_null_file = FALSE;
    char *argv[7];
    char shell_stub[400];
    char quoted_shell_stub[400];
    char null_file[STRPATH+1];
#endif

    if ( strlen(progname) > MAX_SHELL_CMD )
	return(CMD_TOO_LONG);
    strcpy(msg.progname,progname);

    if ( file_stdin && file_stdin[0] ) {
		if (file_stdin[0] == '\"')
			strcpy(msg.file_stdin,file_stdin);
		else
			sprintf(msg.file_stdin, "\"%s\"", file_stdin);
	}
    else
		msg.file_stdin[0] = 0;

    if ( file_stdout && file_stdout[0] ) {
		if (file_stdout[0] == '\"')
			strcpy(msg.file_stdout,file_stdout);
		else
			sprintf(msg.file_stdout, "\"%s\"", file_stdout);
	}
    else
		msg.file_stdout[0] = 0;

    if ( file_stderr && file_stderr[0]) {
		if (file_stderr[0] == '\"')
			strcpy(msg.file_stderr,file_stderr);
		else
			sprintf(msg.file_stderr, "\"%s\"", file_stderr);
	}
    else
		strcpy(msg.file_stderr,msg.file_stdout);
	
#if AX_X11    /*  {  */
    if (AxHaveShellServer && !AxForceFork() ) { /* invoked from thim program */
		LockShellServerToThisTask();
	code = SHC_PROC;
	DoWrite(THIMwtHelper,(char*)&code,sizeof(code));
	task = ThimUIDPid(THIMpid);
	DoWrite(THIMwtHelper,(char*)&task,sizeof(task));
	DoWrite(THIMwtHelper,(char*)&msg,sizeof(msg));
	DoRead(THIMrdHelper,(char*)&code,sizeof(code));
		UnLockShellServerFromThisTask();
    }
    else {                                  /* invoked from dbx */
	int waitHere=1;

	if (waitHere)
	    code = shcmdLocalReturningChildsExitCode(progname,
						     msg.file_stdin,
						     msg.file_stdout,
						     msg.file_stderr);
	else
	    code = shcmdLocalReturningChildsPid(progname,
						msg.file_stdin,
						msg.file_stdout,
						msg.file_stderr);
    }

#else    /*  }  {  */

    /* On the NT side, we run a small program to exec what we want */
    /* to handle "stdin", "stdout", and "stderr" in a portable way */
    /* as opposed to relying on the UNIX'ism that the next file opened */
    /* will always be the lowest possible number, and that there is an */
    /* implied correspondene between "stdin" which is a (FILE *) stream */
    /* and file-descriptor zero.                                       */
    /* Note: even though we specify separate arguments here, it seems */
    /*       that Windows takes everything, builds one big string, and */
    /*      the re-parses it into arguments, so a zero length arg disspapears */
    strcpy( shell_stub, ThimApplixDir );
    strcat( shell_stub, "\\winshell" );
    strcpy( quoted_shell_stub, "\"" );
    strcat( quoted_shell_stub, shell_stub);
    strcat( quoted_shell_stub, "\"");
    argv[0]=quoted_shell_stub;
    strcpy( null_file, "/dev/null" );

    if ( strlen(msg.file_stdin) > 0 )
       argv[1] = msg.file_stdin;
    else
       argv[1] = null_file;

    if ( strlen(msg.file_stdout) > 0 )
       argv[2] = msg.file_stdout;
    else
       argv[2] = null_file;

    /* if a std error file was specified, use it.
     * if no std error but a std out was specified, use that.
     * otherwise use the null file.          -donato
     */
    if ( strlen(msg.file_stderr) > 0 )
       argv[3] = msg.file_stderr;
    else if ( strlen(msg.file_stdout) > 0 )
       argv[3] = msg.file_stdout;
    else
       argv[3] = null_file;

#if 0                                                           /* see bug 91472 */
    AxSysDosUnixPath( progname );
#endif
    argv[4] = progname;
	if ( no_hide )
		{
		argv[5] = "-nohide";
		argv[6] = NULL;
		}
	else
    	argv[5] = (char *)NULL;

    code = _spawnvp( _P_WAIT, shell_stub, argv );

#endif    /*   }  */
    
    return(code);
}

int TaskWaitingOnShell;

/*************************************************************************
function: AxRunBatchProgram - Execute shell command, rtn exit code
If no stdout file is specified, all ouput goes to dev/null.
If a stdout file is specified but no stderr, all output
goes to the stdout file.
*/
AxRunBatchProgram (progname,file_stdin,file_stdout,file_stderr,no_hide)
char *progname;
char *file_stdin;   /* name of stdin input file */
char *file_stdout;  /* NULL or "" for /dev/null */
char *file_stderr;  /* NULL or "" for /dev/null */
int no_hide;
{
    int code;
    int task;
    exec_request msg;
    char *msgP;
    long size, senderType, senderTask, ljunk;

    char errstr[500];   /* more than what is need, enough to accomdate */
			/* the pathname to the program */
#if AXMS
	char *cmd;
#endif

    if ( strlen(progname) > MAX_SHELL_CMD )
	return(CMD_TOO_LONG);
    strcpy(msg.progname,progname);

    if ( file_stdin && (*file_stdin != '\0')) {
		if (file_stdin[0] == '\"')
			strcpy(msg.file_stdin,file_stdin);
		else
			sprintf(msg.file_stdin, "\"%s\"", file_stdin);
	}
    else
		msg.file_stdin[0] = 0;
 
    if ( file_stdout && file_stdout[0] ) {
		if (file_stdout[0] == '\"')
			strcpy(msg.file_stdout,file_stdout);
		else
			sprintf(msg.file_stdout, "\"%s\"", file_stdout);
	}
    else
		msg.file_stdout[0] = 0;

    if ( file_stderr && file_stderr[0]) {
		if (file_stderr[0] == '\"')
			strcpy(msg.file_stderr,file_stderr);
		else
			sprintf(msg.file_stderr, "\"%s\"", file_stderr);
	}
    else
		strcpy(msg.file_stderr,msg.file_stdout);
	
#if AXMS
	/* we need to pass a special flag to winshell to get it to start processes
	 * unpended (it was designed for running shell commands and filters).
	 * And we can't call axmsStartProcessNopend because we need the winshell
	 * interface to handle the std file descriptors.
	 */
	cmd = (char*)malloc(strlen(progname)+8+1);
	sprintf(cmd, "%s -nopend", progname);
	code = AxRunProgram(cmd, msg.file_stdin, msg.file_stdout, msg.file_stderr,
						no_hide);
	free(cmd);
#else
    if (AxHaveShellServer && !AxForceFork() ) {   /* invoked from thim program */
		LockShellServerToThisTask();
	code = SHC_PROC;
	DoWrite(THIMwtHelper,(char*)&code,sizeof(code));
	task = TaskWaitingOnShell = ThimUIDPid(THIMpid);
	DoWrite(THIMwtHelper,(char*)&task,sizeof(task));
	DoWrite(THIMwtHelper,(char*)&msg,sizeof(msg));

    /* Wait for message */

	AxShellWaiting = TRUE;
	code = AxGetMessage(IPC_INF,&msgP,&size,&senderType,&senderTask,&ljunk);
		UnLockShellServerFromThisTask();
	if ( code == IPC_TO )   {
	    strcpy(errstr,NXLT("RUN_PROGRAM_UNPENDED@ -- "));
	    strcat(errstr,progname);
	    AxMessageAbsorbed();    /* message doesn't really contain anything, just a notify */
	    ElfStrAbort(ERR_UNEXPECTED, NULL, errstr);
	}
	if ( senderType != APSC )   {
	    strcpy(errstr,NXLT("RUN_PROGRAM_UNPENDED@ -- "));
	    strcat(errstr,progname);
	    AxMessageAbsorbed();    /* message doesn't really contain anything, just a notify */
	    ElfStrAbort(ERR_UNEXPECTED, NULL, errstr);
	}
	code = *((int*)msgP);
	AxMessageAbsorbed();
    }
    else {                                  /* invoked from dbx */
	int waitHere=1;

	if (waitHere)
	    code = shcmdLocalReturningChildsExitCode(progname, msg.file_stdin,
						     msg.file_stdout,
						     msg.file_stderr);
	else
	    code = shcmdLocalReturningChildsPid(progname, msg.file_stdin,
						msg.file_stdout,
						msg.file_stderr);
    }
#endif
    return(code);
}

/*************************************************************************
function: AxShellResponse - Response to a RunProgram request
Get return code from pipe, forward to task waiting.
*/
AxShellResponse ()
{
    char msg[sizeof(int)];
    int task;

    DoRead(THIMrdHelper,msg,sizeof(int));
    AxShellWaiting = FALSE;
    task = ThimPidFromUID(TaskWaitingOnShell);
    if ( task != 0 )
	IPCXsendMes(IPC_SRT,task,msg,sizeof(int),APSC,0);
}

/*************************************************************************
function: shcmdLocalReturningChildsExitCode - Adapted from startup.c
*/
shcmdLocalReturningChildsExitCode(cmd, stdin_file, stdout_file, stderr_file)
char *cmd;
char *stdin_file, *stdout_file, *stderr_file;
{
    register int rtnv;
    char cmd2[STRPATH+32], filename[STRPATH+32];
    register FILE *fp;

#if AXMS     /*  {  */
    rtnv = AxRunProgram(cmd, stdin_file, stdout_file, stderr_file, FALSE);
    return( rtnv );

#else        /* }  {  */

    strcpy( cmd2, NXLT("/bin/sh \""));	/* note leading quote for tempfile */

    /* Need to have a template specified here or else rpc services
       started by directly invoking axmain fail.
       See calls to create_temp_file@() in rpc_connector$() to see why.
       Temp names used to 'clash' because the calling elf layer did the
       following:
       * create a temp file.
       * note it's name.
       * delete the temp file.
       * fire up a rpc-server (via here), giving it the noted temp file
       as the socket file.
       * the code here (without the template) used to create the exact
       same file; run the rpc-server (which used the same file as the
       socket-file!.  rpc-server creates a socket-file by that name).
       * the temp file is removed here (now a socket file).
       * No socket-file --> rpc-server fails (e.g. ax400).
       Note, that launching the rpc-server thru 'thim' does not create
       a tmp command file ... */
    if (!AxTempFile(NXLT("axr%%%%%"), filename)) /* 8.3 friendly */
		return(0x7f);
	strcat( cmd2, filename);
	strcat( cmd2, "\"");			/* trailing quote for tempfile */

    /* Dump desired command into temp file and have it run by shell */
    fp = fopen(filename, "w");
    fprintf(fp, "(%s)", cmd);
    /* Construct requested redirections in bourne-shell syntax:
       cmd 1> out 2> err < in */
    if (stdout_file && stdout_file[0])
	fprintf(fp, " 1> %s", stdout_file);
    else
	fprintf(fp, " 1> /dev/null");

    if (stderr_file && stderr_file[0])
	fprintf(fp, " 2> %s", stderr_file);
    else
	fprintf(fp, " 2> /dev/null");

    if (stdin_file && stdin_file[0])
	fprintf(fp, " < %s", stdin_file);
    else
	fprintf(fp, " < /dev/null");

    fprintf(fp, "\n");
    fclose(fp);
    chmod(filename,0777);

    rtnv = shcmdLocal(cmd2) & 0xffff; /*only 16 bits are defined*/
    unlink(filename);

    /* If it is 0x7f then either the child stopped (whatever that means) or 
    the shcmd() routine failed and explicitly returned this value.  Don't do 
    any further interpretation. 
    If not 0x7f then, the child terminated due to a signal; treat this as a
    serious error. */

    if(rtnv&0xff)
	return(0x7f);

    /* The child exited and exit code is in upper byte. */
    return(rtnv >> 8);
#endif                        /*  }  not AXMS  */

}

/*************************************************************************
function: shcmdLocal - Perform shell command, return status
Issue a shell command and return its status
127 is returned if the command couldn't be exec'ed.
Double quoted args are passed as a single string after
the quotes are removed to make sh -c "string" work

Copied from startup.c
*/
int shcmdLocal(cmd)
char *cmd;
{         
    register int i;
    register char *c,*ccmd;
    int shpid;
    int status;
#if AX_X11
    int sig = SIGCHLD;
    void (*sfunc)();
#endif

    char *argv[32];
    int jpid;
    extern char *strtok();
	int len;

    i = strlen(cmd) + 1;
	len = i-1;
    if((c = ccmd = malloc((unsigned int)i)) == NULL)
		return(1);
    memcpy(ccmd,cmd,i);
    for(i = 0; i < 31 && (argv[i] = strtok(c," ")); i++) {
		if(*argv[i] != '\"') { 
	    	c = NULL; 
	    	continue; 
		}    
		argv[i]++; 			/* Bypass the leading quote */
		c = argv[i] + strlen(argv[i]);
		if ((c - ccmd) < len)
			*c++ = ' '; 	/* restore the original space (nulled by strtok) */
		c = strtok(argv[i],"\"");	/* Null the trailing quote */
		c = NULL;			/* So next strtok continues beyond that quote */
    }
    argv[i] = NULL;

#if AX_X11
    sfunc = signal(sig,SIG_DFL);
    if(!(shpid = fork())) {
	execvp(argv[0],argv);
	_exit(127);
    }
    free(ccmd);

    /* wait for exec'ed process to finish */
    while((jpid = wait(&status)) != shpid && jpid > 0);

    signal(sig,sfunc);
#else
    /* windows NT */
    status = _spawnvp( _P_WAIT, argv[0], argv );
    if ( status != -1 )
       {
       /* the process ran, for compatibility with UNIX, push the return */
       /* status into the upper byte                                    */
       status = (status & 0xFFFF) << 8;
       }
#endif
    return(status);
}


/*************************************************************************
function: shcmdLocalReturningChildsPid - Adapted from startup.c
*/
static int shcmdLocalReturningChildsPid(cmd, 
				       stdin_file, stdout_file, stderr_file)
char *cmd;
char *stdin_file;
char *stdout_file;
char *stderr_file;
{
    register int rtnv;
    char cmd2[STRPATH+32], filename[STRPATH+32];
    register FILE *fp;

    strcpy( cmd2, "/bin/sh \"");	/* note leading quote for tempfile */
    if (!AxTempFile(NXLT("axr%%%%%"),filename)) /* 8.3 friendly */
		return(0x7f);
	strcat( cmd2, filename);
	strcat( cmd2, "\"");			/* trailing quote for tempfile */

    /* Dump desired command into temp file and have it run by shell */
    fp = fopen(filename, "w");
    fprintf(fp, "(%s)", cmd);
    /* Construct requested redirections in bourne-shell syntax:
       cmd 1> out 2> err < in */
    if (stdout_file && stdout_file[0])
	fprintf(fp, " 1> %s", stdout_file);
    else
	fprintf(fp, " 1> /dev/null");

    if (stderr_file && stderr_file[0])
	fprintf(fp, " 2> %s", stderr_file);
    else
	fprintf(fp, " 2> /dev/null");

    if (stdin_file && stdin_file[0])
	fprintf(fp, " < %s", stdin_file);
    else
	fprintf(fp, " < /dev/null");

    fprintf(fp, "\n");
    fclose(fp);
    chmod(filename,0777);

    rtnv = shcmdLocalPid(cmd2) & 0xffff; /*only 16 bits are defined*/
    unlink(filename);

    /* If it is 0x7f then either the child stopped (whatever that means) or 
    the shcmd() routine failed and explicitly returned this value.  Don't do 
    any further interpretation. 
    If not 0x7f then, the child terminated due to a signal; treat this as a
    serious error. */

    if(rtnv&0xff)
	return(0x7f);

    /* The child exited and exit code is in upper byte. */
    return(rtnv >> 8);

}

/*************************************************************************
function: shcmdLocalPid - Kick off shell command, return pid
Don't wait around for shell command to finish.
127 is returned if the command couldn't be exec'ed.
Double quoted args are passed as a single string after
the quotes are removed to make sh -c "string" work

Copied from startup.c
*/
static int shcmdLocalPid(cmd)
char *cmd;
{         
    register int i;
    register char *c,*ccmd;
    int shpid;
    char *argv[32];
    extern char *strtok();
#if AX_X11
    int sig = SIGCHLD;
    void (*sfunc)();
#endif

    i = strlen(cmd) + 1;
    if((c = ccmd = malloc((unsigned int)i)) == NULL)
	return(1);
    memcpy(ccmd,cmd,i);
    for(i = 0; i < 31 && (argv[i] = strtok(c," ")); i++) {
	if(*argv[i] != '\"') { 
	    c = NULL; 
	    continue; 
	}    
	argv[i]++; 
	c = argv[i] + strlen(argv[i]);
	*c++ = ' '; 
	c = strtok(c,"\"");
    }
    argv[i] = NULL;

#if AX_X11
    sfunc = signal(sig,SIG_DFL);
    if(!(shpid = fork())) {
	execvp(argv[0],argv);
	_exit(127);
    }
    else
	ShellPid = shpid;
#else
    /* NT */
    shpid = ShellPid = _spawnvp( _P_NOWAIT, argv[0], argv );
#endif

    free(ccmd);
    return(shpid);

}

/*************************************************************************
function: LockShellServerToThisTask - Prevent other tasks from hitting "applix"
If this is called by task-0 code, does nothing... we can only hope for the
best!
*/
void LockShellServerToThisTask ()
{
	if ( AxInTask )
		AxLockResource("shell$server$",ThimUIDPid(AxInTask));
}

/*************************************************************************
function: UnLockShellServerFromThisTask - Prevent other tasks from hitting "applix"
*/
void UnLockShellServerFromThisTask ()
{
	if ( AxInTask )
		AxUnlockResource("shell$server$",ThimUIDPid(AxInTask));
}

/*************************************************************************
function: AxForceFork - Return TRUE if need to fork for task-0 shell commands
This is only used to avoid channel sharing where task-0 needs to use the
shell server but another task is waiting for a response from shell server.
*/
int AxForceFork ()
{
#if 0
	if ( !AxInTask && AxIsResourceLocked("shell$server$") )
		return(TRUE);
	else
		return(FALSE);
#else
	return(FALSE);
#endif
}

/*************************************************************************
function: DoRead - Read passed amount, handle interrupts
*/
DoRead (fd,p,cnt)
int fd;
char *p;
int cnt;
{
	int code;
	int passed = cnt;

	while ( cnt > 0 ) {
again:  errno = 0;
		code = read(fd,p,cnt);
		if ( code <= 0 ) {
			if ( errno == EINTR )
				goto again;
			return(code);
		}
		p += code;
		cnt -= code;
	}
	return(passed);
}


/*************************************************************************
function: DoWrite - Write passed amount, handle interrupts
*/
DoWrite (fd,p,cnt)
int fd;
char *p;
int cnt;
{
	int code;
	int passed = cnt;

	while ( cnt > 0 ) {
again:  errno = 0;
		code = write(fd,p,cnt);
		if ( code <= 0 ) {
			if ( errno == EINTR )
				goto again;
			return(code);
		}
		p += code;
		cnt -= code;
	}
	return(passed);
}

