/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axshlib.c
   Applix ASC version: 22.5
   Date: Mon Jun 14 21:12:40 1999
*/
#include <axshlib.h>

#if RIOS
#include <sys/ldr.h>
#endif

#include <axtime.h>
#include <adld.h>
#include <elf_strs.h>
#include <thim_gen.h>

#include <axshlib.p>


int GEElfInstall(void);
int WPElfInstall(void);
int SSElfInstall(void);
int DBElfInstall(void);
int OMailElfInstall(void);
int MailElfInstall(void);
int RTElfInstall(void);
int TM1ElfInstall(void);
int LMElfInstall(void);
#if SUN4S || LINUX
int MCElfInstall(void);
#endif

extern int AxTestMode;     /* set to TRUE if axmain started with -test switch */
extern int AxTestVerbose;	/* set TRUE is -amake_verbose on command line */

static struct shlibInfo {
	int libId;
	char* libName;
	char* entryFuncName;
	int (*libLoader)(int);
	int (*elfInstaller)();
	char* elfExportsFile;
	int isLoaded;
	int needGlobalRefs;
	char* firstCall;
	int isMacroCall;
	time_t loadTime;
	ShlibFuncInfo* exportFuncTable;
} libInfo[] = {
	/* libaxel.so, the elf packages are loaded during init.
	   so we don't need elfInstaller for it */
	{ LIBID_EL, AXSHLIB_EL, "ELGetExportTable", SHLoadAndSnap_EL,        NULL,             NULL, FALSE, FALSE },
	{ LIBID_GE, AXSHLIB_GE, "GEGetExportTable", SHLoadAndSnap_GE, GEElfInstall, "elfexports.ge", FALSE, FALSE },
	{ LIBID_WP, AXSHLIB_WP, "WPGetExportTable", SHLoadAndSnap_WP, WPElfInstall, "elfexports.wp", FALSE, FALSE },
	{ LIBID_SS, AXSHLIB_SS, "SSGetExportTable", SHLoadAndSnap_SS, SSElfInstall, "elfexports.ss", FALSE, FALSE },
	{ LIBID_DB, AXSHLIB_DB, "DBGetExportTable", SHLoadAndSnap_DB, DBElfInstall, "elfexports.db", FALSE, FALSE },
	{ LIBID_OM, AXSHLIB_OM, "OMGetExportTable", SHLoadAndSnap_OM, OMailElfInstall, "elfexports.om", FALSE, FALSE },
	{ LIBID_MAIL, AXSHLIB_MAIL, "MAILGetExportTable", SHLoadAndSnap_MAIL, MailElfInstall, "elfexports.ml", FALSE, FALSE },
	{ LIBID_CORBA, AXSHLIB_CORBA, "CORBAGetExportTable", SHLoadAndSnap_CORBA, NULL, NULL, FALSE, FALSE },
	{ LIBID_TM1, AXSHLIB_TM1, "TM1GetExportTable", SHLoadAndSnap_TM1, TM1ElfInstall, "elfexports.tm1", FALSE, TRUE },
	{ LIBID_PROX, AXSHLIB_PROX, "PROXGetExportTable", SHLoadAndSnap_PROX, NULL, NULL, FALSE, FALSE },
	{ LIBID_RT, AXSHLIB_RT, "RTGetExportTable", SHLoadAndSnap_RT, RTElfInstall, "elfexports.rt", FALSE, FALSE },
	{ LIBID_LM, AXSHLIB_LM, "LMGetExportTable", SHLoadAndSnap_LM, LMElfInstall, "elfexports.lm", FALSE, FALSE },
#if 0
#if SUN4S || LINUX
	{ LIBID_MC, AXSHLIB_MC, "MCGetExportTable", SHLoadAndSnap_MC, MCElfInstall,
"elfexports.mc", FALSE },
#endif
#endif	
	{ -1}
};

/*************************************************************************
function: AxLoadShlib - Return ptr to export function table... load if needed
*/
ShlibFuncInfo* AxLoadShlib (int libid, char* firstCall)
{
	void* handle;
	typeErr err;
	char *Adlerror(), *p, *errorP, *tmpP;
	int ix,len, mode;
	char* libname;
	ShlibFuncInfo* fptr;
	ShlibFuncInfo* (*entry_func)();
	for(ix=0; libInfo[ix].libId >= 0; ix++) {
		if (libInfo[ix].libId == libid)
			break;
	}
	if (libInfo[ix].isLoaded)
		return(libInfo[ix].exportFuncTable);
	if (libInfo[ix].libId < 0) {
		tempAbort("Invalid libId in AxLoadShlib");
	}
		
	libname = libInfo[ix].libName;
	
	mode = ADL_BIND_IMMEDIATE | ADL_BIND_FIRST;
	if ( libInfo[ix].needGlobalRefs )
		mode |= ADL_BIND_GLOBAL;

	handle = Adlopen(libname, mode);
	
	if ( !handle )	{
		errorP = XLT("Could not Adlopen library ");	
		p = Adlerror();
		if(p && (*p != '\0'))	{
			len = strlen(errorP) + strlen(p) + 1;
			tmpP = (char*)malloc(len);
			strcpy(tmpP, errorP);
			strcat(tmpP, p);
			errorP = tmpP;
			ElfStrAbort(ERR_FUNCNOEXISTS, errorP, libname);
			free(errorP);
		}else	{
			ElfStrAbort(ERR_FUNCNOEXISTS, errorP, libname);
		}
	}
#if RIOS
	entry_func = (ShlibFuncInfo* (*)()) handle;
#else
	entry_func = (ShlibFuncInfo* (*)()) Adlsym(handle, libInfo[ix].entryFuncName, ADL_TYPE_PROCEDURE);
	if ( !entry_func )
		ElfStrAbort(ERR_FUNCNOEXISTS,XLT("Could not find (Adlsym) in library"),libInfo[ix].entryFuncName);
#endif
	fptr = (*entry_func)();
	libInfo[ix].exportFuncTable = fptr;
	libInfo[ix].loadTime = time(NULL);
	if (!libInfo[ix].isMacroCall) {
		libInfo[ix].firstCall = (char*)malloc(strlen(firstCall)+1);
		strcpy(libInfo[ix].firstCall,firstCall);
	}
	/* NOTE: isLoaded flag will be set in AxElfInstallForShlib()
	 * elf install cannot be done here, as the import functions are not yet
	 * bound to their addresses
	 */
	ElfbSetSysVar("LibStatusChanged$",AxMakeBoolData(BTRUE));
	if ( AxTestVerbose || (!AxTestMode && !TURBO) )
		printf("%s loaded to call %s(%s) on %s",
		   libInfo[ix].libName,libInfo[ix].firstCall,
		   (libInfo[ix].isMacroCall ? "ELF":"C"),ctime(&libInfo[ix].loadTime));
	return(fptr);
}

/*************************************************************************
function: AxElfLoadShlibForId - Load an applixware shared library given
its id.
*/
int AxElfLoadShlibForId(int libId, char* macroCalled)
{
	int ix;
	for(ix=0; ix < (sizeof(libInfo)/sizeof(libInfo[0])); ix++) {
		if (libId == libInfo[ix].libId && !libInfo[ix].isLoaded) {
			libInfo[ix].firstCall = (char*)malloc(strlen(macroCalled)+1);
			strcpy(libInfo[ix].firstCall,macroCalled);
			libInfo[ix].isMacroCall = TRUE;
			if(!(*libInfo[ix].libLoader)(0))
				return(FALSE);
			return(TRUE);
		}
	}
	return(TRUE);
}


/*************************************************************************
function: AxBinShlibImportFuncs - search import functions in the export
table and get their address.
*/
int AxBindShlibImportFuncs(ShlibFuncInfo* imp_tbl, ShlibFuncInfo* exp_tbl)
{
	int ix,fx=0, retval=TRUE;
	int res;
	if (imp_tbl == NULL || exp_tbl == NULL)
		return(FALSE);
	while (imp_tbl->fname != NULL && exp_tbl->fname != NULL) {
			res = strcmp(imp_tbl->fname,exp_tbl->fname);
			if (res == 0) {
				imp_tbl->fptr = exp_tbl->fptr;
				imp_tbl++;
				exp_tbl++;
			}
			else if (res > 0) {
				exp_tbl++;
			}
			else { /* error import function not in export table */
				ElfStrAbort(ERR_FUNCNOEXISTS,XLT("Could not find in library"),imp_tbl->fname);
				retval = FALSE;
				imp_tbl++;
			}
	}
	if (imp_tbl->fname != NULL) {
		ElfStrAbort(ERR_FUNCNOEXISTS,XLT("Could not find in library"),imp_tbl->fname);
		retval = FALSE;
	}
	return(retval);
}

/*************************************************************************
function: AxElfInstallForShlib - install Elf packages in the given
applixware shared library.
*/
void AxElfInstallForShlib(int libId)
{
	int ix;
	for(ix=0; ix < (sizeof(libInfo)/sizeof(libInfo[0])); ix++) {
		if (libId == libInfo[ix].libId && !libInfo[ix].isLoaded) {
			libInfo[ix].isLoaded = TRUE;
			if (libInfo[ix].elfInstaller != NULL) {
				ElfUnloadDummyPackagesForShlib(libId);
				(*libInfo[ix].elfInstaller)();
			}
		}
	}
}


/*************************************************************************
function: AxShlibElfInit - Initialize Elf for the given applixware shared
library. It installs dummy packages for the shared library, which will be
removed before loading the shared library.
*/
void AxShlibElfInit(void)
{
	char filename[STRPATH+1];
	int ix;
	extern char *ThimApplixDir;
	for(ix=0; ix < (sizeof(libInfo)/sizeof(libInfo[0])); ix++) {
		if (libInfo[ix].elfInstaller == NULL)
			continue;
		sprintf(filename,"%s/elf/%s",ThimApplixDir, libInfo[ix].elfExportsFile); 
		AxSysDosUnixPath(filename);
		ElfLoadDummyPackagesForShlib(libInfo[ix].libId,filename);
	}
}

/*************************************************************************
function: getLoadedLibInfo - return loaded (internal and addin) libraries 
          info
*/
data_ptr AxElfGetLoadedLibInfo(void)
{
	int ix,jx;
	int nlibs = (sizeof(libInfo)/sizeof(libInfo[0]));
	int nlibsLoaded=0;
	int nAddins;
	data_ptr retP, aP, addinP;
	char filename[STRPATH+1];
	for(ix=0; ix < nlibs; ix++) {
		if (libInfo[ix].isLoaded)
			nlibsLoaded++;
	}
	addinP = AxElfGetAddinLibInfo();
	nAddins = AxArraySize(addinP);
	retP = AxMakeArray(nlibsLoaded+nAddins);
	for(ix=0,jx=0; ix < nlibs; ix++) {
		if (libInfo[ix].isLoaded) {
			aP = AxMakeArray(4);
			sprintf(filename,"%s/axshlib/%s",ThimApplixDir, 
					libInfo[ix].libName); 
			AxSysDosUnixPath(filename);
			
			AxAddStrToArray(aP,0,filename);
			AxAddStrToArray(aP,1,libInfo[ix].firstCall);
			AxAddIntToArray(aP,2,libInfo[ix].isMacroCall);
			AxAddIntToArray(aP,3,libInfo[ix].loadTime);
			
			AxAddArrayToArray(retP,jx,aP);
			jx++;
		}
	}
	for (ix=0; ix < nAddins; ix++) {
		AxAddArrayToArray(retP,nlibsLoaded+ix,AxArrayElement(addinP,ix));
	}
	return(retP);
}

/*************************************************************************
function: getLoadedLibInfo - return loaded (internal and addin) libraries 
          names
*/
data_ptr AxElfGetLoadedLibNames(void)
{
	int ix,jx;
	int nlibs = (sizeof(libInfo)/sizeof(libInfo[0]));
	int nlibsLoaded=0;
	int nAddins;
	data_ptr retP, addinP; 
	char filename[STRPATH+1];
	for(ix=0; ix < nlibs; ix++) {
		if (libInfo[ix].isLoaded)
			nlibsLoaded++;
	}
	addinP = AxElfGetInstalledSharedLibs();
	nAddins = AxArraySize(addinP);
	retP = AxMakeArray(nlibsLoaded+nAddins);
	for(ix=0,jx=0; ix < nlibs; ix++) {
		if (libInfo[ix].isLoaded) {
			sprintf(filename,"%s/axshlib/%s",ThimApplixDir, 
					libInfo[ix].libName); 
			AxSysDosUnixPath(filename);
			AxAddStrToArray(retP,jx,filename);
			jx++;
		}
	}
	for (ix=0; ix < nAddins; ix++) {
		AxAddDataToArray(retP,nlibsLoaded+ix,AxArrayElement(addinP,ix));
	}
	return(retP);
}
