/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axsocket.c
   Applix ASC version: 67
   Date: Fri Jan 22 21:12:16 1999
*/
#include <stdio.h>
#include <errno.h>
#if !M_88K && !PMAX
#include <sys/stat.h>
#endif
#include <axtime.h>
#if RIOS
#include <sys/select.h>
#endif
#include <aos.h>
#include <setjmp.h>
#include <thimble.h>
#include <getstr4.h>
#include <eldata.h>

#define CHECKSUM_MSGS 0
#define AXIOLOG 0

#if AXMS
#include <basicwin.h>
#define ENOTCONN WSAENOTCONN
extern HWND WXDummyWnd;
#endif

#define NOISE_BUFFER_SIZE 10000	/* must be same on JAVA client & server side */

#define MAX_SOCKETS 200

typedef struct {
    int in_use;     /* slot in use */
    char sockname[STRPATH+1];   /* server socket number or Unix pathname */
    int sockfd;     /* the fd the accept occurs on */
    int uidfd;      /* the unique fd handed back to the Elf layer */
    int task;       /* the task doing the accept on socket (or channel owner) */
    int readUp;     /* TRUE if reading/listening */
    int writeUp;    /* TRUE if writing */
    int taskUp;     /* task doing the read/write (may differ from owner) */
    int isSlow; /* If TRUE, show hourgalss on reads */
	int packedElfIO;	/* If true, uses tighter IO algorithm */
	int useChecksum;	/* If true, tag socket for checksums on rpc_respond */
	int xorNoise;		/* If true, noice is XOR'd in for Anyware */
		char *cliNoiseBuf;
		char *cliNextNoiseBuf;
		int cliNoiseDex;
	int bufferIt;		/* If true, doing buffered IO */
	void *buffer;		/* if bufferIt, buffer in progress */
	int bufferSize;		/* if bufferIt, size of buffer in progress */
	void (*readCB)(int);	/* Call back routine for reads */
	void *cbInfo;		/* call back info */
} elf_sock_t;

static elf_sock_t *AxSockets;
static int NumSockets;		/* high water size of AxSockets */

int AxShellWaiting;     /* TRUE if waiting for an external program */

extern int THIMwtHelper, THIMrdHelper;  /* pipe fds to thimble helper */
extern int AxSockFd;    /* LISTENER$ socket */

static int UidFd = 100000;  /* unique channel id counter */

extern jchannel_t *JavaChannels;	/* list of client channels */

#include <axsocket.p>



/*************************************************************************
function: AxSelectForInput - See if there is input
*/
int AxSelectForInput (xfd,timeP,haveXserverInputP)
int xfd;        /* socket for X-server, not used for windows */
struct timeval *timeP;
int *haveXserverInputP;
{
    static fd_set read_fds, wtfds, nofds;
    int ix, events;
	jchannel_t *jc, *jcnext;	

    *haveXserverInputP = FALSE;

    /* clear the read descriptors, and then set the ones we want */
    FD_ZERO( &read_fds );

#if AX_X11
    FD_SET(xfd,&read_fds);
#endif
    for ( ix = 0; ix < NumSockets; ix++ ) {
        if ( AxSockets[ix].in_use
            && AxSockets[ix].readUp
			&& ( TaskInfo[AxSockets[ix].taskUp].waiting_on_socket 
				 || AxSockets[ix].readCB)
            && AxSockets[ix].sockfd ) {
            FD_SET(AxSockets[ix].sockfd,&read_fds);
        }
    }

    /* clear the write descriptors, and then set the ones we want */
    FD_ZERO( &wtfds );

	for ( jc = JavaChannels; jc; jc = jc->next ) {
		if ( AxJavaOutputBytesQueued(jc) ) {
		    FD_SET(jc->fd,&wtfds);
		}
	}
    for ( ix = 0; ix < NumSockets; ix++ ) {
        if ( AxSockets[ix].in_use
            && AxSockets[ix].writeUp
			&& TaskInfo[AxSockets[ix].taskUp].waiting_on_socket
            && AxSockets[ix].sockfd ) {
            FD_SET(AxSockets[ix].sockfd,&wtfds);
        }
    }
    if ( AxShellWaiting ) {
        FD_SET(THIMrdHelper,&read_fds);
    }

#if AXMS

    for ( ix = 0; ix < NumSockets; ix++)
	if ( AxSockets[ix].in_use)
		break;

    if ( ix >= NumSockets )
        {
        /* no active channels, clear the timer we may have started earlier */
        ClearWindowsTimer();
        return(0);
        }
#endif

    events = select(FD_SETSIZE, &read_fds, &wtfds, &nofds, timeP);
    if ( events <= 0 )  /* error or time-out expired */
        return(events);

    for ( ix = 0; ix < NumSockets; ix++ ) {
        if ( AxSockets[ix].sockfd ) {
            if ( AxSockets[ix].readUp && FD_ISSET(AxSockets[ix].sockfd,&read_fds) ) {
					if (AxSockets[ix].readCB) {
	                	AxSockets[ix].readUp = FALSE;
						AxSockets[ix].readCB(AxSockets[ix].uidfd);
					}
					else {
	                	TaskInfo[AxSockets[ix].taskUp].haveSocketInput = TRUE;
	                	AxSockets[ix].readUp = FALSE;
					}
            }
            if ( AxSockets[ix].writeUp && FD_ISSET(AxSockets[ix].sockfd,&wtfds) ) {
	                TaskInfo[AxSockets[ix].taskUp].haveSocketInput = TRUE;
					AxSockets[ix].writeUp = FALSE;
            }
        }
    }

    if ( AxShellWaiting ) {
        if ( FD_ISSET(THIMrdHelper,&read_fds) )
            AxShellResponse();
    }

	for ( jc = JavaChannels; jc; jc = jcnext ) {
		jcnext = jc->next;	/* must use tempvar because WriteSome can free jc */
		if ( FD_ISSET(jc->fd,&wtfds) ) {
			AxJavaWriteSomeStuff(events,jc);
		}
	}

#if AX_X11
    if ( FD_ISSET(xfd,&read_fds) )
        *haveXserverInputP = TRUE;
    else
#endif
        *haveXserverInputP = FALSE;
    return(events);
}

/*************************************************************************
function: AxFindSocketByName - Lookup by name, return 0 if not found
*/
int AxFindSocketByName (sockname)
char *sockname;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( streq(AxSockets[dex].sockname,sockname) )
            return(AxSockets[dex].uidfd);
    }
    return(0);
}

/*************************************************************************
function: AxCanCreateAnotherChannel - Return whether we have reached socket limit
*/
int AxCanCreateAnotherChannel ()
{
    int dex;

	if ( NumSockets < MAX_SOCKETS )
		return(TRUE);
    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            return(TRUE);
    }
    return(FALSE);
}

/*************************************************************************
function: AxNoteSocket - Create socket struct
*/
int AxNoteSocket (sockfd,sockname,task,readUp)
int sockfd;
char *sockname;
int task;   /* As in AxInTask */
int readUp;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            break;
    }
    if ( dex == NumSockets ) {
		NumSockets++;
		if ( !AxSockets ) AxSockets = (elf_sock_t*)malloc(1); /* paranoid */
		AxSockets = (elf_sock_t*)realloc(AxSockets,sizeof(elf_sock_t)*NumSockets);
	}
    memset(&AxSockets[dex],0,sizeof(elf_sock_t));
    AxSockets[dex].in_use = TRUE;
    AxSockets[dex].sockfd = sockfd;
    if ( sockname )
        strcpy(AxSockets[dex].sockname,sockname);
    else
        AxSockets[dex].sockname[0] = 0;
    AxSockets[dex].task = task;
    AxSockets[dex].readUp = readUp;
    AxSockets[dex].taskUp = task;

    return(AxSockets[dex].uidfd = ++UidFd);
}

/*************************************************************************
function: AxChannelMaster - Detahc channel from specific task
*/
int AxChannelMaster(uid,task)
int uid, task;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
            AxSockets[dex].task = task;
            return(TRUE);
        }
    }
    return(FALSE);
}

/*************************************************************************
function: AxSocketSetSlowness - Set the "user hourglass" flag
*/
void AxSocketSetSlowness (uid,isSlow)
int uid, isSlow;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
            AxSockets[dex].isSlow = isSlow;
            return;
        }
    }
}

/*************************************************************************
function: AxSocketSetCallBackInfo - Set the callback info
*/
void AxSocketSetCallBackInfo (uid, cbInfo)
int uid;
void *cbInfo;
{
    int dex;
 
    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
            AxSockets[dex].cbInfo = cbInfo;
            return;
        }
    }
}

/*************************************************************************
function: AxSocketGetCallBackInfo - return the call back info
*/
void *AxSocketGetCallBackInfo (uid)
int uid;
{
    int dex;
 
    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) 
            return(AxSockets[dex].cbInfo);
    }
}
 
/*************************************************************************
function: AxSocketSetReadCallBack - Set the read callback function
*/
void AxSocketSetReadCallBack (uid,readCB)
int uid;
void (*readCB)(int);
{
    int dex;
 
    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
            AxSockets[dex].readCB = readCB;
            return;
        }
    }
}
 
/*************************************************************************
function: AxSockBufferWrites - Enable/Flush buffering
*/
void AxSockBufferWrites (uid, bufferIt)
int uid, bufferIt;
{
    int dex, code;

/* printf("buffered mode %d\n",bufferIt); */
    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
			if ( bufferIt ) {
				if ( !AxSockets[dex].bufferIt ) {
					AxSockets[dex].buffer = (void*)TaskAlloc(0,0);
					AxSockets[dex].bufferSize = 0;
					AxSockets[dex].bufferIt = TRUE;
				}
			}
			else {
				if ( AxSockets[dex].bufferIt ) {
					AxSockets[dex].bufferIt = FALSE;
					if ( AxSockets[dex].bufferSize != 0 ) {
						code = AxSocketWriter(uid,
							AxSockets[dex].buffer,
							AxSockets[dex].bufferSize);
						TaskFree(0,AxSockets[dex].buffer);
						AxSockets[dex].bufferSize = 0;
					    if ( code == -1 )
					        ElfStrAbort(ErrnoErr(),NULL,"write socket");
					}
				}
			}
			return;
        }
    }
}

/*************************************************************************
function: AxSocketSetPacking - Set the packedElfIo flag
*/
void AxSocketSetPacking (uid,packedElfIO)
int uid, packedElfIO;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
            AxSockets[dex].packedElfIO = packedElfIO;
            AxSockets[dex].useChecksum = packedElfIO;
            return;
        }
    }
}

/*************************************************************************
function: AxSocketIsSlow - Get the "user hourglass" flag
*/
int AxSocketIsSlow (uid)
int uid;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid )
            return(AxSockets[dex].isSlow);
    }
    return(FALSE);
}

/*************************************************************************
function: AxSockReadUp - Set the readUp flag
*/
int AxSockReadUp (uid)
int uid;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
            AxSockets[dex].readUp = TRUE;
            AxSockets[dex].taskUp = AxInTask;
            return(0);
        }
    }
    errno = ENOTCONN;
    return(-1);
}

/*************************************************************************
function: AxSockWriteUp - Set the writeUp flag
*/
int AxSockWriteUp (uid)
int uid;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
            AxSockets[dex].writeUp = TRUE;
            AxSockets[dex].taskUp = AxInTask;
            return(0);
        }
    }
    errno = ENOTCONN;
    return(-1);
}

/*************************************************************************
function: AxSockNoReadUp - ReSet the readUp flag
*/
int AxSockNoReadUp (uid)
int uid;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( (AxSockets[dex].uidfd == uid)
            && (AxSockets[dex].taskUp == AxInTask) ) {
            AxSockets[dex].readUp = FALSE;
            return(0);
        }
    }
    errno = ENOTCONN;
    return(-1);
}

/*************************************************************************
function: AxReleaseChannel - Release struct, but leave channel OPEN
*/
void AxReleaseChannel (uid)
int uid;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
            AxSockets[dex].in_use = FALSE;
            AxSockets[dex].sockfd = 0;
            AxSockets[dex].task = 0;
            AxSockets[dex].isSlow = 0;
            AxSockets[dex].packedElfIO = 0;
            AxSockets[dex].useChecksum = 0;
            return;
        }
    }
}

/*************************************************************************
function: sockPFromUID - Lookup socket struct
*/
static
elf_sock_t *sockPFromUID (int uid)
{
	int dex;

	for ( dex = 0; dex < NumSockets; dex++ ) {
		if ( AxSockets[dex].in_use && AxSockets[dex].uidfd == uid )
			return(&AxSockets[dex]);
	}
	return(NULL);
}

/*************************************************************************
function: AxSockFdFromUid - Given channel id, get real file desc
*/
int AxSockFdFromUid (uid)
int uid;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid )
            return(AxSockets[dex].sockfd);
    }
    return(0);
}

/*************************************************************************
function: AxSocketUseCheckSums - Set the useChecksum flag
*/
void AxSocketUseCheckSums (uid)
int uid;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( AxSockets[dex].in_use && (AxSockets[dex].uidfd == uid) ) {
            AxSockets[dex].useChecksum = TRUE;
			return;
		}
    }
}

/*************************************************************************
function: AxCksWriter - If socket enabled for checksums, checksum prior message
*/
void AxCksWriter(uid,outBuf,size)
unsigned char *outBuf;
{
#if CHECKSUM_MSGS
	int cks;
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
			if ( AxSockets[dex].useChecksum ) {
				cks = 0;
				while ( --size >= 0 ) cks += *outBuf++;
			    AxSocketWriter(uid,&cks,sizeof(cks));
			}
			return;
		}
    }
#endif
}

/*************************************************************************
function: AxSockFdAndPackerFromUid - Given channel id, get real file desc
*/
int AxSockFdAndPackerFromUid (uid,packedElfIOP)
int uid;
int *packedElfIOP;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
			*packedElfIOP = AxSockets[dex].packedElfIO;
            return(AxSockets[dex].sockfd);
		}
    }
    return(0);
}

/*************************************************************************
function: AxSockUidFromFd - Given physical file desc, get channel id
*/
int AxSockUidFromFd (fd)
int fd;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].sockfd == fd )
            return(AxSockets[dex].uidfd);
    }
    return(0);
}

/*************************************************************************
function: AxSocketClose - Close Socket
*/
int AxSocketClose (uid)
int uid;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid )
            break;
    }
    if ( dex == NumSockets )
        return(FALSE);

    ShutDownSocket( AxSockets[dex].sockfd );

    if ( (AxSockets[dex].sockname[0] == '/') ||
         (AxSockets[dex].sockname[0] == '\\') )
        unlink(AxSockets[dex].sockname);
    AxSockets[dex].in_use = FALSE;
    AxSockets[dex].sockname[0] = 0;
    AxSockets[dex].task = 0;
    AxSockets[dex].sockfd = 0;
    AxSockets[dex].isSlow = 0;
    AxSockets[dex].packedElfIO = 0;
    AxSockets[dex].useChecksum = 0;
    return(TRUE);
}

/*************************************************************************
function: ElfCloseAllSockets - Part of task & process termination only!
*/
void ElfCloseAllSockets (task)
int task;
{
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( AxSockets[dex].in_use
            && ( !task || AxSockets[dex].sockfd != AxSockFd )
            && ( !task || (task == AxSockets[dex].task) ) ) {
                AxSocketClose(AxSockets[dex].uidfd);
        }
    }
    return;
}

/*************************************************************************
function: AxTimedReadBytes - Reads bytes, rescheduling task if needed
read_bytes is used to read the incoming data on a connection.
That data is there has been established by a select().
This has the hack that if no bytes get read, the errno is set
here and this returns -1.  This hack is needed because is seems to be
the only way to detect that a server has gone away.
*/
int AxTimedReadBytes (uid,p,passed_cnt,timer,infinite_after_first)
int uid;
char *p;
int passed_cnt;
int timer;
int infinite_after_first;
{
    int fd;
    int cnt = passed_cnt;
    int got, total_got = 0;
    static fd_set read_fds, wtfds, nofds;
    static struct timeval timeout;
    struct timeval *timeP = &timeout;
    int events;
    int haveSlept = FALSE;
    long senderType, senderTask, size, msgType;
	int code, ix;
    char *msgP;
	elf_sock_t *sockP = sockPFromUID(uid);

    if ( cnt <= 0 )
        return(0);
	if ( !sockP ) {
        errno = ENOTCONN;
        return(-1);
	}
    fd = sockP->sockfd;

    /* test whether any input on fd */
    FD_ZERO(&read_fds );
    FD_SET(fd,&read_fds);
    events = select(FD_SETSIZE, &read_fds, &wtfds, &nofds, timeP);
    if ( events == -1 )
        goto getit;     /* handle failures in the one place */
    if ( events == 0 )
        goto sleeper;

/* Attempt to read amount of data requested */

getit:
    while ( cnt > 0 ) {
        errno = 0;
#if AX_X11
#if AXIOLOG
		printf("%s reading %d on %d\n",TaskInfo[AxInTask].argv[0],cnt,fd);
#endif
        got = read(fd,p,cnt);
#if AXIOLOG
		printf("%s got %d on %d\n",TaskInfo[AxInTask].argv[0],got,fd);
#endif
        if ( got == -1 ) {
			if (errno == EINTR || errno == EAGAIN)
				goto getit;
            if ( errno == ECONNRESET || errno == ECONNABORTED) {
                shutdown(fd, 0);
                close(fd);
            }
            if ( errno == EWOULDBLOCK )
                goto sleeper;
            AxLoadDeferedMessages();
            return(-1);
        }
#else
        /* Windows-NT */

#if 0
        /* turn off notifier while we are reading */
        WSAAsyncSelect( fd, WXDummyWnd, 0, 0 );
#endif

        got = recv(fd, p, cnt, 0 );

        /* now turn the notifier back on.  If there is more data in the */
#if 0
        /* socket, this will generate another message, so we will get back */
        /* here.                                              */
        SetSocketAsyncCallback( fd );     /* restore async messages */
#endif

        if ( got == -1 ) {
            errno = WSAGetLastError();
            if ( errno == WSAECONNRESET ) {
                shutdown(fd, 0);
                closesocket(fd);
            }
            if ( errno == WSAEWOULDBLOCK )
                 goto sleeper;
            AxLoadDeferedMessages();
            return(-1);
        }
#endif
        if ( got == 0 ) {   /* Pipe must be broken */
            if ( !haveSlept )
                goto sleeper;
            close(fd);
            errno = EPIPE;
            AxLoadDeferedMessages();
            return(-1);
        }
		if ( sockP->xorNoise ) {
			for ( ix = 0; ix < got; ix++ )
				*p++ = *p ^ fetchClientNoise(sockP);
		}
		else
	        p += got;
        cnt -= got;
        total_got += got;
        haveSlept = FALSE;
    }
    AxLoadDeferedMessages();
    return(passed_cnt);

/* If got here, socket would block, sleep for input */

sleeper:
    if ( (total_got > 0) && infinite_after_first )
        timer = -1;
    ACImOk();       /* unpend any task waiting on this one */

#if AXMS
    /* we start a timer so that we wake up periodically to check the socket */
    /* This is done in addition to using the WSAAsyncSelect() callback      */
    /*   it seems that depending on how the socket is read, or perhaps the  */
    /*   presence of multiple open connection, we do not always get the     */
    /*   expected message in the main loop so that we wake up from the      */
    /*   MS_GetMessage() call.  The timer is a failsafe device.  It will be */
    /*   taken down if AxSelectForInput() finds that there are no sockets   */
    /*   in use.                                                            */
    axmsStartWindowsTimer();
#endif

/* Wait for message */

msg_getter:
    haveSlept = TRUE;
    if ( -1 == AxSockReadUp(uid) ) {
        AxLoadDeferedMessages();
        return(-1);
    }
    if ( AxSocketIsSlow(uid) )
        AxmHourglassTaskWindows(TaskInfo[AxInTask].uid);

#if AXIOLOG
	printf("%s read waiting on socket %d\n",TaskInfo[AxInTask].argv[0],fd);
#endif
	TaskInfo[AxInTask].waiting_on_socket = TRUE;/* prevents RPC interleave */
	TaskInfo[THIMpid].waiting_on_socket = TRUE;/* prevents RPC interleave */
    code = AxGetMessage(timer,&msgP,&size,&senderType,&senderTask,&msgType);
	TaskInfo[AxInTask].waiting_on_socket = FALSE;
	TaskInfo[THIMpid].waiting_on_socket = FALSE;
	TaskInfo[AxInTask].haveSocketInput = FALSE;
	TaskInfo[THIMpid].haveSocketInput = FALSE;
#if AXIOLOG
	printf("%s read awoke on socket %d\n",TaskInfo[AxInTask].argv[0],fd);
#endif
	sockP = sockPFromUID(uid);

    switch ( code ) {
    case IPC_SOCKET:    /* expected response */
/*        AxMessageAbsorbed(); NO MESSAGE, don't discard it! */
        break;
    case IPC_TO:
        AxSockNoReadUp(uid);
#if AXMS
        errno = WSAETIMEDOUT;
#else
        errno = ETIMEDOUT;
#endif
        AxLoadDeferedMessages();
        if ( timer == -1 )
            return(-1);
        else
            return(total_got);
    default:
        AxDeferMessage();
        haveSlept = FALSE;      /* allow re-entry to sleeper */
        goto msg_getter;
    }
    goto getit; /* now read the stuff */
}

/*************************************************************************
function: AxSocketWriter - Write, allow re-scheduling
API like system write()
*/
int AxSocketWriter (uid,buf,passed_cnt)
int uid;        /* logical channel, not real fd */
char *buf;
int passed_cnt;
{
    int fd = 0, code;
    int cnt = passed_cnt;
    static fd_set read_fds, wtfds, nofds;
    static struct timeval timeout;
    struct timeval *timeP = &timeout;
    int events;
    int leftover = 0;
    long senderType, senderTask, size, msgType;
    char *msgP;
    int did;
	int dex;

    if ( cnt == 0 )
        return(0);

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
            fd = AxSockets[dex].sockfd;
			if ( AxSockets[dex].bufferIt ) {
				AxSockets[dex].buffer = (void*)TaskRealloc(0,AxSockets[dex].buffer,
									AxSockets[dex].bufferSize+passed_cnt);
				memcpy(((char*)AxSockets[dex].buffer)+AxSockets[dex].bufferSize,
							buf,passed_cnt);
				AxSockets[dex].bufferSize += passed_cnt;
/* printf("buffered write %d bytes\n",passed_cnt); */
				return(passed_cnt);
			}
		}
    }
    if ( !fd ) {
        errno = ENOTCONN;
        return(-1);
    }

/* see if write channel ready */

try_some:
 
    /* clear the write descriptors, and then set the ones we want */
    FD_ZERO( &wtfds );

    FD_SET(fd,&wtfds);
    events = select(FD_SETSIZE, &read_fds, &wtfds, &nofds, timeP);
    if ( events == -1 ) {
        AxLoadDeferedMessages();
        return(-1);
    }

/* if not ready, tell waiter about this channel and suspend */

    if ( events == 0 ) {
        if ( -1 == AxSockWriteUp(uid) ) {
            AxLoadDeferedMessages();
            return(-1);
        }
        if ( AxSocketIsSlow(uid) )
            AxmHourglassTaskWindows(TaskInfo[AxInTask].uid);

#if AXIOLOG
		printf("%s write waiting on socket %d\n",TaskInfo[AxInTask].argv[0],fd);
#endif
		TaskInfo[AxInTask].waiting_on_socket = TRUE;/* prevents RPC interleave */
		TaskInfo[THIMpid].waiting_on_socket = TRUE;/* prevents RPC interleave */
        code = AxGetMessage(-1,&msgP,&size,&senderType,&senderTask,&msgType);
		TaskInfo[AxInTask].waiting_on_socket = FALSE;
		TaskInfo[THIMpid].waiting_on_socket = FALSE;
		TaskInfo[AxInTask].haveSocketInput = FALSE;
		TaskInfo[THIMpid].haveSocketInput = FALSE;
#if AXIOLOG
		printf("%s write awoke on socket %d\n",TaskInfo[AxInTask].argv[0],fd);
#endif

        switch ( code ) {
        case IPC_SOCKET:    /* expected response */
/*        AxMessageAbsorbed(); NO MESSAGE, don't discard it! */
            goto try_some;
        case IPC_TO:        /* cannot happen */
            goto try_some;
        default:
            AxDeferMessage();
            if ( -1 == AxSockWriteUp(uid) ) {
                AxLoadDeferedMessages();
                return(-1);
            }
            goto try_some;
        }
    }

#if AX_X11
#if AXIOLOG
	printf("%s writing %d on %d\n",TaskInfo[AxInTask].argv[0],cnt,fd);
#endif
    did = write(fd,buf,cnt);
#if AXIOLOG
	printf("%s did %d on %d\n",TaskInfo[AxInTask].argv[0],did,fd);
#endif
    if ( -1 == did ) {
        if ( errno == EWOULDBLOCK || errno == EINTR || errno == EAGAIN )
            goto try_some;
        AxLoadDeferedMessages();
        return(-1);
    }
#else
    /*  Windows-NT */
    did = send(fd, buf, cnt, 0 );
    if ( -1 == did ) 
        {
        errno = WSAGetLastError();
        if ( errno == WSAEWOULDBLOCK )
            goto try_some;
        AxLoadDeferedMessages();
        return(-1);
        }
#endif
    else if ( did == 0 ) {      /* Pipe must be broken */
        close(fd);
        errno = EPIPE;
        AxLoadDeferedMessages();
        return(-1);
    }
    else if ( did < cnt ) {
        buf += did;
        cnt -= did;
        goto try_some;
    }
    else {
        AxLoadDeferedMessages();
        return(passed_cnt);
    }
}

/*************************************************************************
function: AxFdTimedReadBytes - Hack for Axnet socket comms
Creates a quickie logical channel if needed by axnet internals!
*/
int AxFdTimedReadBytes (fd,p,passed_cnt,timer,infinite_after_first)
int fd;
char *p;
int passed_cnt;
int infinite_after_first;
{
    int uid = AxSockUidFromFd(fd);
    int retval;
    int created_channel = FALSE;

    if ( !uid ) {
        uid = AxNoteSocket(fd,NULL,AxInTask,FALSE);
        created_channel = TRUE;
    }
    retval = AxTimedReadBytes(uid,p,passed_cnt,timer,infinite_after_first);
    if ( created_channel )
        AxReleaseChannel(uid);
    return(retval);
}

/*************************************************************************
function: AxFdSocketWrite - Hack for Axnet socket comms
Creates a quickie logical channel if needed by axnet internals!
*/
int AxFdSocketWriter(fd,buf,cnt)
int fd;
char *buf;
int cnt;
{
    int uid = AxSockUidFromFd(fd);
    int retval;
    int created_channel = FALSE;

    if ( !uid ) {
        uid = AxNoteSocket(fd,NULL,AxInTask,FALSE);
        created_channel = TRUE;
    }
    retval = AxSocketWriter(uid,buf,cnt);
    if ( created_channel )
        AxReleaseChannel(uid);
    return(retval);
}

/*************************************************************************
function: ElfFdStatus - Return info about channels
*/
data_ptr ElfFdStatus ()
{
    int ix;
    data_ptr aP, pP, sP;
    int sizepp = 0;

    aP = AxMakeArray(0);
    aP = AxAddIntToArray(aP,0,AxNumOpenFiles());
    pP = AxMakeArray(0);
    for ( ix = 0; ix < NumSockets; ix++ ) {
        if ( AxSockets[ix].in_use ) {
            sP = AxMakeArray(0);
            sP = AxAddIntToArray(sP,0,AxSockets[ix].uidfd);
            sP = AxAddIntToArray(sP,1,TaskInfo[AxSockets[ix].task].uid);
            if ( AxSockets[ix].readUp || AxSockets[ix].writeUp ) {
                sP = AxAddIntToArray(sP,2,TaskInfo[AxSockets[ix].taskUp].uid);
                sP = AxAddIntToArray(sP,3,AxSockets[ix].writeUp ? -1 : 0);
            }
            sP = AxAddStrToArray(sP,4,AxSockets[ix].sockname);
            pP = AxWriteArray(pP,sizepp++,sP);
        }
    }
    aP = AxWriteArray(aP,1,pP);
    return(aP);
}

/***************************************************************************
function: AxNumOpenFiles 
*/
int AxNumOpenFiles ()
{
	int ix;
	int numOpenFiles = 0;
    struct stat status;
    int num_files;

#if AX_X11
    num_files = FD_SETSIZE;
#else
    /* I do not know what the max is, but on Windows/NT "FD_SETSIZE" is */
    /* the size of a "select" array, and not a count of file descriptors */
    num_files = 256;
#endif

    for ( ix = 0; ix < num_files; ix++) {
        if ( -1 != fstat(ix, &status) )
            numOpenFiles++;
    }

	return(numOpenFiles);
}



#if AXMS
static UINT win_timer_id = 0;


void 
axmsStartWindowsTimer(void)
{
    UINT timeout_value;

    if ( win_timer_id != 0 )
       /* already set */
       return;

    /* set a 1/2 second timer */
    win_timer_id = 1;
    timeout_value = 500;     /* time in miliseconds */
    win_timer_id = MS_SetTimer( (HWND)NULL, 0, timeout_value, (TIMERPROC)NULL );
    return;
}


static void ClearWindowsTimer()
{
    if ( win_timer_id == 0 )
       /* already clear */
       return;

    MS_KillTimer( (HWND)NULL, win_timer_id );
    win_timer_id = 0;
    return;
}

#endif

/*************************************************************************
function: fetchClientNoise - fetch a byte of encrytable noice
If the current buffer goes empty, move to the scond buffer and
tell Elf code to fetch another buffer full from aagwty.

The ELF noise fetcher is waiting on a resource "noise_fetcher$".
When this code release the resource, the Elf code calls up the
aagwy to get the noise.  It gets the noise and calls post_noise$
to the socket.
*/
static
int fetchClientNoise (elf_sock_t *sockP) {
	if ( sockP->cliNoiseDex == NOISE_BUFFER_SIZE ) {
		AxNotifyResource("post_noise$");
		while ( !sockP->cliNextNoiseBuf )	/* should never happen! */
			AxSleepMilliseconds(100);	/* maybe make this an assertion! */
		sockP->cliNoiseBuf = sockP->cliNextNoiseBuf;
		sockP->cliNextNoiseBuf = NULL;
		sockP->cliNoiseDex = 0;
	}
	return(sockP->cliNoiseBuf[sockP->cliNoiseDex++]);
}

/*************************************************************************
function: ElfClientNoiseDelivery - Accept noise from aagtwy for client channel
*/
void ElfClientNoiseDelivery (int uid, char *noise) {
	elf_sock_t *sockP = sockPFromUID(uid);
}

/*************************************************************************
function: AxSocketSetXORNoise - Set the xorNoise flag
*/
void AxSocketSetXORNoise (int uid, int xorNoise) {
    int dex;

    for ( dex = 0; dex < NumSockets; dex++ ) {
        if ( !AxSockets[dex].in_use )
            continue;
        if ( AxSockets[dex].uidfd == uid ) {
            AxSockets[dex].xorNoise = xorNoise;
            return;
        }
    }
}

