/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: doclink.c
   Applix ASC version: 24
   Date: Mon Nov 16 13:53:26 1998
*/
#include <aos.h>
#include <scac.h>
#include <fsrts.h>
#include <eldata.h>
#include <doclinks.h>
#include <doctypes.h>
#include <afile.h>			/* for ATOK_START, etc */
#include <getstr10.h>

#define ELF_TRUE (-1)
#define ELF_FALSE 0

extern char *TaskAlloc();

doc_links_info_t *AxGetDocLinksInfo();
data_ptr AxDocLinksInfoToElf();

extern data_ptr ElfRetData;

/**************************************************************************\
 *                             utility funcs                              *
\**************************************************************************/

/***************************************************************************
function: is_start_token
*/
static int
is_start_token(buf)
mbchar *buf;
{
	if ( !strncmp(buf, ATOK_START, strlen(ATOK_START)) )
		return(TRUE);
	if ( !strncmp(buf, ATOK_OLDSTART, strlen(ATOK_OLDSTART)) )
		return(TRUE);
	return(FALSE);
}

/***************************************************************************
function: is_end_token
*/
static int
is_end_token(buf)
mbchar *buf;
{
	if ( !strncmp(buf, ATOK_END, strlen(ATOK_END)) )
		return(TRUE);
	return(FALSE);
}

/***************************************************************************
function: AxTestDocLinksInfoForLinks - Check passed doc links info for links.
Return TRUE if this doc or any child object has links.
*/
int
AxTestDocLinksInfoForLinks (doc_links)
doc_links_info_t *doc_links;
{
	int ix;

	if ( doc_links )
	{
		/* check for top level links */
		if ( doc_links->numLinks > 0 )
			return(TRUE);  /* document has links */

		/* no links - check objects for links */
		for ( ix=0; ix<doc_links->numObjects; ix++ )
		{
			if ( AxTestDocLinksInfoForLinks(doc_links->objects[ix].info) )
				return(TRUE);  /* found an object with links */
		}
	}
	return(FALSE);
}

/**************************************************************************\
 *               convert doc link info to/from ELF                        *
\**************************************************************************/

/***************************************************************************
function: AxDocLinksInfoFromElf 
*/
doc_links_info_t *
AxDocLinksInfoFromElf (elfDocLinksInfo)
data_ptr elfDocLinksInfo;
{
	int ix;
	int size;
	data_ptr aP;
	data_ptr bP;
	link_info_t *linkP;
	object_info_t *objectP;
	doc_links_info_t *doc_links;

	if ( (elfDocLinksInfo == (data_ptr)NULL) ||
									(AxArraySize(elfDocLinksInfo) <= 0) ) {
		return((doc_links_info_t *)NULL);
	}

	doc_links = (doc_links_info_t *)TrueCalloc(1, sizeof(doc_links_info_t));

	/* get links info */
	aP = AxArrayFromArray(elfDocLinksInfo, 0);
	if ( (aP != (data_ptr)NULL) && (size = AxArraySize(aP)) ) {
		doc_links->numLinks = size;
		doc_links->links = (link_info_t *)TrueCalloc(1,
								doc_links->numLinks * sizeof(link_info_t));
		for ( ix = 0; ix < doc_links->numLinks; ix++ ) {
			linkP = &doc_links->links[ix];
			bP = AxArrayFromArray(aP, ix);

			/* get this link's info */
			strncpy(linkP->name, AxStrFromArray(bP, AXELFDEX_LINK_name), STRPATH);
			linkP->name[STRPATH] = 0;
			linkP->docType = AxIntFromArray(bP, AXELFDEX_LINK_docType);
			linkP->appType = AxIntFromArray(bP, AXELFDEX_LINK_appType);
			strncpy(linkP->filterMacro,
						AxStrFromArray(bP, AXELFDEX_LINK_filterMacro), STRLONG);
			linkP->filterMacro[STRLONG] = 0;
			strncpy(linkP->launchMacro,
						AxStrFromArray(bP, AXELFDEX_LINK_launchMacro), STRLONG);
			linkP->launchMacro[STRLONG] = 0;
			strncpy(linkP->externalRefName,
						AxStrFromArray(bP, AXELFDEX_LINK_externalRef), STRLONG);
			linkP->externalRefName[STRLONG] = 0;
			strncpy(linkP->internalRefName,
						AxStrFromArray(bP, AXELFDEX_LINK_internalRef), STRLONG);
			linkP->internalRefName[STRLONG] = 0;

			/* get doc links info for this link */
			linkP->info =
				AxDocLinksInfoFromElf(AxArrayFromArray(bP, AXELFDEX_LINK_info));
				/* recurse */

			strncpy(linkP->override,
						AxStrFromArray(bP, AXELFDEX_LINK_override), STRPATH);
			linkP->override[STRPATH] = 0;
		}
	}

	/* get objects info */
	aP = AxArrayFromArray(elfDocLinksInfo, 1);
	if ( (aP != (data_ptr)NULL) && (size = AxArraySize(aP)) ) {
		doc_links->numObjects = size;
		doc_links->objects = (object_info_t *)TrueCalloc(1,
								doc_links->numObjects * sizeof(object_info_t));
		for ( ix = 0; ix < doc_links->numObjects; ix++ ) {
			objectP = &doc_links->objects[ix];
			bP = AxArrayFromArray(aP, ix);

			/* get this object's info */
			strncpy(objectP->name,
						AxStrFromArray(bP, AXELFDEX_OBJECT_name), STRPATH);
			objectP->name[STRPATH] = 0;
			objectP->docType = AxIntFromArray(bP, AXELFDEX_OBJECT_docType);
			objectP->appType = AxIntFromArray(bP, AXELFDEX_OBJECT_appType);
			strncpy(objectP->filterMacro,
					AxStrFromArray(bP, AXELFDEX_OBJECT_filterMacro), STRLONG);
			objectP->filterMacro[STRLONG] = 0;
			strncpy(objectP->launchMacro,
					AxStrFromArray(bP, AXELFDEX_OBJECT_launchMacro), STRLONG);
			objectP->launchMacro[STRLONG] = 0;
			objectP->allowUnreferenced =
						AxBoolFromArray(bP, AXELFDEX_OBJECT_allowUnreferenced);
			strncpy(objectP->internalRefName,
					AxStrFromArray(bP, AXELFDEX_OBJECT_internalRef), STRLONG);
			objectP->internalRefName[STRLONG] = 0;

			/* get doc links info for this object */
			objectP->info =
				AxDocLinksInfoFromElf(AxArrayFromArray(bP, 
														AXELFDEX_OBJECT_info));
				/* recurse */
		}
	}

	return(doc_links);
}

/***************************************************************************
function: AxFreeDocLinksInfo 
*/
void
AxFreeDocLinksInfo (doc_links)
doc_links_info_t *doc_links;
{
	int ix;
	link_info_t *linkP;
	object_info_t *objectP;

	if ( !doc_links )
		return;

	/* free link info */
	for ( ix = 0; ix < doc_links->numLinks; ix++ ) {
		linkP = &doc_links->links[ix];
		AxFreeDocLinksInfo(linkP->info);  /* recurse */
	}
	TrueFree(doc_links->links);

	/* free object info */
	for ( ix = 0; ix < doc_links->numObjects; ix++ ) {
		objectP = &doc_links->objects[ix];
		AxFreeDocLinksInfo(objectP->info);  /* recurse */
	}
	TrueFree(doc_links->objects);
}

/***************************************************************************
function: AxLinksInfoToElf 
*/
data_ptr
AxLinksInfoToElf (linksP, numLinks)
link_info_t *linksP;
int numLinks;
{
	int ix;
	data_ptr info;
	data_ptr elfDocLinksInfo;
	data_ptr aP;

	if ( !linksP || (numLinks == 0) )
		return((data_ptr)NULL);

	aP = AxMakeArray(numLinks);
	for ( ix = 0; ix < numLinks; ix++ ) {
		info = AxMakeArray(AXELFSIZE_LINK);
		if ( linksP[ix].name && linksP[ix].name[0] ) {
			info = AxAddStrToArray(info, AXELFDEX_LINK_name, linksP[ix].name);
		}
		info = AxAddIntToArray(info, AXELFDEX_LINK_docType, linksP[ix].docType);
		info = AxAddIntToArray(info, AXELFDEX_LINK_appType, linksP[ix].appType);
		if ( linksP[ix].filterMacro && linksP[ix].filterMacro[0] ) {
			info = AxAddStrToArray(info, AXELFDEX_LINK_filterMacro,
													linksP[ix].filterMacro);
		}
		if ( linksP[ix].launchMacro && linksP[ix].launchMacro[0] ) {
			info = AxAddStrToArray(info, AXELFDEX_LINK_launchMacro,
													linksP[ix].launchMacro);
		}
		if ( linksP[ix].externalRefName && linksP[ix].externalRefName[0] ) {
			info = AxAddStrToArray(info, AXELFDEX_LINK_externalRef,
													linksP[ix].externalRefName);
		}
		if ( linksP[ix].internalRefName && linksP[ix].internalRefName[0] ) {
			info = AxAddStrToArray(info, AXELFDEX_LINK_internalRef,
													linksP[ix].internalRefName);
		}
		elfDocLinksInfo = AxDocLinksInfoToElf(linksP[ix].info);
		if ( elfDocLinksInfo )
			info = AxAddArrayToArray(info, AXELFDEX_LINK_info, elfDocLinksInfo);
		aP = AxAddArrayToArray(aP, ix, info);
	}

	return(aP);
}

/***************************************************************************
function: AxObjectInfoToElf 
*/
data_ptr
AxObjectInfoToElf (objectP, numObjects)
object_info_t *objectP;
int numObjects;
{
	int ix;
	data_ptr info;
	data_ptr elfDocLinksInfo;
	data_ptr aP;

	if ( !objectP || (numObjects == 0) )
		return((data_ptr)NULL);

	aP = AxMakeArray(numObjects);
	for ( ix = 0; ix < numObjects; ix++ ) {
		info = AxMakeArray(AXELFSIZE_OBJECT);
		if ( objectP[ix].name && objectP[ix].name[0] ) {
			info = AxAddStrToArray(info, AXELFDEX_OBJECT_name, objectP[ix].name);
		}
		info = AxAddIntToArray(info, AXELFDEX_OBJECT_docType, objectP[ix].docType);
		info = AxAddIntToArray(info, AXELFDEX_OBJECT_appType, objectP[ix].appType);
		if ( objectP[ix].filterMacro && objectP[ix].filterMacro[0] ) {
			info = AxAddStrToArray(info, AXELFDEX_OBJECT_filterMacro,
													objectP[ix].filterMacro);
		}
		if ( objectP[ix].launchMacro && objectP[ix].launchMacro[0] ) {
			info = AxAddStrToArray(info, AXELFDEX_OBJECT_launchMacro,
													objectP[ix].launchMacro);
		}
		if ( objectP[ix].internalRefName && objectP[ix].internalRefName[0] ) {
			info = AxAddStrToArray(info, AXELFDEX_OBJECT_internalRef,
												objectP[ix].internalRefName);
		}
		info = AxAddBoolToArray(info, AXELFDEX_OBJECT_allowUnreferenced,
												objectP[ix].allowUnreferenced);
		elfDocLinksInfo = AxDocLinksInfoToElf(objectP[ix].info);
		if ( elfDocLinksInfo )
			info = AxAddArrayToArray(info, AXELFDEX_OBJECT_info, elfDocLinksInfo);
		aP = AxAddArrayToArray(aP, ix, info);
	}

	return(aP);
}


/***************************************************************************
function: AxDocLinksInfoToElf
*/
data_ptr
AxDocLinksInfoToElf (doc_links)
doc_links_info_t *doc_links;
{
	data_ptr elfLinkInfo;
	data_ptr elfEmbedInfo;
	data_ptr aP;

	if ( !doc_links )
		return((data_ptr)NULL);

	elfLinkInfo = AxLinksInfoToElf(doc_links->links, doc_links->numLinks);
	elfEmbedInfo = AxObjectInfoToElf(doc_links->objects, doc_links->numObjects);

	aP = (data_ptr)NULL;
	if ( elfLinkInfo || elfEmbedInfo ) {
		aP = AxMakeArray(2);
		if ( elfLinkInfo )
			aP = AxAddArrayToArray(aP, 0, elfLinkInfo);
		if ( elfEmbedInfo )
			aP = AxAddArrayToArray(aP, 1, elfEmbedInfo);
	}
	return(aP);
}

/**************************************************************************\
 *               determine link presence from pathname                    *
\**************************************************************************/

/***************************************************************************
function: doc_has_links 
          May fail if doc contains embedded object with links AND
          embedded object rev was not updated with container.
          Such a file is corrupt, however, and will encounter other
          problems as well!
*/
static int
doc_has_links(name, has_links)
mbchar *name;
int *has_links;
{
	int pid, read_only, err;
	char buf[AREADBUFSIZE];
	FILE *fp;

	/* open the file for reading */
    pid = THIMpid;
    read_only = TRUE;
    err = AFTaskLockFile(name, pid, &read_only, FALSE);
    if ( err != AOK )
	{
		*has_links = FALSE;
		return(FALSE);
	}
    fp = (FILE *)AFTaskfopen(pid, name, READ_BINARY_STR);
    if ( !fp )
    {
		*has_links = FALSE;
        AFTaskUnlockFile(name, pid, read_only);
		return(FALSE);
    }

	/* read the header line */
	fgets(buf, AREADBUFSIZE, fp);
	if ( !buf[0] )
	{
		*has_links = FALSE;
		AFfclose(name, read_only, fp);
		AFTaskUnlockFile(name, pid, read_only);
		return(FALSE);
	}

	/* proceed only if it's a valid APPLIXware doc (*START or *BEGIN) */
	if ( is_start_token(buf) )
	{
		/* search for a link */
		while ( fgets ((char *)buf, AREADBUFSIZE, fp) )
		{
			/* if we find a link, we're done */
			if ( !strncmp(buf, ATOK_LINK, strlen(ATOK_LINK)) )
			{
				*has_links = TRUE;
				AFfclose(name, read_only, fp);
				AFTaskUnlockFile(name, pid, read_only);
				return(TRUE);
			}
		}
	}

	/* no links */
	*has_links = FALSE;
	AFfclose(name, read_only, fp);
	AFTaskUnlockFile(name, pid, read_only);
	return(TRUE);
}

/***************************************************************************
function: WPDocHasLinks 
*/
int
WPDocHasLinks(name)
mbchar *name;
{
	int has_links;
	doc_has_links(name, &has_links);
	return(has_links);
}

/***************************************************************************
function: SSDocHasLinks 
*/
int
SSDocHasLinks(name)
mbchar *name;
{
	int has_links;
	doc_has_links(name, &has_links);
	return(has_links);
}

/***************************************************************************
function: GEDocHasLinks 
*/
int
GEDocHasLinks(name)
mbchar *name;
{
	int has_links;
	doc_has_links(name, &has_links);
	return(has_links);
}

/***************************************************************************
function: AxDocHasLinks 
*/
int
AxDocHasLinks (name)
mbchar *name;
{
	int doc_type, has_links;

	doc_type = (name && name[0]) ? DocTypeFromName(name) : 0;
	switch ( doc_type )
	{
		default:
			has_links = FALSE;
		break;
		case MText:
			has_links = WPDocHasLinks(name);
		break;
		case MSpreadsheet:
			has_links = SSDocHasLinks(name);
		break;
		case MDrawing:
		case MApplixSlideShow:
			has_links = GEDocHasLinks(name);
		break;
	}
	return(has_links);
}

/***************************************************************************
function: AxElfDocHasLinks 
*/
data_ptr AxElfDocHasLinks (name)
mbchar *name;
{
	return(AxMakeNumData(AxDocHasLinks(name) ? ELF_TRUE : ELF_FALSE)); 
}

/**************************************************************************\
 *                   get link info from pathname                          *
\**************************************************************************/

/***************************************************************************
function: get_doc_link_info 
          May fail if doc contains embedded object with links AND
          embedded object rev was not updated with container.
          Such a file is corrupt, however, and will encounter other
          problems as well!
*/
static int
get_doc_link_info(name, ret_info)
mbchar *name;
doc_links_info_t **ret_info;
{
	char template[STRPATH+1], fname[STRPATH+1], buf[AREADBUFSIZE], *link_name;
	AFileInfo file_info;
	int pid, read_only, err, n, sz;
	doc_links_info_t *info;
	object_info_t *objP;
	link_info_t *linkP;
	FILE *fp, *tmp_fp;

	/* open the file for reading */
    pid = THIMpid;
    read_only = TRUE;
    err = AFTaskLockFile(name, pid, &read_only, FALSE);
    if ( err != AOK )
	{
		*ret_info = (doc_links_info_t *)0;
		return(FALSE);
	}
    fp = (FILE *)AFTaskfopen(pid, name, READ_BINARY_STR);
    if ( !fp )
    {
		*ret_info = (doc_links_info_t *)0;
        AFTaskUnlockFile(name, pid, read_only);
		return(FALSE);
    }

	/* read the header line */
	fgets(buf, AREADBUFSIZE, fp);
	if ( !buf[0] )
	{
		*ret_info = (doc_links_info_t *)0;
		AFfclose(name, read_only, fp);
		AFTaskUnlockFile(name, pid, read_only);
		return(FALSE);
	}

	/* proceed only if it's a valid APPLIXware doc (*START or *BEGIN) */
	if ( is_start_token(buf) == FALSE )
		*ret_info = (doc_links_info_t *)0;
	else
	{
		info = (doc_links_info_t *)TaskAlloc(0, sizeof(doc_links_info_t));
		memset(info, 0, sizeof(doc_links_info_t));

		/* search the file, a line at a time */
		while(fgets ((char *)buf, AREADBUFSIZE, fp))
		{
			/* if linked, add link */
			if ( !strncmp(buf, ATOK_LINK, strlen(ATOK_LINK)) )
			{
				link_name = &buf[strlen(ATOK_LINK)+1];
				n = strlen(link_name);
				info->numLinks++;
				sz = info->numLinks * sizeof(link_info_t);
				info->links = (link_info_t *)TaskRealloc(0, info->links, sz);
				linkP = &info->links[info->numLinks-1];
				memset(linkP, 0, sizeof(link_info_t));
				strncpy(linkP->name, link_name, n+1);
				linkP->name[n-1] = (char)0;
				linkP->docType = RecognizeFile(linkP->name);
				linkP->appType = GetAppidFromDocType(linkP->docType);
				linkP->info = AxGetDocLinksInfo(linkP->name);
			}

			/* if embedded, add object */
			else if ( is_start_token(buf) )
			{
				err = AFileInfoFromHeader(buf, &file_info);
				if ( err != AOK )
				{
					/* scan past end of embedded object */
					while(fgets((char *)buf, AREADBUFSIZE, fp))
					{
						if ( is_end_token(buf) )
							break;
					}
				}
				else
				{
					template[0] = (char)0;
					AxAppendTemplateStr(template, file_info.docType);
					if ( AxTempFile(template, fname) == FALSE )
					{
						/* scan past end of embedded object */
						while(fgets((char *)buf, AREADBUFSIZE, fp))
						{
							if ( is_end_token(buf) )
								break;
						}
					}
					else
					{
						/* write the embedded file to disk */
						tmp_fp = fopen(fname, WRITE_TRUNC_STR);
						if ( !tmp_fp )
						{
							/* scan past end of embedded object */
							while(fgets((char *)buf, AREADBUFSIZE, fp))
							{
								if ( is_end_token(buf) )
									break;
							}
						}
						else
						{
							while(fgets((char *)buf, AREADBUFSIZE, fp))
							{
								/* write it */
								fputs(buf, tmp_fp);
								fputc((int)'\n', tmp_fp);
								if ( is_end_token(buf) )
									break;
							}
							fclose(tmp_fp);

							info->numObjects++;
							sz = info->numObjects * sizeof(object_info_t);
							info->objects = (object_info_t *)
							      TaskRealloc(0, info->objects, sz);
							objP = &info->objects[info->numObjects-1];
							memset(objP, 0, sizeof(object_info_t));
							objP->name[0] = (char)0;
							objP->docType = file_info.docType;
							objP->appType = GetAppidFromDocType(objP->docType);
							objP->info = AxGetDocLinksInfo(fname);
						}

						/* remove tempfile */
						AxDeleteTempfile(fname);
					}
				}
			}
		}
		*ret_info = info;
	}

	/* done */
	AFfclose(name, read_only, fp);
	AFTaskUnlockFile(name, pid, read_only);
	return(TRUE);
}

/***************************************************************************
function: WPGetDocLinksInfo 
*/
doc_links_info_t *
WPGetDocLinksInfo(name)
mbchar *name;
{
	doc_links_info_t *info;
	data_ptr aP, uidP;
	typeErr err;

	/* Create an new WP instance */
	err = ElfbCreateWpInstance((data_ptr)0);
	if ( err != AOK )
		info = (doc_links_info_t *)0;
	else
	{
		/* Read in the specified file */
		uidP = ElfRetData;
		aP = AxMakeArray(2); /* need to free? */
		AxAddDataToArray(aP, 0, uidP);
		AxAddStrToArray(aP, 1, name);
		err = ElfbReadWpInstanceFile(aP);

		/* Fill the doclinks struct */
		if ( err != AOK )
			info = (doc_links_info_t *)0;
		else
		{
			info = (doc_links_info_t *)TrueMalloc(sizeof(doc_links_info_t));
			memset(info, 0, sizeof(doc_links_info_t));
			err = WpxGetDocLinksInfo(uidP, info); /* See wpinset.c */
		}

		/* Delete the instance */
		aP = AxMakeArray(1); /* need to free? */
		AxAddDataToArray(aP, 0, uidP);
		ElfbDestroyWpInstance(aP);
	}

	/* done */
	return(info);
}

/***************************************************************************
function: SSGetDocLinksInfo 
*/
doc_links_info_t *
SSGetDocLinksInfo (name)
mbchar *name;
{
	doc_links_info_t *doc_links;
	typeErr err;
	int uid; 
	void *ssP;

	doc_links = (doc_links_info_t *)TrueMalloc(sizeof(doc_links_info_t));
	memset(doc_links, 0, sizeof(doc_links_info_t));

	/* create SS instance */
	uid = SSCreateSSX(0, FALSE);

	/* read file */
	err = SSXReadFile(uid, name, name);
	if(err != AOK)	{
		SSFreeSSXbyUID(uid);
		return(NULL);
	}

	/* get link info */
	SSGetSSXFromUid(uid, &ssP);
	SSGetLinks(ssP, name, doc_links);

	/* free ss instance */
	SSFreeSSXbyUID(uid);

	return(doc_links);
}

/***************************************************************************
function: GEGetDocLinksInfo
*/
doc_links_info_t *
GEGetDocLinksInfo(name)
mbchar *name;
{
	doc_links_info_t *info, *GrGetDocLinksInfo();
	data_ptr aP, uidP;
	typeErr err;
	int uid;

	/* Create an new GE instance (owned by task/uid zero) */
	aP = AxMakeArray(1);
	AxAddIntToArray(aP, 0, 0);
	err = ElfbCreateGraphic((data_ptr)0);
	AxFreeData(aP);
	if ( err != AOK )
		info = (doc_links_info_t *)0;
	else
	{
		/* get the uid of the graphic from the global return */
		uid = AxIntFromDataPtr(ElfRetData);
		AxFreeData(ElfRetData);
		ElfRetData = (data_ptr)0;

		/* Read in the specified file */
		aP = AxMakeArray(5);
		AxAddIntToArray(aP, 0, uid);  /* uid                   */
		AxAddStrToArray(aP, 1, name); /* pathname of file      */
		AxAddIntToArray(aP, 2, (-1)); /* read only set to true */
		AxAddStrToArray(aP, 4, name); /* pathname of file      */
		err = ElfbReadGraphicFile(aP);
		AxFreeData(aP);

		/* Fill the doclinks struct */
		if ( err != AOK )
			info = (doc_links_info_t *)0;
		else
			info = GrGetDocLinksInfo(AxGetGFXfromUID(uid));

		/* Delete the instance */
		aP = AxMakeArray(1);
		AxAddIntToArray(aP, 0, uid);
		ElfbDestroyGraphic(aP);
		AxFreeData(aP);
	}

	/* done */
	return(info);
}

/***************************************************************************
function: AxGetDocLinksInfo 
*/
doc_links_info_t *
AxGetDocLinksInfo(name)
mbchar *name;
{
	doc_links_info_t *doc_links;
	int doc_type;

	/* If we can't access the file, assume it has no links.
		Note moving a file with links to another location may
		well put us in this situation. We do NOT want to throw
		an error and terminate the calling task (see 122783). */
	/* We do NOT handle URLs here. The caller should not pass URLs
		to us. */
	if (Aaccess(name,R_OK)) {
		doc_links = (doc_links_info_t *)0;
		return(doc_links);
	}

	doc_type = (name && name[0]) ? DocTypeFromName(name) : 0;
	switch ( doc_type )
	{
		default:
			doc_links = (doc_links_info_t *)0;
		break;
		case MText:
			doc_links = WPGetDocLinksInfo(name);
		break;
		case MSpreadsheet:
			doc_links = SSGetDocLinksInfo(name);
		break;
		case MDrawing:
		case MApplixSlideShow:
			doc_links = GEGetDocLinksInfo(name);
		break;
	}
	return(doc_links);
}

/***************************************************************************
function: AxElfGetDocLinksInfo 
*/
data_ptr AxElfGetDocLinksInfo (name)
mbchar *name;	/* doc name */
{
	doc_links_info_t *doc_links;
	data_ptr retP;

	doc_links = AxGetDocLinksInfo(name);
	if ( doc_links == (doc_links_info_t *)0 )
		retP = (data_ptr)0;
	else
	{
		retP = AxDocLinksInfoToElf(doc_links);
		AxFreeDocLinksInfo(doc_links);
	}
	return(retP);
}

/**************************************************************************\
 *                        change link info                                *
\**************************************************************************/

/***************************************************************************
function: WPChangeDocLinksInfo 
*/
typeErr
WPChangeDocLinksInfo (name, doc_links)
mbchar *name;
doc_links_info_t *doc_links;
{
	data_ptr uidP, aP;
	typeErr err;

	if ( !doc_links )
		return(AOK);

	/* Create an new WP instance */
	err = ElfbCreateWpInstance( (data_ptr)NULL);
	if (err != AOK)
		return(err);

	/* Read in the specified file */
	uidP = ElfRetData;
	aP = AxMakeArray(2); /* need to free */
	AxAddDataToArray(aP, 0, uidP);
	AxAddStrToArray(aP, 1, name);
	err = ElfbReadWpInstanceFile( aP);

	/* Perform the alterations */
	if ( err == AOK )
		err = WpxChangeDocLinksInfo(uidP, doc_links, NULL); /* See wpinset.c */

	/* Delete the instance */
	aP = AxMakeArray(1); /* need to free */
	AxAddDataToArray(aP, 0, uidP);
	ElfbDestroyWpInstance(aP);

	/* done */
	return(err);
}

/***************************************************************************
function: SSChangeDocLinksInfo - changes the links in the specified document
		and its children. 
		Note: It is assumed that objects don't have links.
*/
typeErr
SSChangeDocLinksInfo (name, doc_links)
mbchar *name;
doc_links_info_t *doc_links;
{
	link_info_t *linkP;
	typeErr err;
	int uid;
	void *ssP;

	if(!doc_links)
		return(AOK);

	/* create SS instance */
	uid = SSCreateSSX(0, FALSE);

	/* read file */
	err = SSXReadFile(uid, name, name);
	if(err != AOK)	{
		SSFreeSSXbyUID(uid);
		return(err);
	}

	/* change links */
	SSGetSSXFromUid(uid, &ssP);
	SSChangeLinks(ssP, doc_links);

	/* write file */
	err = SSXWriteFile(uid, name);

	/* free ss instance */
	SSFreeSSXbyUID(uid);
	
	return(err);
}

/***************************************************************************
function: GEChangeDocLinksInfo - changes the links in the specified document
*/
typeErr
GEChangeDocLinksInfo(name, doc_links)
mbchar *name;
doc_links_info_t *doc_links;
{
	int uid, GrSetDocLinksInfo();
	data_ptr aP;
	typeErr err;

	if ( !doc_links )
		return(AOK);

	/* Create an new GE instance (owned by task/uid zero) */
	aP = AxMakeArray(1);
	AxAddIntToArray(aP, 0, 0);
	err = ElfbCreateGraphic((data_ptr)0);
	AxFreeData(aP);
	if ( err == AOK )
	{
		/* get the uid of the graphic from the global return */
		uid = AxIntFromDataPtr(ElfRetData);
		AxFreeData(ElfRetData);
		ElfRetData = (data_ptr)0;

		/* Read in the specified file */
		aP = AxMakeArray(5);
		AxAddIntToArray(aP, 0, uid);  /* uid                    */
		AxAddStrToArray(aP, 1, name); /* pathname of file       */
		AxAddIntToArray(aP, 2, 0);    /* read only set to false */
		AxAddStrToArray(aP, 4, name); /* pathname of file       */
		err = ElfbReadGraphicFile(aP);
		AxFreeData(aP);

		/* Perform the alterations */
		if ( err == AOK )
		{
			err = GrSetDocLinksInfo(AxGetGFXfromUID(uid), doc_links);
			if ( err == AOK )
			{
				aP = AxMakeArray(2);
				AxAddIntToArray(aP, 0, uid);
				AxAddStrToArray(aP, 1, name);
				ElfbWriteGraphicFile(aP);
				AxFreeData(aP);
			}
		}

		/* Delete the instance */
		aP = AxMakeArray(1);
		AxAddIntToArray(aP, 0, uid);
		ElfbDestroyGraphic(aP);
		AxFreeData(aP);
	}

	/* done */
	return(err);
}

/***************************************************************************
function: AxChangeDocLinksInfo 
*/
typeErr
AxChangeDocLinksInfo (name, doc_links)
mbchar *name;
doc_links_info_t *doc_links;
{
	int doc_type;
	typeErr err;

	doc_type = (name && name[0]) ? DocTypeFromName(name) : 0;
	switch ( doc_type )
	{
		default:
			err = FEENOENT;
		break;
		case MText:
			err = WPChangeDocLinksInfo(name, doc_links);
		break;
		case MSpreadsheet:
			err = SSChangeDocLinksInfo(name, doc_links);
		break;
		case MDrawing:
		case MApplixSlideShow:
			err = GEChangeDocLinksInfo(name, doc_links);
		break;
	}
	return(err);
}

/***************************************************************************
function: AxElfChangeDocLinksInfo 
*/
void
AxElfChangeDocLinksInfo (name, elfDocLinksInfo)
mbchar *name;				/* doc name */
data_ptr elfDocLinksInfo;	/* elf doc links info */
{
	doc_links_info_t *doc_links;
	typeErr err;

	doc_links = AxDocLinksInfoFromElf(elfDocLinksInfo);
	err = AxChangeDocLinksInfo(name, doc_links);
	AxFreeDocLinksInfo(doc_links);
	if ( err != AOK )
		ElfAbort((int)err);
}

/**************************************************************************\
 *                        localize (embed)                                *
\**************************************************************************/

/***************************************************************************
function: WPLocalizeDocLinks 
*/
typeErr
WPLocalizeDocLinks (name, doc_links, savefile, parentdoc)
mbchar *name;					/* the document to open */
doc_links_info_t *doc_links;	/* doc links info of the doc to localize */
mbchar *savefile;				/* the file to save to - may be NULL or "" */
mbchar *parentdoc;				/* parent doc (used to get original path) */
{
	data_ptr uidP, aP;
	typeErr err;

	/*** No need to bring up a separate task - use built-in Instance fcns */

	/* Create an new WP instance */
	err = ElfbCreateWpInstance( (data_ptr)NULL);
	if (err != AOK)
		return(err);
	uidP = ElfRetData; /* need to free */

	/* Read in the specified file */
	aP = AxMakeArray(2); /* need to free */
	AxAddDataToArray(aP,0,uidP);
	AxAddStrToArray(aP,1,name);
	err = ElfbReadWpInstanceFile( aP);
	if ( err == AOK ) /* Perform the localization - See wpinset.c */
		err = WpxLocalizeDocLinks (uidP, name, doc_links, savefile, parentdoc);

	/* Delete the instance */
	aP = AxMakeArray(1); /* need to free */
	AxAddDataToArray(aP,0,uidP);
	ElfbDestroyWpInstance( aP);

	return( err);
}

/***************************************************************************
function: SSLocalizeDocLinks 
*/
typeErr
SSLocalizeDocLinks (name, savefile, parentdoc)
mbchar *name;      /* the document to open */
mbchar *savefile;  /* the file to save to - may be NULL or "" */
mbchar *parentdoc; /* parent doc (used to get original path) */
{
	typeErr err;
	int uid;
	char *fileToWrite;
	void *ssP;

	if ( savefile && savefile[0] ) {
		fileToWrite = savefile;
	}	else {
		fileToWrite = name;
	}

	/* create SS instance */
	uid = SSCreateSSX(0, FALSE);


	/* read file */
	err = SSXReadFile(uid, name, parentdoc);
	if(err != AOK)	{
		SSFreeSSXbyUID(uid);
		return(err);
	}

	/* change links */
	SSGetSSXFromUid(uid, &ssP);
	SSLocalizeLinks(ssP, parentdoc);

	/* write file */
	err = SSXWriteFile(uid, fileToWrite);

	/* free ss instance */
	SSFreeSSXbyUID(uid);
	return(err);
}

/***************************************************************************
function: GELocalizeDocLinks 
*/
typeErr
GELocalizeDocLinks (name, savefile, parentdoc)
mbchar *name;      /* the document to open */
mbchar *savefile;  /* the file to save to - may be NULL or "" */
mbchar *parentdoc; /* parent doc (used to get original path) */
{
	int uid, GrLocalizeDocLinks();
	data_ptr aP;
	typeErr err;

	/* Create an new GE instance (owned by task/uid zero) */
	aP = AxMakeArray(1);
	AxAddIntToArray(aP, 0, 0);
	err = ElfbCreateGraphic((data_ptr)0);
	AxFreeData(aP);
	if ( err == AOK )
	{
		/* get the uid of the graphic from the global return */
		uid = AxIntFromDataPtr(ElfRetData);
		AxFreeData(ElfRetData);
		ElfRetData = (data_ptr)0;

		/* Read in the specified file */
		aP = AxMakeArray(5);
		AxAddIntToArray(aP, 0, uid);       /* uid                    */
		AxAddStrToArray(aP, 1, name);      /* pathname of file       */
		AxAddIntToArray(aP, 2, 0);         /* read only set to false */
		AxAddStrToArray(aP, 4, parentdoc); /* original path of file  */
		err = ElfbReadGraphicFile(aP);
		AxFreeData(aP);

		/* Perform the alterations */
		if ( err == AOK )
		{
			err = GrLocalizeDocLinks(AxGetGFXfromUID(uid));
			if ( err == AOK )
			{
				aP = AxMakeArray(2);
				AxAddIntToArray(aP, 0, uid);
				if ( savefile && savefile[0] )
					AxAddStrToArray(aP, 1, savefile);
				else
					AxAddStrToArray(aP, 1, name);
				ElfbWriteGraphicFile(aP);
				AxFreeData(aP);
			}
		}

		/* Delete the instance */
		aP = AxMakeArray(1);
		AxAddIntToArray(aP, 0, uid);
		ElfbDestroyGraphic(aP);
		AxFreeData(aP);
	}

	/* done */
	return(err);
}

/***************************************************************************
function: AxLocalizeDocLinks - Embed links in document
* 
* Before the document's links are localized, the links themselves
* are localized (in case they contain links).  If the document
* contains objects, the objects' links are also localized.
* If savefile is passed, the actual documents are not modified.
*
* IF THERE ARE NO LINKS, THE DOC IS JUST COPIED TO THE PASSED SAVEFILE.
*/
typeErr
AxLocalizeDocLinks (name, doc_links, savefile, parentdoc)
mbchar *name;					/* the doc to localize */
doc_links_info_t *doc_links;	/* doc links info of the doc to localize */
mbchar *savefile;				/* the file to save to after localizing */
mbchar *parentdoc;				/* parent document(used to get original path) */
{
    int doc_type, has_links;
    typeErr err;

	/* files with no links are simply copied (if save name is passed) */
	if ( !name || !name[0] )
		return(AOK);

	doc_type = DocTypeFromName(name);
	switch ( doc_type )
	{
		default:
			err = AOK;
		break;
		case MText:
			err = WPLocalizeDocLinks(name, doc_links, savefile, parentdoc);
		break;
		case MSpreadsheet:
			err = SSLocalizeDocLinks(name, savefile, parentdoc);
		break;
		case MDrawing:
		case MApplixSlideShow:
			err = GELocalizeDocLinks(name, savefile, parentdoc);
		break;
	}
    return(err);
}

/***************************************************************************
function: AxElfLocalizeDocLinks 
*/
void
AxElfLocalizeDocLinks (name, elfDocLinksInfo, savefile, parentdoc)
mbchar *name;				/* doc name */
data_ptr elfDocLinksInfo;	/* elf doc links info */
mbchar *savefile;			/* save file */
mbchar *parentdoc;			/* parent document (used to get original path) */
{
	typeErr err;
	doc_links_info_t *doc_links;

	err = AOK;
	doc_links = AxDocLinksInfoFromElf(elfDocLinksInfo);
	err = AxLocalizeDocLinks(name, doc_links, savefile, parentdoc);
	AxFreeDocLinksInfo(doc_links);
	if ( err != AOK )
		ElfAbort((int)err);
}

/**************************************************************************\
 *  routines for merging dir and relative path into an absolute pathname  *
\**************************************************************************/

/***************************************************************************
function: AxTaskMergePaths - Merge 'partial' pathname into base pathname.

The 'base' pathname is the full pathname of a document (must be an absolute
path or URL); e.g. '/home/me/foo.aw' or 'http://www.foo.com/foo.aw'

The 'partial' pathname is a reference to a file and can be in absolute or
relative format; e.g. '/home/me/foo.aw' or 'http://www.foo.com/foo.aw' or
'foo.aw' or '../foo.aw'

The merged pathname is returned as an allocated string
*/
char *
AxTaskMergePaths(pid, base_path, partial_path)
int pid;
char *base_path;
char *partial_path;
{
	data_ptr baseP, pathP, mergedP = NULL;
	char *merged_path, *result;
	typeErr err, size;

	baseP = AxMakeStrData((-1), base_path);
	pathP = AxMakeStrData((-1), partial_path);
	result = (char *)0;
	err = ElfCallByName(NXLT("WWW_MERGE_URL@"), &mergedP, 2, pathP, baseP);
	if ( err == AOK )
	{
		merged_path = AxStrPtrFromDataPtr(mergedP);
		if ( merged_path && merged_path[0] != (char)0 )
		{
			size = (strlen(merged_path)+1) * sizeof(char);
			result = (char *)TaskAlloc(pid, size);
			strcpy(result, merged_path);
		}
	}
	if (mergedP)
		AxFreeData(mergedP);
	AxFreeData(baseP);
	AxFreeData(pathP);
	return(result);
}

/***************************************************************************
function: AxMergePaths - see AxTaskMergePaths above
*/
char *
AxMergePaths(base, partial)
char *base, *partial;
{
	return(AxTaskMergePaths(THIMpid, base, partial));
}
