/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pathtool.c
   Applix ASC version: 16
   Date: Mon Nov 16 13:53:25 1998
*/
#include <aos.h>

#define DOS_SLASH '\\'
#define UNIX_SLASH '/'
#if AX_NT
#define SLASH_CHAR DOS_SLASH
#else
#define SLASH_CHAR UNIX_SLASH
#endif


/*************************************************************************
function: AxSysDosUnixPath - Convert / to \ or visa versa
                             Also strip a leading dot from a file name for NT
*/
void AxSysDosUnixPath (name)
char *name;
{
#if AX_NT
    char *p1, *p2;
    p1 = p2 = name;
    while ( *p2 ) 
        {
#if 0
        *** we should not longer need this hack, put in to handle issues when
        *** the profile files were 'dot' files
        if ( (*p2 == '.') && (p2 > name) && (*(p2-1) == DOS_SLASH) )
            {
            /* we have a dot preceeded by a slash, skip the dot to strip off */
            /*  the leading dot char from the filename                       */
            p2++;
            }
#endif
        if ( *p2 == UNIX_SLASH )
            {
            *p1++ = DOS_SLASH;
            p2++;
            }
        else
            /* a regular character */
            *p1++ = *p2++;
        }
     *p1 = '\0';

    if ( (strlen(name) == 2) && (*(name+1) == ':') )
        {
        /* all we have is a drive specifier, make it a "real" location */
        /* by adding a trailing slash                                  */
        strcat( name, "\\" );
        }
  
#else
    while ( *name ) {
        if ( *name == DOS_SLASH )
            *name = UNIX_SLASH;
        name++;
    }
#endif
}

/*************************************************************************
function: AxMakePath - Join directory and Entry names into a pathname
Result is a passed buffer ptr, this returns same ptr (like strcpy).
On Unix, this just cats with a "/" separator.
*/
char *AxMakePath (result,dirname,entryname)
char *result;      /* resulting full pathname */
char *dirname;     /* orginal directory name */
char *entryname;   /* file entry name */
{
    if (dirname == NULL) dirname =".";
#if AX_NT
    Asprintf(result,"%s\\%s",dirname,entryname);
#else
    Asprintf(result,"%s/%s",dirname,entryname);
#endif
    AxSysDosUnixPath( result );    /* just being safe! */
    return(result);
}

/*************************************************************************
function: AxParsePath - Convert a pathname to Dir & entry names
Works fastest if dir & path are same variable
*/
void AxParsePath (path,dir,ent)
char *path;
char *dir;  /* result directory */
char *ent;  /* result entry file name */
{
    register char *p;
    register int len;

    AxSysDosUnixPath( path );    /* just being safe! */
    p = strrchr(path,SLASH_CHAR);
    if ( !p ) {
        strcpy(dir,".");
        strcpy(ent,path);
        return;
    }
    strcpy(ent,p+1);
    if ( path == dir )
        *p = 0;
    else {
        len = p-path;
        memcpy(dir,path,len);
        dir[len] = 0;
    }
}

/*************************************************************************
function: AxReplaceFileExtension - Change file's extension
The passed extension should NOT have a dot
Adds new extension if none exists.
*/
char *AxReplaceFileExtension (name,ext)
char *name;
char *ext;
{
    register char *p;

    p = strrchr(name,'.');
    if ( !p ) {
        strcat(name,".");
        strcat(name,ext);
    }
    else
        strcpy(p+1,ext);
    return(name);
}

#if 0
    /* 8.3 filename issue is OBSOLETE! ***
     * (or do we still want to support it?)  -donato **/
/*************************************************************************
function: IsLegalPath - Check for 8.3 violations
*/
IsLegalPath (name)
char *name;
{
#if AX_NT
    int cnt=0, extcnt=0, got_dot=FALSE;

    AxSysDosUnixPath( name );    /* being safe */

/* look for names too long */

    while ( *name ) {
        if ( (*name == SLASH_CHAR) || (*name == ':') ) {
            cnt=0;
            extcnt=0;
            got_dot = FALSE;
        }
        else if ( *name == '.' ) {
            if ( got_dot )
                return(FALSE);
            got_dot = TRUE;
        }
        else if ( got_dot ) {
            extcnt++;
            if ( extcnt > 3 )
                return(FALSE);
        }
        else {
            cnt++;
            if ( cnt > 8 )
                return(FALSE);
        }
        name++;
    }
#endif
    return(TRUE);
}
#endif
