/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: rpc_svr.c
   Applix ASC version: 85
   Date: Sun Jan 31 21:12:24 1999
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <axtime.h>
#include <signal.h>

#if !AXMS
#include <sys/param.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#endif

#if RIOS
#include <sys/select.h>
#endif

#if ALPHA
#include <xti.h>				/* For t_open, t_bind, t_close */
#else
#if !AXMS && !LINUX /* TLI interface yet on Linux!! */
#include <tiuser.h>				/* For t_open, t_bind, t_close */
#endif
#endif

#include <axsocket.h>
#include <errno.h>
#include <axnet.h>

#define THIM_STATIC 1    /** do not redefine such things as "exit()"  */
#include <aos.h>
#include <elfapi.h>
#if !ELF_COMM_STAND_ALONE
#include <thim_gen.h>
#endif

#if !AXMS
#include <arpa/inet.h>
#endif

#if !LINUX
extern char *sys_errlist[]; /* Unix errro strings */
#endif

#define NEED_TO_WRITE(x) x && (x->written < x->read)
#define MUX_BLOCK_SIZE 2 /* MUST BE EVEN NUM: THERE ARE 2 MUX ENTRIES PER JOB */
#define MSG_SIZE 8*1024 /* 8k for all message buffers */
#define MAX_MUX 10000 /* if we hit this number something has gone awry! */
#define NXLT(str) str   /*  Keep xlt scans sane by NXLT'g where necess */
#define XLT(str) str

/* if you add items here be sure to clear them in AxProxyClear! */
typedef struct buf_list_struct {
    char msg[MSG_SIZE]; 
    int written;            /* how much stuff has been written out */
    int read;               /* how much stuff has been read in */
    struct buf_list_struct *next;
} buf_list_t;

/* if you add items here be sure to clear them in AxProxyClear! */
typedef struct {
    /* the guy we're writing to, eg a channel someone opened
       to us; the java client */
    int clientCh;
    buf_list_t *clientBuf;

    /* the guy we're reading from, eg a channel we opened to
       somewhere else; iasnet or applixware on a remote box */
    int serverCh;
    buf_list_t *serverBuf;
    int cookie;
    int isRemoteSvr;        /* if TRUE, proxy to other aagtwy */
    int inUse;
} mux_entry_t;

#include <rpc_svr.p>

static mux_entry_t *ProxyList = NULL;
static int ProxyListSize = 0; /* size of array */
static int ProxyListLast = 0; /* last item in use in array */
static int Proxying = FALSE;

/*************************************************************************
function: AxProxy - 
*/
/*************************************************************************
function: AxProxyNonBlock - make channel non blocking 
*/
int AxProxyNonBlock(channel)
int channel;
{
    int flags;
    
#if AX_X11   /* {  */
#if SUN4 && !SUN4S  /* Posix style nonblock does not seem to work for SunOS */
    return(0 == fcntl(channel,F_SETFL,O_NDELAY));   /* bsd4.2 non-blocking only! */
#else
    flags = fcntl(channel,F_GETFL,0);
    if( flags < 0 ){
        perror("Error getting flags on channel");
        return(0);
    }
    flags |= O_NONBLOCK;
    if( fcntl(channel,F_SETFL,flags) < 0){
        perror("Error getting flags on channel");
        return(0);
    }
    return(TRUE);

#endif
#else     /*  }  {  */
     /* MS-Windows,  NT  */
    /* setting up the use of async messages puts the socket in non-blocking */
    /*   mode.                                                   */
#if !ELF_COMM_STAND_ALONE
    SetSocketAsyncCallback( channel );
#else
    unsigned long non_block = 1;
    return (0 == ioctlsocket(channel, FIONBIO, &non_block));
#endif
    return( TRUE );
#endif    /*  }  AX_X11 / AXMS  */

}

/*************************************************************************
function: AxProxyRealloc - expand array of channels to proxy
*/
int AxProxyRealloc()
{
    int ix;

    ProxyListSize += MUX_BLOCK_SIZE;
    if( ProxyListSize >= MAX_MUX ){
        fprintf(stderr,XLT("Fatal error: exceeded sock proxy limit of %d items\n"),MAX_MUX);
        return(0);
    }
    if( !ProxyList )
        ProxyList = (mux_entry_t *)AxMalloc(ProxyListSize*sizeof(mux_entry_t));
    else
        ProxyList = (mux_entry_t *)AxRealloc((char*)ProxyList,ProxyListSize*sizeof(mux_entry_t) );
    if( !ProxyList ){
        fprintf(stderr,XLT("Fatal error: unable to malloc %d bytes\n"),
                ProxyListSize*sizeof(mux_entry_t));
        return(FALSE);
    }
    for( ix=ProxyListSize-MUX_BLOCK_SIZE; ix<ProxyListSize; ix++ )
        memset( ProxyList+ix, 0, sizeof(mux_entry_t));
    return(TRUE);
}

/*************************************************************************
function: AxProxyResetCounter() - set index of last entry in use
For performance intensive functions that need to loop over the list, this
minimizes loops over unused entries.
*/
void AxProxyResetCounter()
{
    /* adjust item count - helps optimize search loops in other AxProxy code */
    for( ProxyListLast=ProxyListSize; ProxyListLast > 1; ProxyListLast-- )
        if( ProxyList[ProxyListLast-1].inUse )
            return;
    ProxyListLast=1; /* empty list */
}

/*************************************************************************
function: AxProxyClear - delete entry in array of channels to proxy
*/
void AxProxyClear(idx)
int idx;
{
    buf_list_t *bP,*tmpP;
    int ix;

    /* clear client */
    if ( ProxyList[idx].clientCh )
        CLOSE_CHANNEL(ProxyList[idx].clientCh);
    for (bP = ProxyList[idx].clientBuf; bP; bP = tmpP ) {
        tmpP = bP->next;
        AxFree((char*)bP);
    }

    /* clear server */
    if ( ProxyList[idx].serverCh )
        CLOSE_CHANNEL(ProxyList[idx].serverCh);
    for ( bP = ProxyList[idx].serverBuf; bP; bP = tmpP ) {
        tmpP = bP->next;
        AxFree((char*)bP);
    }
    memset(&ProxyList[idx],0,sizeof(mux_entry_t));
    AxProxyResetCounter();
}

int CookieUid = 500;    /* for now, just increments, should rand() */

/*************************************************************************
function: AxProxyGetSvrCount - Return number of server channels
*/
int AxProxyGetSvrCount ()
{
    int ix, cnt = 0;

    for ( ix=0; ix<ProxyListLast; ix++ )
        if ( ProxyList[ix].inUse )
            cnt++;
    return(cnt);
}


/*************************************************************************
function: AxProxyGetLocalSvrCount - Return number of local server channels
*/
int AxProxyGetLocalSvrCount ()
{
    int ix, cnt = 0;

    for ( ix=0; ix<ProxyListLast; ix++ )
        if ( ProxyList[ix].inUse && !ProxyList[ix].isRemoteSvr )
            cnt++;
    return(cnt);
}

/*************************************************************************
function: AxProxyAddFD - Add matching channels to proxy table
Code that needs proxying should call this routine with two open channels
and set *quitNow = 3.  Returns 0 on failure.
*/
int AxProxyAddFD(serverFD,clientFD)
int serverFD;
int clientFD;
{
    int ix;

    /* find a free entry */
    for( ix=0; ix<ProxyListSize && ProxyList[ix].inUse; ix++ )
        ;
    if( ProxyListSize == ix )
        /* list is full, attempt to expand list */
        if( !AxProxyRealloc() )
            return(0);

    if( !AxProxyNonBlock(serverFD) ||
        !AxProxyNonBlock(clientFD) ){
        fprintf(stderr,XLT("Fatal error: unable to make sock non-blocking\n"));
        return(0);
    }

    Proxying = TRUE;

    ProxyList[ix].serverBuf = NULL;
    ProxyList[ix].serverCh = serverFD; /* generally will be axmain */
    ProxyList[ix].clientBuf = NULL;
    ProxyList[ix].clientCh = clientFD; /* AxChannel(), or java client */
    ProxyList[ix].isRemoteSvr = (0 != serverFD);
    ProxyList[ix].inUse = TRUE;
	ProxyList[ix].cookie = ++CookieUid;
    AxProxyResetCounter();
	return(ProxyList[ix].cookie);
}

/*************************************************************************
function: AxProxySet - for select(), add mux channels to main server channels
*/
void AxProxySet(rfds,wfds)
fd_set *rfds;
fd_set *wfds;
{
    int ix;

    for( ix=0; ix<ProxyListLast; ix++ )
        if( ProxyList[ix].inUse ){
            /* we always read on these */
            FD_SET(ProxyList[ix].serverCh,rfds);
            FD_SET(ProxyList[ix].clientCh,rfds);
            /* we only write on these if we have read something */
            if( NEED_TO_WRITE(ProxyList[ix].serverBuf) )
                FD_SET(ProxyList[ix].serverCh,wfds);
            else
                FD_CLR(ProxyList[ix].serverCh,wfds);
            if( NEED_TO_WRITE(ProxyList[ix].clientBuf) )
                FD_SET(ProxyList[ix].clientCh,wfds);
            else
                FD_CLR(ProxyList[ix].clientCh,wfds);
        }
}
  

/*************************************************************************
function: AxProxyFindBad - find && delete any mux channels that have gone bad
*/
int AxProxyFindBad()
{
    int ix;
    fd_set rfds, wfds;
    struct timeval timeout;
    int events;

    /* instant timeout for busy times */
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    for( ix=0; ix<ProxyListLast; ix++ )
        if( ProxyList[ix].inUse ){

            FD_ZERO(&rfds);
            FD_ZERO(&wfds);
            
            /* we always read on these */
            FD_SET(ProxyList[ix].serverCh,&rfds);
            FD_SET(ProxyList[ix].clientCh,&rfds);

            /* we only write on these if we have read something */
            if( NEED_TO_WRITE(ProxyList[ix].serverBuf) )
                FD_SET(ProxyList[ix].serverCh,&wfds);
            if( NEED_TO_WRITE(ProxyList[ix].clientBuf) )
                FD_SET(ProxyList[ix].clientCh,&wfds);
            events = select(FD_SETSIZE,&rfds,&wfds,NULL,&timeout);
            if ( events <= 0 && errno == EBADF ){
                AxProxyClear(ix);
                return(TRUE);
                }
        }
    return(FALSE);
}

/*************************************************************************
function: AxProxySelect - between calls to AxDispatch, route mux messages
*/
void AxProxySelect()
{
    int ix;
    fd_set rfds, wfds;
    struct timeval timeout;
    int events;

    /* instant timeout for busy times */
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;

    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    AxProxySet(&rfds,&wfds);
    events = select(FD_SETSIZE,&rfds,&wfds,NULL,&timeout);
    if( events > 0 )
        AxProxyMain(&rfds,&wfds);
}

/*************************************************************************
function: AxProxyWrite - a write channel is active, give it the waiting data
Return false if fatal error occurs
*/
int AxProxyWrite(channel,headP)
int channel;
buf_list_t **headP;
{
    int cnt = 0;
    buf_list_t *bP;

    bP = *headP;
    
    /* now the message itself */
    while( bP->written < bP->read ){
#if AX_X11
        cnt = write(channel, bP->msg + bP->written, bP->read - bP->written);
#else
        cnt = send(channel, bP->msg + bP->written, bP->read - bP->written,0);
#endif
        /* !!0 this would block stuff is very OS dependent! See reference
           code in axsocket.c (look for reqv & readv) and elsock.c */
        if( cnt == -1 ) {
#if AXMS
            DWORD last_err = WSAGetLastError();
            if ( last_err == WSAEWOULDBLOCK ) {
#else
            if( errno == EWOULDBLOCK ) {
#endif
                return(TRUE);
            } else {
                perror(XLT("Proxy socket error"));
                return(FALSE);
            }
        }
        bP->written += cnt;
    }

    /* we have now written out every piece of data we have read for
       this message. If we have filled the entire message buffer, then
       we are done with it. */
    if( bP->read == MSG_SIZE ){
        *headP = bP->next;
        AxFree((char*)bP);
    }

    return(TRUE);
}

/*************************************************************************
function: AxAddMessageForProxy - Apply passed message buffer to a proxy channel
*/
void AxAddMessageForProxy (int cookie, elfData ebP)
{
    int size;
    char *buf;
    int ix, xferCnt;
    buf_list_t *bP;

    size = AxBinaryFromDataPtr(ebP, &buf);

    for ( ix = 0; ix < ProxyListLast; ix++ )
        if ( ProxyList[ix].inUse && ProxyList[ix].cookie == cookie )
            break;
    if ( ix == ProxyListLast )
        return; /* just ignore request to send message deleted proxy */

    /* handle empty list */
    if ( !ProxyList[ix].serverBuf )
        ProxyList[ix].serverBuf = (buf_list_t *)calloc(1,sizeof(buf_list_t));

again:
    /* find the end of the list */
    for( bP = ProxyList[ix].serverBuf; bP->next; bP = bP->next )
        ;
    
    /* allocate a new chunk if this one is full */
    if( MSG_SIZE == bP->read ){
        bP->next = (buf_list_t *)calloc(1,sizeof(buf_list_t));
        bP = bP->next;
    }

    /* copy MIN of size vs. whats left in current buf */
    xferCnt = MIN(size,MSG_SIZE-bP->read);
    memcpy(bP->msg+bP->read,buf,xferCnt);
    bP->read += xferCnt;
    buf += xferCnt;

    /* if didn't copy whole message, get a new buffer */
    if ( xferCnt < size )
        goto again;
}

/*************************************************************************
function: AxProxyRead - a read channel is active, pick up any waiting data
*/
static int
AxProxyRead(int channel, buf_list_t **headP)
{
    int32 cnt;    /* counter */
    buf_list_t *bP;

    /* handle empty list */
    if( !*headP )
        *headP = (buf_list_t *)calloc(1,sizeof(buf_list_t));

    /* find the end of the list */
    for( bP = *headP; bP->next; bP = bP->next )
        ;
    
    /* allocate a new chunk if this one is full */
    if( MSG_SIZE == bP->read ){
        bP->next = (buf_list_t *)calloc(1,sizeof(buf_list_t));
        bP = bP->next;
    }

    while( bP->read < MSG_SIZE ){
#if AX_X11
        cnt = read(channel,bP->msg+bP->read,MSG_SIZE-bP->read );
#else
        cnt = recv(channel,bP->msg+bP->read,MSG_SIZE-bP->read,0);
#endif
        if( cnt == 0 ){
            return(FALSE);
        } else if( cnt == -1 ) {
#if AXMS
            DWORD last_err = WSAGetLastError();
            if ( last_err == WSAEWOULDBLOCK )
#else
            if( errno == EWOULDBLOCK)
#endif
                return(TRUE);
            return(FALSE);
        } else 
            bP->read += cnt;
    }

    return(TRUE);
}

/*************************************************************************
function: AxProxyMain - check all mux channels, if active do read or write
message written to serverCh is read out of serverBuf
message written to clientCh is read out of clientBuf
message read from serverCh gets stored in clientBuf
message read from clientCh gets stored in serverBuf
*/
void AxProxyMain(rfds,wfds)
fd_set *rfds;
fd_set *wfds;
{
    register int ix;

    for( ix=0; ix<ProxyListLast; ix++ )
        if( ProxyList[ix].inUse ){

            if( FD_ISSET(ProxyList[ix].serverCh,rfds))
                if( !AxProxyRead( ProxyList[ix].serverCh, &(ProxyList[ix].clientBuf))){
                    AxProxyClear(ix);
                    continue;
                }

            if( FD_ISSET(ProxyList[ix].clientCh,rfds))
                if( !AxProxyRead( ProxyList[ix].clientCh, &(ProxyList[ix].serverBuf))){
                    AxProxyClear(ix);
                    continue;
                }

            if( FD_ISSET(ProxyList[ix].serverCh,wfds))
                if( !AxProxyWrite( ProxyList[ix].serverCh, &(ProxyList[ix].serverBuf))){
                    AxProxyClear(ix);
                    continue;
                }

            if( FD_ISSET(ProxyList[ix].clientCh,wfds))
                if( !AxProxyWrite( ProxyList[ix].clientCh, &(ProxyList[ix].clientBuf))){
                    AxProxyClear(ix);
                    continue;
                }
        }

}
  

/*************************************************************************
function: AxOpenServerAndSignalParent - Create & Bind to sock, let parent process proceed
*/
int AxOpenServerAndSignalParent (sockname)
char *sockname;
{
    int sockfd, altsock;

    sockfd = altsock = 0;

    AxOpenDualServer(sockname,NULL,FALSE,&sockfd,&altsock,TRUE);
    return(sockfd);
}

/*************************************************************************
function: AxOpenServer - Create & Bind to socket
*/
int AxOpenServer (sockname,forkIt)
char *sockname;
int forkIt;
{
    int sockfd, altsock;

    sockfd = altsock = 0;

    AxOpenDualServer(sockname,NULL,forkIt,&sockfd,&altsock,FALSE);
    return(sockfd);
}

/*************************************************************************
function: AxOpenServerWithHandle - Create & Bind to socket, where a specific
     socket handle may be passed in.  This will only be true is a Window/NT
     environment.  In UNIX the input socK_handle must always be zero.
*/
int AxOpenServerWithHandle(sockname, sock_handle, forkIt)
char *sockname;
int sock_handle;
int forkIt;
{
    int altsock;

    altsock = 0;

    AxOpenDualServer(sockname,NULL,forkIt,&sock_handle,&altsock,FALSE);
    return( sock_handle );
}

/*************************************************************************
function: AxOpenDualServer - Opens sockets on both named pipe & tcp
*
*   *sockfdP must be initialized to zero to create a new socket
*/
void AxOpenDualServer (sockname,altname,forkIt,sockfdP,altsockfdP,signalParent)
char *sockname;
char *altname;
int forkIt;
int *sockfdP,*altsockfdP;
int signalParent;
{
    SOCKET sockfd, altsockfd;
    int sockno, mask;
    struct sockaddr_in in_serv_addr;
#if !AXMS 
    struct sockaddr_un un_serv_addr;
#endif
    int isUnix = 0;
    int trueVal = TRUE;
    int code;
    int ret;
    
    /* In the NT environment we will be passed an inherited, duplicated */
    /* socket handle in "*sockfdP" so we do not need to create a socket. */
    /* In the UNIX  case, we create the socket here                      */
    sockfd = *sockfdP;
    if ( *sockfdP == 0 )
       {
       /* create and bind to socket */
       sockno = atoi(sockname);
       if ( sockname[0] == '/' ) {
           isUnix = 1;
#if !AXMS 
           unlink(sockname);
           if ( (sockfd = socket(AF_UNIX, SOCK_STREAM, 0)) == INVALID_SOCKET )
               AxErrnoError(NXLT("socket"));
#endif
       }
       else {
           if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET )
               {
#if AXMS
               errno = WSAGetLastError();
#endif
               AxErrnoError(NXLT("socket"));
               }
       }
       *sockfdP = sockfd;

       if ( isUnix ) {
#if AXMS 
           BadError( NXLT("UNIX sockets not supported on this platform") );
#else
           bzero(&un_serv_addr, sizeof(un_serv_addr));
           un_serv_addr.sun_family = AF_UNIX;
           strcpy(un_serv_addr.sun_path,sockname);
           if ( bind( sockfd, (struct sockaddr*)&un_serv_addr,
                                    sizeof(un_serv_addr)) == SOCKET_ERROR )
            AxErrnoError(NXLT("bind"));
#if !SGI
           chmod(sockname, 0777);
#endif
#endif
       }
       else {
           bzero(&in_serv_addr, sizeof(in_serv_addr));
           in_serv_addr.sin_family = AF_INET;
           in_serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
           in_serv_addr.sin_port = htons(sockno);
           in_serv_addr.sin_port = htons(sockno);
           setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR,
                                     (char *)(&trueVal), sizeof(trueVal));
           /* Allow the bind to this socket, since "winshell" parent may */
           /* still appear to be using it.                                */
#if AXMS
           if ( bind( sockfd, (struct sockaddr*)&in_serv_addr,
                                      sizeof(in_serv_addr)) == SOCKET_ERROR )
               {
               errno = WSAGetLastError();
               AxErrnoError(NXLT("bind"));
               }
#else
         if ( bind( sockfd, (struct sockaddr*)&in_serv_addr,
                                    sizeof(in_serv_addr)) == SOCKET_ERROR )
             AxErrnoError(NXLT("bind"));
#endif

         }
     }    /* end of *sockfdP == 0 */

#if 0
     /* check out the handle (for debugging purposes */
     code = sizeof(in_serv_addr);
     ret = getsockname( sockfd, (struct sockaddr*)&in_serv_addr, &code );
     if ( ret == SOCKET_ERROR )
        {
        errno = WSAGetLastError();
        AxErrnoError( NXLT("getsockname") );
        }
     sockno = ntohs(in_serv_addr.sin_port);
#endif

    /*
    Set REUSEADDR so that we can recycle the server even if there are living
    clients.  This is VERY helpful for debugging servers... bypassing huge
    timeouts.
    */
    if (!isUnix) {
        trueVal = TRUE;
        ret = setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR,
                         (char *)(&trueVal), sizeof(trueVal));
    }
#if AXMS
    if ( ret == SOCKET_ERROR )
        {
        errno = WSAGetLastError();
        AxErrnoError(NXLT("setsockopt error:"));
        }
#endif

    listen(sockfd,50);

#if !AXMS 
    if ( altname ) {
        unlink(altname);
        if ( (altsockfd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0 )
            AxErrnoError(altname);
        *altsockfdP = altsockfd;
         trueVal = TRUE;
        setsockopt( altsockfd, SOL_SOCKET, SO_REUSEADDR,
                                     (char *)(&trueVal), sizeof(trueVal));
        bzero(&un_serv_addr, sizeof(un_serv_addr));
        un_serv_addr.sun_family = AF_UNIX;
        strcpy(un_serv_addr.sun_path,altname);
        if ( bind(altsockfd,(struct sockaddr*)&un_serv_addr,sizeof(un_serv_addr)) < 0 )
            AxErrnoError(altname);
#if !SGI
        chmod(altname, 0777);
#endif
        listen(altsockfd,50);
    }
#endif

#if AX_X11  
    if ( forkIt || signalParent )
        ForkOurselvesOrSignalParent(signalParent);
#endif


}

/*************************************************************************
function: AxChoosePort - Find available port
This function requests a port, opens it and immediately closes it.
It is not reliable if another server does the same thing!!!  Caveat Emptor
*/
int AxChoosePort ()
{
#if AX_X11 && !LINUX                          /*  {  */
#if SUN4S || SUNIS || SGI
#define DEV_TCP "/dev/tcp"
#else
#if HP800V3
#define DEV_TCP "/dev/ositpi"
#else
#if ALPHA
#define DEV_TCP "/dev/streams/xtiso/tcp"
#else
#if RIOS
#define DEV_TCP "/dev/xti/tcp"
#else
#error You need to find where the "tcp" file is ; should be one of the above!;
#endif  /* RIOS */
#endif  /* ALPHA */
#endif  /* HP800V3 */
#endif  /* sun4s || SUNIS */

	struct t_bind reqrsp;
	struct sockaddr_in in_serv_addr, *sP;
	int sockfd, port;

	if ( (sockfd = t_open(DEV_TCP,O_RDWR,NULL)) == INVALID_SOCKET )
		return(-1);
	bzero(&in_serv_addr, sizeof(in_serv_addr));
	in_serv_addr.sin_family = AF_INET;
	in_serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	in_serv_addr.sin_port = 0;
	reqrsp.addr.maxlen = sizeof(in_serv_addr);
	reqrsp.addr.len = sizeof(in_serv_addr);
	reqrsp.addr.buf = (char*)&in_serv_addr;
	reqrsp.qlen = 50;
	if ( t_bind(sockfd,NULL,&reqrsp) == SOCKET_ERROR )
		return(-1);
	sP = (struct sockaddr_in*)reqrsp.addr.buf;
	port = sP->sin_port;
	t_close(sockfd);
	return(port);
#else                                 /* }  {  */

    SOCKET sockH;
    struct sockaddr_in in_serv_addr;
    int ret;
    int port;
    int trueVal = TRUE;
    int i;

    /* create the socket handle */
#if LINUX
    sockH = socket( AF_INET, SOCK_STREAM, 0 );
	if ( sockH == -1 ) {
		return( -1 );
	}
#else
    sockH = socket( PF_INET, SOCK_STREAM, 0 );
    if ( sockH == INVALID_SOCKET ) {
		ret = WSAGetLastError();   /* for debugging */
		return( -1 );
	}
#endif

    /* Allow the quick re-use of this socket */
    setsockopt( sockH, SOL_SOCKET, SO_REUSEADDR,
                                     (char *)(&trueVal), sizeof(trueVal));

    /* bind to it to get a port number */
    memset( &in_serv_addr, 0, sizeof(in_serv_addr) );
    in_serv_addr.sin_family = AF_INET;
    in_serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    in_serv_addr.sin_port = 0;
    ret = bind( sockH, (struct sockaddr*)&in_serv_addr,
                                                    sizeof(in_serv_addr));

#if LINUX
    if ( ret == -1 )
       {
       return( -1 );
       }
#else
    if ( ret == SOCKET_ERROR )
       {
       ret = WSAGetLastError();   /* for debugging */
       return( -1 );
       }
#endif

    i = sizeof(in_serv_addr);    
    ret = getsockname( sockH, (struct sockaddr*)&in_serv_addr, &i );
#if LINUX
    if ( ret == -1 )
       {
       return( -1 );
       }
#else
    if ( ret == SOCKET_ERROR )
       {
       ret = WSAGetLastError();   /* for debugging */
       return( -1 );
       }
#endif

    port = (int)ntohs(in_serv_addr.sin_port);
    
    CLOSE_CHANNEL( sockH );

    return( port );

#endif                                /*  }  */
}


/*************************************************************************
function: AxOpenBasicServer - Open port, return fd
*/
int AxOpenBasicServer (int sockno)
{
	return(AxTryOpenBasicServer(sockno, TRUE));
}

/*************************************************************************
function: AxTryOpenBasicServer 
*/
int AxTryOpenBasicServer (int sockno, int reuse_addr)
{
	SOCKET sockfd;
	int mask;
	struct sockaddr_in in_serv_addr;
	int trueVal = TRUE;
	int code;
	int ret;
	
	if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET ) {
#if AXMS
		errno = WSAGetLastError();
#endif
		return(-1);
	}
	bzero(&in_serv_addr, sizeof(in_serv_addr));
	in_serv_addr.sin_family = AF_INET;
	in_serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	in_serv_addr.sin_port = htons(sockno);
#if AXMS
	if ( reuse_addr ) {
		/* Allow the bind to this socket, since "winshell" parent may */
		/* still appear to be using it.								*/
		setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR,
							 (char *)(&trueVal), sizeof(trueVal));
	}
	if ( bind( sockfd, (struct sockaddr*)&in_serv_addr,
							  sizeof(in_serv_addr)) == SOCKET_ERROR ) {
		errno = WSAGetLastError();
		CLOSE_CHANNEL(sockfd);
		return(-1);
	}
#else
	if ( bind( sockfd, (struct sockaddr*)&in_serv_addr,
					sizeof(in_serv_addr)) == SOCKET_ERROR ) {
		CLOSE_CHANNEL(sockfd);
		return(-1);
	}
#endif
	/*
	Set REUSEADDR so that we can recycle the server even if there are living
	clients.  This is VERY helpful for debugging servers... bypassing huge
	timeouts.  We ignore fails opn this call.
	*/
	trueVal = TRUE;
	setsockopt( sockfd, SOL_SOCKET, SO_REUSEADDR,
				 (char *)(&trueVal), sizeof(trueVal));

	listen(sockfd,50);
	return(sockfd);
}


static SOCKET SvrSockChannel = INVALID_SOCKET;
int AxSvrSockFd;
int AxSvrAltSockFd;

#if ELF_COMM_STAND_ALONE

extern void (*AxBrokenPipeCB)(/* int fd */);

/***********************************************************************
function: AxSetBrokenPipeCB - Set the broken pipe handler.  This
handler is called with one arg, the 'fd' on which EPIPE occured.
If the handler returns, normal processing is resumed.  This handler
will be called from AxMessageTop() on EPIPE write error.

If an RPC server needs to set this handler, it must do so in the 0
case of it's AxDispatch() switch table.
*/
void
AxSetBrokenPipeCB(cbfun)
void (*cbfun)();
{
    AxBrokenPipeCB = cbfun;
    return;
}

/***********************************************************************
function: AxDfltBrokenPipeCB - This is the default EPIPE handler when
the write fails in AxMessageTop().  shutdown 'fd' and exit with code 1.
*/
static void
AxDfltBrokenPipeCB(fd)
int fd;
{
    shutdown(fd, 2);
    CLOSE_CHANNEL(fd);
    exit(1);
}

/*************************************************************************
function: AxServerMainLoop - Wqait for a connection, dispatch messages
*/
void AxServerMainLoop(sockfd,isUnix,keepOpen)
SOCKET sockfd;
int isUnix, keepOpen;
{
    AxSetBrokenPipeCB(AxDfltBrokenPipeCB);
    AxServerDualMainLoop(sockfd,0,isUnix,keepOpen);
}

void AxServerDualMainLoop(mainSockFd,altSockFd,isUnix,keepOpen)
SOCKET mainSockFd, altSockFd;
int isUnix, keepOpen;
{
    int clilen;
    struct sockaddr_in in_cli_addr;
#if !AXMS 
    struct sockaddr_un un_cli_addr;
#endif
    int code, sockfd, events;
    static fd_set read_fds, wtfds;

    AxSvrSockFd = mainSockFd;   /* arg is always the main channel */
    AxSvrAltSockFd = altSockFd;
    if( !ProxyListSize )
        AxProxyRealloc();

    while ( 1 ) {

        /* If we are running dual channel mode, select for first to wake up */

    again:
        if ( Proxying || (AxSvrSockFd && AxSvrAltSockFd) ) {
            FD_ZERO(&read_fds);
            FD_ZERO(&wtfds);
            FD_SET(AxSvrSockFd, &read_fds);
            /* For aagtwy, AxSvrAltSockFd will always be zero, which is an
             * invalid socket on NT and causes the subsequent select to fail.
             */
            if ( AxSvrAltSockFd )
                FD_SET(AxSvrAltSockFd, &read_fds);
            AxProxySet(&read_fds,&wtfds); /* add mux fds to the select list */
            events = select(FD_SETSIZE, &read_fds, &wtfds, NULL, NULL);
            if ( events <= 0 ) {
                if ( errno != EBADF )
                    goto again;
                /* if a mux fd went dead, that's normal.  Remove it from
                   the select and keep going. */
                if( AxProxyFindBad() )
                    goto again;
                /* the server sock must have been bad.  bail. */
                fprintf(stderr,"rpc_svr.c - select failed, errno %d\n",errno);
                exit(1);
            }
            if ( FD_ISSET(AxSvrAltSockFd,&read_fds) )
                sockfd = AxSvrAltSockFd;
            else if ( FD_ISSET(AxSvrSockFd,&read_fds) )
                sockfd = AxSvrSockFd;
            else {
                /* attend to any active mux channels */
                AxProxyMain(&read_fds,&wtfds);
                goto again;
            }
        }
        else {
            sockfd = mainSockFd;
            /* add mux fds to the select list */
            AxProxySet(&read_fds,&wtfds);
        }

        if ( isUnix ) {
#if AXMS
            errno = -1;
            fprintf(stderr,NXLT("UNIX sockets not supported on this platform\n") );
            exit(1);
#else
            clilen = sizeof(un_cli_addr);
            SvrSockChannel = accept(sockfd, (struct sockaddr *)(&un_cli_addr),
                                    &clilen);
#endif
        }
        else {
            clilen = sizeof(in_cli_addr);
            SvrSockChannel = accept(sockfd, (struct sockaddr *)(&in_cli_addr),
                                    &clilen);
        }
        if ( SvrSockChannel == INVALID_SOCKET ) {
            if ( errno == EINTR )
                goto again;
            fprintf(stderr,"rpc_svr.c - accept failed, sockfd=%d, errno %d\n",
                    sockfd, errno);
            if ( errno != EBADF )
                goto again;
            exit(1);        /* something bad thing has happend */
        }

#if 0	/* this is limiting client connections too, so disabled */
        if (!isUnix) {
            if (0 != AxCheckAccessByIP(inet_ntoa(in_cli_addr.sin_addr))) {
                code = 2;
                goto close_sock;
            }
        }
#endif
        do {
            code = AxSockTop(SvrSockChannel);
            /* use break in the action to take care of proxy channels */
            AxProxySelect();
        }
        while ( code == 0 );

        if ( code == 3 )
            continue;

      close_sock:
        if ( SvrSockChannel != INVALID_SOCKET ) {   /* -1 only is AxCloseChannel called by forker */
            shutdown(SvrSockChannel,2);
            CLOSE_CHANNEL(SvrSockChannel);
            if ( code == -1 ) {
                exit(1);
            }
            else if ( (code == 1) || ((code == 2) && !keepOpen) ) {
                if ( AxSvrSockFd ) {
                    shutdown(AxSvrSockFd,2);
                    CLOSE_CHANNEL(AxSvrSockFd);
                }
                if ( AxSvrAltSockFd ) {
                    shutdown(AxSvrAltSockFd,2);
                    CLOSE_CHANNEL(AxSvrAltSockFd);
                }
                return;
            }
        }
    }
}
#endif

/*************************************************************************
function: AxChannel - Return current SvrSockChannel number
This function is ONLY useful if the server is runnign against a Berkely socket.
Not useful if the server if talking with the client thrue a private pipe.
*/
int AxChannel ()
{
    return(SvrSockChannel);
}

/*************************************************************************
function: AxCloseChannel - Used ONLY by internal forking logic
This function is ONLY useful if the server is runnign against a Berkely socket.
Not useful if the server if talking with the client thrue a private pipe.
*/
void AxCloseChannel ()
{
    if ( SvrSockChannel == INVALID_SOCKET )
        return;
/*  shutdown(SvrSockChannel,2); */
    CLOSE_CHANNEL(SvrSockChannel);
    SvrSockChannel = INVALID_SOCKET;
}

void AxInvalidateChannel()
{
/* just set SvrSockChannel to an invalid value,
 * but don't actually close it. NT needs this when axnet
 * starts service threads.
 */
    SvrSockChannel = INVALID_SOCKET;
}

int AxIsSvrChannel(sockfd)
int sockfd;
{
    if ( sockfd == SvrSockChannel )
        return TRUE;
    else
        return FALSE;
}

/*************************************************************************
function: AxNetRegisterService - Resgiert with axnet, return socket name
*/
void AxNetRegisterService (sockname,name,isPrivate,passwd,code,str,obj)
char *sockname;
char *name;
int isPrivate;
char *passwd;
int *code;
char **str, **obj;
{
    int channel;
    elfData request;

    channel = AxNetConnect(NULL,code,str,obj);
    if ( *code != 0 ) return;
    request = AxBuildArray(NXLT("ssis"),name,sockname,isPrivate,passwd);
    AxSendMessageToServer(channel,AXNET_REGISTER,request,code,str,obj);
    if ( *code != 0 ) return;
    AxGetResponseFromServer(channel,code,str,obj);
    if ( *code != 0 ) return;
    AxClientDisConnect(channel);
}

/*************************************************************************
function: AxNetUnRegisterService - Drop service
*/
void AxNetUnRegisterService (name,code,str,obj)
char *name;
int *code;
char **str, **obj;
{
    int channel;
    elfData request;

    channel = AxNetConnect(NULL,code,str,obj);
    if ( *code != 0 ) return;
    request = AxMakeStrData(-1,name);
    AxSendMessageToServer(channel,AXNET_UNREGISTER,request,code,str,obj);
    if ( *code != 0 ) return;
    AxGetResponseFromServer(channel,code,str,obj);
    if ( *code != 0 ) return;
    AxClientDisConnect(channel);
}



void BadError( char *msg )
{

#if ELF_COMM_STAND_ALONE
    errno = -1;
    AxErrnoError( msg );
#else
    tempAbort( msg );
#endif

    return;
}



#if AX_X11  /*  {  */

ForkOurselves() {   /* Obsolete? do RT gateways really ever ForkIt when remote */
    ForkOurselvesOrSignalParent(FALSE);
}

/***********************************************************************/
/* fork to free up the parent, and to take on the identity of the user */
/* This is a common routine which can be called from a couple of levels */
/***********************************************************************/
static ForkOurselvesOrSignalParent(signalParent)
int signalParent;
{
    int code;
    int tty_fd = -1;

    /* Ignore stop signals from tty */
#ifdef SIGTTOU
    signal(SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
    signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
    signal(SIGTSTP, SIG_IGN);
#endif

    /* fork server as child, return parent started ok */
    if ( signalParent ) {
        kill(getppid(),SIGUSR1);
    }
    else {
        if ( fork() != 0 )
            exit(0);            /* parent exits */
    }

    /* child continues */
    /* Disassociate from controlling tty and process group */
#if defined (TIOCNOTTY) && !RIOS /* bsd */
    (void) setpgrp(0, getpid());
    signal(SIGHUP, SIG_IGN);
    tty_fd = open(NXLT("/dev/tty"), O_RDWR);
    if (tty_fd >= 0) {
        (void) ioctl(tty_fd, TIOCNOTTY, (char *) NULL);
        (void) close(tty_fd);
    }
#else /* sys-v nonsense */
    (void) setpgrp();
    signal(SIGHUP, SIG_IGN);
#endif


#if AX_X11
    if ( !access( "/tmp/ForkOurselves", 0 ) )
       {
       /* pause till we get a hold with dbx */
       code = 99;
       while ( code >= 99 )
         {
         code++;
         sleep(2);
         }
       }
#endif
}
#endif   /* }  */
