/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tempfile.c
   Applix ASC version: 14
   Date: Mon Nov 16 13:53:26 1998
*/
#include <aos.h>
#include <thimble.h>

#if AXMS
#include <winmap.h>
#endif

char ThimTempDir[STRPATH+1];

char *AxTempFile AXP_((char *proto_name,char *resultName));
char *AxCreateTempfile AXP_((char *proto_name));
char *Amktemp();

typedef struct tempfile_t {	/* Tempfiel info struct */
	struct tempfile_t *next;	/* next in list */
	char *name;			/* pathname of file */
	int task;			/* owning task */
} tempfile_t;

static tempfile_t *AxTempFiles;	/* List of existant tempfiles */
static char tempfile_name[STRPATH+1];

/*************************************************************************
function: AxExpandPrototypePath - must contain from 5 to 14 %'s, inclusively.
*/
char *
AxExpandPrototypePath ( proto_name )
char *proto_name;
{
	char *inpath;

	if ( proto_name == (char *)0 )
		proto_name = NXLT("%%%%%%%%");

	/* if no absolute path is given, put it in the temp dir */
	inpath = malloc(STRPATH+1); /* caller must free!!! */
	if ( (*proto_name != '/') &&
	     (*proto_name != '\\') &&
	     (*(proto_name+1) != ':') )
		sprintf(inpath, "%s/%s", ThimTempDir, proto_name);
	else
		strcpy(inpath, proto_name);
	AxSysDosUnixPath(inpath);
	return(inpath);
}

/*************************************************************************
function: AxCreateTasklessTempfile - create a tempfile
*/
static char *
AxCreateTasklessTempfile ( proto_name )
char *proto_name; /* NULL or proto name */
{
	char *name, *p;
	int ret;

	/* produce tempname into static data - this MUST NOT give up the thread */
	name = tempfile_name;
	strcpy(name, proto_name);
	p = Amktemp(name);

	/* Create file */
	if ( p == (char *)0 )
		name = (char *)0;
	else
	{
#if AXMS
		HANDLE file_handle;

		file_handle = MS_CreateFile(p, (GENERIC_READ | GENERIC_WRITE), 
		                               (FILE_SHARE_WRITE | FILE_SHARE_READ),
		                               (LPSECURITY_ATTRIBUTES)NULL,
		                               CREATE_ALWAYS, 
		                               FILE_ATTRIBUTE_TEMPORARY,
		                               (HANDLE)NULL );
		if ( file_handle == (HANDLE)NULL )
			name = (char *)0;
		else
		{
			ret = MS_CloseHandle(file_handle);
			if ( ret == 0 )
				ret = MS_GetLastError();
		}
#else
		int fd;

	    fd = creat(p, 0600); /* rw-self, others none */
		if ( -1 == fd )
			name = (char *)0;
		else
			ret = close(fd);
#endif
	}
	return(name);
}

/*************************************************************************
function: AxDeleteTasklessTempfile - Delete a tempfile
*/
int
AxDeleteTasklessTempfile ( name )
char *name;
{
	int ret;
#if AXMS
	ret = DeleteFile(name);
	if ( ret == 0 )
		ret = GetLastError();
#else
	ret = unlink(name);
#endif
	return(ret);
}

/*************************************************************************
function: AxCreateTempfile - Create a tempfile, add to tempfiles list.
This returns the name of the file created.
On failure, this returns NULL and errno is set.
*/
char *
AxCreateTempfile (proto_name)
char *proto_name;	/* /asd/asdasd/zz%%%%%%.sfx */
{
	char *name;
	int fd, pid;
	tempfile_t *tp;

	/* create tempname (static data - do not free name!) */
	name = AxCreateTasklessTempfile(proto_name);
	if ( name ) /* Add to files and task's list */
	{
		pid = ThimTopPidFromPid(THIMpid);
		TaskInfo[pid].created_tempfiles = TRUE;
		tp = (tempfile_t *)malloc(sizeof(tempfile_t));
		tp->name = strcpy(malloc(strlen(name)+1),name);
		tp->task = pid;
		tp->next = AxTempFiles;
		AxTempFiles = tp;
	}
	return(name);
}

/*************************************************************************
function: AxTempFile - Create tempfile, add to tempfiles list
This creates the file in the standard temp directory.
If the passed proto is NULL, the software chooses a name.
This returns the name of the file created.
If the passed result buf is not NULL, the software also writes back the name.
On failure, this returns NULL and errno is set.
*/
char *
AxTempFile ( proto_name, resultName )
char *proto_name; /* NULL or proto name */
char *resultName; /* NULL or buf to write filename to */
{
	char *p, *inpath;

	inpath = AxExpandPrototypePath(proto_name);
	p = AxCreateTempfile(inpath);
	if ( p && resultName )
		strcpy(resultName, p);
	free(inpath);
	return(p); /* static data - do not free p! */
}

/*************************************************************************
function: AxTasklessTempFile - Create tempfile, add to tempfiles list
This creates the file in the standard temp directory.
If the passed proto is NULL, the software chooses a name.
This returns the name of the file created.
If the passed result buf is not NULL, the software also writes back the name.
On failure, this returns NULL and errno is set.
*/
char *
AxTasklessTempFile ( proto_name, resultName )
char *proto_name; /* NULL or proto name */
char *resultName; /* NULL or buf to write filename to */
{
	char *p, *inpath;

	inpath = AxExpandPrototypePath(proto_name);
	p = AxCreateTasklessTempfile(inpath);
	if ( p && resultName )
		strcpy(resultName, p);
	free(inpath);
	return(p);
}

/*************************************************************************
function: AxDeleteAllTemps - Called by shutdown
*/
void
AxDeleteAllTemps ()
{
	tempfile_t *p;
    int count;

	for ( p = AxTempFiles; p; p = p->next )
	{
#if AXMS
		count = 0;
		while ( AxDeleteTasklessTempfile(p->name) == ERROR_SHARING_VIOLATION )
		{
			if ( count >= 5 ) /* try not to go over 1 sec / file */
				break;
			Sleep(250); /* give up process timeslice for 1/4 second */
			count++;
		}
#else
		AxDeleteTasklessTempfile(p->name);
#endif
	}
}

/*************************************************************************
function: AxDeleteTempfile - Delete a tempfile & remove from tempfile's list
*/
void
AxDeleteTempfile (name)
char *name;
{
	tempfile_t *p, *next, *parent;
    int ret;

	AFDoneWithFile(name);
	ret = AxDeleteTasklessTempfile(name);
	parent = (tempfile_t *)0;;
	for ( p = AxTempFiles; p; p = next )
	{
		if ( streq(name, p->name) )
		{
#if AXMS
			/* the delete failed because the file is still open by     */
			/* some child.  Set the pid to zero, so that this will not */
			/* be seen by other task action, but will get deleted when */
			/* applixware exits.                                       */
			if ( ret == ERROR_SHARING_VIOLATION )
			{
				p->task = 0;
				break;
			}
#endif
			if ( parent == (tempfile_t *)0 )
				AxTempFiles = p->next;
			else
				parent->next = p->next;
			free(p->name);
			free(p);
			break;
		}
		next = p->next;
		parent = p;
	}
}

/*************************************************************************
function: AxDeleteTemps - Delete task's tempfiles, call by THIMexit
*/
void
AxDeleteTemps (task)
int task;
{
	tempfile_t *p, *next;

	for ( p = AxTempFiles; p; p = next )
	{
		next = p->next;
		if ( p->task == task )
			AxDeleteTempfile(p->name);
	}
}

/*************************************************************************
function: AxDumpTempFiles - Dump names of all temp files to stdout (for dbx)
*/
void
AxDumpTempFiles ()
{
	tempfile_t *p;

	printf("Here are the temp files currently in use:\n");
	for ( p = AxTempFiles; p; p = p->next )
		printf("  Task %d: '%s'\n", p->task, p->name);
}
