/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tmzmallo.c
   Applix ASC version: 124
   Date: Wed Mar 17 21:12:49 1999
*/
#if BUILD_PURIFY
#ifdef THIM_STATIC
#undef THIM_STATIC
#endif
#define THIM_STATIC 1       /* do not re-map "free" */
#endif

#if (LINUX && __alpha__)
#include <unistd.h>
#endif

#include <stdio.h>
#include <ctype.h>
#if AXMS
#include <stdlib.h>
#include <windows.h>
#include <process.h>    /* for getpid */
#define MS_GetEnvironmentVariable GetEnvironmentVariable
#else
#include <sys/param.h>
#endif

#include <aos.h>
#include <amemtrak.h>
#include <thimble.h>
#include <eldata.h>

#if BUILD_PURIFY
/* undo the effects of thim_gen.h */
#undef malloc
#undef calloc
#undef realloc
#undef free
#endif

char *TaskAlloc();
char *TaskCalloc();
char *TaskRealloc();
void TaskFree();

char *GetProfileByStr();
static void AddMemoryToFreeList();
static void AddChunkToList();
int PrintMemoryStatsFlag;   /* set to cause memory statistics to be dumped */

 /* when FreeUpMemory() is called               */

int ThimMemClearOnFree = -1;    /* used for debug */
unsigned char ThimMemClearChar = 0xEE;  /* char to memset area */


 /* THimMemCheck() on every memory action */
int CheckMemory = FALSE;    /* set this to true to enable calling */
#define CHECK_MEMORY if ( CheckMemory )  ThimMemCheck();

/* Trap the alloc or free of a given memory address */
static void MemTrapTest();
void ThimMemTrap();
char *MemoryTrap = NULL;        /* Set to addr in question in debugger */
#define TRAP_ADDR(p)    if (MemoryTrap) MemTrapTest(p);

static char *msg =
NXLT("Sorry, ran out of virtual memory (swap space), recovery failed.\n");


int *AxMemRet;          /* for memory traker */
int AxSysOnlyMemTrace = FALSE;  /* false also reports unfreed task memory */
int AxOutOfMemory;      /* TRUE while out of memory box is up */
int AxAllowStop;        /* TRUE is meory box up and stop is allowed */

#if DBX_MODE
static FILE *AlcLogFp;
static int AlcLogging = FALSE;

#define LOGGER(body) if ( AlcLogging ) AlcLogger body
#else
#define LOGGER(body)
#endif




/* ALWAYS keep this structure size a multiple of 8 */
typedef struct mem_header
{
    short pid;      /* task that owns this data */
    short free_dex;     /* free list index number + dex */
    struct mem_header *next;/* next of this app type or NULL */
    struct mem_header *prev;/* prev of this app type or NULL */
#if ALPHA || (LINUX && __alpha__) || SGI
    int dummy;
#if MEM_CALL_TRACE
    int *ret_addr;
#endif
#else
    int *ret_addr;      /* return address to caller */
#endif
}   mem_header;

/** The following macros are used to stuff and retrieve the free-list-index */
/** and the "dex" from the single short "free_dex" in the mem_header struct */
/** The "dex" is 5 bits, and the free-list-index is 11, allowing for 2048   */
/** chunks to be allocated.  Note: the subtraction of the two pointers      */
/** below yields an index, not a byte difference.  'C' defines the result   */
/** of pointer subtraction to be in terms of the pointer units, not raw bytes */
#define GET_DEX( xx )  ( xx & 0x1F )    /* dex is in lower 5 bits */
#define GET_INDEX( xx ) ( xx >> 5 ) /* chunk index is in upper 11 bits */
#define MAKE_FREE_DEX( listP, dex )  ( ((listP - ChunkListPP) << 5) | dex )

/** make sure that this structure is a multiple of 8 bytes */
typedef struct _chunk_descr
{
    int chunk_size;
    int reserved;
    mem_header *FreeList[32];
}   chunk_descr;

/** The first 4 entries in the "FreeList" are never used, since all */
/** blocks are at least 16 bytes big. So we stuff a few things we do need   */
/*  into the start of the array  as follows:                                */
#define BASE_CHUNK_LOC  0   /* address of the whole chunk */
#define TASK_ZERO_EXT   1   /* chunk claimed by task zero as extension */
                            /* counts blocks allocated out of the chunk */
                            /* for task zero                            */
#define BYTES_ALLOCATED 2   /* how many bytes are in use from this chunk */
#define MAX_FREEDEX     3   /* Dex of highest possible freelist in this chunk */




chunk_descr **ChunkListPP;  /* pointer to array of pointers to chunk descs */
int Topmost_chunk;      /* index to last valid chunk */

#define TASK_ZERO_CHUNKS 10 /* number of chunks pre-allocated for task 0 */
static int TaskZeroChunks;
static char *MinFreeUpAddress;  /* top of "PermAlloc" memory */


#if BUILD_DEBUG
int alloc_count, chunk_search, found_on_freelist;
int split_search_count, split_count, true_splits;
int found_on_freelist, free_count, free_tests;
int task_reallocs, task_realloc_adjusts;
#define register
#endif


static struct
{
    long TotalMemoryUsage;
    int MemPermanent;   /* how much non-freeable memory alloced */
}   tmzmalloStruct;

static int max_request_size = -1;
static mem_header *MemHeads[MAX_PIDS];  /* List of Allocated blocks */
int AxMemUsage[MAX_PIDS];   /* How much memory task uses */


static unsigned Undex[] =     {
    1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768, 65536,
    131072, 262144, 524288, 1048576, 2097152, 4194304, 8388608, 16777216,
    33554432, 67108864, 134217728, 268435456, 536870912, 1073741824,
};

extern int AxMemoryWarning; /* in bytes before warning box */
extern int AxMemoryLimit;   /* in bytes before tearing task down */
extern int AxPostMemoryWarning; /* tells schedular to post memory warning box */

#if BUILD_PURIFY
#define mem_header char
#endif

#define MAX_SIZE_REQ 16777216   /* plenty! */
#define MAX_DEX 32
#define FIRST_REAL_DEX 5    /* smallest dex actually used */
#define NO_SPLIT_DEX 18
#define CHUNK_SIZE 131072   /* the default,  2 to the 17'th */
static int ChunkSize;       /* set at runtime */


/*
*   Prem. 1/24/93
*
*   (a) a table of size 32 is created for the first 32 numbers (0-31)
*       containing the powers of 2 to which each of the numbers 0-31 belong.
*
*   (b) based on the fact that, all powers of 2 that are >= 32 are multiples
*       of 32, we can create a table of 64 elements (2048/32) that will
*       contain the powers of 2 for all the numbers in the range 32 to 2048.
*       This is done by dividing the requested size by 32, and using the
*       the result as an index rather than the number itself.
*
*   This brings the total table size to 32+64 = 96 elements.
*
*   Note: Better performance may be achieved by reversing the two tables,
*         i.e, a table of size 64 for first 64 numbers and a table of size
*         32 (2048/64) for numbers in the range 64 to 2048 because we will
*         have a better chance of succeeding in the first comparision in
*         the function memdex.
*
*/

static unsigned int memdex();

#define QUICK_SIZE 2048
#define QUICK_BITS 11       /* 11th bit is set for 2048 */
#define MY_TABLE_SIZE   64
/* use this table for numbers >= SMALL_NUMBER and < QUICK_SIZE */
static unsigned char PowersOf2Table[MY_TABLE_SIZE] =     {
    5, 6, 7, 7, 8, 8, 8, 8,
    9, 9, 9, 9, 9, 9, 9, 9,
    10, 10, 10, 10, 10, 10, 10, 10,
    10, 10, 10, 10, 10, 10, 10, 10,
    11, 11, 11, 11, 11, 11, 11, 11,
    11, 11, 11, 11, 11, 11, 11, 11,
    11, 11, 11, 11, 11, 11, 11, 11,
11, 11, 11, 11, 11, 11, 11, 11};

#define SMALL_NUMBER 32
#define SMALL_SHIFT  5
static unsigned char myNewTable[SMALL_NUMBER] =     {   
    /* use this table for numbers < SMALL_NUMBER */
    0, 1, 1, 2, 2,      /* 0 - 4 */
    3, 3, 3, 3,     /* 5 - 8 */
    4, 4, 4, 4, 4, 4, 4, 4, /* 9 - 16 */
    5, 5, 5, 5, 5, 5, 5, 5, /* 17 -24 */
5, 5, 5, 5, 5, 5, 5};       /* 25 - 31 */

#define MEMDEX(s) ( ((unsigned)(s)) < (unsigned)SMALL_NUMBER ? myNewTable[s] : \
           (((unsigned)(s)) <= (unsigned)QUICK_SIZE ? (PowersOf2Table[((s)-1)>>SMALL_SHIFT]) : \
                                                           memdex(s)) )
#ifndef PAGESIZE
#ifdef  NBPG
#define PAGESIZE NBPG       /* HP/UX, others? */
#else
#define PAGESIZE 4096       /* Reasonable Assumption */
#endif
#endif




/*************************************************************************
function: memdex - Compute free list index based on size
This rounds up to the nearest power of 2
This assums that the size is greater that QUICK_SIZE big.
*/
static unsigned int memdex(size)
register unsigned int size;
{
    register unsigned int i;

    if (size > max_request_size)
        taskAbort(NXLT("Memory block size request exceeds maximum allowed"));

    size--;         /* correctly handle exact powers of 2 */
    size = size >> QUICK_BITS;  /* we are not interested in the least */
    /* significant 11 bits */
    for (i = QUICK_BITS; size; i++)
        {
        size = size >> 1;
        }
    return (i);
}




#if HP800 || M_88K || CLIX || LINUX || SGI         /*  { */
/* the pointer returned by sbrk() is */
/* 'not necessarily word aligned'    */
static char *unaligned_sbrk(n)
int n;
{
    long p;

    if (n > 0)
        {
        n += 8;     /* Wastes 8 bytes for EVERY SBRK! */
        p = (long)sbrk(n);
        if (-1 == p)
            return ((char *)p);
        while (p & 7)
            p++;
        return ((char *)p);
        }
    else
        return ((char *)sbrk(n));
}

#define SBRK unaligned_sbrk
#endif       /*    } HP800 || M_88K || CLIX || LINUX || SGI */


#if RIOS            /*     {  */
static char *get_rs6000_memory(size)
int size;
{
    /* this is an interface to the small shared library stub on RS6000 */
    /* see the long comment in "rs_mallo.c"                           */
    char *p;

    p = rs6000_malloc(size);
    if (p == (char *) NULL)
        return ((char *) -1);   /* make use look like 'sbrk'  */
    return (p);
}


static void free_rs6000_memory(ptr)
char *ptr;
{
#if 0
    rs6000_free(ptr);
#endif
    return;
}

#define SBRK get_rs6000_memory

#endif              /*     } RIOS  */


#ifndef SBRK        /*  {  */
/* not one of the "wierd" sbrk cases above */

#if AXMS         /*  {  */
#define SBRK alloc_NT_memory
void free_NT_memory();
void *alloc_NT_memory();
#else             /* } {  */
#define SBRK sbrk   
#endif            /*  } !AXMS */
#endif            /*  } SBRK not defined  */


/*******************************************************************
 FUNCTION:  ax_getenv  -  a wrapper for getenv.

 REASON for existing: the getenv calls in InitializeMemoryStructure
            fail on NT/95, possibly because it needs to do a malloc
            and can't. The Win32 API call GetEnvironmentVariable,
            however, does work.      -donato
*********/
static char *
ax_getenv(env_var)
char *env_var;
{
#if AX_X11
    return getenv(env_var);
#else
    static char env_str[256];
    if ( MS_GetEnvironmentVariable(env_var, env_str, 256) > 0 )
        return env_str;
    else
        return NULL;
#endif
}                     



/*********************************************************************/
/* Initialize the basic memory block structure.  As a bit of a heuristic  */
/* optimization we allocate (via sbrk since we are starting out) the     */
/* array of pointers to the chunk descriptors, and also the first set of */
/* "NumBaseChunks" chunk descriptors.                                    */
/* By pre-allocating the first set of chunk descriptors we can search    */
/* the free lists of these chunks without having to page a lot of pages     */
/* as might happen if we allocated the chunk descritor as part of the chunk */
/* itself.  The overflow case allocates the chunk descriptor when the new   */
/* chunk is allocated.  This descriptor will be free if the chunk is        */
/* returned to the system later.                                            */
/*********************************************************************/

#define NUM_BASE_CHUNKS 128 /* the default */
static int NumBaseChunks;   /* can be set from the environment */

InitializeMemoryStructure()
{
    int size_preallocate;
    int max_num_chunks;
    chunk_descr *chunk_addr;
    chunk_descr **chunk_descrP;
    int i;
    char *p;
    char *prof_str;

#if !BUILD_PURIFY

    /* sanity check */
    if ( ChunkListPP )
        return;

    TaskInfo[0].deamon = TRUE;      /* anything for task-0 is low memory */

    prof_str = ax_getenv("AX_MAX_ALLOCATE");
    if (prof_str == (char *) NULL)
        max_request_size = MAX_SIZE_REQ;
    else
        max_request_size = atoi(prof_str);

    if (max_request_size < 32768)
        /* lets not be silly */
        max_request_size = MAX_SIZE_REQ;


    if (ThimMemClearOnFree < 0)
        {
        /* define an environment variable to turn on memory debug mode */
        ThimMemClearOnFree = 0;
        if (ax_getenv("AX_MEM_DEBUG") != (char *) NULL)
            ThimMemClearOnFree = 1;
#if 0
        /* additionaly, if we are running at applix, for now, set this */
        /* "clear free'd memory" flag                                   */
        if (AreRunningAtApplix())
            ThimMemClearOnFree = 1;
#endif

        /* See if we are to turn on the complete memory checker  */
        /* Note: this makes things VERY VERY VERY slow !!        */
        if (ax_getenv("AX_MEM_CHECK") != (char *) NULL)
            CheckMemory = TRUE;
        }

    ChunkSize = CHUNK_SIZE; /* set the default */
    p = ax_getenv(NXLT("AX_MEM_CHUNK_SIZE"));
    if (p != (char *) NULL)
        {
        /* get the size and round up to a power of 2 */
        ChunkSize = atoi(p);
        ChunkSize = MEMDEX(ChunkSize);
        ChunkSize = Undex[ChunkSize];
        }

    TaskZeroChunks = TASK_ZERO_CHUNKS;
    p = ax_getenv(NXLT("AX_MEM_ZERO_CHUNKS"));
    if (p != (char *) NULL)
        {
        /* get the size and round up to a power of 2 */
        TaskZeroChunks = atoi(p);
        if (TaskZeroChunks < 1)
            TaskZeroChunks = 1; /* we'll easily need this much! */
        }

    NumBaseChunks = NUM_BASE_CHUNKS;
    p = ax_getenv(NXLT("AX_MEM_NUM_BASE_CHUNKS"));
    if (p != (char *) NULL)
        {
        NumBaseChunks = atoi(p);
        if ((NumBaseChunks < 0) || (NumBaseChunks > 1024))
            /* lets be reasonable */
            NumBaseChunks = NUM_BASE_CHUNKS;
        if (NumBaseChunks < TaskZeroChunks + 1)
            /* need at least the number reserved for task zero */
            NumBaseChunks = TaskZeroChunks + 1;
        }

    /**-------------------------------------------------------------------**/
    /* get space for the list of chunk pointers, and a set of pre-allocated */
    /* chunk descriptors                                                    */
    /**-------------------------------------------------------------------**/
    max_num_chunks = AxMemoryLimit / ChunkSize + 1;
    if (max_num_chunks > 2048)
        {
        /* would make chunk index exceed the 11 bits allocated in the */
        /* mem_header "free_dex" word                                 */
        max_num_chunks = 2048;
        AxMemoryLimit = 2047 * ChunkSize;
        }

    size_preallocate = (max_num_chunks * sizeof(chunk_descr *)) +
        (NumBaseChunks * sizeof(chunk_descr));

    ChunkListPP = (chunk_descr **) SBRK(size_preallocate);
	if (-1 == ((long) ChunkListPP)) {
		fprintf(stderr, "Could not allocate base memory, exiting\n");
		exit(1);
	}

    /**-------------------------------------------------------------------**/
    /* now fill in the first set of pointers to the chunk descriptors */
    /* initialize the chunk list to point to the pre-allocated chunk descrs */
    /**-------------------------------------------------------------------**/
    chunk_descrP = ChunkListPP;
    chunk_addr = (chunk_descr *) (((char *) ChunkListPP) +
                      (max_num_chunks * sizeof(char *)));

    for (i = 0; i < NumBaseChunks; i++)
        {
        *chunk_descrP = chunk_addr;
        chunk_descrP++;
        chunk_addr++;
        }

    /**-------------------------------------------------------------------**/
    /* now pre-allocate the chunks for task 0 */
    /**-------------------------------------------------------------------**/
    Topmost_chunk = -1; /* no chunks allocated so far */
    for (i = 0; i < TaskZeroChunks; i++)
        {
        p = (char *) SBRK(ChunkSize);
        if (-1 == ((long) p))
            {
            /* could not get even task zero stuff, bail out */
            printf(NXLT("Could not allocate base memory, exiting\n"));
            exit(1);
            }

        AddChunkToList(p, ChunkSize);
        }
#endif  /* if !BUILD_PURIFY */
    return;
}







/*************************************************************************
function: AxOwnerOfThisMemory - Return task the allocated memory
*/
int AxOwnerOfThisMemory(void *inP)
{
    register mem_header *p = (mem_header*)inP;
#if BUILD_PURIFY
    int id;

    id = purify_get_pool_id(p);
    return (id);
#else
    return ((p - 1)->pid);
#endif
}




#if LEAK_CHECK               /*  {  */

/* the following small functions record memeory allocatins for task zero */
/* in a sequential array, and zero out the array when the item is freed.  */
/* To look for leaks, run things once, then look to see which elements    */
/* of the array still have a value in them.  These elements were not freed. */
/* To find out who allocated the element at position 49 in the array,       */
/*    set a conditional breakpoint in "rrrr" to look for the 49'th call,    */
/*    re-run the script, and when the breakpoint is hit, look up the stack. */

#define RECORD_FOR_LEAKS rrrr( pid, p );
#define FREE_FOR_LEAKS   ffff( pid, p );
   /* record things to look for leaks  */
#define RRRR_SIZE 1000

void *rrrr_data[RRRR_SIZE];
int rrr_ix = 0;
int do_rrrr = 0;

void rrrr( int pid, void *ptr )
{
    int ix, pp;

    if ( pid != 0 )
       return;
    if ( !do_rrrr )
       return;
    
    if ( rrr_ix >= RRRR_SIZE )
       return;

    for ( ix=0; ix<RRRR_SIZE; ix++ )
        {
        if ( rrrr_data[ix] != (void*)NULL )
           pp++;
        }


    rrrr_data[rrr_ix] = ptr;
    rrr_ix++;
    return;
}

void ffff( int pid, void *ptr )
{
    int ix;
    if ( pid != 0 )
       return;
    if ( !do_rrrr )
       return;
    for ( ix = 0; ix < RRRR_SIZE; ix++ )
        {
        if ( rrrr_data[ix] == ptr )
            {
            rrrr_data[ix] = (void *)NULL;
            return;
            }
        }
    ptr = (void *)NULL;
}
#else                        /* } LEAK_CHECK { */
#define RECORD_FOR_LEAKS
#define FREE_FOR_LEAKS
#endif                       /*  }  LEAK_CHECK  */







/*************************************************************************
function: THIMfree - Release memory in current task
*/
void *THIMfree(p)
void *p;
{
    TaskFree(THIMpid,p);    /* type fixing -- lhunt */
    return ((void *) 1);
}

/*************************************************************************
function: SAFEfree - Realease memory owned by any task
*/
SAFEfree(p)
void *p;
{
    TaskFree(AxOwnerOfThisMemory((mem_header *) p), p);
}

/*************************************************************************
function: AxReallocToAnotherTask - Move ownership of data
*/
void AxReallocToAnotherTask(void *inP, int new_pid)
{
    register mem_header *p = (mem_header*)inP;
    register mem_header *p2;
    register int old_pid;
    int chunk_index;
    chunk_descr *chunkP;            

    CHECK_MEMORY

    if (!p)
        return;

#if BUILD_PURIFY        /*     {  */
    old_pid = purify_get_pool_id(p);
#if 0
    ***purify complains about this
        if (old_pid != new_pid)
        {
        purify_set_pool_id(p, new_pid);
        }
#endif

    return;
#else
#if BUILD_DEBUG
    task_reallocs++;
#endif
    
    /* backup the pointer so it points to the header */
    p--;
    if (p->free_dex == -1)
        tempAbort(NXLT("memory error copying data already free'd"));

    /* change the ownership of the block of memory */
    old_pid = p->pid;
    if (old_pid != new_pid)
        {
        /* remove from old's alloc list */
        if (p->next)
            p->next->prev = p->prev;
        if (p->prev)
            p->prev->next = p->next;
        if (MemHeads[old_pid] == p)
            MemHeads[old_pid] = p->next;
        AxMemUsage[old_pid] = AxMemUsage[old_pid] - Undex[GET_DEX(p->free_dex)];
        
        /* adjust TASK_ZERO_EXTension appropriately if old or new is Task 0 */
        if (TaskInfo[old_pid].deamon ^ TaskInfo[new_pid].deamon)
            {
#if BUILD_DEBUG
            task_realloc_adjusts++;
#endif
            if (TaskInfo[old_pid].deamon)
                {
                chunk_index = GET_INDEX(p->free_dex);
                chunkP = *(ChunkListPP + chunk_index);
                (chunkP->FreeList[TASK_ZERO_EXT])--;
                }
            else
                {
                if (TaskInfo[new_pid].deamon)
                    {
                    chunk_index = GET_INDEX(p->free_dex);
                    chunkP = *(ChunkListPP + chunk_index);
                    (chunkP->FreeList[TASK_ZERO_EXT])++;
                    }
                }
            }
        
        /* add to new task's alloc list */
        p->prev = NULL; /* I am head of pid's memory */
        p->pid = new_pid;
        p2 = MemHeads[new_pid]; /* Get old head */
        if (p->next = p2)   /* I point to it as my next */
            p2->prev = p;   /* It points back to me as prev */
        MemHeads[new_pid] = p;  /* register this new storage */
        AxMemUsage[new_pid] = AxMemUsage[new_pid] + Undex[GET_DEX(p->free_dex)];

        CHECK_MEMORY
        }
#endif              /*     }  BUILD_PURIFY */

}

int merge_count, merge_forward, merge_back;

/*************************************************************************
function: Taskfree - Release some memory
 Warning: do not make changes here which are not also made in exactly
          the same manner in THIMpid_free

*/
void TaskFree(pid, inP)
register int pid;
register void *inP;
{
    register mem_header *p = (mem_header*)inP;
    register unsigned int dex;
    int idex;
    register mem_header *pp, *np;
    mem_header *ppp;
    static int debug_flag = 1;  /* always on to check incorrect free'er */
    int block_pid;
    chunk_descr *chunkP;
    char *chunk_start;
    int chunk_index;
    int undexed_dex;
    int merge_it;
    int max_chunk_dex;
    int use_sys_mem;
    int pdelta;

    CHECK_MEMORY


    if (!p)
        return;
#if BUILD_PURIFY        /*     {  */
    free(p);
#else

    TRAP_ADDR(p);           /* See MemTrapTest */

    p--;            /* point to header */

    FREE_FOR_LEAKS

    if (p->free_dex == -1)
        return;

    dex = GET_DEX(p->free_dex);
    chunk_index = GET_INDEX(p->free_dex);
    chunkP = *(ChunkListPP + chunk_index);

#if BUILD_DEBUG
    free_count++;
    if (dex <= 4)
        pp = NULL;  /* DEBUG - a place to stop */
#endif

    if (pid != p->pid)
        {
        char buff[120];

        if ( (pid < 0) || (pid > MAX_PIDS) || 
             (p->pid < 0) || (p->pid > MAX_PIDS) )
            {
            sprintf(buff, "memory error, incorrect free'er, pid=%d   p->pid=%d",
                  pid, p->pid);
            taskAbort(buff);
            }
        }

    if ((dex < 0) || (dex > MAX_DEX))
        taskAbort(NXLT("memory corrupted, free list index damaged"));

    if (ThimMemClearOnFree)
        memset(p + 1, ThimMemClearChar, Undex[dex] - sizeof(mem_header));

    /************************************/
    /* unthread from the Allocated list */
    /************************************/
    if (p->next)
        p->next->prev = p->prev;
    if (p->prev)
        p->prev->next = p->next;

    /* as a failsafe, use the pid out of the block itself, rather than */
    /* the pid of the caller, since we must get this block off the  */
    /* correct per-task chain, regardless of who is actually freeing the */
    /* block - jon                                                  */
    block_pid = p->pid;
    use_sys_mem = TaskInfo[block_pid].deamon;

    if (MemHeads[block_pid] == p)
        /* block being free'd is the first block on the chain */
        MemHeads[block_pid] = p->next;
    AxMemUsage[block_pid] = AxMemUsage[block_pid] - Undex[dex];

    chunkP->FreeList[BYTES_ALLOCATED] =
        (mem_header *)(((long) chunkP->FreeList[BYTES_ALLOCATED]) - Undex[dex]);

    if (chunkP->FreeList[BYTES_ALLOCATED] == 0)
        {
        /* this chunk is entirely free, quickly put the entire chunk on */
        /* the FreeList[MEMDEX(chunk_size)] */
        memset(&chunkP->FreeList[4], 0, 28 * sizeof(mem_header *));

        if (chunk_index >= NumBaseChunks)
            {
            /* chunk is beyond the range of pre-allocated chunk      */
            /* descriptors so the first part of the chunk contains   */
            /* the descriptor      */
            p = (mem_header *)(((char *) chunkP->FreeList[BASE_CHUNK_LOC]) + sizeof(chunk_descr));
            }
        else
            {
            /* Start of chunk is start of chunksized block */
            p = (mem_header *) chunkP->FreeList[BASE_CHUNK_LOC];
            }
        idex = MEMDEX(chunkP->chunk_size);
        chunkP->FreeList[idex] = p;
        chunkP->FreeList[TASK_ZERO_EXT] = 0;    /* unclaim this chunk */
        p->next = p->prev = NULL;
        p->free_dex = -1;
        }
    else
        {           /* Thread indiviual block */
        /* Thread this block onto the appropriate FreeList. If possible, */
        /* Merge this block with the appropriately-aligned adjacent block, */
        /* thus assuring we have free blocks of the maximum possible size */
        /* on the FreeLists... */
        p->free_dex = -1;   /* prevents multi-free */
        if ( use_sys_mem )
            {
            /* decrement the count of task zero blocks allocated out */
            /* of chunk */
            (chunkP->FreeList[TASK_ZERO_EXT])--;
            }

        max_chunk_dex = MEMDEX(chunkP->chunk_size);
        if (chunk_index >= NumBaseChunks)
            chunk_start = (char *)(((char *)chunkP->FreeList[BASE_CHUNK_LOC]) +
                                                           sizeof(chunk_descr));
        else
            chunk_start = (char *) chunkP->FreeList[BASE_CHUNK_LOC];
        pdelta = (long)((char *)p - chunk_start); /* init the delta */

        for (dex; dex <= max_chunk_dex; dex++)
            {
#if BUILD_DEBUG
            free_tests++;
#endif
            if (chunkP->FreeList[dex] == NULL)
                {   /* Empty Free List */
                p->next = p->prev = NULL;
                chunkP->FreeList[dex] = p;
                goto RETURN;
                }
            else
                {   /* More than one thing on free list */
                if (chunkP->FreeList[dex] > p)
                    {   /* First on the list */
                    p->next = chunkP->FreeList[dex];
                    p->next->prev = p;
                    p->prev = NULL;
                    chunkP->FreeList[dex] = p;
                    }
                else
                    {   /* Thread onto free list in order */
                    for (np = chunkP->FreeList[dex]; np; np = np->next)
                        {
#if BUILD_DEBUG
                        free_tests++;
#endif
                        if (p < np)
                            {   /* We go here */
                            p->prev = np->prev;
                            if (p->prev)
                                p->prev->next = p;
                            else
                                chunkP->FreeList[dex] = p;
                            p->next = np;
                            np->prev = p;
                            goto FREEDONE;
                            }
                        else
                            {
                            if (np->next == NULL)
                                {   /* Check if at the end */
                                np->next = p;
                                p->next = NULL;
                                p->prev = np;
                                goto FREEDONE;
                                }
                            }
                        }
                    }

                /* Now, see if we can merge this block. Must do  */
                /* "aligned" merges only connecting blocks so that "p" */
                /* remains the head of the next (dex+1)-sized block. */
FREEDONE:
                undexed_dex = Undex[dex];

                /* We can AND the Undex[dex] with our offset from the */
                /* start of the Chunk's memory to see if we're aligned. */
                /* If 0, the current "p" is the front of a dex+1 block, */
                /* else it is the second half of a dex+1 block */
#if 0
                if ( ((long)((char *)p - chunk_start)) & undexed_dex)
#else
                if ( pdelta & undexed_dex)
#endif
                    {
                    /* Not aligned, check backwards */
                    if (p->prev)
                        {
                        if ((char *) p->prev == (char *) p - undexed_dex)
                            {
                            /* Prev block is head of dex+1 block */
                            p = p->prev;    
                            /* p changed, update the delta */
                            pdelta = (long)((char *)p - chunk_start);
                            goto PROMOTEFREE;
                            }
                        }
                    goto RETURN;
                    }
                else
                    {   /* This p is aligned, check forwards */
                    if (p->next)
                        {
                        if ((char *) p->next == (char *) p + undexed_dex)
                            /* we can merge it */
                            goto PROMOTEFREE;
                        }
                    }
                goto RETURN;    /* No next or prev, or not adjacent */
                }
            /* p is now the pointer to a dex+1 sized block - remove */
            /* from this free list, and drop through the loop to get */
            /* it on the dex+1 free list */
PROMOTEFREE:    
            merge_count++;
            if (p->prev)
                p->prev->next = p->next->next;
            else
                chunkP->FreeList[dex] = p->next->next;
            if (p->next->next)
                p->next->next->prev = p->prev;
            }       /* End of FOR dex loop */
        dex--;
        }

RETURN:
    LOGGER(("FREE(%d) %lx dex %d", pid, p + 1, dex));

    CHECK_MEMORY
#endif              /*     }  BUILD_PURIFY */

}



/*************************************************************************
function: THIMpid_free - Release some memory for task == THIMpid
 Warning: do not make changes here which are not also made in exactly
          the same manner in TaskFree
*/
void *THIMpid_free(void *inP)
{
    TaskFree(THIMpid, inP);
    return(NULL);
}




/** The following will return null, rather than causing a TaskAbort if the */
/*  request is for too big a chunk of memory */
char *THIMhugeAlloc(pid, size)
int pid;
int size;
{

    char *p;

    if ((size + sizeof(mem_header)) > max_request_size)
        return ((char *) NULL);

    p = TaskAlloc(pid, size);

    return (p);
}





/** The following will return null, rather than causeing a TaskAbort if the */
/*  request is for too big a chunk of memory */
char *THIMhugeRealloc(p, newsize)
void *p;
unsigned newsize;
{

    if ((newsize + sizeof(mem_header)) > max_request_size)
        return ((char *) NULL);

    p = TaskRealloc(THIMpid, p, newsize);

    return (p);
}





/*************************************************************************
function: THIMalloc - Create some memory in current task
*/
char *THIMalloc(unsigned int size)
{
    RETADDR(size);
    return (TaskAlloc(THIMpid, size));
}

/*************************************************************************
function: TaskAlloc - Create some memory
 Warning: do not make changes here which are not also made in exactly
          the same manner in THIMpid_alloc
*/
char *TaskAlloc(pid, size)
register int pid;
unsigned int size;
{
#if BUILD_PURIFY        /*     {  */
    char *p;

    p = (char *) malloc(size);
    purify_set_pool_id(p, pid);
    purify_set_user_data(p, size);
    return (p);
#else

    register mem_header *p, *p2;
    register long dex;
    register long ix, new_ix, jx;
    register long oldsize, newsize;
    register mem_header *highP;
    int rqst;
    int loopcount = 0;
    extern int AxMemoryWindow;
    static int ShutdownUid;
    int bust_block_limit = NO_SPLIT_DEX;
    int undexed_size, i, check_chunk;
    chunk_descr **chunkPP;
    chunk_descr *chunkP;
    int top_chunk;
    int skipped_some, skipped_chunk;
    int use_sys_mem;


    InitializeMemoryStructure();

    RETADDR(pid);

    CHECK_MEMORY

    if (size == 0)
        /* force a minimum size block */
        size = 1;

    size = size + sizeof(mem_header);
    dex = MEMDEX(size);
    undexed_size = Undex[dex];
    use_sys_mem = TaskInfo[pid].deamon;
    skipped_some = 0;   /* -1 = use anything, */
                        /* 0 = no skips yet, so should skip "big" blocks */
                        /* 1 = skipped something on this scan */
    skipped_chunk = -1;

#if BUILD_DEBUG
    alloc_count++;      /** DEBUG **/
#endif

    /**--------------------------------------------------------------*/
    /* Attempt to find some unallocated space, or split a larger block */
    /* This will split the first larger block in the first chunk found */
    /**--------------------------------------------------------------*/
try_allocated:
    if (use_sys_mem)
        check_chunk = 0;/* there is a set of reserved chunks for task zero */
    else
        check_chunk = TaskZeroChunks;   /* start looking here  */

retry_allocated:
    chunkPP = ChunkListPP + check_chunk;

    for (i = check_chunk; i <= Topmost_chunk; i++, chunkPP++)
        {
        chunkP = *chunkPP;

        if ( (chunkP->chunk_size - (long)chunkP->FreeList[BYTES_ALLOCATED]) <
                                                                undexed_size )
            {
            /* not enough space free in this chunk */
            continue; 
            }


        /* Non-system tasks can't use chunks claimed for system memory */
        /* System tasks can only use non-system chunks if no system chunk */
        /*     can service the required block */
        if ( !use_sys_mem )
            {
            /* not a system task */
            if (chunkP->FreeList[TASK_ZERO_EXT])
                continue;   /* skip chunks claimed by system tasks */
            }
        else 
            {
            /* a system task */
            if (skipped_some < 0) 
                /* second time through, use anything available */
                goto check_allocated;

            if ( chunkP->FreeList[TASK_ZERO_EXT] == 0 ) 
                {
                /* this is not currently in use as a "system" chunk */
                if (skipped_some == 0) 
                    {
                    skipped_some = 1;
                    skipped_chunk = i;
                    }
                continue;
                }
            else 
                { 
                /* this chunk is in use as a "system" chunk */
                /* for system tasks, we look for perfect fit first */
                /* rather than splitting the first available       */
                if ( p = chunkP->FreeList[dex] ) 
                    /* perfect match */
                    goto use_it;
                else 
                    {  /* skip it for now - split this one later */
                    if (skipped_some == 0) 
                        {
                        skipped_some = 1;
                        skipped_chunk = i;
                        }
                    continue;
                    }
                }
             }


        /**===================================================*/
        /* If we get here, we know the task can use this chunk */
        /* We go for the freelist and failing that, look for   */
        /* a larger block to split                             */
        /**===================================================*/
check_allocated:
        
        if (p = chunkP->FreeList[dex])
            /* a perfect match */
            goto use_it;

        for (ix = dex + 1; ix <= (long)(chunkP->FreeList[MAX_FREEDEX]); ix++)
            {
            p = chunkP->FreeList[ix];
#if BUILD_DEBUG
            split_search_count++;  /* DEBUG */
#endif
            if (p)
                {
#if BUILD_DEBUG
                split_count++;
#endif
                /* chunk has a block big enough, but we don't want to use */
                /* "big" blocks if the block is already trapped below     */
                /* some other ones */
                if (skipped_some >= 0)
                    {
                    /* first pass through, try to avoid busting big blocks */
                    if (ix >= bust_block_limit)
                        {
                        /* we are looking at a "big" block */
                        if ( !use_sys_mem ) /* sys tasks can bust big chunks */
                            {
                            if ( skipped_chunk < 0 )
                                {
                                /* record for later */
                                skipped_chunk = i;
                                skipped_some = 1;
                                }
                            check_chunk = i + 1;
                            goto retry_allocated;
                            }
                        }
                    }

                /* found a larger block to split */
                chunkP->FreeList[ix] = p->next; /* Remove from free list */
                if (p->next)
                    p->next->prev = NULL;   /* Doubly linked free list now */

                /* Fast Splitter, this will cut up the block into all */
                /* of the larger sizes between what we want and what we found */
                highP = (mem_header *) (((char *) p) + undexed_size);
                for (jx = dex; jx < ix; jx++)
                    {
                    chunkP->FreeList[jx] = highP;
                    highP->prev = highP->next = NULL;
                    highP->free_dex = -1;
                    highP->pid = 0;
                    highP = (mem_header *) (((char *) highP) + Undex[jx]);
#if BUILD_DEBUG
                    true_splits++;
#endif
                    }
                /* p is the block we needed */
                goto got_it;
                }
            }
        }    /* end of for each chunk */


    /*=====================================================================*/
    /* If we drop through to here, we couldn't find an appropriately-owned */
    /* chunk (system vs. non-system) to satisfy the request, or we would */
    /* have had to split a NO_SPLIT_DEX chunk. Before we get more memory, */
    /* go back and try to use any chunks we might have skipped */
    /*=====================================================================*/
   
    if ( skipped_some > 0 ) 
        { 
        /* we skiped some big blocks along the way */
        /* go back and use those first */
        skipped_some = -1;
        check_chunk = skipped_chunk;
        goto retry_allocated;  
        }
 


    /**------------------------------------------------*/
    /* If got here, needs to get more memory from sbrk */
    /* This means that we will be creating a new chunk */
    /**------------------------------------------------*/
sbrk_more:
    if (tmzmalloStruct.TotalMemoryUsage >= AxMemoryWarning)
        {
        if (bust_block_limit == NO_SPLIT_DEX)
            {
            /* we have not tried all possible blocks, do so now before */
            /* going to any warning code                               */
            bust_block_limit = MAX_DEX;
            goto try_allocated;
            }
        AxPostMemoryWarning = TRUE;
        AxMemoryWarning += AxMemoryWarning / 8; /* don't nag too much! */
        }
    if (pid && AxInTask &&
        (tmzmalloStruct.TotalMemoryUsage >= AxMemoryLimit) &&
        (ShutdownUid != ThimUIDPid(AxInTask)))
        {
        ShutdownUid = ThimUIDPid(AxInTask);
        taskAbort(XLT("You have exceeded axMemoryLimit."));
        }


    /*************************/
    /** Allocate the memory **/
    /*************************/
    rqst = (undexed_size <= ChunkSize ? ChunkSize : undexed_size);

    if ((Topmost_chunk + 1) >= NumBaseChunks)
        {
        /* this will create a chunk beyond the range of pre-allocated descrs */
        /* so we must add in space for the chunk descriptor at the head of   */
        /* this block                                                        */
        rqst = rqst + sizeof(chunk_descr);
        }

    p = (mem_header *) SBRK(rqst);
    if (-1 != ((long) p))
        {
sbrk_success:
        /* we got what we wanted */
        AddChunkToList(p, rqst);
        check_chunk = Topmost_chunk;
        top_chunk = Topmost_chunk;
        skipped_some = -1;   /* force use of this block on first pass through */
        goto retry_allocated;
        }

    /**------------------------------------------------*/
    /* Cannot get memory from sbrk! */
    /**------------------------------------------------*/
    if (bust_block_limit == NO_SPLIT_DEX)
        {
        /* we have not tried all possible blocks, do so now before */
        /* going to any warning code                               */
        bust_block_limit = MAX_DEX;
        goto try_allocated;
        }
    if (AxOutOfMemory || !AxMemoryWindow)
        {           /* recursive recovery */
#if AXMS
        tempAbort(XLT(msg));
#else
        write(1, XLT(msg), strlen(XLT(msg)));   /* say goodnight! */
        kill(getpid(), 9);  /* buy farm */
#endif
        }
    AxOutOfMemory = TRUE;   /* tailbite preventer */
    RingTheBell();      /* wake up user */
    AxClearAllWindows();

loop_disaster:
    if (++loopcount == 7)
        {           /* ring bell every 14 seconds */
        loopcount = 0;
        RingTheBell();
        }
    if (AxInTask && (pid != 0))
        {
        AxShowMemoryWindow(TRUE, TRUE); /* complain & allow task kill */
        AxAllowStop = TRUE;
        }
    else
        {
        AxShowMemoryWindow(TRUE, FALSE);    /* complain */
        AxAllowStop = FALSE;
        }

#ifdef sleep
#undef sleep
#endif
    sleep(2);       /* pause */
    p = (mem_header *) SBRK(rqst);  /* try again */
    if (-1 == ((long) p))    /* and again... */
        goto loop_disaster;

    /* we got it, get on with life */
    AxOutOfMemory = FALSE;
    AxShowMemoryWindow(FALSE, FALSE);   /* assumes this does not malloc */
    AxRePaintAll(0);

    /* we got what we wanted */
    goto sbrk_success;



/*-------------------------------------------------------------------------*/
/* Here lies the code that handles the assignment of the block to the task */
/*-------------------------------------------------------------------------*/
   
use_it:
#if BUILD_DEBUG
    found_on_freelist++;
#endif
    (*chunkPP)->FreeList[dex] = p->next;    /* Remove from free list */
    if (p->next)
        p->next->prev = NULL;    /* Doubly linked free list now */
 
got_it:
    p->prev = NULL; /* I am head of pid's memory */
    p->free_dex = MAKE_FREE_DEX(chunkPP, dex);
    p->pid = pid;
    if ( use_sys_mem )
        {
        /* increment the count of system blocks allocated from this chunk */
        ((*chunkPP)->FreeList[TASK_ZERO_EXT])++;
        }
 
#if MEM_CALL_TRACE
    p->ret_addr = AxMemRet; /* return address to caller */
    AxMemRet = NULL;/* reset for next use */
#endif
    p2 = MemHeads[pid]; /* Get old head */
    if (p->next = p2)   /* I point to it as my next */
        p2->prev = p;   /* It points back to me as prev */
    MemHeads[pid] = p;  /* register this new storage */
    AxMemUsage[pid] = AxMemUsage[pid] + undexed_size;
    (*chunkPP)->FreeList[BYTES_ALLOCATED] =
        (mem_header *)(((long)(*chunkPP)->FreeList[BYTES_ALLOCATED]) +
                       undexed_size);
 
    CHECK_MEMORY

    RECORD_FOR_LEAKS

    TRAP_ADDR((char *) (p + 1));            /* See MemTrapTest */

    return ((char *) (p + 1));  /* return addr right after the header */
 

#endif              /*     }  BUILD_PURIFY */

}








/*************************************************************************
function: THIMpid_alloc - Create some memory for task == THIMpid
 Warning: do not make changes here which are not also made in exactly
          the same manner in TaskAlloc
*/
char *THIMpid_alloc(size)
unsigned int size;
{
    return(TaskAlloc(THIMpid, size));
}





/*************************************************************************
function: PermAlloc - Alloc (densely) for an alloced-once object
This saves the powers-of-two overhead.  We must however allocate a chunk
for this block which will appear to always be busy, since we can not sbrk
below this point in the future.
*/
char *PermAlloc(unsigned int size)
{
#if BUILD_PURIFY        /*     {  */
    char *p;

    p = (char *) malloc(size);
    return (p);
#else

    register mem_header *p;

    CHECK_MEMORY

    size = size + sizeof(mem_header);

#if FLOAT_ALIGNED_8BYTE
    size = ((size + 7) >> 3) << 3;
#else
    size = ((size + 3) >> 2) << 2;
#endif

    p = (mem_header *) SBRK(size);
    if (-1 == ((long) p))
        tempAbort(NXLT("Not enough memory"));

    /* we got what we wanted, make a note of the end of this block since */
    /* we will not be able to free memory back to the system below this point */
    MinFreeUpAddress = ((char *) p) + size;

    /* I am head of pid's memory */
    p->free_dex = -1;   /* list as being freed (for other users) */
    p->next = p->prev = NULL;
    tmzmalloStruct.MemPermanent += size;

    CHECK_MEMORY

    return ((char *) (p + 1));  /* return addr right after the header */
#endif              /*     }  BUILD_PURIFY  */
}


/*************************************************************************
function: TaskCopy - Copy allocated memory
*/
char *TaskCopy(int pid, void *inP)
{
    register mem_header *p = (mem_header*)inP;
    register char *new;
    register int size;

#if BUILD_PURIFY
    size = purify_get_user_data(p);
    new = (char *) malloc(size);
    purify_set_pool_id(new, pid);
    purify_set_user_data(new, size);
#else
    size = Undex[GET_DEX((p - 1)->free_dex)] - sizeof(mem_header);
    new = TaskAlloc(pid, size);
#endif

    memcpy(new, p, size);
    return (new);
}

/*************************************************************************
function: THIMrealloc - Adjust memory size up/down in current task
*/
char *THIMrealloc(void *p, unsigned newsize)
{
    if (!p)
        RETADDR(p);
    return (TaskRealloc(THIMpid, p, newsize));
}

/*************************************************************************
function: TaskRealloc - Adjust memory size up/down
 Warning: do not make changes here which are not also made in exactly
          the same manner in THIMpid_realloc
*/
char *TaskRealloc(int pid, void *inP, unsigned int newsize)
{
    mem_header *p = (mem_header*)inP;
#if BUILD_PURIFY        /*     {  */
    char *charP;

    if (!p)
        charP = (char *) malloc(newsize);
    else
        charP = (char *) realloc((char *) p, newsize);
    if(p == NULL)
        purify_set_pool_id( charP, pid );
    purify_set_user_data( charP, newsize );

    return (charP);

#else               /*     }          {  */
    register int old_dex, new_dex, oldsize;
    char *newP;

#if MEM_CALL_TRACE
    int *ret_addr;      /* return address to caller */

#endif

    CHECK_MEMORY

    if (!p)
        {
        RETADDR(pid);
        return (TaskAlloc(pid, newsize));
        }
    p--;
#if MEM_CALL_TRACE
    ret_addr = p->ret_addr;
#endif
    old_dex = GET_DEX(p->free_dex);
    new_dex = MEMDEX(newsize + sizeof(mem_header));
    if (old_dex == new_dex)
        return ((char *) (p + 1));
    newP = TaskAlloc(pid, newsize);
    oldsize = Undex[old_dex] - sizeof(mem_header);
    memcpy(newP, p + 1, MIN(newsize, oldsize));
    TaskFree(pid, p + 1);
#if MEM_CALL_TRACE
    p = (mem_header *) newP;
    (p - 1)->ret_addr = ret_addr;
#endif
    return (newP);
#endif              /*     } BUILD_PURIFY */
}




/*************************************************************************
function: THIMpid_realloc - Adjust memory size up/down for task == THIMpid
 Warning: do not make changes here which are not also made in exactly
          the same manner in TaskRealloc
*/
char *THIMpid_realloc(void *inP, unsigned newsize)
{
    return(TaskRealloc(THIMpid, inP, newsize));
}

/*************************************************************************
function: THIMcopy - Create a duplicate datum
*/
char *THIMcopy(void *inP)
{
    register mem_header *p = (mem_header*)inP;
    register char *newP;
    register int size;

#if BUILD_PURIFY
    size = purify_get_user_data(p);
    newP = THIMalloc(size);
#else

    RETADDR(p);
    size = Undex[GET_DEX((p - 1)->free_dex)] - sizeof(mem_header);
    newP = THIMalloc(size);
#endif
    memcpy(newP, p, size);
    return (newP);
}


/*************************************************************************
function: recalloc - A call to realloc plus zero fills
*/
char *recalloc(px, oldsize, newsize)
char *px;
#if !CLIX
size_t oldsize, newsize;
#else
unsigned int oldsize, newsize;
#endif
{
    char *ret;

    if (oldsize == newsize)
        return ((char *) px);

    if ((ret = THIMrealloc(px, newsize)) == NULL)
        return (NULL);
    if (newsize > oldsize)
        memset(ret + oldsize, 0, newsize - oldsize);
    else if (oldsize > newsize)
        memset(ret + newsize, 0, oldsize - newsize);
    return (ret);
}

/*************************************************************************
function: THIMcalloc - A stupid call to malloc plus zero fills
*/
char *THIMcalloc(n, size)
unsigned int n, size;
{
    RETADDR(n);
    return (TaskCalloc(THIMpid, n, size));
}


/*************************************************************************
function: TaskCalloc - malloc plus zero fills
 Warning: do not make changes here which are not also made in exactly
          the same manner in THIMpid_calloc
*/
char *TaskCalloc(pid, n, size)
int pid;
unsigned int n, size;
{
    register char *p;

    RETADDR(pid);
    size *= n;
    p = TaskAlloc(pid, size);
    memset(p, 0, size);
    return ((char *) p);
}


/*************************************************************************
function: THIMpid_calloc - malloc plus zero fills memory for task == THIMpid
 Warning: do not make changes here which are not also made in exactly
          the same manner in TaskCalloc
*/
char *THIMpid_calloc(n, size)
unsigned int n, size;
{
    register char *p;

    RETADDR(THIMpid);
    size *= n;
    p = TaskAlloc(THIMpid, size);
    memset(p, 0, size);
    return ((char *) p);
}




/*************************************************************************
function: THIMfreeAll - Release all memory associated with a task
*/
THIMfreeAll(pid)
int pid;
{
    register mem_header *p, *next;

#if BUILD_PURIFY
    purify_map_pool(pid, free);
#else
    for (p = MemHeads[pid]; p; p = next)
        {
        next = p->next;
        TaskFree(pid, p + 1);
        }
    MemHeads[pid] = NULL;
#endif
}


MemParanoid()
{
    CHECK_MEMORY
}

#if BUILD_PURIFY        /*     {  */
ThimMemCheck()
{
}

data_ptr ElfMemoryUsage() {return(0);}
AxReportMemUsage(char *filename){}
static void AddChunkToList(){}


#else               /*     }   BUILD_PURIFY       {  */

extern data_ptr AxWriteArray AXP_((data_ptr ap, int dex, data_ptr new_val));
extern data_ptr AxAddStrToArray AXP_((data_ptr ap, int dex, char *str));
extern data_ptr AxAddWideStrToArray AXP_((data_ptr ap, int dex, wchar * str));
extern data_ptr AxAddIntToArray AXP_((data_ptr ap, int dex, int n));
extern data_ptr AxAddFloatToArray AXP_((data_ptr ap, int dex, double dbl));





/*************************************************************************
function: AxReportMemUsage - Create a mem usage file
The memory usage file is suitable for reading by DBX on SUN3 (sunstroke)
*/
AxReportMemUsage(filename)
char *filename;
{
#if MEM_CALL_TRACE
    int fd;
    mem_header *p;
    int top = MAX_PIDS;
    char buf[30];
    unsigned long as_int;
    char *hexit = NXLT("0123456789ABCDEF");
    int pid;
    int size;

    if (AxSysOnlyMemTrace)
        top = 1;
    fd = creat(filename, 0666);
    for (pid = 0; pid < top; pid++)
        {
        for (p = MemHeads[pid]; p; p = p->next)
            {
            if (p->ret_addr)
                {   /* cannot call sprintf here! */
                size = Undex[GET_DEX(p->free_dex)];
                buf[0] = '#';
                buf[1] = ' ';
                buf[2] = size / 10000000 + '0';
                size = size % 10000000;
                buf[3] = size / 1000000 + '0';
                size = size % 1000000;
                buf[4] = size / 100000 + '0';
                size = size % 100000;
                buf[5] = size / 10000 + '0';
                size = size % 10000;
                buf[6] = size / 1000 + '0';
                size = size % 1000;
                buf[7] = size / 100 + '0';
                size = size % 100;
                buf[8] = size / 10 + '0';
                size = size % 10;
                buf[9] = size / 1 + '0';
                size = size % 1;
                buf[10] = '\n';
                buf[11] = 0;
                write(fd, buf, strlen(buf));
                as_int = (unsigned long) (p->ret_addr);
                buf[0] = '0';
                buf[1] = 'x';
                buf[2] = hexit[(as_int & 0xF0000000) >> 28];
                buf[3] = hexit[(as_int & 0x0F000000) >> 24];
                buf[4] = hexit[(as_int & 0x00F00000) >> 20];
                buf[5] = hexit[(as_int & 0x000F0000) >> 16];
                buf[6] = hexit[(as_int & 0x0000F000) >> 12];
                buf[7] = hexit[(as_int & 0x00000F00) >> 8];
                buf[8] = hexit[(as_int & 0x000000F0) >> 4];
                buf[9] = hexit[(as_int & 0x0000000F) >> 0];
                buf[10] = '/';
                buf[11] = 'i';
                buf[12] = '\n';
                buf[13] = 0;
                write(fd, buf, strlen(buf));
                p->ret_addr = 0;    /* keep from next report */
                }
            }
        }
    close(fd);
#endif
}



/*************************************************************************
function: ThimMemCheck - Validate sanity of alloced storage
*/
ThimMemCheck()
{
    register int indx, dex;
    register mem_header *p, *next;
    register int pid;
    int count, i;
    chunk_descr *chunkP;

    /* Check the allocated memory list */

    count = 0;
    for (pid = 0; pid < MAX_PIDS; pid++)
        {
        for (p = MemHeads[pid]; p; p = next)
            {
            next = p->next;
            dex = GET_DEX(p->free_dex);
            if ((p->free_dex == -1) || (pid != p->pid) ||
#if FLOAT_ALIGNED_8BYTE
                ((long) p & 7) ||
#else
                ((long) p & 3) ||
#endif
                (dex <= 0) || (dex > MAX_DEX))
                tempAbort(NXLT("memory corrupted, alloc list has been roached"));
            count++;
            }
        }

    /* Check the free list */
    count = 0;
    for (i = 0; i <= Topmost_chunk; i++)
        {
        chunkP = *(ChunkListPP + i);
        if ( chunkP->FreeList[4] )
            tempAbort(NXLT("memory corrupted, illegal usage of free list"));
        for (indx = FIRST_REAL_DEX; 
                         indx <= (long)(chunkP->FreeList[MAX_FREEDEX]); indx++)
            {
            for (p = chunkP->FreeList[indx]; p; p = next)
                {
                next = p->next;
                if ((p->free_dex != -1) || 
#if FLOAT_ALIGNED_8BYTE
                    ((long) p & 7) )
#else
                    ((long) p & 3) )
#endif
                    tempAbort(NXLT("memory corrupted, free list has been roached"));
                count++;
                }
            }
        }
}


/*************************************************************************
function: MemTrapTest - Trap the alloc or free of a specific location.

HOW TO USE THIS FACILITY:
  - Build (at least) this file in debug mode.
  - Discover the address of memory that is being re-used without being
    freed, or freed while in use, etc. Verify the scenario is reproducible!
  - In the debugger, set MemoryTrap to that address, and set a breakpoint
    in ThimMemTrap. When the specified address is alloc'd or freed,
    ThimMemTrap (a no-op function) will be called.
  - When done, reset MemoryTrap to zero.

LIMITATIONS:
  - This only checks for the direct alloc or free of the address itself.
    It does not check if the address falls within a chunk of alloc'd memory.
*/
static void MemTrapTest( p)
register void *p;
{
    if (p == MemoryTrap)
        ThimMemTrap(p);
}

void ThimMemTrap( p)            /* SET BREAK ON THIS FUNCTION */
register void *p;
{
    ;
}

/*************************************************************************/
/* function: AddChunkToList - Create a new chunk and put a new piece    */
/*    of memory on it.                                                   */
/*************************************************************************/
static void AddChunkToList(in_p, size)
mem_header *in_p;
int size;
{
    chunk_descr *chunkP;
    mem_header *memP;
    register int i;
    int chunk_dex;

    tmzmalloStruct.TotalMemoryUsage = tmzmalloStruct.TotalMemoryUsage + size;

    Topmost_chunk++;
    if (Topmost_chunk >= NumBaseChunks)
        {
        /* we have exceeded the pre-allocated chunk descriptors, so the */
        /* chunk descriptor is actually at the start of the memory block */
        /* passed in.  The caller is responsible for making sure that the */
        /* additional space for the chunk descriptor is there.            */
        chunkP = (chunk_descr *) in_p;
        memP = (mem_header *) (((char *) in_p) + sizeof(chunk_descr));
        *(ChunkListPP + Topmost_chunk) = chunkP;
        size = size - sizeof(chunk_descr);
        }
    else
        {
        /* we are within the pre-allocated range */
        chunkP = *(ChunkListPP + Topmost_chunk);
        memP = in_p;
        }

    memset(chunkP, 0, sizeof(chunk_descr));
    chunkP->chunk_size = size;
    chunkP->FreeList[BASE_CHUNK_LOC] = in_p;

    /* Put entire chunk on Free List, so it can be busted instead of carved */
    chunk_dex = MEMDEX(size);
    chunkP->FreeList[TASK_ZERO_EXT] = 0;
    chunkP->FreeList[chunk_dex] = memP;
    chunkP->FreeList[MAX_FREEDEX] = (mem_header *) chunk_dex;
    chunkP->FreeList[BYTES_ALLOCATED] = (mem_header *) 0;

    memP->next = NULL;
    memP->prev = NULL;  /* Now doubly linked */
    memP->free_dex = -1;
    return;
}




#if !BUILD_TEST         /*     {  */

/*************************************************************************
function: ElfMemoryUsage - Return memory usage info
*/
data_ptr ElfMemoryUsage()
{
    int total;      /* Amount currently allocated */
    int pid;
    data_ptr aP, sP, pP;
    int numPids = 0;
    char str[200], *p;

    aP = AxMakeArray(0);
    aP = AxAddFloatToArray(aP, 0, (double) tmzmalloStruct.TotalMemoryUsage);
    aP = AxAddFloatToArray(aP, 1, (double) AxMemUsage[0]);
    pP = AxMakeArray(0);
    total = AxMemUsage[0];
    for (pid = 1; pid < MAX_PIDS; pid++)
        {
        if (TaskInfo[pid].in_use)
            {
            sP = AxMakeArray(3);
            AxAddIntToArray(sP, 0, TaskInfo[pid].uid);
            strcpy(str, TaskInfo[pid].argv[0]);
            for (p = str; *p; p++)
                {
                if (islower(*p))
                    *p = toupper(*p);
                else
                    *p = *p;
                }
            AxAddStrToArray(sP, 1, str);
            AxAddFloatToArray(sP, 2, (double) AxMemUsage[pid]);
            pP = AxWriteArray(pP, numPids++, sP);
            }
        total += AxMemUsage[pid];
        }
    total += tmzmalloStruct.MemPermanent;
    aP = AxAddFloatToArray(aP, 2, (double) total);
    aP = AxWriteArray(aP, 3, pP);
    return (aP);
}

/************************************************************************
function:   AxReportNumMemoryItems - report the number of currently
            allocated by the task.
*/
AxReportNumMemoryItems()
{
#if MEM_CALL_TRACE
    mem_header *p;
    int num = 0;

    for (p = MemHeads[THIMpid]; p; p = p->next)
        {
        if (p->ret_addr)
            num++;
        }
    return (num);
#endif
}



#endif              /*     }  BUILD_TEST */



#if AX_NT || LINUX
#undef alloca           /* defined to be an intrinsic function in MS C */
#endif
void *alloca(int size)
{
    taskAbort(NXLT("alloca"));
}

#if DBX_MODE
AlcLogger(a, b, c, d, e, f, g)
int a, b, c, d, e, f, g;
{
    if (!AlcLogging)    /* interlock prevents recursive log */
        return;
    AlcLogging = FALSE;
    if (!AlcLogFp)
        AlcLogFp = fopen(NXLT("zzmem.log"), WRITE_STR);
    fprintf(AlcLogFp, a, b, c, d, e, f, g);
    fprintf(AlcLogFp, "\n");
    fflush(AlcLogFp);
    AlcLogging = TRUE;
}

#endif



#if 0    /* { */
/*************** THESE HAVE BEEN MOVED TO axmalloc.c *******************/
/***********************************************************************/
/* Other sys's runtime libraries malloc to the REAL malloc, so         */
/*      free in things like in agetNFname the free doesn't work */
/* Compatabilty funtions so ATM kernel and shared runtimes can malloc... */
/* The reason there are so many flavors here is to make the definintions */
/*   match the definitions in the system include file "malloc.h" which the */
/*   newer compiler will insist upon (and which would seem to be good     */
/*   practive in any case) - jon                                          */
/***********************************************************************/
#if SYSV4 || ALPHA || HP800 || HP300 || RIOS || CLIX || AX_NT || SCOUNIX || SGI  /*     {   */
void *malloc(size)
size_t size;
{
    RETADDR(size);

    return ((void *) TaskAlloc(0, size));
}
void *realloc(px, size)
void *px;
size_t size;
{
    return ((px == NULL) ? (void *) TaskAlloc(0, size) : (void *) TaskRealloc(0, (mem_header *) px, size)); /* type fixing -- lhunt */
}

void free(p)
void *p;
{
    if (p)
        TaskFree(0, p);
}
void *calloc(n, size)
size_t n, size;
{
    RETADDR(n);
    return ((void *) TaskCalloc(0, n, size));
}

#if AX_NT
void* _nh_malloc(size,nh_flag)
size_t size;
int nh_flag;
{
  return(malloc(size));
}
#endif

#else         /*     }  SYSV4 || HP800 || RIOS || CLIX || AX_NT || SCOUNIX { */

#if PMAX            /*     { */
void *malloc(size)
register unsigned size;
{
    RETADDR(size);
    return ((void *) TaskAlloc(0, size));
}

void *realloc(px, size)
register void *px;
register unsigned size;
{
    return ((px == NULL) ? (void *) TaskAlloc(0, size) : (void *) TaskRealloc(0, (mem_header *) px, size)); /* type fixing -- lhunt */
}

void free(p)
register void *p;
{
    if (p)
        TaskFree(0, p);
}

#else               /*     } PMAX         { */

#if LINUX  /* { */

char *malloc(size)
size_t size;
{
    RETADDR(size);

    return ((void *) TaskAlloc(0, size));
}
char *realloc(px, size)
char *px;
size_t size;
{
    return ((px == NULL) ? (void *) TaskAlloc(0, size) : (void *) TaskRealloc(0, (mem_header *) px, size)); /* type fixing -- lhunt */
}

void free(p)
char *p;
{
    if (p)
        TaskFree(0, p);
}
char *calloc(n, size)
size_t n, size;
{
    RETADDR(n);
    return ((void *) TaskCalloc(0, n, size));
}

#else    /*   }   LINUX  {  */

char *malloc(size)
register unsigned size;
{
    RETADDR(size);
    return (TaskAlloc(0, size));
}

char *realloc(px, size)
register void *px;
register unsigned size;
{
    return ((px == NULL) ? TaskAlloc(0, size) : TaskRealloc(0, (mem_header *) px, size));   /* type fixing -- lhunt */
}

#if M_88K || ALPHA /* Need void return type for ALPHA - SCM */
void
#endif
free(p)
register void *p;
{
    if (p)
        TaskFree(0, p);
}

#endif      /* end of if LINUX         } */
#endif      /* end of if PMAX         } */
#endif      /* end of SYSV4 || HP800 || RIOS || CLIX || AX_NT || SCOUNIX  }  */


#if !SYSV4 && !ALPHA && !HP800 && !HP300 && !RIOS && !CLIX && !AX_NT && !SCOUNIX && !SGI && !LINUX /* { */ 
char *calloc(n, size)
register unsigned n, size;
{
    RETADDR(n);
    return (TaskCalloc(0, n, size));
}

#endif /* } */


#if RIOS || ALPHA
#ifdef valloc
#undef valloc
#endif
#endif

void *valloc(size)
register size_t size;
{
    register char *p;
    register short oldpid = THIMpid;

    tempAbort(NXLT("called valloc"));

}
#endif /* } #if 0 */


#endif              /*     }  !BUILD_PURIFY */





/* Calls so that normally THIMpid compilation units can do permanent mallocs */

char *TrueMalloc(size)
int size;
{
    RETADDR(size);
    return (TaskAlloc(0, size));
}

char *TrueCalloc(n, size)
int n;
int size;
{
    RETADDR(n);
    return (TaskCalloc(0, n, size));
}

char *TrueRealloc(p, size)
void *p;
int size;
{
    if (p == (char *) NULL)
        return (TaskAlloc(0, size));
    else
        return (TaskRealloc(0, (mem_header *) p, size));
}



TrueFree(void *p)
{
    if (p)
        TaskFree(0, p);
}


/************************************************************/
/* As a bit of a "local" hack, here we read the "/etc/hosts" file */
/* to see if it contains an entry for nova89 which matches what we  */
/* expect here at Applix                                          */

int AreRunningAtApplix()
{
    FILE *fp;
    char buff[101];
    static int running_value = -1;

    if (running_value >= 0)
        /* we have already checked */
        return (running_value);

    /* look in /etc/hiosts for "89.0.00.149 nova89"    */
    running_value = FALSE;  /* assume we are not */

    fp = fopen("/etc/hosts", "r");
    if (fp == (FILE *) NULL)
        return (running_value);

    while (fgets(buff, 100, fp) != (char *) NULL)
        {
        if (strncmp(buff, "89.0.00.149 nova89", 18) == 0)
            {
            /* we found it, we are running at applix */
            running_value = TRUE;
            break;
            }
        }
    fclose(fp);
    return (running_value);
}


#if !BUILD_PURIFY       /*     {   */

/*********************************************************************/
/** Dump out the current memory picture to "/tmp/mem_stat.<num>"     */
/*********************************************************************/
Print_Memory_Stats()
{

#if BUILD_DEBUG    /*  {  */
    static int file_number = 0;
    char pathname[60];
    int dex_count[32];
    int chunk_count[200];
    FILE *fp;
    chunk_descr *chunkP;
    int ix, jx, i, total_usage;
    mem_header *p;
    char *strP;
    char buff[30], outline[200];
    int dex_by_chunk[200][32];
    char task_name[100];
    int total_free_space;

    sprintf(pathname, "/tmp/mem_stat.%d", file_number);
    file_number++;

    fp = fopen(pathname, "w+");
    if (fp == (FILE *) NULL)
        return;

    fprintf(fp, "ChunkSize = %d\n", ChunkSize);
    fprintf(fp, "Topmost_chunk = %d     TaskZeroChunks = %d\n",
        Topmost_chunk, TaskZeroChunks);
    fprintf(fp, "\nUsage counts are:\n");

    fprintf(fp, "    number of call to allocate = %d\n", alloc_count);
    fprintf(fp, "    number of times request found on free list = %d\n",
                                                            found_on_freelist );
    fprintf(fp, "    number of times we look for bigger blocks = %d\n",
                                                     split_search_count );
    fprintf(fp, "    number of times we split a larger block = %d\n", 
                                                     split_count );
    fprintf(fp, "    number of actual splits = %d\n", true_splits );

    fprintf(fp, "\n    number of call to free = %d\n", free_count );
    fprintf(fp, "    number of merges done = %d\n", merge_count );
    fprintf(fp, "    number of tests done to sort free lists = %d\n", free_tests );

    fprintf(fp, "\n    number of task reallocs = %d\n", task_reallocs );
    fprintf(fp, "    number of system task realloc adjustments = %d\n", task_realloc_adjusts);
    
    fprintf(fp, "\n    total memory = %d\n", tmzmalloStruct.TotalMemoryUsage);
    fprintf(fp, "    memwarn = %d   memlimit=%d", AxMemoryWarning, AxMemoryLimit);

    /**-------------------------------------------------------**/
    /** dump the counts of the allocated blocks for each task **/
    /**-------------------------------------------------------**/
    for (i = 0; i < MAX_PIDS; i++)
        {
        p = MemHeads[i];
        if (p == (mem_header *) NULL)
            {
            continue;
            }

        task_name[0] = '\0';
#if !BUILD_TEST
        if (TaskInfo[i].argv != NULL)
            {
            if (TaskInfo[i].argv[0] != NULL)
                {
                strcpy(task_name, TaskInfo[i].argv[0]);
                strP = task_name;
                for (strP = task_name; *strP; strP++)
                    {
                    if (islower(*strP))
                        *strP = toupper(*strP);
                    }
                }
            }
#endif

        fprintf(fp, "\n\n***   Memory usage for task <%d>  \"%s\"***\n",
            i, task_name);
        memset(dex_count, 0, sizeof(dex_count));
        memset(chunk_count, 0, sizeof(chunk_count));
        memset(dex_by_chunk, 0, sizeof(dex_by_chunk));
        total_usage = 0;
        while (p != (mem_header *) NULL)
            {
            jx = GET_DEX(p->free_dex);
            ix = GET_INDEX(p->free_dex);
            dex_count[jx]++;
            chunk_count[ix]++;
            dex_by_chunk[ix][jx]++;
            total_usage = total_usage + Undex[jx];
            p = p->next;
            }
        fprintf(fp, "   total memory usage = %d\n", total_usage);
        fprintf(fp, "\t\tdex\tsize\tcount\ttotal bytes\n");
        for (ix = FIRST_REAL_DEX; ix < 32; ix++)
            {
            if (dex_count[ix] > 0)
                fprintf(fp, "\t\t%d\t%d\t%d\t%d\n", ix, Undex[ix], dex_count[ix], Undex[ix] * dex_count[ix]);
            }

        fprintf(fp, "    \n");
        for (ix = 0; ix <= Topmost_chunk; ix++)
            {
            if (chunk_count[ix] > 0)
                {
                fprintf(fp, "   blocks used within chunk (%d) = %d\n",
                    ix, chunk_count[ix]);
                fprintf(fp, "     distribution of blocks used <dex>-<count>\n");
                strcpy(outline, "    ");
                for (jx = 0; jx < 32; jx++)
                    {
                    if (dex_by_chunk[ix][jx] > 0)
                        {
                        sprintf(buff, "%d-%d  ", jx, dex_by_chunk[ix][jx]);
                        strcat(outline, buff);
                        }
                    }
                fprintf(fp, "          %s\n", outline);
                }
            }
        }

    /**-------------------------------------------------------**/
    /** Dump out the chunks themselves                        **/
    /**-------------------------------------------------------**/
    fprintf(fp, "\n\n\n");
    fprintf(fp, "Free lists by chunk:\n");
    for (ix = 0; ix <= Topmost_chunk; ix++)
        {
        chunkP = *(ChunkListPP + ix);
        memset(dex_count, 0, sizeof(dex_count));
        total_free_space = 0;
        for (jx = FIRST_REAL_DEX; jx < 32; jx++)
            {
            p = chunkP->FreeList[jx];
            while (p != (mem_header *) NULL)
                {
                dex_count[jx]++;
                total_free_space = total_free_space + Undex[jx];
                p = p->next;
                }
            }
        fprintf(fp,
            "Chunk %d      allocated = %d, on-free-list=%d\n",
               ix, (long) chunkP->FreeList[BYTES_ALLOCATED], total_free_space);

        if (chunkP->FreeList[TASK_ZERO_EXT])
            fprintf(fp, "              Chunk used by System Tasks/deamons\n");

        if (total_free_space > 0)
            {
            fprintf(fp, "   free count by dex:   dex     count\n");
            for (jx = FIRST_REAL_DEX; jx < 32; jx++)
                {
                if (dex_count[jx] > 0)
                    fprintf(fp,
                    "                       %d        %d\n", jx, dex_count[jx]);
                }
            }
        }

    fclose(fp);

#endif    /*  }  BUILD_DEBUG */
}


/***********************************************************************/
/* See if we can release meory to the system, and if so, do so         */
/***********************************************************************/

FreeUpMemory()
{
    char *baseP;
    chunk_descr *chunkP;
    int i, new_top, ret;

#if SGI 
    return;
#else

#if RIOS 
    /* not yet implemented */
    return;
#endif

    baseP = (char *) NULL;

    for (i = Topmost_chunk; i >= TaskZeroChunks; i--)
        {
        chunkP = *(ChunkListPP + i);
        if (chunkP->FreeList[BYTES_ALLOCATED] == 0)
            {
            /* this chunk is entirely free, we can move memory back down */
            /* to where this chunk used to start                         */
            if ((char *)(chunkP->FreeList[BASE_CHUNK_LOC]) <= MinFreeUpAddress)
                /* this block is below PermAlloc() memory, can not free it */
                break;
            baseP = (char *) chunkP->FreeList[BASE_CHUNK_LOC];
            tmzmalloStruct.TotalMemoryUsage =
                tmzmalloStruct.TotalMemoryUsage - chunkP->chunk_size;
            new_top = i - 1;
            }
        else
            /* quit here, since the chunk still has things which are busy */
            break;
        }

    if (baseP != (char *) NULL)
        {
        /* move memory back down */
        if (ax_getenv("PRINT_STATS") != NULL)
            Print_Memory_Stats();

#if RIOS || AXMS    /*  {  */
        /* for these platforms we must return individual chunks */
        for (i = new_top + 1; i <= Topmost_chunk; i++)
            {
            chunkP = *(ChunkListPP + i);
#if RIOS
            free_rs6000_memory((char *) chunkP->FreeList[BASE_CHUNK_LOC]);
#else
            free_NT_memory( (void *)chunkP->FreeList[BASE_CHUNK_LOC] );
#endif
            }

#else     /*  }  {  */

        /* just rest the highest valid address to return everything above */
        ret = brk(baseP);
        if (ret != 0)
            {
            char buff[80];

            sprintf(buff, NXLT("brk back down fails, errno=%d\n"), errno);
            tempAbort(buff);
            }
#endif    /* }  */

        Topmost_chunk = new_top;
        }
#endif /* if sgi */
}




#endif              /*     } !BUILD_PURIFY  */

#if BUILD_PURIFY
FreeUpMemory()  {}
Print_Memory_Stats()    {}
#endif





#if AXMS    /*  { */
void *alloc_NT_memory( size )
int  size ;
{
    void *ret;
    ret = GlobalAlloc( GMEM_FIXED, size );
    return( ret == NULL ? (void *)(-1) : ret);
}


void free_NT_memory( ptr )
void *ptr;
{
    HGLOBAL ret;

    ret = GlobalFree( ptr );
    if ( ret != (HGLOBAL)NULL )
       tempAbort("GlobalFree error");
    return;
}


#endif  /*  } AXMS */



#if BUILD_TEST          /*     {  */

short THIMpid = 1;

main()
{
    char *pp[200];
    int i, size;

    InitializeMemoryStructure();

    size = 38990;
    pp[0] = TaskAlloc(1, size);
    pp[1] = TaskAlloc(1, ChunkSize + 1);

    for (i = 2; i < 180; i++)
        {
        pp[i] = TaskAlloc(1, size);
        size = size / 2;
        if (size == 0)
            size = 8990;
        }
    for (i = 31; i >= 0; i--)
        {
        TaskFree(1, pp[i]);
        }
    Print_Memory_Stats();

    size = 70;
    for (i = 0; i < 12; i++)
        {
        pp[i] = TaskAlloc(1, size);
        }

    FreeUpMemory();

    Print_Memory_Stats();

    for (i = 126; i >= 0; i--)
        {
        TaskFree(1, pp[i]);
        }

    Print_Memory_Stats();

    FreeUpMemory();
}

#endif              /*     }  BUILD_TEST  */
