/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: winxfer.c
   Applix ASC version: 121
   Date: Wed Mar 10 21:09:54 1999
*/
#include <stdio.h>
#include <string.h>
#include <basicwin.h>
#include <aos.h>
#include <afile.h>
#include <paripc.h>
#include <scac.h>
#include <getstr5.h>
#include <getstr10.h>
#include <getstr20.h>
#include <getstr27.h>
#include <xferfmt.h>
#include <fsrts.h>
#include <aos_xfer.h>
#include <akeys.h>
#include <setjmp.h>
#include <thimble.h>
#include <doctypes.h>
#include <adlgpoke.h>
#include <xmlib.h>
#include <adocio.h>
#include <java_syms.h>
#if SYSV4
#include <utils.h>
#endif

extern Display *Dpy;    /* Resolution of DisplayWidget */
extern void get_unix_user_name();

doc_info *DocList;       /* Head of documents list */
short *clip1fromShorts();
static long DocId = 200;        /* "current" doc id counter */


#if ELFRT_ONLY /* ELFRT VERSION { -------------------------------*/
Atom AxXA_LENGTH;
Atom AxXA_CLIPBOARD;
Atom AxXA_PRIMARY;
Atom AxXA_GEGE_FILE;
Atom AxXA_GE_FILE;
Atom AxXA_ELF_TABLE;
Atom AxXA_GE_PAGE_FILE;
Atom AxXA_INSET;
Atom AxXA_GE_OLIN_FILE;
int AxSelectionLength = 0;
extern DocInfoType XferDoc; /* transfer bin document wrapper */
void AxClipboardInitialize()
{
    AxXA_LENGTH = XInternAtom(Dpy,"LENGTH", False);
    AxXA_CLIPBOARD = XInternAtom(Dpy, "CLIPBOARD", False);
    AxXA_PRIMARY = XInternAtom(Dpy, "PRIMARY", False);
    AxXA_GEGE_FILE = None;
    AxXA_GE_FILE = None;
    AxXA_ELF_TABLE = None;
    AxXA_GE_PAGE_FILE = None;
    AxXA_INSET = None;
    AxXA_GE_OLIN_FILE = None;
	/* reset the xferdoc info */
    memset(&XferDoc,0, sizeof(DocInfoType));
}

binFromAtom(selection)
Atom selection;
{
    return((selection == XA_PRIMARY) ? XFER_BIN_SELECTION : XFER_BIN_CLIPBOARD);
}
#else /* END ELFRT VERSION } { START REAL VERSION ---------------------------*/


#define WINXLOG 0 /* stdout logging */

/* a fake asterx window id for the dummy window, which holds the exit clipboard
and can hold the selection. */
#define AX_ROOT_WINDOW_ID -21

extern unsigned long TimeOfButtonPress; /* event time of button down */
typeErr ErrnoErr();
extern Window HostWinFromCurWin();
extern DocInfoType XferDoc; /* transfer bin document wrapper */
extern Window AxDummyWin;
extern data_ptr ElfRetData;

/* maximum amount of data for single X transfer */
#if 0 /* for testing -chris 28 jun 93 */
#define MAX_XSTR 200 
#else
#define MAX_XSTR XMaxRequestSize(Dpy)
#endif

int AxSelectionLength = 0;  /* num bytes in selection - used by OL apps */
static int ConvertBufSize = 320000;

#define MAX_ATOM_TYPES 17

/* the information about selections and requests for selections and
on-going incremental transfers is kept in SelectionInfo and
SelectionRequestInfo structures, for each selection type.
So that the same process can be
handle both sides of a transfer thru the window system, each of these
structures is defined for both of the following cases:
    SelectionInfo[type,HERE] -about the ax task that has the selection.
    SelectionInfo[type,THERE] - about the external window that holds the
                                selection and its selection.
    SelectionRequestInfo[type,HERE] - about the ax task that wants data
    SelectionRequestInfo[type,THERE] - about the external window that wants
                                    data.

When transfers that are between ax tasks in the same process are sent through
the window system then the 4 cases are distinct and a single process needs
to know about them all. When transfers between ax tasks are handled internally
then we never simultaneously need HERE and THERE info. So we set
HERE = THERE = 0.
*/

#if AXMS    /* all transfers, even local ones, are thru window system */
#define HERE 0
#define THERE 1
static void SetupWinClipboardForExit(int);

#else       /* local transfers handled locally */
#define HERE 0
#define THERE 0
#endif


/* information about the selection */
typedef struct {
    int AsterxWinId; /* of selection holder, or zero */
    int dataTypeList[MAX_ATOM_TYPES];/* types of data that holder can produce */
    int dataTypeCount; /* count of types */
    unsigned long selectionTime; /* time selection was made */
    XSelectionEvent selection; /* when selection is coming from X, this is the
                                associated selection event */
} SelectionInfo;

/* For table that tells what conversions we can do when root window owns
the clipboard, or when material can not otherwise be obtained */
#define CANCONVERT_SIZE 6 /* see AxClipboardInitialize */
static struct CanConvert {
    Atom from;
    Atom to;
    int (*func)();
} CanConvert[CANCONVERT_SIZE];

/* information about request for selection contents */
typedef struct {
    int requestingPid; /* application that requested selection */
    int AsterxWinId; /* of selection requester, or zero */
    int dataTypeList[MAX_ATOM_TYPES];/* types of data that can be accepted by requestor */
    struct CanConvert *convertTo[MAX_ATOM_TYPES]; /* for each in above array,
    the corresponding entry here is NULL, if requestor can accept data directly,
    or points an entry in the CanConvert table which we use to produce what
    the requestor wants. */
    int dataTypeCount; /* count of types */
    int dataTypeRequested;
            /* index into dataTypeList for last type requested from an X app. */
    XSelectionRequestEvent request; /* when request came from X, this is it. */
    XSelectionEvent selection; /* when selection is coming from X, this is the
                        associated selection event. Needed for incremental
                        transfer. */
    int incrementalXfer; /* 0 when not in process of incremental transfer */
#define INCR_XFER_OFF 0
#define INCR_XFER_IN 1
#define INCR_XFER_OUT 2
    int increment; /* max amount of data to send at one time, for incremental
                    transfer */
    char incrBuffer[STRPATH+1]; /* incremental xfer buffer,
                                clipboard file object */
    long filepos; /* current position during incremental xfer. */
    long priorEventMask; /* event mask before we started to look for certain
                        events during incrementalXfer */
} SelectionRequestInfo;

#include <winxfer.p>  /* prototypes */

static int convertAxObject();

/* first dimension is selection type. second dimension is HERE or THERE */
static SelectionInfo XferSelectionInfo[2][2];
static SelectionRequestInfo XferSelectionRequestInfo[2][2];

static SelectionInfo *getSelectionInfo AXP_((int selectionType, int hOt));
static SelectionRequestInfo *getSelectionRequestInfo AXP_((int selectionType, int hOt));
static char *exitClipboardFileObject AXP_((Atom selType, int dataType));
static Atom propertyForSelectionTransfer AXP_((Atom selType));
static void addConversionPossibilitiesToRequest();
static void clearExitBoard ();
Atom selectionFromBinType();
static Atom atomFromDoctype();
char *StrFromDataPtr();
static int allocError; /* set if BadAlloc error occurs */
static void *ElfXferIn();
static void *getSelectionString();
static int waitForSelection();
static void sendXferMessage();
static int getSelectionIncremental();
static int sendSelectionIncremental();
static long getEventMask();
static Window ClpXWinFromId ();

extern int ThimXErrorCode;
extern data_ptr AxReadDataFromFileObject();
extern Window XWinFromATMWin();
extern char *GetProfileByStr();

#if WINXLOG
extern char *XGetAtomName();
#endif

/* used in xferDoc file object names */
#define SELECTION_CODE(selType) ((selType == XA_CLIPBOARD) \
                        ? NXLT("C") \
                        : (selType == XA_PRIMARY) \
                                ? NXLT("P") \
                                : NXLT("X"))

Atom AxXA_LENGTH;
Atom AxXA_WP_FILE;
Atom AxXA_GE_FILE;
Atom AxXA_GEGE_FILE;
Atom AxXA_SS_INSET_FILE;
Atom AxXA_SS_FILE;
Atom AxXA_EQN_FILE;
Atom AxXA_ELF_DATA;
Atom AxXA_QUERY_DATA;
Atom AxXA_ELF_TABLE;
Atom AxXA_INSET;
Atom AxXFER_PRIMARY;
Atom AxXFER_CLIPBOARD;
Atom AxINCR;
Atom AxTIMESTAMP;
Atom AxTARGETS;
Atom AxMULTIPLE;
Atom AxXA_CLIPBOARD;
Atom AxXA_PRIMARY;
Atom AxXA_BUILDER_OBJECT;
Atom AxXA_GE_PAGE_FILE;
Atom AxXA_GE_OLIN_FILE;

/* used for allocating arrays of datatypes used as args to functions */
#define MAX_DATA_TYPES 64

#define WAITFORX 60
/* standard time (sec) to wait for an X response, when we need to wait. */

extern int JavaMode;

/* : AXInternAtom - wrapper for XInternAtom
 */
static Atom AXInternAtom(dpy, atomstring)
Display *dpy;
char *atomstring;
{
#if AX_X11
	if (JavaMode) {
		/* modify the atomstring by appending the username.
		 * This is to keep action partitioned according to
		 * users using the same anyware server.
		 */
		int namelen;
		char *fullname;
		char username[STRPATH];
		Atom aa;
		
		get_unix_user_name(username, STRPATH);
		namelen = strlen(atomstring) + strlen(username) + 3;
		fullname = malloc(namelen);
		strcpy(fullname, atomstring);
		strcat(fullname, ".");
		strcat(fullname, username);
		aa = XInternAtom(dpy, fullname, False);
		free(fullname);
		return(aa);
	}
	else {
		return(XInternAtom(dpy, atomstring, False));
	}
#else
	return(XInternAtom(dpy, atomstring, False));
#endif
}

/*************************************************************************
function: AxClipboardInitialize - initialize for inter-window transfer
*/
void AxClipboardInitialize()
{
    int ix;
    char *tprof;

    AxXA_LENGTH = AXInternAtom(Dpy,"LENGTH");
    AxXA_WP_FILE = AXInternAtom(Dpy,"_AX_WP_FILE");
    AxXA_GE_FILE = AXInternAtom(Dpy,"_AX_GR_FILE");
    AxXA_GEGE_FILE = AXInternAtom(Dpy,"_AX_GRGR_FILE");
    AxXA_SS_FILE = AXInternAtom(Dpy,"_AX_SS_FILE");
    AxXA_SS_INSET_FILE = AXInternAtom(Dpy,"_AX_SS_INSET_FILE");
    AxXA_EQN_FILE = AXInternAtom(Dpy,"_AX_EQN_FILE");
    AxXA_ELF_DATA = AXInternAtom(Dpy,"_AX_ELF_DATA");
    AxXA_QUERY_DATA = AXInternAtom(Dpy,"_AX_QUERY_DATA");
    AxXA_ELF_TABLE = AXInternAtom(Dpy,"_AX_ELF_TABLE");
    AxXA_INSET = AXInternAtom(Dpy,"_AX_INSET");
    AxXFER_PRIMARY = AXInternAtom(Dpy, "AX_PRIMARY_XFER");
    AxXFER_CLIPBOARD = AXInternAtom(Dpy, "AX_CLIPBOARD_XFER");
    AxINCR = AXInternAtom(Dpy,"INCR");
    AxTIMESTAMP = AXInternAtom(Dpy, "TIMESTAMP");
    AxTARGETS = AXInternAtom(Dpy, "TARGETS");
    AxMULTIPLE = AXInternAtom(Dpy, "MULTIPLE");
    AxXA_CLIPBOARD = AXInternAtom(Dpy, "CLIPBOARD");
	if (JavaMode)
		AxXA_PRIMARY = AXInternAtom(Dpy, "PRIMARY");
	else
		AxXA_PRIMARY = (Atom)1; /* temp Hack!!0, it should be XA_PRIMARY
								   and all references to XA_PRIMARY
								   should be changed to AxXA_PRIMARY */
    AxXA_BUILDER_OBJECT = AXInternAtom(Dpy, "_AX_BUILDER_OBJECT");
    AxXA_GE_PAGE_FILE = AXInternAtom(Dpy, "_AX_GR_PAGE_FILE");

/* build table of conversions that we can do. Ordered by "richness" of the
"from" data type. */
/* WARNING: if this table is expanded, update CANCONVERT_SIZE above! */
    ix = 0;
    CanConvert[ix].from= AxXA_ELF_TABLE;
                CanConvert[ix].func = convertAxObject;
                CanConvert[ix++].to= AxXA_QUERY_DATA;
    CanConvert[ix].from= AxXA_ELF_TABLE;
                CanConvert[ix].func = convertAxObject;
                CanConvert[ix++].to= AxXA_ELF_DATA;
    CanConvert[ix].from= AxXA_ELF_DATA;
                CanConvert[ix].func = convertAxObject;
                CanConvert[ix++].to= AxXA_ELF_TABLE;
    CanConvert[ix].from= AxXA_ELF_DATA;
                CanConvert[ix].func = convertAxObject;
                CanConvert[ix++].to= XA_STRING;
    CanConvert[ix].from= AxXA_QUERY_DATA;
                CanConvert[ix].func = convertAxObject;
                CanConvert[ix++].to= AxXA_ELF_TABLE;
    CanConvert[ix].from= XA_STRING;
                CanConvert[ix].func = convertAxObject;
                CanConvert[ix++].to= AxXA_ELF_DATA;

    /* allow the default conversion buffer size to be overridden
     * via profile.  This size affects whether to do large
     * transfers incrementally
     */
    tprof = GetProfileByStr("axPasteIncrementSize");
    if (tprof && *tprof) {
        ConvertBufSize = atoi(tprof);
		ConvertBufSize = MAX(ConvertBufSize,1000);
    }
	ConvertBufSize = MIN(XMaxRequestSize(Dpy),ConvertBufSize);
}

/***************************************************************************
function: AxClipboardToFile - paste clipboard to a file. returns error code,
*/
int AxClipboardToFile(binType, file, dataTypeList)
int binType; /* XFER_BIN_CLIPBOARD or XFER_BIN_SELECTION */
mbchar *file;
int *dataTypeList;  /* acceptable data types, in preference order, 
                    terminated by None. */
{
    int err = AOK ;  /* initialize err to AOK */
    int gotData;
    xfer_data_convert_msg_t info;

#if WINXLOG
printf("AxClipboardToFile: binType=%d file=%d dataTypes",binType,file);
printDataTypeList(dataTypeList);
#endif

    if (!AxGetSelection(binType, AX_ROOT_WINDOW_ID, dataTypeList, &info) )
        return -1; /* can't get this type of data */

    if (info.type == None) {
        /* request was sent to X, wait for response here. */
        gotData = waitForSelection(WAITFORX,&info);
    }
    else {
        gotData = TRUE;
    }

    if (gotData) {
        err = FSWriteFileObjAsFile(file, AxChmodValueFromCodes(file, 2, 2),
                                &XferDoc, info.pathname);
    }

#if WINXLOG
printf("AxClipboardToFile returns %d\n",err);
#endif
             
    return err;
}

#if WINXLOG
printDataTypeList(dataTypeList)
int *dataTypeList;
{
    int ix;
    for (ix=0; dataTypeList[ix]!=None; ix++)
        printf(" %s",XGetAtomName(Dpy,dataTypeList[ix]));
    printf("\n");
}
#endif

/***************************************************************************
function: AxFileToClipboard - to copy an Asterx, or other, file to the
clipboard.
*/
AxFileToClipboard(binType, file, docType)
int binType; /* XFER_BIN_CLIPBOARD or XFER_BIN_SELECTION */
mbchar *file;
int docType; /* if doctype is AX_DOC_TYPE_UNKNOWN, use file recognition */
{
    xfer_data_convert_msg_t info;
    Atom dataType;
    int err;
    char tmpfile[STRPATH+1];
    int filtered = FALSE;

#if WINXLOG
printf("AxFileToClipboard: binType=%d file=%s docType=%d\n",binType,file,docType);
#endif

    /* what kind of file is it? */
    if (docType == AX_DOC_TYPE_UNKNOWN)
        docType = RecognizeFile(file);

    dataType = atomFromDoctype(docType);

    AxExitClipboard(selectionFromBinType(binType), dataType, &info);
    filtered = FSFilterDocToCurRev(file,tmpfile);
    err = FSReadFileIntoFileObj(tmpfile, &XferDoc, info.pathname);
    if (filtered)
        AxDeleteTempfile(tmpfile);

#if WINXLOG
printf("AxFileToClipboard returns %d\n",err);
#endif

    return err;
}

/***************************************************************************
function: AxStringToClipboard - put string in central clipboard
*/
void AxStringToClipboard(binType, str)
int binType; /* XFER_BIN_CLIPBOARD or XFER_BIN_SELECTION */
mbchar *str;
{
    xfer_data_convert_msg_t info;
    int fd;

    AxExitClipboard(selectionFromBinType(binType), XA_STRING, &info);
    fd = (int) OpenFileObjFmode(&XferDoc,info.pathname, "w");
    FSfputs(str,fd);
    CloseFileObject(fd);
}

/***************************************************************************
function: AxClipboardToString - get string from clipboard.
    caller must free string.
*/
mbchar *AxClipboardToString(binType)
int binType; /* XFER_BIN_CLIPBOARD or XFER_BIN_SELECTION */
{
    int dataTypes;
    mbchar *str;
    int strsize;
    int fd, gotData;
    int dataTypeList[6];
    xfer_data_convert_msg_t info;

#if WINXLOG
printf("AxClipboardToString: binType=%d\n",binType);
#endif

    str = 0;
    dataTypeList[0] = XA_STRING;
    dataTypeList[1] = None;
    gotData = FALSE;
    if (AxGetSelection(binType, AX_ROOT_WINDOW_ID, dataTypeList, &info) ) {
        gotData = TRUE;
        if (info.type == None) {
            gotData = waitForSelection(WAITFORX,&info);
                                                /* wait WAITFORX seconds */
        }
    }
    if (gotData) {
        str = (mbchar *)getSelectionString(&info);
    }

#if WINXLOG
if (str)
printf("AxClipboardToString returns %s\n",str);
else
printf("AxClipboardToString returns NULL\n");
#endif

    return str;
}

/***************************************************************************
function: getSelectionString - get selection as string. caller muust free
the string. Returns NULL on error;
*/

static void *getSelectionString(infoP)
xfer_data_convert_msg_t *infoP;
{
    mbchar *str;
    long strsize;
    int fd;
    
    fd = (int)OpenFileObjFmode(&XferDoc,infoP->pathname, "r");
    if ( !fd )
        return 0;

    strsize = FSfsize(fd)+1;
    str = malloc(strsize);
    if (str) {
        FSfgets(str,strsize,fd);
        str[strsize]=0;
    }
    CloseFileObject(fd);
    return (void *)str;
}


/*************************************************************************
function: ClpXWinFromId - Given ATM id or Elf id, return X window
*/
static Window ClpXWinFromId (winId)
int winId;
{
    if (winId == AX_ROOT_WINDOW_ID)
        return(AxDummyWin);
    else
        return(XWinFromATMWin(winId));
}

/*************************************************************************
function: ClpIdFromXWin - Given X window, return ATM or ELF id
*/
static int ClpIdFromXWin (winId)
Window winId;
{
    int ret;

    if ( winId == None )
       ret = 0;
    else if ( (winId == AxDummyWin) || winId == (AX_ROOT_WINDOW_ID) )
       ret = AX_ROOT_WINDOW_ID;
    else
       ret = ATMWindowFromXWindow(winId);
  
    return( ret );
}

/*************************************************************************
function: AxRegisterSelectionAndConverters - Grab selection ownership and
indicate what types we can convert it to.
*/
AxRegisterSelectionAndConverters (whichBin,winId,dataTypeList)
int whichBin; /* XFER_BIN_CLIPBOARD or XFER_BIN_SELECTION */
int winId;  /* ATM window id */
int *dataTypeList;  /* producible data types, in preference order, 
                    terminated by None. */
{
    Atom selType;
    SelectionInfo *siP;
    int ix;
    Window win;

#if WINXLOG
printf(NXLT("AxRegisterSelectionAndConverters: whichBin %d winId %d dataTypes"),
whichBin,winId);
printDataTypeList(dataTypeList);
#endif

#if AXMS
    /* concept of "selection" doesn't exist on Windows.   -donato */
    if ( whichBin == XFER_BIN_SELECTION )
        return FALSE;
#endif
	
	if (JavaMode && (XFER_BIN_CLIPBOARD == whichBin)) {
		/* comm to java client regarding clipboard.
		 */
		data_ptr msg, submsg, wrapper, sysv;
		mbchar *svar_value = (mbchar *) NULL;

		ElfbGetSysVar(NXLT("ax_jclient_clipboard"));
		sysv = ElfRetData;
		if (sysv) {
			svar_value = AxStrFromDataPtr(sysv);
		}
		if (svar_value && !strcmp(svar_value, NXLT("client"))) {
			/* java client-side has just acquired clipboard ownership;
			 * don't send set-selection-owner to client, instead, clear var.
			 */
			ElfbSetSysVar(NXLT("ax_jclient_clipboard"),
						  AxMakeStrData(-1, ""));
		}
		else {
			/* do: 	java_notify$({JV#CLIPBOARD#EVENT,
			 * {EV_SET_SELECTION_OWNER_}})
			 */
			wrapper = AxMakeArray(1);
			msg = AxMakeArray(2);
			submsg = AxMakeArray(1);
			msg = AxAddIntToArray(msg, 0, JV_CLIPBOARD_EVENT);
			submsg = AxAddIntToArray(submsg, 0, EV_SET_SELECTION_OWNER_);
			msg = AxAddArrayToArray(msg, 1, submsg);
			wrapper = AxAddArrayToArray(wrapper, 0, msg);
			AxfJavaNotify(wrapper);
			AxFreeData(wrapper);
		}
	}

    selType = selectionFromBinType(whichBin);

    if (!winId) winId = AX_ROOT_WINDOW_ID; 

    /* Make this the selection owner, notify other X-windows */
    if ( !AxIAmSelectionOwner (selType,winId) ) {
        win = ClpXWinFromId(winId);
        AxClearSelection(selType,XGetSelectionOwner(Dpy,selType));
        XSetSelectionOwner(Dpy,selType,win,TimeOfButtonPress);
        XSync(Dpy,FALSE);   /* just to be safe! */
    }

#if AXMS
    /* ideally XSetSelectionOwner() should take ownership of
       the clipboard, but at that time we also need to tell Windows
       the formats we know about. Thus this separate call.
    */
    SetWinClipboardOwner(winId, dataTypeList);
#endif

    siP = getSelectionInfo(selType, HERE);
    siP->AsterxWinId = winId;
    for(ix=0; ix<MAX_ATOM_TYPES && dataTypeList[ix] != None; ix++) {
        siP->dataTypeList[ix] = dataTypeList[ix];
    }
    siP->dataTypeCount = ix;
    siP->selectionTime = TimeOfButtonPress;

#if WINXLOG
printf("AxRegisterSelectionAndConverters done.\n");
#endif
}

/*************************************************************************
function: clearExitBoard - Clear the data stored by prev apps's exit
*/
static void clearExitBoard (selType)
Atom selType;
{
    char *name, *ScanForNthFileObject();
    int ix;
    char *selectionCode;

    selectionCode = SELECTION_CODE(selType);

    /* count file objects */
    ix = 0;
    while(ScanForNthFileObject(&XferDoc, ix)) ix++;
    ix--;

    /* delete all file objects with name containing selection code */
    for (; ix>=0; ix--) {
        name = ScanForNthFileObject(&XferDoc, ix);
        if (name && MBstreq(selectionCode,&name[MBstrlen(name)-1])) {
            DeleteFileObject(&XferDoc,name);
        }
    }
}

/*************************************************************************
function: AxIsTypeInXferList - Return TRUE if passed type in OUR convert list
*/
int AxIsTypeInXferList (int target)
{
    SelectionInfo *siP;
    int ix;

    siP = getSelectionInfo(XFER_BIN_CLIPBOARD, HERE);
    for ( ix=0; ix < siP->dataTypeCount; ix++ )
		if ( siP->dataTypeList[ix] == target )
			return(TRUE);
	return(FALSE);
}

/*************************************************************************
function: AxIAmSelectionOwner - Return whether this app supplying the data
*/
AxIAmSelectionOwner (selType,winId)
Atom selType;
int winId;  /* ATM window id */
{
    int ret;
    SelectionInfo *siP;

    siP = getSelectionInfo(selType,HERE);
    ret = (winId != 0 && winId == siP->AsterxWinId);

#if WINXLOG
printf("AxIAmSelectionOwner: selType %d winId %d returns %d\n",selType,winId,ret);
#endif
    return ret;
}

/*************************************************************************
function: AxIOwnSelection - Return whether this app supplying the data
*   same as AxIAMSelectionOwner but takes a window list instead of a single
*   window.
*/
AxIOwnSelection (selType, winList)
Atom selType;
Window winList[];   /* xwindow list terminated by 'None' */
{
    Window win;
    int ix;
    /* get the selection owner */
    win =  XGetSelectionOwner(Dpy,selType);
    for(ix=0; winList[ix] != None; ix++)
        {
        if(win == winList[ix])
            return(TRUE);
        }
    return(FALSE);
}



/*************************************************************************
function: AxClearSelection - Tell selecting app to clear its selections
*/
AxClearSelection (selType,win)
Atom selType;
Window win;
{   
    int wid;
    SelectionInfo *siP;

    siP = getSelectionInfo(selType,HERE);
    siP->AsterxWinId = 0;
    siP->dataTypeCount = 0;

    if (win == AxDummyWin) {
        clearExitBoard(selType);
    }

    else if (win) {

        wid = ClpIdFromXWin(win);

        if ( !wid ) {
            editboxClearSelection(win, selType);
        }

        else { /* send signal to current selection holder */
            xfer_data_convert_msg_t info;

            memset((char *)&info,0,sizeof(info));
            info.code = XFER_CODE_CLEAR_SELECTION;
            info.binType = binFromAtom(selType);
            sendXferMessage(getPidFromATMWindow(wid), &info);
        }
    }
}


/***************************************************************************
function: AxGetSelection - Get clipoard or selection from another Asterx
 task or X window.
*/
AxGetSelection(whichBin,myWinId,dataTypeList,infoP)
int whichBin; /* XFER_BIN_CLIPBOARD or XFER_BIN_SELECTION */
int myWinId;    /* requesters Asterx window */
int *dataTypeList;  /* acceptable data types, in preference order, terminated
                    by None. */
xfer_data_convert_msg_t *infoP;

/* If selection holder is an Asterx task, and it has registered the
selection, then we compare the list of dataTypes that it can produce
with the list of types requested. Find the most desirable and ask
selection holder to produce it.

It the selection holder is an X window, send it a message and handle the 
response in GetXSelectionToXferBin.

Returns TRUE if (1) data is provided in buffer named in infoP, or
                    (infoP->type is set to type of data returned)
                (2) request has been made to X for the data
                    (infoP->type is set to None)
        FALSE if caller can not get any of requested types of data.
*/
{
    Atom selectionType;
    int ix, jx, ret;
    SelectionInfo *siP;
    SelectionRequestInfo *sriP;
    Atom non_applix_datatype;

#if WINXLOG
printf("AxGetSelection: whichBin %d winId %d dataTypes",whichBin,myWinId);
printDataTypeList(dataTypeList);
#endif

    ret = FALSE;
    memset((char *)infoP,0,sizeof(xfer_data_convert_msg_t)); /* zero structure */
    infoP->binType = whichBin;
    infoP->type = None;
    infoP->binDoc = (char*)(&XferDoc);
    infoP->code = XFER_CODE_PASTE; /* in case we respond with a message */
    selectionType = selectionFromBinType(whichBin);

#if AXMS
    /* on Windows, HERE and THERE are different; so since the data was *
     * placed HERE, we must retrieve it from HERE.       -donato       */
    siP = getSelectionInfo(selectionType, HERE);
#else
    siP = getSelectionInfo(selectionType,THERE);       
#endif

    /* save info about this request */
    sriP = getSelectionRequestInfo(selectionType,HERE);
    sriP->AsterxWinId = myWinId;
    /* list of datatypes requested */
    for(ix=0; ix<MAX_ATOM_TYPES && dataTypeList[ix] != None; ix++) {
        sriP->dataTypeList[ix] = dataTypeList[ix];
        sriP->convertTo[ix] = 0;
    }
    sriP->dataTypeCount = ix;

    addConversionPossibilitiesToRequest(sriP);


/* get from asterx task */
/* on Windows, everything will be "external" except when the root window *
 * has ownership, thus the separate check for ax_root_window_id. -donato */
    if ( ((HERE == THERE) && siP->AsterxWinId) ||
            (siP->AsterxWinId == AX_ROOT_WINDOW_ID) ) {
        for(ix=0; ix<sriP->dataTypeCount; ix++) {
            for (jx=0; jx<siP->dataTypeCount; jx++) {
                if (sriP->dataTypeList[ix] == siP->dataTypeList[jx]) {

                    /* we have a requested type that the holder can produce.
                    Get that data from holder */
                    sriP->dataTypeRequested = ix;
                    sriP->requestingPid = AxInTask; /* reply should come to
                                                    this task */
                    ret = convertAxSelection(siP, sriP, infoP);
                    if (ret) return ret;
                }
            }
        }
    }

/* get from external X window. Request now. Obtain later. */
    else {
        sriP->dataTypeRequested = 0; /* request in preference order */
        /* figure task upon whose behalf this is happening.
        This is the task that will get the paste message.
        One hopes that this is usually THIMpid, but
        if these central support services are getting
        the message on their own, then use AxInTask */
        sriP->requestingPid = (sriP->AsterxWinId == AX_ROOT_WINDOW_ID)
                                ? AxInTask 
                                : THIMpid;
        ret = requestSelectionFromX(selectionType,
                                myWinId,
                                sriP->dataTypeList[sriP->dataTypeRequested]);
    }

#if WINXLOG
printf("AxGetSelection returns %d info.type=%d\n",ret, infoP->type);
#endif

    return ret;
}

/*************************************************************************
function: addConversionPossibilitiesToRequest
*/
static void addConversionPossibilitiesToRequest(sriP)
SelectionRequestInfo *sriP;
{
    int ix,jx,kx;

    /* Add to list those datatypes that we can convert into the requested
    types. These are added at end of list, lowest priority. If requestor is
    aware of conversions we can do, she can construct the list accordingly.
     */
    kx = sriP->dataTypeCount;
    for (ix=0; ix<sriP->dataTypeCount; ix++) {
        if (kx == MAX_ATOM_TYPES) break;
        for (jx = 0; jx < CANCONVERT_SIZE; jx++) {
            if (kx == MAX_ATOM_TYPES) break;
            if (CanConvert[jx].to == sriP->dataTypeList[ix]) {
                sriP->dataTypeList[kx] = CanConvert[jx].from;
                sriP->convertTo[kx] = &(CanConvert[jx]);
                kx++;
            }
        }
    }
    sriP->dataTypeCount = kx;
}

/*************************************************************************
function: convertAxSelection - Attempt to convert selection to requested type
    returns TRUE if conversion succeeds. Either works in clipboard objects,
    or sends message to Applixware task.
*/
static int convertAxSelection (siP,sriP,infoP)
SelectionInfo *siP; /* in */
SelectionRequestInfo *sriP; /* in */
xfer_data_convert_msg_t *infoP; /* in: code, binDoc, binType filled in,
                                out: type and pathname set. */
{   
    Atom selType;
    int ix, wid;
    int fd;
    int pid;        /* TaskInfo index */
    xfer_data_convert_msg_t fromInfo; /* convert from this */
    int err;
    data_ptr aP;
    data_ptr filenameP;
    char tmpfile[STRPATH+1];
    struct CanConvert *convP;
    int converted = FALSE;

/* If the data is being supplied by an Applix application,
    send off a convert message to app, pend until app supplies data.
*/

    selType = selectionFromBinType(infoP->binType);
    /* construct file object name for root window object of requested type */
    infoP->type = sriP->dataTypeList[sriP->dataTypeRequested];
    infoP->docType = docTypeFromAtom(infoP->type);
    strcpy(infoP->pathname,exitClipboardFileObject(selType,infoP->type));

    wid = siP->AsterxWinId;
    if ( wid == AX_ROOT_WINDOW_ID ) {

        /* If this file object exists we return it, we return it, or
        convert it into the type that the caller really wants. */
        if( TestForFileObject(&XferDoc, infoP->pathname)) {
            converted = TRUE; /* we already have it as a clipboard object */
        }
    }

    else if (wid) { /* get stuff from a running application */
        pid = getPidFromATMWindow(wid);
        fromInfo = *infoP;
        fromInfo.code = XFER_CODE_COPY;
        IPCXsendMes(IPC_SRT,pid,&fromInfo,sizeof(fromInfo),APWINXFER,
                                            sriP->requestingPid);
        if ( IPC_OK == AxGetResponse(pid,infoP,sizeof(*infoP)) ) { /* wait */
            converted = TRUE;
        }
    }

    /* if we have obtained data into the clipboard object, we either
    have what caller wants, or we can convert what we have obtained
    into what caller wants. */
    if (converted) {
        convP = sriP->convertTo[sriP->dataTypeRequested];
        if (convP) {
            fromInfo = *infoP;
            strcpy(infoP->pathname,
                exitClipboardFileObject(selType,convP->to));
            infoP->type = convP->to;
            infoP->docType = docTypeFromAtom(infoP->type);
            converted = (*convP->func)(&fromInfo, infoP);
        }
    }

    /* return with indication of success or failure */
    if (!converted)
        infoP->type = None;
    return(converted);
}


/*************************************************************************
function: convertAxObject - convert data on exit clipboard 
Returns TRUE for success. fromType must be on clipboard.
*/
static int convertAxObject(fromInfoP, toInfoP)
xfer_data_convert_msg_t *fromInfoP; /* in: this material exists on clipboard */
xfer_data_convert_msg_t *toInfoP; /* in: specifies type and pathname
                                    for result of conversion. */
{
    char tmpfile[STRPATH+1];
    int ix,fd,err;
    int success = TRUE; /* assume success */
    data_ptr filenameP, dataP, aP;

/* ELF_DATA requested **********************************************/
    if (toInfoP->type == AxXA_ELF_DATA) {

/* convert STRING to ELF_DATA */
        if ( fromInfoP->type == XA_STRING ) {

            /* convert strings to elf array, as if pasting in ELF */
            dataP = (data_ptr) ElfXferIn(fromInfoP);

            /* copy elf array to file object */
            success = copyElfDataToClipboard(dataP, toInfoP);

            AxFreeData(dataP);
        }

/* convert ELF_TABLE  to ELF_DATA */
        if ( fromInfoP->type == AxXA_ELF_TABLE) {

            aP = (data_ptr) ElfXferIn(fromInfoP);
            dataP = AxArrayElement(aP,0);
            
            /* write elf array to file object */
            success = copyElfDataToClipboard(dataP, toInfoP);

            AxFreeData(aP);
        }
    }

/* ASCII strings requested ****************************************/
    else if (toInfoP->type == XA_STRING) {

/* convert ELF_DATA to STRING */
        if (fromInfoP->type == AxXA_ELF_DATA) {
            data_ptr elfDataP, aP;
            data_ptr bP;
            int size, jx, jsize;
            mbchar *AxStrPtrFromArray(), *str;

            aP = AxReadDataFromFileObject(THIMpid,&XferDoc,fromInfoP->pathname);

            /* convert to ascii string file object */
            fd = (int)OpenFileObjFmode(&XferDoc,toInfoP->pathname, "w");
            if ( !fd ) {
                success = FALSE;
            }
            else if (AxIsArray(aP)) {
                size = AxArraySize(aP);
                for ( ix = 0; ix < size; ix++ ) {
                    if (AxIsArray(AxArrayElement(aP,ix))) {
                        bP = AxArrayFromArray(aP,ix);
                        jsize = AxArraySize(bP);
                        for (jx=0; jx < jsize; jx++) {
                            str = AxStrPtrFromArray(bP,jx);
                            FSfputs(str,fd);
                            if ( (jx != (jsize-1)) || (strlen(str) != 0) )
                                FSfputc('\n',fd);
                        }
                    }
                    else {
                        str = AxStrPtrFromArray(aP,ix);
                        FSfputs(str,fd);
                        if ( (ix != (size-1)) || (strlen(str) != 0) )
                            FSfputc('\n',fd);
                    }
                }
                CloseFileObject(fd);
            }
            else {
                str = AxStrPtrFromDataPtr(aP);
                FSfputs(str,fd);
            }

            AxFreeData(aP);
        }
    }

/* QUERY doc requested *********************************************/
    else if (toInfoP->type == AxXA_QUERY_DATA) {

/* convert ELF_TABLE to QUERY_DATA (for version 3 Applixware) */
        if ( fromInfoP->type == AxXA_ELF_TABLE) {
            data_ptr resultP;

            /* QUERY_DATA is a sql_state@ with only savedData and headers set */
            dataP = AxReadDataFromFileObject(THIMpid,&XferDoc,fromInfoP->pathname);
            err = ElfCallByName(NXLT("dbase_elf_table_to_sql_state$"),
                            &resultP,
                            1,
                            dataP);
            success = !err;
            if (success) 
                success = copyElfDataToClipboard(resultP, toInfoP);
            AxFreeData(resultP);
            AxFreeData(dataP);
        }
    }

/* ELF_TABLE requested *********************************************/
    else if (toInfoP->type == AxXA_ELF_TABLE) {

/* convert QUERY_DATA to ELF_TABLE (copy from version 3 Applixware) */
        if ( fromInfoP->type == AxXA_QUERY_DATA) {
            data_ptr resultP;

            /* QUERY_DATA is a sql_state@ with only savedData and headers set */
            dataP = AxReadDataFromFileObject(THIMpid,&XferDoc,fromInfoP->pathname);
            err = ElfCallByName(NXLT("dbase_sql_state_to_elf_table$"),
                            &resultP,
                            1,
                            dataP);
            success = !err;
            if (success) 
                success = copyElfDataToClipboard(resultP, toInfoP);
            AxFreeData(resultP);
            AxFreeData(dataP);
        }
/* convert ELF_DATA to ELF_TABLE */
        else if ( fromInfoP->type == AxXA_ELF_DATA) {
            dataP = (data_ptr) ElfXferIn(fromInfoP);
            /* if this is 2-d data, then we can make ELF_TABLE */
            if (AxIsArray(dataP) && AxIsArray(AxArrayElement(dataP,0))) {
                aP = AxMakeArray(1);
                aP = AxAddArrayToArray(aP,0,dataP);
                /* write elf array to file object */
                success = copyElfDataToClipboard(aP, toInfoP);
                AxFreeData(aP);
            }
            else {
                success = FALSE;
                AxFreeData(dataP);
            }
        }
    }

    return success;
}

/*************************************************************************
function: copyElfDataToClipboard
*/
static int copyElfDataToClipboard(dataP,info)
data_ptr dataP;
xfer_data_convert_msg_t *info; /* in: destination file object */

{
    char tmpfile[STRPATH+1];
    data_ptr filenameP;
    int success = TRUE;
    int err;

    /* write elf array to file object */
    AxTempFile(NULL, tmpfile);
    filenameP = AxMakeStrData(-1, tmpfile);
    ElfbWriteDataFile(filenameP, dataP, (data_ptr)NULL);
    err = FSReadFileIntoFileObj(tmpfile, &XferDoc, info->pathname);
    if (err) 
        success = FALSE;
    AxDeleteTempfile(tmpfile);
    AxFreeData(filenameP);

    return success;
}

/*************************************************************************
function: requestSelectionFromX - Request X to convert selection
Note that this function always returns immediately.
Some time later, a SelectionNotify may be sent to the caller of
this routine.  The selection notify comes in as an APWINXFER message.
*/
static requestSelectionFromX (selType,winId,XtypeWanted)
Atom selType;
int winId;          /* window of requester */
Atom XtypeWanted;       /* desired type */
{
    int ix;
    Window requestor_win;
    Atom property;

/* Ask X program to convert selection.  If the X program does not go
    down, it will normally either convert to the type requested
    or convert to None (reject).
*/
    requestor_win = ClpXWinFromId(winId);
    if (requestor_win == None) return FALSE;

    property = propertyForSelectionTransfer(selType);
    XDeleteProperty(Dpy,requestor_win,property);    /* start clean */
    XConvertSelection(Dpy,selType,XtypeWanted,property,
                requestor_win,TimeOfButtonPress);
    XSync(Dpy,FALSE);   /* just to be safe! */
    return(TRUE);
}

/*************************************************************************
function: AxDoneConvert - Tell requester that conversion completed
*/
AxDoneConvert (infoP)
xfer_data_convert_msg_t *infoP;
{
    extern char ThimTempDir[];
    XSelectionEvent event;
    long nbytes;
	int created;
    int fd;
    SelectionRequestInfo *sriP;

#if WINXLOG
printf("AxDoneConvert: info.type=%s\n",XGetAtomName(Dpy,infoP->type));
#endif

    if (infoP->code != XFER_CODE_COPY) return; 

    sriP = getSelectionRequestInfo(selectionFromBinType(infoP->binType),THERE);

#if AXMS
    /* we may need to filter the data. */
    convert_to_non_applix_format(infoP, sriP->request.target);
#endif

/* If the selection is to another Thimble task, just unpend paster,
    letting it read the info from msgbuf */

    if ( sriP->requestingPid ) {
#if WINXLOG
printf(NXLT("sending APWINXFER message %s type %s to pid=%d\n"),
((infoP->code == XFER_CODE_COPY) ? NXLT("XFER_CODE_COPY")
: (infoP->code == XFER_CODE_PASTE) ? NXLT("XFER_CODE_PASTE")
: (infoP->code == XFER_CODE_CLEAR_SELECTION) ? NXLT("XFER_CODE_CLEAR_SELECTION")
: NXLT("UNKNOWN")),
XGetAtomName(Dpy,infoP->type),
sriP->requestingPid);
#endif
        IPCXsendMes(IPC_RES,sriP->requestingPid,infoP,sizeof(*infoP),APWINXFER,THIMpid);
        return;
    }

    else if (sriP->request.requestor) {

    /* Selection out to another X application. 
    Large amounts of data are sent incrementally */

        event.type = SelectionNotify;
        event.serial = sriP->request.serial;
        event.send_event = sriP->request.send_event;
        event.display = sriP->request.display;
        event.requestor = sriP->request.requestor;
        event.time = sriP->request.time;
        event.selection = sriP->request.selection;
        event.target = sriP->request.target;
        event.property = sriP->request.property;

        /* supply # of bytes in selection - requested by OpenLook apps */
        if (infoP->type == None || event.property == None) {
            event.property = None; /* no data */
        }
        else if (infoP->type == AxXA_LENGTH) {
            unsigned char *sel_nbytes;

            sel_nbytes = (unsigned char *)malloc(sizeof(int));
            bcopy((char *)&AxSelectionLength, (char *)sel_nbytes, sizeof(int));
            XChangeProperty(sriP->request.display, sriP->request.requestor, 
                                sriP->request.property, sriP->request.target, 
                                8, PropModeReplace, sel_nbytes, sizeof(int));
            free(sel_nbytes);
        }

        else {

            /* how much data is there to send */
            fd = OpenFileObject(&XferDoc,
                                infoP->pathname,
                                DFM_READ,
                                &created,
                                &nbytes);
            if (fd == -1 || created) { /* no data available */
                if (created) {
                    CloseFileObject(fd);
                    DeleteFileObject((DocInfoType *)(infoP->binDoc),
									infoP->pathname);
                }
                event.property = None; /* failure */
            }
#if AX_X11
    /* no incremental transfers on Windows. */
            /* if there is too much data, send it incrementally */
            else if (nbytes > MIN(MAX_XSTR,ConvertBufSize-2)) {
				Window sourceWin;
				SelectionInfo *siP;

                CloseFileObject(fd);
                /* initialiase for incremental transfer through property */
                /* save window, requested data type, and initial position
                in data object.
                FOR NOW, AND MAYBE FOR EVER, ONLY ONE REQUEST
                CAN BE PROCESSED AT A TIME, for each selection. */
                sriP->incrementalXfer = INCR_XFER_OUT;
                sriP->filepos = 0;
                MBstrcpy(sriP->incrBuffer, infoP->pathname);
                sriP->increment = MIN(MAX_XSTR,ConvertBufSize-2);

#if WINXLOG
printf("AxDoneConvert starts incremental transfer of %d bytes\n",nbytes);
#endif

                /* we want to see PropertyNotify events when this
                property on the requestor's window is changed. */
				siP =
				 getSelectionInfo(selectionFromBinType(infoP->binType), HERE);
				sourceWin = ClpXWinFromId(siP->AsterxWinId);
                sriP->priorEventMask = getEventMask(sourceWin);
                                    /* save to restore later */
#if WINXLOG
printf("XSelectInput window=%d PropertyChangeMask\n",sourceWin);
#endif
                XSelectInput(Dpy, sourceWin,
                            sriP->priorEventMask | PropertyChangeMask);
                /* indicate to requestor that data will be sent incrementally
                by sending the size of the data as a type INCR item */
                XChangeProperty(sriP->request.display, sriP->request.requestor,
                             sriP->request.property,
                             AxINCR,
                             32, PropModeReplace, (unsigned char *)&nbytes, 1);

            }
#endif
            else { /* send data now */
                char *conversionBuf;
#if AXMS
                /* on Windows, always supply entire selection in single shot */
                conversionBuf = malloc(nbytes);
                nbytes = DocSegRead(fd, conversionBuf, nbytes, -1);
#else
                conversionBuf = malloc(ConvertBufSize);
                nbytes = DocSegRead(fd,conversionBuf,ConvertBufSize-1,-1);
#endif
                CloseFileObject(fd);
                conversionBuf[nbytes] = '\0';

                ThimXErrorCode = Success;
                XChangeProperty(sriP->request.display, sriP->request.requestor,
                                     sriP->request.property,
                                     sriP->request.target, 8,
                                     PropModeReplace,
									 (unsigned char *)conversionBuf, nbytes);
                if (ThimXErrorCode == BadAlloc) {
                    /* delete property if alloc failed. */
                    XDeleteProperty(sriP->request.display,
                                sriP->request.requestor,sriP->request.property);
                    event.property = None; /* no data */
                }
                ThimXErrorCode = Success;
                free(conversionBuf);
            }
        }

#if WINXLOG 
printf(NXLT("XSendEvent SelectionNotify property=%s target=%s\n"),
XGetAtomName(Dpy,event.property),
XGetAtomName(Dpy,event.target));
#endif
        XSendEvent(event.display,event.requestor, True,0,(XEvent *)&event);
    }
}

/*************************************************************************
function: SupplySelection - This converts the transfer bin to X-data
Call up application to xfer out in desired form.
*/
SupplySelection(p)
XSelectionRequestEvent *p;
{
    int ix, wid;
    static XSelectionEvent event;
    xfer_data_convert_msg_t info;
    SelectionRequestInfo *sriP;
    SelectionInfo *siP;

    siP = getSelectionInfo(p->selection,HERE);
    sriP = getSelectionRequestInfo(p->selection,THERE);
    sriP->request = *p;

    /* if requestor has not supplied a property name, use target atom as
    property name */
    if (sriP->request.property == None)
        sriP->request.property = sriP->request.target;

    sriP->requestingPid = 0; /* from outside Asterx */
    sriP->AsterxWinId = 0;
    sriP->incrementalXfer = INCR_XFER_OFF;
    sriP->dataTypeCount = 1;
    sriP->dataTypeRequested = 0;
    sriP->dataTypeList[0] = sriP->request.target;
    addConversionPossibilitiesToRequest(sriP);

    event.property = None; /* signals rejection of request */

    if (!siP->AsterxWinId)
        goto done; /* we have no selection */

    /* file object for data */
    memset((char *)&info,0,sizeof(xfer_data_convert_msg_t));/* zero structure */
    info.type = sriP->request.target;
    info.binDoc = (char*)(&XferDoc);
    info.binType = binFromAtom(sriP->request.selection);
    info.code = XFER_CODE_COPY;
    strcpy(info.pathname,exitClipboardFileObject(sriP->request.selection,sriP->request.target));

    if (sriP->request.target == AxTIMESTAMP) {

        XChangeProperty(Dpy, sriP->request.requestor, sriP->request.property, sriP->request.target, 
                            32, PropModeReplace, (unsigned char *)&siP->selectionTime,
                            1);
        event.property = sriP->request.property;
        goto done;
    }
    else if (sriP->request.target == AxTARGETS) {
        int ix;

        for(ix=0; siP->dataTypeList[ix] != None; ix++); /* count types */
        XChangeProperty(Dpy, sriP->request.requestor, sriP->request.property,
							sriP->request.target, 
                            32, PropModeReplace,
							(unsigned char *)siP->dataTypeList,
                            ix);
        event.property = sriP->request.property;
        goto done;
    }
    else if (sriP->request.target == AxMULTIPLE) {
    /* ???????????????? Fail, for now. */
        goto done;
    }

    else if ( siP->AsterxWinId == AX_ROOT_WINDOW_ID ) {
        int converted, fd;

        /* get or make requested data */
        for (converted=FALSE;
                sriP->dataTypeRequested < sriP->dataTypeCount && !converted;
                sriP->dataTypeRequested++) {
            converted = convertAxSelection(siP, sriP,&info);
        }
            
        if (!converted)
            goto done; /* don't have and can't make requested data */

        /* we have data */
        fd = (int)OpenFileObjFmode(&XferDoc,info.pathname,"r");
        if (!fd) 
            goto done; /* fail. */

        DocSegSeek(fd,0,2); /* goto end */
        AxSelectionLength = FSftell(fd);
        CloseFileObject(fd);
        AxDoneConvert(&info);
        return;
    }


    else {
        /* Send message to application that holds selection,
        requesting conversion. Recipient of message will call AxDoneConvert()
        when it has copied data to clipboard object. */
        info.code = XFER_CODE_COPY;
        sendXferMessage(getPidFromATMWindow(siP->AsterxWinId),&info);
        return;
    }

done:
        event.type = SelectionNotify;
        event.serial = sriP->request.serial;
        event.send_event = sriP->request.send_event;
        event.display = sriP->request.display;
        event.requestor = sriP->request.requestor;
        event.time = sriP->request.time;
        event.selection = sriP->request.selection;
        event.target = sriP->request.target;
#if WINXLOG 
printf(NXLT("XSendEvent SelectionNotify selection=%s property=%s target=%s\n"),
XGetAtomName(Dpy,event.selection),
XGetAtomName(Dpy,event.property),
XGetAtomName(Dpy,event.target));
#endif
        XSendEvent(Dpy,sriP->request.requestor,True,0,(XEvent *)&event);
        return;
}

/*************************************************************************
function: GetXSelectionToXferBin - Read X-window clipboard to transfer bin
Produces an ascii & udf file of clipboard and calls Elf handler.
*/
GetXSelectionToXferBin (p)
XSelectionEvent *p;
{
    Atom req_type, actual_type;
    int actual_format;
	unsigned long nitems, bytes_after;
    unsigned char *dataP;
    int retval, filtered;
    int fd;
    xfer_data_convert_msg_t msgbuf;
    SelectionRequestInfo *sriP;
    mbchar tmpfile[STRPATH+1],curtmpfile[STRPATH+1];

/* Determine who gets the data */

    sriP = getSelectionRequestInfo(p->selection,HERE);
    if (!sriP->AsterxWinId || p->requestor != ClpXWinFromId(sriP->AsterxWinId))
        return; /* don't know who this message is for */

/* Prepare basic message parts */

    memset((char *)&msgbuf,0,sizeof(msgbuf));
    msgbuf.code = XFER_CODE_PASTE;
    msgbuf.type = sriP->dataTypeList[sriP->dataTypeRequested];/* type we want */
    msgbuf.binDoc = (char*)(&XferDoc);
    strcpy(msgbuf.pathname, exitClipboardFileObject(p->selection,p->target));
    msgbuf.binType = binFromAtom(p->selection);
    msgbuf.docType = docTypeFromAtom(msgbuf.type);

    if ( p->property == None ) {
        
        /* Did not get requested data. If we have more types to request,
        do so now, otherwise send a reject message to requestor. */
        sriP->dataTypeRequested++;
        if (sriP->dataTypeRequested < sriP->dataTypeCount) {
            requestSelectionFromX(p->selection,
                                sriP->AsterxWinId,
                                sriP->dataTypeList[sriP->dataTypeRequested]);
            return;
        }
        else
            goto fail;

    }
    req_type = AnyPropertyType;

    retval = XGetWindowProperty(Dpy,p->requestor,p->property,
            0,XMaxRequestSize(Dpy),False,
            req_type,&actual_type,&actual_format,
            &nitems,&bytes_after,&dataP);
    if ( retval != Success )
        goto fail;

    if (actual_type == AxINCR) {

        /* incremental transfer will occur */

        /* we want to see a PropertyNotify event when the selection holder
        puts an increment of data in that property. */
        sriP->priorEventMask = getEventMask(p->requestor);
                                    /* save to restore later */
#if WINXLOG
printf("XSelectInput window=%d PropertyChangeMask\n",p->requestor);
#endif
        XSelectInput(Dpy, p->requestor,
                            sriP->priorEventMask | PropertyChangeMask);

        sriP->incrementalXfer = INCR_XFER_IN;
        sriP->selection = *p;
        strcpy(sriP->incrBuffer,
                exitClipboardFileObject(p->selection,p->target));

        /* create file object for data */
        fd = (int)OpenFileObjFmode(&XferDoc,sriP->incrBuffer,"w");
        if (!fd) {
#if WINXLOG
printf("failure to open file object %s\n",sriP->incrBuffer);
#endif
            goto fail;
        }
        CloseFileObject(fd);

        /* delete property to request selection holder to start
        sending pieces of data */
        XDeleteProperty(Dpy,p->requestor,p->property);
        XFree((char *)dataP);
        return; /* get data later */
    }
   else if (actual_type == sriP->dataTypeList[sriP->dataTypeRequested]) {
        /* get data into file object */
        fd = (int)OpenFileObjFmode(&XferDoc,msgbuf.pathname,"w");
        if (!fd) {
#if WINXLOG
printf("failure to open file object %s\n",sriP->incrBuffer);
#endif
            goto fail;
        }
        DocSegWrite(fd,(char *)dataP,nitems,-1);
        if ( actual_type == XA_STRING && dataP[nitems] != 0 )
            DocSegWrite(fd,"",1,-1);    /* add null term */
        CloseFileObject(fd);

        /* if this is applix-type data, check rev and rev-convert if
        necessary, but only if running in a task. */
        if (AxInTask && AxIsApplixClipboardData(actual_type)) {
            AFileInfo info,curinfo;
            AFileInfoFromHeader((mbchar *)NULL,&curinfo); /* get current vers */
            if (FENoHeader != AFileInfoFromHeader((char *)dataP,&info)) {
                if (info.version != curinfo.version) {
                    int docType;
                    int filtered = FALSE;
                    int err = AOK;

                    AxTempFile(NULL, tmpfile);
                    err = FSWriteFileObjAsFile(tmpfile, 0,
                                &XferDoc, msgbuf.pathname);
                    if (err == AOK) {
                        filtered = FSFilterDocToCurRev(tmpfile,curtmpfile);
                        if (filtered) {
                            /* conversion occured, read converted data */
                            err = FSReadFileIntoFileObj(curtmpfile,
                                            &XferDoc, msgbuf.pathname);
                            AxDeleteTempfile(curtmpfile);
                        }
                    }
                    AxDeleteTempfile(tmpfile);
                    if (err) goto fail;
                }
            }
        }

        XFree((char *)dataP);
    }
    else {
        XFree((char *)dataP);
        goto fail;
    }

#if AXMS
    /* we may need to filter the data. */
    filtered = convert_to_applix_format(&msgbuf);
#else
    filtered = FALSE;
#endif
    /* if all data has been transferred, convert to requestor's type
    and send message to requestor */
    if ((msgbuf.type != None) && !filtered) {
        SelectionInfo sinfo;
        sinfo.AsterxWinId = AX_ROOT_WINDOW_ID; /* pretend we own clipboard,
                            since we have copied data into our clipboard
                            object. */
        convertAxSelection(&sinfo,sriP,&msgbuf);
    }
    goto sendmes;

/* Send rejection message if need be */
fail:
    msgbuf.type = None;

sendmes:
    XDeleteProperty(Dpy,p->requestor,p->property); /* finished with property */
    sendXferMessage(sriP->requestingPid, &msgbuf);
    return;
}

/*************************************************************************
function: AxIsApplixClipboardData
*/
int AxIsApplixClipboardData(dataType)
Atom dataType;
{
    return
        dataType == AxXA_WP_FILE
        || dataType == AxXA_GE_FILE
        || dataType == AxXA_GEGE_FILE
        || dataType == AxXA_SS_FILE
        || dataType == AxXA_SS_INSET_FILE
        || dataType == AxXA_EQN_FILE
        || dataType == AxXA_ELF_DATA
        || dataType == AxXA_QUERY_DATA
        || dataType == AxXA_ELF_TABLE
        || dataType == AxXA_INSET
        || dataType == AxXA_BUILDER_OBJECT
        || dataType == AxXA_GE_PAGE_FILE
        ;
}

static int ElfRequestingXferType;   /* broadcast request type of data */

extern data_ptr ElfRetData;

/*************************************************************************
function: ElfbGetSelection - Return array of strs (or null) if no selection
*/
ElfbGetSelection (nameP,timeP,doctypeP)
data_ptr nameP;     /* atom name */
data_ptr timeP;     /* timeout. default is WAITFORX seconds */
data_ptr doctypeP;  /* optional list of doctypes. single number or array  */
{
    char *name;     /* atom name */
    int time;
    xfer_data_convert_msg_t xfer_info;
    long atom;
    int myTypes[MAX_ATOM_TYPES+1];
    int ix, gotData;
    int holder_pid;
    int parent_pid;
    int wid;

    name = StrFromDataPtr(nameP);
    /* time (in seconds) to wait for response from X */
    if (timeP) {
        time = IntFromDataPtr(timeP); 
    }
    else {
        time = WAITFORX;
    }

    if ( 0 == (atom = atoi(name)) )
        atom = AXInternAtom(Dpy,name);

    wid = ClpIdFromXWin(XGetSelectionOwner(Dpy, atom));
    holder_pid = getPidFromATMWindow(wid);
    parent_pid = ThimPidFromUID(TaskInfo[AxInTask].parentTask);
    if (holder_pid == parent_pid)
        ACImOk();

    /* construct list of acceptable types of materal */
    ix = 0;
    if (doctypeP) {
        if (AxIsArray(doctypeP)) {
            int jx;

            for (jx=0; jx<AxArraySize(doctypeP); jx++) {
                myTypes[ix++] = atomFromDoctype(AxIntFromArray(doctypeP,jx));
            }
        }
        else if (AxIsInt(doctypeP)) {
            myTypes[ix++] = atomFromDoctype(AxIntFromDataPtr(doctypeP));
        }
    }
    else { /* default */
        myTypes[ix++] = AxXA_ELF_DATA;
        myTypes[ix++] = XA_STRING;
    }
    myTypes[ix] = None;

    gotData = FALSE;
    if (AxGetSelection(binFromAtom(atom),
                     AX_ROOT_WINDOW_ID, myTypes, &xfer_info)) {
        gotData = TRUE;
        if (xfer_info.type == None) {
            /* wait for response from X */
            gotData = waitForSelection(time,&xfer_info);
        }
        else {
            gotData = TRUE;
        }
    }

    if (gotData) {
        ElfRetData = (data_ptr) ElfXferIn(&xfer_info);
    }
    else {
        /* no data */
        ElfRetData = MkArray(0);
    }
}

/*************************************************************************
function: waitForSelection - Wait for X to supply data
returns TRUE if we get a "selection paste" message
What do we do with non-paste messages?????
*/
static int waitForSelection (time,infoP)
int time;   /* how long to wait, in seconds. */
xfer_data_convert_msg_t *infoP;
{
    int code;
    xfer_data_convert_msg_t *msgP;
    long size, senderType, senderTask, ljunk;
    short temp;

#if WINXLOG
printf("waitForSelection: time=%d\n",time);
#endif
again:
    code = AxGetMessage(time,(char **)&msgP,&size,&senderType,&senderTask,&ljunk);
    if ( code == IPC_TO ) {
#if WINXLOG
printf("waitForSelection gets IPC_TO message\n");
#endif
        infoP->type = None;
        return FALSE; /* timeout */
    }
    if ( senderType != APWINXFER ) {
#if WINXLOG
printf("waitForSelection gets non-APWINXFER message (senderType=%d)\n",senderType);
#endif
        AxMessageAbsorbed();    /* !!1 sigh! */
        goto again;
    }
#if WINXLOG
printf("waitForSelection gets APWINXFER message code=%d\n",msgP->code);
#endif
    if (msgP->code != XFER_CODE_PASTE) {
#if WINXLOG
printf("waitForSelection gets APWINXFER Non-Paste message (senderType=%d)\n",senderType);
#endif
        AxMessageAbsorbed();    /* !!1 sigh! */
        goto again;
    }
    *infoP = *msgP;
    AxMessageAbsorbed();
    return ( infoP->type != None );
}

/*************************************************************************
function: ElfXferIn - Transfer ELF data and strings in.
*/
static void *ElfXferIn (infoP)
xfer_data_convert_msg_t *infoP;
{
    int fp;
    char buf[1001]; /* plenty */
    int len, ix;
    int got_newline = FALSE;    /* relavent only for last line */
    data_ptr dataP, AxCopyData(), elfDataP;


    elfDataP = MkArray(1);
    if ( infoP->type == XA_STRING ) {

        /* Read ascii string file, make array of strings */
        dataP = MkArray(0);
        fp = (int)OpenFileObjFmode(&XferDoc,infoP->pathname,"r");
        if ( fp ) {
            ix = 0;
            while ( FSfgets(buf,1000,fp) ) {
                got_newline = FALSE;
                if (buf[len = strlen(buf)-1] == '\n' ) {
                    buf[len] = 0;
                    got_newline = TRUE;
                }
                dataP = AxAddStrToArray(dataP,ix,buf);
                ix++;
            }
            if ( got_newline )
                dataP = AxAddStrToArray(dataP,ix,"");
            CloseFileObject(fp);
        }
        AxAddArrayToArray(elfDataP,0,dataP); /* data, no header */
    }
    else if (infoP->type == AxXA_ELF_DATA
                || infoP->type == AxXA_ELF_TABLE) {

        /* get elf data from file object, task THIMpid */
        elfDataP = AxReadDataFromFileObject(THIMpid,
							(DocInfoType *)infoP->binDoc,infoP->pathname);
    }
    return (void *)elfDataP;
}

/*************************************************************************
function: ElfbSetSelection - Set X clipboard selection
*/
ElfbSetSelection (nameP,aP,doctypeP)
data_ptr nameP; /* either atom number or name */
data_ptr aP; /* array of data */
data_ptr doctypeP; /* doctype of data (default is ELF_DATA) */
{
    char *name;
    Atom atom;
    Atom dataType;
    xfer_data_convert_msg_t info;
    int wid;
    int holder_pid, parent_pid;
    int myTypes[MAX_ATOM_TYPES+1];
    int success;

    name = AxStrPtrFromDataPtr(nameP);
	
    if ( 0 == (atom = atoi(name)) )
        atom = AXInternAtom(Dpy,name);

    /* why is this here? -chris 18-nov-94 */
    wid = ClpIdFromXWin(XGetSelectionOwner(Dpy, atom));
    holder_pid = getPidFromATMWindow(wid);
    parent_pid = ThimPidFromUID(TaskInfo[AxInTask].parentTask);
    if (holder_pid == parent_pid)
        ACImOk();

    if (doctypeP)
        dataType = atomFromDoctype( AxIntFromDataPtr(doctypeP));
    else
        dataType = AxXA_ELF_DATA; /* default */

    AxExitClipboard(atom,dataType,&info);
    success = copyElfDataToClipboard(aP, &info);
}

/***************************************************************************
function: getSelectionInfo - return pointer to global data
*/
static SelectionInfo *getSelectionInfo(selectionType,hereOrThere)
int selectionType;
int hereOrThere;
{
    Window win;
	SelectionInfo *siP;

    /* get the selection owner */
    win =  XGetSelectionOwner(Dpy,selectionType);
	siP = 
        &XferSelectionInfo[(selectionType == XA_PRIMARY ? 0 : 1)][hereOrThere];
	if (win != ClpXWinFromId(siP->AsterxWinId)) /* Sanity Check : Make sure the information is up to date. */
		siP->AsterxWinId = 0;
	return siP;
}

/***************************************************************************
function: getSelectionRequestInfo - return pointer to global data
*/
static SelectionRequestInfo *getSelectionRequestInfo(selectionType,hereOrThere)
int hereOrThere;
{
    return
  &XferSelectionRequestInfo[(selectionType == XA_PRIMARY ? 0 : 1)][hereOrThere];
}

/***************************************************************************
function: AxSelectionEvent - Standard selection event processing
*/
AxSelectionEvent(event)
XEvent *event;
{

#if WINXLOG
printf("AxSelectionEvent: event.type=");
switch ( event->type ) {
case SelectionRequest:
printf("SelectionRequest\n");
break;
case SelectionNotify:
printf("SelectionNotify\n");
break;
case SelectionClear:
printf("SelectionClear\n");
break;
case PropertyNotify:
printf("PropertyNotify\n");
break;
}
#endif

    switch ( event->type ) {
    case SelectionRequest:
        SupplySelection(&event->xselectionrequest);
        break;
    case SelectionNotify:
        GetXSelectionToXferBin(&event->xselection);
        break;
    case SelectionClear:
        AxClearSelection(event->xselectionclear.selection,
                             event->xselectionclear.window);
        break;
    case PropertyNotify:
        AxPropertyChanged(&event->xproperty);
        break;
    }
}

/***************************************************************************
function: AxPropertyChanged - handle property change event 
*/
int AxPropertyChanged(event)
XPropertyEvent *event;
{
    Atom selType;
    int retval;

#if WINXLOG
printf(NXLT("property for %s transfer. Event is %s\n"),
    ((selType == XA_PRIMARY) ? NXLT("XA_PRIMARY") : NXLT("XA_CLIPBOARD")),
    ((event->state == PropertyDelete) ? NXLT("PropertyDelete")
        : ((event->state == PropertyNewValue) ? NXLT("PropertyNewValue")
                                    : NXLT("Unknown"))));
#endif
    /* is this a selection request! The name of the property that
    has been changed tells us whether this is primary or clipboard selection */
    if (event->atom == propertyForSelectionTransfer(XA_PRIMARY)) {
        selType = XA_PRIMARY;
    }
    else if (event->atom == propertyForSelectionTransfer(XA_CLIPBOARD)) {
        selType = XA_CLIPBOARD;
    }
    else {
        return FALSE; /* nothing we know about here. */
    }

    if (event->state == PropertyDelete) {
        retval = sendSelectionIncremental(selType, event);
    }
    else if (event->state == PropertyNewValue) {
        retval = getSelectionIncremental(selType, event);
    }
    else {
        retval = FALSE;
    }
    return retval;
}

/***************************************************************************
function: AxMoveClipboardToExitClipboard - to give clipboard contents to ax
base services because application is going away or resetting itself.
*/
AxMoveClipboardToExitClipboard(winId, cvtfun) 
int winId; /* ATM window id */
void (*cvtfun)(); /* Function which, given xfer_data_convert_msg_t will convert
                clipboard as indicated in the packet. */
{
    SelectionInfo *siP;
    int ix,jx;
    xfer_data_convert_msg_t info;
    int myTypes[MAX_ATOM_TYPES+1];

#if WINXLOG
printf("AxMoveClipboardToExitClipboard winId %d\n",winId);
#endif

    if (!AxIAmSelectionOwner(XA_CLIPBOARD, winId) ) return;

    siP = getSelectionInfo(XA_CLIPBOARD,HERE);
    if (siP->AsterxWinId != winId) return;

    /* get all possible types of data in case they are requested later */
    for(ix=0,jx=0; ix<siP->dataTypeCount; ix++) {
        memset((char *)&info,0, sizeof(info));
        info.type = siP->dataTypeList[ix];
        info.binDoc = (char *)&XferDoc;
        info.binType = XFER_BIN_CLIPBOARD;
        info.code = XFER_CODE_CLEAR_SELECTION; /* a signal to AxDoneConvert that
                                            nobody but us is waiting for this */
        strcpy(info.pathname,exitClipboardFileObject(XA_CLIPBOARD,info.type));
        (*cvtfun)(winId, &info); /* get clipboard data */
        if (info.type == siP->dataTypeList[ix])
            myTypes[jx++] = info.type;
    }
    myTypes[jx] = None;
    AxRegisterSelectionAndConverters(binFromAtom(XA_CLIPBOARD),
                                    AX_ROOT_WINDOW_ID,myTypes);
}

/*************************************************************************
function: AxExitClipboard - Apply data to clipboard on exit
caller gets an xfer_data_convert_msg_t back which has name of file object to
which caller must then write data of type dataType.
*/
AxExitClipboard (selType,dataType,infoP)
Atom selType;
Atom dataType;      /* XA_GE_FILE, XA_ASCII_FILE... */
xfer_data_convert_msg_t *infoP; /* returned */
{
    int myTypes[MAX_ATOM_TYPES+1];

#if WINXLOG
printf("AxExitClipboard: selType %d dataType %d\n",selType,dataType);
#endif

/* Make this the selection owner, notify other X-windows */

    clearExitBoard(selType);
    AxClearSelection(selType,XGetSelectionOwner(Dpy,selType));
    XSetSelectionOwner(Dpy,selType,AxDummyWin,TimeOfButtonPress);
    XSync(Dpy,FALSE);   /* just to be safe! */
    memset((char *)infoP,0,sizeof(xfer_data_convert_msg_t));
    strcpy(infoP->pathname, exitClipboardFileObject(selType,dataType));
    infoP->binDoc = (char*)&XferDoc;
    infoP->binType = binFromAtom(selType);
    infoP->type = dataType;
    myTypes[0] = dataType;
    myTypes[1] = None;
    AxRegisterSelectionAndConverters(binFromAtom(selType),
                                    AX_ROOT_WINDOW_ID,myTypes);
}

/***************************************************************************
function: exitClipboardFileObject - construct name of exit cliboard file object
given data type, and type of selection
*/
static char *exitClipboardFileObject(selType,dataType)
Atom selType;
int dataType;
{
    static char pathname[STRPATH+1];
    char *selection;

    selection = SELECTION_CODE(selType);
    sprintf(pathname,NXLT("%s/%d%s"),XferDoc.pathname,dataType,selection);
    return pathname;
}


/***************************************************************************
function: ElfbCopyFileToClipboard - copy to Asterx central clipboard
*/
ElfbCopyFileToClipboard(s, docType)
char *s;
int docType;
{
    int err;
    err = AxFileToClipboard(XFER_BIN_CLIPBOARD,s,docType);
    if (err)
        ElfStrAbort(ErrnoErr(), NULL, s);
}

/***************************************************************************
function: ElfbCopyClipboardToFile - copy from Asterx central clipboard
*/
ElfbCopyClipboardToFile(fileP, docTypeP)
data_ptr fileP;
data_ptr docTypeP;
{
    char *s;
    int err, docType;
    int dataTypes[32];
    int ix;

    s = StrFromDataPtr(fileP);
    ix = 0;
    if (docTypeP) {
        if (AxIsArray(docTypeP)) {
            for (ix=0; ix<AxArraySize(docTypeP); ix++) {
                docType = IntFromArray(docTypeP, ix);
                dataTypes[ix] = atomFromDoctype(docType);
            }
        }
        else {
            docType = IntFromDataPtr(docTypeP);
            dataTypes[ix++] = atomFromDoctype(docType);
        }
    }
    else {
        dataTypes[ix++] = XA_STRING;
    }
    dataTypes[ix] = None;
    err = AxClipboardToFile(XFER_BIN_CLIPBOARD,s,dataTypes);
    if (err && err != -1) /* -1 means no data written */
        ElfStrAbort(ErrnoErr(), NULL, s);
}

/***************************************************************************
function: binFromAtom - convert atom to bin code
*/
int binFromAtom(selection)
Atom selection;
{
    return((selection == XA_PRIMARY) ? XFER_BIN_SELECTION : XFER_BIN_CLIPBOARD);
}

/***************************************************************************
function: selectionFromBinType - convert bin code to atom
*/
Atom selectionFromBinType(binType)
{
    return((binType == XFER_BIN_SELECTION) ? XA_PRIMARY : XA_CLIPBOARD);
}

/***************************************************************************
function: atomFromDoctype - convert doctype to atom
*/
static Atom atomFromDoctype(docType)
int docType;
{
    Atom dataType;

    switch (docType) {
    case AX_DOC_TYPE_ASTERX_WORDS:
        dataType = AxXA_WP_FILE;
        break;
    case AX_DOC_TYPE_ASTERX_GRAPHICS:
        dataType = AxXA_GEGE_FILE;
        break;
    case AX_DOC_TYPE_ASTERX_SS:
        dataType = AxXA_SS_FILE;
        break;
    case AX_DOC_TYPE_ASTERX_MACROS:
        dataType = AxXA_WP_FILE;
        break;
#if 0 /* can't xfer these. */
    case AX_DOC_TYPE_ASTERX_BINARY_SS:
        dataType = ???;
        break;
    case AX_DOC_TYPE_ASTERX_BITMAP:
        dataType = ???;
        break;
#endif
    case AX_DOC_TYPE_ASTERX_EQUATIONS:
        dataType = AxXA_EQN_FILE;
        break;
    case AX_DOC_TYPE_ASTERX_ELF_TABLE:
        dataType = AxXA_ELF_TABLE;
        break;
    case AX_DOC_TYPE_ASTERX_ELFARRAY:
        dataType = AxXA_ELF_DATA;
        break;
    case AX_DOC_TYPE_ASTERX_QUERYDATA:
        dataType = AxXA_QUERY_DATA;
        break;
    case AX_DOC_TYPE_ASTERX_OBJ_CLASS:
        dataType = AxXA_BUILDER_OBJECT;
        break;
    default:
        dataType = XA_STRING; /* will handle simple ascii data */
        break;
    }
    return dataType;
}

/***************************************************************************
function: docTypeFromAtom - convert some atoms to doc (object) type. In some
cases we don't know a doctype from Atom.
*/
int docTypeFromAtom(dataType)
Atom dataType;
{
    int docType;

    docType = (dataType == AxXA_WP_FILE) ? AX_DOC_TYPE_ASTERX_WORDS
                : (dataType == AxXA_GEGE_FILE) ? AX_DOC_TYPE_ASTERX_GRAPHICS
                : (dataType == AxXA_GE_FILE) ? AX_DOC_TYPE_ASTERX_GRAPHICS
                : (dataType == AxXA_SS_FILE) ? AX_DOC_TYPE_ASTERX_SS
                : (dataType == AxXA_SS_INSET_FILE) ? AX_DOC_TYPE_ASTERX_SS
                : (dataType == AxXA_EQN_FILE) ? AX_DOC_TYPE_ASTERX_EQUATIONS
                : (dataType == AxXA_ELF_DATA) ? AX_DOC_TYPE_ASTERX_ELFARRAY
                : (dataType == AxXA_ELF_TABLE) ? AX_DOC_TYPE_ASTERX_ELF_TABLE
                : (dataType == AxXA_QUERY_DATA) ? AX_DOC_TYPE_ASTERX_QUERYDATA
                : (dataType == XA_STRING) ? AX_DOC_TYPE_ASCII
                : (dataType == AxXA_BUILDER_OBJECT)?AX_DOC_TYPE_ASTERX_OBJ_CLASS
                : (dataType == AxXA_GE_PAGE_FILE) ? AX_DOC_TYPE_ASTERX_GRAPHICS
                : AX_DOC_TYPE_UNKNOWN;
    return docType;
}

/***************************************************************************
function: propertyForSelectionTransfer - atom which is name of property that
we want to get selection contents in.
*/
static Atom propertyForSelectionTransfer(selType)
Atom selType; /* selection Type */
{
    Atom property;

    if (selType == XA_PRIMARY) {
        property = AxXFER_PRIMARY;
    }
    else {
        property = AxXFER_CLIPBOARD;
    }
    return (property);
}

/***************************************************************************
function: getPidFromATMWindow -  Return pid associated with an atm window.
*/
static int getPidFromATMWindow(wid)
int wid;
{
    if (wid == AX_ROOT_WINDOW_ID) {
        /* special case for no window. */
        return THIMpid; /* watch out! */
    }
    else {
        return PidFromATMWindow(wid);
    }
}

/***************************************************************************
function: sendXferMessage - send a inter-window xfer message to an application
*/
static void sendXferMessage(pid,msgP)
int pid;
xfer_data_convert_msg_t *msgP;
{

#if WINXLOG
printf(NXLT("sending APWINXFER message %s type %d to pid=%d\n"),
((msgP->code == XFER_CODE_COPY) ? NXLT("XFER_CODE_COPY")
: (msgP->code == XFER_CODE_PASTE) ? NXLT("XFER_CODE_PASTE")
: (msgP->code == XFER_CODE_CLEAR_SELECTION) ? NXLT("XFER_CODE_CLEAR_SELECTION")
: NXLT("UNKNOWN")),
msgP->type,
pid);
#endif
    IPCXsendMes(IPC_SRT,
                pid,
                msgP, sizeof(xfer_data_convert_msg_t),APWINXFER,THIMpid);
    return;
}

/***************************************************************************
function: getSelectionIncremental - get a piece of a selection, being transferred incrementally returns TRUE on success.
*/
static int getSelectionIncremental(selType, event)
Atom selType;
XPropertyEvent *event;
/* get a piece of a selection, being transferred incrementally */
/* do this when a ProperyNotify event of type PropertyNewValue occurs */
{
    
    SelectionRequestInfo *sriP;
    Atom actual_type;
    int actual_format, retval;
	unsigned long nitems, bytes_after;
    int fd;
    unsigned char *dataP;

#if WINXLOG
printf("getSelectionIncremental() called\n");
#endif
    if (event->state != PropertyNewValue) return FALSE; /* wrong event type */

    sriP = getSelectionRequestInfo(selType,HERE);

    if (sriP->incrementalXfer != INCR_XFER_IN) {
        return FALSE; /* incremental transfer in is not in process */
    }
    if (event->window != sriP->selection.requestor
            || event->display != sriP->selection.display
            || event->display != Dpy
            || event->atom != sriP->selection.property) {
        return FALSE; /* this is not the property that we are interested in. */
    }

    /* get the property value, delete property */
    retval = XGetWindowProperty(event->display,event->window,event->atom,
            0,XMaxRequestSize(Dpy),True,
            AnyPropertyType,&actual_type,&actual_format,
            &nitems,&bytes_after,&dataP);
    if (retval != Success) return FALSE;
    
    /* handle end of data. send it to requesting application */
    if (nitems == 0 || !dataP) {
        xfer_data_convert_msg_t msgbuf;

        /* no longer interested in these events */
#if WINXLOG
printf("XSelectInput window=%d priorEventMask\n",event->window);
#endif
        XSelectInput(event->display, event->window, sriP->priorEventMask);
        sriP->incrementalXfer = INCR_XFER_OFF; /* end of transfer */

        /* append a null to string data */
        if (actual_type == XA_STRING) {
            fd = (int)OpenFileObjFmode(&XferDoc,sriP->incrBuffer,"r+");
            /* ???????? check for errors */
            if (!fd) tempAbort("incrementalXfer Failure");
            if (fd) {
                DocSegSeek(fd,0,2); /* go to end of file */
                DocSegWrite(fd,"",1,-1);    /* add null term */
                CloseFileObject(fd);
            }
        }
        memset((char *)&msgbuf,0,sizeof(msgbuf));
        msgbuf.code = XFER_CODE_PASTE;
        msgbuf.type = sriP->dataTypeList[sriP->dataTypeRequested];
        msgbuf.docType = docTypeFromAtom(msgbuf.type);
        msgbuf.binDoc = (char*)(&XferDoc);
        strcpy(msgbuf.pathname, sriP->incrBuffer);
        msgbuf.binType = binFromAtom(selType);
        sendXferMessage(sriP->requestingPid, &msgbuf);
    }

    /* save this piece of data */
    else {
        /* append data to file object */
        fd = (int)OpenFileObjFmode(&XferDoc,sriP->incrBuffer,"r+");
        /* ???????? check for errors */
        if (!fd) tempAbort("incrementalXfer Failure");
        if (fd) {
            DocSegSeek(fd,0,2); /* go to end of file */
            DocSegWrite(fd,(char *)dataP,nitems,-1);
            CloseFileObject(fd);
#if WINXLOG
dataP[nitems] = 0;
printf("data increment received=<%s>\n",dataP);
#endif

        }
        XFree((char *)dataP);
    }
    return TRUE;
}

/***************************************************************************
function: sendSelectionIncremental - incrementally supply selection 
*/
static int sendSelectionIncremental(selType, event)
Atom selType;
XPropertyEvent *event;
{
    SelectionRequestInfo *sriP;
    int fd;
    int nbytes;

#if WINXLOG
printf("sendSelectionIncremental called\n");
#endif
    if (event->state != PropertyDelete) return FALSE; /* wrong event type */
    
    sriP = getSelectionRequestInfo(selType,THERE);
    if (sriP->incrementalXfer != INCR_XFER_OUT) {
        return FALSE; /* incremental transfer out is not occuring now */
    }
    
    if (sriP->request.display != event->display 
            || event->display != Dpy
            || sriP->request.requestor != event->window
            || sriP->request.property != event->atom)
        return FALSE;
            /* this is not related to the selection being transferred */

    /* get a piece of data to send */
    fd = (int)OpenFileObjFmode(&XferDoc,sriP->incrBuffer,"r");
    if (fd) {
        char *conversionBuf;
        
        conversionBuf = (char *) malloc (ConvertBufSize + 1);
        DocSegSeek(fd, sriP->filepos, 0);
        nbytes = DocSegRead(fd,conversionBuf,sriP->increment, -1);
        sriP->filepos = FSftell(fd);
        CloseFileObject(fd);
        conversionBuf[nbytes] = '\0';
        if (nbytes == 0) {

            /* no longer interested in these events */
#if WINXLOG
printf("XSelectInput window=%d priorEventMask\n",event->window);
#endif
            XSelectInput(event->display, event->window, sriP->priorEventMask);
            sriP->incrementalXfer = INCR_XFER_OFF; /* end of transfer */

        }

        /* send data or indication of end of data */
        XChangeProperty(sriP->request.display,
                            sriP->request.requestor,
                            sriP->request.property,
                            sriP->request.target, 8,
                            PropModeReplace,
							(unsigned char *)conversionBuf, nbytes);

#if WINXLOG
printf("data increment sent=<%s>\n",conversionBuf);
#endif
        free(conversionBuf);
    }
}

/***************************************************************************
function: geteventMask - get event mask for a window
*/
static long getEventMask(win)
Window win;
{
    XWindowAttributes windowAtts;
    long mask;

    if (XGetWindowAttributes(Dpy,win,&windowAtts))
        mask = windowAtts.your_event_mask;
    else
        mask = 0;
    return mask;
}
/***************************************************************************
function: AxClearSelectionsIfOwnedByWindow - window is going away and
can not own any selection.
*/
void AxClearSelectionsIfOwnedByWindow(wid)
int wid;
{
    if (AxIAmSelectionOwner(XA_CLIPBOARD,wid))
        AxClearSelection(XA_CLIPBOARD,(Window)NULL);
    if (AxIAmSelectionOwner(XA_PRIMARY,wid))
        AxClearSelection(XA_PRIMARY,(Window)NULL);
}


/*************************************************************************
function: ACinitXferBin - Create transfer bin document
This should only be called once.
THIMpid MUST be 0 when this is called.
*/
ACinitXferBin ()
{
    char buf[STRPATH+1];
    typeErr err, FSOpenDocument();

    if ( THIMpid != 0 )
        tempAbort(NXLT("ACinitXferBin"));
    sprintf(buf,NXLT("%s/xfer"),ThimTempDir);
/*  mktemp(buf); */
    err = FSOpenDocument(buf,FSO_OWNER+FSO_FILEOK,&XferDoc);
    if ( err != AOK )
        tempAbort(NXLT("Could not initialize clipboard in /tmp"));
    ACxferClearBin();
}


/*************************************************************************
function: ACxferClearBin - Delete virtual files from xfer bin
*/
ACxferClearBin ()
{
    char *segname;
    int ix = 0;
    char delname[STRPATH+1];
    char *ScanForNthFileObject();

    while ( segname = ScanForNthFileObject(&XferDoc,ix) ) {
        sprintf(delname,"%s/%s",XferDoc.pathname,segname);
        DeleteFileObject(&XferDoc,delname);
        ix++;
    }
}


/*************************************************************************
function: FSOpenDocument - Open a docuemnt
*/
typeErr FSOpenDocument (name,mode,infoP)
char *name;
int mode;
DocInfoType *infoP; /* returned */
{
    typeErr FSDoOpenDocument();
    return(FSDoOpenDocument(name,mode,infoP,FALSE,0));
}

/*************************************************************************
function: FSDoOpenDocument - Open a docuemnt
*/
static typeErr FSDoOpenDocument (name,mode,infoP,forceFlag,type)
char *name;
int mode;
DocInfoType *infoP; /* returned */
int forceFlag;  /* if FALSE, assume type from suffix, else force to type */
{
    char pathname[STRPATH+1];
    char junk[STRPATH+1];
    register doc_info *p;
    typeErr err;
    int fmode;
    int created;
    char *rev = NXLT("12345678901234567890");

    AxSysDosUnixPath(name); /* DANGER! DANGER! affects caller! */
    err = IsValidPathFilename(name);
    if ( err != AOK )
        return(err);
    strcpy( pathname, name );

    p = FindDocStruct(pathname);
    memset(infoP,0,sizeof(*infoP));
    strcpy(infoP->pathname,pathname);
    infoP->DocId = p->id;
    infoP->sysFileId = p;
    if ( mode & FSO_OWNER ) {
        if ( p->owner )
            return(FEalreadyOpen);
        p->owner = THIMpid;
        fmode = DFM_CREATE;
    }
    else
        fmode = DFM_READ;
    if ( mode & FSO_FILEOK )
        fmode |= DFM_FILEOK;
    p->num_openers++;
    p->openers = (short*)TaskRealloc(0,p->openers,p->num_openers * sizeof(short));
    p->openers[p->num_openers-1] = THIMpid;
    if ( forceFlag )
        infoP->DocMatType = p->type = type;
    else
        infoP->DocMatType = p->type = type = DocTypeFromName(pathname);
    ParsePath(pathname,junk,infoP->DocName);
    ClipDocSuffix(infoP->DocName);
    if ( (infoP->DocMatType == MUserCommand) || (infoP->DocMatType == MNone) )
        fmode |= DFM_FILEOK;    /* HACK for editor revert */
    err = OpenDocumentFile(THIMpid,
            infoP->DocId,
            pathname,
            fmode,
            &infoP->DocMatType,
            &rev,
            &created,
            &infoP->docFileId);
    if ( err == AOK )
        err = ReadKSObject(infoP);
    if ( err != AOK ) {
        FSClsDoc(infoP,-1);
        return(err);
    }
    if ( created ) {
        infoP->onDisk = FALSE;
        DfltSetupsToDoc(infoP);
    }
    else
        infoP->onDisk = TRUE;
    p->type = infoP->DocMatType;
    return(AOK);
}


/*************************************************************************
function: FindDocStruct - Find or create doc_info struct
*/
doc_info *FindDocStruct (pathname)
char *pathname;
{
    register doc_info *p;
    register int ix;

/* search thru doc list to see if currently referenced */

    for ( p = DocList; p; p = p->next ) {
        if ( streq(pathname,p->pathname) )
            goto got_it;
    }

/* if not, create and thread to doclist */

    p = (doc_info*)TaskCalloc(0,1,sizeof(doc_info));
    strcpy(p->pathname,pathname);
    p->openers = (short*)TaskAlloc(0,sizeof(short));
    p->lookers = (short*)TaskAlloc(0,sizeof(short));
    p->refers = (short*)TaskAlloc(0,sizeof(short));
    p->id = DocId++;
    p->next = DocList;
    DocList = p;


/* Note the referncing to keep this struct in memory till task goes away */

got_it:

    for ( ix = 0; ix < p->num_refers; ix++ ) {
        if ( p->refers[ix] == THIMpid )
            return(p);      /* already entered, all done! */
    }
    p->num_refers++;
    p->refers = (short*)TaskRealloc(0,p->refers,p->num_refers*sizeof(short));
    p->refers[ix] = THIMpid;
    return(p);
}

/*************************************************************************
function: FSClsDoc - Close doc with abandon/save of KS file
*/
typeErr FSClsDoc (infoP,LockFlag)
DocInfoType *infoP;
short LockFlag;
{
    register typeErr code;
    doc_info *p;
    register int ix;

    p = infoP->sysFileId;
    if ( LockFlag == 1 ) {
        if ( AOK != (code = FSChkPtDoc(infoP)) )
            return(code);
    }
    if ( p->owner == THIMpid )
        p->owner = 0;       /* orphan it */
    for ( ix = 0; ix < p->num_openers; ix++ ) {
        if ( p->openers[ix] == THIMpid ) {
            p->openers = clip1fromShorts(p->openers,ix,p->num_openers);
            p->num_openers--;
        }
    }
/* if nobody is using the image, toss it */
    if ( !p->owner && (p->num_openers == 0) ) {
        AbandonKSObject(infoP);
        AbandonDocumentFile(infoP->docFileId);
        infoP->docFileId = NULL;
    }

    RmBackupDoc(infoP);
    return(AOK);
}


void
DfltSetupsToDoc(infoP)
DocInfoType *infoP;
{
    double height, width;
    char tmp[30];
    int i;
    extern int AxMetric;

    height = UPdblVal(PROF_PAGE_HEIGHT);
    width = UPdblVal(PROF_PAGE_WIDTH);
    if ( AxMetric ) {
        if ( (height < 1) || (height > 60) )
            height = 29.7;
        if ( (width < 1) || (width > 60) )
            width = 21.0;
        height = height / 2.54;
        width = width / 2.54;
    }
    else {
        if ( (height < 0.5) || (height > 30) )
            height = 11.0;
        if ( (width < 0.5) || (width > 30) )
            width = 8.5;
    }
    i = height * 1000;
    sprintf(tmp,"%d",i);
    FSKSAdd(infoP,SETUP_HEIGHT_STR,tmp);
    i = width * 1000;
    sprintf(tmp,"%d",i);
    FSKSAdd(infoP,SETUP_WIDTH_STR,tmp);
    FSKSAdd(infoP,SETUP_CLASS_STR,NXLT("Printer"));
    FSKSAdd(infoP,SETUP_NUMBER_STR,"70");
    FSKSAdd(infoP,SETUP_LEFT_OFS_STR,"0");
    FSKSAdd(infoP,SETUP_TOP_OFS_STR,"0");
    FSKSAdd(infoP,SETUP_P1_ORG_STR,"0");
    FSKSAdd(infoP,SETUP_EVEN_ORG_STR,"0");
    FSKSAdd(infoP,SETUP_ODD_STR,"0");
    FSKSAdd(infoP,SETUP_LANDSCAPE_STR,"0");
    FSKSAdd(infoP,SETUP_TIME_STR,"1");
}

/*************************************************************************
function: FSChkPtDoc - Save the document file.
*/
typeErr FSChkPtDoc (infoP)
DocInfoType *infoP;
{
    register doc_info *p;

    p = infoP->sysFileId;

    if ( THIMpid != p->owner )  /* ignore non-owner checkpoint */
        return(AOK);        /* checkpoint non-owner ok */

    WriteKSObject(infoP);
    return(AOK);
}

short *clip1fromShorts(p,pos,cnt)
register short *p;
register int pos, cnt;
{
    for ( ++pos; pos < cnt; pos++ )
        p[pos-1] = p[pos];
    return((short*)TaskRealloc(0,p,(cnt-1)*sizeof(short)));
}

/*************************************************************************
function: ClipDocSuffix - Remove doc identifying suffix
*/
void ClipDocSuffix (name)
register char *name;
{
    register char *p;

    for ( p = name+strlen(name)-1; *p && *p != '/'; p-- )
        {
        if ( *p == '.' ) 
            {
            *p = 0;
            return;
            }
        }
}


/**************************************************************************/
#if AXMS    /* { */
/*** this function is called when an app requests data (pastes) 
     and the owner of the clipboard is AX_ROOT_WINDOW.
     the "exit clipboard" needs to supply the data.
***/
int
GetExitClipboardSelection(int data_type, xfer_data_convert_msg_t *infoP)
{
    Atom selectionType;
    int ix, jx, ret;
    SelectionInfo *siP;
    SelectionRequestInfo *sriP;


    ret = FALSE;
    memset(infoP,0,sizeof(xfer_data_convert_msg_t)); /* zero structure */
    infoP->type = None;
    infoP->binType = XFER_BIN_CLIPBOARD;
    infoP->code = XFER_CODE_COPY;
    selectionType = selectionFromBinType(XFER_BIN_CLIPBOARD);
    siP = getSelectionInfo(selectionType, HERE);

    /* save info about this request */
    sriP = getSelectionRequestInfo(selectionType, HERE);
    sriP->AsterxWinId = AX_ROOT_WINDOW_ID;
    sriP->dataTypeList[0] = data_type;
    sriP->dataTypeCount = 1;

    for ( ix=0; ix < siP->dataTypeCount; ix++ )
        {
        if ( data_type == siP->dataTypeList[ix] )
            {
            /* we have a requested type that the holder can produce.
                Get that data from holder */
            sriP->dataTypeRequested = 0;
            sriP->requestingPid = AxInTask;/* reply should come to this task */
            ret = convertAxSelection(siP, sriP, infoP);
            if (ret) return ret;
            }
        }
    return ret;
}

#if 0 /* { may need this later. */
/* this is called when an app exits while still owning the clipboard. */
/* would normally be during processing of WM_RENDERALLFORMATS, but is
 * instead called by the application on exit.                         */
/* caller needs to close the clipboard!! */
static void
SetupWinClipboardForExit(int winID)
{
    int save_winid, save_count;
    HWND hwnd;
    Window win;
    SelectionInfo *siP;


    win = XWinFromATMWin(winID);
    hwnd = AxGetHWNDFromHDC(win);
    MS_OpenClipboard(hwnd);

    /* bit of a necessary hack... */
    /* MS_EmptyClipboard will send a WM_DESTROYCLIPBOARD message which  *
     * will send a SelectionClear event to winxfer. This will clear the *
     * global Selection pointer, which we don't want to do because the  *
     * exit code (AxMoveClipboardToExitClipboard) needs it. So we get   *
     * the current pointer and save the values before calling           *
     * MS_EmptyClipboard and restore them after.            -donato     */
    siP = getSelectionInfo(XA_CLIPBOARD, HERE);
    save_winid = siP->AsterxWinId;
    save_count = siP->dataTypeCount;
    MS_EmptyClipboard();
    siP = getSelectionInfo(XA_CLIPBOARD, HERE);
    siP->AsterxWinId = save_winid;
    siP->dataTypeCount = save_count;
}
#endif    /* } if zero */

#endif    /* } AXMS */
/*************************************************************************/

#endif /* END real VERSION } ---------------------------*/

