/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: xxfer.c
   Applix ASC version: 5
   Date: Mon Nov 16 13:53:25 1998
*/
#include <stdio.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <aos.h>
#include <scac.h>
#include <getstr20.h>
#include <getstr27.h>
#include <xferfmt.h>
#include <fsrts.h>

/* From DXferOut.c */
extern char	*CTFAPrevText();

#define MAX_XSTR 1000	/* good number?, depends on X-server! */
#define BTRUE -1

extern Display *Dpy;	/* Resolution of DisplayWidget */
extern unsigned long TimeOfButtonPress;	/* event time of button down */
typeErr ErrnoErr();
DocInfoType *ACxferDoc();
extern Window HostWinFromCurWin();

typedef struct {
	int registered;	/* boolean whether X knows about it */
	Window window;		/* who owns current X selection */
} selection_info;
selection_info Selection;

/*************************************************************************
function: RegisterSelection - Notify X that something is in xfer bin
*/
RegisterSelection()
{
	Window win;

	win = HostWinFromCurWin();
	XSetSelectionOwner(Dpy,XA_PRIMARY,win,TimeOfButtonPress);
	XSync(Dpy,FALSE);	/* just to be safe! */
	if ( XGetSelectionOwner(Dpy,XA_PRIMARY) == win ) {
		Selection.registered = TRUE;
		Selection.window = win;
	}
	else {
		Selection.registered = FALSE;
		Selection.window = None;
	}
}

DeRegisterSelection()
{
	if ( Selection.registered ) {
		XSetSelectionOwner(Dpy,XA_PRIMARY,None,CurrentTime);
		Selection.registered = FALSE;
	}
}

static char *XferFunction;	/* Elf func to call when X-xfer becomes available */

/*************************************************************************
function: ElfbFetchAndRun - Get register xfer function
*/
ElfbFetchAndRun (name)
char *name;
{
	Window win;

	if ( !XferFunction )
		XferFunction = malloc(STRPATH+1);
	strcpy(XferFunction,name);
	win = HostWinFromCurWin();
	XConvertSelection(Dpy,XA_PRIMARY,XA_STRING,
							XA_STRING,win,TimeOfButtonPress);
}

/*************************************************************************
function: ACUpdateXferState - Check for external selection for pasting
*/
ACUpdateXferState (xferStateP)
XferDesc *xferStateP;
{
	Window win;

	win = XGetSelectionOwner(Dpy,XA_PRIMARY);
	if ( win == Selection.window ) {
		Selection.window = None;
		return(FALSE);
	}
	if ( win == None ) {
		ACxferClearBin();
		Selection.window = None;
		return(FALSE);
	}
	return(TRUE);	/* X has something */
}

#define XCLOSE(x) CloseFileObject(x)
#define XLSEEK(a,b,c) DocSegSeek(a,b,c)
#define read DocHardRead
#define MAX_COLS 256

/*************************************************************************
function: ReadAscii - Read ascii file from xfer bin
*/
typeErr ReadAscii (filename,buf,bufsize,got_bytes)
char *filename;	/* name of file in XferDoc */
char *buf;		/* result */
int bufsize;	/* max chars in buffer */
int *got_bytes;	/* Returns bufsize or -1 for failure */
{
	int fd;

	*got_bytes = 0;
	fd = (int)OpenFileObjFmode(ACxferDoc(),filename,"r");
	if ( fd <= 0 )
		return(ErrnoErr());
	*got_bytes = DocSegRead(fd,buf,bufsize,-1);
	CloseFileObject(fd);
	return(AOK);
}

/*************************************************************************
function: ReadUDFToString - Reads UDF strings into passed memory buffer
*/
typeErr ReadUDFToString (filename,buf,bufsize,got_bytes)
char *filename;	/* name of file in XferDoc */
char *buf;		/* result */
int bufsize;	/* max chars in buffer */
int *got_bytes;	/* Returns bufsize or -1 for failure */
{
	register short *colnumP, *offsetsP;
	register unsigned char *lengthsP;
	int NumRecords, recnum, len;
	long pos;
    int UDFfd;          /* file descriptor for open udf file */
    short HdrCols;      /* How many are in the header's cols array */
    short RecSize;      /* size of each UDF record in chars */
    long LocColDef;     /* where in the UDF file the coldefs are */
    long LocRngTbl;     /* where in the UDF file the range table is */
    long LocRecords;        /* where the records are */
    int SkipCol0;       /* TRUE if col0 is to be IMP formatted */
	short col_array[MAX_COLS];
	unsigned char col_lengths[MAX_COLS];
	short col_offsets[MAX_COLS];
	struct impMD hdr;			/* metadata header */
	struct tblRange rngtab;
	char rcd[MAX_XSTR];
	int cols_visible;
	struct colDefinition coldef;
	int col;

	*got_bytes = 0;
	UDFfd = (int)OpenFileObjFmode(ACxferDoc(),filename,"r");
	if ( UDFfd <= 0 )
		return(ErrnoErr());

/* read the udf header and set the easy vars */

	read(UDFfd,(char*)&hdr,sizeof(hdr));
	HdrCols = hdr.colCount;		/* gets fixed for hidden stuff in CountCols */
	RecSize = hdr.recSize;
	LocColDef = hdr.locColDef;
	LocRngTbl = hdr.locRangeTbl;
	SkipCol0 = !hdr.fmtMark;

	if ( (RecSize >= MAX_XSTR) || (hdr.rngCount != 1) ) {
		XCLOSE(UDFfd);
		return(BADXFERFILE);
	}
   
/* Collect the per-column information */

	cols_visible = 0;
	XLSEEK(UDFfd,(long)LocColDef,0);
	for ( col = 0; col < HdrCols; col++ ) {
		read(UDFfd,(char*)&coldef,sizeof(coldef));
		if ( coldef.hidColumn )
			continue;
		if ( (col == 0) && SkipCol0 )
			continue;
		col_array[cols_visible] = col;
		col_lengths[cols_visible] = coldef.lenColumn;
		col_offsets[cols_visible] = coldef.ofsColumn;
		cols_visible++;
	}
	col_array[cols_visible] = -1;

/* Get the Range Table info - where records are and how many */

	XLSEEK(UDFfd,(long)LocRngTbl,0);
	read(UDFfd,(char*)&rngtab,sizeof(rngtab));
	NumRecords = (rngtab.rngEnd - rngtab.rngStart + 1);
	LocRecords = rngtab.rngOffset;

/* read the data into buf */

	pos = LocRecords;
	XLSEEK(UDFfd,pos,0);
	for ( recnum = 0; recnum < NumRecords; recnum++ ) {
		read(UDFfd,rcd,RecSize);
		offsetsP = col_offsets;
		lengthsP = col_lengths;
		for ( colnumP = col_array; *colnumP != -1; colnumP++,lengthsP++ ) {
			len = *lengthsP;
			if ( *got_bytes + len >= bufsize )
				return(AOK);		/* partial copy better than nothing */
			memcpy(buf,&rcd[*offsetsP++],len & 255);
			*got_bytes += len;
			buf += len;
		}
		*buf++ = '\n';
		*got_bytes += 1;
	}
	XCLOSE(UDFfd);
	return(AOK);
}

