/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tm1.h
   Applix ASC version: 22
   Date: Sat Jan 23 21:15:05 1999
*/

/******************************************/
/* Formats for Controlling Browser Layout */
/******************************************/

/* IMPORTANT!!!:  Note that various C utility functions used by the UI
   depend upon the structures of the datatype in this section.  If you
   want to modify any of these, please either:
   1.  Add new fields on to the *end* of the structures, or
   2.  Modify the TM1_Browz_info$ parsing C procedures in tm1ui.c
*/
format TM1_Browz_info$
	cube_name,
	format tm1_view_info$ view 'the current view

format tm1_view_info$     'an array of these for each cube
	view_name,        'may be null if "(Unnamed)"
	format tm1_layout_info$ layout,
	format tm1_option_info$ opts

format tm1_layout_info$
	format arrayof tm1_dim_info$ rows,
	format arrayof tm1_dim_info$ columns,
	format arrayof tm1_dim_info$ unallocd

format tm1_dim_info$
	dim_name,
	list_of_all_elements,
	format tm1_subset_info$ subset,
	format tm1_subset_opinfo$ opinfo

format tm1_subset_opinfo$
	sstaskid,
	exit_op_was_cancel,
	in_advanced_mode

/* Note if a subset_name is given, the elements may be null (forcing a read
   of the subset file), or may be a list of names.  If subset_name is not
   given, but a list of elements is, the subset is assumed to be "(Unnamed)".
   If neither a name nor a list of elements is given, the subset is presumed
   to be "ALL".  "subset_single_element" is the *name* of the element
   selected if the dimension is currently located in the static layout
   area. */
format tm1_subset_info$   'an array of these for each dimension
	subset_name,      'may be null if "(Unnamed)"
	subset_elements,  'may be null if a named subset is given
        subset_single_element

format tm1_option_info$
	suppress_zeroes,         'boolean
	show_automatically,      'boolean
	show_cell_borders,       'boolean
	show_row_col_headings,   'boolean
	num_format,              'number specifying format in file ???.h
	num_precision,           'for numeric formatting
	min_col_width,
        format rt_attrs_ realtime_attrs

/*************************/
/* Query Data Structures */
/*************************/

format tm1_query_info$        'an array of these for each cube
	query_name,
	cube_name,
	format arrayof tm1_subset_info$ dim_subsets,
	operator_number,      'see "Query Operators Enumeration", below
	skip_computed_values, 'boolean
	skip_zero_values,     'boolean
	range_a_val,
	range_b_val,
 	format tm1_query_opinfo$ opinfo

format tm1_query_opinfo$
	sstaskid,
	exit_op_was_cancel,
	index_of_leftout_dim

/**********************/
/* Option Enumeration */
/**********************/

#define TM1#OPT_SUPPRESS_ZEROES            1
#define TM1#OPT_AUTO_SHOW                  2
#define TM1#OPT_SHOW_CELL_BORDERS          3
#define TM1#OPT_SHOW_ROW_COL_HEADINGS      4
#define TM1#OPT_NUMBER_FORMAT              5
#define TM1#OPT_NUMBER_FORMAT_INFO         6
#define TM1#OPT_MIN_COL_WIDTH              7

#define TM1#OPT_DEFAULT_MIN_COL_WIDTH_VAL     9

/*****************************/
/* Subset String Definitions */
/*****************************/

#define TM1#STR_SUBSET_ALL     XLT("ALL")
#define TM1#STR_SUBSET_UNNAMED XLT("(Unnamed)")

/******************/
/* Number Formats */
/******************/

#define TM1#OPT_NUMFMT_UNSTYLED  0
#define TM1#OPT_NUMFMT_GENERAL   2
#define TM1#OPT_NUMFMT_FIXED     3
#define TM1#OPT_NUMFMT_SCIENCE   4

/*******************************/
/* Query Operators Enumeration */
/*  and Corresponding Strings  */
/*******************************/

/* these parallel the enumeration in TM1's "db_qry.h" file */
#define TM1#OPT_QRY_BLANK     0
#define TM1#OPT_QRY_EQ        1
#define TM1#OPT_QRY_GE        2
#define TM1#OPT_QRY_LE        3
#define TM1#OPT_QRY_GT        4
#define TM1#OPT_QRY_LT        5
#define TM1#OPT_QRY_NE        6
#define TM1#OPT_QRY_GEA_LEB   7
#define TM1#OPT_QRY_GTA_LTB   8

#define TM1#STR_QRYOP_BLANK     TM1#STR_SUBSET_ALL
#define TM1#STR_QRYOP_EQ        NXLT("= a")
#define TM1#STR_QRYOP_GE        NXLT(">= a")
#define TM1#STR_QRYOP_LE        NXLT("<= a")
#define TM1#STR_QRYOP_GT        NXLT("> a")
#define TM1#STR_QRYOP_LT        NXLT("< a")
#define TM1#STR_QRYOP_NE        NXLT("~= a")
#define TM1#STR_QRYOP_GEA_LEB   NXLT(">= a & <= b")
#define TM1#STR_QRYOP_GTA_LTB   NXLT("> a & < b")

#define TM1#STR_QRY_OPS {TM1#STR_QRYOP_BLANK, TM1#STR_QRYOP_EQ, TM1#STR_QRYOP_GE, TM1#STR_QRYOP_LE, TM1#STR_QRYOP_GT, TM1#STR_QRYOP_LT, TM1#STR_QRYOP_NE, TM1#STR_QRYOP_GEA_LEB, TM1#STR_QRYOP_GTA_LTB }

/***********************************************************/
/* Stuff for Managing Dialogs Implemented as Builder Tasks */
/***********************************************************/

format tm1_ss_task_list_info$
	ss_task,
	browse_task,
	subset_task_can_query,
	subset_task_no_query,
	query_task

#define TM1#BROWSE_TASK              1
#define TM1#SUBSET_TASK_CAN_QUERY    2
#define TM1#SUBSET_TASK_NO_QUERY     3
#define TM1#QUERY_TASK               4

#define TM1#TASK_MESSAGE_PROMOTE_WINDOW          473
#define TM1#TASK_MESSAGE_SUBSET_CHANGED          474
#define TM1#TASK_MESSAGE_UPDATE_QUERY_SUBSETS    475

/*****************/
/* Other Formats */
/*****************/

format tm1_formula_structure
	ss_task_id,
	cell_addr,
	value_type,
	formula_type

format tm1_connect_list
	servername,
	username,
        password

#define TM1#CURRENT#CUBE	"TM1CurrentTable"
#define TM1#FORMULA	1001		'poke
#define TM1#NUMERIC	1
#define TM1#STRING	2
#define TM1#AUTOMATIC 3

#define TM1_SUBSET_DLG_ADV_MODE        "TM1SubsetAdvMode"
#define TM1_POLL_CUBE_CHANGES          "TM1PollCube"

format tm1_server_info$
	server_name,
	username,
	password,
	num_users,
	ss_task_id


#define TM1_ADMIN_FILE_NAME NXLT("tm1.adm")
#define TM1_DATA_DIR_SYS_VAR NXLT("tm1_data_dir@")
#define TM1_ADMIN_DIR_SYS_VAR NXLT("tm1_admin_dir@")
#define TM1_HOME_DIR_SYS_VAR NXLT("tm1_home_dir@")
#define TM1_CUBE_DIMENSIONS_LIST_FOR_BROWSE NXLT("TM1_cube_dim_list@")
#define TM1_SERVER_LIST	NXLT("TM1ServerList@")
#define TM1_USERNAME	NXLT("TM1UserName@")
#define TM1_PASSWORD	NXLT("TM1Password@")
#define TM1_CONNECT_STRING	NXLT("TM1ConnString@")
#define TM1_INSERT_ELEMENT_NAME	NXLT("TM1InsertElementName@")
#define TM1_INSERT_CUBE_NAME	NXLT("TM1InsertCubeName@")
#define TM1_INSERT_DIMENSION_NAME	NXLT("TM1InsertDimensionName@")
#define TM1_KILL_BUTTON_TASK	NXLT("TM1killButtonTask$")
#define TM1_FORMULA_STRUCT	NXLT("TM1FormulaStruct")
#define TM1_SUBSET_INFO_LIST NXLT("TM1_subset_dim@")
#define TM1_QUERY_INFO_LIST NXLT("TM1_query_list@")
#define SS_TM1_TASKLIST	NXLT("TM1_ss_task_list@")
#define TM1_BROWSE_UPDATE_DATA NXLT("TM1BrowseUpdateData@")

#define SINGLE_SELECT   1
#define MULTI_SELECT    2

/*****************/
/* File Suffixes */
/*****************/

#define TM1_FILE_SUFFIX_ASCII_VIEW    NXLT(".tvu")
#define TM1_FILE_SUFFIX_ASCII_QUERY   NXLT(".tqu")
#define TM1_FILE_SUFFIX_ASCII_SUBSET  NXLT(".sub")

#define TM1_FILE_SUFFIX_ASCII_DIM     NXLT(".dit")

/* obsolete and should be removed */
format dimension_member_tree$
     member_name,
     level_number,
     generation_number,
     has_kids,
     entry_list_idx,
     format arrayof dimension_member_tree$ next_level

format dimension_info$
	names,
	format arrayof dimension_member_tree$ trees

format label_stuff
	dim_name,
	elem_name

#define DLG#LABEL_ 5
#define DLG#BUTTON_ 3

/**************************************/
/* Stuff for Handling Caching Subsets */
/**************************************/

/* note that the various entries are actually calculated in the C
   code in "tm1ui.c".  Any changes to this data structure *must* be
   duplicated in that code */
format tm1_subset_canvas_entry$
	member_name,
	level_number,
	start_point_y,
	is_selected,
	was_selected,       'used for controlled mouse selection
	has_kids_displayed, 'used for double-click expand/collapse
	first_kid_idx,     'used for double-click expand/collapse
	last_kid_idx,      'used for double-click expand/collapse
	format gr_area@ boxarea, 'x,y,width,height
	total_width,
	line_array,	'used for drawing vertical hierarchy -- parallels
				' the box_x_pos_by_level array
	level_of_hline	'used for drawing horizontal hierarchy
