/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: tm1sbst.c
   Applix ASC version: 2
   Date: Mon Nov 16 14:27:45 1998
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#define AX_USE_PROTOS 1
#include <aos.h>
#include <elfapi.h>
#include <tm1ui.h>
#include <tm1sbst.h>

/***************/
/* Definitions */
/***************/

#define FALSE 0
#define TRUE  1

#define MAX_DIMNAME_CHARS      8    /* TM1 constant */
#define MAX_SVRNAME_CHARS      100  /* arbitrary */
#define MAX_DIM_AND_SVR_CHARS  (MAX_DIMNAME_CHARS + MAX_SVRNAME_CHARS + 1)

#define MAX_SUBSET_NAME_LEN    150  /* arbitrary */
#define MAX_CACHED_SUBSETS     20   /* arbitrary */

#define TM1_SUBSET_UNNAMED_NAME  "(Unnamed)"  /* TM1 constant */

/**************/
/* Data Types */
/**************/

typedef struct dimension_information {
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  data_ptr canonical_list_of_element_names;
  data_ptr parallel_list_of_levels;
  data_ptr parallel_list_of_parents;
  data_ptr parallel_list_of_children;
} dimension_information;

typedef struct cached_subset_info {
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  data_ptr element_list;
  data_ptr layout_info;
  dimension_information *dim_info; /* this pointer must be valid (non-NULL) */
  int max_total_width;
  int max_total_height;
} cached_subset_info;

/********************/
/* Global Variables */
/********************/

cached_subset_info cached_subsets[MAX_CACHED_SUBSETS];
int indexes_of_most_recently_used_subsets[MAX_CACHED_SUBSETS]; /* used for caching algm */
int num_current_active_subsets = 0;

dimension_information stored_dimension_info[MAX_CACHED_SUBSETS+1];
int parallel_array_of_num_pointers[MAX_CACHED_SUBSETS+1];
int num_current_active_dimensions = 0;

/**********************************/
/* Prototypes of Local Procedures */
/**********************************/

int tm1subset_name_maps_to_all(char *subset_name);   /* IN */
int tm1subset_find_subset_data(char *dimname_with_server,   /* IN */
			       char *subset_name);          /* IN */
void tm1subset_move_this_subset_index_to_top(int idx); /* IN */
void tm1subset_add_dimension_reference_by_index(int idx);/* IN */
void tm1subset_subtract_dimension_reference_by_index(int idx); /* IN */
int tm1subset_find_next_free_dim_index();
void tm1subset_record_initial_dimension_reference(int diminfo_idx); /* IN */
int tm1subset_retrieve_dimension_info(char *dimname_with_server); /* IN */
int tm1subset_find_next_free_subset_index();
data_ptr tm1subset_retrieve_dimension_element_list(
			  char *dimname_with_server); /* IN */
data_ptr tm1subset_find_layout_info(char *dimname_with_server, 
				   char *subset_name);
void tm1subset_fetch_all_parallel_lists(char *dimname, 
					dimension_information *diminfo);
int tm1subset_get_stored_subset_index(char *dimname_with_server,
				      char *subset_name);
data_ptr tm1subset_find_layout_info_entry_by_index(
				   char *dimname_with_server, 
				   char *subset_name,
				   int entry_number);
int tm1subset_parse_dimname_and_subname(data_ptr funcData, /* IN */
					char *dimname,     /* OUT */
					char *subname);    /* OUT */
int tm1subset_parse_dimname(data_ptr funcData, /* IN */
			    char *dimname);    /* OUT */
int tm1subset_parse_dimname_subname_and_int(data_ptr funcData, /* IN */
					    char *dimname,     /* OUT */
					    char *subname,     /* OUT */
					    int *out_ival);    /* OUT */
int tm1subset_parse_dimname_subname_and_elfData(data_ptr funcData, /* IN */
						char *dimname,     /* OUT */
						char *subname,     /* OUT */
						data_ptr *out_dp);  /* OUT */

/***************************/
/* Input-Parsing Functions */
/***************************/

/**************************************************************************
 * tm1subset_parse_dimname - parses the first element of the array
 *     into the "dimname" string
 */
int
tm1subset_parse_dimname(data_ptr funcData, /* IN */
			char *dimname)     /* OUT */
{
  char *cptr;

  /* make sure we have an array of at least one element1 */
  if (! AxIsArray(funcData)) {
    return(FALSE);
  }
  if (AxArraySize(funcData) < 1) {
    return(FALSE);
  }

  /* parse the first element into the dimension name */
  if (! AxIsString(AxArrayElement(funcData, 0))) {
    return(FALSE);
  }
  cptr = AxStrPtrFromArray(funcData, 0);
  strcpy(dimname, cptr);
  return(TRUE);
} /* tm1subset_parse_dimname */

/**************************************************************************
 * tm1subset_parse_dimname_and_subname - takes the input data_ptr from
 *    the util$$ call, and parses the first two array arguments into
 *    the provided character strings for the dimension and subset names
 */
int
tm1subset_parse_dimname_and_subname(data_ptr funcData, /* IN */
				    char *dimname,     /* OUT */
				    char *subname)     /* OUT */
{
  char *cptr;
  int rval;

  rval = tm1subset_parse_dimname(funcData, dimname);
  if (! rval) {
    return(FALSE);
  }

  /* make sure we have an array of at least two elements */
  if (AxArraySize(funcData) < 2) {
    return(FALSE);
  }

  /* parse the first element into the dimension name */
  if (! AxIsString(AxArrayElement(funcData, 0))) {
    return(FALSE);
  }
  cptr = AxStrPtrFromArray(funcData, 0);
  strcpy(dimname, cptr);

  /* parse the second element into the subset name */
  if (! AxIsString(AxArrayElement(funcData, 1))) {
    return(FALSE);
  }
  cptr = AxStrPtrFromArray(funcData, 1);
  strcpy(subname, cptr);

  /* return the "okay" flag */
  return(TRUE);
} /* tm1subset_parse_dimname_and_subname */

/**************************************************************************
 * tm1subset_parse_dimname_subname_and_elfData - parses the first tow
 *     elements fo funcData into the dimension and subset names, and
 *     parses the third into the output data_ptr value
 */
int
tm1subset_parse_dimname_subname_and_elfData(data_ptr funcData, /* IN */
					    char *dimname,     /* OUT */
					    char *subname,     /* OUT */
					    data_ptr *out_dp)  /* OUT */
{
  int rval;

  /* parse the first two arguments */
  rval = tm1subset_parse_dimname_and_subname(funcData, dimname, subname);
  if (! rval) {
    return(FALSE);
  }

  /* parse the third argument */
  if (AxArraySize(funcData) < 3) {
    return(FALSE);
  }
  *out_dp = AxArrayElement(funcData, 2);
  return(TRUE);
} /* tm1subset_parse_dimname_subname_and_elfData */
 
/**************************************************************************
 * tm1subset_parse_dimname_subname_and_int - parses the first two
 *     elements of funcData into the dimension and subset names, and
 *     parses the third into the integer output value
 */
int
tm1subset_parse_dimname_subname_and_int(data_ptr funcData, /* IN */
					char *dimname,     /* OUT */
					char *subname,     /* OUT */
					int *out_ival)     /* OUT */
{
  int rval;
  elfData third_arg;

  /* parse the first two arguments */
  rval = tm1subset_parse_dimname_subname_and_elfData(funcData,
		     dimname,  subname, &third_arg);
  if (! rval) {
    return(FALSE);
  }

  if (! AxIsInt(third_arg)) {
    return(FALSE);
  }
  *out_ival = AxIntFromDataPtr(third_arg);
  return(TRUE);
} /* tm1subset_parse_dimname_subname_and_int */
 
/************************************/
/* Functions for Storing/Retrieving */
/*     max_total_{width_height}     */
/************************************/

/*
The functions in this section store/retrieve the
max_total_{width,height} portion of the cached subset definition.
*/

/*****************************************************
 * tm1subset_retrieve_subset_max_width
 */
data_ptr
tm1subset_retrieve_subset_max_width(data_ptr funcData)
{ 
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  int subset_idx, max_width, rval;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_and_subname(funcData, dimname, subset_name);
  if (! rval) {
    return(NULL);
  }

  subset_idx = tm1subset_find_subset_data(dimname, subset_name);
  max_width = cached_subsets[subset_idx].max_total_width;
  return(AxMakeIntData(max_width));
} /* tm1subset_retrieve_subset_max_width */

/*****************************************************
 * tm1subset_retrieve_subset_max_height
 */
data_ptr
tm1subset_retrieve_subset_max_height(data_ptr funcData)
{
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  int subset_idx, max_height, rval;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_and_subname(funcData, dimname, subset_name);
  if (! rval) {
    return(NULL);
  }

  subset_idx = tm1subset_find_subset_data(dimname, subset_name);
  max_height = cached_subsets[subset_idx].max_total_height;
  return(AxMakeIntData(max_height));
} /* tm1subset_retrieve_subset_max_height */

/*****************************************************
 * tm1subset_store_subset_max_width
 */
data_ptr
tm1subset_store_subset_max_width(data_ptr funcData)
{
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  int subset_idx, max_width, rval;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_subname_and_int(funcData, dimname,
			 subset_name, &max_width);
  if (! rval) {
    return(NULL);
  }

  subset_idx = tm1subset_find_subset_data(dimname, subset_name);
  cached_subsets[subset_idx].max_total_width = max_width;
  return(NULL);
} /* tm1subset_store_subset_max_width */

/*****************************************************
 * tm1subset_store_subset_max_height
 */
data_ptr
tm1subset_store_subset_max_height(data_ptr funcData)
{
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  int subset_idx, max_height, rval;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_subname_and_int(funcData, dimname,
			 subset_name, &max_height);
  if (! rval) {
    return(NULL);
  }

  subset_idx = tm1subset_find_subset_data(dimname, subset_name);
  cached_subsets[subset_idx].max_total_height = max_height;
  return(NULL);
} /* tm1subset_store_subset_max_height */

/****************************************/
/* Functions for Retrieving Subset Info */
/*        to the data_ptr Layer         */
/****************************************/

/**************************************************************************
 * tm1subset_get_subset_elements - retrieves the list of elements
 *      associated with the given subset in the given dimension
 */
data_ptr
tm1subset_get_subset_elements(data_ptr funcData)
{
  data_ptr rData;
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  int rval;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_and_subname(funcData, dimname, subset_name);
  if (! rval) {
    return(NULL);
  }
  rData = tm1subset_get_pointer_to_subset_elements(dimname, subset_name);
  return(AxCopyData(rData));
} /* tm1subset_get_subset_elements */

/**************************************************************************
 * tm1subset_get_pointer_to_subset_elements
 */
data_ptr
tm1subset_get_pointer_to_subset_elements(char *dimname, 
					 char *subset_name)
{
  data_ptr rData;
  int subset_idx;

  subset_idx = tm1subset_find_subset_data(dimname, subset_name);
  rData = cached_subsets[subset_idx].element_list;
  return(rData);
} /* tm1subset_get_pointer_to_subset_elements */

/**************************************************************************
 * tm1subset_get_subset_size - retrieves the number of elements in the
 *      given subset in the given dimension
 */
data_ptr
tm1subset_get_subset_size(data_ptr funcData)
{
  data_ptr rData;
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  int subset_idx, subset_size, rval;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_and_subname(funcData, dimname, subset_name);
  if (! rval) {
    return(NULL);
  }

  subset_idx = tm1subset_find_subset_data(dimname, subset_name);
  rData = cached_subsets[subset_idx].element_list;
  subset_size = AxArraySize(rData);
  return(AxMakeIntData(subset_size));
} /* tm1subset_get_subset_size */

/**************************************************************************
 * tm1subset_get_subset_element_by_index - for the given subset in the
 *     given dimension, retrieves the element specified by the given
 *     index (one-based); if the index is out-of-bounds, returns ""
 */
data_ptr
tm1subset_get_subset_element_by_index(data_ptr funcData)
{
  data_ptr rData, elemlist;
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  char *cptr;
  int subset_idx, elem_idx, rval;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_subname_and_int(funcData, dimname,
			 subset_name, &elem_idx);
  if (! rval) {
    return(NULL);
  }

  subset_idx = tm1subset_find_subset_data(dimname, subset_name);
  elemlist = cached_subsets[subset_idx].element_list;
  if ((elem_idx <= 0) || (AxArraySize(elemlist) < elem_idx)) {
    return(AxMakeStrData(-1, ""));
  }
  cptr = AxStrPtrFromArray(elemlist, elem_idx-1);
  rData = AxMakeStrData(-1, cptr);
  return(rData);
} /* tm1subset_get_subset_element_by_index */

/**********************************/
/*   Functions for Storing the    */
/* Subset Dialog's Canvas Layout  */
/**********************************/

/**************************************************************************
 * tm1subset_set_subset_layout - for the given subset in the given
 *     dimension, stores the provided elfData array with the cached
 *     subset information
 */
data_ptr
tm1subset_set_subset_layout(data_ptr funcData)
{
  data_ptr rData, layout_info;
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  int rval, subset_idx;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_subname_and_elfData(funcData,
	     dimname, subset_name, &layout_info);
  if (! rval) {
    return(NULL);
  }

  subset_idx =
    tm1subset_find_subset_data(dimname, subset_name);
  cached_subsets[subset_idx].layout_info = 
    (data_ptr)AxTaskCopyData(0, layout_info);

  return(NULL);
} /* tm1subset_set_subset_layout */

/**********************************/
/*  Functions for Retrieving the  */
/* Subset Dialog's Canvas Layout  */
/**********************************/

/**************************************************************************
 * tm1subset_get_subset_layout
 */
data_ptr
tm1subset_get_subset_layout(data_ptr funcData)
{
  data_ptr rData;
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  int rval;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_and_subname(funcData, dimname, subset_name);
  if (! rval) {
    return(NULL);
  }

  rData = tm1subset_find_layout_info(dimname, subset_name);
  return(AxCopyData(rData));
} /* tm1subset_get_subset_layout */

/**************************************************************************
 * tm1subset_find_layout_info_entry_by_index - returns the actual
 *     data_ptr array corresponding to the tm1_subset_canvas_entry$
 *     found in the layout information for the given entry_number
 *     (zero-based) in the layout_info data_ptr array; warning:  do
 *     *not* return this value to the elfData layer -- use
 *     "tm1subset_get_layout_info_entry_by_index"
 *                ^^^
 */
data_ptr
tm1subset_find_layout_info_entry_by_index(char *dimname_with_server, 
					  char *subset_name,
					  int entry_number)
{
  data_ptr layout_info, entry;

  layout_info = tm1subset_find_layout_info(dimname_with_server,
					   subset_name);
  if (NULL == layout_info) {
    return(NULL);
  }
  if (! AxIsArray(layout_info)) {
    return(NULL);
  }
  if (AxArraySize(layout_info) <= entry_number) {
    return(NULL);
  }
  entry = AxArrayElement(layout_info, entry_number);
  return(entry);
} /* tm1subset_find_layout_info_entry_by_index */

/**************************************************************************
 * tm1subset_get_layout_info_entry_by_index - returns a copy of the
 *     data_ptr array corresponding to the tm1_subset_canvas_entry$
 *     found in the layout information for the given entry_number
 *     (zero-based) in the layout_info data_ptr array
 */
data_ptr
tm1subset_get_layout_info_entry_by_index(data_ptr funcData)
{
  data_ptr entry;
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  int subset_idx, rval, entry_number;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_subname_and_int(funcData, dimname,
			 subset_name, &entry_number);
  if (! rval) {
    return(NULL);
  }

  subset_idx = tm1subset_find_subset_data(dimname, subset_name);
  entry = tm1subset_find_layout_info_entry_by_index(dimname,
					  subset_name, entry_number);
  return(AxCopyData(entry));
} /* tm1subset_get_layout_info_entry_by_index */

/**************************************************************************
 * tm1subset_find_layout_info - returns the actual data_ptr array
 *    array stored in the subset cache; if this data or any
 *    constituent part of it is returned to the Elf layer, it must
 *    first be copied and the copy returned; do *not* return the
 *    actual, stored data_ptr array, as this will be freed by the Elf
 *    layer when the calling macro terminates, and will leave the
 *    cached subset structure with an invalid pointer
 */
data_ptr
tm1subset_find_layout_info(char *dimname_with_server, 
			   char *subset_name)
{
  int subset_idx;
  data_ptr layout_info;

  subset_idx =
    tm1subset_find_subset_data(dimname_with_server, subset_name);
  layout_info = cached_subsets[subset_idx].layout_info;
  return(layout_info);
} /* tm1subset_find_layout_info */

/***********************************************************/
/*                Functions For Getting the                */
/* Various Lists that Parallel the Canonical Elements List */
/***********************************************************/

/**************************************************************************
 * tm1subset_get_level_list
 */
data_ptr
tm1subset_get_level_list(data_ptr funcData)
{
  data_ptr levlist;
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  int rval;

  rval = tm1subset_parse_dimname(funcData, dimname);
  if (! rval) {
    return(NULL);
  }
  levlist = tm1subset_get_pointer_to_level_list(dimname);
  return(AxCopyData(levlist));
} /* tm1subset_get_level_list */

/**************************************************************************
 * tm1subset_get_pointer_to_level_list
 */
data_ptr
tm1subset_get_pointer_to_level_list(char *dimname)
{
  int subset_idx;
  dimension_information *diminfo;
  data_ptr levlist;

  subset_idx = tm1subset_find_subset_data(dimname,
					   TM1_SUBSET_ALL_NAME);
  diminfo = cached_subsets[subset_idx].dim_info;
  levlist = diminfo->parallel_list_of_levels;
  if (NULL == levlist) {
    tm1subset_fetch_all_parallel_lists(dimname, diminfo);
    levlist = diminfo->parallel_list_of_levels;
  }
  return(levlist);
} /* tm1subset_get_pointer_to_level_list */

/**************************************************************************
 * tm1subset_get_parent_list
 */
data_ptr
tm1subset_get_parent_list(data_ptr funcData)
{
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  data_ptr parentlist;
  int rval;

  rval = tm1subset_parse_dimname(funcData, dimname);
  if (! rval) {
    return(NULL);
  }

  parentlist = tm1subset_get_pointer_to_parent_list(dimname);
  return(AxCopyData(parentlist));
} /* tm1subset_get_parent_list */

/**************************************************************************
 * tm1subset_get_pointer_to_parent_list
 */
data_ptr
tm1subset_get_pointer_to_parent_list(char *dimname)
{
  data_ptr rData, parentlist;
  int subset_idx;
  dimension_information *diminfo;

  subset_idx = tm1subset_find_subset_data(dimname,
					   TM1_SUBSET_ALL_NAME);
  diminfo = cached_subsets[subset_idx].dim_info;
  parentlist = diminfo->parallel_list_of_parents;
  if (NULL == parentlist) {
    tm1subset_fetch_all_parallel_lists(dimname, diminfo);
    parentlist = diminfo->parallel_list_of_parents;
  }
  return(parentlist);
} /* tm1subset_get_pointer_to_parent_list */

/**************************************************************************
 * tm1subset_get_children_list
 */
data_ptr
tm1subset_get_children_list(data_ptr funcData)
{
  data_ptr rData, kidlist, elemlists;
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  int subset_idx, rval;
  dimension_information *diminfo;

  rval = tm1subset_parse_dimname(funcData, dimname);
  if (! rval) {
    return(NULL);
  }

  subset_idx = tm1subset_find_subset_data(dimname,
					   TM1_SUBSET_ALL_NAME);
  diminfo = cached_subsets[subset_idx].dim_info;
  kidlist = diminfo->parallel_list_of_children;
  if (NULL == kidlist) {
    tm1subset_fetch_all_parallel_lists(dimname, diminfo);
    kidlist = diminfo->parallel_list_of_children;
  }
  return(AxCopyData(kidlist));
} /* tm1subset_get_children_list */

/**************************************************************************
 * tm1subset_fetch_all_parallel_lists
 */
void
tm1subset_fetch_all_parallel_lists(char *dimname, 
				   dimension_information *diminfo)
{
  int rval, has_null_return_value;
  data_ptr rlists, elemlist, levlist, parentlist, kidlist;

  rval =
    ElfCallByName("tm_all_element_hierarchy_info_must_call_gateway$",
		  &rlists, 1, AxMakeStrData(-1, dimname));
  has_null_return_value = ((NULL == rlists) ||
			   (AxIsArray(rlists) &&
			    (0 == AxArraySize(rlists))));
  if (! has_null_return_value) {
    elemlist = (data_ptr)AxTaskCopyData(0,AxArrayElement(rlists, 0));
    levlist = (data_ptr)AxTaskCopyData(0,AxArrayElement(rlists, 1));
    parentlist = (data_ptr)AxTaskCopyData(0,AxArrayElement(rlists, 2));
    kidlist = (data_ptr)AxTaskCopyData(0,AxArrayElement(rlists, 3));
    diminfo->canonical_list_of_element_names = elemlist;
    diminfo->parallel_list_of_levels = levlist;
    diminfo->parallel_list_of_parents = parentlist;
    diminfo->parallel_list_of_children = kidlist;
  }
} /* tm1subset_fetch_all_parallel_lists */

/*******************************************/
/* Functions for Adding/Replacing/Removing */
/*        Entire Subset Definitions        */
/*******************************************/

/**************************************************************************
 * tm1subset_add_new_subset - for the given subset in the given
 *      dimension, add an entry (or replace an existing entry) in the
 *      subset cache -- NULL out any associated layout information
 */
data_ptr
tm1subset_add_new_subset(data_ptr funcData)
{
  data_ptr rData, elemlist, new_elemlist;
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  int subset_idx, rval;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_subname_and_elfData(funcData,
	     dimname, subset_name, &elemlist);
  if (! rval) {
    return(NULL);
  }
  
  subset_idx = tm1subset_get_stored_subset_index(dimname, subset_name);
  new_elemlist = (data_ptr)AxTaskCopyData(0, elemlist);
  if (-1 == subset_idx) {
    subset_idx = tm1subset_add_new_subset_value(dimname,
	  subset_name, new_elemlist);
  } else {
    cached_subsets[subset_idx].element_list = new_elemlist;
    cached_subsets[subset_idx].layout_info = NULL;
    cached_subsets[subset_idx].max_total_width = 0;
    cached_subsets[subset_idx].max_total_height = 0;
  }

  return(NULL);  
} /* tm1subset_add_new_subset */

/**************************************************************************
 * tm1subset_remove_subset_info - remove the information for the given
 *      subset in the given dimension from the subset cache
 */
data_ptr
tm1subset_remove_subset_info(data_ptr funcData)
{
  data_ptr rData, layout_info;
  char dimname[MAX_DIM_AND_SVR_CHARS+1];
  char subset_name[MAX_SUBSET_NAME_LEN+1];
  int subset_idx, ix, jx, rval, dimidx;

  /* parse the arguments */
  rval = tm1subset_parse_dimname_and_subname(funcData, dimname, subset_name);
  if (! rval) {
    return(NULL);
  }

  /* if the subset is no longer stored, then we're done */
  subset_idx = tm1subset_get_stored_subset_index(dimname, subset_name);
  if (-1 == subset_idx) {
    return(NULL);
  }

  /* since we're removing the subset, we need to decrement the 
     corresponding dimension reference count */
  dimidx = tm1subset_get_stored_dimension_information(
	      (cached_subsets[subset_idx].dim_info)->dimname);
  tm1subset_subtract_dimension_reference_by_index(dimidx);

  /* release any allocated memory */
  if ((NULL != cached_subsets[subset_idx].element_list) &&
      (0 == strcasecmp(subset_name, TM1_SUBSET_ALL_NAME))) {
    AxTaskFreeData(0,cached_subsets[subset_idx].element_list);
  }
  if (NULL != cached_subsets[subset_idx].layout_info) {
    AxTaskFreeData(0,cached_subsets[subset_idx].layout_info);
  }

  /* move the stuff that is lower down in the array upwards */
  for (ix=subset_idx; (ix < num_current_active_subsets); ix++) {
    memcpy(&(cached_subsets[ix]), &(cached_subsets[ix+1]),
	   sizeof(cached_subset_info));
  }

  /* record that we've deleted a subset */
  num_current_active_subsets--;

  /* adjust the "most recent use" array */
  for (ix=0; (ix <= num_current_active_subsets); ix++) {
    if (indexes_of_most_recently_used_subsets[ix] == subset_idx) {
      for (jx=ix; (jx < num_current_active_subsets); jx++) {
	indexes_of_most_recently_used_subsets[jx] =
	  indexes_of_most_recently_used_subsets[jx+1];
	if (indexes_of_most_recently_used_subsets[jx] > subset_idx) {
	  indexes_of_most_recently_used_subsets[jx] -= 1;
	}
      }
      return(NULL);
    } else if (indexes_of_most_recently_used_subsets[ix] > subset_idx) {
      indexes_of_most_recently_used_subsets[ix] -= 1;
    }
  }

  /* return */
  return(NULL);
} /* tm1subset_remove_subset_info */

/**************************************/
/* Miscellaneous Interface Procedures */
/**************************************/

/**************************************************************************
 * tm1subset_clear_subset_caches
 */
data_ptr
tm1subset_clear_subset_caches()
{
  int ix;

  for (ix=0; (ix < num_current_active_subsets); ix++) {
    if (NULL != cached_subsets[ix].element_list) {
      /* if this is subset "ALL", then the element list is just a
         pointer to the element list in the dimension, so it will be
	 freed there */
      if (0 != strcasecmp(cached_subsets[ix].subset_name,
			  TM1_SUBSET_ALL_NAME)) {
	AxTaskFreeData(0,cached_subsets[ix].element_list);
      }
    }
    if (NULL != cached_subsets[ix].layout_info) {
      AxTaskFreeData(0,cached_subsets[ix].layout_info);
    }
  }

  for (ix=0; (ix < num_current_active_dimensions); ix++) {
    if (NULL !=
	stored_dimension_info[ix].canonical_list_of_element_names) {
      AxTaskFreeData(0,stored_dimension_info[ix].canonical_list_of_element_names);
    }
    if (NULL != stored_dimension_info[ix].parallel_list_of_levels) {
      AxTaskFreeData(0,stored_dimension_info[ix].parallel_list_of_levels);
    }
    if (NULL != stored_dimension_info[ix].parallel_list_of_parents) {
      AxTaskFreeData(0,stored_dimension_info[ix].parallel_list_of_parents);
    }
    if (NULL != stored_dimension_info[ix].parallel_list_of_children) {
      AxTaskFreeData(0,stored_dimension_info[ix].parallel_list_of_children);
    }
  }

  num_current_active_subsets = 0;
  num_current_active_dimensions = 0;

  return(NULL);
} /* tm1subset_clear_subset_caches */

/************************************/
/* MAIN SUBSET RETRIEVAL PROCEDURES */
/************************************/

/**************************************************************************
 * tm1subset_get_stored_subset_index - retrieves the index in the
 *    subset cache at which the entry for the given subset in the
 *    given dimension can be found; returns -1 if the subset is not
 *    currently cached
 */
int
tm1subset_get_stored_subset_index(char *dimname_with_server,
				  char *subset_name)
{
  int ix;
  dimension_information *dinfo;

  for (ix=0; (ix < num_current_active_subsets); ix++) {
    if (0 == strcasecmp(subset_name, cached_subsets[ix].subset_name)) {
      dinfo = cached_subsets[ix].dim_info;
      if (0 == strcasecmp(dinfo->dimname, dimname_with_server)) {
	return(ix);
      }
    }
  }
  return(-1);
} /* tm1subset_get_stored_subset_index */

/**************************************************************************
 * tm1subset_find_subset_data - finds the index in the
 *    "cached_subsets" data structure at which the information for the
 *    specified subset is cached; if the information is not already in
 *    the subset cache, the procedure retrieves it, stores it in the
 *    cache, then returns the index of the newly stored material; note
 *    that this procedure also handles cache replacement/management
 */
int
tm1subset_find_subset_data(char *dimname_with_server, /* IN */
			   char *subset_name)         /* IN */
{
  int ix, rval, diminfo_idx, has_null_return_value;
  data_ptr elemlist, new_elemlist;

  /* if we're looking for something that will default to the subset
     "ALL", then look for the "ALL" subset in the cache explicitly */
  if (tm1subset_name_maps_to_all(subset_name)) {
    subset_name = TM1_SUBSET_ALL_NAME;
  }

  /* check to see if the subset is currently in the cached subset array */
  ix = tm1subset_get_stored_subset_index(dimname_with_server, subset_name);
  if (-1 != ix) {
    /* note here, that since we're accessing the data, we want to
       record that fact, so that the caching algorithm caches out
       the *least* recently used data */
    tm1subset_move_this_subset_index_to_top(ix);
    return(ix);
  }

  /* if the subset is not currently in the cached subset array, and if
     we're not looking for the special subsets "ALL" or "(Unnamed)",
     try reading its elements from the file */ 
  if ((0 != strcasecmp(subset_name, TM1_SUBSET_ALL_NAME)) &&
      (0 != strcasecmp(subset_name, TM1_SUBSET_UNNAMED_NAME))) {
    rval = ElfCallByName("tm1_read_subset_file_for_element_list$", &elemlist, 2,
			 AxMakeStrData(-1, dimname_with_server),
			 AxMakeStrData(-1, subset_name));
    has_null_return_value = ((NULL == elemlist) ||
			     (AxIsArray(elemlist) &&
			      (0 == AxArraySize(elemlist))));
    if (! has_null_return_value) {
      new_elemlist = (data_ptr)AxTaskCopyData(0, elemlist);
      ix = tm1subset_add_new_subset_value(dimname_with_server, 
				  subset_name, new_elemlist);
      return(ix);
    }
    /* else, the subset is not defined, so default to the subset
       "ALL" case */
    return(tm1subset_find_subset_data(dimname_with_server,
				      TM1_SUBSET_ALL_NAME));
  }

  /* case of subset name "ALL":
     we don't currently have this subset cached, so go find the dimension
     information, and create a subset entry for the subset "ALL" of
     that dimension */
  elemlist = tm1subset_retrieve_dimension_element_list(dimname_with_server);
  ix = tm1subset_add_new_subset_value(dimname_with_server, 
				      subset_name, elemlist);
  return(ix);
} /* tm1subset_find_subset_data */

/************************************/
/* MAIN SUBSET ALLOCATION PROCEDURE */
/************************************/

/**************************************************************************
 * tm1subset_add_new_subset_value - creates an entry in the
 *     "cached_subsets" array which will store the values provided;
 *     note that this procedures also adjusts the "most recently used"
 *     array to reflect the "use" of this new subset
 */
int
tm1subset_add_new_subset_value(char *dimname_with_server, 
			       char *subset_name,
			       data_ptr elemlist)
{
  int next_free, dimidx;

  /* find where to put the new subset */
  next_free = tm1subset_find_next_free_subset_index();

  /* fill in the data structure */
  strcpy(cached_subsets[next_free].subset_name, subset_name);
  cached_subsets[next_free].element_list = elemlist;
  cached_subsets[next_free].layout_info = NULL;
  cached_subsets[next_free].max_total_width = 0;
  cached_subsets[next_free].max_total_height = 0;
  /* dimidx is guaranteed to be valid */
  dimidx = tm1subset_retrieve_dimension_info(dimname_with_server);
  cached_subsets[next_free].dim_info = &(stored_dimension_info[dimidx]);
  /* up the reference counter for the corresponding dimension */
  tm1subset_add_dimension_reference_by_index(dimidx);

  /* indicate that this subset was the last one looked at */
  tm1subset_move_this_subset_index_to_top(next_free);

  /* return the index of the subset */
  return(next_free);
} /* tm1subset_add_new_subset_value */

/*******************************************************/
/* Functions For Managing the Subset Caching Algorithm */
/*******************************************************/

/**************************************************************************
 * tm1subset_move_this_subset_index_to_top - the index refers to the
 *    location in the "cached_subsets" array; this procedure moves
 *    that index to the top of the "most recently used" stack
 */
void
tm1subset_move_this_subset_index_to_top(int idx) /* IN */
{
  int ix;
  int where_idx = -1;

  for (ix=0; (ix < num_current_active_subsets); ix++) {
    if (idx == indexes_of_most_recently_used_subsets[ix]) {
      where_idx = idx;
      break;
    }
  }
  /* if it's already at the top, then we're done */
  if (0 == where_idx) {
    return;
  }
  /* if the index isn't currently in the index array, then we can
     pretend that it's at the bottom of the array */
  if (-1 == where_idx) {
    where_idx = num_current_active_subsets;
  }
  /* once we've found where the index currently is, we shift 
     everything between the top of the array and that index down one,
     then put the index at the top of the array */
  for (ix=where_idx; (ix >= 1); ix--) {
    indexes_of_most_recently_used_subsets[ix] =
      indexes_of_most_recently_used_subsets[ix-1];
  }
  indexes_of_most_recently_used_subsets[0] = idx;
} /* tm1subset_move_this_subset_index_to_top */

/**************************************************************************
 * tm1subset_find_next_free_subset_index - returns the index in the
 *    "cached_subsets" array at which the next subset should be
 *    cached
 */
int
tm1subset_find_next_free_subset_index()
{
  int ix, replace_idx, dimidx;

  /* if we still have space to allocate, just return the next
     available */
  if (num_current_active_subsets < MAX_CACHED_SUBSETS) {
    return(num_current_active_subsets++);
  }

  /* else, take the index at the bottom of the "recent use" stack */
  replace_idx =
    indexes_of_most_recently_used_subsets[MAX_CACHED_SUBSETS];

  /* clear out the corresponding memory and adjust the dimension
     reference counter accordingly */
  dimidx = tm1subset_get_stored_dimension_information(
	      (cached_subsets[replace_idx].dim_info)->dimname);
  tm1subset_subtract_dimension_reference_by_index(dimidx);

  /* shift the rest of the indices down one */
  for (ix=MAX_CACHED_SUBSETS-1; (ix >= 1); ix--) {
    indexes_of_most_recently_used_subsets[ix] =
      indexes_of_most_recently_used_subsets[ix-1];
  }

  /* place this index at the top */
  indexes_of_most_recently_used_subsets[0] = replace_idx;

  /* return the index */
  return(replace_idx);
} /* tm1subset_find_next_free_subset_index */

/******************************************************/
/* Functions For Handling Dimension Caching Algorithm */
/******************************************************/

/**************************************************************************
 * tm1subset_add_dimension_reference_by_index
 */
void
tm1subset_add_dimension_reference_by_index(int idx) /* IN */
{
  if (-1 == idx) {
    return;
  }
  parallel_array_of_num_pointers[idx] += 1;
} /* tm1subset_add_dimension_reference_by_index */

/**************************************************************************
 * tm1subset_subtract_dimension_reference_by_index
 */
void
tm1subset_subtract_dimension_reference_by_index(int idx) /* IN */
{
  if (-1 == idx) {
    return;
  }
  parallel_array_of_num_pointers[idx] -= 1;
} /* tm1subset_subtract_dimension_reference_by_index */

/**************************************************************************
 * tm1subset_find_next_free_dim_index
 */
int
tm1subset_find_next_free_dim_index()
{
  int ix, dimidx;
  dimension_information *diminfo;
  char *dimname;

  if (num_current_active_dimensions < (MAX_CACHED_SUBSETS+1)) {
    ix = num_current_active_dimensions++;
    parallel_array_of_num_pointers[ix] = 0;
    return(ix);
  }
  for (ix=0; (ix < (MAX_CACHED_SUBSETS+1)); ix++) {
    if (0 == parallel_array_of_num_pointers[ix]) {
      return(ix);
    }
  }

  /* this should never happen -- something has gotten confused, so
     recompute the reference counter array */
  for (ix=0; (ix < (MAX_CACHED_SUBSETS+1)); ix++) {
    parallel_array_of_num_pointers[ix] = 0;
  }
  for (ix=0; (ix < num_current_active_subsets); ix++) {
    diminfo = cached_subsets[ix].dim_info;
    dimname = diminfo->dimname;
    dimidx = tm1subset_get_stored_dimension_information(dimname);
    if (-1 != dimidx) {
      parallel_array_of_num_pointers[dimidx] += 1;
    }
  }
  for (ix=0; (ix < (MAX_CACHED_SUBSETS+1)); ix++) {
    if (0 == parallel_array_of_num_pointers[ix]) {
      return(ix);
    }
  }
  /* THIS SHOULD *REALLY* NEVER HAPPEN!!!!!! */
  return(MAX_CACHED_SUBSETS);
} /* tm1subset_find_next_free_dim_index */

/***********************************/
/* Miscellaneous Utility Functions */
/***********************************/

/**************************************************************************
 * tm1subset_name_maps_to_all
 */
int
tm1subset_name_maps_to_all(char *subset_name) /* IN */
{
  if (NULL == subset_name) {
    return(TRUE);
  }
  if (0 == strlen(subset_name)) {
    return(TRUE);
  }
  if (0 == strcasecmp(subset_name, TM1_SUBSET_ALL_NAME)) {
    return(TRUE);
  }
  return(FALSE);
} /* tm1subset_name_maps_to_all */

/********************************/
/* Functions for Mucking Around */ 
/*  with Basic Dimension Info   */
/********************************/

/**************************************************************************
 * tm1subset_get_stored_dimension_information - retrieves the index at
 *     which information for the given dimension is stored; returns -1
 *     if the information is not currently loaded into the memory cache
 */
int
tm1subset_get_stored_dimension_information(char *dimname_with_server)
{
  int ix;

  for (ix=0; (ix < num_current_active_dimensions); ix++) {
    if (0 == strcasecmp(dimname_with_server,
			stored_dimension_info[ix].dimname)) {
      return(ix);
    }
  }
  return(-1);
} /* tm1subset_get_stored_dimension_information */

/**************************************************************************
 * tm1subset_retrieve_dimension_info - looks in the dimension
 *     information cache first, to see if the information is cached;
 *     if so, it returns the index at which the information is stored;
 *     if not, it creates an entry and returns the index; note that
 *     if this is a *new* entry, nothing but the dimension name will
 *     be filled in
 */
int
tm1subset_retrieve_dimension_info(char *dimname_with_server) /* IN */
{
  dimension_information *dinfo;
  int diminfo_idx, rval;

  /* check to see if we've cached the information somewhere */
  diminfo_idx = tm1subset_get_stored_dimension_information(dimname_with_server);
  if (-1 != diminfo_idx) {
    return(diminfo_idx);
  }

  diminfo_idx = tm1subset_find_next_free_dim_index();

  strcpy(stored_dimension_info[diminfo_idx].dimname, dimname_with_server);
  stored_dimension_info[diminfo_idx].canonical_list_of_element_names = NULL;
  stored_dimension_info[diminfo_idx].parallel_list_of_levels = NULL;
  stored_dimension_info[diminfo_idx].parallel_list_of_parents = NULL;
  stored_dimension_info[diminfo_idx].parallel_list_of_children = NULL;

  return(diminfo_idx);
} /* tm1subset_retrieve_dimension_info */

/**************************************************************************
 * tm1subset_retrieve_dimension_element_list -- returns the data_ptr
 *    stored in the appropriate cached dimension information -- *not*
 *    a copy of that data_ptr
 */
data_ptr
tm1subset_retrieve_dimension_element_list(char *dimname_with_server) /* IN */
{
  int rval, diminfo_idx;
  data_ptr elemlist, new_elemlist;

  /* find the cached dimension information */
  diminfo_idx =
    tm1subset_retrieve_dimension_info(dimname_with_server);
  elemlist =
    stored_dimension_info[diminfo_idx].canonical_list_of_element_names;

  /* if necessary, retrieve the element list and store it away */
  if (NULL == elemlist) {
    rval = ElfCallByName("tm_elemlist_must_call_gateway$", &elemlist, 1,
			 AxMakeStrData(-1, dimname_with_server));
    new_elemlist = (data_ptr)AxTaskCopyData(0,elemlist);
    stored_dimension_info[diminfo_idx].canonical_list_of_element_names
      = new_elemlist;
    return(new_elemlist);
  }

  /* return the element list */
  return(elemlist);
} /* tm1subset_retrieve_dimension_element_list */
