/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: head_off.c
   Applix ASC version: 4
   Date: Wed Nov 18 16:29:20 1998
*/
#include <stdio.h>

#ifdef AXMS
#define  ENV_USER       "USERNAME"
#define READ_BINARY_STR "rb"
#define WRITE_TRUNC_STR "wb+"
#else
#define READ_BINARY_STR "r"
#define WRITE_TRUNC_STR "w"
#define  ENV_USER       "USER"
#endif
 


#define SIZE_ASC_BINARY_HEADER 700 
#define BINARY_HEADER_STRING "BINARY~HEADER~SPECIAL~STRING\n"
#define SIZE_HEADER_STRING strlen(BINARY_HEADER_STRING)


#define IO_BUFF_SIZE 4096

main(argc,argv)
int argc;
char *argv[];
{
	FILE *rfp, *wfp;
	char *infile, *outfile;
	char buf[IO_BUFF_SIZE];
	unsigned short checksum = 0;
	int bytecounter = 0;
	char *p;
    char header_buffer[SIZE_ASC_BINARY_HEADER];
    int bytes_read, bytes_written;

	infile = argv[2];
	outfile = argv[1];

/* Open input file */

	rfp = fopen(infile,READ_BINARY_STR);
	if ( !rfp ) {
		fprintf(stderr,"Could not open input file %s\n",infile);
		exit(1);
	}

/* Open output file */

    if ( strcmp( outfile, "-" ) == 0 )
      /* write to standard out */
      wfp = stdout;
    else
       {
	   wfp = fopen(outfile,WRITE_TRUNC_STR);
	   if ( !wfp ) 
           {
		   fprintf(stderr,"Could not open output file %s\n",outfile);
		   exit(1);
	       }
       }

    /***********************************************************************/
    /* Read the file, check the first block for the header, and if it is   */
    /*  there, write out the bulk of the file, without the header          */
    /***********************************************************************/
    bytes_read = fread( buf, 1, IO_BUFF_SIZE, rfp );


/* the "-1" below is because of an earlier bug which placed the tag string */
/* 1 byte before the actual end of the buffer.  The bug has been fixed     */
/* but to maintain compatibility with old stuff in source control we need  */
/* to compensate here                                                      */
    p = buf + SIZE_ASC_BINARY_HEADER - SIZE_HEADER_STRING - 1;
    if ( bytes_read >= SIZE_ASC_BINARY_HEADER && 
		 strncmp( p, BINARY_HEADER_STRING, SIZE_HEADER_STRING ) == 0 )
	{
		/** The header is there, don't write it out */
		fwrite( (buf+SIZE_ASC_BINARY_HEADER), 1, 
				(bytes_read-SIZE_ASC_BINARY_HEADER), wfp );
		bytes_read = fread( buf, 1, IO_BUFF_SIZE, rfp );
	}
	while ( bytes_read >0 ) 
        {
        bytes_written = fwrite( buf, 1, bytes_read, wfp );
		if ( bytes_written != bytes_read ) 
            goto errout;
        bytes_read = fread( buf, 1, IO_BUFF_SIZE, rfp );
	    }

	fclose(rfp);
	fclose(wfp);
	exit(0);

errout:
	fprintf(stderr,"Write failed on %s\n",outfile);
	exit(1);
}

