/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: acpytre.c
   Applix ASC version: 7
   Date: Mon Nov 16 13:57:57 1998
*/
#include <utloc.h>
#include <getstr9.h>
#include <sys/types.h>
#include <sys/stat.h>


#if AXMS
#define SLASH_CHAR '\\'
#else
#define SLASH_CHAR '/'
#endif

/*
.BEGIN MANPAGE
.NAME AcopyTree - Copy a directory tree.
.DESCRIPTION
	AcopyTree() copies a specified directory to specified destination.
	This routine emulates the Unix command "cp -r".
.ARGUMENT acpy_source
	The address of a string which identifies the source directory.
.ARGUMENT acpy_target
	The address of a string which identifies the destination directory.
.RETURN VALUE
	AcopyTree() returns 0 if it succeeds.
.SEE ALSO
	AcopyFile(3A)
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
typeErr AcopyTree(acpy_source,acpy_target)
char *acpy_source;
char *acpy_target;
/* END SYNOPSIS */
{
	int level = 0,start_dir;
	char *place;
	char src_path[STRPATH + 1],dst_path[STRPATH + 1];
	char fname[STRNORMAL + 1],name_for_stat[STRPATH + 1];
	char dst_name[STRPATH + 1];
	struct stat file_info;

	strcpy(name_for_stat,acpy_source);
	if(Aaccess(acpy_target,0))
		strcpy(dst_name,acpy_target);
	else {
		if(place = rindex(name_for_stat,SLASH_CHAR))
			place++;
		else
			place = name_for_stat;
		sprintf(dst_name,"%s%c%s",acpy_target,SLASH_CHAR,place);
	}
	place = NULL;

get_info:
	if(Astat(name_for_stat,&file_info))
		return(S9COPYTREE);
	if(file_info.st_mode&S_IFDIR) {
		start_dir = TRUE;
		level++;
		if(Amkdir(dst_name))
			return(S9COPYTREE);
		strcpy(src_path,name_for_stat);
		strcpy(dst_path,dst_name);
	}
	else {
		start_dir = FALSE;
		if(AcopyFile(name_for_stat,dst_name))
			return(S9COPYTREE);
	}

	while(level) {
		if(AgetNfilename(start_dir,src_path,fname,&place)) {
			sprintf(name_for_stat,"%s%c%s",src_path,SLASH_CHAR,fname);
			sprintf(dst_name,"%s%c%s",dst_path,SLASH_CHAR,fname);
			goto get_info;
		}
		if(--level) {
			start_dir = FALSE;
			*rindex(src_path,SLASH_CHAR) = '\0';
			*rindex(dst_path,SLASH_CHAR) = '\0';
		}
	}

	return(AOK);
}


