/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: adelf.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:57:58 1998
*/
#include <utloc.h>
#include <getstr9.h>

#if SYS&VMSSYS
struct del_file_tag {
	struct del_file_tag *next;
	char name[1];           /* Actually variable length. */
};
static int add_to_list();
static struct del_file_tag *del_list;
#endif

/*
.BEGIN MANPAGE
.NAME AdelFile - Delete files.
.DESCRIPTION
	AdelFile() deletes the possibly wildcarded source files adlf_paths.
.ARGUMENT adlf_paths
	The address of a string which identifies at least one file.
	Adlf_paths may identify multiple files by using wildcards and
	multiple pathnames separated by spaces or tabs.  "*" and "?" are
	legal wildcards.
.RETURN VALUE
	AdelFile() returns 0.
.SEE ALSO
	Armdir(3A)
.END MANPAGE
*/
/* BEGIN SYNOPSIS */
typeErr AdelFile(adlf_paths)
char *adlf_paths;
/* END SYNOPSIS */
{
	if(strpbrk(adlf_paths," \t?*")) {
#if SYS&AEGISO
    	char paths[2 * STRPATH];

    	str_aegis_wild(paths,adlf_paths);
    	pgm_system("dlf",3,"-f","-nq",paths);
#else    
#if SYS&VMSSYS
		struct del_file_tag *next;

/*
 * First make a list of files to delete.
 */
		del_list = NULL;

		shell$to_vms(adlf_paths,add_to_list,1);

/*
 * Now delete the files.
 * This foolishness is required because shell$to_vms malfunctions
 * when used recursively and delete calls it.
 */
		while(del_list) {
			delete(del_list->name);
			next = del_list->next;
			free(del_list);
			del_list = next;
		}
#else
    	char adlf_cmd[STRPATH + 19]; 

    	sprintf(adlf_cmd,"rm -f %s 2>/dev/null",adlf_paths);
    	system(adlf_cmd);
#endif
#endif
		return(AOK);
	}

	unlink(adlf_paths);
	return(AOK);
}

#if SYS&VMSSYS
/* For a wildcarded file specification, this is */
/* called once for each matching file name. */
/* Stick them all in a linked list. */

globalvalue int SHELL$K_FILE;

static int add_to_list(fname,type)
char *fname;
int type;
{
	struct del_file_tag *new;
	
	/* The rm shell command does not delete directories. */
	if(type == SHELL$K_FILE) {
		new = (struct del_file_tag *)
			malloc(sizeof(struct del_file_tag) + strlen(fname));
		new->next = del_list;
		strcpy(new->name,fname);
		del_list = new;
	}
	return(1);
}
#endif
