/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: amktemp.c
   Applix ASC version: 10
   Date: Tue Jan 19 21:10:52 1999
*/
#include <errno.h>
#include <aos.h>

#if AXMS
#include <process.h>
#define MAXUNIQUE    8
#define PIDLENGTH    4
#else
#define MAXUNIQUE    14
#define PIDLENGTH    5
#endif


char *Amktemp(in_path)
char *in_path;
{
	static char pidString[PIDLENGTH + 1];
    static int init_flag = 0;
    static int starting_offset = 0;
    register char *firstP, *lastP, *leftP;
    int num_percents, pid_fits;

	errno = EBADF;	/* failure assumption */

	if(!init_flag) 
        {
		sprintf(pidString,"%0*ld",PIDLENGTH,getpid() & 0x7fff);
	    }
	
	if( strlen(in_path) > STRPATH)
		return(NULL);

    lastP = strrchr( in_path, '%' );
    if ( lastP == (char *)NULL )
        /* no percents */
        return(NULL);

    /* find the first percent */
    firstP = lastP;
    while ( *firstP == '%' )
        {
        firstP--;
        }
    firstP++;   

    num_percents = lastP - firstP + 1;

    /* copy as much of the pid string as will fit */
    pid_fits = PIDLENGTH;
    if ( num_percents < PIDLENGTH )
       pid_fits =  num_percents;
    strncpy( firstP, pidString, pid_fits );

    /* fill the other percents with 'a' chars to start with */
    leftP = firstP + pid_fits;
    while ( leftP <= lastP )
        { 
        if ( leftP == lastP )
           {
           /* we start out with an offset in the starting character, */
           /* because some code relies on the fact that it can create a name,*/
           /* delete that file, create another name, and not get the first   */
           /* name.  By specifying an offset, we get to go around 26 times   */
           /* before name duplication.                                       */
           *leftP = 'a' + starting_offset;
           starting_offset++;
           if ( starting_offset >= 26 )
               starting_offset = 0;
           }
        else
           *leftP = 'a';
        leftP++;
        }

    /* we now have the basic starting point for the temp name creation */
    /* so we can cycle through, starting at the right end, looking for */
    /* a file wich does not exist.  We create names by treating the percent */
    /* range as a "number" in base-26, and incremnent until we get to  */
    /* "zzzzzzzz"                                                      */
    if ( *lastP < 'a' )
       {
       /* we are starting in the PID digits, start with 'a' */
       *lastP = 'a';
       }
    *lastP = *lastP - 1;   /* so loop can increment to start */

    while ( TRUE )
        {
        *lastP = *lastP + 1;

        if ( *lastP > 'z' )
           {
           /* carry a one to the position to the left */
           *lastP = 'a';
           leftP = lastP - 1;
           while ( TRUE )
              {
              /* continue to carry 1 to the left */
              *leftP = *leftP + 1;
              if ( *leftP < 'a' )
                 {
                 /* we have crossed over into the PID digits, start with 'a' */
                 *leftP = 'a';
                 }
              if ( *leftP <= 'z' )
                 break;
              if ( *leftP > 'z' )
                 {
                 *leftP = 'a';
                 leftP--;
                 if ( leftP < firstP )
                     {
                     /* we incremented the first char of the percent range */
                     /* past a 'z'.  This means we have tried them all!    */
                     memset( firstP, '%', num_percents ); /* restore original */
                     return( (char *)NULL );
                     }
                 }
              }
           }
        /* see if the name just created exists */
	    if( access(in_path, 0) ) 
		{
			if(errno != ENOENT)
			{
				/* an error on the acces, but not "file does not exist" */
				/* something seeems to be seriously wrong, bail out     */
				return(NULL);
			}
			/* file does not exit, we're all done */
			errno = 0;	/* in case anybody looking for zero errno */
			return( in_path );
		}
    }
}
