/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: amove.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:57:58 1998
*/
#include <utloc.h>
#include <getstr9.h>

#if SYS&VMSSYS
#include <stat.h>
struct move_file_tag {
	struct move_file_tag *next;
	char name[1];
};
static int add_to_list();
static struct move_file_tag *move_list;
#endif

#if SYS&AEGISO
extern long pgm_sys_stat;
#endif

/*
.BEGIN MANPAGE
.NAME Amove - Move files.
.DESCRIPTION
	Amove() moves specified files to a specified destination.
.ARGUMENT amve_source
	The address of a string which identifies at least one file or
	directory.  Amve_source may identify multiple files or directories
	by using wildcards and multiple pathnames separated by spaces or tabs.
	"*" and "?" are legal wildcards.
.ARGUMENT amve_target
	The address of a string which identifies a destination directory.
    If amve_source identifies a single file, amve_target may identify a
    file or directory.
.SEE ALSO
	Arename(3A)
.NOTES
	Amove() can not move directories across disk partitions. 
.RETURN VALUE
	Amove() returns 0 if it succeeds.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
typeErr Amove(amve_source, amve_target)
char *amve_source;
char *amve_target;
/* END SYNOPSIS */
{
#if SYS&VMSSYS
    char buffer[STRPATH + 1];
	struct move_file_tag *next;
    char *p;
	struct stat lstat;

    if(strpbrk(amve_source," \t?*")) {

		strcpy(buffer,amve_source);
		move_list = NULL;
		p = buffer;

		while(p = strtok(p," \t")) {
			shell$to_vms(p,add_to_list,1);
			p = NULL;
		}

		if(move_list && move_list->next &&
			(Astat(amve_target,&lstat) || !(lstat.st_mode&S_IFDIR)))
			return(S9MOVE);

        while(move_list) {
            Arename(move_list->name,amve_target);
            next = move_list->next;
            free((char *)move_list);
            move_list = next;
        }

        return(AOK);
    }

	if(Arename(amve_source,amve_target))
		return(S9MOVE);

	return(AOK);
#else
#if SYS&AEGISO
    char sources[1000];
    char target[256];
	char *sp;

	sp = sources;
	*sp++ = '(';
	*sp++ = ' ';
    str_aegis_wild(sp,amve_source);
	sp += strlen(sp);
	*sp++ = ' ';
	*sp++ = ')';
	*sp++ = '\0';
    str_aegis_wild(target,amve_target);

    if(pgm_system("mvf",4,"-r","-nq",sources,target))
        return(S9MOVE);

   	return(AOK);

#else    

    char  amve_cmd[2 * STRPATH + 4]; 

	if(!strpbrk(amve_source," \t*?")) {
		if(Arename(amve_source,amve_target))
			return(S9MOVE);
		return(AOK);
	}

	sprintf(amve_cmd,"mv %s %s 2>/dev/null",amve_source,amve_target);

    if(system(amve_cmd))     
        return(S9MOVE);

	return(AOK);

#endif
#endif
}

#if SYS&VMSSYS
/* For a wildcarded file specification, this is */
/* called once for each matching file name. */
/* Stick them all in a linked list. */
 
globalvalue int SHELL$K_FILE;
 
static int add_to_list(fname,type)
char *fname;
int type;
{    
    struct move_file_tag *new;
 
    if(type == SHELL$K_FILE) {
        new = (struct move_file_tag *)
            malloc(sizeof(struct move_file_tag) + strlen(fname));
        new->next = move_list;
        strcpy(new->name,fname);
        move_list = new;
    }
    return(1);
}    
#endif
