/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: aregcmp.c
   Applix ASC version: 5
   Date: Fri Jan  8 21:11:43 1999
*/
#include <utloc.h>
#include <aregexp.h>
#include <varargs.h>

static char **ut_sp;
static char **ut_stmax;
static long ut_i_size;

static char *ut_rpop();
static ut_rpush(/* char *ptr */);

/*
.BEGIN MANPAGE
.NAME Aregcmp - Compile a regular expression.
.SYNOPSIS
char *Aregcmp(string1[, string2, ...], (char *)0)
char *string1,*string2, ...
.DESCRIPTION
	Aregcmp() pprovides a generic interface to compile a regular expression.
.ARGUMENT string1
	The address of a regular expression string to be compiled.
	Additional addresses of strings may be presented for compilation.
	The list of addresses must be terminated by a 0 address.
.SEE ALSO
	Aregex(3A) regcmp(3) ed(1) free(3) malloc(3)
.RETURN VALUE
	Aregcmp() returns a pointer to the compiled form.  Malloc() is used to
	create space for the vector.  It is the caller's responsibility to
	free unneeded space so allocated.  A zero return indicates an 
	incorrect argument.
.END MANPAGE
*/
/* BEGIN SYNOPSIS */
/* END SYNOPSIS */
char *Aregcmp(va_alist)
va_dcl
{
	va_list ap;
	register int c;
	register char *ep,*sp;
	char **adx;
	int i,cflg;
	char *lastep,*sep,*eptr;
	int nbra,ngrp;
	int cclcnt;
	char *stack[SSIZE];
	char *args[SSIZE];
 
	va_start(ap);
	i = 0;
	while(args[i++] = va_arg(ap,char *));
	va_end(ap);

	ut_sp = stack;
	*ut_sp = NULL;
	ut_stmax = &stack[SSIZE];
 
	adx = args;
	i = nbra = ngrp = 0;
	while(*adx)
		i += strlen(*adx++) + 1;
	adx = args;
	sp = *adx++;
	if(!(sep = ep = malloc((unsigned)(2 * i + SLOP))))
		return(ep);
	if(!(c = *sp++))
		goto cerror;
	if(c=='^') {
		c = *sp++;
		*ep++ = CIRCFL;
	}
	if(c == '*' || c == '+' || c == '{')
		goto cerror;
	sp--;
	while(1) {
		if(!(c = *sp++)) {
			if(*adx) {
				sp = (char *)*adx++;
				continue;
			}
			*ep++ = CEOF;
			if(--nbra > NBRA || *ut_sp)
				goto cerror;
			ut_i_size = ep - sep;
			return(sep);
		}
		if(c != '*' && c != '{' && c != '+')
			lastep = ep;
		switch(c) {
		  case '(':
			if(!ut_rpush(ep))
				goto cerror;
			*ep++ = CBRA;
			*ep++ = 0xff;
			continue;
		  case ')':
			if(!(eptr = ut_rpop()))
				goto cerror;
			if((c = *sp++) == '$') {
				if('0' > (c = *sp++) || c > '9')
					goto cerror;
				*ep++ = CKET;
				*ep++ = *++eptr = nbra++;
				*ep++ = c - '0';
				continue;
			}
			*ep++ = EGRP;
			*ep++ = ngrp++;
			sp--;
			switch(c) {
			  case '+':
				*eptr = PGRP;
				break;
			  case '*':
				*eptr = SGRP;
				break;
			  case '{':
				*eptr = TGRP;
				break;
			  default:
				*eptr = GRP;
				continue;
			}
			i = ep - eptr - 2;
			for(cclcnt = 0; i >= 256; cclcnt++)
				i -= 256;
			if(cclcnt > 3)
				goto cerror;
			*eptr |= cclcnt;
			*++eptr = i;
			continue;
 
		  case '\\':
			*ep++ = CCHR;
			if(!(c = *sp++))
				goto cerror;
			*ep++ = c;
			continue;
 
		  case '{':
			*lastep |= RNGE;
			cflg = 0;
		  nlim:
			if((c = *sp++) == '}')
				goto cerror;
			i = 0;
			do {
				if('0' <= c && c <= '9')
					i = i * 10 + (c - '0');
				else
					goto cerror;
			} while((c = *sp++) != '}' && c != ',');
			if(i > 255)
				goto cerror;
			*ep++ = i;
			if(c == ',') {
				if(cflg++)
					goto cerror;
				if((c = *sp++) == '}') {
					*ep++ = 0xff;
					continue;
				}
				sp--;
				goto nlim;
			}
			if(!cflg)
				*ep++ = i;
			else if((ep[-1] & 0xff) < (ep[-2] & 0xff))
				goto cerror;
			continue;
 
		  case '.':
			*ep++ = CDOT;
			continue;
 
		  case '+':
			if(*lastep == CBRA || *lastep == CKET)
				goto cerror;
			*lastep |= PLUS;
			continue;
 
		  case '*':
			if(*lastep == CBRA || *lastep == CKET)
				goto cerror;
			*lastep |= STAR;
			continue;
 
		  case '$':
			if(*sp || *adx)
				goto defchar;
			*ep++ = CDOL;
			continue;
 
		  case '[':
			*ep++ = CCL;
			*ep++ = 0;
			cclcnt = 1;
			if((c = *sp++) == '^') {
				c = *sp++;
				ep[-2] = NCCL;
			}
			do {
				if(!c)
					goto cerror;
				if(c == '-' && cclcnt > 1 && *sp != ']') {
					c = ep[-1];
					ep[-1] = MINUS;
					*ep++ = c;
					cclcnt++;
					continue;
				}
				*ep++ = c;
				cclcnt++;
			} while((c = *sp++) != ']');
			lastep[1] = cclcnt;
			continue;
 
		  defchar:
		  default:
			*ep++ = CCHR;
			*ep++ = c;
		}
	}
cerror:
	free(sep);
	return((char *)NULL);
}

static char *ut_rpop()
{
	return(*ut_sp--);
}

static ut_rpush(ptr)
char *ptr;
{
	if(++ut_sp > ut_stmax)
		return(0);
	*ut_sp = ptr;
	return(1);
}
