/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: astrncm.c
   Applix ASC version: 5
   Date: Mon Nov 16 13:57:58 1998
*/
#include <ctype.h>
#include <utloc.h>
#include <wchars.h>


/*
.BEGIN MANPAGE
.NAME Astrncmp - Case-insensitive limited size string compare.
.ARGUMENT str1
    The address of a string.
.ARGUMENT str2
    The address of a string.
.ARGUMENT n
	The maximum number of characters to compare
.DESCRIPTION
    Similar to strncmp() but treats the same alphabetic
    character of different cases as the same character.
.SEE ALSO
    strncmp(3)
.RETURN VALUE
    Astrncmp() returns an integer greater than, equal to, or less than 0,
    accordingly as str1 is lexicographically greater than, equal to, or
    less than str2 within n characters.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
Astrncmp(str1,str2,n)
REGISTER char *str1;
REGISTER char *str2;
REGISTER int n;
/* END SYNOPSIS */
{
	REGISTER int char1,char2;
	
	if(str1 == str2)
		return(0);

	while(n-- &&
		((char1 = *str1++ & 0xff) != 0) & ((char2 = *str2++ & 0xff) != 0)) {
		char1 = Atoupper(char1);
		char2 = Atoupper(char2);
		if(char1 != char2) 
			break;
	}

	return((n < 0) ? 0 : (char1 - char2));
}





#if BUILD_WIDE   /*  }  */
AWstrncmp(str1,str2,n)
REGISTER wchar *str1;
REGISTER wchar *str2;
REGISTER int n;
/* END SYNOPSIS */
{
	REGISTER int char1,char2;
	
	if(str1 == str2)
		return(0);

	while(n-- &&
		((char1 = *str1++ & 0xff) != 0) & ((char2 = *str2++ & 0xff) != 0)) {
		   char1 = Atoupper(char1);
		   char2 = Atoupper(char2);

		   if(char1 != char2) 
			   break;
	}

	return((n < 0) ? 0 : (char1 - char2));
}
#endif   /*  }  BUILD_WIDE  */
