/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: axgtstr.c
   Applix ASC version: 21
   Date: Mon Nov 16 13:57:59 1998
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>


#if ALPHA
#include <unistd.h>
#endif


#if SYSV4 || AX_NT || CLIX || M_88K || SCOUNIX
#include <fcntl.h>
#else
#include <sys/file.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#include <aos.h>
#include <utloc.h>
#include <gengetst.h>
#include <scac.h>

#define IXHANDLE  1
#define TXTHANDLE 0 

static int getSrFD, open_flag = FALSE;
static int first_time = TRUE;

#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG 1

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
static short getSr_categIx[MAXCATEG]; 
	/* Each entry is the disc blk number of the index blk for a category. */

static union strIxBlk {
	struct {
		short curr0Categ;
		short ixValue0 [BLK0MAX];
		short categBlk[MAXCATEG];
	} blk0 ;
	struct {
		short currCateg;
		short ixValue[BLK1MAX];
	} ixblks ;
} IxBlk;

struct strDataBlk {
	short dataBlkNum;
	unsigned char stringStuff[DATABLKSZ];
} DataBlk;


static char *strflbuf;
static long seekpos = 0;

static openxx(filename)
char *filename;
{
#if DATA_PRELOAD && !ELFRT_ONLY
	extern char AxStrflData[];

	strflbuf = AxStrflData;
	return(0);
#else
	FILE *fp;
	struct stat status;
	static int called;
	char *PermAlloc();
	extern int AxElfrtMode;
	extern char *ElfrtGetStrfl();

	if ( called )
		return(0);
	errno = ENOENT;
#if AXMS
	if ( NULL == (fp = fopen(filename,"rb")) )
#else	
	if ( NULL == (fp = fopen(filename,"r")) )
#endif  /* DAMNT */
		return(-1);
    if ( -1 == stat(filename,&status) )
		goto fail;
	if ( NULL == (strflbuf = PermAlloc(status.st_size)) )
		goto fail;
	if ( 1 != fread(strflbuf,status.st_size,1,fp) )
		goto fail;
	fclose(fp);
	called++;
	return(0);
fail:
	tempAbort("Not enough memory to initialize");
#endif
}
static axlseek(fd,pos,mode)
int fd,mode;
long pos;
{
	seekpos = pos;
	return(0);
}
static void readxx(fd,buf,cnt,shorts_to_fix)
int fd;
char *buf;
int cnt;
int shorts_to_fix;
{
    unsigned short *shortP;
    unsigned char *byteP;
    short short_val;
    int i;

	memcpy(buf,strflbuf+seekpos,cnt);
	seekpos += cnt;

    /* the block is assumed to consist partially or entirely of shorts.  */
    /* These are written to the file in big-endian order.                */
    /* Here we convert them to the native machine representation of a short.  */
    byteP = (unsigned char *)buf;
    shortP = (unsigned short *)buf;
    for ( i=0; i<shorts_to_fix; i++ )
       {
       short_val = (*byteP << 8) + *(byteP+1);
       *shortP = short_val;
       shortP++;
       byteP = byteP + 2;
       }

	return;
}

getStrClose(){}		/* OBSOLETE */
getStrCache(){}

/*--------------------------- getStr -------------------------------------*/

typeErr getStr(strNum,strTxt)  /*	GET STRING FROM FILE  */
unsigned int strNum;
char strTxt[];	/* pointer to an array of characters. */
{
	REGISTER int i,cat,snum;
	short offs;
	long l,blkL;
        
	cat = strNum/(unsigned int)ERRCOFFSET;
	snum = strNum%(unsigned int)ERRCOFFSET;

	i = (cat == 0) ? BLK0MAX : BLK1MAX;
	/*string # larger than allowed in this category or category # too large*/
	if(snum > i || cat > MAXCATEG)
		goto bumStrnum;

	if(first_time || cat != IxBlk.ixblks.currCateg) {
		/* Then we don't have the index block for this category in memory. */
		if(!open_flag && getStrOpen())
			goto bumStrnum;
		if((l = getSr_categIx[cat]) == -1)
			goto bumStrnum;	  /*no strings in this category. */
		if(axlseek(getSrFD,(long)(l*DISCBLKSZ),0) == -1)
			goto bumStrnum;
		readxx(getSrFD,(char *)&IxBlk,DISCBLKSZ, (DISCBLKSZ/sizeof(short)) );
	}

	l = IxBlk.ixblks.ixValue[snum];
	blkL = l << OFFSSHIFT;
	if(blkL == 0)		  /* no string by this number in this category. */
		goto bumStrnum;
	offs = blkL%DISCBLKSZ;
	blkL /= DISCBLKSZ;

#if AXMS	
	if(blkL <= 0)	        /* hack for NT build */
		goto bumStrnum;
#endif
	
	if(blkL != DataBlk.dataBlkNum ) {
		if(!open_flag && getStrOpen())
			goto bumStrnum;
		if(axlseek(getSrFD,blkL*DISCBLKSZ,0) == -1)
			goto bumStrnum;
		readxx(getSrFD,(char *)&DataBlk,DISCBLKSZ, 1);
	}

	bcopy(&DataBlk.stringStuff[offs + 1],strTxt,DataBlk.stringStuff[offs] - 1);
	return(AOK);
	
bumStrnum:
	sprintf(strTxt,"...code %d...",strNum);
	return(ERRABSTR);
}
	
/*- - - - - - - - - - - - - getStrOpen - - - - - - - - - - - - - - - - -*/
int getStrOpen()
/* RETURN VALUE: 0 if file opened OK, otherwise, -1 */
{
	REGISTER int i;
	extern char *AxStrflName;

	if ( !AxStrflName ) {
        tempAbort( NXLT("No axstrfl (Applixstrfl) specified") );
	}
	if ( (getSrFD = openxx(AxStrflName, O_RDONLY)) == -1) {
		char *tmpmsg;
		tmpmsg = (char *)malloc(32 + strlen(AxStrflName));
		if(tmpmsg) {
			sprintf(tmpmsg, NXLT("getStrOpen: Could not open %s"), AxStrflName);
    	    tempAbort( tmpmsg );
		}else{
			tempAbort( NXLT("getStrOpen: Could not allocate memory for error message."));
		}
	}

	open_flag = TRUE;

	if(first_time) {
		/* Establish the list of index blocks */
		DataBlk.dataBlkNum = 0;
		readxx(getSrFD,(char *)&IxBlk,DISCBLKSZ, (DISCBLKSZ/sizeof(short)));
		for(i = 0; i < MAXCATEG; i++)
			getSr_categIx[i] = IxBlk.blk0.categBlk[i];
		first_time = FALSE;
	}
	return(0);

}

char *QuickGetStr(strNum)
{
	static char text[80];
	getStr(strNum,text);
	return(text);
}

char *AxGetStr(strNum)
unsigned int strNum;
{
	static char buf[256];

	if ( AOK != getStr(strNum,buf) )
		strcpy(buf,"***");
	return(buf);
}

