/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: coding8.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:57:58 1998
*/
#include <stdio.h>



/****************************************************************/
/*                                                              */
/* Function: Encode_8bit_Chars                                  */
/*                                                              */
/*           This function will take a string and return that   */
/*           string with all 8 bit characters encoded in a      */
/*           \<hex digit><hex digit> format.  This is necessary */
/*           to do because 8 bit characters cannot travel threw */
/*           unix mail. Also, any encountered backslash char    */
/*           in the string will be represented by a double \\   */
/*                                                              */
/* Inputs: String to be encoded, and returned encoded string.   */
/* Outputs: New string with encoded 8 bit characters.           */
/*                                                              */
/****************************************************************/

Encode_8bit_Chars(Orig_string, Encoded_string)
char *Orig_string;
char *Encoded_string;
{
	char *Orig_ptr, *Encoded_ptr;
	int val;

	/* Go through strings and encode any 8 bit chars */

	Orig_ptr = Orig_string;
	Encoded_ptr = Encoded_string;

	while ( *Orig_ptr != '\0' )
		{
		/* Now that we have finished checking for  */
		/* new lines let's look for 8-bit chars    */

		if (*Orig_ptr == '\\')
			/* Add an extra backslash to double it */
			*Encoded_ptr++ = '\\';

		if (*Orig_ptr & 0x80)
			{
			/* We found a 8 bit character */
			*Encoded_ptr++ = '\\';

			/* Do High hex digit */

			val = (*Orig_ptr & 0xf0) >> 4;
			if (val <= 9)
				*Encoded_ptr++ = '0' + val;
			else
				*Encoded_ptr++ = 'a' + val - 10;

			/* Now do low hex digit */

			val = *Orig_ptr & 0x0f;
			if (val <= 9)
				*Encoded_ptr++ = '0' + val;
			else
				*Encoded_ptr++ = 'a' + val - 10;

			Orig_ptr++;
			}
		else
			{
			/* Normal character */

			*Encoded_ptr++ = *Orig_ptr++;
			}
		}
	*Encoded_ptr = '\0';
}


/************************************************************/
/*                                                          */
/* Function: Decode_8bit_Chars                              */
/*                                                          */
/*           This function will go through a strings and    */
/*           convert any encoded and escaped 8-bit chars    */
/*           back into real 8-bit characters. The escape    */
/*           sequences begin with a backslash followed by   */
/*           two hex digits. A backslash is represented by  */
/*           a double backslash.                            */
/*                                                          */
/* Inputs: String to be decoded.                            */
/*                                                          */
/************************************************************/

Decode_8bit_Chars(Str)
char *Str;
{
	char *ptr, *ptr2;
	int val;

	/* Go through strings and convert any escaped 8-bit */
	/* characters into real 8-bit characters.           */

	ptr = ptr2 = Str;

	while( *ptr2 != '\0' )
	{
		if (*ptr2 == '\\')
		{
			/* Found a possible 8-bit char */

			if (*(ptr2+1) == '\\')
			{
				/* Found double backslash skip it */

				ptr2++;
				*ptr++ = *ptr2++;
			}
			else
			if ( *(ptr2+1) == '~')
			{
				/* check for "~" or "~foo" */
				ptr2++;
				*ptr++ = *ptr2++;
			}
            else if ( (((*(ptr2+1) >= '0') && (*(ptr2+1) <= '9')) ||
                      ((*(ptr2+1) >= 'a') && (*(ptr2+1) <= 'f'))) &&
                      (((*(ptr2+2) >= '0') && (*(ptr2+2) <= '9')) ||
                      ((*(ptr2+2) >= 'a') && (*(ptr2+2) <= 'f'))) ) 
                     
			{
                /* both of the characters following the backslash are in
                *    the allowable range for hex digits, therefore we have
				*  Found 8-bit character                                  */
				/* lets get first char of*/
				/* sequence.             */

				ptr2++;
				if ( *ptr2 >= 'a')
					val = *ptr2 - 'a' + 10;
				else
					val = *ptr2 - '0';

				val = val << 4;

				/* next sequence */

				ptr2++;

				if (*ptr2 >= 'a')
					val = val + (*ptr2 - 'a' + 10);
				else
					val = val + (*ptr2 - '0');

				*ptr++ = val & 0xff;
				ptr2++;
			}
            else
               {
               /* the backslash is not followed by valid hex digits, 
               *   leave it alone   */
               *ptr++ = *ptr2++;
               }
		}
		else
		{
			/* normal character */

			*ptr++ = *ptr2++;
		}
	} /* end while */
	*ptr = '\0';
}
		
