/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: compstr.c
   Applix ASC version: 7
   Date: Fri Jan  8 21:11:44 1999
*/
#include <varargs.h>
#include <aos.h>
extern char *strcat();
char *temp = NULL;

#define MAX_ARGS	100

char *comp_str(va_alist)
va_dcl
{

	/* The first arg is the pattern, all following are the replacement strings.
		The substitutes array holds the pointers to these
		replacement strings in the order they were passed in */

	va_list ap;
	char *pattern = NULL;
	char *substitutes[MAX_ARGS];
	char *out_ptr = NULL;
	char out_buf[512];
	char pattern_buf[512];
	int indx = 0,
		hi_indx = 0;

	va_start (ap);
	pattern = va_arg(ap, char *);
	strcpy (pattern_buf, pattern);

	/* Store the arguments in an easy-to-access array and save the array size */
	memset (substitutes, 0, MAX_ARGS * sizeof(char *));
	while ( (substitutes[indx] = va_arg(ap, char *)) != (char *) 0)
			indx++;
	hi_indx = indx;

	/* loop through the pattern copying it into out_buf until a
	   '$' is met. */
	pattern = pattern_buf;
	out_ptr = out_buf;
	
	while (*pattern != '\0') {

		/* By initially setting the stepping pointer (out_ptr) to
		   the null char we avoid any unpredictable results that
		   may be the result of some garbage info present initially
		   in out_buf.  */
		*out_ptr = '\0';

		if (*pattern != '$') {

		/* normal character, just copy it */
			*out_ptr = *pattern;

		/* increment pointers accordingly */
			out_ptr++;
			pattern++;

		}
		else {
		/* Encountered a $. the following character will be the 
		   number of the substitute argument. To save this as an
		   integer subtract the character '0' from it. */

			pattern++;

			indx = *pattern - '0';

			pattern++;

		/* Use this indx (decrement since C starts from 0) to get 
		   the proper substitute string from the array and append 
	   	   this to out_ptr */

			indx -= 1;

		/* Check that the value retrieved is valid array indx */
		   if ( (indx < 0) || (indx > hi_indx - 1) ) {
				printf("An invalid argument number %d found in compose_string. Using arg 1\n",indx+1);

				indx = 1;
		    }

			if (substitutes[indx] != NULL) {

			    strcat(out_buf,substitutes[indx]);

				out_ptr = out_ptr + strlen(substitutes[indx]);
			}

			else {
			/* if the substitute is NULL, then a blank space will
			   be written into the out_buf. */
				*out_ptr = ' ';
				out_ptr++;
				printf("Index %d is a null string. Writing a blank space to the out_buf. \n", indx);
			}

		} /* outer else */

			
	} /* while */

	/* null terminate the new string (for tidiness) */
	/* remember that out_ptr (the stepping pointer) points to the
	   end of the out_buf string */
	*out_ptr = '\0';
	temp = out_buf;

	return(temp);

}
