/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: dia.c
   Applix ASC version: 2
   Date: Mon Nov 16 13:57:59 1998
*/
#include <utloc.h>
#include <dia.h>
#include <paripc.h>

#define MAXSTACK 8	/* Maximum nesting level of constructor data items*/
#define DIAHEADLN (unsigned short)(sizeof(unsigned short)+sizeof(typeDt))
			/* This is the size in bytes of the length and data
			type info that precedes each data value in a DIA
			format message. */
#define CHARSIZE 8

/* ------------------------------------------------------------- */

static sh2str (msgP, c)
REGISTER char *msgP;
REGISTER char *c;
/*
 * Copy a short integer to a string such that the high order byte
 * preceeds the low order byte.  Assumes no restrictions on alignment
 * of msgP.
 */
{
#if VAXORDER
	*msgP++ = *(c + 1);
	*msgP = *c;
#else
	*msgP++ = *c++;
	*msgP = *c;
#endif
	return;
}

/* ------------------------------------------------------------- */

static unsigned short str2sh (msg)
REGISTER char *msg;
/*
 * Extract a short integer from a string.  No alignment restrictions.
 */
{
	REGISTER unsigned short s;

	s = *((unsigned char *)msg++) << CHARSIZE;
	s |= *((unsigned char *)msg);
	return(s);
}

/* ------------------------------------------------------------- */

static lg2str (msgP, c)
REGISTER char *msgP;
REGISTER char *c;
/*
 * Copy a long integer to a string such that the high order byte
 * preceeds the low order byte.  Assumes no restrictions on alignment
 * of msgP.
 */
{
	REGISTER int i;

#if VAXORDER
	c += sizeof(long) - 1;
	for (i = 0; i < sizeof(long); i++)
		*msgP++ = *c--;
#else
#if PDP11ORDER
	*msgP++ = *(c + 2);
	*msgP++ = *(c + 3);
	*msgP++ = *c;
	*msgP = *(c + 1);
#else
	for (i = 0; i < sizeof(long); i++)
		*msgP++ = *c++;
#endif
#endif
	return;
}

/* ------------------------------------------------------------- */

static unsigned long str2lg (msg)
REGISTER char *msg;
/*
 * Extract a long integer from a string.  No alignment restrictions.
 */
{
	unsigned long s;

	s  = ((long) *((unsigned char *) msg++)) << CHARSIZE*3;
	s |= ((long) *((unsigned char *) msg++)) << CHARSIZE*2;
	s |= ((long) *((unsigned char *) msg++)) << CHARSIZE;
	s |= *((unsigned char *)msg);
	return(s);
}

/*
.BEGIN MANPAGE
.NAME DIAgettype - Identify the next element in a DIA message.
.SYNOPSIS
#include <dia.h>
.DESCRIPTION
	DIAgettype() identifies the next datum whose location pointed to by msgP.
	The type, length, and location of the data are returned in the
	AI_desc structure.  It is assumed that msgP points to the beginning
	of a correctly defined datum on input, e.g., the start of a
	DIA message or submessage. It is often obtained by a previous call
	to this function or DIAgetval.
.ARGUMENT msgP
	The address of the beginning of a (length/type, data) tuple in a
	DIA message.  Input, not altered.
.ARGUMENT descP
	The address of a AI_desc structure in which the return values from
	this function will appear.  In particular, AIdlen, AItype, and AIptr
	will be set.  AIptr points to the first byte of the data portion
	of the DIA tuple.  If AItype is of the fundamental data type DtNODATA,
	AIptr is undefined.  If AItype is of fundamental data type DtCONSTRUCTOR,
	then AIptr points to the beginning of the first tuple of the submessage
	and AIdlen is the total length of the submessage.
.RETURN VALUE
	Normally the address of the beginning of the next data tuple is returned.  
	The returned value may be resubmitted to this routine as a msgP value.
  
	If the AItype returned is DtEOM (end of message), then the return value is
	undefined.

	If the AItype returned is of the fundamental type DtCONSTRUCTOR, then the
	return value refers to the data tuple following the end of the 
	submessage, not the first element of the submessage.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
char *DIAgettype(msgP, descP)
REGISTER char *msgP;
REGISTER AI_desc *descP;
/* END SYNOPSIS */
{
	descP -> AIdlen = str2sh(msgP) - (unsigned short)DIAHEADLN;
	msgP += sizeof(short);
	descP -> AItype = str2sh(msgP);/*NOTE!!! assumes that typeDt is short*/
	msgP += sizeof(short);
	descP -> AIptr = msgP;
	return( msgP + descP->AIdlen );
}


/*
.BEGIN MANPAGE
.NAME DIAsearch - Search for element of specified type.
.SYNOPSIS
#includes <dia.h>
.DESCRIPTION
	DIAsearch() searches for an element of the specified type stopping when 
 	type = DtEOM is reached.
.ARGUMENT ptr
	Ptr is a pointer to where to start search (must point to beginning
	of an DIA datum).
.ARGUMENT type
	Type is an DIA datum type (typeDt) to search for.
.ARGUMENT cdesc
	Cdesc is a pointer to an 'AIdesc' descriptor for returning values of 
	the found datum.
.RETURN VALUE
	The address of the beginning of the matching tuple in the 
	DIA message (if found) or NULL (if not found).  Note the
	difference between the address returned in AIptr (which is
	a data address) and returned by this function proper (which
	is a tuple address).
.NOTES
	This routine does NOT do a depth search of constructor data items
	(i.e., submessages).
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
char *DIAsearch(ptr, type, cdesc)
REGISTER char *ptr;
REGISTER typeDt type;
REGISTER AI_desc *cdesc;
/* END SYNOPSIS */
{
	AI_desc desc;	/* our own personal descriptor.	*/
	REGISTER char *s, *nP;

	s = ptr;
	for (;;) {
		nP = DIAgettype (s, &desc);
		if (desc.AItype == type) {
			cdesc -> AItype = desc.AItype;
			cdesc -> AIdlen = desc.AIdlen;
			cdesc -> AIptr = desc.AIptr;	
			return(s);
		}
		if (desc.AItype == DtEOM) return(NULL);
		s = nP;
	}
}


/*
.BEGIN MANPAGE
.NAME DIAgetval - Copy the data portion of a DIA tuple to caller's buffer.
.SYNOPSIS
#include <dia.h>
.DESCRIPTION
	DIAgetval() copies the DIA datum referenced by descP into val,
	unless valP = NULL.  

	This call is primarily used to move text data from the DIA message to
	the caller's buffer.  The text may be references directly using
	"descP->AIptr".  If the data is not text, the appropriate type
	specific routine should be used (e.g., DIAgetlong()).
.ARGUMENT descP
	The address of AI_desc structure which identifies the tuple whose data is to
	be copied. AIptr and AIdlen must be properly set (usually via a DIAgettype
	call).  Input only.
.ARGUMENT valP
	The address of a buffer into which the data is to be moved.  It is the
	caller's responsibility to provide a buffer of at least AIdlen bytes.
	No buffer overflow checks are made.  If valP is NULL, no data is copied
	but this function's return value is set as though data had been moved.
.RETURN VALUE
	The address of the beginning of the next data tuple is returned.
	The returned value may be submitted to DIAgettype as the msgP argument.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
char *DIAgetval(descP, valP)
REGISTER AI_desc *descP;
REGISTER char *valP;
/* END SYNOPSIS */
{
	REGISTER short * shortP;
	REGISTER long * longP;

	if (valP != NULL) {
		switch (DtFundamental (descP->AItype)) {
			case (DtSHORT):
				shortP = (short *)valP;
				*shortP = str2sh (descP->AIptr);
				break;

			case (DtLONG):
				longP = (long *)valP;
				*longP = str2lg (descP->AIptr);
				break;

			default:
				memcpy (valP, descP->AIptr, (int)descP->AIdlen);
		}
	}
	return (descP->AIptr + descP->AIdlen);

}


/*
.BEGIN MANPAGE
.NAME DIAgetlong - Get value of DIA data item whose fundamental type is DtLONG.
.SYNOPSIS
#include <dia.h>
.DESCRIPTION
	Interpret the datum referenced by descP as a long integer
	and copy its value into intval.  
.ARGUMENT descP
	The address of a AI_desc structure which contains data usually set by
	a previous call to DIAgettype.  The fundamental type of the AItype value
	should be DtLONG.
.ARGUMENT intvalP
	The address of a long integer to be set by this procedure.
.RETURN VALUE
	The address of the beginning of the next data tuple is returned.
	The returned value may be submitted to DIAgettype as the msgP argument.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
char *DIAgetlong(descP, intvalP)
REGISTER AI_desc *descP;
REGISTER unsigned long *intvalP; 
/* END SYNOPSIS */
{
	*intvalP = str2lg(descP->AIptr);
	return (descP->AIptr + sizeof(long));
}


/*
.BEGIN MANPAGE
.NAME DIAgetshort - Get value of DIA data item whose fundamental type is
	DtSHORT.
.SYNOPSIS
#include <dia.h>
.DESCRIPTION
	Interpret the datum referenced by descP as a short integer
	and copy its value into intval.  
.ARGUMENT descP
	The address of a AI_desc structure which contains data usually set by a
	previous call to DIAgettype.  The fundamental type of the AItype value
	should be DtSHORT.
.ARGUMENT intvalP
	The address of a short integer to be set by this procedure.
.RETURN VALUE
	The address of the beginning of the next data tuple is returned.
	The returned value may be submitted to DIAgettype as the msgP argument.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */ 
char *DIAgetshort(descP, intvalP)
REGISTER AI_desc *descP;
REGISTER unsigned short *intvalP;
/* END SYNOPSIS */
{
	*intvalP = str2sh(descP->AIptr);
	return (descP->AIptr + sizeof(short));
}

/*
.BEGIN MANPAGE
.NAME DIAgetdouble - Get value of DIA data item whose fundamental type is 
	DtDOUBLE.
.SYNOPSIS
#include <dia.h>
.DESCRIPTION
	Interpret the datum referenced by descP as a double precision integer
	and copy its value into intval.  
.ARGUMENT descP
	The address of a AI_desc structure which contains data usually set by
	a previous call to DIAgettype.  The fundamental type of the AItype
	value should be DtDOUBLE.
.ARGUMENT doublevalP
	The address of a double precision integer to be set by this procedure.
.RETURN VALUE
	The address of the beginning of the next data tuple is returned. 
	The returned value may be submitted to DIAgettype as the msgP argument.
.WARNING
	Assumes that doubles are not transmitted between machines
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
char *DIAgetdouble(descP, doublevalP)
REGISTER AI_desc *descP;
REGISTER double *doublevalP; 
/* END SYNOPSIS */
{
	memcpy( (char *)doublevalP, descP->AIptr, sizeof(double) );
	return (descP->AIptr + sizeof(double));
}


/*
.BEGIN MANPAGE
.NAME DIAputval - Add a data value to a DIA message being built.
.SYNOPSIS
#include <dia.h>
.DESCRIPTION
	Append the datum described by the arguments to the message.  The
	fundamental data type subfield of the Dt value in AItype is used
    to determine the number of bytes of data moved into the message for
    all fundamental types except DtCHAR and DtCONSTRUCTOR.  In these
    cases, AIdlen determines the data size.  AIptr is the address of the
    lowest byte of the data.  It is ignored if Dt is of fundamental type
    DtNODATA.
.ARGUMENT msgP
	The address of the 1st available byte in the buffer where DIA format
	message is being built.  Input parameter.
.ARGUMENT msgendP
	The address of the last usable byte of the buffer allocated by the caller
	in which the message is being built.  If no check for buffer overflow is
	needed, set this NULL.  Input parameter.
.ARGUMENT descP
	The address of an AI_desc structure in which the Dt value, length and 
	address for the new datum is located.  Input paramter.
.RETURN VALUE
	Return a revised address of the first available byte in the message
	buffer where the next datum can be appended if this call is successful.

	If msgendP is not NULL and adding the designated data, including DIA tuple
	header info, would exceed msgendP then no action is taken by this routine
	and NULL is returned.  (Boundary condition note:  If the data exactly fills
	the buffer, then the address returned will be one greater than msgendP.)

	If the caller does not check for a NULL return from a DIAputval, DIAbegcon,
	or DIAendcon call and thus (in error) passes this call a msgP parameter
	of NULL, this procedure returns a NULL and does nothing else; i.e. it
	"pegs" to the buffer full condition.
.END MANPAGE
*/

static struct {
	char * ptr;
	unsigned short Sum;
} AIstack[MAXSTACK];
static int AIlevel = -1;

/* BEGIN SYNOPSIS */
char *DIAputval (msgP, descP, msgendP )
REGISTER char *msgP;
REGISTER char *msgendP;
REGISTER AI_desc *descP;
/* END SYNOPSIS */
{

	REGISTER unsigned short len;
	unsigned short slen;

	if( msgP == NULL )  
		return( NULL );
	switch( DtFundamental(descP->AItype) ) {
	case DtSHORT:
		len = sizeof(short);
		break;
	case DtLONG:
		len = sizeof(long);
		break;
	case DtDOUBLE:
		len = sizeof(double);
		break;
	case DtNODATA:
		len = 0;
		break;
	default:   /* must be char or constructor */
		len = descP->AIdlen;
	}
	if (msgendP != NULL) 
		if((msgP + len - 1 + DIAHEADLN) > msgendP) return(NULL);
	slen = len + DIAHEADLN;
	sh2str(msgP, (char *) &slen);
	msgP += sizeof(short);
	sh2str(msgP, (char *) &(descP -> AItype));
	msgP += sizeof(typeDt);
	if (len > 0)
		switch( DtFundamental(descP->AItype) ) {
		case DtSHORT:
			sh2str(msgP, descP->AIptr);
			break;
		case DtLONG:
			lg2str(msgP, descP->AIptr);
			break;
		default:
			memcpy(msgP, descP -> AIptr, len);
		}
	if (AIlevel >= 0) AIstack[AIlevel].Sum += (len + DIAHEADLN);
	return(msgP + len);

}


/*
.BEGIN MANPAGE
.NAME DIAbegcon - Begin a new DIA submessage.
.SYNOPSIS
#include <dia.h>
.DESCRIPTION
	Begin a (new) constructor datum.  Constructor data allows nesting of
	DIA format (sub)messages within a DIA message.  The submessage is appears
	in the containing message as a tuple whose data type (DtXXXX) is of
	fundamental type DtCONSTRUCTOR and whose length reflects all of the space
	occupied by the tuples within the submessage and whose data is the
	submessage itself.

	A submessage is normally build by issuing a call to DIAbegcon, followed by
	some number of calls to DIAputval followed by a call to DIAendcon.  If the
	entire submessage has been prepared in another buffer, the DIAbegcon call
	can be used to copy it in to the message being built in which case
	DIAbegcon would be immediately followed by a call to DIAendcon.

	This routine appends the specified constructor Dt value to the message and
	saves a 'place holder' for the length field internally, which will be
	filled-in when 'DIAendcon' function is called.  

	As subsequent calls to 'DIAputval' are made, the lengths of the elementary
	datum are summed and saved internally.  (Pushes down a level for any nested
	constructor datum.  Constructors may may be nested to a maximum level of
	MAXSTACK, typically a value of 8.)
.ARGUMENT msgP
	Pointer to the current end of the DIA message being built.  Input parameter.
.ARGUMENT msgendP
	Pointer to the last usable byte in the buffer provided by the caller to
	contain the DIA msg being built or NULL.  If msgendP is not NULL, then
	this function will return a NULL if there is not enough room to place the
	entire datum.  msgendP=NULL means don't bother with the check.  Input
	parameter.
.ARGUMENT descP
	The address of an AI_desc structure in which the Dt value (of fundamental
	type DtCONSTRUCTOR) is located.  Optionally, AIptr and AIdlen may be
	set if submessage data is being provided by this call.  If no data is
	being provided, set AIdlen = 0.
.RETURN VALUE
	Normally returns the address of the 1st available byte in the message
	buffer.  Returns NULL if stack limit has been reached.  Returns NULL if
	there is insufficient space in the buffer (based on msgendP).

	If the caller does not check for a NULL return from a DIAputval, DIAendcon
	call, etc. and thus (in error) passes this call a msgP parameter of NULL,
	this procedure returns a NULL and does nothing else; i.e. it "pegs" to the
	buffer full condition.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
char *DIAbegcon(msgP, descP, msgendP )
REGISTER char *msgP;
REGISTER char *msgendP;
REGISTER AI_desc *descP;
/* END SYNOPSIS */
{
	unsigned short zero = 0;

	if(AIlevel == (MAXSTACK-1)) return(NULL);
	if( msgP == NULL)
		return( NULL );
	if(msgendP != NULL) 
		if((msgP + descP->AIdlen - 1 + DIAHEADLN) > msgendP) 
			return(NULL);
	AIstack[++AIlevel].ptr = msgP;

/* set constructor length in msgP to 0 (for diag.)*/

	sh2str(msgP, (char *) &zero);	
	msgP += sizeof(short);
	sh2str(msgP, (char *) &(descP -> AItype)); /* put in datum type	*/
	msgP += sizeof(short);
	if (descP -> AIdlen != 0)	/* append data, if any	*/
		memcpy (msgP, descP ->AIptr, (int)descP -> AIdlen);

/* Compute length  of this datum, so far
*/
	AIstack[AIlevel].Sum = DIAHEADLN + descP -> AIdlen;
	return(msgP + descP -> AIdlen);
/* bump past type and/or data	*/
}



/*
.BEGIN MANPAGE
.NAME DIAendcon - End a DIA submessage.
.DESCRIPTION
	End the current constructor datum.  Appends a data item of DtEOM (end of
	message) which is of fundamental type DtNODATA, space permitting, and
	fills in the appropriate length information for the constructor datum
	which was most recently begun via a previous 'DIAbegcon' call.  (Pops
	back a level from any nested constructor datum.)
.ARGUMENT msgP
	Address of the 1st available byte in msg.
.ARGUMENT msgendP
	address of the last usable byte in buffer in which the message is being
	built or NULL if no buffer overflow check is desired.  Input parameter.
.RETURN VALUE
	Returns NULL if there was no corresponding 'DIAbegcon' call. Returns NULL
	if there is insufficient space in the buffer for the DtEOM data item
	which this call appends.  In both these cases, this procedure took no
	other action.  Normally returns pointer to the next available byte in
	the buffer where the message is being built.

	If the caller does not check for a NULL return from a DIAputval or
	DIAendcon call and thus (in error) passes this call a msgP parameter
	of NULL, this procedure returns a NULL and does nothing else;
	i.e. it "pegs" to the buffer full condition.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
char *DIAendcon( msgP, msgendP )
REGISTER char *msgP;
REGISTER char *msgendP;
/* END SYNOPSIS */
{
	REGISTER char * stcon;
	unsigned short total;
	AI_desc  desc;

	if (AIlevel == -1) {
		 return(NULL);
	}
	desc.AItype = DtEOM;
	if (DIAputval( msgP, &desc, msgendP ) == NULL)  {
		return( NULL );
	}
	total = AIstack[AIlevel].Sum;
	sh2str((stcon = AIstack[AIlevel].ptr), (char *) &total);
	--AIlevel;

/* if this was a nested constructor, sum its length into the previous level */

	if (AIlevel >=0) AIstack[AIlevel].Sum += total;
	return(stcon + total);
}

