/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: dt.h
   Applix ASC version: 14
   Date: Sat Jan 23 21:14:35 1999
*/

/*	These values allow ParseDate() and Formatdt() to run on a token
 *	recognition basis
 */



typedef enum {
	DU_UNIDENTIF = 1000,	/*	unidentifiable material has been found.  */

	DU_UNINIT = -1,		/*	Uninitialized */
	DU_1MON = 0,		/*	Numerical month expressed as 1 digit if
						 *	less than 10 */
	DU_2MON,			/*	Numerical month expressed as 2 digits */
	DU_1DAY,			/*	Numerical day expressed as 1 digit if
						 *	less than 10 */
	DU_2DAY,			/*	Numerical day expressed as 2 digits */
	DU_3DAY,			/* numerical day expressed as 3 digits */
	DU_4YR,				/*	Year expressed as 4 digits */
	DU_2YR,				/*	Year expressed as 4 digits */
	DU_1DATE2YR,		/*	Date expressed as yymmdd, all one number */
	DU_1DATE4YR,		/*	Date expressed as yyyymmdd, all one number */
	DU_1HR,				/*	Hour expressed as 1 digit if
						 *	less than 10 */
	DU_2HR,				/*	Hour expressed as 2 digits */
	DU_24HR,			/*	Hour expressed as 2 digits 00-23 */
	DU_MIN,				/*	Minutes expressed as 2 digits */
	DU_SEC,				/*	Seconds expressed as 2 digits */
	DU_MSEC,			/*	milliseconds expressed as 3 digits */
	DU_1TIME,			/*	Time expressed as hhmm, all one number */
	DU_EMP,				/*	Emperor date, expressed as <string,number> */
    DU_WST,             /*  Western date, expressed in Japanese */

	DU_NUMBER,			/*	A return value for an id'd number */
	DU_FMON,			/*	Full month name */
	DU_AMON,			/*	Abbrev. month name */
	DU_FWKDY,			/*	Full weekday name */
	DU_AWKDY,			/*	Abbrev. weekday name */
	DU_AM,				/*	Am */
	DU_PM,				/*	Pm */
	DU_TH,				/*	Ordinal abbreviation */
	DU_SLASH,			/*	Slash */
	DU_DASH,			/*	Dash */
	DU_PERIOD,			/*	Period */
	DU_COLON,			/*	Colon */
	DU_SPACE,			/*	Space */
	DU_COMMA,			/*	Comma */
	DU_ENDOFSTRING 		/*	EOS */
} DUTYPE;


typedef struct {
	char			fmt[120];		/*	fmt in plain expression */
	DUTYPE			tok[120],		/*	fmt as tokens */
					sep;			/*	separator char, default DU_SPACE */
	int				toklen;			/*	len of tok array */
} dttmfmt;


#define FMON_NAME	"Mmmm"	/*	Full month name */
#define FMON_LEN	4
#define AMON_NAME	"Mm"	/*	Abbreviated month name */
#define AMON_LEN	2
#define MON1_NAME	"m"		/*	Month as a single or dbl digit number */
#define MON1_LEN	1
#define MON2_NAME	"mm"	/*	Month as a 0-padded number */
#define MON2_LEN	2
#define DAY1_NAME	"d"		/*	Day as a single or dbl digit number */
#define DAY1_LEN	1
#define DAY2_NAME	"dd"	/*	Day as a 0-padded number */
#define DAY2_LEN	2
#define TH_NAME		"th"	/*	Ordinal abbreviation */
#define TH_LEN		2
#define FWKDY_NAME	"Wwww"	/*	Full weekday name */
#define FWKDY_LEN	4
#define AWKDY_NAME	"Ww"	/*	Abbreviated weekday name */
#define AWKDY_LEN	2
#define YR2_NAME	"yy"	/*	Year as a two-digit number */
#define YR2_LEN		2
#define YR4_NAME	"yyyy"	/*	Year as a four-digit number */
#define YR4_LEN		4
#define HR2_NAME	"hh"	/*	Hour */
#define HR2_LEN		2
#define	HR24_NAME	"HH"	/*	Hour expressed as 2 digits 00-23 */
#define	HR24_LEN	2
#define MIN_NAME	"mi"	/*	Minutes */
#define MIN_LEN		2
#define SEC_NAME	"ss"	/*	Seconds */
#define SEC_LEN		2
#define MSEC_NAME	"fff"	/*	milliseconds */
#define MSEC_LEN	3
#define PM_NAME		"pm"	/*	Am/pm abbreviation and a 12-hour clock */
#define PM_LEN		2
#define EMP_NAME	"EE"	/*	Emperor date fmt */
#define EMP_LEN		2
#define WST_NAME  "CCCC"  /*  Christian Era date fmt */
#define WST_LEN       4
#define DAY3_NAME  "ddd"  /* number of days passed in the year instead a month*/
#define DAY3_LEN       3
