/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: intsarr.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:57:58 1998
*/
#include <aos.h>

/* Note: IntsArrays are used all over the code to store pointers.
   So IntsArrays are really LongsArrays! */

long FirstInIntsArray(p)		/* returns 0 if array empty */
register long *p;
{
	register int cnt;

	if (!p || (*p == 0))
		return (0);
	return (*(p + 1));
}

InIntsArray(p, n)
register long *p;
register long n;
{
	register long cnt;

	if (!p)
		return (FALSE);
	cnt = *p++;
	while (--cnt >= 0)
		if (*p++ == n)
			return (TRUE);
	return (FALSE);
}

long *CopyIntsArray(p)
register long *p; /*changed to long ptr - ALPHA bug. -sean*/
{
	register long cnt, *new, *px;

	px = new = (long *) malloc((*p + 1) * sizeof(long));
	*new++ = cnt = *p++;
	while (--cnt >= 0)
		*new++ = *p++;
	return (px);
}

long *CreateIntsArray(n)
int n;
{
	register long *px;

	px = (long *) malloc((n + 1) * sizeof(long));
	*px = n;
	return (px);
}

void WriteIntsArray(p, n, val)
register long *p;
register long n, val;
{
    if ( *p <= n )
       /* allocated space is not big enough, do not overwrite something else! */
       return;
    *(p+n+1) = val;
    return;
}

long ReadIntsArray(p, n)
register long *p;
register long n;
{
	return (p[n + 1]);
}

long IntsArraySize(p)
register long *p;
{
	if (!p)
		return (0);
	return (*p);
}

/* Delete first instance of VAL from ints array P.
   Always returns new ints array NP (caller must free P). */
long *
DeleteIntsArray(p, val)
register long *p;
register long val;
{
	register long *np = (long *) NULL;
	register int found = FALSE;
	register int ix, nx;
	register int size;

	if (!InIntsArray(p, val)) {
		np = CopyIntsArray(p);
		return(np);
	}
	size = IntsArraySize(p);
	if (size == 1) {
		np = CreateIntsArray(0);
		return(np);
	}
	np = CreateIntsArray(size - 1);
	for (ix = 0, nx = 0; ix < size; ix++) {
		long thisVal;

		thisVal = ReadIntsArray(p, ix);
		if (!found && (thisVal == val)) {
			found = TRUE;
			continue;
		}
		WriteIntsArray(np, nx, thisVal);
		nx++;
	}
	return(np);
}

/* Add VAL to the ints array P. Return a new ints array NP.
   NOTE: Allows duplicates.  Caller must call `InIntsArray ()' to
   ensure no dups. */
long *
AddIntsArray(p, val)
register long *p;
register long val;
{
	register int ix, nx;
	register long *np = (long *) NULL;
	int size;

	size = IntsArraySize(p);
	if (!size) {
		np = CreateIntsArray(1);
		WriteIntsArray(np, 0, val);
		return(np);
	}
	np = CreateIntsArray(size + 1);
	for (ix = 0; ix < size; ix++)
		WriteIntsArray(np, ix, ReadIntsArray(p, ix));
	WriteIntsArray(np, size, val);
	return(np);
}


/* GNU Emacs settings - This (including the "\n^L") MUST be the last thing
 *                      in this file!
 * Local variables:
 * mode: c
 * tab-width: 4
 * end:
 */
