/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: istrncm.c
   Applix ASC version: 8
   Date: Mon Nov 16 13:58:00 1998
*/
#include <utloc.h>
#include <wchars.h>

#undef strncmp

unsigned int Atoupper();

#if 1
extern unsigned char Imap[][256];

/* 
.BEGIN MANPAGE 
.NAME Istrncmp - Perform Applix international string comparison.
.ARGUMENTS s1
	The address of a string.
.ARGUMENTS s2 
	The address of a string. 
.ARGUMENT n
	The maximum number of characters to compare
.ARGUMENTS langum 
	A Proximity language number
.ARGUMENTS case_sensitive 
	TRUE/FALSE
.DESCRIPTION 
	Istrncmp() is similar to strncmp() except characters get mapped to
	their international equivalents prior to comparison.
.RETURN VALUE 
	Istrncmp() returns an integer greater than, equal to, or less than 0, 
	accordingly as s1 is lexicographically greater than, equal to or less 
	than s2 within n characters.
.END MANPAGE 
*/

/* BEGIN SYNOPSIS */ 
Istrncmp(s1,s2,n,langnum, case_sensitive)
REGISTER unsigned char *s1;
REGISTER unsigned char *s2;
REGISTER int n;
REGISTER int langnum;
REGISTER int case_sensitive;
/* END SYNOPSIS */ 
{
	REGISTER int c1,c2;
	
	if ((s1 == s2) || n <= 0)	/* don't compare 0 or negative nb of chars */
		return(0);

	if(Imap[langnum][0] == 0)
		langnum = 1;

/*	if((langnum == 1 && case_sensitive))
 *		return(strncmp(s1,s2,n));
 */
	
	if (case_sensitive)
	{
		while(n-- &&
			(c1 = *(Imap[langnum] + *s1++ )) ==
			(c2 = *(Imap[langnum] + *s2++ )) && c1)
			{
				;
			}
	}
	else
	{
		while( n-- &&
				((c1 = *s1++ ) != 0) &&
				((c2 = *s2++ ) != 0) )
		{
			/*	Atoupper correctly toupper's accented chars, but is
			 *	very dependent on specific char tests for its subroutine,
			 *	Aislower() which give incorrect results if we set c1 and c2
			 *	to the imap value in the loop then Atoupper() that value.
			 *	So we'll Atoupper() our char here instead of in the while loop
			 *	because the correct result is still returned if the while
			 *	loop fails for lack of c1 or c2.
			 */

			c1 = *(Imap[langnum] + Atoupper(c1));
			c2 = *(Imap[langnum] + Atoupper(c2));
			if (c1 != c2)
				break;
		}
	}

	return((n < 0) ? 0 : (c1 - c2));
}


#if BUILD_WIDE    /*  {  */


/*** do a wide string comparison */
int IWstrncmp(str1, str2,n, langnum, case_sensitive)
register wchar *str1;
register wchar *str2;
register int n;
register int langnum;
register int case_sensitive;
{
    register unsigned int char1,char2;
 
    if (str1 == str2)
        return(0);
 
	if (case_sensitive)
	{
    while ( n-- &&
			((char1 = *str1++) != 0) && ((char2 = *str2++) != 0) )
        {
        if ( !IS_ORIENT_CHAR(char1) )
            char1 = *(Imap[langnum] + char1);
        if ( !IS_ORIENT_CHAR(char2) )
            char2 = *(Imap[langnum] + char2);
        if (char1 != char2)
            break;
        }
	}
	else
	{
    while (n-- &&
			((char1 = *str1++) != 0) && ((char2 = *str2++) != 0) )
        {
        if ( !IS_ORIENT_CHAR(char1) )
            char1 = Atoupper(*(Imap[langnum] + char1));
        if ( !IS_ORIENT_CHAR(char2) )
            char2 = Atoupper(*(Imap[langnum] + char2));
        if (char1 != char2)
            break;
        }
	}
    return(char1 - char2);
}
 

#endif    /*  } BUILD_WIDE */

#else	/*}{*/

extern unsigned char	*Imap;

/* 
.BEGIN MANPAGE 
.NAME Istrncmp - Perform Applix international string comparison.
.ARGUMENTS s1
	The address of a string.
.ARGUMENTS s2 
	The address of a string. 
.ARGUMENT n
	The maximum number of characters to compare
.DESCRIPTION 
	Istrncmp() is similar to strncmp() except characters get mapped to
	their non-international equivalents prior to comparison.  For example,
	if "o" is equivalent to "o-umlaut", "o-umlaut" will get mapped to "o".
.RETURN VALUE 
	Istrncmp() returns an integer greater than, equal to, or less than 0, 
	accordingly as s1 is lexicographically greater than, equal to or less 
	than s2 within n characters.
.END MANPAGE 
*/

/* BEGIN SYNOPSIS */ 
Istrncmp(s1,s2,n)
REGISTER char *s1;
REGISTER char *s2;
REGISTER int n;
/* END SYNOPSIS */ 
{
	REGISTER int c1,c2;
	
	if(s1 == s2)
		return(0);

	if(!Imap)
		return(strncmp(s1,s2,n));
	
	while(n-- &&
	   	(c1 = (*(Imap + (*s1++ & 0xff)) & 0xff)) ==
		(c2 = (*(Imap + (*s2++ & 0xff)) & 0xff)) && c1);

	return((n < 0) ? 0 : (c1 - c2));
}
#endif	/*}*/
