/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: listck.c
   Applix ASC version: 2
   Date: Mon Nov 16 13:57:59 1998
*/
#include <utloc.h>

/*
.BEGIN MANPAGE
.NAME ListCk - Search a string list.
.ARGUMENT valStr
	The address of the string to be searched for.
.ARGUMENT listStr
	The address of the string list.
.ARGUMENT caseflag
	0 for case insensitive comparison.

	1 for case sensitive comparison.
.DESCRIPTION
	ListCk() is used to search for a string among a list of values.
	If the string is found, ListCk() will return an integer value
	associated with the string.  The format of the string list is
	illustrated below:

    #aJAN#bFEB#cMAR#dAPR#eMAY#fJUN#gJUL#hAUG#iSEP#jOCT#kNOV#lDEC#fJUNE##

	The first char of the string list declares the delimiter character to be 
	used in this string list.  In the case shown, # will separate each entry.

	The character immediately following each delimiter character is the value
	returned by ListCk() if the following entry is matched.  In the case shown,
	if the string being searched for is "FEB", ListCk() will return 'b'.

	The string list must end with two consecutive occurances of the delimiter
	character.  The null string can have an associated value.  The following
	string list illustrates this:

    #a#bNO#cYES##
.SEE ALSO
	ListExtract(3A)
.RETURN VALUE
	If valStr is found in the string list, the integer value associated
	with the string is returned.  If valStr is not found, then -1 is returned.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
ListCk(valStr,listStr,caseflag)
unsigned char *valStr;
unsigned char *listStr;
int	caseflag;
/* END SYNOPSIS */
{
	unsigned char *lp;
	int del,val,len;
#if SYS&SYS5 || SYS&VMSSYS || SYSV4
	extern int memcmp(),Astrncmp();
	int (*compare)() = caseflag ? memcmp : Astrncmp;
#else
	extern int bcmp(),Astrncmp();
	int (*compare)() = caseflag ? bcmp : Astrncmp;
#endif

    if ( (listStr == (unsigned char *)NULL) || (valStr == (unsigned char *)NULL) )
        return( -1 );

	len = strlen(valStr);
	del = *listStr++;

	while((val = *listStr++) != del) {
		lp = listStr;
		while(*listStr++ != del);
		if(listStr != lp + len + 1 || (*compare)(lp,valStr,len))
			continue;
			/* added for mbp */
		if((len == 0) && (val == '0')) val = 0;
           /* Returned 48 if empty string in spell menu
           * see qmenu.c
           * Used as index to Spell_Map causes core dump
           * because of segmentation violation on UNISYS and PCS
           * See Spell_Map in dc/sources/controller/CMenu2.c
           * mbp_PR#1010 and mbp_PR#1544 Sep-14,1989  am
           */
		return(val);
	}
	return(-1);
}
