/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: listext.c
   Applix ASC version: 4
   Date: Mon Nov 16 13:57:59 1998
*/
#include <ctype.h>
#include <utloc.h>

/*
.BEGIN MANPAGE
.NAME ListExtract - Extract a string from a string list.
.ARGUMENT valu
	The numeric value associated with the string in the string list.
.ARGUMENT searchstrP
	The address of the string list from which the string is to be extracted.
.ARGUMENT caseopt
	0 for extracting the string as it exists in the string list.

	1 for extracting the string in upper case.

	-1 for extracting the string in lower case.
.ARGUMENT strP
	The address of a buffer which will receive the string.
.DESCRIPTION
	ListExtract() obtains a string associated with a specified value
	from a string list.  The format of the string list is
	illustrated below:

	#aJAN#bFEB#cMAR#dAPR#eMAY#fJUN#gJUL#hAUG#iSEP#jOCT#kNOV#lDEC##

	The first char of the string list declares the delimiter character to be
	used in this string list.  In the case shown, # will separate each entry.

	The character immediately following each delimiter character is the value
	associated with the string which follows.  In the case shown,
	if the value specified is 'f', ListExtract() will retrieve the string "JUN".

	The string list must end with two consecutive occurances of the delimiter
	character.  The null string can have an associated value.  The following
	string list illustrates this:

	#a#bNO#cYES##
.SEE ALSO
	ListCk(3A)
.RETURN VALUE
	ListExtract() will return the length of the string extracted.

	If valu is not in the string list, ListExtract() will extract a
	null string and return 0.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
ListExtract(valu,searchstrP,caseopt,strP)
int	valu;		
unsigned char *searchstrP;	
int caseopt;
char *strP;
/* END SYNOPSIS */
{
	int	c,dchr,len=0;

	dchr = *searchstrP++;

	while((c = *searchstrP++) != valu) {
		if(c == dchr)
			goto procexit;
		while(*searchstrP++ != dchr);
	}

	for(len = 0; (c = *searchstrP++) != dchr; len++) {
		if(caseopt < 0)
			c = tolower(c);
		else if(caseopt > 0)
			c = toupper(c);
		*strP++ = c;
	}

procexit:
	*strP = '\0';
	return(len); 
}
