/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: shcmd.c
   Applix ASC version: 5
   Date: Mon Nov 16 13:57:59 1998
*/
#include <utloc.h>

#include <signal.h>
#if SYS&B4_2
#include <sys/wait.h>
#endif

    /* MOSHE 4/27/87 - global pid of the invoked cmd - in case
	it hangs - needed for SUN NFS */
int Ashpid = -1;

/*
.BEGIN MANPAGE
.NAME shcmd - Execute a file.
.SYNOPSIS
shcmd(cmd)
char *cmd;
.ARGUMENT cmd
	The address of a string which contains the command to invoke.
	The command may not includes wildcards or I/O redirection.
	(If wildcards or I/O redirection is requires use system().)
.DESCRIPTION
	Shcmd() and shcmdReturningChildsExitCode() provide a generic
	interface to fork() and execvp() a command.
.SEE ALSO
	fork(2) execvp(3) system(3) wait(2) exit(2)
.RETURN VALUE
	Shcmd() returns the status obtained from wait().

	ShcmdReturningChildsExitCode returns the low byte of the value the
	command actually supplied to exit() if the command terminated
	normally.  Otherwise it returns 127.
.END MANPAGE
*/

/* This procedure starts a child process, waits for its termination and
returns the exit code with which the process terminated.  If the child
is never started, aborts, terminates due to a signal, or otherwise dies 
without executing an exit() statement, this procedure returns 0177. */

/* BEGIN SYNOPSIS */
shcmdReturningChildsExitCode(cmd)
char *cmd;
/* END SYNOPSIS */
{
	register int rtnv;
	rtnv = shcmd(cmd) & 0xffff; /*only 16 bits are defined*/

	/* If it is 0x7f then either the child stopped (whatever that means) or 
	the shcmd() routine failed and explicitly returned this value.  Don't do 
	any further interpretation. 
	If not 0x7f then, the child terminated due to a signal; treat this as a
	serious error. */

	if(rtnv&0xff)
		return(0x7f);

	/* The child exited and exit code is in upper byte. */
	return(rtnv >> 8);

}


/*
 * Issue a shell command and return its status
 * 127 is returned if the command couldn't be exec'ed.
 * Double quoted args are passed as a single string after
 * the quotes are removed to make sh -c "string" work
 */

shcmd(cmd)
char *cmd;
{         
#if SYS&VMSSYS
	extern char *SCAOSDIR;

	return(vms_cmd(SCAOSDIR,cmd,0));
#else
#if SYS&AEGISO
/*
 * For aegis we must copy the command, have it parsed in place, and then
 * have a pended AEGIS process created to do the work.
 */

#define	MAXARGS	20

	char *args[MAXARGS];
	char *cp;
	register int status;

	if(!cmd || *cmd == '\0')
		return(1);

	status = strlen(cmd) + 1;
	if((cp = malloc((unsigned int)status)) == NULL)
		return(1);
	memcpy(cp,cmd,status);
	cmdparse(cp,args,MAXARGS);
	status = pgm_shcmd(args);
	free(cp);
	return(status);

#else

	register int i;
	register char *c,*ccmd;
#if SYS&B4_2
	int sig = SIGCHLD;
#else
#if !AXMS
	int sig = SIGCLD;
#else
        int sig = 20;   /* BOGUS, for now! */
#endif  /* if !AXMS, DAMNT */
#endif
	int status;

#if 0
   *** the old way, only SYSV3 defined it as 'void', now
   ***  everyone does - jon   9/25/90
#ifdef SYSV3
	void (*sfunc)();
#else
	int (*sfunc)();
#endif
#endif
    void (*sfunc)();

	char *argv[32];
	int jpid;
	extern char *strtok();

	i = strlen(cmd) + 1;
	if((c = ccmd = malloc((unsigned int)i)) == NULL)
		return(1);
	memcpy(ccmd,cmd,i);
	for(i = 0; i < 31 && (argv[i] = strtok(c," ")); i++) {
		if(*argv[i] != '\"') { 
			c = NULL; 
			continue; 
		}    
		argv[i]++; 
		c = argv[i] + strlen(argv[i]);
		*c++ = ' '; 
		c = strtok(c,"\"");
	}
	argv[i] = NULL;
	sfunc = signal(sig,SIG_DFL);
#if SYS&B4_2
	if(!(Ashpid = vfork())) {
#else
	if(!(Ashpid = fork())) {
#endif
		execvp(argv[0],argv);
		_exit(127);
	}
	free(ccmd);

	/* wait for exec'ed process to finish */
	while((jpid = wait(&status)) != Ashpid && jpid > 0);

	signal(sig,sfunc);
	Ashpid = -1;
	return(status);
#endif
#endif
}
/* ------------------------------------------------------ */

#if SYS&AEGISO

/*
 * The command string is parsed in place.
 * Arguments are separated by spaces and/or tabs.
 * Single and double quoted strings represent a single argument.
 * Embedded single and double quotes are allowed if escaped.
 * The argument vector is filled out and returned.
 */

cmdparse(cp,args,maxargs)
register char *cp;	/* command string pointer */
char *args[];		/* argument vector */
int maxargs;
{
	register char tc;		/* terminating char for argument */
	register int i;


	if(!cp || *cp == '\0')
		return;

	maxargs--;
	for(i = 0; *cp && i < maxargs; i++) {

		while(*cp == ' ' || *cp == '\t')	/* white space */
			cp++;

							/* det. arg type */
		if(*cp == '"') {			/* double quoted string */
			tc = '"';
			cp++;
		}
		else if(*cp == '\'') {		/* single quoted string */
			tc = '\'';
			cp++;
		}
		else {					/* simple arg */
			tc = ' ';
		}


		args [i] = cp;				/* set arg ptr */
		if(tc == ' ') {			/* traverse simple arg */
			while(*cp && *cp != ' ' && *cp != '\t')
				cp++;
		}
		else {					/* traverse quoted string,
							   skipping embedded tc */

			while(1) {				
				while(*cp && *cp != tc)
					cp++;
				if(!*cp || *(cp - 1) != '\\')
					break;
				cp++;
			}
		}
		if(*cp)			/* make arg NULL terminated */
			*cp++ = '\0';

	}

	args[i] = NULL;

}

#endif
