/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: utchkpt.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:58:00 1998
*/
#include <utloc.h>

/*
.BEGIN MANPAGE
.NAME CheckPathName - Convert a VMS path to a Unix path
.DESCRIPTION
	CheckPathName() converts a VMS path to a Unix path.
.ARGUMENT source
	The address of the path string to be converted.
.ARGUMENT dest
	The address of a buffer to which a reformatted version of the path will be
	returned.
.RETURN VALUE
	If the path contains a '/' or is successfully converted to Unix format,
	then TRUE is returned, otherwise FALSE.
.NOTES
	CheckPathName() was written to convert VMS path syntax to Unix path syntax
	for the VMS porting effort.  The interface is provided on all platforms.
.END MANPAGE
*/
 
/* BEGIN SYNOPSIS */
CheckPathName(source,dest)
char *source;
char *dest;
/* END SYNOPSIS */
{
#if SYS&VMSSYS
	char *p;
	extern char *Agetcwd();
	extern char *shell$translate_vms();

	if(index(source,'/')) {
		if(source != dest)
			strcpy(dest,source);
		return(TRUE);
	}

	if(source[0] == '.' &&
		(source[1] == '\0' || (source[1] == '.' && source[2] == '\0'))) {
		sprintf(dest,"%s%s",Agetcwd(),source[1] ? "/.." : "");
		return(TRUE);
	}

	if(p = shell$translate_vms(source)) {
		if(p[0] == '.' && (p[1] == '\0' || (p[1] == '.' && p[2] == '\0')))
			sprintf(dest,"%s%s",Agetcwd(),p[1] ? "/.." : "");
		else
			strcpy(dest,p);
		return(TRUE);
	}

	if(source != dest)
		strcpy(dest,source);
	return(FALSE);

#else
	if(source != dest)
		strcpy(dest,source);
	return(index(source,'/') ? TRUE : FALSE);
#endif
}
