/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: uttimcm.c
   Applix ASC version: 3
   Date: Mon Nov 16 13:58:01 1998
*/
#include <aos.h>
#include <datetime.h>

/*
.BEGIN MANPAGE
.NAME ut_timecmp - Compare Applix format date-time values.
.SYNOPSIS
#include <datetime.h>
.ARGUMENT arg1
    An Applix format date-time as obtained from ParseDate(3A).
.ARGUMENT arg2
    An Applix format date-time as obtained from ParseDate(3A).
.DESCRIPTION
    Ut_timecmp() compares two times in Applix format.
    Only the time portion of the two arguments are compared.
    The date and seconds value never enter into the comparison.
.SEE ALSO
    ParseDate(3A) datecmp(3A) ut_daycmp(3A)
.RETURN VALUE
    Ut_timecmp() -1, 0 or 1 depending on whether arg1 is less than, equal to,
    or greater than arg2.
.END MANPAGE
*/

/* BEGIN SYNOPSIS */
ut_timecmp( arg1, arg2 )
unsigned long arg1;
unsigned long arg2;
/* END SYNOPSIS */
{
	unsigned long lc;

	if( arg1 == arg2 )
		return( 0 );

	/* base comparison on time values only. */

	lc = 60/DTTIMEGRANULARITY;

	arg1 = (arg1/lc)%1440;
	arg2 = (arg2/lc)%1440;

	return( (arg1 < arg2) ? -1 : ( (arg1 == arg2) ? 0 : 1 ) );
}
