/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: elintr.c
   Applix ASC version: 420
   Date: Thu May 13 21:13:49 1999
*/

/* Kernel of Elf Interpreter */

#define FOR_INTERPRETER 1
#define ELF_INTR_LOG 0          /* for debug */
#include <math.h>
#include <axvalues.h>
#include <elf_env.h>
#include <paripc.h>
#include <scac.h>
#include <thimble.h>
#include <elcodes.h>
#include <sqlops.h>
#include <elf_strs.h>
#include <style.h>              /* for parsenum */
#include <fsrts.h>
#include <getstr5.h>
#include <elrecarg.h>
#include <elhelper.h>
#include <eldebug.h>
#include <adlgpoke.h>
#include <axtrace.h>
#include <elintr.p>

#define APX_MASTER_REVSTAMP 300
#define EPSILON     (1e-150)  
#define MIN_PACKAGES 250

extern int      AxInTask;
extern char *DirSlash;
extern bint     ElfHelperPid;

bint ElfInLearnMode;    /* TRUE while in learn mode */

/************* MACROS *************************************/

#define MAX_ELF_RECURSION 100
#define MAX_ELF_ERRSTR 2000

#define GLOVARDEX (Elf.curFrame->packP->gvar_list[word2]->value)
#define EXTPTR (Elf.curFrame->packP->ext_list[word2])
#define EXTDEX SnapData(EXTPTR)

#define ELF_STACK_SIZE 2000     /* size of elf stack (in longs) */
#define ElfDumpDatum(x)
#define DUMPTOS()
#define POP_RAW (*(--Elf.sp))
#define POP_PTR ((data_ptr*)((long)POP_RAW & ~FAULT_BIT))
#define POP_VAL (*(POP_PTR))
#define POP_INT (IS_INTEGER((p1 = POP_VAL)) ? GET_INTEGER(p1) : RePopIntVal())

#define PUSHQUICK(locid,ival) Assign(PushPtr(&Elf.curFrame->locvars[locid]),MK_INTEGER((ival)))

#define FAULTED(p) ( LX(p) & FAULT_BIT )
#define UNFAULT(p) ((data_ptr)((long)p & ~FAULT_BIT))

#define NULLP (char*)0

/* Memory validator should be periodically enabled via #if DEBUG */
 /* ??c should be based on DEBUG */
#define MEMTEST 0
#if MEMTEST
#define DBG_RMV_MEM(p) ElfRmvMem(p)
#define DBG_ADD_MEM(p) ElfAddMem(p)
#else
#define DBG_RMV_MEM(p)
#define DBG_ADD_MEM(p)
#endif

#define MAX_ELFNUM 1e151         /* maximum value of a floating point number */

/************* GLOBAL VARS ********************************/

typedef int (*QSORT_FUNCPTR)(const void*, const void*);

elf_info        Elf;
ext_link *LinkByNameArray;		/* array of ext_links for CallByName */
int SizeLinkByNameArray;		/* number of ext_links in array */
int ElfProfiling;				/* true if recording calls for profiler */
static package_info **Packages; /* Array of packages installed */
static int NextPackageInd;         /* Index for installing a new package */
static int MaxPackages;         /* Max. packages for which memory is already allocated */
static short    GvarFreeList[MAX_GVARS];        /* Free list for gvars (maps to
												 * TaskFrame's) */
#if ELF_PRELOAD
extern int      GvarFreeHead;   /* emitted in last package */
#else
static int      GvarFreeHead;   /* first avail gvar */
#endif
axglobal data_ptr ElfRetData;   /* Alis func generic return value */
axglobal double ElfNumRet;      /* Alis func number return value */
axglobal char  *ElfStrPtrRet;   /* Alis func long string return value */
axglobal data_ptr NullDataPtr;  /* for better linting */
axglobal FILE  *ElfFpAbort;     /* Abort stack dump file */
breakpoint_t ElfBreakData[MAX_ELF_BREAKS];	/* breakpoint info */
int ElfNumBreakpoints;			/* number of breakpoints active */
debugger_info ElfDebugger[MAX_DEBUGGERS];	/* debugger info */
int ElfNumDebuggers;			/* number of attached debuggers */
int ElfStrsForInterpreter = TRUE;
data_ptr ElfQNumbers[MAX_QPUSH+2];	/* for the OP_PUSH_NUM small numbers */
int ElfLogging = FALSE;         /* TRUE enables logging */

extern short    MenulessMode;
extern char    *MainLibBuf;
extern elf_obj_grp *ElfObjData();
extern data_ptr *ElfGetObjData();

int      AxElfrtMode;
int      AxMailrtMode;
static char     CurExtMacroName[ELFIDSTR+1];
static int NDummyPackages;   /* used to keep track of all dummy packages */

/************* FORWARD REFERENCES ************************/
#if DBX_MODE
data_ptr       *PushPtr();
#else
#define PushPtr(p) (*Elf.sp++ = (data_ptr*)(p))
#endif
double log10(), pow();	/* log10 fails on linux - not in math.h? */


/*************************************************************************
function: ElfInitLibs - One-only init
*/
void 
ElfInitLibs()
{
	int             ix;
	int             cnt = 0;

/* Global vars free list init */

	for (ix = 1; ix < MAX_GVARS - 1; ix++)
		GvarFreeList[ix] = ix + 1;
	GvarFreeList[ix] = 0;
	GvarFreeHead = 1;

	ElfStrsForInterpreter = TRUE;       /* needed by InitStrs */
	ElfInitStrs();

/* Immediate mode literal numbers */

	for (ix = -1; ix <= MAX_QPUSH; ix++)
		ElfQNumbers[ix + 1] = MK_INTEGER(ix);

/* Libraries */
    MaxPackages = MIN_PACKAGES;
	Packages = (package_info **) TaskCalloc(0, sizeof(package_info *), MaxPackages);
#if !FOR_TESTER
    ElfInit();
    if ( setjmp(Elf.abortBuf) ) {
        /* tempAborts here if libaxel.so is not found */
        char buff[3000];
        if ( Elf.ErrorStr != NULL )
           strcpy( buff, Elf.ErrorStr  );
        else
           strcpy( buff, "Could not init Elf" );

        if ( Elf.ErrorObject != NULL )
           {
           strcat( buff, " " );
           strcat( buff, Elf.ErrorObject );
           }

		tempAbort( buff );
    }
	AxShlibElfInit();
	ElfInstallLibModules();
#endif	
}

#if !FOR_TESTER
AxFindTaskMaster(mypid)
int mypid;
{
/*  TaskMaster Criterion (For ELF tasks):
0) If called from within app (e.g. WP), the app is task master
1) If launched from kernel (e.g. menubar...), use the AxActiveTask global.
2) If launched by a non-Elf task, the parent task is TaskMaster
3) If launched by a NON-top-level Elf task, inherits parent's task master
4) If launched by a top-level Elf task parent is task master
   For Non-Elf tasks there is no taskMaster, they are their masters
*/

	if (APELFM != TaskInfo[mypid].appid)
		AxSetTaskMaster(mypid, TaskInfo[mypid].uid);
	else if (!TaskInfo[mypid].parentTask)
		AxSetTaskMaster(mypid, AxActiveTask);
	else {
		int             parentPid = ThimPidFromUID(TaskInfo[mypid].parentTask);
		if (APELFM != TaskInfo[parentPid].appid)
			AxSetTaskMaster(mypid, TaskInfo[mypid].parentTask);
		else {
			if (!TaskInfo[parentPid].taskMaster)
				AxSetTaskMaster(mypid, TaskInfo[mypid].parentTask);
			else
				AxSetTaskMaster(mypid, TaskInfo[parentPid].taskMaster);
		}
	}
}
#endif

/*************************************************************************
function:  ElfInit - Init runtime environment
*/
ElfInit()
{
	extern char    *ErrnoTextFromUnix();
	extern package_info *PACK_mainlib;

	if (Elf.inited)
		return (TRUE);
	Elf.inited = TRUE;

/* ??c The task frame is created with a fixed size.  A better approach would be
		to create it small and extend it as needed based on package loading */

	Elf.taskFrame = (data_ptr *) calloc(MAX_GVARS, sizeof(data_ptr));

	if (-1 == ElfReadSymTab(1)) {
		char            buf[500];
		sprintf(buf, NXLT("%s: initializing Elf"), ErrnoTextFromUnix(errno));
		tempAbort(buf);
	}
#if !FOR_TESTER
	if(!TaskInfo[THIMpid].taskMaster)
		AxFindTaskMaster(THIMpid);

	Elf.sp = (data_ptr **) malloc(sizeof(data_ptr) * ELF_STACK_SIZE);
	if (!Elf.ErrorStr)
		Elf.ErrorStr = (char*)malloc(MAX_ELF_ERRSTR + 1);
	if (!Elf.ErrorObject)
		Elf.ErrorObject = (char*)malloc(MAX_ELF_ERRSTR + 1);
	if (!Elf.ErrorFile)
		Elf.ErrorFile = (char*)malloc(STRPATH + 1);

	/* Error handler - want ElfStrAbort to be called to throw error */
	/* However, don't do this if the current task already has an error
		handler installed - e.g. if called by one of the main apps */
	if (!TaskInfo[THIMpid].elfAbortHandler)
		AxSetElfAbortHandler( ElfStrAbort);
#endif
	return (TRUE);
}

static 
load_lib(name)
	char           *name;
{
	char            libname[STRPATH + 1];
	package_info   *packP;

	sprintf(libname, NXLT("%s/elf/%s.elo"), ThimApplixDir, name);
	AxSysDosUnixPath(libname);
	ElfbReadDataFile(libname);
	packP = ElfLoadPackage(ElfRetData,FALSE);
	AxTaskFreeElfData(0, ElfRetData);
	ElfRetData = NULL;
}

/*************************************************************************
function: ElfIntrJanitor - Release temp allocs, check for breakpoints
This is not a dumping ground!!!  It only gets gets when a debugger
is present or temp memory has been allocated as a result of
coercing datatypes.  This has a non-trivial impact on interpreter
performance.
The RunJanitoring flag is set when temp data is created
or when a debugger comes on-line.
*/
ElfIntrJanitor()
{
	register garbage_info *next;

	while (Elf.GarbageList) {   /* Free per-instruction malloced data */
		next = Elf.GarbageList->next;
		free(Elf.GarbageList);
		Elf.GarbageList = next;
	}
	if (ElfNumDebuggers == 0)
		Elf.RunJanitoring = FALSE;
}





#if AX_X11
#define CHECK_FOR_SPECIAL_WINDOWS_MESSAGES
#else
#if !FOR_TESTER
extern int Special_Check_Limit;
#define CHECK_FOR_SPECIAL_WINDOWS_MESSAGES if ( Special_Check_Limit > 0 ) \
                     CheckForSpecialMessages();
#else
#define CHECK_FOR_SPECIAL_WINDOWS_MESSAGES
#endif
#endif



/*************************************************************************
function:  ElfRun - Execute ELF command
*/
static data_ptr 
ElfRun(retCode)
	int            *retCode;    /* Error code or AOK */
{
	register int    word2;
	register shrt   opcode;
	register data_ptr p, *pp;
	register data_ptr p1, p2;
	register int    x, y;
	register char  *name;
	register int    hashval;
	double          dx, dy;
	shrt           *oldpc;
	int             memTask;
	char           *str1, *str2;

	*retCode = AOK;             /* AOK assumption */
	if (setjmp(Elf.abortBuf)) { /* so much for assumptions */
		if (!Elf.curFrame || Elf.curFrame->stop_here) {
			p = ELF_CREATE_ARRAY(4);
			*retCode = Elf.ErrorCode;
			ElfAddStrToArray(p, 0, Elf.ErrorStr);
			ElfAddStrToArray(p, 1, Elf.ErrorObject);
			ElfAddNumToArray(p, 2, Elf.ErrorLine);
			ElfAddStrToArray(p, 3, Elf.ErrorFile);
			return (p);
		}
	}
next_step:
	if (Elf.RunJanitoring)
		ElfIntrJanitor();
no_garbage:
do_quick:

#if ELF_INTR_LOG
	oldpc = Elf.pc;
#endif

	opcode = *Elf.pc++;
	word2 = *((short *) Elf.pc++);

#if ELF_INTR_LOG
	ElfIntrLogger(oldpc, opcode, word2);
#endif

	switch (opcode) {

	case OP_NOP:
		Elf.pc--;               /* fixup because word2 was pre-fetched */

		goto next_step;
	case OP_BCALL0:
		CallElfb(ElfCoreFuncs + word2, 0, 0);
		goto next_step;
	case OP_BCALL:
		CallElfb(ElfCoreFuncs + word2, *Elf.pc, 0);
		goto next_step;
	case OP_KCALL0:
		abort();
		goto next_step;
	case OP_KCALL:
		abort();
		goto next_step;
	case OP_RBCALL0:
		CallElfb(ElfCoreFuncs + word2, 0, *Elf.pc);
		goto next_step;
	case OP_RBCALL:
		CallElfb(ElfCoreFuncs + word2, *Elf.pc, *(Elf.pc + 1));
		goto next_step;
	case OP_TRCALL:
		p = ELF_CREATE_STRING(XLT(PopStr()));
		Assign(PushPtr(Elf.curFrame->locvars + word2), p);
		goto next_step;
	case OP_XTRCALL:
		str1 = PopStr();        /* first arg == "tag"  */
		str2 = PopStr();        /* second arg == "string" */
		p = ELF_CREATE_STRING(XLT2(str1, str2));
		Assign(PushPtr(Elf.curFrame->locvars + word2), p);
		goto next_step;
	case OP_SQL:
		Elf.pc += ElfSQL(Elf.pc, word2);
		goto next_step;
	case OP_RKCALL0:
		abort();
		goto next_step;
	case OP_RKCALL:
		abort();
		goto next_step;
	case OP_CALL0:
		CallSub(word2, 0, 0, Elf.curFrame->inObj);
		goto next_step;
	case OP_CALL0_EXT:
		CallExtSub(EXTPTR, 0, 0, Elf.curFrame->inObj, NULL);
		goto next_step;
	case OP_0ENTRYCALL:
		CallEntrySub(EX(POP_VAL), 0, 0, Elf.curFrame->inObj);
		goto next_step;
	case OP_CALL:
		CallSub(word2, *Elf.pc, 0, Elf.curFrame->inObj);
		goto next_step;
	case OP_CALL_EXT:
		CallExtSub(EXTPTR, *Elf.pc, 0, Elf.curFrame->inObj, NULL);
		goto next_step;
	case OP_ENTRYCALL:
		CallEntrySub(EX(POP_VAL), word2, 0, Elf.curFrame->inObj);
		goto next_step;
	case OP_RCALL0:
		CallSub(word2, 0, *Elf.pc, Elf.curFrame->inObj);
		goto next_step;
	case OP_RCALL0_EXT:
		CallExtSub(EXTPTR, 0, *Elf.pc, Elf.curFrame->inObj, NULL);
		goto next_step;
	case OP_0RENTRYCALL:
		CallEntrySub(EX(POP_VAL), 0, word2, Elf.curFrame->inObj);
		goto next_step;
	case OP_RCALL:
		CallSub(word2, *Elf.pc, *(Elf.pc + 1), Elf.curFrame->inObj);
		goto next_step;
	case OP_RCALL_EXT:
		CallExtSub(EXTPTR, *Elf.pc, *(Elf.pc + 1), Elf.curFrame->inObj, NULL);
		goto next_step;
	case OP_RENTRYCALL:
		CallEntrySub(EX(POP_VAL), word2, *Elf.pc, Elf.curFrame->inObj);

		goto next_step;
	case OP_JZ:
	case OP_JZ_SHRT:
		p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
		if (IS_INTEGER(p1)) {
			Elf.sp--;
			if (GET_INTEGER(p1) == 0)
				Elf.pc += word2;
		} else {
			if (PopNumVal() == 0.0)
				Elf.pc += word2;
		}
		goto next_step;
	case OP_JMP:
	case OP_JMP_SHRT:
		Elf.pc += word2;

		goto next_step;
	case OP_SLESS:
		PushTempInt(word2, StrDiffTester() < 0 ? BTRUE : BFALSE);
		goto next_step;
	case OP_SGRTR:
		PushTempInt(word2, StrDiffTester() > 0 ? BTRUE : BFALSE);
		goto next_step;
	case OP_SLEQ:
		PushTempInt(word2, StrDiffTester() <= 0 ? BTRUE : BFALSE);
		goto next_step;
	case OP_SGEQ:
		PushTempInt(word2, StrDiffTester() >= 0 ? BTRUE : BFALSE);

/* For the following OP_PUSH instructions, a ptr to the data ptr is pushed. */

		goto next_step;
	case OP_PUSH:
		PushPtr(&Elf.curFrame->locvars[word2]);
		DUMPTOS();
		goto do_quick;
	case OP_PUSH_ARG:
		PushPtr(Elf.curFrame->argvars[word2]);
		DUMPTOS();
		goto do_quick;
	case OP_PUSH_GLO:
		PushPtr(&Elf.taskFrame[GLOVARDEX]);
		DUMPTOS();
		goto do_quick;
	case OP_PUSH_EXT:
		PushPtr(Elf.taskFrame + EXTDEX);
		DUMPTOS();
		goto do_quick;
	case OP_PUSH_OBJ:
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		PushPtr((long)ElfGetObjData(Elf.curFrame->inObj, name, hashval)|FAULT_BIT);
		DUMPTOS();
		goto do_quick;
	case OP_PUSH_SYS:
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		PushPtr((long)ElfGetSysVarPP(name,hashval)|FAULT_BIT);
		DUMPTOS();
		goto do_quick;
	case OP_PUSH_LIT:
		PushPtr((long) (Elf.curFrame->packP->literals_table + word2) | FAULT_BIT);
		DUMPTOS();
		goto do_quick;
	case OP_PUSH_NUM:
		PushPtr((long) (ElfQNumbers + word2) | FAULT_BIT);
		goto do_quick;
	case OP_PUSH_ENT:
		{
			shrt            word3;
			word3 = *Elf.pc++;
			PushEntryBlock(word3, Elf.curFrame->packP->literals_table + word2);
			goto do_quick;
		}

/* In the following OP_POP instructions, the tos is assumed to be a dataptr
	to the data to assign to the variable.  */

	case OP_POPTO:
		{
			register data_ptr old_val;
			pp = &Elf.curFrame->locvars[word2];
	popto: old_val = *pp;
			p = POP_VAL;
			*pp = (!XX(p) || IS_INTEGER(p)) ? p : ElfDataCopy(p);
			if (XX(old_val) && !IS_INTEGER(old_val))
				ELF_DATA_FREE(old_val);
			goto no_garbage;
		}
	case OP_POPTO_ARG:
		pp = Elf.curFrame->argvars[word2];
		VERIFY((pp != &NullDataPtr) && (((long) pp & FAULT_BIT) == 0), ERR_NOTVAR);
		goto popto;
	case OP_POPTO_GLO:
		pp = &Elf.taskFrame[GLOVARDEX];
		goto popto;
	case OP_POPTO_EXT:
		pp = Elf.taskFrame + EXTDEX;
		goto popto;
	case OP_POPTO_OBJ:
		{
			register data_ptr old_val;
			name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
			hashval = *Elf.pc++;
			pp = &ElfObjData(Elf.curFrame->inObj, name, hashval)->value;
			old_val = *pp;
			p = POP_VAL;
			*pp = (!XX(p) || IS_INTEGER(p)) ? p : AxTaskCopyElfData(0, p);
			if (XX(old_val) && !IS_INTEGER(old_val))
				AxTaskFreeElfData(0, old_val);
			goto no_garbage;
		}
	case OP_POPTO_SYS:
		{
			register data_ptr old_val;
			name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
			hashval = *Elf.pc++;
			pp = ElfGetSysVarPP(name, hashval);
			old_val = *pp;
			p = POP_VAL;
			*pp = (!XX(p) || IS_INTEGER(p)) ? p : AxTaskCopyElfData(0, p);
			if (XX(old_val) && !IS_INTEGER(old_val))
				AxTaskFreeElfData(0, old_val);
			goto no_garbage;
		}

/* For the return instruction, the optional return value is a dataptr on
	the tos.  */

	case OP_RETURN:
        CHECK_FOR_SPECIAL_WINDOWS_MESSAGES
		DoReturn(NULL);
		if (!Elf.curFrame || Elf.curFrame->stop_here)
			return (NULL);
		if (Elf.curFrame->objVals) {
			Elf.pc = Elf.curFrame->objLoopPC;
			pp = &(AX(Elf.curFrame->objVals)->
				   vals[Elf.curFrame->currentObjVal++]);
			goto objloop;
		}
		goto next_step;
	case OP_RETVAL:
        CHECK_FOR_SPECIAL_WINDOWS_MESSAGES
		p = DoReturn(POP_VAL);
		if (!Elf.curFrame || Elf.curFrame->stop_here)
			return (p);
		if (Elf.curFrame->objVals) {
			Elf.pc = Elf.curFrame->objLoopPC;
			pp = &(AX(Elf.curFrame->objVals)->
				   vals[Elf.curFrame->currentObjVal++]);
			goto objloop;
		}
		goto next_step;
	case OP_PUSH_NULL:
		Elf.pc--;               /* fixup because word2 was pre-fetched */
		PushPtr(&NullDataPtr);
		goto do_quick;

/* In the following OP_PUSHARR instructions, an array reference is pushed.
	The pushed data is address of the dataptr to the actual data.
	The third word of the instruction is a "number of indexes" word. */

	case OP_PUSH1ARR:
		{
			register int    dex;

			p = Elf.curFrame->locvars[word2];
push1p:     dex = POP_INT;
			if (XX(p)) {
				VERIFY(!FAULTED(p) &&
					   (AX(p)->type == TYPE_ARRAY), ERR_NOTARRAY);
				VERIFY(dex >= 0, ERR_BADARRDEX);
				if (dex >= AX(p)->size)
					PushPtr(&NullDataPtr);
				else
					PushPtr(AX(p)->vals + dex);
				DUMPTOS();
			} else
				PushPtr(&NullDataPtr);
		}
		goto next_step;
	case OP_PUSH1ARR_ARG:
		p = *((data_ptr *) ((long) Elf.curFrame->argvars[word2] & ~FAULT_BIT));
		goto push1p;
	case OP_PUSH1ARR_GLO:
		p = Elf.taskFrame[GLOVARDEX];
		goto push1p;
	case OP_PUSH1ARR_EXT:
		p = Elf.taskFrame[EXTDEX];
		goto push1p;
	case OP_PUSH1ARR_OBJ:
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		p = *ElfGetObjData(Elf.curFrame->inObj, name, hashval);
		goto push1p;
	case OP_PUSH1ARR_SYS:
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		p = *ElfGetSysVarPP(name, hashval);
		goto push1p;

	case OP_PUSHARR:
		{
			register int    num_refs, dex;

			p = Elf.curFrame->locvars[word2];
	pushap:num_refs = *Elf.pc++;
			while (--num_refs >= 0) {
				dex = POP_INT;
				if (XX(p)) {
					VERIFY(!FAULTED(p) &&
						   (AX(p)->type == TYPE_ARRAY), ERR_NOTARRAY);
					VERIFY(dex >= 0, ERR_BADARRDEX);
					if (dex >= AX(p)->size)
						p = *(pp = &NullDataPtr);
					else
						p = *(pp = AX(p)->vals + dex);
				} else
					p = *(pp = &NullDataPtr);
			}
			if (pp == &NullDataPtr)
				PushPtr(&NullDataPtr);
			else
				PushPtr(pp);
			DUMPTOS();
		}
		goto next_step;
	case OP_PUSHARR_ARG:
		p = *((data_ptr *) ((long) Elf.curFrame->argvars[word2] & ~FAULT_BIT));
		goto pushap;
	case OP_PUSHARR_GLO:
		p = Elf.taskFrame[GLOVARDEX];
		goto pushap;
	case OP_PUSHARR_EXT:
		p = Elf.taskFrame[EXTDEX];
		goto pushap;
	case OP_PUSHARR_OBJ:
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		p = *ElfGetObjData(Elf.curFrame->inObj, name, hashval);
		goto pushap;
	case OP_PUSHARR_SYS:
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		p = *ElfGetSysVarPP(name, hashval);
		goto pushap;

/* In the following OP_POPDATA instructions, an array is created by popping
	a bunch of datums from the stack */

	case OP_POPDATA:
		memTask = THIMpid;
		pp = &Elf.curFrame->locvars[word2];
popdat:word2 = *Elf.pc++;
        p = AxTaskCreateElfArray(memTask, word2);
		for (word2 = word2 - 1; word2 >= 0; word2--)
			WriteArray(p, word2, AxTaskCopyElfData(memTask,POP_VAL));
		TaskAssign(memTask, pp, p);

		goto next_step;
	case OP_POPDATA_ARG:
		memTask = THIMpid;
		pp = Elf.curFrame->argvars[word2];
		VERIFY((pp != &NullDataPtr) && (((long) pp & FAULT_BIT) == 0), ERR_NOTVAR);
		goto popdat;
	case OP_POPDATA_GLO:
		memTask = THIMpid;
		pp = &Elf.taskFrame[GLOVARDEX];
		goto popdat;
	case OP_POPDATA_EXT:
		memTask = THIMpid;
		pp = Elf.taskFrame + EXTDEX;
		goto popdat;
	case OP_POPDATA_OBJ:
		memTask = 0;
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		pp = &ElfObjData(Elf.curFrame->inObj, name, hashval)->value;
		goto popdat;
	case OP_POPDATA_SYS:
		memTask = 0;
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		pp = ElfGetSysVarPP(name, hashval);
		goto popdat;

	case OP_POPDARR:
		memTask = THIMpid;
		pp = &Elf.curFrame->locvars[word2];
		goto popdar;
	case OP_POPDARR_ARG:
		memTask = THIMpid;
		pp = Elf.curFrame->argvars[word2];
		VERIFY((pp != &NullDataPtr) && (((long) pp & FAULT_BIT) == 0), ERR_NOTVAR);
		goto popdar;
	case OP_POPDARR_GLO:
		memTask = THIMpid;
		pp = &Elf.taskFrame[GLOVARDEX];
		goto popdar;
	case OP_POPDARR_EXT:
		memTask = THIMpid;
		pp = Elf.taskFrame + EXTDEX;
		goto popdar;
	case OP_POPDARR_OBJ:
		memTask = 0;
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		pp = &ElfObjData(Elf.curFrame->inObj, name, hashval)->value;
		goto popdar;
	case OP_POPDARR_SYS:
		memTask = 0;
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		pp = ElfGetSysVarPP(name, hashval);
		goto popdar;

/* In the following OP_POPTOARR instructions, a datum is popped into
	an array reference as in "foo[1,5] = 9"*/

	case OP_POPTO1AR:
		memTask = THIMpid;
		pp = &Elf.curFrame->locvars[word2];
		goto poptar1;
	case OP_POPTO1AR_ARG:
		memTask = THIMpid;
		pp = Elf.curFrame->argvars[word2];
		VERIFY((pp != &NullDataPtr) && (((long) pp & FAULT_BIT) == 0), ERR_NOTVAR);
		goto poptar1;
	case OP_POPTO1AR_GLO:
		memTask = THIMpid;
		pp = &Elf.taskFrame[GLOVARDEX];
		goto poptar1;
	case OP_POPTO1AR_EXT:
		memTask = THIMpid;
		pp = Elf.taskFrame + EXTDEX;
		goto poptar1;
	case OP_POPTO1AR_OBJ:
		memTask = 0;
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		pp = &ElfObjData(Elf.curFrame->inObj, name, hashval)->value;
		goto poptar1;
	case OP_POPTO1AR_SYS:
		memTask = 0;
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		pp = ElfGetSysVarPP(name, hashval);
		goto poptar1;
	case OP_POPTOARR:
		{
			data_ptr        workP, *finishP;
			register data_ptr dP;
			register data_ptr *refP;
			register int    num_refs, data_block, dex;
			data_ptr        to_free;

			memTask = THIMpid;
			pp = &Elf.curFrame->locvars[word2];

	poptar:num_refs = *Elf.pc++;        /* pop 1 datum to multi-dim array */
			data_block = FALSE;
			goto dopoptoar;

	poptar1:num_refs = 1;       /* pop 1 datum to 1-dim array */
			data_block = FALSE;
			goto dopoptoar;

	popdar:num_refs = *Elf.pc++;        /* pop block to multi-dim array */
			data_block = TRUE;
			goto dopoptoar;

	dopoptoar:

/* create array if var is null or not currently an array */

			finishP = pp;       /* note where array data_ptr will go */
			workP = *pp;        /* make a temp data_ptr to work with */
			pp = &workP;        /* because the POP_VAL can hammer foo[0] = foo */
			if (!(XX(*pp)) ||
				FAULTED(*pp) ||
				(XX(*pp)->type != TYPE_ARRAY)) {
				to_free = *pp;
				*pp = AxTaskCreateElfArray(memTask, 0);
			} else
				to_free = NULL;

/* walk arrays, optionally creating or resizing */

			refP = pp;
			dP = *refP;
			while (--num_refs > 0) {
				dex = POP_INT;
				VERIFY(dex >= 0, ERR_BADARRDEX);
				if ((AX(dP)->size <= dex) ||
					!AX(AX(dP)->vals[dex]) ||
					(LX(AX(dP)->vals[dex]) & FAULT_BIT) ||
					(XX(AX(dP)->vals[dex])->type != TYPE_ARRAY)) {
					p = AxTaskCreateElfArray(memTask, 0);
					*refP = dP = TaskWriteArray(memTask, dP, dex, p);
				}
				refP = AX(dP)->vals + dex;
				dP = *refP;
			}
			dex = POP_INT;
			VERIFY(dex >= 0, ERR_BADARRDEX);
			if (!data_block)
				*refP = TaskWriteArray(memTask, dP, dex,
									   AxTaskCopyElfData(memTask, POP_VAL));
			else {
				register data_ptr aP;
				register int    count;
				aP = AxTaskCreateElfArray(memTask, *Elf.pc);
				for (count = (*Elf.pc++) - 1; count >= 0; count--)
					aP = TaskWriteArray(memTask, aP, count,
										AxTaskCopyElfData(memTask, POP_VAL));
				*refP = TaskWriteArray(memTask, dP, dex, aP);
			}
			if (XX(to_free))
				AxTaskFreeElfData(memTask, to_free);
			*finishP = workP;
		}
		goto next_step;
	case OP_POPTOARR_ARG:
		memTask = THIMpid;
		pp = Elf.curFrame->argvars[word2];
		VERIFY((pp != &NullDataPtr) && (((long) pp & FAULT_BIT) == 0), ERR_NOTVAR);
		goto poptar;
	case OP_POPTOARR_GLO:
		memTask = THIMpid;
		pp = &Elf.taskFrame[GLOVARDEX];
		goto poptar;
	case OP_POPTOARR_EXT:
		memTask = THIMpid;
		pp = Elf.taskFrame + EXTDEX;
		goto poptar;
	case OP_POPTOARR_OBJ:
		memTask = 0;
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		pp = &ElfObjData(Elf.curFrame->inObj, name, hashval)->value;
		goto poptar;
	case OP_POPTOARR_SYS:
		memTask = 0;
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		pp = ElfGetSysVarPP(name, hashval);
		goto poptar;

/* OP_COMPCALL implements the computed call statement.  A routines count
	follows the opcode, followed by the routine calls themselves.  Each routine
		call is followed by a JMP to the next instruction.  Routine calls that
		are single word are padded out with OP_NOP.  Thus the total storage for
		each routine call is 4 words (2 for call/nop, 2 for jmp). */


	case OP_COMPCALL:
		{
			register bint   expr;

			expr = POP_INT;
			VERIFY((expr >= 0) && (expr <= 255), ERR_BADONCALL);
			if ((expr != 0) && (expr <= word2))
				Elf.pc += 4 * (expr - 1);
			else
				Elf.pc += 4 * word2;
		}

/* OP_COMPGOTO performs the computed goto. Like OP_COMPCALL, the tos val must
	be in the range of 0 to 255 and is used (one based) to select the label
	to branch to. */

		goto next_step;
	case OP_COMPGOTO:
		{
			register bint   expr;
			expr = POP_INT;
			VERIFY((expr >= 0) && (expr <= 255), ERR_BADONGOTO);
			if ((expr != 0) && (expr <= word2)) {
				expr--;
				Elf.pc += *((short *) (Elf.pc + expr)) + expr + 1;
			} else
				Elf.pc += word2;
		}

/* The arithmetic operations all work pretty much the same.  They are
	implemented in floating point except where noted. */

		goto next_step;
	case OP_IADD:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, x + y);
		goto next_step;
	case OP_ISUB:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, x - y);
		goto next_step;
	case OP_IMULT:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, x * y);
		goto next_step;
	case OP_IDIV:
		dx = PopNumVal();
		dy = PopNumVal();
		VERIFY(fabs(dy) > EPSILON, ERR_ZERODIVIDE);
		x = (int)(dx/dy);
		PushTempInt(word2, x);
		goto next_step;
	case OP_ADD:
		p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
		p2 = *((data_ptr *) ((long) (*(Elf.sp - 2)) & ~FAULT_BIT));
		if (IS_INTEGER(p1) && IS_INTEGER(p2)) {
			Elf.sp -= 2;
			x = GET_INTEGER(p1) + GET_INTEGER(p2);
			if (C_INT_IS_ELF_INT(x))
				PUSHQUICK(word2, x);
			else
				PushTempNum(word2, (double) x);
		} else {
			dx = PopNumVal();
			dy = PopNumVal();
			PushTempNum(word2, dx + dy);
		}
		goto next_step;
	case OP_SUB:
		p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
		p2 = *((data_ptr *) ((long) (*(Elf.sp - 2)) & ~FAULT_BIT));
		if (IS_INTEGER(p1) && IS_INTEGER(p2)) {
			Elf.sp -= 2;
			x = GET_INTEGER(p1) - GET_INTEGER(p2);
			if (C_INT_IS_ELF_INT(x))
				PUSHQUICK(word2, x);
			else
				PushTempNum(word2, (double) x);
		} else {
			dx = PopNumVal();
			dy = PopNumVal();
			PushTempNum(word2, dx - dy);
		}
		goto next_step;
	case OP_EXP:
		dx = PopNumVal();
		dy = PopNumVal();
		PushTempNum(word2, do_pow(dx, dy));
		goto next_step;
	case OP_UMINUS:
		p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
		if (IS_INTEGER(p1)) {
			Elf.sp--;
			PushTempInt(word2, -GET_INTEGER(p1));
		} else
			PushTempNum(word2, -PopNumVal());
		goto next_step;
	case OP_IUMINUS:
		PushTempInt(word2, -POP_INT);
		goto next_step;
	case OP_MULT:
		p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
		p2 = *((data_ptr *) ((long) (*(Elf.sp - 2)) & ~FAULT_BIT));
		if (IS_INTEGER(p1) && IS_INTEGER(p2)) {
			x = GET_INTEGER(p1);
			y = GET_INTEGER(p2);
			if (((x >= -16384) && (x <= 16383))
				&& ((y >= -16384) && (y <= 16383))) {
				Elf.sp -= 2;
				x = x * y;
				if (C_INT_IS_ELF_INT(x))
					PUSHQUICK(word2, x);
				else
					PushTempNum(word2, (double) x);
				goto do_quick;
			}
		}
		dx = PopNumVal();
		dy = PopNumVal();
		VERIFY((fabs(dx) < MAX_ELFNUM) && (fabs(dy) < MAX_ELFNUM), ERR_ARITHMETIC);
		PushTempNum(word2, dx * dy);
		goto next_step;
	case OP_DIV:
		dx = PopNumVal();
		dy = PopNumVal();
		VERIFY(fabs(dy) > EPSILON, ERR_ZERODIVIDE);
		PushTempNum(word2, dx / dy);
		goto next_step;
	case OP_REM:
		x = POP_INT;
		y = POP_INT;
		VERIFY(y != 0, ERR_ZERODIVIDE)
			PushTempInt(word2, x % y);

/* The logical operators all work with 16 bit integer values.  They perform
	bitwise operations. Note that TRUE is 1111111 and FALSE is 0000000. */

		goto next_step;
	case OP_NOT:
		PushTempInt(word2, ~POP_INT);
		goto next_step;
	case OP_EQV:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, ~(x ^ y));
		goto next_step;
	case OP_IMP:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, ~(x & ~y));
		goto next_step;
	case OP_AND:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, x & y);
		goto next_step;
	case OP_OR:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, x | y);
		goto next_step;
	case OP_XOR:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, x ^ y);


/* The relational operators are primarily floating point except for OP_EQUALS
	and OP_NEQ which perform string tests is either expression is a string.
	The pushed value for all relation ops is either TRUE (111...) or FALSE (0).
*/

		goto next_step;
	case OP_GEQ:
		p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
		p2 = *((data_ptr *) ((long) (*(Elf.sp - 2)) & ~FAULT_BIT));
		if (IS_INTEGER(p1) && IS_INTEGER(p2)) {
			Elf.sp -= 2;
			PUSHQUICK(word2, GET_INTEGER(p1) >= GET_INTEGER(p2) ? BTRUE : FALSE);
		} else {
			dx = PopNumVal();
			dy = PopNumVal();
			PushTempInt(word2, dx >= dy ? BTRUE : BFALSE);
		}
		goto next_step;
	case OP_GRTR:
		p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
		p2 = *((data_ptr *) ((long) (*(Elf.sp - 2)) & ~FAULT_BIT));
		if (IS_INTEGER(p1) && IS_INTEGER(p2)) {
			Elf.sp -= 2;
			PUSHQUICK(word2, GET_INTEGER(p1) > GET_INTEGER(p2) ? BTRUE : FALSE);
		} else {
			dx = PopNumVal();
			dy = PopNumVal();
			PushTempInt(word2, dx > dy ? BTRUE : BFALSE);
		}
		goto next_step;
	case OP_LEQ:
		p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
		p2 = *((data_ptr *) ((long) (*(Elf.sp - 2)) & ~FAULT_BIT));
		if (IS_INTEGER(p1) && IS_INTEGER(p2)) {
			Elf.sp -= 2;
			PUSHQUICK(word2, GET_INTEGER(p1) <= GET_INTEGER(p2) ? BTRUE : FALSE);
		} else {
			dx = PopNumVal();
			dy = PopNumVal();
			PushTempInt(word2, dx <= dy ? BTRUE : BFALSE);
		}
		goto next_step;
	case OP_LESS:
		p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
		p2 = *((data_ptr *) ((long) (*(Elf.sp - 2)) & ~FAULT_BIT));
		if (IS_INTEGER(p1) && IS_INTEGER(p2)) {
			Elf.sp -= 2;
			PUSHQUICK(word2, GET_INTEGER(p1) < GET_INTEGER(p2) ? BTRUE : FALSE);
		} else {
			dx = PopNumVal();
			dy = PopNumVal();
			PushTempInt(word2, dx < dy ? BTRUE : BFALSE);
		}
		goto next_step;
	case OP_IGEQ:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, x >= y ? BTRUE : BFALSE);
		goto next_step;
	case OP_IGRTR:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, x > y ? BTRUE : BFALSE);
		goto next_step;
	case OP_ILEQ:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, x <= y ? BTRUE : BFALSE);
		goto next_step;
	case OP_ILESS:
		x = POP_INT;
		y = POP_INT;
		PushTempInt(word2, x < y ? BTRUE : BFALSE);

		goto next_step;
	case OP_EQUALS:
		{
			p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
			p2 = *((data_ptr *) ((long) (*(Elf.sp - 2)) & ~FAULT_BIT));

			if (p1 == p2) {
				Elf.sp -= 2;
				PUSHQUICK(word2, BTRUE);
				goto do_quick;
			}
/* attempt to compare two small integers */

			if (IS_INTEGER(p1)) {
				if (IS_INTEGER(p2)) {
					Elf.sp -= 2;
					PUSHQUICK(word2, FALSE);
					goto do_quick;
				} else {
					p2 = (data_ptr) ((long) p2 & ~FAULT_BIT);
					if (!p2) {
						Elf.sp -= 2;
						PUSHQUICK(word2, (GET_INTEGER(p1) == 0) ? BTRUE : FALSE);
						goto do_quick;
					} else if (p2->type == TYPE_STRING)
						goto str_eq;
					else
						goto try_floats;
				}
			}
			if (IS_INTEGER(p2)) {
				if (IS_INTEGER(p1)) {
					Elf.sp -= 2;
					PUSHQUICK(word2, FALSE);
					goto do_quick;
				} else {
					p1 = (data_ptr) ((long) p1 & ~FAULT_BIT);
					if (!p1) {
						Elf.sp -= 2;
						PUSHQUICK(word2, (GET_INTEGER(p2) == 0) ? BTRUE : FALSE);
						goto do_quick;
					} else if (p1->type == TYPE_STRING)
						goto str_eq;
					else
						goto try_floats;
				}
			}
			p1 = (data_ptr) ((long) p1 & ~FAULT_BIT);
			p2 = (data_ptr) ((long) p2 & ~FAULT_BIT);

/* not small integers, try as two strings if either are strings */

			if ((p1 && (p1->type == TYPE_STRING)) ||
				(p2 && (p2->type == TYPE_STRING))) {
		str_eq:PushTempInt(word2, StrEqualTester() == 0 ? BTRUE : BFALSE);
				goto next_step;
			}
/* not strings, try as arrays if either is an array */

			if ((p1 && (p1->type == TYPE_ARRAY)) ||
				(p2 && (p2->type == TYPE_ARRAY))) {
				Elf.sp -= 2;
				PushTempInt(word2, ElfSameArray(p1, p2) ? BTRUE : BFALSE);
				goto next_step;
			}
/* not arrays, try as binary data */

			if ((p1 && (p1->type == TYPE_BINARY)) ||
				(p2 && (p2->type == TYPE_BINARY))) {
				Elf.sp -= 2;
				PushTempInt(word2, ElfSameBinary(p1, p2) ? BTRUE : BFALSE);
				goto next_step;
			}
/* not binaries, try as objects */

			if ((p1 && (p1->type == TYPE_OBJECT)) ||
				(p2 && (p2->type == TYPE_OBJECT))) {
				Elf.sp -= 2;
				PushTempInt(word2, ElfSameObject(p1, p2) ? BTRUE : BFALSE);
				goto next_step;
			}
/* not objects, assume floating point numbers */

			else {
		try_floats:
				PushTempInt(word2, PopNumVal() == PopNumVal() ? BTRUE : BFALSE);
			}
		}

		goto next_step;
	case OP_NEQ:                /* ??x change algorthim to match OP_EQUALS */
		{
/* first attempt to compare two small integers */

			p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
			p2 = *((data_ptr *) ((long) (*(Elf.sp - 2)) & ~FAULT_BIT));

			if (p1 == p2) {
				Elf.sp -= 2;
				PUSHQUICK(word2, FALSE);
				goto do_quick;
			}
			if (IS_INTEGER(p1)) {
				if (IS_INTEGER(p2)) {
					Elf.sp -= 2;
					PUSHQUICK(word2, BTRUE);
					goto do_quick;
				} else {
					p2 = (data_ptr) ((long) p2 & ~FAULT_BIT);
					if (!p2) {
						Elf.sp -= 2;
						PUSHQUICK(word2, (GET_INTEGER(p1) != 0) ? BTRUE : FALSE);
						goto do_quick;
					} else if (p2->type == TYPE_STRING)
						goto str_neq;
					else
						goto try_nfloats;
				}
			}
			if (IS_INTEGER(p2)) {
				if (IS_INTEGER(p1)) {
					Elf.sp -= 2;
					PUSHQUICK(word2, BTRUE);
					goto do_quick;
				} else {
					p1 = (data_ptr) ((long) p1 & ~FAULT_BIT);
					if (!p1) {
						Elf.sp -= 2;
						PUSHQUICK(word2, (GET_INTEGER(p2) != 0) ? BTRUE : FALSE);
						goto do_quick;
					} else if (p1->type == TYPE_STRING)
						goto str_neq;
					else
						goto try_nfloats;
				}
			}
			p1 = (data_ptr) ((long) p1 & ~FAULT_BIT);
			p2 = (data_ptr) ((long) p2 & ~FAULT_BIT);

/* not small integers, try as two strings if either are strings */

			if ((p1 && (p1->type == TYPE_STRING)) ||
				(p2 && (p2->type == TYPE_STRING))) {
		str_neq:PushTempInt(word2, StrEqualTester() != 0 ? BTRUE : BFALSE);
				goto next_step;
			}
/* not strings, try as arrays if either is an array */

			if ((p1 && (p1->type == TYPE_ARRAY)) ||
				(p2 && (p2->type == TYPE_ARRAY))) {
				Elf.sp -= 2;
				PushTempInt(word2, !ElfSameArray(p1, p2) ? BTRUE : BFALSE);
				goto next_step;
			}
/* not arrays, try as binary if either is an array */

			if ((p1 && (p1->type == TYPE_BINARY)) ||
				(p2 && (p2->type == TYPE_BINARY))) {
				Elf.sp -= 2;
				PushTempInt(word2, !ElfSameBinary(p1, p2) ? BTRUE : BFALSE);
				goto next_step;
			}
/* not binaries, try as objects */

			if ((p1 && (p1->type == TYPE_OBJECT)) ||
				(p2 && (p2->type == TYPE_OBJECT))) {
				Elf.sp -= 2;
				PushTempInt(word2, !ElfSameObject(p1, p2) ? BTRUE : BFALSE);
				goto next_step;
			}
/* not objects, assume floating point numbers */

			else {
		try_nfloats:
				PushTempInt(word2, PopNumVal() != PopNumVal() ? BTRUE : BFALSE);
			}
		}

		goto next_step;
	case OP_IEQUALS:
		{
			p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
			p2 = *((data_ptr *) ((long) (*(Elf.sp - 2)) & ~FAULT_BIT));
			if (p1 == p2) {
				PushTempInt(word2, BTRUE);
				goto do_quick;
			}
			if (IS_INTEGER(p1) && IS_INTEGER(p2)) {     /* !!0 optimize */
				Elf.sp -= 2;
				PushTempInt(word2, (GET_INTEGER(p1) == GET_INTEGER(p2)) ? BTRUE : FALSE);
			} else {
				if ((!IS_INTEGER(p1) &&
					 (p1 = (data_ptr) (LX(p1) & ~FAULT_BIT)) &&
					 (SX(p1)->type == TYPE_STRING)) ||
					(!IS_INTEGER(p2) &&
					 (p2 = (data_ptr) (LX(p2) & ~FAULT_BIT)) &&
					 (SX(p2)->type == TYPE_STRING)))
					PushTempInt(word2, StrEqualTester() == 0 ? BTRUE : BFALSE);
				else
					PushTempInt(word2, POP_INT == POP_INT ? BTRUE : BFALSE);
			}
		}

		goto next_step;
	case OP_INEQ:
		{
			p1 = *((data_ptr *) ((long) (*(Elf.sp - 1)) & ~FAULT_BIT));
			p2 = *((data_ptr *) ((long) (*(Elf.sp - 2)) & ~FAULT_BIT));
			if (IS_INTEGER(p1) && IS_INTEGER(p2)) {
				Elf.sp -= 2;
				PushTempInt(word2, (GET_INTEGER(p1) != GET_INTEGER(p2)) ? BTRUE : FALSE);
			} else {
				if ((!IS_INTEGER(p1) &&
					 (p1 = (data_ptr) (LX(p1) & ~FAULT_BIT)) &&
					 (SX(p1)->type == TYPE_STRING)) ||
					(!IS_INTEGER(p2) &&
					 (p2 = (data_ptr) (LX(p2) & ~FAULT_BIT)) &&
					 (SX(p2)->type == TYPE_STRING)))
					PushTempInt(word2, StrEqualTester() != 0 ? BTRUE : BFALSE);
				else
					PushTempInt(word2, POP_INT != POP_INT ? BTRUE : BFALSE);
			}
		}

		goto next_step;
	case OP_SADD:
		{
			register char  *cp1, *cp2;
			register data_ptr p3;
			register int    len;
			cp1 = PopStr();
			cp2 = PopStr();
			len = strlen(cp1) + strlen(cp2);
			p3 = AxTaskCreateSizedElfString(THIMpid, len, "");
			strcpy(SX(p3)->text, cp1);
			strcat(SX(p3)->text, cp2);
			Assign(PushPtr(Elf.curFrame->locvars + word2), p3);
			DUMPTOS();
		}

		goto next_step;
	case OP_SWITCH:{
			register int    ix, cnt;
			register shrt  *sp;

			cnt = *(Elf.pc + word2);
			pp = Elf.curFrame->packP->literals_table;
			sp = Elf.pc + word2 + 1;
			x = POP_INT;
			p = MK_INTEGER(x);
			for (ix = 0; ix < cnt; ix++, sp += 2) {
				if (p == pp[*sp]) {
					Elf.pc = (sp + 2) + *(((short *) sp) + 1);
					goto do_quick;
				}
			}
			Elf.pc += *((short *) Elf.pc) + 1;
		}

		goto next_step;
	case OP_SSWITCH:{
			register int    ix, cnt;
			register shrt  *sp;
			char           *str;

			cnt = *(Elf.pc + word2);
			pp = Elf.curFrame->packP->literals_table;
			sp = Elf.pc + word2 + 1;
			str = PopStr();
			for (ix = 0; ix < cnt; ix++, sp += 2) {
				if (0 == strcmp(str, SX(UNFAULT(pp[*sp]))->text)) {
					Elf.pc = (sp + 2) + *(((short *) sp) + 1);
					goto do_quick;
				}
			}
			Elf.pc += *((short *) Elf.pc) + 1;
		}

/* To make any sense of FOR loops, we need to look at the code the compiler
	emits:

		OP_PUSH stop        ' push the stop value
		OP_PUSH step        ' The step value
		OP_PUSH initval     ' Push initial value expression
		OP_FORLOOP var      ' Set initial value of var
				   stepvar  '   local var alloced as step val
				   stopvar  '   local var alloced as stop val
looper: OP_PUSH var         ' Push var ptr on stack for next inst
		OP_FORTEST endlab   ' Pop val, add step, NULL step & branch out if done
				   stepvar  '  (error if step NULL - illegal loop entry)
				   stopvar

	... statements ...

		OP_PUSH var         ' Push the current value of the loop var
		OP_ADD step         ' Add & push the step value based on control array
		OP_POPTO var        ' Store val into var
		OP_JMP looper       ' goto top of loop
*/

		goto next_step;
	case OP_FORLOOP:
		DupAssign(Elf.curFrame->locvars + word2, POP_VAL);
		DupAssign(Elf.curFrame->locvars + *Elf.pc++, POP_VAL);
		DupAssign(Elf.curFrame->locvars + *Elf.pc++, POP_VAL);
		goto next_step;
	case OP_FORTEST:
		{
			double          total, step, stop;

			total = PopNumVal();
			step = NumFromDataPtr(*(pp = Elf.curFrame->locvars + *Elf.pc++));
			VERIFY(step != 0.0, ERR_BADSTEP);
			stop = NumFromDataPtr(Elf.curFrame->locvars[*Elf.pc++]);
			if (step >= 0.0) {
				if (total > stop) {
					Elf.pc += word2 - 2;        /* -2 cause of Elf.pc++'s above */
					Assign(pp, ELF_CREATE_NUMBER(0.0));
				}
			} else {
				if (total < stop) {
					Elf.pc += word2 - 2;        /* -2 cause of Elf.pc++'s above */
					Assign(pp, ELF_CREATE_NUMBER(0.0));
				}
			}
		}
		goto next_step;
	case OP_IFORTEST:
		{
			bint            total, step, stop;

			total = POP_INT;
			step = IntFromDataPtr(*(pp = Elf.curFrame->locvars + *Elf.pc++));
			VERIFY(step != 0.0, ERR_BADSTEP);
			stop = IntFromDataPtr(Elf.curFrame->locvars[*Elf.pc++]);
			if (step >= 0.0) {
				if (total > stop) {
					Elf.pc += word2 - 2;        /* -2 cause of Elf.pc++'s above */
					Assign(pp, ELF_CREATE_NUMBER(0.0));
				}
			} else {
				if (total < stop) {
					Elf.pc += word2 - 2;        /* -2 cause of Elf.pc++'s above */
					Assign(pp, ELF_CREATE_NUMBER(0.0));
				}
			}
		}

/* ERROR support stuff */

		goto next_step;
	case OP_ONERROR:            /* set error catcher */
		Elf.curFrame->error_catcher = Elf.pc;
		Elf.curFrame->catcher_sp = Elf.sp;
		Elf.curFrame->error_caught = FALSE;
		Elf.pc += word2;

		goto next_step;
	case OP_ERRERR:             /* reset error catcher */
		Elf.curFrame->error_catcher = 0;
		Elf.pc--;               /* fixup because word2 was pre-fetched */

/* OP_ENTER is followed by several size words:

	OP_ENTER
	word2       number of args declared
	word3       number of local vars
	word4       number of registers used
*/
		goto next_step;
	case OP_ENTER:
		{
			register int    args, locs, regs, ix, size;

			Elf.curFrame->funcstarts = Elf.pc - 3;
			Elf.curFrame->functype = *(Elf.pc - 3);     /* type preceeds
														 * OP_ENTER */
			Elf.curFrame->numargs = word2;
			Elf.curFrame->numlocs = locs = *Elf.pc++;
			Elf.curFrame->numregs = regs = *Elf.pc++;   /* registers actually
														 * used */
			Elf.curFrame->varnames = *Elf.pc++;
			args = MAX(word2,Elf.curFrame->num_args_passed);
			size = sizeof(data_ptr) * (args + locs);
			pp = (data_ptr *) ElfMalloc(size);
			memset(pp + MAX_REGS - regs, 0, sizeof(data_ptr) * (locs - MAX_REGS + regs));
			*pp = NULL;         /* return value location */
			Elf.curFrame->locvars = pp;
			Elf.curFrame->argvars = (data_ptr **) (Elf.curFrame->locvars + locs);
			if (ElfInLearnMode)
				ElfRecordCall(Elf.curFrame->functype, Elf.curFrame,
							  Elf.curFrame->funcstarts,
							  MIN(args,Elf.curFrame->num_args_passed));
			for (ix = 0; ix < Elf.curFrame->num_args_passed; ix++)
				Elf.curFrame->argvars[ix] = POP_RAW;
			while (ix < Elf.curFrame->numargs)
				Elf.curFrame->argvars[ix++] = &NullDataPtr;
#if !FOR_TESTER
			if ( ElfProfiling )
				ElfProfileNote(Elf.curFrame);				
#endif
		}
		goto next_step;

	case OP_TOSSTOS:
		Elf.sp--;
		Elf.pc--;               /* fixup because word2 was pre-fetched */
		goto do_quick;

/*
		OP_SETOBJ Instruction.
		The opcode word contains the EA of the object datum just
		like OP_POPTO.

		The words that follow:

				queryDestination        0 if changeMenthod, else result register
				methodName literal offset
				methodHash literal offset
				arg_count of args that where pushed before the OP_SETOBJ instruction
*/
	case OP_SETOBJ:
		{
			register short  literalAt;
			register short  args_passed;
			register shrt   queryDestination;
			register elf_object *obj; /* uninit? */
			register data_ptr runMe; /* uninit? */
			register int    is_broadcast, call_inherit;
			char            nameBuf[ELFIDSTR + 1];
			data_entry      entdata;
			int             isRemote;
			cpp_method_info_t	*cppMethodP;
			data_ptr		ElfCallCppMember();
			data_ptr		ElfCallCorbaMethod();
			data_ptr        remoteDatum, elfCallRemoteObjMethod();
			int 			corbaIx;

			/* get ptr to object to manipulate */

			pp = &Elf.curFrame->locvars[word2];
	setobj:
			p = UNFAULT(*pp);
			if (!IS_INTEGER(*pp) && p && (p->type == TYPE_ARRAY)) {
				if (Elf.curFrame->objVals)
					ElfAbort(ERR_NOTOBJECT);
				Elf.curFrame->objVals = ElfDataCopy(p);
				if (AX(p)->size == 0)
					pp = &NullDataPtr;
				else
					pp = &(AX(p)->vals[0]);
				Elf.curFrame->currentObjVal = 1;        /* primed for next in
														 * array */
				Elf.curFrame->objLoopPC = Elf.pc;
			}
	objloop:
#if !FOR_TESTER
			if ( TaskInfo[AxInTask].alarm_has_fired &&
					(TaskInfo[AxInTask].alarm_mode == ALARM_MODE_RECYCLE) ) {
				AxSleepSeconds(0);
				AxSetTaskAlarm(AxInTask,ALARM_MODE_RECYCLE,
							TaskInfo[AxInTask].alarm_millisec);
			}
#endif	
			queryDestination = *Elf.pc++;
			if (queryDestination && Elf.curFrame->objVals)      /* !!1 resolve */
				ElfStrAbort(ERR_ILLVAL, XLT("Array of objects cannot return a value"), NULL);
			call_inherit = FALSE;
			if ( is_broadcast = *Elf.pc++ ) {	/* call_inherit encoded in */
				if ( is_broadcast & 2 ) {
					call_inherit = TRUE;
					is_broadcast = is_broadcast & ~2;
				}
			}
			

			/* get the method */

			literalAt = (short) (*Elf.pc++);
			if (literalAt == 0) {       /* dynamic method selection */
				name = PopStr();
				while (*name && (literalAt < ELFIDSTR))
					nameBuf[literalAt++] = toupper(*name++);
				nameBuf[literalAt] = 0;
				hashval = hash(name = nameBuf);
			}
			else {
				name = StrFromDataPtr(Elf.curFrame->packP->literals_table[literalAt]);
				hashval = *Elf.pc++;
			}

			/* get the arg count, if negative, flip argument order */

			args_passed = (short) (*Elf.pc++);  /* pushed on stack */
			if (args_passed < 0) {
				args_passed = -args_passed;
				if (!Elf.curFrame->objVals || (Elf.curFrame->currentObjVal == 1)) {
					register data_ptr **lowest, **highest;
					register data_ptr *tmp;

					highest = Elf.sp - 1;
					lowest = highest - args_passed + 1;
					while (lowest < highest) {
						tmp = *highest;
						*highest = *lowest;
						*lowest = tmp;
						highest--;
						lowest++;
					}

				}
			}
			/* if this a subsequent broadcast, copy args from backup area */

			if (Elf.curFrame->objVals && (Elf.curFrame->currentObjVal != 1)) {
				register int    ix;
				for (ix = 0; ix < args_passed; ix++)
					*Elf.sp++ = Elf.curFrame->loopArgs[ix];
			}
			else {

				/* set the fault bit... methods cannot modify args! */

				register data_ptr **lowest, **highest;
				highest = Elf.sp - 1;
				lowest = highest - args_passed + 1;
				while (lowest <= highest) {
					*lowest = (data_ptr *) (((long) *lowest) | FAULT_BIT);
					lowest++;
				}

				if (Elf.curFrame->objVals && (Elf.curFrame->currentObjVal == 1)) {
					register int    ix;
					if (args_passed > 0) {
						Elf.curFrame->loopArgs = 
							(data_ptr **) malloc(args_passed * sizeof(data_ptr *));
					}
					for (ix = 0; ix < args_passed; ix++)
						Elf.curFrame->loopArgs[ix] = *(Elf.sp - args_passed + ix);
				}
			}

			p = UNFAULT(*pp);
			if (IS_INTEGER(*pp) || (p && p->type != TYPE_OBJECT))
				ElfAbort(ERR_NOTOBJECT);

			cppMethodP = NULL;
			isRemote = FALSE;
			corbaIx = FALSE;
			obj = NULL;
			if (p) {
				if (OX(p)->objHostname) {
					entdata.linkage.addr = NULL;
					isRemote = TRUE;
				}
				else {
					obj = ElfObjPtr(p);
					entdata.type = TYPE_ENTRY;
					entdata.linkage.addr =
						ElfFindMethod(obj, name, hashval, queryDestination,
						call_inherit,&entdata.linkage.packP,&cppMethodP,&corbaIx);
				}
			}
			else
				entdata.linkage.addr = NULL;

			if (entdata.linkage.addr) {
				runMe = (data_ptr) (&entdata);

				if ( *entdata.linkage.addr == TYPE_OBJVAR ) {
					data_ptr first_arg;
					elf_obj_grp		*public_objvarP;

					if (args_passed)
						first_arg =
							*((data_ptr *) ((long) (*(Elf.sp - args_passed)) & ~FAULT_BIT));
					else
						first_arg = NullDataPtr;
					Elf.sp -= args_passed;	/* pop args */
	
					if (is_broadcast) {
						if (!Elf.curFrame->objVals) {
							if (queryDestination) {
								public_objvarP = ElfFindObjLocalData(obj, name, hashval);
								DupAssign(PushPtr(&Elf.curFrame->locvars[queryDestination]),
										  ((public_objvarP) ? public_objvarP->value : NULL));
							}
							else {
								public_objvarP = ElfObjData(obj,name,hashval);
								AxTaskFreeElfData(0,public_objvarP->value);
								public_objvarP->value = AxTaskCopyElfData(0, first_arg);
							}
							break;
						}
						else {
							public_objvarP = ElfObjData(obj,name,hashval);
							AxTaskFreeElfData(0,public_objvarP->value);
							public_objvarP->value = AxTaskCopyElfData(0, first_arg);
	
							if (Elf.curFrame->currentObjVal >= AX(Elf.curFrame->objVals)->size) {
								ELF_DATA_FREE(Elf.curFrame->objVals);
								free(Elf.curFrame->loopArgs);
								Elf.curFrame->objVals = NULL;
								Elf.curFrame->loopArgs = NULL;
								break;      /* finished looping, all done */
							}
							Elf.pc = Elf.curFrame->objLoopPC;       /* send to next in loop */
							pp = &(AX(Elf.curFrame->objVals)->
								   vals[Elf.curFrame->currentObjVal++]);
							goto objloop;
						}
					}
					else {
						if (queryDestination) {
							public_objvarP = ElfFindObjLocalData(obj, name, hashval);
							DupAssign(PushPtr(&Elf.curFrame->locvars[queryDestination]),
									  ((public_objvarP) ? public_objvarP->value : NULL));
						}
						else {
							public_objvarP = ElfObjData(obj,name,hashval);
							AxTaskFreeElfData(0,public_objvarP->value);
							public_objvarP->value = AxTaskCopyElfData(0, first_arg);
						}
						break;
					}
				}
			}
#if !FOR_TESTER
			else if (cppMethodP) {
				remoteDatum = ElfCallCppMember(obj,name,hashval,args_passed);
				if (queryDestination) {
					DupAssign(PushPtr(&Elf.curFrame->locvars[queryDestination]),remoteDatum);
				}
				ELF_DATA_FREE(remoteDatum);
				if (is_broadcast && Elf.curFrame->objVals) {
					if (Elf.curFrame->currentObjVal >= AX(Elf.curFrame->objVals)->size) {
						ELF_DATA_FREE(Elf.curFrame->objVals);
						free(Elf.curFrame->loopArgs);
						Elf.curFrame->objVals = NULL;
						Elf.curFrame->loopArgs = NULL;
						break;      /* finished looping, all done */
					}
					Elf.pc = Elf.curFrame->objLoopPC;       /* send to next in loop */
					pp = &(AX(Elf.curFrame->objVals)->
						   vals[Elf.curFrame->currentObjVal++]);
					goto objloop;
				}
				break;
			}
			
			else if (corbaIx) {
				remoteDatum = ElfCallCorbaMethod(obj,name,args_passed);
				if (queryDestination) {
					DupAssign(PushPtr(&Elf.curFrame->locvars[queryDestination]),
                              remoteDatum);
				}
				ELF_DATA_FREE(remoteDatum);
				if (is_broadcast && Elf.curFrame->objVals) {
                    if (Elf.curFrame->currentObjVal >= AX(Elf.curFrame->objVals)->size) {
                        ELF_DATA_FREE(Elf.curFrame->objVals);
                        free(Elf.curFrame->loopArgs);
                        Elf.curFrame->objVals = NULL;
                        Elf.curFrame->loopArgs = NULL;
                        break;      /* finished looping, all done */
                    }
                    Elf.pc = Elf.curFrame->objLoopPC;   /* send to next in loop */
                    pp = &(AX(Elf.curFrame->objVals)->
                           vals[Elf.curFrame->currentObjVal++]);
                    goto objloop;
                }
                break;
            }

#endif
			else if (isRemote) {
				remoteDatum = elfCallRemoteObjMethod(p, name, args_passed,
												 queryDestination, is_broadcast);
				if (queryDestination) {
					DupAssign(PushPtr(&Elf.curFrame->locvars[queryDestination]),
							  remoteDatum);
				}
				ELF_DATA_FREE(remoteDatum);
				if (is_broadcast && Elf.curFrame->objVals) {
					if (Elf.curFrame->currentObjVal >= AX(Elf.curFrame->objVals)->size) {
						ELF_DATA_FREE(Elf.curFrame->objVals);
						free(Elf.curFrame->loopArgs);
						Elf.curFrame->objVals = NULL;
						Elf.curFrame->loopArgs = NULL;
						break;      /* finished looping, all done */
					}
					Elf.pc = Elf.curFrame->objLoopPC;   /* send to next in loop */
					pp = &(AX(Elf.curFrame->objVals)->
						   vals[Elf.curFrame->currentObjVal++]);
					goto objloop;
				}
				break;
			}
			else if (is_broadcast) {
				Elf.sp -= args_passed;
				if (!Elf.curFrame->objVals) {
					if (queryDestination)
						Assign(PushPtr(Elf.curFrame->locvars + queryDestination), NULL);
					break;      /* not looping, all done */
				}
				else {
					if (Elf.curFrame->currentObjVal >= AX(Elf.curFrame->objVals)->size) {
						ELF_DATA_FREE(Elf.curFrame->objVals);
						free(Elf.curFrame->loopArgs);
						Elf.curFrame->objVals = NULL;
						Elf.curFrame->loopArgs = NULL;
						break;  /* finished looping, all done */
					}
					Elf.pc = Elf.curFrame->objLoopPC;   /* send to next in loop */
					pp = &(AX(Elf.curFrame->objVals)->
						   vals[Elf.curFrame->currentObjVal++]);
					goto objloop;
				}
			}
			else
				ElfObjAbort(ERR_METHNOTINOBJECT, NULL, name, obj);

			/* push the object as first arg */

			PushPtr(pp);

			/* call the method */

			args_passed++;      /* add in the objptr we just pushed */
			if (queryDestination)
				Elf.pc--;
			CallEntrySub(EX(runMe), args_passed, queryDestination, obj);
		}
		goto next_step;
	case OP_SETOBJ_ARG:
		pp = Elf.curFrame->argvars[word2];
		pp = (data_ptr *) ((long) pp & ~FAULT_BIT);
		goto setobj;
	case OP_SETOBJ_GLO:
		pp = &Elf.taskFrame[GLOVARDEX];
		goto setobj;
	case OP_SETOBJ_EXT:
		pp = Elf.taskFrame + EXTDEX;
		goto setobj;
	case OP_SETOBJ_OBJ:
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		pp = &ElfObjData(Elf.curFrame->inObj, name, hashval)->value;
		goto setobj;
	case OP_SETOBJ_SYS:
		name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
		hashval = *Elf.pc++;
		pp = ElfGetSysVarPP(name, hashval);
		goto setobj;

	case OP_SRCMAP:
		Elf.curFrame->curline = word2;
		if (ElfNumDebuggers != 0)
			elfDbgCheckForBreak(Elf.curFrame->inObj);
		goto do_quick;          /* src line markers not counted in run limit */

	case OP_SRCTIMECHECK:
#if !FOR_TESTER
		if ( TaskInfo[AxInTask].alarm_has_fired &&
			 (TaskInfo[AxInTask].alarm_mode == ALARM_MODE_RECYCLE) ) {
			AxSleepSeconds(0);
			AxSetTaskAlarm(AxInTask,ALARM_MODE_RECYCLE,
						TaskInfo[AxInTask].alarm_millisec);
		}
#endif
		Elf.curFrame->curline = word2;
		if (ElfNumDebuggers != 0)
			elfDbgCheckForBreak(Elf.curFrame->inObj);
		goto do_quick;

	case OP_TIMECHECK:
#if !FOR_TESTER
		Elf.pc--;               /* fixup because word2 was pre-fetched */
		if ( TaskInfo[AxInTask].alarm_has_fired &&
			 (TaskInfo[AxInTask].alarm_mode == ALARM_MODE_RECYCLE) ) {
			AxSleepSeconds(0);
			AxSetTaskAlarm(AxInTask,ALARM_MODE_RECYCLE,
						TaskInfo[AxInTask].alarm_millisec);
		}
#endif
		goto do_quick;

	default:
		ElfAbort(ERR_WENTWEST);
	}

	goto next_step;
}

/***************************************************************************
function: ElfSameData - compare two ELF datums, return TRUE if equal
*/
int
ElfSameData (p1, p2)
data_ptr p1, p2;
{
	if (p1 == p2)
		return(TRUE);

/* attempt to compare two small integers */

	if (IS_INTEGER(p1)) {
		if (IS_INTEGER(p2))
			return(FALSE);
		else {
			if (!p2)
				return(GET_INTEGER(p1) == 0);
			else if (p2->type == TYPE_STRING)
				goto dstr_eq;
			else if (p2->type == TYPE_NUMBER)
				goto dtry_floats;
			else
				return(FALSE);
		}
	}
	if (IS_INTEGER(p2)) {
		if (IS_INTEGER(p1)) {
			return(FALSE);
		}
		else {
			if (!p1)
				return(GET_INTEGER(p2) == 0);
			else if (p1->type == TYPE_STRING)
				goto dstr_eq;
			else if (p1->type == TYPE_NUMBER)
				goto dtry_floats;
			else
				return(FALSE);
		}
	}
/* if both are not null, if types dont match, they can't be same */

	if (p1 && p2 && p1->type != p2->type)
		return(FALSE);

/* not small integers, try as two strings if either are strings */

	if ((p1 && (p1->type == TYPE_STRING)) ||
		(p2 && (p2->type == TYPE_STRING))) {
  dstr_eq:
		return(streq(StrFromDataPtr(p1), StrFromDataPtr(p2)));
	}
/* not strings, try as arrays if either is an array */

	if ((p1 && (p1->type == TYPE_ARRAY)) ||
		(p2 && (p2->type == TYPE_ARRAY))) {
		return(ElfSameArray(p1, p2));
	}
/* not arrays, try as binary data */

	if ((p1 && (p1->type == TYPE_BINARY)) ||
		(p2 && (p2->type == TYPE_BINARY))) {
		return(ElfSameBinary(p1, p2));
	}
/* not binaries, try as objects */

	if ((p1 && (p1->type == TYPE_OBJECT)) ||
		(p2 && (p2->type == TYPE_OBJECT))) {
		return(ElfSameObject(p1, p2));
	}
/* not objects, assume floating point numbers */

	if ((p1 && (p1->type == TYPE_NUMBER)) ||
		(p2 && (p2->type == TYPE_NUMBER))) {
  dtry_floats:
		return(NumFromDataPtr(p1) == NumFromDataPtr(p2));
	}
	return(FALSE);
}


/***************************************************************************
function: ElfSameObject - Return whether two objects are equal. This is only used by
                          OP_EQUAL & OP_NEQ instructions.
*/
static int
ElfSameObject (o1, o2)
data_ptr o1, o2;
{
	if ( !o1 && !o2 )		/* two nulls are equal */
		return(TRUE);
	if ( !o1 || !o2 )		/* one of them is null */
		return(FALSE);
	if ( o1->type != o2->type || o1->type != TYPE_OBJECT )
		return(FALSE);

	if ( OX(o1)->objHostname || OX(o2)->objHostname ) {	/* remote objects */
		if ( (OX(o1)->objDex == OX(o2)->objDex) && (OX(o1)->objUid == OX(o2)->objUid)
			&& (OX(o1)->objHostname == OX(o2)->objHostname) )
			return(TRUE);
		else
			return(FALSE);
	}
	else {
		if ( ElfObjPtr(o1) == ElfObjPtr(o2) )
			return(TRUE);
		else
			return(FALSE);
	}
}

/*************************************************************************
function: ElfSameBinary - Return whether two binary objects are equal
*/
ElfSameBinary(a1, a2)
	data_ptr        a1, a2;
{
	register char  *p1, *p2;
	register int    size;

	size = ElfbBinarySize(a1);  /* throws error on failure */
	if (size != ElfbBinarySize(a2))
		return (FALSE);
	if (!size)
		return(TRUE);
	p1 = ((data_binary *) a1)->bytes;
	p2 = ((data_binary *) a2)->bytes;
	while (--size) {
		if (*p1++ != *p2++)
			return (FALSE);
	}
	return (TRUE);
}

/*************************************************************************
function: ElfSameArray - Return whether arrays are equal
This is only used by the OP_EQUAL & OP_NEQ instructions.
Note that arrays are regarded the same based on equivalence of
internal elements.  Thus, an array of nulls is equal to an array of zeros
and likewise is equivalent to an array of empty strings.  This seems to
make the most sense.
*/
ElfSameArray(a1, a2)
	register data_ptr a1, a2;
{
	register data_ptr p1, p2;
	register int    ix, size;
	int             size1, size2;
	char           *StrFromDataPtr();

	size1 = ElfbArraySize(a1);  /* throws error on failure */
	size2 = ElfbArraySize(a2);  /* throws error on failure */
	size = MAX(size1, size2);
	if ((size > 0) && (!a1 || !a2))
		return (FALSE);

	for (ix = 0; ix < size; ix++) {
		p1 = ReadArray(AX(a1), ix);
		p2 = ReadArray(AX(a2), ix);

		if (p1 == p2)
			continue;

/* if either is null and the other is 0 or "", okay */

		if (p1 == NULL) {
			if (p2 == NULL)     /* ok, null == null */
				continue;
			else if (IS_INTEGER(p2)) {
				if (GET_INTEGER(p2) == 0)       /* ok, null == 0 */
					continue;
				else            /* bad, null != 5 */
					return (FALSE);
			} else if ((SX(p2)->type == TYPE_STRING) &&
					   (SX(p2)->text[0] == 0))
				continue;       /* ok, null == "" */
			else
				return (FALSE);
		} else if (p2 == NULL) {
			if (p1 == NULL)     /* ok, null == null */
				continue;
			else if (IS_INTEGER(p1)) {
				if (GET_INTEGER(p1) == 0)       /* ok, null == 0 */
					continue;
				else            /* bad, null != 5 */
					return (FALSE);
			} else if ((SX(p1)->type == TYPE_STRING) &&
					   (SX(p1)->text[0] == 0))
				continue;       /* ok, null == "" */
			else
				return (FALSE);
		}
/* Try integer comparason */

		if (IS_INTEGER(p1)) {
			if (IS_INTEGER(p2))
				return (FALSE); /* p1 == p2 test would have hit */
			else if ((p2->type == TYPE_NUMBER)
					 && (NumFromDataPtr(p1) == NumFromDataPtr(p2)))
				continue;
			else
				return (FALSE);
		} else if (IS_INTEGER(p2)) {
			if (IS_INTEGER(p1))
				return (FALSE); /* p1 == p2 test would have hit */
			else if ((p1->type == TYPE_NUMBER)
					 && (NumFromDataPtr(p1) == NumFromDataPtr(p2)))
				continue;
			else
				return (FALSE);
		}
/* If they are not of same type, must be bad */

		if (XX(p1)->type != XX(p2)->type)
			return (FALSE);

		switch (XX(p1)->type) {
		case TYPE_STRING:
			if (streq(StrFromDataPtr(p2), SX(p1)->text))
				continue;
			else
				return (FALSE);
		case TYPE_ARRAY:
			if (ElfSameArray(p1, p2))
				continue;
			else
				return (FALSE);
		case TYPE_BINARY:
			if (ElfSameBinary(p1, p2))
				continue;
			else
				return (FALSE);

		case TYPE_OBJECT:
			if (ElfSameObject(p1, p2))
				continue;
			else
				return (FALSE);

		case TYPE_NUMBER:
			if (NX(p1)->value == NX(p2)->value)
				continue;
			else
				return (FALSE);
		default:
			return (FALSE);     /* ??1 fails on entry var */
		}

	}
	return (TRUE);
}

ElfDumpArray(a1, level)         /* call initially with level == 0 */
	register data_ptr a1;
	int             level;
{
	register data_ptr p1;
	register int    ix, size;
	char           *StrFromDataPtr();
	char            tabString[100], lofFname[256];
	static FILE    *fp;
	static int      timesCalled = 0;

	if (a1 == NULL)
		return;

	for (ix = 0; ix < level; ix++)
		tabString[ix] = '\t';
	tabString[level] = 0;

	size = ElfbArraySize(a1);   /* throws error on failure */

	if (level == 0) {
		sprintf(lofFname, "elflog.%d", timesCalled++);
		fp = fopen(lofFname, "w+");
	}
	for (ix = 0; ix < size; ix++) {

		p1 = ReadArray(AX(a1), ix);

		if (p1 == NULL) {
			fprintf(fp, "%sNULL\n", tabString);
		}
/* Try integer comparason */

		else if (IS_INTEGER(p1)) {
			fprintf(fp, "%sINTEGER %d\n", tabString, (int) GET_INTEGER(p1));
		} else
			switch (XX(p1)->type) {
			case TYPE_STRING:
				fprintf(fp, "%sSTRING %s\n", tabString, SX(p1)->text);
				break;
			case TYPE_ARRAY:
				fprintf(fp, "%sARRAY\n", tabString);
				ElfDumpArray(p1, level + 1);
				break;
			case TYPE_BINARY:
				fprintf(fp, "%sBINARY\n", tabString);
				break;
			case TYPE_OBJECT:
				fprintf(fp, "%sOBJECT\n", tabString);
				break;
			case TYPE_NUMBER:
				fprintf(fp, "%sNUMBER %f\n", tabString, NX(p1)->value);
				break;
			default:
				fprintf(fp, "%sUNKNOWN\n", tabString);
			}

	}

	fprintf(fp, "\n\n");

	if (level == 0) {
		fclose(fp);
		fp = NULL;
	}
}

/*************************************************************************
function: AxTaskFreeElfData - Free data from a specific task
*/
void 
AxTaskFreeElfData(task, aP)
	register int    task;
	register data_ptr aP;
{
	register data_ptr *pp, *endp;

	if (!aP || (LX(aP) & FAULT_BIT))
		return;
	switch (aP->type) {
	case TYPE_ARRAY:
		pp = AX(aP)->vals;
		endp = pp + AX(aP)->size;
		while (pp < endp) {
			if (*pp && !IS_INTEGER(*pp))
				AxTaskFreeElfData(task, *pp);
			pp++;
		}
		AxTaskPoolFree(task, (char *) aP,
				  sizeof(data_array) + (AX(aP)->size - 1) * sizeof(data_ptr));
		break;
	case TYPE_OBJECT:
		ElfObjDecRef(OX(aP));
		AxTaskPoolFree(task, (char *) aP, sizeof(data_object));
		break;
	case TYPE_NUMBER:
		AxTaskPoolFree(task, (char *) aP, sizeof(data_number));
		break;
	case TYPE_STRING:
		AxTaskPoolFree(task, (char *) aP,
					   sizeof(data_string) + strlen(SX(aP)->text));
		break;
	default:
		TaskFree(task, (char *) aP);
	}
}

/*************************************************************************
function: AxTaskCopyElfData - Copy data for a specific task
*/
data_ptr 
AxTaskCopyElfData(task, aP)
	register int    task;
	register data_ptr aP;
{
	register data_ptr *px, *pends, *pdst;
	register data_ptr new;
	register int size;
    mbchar buff[200];

	if (!aP || IS_INTEGER(aP))
		return (aP);
	aP = (data_ptr) ((long) aP & ~FAULT_BIT);

	if (aP->type == TYPE_NUMBER) {
		new = (data_ptr) AxTaskPoolAllocate(task, sizeof(data_number));
		*NX(new) = *NX(aP);
		return (new);
	}
	if (aP->type == TYPE_STRING)
		return (AxTaskCreateElfString(task, SX(aP)->text));

	if (aP->type == TYPE_ARRAY) {
		new = AxTaskCreateElfArray(task, AX(aP)->size);
		pends = (px = AX(aP)->vals) + AX(aP)->size;
		pdst = AX(new)->vals;
		while (px < pends) {    /* fastest possible array copy? */
			aP = *px++;
			if (IS_INTEGER(aP) || !aP)
				*pdst++ = aP;
			else
				*pdst++ = AxTaskCopyElfData(task, aP);
		}
		return (new);
	}
	if (aP->type == TYPE_OBJECT) {
		if (OX(aP)->objHostname)
			return(AxTaskCreateRemoteElfObject(task,OX(aP)->objDex,OX(aP)->objUid, OX(aP)->objHostname));
		else
			return(AxTaskCreateElfObject(task,OX(aP)->objDex,OX(aP)->objUid,OX(aP)->dynBinder));
	}
	if (aP->type == TYPE_BINARY)
		return(AxTaskCreateElfBinaryWithData(task, BX(aP)->size, BX(aP)->bytes));

	if (aP->type == TYPE_ENTRY)
		return ((data_ptr) TaskCopy(task, aP));

    sprintf(buff, "invalid type %d passed to AxTaskCopyElfData", aP->type );
    tempAbort( buff );

}

/*************************************************************************
function: Assign - Free old val, splice in new val
Called only when the new value is not already owned by a variable/register.
In other words, the data was CREATED just for this assignment.
*/
Assign(dst, src)
	register data_ptr *dst, src;
{
	register data_ptr p;

	p = *dst;
	*dst = src;
	if (p && !IS_INTEGER(p))
		ELF_DATA_FREE(p);
}
TaskAssign(taskId, dst, src)
	register int      taskId;   /* was missing type */
	register data_ptr *dst, src;
{
	register data_ptr p;

	p = *dst;
	*dst = src;
	if (p && !IS_INTEGER(p))
		AxTaskFreeElfData(taskId, p);
}

/*************************************************************************
function: DupAssign - Free old val, copy new val
*/
DupAssign(dst, src)
	register data_ptr *dst, src;
{
	register data_ptr p;

	p = *dst;
	*dst = (IS_INTEGER(src) || !XX(src)) ? src : ElfDataCopy(src);
	if (p && !IS_INTEGER(p))
		ELF_DATA_FREE(p);
}

/*************************************************************************
function: PushTempNum - Assign temp var to dbl & push val
*/
PushTempNum(locid, val)
	int             locid;
	double          val;
{
	VERIFY(fabs(val) < MAX_ELFNUM, ERR_ARITHMETIC);
	Assign(PushPtr(&Elf.curFrame->locvars[locid]), ELF_CREATE_NUMBER(val));
	DUMPTOS();
}

/*************************************************************************
function: PushTempInt - Assign temp var to int & push val
*/
PushTempInt(locid, val)
	int             locid;
	int             val;
{
	if ((val <= 32767) && (val > -32768))
		Assign(PushPtr(&Elf.curFrame->locvars[locid]), MK_INTEGER(val));
	else
		Assign(PushPtr(&Elf.curFrame->locvars[locid]), ELF_CREATE_NUMBER((double) val));
	DUMPTOS();
}

/*************************************************************************
function: PushEntryBlock - Assign temp var to entry block and push val
*/
static 
PushEntryBlock(locid, name)
	int             locid;
	data_ptr       *name;
{
	data_ptr        dP;

	dP = *name;
	dP = (data_ptr) ((long) dP & ~FAULT_BIT);
	Assign(PushPtr(&Elf.curFrame->locvars[locid]), ELF_CREATE_ENTRY(SX(dP)->text));
	DUMPTOS();
}

/*************************************************************************
function: StrEqualfTester - Pop two strings, return result of comparison
                            testing only for byte-by-byte equality which is
                            not language sensitive.
*/
StrEqualTester()
{
	register int    retval;
	register char  *p1, *p2;

	p1 = PopStr();
	p2 = PopStr();

	retval = strcmp(p1, p2);
	return (retval);
}


/*************************************************************************
function: StrDiffTester - Pop two strings, return result of comparison
                          using the per-language collating sequence
*/
#if !FOR_TESTER
    extern int AxLanguage;          /* set in context.c by getenv("AXLANG") */
#endif

StrDiffTester()
{
	register int    retval;
	register char  *p1, *p2;

	p1 = PopStr();
	p2 = PopStr();

#if FOR_TESTER   /*  {  */
    /* stand-alone compiler always runs in "English" */
	retval = strcmp(p1, p2);
#else            /*  } { */
    /* running in "axmain"  */
    retval = Istrcmp( (unsigned char*)p1, (unsigned char*)p2, AxLanguage, TRUE );

#if BUILD_DEBUG
    {
    int retval2;
    retval2 = strcmp(p1, p2);
    if ( ((retval > 0) && (retval2 <= 0)) ||
         ((retval == 0) && (retval2 != 0)) ||
         ((retval < 0)  && (retval2 >= 0)) )
       retval2++;  /* results are different!! thisis trouble if English - jon */
    }
#endif
#endif          /* }  */
	return (retval);
}

/*************************************************************************
function: PopNumVal - Get a double based on Elf.sp
*/
double 
PopNumVal()
{
	register data_ptr *pp;

	pp = POP_PTR;
	if (IS_INTEGER(*pp))
		return (GET_INTEGER(*pp));
	return (NumFromDataPtr(*pp));
}

/*************************************************************************
function: PopDatum - Pop data object
*/
data_ptr 
PopDatum()
{
	register data_ptr *pp;

	pp = POP_PTR;
	if (IS_INTEGER(*pp))
		return (*pp);
	else
		return ((data_ptr) (LX(*pp) & ~FAULT_BIT));
}

/*************************************************************************
function: IntFromDataPtr - Given a DataPtr perform convs needed to ret bint
*/
IntFromDataPtr(dP)
	register data_ptr dP;
{
	double          dval;

	if (IS_INTEGER(dP))
		return (GET_INTEGER(dP));
	dP = (data_ptr) ((long) dP & ~FAULT_BIT);   /* allow pops of read-only
												 * values */
	if (!XX(dP))
		return (0);
	dval = NumFromDataPtr(dP);
	if (dval >= 0.0)
		dval += 0.5;
	else
		dval -= 0.5;
	return (dval);
}

/*************************************************************************
function: NumFromDataPtr - Given a DataPtr perform convs needed to ret double
*/
double 
NumFromDataPtr(dP)
	register data_ptr dP;
{
	double          retval;
	double          Astrtod();
	register char  *p;
	char           *dst;

	if (IS_INTEGER(dP))
		return (GET_INTEGER(dP));
	dP = (data_ptr) ((long) dP & ~FAULT_BIT);   /* allow pops of read-only
												 * values */
	if (!XX(dP))
		return (0.0);
	if (XX(dP)->type == TYPE_NUMBER)
		return (NX(dP)->value);
	if (XX(dP)->type == TYPE_STRING) {
		p = SX(dP)->text;       /* convert spaces to 0.0 */
		if (!(*p))
			return (0.0);
		retval = Astrtod(p, &dst);
		while ((*dst != 0) && isspace(*dst))
			dst++;
		if (*dst != 0)
			ElfStrAbort(ERR_NOTNUMBER, NULL, SX(dP)->text);
		return (retval);
	}
	ElfAbort(ERR_NOTNUMBER);
}

/*************************************************************************
function: RePopIntVal - Fix SP from POP_INT's tentative pop and pop int
This is ONLY called by POP_INT
*/
static int 
RePopIntVal()
{
	register data_ptr p;
	double          dval;

	Elf.sp++;
	p = POP_VAL;
	if (IS_INTEGER(p))
		return (GET_INTEGER(p));
	if (!p)
		return (0);
	dval = NumFromDataPtr(p);
	if (dval >= 0.0)
		dval += 0.5;
	else
		dval -= 0.5;
	return (dval);
}

/*************************************************************************
function: PopIntVal - Get a 32 bit long based on Elf.sp
*/
static int 
PopIntVal()
{
	register data_ptr p;
	double          dval;

	p = POP_VAL;
	if (IS_INTEGER(p))
		return (GET_INTEGER(p));
	if (!p)
		return (0);
	dval = NumFromDataPtr(p);
	if (dval >= 0.0)
		dval += 0.5;
	else
		dval -= 0.5;
	return (dval);
}

/*************************************************************************
function: PopElfbArgVal - Pop value from call stack
This is only used by functions called in '*' mode.
If a function is called in '*' mode, it must pop ALL args
based on the passed arg_count.
*/
data_ptr 
PopElfbArgVal()
{
	register data_ptr *pp;

	pp = POP_PTR;
	if (IS_INTEGER(*pp))
		return (*pp);
	else
		return ((data_ptr) (LX(*pp) & ~FAULT_BIT));
}

/*************************************************************************
function: PopStr - Get a string datum based on Elf.sp, create if appropriate
*/
char           *
PopStr()
{
	char           *StrFromDataPtr();
	return (StrFromDataPtr(POP_VAL));
}

/*************************************************************************
function: StrFromDataPtr - Get a string from a data ptr - do conversions
*/
char           *
StrFromDataPtr(dP)
	register data_ptr dP;
{
	char           *buf;

	if (!XX(dP))
		return ("");
	if (IS_INTEGER(dP)) {
		buf = ElfInstMalloc(50);
		sprintf(buf, "%d", GET_INTEGER(dP)); /* long int into %d?... */
		WORK(WORK_CONVERT);
		return (buf);
	}
	dP = (data_ptr) ((long) dP & ~FAULT_BIT);   /* allow pops of read-only
												 * values */
	if (XX(dP)->type == TYPE_STRING)
		return (SX(dP)->text);
	if (XX(dP)->type == TYPE_NUMBER) {
		double          val;
		int             valint;
		buf = ElfInstMalloc(50);
		val = NX(dP)->value;
		if ( (val > MAXINT) || (val <= -MAXINT) )	/* direct assign crashes HP */
			valint = 0;
		else
			valint = val;		/* would crash on HP if val == -8.0e+306 */
		if (valint == val)
			sprintf(buf, "%d", valint);
		else {
			sprintf(buf, "%.15g", val);
			ConvertUSDotsToEuropeComma(buf);
		}
		WORK(WORK_CONVERT);
		return (buf);
	}
	ElfAbort(ERR_NOTSTRING);
}

#if DBX_MODE
/*************************************************************************
function: PushPtr - Push a datum's addr onto stack, checking for overflow
*/
data_ptr       *
PushPtr(p)
	register data_ptr *p;
{
	register data_ptr dp;

	if (!(((long) p) & FAULT_BIT))
		VerifyNotFreed(*p);
	return (*Elf.sp++ = p);
}

VerifyNotFreed(dp)
	register data_ptr dp;
{
	if (XX(p) && !(LX(dp) & FAULT_BIT)) {
		if (*(((long *) XX(p)) - 3) == -1)
			tempAbort(NXLT("memory management error - data re-used"));
	}
}
#endif

/*************************************************************************
function: AxTaskCreateElfEntry - Create an entry block
Threads via AxInTask but allocates based on passed task!
*/
data_ptr 
AxTaskCreateElfEntry(task, name)
	register int    task;
	register char  *name;
{
	data_entry     *p;
	int             hashval;

	p = (data_entry *) TaskAlloc(task, sizeof(data_entry));
	name = elfHashNameAndVal(name, &hashval);
	p->type = TYPE_ENTRY;
	p->linkage.nameP = name;
	p->linkage.moduleName = NULL;       /* !!2 correct? */
	p->linkage.hashval = hashval;
	p->linkage.addr = NULL;     /* will get snapped in as needed */
	return ((data_ptr) p);
}

/*************************************************************************
function: AxTaskCreateElfNumber - Allocate a number object
*/
data_ptr 
AxTaskCreateElfNumber(task, val)
	register int    task;
	double          val;
{
	register data_ptr p;
	register int    num;

	if (C_DBL_IS_ELF_INT(val, num)) {
		p = (data_ptr) MK_INTEGER(num);
		return (p);
	}
	p = (data_ptr) AxTaskPoolAllocate(task, sizeof(data_number));
	NX(p)->type = TYPE_NUMBER;
	NX(p)->value = val;
	return (p);
}

/*************************************************************************
function: AxTaskCreateElfString - Allocate a string object
*/
data_ptr 
AxTaskCreateElfString(task, str)
	register int    task;
	register char  *str;        /* null terminated intial string */
{
	register data_string *p;
	register unsigned int size;
	register int    len;

	if (str)
		len = strlen(str);
	else
		len = 0;
	p = (data_string *) AxTaskPoolAllocate(task, size = sizeof(data_string) + len);
	p->type = TYPE_STRING;
	if (str)
		strcpy(p->text, str);
	else
		strcpy(p->text, NXLT(""));
	return ((data_ptr) p);
}

/*************************************************************************
function: AxTaskCreateSizedElfString - Allocate a string object, len specified
*/
data_ptr 
AxTaskCreateSizedElfString(task, len, str)
	register int    task;
	register int    len;
	register char  *str;        /* null terminated intial string */
{
	register data_string *p;
	register unsigned int size;

	if (len == -1) {
		if (str)
			len = strlen(str);
		else
			len = 0;
	}
	p = (data_string *) AxTaskPoolAllocate(task, size = sizeof(data_string) + len);
	p->type = TYPE_STRING;
	if (str)
		strcpy(p->text, str);
	else
		strcpy(p->text, NXLT(""));
	return ((data_ptr) p);
}

/*************************************************************************
function: AxTaskCreateElfArray - Create array with initial size
*/
data_ptr 
AxTaskCreateElfArray(task, elements)
	register int    task;
	register int    elements;
{
	register data_ptr p;
	register unsigned int size;

	size = sizeof(data_array) + (sizeof(data_ptr) * (elements - 1));
	p = (data_ptr) AxTaskPoolAllocate(task, size);
	MEMSET(AX(p), 0, size);
	AX(p)->type = TYPE_ARRAY;
	AX(p)->size = elements;
	return (p);
}

/*************************************************************************
function: AxTaskCreateElfBinary - Create binary object with initial size
*/
data_ptr 
AxTaskCreateElfBinary(task, elements)
	register int    task;
	register int    elements;
{
	register data_binary *p;
	register unsigned int size;

	size = sizeof(data_binary) + elements - 1;
	p = (data_binary *) TaskAlloc(task, size);
	MEMSET(p, 0, size);
	p->type = TYPE_BINARY;
	p->size = elements;
	return ((data_ptr) p);
}

/*************************************************************************
function: AxTaskCreateElfBinaryWithData - Create binary object with initial data
*/
data_ptr 
AxTaskCreateElfBinaryWithData(task, elements, data)
	register int    task;
	register int    elements;
	register char  *data;
{
	register data_binary *p;
	register unsigned int size;

	size = sizeof(data_binary) + elements - 1;
	p = (data_binary *) TaskAlloc(task, size);
	p->type = TYPE_BINARY;
	p->size = elements;
	memcpy(p->bytes,data,elements);
	return ((data_ptr) p);
}

/*************************************************************************
function: AxTaskCreateElfObject - Create a reference to an object
*/
data_ptr 
AxTaskCreateElfObject(task, dex, uid, binderList)
	register int    task;
	register int    dex, uid;
	register data_ptr binderList;
{
#if !FOR_TESTER
	register data_object *dP;

	dP = (data_object *) AxTaskPoolAllocate(task, sizeof(data_object));
	MEMSET(dP,0,sizeof(data_object));
	dP->type = TYPE_OBJECT;
	dP->objDex = dex;
	dP->objUid = uid;
	dP->objHostname = NULL;         /* reference to local object */
	dP->dynBinder = AxTaskCopyElfData(task, binderList);
	if (uid == 0)
		ElfAddUnsnappedObjRefToList(dP);
	return ((data_ptr) dP);
#endif
}

/***************************************************************************
function: AxTaskCreateRemoteElfObject - create reference to a remote object
*/
data_ptr 
AxTaskCreateRemoteElfObject(task, dex, uid, hostname)
	register int    task;
	register int    dex, uid;
	register char   *hostname;
{
#if !FOR_TESTER
	register data_object *dP;

	dP = (data_object *) AxTaskPoolAllocate(task, sizeof(data_object));
	MEMSET(dP, 0, sizeof(data_object));
	dP->type = TYPE_OBJECT;
	dP->objDex = dex;
	dP->objUid = uid;
	dP->objHostname = elfHashName(hostname);   /* reference to remote object */
	return ((data_ptr) dP);
#endif
}

/*************************************************************************
function: WriteArray - Replace element, optionally increase size of array
*/
data_ptr 
WriteArray(p, dex, new_val)
	register data_ptr p;
	register int    dex;
	register data_ptr new_val;
{
	register int    old_bytes, new_bytes;

	old_bytes = sizeof(data_array) + (AX(p)->size - 1) * sizeof(data_ptr);
	new_bytes = sizeof(data_array) + dex * sizeof(data_ptr);
	if (new_bytes > old_bytes) {
		p = (data_ptr) AxTaskPoolRealloc(THIMpid,
							 (char *) XX(p), (unsigned) new_bytes, old_bytes);
		MEMSET(((char *) XX(p)) + old_bytes, 0, new_bytes - old_bytes);
		AX(p)->size = dex + 1;
	}
	ELF_DATA_FREE(AX(p)->vals[dex]);
	AX(p)->vals[dex] = new_val;
	return (p);
}
data_ptr 
TaskWriteArray(taskId, p, dex, new_val)
	register int      taskId;   /* was missing type => int*/
	register data_ptr p;
	register int    dex;
	register data_ptr new_val;
{
	register int    old_bytes, new_bytes;

	old_bytes = sizeof(data_array) + (AX(p)->size - 1) * sizeof(data_ptr);
	new_bytes = sizeof(data_array) + dex * sizeof(data_ptr);
	if (new_bytes > old_bytes) {
		p = (data_ptr) AxTaskPoolRealloc(taskId,
							 (char *) XX(p), (unsigned) new_bytes, old_bytes);
		MEMSET(((char *) XX(p)) + old_bytes, 0, new_bytes - old_bytes);
		AX(p)->size = dex + 1;
	}
	AxTaskFreeElfData(taskId, AX(p)->vals[dex]);
	AX(p)->vals[dex] = new_val;
	return (p);
}

/*************************************************************************
function: TruncArray - Decrease size of an array
*/
data_ptr 
TruncArray(p, newsize)
	register data_ptr p;
	register int    newsize;
{
	register int    ix, new_bytes, old_bytes;

	VERIFY(newsize >= 0, ERR_ILLVAL);
	if (AX(p)->size <= newsize)
		return (p);
	for (ix = newsize; ix < AX(p)->size; ix++)
		ELF_DATA_FREE(AX(p)->vals[ix]);
	new_bytes = sizeof(data_array) + (newsize - 1) * sizeof(data_ptr);
	old_bytes = sizeof(data_array) + (AX(p)->size - 1) * sizeof(data_ptr);
	p = (data_ptr) AxTaskPoolRealloc(THIMpid,
							 (char *) XX(p), (unsigned) new_bytes, old_bytes);
	AX(p)->size = newsize;
	return (p);
}

/*************************************************************************
function: TaskTruncArray - Decrease size of an array
*/
data_ptr 
TaskTruncArray(taskId, p, newsize)
	register int      taskId;
	register data_ptr p;
	register int      newsize;
{
	register int    ix, new_bytes, old_bytes;

	VERIFY(newsize >= 0, ERR_ILLVAL);
	if (AX(p)->size <= newsize)
		return (p);
	for (ix = newsize; ix < AX(p)->size; ix++)
		ELF_DATA_FREE(AX(p)->vals[ix]);
	new_bytes = sizeof(data_array) + (newsize - 1) * sizeof(data_ptr);
	old_bytes = sizeof(data_array) + (AX(p)->size - 1) * sizeof(data_ptr);
	p = (data_ptr) AxTaskPoolRealloc(taskId,
							 (char *) XX(p), (unsigned) new_bytes, old_bytes);
	AX(p)->size = newsize;
	return (p);
}

/*************************************************************************
function: ReadArray - Get an element from an array (or null if bounds error)
*/
data_ptr 
ReadArray(p, dex)
	register data_array *p;
	register int    dex;
{
	VERIFY(p && (p->type == TYPE_ARRAY), ERR_NOTARRAY);
	VERIFY(dex >= 0, ERR_BADARRDEX);
	return (dex >= p->size ? NULL : p->vals[dex]);
}

/*************************************************************************
function: FreeLocals - Free local variables & args
*/
FreeLocals(fp)
	register stack_frame *fp;
{
	register int    ix;

	if (!fp->locvars)
		return;                 /* happens on malloc failure in OP_ENTER */
	for (ix = MAX_REGS - fp->numregs; ix < fp->numlocs; ix++)
		ELF_DATA_FREE(fp->locvars[ix]);
	ElfFree(fp->locvars);       /* memory leak fixed 4/1/90 */
	fp->locvars = 0;
}

/*************************************************************************
function: NewStackFrame - Create a new stack fram
*/
static stack_frame *
NewStackFrame(args_passed, retloc, functype, inObj, packP)
	register int    args_passed;/* how many args are pushed on stack */
	register int    retloc;     /* tempvar id to put return value or 0 for no
								 * ret */
    register shrt   functype;   /* was missing type => int */
	register elf_object *inObj; /* NULL or object ptr */
	register package_info *packP;
{
	register stack_frame *fp;

	VERIFY(Elf.recursion < MAX_ELF_RECURSION, ERR_RECURS);
	Elf.recursion++;
	fp = (stack_frame *) ElfMalloc(sizeof(stack_frame) + 25);   /* Purify 12.26 */
	MEMSET(fp, 0, sizeof(stack_frame));
	fp->caller = Elf.curFrame;
	fp->return_addr = Elf.pc + (args_passed != 0) + (retloc != 0);
	fp->num_args_passed = args_passed;
	fp->retloc = retloc;
	fp->functype = functype;
	fp->funcstarts = NULL;
	fp->inObj = inObj;
	fp->objVals = NULL;
	fp->loopArgs = NULL;
	fp->packP = packP;
	Elf.curFrame = fp;
}

/*************************************************************************
function: SaveCurMacroName - Save the name of the Elf macro being called. 
*/
static void
SaveCurMacroName(name)
    char* name;
{
	strncpy(CurExtMacroName,name,ELFIDSTR);
	CurExtMacroName[ELFIDSTR] = '\0';
}

/*************************************************************************
function: ElfGetCurMacroName - Returns most recently called external 
Elf (non builtin) macro name. Temporary func for SS use only.
BUG: This function can only return the current External Macro Name. It is a temporary
solution to give the macro name calling ss_call$ to SS. 
Permanent solution will be to maintain the macro name/index in the StackFrame.
Todo that currently the information is not readily available for Elf routines
called from the same package (I mean without doing a expensive search).
*/
char*
ElfGetCurMacroName()
{
	return(CurExtMacroName);
}


/*************************************************************************
function: CallSub - Call an Elf routine in the same package
*/
static 
CallSub(cf_offset, args_passed, retloc, inObj)
	int             cf_offset;  /* to function to call */
	int             args_passed;/* how many args are pushed on stack */
	int             retloc;     /* tempvar id to put return value or 0 for no
								 * ret */
	elf_object     *inObj;      /* NULL or object ptr */
{
	NewStackFrame(args_passed, retloc, 0, inObj, Elf.curFrame->packP);
	Elf.pc = Elf.curFrame->packP->code_starts + (cf_offset << 2)
		+ 1;                    /* +1 gets past type word */
	Elf.curFrame->pc = Elf.pc;
}

/*************************************************************************
function: CallExtSub - Call an Elf/C routine
If this calls a C function, returns TRUE.
*/
static int
CallExtSub(eP, args_passed, retloc, inObj, retptr)
	ext_link       *eP;
	int             args_passed;/* how many args are pushed on stack */
	int             retloc;     /* tempvar id to put return value or 0 for no
								 * ret */
	elf_object     *inObj;
	data_ptr       *retptr;     /* where to write back Cfunc retval if retloc
								 * -1 */
{
	register int    type;

	Snapper(0, eP, inObj, TRUE);
	type = *(eP->value.ptr);
	switch (type) {
	case TYPE_QFUNC:
	case TYPE_MACRO:
	case TYPE_UIFUNC:
	case TYPE_NRFUNC:
	case TYPE_SETOBJ:
	case TYPE_GETOBJ:
		if (retloc == -1)       /* data will return back normally in the ElfRun */
			retloc = 0;
		NewStackFrame(args_passed, retloc, (shrt)type, inObj, eP->packP);
		SaveCurMacroName(eP->nameP);
		Elf.pc = eP->value.ptr + 1;
		Elf.curFrame->pc = Elf.pc;
		return (FALSE);

	case TYPE_C_FUNC:
	case TYPE_C_MACRO:
		if (retloc == -1)
			*retptr = CallElfb((bltin_desc *)eP->value.ptr, args_passed, retloc);
		else
			CallElfb((bltin_desc *)eP->value.ptr, args_passed, retloc);
		return (TRUE);

	case TYPE_U_FUNC:
	case TYPE_R_FUNC:
		if ( retloc == -1 )
			*retptr = CallUserElfb((data_user_func *)eP->value.ptr,args_passed,retloc);
		else
			CallUserElfb((data_user_func *)eP->value.ptr,args_passed,retloc);
		return(TRUE);

	default:
		ElfAbort(ERR_NOTFUNC);
	}
        /* NOTREACHED */
	return(FALSE);
}

/*************************************************************************
function: ElfFindLinkForByNameCalls - Find/create dynamic link
*/
ext_link *ElfFindLinkForByNameCalls (passed_name)
char *passed_name;
{
	register int    ix;
	ext_link		*linkP;
	register char  *p;
	int hashval;
	char            name[STRLONG + 1];

/* convert name to upper case */

	VERIFY(strlen(passed_name) < STRLONG, ERR_STRTOOLONG);
	FullTrim(name, passed_name);
	VERIFY(strlen(name) <= ELFIDSTR, ERR_STRTOOLONG);
	for (p = name; *p; p++)
		*p = toupper(*p);

/* Keeping a true hash table would be much better! ??2 */

	hashval = hash(name);
	for ( ix = 0; ix < SizeLinkByNameArray; ix++ ) {
		if ( (LinkByNameArray[ix].hashval == hashval)
				&& streq(LinkByNameArray[ix].nameP,name) )
			break;		/* found, use it! */
	}
	if ( ix == SizeLinkByNameArray ) {		/* create link */
		if ( 0 == ix % 100 ) {
			LinkByNameArray = (ext_link*)TaskRealloc(0,LinkByNameArray,
						(ix+100) * sizeof(ext_link));
		}
		LinkByNameArray[ix].nameP = strcpy(TaskAlloc(0,strlen(name)+1),name);
		LinkByNameArray[ix].moduleName = NULL;
		LinkByNameArray[ix].hashval = hashval;
		LinkByNameArray[ix].value.ptr = NULL;	/* snap this! */
		SizeLinkByNameArray++;
	}
	linkP = &LinkByNameArray[ix];
	return(linkP);
}

/*************************************************************************
function: CallEntrySub - Call an Elf/Alis routine via entry struct
If the passed_eP is a string object, a dynamic link is fabricated.
*/
static 
CallEntrySub(passed_eP, args_passed, retloc, inObj)
	data_entry     *passed_eP;
	int             args_passed;/* how many args are pushed on stack */
	int             retloc;     /* tempvar id to put return value or 0 for no
								 * ret */
	elf_object     *inObj;      /* obejfct whose method is being executed */
{
	data_entry     *eP = passed_eP;
	ext_link       *linkP = NULL;
	char            nameBuf[ELFIDSTR + 1];
	package_info   *packP;

	register shrt  *addr;       /* addr of function to run */
	char           *p;

	Elf.pc -= 1;                /* fixup cause of how NewStackFrame compute ret
								 * addr */
	VERIFY(eP && !IS_INTEGER(eP), ERR_NOTFUNC);
	eP = (data_entry *) UNFAULT(eP);
	if (eP->type == TYPE_STRING) {      /* True call by name */
		p = ((data_string *) (eP))->text;
		VERIFY(0 != strlen(p), ERR_ILLVAL);
		strcpy(nameBuf, p);
		for (p = nameBuf; *p; p++)
			*p = toupper(*p);
		linkP = ElfFindLinkForByNameCalls(nameBuf);
		Snapper(0, linkP, inObj, TRUE);
		addr = linkP->value.ptr;
		packP = linkP->packP;
	} else {
		VERIFY(eP->type == TYPE_ENTRY, ERR_NOTFUNC);
		Snapper(0, (ext_link*)&(eP->linkage), inObj, TRUE);
		addr = eP->linkage.addr;
		packP = eP->linkage.packP;
	}

	switch (*addr) {
	case TYPE_QFUNC:
	case TYPE_MACRO:
	case TYPE_UIFUNC:
	case TYPE_NRFUNC:
	case TYPE_SETOBJ:
	case TYPE_GETOBJ:
		NewStackFrame(args_passed, retloc, *addr, inObj, packP);
		Elf.pc = addr + 1;
		Elf.curFrame->pc = Elf.pc;
		break;
	case TYPE_C_FUNC:
	case TYPE_C_MACRO:
		CallElfb((bltin_desc *)addr, args_passed, retloc);
		break;
	case TYPE_U_FUNC:
	case TYPE_R_FUNC:
        CallUserElfb((data_user_func *)addr,args_passed,retloc);
		break;
	default:
		ElfAbort(ERR_NOTFUNC);
	}
}

typedef union risc_args {
	int             i;
	double          n;
	char           *s;
	data_ptr        v;
	long            l;
}               risc_args;

/*************************************************************************
function: ElfbFuncInstalled - Determine whether passed name is installed
*/
void 
ElfbFuncInstalled(nameP)
	data_ptr        nameP;
{
	char           *p = StrFromDataPtr(nameP);
	data_entry      linker;
	int             hashval;
	char            nameBuf[ELFIDSTR + 1];

	VERIFY(0 != strlen(p), ERR_ILLVAL);
	VERIFY(strlen(p) <= ELFIDSTR, ERR_ILLVAL);
	strcpy(nameBuf, p);
	for (p = nameBuf; *p; p++)
		*p = toupper(*p);
	linker.linkage.nameP = elfHashNameAndVal(nameBuf, &hashval);
	linker.linkage.moduleName = NULL;
	linker.linkage.hashval = hashval;
	linker.linkage.addr = NULL;
	Snapper(0, (ext_link *)&(linker.linkage), FALSE, FALSE);
	if (linker.linkage.addr)
		ElfRetData = MK_INTEGER(-1);
	else
		ElfRetData = MK_INTEGER(0);
}

/*************************************************************************
function: CallElfb - Call an Elf routine
If retloc is specified as -1, the return value from the C func is returned
by this func rather than pushed onto the stack.  This form of call is only
used by a C program which wants to call an Elf bltin... like the SS recalc.
*/
data_ptr CallElfb(lp, args_passed, retloc)
register bltin_desc *lp;
register int    args_passed;/* how many args are pushed on stack */
register int    retloc;     /* tempvar id to put return value or 0 for no ret */
{
	register data_ptr retval = NULL;
	register int    ix;
	stack_frame frm;
	register data_ptr aP;
	register data_ptr (*func) ();

	if ( ElfProfiling ) {
		frm.caller = Elf.curFrame;
		frm.packP = NULL;
		frm.funcstarts = (shrt*)lp;
		ElfProfileNote(&frm);
	}

	if (ElfInLearnMode && (lp->type == TYPE_C_MACRO)) {
		stack_frame    *fp;
		char            name[ELFIDSTR + 1];
		data_ptr        *args;

		for (fp = Elf.curFrame; fp; fp = fp->caller) {
			if (fp->functype == TYPE_QFUNC)
				goto no_record;
			if (fp->recorded)
				goto no_record;
		}

		args = (data_ptr*)calloc(args_passed,sizeof(data_ptr));
		for (ix = 0; ix < args_passed; ix++)
			args[ix] = POP_VAL;
		FetchNameOfFunction(name, (shrt*)lp);
		ElfRecord(ELFREC_FCALL, name, args_passed, args, 0, NULL, NULL, NULL);
		Elf.sp += args_passed;
		free(args);
	}
no_record:

	VERIFY((lp->type == TYPE_C_FUNC) || (lp->type == TYPE_C_MACRO), ERR_NOTFUNC);
	VERIFY(lp->func, ERR_NOTFUNC);
	if (retloc == -1)
		Elf.pc += (args_passed != 0);
	else
		Elf.pc += (args_passed != 0) + (retloc != 0);

	ElfRetData = 0;	/* sanity for old-style calls */

	func = lp->func;
	aP = MkArray(args_passed);
	for ( ix = 0; ix < args_passed; ix++ )
		AX(aP)->vals[ix] = PopElfbArgVal();
	retval = (*func)(aP);
	AxTaskFreeArrayButNotContents(THIMpid,aP);

	if ( ElfProfiling ) {
		frm.caller = Elf.curFrame;
		frm.packP = NULL;
		frm.funcstarts = (shrt*)lp;
		ElfProfileExit(&frm);
	}

	if (!retloc) {              /* just exec function & return */
		ELF_DATA_FREE(retval);
		return(NULL);           /* was return; */
	}

	if (retloc == -1)
		return (retval);
	else {
		Assign(PushPtr(&Elf.curFrame->locvars[retloc]), retval);
		return (NULL);
	}
}

/*************************************************************************
function: ELthimble - Start a new task from C code
*/
ELthimble(argc, argv, elfArgs, elfArgv)
	int             argc;
	char          **argv;
	int             elfArgs;    /* if TRUE, argv is really data_ptr** */
	data_ptr        elfArgv;
{
	data_ptr        aP;
	int             ix;
	int             retCode;
	data_ptr        retData;
	int             size;
	data_ptr argsP;

	if (!ElfInit())
		taskAbort(NXLT("Could not initialize new Elf task."));

	Elf.pc = 0;
	if (elfArgs) {
		size = ElfbArraySize(elfArgv);
		retCode = elfDoCallByName(argv[0], &retData, size, elfArgv);
	} else if (argc > 1) {
		aP = ELF_CREATE_ARRAY(argc - 1);
		for (ix = 0; ix < argc - 1; ix++)
			ElfAddStrToArray(aP, ix, argv[ix + 1]);
		argsP = ELF_CREATE_ARRAY(1);
		WriteArray(argsP, 0, aP);
		retCode = elfDoCallByName(argv[0], &retData, 1, argsP);
	} else {
		retCode = elfDoCallByName(argv[0], &retData, 0, NULL);
	}
	AxExitWithData(retCode, retData);
}

/*************************************************************************
function: ElfCallFromC - A wrapper to allow C apps & widget layer to call Elf code
This allows for the fact the elfDoCallByName calls CallExtSub which calls
NewStackFrame which bumps the returned-to PC by one if there are passed args.
In a C-to-Elf routine, the PC should be left alone.
*/
ElfCallFromC(name, retP, args)
	register char  *name;
	register data_ptr *retP;
	register data_ptr args;
{
	shrt           *savedPC;
	int             retval;
	jmp_buf         savedAborter;
	int argcnt = ElfbArraySize(args);

	savedPC = Elf.pc;
	memcpy(savedAborter, Elf.abortBuf, sizeof(jmp_buf));
	retval = elfDoCallByName(name, retP, argcnt, args);
	memcpy(Elf.abortBuf, savedAborter, sizeof(jmp_buf));
	Elf.pc = savedPC;
	return (retval);
}

/*************************************************************************
function: elfDoCallByName - Perform the call by name
*/
int 
elfDoCallByName(name, retP, argcnt, passed_args)
	register char  *name;		/* Allowed to be mixed case */
	register data_ptr *retP;
	register int    argcnt;
	register data_ptr passed_args;
{
	register data_ptr *pp;
	register int    ix;
	ext_link		*linkP;
	register char  *p;
	int retCode;
	int             calledC;
	elf_object     *inObj;
	register data_ptr *args = NULL; /* need copy in heap for debugger */ /* clobbered by longjmp */
	data_ptr **savedSP;

	if (!Elf.inited)
		ElfInit();

/* set up the top level error catcher */

	savedSP = Elf.sp;
    if ( setjmp(Elf.abortBuf) ) {
		Elf.sp = savedSP;
		*retP = ELF_CREATE_ARRAY(4);
		ElfAddStrToArray(*retP, 0, Elf.ErrorStr);
		ElfAddStrToArray(*retP, 1, Elf.ErrorObject);
		ElfAddNumToArray(*retP, 2, Elf.ErrorLine);
		ElfAddStrToArray(*retP, 3, Elf.ErrorFile);
		if ( args )
			free(args);
		return (Elf.ErrorCode);
	}

/* Find/create dynamic link */

	linkP = ElfFindLinkForByNameCalls(name);

/* push args and call it! */

	if (argcnt > 0)
		args = (data_ptr *) malloc(argcnt * sizeof(data_ptr *));

	for (ix = argcnt - 1; ix >= 0; ix--) {
		args[ix] = ReadArray(AX(passed_args), ix);
		pp = &args[ix];
		pp = (data_ptr *) ((long) pp | FAULT_BIT);
		PushPtr(pp);
	}

	if (Elf.curFrame) {
		Elf.curFrame->stop_here = TRUE;
		inObj = Elf.curFrame->inObj;
	} else
		inObj = NULL;

	if (retP)
		calledC = CallExtSub(linkP, argcnt, -1, inObj, retP);
	else
		calledC = CallExtSub(linkP, argcnt, 0, inObj, retP);

	if (calledC) {
		ElfIntrJanitor();
		retCode = 0;
	} else
		*retP = ElfRun(&retCode);

	if (args)
		free(args);

	if (Elf.curFrame)
		Elf.curFrame->stop_here = FALSE;
	Elf.sp = savedSP;
	return (retCode);
}

/*************************************************************************
function: FullTrim - Trim leading and trailing spaces
*/
FullTrim(dst, src)
	register char  *dst, *src;
{
#define TAB 9

	/* trim leading spaces */

	while ((*src == ' ') || (*src == TAB))
		src++;
#if AT == 0                     /* test no good on AT's */
	if (src != dst)
#endif
		strcpy(dst, src);

	/* trim trailing spaces */

	for (src = dst + strlen(dst) - 1; src > dst; src--)
		if ((*src != ' ') && (*src != TAB))
			break;
	*(src + 1) = 0;
}

/*************************************************************************
function: DoReturn - Return from an Elf/Alis routine
If there is return data, it is COPIED from a register to another register.
Then the source register is freed. One would think it could be simply transferred,
saving many cpu cycles.  Naturally, we could only do that if the source
register was a local register rather than an argument.  One day...
*/
static data_ptr 
DoReturn(retP)
	data_ptr        retP;       /* return value or NULL */
{
	register stack_frame *fp;

retry:
	fp = Elf.curFrame;
#if !FOR_TESTER
	if ( ElfProfiling )
		ElfProfileExit(fp);
#endif
	if (fp->objVals) {          /* only the case during error crawl-back */
		ELF_DATA_FREE(fp->objVals);
		free(fp->loopArgs);
	}
	Elf.pc = fp->return_addr;

	elfDbgSetNexterFrame(fp, fp->caller);

	Elf.curFrame = fp->caller;
	if (fp->retloc) {
		DupAssign(PushPtr(&Elf.curFrame->locvars[fp->retloc]), retP);
		DUMPTOS();
	}
	if (!Elf.curFrame || Elf.curFrame->stop_here)
		retP = ElfDataCopy(retP);
	else if (Elf.curFrame->objVals) {   /* done with objects loop? */
		if (Elf.curFrame->currentObjVal >= AX(Elf.curFrame->objVals)->size) {
			ELF_DATA_FREE(Elf.curFrame->objVals);
			Elf.curFrame->objVals = NULL;
			free(Elf.curFrame->loopArgs);
			Elf.curFrame->loopArgs = NULL;
		}
	}
	FreeLocals(fp);
	ElfFree(fp);
	Elf.recursion--;
	if (!Elf.curFrame || Elf.curFrame->stop_here) {
		return (retP);          /* pass value back to caller of ElfRun */
	} else if (!Elf.pc)         /* returning thru a call-by-name */
		goto retry;
	return (NULL);
}

/*************************************************************************
function: SnapData - Snap link to external variable, return Elf.taskFrame offset
If link already snapped, just return the snap value
*/
int 
SnapData(lp)
	register ext_link *lp;
{
	register int    val;
	gvar_info      *gv;         /* ignored returned addr of gvar struct */

	if (lp->value.dex)
		return (lp->value.dex);
	val = SnapDataDoIt(lp, &gv);
	if (val == -1)
		ElfStrAbort(ERR_NOTVAR, NULL, lp->nameP);
	return (val);
}

static int 
SnapDataDoIt(lp, gvP)
	register ext_link *lp;
	register gvar_info **gvP;
{
	register char  *name;
	register int    ix, px;
	register gvar_info *sP;
	register package_info *packP;
	int             retval = 0;
	int             hashval;

	if (lp->value.dex)
		return (lp->value.dex);
	name = lp->nameP;
	hashval = hash(name);
	for (px = 0; packP = Packages[px]; px++) {
		if (packP->gvar_list != NULL)
			for (ix = 0; sP = packP->gvar_list[ix]; ix++) {
#if AXMS
				if (lp->moduleName && Astrcmp(lp->moduleName, packP->moduleName))
					continue;
#else
				if (lp->moduleName && !streq(lp->moduleName, packP->moduleName))
					continue;
#endif
				if ((sP->hashval == hashval) && streq(name, sP->name)) {
					if (!retval || (packP->task_uid == TaskInfo[AxInTask].uid)) {
						*gvP = sP;
						retval = sP->value;
					}
					if (packP->task_uid == TaskInfo[AxInTask].uid) {
						*gvP = sP;
						return (lp->value.dex = retval);
					}
				}
			}
	}
	if (retval)
		return (lp->value.dex = retval);
	return (-1);
}

/*************************************************************************
function: DumpElfSymNames - for debug
*/
DumpElfSymNames(arg)
data_ptr arg;
{
	register gfunc_info *sP;
	register gvar_info *vsP;
	register package_info *packP;
	register bltin_desc *bP;
	int             ix, px, an;
	FILE           *fp;
	int             tmp, numArgs;
	char           *p1, *p2, *varNames, *name, *sourceName;

arg = NULL;		/* just until callmode fixed in elstatic.c to allow str! */

	if ( !arg ) {
		fp = fopen(NXLT("/tmp/zznames"), "w");
	}
	else {
		fp = fopen(StrFromDataPtr(arg), "w");
	}

	for (bP = &ElfCoreFuncs[1]; bP->type; bP++) {
		tmp = bP->symid;
		p2 = ElfSymbolNameFromId(tmp, 1);
		fprintf(fp, NXLT("C: %s\n"), p2);
	}

	for (px = 0; packP = Packages[px]; px++) {
		if (packP->code_starts == NULL)
			continue;
		sourceName = strrchr(packP->source_name,'/');
		if ( !sourceName )
			sourceName = packP->source_name;
		else
			sourceName++;
		for (ix = 0; vsP = packP->gvar_list[ix]; ix++)
			fprintf(fp, NXLT("%s data: %s\n"), sourceName, vsP->name);
		for (ix = 0; sP = packP->gfun_list[ix]; ix++) {
			fprintf(fp, "%s: %s(", sourceName, sP->name);
			numArgs = *(sP->value+2);
			varNames = packP->varNames + *(sP->value+5);
			an = 0;
			for ( an = 0; *varNames && (an < numArgs); an++) {
				if ( an != 0 )
					fprintf(fp,",");
				name = varNames;
				varNames += strlen(varNames) + 1;
				if ( *varNames ) {
					if ( *varNames == '[' ) {
						varNames++;
						if ( *varNames == '[' ) {
							varNames++;
							fprintf(fp,"format arrayof2 %s %s",varNames,name);
						}
						else
							fprintf(fp,"format arrayof %s %s",varNames,name);
					}
					else
						fprintf(fp,"format %s %s",varNames,name);
				}
				else
					fprintf(fp,"%s",name);
				varNames += strlen(varNames) + 1;
			}
			fprintf(fp,")\n");
		}
	}
	fclose(fp);
}


/*************************************************************************
function: BinarySearchFunc -- Search for a given function in the list and
                              return its index
                              (Non-recursive binary search)
*/
static int 
BinarySearchFunc( name, fun_list)
register char* name;
gfunc_info** fun_list;
{
    int nfuncs;
	int res, start_ix;
	register int ix;
	for(nfuncs=0; fun_list[nfuncs]; nfuncs++)
		; /* TODO-- save 'nfuncs' in the package_info */ 

	for( start_ix = 0; nfuncs > 1; ) {
		ix = start_ix + nfuncs/2;
		res = strcmp(name,fun_list[ix]->name);
		if(res == 0)
			return(ix);
		if(res < 0) {
			nfuncs = (ix-start_ix);
		}
		else {
			nfuncs -= (ix-start_ix)+1;
			start_ix = ix+1;
		}
	}
	if(nfuncs==1 && strcmp(name,fun_list[start_ix]->name)==0)
		return(start_ix);
	return(-1);
}

/*************************************************************************
function: Snapper - Snap link to external symbol return result addr
If link already snapped, just return the snap addr
??x This seems to be getting called FAR TOO MUCH
If this finds a linkage who package taskuid matches the current task,
it returns immediately... other wise it keeps searching, hoping
for a taskuid match.
*/
static data_ptr *
Snapper(mode, lp, inObj, allowInstall)
	int             mode;       /* ignored */
	register ext_link *lp;
	elf_object     *inObj;
	int             allowInstall;       /* if TRUE, allows autoinstall */
{
	register char  *name;
	register int    ix, px;
	register gfunc_info *sP;
	register package_info *packP;
	int             installed_it = FALSE;
	char           *ElfSymbolNameFromId();
	shrt           *retval = NULL;
	package_info   *retpack = NULL; /* initialized! */
	int             applix_sym;
	data_user_func *shlibP;

	if (lp->value.ptr)
		return ((data_ptr *) (lp->value.ptr));
	name = lp->nameP;

/* Check for shared library reference */

	shlibP = ElfFindUserFunc(name);
	if ( shlibP ) {
		ElfBindUserFunc(shlibP);	/* throws error if there is a problem */
		return((data_ptr*)(lp->value.ptr = &(shlibP->type)));
	}

retry:
	for (px = 0; packP = Packages[px]; px++) {
		if ( !packP->gfun_list )	/* skip placeholders */
			continue;
#if AXMS
		if (lp->moduleName && Astrcmp(lp->moduleName, packP->moduleName))
			continue;
#else
		if (lp->moduleName && !streq(lp->moduleName, packP->moduleName))
			continue;
#endif
		if (retval && packP->task_uid != TaskInfo[AxInTask].uid)
			continue;
		if((ix = BinarySearchFunc(name, packP->gfun_list)) >= 0)
		{
				sP = packP->gfun_list[ix];
				if (packP->code_starts == NULL) {
					/* it's a dummy package. load the shared library that
					   contains the package now */
					if(!AxElfLoadShlibForId(packP->dynamic_load,name))
						return(NULL);
					return(Snapper(mode, lp, inObj, allowInstall));
				}
				else if (!retval 
						 || (packP->task_uid == TaskInfo[AxInTask].uid)) {
					retval = (shrt *) sP->value;
					retpack = packP;
					if (packP->task_uid == TaskInfo[AxInTask].uid) {
						lp->packP = retpack;
						return ((data_ptr *) (lp->value.ptr = retval));
					}
					/* else continue searching, hoping for a taskuid match */
				}
		 }
	}
	if (retval) {               /* handles non-task_uid snaps */
		lp->packP = retpack;
		return ((data_ptr *) (lp->value.ptr = retval));
	}
/* Check for call by name of a bltin */

	applix_sym = ElfLookupSymbol(name, ElfLanguage);
	if (applix_sym && IsBltinSymCode(applix_sym))
		return ((data_ptr *) (lp->value.ptr = &ElfCoreFuncs[applix_sym].type));

/* Try auto-installing */

	if (allowInstall) {
		if (!installed_it && InstallCommandByName(name)) {
			installed_it = TRUE;
			goto retry;
		}
		ElfStrAbort(ERR_FUNCNOEXISTS, NULL, name);
		return (NULL);          /* just for lint... can never get here */
	} else
		return (NULL);
}

/*************************************************************************
function: ElfUnloadTaskPackages - Remove this task's transient packages
*/
void 
ElfUnloadTaskPackages(taskId)
	int             taskId;     /* task's UID */
{
	register package_info *packP;
	register int    px;

start_over:
	for (px = 0; packP = Packages[px]; px++) {
		if (packP->task_uid == taskId) {
			ElfUnloadPackage(packP);
			goto start_over;    /* link list damaged, do again */
		}
	}
}

/*************************************************************************
function: ElfUnloadPackage - Remove a package from memory
*/
void
ElfUnloadPackage(packP)
	register package_info *packP;
{
	register stack_frame *frp;
	register int    jx, ix;
	gvar_info      *sP;
	int             release_count = 0;
	char            vars_to_free[MAX_GVARS];    /* used when replacing existing
												 * package */
	int             old_context = THIMpid;
	int             did_kill = FALSE;
	data_ptr dP;

	if (packP->code_starts == NULL) 
		return;/* dummy package - should never come here */
/* Clear all breakpoints in this package */

	for (ix = ElfNumBreakpoints - 1; ix >= 0; ix--) {
		if ( packP == ElfBreakData[ix].packP )
			ElfClrBreakPoint(ElfBreakData[ix].packP->source_name,
					 ElfBreakData[ix].line, ElfBreakData[ix].obj);
	}
	for (ix = ElfNumDebuggers-1; ix >= 0; --ix) {
		if (!ThimPidFromUID(ElfDebugger[ix].dbg_uid)) { /* debuger crashed */
			elfDbgCloseDebugger(&ElfDebugger[ix],FALSE);
			continue;
		}
		if (!ThimPidFromUID(ElfDebugger[ix].debuggee_uid)) { /* debugee crashed */
			ElfDebugger[ix].is_stopped = FALSE;
			ElfDebugger[ix].is_nexting = FALSE;
			ElfDebugger[ix].is_stepping = FALSE;
			ElfDebugger[ix].breakFrame = NULL;
			ElfDebugger[ix].breakGlobals = NULL;
			ElfDebugger[ix].exbrk.packP = NULL;
		}
		if (ElfDebugger[ix].breakFrame &&
			ElfDebugger[ix].breakFrame->packP == packP)
			ElfDebugger[ix].breakFrame = NULL;
		if (ElfDebugger[ix].exbrk.packP == packP)
			ElfDebugger[ix].exbrk.packP = NULL;
	}

/* Shoot down any tasks using this package */

	for (ix = 0; ix < MAX_PIDS; ix++) {
		if (TaskInfo[ix].in_use && IsAppidElf(TaskInfo[ix].appid)) {
			VfyContext(ix);

			for (frp = Elf.curFrame; frp; frp = frp->caller) {
				if (frp->packP == packP) {

					/* clear debuggers break info */
					for (jx = 0; jx < ElfNumDebuggers; ++jx) {
						if (ElfDebugger[jx].breakGlobals == Elf.taskFrame) {
							ElfDebugger[jx].breakGlobals = NULL;
							ElfDebugger[jx].breakFrame = NULL;
						}
					}

					if (old_context == ix)
						taskAbort(NXLT("Package install overloads current task."));
					else
						KillTask(ix);
					did_kill = TRUE;
				}
			}
		}
	}
	VfyContext(old_context);
	if (did_kill)
		TaskRecycle();          /* terminate kill-tagged tasks NOW */

/* Note (for per-task release) and release global vars from alloc table */

	memset(vars_to_free, 0, sizeof(vars_to_free));
	for (ix = 0; sP = packP->gvar_list[ix]; ix++) {
		vars_to_free[LX(sP->value)] = TRUE;
		GvarFreeList[LX(sP->value)] = GvarFreeHead;
		GvarFreeHead = LX(sP->value);
		release_count++;
	}

/* If have released globals, release all of the globals
		associated with that package in all of the Elf tasks. */

	if (release_count) {
		for (ix = 0; ix < MAX_PIDS; ix++) {
			if (TaskInfo[ix].in_use && IsAppidElf(TaskInfo[ix].appid)) {
				VfyContext(ix);
				for (jx = 0; jx < MAX_GVARS; jx++) {
					if (vars_to_free[jx]) {
						ELF_DATA_FREE(Elf.taskFrame[jx]);
						Elf.taskFrame[jx] = NULL;
					}
				}
			}
		}
		VfyContext(old_context);
	}
/* Remove package reference */

	for (ix = 0, jx = 0; Packages[ix]; ix++) {
		if (Packages[ix] != packP)
			Packages[jx++] = Packages[ix];
	}
	Packages[jx] = NULL;		/* terminate with NULL */
	NextPackageInd = jx;
	
	ElfResetAllDynamicLinks();

	/* Free all memory that was allocated at LoadPackage time */
	
	TaskFree(0, packP->gvar_list[0]);	/* gvarsDataP */
	TaskFree(0, packP->gvar_list);		/* gvarsArrayP */

	TaskFree(0, packP->ovar_list[0]);	/* ovarsDataP */
	TaskFree(0, packP->ovar_list);		/* ovarsArrayP */

	TaskFree(0, packP->svar_list[0]);	/* svarsDataP */
	TaskFree(0, packP->svar_list);		/* svarsArrayP */

	TaskFree(0, packP->gfun_list[0]);	/* gfuncsDataP */
	TaskFree(0, packP->gfun_list);		/* gfuncsArrayP */

	TaskFree(0, packP->ext_list[0]);	/* extsDataP */
	TaskFree(0, packP->ext_list);		/* extsArrayP */

	/* codeP */
	dP = (data_ptr) (((char *) packP->code_starts) - 8);
	AxTaskFreeElfData(0,  dP);

	/* litsP */
	dP = (data_ptr) (((char *)packP->literals_table) - 8);
	for (ix = 0; ix < AX(dP)->size; ++ix)		/* reset fault bit */
		if (!IS_INTEGER(AX(dP)->vals[ix]))
			AX(dP)->vals[ix] = (data_ptr) (((long) AX(dP)->vals[ix]) & ~FAULT_BIT);
	AxTaskFreeElfData(0,  dP);

	/* varnamesP */
	dP = (data_ptr) (((char *)packP->varNames) - 8);
	AxTaskFreeElfData(0,  dP);

	/* formatsP */
	AxTaskFreeElfData(0, packP->formats);

	/* package itself */
	TaskFree(0, packP);
}

/*************************************************************************
function: ElfResetAllDynamicLinks - Both code & data links
*/
ElfResetAllDynamicLinks()
{
	register int    px, ix;
	register package_info *packP;
	register ext_link *eP;

	/* reset all dynamic links in all packages */

	for (px = 0; packP = Packages[px]; px++) {
		if (packP->ext_list != NULL)
			for (ix = 0; eP = packP->ext_list[ix]; ix++)
				eP->value.ptr = NULL;
	}

	/* reset the CallByName array */

	for ( ix = 0; ix < SizeLinkByNameArray; ix++ )
		LinkByNameArray[ix].value.ptr = NULL;
}

/*************************************************************************
function: ElfFindMethod - Return addr of entry block
Note tah dynamically public'ed objvars do not have ?! tags.  Therefore,
the lookup compares to the entire string name if the value_ptr is TYPE_OBJVAR.
Compiler public'ed objvars have ?! tags.
*/
static shrt    *
ElfFindMethod (obj, name, hashval, query, call_inherit, packPP, cppMethodP,corbaIx)
register elf_object *obj;
register char  *name;
register int    hashval;
register int    query;
register int call_inherit;	/* support for obj.^method */
register package_info **packPP;
cpp_method_info_t **cppMethodP;
int *corbaIx;
{
#if !FOR_TESTER
	register int    ix;
	register package_info *packP;
	register gfunc_info *sP, **sPP;
	register char   tag = (query ? '?' : '!');
	elf_obj_grp *ElfFindObjPublicData();
	cpp_method_info_t		*mP;

	while (obj) {
		if ( obj->hashtable ) {
			for ( sPP = obj->hashpool+obj->hashtable[hashval]; *sPP; sPP++ ) {
				if ( (((*sPP)->name[0] == tag) && streq(name, (*sPP)->name+1))
					|| ((*((*sPP)->value) == TYPE_OBJVAR) && streq(name, (*sPP)->name)) ) {
					if ( ((*sPP)->srcLine == -1) && (obj != Elf.curFrame->inObj) )
						continue;	/* private methods encoded with -1 srcLine */
					if ( call_inherit ) {
						if ( (*sPP)->value == Elf.curFrame->funcstarts )
							call_inherit--;	/* get the next, not this one */
					}
					else {
						*packPP = obj->packP;
						return ((*sPP)->value);
					}
				}
			}
		}
		else if (packP = obj->packP) {
			for (ix = 0; sP = packP->gfun_list[ix]; ix++) {
				if ((sP->hashval == hashval)
					&& (sP->name[0] == tag) && streq(name, sP->name + 1)) {
					if ( (sP->srcLine == -1) && (obj != Elf.curFrame->inObj) )
						continue;	/* private methods encoded with -1 srcLine */
					if ( call_inherit ) {
						if ( sP->value == Elf.curFrame->funcstarts )
							call_inherit--;	/* get the next, not this one */
					}
					else {
						*packPP = packP;
						return (sP->value);
					}
				}
			}
			/* "large" nodes flesh out even if they aren't classes */
			if ( ix > 10 )	
				ElfFleshOutMethodHash(obj);
		}

		if (obj->cpp_obj_info && obj->cpp_obj_info->objP
								&& obj->cpp_obj_info->classP) {
			elf_object *cP;

			cP = ElfObjPtr(obj->cpp_obj_info->classP);
			if (cP && cP->cpp_class_info)
				mP = cP->cpp_class_info->methods;
			else
				mP = NULL;
			for (;  mP && mP->methodName; ++mP) {
				if (hashval == mP->hashval && streq(name, mP->methodName)) {
					if ( call_inherit ) {
						call_inherit--;	/* get the next, not this one */
					}
					else {
						*cppMethodP = mP;
						return(NULL);
					}
				}
			}
		}
		
		*corbaIx = 0;
		if (obj->ext_obj) {
			if (elfExtMethodExists(obj->ext_obj,name)) {
				*corbaIx = 1;
				return(NULL);
			}
			obj = obj->inheritFromP;
		}
		obj = obj->inheritFromP;
	}
	return (NULL);
#endif
}

/*************************************************************************
function: ElfbInstallBinaryPatch - Load patch file into memory & Branch to it
*/
void 
ElfbInstallBinaryPatch(bP)
	data_ptr        bP;
{
	static data_ptr codeP;
	shrt           *code_starts;
	register char  *cp;
	register int    ix, code_size;

	if (codeP)                  /* only one patch in memory at a time */
		AxTaskFreeElfData(0, codeP);
	codeP = ReadArray(AX(bP), 5);
	codeP = AxTaskCopyElfData(0, codeP);
	code_starts = (shrt *) (BX(codeP)->bytes);
	code_size = (BX(codeP)->size) / sizeof(shrt);
	cp = (char *) code_starts;
	for (ix = 0; ix < code_size; ix++, cp += 2)
		code_starts[ix] = (*(cp + 1) & 0xFF) | (*cp << 8);
	Elf.pc = code_starts;
}

/*************************************************************************
function: ElfbInstallBinary - Load binary package to memory
*/
void 
ElfbInstallBinary(bP, slaved)
	data_ptr        bP;
	int             slaved;     /* if true, slave to current task */
{
	int             task_uid;

	if (slaved)
		task_uid = TaskInfo[AxInTask].uid;
	else
		task_uid = 0;

	ElfbArraySize(bP);          /* verifies is a program object */
	ElfbArraySize(ReadArray(AX(bP), 0));
	ElfLoadPackage(bP,slaved);
}

/*************************************************************************
function: ElfLoadPackage - Load package array
*/
package_info   *
ElfLoadPackage(aP, slaved)
data_ptr aP;
int slaved;
{
	data_ptr        headerP, gvarsP, ovarsP, svarsP, gfuncsP, afuncP, avarP, litsP,
					extsP, codeP, varnamesP, formatsP;
	package_info   *packP;
	register int    ix, num_gvars, num_ovars, num_svars, num_gfuncs, num_exts, type,
					px, code_size;
	int             hashval,rev_num;
	gvar_info      *gvarsDataP, **gvarsArrayP;
	gvar_info      *ovarsDataP, **ovarsArrayP;
	gvar_info      *svarsDataP, **svarsArrayP;
	gfunc_info     *gfuncsDataP, **gfuncsArrayP;
	ext_link      **extsArrayP, *extsDataP;
	char           *name;
	char            filename[STRPATH + 1];

	headerP = ReadArray(AX(aP), 0);
	gvarsP = ReadArray(AX(aP), 1);
	gfuncsP = ReadArray(AX(aP), 2);
	extsP = ReadArray(AX(aP), 3);
	litsP = ReadArray(AX(aP), 4);
	litsP = AxTaskCopyElfData(0, litsP);
	codeP = ReadArray(AX(aP), 5);
	codeP = AxTaskCopyElfData(0, codeP);
	varnamesP = ReadArray(AX(aP), 6);
	varnamesP = AxTaskCopyElfData(0, varnamesP);
	formatsP = ReadArray(AX(aP), 7);
	formatsP = AxTaskCopyElfData(0, formatsP);
	ovarsP = ReadArray(AX(aP), 8);
	svarsP = ReadArray(AX(aP), 9);

    rev_num = IntFromArray(headerP,0);
	strcpy(filename, StrFromArray(headerP, 1));

/* Check package list to see if already loaded, if so, unload it */

	if (filename[0]) {          /* object methods uses empty filenames */
		for (px = 0; packP = Packages[px]; px++) {
			if (packP->dynamic_load && packP->code_starts) {
				if (streq(packP->source_name, filename)) {
					/* Throw this package away */
					ElfUnloadPackage(packP);
					break;      /* don't keep looping, packP now garbage */
				}
			}
		}
	}
/* reset all dynamic links in all packages */

	ElfResetAllDynamicLinks();

/* code */

	packP = (package_info *) TaskCalloc(0, 1, sizeof(package_info));
	packP->dynamic_load = TRUE;
	packP->code_starts = (shrt *) (BX(codeP)->bytes);
	code_size = (BX(codeP)->size) / sizeof(shrt);
	packP->code_ends = packP->code_starts + code_size;
	packP->task_uid = (slaved) ?  TaskInfo[AxInTask].uid : 0;
	{
		register char  *cp;
		cp = (char *) packP->code_starts;
		for (ix = 0; ix < code_size; ix++) {
			packP->code_starts[ix] = (*(cp + 1) & 0xFF) | (*cp << 8);
			cp += 2;
		}
	}

/* globals */

	num_gvars = ElfbArraySize(gvarsP);
	gvarsDataP = (gvar_info *) TaskCalloc(0, sizeof(gvar_info), num_gvars);
	gvarsArrayP = (gvar_info **) TaskCalloc(0, sizeof(gvar_info *), (num_gvars + 1));
	for (ix = 0; ix < num_gvars; ix++) {
		avarP = ReadArray(AX(gvarsP), ix);
		name = StrFromArray(avarP, 0);
		name = elfHashNameAndVal(name, &hashval);
		gvarsDataP[ix].hashval = hashval;
		gvarsDataP[ix].name = name;
		if (AX(avarP)->size > 1)
			gvarsDataP[ix].formatP = elfHashName(StrFromArray(avarP, 1));
		AllocateGvar(&gvarsDataP[ix]);
		gvarsArrayP[ix] = &gvarsDataP[ix];
	}
	packP->gvar_list = gvarsArrayP;

/* object variables - for the debugger */

	num_ovars = ElfbArraySize(ovarsP);
	ovarsDataP = (gvar_info *) TaskCalloc(0, sizeof(gvar_info), num_ovars);
	ovarsArrayP = (gvar_info **) TaskCalloc(0, sizeof(gvar_info *), (num_ovars + 1));
	for (ix = 0; ix < num_ovars; ix++) {
		avarP = ReadArray(AX(ovarsP), ix);
		name = StrFromArray(avarP, 0);
		name = elfHashNameAndVal(name, &hashval);
		ovarsDataP[ix].hashval = hashval;
		ovarsDataP[ix].name = name;
		if (AX(avarP)->size > 1)
			ovarsDataP[ix].formatP = elfHashName(StrFromArray(avarP, 1));
		ovarsArrayP[ix] = &ovarsDataP[ix];
	}
	packP->ovar_list = ovarsArrayP;

/* sysvar variables - for the debugger */

	num_svars = ElfbArraySize(svarsP);
	svarsDataP = (gvar_info *) TaskCalloc(0, sizeof(gvar_info), num_svars);
	svarsArrayP = (gvar_info **) TaskCalloc(0, sizeof(gvar_info *), (num_svars + 1));
	for (ix = 0; ix < num_svars; ix++) {
		avarP = ReadArray(AX(svarsP), ix);
		name = StrFromArray(avarP, 0);
		name = elfHashNameAndVal(name, &hashval);
		svarsDataP[ix].hashval = hashval;
		svarsDataP[ix].name = name;
		if (AX(avarP)->size > 1)
			svarsDataP[ix].formatP = elfHashName(StrFromArray(avarP, 1));
		svarsArrayP[ix] = &svarsDataP[ix];
	}
	packP->svar_list = svarsArrayP;

/* func entries for others to link to */

	if(rev_num < 301) { /* gfunc list is not sorted.
						  sort it now, as Snapper() expects them sorted */
		ElfSortGfuncs(gfuncsP);
	}
	num_gfuncs = ElfbArraySize(gfuncsP);
	gfuncsDataP = (gfunc_info *) TaskCalloc(0, sizeof(gfunc_info), num_gfuncs);
	gfuncsArrayP = (gfunc_info **) TaskCalloc(0, sizeof(gfunc_info *), (num_gfuncs + 1));
	for (ix = 0; ix < num_gfuncs; ix++) {
		int loc_info;
		afuncP = ReadArray(AX(gfuncsP), ix);
		name = StrFromArray(afuncP, 0);
		name = elfHashNameAndVal(name, &hashval);
		type = IntFromArray(afuncP, 1);
		loc_info = IntFromArray(afuncP, 3);
		gfuncsDataP[ix].value = packP->code_starts + loc_info;	/* presumption */
		if ((type == TYPE_SETOBJ) || (type == TYPE_GETOBJ)) {
			static shrt objvar_val = TYPE_OBJVAR;
			static shrt *objvar_ptr = &objvar_val;
			gfuncsDataP[ix].hashval = hash(name + 1);
			if ( loc_info == 0 )		/* an OBJVAR hiding as a method */
				gfuncsDataP[ix].value = objvar_ptr;
		}
		else
			gfuncsDataP[ix].hashval = hashval;
		gfuncsDataP[ix].name = name;
		gfuncsDataP[ix].srcLine = IntFromArray(afuncP, 2);
		gfuncsArrayP[ix] = &gfuncsDataP[ix];
	}
	packP->gfun_list = gfuncsArrayP;

/* external variable/function linkage to snap thru  */

	num_exts = ElfbArraySize(extsP);
	extsDataP = (ext_link *) TaskCalloc(0, sizeof(ext_link), num_exts);
	extsArrayP = (ext_link **) TaskCalloc(0, sizeof(ext_link *), (num_exts + 1));
	for (ix = 0; ix < num_exts; ix++) {
		afuncP = ReadArray(AX(extsP), ix);
		name = StrFromArray(afuncP, 0);
		extsDataP[ix].nameP = elfHashNameAndVal(name, &hashval);
		extsDataP[ix].hashval = hashval;
		if (AX(afuncP)->size > 2)
			extsDataP[ix].moduleName = elfHashName(StrFromArray(afuncP, 2));
		extsArrayP[ix] = &extsDataP[ix];
	}
	packP->ext_list = extsArrayP;

/* literals... not tightly packed for now */

	packP->literals_table = AX(litsP)->vals;
	for (ix = 0; ix < AX(litsP)->size; ix++)    /* set fault bit in ptrs */
		AX(litsP)->vals[ix] = (data_ptr) (((long) AX(litsP)->vals[ix]) | FAULT_BIT);

/* variable names for debugger & recording */

	packP->varNames = BX(varnamesP)->bytes;
	packP->formats = formatsP;
	packP->source_name = elfHashName(filename);
	packP->src_map = IntFromArray(headerP, 2);

/* compute module name from file name... clip off trailing dot & leading path */

	name = strrchr(filename, '.');
	if (name)
		*name = 0;
	name = strrchr(filename, '/');
	if (!name)
		name = strrchr(filename, '\\');
	if (name)
		name = name + 1;
	else
		name = filename;
	packP->moduleName = elfHashName(name);

/* Add to Packages array... at head so new funcs found first */

	if(NextPackageInd+1 >= MaxPackages-1) {
		MaxPackages = NextPackageInd+1+20;
		Packages = (package_info **) TaskRealloc(0, Packages, 
												 sizeof(package_info *) * MaxPackages);
	}
	ix = NextPackageInd;
	while (ix >= 0) {
		Packages[ix + 1] = Packages[ix];
		ix--;
	}
	Packages[0] = packP;
	Packages[++NextPackageInd] = NULL;
	return (packP);
}


/* 
 *  This function will load ONE dummy package for each shared lib. Once a 
 *  macro call is identified to exist  in this dummy package, it will be 
 *  unloaded/deleted  and the real packages for that dummy libpackage are 
 *  loaded in its position in the Packages[]- done by ElfInstallModule()
 */

void ElfLoadDummyPackagesForShlib(lib_id,filename) 
int lib_id;
char* filename; /* exports file name for the library */
{
	int ix,jx,npackages;
	package_info* packP;
	data_ptr gfuncsP, exports_listP;
	gfunc_info     *gfuncsDataP, **gfuncsArrayP;
	char* name;
	int hashval,num_gfuncs,type=0;

	if ( -1 == access(filename,0) )
		return;
	ElfbReadDataFile(filename);
	exports_listP = ElfRetData;
	ElfRetData = NULL;
	if (exports_listP == NULL)
		return;
	npackages = ElfbArraySize(exports_listP);
	MaxPackages = NextPackageInd+npackages+20;
	Packages = (package_info **) TaskRealloc(0, Packages, MaxPackages * sizeof(package_info *));
	for (ix=0; ix < npackages; ix++) {
		gfuncsP = ReadArray(AX(exports_listP),ix);
		packP = (package_info *) TaskCalloc(0, 1, sizeof(package_info));
		packP->dynamic_load = lib_id; 
		packP->code_starts = (shrt *) (NULL);
		packP->code_ends = NULL;
		packP->task_uid = 0;
		packP->ext_list = NULL;
		packP->literals_table = NULL;
		packP->varNames = NULL;
		packP->formats = NULL;
		packP->src_map = FALSE;
		packP->moduleName = elfHashName(StrFromArray(gfuncsP,0));
		packP->source_name = packP->moduleName;
		num_gfuncs = ElfbArraySize(gfuncsP)-1;
		gfuncsDataP = (gfunc_info *) TaskCalloc(0, sizeof(gfunc_info), num_gfuncs);
		gfuncsArrayP = (gfunc_info **) TaskCalloc(0, sizeof(gfunc_info *), (num_gfuncs + 1));
		for (jx = 0; jx < num_gfuncs; jx++) {
			int loc_info;
			name = StrFromArray(gfuncsP, jx+1);
			name = elfHashNameAndVal(name, &hashval);
			gfuncsDataP[jx].value = NULL;
			if ((type == TYPE_SETOBJ) || (type == TYPE_GETOBJ)) {
				static shrt objvar_val = TYPE_OBJVAR;
				static shrt *objvar_ptr = &objvar_val;
				gfuncsDataP[jx].hashval = hash(name + 1);
			}
			else
				gfuncsDataP[jx].hashval = hashval;
			gfuncsDataP[jx].name = name;
			gfuncsDataP[jx].srcLine = 0;
			gfuncsArrayP[jx] = &gfuncsDataP[jx];
		}
		packP->gfun_list = gfuncsArrayP;
		Packages[NextPackageInd++] = packP;
	}
	Packages[NextPackageInd] = NULL;
	ELF_DATA_FREE(exports_listP);
}


/* This function will be called only when the packages for libId, have
   been identified. This will then reduce the number of dummy packages.
   */
void ElfUnloadDummyPackagesForShlib(libId)
{
	int ix, jx;
	package_info* p;
	for(ix=0, jx=0; p = Packages[ix]; ix++) {
		if (p->dynamic_load == libId) {
			TaskFree(0, p->gfun_list[0]);
			TaskFree(0, p->gfun_list);
			TaskFree(0, p);
			Packages[jx] = NULL;
		}
		else {
			if (ix != jx) {
				Packages[ix] = NULL;
				Packages[jx] = p;
			}
			jx++;
		}
	}
	NextPackageInd = jx;
	Packages[NextPackageInd] = NULL;

	NDummyPackages = AxGetNumDummyPackages();
}

/*************************************************************************
function: ElfbListPackages - Return array of loaded packages info
*/
ElfbListPackages()
{
	register data_ptr aP;
	register package_info *p;
	register int    ix = 0, px;

	aP = ELF_CREATE_ARRAY(0);
	for (px = 0; p = Packages[px]; px++)
		if ( p->source_name )
			aP = ElfAddStrToArray(aP, ix++, p->source_name);
	ElfRetData = aP;            /* Alis func generic return value */
}

static int 
CmpFuncs(fn1, fn2)
    data_ptr *fn1, *fn2;
{
    register data_ptr f1, f2;
	register char *s1, *s2;
	f1 = *fn1;
	f2 = *fn2;
	if(!XX(f1))
		s1 = "";
	else
		s1 = SX(AX(f1)->vals[0])->text;
	if(!XX(fn2))
		s2 = "";
	else
		s2 = SX(AX(f2)->vals[0])->text;
	return(strcmp(s1,s2));
}	

void
ElfSortGfuncs(gfuncsP)
register data_ptr gfuncsP;
{
	qsort(AX(gfuncsP)->vals,ElfbArraySize(gfuncsP),
		  sizeof(char*),(QSORT_FUNCPTR)CmpFuncs);
}

/*************************************************************************
function: AllocateGvar - Get storage for a global var
*/
static 
AllocateGvar(gP)
	register gvar_info *gP;
{
	register int    pos;
#define ERR_TOO_MANY_GVARS 999  /* ??x */
	gP->value = GvarFreeHead;
	VERIFY(pos = LX(gP->value), ERR_TOO_MANY_GVARS);
	GvarFreeHead = GvarFreeList[pos];
}

/*************************************************************************
function: FindPackage - Return a package ptr based on PC
*/
package_info   *
FindPackage(pc)
	register shrt  *pc;
{
	register package_info *p;
	register int    px;

	for (px = 0; p = Packages[px]; px++)
		if ((pc >= p->code_starts) && (pc < p->code_ends))
			break;
	return (p);
}

/*************************************************************************
function: FindPackageByName - Return a package ptr based on source_name
*/
package_info   *
FindPackageByName(name)
	register char  *name;
{
	register package_info *p;
	register int    px;

	for (px = 0; p = Packages[px]; px++)
		if (p->source_name && streq(p->source_name, name))
			break;
	return (p);
}

/*************************************************************************
function: FindPackageByModuleName - Return a package ptr based on module name
*/
package_info   *
FindPackageByModuleName(name)
	register char  *name;
{
	register package_info *p;
	register int    px;

	for (px = 0; p = Packages[px]; px++)
        {
		if ( !p->moduleName )
			continue;
#if AXMS
        /* On Windows, all filenames are upper case */
        if ( Astrcmp(p->moduleName, name) == 0 )
            break;
#else
		if (streq(p->moduleName, name))
			break;
#endif
        }
	return (p);
}

/*************************************************************************
function: ElfWhere - Print to stdout ELF's execution point
This is only called by DBX users to figure out what macroi we crashed in.
*/
ElfWhere()
{
	int             line;
	package_info   *packP;
	char           *name;
	stack_frame    *fp;
	shrt           *pc;
	char			funcName[ELFIDSTR+1];

	pc = Elf.pc;
	for (fp = Elf.curFrame; fp && !fp->stop_here; fp = fp->caller) {
		line = fp->curline;
		packP = fp->packP;
		name = packP->source_name;
    	if ( fp->funcstarts )
        	FetchNameOfFunction(funcName, fp->funcstarts);
    	else
			strcpy(funcName,"???");
		printf(NXLT("line %d in function %s in file %s\n"), line, funcName, name);
		fflush(stdout);
		pc = fp->return_addr;
	}
	return (0);
}

/*************************************************************************
function: AxfPrintDatum - Write datum to stdout
*/
data_ptr AxfPrintDatum (argDatum)
data_ptr argDatum;
{
#if !FOR_TESTER
    data_ptr aP = AnyFromArray(argDatum,0);
	AxPrintfDatum(aP);
#endif
	return(NULL);
}

/*************************************************************************
function: AxfElfStack - return "where" string array
*/
data_ptr AxfElfStack ()
{
	int             line;
	package_info   *packP;
	char           *name;
	stack_frame    *fp;
	shrt           *pc;
	char			funcName[ELFIDSTR+1];
	char buf[STRPATH+200];
	data_ptr aP;
	int lines = 0;

	aP = ELF_CREATE_ARRAY(0);
#if !FOR_TESTER
	pc = Elf.pc;
	for (fp = Elf.curFrame; fp && !fp->stop_here; fp = fp->caller) {
		line = fp->curline;
		packP = fp->packP;
		name = packP->source_name;
    	if ( fp->funcstarts )
        	FetchNameOfFunction(funcName, fp->funcstarts);
    	else
			strcpy(funcName,"???");
		sprintf(buf,NXLT("line %d in function %s in file %s\n"), line, funcName, name);
		aP = ElfAddStrToArray(aP, lines++, buf);
		pc = fp->return_addr;
	}
#endif
	return(aP);
}

/*************************************************************************
function: ElfWhereLog -  like ElfWhere() bug log in to the AXTrace file
This is only called by SS to log errors
*/
ElfWhereLog()
{
    stack_frame    *fp;
	char			funcName[ELFIDSTR+1];

	if(!AXTrace)
		return;

    for (fp = Elf.curFrame; fp && !fp->stop_here; fp = fp->caller) {
    	if (fp && fp->funcstarts)
        	FetchNameOfFunction(funcName, fp->funcstarts);
    	else
        	funcName[0] = '\0';
        AXTRACE((NXLT("macro <%s> at line %d in file <%s>\n"), funcName, 
			fp->curline, fp->packP->source_name));
    }
    return (0);
}

#if WIN_DEBUG                   /* {  */

static char     WhereBuff[3000];


/*************************************************************************
function: ElfStaticWhere - stuff the  ELF's execution point into some
static storage so that we can find it from the windows debugger.
This is only called by DBX users to figure out what macro we crashed in.
*/
int 
ElfStaticWhere()
{
	int             line;
	package_info   *packP;
	char           *name;
	stack_frame    *fp;
	shrt           *pc;
	char           *buffP;
	int             num_chars;

	buffP = WhereBuff;
	pc = Elf.pc;
	for (fp = Elf.curFrame; fp && !fp->stop_here; fp = fp->caller) {
		line = fp->curline;
		packP = fp->packP;
		name = packP->source_name;
		num_chars = sprintf(buffP, NXLT("line %d in %s  ||  "), line, name);
		buffP = buffP + num_chars;
		pc = fp->return_addr;
	}
	return (0);
}
#endif                          /* } WIN_DEBUG  */


/*************************************************************************
function: ElfAbort - Throw an error, setting Elf.errorLine and ErrorPackage
*/
ElfAbort(code)
	int             code;
{
	Elf.errorLine = Elf.curFrame ? Elf.curFrame->curline : 0;
	Elf.errorPackage = FindPackage(Elf.pc);
	if (Elf.curFrame && Elf.curFrame->funcstarts)
		FetchNameOfFunction(Elf.ErrorFunc, Elf.curFrame->funcstarts);
	else
		Elf.ErrorFunc[0] = '\0';
	Elf.ErrorElfObj = Elf.curFrame ? Elf.curFrame->inObj : NULL;
	ElfDoAbort(code, NULL, NULL, Elf.errorLine, Elf.errorPackage, Elf.ErrorFunc, Elf.ErrorElfObj);
}

/*************************************************************************
function: ElfObjAbort - Throw an error, override error object
*/
ElfObjAbort(code, str, prepend, obj)
	int             code;
	char           *str;        /* NULL - use Applixstrfl, "" - use current */
	char           *prepend;    /* NULL/"" - no prepend info */
	elf_object	   *obj;
{
	char error_str[256];
	
	Elf.errorLine = Elf.curFrame ? Elf.curFrame->curline : 0;
	Elf.errorPackage = FindPackage(Elf.pc);
	if (Elf.curFrame && Elf.curFrame->funcstarts)
		FetchNameOfFunction(Elf.ErrorFunc, Elf.curFrame->funcstarts);
	else
		Elf.ErrorFunc[0] = '\0';
	if (!str || !*str)
		StrGet(code,error_str);
	else
		strcpy(error_str, str);
	/* append object name to error string */
	strcat(error_str, " ");
	if (obj)
		strcat(error_str, obj->nameTag);
	else
		strcat(error_str, XLT("<null object>"));
	
	Elf.ErrorElfObj = Elf.curFrame ? Elf.curFrame->inObj : NULL;
	ElfDoAbort(code, error_str, prepend, Elf.errorLine, Elf.errorPackage, Elf.ErrorFunc, Elf.ErrorElfObj);
}

/*************************************************************************
function: ElfStrAbort - Throw an error, override error string
*/
void ElfStrAbort(code, str, prepend)
	int             code;
	char           *str;        /* NULL - use Applixstrfl, "" - use current */
	char           *prepend;    /* NULL/"" - no prepend info */
{
	Elf.errorLine = Elf.curFrame ? Elf.curFrame->curline : 0;
	Elf.errorPackage = FindPackage(Elf.pc);
	if (Elf.curFrame && Elf.curFrame->funcstarts)
		FetchNameOfFunction(Elf.ErrorFunc, Elf.curFrame->funcstarts);
	else
		Elf.ErrorFunc[0] = '\0';
	Elf.ErrorElfObj = Elf.curFrame ? Elf.curFrame->inObj : NULL;
	ElfDoAbort(code, str, prepend, Elf.errorLine, Elf.errorPackage, Elf.ErrorFunc, Elf.ErrorElfObj);
}

/*************************************************************************
function: ElfbReAbort - Re-abort using old error values
*/
ElfbReAbort()
{
	if (!Elf.ErrorCode)
		ElfAbort(ERR_ILLVAL);
	else
		ElfDoAbort(Elf.ErrorCode, Elf.ErrorStr, Elf.ErrorObject,
				   Elf.errorLine, Elf.errorPackage, Elf.ErrorFunc, Elf.ErrorElfObj);
}

#define ELFAB_THROWER 0
#define ELFAB_MEMBER 1
#define ELFAB_CATCHER 2

/*************************************************************************
function: ElfDoAbort - Throw an error
This unwinds the stack, "returning" from all functions.  Detects
and redispatches on error catchers.
*/
static 
ElfDoAbort(code, errstr, prepend, err_line, err_packp, err_func, err_elf_object)
	int             code;
	char           *errstr;     /* NULL - use Applixstrfl, "" - use current */
	char           *prepend;
	shrt            err_line;
	register package_info *err_packp;
	char			*err_func;
	elf_object		*err_elf_object;
{
	extern int      ElfAllowAbort;
	char           *p;

#if !FOR_TESTER
	AxRecycleIfTaskLockingActive();     /* Only really needed for the socket
										 * code */
#endif

	Elf.ErrorCode = code;
	if ((errstr == NULL) || (*errstr == 0))
		StrGet(code, Elf.ErrorStr);
    else if (*errstr != 0) {
        strncpy(Elf.ErrorStr, errstr,MAX_ELF_ERRSTR);
        Elf.ErrorStr[MAX_ELF_ERRSTR] = 0;
    }
	if (Elf.ErrorStr[0] == '!') /* suppress leading beep for now */
		strcpy(Elf.ErrorStr, Elf.ErrorStr + 1);

/*
If there is a prepend string, replace current prepend str with it.
*/
	if (prepend && *prepend)
		strcpy(Elf.ErrorObject, prepend);
	else
		Elf.ErrorObject[0] = 0;

	/*
	 * if there is a macro name, replace it with err_func
	 */
	if (err_func && *err_func)
		strcpy(Elf.ErrorFunc, err_func);
	else
		Elf.ErrorFunc[0] = '\0';
	
	Elf.ErrorElfObj = err_elf_object;

/*
Note error file and line
*/

/*
If unwound entire stack, produce & post error message from here.
*/
	Elf.ErrorFile[0] = 0;
	Elf.ErrorLine = err_line;

	if (err_packp && err_packp->src_map) {
		if (err_packp->source_name[0]) {
			p = strrchr(err_packp->source_name, *DirSlash );
			if (!p)
				p = err_packp->source_name;
			else
				p++;
			strcpy(Elf.ErrorFile, p);
		}
	}
	if (ElfNumDebuggers && err_packp && err_packp->src_map)
		elfDbgHitBreak(ELFBRK_ERROR, err_packp, err_line, err_elf_object,NULL);

	while (Elf.curFrame && !Elf.curFrame->stop_here) {
/*
While un-winding, attempt to set error info if the error package
was compiled DEBUG mode.  Stop at lowest debug package.
*/
		if (!err_packp || !err_packp->src_map) {
			err_packp = Elf.errorPackage = FindPackage(Elf.pc);
			err_line = Elf.errorLine = Elf.curFrame->curline;
			Elf.ErrorLine = err_line;
			FetchNameOfFunction(Elf.ErrorFunc, Elf.curFrame->funcstarts);
			if (err_packp && err_packp->src_map) {
				if (err_packp->source_name[0]) {
					p = strrchr(err_packp->source_name, '/');
					if (!p)
						p = err_packp->source_name;
					strcpy(Elf.ErrorFile, p);
				}
				if (ElfNumDebuggers)
					elfDbgHitBreak(ELFBRK_ERROR, err_packp, err_line, err_elf_object,NULL);
			}
		}
/*
If an error catcher is encountered, change Elf.pc & stuff.  Restart
C execution from the Elf.abortBuf (normally ElfRun).
*/
		if (Elf.curFrame->error_catcher && !Elf.curFrame->error_caught) {
			Elf.curFrame->error_caught = TRUE;
			Elf.pc = Elf.curFrame->error_catcher;
			Elf.sp = Elf.curFrame->catcher_sp;

			/* reset broadcast info if any */
			if (Elf.curFrame->objVals) {
				ELF_DATA_FREE(Elf.curFrame->objVals);
				free(Elf.curFrame->loopArgs);
				Elf.curFrame->objVals = NULL;
				Elf.curFrame->loopArgs = NULL;
			}

			longjmp(Elf.abortBuf, 1);
		} else {
			DoReturn(NULL);     /* pop back up a level */
		}
	}
/*
If unwound entire stack, produce & post error message from here.
*/
	longjmp(Elf.abortBuf, 1);
}

/*************************************************************************
function: ElfbCurLineInfo - Return array of current line & file
*/
ElfbCurLineInfo()
{
	long            line;
	package_info   *packP;
	char            name[STRPATH + 1];

	line = Elf.curFrame ? Elf.curFrame->curline : 0;
	packP = FindPackage(Elf.pc);
	if (packP)
		strcpy(name, packP->source_name);
	else
		name[0] = 0;
	ElfRetData = ELF_CREATE_ARRAY(2);
	ElfAddNumToArray(ElfRetData, 0, line);
	ElfAddStrToArray(ElfRetData, 1, name);
}


/*************************************************************************
function: ElfInstMalloc - Allocate some VERY TEMPORARY memory
This memory goes away when the current Elf instruction completes!
*/
char           *
ElfInstMalloc(size)
	int             size;
{
	register garbage_info *p;

	p = (garbage_info *) malloc(size + sizeof(garbage_info) - 1);
	p->next = Elf.GarbageList;
	Elf.GarbageList = p;
	Elf.RunJanitoring = TRUE;
	return (p->data);
}

#if SYS&VMSSYS == 0
/*************************************************************************
function:  matherr - To get AT&T SYS-5 arithmetic exceptions correct
*/
matherr(x)
	struct exception *x;
{
	ElfAbort(ERR_ARITHMETIC);
}
#endif

/*************************************************************************
function: do_pow - Perform a pow() if values are kosher
Modeled from SS on 2/17/88
*/
double 
do_pow(value, exponent)
	double          value, exponent;
{
	double          z;

	/* for negative values we only allow integer exponents */
	VERIFY((value >= 0) || (exponent == floor(exponent)), ERR_ARITHMETIC);

	z = fabs(value);
	if (z < EPSILON)
		return (0.0);

	/* use logarithms to check for numeric overflow, before we raise  */
	/* anything to a power                                         */
#if !LINUX	/* for some reason, the log10 is failing on Linux */
	VERIFY((log10(z) * exponent) <= 151, ERR_ARITHMETIC);
#endif
	return (pow(value, exponent));
}

#if MEMTEST
typedef struct dmem {
	struct dmem    *next;
	struct dmem    *prev;
	char           *this;
}               dmem;
dmem           *ElfMemList;

/*************************************************************************
function: ElfRmvMem - Verify a free
This is a sanity test to make sure ElfFree is used only on ElfMalloc.
*/
ElfRmvMem(p)
	register char  *p;
{
	register dmem  *p2;

	for (p2 = ElfMemList; p2 && (p2->this != p); p2 = p2->next);
	if (!p2)
		abort();
	if (p2->prev)
		p2->prev->next = p2->next;
	else
		ElfMemList = p2->next;
	if (p2->next)
		p2->next->prev = p2->prev;
	free(p2);
}
/*************************************************************************
function: ElfAddMem - Add a malloc ptr
*/
ElfAddMem(p)
	register char  *p;
{
	register dmem  *p2;

	p2 = (dmem *) malloc(sizeof(dmem));
	p2->next = ElfMemList;
	p2->prev = NULL;
	p2->this = p;
	if (ElfMemList)
		ElfMemList->prev = p2;
	ElfMemList = p2;
}

#endif

/*************************************************************************
function: InstallCommandByName - Attempt to locate & install file
Searches user's macros dir, then site dir, then system dir.
*/
InstallCommandByName(name)
	char           *name;       /* name of command, also presumed name of file
								 * (minus extension) */
{
#if !FOR_TESTER
	char            filename[STRPATH + 1];
	char            retname[STRPATH + 1];
	char           *ThimSfxFromType();
	char           *UPneedByCode();
	int             ix;
	extern char    *ThimApplixDir;      /* Where the THIMBLE executable lies */
	data_ptr        searchList;
	int             searchListCount;
	char           *dir;

again:
	ElfbGetSysVar(NXLT("elf_search_list@"));
	searchList = ElfRetData;
	searchListCount = ElfbArraySize(searchList);
	if (0 == searchListCount) {
		ElfSetStandardSearchList();
		goto again;
	}
	sprintf(filename, "%s", name);
	for (ix = 0; ix < searchListCount; ix++) {
		dir = StrFromArray(searchList, ix);
		if (CaseLessFindMacroFile(dir, filename, retname)) {
			sprintf(filename, "%s/%s", dir, retname);
			AxSysDosUnixPath(filename);
			ElfDataFree(searchList);
			ElfbInstall(filename);
			return (TRUE);
		}
	}
	ElfDataFree(searchList);
	return (FALSE);
#endif
}

#if ELF_INTR_LOG
ElfIntrLogger(pc, opcode, word2)
{
	char           *p;

	switch (opcode) {
	case OP_SRC0:
		p = NXLT("OP_SRC0");
		break;
	case OP_SRC1:
		p = NXLT("OP_SRC1");
		break;
	case OP_SRC2:
		p = NXLT("OP_SRC2");
		break;
	case OP_SRC3:
		p = NXLT("OP_SRC3");
		break;
	case OP_SRC4:
		p = NXLT("OP_SRC4");
		break;
	case OP_SRC5:
		p = NXLT("OP_SRC5");
		break;
	case OP_SRC6:
		p = NXLT("OP_SRC6");
		break;
	case OP_SRC7:
		p = NXLT("OP_SRC7");
		break;
	case OP_PUSH:
		p = NXLT("OP_PUSH");
		break;
	case OP_PUSH_ARG:
		p = NXLT("OP_PUSH_ARG");
		break;
	case OP_POPTO:
		p = NXLT("OP_POPTO");
		break;
	case OP_POPTO_ARG:
		p = NXLT("OP_POPTO_ARG");
		break;
	case OP_PUSHARR:
		p = NXLT("OP_PUSHARR");
		break;
	case OP_PUSHARR_ARG:
		p = NXLT("OP_PUSHARR_ARG");
		break;
	case OP_POPDATA:
		p = NXLT("OP_POPDATA");
		break;
	case OP_POPDATA_ARG:
		p = NXLT("OP_POPDATA_ARG");
		break;
	case OP_POPTOARR:
		p = NXLT("OP_POPTOARR");
		break;
	case OP_POPTOARR_ARG:
		p = NXLT("OP_POPTOARR_ARG");
		break;
	case OP_POPADD:
		p = NXLT("OP_POPADD");
		break;
	case OP_POPADD_ARG:
		p = NXLT("OP_POPADD_ARG");
		break;
	case OP_POPSUB:
		p = NXLT("OP_POPSUB");
		break;
	case OP_POPSUB_ARG:
		p = NXLT("OP_POPSUB_ARG");
		break;
	case OP_PUSH1ARR:
		p = NXLT("OP_PUSH1ARR");
		break;
	case OP_PUSH1ARR_ARG:
		p = NXLT("OP_PUSH1ARR_ARG");
		break;
	case OP_POPTO1AR:
		p = NXLT("OP_POPTO1AR");
		break;
	case OP_POPTO1AR_ARG:
		p = NXLT("OP_POPTO1AR_ARG");
		break;
	case OP_POPDARR:
		p = NXLT("OP_POPDARR");
		break;
	case OP_POPDARR_ARG:
		p = NXLT("OP_POPDARR_ARG");
		break;
	case OP_SETOBJ:
		p = NXLT("OP_SETOBJ");
		break;
	case OP_SETOBJ_ARG:
		p = NXLT("OP_SETOBJ_ARG");
		break;
	case OP_INC:
		p = NXLT("OP_INC");
		break;
	case OP_DEC:
		p = NXLT("OP_DEC");
		break;
	case OP_CASE:
		p = NXLT("OP_CASE");
		break;
	case OP_JZ_SHRT:
		p = NXLT("OP_JZ_SHRT");
		break;
	case OP_JMP_SHRT:
		p = NXLT("OP_JMP_SHRT");
		break;
	case OP_ADD:
		p = NXLT("OP_ADD");
		break;
	case OP_SUB:
		p = NXLT("OP_SUB");
		break;
	case OP_MULT:
		p = NXLT("OP_MULT");
		break;
	case OP_DIV:
		p = NXLT("OP_DIV");
		break;
	case OP_EXP:
		p = NXLT("OP_EXP");
		break;
	case OP_LESS:
		p = NXLT("OP_LESS");
		break;
	case OP_GRTR:
		p = NXLT("OP_GRTR");
		break;
	case OP_LEQ:
		p = NXLT("OP_LEQ");
		break;
	case OP_GEQ:
		p = NXLT("OP_GEQ");
		break;
	case OP_IDIV:
		p = NXLT("OP_IDIV");
		break;
	case OP_EQV:
		p = NXLT("OP_EQV");
		break;
	case OP_IMP:
		p = NXLT("OP_IMP");
		break;
	case OP_REM:
		p = NXLT("OP_REM");
		break;
	case OP_AND:
		p = NXLT("OP_AND");
		break;
	case OP_OR:
		p = NXLT("OP_OR");
		break;
	case OP_XOR:
		p = NXLT("OP_XOR");
		break;
	case OP_IADD:
		p = NXLT("OP_IADD");
		break;
	case OP_ISUB:
		p = NXLT("OP_ISUB");
		break;
	case OP_IMULT:
		p = NXLT("OP_IMULT");
		break;
	case OP_ILESS:
		p = NXLT("OP_ILESS");
		break;
	case OP_IGRTR:
		p = NXLT("OP_IGRTR");
		break;
	case OP_ILEQ:
		p = NXLT("OP_ILEQ");
		break;
	case OP_IGEQ:
		p = NXLT("OP_IGEQ");
		break;
	case OP_SADD:
		p = NXLT("OP_SADD");
		break;
	case OP_SLESS:
		p = NXLT("OP_SLESS");
		break;
	case OP_SGRTR:
		p = NXLT("OP_SGRTR");
		break;
	case OP_SLEQ:
		p = NXLT("OP_SLEQ");
		break;
	case OP_SGEQ:
		p = NXLT("OP_SGEQ");
		break;
	case OP_EQUALS:
		p = NXLT("OP_EQUALS");
		break;
	case OP_NEQ:
		p = NXLT("OP_NEQ");
		break;
	case OP_UMINUS:
		p = NXLT("OP_UMINUS");
		break;
	case OP_NOT:
		p = NXLT("OP_NOT");
		break;
	case OP_PUSH_GLO:
		p = NXLT("OP_PUSH_GLO");
		break;
	case OP_PUSH_EXT:
		p = NXLT("OP_PUSH_EXT");
		break;
	case OP_POPTO_GLO:
		p = NXLT("OP_POPTO_GLO");
		break;
	case OP_POPTO_EXT:
		p = NXLT("OP_POPTO_EXT");
		break;
	case OP_PUSHARR_GLO:
		p = NXLT("OP_PUSHARR_GLO");
		break;
	case OP_PUSHARR_EXT:
		p = NXLT("OP_PUSHARR_EXT");
		break;
	case OP_POPDATA_GLO:
		p = NXLT("OP_POPDATA_GLO");
		break;
	case OP_POPDATA_EXT:
		p = NXLT("OP_POPDATA_EXT");
		break;
	case OP_POPTOARR_GLO:
		p = NXLT("OP_POPTOARR_GLO");
		break;
	case OP_POPTOARR_EXT:
		p = NXLT("OP_POPTOARR_EXT");
		break;
	case OP_POPADD_GLO:
		p = NXLT("OP_POPADD_GLO");
		break;
	case OP_POPADD_EXT:
		p = NXLT("OP_POPADD_EXT");
		break;
	case OP_POPSUB_GLO:
		p = NXLT("OP_POPSUB_GLO");
		break;
	case OP_POPSUB_EXT:
		p = NXLT("OP_POPSUB_EXT");
		break;
	case OP_PUSH1ARR_GLO:
		p = NXLT("OP_PUSH1ARR_GLO");
		break;
	case OP_PUSH1ARR_EXT:
		p = NXLT("OP_PUSH1ARR_EXT");
		break;
	case OP_POPTO1AR_GLO:
		p = NXLT("OP_POPTO1AR_GLO");
		break;
	case OP_POPTO1AR_EXT:
		p = NXLT("OP_POPTO1AR_EXT");
		break;
	case OP_POPDARR_GLO:
		p = NXLT("OP_POPDARR_GLO");
		break;
	case OP_POPDARR_EXT:
		p = NXLT("OP_POPDARR_EXT");
		break;
	case OP_SETOBJ_GLO:
		p = NXLT("OP_SETOBJ_GLO");
		break;
	case OP_SETOBJ_EXT:
		p = NXLT("OP_SETOBJ_EXT");
		break;
	case OP_CALL:
		p = NXLT("OP_CALL");
		break;
	case OP_CALL_EXT:
		p = NXLT("OP_CALL_EXT");
		break;
	case OP_RCALL:
		p = NXLT("OP_RCALL");
		break;
	case OP_RCALL_EXT:
		p = NXLT("OP_RCALL_EXT");
		break;
	case OP_CALL0:
		p = NXLT("OP_CALL0");
		break;
	case OP_CALL0_EXT:
		p = NXLT("OP_CALL0_EXT");
		break;
	case OP_RCALL0:
		p = NXLT("OP_RCALL0");
		break;
	case OP_RCALL0_EXT:
		p = NXLT("OP_RCALL0_EXT");
		break;
	case OP_ONERROR:
		p = NXLT("OP_ONERROR");
		break;
	case OP_PUSH_LIT:
		p = NXLT("OP_PUSH_LIT");
		break;
	case OP_JZ:
		p = NXLT("OP_JZ");
		break;
	case OP_JMP:
		p = NXLT("OP_JMP");
		break;
	case OP_FORLOOP:
		p = NXLT("OP_FORLOOP");
		break;
	case OP_FORTEST:
		p = NXLT("OP_FORTEST");
		break;
	case OP_RETVAL:
		p = NXLT("OP_RETVAL");
		break;
	case OP_RETURN:
		p = NXLT("OP_RETURN");
		break;
	case OP_NOP:
		p = NXLT("OP_NOP");
		break;
	case OP_COMPCALL:
		p = NXLT("OP_COMPCALL");
		break;
	case OP_COMPGOTO:
		p = NXLT("OP_COMPGOTO");
		break;
	case OP_ERRERR:
		p = NXLT("OP_ERRERR");
		break;
	case OP_ERROR:
		p = NXLT("OP_ERROR");
		break;
	case OP_RETERR:
		p = NXLT("OP_RETERR");
		break;
	case OP_POPSTEP:
		p = NXLT("OP_POPSTEP");
		break;
	case OP_POPSTOP:
		p = NXLT("OP_POPSTOP");
		break;
	case OP_ADDSTEP:
		p = NXLT("OP_ADDSTEP");
		break;
	case OP_SRCMAP:
		p = NXLT("OP_SRCMAP");
		break;
	case OP_SELECTER:
		p = NXLT("OP_SELECTER");
		break;
	case OP_ENTER:
		p = NXLT("OP_ENTER");
		break;
	case OP_PUSH_ENT:
		p = NXLT("OP_PUSH_ENT");
		break;
	case OP_PUSH_EXTENT:
		p = NXLT("OP_PUSH_EXTENT");
		break;
	case OP_ENTRYCALL:
		p = NXLT("OP_ENTRYCALL");
		break;
	case OP_RENTRYCALL:
		p = NXLT("OP_RENTRYCALL");
		break;
	case OP_0ENTRYCALL:
		p = NXLT("OP_0ENTRYCALL");
		break;
	case OP_0RENTRYCALL:
		p = NXLT("OP_0RENTRYCALL");
		break;
	case OP_IEQUALS:
		p = NXLT("OP_IEQUALS");
		break;
	case OP_INEQ:
		p = NXLT("OP_INEQ");
		break;
	case OP_IFORTEST:
		p = NXLT("OP_IFORTEST");
		break;
	case OP_IUMINUS:
		p = NXLT("OP_IUMINUS");
		break;
	case OP_PUSH_NUM:
		p = NXLT("OP_PUSH_NUM");
		break;
	case OP_BCALL0:
		p = NXLT("OP_BCALL0");
		break;
	case OP_BCALL:
		p = NXLT("OP_BCALL");
		break;
	case OP_KCALL0:
		p = NXLT("OP_KCALL0");
		break;
	case OP_KCALL:
		p = NXLT("OP_KCALL");
		break;
	case OP_RBCALL0:
		p = NXLT("OP_RBCALL0");
		break;
	case OP_RBCALL:
		p = NXLT("OP_RBCALL");
		break;
	case OP_RKCALL0:
		p = NXLT("OP_RKCALL0");
		break;
	case OP_RKCALL:
		p = NXLT("OP_RKCALL");
		break;
	case OP_PUSH_NULL:
		p = NXLT("OP_PUSH_NULL");
		break;
	case OP_TRCALL:
		p = NXLT("OP_TRCALL");
		break;
	case OP_XTRCALL:
		p = NXLT("OP_XTRCALL");
		break;
	case OP_SQL:
		p = NXLT("OP_SQL");
		break;
		/* case OP_SPARE: p = NXLT("OP_SPARE"); break; */
	case OP_PUSH_OBJ:
		p = NXLT("OP_PUSH_OBJ");
		break;
	case OP_PUSH_SYS:
		p = NXLT("OP_PUSH_SYS");
		break;
	case OP_POPTO_OBJ:
		p = NXLT("OP_POPTO_OBJ");
		break;
	case OP_POPTO_SYS:
		p = NXLT("OP_POPTO_SYS");
		break;
	case OP_PUSHARR_OBJ:
		p = NXLT("OP_PUSHARR_OBJ");
		break;
	case OP_PUSHARR_SYS:
		p = NXLT("OP_PUSHARR_SYS");
		break;
	case OP_POPDATA_OBJ:
		p = NXLT("OP_POPDATA_OBJ");
		break;
	case OP_POPDATA_SYS:
		p = NXLT("OP_POPDATA_SYS");
		break;
	case OP_POPTOARR_OBJ:
		p = NXLT("OP_POPTOARR_OBJ");
		break;
	case OP_POPTOARR_SYS:
		p = NXLT("OP_POPTOARR_SYS");
		break;
	case OP_POPADD_OBJ:
		p = NXLT("OP_POPADD_OBJ");
		break;
	case OP_POPADD_SYS:
		p = NXLT("OP_POPADD_SYS");
		break;
	case OP_POPSUB_OBJ:
		p = NXLT("OP_POPSUB_OBJ");
		break;
	case OP_POPSUB_SYS:
		p = NXLT("OP_POPSUB_SYS");
		break;
	case OP_PUSH1ARR_OBJ:
		p = NXLT("OP_PUSH1ARR_OBJ");
		break;
	case OP_PUSH1ARR_SYS:
		p = NXLT("OP_PUSH1ARR_SYS");
		break;
	case OP_POPTO1AR_OBJ:
		p = NXLT("OP_POPTO1AR_OBJ");
		break;
	case OP_POPTO1AR_SYS:
		p = NXLT("OP_POPTO1AR_SYS");
		break;
	case OP_POPDARR_OBJ:
		p = NXLT("OP_POPDARR_OBJ");
		break;
	case OP_POPDARR_SYS:
		p = NXLT("OP_POPDARR_SYS");
		break;
	case OP_SETOBJ_OBJ:
		p = NXLT("OP_SETOBJ_OBJ");
		break;
	case OP_SETOBJ_SYS:
		p = NXLT("OP_SETOBJ_SYS");
		break;
	default:
		p = "???";
	}
	printf(NXLT("%X: %s %d at line %d in %s\n"), pc, p, word2, Elf.curFrame->curline,
		   Elf.curFrame->packP->source_name);
	fflush(stdout);
}
#endif

/*************************************************************************
function: ElfRecordCall - Record function call
??x special hook needed for select_window()
*/
ElfRecordCall(functype, fp, pc, num_args)
	int             functype;
	stack_frame    *fp;
	shrt           *pc;
	int             num_args;
{
	char            name[STRLONG];
	stack_frame    *fpx;
	int             ix;
	data_ptr        args[100];
	package_info   *packP;
	char           *varsBeginP;
	char           *vp;
	char           *names[100];
	char           *formats[100];
	int             dst_args;

	if ((functype == TYPE_QFUNC) || (functype == TYPE_UIFUNC))
		return;

	/* Don't record macro calls within funcs or other macro calls */

	if (fp->caller) {
		for (fpx = fp->caller; fpx; fpx = fpx->caller) {
			if (fpx->functype == TYPE_QFUNC)
				return;
			if (fpx->recorded)
				return;
		}
	}
	fp->recorded = TRUE;
	FetchNameOfFunction(name, pc);

	/* reject object calls and NRMACRO calls */

	if ((functype == TYPE_NRFUNC)
		|| (functype == TYPE_SETOBJ)
		|| (functype == TYPE_GETOBJ))
		ElfStrAbort(ERR_NORECORD, NULL, name);

/* get names & formats for args vars */

	packP = fp->packP;
	varsBeginP = packP->varNames;

	ix = 0;
	for (vp = varsBeginP + fp->varnames; *vp && (ix < fp->numargs); ix++) {
		names[ix] = vp;
		vp += strlen(vp) + 1;
		formats[ix] = vp;
		vp += strlen(vp) + 1;
	}
	dst_args = ix;

	num_args = MIN(100, num_args);
	for (ix = 0; ix < num_args; ix++)
		args[ix] = POP_VAL;
	ElfRecord(ELFREC_FCALL, name, num_args, args, dst_args, names, formats, packP);
	Elf.sp += num_args;
	return;
}

/*************************************************************************
function: FetchNameOfFunction - Return called function name
*/
FetchNameOfFunction(name, pc)
	char           *name;       /* returned name */
	shrt           *pc;         /* addr of TYPE_XXX word */
{
	package_info   *packP;      /* used uninitialized? */
	register int    ix;
	register gfunc_info *sP;
	int             tmp;

	if (!pc) {
		strcpy(name, "???");    /* sanity */
		return;
	}
	switch (*pc) {
	case TYPE_QFUNC:
	case TYPE_MACRO:
	case TYPE_UIFUNC:
	case TYPE_NRFUNC:
	case TYPE_SETOBJ:
	case TYPE_GETOBJ:
		packP = FindPackage(pc);
		break;

	case TYPE_C_FUNC:
	case TYPE_C_MACRO:
		tmp = *((int *) (pc + 2));
		strcpy(name, ElfSymbolNameFromId(tmp, ElfLanguage));
		return;

	default:
		ElfAbort(ERR_NOTFUNC);
	}

	for (ix = 0; sP = packP->gfun_list[ix]; ix++) {
		if (pc == sP->value) {
			strcpy(name, sP->name);
			return;
		}
	}
	strcpy(name, "???");
}

/*************************************************************************
function: ElfAttachDebugger - Add a debugger to the list
Called directly by macro editor
*/
ElfAttachDebugger()
{
	int ix;
	int uid;

	ElfDetachDebugger();        /* in case already listed */
	uid = TaskInfo[THIMpid].uid;
	if (ElfNumDebuggers >= MAX_DEBUGGERS-1)
		return(FALSE);
	ElfDebugger[ElfNumDebuggers].dbg_uid = uid;
	ElfDebugger[ElfNumDebuggers].is_error_break = TRUE;	/* catch errors */
	ElfNumDebuggers++;
	return (TRUE);
}

/*************************************************************************
function: ElfDetachDebugger - Remove debugger from list
*/
ElfDetachDebugger()
{
	int ix;
	int uid;

	uid = TaskInfo[THIMpid].uid;
	for (ix = 0; ix < ElfNumDebuggers; ix++) {
		if (ElfDebugger[ix].dbg_uid == uid) {
			elfDbgCloseDebugger(&ElfDebugger[ix],TRUE);
			return(TRUE);
		}
	}
	return(FALSE);
}

/***************************************************************************
function: elfDbgCloseDebugger - remove debugger from the list of debuggers
and reset debugger count.
*/
static
elfDbgCloseDebugger (dbginfoP, detach_only)
debugger_info *dbginfoP;
int detach_only;		/* if TRUE, do not clear break points */
{
	register int ix;
	
	if (!detach_only) {
		/* clear all breakpoints of this debugger */
		for (ix = ElfNumBreakpoints-1; ix >= 0; --ix) {
			if (ElfBreakData[ix].dbg_uid == dbginfoP->dbg_uid)
				ElfClrBreakPoint(ElfBreakData[ix].packP->source_name,
								 ElfBreakData[ix].line, ElfBreakData[ix].obj);
		}
	}
	/* clear debugger info */
	memset(dbginfoP, 0, sizeof(debugger_info));

	for (ix = 0; ix < ElfNumDebuggers; ++ix) {
		if (!ElfDebugger[ix].dbg_uid) {
			bcopy(&ElfDebugger[ix+1],&ElfDebugger[ix],
				  sizeof(debugger_info)*(ElfNumDebuggers-ix));
			ElfNumDebuggers--;
			break;
		}
	}
}

/***************************************************************************
function: elfDbgSetNexterFrame - change nexter frame of debugger
*/
static
elfDbgSetNexterFrame (old_fp, new_fp)
stack_frame *old_fp, *new_fp;
{
	register int ix;
	
	for (ix = 0; ix < ElfNumDebuggers; ix++)
		if (ElfDebugger[ix].nexterFrame == old_fp)
			ElfDebugger[ix].nexterFrame = new_fp;
}


/***************************************************************************
function: ElfSetErrorBreak - set catch errors flag for debugger
*/
int
ElfSetErrorBreak (is_error_break)
int is_error_break;
{
	int ix;
	int uid;
	
	uid = TaskInfo[THIMpid].uid;
	for (ix = 0; ix < ElfNumDebuggers; ix++) {
		if (ElfDebugger[ix].dbg_uid == uid) {
			ElfDebugger[ix].is_error_break = is_error_break;
			return(TRUE);
		}
	}
	return(FALSE);
}

/***************************************************************************
function: ElfGetErrorBreak - return catch errors flag of debugger
*/
int
ElfGetErrorBreak ()
{
	int ix;
	int uid;
	
	uid = TaskInfo[THIMpid].uid;
	for (ix = 0; ix < ElfNumDebuggers; ix++) {
		if (ElfDebugger[ix].dbg_uid == uid)
			return(ElfDebugger[ix].is_error_break);
	}
	return(FALSE);
}

/***************************************************************************
function: ElfDbgGetStopStatus - return TRUE if debugger is stopped and
waiting for response.
*/
int
ElfDbgGetStopStatus ()
{
	register int ix;
	int uid;
	
	uid = TaskInfo[THIMpid].uid;
	for (ix = 0; ix < ElfNumDebuggers; ix++) {
		if (ElfDebugger[ix].dbg_uid == uid)
			return(ElfDebugger[ix].is_stopped);
	}
	return(FALSE);
}


/***************************************************************************
function: ElfDbgExecuteToHere - execute until this packP and line reached
*/
data_ptr
ElfDbgExecuteToHere (task, packname, lineno)
int task;		/* task under debug */
char *packname;	/* package name */
int lineno;		/* line no */
{
	register int ix, uid, pid;
	int saved_num;
	char msg_buf[256];
	debugger_info *dbginfoP = NULL;
	
	uid = TaskInfo[THIMpid].uid;
	for (ix = 0; ix < ElfNumDebuggers; ix++) {
		if (ElfDebugger[ix].dbg_uid == uid) {
			dbginfoP = &ElfDebugger[ix];
			break;
		}
	}
	if (!dbginfoP || !ThimPidFromUID(dbginfoP->dbg_uid))
		return(NULL);
	saved_num = ElfNumBreakpoints;
	ElfSetBreakPoint(packname,lineno,NULL);
	memcpy(&dbginfoP->exbrk, &ElfBreakData[ElfNumBreakpoints-1],
		   sizeof(breakpoint_t));
	ElfNumBreakpoints = saved_num;
	if (!ElfDebugResponse(task,ELFBRK_CONTINUE)) {
		dbginfoP->exbrk.packP = NULL;
		pid = ThimPidFromUID(task);
		if (!pid) {
			strcpy(msg_buf,
			  XLT("Application already completed or hasn't been started yet."));
		}
		else if (TaskInfo[pid].rsp || !TaskInfo[pid].ipc_response_waiting) {
			strcpy(msg_buf, XLT("Application not listening now."));
		}
		return(ELF_CREATE_STRING(msg_buf));
	}
	return(NULL);
}


/***************************************************************************
function: ElfDbgStopInFunc - set a break in function.
*/
data_ptr
ElfDbgStopInFunc (funcP, dP)
data_ptr funcP, dP;
{
	register int ix, px;
	int hashval;
	char *packnameP = NULL;
	elf_object *oP = NULL;
	package_info *packP = NULL;
	char *p, funcname[STRLONG + 1];
	char packname[STRPATH + 1];
	char cmpname[STRPATH + 1];
	char packdir[STRPATH + 1];
	register gfunc_info *sP = NULL;
	data_ptr aP;

	strcpy(funcname, StrFromDataPtr(funcP));
	for (p = funcname; *p; p++)	/* convert to uppercase */
		*p = toupper(*p);
	hashval = hash(funcname);
	
	if (!dP) {
	no_pack_name:
		for (px = 0; packP = Packages[px]; px++) {
			if (packP->src_map) {
				for (ix = 0; sP = packP->gfun_list[ix]; ++ix) {
					if (sP->hashval == hashval && streqi(sP->name,funcname))
						goto found;
				}
			}
		}
	}
	else if (ElfbIsString(dP)) {
		packnameP = StrFromDataPtr(dP);
		if (!packnameP || !*packnameP)
			goto no_pack_name;
		ParsePath(packnameP,packdir,packname);
		for (px = 0; packP = Packages[px]; px++) {
			if (packP->src_map) {
				ParsePath(packP->source_name, packdir, cmpname);
#if AXMS
				if (streqi(cmpname, packname))
#else
				if (streq(cmpname, packname))
#endif
				{
					for (ix = 0; sP = packP->gfun_list[ix]; ++ix) {
						if (sP->hashval == hashval && streqi(sP->name,funcname))
							goto found;
					}
					goto found;
				}
			}
		}
	}
	else if (ElfbIsObject(dP)) {
		oP = ElfObjPtr(dP);
		for (; oP; oP = oP->inheritFromP) {
			packP = oP->packP;
			if (packP && packP->src_map) {
				for (ix = 0; sP = packP->gfun_list[ix]; ++ix) {
					if (sP->hashval == hashval && streqi(sP->name,funcname))
						goto found;
				}
			}
		}
	}
 found:
	if (!packP || !sP)
		return(NULL);
	if (ElfSetBreakPoint(packP->source_name, sP->value[7], NULL)) {
		aP = ELF_CREATE_ARRAY(2);
		aP = ElfAddStrToArray(aP, 0, packP->source_name);
		aP = ElfAddNumToArray(aP, 0, sP->value[7]);
		return(aP);
	}
	return(NULL);
}


/*************************************************************************
function: ElfSetBreakPoint - Set a breakpoint for the debuggger
*/
ElfSetBreakPoint (packnameP, line, obj)
	char           *packnameP;
	int             line;
	data_ptr        obj;
{
#if !FOR_TESTER
	package_info   *packP;
	int             ix, px;
	char            packname[STRPATH + 1];
	char            cmpname[STRPATH + 1];

	MakeAbsolutePathname(packname, packnameP, TRUE);
	ElfClrBreakPoint(packname, line, obj);

	if (ElfNumBreakpoints == MAX_ELF_BREAKS)
		return (FALSE);

	for (px = 0; packP = Packages[px]; px++) {
		if (packP->src_map) {
			MakeAbsolutePathname(cmpname, packP->source_name, TRUE);
#if AXMS
			if (streqi(cmpname, packname))
#else
			if (streq(cmpname, packname))
#endif
			{
				ix = ElfNumBreakpoints++;
				ElfBreakData[ix].packP = packP;
				ElfBreakData[ix].line = line;
				ElfBreakData[ix].obj = AxTaskCopyElfData(0, obj);
				ElfBreakData[ix].dbg_uid = TaskInfo[THIMpid].uid;
				return (TRUE);
			}
		}
	}
	return (FALSE);
#endif
}

/*************************************************************************
function: ElfSetBreakPoint - Set a breakpoint for the debuggger
*/
ElfClrBreakPoint (packnameP, line, obj)
	char           *packnameP;
	int             line;
	data_ptr        obj;
{
#if !FOR_TESTER
	int             ix, px;
	package_info   *packP;
	char            packname[STRPATH + 1];
	char            cmpname[STRPATH + 1];

    if ( packnameP == (char *)NULL )
        /* can happen with rapid double click on "clear breakpoints" button */
        /* under NT - jon                                            */
        return;

	MakeAbsolutePathname(packname, packnameP, TRUE);
	for (px = 0; packP = Packages[px]; px++) {
		if ( !packP->source_name )	/* skip placeholders */
			continue;
		MakeAbsolutePathname(cmpname, packP->source_name, TRUE);
		if (streq(cmpname, packname))
			break;
	}
	if (!packP)
		return (FALSE);

	for (ix = 0; ix < ElfNumBreakpoints; ix++) {
		if ((ElfBreakData[ix].packP == packP) &&
			(ElfBreakData[ix].line == line)) {
			AxTaskFreeElfData(0,ElfBreakData[ix].obj);	/* free breakpt object */
			bcopy(&ElfBreakData[ix+1], &ElfBreakData[ix],
				  sizeof(breakpoint_t)*(ElfNumBreakpoints-ix));
			ElfNumBreakpoints--;
			return (TRUE);
		}
	}
	return (FALSE);
#endif
}

/*************************************************************************
function: ElfClrAllBreaks - Clear all breakpoints for this debugger
*/
ElfClrAllBreaks ()
{
	register int ix;
	int uid;

	uid = TaskInfo[THIMpid].uid;	/* clear all breakpts in this debugger */
	for (ix = ElfNumBreakpoints - 1; ix >= 0; ix--) {
		if (ElfBreakData[ix].dbg_uid == uid)
			ElfClrBreakPoint(ElfBreakData[ix].packP->source_name,
							 ElfBreakData[ix].line, ElfBreakData[ix].obj);
	}
}

/*************************************************************************
function: elfDbgCheckForBreak - Determine whether current line is a
breakpoint line
*/
static 
elfDbgCheckForBreak (inObj)
register elf_object *inObj; /* NULL or object ptr */
{
#if !FOR_TESTER
	register int    ix, jx;
	register int    line;
	register package_info *packP;

	if (ElfNumDebuggers <= 0)
		return;
	
	packP = Elf.curFrame->packP;
	line = Elf.curFrame->curline;

	if (!packP || !packP->src_map)		/* can not break in this package */
		return;
	
	for (ix = 0; ix < ElfNumDebuggers; ++ix) {
		if (ElfDebugger[ix].is_stepping &&
			ElfDebugger[ix].debuggee_uid == TaskInfo[AxInTask].uid) {
			elfDbgHitBreak(ELFBRK_STEP, packP, line,
						   inObj,&ElfDebugger[ix]);
			return;
		}
		else if (ElfDebugger[ix].is_nexting &&
				 ElfDebugger[ix].nexterFrame == Elf.curFrame &&
				 ElfDebugger[ix].debuggee_uid == TaskInfo[AxInTask].uid) {
			elfDbgHitBreak(ELFBRK_NEXT, packP, line,
						   inObj,&ElfDebugger[ix]);
			return;
		}
		else if (ElfDebugger[ix].exbrk.packP == packP &&
				 ElfDebugger[ix].exbrk.line == line &&
				 ElfDebugger[ix].debuggee_uid == TaskInfo[AxInTask].uid) {
			ElfDebugger[ix].exbrk.packP = NULL; /* done with this */
			elfDbgHitBreak(ELFBRK_BRK, packP, line,
						   inObj,&ElfDebugger[ix]);
		}
			
	}

	for (ix = 0; ix < ElfNumBreakpoints; ix++) {	/* find break point */
		if ((ElfBreakData[ix].line == line) &&
			(ElfBreakData[ix].packP == packP)) {
			for (jx = 0; jx < ElfNumDebuggers; ++jx) { /* find the debugger */
				if (ElfBreakData[ix].dbg_uid == ElfDebugger[jx].dbg_uid) {
					elfDbgHitBreak(ELFBRK_BRK, packP, line,
								   inObj,&ElfDebugger[jx]);
					return;
				}
			}
			return;
		}
	}
#endif
}

/*************************************************************************
function: ElfDebugResponse - Send message to task under debug
*/
ElfDebugResponse (task, cmd)
	int             task;       /* presumed Uid of task being debugged */
	int             cmd;
{
#if !FOR_TESTER
	int             pid;

	pid = ThimPidFromUID(task);
	if (!pid) {
		AxInfoMessage(-1, ERR_ILLVAL, NULL,
				XLT("?tMacro already completed or hasn't been started yet."));
		return (FALSE);
	}
	if (TaskInfo[pid].rsp || !TaskInfo[pid].ipc_response_waiting) {
		AxInfoMessage(-1, ERR_ILLVAL, NULL, XLT("?tMacro not listening now."));
		return (FALSE);
	}
	IPCXsendMes(IPC_RES, pid, &cmd, sizeof(cmd), APELFDEBUG, AxInTask);
	return (TRUE);
#endif
}

/*************************************************************************
function: ElfGetBreakPackP - Return package of breakpoint (with stack level)
*/
package_info   *
ElfGetBreakPackP (level)
	int             level;      /* stack level, 0 is at breakpoint */
{
	stack_frame    *fp = NULL;
	int             ix;

	for (ix = 0; ix < ElfNumDebuggers; ix++) {
		if (ElfDebugger[ix].dbg_uid == TaskInfo[THIMpid].uid) {
			if (!ThimPidFromUID(ElfDebugger[ix].debuggee_uid)) {
				/* debugee crashed */
				ElfDebugger[ix].is_stopped = FALSE;
				ElfDebugger[ix].is_nexting = FALSE;
				ElfDebugger[ix].is_stepping = FALSE;
				ElfDebugger[ix].breakFrame = NULL;
				ElfDebugger[ix].breakGlobals = NULL;
				ElfDebugger[ix].exbrk.packP = NULL;
			}
			fp = ElfDebugger[ix].breakFrame;
			break;
		}
	}

	if (!fp)
		return (NULL);

	for (ix = 0; ix < level; ix++) {
		fp = fp->caller;
		if (!fp)
			return (NULL);
	}
	return (FindPackage(fp->pc));
}

/*************************************************************************
function: ElfFetchVar - Return a var's value to debugger
*/
typeErr 
ElfFetchVar (varP, formatNameP, formatInfoP, varname, level)
	data_ptr       *varP;       /* OUT, variable's data */
	char          **formatNameP;/* OUT, variable's format name */
	data_ptr       *formatInfoP;/* OUT, variable's format definition */
	char           *varname;    /* IN varname */
	int             level;      /* stack level, 0 is at breakpoint */
{
	char           *p;
	package_info   *packP;
	int             ix, px;
	stack_frame    *fp;
	data_ptr       *dp;
	int             locals_count, args_count, total_vars;
	data_ptr       *globalsP;
	char           *varsBeginP;
	int             val;        /* offset in Elf.taskFrame for global vars */
	ext_link        info;
	int             hashval;
	elf_obj_grp    *gP;
	gvar_info      *gv;

/* find frame/package */


	fp = NULL;
	for (ix = 0; ix < ElfNumDebuggers; ix++) {
		if (ElfDebugger[ix].dbg_uid == TaskInfo[THIMpid].uid) {
			if (!ThimPidFromUID(ElfDebugger[ix].debuggee_uid)) {
				/* debugee crashed */
				ElfDebugger[ix].is_stopped = FALSE;
				ElfDebugger[ix].is_nexting = FALSE;
				ElfDebugger[ix].is_stepping = FALSE;
				ElfDebugger[ix].breakFrame = NULL;
				ElfDebugger[ix].breakGlobals = NULL;
				ElfDebugger[ix].exbrk.packP = NULL;
			}
			fp = ElfDebugger[ix].breakFrame;
			globalsP = ElfDebugger[ix].breakGlobals;
			break;
		}
	}

	if (!fp)
		goto all_sysvars;

	for (ix = 0; ix < level; ix++) {
		fp = fp->caller;
		if (!fp)
			return (1);
	}

	locals_count = fp->numlocs - MAX_REGS;
	args_count = fp->numargs;
	total_vars = locals_count + args_count;

	packP = fp->packP;
	if (!packP)
		return (1);
	varsBeginP = packP->varNames;
	*formatInfoP = packP->formats;

/* find var in package list */

	ix = 0;
	for (p = varsBeginP + fp->varnames; *p && (ix < total_vars); ix++) {
		if (streq(p, varname)) {
			p += strlen(p) + 1;
			*formatNameP = p;
			break;
		}
		p += strlen(p) + 1;     /* get past this var name */
		p += strlen(p) + 1;     /* get past its format */
	}

	if (ix < total_vars) {      /* is a local var/arg */
		if (ix < fp->numargs) {
			dp = fp->argvars[ix];
			dp = (data_ptr *) (((long) dp) & ~FAULT_BIT);
			*varP = *dp;
		} else {
			ix -= fp->numargs;
			ix += MAX_REGS;
			if (ix < 0)
				return (1);
			*varP = fp->locvars[ix];
		}
	} else {                    /* not local, try for object/global var */
#if !FOR_TESTER
		if (fp->inObj) {
			hashval = hash(varname);
			gP = ElfFindObjLocalData(fp->inObj, varname, hashval);
			if (gP) {
				*formatNameP = "";
				for (ix = 0; fp->packP->ovar_list[ix]; ix++) {
					if (streq(fp->packP->ovar_list[ix]->name, varname)) {
						*formatNameP = fp->packP->ovar_list[ix]->formatP;
						if (!(*formatNameP))    /* don't return NULL */
							*formatNameP = "";  /* do ptr to empty str instead */
						break;
					}
				}
				*varP = gP->value;
				return (0);
			}
		}
#endif

		/* Check for SYSVAR in current frame */

		packP = fp->packP;
		for (ix = 0; packP->svar_list[ix]; ix++) {
			if (streq(packP->svar_list[ix]->name, varname)) {
				*formatNameP = packP->svar_list[ix]->formatP;
				if (!(*formatNameP))    /* don't return NULL */
					*formatNameP = "";  /* do ptr to empty str instead */
				*varP = ElfGetSysVar(varname);
				return(0);
			}
		}

		/* Check for SYSVAR in ANY package */

all_sysvars:
		for (px = 0; packP = Packages[px]; px++) {
			if (packP->code_starts==NULL)
				continue; /* dummy package */
			for (ix = 0; packP->svar_list[ix]; ix++) {
				if (streq(packP->svar_list[ix]->name, varname)) {
					*formatNameP = packP->svar_list[ix]->formatP;
					if (!(*formatNameP))    /* don't return NULL */
						*formatNameP = "";  /* do ptr to empty str instead */
					*varP = ElfGetSysVar(varname);
					return(0);
				}
			}
		}

		/* Assume Global Var... */

		if (!fp)
			return (-1);
		memset(&info, 0, sizeof(info));
		info.nameP = elfHashNameAndVal(varname, &hashval);
		info.hashval = hashval;
		info.moduleName = fp->packP->moduleName;
		val = SnapDataDoIt(&info, &gv);
		if (val == -1) {	/* not a global in this file */
			info.moduleName = NULL;	/* try all globals */
			val = SnapDataDoIt(&info, &gv);
		}
		if (val == -1)
			return (1);
		if (gv->formatP)
			*formatNameP = gv->formatP;
		else
			*formatNameP = "";
		*varP = globalsP[val];
	}
	return (0);
}

/*************************************************************************
function: ElfWhereBreakFrame - Return frame info of a break level
This is used by the debugger's where command
*/
ElfWhereBreakFrame (funcname, line, package, inObj, srcObject, level)
	int             level;      /* IN: how many frames to pop up before
								 * reporting */
	char           *funcname;   /* OUT: Name of function stoped in */
	int            *line;       /* OUT: Line of source */
	char           *package;    /* OUT: Source */
	elf_object    **inObj;      /* OUT: Object breakpointd in or NULL */
	elf_object    **srcObject;  /* OUT: Object that contains the relavent
								 * source */
{
	int             ix;
	stack_frame    *fp = NULL;
	package_info   *packP;
	elf_object     *obj;

	strcpy(funcname, "???");    /* assume failure */
	strcpy(package, "???");
	*line = 1;

	for (ix = 0; ix < ElfNumDebuggers; ix++) {
		if (ElfDebugger[ix].dbg_uid == TaskInfo[THIMpid].uid) {
			if (!ThimPidFromUID(ElfDebugger[ix].debuggee_uid)) {
				/* debugee crashed */
				ElfDebugger[ix].is_stopped = FALSE;
				ElfDebugger[ix].is_nexting = FALSE;
				ElfDebugger[ix].is_stepping = FALSE;
				ElfDebugger[ix].breakFrame = NULL;
				ElfDebugger[ix].breakGlobals = NULL;
				ElfDebugger[ix].exbrk.packP = NULL;
			}
			fp = ElfDebugger[ix].breakFrame;
			break;
		}
	}
	if (!fp)
		return (FALSE);
	for (ix = 0; ix < level; ix++) {
		fp = fp->caller;
		if (!fp)
			return (FALSE);
	}
	*line = fp->curline;
	packP = fp->packP;
	if (packP && packP->src_map)
		strcpy(package, packP->source_name);
	else
		strcpy(package, "???");
	FetchNameOfFunction(funcname, fp->funcstarts);

	*inObj = obj = fp->inObj;
	while (obj) {
		if (obj->packP == packP)
			break;
		obj = obj->inheritFromP;
	}
	*srcObject = obj;
	return (TRUE);
}

ElfWhereBreak (funcname, line, package, level)   /* used by wpdebug.c */
	int             level;      /* IN: how many frames to pop up before
								 * reporting */
	char           *funcname;   /* OUT: Name of function stoped in */
	int            *line;       /* OUT: Line of source */
	char           *package;    /* OUT: Source */
{
	elf_object    *obj;

	return (ElfWhereBreakFrame(funcname, line, package, &obj, &obj, level));
}

/*************************************************************************
function: elfDbgHitBreak - Hit a breakpoint, report info
*/
static
elfDbgHitBreak (type, packP, line, inObj, dbginfoP)
int type;       /* ELFBRK_BRK, ELFBRK_STEP, ELFBRK_NEXT, ELFBRK_ERROR, ... */
package_info *packP;
int line;
elf_object *inObj;      /* Object breakpointd in or NULL */
debugger_info *dbginfoP;
{
#if !FOR_TESTER
	int             pid, ix, code;
	int             msgcount = 0;
	int             cmd;
	char            funcname[ELFIDSTR + 1];
	char           *msg;
	int             size;
	int				rsp_size;
	data_ptr        aP;
	stack_frame    *fp;
	int				response_size;
	char			*response_buf;
	int 			rsp_appid;

/* forbidden macros - ??? */
	for (fp = Elf.curFrame; fp; fp = fp->caller) {
		FetchNameOfFunction(funcname, fp->funcstarts);
		if (streq(funcname, "MENU_STATE_RQST$") ||
			streq(funcname, "MENU_NAME_RQST$") ||
			streq(funcname, "BUTTON3_MENU_STATE_RQST$")) {
			return;
		}
	}

	Elf.curFrame->pc = Elf.pc;
	FetchNameOfFunction(funcname, Elf.curFrame->funcstarts);

/* if task under debug is crashed, make debugger re-available */
	for (ix = 0; ix < ElfNumDebuggers; ++ix) {
		if (!ThimPidFromUID(ElfDebugger[ix].debuggee_uid)) { /* debugee crashed */
			ElfDebugger[ix].is_stopped = FALSE;
			ElfDebugger[ix].is_nexting = FALSE;
			ElfDebugger[ix].is_stepping = FALSE;
			ElfDebugger[ix].breakFrame = NULL;
			ElfDebugger[ix].breakGlobals = NULL;
			ElfDebugger[ix].exbrk.packP = NULL;
		}
	}

/* find the debugger */
	if (type == ELFBRK_ERROR) {
		/* find first available debugger with error catching enabled */
		dbginfoP = NULL;
		for (ix = 0; ix < ElfNumDebuggers; ++ix) {
			if (ElfDebugger[ix].dbg_uid && ElfDebugger[ix].is_error_break
				&& !ElfDebugger[ix].is_stopped) {
				dbginfoP = &ElfDebugger[ix];
				break;
			}
		}
	}

	if (!dbginfoP)		/* no debuggers found to catch errors */
		return;

	if (dbginfoP->is_stopped)
		return;			/* should never happen, sanity check */
	
	pid = ThimPidFromUID(dbginfoP->dbg_uid);
	if (!pid) {			/* debugger crashed */
		elfDbgCloseDebugger(dbginfoP,FALSE);
		return;
	}

	dbginfoP->breakFrame = Elf.curFrame;
	dbginfoP->breakGlobals = Elf.taskFrame;
	dbginfoP->is_stepping = FALSE;
	dbginfoP->is_nexting = FALSE;
	dbginfoP->debuggee_uid = TaskInfo[AxInTask].uid;

	dbginfoP->is_stopped = TRUE;

	aP = ELF_CREATE_ARRAY(6);
	aP = ElfAddNumToArray(aP, 0, type);
	aP = ElfAddNumToArray(aP, 1, line);
	aP = ElfAddNumToArray(aP, 2, TaskInfo[AxInTask].uid);
	aP = ElfAddStrToArray(aP, 3, packP->source_name);
	aP = ElfAddStrToArray(aP, 4, funcname);
	if (inObj)
		aP = WriteArray(aP, 5, ELF_CREATE_OBJECT(inObj->dex, inObj->uid, NULL));

	cmd = POKE_DEBUGGER;
	size = AxMemWrite(aP, sizeof(int), &msg);	/* pack into a single buffer */
	memcpy(msg, &cmd, sizeof(int));
	AxSendMes(pid, msg, size, APELFDEBUG);
	THIMfree(msg);
	ELF_DATA_FREE(aP);

/* ready: */
	ACImOk();	/* Okay debugger task, if not alreay okayed */
	size = sizeof(cmd);			/* Expected size of response */
	while (TRUE) {
		/* 
		 * Accept ANY response (potentially of unexpected size), but
		 *	discard all those except those from ME debugger 
		 */


		/* The following fcn waits for a response, setting the
			task's ipc_response_waiting flag till it arrives */
		response_size = IPCresponseSizeAndSender(pid, &rsp_appid);
		response_buf = ElfMalloc(response_size);

		/* The following fcn clears the response and clears the
			task's ipc_response_waiting flag */
		code = AxGetResponse(pid, response_buf, response_size);

		if (rsp_appid == APELFDEBUG) {
			if ((response_size == size) && (code == IPC_OK)) 
				memcpy( &cmd, response_buf, size);
			else
				cmd = ELFBRK_QUIT;
			break;
		}
	}

	dbginfoP->is_stopped = FALSE;
	dbginfoP->breakFrame = NULL;
	dbginfoP->breakGlobals = NULL;
	dbginfoP->debuggee_uid = TaskInfo[AxInTask].uid;

	switch (cmd) {
	case ELFBRK_QUIT:
		exit(0);

	case ELFBRK_CONTINUE:
		return;

	case ELFBRK_STEP:
		dbginfoP->is_stepping = TRUE;
		return;

	case ELFBRK_NEXT:
		dbginfoP->is_nexting = TRUE;
		dbginfoP->nexterFrame = Elf.curFrame;
		return;
	}
#endif
}

static ax_mem_block_info_t ElfMemData[2];
ax_mem_block_info_t *
ElfMemBlockInfo()
{
	ElfMemData[0].addr = (char *) &Elf;
	ElfMemData[0].size = sizeof(Elf);
	ElfMemData[0].is_zeros = TRUE;
	return (ElfMemData);
}

/*************************************************************************
function: hash - Comput a string's hash value

From Donald E. Knuth's (The Art of Programming) Searching & Sorting Algorithms
Uses only additions and subtractions - babu 12/5/94
*/
int 
hash (s)
	register char *s;
{
	register int sum = 0;
	register int chr = 0;

	while (chr = *s++) {
		sum = sum + sum + (chr & 0xFF);
		while (sum >= ELF_HASH_PRIME)
			sum -= ELF_HASH_PRIME;
	}
	return(sum);
}

/*************************************************************************
function: streq - Compare two strings, return TRUE if equal
*/
streq(s1, s2)
	register char  *s1;
	register char  *s2;
{
	for (; *s1 == *s2; s1++, s2++)
		if (*s1 == '\0')
			return (TRUE);
	return (FALSE);
}

/*************************************************************************
function: streqi - Compare two strings case insensitive, return TRUE if equal
*/
streqi(s1, s2)
	register char  *s1;
	register char  *s2;
{
	for (; tolower(*s1) == tolower(*s2); s1++, s2++)
		if (*s1 == '\0')
			return (TRUE);
	return (FALSE);
}

axglobal elf_str *ElfStrings;   /* Head of user "getStr" string  objects list */

/*************************************************************************
function: StrGet - Low level Interface to getStr
*/
typeErr 
StrGet(n, p)
	int             n;
	char           *p;
{
	elf_str        *ep;

	for (ep = ElfStrings; ep; ep = ep->next) {
		if (ep->num == n) {
			strcpy(p, ep->text);
			return (AOK);
		}
	}
	return (getStr(n, p));
}

#define SQS_COLS 0
#define SQS_TABS 1

/*************************************************************************
function: ElfSQL - Process a SQL related instruction
This returns the number of words after the OP_SQL token that it absorbes.
*/
int 
ElfSQL(pc, subcode)
	shrt           *pc;         /* Elf.pc, points to word after the subcode */
	shrt            subcode;    /* The word that follows the OP_SQL token */
{
	int             absorbedWords = 0;
	data_ptr        aP;
	data_ptr        ElfSQLStructToStr();
	data_ptr        ElfSQLExec();
	int             locid, varDex, hashval;
	int             dex;
	data_ptr        bindInfo;
	char           *funcName;
	int             code;
	char            str[1001];
	char            obj[1001];
	int             count, ix;
	data_ptr        bindValues, stmtId, /* channel, stmtNum */
					cursorId;   /* channel, cursorNum */

	switch (subcode) {

	case SQLOP_START:           /* beginning of SQL statement */
		ELF_DATA_FREE(Elf.sqlStruct);
		Elf.sqlStruct = ELF_CREATE_ARRAY(0);
		break;

	case SQLOP_TOKEN:
		dex = AX(Elf.sqlStruct)->size;
		aP = PopDatum();
/* printf("tok: %s\n",StrFromDataPtr(aP)); */
		if (!aP)
			Elf.sqlStruct = ElfAddStrToArray(Elf.sqlStruct, dex, NXLT("null"));
		else
			Elf.sqlStruct = ElfSQLAddToken(Elf.sqlStruct, aP, 0);
		break;

	case SQLOP_DATUM:
		dex = AX(Elf.sqlStruct)->size;
		aP = PopDatum();
/* printf("dtm: %s\n",StrFromDataPtr(aP)); */
		if (!aP)
			Elf.sqlStruct = ElfAddStrToArray(Elf.sqlStruct, dex, NXLT("null"));
		else
			Elf.sqlStruct = ElfSQLAddDatum(Elf.sqlStruct, aP, 0);
		break;

	case SQLOP_REPL_TOKEN:      /* for compatibilty */
		subcode = SQLOP_EA_REPL_TOKEN;
	case SQLOP_EA_REPL_TOKEN:
	case SQLOP_EA_REPL_TOKEN_ARG:
	case SQLOP_EA_REPL_TOKEN_GLO:
	case SQLOP_EA_REPL_TOKEN_EXT:
		bindInfo = ELF_CREATE_ARRAY(2);
		WriteArray(bindInfo, 0, MK_INTEGER(subcode));
		varDex = *pc++;
		absorbedWords++;
		WriteArray(bindInfo, 1, MK_INTEGER(varDex));
/* printf("repled sql var %d\n",varDex); */
/* printf("val '%s'\n",StrFromDataPtr(Elf.curFrame->locvars[varDex])); */
		dex = AX(Elf.sqlStruct)->size;
		Elf.sqlStruct = WriteArray(Elf.sqlStruct, dex, bindInfo);
		break;
	case SQLOP_EA_REPL_TOKEN_OBJ:
		bindInfo = ELF_CREATE_ARRAY(3);
		WriteArray(bindInfo, 0, MK_INTEGER(subcode));
		varDex = *pc++;
		absorbedWords++;
		WriteArray(bindInfo, 1, MK_INTEGER(varDex));
		hashval = *pc++;
		absorbedWords++;
		WriteArray(bindInfo, 2, MK_INTEGER(hashval));
		dex = AX(Elf.sqlStruct)->size;
		Elf.sqlStruct = WriteArray(Elf.sqlStruct, dex, bindInfo);
		break;

	case SQLOP_REPL_DATUM:      /* for compatibilty */
		subcode = SQLOP_EA_REPL_DATUM;
	case SQLOP_EA_REPL_DATUM:
	case SQLOP_EA_REPL_DATUM_ARG:
	case SQLOP_EA_REPL_DATUM_GLO:
	case SQLOP_EA_REPL_DATUM_EXT:
		bindInfo = ELF_CREATE_ARRAY(2);
		WriteArray(bindInfo, 0, MK_INTEGER(subcode));
		varDex = *pc++;
		absorbedWords++;
		WriteArray(bindInfo, 1, MK_INTEGER(varDex));
/* printf("repled datum var %d\n",varDex); */
/* printf("val '%s'\n",StrFromDataPtr(Elf.curFrame->locvars[varDex])); */
		dex = AX(Elf.sqlStruct)->size;
		Elf.sqlStruct = WriteArray(Elf.sqlStruct, dex, bindInfo);
		break;
	case SQLOP_EA_REPL_DATUM_OBJ:
		bindInfo = ELF_CREATE_ARRAY(3);
		WriteArray(bindInfo, 0, MK_INTEGER(subcode));
		varDex = *pc++;
		absorbedWords++;
		WriteArray(bindInfo, 1, MK_INTEGER(varDex));
		hashval = *pc++;
		absorbedWords++;
		WriteArray(bindInfo, 2, MK_INTEGER(hashval));
		dex = AX(Elf.sqlStruct)->size;
		Elf.sqlStruct = WriteArray(Elf.sqlStruct, dex, bindInfo);
		break;

	case SQLOP_BIND_TOKEN:
		bindInfo = ELF_CREATE_ARRAY(1);
		WriteArray(bindInfo, 0, MK_INTEGER(3));
/* printf("binded sql var\n"); */
		dex = AX(Elf.sqlStruct)->size;
		Elf.sqlStruct = WriteArray(Elf.sqlStruct, dex, bindInfo);
		break;

	case SQLOP_BIND_DATUM:
		bindInfo = ELF_CREATE_ARRAY(1);
		WriteArray(bindInfo, 0, MK_INTEGER(4));
/* printf("binded datum var\n"); */
		dex = AX(Elf.sqlStruct)->size;
		Elf.sqlStruct = WriteArray(Elf.sqlStruct, dex, bindInfo);
		break;

	case SQLOP_EXECUTE:
		stmtId = POP_VAL;
		count = *pc;
		absorbedWords++;
		bindValues = ELF_CREATE_ARRAY(count);
/* printf("%d binding values\n"); */
		for (ix = count - 1; ix >= 0; ix--)
			WriteArray(bindValues, ix, ElfDataCopy(PopDatum()));

		code = ElfCallByName(NXLT("sql_execute_stmt@"), &aP, 3,
							 ReadArray(AX(stmtId), 0),      /* channel */
							 ReadArray(AX(stmtId), 1),      /* stmt */
							 bindValues);
		ELF_DATA_FREE(bindValues);
		if (code != 0) {
			strncpy(str, StrFromArray(aP, 0), 1000);
			str[1000] = 0;
			strncpy(obj, StrFromArray(aP, 1), 1000);
			obj[1000] = 0;
			ELF_DATA_FREE(aP);
			ElfStrAbort(code, str, obj);
		}
		ELF_DATA_FREE(aP);
		break;

	case SQLOP_DCLCURSOR:
		stmtId = POP_VAL;
		code = ElfCallByName(NXLT("sql_declare_cursor@"), &aP, 2,
							 ReadArray(AX(stmtId), 0),      /* channel */
							 ReadArray(AX(stmtId), 1));     /* stmt */
		if (code != 0) {
			strncpy(str, StrFromArray(aP, 0), 1000);
			str[1000] = 0;
			strncpy(obj, StrFromArray(aP, 1), 1000);
			obj[1000] = 0;
			ELF_DATA_FREE(aP);
			ElfStrAbort(code, str, obj);
		}
		locid = *pc;
		absorbedWords++;
		cursorId = MkArray(2);
		WriteArray(cursorId, 0, ElfDataCopy(ReadArray(AX(stmtId), 0)));
		WriteArray(cursorId, 1, aP);
		Assign(PushPtr(&Elf.curFrame->locvars[locid]), cursorId);
		break;

	case SQLOP_UNPREPARE:
		stmtId = POP_VAL;
		code = ElfCallByName(NXLT("sql_unprepare@"), &aP, 2,
							 ReadArray(AX(stmtId), 0),      /* channel */
							 ReadArray(AX(stmtId), 1));     /* stmt */
		if (code != 0) {
			strncpy(str, StrFromArray(aP, 0), 1000);
			str[1000] = 0;
			strncpy(obj, StrFromArray(aP, 1), 1000);
			obj[1000] = 0;
			ELF_DATA_FREE(aP);
			ElfStrAbort(code, str, obj);
		}
		break;

	case SQLOP_OPENCURSOR:
		cursorId = POP_VAL;
		count = *pc;
		absorbedWords++;
		bindValues = ELF_CREATE_ARRAY(count);
/* printf("%d binding values\n"); */
		for (ix = count - 1; ix >= 0; ix--)
			WriteArray(bindValues, ix, ElfDataCopy(PopDatum()));

		code = ElfCallByName(NXLT("sql_open_cursor@"), &aP, 3,
							 ReadArray(AX(cursorId), 0),    /* channel */
							 ReadArray(AX(cursorId), 1),    /* cursorNum */
							 bindValues);
		ELF_DATA_FREE(bindValues);
		if (code != 0) {
			strncpy(str, StrFromArray(aP, 0), 1000);
			str[1000] = 0;
			strncpy(obj, StrFromArray(aP, 1), 1000);
			obj[1000] = 0;
			ELF_DATA_FREE(aP);
			ElfStrAbort(code, str, obj);
		}
		break;

	case SQLOP_CLOSECURSOR:
		cursorId = POP_VAL;
		code = ElfCallByName(NXLT("sql_close_cursor@"), &aP, 2,
							 ReadArray(AX(cursorId), 0),    /* channel */
							 ReadArray(AX(cursorId), 1));   /* cursorNum */
		if (code != 0) {
			strncpy(str, StrFromArray(aP, 0), 1000);
			str[1000] = 0;
			strncpy(obj, StrFromArray(aP, 1), 1000);
			obj[1000] = 0;
			ELF_DATA_FREE(aP);
			ElfStrAbort(code, str, obj);
		}
		break;

	case SQLOP_FETCHCURSOR:
		cursorId = POP_VAL;
		code = ElfCallByName(NXLT("sql_fetch_cursor@"), &aP, 2,
							 ReadArray(AX(cursorId), 0),    /* channel */
							 ReadArray(AX(cursorId), 1));   /* cursorNum */
		if (code != 0) {
			strncpy(str, StrFromArray(aP, 0), 1000);
			str[1000] = 0;
			strncpy(obj, StrFromArray(aP, 1), 1000);
			obj[1000] = 0;
			ELF_DATA_FREE(aP);
			ElfStrAbort(code, str, obj);
		}
		locid = *pc;
		absorbedWords++;
		Assign(PushPtr(&Elf.curFrame->locvars[locid]), aP);
		break;

	case SQLOP_ARRAYFETCHCURSOR:
		count = PopIntVal();
		cursorId = POP_VAL;
		code = ElfCallByName(NXLT("sql_fetch_cursor@"), &aP, 3,
							 ReadArray(AX(cursorId), 0),    /* channel */
							 ReadArray(AX(cursorId), 1),    /* cursorNum */
							 MK_INTEGER(count));
		if (code != 0) {
			strncpy(str, StrFromArray(aP, 0), 1000);
			str[1000] = 0;
			strncpy(obj, StrFromArray(aP, 1), 1000);
			obj[1000] = 0;
			ELF_DATA_FREE(aP);
			ElfStrAbort(code, str, obj);
		}
		locid = *pc;
		absorbedWords++;
		Assign(PushPtr(&Elf.curFrame->locvars[locid]), aP);
		break;

	case SQLOP_PUSH:
		funcName = PopStr();
		aP = ElfSQLExec(funcName, Elf.sqlStruct);
		locid = *pc;
		absorbedWords++;
		Assign(PushPtr(&Elf.curFrame->locvars[locid]), aP);
		break;

	case SQLOP_EXEC:
		funcName = PopStr();
		aP = ElfSQLExec(funcName, Elf.sqlStruct);
		break;

	default:
		abort();
	}
	return (absorbedWords);
}

/*************************************************************************
function: ElfSQLAddToken - Add an unquoted datum to the aP
*/
data_ptr 
ElfSQLAddToken(aP, dP, level)
	data_ptr        aP;
	data_ptr        dP;
	int             level;
{
	int             dex = AX(aP)->size;
	int             ix;

	if (!dP)
		aP = ElfAddStrToArray(aP, dex, NXLT("null"));
	else if (IS_INTEGER(dP))
		aP = WriteArray(aP, dex, dP);
	else if ((dP->type == TYPE_ARRAY) && (level == 0)) {
		for (ix = 0; ix < AX(dP)->size; ix++) {
			aP = ElfSQLAddToken(aP, AX(dP)->vals[ix], level + 1);
			if (ix != AX(dP)->size - 1) {
				dex = AX(aP)->size;
				aP = ElfAddStrToArray(aP, dex++, " , ");
			}
		}
	} else if (dP->type == TYPE_STRING)
		aP = ElfAddStrToArray(aP, dex, SX(dP)->text);
	else if (dP->type == TYPE_NUMBER) {
		aP = WriteArray(aP, dex, ElfDataCopy(dP));
	} else
		ElfAbort(ERR_ILLVAL);   /* !!0 better code */
	return (aP);
}

/*************************************************************************
function: ElfSQLAddDatum - Add a datum to the aP
*/
data_ptr 
ElfSQLAddDatum(aP, dP, level)
	data_ptr        aP;
	data_ptr        dP;
	int             level;
{
	int             dex = AX(aP)->size;
	int             ix;

	if (!dP)
		aP = ElfAddStrToArray(aP, dex, NXLT("null"));
	else if (IS_INTEGER(dP))
		aP = WriteArray(aP, dex, dP);
	else if ((dP->type == TYPE_ARRAY) && (level == 0)) {
		for (ix = 0; ix < AX(dP)->size; ix++) {
			aP = ElfSQLAddDatum(aP, AX(dP)->vals[ix], level + 1);
			if (ix != AX(dP)->size - 1) {
				dex = AX(aP)->size;
				aP = ElfAddStrToArray(aP, dex++, " , ");
			}
		}
	} else if (dP->type == TYPE_STRING) {
		char           *p;
		char           *buf;

		p = SX(dP)->text;
		buf = (char*)malloc(strlen(p) + 3);
		buf[0] = '\'';
		strcpy(buf + 1, p);
		strcat(buf, "'");
		aP = ElfAddStrToArray(aP, dex, buf);
		free(buf);
	} else if (dP->type == TYPE_NUMBER) {
		aP = WriteArray(aP, dex, ElfDataCopy(dP));
	} else
		ElfAbort(ERR_ILLVAL);   /* !!0 better code */
	return (aP);
}

/*************************************************************************
function: ElfSQLExec - Run SQL
!!0 Should use entry var for call!
!!1 Call should be in C
The return data is currently an array-2 of headers and data.
This routine gets rid of the headers and just returns the data.
*/
data_ptr 
ElfSQLExec(funcName, aP)
	char           *funcName;
	data_ptr        aP;
{
	data_ptr        retData;
	int             code;
	char            str[1001];
	char            obj[1001];
	data_ptr        bindInfo;
	data_ptr        execP, dP, xP;
	int             word2, ix, cnt;
	int             bindCode;
	char           *name;
	int             hashval;

/* copy the data structure */

	cnt = ElfbArraySize(aP);
	execP = ELF_CREATE_ARRAY(0);
	for (ix = 0; ix < cnt; ix++) {
		dP = AX(aP)->vals[ix];
		if (IS_INTEGER(dP))
			execP = WriteArray(execP, AX(execP)->size, dP);
		else if (AX(aP)->vals[ix]->type == TYPE_ARRAY) {
			bindInfo = AX(aP)->vals[ix];
			bindCode = IntFromDataPtr(AX(bindInfo)->vals[0]);

			switch (bindCode) {

			case SQLOP_EA_REPL_TOKEN:   /* repl as sql token */
				word2 = IntFromDataPtr(AX(bindInfo)->vals[1]);
				xP = Elf.curFrame->locvars[word2];
				execP = ElfSQLAddToken(execP, ElfDataCopy(xP), 0);
/* printf("unrepled token var %d\n",word2); */
				break;
			case SQLOP_EA_REPL_TOKEN_ARG:       /* repl as sql token */
				word2 = IntFromDataPtr(AX(bindInfo)->vals[1]);
				xP = *((data_ptr *) UNFAULT(Elf.curFrame->argvars[word2]));
				execP = ElfSQLAddToken(execP, ElfDataCopy(xP), 0);
				break;
			case SQLOP_EA_REPL_TOKEN_GLO:       /* repl as sql token */
				word2 = IntFromDataPtr(AX(bindInfo)->vals[1]);
				xP = Elf.taskFrame[GLOVARDEX];
				execP = ElfSQLAddToken(execP, ElfDataCopy(xP), 0);
				break;
			case SQLOP_EA_REPL_TOKEN_EXT:       /* repl as sql token */
				word2 = IntFromDataPtr(AX(bindInfo)->vals[1]);
				xP = *(Elf.taskFrame + EXTDEX);
				execP = ElfSQLAddToken(execP, ElfDataCopy(xP), 0);
				break;
			case SQLOP_EA_REPL_TOKEN_OBJ:       /* repl as sql token */
				word2 = IntFromDataPtr(AX(bindInfo)->vals[1]);
				name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
				hashval = IntFromDataPtr(AX(bindInfo)->vals[2]);
				xP = *ElfGetObjData(Elf.curFrame->inObj, name, hashval);
				execP = ElfSQLAddToken(execP, ElfDataCopy(xP), 0);
				break;

			case SQLOP_EA_REPL_DATUM:   /* repl as datum */
				word2 = IntFromDataPtr(AX(bindInfo)->vals[1]);
				xP = Elf.curFrame->locvars[word2];
				execP = ElfSQLAddDatum(execP, ElfDataCopy(xP), 0);
/* printf("unrepled datum var %d\n",word2); */
				break;
			case SQLOP_EA_REPL_DATUM_ARG:       /* repl as datum */
				word2 = IntFromDataPtr(AX(bindInfo)->vals[1]);
				xP = *((data_ptr *) UNFAULT(Elf.curFrame->argvars[word2]));
				execP = ElfSQLAddDatum(execP, ElfDataCopy(xP), 0);
				break;
			case SQLOP_EA_REPL_DATUM_GLO:       /* repl as sql datum */
				word2 = IntFromDataPtr(AX(bindInfo)->vals[1]);
				xP = Elf.taskFrame[GLOVARDEX];
				execP = ElfSQLAddDatum(execP, ElfDataCopy(xP), 0);
				break;
			case SQLOP_EA_REPL_DATUM_EXT:       /* repl as sql datum */
				word2 = IntFromDataPtr(AX(bindInfo)->vals[1]);
				xP = *(Elf.taskFrame + EXTDEX);
				execP = ElfSQLAddDatum(execP, ElfDataCopy(xP), 0);
				break;
			case SQLOP_EA_REPL_DATUM_OBJ:       /* repl as sql datum */
				word2 = IntFromDataPtr(AX(bindInfo)->vals[1]);
				name = StrFromDataPtr(Elf.curFrame->packP->literals_table[word2]);
				hashval = IntFromDataPtr(AX(bindInfo)->vals[2]);
				xP = *ElfGetObjData(Elf.curFrame->inObj, name, hashval);
				execP = ElfSQLAddDatum(execP, ElfDataCopy(xP), 0);
				break;

			case 3:             /* bind as sql token */
			case 4:             /* bind as datum */
/* printf("transmitted placeholder\n"); */
				execP = WriteArray(execP, AX(execP)->size, ElfDataCopy(bindInfo));
				break;
			}
		} else {
			dP = ElfDataCopy(dP);
			execP = WriteArray(execP, AX(execP)->size, dP);
		}
	}

	code = ElfCallByName(funcName, &retData, 1, execP);
	ELF_DATA_FREE(execP);
	if (code != 0) {
		strncpy(str, StrFromArray(retData, 0), 1000);
		str[1000] = 0;
		strncpy(obj, StrFromArray(retData, 1), 1000);
		obj[1000] = 0;
		ELF_DATA_FREE(retData);
		ElfStrAbort(code, str, obj);
	}
	return (retData);
}

typedef struct name_entry {
	struct name_entry *next;
	char           *name;
}               name_entry;

/*************************************************************************
function: elfHashName - Given name, return system's one & only copy of name
*/
char           *
elfHashName(name)
	char           *name;
{
	int             hashval;
	char           *elfHashNameAndVal();

	return (elfHashNameAndVal(name, &hashval));
}

char           *
elfHashNameAndVal(name, hashvalP)
	char           *name;
	int            *hashvalP;
{
	register name_entry *vP;
	static name_entry **hashTable;      /* ALL WP's share it */

	if (!hashTable)
		hashTable = (name_entry **) TaskCalloc(0, ELF_HASH_PRIME + 1, sizeof(name_entry *));
	for (vP = hashTable[*hashvalP = hash(name)]; vP; vP = vP->next) {
		if (streq(name, vP->name))
			return (vP->name);
	}
	vP = (name_entry *) TaskAlloc(0, sizeof(name_entry));
	vP->name = strcpy((char *) TaskAlloc(0, strlen(name) + 1), name);
	vP->next = (name_entry *) hashTable[*hashvalP];
	hashTable[*hashvalP] = vP;
	return (vP->name);
}

#if 0
int
print_hash_stats()
{
	register int ix, hits;
	register name_entry *vP;
	register int hit_counts[20];
	register int more20s;
	register int total_entries;
	
	if (!hashTable)
		return;

	for (ix = 0; ix < 20; ++ix) {
		hit_counts[ix] = 0;
		more20s = 0;
	}
	
	total_entries = 0;
	for (ix = 0; ix < ELF_HASH_PRIME+1; ++ix) {
		hits = 0;
		for (vP = hashTable[ix]; vP; vP = vP->next)
			hits++;
		if (hits < 20)
			hit_counts[hits]++;
		else
			more20s++;
		total_entries += hits;
	}

	
	for (ix = 20-1; ix >= 0; --ix)
		printf("%d hits = %d\n", ix, hit_counts[ix]);
	printf(">20 hits = %d\n", more20s);

	printf("Total Entries = %d\n", total_entries);

	return(total_entries);
}
#endif

axElfGetSourceName(macroName, sourceName)
	char *macroName;
	char *sourceName;	/* caller should allocate of size STRPATH+1 */
{
	ext_link		*eP;
	int len,type;
	data_user_func *funcP;
	data_ptr retData, nameP;
	char *p;
	typeErr err;

	*sourceName = '\0';
	eP = ElfFindLinkForByNameCalls(macroName);
	if(NULL == eP)
		return(TYPE_UNKNOWN);
	if(NULL == Snapper(0, eP, NULL, FALSE))	/* just check, no auto-install */
		return(TYPE_UNKNOWN);
	type = *(eP->value.ptr);
	switch (type) {
	case TYPE_QFUNC:
	case TYPE_MACRO:
	case TYPE_UIFUNC:
	case TYPE_NRFUNC:
	case TYPE_SETOBJ:
	case TYPE_GETOBJ:
		strncpy(sourceName, eP->packP->source_name, STRPATH);
		len = MIN(strlen(eP->packP->source_name), STRPATH);
		sourceName[len] = '\0';
		break;
	case TYPE_C_FUNC:
	case TYPE_C_MACRO:
		strcpy(sourceName, XLT("ApplixWare builtin function"));
		break;
	case TYPE_U_FUNC:
		funcP = (data_user_func *)eP->value.ptr;
		strncpy(sourceName, funcP->libname, STRPATH);
		len = MIN(strlen(funcP->libname), STRPATH);
		sourceName[len] = '\0';
		break;
    case TYPE_R_FUNC:
        funcP = (data_user_func *)eP->value.ptr;
        nameP = ELF_CREATE_STRING(funcP->libname);
        err = ElfCallByName(NXLT("rpc_prog_name_from_channel$"), &retData,
            1, nameP);
        ELF_DATA_FREE(nameP);
        if(err == AOK) {
            p = StrFromDataPtr(retData);
            strncpy(sourceName, p, STRPATH);
            len = MIN(strlen(p), STRPATH);
            sourceName[len] = '\0';
        }
        ELF_DATA_FREE(retData);
        break;
    }
	return(type);
}




/* 
 *  Called by ElfInstallModule(). Called once so that NDummyPackages can 
 *  be set. This value is decremented every time a dummyPackage is unloaded 
 *  in ElfUnloadDummyPackagesForShlib.
 */
int AxGetNumDummyPackages(void)
{
	int i;
	
	NDummyPackages = 0;
	for(i = 0; i < NextPackageInd; i++) {
		if(Packages[i] != NULL && Packages[i]->code_starts == NULL)
			NDummyPackages++;
	}
	return(NDummyPackages);
}


/**
  * function: ElfInstallModule --
  *  Install the given Elf packages, by adding at the end of 
  *  already installed  packages, but before the dummy packages,
  *  so that installed packages are searched before dummy packages.
  *  For dummy packages code_starts is NULL
  *
  * Inputs:
  *   num_packages -- No. of packages to be installed
  *   mod_packages -- Array of packages that are to be installed
  *   
  * Input/Output's:    
  *   NextPackageInd  -- static -- Index where next package can be installed
  *   MaxPackages     -- static -- No. of  packages for which memory is already allocated
  *
  * Outputs:
  *   Packages        -- static -- Array of Installed packages
**/
int
ElfInstallModule(num_packages, mod_packages)
int num_packages;
package_info** mod_packages;
{
   gvar_info      *sP;
   data_ptr dP;
   data_ptr *litsP;
   int i,px,ix,vx,insert_pos;
   
   if(NextPackageInd+num_packages >= MaxPackages-1) {
	  MaxPackages = NextPackageInd+num_packages+20;
	  Packages = (package_info **) TaskRealloc(0, Packages, 
											   sizeof(package_info *) * MaxPackages);
   }

   for(px=0; px < NextPackageInd; px++)
	   if(Packages[px] != NULL && Packages[px]->code_starts == NULL)
		   /* start of dummy packages */
		   break; 
   insert_pos = px;

   /* Start off by computing the existing dummy packages */
   if ( !NDummyPackages )
	   AxGetNumDummyPackages();

   /* move dummy packages to the end */
   for ( i=0; i<NDummyPackages; i++ ) {
	   Packages[NextPackageInd+num_packages-i-1] = 
		   Packages[NextPackageInd-i-1];
   }

   /* insert new packages at insert_pos */
   for(px=insert_pos,ix=0; ix < num_packages && mod_packages[ix]; px++,ix++) {
      Packages[px] = mod_packages[ix];	   
	  for(vx = 0; sP = mod_packages[ix]->gvar_list[vx]; vx++)
		 AllocateGvar(sP);
	if (mod_packages[ix]->literals_table != NULL) {
		litsP = mod_packages[ix]->literals_table;
		for(vx = 1; (dP = UNFAULT(litsP[vx])) != NULL; vx++) {
			if ((*(int *)dP) == TYPE_INTEGER)
				litsP[vx] = (data_ptr)(*((int *)dP+1));	
		}
	}
   }
   NextPackageInd += num_packages;
   Packages[NextPackageInd] = NULL;
   return(0);
}

typedef struct sysvars_entry {
	struct sysvars_entry *next;
	char *name;
	data_ptr value;
} sysvars_entry;
static sysvars_entry *VarsTab[ELF_HASH_PRIME+1];

/*************************************************************************
function: ElfGetSysVar - a C version which returns the value of a sysvar
*/
data_ptr
ElfGetSysVar (name)
register char *name;
{
	register sysvars_entry *vP;

	for ( vP = VarsTab[hash(name)]; vP; vP = vP->next ) {
		if ( streq(name,vP->name) ) {
			return(vP->value);
		}
	}
	return(NULL);
}

/*************************************************************************
function: ElfGetSysVarPP - returns address of ptr to a sysvar
*/
data_ptr *
ElfGetSysVarPP (name,hashval)
register char *name;
register int hashval;
{
	register sysvars_entry *vP;

	for ( vP = VarsTab[hashval]; vP; vP = vP->next ) {
		if ( streq(name,vP->name) ) {
			return(&vP->value);
		}
	}
	vP = (sysvars_entry *)TaskAlloc(0,sizeof(sysvars_entry));
	vP->name = strcpy(TaskAlloc(0,strlen(name)+1),name);
	vP->next = VarsTab[hashval];
	VarsTab[hashval] = vP;
	vP->value = NULL;
	return(&vP->value);
}

/*************************************************************************
function: ElfbSetSysVar - Set an environment (system) var
*/
ElfbSetSysVar (name,value)
register char *name;
register data_ptr value;
{
	register sysvars_entry *vP;
	register int hashval;
	register int savedContext = THIMpid;

	for ( vP = VarsTab[hashval = hash(name)]; vP; vP = vP->next ) {
		if ( streq(name,vP->name) ) {		/* re-use existing slot */
			AxTaskFreeElfData(0,vP->value);
			goto wrap_up;
		}
	}
	vP = (sysvars_entry *)TaskAlloc(0,sizeof(sysvars_entry));
	vP->name = strcpy(TaskAlloc(0,strlen(name)+1),name);
	vP->next = VarsTab[hashval];
	VarsTab[hashval] = vP;
wrap_up:
	vP->value = AxTaskCopyElfData(0,value);
}

/*************************************************************************
function: ElfbGetSysVar - Get the value of a system var
*/
ElfbGetSysVar (name)
register char *name;
{
	register sysvars_entry *vP;

	for ( vP = VarsTab[hash(name)]; vP; vP = vP->next ) {
		if ( streq(name,vP->name) ) {
			ElfRetData = ElfDataCopy(vP->value);
			return;
		}
	}
	ElfRetData = NULL;
}

/* stubs so elfc can be built without elstatic.c */
#if FOR_TESTER
bltin_desc ElfCoreFuncs[2];
ElfNumBuiltins (){}
void ElfProfileNote(stack_frame *fp){}
void AxTaskFreeArrayButNotContents(int pid, data_ptr p){}
void ElfProfileExit(stack_frame *fp){}
#endif
