/* Copyright (C) 1983-1999 Applix, Inc.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
   
   Applix Inc. is the original author of this material.  Please send
   generally useful enhancements to Applix in the form of all modified 
   source files including an explanation of the changes.  For more
   information see our website at www.applixware.org.
   
   Do not remove or modify the license information above or the following
   Applix source control information from this file.
   
   Applix filename: pf03prim.c
   Applix ASC version: 111
   Date: Mon Dec 21 21:12:22 1998
*/
#include <stdio.h>

#if AX_X11  /* see comment below */
#if !M_88K
#include <varargs.h>
#endif
#endif

#include    <sys/stat.h>
#include    <axsocket.h>
#include    <math.h>
#include    <aos.h>
#include    <pf.h>
#include    <paratm.h>
#include    <axtext.h>
#include    <devsetup.h>
#include    <pfint.h>
#include    <getstr2.h>
#include    <errno.h>
#include    <sff.h>
#include    <pfpat.h>
#include    <axprint.h>
#include    <axbitmap.h>
#include    <revstamp.h>
#include    <axhtab.h>
#include    <axgfs.h>
#include    <pfsoft.h>

typedef struct HPGL_font_desc
{
    char symbol_set[8];
    char font_spacing[8];
    char pitch[8];
    char height[8];
    char posture[8];
    char stroke_weight[8];
    char typeface[16];
} HPGL_font_desc;

typedef unsigned short  UI ;
typedef short           SI ;
typedef unsigned char   UB ;
typedef char            SB ;
typedef unsigned char   B ;
typedef unsigned int    ULI ;

typedef struct {
    UI      size ;
    UB      format ;
    UB      type ;
    UB      style_MSB ;
    UB      reserved_1 ;
    UI      baseline ;
    UI      cell_width ;
    UI      cell_height ;
    UB      orientation ;
    B       spacing ;
    UI      symbol_set ;
    UI      pitch ;
    UI      height ;
    UI      x_height ;
    SB      width_type ;
    UB      style_LSB ;
    SB      stroke_weight ;
    UB      typeface_LSB ;
    UB      typeface_MSB ;
    UB      serif_style ;
    UB      quality ;
    SB      placement ;
    SB      underline_position ;
    UB      underline_thickness ;
    UI      text_height ;
    UI      text_width ;
    UI      first_code ;
    UI      Last_code ;
    UB      pitch_extended ;
    UB      height_extended ;
    UI      cap_height ;
    ULI     font_number ;
    char    font_name[16] ;
    UI      X_resolution ;
    UI      Y_resolution ;
    } PCL20_header ;

typedef struct {
    UI      size ;
    UB      format ;
    UB      type ;
    UB      style_MSB ;
    UB      reserved_1 ;
    UI      baseline ;
    UI      cell_width ;
    UI      cell_height ;
    UB      orientation ;
    B       spacing ;
    UI      symbol_set ;
    UI      pitch ;
    UI      height ;
    UI      x_height ;
    SB      width_type ;
    UB      style_LSB ;
    SB      stroke_weight ;
    UB      typeface_LSB ;
    UB      typeface_MSB ;
    UB      serif_style ;
    UB      quality ;
    SB      placement ;
    SB      underline_position ;
    UB      underline_thickness ;
    UI      text_height ;
    UI      text_width ;
    UI      first_code ;
    UI      last_code ;
    UB      pitch_extended ;
    UB      height_extended ;
    UI      cap_height ;
    ULI     font_number ;
    char    font_name[16] ;
    } PCL0_header ;

typedef struct {
    UB      format ;
    B       continuation ;
    UB      size ;
    UB      class ;
    UB      orientation ;
    UB      reserved_1 ;
    SI      left_offset ;
    SI      top_offset ;
    UI      width ;
    UI      height ;
    SI      delta_x ;
    } PCL_char ;

typedef struct {
    UB      format ;
    B       continuation ;
    } PCL_char_cont ;

#include    <pf03prim.p>
 

/* define -------------------------------------------------------------------*/
#define HPGL_PATTERNS       TRUE    
#define ABSOLUTE_POSITIONS  TRUE    
#define USE_DECIPTS         TRUE    
#define USE_PLU_IN_HPGL     TRUE
#define Y_MOVE_GRAPHICS     TRUE

#define MARK_MARGIN         125
#define TEXT_RES            300
#define RESOLUTION          300
#define SCREEN_RESOLUTION   75
#define HARD_HPGL_MARG      75          /* 71 dots for A4... */
#define PLU_TO_INCH         1016
#define MINIMUM_PEN_WIDTH   0.20
#define PFWHITE(ch)         ((int)(ch) == ' ' || (int)(ch) == '\t')
#define COLOR_GRAY_PERCENT(gr) ((gr) * cur_grayness)

#define MAX_HPGL_PAT        8           /* HPGL allows user def rasters 1..8 */
#define IMAGE_LIMIT         131072
#define COLOR_IMAGE_LIMIT   131072/8

#define TWICE_MAX_PATTERN   2048
#define PCL_COMPR_NONE      0
#define PCL_COMPR_RLE       1
#define FONT_ID             13          /* for subsequent Font Management commands */


/* these are tied to axPrintDitherStyle */
#define PF_DITHER_HT4       0
#define PF_DITHER_FS        -1

#define COLOR_MODEL_CMYK    4
#define COLOR_MODEL_RGB     3
#define COLOR_MODEL_GRAY    1

#if 1
#define PF_MEM_CHECK
#else
#define PF_MEM_CHECK        ThimMemCheck()
#endif

/* macros -------------------------------------------------------------------*/
#define ABS(x)              (((x) > 0) ? (x) : -(x))
#define PFPAGETOP_MIL       (PF.pagetop * 1000.0 / RESOLUTION)
#define MIL_TO_DECIPT(mil)  (((mil) * 720.0) / 1000.0)
#if 0
#define MIL_TO_DOT(mil)     (((mil) * (double)RESOLUTION) / 1000.0)
#else
#define MIL_TO_DOT(mil)     ((double)xmiltopix(mil))
#endif
#define MIL_TO_PLU(mil)     (((mil) * (double)PLU_TO_INCH) / 1000.0)
#define MIL_TO_MM(mil)      (((mil) * 25.4) / 1000.0)
#define X_MIL_TO_PLU(mil)   ((((mil) - PF.PfLeftMarginMil) * (double)PLU_TO_INCH) / 1000.0)
#define Y_MIL_TO_PLU(mil)   (((PF.PfPageMilHeight - (mil)) * (double)PLU_TO_INCH) / 1000.0)
#define DOT_TO_DECIPT(dot)  (((dot) * 720.0) / (double)RESOLUTION)
#define DOT_TO_MM(dot)      (((dot) * 25.4) / (double)RESOLUTION)
#define DEG_TO_RAD(degree)  ((degree)*M_PI / 180.0)

#define DOT_TO_PLU(dot)     (((dot) * (double)PLU_TO_INCH) / (double)RESOLUTION)
#define X_DOT_TO_PLU(dot)   ((((dot) - MIL_TO_DOT(PF.PfLeftMarginMil)) * (double)PLU_TO_INCH) / (double)RESOLUTION)
#define Y_DOT_TO_PLU(dot)   (((PF.pagetop - (dot)) * (double)PLU_TO_INCH) / (double)RESOLUTION)
#define Y(y)                (PF.pagetop - (y))
#define min(a, b)			((a) < (b) ? (a) : (b))

/* external data --------------------------------------------------------------*/
extern  short   THIMpid;
extern  int     TargetRes;
extern  int     PfJobToPSPrinter;


/* external procedures --------------------------------------------------------*/
extern  void    pfRasterLineBytewiseCopier();
extern  void    pfRasterLineMasker();
extern  void    pfRasterLineBitwiseCopier();

/* local data -----------------------------------------------------------------*/
static  char    NameOfThisSourceFile[] = __FILE__ ;     /* for the assert code */
static  int     HPGL_pen =          -1;
static  int     HPGL_pen_up =       TRUE;
static  int     HPGL_cfill =        -1;
static  double cur_grayness =       1.0;        /* proportion of BLACK */
static  int     In_HPGL_Mode =      FALSE;
static  int     HPGL_Transparent =  FALSE;
static  int     HPGL_VectorScreen = 0;
static  int     HPGL_FillType =     1;
static  int     HPGL_GrayFillValue = 100;
static  int     HPGL_GrayVectorValue = 100;
static  int     HPGL_LineEnds =     1;
static  int     HPGL_LineType =     0;
static  int     HPGL_LineJoins =    1;
static  int     HPGL_MiterLimit =   5;
static  int     HPGL_UsePolygonMode = FALSE;

static  int     duplex ;
static  int     duplex_page ;
static  int     page_height ;
static  int     page_width ;
static  int     input_tray ;
static  int     output_tray ;
static  int     banner_point_size =	14 ;

static struct {
    double x0, y0, x1, y1 ;
    } HPGL_IW = {0, 0, 0, 0} ;
    
static struct {
    int cyan ;
    int magenta ;
    int yellow ;
    int black ;
    int color ;                 /* TRUE if color has a non-0 cmy component */
    } palette[256] ;

static  int     palette_count ;

static  textAttrs pf03CurFont;

static  int     pf_color_model =    COLOR_MODEL_CMYK;

static  int     numColorComponents = COLOR_MODEL_CMYK;  /* # components in dest image */
static  int     bitsPerColorComponent = 4;          /* # bits per color per pixel in dest */
static  int     bitsPerGrayComponent =  4;          /* # bits per pixel for grayscale image */


static  char    pf03PatBuf[TWICE_MAX_PATTERN];
static  char    *pf03PatScannedPtr;
static  char    *pf03PatInputPtr;

static  int     PCL_compression_mode = PCL_COMPR_RLE;
static  int     pf_dither_style ;
static  int     max_poly_points =       0;
static  int     HPGL_pattern_table[MAX_HPGL_PAT+1];

#define NUM_HP_SYMBOL_SET_STRINGS       41
static  char *D03SymbolSetStrings[NUM_HP_SYMBOL_SET_STRINGS][2] = {
    NXLT("0N"), "14",           /* latin1 */
    NXLT("19M"),"621",          /* symbol */
    NXLT("2N"), "78",           /* latin 2 */
    NXLT("5N"), "174",          /* latin 5 */
    NXLT("1U"), "53",           /* legal */
    NXLT("8M"), "269",          /* math-8 */
    NXLT("8U"), "277",          /* roman-8 */
    NXLT("8K"), "267",          /* kana-8 */
    NXLT("9K"), "299",          /* korean-8 */
    NXLT("6J"), "202",          /* microsoft publishing */
    NXLT("1F"), "38",           /* iso 69 (french) */
    NXLT("1G"), "39",           /* iso 21 (german) */
    NXLT("0I"), "9",            /* iso 15 (italian) */
    NXLT("5M"), "173",          /* ps math */
    NXLT("10U"), "341",         /* pc-8 code page 437 */
    NXLT("12G"), "391",         /* one of the pc-8 sets */
    NXLT("31L"), "1004",        /* pi set #1 */
    NXLT("32L"), "1036",        /* pi set #2 */
    NXLT("33L"), "1068",
    NXLT("34L"), "1100",
    NXLT("35L"), "1132",
    NXLT("36L"), "1164",        /* pi set #6 */
    NXLT("579L"),"18540",       /* wingdings */
    NXLT("12U"), "405",         /* pc-850 multilingual */
    NXLT("0S"), "19",           /* iso 11 (swedish for names) */
    NXLT("10J"), "330",         /* ps text */
    NXLT("1E"), "37",           /* iso 4 (uk) */
    NXLT("13J"), "426",         /* ventura international */
    NXLT("6M"), "205",          /* ventura math */
    NXLT("14J"), "458",         /* ventura us */
    NXLT("9E"), "293",          /* windows 3.1 latin 2 */
    NXLT("9G"), "295",          /* windows 3.1 latin / greek */
    NXLT("9V"), "310",          /* windows 3.1 latin / arabic */
    NXLT("9U"), "209",          /* windows 3.0 latin 1 */
    NXLT("15U"), "501",         /* pi font */
    NXLT("8V"), "278",          /* arabic--8 */
    NXLT("7J"), "234",          /* desktop */
    NXLT("10L"),"332",          /* PS ITC Zapf Dingbats */
    NXLT("11L"),"364",          /* ITC Zapf Dingbats Series 100 */
    NXLT("12L"),"396",          /* ITC Zapf Dingbats Series 200 */
    NXLT("13L"),"428",          /* ITC Zapf Dingbats Series 300 */
};

static PCL0_header PCL_font_header ;

static PCL_char PCLchar ;
static PCL_char_cont PCLcharcont ;
    
static TmRect32 current_clip ;

/******************************************************************************
*
*
******************************************************************************/
static void put(va_alist)
va_dcl
{
    va_list argp;
    char *format;

    va_start(argp);
    format = va_arg(argp, char *);

    vsprintf(pf03PatBuf, format, argp);

    va_end(argp);
    PfOutString(pf03PatBuf);
}

/******************************************************************************
*
*
******************************************************************************/
void PCL_command(va_alist)
va_dcl
{
    va_list argp;
    char *format;

    if (In_HPGL_Mode)
        SwitchTo_PCL() ;

    va_start(argp);
    format = va_arg(argp, char *);

    vsprintf(pf03PatBuf, format, argp);

    va_end(argp);
    PfOutString(NXLT("\033")) ;
    PfOutString(pf03PatBuf);
}


/******************************************************************************
*
*
******************************************************************************/
static void
HPGL_command(va_alist)
va_dcl
{
    va_list argp;
    char *format;

    if (!In_HPGL_Mode)
        SwitchTo_HPGL() ;

    va_start(argp);
    format = va_arg(argp, char *);

    vsprintf(pf03PatBuf, format, argp);

    va_end(argp);
    PfOutString(pf03PatBuf);
}

/******************************************************************************
*
*   pf03InitPatBuf
*
******************************************************************************/
static void pf03InitPatBuf()
{
    pf03PatScannedPtr = pf03PatBuf;
    pf03PatInputPtr = pf03PatBuf;
}


/************************************************************************
 *
 *  StartJob for PCL5 driver
 *
 ************************************************************************/
int D03StartJob()
{
    int pat;
    char *prof_str;

    pfClearPattern() ;

    for (pat=0; pat <= MAX_HPGL_PAT; pat++)
        HPGL_pattern_table[pat] = -1;

    /* we might need the 50% gray fill for gradient of line and chars */
    pattern[getIXfromPercent(500)].needed = TRUE ;

    PF.left = 0;
    PF.right = 0;
    PF.top = 0;
    PF.bottom = 0;

    PF.pagetop = PF.PfPagePixHeight;    /* pagetop is defined in pf70prim.c */
    PF.cur_thick = -1;
    PF.cur_style = TM_SOLID;
    HPGL_pen = -1;
    HPGL_pen_up = TRUE;
    HPGL_Transparent = FALSE;
    HPGL_VectorScreen = (char)0;
    HPGL_FillType = 1;
    HPGL_GrayFillValue = 100;
    HPGL_GrayVectorValue = 100;
    HPGL_LineEnds = 1;
    HPGL_LineType = 0;
    HPGL_LineJoins = 1;
    HPGL_MiterLimit = 5;
    HPGL_UsePolygonMode = FALSE;

    duplex = -1 ;
    duplex_page = 0 ;
    page_height = 0 ;
    page_width = 0 ;
    input_tray = -1 ;
    output_tray = -1 ;

    /* get user preference for dithering style */
    prof_str = GetProfileByStr(XLT("axPrintDitherStyle"));
    pf_dither_style = (prof_str == NULL) ?
                        PF_DITHER_HT4 : atoi(prof_str);

    /* get color model preference */
    prof_str = GetProfileByCode(PROF_PRINT_COLOR_MODEL_);
    if (prof_str != NULL)
    {
        pf_color_model = atoi(prof_str);
        switch (pf_color_model)
        {
        default:
            pf_color_model = COLOR_MODEL_CMYK ;
        case COLOR_MODEL_CMYK :
            numColorComponents = 4;                 
            break;
        case COLOR_MODEL_RGB :
            numColorComponents = 3;
            break;
        case COLOR_MODEL_GRAY :
            numColorComponents = 1;
            break;
        }
    }
    else
    {
        numColorComponents = 4;
    }

    bitsPerColorComponent = 4;
    prof_str = GetProfileByCode(PROF_PRINT_BITS_);

    if (prof_str != NULL)
    {
        switch (bitsPerColorComponent = atoi(prof_str))
        {
        case 8:
            break ;
        default:
            bitsPerColorComponent = 4;
            break ;
        }
    }

    bitsPerGrayComponent = bitsPerColorComponent;

    D03PreScan();

    PF.PfNumTotalPages = PF.PfNumDocPages;
    if (PF.PfBanner)
        PF.PfNumTotalPages++;

    memset((char *) &pf03CurFont, 0, sizeof(textAttrs));

    PF.PfDev_align = TM_BASELEFT;       /* was topleft before 6/25 */

    In_HPGL_Mode = FALSE;

    PCL_initialize() ;

    HPGL_initialize() ;

    /* down load any required soft fonts */
    D03SoftDownLoad() ;

    if (PF.PfBanner)
    {
        PF.PfLeftMarginMil = 500 ;
        StdBanner (NULL, banner_point_size, 500);
        PF.PfHardX = -1;
        PF.PfHardY = -1;
    }

    return(0);
}

/************************************************************************
 *
 *  EndJob
 *
 ************************************************************************/
int D03EndJob()
{
    int pat;

    pfResetPattern();

    for (pat=0; pat <= MAX_HPGL_PAT; pat++)
        HPGL_pattern_table[pat] = -1;

    PF_MEM_CHECK;

	/* this stuff seems to be required in Windows */
#if AXMS
    if (In_HPGL_Mode)
        SwitchTo_PCL() ;
    PfOutString(NXLT("\014")) ;         /* form feed */
#endif
    PCL_command(NXLT("&l1T"));			/* Job Separation command */
#if AXMS
    PCL_command(NXLT("E"));             /* ESC E - Master reset */
#endif

    PF_MEM_CHECK;
    max_poly_points = 0;

    return(0);
}

/************************************************************************
 *
 * NewPage function
 *
 ************************************************************************/
int D03NewPage(firstPage, clip)
    int firstPage;          /* TRUE for 1st physical page of printed output */
    TmRect32 *clip ;
{
	double left_offset, top_offset ;


    if ((PF.LastPage == TRUE) || (PF.PfPageNum > PF.PfNumDocPages))
    {
        /* don't need a paper feed here because there are no more pages */
        return(0);
    }

    /* print the previous page */
    if (!firstPage)
    {
#if 1
        if (In_HPGL_Mode)
            SwitchTo_PCL() ;
        PfOutString(NXLT("\014")) ;
#else
        PCL_command(NXLT("&l0H"));
#endif
    }

    /* select source for this page */
    if (PF.PfPageNum > 0)
        select_source_tray() ;

    PF.PfHardX = -1;
    PF.PfHardY = -1;

    if (PF.PfPageNum == 1 && PF.PfPrintMultiCopiesInLang > 1)
        PCL_command(NXLT("&l%dX"), PF.PfPrintMultiCopiesInLang);

    /* select plex */
    if (PF.pi.duplex == -1 && PF.PfPageNum > 0)
        select_plex(PF.duplex) ;
    duplex_page++ ;

	/* let's grasp at straws */
    PCL_command(NXLT("&l+0Z")) ;
    PCL_command(NXLT("&l+0U")); 

    /* set page size, orientation and registration offsets */
    if (PF.PfLandscape)
    {
        select_size(PF.PfPageMilHeight, PF.PfPageMilWidth) ;
        PCL_command(NXLT("&l1O"));
   
        /* if duplex short side binding and even page use -top offset */
        if (duplex == 2 && (duplex_page & 1) == 0)
			top_offset = -MIL_TO_DECIPT(167 - PF.PfLeftMarginMil) ;
        else
			top_offset = MIL_TO_DECIPT(167 - PF.PfLeftMarginMil) ;

        /* if duplex long side binding and even page use -left offset */
        if (duplex == 1 && (duplex_page & 1) == 0)
			left_offset = 400 ;
		else
			left_offset = -400 ;
		left_offset = 0 ;
    }
    else
    {
        select_size(PF.PfPageMilWidth, PF.PfPageMilHeight) ;
        PCL_command(NXLT("&l0O"));

        /* if duplex long side binding and even page, use -left offset */
        if (duplex == 1 && (duplex_page & 1) == 0)
			left_offset = -MIL_TO_DECIPT(PF.PfLeftMarginMil - 250) ;
        else
			left_offset = MIL_TO_DECIPT(PF.PfLeftMarginMil - 250) ;
        /* if duplex short side binding and even page use -top offset */
        if (duplex == 2 && (duplex_page & 1) == 0)
			top_offset = 300 ;
		else
			top_offset = -300 ;
		top_offset = 0 ;
    }
    PCL_command(NXLT("&l%+.2fZ"), top_offset) ;
    PCL_command(NXLT("&l%+.2fU"), left_offset); 

    if (PF.PfPageNum > 0)
        select_destination() ;

    PF.ClipGSaveLevel = 0;  /* just making sure */

    PF.left = 0;
    PF.right = 0;
    PF.top = 0;
    PF.bottom = 0;

    PF.cur_thick = -1;

    PCL_command(NXLT("&l0E")); /* Top Margin, PCL5 Ref. 5-18 */
    PCL_command(NXLT("&a0L")); /* Left Margin, PCL5 Ref. 5-15 */
    PCL_command(NXLT("&l0L")); /* Perf. Skip, PCL5 Ref. 5-21 */

    /* reset the appclip to the page clip */
    PF.PfPage.appClip = PF.PfPage.inner;

    /* set the new page top */
    PF.PfPagePixHeight = ymiltogpix(PF.PfPageMilHeight);
    PF.PfPagePixWidth = ymiltogpix(PF.PfPageMilWidth);
    PF.pagetop = PF.PfPagePixHeight;

    /* See PCL5 Tech Ref. 2-9, 18-5 (in decipoints) 
     This should reflect the paper size... */
    PCL_command(NXLT("*c%dX"), (int)((PF.PfPageMilWidth * 72) / 100));
    PCL_command(NXLT("*c%dY"), (int)((PF.PfPageMilHeight * 72) / 100));

    PCL_command(NXLT("*p0X"));  /* cursor position (PCL units) */
    PCL_command(NXLT("*p0Y"));
    PCL_command(NXLT("*c0T"));  /* Picture Frame Anchor Point */

    D03Clip();
    clip_box(clip) ;
    HPGL_command(NXLT("AC;")) ;

	/* forget current font, for those Genicom dips */
    memset((char *) &pf03CurFont, 0, sizeof(textAttrs));
    return(0);
}


/************************************************************************

    select_source_tray

 ************************************************************************/
static void select_source_tray()
{
    char *s ;

    if (PF.PfPageNum == 1 && PF.pi.source_1st)
    {
        if (In_HPGL_Mode)
            SwitchTo_PCL() ;
        PfOutString(PF.pi.source_1st) ;
    }
    else if (PF.PfPageNum == PF.PfNumDocPages && PF.pi.source_last)
    {       
        if (In_HPGL_Mode)
            SwitchTo_PCL() ;
        PfOutString(PF.pi.source_last) ;
    }
    else if (PF.pi.source_all)
    {       
        if (In_HPGL_Mode)
            SwitchTo_PCL() ;
        PfOutString(PF.pi.source_all) ;
    }
    else if (input_tray != PF.inputTray)
    {
        s = PCL_source_tray(PF.inputTray) ;
        if (s)
        {
            if (In_HPGL_Mode)
                SwitchTo_PCL() ;
            PfOutString(s) ;
        }
#if 0
        if (PF.inputTray == 0)
            PCL_command(NXLT("&l1H")) ;
        else if (PF.inputTray == 1)
            PCL_command(NXLT("&l4H")) ;
#endif
        input_tray = PF.inputTray ;
    }
}

/************************************************************************

    PCL_source_tray

 ************************************************************************/
static char *PCL_source_tray(n)
    int n ;
{
    int k ;
    char *p = NULL ;

    if (PF.pi.source)
    {
        k = AxArraySize(PF.pi.source) ;
        if (n < k)
            p = AxStrFromArray(PF.pi.source, n) ;
    }
    return p ;
}

/************************************************************************

    select_plex

*************************************************************************/
static void select_plex(p)
    int p ;
{
    if (p != duplex)
    {
        if (p >= 0)
            PCL_command(NXLT("&l%dS"), p) ;
        duplex = p ;
        duplex_page = 0 ;
    }
}

/************************************************************************

    select_destination

*************************************************************************/
static void select_destination()
{
    char *s ;

    if (PF.PfPageNum == 1 && PF.pi.dest_1st)
    {
        if (In_HPGL_Mode)
            SwitchTo_PCL() ;
        PfOutString(PF.pi.dest_1st) ;
    }
    else if (PF.PfPageNum == PF.PfNumDocPages && PF.pi.dest_last)
    {       
        if (In_HPGL_Mode)
            SwitchTo_PCL() ;
        PfOutString(PF.pi.dest_last) ;
    }
    else if (PF.pi.dest_all)
    {       
        if (In_HPGL_Mode)
            SwitchTo_PCL() ;
        PfOutString(PF.pi.dest_all) ;
    }

    else if (output_tray != PF.outputTray)
    {
        s = PCL_dest_tray(PF.outputTray) ;
        if (s)
        {
            if (In_HPGL_Mode)
                SwitchTo_PCL() ;
            PfOutString(s) ;
        }
        output_tray = PF.outputTray ;
    }
}


/************************************************************************

    PCL_dest_tray

*************************************************************************/
static char *PCL_dest_tray(n)
    int n ;
{
    int k ;
    char *p = NULL ;

    if (PF.pi.dest)
    {
        k = AxArraySize(PF.pi.dest) ;
        if (n < k)
            p = AxStrFromArray(PF.pi.dest, n) ;
    }
    return p ;
}

/************************************************************************

    page_size

*************************************************************************/
static void select_size(w, h)
    int w, h ;
{
    int code = -1 ;
    int i, k ;
    data_ptr z ;
    int wz, hzl ;
    char *p ;

    if (w == page_width && h == page_height)
        return ;

    page_width = w ;
    page_height = h ;

    if (PF.pi.size)
    {
        k = AxArraySize(PF.pi.size) ;
        for (i = 0 ; i < k ; i++)
        {
            z = AxArrayFromArray(PF.pi.size, i) ;
            wz = AxIntFromArray(z, 0) ;
            hzl = AxIntFromArray(z, 1) ;
            if (quite_close(w, wz) && quite_close(h, hzl))
            {
                p = AxStrFromArray(z, 2) ;
                if (p && *p != 0)
                {
                    if (In_HPGL_Mode)
                        SwitchTo_PCL() ;
                    PfOutString(p) ;
                    return ;
                }
            }
        }
    }

    if (w == 7250 && h == 10500)
        code = 1 ;
    else if (w == 8500 && h == 11000)
        code = 2 ;
    else if (w == 8500 && h == 14000)
        code = 3 ;
    else if (w == 11000 && h == 17000)
        code = 6 ;
    /* A4, more or lesss */
    else if (w >= 8264 && w <= 8268 && h >= 11692 && h <= 11694)
        code = 26 ;
    /* A3, more or lesss */
    else if (w >= 11692 && w <= 11694 && h >= 16528 && h <= 16536)
        code = 27 ;
#if 0
    else if (w == 0 && h == 0)
        code = 45 ;
    else if (w == 0 && h == 0)
        code = 71 ;
    else if (w == 0 && h == 0)
        code = 72 ;
#endif
    else if (h == 7500 && w == 3875)
        code = 80 ;
    else if (h == 9500 && w == 4125)
        code = 81 ;
    else if (h ==  4330 && w == 8661)
        code = 90 ;
    else if (h == 6377 && w == 9015)
        code = 91 ;
    else if (h == 6929 && w == 9842)
        code = 100 ;
    if (code != -1)
        PCL_command(NXLT("&l%dA"), code) ;
}
        
/************************************************************************
    quite_close
*************************************************************************/
static int quite_close(a, b)
    int a, b ;
{
    return a >= b-4 && a <= b+4 ;
}

/************************************************************************
    D03DrawClip
*************************************************************************/
int D03DrawClip(clip)
    TmRect32 *clip ;
{
    clip_box(clip) ;
    return(0);
}

/************************************************************************
 ************************************************************************/
int D03PrControl(n, buf)
    int n ;
    char *buf ;
{
    if (In_HPGL_Mode)
        SwitchTo_PCL() ;
    PfOutString(buf);
    return(0);
}

/************************************************************************
 *
 *  Make registration marks.
 *
 ************************************************************************/

int D03Printmarks(use_crops, use_regis, use_bounds, bounds_mil)
    int use_crops, use_regis, use_bounds;
    TmRect32 *bounds_mil;
{
    int i;
    int edge_x1_mil, edge_y1_mil, edge_x2_mil, edge_y2_mil ;
    int page_wid_mil, page_hyt_mil, margin_mil;
    TmRect32 pf_bounds;
    TmRect32 pf_bounds_mil;

    /* shading fills for registration 'color' boxes */
    static int HP_fills[] = {0, 1, 3, 11, 21, 36, 56, 81, 99, 100} ;

    if (!( use_crops || use_regis || use_bounds ))
        return(0);

    hpgl_IW(0.0, 0.0, 0.0, 0.0) ;

    HPGL_command(NXLT("PW;"));          /* default the line width */
    pf_bounds_mil.x_position = bounds_mil->x_position;
    pf_bounds_mil.y_position = bounds_mil->y_position;
    pf_bounds_mil.x_size = bounds_mil->x_size;
    pf_bounds_mil.y_size = bounds_mil->y_size;
    page_wid_mil = PF.PfPageMilWidth;
    page_hyt_mil = PF.PfPageMilHeight;
    margin_mil = MARK_MARGIN;

    if ( use_bounds )
    {
        edge_x1_mil = pf_bounds_mil.x_position;
        edge_y1_mil = pf_bounds_mil.y_position;
        edge_x2_mil = edge_x1_mil + pf_bounds_mil.x_size;
        edge_y2_mil = edge_y1_mil + pf_bounds_mil.y_size;

        /* Enter polygon mode, Pen down */

        HPGL_command(NXLT("PM;PU%.2f,%.2f;"),
            X_MIL_TO_PLU(edge_x1_mil),
            Y_MIL_TO_PLU(edge_y1_mil));
        HPGL_command(NXLT("PD%.2f,%.2f"),
            X_MIL_TO_PLU(edge_x2_mil),
            Y_MIL_TO_PLU(edge_y1_mil));
        HPGL_command(NXLT(",%.2f,%.2f"),
            X_MIL_TO_PLU(edge_x2_mil),
            Y_MIL_TO_PLU(edge_y2_mil));
        HPGL_command(NXLT(",%.2f,%.2f"),
            X_MIL_TO_PLU(edge_x1_mil),
            Y_MIL_TO_PLU(edge_y2_mil));
        HPGL_command(NXLT(",%.2f,%.2f"),
            X_MIL_TO_PLU(edge_x1_mil),
            Y_MIL_TO_PLU(edge_y1_mil));

        /* Pen up, close and exit polygon mode, edge the polygon */
        HPGL_command(NXLT(";PU;PM2;EP;"));

    }
    if ( use_crops )
    {
        /* top crops */
        if ( bounds_mil->y_position > 100 ) /* was 250 */
        {
            edge_x1_mil = pf_bounds_mil.x_position;
            edge_y1_mil = margin_mil;
            edge_x2_mil = pf_bounds_mil.x_position + pf_bounds_mil.x_size;
            edge_y2_mil = pf_bounds_mil.y_position - margin_mil;

            /* left */
            HPGL_command(NXLT("PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y2_mil));

            /* right */
            HPGL_command(NXLT("PU%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x2_mil),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;PU;"),
                X_MIL_TO_PLU(edge_x2_mil),
                Y_MIL_TO_PLU(edge_y2_mil));
        }

        /* right crops */
        if ( PF.PfPageMilWidth - (bounds_mil->x_position+bounds_mil->x_size) > 100 )
        {
            edge_x1_mil = pf_bounds_mil.x_position + pf_bounds_mil.x_size + margin_mil;
            edge_y1_mil = pf_bounds_mil.y_position;
            edge_x2_mil = page_wid_mil - margin_mil;
            edge_y2_mil = pf_bounds_mil.y_position + pf_bounds_mil.y_size;

            /* top */
            HPGL_command(NXLT("PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x2_mil),
                Y_MIL_TO_PLU(edge_y1_mil));
            /* right */
            HPGL_command(NXLT("PU%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y2_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;PU;"),
                X_MIL_TO_PLU(edge_x2_mil),
                Y_MIL_TO_PLU(edge_y2_mil));
        }

        /* bottom crops */
        if ( PF.PfPageMilHeight - (bounds_mil->y_position+bounds_mil->y_size) > 100 )
        {
            edge_x1_mil = pf_bounds_mil.x_position;
            edge_y1_mil = pf_bounds_mil.y_position + pf_bounds_mil.y_size + margin_mil;
            edge_x2_mil = pf_bounds_mil.x_position + pf_bounds_mil.x_size;
            edge_y2_mil = page_hyt_mil - margin_mil;

            /* right */
            HPGL_command(NXLT("PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x2_mil),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x2_mil),
                Y_MIL_TO_PLU(edge_y2_mil));
            /* left */
            HPGL_command(NXLT("PU%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;PU;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y2_mil));
        }

        /* left crops */
        if ( bounds_mil->x_position > 100 )
        {
            edge_x1_mil = margin_mil;
            edge_y1_mil = pf_bounds_mil.y_position;
            edge_x2_mil = pf_bounds_mil.x_position - margin_mil;
            edge_y2_mil = pf_bounds_mil.y_position + pf_bounds_mil.y_size;

            /* bottom */
            HPGL_command(NXLT("PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y2_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x2_mil),
                Y_MIL_TO_PLU(edge_y2_mil));
            /* top */
            HPGL_command(NXLT("PU%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;PU;"),
                X_MIL_TO_PLU(edge_x2_mil),
                Y_MIL_TO_PLU(edge_y1_mil));
        }
    }

    if ( use_regis )
    {
        /* left mark */
        if ( bounds_mil->x_position > 100 )
        {
            edge_x1_mil = pf_bounds_mil.x_position / 2;
            edge_y1_mil = page_hyt_mil / 2;

            /* draw the registration circle */
            /* centered at edge_x1,edge_y1 radius margin */

            HPGL_command(NXLT("PU;PA%.2f,%.2f;CI%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil),
                MIL_TO_PLU(margin_mil));

            HPGL_command(NXLT("PU;FT;WG%.2f,0,360"),
                MIL_TO_PLU(margin_mil/2));

            HPGL_command(NXLT("PU;PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil - (margin_mil+10)),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil + (margin_mil+10)),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PU;PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil - (margin_mil+10)));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil + (margin_mil+10)));

            HPGL_command(NXLT("TR0;SP0")) ;
            HPGL_command(NXLT("PU;PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil - (margin_mil/2)),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil + (margin_mil/2)),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PU;PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil - (margin_mil/2)));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil + (margin_mil/2)));
            HPGL_command(NXLT("PU;TR1;SP1")) ;
        }
        /* right mark */
        if ( PF.PfPageMilWidth - (bounds_mil->x_position+bounds_mil->x_size) > 100 )
        {
            edge_x2_mil = (page_wid_mil - (pf_bounds_mil.x_position+pf_bounds_mil.x_size))/2.0;
            edge_x1_mil = pf_bounds_mil.x_position + pf_bounds_mil.x_size + edge_x2_mil;
            edge_y1_mil = page_hyt_mil / 2;

            /* draw the registration circle */
            /* centered at edge_x1,edge_y1 radius margin */

            HPGL_command(NXLT("PU;PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("CI%.2f;PU;"),
                MIL_TO_PLU(margin_mil));

            HPGL_command(NXLT("FT;WG%.2f,0,360"),
                MIL_TO_PLU(margin_mil/2));

            HPGL_command(NXLT("PU;PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil - (margin_mil+10)),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil + (margin_mil+10)),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PU;PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil - (margin_mil+10)));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil + (margin_mil+10)));

            HPGL_command(NXLT("TR0;SP0")) ;
            HPGL_command(NXLT("PU;PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil - (margin_mil/2)),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil + (margin_mil/2)),
                Y_MIL_TO_PLU(edge_y1_mil));
            HPGL_command(NXLT("PU;PA%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil - (margin_mil/2)));
            HPGL_command(NXLT("PD%.2f,%.2f;"),
                X_MIL_TO_PLU(edge_x1_mil),
                Y_MIL_TO_PLU(edge_y1_mil + (margin_mil/2)));
            HPGL_command(NXLT("PU;TR1;SP1")) ;

            /* draw the color boxes */
            /* 10 small squares of intensities 0..100%, sort of */
            margin_mil = margin_mil * 2;
            edge_x1_mil -= margin_mil/2;
            edge_y1_mil -= 12*margin_mil;
            for ( i=0; i<10; i++ )
            {
                HPGL_command(NXLT("PU;PA%.2f,%.2f;PM0;"),
                    X_MIL_TO_PLU(edge_x1_mil),
                    Y_MIL_TO_PLU(edge_y1_mil)) ;
                HPGL_command(NXLT("PD%.2f,%.2f,%.2f,%.2f,%.2f,%.2f;PM2;FT10,%d;FP;EP;"),
                    X_MIL_TO_PLU(edge_x1_mil + margin_mil),
                    Y_MIL_TO_PLU(edge_y1_mil),
                    X_MIL_TO_PLU(edge_x1_mil + margin_mil),
                    Y_MIL_TO_PLU(edge_y1_mil + margin_mil),
                    X_MIL_TO_PLU(edge_x1_mil),
                    Y_MIL_TO_PLU(edge_y1_mil + margin_mil),
                    HP_fills[i]) ;
                edge_y1_mil += margin_mil;
            }

            /* draw the registration string */
            edge_x2_mil = PF.PfPageMilWidth -
                    (bounds_mil->x_position + bounds_mil->x_size);
            edge_x1_mil = bounds_mil->x_position + bounds_mil->x_size +
                    edge_x2_mil/2.0 - 120;
            edge_y1_mil = PF.PfPageMilHeight / 2.0 + 3*MARK_MARGIN;
			pf_display_reg_string(NXLT("Univers"), edge_x1_mil, edge_y1_mil);
        }
        D03Clip();
        PCL_command(NXLT("*p0X"));  /* cursor position (PCL units) */
        PCL_command(NXLT("*p0Y"));
        PCL_command(NXLT("*c0T"));  /* Picture Frame Anchor Point */
    }

    return(0);

} /* D03Printmarks */


/************************************************************************
 *
 *  SetPos function for device 03
 *
 ************************************************************************/
int D03SetPos(new_x, new_y)
    int new_x, new_y;
{
    double tmp_x, tmp_y;

    if (new_x < 0)
        new_x = 0 ;

    if (new_y < 0)
        new_y = 0 ;

    PF.PfHardY = new_y;
    PF.PfHardX = new_x;

    if (In_HPGL_Mode)
    {
        tmp_x = X_DOT_TO_PLU(new_x);
        tmp_y = Y_DOT_TO_PLU(new_y);
        HPGL_command(NXLT("PA%.2f,%.2f;"), tmp_x, tmp_y);
    }
    else
    {
        tmp_x = DOT_TO_DECIPT(new_x - MIL_TO_DOT(PF.PfLeftMarginMil));
        tmp_y = DOT_TO_DECIPT(new_y);
        PCL_command(NXLT("&a%.2fV"), tmp_y);
        PCL_command(NXLT("&a%.2fH"), tmp_x);
    }

    return(0);
} /* SetPos */


/************************************************************************
 ************************************************************************/
int D03milSetPos(new_x, new_y)
    int new_x, new_y;
{
    double tmp_x, tmp_y;
    int dot_x, dot_y;

    PF.PfHardY = (int)ymiltopix(new_y);
    PF.PfHardX = (int)xmiltopix(new_x);

    if (In_HPGL_Mode)
    {
#if USE_PLU_IN_HPGL
        tmp_x = X_MIL_TO_PLU(new_x);
        tmp_y = Y_MIL_TO_PLU(new_y);
        HPGL_command(NXLT("PA%.2f,%.2f;"), tmp_x, tmp_y);
#else
        HPGL_command(NXLT("PA%d,%d;"), PF.PfHardX, PF.PfHardY);
#endif
    }
    else
    {
        tmp_x = MIL_TO_DECIPT(new_x - PF.PfLeftMarginMil);
        tmp_y = MIL_TO_DECIPT(new_y);
        PCL_command(NXLT("&a%.2fV"), tmp_y);
        PCL_command(NXLT("&a%.2fH"), tmp_x);
    }
    return(0);
} /* milSetPos */


/************************************************************************
 *
 *  Draw polyline
 *
 ************************************************************************/
int D03Line (nPolys, attr, pol, clip)
    int nPolys;
    attr_type *attr;
    poly_type *pol;
    TmRect32 *clip ;
{
    int polyN;
    int NeedToFill = TRUE;
    int NeedToStroke = TRUE;
    TmPoint32 boundingBox[4];
    poly_type *bezier_pol ;

    HPGL_UsePolygonMode = TRUE ;

    /* weed out the 'no print' line types, if you can */
    if (attr->l_wid == 0 ||
         (attr->l_type == BUILTIN_FILL_TYPE && attr->l_pat == -1))
        NeedToStroke = FALSE;

    /* ditto for the fills */
    if (attr->fill_flag == 0)
        NeedToFill = FALSE;

    if (!(NeedToFill || NeedToStroke))
        return 0;

    clip_box(clip) ;

    /* find bounding box! */
    D03GetBoundingBox(nPolys, attr, pol, boundingBox);


    if (NeedToStroke)
    {
        reset_line_ends(attr->l_cap) ;

        /* set miter limit and join */
        hpgl_LA(2, 4);
        hpgl_LA(3, 2);

        /* set the line style */
        /* See PCL5 Ref. 20.28 LT, Line Type */
        /* There are other things that change the LT.   Safest is to set it
            explicitly everytime we need it. */
        PF.cur_style = attr->l_style;   
        switch (PF.cur_style)
        {
        case TM_SOLID:
            /* ################ */
            hpgl_LT(0) ;
            break;

        case TM_DASHED:
            /* #######_____#######_____ */
            hpgl_LA(1, 1);
            hpgl_LT(1) ;
            break;

        case TM_DOTTED:
            /* ###_____###_____ */
            hpgl_LA(1, 1);
            hpgl_LT(2) ;
            break;

        case TM_DADOT:
            /* #######_____###_____ */
            hpgl_LA(1, 1);
            hpgl_LT(3) ;
            break;

        case TM_THIN_DOTTED:
            /* #__#__ */
            hpgl_LA(1, 1);
            hpgl_LT(4) ;
            break;

        case 5:
            /* #######_____###_____###_____ */
            hpgl_LA(1, 1);
            hpgl_LT(5) ;
            break;

        case 6:
            /* ##############_____ */
            hpgl_LA(1, 1);
            hpgl_LT(6) ;
            break;

        default:
            hpgl_LT(0) ;
            break;
        }

#if 0
        if ( !attr->lfg.see_thru && !attr->lbg.see_thru )
            hpgl_TR(FALSE);
        else if ( attr->lfg.see_thru )
            hpgl_TR(FALSE);
        else if ( attr->lbg.see_thru )
            hpgl_TR(TRUE);
        else
            hpgl_TR(FALSE);
#endif

        if (PF.cur_thick != attr->l_wid)
        {
            PF.cur_thick = attr->l_wid;
            if (PF.cur_thick == 0)
                HPGL_command(NXLT("PW0;"));
            else
                HPGL_command(NXLT("PW%.4f;"), pen_width(PF.cur_thick));
        }
    }


    bezier_pol = (poly_type *) malloc(nPolys * sizeof(poly_type)) ;
    memset(bezier_pol, 0, nPolys * sizeof(poly_type)) ;

    for (polyN=0; polyN<nPolys; polyN++)
    {
        HPGL_one_line(polyN, attr, &pol[polyN], &bezier_pol[polyN], boundingBox);
        if (HPGL_UsePolygonMode && polyN < nPolys-1)
            hpgl_PM(1) ;
    }

    if (HPGL_UsePolygonMode)
        hpgl_PM(2) ;        /* exit Polygon Mode */


    /* fill the path according to pattern info passed in */
    if ( NeedToFill )
         D03FillPattern(attr, boundingBox, clip,
                 bezier_pol[0].nPoints != 0 ?
                &bezier_pol[0] : &pol[0]);

    if ( NeedToStroke )
         D03FillStroke(attr, boundingBox,
                 bezier_pol[0].nPoints != 0 ?
                &bezier_pol[0] : &pol[0]);


    /* Now free any memory malloc()ed in convertPols() */
    for (polyN=0; polyN<nPolys; polyN++)
    {
        if (bezier_pol[polyN].nPoints != 0)
            axFreePolylinePoints(&(bezier_pol[polyN].points), 1);
    }

    free(bezier_pol) ;
    return 0 ;
}


/************************************************************************
 *
 *
 ************************************************************************/
static void HPGL_one_line(n, attr, pol, bezier_pol, boundingBox)
    int n ;
    attr_type *attr;
    poly_type *pol, *bezier_pol ;
    TmPoint32 *boundingBox;
{

    if ( pol->nCtrls == 0 ) /* draw a simple poly */
    {
        if (pol->nPoints > max_poly_points)
            max_poly_points = pol->nPoints;


        HPGL_points(n, attr, pol, boundingBox);


    }
    else
    {
        /* Bezier curves... */
        HPGL_command(NXLT("PA%.2f,%.2f;"),
            X_DOT_TO_PLU(pol->points[0].x), Y_DOT_TO_PLU(pol->points[0].y));

        /* create a polygon of the Bezier points */
        convertPols(1, pol, &(bezier_pol->nPoints), &(bezier_pol->points), 0);
        HPGL_points(n, attr, bezier_pol, boundingBox);

    }

    PF_MEM_CHECK;
}

#define TX(x, y)    ((x-orgx)*cosa - (y-orgy)*sina)
#define TY(x, y)    ((x-orgx)*sina + (y-orgy)*cosa)

/************************************************************************

    D03GetBoundingBox

    accumulate an enclosing rectangle for a poly_type instance
    and also the rectangle for the rotated poly
    return:
    
        box[0] = LLC of rect
        box[1] = URC of rect
        box[2] = LLC of rotated rect
        box[3] = width, height pf rotated rect
 ************************************************************************/
static void D03GetBoundingBox(nPolys, attr, pol, box)
    int nPolys;
    attr_type *attr;
    poly_type *pol;
    TmPoint32 *box;
{
    register int ix, jx, half_line_width;
    double minx, miny, maxx, maxy, xr, yr ;
    double angle, sina, cosa ;
    double orgx, orgy ;
        

    /* init boundary */
    box[0].x = box[1].x = (int) pol[0].points[0].x;
    box[0].y = box[1].y = (int) (pol[0].points[0].y);

    /* scan poly info */
    for (ix = 0; ix < nPolys; ix++) {
        /* scan points */
        for (jx = 0; jx < pol[ix].nPoints; jx++)
        {
            if (pol[ix].points[jx].x < box[0].x)
                box[0].x = pol[ix].points[jx].x;
            else if (pol[ix].points[jx].x > box[1].x)
                box[1].x = pol[ix].points[jx].x;
            if (pol[ix].points[jx].y < box[0].y)
                box[0].y = pol[ix].points[jx].y;
            else if (pol[ix].points[jx].y > box[1].y)
                box[1].y = pol[ix].points[jx].y;
        }
        /* this is the case for round ends, need to do square, butt */
        half_line_width = MIL_TO_DOT(attr->l_wid)/2;
        box[0].x -= half_line_width;
        box[0].y -= half_line_width;
        box[1].x += half_line_width;
        box[1].y += half_line_width;

        /* scan control points also */
        for (jx = 0; jx < pol[ix].nCtrls; jx++)
        {
            if (pol[ix].controls[jx].x < box[0].x)
                box[0].x = pol[ix].controls[jx].x;
            else if (pol[ix].controls[jx].x > box[1].x)
                box[1].x = pol[ix].controls[jx].x;
            if (pol[ix].controls[jx].y < box[0].y)
                box[0].y = pol[ix].controls[jx].y;
            else if (pol[ix].controls[jx].y > box[1].y)
                box[1].y = pol[ix].controls[jx].y;
        }
    }

    /* now find the rotated box */

    if (attr->f_ang == 0)
    {
        box[2].x = 0 ;
        box[2].y = 0 ;
        box[3].x = ABS(box[1].x - box[0].x) ;
        box[3].y = ABS(box[1].y - box[0].y) ;
        return ;
    }       

    angle = ((double)(-attr->f_ang) * M_PI) / 180.0 ;
    sina = sin(angle) ;
    cosa = cos(angle) ;
    orgx = box[0].x ;
    orgy = box[0].y ;

    /* init boundary */
#if 0
    minx = maxx = TX(pol[0].points[0].x, pol[0].points[0].y) ;
    miny = maxy = TY(pol[0].points[0].x, pol[0].points[0].y) ;

    /* scan poly info */
    for (ix = 0; ix < nPolys; ix++)
    {
        /* scan points */
        for (jx = 0; jx < pol[ix].nPoints; jx++)
        {
            xr = TX(pol[ix].points[jx].x, pol[ix].points[jx].y) ;
            yr = TY(pol[ix].points[jx].x, pol[ix].points[jx].y) ;
            if (xr < minx)
                minx = xr ;
            else if (xr > maxx)
                maxx = xr ;
            if (yr < miny)
                miny = yr ;
            else if (yr > maxy)
                maxy = yr ;

        }

        /* this is the case for round ends, need to do square, butt */
        half_line_width = attr->l_wid/2;
        minx -= half_line_width;
        miny -= half_line_width;
        maxx += half_line_width;
        maxy += half_line_width;

        /* scan control points also */
        for (jx = 0; jx < pol[ix].nCtrls; jx++)
        {
            xr = TX(pol[ix].controls[jx].x, pol[ix].controls[jx].y) ;
            yr = TY(pol[ix].controls[jx].x, pol[ix].controls[jx].y) ;
            if (xr < minx)
                minx = xr ;
            else if (xr > maxx)
                maxx = xr ;
            if (yr < miny)
                miny = yr ;
            else if (yr > maxy)
                maxy = yr ;
        }
    }
#endif
    minx = maxx = TX(box[0].x, box[0].y) ;
    miny = maxy = TY(box[0].x, box[0].y) ;

    xr = TX(box[0].x, box[1].y) ;
    yr = TY(box[0].x, box[1].y) ;

    if (xr < minx)
        minx = xr ;
    else if (xr > maxx)
        maxx = xr ;
    if (yr < miny)
        miny = yr ;
    else if (yr > maxy)
        maxy = yr ;

    xr = TX(box[1].x, box[1].y) ;
    yr = TY(box[1].x, box[1].y) ;

    if (xr < minx)
        minx = xr ;
    else if (xr > maxx)
        maxx = xr ;
    if (yr < miny)
        miny = yr ;
    else if (yr > maxy)
        maxy = yr ;

    xr = TX(box[1].x, box[0].y) ;
    yr = TY(box[1].x, box[0].y) ;

    if (xr < minx)
        minx = xr ;
    else if (xr > maxx)
        maxx = xr ;
    if (yr < miny)
        miny = yr ;
    else if (yr > maxy)
        maxy = yr ;
#if 0
    /* now rotate back */
    angle = ((double)(attr->f_ang) * M_PI) / 180.0 ;
    sina = sin(angle) ;
    cosa = cos(angle) ;
    orgx = orgy = 0 ;

    box[2].x = (int)TX(minx, miny) + box[0].x; 
    box[2].y = (int)TY(minx, miny) + box[0].y ;
    box[3].x = maxx - minx ;
    box[3].y = maxy - miny ;
#endif
    box[2].x = minx ;
    box[2].y = miny ;
    box[3].x = maxx ;
    box[3].y = maxy ;

}

/************************************************************************
 *
 *
 ************************************************************************/
static void HPGL_points(n, attr, pol, boundingBox)
    int n ;
    attr_type *attr;
    poly_type *pol;
    TmPoint32 *boundingBox;
{
    int indx;
    int last_y = -1;
    int last_x = -1;
/*
 * Position Absolute, 0th point
 */

    hpgl_PU() ;
    HPGL_command(NXLT("PA%.2f,%.2f;"),
        X_DOT_TO_PLU(pol->points[0].x), Y_DOT_TO_PLU(pol->points[0].y));

    if (pol->nPoints <= 1)
        return;

    /* Enter polygon mode, Pen down */
    if (HPGL_UsePolygonMode)
        if (n == 0)
            hpgl_PM(0) ;        /* new Polygon */
    hpgl_PD(X_DOT_TO_PLU(pol->points[1].x), Y_DOT_TO_PLU(pol->points[1].y));
    last_x = pol->points[1].x;
    last_y = pol->points[1].y;
    for (indx=2; indx<pol->nPoints; indx++)
    {
        if ((ABS(last_x - pol->points[indx].x) > 2) ||
            (ABS(last_y - pol->points[indx].y) > 2))
        {
            HPGL_command(NXLT(",%.2f,%.2f"), 
                X_DOT_TO_PLU(pol->points[indx].x),
                Y_DOT_TO_PLU(pol->points[indx].y));
            last_x = pol->points[indx].x;
            last_y = pol->points[indx].y;
        }
    }
    /* Pen up */
    HPGL_command(NXLT(";"));
    hpgl_PU() ;
}


/************************************************************************
 *
 *  Clear Box
 *
 *  Changes coordinates for 75 dpi mode.
 *
 ************************************************************************/

int
D03ClrBox (x, y, w, h)
int x, y, w, h;
{
    AxRectangle old_clip;

    old_clip = PF.PfPage.appClip;
#if 0
    clearbox (x, y, w, h);
#endif
    PF.PfPage.appClip = old_clip;

    return(0);
} /* D03ClrBox */


/************************************************************************
 *
 *
 *
 ************************************************************************/
static int find_next_free_cache_slot(id)
    int id;
{

    int index;
    int old_id;
    int candidate;

    for (index=1; index<=MAX_HPGL_PAT; index++)
    {
        if (HPGL_pattern_table[index] == -1)    /* is free... */
        {
            HPGL_pattern_table[index] = id; /* assign it */
            return(index);
        }
    }
    /* have to spill an entry... */
    /* throw out the least used entry... LATER */
    candidate = 8;  /* select one to trash... */
    old_id = HPGL_pattern_table[candidate];
    pattern[old_id].ref_index = -1;
        /* tell old_id that we need to reload it before use */

    HPGL_pattern_table[candidate] = id;
    pattern[id].ref_index = candidate;
    return(index);
}

/************************************************************************
 *
 * This function gets the pattern to the printer file
 * as a raster fill.
 *
 ************************************************************************/
static int D03LoadPattern(id, fg, bg)
    int id;
    color_type *fg, *bg ;
{
    char *pat_data;
    XImage *Xgray = NULL, *Xmono = NULL ;
    int HPGL_pattern_index;
    int bytewidth ;
    int pat_wid, pat_hyt;
    int p0, p1 ;
    int fpen, bpen ;
    short *FSbuffer;
    int FS_width, FS_height ;
    int x, y, p, b, z, i, j ;

    pat_data = pattern[id].data;
    pat_wid = pattern[id].wid;
    pat_hyt = pattern[id].hyt;
    if ((pat_wid==0) || (pat_hyt==0) || (pat_data == NULL))
        return(-1);

    /* if this is mono, but new fg or bg color, don't use old */
    if (pattern[id].dep == 1)
        if (memcmp(fg, &pattern[id].fg, sizeof(color_type)) != 0  ||
            memcmp(bg, &pattern[id].bg, sizeof(color_type)) != 0)
            pattern[id].ref_index = -1 ;

    /* since the HP 1200 DJ can't remember palettes by ID no, we have to
       create the palette every time we use it ... bummer */
    if (PF.PfColordev && pattern[id].dep == 8)
    {
        if (pf_color_model == COLOR_MODEL_GRAY)
            PCL_command(NXLT("*r1U")) ;         /* go into BW mode */
        else
            create_palette(&pattern[id].cmap, pat_data, pat_wid, pat_hyt) ;
    }


    HPGL_pattern_index = pattern[id].ref_index;

    /* if already loaded, use it */
    if (HPGL_pattern_index != -1)
        return (HPGL_pattern_index);

    pattern[id].ref_index = find_next_free_cache_slot(id);
    pattern[id].fg = *fg ;
    pattern[id].bg = *bg ;
    bytewidth = pattern[id].byte_wid;
    if (bytewidth == 0)
        bytewidth = 2*((pat_wid + 15)/16);

    HPGL_pattern_index = pattern[id].ref_index;

    /* load the bits into HPGL_pattern_index */
    switch (pattern[id].dep)
    {
    case 8:
        /* color pattern fill */
        if (!PF.PfColordev || pf_color_model == COLOR_MODEL_GRAY)
        {
            /* color fill to b&w printer */
            z = id < MAX_PATS ? 1 : 300 / SCREEN_RESOLUTION ;
            Xgray = (XImage *)calloc(1, sizeof(XImage));
            Xgray->depth = 8 ;
            Xgray->width = z * pat_wid ;
            Xgray->height = z * pat_hyt ;
            Xgray->bytes_per_line = Xgray->width ;
            Xgray->data =
                (char *)calloc(1, Xgray->bytes_per_line * Xgray->height) ;
            FS_width = (pat_wid + 1) & ~1; /* 16 bit aligned */
            FS_height = pat_hyt ;
            FSbuffer = (short *)malloc(FS_width * FS_height * sizeof(short)) ;
        
            ColorToGray(pat_data, pat_wid, pat_hyt, &pattern[id].cmap, FSbuffer, 16) ;
    
            for (y = 0 ; y < FS_height ; y++)
                for (j = 0 ; j < z ; j++)
                    for (x = 0; x < FS_width; x++)
                    {
                        p = 0xFF - (unsigned char)(FSbuffer[(y*FS_width)+x] & 0xFF);
                        for (i = 0 ; i < z ; i++)
                            Xgray->data[((y * z + j) * Xgray->bytes_per_line) + (x*z + i)] = p ;
                    }

#if 1
            Xmono = HalfTone(Xgray);
            Xmono->bytes_per_line = 2*((Xmono->width + 15)/16);

            pfFreeXImage(Xgray) ;   
            MonoRasterFill(0, HPGL_pattern_index, Xmono->width, Xmono->height,
                    Xmono->bytes_per_line, Xmono->data, 0, 1) ;
#else
            Xmono = (XImage *)calloc(1, sizeof(XImage));
            Xmono->depth = 1 ;
            Xmono->width = pat_wid ;
            Xmono->height = pat_hyt ;
            Xmono->bytes_per_line = bytewidth ;
            Xmono->data =
                (char *)calloc(1, Xmono->bytes_per_line * Xmono->height) ;

            for (y = 0 ; y < Xgray->height ; y++)
                for (x = 0 ; x < Xgray->width ; x++)
                {
                    p = 0xff & Xgray->data[y * Xgray->bytes_per_line + x] ;
                    if (p < 127)
                        continue ;
                    b = 0x80 >> (x % 8) ;
                    Xmono->data[y * Xmono->bytes_per_line + x/8] |= b ;
                }
#endif

            pfFreeXImage(Xmono) ;   
            free(FSbuffer);
        }
        else
        {
            /* color fill to color printer */
            bytewidth = (pat_wid + 1) & ~1 ;
            ColorRasterFill(id, HPGL_pattern_index, pat_wid, pat_hyt,
                bytewidth, pat_data) ;
        }
        break ;

    default:
    case 1:
        /* mono pattern fill */
        if (PF.PfColordev)
        {
            fpen = find_pen(fg->c, fg->m, fg->y, fg->k) ;
            bpen = find_pen(bg->c, bg->m, bg->y, bg->k) ;
            p0 = bpen ;
            p1 = fpen ;
        }
        else
        {
            /* decide who's gonna be black */
            if (IntensFromCMYKColor(fg->c, fg->m, fg->y, fg->k, 255) >=
                IntensFromCMYKColor(bg->c, bg->m, bg->y, bg->k, 255))
            {
                p0 = 0 ;
                p1 = 1 ;
            }
            else
            {
                p0 = 1 ;
                p1 = 0 ;
            }
        }

        MonoRasterFill(id, HPGL_pattern_index, pat_wid, pat_hyt, bytewidth,
                 pat_data, p0, p1) ;
    }
    return (HPGL_pattern_index);
}

/************************************************************************
 *
 *
 ************************************************************************/
static void MonoRasterFill(id, index, w, h, bw, data, p0, p1)
    int id, index, w, h, bw, p0, p1 ;
    char *data ;
{
    int i, j, k, m, q, z ;
    int bytewidth ;
    char tc ;
	int zw, zh ;
	int comma ;

    /* scale pattern from screen res. up to 300 dpi if it's user defined */
    z = id < MAX_PATS ? 1 : 300 / SCREEN_RESOLUTION ;
	zw = min(z*w, 255) ;
	zh = min(z*h, 255) ;

    HPGL_command(NXLT("RF%d,%d,%d,"), index, zw,  zh);
	comma = FALSE ;

    for (i = 0 ; i < h ; i++)
    {
        for (k = 0 ; k < z ; k++)
        {
			if (i*k >= 255)
				goto all_done ;
            for (j = 0 ; j < w ; j++)
            {
                /* i * bw is the vertical scan line in bytes
                    j/8 is the number of bytes from the start of the
                    scan line */
                tc = data[(i * bw) + (j/8)];
                if (tc & (0x80 >> (j%8)))
                    q = p1 ;
                else
                    q = p0 ;
                for (m = 0 ; m < z ; m++)
                {
					if (j*m >= 255)
						goto next_row ;
					if (comma)
	                    put(NXLT(",")) ;
                    put(NXLT("%d"), q) ;
					comma = TRUE ;
                }
            }
next_row:
		;
        }
    }

all_done:
    put(NXLT(";"));         /* end of pattern data, terminate command */
}

/************************************************************************
 *
 *
 ************************************************************************/
static void ColorRasterFill(id, index, w, h, bw, data)
    int id, index, w, h, bw ;
    char *data ;
{
    char *t ;
    int i, j, k, m ;
    int z ;
    int p ;
	int zw, zh ;
	int comma ;

    /* scale pattern from screen res. up to 300 dpi if it's user defined */
    z = id < MAX_PATS ? 1 : 300 / SCREEN_RESOLUTION ;
	zw = min(z*w, 255) ;
	zh = min(z*h, 255) ;

    HPGL_command(NXLT("RF%d,%d,%d,"), index, zw,  zh);
	comma = FALSE ;

    for (i = 0 ; i < h ; i++)
    {
        for (k = 0 ; k < z ; k++)
        {
			if (i*k >= 255)
				goto all_done ;
            t = &data[i*bw] ;
            for (j = 0 ; j < w ; j++)
            {
                p = *t++ & 0xff ;
                for (m = 0 ; m < z ; m++)
                {
					if (j*m >= 255)
						goto next_row ;
					if (comma)
	                    put(NXLT(",")) ;
                    put(NXLT("%d"), p) ;
					comma = TRUE ;
                }
            }
next_row:
		;
        }
    }
all_done:
    put(NXLT(";"));         /* end of pattern data, terminate command */
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03PreLoadPatterns()
{

    int pat, bytewidth, bytesize;
    int hasMask;
    char *ip;
    int ii;
    bitmap_struct Bitmap;

    extern char *printer_fills[];

    /* load the built in patterns used in this doc into the pattern cache */
    for ( pat=0; pat<MAX_PATS; pat++ )
    {
        /* don't send unused patterns to the print ready file */
        if ( pattern[pat].needed && pat != 0 && pat != 4 )
        {
            memset((char *)&Bitmap, '\0', sizeof(bitmap_struct));
            if (AxIMtoXImage(printer_fills[pat], printer_fills[pat], &Bitmap) == AOK)
            {
                pattern[pat].wid = Bitmap.image.width;
                pattern[pat].hyt = Bitmap.image.height;
                pattern[pat].dep = Bitmap.image.depth;
                pattern[pat].byte_wid = Bitmap.image.bytes_per_line;
                bytewidth = Bitmap.image.bytes_per_line;
                bytesize = bytewidth * pattern[pat].hyt;
                pattern[pat].data = (char *)malloc(bytesize);
                memcpy(pattern[pat].data, Bitmap.image.data, bytesize);
                memset((char *)&pattern[pat].fg, 0, sizeof(pattern[pat].fg)) ;
                memset((char *)&pattern[pat].bg, 0, sizeof(pattern[pat].bg)) ;
                if (Bitmap.image.bitmap_bit_order != MSBFirst)
                {
                    ip = pattern[pat].data;
                    for (ii = bytesize; ii>0; ii--)
                    {
                        *ip = BitSpin(*ip);
                        ip++;
                    }
                }
            }
        }
    }
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03BitmapFill(sattr, boundingBox, id, eoflag, fillflag)
    sub_attr_type *sattr;
    TmPoint32 boundingBox[];
    int id;
    int eoflag; 
    int fillflag ;
{
    int HPGL_pattern_index ;
    int xpos, ypos;

    ASSERT(boundingBox != NULL);

    /* don't do anything if solid fill when color is transparent */

    if ((sattr->type == BUILTIN_FILL_TYPE) &&
        ((id == 0 && sattr->mode == TMS_O_D) ||
         (id == 4 && sattr->mode == TMS_A_D)))
        return;

    if ( sattr->mode == TMD )
        return;

/* Apparently, not all PCL5 printers support user defined raster
    fills in PCL... ! But HP-GL/2 always does... besides, (1) all
    our fills will be of graphics things created in HP-GL mode
    anyway, (2) HP-GL/2 allows definition of up to 8 "user defined"
    patterns.
    One day we'll have to specify a pen color here
*/

    if (PF.PfColordev && pattern[id].dep == 8)
        PCL_command(NXLT("*p0P")) ;
        
    HPGL_pattern_index = D03LoadPattern(id, &sattr->fg, &sattr->bg);

    /* Anchor Corner */
    if (boundingBox != NULL)
    {
        if (sattr->fill_abs)
            hpgl_AC(sattr->xoff, sattr->yoff);
        else
        {
            xpos = xmiltopix(boundingBox[0].x);
            ypos = ymiltopix(boundingBox[0].y);
            hpgl_AC(xpos + sattr->xoff, ypos + sattr->yoff);
        }
    }

    switch (fillflag)
    {
    case 0:                             /* polygon fill */
        hpgl_FT(11, HPGL_pattern_index);
        HPGL_command(NXLT("FP;"));
        break ;
    case 1:                             /* line fill */
        hpgl_SV(2, HPGL_pattern_index, 1) ;
        break ;
    case 2:                             /* character fill */
        hpgl_FT(11, HPGL_pattern_index);
        break ;
    }

    if (PF.PfColordev && pattern[id].dep == 8)
        PCL_command(NXLT("*p1P")) ;
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03BuiltinPatternFill(sattr, pol, boundingBox, eoflag, fillflag)
    sub_attr_type *sattr;
    poly_type *pol;
    TmPoint32 boundingBox[];
    int eoflag;
    int fillflag ;
{
    int id;

    id = sattr->pat;
    if (id == -1)
        return;

    /* Do transfer mode */
    switch (sattr->mode)
    {
    case TMS:
        hpgl_TR(FALSE); 
        break;
    case TMS_O_D:
    case TMIS_O_D:
        hpgl_TR(TRUE);
         break;
    default:
        hpgl_TR(FALSE);
         break;
    }

    D03BitmapFill(sattr, boundingBox, id, eoflag, fillflag);
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03UserdefPatternFill(sattr, pol, boundingBox, eoflag, fillflag)
    sub_attr_type *sattr;
    poly_type *pol;
    TmPoint32 boundingBox[];
    int eoflag;
    int fillflag ;
{
    int id = pfFindPattern(sattr->pat) ;

    if (id != -1) 
    {
        /* Do transfer mode */
        switch (sattr->mode)
        {
        case TMS:
            hpgl_TR(FALSE);
            break;
        case TMS_O_D:
            hpgl_TR(TRUE);
            break;
        default:
            hpgl_TR(FALSE);
            break;
        }

        D03BitmapFill(sattr, boundingBox, id, eoflag, fillflag);

    }
}

/************************************************************************
 *
 *
 ************************************************************************/
static poly_type *ConvertPolToPel(pol, res)
    poly_type *pol;
    int res;
{
    poly_type *p;
    int point;

    /* this is kind of a no op, lines are already in 300 dpi units... */
    p = (poly_type *)malloc(sizeof(poly_type));
    p->nPoints = pol->nPoints;
    p->nCtrls = pol->nCtrls;
    if ((pol->controls == NULL) || (pol->nCtrls == 0))
        p->controls = NULL;
    else
    {
        p->controls =
            (TmPoint32 *)malloc(pol->nCtrls * sizeof(TmPoint32));
        for (point=0; point < pol->nCtrls; point++)
        {
            p->controls[point].x = pol->controls[point].x;
            p->controls[point].y = pol->controls[point].y;
        }
        
    }
    if ((pol->points == NULL) || (pol->nPoints == 0))
        p->points = NULL;
    else
    {
        p->points =
            (TmPoint32 *)malloc(pol->nPoints * sizeof(TmPoint32));
        for (point=0; point < pol->nPoints; point++)
        {
            p->points[point].x = pol->points[point].x;
            p->points[point].y = pol->points[point].y;
        }

    }
    return(p);
}

/************************************************************************
 *
 *  Fill the polygon pol with a gradient
 *
 ************************************************************************/
static void D03GradientFill(sattr, pol, boundingBox, clip, eoflag)
    sub_attr_type *sattr;
    poly_type *pol;
    TmRect32 *clip ;
    TmPoint32 boundingBox[];
    int eoflag;
{
    int x_pos, y_pos;
	int pen1, pen2 ;
    XImage *Xgray;
    XImage *Xmono;
    XImage *Xclip;
    XImage *Xmask;
    poly_type *p;           /* the (clipping) pol in device space */
    int pix;

    
    /* determine the starting point for the gradient */
    x_pos = boundingBox[0].x;
    y_pos = boundingBox[0].y;

    /* build a grey rectangle gradient in Xgray */
    switch (sattr->type)
    {
    case LINEAR_FILL_TYPE:
        Xgray = LinearGradientXImage(sattr, boundingBox) ;
        break ;
    case BILINEAR_FILL_TYPE:
        Xgray = BilinearGradientXImage(sattr, boundingBox) ;
        break ;
    case RADIAL_FILL_TYPE:
        Xgray = RadialGradientXImage(sattr, boundingBox) ;
        break ;
    case RECT_FILL_TYPE:
        Xgray = RectGradientXImage(sattr, boundingBox) ;
        break ;
    }
    if (Xgray == NULL)
        return;

#if 0
    /* Do transfer mode */
    switch (sattr->mode)
    {
    case TMS:
        hpgl_TR(FALSE);
        break;
    case TMS_O_D:
        hpgl_TR(TRUE);
        break;
    default:
        hpgl_TR(FALSE);
        break;
    }
#endif
    p = ConvertPolToPel(pol, TargetRes);

    /* build the clipped image in Xclip */
    Xclip = (XImage *)malloc(sizeof(XImage));
    memcpy(Xclip, Xgray, sizeof(XImage));
    Xclip->data = (char *)malloc(Xclip->bytes_per_line * Xclip->height);
    memset(Xclip->data, 0, Xclip->bytes_per_line * Xclip->height);
    
    poly_fill(Xgray, Xclip,
                p, boundingBox,
                0, Xclip->height,
                pfRasterLineBytewiseCopier);

    /* build a mask image in Xmask */
    Xmask = (XImage *)malloc(sizeof(XImage));
    memcpy(Xmask, Xgray, sizeof(XImage));
    Xmask->depth = 1 ;
    Xmask->bytes_per_line = (Xmask->bytes_per_line + 7) / 8 ;
    Xmask->data = (char *)malloc(Xmask->bytes_per_line * Xmask->height);
    memset(Xmask->data, 0, Xmask->bytes_per_line * Xmask->height);
    
    poly_fill(Xgray, Xmask,
                p, boundingBox,
                0, Xmask->height,
                pfRasterLineMasker);

    pfFreeXImage(Xgray);

    Xmono = pf_dither_style == PF_DITHER_FS ?
                    FS_HalfTone(Xclip) : HalfTone(Xclip) ;

    pfFreeXImage(Xclip);

    if (Xmono == NULL)
        return;


    if (p != NULL)
    {
        if (p->controls != NULL)
        {
            free(p->controls);
            p->controls = NULL;
        }
        if (p->points != NULL)
        {
            free(p->points);
            p->points = NULL;
        }
        free(p);
    }
    PF_MEM_CHECK;
#if 0
    if ( (sattr->mode != TMS_O_D) && (sattr->mode != TMIS_O_D) )
    {
        PCL_command(NXLT("*v1N"));  /* source transparency mode => opaque */
    }
    else
    {
        PCL_command(NXLT("*v0N"));
    }
#else
    /* always OR the raster */
    PCL_command(NXLT("*v0N"));
#endif

    D03SetPos(x_pos, y_pos);

    if (PF.PfColordev)
	{
	    pen1 = find_pen(sattr->fg.c, sattr->fg.m, sattr->fg.y, sattr->fg.k) ;
	    pen2 = find_pen(sattr->bg.c, sattr->bg.m, sattr->bg.y, sattr->bg.k) ;
        D03MonoColorXImage(Xmono, Xmask, TargetRes, boundingBox, clip, x_pos, pen1, pen2);
	}
    else
        D03MonoXImage(Xmono, Xmask, TargetRes, boundingBox, clip, x_pos);

    pfFreeXImage(Xmono);
    pfFreeXImage(Xmask);
    return;

}


/************************************************************************
 *
 *
 ************************************************************************/
static void D03MonoXImage(Xp, Xm, image_resolution, boundingBox, clip, x_pos)
    XImage *Xp, *Xm;
    int image_resolution;
    TmPoint32 boundingBox[] ;
    TmRect32 *clip ;
    int x_pos ;
{
    int dw, iw, y, n, k ;
    char *scan_data;
    char *compression_buffer;
    int rh, rw ;
    int x_off ;

    k = (int)MIL_TO_DOT(PF.PfLeftMarginMil) ;
    if (x_pos < k)
        x_off = k - x_pos ;
    else
        x_off = 0 ;

    x_off = (x_off >> 3) << 3 ;

    rh = MIL_TO_DOT(clip->y_position + clip->y_size) - boundingBox[0].y ;
    rh = MAX(0, MIN(rh, Xp->height)) ;

    rw = MIL_TO_DOT(clip->x_position + clip->x_size) - boundingBox[0].x ;
    rw = MAX(0, MIN(rw, Xp->width)) ;

    rw -= x_off ;
    if (rw <= 0)
        return ;

    PCL_command(NXLT("*r0F"));  /* raster image prints in direction of logical page. */
    PCL_command(NXLT("*t%dR"), image_resolution);
    
    PCL_command(NXLT("*r%dT"), rh);     /* height in raster rows of the image. */
    PCL_command(NXLT("*r%dS"), rw);     /* width in pixels of specified resolution. */
    PCL_command(NXLT("*r1A"));          /* start raster graphics at current cursor position */

    switch(PCL_compression_mode)
    {
    case PCL_COMPR_RLE:
        PCL_command(NXLT("*b1M"));      /* RLE compression transfer mode */

        dw = Xp->bytes_per_line;
        iw = (rw + 7) /8;
        scan_data = Xp->data;
        compression_buffer = (char *)malloc(2 * iw);

        /* must take account of the image mask */
        for (y = 0; y < Xp->height; y++, scan_data += dw)
        {
            n = run_length_encode((unsigned char *)compression_buffer,
									(unsigned char *)scan_data + (x_off >> 3), iw);
            PCL_command(NXLT("*b%dW"), n);
            PfOutCString(compression_buffer, n);
        }
        free(compression_buffer);
        break;

    default:
    case PCL_COMPR_NONE:
        PCL_command(NXLT("*b0M"));  /* no compression transfer mode */

        /* note the difference between iw (=> how much we
        want to put out) and dw (=> the stride through the
        source data) */

        dw = Xp->bytes_per_line;
        iw = (Xp->width - x_off + 7) /8;
        scan_data = Xp->data;

        for (y = 0 ; y < Xp->height; y++, scan_data += dw)
        {
            PCL_command(NXLT("*b%dW"), iw);
            PfOutCString(scan_data + (x_off >> 3), iw);
        }
    }
    PCL_command(NXLT("*rB"));   /* End raster graphics */
}

/************************************************************************
 *
 *  Like D03MonoXImage, except use foreground/background colors
 *
 ************************************************************************/
static void D03MonoColorXImage(Xp, Xm, image_resolution, boundingBox, clip, x_pos, pen1, pen2)
    XImage *Xp, *Xm;
    int image_resolution;
    TmPoint32 boundingBox[] ;
    TmRect32 *clip ;
    int x_pos ;
    int pen1, pen2 ;
{
    int dw, iw, y, n, k;
    char *scan_data, *mask_data;
    char *compression_buffer;
    char *color_buffer;
    int rh, rw ;
    int x_off ;

    k = (int)MIL_TO_DOT(PF.PfLeftMarginMil) ;
    if (x_pos < k)
        x_off = k - x_pos ;
    else
        x_off = 0 ;

    x_off = (x_off >> 3) << 3 ;

    rh = MIL_TO_DOT(clip->y_position + clip->y_size) - boundingBox[0].y ;
    rh = MAX(0, MIN(rh, Xp->height)) ;

    rw = MIL_TO_DOT(clip->x_position + clip->x_size) - boundingBox[0].x ;
    rw = MAX(0, MIN(rw, Xp->width)) ;

    rw -= x_off ;
    if (rw <= 0)
        return ;

    PCL_command(NXLT("*r0F"));  /* raster image prints in direction of logical page. */
    PCL_command(NXLT("*t%dR"), image_resolution);
    
    PCL_command(NXLT("*r%dT"), rh);     /* height in raster rows of the image. */
    PCL_command(NXLT("*r%dS"), rw);     /* width in pixels of specified resolution. */
    PCL_command(NXLT("*r1A"));          /* start raster graphics at current cursor position */

    switch(PCL_compression_mode)
    {
    case PCL_COMPR_RLE:
        PCL_command(NXLT("*b1M"));      /* RLE compression transfer mode */

        dw = Xp->bytes_per_line;
        scan_data = Xp->data;
        mask_data = Xm ? Xm->data : NULL ;
        compression_buffer = (char *)malloc(2 * Xp->width);
        color_buffer = (char *)malloc(Xp->width);

        /* must take account of the image mask */
        for (y = 0 ; y < Xp->height ; y++, scan_data += dw)
        {
            colorize(color_buffer, scan_data, mask_data, Xp->width, pen1, pen2) ;
            n = run_length_encode((unsigned char *)compression_buffer,
									(unsigned char *)color_buffer + x_off, rw) ;
            PCL_command(NXLT("*b%dW"), n);
            PfOutCString(compression_buffer, n);
            if (mask_data)
                mask_data += Xm->bytes_per_line ;
        }
        free(compression_buffer);
        free(color_buffer);
        break;

    default:
    case PCL_COMPR_NONE:
        PCL_command(NXLT("*b0M"));  /* no compression transfer mode */

        /* note the difference between iw (=> how much we
        want to put out) and dw (=> the stride through the
        source data) */

        dw = Xp->bytes_per_line;
        iw = (rw + 7) /8;
        scan_data = Xp->data;

        for (y = 0 ; y < Xp->height ; y++, scan_data += dw)
        {
            PCL_command(NXLT("*b%dW"), iw);
            PfOutCString(scan_data, iw);
        }
    }
    PCL_command(NXLT("*rB"));   /* End raster graphics */
}

/************************************************************************
 *
 *
 ************************************************************************/
static void colorize(d, s, m, n, pen1, pen2)
    char *d, *s, *m ;
    int n ;
    int pen1, pen2 ;
{
    int k ;

    while (TRUE)
    {
        for (k = 0x80 ; k != 0 ; k >>= 1)
        {
            if (m && ((*m & k) == 0))
                *d++ = 0 ;
            else if (*s & k)
                *d++ = pen1 ;
            else
                *d++ = pen2 ; 
            if (--n == 0)
                return ;
        }
    if (m)
        m++ ;
    s++ ;
    }
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03ColorXImage(Xp, image_resolution, boundingBox, clip, x_pos)
    XImage *Xp;
    int image_resolution;
    TmPoint32 boundingBox[] ;
    TmRect32 *clip ;
{
    int dw, iw, y, n, k;
    char *scan_data;
    char *compression_buffer;
    int rh, rw ;
    int x_off ;

    k = (int)MIL_TO_DOT(PF.PfLeftMarginMil) ;
    if (x_pos < k)
        x_off = k - x_pos ;
    else
        x_off = 0 ;

    rh = MIL_TO_DOT(clip->y_position + clip->y_size) - boundingBox[0].y ;
    rh = MAX(0, MIN(rh, Xp->height)) ;

    rw = MIL_TO_DOT(clip->x_position + clip->x_size) - boundingBox[0].x ;
    rw = MAX(0, MIN(rw, Xp->width)) ;

    rw -= x_off ;
    if (rw <= 0)
        return ;

    PCL_command(NXLT("*r0F"));  /* raster image prints in direction of logical page. */
    PCL_command(NXLT("*t%dR"), image_resolution);
    
    PCL_command(NXLT("*r%dT"), rh);      /* height in raster rows of the image. */
    PCL_command(NXLT("*r%dS"), rw);     /* width in pixels of specified resolution. */
    PCL_command(NXLT("*r1A"));  /* start raster graphics at current cursor position */

    switch(PCL_compression_mode)
    {
    case PCL_COMPR_RLE:
        PCL_command(NXLT("*b1M"));  /* RLE compression transfer mode */
        dw = Xp->bytes_per_line;
        scan_data = Xp->data;
        compression_buffer = (char *)malloc(2 * dw);

        /* must take account of the image mask */
        for (y = 0; y < Xp->height; y++, scan_data += dw)
        {
            n = run_length_encode((unsigned char *)compression_buffer,
									(unsigned char *)scan_data + x_off, rw);
            PCL_command(NXLT("*b%dW"), n);
            PfOutCString(compression_buffer, n);
        }
        free(compression_buffer);
        break;

    default:
    case PCL_COMPR_NONE:
        PCL_command(NXLT("*b0M"));  /* no compression transfer mode */

        dw = Xp->bytes_per_line;
        scan_data = Xp->data;

        for (y=0; y<Xp->height; y++, scan_data += dw)
        {
            PCL_command(NXLT("*b%dW"), rw);
            PfOutCString(scan_data + x_off, rw);
        }
    }
    PCL_command(NXLT("*rB"));   /* End raster graphics */
}


/************************************************************************
 *
 *  D03GrayFill()
 *
 *  Calculate gray (0-100)
 *
 *  Either 
 *  (1) one color is white (or transparent)
 *  (2) the colors are arbitrary but the fill is either 100% fg or 100% bg.
 *
 *  select the pen.
 *
 ************************************************************************/
static int D03GrayFill(sattr, boundingBox, eoflag)
    sub_attr_type *sattr;
    TmPoint32 boundingBox[];
    int eoflag;
{

    int gf ;
    int local_gray_percent ;
    int c_diff, m_diff, y_diff, k_diff;
    int pen1, pen2 ;
    sub_attr_type tattr;

    tattr = *sattr ;

    if ( tattr.use_f_black )
    {
        tattr.fg.c = 0;
        tattr.fg.m = 0;
        tattr.fg.y = 0;
        tattr.fg.k = 255;
    }
    if ( tattr.use_b_white )
    {
        tattr.bg.c = 0;
        tattr.bg.m = 0;
        tattr.bg.y = 0;
        tattr.bg.k = 0;
    }

    if (tattr.fg.see_thru)
    {
        tattr.fg = sattr->bg ;
        tattr.bg = sattr->fg ;
        tattr.pat = 1000 - tattr.pat ;
    }
        
    tattr.mode = TMS ;
    if (tattr.bg.see_thru &&tattr.pat != 1000)
        tattr.mode = TMS_O_D ;

        
    pen1 = find_pen(tattr.fg.c, tattr.fg.m, tattr.fg.y, tattr.fg.k) ;
    pen2 = find_pen(tattr.bg.c, tattr.bg.m, tattr.bg.y, tattr.bg.k) ;

    if (palette[pen1].color || palette[pen2].color)
    {
        if (tattr.pat == 0 || pen1 == 0)
        {
            hpgl_SP(pen2) ;
            tattr.pat = 1000 - tattr.pat ;
        }
        else if (tattr.pat == 1000 || pen2 == 0)
            hpgl_SP(pen1) ;

        local_gray_percent = tattr.pat / 10 ;
    }

    else
    {
        c_diff = tattr.bg.c - tattr.fg.c;
        m_diff = tattr.bg.m - tattr.fg.m;
        y_diff = tattr.bg.y - tattr.fg.y;
        k_diff = tattr.bg.k - tattr.fg.k;

        gf = 1000 - tattr.pat;

        local_gray_percent = (int)(100 * IntensFromCMYKColor(
                                    tattr.fg.c + (c_diff * gf)/1000,
                                    tattr.fg.m + (m_diff * gf)/1000,
                                    tattr.fg.y + (y_diff * gf)/1000,
                                    tattr.fg.k + (k_diff * gf)/1000,
                                    255));

        if ((tattr.mode == TMS_O_D) && (local_gray_percent == 0))
            return -1;
        hpgl_SP(1) ;
    }

    /* Do transfer mode */
    switch (tattr.mode)
    {
    default:
    case TMS:
        hpgl_TR(FALSE);
        break;
    case TMS_O_D:
        hpgl_TR(TRUE);
        break;
    }

    return local_gray_percent ;      
}

/************************************************************************
 *
 *
 ************************************************************************/
static void  D03FillPattern(attr, boundingBox, clip, pol)
    attr_type *attr;
    TmPoint32 boundingBox[];
    TmRect32 *clip ;
    poly_type *pol;
{
    sub_attr_type sattr;
    int pen1, pen2, g ;

    /* extract the attributes */
    sattr.mode = attr->mode;
    sattr.type = attr->f_type;
    sattr.pat = attr->f_pat;
    sattr.ang = attr->f_ang;
    sattr.data = attr->f_data;
    sattr.dhyt = attr->f_dhyt;
    sattr.dwid = attr->f_dwid;
    sattr.ddep = attr->f_ddep;
    sattr.fill_abs = attr->fill_abs;
    sattr.xoff = attr->f_xoff;
    sattr.yoff = attr->f_yoff;
    sattr.fg = attr->ffg;
    sattr.bg = attr->fbg;
    sattr.use_f_black = attr->use_ffg;
    sattr.use_b_white = attr->use_fbg;

    /* compatability hacks... */
    if (sattr.fg.see_thru && sattr.bg.see_thru)
        return ;

    if (sattr.use_f_black)
    {
        sattr.fg.c = 0;
        sattr.fg.m = 0;
        sattr.fg.y = 0;
        sattr.fg.k = 255;
    }
    if (sattr.use_b_white)
    {
        sattr.bg.c = 0;
        sattr.bg.m = 0;
        sattr.bg.y = 0;
        sattr.bg.k = 0;
    }

    sattr.mode = TMS ;

    if (sattr.bg.see_thru)
        if (!((sattr.type == BUILTIN_FILL_TYPE && sattr.pat == 4) ||
                (sattr.type == GRAY_FILL_TYPE && sattr.pat == 1000)))
        sattr.mode = TMS_O_D ;


    if (sattr.fg.see_thru)
        if (!((sattr.type == BUILTIN_FILL_TYPE && sattr.pat == 0) ||
                (sattr.type == GRAY_FILL_TYPE && sattr.pat == 0)))
        sattr.mode = TMIS_O_D ;

    pen1 = find_pen(sattr.fg.c, sattr.fg.m, sattr.fg.y, sattr.fg.k) ;
    pen2 = find_pen(sattr.bg.c, sattr.bg.m, sattr.bg.y, sattr.bg.k) ;

    if (sattr.type == GRAY_FILL_TYPE && sattr.pat != 1000 && sattr.pat != 0 &&
/*          (sattr.mode == TMS_O_D || */
            (
/*              (pen1 > 1 && pen2 > 0) || (pen1 > 0 && pen2 > 1))) */
                (palette[pen1].color && pen2 > 0) || (pen1 > 0 && palette[pen2].color))) 
    {
        sattr.pat = getIXfromPercent(sattr.pat);
        sattr.type = BUILTIN_FILL_TYPE;
    }

    /* convert these to gray */
    else if (sattr.mode == TMS && sattr.type == BUILTIN_FILL_TYPE)
    {
        switch (sattr.pat)
        {
        case 0:                 /* 0% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 0;
            break;
        case 12:                /* 12.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 125;
            break;
        case 1:                 /* 25% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 250;
            break;
        case 10:                /* 37.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 375;
            break;
        case 2:                 /* 50% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 500;
            break;
        case 21:                /* 62.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 625;
            break;
        case 3:                 /* 75% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 750;
            break;
        case 13:                /* 87.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 875;
            break;
        case 4:                 /* 100% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 1000;
            break;
        }
    }
#if 0
    /* Do transfer mode */
    switch (sattr.mode)
    {
    case TMS:
        hpgl_TR(0); break;
    case TMS_O_D:
        hpgl_TR(1); break;
    default:
        hpgl_TR(0); break;
    }
#endif
    /* now do the fill */
    switch (sattr.type)
    {
    case BUILTIN_FILL_TYPE:
        D03BuiltinPatternFill(&sattr, pol, boundingBox, TRUE, 0);
        break;

    case USERDEF_FILL_TYPE:
        D03UserdefPatternFill(&sattr, pol, boundingBox, TRUE, 0);
        break;

    case LINEAR_FILL_TYPE:
    case BILINEAR_FILL_TYPE:
    case RADIAL_FILL_TYPE:
    case RECT_FILL_TYPE:
        if (sattr.mode == TMS) /* clear the area */
        {
            hpgl_FT(10, 0);
            HPGL_command(NXLT("FP;"));
        }
        D03GradientFill(&sattr, pol, boundingBox, clip, TRUE);
        break;

    case GRAY_FILL_TYPE:
        g = D03GrayFill(&sattr, boundingBox, TRUE);
        if (g >= 0)
        {
#if 1
            hpgl_AC(0, 0) ;
#else
            hpgl_AC(boundingBox[0].x, boundingBox[0].y) ;
#endif
            hpgl_FT(10, g);
            HPGL_command(NXLT("FP;"));
        }
        break;
    }

}


/*************************************************************************
 *
 * D03FillStroke()
 
 *  fill the polygon edge
 *
 ************************************************************************/
static void D03FillStroke(attr, boundingBox, pol)
    attr_type *attr;
    TmPoint32 boundingBox[];
    poly_type *pol;
{
    int c_diff, m_diff, y_diff, k_diff, gray_factor;
    int local_gray_percent;
    sub_attr_type sattr;
    int pen1, pen2, g ;

    if (attr->l_wid == 0)
        return;

    /* extract the attributes */
    sattr.mode = attr->mode;
    sattr.type = attr->l_type;
    sattr.pat = attr->l_pat;
    sattr.ang = attr->l_ang;
    sattr.data = attr->l_data;
    sattr.dhyt = attr->l_dhyt;
    sattr.dwid = attr->l_dwid;
    sattr.ddep = attr->l_ddep;
    sattr.fill_abs = attr->line_abs;
    sattr.xoff = attr->l_xoff;
    sattr.yoff = attr->l_yoff;
    sattr.fg = attr->lfg;
    sattr.bg = attr->lbg;
    sattr.use_f_black = attr->use_lfg;
    sattr.use_b_white = attr->use_lbg;

    if (sattr.fg.see_thru && sattr.bg.see_thru)
        return ;

    sattr.mode = TMS ;

    if (sattr.use_f_black)
    {
        sattr.fg.c = 0;
        sattr.fg.m = 0;
        sattr.fg.y = 0;
        sattr.fg.k = 255;
    }
    if (sattr.use_b_white)
    {
        sattr.bg.c = 0;
        sattr.bg.m = 0;
        sattr.bg.y = 0;
        sattr.bg.k = 0;
    }

    if (sattr.bg.see_thru)
        if (!((sattr.type == BUILTIN_FILL_TYPE && sattr.pat == 4) ||
                (sattr.type == GRAY_FILL_TYPE && sattr.pat == 1000)))
        sattr.mode = TMS_O_D ;


    if (sattr.fg.see_thru)
        if (!((sattr.type == BUILTIN_FILL_TYPE && sattr.pat == 0) ||
                (sattr.type == GRAY_FILL_TYPE && sattr.pat == 0)))
        sattr.mode = TMIS_O_D ;

    pen1 = find_pen(sattr.fg.c, sattr.fg.m, sattr.fg.y, sattr.fg.k) ;
    pen2 = find_pen(sattr.bg.c, sattr.bg.m, sattr.bg.y, sattr.bg.k) ;

    if (sattr.type == GRAY_FILL_TYPE && sattr.pat != 1000 && sattr.pat != 0 &&
/*          (sattr.mode == TMS_O_D || */
            (
/*              (pen1 > 1 && pen2 > 0) || (pen1 > 0 && pen2 > 1))) */
                (palette[pen1].color && pen2 > 0) || (pen1 > 0 && palette[pen2].color)))
    {
        sattr.pat = getIXfromPercent(sattr.pat);
        sattr.type = BUILTIN_FILL_TYPE;
    }

    /* convert these to gray */
    else if (sattr.mode == TMS && sattr.type == BUILTIN_FILL_TYPE)
    {
        switch (sattr.pat)
        {
        case 0:                 /* 0% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 0;
            break;
        case 12:                /* 12.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 125;
            break;
        case 1:                 /* 25% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 250;
            break;
        case 10:                /* 37.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 375;
            break;
        case 2:                 /* 50% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 500;
            break;
        case 21:                /* 62.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 625;
            break;
        case 3:                 /* 75% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 750;
            break;
        case 13:                /* 87.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 875;
            break;
        case 4:                 /* 100% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 1000;
            break;
        }
    }

    if (!sattr.bg.see_thru)
        hpgl_TR(0);

    switch (sattr.type)
    {
    case BUILTIN_FILL_TYPE:
        D03BuiltinPatternFill(&sattr, pol, boundingBox, TRUE, 1);
        break ;

    case USERDEF_FILL_TYPE:
        D03UserdefPatternFill(&sattr, pol, boundingBox, TRUE, 1);
        break;

    case LINEAR_FILL_TYPE:
    case RADIAL_FILL_TYPE:
    case RECT_FILL_TYPE:
#if 0
        c_diff = sattr.bg.c - sattr.fg.c;
        m_diff = sattr.bg.m - sattr.fg.m;
        y_diff = sattr.bg.y - sattr.fg.y;
        k_diff = sattr.bg.k - sattr.fg.k;
        D03SetCMYKColor(sattr.fg.c + c_diff/2,
                        sattr.fg.m + m_diff/2,
                        sattr.fg.y + y_diff/2,
                        sattr.fg.k + k_diff/2);

        hpgl_SV(1, (int)(COLOR_GRAY_PERCENT(100.0)));
#endif
        /* can't do the real thing - try 50% fill of the two colors */
        sattr.type = BUILTIN_FILL_TYPE;
        sattr.pat = getIXfromPercent(500);
        D03BuiltinPatternFill(&sattr, pol, boundingBox, TRUE, 1);
        break;

    case GRAY_FILL_TYPE:
        g = D03GrayFill(&sattr, boundingBox, TRUE) ;
        if (g >= 0)
        {
            if (g < 99)
                hpgl_SV(1, g);
            else
                hpgl_SV(0);
        }
        break;
    }
    if (HPGL_UsePolygonMode)
        HPGL_command(NXLT("EP;"));
}

#if 0
/************************************************************************
 *
 *
 ************************************************************************/
static void ellipsePath(bbox, startAng, arcAng, toBeFilled)
    TmRect32 *bbox;
    int startAng, arcAng;
    int toBeFilled; /* bool */
{
    double xAxis, yAxis, centerX, centerY;

    xAxis = bbox->x_size;
    yAxis = bbox->y_size;
    centerX = bbox->x_position + (bbox->x_size / 2);
    centerY = bbox->y_position + (bbox->y_size / 2);

    /* if it's a circle, use the arc call */
    if ( xAxis == yAxis )
    {
        D03SetPos((int)centerX, (int)centerY);
        /* if filled, moveto center to create 'wedge edge' in the path */
        if ( toBeFilled && arcAng < 3600 )
        {
/*          put("%.3f %.3f M\n", centerX, Y(centerY)); */
        }
        /* now the rest of the arc */

        /* if filled, do the other 'wedge edge' */
        if (toBeFilled && arcAng < 3600)
        {
/*          put("%.3f %.3f L\n", centerX, Y(centerY));  */
        }

        if (toBeFilled)
        {
            HPGL_command(NXLT("FW%3.3f,%d,%d;"), yAxis, startAng, arcAng);
        }
        else
        {
            HPGL_command(NXLT("EW%3.3f,%d,%d;"), yAxis, startAng, arcAng);
        }
    }
    else
    {
        /* ellipse, not yet... */
    }
}

/************************************************************************
 *
 *
 ************************************************************************/
int D03Ellipse(bbox, pol, start_angle, arc_angle, attr)
    TmRect32 *bbox;
    poly_type pol;
    int start_angle;
    int arc_angle;
    attr_type *attr;
{
    TmPoint32 boundingBox[2];

    /* create the path */
    ellipsePath(bbox, start_angle, arc_angle, attr->fill_flag);

    /* init boundary */
    boundingBox[0].x = bbox->x_position;
    boundingBox[0].y = bbox->y_position;
    boundingBox[1].x = bbox->x_position + bbox->x_size;
    boundingBox[1].y = bbox->y_position + bbox->y_size;

    /* fill the path according to pattern info passed in */
    if ( attr->fill_flag )
        D03FillPattern(attr, boundingBox, pol);

    /* stroke the path according to line info passed in */
    if ( attr->l_wid )
        D03FillStroke(attr, boundingBox, pol);


    return(0);
}
#endif


/************************************************************************
 *
 *
 ************************************************************************/
static int DecodePCLPrintName(ftab, txProps, HPGL_font_p)
    fontTabEnt *ftab;
    textAttrs *txProps;
    HPGL_font_desc *HPGL_font_p;
{
    char *symbolSet;
    char PCLPrintNameCopy[32];
    char *comp_del_ptr = PCLPrintNameCopy;
    char *comp_val_ptr = PCLPrintNameCopy;
    char *PCL_style;
    char *PCL_weight;
    char *PCL_typeface;
    int style_int;

    symbolSet = pf03SymbolSetSequence(ftab);
    pf03PCLSymSeqtoHPGL(symbolSet, HPGL_font_p);

    if (ftab->fixedWidth)
    {
        strcpy(HPGL_font_p->font_spacing, "0");
        sprintf(HPGL_font_p->pitch,
                NXLT("%.3f"),
                pitchFromPointsize(txProps->pointsize, ftab));
    }
    else
    {
        strcpy(HPGL_font_p->font_spacing, "1");
        sprintf(HPGL_font_p->height,
                NXLT("%.3f"),
                (double)(txProps->pointsize / 10000.0));
    }

    strcpy(PCLPrintNameCopy, D03PrintName(ftab));
    comp_del_ptr = PCLPrintNameCopy;
    comp_del_ptr++;

    while(*comp_del_ptr != '\0')
    {
        switch (*comp_del_ptr)
        {
        case 's':
            *comp_del_ptr = '\0';       /* cauterize! */
            PCL_style = comp_val_ptr;
            comp_del_ptr++;             /* look for next token */
            comp_val_ptr = comp_del_ptr; /* start again */

            style_int = atoi(PCL_style);
            switch(style_int)
            {
            default:
            case 0:             /* upright */
                strcpy(HPGL_font_p->posture, "0");
                break;
            case 1:             /* italic */
                strcpy(HPGL_font_p->posture, "1");
                break;
            case 4:             /* condensed */ /* not real HPGL/2 */
                strcpy(HPGL_font_p->posture, "4");
                break;
            case 5:             /* condensed italic */
                strcpy(HPGL_font_p->posture, "5");
                break;
            case 8:             /* compressed, or extra condensed */
                strcpy(HPGL_font_p->posture, "8");
                break;
            case 24:            /* expanded */
                strcpy(HPGL_font_p->posture, "0");
                break;
            case 32:            /* outline */
                strcpy(HPGL_font_p->posture, "0");
                break;
            case 64:            /* inline */
                strcpy(HPGL_font_p->posture, "0");
                break;
            case 128:           /* shadowed */
                strcpy(HPGL_font_p->posture, "0");
                break;
            case 160:           /* outline shadowed */
                strcpy(HPGL_font_p->posture, "0");
                break;
            }
            break;

        case 'b':
            /* These are 1:1 between PCL and HPGL */
            *comp_del_ptr = '\0';
            PCL_weight = comp_val_ptr;
            comp_del_ptr++;
            comp_val_ptr = comp_del_ptr;

            strcpy(HPGL_font_p->stroke_weight, PCL_weight);
            break;

        case 'T':
            *comp_del_ptr = '\0';
            PCL_typeface = comp_val_ptr;
            comp_del_ptr++;
            comp_val_ptr = comp_del_ptr;

            strcpy(HPGL_font_p->typeface, PCL_typeface);
            break;
        }
        comp_del_ptr++;
    }
    return(!ftab->fixedWidth);
}

/************************************************************************
 *
 * pf03PCLSymSeqtoHPGL - This is a hack which attempts to
 * map a PCL symbol set selection sequence into an equivalent
 * HPGL one.    (a pity that the two domains have totally
 * different naming conventions.. that's life).
 *
 * Our problem here is that PCL printers report their typeface
 * info in terms of PCL selection strings, not HPGL;    however,
 * we need to use HPGL for text setting.    There are 100's
 * of these sequences defined (at least); only a few of those that
 * are supported in the .tfm files we distribute are included here.
 *
 ************************************************************************/
static int pf03PCLSymSeqtoHPGL(symbolSet, HPGL_font_p)
    char *symbolSet;
    HPGL_font_desc *HPGL_font_p;
{
    int ix;

    for (ix = 0; ix < NUM_HP_SYMBOL_SET_STRINGS; ix++) {
        if (strcmp(symbolSet, D03SymbolSetStrings[ix][0]) == 0) {
            strcpy(HPGL_font_p->symbol_set,
                    D03SymbolSetStrings[ix][1]);
            return TRUE;
        }
    }
    strcpy(HPGL_font_p->symbol_set, "14");  /* hpgl for latin-1 */
    return FALSE;
}



#if AXMS                            /*  {  */
static int download_to_PCL( fontTabEnt *ftab )
{
    char *p;

    if ( (ftab->PCLPrintName != (char *)NULL) )
        /* substitute some other font */
        return( FALSE );
    p = ftab->targetMetricsNames[0];
    if ( !p || (strlen(p) < 4) )
        return( FALSE );

    p = p + strlen(p) - 4;
    if ( strcmp(p, ".win") != 0 )
        /* not a special truetype metrics file name as we construct them */
        return( FALSE );

    return( TRUE );
}
#endif                            /*  }  */





/************************************************************************
 *
 *
 *
 ************************************************************************/
int D03AxText(textstr, txlen, pos, width, fontName,
            underPos, underThick, strikePos,
            txProps, fieldProps, xZoom, yZoom, clip)
    tx_char *textstr;
    register int txlen;
    p32 *pos;               /* position in mils */
    int width;              /* setwidth in mils */
    char *fontName;
    int underPos, underThick, strikePos;
    textAttrs *txProps;     /* properties for this segment of text */
    textAttrs *fieldProps;  /* props of the overall field of text */
    int xZoom, yZoom;       /* zoom factor relative to ref val 100 */
    TmRect32 *clip ;
{
    char *errString, *symbolSet, *narrowStr;
    register int ix;
    int xpos, ypos, tmpx, tmpy, milx, mily, seglen;
    int lastx, lasty, lx;
    int xpos_mil, ypos_mil;
    int lbearAdjust;
    int unity;
    int retval = 0;
    axform xf, sxf;
    axformTerms xfTerms;
    fontTabEnt *ftab, *altRec;
    attr_type *fill_attr;
    HPGL_font_desc HPGL_font;
    axtext txt;
    textAttrs sfieldProps, stxProps;
    PF_soft *soft ;
    int angle ;

    if (txlen <= 0)
        return(0);

/* Must handle bogus (empty?) font names */
    if ((fontName == NULL) || (txProps == NULL) || (fieldProps == NULL))
    {
        /* these should really be set to defaults... */
        return(0);
    }

    D03Clip();
    clip_box(clip) ;

    xpos_mil = pos->x;
    ypos_mil = pos->y;
    xpos = xmiltopix(pos->x);
    ypos = ymiltopix(pos->y);
    D03milSetPos(xpos_mil, ypos_mil);

    hpgl_SP(find_pen(axCyanComponent(txProps->fillcolor),
                    axMagentaComponent(txProps->fillcolor),
                    axYellowComponent(txProps->fillcolor),
                    axBlackComponent(txProps->fillcolor)));


    /* do line_attr too? */
	fill_attr = txProps->fillpattern;

    /* ignore if this all transparent */
    if (pfIsTransparent(fill_attr))
        return ;

	D03CharacterFill(fill_attr, pos) ;


	/* combine whole-field and segment xforms plus zoom, onto xf */
	unity = pfGetTextSegTransform(txProps, fieldProps, xZoom, yZoom, &xf);

	/* extract physical terms from that composite, onto xfTerms */
	pfGetTransformTerms(&xf, &xfTerms);

	PF_MEM_CHECK;

	/* see whether this font is based on .tfm metrics */
	if (errString = axGetFontRecord(fontName, &ftab))
	{
		freeErrString(errString);
		retval = 0;
		goto leave_D03AxText;
	}

    if (!ftab->PCLAlternate)
    {
            /* we assume that the font is available on the
                output device (whether or not true); output directly */

#if AXMS
        if ( download_to_PCL( ftab ) )
#else
        if (ftab->gallium)
#endif
        {
            soft_xform(&sfieldProps, fieldProps) ;
            soft_xform(&stxProps, txProps) ;
            pfGetTextSegTransform(&stxProps, &sfieldProps, xZoom, yZoom, &sxf);
            soft = pfSoftFontRegister(ftab, txProps->pointsize, &sxf.forward) ;
            HPGL_command(NXLT("SB1;"));
            HPGL_command(NXLT("FI%d;"), soft->id) ;
            HPGL_command(NXLT("SS;"));
            pf03CurFont.pointsize = 0 ;         /* next time it's a different font */
        }

        else if (pf03DifferentFont(txProps))
        {

            memset(&HPGL_font, 0, sizeof(HPGL_font_desc));
            if (DecodePCLPrintName(ftab, txProps, &HPGL_font))
            {
                /* non-fixed pitch... use only scalables */
                HPGL_command(NXLT("SB0;"));
                HPGL_command(NXLT("SD1,%s,2,%s,4,%s,5,%s,6,%s,7,%s;"),
                    HPGL_font.symbol_set,       /* 1 */
                    HPGL_font.font_spacing,     /* 2 */
                    HPGL_font.height,           /* 4 */
                    HPGL_font.posture,          /* 5 */
                    HPGL_font.stroke_weight,    /* 6 */
                    HPGL_font.typeface);        /* 7 */
                HPGL_command(NXLT("SS;"));
            }
            else
            {
                /* fixed pitch font... use bitmapped too */
#if 1
                HPGL_command(NXLT("SB1;"));
#else
                HPGL_command(NXLT("SB0;"));
#endif
                HPGL_command(NXLT("SD1,%s,2,%s,3,%s,5,%s,6,%s,7,%s;"),
                    HPGL_font.symbol_set,       /* 1 */
                    HPGL_font.font_spacing,     /* 2 */
                    HPGL_font.pitch,            /* 3 */
                    HPGL_font.posture,          /* 5 */
                    HPGL_font.stroke_weight,    /* 6 */
                    HPGL_font.typeface);        /* 7 */
                HPGL_command(NXLT("SS;"));
            }
        }

        /* output the text */

        if (!unity)
            pf03SetTextTransform(txProps, &xfTerms, ftab->gallium);
            
        /* %%%%%%%%%%%%%WIDECHAR %%%%%%%*/
        narrowStr = (char *) malloc(txlen + 1);
        for (ix = 0; ix < txlen; ix++)
        {
            if (textstr[ix] == NONBREAK_SPACE)
                narrowStr[ix] = SPACE;
            else
                narrowStr[ix] = (char) textstr[ix];
        }
        narrowStr[txlen] = '\0';            /* paranoia! */

        /* position each word independently */

        /* get full target metrics info for the text */
        memset((char *) &txt, 0, sizeof(axtext));
        axtDefaultText(&txt);               /* set hardcoded defs */
        insChars(&txt, textstr, 0, txlen);  /* add text */

        setTextAttrs(&txt, txProps, (unsigned int) TXPrintAtts, 0, txlen);
        axtSetFieldTransform(&txt, fieldProps, (unsigned int)
                             (TXangle | TXxscale | TXyscale | TXxshear));
		axtSetZoom(&txt, xZoom, yZoom) ;
        constructForPrint(&txt);    /* get target metrics info onto text */

        HPGL_command(NXLT("PA%.2f,%.2f;"), X_DOT_TO_PLU(xpos), Y_DOT_TO_PLU(ypos));

        milx = lastx = pos->x;
        mily = lasty = pos->y;
        seglen = 0;
        for (ix = lx = 0; ix < txlen; ix++)
        {
            if (PFWHITE(txt.items[ix].charcode))
            {
                if (seglen > 0)
                {
#if 0
                    if (unity && (lx < txlen - 1))
                    {
                        /* in HPGL, seems we need to account for left
                            bearings??  do so for the simple case */
                        lbearAdjust = txt.items[lx + 1].orig.x;
                        if (lbearAdjust > 0)
                            lbearAdjust = 0;    /* adjust only if negative */
                    }
                    else
#endif
                        lbearAdjust = 0;
                    
                    tmpx = xmiltopix(lastx + lbearAdjust);
                    tmpy = ymiltopix(lasty);
                    if (tmpx >= 0 && tmpy >= 0)
                    {
                            /* set the position to tmpx, ypos */
#if USE_PLU_IN_HPGL
                            HPGL_command(NXLT("PA%.2f,%.2f;"),
                                X_DOT_TO_PLU(tmpx),
                                Y_DOT_TO_PLU(tmpy));
/*
                                X_MIL_TO_PLU(tmpx),
                                Y_MIL_TO_PLU(tmpy));
*/
#else
                            HPGL_command(NXLT("PA%d,%d;"), tmpx, tmpy);
#endif
                    }
                    
                    HPGL_command(NXLT("LB"));
                    PfOutCString(&narrowStr[lx], seglen);
                    put(NXLT("\003;")); /* default label terminator */

                    lastx = milx;
                    lasty = mily;
                    lx = ix;
                    seglen = 0;
                }
            }

            milx += txt.items[ix].advance.x;
            mily += txt.items[ix].advance.y;

            seglen++;
        }
        if (seglen > 0)
        {
#if 0
            if (unity && (lx < txlen - 1))
            {
                /* in HPGL, seems we need to account for left
                    bearings??  do so for the simple case */
                lbearAdjust = txt.items[lx + 1].orig.x;
                if (lbearAdjust > 0)
                    lbearAdjust = 0;    /* adjust only if negative */
            }
            else
#endif
                lbearAdjust = 0;
            
            tmpx = xmiltopix(lastx + lbearAdjust);
            tmpy = ymiltopix(lasty);
            if (tmpx >= 0 && tmpy >= 0)
            {
                    /* set the position to tmpx, ypos */
                    HPGL_command(NXLT("PA%.2f,%.2f;"),
                        X_DOT_TO_PLU(tmpx),
                        Y_DOT_TO_PLU(tmpy));
/*
                        X_MIL_TO_PLU(tmpx),
                        Y_MIL_TO_PLU(tmpy));
*/
            }
                    
            HPGL_command(NXLT("LB"));
            PfOutCString(&narrowStr[lx], seglen);
            put(NXLT("\003;")); /* default label terminator is \003 */
        }

        if (!unity)
            pf03ResetTextTransform(&xfTerms);

        daxtext(&txt);
        free(narrowStr);
        PF_MEM_CHECK;
        narrowStr = NULL;   /* SJP */
    }
    else
    {
            /* this is the print-bash code; set the text in
                a substitue font, adjusting layout to reflect
                the ideal font's metrics */
        textAttrs altAtts;
        axtext alt; /* represents passed info, in substitue font */
        int *deltas = (int *) NULL;

        PF_MEM_CHECK;
        if (!ftab->PCLAlternate)
        {
            /* try for default */
            if (errString = axGetFontRecord(XLT("axDefaultFontRecord"),
                                            &altRec))
            {
                freeErrString(errString);
                goto leave_D03AxText;   /* drop if all fails */
            }
        }
        /* get the alternate's record */
        if (errString = axGetFontRecord(ftab->PCLAlternate, &altRec))
        {
            freeErrString(errString);
            retval = 0;
            goto leave_D03AxText;
        }

        /* get the deltas array describing layout adjustments */
        memset((char *) &alt, 0, sizeof(axtext));
        deltas = pfGetTextLayoutDeltas(ftab, textstr, txlen, width, txProps, altRec, &alt);
        if (!deltas)
        {
            daxtext(&alt);
            retval = 0;
            goto leave_D03AxText;
        }

        /* get the alternate's attrs to test whether to select a new font */
        memset((char *) &altAtts, 0, sizeof(textAttrs));
        altAtts.fontFamily = alt.atts[0].fontAttr.familyName;
        altAtts.pointsize = (int) (alt.atts[0].fontAttr.ptsize * 10000.0);
        altAtts.weight = alt.atts[0].fontAttr.weight;
        altAtts.slant = alt.atts[0].fontAttr.slant;

        if (pf03DifferentFont(&altAtts))
        {
            /* select the font from the alternate record */
            memset(&HPGL_font, 0, sizeof(HPGL_font_desc));
            if (DecodePCLPrintName(altRec, txProps, &HPGL_font))
            {
                HPGL_command(NXLT("SB0;"));
                HPGL_command(NXLT("SD1,%s,2,%s,4,%s,5,%s,6,%s,7,%s;"),
                    HPGL_font.symbol_set,           /* 1 */
                    HPGL_font.font_spacing,         /* 2 */
                    HPGL_font.height,               /* 4 */
                    HPGL_font.posture,              /* 5 */
                    HPGL_font.stroke_weight,        /* 6 */
                    HPGL_font.typeface);            /* 7 */
            }
            else
            {
                HPGL_command(NXLT("SB0;"));
                HPGL_command(NXLT("SD1,%s,2,%s,3,%s,5,%s,6,%s,7,%s;"),
                    HPGL_font.symbol_set, /* 1 */
                    HPGL_font.font_spacing, /* 2 */
                    HPGL_font.pitch,        /* 3 */
                    HPGL_font.posture,  /* 5 */
                    HPGL_font.stroke_weight, /* 6 */
                    HPGL_font.typeface);     /* 7 */
            }
            HPGL_command(NXLT("SS;"));
        }

        D03HPGL_delta_text(textstr, txlen, &alt, pos->x, pos->y,
                            deltas, unity, &altAtts, &xf, &xfTerms);

        if (deltas)
        {
            free(deltas);
            PF_MEM_CHECK;
            deltas = NULL;
        }
        daxtext(&alt);
    }
    retval = 0;

leave_D03AxText:

    PF_MEM_CHECK;
    return(retval);
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03CharacterFill(fill_attr, pos)
    attr_type *fill_attr;
    p32 *pos ;
{
    int c_diff, m_diff, y_diff, k_diff, gray_factor;
    int local_gray_percent = 100;
    sub_attr_type sattr;
    TmPoint32 box[2];
    int pen1, pen2, g ;

    /* cur_grayness got from SetCMYKColor() */
    local_gray_percent = (int)(cur_grayness * 100.0);

    if (fill_attr == NULL || fill_attr->f_pat == -1)
    {
        character_fill(local_gray_percent) ;
        return ;
    }

    box[0].x = box[1].x = pos->x;
    box[0].y = box[1].y = pos->y;

    /* extract the attributes */
    sattr.mode = fill_attr->mode;
    sattr.type = fill_attr->f_type;
    sattr.pat = fill_attr->f_pat;
    sattr.ang = fill_attr->f_ang;
    sattr.data = fill_attr->f_data;
    sattr.dhyt = fill_attr->f_dhyt;
    sattr.dwid = fill_attr->f_dwid;
    sattr.fill_abs = fill_attr->fill_abs;
    sattr.xoff = fill_attr->f_xoff;
    sattr.yoff = fill_attr->f_yoff;
    sattr.fg = fill_attr->ffg;
    sattr.bg = fill_attr->fbg;
    sattr.use_f_black = fill_attr->use_ffg;
    sattr.use_b_white = fill_attr->use_fbg;

    if (sattr.fg.see_thru && sattr.bg.see_thru)
        return ;

    if (!(sattr.fg.see_thru || sattr.bg.see_thru))
        sattr.mode = TMS ;
    else
    {
        sattr.mode = TMS_O_D ;

        if (sattr.fg.see_thru)
        {
            sattr.mode = TMIS_O_D;
            /* flip the sense of the colors */
#if 0
            sattr.fg = attr->fbg;
            sattr.bg = attr->ffg;
#endif
        }
    }

    pen1 = find_pen(sattr.fg.c, sattr.fg.m, sattr.fg.y, sattr.fg.k) ;
    pen2 = find_pen(sattr.bg.c, sattr.bg.m, sattr.bg.y, sattr.bg.k) ;

    if (sattr.type == GRAY_FILL_TYPE && sattr.pat != 1000 && sattr.pat != 0 &&
/*          (sattr.mode == TMS_O_D || */
            (
/*              (pen1 > 1 && pen2 > 0) || (pen1 > 0 && pen2 > 1))) */
                (palette[pen1].color && pen2 > 0) || (pen1 > 0 && palette[pen2].color)))
    {
        sattr.pat = getIXfromPercent(sattr.pat);
        sattr.type = BUILTIN_FILL_TYPE;
    }

    /* convert these to gray */
    else if (sattr.mode == TMS && sattr.type == BUILTIN_FILL_TYPE)
    {
        switch (sattr.pat)
        {
        case 0:                 /* 0% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 0;
            break;
        case 12:                /* 12.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 125;
            break;
        case 1:                 /* 25% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 250;
            break;
        case 10:                /* 37.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 375;
            break;
        case 2:                 /* 50% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 500;
            break;
        case 21:                /* 62.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 625;
            break;
        case 3:                 /* 75% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 750;
            break;
        case 13:                /* 87.5% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 875;
            break;
        case 4:                 /* 100% */
            sattr.type = GRAY_FILL_TYPE;
            sattr.pat = 1000;
            break;
        }
    }

    switch (sattr.type)
    {
    case BUILTIN_FILL_TYPE:
        D03BuiltinPatternFill(&sattr, NULL, box, TRUE, 2) ;
        break;

    case USERDEF_FILL_TYPE:
        D03UserdefPatternFill(&sattr, NULL, box, TRUE, 2);
        break;

    case LINEAR_FILL_TYPE:
    case RADIAL_FILL_TYPE:
    case RECT_FILL_TYPE:
#if 0
        c_diff = sattr.bg.c - sattr.fg.c;
        m_diff = sattr.bg.m - sattr.fg.m;
        y_diff = sattr.bg.y - sattr.fg.y;
        k_diff = sattr.bg.k - sattr.fg.k;
        D03SetCMYKColor(sattr.fg.c + c_diff/2,
                        sattr.fg.m + m_diff/2,
                        sattr.fg.y + y_diff/2,
                        sattr.fg.k + k_diff/2);

        g = (int)COLOR_GRAY_PERCENT(100.0) ;
        hpgl_FT(10, g);
#endif
        sattr.type = BUILTIN_FILL_TYPE;
        sattr.pat = getIXfromPercent(500);
        D03BuiltinPatternFill(&sattr, NULL, box, TRUE, 2) ;
        break;

    case GRAY_FILL_TYPE:
        g = D03GrayFill(&sattr, box, TRUE) ;
        if (g >= 0)
        {
            if (g < 99)
                hpgl_FT(10, g);
            else
                hpgl_FT(1);
        }
        break;
    }

    HPGL_cfill = -1 ;
    HPGL_command(NXLT("CF3,0;"));   /* Character Fill */
}

/************************************************************************
 *
 * output the text using the deltas array to adjust layout
 *
 ************************************************************************/
static int D03HPGL_delta_text(textstr, txlen, txt, xpos_mil, ypos_mil,
                    deltas, unity, txProps, xf, xfTerms)
    tx_char *textstr;
    register int txlen;
    axtext *txt;
    int xpos_mil;
    int ypos_mil;
    int *deltas;
    int unity;
    textAttrs *txProps;
    axform *xf;
    axformTerms *xfTerms;
{
    int ix;
    int jx;
    int kx;
    int milx = xpos_mil;
    int mily = ypos_mil;
    int lbearAdjust;
    int seglen = 0;
    char *narrowStr;
    p32 xfdelta, xfadv;


    if (!unity)
        pf03SetTextTransform(txProps, xfTerms, FALSE);

    for (ix = 0, jx = 0; ix < txlen; ix++)
    {
        seglen++;
        if ((deltas != NULL) && deltas[ix])
        {
            /* set text segment, adjust position */
/* %%%%%%%%%%%%%WIDECHAR %%%%%%%*/
            narrowStr = (char *) malloc(seglen + 1);
            for (kx = 0; kx < seglen; kx++)
                narrowStr[kx] = (char) textstr[jx + kx];
            HPGL_command(NXLT("LB"));
            PfOutCString(narrowStr, (short)seglen);
            PfOutString(NXLT("\003"));
            free(narrowStr);
            PF_MEM_CHECK;
            narrowStr = NULL;
            jx = ix + 1;
            seglen = 0;

            /* add the transformed delta and advance */
            forwardSXRT(deltas[ix], 0, &xfdelta.x, &xfdelta.y, xf);
            milx += xfdelta.x;
            mily += xfdelta.y;

            forwardSXRT(txt->items[ix].padvance, 0, &xfadv.x, &xfadv.y, xf);
            milx += xfadv.x;
            mily += xfadv.y;

            if (unity && (ix < txlen - 1))
            {
                /* in HPGL, seems we need to account for left
                bearings??  do so for the simple case */
                lbearAdjust = txt->items[ix + 1].orig.x;
                if (lbearAdjust > 0)
                    lbearAdjust = 0;    /* adjust only if negative */
            }
            else
            {
                lbearAdjust = 0;
            }
#if 1
            HPGL_command(NXLT("PR%.2f,%.2f;"),
                MIL_TO_PLU(xfdelta.x + lbearAdjust),
                MIL_TO_PLU(-xfdelta.y));
#else
#if USE_PLU_IN_HPGL
            HPGL_command(NXLT("PA%.2f,%.2f;"),
                X_MIL_TO_PLU(milx + lbearAdjust),
                Y_MIL_TO_PLU(mily));
#else
            tmpx = xmiltopix(milx + lbearAdjust);
            tmpy = ymiltopix(mily);
            
            if (tmpx >= 0 && tmpy >= 0)
            {
                /* set the position to tmpx, ypos */
                HPGL_command(NXLT("PA%d,%d;"), tmpx, tmpy);
            }
#endif
#endif
        }
        else
        {
            /* add the transformed advance */
            forwardSXRT(txt->items[ix].padvance, 0, &xfadv.x, &xfadv.y, xf);
            milx += xfadv.x;
            mily += xfadv.y;
        }
    }

    if (seglen > 0)
    {
        /* output the caboose */
/* %%%%%%%%%%%%%WIDECHAR %%%%%%%*/
        narrowStr = (char *) malloc(seglen + 3);
        for (kx = 0; kx < seglen; kx++)
            narrowStr[kx] = (char) textstr[jx + kx];
        narrowStr[seglen] = '\0';               /* paranoia! */
        
        HPGL_command(NXLT("LB"));
        PfOutCString(narrowStr, (short) seglen);
        put(NXLT("\003;"));
        free(narrowStr);
        PF_MEM_CHECK;
        narrowStr = NULL;
    }

    if (!unity)
        pf03ResetTextTransform(xfTerms);
}


/************************************************************************
 *
 *
 ************************************************************************/
static int pf03DifferentFont(atts)
    textAttrs *atts;    /* for the text about to be set */
{
    if (atts->weight != pf03CurFont.weight ||
        atts->slant != pf03CurFont.slant ||
        atts->pointsize != pf03CurFont.pointsize ||
        !pf03CurFont.fontFamily ||
        strcmp(atts->fontFamily, pf03CurFont.fontFamily))
    {
        /* a font change is needed; update pf03CurFont */
        if (pf03CurFont.fontFamily)
        {
            free(pf03CurFont.fontFamily);
            PF_MEM_CHECK;
            pf03CurFont.fontFamily = NULL;
        }
        pf03CurFont.fontFamily = (char *) malloc(strlen(atts->fontFamily) + 1);
        strcpy(pf03CurFont.fontFamily, atts->fontFamily);
        pf03CurFont.weight = atts->weight;
        pf03CurFont.slant = atts->slant;
        pf03CurFont.pointsize = atts->pointsize;
        return TRUE;
    }
    return FALSE;
}

/************************************************************************
 *
 *
 ************************************************************************/
static double pitchFromPointsize(pointsize, fontRec)
    int pointsize;  /* in 10000ths point units */
    fontTabEnt *fontRec;
{
    fontMetrics *fm;
    char *errString;
    double dpoints, milliemWid;

    if(errString = axGetTargetMetrics(fontRec->targetMetricsNames[0], fontRec, &fm))
    {
        freeErrString(errString);
        return 10.0;
    }
    milliemWid = (double) fm->xfont->min_bounds.width;
    dpoints = ((double) pointsize) / 10000.0;
    return ( 72000.0 / (milliemWid * dpoints));
}

/************************************************************************
 *
 *
 ************************************************************************/
static char *pf03SymbolSetSequence(fontRec)
    fontTabEnt *fontRec;    /* tfm-based font record */
{
    int ix, len;
    char *errString;
    fontMetrics *fm;
    char *tfmName = (char *) NULL;
    char *symString = (char *) NULL;

    /* find the name of the .tfm metrics file */
    if (fontRec->numTargetMetricsNames == 1)
        tfmName = fontRec->targetMetricsNames[0];
    else
    {
        for (ix = 0; ix < fontRec->numTargetMetricsNames; ix++)
        {
            len = strlen(fontRec->targetMetricsNames[ix]);
            if (len >= 4)
            {
                if (strncmp((char *)&fontRec->targetMetricsNames[len - 4],
                            ".tfm", 4) == 0)
                {
                    tfmName = fontRec->targetMetricsNames[ix];
                    break;
                }
            }
        }
    }
    if (tfmName)
    {
        /* lookup the target metrics record, which should have
                the symbol set id on it */
        errString = axGetTargetMetrics(tfmName, fontRec, &fm);
        if (errString)
            freeErrString(errString);
        else if (fm->PCLSymbolSetID)
            symString = fm->PCLSymbolSetID;
    }
    if (!symString)
    {
        if (!strcmp(fontRec->globalName, NXLT("Symbol")))
        {
            /* Hack. Supply seq for the truetype symbol font.. */
            symString = NXLT("19M");
        }
        else
        {
            symString = "0N";   /* ECMA-94 Latin 1 */
        }
    }
    return (symString);
}



/************************************************************************
 *
 *      bytes_this_line = rle(rle_buffer, scan_data, iw)
 *
 ************************************************************************/
static int run_length_encode(out_str, in_str, bytes_rem)
    unsigned char *out_str;
    unsigned char *in_str;
    int bytes_rem;
{
    unsigned char current_byte;
    int in_index = 0;
    int out_index = 0;
    int repetitions = 0;

    /* do 1 & 2 byte cases... */

    current_byte = in_str[in_index];
    bytes_rem--;
    in_index++;
    while (bytes_rem > 0)
    {
        while ((current_byte == in_str[in_index]) &&
                (bytes_rem > 0) &&
                (repetitions <255))
        {
            repetitions++;
            bytes_rem--;
            in_index++;
        }
        if (bytes_rem > 0)  /* must have found a differing byte */
        {
            out_str[out_index++] = (unsigned char)repetitions;
            out_str[out_index++] = current_byte;
            current_byte = in_str[in_index++];
            repetitions=0;
            bytes_rem--;
        }
    }
    out_str[out_index++] = (unsigned char)repetitions;
    out_str[out_index++] = current_byte;

    return (out_index);
}



/************************************************************************
 *
 *
 ************************************************************************/
int D03PutImage(x_pos_mil, y_pos_mil, imgP, mode, xform, cmap, attr, clip)
    int x_pos_mil;
    int y_pos_mil;
    ge_image_type *imgP;
    int mode;
    poly_type *xform;
    cmap_type *cmap;
    attr_type *attr;
    TmRect32 *clip ;
{
    int x1, y1, x2, y2 ;

    /* assert(imgP != NULL); */

    ASSERT( (imgP->data != NULL) && (imgP->wid != 0 && imgP->hyt != 0) );
    ASSERT( ((imgP->eps_seg == NULL) && (imgP->eps_seg_size == 0)) ||
            ((imgP->eps_seg != NULL) && (imgP->eps_seg_size != 0)));
    ASSERT( (imgP->dep == 1) || (imgP->dep == 8) );

    if ((imgP->data == NULL) && (imgP->wid && imgP->hyt))
        return(-1);

    if (imgP->dep != 1 && imgP->dep !=8)
        return(-1);

    clip_box(clip) ;

#if 0
    if (( (mode != TMS_O_D) && (mode != TMIS_O_D) ) || (imgP->dep > 1))
        /* opaque */
        PCL_command(NXLT("*v1N"));
    else
        /* transparent */
        PCL_command(NXLT("*v0N"));
#endif

    if (imgP->dep == 1)
    {
        if (mode != TMS_O_D && mode != TMIS_O_D)
            PCL_command(NXLT("*v1N"));
        else
            PCL_command(NXLT("*v0N"));
        D03MonoImage(x_pos_mil, y_pos_mil, imgP, xform, cmap, attr, clip);
    }
    else if (imgP->dep == 8)
    {
        if (PF.PfColordev)
        {
            if (pf_color_model != COLOR_MODEL_GRAY)
            {
                PCL_command(NXLT("*v0N"));
                D03ColorImage(x_pos_mil, y_pos_mil, imgP, xform, cmap, attr, clip);
            }
            else
            {
                PCL_command(NXLT("*p0P")) ;         /* push current palette */
                PCL_command(NXLT("*r1U")) ;         /* go into BW mode */
                if (ImageTransparent(imgP, cmap))
                    PCL_command(NXLT("*v0N"));
                else
                    PCL_command(NXLT("*v1N"));
                D03ColorAsMonoImage(x_pos_mil, y_pos_mil, imgP, xform, cmap, attr, clip);
                PCL_command(NXLT("*p1P")) ;         /* pop current palette */
            }
        }
        else
        {
            if (ImageTransparent(imgP, cmap))
                PCL_command(NXLT("*v0N"));
            else
                PCL_command(NXLT("*v1N"));
            D03ColorAsMonoImage(x_pos_mil, y_pos_mil, imgP, xform, cmap, attr, clip);
        }
    }
    else
        return(-1);
    return(0);
}



/*************************************************************************
 *
 *  If this image has any tranparent pixels return TRUE,
 * 
************************************************************************/
static int ImageTransparent(imgP, cmap)
    ge_image_type *imgP;
    cmap_type *cmap;
{
    int i, j ;
    unsigned char *dP ;
    unsigned char x ;
    int x_size, y_size;
    int data_width_bytes ;

    x_size = imgP->wid;
    y_size = imgP->hyt;
    data_width_bytes = imgP->data_byte_wid;

    /* determine if any are transparent */
    for (i = 0, dP = imgP->data ; i < y_size ; i++, dP += data_width_bytes)
        for (j = 0 ; j < x_size ; j++)
        {
            x = dP[j] ;
            if (cmap->color[x].see_thru)
                return TRUE ;
        }
    return FALSE ;
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03MonoImage(x_pos_mil, y_pos_mil, imgP, xform, cmap, attr, clip)
    int x_pos_mil;
    int y_pos_mil;
    ge_image_type *imgP;
    poly_type *xform;
    cmap_type *cmap;
    attr_type *attr;
    TmRect32 *clip ;
{
    int dw;
    int x_pos ;
    ge_image_transform_type tr;
    XImage *Xp = NULL;
    TmPoint32 bbox[2] ;
    int best_res = 300 ;
	int pen1, pen2 ;

    tr.x_scale = 1.0;
    tr.y_scale = 1.0;

    x_pos = xmiltopix(x_pos_mil);

    bbox[0].x = xmiltopix(x_pos_mil);
    bbox[0].y = ymiltopix(y_pos_mil);

    extract_factors_for_device(imgP, xform, &tr, PCL_CLASS_NUM);
    Xp = (XImage *)pf_tr_sys_mono_image(THIMpid, imgP, 0, &tr, &best_res);

    pen1 = find_pen(attr->ffg.c, attr->ffg.m, attr->ffg.y, attr->ffg.k) ;
    pen2 = find_pen(attr->fbg.c, attr->fbg.m, attr->fbg.y, attr->fbg.k) ;

    if (In_HPGL_Mode)
        SwitchTo_PCL() ;
    D03milSetPos(x_pos_mil, y_pos_mil);
    if (PF.PfColordev)
        D03MonoColorXImage(Xp, (XImage *)NULL, best_res, bbox, clip, x_pos, pen1, pen2);
    else
        D03MonoXImage(Xp, (XImage *)NULL, best_res, bbox, clip, x_pos);

    /* "use ESC*rB for all printers except IIIP, IIISi and 4." */

    PCL_command(NXLT("*t%dR"), TargetRes);  /* revert to 300 dpi (PCL5 Ref. 14-9) */

    pfFreeXImage(Xp);

    PF_MEM_CHECK;
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03ColorAsMonoImage(x_pos_mil, y_pos_mil, imgP, xform, cmap, attr, clip)
    int x_pos_mil;
    int y_pos_mil;
    ge_image_type *imgP;
    poly_type *xform;
    cmap_type *cmap;
    attr_type *attr;
    TmRect32 *clip ;
{
    char *data_src;
    int x_pos, y_pos;
    int x, y;
    int slice_height, num_slices, last_slice, cur_slice;
    ge_image_transform_type tr;
    XImage *Xgray = NULL;
    XImage *Xmono = NULL;
    XImage *Xp = NULL;
    short *FSbuffer;
    int FS_width;
    int best_res;
    TmPoint32 bbox[2] ;

    tr.x_scale = 1.0;
    tr.y_scale = 1.0;

    extract_factors_for_device(imgP, xform, &tr, PCL_CLASS_NUM);

    x_pos = xmiltopix(x_pos_mil);
    y_pos = ymiltopix(y_pos_mil);
    bbox[0].x = x_pos ;
    bbox[0].y = y_pos ;


    Xp = pf_tr_sys_color_image(THIMpid, imgP, cmap, &tr, &best_res);

/*
    Xp->bytes_per_line is always even, i.e. rounded up to
    16 bit alignment
    We somehow need to scale the image mask imP.mask... which
            can be considered a kind of mono-image... 
*/

    slice_height = Xp->height;
    last_slice = Xp->height;
    num_slices = 0;
    data_src = Xp->data;

    if (Xp->bytes_per_line * Xp->height > IMAGE_LIMIT)
    {
        slice_height = (IMAGE_LIMIT / Xp->bytes_per_line) & ~3;
        num_slices = Xp->height / slice_height;
        last_slice = Xp->height % slice_height;
    }

    FSbuffer = (short *)malloc(Xp->bytes_per_line * slice_height *
                                sizeof(short));

    Xgray = (XImage *)calloc(1, sizeof(XImage));
    Xgray->depth = 8;
    Xgray->width = Xp->width;
    Xgray->height = slice_height;
    Xgray->bytes_per_line = Xp->bytes_per_line;
    Xgray->data =
        (char *)malloc(Xgray->bytes_per_line * Xgray->height);

    for (cur_slice=0; cur_slice<num_slices; cur_slice++)
    {
        /* convert Xp->data into 0..255 _intensity_ values in an
        array of 16 bit values (so that FloydSteinbergDither() can
        work on the array) */
        data_src = &(Xp->data[Xp->bytes_per_line * (slice_height * cur_slice)]);

        /* Assume input data is padded to even byte boundary */
        ColorToGray(data_src, Xp->width, slice_height, cmap, FSbuffer, 16);
        
        /* go from FSbuffer[] intensities to Xmono->data which
        is aligned to 16 bits */
#define DITHER 0
        
#if DITHER
        Xmono = (XImage *)calloc(1, sizeof(XImage));
        Xmono->depth = 1;
        Xmono->width = Xp->width;
        Xmono->height = slice_height;
        Xmono->bytes_per_line = Xp->bytes_per_line;
        Xmono->data =
            (char *)malloc(Xmono->bytes_per_line * Xmono->height);

        FloydSteinbergDither(FSbuffer, Xmono->bytes_per_line, slice_height,
                             16, Xmono->data);
        
        /* some assumptions for tidyness... */
        Xmono->depth = 1;
        Xmono->bytes_per_line = EVEN_BYTES_FROM_BITS(Xp->width);
#else   /* halftoner */
        /* convert the FSbuffer into Xgray */

        memset(Xgray->data, 0, (Xgray->bytes_per_line * Xgray->height));
    
        FS_width = (Xp->width +1) & ~1; /* 16 bit aligned */
        for (y=0; y<Xgray->height; y++)
        {
            for (x=0; x<FS_width; x++)
            {
                Xgray->data[(y*Xp->bytes_per_line)+x] =
                    0xFF - (unsigned char)(FSbuffer[(y*FS_width)+x] & 0xFF);
            }
        }

        Xmono = HalfTone(Xgray);    /* Xgray is 8bit deep... */
#endif

        if (Xmono != NULL)
        {
            D03SetPos(x_pos, y_pos + (cur_slice * Xmono->height));

            D03MonoXImage(Xmono, NULL, TargetRes, bbox, clip, x_pos);
            bbox[0].y += slice_height ;
            pfFreeXImage(Xmono);
        }
    }

    /* remainder */

    data_src = &(Xp->data[Xp->bytes_per_line * (slice_height * cur_slice)]);

    ColorToGray(data_src, Xp->width, last_slice, cmap, FSbuffer, 16);
        
#define DITHER 0
#if DITHER
    Xmono = (XImage *)calloc(1, sizeof(XImage));
    Xmono->depth = 1;
    Xmono->width = Xp->width;
    Xmono->height = last_slice;
    Xmono->bytes_per_line = Xp->bytes_per_line;
    Xmono->data =
        malloc(Xmono->bytes_per_line * Xmono->height);

    FloydSteinbergDither(FSbuffer, Xmono->bytes_per_line, last_slice,
                         16, Xmono->data);

#else   /* halftoner */
    Xgray->height = last_slice;
    memset(Xgray->data, 0, (Xgray->bytes_per_line * Xgray->height));

    FS_width = (Xp->width +1) & ~1; /* 16 bit aligned */
    for (y=0; y<Xgray->height; y++)
    {
        for (x=0; x<FS_width; x++)
        {
            Xgray->data[(y*Xp->bytes_per_line)+x] =
                0xFF - (unsigned char)(FSbuffer[(y*FS_width)+x] & 0xFF);
        }
    }

    Xmono = HalfTone(Xgray);    /* Xgray is 8bit deep... */
    pfFreeXImage(Xgray);

#endif

    free(FSbuffer);
    FSbuffer = NULL;
    D03SetPos(x_pos, y_pos + (cur_slice * slice_height));
    D03MonoXImage(Xmono, NULL, TargetRes, bbox, clip, x_pos);   /* last_slice */
    pfFreeXImage(Xmono);
    PCL_command(NXLT("*t%dR"), TargetRes);  /* revert to 300 dpi (PCL5 Ref. 14-9) */
    pfFreeXImage(Xp);

    PF_MEM_CHECK;
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03ColorImage(x_pos_mil, y_pos_mil, imgP, xform, cmap, attr, clip)
    int x_pos_mil;
    int y_pos_mil;
    ge_image_type *imgP;
    poly_type *xform;
    cmap_type *cmap;
    attr_type *attr;
    TmRect32 *clip ;
{
    int x_pos, y_pos;
    int slice_height, num_slices, last_slice, cur_slice;
    ge_image_transform_type tr;
    XImage *Xp = NULL;
    XImage xS ;
    int best_res;
    TmPoint32 bbox[2] ;

    tr.x_scale = 1.0;
    tr.y_scale = 1.0;

    extract_factors_for_device(imgP, xform, &tr, PCL_CLASS_NUM);

    x_pos = xmiltopix(x_pos_mil);
    y_pos = ymiltopix(y_pos_mil);
    bbox[0].x = x_pos ;
    bbox[0].y = y_pos ;


    /* then the image data */
    Xp = pf_tr_sys_color_image(THIMpid, imgP, cmap, &tr, &best_res);

    PCL_command(NXLT("*p0P")) ;         /* push current palette*/
    create_palette(cmap, Xp->data, Xp->bytes_per_line, Xp->height) ;    /* and make a new one */

    if (In_HPGL_Mode)
        SwitchTo_PCL() ;

/*
    Xp->bytes_per_line is always even, i.e. rounded up to
    16 bit alignment
*/

    slice_height = Xp->height;
    last_slice = Xp->height;
    num_slices = 0;


    if (Xp->bytes_per_line * Xp->height > COLOR_IMAGE_LIMIT)
    {
        slice_height = (COLOR_IMAGE_LIMIT / Xp->bytes_per_line) & ~3;
        num_slices = Xp->height / slice_height;
        last_slice = Xp->height % slice_height;
    }
    xS.bytes_per_line = Xp->bytes_per_line ;
    xS.width = Xp->width ;
    xS.height = slice_height ;

    for (cur_slice = 0; cur_slice < num_slices ; cur_slice++)
    {
        xS.data = &(Xp->data[Xp->bytes_per_line * (slice_height * cur_slice)]);

        if (Xp != NULL)
        {
            D03SetPos(x_pos, y_pos + (cur_slice * slice_height));

            D03ColorXImage(&xS, TargetRes, bbox, clip, x_pos);
            bbox[0].y += slice_height ;
        }
    }

    /* remainder */
    xS.data = &(Xp->data[Xp->bytes_per_line * (slice_height * cur_slice)]);
    xS.height = last_slice ;
    D03SetPos(x_pos, y_pos + (cur_slice * slice_height));
    D03ColorXImage(&xS, TargetRes, bbox, clip, x_pos);  /* last_slice */

    PCL_command(NXLT("*t%dR"), TargetRes);  /* revert to 300 dpi (PCL5 Ref. 14-9) */
    pfFreeXImage(Xp);
    PCL_command(NXLT("*p1P")) ;             /* pop previous palette */
    PF_MEM_CHECK;
}

/************************************************************************
 *
 * for compatability
 *
 ************************************************************************/
int D03SetColor(color_request)
    short color_request;
{

    switch (color_request)
    {
    case TMA_WHITE:
        cur_grayness = 0.0;
        break;
    case TMA_RED:
        cur_grayness = IntensFromCMYKColor(0, 223, 223, 0, 256);
        break;
    case TMA_LBLUE:
        cur_grayness = IntensFromCMYKColor(256, 0, 0, 0, 256); 
        break;
    case TMA_GREEN:
        cur_grayness = IntensFromCMYKColor(189, 0, 161, 0, 256); 
        break;
    case TMA_PURPLE:
        cur_grayness = IntensFromCMYKColor(0, 223, 223, 0, 256);
        break; 
    case TMA_BLUE:
        cur_grayness = IntensFromCMYKColor(256, 76, 0, 0, 256); 
        break;
    case TMA_YELLOW:
        cur_grayness = IntensFromCMYKColor(0, 0, 256, 0, 256); 
        break;
    case TMA_DGRAY:
        cur_grayness = IntensFromCMYKColor(0, 0, 0, 128, 256); 
        break;
    case TMA_OLIVE:
        cur_grayness = IntensFromCMYKColor(163, 128, 161, 0, 256);
        break; 
    case TMA_LGRAY:
        cur_grayness = IntensFromCMYKColor(0, 0, 0, 64, 256);
        break; 
    case TMA_DRED: 
        cur_grayness = IntensFromCMYKColor(0, 189, 207, 94, 256);
        break;
    case TMA_ORANGE:
        cur_grayness = IntensFromCMYKColor(0, 97, 176, 0, 256); 
        break;
    case TMA_DBLUE:
        cur_grayness = IntensFromCMYKColor(256, 184, 0, 28, 256);
        break; 
    case TMA_BLACK:
    default:
        cur_grayness = 1.0;
        break;
    }
    return (int)color_request;
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03SetCMYKColor(c, m, y, k)
    int c, m, y, k;
{
    int p ;

    if (PF.PfColordev)
        cur_grayness = 1.0;
    else if ( c || m || y || k )
        cur_grayness = IntensFromCMYKColor(c, m, y, k, 255);
    else
        cur_grayness = 0.0;
}

/************************************************************************
 *
 *
 ************************************************************************/
static int find_pen(c, m, y, k)
    int c, m, y, k ;
{
    int i ;
    int red, green, blue ;
    int cyan, magenta, yellow ;

    /* if not color, return 0 for white and 1 for non-white */
    if (!PF.PfColordev)
    {
        D03SetCMYKColor(c, m, y, k) ;
        if (c == 0 && m == 0 && y == 0 && k == 0)
            return 0 ;
        else
            return 1 ;
    }

    cur_grayness = 1.0;

    /* look in the current palette */
    for (i = 0 ; i < palette_count ; i++)
    {
        if (palette[i].cyan == c &&
            palette[i].magenta == m &&
            palette[i].yellow == y &&
            palette[i].black == k)
            return i ;
    }

    /* palette count limited to 256: reuse last slot */
    if (i >= 256)
        i = 255 ;

    /* make a new pen */
    palette[i].cyan = c ;
    palette[i].magenta = m ;
    palette[i].yellow = y ;
    palette[i].black = k ;
    palette[i].color = (c > 0) || (m > 0) || (y > 0) ;

    cyan = MIN(c + k, 255);
    magenta = MIN(m + k, 255);
    yellow = MIN(y + k, 255);

    if (pf_color_model == COLOR_MODEL_RGB)
    {
        red = 255 - cyan ;
        green = 255 - magenta ;
        blue = 255 - yellow ;
        HPGL_command(NXLT("PC%d,%d,%d,%d"), i, red, green, blue) ;
    }
    else
    {
        HPGL_command(NXLT("PC%d,%d,%d,%d"), i, cyan, magenta, yellow) ;
    }
    palette_count++ ;
    return i ;
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03PreScanText()
{
    int segCount, piececount, width, txlen;
    int ix, eightBit, wideChar;
    int underPos, underThick, strikePos, xZoom, yZoom;
	int sdx, sdy, sfg, no_prnt, draw_shadow, shadow_only; 
	int shadow_type, draw_emboss;
    p32 pos;
    tx_char *textstr;
    textAttrs *txProps, fieldProps, stxProps, sfieldProps;
    attr_type *attr ;
    int equivalent_fill ;
    char *fontName;
    char *errString ;
    r32 clip ;
    fontTabEnt *ftab;
    PF_soft *soft ;
    axform sxf ;
	color_type new_color;

	shadow_type = InInt();
    InTxFieldProps(&fieldProps, &xZoom, &yZoom,
					&draw_shadow, &shadow_only, &draw_emboss);

    /* clipping rectangle */
    InBlock((char *)&clip, sizeof(clip)) ;

    segCount = InInt();             /* # mono-attr segments in this text */
    while (segCount--)
    {
        piececount = InInt();       /* # 'default-positioned' pieces in seg */
   
        /* read attr info for this seg */
        InTextAttrs(&txProps, &fontName, &underPos, &underThick, &strikePos);

		if ( shadow_type == 2 ) /* dark shadow */
		{
			no_prnt = axtGetShadowPrintInfo(txProps, draw_shadow, draw_emboss, shadow_type);
			if ( no_prnt )
			{
				sdx = 0;
				sdy = 0;
			}
			else
			{
				sdx = txProps->fillpattern->s_xoff;
				sdy = txProps->fillpattern->s_yoff;
				txProps->fillpattern->ffg = txProps->fillpattern->sfg;
				txProps->fillpattern->fbg = txProps->fillpattern->sfg;
				txProps->fillpattern->lfg = txProps->fillpattern->sfg;
				txProps->fillpattern->lbg = txProps->fillpattern->sfg;
				sfg = 0;
				sfg |= txProps->fillpattern->ffg.c << 24;
				sfg |= txProps->fillpattern->ffg.m << 16;
				sfg |= txProps->fillpattern->ffg.y << 8;
				sfg |= txProps->fillpattern->ffg.k;
				txProps->fillcolor = sfg;
			}
		}
		else if ( shadow_type == 1 ) /* emboss shadow */
		{
			no_prnt = axtGetShadowPrintInfo(txProps, draw_shadow, draw_emboss, shadow_type);
			if ( no_prnt )
			{
				sdx = 0;
				sdy = 0;
			}
			else if ( txProps->fillpattern->s_type < 3 )
			{
				sdx = 0;
				sdy = 0;
				no_prnt = TRUE;
			}
			else
			{
				sdx = (-1)*txProps->fillpattern->s_xoff;
				sdy = (-1)*txProps->fillpattern->s_yoff;

				GetWashoutColor(&txProps->fillpattern->sfg, &new_color);
				txProps->fillpattern->sfg.c = new_color.c;
				txProps->fillpattern->sfg.m = new_color.m;
				txProps->fillpattern->sfg.y = new_color.y;
				txProps->fillpattern->sfg.k = new_color.k;

				txProps->fillpattern->ffg = txProps->fillpattern->sfg;
				txProps->fillpattern->fbg = txProps->fillpattern->sfg;
				txProps->fillpattern->lfg = txProps->fillpattern->sfg;
				txProps->fillpattern->lbg = txProps->fillpattern->sfg;
				sfg = 0;
				sfg |= txProps->fillpattern->ffg.c << 24;
				sfg |= txProps->fillpattern->ffg.m << 16;
				sfg |= txProps->fillpattern->ffg.y << 8;
				sfg |= txProps->fillpattern->ffg.k;
				txProps->fillcolor = sfg;
			}
		}
		else if ( shadow_only == FALSE )/* normal text */
		{
			sdx = 0;
			sdy = 0;
			no_prnt = FALSE;
		}

        /* check for soft (download) font */
        if ((errString = axGetFontRecord(fontName, &ftab)) == NULL)
        {
#if AXMS
            if ( download_to_PCL( ftab ) )
#else
            if (ftab->gallium)
#endif
            {
                soft_xform(&sfieldProps, &fieldProps) ;
                soft_xform(&stxProps, txProps) ;
                pfGetTextSegTransform(&stxProps, &sfieldProps, xZoom, yZoom, &sxf);
                soft = pfSoftFontRegister(ftab, txProps->pointsize, &sxf.forward) ;
            }
            else
                soft = NULL ;
        }
        else
        {
            freeErrString(errString);
            soft = NULL ;
        }

        attr = txProps->fillpattern ;
        if (attr)
        {
            if (attr->fill_flag)
            {
                if (attr->f_type == GRAY_FILL_TYPE)
                {
                    equivalent_fill = getIXfromPercent(attr->f_pat);
                    pattern[equivalent_fill].needed = TRUE;
                }
                if ( attr->f_type == BUILTIN_FILL_TYPE)
                {
                    pattern[attr->f_pat].needed = TRUE;
                    pattern[attr->f_pat].hyt = attr->f_dhyt;
                    pattern[attr->f_pat].wid = attr->f_dwid;
                }
                if ( attr->f_type == USERDEF_FILL_TYPE)
                {
                    if (attr->f_data && attr->f_dhyt && attr->f_dwid)
                        pfSetPattern(attr->f_type,
                                    attr->f_pat,
                                    attr->f_dhyt,
                                    attr->f_dwid,
                                    attr->f_ddep,
                                    attr->f_data,
                                    attr->f_cmap);
                }
            }
        }
        eightBit = wideChar = FALSE;
        while (piececount--)
        {
            pos.x = InInt();
            pos.y = InInt();
            width = InInt();
            txlen = InTxCharString(&textstr);
            if (txProps->ruling || txProps->strike_thru) {
                InInt();
                InInt();
                InInt();
                InInt();
            }
            for (ix = 0; ix < txlen; ix++)
            {
                if (textstr[ix] > 0x7f)
                    eightBit = TRUE;
/*%%%%%%%%%%%%%% following test doesn't prove a thing */
                if (textstr[ix] > 0xff)
                    wideChar = TRUE;
                if (soft)
                    pfSoftCharRegister(soft, textstr[ix]) ;
            }
            free(textstr);
            textstr = NULL;
        }
        txProps->pointsize /= 10000;

        free(fontName);
        fontName = NULL;
        pfFreeTextAttrs(txProps, 1);
    }

    PF.fontNeeded = TRUE;
    PF_MEM_CHECK;
}

/************************************************************************
 *
 *
 ************************************************************************/
static void D03PreScan()
{
    int error_check;
    long mark;
    int y;
    int PageMilWidth = -1;
    int PageMilHeight = -1;
    register int i;
    register int j;
    register int x;
    int dep;
    int nCtrls = 0, nPts = 0, nPolys = 0, fill_flag;
    int byte_wid = 0;
    char buf[512];
    int hyt, wid, f_type, f_pat, l_type, l_pat;
    int equivalent_fill;
    int print_stats_pline = 0;
    int print_stats_text = 0;
    unsigned char *data;
    attr_type attr;
    TmRect32 clip ;
         
    extern long InMark();
    extern long InLong();


    mark = InMark();
 
    PF.PfJob_done = FALSE;
    while (!PF.PfJob_done)
    {
        switch (InWord ())
        {
        case SfCreate:
            if (PF.PfReverse)
                pf_markpage() ;
            PF.PfNumDocPages++;
            InWord();                       /* orientation */
            if (PageMilWidth == -1)
                PageMilWidth = InInt();     /* page width */
            else
                InInt();                    /* page width */
            if (PageMilHeight == -1)
                PageMilHeight = InInt();    /* page height */
            else
                InInt();                    /* page height */

            InInt();                        /* left margin */

            InInt();                        /* use crops */
            InInt();                        /* use registration marks */
            InInt();                        /* use bounder */
            InInt();                        /* bounder x */
            InInt();                        /* bounder y */
            InInt();                        /* bounder width */
            InInt();                        /* bounder height */
            InInt();                        /* clip x */
            InInt();                        /* clip y */
            InInt();                        /* clip width */
            InInt();                        /* clip height */
            InInt();                        /* input tray */
            InInt();                        /* output tray */
            InInt();                        /* duplex */
            break;

        case SfTerm:
            if (PF.PfReverse)
                pf_markpage() ;
            PF.PfJob_done = TRUE;
            break;

        case SfPolyline:
            print_stats_pline++;

            nPolys = InWord ();

            for ( i=0; i<nPolys; i++ )
            {
                /* read in the poly points */
                nPts = InWord ();
                for ( j=0; j<nPts; j++ )
                {
                    InInt ();
                    InInt ();
                }

                /* read in the control points */
                nCtrls = InWord ();
                for ( j=0; j<nCtrls; j++ )
                {
                    InInt ();
                    InInt ();
                }
            }

            memset((char *)&attr, 0, sizeof(attr_type));
            pf_read_poly_attr(&attr);

            /* clipping rectangle */
            InBlock((char *)&clip, sizeof(clip)) ;

			/* check for clpping info */
            nPolys = InWord ();

			if ( nPolys > 0 )
			{
				/* pass the points */
				for ( i=0; i<nPolys; i++ )
				{
					/* read in the poly points */
					nPts = InWord ();

					for ( j=0; j<nPts; j++ )
					{
						InInt ();
						InInt ();
					}
				
					/* read in the control points */
					nCtrls = InWord ();
					for ( j=0; j<nCtrls; j++ )
					{
						InInt ();
						InInt ();
					}
				}
			}
            if (attr.fill_flag)
            {
                switch (attr.f_type)
                {
                case GRAY_FILL_TYPE:
                    equivalent_fill = getIXfromPercent(attr.f_pat);
                    pattern[equivalent_fill].needed = TRUE ;
                    break ;

                case BUILTIN_FILL_TYPE:
                    pattern[attr.f_pat].needed = TRUE ;
                    pattern[attr.f_pat].hyt = attr.f_dhyt ;
                    pattern[attr.f_pat].wid = attr.f_dwid ;
                    break ;

                case USERDEF_FILL_TYPE:
                    pfSetPattern(attr.f_type,
                                attr.f_pat,
                                attr.f_dhyt,
                                attr.f_dwid,
                                attr.f_ddep,
                                attr.f_data,
                                attr.f_cmap);
                    break ;
                }

            }
            if (attr.line_flag)
            {
                switch (attr.l_type)
                {
                case GRAY_FILL_TYPE:
                    equivalent_fill = getIXfromPercent(attr.l_pat);
                    pattern[equivalent_fill].needed = TRUE ;
                    break ;

                case BUILTIN_FILL_TYPE:
                    pattern[attr.l_pat].needed = TRUE ;
                    pattern[attr.l_pat].hyt = attr.l_dhyt ;
                    pattern[attr.l_pat].wid = attr.l_dwid ;
                    break ;

                case USERDEF_FILL_TYPE:
                    pfSetPattern(attr.l_type,
                                attr.l_pat,
                                attr.l_dhyt,
                                attr.l_dwid,
                                attr.l_ddep,
                                attr.l_data,
                                attr.l_cmap);
                    break ;
                }
            }
            if (attr.f_data)
                free(attr.f_data);
            if (attr.l_data)
                free(attr.l_data);
            break;

        case SfArc:
#if 0
                /* read the ellipse data */
            InWord (); InWord (); InWord ();
            InWord (); InWord (); InWord ();
 
                /* read in the fill attributes */
            fill_flag = InWord();
            InWord();
            f_type = InWord();
            f_pat = InWord();
            if ( f_type == 0 ) /* built in fill */
                pat_needed[f_pat] = TRUE;
            InWord(); InWord(); InWord();
 
                /* read any passed in fill fill pattern */
            hyt = InWord();
            wid = InWord();
            byte_wid = (wid+7) / 8;
            data = (unsigned char *)malloc(hyt * byte_wid);
            for ( i=0; i<hyt; i++ )
            {
                for (j=0; j<byte_wid; j++ )
                {
                    data[i*byte_wid + j] = (unsigned char)InByte();
                }
            }
            InAlign();
            pfAddFillToList(f_type, hyt, wid, 1, data, NULL);
            free(data);
            data = NULL;
 
            /* read the line attributes */
            InWord(); InWord(); InWord();
            l_type = InWord ();
            l_pat = InWord ();
            if ( l_type == 0 ) /* built in fill */
                pat_needed[l_pat] = TRUE;
            InWord(); InWord(); InWord();

            /* read any passed in line fill pattern */
            hyt = InWord();
            wid = InWord();
            byte_wid = (wid+7) / 8;
            data = (unsigned char *)malloc(hyt * byte_wid);
            for ( i=0; i<hyt; i++ )
            {
                for (j=0; j<byte_wid; j++ )
                {
                    data[i*byte_wid + j] = (unsigned char)InByte();
                }
            }
            InAlign();
            pfAddFillToList(f_type, hyt, wid, 1, data, NULL);
            free(data);
            data = NULL;
                /* read the color attributes and their overrides */
            InWord (); InWord (); InWord (); InWord ();
            InWord (); InWord (); InWord (); InWord ();
            InWord (); InWord (); InWord (); InWord ();
            InWord (); InWord (); InWord (); InWord ();
            InWord (); InWord (); InWord (); InWord ();
#endif
            break;

        case SfSetWinAttr:
            InWord ();      /* line style */
            InWord ();      /* line thickness */
            InLong ();      /* text descriptor */
            InWord ();      /* transfer mode */
            InWord ();      /* text alignment */
            if (PF.PfSe_version >= 4)
            {
                InWord();   /* foreground color */
                InWord();   /* background color */
            }
            InWord ();      /* clip rect */
            InWord ();
            InWord ();
            InWord ();
            break;

        case SfGMode:
            InWord ();      /* On off flag */
            InWord ();      /* graphics rectangle */
            InWord ();
            InWord ();
            InWord ();
            break;

        case SfImage:
            InWord ();      /* pixel flag */
            InWord ();      /* image mode */
            InByte();       /* attr->ffg.c */
            InByte();       /* attr->ffg.m */
            InByte();       /* attr->ffg.y */
            InByte();       /* attr->ffg.k */
            InByte();       /* attr->fbg.c */
            InByte();       /* attr->fbg.m */
            InByte();       /* attr->fbg.y */
            InByte();       /* attr->fbg.k */
            InInt ();       /* xpos */
            InInt ();       /* ypos */
            x = InInt ();   /* xsize */
            i = InInt ();   /* ysize */
            dep = InWord ();    /* image_depth */

            InLong (); InLong ();
            InLong (); InLong (); 
            InLong (); InLong ();
            InLong (); InLong ();
            InLong (); InLong ();

#if BYTE_WID_IN_SE
            j = InWord ();x  /* image width in bytes */
#else
            j = EVEN_BYTES_FROM_BITS(x * dep);
#endif

            for (y=0; y<i; y++)
                for (x=0; x<j; x++)
                    InByte();
            /* MASK_IN_STREAM */
            InAlign();
            j = InWord();   /* BYTE wid; not bit! */
            if (j != 0)
                for (y=0; y<i; y++)
                    for (x=0; x<j; x++)
                        InByte();

            i = (unsigned int)InLong(); /* cmap->n_colors */
            for (y=0; y<i; y++)
            {
                InInt();
                InByte();   /* c */
                InByte();   /* m */
                InByte();   /* y */
                InByte();   /* k */
            }

            /* clipping rectangle */
            InBlock((char *)&clip, sizeof(clip)) ;

            /* Now for the EPS data */
            i = InLong();
            if (i > 0)
            {
                for (y=0; y<i; y++)
                    InByte();
                if (i & 1)
                    InByte();
            }
            break;

        case SfTabs:
            pf_tabs();
            break;

        case SfScaledTxt:
            InWord();       /* on/off */
            InWord();       /* imageBox[0].x */
            InWord();       /* imageBox[0].y */
            InWord();       /* imageBox[1].x */
            InWord();       /* imageBox[1].y */
            InWord();       /* imageBox[2].x */
            InWord();       /* imageBox[2].y */
            InWord();       /* imageBox[3].x */
            InWord();       /* imageBox[3].y */
            break;

        case SfInclude:
            InWord();       /* width */
            InWord();       /* height */
            InWord();
            InWord();
            InWord();
            InWord();
            InWord();
            InWord();
            InWord();
            InWord();
 
            InNString(buf);
            InNString(buf);
 
            InAlign();
            break;

        case SfText:
            D03PreScanText();
            break;

        case SfClip:
            InBlock((char *)&clip, sizeof(clip)) ;
            break;

        case SfPrControl:
            InInt() ;
            i = InInt() ;
            InBlock(buf, i) ;
            InAlign();
            break;
            
        default:
            break;
        }
 
        error_check = InWord ();
        if (error_check != SfEndCall)
        {
            Debugf ("Expected end of call, found %d\n", error_check);
            PFabort ("File out of synch - didn't read end of call");
        }
    }
    InGotoMark(mark);
    PF.PfJob_done = FALSE;

#if 0
    PF.PfPageMilWidth = PageMilWidth;
    PF.PfPageMilHeight = PageMilHeight;
    PF.PfPagePixHeight = ymiltogpix(PF.PfPageMilHeight);
    PF.PfPagePixWidth = xmiltogpix(PF.PfPageMilWidth);
    PF.pagetop = PF.PfPagePixHeight;
#endif

}   /* D03PreScan() */

/************************************************************************
 *
 *
 ************************************************************************/
int D03Save()
{
}

/************************************************************************
 *
 *
 ************************************************************************/
int
D03Restore()
{
    hpgl_IW(0.0, 0.0, 0.0, 0.0) ;
    HPGL_command(NXLT("LT;"));
    PF.cur_thick = -1;
    PF.cur_style = TM_SOLID;
}

/************************************************************************
 *
 *
 ************************************************************************/
static int D03Clip()
{
    if ( PF.left != PF.PfPage.appClip.origin.x ||
         PF.top != PF.PfPage.appClip.origin.y ||
         PF.right != PF.PfPage.appClip.corner.x ||
         PF.bottom != PF.PfPage.appClip.corner.y )
    {
        PF.left = PF.PfPage.appClip.origin.x;
        PF.top = PF.PfPage.appClip.origin.y;
        PF.right = PF.PfPage.appClip.corner.x;
        PF.bottom = PF.PfPage.appClip.corner.y;

        hpgl_IW(X_DOT_TO_PLU(PF.left),
                 Y_DOT_TO_PLU(PF.bottom),
                 X_DOT_TO_PLU(PF.right),
                 Y_DOT_TO_PLU(PF.top));

        return(TRUE);
    }
    return(FALSE);
}

/************************************************************************
 *
 *
 ************************************************************************/
int D03SetClip(clipRect, x, y, last_x, last_y)
    AxRectangle *clipRect;  /* current app clip */
    int x, y, last_x, last_y;   /* new appclip requested */
{
    /* set app clip */
    if ( (x != clipRect->origin.x) ||
         (y != clipRect->origin.y) ||
         (last_x != clipRect->corner.x) ||
         (last_y != clipRect->corner.y) )
    {
        PF.PfPage.appClip.origin.x = x;
        PF.PfPage.appClip.origin.y = y;
        PF.PfPage.appClip.corner.x = last_x;
        PF.PfPage.appClip.corner.y = last_y;

        PF.left = x;
        PF.top = y;
        PF.right = last_x;
        PF.bottom = last_y;

        hpgl_IW(X_DOT_TO_PLU(x),
                Y_DOT_TO_PLU(y),
                X_DOT_TO_PLU(last_x),
                Y_DOT_TO_PLU(last_y));
        return(TRUE);
    }
    return(FALSE);
}

/************************************************************************
 *
 * This sets the landscape mode for the first page
 *
 ************************************************************************/
int D03Landscape()
{
 

    if (PF.PfLandscape)
        PCL_command(NXLT("&l1O"));
    else
        PCL_command(NXLT("&l0O"));
 
    PCL_command(NXLT("&l0E"));  /* Top Margin, PCL5 Ref. 5-18 */
    PCL_command(NXLT("&a0L"));  /* Left Margin, PCL5 Ref. 5-15 */

    PCL_command(NXLT("*p0X"));  /* cursor position (PCL units) */
    PCL_command(NXLT("*p0Y"));
    PCL_command(NXLT("*c0T"));  /* Picture Frame Anchor Point */


    return(0);
}

/************************************************************************
 *
 * VARARGS
 *
 ************************************************************************/
static void hpgl_SV(va_alist)
va_dcl
{
    va_list argp;
    int screen_type;
    int option1, option2;

    va_start(argp);
    screen_type = va_arg(argp, int);

    switch (screen_type)
    {
    default:
    case 0:                 /* No screening */
        if (HPGL_VectorScreen != screen_type)
        {
            HPGL_VectorScreen = screen_type;
            HPGL_command(NXLT("SV;"));
        }
        break;
    case 1:
/*      HPGL_VectorScreen = screen_type; */
        option1 = va_arg(argp, int);
        if ((HPGL_VectorScreen != screen_type) ||
            (HPGL_GrayVectorValue != option1))
        {
            HPGL_VectorScreen = screen_type;
            HPGL_GrayVectorValue = option1;
            if (HPGL_GrayVectorValue <= 99)
                HPGL_command(NXLT("SV1,%d;"), HPGL_GrayVectorValue);
            else
                HPGL_command(NXLT("SV;"));
        }
        break;
    case 2:
        HPGL_VectorScreen = screen_type;
        option1 = va_arg(argp, int);
        option2 = va_arg(argp, int);
        HPGL_command(NXLT("SV2,%d,%d;"), option1, option2);
        break;
    case 21:
        HPGL_VectorScreen = screen_type;
        option1 = va_arg(argp, int);
        HPGL_command(NXLT("SV21,%d;"), option1);
        break;
    case 22:
        HPGL_VectorScreen = screen_type;
        option1 = va_arg(argp, int);
        HPGL_command(NXLT("SV22,%d;"), option1);
        break;
    }
    va_end(argp);

}

/************************************************************************
 *
 *
 ************************************************************************/
static void  hpgl_TR(transp)
    int transp;
{
    if (transp != HPGL_Transparent)
    {
        HPGL_command(NXLT("TR%d;"), transp ? 1 : 0);
        HPGL_Transparent = transp;
        return;
    }
}

/************************************************************************
 *
 *
 ************************************************************************/
static void hpgl_AC(xpos, ypos)
    int xpos;
    int ypos;
{
    if (xpos == 0 && ypos == 0)
        HPGL_command(NXLT("AC;"));
    else
        HPGL_command(NXLT("AC%.2f,%.2f;"), X_DOT_TO_PLU(xpos), Y_DOT_TO_PLU(ypos));
}

/************************************************************************
 *
 *
 ************************************************************************/
static void hpgl_LA(attribute, value)
    int attribute;
    int value;
{
    switch (attribute)
    {
    case 1:                                 /* Line Ends */
        if (value != HPGL_LineEnds) /* current value */
        {
            HPGL_LineEnds = value;
            HPGL_command(NXLT("LA1,%d;"), value);
        }
        break ;

    case 2:                                 /* Line Joins */
        if (value != HPGL_LineJoins)
        {
            HPGL_LineJoins = value;
            HPGL_command(NXLT("LA2,%d;"), value);
        }
        break ;

    case 3:                                 /* Miter Limit */
        if (value != HPGL_MiterLimit)
        {
            HPGL_MiterLimit = value;
            HPGL_command(NXLT("LA3,%d;"), value);
        }
        break ;
    }
}
/************************************************************************
 *
 *
 ************************************************************************/
static void hpgl_LT(type)
    int type ;
{
    if (type == HPGL_LineType)
        return ;

    HPGL_LineType = type ;
    switch (type)
    {
    case 0:
        HPGL_command(NXLT("LT;"));
        break ;
    case 1:
        HPGL_command(NXLT("LT1,%.4f,1;"), 4.00);
        break ;
    case 2:
        HPGL_command(NXLT("LT2,%.4f,1;"), 2.75) ;
        break ;
    case 3:
        HPGL_command(NXLT("LT3,%.4f,1;"), 7.50);
        break ;
    case 4:
        HPGL_command(NXLT("LT4,%.4f,1;"), 0.75);
        break ;
    case 5:
        HPGL_command(NXLT("LT5,%.4f,1;"), 9.00);
        break ;
    case 6:
        HPGL_command(NXLT("LT6,%.4f,1;"), 6.35);
        break ;
    }
}


/************************************************************************
 *
 * VARARGS
 *
 ************************************************************************/
static void hpgl_FT(va_alist)
va_dcl
{
    va_list argp;
    int filltype;
    int option1, option2;

    va_start(argp);
    filltype = va_arg(argp, int);
    HPGL_cfill = -1 ;

    switch (filltype)
    {
    default:
    case 0:
    case 1:
    case 2:
        if (HPGL_FillType != 1)
        {
            HPGL_FillType = 1;
            HPGL_GrayFillValue = 100;
            
            HPGL_command(NXLT("FT;"));
        }
        break;
    case 3:                 /* hatching */
        HPGL_FillType = filltype;
        option1 = va_arg(argp, int);
        option2 = va_arg(argp, int);
        HPGL_command(NXLT("FT3,%d,%d;"), option1, option2);
        break;
    case 4:                 /* cross-hatching */
        HPGL_FillType = filltype;
        option1 = va_arg(argp, int);
        option2 = va_arg(argp, int);
        HPGL_command(NXLT("FT4,%d,%d;"), option1, option2);
        break;
    case 10:                    /* gray shading */
        option1 = va_arg(argp, int);
        if ((HPGL_FillType != filltype) ||
            (HPGL_GrayFillValue != option1))
        {
            HPGL_FillType = filltype;
            HPGL_GrayFillValue = option1;
            if (HPGL_GrayFillValue <= 99)
                HPGL_command(NXLT("FT10,%d;"), HPGL_GrayFillValue);
            else
                HPGL_command(NXLT("FT;"));
        }                       /* else same, do nothing */
        break;
    case 11:                    /* HPGL user-defined pattern */
        HPGL_FillType = filltype;
        option1 = va_arg(argp, int);
        HPGL_command(NXLT("FT11,%d;"), option1);
        break;
    case 21:                    /* PCL cross-hatch */
        HPGL_FillType = filltype;
        option1 = va_arg(argp, int);
        HPGL_command(NXLT("FT21,%d;"), option1);
        break;
    case 22:                    /* PCL user-defined pattern */
        HPGL_FillType = filltype;
        option1 = va_arg(argp, int);
        HPGL_command(NXLT("FT22,%d;"), option1);
        break;
    }
    va_end(argp);

}

/************************************************************************

    hpgl_IW(llx, lly, urx, ury)

    Generate IW command (if necessary) for clip rectangle in PLUs

    For all zeros, cancel  clip rectangle
 ************************************************************************/
static void hpgl_IW(llx, lly, urx, ury)
    double llx, lly, urx, ury ;
{

    if (llx != HPGL_IW.x0 ||
        lly != HPGL_IW.y0 ||
        urx != HPGL_IW.x1 ||
        ury != HPGL_IW.y1)
    {
        if (llx == 0.0 && lly == 0.0 && urx == 0.0 && ury == 0.0)
            HPGL_command(NXLT("IW;")) ;
        else
            HPGL_command(NXLT("IW%.2f,%.2f,%.2f,%.2f;"), llx, lly, urx, ury) ;
        HPGL_IW.x0 = llx ;
        HPGL_IW.y0 = lly ;
        HPGL_IW.x1 = urx ;
        HPGL_IW.y1 = ury ;
    }
    return;
}


/************************************************************************

    hpgl_PM(m)

    Generate PM command

 ************************************************************************/
static void hpgl_PM(m)
    int m ;
{

    if (m == 0)
        HPGL_command(NXLT("PM;")) ;
    else
        HPGL_command(NXLT("PM%d;"), m) ;

}
/************************************************************************

    hpgl_PD(x, y)

    Generate PD command

 ************************************************************************/
static void hpgl_PD(x, y)
    double x, y ;
{
    HPGL_pen_up = FALSE ;
    HPGL_command(NXLT("PD%.2f,%.2f"), x, y) ;
}

/************************************************************************

    hpgl_PU


    Generate PU command

 ************************************************************************/
static void hpgl_PU()
{
    if (!HPGL_pen_up)
    {
        HPGL_pen_up = TRUE ;
        HPGL_command(NXLT("PU;")) ;
    }
}

/************************************************************************

    hpgl_SP


    Generate SP command

 ************************************************************************/
static void hpgl_SP(p)
{
    if (PF.PfColordev && p != HPGL_pen)
    {
        HPGL_pen = p ;
        HPGL_command(NXLT("SP%d;"), p) ;
    }
}

/************************************************************************
 *
 *
 ************************************************************************/
static void SwitchTo_HPGL()
{
    PfOutString(NXLT("\033%1B"));   /* Enter HP-GL/2 mode   */
    In_HPGL_Mode = TRUE;
}

/************************************************************************
 *
 *
 ************************************************************************/
static void SwitchTo_PCL()
{
    PfOutString(NXLT("\033%1A"));   /* retain current cursor pos. */
    In_HPGL_Mode = FALSE;
}




/************************************************************************
 *
 *
 ************************************************************************/
static void pf03SetTextTransform(txProps, xfTerms, soft)
    textAttrs *txProps;     /* we'll use font, size, slant, weight */
    axformTerms *xfTerms;       /* overall transform info */
{
    /* apply scale, xshear, and rotate terms */

    if (!soft && (xfTerms->xscale != 10000 || xfTerms->yscale != 10000)) {
            /* some scale factor is present; this
                requires a reference measurement;
                measure an untransformed "A" */
        axtext txt;
        unsigned int tflags;
        tx_char tch;
        double rwid, rht;

            /* need to get the cap height and the nominal
                width for the specified text font */
        memset((char *) &txt, 0, sizeof(txt));

            /* use a dummy character to provide some real text */
        tch = (tx_char) 'A'; /* what if no such codepoint? */
        insChars(&txt, &tch, 0, 1);
        tflags = TXfontFamily | TXpointsize | TXslant | TXweight;
        setTextAttrs(&txt, txProps, tflags, 0, 1);
        constructForPrint(&txt);    /* get target metrics info onto text */

            /* just read fconx->target directly, without worrying
                whether target could've been kicked out.. */
        /* use half the pointsize as the nominal charwidth */
        rwid = (double) mToMils(1000, txt.atts[0].fontAttr.ptsize) * 0.5;
        rht = (double) mToMils(txt.atts[0].fconx.target->capHeight,
                                txt.atts[0].fontAttr.ptsize);
        if (!rwid)
            rwid = (double) txt.items[0].width;
        if (!rht)
            rht = (double) txt.items[0].height;

            /* convert to cm */
        rwid = (rwid * 2.54) / 1000.0;
        rht = (rht * 2.54) / 1000.0;

            /* apply scaling to the reference meas */
        rwid *= (double) xfTerms->xscale / 10000.0;
        rht *= (double) xfTerms->yscale / 10000.0;

        HPGL_command(NXLT("SI%.3f,%.3f;"), rwid, rht);  /* charplot cell size */
        daxtext(&txt);
    }

    if (xfTerms->xshear) {
        double skew;
        if (xfTerms->yscale < 0)
            skew = (double) xfTerms->xshear / 10000.0;
        else
            skew = (double) -xfTerms->xshear / 10000.0;
        HPGL_command(NXLT("SL%.3f;"), skew);
    }

    if (xfTerms->angle) {
        double tangle, sinAng, cosAng;
        double radians;

        tangle = (double) xfTerms->angle / 10000.0;
        radians = (double) DEG_TO_RAD(tangle);
        sinAng = (double) sin(radians);
        cosAng = (double) cos(radians);
        
        if (fabs(sinAng) < 0.00005)
            sinAng = 0.0 ;
        if (fabs(cosAng) < 0.00005)
            cosAng = 0.0 ;
        HPGL_command(NXLT("DI%.4f,%.4f;"), cosAng, sinAng);
    }
}


/************************************************************************
 *
 *
 ************************************************************************/
static void pf03ResetTextTransform(xfTerms)
    axformTerms *xfTerms;       /* overall transform info */
{
    if (xfTerms->xscale != 10000 || xfTerms->yscale != 10000)
        HPGL_command(NXLT("SI;"));

    if (xfTerms->xshear)
        HPGL_command(NXLT("SL;"));

    if (xfTerms->angle)
        HPGL_command(NXLT("DI;"));
}



/************************************************************************
 *
 *
 ************************************************************************/
static void PCL_initialize()
{
#if 0
	/* this don't work for non-PJL printers, e.g., 3si */
    PCL_command(NXLT("%%-12345X@PJL\n"));
    PfOutString(NXLT("@PJL ENTER LANGUAGE=PCL\n")) ;
#endif
    PCL_command(NXLT("E"));             /* ESC E - Master reset */

/* shouldn't this be ((PfNumDocPages == 1) || Collate) ? */
#if 0
    if (PfJobToPSPrinter &&
        (PF.PfNumDocPages == 1) &&
        (PF.PfCopies > 1))
    {
        PF.PfPrintMultiCopiesInLang = TRUE; /* compat w. PS */
    }
#endif
#if 0
    if (PF.PfPrintMultiCopiesInLang > 1)
        PCL_command(NXLT("&l%dX"), PF.PfPrintMultiCopiesInLang);
#endif

    PCL_command(NXLT("&u%dD"), TargetRes); /* define PCL Units for ESC*p#X */

    PCL_command(NXLT("*t%dR"), TargetRes);
    D03PreLoadPatterns();

    PF.PfHardX = -1;
    PF.PfHardY = -1;
#if 0 
    PCL_command(NXLT("&l1H"));              /* Paper Source from printer
                                                specified tray */
#endif
    PCL_command(NXLT("&l0E"));              /* Top Margin, PCL5 Ref. 5-18 */
    PCL_command(NXLT("&a0L"));              /* Left Margin, PCL5 Ref. 5-15 */
    PCL_command(NXLT("&l0L"));              /* Perf. Skip, PCL5 Ref. 5-21 */

    /* See PCL5 Tech Ref. 2-9, 18-5 (in decipoints) 
     This should reflect the paper size... */
    PCL_command(NXLT("*c%dX"), (int)((PF.PfPageMilWidth * 72) / 100));
    PCL_command(NXLT("*c%dY"), (int)((PF.PfPageMilHeight * 72) / 100));

    PCL_command(NXLT("*p0X"));              /* cursor position (PCL units) */
    PCL_command(NXLT("*p0Y"));
    PCL_command(NXLT("*c0T"));              /* Picture Frame Anchor Point */

    if (PF.pi.duplex >= 0)
        select_plex(PF.pi.duplex) ;

    /* make the default color palette */
    palette_count = 0 ;

    if (PF.PfColordev)
    {
        palette[0].cyan = 0 ;               /* pen 0 is white */
        palette[0].magenta = 0 ;
        palette[0].yellow = 0 ;
        palette[0].black = 0 ;
        palette[0].color = FALSE ;
        palette[1] = palette[0] ;           /* pen 1 is black */
        palette[1].black = 255 ;
        PCL_command(NXLT("*v6W")) ;
        if (pf_color_model == COLOR_MODEL_RGB)
            PfOutByte(0) ;                      /* RGB color space */
        else
            PfOutByte(1) ;                      /* CMY color space */
        PfOutByte(1) ;                      /* color is indexed by pixel */
        PfOutByte(8) ;                      /* 8 bits per index => 256 colors */
        PfOutByte(8) ;                      /* 8 bits for each primary color */
        PfOutByte(8) ;
        PfOutByte(8) ;
        palette_count = 2 ;
        PCL_command(NXLT("*t9J")) ;         /* this dither looks good, but what .? */
    }
}

/************************************************************************
 *
 *
 ************************************************************************/
static void HPGL_initialize()
{
    char *s, *t ;

    /* comments, if profile says OK */
    s = GetProfileByStr(NXLT("aXprintHeaderComments"));
    HPGL_command(NXLT("CO\"\nCreator: Applixware\n"));
    while (s != NULL)
    {
        s = strchr(s, '<');
        if (s == NULL)
            break ; 
        s++ ;
        switch (*s)
        {
        case 'T':
            put(NXLT("Title: %s\n"), PF.PfDoc_name);
            break;
        case 'D':
            put(NXLT("CreationDate: %s\n"), PF.PfDoc_date);
            break;
        case 'U':
            put(NXLT("For: %s\n"), PF.PfUser_name);
            break;
        case 'H':
            put(NXLT("Host: %s\n"), PF.PfHost_name);
            break;
        case 'V':
            put(NXLT("Revision: %.3f (%s)\n"),
                (double)BUILD_REVISION, BUILD_VERSION_STR);
            break;
        default:
            t = strchr(s, '>') ;
            if (t != NULL)
            {               
                *t = 0 ;
                put(NXLT("%s\n"), s);
                *t = '>' ;
                s = t + 1 ;
            }
            break;
        }
    }
    put(NXLT("Pages: %d\n\""), PF.PfNumTotalPages);

    /* other HPGL stuff */
/*  HPGL_command(NXLT("IN;")); */
    HPGL_command(NXLT("SP1;")) ;
    HPGL_command(NXLT("CF;")) ;
    HPGL_command(NXLT("UL1,60,40;")) ;
    HPGL_command(NXLT("UL2,30,70;")) ;
    HPGL_command(NXLT("UL3,40,25,10,25;")) ;
    HPGL_command(NXLT("UL4,40,60;")) ;
    HPGL_command(NXLT("UL5,26,18,10,18,10,18;")) ;
    HPGL_command(NXLT("UL6,75,25;")) ;
    HPGL_command(NXLT("LT;")) ;
    HPGL_command(NXLT("TR0;")) ;
    HPGL_command(NXLT("WU0;")) ;
    HPGL_pen = 1 ;
    HPGL_pen_up = TRUE ;
    HPGL_cfill = 100 ;
    HPGL_Transparent = FALSE;
    HPGL_VectorScreen = 0;
    HPGL_LineType = 0;
    HPGL_LineEnds = 1;
    HPGL_LineJoins = 1;
    HPGL_MiterLimit = 5;
    HPGL_UsePolygonMode = FALSE;
    HPGL_IW.x0 = 0.0 ;
    HPGL_IW.x1 = 0.0 ;
    HPGL_IW.y0 = 0.0 ;
    HPGL_IW.y1 = 0.0 ;


    if (PF.PfColordev)
    {
        HPGL_command(NXLT("NP256;")) ;                  /* 256 pens */
        if (pf_color_model == COLOR_MODEL_RGB)
        {
            HPGL_command(NXLT("PC0,255,255,255;")) ;    /* pen 0: white */
            HPGL_command(NXLT("PC1,0,0,0;")) ;          /* pen 1: black */
        }
        else
        {
            HPGL_command(NXLT("PC0,0,0,0;")) ;          /* pen 0: white */
            HPGL_command(NXLT("PC1,255,255,255;")) ;    /* pen 1: black */
        }
    }
}

/************************************************************************
 *
 *
 ************************************************************************/
static void reset_line_ends(cap)
int cap ;
{
    /* reset the line ends */
    switch ( cap )
    {
    default:
    case TM_CAP_ROUND:
        hpgl_LA(1, 4) ;
        break ;
    case TM_CAP_BUTT:
        hpgl_LA(1, 1) ;
        break ;
    case TM_CAP_PROJECTING:
        hpgl_LA(1, 2) ;
        break ;
    }
}


/************************************************************************
 *
 *
 ************************************************************************/
static void  clip_box(clip)
    TmRect32 *clip ;
{
    if (clip != NULL && clip->x_size != 0 && clip->y_size != 0)
        hpgl_IW(X_MIL_TO_PLU(clip->x_position),
                Y_MIL_TO_PLU(clip->y_position + clip->y_size),
                X_MIL_TO_PLU(clip->x_position + clip->x_size),
                Y_MIL_TO_PLU(clip->y_position)) ;

	else
	    hpgl_IW(0.0, 0.0, 0.0, 0.0) ;
		
}

/************************************************************************
 *
 *
 ************************************************************************/
static double pen_width(w)
    int w ;
{
    double x = MIL_TO_MM(w) ;

    if (x < MINIMUM_PEN_WIDTH)
        x = MINIMUM_PEN_WIDTH ;
    return x ;
}

/************************************************************************
 *
 *
 ************************************************************************/
static void character_fill(percent)
    int percent ;
{
    if (percent != HPGL_cfill)
    {
        if (percent == 100)
            HPGL_command(NXLT("CF;"));
        else
        {
            hpgl_FT(10, percent);
            HPGL_command(NXLT("CF3,0;"));           /* shaded character Fill */
        }
        HPGL_cfill = percent ;
    }
}


/************************************************************************
 ************************************************************************/
static void create_palette(cmap, data, w, h)
    cmap_type *cmap;
    char *data ;
    int w, h ;
{
    int i, j, red, green, blue, cyan, magenta, yellow, k ;
    int need[256] ;
    
    PCL_command(NXLT("*v6W")) ;         /* make a palette with indexed 8-bit color */
    if (pf_color_model == COLOR_MODEL_RGB)
        PfOutByte(0) ;                  /* RGB */
    else
        PfOutByte(1) ;                  /* CMY */
    PfOutByte(1) ;                      /* color is indexed by pixel */
    PfOutByte(8) ;                      /* 8 bits per index => 256 colors */
    PfOutByte(8) ;                      /* 8 bits per component */
    PfOutByte(8) ;
    PfOutByte(8) ;

    /* initialize need[] to no */
    for (i = 0 ; i < 256 ; i++)
        need[i] = FALSE ;

    /* for every color actually needed in bitmap*/
    for (j = 0 ; j < h ; j++)
        for (i = 0 ; i < w ; i++)
        {
            k = data[j*w + i] & 0xff ;
            need[k] = TRUE ;
        }

    HPGL_command(NXLT("NP256;")) ;      /* 256 pens */
    for (i = 0 ; i < cmap->n_colors ; i++)
    {
        /* if color needed, make a pen */
        if (need[i])
        {
            cyan = cmap->color[i].c ;
            magenta = cmap->color[i].m ;
            yellow = cmap->color[i].y ;
            k = cmap->color[i].k ;

            /* compute cmy */
            cyan = MIN(255, cyan+k) ;
            magenta = MIN(255, magenta+k) ;
            yellow = MIN(255, yellow+k) ;

            /* compute rgb */
            red = 255 - cyan ;
            green = 255 - magenta ;
            blue = 255 - yellow ;

            /* here's a nice hack: white will be taken as transparent if transparency
               mode is "transparent", so if you want this white to be opaque, make
               it damn near white instead */
            if (!cmap->color[i].see_thru &&
                red == 255 && green == 255 && blue == 255)
            {
                red = green = blue = 254 ;
                cyan = magenta = yellow = 1 ;
            }

            if (pf_color_model == COLOR_MODEL_RGB)
                HPGL_command(NXLT("PC%d,%d,%d,%d"), i, red, green, blue) ;
            else
                HPGL_command(NXLT("PC%d,%d,%d,%d"), i, cyan, magenta, yellow) ;
        }
    }
}

/************************************************************************
 *
 * D03PrintName - a safety wrapper for making a
 * <fontTabEnt>->PCLPrintName reference
 *
 ************************************************************************/
static char * D03PrintName(fontRec)
    fontTabEnt *fontRec;
{
    if (!fontRec || !fontRec->PCLPrintName)
        return (NXLT("0s0b4101T"));            /* cg times */
    else
        return (fontRec->PCLPrintName);
}

/*-----------------------------------------------------------------------------
    D03SoftDownLoad

    Down load all the PCL fonts
-----------------------------------------------------------------------------*/
static void D03SoftDownLoad()
{
    int ix, jx, k, z ;
    char *err;
    PF_soft *s ;
    PF_soft *sl = PF.slist ;
    axpfs_ext *x ;
    fontMetrics *fm ;
    axletter *al;
    plettername pl;
    XCharStruct *xc ;
    fontTabEnt *ftab;
	int h, w ;
	char *zbits ;

    for (ix = 0 ; ix < PF.scount ; ix++)
    {
        s = &sl[ix] ;
        ftab = (fontTabEnt *)s->ftab ;
        err = axGetTargetMetrics(ftab->targetMetricsNames[0], ftab, &fm);
        if (err)
        {
            freeErrString(err);
            break ;
        }

        /* down load font header */
        PCL_font_header.size = htons(sizeof(PCL_font_header)) ;
#if 0
        PCL_font_header.format = 20 ;
#else
        PCL_font_header.format = 00 ;
#endif
        PCL_font_header.type = 2 ;
        PCL_font_header.style_MSB = 0 ;
        PCL_font_header.reserved_1 = 0 ;
        PCL_font_header.baseline = htons(mToPix(fm->xfont->max_bounds.ascent,
                            ((float)s->pointsize)/10000.0, 300)) ;
        PCL_font_header.cell_width = htons(mToPix(fm->xfont->max_bounds.width,
                            ((float)s->pointsize)/10000.0, 300)) ;
        PCL_font_header.cell_height = htons(mToPix(fm->xfont->max_bounds.ascent +
                            fm->xfont->max_bounds.descent,
                            ((float)s->pointsize)/10000.0, 300)) ;
        PCL_font_header.orientation = 0 ;
        PCL_font_header.spacing = 1 ;
        PCL_font_header.symbol_set = htons(14) ;
        PCL_font_header.pitch = htons(0) ;
        PCL_font_header.height = htons(0) ;
        PCL_font_header.x_height = htons(0) ;
        PCL_font_header.width_type = 0 ;
        PCL_font_header.style_LSB = 0 ;
        PCL_font_header.stroke_weight = 0 ;
        PCL_font_header.typeface_LSB = 0 ;
        PCL_font_header.typeface_MSB = 0 ;
        PCL_font_header.serif_style = 0;
        PCL_font_header.quality = 2 ;
        PCL_font_header.placement = 0 ;
        PCL_font_header.underline_position = 0 ;
        PCL_font_header.underline_thickness = 0 ;
        PCL_font_header.text_height = htons(0) ;
        PCL_font_header.text_width = htons(0) ;
        PCL_font_header.first_code = htons(0) ;
        PCL_font_header.last_code = htons(255) ;
        PCL_font_header.pitch_extended = 0 ;
        PCL_font_header.height_extended = 0 ;
        PCL_font_header.cap_height = htons(0) ;
        PCL_font_header.font_number = htonl(0) ;
#if 0
        PCL_font_header.font_name = 0 ;
#endif
#if 0
        PCL_font_header.X_resolution = 300 ;
        PCL_font_header.Y_resolution = 300 ;
#endif

        /* get the width of a space char to use as the pitch */
        xc = getCharStruct(' ', fm->xfont) ;
#if 0
        PCL_font_header.pitch = 4*mToPix(xc->width,
                            ((float)s->pointsize)/10000.0, 300) ;
#endif
        PCL_font_header.pitch = htons(4*mToPix((int)(xc->width * extractXscale(&(s->ctm))),
                            ((float)s->pointsize)/10000.0, 300)) ;

        /* mail the font header */
        PCL_command(NXLT("*c%dD"), s->id) ;
        PCL_command(NXLT(")s%dW"), sizeof(PCL_font_header)) ;
        PfOutCString((char *)&PCL_font_header, sizeof(PCL_font_header)) ;

        /* find all the characters we need */
#if AXMS
        /* if the soft-font structures have not bee set up yet, do so now */
        if ( !fm->xfont->ext_data )
            axgfsAddFontExtension( fm->xfont, TargetRes, AXGFS_TARGET_POINTSIZE,
                                                ftab->targetMetricsNames[0], ftab );
#endif
        x = (axpfs_ext *) fm->xfont->ext_data->private_data;
        for (jx = 0 ; jx < 256 ; jx++)
        {
            if (s->needed[jx])
            {
				memset((void *) &pl, 0, sizeof(pl));
                pl.charcode = (unsigned short) jx;
                pl.pfid.pointsize = s->pointsize / 10000;
                pl.pfid.ctm = s->ctm;
                axhGetItemI(&pl, sizeof(plettername), &x->com.letters,
                            (char *) fm->xfont, (char **)&al);

                PCL_command(NXLT("*c%dE"), jx) ;

                /* character descriptor */
                xc = getCharStruct(jx, fm->xfont) ;
                PCLchar.format = 4 ;
                PCLchar.continuation = 0 ;
                PCLchar.size = 14 ;
                PCLchar.class = 1 ;
                PCLchar.orientation = 0 ;
                PCLchar.reserved_1 = 0 ;
                if (al == NULL || al->bits == NULL)
                {
                    PCLchar.left_offset = htons(0) ;
                    PCLchar.top_offset = htons(0) ;
					w = h = 0 ;
                    PCLchar.width = htons(w) ;
                    PCLchar.height = htons(h) ;
                    PCLchar.delta_x = htons(0) ;
                }
                else
                {
                    PCLchar.left_offset = htons(al->bbox->origin.x) ;
                    PCLchar.top_offset = htons(-al->bbox->origin.y) ;
                    w = al->bbox->width ;
                    PCLchar.width = htons(w) ;
                    h = al->bbox->height ;
                    PCLchar.height = htons(h) ;

#if 0
                    PCLchar.delta_x = 4*mToPix(xc->width,
                                ((float)s->pointsize)/10000.0, 300) ;
                    PCLchar.delta_x = 4*al->width ;
#endif
                    PCLchar.delta_x = htons((int)(extractXscale(&(s->ctm)) * 4 *
                                mToPix(xc->width, ((float)s->pointsize)/10000.0, 300))) ;
                }
                k = h * ((w + 7) / 8) ;
				if ((k + sizeof(PCL_char)) <= 32767)
				{
	                PCL_command(NXLT("(s%dW"), k + sizeof(PCL_char)) ;
	                PfOutCString((char *)&PCLchar, sizeof(PCL_char)) ;
					if (al)
		                PfOutCString(al->bits, k) ;
				}
				else
				{
					zbits = (char *)al->bits ;
					z = 32767 - sizeof(PCL_char) ;
	                PCL_command(NXLT("(s%dW"), z + sizeof(PCL_char)) ;
	                PfOutCString((char *)&PCLchar, sizeof(PCL_char)) ;
	                PfOutCString(zbits, z) ;
					zbits += z ;
					k -= z ;
	                PCLcharcont.format = 4 ;
	                PCLcharcont.continuation = 1 ;
					while ((k + sizeof(PCL_char_cont)) > 32767)
					{
		                PCL_command(NXLT("(s%dW"), z + sizeof(PCL_char_cont)) ;
		                PfOutCString((char *)&PCLcharcont, sizeof(PCL_char_cont)) ;
		                PfOutCString(zbits, z) ;
						zbits += z ;
						k -= z ;
					}
					if (k > 0)
					{
		                PCL_command(NXLT("(s%dW"), k + sizeof(PCL_char_cont)) ;
		                PfOutCString((char *)&PCLcharcont, sizeof(PCL_char_cont)) ;
		                PfOutCString(zbits, k) ;
					}
				}
            }
        }
    }
}

/*-----------------------------------------------------------------------------
    soft_xform

    Create a textAttrs structure suitable for computing the
    transform matrix for soft fonts
-----------------------------------------------------------------------------*/
static void soft_xform(st, ft)
    textAttrs *st, *ft ;
{
    int angle ;

    *st = *ft ;

    /* different strokes ... */
    st->xshear = -st->xshear ;


    /* reduce angle to 0-360 */
    angle = st->angle ;
    while (angle > 3600000)
        angle -= 3600000 ;
    while (angle < 0)
        angle += 3600000 ;

    /* compensate for PCL's idea to rotate bit-mapped fonts */
    if (angle > 450000 && angle <= 1350000)
        angle -= 900000 ;
    else if (angle > 1350000 && angle < 2250000)
        angle -= 1800000 ;
    else if (angle >= 2250000 && angle <= 3150000)
        angle -= 2700000 ;

    /* different strokes ... */
    angle = -angle ;

    /* reduce again, for ease of thinking about it */
    while (angle > 3600000)
        angle -= 3600000 ;
    while (angle < 0)
        angle += 3600000 ;

    st->angle = angle ;

}

/*-----------------------------------------------------------------------------
	extractXscale

-----------------------------------------------------------------------------*/
static float extractXscale(ctm)
    axctm *ctm; /* composite forward transform */
{
    p32 inp, outp;
    axform targetCtm;
#if 0
    if (targetCtm->unity) {
        *xscale = 1.0;
        return;
    }
#endif
    inp.x = 100000;
    inp.y = 0;
	targetCtm.unity = FALSE ;
    targetCtm.forward = *ctm ;
    forwardSXRT(inp.x, inp.y, &outp.x, &outp.y, &targetCtm);
    return (float) Axhypot((double) outp.x, (double) outp.y) / 100000.0;
}
